/* Generated by Cython 0.29.13 */

/* BEGIN: Cython Metadata
{
    "distutils": {
        "define_macros": [
            [
                "NPY_NO_DEPRECATED_API",
                "0"
            ]
        ],
        "depends": [
            "/opt/conda/lib/python3.7/site-packages/numpy/core/include/numpy/arrayobject.h",
            "/opt/conda/lib/python3.7/site-packages/numpy/core/include/numpy/ndarrayobject.h",
            "/opt/conda/lib/python3.7/site-packages/numpy/core/include/numpy/ndarraytypes.h",
            "/opt/conda/lib/python3.7/site-packages/numpy/core/include/numpy/npy_common.h",
            "/opt/conda/lib/python3.7/site-packages/numpy/core/include/numpy/ufuncobject.h",
            "pandas/_libs/intervaltree.pxi.in",
            "pandas/_libs/src/headers/stdint.h",
            "pandas/_libs/src/klib/khash_python.h"
        ],
        "extra_compile_args": [
            "-Wno-unused-function"
        ],
        "include_dirs": [
            "./pandas/_libs",
            "pandas/_libs/src/klib",
            "pandas/_libs/src",
            "/opt/conda/lib/python3.7/site-packages/numpy/core/include"
        ],
        "language": "c",
        "name": "pandas._libs.interval",
        "sources": [
            "pandas/_libs/interval.pyx"
        ]
    },
    "module_name": "pandas._libs.interval"
}
END: Cython Metadata */

#define PY_SSIZE_T_CLEAN
#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02060000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03030000)
    #error Cython requires Python 2.6+ or Python 3.3+.
#else
#define CYTHON_ABI "0_29_13"
#define CYTHON_HEX_VERSION 0x001D0DF0
#define CYTHON_FUTURE_DIVISION 1
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef HAVE_LONG_LONG
  #if PY_VERSION_HEX >= 0x02070000
    #define HAVE_LONG_LONG
  #endif
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#ifdef PYPY_VERSION
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#elif defined(PYSTON_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 0
  #elif !defined(CYTHON_USE_PYTYPE_LOOKUP)
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #if PY_MAJOR_VERSION < 3
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 0
  #elif !defined(CYTHON_USE_PYLONG_INTERNALS)
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if PY_VERSION_HEX < 0x030300F0
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #ifndef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL 1
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT (PY_VERSION_HEX >= 0x03050000)
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE (PY_VERSION_HEX >= 0x030400a1)
  #endif
  #ifndef CYTHON_USE_DICT_VERSIONS
    #define CYTHON_USE_DICT_VERSIONS (PY_VERSION_HEX >= 0x030600B1)
  #endif
  #ifndef CYTHON_USE_EXC_INFO_STACK
    #define CYTHON_USE_EXC_INFO_STACK (PY_VERSION_HEX >= 0x030700A3)
  #endif
#endif
#if !defined(CYTHON_FAST_PYCCALL)
#define CYTHON_FAST_PYCCALL  (CYTHON_FAST_PYCALL && PY_VERSION_HEX >= 0x030600B1)
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #include "longintrepr.h"
  #undef SHIFT
  #undef BASE
  #undef MASK
  #ifdef SIZEOF_VOID_P
    enum { __pyx_check_sizeof_voidp = 1 / (int)(SIZEOF_VOID_P == sizeof(void*)) };
  #endif
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_MAYBE_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_MAYBE_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#ifdef _MSC_VER
    #ifndef _MSC_STDINT_H_
        #if _MSC_VER < 1300
           typedef unsigned char     uint8_t;
           typedef unsigned int      uint32_t;
        #else
           typedef unsigned __int8   uint8_t;
           typedef unsigned __int32  uint32_t;
        #endif
    #endif
#else
   #include <stdint.h>
#endif
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus) && __cplusplus >= 201103L
    #if __has_cpp_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH [[fallthrough]]
    #elif __has_cpp_attribute(clang::fallthrough)
      #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
    #elif __has_cpp_attribute(gnu::fallthrough)
      #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__ ) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif

#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #elif defined(__GNUC__)
    #define CYTHON_INLINE __inline__
  #elif defined(_MSC_VER)
    #define CYTHON_INLINE __inline
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_INLINE inline
  #else
    #define CYTHON_INLINE
  #endif
#endif

#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x02070600 && !defined(Py_OptimizeFlag)
  #define Py_OptimizeFlag 0
#endif
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyClass_Type
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
#if PY_VERSION_HEX >= 0x030800A4 && PY_VERSION_HEX < 0x030800B2
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, 0, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#else
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#endif
  #define __Pyx_DefaultClassType PyType_Type
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef METH_STACKLESS
  #define METH_STACKLESS 0
#endif
#if PY_VERSION_HEX <= 0x030700A3 || !defined(METH_FASTCALL)
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #define __Pyx_PyCFunctionFast _PyCFunctionFast
  #define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
#endif
#if CYTHON_FAST_PYCCALL
#define __Pyx_PyFastCFunction_Check(func)\
    ((PyCFunction_Check(func) && (METH_FASTCALL == (PyCFunction_GET_FLAGS(func) & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)))))
#else
#define __Pyx_PyFastCFunction_Check(func) 0
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030400A1
  #define PyMem_RawMalloc(n)           PyMem_Malloc(n)
  #define PyMem_RawRealloc(p, n)       PyMem_Realloc(p, n)
  #define PyMem_RawFree(p)             PyMem_Free(p)
#endif
#if CYTHON_COMPILING_IN_PYSTON
  #define __Pyx_PyCode_HasFreeVars(co)  PyCode_HasFreeVars(co)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) PyFrame_SetLineNumber(frame, lineno)
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if !CYTHON_FAST_THREAD_STATE || PY_VERSION_HEX < 0x02070000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x03060000
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#elif PY_VERSION_HEX >= 0x03000000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_Current
#endif
#if PY_VERSION_HEX < 0x030700A2 && !defined(PyThread_tss_create) && !defined(Py_tss_NEEDS_INIT)
#include "pythread.h"
#define Py_tss_NEEDS_INIT 0
typedef int Py_tss_t;
static CYTHON_INLINE int PyThread_tss_create(Py_tss_t *key) {
  *key = PyThread_create_key();
  return 0;
}
static CYTHON_INLINE Py_tss_t * PyThread_tss_alloc(void) {
  Py_tss_t *key = (Py_tss_t *)PyObject_Malloc(sizeof(Py_tss_t));
  *key = Py_tss_NEEDS_INIT;
  return key;
}
static CYTHON_INLINE void PyThread_tss_free(Py_tss_t *key) {
  PyObject_Free(key);
}
static CYTHON_INLINE int PyThread_tss_is_created(Py_tss_t *key) {
  return *key != Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE void PyThread_tss_delete(Py_tss_t *key) {
  PyThread_delete_key(*key);
  *key = Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE int PyThread_tss_set(Py_tss_t *key, void *value) {
  return PyThread_set_key_value(*key, value);
}
static CYTHON_INLINE void * PyThread_tss_get(Py_tss_t *key) {
  return PyThread_get_key_value(*key);
}
#endif
#if CYTHON_COMPILING_IN_CPYTHON || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#if PY_MAJOR_VERSION >= 3 || CYTHON_FUTURE_DIVISION
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1 && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStr(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
#else
#define __Pyx_PyDict_GetItemStr(dict, name)  PyDict_GetItem(dict, name)
#endif
#if PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                              0 : _PyUnicode_Ready((PyObject *)(op)))
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         PyUnicode_KIND(u)
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
#else
  #define CYTHON_PEP393_ENABLED 0
  #define PyUnicode_1BYTE_KIND  1
  #define PyUnicode_2BYTE_KIND  2
  #define PyUnicode_4BYTE_KIND  4
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((sizeof(Py_UNICODE) == 2) ? 65535 : 1114111)
  #define __Pyx_PyUnicode_KIND(u)         (sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  (((void)(k)), ((Py_UNICODE*)d)[i] = ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_SIZE(u))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_Contains)
  #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyByteArray_Check)
  #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Format)
  #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None || (PyString_Check(b) && !PyString_CheckExact(b)))) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None || (PyUnicode_Check(b) && !PyUnicode_CheckExact(b)))) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION < 3 && !defined(PyObject_ASCII)
  #define PyObject_ASCII(o)            PyObject_Repr(o)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
  #define PyObject_Unicode             PyObject_Str
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        (Py_TYPE(obj) == &PySet_Type)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
#else
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   PyInt_AsLong
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   PyInt_AsSsize_t
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyMethod_New(func, self, klass) ((self) ? PyMethod_New(func, self) : (Py_INCREF(func), func))
#else
  #define __Pyx_PyMethod_New(func, self, klass) PyMethod_New(func, self, klass)
#endif
#if CYTHON_USE_ASYNC_SLOTS
  #if PY_VERSION_HEX >= 0x030500B1
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
    #define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
  #else
    #define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
  #endif
#else
  #define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef __Pyx_PyAsyncMethodsStruct
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
    } __Pyx_PyAsyncMethodsStruct;
#endif

#if defined(WIN32) || defined(MS_WINDOWS)
  #define _USE_MATH_DEFINES
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif


#define __PYX_ERR(f_index, lineno, Ln_error) \
{ \
  __pyx_filename = __pyx_f[f_index]; __pyx_lineno = lineno; __pyx_clineno = __LINE__; goto Ln_error; \
}

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__pandas___libs__interval
#define __PYX_HAVE_API__pandas___libs__interval
/* Early includes */
#include <string.h>
#include <stdio.h>
#include "numpy/arrayobject.h"
#include "numpy/ufuncobject.h"
#include "pythread.h"

    PyObject* char_to_string(const char* data) {
    #if PY_VERSION_HEX >= 0x03000000
        return PyUnicode_FromString(data);
    #else
        return PyString_FromString(data);
    #endif
    }
    
#include "numpy/ndarrayobject.h"
#include "numpy/npy_common.h"
#include "numpy/ndarraytypes.h"
#include "src/headers/stdint.h"
#include "khash_python.h"
#include <stdlib.h>
#include "pystate.h"
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

typedef struct {PyObject **p; const char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_UTF8 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT (PY_MAJOR_VERSION >= 3 && __PYX_DEFAULT_STRING_ENCODING_IS_UTF8)
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
static CYTHON_INLINE int __Pyx_is_valid_index(Py_ssize_t i, Py_ssize_t limit) {
    return (size_t) i < (size_t) limit;
}
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
#define __Pyx_PyByteArray_FromString(s) PyByteArray_FromStringAndSize((const char*)s, strlen((const char*)s))
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyObject_AsWritableString(s)    ((char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const Py_UNICODE *u) {
    const Py_UNICODE *u_end = u;
    while (*u_end++) ;
    return (size_t)(u_end - u - 1);
}
#define __Pyx_PyUnicode_FromUnicode(u)       PyUnicode_FromUnicode(u, __Pyx_Py_UNICODE_strlen(u))
#define __Pyx_PyUnicode_FromUnicodeAndLength PyUnicode_FromUnicode
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
#if CYTHON_ASSUME_SAFE_MACROS
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#else
#define __Pyx_PyNumber_Int(x) (PyInt_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Int(x))
#endif
#define __Pyx_PyNumber_Float(x) (PyFloat_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Float(x))
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c) + 1);
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }

static PyObject *__pyx_m = NULL;
static PyObject *__pyx_d;
static PyObject *__pyx_b;
static PyObject *__pyx_cython_runtime = NULL;
static PyObject *__pyx_empty_tuple;
static PyObject *__pyx_empty_bytes;
static PyObject *__pyx_empty_unicode;
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm= __FILE__;
static const char *__pyx_filename;

/* Header.proto */
#if !defined(CYTHON_CCOMPLEX)
  #if defined(__cplusplus)
    #define CYTHON_CCOMPLEX 1
  #elif defined(_Complex_I)
    #define CYTHON_CCOMPLEX 1
  #else
    #define CYTHON_CCOMPLEX 0
  #endif
#endif
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    #include <complex>
  #else
    #include <complex.h>
  #endif
#endif
#if CYTHON_CCOMPLEX && !defined(__cplusplus) && defined(__sun__) && defined(__GNUC__)
  #undef _Complex_I
  #define _Complex_I 1.0fj
#endif


static const char *__pyx_f[] = {
  "pandas/_libs/interval.pyx",
  "pandas/_libs/intervaltree.pxi",
  "stringsource",
  "__init__.pxd",
  "pandas/_libs/tslibs/util.pxd",
  "pandas/_libs/util.pxd",
  "type.pxd",
  "bool.pxd",
  "complex.pxd",
  "pandas/_libs/hashtable.pxd",
};
/* MemviewSliceStruct.proto */
struct __pyx_memoryview_obj;
typedef struct {
  struct __pyx_memoryview_obj *memview;
  char *data;
  Py_ssize_t shape[8];
  Py_ssize_t strides[8];
  Py_ssize_t suboffsets[8];
} __Pyx_memviewslice;
#define __Pyx_MemoryView_Len(m)  (m.shape[0])

/* Atomics.proto */
#include <pythread.h>
#ifndef CYTHON_ATOMICS
    #define CYTHON_ATOMICS 1
#endif
#define __pyx_atomic_int_type int
#if CYTHON_ATOMICS && __GNUC__ >= 4 && (__GNUC_MINOR__ > 1 ||\
                    (__GNUC_MINOR__ == 1 && __GNUC_PATCHLEVEL >= 2)) &&\
                    !defined(__i386__)
    #define __pyx_atomic_incr_aligned(value, lock) __sync_fetch_and_add(value, 1)
    #define __pyx_atomic_decr_aligned(value, lock) __sync_fetch_and_sub(value, 1)
    #ifdef __PYX_DEBUG_ATOMICS
        #warning "Using GNU atomics"
    #endif
#elif CYTHON_ATOMICS && defined(_MSC_VER) && 0
    #include <Windows.h>
    #undef __pyx_atomic_int_type
    #define __pyx_atomic_int_type LONG
    #define __pyx_atomic_incr_aligned(value, lock) InterlockedIncrement(value)
    #define __pyx_atomic_decr_aligned(value, lock) InterlockedDecrement(value)
    #ifdef __PYX_DEBUG_ATOMICS
        #pragma message ("Using MSVC atomics")
    #endif
#elif CYTHON_ATOMICS && (defined(__ICC) || defined(__INTEL_COMPILER)) && 0
    #define __pyx_atomic_incr_aligned(value, lock) _InterlockedIncrement(value)
    #define __pyx_atomic_decr_aligned(value, lock) _InterlockedDecrement(value)
    #ifdef __PYX_DEBUG_ATOMICS
        #warning "Using Intel atomics"
    #endif
#else
    #undef CYTHON_ATOMICS
    #define CYTHON_ATOMICS 0
    #ifdef __PYX_DEBUG_ATOMICS
        #warning "Not using atomics"
    #endif
#endif
typedef volatile __pyx_atomic_int_type __pyx_atomic_int;
#if CYTHON_ATOMICS
    #define __pyx_add_acquisition_count(memview)\
             __pyx_atomic_incr_aligned(__pyx_get_slice_count_pointer(memview), memview->lock)
    #define __pyx_sub_acquisition_count(memview)\
            __pyx_atomic_decr_aligned(__pyx_get_slice_count_pointer(memview), memview->lock)
#else
    #define __pyx_add_acquisition_count(memview)\
            __pyx_add_acquisition_count_locked(__pyx_get_slice_count_pointer(memview), memview->lock)
    #define __pyx_sub_acquisition_count(memview)\
            __pyx_sub_acquisition_count_locked(__pyx_get_slice_count_pointer(memview), memview->lock)
#endif

/* BufferFormatStructs.proto */
#define IS_UNSIGNED(type) (((type) -1) > 0)
struct __Pyx_StructField_;
#define __PYX_BUF_FLAGS_PACKED_STRUCT (1 << 0)
typedef struct {
  const char* name;
  struct __Pyx_StructField_* fields;
  size_t size;
  size_t arraysize[8];
  int ndim;
  char typegroup;
  char is_unsigned;
  int flags;
} __Pyx_TypeInfo;
typedef struct __Pyx_StructField_ {
  __Pyx_TypeInfo* type;
  const char* name;
  size_t offset;
} __Pyx_StructField;
typedef struct {
  __Pyx_StructField* field;
  size_t parent_offset;
} __Pyx_BufFmt_StackElem;
typedef struct {
  __Pyx_StructField root;
  __Pyx_BufFmt_StackElem* head;
  size_t fmt_offset;
  size_t new_count, enc_count;
  size_t struct_alignment;
  int is_complex;
  char enc_type;
  char new_packmode;
  char enc_packmode;
  char is_valid_array;
} __Pyx_BufFmt_Context;

/* ForceInitThreads.proto */
#ifndef __PYX_FORCE_INIT_THREADS
  #define __PYX_FORCE_INIT_THREADS 0
#endif

/* NoFastGil.proto */
#define __Pyx_PyGILState_Ensure PyGILState_Ensure
#define __Pyx_PyGILState_Release PyGILState_Release
#define __Pyx_FastGIL_Remember()
#define __Pyx_FastGIL_Forget()
#define __Pyx_FastGilFuncInit()


/* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":776
 * # in Cython to enable them only on the right systems.
 * 
 * ctypedef npy_int8       int8_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t
 */
typedef npy_int8 __pyx_t_5numpy_int8_t;

/* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":777
 * 
 * ctypedef npy_int8       int8_t
 * ctypedef npy_int16      int16_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int32      int32_t
 * ctypedef npy_int64      int64_t
 */
typedef npy_int16 __pyx_t_5numpy_int16_t;

/* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":778
 * ctypedef npy_int8       int8_t
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int64      int64_t
 * #ctypedef npy_int96      int96_t
 */
typedef npy_int32 __pyx_t_5numpy_int32_t;

/* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":779
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t
 * ctypedef npy_int64      int64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_int96      int96_t
 * #ctypedef npy_int128     int128_t
 */
typedef npy_int64 __pyx_t_5numpy_int64_t;

/* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":783
 * #ctypedef npy_int128     int128_t
 * 
 * ctypedef npy_uint8      uint8_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t
 */
typedef npy_uint8 __pyx_t_5numpy_uint8_t;

/* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":784
 * 
 * ctypedef npy_uint8      uint8_t
 * ctypedef npy_uint16     uint16_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint32     uint32_t
 * ctypedef npy_uint64     uint64_t
 */
typedef npy_uint16 __pyx_t_5numpy_uint16_t;

/* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":785
 * ctypedef npy_uint8      uint8_t
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint64     uint64_t
 * #ctypedef npy_uint96     uint96_t
 */
typedef npy_uint32 __pyx_t_5numpy_uint32_t;

/* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":786
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t
 * ctypedef npy_uint64     uint64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_uint96     uint96_t
 * #ctypedef npy_uint128    uint128_t
 */
typedef npy_uint64 __pyx_t_5numpy_uint64_t;

/* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":790
 * #ctypedef npy_uint128    uint128_t
 * 
 * ctypedef npy_float32    float32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_float64    float64_t
 * #ctypedef npy_float80    float80_t
 */
typedef npy_float32 __pyx_t_5numpy_float32_t;

/* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":791
 * 
 * ctypedef npy_float32    float32_t
 * ctypedef npy_float64    float64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_float80    float80_t
 * #ctypedef npy_float128   float128_t
 */
typedef npy_float64 __pyx_t_5numpy_float64_t;

/* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":800
 * # The int types are mapped a bit surprising --
 * # numpy.int corresponds to 'l' and numpy.long to 'q'
 * ctypedef npy_long       int_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longlong   long_t
 * ctypedef npy_longlong   longlong_t
 */
typedef npy_long __pyx_t_5numpy_int_t;

/* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":801
 * # numpy.int corresponds to 'l' and numpy.long to 'q'
 * ctypedef npy_long       int_t
 * ctypedef npy_longlong   long_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longlong   longlong_t
 * 
 */
typedef npy_longlong __pyx_t_5numpy_long_t;

/* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":802
 * ctypedef npy_long       int_t
 * ctypedef npy_longlong   long_t
 * ctypedef npy_longlong   longlong_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_ulong      uint_t
 */
typedef npy_longlong __pyx_t_5numpy_longlong_t;

/* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":804
 * ctypedef npy_longlong   longlong_t
 * 
 * ctypedef npy_ulong      uint_t             # <<<<<<<<<<<<<<
 * ctypedef npy_ulonglong  ulong_t
 * ctypedef npy_ulonglong  ulonglong_t
 */
typedef npy_ulong __pyx_t_5numpy_uint_t;

/* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":805
 * 
 * ctypedef npy_ulong      uint_t
 * ctypedef npy_ulonglong  ulong_t             # <<<<<<<<<<<<<<
 * ctypedef npy_ulonglong  ulonglong_t
 * 
 */
typedef npy_ulonglong __pyx_t_5numpy_ulong_t;

/* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":806
 * ctypedef npy_ulong      uint_t
 * ctypedef npy_ulonglong  ulong_t
 * ctypedef npy_ulonglong  ulonglong_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_intp       intp_t
 */
typedef npy_ulonglong __pyx_t_5numpy_ulonglong_t;

/* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":808
 * ctypedef npy_ulonglong  ulonglong_t
 * 
 * ctypedef npy_intp       intp_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uintp      uintp_t
 * 
 */
typedef npy_intp __pyx_t_5numpy_intp_t;

/* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":809
 * 
 * ctypedef npy_intp       intp_t
 * ctypedef npy_uintp      uintp_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_double     float_t
 */
typedef npy_uintp __pyx_t_5numpy_uintp_t;

/* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":811
 * ctypedef npy_uintp      uintp_t
 * 
 * ctypedef npy_double     float_t             # <<<<<<<<<<<<<<
 * ctypedef npy_double     double_t
 * ctypedef npy_longdouble longdouble_t
 */
typedef npy_double __pyx_t_5numpy_float_t;

/* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":812
 * 
 * ctypedef npy_double     float_t
 * ctypedef npy_double     double_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longdouble longdouble_t
 * 
 */
typedef npy_double __pyx_t_5numpy_double_t;

/* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":813
 * ctypedef npy_double     float_t
 * ctypedef npy_double     double_t
 * ctypedef npy_longdouble longdouble_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_cfloat      cfloat_t
 */
typedef npy_longdouble __pyx_t_5numpy_longdouble_t;
/* Declarations.proto */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    typedef ::std::complex< float > __pyx_t_float_complex;
  #else
    typedef float _Complex __pyx_t_float_complex;
  #endif
#else
    typedef struct { float real, imag; } __pyx_t_float_complex;
#endif
static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float, float);

/* Declarations.proto */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    typedef ::std::complex< double > __pyx_t_double_complex;
  #else
    typedef double _Complex __pyx_t_double_complex;
  #endif
#else
    typedef struct { double real, imag; } __pyx_t_double_complex;
#endif
static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double, double);


/*--- Type declarations ---*/
struct __pyx_obj_6pandas_5_libs_9hashtable_HashTable;
struct __pyx_obj_6pandas_5_libs_9hashtable_UInt64HashTable;
struct __pyx_obj_6pandas_5_libs_9hashtable_Int64HashTable;
struct __pyx_obj_6pandas_5_libs_9hashtable_Float64HashTable;
struct __pyx_obj_6pandas_5_libs_9hashtable_PyObjectHashTable;
struct __pyx_obj_6pandas_5_libs_9hashtable_StringHashTable;
struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector;
struct __pyx_obj_6pandas_5_libs_8interval_IntervalMixin;
struct __pyx_obj_6pandas_5_libs_8interval_Interval;
struct __pyx_obj_6pandas_5_libs_8interval_IntervalTree;
struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode;
struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode;
struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode;
struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode;
struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode;
struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode;
struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode;
struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode;
struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode;
struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode;
struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode;
struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode;
struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode;
struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode;
struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode;
struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode;
struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode;
struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode;
struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode;
struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode;
struct __pyx_array_obj;
struct __pyx_MemviewEnum_obj;
struct __pyx_memoryview_obj;
struct __pyx_memoryviewslice_obj;

/* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":815
 * ctypedef npy_longdouble longdouble_t
 * 
 * ctypedef npy_cfloat      cfloat_t             # <<<<<<<<<<<<<<
 * ctypedef npy_cdouble     cdouble_t
 * ctypedef npy_clongdouble clongdouble_t
 */
typedef npy_cfloat __pyx_t_5numpy_cfloat_t;

/* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":816
 * 
 * ctypedef npy_cfloat      cfloat_t
 * ctypedef npy_cdouble     cdouble_t             # <<<<<<<<<<<<<<
 * ctypedef npy_clongdouble clongdouble_t
 * 
 */
typedef npy_cdouble __pyx_t_5numpy_cdouble_t;

/* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":817
 * ctypedef npy_cfloat      cfloat_t
 * ctypedef npy_cdouble     cdouble_t
 * ctypedef npy_clongdouble clongdouble_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_cdouble     complex_t
 */
typedef npy_clongdouble __pyx_t_5numpy_clongdouble_t;

/* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":819
 * ctypedef npy_clongdouble clongdouble_t
 * 
 * ctypedef npy_cdouble     complex_t             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew1(a):
 */
typedef npy_cdouble __pyx_t_5numpy_complex_t;
struct __pyx_t_6pandas_5_libs_9hashtable_Int64VectorData;

/* "pandas/_libs/hashtable.pxd":42
 *     cpdef set_item(self, object key, Py_ssize_t val)
 * 
 * cdef struct Int64VectorData:             # <<<<<<<<<<<<<<
 *     int64_t *data
 *     Py_ssize_t n, m
 */
struct __pyx_t_6pandas_5_libs_9hashtable_Int64VectorData {
  __pyx_t_5numpy_int64_t *data;
  Py_ssize_t n;
  Py_ssize_t m;
};

/* "pandas/_libs/hashtable.pxd":8
 * # prototypes for sharing
 * 
 * cdef class HashTable:             # <<<<<<<<<<<<<<
 *     pass
 * 
 */
struct __pyx_obj_6pandas_5_libs_9hashtable_HashTable {
  PyObject_HEAD
};


/* "pandas/_libs/hashtable.pxd":11
 *     pass
 * 
 * cdef class UInt64HashTable(HashTable):             # <<<<<<<<<<<<<<
 *     cdef kh_uint64_t *table
 * 
 */
struct __pyx_obj_6pandas_5_libs_9hashtable_UInt64HashTable {
  struct __pyx_obj_6pandas_5_libs_9hashtable_HashTable __pyx_base;
  struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_UInt64HashTable *__pyx_vtab;
  kh_uint64_t *table;
};


/* "pandas/_libs/hashtable.pxd":17
 *     cpdef set_item(self, uint64_t key, Py_ssize_t val)
 * 
 * cdef class Int64HashTable(HashTable):             # <<<<<<<<<<<<<<
 *     cdef kh_int64_t *table
 * 
 */
struct __pyx_obj_6pandas_5_libs_9hashtable_Int64HashTable {
  struct __pyx_obj_6pandas_5_libs_9hashtable_HashTable __pyx_base;
  struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64HashTable *__pyx_vtab;
  kh_int64_t *table;
};


/* "pandas/_libs/hashtable.pxd":23
 *     cpdef set_item(self, int64_t key, Py_ssize_t val)
 * 
 * cdef class Float64HashTable(HashTable):             # <<<<<<<<<<<<<<
 *     cdef kh_float64_t *table
 * 
 */
struct __pyx_obj_6pandas_5_libs_9hashtable_Float64HashTable {
  struct __pyx_obj_6pandas_5_libs_9hashtable_HashTable __pyx_base;
  struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Float64HashTable *__pyx_vtab;
  kh_float64_t *table;
};


/* "pandas/_libs/hashtable.pxd":29
 *     cpdef set_item(self, float64_t key, Py_ssize_t val)
 * 
 * cdef class PyObjectHashTable(HashTable):             # <<<<<<<<<<<<<<
 *     cdef kh_pymap_t *table
 * 
 */
struct __pyx_obj_6pandas_5_libs_9hashtable_PyObjectHashTable {
  struct __pyx_obj_6pandas_5_libs_9hashtable_HashTable __pyx_base;
  struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_PyObjectHashTable *__pyx_vtab;
  kh_pymap_t *table;
};


/* "pandas/_libs/hashtable.pxd":36
 * 
 * 
 * cdef class StringHashTable(HashTable):             # <<<<<<<<<<<<<<
 *     cdef kh_str_t *table
 * 
 */
struct __pyx_obj_6pandas_5_libs_9hashtable_StringHashTable {
  struct __pyx_obj_6pandas_5_libs_9hashtable_HashTable __pyx_base;
  struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_StringHashTable *__pyx_vtab;
  kh_str_t *table;
};


/* "pandas/_libs/hashtable.pxd":46
 *     Py_ssize_t n, m
 * 
 * cdef class Int64Vector:             # <<<<<<<<<<<<<<
 *     cdef Int64VectorData *data
 *     cdef ndarray ao
 */
struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector {
  PyObject_HEAD
  struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *__pyx_vtab;
  struct __pyx_t_6pandas_5_libs_9hashtable_Int64VectorData *data;
  PyArrayObject *ao;
  int external_view_exists;
};


/* "pandas/_libs/interval.pyx":32
 * 
 * 
 * cdef class IntervalMixin:             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
struct __pyx_obj_6pandas_5_libs_8interval_IntervalMixin {
  PyObject_HEAD
};


/* "pandas/_libs/interval.pyx":190
 * 
 * 
 * cdef class Interval(IntervalMixin):             # <<<<<<<<<<<<<<
 *     """
 *     Immutable object implementing an Interval, a bounded slice-like interval.
 */
struct __pyx_obj_6pandas_5_libs_8interval_Interval {
  struct __pyx_obj_6pandas_5_libs_8interval_IntervalMixin __pyx_base;
  PyObject *left;
  PyObject *right;
  PyObject *closed;
};


/* "pandas/_libs/intervaltree.pxi":20
 * # ----------------------------------------------------------------------
 * 
 * cdef class IntervalTree(IntervalMixin):             # <<<<<<<<<<<<<<
 *     """A centered interval tree
 * 
 */
struct __pyx_obj_6pandas_5_libs_8interval_IntervalTree {
  struct __pyx_obj_6pandas_5_libs_8interval_IntervalMixin __pyx_base;
  PyObject *left;
  PyObject *right;
  PyObject *root;
  PyObject *dtype;
  PyObject *closed;
  PyObject *_is_overlapping;
  PyObject *_left_sorter;
  PyObject *_right_sorter;
};


/* "pandas/_libs/intervaltree.pxi":233
 * NODE_CLASSES = {}
 * 
 * cdef class Float32ClosedLeftIntervalNode:             # <<<<<<<<<<<<<<
 *     """Non-terminal node for an IntervalTree
 * 
 */
struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode {
  PyObject_HEAD
  struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode *__pyx_vtab;
  struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode *left_node;
  struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode *right_node;
  __Pyx_memviewslice center_left_values;
  __Pyx_memviewslice center_right_values;
  __Pyx_memviewslice left;
  __Pyx_memviewslice right;
  __Pyx_memviewslice center_left_indices;
  __Pyx_memviewslice center_right_indices;
  __Pyx_memviewslice indices;
  __pyx_t_5numpy_float32_t min_left;
  __pyx_t_5numpy_float32_t max_right;
  __pyx_t_5numpy_float32_t pivot;
  __pyx_t_5numpy_int64_t n_elements;
  __pyx_t_5numpy_int64_t n_center;
  __pyx_t_5numpy_int64_t leaf_size;
  int is_leaf_node;
};


/* "pandas/_libs/intervaltree.pxi":403
 *              'left'] = Float32ClosedLeftIntervalNode
 * 
 * cdef class Float32ClosedRightIntervalNode:             # <<<<<<<<<<<<<<
 *     """Non-terminal node for an IntervalTree
 * 
 */
struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode {
  PyObject_HEAD
  struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode *__pyx_vtab;
  struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode *left_node;
  struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode *right_node;
  __Pyx_memviewslice center_left_values;
  __Pyx_memviewslice center_right_values;
  __Pyx_memviewslice left;
  __Pyx_memviewslice right;
  __Pyx_memviewslice center_left_indices;
  __Pyx_memviewslice center_right_indices;
  __Pyx_memviewslice indices;
  __pyx_t_5numpy_float32_t min_left;
  __pyx_t_5numpy_float32_t max_right;
  __pyx_t_5numpy_float32_t pivot;
  __pyx_t_5numpy_int64_t n_elements;
  __pyx_t_5numpy_int64_t n_center;
  __pyx_t_5numpy_int64_t leaf_size;
  int is_leaf_node;
};


/* "pandas/_libs/intervaltree.pxi":573
 *              'right'] = Float32ClosedRightIntervalNode
 * 
 * cdef class Float32ClosedBothIntervalNode:             # <<<<<<<<<<<<<<
 *     """Non-terminal node for an IntervalTree
 * 
 */
struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode {
  PyObject_HEAD
  struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode *__pyx_vtab;
  struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode *left_node;
  struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode *right_node;
  __Pyx_memviewslice center_left_values;
  __Pyx_memviewslice center_right_values;
  __Pyx_memviewslice left;
  __Pyx_memviewslice right;
  __Pyx_memviewslice center_left_indices;
  __Pyx_memviewslice center_right_indices;
  __Pyx_memviewslice indices;
  __pyx_t_5numpy_float32_t min_left;
  __pyx_t_5numpy_float32_t max_right;
  __pyx_t_5numpy_float32_t pivot;
  __pyx_t_5numpy_int64_t n_elements;
  __pyx_t_5numpy_int64_t n_center;
  __pyx_t_5numpy_int64_t leaf_size;
  int is_leaf_node;
};


/* "pandas/_libs/intervaltree.pxi":743
 *              'both'] = Float32ClosedBothIntervalNode
 * 
 * cdef class Float32ClosedNeitherIntervalNode:             # <<<<<<<<<<<<<<
 *     """Non-terminal node for an IntervalTree
 * 
 */
struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode {
  PyObject_HEAD
  struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode *__pyx_vtab;
  struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode *left_node;
  struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode *right_node;
  __Pyx_memviewslice center_left_values;
  __Pyx_memviewslice center_right_values;
  __Pyx_memviewslice left;
  __Pyx_memviewslice right;
  __Pyx_memviewslice center_left_indices;
  __Pyx_memviewslice center_right_indices;
  __Pyx_memviewslice indices;
  __pyx_t_5numpy_float32_t min_left;
  __pyx_t_5numpy_float32_t max_right;
  __pyx_t_5numpy_float32_t pivot;
  __pyx_t_5numpy_int64_t n_elements;
  __pyx_t_5numpy_int64_t n_center;
  __pyx_t_5numpy_int64_t leaf_size;
  int is_leaf_node;
};


/* "pandas/_libs/intervaltree.pxi":913
 *              'neither'] = Float32ClosedNeitherIntervalNode
 * 
 * cdef class Float64ClosedLeftIntervalNode:             # <<<<<<<<<<<<<<
 *     """Non-terminal node for an IntervalTree
 * 
 */
struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode {
  PyObject_HEAD
  struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode *__pyx_vtab;
  struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode *left_node;
  struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode *right_node;
  __Pyx_memviewslice center_left_values;
  __Pyx_memviewslice center_right_values;
  __Pyx_memviewslice left;
  __Pyx_memviewslice right;
  __Pyx_memviewslice center_left_indices;
  __Pyx_memviewslice center_right_indices;
  __Pyx_memviewslice indices;
  __pyx_t_5numpy_float64_t min_left;
  __pyx_t_5numpy_float64_t max_right;
  __pyx_t_5numpy_float64_t pivot;
  __pyx_t_5numpy_int64_t n_elements;
  __pyx_t_5numpy_int64_t n_center;
  __pyx_t_5numpy_int64_t leaf_size;
  int is_leaf_node;
};


/* "pandas/_libs/intervaltree.pxi":1083
 *              'left'] = Float64ClosedLeftIntervalNode
 * 
 * cdef class Float64ClosedRightIntervalNode:             # <<<<<<<<<<<<<<
 *     """Non-terminal node for an IntervalTree
 * 
 */
struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode {
  PyObject_HEAD
  struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode *__pyx_vtab;
  struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode *left_node;
  struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode *right_node;
  __Pyx_memviewslice center_left_values;
  __Pyx_memviewslice center_right_values;
  __Pyx_memviewslice left;
  __Pyx_memviewslice right;
  __Pyx_memviewslice center_left_indices;
  __Pyx_memviewslice center_right_indices;
  __Pyx_memviewslice indices;
  __pyx_t_5numpy_float64_t min_left;
  __pyx_t_5numpy_float64_t max_right;
  __pyx_t_5numpy_float64_t pivot;
  __pyx_t_5numpy_int64_t n_elements;
  __pyx_t_5numpy_int64_t n_center;
  __pyx_t_5numpy_int64_t leaf_size;
  int is_leaf_node;
};


/* "pandas/_libs/intervaltree.pxi":1253
 *              'right'] = Float64ClosedRightIntervalNode
 * 
 * cdef class Float64ClosedBothIntervalNode:             # <<<<<<<<<<<<<<
 *     """Non-terminal node for an IntervalTree
 * 
 */
struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode {
  PyObject_HEAD
  struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode *__pyx_vtab;
  struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode *left_node;
  struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode *right_node;
  __Pyx_memviewslice center_left_values;
  __Pyx_memviewslice center_right_values;
  __Pyx_memviewslice left;
  __Pyx_memviewslice right;
  __Pyx_memviewslice center_left_indices;
  __Pyx_memviewslice center_right_indices;
  __Pyx_memviewslice indices;
  __pyx_t_5numpy_float64_t min_left;
  __pyx_t_5numpy_float64_t max_right;
  __pyx_t_5numpy_float64_t pivot;
  __pyx_t_5numpy_int64_t n_elements;
  __pyx_t_5numpy_int64_t n_center;
  __pyx_t_5numpy_int64_t leaf_size;
  int is_leaf_node;
};


/* "pandas/_libs/intervaltree.pxi":1423
 *              'both'] = Float64ClosedBothIntervalNode
 * 
 * cdef class Float64ClosedNeitherIntervalNode:             # <<<<<<<<<<<<<<
 *     """Non-terminal node for an IntervalTree
 * 
 */
struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode {
  PyObject_HEAD
  struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode *__pyx_vtab;
  struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode *left_node;
  struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode *right_node;
  __Pyx_memviewslice center_left_values;
  __Pyx_memviewslice center_right_values;
  __Pyx_memviewslice left;
  __Pyx_memviewslice right;
  __Pyx_memviewslice center_left_indices;
  __Pyx_memviewslice center_right_indices;
  __Pyx_memviewslice indices;
  __pyx_t_5numpy_float64_t min_left;
  __pyx_t_5numpy_float64_t max_right;
  __pyx_t_5numpy_float64_t pivot;
  __pyx_t_5numpy_int64_t n_elements;
  __pyx_t_5numpy_int64_t n_center;
  __pyx_t_5numpy_int64_t leaf_size;
  int is_leaf_node;
};


/* "pandas/_libs/intervaltree.pxi":1593
 *              'neither'] = Float64ClosedNeitherIntervalNode
 * 
 * cdef class Int32ClosedLeftIntervalNode:             # <<<<<<<<<<<<<<
 *     """Non-terminal node for an IntervalTree
 * 
 */
struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode {
  PyObject_HEAD
  struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode *__pyx_vtab;
  struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode *left_node;
  struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode *right_node;
  __Pyx_memviewslice center_left_values;
  __Pyx_memviewslice center_right_values;
  __Pyx_memviewslice left;
  __Pyx_memviewslice right;
  __Pyx_memviewslice center_left_indices;
  __Pyx_memviewslice center_right_indices;
  __Pyx_memviewslice indices;
  __pyx_t_5numpy_int32_t min_left;
  __pyx_t_5numpy_int32_t max_right;
  __pyx_t_5numpy_int32_t pivot;
  __pyx_t_5numpy_int64_t n_elements;
  __pyx_t_5numpy_int64_t n_center;
  __pyx_t_5numpy_int64_t leaf_size;
  int is_leaf_node;
};


/* "pandas/_libs/intervaltree.pxi":1763
 *              'left'] = Int32ClosedLeftIntervalNode
 * 
 * cdef class Int32ClosedRightIntervalNode:             # <<<<<<<<<<<<<<
 *     """Non-terminal node for an IntervalTree
 * 
 */
struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode {
  PyObject_HEAD
  struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode *__pyx_vtab;
  struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode *left_node;
  struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode *right_node;
  __Pyx_memviewslice center_left_values;
  __Pyx_memviewslice center_right_values;
  __Pyx_memviewslice left;
  __Pyx_memviewslice right;
  __Pyx_memviewslice center_left_indices;
  __Pyx_memviewslice center_right_indices;
  __Pyx_memviewslice indices;
  __pyx_t_5numpy_int32_t min_left;
  __pyx_t_5numpy_int32_t max_right;
  __pyx_t_5numpy_int32_t pivot;
  __pyx_t_5numpy_int64_t n_elements;
  __pyx_t_5numpy_int64_t n_center;
  __pyx_t_5numpy_int64_t leaf_size;
  int is_leaf_node;
};


/* "pandas/_libs/intervaltree.pxi":1933
 *              'right'] = Int32ClosedRightIntervalNode
 * 
 * cdef class Int32ClosedBothIntervalNode:             # <<<<<<<<<<<<<<
 *     """Non-terminal node for an IntervalTree
 * 
 */
struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode {
  PyObject_HEAD
  struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode *__pyx_vtab;
  struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode *left_node;
  struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode *right_node;
  __Pyx_memviewslice center_left_values;
  __Pyx_memviewslice center_right_values;
  __Pyx_memviewslice left;
  __Pyx_memviewslice right;
  __Pyx_memviewslice center_left_indices;
  __Pyx_memviewslice center_right_indices;
  __Pyx_memviewslice indices;
  __pyx_t_5numpy_int32_t min_left;
  __pyx_t_5numpy_int32_t max_right;
  __pyx_t_5numpy_int32_t pivot;
  __pyx_t_5numpy_int64_t n_elements;
  __pyx_t_5numpy_int64_t n_center;
  __pyx_t_5numpy_int64_t leaf_size;
  int is_leaf_node;
};


/* "pandas/_libs/intervaltree.pxi":2103
 *              'both'] = Int32ClosedBothIntervalNode
 * 
 * cdef class Int32ClosedNeitherIntervalNode:             # <<<<<<<<<<<<<<
 *     """Non-terminal node for an IntervalTree
 * 
 */
struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode {
  PyObject_HEAD
  struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode *__pyx_vtab;
  struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode *left_node;
  struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode *right_node;
  __Pyx_memviewslice center_left_values;
  __Pyx_memviewslice center_right_values;
  __Pyx_memviewslice left;
  __Pyx_memviewslice right;
  __Pyx_memviewslice center_left_indices;
  __Pyx_memviewslice center_right_indices;
  __Pyx_memviewslice indices;
  __pyx_t_5numpy_int32_t min_left;
  __pyx_t_5numpy_int32_t max_right;
  __pyx_t_5numpy_int32_t pivot;
  __pyx_t_5numpy_int64_t n_elements;
  __pyx_t_5numpy_int64_t n_center;
  __pyx_t_5numpy_int64_t leaf_size;
  int is_leaf_node;
};


/* "pandas/_libs/intervaltree.pxi":2273
 *              'neither'] = Int32ClosedNeitherIntervalNode
 * 
 * cdef class Int64ClosedLeftIntervalNode:             # <<<<<<<<<<<<<<
 *     """Non-terminal node for an IntervalTree
 * 
 */
struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode {
  PyObject_HEAD
  struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode *__pyx_vtab;
  struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode *left_node;
  struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode *right_node;
  __Pyx_memviewslice center_left_values;
  __Pyx_memviewslice center_right_values;
  __Pyx_memviewslice left;
  __Pyx_memviewslice right;
  __Pyx_memviewslice center_left_indices;
  __Pyx_memviewslice center_right_indices;
  __Pyx_memviewslice indices;
  __pyx_t_5numpy_int64_t min_left;
  __pyx_t_5numpy_int64_t max_right;
  __pyx_t_5numpy_int64_t pivot;
  __pyx_t_5numpy_int64_t n_elements;
  __pyx_t_5numpy_int64_t n_center;
  __pyx_t_5numpy_int64_t leaf_size;
  int is_leaf_node;
};


/* "pandas/_libs/intervaltree.pxi":2443
 *              'left'] = Int64ClosedLeftIntervalNode
 * 
 * cdef class Int64ClosedRightIntervalNode:             # <<<<<<<<<<<<<<
 *     """Non-terminal node for an IntervalTree
 * 
 */
struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode {
  PyObject_HEAD
  struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode *__pyx_vtab;
  struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode *left_node;
  struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode *right_node;
  __Pyx_memviewslice center_left_values;
  __Pyx_memviewslice center_right_values;
  __Pyx_memviewslice left;
  __Pyx_memviewslice right;
  __Pyx_memviewslice center_left_indices;
  __Pyx_memviewslice center_right_indices;
  __Pyx_memviewslice indices;
  __pyx_t_5numpy_int64_t min_left;
  __pyx_t_5numpy_int64_t max_right;
  __pyx_t_5numpy_int64_t pivot;
  __pyx_t_5numpy_int64_t n_elements;
  __pyx_t_5numpy_int64_t n_center;
  __pyx_t_5numpy_int64_t leaf_size;
  int is_leaf_node;
};


/* "pandas/_libs/intervaltree.pxi":2613
 *              'right'] = Int64ClosedRightIntervalNode
 * 
 * cdef class Int64ClosedBothIntervalNode:             # <<<<<<<<<<<<<<
 *     """Non-terminal node for an IntervalTree
 * 
 */
struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode {
  PyObject_HEAD
  struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode *__pyx_vtab;
  struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode *left_node;
  struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode *right_node;
  __Pyx_memviewslice center_left_values;
  __Pyx_memviewslice center_right_values;
  __Pyx_memviewslice left;
  __Pyx_memviewslice right;
  __Pyx_memviewslice center_left_indices;
  __Pyx_memviewslice center_right_indices;
  __Pyx_memviewslice indices;
  __pyx_t_5numpy_int64_t min_left;
  __pyx_t_5numpy_int64_t max_right;
  __pyx_t_5numpy_int64_t pivot;
  __pyx_t_5numpy_int64_t n_elements;
  __pyx_t_5numpy_int64_t n_center;
  __pyx_t_5numpy_int64_t leaf_size;
  int is_leaf_node;
};


/* "pandas/_libs/intervaltree.pxi":2783
 *              'both'] = Int64ClosedBothIntervalNode
 * 
 * cdef class Int64ClosedNeitherIntervalNode:             # <<<<<<<<<<<<<<
 *     """Non-terminal node for an IntervalTree
 * 
 */
struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode {
  PyObject_HEAD
  struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode *__pyx_vtab;
  struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode *left_node;
  struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode *right_node;
  __Pyx_memviewslice center_left_values;
  __Pyx_memviewslice center_right_values;
  __Pyx_memviewslice left;
  __Pyx_memviewslice right;
  __Pyx_memviewslice center_left_indices;
  __Pyx_memviewslice center_right_indices;
  __Pyx_memviewslice indices;
  __pyx_t_5numpy_int64_t min_left;
  __pyx_t_5numpy_int64_t max_right;
  __pyx_t_5numpy_int64_t pivot;
  __pyx_t_5numpy_int64_t n_elements;
  __pyx_t_5numpy_int64_t n_center;
  __pyx_t_5numpy_int64_t leaf_size;
  int is_leaf_node;
};


/* "pandas/_libs/intervaltree.pxi":2953
 *              'neither'] = Int64ClosedNeitherIntervalNode
 * 
 * cdef class Uint64ClosedLeftIntervalNode:             # <<<<<<<<<<<<<<
 *     """Non-terminal node for an IntervalTree
 * 
 */
struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode {
  PyObject_HEAD
  struct __pyx_vtabstruct_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode *__pyx_vtab;
  struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode *left_node;
  struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode *right_node;
  __Pyx_memviewslice center_left_values;
  __Pyx_memviewslice center_right_values;
  __Pyx_memviewslice left;
  __Pyx_memviewslice right;
  __Pyx_memviewslice center_left_indices;
  __Pyx_memviewslice center_right_indices;
  __Pyx_memviewslice indices;
  __pyx_t_5numpy_uint64_t min_left;
  __pyx_t_5numpy_uint64_t max_right;
  __pyx_t_5numpy_uint64_t pivot;
  __pyx_t_5numpy_int64_t n_elements;
  __pyx_t_5numpy_int64_t n_center;
  __pyx_t_5numpy_int64_t leaf_size;
  int is_leaf_node;
};


/* "pandas/_libs/intervaltree.pxi":3123
 *              'left'] = Uint64ClosedLeftIntervalNode
 * 
 * cdef class Uint64ClosedRightIntervalNode:             # <<<<<<<<<<<<<<
 *     """Non-terminal node for an IntervalTree
 * 
 */
struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode {
  PyObject_HEAD
  struct __pyx_vtabstruct_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode *__pyx_vtab;
  struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode *left_node;
  struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode *right_node;
  __Pyx_memviewslice center_left_values;
  __Pyx_memviewslice center_right_values;
  __Pyx_memviewslice left;
  __Pyx_memviewslice right;
  __Pyx_memviewslice center_left_indices;
  __Pyx_memviewslice center_right_indices;
  __Pyx_memviewslice indices;
  __pyx_t_5numpy_uint64_t min_left;
  __pyx_t_5numpy_uint64_t max_right;
  __pyx_t_5numpy_uint64_t pivot;
  __pyx_t_5numpy_int64_t n_elements;
  __pyx_t_5numpy_int64_t n_center;
  __pyx_t_5numpy_int64_t leaf_size;
  int is_leaf_node;
};


/* "pandas/_libs/intervaltree.pxi":3293
 *              'right'] = Uint64ClosedRightIntervalNode
 * 
 * cdef class Uint64ClosedBothIntervalNode:             # <<<<<<<<<<<<<<
 *     """Non-terminal node for an IntervalTree
 * 
 */
struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode {
  PyObject_HEAD
  struct __pyx_vtabstruct_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode *__pyx_vtab;
  struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode *left_node;
  struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode *right_node;
  __Pyx_memviewslice center_left_values;
  __Pyx_memviewslice center_right_values;
  __Pyx_memviewslice left;
  __Pyx_memviewslice right;
  __Pyx_memviewslice center_left_indices;
  __Pyx_memviewslice center_right_indices;
  __Pyx_memviewslice indices;
  __pyx_t_5numpy_uint64_t min_left;
  __pyx_t_5numpy_uint64_t max_right;
  __pyx_t_5numpy_uint64_t pivot;
  __pyx_t_5numpy_int64_t n_elements;
  __pyx_t_5numpy_int64_t n_center;
  __pyx_t_5numpy_int64_t leaf_size;
  int is_leaf_node;
};


/* "pandas/_libs/intervaltree.pxi":3463
 *              'both'] = Uint64ClosedBothIntervalNode
 * 
 * cdef class Uint64ClosedNeitherIntervalNode:             # <<<<<<<<<<<<<<
 *     """Non-terminal node for an IntervalTree
 * 
 */
struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode {
  PyObject_HEAD
  struct __pyx_vtabstruct_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode *__pyx_vtab;
  struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode *left_node;
  struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode *right_node;
  __Pyx_memviewslice center_left_values;
  __Pyx_memviewslice center_right_values;
  __Pyx_memviewslice left;
  __Pyx_memviewslice right;
  __Pyx_memviewslice center_left_indices;
  __Pyx_memviewslice center_right_indices;
  __Pyx_memviewslice indices;
  __pyx_t_5numpy_uint64_t min_left;
  __pyx_t_5numpy_uint64_t max_right;
  __pyx_t_5numpy_uint64_t pivot;
  __pyx_t_5numpy_int64_t n_elements;
  __pyx_t_5numpy_int64_t n_center;
  __pyx_t_5numpy_int64_t leaf_size;
  int is_leaf_node;
};


/* "View.MemoryView":105
 * 
 * @cname("__pyx_array")
 * cdef class array:             # <<<<<<<<<<<<<<
 * 
 *     cdef:
 */
struct __pyx_array_obj {
  PyObject_HEAD
  struct __pyx_vtabstruct_array *__pyx_vtab;
  char *data;
  Py_ssize_t len;
  char *format;
  int ndim;
  Py_ssize_t *_shape;
  Py_ssize_t *_strides;
  Py_ssize_t itemsize;
  PyObject *mode;
  PyObject *_format;
  void (*callback_free_data)(void *);
  int free_data;
  int dtype_is_object;
};


/* "View.MemoryView":279
 * 
 * @cname('__pyx_MemviewEnum')
 * cdef class Enum(object):             # <<<<<<<<<<<<<<
 *     cdef object name
 *     def __init__(self, name):
 */
struct __pyx_MemviewEnum_obj {
  PyObject_HEAD
  PyObject *name;
};


/* "View.MemoryView":330
 * 
 * @cname('__pyx_memoryview')
 * cdef class memoryview(object):             # <<<<<<<<<<<<<<
 * 
 *     cdef object obj
 */
struct __pyx_memoryview_obj {
  PyObject_HEAD
  struct __pyx_vtabstruct_memoryview *__pyx_vtab;
  PyObject *obj;
  PyObject *_size;
  PyObject *_array_interface;
  PyThread_type_lock lock;
  __pyx_atomic_int acquisition_count[2];
  __pyx_atomic_int *acquisition_count_aligned_p;
  Py_buffer view;
  int flags;
  int dtype_is_object;
  __Pyx_TypeInfo *typeinfo;
};


/* "View.MemoryView":965
 * 
 * @cname('__pyx_memoryviewslice')
 * cdef class _memoryviewslice(memoryview):             # <<<<<<<<<<<<<<
 *     "Internal class for passing memoryview slices to Python"
 * 
 */
struct __pyx_memoryviewslice_obj {
  struct __pyx_memoryview_obj __pyx_base;
  __Pyx_memviewslice from_slice;
  PyObject *from_object;
  PyObject *(*to_object_func)(char *);
  int (*to_dtype_func)(char *, PyObject *);
};



/* "pandas/_libs/hashtable.pxd":11
 *     pass
 * 
 * cdef class UInt64HashTable(HashTable):             # <<<<<<<<<<<<<<
 *     cdef kh_uint64_t *table
 * 
 */

struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_UInt64HashTable {
  PyObject *(*get_item)(struct __pyx_obj_6pandas_5_libs_9hashtable_UInt64HashTable *, __pyx_t_5numpy_uint64_t, int __pyx_skip_dispatch);
  PyObject *(*set_item)(struct __pyx_obj_6pandas_5_libs_9hashtable_UInt64HashTable *, __pyx_t_5numpy_uint64_t, Py_ssize_t, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_UInt64HashTable *__pyx_vtabptr_6pandas_5_libs_9hashtable_UInt64HashTable;


/* "pandas/_libs/hashtable.pxd":17
 *     cpdef set_item(self, uint64_t key, Py_ssize_t val)
 * 
 * cdef class Int64HashTable(HashTable):             # <<<<<<<<<<<<<<
 *     cdef kh_int64_t *table
 * 
 */

struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64HashTable {
  PyObject *(*get_item)(struct __pyx_obj_6pandas_5_libs_9hashtable_Int64HashTable *, __pyx_t_5numpy_int64_t, int __pyx_skip_dispatch);
  PyObject *(*set_item)(struct __pyx_obj_6pandas_5_libs_9hashtable_Int64HashTable *, __pyx_t_5numpy_int64_t, Py_ssize_t, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64HashTable *__pyx_vtabptr_6pandas_5_libs_9hashtable_Int64HashTable;


/* "pandas/_libs/hashtable.pxd":23
 *     cpdef set_item(self, int64_t key, Py_ssize_t val)
 * 
 * cdef class Float64HashTable(HashTable):             # <<<<<<<<<<<<<<
 *     cdef kh_float64_t *table
 * 
 */

struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Float64HashTable {
  PyObject *(*get_item)(struct __pyx_obj_6pandas_5_libs_9hashtable_Float64HashTable *, __pyx_t_5numpy_float64_t, int __pyx_skip_dispatch);
  PyObject *(*set_item)(struct __pyx_obj_6pandas_5_libs_9hashtable_Float64HashTable *, __pyx_t_5numpy_float64_t, Py_ssize_t, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Float64HashTable *__pyx_vtabptr_6pandas_5_libs_9hashtable_Float64HashTable;


/* "pandas/_libs/hashtable.pxd":29
 *     cpdef set_item(self, float64_t key, Py_ssize_t val)
 * 
 * cdef class PyObjectHashTable(HashTable):             # <<<<<<<<<<<<<<
 *     cdef kh_pymap_t *table
 * 
 */

struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_PyObjectHashTable {
  PyObject *(*get_item)(struct __pyx_obj_6pandas_5_libs_9hashtable_PyObjectHashTable *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*set_item)(struct __pyx_obj_6pandas_5_libs_9hashtable_PyObjectHashTable *, PyObject *, Py_ssize_t, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_PyObjectHashTable *__pyx_vtabptr_6pandas_5_libs_9hashtable_PyObjectHashTable;


/* "pandas/_libs/hashtable.pxd":36
 * 
 * 
 * cdef class StringHashTable(HashTable):             # <<<<<<<<<<<<<<
 *     cdef kh_str_t *table
 * 
 */

struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_StringHashTable {
  PyObject *(*get_item)(struct __pyx_obj_6pandas_5_libs_9hashtable_StringHashTable *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*set_item)(struct __pyx_obj_6pandas_5_libs_9hashtable_StringHashTable *, PyObject *, Py_ssize_t, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_StringHashTable *__pyx_vtabptr_6pandas_5_libs_9hashtable_StringHashTable;


/* "pandas/_libs/hashtable.pxd":46
 *     Py_ssize_t n, m
 * 
 * cdef class Int64Vector:             # <<<<<<<<<<<<<<
 *     cdef Int64VectorData *data
 *     cdef ndarray ao
 */

struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector {
  PyObject *(*resize)(struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *);
  PyObject *(*to_array)(struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, int __pyx_skip_dispatch);
  void (*append)(struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_int64_t);
  PyObject *(*extend)(struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __Pyx_memviewslice);
};
static struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *__pyx_vtabptr_6pandas_5_libs_9hashtable_Int64Vector;


/* "pandas/_libs/intervaltree.pxi":233
 * NODE_CLASSES = {}
 * 
 * cdef class Float32ClosedLeftIntervalNode:             # <<<<<<<<<<<<<<
 *     """Non-terminal node for an IntervalTree
 * 
 */

struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode {
  PyObject *(*classify_intervals)(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode *, __Pyx_memviewslice, __Pyx_memviewslice);
  PyObject *(*new_child_node)(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode *, PyArrayObject *, PyArrayObject *, PyArrayObject *, PyArrayObject *);
  PyObject *(*__pyx_fuse_0query)(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_float64_t, int __pyx_skip_dispatch);
  PyObject *(*__pyx_fuse_1query)(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_float32_t, int __pyx_skip_dispatch);
  PyObject *(*__pyx_fuse_2query)(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_int64_t, int __pyx_skip_dispatch);
  PyObject *(*__pyx_fuse_3query)(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_int32_t, int __pyx_skip_dispatch);
  PyObject *(*__pyx_fuse_4query)(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_uint64_t, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode *__pyx_vtabptr_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode;


/* "pandas/_libs/intervaltree.pxi":403
 *              'left'] = Float32ClosedLeftIntervalNode
 * 
 * cdef class Float32ClosedRightIntervalNode:             # <<<<<<<<<<<<<<
 *     """Non-terminal node for an IntervalTree
 * 
 */

struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode {
  PyObject *(*classify_intervals)(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode *, __Pyx_memviewslice, __Pyx_memviewslice);
  PyObject *(*new_child_node)(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode *, PyArrayObject *, PyArrayObject *, PyArrayObject *, PyArrayObject *);
  PyObject *(*__pyx_fuse_0query)(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_float64_t, int __pyx_skip_dispatch);
  PyObject *(*__pyx_fuse_1query)(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_float32_t, int __pyx_skip_dispatch);
  PyObject *(*__pyx_fuse_2query)(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_int64_t, int __pyx_skip_dispatch);
  PyObject *(*__pyx_fuse_3query)(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_int32_t, int __pyx_skip_dispatch);
  PyObject *(*__pyx_fuse_4query)(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_uint64_t, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode *__pyx_vtabptr_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode;


/* "pandas/_libs/intervaltree.pxi":573
 *              'right'] = Float32ClosedRightIntervalNode
 * 
 * cdef class Float32ClosedBothIntervalNode:             # <<<<<<<<<<<<<<
 *     """Non-terminal node for an IntervalTree
 * 
 */

struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode {
  PyObject *(*classify_intervals)(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode *, __Pyx_memviewslice, __Pyx_memviewslice);
  PyObject *(*new_child_node)(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode *, PyArrayObject *, PyArrayObject *, PyArrayObject *, PyArrayObject *);
  PyObject *(*__pyx_fuse_0query)(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_float64_t, int __pyx_skip_dispatch);
  PyObject *(*__pyx_fuse_1query)(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_float32_t, int __pyx_skip_dispatch);
  PyObject *(*__pyx_fuse_2query)(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_int64_t, int __pyx_skip_dispatch);
  PyObject *(*__pyx_fuse_3query)(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_int32_t, int __pyx_skip_dispatch);
  PyObject *(*__pyx_fuse_4query)(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_uint64_t, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode *__pyx_vtabptr_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode;


/* "pandas/_libs/intervaltree.pxi":743
 *              'both'] = Float32ClosedBothIntervalNode
 * 
 * cdef class Float32ClosedNeitherIntervalNode:             # <<<<<<<<<<<<<<
 *     """Non-terminal node for an IntervalTree
 * 
 */

struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode {
  PyObject *(*classify_intervals)(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode *, __Pyx_memviewslice, __Pyx_memviewslice);
  PyObject *(*new_child_node)(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode *, PyArrayObject *, PyArrayObject *, PyArrayObject *, PyArrayObject *);
  PyObject *(*__pyx_fuse_0query)(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_float64_t, int __pyx_skip_dispatch);
  PyObject *(*__pyx_fuse_1query)(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_float32_t, int __pyx_skip_dispatch);
  PyObject *(*__pyx_fuse_2query)(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_int64_t, int __pyx_skip_dispatch);
  PyObject *(*__pyx_fuse_3query)(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_int32_t, int __pyx_skip_dispatch);
  PyObject *(*__pyx_fuse_4query)(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_uint64_t, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode *__pyx_vtabptr_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode;


/* "pandas/_libs/intervaltree.pxi":913
 *              'neither'] = Float32ClosedNeitherIntervalNode
 * 
 * cdef class Float64ClosedLeftIntervalNode:             # <<<<<<<<<<<<<<
 *     """Non-terminal node for an IntervalTree
 * 
 */

struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode {
  PyObject *(*classify_intervals)(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode *, __Pyx_memviewslice, __Pyx_memviewslice);
  PyObject *(*new_child_node)(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode *, PyArrayObject *, PyArrayObject *, PyArrayObject *, PyArrayObject *);
  PyObject *(*__pyx_fuse_0query)(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_float64_t, int __pyx_skip_dispatch);
  PyObject *(*__pyx_fuse_1query)(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_float32_t, int __pyx_skip_dispatch);
  PyObject *(*__pyx_fuse_2query)(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_int64_t, int __pyx_skip_dispatch);
  PyObject *(*__pyx_fuse_3query)(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_int32_t, int __pyx_skip_dispatch);
  PyObject *(*__pyx_fuse_4query)(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_uint64_t, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode *__pyx_vtabptr_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode;


/* "pandas/_libs/intervaltree.pxi":1083
 *              'left'] = Float64ClosedLeftIntervalNode
 * 
 * cdef class Float64ClosedRightIntervalNode:             # <<<<<<<<<<<<<<
 *     """Non-terminal node for an IntervalTree
 * 
 */

struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode {
  PyObject *(*classify_intervals)(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode *, __Pyx_memviewslice, __Pyx_memviewslice);
  PyObject *(*new_child_node)(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode *, PyArrayObject *, PyArrayObject *, PyArrayObject *, PyArrayObject *);
  PyObject *(*__pyx_fuse_0query)(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_float64_t, int __pyx_skip_dispatch);
  PyObject *(*__pyx_fuse_1query)(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_float32_t, int __pyx_skip_dispatch);
  PyObject *(*__pyx_fuse_2query)(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_int64_t, int __pyx_skip_dispatch);
  PyObject *(*__pyx_fuse_3query)(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_int32_t, int __pyx_skip_dispatch);
  PyObject *(*__pyx_fuse_4query)(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_uint64_t, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode *__pyx_vtabptr_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode;


/* "pandas/_libs/intervaltree.pxi":1253
 *              'right'] = Float64ClosedRightIntervalNode
 * 
 * cdef class Float64ClosedBothIntervalNode:             # <<<<<<<<<<<<<<
 *     """Non-terminal node for an IntervalTree
 * 
 */

struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode {
  PyObject *(*classify_intervals)(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode *, __Pyx_memviewslice, __Pyx_memviewslice);
  PyObject *(*new_child_node)(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode *, PyArrayObject *, PyArrayObject *, PyArrayObject *, PyArrayObject *);
  PyObject *(*__pyx_fuse_0query)(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_float64_t, int __pyx_skip_dispatch);
  PyObject *(*__pyx_fuse_1query)(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_float32_t, int __pyx_skip_dispatch);
  PyObject *(*__pyx_fuse_2query)(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_int64_t, int __pyx_skip_dispatch);
  PyObject *(*__pyx_fuse_3query)(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_int32_t, int __pyx_skip_dispatch);
  PyObject *(*__pyx_fuse_4query)(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_uint64_t, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode *__pyx_vtabptr_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode;


/* "pandas/_libs/intervaltree.pxi":1423
 *              'both'] = Float64ClosedBothIntervalNode
 * 
 * cdef class Float64ClosedNeitherIntervalNode:             # <<<<<<<<<<<<<<
 *     """Non-terminal node for an IntervalTree
 * 
 */

struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode {
  PyObject *(*classify_intervals)(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode *, __Pyx_memviewslice, __Pyx_memviewslice);
  PyObject *(*new_child_node)(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode *, PyArrayObject *, PyArrayObject *, PyArrayObject *, PyArrayObject *);
  PyObject *(*__pyx_fuse_0query)(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_float64_t, int __pyx_skip_dispatch);
  PyObject *(*__pyx_fuse_1query)(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_float32_t, int __pyx_skip_dispatch);
  PyObject *(*__pyx_fuse_2query)(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_int64_t, int __pyx_skip_dispatch);
  PyObject *(*__pyx_fuse_3query)(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_int32_t, int __pyx_skip_dispatch);
  PyObject *(*__pyx_fuse_4query)(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_uint64_t, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode *__pyx_vtabptr_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode;


/* "pandas/_libs/intervaltree.pxi":1593
 *              'neither'] = Float64ClosedNeitherIntervalNode
 * 
 * cdef class Int32ClosedLeftIntervalNode:             # <<<<<<<<<<<<<<
 *     """Non-terminal node for an IntervalTree
 * 
 */

struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode {
  PyObject *(*classify_intervals)(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode *, __Pyx_memviewslice, __Pyx_memviewslice);
  PyObject *(*new_child_node)(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode *, PyArrayObject *, PyArrayObject *, PyArrayObject *, PyArrayObject *);
  PyObject *(*__pyx_fuse_0query)(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_float64_t, int __pyx_skip_dispatch);
  PyObject *(*__pyx_fuse_1query)(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_float32_t, int __pyx_skip_dispatch);
  PyObject *(*__pyx_fuse_2query)(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_int64_t, int __pyx_skip_dispatch);
  PyObject *(*__pyx_fuse_3query)(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_int32_t, int __pyx_skip_dispatch);
  PyObject *(*__pyx_fuse_4query)(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_uint64_t, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode *__pyx_vtabptr_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode;


/* "pandas/_libs/intervaltree.pxi":1763
 *              'left'] = Int32ClosedLeftIntervalNode
 * 
 * cdef class Int32ClosedRightIntervalNode:             # <<<<<<<<<<<<<<
 *     """Non-terminal node for an IntervalTree
 * 
 */

struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode {
  PyObject *(*classify_intervals)(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode *, __Pyx_memviewslice, __Pyx_memviewslice);
  PyObject *(*new_child_node)(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode *, PyArrayObject *, PyArrayObject *, PyArrayObject *, PyArrayObject *);
  PyObject *(*__pyx_fuse_0query)(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_float64_t, int __pyx_skip_dispatch);
  PyObject *(*__pyx_fuse_1query)(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_float32_t, int __pyx_skip_dispatch);
  PyObject *(*__pyx_fuse_2query)(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_int64_t, int __pyx_skip_dispatch);
  PyObject *(*__pyx_fuse_3query)(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_int32_t, int __pyx_skip_dispatch);
  PyObject *(*__pyx_fuse_4query)(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_uint64_t, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode *__pyx_vtabptr_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode;


/* "pandas/_libs/intervaltree.pxi":1933
 *              'right'] = Int32ClosedRightIntervalNode
 * 
 * cdef class Int32ClosedBothIntervalNode:             # <<<<<<<<<<<<<<
 *     """Non-terminal node for an IntervalTree
 * 
 */

struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode {
  PyObject *(*classify_intervals)(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode *, __Pyx_memviewslice, __Pyx_memviewslice);
  PyObject *(*new_child_node)(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode *, PyArrayObject *, PyArrayObject *, PyArrayObject *, PyArrayObject *);
  PyObject *(*__pyx_fuse_0query)(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_float64_t, int __pyx_skip_dispatch);
  PyObject *(*__pyx_fuse_1query)(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_float32_t, int __pyx_skip_dispatch);
  PyObject *(*__pyx_fuse_2query)(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_int64_t, int __pyx_skip_dispatch);
  PyObject *(*__pyx_fuse_3query)(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_int32_t, int __pyx_skip_dispatch);
  PyObject *(*__pyx_fuse_4query)(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_uint64_t, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode *__pyx_vtabptr_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode;


/* "pandas/_libs/intervaltree.pxi":2103
 *              'both'] = Int32ClosedBothIntervalNode
 * 
 * cdef class Int32ClosedNeitherIntervalNode:             # <<<<<<<<<<<<<<
 *     """Non-terminal node for an IntervalTree
 * 
 */

struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode {
  PyObject *(*classify_intervals)(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode *, __Pyx_memviewslice, __Pyx_memviewslice);
  PyObject *(*new_child_node)(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode *, PyArrayObject *, PyArrayObject *, PyArrayObject *, PyArrayObject *);
  PyObject *(*__pyx_fuse_0query)(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_float64_t, int __pyx_skip_dispatch);
  PyObject *(*__pyx_fuse_1query)(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_float32_t, int __pyx_skip_dispatch);
  PyObject *(*__pyx_fuse_2query)(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_int64_t, int __pyx_skip_dispatch);
  PyObject *(*__pyx_fuse_3query)(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_int32_t, int __pyx_skip_dispatch);
  PyObject *(*__pyx_fuse_4query)(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_uint64_t, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode *__pyx_vtabptr_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode;


/* "pandas/_libs/intervaltree.pxi":2273
 *              'neither'] = Int32ClosedNeitherIntervalNode
 * 
 * cdef class Int64ClosedLeftIntervalNode:             # <<<<<<<<<<<<<<
 *     """Non-terminal node for an IntervalTree
 * 
 */

struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode {
  PyObject *(*classify_intervals)(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode *, __Pyx_memviewslice, __Pyx_memviewslice);
  PyObject *(*new_child_node)(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode *, PyArrayObject *, PyArrayObject *, PyArrayObject *, PyArrayObject *);
  PyObject *(*__pyx_fuse_0query)(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_float64_t, int __pyx_skip_dispatch);
  PyObject *(*__pyx_fuse_1query)(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_float32_t, int __pyx_skip_dispatch);
  PyObject *(*__pyx_fuse_2query)(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_int64_t, int __pyx_skip_dispatch);
  PyObject *(*__pyx_fuse_3query)(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_int32_t, int __pyx_skip_dispatch);
  PyObject *(*__pyx_fuse_4query)(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_uint64_t, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode *__pyx_vtabptr_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode;


/* "pandas/_libs/intervaltree.pxi":2443
 *              'left'] = Int64ClosedLeftIntervalNode
 * 
 * cdef class Int64ClosedRightIntervalNode:             # <<<<<<<<<<<<<<
 *     """Non-terminal node for an IntervalTree
 * 
 */

struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode {
  PyObject *(*classify_intervals)(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode *, __Pyx_memviewslice, __Pyx_memviewslice);
  PyObject *(*new_child_node)(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode *, PyArrayObject *, PyArrayObject *, PyArrayObject *, PyArrayObject *);
  PyObject *(*__pyx_fuse_0query)(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_float64_t, int __pyx_skip_dispatch);
  PyObject *(*__pyx_fuse_1query)(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_float32_t, int __pyx_skip_dispatch);
  PyObject *(*__pyx_fuse_2query)(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_int64_t, int __pyx_skip_dispatch);
  PyObject *(*__pyx_fuse_3query)(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_int32_t, int __pyx_skip_dispatch);
  PyObject *(*__pyx_fuse_4query)(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_uint64_t, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode *__pyx_vtabptr_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode;


/* "pandas/_libs/intervaltree.pxi":2613
 *              'right'] = Int64ClosedRightIntervalNode
 * 
 * cdef class Int64ClosedBothIntervalNode:             # <<<<<<<<<<<<<<
 *     """Non-terminal node for an IntervalTree
 * 
 */

struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode {
  PyObject *(*classify_intervals)(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode *, __Pyx_memviewslice, __Pyx_memviewslice);
  PyObject *(*new_child_node)(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode *, PyArrayObject *, PyArrayObject *, PyArrayObject *, PyArrayObject *);
  PyObject *(*__pyx_fuse_0query)(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_float64_t, int __pyx_skip_dispatch);
  PyObject *(*__pyx_fuse_1query)(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_float32_t, int __pyx_skip_dispatch);
  PyObject *(*__pyx_fuse_2query)(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_int64_t, int __pyx_skip_dispatch);
  PyObject *(*__pyx_fuse_3query)(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_int32_t, int __pyx_skip_dispatch);
  PyObject *(*__pyx_fuse_4query)(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_uint64_t, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode *__pyx_vtabptr_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode;


/* "pandas/_libs/intervaltree.pxi":2783
 *              'both'] = Int64ClosedBothIntervalNode
 * 
 * cdef class Int64ClosedNeitherIntervalNode:             # <<<<<<<<<<<<<<
 *     """Non-terminal node for an IntervalTree
 * 
 */

struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode {
  PyObject *(*classify_intervals)(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode *, __Pyx_memviewslice, __Pyx_memviewslice);
  PyObject *(*new_child_node)(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode *, PyArrayObject *, PyArrayObject *, PyArrayObject *, PyArrayObject *);
  PyObject *(*__pyx_fuse_0query)(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_float64_t, int __pyx_skip_dispatch);
  PyObject *(*__pyx_fuse_1query)(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_float32_t, int __pyx_skip_dispatch);
  PyObject *(*__pyx_fuse_2query)(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_int64_t, int __pyx_skip_dispatch);
  PyObject *(*__pyx_fuse_3query)(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_int32_t, int __pyx_skip_dispatch);
  PyObject *(*__pyx_fuse_4query)(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_uint64_t, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode *__pyx_vtabptr_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode;


/* "pandas/_libs/intervaltree.pxi":2953
 *              'neither'] = Int64ClosedNeitherIntervalNode
 * 
 * cdef class Uint64ClosedLeftIntervalNode:             # <<<<<<<<<<<<<<
 *     """Non-terminal node for an IntervalTree
 * 
 */

struct __pyx_vtabstruct_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode {
  PyObject *(*classify_intervals)(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode *, __Pyx_memviewslice, __Pyx_memviewslice);
  PyObject *(*new_child_node)(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode *, PyArrayObject *, PyArrayObject *, PyArrayObject *, PyArrayObject *);
  PyObject *(*__pyx_fuse_0query)(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_float64_t, int __pyx_skip_dispatch);
  PyObject *(*__pyx_fuse_1query)(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_float32_t, int __pyx_skip_dispatch);
  PyObject *(*__pyx_fuse_2query)(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_int64_t, int __pyx_skip_dispatch);
  PyObject *(*__pyx_fuse_3query)(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_int32_t, int __pyx_skip_dispatch);
  PyObject *(*__pyx_fuse_4query)(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_uint64_t, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode *__pyx_vtabptr_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode;


/* "pandas/_libs/intervaltree.pxi":3123
 *              'left'] = Uint64ClosedLeftIntervalNode
 * 
 * cdef class Uint64ClosedRightIntervalNode:             # <<<<<<<<<<<<<<
 *     """Non-terminal node for an IntervalTree
 * 
 */

struct __pyx_vtabstruct_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode {
  PyObject *(*classify_intervals)(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode *, __Pyx_memviewslice, __Pyx_memviewslice);
  PyObject *(*new_child_node)(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode *, PyArrayObject *, PyArrayObject *, PyArrayObject *, PyArrayObject *);
  PyObject *(*__pyx_fuse_0query)(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_float64_t, int __pyx_skip_dispatch);
  PyObject *(*__pyx_fuse_1query)(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_float32_t, int __pyx_skip_dispatch);
  PyObject *(*__pyx_fuse_2query)(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_int64_t, int __pyx_skip_dispatch);
  PyObject *(*__pyx_fuse_3query)(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_int32_t, int __pyx_skip_dispatch);
  PyObject *(*__pyx_fuse_4query)(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_uint64_t, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode *__pyx_vtabptr_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode;


/* "pandas/_libs/intervaltree.pxi":3293
 *              'right'] = Uint64ClosedRightIntervalNode
 * 
 * cdef class Uint64ClosedBothIntervalNode:             # <<<<<<<<<<<<<<
 *     """Non-terminal node for an IntervalTree
 * 
 */

struct __pyx_vtabstruct_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode {
  PyObject *(*classify_intervals)(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode *, __Pyx_memviewslice, __Pyx_memviewslice);
  PyObject *(*new_child_node)(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode *, PyArrayObject *, PyArrayObject *, PyArrayObject *, PyArrayObject *);
  PyObject *(*__pyx_fuse_0query)(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_float64_t, int __pyx_skip_dispatch);
  PyObject *(*__pyx_fuse_1query)(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_float32_t, int __pyx_skip_dispatch);
  PyObject *(*__pyx_fuse_2query)(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_int64_t, int __pyx_skip_dispatch);
  PyObject *(*__pyx_fuse_3query)(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_int32_t, int __pyx_skip_dispatch);
  PyObject *(*__pyx_fuse_4query)(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_uint64_t, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode *__pyx_vtabptr_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode;


/* "pandas/_libs/intervaltree.pxi":3463
 *              'both'] = Uint64ClosedBothIntervalNode
 * 
 * cdef class Uint64ClosedNeitherIntervalNode:             # <<<<<<<<<<<<<<
 *     """Non-terminal node for an IntervalTree
 * 
 */

struct __pyx_vtabstruct_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode {
  PyObject *(*classify_intervals)(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode *, __Pyx_memviewslice, __Pyx_memviewslice);
  PyObject *(*new_child_node)(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode *, PyArrayObject *, PyArrayObject *, PyArrayObject *, PyArrayObject *);
  PyObject *(*__pyx_fuse_0query)(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_float64_t, int __pyx_skip_dispatch);
  PyObject *(*__pyx_fuse_1query)(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_float32_t, int __pyx_skip_dispatch);
  PyObject *(*__pyx_fuse_2query)(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_int64_t, int __pyx_skip_dispatch);
  PyObject *(*__pyx_fuse_3query)(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_int32_t, int __pyx_skip_dispatch);
  PyObject *(*__pyx_fuse_4query)(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_uint64_t, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode *__pyx_vtabptr_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode;


/* "View.MemoryView":105
 * 
 * @cname("__pyx_array")
 * cdef class array:             # <<<<<<<<<<<<<<
 * 
 *     cdef:
 */

struct __pyx_vtabstruct_array {
  PyObject *(*get_memview)(struct __pyx_array_obj *);
};
static struct __pyx_vtabstruct_array *__pyx_vtabptr_array;


/* "View.MemoryView":330
 * 
 * @cname('__pyx_memoryview')
 * cdef class memoryview(object):             # <<<<<<<<<<<<<<
 * 
 *     cdef object obj
 */

struct __pyx_vtabstruct_memoryview {
  char *(*get_item_pointer)(struct __pyx_memoryview_obj *, PyObject *);
  PyObject *(*is_slice)(struct __pyx_memoryview_obj *, PyObject *);
  PyObject *(*setitem_slice_assignment)(struct __pyx_memoryview_obj *, PyObject *, PyObject *);
  PyObject *(*setitem_slice_assign_scalar)(struct __pyx_memoryview_obj *, struct __pyx_memoryview_obj *, PyObject *);
  PyObject *(*setitem_indexed)(struct __pyx_memoryview_obj *, PyObject *, PyObject *);
  PyObject *(*convert_item_to_object)(struct __pyx_memoryview_obj *, char *);
  PyObject *(*assign_item_from_object)(struct __pyx_memoryview_obj *, char *, PyObject *);
};
static struct __pyx_vtabstruct_memoryview *__pyx_vtabptr_memoryview;


/* "View.MemoryView":965
 * 
 * @cname('__pyx_memoryviewslice')
 * cdef class _memoryviewslice(memoryview):             # <<<<<<<<<<<<<<
 *     "Internal class for passing memoryview slices to Python"
 * 
 */

struct __pyx_vtabstruct__memoryviewslice {
  struct __pyx_vtabstruct_memoryview __pyx_base;
};
static struct __pyx_vtabstruct__memoryviewslice *__pyx_vtabptr__memoryviewslice;

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, int);
    void (*DECREF)(void*, PyObject*, int);
    void (*GOTREF)(void*, PyObject*, int);
    void (*GIVEREF)(void*, PyObject*, int);
    void* (*SetupContext)(const char*, int, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__)
#endif
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_XINCREF(r)  do { if((r) != NULL) {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) != NULL) {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) != NULL) {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) != NULL) {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyObjectGetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* IncludeStringH.proto */
#include <string.h>

/* BytesEquals.proto */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals);

/* UnicodeEquals.proto */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals);

/* GetTopmostException.proto */
#if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem * __Pyx_PyErr_GetTopmostException(PyThreadState *tstate);
#endif

/* PyThreadStateGet.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#define __Pyx_PyErr_Occurred()  __pyx_tstate->curexc_type
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  PyErr_Occurred()
#endif

/* SaveResetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSave(type, value, tb)  __Pyx__ExceptionSave(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#define __Pyx_ExceptionReset(type, value, tb)  __Pyx__ExceptionReset(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
#else
#define __Pyx_ExceptionSave(type, value, tb)   PyErr_GetExcInfo(type, value, tb)
#define __Pyx_ExceptionReset(type, value, tb)  PyErr_SetExcInfo(type, value, tb)
#endif

/* PyErrExceptionMatches.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_ExceptionMatches(err) __Pyx_PyErr_ExceptionMatchesInState(__pyx_tstate, err)
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err);
#else
#define __Pyx_PyErr_ExceptionMatches(err)  PyErr_ExceptionMatches(err)
#endif

/* GetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_GetException(type, value, tb)  __Pyx__GetException(__pyx_tstate, type, value, tb)
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* RaiseDoubleKeywords.proto */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywords.proto */
static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject **argnames[],\
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args,\
    const char* function_name);

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* PyFunctionFastCall.proto */
#if CYTHON_FAST_PYCALL
#define __Pyx_PyFunction_FastCall(func, args, nargs)\
    __Pyx_PyFunction_FastCallDict((func), (args), (nargs), NULL)
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs);
#else
#define __Pyx_PyFunction_FastCallDict(func, args, nargs, kwargs) _PyFunction_FastCallDict(func, args, nargs, kwargs)
#endif
#define __Pyx_BUILD_ASSERT_EXPR(cond)\
    (sizeof(char [1 - 2*!(cond)]) - 1)
#ifndef Py_MEMBER_SIZE
#define Py_MEMBER_SIZE(type, member) sizeof(((type *)0)->member)
#endif
  static size_t __pyx_pyframe_localsplus_offset = 0;
  #include "frameobject.h"
  #define __Pxy_PyFrame_Initialize_Offsets()\
    ((void)__Pyx_BUILD_ASSERT_EXPR(sizeof(PyFrameObject) == offsetof(PyFrameObject, f_localsplus) + Py_MEMBER_SIZE(PyFrameObject, f_localsplus)),\
     (void)(__pyx_pyframe_localsplus_offset = ((size_t)PyFrame_Type.tp_basicsize) - Py_MEMBER_SIZE(PyFrameObject, f_localsplus)))
  #define __Pyx_PyFrame_GetLocalsplus(frame)\
    (assert(__pyx_pyframe_localsplus_offset), (PyObject **)(((char *)(frame)) + __pyx_pyframe_localsplus_offset))
#endif

/* PyCFunctionFastCall.proto */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject *__Pyx_PyCFunction_FastCall(PyObject *func, PyObject **args, Py_ssize_t nargs);
#else
#define __Pyx_PyCFunction_FastCall(func, args, nargs)  (assert(0), NULL)
#endif

/* PyObjectCall.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* PyObjectCallMethO.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectCallOneArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* PyErrFetchRestore.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* RaiseException.proto */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* GetAttr.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *, PyObject *);

/* GetAttr3.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *, PyObject *, PyObject *);

/* PyDictVersioning.proto */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
#define __PYX_DICT_VERSION_INIT  ((PY_UINT64_T) -1)
#define __PYX_GET_DICT_VERSION(dict)  (((PyDictObject*)(dict))->ma_version_tag)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)\
    (version_var) = __PYX_GET_DICT_VERSION(dict);\
    (cache_var) = (value);
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP) {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    if (likely(__PYX_GET_DICT_VERSION(DICT) == __pyx_dict_version)) {\
        (VAR) = __pyx_dict_cached_value;\
    } else {\
        (VAR) = __pyx_dict_cached_value = (LOOKUP);\
        __pyx_dict_version = __PYX_GET_DICT_VERSION(DICT);\
    }\
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj);
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj);
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version);
#else
#define __PYX_GET_DICT_VERSION(dict)  (0)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP)  (VAR) = (LOOKUP);
#endif

/* GetModuleGlobalName.proto */
#if CYTHON_USE_DICT_VERSIONS
#define __Pyx_GetModuleGlobalName(var, name)  {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    (var) = (likely(__pyx_dict_version == __PYX_GET_DICT_VERSION(__pyx_d))) ?\
        (likely(__pyx_dict_cached_value) ? __Pyx_NewRef(__pyx_dict_cached_value) : __Pyx_GetBuiltinName(name)) :\
        __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
}
#define __Pyx_GetModuleGlobalNameUncached(var, name)  {\
    PY_UINT64_T __pyx_dict_version;\
    PyObject *__pyx_dict_cached_value;\
    (var) = __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
}
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value);
#else
#define __Pyx_GetModuleGlobalName(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name);
#endif

/* HasAttr.proto */
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *, PyObject *);

/* ArgTypeTest.proto */
#define __Pyx_ArgTypeTest(obj, type, none_allowed, name, exact)\
    ((likely((Py_TYPE(obj) == type) | (none_allowed && (obj == Py_None)))) ? 1 :\
        __Pyx__ArgTypeTest(obj, type, name, exact))
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact);

/* PyObjectCall2Args.proto */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2);

/* PySequenceContains.proto */
static CYTHON_INLINE int __Pyx_PySequence_ContainsTF(PyObject* item, PyObject* seq, int eq) {
    int result = PySequence_Contains(seq, item);
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

/* DictGetItem.proto */
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key);
#define __Pyx_PyObject_Dict_GetItem(obj, name)\
    (likely(PyDict_CheckExact(obj)) ?\
     __Pyx_PyDict_GetItem(obj, name) : PyObject_GetItem(obj, name))
#else
#define __Pyx_PyDict_GetItem(d, key) PyObject_GetItem(d, key)
#define __Pyx_PyObject_Dict_GetItem(obj, name)  PyObject_GetItem(obj, name)
#endif

/* PyObjectCallNoArg.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func);
#else
#define __Pyx_PyObject_CallNoArg(func) __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL)
#endif

/* RaiseTooManyValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected);

/* RaiseNeedMoreValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index);

/* IterFinish.proto */
static CYTHON_INLINE int __Pyx_IterFinish(void);

/* UnpackItemEndCheck.proto */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected);

/* ExtTypeTest.proto */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type);

/* GetItemInt.proto */
#define __Pyx_GetItemInt(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Fast(o, (Py_ssize_t)i, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL) :\
               __Pyx_GetItemInt_Generic(o, to_py_func(i))))
#define __Pyx_GetItemInt_List(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_List_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
#define __Pyx_GetItemInt_Tuple(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Tuple_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "tuple index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i,
                                                     int is_list, int wraparound, int boundscheck);

/* SetItemInt.proto */
#define __Pyx_SetItemInt(o, i, v, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_SetItemInt_Fast(o, (Py_ssize_t)i, v, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list assignment index out of range"), -1) :\
               __Pyx_SetItemInt_Generic(o, to_py_func(i), v)))
static int __Pyx_SetItemInt_Generic(PyObject *o, PyObject *j, PyObject *v);
static CYTHON_INLINE int __Pyx_SetItemInt_Fast(PyObject *o, Py_ssize_t i, PyObject *v,
                                               int is_list, int wraparound, int boundscheck);

/* ObjectGetItem.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject* key);
#else
#define __Pyx_PyObject_GetItem(obj, key)  PyObject_GetItem(obj, key)
#endif

/* SliceObject.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetSlice(
        PyObject* obj, Py_ssize_t cstart, Py_ssize_t cstop,
        PyObject** py_start, PyObject** py_stop, PyObject** py_slice,
        int has_cstart, int has_cstop, int wraparound);

/* PyDictContains.proto */
static CYTHON_INLINE int __Pyx_PyDict_ContainsTF(PyObject* item, PyObject* dict, int eq) {
    int result = PyDict_Contains(dict, item);
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

/* PyObjectGetMethod.proto */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method);

/* PyObjectCallMethod0.proto */
static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name);

/* RaiseNoneIterError.proto */
static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void);

/* UnpackTupleError.proto */
static void __Pyx_UnpackTupleError(PyObject *, Py_ssize_t index);

/* UnpackTuple2.proto */
#define __Pyx_unpack_tuple2(tuple, value1, value2, is_tuple, has_known_size, decref_tuple)\
    (likely(is_tuple || PyTuple_Check(tuple)) ?\
        (likely(has_known_size || PyTuple_GET_SIZE(tuple) == 2) ?\
            __Pyx_unpack_tuple2_exact(tuple, value1, value2, decref_tuple) :\
            (__Pyx_UnpackTupleError(tuple, 2), -1)) :\
        __Pyx_unpack_tuple2_generic(tuple, value1, value2, has_known_size, decref_tuple))
static CYTHON_INLINE int __Pyx_unpack_tuple2_exact(
    PyObject* tuple, PyObject** value1, PyObject** value2, int decref_tuple);
static int __Pyx_unpack_tuple2_generic(
    PyObject* tuple, PyObject** value1, PyObject** value2, int has_known_size, int decref_tuple);

/* dict_iter.proto */
static CYTHON_INLINE PyObject* __Pyx_dict_iterator(PyObject* dict, int is_dict, PyObject* method_name,
                                                   Py_ssize_t* p_orig_length, int* p_is_dict);
static CYTHON_INLINE int __Pyx_dict_iter_next(PyObject* dict_or_iter, Py_ssize_t orig_length, Py_ssize_t* ppos,
                                              PyObject** pkey, PyObject** pvalue, PyObject** pitem, int is_dict);

/* ListAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_PyList_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len) & likely(len > (L->allocated >> 1))) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        Py_SIZE(list) = len+1;
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_PyList_Append(L,x) PyList_Append(L,x)
#endif

/* Import.proto */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level);

/* UnicodeAsUCS4.proto */
static CYTHON_INLINE Py_UCS4 __Pyx_PyUnicode_AsPy_UCS4(PyObject*);

/* object_ord.proto */
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyObject_Ord(c)\
    (likely(PyUnicode_Check(c)) ? (long)__Pyx_PyUnicode_AsPy_UCS4(c) : __Pyx__PyObject_Ord(c))
#else
#define __Pyx_PyObject_Ord(c) __Pyx__PyObject_Ord(c)
#endif
static long __Pyx__PyObject_Ord(PyObject* c);

/* BufferIndexError.proto */
static void __Pyx_RaiseBufferIndexError(int axis);

/* MemviewSliceInit.proto */
#define __Pyx_BUF_MAX_NDIMS %(BUF_MAX_NDIMS)d
#define __Pyx_MEMVIEW_DIRECT   1
#define __Pyx_MEMVIEW_PTR      2
#define __Pyx_MEMVIEW_FULL     4
#define __Pyx_MEMVIEW_CONTIG   8
#define __Pyx_MEMVIEW_STRIDED  16
#define __Pyx_MEMVIEW_FOLLOW   32
#define __Pyx_IS_C_CONTIG 1
#define __Pyx_IS_F_CONTIG 2
static int __Pyx_init_memviewslice(
                struct __pyx_memoryview_obj *memview,
                int ndim,
                __Pyx_memviewslice *memviewslice,
                int memview_is_new_reference);
static CYTHON_INLINE int __pyx_add_acquisition_count_locked(
    __pyx_atomic_int *acquisition_count, PyThread_type_lock lock);
static CYTHON_INLINE int __pyx_sub_acquisition_count_locked(
    __pyx_atomic_int *acquisition_count, PyThread_type_lock lock);
#define __pyx_get_slice_count_pointer(memview) (memview->acquisition_count_aligned_p)
#define __pyx_get_slice_count(memview) (*__pyx_get_slice_count_pointer(memview))
#define __PYX_INC_MEMVIEW(slice, have_gil) __Pyx_INC_MEMVIEW(slice, have_gil, __LINE__)
#define __PYX_XDEC_MEMVIEW(slice, have_gil) __Pyx_XDEC_MEMVIEW(slice, have_gil, __LINE__)
static CYTHON_INLINE void __Pyx_INC_MEMVIEW(__Pyx_memviewslice *, int, int);
static CYTHON_INLINE void __Pyx_XDEC_MEMVIEW(__Pyx_memviewslice *, int, int);

/* IsLittleEndian.proto */
static CYTHON_INLINE int __Pyx_Is_Little_Endian(void);

/* BufferFormatCheck.proto */
static const char* __Pyx_BufFmt_CheckString(__Pyx_BufFmt_Context* ctx, const char* ts);
static void __Pyx_BufFmt_Init(__Pyx_BufFmt_Context* ctx,
                              __Pyx_BufFmt_StackElem* stack,
                              __Pyx_TypeInfo* type);

/* BufferGetAndValidate.proto */
#define __Pyx_GetBufferAndValidate(buf, obj, dtype, flags, nd, cast, stack)\
    ((obj == Py_None || obj == NULL) ?\
    (__Pyx_ZeroBuffer(buf), 0) :\
    __Pyx__GetBufferAndValidate(buf, obj, dtype, flags, nd, cast, stack))
static int  __Pyx__GetBufferAndValidate(Py_buffer* buf, PyObject* obj,
    __Pyx_TypeInfo* dtype, int flags, int nd, int cast, __Pyx_BufFmt_StackElem* stack);
static void __Pyx_ZeroBuffer(Py_buffer* buf);
static CYTHON_INLINE void __Pyx_SafeReleaseBuffer(Py_buffer* info);
static Py_ssize_t __Pyx_minusones[] = { -1, -1, -1, -1, -1, -1, -1, -1 };
static Py_ssize_t __Pyx_zeros[] = { 0, 0, 0, 0, 0, 0, 0, 0 };

/* BufferFallbackError.proto */
static void __Pyx_RaiseBufferFallbackError(void);

/* PyObjectFormatAndDecref.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_FormatSimpleAndDecref(PyObject* s, PyObject* f);
static CYTHON_INLINE PyObject* __Pyx_PyObject_FormatAndDecref(PyObject* s, PyObject* f);

/* JoinPyUnicode.proto */
static PyObject* __Pyx_PyUnicode_Join(PyObject* value_tuple, Py_ssize_t value_count, Py_ssize_t result_ulength,
                                      Py_UCS4 max_char);

/* ImportFrom.proto */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name);

/* WriteUnraisableException.proto */
static void __Pyx_WriteUnraisable(const char *name, int clineno,
                                  int lineno, const char *filename,
                                  int full_traceback, int nogil);

/* StrEquals.proto */
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyString_Equals __Pyx_PyUnicode_Equals
#else
#define __Pyx_PyString_Equals __Pyx_PyBytes_Equals
#endif

/* None.proto */
static CYTHON_INLINE Py_ssize_t __Pyx_div_Py_ssize_t(Py_ssize_t, Py_ssize_t);

/* UnaryNegOverflows.proto */
#define UNARY_NEG_WOULD_OVERFLOW(x)\
        (((x) < 0) & ((unsigned long)(x) == 0-(unsigned long)(x)))

static CYTHON_UNUSED int __pyx_array_getbuffer(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /*proto*/
static PyObject *__pyx_array_get_memview(struct __pyx_array_obj *); /*proto*/
/* decode_c_string_utf16.proto */
static CYTHON_INLINE PyObject *__Pyx_PyUnicode_DecodeUTF16(const char *s, Py_ssize_t size, const char *errors) {
    int byteorder = 0;
    return PyUnicode_DecodeUTF16(s, size, errors, &byteorder);
}
static CYTHON_INLINE PyObject *__Pyx_PyUnicode_DecodeUTF16LE(const char *s, Py_ssize_t size, const char *errors) {
    int byteorder = -1;
    return PyUnicode_DecodeUTF16(s, size, errors, &byteorder);
}
static CYTHON_INLINE PyObject *__Pyx_PyUnicode_DecodeUTF16BE(const char *s, Py_ssize_t size, const char *errors) {
    int byteorder = 1;
    return PyUnicode_DecodeUTF16(s, size, errors, &byteorder);
}

/* decode_c_string.proto */
static CYTHON_INLINE PyObject* __Pyx_decode_c_string(
         const char* cstring, Py_ssize_t start, Py_ssize_t stop,
         const char* encoding, const char* errors,
         PyObject* (*decode_func)(const char *s, Py_ssize_t size, const char *errors));

/* SwapException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSwap(type, value, tb)  __Pyx__ExceptionSwap(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
#define __Pyx_PyErr_GivenExceptionMatches2(err, type1, type2) (PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2))
#endif
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)

static CYTHON_UNUSED int __pyx_memoryview_getbuffer(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /*proto*/
/* ListCompAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_ListComp_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len)) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        Py_SIZE(list) = len+1;
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_ListComp_Append(L,x) PyList_Append(L,x)
#endif

/* PyIntBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_AddObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyInt_AddObjC(op1, op2, intval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceAdd(op1, op2) : PyNumber_Add(op1, op2))
#endif

/* ListExtend.proto */
static CYTHON_INLINE int __Pyx_PyList_Extend(PyObject* L, PyObject* v) {
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject* none = _PyList_Extend((PyListObject*)L, v);
    if (unlikely(!none))
        return -1;
    Py_DECREF(none);
    return 0;
#else
    return PyList_SetSlice(L, PY_SSIZE_T_MAX, PY_SSIZE_T_MAX, v);
#endif
}

/* None.proto */
static CYTHON_INLINE void __Pyx_RaiseUnboundLocalError(const char *varname);

/* None.proto */
static CYTHON_INLINE long __Pyx_div_long(long, long);

/* CallNextTpTraverse.proto */
static int __Pyx_call_next_tp_traverse(PyObject* obj, visitproc v, void *a, traverseproc current_tp_traverse);

/* CallNextTpClear.proto */
static void __Pyx_call_next_tp_clear(PyObject* obj, inquiry current_tp_dealloc);

/* PyObject_GenericGetAttrNoDict.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttrNoDict PyObject_GenericGetAttr
#endif

/* PyObject_GenericGetAttr.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttr PyObject_GenericGetAttr
#endif

/* SetupReduce.proto */
static int __Pyx_setup_reduce(PyObject* type_obj);

/* SetVTable.proto */
static int __Pyx_SetVtable(PyObject *dict, void *vtable);

/* TypeImport.proto */
#ifndef __PYX_HAVE_RT_ImportType_proto
#define __PYX_HAVE_RT_ImportType_proto
enum __Pyx_ImportType_CheckSize {
   __Pyx_ImportType_CheckSize_Error = 0,
   __Pyx_ImportType_CheckSize_Warn = 1,
   __Pyx_ImportType_CheckSize_Ignore = 2
};
static PyTypeObject *__Pyx_ImportType(PyObject* module, const char *module_name, const char *class_name, size_t size, enum __Pyx_ImportType_CheckSize check_size);
#endif

/* GetVTable.proto */
static void* __Pyx_GetVtable(PyObject *dict);

/* pyfrozenset_new.proto */
static CYTHON_INLINE PyObject* __Pyx_PyFrozenSet_New(PyObject* it);

/* FetchCommonType.proto */
static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type);

/* CythonFunction.proto */
#define __Pyx_CyFunction_USED 1
#define __Pyx_CYFUNCTION_STATICMETHOD  0x01
#define __Pyx_CYFUNCTION_CLASSMETHOD   0x02
#define __Pyx_CYFUNCTION_CCLASS        0x04
#define __Pyx_CyFunction_GetClosure(f)\
    (((__pyx_CyFunctionObject *) (f))->func_closure)
#define __Pyx_CyFunction_GetClassObj(f)\
    (((__pyx_CyFunctionObject *) (f))->func_classobj)
#define __Pyx_CyFunction_Defaults(type, f)\
    ((type *)(((__pyx_CyFunctionObject *) (f))->defaults))
#define __Pyx_CyFunction_SetDefaultsGetter(f, g)\
    ((__pyx_CyFunctionObject *) (f))->defaults_getter = (g)
typedef struct {
    PyCFunctionObject func;
#if PY_VERSION_HEX < 0x030500A0
    PyObject *func_weakreflist;
#endif
    PyObject *func_dict;
    PyObject *func_name;
    PyObject *func_qualname;
    PyObject *func_doc;
    PyObject *func_globals;
    PyObject *func_code;
    PyObject *func_closure;
    PyObject *func_classobj;
    void *defaults;
    int defaults_pyobjects;
    int flags;
    PyObject *defaults_tuple;
    PyObject *defaults_kwdict;
    PyObject *(*defaults_getter)(PyObject *);
    PyObject *func_annotations;
} __pyx_CyFunctionObject;
static PyTypeObject *__pyx_CyFunctionType = 0;
#define __Pyx_CyFunction_Check(obj)  (__Pyx_TypeCheck(obj, __pyx_CyFunctionType))
#define __Pyx_CyFunction_NewEx(ml, flags, qualname, self, module, globals, code)\
    __Pyx_CyFunction_New(__pyx_CyFunctionType, ml, flags, qualname, self, module, globals, code)
static PyObject *__Pyx_CyFunction_New(PyTypeObject *, PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *self,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *m,
                                                         size_t size,
                                                         int pyobjects);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *m,
                                                            PyObject *tuple);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *m,
                                                             PyObject *dict);
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *m,
                                                              PyObject *dict);
static int __pyx_CyFunction_init(void);

/* FusedFunction.proto */
typedef struct {
    __pyx_CyFunctionObject func;
    PyObject *__signatures__;
    PyObject *type;
    PyObject *self;
} __pyx_FusedFunctionObject;
#define __pyx_FusedFunction_NewEx(ml, flags, qualname, self, module, globals, code)\
        __pyx_FusedFunction_New(__pyx_FusedFunctionType, ml, flags, qualname, self, module, globals, code)
static PyObject *__pyx_FusedFunction_New(PyTypeObject *type,
                                         PyMethodDef *ml, int flags,
                                         PyObject *qualname, PyObject *self,
                                         PyObject *module, PyObject *globals,
                                         PyObject *code);
static int __pyx_FusedFunction_clear(__pyx_FusedFunctionObject *self);
static PyTypeObject *__pyx_FusedFunctionType = NULL;
static int __pyx_FusedFunction_init(void);
#define __Pyx_FusedFunction_USED

/* CLineInTraceback.proto */
#ifdef CYTHON_CLINE_IN_TRACEBACK
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#else
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#endif

/* CodeObjectCache.proto */
typedef struct {
    PyCodeObject* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

#if PY_MAJOR_VERSION < 3
    static int __Pyx_GetBuffer(PyObject *obj, Py_buffer *view, int flags);
    static void __Pyx_ReleaseBuffer(Py_buffer *view);
#else
    #define __Pyx_GetBuffer PyObject_GetBuffer
    #define __Pyx_ReleaseBuffer PyBuffer_Release
#endif


/* BufferStructDeclare.proto */
typedef struct {
  Py_ssize_t shape, strides, suboffsets;
} __Pyx_Buf_DimInfo;
typedef struct {
  size_t refcount;
  Py_buffer pybuffer;
} __Pyx_Buffer;
typedef struct {
  __Pyx_Buffer *rcbuffer;
  char *data;
  __Pyx_Buf_DimInfo diminfo[8];
} __Pyx_LocalBuf_ND;

/* MemviewSliceIsContig.proto */
static int __pyx_memviewslice_is_contig(const __Pyx_memviewslice mvs, char order, int ndim);

/* OverlappingSlices.proto */
static int __pyx_slices_overlap(__Pyx_memviewslice *slice1,
                                __Pyx_memviewslice *slice2,
                                int ndim, size_t itemsize);

/* Capsule.proto */
static CYTHON_INLINE PyObject *__pyx_capsule_create(void *p, const char *sig);

/* TypeInfoCompare.proto */
static int __pyx_typeinfo_cmp(__Pyx_TypeInfo *a, __Pyx_TypeInfo *b);

/* MemviewSliceValidateAndInit.proto */
static int __Pyx_ValidateAndInit_memviewslice(
                int *axes_specs,
                int c_or_f_flag,
                int buf_flags,
                int ndim,
                __Pyx_TypeInfo *dtype,
                __Pyx_BufFmt_StackElem stack[],
                __Pyx_memviewslice *memviewslice,
                PyObject *original_obj);

/* ObjectToMemviewSlice.proto */
static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_float64_t(PyObject *, int writable_flag);

/* ObjectToMemviewSlice.proto */
static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_float32_t(PyObject *, int writable_flag);

/* ObjectToMemviewSlice.proto */
static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(PyObject *, int writable_flag);

/* ObjectToMemviewSlice.proto */
static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int32_t(PyObject *, int writable_flag);

/* ObjectToMemviewSlice.proto */
static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_uint64_t(PyObject *, int writable_flag);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_npy_int64(npy_int64 value);

/* MemviewDtypeToObject.proto */
static CYTHON_INLINE PyObject *__pyx_memview_get_nn___pyx_t_5numpy_int64_t(const char *itemp);
static CYTHON_INLINE int __pyx_memview_set_nn___pyx_t_5numpy_int64_t(const char *itemp, PyObject *obj);

/* MemviewDtypeToObject.proto */
static CYTHON_INLINE PyObject *__pyx_memview_get_nn___pyx_t_5numpy_float32_t(const char *itemp);
static CYTHON_INLINE int __pyx_memview_set_nn___pyx_t_5numpy_float32_t(const char *itemp, PyObject *obj);

/* MemviewDtypeToObject.proto */
static CYTHON_INLINE PyObject *__pyx_memview_get_nn___pyx_t_5numpy_float64_t(const char *itemp);
static CYTHON_INLINE int __pyx_memview_set_nn___pyx_t_5numpy_float64_t(const char *itemp, PyObject *obj);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_npy_int32(npy_int32 value);

/* MemviewDtypeToObject.proto */
static CYTHON_INLINE PyObject *__pyx_memview_get_nn___pyx_t_5numpy_int32_t(const char *itemp);
static CYTHON_INLINE int __pyx_memview_set_nn___pyx_t_5numpy_int32_t(const char *itemp, PyObject *obj);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_npy_uint64(npy_uint64 value);

/* MemviewDtypeToObject.proto */
static CYTHON_INLINE PyObject *__pyx_memview_get_nn___pyx_t_5numpy_uint64_t(const char *itemp);
static CYTHON_INLINE int __pyx_memview_set_nn___pyx_t_5numpy_uint64_t(const char *itemp, PyObject *obj);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

/* RealImag.proto */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    #define __Pyx_CREAL(z) ((z).real())
    #define __Pyx_CIMAG(z) ((z).imag())
  #else
    #define __Pyx_CREAL(z) (__real__(z))
    #define __Pyx_CIMAG(z) (__imag__(z))
  #endif
#else
    #define __Pyx_CREAL(z) ((z).real)
    #define __Pyx_CIMAG(z) ((z).imag)
#endif
#if defined(__cplusplus) && CYTHON_CCOMPLEX\
        && (defined(_WIN32) || defined(__clang__) || (defined(__GNUC__) && (__GNUC__ >= 5 || __GNUC__ == 4 && __GNUC_MINOR__ >= 4 )) || __cplusplus >= 201103)
    #define __Pyx_SET_CREAL(z,x) ((z).real(x))
    #define __Pyx_SET_CIMAG(z,y) ((z).imag(y))
#else
    #define __Pyx_SET_CREAL(z,x) __Pyx_CREAL(z) = (x)
    #define __Pyx_SET_CIMAG(z,y) __Pyx_CIMAG(z) = (y)
#endif

/* Arithmetic.proto */
#if CYTHON_CCOMPLEX
    #define __Pyx_c_eq_float(a, b)   ((a)==(b))
    #define __Pyx_c_sum_float(a, b)  ((a)+(b))
    #define __Pyx_c_diff_float(a, b) ((a)-(b))
    #define __Pyx_c_prod_float(a, b) ((a)*(b))
    #define __Pyx_c_quot_float(a, b) ((a)/(b))
    #define __Pyx_c_neg_float(a)     (-(a))
  #ifdef __cplusplus
    #define __Pyx_c_is_zero_float(z) ((z)==(float)0)
    #define __Pyx_c_conj_float(z)    (::std::conj(z))
    #if 1
        #define __Pyx_c_abs_float(z)     (::std::abs(z))
        #define __Pyx_c_pow_float(a, b)  (::std::pow(a, b))
    #endif
  #else
    #define __Pyx_c_is_zero_float(z) ((z)==0)
    #define __Pyx_c_conj_float(z)    (conjf(z))
    #if 1
        #define __Pyx_c_abs_float(z)     (cabsf(z))
        #define __Pyx_c_pow_float(a, b)  (cpowf(a, b))
    #endif
 #endif
#else
    static CYTHON_INLINE int __Pyx_c_eq_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_sum_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_diff_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_prod_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quot_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_neg_float(__pyx_t_float_complex);
    static CYTHON_INLINE int __Pyx_c_is_zero_float(__pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_conj_float(__pyx_t_float_complex);
    #if 1
        static CYTHON_INLINE float __Pyx_c_abs_float(__pyx_t_float_complex);
        static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_pow_float(__pyx_t_float_complex, __pyx_t_float_complex);
    #endif
#endif

/* Arithmetic.proto */
#if CYTHON_CCOMPLEX
    #define __Pyx_c_eq_double(a, b)   ((a)==(b))
    #define __Pyx_c_sum_double(a, b)  ((a)+(b))
    #define __Pyx_c_diff_double(a, b) ((a)-(b))
    #define __Pyx_c_prod_double(a, b) ((a)*(b))
    #define __Pyx_c_quot_double(a, b) ((a)/(b))
    #define __Pyx_c_neg_double(a)     (-(a))
  #ifdef __cplusplus
    #define __Pyx_c_is_zero_double(z) ((z)==(double)0)
    #define __Pyx_c_conj_double(z)    (::std::conj(z))
    #if 1
        #define __Pyx_c_abs_double(z)     (::std::abs(z))
        #define __Pyx_c_pow_double(a, b)  (::std::pow(a, b))
    #endif
  #else
    #define __Pyx_c_is_zero_double(z) ((z)==0)
    #define __Pyx_c_conj_double(z)    (conj(z))
    #if 1
        #define __Pyx_c_abs_double(z)     (cabs(z))
        #define __Pyx_c_pow_double(a, b)  (cpow(a, b))
    #endif
 #endif
#else
    static CYTHON_INLINE int __Pyx_c_eq_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_sum_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_diff_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_prod_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_neg_double(__pyx_t_double_complex);
    static CYTHON_INLINE int __Pyx_c_is_zero_double(__pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_conj_double(__pyx_t_double_complex);
    #if 1
        static CYTHON_INLINE double __Pyx_c_abs_double(__pyx_t_double_complex);
        static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_pow_double(__pyx_t_double_complex, __pyx_t_double_complex);
    #endif
#endif

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_enum__NPY_TYPES(enum NPY_TYPES value);

/* MemviewSliceCopyTemplate.proto */
static __Pyx_memviewslice
__pyx_memoryview_copy_new_contig(const __Pyx_memviewslice *from_mvs,
                                 const char *mode, int ndim,
                                 size_t sizeof_dtype, int contig_flag,
                                 int dtype_is_object);

/* CIntFromPy.proto */
static CYTHON_INLINE npy_int64 __Pyx_PyInt_As_npy_int64(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE npy_int32 __Pyx_PyInt_As_npy_int32(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE npy_uint64 __Pyx_PyInt_As_npy_uint64(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

/* BytesContains.proto */
static CYTHON_INLINE int __Pyx_BytesContains(PyObject* bytes, char character);

/* ImportNumPyArray.proto */
static PyObject *__pyx_numpy_ndarray = NULL;
static PyObject* __Pyx_ImportNumPyArrayTypeIfAvailable(void);

/* CIntFromPy.proto */
static CYTHON_INLINE char __Pyx_PyInt_As_char(PyObject *);

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(void);

/* FunctionImport.proto */
static int __Pyx_ImportFunction(PyObject *module, const char *funcname, void (**f)(void), const char *sig);

/* InitStrings.proto */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);

static PyObject *__pyx_f_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_classify_intervals(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode *__pyx_v_self, __Pyx_memviewslice __pyx_v_left, __Pyx_memviewslice __pyx_v_right); /* proto*/
static PyObject *__pyx_f_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_new_child_node(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode *__pyx_v_self, PyArrayObject *__pyx_v_left, PyArrayObject *__pyx_v_right, PyArrayObject *__pyx_v_indices, PyArrayObject *__pyx_v_subset); /* proto*/
static PyObject *__pyx_fuse_0__pyx_f_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float64_t __pyx_v_point, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_fuse_1__pyx_f_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float32_t __pyx_v_point, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_fuse_2__pyx_f_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int64_t __pyx_v_point, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_fuse_3__pyx_f_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int32_t __pyx_v_point, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_fuse_4__pyx_f_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_uint64_t __pyx_v_point, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_classify_intervals(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode *__pyx_v_self, __Pyx_memviewslice __pyx_v_left, __Pyx_memviewslice __pyx_v_right); /* proto*/
static PyObject *__pyx_f_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_new_child_node(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode *__pyx_v_self, PyArrayObject *__pyx_v_left, PyArrayObject *__pyx_v_right, PyArrayObject *__pyx_v_indices, PyArrayObject *__pyx_v_subset); /* proto*/
static PyObject *__pyx_fuse_0__pyx_f_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float64_t __pyx_v_point, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_fuse_1__pyx_f_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float32_t __pyx_v_point, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_fuse_2__pyx_f_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int64_t __pyx_v_point, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_fuse_3__pyx_f_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int32_t __pyx_v_point, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_fuse_4__pyx_f_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_uint64_t __pyx_v_point, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_classify_intervals(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode *__pyx_v_self, __Pyx_memviewslice __pyx_v_left, __Pyx_memviewslice __pyx_v_right); /* proto*/
static PyObject *__pyx_f_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_new_child_node(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode *__pyx_v_self, PyArrayObject *__pyx_v_left, PyArrayObject *__pyx_v_right, PyArrayObject *__pyx_v_indices, PyArrayObject *__pyx_v_subset); /* proto*/
static PyObject *__pyx_fuse_0__pyx_f_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float64_t __pyx_v_point, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_fuse_1__pyx_f_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float32_t __pyx_v_point, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_fuse_2__pyx_f_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int64_t __pyx_v_point, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_fuse_3__pyx_f_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int32_t __pyx_v_point, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_fuse_4__pyx_f_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_uint64_t __pyx_v_point, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_classify_intervals(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode *__pyx_v_self, __Pyx_memviewslice __pyx_v_left, __Pyx_memviewslice __pyx_v_right); /* proto*/
static PyObject *__pyx_f_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_new_child_node(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode *__pyx_v_self, PyArrayObject *__pyx_v_left, PyArrayObject *__pyx_v_right, PyArrayObject *__pyx_v_indices, PyArrayObject *__pyx_v_subset); /* proto*/
static PyObject *__pyx_fuse_0__pyx_f_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float64_t __pyx_v_point, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_fuse_1__pyx_f_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float32_t __pyx_v_point, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_fuse_2__pyx_f_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int64_t __pyx_v_point, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_fuse_3__pyx_f_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int32_t __pyx_v_point, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_fuse_4__pyx_f_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_uint64_t __pyx_v_point, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_classify_intervals(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode *__pyx_v_self, __Pyx_memviewslice __pyx_v_left, __Pyx_memviewslice __pyx_v_right); /* proto*/
static PyObject *__pyx_f_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_new_child_node(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode *__pyx_v_self, PyArrayObject *__pyx_v_left, PyArrayObject *__pyx_v_right, PyArrayObject *__pyx_v_indices, PyArrayObject *__pyx_v_subset); /* proto*/
static PyObject *__pyx_fuse_0__pyx_f_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float64_t __pyx_v_point, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_fuse_1__pyx_f_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float32_t __pyx_v_point, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_fuse_2__pyx_f_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int64_t __pyx_v_point, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_fuse_3__pyx_f_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int32_t __pyx_v_point, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_fuse_4__pyx_f_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_uint64_t __pyx_v_point, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_classify_intervals(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode *__pyx_v_self, __Pyx_memviewslice __pyx_v_left, __Pyx_memviewslice __pyx_v_right); /* proto*/
static PyObject *__pyx_f_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_new_child_node(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode *__pyx_v_self, PyArrayObject *__pyx_v_left, PyArrayObject *__pyx_v_right, PyArrayObject *__pyx_v_indices, PyArrayObject *__pyx_v_subset); /* proto*/
static PyObject *__pyx_fuse_0__pyx_f_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float64_t __pyx_v_point, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_fuse_1__pyx_f_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float32_t __pyx_v_point, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_fuse_2__pyx_f_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int64_t __pyx_v_point, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_fuse_3__pyx_f_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int32_t __pyx_v_point, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_fuse_4__pyx_f_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_uint64_t __pyx_v_point, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_classify_intervals(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode *__pyx_v_self, __Pyx_memviewslice __pyx_v_left, __Pyx_memviewslice __pyx_v_right); /* proto*/
static PyObject *__pyx_f_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_new_child_node(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode *__pyx_v_self, PyArrayObject *__pyx_v_left, PyArrayObject *__pyx_v_right, PyArrayObject *__pyx_v_indices, PyArrayObject *__pyx_v_subset); /* proto*/
static PyObject *__pyx_fuse_0__pyx_f_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float64_t __pyx_v_point, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_fuse_1__pyx_f_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float32_t __pyx_v_point, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_fuse_2__pyx_f_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int64_t __pyx_v_point, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_fuse_3__pyx_f_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int32_t __pyx_v_point, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_fuse_4__pyx_f_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_uint64_t __pyx_v_point, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_classify_intervals(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode *__pyx_v_self, __Pyx_memviewslice __pyx_v_left, __Pyx_memviewslice __pyx_v_right); /* proto*/
static PyObject *__pyx_f_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_new_child_node(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode *__pyx_v_self, PyArrayObject *__pyx_v_left, PyArrayObject *__pyx_v_right, PyArrayObject *__pyx_v_indices, PyArrayObject *__pyx_v_subset); /* proto*/
static PyObject *__pyx_fuse_0__pyx_f_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float64_t __pyx_v_point, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_fuse_1__pyx_f_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float32_t __pyx_v_point, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_fuse_2__pyx_f_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int64_t __pyx_v_point, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_fuse_3__pyx_f_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int32_t __pyx_v_point, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_fuse_4__pyx_f_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_uint64_t __pyx_v_point, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_classify_intervals(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode *__pyx_v_self, __Pyx_memviewslice __pyx_v_left, __Pyx_memviewslice __pyx_v_right); /* proto*/
static PyObject *__pyx_f_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_new_child_node(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode *__pyx_v_self, PyArrayObject *__pyx_v_left, PyArrayObject *__pyx_v_right, PyArrayObject *__pyx_v_indices, PyArrayObject *__pyx_v_subset); /* proto*/
static PyObject *__pyx_fuse_0__pyx_f_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float64_t __pyx_v_point, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_fuse_1__pyx_f_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float32_t __pyx_v_point, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_fuse_2__pyx_f_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int64_t __pyx_v_point, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_fuse_3__pyx_f_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int32_t __pyx_v_point, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_fuse_4__pyx_f_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_uint64_t __pyx_v_point, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_classify_intervals(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode *__pyx_v_self, __Pyx_memviewslice __pyx_v_left, __Pyx_memviewslice __pyx_v_right); /* proto*/
static PyObject *__pyx_f_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_new_child_node(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode *__pyx_v_self, PyArrayObject *__pyx_v_left, PyArrayObject *__pyx_v_right, PyArrayObject *__pyx_v_indices, PyArrayObject *__pyx_v_subset); /* proto*/
static PyObject *__pyx_fuse_0__pyx_f_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float64_t __pyx_v_point, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_fuse_1__pyx_f_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float32_t __pyx_v_point, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_fuse_2__pyx_f_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int64_t __pyx_v_point, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_fuse_3__pyx_f_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int32_t __pyx_v_point, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_fuse_4__pyx_f_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_uint64_t __pyx_v_point, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_classify_intervals(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode *__pyx_v_self, __Pyx_memviewslice __pyx_v_left, __Pyx_memviewslice __pyx_v_right); /* proto*/
static PyObject *__pyx_f_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_new_child_node(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode *__pyx_v_self, PyArrayObject *__pyx_v_left, PyArrayObject *__pyx_v_right, PyArrayObject *__pyx_v_indices, PyArrayObject *__pyx_v_subset); /* proto*/
static PyObject *__pyx_fuse_0__pyx_f_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float64_t __pyx_v_point, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_fuse_1__pyx_f_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float32_t __pyx_v_point, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_fuse_2__pyx_f_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int64_t __pyx_v_point, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_fuse_3__pyx_f_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int32_t __pyx_v_point, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_fuse_4__pyx_f_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_uint64_t __pyx_v_point, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_classify_intervals(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode *__pyx_v_self, __Pyx_memviewslice __pyx_v_left, __Pyx_memviewslice __pyx_v_right); /* proto*/
static PyObject *__pyx_f_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_new_child_node(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode *__pyx_v_self, PyArrayObject *__pyx_v_left, PyArrayObject *__pyx_v_right, PyArrayObject *__pyx_v_indices, PyArrayObject *__pyx_v_subset); /* proto*/
static PyObject *__pyx_fuse_0__pyx_f_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float64_t __pyx_v_point, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_fuse_1__pyx_f_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float32_t __pyx_v_point, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_fuse_2__pyx_f_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int64_t __pyx_v_point, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_fuse_3__pyx_f_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int32_t __pyx_v_point, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_fuse_4__pyx_f_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_uint64_t __pyx_v_point, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_classify_intervals(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode *__pyx_v_self, __Pyx_memviewslice __pyx_v_left, __Pyx_memviewslice __pyx_v_right); /* proto*/
static PyObject *__pyx_f_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_new_child_node(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode *__pyx_v_self, PyArrayObject *__pyx_v_left, PyArrayObject *__pyx_v_right, PyArrayObject *__pyx_v_indices, PyArrayObject *__pyx_v_subset); /* proto*/
static PyObject *__pyx_fuse_0__pyx_f_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float64_t __pyx_v_point, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_fuse_1__pyx_f_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float32_t __pyx_v_point, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_fuse_2__pyx_f_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int64_t __pyx_v_point, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_fuse_3__pyx_f_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int32_t __pyx_v_point, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_fuse_4__pyx_f_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_uint64_t __pyx_v_point, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_classify_intervals(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode *__pyx_v_self, __Pyx_memviewslice __pyx_v_left, __Pyx_memviewslice __pyx_v_right); /* proto*/
static PyObject *__pyx_f_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_new_child_node(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode *__pyx_v_self, PyArrayObject *__pyx_v_left, PyArrayObject *__pyx_v_right, PyArrayObject *__pyx_v_indices, PyArrayObject *__pyx_v_subset); /* proto*/
static PyObject *__pyx_fuse_0__pyx_f_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float64_t __pyx_v_point, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_fuse_1__pyx_f_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float32_t __pyx_v_point, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_fuse_2__pyx_f_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int64_t __pyx_v_point, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_fuse_3__pyx_f_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int32_t __pyx_v_point, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_fuse_4__pyx_f_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_uint64_t __pyx_v_point, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_classify_intervals(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode *__pyx_v_self, __Pyx_memviewslice __pyx_v_left, __Pyx_memviewslice __pyx_v_right); /* proto*/
static PyObject *__pyx_f_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_new_child_node(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode *__pyx_v_self, PyArrayObject *__pyx_v_left, PyArrayObject *__pyx_v_right, PyArrayObject *__pyx_v_indices, PyArrayObject *__pyx_v_subset); /* proto*/
static PyObject *__pyx_fuse_0__pyx_f_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float64_t __pyx_v_point, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_fuse_1__pyx_f_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float32_t __pyx_v_point, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_fuse_2__pyx_f_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int64_t __pyx_v_point, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_fuse_3__pyx_f_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int32_t __pyx_v_point, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_fuse_4__pyx_f_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_uint64_t __pyx_v_point, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_classify_intervals(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode *__pyx_v_self, __Pyx_memviewslice __pyx_v_left, __Pyx_memviewslice __pyx_v_right); /* proto*/
static PyObject *__pyx_f_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_new_child_node(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode *__pyx_v_self, PyArrayObject *__pyx_v_left, PyArrayObject *__pyx_v_right, PyArrayObject *__pyx_v_indices, PyArrayObject *__pyx_v_subset); /* proto*/
static PyObject *__pyx_fuse_0__pyx_f_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float64_t __pyx_v_point, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_fuse_1__pyx_f_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float32_t __pyx_v_point, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_fuse_2__pyx_f_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int64_t __pyx_v_point, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_fuse_3__pyx_f_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int32_t __pyx_v_point, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_fuse_4__pyx_f_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_uint64_t __pyx_v_point, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_classify_intervals(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode *__pyx_v_self, __Pyx_memviewslice __pyx_v_left, __Pyx_memviewslice __pyx_v_right); /* proto*/
static PyObject *__pyx_f_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_new_child_node(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode *__pyx_v_self, PyArrayObject *__pyx_v_left, PyArrayObject *__pyx_v_right, PyArrayObject *__pyx_v_indices, PyArrayObject *__pyx_v_subset); /* proto*/
static PyObject *__pyx_fuse_0__pyx_f_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float64_t __pyx_v_point, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_fuse_1__pyx_f_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float32_t __pyx_v_point, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_fuse_2__pyx_f_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int64_t __pyx_v_point, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_fuse_3__pyx_f_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int32_t __pyx_v_point, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_fuse_4__pyx_f_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_uint64_t __pyx_v_point, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_classify_intervals(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode *__pyx_v_self, __Pyx_memviewslice __pyx_v_left, __Pyx_memviewslice __pyx_v_right); /* proto*/
static PyObject *__pyx_f_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_new_child_node(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode *__pyx_v_self, PyArrayObject *__pyx_v_left, PyArrayObject *__pyx_v_right, PyArrayObject *__pyx_v_indices, PyArrayObject *__pyx_v_subset); /* proto*/
static PyObject *__pyx_fuse_0__pyx_f_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float64_t __pyx_v_point, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_fuse_1__pyx_f_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float32_t __pyx_v_point, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_fuse_2__pyx_f_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int64_t __pyx_v_point, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_fuse_3__pyx_f_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int32_t __pyx_v_point, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_fuse_4__pyx_f_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_uint64_t __pyx_v_point, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_classify_intervals(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode *__pyx_v_self, __Pyx_memviewslice __pyx_v_left, __Pyx_memviewslice __pyx_v_right); /* proto*/
static PyObject *__pyx_f_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_new_child_node(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode *__pyx_v_self, PyArrayObject *__pyx_v_left, PyArrayObject *__pyx_v_right, PyArrayObject *__pyx_v_indices, PyArrayObject *__pyx_v_subset); /* proto*/
static PyObject *__pyx_fuse_0__pyx_f_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float64_t __pyx_v_point, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_fuse_1__pyx_f_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float32_t __pyx_v_point, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_fuse_2__pyx_f_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int64_t __pyx_v_point, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_fuse_3__pyx_f_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int32_t __pyx_v_point, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_fuse_4__pyx_f_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_uint64_t __pyx_v_point, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_classify_intervals(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode *__pyx_v_self, __Pyx_memviewslice __pyx_v_left, __Pyx_memviewslice __pyx_v_right); /* proto*/
static PyObject *__pyx_f_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_new_child_node(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode *__pyx_v_self, PyArrayObject *__pyx_v_left, PyArrayObject *__pyx_v_right, PyArrayObject *__pyx_v_indices, PyArrayObject *__pyx_v_subset); /* proto*/
static PyObject *__pyx_fuse_0__pyx_f_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float64_t __pyx_v_point, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_fuse_1__pyx_f_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float32_t __pyx_v_point, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_fuse_2__pyx_f_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int64_t __pyx_v_point, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_fuse_3__pyx_f_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int32_t __pyx_v_point, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_fuse_4__pyx_f_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_uint64_t __pyx_v_point, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_array_get_memview(struct __pyx_array_obj *__pyx_v_self); /* proto*/
static char *__pyx_memoryview_get_item_pointer(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_index); /* proto*/
static PyObject *__pyx_memoryview_is_slice(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_obj); /* proto*/
static PyObject *__pyx_memoryview_setitem_slice_assignment(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_dst, PyObject *__pyx_v_src); /* proto*/
static PyObject *__pyx_memoryview_setitem_slice_assign_scalar(struct __pyx_memoryview_obj *__pyx_v_self, struct __pyx_memoryview_obj *__pyx_v_dst, PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_memoryview_setitem_indexed(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_index, PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_memoryview_convert_item_to_object(struct __pyx_memoryview_obj *__pyx_v_self, char *__pyx_v_itemp); /* proto*/
static PyObject *__pyx_memoryview_assign_item_from_object(struct __pyx_memoryview_obj *__pyx_v_self, char *__pyx_v_itemp, PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_memoryviewslice_convert_item_to_object(struct __pyx_memoryviewslice_obj *__pyx_v_self, char *__pyx_v_itemp); /* proto*/
static PyObject *__pyx_memoryviewslice_assign_item_from_object(struct __pyx_memoryviewslice_obj *__pyx_v_self, char *__pyx_v_itemp, PyObject *__pyx_v_value); /* proto*/

/* Module declarations from 'libc.string' */

/* Module declarations from 'libc.stdio' */

/* Module declarations from '__builtin__' */

/* Module declarations from 'cpython.type' */
static PyTypeObject *__pyx_ptype_7cpython_4type_type = 0;

/* Module declarations from 'cpython.buffer' */

/* Module declarations from 'cpython.version' */

/* Module declarations from 'cpython.ref' */

/* Module declarations from 'cpython.exc' */

/* Module declarations from 'cpython.module' */

/* Module declarations from 'cpython.mem' */

/* Module declarations from 'cpython.tuple' */

/* Module declarations from 'cpython.list' */

/* Module declarations from 'cpython.sequence' */

/* Module declarations from 'cpython.mapping' */

/* Module declarations from 'cpython.iterator' */

/* Module declarations from 'cpython.number' */

/* Module declarations from 'cpython.int' */

/* Module declarations from '__builtin__' */

/* Module declarations from 'cpython.bool' */
static PyTypeObject *__pyx_ptype_7cpython_4bool_bool = 0;

/* Module declarations from 'cpython.long' */

/* Module declarations from 'cpython.float' */

/* Module declarations from '__builtin__' */

/* Module declarations from 'cpython.complex' */
static PyTypeObject *__pyx_ptype_7cpython_7complex_complex = 0;

/* Module declarations from 'cpython.string' */

/* Module declarations from 'cpython.unicode' */

/* Module declarations from 'cpython.dict' */

/* Module declarations from 'cpython.instance' */

/* Module declarations from 'cpython.function' */

/* Module declarations from 'cpython.method' */

/* Module declarations from 'cpython.weakref' */

/* Module declarations from 'cpython.getargs' */

/* Module declarations from 'cpython.pythread' */

/* Module declarations from 'cpython.pystate' */

/* Module declarations from 'cpython.cobject' */

/* Module declarations from 'cpython.oldbuffer' */

/* Module declarations from 'cpython.set' */

/* Module declarations from 'cpython.bytes' */

/* Module declarations from 'cpython.pycapsule' */

/* Module declarations from 'cpython' */

/* Module declarations from 'cpython.object' */

/* Module declarations from 'cython.view' */

/* Module declarations from 'cython' */

/* Module declarations from 'numpy' */

/* Module declarations from 'numpy' */
static PyTypeObject *__pyx_ptype_5numpy_dtype = 0;
static PyTypeObject *__pyx_ptype_5numpy_flatiter = 0;
static PyTypeObject *__pyx_ptype_5numpy_broadcast = 0;
static PyTypeObject *__pyx_ptype_5numpy_ndarray = 0;
static PyTypeObject *__pyx_ptype_5numpy_ufunc = 0;
static CYTHON_INLINE char *__pyx_f_5numpy__util_dtypestring(PyArray_Descr *, char *, char *, int *); /*proto*/
static CYTHON_INLINE int __pyx_f_5numpy_import_array(void); /*proto*/

/* Module declarations from 'pandas._libs.tslibs.util' */
static CYTHON_INLINE int __pyx_f_6pandas_5_libs_6tslibs_4util_is_integer_object(PyObject *); /*proto*/
static CYTHON_INLINE int __pyx_f_6pandas_5_libs_6tslibs_4util_is_float_object(PyObject *); /*proto*/
static CYTHON_INLINE int __pyx_f_6pandas_5_libs_6tslibs_4util_is_complex_object(PyObject *); /*proto*/
static CYTHON_INLINE int __pyx_f_6pandas_5_libs_6tslibs_4util_is_timedelta64_object(PyObject *); /*proto*/
static CYTHON_INLINE int __pyx_f_6pandas_5_libs_6tslibs_4util_is_nan(PyObject *); /*proto*/
static CYTHON_INLINE char const *__pyx_f_6pandas_5_libs_6tslibs_4util_get_c_string_buf_and_size(PyObject *, Py_ssize_t *); /*proto*/

/* Module declarations from 'pandas._libs.util' */
static CYTHON_INLINE Py_ssize_t __pyx_f_6pandas_5_libs_4util_validate_indexer(PyArrayObject *, PyObject *); /*proto*/

/* Module declarations from 'pandas._libs.khash' */

/* Module declarations from 'pandas._libs.hashtable' */
static PyTypeObject *__pyx_ptype_6pandas_5_libs_9hashtable_HashTable = 0;
static PyTypeObject *__pyx_ptype_6pandas_5_libs_9hashtable_UInt64HashTable = 0;
static PyTypeObject *__pyx_ptype_6pandas_5_libs_9hashtable_Int64HashTable = 0;
static PyTypeObject *__pyx_ptype_6pandas_5_libs_9hashtable_Float64HashTable = 0;
static PyTypeObject *__pyx_ptype_6pandas_5_libs_9hashtable_PyObjectHashTable = 0;
static PyTypeObject *__pyx_ptype_6pandas_5_libs_9hashtable_StringHashTable = 0;
static PyTypeObject *__pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector = 0;

/* Module declarations from 'pandas._libs.tslibs.timezones' */
static int (*__pyx_f_6pandas_5_libs_6tslibs_9timezones_tz_compare)(PyObject *, PyObject *, int __pyx_skip_dispatch); /*proto*/

/* Module declarations from 'pandas._libs.interval' */
static PyTypeObject *__pyx_ptype_6pandas_5_libs_8interval_IntervalMixin = 0;
static PyTypeObject *__pyx_ptype_6pandas_5_libs_8interval_Interval = 0;
static PyTypeObject *__pyx_ptype_6pandas_5_libs_8interval_IntervalTree = 0;
static PyTypeObject *__pyx_ptype_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode = 0;
static PyTypeObject *__pyx_ptype_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode = 0;
static PyTypeObject *__pyx_ptype_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode = 0;
static PyTypeObject *__pyx_ptype_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode = 0;
static PyTypeObject *__pyx_ptype_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode = 0;
static PyTypeObject *__pyx_ptype_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode = 0;
static PyTypeObject *__pyx_ptype_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode = 0;
static PyTypeObject *__pyx_ptype_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode = 0;
static PyTypeObject *__pyx_ptype_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode = 0;
static PyTypeObject *__pyx_ptype_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode = 0;
static PyTypeObject *__pyx_ptype_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode = 0;
static PyTypeObject *__pyx_ptype_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode = 0;
static PyTypeObject *__pyx_ptype_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode = 0;
static PyTypeObject *__pyx_ptype_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode = 0;
static PyTypeObject *__pyx_ptype_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode = 0;
static PyTypeObject *__pyx_ptype_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode = 0;
static PyTypeObject *__pyx_ptype_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode = 0;
static PyTypeObject *__pyx_ptype_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode = 0;
static PyTypeObject *__pyx_ptype_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode = 0;
static PyTypeObject *__pyx_ptype_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode = 0;
static PyTypeObject *__pyx_array_type = 0;
static PyTypeObject *__pyx_MemviewEnum_type = 0;
static PyTypeObject *__pyx_memoryview_type = 0;
static PyTypeObject *__pyx_memoryviewslice_type = 0;
static PyObject *generic = 0;
static PyObject *strided = 0;
static PyObject *indirect = 0;
static PyObject *contiguous = 0;
static PyObject *indirect_contiguous = 0;
static int __pyx_memoryview_thread_locks_used;
static PyThread_type_lock __pyx_memoryview_thread_locks[8];
static PyObject *__pyx_f_6pandas_5_libs_8interval__interval_like(PyObject *); /*proto*/
static PyObject *__pyx_f_6pandas_5_libs_8interval_take(PyArrayObject *, PyArrayObject *); /*proto*/
static PyObject *__pyx_f_6pandas_5_libs_8interval_sort_values_and_indices(PyObject *, PyObject *, PyObject *); /*proto*/
static PyObject *__pyx_f_6pandas_5_libs_8interval___pyx_unpickle_IntervalMixin__set_state(struct __pyx_obj_6pandas_5_libs_8interval_IntervalMixin *, PyObject *); /*proto*/
static PyObject *__pyx_f_6pandas_5_libs_8interval___pyx_unpickle_IntervalTree__set_state(struct __pyx_obj_6pandas_5_libs_8interval_IntervalTree *, PyObject *); /*proto*/
static PyObject *__pyx_f_6pandas_5_libs_8interval___pyx_unpickle_Float32ClosedLeftIntervalNode__set_state(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode *, PyObject *); /*proto*/
static PyObject *__pyx_f_6pandas_5_libs_8interval___pyx_unpickle_Float32ClosedRightIntervalNode__set_state(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode *, PyObject *); /*proto*/
static PyObject *__pyx_f_6pandas_5_libs_8interval___pyx_unpickle_Float32ClosedBothIntervalNode__set_state(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode *, PyObject *); /*proto*/
static PyObject *__pyx_f_6pandas_5_libs_8interval___pyx_unpickle_Float32ClosedNeitherIntervalNode__set_state(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode *, PyObject *); /*proto*/
static PyObject *__pyx_f_6pandas_5_libs_8interval___pyx_unpickle_Float64ClosedLeftIntervalNode__set_state(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode *, PyObject *); /*proto*/
static PyObject *__pyx_f_6pandas_5_libs_8interval___pyx_unpickle_Float64ClosedRightIntervalNode__set_state(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode *, PyObject *); /*proto*/
static PyObject *__pyx_f_6pandas_5_libs_8interval___pyx_unpickle_Float64ClosedBothIntervalNode__set_state(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode *, PyObject *); /*proto*/
static PyObject *__pyx_f_6pandas_5_libs_8interval___pyx_unpickle_Float64ClosedNeitherIntervalNode__set_state(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode *, PyObject *); /*proto*/
static PyObject *__pyx_f_6pandas_5_libs_8interval___pyx_unpickle_Int32ClosedLeftIntervalNode__set_state(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode *, PyObject *); /*proto*/
static PyObject *__pyx_f_6pandas_5_libs_8interval___pyx_unpickle_Int32ClosedRightIntervalNode__set_state(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode *, PyObject *); /*proto*/
static PyObject *__pyx_f_6pandas_5_libs_8interval___pyx_unpickle_Int32ClosedBothIntervalNode__set_state(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode *, PyObject *); /*proto*/
static PyObject *__pyx_f_6pandas_5_libs_8interval___pyx_unpickle_Int32ClosedNeitherIntervalNode__set_state(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode *, PyObject *); /*proto*/
static PyObject *__pyx_f_6pandas_5_libs_8interval___pyx_unpickle_Int64ClosedLeftIntervalNode__set_state(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode *, PyObject *); /*proto*/
static PyObject *__pyx_f_6pandas_5_libs_8interval___pyx_unpickle_Int64ClosedRightIntervalNode__set_state(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode *, PyObject *); /*proto*/
static PyObject *__pyx_f_6pandas_5_libs_8interval___pyx_unpickle_Int64ClosedBothIntervalNode__set_state(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode *, PyObject *); /*proto*/
static PyObject *__pyx_f_6pandas_5_libs_8interval___pyx_unpickle_Int64ClosedNeitherIntervalNode__set_state(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode *, PyObject *); /*proto*/
static PyObject *__pyx_f_6pandas_5_libs_8interval___pyx_unpickle_Uint64ClosedLeftIntervalNode__set_state(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode *, PyObject *); /*proto*/
static PyObject *__pyx_f_6pandas_5_libs_8interval___pyx_unpickle_Uint64ClosedRightIntervalNode__set_state(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode *, PyObject *); /*proto*/
static PyObject *__pyx_f_6pandas_5_libs_8interval___pyx_unpickle_Uint64ClosedBothIntervalNode__set_state(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode *, PyObject *); /*proto*/
static PyObject *__pyx_f_6pandas_5_libs_8interval___pyx_unpickle_Uint64ClosedNeitherIntervalNode__set_state(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode *, PyObject *); /*proto*/
static struct __pyx_array_obj *__pyx_array_new(PyObject *, Py_ssize_t, char *, char *, char *); /*proto*/
static void *__pyx_align_pointer(void *, size_t); /*proto*/
static PyObject *__pyx_memoryview_new(PyObject *, int, int, __Pyx_TypeInfo *); /*proto*/
static CYTHON_INLINE int __pyx_memoryview_check(PyObject *); /*proto*/
static PyObject *_unellipsify(PyObject *, int); /*proto*/
static PyObject *assert_direct_dimensions(Py_ssize_t *, int); /*proto*/
static struct __pyx_memoryview_obj *__pyx_memview_slice(struct __pyx_memoryview_obj *, PyObject *); /*proto*/
static int __pyx_memoryview_slice_memviewslice(__Pyx_memviewslice *, Py_ssize_t, Py_ssize_t, Py_ssize_t, int, int, int *, Py_ssize_t, Py_ssize_t, Py_ssize_t, int, int, int, int); /*proto*/
static char *__pyx_pybuffer_index(Py_buffer *, char *, Py_ssize_t, Py_ssize_t); /*proto*/
static int __pyx_memslice_transpose(__Pyx_memviewslice *); /*proto*/
static PyObject *__pyx_memoryview_fromslice(__Pyx_memviewslice, int, PyObject *(*)(char *), int (*)(char *, PyObject *), int); /*proto*/
static __Pyx_memviewslice *__pyx_memoryview_get_slice_from_memoryview(struct __pyx_memoryview_obj *, __Pyx_memviewslice *); /*proto*/
static void __pyx_memoryview_slice_copy(struct __pyx_memoryview_obj *, __Pyx_memviewslice *); /*proto*/
static PyObject *__pyx_memoryview_copy_object(struct __pyx_memoryview_obj *); /*proto*/
static PyObject *__pyx_memoryview_copy_object_from_slice(struct __pyx_memoryview_obj *, __Pyx_memviewslice *); /*proto*/
static Py_ssize_t abs_py_ssize_t(Py_ssize_t); /*proto*/
static char __pyx_get_best_slice_order(__Pyx_memviewslice *, int); /*proto*/
static void _copy_strided_to_strided(char *, Py_ssize_t *, char *, Py_ssize_t *, Py_ssize_t *, Py_ssize_t *, int, size_t); /*proto*/
static void copy_strided_to_strided(__Pyx_memviewslice *, __Pyx_memviewslice *, int, size_t); /*proto*/
static Py_ssize_t __pyx_memoryview_slice_get_size(__Pyx_memviewslice *, int); /*proto*/
static Py_ssize_t __pyx_fill_contig_strides_array(Py_ssize_t *, Py_ssize_t *, Py_ssize_t, int, char); /*proto*/
static void *__pyx_memoryview_copy_data_to_temp(__Pyx_memviewslice *, __Pyx_memviewslice *, char, int); /*proto*/
static int __pyx_memoryview_err_extents(int, Py_ssize_t, Py_ssize_t); /*proto*/
static int __pyx_memoryview_err_dim(PyObject *, char *, int); /*proto*/
static int __pyx_memoryview_err(PyObject *, char *); /*proto*/
static int __pyx_memoryview_copy_contents(__Pyx_memviewslice, __Pyx_memviewslice, int, int, int); /*proto*/
static void __pyx_memoryview_broadcast_leading(__Pyx_memviewslice *, int, int); /*proto*/
static void __pyx_memoryview_refcount_copying(__Pyx_memviewslice *, int, int, int); /*proto*/
static void __pyx_memoryview_refcount_objects_in_slice_with_gil(char *, Py_ssize_t *, Py_ssize_t *, int, int); /*proto*/
static void __pyx_memoryview_refcount_objects_in_slice(char *, Py_ssize_t *, Py_ssize_t *, int, int); /*proto*/
static void __pyx_memoryview_slice_assign_scalar(__Pyx_memviewslice *, int, size_t, void *, int); /*proto*/
static void __pyx_memoryview__slice_assign_scalar(char *, Py_ssize_t *, Py_ssize_t *, int, size_t, void *); /*proto*/
static PyObject *__pyx_unpickle_Enum__set_state(struct __pyx_MemviewEnum_obj *, PyObject *); /*proto*/
static __Pyx_TypeInfo __Pyx_TypeInfo_nn___pyx_t_5numpy_float32_t = { "float32_t", NULL, sizeof(__pyx_t_5numpy_float32_t), { 0 }, 0, 'R', 0, 0 };
static __Pyx_TypeInfo __Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t = { "int64_t", NULL, sizeof(__pyx_t_5numpy_int64_t), { 0 }, 0, IS_UNSIGNED(__pyx_t_5numpy_int64_t) ? 'U' : 'I', IS_UNSIGNED(__pyx_t_5numpy_int64_t), 0 };
static __Pyx_TypeInfo __Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t = { "float64_t", NULL, sizeof(__pyx_t_5numpy_float64_t), { 0 }, 0, 'R', 0, 0 };
static __Pyx_TypeInfo __Pyx_TypeInfo_nn___pyx_t_5numpy_int32_t = { "int32_t", NULL, sizeof(__pyx_t_5numpy_int32_t), { 0 }, 0, IS_UNSIGNED(__pyx_t_5numpy_int32_t) ? 'U' : 'I', IS_UNSIGNED(__pyx_t_5numpy_int32_t), 0 };
static __Pyx_TypeInfo __Pyx_TypeInfo_nn___pyx_t_5numpy_uint64_t = { "uint64_t", NULL, sizeof(__pyx_t_5numpy_uint64_t), { 0 }, 0, IS_UNSIGNED(__pyx_t_5numpy_uint64_t) ? 'U' : 'I', IS_UNSIGNED(__pyx_t_5numpy_uint64_t), 0 };
#define __Pyx_MODULE_NAME "pandas._libs.interval"
extern int __pyx_module_is_main_pandas___libs__interval;
int __pyx_module_is_main_pandas___libs__interval = 0;

/* Implementation of 'pandas._libs.interval' */
static PyObject *__pyx_builtin_TypeError;
static PyObject *__pyx_builtin_ValueError;
static PyObject *__pyx_builtin_NotImplemented;
static PyObject *__pyx_builtin_range;
static PyObject *__pyx_builtin_KeyError;
static PyObject *__pyx_builtin_RuntimeError;
static PyObject *__pyx_builtin_ImportError;
static PyObject *__pyx_builtin_IndexError;
static PyObject *__pyx_builtin_MemoryError;
static PyObject *__pyx_builtin_enumerate;
static PyObject *__pyx_builtin_Ellipsis;
static PyObject *__pyx_builtin_id;
static const char __pyx_k_O[] = "O";
static const char __pyx_k_c[] = "c";
static const char __pyx_k_i[] = "i";
static const char __pyx_k_n[] = "n";
static const char __pyx_k_s[] = "s";
static const char __pyx_k__3[] = "<";
static const char __pyx_k__4[] = "<=";
static const char __pyx_k__5[] = ">";
static const char __pyx_k__6[] = ">=";
static const char __pyx_k__7[] = "[";
static const char __pyx_k__8[] = "(";
static const char __pyx_k__9[] = "]";
static const char __pyx_k_id[] = "id";
static const char __pyx_k_iv[] = "iv";
static const char __pyx_k_le[] = "le";
static const char __pyx_k_lt[] = "lt";
static const char __pyx_k_np[] = "np";
static const char __pyx_k_op[] = "op";
static const char __pyx_k__10[] = ")";
static const char __pyx_k__13[] = "()";
static const char __pyx_k__14[] = "|";
static const char __pyx_k__18[] = ", ";
static const char __pyx_k_any[] = "any";
static const char __pyx_k_end[] = "end";
static const char __pyx_k_key[] = "key";
static const char __pyx_k_max[] = "max";
static const char __pyx_k_min[] = "min";
static const char __pyx_k_msg[] = "msg";
static const char __pyx_k_nan[] = "nan";
static const char __pyx_k_new[] = "__new__";
static const char __pyx_k_obj[] = "obj";
static const char __pyx_k_typ[] = "typ";
static const char __pyx_k_args[] = "args";
static const char __pyx_k_base[] = "base";
static const char __pyx_k_both[] = "both";
static const char __pyx_k_dict[] = "__dict__";
static const char __pyx_k_intp[] = "intp";
static const char __pyx_k_kind[] = "kind";
static const char __pyx_k_left[] = "left";
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_mode[] = "mode";
static const char __pyx_k_name[] = "name";
static const char __pyx_k_ndim[] = "ndim";
static const char __pyx_k_pack[] = "pack";
static const char __pyx_k_self[] = "self";
static const char __pyx_k_side[] = "side";
static const char __pyx_k_size[] = "size";
static const char __pyx_k_step[] = "step";
static const char __pyx_k_stop[] = "stop";
static const char __pyx_k_test[] = "__test__";
static const char __pyx_k_ASCII[] = "ASCII";
static const char __pyx_k_class[] = "__class__";
static const char __pyx_k_dtype[] = "dtype";
static const char __pyx_k_empty[] = "empty";
static const char __pyx_k_error[] = "error";
static const char __pyx_k_flags[] = "flags";
static const char __pyx_k_int32[] = "int32";
static const char __pyx_k_int64[] = "int64";
static const char __pyx_k_isnan[] = "isnan";
static const char __pyx_k_numpy[] = "numpy";
static const char __pyx_k_other[] = "other";
static const char __pyx_k_point[] = "point";
static const char __pyx_k_query[] = "query";
static const char __pyx_k_range[] = "range";
static const char __pyx_k_right[] = "right";
static const char __pyx_k_shape[] = "shape";
static const char __pyx_k_split[] = "split";
static const char __pyx_k_start[] = "start";
static const char __pyx_k_strip[] = "strip";
static const char __pyx_k_typ_2[] = "_typ";
static const char __pyx_k_Number[] = "Number";
static const char __pyx_k_arange[] = "arange";
static const char __pyx_k_astype[] = "astype";
static const char __pyx_k_closed[] = "closed";
static const char __pyx_k_counts[] = "counts";
static const char __pyx_k_encode[] = "encode";
static const char __pyx_k_format[] = "format";
static const char __pyx_k_import[] = "__import__";
static const char __pyx_k_kwargs[] = "kwargs";
static const char __pyx_k_left_2[] = " left, ";
static const char __pyx_k_length[] = "length";
static const char __pyx_k_median[] = "median";
static const char __pyx_k_name_2[] = "__name__";
static const char __pyx_k_pandas[] = "pandas";
static const char __pyx_k_period[] = "period";
static const char __pyx_k_pickle[] = "pickle";
static const char __pyx_k_reduce[] = "__reduce__";
static const char __pyx_k_result[] = "result";
static const char __pyx_k_sorter[] = "sorter";
static const char __pyx_k_struct[] = "struct";
static const char __pyx_k_target[] = "target";
static const char __pyx_k_tzinfo[] = "tzinfo";
static const char __pyx_k_uint64[] = "uint64";
static const char __pyx_k_unique[] = "unique";
static const char __pyx_k_unpack[] = "unpack";
static const char __pyx_k_update[] = "update";
static const char __pyx_k_argsort[] = "argsort";
static const char __pyx_k_asarray[] = "asarray";
static const char __pyx_k_float32[] = "float32";
static const char __pyx_k_float64[] = "float64";
static const char __pyx_k_fortran[] = "fortran";
static const char __pyx_k_get_loc[] = "get_loc";
static const char __pyx_k_indices[] = "indices";
static const char __pyx_k_int32_t[] = "int32_t";
static const char __pyx_k_int64_t[] = "int64_t";
static const char __pyx_k_left_tz[] = "left_tz";
static const char __pyx_k_lexsort[] = "lexsort";
static const char __pyx_k_memview[] = "memview";
static const char __pyx_k_missing[] = "missing";
static const char __pyx_k_neither[] = "neither";
static const char __pyx_k_numbers[] = "numbers";
static const char __pyx_k_old_len[] = "old_len";
static const char __pyx_k_right_2[] = " right, ";
static const char __pyx_k_Ellipsis[] = "Ellipsis";
static const char __pyx_k_Interval[] = "Interval";
static const char __pyx_k_KeyError[] = "KeyError";
static const char __pyx_k_defaults[] = "defaults";
static const char __pyx_k_elements[] = " elements (";
static const char __pyx_k_getstate[] = "__getstate__";
static const char __pyx_k_interval[] = "interval";
static const char __pyx_k_itemsize[] = "itemsize";
static const char __pyx_k_key_left[] = "key_left";
static const char __pyx_k_operator[] = "operator";
static const char __pyx_k_pyx_type[] = "__pyx_type";
static const char __pyx_k_right_tz[] = "right_tz";
static const char __pyx_k_setstate[] = "__setstate__";
static const char __pyx_k_uint64_t[] = "uint64_t";
static const char __pyx_k_Timedelta[] = "Timedelta";
static const char __pyx_k_Timestamp[] = "Timestamp";
static const char __pyx_k_TypeError[] = "TypeError";
static const char __pyx_k_enumerate[] = "enumerate";
static const char __pyx_k_float32_t[] = "float32_t";
static const char __pyx_k_float64_t[] = "float64_t";
static const char __pyx_k_intervals[] = "intervals";
static const char __pyx_k_key_right[] = "key_right";
static const char __pyx_k_leaf_size[] = "leaf_size";
static const char __pyx_k_open_left[] = "open_left";
static const char __pyx_k_pyx_state[] = "__pyx_state";
static const char __pyx_k_reduce_ex[] = "__reduce_ex__";
static const char __pyx_k_repr_base[] = "_repr_base";
static const char __pyx_k_IndexError[] = "IndexError";
static const char __pyx_k_ValueError[] = "ValueError";
static const char __pyx_k_dateoffset[] = "dateoffset";
static const char __pyx_k_n_elements[] = "n_elements";
static const char __pyx_k_open_right[] = "open_right";
static const char __pyx_k_pyx_result[] = "__pyx_result";
static const char __pyx_k_pyx_vtable[] = "__pyx_vtable__";
static const char __pyx_k_short_repr[] = "_short_repr";
static const char __pyx_k_ImportError[] = "ImportError";
static const char __pyx_k_MemoryError[] = "MemoryError";
static const char __pyx_k_PickleError[] = "PickleError";
static const char __pyx_k_closed_left[] = "closed_left";
static const char __pyx_k_concatenate[] = "concatenate";
static const char __pyx_k_get_indexer[] = "get_indexer";
static const char __pyx_k_left_sorter[] = "left_sorter";
static const char __pyx_k_overlapping[] = " overlapping)>";
static const char __pyx_k_result_type[] = "result_type";
static const char __pyx_k_seen_closed[] = "seen_closed";
static const char __pyx_k_IntervalTree[] = "IntervalTree";
static const char __pyx_k_NODE_CLASSES[] = "NODE_CLASSES";
static const char __pyx_k_RuntimeError[] = "RuntimeError";
static const char __pyx_k_VALID_CLOSED[] = "_VALID_CLOSED";
static const char __pyx_k_closed_right[] = "closed_right";
static const char __pyx_k_is_monotonic[] = "is_monotonic";
static const char __pyx_k_pyx_checksum[] = "__pyx_checksum";
static const char __pyx_k_right_sorter[] = "right_sorter";
static const char __pyx_k_searchsorted[] = "searchsorted";
static const char __pyx_k_stringsource[] = "stringsource";
static const char __pyx_k_IntervalMixin[] = "IntervalMixin";
static const char __pyx_k_pyx_getbuffer[] = "__pyx_getbuffer";
static const char __pyx_k_reduce_cython[] = "__reduce_cython__";
static const char __pyx_k_NotImplemented[] = "NotImplemented";
static const char __pyx_k_View_MemoryView[] = "View.MemoryView";
static const char __pyx_k_allocate_buffer[] = "allocate_buffer";
static const char __pyx_k_dtype_is_object[] = "dtype_is_object";
static const char __pyx_k_pyx_PickleError[] = "__pyx_PickleError";
static const char __pyx_k_pyx_fuse_0query[] = "__pyx_fuse_0query";
static const char __pyx_k_pyx_fuse_1query[] = "__pyx_fuse_1query";
static const char __pyx_k_pyx_fuse_2query[] = "__pyx_fuse_2query";
static const char __pyx_k_pyx_fuse_3query[] = "__pyx_fuse_3query";
static const char __pyx_k_pyx_fuse_4query[] = "__pyx_fuse_4query";
static const char __pyx_k_setstate_cython[] = "__setstate_cython__";
static const char __pyx_k_validate_closed[] = "validate_closed";
static const char __pyx_k_pyx_unpickle_Enum[] = "__pyx_unpickle_Enum";
static const char __pyx_k_validate_endpoint[] = "_validate_endpoint";
static const char __pyx_k_cline_in_traceback[] = "cline_in_traceback";
static const char __pyx_k_closed_is_expected[] = "'{}.closed' is '{}', expected '{}'.";
static const char __pyx_k_pandas__libs_algos[] = "pandas._libs.algos";
static const char __pyx_k_strided_and_direct[] = "<strided and direct>";
static const char __pyx_k_get_partial_overlap[] = "_get_partial_overlap";
static const char __pyx_k_index_out_of_bounds[] = "index out of bounds";
static const char __pyx_k_pandas__libs_tslibs[] = "pandas._libs.tslibs";
static const char __pyx_k_IntervalTree_get_loc[] = "IntervalTree.get_loc";
static const char __pyx_k_start_left_right_end[] = "{start}{left}, {right}{end}";
static const char __pyx_k_strided_and_indirect[] = "<strided and indirect>";
static const char __pyx_k_contiguous_and_direct[] = "<contiguous and direct>";
static const char __pyx_k_pandas__libs_interval[] = "pandas._libs.interval";
static const char __pyx_k_MemoryView_of_r_object[] = "<MemoryView of %r object>";
static const char __pyx_k_get_indexer_non_unique[] = "get_indexer_non_unique";
static const char __pyx_k_MemoryView_of_r_at_0x_x[] = "<MemoryView of %r at 0x%x>";
static const char __pyx_k_contiguous_and_indirect[] = "<contiguous and indirect>";
static const char __pyx_k_Cannot_index_with_type_s[] = "Cannot index with type '%s'";
static const char __pyx_k_IntervalTree_get_indexer[] = "IntervalTree.get_indexer";
static const char __pyx_k_Invalid_shape_in_axis_d_d[] = "Invalid shape in axis %d: %d.";
static const char __pyx_k_pandas__libs_interval_pyx[] = "pandas/_libs/interval.pyx";
static const char __pyx_k_pyx_unpickle_IntervalTree[] = "__pyx_unpickle_IntervalTree";
static const char __pyx_k_Interval_overlaps_line_427[] = "Interval.overlaps (line 427)";
static const char __pyx_k_pyx_unpickle_IntervalMixin[] = "__pyx_unpickle_IntervalMixin";
static const char __pyx_k_Int32ClosedBothIntervalNode[] = "Int32ClosedBothIntervalNode";
static const char __pyx_k_Int32ClosedLeftIntervalNode[] = "Int32ClosedLeftIntervalNode";
static const char __pyx_k_Int64ClosedBothIntervalNode[] = "Int64ClosedBothIntervalNode";
static const char __pyx_k_Int64ClosedLeftIntervalNode[] = "Int64ClosedLeftIntervalNode";
static const char __pyx_k_No_matching_signature_found[] = "No matching signature found";
static const char __pyx_k_invalid_option_for_closed_s[] = "invalid option for 'closed': %s";
static const char __pyx_k_itemsize_0_for_cython_array[] = "itemsize <= 0 for cython.array";
static const char __pyx_k_ndarray_is_not_C_contiguous[] = "ndarray is not C contiguous";
static const char __pyx_k_Int32ClosedRightIntervalNode[] = "Int32ClosedRightIntervalNode";
static const char __pyx_k_Int64ClosedRightIntervalNode[] = "Int64ClosedRightIntervalNode";
static const char __pyx_k_Uint64ClosedBothIntervalNode[] = "Uint64ClosedBothIntervalNode";
static const char __pyx_k_Uint64ClosedLeftIntervalNode[] = "Uint64ClosedLeftIntervalNode";
static const char __pyx_k_intervals_to_interval_bounds[] = "intervals_to_interval_bounds";
static const char __pyx_k_Float32ClosedBothIntervalNode[] = "Float32ClosedBothIntervalNode";
static const char __pyx_k_Float32ClosedLeftIntervalNode[] = "Float32ClosedLeftIntervalNode";
static const char __pyx_k_Float64ClosedBothIntervalNode[] = "Float64ClosedBothIntervalNode";
static const char __pyx_k_Float64ClosedLeftIntervalNode[] = "Float64ClosedLeftIntervalNode";
static const char __pyx_k_Uint64ClosedRightIntervalNode[] = "Uint64ClosedRightIntervalNode";
static const char __pyx_k_pandas__libs_intervaltree_pxi[] = "pandas/_libs/intervaltree.pxi";
static const char __pyx_k_unable_to_allocate_array_data[] = "unable to allocate array data.";
static const char __pyx_k_Float32ClosedRightIntervalNode[] = "<Float32ClosedRightIntervalNode: %s elements (terminal)>";
static const char __pyx_k_Float64ClosedRightIntervalNode[] = "<Float64ClosedRightIntervalNode: %s elements (terminal)>";
static const char __pyx_k_Int32ClosedNeitherIntervalNode[] = "<Int32ClosedNeitherIntervalNode: %s elements (terminal)>";
static const char __pyx_k_Int32ClosedRightIntervalNode_s[] = "<Int32ClosedRightIntervalNode: %s elements (terminal)>";
static const char __pyx_k_Int64ClosedNeitherIntervalNode[] = "<Int64ClosedNeitherIntervalNode: %s elements (terminal)>";
static const char __pyx_k_Int64ClosedRightIntervalNode_s[] = "<Int64ClosedRightIntervalNode: %s elements (terminal)>";
static const char __pyx_k_Uint64ClosedBothIntervalNode_s[] = "<Uint64ClosedBothIntervalNode: %s elements (terminal)>";
static const char __pyx_k_Uint64ClosedLeftIntervalNode_s[] = "<Uint64ClosedLeftIntervalNode: %s elements (terminal)>";
static const char __pyx_k_contains___not_defined_for_two[] = "__contains__ not defined for two intervals";
static const char __pyx_k_pandas__libs_tslibs_timedeltas[] = "pandas._libs.tslibs.timedeltas";
static const char __pyx_k_pyx_unpickle_Float32ClosedBoth[] = "__pyx_unpickle_Float32ClosedBothIntervalNode";
static const char __pyx_k_pyx_unpickle_Float32ClosedLeft[] = "__pyx_unpickle_Float32ClosedLeftIntervalNode";
static const char __pyx_k_pyx_unpickle_Float32ClosedNeit[] = "__pyx_unpickle_Float32ClosedNeitherIntervalNode";
static const char __pyx_k_pyx_unpickle_Float32ClosedRigh[] = "__pyx_unpickle_Float32ClosedRightIntervalNode";
static const char __pyx_k_pyx_unpickle_Float64ClosedBoth[] = "__pyx_unpickle_Float64ClosedBothIntervalNode";
static const char __pyx_k_pyx_unpickle_Float64ClosedLeft[] = "__pyx_unpickle_Float64ClosedLeftIntervalNode";
static const char __pyx_k_pyx_unpickle_Float64ClosedNeit[] = "__pyx_unpickle_Float64ClosedNeitherIntervalNode";
static const char __pyx_k_pyx_unpickle_Float64ClosedRigh[] = "__pyx_unpickle_Float64ClosedRightIntervalNode";
static const char __pyx_k_pyx_unpickle_Int32ClosedBothIn[] = "__pyx_unpickle_Int32ClosedBothIntervalNode";
static const char __pyx_k_pyx_unpickle_Int32ClosedLeftIn[] = "__pyx_unpickle_Int32ClosedLeftIntervalNode";
static const char __pyx_k_pyx_unpickle_Int32ClosedNeithe[] = "__pyx_unpickle_Int32ClosedNeitherIntervalNode";
static const char __pyx_k_pyx_unpickle_Int32ClosedRightI[] = "__pyx_unpickle_Int32ClosedRightIntervalNode";
static const char __pyx_k_pyx_unpickle_Int64ClosedBothIn[] = "__pyx_unpickle_Int64ClosedBothIntervalNode";
static const char __pyx_k_pyx_unpickle_Int64ClosedLeftIn[] = "__pyx_unpickle_Int64ClosedLeftIntervalNode";
static const char __pyx_k_pyx_unpickle_Int64ClosedNeithe[] = "__pyx_unpickle_Int64ClosedNeitherIntervalNode";
static const char __pyx_k_pyx_unpickle_Int64ClosedRightI[] = "__pyx_unpickle_Int64ClosedRightIntervalNode";
static const char __pyx_k_pyx_unpickle_Uint64ClosedBothI[] = "__pyx_unpickle_Uint64ClosedBothIntervalNode";
static const char __pyx_k_pyx_unpickle_Uint64ClosedLeftI[] = "__pyx_unpickle_Uint64ClosedLeftIntervalNode";
static const char __pyx_k_pyx_unpickle_Uint64ClosedNeith[] = "__pyx_unpickle_Uint64ClosedNeitherIntervalNode";
static const char __pyx_k_pyx_unpickle_Uint64ClosedRight[] = "__pyx_unpickle_Uint64ClosedRightIntervalNode";
static const char __pyx_k_strided_and_direct_or_indirect[] = "<strided and direct or indirect>";
static const char __pyx_k_Check_whether_two_Interval_obje[] = "\n        Check whether two Interval objects overlap.\n\n        Two intervals overlap if they share a common point, including closed\n        endpoints. Intervals that only have an open endpoint in common do not\n        overlap.\n\n        .. versionadded:: 0.24.0\n\n        Parameters\n        ----------\n        other : Interval\n            The interval to check against for an overlap.\n\n        Returns\n        -------\n        bool\n            ``True`` if the two intervals overlap, else ``False``.\n\n        See Also\n        --------\n        IntervalArray.overlaps : The corresponding method for IntervalArray.\n        IntervalIndex.overlaps : The corresponding method for IntervalIndex.\n\n        Examples\n        --------\n        >>> i1 = pd.Interval(0, 2)\n        >>> i2 = pd.Interval(1, 3)\n        >>> i1.overlaps(i2)\n        True\n        >>> i3 = pd.Interval(4, 5)\n        >>> i1.overlaps(i3)\n        False\n\n        Intervals that share closed endpoints overlap:\n\n        >>> i4 = pd.Interval(0, 1, closed='both')\n        >>> i5 = pd.Interval(1, 2, closed='both')\n        >>> i4.overlaps(i5)\n        True\n\n        Intervals that only have an open endpoint in common do not overlap:\n\n        >>> i6 = pd.Interval(1, 2, closed='neither')\n        >>> i4.overlaps(i6)\n        False\n        ";
static const char __pyx_k_Float32ClosedBothIntervalNode_2[] = "Float32ClosedBothIntervalNode.__pyx_fuse_0query";
static const char __pyx_k_Float32ClosedBothIntervalNode_3[] = "Float32ClosedBothIntervalNode.__pyx_fuse_1query";
static const char __pyx_k_Float32ClosedBothIntervalNode_4[] = "Float32ClosedBothIntervalNode.__pyx_fuse_2query";
static const char __pyx_k_Float32ClosedBothIntervalNode_5[] = "Float32ClosedBothIntervalNode.__pyx_fuse_3query";
static const char __pyx_k_Float32ClosedBothIntervalNode_6[] = "Float32ClosedBothIntervalNode.__pyx_fuse_4query";
static const char __pyx_k_Float32ClosedBothIntervalNode_p[] = "<Float32ClosedBothIntervalNode: pivot ";
static const char __pyx_k_Float32ClosedBothIntervalNode_s[] = "<Float32ClosedBothIntervalNode: %s elements (terminal)>";
static const char __pyx_k_Float32ClosedLeftIntervalNode_2[] = "Float32ClosedLeftIntervalNode.__pyx_fuse_0query";
static const char __pyx_k_Float32ClosedLeftIntervalNode_3[] = "Float32ClosedLeftIntervalNode.__pyx_fuse_1query";
static const char __pyx_k_Float32ClosedLeftIntervalNode_4[] = "Float32ClosedLeftIntervalNode.__pyx_fuse_2query";
static const char __pyx_k_Float32ClosedLeftIntervalNode_5[] = "Float32ClosedLeftIntervalNode.__pyx_fuse_3query";
static const char __pyx_k_Float32ClosedLeftIntervalNode_6[] = "Float32ClosedLeftIntervalNode.__pyx_fuse_4query";
static const char __pyx_k_Float32ClosedLeftIntervalNode_p[] = "<Float32ClosedLeftIntervalNode: pivot ";
static const char __pyx_k_Float32ClosedLeftIntervalNode_s[] = "<Float32ClosedLeftIntervalNode: %s elements (terminal)>";
static const char __pyx_k_Float32ClosedNeitherIntervalNod[] = "<Float32ClosedNeitherIntervalNode: %s elements (terminal)>";
static const char __pyx_k_Float64ClosedBothIntervalNode_2[] = "Float64ClosedBothIntervalNode.__pyx_fuse_0query";
static const char __pyx_k_Float64ClosedBothIntervalNode_3[] = "Float64ClosedBothIntervalNode.__pyx_fuse_1query";
static const char __pyx_k_Float64ClosedBothIntervalNode_4[] = "Float64ClosedBothIntervalNode.__pyx_fuse_2query";
static const char __pyx_k_Float64ClosedBothIntervalNode_5[] = "Float64ClosedBothIntervalNode.__pyx_fuse_3query";
static const char __pyx_k_Float64ClosedBothIntervalNode_6[] = "Float64ClosedBothIntervalNode.__pyx_fuse_4query";
static const char __pyx_k_Float64ClosedBothIntervalNode_p[] = "<Float64ClosedBothIntervalNode: pivot ";
static const char __pyx_k_Float64ClosedBothIntervalNode_s[] = "<Float64ClosedBothIntervalNode: %s elements (terminal)>";
static const char __pyx_k_Float64ClosedLeftIntervalNode_2[] = "Float64ClosedLeftIntervalNode.__pyx_fuse_0query";
static const char __pyx_k_Float64ClosedLeftIntervalNode_3[] = "Float64ClosedLeftIntervalNode.__pyx_fuse_1query";
static const char __pyx_k_Float64ClosedLeftIntervalNode_4[] = "Float64ClosedLeftIntervalNode.__pyx_fuse_2query";
static const char __pyx_k_Float64ClosedLeftIntervalNode_5[] = "Float64ClosedLeftIntervalNode.__pyx_fuse_3query";
static const char __pyx_k_Float64ClosedLeftIntervalNode_6[] = "Float64ClosedLeftIntervalNode.__pyx_fuse_4query";
static const char __pyx_k_Float64ClosedLeftIntervalNode_p[] = "<Float64ClosedLeftIntervalNode: pivot ";
static const char __pyx_k_Float64ClosedLeftIntervalNode_s[] = "<Float64ClosedLeftIntervalNode: %s elements (terminal)>";
static const char __pyx_k_Float64ClosedNeitherIntervalNod[] = "<Float64ClosedNeitherIntervalNode: %s elements (terminal)>";
static const char __pyx_k_Indicates_if_an_interval_is_emp[] = "\n        Indicates if an interval is empty, meaning it contains no points.\n\n        .. versionadded:: 0.25.0\n\n        Returns\n        -------\n        bool or ndarray\n            A boolean indicating if a scalar :class:`Interval` is empty, or a\n            boolean ``ndarray`` positionally indicating if an ``Interval`` in\n            an :class:`~arrays.IntervalArray` or :class:`IntervalIndex` is\n            empty.\n\n        Examples\n        --------\n        An :class:`Interval` that contains points is not empty:\n\n        >>> pd.Interval(0, 1, closed='right').is_empty\n        False\n\n        An ``Interval`` that does not contain any points is empty:\n\n        >>> pd.Interval(0, 0, closed='right').is_empty\n        True\n        >>> pd.Interval(0, 0, closed='left').is_empty\n        True\n        >>> pd.Interval(0, 0, closed='neither').is_empty\n        True\n\n        An ``Interval`` that contains a single point is not empty:\n\n        >>> pd.Interval(0, 0, closed='both').is_empty\n        False\n\n        An :class:`~arrays.IntervalArray` or :class:`IntervalIndex` returns a\n        boolean ``ndarray`` positionally indicating if an ``Interval`` is\n        empty:\n\n        >>> ivs = [pd.Interval(0, 0, closed='neither'),\n        ...        pd.Interval(1, 2, closed='neither')]\n        >>> pd.arrays.IntervalArray(ivs).is_empty\n        array([ True, False])\n\n        Missing values are not considered empty:\n\n        >>> ivs = [pd.Interval(0, 0, closed='neither'), np.nan]\n        >>> pd.IntervalIndex(ivs).is_empty\n        array([ True, False])\n        ";
static const char __pyx_k_Int32ClosedBothIntervalNode_piv[] = "<Int32ClosedBothIntervalNode: pivot ";
static const char __pyx_k_Int32ClosedBothIntervalNode_s_e[] = "<Int32ClosedBothIntervalNode: %s elements (terminal)>";
static const char __pyx_k_Int32ClosedLeftIntervalNode_piv[] = "<Int32ClosedLeftIntervalNode: pivot ";
static const char __pyx_k_Int32ClosedLeftIntervalNode_s_e[] = "<Int32ClosedLeftIntervalNode: %s elements (terminal)>";
static const char __pyx_k_Int32ClosedRightIntervalNode_pi[] = "<Int32ClosedRightIntervalNode: pivot ";
static const char __pyx_k_Int64ClosedBothIntervalNode_piv[] = "<Int64ClosedBothIntervalNode: pivot ";
static const char __pyx_k_Int64ClosedBothIntervalNode_s_e[] = "<Int64ClosedBothIntervalNode: %s elements (terminal)>";
static const char __pyx_k_Int64ClosedLeftIntervalNode_piv[] = "<Int64ClosedLeftIntervalNode: pivot ";
static const char __pyx_k_Int64ClosedLeftIntervalNode_s_e[] = "<Int64ClosedLeftIntervalNode: %s elements (terminal)>";
static const char __pyx_k_Int64ClosedRightIntervalNode_pi[] = "<Int64ClosedRightIntervalNode: pivot ";
static const char __pyx_k_IntervalTree_dtype_closed_n_ele[] = "<IntervalTree[{dtype},{closed}]: {n_elements} elements>";
static const char __pyx_k_Uint64ClosedBothIntervalNode_pi[] = "<Uint64ClosedBothIntervalNode: pivot ";
static const char __pyx_k_Uint64ClosedLeftIntervalNode_pi[] = "<Uint64ClosedLeftIntervalNode: pivot ";
static const char __pyx_k_Uint64ClosedNeitherIntervalNode[] = "<Uint64ClosedNeitherIntervalNode: %s elements (terminal)>";
static const char __pyx_k_Uint64ClosedRightIntervalNode_2[] = "Uint64ClosedRightIntervalNode.__pyx_fuse_0query";
static const char __pyx_k_Uint64ClosedRightIntervalNode_3[] = "Uint64ClosedRightIntervalNode.__pyx_fuse_1query";
static const char __pyx_k_Uint64ClosedRightIntervalNode_4[] = "Uint64ClosedRightIntervalNode.__pyx_fuse_2query";
static const char __pyx_k_Uint64ClosedRightIntervalNode_5[] = "Uint64ClosedRightIntervalNode.__pyx_fuse_3query";
static const char __pyx_k_Uint64ClosedRightIntervalNode_6[] = "Uint64ClosedRightIntervalNode.__pyx_fuse_4query";
static const char __pyx_k_Uint64ClosedRightIntervalNode_p[] = "<Uint64ClosedRightIntervalNode: pivot ";
static const char __pyx_k_Uint64ClosedRightIntervalNode_s[] = "<Uint64ClosedRightIntervalNode: %s elements (terminal)>";
static const char __pyx_k_intervals_must_all_be_closed_on[] = "intervals must all be closed on the same side";
static const char __pyx_k_name_left_r_right_r_closed_clos[] = "{name}({left!r}, {right!r}, closed={closed!r})";
static const char __pyx_k_numpy_core_multiarray_failed_to[] = "numpy.core.multiarray failed to import";
static const char __pyx_k_other_must_be_an_Interval_got_o[] = "`other` must be an Interval, got {other}";
static const char __pyx_k_unknown_dtype_code_in_numpy_pxd[] = "unknown dtype code in numpy.pxd (%d)";
static const char __pyx_k_unorderable_types_name_op_other[] = "unorderable types: {name}() {op} {other}()";
static const char __pyx_k_Buffer_view_does_not_expose_stri[] = "Buffer view does not expose strides";
static const char __pyx_k_Can_only_create_a_buffer_that_is[] = "Can only create a buffer that is contiguous in memory.";
static const char __pyx_k_Cannot_assign_to_read_only_memor[] = "Cannot assign to read-only memoryview";
static const char __pyx_k_Cannot_create_writable_memory_vi[] = "Cannot create writable memory view from read-only memoryview";
static const char __pyx_k_Empty_shape_tuple_for_cython_arr[] = "Empty shape tuple for cython.array";
static const char __pyx_k_Expected_at_least_d_argument_s_g[] = "Expected at least %d argument%s, got %d";
static const char __pyx_k_Float32ClosedBothIntervalNode_qu[] = "Float32ClosedBothIntervalNode.query";
static const char __pyx_k_Float32ClosedLeftIntervalNode_qu[] = "Float32ClosedLeftIntervalNode.query";
static const char __pyx_k_Float32ClosedNeitherIntervalNode[] = "Float32ClosedNeitherIntervalNode";
static const char __pyx_k_Float32ClosedRightIntervalNode_2[] = "<Float32ClosedRightIntervalNode: pivot ";
static const char __pyx_k_Float32ClosedRightIntervalNode_3[] = "Float32ClosedRightIntervalNode";
static const char __pyx_k_Float32ClosedRightIntervalNode_4[] = "Float32ClosedRightIntervalNode.__pyx_fuse_0query";
static const char __pyx_k_Float32ClosedRightIntervalNode_5[] = "Float32ClosedRightIntervalNode.__pyx_fuse_1query";
static const char __pyx_k_Float32ClosedRightIntervalNode_6[] = "Float32ClosedRightIntervalNode.__pyx_fuse_2query";
static const char __pyx_k_Float32ClosedRightIntervalNode_7[] = "Float32ClosedRightIntervalNode.__pyx_fuse_3query";
static const char __pyx_k_Float32ClosedRightIntervalNode_8[] = "Float32ClosedRightIntervalNode.__pyx_fuse_4query";
static const char __pyx_k_Float32ClosedRightIntervalNode_q[] = "Float32ClosedRightIntervalNode.query";
static const char __pyx_k_Float64ClosedBothIntervalNode_qu[] = "Float64ClosedBothIntervalNode.query";
static const char __pyx_k_Float64ClosedLeftIntervalNode_qu[] = "Float64ClosedLeftIntervalNode.query";
static const char __pyx_k_Float64ClosedNeitherIntervalNode[] = "Float64ClosedNeitherIntervalNode";
static const char __pyx_k_Float64ClosedRightIntervalNode_2[] = "<Float64ClosedRightIntervalNode: pivot ";
static const char __pyx_k_Float64ClosedRightIntervalNode_3[] = "Float64ClosedRightIntervalNode";
static const char __pyx_k_Float64ClosedRightIntervalNode_4[] = "Float64ClosedRightIntervalNode.__pyx_fuse_0query";
static const char __pyx_k_Float64ClosedRightIntervalNode_5[] = "Float64ClosedRightIntervalNode.__pyx_fuse_1query";
static const char __pyx_k_Float64ClosedRightIntervalNode_6[] = "Float64ClosedRightIntervalNode.__pyx_fuse_2query";
static const char __pyx_k_Float64ClosedRightIntervalNode_7[] = "Float64ClosedRightIntervalNode.__pyx_fuse_3query";
static const char __pyx_k_Float64ClosedRightIntervalNode_8[] = "Float64ClosedRightIntervalNode.__pyx_fuse_4query";
static const char __pyx_k_Float64ClosedRightIntervalNode_q[] = "Float64ClosedRightIntervalNode.query";
static const char __pyx_k_Format_string_allocated_too_shor[] = "Format string allocated too short, see comment in numpy.pxd";
static const char __pyx_k_Function_call_with_ambiguous_arg[] = "Function call with ambiguous argument types";
static const char __pyx_k_Incompatible_checksums_s_vs_0x05[] = "Incompatible checksums (%s vs 0x05502fb = (_is_overlapping, _left_sorter, _right_sorter, closed, dtype, left, right, root))";
static const char __pyx_k_Incompatible_checksums_s_vs_0x7a[] = "Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))";
static const char __pyx_k_Incompatible_checksums_s_vs_0xb0[] = "Incompatible checksums (%s vs 0xb068931 = (name))";
static const char __pyx_k_Incompatible_checksums_s_vs_0xd4[] = "Incompatible checksums (%s vs 0xd41d8cd = ())";
static const char __pyx_k_Indirect_dimensions_not_supporte[] = "Indirect dimensions not supported";
static const char __pyx_k_Int32ClosedBothIntervalNode___py[] = "Int32ClosedBothIntervalNode.__pyx_fuse_0query";
static const char __pyx_k_Int32ClosedBothIntervalNode_quer[] = "Int32ClosedBothIntervalNode.query";
static const char __pyx_k_Int32ClosedLeftIntervalNode___py[] = "Int32ClosedLeftIntervalNode.__pyx_fuse_0query";
static const char __pyx_k_Int32ClosedLeftIntervalNode_quer[] = "Int32ClosedLeftIntervalNode.query";
static const char __pyx_k_Int32ClosedNeitherIntervalNode_2[] = "<Int32ClosedNeitherIntervalNode: pivot ";
static const char __pyx_k_Int32ClosedNeitherIntervalNode_3[] = "Int32ClosedNeitherIntervalNode";
static const char __pyx_k_Int32ClosedNeitherIntervalNode_4[] = "Int32ClosedNeitherIntervalNode.__pyx_fuse_0query";
static const char __pyx_k_Int32ClosedNeitherIntervalNode_5[] = "Int32ClosedNeitherIntervalNode.__pyx_fuse_1query";
static const char __pyx_k_Int32ClosedNeitherIntervalNode_6[] = "Int32ClosedNeitherIntervalNode.__pyx_fuse_2query";
static const char __pyx_k_Int32ClosedNeitherIntervalNode_7[] = "Int32ClosedNeitherIntervalNode.__pyx_fuse_3query";
static const char __pyx_k_Int32ClosedNeitherIntervalNode_8[] = "Int32ClosedNeitherIntervalNode.__pyx_fuse_4query";
static const char __pyx_k_Int32ClosedNeitherIntervalNode_q[] = "Int32ClosedNeitherIntervalNode.query";
static const char __pyx_k_Int32ClosedRightIntervalNode___p[] = "Int32ClosedRightIntervalNode.__pyx_fuse_0query";
static const char __pyx_k_Int32ClosedRightIntervalNode_que[] = "Int32ClosedRightIntervalNode.query";
static const char __pyx_k_Int64ClosedBothIntervalNode___py[] = "Int64ClosedBothIntervalNode.__pyx_fuse_0query";
static const char __pyx_k_Int64ClosedBothIntervalNode_quer[] = "Int64ClosedBothIntervalNode.query";
static const char __pyx_k_Int64ClosedLeftIntervalNode___py[] = "Int64ClosedLeftIntervalNode.__pyx_fuse_0query";
static const char __pyx_k_Int64ClosedLeftIntervalNode_quer[] = "Int64ClosedLeftIntervalNode.query";
static const char __pyx_k_Int64ClosedNeitherIntervalNode_2[] = "<Int64ClosedNeitherIntervalNode: pivot ";
static const char __pyx_k_Int64ClosedNeitherIntervalNode_3[] = "Int64ClosedNeitherIntervalNode";
static const char __pyx_k_Int64ClosedNeitherIntervalNode_4[] = "Int64ClosedNeitherIntervalNode.__pyx_fuse_0query";
static const char __pyx_k_Int64ClosedNeitherIntervalNode_5[] = "Int64ClosedNeitherIntervalNode.__pyx_fuse_1query";
static const char __pyx_k_Int64ClosedNeitherIntervalNode_6[] = "Int64ClosedNeitherIntervalNode.__pyx_fuse_2query";
static const char __pyx_k_Int64ClosedNeitherIntervalNode_7[] = "Int64ClosedNeitherIntervalNode.__pyx_fuse_3query";
static const char __pyx_k_Int64ClosedNeitherIntervalNode_8[] = "Int64ClosedNeitherIntervalNode.__pyx_fuse_4query";
static const char __pyx_k_Int64ClosedNeitherIntervalNode_q[] = "Int64ClosedNeitherIntervalNode.query";
static const char __pyx_k_Int64ClosedRightIntervalNode___p[] = "Int64ClosedRightIntervalNode.__pyx_fuse_0query";
static const char __pyx_k_Int64ClosedRightIntervalNode_que[] = "Int64ClosedRightIntervalNode.query";
static const char __pyx_k_IntervalMixin_is_empty___get___l[] = "IntervalMixin.is_empty.__get__ (line 111)";
static const char __pyx_k_IntervalTree_get_indexer_non_uni[] = "IntervalTree.get_indexer_non_unique";
static const char __pyx_k_Invalid_mode_expected_c_or_fortr[] = "Invalid mode, expected 'c' or 'fortran', got %s";
static const char __pyx_k_Non_native_byte_order_not_suppor[] = "Non-native byte order not supported";
static const char __pyx_k_Only_numeric_Timestamp_and_Timed[] = "Only numeric, Timestamp and Timedelta endpoints are allowed when constructing an Interval.";
static const char __pyx_k_Out_of_bounds_on_buffer_access_a[] = "Out of bounds on buffer access (axis %d)";
static const char __pyx_k_Uint64ClosedBothIntervalNode___p[] = "Uint64ClosedBothIntervalNode.__pyx_fuse_0query";
static const char __pyx_k_Uint64ClosedBothIntervalNode_que[] = "Uint64ClosedBothIntervalNode.query";
static const char __pyx_k_Uint64ClosedLeftIntervalNode___p[] = "Uint64ClosedLeftIntervalNode.__pyx_fuse_0query";
static const char __pyx_k_Uint64ClosedLeftIntervalNode_que[] = "Uint64ClosedLeftIntervalNode.query";
static const char __pyx_k_Uint64ClosedRightIntervalNode_qu[] = "Uint64ClosedRightIntervalNode.query";
static const char __pyx_k_Unable_to_convert_item_to_object[] = "Unable to convert item to object";
static const char __pyx_k_got_differing_extents_in_dimensi[] = "got differing extents in dimension %d (got %d and %d)";
static const char __pyx_k_indexer_does_not_intersect_a_uni[] = "indexer does not intersect a unique set of intervals";
static const char __pyx_k_invalid_option_for_closed_closed[] = "invalid option for 'closed': {closed}";
static const char __pyx_k_left_and_right_must_have_the_sam[] = "left and right must have the same time zone, got '{left_tz}' and '{right_tz}'";
static const char __pyx_k_left_side_of_interval_must_be_ri[] = "left side of interval must be <= right side";
static const char __pyx_k_ndarray_is_not_Fortran_contiguou[] = "ndarray is not Fortran contiguous";
static const char __pyx_k_no_default___reduce___due_to_non[] = "no default __reduce__ due to non-trivial __cinit__";
static const char __pyx_k_numpy_core_umath_failed_to_impor[] = "numpy.core.umath failed to import";
static const char __pyx_k_type_typ_with_value_iv_is_not_an[] = "type {typ} with value {iv} is not an interval";
static const char __pyx_k_unable_to_allocate_shape_and_str[] = "unable to allocate shape and strides.";
static const char __pyx_k_Float32ClosedNeitherIntervalNod_2[] = "<Float32ClosedNeitherIntervalNode: pivot ";
static const char __pyx_k_Float64ClosedNeitherIntervalNod_2[] = "<Float64ClosedNeitherIntervalNode: pivot ";
static const char __pyx_k_Uint64ClosedNeitherIntervalNode_2[] = "<Uint64ClosedNeitherIntervalNode: pivot ";
static const char __pyx_k_Uint64ClosedNeitherIntervalNode_3[] = "Uint64ClosedNeitherIntervalNode";
static const char __pyx_k_Uint64ClosedNeitherIntervalNode_4[] = "Uint64ClosedNeitherIntervalNode.__pyx_fuse_0query";
static const char __pyx_k_Uint64ClosedNeitherIntervalNode_5[] = "Uint64ClosedNeitherIntervalNode.__pyx_fuse_1query";
static const char __pyx_k_Uint64ClosedNeitherIntervalNode_6[] = "Uint64ClosedNeitherIntervalNode.__pyx_fuse_2query";
static const char __pyx_k_Uint64ClosedNeitherIntervalNode_7[] = "Uint64ClosedNeitherIntervalNode.__pyx_fuse_3query";
static const char __pyx_k_Uint64ClosedNeitherIntervalNode_8[] = "Uint64ClosedNeitherIntervalNode.__pyx_fuse_4query";
static const char __pyx_k_Uint64ClosedNeitherIntervalNode_9[] = "Uint64ClosedNeitherIntervalNode.query";
static const char __pyx_k_Float32ClosedNeitherIntervalNode_2[] = "Float32ClosedNeitherIntervalNode.__pyx_fuse_0query";
static const char __pyx_k_Float32ClosedNeitherIntervalNode_3[] = "Float32ClosedNeitherIntervalNode.__pyx_fuse_1query";
static const char __pyx_k_Float32ClosedNeitherIntervalNode_4[] = "Float32ClosedNeitherIntervalNode.__pyx_fuse_2query";
static const char __pyx_k_Float32ClosedNeitherIntervalNode_5[] = "Float32ClosedNeitherIntervalNode.__pyx_fuse_3query";
static const char __pyx_k_Float32ClosedNeitherIntervalNode_6[] = "Float32ClosedNeitherIntervalNode.__pyx_fuse_4query";
static const char __pyx_k_Float32ClosedNeitherIntervalNode_7[] = "Float32ClosedNeitherIntervalNode.query";
static const char __pyx_k_Float64ClosedNeitherIntervalNode_2[] = "Float64ClosedNeitherIntervalNode.__pyx_fuse_0query";
static const char __pyx_k_Float64ClosedNeitherIntervalNode_3[] = "Float64ClosedNeitherIntervalNode.__pyx_fuse_1query";
static const char __pyx_k_Float64ClosedNeitherIntervalNode_4[] = "Float64ClosedNeitherIntervalNode.__pyx_fuse_2query";
static const char __pyx_k_Float64ClosedNeitherIntervalNode_5[] = "Float64ClosedNeitherIntervalNode.__pyx_fuse_3query";
static const char __pyx_k_Float64ClosedNeitherIntervalNode_6[] = "Float64ClosedNeitherIntervalNode.__pyx_fuse_4query";
static const char __pyx_k_Float64ClosedNeitherIntervalNode_7[] = "Float64ClosedNeitherIntervalNode.query";
static const char __pyx_k_Format_string_allocated_too_shor_2[] = "Format string allocated too short.";
static const char __pyx_k_Int32ClosedBothIntervalNode___py_2[] = "Int32ClosedBothIntervalNode.__pyx_fuse_1query";
static const char __pyx_k_Int32ClosedBothIntervalNode___py_3[] = "Int32ClosedBothIntervalNode.__pyx_fuse_2query";
static const char __pyx_k_Int32ClosedBothIntervalNode___py_4[] = "Int32ClosedBothIntervalNode.__pyx_fuse_3query";
static const char __pyx_k_Int32ClosedBothIntervalNode___py_5[] = "Int32ClosedBothIntervalNode.__pyx_fuse_4query";
static const char __pyx_k_Int32ClosedLeftIntervalNode___py_2[] = "Int32ClosedLeftIntervalNode.__pyx_fuse_1query";
static const char __pyx_k_Int32ClosedLeftIntervalNode___py_3[] = "Int32ClosedLeftIntervalNode.__pyx_fuse_2query";
static const char __pyx_k_Int32ClosedLeftIntervalNode___py_4[] = "Int32ClosedLeftIntervalNode.__pyx_fuse_3query";
static const char __pyx_k_Int32ClosedLeftIntervalNode___py_5[] = "Int32ClosedLeftIntervalNode.__pyx_fuse_4query";
static const char __pyx_k_Int32ClosedRightIntervalNode___p_2[] = "Int32ClosedRightIntervalNode.__pyx_fuse_1query";
static const char __pyx_k_Int32ClosedRightIntervalNode___p_3[] = "Int32ClosedRightIntervalNode.__pyx_fuse_2query";
static const char __pyx_k_Int32ClosedRightIntervalNode___p_4[] = "Int32ClosedRightIntervalNode.__pyx_fuse_3query";
static const char __pyx_k_Int32ClosedRightIntervalNode___p_5[] = "Int32ClosedRightIntervalNode.__pyx_fuse_4query";
static const char __pyx_k_Int64ClosedBothIntervalNode___py_2[] = "Int64ClosedBothIntervalNode.__pyx_fuse_1query";
static const char __pyx_k_Int64ClosedBothIntervalNode___py_3[] = "Int64ClosedBothIntervalNode.__pyx_fuse_2query";
static const char __pyx_k_Int64ClosedBothIntervalNode___py_4[] = "Int64ClosedBothIntervalNode.__pyx_fuse_3query";
static const char __pyx_k_Int64ClosedBothIntervalNode___py_5[] = "Int64ClosedBothIntervalNode.__pyx_fuse_4query";
static const char __pyx_k_Int64ClosedLeftIntervalNode___py_2[] = "Int64ClosedLeftIntervalNode.__pyx_fuse_1query";
static const char __pyx_k_Int64ClosedLeftIntervalNode___py_3[] = "Int64ClosedLeftIntervalNode.__pyx_fuse_2query";
static const char __pyx_k_Int64ClosedLeftIntervalNode___py_4[] = "Int64ClosedLeftIntervalNode.__pyx_fuse_3query";
static const char __pyx_k_Int64ClosedLeftIntervalNode___py_5[] = "Int64ClosedLeftIntervalNode.__pyx_fuse_4query";
static const char __pyx_k_Int64ClosedRightIntervalNode___p_2[] = "Int64ClosedRightIntervalNode.__pyx_fuse_1query";
static const char __pyx_k_Int64ClosedRightIntervalNode___p_3[] = "Int64ClosedRightIntervalNode.__pyx_fuse_2query";
static const char __pyx_k_Int64ClosedRightIntervalNode___p_4[] = "Int64ClosedRightIntervalNode.__pyx_fuse_3query";
static const char __pyx_k_Int64ClosedRightIntervalNode___p_5[] = "Int64ClosedRightIntervalNode.__pyx_fuse_4query";
static const char __pyx_k_Uint64ClosedBothIntervalNode___p_2[] = "Uint64ClosedBothIntervalNode.__pyx_fuse_1query";
static const char __pyx_k_Uint64ClosedBothIntervalNode___p_3[] = "Uint64ClosedBothIntervalNode.__pyx_fuse_2query";
static const char __pyx_k_Uint64ClosedBothIntervalNode___p_4[] = "Uint64ClosedBothIntervalNode.__pyx_fuse_3query";
static const char __pyx_k_Uint64ClosedBothIntervalNode___p_5[] = "Uint64ClosedBothIntervalNode.__pyx_fuse_4query";
static const char __pyx_k_Uint64ClosedLeftIntervalNode___p_2[] = "Uint64ClosedLeftIntervalNode.__pyx_fuse_1query";
static const char __pyx_k_Uint64ClosedLeftIntervalNode___p_3[] = "Uint64ClosedLeftIntervalNode.__pyx_fuse_2query";
static const char __pyx_k_Uint64ClosedLeftIntervalNode___p_4[] = "Uint64ClosedLeftIntervalNode.__pyx_fuse_3query";
static const char __pyx_k_Uint64ClosedLeftIntervalNode___p_5[] = "Uint64ClosedLeftIntervalNode.__pyx_fuse_4query";
static PyObject *__pyx_n_s_ASCII;
static PyObject *__pyx_kp_s_Buffer_view_does_not_expose_stri;
static PyObject *__pyx_kp_s_Can_only_create_a_buffer_that_is;
static PyObject *__pyx_kp_s_Cannot_assign_to_read_only_memor;
static PyObject *__pyx_kp_s_Cannot_create_writable_memory_vi;
static PyObject *__pyx_kp_s_Cannot_index_with_type_s;
static PyObject *__pyx_kp_u_Check_whether_two_Interval_obje;
static PyObject *__pyx_n_s_Ellipsis;
static PyObject *__pyx_kp_s_Empty_shape_tuple_for_cython_arr;
static PyObject *__pyx_kp_s_Expected_at_least_d_argument_s_g;
static PyObject *__pyx_n_s_Float32ClosedBothIntervalNode;
static PyObject *__pyx_n_s_Float32ClosedBothIntervalNode_2;
static PyObject *__pyx_n_s_Float32ClosedBothIntervalNode_3;
static PyObject *__pyx_n_s_Float32ClosedBothIntervalNode_4;
static PyObject *__pyx_n_s_Float32ClosedBothIntervalNode_5;
static PyObject *__pyx_n_s_Float32ClosedBothIntervalNode_6;
static PyObject *__pyx_kp_u_Float32ClosedBothIntervalNode_p;
static PyObject *__pyx_n_s_Float32ClosedBothIntervalNode_qu;
static PyObject *__pyx_kp_u_Float32ClosedBothIntervalNode_s;
static PyObject *__pyx_n_s_Float32ClosedLeftIntervalNode;
static PyObject *__pyx_n_s_Float32ClosedLeftIntervalNode_2;
static PyObject *__pyx_n_s_Float32ClosedLeftIntervalNode_3;
static PyObject *__pyx_n_s_Float32ClosedLeftIntervalNode_4;
static PyObject *__pyx_n_s_Float32ClosedLeftIntervalNode_5;
static PyObject *__pyx_n_s_Float32ClosedLeftIntervalNode_6;
static PyObject *__pyx_kp_u_Float32ClosedLeftIntervalNode_p;
static PyObject *__pyx_n_s_Float32ClosedLeftIntervalNode_qu;
static PyObject *__pyx_kp_u_Float32ClosedLeftIntervalNode_s;
static PyObject *__pyx_kp_u_Float32ClosedNeitherIntervalNod;
static PyObject *__pyx_kp_u_Float32ClosedNeitherIntervalNod_2;
static PyObject *__pyx_n_s_Float32ClosedNeitherIntervalNode;
static PyObject *__pyx_n_s_Float32ClosedNeitherIntervalNode_2;
static PyObject *__pyx_n_s_Float32ClosedNeitherIntervalNode_3;
static PyObject *__pyx_n_s_Float32ClosedNeitherIntervalNode_4;
static PyObject *__pyx_n_s_Float32ClosedNeitherIntervalNode_5;
static PyObject *__pyx_n_s_Float32ClosedNeitherIntervalNode_6;
static PyObject *__pyx_n_s_Float32ClosedNeitherIntervalNode_7;
static PyObject *__pyx_kp_u_Float32ClosedRightIntervalNode;
static PyObject *__pyx_kp_u_Float32ClosedRightIntervalNode_2;
static PyObject *__pyx_n_s_Float32ClosedRightIntervalNode_3;
static PyObject *__pyx_n_s_Float32ClosedRightIntervalNode_4;
static PyObject *__pyx_n_s_Float32ClosedRightIntervalNode_5;
static PyObject *__pyx_n_s_Float32ClosedRightIntervalNode_6;
static PyObject *__pyx_n_s_Float32ClosedRightIntervalNode_7;
static PyObject *__pyx_n_s_Float32ClosedRightIntervalNode_8;
static PyObject *__pyx_n_s_Float32ClosedRightIntervalNode_q;
static PyObject *__pyx_n_s_Float64ClosedBothIntervalNode;
static PyObject *__pyx_n_s_Float64ClosedBothIntervalNode_2;
static PyObject *__pyx_n_s_Float64ClosedBothIntervalNode_3;
static PyObject *__pyx_n_s_Float64ClosedBothIntervalNode_4;
static PyObject *__pyx_n_s_Float64ClosedBothIntervalNode_5;
static PyObject *__pyx_n_s_Float64ClosedBothIntervalNode_6;
static PyObject *__pyx_kp_u_Float64ClosedBothIntervalNode_p;
static PyObject *__pyx_n_s_Float64ClosedBothIntervalNode_qu;
static PyObject *__pyx_kp_u_Float64ClosedBothIntervalNode_s;
static PyObject *__pyx_n_s_Float64ClosedLeftIntervalNode;
static PyObject *__pyx_n_s_Float64ClosedLeftIntervalNode_2;
static PyObject *__pyx_n_s_Float64ClosedLeftIntervalNode_3;
static PyObject *__pyx_n_s_Float64ClosedLeftIntervalNode_4;
static PyObject *__pyx_n_s_Float64ClosedLeftIntervalNode_5;
static PyObject *__pyx_n_s_Float64ClosedLeftIntervalNode_6;
static PyObject *__pyx_kp_u_Float64ClosedLeftIntervalNode_p;
static PyObject *__pyx_n_s_Float64ClosedLeftIntervalNode_qu;
static PyObject *__pyx_kp_u_Float64ClosedLeftIntervalNode_s;
static PyObject *__pyx_kp_u_Float64ClosedNeitherIntervalNod;
static PyObject *__pyx_kp_u_Float64ClosedNeitherIntervalNod_2;
static PyObject *__pyx_n_s_Float64ClosedNeitherIntervalNode;
static PyObject *__pyx_n_s_Float64ClosedNeitherIntervalNode_2;
static PyObject *__pyx_n_s_Float64ClosedNeitherIntervalNode_3;
static PyObject *__pyx_n_s_Float64ClosedNeitherIntervalNode_4;
static PyObject *__pyx_n_s_Float64ClosedNeitherIntervalNode_5;
static PyObject *__pyx_n_s_Float64ClosedNeitherIntervalNode_6;
static PyObject *__pyx_n_s_Float64ClosedNeitherIntervalNode_7;
static PyObject *__pyx_kp_u_Float64ClosedRightIntervalNode;
static PyObject *__pyx_kp_u_Float64ClosedRightIntervalNode_2;
static PyObject *__pyx_n_s_Float64ClosedRightIntervalNode_3;
static PyObject *__pyx_n_s_Float64ClosedRightIntervalNode_4;
static PyObject *__pyx_n_s_Float64ClosedRightIntervalNode_5;
static PyObject *__pyx_n_s_Float64ClosedRightIntervalNode_6;
static PyObject *__pyx_n_s_Float64ClosedRightIntervalNode_7;
static PyObject *__pyx_n_s_Float64ClosedRightIntervalNode_8;
static PyObject *__pyx_n_s_Float64ClosedRightIntervalNode_q;
static PyObject *__pyx_kp_u_Format_string_allocated_too_shor;
static PyObject *__pyx_kp_u_Format_string_allocated_too_shor_2;
static PyObject *__pyx_kp_s_Function_call_with_ambiguous_arg;
static PyObject *__pyx_n_s_ImportError;
static PyObject *__pyx_kp_s_Incompatible_checksums_s_vs_0x05;
static PyObject *__pyx_kp_s_Incompatible_checksums_s_vs_0x7a;
static PyObject *__pyx_kp_s_Incompatible_checksums_s_vs_0xb0;
static PyObject *__pyx_kp_s_Incompatible_checksums_s_vs_0xd4;
static PyObject *__pyx_n_s_IndexError;
static PyObject *__pyx_kp_u_Indicates_if_an_interval_is_emp;
static PyObject *__pyx_kp_s_Indirect_dimensions_not_supporte;
static PyObject *__pyx_n_s_Int32ClosedBothIntervalNode;
static PyObject *__pyx_n_s_Int32ClosedBothIntervalNode___py;
static PyObject *__pyx_n_s_Int32ClosedBothIntervalNode___py_2;
static PyObject *__pyx_n_s_Int32ClosedBothIntervalNode___py_3;
static PyObject *__pyx_n_s_Int32ClosedBothIntervalNode___py_4;
static PyObject *__pyx_n_s_Int32ClosedBothIntervalNode___py_5;
static PyObject *__pyx_kp_u_Int32ClosedBothIntervalNode_piv;
static PyObject *__pyx_n_s_Int32ClosedBothIntervalNode_quer;
static PyObject *__pyx_kp_u_Int32ClosedBothIntervalNode_s_e;
static PyObject *__pyx_n_s_Int32ClosedLeftIntervalNode;
static PyObject *__pyx_n_s_Int32ClosedLeftIntervalNode___py;
static PyObject *__pyx_n_s_Int32ClosedLeftIntervalNode___py_2;
static PyObject *__pyx_n_s_Int32ClosedLeftIntervalNode___py_3;
static PyObject *__pyx_n_s_Int32ClosedLeftIntervalNode___py_4;
static PyObject *__pyx_n_s_Int32ClosedLeftIntervalNode___py_5;
static PyObject *__pyx_kp_u_Int32ClosedLeftIntervalNode_piv;
static PyObject *__pyx_n_s_Int32ClosedLeftIntervalNode_quer;
static PyObject *__pyx_kp_u_Int32ClosedLeftIntervalNode_s_e;
static PyObject *__pyx_kp_u_Int32ClosedNeitherIntervalNode;
static PyObject *__pyx_kp_u_Int32ClosedNeitherIntervalNode_2;
static PyObject *__pyx_n_s_Int32ClosedNeitherIntervalNode_3;
static PyObject *__pyx_n_s_Int32ClosedNeitherIntervalNode_4;
static PyObject *__pyx_n_s_Int32ClosedNeitherIntervalNode_5;
static PyObject *__pyx_n_s_Int32ClosedNeitherIntervalNode_6;
static PyObject *__pyx_n_s_Int32ClosedNeitherIntervalNode_7;
static PyObject *__pyx_n_s_Int32ClosedNeitherIntervalNode_8;
static PyObject *__pyx_n_s_Int32ClosedNeitherIntervalNode_q;
static PyObject *__pyx_n_s_Int32ClosedRightIntervalNode;
static PyObject *__pyx_n_s_Int32ClosedRightIntervalNode___p;
static PyObject *__pyx_n_s_Int32ClosedRightIntervalNode___p_2;
static PyObject *__pyx_n_s_Int32ClosedRightIntervalNode___p_3;
static PyObject *__pyx_n_s_Int32ClosedRightIntervalNode___p_4;
static PyObject *__pyx_n_s_Int32ClosedRightIntervalNode___p_5;
static PyObject *__pyx_kp_u_Int32ClosedRightIntervalNode_pi;
static PyObject *__pyx_n_s_Int32ClosedRightIntervalNode_que;
static PyObject *__pyx_kp_u_Int32ClosedRightIntervalNode_s;
static PyObject *__pyx_n_s_Int64ClosedBothIntervalNode;
static PyObject *__pyx_n_s_Int64ClosedBothIntervalNode___py;
static PyObject *__pyx_n_s_Int64ClosedBothIntervalNode___py_2;
static PyObject *__pyx_n_s_Int64ClosedBothIntervalNode___py_3;
static PyObject *__pyx_n_s_Int64ClosedBothIntervalNode___py_4;
static PyObject *__pyx_n_s_Int64ClosedBothIntervalNode___py_5;
static PyObject *__pyx_kp_u_Int64ClosedBothIntervalNode_piv;
static PyObject *__pyx_n_s_Int64ClosedBothIntervalNode_quer;
static PyObject *__pyx_kp_u_Int64ClosedBothIntervalNode_s_e;
static PyObject *__pyx_n_s_Int64ClosedLeftIntervalNode;
static PyObject *__pyx_n_s_Int64ClosedLeftIntervalNode___py;
static PyObject *__pyx_n_s_Int64ClosedLeftIntervalNode___py_2;
static PyObject *__pyx_n_s_Int64ClosedLeftIntervalNode___py_3;
static PyObject *__pyx_n_s_Int64ClosedLeftIntervalNode___py_4;
static PyObject *__pyx_n_s_Int64ClosedLeftIntervalNode___py_5;
static PyObject *__pyx_kp_u_Int64ClosedLeftIntervalNode_piv;
static PyObject *__pyx_n_s_Int64ClosedLeftIntervalNode_quer;
static PyObject *__pyx_kp_u_Int64ClosedLeftIntervalNode_s_e;
static PyObject *__pyx_kp_u_Int64ClosedNeitherIntervalNode;
static PyObject *__pyx_kp_u_Int64ClosedNeitherIntervalNode_2;
static PyObject *__pyx_n_s_Int64ClosedNeitherIntervalNode_3;
static PyObject *__pyx_n_s_Int64ClosedNeitherIntervalNode_4;
static PyObject *__pyx_n_s_Int64ClosedNeitherIntervalNode_5;
static PyObject *__pyx_n_s_Int64ClosedNeitherIntervalNode_6;
static PyObject *__pyx_n_s_Int64ClosedNeitherIntervalNode_7;
static PyObject *__pyx_n_s_Int64ClosedNeitherIntervalNode_8;
static PyObject *__pyx_n_s_Int64ClosedNeitherIntervalNode_q;
static PyObject *__pyx_n_s_Int64ClosedRightIntervalNode;
static PyObject *__pyx_n_s_Int64ClosedRightIntervalNode___p;
static PyObject *__pyx_n_s_Int64ClosedRightIntervalNode___p_2;
static PyObject *__pyx_n_s_Int64ClosedRightIntervalNode___p_3;
static PyObject *__pyx_n_s_Int64ClosedRightIntervalNode___p_4;
static PyObject *__pyx_n_s_Int64ClosedRightIntervalNode___p_5;
static PyObject *__pyx_kp_u_Int64ClosedRightIntervalNode_pi;
static PyObject *__pyx_n_s_Int64ClosedRightIntervalNode_que;
static PyObject *__pyx_kp_u_Int64ClosedRightIntervalNode_s;
static PyObject *__pyx_n_s_Interval;
static PyObject *__pyx_n_s_IntervalMixin;
static PyObject *__pyx_kp_u_IntervalMixin_is_empty___get___l;
static PyObject *__pyx_n_s_IntervalTree;
static PyObject *__pyx_kp_u_IntervalTree_dtype_closed_n_ele;
static PyObject *__pyx_n_s_IntervalTree_get_indexer;
static PyObject *__pyx_n_s_IntervalTree_get_indexer_non_uni;
static PyObject *__pyx_n_s_IntervalTree_get_loc;
static PyObject *__pyx_kp_u_Interval_overlaps_line_427;
static PyObject *__pyx_kp_s_Invalid_mode_expected_c_or_fortr;
static PyObject *__pyx_kp_s_Invalid_shape_in_axis_d_d;
static PyObject *__pyx_n_s_KeyError;
static PyObject *__pyx_n_s_MemoryError;
static PyObject *__pyx_kp_s_MemoryView_of_r_at_0x_x;
static PyObject *__pyx_kp_s_MemoryView_of_r_object;
static PyObject *__pyx_n_s_NODE_CLASSES;
static PyObject *__pyx_kp_s_No_matching_signature_found;
static PyObject *__pyx_kp_u_Non_native_byte_order_not_suppor;
static PyObject *__pyx_n_s_NotImplemented;
static PyObject *__pyx_n_s_Number;
static PyObject *__pyx_n_b_O;
static PyObject *__pyx_kp_u_Only_numeric_Timestamp_and_Timed;
static PyObject *__pyx_kp_s_Out_of_bounds_on_buffer_access_a;
static PyObject *__pyx_n_s_PickleError;
static PyObject *__pyx_n_s_RuntimeError;
static PyObject *__pyx_n_s_Timedelta;
static PyObject *__pyx_n_s_Timestamp;
static PyObject *__pyx_n_s_TypeError;
static PyObject *__pyx_n_s_Uint64ClosedBothIntervalNode;
static PyObject *__pyx_n_s_Uint64ClosedBothIntervalNode___p;
static PyObject *__pyx_n_s_Uint64ClosedBothIntervalNode___p_2;
static PyObject *__pyx_n_s_Uint64ClosedBothIntervalNode___p_3;
static PyObject *__pyx_n_s_Uint64ClosedBothIntervalNode___p_4;
static PyObject *__pyx_n_s_Uint64ClosedBothIntervalNode___p_5;
static PyObject *__pyx_kp_u_Uint64ClosedBothIntervalNode_pi;
static PyObject *__pyx_n_s_Uint64ClosedBothIntervalNode_que;
static PyObject *__pyx_kp_u_Uint64ClosedBothIntervalNode_s;
static PyObject *__pyx_n_s_Uint64ClosedLeftIntervalNode;
static PyObject *__pyx_n_s_Uint64ClosedLeftIntervalNode___p;
static PyObject *__pyx_n_s_Uint64ClosedLeftIntervalNode___p_2;
static PyObject *__pyx_n_s_Uint64ClosedLeftIntervalNode___p_3;
static PyObject *__pyx_n_s_Uint64ClosedLeftIntervalNode___p_4;
static PyObject *__pyx_n_s_Uint64ClosedLeftIntervalNode___p_5;
static PyObject *__pyx_kp_u_Uint64ClosedLeftIntervalNode_pi;
static PyObject *__pyx_n_s_Uint64ClosedLeftIntervalNode_que;
static PyObject *__pyx_kp_u_Uint64ClosedLeftIntervalNode_s;
static PyObject *__pyx_kp_u_Uint64ClosedNeitherIntervalNode;
static PyObject *__pyx_kp_u_Uint64ClosedNeitherIntervalNode_2;
static PyObject *__pyx_n_s_Uint64ClosedNeitherIntervalNode_3;
static PyObject *__pyx_n_s_Uint64ClosedNeitherIntervalNode_4;
static PyObject *__pyx_n_s_Uint64ClosedNeitherIntervalNode_5;
static PyObject *__pyx_n_s_Uint64ClosedNeitherIntervalNode_6;
static PyObject *__pyx_n_s_Uint64ClosedNeitherIntervalNode_7;
static PyObject *__pyx_n_s_Uint64ClosedNeitherIntervalNode_8;
static PyObject *__pyx_n_s_Uint64ClosedNeitherIntervalNode_9;
static PyObject *__pyx_n_s_Uint64ClosedRightIntervalNode;
static PyObject *__pyx_n_s_Uint64ClosedRightIntervalNode_2;
static PyObject *__pyx_n_s_Uint64ClosedRightIntervalNode_3;
static PyObject *__pyx_n_s_Uint64ClosedRightIntervalNode_4;
static PyObject *__pyx_n_s_Uint64ClosedRightIntervalNode_5;
static PyObject *__pyx_n_s_Uint64ClosedRightIntervalNode_6;
static PyObject *__pyx_kp_u_Uint64ClosedRightIntervalNode_p;
static PyObject *__pyx_n_s_Uint64ClosedRightIntervalNode_qu;
static PyObject *__pyx_kp_u_Uint64ClosedRightIntervalNode_s;
static PyObject *__pyx_kp_s_Unable_to_convert_item_to_object;
static PyObject *__pyx_n_s_VALID_CLOSED;
static PyObject *__pyx_n_s_ValueError;
static PyObject *__pyx_n_s_View_MemoryView;
static PyObject *__pyx_kp_u__10;
static PyObject *__pyx_kp_s__13;
static PyObject *__pyx_kp_s__14;
static PyObject *__pyx_kp_u__18;
static PyObject *__pyx_kp_u__3;
static PyObject *__pyx_kp_u__4;
static PyObject *__pyx_kp_u__5;
static PyObject *__pyx_kp_u__6;
static PyObject *__pyx_kp_u__7;
static PyObject *__pyx_kp_u__8;
static PyObject *__pyx_kp_u__9;
static PyObject *__pyx_n_s_allocate_buffer;
static PyObject *__pyx_n_s_any;
static PyObject *__pyx_n_s_arange;
static PyObject *__pyx_n_s_args;
static PyObject *__pyx_n_s_argsort;
static PyObject *__pyx_n_s_asarray;
static PyObject *__pyx_n_s_astype;
static PyObject *__pyx_n_s_base;
static PyObject *__pyx_n_u_both;
static PyObject *__pyx_n_s_c;
static PyObject *__pyx_n_u_c;
static PyObject *__pyx_n_s_class;
static PyObject *__pyx_n_s_cline_in_traceback;
static PyObject *__pyx_n_s_closed;
static PyObject *__pyx_n_u_closed;
static PyObject *__pyx_kp_u_closed_is_expected;
static PyObject *__pyx_n_s_closed_left;
static PyObject *__pyx_n_s_closed_right;
static PyObject *__pyx_n_s_concatenate;
static PyObject *__pyx_kp_u_contains___not_defined_for_two;
static PyObject *__pyx_kp_s_contiguous_and_direct;
static PyObject *__pyx_kp_s_contiguous_and_indirect;
static PyObject *__pyx_n_s_counts;
static PyObject *__pyx_n_u_dateoffset;
static PyObject *__pyx_n_s_defaults;
static PyObject *__pyx_n_s_dict;
static PyObject *__pyx_n_s_dtype;
static PyObject *__pyx_n_s_dtype_is_object;
static PyObject *__pyx_kp_u_elements;
static PyObject *__pyx_n_s_empty;
static PyObject *__pyx_n_s_encode;
static PyObject *__pyx_n_s_end;
static PyObject *__pyx_n_s_enumerate;
static PyObject *__pyx_n_s_error;
static PyObject *__pyx_n_s_flags;
static PyObject *__pyx_n_u_float32;
static PyObject *__pyx_n_s_float32_t;
static PyObject *__pyx_n_u_float64;
static PyObject *__pyx_n_s_float64_t;
static PyObject *__pyx_n_s_format;
static PyObject *__pyx_n_s_fortran;
static PyObject *__pyx_n_u_fortran;
static PyObject *__pyx_n_s_get_indexer;
static PyObject *__pyx_n_s_get_indexer_non_unique;
static PyObject *__pyx_n_s_get_loc;
static PyObject *__pyx_n_s_get_partial_overlap;
static PyObject *__pyx_n_s_getstate;
static PyObject *__pyx_kp_s_got_differing_extents_in_dimensi;
static PyObject *__pyx_n_s_i;
static PyObject *__pyx_n_s_id;
static PyObject *__pyx_n_s_import;
static PyObject *__pyx_kp_u_index_out_of_bounds;
static PyObject *__pyx_kp_u_indexer_does_not_intersect_a_uni;
static PyObject *__pyx_n_s_indices;
static PyObject *__pyx_n_u_int32;
static PyObject *__pyx_n_s_int32_t;
static PyObject *__pyx_n_u_int64;
static PyObject *__pyx_n_s_int64_t;
static PyObject *__pyx_n_s_interval;
static PyObject *__pyx_n_u_interval;
static PyObject *__pyx_n_s_intervals;
static PyObject *__pyx_kp_u_intervals_must_all_be_closed_on;
static PyObject *__pyx_n_s_intervals_to_interval_bounds;
static PyObject *__pyx_n_u_intp;
static PyObject *__pyx_kp_u_invalid_option_for_closed_closed;
static PyObject *__pyx_kp_u_invalid_option_for_closed_s;
static PyObject *__pyx_n_s_is_monotonic;
static PyObject *__pyx_n_s_isnan;
static PyObject *__pyx_n_s_itemsize;
static PyObject *__pyx_kp_s_itemsize_0_for_cython_array;
static PyObject *__pyx_n_s_iv;
static PyObject *__pyx_n_s_key;
static PyObject *__pyx_n_s_key_left;
static PyObject *__pyx_n_s_key_right;
static PyObject *__pyx_n_s_kind;
static PyObject *__pyx_n_s_kwargs;
static PyObject *__pyx_n_s_le;
static PyObject *__pyx_n_s_leaf_size;
static PyObject *__pyx_n_s_left;
static PyObject *__pyx_n_u_left;
static PyObject *__pyx_kp_u_left_2;
static PyObject *__pyx_kp_u_left_and_right_must_have_the_sam;
static PyObject *__pyx_kp_u_left_side_of_interval_must_be_ri;
static PyObject *__pyx_n_s_left_sorter;
static PyObject *__pyx_n_s_left_tz;
static PyObject *__pyx_n_s_length;
static PyObject *__pyx_n_s_lexsort;
static PyObject *__pyx_n_s_lt;
static PyObject *__pyx_n_s_main;
static PyObject *__pyx_n_s_max;
static PyObject *__pyx_n_s_median;
static PyObject *__pyx_n_s_memview;
static PyObject *__pyx_n_s_min;
static PyObject *__pyx_n_s_missing;
static PyObject *__pyx_n_s_mode;
static PyObject *__pyx_n_s_msg;
static PyObject *__pyx_n_s_n;
static PyObject *__pyx_n_s_n_elements;
static PyObject *__pyx_n_s_name;
static PyObject *__pyx_n_s_name_2;
static PyObject *__pyx_kp_u_name_left_r_right_r_closed_clos;
static PyObject *__pyx_n_s_nan;
static PyObject *__pyx_kp_u_ndarray_is_not_C_contiguous;
static PyObject *__pyx_kp_u_ndarray_is_not_Fortran_contiguou;
static PyObject *__pyx_n_s_ndim;
static PyObject *__pyx_n_u_ndim;
static PyObject *__pyx_n_u_neither;
static PyObject *__pyx_n_s_new;
static PyObject *__pyx_kp_s_no_default___reduce___due_to_non;
static PyObject *__pyx_n_s_np;
static PyObject *__pyx_n_s_numbers;
static PyObject *__pyx_n_s_numpy;
static PyObject *__pyx_kp_u_numpy_core_multiarray_failed_to;
static PyObject *__pyx_kp_u_numpy_core_umath_failed_to_impor;
static PyObject *__pyx_n_s_obj;
static PyObject *__pyx_n_s_old_len;
static PyObject *__pyx_n_s_op;
static PyObject *__pyx_n_s_open_left;
static PyObject *__pyx_n_s_open_right;
static PyObject *__pyx_n_s_operator;
static PyObject *__pyx_n_s_other;
static PyObject *__pyx_n_u_other;
static PyObject *__pyx_kp_u_other_must_be_an_Interval_got_o;
static PyObject *__pyx_kp_u_overlapping;
static PyObject *__pyx_n_s_pack;
static PyObject *__pyx_n_s_pandas;
static PyObject *__pyx_n_s_pandas__libs_algos;
static PyObject *__pyx_n_s_pandas__libs_interval;
static PyObject *__pyx_kp_s_pandas__libs_interval_pyx;
static PyObject *__pyx_kp_s_pandas__libs_intervaltree_pxi;
static PyObject *__pyx_n_s_pandas__libs_tslibs;
static PyObject *__pyx_n_s_pandas__libs_tslibs_timedeltas;
static PyObject *__pyx_n_u_period;
static PyObject *__pyx_n_s_pickle;
static PyObject *__pyx_n_s_point;
static PyObject *__pyx_n_s_pyx_PickleError;
static PyObject *__pyx_n_s_pyx_checksum;
static PyObject *__pyx_n_s_pyx_fuse_0query;
static PyObject *__pyx_n_s_pyx_fuse_1query;
static PyObject *__pyx_n_s_pyx_fuse_2query;
static PyObject *__pyx_n_s_pyx_fuse_3query;
static PyObject *__pyx_n_s_pyx_fuse_4query;
static PyObject *__pyx_n_s_pyx_getbuffer;
static PyObject *__pyx_n_s_pyx_result;
static PyObject *__pyx_n_s_pyx_state;
static PyObject *__pyx_n_s_pyx_type;
static PyObject *__pyx_n_s_pyx_unpickle_Enum;
static PyObject *__pyx_n_s_pyx_unpickle_Float32ClosedBoth;
static PyObject *__pyx_n_s_pyx_unpickle_Float32ClosedLeft;
static PyObject *__pyx_n_s_pyx_unpickle_Float32ClosedNeit;
static PyObject *__pyx_n_s_pyx_unpickle_Float32ClosedRigh;
static PyObject *__pyx_n_s_pyx_unpickle_Float64ClosedBoth;
static PyObject *__pyx_n_s_pyx_unpickle_Float64ClosedLeft;
static PyObject *__pyx_n_s_pyx_unpickle_Float64ClosedNeit;
static PyObject *__pyx_n_s_pyx_unpickle_Float64ClosedRigh;
static PyObject *__pyx_n_s_pyx_unpickle_Int32ClosedBothIn;
static PyObject *__pyx_n_s_pyx_unpickle_Int32ClosedLeftIn;
static PyObject *__pyx_n_s_pyx_unpickle_Int32ClosedNeithe;
static PyObject *__pyx_n_s_pyx_unpickle_Int32ClosedRightI;
static PyObject *__pyx_n_s_pyx_unpickle_Int64ClosedBothIn;
static PyObject *__pyx_n_s_pyx_unpickle_Int64ClosedLeftIn;
static PyObject *__pyx_n_s_pyx_unpickle_Int64ClosedNeithe;
static PyObject *__pyx_n_s_pyx_unpickle_Int64ClosedRightI;
static PyObject *__pyx_n_s_pyx_unpickle_IntervalMixin;
static PyObject *__pyx_n_s_pyx_unpickle_IntervalTree;
static PyObject *__pyx_n_s_pyx_unpickle_Uint64ClosedBothI;
static PyObject *__pyx_n_s_pyx_unpickle_Uint64ClosedLeftI;
static PyObject *__pyx_n_s_pyx_unpickle_Uint64ClosedNeith;
static PyObject *__pyx_n_s_pyx_unpickle_Uint64ClosedRight;
static PyObject *__pyx_n_s_pyx_vtable;
static PyObject *__pyx_n_s_query;
static PyObject *__pyx_n_s_range;
static PyObject *__pyx_n_s_reduce;
static PyObject *__pyx_n_s_reduce_cython;
static PyObject *__pyx_n_s_reduce_ex;
static PyObject *__pyx_n_s_repr_base;
static PyObject *__pyx_n_s_result;
static PyObject *__pyx_n_s_result_type;
static PyObject *__pyx_n_s_right;
static PyObject *__pyx_n_u_right;
static PyObject *__pyx_kp_u_right_2;
static PyObject *__pyx_n_s_right_sorter;
static PyObject *__pyx_n_s_right_tz;
static PyObject *__pyx_n_s_s;
static PyObject *__pyx_n_s_searchsorted;
static PyObject *__pyx_n_s_seen_closed;
static PyObject *__pyx_n_s_self;
static PyObject *__pyx_n_s_setstate;
static PyObject *__pyx_n_s_setstate_cython;
static PyObject *__pyx_n_s_shape;
static PyObject *__pyx_n_s_short_repr;
static PyObject *__pyx_n_s_side;
static PyObject *__pyx_n_s_size;
static PyObject *__pyx_n_s_sorter;
static PyObject *__pyx_n_s_split;
static PyObject *__pyx_n_s_start;
static PyObject *__pyx_kp_u_start_left_right_end;
static PyObject *__pyx_n_s_step;
static PyObject *__pyx_n_s_stop;
static PyObject *__pyx_kp_s_strided_and_direct;
static PyObject *__pyx_kp_s_strided_and_direct_or_indirect;
static PyObject *__pyx_kp_s_strided_and_indirect;
static PyObject *__pyx_kp_s_stringsource;
static PyObject *__pyx_n_s_strip;
static PyObject *__pyx_n_s_struct;
static PyObject *__pyx_n_s_target;
static PyObject *__pyx_n_s_test;
static PyObject *__pyx_n_s_typ;
static PyObject *__pyx_n_s_typ_2;
static PyObject *__pyx_n_u_typ_2;
static PyObject *__pyx_kp_u_type_typ_with_value_iv_is_not_an;
static PyObject *__pyx_n_s_tzinfo;
static PyObject *__pyx_n_u_uint64;
static PyObject *__pyx_n_s_uint64_t;
static PyObject *__pyx_kp_s_unable_to_allocate_array_data;
static PyObject *__pyx_kp_s_unable_to_allocate_shape_and_str;
static PyObject *__pyx_n_s_unique;
static PyObject *__pyx_kp_u_unknown_dtype_code_in_numpy_pxd;
static PyObject *__pyx_kp_u_unorderable_types_name_op_other;
static PyObject *__pyx_n_s_unpack;
static PyObject *__pyx_n_s_update;
static PyObject *__pyx_n_s_validate_closed;
static PyObject *__pyx_n_s_validate_endpoint;
static PyObject *__pyx_pf_6pandas_5_libs_8interval_13IntervalMixin_11closed_left___get__(struct __pyx_obj_6pandas_5_libs_8interval_IntervalMixin *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_13IntervalMixin_12closed_right___get__(struct __pyx_obj_6pandas_5_libs_8interval_IntervalMixin *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_13IntervalMixin_9open_left___get__(struct __pyx_obj_6pandas_5_libs_8interval_IntervalMixin *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_13IntervalMixin_10open_right___get__(struct __pyx_obj_6pandas_5_libs_8interval_IntervalMixin *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_13IntervalMixin_3mid___get__(struct __pyx_obj_6pandas_5_libs_8interval_IntervalMixin *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_13IntervalMixin_6length___get__(struct __pyx_obj_6pandas_5_libs_8interval_IntervalMixin *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_13IntervalMixin_8is_empty___get__(struct __pyx_obj_6pandas_5_libs_8interval_IntervalMixin *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_13IntervalMixin__check_closed_matches(struct __pyx_obj_6pandas_5_libs_8interval_IntervalMixin *__pyx_v_self, PyObject *__pyx_v_other, PyObject *__pyx_v_name); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_13IntervalMixin_2__reduce_cython__(struct __pyx_obj_6pandas_5_libs_8interval_IntervalMixin *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_13IntervalMixin_4__setstate_cython__(struct __pyx_obj_6pandas_5_libs_8interval_IntervalMixin *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_6pandas_5_libs_8interval_8Interval___init__(struct __pyx_obj_6pandas_5_libs_8interval_Interval *__pyx_v_self, PyObject *__pyx_v_left, PyObject *__pyx_v_right, PyObject *__pyx_v_closed); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_8Interval_2_validate_endpoint(CYTHON_UNUSED struct __pyx_obj_6pandas_5_libs_8interval_Interval *__pyx_v_self, PyObject *__pyx_v_endpoint); /* proto */
static Py_hash_t __pyx_pf_6pandas_5_libs_8interval_8Interval_4__hash__(struct __pyx_obj_6pandas_5_libs_8interval_Interval *__pyx_v_self); /* proto */
static int __pyx_pf_6pandas_5_libs_8interval_8Interval_6__contains__(struct __pyx_obj_6pandas_5_libs_8interval_Interval *__pyx_v_self, PyObject *__pyx_v_key); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_8Interval_8__richcmp__(struct __pyx_obj_6pandas_5_libs_8interval_Interval *__pyx_v_self, PyObject *__pyx_v_other, PyObject *__pyx_v_op); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_8Interval_10__reduce__(struct __pyx_obj_6pandas_5_libs_8interval_Interval *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_8Interval_12_repr_base(struct __pyx_obj_6pandas_5_libs_8interval_Interval *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_8Interval_14__repr__(struct __pyx_obj_6pandas_5_libs_8interval_Interval *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_8Interval_16__str__(struct __pyx_obj_6pandas_5_libs_8interval_Interval *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_8Interval_18__add__(PyObject *__pyx_v_self, PyObject *__pyx_v_y); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_8Interval_20__sub__(PyObject *__pyx_v_self, PyObject *__pyx_v_y); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_8Interval_22__mul__(PyObject *__pyx_v_self, PyObject *__pyx_v_y); /* proto */
#if PY_MAJOR_VERSION < 3 || (CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x03050000)
static PyObject *__pyx_pf_6pandas_5_libs_8interval_8Interval_24__div__(PyObject *__pyx_v_self, PyObject *__pyx_v_y); /* proto */
#endif
static PyObject *__pyx_pf_6pandas_5_libs_8interval_8Interval_26__truediv__(PyObject *__pyx_v_self, PyObject *__pyx_v_y); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_8Interval_28__floordiv__(PyObject *__pyx_v_self, PyObject *__pyx_v_y); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_8Interval_30overlaps(struct __pyx_obj_6pandas_5_libs_8interval_Interval *__pyx_v_self, PyObject *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_8Interval_4left___get__(struct __pyx_obj_6pandas_5_libs_8interval_Interval *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_8Interval_5right___get__(struct __pyx_obj_6pandas_5_libs_8interval_Interval *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_8Interval_6closed___get__(struct __pyx_obj_6pandas_5_libs_8interval_Interval *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_intervals_to_interval_bounds(CYTHON_UNUSED PyObject *__pyx_self, PyArrayObject *__pyx_v_intervals, int __pyx_v_validate_closed); /* proto */
static int __pyx_pf_6pandas_5_libs_8interval_12IntervalTree___init__(struct __pyx_obj_6pandas_5_libs_8interval_IntervalTree *__pyx_v_self, PyObject *__pyx_v_left, PyObject *__pyx_v_right, PyObject *__pyx_v_closed, PyObject *__pyx_v_leaf_size); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_12IntervalTree_11left_sorter___get__(struct __pyx_obj_6pandas_5_libs_8interval_IntervalTree *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_12IntervalTree_12right_sorter___get__(struct __pyx_obj_6pandas_5_libs_8interval_IntervalTree *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_12IntervalTree_14is_overlapping___get__(struct __pyx_obj_6pandas_5_libs_8interval_IntervalTree *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_12IntervalTree_23is_monotonic_increasing___get__(struct __pyx_obj_6pandas_5_libs_8interval_IntervalTree *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_12IntervalTree_2get_loc(struct __pyx_obj_6pandas_5_libs_8interval_IntervalTree *__pyx_v_signatures, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs, CYTHON_UNUSED PyObject *__pyx_v_defaults); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_12IntervalTree_16get_loc(struct __pyx_obj_6pandas_5_libs_8interval_IntervalTree *__pyx_v_self, __pyx_t_5numpy_float64_t __pyx_v_key); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_12IntervalTree_18get_loc(struct __pyx_obj_6pandas_5_libs_8interval_IntervalTree *__pyx_v_self, __pyx_t_5numpy_float32_t __pyx_v_key); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_12IntervalTree_20get_loc(struct __pyx_obj_6pandas_5_libs_8interval_IntervalTree *__pyx_v_self, __pyx_t_5numpy_int64_t __pyx_v_key); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_12IntervalTree_22get_loc(struct __pyx_obj_6pandas_5_libs_8interval_IntervalTree *__pyx_v_self, __pyx_t_5numpy_int32_t __pyx_v_key); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_12IntervalTree_24get_loc(struct __pyx_obj_6pandas_5_libs_8interval_IntervalTree *__pyx_v_self, __pyx_t_5numpy_uint64_t __pyx_v_key); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_12IntervalTree_4_get_partial_overlap(struct __pyx_obj_6pandas_5_libs_8interval_IntervalTree *__pyx_v_self, PyObject *__pyx_v_key_left, PyObject *__pyx_v_key_right, PyObject *__pyx_v_side); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_12IntervalTree_6get_loc_interval(struct __pyx_obj_6pandas_5_libs_8interval_IntervalTree *__pyx_v_self, PyObject *__pyx_v_key_left, PyObject *__pyx_v_key_right); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_12IntervalTree_8get_indexer(struct __pyx_obj_6pandas_5_libs_8interval_IntervalTree *__pyx_v_signatures, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs, CYTHON_UNUSED PyObject *__pyx_v_defaults); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_12IntervalTree_28get_indexer(struct __pyx_obj_6pandas_5_libs_8interval_IntervalTree *__pyx_v_self, __Pyx_memviewslice __pyx_v_target); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_12IntervalTree_30get_indexer(struct __pyx_obj_6pandas_5_libs_8interval_IntervalTree *__pyx_v_self, __Pyx_memviewslice __pyx_v_target); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_12IntervalTree_32get_indexer(struct __pyx_obj_6pandas_5_libs_8interval_IntervalTree *__pyx_v_self, __Pyx_memviewslice __pyx_v_target); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_12IntervalTree_34get_indexer(struct __pyx_obj_6pandas_5_libs_8interval_IntervalTree *__pyx_v_self, __Pyx_memviewslice __pyx_v_target); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_12IntervalTree_36get_indexer(struct __pyx_obj_6pandas_5_libs_8interval_IntervalTree *__pyx_v_self, __Pyx_memviewslice __pyx_v_target); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_12IntervalTree_10get_indexer_non_unique(struct __pyx_obj_6pandas_5_libs_8interval_IntervalTree *__pyx_v_signatures, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs, CYTHON_UNUSED PyObject *__pyx_v_defaults); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_12IntervalTree_40get_indexer_non_unique(struct __pyx_obj_6pandas_5_libs_8interval_IntervalTree *__pyx_v_self, __Pyx_memviewslice __pyx_v_target); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_12IntervalTree_42get_indexer_non_unique(struct __pyx_obj_6pandas_5_libs_8interval_IntervalTree *__pyx_v_self, __Pyx_memviewslice __pyx_v_target); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_12IntervalTree_44get_indexer_non_unique(struct __pyx_obj_6pandas_5_libs_8interval_IntervalTree *__pyx_v_self, __Pyx_memviewslice __pyx_v_target); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_12IntervalTree_46get_indexer_non_unique(struct __pyx_obj_6pandas_5_libs_8interval_IntervalTree *__pyx_v_self, __Pyx_memviewslice __pyx_v_target); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_12IntervalTree_48get_indexer_non_unique(struct __pyx_obj_6pandas_5_libs_8interval_IntervalTree *__pyx_v_self, __Pyx_memviewslice __pyx_v_target); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_12IntervalTree_12__repr__(struct __pyx_obj_6pandas_5_libs_8interval_IntervalTree *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_12IntervalTree_14clear_mapping(CYTHON_UNUSED struct __pyx_obj_6pandas_5_libs_8interval_IntervalTree *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_12IntervalTree_4left___get__(struct __pyx_obj_6pandas_5_libs_8interval_IntervalTree *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_12IntervalTree_5right___get__(struct __pyx_obj_6pandas_5_libs_8interval_IntervalTree *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_12IntervalTree_4root___get__(struct __pyx_obj_6pandas_5_libs_8interval_IntervalTree *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_12IntervalTree_5dtype___get__(struct __pyx_obj_6pandas_5_libs_8interval_IntervalTree *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_12IntervalTree_6closed___get__(struct __pyx_obj_6pandas_5_libs_8interval_IntervalTree *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_12IntervalTree_52__reduce_cython__(struct __pyx_obj_6pandas_5_libs_8interval_IntervalTree *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_12IntervalTree_54__setstate_cython__(struct __pyx_obj_6pandas_5_libs_8interval_IntervalTree *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode___init__(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode *__pyx_v_self, PyArrayObject *__pyx_v_left, PyArrayObject *__pyx_v_right, PyArrayObject *__pyx_v_indices, __pyx_t_5numpy_int64_t __pyx_v_leaf_size); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_2query(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode *__pyx_v_signatures, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs, CYTHON_UNUSED PyObject *__pyx_v_defaults); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_8__pyx_fuse_0query(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float64_t __pyx_v_point); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_10__pyx_fuse_1query(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float32_t __pyx_v_point); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_12__pyx_fuse_2query(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int64_t __pyx_v_point); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_14__pyx_fuse_3query(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int32_t __pyx_v_point); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_16__pyx_fuse_4query(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_uint64_t __pyx_v_point); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_4__repr__(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_6counts(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_5pivot___get__(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_10n_elements___get__(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_8n_center___get__(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_9leaf_size___get__(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_12is_leaf_node___get__(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_20__reduce_cython__(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_22__setstate_cython__(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode___init__(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode *__pyx_v_self, PyArrayObject *__pyx_v_left, PyArrayObject *__pyx_v_right, PyArrayObject *__pyx_v_indices, __pyx_t_5numpy_int64_t __pyx_v_leaf_size); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_2query(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode *__pyx_v_signatures, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs, CYTHON_UNUSED PyObject *__pyx_v_defaults); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_8__pyx_fuse_0query(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float64_t __pyx_v_point); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_10__pyx_fuse_1query(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float32_t __pyx_v_point); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_12__pyx_fuse_2query(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int64_t __pyx_v_point); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_14__pyx_fuse_3query(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int32_t __pyx_v_point); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_16__pyx_fuse_4query(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_uint64_t __pyx_v_point); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_4__repr__(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_6counts(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_5pivot___get__(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_10n_elements___get__(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_8n_center___get__(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_9leaf_size___get__(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_12is_leaf_node___get__(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_20__reduce_cython__(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_22__setstate_cython__(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode___init__(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode *__pyx_v_self, PyArrayObject *__pyx_v_left, PyArrayObject *__pyx_v_right, PyArrayObject *__pyx_v_indices, __pyx_t_5numpy_int64_t __pyx_v_leaf_size); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_2query(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode *__pyx_v_signatures, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs, CYTHON_UNUSED PyObject *__pyx_v_defaults); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_8__pyx_fuse_0query(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float64_t __pyx_v_point); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_10__pyx_fuse_1query(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float32_t __pyx_v_point); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_12__pyx_fuse_2query(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int64_t __pyx_v_point); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_14__pyx_fuse_3query(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int32_t __pyx_v_point); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_16__pyx_fuse_4query(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_uint64_t __pyx_v_point); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_4__repr__(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_6counts(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_5pivot___get__(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_10n_elements___get__(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_8n_center___get__(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_9leaf_size___get__(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_12is_leaf_node___get__(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_20__reduce_cython__(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_22__setstate_cython__(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode___init__(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode *__pyx_v_self, PyArrayObject *__pyx_v_left, PyArrayObject *__pyx_v_right, PyArrayObject *__pyx_v_indices, __pyx_t_5numpy_int64_t __pyx_v_leaf_size); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_2query(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode *__pyx_v_signatures, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs, CYTHON_UNUSED PyObject *__pyx_v_defaults); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_8__pyx_fuse_0query(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float64_t __pyx_v_point); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_10__pyx_fuse_1query(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float32_t __pyx_v_point); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_12__pyx_fuse_2query(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int64_t __pyx_v_point); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_14__pyx_fuse_3query(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int32_t __pyx_v_point); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_16__pyx_fuse_4query(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_uint64_t __pyx_v_point); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_4__repr__(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_6counts(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_5pivot___get__(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_10n_elements___get__(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_8n_center___get__(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_9leaf_size___get__(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_12is_leaf_node___get__(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_20__reduce_cython__(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_22__setstate_cython__(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode___init__(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode *__pyx_v_self, PyArrayObject *__pyx_v_left, PyArrayObject *__pyx_v_right, PyArrayObject *__pyx_v_indices, __pyx_t_5numpy_int64_t __pyx_v_leaf_size); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_2query(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode *__pyx_v_signatures, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs, CYTHON_UNUSED PyObject *__pyx_v_defaults); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_8__pyx_fuse_0query(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float64_t __pyx_v_point); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_10__pyx_fuse_1query(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float32_t __pyx_v_point); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_12__pyx_fuse_2query(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int64_t __pyx_v_point); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_14__pyx_fuse_3query(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int32_t __pyx_v_point); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_16__pyx_fuse_4query(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_uint64_t __pyx_v_point); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_4__repr__(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_6counts(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_5pivot___get__(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_10n_elements___get__(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_8n_center___get__(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_9leaf_size___get__(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_12is_leaf_node___get__(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_20__reduce_cython__(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_22__setstate_cython__(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode___init__(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode *__pyx_v_self, PyArrayObject *__pyx_v_left, PyArrayObject *__pyx_v_right, PyArrayObject *__pyx_v_indices, __pyx_t_5numpy_int64_t __pyx_v_leaf_size); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_2query(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode *__pyx_v_signatures, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs, CYTHON_UNUSED PyObject *__pyx_v_defaults); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_8__pyx_fuse_0query(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float64_t __pyx_v_point); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_10__pyx_fuse_1query(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float32_t __pyx_v_point); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_12__pyx_fuse_2query(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int64_t __pyx_v_point); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_14__pyx_fuse_3query(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int32_t __pyx_v_point); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_16__pyx_fuse_4query(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_uint64_t __pyx_v_point); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_4__repr__(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_6counts(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_5pivot___get__(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_10n_elements___get__(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_8n_center___get__(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_9leaf_size___get__(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_12is_leaf_node___get__(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_20__reduce_cython__(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_22__setstate_cython__(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode___init__(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode *__pyx_v_self, PyArrayObject *__pyx_v_left, PyArrayObject *__pyx_v_right, PyArrayObject *__pyx_v_indices, __pyx_t_5numpy_int64_t __pyx_v_leaf_size); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_2query(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode *__pyx_v_signatures, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs, CYTHON_UNUSED PyObject *__pyx_v_defaults); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_8__pyx_fuse_0query(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float64_t __pyx_v_point); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_10__pyx_fuse_1query(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float32_t __pyx_v_point); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_12__pyx_fuse_2query(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int64_t __pyx_v_point); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_14__pyx_fuse_3query(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int32_t __pyx_v_point); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_16__pyx_fuse_4query(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_uint64_t __pyx_v_point); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_4__repr__(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_6counts(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_5pivot___get__(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_10n_elements___get__(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_8n_center___get__(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_9leaf_size___get__(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_12is_leaf_node___get__(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_20__reduce_cython__(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_22__setstate_cython__(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode___init__(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode *__pyx_v_self, PyArrayObject *__pyx_v_left, PyArrayObject *__pyx_v_right, PyArrayObject *__pyx_v_indices, __pyx_t_5numpy_int64_t __pyx_v_leaf_size); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_2query(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode *__pyx_v_signatures, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs, CYTHON_UNUSED PyObject *__pyx_v_defaults); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_8__pyx_fuse_0query(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float64_t __pyx_v_point); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_10__pyx_fuse_1query(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float32_t __pyx_v_point); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_12__pyx_fuse_2query(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int64_t __pyx_v_point); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_14__pyx_fuse_3query(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int32_t __pyx_v_point); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_16__pyx_fuse_4query(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_uint64_t __pyx_v_point); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_4__repr__(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_6counts(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_5pivot___get__(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_10n_elements___get__(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_8n_center___get__(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_9leaf_size___get__(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_12is_leaf_node___get__(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_20__reduce_cython__(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_22__setstate_cython__(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode___init__(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode *__pyx_v_self, PyArrayObject *__pyx_v_left, PyArrayObject *__pyx_v_right, PyArrayObject *__pyx_v_indices, __pyx_t_5numpy_int64_t __pyx_v_leaf_size); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_2query(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode *__pyx_v_signatures, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs, CYTHON_UNUSED PyObject *__pyx_v_defaults); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_8__pyx_fuse_0query(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float64_t __pyx_v_point); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_10__pyx_fuse_1query(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float32_t __pyx_v_point); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_12__pyx_fuse_2query(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int64_t __pyx_v_point); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_14__pyx_fuse_3query(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int32_t __pyx_v_point); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_16__pyx_fuse_4query(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_uint64_t __pyx_v_point); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_4__repr__(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_6counts(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_5pivot___get__(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_10n_elements___get__(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_8n_center___get__(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_9leaf_size___get__(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_12is_leaf_node___get__(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_20__reduce_cython__(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_22__setstate_cython__(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode___init__(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode *__pyx_v_self, PyArrayObject *__pyx_v_left, PyArrayObject *__pyx_v_right, PyArrayObject *__pyx_v_indices, __pyx_t_5numpy_int64_t __pyx_v_leaf_size); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_2query(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode *__pyx_v_signatures, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs, CYTHON_UNUSED PyObject *__pyx_v_defaults); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_8__pyx_fuse_0query(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float64_t __pyx_v_point); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_10__pyx_fuse_1query(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float32_t __pyx_v_point); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_12__pyx_fuse_2query(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int64_t __pyx_v_point); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_14__pyx_fuse_3query(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int32_t __pyx_v_point); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_16__pyx_fuse_4query(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_uint64_t __pyx_v_point); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_4__repr__(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_6counts(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_5pivot___get__(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_10n_elements___get__(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_8n_center___get__(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_9leaf_size___get__(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_12is_leaf_node___get__(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_20__reduce_cython__(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_22__setstate_cython__(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode___init__(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode *__pyx_v_self, PyArrayObject *__pyx_v_left, PyArrayObject *__pyx_v_right, PyArrayObject *__pyx_v_indices, __pyx_t_5numpy_int64_t __pyx_v_leaf_size); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_2query(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode *__pyx_v_signatures, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs, CYTHON_UNUSED PyObject *__pyx_v_defaults); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_8__pyx_fuse_0query(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float64_t __pyx_v_point); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_10__pyx_fuse_1query(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float32_t __pyx_v_point); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_12__pyx_fuse_2query(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int64_t __pyx_v_point); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_14__pyx_fuse_3query(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int32_t __pyx_v_point); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_16__pyx_fuse_4query(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_uint64_t __pyx_v_point); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_4__repr__(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_6counts(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_5pivot___get__(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_10n_elements___get__(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_8n_center___get__(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_9leaf_size___get__(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_12is_leaf_node___get__(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_20__reduce_cython__(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_22__setstate_cython__(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode___init__(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode *__pyx_v_self, PyArrayObject *__pyx_v_left, PyArrayObject *__pyx_v_right, PyArrayObject *__pyx_v_indices, __pyx_t_5numpy_int64_t __pyx_v_leaf_size); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_2query(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode *__pyx_v_signatures, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs, CYTHON_UNUSED PyObject *__pyx_v_defaults); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_8__pyx_fuse_0query(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float64_t __pyx_v_point); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_10__pyx_fuse_1query(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float32_t __pyx_v_point); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_12__pyx_fuse_2query(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int64_t __pyx_v_point); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_14__pyx_fuse_3query(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int32_t __pyx_v_point); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_16__pyx_fuse_4query(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_uint64_t __pyx_v_point); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_4__repr__(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_6counts(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_5pivot___get__(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_10n_elements___get__(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_8n_center___get__(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_9leaf_size___get__(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_12is_leaf_node___get__(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_20__reduce_cython__(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_22__setstate_cython__(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode___init__(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode *__pyx_v_self, PyArrayObject *__pyx_v_left, PyArrayObject *__pyx_v_right, PyArrayObject *__pyx_v_indices, __pyx_t_5numpy_int64_t __pyx_v_leaf_size); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_2query(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode *__pyx_v_signatures, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs, CYTHON_UNUSED PyObject *__pyx_v_defaults); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_8__pyx_fuse_0query(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float64_t __pyx_v_point); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_10__pyx_fuse_1query(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float32_t __pyx_v_point); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_12__pyx_fuse_2query(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int64_t __pyx_v_point); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_14__pyx_fuse_3query(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int32_t __pyx_v_point); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_16__pyx_fuse_4query(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_uint64_t __pyx_v_point); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_4__repr__(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_6counts(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_5pivot___get__(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_10n_elements___get__(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_8n_center___get__(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_9leaf_size___get__(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_12is_leaf_node___get__(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_20__reduce_cython__(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_22__setstate_cython__(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode___init__(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode *__pyx_v_self, PyArrayObject *__pyx_v_left, PyArrayObject *__pyx_v_right, PyArrayObject *__pyx_v_indices, __pyx_t_5numpy_int64_t __pyx_v_leaf_size); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_2query(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode *__pyx_v_signatures, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs, CYTHON_UNUSED PyObject *__pyx_v_defaults); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_8__pyx_fuse_0query(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float64_t __pyx_v_point); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_10__pyx_fuse_1query(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float32_t __pyx_v_point); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_12__pyx_fuse_2query(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int64_t __pyx_v_point); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_14__pyx_fuse_3query(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int32_t __pyx_v_point); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_16__pyx_fuse_4query(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_uint64_t __pyx_v_point); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_4__repr__(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_6counts(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_5pivot___get__(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_10n_elements___get__(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_8n_center___get__(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_9leaf_size___get__(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_12is_leaf_node___get__(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_20__reduce_cython__(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_22__setstate_cython__(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode___init__(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode *__pyx_v_self, PyArrayObject *__pyx_v_left, PyArrayObject *__pyx_v_right, PyArrayObject *__pyx_v_indices, __pyx_t_5numpy_int64_t __pyx_v_leaf_size); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_2query(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode *__pyx_v_signatures, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs, CYTHON_UNUSED PyObject *__pyx_v_defaults); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_8__pyx_fuse_0query(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float64_t __pyx_v_point); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_10__pyx_fuse_1query(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float32_t __pyx_v_point); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_12__pyx_fuse_2query(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int64_t __pyx_v_point); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_14__pyx_fuse_3query(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int32_t __pyx_v_point); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_16__pyx_fuse_4query(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_uint64_t __pyx_v_point); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_4__repr__(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_6counts(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_5pivot___get__(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_10n_elements___get__(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_8n_center___get__(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_9leaf_size___get__(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_12is_leaf_node___get__(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_20__reduce_cython__(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_22__setstate_cython__(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode___init__(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode *__pyx_v_self, PyArrayObject *__pyx_v_left, PyArrayObject *__pyx_v_right, PyArrayObject *__pyx_v_indices, __pyx_t_5numpy_int64_t __pyx_v_leaf_size); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_2query(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode *__pyx_v_signatures, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs, CYTHON_UNUSED PyObject *__pyx_v_defaults); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_8__pyx_fuse_0query(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float64_t __pyx_v_point); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_10__pyx_fuse_1query(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float32_t __pyx_v_point); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_12__pyx_fuse_2query(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int64_t __pyx_v_point); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_14__pyx_fuse_3query(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int32_t __pyx_v_point); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_16__pyx_fuse_4query(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_uint64_t __pyx_v_point); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_4__repr__(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_6counts(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_5pivot___get__(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_10n_elements___get__(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_8n_center___get__(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_9leaf_size___get__(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_12is_leaf_node___get__(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_20__reduce_cython__(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_22__setstate_cython__(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode___init__(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode *__pyx_v_self, PyArrayObject *__pyx_v_left, PyArrayObject *__pyx_v_right, PyArrayObject *__pyx_v_indices, __pyx_t_5numpy_int64_t __pyx_v_leaf_size); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_2query(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode *__pyx_v_signatures, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs, CYTHON_UNUSED PyObject *__pyx_v_defaults); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_8__pyx_fuse_0query(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float64_t __pyx_v_point); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_10__pyx_fuse_1query(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float32_t __pyx_v_point); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_12__pyx_fuse_2query(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int64_t __pyx_v_point); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_14__pyx_fuse_3query(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int32_t __pyx_v_point); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_16__pyx_fuse_4query(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_uint64_t __pyx_v_point); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_4__repr__(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_6counts(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_5pivot___get__(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_10n_elements___get__(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_8n_center___get__(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_9leaf_size___get__(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_12is_leaf_node___get__(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_20__reduce_cython__(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_22__setstate_cython__(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode___init__(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode *__pyx_v_self, PyArrayObject *__pyx_v_left, PyArrayObject *__pyx_v_right, PyArrayObject *__pyx_v_indices, __pyx_t_5numpy_int64_t __pyx_v_leaf_size); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_2query(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode *__pyx_v_signatures, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs, CYTHON_UNUSED PyObject *__pyx_v_defaults); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_8__pyx_fuse_0query(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float64_t __pyx_v_point); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_10__pyx_fuse_1query(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float32_t __pyx_v_point); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_12__pyx_fuse_2query(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int64_t __pyx_v_point); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_14__pyx_fuse_3query(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int32_t __pyx_v_point); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_16__pyx_fuse_4query(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_uint64_t __pyx_v_point); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_4__repr__(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_6counts(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_5pivot___get__(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_10n_elements___get__(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_8n_center___get__(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_9leaf_size___get__(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_12is_leaf_node___get__(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_20__reduce_cython__(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_22__setstate_cython__(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode___init__(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode *__pyx_v_self, PyArrayObject *__pyx_v_left, PyArrayObject *__pyx_v_right, PyArrayObject *__pyx_v_indices, __pyx_t_5numpy_int64_t __pyx_v_leaf_size); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_2query(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode *__pyx_v_signatures, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs, CYTHON_UNUSED PyObject *__pyx_v_defaults); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_8__pyx_fuse_0query(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float64_t __pyx_v_point); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_10__pyx_fuse_1query(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float32_t __pyx_v_point); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_12__pyx_fuse_2query(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int64_t __pyx_v_point); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_14__pyx_fuse_3query(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int32_t __pyx_v_point); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_16__pyx_fuse_4query(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_uint64_t __pyx_v_point); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_4__repr__(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_6counts(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_5pivot___get__(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_10n_elements___get__(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_8n_center___get__(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_9leaf_size___get__(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_12is_leaf_node___get__(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_20__reduce_cython__(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_22__setstate_cython__(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode___init__(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode *__pyx_v_self, PyArrayObject *__pyx_v_left, PyArrayObject *__pyx_v_right, PyArrayObject *__pyx_v_indices, __pyx_t_5numpy_int64_t __pyx_v_leaf_size); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_2query(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode *__pyx_v_signatures, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs, CYTHON_UNUSED PyObject *__pyx_v_defaults); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_8__pyx_fuse_0query(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float64_t __pyx_v_point); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_10__pyx_fuse_1query(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float32_t __pyx_v_point); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_12__pyx_fuse_2query(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int64_t __pyx_v_point); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_14__pyx_fuse_3query(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int32_t __pyx_v_point); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_16__pyx_fuse_4query(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_uint64_t __pyx_v_point); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_4__repr__(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_6counts(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_5pivot___get__(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_10n_elements___get__(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_8n_center___get__(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_9leaf_size___get__(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_12is_leaf_node___get__(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_20__reduce_cython__(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_22__setstate_cython__(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_2__pyx_unpickle_IntervalMixin(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_4__pyx_unpickle_IntervalTree(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_6__pyx_unpickle_Float32ClosedLeftIntervalNode(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_8__pyx_unpickle_Float32ClosedRightIntervalNode(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_10__pyx_unpickle_Float32ClosedBothIntervalNode(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_12__pyx_unpickle_Float32ClosedNeitherIntervalNode(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_14__pyx_unpickle_Float64ClosedLeftIntervalNode(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_16__pyx_unpickle_Float64ClosedRightIntervalNode(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_18__pyx_unpickle_Float64ClosedBothIntervalNode(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_20__pyx_unpickle_Float64ClosedNeitherIntervalNode(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_22__pyx_unpickle_Int32ClosedLeftIntervalNode(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_24__pyx_unpickle_Int32ClosedRightIntervalNode(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_26__pyx_unpickle_Int32ClosedBothIntervalNode(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_28__pyx_unpickle_Int32ClosedNeitherIntervalNode(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_30__pyx_unpickle_Int64ClosedLeftIntervalNode(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_32__pyx_unpickle_Int64ClosedRightIntervalNode(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_34__pyx_unpickle_Int64ClosedBothIntervalNode(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_36__pyx_unpickle_Int64ClosedNeitherIntervalNode(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_38__pyx_unpickle_Uint64ClosedLeftIntervalNode(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_40__pyx_unpickle_Uint64ClosedRightIntervalNode(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_42__pyx_unpickle_Uint64ClosedBothIntervalNode(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_6pandas_5_libs_8interval_44__pyx_unpickle_Uint64ClosedNeitherIntervalNode(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_5numpy_7ndarray___getbuffer__(PyArrayObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /* proto */
static void __pyx_pf_5numpy_7ndarray_2__releasebuffer__(PyArrayObject *__pyx_v_self, Py_buffer *__pyx_v_info); /* proto */
static int __pyx_array___pyx_pf_15View_dot_MemoryView_5array___cinit__(struct __pyx_array_obj *__pyx_v_self, PyObject *__pyx_v_shape, Py_ssize_t __pyx_v_itemsize, PyObject *__pyx_v_format, PyObject *__pyx_v_mode, int __pyx_v_allocate_buffer); /* proto */
static int __pyx_array___pyx_pf_15View_dot_MemoryView_5array_2__getbuffer__(struct __pyx_array_obj *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /* proto */
static void __pyx_array___pyx_pf_15View_dot_MemoryView_5array_4__dealloc__(struct __pyx_array_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_5array_7memview___get__(struct __pyx_array_obj *__pyx_v_self); /* proto */
static Py_ssize_t __pyx_array___pyx_pf_15View_dot_MemoryView_5array_6__len__(struct __pyx_array_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_array___pyx_pf_15View_dot_MemoryView_5array_8__getattr__(struct __pyx_array_obj *__pyx_v_self, PyObject *__pyx_v_attr); /* proto */
static PyObject *__pyx_array___pyx_pf_15View_dot_MemoryView_5array_10__getitem__(struct __pyx_array_obj *__pyx_v_self, PyObject *__pyx_v_item); /* proto */
static int __pyx_array___pyx_pf_15View_dot_MemoryView_5array_12__setitem__(struct __pyx_array_obj *__pyx_v_self, PyObject *__pyx_v_item, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf___pyx_array___reduce_cython__(CYTHON_UNUSED struct __pyx_array_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf___pyx_array_2__setstate_cython__(CYTHON_UNUSED struct __pyx_array_obj *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_MemviewEnum___pyx_pf_15View_dot_MemoryView_4Enum___init__(struct __pyx_MemviewEnum_obj *__pyx_v_self, PyObject *__pyx_v_name); /* proto */
static PyObject *__pyx_MemviewEnum___pyx_pf_15View_dot_MemoryView_4Enum_2__repr__(struct __pyx_MemviewEnum_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf___pyx_MemviewEnum___reduce_cython__(struct __pyx_MemviewEnum_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf___pyx_MemviewEnum_2__setstate_cython__(struct __pyx_MemviewEnum_obj *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview___cinit__(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_obj, int __pyx_v_flags, int __pyx_v_dtype_is_object); /* proto */
static void __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_2__dealloc__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_4__getitem__(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_index); /* proto */
static int __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_6__setitem__(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_index, PyObject *__pyx_v_value); /* proto */
static int __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_8__getbuffer__(struct __pyx_memoryview_obj *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_1T___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_4base___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_5shape___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_7strides___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_10suboffsets___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_4ndim___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_8itemsize___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_6nbytes___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_4size___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static Py_ssize_t __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_10__len__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_12__repr__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_14__str__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_16is_c_contig(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_18is_f_contig(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_20copy(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_22copy_fortran(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf___pyx_memoryview___reduce_cython__(CYTHON_UNUSED struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf___pyx_memoryview_2__setstate_cython__(CYTHON_UNUSED struct __pyx_memoryview_obj *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static void __pyx_memoryviewslice___pyx_pf_15View_dot_MemoryView_16_memoryviewslice___dealloc__(struct __pyx_memoryviewslice_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_16_memoryviewslice_4base___get__(struct __pyx_memoryviewslice_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf___pyx_memoryviewslice___reduce_cython__(CYTHON_UNUSED struct __pyx_memoryviewslice_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf___pyx_memoryviewslice_2__setstate_cython__(CYTHON_UNUSED struct __pyx_memoryviewslice_obj *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView___pyx_unpickle_Enum(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_tp_new_6pandas_5_libs_8interval_IntervalMixin(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6pandas_5_libs_8interval_Interval(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6pandas_5_libs_8interval_IntervalTree(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_array(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_Enum(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_memoryview(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new__memoryviewslice(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_float_0_5;
static PyObject *__pyx_int_0;
static PyObject *__pyx_int_1;
static PyObject *__pyx_int_2;
static PyObject *__pyx_int_3;
static PyObject *__pyx_int_100;
static PyObject *__pyx_int_5571323;
static PyObject *__pyx_int_127993061;
static PyObject *__pyx_int_184977713;
static PyObject *__pyx_int_222419149;
static PyObject *__pyx_int_neg_1;
static PyObject *__pyx_tuple_;
static PyObject *__pyx_tuple__2;
static PyObject *__pyx_slice__11;
static PyObject *__pyx_slice__12;
static PyObject *__pyx_slice__41;
static PyObject *__pyx_tuple__15;
static PyObject *__pyx_tuple__16;
static PyObject *__pyx_tuple__17;
static PyObject *__pyx_tuple__19;
static PyObject *__pyx_tuple__20;
static PyObject *__pyx_tuple__21;
static PyObject *__pyx_tuple__22;
static PyObject *__pyx_tuple__23;
static PyObject *__pyx_tuple__24;
static PyObject *__pyx_tuple__25;
static PyObject *__pyx_tuple__26;
static PyObject *__pyx_tuple__27;
static PyObject *__pyx_tuple__28;
static PyObject *__pyx_tuple__29;
static PyObject *__pyx_tuple__30;
static PyObject *__pyx_tuple__31;
static PyObject *__pyx_tuple__32;
static PyObject *__pyx_tuple__33;
static PyObject *__pyx_tuple__34;
static PyObject *__pyx_tuple__35;
static PyObject *__pyx_tuple__36;
static PyObject *__pyx_tuple__37;
static PyObject *__pyx_tuple__38;
static PyObject *__pyx_tuple__39;
static PyObject *__pyx_tuple__40;
static PyObject *__pyx_tuple__42;
static PyObject *__pyx_tuple__43;
static PyObject *__pyx_tuple__44;
static PyObject *__pyx_tuple__45;
static PyObject *__pyx_tuple__46;
static PyObject *__pyx_tuple__48;
static PyObject *__pyx_tuple__50;
static PyObject *__pyx_tuple__52;
static PyObject *__pyx_tuple__54;
static PyObject *__pyx_tuple__56;
static PyObject *__pyx_tuple__57;
static PyObject *__pyx_tuple__59;
static PyObject *__pyx_tuple__60;
static PyObject *__pyx_tuple__62;
static PyObject *__pyx_tuple__63;
static PyObject *__pyx_tuple__65;
static PyObject *__pyx_tuple__66;
static PyObject *__pyx_tuple__68;
static PyObject *__pyx_tuple__69;
static PyObject *__pyx_tuple__71;
static PyObject *__pyx_tuple__72;
static PyObject *__pyx_tuple__74;
static PyObject *__pyx_tuple__75;
static PyObject *__pyx_tuple__77;
static PyObject *__pyx_tuple__78;
static PyObject *__pyx_tuple__80;
static PyObject *__pyx_tuple__81;
static PyObject *__pyx_tuple__83;
static PyObject *__pyx_tuple__84;
static PyObject *__pyx_tuple__86;
static PyObject *__pyx_tuple__87;
static PyObject *__pyx_tuple__89;
static PyObject *__pyx_tuple__90;
static PyObject *__pyx_tuple__92;
static PyObject *__pyx_tuple__93;
static PyObject *__pyx_tuple__95;
static PyObject *__pyx_tuple__96;
static PyObject *__pyx_tuple__98;
static PyObject *__pyx_tuple__99;
static PyObject *__pyx_tuple__101;
static PyObject *__pyx_tuple__102;
static PyObject *__pyx_tuple__104;
static PyObject *__pyx_tuple__105;
static PyObject *__pyx_tuple__107;
static PyObject *__pyx_tuple__108;
static PyObject *__pyx_tuple__110;
static PyObject *__pyx_tuple__111;
static PyObject *__pyx_tuple__113;
static PyObject *__pyx_tuple__114;
static PyObject *__pyx_tuple__116;
static PyObject *__pyx_tuple__118;
static PyObject *__pyx_tuple__120;
static PyObject *__pyx_tuple__122;
static PyObject *__pyx_tuple__124;
static PyObject *__pyx_tuple__126;
static PyObject *__pyx_tuple__128;
static PyObject *__pyx_tuple__130;
static PyObject *__pyx_tuple__132;
static PyObject *__pyx_tuple__134;
static PyObject *__pyx_tuple__136;
static PyObject *__pyx_tuple__138;
static PyObject *__pyx_tuple__140;
static PyObject *__pyx_tuple__142;
static PyObject *__pyx_tuple__144;
static PyObject *__pyx_tuple__146;
static PyObject *__pyx_tuple__148;
static PyObject *__pyx_tuple__150;
static PyObject *__pyx_tuple__152;
static PyObject *__pyx_tuple__154;
static PyObject *__pyx_tuple__156;
static PyObject *__pyx_tuple__158;
static PyObject *__pyx_tuple__159;
static PyObject *__pyx_tuple__160;
static PyObject *__pyx_tuple__161;
static PyObject *__pyx_tuple__162;
static PyObject *__pyx_tuple__163;
static PyObject *__pyx_codeobj__47;
static PyObject *__pyx_codeobj__49;
static PyObject *__pyx_codeobj__51;
static PyObject *__pyx_codeobj__53;
static PyObject *__pyx_codeobj__55;
static PyObject *__pyx_codeobj__58;
static PyObject *__pyx_codeobj__61;
static PyObject *__pyx_codeobj__64;
static PyObject *__pyx_codeobj__67;
static PyObject *__pyx_codeobj__70;
static PyObject *__pyx_codeobj__73;
static PyObject *__pyx_codeobj__76;
static PyObject *__pyx_codeobj__79;
static PyObject *__pyx_codeobj__82;
static PyObject *__pyx_codeobj__85;
static PyObject *__pyx_codeobj__88;
static PyObject *__pyx_codeobj__91;
static PyObject *__pyx_codeobj__94;
static PyObject *__pyx_codeobj__97;
static PyObject *__pyx_codeobj__100;
static PyObject *__pyx_codeobj__103;
static PyObject *__pyx_codeobj__106;
static PyObject *__pyx_codeobj__109;
static PyObject *__pyx_codeobj__112;
static PyObject *__pyx_codeobj__115;
static PyObject *__pyx_codeobj__117;
static PyObject *__pyx_codeobj__119;
static PyObject *__pyx_codeobj__121;
static PyObject *__pyx_codeobj__123;
static PyObject *__pyx_codeobj__125;
static PyObject *__pyx_codeobj__127;
static PyObject *__pyx_codeobj__129;
static PyObject *__pyx_codeobj__131;
static PyObject *__pyx_codeobj__133;
static PyObject *__pyx_codeobj__135;
static PyObject *__pyx_codeobj__137;
static PyObject *__pyx_codeobj__139;
static PyObject *__pyx_codeobj__141;
static PyObject *__pyx_codeobj__143;
static PyObject *__pyx_codeobj__145;
static PyObject *__pyx_codeobj__147;
static PyObject *__pyx_codeobj__149;
static PyObject *__pyx_codeobj__151;
static PyObject *__pyx_codeobj__153;
static PyObject *__pyx_codeobj__155;
static PyObject *__pyx_codeobj__157;
static PyObject *__pyx_codeobj__164;
/* Late includes */

/* "pandas/_libs/interval.pyx":35
 * 
 *     @property
 *     def closed_left(self):             # <<<<<<<<<<<<<<
 *         """
 *         Check if the interval is closed on the left side.
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_13IntervalMixin_11closed_left_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_13IntervalMixin_11closed_left_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_13IntervalMixin_11closed_left___get__(((struct __pyx_obj_6pandas_5_libs_8interval_IntervalMixin *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_13IntervalMixin_11closed_left___get__(struct __pyx_obj_6pandas_5_libs_8interval_IntervalMixin *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pandas/_libs/interval.pyx":47
 *             ``False``.
 *         """
 *         return self.closed in ('left', 'both')             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_closed); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 47, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = (__Pyx_PyUnicode_Equals(__pyx_t_1, __pyx_n_u_left, Py_EQ)); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 47, __pyx_L1_error)
  if (!__pyx_t_3) {
  } else {
    __pyx_t_2 = __pyx_t_3;
    goto __pyx_L3_bool_binop_done;
  }
  __pyx_t_3 = (__Pyx_PyUnicode_Equals(__pyx_t_1, __pyx_n_u_both, Py_EQ)); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 47, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_3;
  __pyx_L3_bool_binop_done:;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 47, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pandas/_libs/interval.pyx":35
 * 
 *     @property
 *     def closed_left(self):             # <<<<<<<<<<<<<<
 *         """
 *         Check if the interval is closed on the left side.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.IntervalMixin.closed_left.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/interval.pyx":50
 * 
 *     @property
 *     def closed_right(self):             # <<<<<<<<<<<<<<
 *         """
 *         Check if the interval is closed on the right side.
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_13IntervalMixin_12closed_right_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_13IntervalMixin_12closed_right_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_13IntervalMixin_12closed_right___get__(((struct __pyx_obj_6pandas_5_libs_8interval_IntervalMixin *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_13IntervalMixin_12closed_right___get__(struct __pyx_obj_6pandas_5_libs_8interval_IntervalMixin *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pandas/_libs/interval.pyx":62
 *             ``False``.
 *         """
 *         return self.closed in ('right', 'both')             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_closed); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 62, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = (__Pyx_PyUnicode_Equals(__pyx_t_1, __pyx_n_u_right, Py_EQ)); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 62, __pyx_L1_error)
  if (!__pyx_t_3) {
  } else {
    __pyx_t_2 = __pyx_t_3;
    goto __pyx_L3_bool_binop_done;
  }
  __pyx_t_3 = (__Pyx_PyUnicode_Equals(__pyx_t_1, __pyx_n_u_both, Py_EQ)); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 62, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_3;
  __pyx_L3_bool_binop_done:;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 62, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pandas/_libs/interval.pyx":50
 * 
 *     @property
 *     def closed_right(self):             # <<<<<<<<<<<<<<
 *         """
 *         Check if the interval is closed on the right side.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.IntervalMixin.closed_right.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/interval.pyx":65
 * 
 *     @property
 *     def open_left(self):             # <<<<<<<<<<<<<<
 *         """
 *         Check if the interval is open on the left side.
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_13IntervalMixin_9open_left_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_13IntervalMixin_9open_left_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_13IntervalMixin_9open_left___get__(((struct __pyx_obj_6pandas_5_libs_8interval_IntervalMixin *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_13IntervalMixin_9open_left___get__(struct __pyx_obj_6pandas_5_libs_8interval_IntervalMixin *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pandas/_libs/interval.pyx":77
 *             ``False``.
 *         """
 *         return not self.closed_left             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_closed_left); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 77, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 77, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyBool_FromLong((!__pyx_t_2)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 77, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pandas/_libs/interval.pyx":65
 * 
 *     @property
 *     def open_left(self):             # <<<<<<<<<<<<<<
 *         """
 *         Check if the interval is open on the left side.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.IntervalMixin.open_left.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/interval.pyx":80
 * 
 *     @property
 *     def open_right(self):             # <<<<<<<<<<<<<<
 *         """
 *         Check if the interval is open on the right side.
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_13IntervalMixin_10open_right_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_13IntervalMixin_10open_right_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_13IntervalMixin_10open_right___get__(((struct __pyx_obj_6pandas_5_libs_8interval_IntervalMixin *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_13IntervalMixin_10open_right___get__(struct __pyx_obj_6pandas_5_libs_8interval_IntervalMixin *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pandas/_libs/interval.pyx":92
 *             ``False``.
 *         """
 *         return not self.closed_right             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_closed_right); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 92, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 92, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyBool_FromLong((!__pyx_t_2)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 92, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pandas/_libs/interval.pyx":80
 * 
 *     @property
 *     def open_right(self):             # <<<<<<<<<<<<<<
 *         """
 *         Check if the interval is open on the right side.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.IntervalMixin.open_right.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/interval.pyx":95
 * 
 *     @property
 *     def mid(self):             # <<<<<<<<<<<<<<
 *         """
 *         Return the midpoint of the Interval
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_13IntervalMixin_3mid_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_13IntervalMixin_3mid_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_13IntervalMixin_3mid___get__(((struct __pyx_obj_6pandas_5_libs_8interval_IntervalMixin *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_13IntervalMixin_3mid___get__(struct __pyx_obj_6pandas_5_libs_8interval_IntervalMixin *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pandas/_libs/interval.pyx":99
 *         Return the midpoint of the Interval
 *         """
 *         try:             # <<<<<<<<<<<<<<
 *             return 0.5 * (self.left + self.right)
 *         except TypeError:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "pandas/_libs/interval.pyx":100
 *         """
 *         try:
 *             return 0.5 * (self.left + self.right)             # <<<<<<<<<<<<<<
 *         except TypeError:
 *             # datetime safe version
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_left); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 100, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_right); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 100, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = PyNumber_Add(__pyx_t_4, __pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 100, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = PyNumber_Multiply(__pyx_float_0_5, __pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 100, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_r = __pyx_t_5;
      __pyx_t_5 = 0;
      goto __pyx_L7_try_return;

      /* "pandas/_libs/interval.pyx":99
 *         Return the midpoint of the Interval
 *         """
 *         try:             # <<<<<<<<<<<<<<
 *             return 0.5 * (self.left + self.right)
 *         except TypeError:
 */
    }
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "pandas/_libs/interval.pyx":101
 *         try:
 *             return 0.5 * (self.left + self.right)
 *         except TypeError:             # <<<<<<<<<<<<<<
 *             # datetime safe version
 *             return self.left + 0.5 * self.length
 */
    __pyx_t_7 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_TypeError);
    if (__pyx_t_7) {
      __Pyx_AddTraceback("pandas._libs.interval.IntervalMixin.mid.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_4) < 0) __PYX_ERR(0, 101, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_4);

      /* "pandas/_libs/interval.pyx":103
 *         except TypeError:
 *             # datetime safe version
 *             return self.left + 0.5 * self.length             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_8 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_left); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 103, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_length); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 103, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_10 = PyNumber_Multiply(__pyx_float_0_5, __pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 103, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_t_9 = PyNumber_Add(__pyx_t_8, __pyx_t_10); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 103, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __pyx_r = __pyx_t_9;
      __pyx_t_9 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      goto __pyx_L6_except_return;
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "pandas/_libs/interval.pyx":99
 *         Return the midpoint of the Interval
 *         """
 *         try:             # <<<<<<<<<<<<<<
 *             return 0.5 * (self.left + self.right)
 *         except TypeError:
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L7_try_return:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L0;
    __pyx_L6_except_return:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L0;
  }

  /* "pandas/_libs/interval.pyx":95
 * 
 *     @property
 *     def mid(self):             # <<<<<<<<<<<<<<
 *         """
 *         Return the midpoint of the Interval
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("pandas._libs.interval.IntervalMixin.mid.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/interval.pyx":106
 * 
 *     @property
 *     def length(self):             # <<<<<<<<<<<<<<
 *         """Return the length of the Interval"""
 *         return self.right - self.left
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_13IntervalMixin_6length_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_13IntervalMixin_6length_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_13IntervalMixin_6length___get__(((struct __pyx_obj_6pandas_5_libs_8interval_IntervalMixin *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_13IntervalMixin_6length___get__(struct __pyx_obj_6pandas_5_libs_8interval_IntervalMixin *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pandas/_libs/interval.pyx":108
 *     def length(self):
 *         """Return the length of the Interval"""
 *         return self.right - self.left             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_right); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 108, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_left); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 108, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyNumber_Subtract(__pyx_t_1, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 108, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "pandas/_libs/interval.pyx":106
 * 
 *     @property
 *     def length(self):             # <<<<<<<<<<<<<<
 *         """Return the length of the Interval"""
 *         return self.right - self.left
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pandas._libs.interval.IntervalMixin.length.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/interval.pyx":111
 * 
 *     @property
 *     def is_empty(self):             # <<<<<<<<<<<<<<
 *         """
 *         Indicates if an interval is empty, meaning it contains no points.
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_13IntervalMixin_8is_empty_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_13IntervalMixin_8is_empty_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_13IntervalMixin_8is_empty___get__(((struct __pyx_obj_6pandas_5_libs_8interval_IntervalMixin *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_13IntervalMixin_8is_empty___get__(struct __pyx_obj_6pandas_5_libs_8interval_IntervalMixin *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pandas/_libs/interval.pyx":161
 *         array([ True, False])
 *         """
 *         return (self.right == self.left) & (self.closed != 'both')             # <<<<<<<<<<<<<<
 * 
 *     def _check_closed_matches(self, other, name='other'):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_right); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 161, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_left); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 161, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyObject_RichCompare(__pyx_t_1, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 161, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_closed); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 161, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = PyObject_RichCompare(__pyx_t_2, __pyx_n_u_both, Py_NE); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 161, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyNumber_And(__pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 161, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pandas/_libs/interval.pyx":111
 * 
 *     @property
 *     def is_empty(self):             # <<<<<<<<<<<<<<
 *         """
 *         Indicates if an interval is empty, meaning it contains no points.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pandas._libs.interval.IntervalMixin.is_empty.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/interval.pyx":163
 *         return (self.right == self.left) & (self.closed != 'both')
 * 
 *     def _check_closed_matches(self, other, name='other'):             # <<<<<<<<<<<<<<
 *         """Check if the closed attribute of `other` matches.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_13IntervalMixin_1_check_closed_matches(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6pandas_5_libs_8interval_13IntervalMixin__check_closed_matches[] = "Check if the closed attribute of `other` matches.\n\n        Note that 'left' and 'right' are considered different from 'both'.\n\n        Parameters\n        ----------\n        other : Interval, IntervalIndex, IntervalArray\n        name : str\n            Name to use for 'other' in the error message.\n\n        Raises\n        ------\n        ValueError\n            When `other` is not closed exactly the same as self.\n        ";
static PyObject *__pyx_pw_6pandas_5_libs_8interval_13IntervalMixin_1_check_closed_matches(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_other = 0;
  PyObject *__pyx_v_name = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_check_closed_matches (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_other,&__pyx_n_s_name,0};
    PyObject* values[2] = {0,0};
    values[1] = ((PyObject *)__pyx_n_u_other);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_other)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_name);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_check_closed_matches") < 0)) __PYX_ERR(0, 163, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_other = values[0];
    __pyx_v_name = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_check_closed_matches", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 163, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.IntervalMixin._check_closed_matches", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_13IntervalMixin__check_closed_matches(((struct __pyx_obj_6pandas_5_libs_8interval_IntervalMixin *)__pyx_v_self), __pyx_v_other, __pyx_v_name);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_13IntervalMixin__check_closed_matches(struct __pyx_obj_6pandas_5_libs_8interval_IntervalMixin *__pyx_v_self, PyObject *__pyx_v_other, PyObject *__pyx_v_name) {
  PyObject *__pyx_v_msg = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  __Pyx_RefNannySetupContext("_check_closed_matches", 0);

  /* "pandas/_libs/interval.pyx":179
 *             When `other` is not closed exactly the same as self.
 *         """
 *         if self.closed != other.closed:             # <<<<<<<<<<<<<<
 *             msg = "'{}.closed' is '{}', expected '{}'."
 *             raise ValueError(msg.format(name, other.closed, self.closed))
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_closed); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 179, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_other, __pyx_n_s_closed); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 179, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyObject_RichCompare(__pyx_t_1, __pyx_t_2, Py_NE); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 179, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 179, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(__pyx_t_4)) {

    /* "pandas/_libs/interval.pyx":180
 *         """
 *         if self.closed != other.closed:
 *             msg = "'{}.closed' is '{}', expected '{}'."             # <<<<<<<<<<<<<<
 *             raise ValueError(msg.format(name, other.closed, self.closed))
 * 
 */
    __Pyx_INCREF(__pyx_kp_u_closed_is_expected);
    __pyx_v_msg = __pyx_kp_u_closed_is_expected;

    /* "pandas/_libs/interval.pyx":181
 *         if self.closed != other.closed:
 *             msg = "'{}.closed' is '{}', expected '{}'."
 *             raise ValueError(msg.format(name, other.closed, self.closed))             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_msg, __pyx_n_s_format); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 181, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_other, __pyx_n_s_closed); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 181, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_closed); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 181, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = NULL;
    __pyx_t_7 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
        __pyx_t_7 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_2)) {
      PyObject *__pyx_temp[4] = {__pyx_t_6, __pyx_v_name, __pyx_t_1, __pyx_t_5};
      __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_7, 3+__pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 181, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
      PyObject *__pyx_temp[4] = {__pyx_t_6, __pyx_v_name, __pyx_t_1, __pyx_t_5};
      __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_7, 3+__pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 181, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    } else
    #endif
    {
      __pyx_t_8 = PyTuple_New(3+__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 181, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      if (__pyx_t_6) {
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_6); __pyx_t_6 = NULL;
      }
      __Pyx_INCREF(__pyx_v_name);
      __Pyx_GIVEREF(__pyx_v_name);
      PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_7, __pyx_v_name);
      __Pyx_GIVEREF(__pyx_t_1);
      PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_7, __pyx_t_1);
      __Pyx_GIVEREF(__pyx_t_5);
      PyTuple_SET_ITEM(__pyx_t_8, 2+__pyx_t_7, __pyx_t_5);
      __pyx_t_1 = 0;
      __pyx_t_5 = 0;
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_8, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 181, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 181, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 181, __pyx_L1_error)

    /* "pandas/_libs/interval.pyx":179
 *             When `other` is not closed exactly the same as self.
 *         """
 *         if self.closed != other.closed:             # <<<<<<<<<<<<<<
 *             msg = "'{}.closed' is '{}', expected '{}'."
 *             raise ValueError(msg.format(name, other.closed, self.closed))
 */
  }

  /* "pandas/_libs/interval.pyx":163
 *         return (self.right == self.left) & (self.closed != 'both')
 * 
 *     def _check_closed_matches(self, other, name='other'):             # <<<<<<<<<<<<<<
 *         """Check if the closed attribute of `other` matches.
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("pandas._libs.interval.IntervalMixin._check_closed_matches", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_msg);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_13IntervalMixin_3__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_13IntervalMixin_3__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_13IntervalMixin_2__reduce_cython__(((struct __pyx_obj_6pandas_5_libs_8interval_IntervalMixin *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_13IntervalMixin_2__reduce_cython__(struct __pyx_obj_6pandas_5_libs_8interval_IntervalMixin *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = ()             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __Pyx_INCREF(__pyx_empty_tuple);
  __pyx_v_state = __pyx_empty_tuple;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict);
    __pyx_t_4 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_4));
    __pyx_t_4 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = False
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = False             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_IntervalMixin, (type(self), 0xd41d8cd, None), state
 */
  /*else*/ {
    __pyx_v_use_setstate = 0;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_IntervalMixin, (type(self), 0xd41d8cd, None), state
 *     else:
 */
  __pyx_t_3 = (__pyx_v_use_setstate != 0);
  if (__pyx_t_3) {

    /* "(tree fragment)":13
 *         use_setstate = False
 *     if use_setstate:
 *         return __pyx_unpickle_IntervalMixin, (type(self), 0xd41d8cd, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_IntervalMixin, (type(self), 0xd41d8cd, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_pyx_unpickle_IntervalMixin); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_222419149);
    __Pyx_GIVEREF(__pyx_int_222419149);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_222419149);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state);
    __pyx_t_4 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_IntervalMixin, (type(self), 0xd41d8cd, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_IntervalMixin, (type(self), 0xd41d8cd, None), state
 *     else:
 *         return __pyx_unpickle_IntervalMixin, (type(self), 0xd41d8cd, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_IntervalMixin__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_pyx_unpickle_IntervalMixin); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_222419149);
    __Pyx_GIVEREF(__pyx_int_222419149);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_222419149);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state);
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1);
    __pyx_t_5 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pandas._libs.interval.IntervalMixin.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_IntervalMixin, (type(self), 0xd41d8cd, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_IntervalMixin__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_13IntervalMixin_5__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_13IntervalMixin_5__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_13IntervalMixin_4__setstate_cython__(((struct __pyx_obj_6pandas_5_libs_8interval_IntervalMixin *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_13IntervalMixin_4__setstate_cython__(struct __pyx_obj_6pandas_5_libs_8interval_IntervalMixin *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_IntervalMixin, (type(self), 0xd41d8cd, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_IntervalMixin__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(2, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_6pandas_5_libs_8interval___pyx_unpickle_IntervalMixin__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_IntervalMixin, (type(self), 0xd41d8cd, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_IntervalMixin__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.IntervalMixin.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/interval.pyx":184
 * 
 * 
 * cdef _interval_like(other):             # <<<<<<<<<<<<<<
 *     return (hasattr(other, 'left')
 *             and hasattr(other, 'right')
 */

static PyObject *__pyx_f_6pandas_5_libs_8interval__interval_like(PyObject *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("_interval_like", 0);

  /* "pandas/_libs/interval.pyx":185
 * 
 * cdef _interval_like(other):
 *     return (hasattr(other, 'left')             # <<<<<<<<<<<<<<
 *             and hasattr(other, 'right')
 *             and hasattr(other, 'closed'))
 */
  __Pyx_XDECREF(__pyx_r);

  /* "pandas/_libs/interval.pyx":186
 * cdef _interval_like(other):
 *     return (hasattr(other, 'left')
 *             and hasattr(other, 'right')             # <<<<<<<<<<<<<<
 *             and hasattr(other, 'closed'))
 * 
 */
  __pyx_t_2 = __Pyx_HasAttr(__pyx_v_other, __pyx_n_u_left); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(0, 185, __pyx_L1_error)
  if (__pyx_t_2) {
  } else {

    /* "pandas/_libs/interval.pyx":185
 * 
 * cdef _interval_like(other):
 *     return (hasattr(other, 'left')             # <<<<<<<<<<<<<<
 *             and hasattr(other, 'right')
 *             and hasattr(other, 'closed'))
 */
    __pyx_t_3 = __Pyx_PyBool_FromLong(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 185, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L3_bool_binop_done;
  }

  /* "pandas/_libs/interval.pyx":186
 * cdef _interval_like(other):
 *     return (hasattr(other, 'left')
 *             and hasattr(other, 'right')             # <<<<<<<<<<<<<<
 *             and hasattr(other, 'closed'))
 * 
 */
  __pyx_t_2 = __Pyx_HasAttr(__pyx_v_other, __pyx_n_u_right); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(0, 186, __pyx_L1_error)
  if (__pyx_t_2) {
  } else {
    __pyx_t_3 = __Pyx_PyBool_FromLong(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 186, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L3_bool_binop_done;
  }

  /* "pandas/_libs/interval.pyx":187
 *     return (hasattr(other, 'left')
 *             and hasattr(other, 'right')
 *             and hasattr(other, 'closed'))             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = __Pyx_HasAttr(__pyx_v_other, __pyx_n_u_closed); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(0, 187, __pyx_L1_error)
  __pyx_t_3 = __Pyx_PyBool_FromLong(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 187, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = __pyx_t_3;
  __pyx_t_3 = 0;
  __pyx_L3_bool_binop_done:;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pandas/_libs/interval.pyx":184
 * 
 * 
 * cdef _interval_like(other):             # <<<<<<<<<<<<<<
 *     return (hasattr(other, 'left')
 *             and hasattr(other, 'right')
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pandas._libs.interval._interval_like", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/interval.pyx":299
 *     """
 * 
 *     def __init__(self, left, right, str closed='right'):             # <<<<<<<<<<<<<<
 *         # note: it is faster to just do these checks than to use a special
 *         # constructor (__cinit__/__new__) to avoid them
 */

/* Python wrapper */
static int __pyx_pw_6pandas_5_libs_8interval_8Interval_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6pandas_5_libs_8interval_8Interval_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_left = 0;
  PyObject *__pyx_v_right = 0;
  PyObject *__pyx_v_closed = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_left,&__pyx_n_s_right,&__pyx_n_s_closed,0};
    PyObject* values[3] = {0,0,0};
    values[2] = ((PyObject*)__pyx_n_u_right);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_left)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_right)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 3, 1); __PYX_ERR(0, 299, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_closed);
          if (value) { values[2] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 299, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_left = values[0];
    __pyx_v_right = values[1];
    __pyx_v_closed = ((PyObject*)values[2]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 299, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Interval.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_closed), (&PyUnicode_Type), 1, "closed", 1))) __PYX_ERR(0, 299, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_8Interval___init__(((struct __pyx_obj_6pandas_5_libs_8interval_Interval *)__pyx_v_self), __pyx_v_left, __pyx_v_right, __pyx_v_closed);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6pandas_5_libs_8interval_8Interval___init__(struct __pyx_obj_6pandas_5_libs_8interval_Interval *__pyx_v_self, PyObject *__pyx_v_left, PyObject *__pyx_v_right, PyObject *__pyx_v_closed) {
  PyObject *__pyx_v_msg = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pandas/_libs/interval.pyx":303
 *         # constructor (__cinit__/__new__) to avoid them
 * 
 *         self._validate_endpoint(left)             # <<<<<<<<<<<<<<
 *         self._validate_endpoint(right)
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_validate_endpoint); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 303, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_v_left) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_left);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 303, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pandas/_libs/interval.pyx":304
 * 
 *         self._validate_endpoint(left)
 *         self._validate_endpoint(right)             # <<<<<<<<<<<<<<
 * 
 *         if closed not in _VALID_CLOSED:
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_validate_endpoint); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 304, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_v_right) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_right);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 304, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pandas/_libs/interval.pyx":306
 *         self._validate_endpoint(right)
 * 
 *         if closed not in _VALID_CLOSED:             # <<<<<<<<<<<<<<
 *             msg = "invalid option for 'closed': {closed}".format(closed=closed)
 *             raise ValueError(msg)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_VALID_CLOSED); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 306, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = (__Pyx_PySequence_ContainsTF(__pyx_v_closed, __pyx_t_1, Py_NE)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 306, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_5 = (__pyx_t_4 != 0);
  if (unlikely(__pyx_t_5)) {

    /* "pandas/_libs/interval.pyx":307
 * 
 *         if closed not in _VALID_CLOSED:
 *             msg = "invalid option for 'closed': {closed}".format(closed=closed)             # <<<<<<<<<<<<<<
 *             raise ValueError(msg)
 *         if not left <= right:
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_invalid_option_for_closed_closed, __pyx_n_s_format); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 307, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 307, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_closed, __pyx_v_closed) < 0) __PYX_ERR(0, 307, __pyx_L1_error)
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_empty_tuple, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 307, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_msg = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "pandas/_libs/interval.pyx":308
 *         if closed not in _VALID_CLOSED:
 *             msg = "invalid option for 'closed': {closed}".format(closed=closed)
 *             raise ValueError(msg)             # <<<<<<<<<<<<<<
 *         if not left <= right:
 *             raise ValueError('left side of interval must be <= right side')
 */
    __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_v_msg); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 308, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 308, __pyx_L1_error)

    /* "pandas/_libs/interval.pyx":306
 *         self._validate_endpoint(right)
 * 
 *         if closed not in _VALID_CLOSED:             # <<<<<<<<<<<<<<
 *             msg = "invalid option for 'closed': {closed}".format(closed=closed)
 *             raise ValueError(msg)
 */
  }

  /* "pandas/_libs/interval.pyx":309
 *             msg = "invalid option for 'closed': {closed}".format(closed=closed)
 *             raise ValueError(msg)
 *         if not left <= right:             # <<<<<<<<<<<<<<
 *             raise ValueError('left side of interval must be <= right side')
 *         if (isinstance(left, Timestamp) and
 */
  __pyx_t_3 = PyObject_RichCompare(__pyx_v_left, __pyx_v_right, Py_LE); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 309, __pyx_L1_error)
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 309, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_4 = ((!__pyx_t_5) != 0);
  if (unlikely(__pyx_t_4)) {

    /* "pandas/_libs/interval.pyx":310
 *             raise ValueError(msg)
 *         if not left <= right:
 *             raise ValueError('left side of interval must be <= right side')             # <<<<<<<<<<<<<<
 *         if (isinstance(left, Timestamp) and
 *                 not tz_compare(left.tzinfo, right.tzinfo)):
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple_, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 310, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 310, __pyx_L1_error)

    /* "pandas/_libs/interval.pyx":309
 *             msg = "invalid option for 'closed': {closed}".format(closed=closed)
 *             raise ValueError(msg)
 *         if not left <= right:             # <<<<<<<<<<<<<<
 *             raise ValueError('left side of interval must be <= right side')
 *         if (isinstance(left, Timestamp) and
 */
  }

  /* "pandas/_libs/interval.pyx":311
 *         if not left <= right:
 *             raise ValueError('left side of interval must be <= right side')
 *         if (isinstance(left, Timestamp) and             # <<<<<<<<<<<<<<
 *                 not tz_compare(left.tzinfo, right.tzinfo)):
 *             # GH 18538
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_Timestamp); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 311, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = PyObject_IsInstance(__pyx_v_left, __pyx_t_3); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 311, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_6 = (__pyx_t_5 != 0);
  if (__pyx_t_6) {
  } else {
    __pyx_t_4 = __pyx_t_6;
    goto __pyx_L6_bool_binop_done;
  }

  /* "pandas/_libs/interval.pyx":312
 *             raise ValueError('left side of interval must be <= right side')
 *         if (isinstance(left, Timestamp) and
 *                 not tz_compare(left.tzinfo, right.tzinfo)):             # <<<<<<<<<<<<<<
 *             # GH 18538
 *             msg = ("left and right must have the same time zone, got "
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_left, __pyx_n_s_tzinfo); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 312, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_right, __pyx_n_s_tzinfo); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 312, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_6 = ((!(__pyx_f_6pandas_5_libs_6tslibs_9timezones_tz_compare(__pyx_t_3, __pyx_t_2, 0) != 0)) != 0);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = __pyx_t_6;
  __pyx_L6_bool_binop_done:;

  /* "pandas/_libs/interval.pyx":311
 *         if not left <= right:
 *             raise ValueError('left side of interval must be <= right side')
 *         if (isinstance(left, Timestamp) and             # <<<<<<<<<<<<<<
 *                 not tz_compare(left.tzinfo, right.tzinfo)):
 *             # GH 18538
 */
  if (unlikely(__pyx_t_4)) {

    /* "pandas/_libs/interval.pyx":314
 *                 not tz_compare(left.tzinfo, right.tzinfo)):
 *             # GH 18538
 *             msg = ("left and right must have the same time zone, got "             # <<<<<<<<<<<<<<
 *                    "'{left_tz}' and '{right_tz}'")
 *             raise ValueError(msg.format(left_tz=left.tzinfo,
 */
    __Pyx_INCREF(__pyx_kp_u_left_and_right_must_have_the_sam);
    __pyx_v_msg = __pyx_kp_u_left_and_right_must_have_the_sam;

    /* "pandas/_libs/interval.pyx":316
 *             msg = ("left and right must have the same time zone, got "
 *                    "'{left_tz}' and '{right_tz}'")
 *             raise ValueError(msg.format(left_tz=left.tzinfo,             # <<<<<<<<<<<<<<
 *                                         right_tz=right.tzinfo))
 *         self.left = left
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_msg, __pyx_n_s_format); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 316, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 316, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_left, __pyx_n_s_tzinfo); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 316, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_left_tz, __pyx_t_1) < 0) __PYX_ERR(0, 316, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "pandas/_libs/interval.pyx":317
 *                    "'{left_tz}' and '{right_tz}'")
 *             raise ValueError(msg.format(left_tz=left.tzinfo,
 *                                         right_tz=right.tzinfo))             # <<<<<<<<<<<<<<
 *         self.left = left
 *         self.right = right
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_right, __pyx_n_s_tzinfo); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 317, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_right_tz, __pyx_t_1) < 0) __PYX_ERR(0, 316, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "pandas/_libs/interval.pyx":316
 *             msg = ("left and right must have the same time zone, got "
 *                    "'{left_tz}' and '{right_tz}'")
 *             raise ValueError(msg.format(left_tz=left.tzinfo,             # <<<<<<<<<<<<<<
 *                                         right_tz=right.tzinfo))
 *         self.left = left
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_empty_tuple, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 316, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 316, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 316, __pyx_L1_error)

    /* "pandas/_libs/interval.pyx":311
 *         if not left <= right:
 *             raise ValueError('left side of interval must be <= right side')
 *         if (isinstance(left, Timestamp) and             # <<<<<<<<<<<<<<
 *                 not tz_compare(left.tzinfo, right.tzinfo)):
 *             # GH 18538
 */
  }

  /* "pandas/_libs/interval.pyx":318
 *             raise ValueError(msg.format(left_tz=left.tzinfo,
 *                                         right_tz=right.tzinfo))
 *         self.left = left             # <<<<<<<<<<<<<<
 *         self.right = right
 *         self.closed = closed
 */
  __Pyx_INCREF(__pyx_v_left);
  __Pyx_GIVEREF(__pyx_v_left);
  __Pyx_GOTREF(__pyx_v_self->left);
  __Pyx_DECREF(__pyx_v_self->left);
  __pyx_v_self->left = __pyx_v_left;

  /* "pandas/_libs/interval.pyx":319
 *                                         right_tz=right.tzinfo))
 *         self.left = left
 *         self.right = right             # <<<<<<<<<<<<<<
 *         self.closed = closed
 * 
 */
  __Pyx_INCREF(__pyx_v_right);
  __Pyx_GIVEREF(__pyx_v_right);
  __Pyx_GOTREF(__pyx_v_self->right);
  __Pyx_DECREF(__pyx_v_self->right);
  __pyx_v_self->right = __pyx_v_right;

  /* "pandas/_libs/interval.pyx":320
 *         self.left = left
 *         self.right = right
 *         self.closed = closed             # <<<<<<<<<<<<<<
 * 
 *     def _validate_endpoint(self, endpoint):
 */
  __Pyx_INCREF(__pyx_v_closed);
  __Pyx_GIVEREF(__pyx_v_closed);
  __Pyx_GOTREF(__pyx_v_self->closed);
  __Pyx_DECREF(__pyx_v_self->closed);
  __pyx_v_self->closed = __pyx_v_closed;

  /* "pandas/_libs/interval.pyx":299
 *     """
 * 
 *     def __init__(self, left, right, str closed='right'):             # <<<<<<<<<<<<<<
 *         # note: it is faster to just do these checks than to use a special
 *         # constructor (__cinit__/__new__) to avoid them
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pandas._libs.interval.Interval.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_msg);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/interval.pyx":322
 *         self.closed = closed
 * 
 *     def _validate_endpoint(self, endpoint):             # <<<<<<<<<<<<<<
 *         # GH 23013
 *         if not (is_integer_object(endpoint) or is_float_object(endpoint) or
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_8Interval_3_validate_endpoint(PyObject *__pyx_v_self, PyObject *__pyx_v_endpoint); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_8Interval_3_validate_endpoint(PyObject *__pyx_v_self, PyObject *__pyx_v_endpoint) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_validate_endpoint (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_8Interval_2_validate_endpoint(((struct __pyx_obj_6pandas_5_libs_8interval_Interval *)__pyx_v_self), ((PyObject *)__pyx_v_endpoint));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_8Interval_2_validate_endpoint(CYTHON_UNUSED struct __pyx_obj_6pandas_5_libs_8interval_Interval *__pyx_v_self, PyObject *__pyx_v_endpoint) {
  PyObject *__pyx_v_msg = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  __Pyx_RefNannySetupContext("_validate_endpoint", 0);

  /* "pandas/_libs/interval.pyx":324
 *     def _validate_endpoint(self, endpoint):
 *         # GH 23013
 *         if not (is_integer_object(endpoint) or is_float_object(endpoint) or             # <<<<<<<<<<<<<<
 *                 isinstance(endpoint, (Timestamp, Timedelta))):
 *             msg = ("Only numeric, Timestamp and Timedelta endpoints "
 */
  __pyx_t_2 = (__pyx_f_6pandas_5_libs_6tslibs_4util_is_integer_object(__pyx_v_endpoint) != 0);
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = (__pyx_f_6pandas_5_libs_6tslibs_4util_is_float_object(__pyx_v_endpoint) != 0);
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }

  /* "pandas/_libs/interval.pyx":325
 *         # GH 23013
 *         if not (is_integer_object(endpoint) or is_float_object(endpoint) or
 *                 isinstance(endpoint, (Timestamp, Timedelta))):             # <<<<<<<<<<<<<<
 *             msg = ("Only numeric, Timestamp and Timedelta endpoints "
 *                    "are allowed when constructing an Interval.")
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_Timestamp); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 325, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_Timedelta); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 325, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = PyObject_IsInstance(__pyx_v_endpoint, __pyx_t_3); 
  __pyx_t_6 = (__pyx_t_5 != 0);
  if (!__pyx_t_6) {
  } else {
    __pyx_t_2 = __pyx_t_6;
    goto __pyx_L7_bool_binop_done;
  }
  __pyx_t_6 = PyObject_IsInstance(__pyx_v_endpoint, __pyx_t_4); 
  __pyx_t_5 = (__pyx_t_6 != 0);
  __pyx_t_2 = __pyx_t_5;
  __pyx_L7_bool_binop_done:;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_5 = (__pyx_t_2 != 0);
  __pyx_t_1 = __pyx_t_5;
  __pyx_L4_bool_binop_done:;

  /* "pandas/_libs/interval.pyx":324
 *     def _validate_endpoint(self, endpoint):
 *         # GH 23013
 *         if not (is_integer_object(endpoint) or is_float_object(endpoint) or             # <<<<<<<<<<<<<<
 *                 isinstance(endpoint, (Timestamp, Timedelta))):
 *             msg = ("Only numeric, Timestamp and Timedelta endpoints "
 */
  __pyx_t_5 = ((!__pyx_t_1) != 0);
  if (unlikely(__pyx_t_5)) {

    /* "pandas/_libs/interval.pyx":326
 *         if not (is_integer_object(endpoint) or is_float_object(endpoint) or
 *                 isinstance(endpoint, (Timestamp, Timedelta))):
 *             msg = ("Only numeric, Timestamp and Timedelta endpoints "             # <<<<<<<<<<<<<<
 *                    "are allowed when constructing an Interval.")
 *             raise ValueError(msg)
 */
    __Pyx_INCREF(__pyx_kp_u_Only_numeric_Timestamp_and_Timed);
    __pyx_v_msg = __pyx_kp_u_Only_numeric_Timestamp_and_Timed;

    /* "pandas/_libs/interval.pyx":328
 *             msg = ("Only numeric, Timestamp and Timedelta endpoints "
 *                    "are allowed when constructing an Interval.")
 *             raise ValueError(msg)             # <<<<<<<<<<<<<<
 * 
 *     def __hash__(self):
 */
    __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_v_msg); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 328, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 328, __pyx_L1_error)

    /* "pandas/_libs/interval.pyx":324
 *     def _validate_endpoint(self, endpoint):
 *         # GH 23013
 *         if not (is_integer_object(endpoint) or is_float_object(endpoint) or             # <<<<<<<<<<<<<<
 *                 isinstance(endpoint, (Timestamp, Timedelta))):
 *             msg = ("Only numeric, Timestamp and Timedelta endpoints "
 */
  }

  /* "pandas/_libs/interval.pyx":322
 *         self.closed = closed
 * 
 *     def _validate_endpoint(self, endpoint):             # <<<<<<<<<<<<<<
 *         # GH 23013
 *         if not (is_integer_object(endpoint) or is_float_object(endpoint) or
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pandas._libs.interval.Interval._validate_endpoint", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_msg);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/interval.pyx":330
 *             raise ValueError(msg)
 * 
 *     def __hash__(self):             # <<<<<<<<<<<<<<
 *         return hash((self.left, self.right, self.closed))
 * 
 */

/* Python wrapper */
static Py_hash_t __pyx_pw_6pandas_5_libs_8interval_8Interval_5__hash__(PyObject *__pyx_v_self); /*proto*/
static Py_hash_t __pyx_pw_6pandas_5_libs_8interval_8Interval_5__hash__(PyObject *__pyx_v_self) {
  Py_hash_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__hash__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_8Interval_4__hash__(((struct __pyx_obj_6pandas_5_libs_8interval_Interval *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static Py_hash_t __pyx_pf_6pandas_5_libs_8interval_8Interval_4__hash__(struct __pyx_obj_6pandas_5_libs_8interval_Interval *__pyx_v_self) {
  Py_hash_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_hash_t __pyx_t_2;
  __Pyx_RefNannySetupContext("__hash__", 0);

  /* "pandas/_libs/interval.pyx":331
 * 
 *     def __hash__(self):
 *         return hash((self.left, self.right, self.closed))             # <<<<<<<<<<<<<<
 * 
 *     def __contains__(self, key):
 */
  __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 331, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_self->left);
  __Pyx_GIVEREF(__pyx_v_self->left);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_self->left);
  __Pyx_INCREF(__pyx_v_self->right);
  __Pyx_GIVEREF(__pyx_v_self->right);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_self->right);
  __Pyx_INCREF(__pyx_v_self->closed);
  __Pyx_GIVEREF(__pyx_v_self->closed);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_self->closed);
  __pyx_t_2 = PyObject_Hash(__pyx_t_1); if (unlikely(__pyx_t_2 == ((Py_hash_t)-1))) __PYX_ERR(0, 331, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  goto __pyx_L0;

  /* "pandas/_libs/interval.pyx":330
 *             raise ValueError(msg)
 * 
 *     def __hash__(self):             # <<<<<<<<<<<<<<
 *         return hash((self.left, self.right, self.closed))
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Interval.__hash__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  if (unlikely(__pyx_r == -1) && !PyErr_Occurred()) __pyx_r = -2;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/interval.pyx":333
 *         return hash((self.left, self.right, self.closed))
 * 
 *     def __contains__(self, key):             # <<<<<<<<<<<<<<
 *         if _interval_like(key):
 *             raise TypeError('__contains__ not defined for two intervals')
 */

/* Python wrapper */
static int __pyx_pw_6pandas_5_libs_8interval_8Interval_7__contains__(PyObject *__pyx_v_self, PyObject *__pyx_v_key); /*proto*/
static int __pyx_pw_6pandas_5_libs_8interval_8Interval_7__contains__(PyObject *__pyx_v_self, PyObject *__pyx_v_key) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__contains__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_8Interval_6__contains__(((struct __pyx_obj_6pandas_5_libs_8interval_Interval *)__pyx_v_self), ((PyObject *)__pyx_v_key));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6pandas_5_libs_8interval_8Interval_6__contains__(struct __pyx_obj_6pandas_5_libs_8interval_Interval *__pyx_v_self, PyObject *__pyx_v_key) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  __Pyx_RefNannySetupContext("__contains__", 0);

  /* "pandas/_libs/interval.pyx":334
 * 
 *     def __contains__(self, key):
 *         if _interval_like(key):             # <<<<<<<<<<<<<<
 *             raise TypeError('__contains__ not defined for two intervals')
 *         return ((self.left < key if self.open_left else self.left <= key) and
 */
  __pyx_t_1 = __pyx_f_6pandas_5_libs_8interval__interval_like(__pyx_v_key); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 334, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 334, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(__pyx_t_2)) {

    /* "pandas/_libs/interval.pyx":335
 *     def __contains__(self, key):
 *         if _interval_like(key):
 *             raise TypeError('__contains__ not defined for two intervals')             # <<<<<<<<<<<<<<
 *         return ((self.left < key if self.open_left else self.left <= key) and
 *                 (key < self.right if self.open_right else key <= self.right))
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 335, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 335, __pyx_L1_error)

    /* "pandas/_libs/interval.pyx":334
 * 
 *     def __contains__(self, key):
 *         if _interval_like(key):             # <<<<<<<<<<<<<<
 *             raise TypeError('__contains__ not defined for two intervals')
 *         return ((self.left < key if self.open_left else self.left <= key) and
 */
  }

  /* "pandas/_libs/interval.pyx":336
 *         if _interval_like(key):
 *             raise TypeError('__contains__ not defined for two intervals')
 *         return ((self.left < key if self.open_left else self.left <= key) and             # <<<<<<<<<<<<<<
 *                 (key < self.right if self.open_right else key <= self.right))
 * 
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_open_left); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 336, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 336, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__pyx_t_2) {
    __pyx_t_4 = PyObject_RichCompare(__pyx_v_self->left, __pyx_v_key, Py_LT); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 336, __pyx_L1_error)
    __pyx_t_1 = __pyx_t_4;
    __pyx_t_4 = 0;
  } else {
    __pyx_t_4 = PyObject_RichCompare(__pyx_v_self->left, __pyx_v_key, Py_LE); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 336, __pyx_L1_error)
    __pyx_t_1 = __pyx_t_4;
    __pyx_t_4 = 0;
  }
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 336, __pyx_L1_error)
  if (__pyx_t_2) {
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  } else {
    __pyx_t_5 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 336, __pyx_L1_error)
    __pyx_t_3 = __pyx_t_5;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    goto __pyx_L4_bool_binop_done;
  }

  /* "pandas/_libs/interval.pyx":337
 *             raise TypeError('__contains__ not defined for two intervals')
 *         return ((self.left < key if self.open_left else self.left <= key) and
 *                 (key < self.right if self.open_right else key <= self.right))             # <<<<<<<<<<<<<<
 * 
 *     def __richcmp__(self, other, op: int):
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_open_right); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 337, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 337, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__pyx_t_2) {
    __pyx_t_4 = PyObject_RichCompare(__pyx_v_key, __pyx_v_self->right, Py_LT); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 337, __pyx_L1_error)
    __pyx_t_1 = __pyx_t_4;
    __pyx_t_4 = 0;
  } else {
    __pyx_t_4 = PyObject_RichCompare(__pyx_v_key, __pyx_v_self->right, Py_LE); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 337, __pyx_L1_error)
    __pyx_t_1 = __pyx_t_4;
    __pyx_t_4 = 0;
  }
  __pyx_t_5 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 337, __pyx_L1_error)
  __pyx_t_3 = __pyx_t_5;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_L4_bool_binop_done:;
  __pyx_r = __pyx_t_3;
  goto __pyx_L0;

  /* "pandas/_libs/interval.pyx":333
 *         return hash((self.left, self.right, self.closed))
 * 
 *     def __contains__(self, key):             # <<<<<<<<<<<<<<
 *         if _interval_like(key):
 *             raise TypeError('__contains__ not defined for two intervals')
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pandas._libs.interval.Interval.__contains__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/interval.pyx":339
 *                 (key < self.right if self.open_right else key <= self.right))
 * 
 *     def __richcmp__(self, other, op: int):             # <<<<<<<<<<<<<<
 *         if hasattr(other, 'ndim'):
 *             # let numpy (or IntervalIndex) handle vectorization
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_8Interval_9__richcmp__(PyObject *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_arg_op); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_8Interval_9__richcmp__(PyObject *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_arg_op) {
  PyObject *__pyx_v_op = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__richcmp__ (wrapper)", 0);
  __pyx_v_op = __Pyx_PyInt_From_int(__pyx_arg_op); if (unlikely(!__pyx_v_op)) __PYX_ERR(0, 339, __pyx_L3_error)
  __Pyx_GOTREF(__pyx_v_op);
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Interval.__richcmp__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_8Interval_8__richcmp__(((struct __pyx_obj_6pandas_5_libs_8interval_Interval *)__pyx_v_self), ((PyObject *)__pyx_v_other), ((PyObject *)__pyx_v_op));

  /* function exit code */
  __Pyx_XDECREF(__pyx_v_op);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_8Interval_8__richcmp__(struct __pyx_obj_6pandas_5_libs_8interval_Interval *__pyx_v_self, PyObject *__pyx_v_other, PyObject *__pyx_v_op) {
  PyObject *__pyx_v_self_tuple = NULL;
  PyObject *__pyx_v_other_tuple = NULL;
  PyObject *__pyx_v_name = NULL;
  PyObject *__pyx_v_op_str = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  __Pyx_RefNannySetupContext("__richcmp__", 0);
  __Pyx_INCREF(__pyx_v_other);

  /* "pandas/_libs/interval.pyx":340
 * 
 *     def __richcmp__(self, other, op: int):
 *         if hasattr(other, 'ndim'):             # <<<<<<<<<<<<<<
 *             # let numpy (or IntervalIndex) handle vectorization
 *             return NotImplemented
 */
  __pyx_t_1 = __Pyx_HasAttr(__pyx_v_other, __pyx_n_u_ndim); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 340, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pandas/_libs/interval.pyx":342
 *         if hasattr(other, 'ndim'):
 *             # let numpy (or IntervalIndex) handle vectorization
 *             return NotImplemented             # <<<<<<<<<<<<<<
 * 
 *         if _interval_like(other):
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_builtin_NotImplemented);
    __pyx_r = __pyx_builtin_NotImplemented;
    goto __pyx_L0;

    /* "pandas/_libs/interval.pyx":340
 * 
 *     def __richcmp__(self, other, op: int):
 *         if hasattr(other, 'ndim'):             # <<<<<<<<<<<<<<
 *             # let numpy (or IntervalIndex) handle vectorization
 *             return NotImplemented
 */
  }

  /* "pandas/_libs/interval.pyx":344
 *             return NotImplemented
 * 
 *         if _interval_like(other):             # <<<<<<<<<<<<<<
 *             self_tuple = (self.left, self.right, self.closed)
 *             other_tuple = (other.left, other.right, other.closed)
 */
  __pyx_t_3 = __pyx_f_6pandas_5_libs_8interval__interval_like(__pyx_v_other); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 344, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 344, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_2) {

    /* "pandas/_libs/interval.pyx":345
 * 
 *         if _interval_like(other):
 *             self_tuple = (self.left, self.right, self.closed)             # <<<<<<<<<<<<<<
 *             other_tuple = (other.left, other.right, other.closed)
 *             return PyObject_RichCompare(self_tuple, other_tuple, op)
 */
    __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 345, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_v_self->left);
    __Pyx_GIVEREF(__pyx_v_self->left);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_self->left);
    __Pyx_INCREF(__pyx_v_self->right);
    __Pyx_GIVEREF(__pyx_v_self->right);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self->right);
    __Pyx_INCREF(__pyx_v_self->closed);
    __Pyx_GIVEREF(__pyx_v_self->closed);
    PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_v_self->closed);
    __pyx_v_self_tuple = ((PyObject*)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "pandas/_libs/interval.pyx":346
 *         if _interval_like(other):
 *             self_tuple = (self.left, self.right, self.closed)
 *             other_tuple = (other.left, other.right, other.closed)             # <<<<<<<<<<<<<<
 *             return PyObject_RichCompare(self_tuple, other_tuple, op)
 * 
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_other, __pyx_n_s_left); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 346, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_other, __pyx_n_s_right); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 346, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_other, __pyx_n_s_closed); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 346, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = PyTuple_New(3); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 346, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_t_5);
    __pyx_t_3 = 0;
    __pyx_t_4 = 0;
    __pyx_t_5 = 0;
    __pyx_v_other_tuple = ((PyObject*)__pyx_t_6);
    __pyx_t_6 = 0;

    /* "pandas/_libs/interval.pyx":347
 *             self_tuple = (self.left, self.right, self.closed)
 *             other_tuple = (other.left, other.right, other.closed)
 *             return PyObject_RichCompare(self_tuple, other_tuple, op)             # <<<<<<<<<<<<<<
 * 
 *         # nb. could just return NotImplemented now, but handling this
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_7 = __Pyx_PyInt_As_int(__pyx_v_op); if (unlikely((__pyx_t_7 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 347, __pyx_L1_error)
    __pyx_t_6 = PyObject_RichCompare(__pyx_v_self_tuple, __pyx_v_other_tuple, __pyx_t_7); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 347, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_r = __pyx_t_6;
    __pyx_t_6 = 0;
    goto __pyx_L0;

    /* "pandas/_libs/interval.pyx":344
 *             return NotImplemented
 * 
 *         if _interval_like(other):             # <<<<<<<<<<<<<<
 *             self_tuple = (self.left, self.right, self.closed)
 *             other_tuple = (other.left, other.right, other.closed)
 */
  }

  /* "pandas/_libs/interval.pyx":352
 *         # explicitly allows us to opt into the Python 3 behavior, even on
 *         # Python 2.
 *         if op == Py_EQ or op == Py_NE:             # <<<<<<<<<<<<<<
 *             return NotImplemented
 *         else:
 */
  __pyx_t_6 = __Pyx_PyInt_From_int(Py_EQ); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 352, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_5 = PyObject_RichCompare(__pyx_v_op, __pyx_t_6, Py_EQ); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 352, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 352, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (!__pyx_t_1) {
  } else {
    __pyx_t_2 = __pyx_t_1;
    goto __pyx_L6_bool_binop_done;
  }
  __pyx_t_5 = __Pyx_PyInt_From_int(Py_NE); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 352, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = PyObject_RichCompare(__pyx_v_op, __pyx_t_5, Py_EQ); __Pyx_XGOTREF(__pyx_t_6); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 352, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_6); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 352, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_2 = __pyx_t_1;
  __pyx_L6_bool_binop_done:;
  if (likely(__pyx_t_2)) {

    /* "pandas/_libs/interval.pyx":353
 *         # Python 2.
 *         if op == Py_EQ or op == Py_NE:
 *             return NotImplemented             # <<<<<<<<<<<<<<
 *         else:
 *             name = type(self).__name__
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_builtin_NotImplemented);
    __pyx_r = __pyx_builtin_NotImplemented;
    goto __pyx_L0;

    /* "pandas/_libs/interval.pyx":352
 *         # explicitly allows us to opt into the Python 3 behavior, even on
 *         # Python 2.
 *         if op == Py_EQ or op == Py_NE:             # <<<<<<<<<<<<<<
 *             return NotImplemented
 *         else:
 */
  }

  /* "pandas/_libs/interval.pyx":355
 *             return NotImplemented
 *         else:
 *             name = type(self).__name__             # <<<<<<<<<<<<<<
 *             other = type(other).__name__
 *             op_str = {Py_LT: '<', Py_LE: '<=', Py_GT: '>', Py_GE: '>='}[op]
 */
  /*else*/ {
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))), __pyx_n_s_name_2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 355, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_v_name = __pyx_t_6;
    __pyx_t_6 = 0;

    /* "pandas/_libs/interval.pyx":356
 *         else:
 *             name = type(self).__name__
 *             other = type(other).__name__             # <<<<<<<<<<<<<<
 *             op_str = {Py_LT: '<', Py_LE: '<=', Py_GT: '>', Py_GE: '>='}[op]
 *             raise TypeError('unorderable types: {name}() {op} {other}()'
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)Py_TYPE(__pyx_v_other)), __pyx_n_s_name_2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 356, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF_SET(__pyx_v_other, __pyx_t_6);
    __pyx_t_6 = 0;

    /* "pandas/_libs/interval.pyx":357
 *             name = type(self).__name__
 *             other = type(other).__name__
 *             op_str = {Py_LT: '<', Py_LE: '<=', Py_GT: '>', Py_GE: '>='}[op]             # <<<<<<<<<<<<<<
 *             raise TypeError('unorderable types: {name}() {op} {other}()'
 *                             .format(name=name, op=op_str, other=other))
 */
    __pyx_t_6 = __Pyx_PyDict_NewPresized(4); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 357, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_5 = __Pyx_PyInt_From_int(Py_LT); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 357, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (PyDict_SetItem(__pyx_t_6, __pyx_t_5, __pyx_kp_u__3) < 0) __PYX_ERR(0, 357, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyInt_From_int(Py_LE); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 357, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (PyDict_SetItem(__pyx_t_6, __pyx_t_5, __pyx_kp_u__4) < 0) __PYX_ERR(0, 357, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyInt_From_int(Py_GT); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 357, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (PyDict_SetItem(__pyx_t_6, __pyx_t_5, __pyx_kp_u__5) < 0) __PYX_ERR(0, 357, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyInt_From_int(Py_GE); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 357, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (PyDict_SetItem(__pyx_t_6, __pyx_t_5, __pyx_kp_u__6) < 0) __PYX_ERR(0, 357, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyDict_GetItem(__pyx_t_6, __pyx_v_op); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 357, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_v_op_str = __pyx_t_5;
    __pyx_t_5 = 0;

    /* "pandas/_libs/interval.pyx":359
 *             op_str = {Py_LT: '<', Py_LE: '<=', Py_GT: '>', Py_GE: '>='}[op]
 *             raise TypeError('unorderable types: {name}() {op} {other}()'
 *                             .format(name=name, op=op_str, other=other))             # <<<<<<<<<<<<<<
 * 
 *     def __reduce__(self):
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_unorderable_types_name_op_other, __pyx_n_s_format); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 359, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 359, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (PyDict_SetItem(__pyx_t_6, __pyx_n_s_name, __pyx_v_name) < 0) __PYX_ERR(0, 359, __pyx_L1_error)
    if (PyDict_SetItem(__pyx_t_6, __pyx_n_s_op, __pyx_v_op_str) < 0) __PYX_ERR(0, 359, __pyx_L1_error)
    if (PyDict_SetItem(__pyx_t_6, __pyx_n_s_other, __pyx_v_other) < 0) __PYX_ERR(0, 359, __pyx_L1_error)
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_empty_tuple, __pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 359, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "pandas/_libs/interval.pyx":358
 *             other = type(other).__name__
 *             op_str = {Py_LT: '<', Py_LE: '<=', Py_GT: '>', Py_GE: '>='}[op]
 *             raise TypeError('unorderable types: {name}() {op} {other}()'             # <<<<<<<<<<<<<<
 *                             .format(name=name, op=op_str, other=other))
 * 
 */
    __pyx_t_6 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_4); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 358, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(0, 358, __pyx_L1_error)
  }

  /* "pandas/_libs/interval.pyx":339
 *                 (key < self.right if self.open_right else key <= self.right))
 * 
 *     def __richcmp__(self, other, op: int):             # <<<<<<<<<<<<<<
 *         if hasattr(other, 'ndim'):
 *             # let numpy (or IntervalIndex) handle vectorization
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pandas._libs.interval.Interval.__richcmp__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_self_tuple);
  __Pyx_XDECREF(__pyx_v_other_tuple);
  __Pyx_XDECREF(__pyx_v_name);
  __Pyx_XDECREF(__pyx_v_op_str);
  __Pyx_XDECREF(__pyx_v_other);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/interval.pyx":361
 *                             .format(name=name, op=op_str, other=other))
 * 
 *     def __reduce__(self):             # <<<<<<<<<<<<<<
 *         args = (self.left, self.right, self.closed)
 *         return (type(self), args)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_8Interval_11__reduce__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_8Interval_11__reduce__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_8Interval_10__reduce__(((struct __pyx_obj_6pandas_5_libs_8interval_Interval *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_8Interval_10__reduce__(struct __pyx_obj_6pandas_5_libs_8interval_Interval *__pyx_v_self) {
  PyObject *__pyx_v_args = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__reduce__", 0);

  /* "pandas/_libs/interval.pyx":362
 * 
 *     def __reduce__(self):
 *         args = (self.left, self.right, self.closed)             # <<<<<<<<<<<<<<
 *         return (type(self), args)
 * 
 */
  __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 362, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_self->left);
  __Pyx_GIVEREF(__pyx_v_self->left);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_self->left);
  __Pyx_INCREF(__pyx_v_self->right);
  __Pyx_GIVEREF(__pyx_v_self->right);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_self->right);
  __Pyx_INCREF(__pyx_v_self->closed);
  __Pyx_GIVEREF(__pyx_v_self->closed);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_self->closed);
  __pyx_v_args = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pandas/_libs/interval.pyx":363
 *     def __reduce__(self):
 *         args = (self.left, self.right, self.closed)
 *         return (type(self), args)             # <<<<<<<<<<<<<<
 * 
 *     def _repr_base(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 363, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
  __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
  PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
  __Pyx_INCREF(__pyx_v_args);
  __Pyx_GIVEREF(__pyx_v_args);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_args);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pandas/_libs/interval.pyx":361
 *                             .format(name=name, op=op_str, other=other))
 * 
 *     def __reduce__(self):             # <<<<<<<<<<<<<<
 *         args = (self.left, self.right, self.closed)
 *         return (type(self), args)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Interval.__reduce__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_args);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/interval.pyx":365
 *         return (type(self), args)
 * 
 *     def _repr_base(self):             # <<<<<<<<<<<<<<
 *         left = self.left
 *         right = self.right
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_8Interval_13_repr_base(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_8Interval_13_repr_base(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_repr_base (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_8Interval_12_repr_base(((struct __pyx_obj_6pandas_5_libs_8interval_Interval *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_8Interval_12_repr_base(struct __pyx_obj_6pandas_5_libs_8interval_Interval *__pyx_v_self) {
  PyObject *__pyx_v_left = NULL;
  PyObject *__pyx_v_right = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  __Pyx_RefNannySetupContext("_repr_base", 0);

  /* "pandas/_libs/interval.pyx":366
 * 
 *     def _repr_base(self):
 *         left = self.left             # <<<<<<<<<<<<<<
 *         right = self.right
 * 
 */
  __pyx_t_1 = __pyx_v_self->left;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_left = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pandas/_libs/interval.pyx":367
 *     def _repr_base(self):
 *         left = self.left
 *         right = self.right             # <<<<<<<<<<<<<<
 * 
 *         # TODO: need more general formatting methodology here
 */
  __pyx_t_1 = __pyx_v_self->right;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_right = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pandas/_libs/interval.pyx":370
 * 
 *         # TODO: need more general formatting methodology here
 *         if isinstance(left, Timestamp) and isinstance(right, Timestamp):             # <<<<<<<<<<<<<<
 *             left = left._short_repr
 *             right = right._short_repr
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_Timestamp); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 370, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyObject_IsInstance(__pyx_v_left, __pyx_t_1); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 370, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_4 = (__pyx_t_3 != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_Timestamp); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 370, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = PyObject_IsInstance(__pyx_v_right, __pyx_t_1); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(0, 370, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = (__pyx_t_4 != 0);
  __pyx_t_2 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "pandas/_libs/interval.pyx":371
 *         # TODO: need more general formatting methodology here
 *         if isinstance(left, Timestamp) and isinstance(right, Timestamp):
 *             left = left._short_repr             # <<<<<<<<<<<<<<
 *             right = right._short_repr
 * 
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_left, __pyx_n_s_short_repr); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 371, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF_SET(__pyx_v_left, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "pandas/_libs/interval.pyx":372
 *         if isinstance(left, Timestamp) and isinstance(right, Timestamp):
 *             left = left._short_repr
 *             right = right._short_repr             # <<<<<<<<<<<<<<
 * 
 *         return left, right
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_right, __pyx_n_s_short_repr); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 372, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF_SET(__pyx_v_right, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "pandas/_libs/interval.pyx":370
 * 
 *         # TODO: need more general formatting methodology here
 *         if isinstance(left, Timestamp) and isinstance(right, Timestamp):             # <<<<<<<<<<<<<<
 *             left = left._short_repr
 *             right = right._short_repr
 */
  }

  /* "pandas/_libs/interval.pyx":374
 *             right = right._short_repr
 * 
 *         return left, right             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 374, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_left);
  __Pyx_GIVEREF(__pyx_v_left);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_left);
  __Pyx_INCREF(__pyx_v_right);
  __Pyx_GIVEREF(__pyx_v_right);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_right);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pandas/_libs/interval.pyx":365
 *         return (type(self), args)
 * 
 *     def _repr_base(self):             # <<<<<<<<<<<<<<
 *         left = self.left
 *         right = self.right
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Interval._repr_base", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_left);
  __Pyx_XDECREF(__pyx_v_right);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/interval.pyx":376
 *         return left, right
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 * 
 *         left, right = self._repr_base()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_8Interval_15__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_8Interval_15__repr__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_8Interval_14__repr__(((struct __pyx_obj_6pandas_5_libs_8interval_Interval *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_8Interval_14__repr__(struct __pyx_obj_6pandas_5_libs_8interval_Interval *__pyx_v_self) {
  PyObject *__pyx_v_left = NULL;
  PyObject *__pyx_v_right = NULL;
  PyObject *__pyx_v_name = NULL;
  PyObject *__pyx_v_repr_str = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *(*__pyx_t_5)(PyObject *);
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "pandas/_libs/interval.pyx":378
 *     def __repr__(self):
 * 
 *         left, right = self._repr_base()             # <<<<<<<<<<<<<<
 *         name = type(self).__name__
 *         repr_str = '{name}({left!r}, {right!r}, closed={closed!r})'.format(
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_repr_base); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 378, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 378, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if ((likely(PyTuple_CheckExact(__pyx_t_1))) || (PyList_CheckExact(__pyx_t_1))) {
    PyObject* sequence = __pyx_t_1;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 2)) {
      if (size > 2) __Pyx_RaiseTooManyValuesError(2);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 378, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_2 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_3 = PyTuple_GET_ITEM(sequence, 1); 
    } else {
      __pyx_t_2 = PyList_GET_ITEM(sequence, 0); 
      __pyx_t_3 = PyList_GET_ITEM(sequence, 1); 
    }
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_3);
    #else
    __pyx_t_2 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 378, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 378, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  } else {
    Py_ssize_t index = -1;
    __pyx_t_4 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 378, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_5 = Py_TYPE(__pyx_t_4)->tp_iternext;
    index = 0; __pyx_t_2 = __pyx_t_5(__pyx_t_4); if (unlikely(!__pyx_t_2)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_2);
    index = 1; __pyx_t_3 = __pyx_t_5(__pyx_t_4); if (unlikely(!__pyx_t_3)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_3);
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_5(__pyx_t_4), 2) < 0) __PYX_ERR(0, 378, __pyx_L1_error)
    __pyx_t_5 = NULL;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    goto __pyx_L4_unpacking_done;
    __pyx_L3_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_5 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(0, 378, __pyx_L1_error)
    __pyx_L4_unpacking_done:;
  }
  __pyx_v_left = __pyx_t_2;
  __pyx_t_2 = 0;
  __pyx_v_right = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "pandas/_libs/interval.pyx":379
 * 
 *         left, right = self._repr_base()
 *         name = type(self).__name__             # <<<<<<<<<<<<<<
 *         repr_str = '{name}({left!r}, {right!r}, closed={closed!r})'.format(
 *             name=name, left=left, right=right, closed=self.closed)
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))), __pyx_n_s_name_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 379, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_name = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pandas/_libs/interval.pyx":380
 *         left, right = self._repr_base()
 *         name = type(self).__name__
 *         repr_str = '{name}({left!r}, {right!r}, closed={closed!r})'.format(             # <<<<<<<<<<<<<<
 *             name=name, left=left, right=right, closed=self.closed)
 *         return repr_str
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_name_left_r_right_r_closed_clos, __pyx_n_s_format); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 380, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "pandas/_libs/interval.pyx":381
 *         name = type(self).__name__
 *         repr_str = '{name}({left!r}, {right!r}, closed={closed!r})'.format(
 *             name=name, left=left, right=right, closed=self.closed)             # <<<<<<<<<<<<<<
 *         return repr_str
 * 
 */
  __pyx_t_3 = __Pyx_PyDict_NewPresized(4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 381, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_name, __pyx_v_name) < 0) __PYX_ERR(0, 381, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_left, __pyx_v_left) < 0) __PYX_ERR(0, 381, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_right, __pyx_v_right) < 0) __PYX_ERR(0, 381, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_closed, __pyx_v_self->closed) < 0) __PYX_ERR(0, 381, __pyx_L1_error)

  /* "pandas/_libs/interval.pyx":380
 *         left, right = self._repr_base()
 *         name = type(self).__name__
 *         repr_str = '{name}({left!r}, {right!r}, closed={closed!r})'.format(             # <<<<<<<<<<<<<<
 *             name=name, left=left, right=right, closed=self.closed)
 *         return repr_str
 */
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_empty_tuple, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 380, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_repr_str = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "pandas/_libs/interval.pyx":382
 *         repr_str = '{name}({left!r}, {right!r}, closed={closed!r})'.format(
 *             name=name, left=left, right=right, closed=self.closed)
 *         return repr_str             # <<<<<<<<<<<<<<
 * 
 *     def __str__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_repr_str);
  __pyx_r = __pyx_v_repr_str;
  goto __pyx_L0;

  /* "pandas/_libs/interval.pyx":376
 *         return left, right
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 * 
 *         left, right = self._repr_base()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pandas._libs.interval.Interval.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_left);
  __Pyx_XDECREF(__pyx_v_right);
  __Pyx_XDECREF(__pyx_v_name);
  __Pyx_XDECREF(__pyx_v_repr_str);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/interval.pyx":384
 *         return repr_str
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 * 
 *         left, right = self._repr_base()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_8Interval_17__str__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_8Interval_17__str__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__str__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_8Interval_16__str__(((struct __pyx_obj_6pandas_5_libs_8interval_Interval *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_8Interval_16__str__(struct __pyx_obj_6pandas_5_libs_8interval_Interval *__pyx_v_self) {
  PyObject *__pyx_v_left = NULL;
  PyObject *__pyx_v_right = NULL;
  PyObject *__pyx_v_start_symbol = NULL;
  PyObject *__pyx_v_end_symbol = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *(*__pyx_t_5)(PyObject *);
  int __pyx_t_6;
  __Pyx_RefNannySetupContext("__str__", 0);

  /* "pandas/_libs/interval.pyx":386
 *     def __str__(self):
 * 
 *         left, right = self._repr_base()             # <<<<<<<<<<<<<<
 *         start_symbol = '[' if self.closed_left else '('
 *         end_symbol = ']' if self.closed_right else ')'
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_repr_base); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 386, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 386, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if ((likely(PyTuple_CheckExact(__pyx_t_1))) || (PyList_CheckExact(__pyx_t_1))) {
    PyObject* sequence = __pyx_t_1;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 2)) {
      if (size > 2) __Pyx_RaiseTooManyValuesError(2);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 386, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_2 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_3 = PyTuple_GET_ITEM(sequence, 1); 
    } else {
      __pyx_t_2 = PyList_GET_ITEM(sequence, 0); 
      __pyx_t_3 = PyList_GET_ITEM(sequence, 1); 
    }
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_3);
    #else
    __pyx_t_2 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 386, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 386, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  } else {
    Py_ssize_t index = -1;
    __pyx_t_4 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 386, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_5 = Py_TYPE(__pyx_t_4)->tp_iternext;
    index = 0; __pyx_t_2 = __pyx_t_5(__pyx_t_4); if (unlikely(!__pyx_t_2)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_2);
    index = 1; __pyx_t_3 = __pyx_t_5(__pyx_t_4); if (unlikely(!__pyx_t_3)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_3);
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_5(__pyx_t_4), 2) < 0) __PYX_ERR(0, 386, __pyx_L1_error)
    __pyx_t_5 = NULL;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    goto __pyx_L4_unpacking_done;
    __pyx_L3_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_5 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(0, 386, __pyx_L1_error)
    __pyx_L4_unpacking_done:;
  }
  __pyx_v_left = __pyx_t_2;
  __pyx_t_2 = 0;
  __pyx_v_right = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "pandas/_libs/interval.pyx":387
 * 
 *         left, right = self._repr_base()
 *         start_symbol = '[' if self.closed_left else '('             # <<<<<<<<<<<<<<
 *         end_symbol = ']' if self.closed_right else ')'
 *         return '{start}{left}, {right}{end}'.format(
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_closed_left); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 387, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 387, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_6) {
    __Pyx_INCREF(__pyx_kp_u__7);
    __pyx_t_1 = __pyx_kp_u__7;
  } else {
    __Pyx_INCREF(__pyx_kp_u__8);
    __pyx_t_1 = __pyx_kp_u__8;
  }
  __pyx_v_start_symbol = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pandas/_libs/interval.pyx":388
 *         left, right = self._repr_base()
 *         start_symbol = '[' if self.closed_left else '('
 *         end_symbol = ']' if self.closed_right else ')'             # <<<<<<<<<<<<<<
 *         return '{start}{left}, {right}{end}'.format(
 *             start=start_symbol, left=left, right=right, end=end_symbol)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_closed_right); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 388, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 388, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_6) {
    __Pyx_INCREF(__pyx_kp_u__9);
    __pyx_t_1 = __pyx_kp_u__9;
  } else {
    __Pyx_INCREF(__pyx_kp_u__10);
    __pyx_t_1 = __pyx_kp_u__10;
  }
  __pyx_v_end_symbol = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pandas/_libs/interval.pyx":389
 *         start_symbol = '[' if self.closed_left else '('
 *         end_symbol = ']' if self.closed_right else ')'
 *         return '{start}{left}, {right}{end}'.format(             # <<<<<<<<<<<<<<
 *             start=start_symbol, left=left, right=right, end=end_symbol)
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_start_left_right_end, __pyx_n_s_format); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 389, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "pandas/_libs/interval.pyx":390
 *         end_symbol = ']' if self.closed_right else ')'
 *         return '{start}{left}, {right}{end}'.format(
 *             start=start_symbol, left=left, right=right, end=end_symbol)             # <<<<<<<<<<<<<<
 * 
 *     def __add__(self, y):
 */
  __pyx_t_3 = __Pyx_PyDict_NewPresized(4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 390, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_start, __pyx_v_start_symbol) < 0) __PYX_ERR(0, 390, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_left, __pyx_v_left) < 0) __PYX_ERR(0, 390, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_right, __pyx_v_right) < 0) __PYX_ERR(0, 390, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_end, __pyx_v_end_symbol) < 0) __PYX_ERR(0, 390, __pyx_L1_error)

  /* "pandas/_libs/interval.pyx":389
 *         start_symbol = '[' if self.closed_left else '('
 *         end_symbol = ']' if self.closed_right else ')'
 *         return '{start}{left}, {right}{end}'.format(             # <<<<<<<<<<<<<<
 *             start=start_symbol, left=left, right=right, end=end_symbol)
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_empty_tuple, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 389, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pandas/_libs/interval.pyx":384
 *         return repr_str
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 * 
 *         left, right = self._repr_base()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pandas._libs.interval.Interval.__str__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_left);
  __Pyx_XDECREF(__pyx_v_right);
  __Pyx_XDECREF(__pyx_v_start_symbol);
  __Pyx_XDECREF(__pyx_v_end_symbol);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/interval.pyx":392
 *             start=start_symbol, left=left, right=right, end=end_symbol)
 * 
 *     def __add__(self, y):             # <<<<<<<<<<<<<<
 *         if isinstance(y, numbers.Number):
 *             return Interval(self.left + y, self.right + y, closed=self.closed)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_8Interval_19__add__(PyObject *__pyx_v_self, PyObject *__pyx_v_y); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_8Interval_19__add__(PyObject *__pyx_v_self, PyObject *__pyx_v_y) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__add__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_8Interval_18__add__(((PyObject *)__pyx_v_self), ((PyObject *)__pyx_v_y));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_8Interval_18__add__(PyObject *__pyx_v_self, PyObject *__pyx_v_y) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  __Pyx_RefNannySetupContext("__add__", 0);

  /* "pandas/_libs/interval.pyx":393
 * 
 *     def __add__(self, y):
 *         if isinstance(y, numbers.Number):             # <<<<<<<<<<<<<<
 *             return Interval(self.left + y, self.right + y, closed=self.closed)
 *         elif isinstance(y, Interval) and isinstance(self, numbers.Number):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_numbers); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 393, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_Number); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 393, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = PyObject_IsInstance(__pyx_v_y, __pyx_t_2); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 393, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = (__pyx_t_3 != 0);
  if (__pyx_t_4) {

    /* "pandas/_libs/interval.pyx":394
 *     def __add__(self, y):
 *         if isinstance(y, numbers.Number):
 *             return Interval(self.left + y, self.right + y, closed=self.closed)             # <<<<<<<<<<<<<<
 *         elif isinstance(y, Interval) and isinstance(self, numbers.Number):
 *             return Interval(y.left + self, y.right + self, closed=y.closed)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_left); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 394, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = PyNumber_Add(__pyx_t_2, __pyx_v_y); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 394, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_right); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 394, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = PyNumber_Add(__pyx_t_2, __pyx_v_y); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 394, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 394, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_5);
    __pyx_t_1 = 0;
    __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 394, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_closed); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 394, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_closed, __pyx_t_1) < 0) __PYX_ERR(0, 394, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6pandas_5_libs_8interval_Interval), __pyx_t_2, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 394, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "pandas/_libs/interval.pyx":393
 * 
 *     def __add__(self, y):
 *         if isinstance(y, numbers.Number):             # <<<<<<<<<<<<<<
 *             return Interval(self.left + y, self.right + y, closed=self.closed)
 *         elif isinstance(y, Interval) and isinstance(self, numbers.Number):
 */
  }

  /* "pandas/_libs/interval.pyx":395
 *         if isinstance(y, numbers.Number):
 *             return Interval(self.left + y, self.right + y, closed=self.closed)
 *         elif isinstance(y, Interval) and isinstance(self, numbers.Number):             # <<<<<<<<<<<<<<
 *             return Interval(y.left + self, y.right + self, closed=y.closed)
 *         return NotImplemented
 */
  __pyx_t_3 = __Pyx_TypeCheck(__pyx_v_y, __pyx_ptype_6pandas_5_libs_8interval_Interval); 
  __pyx_t_6 = (__pyx_t_3 != 0);
  if (__pyx_t_6) {
  } else {
    __pyx_t_4 = __pyx_t_6;
    goto __pyx_L4_bool_binop_done;
  }
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_numbers); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 395, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_Number); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 395, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_6 = PyObject_IsInstance(__pyx_v_self, __pyx_t_5); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 395, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_3 = (__pyx_t_6 != 0);
  __pyx_t_4 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_4) {

    /* "pandas/_libs/interval.pyx":396
 *             return Interval(self.left + y, self.right + y, closed=self.closed)
 *         elif isinstance(y, Interval) and isinstance(self, numbers.Number):
 *             return Interval(y.left + self, y.right + self, closed=y.closed)             # <<<<<<<<<<<<<<
 *         return NotImplemented
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_y, __pyx_n_s_left); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 396, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = PyNumber_Add(__pyx_t_5, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 396, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_y, __pyx_n_s_right); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 396, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_2 = PyNumber_Add(__pyx_t_5, __pyx_v_self); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 396, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 396, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_2);
    __pyx_t_1 = 0;
    __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 396, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_y, __pyx_n_s_closed); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 396, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_closed, __pyx_t_1) < 0) __PYX_ERR(0, 396, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6pandas_5_libs_8interval_Interval), __pyx_t_5, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 396, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "pandas/_libs/interval.pyx":395
 *         if isinstance(y, numbers.Number):
 *             return Interval(self.left + y, self.right + y, closed=self.closed)
 *         elif isinstance(y, Interval) and isinstance(self, numbers.Number):             # <<<<<<<<<<<<<<
 *             return Interval(y.left + self, y.right + self, closed=y.closed)
 *         return NotImplemented
 */
  }

  /* "pandas/_libs/interval.pyx":397
 *         elif isinstance(y, Interval) and isinstance(self, numbers.Number):
 *             return Interval(y.left + self, y.right + self, closed=y.closed)
 *         return NotImplemented             # <<<<<<<<<<<<<<
 * 
 *     def __sub__(self, y):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_builtin_NotImplemented);
  __pyx_r = __pyx_builtin_NotImplemented;
  goto __pyx_L0;

  /* "pandas/_libs/interval.pyx":392
 *             start=start_symbol, left=left, right=right, end=end_symbol)
 * 
 *     def __add__(self, y):             # <<<<<<<<<<<<<<
 *         if isinstance(y, numbers.Number):
 *             return Interval(self.left + y, self.right + y, closed=self.closed)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pandas._libs.interval.Interval.__add__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/interval.pyx":399
 *         return NotImplemented
 * 
 *     def __sub__(self, y):             # <<<<<<<<<<<<<<
 *         if isinstance(y, numbers.Number):
 *             return Interval(self.left - y, self.right - y, closed=self.closed)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_8Interval_21__sub__(PyObject *__pyx_v_self, PyObject *__pyx_v_y); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_8Interval_21__sub__(PyObject *__pyx_v_self, PyObject *__pyx_v_y) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__sub__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_8Interval_20__sub__(((PyObject *)__pyx_v_self), ((PyObject *)__pyx_v_y));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_8Interval_20__sub__(PyObject *__pyx_v_self, PyObject *__pyx_v_y) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__sub__", 0);

  /* "pandas/_libs/interval.pyx":400
 * 
 *     def __sub__(self, y):
 *         if isinstance(y, numbers.Number):             # <<<<<<<<<<<<<<
 *             return Interval(self.left - y, self.right - y, closed=self.closed)
 *         return NotImplemented
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_numbers); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 400, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_Number); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 400, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = PyObject_IsInstance(__pyx_v_y, __pyx_t_2); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 400, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = (__pyx_t_3 != 0);
  if (__pyx_t_4) {

    /* "pandas/_libs/interval.pyx":401
 *     def __sub__(self, y):
 *         if isinstance(y, numbers.Number):
 *             return Interval(self.left - y, self.right - y, closed=self.closed)             # <<<<<<<<<<<<<<
 *         return NotImplemented
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_left); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 401, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = PyNumber_Subtract(__pyx_t_2, __pyx_v_y); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 401, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_right); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 401, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = PyNumber_Subtract(__pyx_t_2, __pyx_v_y); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 401, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 401, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_5);
    __pyx_t_1 = 0;
    __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 401, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_closed); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 401, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_closed, __pyx_t_1) < 0) __PYX_ERR(0, 401, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6pandas_5_libs_8interval_Interval), __pyx_t_2, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 401, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "pandas/_libs/interval.pyx":400
 * 
 *     def __sub__(self, y):
 *         if isinstance(y, numbers.Number):             # <<<<<<<<<<<<<<
 *             return Interval(self.left - y, self.right - y, closed=self.closed)
 *         return NotImplemented
 */
  }

  /* "pandas/_libs/interval.pyx":402
 *         if isinstance(y, numbers.Number):
 *             return Interval(self.left - y, self.right - y, closed=self.closed)
 *         return NotImplemented             # <<<<<<<<<<<<<<
 * 
 *     def __mul__(self, y):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_builtin_NotImplemented);
  __pyx_r = __pyx_builtin_NotImplemented;
  goto __pyx_L0;

  /* "pandas/_libs/interval.pyx":399
 *         return NotImplemented
 * 
 *     def __sub__(self, y):             # <<<<<<<<<<<<<<
 *         if isinstance(y, numbers.Number):
 *             return Interval(self.left - y, self.right - y, closed=self.closed)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pandas._libs.interval.Interval.__sub__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/interval.pyx":404
 *         return NotImplemented
 * 
 *     def __mul__(self, y):             # <<<<<<<<<<<<<<
 *         if isinstance(y, numbers.Number):
 *             return Interval(self.left * y, self.right * y, closed=self.closed)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_8Interval_23__mul__(PyObject *__pyx_v_self, PyObject *__pyx_v_y); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_8Interval_23__mul__(PyObject *__pyx_v_self, PyObject *__pyx_v_y) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__mul__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_8Interval_22__mul__(((PyObject *)__pyx_v_self), ((PyObject *)__pyx_v_y));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_8Interval_22__mul__(PyObject *__pyx_v_self, PyObject *__pyx_v_y) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  __Pyx_RefNannySetupContext("__mul__", 0);

  /* "pandas/_libs/interval.pyx":405
 * 
 *     def __mul__(self, y):
 *         if isinstance(y, numbers.Number):             # <<<<<<<<<<<<<<
 *             return Interval(self.left * y, self.right * y, closed=self.closed)
 *         elif isinstance(y, Interval) and isinstance(self, numbers.Number):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_numbers); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 405, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_Number); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 405, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = PyObject_IsInstance(__pyx_v_y, __pyx_t_2); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 405, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = (__pyx_t_3 != 0);
  if (__pyx_t_4) {

    /* "pandas/_libs/interval.pyx":406
 *     def __mul__(self, y):
 *         if isinstance(y, numbers.Number):
 *             return Interval(self.left * y, self.right * y, closed=self.closed)             # <<<<<<<<<<<<<<
 *         elif isinstance(y, Interval) and isinstance(self, numbers.Number):
 *             return Interval(y.left * self, y.right * self, closed=y.closed)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_left); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 406, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = PyNumber_Multiply(__pyx_t_2, __pyx_v_y); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 406, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_right); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 406, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = PyNumber_Multiply(__pyx_t_2, __pyx_v_y); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 406, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 406, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_5);
    __pyx_t_1 = 0;
    __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 406, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_closed); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 406, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_closed, __pyx_t_1) < 0) __PYX_ERR(0, 406, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6pandas_5_libs_8interval_Interval), __pyx_t_2, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 406, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "pandas/_libs/interval.pyx":405
 * 
 *     def __mul__(self, y):
 *         if isinstance(y, numbers.Number):             # <<<<<<<<<<<<<<
 *             return Interval(self.left * y, self.right * y, closed=self.closed)
 *         elif isinstance(y, Interval) and isinstance(self, numbers.Number):
 */
  }

  /* "pandas/_libs/interval.pyx":407
 *         if isinstance(y, numbers.Number):
 *             return Interval(self.left * y, self.right * y, closed=self.closed)
 *         elif isinstance(y, Interval) and isinstance(self, numbers.Number):             # <<<<<<<<<<<<<<
 *             return Interval(y.left * self, y.right * self, closed=y.closed)
 *         return NotImplemented
 */
  __pyx_t_3 = __Pyx_TypeCheck(__pyx_v_y, __pyx_ptype_6pandas_5_libs_8interval_Interval); 
  __pyx_t_6 = (__pyx_t_3 != 0);
  if (__pyx_t_6) {
  } else {
    __pyx_t_4 = __pyx_t_6;
    goto __pyx_L4_bool_binop_done;
  }
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_numbers); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 407, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_Number); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 407, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_6 = PyObject_IsInstance(__pyx_v_self, __pyx_t_5); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 407, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_3 = (__pyx_t_6 != 0);
  __pyx_t_4 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_4) {

    /* "pandas/_libs/interval.pyx":408
 *             return Interval(self.left * y, self.right * y, closed=self.closed)
 *         elif isinstance(y, Interval) and isinstance(self, numbers.Number):
 *             return Interval(y.left * self, y.right * self, closed=y.closed)             # <<<<<<<<<<<<<<
 *         return NotImplemented
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_y, __pyx_n_s_left); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 408, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = PyNumber_Multiply(__pyx_t_5, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 408, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_y, __pyx_n_s_right); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 408, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_2 = PyNumber_Multiply(__pyx_t_5, __pyx_v_self); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 408, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 408, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_2);
    __pyx_t_1 = 0;
    __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 408, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_y, __pyx_n_s_closed); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 408, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_closed, __pyx_t_1) < 0) __PYX_ERR(0, 408, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6pandas_5_libs_8interval_Interval), __pyx_t_5, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 408, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "pandas/_libs/interval.pyx":407
 *         if isinstance(y, numbers.Number):
 *             return Interval(self.left * y, self.right * y, closed=self.closed)
 *         elif isinstance(y, Interval) and isinstance(self, numbers.Number):             # <<<<<<<<<<<<<<
 *             return Interval(y.left * self, y.right * self, closed=y.closed)
 *         return NotImplemented
 */
  }

  /* "pandas/_libs/interval.pyx":409
 *         elif isinstance(y, Interval) and isinstance(self, numbers.Number):
 *             return Interval(y.left * self, y.right * self, closed=y.closed)
 *         return NotImplemented             # <<<<<<<<<<<<<<
 * 
 *     def __div__(self, y):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_builtin_NotImplemented);
  __pyx_r = __pyx_builtin_NotImplemented;
  goto __pyx_L0;

  /* "pandas/_libs/interval.pyx":404
 *         return NotImplemented
 * 
 *     def __mul__(self, y):             # <<<<<<<<<<<<<<
 *         if isinstance(y, numbers.Number):
 *             return Interval(self.left * y, self.right * y, closed=self.closed)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pandas._libs.interval.Interval.__mul__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/interval.pyx":411
 *         return NotImplemented
 * 
 *     def __div__(self, y):             # <<<<<<<<<<<<<<
 *         if isinstance(y, numbers.Number):
 *             return Interval(self.left / y, self.right / y, closed=self.closed)
 */

/* Python wrapper */
#if PY_MAJOR_VERSION < 3 || (CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x03050000)
static PyObject *__pyx_pw_6pandas_5_libs_8interval_8Interval_25__div__(PyObject *__pyx_v_self, PyObject *__pyx_v_y); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_8Interval_25__div__(PyObject *__pyx_v_self, PyObject *__pyx_v_y) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__div__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_8Interval_24__div__(((PyObject *)__pyx_v_self), ((PyObject *)__pyx_v_y));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
#endif /*!(#if PY_MAJOR_VERSION < 3 || (CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x03050000))*/

#if PY_MAJOR_VERSION < 3 || (CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x03050000)
static PyObject *__pyx_pf_6pandas_5_libs_8interval_8Interval_24__div__(PyObject *__pyx_v_self, PyObject *__pyx_v_y) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__div__", 0);

  /* "pandas/_libs/interval.pyx":412
 * 
 *     def __div__(self, y):
 *         if isinstance(y, numbers.Number):             # <<<<<<<<<<<<<<
 *             return Interval(self.left / y, self.right / y, closed=self.closed)
 *         return NotImplemented
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_numbers); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 412, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_Number); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 412, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = PyObject_IsInstance(__pyx_v_y, __pyx_t_2); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 412, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = (__pyx_t_3 != 0);
  if (__pyx_t_4) {

    /* "pandas/_libs/interval.pyx":413
 *     def __div__(self, y):
 *         if isinstance(y, numbers.Number):
 *             return Interval(self.left / y, self.right / y, closed=self.closed)             # <<<<<<<<<<<<<<
 *         return NotImplemented
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_left); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 413, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = __Pyx_PyNumber_Divide(__pyx_t_2, __pyx_v_y); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 413, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_right); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 413, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = __Pyx_PyNumber_Divide(__pyx_t_2, __pyx_v_y); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 413, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 413, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_5);
    __pyx_t_1 = 0;
    __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 413, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_closed); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 413, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_closed, __pyx_t_1) < 0) __PYX_ERR(0, 413, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6pandas_5_libs_8interval_Interval), __pyx_t_2, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 413, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "pandas/_libs/interval.pyx":412
 * 
 *     def __div__(self, y):
 *         if isinstance(y, numbers.Number):             # <<<<<<<<<<<<<<
 *             return Interval(self.left / y, self.right / y, closed=self.closed)
 *         return NotImplemented
 */
  }

  /* "pandas/_libs/interval.pyx":414
 *         if isinstance(y, numbers.Number):
 *             return Interval(self.left / y, self.right / y, closed=self.closed)
 *         return NotImplemented             # <<<<<<<<<<<<<<
 * 
 *     def __truediv__(self, y):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_builtin_NotImplemented);
  __pyx_r = __pyx_builtin_NotImplemented;
  goto __pyx_L0;

  /* "pandas/_libs/interval.pyx":411
 *         return NotImplemented
 * 
 *     def __div__(self, y):             # <<<<<<<<<<<<<<
 *         if isinstance(y, numbers.Number):
 *             return Interval(self.left / y, self.right / y, closed=self.closed)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pandas._libs.interval.Interval.__div__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
#endif /*!(#if PY_MAJOR_VERSION < 3 || (CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x03050000))*/

/* "pandas/_libs/interval.pyx":416
 *         return NotImplemented
 * 
 *     def __truediv__(self, y):             # <<<<<<<<<<<<<<
 *         if isinstance(y, numbers.Number):
 *             return Interval(self.left / y, self.right / y, closed=self.closed)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_8Interval_27__truediv__(PyObject *__pyx_v_self, PyObject *__pyx_v_y); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_8Interval_27__truediv__(PyObject *__pyx_v_self, PyObject *__pyx_v_y) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__truediv__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_8Interval_26__truediv__(((PyObject *)__pyx_v_self), ((PyObject *)__pyx_v_y));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_8Interval_26__truediv__(PyObject *__pyx_v_self, PyObject *__pyx_v_y) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__truediv__", 0);

  /* "pandas/_libs/interval.pyx":417
 * 
 *     def __truediv__(self, y):
 *         if isinstance(y, numbers.Number):             # <<<<<<<<<<<<<<
 *             return Interval(self.left / y, self.right / y, closed=self.closed)
 *         return NotImplemented
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_numbers); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 417, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_Number); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 417, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = PyObject_IsInstance(__pyx_v_y, __pyx_t_2); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 417, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = (__pyx_t_3 != 0);
  if (__pyx_t_4) {

    /* "pandas/_libs/interval.pyx":418
 *     def __truediv__(self, y):
 *         if isinstance(y, numbers.Number):
 *             return Interval(self.left / y, self.right / y, closed=self.closed)             # <<<<<<<<<<<<<<
 *         return NotImplemented
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_left); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 418, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = __Pyx_PyNumber_Divide(__pyx_t_2, __pyx_v_y); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 418, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_right); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 418, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = __Pyx_PyNumber_Divide(__pyx_t_2, __pyx_v_y); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 418, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 418, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_5);
    __pyx_t_1 = 0;
    __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 418, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_closed); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 418, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_closed, __pyx_t_1) < 0) __PYX_ERR(0, 418, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6pandas_5_libs_8interval_Interval), __pyx_t_2, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 418, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "pandas/_libs/interval.pyx":417
 * 
 *     def __truediv__(self, y):
 *         if isinstance(y, numbers.Number):             # <<<<<<<<<<<<<<
 *             return Interval(self.left / y, self.right / y, closed=self.closed)
 *         return NotImplemented
 */
  }

  /* "pandas/_libs/interval.pyx":419
 *         if isinstance(y, numbers.Number):
 *             return Interval(self.left / y, self.right / y, closed=self.closed)
 *         return NotImplemented             # <<<<<<<<<<<<<<
 * 
 *     def __floordiv__(self, y):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_builtin_NotImplemented);
  __pyx_r = __pyx_builtin_NotImplemented;
  goto __pyx_L0;

  /* "pandas/_libs/interval.pyx":416
 *         return NotImplemented
 * 
 *     def __truediv__(self, y):             # <<<<<<<<<<<<<<
 *         if isinstance(y, numbers.Number):
 *             return Interval(self.left / y, self.right / y, closed=self.closed)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pandas._libs.interval.Interval.__truediv__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/interval.pyx":421
 *         return NotImplemented
 * 
 *     def __floordiv__(self, y):             # <<<<<<<<<<<<<<
 *         if isinstance(y, numbers.Number):
 *             return Interval(
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_8Interval_29__floordiv__(PyObject *__pyx_v_self, PyObject *__pyx_v_y); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_8Interval_29__floordiv__(PyObject *__pyx_v_self, PyObject *__pyx_v_y) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__floordiv__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_8Interval_28__floordiv__(((PyObject *)__pyx_v_self), ((PyObject *)__pyx_v_y));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_8Interval_28__floordiv__(PyObject *__pyx_v_self, PyObject *__pyx_v_y) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__floordiv__", 0);

  /* "pandas/_libs/interval.pyx":422
 * 
 *     def __floordiv__(self, y):
 *         if isinstance(y, numbers.Number):             # <<<<<<<<<<<<<<
 *             return Interval(
 *                 self.left // y, self.right // y, closed=self.closed)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_numbers); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 422, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_Number); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 422, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = PyObject_IsInstance(__pyx_v_y, __pyx_t_2); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 422, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = (__pyx_t_3 != 0);
  if (__pyx_t_4) {

    /* "pandas/_libs/interval.pyx":423
 *     def __floordiv__(self, y):
 *         if isinstance(y, numbers.Number):
 *             return Interval(             # <<<<<<<<<<<<<<
 *                 self.left // y, self.right // y, closed=self.closed)
 *         return NotImplemented
 */
    __Pyx_XDECREF(__pyx_r);

    /* "pandas/_libs/interval.pyx":424
 *         if isinstance(y, numbers.Number):
 *             return Interval(
 *                 self.left // y, self.right // y, closed=self.closed)             # <<<<<<<<<<<<<<
 *         return NotImplemented
 * 
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_left); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 424, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = PyNumber_FloorDivide(__pyx_t_2, __pyx_v_y); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 424, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_right); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 424, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = PyNumber_FloorDivide(__pyx_t_2, __pyx_v_y); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 424, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "pandas/_libs/interval.pyx":423
 *     def __floordiv__(self, y):
 *         if isinstance(y, numbers.Number):
 *             return Interval(             # <<<<<<<<<<<<<<
 *                 self.left // y, self.right // y, closed=self.closed)
 *         return NotImplemented
 */
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 423, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_5);
    __pyx_t_1 = 0;
    __pyx_t_5 = 0;

    /* "pandas/_libs/interval.pyx":424
 *         if isinstance(y, numbers.Number):
 *             return Interval(
 *                 self.left // y, self.right // y, closed=self.closed)             # <<<<<<<<<<<<<<
 *         return NotImplemented
 * 
 */
    __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 424, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_closed); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 424, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_closed, __pyx_t_1) < 0) __PYX_ERR(0, 424, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "pandas/_libs/interval.pyx":423
 *     def __floordiv__(self, y):
 *         if isinstance(y, numbers.Number):
 *             return Interval(             # <<<<<<<<<<<<<<
 *                 self.left // y, self.right // y, closed=self.closed)
 *         return NotImplemented
 */
    __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6pandas_5_libs_8interval_Interval), __pyx_t_2, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 423, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "pandas/_libs/interval.pyx":422
 * 
 *     def __floordiv__(self, y):
 *         if isinstance(y, numbers.Number):             # <<<<<<<<<<<<<<
 *             return Interval(
 *                 self.left // y, self.right // y, closed=self.closed)
 */
  }

  /* "pandas/_libs/interval.pyx":425
 *             return Interval(
 *                 self.left // y, self.right // y, closed=self.closed)
 *         return NotImplemented             # <<<<<<<<<<<<<<
 * 
 *     def overlaps(self, other):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_builtin_NotImplemented);
  __pyx_r = __pyx_builtin_NotImplemented;
  goto __pyx_L0;

  /* "pandas/_libs/interval.pyx":421
 *         return NotImplemented
 * 
 *     def __floordiv__(self, y):             # <<<<<<<<<<<<<<
 *         if isinstance(y, numbers.Number):
 *             return Interval(
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pandas._libs.interval.Interval.__floordiv__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/interval.pyx":427
 *         return NotImplemented
 * 
 *     def overlaps(self, other):             # <<<<<<<<<<<<<<
 *         """
 *         Check whether two Interval objects overlap.
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_8Interval_31overlaps(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static char __pyx_doc_6pandas_5_libs_8interval_8Interval_30overlaps[] = "\n        Check whether two Interval objects overlap.\n\n        Two intervals overlap if they share a common point, including closed\n        endpoints. Intervals that only have an open endpoint in common do not\n        overlap.\n\n        .. versionadded:: 0.24.0\n\n        Parameters\n        ----------\n        other : Interval\n            The interval to check against for an overlap.\n\n        Returns\n        -------\n        bool\n            ``True`` if the two intervals overlap, else ``False``.\n\n        See Also\n        --------\n        IntervalArray.overlaps : The corresponding method for IntervalArray.\n        IntervalIndex.overlaps : The corresponding method for IntervalIndex.\n\n        Examples\n        --------\n        >>> i1 = pd.Interval(0, 2)\n        >>> i2 = pd.Interval(1, 3)\n        >>> i1.overlaps(i2)\n        True\n        >>> i3 = pd.Interval(4, 5)\n        >>> i1.overlaps(i3)\n        False\n\n        Intervals that share closed endpoints overlap:\n\n        >>> i4 = pd.Interval(0, 1, closed='both')\n        >>> i5 = pd.Interval(1, 2, closed='both')\n        >>> i4.overlaps(i5)\n        True\n\n        Intervals that only have an open endpoint in common do not overlap:\n\n        >>> i6 = pd.Interval(1, 2, closed='neither')\n        >>> i4.overlaps(i6)\n        False\n        ";
static PyObject *__pyx_pw_6pandas_5_libs_8interval_8Interval_31overlaps(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("overlaps (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_8Interval_30overlaps(((struct __pyx_obj_6pandas_5_libs_8interval_Interval *)__pyx_v_self), ((PyObject *)__pyx_v_other));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_8Interval_30overlaps(struct __pyx_obj_6pandas_5_libs_8interval_Interval *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_v_msg = NULL;
  PyObject *__pyx_v_op1 = NULL;
  PyObject *__pyx_v_op2 = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  __Pyx_RefNannySetupContext("overlaps", 0);

  /* "pandas/_libs/interval.pyx":475
 *         False
 *         """
 *         if not isinstance(other, Interval):             # <<<<<<<<<<<<<<
 *             msg = '`other` must be an Interval, got {other}'
 *             raise TypeError(msg.format(other=type(other).__name__))
 */
  __pyx_t_1 = __Pyx_TypeCheck(__pyx_v_other, __pyx_ptype_6pandas_5_libs_8interval_Interval); 
  __pyx_t_2 = ((!(__pyx_t_1 != 0)) != 0);
  if (unlikely(__pyx_t_2)) {

    /* "pandas/_libs/interval.pyx":476
 *         """
 *         if not isinstance(other, Interval):
 *             msg = '`other` must be an Interval, got {other}'             # <<<<<<<<<<<<<<
 *             raise TypeError(msg.format(other=type(other).__name__))
 * 
 */
    __Pyx_INCREF(__pyx_kp_u_other_must_be_an_Interval_got_o);
    __pyx_v_msg = __pyx_kp_u_other_must_be_an_Interval_got_o;

    /* "pandas/_libs/interval.pyx":477
 *         if not isinstance(other, Interval):
 *             msg = '`other` must be an Interval, got {other}'
 *             raise TypeError(msg.format(other=type(other).__name__))             # <<<<<<<<<<<<<<
 * 
 *         # equality is okay if both endpoints are closed (overlap at a point)
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_msg, __pyx_n_s_format); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 477, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 477, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)Py_TYPE(__pyx_v_other)), __pyx_n_s_name_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 477, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_other, __pyx_t_5) < 0) __PYX_ERR(0, 477, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_empty_tuple, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 477, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 477, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __PYX_ERR(0, 477, __pyx_L1_error)

    /* "pandas/_libs/interval.pyx":475
 *         False
 *         """
 *         if not isinstance(other, Interval):             # <<<<<<<<<<<<<<
 *             msg = '`other` must be an Interval, got {other}'
 *             raise TypeError(msg.format(other=type(other).__name__))
 */
  }

  /* "pandas/_libs/interval.pyx":480
 * 
 *         # equality is okay if both endpoints are closed (overlap at a point)
 *         op1 = le if (self.closed_left and other.closed_right) else lt             # <<<<<<<<<<<<<<
 *         op2 = le if (other.closed_left and self.closed_right) else lt
 * 
 */
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_closed_left); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 480, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 480, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__pyx_t_1) {
  } else {
    __pyx_t_2 = __pyx_t_1;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_other, __pyx_n_s_closed_right); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 480, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 480, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_2 = __pyx_t_1;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_le); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 480, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_4 = __pyx_t_5;
    __pyx_t_5 = 0;
  } else {
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_lt); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 480, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_4 = __pyx_t_5;
    __pyx_t_5 = 0;
  }
  __pyx_v_op1 = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "pandas/_libs/interval.pyx":481
 *         # equality is okay if both endpoints are closed (overlap at a point)
 *         op1 = le if (self.closed_left and other.closed_right) else lt
 *         op2 = le if (other.closed_left and self.closed_right) else lt             # <<<<<<<<<<<<<<
 * 
 *         # overlaps is equivalent negation of two interval being disjoint:
 */
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_other, __pyx_n_s_closed_left); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 481, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 481, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__pyx_t_1) {
  } else {
    __pyx_t_2 = __pyx_t_1;
    goto __pyx_L6_bool_binop_done;
  }
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_closed_right); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 481, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 481, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_2 = __pyx_t_1;
  __pyx_L6_bool_binop_done:;
  if (__pyx_t_2) {
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_le); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 481, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_4 = __pyx_t_5;
    __pyx_t_5 = 0;
  } else {
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_lt); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 481, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_4 = __pyx_t_5;
    __pyx_t_5 = 0;
  }
  __pyx_v_op2 = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "pandas/_libs/interval.pyx":486
 *         # disjoint = (A.left > B.right) or (B.left > A.right)
 *         # (simplifying the negation allows this to be done in less operations)
 *         return op1(self.left, other.right) and op2(other.left, self.right)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_other, __pyx_n_s_right); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 486, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_v_op1);
  __pyx_t_6 = __pyx_v_op1; __pyx_t_7 = NULL;
  __pyx_t_8 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_6, function);
      __pyx_t_8 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_6)) {
    PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_v_self->left, __pyx_t_3};
    __pyx_t_5 = __Pyx_PyFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 486, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_6)) {
    PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_v_self->left, __pyx_t_3};
    __pyx_t_5 = __Pyx_PyCFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 486, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else
  #endif
  {
    __pyx_t_9 = PyTuple_New(2+__pyx_t_8); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 486, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    if (__pyx_t_7) {
      __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_7); __pyx_t_7 = NULL;
    }
    __Pyx_INCREF(__pyx_v_self->left);
    __Pyx_GIVEREF(__pyx_v_self->left);
    PyTuple_SET_ITEM(__pyx_t_9, 0+__pyx_t_8, __pyx_v_self->left);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_9, 1+__pyx_t_8, __pyx_t_3);
    __pyx_t_3 = 0;
    __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_9, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 486, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  }
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 486, __pyx_L1_error)
  if (__pyx_t_2) {
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  } else {
    __Pyx_INCREF(__pyx_t_5);
    __pyx_t_4 = __pyx_t_5;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    goto __pyx_L8_bool_binop_done;
  }
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_other, __pyx_n_s_left); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 486, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_INCREF(__pyx_v_op2);
  __pyx_t_9 = __pyx_v_op2; __pyx_t_3 = NULL;
  __pyx_t_8 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_9))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_9);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_9, function);
      __pyx_t_8 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_9)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_6, __pyx_v_self->right};
    __pyx_t_5 = __Pyx_PyFunction_FastCall(__pyx_t_9, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 486, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_9)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_6, __pyx_v_self->right};
    __pyx_t_5 = __Pyx_PyCFunction_FastCall(__pyx_t_9, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 486, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  } else
  #endif
  {
    __pyx_t_7 = PyTuple_New(2+__pyx_t_8); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 486, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_8, __pyx_t_6);
    __Pyx_INCREF(__pyx_v_self->right);
    __Pyx_GIVEREF(__pyx_v_self->right);
    PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_8, __pyx_v_self->right);
    __pyx_t_6 = 0;
    __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_9, __pyx_t_7, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 486, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_INCREF(__pyx_t_5);
  __pyx_t_4 = __pyx_t_5;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_L8_bool_binop_done:;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "pandas/_libs/interval.pyx":427
 *         return NotImplemented
 * 
 *     def overlaps(self, other):             # <<<<<<<<<<<<<<
 *         """
 *         Check whether two Interval objects overlap.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("pandas._libs.interval.Interval.overlaps", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_msg);
  __Pyx_XDECREF(__pyx_v_op1);
  __Pyx_XDECREF(__pyx_v_op2);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/interval.pyx":287
 *     _typ = "interval"
 * 
 *     cdef readonly object left             # <<<<<<<<<<<<<<
 *     """Left bound for the interval"""
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_8Interval_4left_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_8Interval_4left_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_8Interval_4left___get__(((struct __pyx_obj_6pandas_5_libs_8interval_Interval *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_8Interval_4left___get__(struct __pyx_obj_6pandas_5_libs_8interval_Interval *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->left);
  __pyx_r = __pyx_v_self->left;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/interval.pyx":290
 *     """Left bound for the interval"""
 * 
 *     cdef readonly object right             # <<<<<<<<<<<<<<
 *     """Right bound for the interval"""
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_8Interval_5right_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_8Interval_5right_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_8Interval_5right___get__(((struct __pyx_obj_6pandas_5_libs_8interval_Interval *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_8Interval_5right___get__(struct __pyx_obj_6pandas_5_libs_8interval_Interval *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->right);
  __pyx_r = __pyx_v_self->right;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/interval.pyx":293
 *     """Right bound for the interval"""
 * 
 *     cdef readonly str closed             # <<<<<<<<<<<<<<
 *     """
 *     Whether the interval is closed on the left-side, right-side, both or
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_8Interval_6closed_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_8Interval_6closed_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_8Interval_6closed___get__(((struct __pyx_obj_6pandas_5_libs_8interval_Interval *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_8Interval_6closed___get__(struct __pyx_obj_6pandas_5_libs_8interval_Interval *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->closed);
  __pyx_r = __pyx_v_self->closed;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/interval.pyx":491
 * @cython.wraparound(False)
 * @cython.boundscheck(False)
 * def intervals_to_interval_bounds(ndarray intervals,             # <<<<<<<<<<<<<<
 *                                  bint validate_closed=True):
 *     """
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_1intervals_to_interval_bounds(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6pandas_5_libs_8interval_intervals_to_interval_bounds[] = "\n    Parameters\n    ----------\n    intervals : ndarray\n        object array of Intervals / nulls\n\n    validate_closed: boolean, default True\n        boolean indicating if all intervals must be closed on the same side.\n        Mismatching closed will raise if True, else return None for closed.\n\n    Returns\n    -------\n    tuples (left: ndarray object array,\n            right: ndarray object array,\n            closed: str)\n\n    ";
static PyMethodDef __pyx_mdef_6pandas_5_libs_8interval_1intervals_to_interval_bounds = {"intervals_to_interval_bounds", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_1intervals_to_interval_bounds, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_intervals_to_interval_bounds};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_1intervals_to_interval_bounds(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyArrayObject *__pyx_v_intervals = 0;
  int __pyx_v_validate_closed;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("intervals_to_interval_bounds (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_intervals,&__pyx_n_s_validate_closed,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_intervals)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_validate_closed);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "intervals_to_interval_bounds") < 0)) __PYX_ERR(0, 491, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_intervals = ((PyArrayObject *)values[0]);
    if (values[1]) {
      __pyx_v_validate_closed = __Pyx_PyObject_IsTrue(values[1]); if (unlikely((__pyx_v_validate_closed == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 492, __pyx_L3_error)
    } else {

      /* "pandas/_libs/interval.pyx":492
 * @cython.boundscheck(False)
 * def intervals_to_interval_bounds(ndarray intervals,
 *                                  bint validate_closed=True):             # <<<<<<<<<<<<<<
 *     """
 *     Parameters
 */
      __pyx_v_validate_closed = ((int)1);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("intervals_to_interval_bounds", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 491, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.intervals_to_interval_bounds", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_intervals), __pyx_ptype_5numpy_ndarray, 1, "intervals", 0))) __PYX_ERR(0, 491, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_intervals_to_interval_bounds(__pyx_self, __pyx_v_intervals, __pyx_v_validate_closed);

  /* "pandas/_libs/interval.pyx":491
 * @cython.wraparound(False)
 * @cython.boundscheck(False)
 * def intervals_to_interval_bounds(ndarray intervals,             # <<<<<<<<<<<<<<
 *                                  bint validate_closed=True):
 *     """
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_intervals_to_interval_bounds(CYTHON_UNUSED PyObject *__pyx_self, PyArrayObject *__pyx_v_intervals, int __pyx_v_validate_closed) {
  PyObject *__pyx_v_closed = 0;
  PyObject *__pyx_v_interval = 0;
  __pyx_t_5numpy_int64_t __pyx_v_n;
  PyArrayObject *__pyx_v_left = 0;
  PyArrayObject *__pyx_v_right = 0;
  int __pyx_v_seen_closed;
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_v_msg = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  Py_ssize_t __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  int __pyx_t_10;
  __Pyx_RefNannySetupContext("intervals_to_interval_bounds", 0);

  /* "pandas/_libs/interval.pyx":512
 * 
 *     cdef:
 *         object closed = None, interval             # <<<<<<<<<<<<<<
 *         int64_t n = len(intervals)
 *         ndarray left, right
 */
  __Pyx_INCREF(Py_None);
  __pyx_v_closed = Py_None;

  /* "pandas/_libs/interval.pyx":513
 *     cdef:
 *         object closed = None, interval
 *         int64_t n = len(intervals)             # <<<<<<<<<<<<<<
 *         ndarray left, right
 *         bint seen_closed = False
 */
  __pyx_t_1 = PyObject_Length(((PyObject *)__pyx_v_intervals)); if (unlikely(__pyx_t_1 == ((Py_ssize_t)-1))) __PYX_ERR(0, 513, __pyx_L1_error)
  __pyx_v_n = __pyx_t_1;

  /* "pandas/_libs/interval.pyx":515
 *         int64_t n = len(intervals)
 *         ndarray left, right
 *         bint seen_closed = False             # <<<<<<<<<<<<<<
 * 
 *     left = np.empty(n, dtype=intervals.dtype)
 */
  __pyx_v_seen_closed = 0;

  /* "pandas/_libs/interval.pyx":517
 *         bint seen_closed = False
 * 
 *     left = np.empty(n, dtype=intervals.dtype)             # <<<<<<<<<<<<<<
 *     right = np.empty(n, dtype=intervals.dtype)
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 517, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_empty); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 517, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyInt_From_npy_int64(__pyx_v_n); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 517, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 517, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 517, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_intervals), __pyx_n_s_dtype); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 517, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_dtype, __pyx_t_5) < 0) __PYX_ERR(0, 517, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, __pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 517, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (!(likely(((__pyx_t_5) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_5, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 517, __pyx_L1_error)
  __pyx_v_left = ((PyArrayObject *)__pyx_t_5);
  __pyx_t_5 = 0;

  /* "pandas/_libs/interval.pyx":518
 * 
 *     left = np.empty(n, dtype=intervals.dtype)
 *     right = np.empty(n, dtype=intervals.dtype)             # <<<<<<<<<<<<<<
 * 
 *     for i in range(len(intervals)):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_np); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 518, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_empty); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 518, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyInt_From_npy_int64(__pyx_v_n); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 518, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 518, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5);
  __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 518, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_intervals), __pyx_n_s_dtype); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 518, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_dtype, __pyx_t_3) < 0) __PYX_ERR(0, 518, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_4, __pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 518, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 518, __pyx_L1_error)
  __pyx_v_right = ((PyArrayObject *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "pandas/_libs/interval.pyx":520
 *     right = np.empty(n, dtype=intervals.dtype)
 * 
 *     for i in range(len(intervals)):             # <<<<<<<<<<<<<<
 *         interval = intervals[i]
 *         if interval is None or util.is_nan(interval):
 */
  __pyx_t_1 = PyObject_Length(((PyObject *)__pyx_v_intervals)); if (unlikely(__pyx_t_1 == ((Py_ssize_t)-1))) __PYX_ERR(0, 520, __pyx_L1_error)
  __pyx_t_6 = __pyx_t_1;
  for (__pyx_t_7 = 0; __pyx_t_7 < __pyx_t_6; __pyx_t_7+=1) {
    __pyx_v_i = __pyx_t_7;

    /* "pandas/_libs/interval.pyx":521
 * 
 *     for i in range(len(intervals)):
 *         interval = intervals[i]             # <<<<<<<<<<<<<<
 *         if interval is None or util.is_nan(interval):
 *             left[i] = np.nan
 */
    __pyx_t_3 = __Pyx_GetItemInt(((PyObject *)__pyx_v_intervals), __pyx_v_i, Py_ssize_t, 1, PyInt_FromSsize_t, 0, 0, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 521, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_XDECREF_SET(__pyx_v_interval, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "pandas/_libs/interval.pyx":522
 *     for i in range(len(intervals)):
 *         interval = intervals[i]
 *         if interval is None or util.is_nan(interval):             # <<<<<<<<<<<<<<
 *             left[i] = np.nan
 *             right[i] = np.nan
 */
    __pyx_t_9 = (__pyx_v_interval == Py_None);
    __pyx_t_10 = (__pyx_t_9 != 0);
    if (!__pyx_t_10) {
    } else {
      __pyx_t_8 = __pyx_t_10;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_10 = (__pyx_f_6pandas_5_libs_6tslibs_4util_is_nan(__pyx_v_interval) != 0);
    __pyx_t_8 = __pyx_t_10;
    __pyx_L6_bool_binop_done:;
    if (__pyx_t_8) {

      /* "pandas/_libs/interval.pyx":523
 *         interval = intervals[i]
 *         if interval is None or util.is_nan(interval):
 *             left[i] = np.nan             # <<<<<<<<<<<<<<
 *             right[i] = np.nan
 *             continue
 */
      __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 523, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_nan); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 523, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(__Pyx_SetItemInt(((PyObject *)__pyx_v_left), __pyx_v_i, __pyx_t_5, Py_ssize_t, 1, PyInt_FromSsize_t, 0, 0, 0) < 0)) __PYX_ERR(0, 523, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "pandas/_libs/interval.pyx":524
 *         if interval is None or util.is_nan(interval):
 *             left[i] = np.nan
 *             right[i] = np.nan             # <<<<<<<<<<<<<<
 *             continue
 * 
 */
      __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_np); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 524, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_nan); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 524, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(__Pyx_SetItemInt(((PyObject *)__pyx_v_right), __pyx_v_i, __pyx_t_3, Py_ssize_t, 1, PyInt_FromSsize_t, 0, 0, 0) < 0)) __PYX_ERR(0, 524, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "pandas/_libs/interval.pyx":525
 *             left[i] = np.nan
 *             right[i] = np.nan
 *             continue             # <<<<<<<<<<<<<<
 * 
 *         if not isinstance(interval, Interval):
 */
      goto __pyx_L3_continue;

      /* "pandas/_libs/interval.pyx":522
 *     for i in range(len(intervals)):
 *         interval = intervals[i]
 *         if interval is None or util.is_nan(interval):             # <<<<<<<<<<<<<<
 *             left[i] = np.nan
 *             right[i] = np.nan
 */
    }

    /* "pandas/_libs/interval.pyx":527
 *             continue
 * 
 *         if not isinstance(interval, Interval):             # <<<<<<<<<<<<<<
 *             raise TypeError("type {typ} with value {iv} is not an interval"
 *                             .format(typ=type(interval), iv=interval))
 */
    __pyx_t_8 = __Pyx_TypeCheck(__pyx_v_interval, __pyx_ptype_6pandas_5_libs_8interval_Interval); 
    __pyx_t_10 = ((!(__pyx_t_8 != 0)) != 0);
    if (unlikely(__pyx_t_10)) {

      /* "pandas/_libs/interval.pyx":529
 *         if not isinstance(interval, Interval):
 *             raise TypeError("type {typ} with value {iv} is not an interval"
 *                             .format(typ=type(interval), iv=interval))             # <<<<<<<<<<<<<<
 * 
 *         left[i] = interval.left
 */
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_type_typ_with_value_iv_is_not_an, __pyx_n_s_format); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 529, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_5 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 529, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_typ, ((PyObject *)Py_TYPE(__pyx_v_interval))) < 0) __PYX_ERR(0, 529, __pyx_L1_error)
      if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_iv, __pyx_v_interval) < 0) __PYX_ERR(0, 529, __pyx_L1_error)
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_empty_tuple, __pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 529, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "pandas/_libs/interval.pyx":528
 * 
 *         if not isinstance(interval, Interval):
 *             raise TypeError("type {typ} with value {iv} is not an interval"             # <<<<<<<<<<<<<<
 *                             .format(typ=type(interval), iv=interval))
 * 
 */
      __pyx_t_5 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 528, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_Raise(__pyx_t_5, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __PYX_ERR(0, 528, __pyx_L1_error)

      /* "pandas/_libs/interval.pyx":527
 *             continue
 * 
 *         if not isinstance(interval, Interval):             # <<<<<<<<<<<<<<
 *             raise TypeError("type {typ} with value {iv} is not an interval"
 *                             .format(typ=type(interval), iv=interval))
 */
    }

    /* "pandas/_libs/interval.pyx":531
 *                             .format(typ=type(interval), iv=interval))
 * 
 *         left[i] = interval.left             # <<<<<<<<<<<<<<
 *         right[i] = interval.right
 *         if not seen_closed:
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_interval, __pyx_n_s_left); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 531, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (unlikely(__Pyx_SetItemInt(((PyObject *)__pyx_v_left), __pyx_v_i, __pyx_t_5, Py_ssize_t, 1, PyInt_FromSsize_t, 0, 0, 0) < 0)) __PYX_ERR(0, 531, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "pandas/_libs/interval.pyx":532
 * 
 *         left[i] = interval.left
 *         right[i] = interval.right             # <<<<<<<<<<<<<<
 *         if not seen_closed:
 *             seen_closed = True
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_interval, __pyx_n_s_right); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 532, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (unlikely(__Pyx_SetItemInt(((PyObject *)__pyx_v_right), __pyx_v_i, __pyx_t_5, Py_ssize_t, 1, PyInt_FromSsize_t, 0, 0, 0) < 0)) __PYX_ERR(0, 532, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "pandas/_libs/interval.pyx":533
 *         left[i] = interval.left
 *         right[i] = interval.right
 *         if not seen_closed:             # <<<<<<<<<<<<<<
 *             seen_closed = True
 *             closed = interval.closed
 */
    __pyx_t_10 = ((!(__pyx_v_seen_closed != 0)) != 0);
    if (__pyx_t_10) {

      /* "pandas/_libs/interval.pyx":534
 *         right[i] = interval.right
 *         if not seen_closed:
 *             seen_closed = True             # <<<<<<<<<<<<<<
 *             closed = interval.closed
 *         elif closed != interval.closed:
 */
      __pyx_v_seen_closed = 1;

      /* "pandas/_libs/interval.pyx":535
 *         if not seen_closed:
 *             seen_closed = True
 *             closed = interval.closed             # <<<<<<<<<<<<<<
 *         elif closed != interval.closed:
 *             closed = None
 */
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_interval, __pyx_n_s_closed); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 535, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF_SET(__pyx_v_closed, __pyx_t_5);
      __pyx_t_5 = 0;

      /* "pandas/_libs/interval.pyx":533
 *         left[i] = interval.left
 *         right[i] = interval.right
 *         if not seen_closed:             # <<<<<<<<<<<<<<
 *             seen_closed = True
 *             closed = interval.closed
 */
      goto __pyx_L9;
    }

    /* "pandas/_libs/interval.pyx":536
 *             seen_closed = True
 *             closed = interval.closed
 *         elif closed != interval.closed:             # <<<<<<<<<<<<<<
 *             closed = None
 *             if validate_closed:
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_interval, __pyx_n_s_closed); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 536, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_4 = PyObject_RichCompare(__pyx_v_closed, __pyx_t_5, Py_NE); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 536, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_10 < 0)) __PYX_ERR(0, 536, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (__pyx_t_10) {

      /* "pandas/_libs/interval.pyx":537
 *             closed = interval.closed
 *         elif closed != interval.closed:
 *             closed = None             # <<<<<<<<<<<<<<
 *             if validate_closed:
 *                 msg = 'intervals must all be closed on the same side'
 */
      __Pyx_INCREF(Py_None);
      __Pyx_DECREF_SET(__pyx_v_closed, Py_None);

      /* "pandas/_libs/interval.pyx":538
 *         elif closed != interval.closed:
 *             closed = None
 *             if validate_closed:             # <<<<<<<<<<<<<<
 *                 msg = 'intervals must all be closed on the same side'
 *                 raise ValueError(msg)
 */
      __pyx_t_10 = (__pyx_v_validate_closed != 0);
      if (unlikely(__pyx_t_10)) {

        /* "pandas/_libs/interval.pyx":539
 *             closed = None
 *             if validate_closed:
 *                 msg = 'intervals must all be closed on the same side'             # <<<<<<<<<<<<<<
 *                 raise ValueError(msg)
 * 
 */
        __Pyx_INCREF(__pyx_kp_u_intervals_must_all_be_closed_on);
        __pyx_v_msg = __pyx_kp_u_intervals_must_all_be_closed_on;

        /* "pandas/_libs/interval.pyx":540
 *             if validate_closed:
 *                 msg = 'intervals must all be closed on the same side'
 *                 raise ValueError(msg)             # <<<<<<<<<<<<<<
 * 
 *     return left, right, closed
 */
        __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_v_msg); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 540, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_Raise(__pyx_t_4, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __PYX_ERR(0, 540, __pyx_L1_error)

        /* "pandas/_libs/interval.pyx":538
 *         elif closed != interval.closed:
 *             closed = None
 *             if validate_closed:             # <<<<<<<<<<<<<<
 *                 msg = 'intervals must all be closed on the same side'
 *                 raise ValueError(msg)
 */
      }

      /* "pandas/_libs/interval.pyx":536
 *             seen_closed = True
 *             closed = interval.closed
 *         elif closed != interval.closed:             # <<<<<<<<<<<<<<
 *             closed = None
 *             if validate_closed:
 */
    }
    __pyx_L9:;
    __pyx_L3_continue:;
  }

  /* "pandas/_libs/interval.pyx":542
 *                 raise ValueError(msg)
 * 
 *     return left, right, closed             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 542, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(((PyObject *)__pyx_v_left));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_left));
  PyTuple_SET_ITEM(__pyx_t_4, 0, ((PyObject *)__pyx_v_left));
  __Pyx_INCREF(((PyObject *)__pyx_v_right));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_right));
  PyTuple_SET_ITEM(__pyx_t_4, 1, ((PyObject *)__pyx_v_right));
  __Pyx_INCREF(__pyx_v_closed);
  __Pyx_GIVEREF(__pyx_v_closed);
  PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_v_closed);
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "pandas/_libs/interval.pyx":491
 * @cython.wraparound(False)
 * @cython.boundscheck(False)
 * def intervals_to_interval_bounds(ndarray intervals,             # <<<<<<<<<<<<<<
 *                                  bint validate_closed=True):
 *     """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pandas._libs.interval.intervals_to_interval_bounds", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_closed);
  __Pyx_XDECREF(__pyx_v_interval);
  __Pyx_XDECREF((PyObject *)__pyx_v_left);
  __Pyx_XDECREF((PyObject *)__pyx_v_right);
  __Pyx_XDECREF(__pyx_v_msg);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":33
 *         object _is_overlapping, _left_sorter, _right_sorter
 * 
 *     def __init__(self, left, right, closed='right', leaf_size=100):             # <<<<<<<<<<<<<<
 *         """
 *         Parameters
 */

/* Python wrapper */
static int __pyx_pw_6pandas_5_libs_8interval_12IntervalTree_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6pandas_5_libs_8interval_12IntervalTree___init__[] = "\n        Parameters\n        ----------\n        left, right : np.ndarray[ndim=1]\n            Left and right bounds for each interval. Assumed to contain no\n            NaNs.\n        closed : {'left', 'right', 'both', 'neither'}, optional\n            Whether the intervals are closed on the left-side, right-side, both\n            or neither. Defaults to 'right'.\n        leaf_size : int, optional\n            Parameter that controls when the tree switches from creating nodes\n            to brute-force search. Tune this parameter to optimize query\n            performance.\n        ";
#if CYTHON_COMPILING_IN_CPYTHON
struct wrapperbase __pyx_wrapperbase_6pandas_5_libs_8interval_12IntervalTree___init__;
#endif
static int __pyx_pw_6pandas_5_libs_8interval_12IntervalTree_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_left = 0;
  PyObject *__pyx_v_right = 0;
  PyObject *__pyx_v_closed = 0;
  PyObject *__pyx_v_leaf_size = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_left,&__pyx_n_s_right,&__pyx_n_s_closed,&__pyx_n_s_leaf_size,0};
    PyObject* values[4] = {0,0,0,0};
    values[2] = ((PyObject *)__pyx_n_u_right);
    values[3] = ((PyObject *)__pyx_int_100);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_left)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_right)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, 1); __PYX_ERR(1, 33, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_closed);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_leaf_size);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(1, 33, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_left = values[0];
    __pyx_v_right = values[1];
    __pyx_v_closed = values[2];
    __pyx_v_leaf_size = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 33, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.IntervalTree.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_12IntervalTree___init__(((struct __pyx_obj_6pandas_5_libs_8interval_IntervalTree *)__pyx_v_self), __pyx_v_left, __pyx_v_right, __pyx_v_closed, __pyx_v_leaf_size);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6pandas_5_libs_8interval_12IntervalTree___init__(struct __pyx_obj_6pandas_5_libs_8interval_IntervalTree *__pyx_v_self, PyObject *__pyx_v_left, PyObject *__pyx_v_right, PyObject *__pyx_v_closed, PyObject *__pyx_v_leaf_size) {
  PyObject *__pyx_v_indices = NULL;
  PyObject *__pyx_v_mask = NULL;
  PyObject *__pyx_v_node_cls = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  Py_ssize_t __pyx_t_8;
  __Pyx_RefNannySetupContext("__init__", 0);
  __Pyx_INCREF(__pyx_v_left);
  __Pyx_INCREF(__pyx_v_right);

  /* "pandas/_libs/intervaltree.pxi":48
 *             performance.
 *         """
 *         if closed not in ['left', 'right', 'both', 'neither']:             # <<<<<<<<<<<<<<
 *             raise ValueError("invalid option for 'closed': %s" % closed)
 * 
 */
  __Pyx_INCREF(__pyx_v_closed);
  __pyx_t_1 = __pyx_v_closed;
  __pyx_t_3 = (__Pyx_PyUnicode_Equals(__pyx_t_1, __pyx_n_u_left, Py_NE)); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(1, 48, __pyx_L1_error)
  if (__pyx_t_3) {
  } else {
    __pyx_t_2 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = (__Pyx_PyUnicode_Equals(__pyx_t_1, __pyx_n_u_right, Py_NE)); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(1, 48, __pyx_L1_error)
  if (__pyx_t_3) {
  } else {
    __pyx_t_2 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = (__Pyx_PyUnicode_Equals(__pyx_t_1, __pyx_n_u_both, Py_NE)); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(1, 48, __pyx_L1_error)
  if (__pyx_t_3) {
  } else {
    __pyx_t_2 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = (__Pyx_PyUnicode_Equals(__pyx_t_1, __pyx_n_u_neither, Py_NE)); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(1, 48, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (unlikely(__pyx_t_3)) {

    /* "pandas/_libs/intervaltree.pxi":49
 *         """
 *         if closed not in ['left', 'right', 'both', 'neither']:
 *             raise ValueError("invalid option for 'closed': %s" % closed)             # <<<<<<<<<<<<<<
 * 
 *         left = np.asarray(left)
 */
    __pyx_t_1 = __Pyx_PyUnicode_FormatSafe(__pyx_kp_u_invalid_option_for_closed_s, __pyx_v_closed); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 49, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 49, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __PYX_ERR(1, 49, __pyx_L1_error)

    /* "pandas/_libs/intervaltree.pxi":48
 *             performance.
 *         """
 *         if closed not in ['left', 'right', 'both', 'neither']:             # <<<<<<<<<<<<<<
 *             raise ValueError("invalid option for 'closed': %s" % closed)
 * 
 */
  }

  /* "pandas/_libs/intervaltree.pxi":51
 *             raise ValueError("invalid option for 'closed': %s" % closed)
 * 
 *         left = np.asarray(left)             # <<<<<<<<<<<<<<
 *         right = np.asarray(right)
 *         self.dtype = np.result_type(left, right)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 51, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_asarray); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 51, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  __pyx_t_4 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_1, __pyx_v_left) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_v_left);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 51, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF_SET(__pyx_v_left, __pyx_t_4);
  __pyx_t_4 = 0;

  /* "pandas/_libs/intervaltree.pxi":52
 * 
 *         left = np.asarray(left)
 *         right = np.asarray(right)             # <<<<<<<<<<<<<<
 *         self.dtype = np.result_type(left, right)
 *         self.left = np.asarray(left, dtype=self.dtype)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_np); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 52, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_asarray); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 52, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_4 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_5, __pyx_v_right) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_right);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 52, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF_SET(__pyx_v_right, __pyx_t_4);
  __pyx_t_4 = 0;

  /* "pandas/_libs/intervaltree.pxi":53
 *         left = np.asarray(left)
 *         right = np.asarray(right)
 *         self.dtype = np.result_type(left, right)             # <<<<<<<<<<<<<<
 *         self.left = np.asarray(left, dtype=self.dtype)
 *         self.right = np.asarray(right, dtype=self.dtype)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 53, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_result_type); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 53, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_5)) {
    PyObject *__pyx_temp[3] = {__pyx_t_1, __pyx_v_left, __pyx_v_right};
    __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 53, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_GOTREF(__pyx_t_4);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
    PyObject *__pyx_temp[3] = {__pyx_t_1, __pyx_v_left, __pyx_v_right};
    __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 53, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_GOTREF(__pyx_t_4);
  } else
  #endif
  {
    __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 53, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__pyx_t_1) {
      __Pyx_GIVEREF(__pyx_t_1); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_1); __pyx_t_1 = NULL;
    }
    __Pyx_INCREF(__pyx_v_left);
    __Pyx_GIVEREF(__pyx_v_left);
    PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_left);
    __Pyx_INCREF(__pyx_v_right);
    __Pyx_GIVEREF(__pyx_v_right);
    PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_v_right);
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_7, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 53, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_GIVEREF(__pyx_t_4);
  __Pyx_GOTREF(__pyx_v_self->dtype);
  __Pyx_DECREF(__pyx_v_self->dtype);
  __pyx_v_self->dtype = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "pandas/_libs/intervaltree.pxi":54
 *         right = np.asarray(right)
 *         self.dtype = np.result_type(left, right)
 *         self.left = np.asarray(left, dtype=self.dtype)             # <<<<<<<<<<<<<<
 *         self.right = np.asarray(right, dtype=self.dtype)
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 54, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_asarray); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 54, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 54, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(__pyx_v_left);
  __Pyx_GIVEREF(__pyx_v_left);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_left);
  __pyx_t_7 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 54, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_dtype, __pyx_v_self->dtype) < 0) __PYX_ERR(1, 54, __pyx_L1_error)
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_4, __pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 54, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->left);
  __Pyx_DECREF(__pyx_v_self->left);
  __pyx_v_self->left = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":55
 *         self.dtype = np.result_type(left, right)
 *         self.left = np.asarray(left, dtype=self.dtype)
 *         self.right = np.asarray(right, dtype=self.dtype)             # <<<<<<<<<<<<<<
 * 
 *         indices = np.arange(len(left), dtype='int64')
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 55, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_asarray); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 55, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 55, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_right);
  __Pyx_GIVEREF(__pyx_v_right);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_right);
  __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 55, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_dtype, __pyx_v_self->dtype) < 0) __PYX_ERR(1, 55, __pyx_L1_error)
  __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_1, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 55, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_GIVEREF(__pyx_t_5);
  __Pyx_GOTREF(__pyx_v_self->right);
  __Pyx_DECREF(__pyx_v_self->right);
  __pyx_v_self->right = __pyx_t_5;
  __pyx_t_5 = 0;

  /* "pandas/_libs/intervaltree.pxi":57
 *         self.right = np.asarray(right, dtype=self.dtype)
 * 
 *         indices = np.arange(len(left), dtype='int64')             # <<<<<<<<<<<<<<
 * 
 *         self.closed = closed
 */
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_np); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 57, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_arange); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 57, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_8 = PyObject_Length(__pyx_v_left); if (unlikely(__pyx_t_8 == ((Py_ssize_t)-1))) __PYX_ERR(1, 57, __pyx_L1_error)
  __pyx_t_5 = PyInt_FromSsize_t(__pyx_t_8); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 57, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 57, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_5);
  __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 57, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_dtype, __pyx_n_u_int64) < 0) __PYX_ERR(1, 57, __pyx_L1_error)
  __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_1, __pyx_t_5); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 57, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_v_indices = __pyx_t_7;
  __pyx_t_7 = 0;

  /* "pandas/_libs/intervaltree.pxi":59
 *         indices = np.arange(len(left), dtype='int64')
 * 
 *         self.closed = closed             # <<<<<<<<<<<<<<
 * 
 *         # GH 23352: ensure no nan in nodes
 */
  if (!(likely(PyUnicode_CheckExact(__pyx_v_closed))||((__pyx_v_closed) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "unicode", Py_TYPE(__pyx_v_closed)->tp_name), 0))) __PYX_ERR(1, 59, __pyx_L1_error)
  __pyx_t_7 = __pyx_v_closed;
  __Pyx_INCREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_7);
  __Pyx_GOTREF(__pyx_v_self->closed);
  __Pyx_DECREF(__pyx_v_self->closed);
  __pyx_v_self->closed = ((PyObject*)__pyx_t_7);
  __pyx_t_7 = 0;

  /* "pandas/_libs/intervaltree.pxi":62
 * 
 *         # GH 23352: ensure no nan in nodes
 *         mask = ~np.isnan(self.left)             # <<<<<<<<<<<<<<
 *         self.left = self.left[mask]
 *         self.right = self.right[mask]
 */
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_np); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 62, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_isnan); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 62, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_7 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_5, __pyx_v_self->left) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_self->left);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 62, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyNumber_Invert(__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 62, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_v_mask = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":63
 *         # GH 23352: ensure no nan in nodes
 *         mask = ~np.isnan(self.left)
 *         self.left = self.left[mask]             # <<<<<<<<<<<<<<
 *         self.right = self.right[mask]
 *         indices = indices[mask]
 */
  __pyx_t_1 = __Pyx_PyObject_GetItem(__pyx_v_self->left, __pyx_v_mask); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 63, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->left);
  __Pyx_DECREF(__pyx_v_self->left);
  __pyx_v_self->left = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":64
 *         mask = ~np.isnan(self.left)
 *         self.left = self.left[mask]
 *         self.right = self.right[mask]             # <<<<<<<<<<<<<<
 *         indices = indices[mask]
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetItem(__pyx_v_self->right, __pyx_v_mask); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 64, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->right);
  __Pyx_DECREF(__pyx_v_self->right);
  __pyx_v_self->right = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":65
 *         self.left = self.left[mask]
 *         self.right = self.right[mask]
 *         indices = indices[mask]             # <<<<<<<<<<<<<<
 * 
 *         node_cls = NODE_CLASSES[str(self.dtype), closed]
 */
  __pyx_t_1 = __Pyx_PyObject_GetItem(__pyx_v_indices, __pyx_v_mask); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 65, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF_SET(__pyx_v_indices, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":67
 *         indices = indices[mask]
 * 
 *         node_cls = NODE_CLASSES[str(self.dtype), closed]             # <<<<<<<<<<<<<<
 *         self.root = node_cls(self.left, self.right, indices, leaf_size)
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_NODE_CLASSES); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 67, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_7 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyUnicode_Type)), __pyx_v_self->dtype); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 67, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 67, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7);
  __Pyx_INCREF(__pyx_v_closed);
  __Pyx_GIVEREF(__pyx_v_closed);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_v_closed);
  __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_PyObject_GetItem(__pyx_t_1, __pyx_t_5); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 67, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_v_node_cls = __pyx_t_7;
  __pyx_t_7 = 0;

  /* "pandas/_libs/intervaltree.pxi":68
 * 
 *         node_cls = NODE_CLASSES[str(self.dtype), closed]
 *         self.root = node_cls(self.left, self.right, indices, leaf_size)             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_INCREF(__pyx_v_node_cls);
  __pyx_t_5 = __pyx_v_node_cls; __pyx_t_1 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_5)) {
    PyObject *__pyx_temp[5] = {__pyx_t_1, __pyx_v_self->left, __pyx_v_self->right, __pyx_v_indices, __pyx_v_leaf_size};
    __pyx_t_7 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 4+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 68, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_GOTREF(__pyx_t_7);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
    PyObject *__pyx_temp[5] = {__pyx_t_1, __pyx_v_self->left, __pyx_v_self->right, __pyx_v_indices, __pyx_v_leaf_size};
    __pyx_t_7 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 4+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 68, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_GOTREF(__pyx_t_7);
  } else
  #endif
  {
    __pyx_t_4 = PyTuple_New(4+__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 68, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (__pyx_t_1) {
      __Pyx_GIVEREF(__pyx_t_1); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1); __pyx_t_1 = NULL;
    }
    __Pyx_INCREF(__pyx_v_self->left);
    __Pyx_GIVEREF(__pyx_v_self->left);
    PyTuple_SET_ITEM(__pyx_t_4, 0+__pyx_t_6, __pyx_v_self->left);
    __Pyx_INCREF(__pyx_v_self->right);
    __Pyx_GIVEREF(__pyx_v_self->right);
    PyTuple_SET_ITEM(__pyx_t_4, 1+__pyx_t_6, __pyx_v_self->right);
    __Pyx_INCREF(__pyx_v_indices);
    __Pyx_GIVEREF(__pyx_v_indices);
    PyTuple_SET_ITEM(__pyx_t_4, 2+__pyx_t_6, __pyx_v_indices);
    __Pyx_INCREF(__pyx_v_leaf_size);
    __Pyx_GIVEREF(__pyx_v_leaf_size);
    PyTuple_SET_ITEM(__pyx_t_4, 3+__pyx_t_6, __pyx_v_leaf_size);
    __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_4, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 68, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_GIVEREF(__pyx_t_7);
  __Pyx_GOTREF(__pyx_v_self->root);
  __Pyx_DECREF(__pyx_v_self->root);
  __pyx_v_self->root = __pyx_t_7;
  __pyx_t_7 = 0;

  /* "pandas/_libs/intervaltree.pxi":33
 *         object _is_overlapping, _left_sorter, _right_sorter
 * 
 *     def __init__(self, left, right, closed='right', leaf_size=100):             # <<<<<<<<<<<<<<
 *         """
 *         Parameters
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("pandas._libs.interval.IntervalTree.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_indices);
  __Pyx_XDECREF(__pyx_v_mask);
  __Pyx_XDECREF(__pyx_v_node_cls);
  __Pyx_XDECREF(__pyx_v_left);
  __Pyx_XDECREF(__pyx_v_right);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":71
 * 
 *     @property
 *     def left_sorter(self):             # <<<<<<<<<<<<<<
 *         """How to sort the left labels; this is used for binary search
 *         """
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_12IntervalTree_11left_sorter_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_12IntervalTree_11left_sorter_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_12IntervalTree_11left_sorter___get__(((struct __pyx_obj_6pandas_5_libs_8interval_IntervalTree *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_12IntervalTree_11left_sorter___get__(struct __pyx_obj_6pandas_5_libs_8interval_IntervalTree *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pandas/_libs/intervaltree.pxi":74
 *         """How to sort the left labels; this is used for binary search
 *         """
 *         if self._left_sorter is None:             # <<<<<<<<<<<<<<
 *             self._left_sorter = np.argsort(self.left)
 *         return self._left_sorter
 */
  __pyx_t_1 = (__pyx_v_self->_left_sorter == Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pandas/_libs/intervaltree.pxi":75
 *         """
 *         if self._left_sorter is None:
 *             self._left_sorter = np.argsort(self.left)             # <<<<<<<<<<<<<<
 *         return self._left_sorter
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 75, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_argsort); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 75, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_4, __pyx_v_self->left) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_v_self->left);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 75, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GIVEREF(__pyx_t_3);
    __Pyx_GOTREF(__pyx_v_self->_left_sorter);
    __Pyx_DECREF(__pyx_v_self->_left_sorter);
    __pyx_v_self->_left_sorter = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "pandas/_libs/intervaltree.pxi":74
 *         """How to sort the left labels; this is used for binary search
 *         """
 *         if self._left_sorter is None:             # <<<<<<<<<<<<<<
 *             self._left_sorter = np.argsort(self.left)
 *         return self._left_sorter
 */
  }

  /* "pandas/_libs/intervaltree.pxi":76
 *         if self._left_sorter is None:
 *             self._left_sorter = np.argsort(self.left)
 *         return self._left_sorter             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_left_sorter);
  __pyx_r = __pyx_v_self->_left_sorter;
  goto __pyx_L0;

  /* "pandas/_libs/intervaltree.pxi":71
 * 
 *     @property
 *     def left_sorter(self):             # <<<<<<<<<<<<<<
 *         """How to sort the left labels; this is used for binary search
 *         """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pandas._libs.interval.IntervalTree.left_sorter.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":79
 * 
 *     @property
 *     def right_sorter(self):             # <<<<<<<<<<<<<<
 *         """How to sort the right labels
 *         """
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_12IntervalTree_12right_sorter_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_12IntervalTree_12right_sorter_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_12IntervalTree_12right_sorter___get__(((struct __pyx_obj_6pandas_5_libs_8interval_IntervalTree *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_12IntervalTree_12right_sorter___get__(struct __pyx_obj_6pandas_5_libs_8interval_IntervalTree *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pandas/_libs/intervaltree.pxi":82
 *         """How to sort the right labels
 *         """
 *         if self._right_sorter is None:             # <<<<<<<<<<<<<<
 *             self._right_sorter = np.argsort(self.right)
 *         return self._right_sorter
 */
  __pyx_t_1 = (__pyx_v_self->_right_sorter == Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pandas/_libs/intervaltree.pxi":83
 *         """
 *         if self._right_sorter is None:
 *             self._right_sorter = np.argsort(self.right)             # <<<<<<<<<<<<<<
 *         return self._right_sorter
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 83, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_argsort); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 83, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_4, __pyx_v_self->right) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_v_self->right);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 83, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GIVEREF(__pyx_t_3);
    __Pyx_GOTREF(__pyx_v_self->_right_sorter);
    __Pyx_DECREF(__pyx_v_self->_right_sorter);
    __pyx_v_self->_right_sorter = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "pandas/_libs/intervaltree.pxi":82
 *         """How to sort the right labels
 *         """
 *         if self._right_sorter is None:             # <<<<<<<<<<<<<<
 *             self._right_sorter = np.argsort(self.right)
 *         return self._right_sorter
 */
  }

  /* "pandas/_libs/intervaltree.pxi":84
 *         if self._right_sorter is None:
 *             self._right_sorter = np.argsort(self.right)
 *         return self._right_sorter             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_right_sorter);
  __pyx_r = __pyx_v_self->_right_sorter;
  goto __pyx_L0;

  /* "pandas/_libs/intervaltree.pxi":79
 * 
 *     @property
 *     def right_sorter(self):             # <<<<<<<<<<<<<<
 *         """How to sort the right labels
 *         """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pandas._libs.interval.IntervalTree.right_sorter.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":87
 * 
 *     @property
 *     def is_overlapping(self):             # <<<<<<<<<<<<<<
 *         """
 *         Determine if the IntervalTree contains overlapping intervals.
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_12IntervalTree_14is_overlapping_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_12IntervalTree_14is_overlapping_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_12IntervalTree_14is_overlapping___get__(((struct __pyx_obj_6pandas_5_libs_8interval_IntervalTree *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_12IntervalTree_14is_overlapping___get__(struct __pyx_obj_6pandas_5_libs_8interval_IntervalTree *__pyx_v_self) {
  PyObject *__pyx_v_op = NULL;
  PyObject *__pyx_v_current = NULL;
  PyObject *__pyx_v_previous = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pandas/_libs/intervaltree.pxi":92
 *         Cached as self._is_overlapping.
 *         """
 *         if self._is_overlapping is not None:             # <<<<<<<<<<<<<<
 *             return self._is_overlapping
 * 
 */
  __pyx_t_1 = (__pyx_v_self->_is_overlapping != Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pandas/_libs/intervaltree.pxi":93
 *         """
 *         if self._is_overlapping is not None:
 *             return self._is_overlapping             # <<<<<<<<<<<<<<
 * 
 *         # <= when both sides closed since endpoints can overlap
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_self->_is_overlapping);
    __pyx_r = __pyx_v_self->_is_overlapping;
    goto __pyx_L0;

    /* "pandas/_libs/intervaltree.pxi":92
 *         Cached as self._is_overlapping.
 *         """
 *         if self._is_overlapping is not None:             # <<<<<<<<<<<<<<
 *             return self._is_overlapping
 * 
 */
  }

  /* "pandas/_libs/intervaltree.pxi":96
 * 
 *         # <= when both sides closed since endpoints can overlap
 *         op = le if self.closed == 'both' else lt             # <<<<<<<<<<<<<<
 * 
 *         # overlap if start of current interval < end of previous interval
 */
  __pyx_t_2 = (__Pyx_PyUnicode_Equals(__pyx_v_self->closed, __pyx_n_u_both, Py_EQ)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(1, 96, __pyx_L1_error)
  if ((__pyx_t_2 != 0)) {
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_le); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 96, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = __pyx_t_4;
    __pyx_t_4 = 0;
  } else {
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_lt); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 96, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = __pyx_t_4;
    __pyx_t_4 = 0;
  }
  __pyx_v_op = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "pandas/_libs/intervaltree.pxi":100
 *         # overlap if start of current interval < end of previous interval
 *         # (current and previous in terms of sorted order by left/start side)
 *         current = self.left[self.left_sorter[1:]]             # <<<<<<<<<<<<<<
 *         previous = self.right[self.left_sorter[:-1]]
 *         self._is_overlapping = bool(op(current, previous).any())
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_left_sorter); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 100, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetSlice(__pyx_t_3, 1, 0, NULL, NULL, &__pyx_slice__11, 1, 0, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 100, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_v_self->left, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 100, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_v_current = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "pandas/_libs/intervaltree.pxi":101
 *         # (current and previous in terms of sorted order by left/start side)
 *         current = self.left[self.left_sorter[1:]]
 *         previous = self.right[self.left_sorter[:-1]]             # <<<<<<<<<<<<<<
 *         self._is_overlapping = bool(op(current, previous).any())
 * 
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_left_sorter); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 101, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetSlice(__pyx_t_3, 0, -1L, NULL, NULL, &__pyx_slice__12, 0, 1, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 101, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_v_self->right, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 101, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_v_previous = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "pandas/_libs/intervaltree.pxi":102
 *         current = self.left[self.left_sorter[1:]]
 *         previous = self.right[self.left_sorter[:-1]]
 *         self._is_overlapping = bool(op(current, previous).any())             # <<<<<<<<<<<<<<
 * 
 *         return self._is_overlapping
 */
  __Pyx_INCREF(__pyx_v_op);
  __pyx_t_5 = __pyx_v_op; __pyx_t_6 = NULL;
  __pyx_t_7 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
      __pyx_t_7 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_5)) {
    PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_v_current, __pyx_v_previous};
    __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 102, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_GOTREF(__pyx_t_4);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
    PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_v_current, __pyx_v_previous};
    __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 102, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_GOTREF(__pyx_t_4);
  } else
  #endif
  {
    __pyx_t_8 = PyTuple_New(2+__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 102, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    if (__pyx_t_6) {
      __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_6); __pyx_t_6 = NULL;
    }
    __Pyx_INCREF(__pyx_v_current);
    __Pyx_GIVEREF(__pyx_v_current);
    PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_7, __pyx_v_current);
    __Pyx_INCREF(__pyx_v_previous);
    __Pyx_GIVEREF(__pyx_v_previous);
    PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_7, __pyx_v_previous);
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_8, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 102, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  }
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_any); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 102, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 102, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(1, 102, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyBool_FromLong((!(!__pyx_t_2))); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 102, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __Pyx_GOTREF(__pyx_v_self->_is_overlapping);
  __Pyx_DECREF(__pyx_v_self->_is_overlapping);
  __pyx_v_self->_is_overlapping = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "pandas/_libs/intervaltree.pxi":104
 *         self._is_overlapping = bool(op(current, previous).any())
 * 
 *         return self._is_overlapping             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_is_overlapping);
  __pyx_r = __pyx_v_self->_is_overlapping;
  goto __pyx_L0;

  /* "pandas/_libs/intervaltree.pxi":87
 * 
 *     @property
 *     def is_overlapping(self):             # <<<<<<<<<<<<<<
 *         """
 *         Determine if the IntervalTree contains overlapping intervals.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("pandas._libs.interval.IntervalTree.is_overlapping.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_op);
  __Pyx_XDECREF(__pyx_v_current);
  __Pyx_XDECREF(__pyx_v_previous);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":107
 * 
 *     @property
 *     def is_monotonic_increasing(self):             # <<<<<<<<<<<<<<
 *         """
 *         Return True if the IntervalTree is monotonic increasing (only equal or
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_12IntervalTree_23is_monotonic_increasing_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_12IntervalTree_23is_monotonic_increasing_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_12IntervalTree_23is_monotonic_increasing___get__(((struct __pyx_obj_6pandas_5_libs_8interval_IntervalTree *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_12IntervalTree_23is_monotonic_increasing___get__(struct __pyx_obj_6pandas_5_libs_8interval_IntervalTree *__pyx_v_self) {
  PyObject *__pyx_v_values = NULL;
  PyObject *__pyx_v_sort_order = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pandas/_libs/intervaltree.pxi":112
 *         increasing values), else False
 *         """
 *         values = [self.right, self.left]             # <<<<<<<<<<<<<<
 * 
 *         sort_order = np.lexsort(values)
 */
  __pyx_t_1 = PyList_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 112, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_self->right);
  __Pyx_GIVEREF(__pyx_v_self->right);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_v_self->right);
  __Pyx_INCREF(__pyx_v_self->left);
  __Pyx_GIVEREF(__pyx_v_self->left);
  PyList_SET_ITEM(__pyx_t_1, 1, __pyx_v_self->left);
  __pyx_v_values = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":114
 *         values = [self.right, self.left]
 * 
 *         sort_order = np.lexsort(values)             # <<<<<<<<<<<<<<
 *         return is_monotonic(sort_order, False)[0]
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 114, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_lexsort); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 114, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_v_values) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_values);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 114, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_sort_order = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":115
 * 
 *         sort_order = np.lexsort(values)
 *         return is_monotonic(sort_order, False)[0]             # <<<<<<<<<<<<<<
 * 
 *     def get_loc(self, scalar_t key):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_is_monotonic); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 115, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_sort_order, Py_False};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 115, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_sort_order, Py_False};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 115, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 115, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_2) {
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2); __pyx_t_2 = NULL;
    }
    __Pyx_INCREF(__pyx_v_sort_order);
    __Pyx_GIVEREF(__pyx_v_sort_order);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_v_sort_order);
    __Pyx_INCREF(Py_False);
    __Pyx_GIVEREF(Py_False);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, Py_False);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 115, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_GetItemInt(__pyx_t_1, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 115, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "pandas/_libs/intervaltree.pxi":107
 * 
 *     @property
 *     def is_monotonic_increasing(self):             # <<<<<<<<<<<<<<
 *         """
 *         Return True if the IntervalTree is monotonic increasing (only equal or
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pandas._libs.interval.IntervalTree.is_monotonic_increasing.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_values);
  __Pyx_XDECREF(__pyx_v_sort_order);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":117
 *         return is_monotonic(sort_order, False)[0]
 * 
 *     def get_loc(self, scalar_t key):             # <<<<<<<<<<<<<<
 *         """Return all positions corresponding to intervals that overlap with
 *         the given scalar key
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_12IntervalTree_3get_loc(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6pandas_5_libs_8interval_12IntervalTree_2get_loc[] = "Return all positions corresponding to intervals that overlap with\n        the given scalar key\n        ";
static PyMethodDef __pyx_mdef_6pandas_5_libs_8interval_12IntervalTree_3get_loc = {"get_loc", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_12IntervalTree_3get_loc, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_12IntervalTree_2get_loc};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_12IntervalTree_3get_loc(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_args = 0;
  PyObject *__pyx_v_kwargs = 0;
  CYTHON_UNUSED PyObject *__pyx_v_defaults = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fused_cpdef (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_args,&__pyx_n_s_kwargs,&__pyx_n_s_defaults,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_args)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_kwargs)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fused_cpdef", 1, 3, 3, 1); __PYX_ERR(1, 117, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_defaults)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fused_cpdef", 1, 3, 3, 2); __PYX_ERR(1, 117, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fused_cpdef") < 0)) __PYX_ERR(1, 117, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_args = values[0];
    __pyx_v_kwargs = values[1];
    __pyx_v_defaults = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fused_cpdef", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 117, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.IntervalTree.__pyx_fused_cpdef", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_12IntervalTree_2get_loc(((struct __pyx_obj_6pandas_5_libs_8interval_IntervalTree *)__pyx_v_signatures), __pyx_v_args, __pyx_v_kwargs, __pyx_v_defaults);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_12IntervalTree_2get_loc(struct __pyx_obj_6pandas_5_libs_8interval_IntervalTree *__pyx_v_signatures, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs, CYTHON_UNUSED PyObject *__pyx_v_defaults) {
  PyObject *__pyx_v_dest_sig = NULL;
  Py_ssize_t __pyx_v_i;
  CYTHON_UNUSED PyTypeObject *__pyx_v_ndarray = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_v_itemsize;
  PyObject *__pyx_v_arg = NULL;
  PyObject *__pyx_v_candidates = NULL;
  PyObject *__pyx_v_sig = NULL;
  int __pyx_v_match_found;
  PyObject *__pyx_v_src_sig = NULL;
  PyObject *__pyx_v_dst_type = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  Py_ssize_t __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  Py_ssize_t __pyx_t_13;
  Py_ssize_t __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  int __pyx_t_16;
  __Pyx_RefNannySetupContext("get_loc", 0);
  __Pyx_INCREF(__pyx_v_kwargs);
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 117, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  PyList_SET_ITEM(__pyx_t_1, 0, Py_None);
  __pyx_v_dest_sig = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_3 = (__pyx_v_kwargs != Py_None);
  __pyx_t_4 = (__pyx_t_3 != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_kwargs); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(1, 117, __pyx_L1_error)
  __pyx_t_3 = ((!__pyx_t_4) != 0);
  __pyx_t_2 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {
    __Pyx_INCREF(Py_None);
    __Pyx_DECREF_SET(__pyx_v_kwargs, Py_None);
  }
  __pyx_t_1 = ((PyObject *)__Pyx_ImportNumPyArrayTypeIfAvailable()); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 117, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_ndarray = ((PyTypeObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_v_itemsize = -1L;
  if (unlikely(__pyx_v_args == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 117, __pyx_L1_error)
  }
  __pyx_t_5 = PyTuple_GET_SIZE(((PyObject*)__pyx_v_args)); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(1, 117, __pyx_L1_error)
  __pyx_t_2 = ((1 < __pyx_t_5) != 0);
  if (__pyx_t_2) {
    if (unlikely(__pyx_v_args == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 117, __pyx_L1_error)
    }
    __pyx_t_1 = __Pyx_GetItemInt_Tuple(((PyObject*)__pyx_v_args), 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 117, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_arg = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L6;
  }
  __pyx_t_3 = (__pyx_v_kwargs != Py_None);
  __pyx_t_4 = (__pyx_t_3 != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L7_bool_binop_done;
  }
  if (unlikely(__pyx_v_kwargs == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(1, 117, __pyx_L1_error)
  }
  __pyx_t_4 = (__Pyx_PyDict_ContainsTF(__pyx_n_s_key, ((PyObject*)__pyx_v_kwargs), Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(1, 117, __pyx_L1_error)
  __pyx_t_3 = (__pyx_t_4 != 0);
  __pyx_t_2 = __pyx_t_3;
  __pyx_L7_bool_binop_done:;
  if (__pyx_t_2) {
    if (unlikely(__pyx_v_kwargs == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 117, __pyx_L1_error)
    }
    __pyx_t_1 = __Pyx_PyDict_GetItem(((PyObject*)__pyx_v_kwargs), __pyx_n_s_key); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 117, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_arg = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L6;
  }
  /*else*/ {
    if (unlikely(__pyx_v_args == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
      __PYX_ERR(1, 117, __pyx_L1_error)
    }
    __pyx_t_5 = PyTuple_GET_SIZE(((PyObject*)__pyx_v_args)); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(1, 117, __pyx_L1_error)
    __pyx_t_1 = PyInt_FromSsize_t(__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 117, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_6 = PyTuple_New(3); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 117, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_INCREF(__pyx_int_2);
    __Pyx_GIVEREF(__pyx_int_2);
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_int_2);
    __Pyx_INCREF(__pyx_n_s_s);
    __Pyx_GIVEREF(__pyx_n_s_s);
    PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_n_s_s);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Expected_at_least_d_argument_s_g, __pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 117, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_1); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 117, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(1, 117, __pyx_L1_error)
  }
  __pyx_L6:;
  while (1) {
    __pyx_t_2 = PyFloat_Check(__pyx_v_arg); 
    __pyx_t_3 = (__pyx_t_2 != 0);
    if (__pyx_t_3) {
      if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, __pyx_n_s_float64_t, long, 1, __Pyx_PyInt_From_long, 1, 0, 1) < 0)) __PYX_ERR(1, 117, __pyx_L1_error)
      goto __pyx_L10_break;
    }
    __pyx_t_2 = PyInt_Check(__pyx_v_arg); 
    __pyx_t_4 = (__pyx_t_2 != 0);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_3 = __pyx_t_4;
      goto __pyx_L13_bool_binop_done;
    }
    __pyx_t_4 = PyLong_Check(__pyx_v_arg); 
    __pyx_t_2 = (__pyx_t_4 != 0);
    __pyx_t_3 = __pyx_t_2;
    __pyx_L13_bool_binop_done:;
    __pyx_t_2 = (__pyx_t_3 != 0);
    if (__pyx_t_2) {
      if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, __pyx_n_s_int64_t, long, 1, __Pyx_PyInt_From_long, 1, 0, 1) < 0)) __PYX_ERR(1, 117, __pyx_L1_error)
      goto __pyx_L10_break;
    }
    if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, Py_None, long, 1, __Pyx_PyInt_From_long, 1, 0, 1) < 0)) __PYX_ERR(1, 117, __pyx_L1_error)
    goto __pyx_L10_break;
  }
  __pyx_L10_break:;
  __pyx_t_6 = PyList_New(0); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 117, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_v_candidates = ((PyObject*)__pyx_t_6);
  __pyx_t_6 = 0;
  __pyx_t_5 = 0;
  if (unlikely(((PyObject *)__pyx_v_signatures) == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(1, 117, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_dict_iterator(((PyObject*)__pyx_v_signatures), 1, ((PyObject *)NULL), (&__pyx_t_7), (&__pyx_t_8)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 117, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __pyx_t_6 = __pyx_t_1;
  __pyx_t_1 = 0;
  while (1) {
    __pyx_t_9 = __Pyx_dict_iter_next(__pyx_t_6, __pyx_t_7, &__pyx_t_5, &__pyx_t_1, NULL, NULL, __pyx_t_8);
    if (unlikely(__pyx_t_9 == 0)) break;
    if (unlikely(__pyx_t_9 == -1)) __PYX_ERR(1, 117, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_sig, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_v_match_found = 0;
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_v_sig, __pyx_n_s_strip); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 117, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __pyx_t_12 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_11))) {
      __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_11);
      if (likely(__pyx_t_12)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
        __Pyx_INCREF(__pyx_t_12);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_11, function);
      }
    }
    __pyx_t_10 = (__pyx_t_12) ? __Pyx_PyObject_Call2Args(__pyx_t_11, __pyx_t_12, __pyx_kp_s__13) : __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_kp_s__13);
    __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
    if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 117, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_split); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 117, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_t_10 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_11))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_11);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_11, function);
      }
    }
    __pyx_t_1 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_11, __pyx_t_10, __pyx_kp_s__14) : __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_kp_s__14);
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 117, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_XDECREF_SET(__pyx_v_src_sig, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_13 = PyList_GET_SIZE(__pyx_v_dest_sig); if (unlikely(__pyx_t_13 == ((Py_ssize_t)-1))) __PYX_ERR(1, 117, __pyx_L1_error)
    __pyx_t_14 = __pyx_t_13;
    for (__pyx_t_15 = 0; __pyx_t_15 < __pyx_t_14; __pyx_t_15+=1) {
      __pyx_v_i = __pyx_t_15;
      __pyx_t_1 = __Pyx_GetItemInt_List(__pyx_v_dest_sig, __pyx_v_i, Py_ssize_t, 1, PyInt_FromSsize_t, 1, 1, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 117, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_XDECREF_SET(__pyx_v_dst_type, __pyx_t_1);
      __pyx_t_1 = 0;
      __pyx_t_2 = (__pyx_v_dst_type != Py_None);
      __pyx_t_3 = (__pyx_t_2 != 0);
      if (__pyx_t_3) {
        __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_src_sig, __pyx_v_i, Py_ssize_t, 1, PyInt_FromSsize_t, 0, 1, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 117, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_11 = PyObject_RichCompare(__pyx_t_1, __pyx_v_dst_type, Py_EQ); __Pyx_XGOTREF(__pyx_t_11); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 117, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_11); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(1, 117, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        if (__pyx_t_3) {
          __pyx_v_match_found = 1;
          goto __pyx_L20;
        }
        /*else*/ {
          __pyx_v_match_found = 0;
          goto __pyx_L18_break;
        }
        __pyx_L20:;
      }
    }
    __pyx_L18_break:;
    __pyx_t_3 = (__pyx_v_match_found != 0);
    if (__pyx_t_3) {
      __pyx_t_16 = __Pyx_PyList_Append(__pyx_v_candidates, __pyx_v_sig); if (unlikely(__pyx_t_16 == ((int)-1))) __PYX_ERR(1, 117, __pyx_L1_error)
    }
  }
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_3 = (PyList_GET_SIZE(__pyx_v_candidates) != 0);
  __pyx_t_2 = ((!__pyx_t_3) != 0);
  if (__pyx_t_2) {
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__15, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 117, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(1, 117, __pyx_L1_error)
  }
  __pyx_t_7 = PyList_GET_SIZE(__pyx_v_candidates); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1))) __PYX_ERR(1, 117, __pyx_L1_error)
  __pyx_t_2 = ((__pyx_t_7 > 1) != 0);
  if (__pyx_t_2) {
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__16, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 117, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(1, 117, __pyx_L1_error)
  }
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    if (unlikely(((PyObject *)__pyx_v_signatures) == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 117, __pyx_L1_error)
    }
    __pyx_t_6 = __Pyx_GetItemInt_List(__pyx_v_candidates, 0, long, 1, __Pyx_PyInt_From_long, 1, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 117, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_11 = __Pyx_PyDict_GetItem(((PyObject*)__pyx_v_signatures), __pyx_t_6); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 117, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_r = __pyx_t_11;
    __pyx_t_11 = 0;
    goto __pyx_L0;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("pandas._libs.interval.IntervalTree.__pyx_fused_cpdef", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_dest_sig);
  __Pyx_XDECREF(__pyx_v_ndarray);
  __Pyx_XDECREF(__pyx_v_arg);
  __Pyx_XDECREF(__pyx_v_candidates);
  __Pyx_XDECREF(__pyx_v_sig);
  __Pyx_XDECREF(__pyx_v_src_sig);
  __Pyx_XDECREF(__pyx_v_dst_type);
  __Pyx_XDECREF(__pyx_v_kwargs);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_fuse_0__pyx_pw_6pandas_5_libs_8interval_12IntervalTree_17get_loc(PyObject *__pyx_v_self, PyObject *__pyx_arg_key); /*proto*/
static PyMethodDef __pyx_fuse_0__pyx_mdef_6pandas_5_libs_8interval_12IntervalTree_17get_loc = {"__pyx_fuse_0get_loc", (PyCFunction)__pyx_fuse_0__pyx_pw_6pandas_5_libs_8interval_12IntervalTree_17get_loc, METH_O, __pyx_doc_6pandas_5_libs_8interval_12IntervalTree_2get_loc};
static PyObject *__pyx_fuse_0__pyx_pw_6pandas_5_libs_8interval_12IntervalTree_17get_loc(PyObject *__pyx_v_self, PyObject *__pyx_arg_key) {
  __pyx_t_5numpy_float64_t __pyx_v_key;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_loc (wrapper)", 0);
  assert(__pyx_arg_key); {
    __pyx_v_key = __pyx_PyFloat_AsDouble(__pyx_arg_key); if (unlikely((__pyx_v_key == ((npy_float64)-1)) && PyErr_Occurred())) __PYX_ERR(1, 117, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.IntervalTree.get_loc", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_12IntervalTree_16get_loc(((struct __pyx_obj_6pandas_5_libs_8interval_IntervalTree *)__pyx_v_self), ((__pyx_t_5numpy_float64_t)__pyx_v_key));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_12IntervalTree_16get_loc(struct __pyx_obj_6pandas_5_libs_8interval_IntervalTree *__pyx_v_self, __pyx_t_5numpy_float64_t __pyx_v_key) {
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  __Pyx_RefNannySetupContext("__pyx_fuse_0get_loc", 0);

  /* "pandas/_libs/intervaltree.pxi":121
 *         the given scalar key
 *         """
 *         result = Int64Vector()             # <<<<<<<<<<<<<<
 *         self.root.query(result, key)
 *         if not result.data.n:
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 121, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_result = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":122
 *         """
 *         result = Int64Vector()
 *         self.root.query(result, key)             # <<<<<<<<<<<<<<
 *         if not result.data.n:
 *             raise KeyError(key)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->root, __pyx_n_s_query); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 122, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyFloat_FromDouble(__pyx_v_key); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 122, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_4, ((PyObject *)__pyx_v_result), __pyx_t_3};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 122, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_4, ((PyObject *)__pyx_v_result), __pyx_t_3};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 122, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 122, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_4) {
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4); __pyx_t_4 = NULL;
    }
    __Pyx_INCREF(((PyObject *)__pyx_v_result));
    __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, ((PyObject *)__pyx_v_result));
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_t_3);
    __pyx_t_3 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 122, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":123
 *         result = Int64Vector()
 *         self.root.query(result, key)
 *         if not result.data.n:             # <<<<<<<<<<<<<<
 *             raise KeyError(key)
 *         return result.to_array().astype('intp')
 */
  __pyx_t_7 = ((!(__pyx_v_result->data->n != 0)) != 0);
  if (unlikely(__pyx_t_7)) {

    /* "pandas/_libs/intervaltree.pxi":124
 *         self.root.query(result, key)
 *         if not result.data.n:
 *             raise KeyError(key)             # <<<<<<<<<<<<<<
 *         return result.to_array().astype('intp')
 * 
 */
    __pyx_t_1 = PyFloat_FromDouble(__pyx_v_key); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 124, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_KeyError, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 124, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(1, 124, __pyx_L1_error)

    /* "pandas/_libs/intervaltree.pxi":123
 *         result = Int64Vector()
 *         self.root.query(result, key)
 *         if not result.data.n:             # <<<<<<<<<<<<<<
 *             raise KeyError(key)
 *         return result.to_array().astype('intp')
 */
  }

  /* "pandas/_libs/intervaltree.pxi":125
 *         if not result.data.n:
 *             raise KeyError(key)
 *         return result.to_array().astype('intp')             # <<<<<<<<<<<<<<
 * 
 *     def _get_partial_overlap(self, key_left, key_right, side):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->to_array(__pyx_v_result, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 125, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_astype); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 125, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_6);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_6, function);
    }
  }
  __pyx_t_2 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_1, __pyx_n_u_intp) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_n_u_intp);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 125, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pandas/_libs/intervaltree.pxi":117
 *         return is_monotonic(sort_order, False)[0]
 * 
 *     def get_loc(self, scalar_t key):             # <<<<<<<<<<<<<<
 *         """Return all positions corresponding to intervals that overlap with
 *         the given scalar key
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pandas._libs.interval.IntervalTree.get_loc", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_fuse_1__pyx_pw_6pandas_5_libs_8interval_12IntervalTree_19get_loc(PyObject *__pyx_v_self, PyObject *__pyx_arg_key); /*proto*/
static PyMethodDef __pyx_fuse_1__pyx_mdef_6pandas_5_libs_8interval_12IntervalTree_19get_loc = {"__pyx_fuse_1get_loc", (PyCFunction)__pyx_fuse_1__pyx_pw_6pandas_5_libs_8interval_12IntervalTree_19get_loc, METH_O, __pyx_doc_6pandas_5_libs_8interval_12IntervalTree_2get_loc};
static PyObject *__pyx_fuse_1__pyx_pw_6pandas_5_libs_8interval_12IntervalTree_19get_loc(PyObject *__pyx_v_self, PyObject *__pyx_arg_key) {
  __pyx_t_5numpy_float32_t __pyx_v_key;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_loc (wrapper)", 0);
  assert(__pyx_arg_key); {
    __pyx_v_key = __pyx_PyFloat_AsFloat(__pyx_arg_key); if (unlikely((__pyx_v_key == ((npy_float32)-1)) && PyErr_Occurred())) __PYX_ERR(1, 117, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.IntervalTree.get_loc", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_12IntervalTree_18get_loc(((struct __pyx_obj_6pandas_5_libs_8interval_IntervalTree *)__pyx_v_self), ((__pyx_t_5numpy_float32_t)__pyx_v_key));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_12IntervalTree_18get_loc(struct __pyx_obj_6pandas_5_libs_8interval_IntervalTree *__pyx_v_self, __pyx_t_5numpy_float32_t __pyx_v_key) {
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  __Pyx_RefNannySetupContext("__pyx_fuse_1get_loc", 0);

  /* "pandas/_libs/intervaltree.pxi":121
 *         the given scalar key
 *         """
 *         result = Int64Vector()             # <<<<<<<<<<<<<<
 *         self.root.query(result, key)
 *         if not result.data.n:
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 121, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_result = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":122
 *         """
 *         result = Int64Vector()
 *         self.root.query(result, key)             # <<<<<<<<<<<<<<
 *         if not result.data.n:
 *             raise KeyError(key)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->root, __pyx_n_s_query); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 122, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyFloat_FromDouble(__pyx_v_key); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 122, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_4, ((PyObject *)__pyx_v_result), __pyx_t_3};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 122, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_4, ((PyObject *)__pyx_v_result), __pyx_t_3};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 122, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 122, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_4) {
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4); __pyx_t_4 = NULL;
    }
    __Pyx_INCREF(((PyObject *)__pyx_v_result));
    __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, ((PyObject *)__pyx_v_result));
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_t_3);
    __pyx_t_3 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 122, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":123
 *         result = Int64Vector()
 *         self.root.query(result, key)
 *         if not result.data.n:             # <<<<<<<<<<<<<<
 *             raise KeyError(key)
 *         return result.to_array().astype('intp')
 */
  __pyx_t_7 = ((!(__pyx_v_result->data->n != 0)) != 0);
  if (unlikely(__pyx_t_7)) {

    /* "pandas/_libs/intervaltree.pxi":124
 *         self.root.query(result, key)
 *         if not result.data.n:
 *             raise KeyError(key)             # <<<<<<<<<<<<<<
 *         return result.to_array().astype('intp')
 * 
 */
    __pyx_t_1 = PyFloat_FromDouble(__pyx_v_key); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 124, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_KeyError, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 124, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(1, 124, __pyx_L1_error)

    /* "pandas/_libs/intervaltree.pxi":123
 *         result = Int64Vector()
 *         self.root.query(result, key)
 *         if not result.data.n:             # <<<<<<<<<<<<<<
 *             raise KeyError(key)
 *         return result.to_array().astype('intp')
 */
  }

  /* "pandas/_libs/intervaltree.pxi":125
 *         if not result.data.n:
 *             raise KeyError(key)
 *         return result.to_array().astype('intp')             # <<<<<<<<<<<<<<
 * 
 *     def _get_partial_overlap(self, key_left, key_right, side):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->to_array(__pyx_v_result, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 125, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_astype); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 125, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_6);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_6, function);
    }
  }
  __pyx_t_2 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_1, __pyx_n_u_intp) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_n_u_intp);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 125, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pandas/_libs/intervaltree.pxi":117
 *         return is_monotonic(sort_order, False)[0]
 * 
 *     def get_loc(self, scalar_t key):             # <<<<<<<<<<<<<<
 *         """Return all positions corresponding to intervals that overlap with
 *         the given scalar key
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pandas._libs.interval.IntervalTree.get_loc", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_fuse_2__pyx_pw_6pandas_5_libs_8interval_12IntervalTree_21get_loc(PyObject *__pyx_v_self, PyObject *__pyx_arg_key); /*proto*/
static PyMethodDef __pyx_fuse_2__pyx_mdef_6pandas_5_libs_8interval_12IntervalTree_21get_loc = {"__pyx_fuse_2get_loc", (PyCFunction)__pyx_fuse_2__pyx_pw_6pandas_5_libs_8interval_12IntervalTree_21get_loc, METH_O, __pyx_doc_6pandas_5_libs_8interval_12IntervalTree_2get_loc};
static PyObject *__pyx_fuse_2__pyx_pw_6pandas_5_libs_8interval_12IntervalTree_21get_loc(PyObject *__pyx_v_self, PyObject *__pyx_arg_key) {
  __pyx_t_5numpy_int64_t __pyx_v_key;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_loc (wrapper)", 0);
  assert(__pyx_arg_key); {
    __pyx_v_key = __Pyx_PyInt_As_npy_int64(__pyx_arg_key); if (unlikely((__pyx_v_key == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(1, 117, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.IntervalTree.get_loc", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_12IntervalTree_20get_loc(((struct __pyx_obj_6pandas_5_libs_8interval_IntervalTree *)__pyx_v_self), ((__pyx_t_5numpy_int64_t)__pyx_v_key));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_12IntervalTree_20get_loc(struct __pyx_obj_6pandas_5_libs_8interval_IntervalTree *__pyx_v_self, __pyx_t_5numpy_int64_t __pyx_v_key) {
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  __Pyx_RefNannySetupContext("__pyx_fuse_2get_loc", 0);

  /* "pandas/_libs/intervaltree.pxi":121
 *         the given scalar key
 *         """
 *         result = Int64Vector()             # <<<<<<<<<<<<<<
 *         self.root.query(result, key)
 *         if not result.data.n:
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 121, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_result = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":122
 *         """
 *         result = Int64Vector()
 *         self.root.query(result, key)             # <<<<<<<<<<<<<<
 *         if not result.data.n:
 *             raise KeyError(key)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->root, __pyx_n_s_query); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 122, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_From_npy_int64(__pyx_v_key); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 122, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_4, ((PyObject *)__pyx_v_result), __pyx_t_3};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 122, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_4, ((PyObject *)__pyx_v_result), __pyx_t_3};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 122, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 122, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_4) {
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4); __pyx_t_4 = NULL;
    }
    __Pyx_INCREF(((PyObject *)__pyx_v_result));
    __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, ((PyObject *)__pyx_v_result));
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_t_3);
    __pyx_t_3 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 122, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":123
 *         result = Int64Vector()
 *         self.root.query(result, key)
 *         if not result.data.n:             # <<<<<<<<<<<<<<
 *             raise KeyError(key)
 *         return result.to_array().astype('intp')
 */
  __pyx_t_7 = ((!(__pyx_v_result->data->n != 0)) != 0);
  if (unlikely(__pyx_t_7)) {

    /* "pandas/_libs/intervaltree.pxi":124
 *         self.root.query(result, key)
 *         if not result.data.n:
 *             raise KeyError(key)             # <<<<<<<<<<<<<<
 *         return result.to_array().astype('intp')
 * 
 */
    __pyx_t_1 = __Pyx_PyInt_From_npy_int64(__pyx_v_key); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 124, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_KeyError, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 124, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(1, 124, __pyx_L1_error)

    /* "pandas/_libs/intervaltree.pxi":123
 *         result = Int64Vector()
 *         self.root.query(result, key)
 *         if not result.data.n:             # <<<<<<<<<<<<<<
 *             raise KeyError(key)
 *         return result.to_array().astype('intp')
 */
  }

  /* "pandas/_libs/intervaltree.pxi":125
 *         if not result.data.n:
 *             raise KeyError(key)
 *         return result.to_array().astype('intp')             # <<<<<<<<<<<<<<
 * 
 *     def _get_partial_overlap(self, key_left, key_right, side):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->to_array(__pyx_v_result, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 125, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_astype); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 125, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_6);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_6, function);
    }
  }
  __pyx_t_2 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_1, __pyx_n_u_intp) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_n_u_intp);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 125, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pandas/_libs/intervaltree.pxi":117
 *         return is_monotonic(sort_order, False)[0]
 * 
 *     def get_loc(self, scalar_t key):             # <<<<<<<<<<<<<<
 *         """Return all positions corresponding to intervals that overlap with
 *         the given scalar key
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pandas._libs.interval.IntervalTree.get_loc", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_fuse_3__pyx_pw_6pandas_5_libs_8interval_12IntervalTree_23get_loc(PyObject *__pyx_v_self, PyObject *__pyx_arg_key); /*proto*/
static PyMethodDef __pyx_fuse_3__pyx_mdef_6pandas_5_libs_8interval_12IntervalTree_23get_loc = {"__pyx_fuse_3get_loc", (PyCFunction)__pyx_fuse_3__pyx_pw_6pandas_5_libs_8interval_12IntervalTree_23get_loc, METH_O, __pyx_doc_6pandas_5_libs_8interval_12IntervalTree_2get_loc};
static PyObject *__pyx_fuse_3__pyx_pw_6pandas_5_libs_8interval_12IntervalTree_23get_loc(PyObject *__pyx_v_self, PyObject *__pyx_arg_key) {
  __pyx_t_5numpy_int32_t __pyx_v_key;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_loc (wrapper)", 0);
  assert(__pyx_arg_key); {
    __pyx_v_key = __Pyx_PyInt_As_npy_int32(__pyx_arg_key); if (unlikely((__pyx_v_key == ((npy_int32)-1)) && PyErr_Occurred())) __PYX_ERR(1, 117, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.IntervalTree.get_loc", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_12IntervalTree_22get_loc(((struct __pyx_obj_6pandas_5_libs_8interval_IntervalTree *)__pyx_v_self), ((__pyx_t_5numpy_int32_t)__pyx_v_key));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_12IntervalTree_22get_loc(struct __pyx_obj_6pandas_5_libs_8interval_IntervalTree *__pyx_v_self, __pyx_t_5numpy_int32_t __pyx_v_key) {
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  __Pyx_RefNannySetupContext("__pyx_fuse_3get_loc", 0);

  /* "pandas/_libs/intervaltree.pxi":121
 *         the given scalar key
 *         """
 *         result = Int64Vector()             # <<<<<<<<<<<<<<
 *         self.root.query(result, key)
 *         if not result.data.n:
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 121, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_result = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":122
 *         """
 *         result = Int64Vector()
 *         self.root.query(result, key)             # <<<<<<<<<<<<<<
 *         if not result.data.n:
 *             raise KeyError(key)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->root, __pyx_n_s_query); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 122, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_From_npy_int32(__pyx_v_key); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 122, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_4, ((PyObject *)__pyx_v_result), __pyx_t_3};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 122, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_4, ((PyObject *)__pyx_v_result), __pyx_t_3};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 122, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 122, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_4) {
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4); __pyx_t_4 = NULL;
    }
    __Pyx_INCREF(((PyObject *)__pyx_v_result));
    __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, ((PyObject *)__pyx_v_result));
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_t_3);
    __pyx_t_3 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 122, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":123
 *         result = Int64Vector()
 *         self.root.query(result, key)
 *         if not result.data.n:             # <<<<<<<<<<<<<<
 *             raise KeyError(key)
 *         return result.to_array().astype('intp')
 */
  __pyx_t_7 = ((!(__pyx_v_result->data->n != 0)) != 0);
  if (unlikely(__pyx_t_7)) {

    /* "pandas/_libs/intervaltree.pxi":124
 *         self.root.query(result, key)
 *         if not result.data.n:
 *             raise KeyError(key)             # <<<<<<<<<<<<<<
 *         return result.to_array().astype('intp')
 * 
 */
    __pyx_t_1 = __Pyx_PyInt_From_npy_int32(__pyx_v_key); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 124, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_KeyError, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 124, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(1, 124, __pyx_L1_error)

    /* "pandas/_libs/intervaltree.pxi":123
 *         result = Int64Vector()
 *         self.root.query(result, key)
 *         if not result.data.n:             # <<<<<<<<<<<<<<
 *             raise KeyError(key)
 *         return result.to_array().astype('intp')
 */
  }

  /* "pandas/_libs/intervaltree.pxi":125
 *         if not result.data.n:
 *             raise KeyError(key)
 *         return result.to_array().astype('intp')             # <<<<<<<<<<<<<<
 * 
 *     def _get_partial_overlap(self, key_left, key_right, side):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->to_array(__pyx_v_result, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 125, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_astype); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 125, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_6);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_6, function);
    }
  }
  __pyx_t_2 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_1, __pyx_n_u_intp) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_n_u_intp);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 125, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pandas/_libs/intervaltree.pxi":117
 *         return is_monotonic(sort_order, False)[0]
 * 
 *     def get_loc(self, scalar_t key):             # <<<<<<<<<<<<<<
 *         """Return all positions corresponding to intervals that overlap with
 *         the given scalar key
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pandas._libs.interval.IntervalTree.get_loc", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_fuse_4__pyx_pw_6pandas_5_libs_8interval_12IntervalTree_25get_loc(PyObject *__pyx_v_self, PyObject *__pyx_arg_key); /*proto*/
static PyMethodDef __pyx_fuse_4__pyx_mdef_6pandas_5_libs_8interval_12IntervalTree_25get_loc = {"__pyx_fuse_4get_loc", (PyCFunction)__pyx_fuse_4__pyx_pw_6pandas_5_libs_8interval_12IntervalTree_25get_loc, METH_O, __pyx_doc_6pandas_5_libs_8interval_12IntervalTree_2get_loc};
static PyObject *__pyx_fuse_4__pyx_pw_6pandas_5_libs_8interval_12IntervalTree_25get_loc(PyObject *__pyx_v_self, PyObject *__pyx_arg_key) {
  __pyx_t_5numpy_uint64_t __pyx_v_key;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_loc (wrapper)", 0);
  assert(__pyx_arg_key); {
    __pyx_v_key = __Pyx_PyInt_As_npy_uint64(__pyx_arg_key); if (unlikely((__pyx_v_key == ((npy_uint64)-1)) && PyErr_Occurred())) __PYX_ERR(1, 117, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.IntervalTree.get_loc", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_12IntervalTree_24get_loc(((struct __pyx_obj_6pandas_5_libs_8interval_IntervalTree *)__pyx_v_self), ((__pyx_t_5numpy_uint64_t)__pyx_v_key));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_12IntervalTree_24get_loc(struct __pyx_obj_6pandas_5_libs_8interval_IntervalTree *__pyx_v_self, __pyx_t_5numpy_uint64_t __pyx_v_key) {
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  __Pyx_RefNannySetupContext("__pyx_fuse_4get_loc", 0);

  /* "pandas/_libs/intervaltree.pxi":121
 *         the given scalar key
 *         """
 *         result = Int64Vector()             # <<<<<<<<<<<<<<
 *         self.root.query(result, key)
 *         if not result.data.n:
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 121, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_result = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":122
 *         """
 *         result = Int64Vector()
 *         self.root.query(result, key)             # <<<<<<<<<<<<<<
 *         if not result.data.n:
 *             raise KeyError(key)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->root, __pyx_n_s_query); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 122, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_From_npy_uint64(__pyx_v_key); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 122, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_4, ((PyObject *)__pyx_v_result), __pyx_t_3};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 122, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_4, ((PyObject *)__pyx_v_result), __pyx_t_3};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 122, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 122, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_4) {
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4); __pyx_t_4 = NULL;
    }
    __Pyx_INCREF(((PyObject *)__pyx_v_result));
    __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, ((PyObject *)__pyx_v_result));
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_t_3);
    __pyx_t_3 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 122, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":123
 *         result = Int64Vector()
 *         self.root.query(result, key)
 *         if not result.data.n:             # <<<<<<<<<<<<<<
 *             raise KeyError(key)
 *         return result.to_array().astype('intp')
 */
  __pyx_t_7 = ((!(__pyx_v_result->data->n != 0)) != 0);
  if (unlikely(__pyx_t_7)) {

    /* "pandas/_libs/intervaltree.pxi":124
 *         self.root.query(result, key)
 *         if not result.data.n:
 *             raise KeyError(key)             # <<<<<<<<<<<<<<
 *         return result.to_array().astype('intp')
 * 
 */
    __pyx_t_1 = __Pyx_PyInt_From_npy_uint64(__pyx_v_key); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 124, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_KeyError, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 124, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(1, 124, __pyx_L1_error)

    /* "pandas/_libs/intervaltree.pxi":123
 *         result = Int64Vector()
 *         self.root.query(result, key)
 *         if not result.data.n:             # <<<<<<<<<<<<<<
 *             raise KeyError(key)
 *         return result.to_array().astype('intp')
 */
  }

  /* "pandas/_libs/intervaltree.pxi":125
 *         if not result.data.n:
 *             raise KeyError(key)
 *         return result.to_array().astype('intp')             # <<<<<<<<<<<<<<
 * 
 *     def _get_partial_overlap(self, key_left, key_right, side):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->to_array(__pyx_v_result, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 125, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_astype); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 125, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_6);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_6, function);
    }
  }
  __pyx_t_2 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_1, __pyx_n_u_intp) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_n_u_intp);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 125, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pandas/_libs/intervaltree.pxi":117
 *         return is_monotonic(sort_order, False)[0]
 * 
 *     def get_loc(self, scalar_t key):             # <<<<<<<<<<<<<<
 *         """Return all positions corresponding to intervals that overlap with
 *         the given scalar key
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pandas._libs.interval.IntervalTree.get_loc", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":127
 *         return result.to_array().astype('intp')
 * 
 *     def _get_partial_overlap(self, key_left, key_right, side):             # <<<<<<<<<<<<<<
 *         """Return all positions corresponding to intervals with the given side
 *         falling between the left and right bounds of an interval query
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_12IntervalTree_5_get_partial_overlap(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6pandas_5_libs_8interval_12IntervalTree_4_get_partial_overlap[] = "Return all positions corresponding to intervals with the given side\n        falling between the left and right bounds of an interval query\n        ";
static PyObject *__pyx_pw_6pandas_5_libs_8interval_12IntervalTree_5_get_partial_overlap(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_key_left = 0;
  PyObject *__pyx_v_key_right = 0;
  PyObject *__pyx_v_side = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_get_partial_overlap (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_key_left,&__pyx_n_s_key_right,&__pyx_n_s_side,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_key_left)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_key_right)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_get_partial_overlap", 1, 3, 3, 1); __PYX_ERR(1, 127, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_side)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_get_partial_overlap", 1, 3, 3, 2); __PYX_ERR(1, 127, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_get_partial_overlap") < 0)) __PYX_ERR(1, 127, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_key_left = values[0];
    __pyx_v_key_right = values[1];
    __pyx_v_side = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_get_partial_overlap", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 127, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.IntervalTree._get_partial_overlap", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_12IntervalTree_4_get_partial_overlap(((struct __pyx_obj_6pandas_5_libs_8interval_IntervalTree *)__pyx_v_self), __pyx_v_key_left, __pyx_v_key_right, __pyx_v_side);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_12IntervalTree_4_get_partial_overlap(struct __pyx_obj_6pandas_5_libs_8interval_IntervalTree *__pyx_v_self, PyObject *__pyx_v_key_left, PyObject *__pyx_v_key_right, PyObject *__pyx_v_side) {
  PyObject *__pyx_v_values = NULL;
  PyObject *__pyx_v_sorter = NULL;
  PyObject *__pyx_v_key = NULL;
  PyObject *__pyx_v_i = NULL;
  PyObject *__pyx_v_j = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *(*__pyx_t_6)(PyObject *);
  __Pyx_RefNannySetupContext("_get_partial_overlap", 0);

  /* "pandas/_libs/intervaltree.pxi":131
 *         falling between the left and right bounds of an interval query
 *         """
 *         if side == 'left':             # <<<<<<<<<<<<<<
 *             values = self.left
 *             sorter = self.left_sorter
 */
  __pyx_t_1 = (__Pyx_PyUnicode_Equals(__pyx_v_side, __pyx_n_u_left, Py_EQ)); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(1, 131, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "pandas/_libs/intervaltree.pxi":132
 *         """
 *         if side == 'left':
 *             values = self.left             # <<<<<<<<<<<<<<
 *             sorter = self.left_sorter
 *         else:
 */
    __pyx_t_2 = __pyx_v_self->left;
    __Pyx_INCREF(__pyx_t_2);
    __pyx_v_values = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "pandas/_libs/intervaltree.pxi":133
 *         if side == 'left':
 *             values = self.left
 *             sorter = self.left_sorter             # <<<<<<<<<<<<<<
 *         else:
 *             values = self.right
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_left_sorter); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 133, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_sorter = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "pandas/_libs/intervaltree.pxi":131
 *         falling between the left and right bounds of an interval query
 *         """
 *         if side == 'left':             # <<<<<<<<<<<<<<
 *             values = self.left
 *             sorter = self.left_sorter
 */
    goto __pyx_L3;
  }

  /* "pandas/_libs/intervaltree.pxi":135
 *             sorter = self.left_sorter
 *         else:
 *             values = self.right             # <<<<<<<<<<<<<<
 *             sorter = self.right_sorter
 *         key = [key_left, key_right]
 */
  /*else*/ {
    __pyx_t_2 = __pyx_v_self->right;
    __Pyx_INCREF(__pyx_t_2);
    __pyx_v_values = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "pandas/_libs/intervaltree.pxi":136
 *         else:
 *             values = self.right
 *             sorter = self.right_sorter             # <<<<<<<<<<<<<<
 *         key = [key_left, key_right]
 *         i, j = values.searchsorted(key, sorter=sorter)
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_right_sorter); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 136, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_sorter = __pyx_t_2;
    __pyx_t_2 = 0;
  }
  __pyx_L3:;

  /* "pandas/_libs/intervaltree.pxi":137
 *             values = self.right
 *             sorter = self.right_sorter
 *         key = [key_left, key_right]             # <<<<<<<<<<<<<<
 *         i, j = values.searchsorted(key, sorter=sorter)
 *         return sorter[i:j]
 */
  __pyx_t_2 = PyList_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 137, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_key_left);
  __Pyx_GIVEREF(__pyx_v_key_left);
  PyList_SET_ITEM(__pyx_t_2, 0, __pyx_v_key_left);
  __Pyx_INCREF(__pyx_v_key_right);
  __Pyx_GIVEREF(__pyx_v_key_right);
  PyList_SET_ITEM(__pyx_t_2, 1, __pyx_v_key_right);
  __pyx_v_key = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "pandas/_libs/intervaltree.pxi":138
 *             sorter = self.right_sorter
 *         key = [key_left, key_right]
 *         i, j = values.searchsorted(key, sorter=sorter)             # <<<<<<<<<<<<<<
 *         return sorter[i:j]
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_values, __pyx_n_s_searchsorted); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 138, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 138, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_v_key);
  __Pyx_GIVEREF(__pyx_v_key);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_key);
  __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 138, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_sorter, __pyx_v_sorter) < 0) __PYX_ERR(1, 138, __pyx_L1_error)
  __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 138, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if ((likely(PyTuple_CheckExact(__pyx_t_5))) || (PyList_CheckExact(__pyx_t_5))) {
    PyObject* sequence = __pyx_t_5;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 2)) {
      if (size > 2) __Pyx_RaiseTooManyValuesError(2);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(1, 138, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_4 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_3 = PyTuple_GET_ITEM(sequence, 1); 
    } else {
      __pyx_t_4 = PyList_GET_ITEM(sequence, 0); 
      __pyx_t_3 = PyList_GET_ITEM(sequence, 1); 
    }
    __Pyx_INCREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_t_3);
    #else
    __pyx_t_4 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 138, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 138, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  } else {
    Py_ssize_t index = -1;
    __pyx_t_2 = PyObject_GetIter(__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 138, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_6 = Py_TYPE(__pyx_t_2)->tp_iternext;
    index = 0; __pyx_t_4 = __pyx_t_6(__pyx_t_2); if (unlikely(!__pyx_t_4)) goto __pyx_L4_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_4);
    index = 1; __pyx_t_3 = __pyx_t_6(__pyx_t_2); if (unlikely(!__pyx_t_3)) goto __pyx_L4_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_3);
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_6(__pyx_t_2), 2) < 0) __PYX_ERR(1, 138, __pyx_L1_error)
    __pyx_t_6 = NULL;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    goto __pyx_L5_unpacking_done;
    __pyx_L4_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_6 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(1, 138, __pyx_L1_error)
    __pyx_L5_unpacking_done:;
  }
  __pyx_v_i = __pyx_t_4;
  __pyx_t_4 = 0;
  __pyx_v_j = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "pandas/_libs/intervaltree.pxi":139
 *         key = [key_left, key_right]
 *         i, j = values.searchsorted(key, sorter=sorter)
 *         return sorter[i:j]             # <<<<<<<<<<<<<<
 * 
 *     def get_loc_interval(self, key_left, key_right):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_5 = __Pyx_PyObject_GetSlice(__pyx_v_sorter, 0, 0, &__pyx_v_i, &__pyx_v_j, NULL, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 139, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_r = __pyx_t_5;
  __pyx_t_5 = 0;
  goto __pyx_L0;

  /* "pandas/_libs/intervaltree.pxi":127
 *         return result.to_array().astype('intp')
 * 
 *     def _get_partial_overlap(self, key_left, key_right, side):             # <<<<<<<<<<<<<<
 *         """Return all positions corresponding to intervals with the given side
 *         falling between the left and right bounds of an interval query
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pandas._libs.interval.IntervalTree._get_partial_overlap", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_values);
  __Pyx_XDECREF(__pyx_v_sorter);
  __Pyx_XDECREF(__pyx_v_key);
  __Pyx_XDECREF(__pyx_v_i);
  __Pyx_XDECREF(__pyx_v_j);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":141
 *         return sorter[i:j]
 * 
 *     def get_loc_interval(self, key_left, key_right):             # <<<<<<<<<<<<<<
 *         """Lookup the intervals enclosed in the given interval bounds
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_12IntervalTree_7get_loc_interval(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6pandas_5_libs_8interval_12IntervalTree_6get_loc_interval[] = "Lookup the intervals enclosed in the given interval bounds\n\n        The given interval is presumed to have closed bounds.\n        ";
static PyObject *__pyx_pw_6pandas_5_libs_8interval_12IntervalTree_7get_loc_interval(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_key_left = 0;
  PyObject *__pyx_v_key_right = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_loc_interval (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_key_left,&__pyx_n_s_key_right,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_key_left)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_key_right)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("get_loc_interval", 1, 2, 2, 1); __PYX_ERR(1, 141, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "get_loc_interval") < 0)) __PYX_ERR(1, 141, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_key_left = values[0];
    __pyx_v_key_right = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("get_loc_interval", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 141, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.IntervalTree.get_loc_interval", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_12IntervalTree_6get_loc_interval(((struct __pyx_obj_6pandas_5_libs_8interval_IntervalTree *)__pyx_v_self), __pyx_v_key_left, __pyx_v_key_right);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_12IntervalTree_6get_loc_interval(struct __pyx_obj_6pandas_5_libs_8interval_IntervalTree *__pyx_v_self, PyObject *__pyx_v_key_left, PyObject *__pyx_v_key_right) {
  PyObject *__pyx_v_pd = NULL;
  PyObject *__pyx_v_left_overlap = NULL;
  PyObject *__pyx_v_right_overlap = NULL;
  PyObject *__pyx_v_enclosing = NULL;
  PyObject *__pyx_v_combined = NULL;
  PyObject *__pyx_v_uniques = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("get_loc_interval", 0);

  /* "pandas/_libs/intervaltree.pxi":146
 *         The given interval is presumed to have closed bounds.
 *         """
 *         import pandas as pd             # <<<<<<<<<<<<<<
 *         left_overlap = self._get_partial_overlap(key_left, key_right, 'left')
 *         right_overlap = self._get_partial_overlap(key_left, key_right, 'right')
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_pandas, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 146, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_pd = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":147
 *         """
 *         import pandas as pd
 *         left_overlap = self._get_partial_overlap(key_left, key_right, 'left')             # <<<<<<<<<<<<<<
 *         right_overlap = self._get_partial_overlap(key_left, key_right, 'right')
 *         enclosing = self.get_loc(0.5 * (key_left + key_right))
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_get_partial_overlap); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 147, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[4] = {__pyx_t_3, __pyx_v_key_left, __pyx_v_key_right, __pyx_n_u_left};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_4, 3+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 147, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[4] = {__pyx_t_3, __pyx_v_key_left, __pyx_v_key_right, __pyx_n_u_left};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_4, 3+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 147, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(3+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 147, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_INCREF(__pyx_v_key_left);
    __Pyx_GIVEREF(__pyx_v_key_left);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_v_key_left);
    __Pyx_INCREF(__pyx_v_key_right);
    __Pyx_GIVEREF(__pyx_v_key_right);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_v_key_right);
    __Pyx_INCREF(__pyx_n_u_left);
    __Pyx_GIVEREF(__pyx_n_u_left);
    PyTuple_SET_ITEM(__pyx_t_5, 2+__pyx_t_4, __pyx_n_u_left);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 147, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_left_overlap = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":148
 *         import pandas as pd
 *         left_overlap = self._get_partial_overlap(key_left, key_right, 'left')
 *         right_overlap = self._get_partial_overlap(key_left, key_right, 'right')             # <<<<<<<<<<<<<<
 *         enclosing = self.get_loc(0.5 * (key_left + key_right))
 *         combined = np.concatenate([left_overlap, right_overlap, enclosing])
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_get_partial_overlap); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 148, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_key_left, __pyx_v_key_right, __pyx_n_u_right};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_4, 3+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 148, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_key_left, __pyx_v_key_right, __pyx_n_u_right};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_4, 3+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 148, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_3 = PyTuple_New(3+__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 148, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (__pyx_t_5) {
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5); __pyx_t_5 = NULL;
    }
    __Pyx_INCREF(__pyx_v_key_left);
    __Pyx_GIVEREF(__pyx_v_key_left);
    PyTuple_SET_ITEM(__pyx_t_3, 0+__pyx_t_4, __pyx_v_key_left);
    __Pyx_INCREF(__pyx_v_key_right);
    __Pyx_GIVEREF(__pyx_v_key_right);
    PyTuple_SET_ITEM(__pyx_t_3, 1+__pyx_t_4, __pyx_v_key_right);
    __Pyx_INCREF(__pyx_n_u_right);
    __Pyx_GIVEREF(__pyx_n_u_right);
    PyTuple_SET_ITEM(__pyx_t_3, 2+__pyx_t_4, __pyx_n_u_right);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 148, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_right_overlap = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":149
 *         left_overlap = self._get_partial_overlap(key_left, key_right, 'left')
 *         right_overlap = self._get_partial_overlap(key_left, key_right, 'right')
 *         enclosing = self.get_loc(0.5 * (key_left + key_right))             # <<<<<<<<<<<<<<
 *         combined = np.concatenate([left_overlap, right_overlap, enclosing])
 *         uniques = pd.unique(combined)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_get_loc); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 149, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyNumber_Add(__pyx_v_key_left, __pyx_v_key_right); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 149, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = PyNumber_Multiply(__pyx_float_0_5, __pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 149, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_5);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 149, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_enclosing = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":150
 *         right_overlap = self._get_partial_overlap(key_left, key_right, 'right')
 *         enclosing = self.get_loc(0.5 * (key_left + key_right))
 *         combined = np.concatenate([left_overlap, right_overlap, enclosing])             # <<<<<<<<<<<<<<
 *         uniques = pd.unique(combined)
 *         return uniques.astype('intp')
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 150, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_concatenate); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 150, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyList_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 150, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_left_overlap);
  __Pyx_GIVEREF(__pyx_v_left_overlap);
  PyList_SET_ITEM(__pyx_t_2, 0, __pyx_v_left_overlap);
  __Pyx_INCREF(__pyx_v_right_overlap);
  __Pyx_GIVEREF(__pyx_v_right_overlap);
  PyList_SET_ITEM(__pyx_t_2, 1, __pyx_v_right_overlap);
  __Pyx_INCREF(__pyx_v_enclosing);
  __Pyx_GIVEREF(__pyx_v_enclosing);
  PyList_SET_ITEM(__pyx_t_2, 2, __pyx_v_enclosing);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_3, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 150, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_v_combined = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":151
 *         enclosing = self.get_loc(0.5 * (key_left + key_right))
 *         combined = np.concatenate([left_overlap, right_overlap, enclosing])
 *         uniques = pd.unique(combined)             # <<<<<<<<<<<<<<
 *         return uniques.astype('intp')
 * 
 */
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_pd, __pyx_n_s_unique); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 151, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_2, __pyx_v_combined) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_v_combined);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 151, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_v_uniques = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":152
 *         combined = np.concatenate([left_overlap, right_overlap, enclosing])
 *         uniques = pd.unique(combined)
 *         return uniques.astype('intp')             # <<<<<<<<<<<<<<
 * 
 *     def get_indexer(self, scalar_t[:] target):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_uniques, __pyx_n_s_astype); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 152, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_2, __pyx_n_u_intp) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_n_u_intp);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 152, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pandas/_libs/intervaltree.pxi":141
 *         return sorter[i:j]
 * 
 *     def get_loc_interval(self, key_left, key_right):             # <<<<<<<<<<<<<<
 *         """Lookup the intervals enclosed in the given interval bounds
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pandas._libs.interval.IntervalTree.get_loc_interval", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_pd);
  __Pyx_XDECREF(__pyx_v_left_overlap);
  __Pyx_XDECREF(__pyx_v_right_overlap);
  __Pyx_XDECREF(__pyx_v_enclosing);
  __Pyx_XDECREF(__pyx_v_combined);
  __Pyx_XDECREF(__pyx_v_uniques);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":154
 *         return uniques.astype('intp')
 * 
 *     def get_indexer(self, scalar_t[:] target):             # <<<<<<<<<<<<<<
 *         """Return the positions corresponding to unique intervals that overlap
 *         with the given array of scalar targets.
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_12IntervalTree_9get_indexer(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6pandas_5_libs_8interval_12IntervalTree_8get_indexer[] = "Return the positions corresponding to unique intervals that overlap\n        with the given array of scalar targets.\n        ";
static PyMethodDef __pyx_mdef_6pandas_5_libs_8interval_12IntervalTree_9get_indexer = {"get_indexer", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_12IntervalTree_9get_indexer, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_12IntervalTree_8get_indexer};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_12IntervalTree_9get_indexer(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_args = 0;
  PyObject *__pyx_v_kwargs = 0;
  CYTHON_UNUSED PyObject *__pyx_v_defaults = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fused_cpdef (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_args,&__pyx_n_s_kwargs,&__pyx_n_s_defaults,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_args)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_kwargs)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fused_cpdef", 1, 3, 3, 1); __PYX_ERR(1, 154, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_defaults)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fused_cpdef", 1, 3, 3, 2); __PYX_ERR(1, 154, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fused_cpdef") < 0)) __PYX_ERR(1, 154, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_args = values[0];
    __pyx_v_kwargs = values[1];
    __pyx_v_defaults = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fused_cpdef", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 154, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.IntervalTree.__pyx_fused_cpdef", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_12IntervalTree_8get_indexer(((struct __pyx_obj_6pandas_5_libs_8interval_IntervalTree *)__pyx_v_signatures), __pyx_v_args, __pyx_v_kwargs, __pyx_v_defaults);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_12IntervalTree_8get_indexer(struct __pyx_obj_6pandas_5_libs_8interval_IntervalTree *__pyx_v_signatures, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs, CYTHON_UNUSED PyObject *__pyx_v_defaults) {
  PyObject *__pyx_v_dest_sig = NULL;
  Py_ssize_t __pyx_v_i;
  PyTypeObject *__pyx_v_ndarray = 0;
  __Pyx_memviewslice __pyx_v_memslice;
  Py_ssize_t __pyx_v_itemsize;
  int __pyx_v_dtype_signed;
  char __pyx_v_kind;
  int __pyx_v____pyx_int64_t_is_signed;
  int __pyx_v____pyx_int32_t_is_signed;
  int __pyx_v____pyx_uint64_t_is_signed;
  PyObject *__pyx_v_arg = NULL;
  PyObject *__pyx_v_dtype = NULL;
  PyObject *__pyx_v_arg_base = NULL;
  PyObject *__pyx_v_candidates = NULL;
  PyObject *__pyx_v_sig = NULL;
  int __pyx_v_match_found;
  PyObject *__pyx_v_src_sig = NULL;
  PyObject *__pyx_v_dst_type = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  long __pyx_t_7;
  __Pyx_memviewslice __pyx_t_8;
  Py_ssize_t __pyx_t_9;
  int __pyx_t_10;
  int __pyx_t_11;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  Py_ssize_t __pyx_t_15;
  Py_ssize_t __pyx_t_16;
  Py_ssize_t __pyx_t_17;
  int __pyx_t_18;
  __Pyx_RefNannySetupContext("get_indexer", 0);
  __Pyx_INCREF(__pyx_v_kwargs);
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 154, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  PyList_SET_ITEM(__pyx_t_1, 0, Py_None);
  __pyx_v_dest_sig = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_3 = (__pyx_v_kwargs != Py_None);
  __pyx_t_4 = (__pyx_t_3 != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_kwargs); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(1, 154, __pyx_L1_error)
  __pyx_t_3 = ((!__pyx_t_4) != 0);
  __pyx_t_2 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {
    __Pyx_INCREF(Py_None);
    __Pyx_DECREF_SET(__pyx_v_kwargs, Py_None);
  }
  __pyx_t_1 = ((PyObject *)__Pyx_ImportNumPyArrayTypeIfAvailable()); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 154, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_ndarray = ((PyTypeObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_v_itemsize = -1L;
  __pyx_v____pyx_int64_t_is_signed = (!((((__pyx_t_5numpy_int64_t)-1L) > 0) != 0));
  __pyx_v____pyx_int32_t_is_signed = (!((((__pyx_t_5numpy_int32_t)-1L) > 0) != 0));
  __pyx_v____pyx_uint64_t_is_signed = (!((((__pyx_t_5numpy_uint64_t)-1L) > 0) != 0));
  if (unlikely(__pyx_v_args == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 154, __pyx_L1_error)
  }
  __pyx_t_5 = PyTuple_GET_SIZE(((PyObject*)__pyx_v_args)); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(1, 154, __pyx_L1_error)
  __pyx_t_2 = ((1 < __pyx_t_5) != 0);
  if (__pyx_t_2) {
    if (unlikely(__pyx_v_args == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 154, __pyx_L1_error)
    }
    __pyx_t_1 = __Pyx_GetItemInt_Tuple(((PyObject*)__pyx_v_args), 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 154, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_arg = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L6;
  }
  __pyx_t_3 = (__pyx_v_kwargs != Py_None);
  __pyx_t_4 = (__pyx_t_3 != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L7_bool_binop_done;
  }
  if (unlikely(__pyx_v_kwargs == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(1, 154, __pyx_L1_error)
  }
  __pyx_t_4 = (__Pyx_PyDict_ContainsTF(__pyx_n_s_target, ((PyObject*)__pyx_v_kwargs), Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(1, 154, __pyx_L1_error)
  __pyx_t_3 = (__pyx_t_4 != 0);
  __pyx_t_2 = __pyx_t_3;
  __pyx_L7_bool_binop_done:;
  if (__pyx_t_2) {
    if (unlikely(__pyx_v_kwargs == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 154, __pyx_L1_error)
    }
    __pyx_t_1 = __Pyx_PyDict_GetItem(((PyObject*)__pyx_v_kwargs), __pyx_n_s_target); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 154, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_arg = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L6;
  }
  /*else*/ {
    if (unlikely(__pyx_v_args == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
      __PYX_ERR(1, 154, __pyx_L1_error)
    }
    __pyx_t_5 = PyTuple_GET_SIZE(((PyObject*)__pyx_v_args)); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(1, 154, __pyx_L1_error)
    __pyx_t_1 = PyInt_FromSsize_t(__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 154, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_6 = PyTuple_New(3); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 154, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_INCREF(__pyx_int_2);
    __Pyx_GIVEREF(__pyx_int_2);
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_int_2);
    __Pyx_INCREF(__pyx_n_s_s);
    __Pyx_GIVEREF(__pyx_n_s_s);
    PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_n_s_s);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Expected_at_least_d_argument_s_g, __pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 154, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_1); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 154, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(1, 154, __pyx_L1_error)
  }
  __pyx_L6:;
  while (1) {
    __pyx_t_2 = (__pyx_v_ndarray != ((PyTypeObject*)Py_None));
    __pyx_t_3 = (__pyx_t_2 != 0);
    if (__pyx_t_3) {
      __pyx_t_3 = __Pyx_TypeCheck(__pyx_v_arg, __pyx_v_ndarray); 
      __pyx_t_2 = (__pyx_t_3 != 0);
      if (__pyx_t_2) {
        __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_arg, __pyx_n_s_dtype); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 154, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_v_dtype = __pyx_t_6;
        __pyx_t_6 = 0;
        goto __pyx_L12;
      }
      __pyx_t_2 = __pyx_memoryview_check(__pyx_v_arg); 
      __pyx_t_3 = (__pyx_t_2 != 0);
      if (__pyx_t_3) {
        __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_arg, __pyx_n_s_base); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 154, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_v_arg_base = __pyx_t_6;
        __pyx_t_6 = 0;
        __pyx_t_3 = __Pyx_TypeCheck(__pyx_v_arg_base, __pyx_v_ndarray); 
        __pyx_t_2 = (__pyx_t_3 != 0);
        if (__pyx_t_2) {
          __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_arg_base, __pyx_n_s_dtype); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 154, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __pyx_v_dtype = __pyx_t_6;
          __pyx_t_6 = 0;
          goto __pyx_L13;
        }
        /*else*/ {
          __Pyx_INCREF(Py_None);
          __pyx_v_dtype = Py_None;
        }
        __pyx_L13:;
        goto __pyx_L12;
      }
      /*else*/ {
        __Pyx_INCREF(Py_None);
        __pyx_v_dtype = Py_None;
      }
      __pyx_L12:;
      __pyx_v_itemsize = -1L;
      __pyx_t_2 = (__pyx_v_dtype != Py_None);
      __pyx_t_3 = (__pyx_t_2 != 0);
      if (__pyx_t_3) {
        __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_dtype, __pyx_n_s_itemsize); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 154, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_5 = __Pyx_PyIndex_AsSsize_t(__pyx_t_6); if (unlikely((__pyx_t_5 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(1, 154, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_v_itemsize = __pyx_t_5;
        __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_dtype, __pyx_n_s_kind); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 154, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_7 = __Pyx_PyObject_Ord(__pyx_t_6); if (unlikely(__pyx_t_7 == ((long)(long)(Py_UCS4)-1))) __PYX_ERR(1, 154, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_v_kind = __pyx_t_7;
        __pyx_v_dtype_signed = (__pyx_v_kind == 'i');
        switch (__pyx_v_kind) {
          case 'i':
          case 'u':
          __pyx_t_2 = (((sizeof(__pyx_t_5numpy_int64_t)) == __pyx_v_itemsize) != 0);
          if (__pyx_t_2) {
          } else {
            __pyx_t_3 = __pyx_t_2;
            goto __pyx_L16_bool_binop_done;
          }
          __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_arg, __pyx_n_s_ndim); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 154, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __pyx_t_5 = __Pyx_PyIndex_AsSsize_t(__pyx_t_6); if (unlikely((__pyx_t_5 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(1, 154, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __pyx_t_2 = ((((Py_ssize_t)__pyx_t_5) == 1) != 0);
          if (__pyx_t_2) {
          } else {
            __pyx_t_3 = __pyx_t_2;
            goto __pyx_L16_bool_binop_done;
          }
          __pyx_t_2 = ((!((__pyx_v____pyx_int64_t_is_signed ^ __pyx_v_dtype_signed) != 0)) != 0);
          __pyx_t_3 = __pyx_t_2;
          __pyx_L16_bool_binop_done:;
          if (__pyx_t_3) {
            if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, __pyx_n_s_int64_t, long, 1, __Pyx_PyInt_From_long, 1, 0, 1) < 0)) __PYX_ERR(1, 154, __pyx_L1_error)
            goto __pyx_L10_break;
          }
          __pyx_t_2 = (((sizeof(__pyx_t_5numpy_int32_t)) == __pyx_v_itemsize) != 0);
          if (__pyx_t_2) {
          } else {
            __pyx_t_3 = __pyx_t_2;
            goto __pyx_L20_bool_binop_done;
          }
          __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_arg, __pyx_n_s_ndim); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 154, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __pyx_t_5 = __Pyx_PyIndex_AsSsize_t(__pyx_t_6); if (unlikely((__pyx_t_5 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(1, 154, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __pyx_t_2 = ((((Py_ssize_t)__pyx_t_5) == 1) != 0);
          if (__pyx_t_2) {
          } else {
            __pyx_t_3 = __pyx_t_2;
            goto __pyx_L20_bool_binop_done;
          }
          __pyx_t_2 = ((!((__pyx_v____pyx_int32_t_is_signed ^ __pyx_v_dtype_signed) != 0)) != 0);
          __pyx_t_3 = __pyx_t_2;
          __pyx_L20_bool_binop_done:;
          if (__pyx_t_3) {
            if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, __pyx_n_s_int32_t, long, 1, __Pyx_PyInt_From_long, 1, 0, 1) < 0)) __PYX_ERR(1, 154, __pyx_L1_error)
            goto __pyx_L10_break;
          }
          __pyx_t_2 = (((sizeof(__pyx_t_5numpy_uint64_t)) == __pyx_v_itemsize) != 0);
          if (__pyx_t_2) {
          } else {
            __pyx_t_3 = __pyx_t_2;
            goto __pyx_L24_bool_binop_done;
          }
          __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_arg, __pyx_n_s_ndim); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 154, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __pyx_t_5 = __Pyx_PyIndex_AsSsize_t(__pyx_t_6); if (unlikely((__pyx_t_5 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(1, 154, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __pyx_t_2 = ((((Py_ssize_t)__pyx_t_5) == 1) != 0);
          if (__pyx_t_2) {
          } else {
            __pyx_t_3 = __pyx_t_2;
            goto __pyx_L24_bool_binop_done;
          }
          __pyx_t_2 = ((!((__pyx_v____pyx_uint64_t_is_signed ^ __pyx_v_dtype_signed) != 0)) != 0);
          __pyx_t_3 = __pyx_t_2;
          __pyx_L24_bool_binop_done:;
          if (__pyx_t_3) {
            if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, __pyx_n_s_uint64_t, long, 1, __Pyx_PyInt_From_long, 1, 0, 1) < 0)) __PYX_ERR(1, 154, __pyx_L1_error)
            goto __pyx_L10_break;
          }
          break;
          case 'f':
          __pyx_t_2 = (((sizeof(__pyx_t_5numpy_float64_t)) == __pyx_v_itemsize) != 0);
          if (__pyx_t_2) {
          } else {
            __pyx_t_3 = __pyx_t_2;
            goto __pyx_L28_bool_binop_done;
          }
          __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_arg, __pyx_n_s_ndim); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 154, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __pyx_t_5 = __Pyx_PyIndex_AsSsize_t(__pyx_t_6); if (unlikely((__pyx_t_5 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(1, 154, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __pyx_t_2 = ((((Py_ssize_t)__pyx_t_5) == 1) != 0);
          __pyx_t_3 = __pyx_t_2;
          __pyx_L28_bool_binop_done:;
          if (__pyx_t_3) {
            if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, __pyx_n_s_float64_t, long, 1, __Pyx_PyInt_From_long, 1, 0, 1) < 0)) __PYX_ERR(1, 154, __pyx_L1_error)
            goto __pyx_L10_break;
          }
          __pyx_t_2 = (((sizeof(__pyx_t_5numpy_float32_t)) == __pyx_v_itemsize) != 0);
          if (__pyx_t_2) {
          } else {
            __pyx_t_3 = __pyx_t_2;
            goto __pyx_L31_bool_binop_done;
          }
          __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_arg, __pyx_n_s_ndim); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 154, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __pyx_t_5 = __Pyx_PyIndex_AsSsize_t(__pyx_t_6); if (unlikely((__pyx_t_5 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(1, 154, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __pyx_t_2 = ((((Py_ssize_t)__pyx_t_5) == 1) != 0);
          __pyx_t_3 = __pyx_t_2;
          __pyx_L31_bool_binop_done:;
          if (__pyx_t_3) {
            if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, __pyx_n_s_float32_t, long, 1, __Pyx_PyInt_From_long, 1, 0, 1) < 0)) __PYX_ERR(1, 154, __pyx_L1_error)
            goto __pyx_L10_break;
          }
          break;
          case 'c':
          break;
          case 'O':
          break;
          default: break;
        }
      }
    }
    __pyx_t_2 = ((__pyx_v_itemsize == -1L) != 0);
    if (!__pyx_t_2) {
    } else {
      __pyx_t_3 = __pyx_t_2;
      goto __pyx_L34_bool_binop_done;
    }
    __pyx_t_2 = ((__pyx_v_itemsize == (sizeof(__pyx_t_5numpy_float64_t))) != 0);
    __pyx_t_3 = __pyx_t_2;
    __pyx_L34_bool_binop_done:;
    if (__pyx_t_3) {
      __pyx_t_8 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_float64_t(__pyx_v_arg, 0); 
      __pyx_v_memslice = __pyx_t_8;
      __pyx_t_3 = (__pyx_v_memslice.memview != 0);
      if (__pyx_t_3) {
        __PYX_XDEC_MEMVIEW((&__pyx_v_memslice), 1); 
        if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, __pyx_n_s_float64_t, long, 1, __Pyx_PyInt_From_long, 1, 0, 1) < 0)) __PYX_ERR(1, 154, __pyx_L1_error)
        goto __pyx_L10_break;
      }
      /*else*/ {
        PyErr_Clear(); 
      }
    }
    __pyx_t_2 = ((__pyx_v_itemsize == -1L) != 0);
    if (!__pyx_t_2) {
    } else {
      __pyx_t_3 = __pyx_t_2;
      goto __pyx_L38_bool_binop_done;
    }
    __pyx_t_2 = ((__pyx_v_itemsize == (sizeof(__pyx_t_5numpy_float32_t))) != 0);
    __pyx_t_3 = __pyx_t_2;
    __pyx_L38_bool_binop_done:;
    if (__pyx_t_3) {
      __pyx_t_8 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_float32_t(__pyx_v_arg, 0); 
      __pyx_v_memslice = __pyx_t_8;
      __pyx_t_3 = (__pyx_v_memslice.memview != 0);
      if (__pyx_t_3) {
        __PYX_XDEC_MEMVIEW((&__pyx_v_memslice), 1); 
        if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, __pyx_n_s_float32_t, long, 1, __Pyx_PyInt_From_long, 1, 0, 1) < 0)) __PYX_ERR(1, 154, __pyx_L1_error)
        goto __pyx_L10_break;
      }
      /*else*/ {
        PyErr_Clear(); 
      }
    }
    __pyx_t_2 = ((__pyx_v_itemsize == -1L) != 0);
    if (!__pyx_t_2) {
    } else {
      __pyx_t_3 = __pyx_t_2;
      goto __pyx_L42_bool_binop_done;
    }
    __pyx_t_2 = ((__pyx_v_itemsize == (sizeof(__pyx_t_5numpy_int64_t))) != 0);
    __pyx_t_3 = __pyx_t_2;
    __pyx_L42_bool_binop_done:;
    if (__pyx_t_3) {
      __pyx_t_8 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_v_arg, 0); 
      __pyx_v_memslice = __pyx_t_8;
      __pyx_t_3 = (__pyx_v_memslice.memview != 0);
      if (__pyx_t_3) {
        __PYX_XDEC_MEMVIEW((&__pyx_v_memslice), 1); 
        if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, __pyx_n_s_int64_t, long, 1, __Pyx_PyInt_From_long, 1, 0, 1) < 0)) __PYX_ERR(1, 154, __pyx_L1_error)
        goto __pyx_L10_break;
      }
      /*else*/ {
        PyErr_Clear(); 
      }
    }
    __pyx_t_2 = ((__pyx_v_itemsize == -1L) != 0);
    if (!__pyx_t_2) {
    } else {
      __pyx_t_3 = __pyx_t_2;
      goto __pyx_L46_bool_binop_done;
    }
    __pyx_t_2 = ((__pyx_v_itemsize == (sizeof(__pyx_t_5numpy_int32_t))) != 0);
    __pyx_t_3 = __pyx_t_2;
    __pyx_L46_bool_binop_done:;
    if (__pyx_t_3) {
      __pyx_t_8 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int32_t(__pyx_v_arg, 0); 
      __pyx_v_memslice = __pyx_t_8;
      __pyx_t_3 = (__pyx_v_memslice.memview != 0);
      if (__pyx_t_3) {
        __PYX_XDEC_MEMVIEW((&__pyx_v_memslice), 1); 
        if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, __pyx_n_s_int32_t, long, 1, __Pyx_PyInt_From_long, 1, 0, 1) < 0)) __PYX_ERR(1, 154, __pyx_L1_error)
        goto __pyx_L10_break;
      }
      /*else*/ {
        PyErr_Clear(); 
      }
    }
    __pyx_t_2 = ((__pyx_v_itemsize == -1L) != 0);
    if (!__pyx_t_2) {
    } else {
      __pyx_t_3 = __pyx_t_2;
      goto __pyx_L50_bool_binop_done;
    }
    __pyx_t_2 = ((__pyx_v_itemsize == (sizeof(__pyx_t_5numpy_uint64_t))) != 0);
    __pyx_t_3 = __pyx_t_2;
    __pyx_L50_bool_binop_done:;
    if (__pyx_t_3) {
      __pyx_t_8 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_uint64_t(__pyx_v_arg, 0); 
      __pyx_v_memslice = __pyx_t_8;
      __pyx_t_3 = (__pyx_v_memslice.memview != 0);
      if (__pyx_t_3) {
        __PYX_XDEC_MEMVIEW((&__pyx_v_memslice), 1); 
        if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, __pyx_n_s_uint64_t, long, 1, __Pyx_PyInt_From_long, 1, 0, 1) < 0)) __PYX_ERR(1, 154, __pyx_L1_error)
        goto __pyx_L10_break;
      }
      /*else*/ {
        PyErr_Clear(); 
      }
    }
    if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, Py_None, long, 1, __Pyx_PyInt_From_long, 1, 0, 1) < 0)) __PYX_ERR(1, 154, __pyx_L1_error)
    goto __pyx_L10_break;
  }
  __pyx_L10_break:;
  __pyx_t_6 = PyList_New(0); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 154, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_v_candidates = ((PyObject*)__pyx_t_6);
  __pyx_t_6 = 0;
  __pyx_t_5 = 0;
  if (unlikely(((PyObject *)__pyx_v_signatures) == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(1, 154, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_dict_iterator(((PyObject*)__pyx_v_signatures), 1, ((PyObject *)NULL), (&__pyx_t_9), (&__pyx_t_10)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 154, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __pyx_t_6 = __pyx_t_1;
  __pyx_t_1 = 0;
  while (1) {
    __pyx_t_11 = __Pyx_dict_iter_next(__pyx_t_6, __pyx_t_9, &__pyx_t_5, &__pyx_t_1, NULL, NULL, __pyx_t_10);
    if (unlikely(__pyx_t_11 == 0)) break;
    if (unlikely(__pyx_t_11 == -1)) __PYX_ERR(1, 154, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_sig, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_v_match_found = 0;
    __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_v_sig, __pyx_n_s_strip); if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 154, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __pyx_t_14 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_13))) {
      __pyx_t_14 = PyMethod_GET_SELF(__pyx_t_13);
      if (likely(__pyx_t_14)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
        __Pyx_INCREF(__pyx_t_14);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_13, function);
      }
    }
    __pyx_t_12 = (__pyx_t_14) ? __Pyx_PyObject_Call2Args(__pyx_t_13, __pyx_t_14, __pyx_kp_s__13) : __Pyx_PyObject_CallOneArg(__pyx_t_13, __pyx_kp_s__13);
    __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
    if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 154, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_t_12, __pyx_n_s_split); if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 154, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    __pyx_t_12 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_13))) {
      __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_13);
      if (likely(__pyx_t_12)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
        __Pyx_INCREF(__pyx_t_12);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_13, function);
      }
    }
    __pyx_t_1 = (__pyx_t_12) ? __Pyx_PyObject_Call2Args(__pyx_t_13, __pyx_t_12, __pyx_kp_s__14) : __Pyx_PyObject_CallOneArg(__pyx_t_13, __pyx_kp_s__14);
    __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 154, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    __Pyx_XDECREF_SET(__pyx_v_src_sig, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_15 = PyList_GET_SIZE(__pyx_v_dest_sig); if (unlikely(__pyx_t_15 == ((Py_ssize_t)-1))) __PYX_ERR(1, 154, __pyx_L1_error)
    __pyx_t_16 = __pyx_t_15;
    for (__pyx_t_17 = 0; __pyx_t_17 < __pyx_t_16; __pyx_t_17+=1) {
      __pyx_v_i = __pyx_t_17;
      __pyx_t_1 = __Pyx_GetItemInt_List(__pyx_v_dest_sig, __pyx_v_i, Py_ssize_t, 1, PyInt_FromSsize_t, 1, 1, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 154, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_XDECREF_SET(__pyx_v_dst_type, __pyx_t_1);
      __pyx_t_1 = 0;
      __pyx_t_3 = (__pyx_v_dst_type != Py_None);
      __pyx_t_2 = (__pyx_t_3 != 0);
      if (__pyx_t_2) {
        __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_src_sig, __pyx_v_i, Py_ssize_t, 1, PyInt_FromSsize_t, 0, 1, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 154, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_13 = PyObject_RichCompare(__pyx_t_1, __pyx_v_dst_type, Py_EQ); __Pyx_XGOTREF(__pyx_t_13); if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 154, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_13); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(1, 154, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
        if (__pyx_t_2) {
          __pyx_v_match_found = 1;
          goto __pyx_L58;
        }
        /*else*/ {
          __pyx_v_match_found = 0;
          goto __pyx_L56_break;
        }
        __pyx_L58:;
      }
    }
    __pyx_L56_break:;
    __pyx_t_2 = (__pyx_v_match_found != 0);
    if (__pyx_t_2) {
      __pyx_t_18 = __Pyx_PyList_Append(__pyx_v_candidates, __pyx_v_sig); if (unlikely(__pyx_t_18 == ((int)-1))) __PYX_ERR(1, 154, __pyx_L1_error)
    }
  }
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_2 = (PyList_GET_SIZE(__pyx_v_candidates) != 0);
  __pyx_t_3 = ((!__pyx_t_2) != 0);
  if (__pyx_t_3) {
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__15, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 154, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(1, 154, __pyx_L1_error)
  }
  __pyx_t_9 = PyList_GET_SIZE(__pyx_v_candidates); if (unlikely(__pyx_t_9 == ((Py_ssize_t)-1))) __PYX_ERR(1, 154, __pyx_L1_error)
  __pyx_t_3 = ((__pyx_t_9 > 1) != 0);
  if (__pyx_t_3) {
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__16, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 154, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(1, 154, __pyx_L1_error)
  }
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    if (unlikely(((PyObject *)__pyx_v_signatures) == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 154, __pyx_L1_error)
    }
    __pyx_t_6 = __Pyx_GetItemInt_List(__pyx_v_candidates, 0, long, 1, __Pyx_PyInt_From_long, 1, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 154, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_13 = __Pyx_PyDict_GetItem(((PyObject*)__pyx_v_signatures), __pyx_t_6); if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 154, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_r = __pyx_t_13;
    __pyx_t_13 = 0;
    goto __pyx_L0;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_AddTraceback("pandas._libs.interval.IntervalTree.__pyx_fused_cpdef", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_dest_sig);
  __Pyx_XDECREF(__pyx_v_ndarray);
  __Pyx_XDECREF(__pyx_v_arg);
  __Pyx_XDECREF(__pyx_v_dtype);
  __Pyx_XDECREF(__pyx_v_arg_base);
  __Pyx_XDECREF(__pyx_v_candidates);
  __Pyx_XDECREF(__pyx_v_sig);
  __Pyx_XDECREF(__pyx_v_src_sig);
  __Pyx_XDECREF(__pyx_v_dst_type);
  __Pyx_XDECREF(__pyx_v_kwargs);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_fuse_0__pyx_pw_6pandas_5_libs_8interval_12IntervalTree_29get_indexer(PyObject *__pyx_v_self, PyObject *__pyx_arg_target); /*proto*/
static PyMethodDef __pyx_fuse_0__pyx_mdef_6pandas_5_libs_8interval_12IntervalTree_29get_indexer = {"__pyx_fuse_0get_indexer", (PyCFunction)__pyx_fuse_0__pyx_pw_6pandas_5_libs_8interval_12IntervalTree_29get_indexer, METH_O, __pyx_doc_6pandas_5_libs_8interval_12IntervalTree_8get_indexer};
static PyObject *__pyx_fuse_0__pyx_pw_6pandas_5_libs_8interval_12IntervalTree_29get_indexer(PyObject *__pyx_v_self, PyObject *__pyx_arg_target) {
  __Pyx_memviewslice __pyx_v_target = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_indexer (wrapper)", 0);
  assert(__pyx_arg_target); {
    __pyx_v_target = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_float64_t(__pyx_arg_target, PyBUF_WRITABLE); if (unlikely(!__pyx_v_target.memview)) __PYX_ERR(1, 154, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.IntervalTree.get_indexer", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_12IntervalTree_28get_indexer(((struct __pyx_obj_6pandas_5_libs_8interval_IntervalTree *)__pyx_v_self), __pyx_v_target);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_12IntervalTree_28get_indexer(struct __pyx_obj_6pandas_5_libs_8interval_IntervalTree *__pyx_v_self, __Pyx_memviewslice __pyx_v_target) {
  size_t __pyx_v_old_len;
  Py_ssize_t __pyx_v_i;
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  size_t __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  Py_ssize_t __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  __Pyx_RefNannySetupContext("__pyx_fuse_0get_indexer", 0);

  /* "pandas/_libs/intervaltree.pxi":165
 *             Int64Vector result
 * 
 *         result = Int64Vector()             # <<<<<<<<<<<<<<
 *         old_len = 0
 *         for i in range(len(target)):
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 165, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_result = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":166
 * 
 *         result = Int64Vector()
 *         old_len = 0             # <<<<<<<<<<<<<<
 *         for i in range(len(target)):
 *             self.root.query(result, target[i])
 */
  __pyx_v_old_len = 0;

  /* "pandas/_libs/intervaltree.pxi":167
 *         result = Int64Vector()
 *         old_len = 0
 *         for i in range(len(target)):             # <<<<<<<<<<<<<<
 *             self.root.query(result, target[i])
 *             if result.data.n == old_len:
 */
  __pyx_t_2 = __Pyx_MemoryView_Len(__pyx_v_target); 
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "pandas/_libs/intervaltree.pxi":168
 *         old_len = 0
 *         for i in range(len(target)):
 *             self.root.query(result, target[i])             # <<<<<<<<<<<<<<
 *             if result.data.n == old_len:
 *                 result.append(-1)
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->root, __pyx_n_s_query); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 168, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __pyx_v_i;
    __pyx_t_7 = -1;
    if (__pyx_t_6 < 0) {
      __pyx_t_6 += __pyx_v_target.shape[0];
      if (unlikely(__pyx_t_6 < 0)) __pyx_t_7 = 0;
    } else if (unlikely(__pyx_t_6 >= __pyx_v_target.shape[0])) __pyx_t_7 = 0;
    if (unlikely(__pyx_t_7 != -1)) {
      __Pyx_RaiseBufferIndexError(__pyx_t_7);
      __PYX_ERR(1, 168, __pyx_L1_error)
    }
    __pyx_t_8 = PyFloat_FromDouble((*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_target.data + __pyx_t_6 * __pyx_v_target.strides[0]) )))); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 168, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_9 = NULL;
    __pyx_t_7 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_7 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[3] = {__pyx_t_9, ((PyObject *)__pyx_v_result), __pyx_t_8};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 168, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[3] = {__pyx_t_9, ((PyObject *)__pyx_v_result), __pyx_t_8};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 168, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    } else
    #endif
    {
      __pyx_t_10 = PyTuple_New(2+__pyx_t_7); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 168, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      if (__pyx_t_9) {
        __Pyx_GIVEREF(__pyx_t_9); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_9); __pyx_t_9 = NULL;
      }
      __Pyx_INCREF(((PyObject *)__pyx_v_result));
      __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
      PyTuple_SET_ITEM(__pyx_t_10, 0+__pyx_t_7, ((PyObject *)__pyx_v_result));
      __Pyx_GIVEREF(__pyx_t_8);
      PyTuple_SET_ITEM(__pyx_t_10, 1+__pyx_t_7, __pyx_t_8);
      __pyx_t_8 = 0;
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_10, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 168, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "pandas/_libs/intervaltree.pxi":169
 *         for i in range(len(target)):
 *             self.root.query(result, target[i])
 *             if result.data.n == old_len:             # <<<<<<<<<<<<<<
 *                 result.append(-1)
 *             elif result.data.n > old_len + 1:
 */
    __pyx_t_11 = ((__pyx_v_result->data->n == __pyx_v_old_len) != 0);
    if (__pyx_t_11) {

      /* "pandas/_libs/intervaltree.pxi":170
 *             self.root.query(result, target[i])
 *             if result.data.n == old_len:
 *                 result.append(-1)             # <<<<<<<<<<<<<<
 *             elif result.data.n > old_len + 1:
 *                 raise KeyError(
 */
      ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, -1LL);

      /* "pandas/_libs/intervaltree.pxi":169
 *         for i in range(len(target)):
 *             self.root.query(result, target[i])
 *             if result.data.n == old_len:             # <<<<<<<<<<<<<<
 *                 result.append(-1)
 *             elif result.data.n > old_len + 1:
 */
      goto __pyx_L5;
    }

    /* "pandas/_libs/intervaltree.pxi":171
 *             if result.data.n == old_len:
 *                 result.append(-1)
 *             elif result.data.n > old_len + 1:             # <<<<<<<<<<<<<<
 *                 raise KeyError(
 *                     'indexer does not intersect a unique set of intervals')
 */
    __pyx_t_11 = ((__pyx_v_result->data->n > (__pyx_v_old_len + 1)) != 0);
    if (unlikely(__pyx_t_11)) {

      /* "pandas/_libs/intervaltree.pxi":172
 *                 result.append(-1)
 *             elif result.data.n > old_len + 1:
 *                 raise KeyError(             # <<<<<<<<<<<<<<
 *                     'indexer does not intersect a unique set of intervals')
 *             old_len = result.data.n
 */
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_KeyError, __pyx_tuple__17, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 172, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_Raise(__pyx_t_1, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __PYX_ERR(1, 172, __pyx_L1_error)

      /* "pandas/_libs/intervaltree.pxi":171
 *             if result.data.n == old_len:
 *                 result.append(-1)
 *             elif result.data.n > old_len + 1:             # <<<<<<<<<<<<<<
 *                 raise KeyError(
 *                     'indexer does not intersect a unique set of intervals')
 */
    }
    __pyx_L5:;

    /* "pandas/_libs/intervaltree.pxi":174
 *                 raise KeyError(
 *                     'indexer does not intersect a unique set of intervals')
 *             old_len = result.data.n             # <<<<<<<<<<<<<<
 *         return result.to_array().astype('intp')
 * 
 */
    __pyx_t_12 = __pyx_v_result->data->n;
    __pyx_v_old_len = __pyx_t_12;
  }

  /* "pandas/_libs/intervaltree.pxi":175
 *                     'indexer does not intersect a unique set of intervals')
 *             old_len = result.data.n
 *         return result.to_array().astype('intp')             # <<<<<<<<<<<<<<
 * 
 *     def get_indexer_non_unique(self, scalar_t[:] target):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_5 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->to_array(__pyx_v_result, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 175, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_astype); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 175, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_10))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_10);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_10, function);
    }
  }
  __pyx_t_1 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_10, __pyx_t_5, __pyx_n_u_intp) : __Pyx_PyObject_CallOneArg(__pyx_t_10, __pyx_n_u_intp);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 175, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pandas/_libs/intervaltree.pxi":154
 *         return uniques.astype('intp')
 * 
 *     def get_indexer(self, scalar_t[:] target):             # <<<<<<<<<<<<<<
 *         """Return the positions corresponding to unique intervals that overlap
 *         with the given array of scalar targets.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("pandas._libs.interval.IntervalTree.get_indexer", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_target, 1);
  __Pyx_XDECREF((PyObject *)__pyx_v_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_fuse_1__pyx_pw_6pandas_5_libs_8interval_12IntervalTree_31get_indexer(PyObject *__pyx_v_self, PyObject *__pyx_arg_target); /*proto*/
static PyMethodDef __pyx_fuse_1__pyx_mdef_6pandas_5_libs_8interval_12IntervalTree_31get_indexer = {"__pyx_fuse_1get_indexer", (PyCFunction)__pyx_fuse_1__pyx_pw_6pandas_5_libs_8interval_12IntervalTree_31get_indexer, METH_O, __pyx_doc_6pandas_5_libs_8interval_12IntervalTree_8get_indexer};
static PyObject *__pyx_fuse_1__pyx_pw_6pandas_5_libs_8interval_12IntervalTree_31get_indexer(PyObject *__pyx_v_self, PyObject *__pyx_arg_target) {
  __Pyx_memviewslice __pyx_v_target = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_indexer (wrapper)", 0);
  assert(__pyx_arg_target); {
    __pyx_v_target = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_float32_t(__pyx_arg_target, PyBUF_WRITABLE); if (unlikely(!__pyx_v_target.memview)) __PYX_ERR(1, 154, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.IntervalTree.get_indexer", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_12IntervalTree_30get_indexer(((struct __pyx_obj_6pandas_5_libs_8interval_IntervalTree *)__pyx_v_self), __pyx_v_target);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_12IntervalTree_30get_indexer(struct __pyx_obj_6pandas_5_libs_8interval_IntervalTree *__pyx_v_self, __Pyx_memviewslice __pyx_v_target) {
  size_t __pyx_v_old_len;
  Py_ssize_t __pyx_v_i;
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  size_t __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  Py_ssize_t __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  __Pyx_RefNannySetupContext("__pyx_fuse_1get_indexer", 0);

  /* "pandas/_libs/intervaltree.pxi":165
 *             Int64Vector result
 * 
 *         result = Int64Vector()             # <<<<<<<<<<<<<<
 *         old_len = 0
 *         for i in range(len(target)):
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 165, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_result = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":166
 * 
 *         result = Int64Vector()
 *         old_len = 0             # <<<<<<<<<<<<<<
 *         for i in range(len(target)):
 *             self.root.query(result, target[i])
 */
  __pyx_v_old_len = 0;

  /* "pandas/_libs/intervaltree.pxi":167
 *         result = Int64Vector()
 *         old_len = 0
 *         for i in range(len(target)):             # <<<<<<<<<<<<<<
 *             self.root.query(result, target[i])
 *             if result.data.n == old_len:
 */
  __pyx_t_2 = __Pyx_MemoryView_Len(__pyx_v_target); 
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "pandas/_libs/intervaltree.pxi":168
 *         old_len = 0
 *         for i in range(len(target)):
 *             self.root.query(result, target[i])             # <<<<<<<<<<<<<<
 *             if result.data.n == old_len:
 *                 result.append(-1)
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->root, __pyx_n_s_query); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 168, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __pyx_v_i;
    __pyx_t_7 = -1;
    if (__pyx_t_6 < 0) {
      __pyx_t_6 += __pyx_v_target.shape[0];
      if (unlikely(__pyx_t_6 < 0)) __pyx_t_7 = 0;
    } else if (unlikely(__pyx_t_6 >= __pyx_v_target.shape[0])) __pyx_t_7 = 0;
    if (unlikely(__pyx_t_7 != -1)) {
      __Pyx_RaiseBufferIndexError(__pyx_t_7);
      __PYX_ERR(1, 168, __pyx_L1_error)
    }
    __pyx_t_8 = PyFloat_FromDouble((*((__pyx_t_5numpy_float32_t *) ( /* dim=0 */ (__pyx_v_target.data + __pyx_t_6 * __pyx_v_target.strides[0]) )))); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 168, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_9 = NULL;
    __pyx_t_7 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_7 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[3] = {__pyx_t_9, ((PyObject *)__pyx_v_result), __pyx_t_8};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 168, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[3] = {__pyx_t_9, ((PyObject *)__pyx_v_result), __pyx_t_8};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 168, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    } else
    #endif
    {
      __pyx_t_10 = PyTuple_New(2+__pyx_t_7); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 168, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      if (__pyx_t_9) {
        __Pyx_GIVEREF(__pyx_t_9); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_9); __pyx_t_9 = NULL;
      }
      __Pyx_INCREF(((PyObject *)__pyx_v_result));
      __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
      PyTuple_SET_ITEM(__pyx_t_10, 0+__pyx_t_7, ((PyObject *)__pyx_v_result));
      __Pyx_GIVEREF(__pyx_t_8);
      PyTuple_SET_ITEM(__pyx_t_10, 1+__pyx_t_7, __pyx_t_8);
      __pyx_t_8 = 0;
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_10, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 168, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "pandas/_libs/intervaltree.pxi":169
 *         for i in range(len(target)):
 *             self.root.query(result, target[i])
 *             if result.data.n == old_len:             # <<<<<<<<<<<<<<
 *                 result.append(-1)
 *             elif result.data.n > old_len + 1:
 */
    __pyx_t_11 = ((__pyx_v_result->data->n == __pyx_v_old_len) != 0);
    if (__pyx_t_11) {

      /* "pandas/_libs/intervaltree.pxi":170
 *             self.root.query(result, target[i])
 *             if result.data.n == old_len:
 *                 result.append(-1)             # <<<<<<<<<<<<<<
 *             elif result.data.n > old_len + 1:
 *                 raise KeyError(
 */
      ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, -1LL);

      /* "pandas/_libs/intervaltree.pxi":169
 *         for i in range(len(target)):
 *             self.root.query(result, target[i])
 *             if result.data.n == old_len:             # <<<<<<<<<<<<<<
 *                 result.append(-1)
 *             elif result.data.n > old_len + 1:
 */
      goto __pyx_L5;
    }

    /* "pandas/_libs/intervaltree.pxi":171
 *             if result.data.n == old_len:
 *                 result.append(-1)
 *             elif result.data.n > old_len + 1:             # <<<<<<<<<<<<<<
 *                 raise KeyError(
 *                     'indexer does not intersect a unique set of intervals')
 */
    __pyx_t_11 = ((__pyx_v_result->data->n > (__pyx_v_old_len + 1)) != 0);
    if (unlikely(__pyx_t_11)) {

      /* "pandas/_libs/intervaltree.pxi":172
 *                 result.append(-1)
 *             elif result.data.n > old_len + 1:
 *                 raise KeyError(             # <<<<<<<<<<<<<<
 *                     'indexer does not intersect a unique set of intervals')
 *             old_len = result.data.n
 */
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_KeyError, __pyx_tuple__17, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 172, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_Raise(__pyx_t_1, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __PYX_ERR(1, 172, __pyx_L1_error)

      /* "pandas/_libs/intervaltree.pxi":171
 *             if result.data.n == old_len:
 *                 result.append(-1)
 *             elif result.data.n > old_len + 1:             # <<<<<<<<<<<<<<
 *                 raise KeyError(
 *                     'indexer does not intersect a unique set of intervals')
 */
    }
    __pyx_L5:;

    /* "pandas/_libs/intervaltree.pxi":174
 *                 raise KeyError(
 *                     'indexer does not intersect a unique set of intervals')
 *             old_len = result.data.n             # <<<<<<<<<<<<<<
 *         return result.to_array().astype('intp')
 * 
 */
    __pyx_t_12 = __pyx_v_result->data->n;
    __pyx_v_old_len = __pyx_t_12;
  }

  /* "pandas/_libs/intervaltree.pxi":175
 *                     'indexer does not intersect a unique set of intervals')
 *             old_len = result.data.n
 *         return result.to_array().astype('intp')             # <<<<<<<<<<<<<<
 * 
 *     def get_indexer_non_unique(self, scalar_t[:] target):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_5 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->to_array(__pyx_v_result, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 175, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_astype); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 175, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_10))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_10);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_10, function);
    }
  }
  __pyx_t_1 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_10, __pyx_t_5, __pyx_n_u_intp) : __Pyx_PyObject_CallOneArg(__pyx_t_10, __pyx_n_u_intp);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 175, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pandas/_libs/intervaltree.pxi":154
 *         return uniques.astype('intp')
 * 
 *     def get_indexer(self, scalar_t[:] target):             # <<<<<<<<<<<<<<
 *         """Return the positions corresponding to unique intervals that overlap
 *         with the given array of scalar targets.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("pandas._libs.interval.IntervalTree.get_indexer", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_target, 1);
  __Pyx_XDECREF((PyObject *)__pyx_v_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_fuse_2__pyx_pw_6pandas_5_libs_8interval_12IntervalTree_33get_indexer(PyObject *__pyx_v_self, PyObject *__pyx_arg_target); /*proto*/
static PyMethodDef __pyx_fuse_2__pyx_mdef_6pandas_5_libs_8interval_12IntervalTree_33get_indexer = {"__pyx_fuse_2get_indexer", (PyCFunction)__pyx_fuse_2__pyx_pw_6pandas_5_libs_8interval_12IntervalTree_33get_indexer, METH_O, __pyx_doc_6pandas_5_libs_8interval_12IntervalTree_8get_indexer};
static PyObject *__pyx_fuse_2__pyx_pw_6pandas_5_libs_8interval_12IntervalTree_33get_indexer(PyObject *__pyx_v_self, PyObject *__pyx_arg_target) {
  __Pyx_memviewslice __pyx_v_target = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_indexer (wrapper)", 0);
  assert(__pyx_arg_target); {
    __pyx_v_target = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_arg_target, PyBUF_WRITABLE); if (unlikely(!__pyx_v_target.memview)) __PYX_ERR(1, 154, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.IntervalTree.get_indexer", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_12IntervalTree_32get_indexer(((struct __pyx_obj_6pandas_5_libs_8interval_IntervalTree *)__pyx_v_self), __pyx_v_target);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_12IntervalTree_32get_indexer(struct __pyx_obj_6pandas_5_libs_8interval_IntervalTree *__pyx_v_self, __Pyx_memviewslice __pyx_v_target) {
  size_t __pyx_v_old_len;
  Py_ssize_t __pyx_v_i;
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  size_t __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  Py_ssize_t __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  __Pyx_RefNannySetupContext("__pyx_fuse_2get_indexer", 0);

  /* "pandas/_libs/intervaltree.pxi":165
 *             Int64Vector result
 * 
 *         result = Int64Vector()             # <<<<<<<<<<<<<<
 *         old_len = 0
 *         for i in range(len(target)):
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 165, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_result = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":166
 * 
 *         result = Int64Vector()
 *         old_len = 0             # <<<<<<<<<<<<<<
 *         for i in range(len(target)):
 *             self.root.query(result, target[i])
 */
  __pyx_v_old_len = 0;

  /* "pandas/_libs/intervaltree.pxi":167
 *         result = Int64Vector()
 *         old_len = 0
 *         for i in range(len(target)):             # <<<<<<<<<<<<<<
 *             self.root.query(result, target[i])
 *             if result.data.n == old_len:
 */
  __pyx_t_2 = __Pyx_MemoryView_Len(__pyx_v_target); 
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "pandas/_libs/intervaltree.pxi":168
 *         old_len = 0
 *         for i in range(len(target)):
 *             self.root.query(result, target[i])             # <<<<<<<<<<<<<<
 *             if result.data.n == old_len:
 *                 result.append(-1)
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->root, __pyx_n_s_query); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 168, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __pyx_v_i;
    __pyx_t_7 = -1;
    if (__pyx_t_6 < 0) {
      __pyx_t_6 += __pyx_v_target.shape[0];
      if (unlikely(__pyx_t_6 < 0)) __pyx_t_7 = 0;
    } else if (unlikely(__pyx_t_6 >= __pyx_v_target.shape[0])) __pyx_t_7 = 0;
    if (unlikely(__pyx_t_7 != -1)) {
      __Pyx_RaiseBufferIndexError(__pyx_t_7);
      __PYX_ERR(1, 168, __pyx_L1_error)
    }
    __pyx_t_8 = __Pyx_PyInt_From_npy_int64((*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_target.data + __pyx_t_6 * __pyx_v_target.strides[0]) )))); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 168, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_9 = NULL;
    __pyx_t_7 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_7 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[3] = {__pyx_t_9, ((PyObject *)__pyx_v_result), __pyx_t_8};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 168, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[3] = {__pyx_t_9, ((PyObject *)__pyx_v_result), __pyx_t_8};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 168, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    } else
    #endif
    {
      __pyx_t_10 = PyTuple_New(2+__pyx_t_7); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 168, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      if (__pyx_t_9) {
        __Pyx_GIVEREF(__pyx_t_9); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_9); __pyx_t_9 = NULL;
      }
      __Pyx_INCREF(((PyObject *)__pyx_v_result));
      __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
      PyTuple_SET_ITEM(__pyx_t_10, 0+__pyx_t_7, ((PyObject *)__pyx_v_result));
      __Pyx_GIVEREF(__pyx_t_8);
      PyTuple_SET_ITEM(__pyx_t_10, 1+__pyx_t_7, __pyx_t_8);
      __pyx_t_8 = 0;
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_10, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 168, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "pandas/_libs/intervaltree.pxi":169
 *         for i in range(len(target)):
 *             self.root.query(result, target[i])
 *             if result.data.n == old_len:             # <<<<<<<<<<<<<<
 *                 result.append(-1)
 *             elif result.data.n > old_len + 1:
 */
    __pyx_t_11 = ((__pyx_v_result->data->n == __pyx_v_old_len) != 0);
    if (__pyx_t_11) {

      /* "pandas/_libs/intervaltree.pxi":170
 *             self.root.query(result, target[i])
 *             if result.data.n == old_len:
 *                 result.append(-1)             # <<<<<<<<<<<<<<
 *             elif result.data.n > old_len + 1:
 *                 raise KeyError(
 */
      ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, -1LL);

      /* "pandas/_libs/intervaltree.pxi":169
 *         for i in range(len(target)):
 *             self.root.query(result, target[i])
 *             if result.data.n == old_len:             # <<<<<<<<<<<<<<
 *                 result.append(-1)
 *             elif result.data.n > old_len + 1:
 */
      goto __pyx_L5;
    }

    /* "pandas/_libs/intervaltree.pxi":171
 *             if result.data.n == old_len:
 *                 result.append(-1)
 *             elif result.data.n > old_len + 1:             # <<<<<<<<<<<<<<
 *                 raise KeyError(
 *                     'indexer does not intersect a unique set of intervals')
 */
    __pyx_t_11 = ((__pyx_v_result->data->n > (__pyx_v_old_len + 1)) != 0);
    if (unlikely(__pyx_t_11)) {

      /* "pandas/_libs/intervaltree.pxi":172
 *                 result.append(-1)
 *             elif result.data.n > old_len + 1:
 *                 raise KeyError(             # <<<<<<<<<<<<<<
 *                     'indexer does not intersect a unique set of intervals')
 *             old_len = result.data.n
 */
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_KeyError, __pyx_tuple__17, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 172, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_Raise(__pyx_t_1, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __PYX_ERR(1, 172, __pyx_L1_error)

      /* "pandas/_libs/intervaltree.pxi":171
 *             if result.data.n == old_len:
 *                 result.append(-1)
 *             elif result.data.n > old_len + 1:             # <<<<<<<<<<<<<<
 *                 raise KeyError(
 *                     'indexer does not intersect a unique set of intervals')
 */
    }
    __pyx_L5:;

    /* "pandas/_libs/intervaltree.pxi":174
 *                 raise KeyError(
 *                     'indexer does not intersect a unique set of intervals')
 *             old_len = result.data.n             # <<<<<<<<<<<<<<
 *         return result.to_array().astype('intp')
 * 
 */
    __pyx_t_12 = __pyx_v_result->data->n;
    __pyx_v_old_len = __pyx_t_12;
  }

  /* "pandas/_libs/intervaltree.pxi":175
 *                     'indexer does not intersect a unique set of intervals')
 *             old_len = result.data.n
 *         return result.to_array().astype('intp')             # <<<<<<<<<<<<<<
 * 
 *     def get_indexer_non_unique(self, scalar_t[:] target):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_5 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->to_array(__pyx_v_result, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 175, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_astype); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 175, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_10))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_10);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_10, function);
    }
  }
  __pyx_t_1 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_10, __pyx_t_5, __pyx_n_u_intp) : __Pyx_PyObject_CallOneArg(__pyx_t_10, __pyx_n_u_intp);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 175, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pandas/_libs/intervaltree.pxi":154
 *         return uniques.astype('intp')
 * 
 *     def get_indexer(self, scalar_t[:] target):             # <<<<<<<<<<<<<<
 *         """Return the positions corresponding to unique intervals that overlap
 *         with the given array of scalar targets.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("pandas._libs.interval.IntervalTree.get_indexer", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_target, 1);
  __Pyx_XDECREF((PyObject *)__pyx_v_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_fuse_3__pyx_pw_6pandas_5_libs_8interval_12IntervalTree_35get_indexer(PyObject *__pyx_v_self, PyObject *__pyx_arg_target); /*proto*/
static PyMethodDef __pyx_fuse_3__pyx_mdef_6pandas_5_libs_8interval_12IntervalTree_35get_indexer = {"__pyx_fuse_3get_indexer", (PyCFunction)__pyx_fuse_3__pyx_pw_6pandas_5_libs_8interval_12IntervalTree_35get_indexer, METH_O, __pyx_doc_6pandas_5_libs_8interval_12IntervalTree_8get_indexer};
static PyObject *__pyx_fuse_3__pyx_pw_6pandas_5_libs_8interval_12IntervalTree_35get_indexer(PyObject *__pyx_v_self, PyObject *__pyx_arg_target) {
  __Pyx_memviewslice __pyx_v_target = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_indexer (wrapper)", 0);
  assert(__pyx_arg_target); {
    __pyx_v_target = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int32_t(__pyx_arg_target, PyBUF_WRITABLE); if (unlikely(!__pyx_v_target.memview)) __PYX_ERR(1, 154, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.IntervalTree.get_indexer", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_12IntervalTree_34get_indexer(((struct __pyx_obj_6pandas_5_libs_8interval_IntervalTree *)__pyx_v_self), __pyx_v_target);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_12IntervalTree_34get_indexer(struct __pyx_obj_6pandas_5_libs_8interval_IntervalTree *__pyx_v_self, __Pyx_memviewslice __pyx_v_target) {
  size_t __pyx_v_old_len;
  Py_ssize_t __pyx_v_i;
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  size_t __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  Py_ssize_t __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  __Pyx_RefNannySetupContext("__pyx_fuse_3get_indexer", 0);

  /* "pandas/_libs/intervaltree.pxi":165
 *             Int64Vector result
 * 
 *         result = Int64Vector()             # <<<<<<<<<<<<<<
 *         old_len = 0
 *         for i in range(len(target)):
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 165, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_result = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":166
 * 
 *         result = Int64Vector()
 *         old_len = 0             # <<<<<<<<<<<<<<
 *         for i in range(len(target)):
 *             self.root.query(result, target[i])
 */
  __pyx_v_old_len = 0;

  /* "pandas/_libs/intervaltree.pxi":167
 *         result = Int64Vector()
 *         old_len = 0
 *         for i in range(len(target)):             # <<<<<<<<<<<<<<
 *             self.root.query(result, target[i])
 *             if result.data.n == old_len:
 */
  __pyx_t_2 = __Pyx_MemoryView_Len(__pyx_v_target); 
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "pandas/_libs/intervaltree.pxi":168
 *         old_len = 0
 *         for i in range(len(target)):
 *             self.root.query(result, target[i])             # <<<<<<<<<<<<<<
 *             if result.data.n == old_len:
 *                 result.append(-1)
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->root, __pyx_n_s_query); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 168, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __pyx_v_i;
    __pyx_t_7 = -1;
    if (__pyx_t_6 < 0) {
      __pyx_t_6 += __pyx_v_target.shape[0];
      if (unlikely(__pyx_t_6 < 0)) __pyx_t_7 = 0;
    } else if (unlikely(__pyx_t_6 >= __pyx_v_target.shape[0])) __pyx_t_7 = 0;
    if (unlikely(__pyx_t_7 != -1)) {
      __Pyx_RaiseBufferIndexError(__pyx_t_7);
      __PYX_ERR(1, 168, __pyx_L1_error)
    }
    __pyx_t_8 = __Pyx_PyInt_From_npy_int32((*((__pyx_t_5numpy_int32_t *) ( /* dim=0 */ (__pyx_v_target.data + __pyx_t_6 * __pyx_v_target.strides[0]) )))); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 168, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_9 = NULL;
    __pyx_t_7 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_7 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[3] = {__pyx_t_9, ((PyObject *)__pyx_v_result), __pyx_t_8};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 168, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[3] = {__pyx_t_9, ((PyObject *)__pyx_v_result), __pyx_t_8};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 168, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    } else
    #endif
    {
      __pyx_t_10 = PyTuple_New(2+__pyx_t_7); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 168, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      if (__pyx_t_9) {
        __Pyx_GIVEREF(__pyx_t_9); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_9); __pyx_t_9 = NULL;
      }
      __Pyx_INCREF(((PyObject *)__pyx_v_result));
      __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
      PyTuple_SET_ITEM(__pyx_t_10, 0+__pyx_t_7, ((PyObject *)__pyx_v_result));
      __Pyx_GIVEREF(__pyx_t_8);
      PyTuple_SET_ITEM(__pyx_t_10, 1+__pyx_t_7, __pyx_t_8);
      __pyx_t_8 = 0;
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_10, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 168, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "pandas/_libs/intervaltree.pxi":169
 *         for i in range(len(target)):
 *             self.root.query(result, target[i])
 *             if result.data.n == old_len:             # <<<<<<<<<<<<<<
 *                 result.append(-1)
 *             elif result.data.n > old_len + 1:
 */
    __pyx_t_11 = ((__pyx_v_result->data->n == __pyx_v_old_len) != 0);
    if (__pyx_t_11) {

      /* "pandas/_libs/intervaltree.pxi":170
 *             self.root.query(result, target[i])
 *             if result.data.n == old_len:
 *                 result.append(-1)             # <<<<<<<<<<<<<<
 *             elif result.data.n > old_len + 1:
 *                 raise KeyError(
 */
      ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, -1LL);

      /* "pandas/_libs/intervaltree.pxi":169
 *         for i in range(len(target)):
 *             self.root.query(result, target[i])
 *             if result.data.n == old_len:             # <<<<<<<<<<<<<<
 *                 result.append(-1)
 *             elif result.data.n > old_len + 1:
 */
      goto __pyx_L5;
    }

    /* "pandas/_libs/intervaltree.pxi":171
 *             if result.data.n == old_len:
 *                 result.append(-1)
 *             elif result.data.n > old_len + 1:             # <<<<<<<<<<<<<<
 *                 raise KeyError(
 *                     'indexer does not intersect a unique set of intervals')
 */
    __pyx_t_11 = ((__pyx_v_result->data->n > (__pyx_v_old_len + 1)) != 0);
    if (unlikely(__pyx_t_11)) {

      /* "pandas/_libs/intervaltree.pxi":172
 *                 result.append(-1)
 *             elif result.data.n > old_len + 1:
 *                 raise KeyError(             # <<<<<<<<<<<<<<
 *                     'indexer does not intersect a unique set of intervals')
 *             old_len = result.data.n
 */
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_KeyError, __pyx_tuple__17, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 172, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_Raise(__pyx_t_1, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __PYX_ERR(1, 172, __pyx_L1_error)

      /* "pandas/_libs/intervaltree.pxi":171
 *             if result.data.n == old_len:
 *                 result.append(-1)
 *             elif result.data.n > old_len + 1:             # <<<<<<<<<<<<<<
 *                 raise KeyError(
 *                     'indexer does not intersect a unique set of intervals')
 */
    }
    __pyx_L5:;

    /* "pandas/_libs/intervaltree.pxi":174
 *                 raise KeyError(
 *                     'indexer does not intersect a unique set of intervals')
 *             old_len = result.data.n             # <<<<<<<<<<<<<<
 *         return result.to_array().astype('intp')
 * 
 */
    __pyx_t_12 = __pyx_v_result->data->n;
    __pyx_v_old_len = __pyx_t_12;
  }

  /* "pandas/_libs/intervaltree.pxi":175
 *                     'indexer does not intersect a unique set of intervals')
 *             old_len = result.data.n
 *         return result.to_array().astype('intp')             # <<<<<<<<<<<<<<
 * 
 *     def get_indexer_non_unique(self, scalar_t[:] target):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_5 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->to_array(__pyx_v_result, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 175, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_astype); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 175, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_10))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_10);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_10, function);
    }
  }
  __pyx_t_1 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_10, __pyx_t_5, __pyx_n_u_intp) : __Pyx_PyObject_CallOneArg(__pyx_t_10, __pyx_n_u_intp);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 175, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pandas/_libs/intervaltree.pxi":154
 *         return uniques.astype('intp')
 * 
 *     def get_indexer(self, scalar_t[:] target):             # <<<<<<<<<<<<<<
 *         """Return the positions corresponding to unique intervals that overlap
 *         with the given array of scalar targets.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("pandas._libs.interval.IntervalTree.get_indexer", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_target, 1);
  __Pyx_XDECREF((PyObject *)__pyx_v_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_fuse_4__pyx_pw_6pandas_5_libs_8interval_12IntervalTree_37get_indexer(PyObject *__pyx_v_self, PyObject *__pyx_arg_target); /*proto*/
static PyMethodDef __pyx_fuse_4__pyx_mdef_6pandas_5_libs_8interval_12IntervalTree_37get_indexer = {"__pyx_fuse_4get_indexer", (PyCFunction)__pyx_fuse_4__pyx_pw_6pandas_5_libs_8interval_12IntervalTree_37get_indexer, METH_O, __pyx_doc_6pandas_5_libs_8interval_12IntervalTree_8get_indexer};
static PyObject *__pyx_fuse_4__pyx_pw_6pandas_5_libs_8interval_12IntervalTree_37get_indexer(PyObject *__pyx_v_self, PyObject *__pyx_arg_target) {
  __Pyx_memviewslice __pyx_v_target = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_indexer (wrapper)", 0);
  assert(__pyx_arg_target); {
    __pyx_v_target = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_uint64_t(__pyx_arg_target, PyBUF_WRITABLE); if (unlikely(!__pyx_v_target.memview)) __PYX_ERR(1, 154, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.IntervalTree.get_indexer", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_12IntervalTree_36get_indexer(((struct __pyx_obj_6pandas_5_libs_8interval_IntervalTree *)__pyx_v_self), __pyx_v_target);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_12IntervalTree_36get_indexer(struct __pyx_obj_6pandas_5_libs_8interval_IntervalTree *__pyx_v_self, __Pyx_memviewslice __pyx_v_target) {
  size_t __pyx_v_old_len;
  Py_ssize_t __pyx_v_i;
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  size_t __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  Py_ssize_t __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  __Pyx_RefNannySetupContext("__pyx_fuse_4get_indexer", 0);

  /* "pandas/_libs/intervaltree.pxi":165
 *             Int64Vector result
 * 
 *         result = Int64Vector()             # <<<<<<<<<<<<<<
 *         old_len = 0
 *         for i in range(len(target)):
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 165, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_result = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":166
 * 
 *         result = Int64Vector()
 *         old_len = 0             # <<<<<<<<<<<<<<
 *         for i in range(len(target)):
 *             self.root.query(result, target[i])
 */
  __pyx_v_old_len = 0;

  /* "pandas/_libs/intervaltree.pxi":167
 *         result = Int64Vector()
 *         old_len = 0
 *         for i in range(len(target)):             # <<<<<<<<<<<<<<
 *             self.root.query(result, target[i])
 *             if result.data.n == old_len:
 */
  __pyx_t_2 = __Pyx_MemoryView_Len(__pyx_v_target); 
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "pandas/_libs/intervaltree.pxi":168
 *         old_len = 0
 *         for i in range(len(target)):
 *             self.root.query(result, target[i])             # <<<<<<<<<<<<<<
 *             if result.data.n == old_len:
 *                 result.append(-1)
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->root, __pyx_n_s_query); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 168, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __pyx_v_i;
    __pyx_t_7 = -1;
    if (__pyx_t_6 < 0) {
      __pyx_t_6 += __pyx_v_target.shape[0];
      if (unlikely(__pyx_t_6 < 0)) __pyx_t_7 = 0;
    } else if (unlikely(__pyx_t_6 >= __pyx_v_target.shape[0])) __pyx_t_7 = 0;
    if (unlikely(__pyx_t_7 != -1)) {
      __Pyx_RaiseBufferIndexError(__pyx_t_7);
      __PYX_ERR(1, 168, __pyx_L1_error)
    }
    __pyx_t_8 = __Pyx_PyInt_From_npy_uint64((*((__pyx_t_5numpy_uint64_t *) ( /* dim=0 */ (__pyx_v_target.data + __pyx_t_6 * __pyx_v_target.strides[0]) )))); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 168, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_9 = NULL;
    __pyx_t_7 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_7 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[3] = {__pyx_t_9, ((PyObject *)__pyx_v_result), __pyx_t_8};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 168, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[3] = {__pyx_t_9, ((PyObject *)__pyx_v_result), __pyx_t_8};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 168, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    } else
    #endif
    {
      __pyx_t_10 = PyTuple_New(2+__pyx_t_7); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 168, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      if (__pyx_t_9) {
        __Pyx_GIVEREF(__pyx_t_9); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_9); __pyx_t_9 = NULL;
      }
      __Pyx_INCREF(((PyObject *)__pyx_v_result));
      __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
      PyTuple_SET_ITEM(__pyx_t_10, 0+__pyx_t_7, ((PyObject *)__pyx_v_result));
      __Pyx_GIVEREF(__pyx_t_8);
      PyTuple_SET_ITEM(__pyx_t_10, 1+__pyx_t_7, __pyx_t_8);
      __pyx_t_8 = 0;
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_10, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 168, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "pandas/_libs/intervaltree.pxi":169
 *         for i in range(len(target)):
 *             self.root.query(result, target[i])
 *             if result.data.n == old_len:             # <<<<<<<<<<<<<<
 *                 result.append(-1)
 *             elif result.data.n > old_len + 1:
 */
    __pyx_t_11 = ((__pyx_v_result->data->n == __pyx_v_old_len) != 0);
    if (__pyx_t_11) {

      /* "pandas/_libs/intervaltree.pxi":170
 *             self.root.query(result, target[i])
 *             if result.data.n == old_len:
 *                 result.append(-1)             # <<<<<<<<<<<<<<
 *             elif result.data.n > old_len + 1:
 *                 raise KeyError(
 */
      ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, -1LL);

      /* "pandas/_libs/intervaltree.pxi":169
 *         for i in range(len(target)):
 *             self.root.query(result, target[i])
 *             if result.data.n == old_len:             # <<<<<<<<<<<<<<
 *                 result.append(-1)
 *             elif result.data.n > old_len + 1:
 */
      goto __pyx_L5;
    }

    /* "pandas/_libs/intervaltree.pxi":171
 *             if result.data.n == old_len:
 *                 result.append(-1)
 *             elif result.data.n > old_len + 1:             # <<<<<<<<<<<<<<
 *                 raise KeyError(
 *                     'indexer does not intersect a unique set of intervals')
 */
    __pyx_t_11 = ((__pyx_v_result->data->n > (__pyx_v_old_len + 1)) != 0);
    if (unlikely(__pyx_t_11)) {

      /* "pandas/_libs/intervaltree.pxi":172
 *                 result.append(-1)
 *             elif result.data.n > old_len + 1:
 *                 raise KeyError(             # <<<<<<<<<<<<<<
 *                     'indexer does not intersect a unique set of intervals')
 *             old_len = result.data.n
 */
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_KeyError, __pyx_tuple__17, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 172, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_Raise(__pyx_t_1, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __PYX_ERR(1, 172, __pyx_L1_error)

      /* "pandas/_libs/intervaltree.pxi":171
 *             if result.data.n == old_len:
 *                 result.append(-1)
 *             elif result.data.n > old_len + 1:             # <<<<<<<<<<<<<<
 *                 raise KeyError(
 *                     'indexer does not intersect a unique set of intervals')
 */
    }
    __pyx_L5:;

    /* "pandas/_libs/intervaltree.pxi":174
 *                 raise KeyError(
 *                     'indexer does not intersect a unique set of intervals')
 *             old_len = result.data.n             # <<<<<<<<<<<<<<
 *         return result.to_array().astype('intp')
 * 
 */
    __pyx_t_12 = __pyx_v_result->data->n;
    __pyx_v_old_len = __pyx_t_12;
  }

  /* "pandas/_libs/intervaltree.pxi":175
 *                     'indexer does not intersect a unique set of intervals')
 *             old_len = result.data.n
 *         return result.to_array().astype('intp')             # <<<<<<<<<<<<<<
 * 
 *     def get_indexer_non_unique(self, scalar_t[:] target):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_5 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->to_array(__pyx_v_result, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 175, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_astype); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 175, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_10))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_10);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_10, function);
    }
  }
  __pyx_t_1 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_10, __pyx_t_5, __pyx_n_u_intp) : __Pyx_PyObject_CallOneArg(__pyx_t_10, __pyx_n_u_intp);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 175, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pandas/_libs/intervaltree.pxi":154
 *         return uniques.astype('intp')
 * 
 *     def get_indexer(self, scalar_t[:] target):             # <<<<<<<<<<<<<<
 *         """Return the positions corresponding to unique intervals that overlap
 *         with the given array of scalar targets.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("pandas._libs.interval.IntervalTree.get_indexer", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_target, 1);
  __Pyx_XDECREF((PyObject *)__pyx_v_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":177
 *         return result.to_array().astype('intp')
 * 
 *     def get_indexer_non_unique(self, scalar_t[:] target):             # <<<<<<<<<<<<<<
 *         """Return the positions corresponding to intervals that overlap with
 *         the given array of scalar targets. Non-unique positions are repeated.
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_12IntervalTree_11get_indexer_non_unique(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6pandas_5_libs_8interval_12IntervalTree_10get_indexer_non_unique[] = "Return the positions corresponding to intervals that overlap with\n        the given array of scalar targets. Non-unique positions are repeated.\n        ";
static PyMethodDef __pyx_mdef_6pandas_5_libs_8interval_12IntervalTree_11get_indexer_non_unique = {"get_indexer_non_unique", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_12IntervalTree_11get_indexer_non_unique, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_12IntervalTree_10get_indexer_non_unique};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_12IntervalTree_11get_indexer_non_unique(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_args = 0;
  PyObject *__pyx_v_kwargs = 0;
  CYTHON_UNUSED PyObject *__pyx_v_defaults = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fused_cpdef (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_args,&__pyx_n_s_kwargs,&__pyx_n_s_defaults,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_args)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_kwargs)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fused_cpdef", 1, 3, 3, 1); __PYX_ERR(1, 177, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_defaults)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fused_cpdef", 1, 3, 3, 2); __PYX_ERR(1, 177, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fused_cpdef") < 0)) __PYX_ERR(1, 177, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_args = values[0];
    __pyx_v_kwargs = values[1];
    __pyx_v_defaults = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fused_cpdef", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 177, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.IntervalTree.__pyx_fused_cpdef", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_12IntervalTree_10get_indexer_non_unique(((struct __pyx_obj_6pandas_5_libs_8interval_IntervalTree *)__pyx_v_signatures), __pyx_v_args, __pyx_v_kwargs, __pyx_v_defaults);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_12IntervalTree_10get_indexer_non_unique(struct __pyx_obj_6pandas_5_libs_8interval_IntervalTree *__pyx_v_signatures, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs, CYTHON_UNUSED PyObject *__pyx_v_defaults) {
  PyObject *__pyx_v_dest_sig = NULL;
  Py_ssize_t __pyx_v_i;
  PyTypeObject *__pyx_v_ndarray = 0;
  __Pyx_memviewslice __pyx_v_memslice;
  Py_ssize_t __pyx_v_itemsize;
  int __pyx_v_dtype_signed;
  char __pyx_v_kind;
  int __pyx_v____pyx_int64_t_is_signed;
  int __pyx_v____pyx_int32_t_is_signed;
  int __pyx_v____pyx_uint64_t_is_signed;
  PyObject *__pyx_v_arg = NULL;
  PyObject *__pyx_v_dtype = NULL;
  PyObject *__pyx_v_arg_base = NULL;
  PyObject *__pyx_v_candidates = NULL;
  PyObject *__pyx_v_sig = NULL;
  int __pyx_v_match_found;
  PyObject *__pyx_v_src_sig = NULL;
  PyObject *__pyx_v_dst_type = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  long __pyx_t_7;
  __Pyx_memviewslice __pyx_t_8;
  Py_ssize_t __pyx_t_9;
  int __pyx_t_10;
  int __pyx_t_11;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  Py_ssize_t __pyx_t_15;
  Py_ssize_t __pyx_t_16;
  Py_ssize_t __pyx_t_17;
  int __pyx_t_18;
  __Pyx_RefNannySetupContext("get_indexer_non_unique", 0);
  __Pyx_INCREF(__pyx_v_kwargs);
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 177, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  PyList_SET_ITEM(__pyx_t_1, 0, Py_None);
  __pyx_v_dest_sig = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_3 = (__pyx_v_kwargs != Py_None);
  __pyx_t_4 = (__pyx_t_3 != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_kwargs); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(1, 177, __pyx_L1_error)
  __pyx_t_3 = ((!__pyx_t_4) != 0);
  __pyx_t_2 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {
    __Pyx_INCREF(Py_None);
    __Pyx_DECREF_SET(__pyx_v_kwargs, Py_None);
  }
  __pyx_t_1 = ((PyObject *)__Pyx_ImportNumPyArrayTypeIfAvailable()); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 177, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_ndarray = ((PyTypeObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_v_itemsize = -1L;
  __pyx_v____pyx_int64_t_is_signed = (!((((__pyx_t_5numpy_int64_t)-1L) > 0) != 0));
  __pyx_v____pyx_int32_t_is_signed = (!((((__pyx_t_5numpy_int32_t)-1L) > 0) != 0));
  __pyx_v____pyx_uint64_t_is_signed = (!((((__pyx_t_5numpy_uint64_t)-1L) > 0) != 0));
  if (unlikely(__pyx_v_args == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 177, __pyx_L1_error)
  }
  __pyx_t_5 = PyTuple_GET_SIZE(((PyObject*)__pyx_v_args)); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(1, 177, __pyx_L1_error)
  __pyx_t_2 = ((1 < __pyx_t_5) != 0);
  if (__pyx_t_2) {
    if (unlikely(__pyx_v_args == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 177, __pyx_L1_error)
    }
    __pyx_t_1 = __Pyx_GetItemInt_Tuple(((PyObject*)__pyx_v_args), 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 177, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_arg = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L6;
  }
  __pyx_t_3 = (__pyx_v_kwargs != Py_None);
  __pyx_t_4 = (__pyx_t_3 != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L7_bool_binop_done;
  }
  if (unlikely(__pyx_v_kwargs == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(1, 177, __pyx_L1_error)
  }
  __pyx_t_4 = (__Pyx_PyDict_ContainsTF(__pyx_n_s_target, ((PyObject*)__pyx_v_kwargs), Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(1, 177, __pyx_L1_error)
  __pyx_t_3 = (__pyx_t_4 != 0);
  __pyx_t_2 = __pyx_t_3;
  __pyx_L7_bool_binop_done:;
  if (__pyx_t_2) {
    if (unlikely(__pyx_v_kwargs == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 177, __pyx_L1_error)
    }
    __pyx_t_1 = __Pyx_PyDict_GetItem(((PyObject*)__pyx_v_kwargs), __pyx_n_s_target); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 177, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_arg = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L6;
  }
  /*else*/ {
    if (unlikely(__pyx_v_args == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
      __PYX_ERR(1, 177, __pyx_L1_error)
    }
    __pyx_t_5 = PyTuple_GET_SIZE(((PyObject*)__pyx_v_args)); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(1, 177, __pyx_L1_error)
    __pyx_t_1 = PyInt_FromSsize_t(__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 177, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_6 = PyTuple_New(3); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 177, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_INCREF(__pyx_int_2);
    __Pyx_GIVEREF(__pyx_int_2);
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_int_2);
    __Pyx_INCREF(__pyx_n_s_s);
    __Pyx_GIVEREF(__pyx_n_s_s);
    PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_n_s_s);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Expected_at_least_d_argument_s_g, __pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 177, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_1); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 177, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(1, 177, __pyx_L1_error)
  }
  __pyx_L6:;
  while (1) {
    __pyx_t_2 = (__pyx_v_ndarray != ((PyTypeObject*)Py_None));
    __pyx_t_3 = (__pyx_t_2 != 0);
    if (__pyx_t_3) {
      __pyx_t_3 = __Pyx_TypeCheck(__pyx_v_arg, __pyx_v_ndarray); 
      __pyx_t_2 = (__pyx_t_3 != 0);
      if (__pyx_t_2) {
        __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_arg, __pyx_n_s_dtype); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 177, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_v_dtype = __pyx_t_6;
        __pyx_t_6 = 0;
        goto __pyx_L12;
      }
      __pyx_t_2 = __pyx_memoryview_check(__pyx_v_arg); 
      __pyx_t_3 = (__pyx_t_2 != 0);
      if (__pyx_t_3) {
        __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_arg, __pyx_n_s_base); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 177, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_v_arg_base = __pyx_t_6;
        __pyx_t_6 = 0;
        __pyx_t_3 = __Pyx_TypeCheck(__pyx_v_arg_base, __pyx_v_ndarray); 
        __pyx_t_2 = (__pyx_t_3 != 0);
        if (__pyx_t_2) {
          __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_arg_base, __pyx_n_s_dtype); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 177, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __pyx_v_dtype = __pyx_t_6;
          __pyx_t_6 = 0;
          goto __pyx_L13;
        }
        /*else*/ {
          __Pyx_INCREF(Py_None);
          __pyx_v_dtype = Py_None;
        }
        __pyx_L13:;
        goto __pyx_L12;
      }
      /*else*/ {
        __Pyx_INCREF(Py_None);
        __pyx_v_dtype = Py_None;
      }
      __pyx_L12:;
      __pyx_v_itemsize = -1L;
      __pyx_t_2 = (__pyx_v_dtype != Py_None);
      __pyx_t_3 = (__pyx_t_2 != 0);
      if (__pyx_t_3) {
        __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_dtype, __pyx_n_s_itemsize); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 177, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_5 = __Pyx_PyIndex_AsSsize_t(__pyx_t_6); if (unlikely((__pyx_t_5 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(1, 177, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_v_itemsize = __pyx_t_5;
        __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_dtype, __pyx_n_s_kind); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 177, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_7 = __Pyx_PyObject_Ord(__pyx_t_6); if (unlikely(__pyx_t_7 == ((long)(long)(Py_UCS4)-1))) __PYX_ERR(1, 177, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_v_kind = __pyx_t_7;
        __pyx_v_dtype_signed = (__pyx_v_kind == 'i');
        switch (__pyx_v_kind) {
          case 'i':
          case 'u':
          __pyx_t_2 = (((sizeof(__pyx_t_5numpy_int64_t)) == __pyx_v_itemsize) != 0);
          if (__pyx_t_2) {
          } else {
            __pyx_t_3 = __pyx_t_2;
            goto __pyx_L16_bool_binop_done;
          }
          __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_arg, __pyx_n_s_ndim); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 177, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __pyx_t_5 = __Pyx_PyIndex_AsSsize_t(__pyx_t_6); if (unlikely((__pyx_t_5 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(1, 177, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __pyx_t_2 = ((((Py_ssize_t)__pyx_t_5) == 1) != 0);
          if (__pyx_t_2) {
          } else {
            __pyx_t_3 = __pyx_t_2;
            goto __pyx_L16_bool_binop_done;
          }
          __pyx_t_2 = ((!((__pyx_v____pyx_int64_t_is_signed ^ __pyx_v_dtype_signed) != 0)) != 0);
          __pyx_t_3 = __pyx_t_2;
          __pyx_L16_bool_binop_done:;
          if (__pyx_t_3) {
            if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, __pyx_n_s_int64_t, long, 1, __Pyx_PyInt_From_long, 1, 0, 1) < 0)) __PYX_ERR(1, 177, __pyx_L1_error)
            goto __pyx_L10_break;
          }
          __pyx_t_2 = (((sizeof(__pyx_t_5numpy_int32_t)) == __pyx_v_itemsize) != 0);
          if (__pyx_t_2) {
          } else {
            __pyx_t_3 = __pyx_t_2;
            goto __pyx_L20_bool_binop_done;
          }
          __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_arg, __pyx_n_s_ndim); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 177, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __pyx_t_5 = __Pyx_PyIndex_AsSsize_t(__pyx_t_6); if (unlikely((__pyx_t_5 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(1, 177, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __pyx_t_2 = ((((Py_ssize_t)__pyx_t_5) == 1) != 0);
          if (__pyx_t_2) {
          } else {
            __pyx_t_3 = __pyx_t_2;
            goto __pyx_L20_bool_binop_done;
          }
          __pyx_t_2 = ((!((__pyx_v____pyx_int32_t_is_signed ^ __pyx_v_dtype_signed) != 0)) != 0);
          __pyx_t_3 = __pyx_t_2;
          __pyx_L20_bool_binop_done:;
          if (__pyx_t_3) {
            if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, __pyx_n_s_int32_t, long, 1, __Pyx_PyInt_From_long, 1, 0, 1) < 0)) __PYX_ERR(1, 177, __pyx_L1_error)
            goto __pyx_L10_break;
          }
          __pyx_t_2 = (((sizeof(__pyx_t_5numpy_uint64_t)) == __pyx_v_itemsize) != 0);
          if (__pyx_t_2) {
          } else {
            __pyx_t_3 = __pyx_t_2;
            goto __pyx_L24_bool_binop_done;
          }
          __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_arg, __pyx_n_s_ndim); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 177, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __pyx_t_5 = __Pyx_PyIndex_AsSsize_t(__pyx_t_6); if (unlikely((__pyx_t_5 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(1, 177, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __pyx_t_2 = ((((Py_ssize_t)__pyx_t_5) == 1) != 0);
          if (__pyx_t_2) {
          } else {
            __pyx_t_3 = __pyx_t_2;
            goto __pyx_L24_bool_binop_done;
          }
          __pyx_t_2 = ((!((__pyx_v____pyx_uint64_t_is_signed ^ __pyx_v_dtype_signed) != 0)) != 0);
          __pyx_t_3 = __pyx_t_2;
          __pyx_L24_bool_binop_done:;
          if (__pyx_t_3) {
            if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, __pyx_n_s_uint64_t, long, 1, __Pyx_PyInt_From_long, 1, 0, 1) < 0)) __PYX_ERR(1, 177, __pyx_L1_error)
            goto __pyx_L10_break;
          }
          break;
          case 'f':
          __pyx_t_2 = (((sizeof(__pyx_t_5numpy_float64_t)) == __pyx_v_itemsize) != 0);
          if (__pyx_t_2) {
          } else {
            __pyx_t_3 = __pyx_t_2;
            goto __pyx_L28_bool_binop_done;
          }
          __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_arg, __pyx_n_s_ndim); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 177, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __pyx_t_5 = __Pyx_PyIndex_AsSsize_t(__pyx_t_6); if (unlikely((__pyx_t_5 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(1, 177, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __pyx_t_2 = ((((Py_ssize_t)__pyx_t_5) == 1) != 0);
          __pyx_t_3 = __pyx_t_2;
          __pyx_L28_bool_binop_done:;
          if (__pyx_t_3) {
            if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, __pyx_n_s_float64_t, long, 1, __Pyx_PyInt_From_long, 1, 0, 1) < 0)) __PYX_ERR(1, 177, __pyx_L1_error)
            goto __pyx_L10_break;
          }
          __pyx_t_2 = (((sizeof(__pyx_t_5numpy_float32_t)) == __pyx_v_itemsize) != 0);
          if (__pyx_t_2) {
          } else {
            __pyx_t_3 = __pyx_t_2;
            goto __pyx_L31_bool_binop_done;
          }
          __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_arg, __pyx_n_s_ndim); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 177, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __pyx_t_5 = __Pyx_PyIndex_AsSsize_t(__pyx_t_6); if (unlikely((__pyx_t_5 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(1, 177, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __pyx_t_2 = ((((Py_ssize_t)__pyx_t_5) == 1) != 0);
          __pyx_t_3 = __pyx_t_2;
          __pyx_L31_bool_binop_done:;
          if (__pyx_t_3) {
            if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, __pyx_n_s_float32_t, long, 1, __Pyx_PyInt_From_long, 1, 0, 1) < 0)) __PYX_ERR(1, 177, __pyx_L1_error)
            goto __pyx_L10_break;
          }
          break;
          case 'c':
          break;
          case 'O':
          break;
          default: break;
        }
      }
    }
    __pyx_t_2 = ((__pyx_v_itemsize == -1L) != 0);
    if (!__pyx_t_2) {
    } else {
      __pyx_t_3 = __pyx_t_2;
      goto __pyx_L34_bool_binop_done;
    }
    __pyx_t_2 = ((__pyx_v_itemsize == (sizeof(__pyx_t_5numpy_float64_t))) != 0);
    __pyx_t_3 = __pyx_t_2;
    __pyx_L34_bool_binop_done:;
    if (__pyx_t_3) {
      __pyx_t_8 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_float64_t(__pyx_v_arg, 0); 
      __pyx_v_memslice = __pyx_t_8;
      __pyx_t_3 = (__pyx_v_memslice.memview != 0);
      if (__pyx_t_3) {
        __PYX_XDEC_MEMVIEW((&__pyx_v_memslice), 1); 
        if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, __pyx_n_s_float64_t, long, 1, __Pyx_PyInt_From_long, 1, 0, 1) < 0)) __PYX_ERR(1, 177, __pyx_L1_error)
        goto __pyx_L10_break;
      }
      /*else*/ {
        PyErr_Clear(); 
      }
    }
    __pyx_t_2 = ((__pyx_v_itemsize == -1L) != 0);
    if (!__pyx_t_2) {
    } else {
      __pyx_t_3 = __pyx_t_2;
      goto __pyx_L38_bool_binop_done;
    }
    __pyx_t_2 = ((__pyx_v_itemsize == (sizeof(__pyx_t_5numpy_float32_t))) != 0);
    __pyx_t_3 = __pyx_t_2;
    __pyx_L38_bool_binop_done:;
    if (__pyx_t_3) {
      __pyx_t_8 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_float32_t(__pyx_v_arg, 0); 
      __pyx_v_memslice = __pyx_t_8;
      __pyx_t_3 = (__pyx_v_memslice.memview != 0);
      if (__pyx_t_3) {
        __PYX_XDEC_MEMVIEW((&__pyx_v_memslice), 1); 
        if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, __pyx_n_s_float32_t, long, 1, __Pyx_PyInt_From_long, 1, 0, 1) < 0)) __PYX_ERR(1, 177, __pyx_L1_error)
        goto __pyx_L10_break;
      }
      /*else*/ {
        PyErr_Clear(); 
      }
    }
    __pyx_t_2 = ((__pyx_v_itemsize == -1L) != 0);
    if (!__pyx_t_2) {
    } else {
      __pyx_t_3 = __pyx_t_2;
      goto __pyx_L42_bool_binop_done;
    }
    __pyx_t_2 = ((__pyx_v_itemsize == (sizeof(__pyx_t_5numpy_int64_t))) != 0);
    __pyx_t_3 = __pyx_t_2;
    __pyx_L42_bool_binop_done:;
    if (__pyx_t_3) {
      __pyx_t_8 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_v_arg, 0); 
      __pyx_v_memslice = __pyx_t_8;
      __pyx_t_3 = (__pyx_v_memslice.memview != 0);
      if (__pyx_t_3) {
        __PYX_XDEC_MEMVIEW((&__pyx_v_memslice), 1); 
        if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, __pyx_n_s_int64_t, long, 1, __Pyx_PyInt_From_long, 1, 0, 1) < 0)) __PYX_ERR(1, 177, __pyx_L1_error)
        goto __pyx_L10_break;
      }
      /*else*/ {
        PyErr_Clear(); 
      }
    }
    __pyx_t_2 = ((__pyx_v_itemsize == -1L) != 0);
    if (!__pyx_t_2) {
    } else {
      __pyx_t_3 = __pyx_t_2;
      goto __pyx_L46_bool_binop_done;
    }
    __pyx_t_2 = ((__pyx_v_itemsize == (sizeof(__pyx_t_5numpy_int32_t))) != 0);
    __pyx_t_3 = __pyx_t_2;
    __pyx_L46_bool_binop_done:;
    if (__pyx_t_3) {
      __pyx_t_8 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int32_t(__pyx_v_arg, 0); 
      __pyx_v_memslice = __pyx_t_8;
      __pyx_t_3 = (__pyx_v_memslice.memview != 0);
      if (__pyx_t_3) {
        __PYX_XDEC_MEMVIEW((&__pyx_v_memslice), 1); 
        if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, __pyx_n_s_int32_t, long, 1, __Pyx_PyInt_From_long, 1, 0, 1) < 0)) __PYX_ERR(1, 177, __pyx_L1_error)
        goto __pyx_L10_break;
      }
      /*else*/ {
        PyErr_Clear(); 
      }
    }
    __pyx_t_2 = ((__pyx_v_itemsize == -1L) != 0);
    if (!__pyx_t_2) {
    } else {
      __pyx_t_3 = __pyx_t_2;
      goto __pyx_L50_bool_binop_done;
    }
    __pyx_t_2 = ((__pyx_v_itemsize == (sizeof(__pyx_t_5numpy_uint64_t))) != 0);
    __pyx_t_3 = __pyx_t_2;
    __pyx_L50_bool_binop_done:;
    if (__pyx_t_3) {
      __pyx_t_8 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_uint64_t(__pyx_v_arg, 0); 
      __pyx_v_memslice = __pyx_t_8;
      __pyx_t_3 = (__pyx_v_memslice.memview != 0);
      if (__pyx_t_3) {
        __PYX_XDEC_MEMVIEW((&__pyx_v_memslice), 1); 
        if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, __pyx_n_s_uint64_t, long, 1, __Pyx_PyInt_From_long, 1, 0, 1) < 0)) __PYX_ERR(1, 177, __pyx_L1_error)
        goto __pyx_L10_break;
      }
      /*else*/ {
        PyErr_Clear(); 
      }
    }
    if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, Py_None, long, 1, __Pyx_PyInt_From_long, 1, 0, 1) < 0)) __PYX_ERR(1, 177, __pyx_L1_error)
    goto __pyx_L10_break;
  }
  __pyx_L10_break:;
  __pyx_t_6 = PyList_New(0); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 177, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_v_candidates = ((PyObject*)__pyx_t_6);
  __pyx_t_6 = 0;
  __pyx_t_5 = 0;
  if (unlikely(((PyObject *)__pyx_v_signatures) == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(1, 177, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_dict_iterator(((PyObject*)__pyx_v_signatures), 1, ((PyObject *)NULL), (&__pyx_t_9), (&__pyx_t_10)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 177, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __pyx_t_6 = __pyx_t_1;
  __pyx_t_1 = 0;
  while (1) {
    __pyx_t_11 = __Pyx_dict_iter_next(__pyx_t_6, __pyx_t_9, &__pyx_t_5, &__pyx_t_1, NULL, NULL, __pyx_t_10);
    if (unlikely(__pyx_t_11 == 0)) break;
    if (unlikely(__pyx_t_11 == -1)) __PYX_ERR(1, 177, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_sig, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_v_match_found = 0;
    __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_v_sig, __pyx_n_s_strip); if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 177, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __pyx_t_14 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_13))) {
      __pyx_t_14 = PyMethod_GET_SELF(__pyx_t_13);
      if (likely(__pyx_t_14)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
        __Pyx_INCREF(__pyx_t_14);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_13, function);
      }
    }
    __pyx_t_12 = (__pyx_t_14) ? __Pyx_PyObject_Call2Args(__pyx_t_13, __pyx_t_14, __pyx_kp_s__13) : __Pyx_PyObject_CallOneArg(__pyx_t_13, __pyx_kp_s__13);
    __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
    if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 177, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_t_12, __pyx_n_s_split); if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 177, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    __pyx_t_12 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_13))) {
      __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_13);
      if (likely(__pyx_t_12)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
        __Pyx_INCREF(__pyx_t_12);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_13, function);
      }
    }
    __pyx_t_1 = (__pyx_t_12) ? __Pyx_PyObject_Call2Args(__pyx_t_13, __pyx_t_12, __pyx_kp_s__14) : __Pyx_PyObject_CallOneArg(__pyx_t_13, __pyx_kp_s__14);
    __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 177, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    __Pyx_XDECREF_SET(__pyx_v_src_sig, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_15 = PyList_GET_SIZE(__pyx_v_dest_sig); if (unlikely(__pyx_t_15 == ((Py_ssize_t)-1))) __PYX_ERR(1, 177, __pyx_L1_error)
    __pyx_t_16 = __pyx_t_15;
    for (__pyx_t_17 = 0; __pyx_t_17 < __pyx_t_16; __pyx_t_17+=1) {
      __pyx_v_i = __pyx_t_17;
      __pyx_t_1 = __Pyx_GetItemInt_List(__pyx_v_dest_sig, __pyx_v_i, Py_ssize_t, 1, PyInt_FromSsize_t, 1, 1, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 177, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_XDECREF_SET(__pyx_v_dst_type, __pyx_t_1);
      __pyx_t_1 = 0;
      __pyx_t_3 = (__pyx_v_dst_type != Py_None);
      __pyx_t_2 = (__pyx_t_3 != 0);
      if (__pyx_t_2) {
        __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_src_sig, __pyx_v_i, Py_ssize_t, 1, PyInt_FromSsize_t, 0, 1, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 177, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_13 = PyObject_RichCompare(__pyx_t_1, __pyx_v_dst_type, Py_EQ); __Pyx_XGOTREF(__pyx_t_13); if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 177, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_13); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(1, 177, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
        if (__pyx_t_2) {
          __pyx_v_match_found = 1;
          goto __pyx_L58;
        }
        /*else*/ {
          __pyx_v_match_found = 0;
          goto __pyx_L56_break;
        }
        __pyx_L58:;
      }
    }
    __pyx_L56_break:;
    __pyx_t_2 = (__pyx_v_match_found != 0);
    if (__pyx_t_2) {
      __pyx_t_18 = __Pyx_PyList_Append(__pyx_v_candidates, __pyx_v_sig); if (unlikely(__pyx_t_18 == ((int)-1))) __PYX_ERR(1, 177, __pyx_L1_error)
    }
  }
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_2 = (PyList_GET_SIZE(__pyx_v_candidates) != 0);
  __pyx_t_3 = ((!__pyx_t_2) != 0);
  if (__pyx_t_3) {
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__15, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 177, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(1, 177, __pyx_L1_error)
  }
  __pyx_t_9 = PyList_GET_SIZE(__pyx_v_candidates); if (unlikely(__pyx_t_9 == ((Py_ssize_t)-1))) __PYX_ERR(1, 177, __pyx_L1_error)
  __pyx_t_3 = ((__pyx_t_9 > 1) != 0);
  if (__pyx_t_3) {
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__16, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 177, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(1, 177, __pyx_L1_error)
  }
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    if (unlikely(((PyObject *)__pyx_v_signatures) == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 177, __pyx_L1_error)
    }
    __pyx_t_6 = __Pyx_GetItemInt_List(__pyx_v_candidates, 0, long, 1, __Pyx_PyInt_From_long, 1, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 177, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_13 = __Pyx_PyDict_GetItem(((PyObject*)__pyx_v_signatures), __pyx_t_6); if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 177, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_r = __pyx_t_13;
    __pyx_t_13 = 0;
    goto __pyx_L0;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_AddTraceback("pandas._libs.interval.IntervalTree.__pyx_fused_cpdef", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_dest_sig);
  __Pyx_XDECREF(__pyx_v_ndarray);
  __Pyx_XDECREF(__pyx_v_arg);
  __Pyx_XDECREF(__pyx_v_dtype);
  __Pyx_XDECREF(__pyx_v_arg_base);
  __Pyx_XDECREF(__pyx_v_candidates);
  __Pyx_XDECREF(__pyx_v_sig);
  __Pyx_XDECREF(__pyx_v_src_sig);
  __Pyx_XDECREF(__pyx_v_dst_type);
  __Pyx_XDECREF(__pyx_v_kwargs);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_fuse_0__pyx_pw_6pandas_5_libs_8interval_12IntervalTree_41get_indexer_non_unique(PyObject *__pyx_v_self, PyObject *__pyx_arg_target); /*proto*/
static PyMethodDef __pyx_fuse_0__pyx_mdef_6pandas_5_libs_8interval_12IntervalTree_41get_indexer_non_unique = {"__pyx_fuse_0get_indexer_non_unique", (PyCFunction)__pyx_fuse_0__pyx_pw_6pandas_5_libs_8interval_12IntervalTree_41get_indexer_non_unique, METH_O, __pyx_doc_6pandas_5_libs_8interval_12IntervalTree_10get_indexer_non_unique};
static PyObject *__pyx_fuse_0__pyx_pw_6pandas_5_libs_8interval_12IntervalTree_41get_indexer_non_unique(PyObject *__pyx_v_self, PyObject *__pyx_arg_target) {
  __Pyx_memviewslice __pyx_v_target = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_indexer_non_unique (wrapper)", 0);
  assert(__pyx_arg_target); {
    __pyx_v_target = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_float64_t(__pyx_arg_target, PyBUF_WRITABLE); if (unlikely(!__pyx_v_target.memview)) __PYX_ERR(1, 177, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.IntervalTree.get_indexer_non_unique", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_12IntervalTree_40get_indexer_non_unique(((struct __pyx_obj_6pandas_5_libs_8interval_IntervalTree *)__pyx_v_self), __pyx_v_target);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_12IntervalTree_40get_indexer_non_unique(struct __pyx_obj_6pandas_5_libs_8interval_IntervalTree *__pyx_v_self, __Pyx_memviewslice __pyx_v_target) {
  size_t __pyx_v_old_len;
  Py_ssize_t __pyx_v_i;
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result = 0;
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_missing = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  size_t __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  Py_ssize_t __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  __Pyx_RefNannySetupContext("__pyx_fuse_0get_indexer_non_unique", 0);

  /* "pandas/_libs/intervaltree.pxi":186
 *             Int64Vector result, missing
 * 
 *         result = Int64Vector()             # <<<<<<<<<<<<<<
 *         missing = Int64Vector()
 *         old_len = 0
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 186, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_result = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":187
 * 
 *         result = Int64Vector()
 *         missing = Int64Vector()             # <<<<<<<<<<<<<<
 *         old_len = 0
 *         for i in range(len(target)):
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 187, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_missing = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":188
 *         result = Int64Vector()
 *         missing = Int64Vector()
 *         old_len = 0             # <<<<<<<<<<<<<<
 *         for i in range(len(target)):
 *             self.root.query(result, target[i])
 */
  __pyx_v_old_len = 0;

  /* "pandas/_libs/intervaltree.pxi":189
 *         missing = Int64Vector()
 *         old_len = 0
 *         for i in range(len(target)):             # <<<<<<<<<<<<<<
 *             self.root.query(result, target[i])
 *             if result.data.n == old_len:
 */
  __pyx_t_2 = __Pyx_MemoryView_Len(__pyx_v_target); 
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "pandas/_libs/intervaltree.pxi":190
 *         old_len = 0
 *         for i in range(len(target)):
 *             self.root.query(result, target[i])             # <<<<<<<<<<<<<<
 *             if result.data.n == old_len:
 *                 result.append(-1)
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->root, __pyx_n_s_query); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 190, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __pyx_v_i;
    __pyx_t_7 = -1;
    if (__pyx_t_6 < 0) {
      __pyx_t_6 += __pyx_v_target.shape[0];
      if (unlikely(__pyx_t_6 < 0)) __pyx_t_7 = 0;
    } else if (unlikely(__pyx_t_6 >= __pyx_v_target.shape[0])) __pyx_t_7 = 0;
    if (unlikely(__pyx_t_7 != -1)) {
      __Pyx_RaiseBufferIndexError(__pyx_t_7);
      __PYX_ERR(1, 190, __pyx_L1_error)
    }
    __pyx_t_8 = PyFloat_FromDouble((*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_target.data + __pyx_t_6 * __pyx_v_target.strides[0]) )))); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 190, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_9 = NULL;
    __pyx_t_7 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_7 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[3] = {__pyx_t_9, ((PyObject *)__pyx_v_result), __pyx_t_8};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 190, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[3] = {__pyx_t_9, ((PyObject *)__pyx_v_result), __pyx_t_8};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 190, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    } else
    #endif
    {
      __pyx_t_10 = PyTuple_New(2+__pyx_t_7); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 190, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      if (__pyx_t_9) {
        __Pyx_GIVEREF(__pyx_t_9); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_9); __pyx_t_9 = NULL;
      }
      __Pyx_INCREF(((PyObject *)__pyx_v_result));
      __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
      PyTuple_SET_ITEM(__pyx_t_10, 0+__pyx_t_7, ((PyObject *)__pyx_v_result));
      __Pyx_GIVEREF(__pyx_t_8);
      PyTuple_SET_ITEM(__pyx_t_10, 1+__pyx_t_7, __pyx_t_8);
      __pyx_t_8 = 0;
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_10, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 190, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "pandas/_libs/intervaltree.pxi":191
 *         for i in range(len(target)):
 *             self.root.query(result, target[i])
 *             if result.data.n == old_len:             # <<<<<<<<<<<<<<
 *                 result.append(-1)
 *                 missing.append(i)
 */
    __pyx_t_11 = ((__pyx_v_result->data->n == __pyx_v_old_len) != 0);
    if (__pyx_t_11) {

      /* "pandas/_libs/intervaltree.pxi":192
 *             self.root.query(result, target[i])
 *             if result.data.n == old_len:
 *                 result.append(-1)             # <<<<<<<<<<<<<<
 *                 missing.append(i)
 *             old_len = result.data.n
 */
      ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, -1LL);

      /* "pandas/_libs/intervaltree.pxi":193
 *             if result.data.n == old_len:
 *                 result.append(-1)
 *                 missing.append(i)             # <<<<<<<<<<<<<<
 *             old_len = result.data.n
 *         return (result.to_array().astype('intp'),
 */
      ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_missing->__pyx_vtab)->append(__pyx_v_missing, __pyx_v_i);

      /* "pandas/_libs/intervaltree.pxi":191
 *         for i in range(len(target)):
 *             self.root.query(result, target[i])
 *             if result.data.n == old_len:             # <<<<<<<<<<<<<<
 *                 result.append(-1)
 *                 missing.append(i)
 */
    }

    /* "pandas/_libs/intervaltree.pxi":194
 *                 result.append(-1)
 *                 missing.append(i)
 *             old_len = result.data.n             # <<<<<<<<<<<<<<
 *         return (result.to_array().astype('intp'),
 *                 missing.to_array().astype('intp'))
 */
    __pyx_t_12 = __pyx_v_result->data->n;
    __pyx_v_old_len = __pyx_t_12;
  }

  /* "pandas/_libs/intervaltree.pxi":195
 *                 missing.append(i)
 *             old_len = result.data.n
 *         return (result.to_array().astype('intp'),             # <<<<<<<<<<<<<<
 *                 missing.to_array().astype('intp'))
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_5 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->to_array(__pyx_v_result, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 195, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_astype); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 195, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_10))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_10);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_10, function);
    }
  }
  __pyx_t_1 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_10, __pyx_t_5, __pyx_n_u_intp) : __Pyx_PyObject_CallOneArg(__pyx_t_10, __pyx_n_u_intp);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 195, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

  /* "pandas/_libs/intervaltree.pxi":196
 *             old_len = result.data.n
 *         return (result.to_array().astype('intp'),
 *                 missing.to_array().astype('intp'))             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
  __pyx_t_5 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_missing->__pyx_vtab)->to_array(__pyx_v_missing, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 196, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_astype); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 196, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_8))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_8);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_8, function);
    }
  }
  __pyx_t_10 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_8, __pyx_t_5, __pyx_n_u_intp) : __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_n_u_intp);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 196, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "pandas/_libs/intervaltree.pxi":195
 *                 missing.append(i)
 *             old_len = result.data.n
 *         return (result.to_array().astype('intp'),             # <<<<<<<<<<<<<<
 *                 missing.to_array().astype('intp'))
 * 
 */
  __pyx_t_8 = PyTuple_New(2); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 195, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_10);
  PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_t_10);
  __pyx_t_1 = 0;
  __pyx_t_10 = 0;
  __pyx_r = __pyx_t_8;
  __pyx_t_8 = 0;
  goto __pyx_L0;

  /* "pandas/_libs/intervaltree.pxi":177
 *         return result.to_array().astype('intp')
 * 
 *     def get_indexer_non_unique(self, scalar_t[:] target):             # <<<<<<<<<<<<<<
 *         """Return the positions corresponding to intervals that overlap with
 *         the given array of scalar targets. Non-unique positions are repeated.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("pandas._libs.interval.IntervalTree.get_indexer_non_unique", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_target, 1);
  __Pyx_XDECREF((PyObject *)__pyx_v_result);
  __Pyx_XDECREF((PyObject *)__pyx_v_missing);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_fuse_1__pyx_pw_6pandas_5_libs_8interval_12IntervalTree_43get_indexer_non_unique(PyObject *__pyx_v_self, PyObject *__pyx_arg_target); /*proto*/
static PyMethodDef __pyx_fuse_1__pyx_mdef_6pandas_5_libs_8interval_12IntervalTree_43get_indexer_non_unique = {"__pyx_fuse_1get_indexer_non_unique", (PyCFunction)__pyx_fuse_1__pyx_pw_6pandas_5_libs_8interval_12IntervalTree_43get_indexer_non_unique, METH_O, __pyx_doc_6pandas_5_libs_8interval_12IntervalTree_10get_indexer_non_unique};
static PyObject *__pyx_fuse_1__pyx_pw_6pandas_5_libs_8interval_12IntervalTree_43get_indexer_non_unique(PyObject *__pyx_v_self, PyObject *__pyx_arg_target) {
  __Pyx_memviewslice __pyx_v_target = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_indexer_non_unique (wrapper)", 0);
  assert(__pyx_arg_target); {
    __pyx_v_target = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_float32_t(__pyx_arg_target, PyBUF_WRITABLE); if (unlikely(!__pyx_v_target.memview)) __PYX_ERR(1, 177, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.IntervalTree.get_indexer_non_unique", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_12IntervalTree_42get_indexer_non_unique(((struct __pyx_obj_6pandas_5_libs_8interval_IntervalTree *)__pyx_v_self), __pyx_v_target);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_12IntervalTree_42get_indexer_non_unique(struct __pyx_obj_6pandas_5_libs_8interval_IntervalTree *__pyx_v_self, __Pyx_memviewslice __pyx_v_target) {
  size_t __pyx_v_old_len;
  Py_ssize_t __pyx_v_i;
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result = 0;
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_missing = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  size_t __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  Py_ssize_t __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  __Pyx_RefNannySetupContext("__pyx_fuse_1get_indexer_non_unique", 0);

  /* "pandas/_libs/intervaltree.pxi":186
 *             Int64Vector result, missing
 * 
 *         result = Int64Vector()             # <<<<<<<<<<<<<<
 *         missing = Int64Vector()
 *         old_len = 0
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 186, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_result = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":187
 * 
 *         result = Int64Vector()
 *         missing = Int64Vector()             # <<<<<<<<<<<<<<
 *         old_len = 0
 *         for i in range(len(target)):
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 187, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_missing = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":188
 *         result = Int64Vector()
 *         missing = Int64Vector()
 *         old_len = 0             # <<<<<<<<<<<<<<
 *         for i in range(len(target)):
 *             self.root.query(result, target[i])
 */
  __pyx_v_old_len = 0;

  /* "pandas/_libs/intervaltree.pxi":189
 *         missing = Int64Vector()
 *         old_len = 0
 *         for i in range(len(target)):             # <<<<<<<<<<<<<<
 *             self.root.query(result, target[i])
 *             if result.data.n == old_len:
 */
  __pyx_t_2 = __Pyx_MemoryView_Len(__pyx_v_target); 
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "pandas/_libs/intervaltree.pxi":190
 *         old_len = 0
 *         for i in range(len(target)):
 *             self.root.query(result, target[i])             # <<<<<<<<<<<<<<
 *             if result.data.n == old_len:
 *                 result.append(-1)
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->root, __pyx_n_s_query); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 190, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __pyx_v_i;
    __pyx_t_7 = -1;
    if (__pyx_t_6 < 0) {
      __pyx_t_6 += __pyx_v_target.shape[0];
      if (unlikely(__pyx_t_6 < 0)) __pyx_t_7 = 0;
    } else if (unlikely(__pyx_t_6 >= __pyx_v_target.shape[0])) __pyx_t_7 = 0;
    if (unlikely(__pyx_t_7 != -1)) {
      __Pyx_RaiseBufferIndexError(__pyx_t_7);
      __PYX_ERR(1, 190, __pyx_L1_error)
    }
    __pyx_t_8 = PyFloat_FromDouble((*((__pyx_t_5numpy_float32_t *) ( /* dim=0 */ (__pyx_v_target.data + __pyx_t_6 * __pyx_v_target.strides[0]) )))); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 190, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_9 = NULL;
    __pyx_t_7 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_7 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[3] = {__pyx_t_9, ((PyObject *)__pyx_v_result), __pyx_t_8};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 190, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[3] = {__pyx_t_9, ((PyObject *)__pyx_v_result), __pyx_t_8};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 190, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    } else
    #endif
    {
      __pyx_t_10 = PyTuple_New(2+__pyx_t_7); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 190, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      if (__pyx_t_9) {
        __Pyx_GIVEREF(__pyx_t_9); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_9); __pyx_t_9 = NULL;
      }
      __Pyx_INCREF(((PyObject *)__pyx_v_result));
      __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
      PyTuple_SET_ITEM(__pyx_t_10, 0+__pyx_t_7, ((PyObject *)__pyx_v_result));
      __Pyx_GIVEREF(__pyx_t_8);
      PyTuple_SET_ITEM(__pyx_t_10, 1+__pyx_t_7, __pyx_t_8);
      __pyx_t_8 = 0;
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_10, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 190, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "pandas/_libs/intervaltree.pxi":191
 *         for i in range(len(target)):
 *             self.root.query(result, target[i])
 *             if result.data.n == old_len:             # <<<<<<<<<<<<<<
 *                 result.append(-1)
 *                 missing.append(i)
 */
    __pyx_t_11 = ((__pyx_v_result->data->n == __pyx_v_old_len) != 0);
    if (__pyx_t_11) {

      /* "pandas/_libs/intervaltree.pxi":192
 *             self.root.query(result, target[i])
 *             if result.data.n == old_len:
 *                 result.append(-1)             # <<<<<<<<<<<<<<
 *                 missing.append(i)
 *             old_len = result.data.n
 */
      ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, -1LL);

      /* "pandas/_libs/intervaltree.pxi":193
 *             if result.data.n == old_len:
 *                 result.append(-1)
 *                 missing.append(i)             # <<<<<<<<<<<<<<
 *             old_len = result.data.n
 *         return (result.to_array().astype('intp'),
 */
      ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_missing->__pyx_vtab)->append(__pyx_v_missing, __pyx_v_i);

      /* "pandas/_libs/intervaltree.pxi":191
 *         for i in range(len(target)):
 *             self.root.query(result, target[i])
 *             if result.data.n == old_len:             # <<<<<<<<<<<<<<
 *                 result.append(-1)
 *                 missing.append(i)
 */
    }

    /* "pandas/_libs/intervaltree.pxi":194
 *                 result.append(-1)
 *                 missing.append(i)
 *             old_len = result.data.n             # <<<<<<<<<<<<<<
 *         return (result.to_array().astype('intp'),
 *                 missing.to_array().astype('intp'))
 */
    __pyx_t_12 = __pyx_v_result->data->n;
    __pyx_v_old_len = __pyx_t_12;
  }

  /* "pandas/_libs/intervaltree.pxi":195
 *                 missing.append(i)
 *             old_len = result.data.n
 *         return (result.to_array().astype('intp'),             # <<<<<<<<<<<<<<
 *                 missing.to_array().astype('intp'))
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_5 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->to_array(__pyx_v_result, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 195, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_astype); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 195, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_10))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_10);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_10, function);
    }
  }
  __pyx_t_1 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_10, __pyx_t_5, __pyx_n_u_intp) : __Pyx_PyObject_CallOneArg(__pyx_t_10, __pyx_n_u_intp);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 195, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

  /* "pandas/_libs/intervaltree.pxi":196
 *             old_len = result.data.n
 *         return (result.to_array().astype('intp'),
 *                 missing.to_array().astype('intp'))             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
  __pyx_t_5 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_missing->__pyx_vtab)->to_array(__pyx_v_missing, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 196, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_astype); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 196, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_8))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_8);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_8, function);
    }
  }
  __pyx_t_10 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_8, __pyx_t_5, __pyx_n_u_intp) : __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_n_u_intp);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 196, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "pandas/_libs/intervaltree.pxi":195
 *                 missing.append(i)
 *             old_len = result.data.n
 *         return (result.to_array().astype('intp'),             # <<<<<<<<<<<<<<
 *                 missing.to_array().astype('intp'))
 * 
 */
  __pyx_t_8 = PyTuple_New(2); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 195, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_10);
  PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_t_10);
  __pyx_t_1 = 0;
  __pyx_t_10 = 0;
  __pyx_r = __pyx_t_8;
  __pyx_t_8 = 0;
  goto __pyx_L0;

  /* "pandas/_libs/intervaltree.pxi":177
 *         return result.to_array().astype('intp')
 * 
 *     def get_indexer_non_unique(self, scalar_t[:] target):             # <<<<<<<<<<<<<<
 *         """Return the positions corresponding to intervals that overlap with
 *         the given array of scalar targets. Non-unique positions are repeated.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("pandas._libs.interval.IntervalTree.get_indexer_non_unique", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_target, 1);
  __Pyx_XDECREF((PyObject *)__pyx_v_result);
  __Pyx_XDECREF((PyObject *)__pyx_v_missing);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_fuse_2__pyx_pw_6pandas_5_libs_8interval_12IntervalTree_45get_indexer_non_unique(PyObject *__pyx_v_self, PyObject *__pyx_arg_target); /*proto*/
static PyMethodDef __pyx_fuse_2__pyx_mdef_6pandas_5_libs_8interval_12IntervalTree_45get_indexer_non_unique = {"__pyx_fuse_2get_indexer_non_unique", (PyCFunction)__pyx_fuse_2__pyx_pw_6pandas_5_libs_8interval_12IntervalTree_45get_indexer_non_unique, METH_O, __pyx_doc_6pandas_5_libs_8interval_12IntervalTree_10get_indexer_non_unique};
static PyObject *__pyx_fuse_2__pyx_pw_6pandas_5_libs_8interval_12IntervalTree_45get_indexer_non_unique(PyObject *__pyx_v_self, PyObject *__pyx_arg_target) {
  __Pyx_memviewslice __pyx_v_target = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_indexer_non_unique (wrapper)", 0);
  assert(__pyx_arg_target); {
    __pyx_v_target = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_arg_target, PyBUF_WRITABLE); if (unlikely(!__pyx_v_target.memview)) __PYX_ERR(1, 177, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.IntervalTree.get_indexer_non_unique", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_12IntervalTree_44get_indexer_non_unique(((struct __pyx_obj_6pandas_5_libs_8interval_IntervalTree *)__pyx_v_self), __pyx_v_target);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_12IntervalTree_44get_indexer_non_unique(struct __pyx_obj_6pandas_5_libs_8interval_IntervalTree *__pyx_v_self, __Pyx_memviewslice __pyx_v_target) {
  size_t __pyx_v_old_len;
  Py_ssize_t __pyx_v_i;
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result = 0;
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_missing = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  size_t __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  Py_ssize_t __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  __Pyx_RefNannySetupContext("__pyx_fuse_2get_indexer_non_unique", 0);

  /* "pandas/_libs/intervaltree.pxi":186
 *             Int64Vector result, missing
 * 
 *         result = Int64Vector()             # <<<<<<<<<<<<<<
 *         missing = Int64Vector()
 *         old_len = 0
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 186, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_result = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":187
 * 
 *         result = Int64Vector()
 *         missing = Int64Vector()             # <<<<<<<<<<<<<<
 *         old_len = 0
 *         for i in range(len(target)):
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 187, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_missing = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":188
 *         result = Int64Vector()
 *         missing = Int64Vector()
 *         old_len = 0             # <<<<<<<<<<<<<<
 *         for i in range(len(target)):
 *             self.root.query(result, target[i])
 */
  __pyx_v_old_len = 0;

  /* "pandas/_libs/intervaltree.pxi":189
 *         missing = Int64Vector()
 *         old_len = 0
 *         for i in range(len(target)):             # <<<<<<<<<<<<<<
 *             self.root.query(result, target[i])
 *             if result.data.n == old_len:
 */
  __pyx_t_2 = __Pyx_MemoryView_Len(__pyx_v_target); 
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "pandas/_libs/intervaltree.pxi":190
 *         old_len = 0
 *         for i in range(len(target)):
 *             self.root.query(result, target[i])             # <<<<<<<<<<<<<<
 *             if result.data.n == old_len:
 *                 result.append(-1)
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->root, __pyx_n_s_query); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 190, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __pyx_v_i;
    __pyx_t_7 = -1;
    if (__pyx_t_6 < 0) {
      __pyx_t_6 += __pyx_v_target.shape[0];
      if (unlikely(__pyx_t_6 < 0)) __pyx_t_7 = 0;
    } else if (unlikely(__pyx_t_6 >= __pyx_v_target.shape[0])) __pyx_t_7 = 0;
    if (unlikely(__pyx_t_7 != -1)) {
      __Pyx_RaiseBufferIndexError(__pyx_t_7);
      __PYX_ERR(1, 190, __pyx_L1_error)
    }
    __pyx_t_8 = __Pyx_PyInt_From_npy_int64((*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_target.data + __pyx_t_6 * __pyx_v_target.strides[0]) )))); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 190, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_9 = NULL;
    __pyx_t_7 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_7 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[3] = {__pyx_t_9, ((PyObject *)__pyx_v_result), __pyx_t_8};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 190, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[3] = {__pyx_t_9, ((PyObject *)__pyx_v_result), __pyx_t_8};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 190, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    } else
    #endif
    {
      __pyx_t_10 = PyTuple_New(2+__pyx_t_7); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 190, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      if (__pyx_t_9) {
        __Pyx_GIVEREF(__pyx_t_9); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_9); __pyx_t_9 = NULL;
      }
      __Pyx_INCREF(((PyObject *)__pyx_v_result));
      __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
      PyTuple_SET_ITEM(__pyx_t_10, 0+__pyx_t_7, ((PyObject *)__pyx_v_result));
      __Pyx_GIVEREF(__pyx_t_8);
      PyTuple_SET_ITEM(__pyx_t_10, 1+__pyx_t_7, __pyx_t_8);
      __pyx_t_8 = 0;
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_10, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 190, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "pandas/_libs/intervaltree.pxi":191
 *         for i in range(len(target)):
 *             self.root.query(result, target[i])
 *             if result.data.n == old_len:             # <<<<<<<<<<<<<<
 *                 result.append(-1)
 *                 missing.append(i)
 */
    __pyx_t_11 = ((__pyx_v_result->data->n == __pyx_v_old_len) != 0);
    if (__pyx_t_11) {

      /* "pandas/_libs/intervaltree.pxi":192
 *             self.root.query(result, target[i])
 *             if result.data.n == old_len:
 *                 result.append(-1)             # <<<<<<<<<<<<<<
 *                 missing.append(i)
 *             old_len = result.data.n
 */
      ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, -1LL);

      /* "pandas/_libs/intervaltree.pxi":193
 *             if result.data.n == old_len:
 *                 result.append(-1)
 *                 missing.append(i)             # <<<<<<<<<<<<<<
 *             old_len = result.data.n
 *         return (result.to_array().astype('intp'),
 */
      ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_missing->__pyx_vtab)->append(__pyx_v_missing, __pyx_v_i);

      /* "pandas/_libs/intervaltree.pxi":191
 *         for i in range(len(target)):
 *             self.root.query(result, target[i])
 *             if result.data.n == old_len:             # <<<<<<<<<<<<<<
 *                 result.append(-1)
 *                 missing.append(i)
 */
    }

    /* "pandas/_libs/intervaltree.pxi":194
 *                 result.append(-1)
 *                 missing.append(i)
 *             old_len = result.data.n             # <<<<<<<<<<<<<<
 *         return (result.to_array().astype('intp'),
 *                 missing.to_array().astype('intp'))
 */
    __pyx_t_12 = __pyx_v_result->data->n;
    __pyx_v_old_len = __pyx_t_12;
  }

  /* "pandas/_libs/intervaltree.pxi":195
 *                 missing.append(i)
 *             old_len = result.data.n
 *         return (result.to_array().astype('intp'),             # <<<<<<<<<<<<<<
 *                 missing.to_array().astype('intp'))
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_5 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->to_array(__pyx_v_result, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 195, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_astype); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 195, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_10))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_10);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_10, function);
    }
  }
  __pyx_t_1 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_10, __pyx_t_5, __pyx_n_u_intp) : __Pyx_PyObject_CallOneArg(__pyx_t_10, __pyx_n_u_intp);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 195, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

  /* "pandas/_libs/intervaltree.pxi":196
 *             old_len = result.data.n
 *         return (result.to_array().astype('intp'),
 *                 missing.to_array().astype('intp'))             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
  __pyx_t_5 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_missing->__pyx_vtab)->to_array(__pyx_v_missing, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 196, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_astype); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 196, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_8))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_8);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_8, function);
    }
  }
  __pyx_t_10 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_8, __pyx_t_5, __pyx_n_u_intp) : __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_n_u_intp);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 196, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "pandas/_libs/intervaltree.pxi":195
 *                 missing.append(i)
 *             old_len = result.data.n
 *         return (result.to_array().astype('intp'),             # <<<<<<<<<<<<<<
 *                 missing.to_array().astype('intp'))
 * 
 */
  __pyx_t_8 = PyTuple_New(2); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 195, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_10);
  PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_t_10);
  __pyx_t_1 = 0;
  __pyx_t_10 = 0;
  __pyx_r = __pyx_t_8;
  __pyx_t_8 = 0;
  goto __pyx_L0;

  /* "pandas/_libs/intervaltree.pxi":177
 *         return result.to_array().astype('intp')
 * 
 *     def get_indexer_non_unique(self, scalar_t[:] target):             # <<<<<<<<<<<<<<
 *         """Return the positions corresponding to intervals that overlap with
 *         the given array of scalar targets. Non-unique positions are repeated.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("pandas._libs.interval.IntervalTree.get_indexer_non_unique", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_target, 1);
  __Pyx_XDECREF((PyObject *)__pyx_v_result);
  __Pyx_XDECREF((PyObject *)__pyx_v_missing);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_fuse_3__pyx_pw_6pandas_5_libs_8interval_12IntervalTree_47get_indexer_non_unique(PyObject *__pyx_v_self, PyObject *__pyx_arg_target); /*proto*/
static PyMethodDef __pyx_fuse_3__pyx_mdef_6pandas_5_libs_8interval_12IntervalTree_47get_indexer_non_unique = {"__pyx_fuse_3get_indexer_non_unique", (PyCFunction)__pyx_fuse_3__pyx_pw_6pandas_5_libs_8interval_12IntervalTree_47get_indexer_non_unique, METH_O, __pyx_doc_6pandas_5_libs_8interval_12IntervalTree_10get_indexer_non_unique};
static PyObject *__pyx_fuse_3__pyx_pw_6pandas_5_libs_8interval_12IntervalTree_47get_indexer_non_unique(PyObject *__pyx_v_self, PyObject *__pyx_arg_target) {
  __Pyx_memviewslice __pyx_v_target = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_indexer_non_unique (wrapper)", 0);
  assert(__pyx_arg_target); {
    __pyx_v_target = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int32_t(__pyx_arg_target, PyBUF_WRITABLE); if (unlikely(!__pyx_v_target.memview)) __PYX_ERR(1, 177, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.IntervalTree.get_indexer_non_unique", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_12IntervalTree_46get_indexer_non_unique(((struct __pyx_obj_6pandas_5_libs_8interval_IntervalTree *)__pyx_v_self), __pyx_v_target);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_12IntervalTree_46get_indexer_non_unique(struct __pyx_obj_6pandas_5_libs_8interval_IntervalTree *__pyx_v_self, __Pyx_memviewslice __pyx_v_target) {
  size_t __pyx_v_old_len;
  Py_ssize_t __pyx_v_i;
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result = 0;
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_missing = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  size_t __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  Py_ssize_t __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  __Pyx_RefNannySetupContext("__pyx_fuse_3get_indexer_non_unique", 0);

  /* "pandas/_libs/intervaltree.pxi":186
 *             Int64Vector result, missing
 * 
 *         result = Int64Vector()             # <<<<<<<<<<<<<<
 *         missing = Int64Vector()
 *         old_len = 0
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 186, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_result = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":187
 * 
 *         result = Int64Vector()
 *         missing = Int64Vector()             # <<<<<<<<<<<<<<
 *         old_len = 0
 *         for i in range(len(target)):
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 187, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_missing = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":188
 *         result = Int64Vector()
 *         missing = Int64Vector()
 *         old_len = 0             # <<<<<<<<<<<<<<
 *         for i in range(len(target)):
 *             self.root.query(result, target[i])
 */
  __pyx_v_old_len = 0;

  /* "pandas/_libs/intervaltree.pxi":189
 *         missing = Int64Vector()
 *         old_len = 0
 *         for i in range(len(target)):             # <<<<<<<<<<<<<<
 *             self.root.query(result, target[i])
 *             if result.data.n == old_len:
 */
  __pyx_t_2 = __Pyx_MemoryView_Len(__pyx_v_target); 
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "pandas/_libs/intervaltree.pxi":190
 *         old_len = 0
 *         for i in range(len(target)):
 *             self.root.query(result, target[i])             # <<<<<<<<<<<<<<
 *             if result.data.n == old_len:
 *                 result.append(-1)
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->root, __pyx_n_s_query); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 190, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __pyx_v_i;
    __pyx_t_7 = -1;
    if (__pyx_t_6 < 0) {
      __pyx_t_6 += __pyx_v_target.shape[0];
      if (unlikely(__pyx_t_6 < 0)) __pyx_t_7 = 0;
    } else if (unlikely(__pyx_t_6 >= __pyx_v_target.shape[0])) __pyx_t_7 = 0;
    if (unlikely(__pyx_t_7 != -1)) {
      __Pyx_RaiseBufferIndexError(__pyx_t_7);
      __PYX_ERR(1, 190, __pyx_L1_error)
    }
    __pyx_t_8 = __Pyx_PyInt_From_npy_int32((*((__pyx_t_5numpy_int32_t *) ( /* dim=0 */ (__pyx_v_target.data + __pyx_t_6 * __pyx_v_target.strides[0]) )))); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 190, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_9 = NULL;
    __pyx_t_7 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_7 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[3] = {__pyx_t_9, ((PyObject *)__pyx_v_result), __pyx_t_8};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 190, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[3] = {__pyx_t_9, ((PyObject *)__pyx_v_result), __pyx_t_8};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 190, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    } else
    #endif
    {
      __pyx_t_10 = PyTuple_New(2+__pyx_t_7); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 190, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      if (__pyx_t_9) {
        __Pyx_GIVEREF(__pyx_t_9); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_9); __pyx_t_9 = NULL;
      }
      __Pyx_INCREF(((PyObject *)__pyx_v_result));
      __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
      PyTuple_SET_ITEM(__pyx_t_10, 0+__pyx_t_7, ((PyObject *)__pyx_v_result));
      __Pyx_GIVEREF(__pyx_t_8);
      PyTuple_SET_ITEM(__pyx_t_10, 1+__pyx_t_7, __pyx_t_8);
      __pyx_t_8 = 0;
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_10, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 190, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "pandas/_libs/intervaltree.pxi":191
 *         for i in range(len(target)):
 *             self.root.query(result, target[i])
 *             if result.data.n == old_len:             # <<<<<<<<<<<<<<
 *                 result.append(-1)
 *                 missing.append(i)
 */
    __pyx_t_11 = ((__pyx_v_result->data->n == __pyx_v_old_len) != 0);
    if (__pyx_t_11) {

      /* "pandas/_libs/intervaltree.pxi":192
 *             self.root.query(result, target[i])
 *             if result.data.n == old_len:
 *                 result.append(-1)             # <<<<<<<<<<<<<<
 *                 missing.append(i)
 *             old_len = result.data.n
 */
      ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, -1LL);

      /* "pandas/_libs/intervaltree.pxi":193
 *             if result.data.n == old_len:
 *                 result.append(-1)
 *                 missing.append(i)             # <<<<<<<<<<<<<<
 *             old_len = result.data.n
 *         return (result.to_array().astype('intp'),
 */
      ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_missing->__pyx_vtab)->append(__pyx_v_missing, __pyx_v_i);

      /* "pandas/_libs/intervaltree.pxi":191
 *         for i in range(len(target)):
 *             self.root.query(result, target[i])
 *             if result.data.n == old_len:             # <<<<<<<<<<<<<<
 *                 result.append(-1)
 *                 missing.append(i)
 */
    }

    /* "pandas/_libs/intervaltree.pxi":194
 *                 result.append(-1)
 *                 missing.append(i)
 *             old_len = result.data.n             # <<<<<<<<<<<<<<
 *         return (result.to_array().astype('intp'),
 *                 missing.to_array().astype('intp'))
 */
    __pyx_t_12 = __pyx_v_result->data->n;
    __pyx_v_old_len = __pyx_t_12;
  }

  /* "pandas/_libs/intervaltree.pxi":195
 *                 missing.append(i)
 *             old_len = result.data.n
 *         return (result.to_array().astype('intp'),             # <<<<<<<<<<<<<<
 *                 missing.to_array().astype('intp'))
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_5 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->to_array(__pyx_v_result, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 195, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_astype); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 195, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_10))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_10);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_10, function);
    }
  }
  __pyx_t_1 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_10, __pyx_t_5, __pyx_n_u_intp) : __Pyx_PyObject_CallOneArg(__pyx_t_10, __pyx_n_u_intp);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 195, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

  /* "pandas/_libs/intervaltree.pxi":196
 *             old_len = result.data.n
 *         return (result.to_array().astype('intp'),
 *                 missing.to_array().astype('intp'))             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
  __pyx_t_5 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_missing->__pyx_vtab)->to_array(__pyx_v_missing, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 196, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_astype); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 196, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_8))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_8);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_8, function);
    }
  }
  __pyx_t_10 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_8, __pyx_t_5, __pyx_n_u_intp) : __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_n_u_intp);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 196, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "pandas/_libs/intervaltree.pxi":195
 *                 missing.append(i)
 *             old_len = result.data.n
 *         return (result.to_array().astype('intp'),             # <<<<<<<<<<<<<<
 *                 missing.to_array().astype('intp'))
 * 
 */
  __pyx_t_8 = PyTuple_New(2); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 195, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_10);
  PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_t_10);
  __pyx_t_1 = 0;
  __pyx_t_10 = 0;
  __pyx_r = __pyx_t_8;
  __pyx_t_8 = 0;
  goto __pyx_L0;

  /* "pandas/_libs/intervaltree.pxi":177
 *         return result.to_array().astype('intp')
 * 
 *     def get_indexer_non_unique(self, scalar_t[:] target):             # <<<<<<<<<<<<<<
 *         """Return the positions corresponding to intervals that overlap with
 *         the given array of scalar targets. Non-unique positions are repeated.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("pandas._libs.interval.IntervalTree.get_indexer_non_unique", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_target, 1);
  __Pyx_XDECREF((PyObject *)__pyx_v_result);
  __Pyx_XDECREF((PyObject *)__pyx_v_missing);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_fuse_4__pyx_pw_6pandas_5_libs_8interval_12IntervalTree_49get_indexer_non_unique(PyObject *__pyx_v_self, PyObject *__pyx_arg_target); /*proto*/
static PyMethodDef __pyx_fuse_4__pyx_mdef_6pandas_5_libs_8interval_12IntervalTree_49get_indexer_non_unique = {"__pyx_fuse_4get_indexer_non_unique", (PyCFunction)__pyx_fuse_4__pyx_pw_6pandas_5_libs_8interval_12IntervalTree_49get_indexer_non_unique, METH_O, __pyx_doc_6pandas_5_libs_8interval_12IntervalTree_10get_indexer_non_unique};
static PyObject *__pyx_fuse_4__pyx_pw_6pandas_5_libs_8interval_12IntervalTree_49get_indexer_non_unique(PyObject *__pyx_v_self, PyObject *__pyx_arg_target) {
  __Pyx_memviewslice __pyx_v_target = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_indexer_non_unique (wrapper)", 0);
  assert(__pyx_arg_target); {
    __pyx_v_target = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_uint64_t(__pyx_arg_target, PyBUF_WRITABLE); if (unlikely(!__pyx_v_target.memview)) __PYX_ERR(1, 177, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.IntervalTree.get_indexer_non_unique", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_12IntervalTree_48get_indexer_non_unique(((struct __pyx_obj_6pandas_5_libs_8interval_IntervalTree *)__pyx_v_self), __pyx_v_target);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_12IntervalTree_48get_indexer_non_unique(struct __pyx_obj_6pandas_5_libs_8interval_IntervalTree *__pyx_v_self, __Pyx_memviewslice __pyx_v_target) {
  size_t __pyx_v_old_len;
  Py_ssize_t __pyx_v_i;
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result = 0;
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_missing = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  size_t __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  Py_ssize_t __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  __Pyx_RefNannySetupContext("__pyx_fuse_4get_indexer_non_unique", 0);

  /* "pandas/_libs/intervaltree.pxi":186
 *             Int64Vector result, missing
 * 
 *         result = Int64Vector()             # <<<<<<<<<<<<<<
 *         missing = Int64Vector()
 *         old_len = 0
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 186, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_result = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":187
 * 
 *         result = Int64Vector()
 *         missing = Int64Vector()             # <<<<<<<<<<<<<<
 *         old_len = 0
 *         for i in range(len(target)):
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 187, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_missing = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":188
 *         result = Int64Vector()
 *         missing = Int64Vector()
 *         old_len = 0             # <<<<<<<<<<<<<<
 *         for i in range(len(target)):
 *             self.root.query(result, target[i])
 */
  __pyx_v_old_len = 0;

  /* "pandas/_libs/intervaltree.pxi":189
 *         missing = Int64Vector()
 *         old_len = 0
 *         for i in range(len(target)):             # <<<<<<<<<<<<<<
 *             self.root.query(result, target[i])
 *             if result.data.n == old_len:
 */
  __pyx_t_2 = __Pyx_MemoryView_Len(__pyx_v_target); 
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "pandas/_libs/intervaltree.pxi":190
 *         old_len = 0
 *         for i in range(len(target)):
 *             self.root.query(result, target[i])             # <<<<<<<<<<<<<<
 *             if result.data.n == old_len:
 *                 result.append(-1)
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->root, __pyx_n_s_query); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 190, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __pyx_v_i;
    __pyx_t_7 = -1;
    if (__pyx_t_6 < 0) {
      __pyx_t_6 += __pyx_v_target.shape[0];
      if (unlikely(__pyx_t_6 < 0)) __pyx_t_7 = 0;
    } else if (unlikely(__pyx_t_6 >= __pyx_v_target.shape[0])) __pyx_t_7 = 0;
    if (unlikely(__pyx_t_7 != -1)) {
      __Pyx_RaiseBufferIndexError(__pyx_t_7);
      __PYX_ERR(1, 190, __pyx_L1_error)
    }
    __pyx_t_8 = __Pyx_PyInt_From_npy_uint64((*((__pyx_t_5numpy_uint64_t *) ( /* dim=0 */ (__pyx_v_target.data + __pyx_t_6 * __pyx_v_target.strides[0]) )))); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 190, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_9 = NULL;
    __pyx_t_7 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_7 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[3] = {__pyx_t_9, ((PyObject *)__pyx_v_result), __pyx_t_8};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 190, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[3] = {__pyx_t_9, ((PyObject *)__pyx_v_result), __pyx_t_8};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 190, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    } else
    #endif
    {
      __pyx_t_10 = PyTuple_New(2+__pyx_t_7); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 190, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      if (__pyx_t_9) {
        __Pyx_GIVEREF(__pyx_t_9); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_9); __pyx_t_9 = NULL;
      }
      __Pyx_INCREF(((PyObject *)__pyx_v_result));
      __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
      PyTuple_SET_ITEM(__pyx_t_10, 0+__pyx_t_7, ((PyObject *)__pyx_v_result));
      __Pyx_GIVEREF(__pyx_t_8);
      PyTuple_SET_ITEM(__pyx_t_10, 1+__pyx_t_7, __pyx_t_8);
      __pyx_t_8 = 0;
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_10, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 190, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "pandas/_libs/intervaltree.pxi":191
 *         for i in range(len(target)):
 *             self.root.query(result, target[i])
 *             if result.data.n == old_len:             # <<<<<<<<<<<<<<
 *                 result.append(-1)
 *                 missing.append(i)
 */
    __pyx_t_11 = ((__pyx_v_result->data->n == __pyx_v_old_len) != 0);
    if (__pyx_t_11) {

      /* "pandas/_libs/intervaltree.pxi":192
 *             self.root.query(result, target[i])
 *             if result.data.n == old_len:
 *                 result.append(-1)             # <<<<<<<<<<<<<<
 *                 missing.append(i)
 *             old_len = result.data.n
 */
      ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, -1LL);

      /* "pandas/_libs/intervaltree.pxi":193
 *             if result.data.n == old_len:
 *                 result.append(-1)
 *                 missing.append(i)             # <<<<<<<<<<<<<<
 *             old_len = result.data.n
 *         return (result.to_array().astype('intp'),
 */
      ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_missing->__pyx_vtab)->append(__pyx_v_missing, __pyx_v_i);

      /* "pandas/_libs/intervaltree.pxi":191
 *         for i in range(len(target)):
 *             self.root.query(result, target[i])
 *             if result.data.n == old_len:             # <<<<<<<<<<<<<<
 *                 result.append(-1)
 *                 missing.append(i)
 */
    }

    /* "pandas/_libs/intervaltree.pxi":194
 *                 result.append(-1)
 *                 missing.append(i)
 *             old_len = result.data.n             # <<<<<<<<<<<<<<
 *         return (result.to_array().astype('intp'),
 *                 missing.to_array().astype('intp'))
 */
    __pyx_t_12 = __pyx_v_result->data->n;
    __pyx_v_old_len = __pyx_t_12;
  }

  /* "pandas/_libs/intervaltree.pxi":195
 *                 missing.append(i)
 *             old_len = result.data.n
 *         return (result.to_array().astype('intp'),             # <<<<<<<<<<<<<<
 *                 missing.to_array().astype('intp'))
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_5 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->to_array(__pyx_v_result, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 195, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_astype); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 195, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_10))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_10);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_10, function);
    }
  }
  __pyx_t_1 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_10, __pyx_t_5, __pyx_n_u_intp) : __Pyx_PyObject_CallOneArg(__pyx_t_10, __pyx_n_u_intp);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 195, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

  /* "pandas/_libs/intervaltree.pxi":196
 *             old_len = result.data.n
 *         return (result.to_array().astype('intp'),
 *                 missing.to_array().astype('intp'))             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
  __pyx_t_5 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_missing->__pyx_vtab)->to_array(__pyx_v_missing, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 196, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_astype); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 196, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_8))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_8);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_8, function);
    }
  }
  __pyx_t_10 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_8, __pyx_t_5, __pyx_n_u_intp) : __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_n_u_intp);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 196, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "pandas/_libs/intervaltree.pxi":195
 *                 missing.append(i)
 *             old_len = result.data.n
 *         return (result.to_array().astype('intp'),             # <<<<<<<<<<<<<<
 *                 missing.to_array().astype('intp'))
 * 
 */
  __pyx_t_8 = PyTuple_New(2); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 195, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_10);
  PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_t_10);
  __pyx_t_1 = 0;
  __pyx_t_10 = 0;
  __pyx_r = __pyx_t_8;
  __pyx_t_8 = 0;
  goto __pyx_L0;

  /* "pandas/_libs/intervaltree.pxi":177
 *         return result.to_array().astype('intp')
 * 
 *     def get_indexer_non_unique(self, scalar_t[:] target):             # <<<<<<<<<<<<<<
 *         """Return the positions corresponding to intervals that overlap with
 *         the given array of scalar targets. Non-unique positions are repeated.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("pandas._libs.interval.IntervalTree.get_indexer_non_unique", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_target, 1);
  __Pyx_XDECREF((PyObject *)__pyx_v_result);
  __Pyx_XDECREF((PyObject *)__pyx_v_missing);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":198
 *                 missing.to_array().astype('intp'))
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return ('<IntervalTree[{dtype},{closed}]: '
 *                 '{n_elements} elements>'.format(
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_12IntervalTree_13__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_12IntervalTree_13__repr__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_12IntervalTree_12__repr__(((struct __pyx_obj_6pandas_5_libs_8interval_IntervalTree *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_12IntervalTree_12__repr__(struct __pyx_obj_6pandas_5_libs_8interval_IntervalTree *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "pandas/_libs/intervaltree.pxi":199
 * 
 *     def __repr__(self):
 *         return ('<IntervalTree[{dtype},{closed}]: '             # <<<<<<<<<<<<<<
 *                 '{n_elements} elements>'.format(
 *                     dtype=self.dtype, closed=self.closed,
 */
  __Pyx_XDECREF(__pyx_r);

  /* "pandas/_libs/intervaltree.pxi":200
 *     def __repr__(self):
 *         return ('<IntervalTree[{dtype},{closed}]: '
 *                 '{n_elements} elements>'.format(             # <<<<<<<<<<<<<<
 *                     dtype=self.dtype, closed=self.closed,
 *                     n_elements=self.root.n_elements))
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_IntervalTree_dtype_closed_n_ele, __pyx_n_s_format); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 200, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "pandas/_libs/intervaltree.pxi":201
 *         return ('<IntervalTree[{dtype},{closed}]: '
 *                 '{n_elements} elements>'.format(
 *                     dtype=self.dtype, closed=self.closed,             # <<<<<<<<<<<<<<
 *                     n_elements=self.root.n_elements))
 * 
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 201, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_dtype, __pyx_v_self->dtype) < 0) __PYX_ERR(1, 201, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_closed, __pyx_v_self->closed) < 0) __PYX_ERR(1, 201, __pyx_L1_error)

  /* "pandas/_libs/intervaltree.pxi":202
 *                 '{n_elements} elements>'.format(
 *                     dtype=self.dtype, closed=self.closed,
 *                     n_elements=self.root.n_elements))             # <<<<<<<<<<<<<<
 * 
 *     # compat with IndexEngine interface
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->root, __pyx_n_s_n_elements); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 202, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_n_elements, __pyx_t_3) < 0) __PYX_ERR(1, 201, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "pandas/_libs/intervaltree.pxi":200
 *     def __repr__(self):
 *         return ('<IntervalTree[{dtype},{closed}]: '
 *                 '{n_elements} elements>'.format(             # <<<<<<<<<<<<<<
 *                     dtype=self.dtype, closed=self.closed,
 *                     n_elements=self.root.n_elements))
 */
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_empty_tuple, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 200, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "pandas/_libs/intervaltree.pxi":198
 *                 missing.to_array().astype('intp'))
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return ('<IntervalTree[{dtype},{closed}]: '
 *                 '{n_elements} elements>'.format(
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pandas._libs.interval.IntervalTree.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":205
 * 
 *     # compat with IndexEngine interface
 *     def clear_mapping(self):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_12IntervalTree_15clear_mapping(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_12IntervalTree_15clear_mapping(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("clear_mapping (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_12IntervalTree_14clear_mapping(((struct __pyx_obj_6pandas_5_libs_8interval_IntervalTree *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_12IntervalTree_14clear_mapping(CYTHON_UNUSED struct __pyx_obj_6pandas_5_libs_8interval_IntervalTree *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("clear_mapping", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":29
 *     """
 *     cdef:
 *         readonly object left, right, root, dtype             # <<<<<<<<<<<<<<
 *         readonly str closed
 *         object _is_overlapping, _left_sorter, _right_sorter
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_12IntervalTree_4left_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_12IntervalTree_4left_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_12IntervalTree_4left___get__(((struct __pyx_obj_6pandas_5_libs_8interval_IntervalTree *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_12IntervalTree_4left___get__(struct __pyx_obj_6pandas_5_libs_8interval_IntervalTree *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->left);
  __pyx_r = __pyx_v_self->left;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_12IntervalTree_5right_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_12IntervalTree_5right_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_12IntervalTree_5right___get__(((struct __pyx_obj_6pandas_5_libs_8interval_IntervalTree *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_12IntervalTree_5right___get__(struct __pyx_obj_6pandas_5_libs_8interval_IntervalTree *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->right);
  __pyx_r = __pyx_v_self->right;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_12IntervalTree_4root_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_12IntervalTree_4root_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_12IntervalTree_4root___get__(((struct __pyx_obj_6pandas_5_libs_8interval_IntervalTree *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_12IntervalTree_4root___get__(struct __pyx_obj_6pandas_5_libs_8interval_IntervalTree *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->root);
  __pyx_r = __pyx_v_self->root;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_12IntervalTree_5dtype_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_12IntervalTree_5dtype_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_12IntervalTree_5dtype___get__(((struct __pyx_obj_6pandas_5_libs_8interval_IntervalTree *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_12IntervalTree_5dtype___get__(struct __pyx_obj_6pandas_5_libs_8interval_IntervalTree *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->dtype);
  __pyx_r = __pyx_v_self->dtype;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":30
 *     cdef:
 *         readonly object left, right, root, dtype
 *         readonly str closed             # <<<<<<<<<<<<<<
 *         object _is_overlapping, _left_sorter, _right_sorter
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_12IntervalTree_6closed_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_12IntervalTree_6closed_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_12IntervalTree_6closed___get__(((struct __pyx_obj_6pandas_5_libs_8interval_IntervalTree *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_12IntervalTree_6closed___get__(struct __pyx_obj_6pandas_5_libs_8interval_IntervalTree *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->closed);
  __pyx_r = __pyx_v_self->closed;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_12IntervalTree_53__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_12IntervalTree_53__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_12IntervalTree_52__reduce_cython__(((struct __pyx_obj_6pandas_5_libs_8interval_IntervalTree *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_12IntervalTree_52__reduce_cython__(struct __pyx_obj_6pandas_5_libs_8interval_IntervalTree *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self._is_overlapping, self._left_sorter, self._right_sorter, self.closed, self.dtype, self.left, self.right, self.root)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = PyTuple_New(8); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_self->_is_overlapping);
  __Pyx_GIVEREF(__pyx_v_self->_is_overlapping);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_self->_is_overlapping);
  __Pyx_INCREF(__pyx_v_self->_left_sorter);
  __Pyx_GIVEREF(__pyx_v_self->_left_sorter);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_self->_left_sorter);
  __Pyx_INCREF(__pyx_v_self->_right_sorter);
  __Pyx_GIVEREF(__pyx_v_self->_right_sorter);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_self->_right_sorter);
  __Pyx_INCREF(__pyx_v_self->closed);
  __Pyx_GIVEREF(__pyx_v_self->closed);
  PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_v_self->closed);
  __Pyx_INCREF(__pyx_v_self->dtype);
  __Pyx_GIVEREF(__pyx_v_self->dtype);
  PyTuple_SET_ITEM(__pyx_t_1, 4, __pyx_v_self->dtype);
  __Pyx_INCREF(__pyx_v_self->left);
  __Pyx_GIVEREF(__pyx_v_self->left);
  PyTuple_SET_ITEM(__pyx_t_1, 5, __pyx_v_self->left);
  __Pyx_INCREF(__pyx_v_self->right);
  __Pyx_GIVEREF(__pyx_v_self->right);
  PyTuple_SET_ITEM(__pyx_t_1, 6, __pyx_v_self->right);
  __Pyx_INCREF(__pyx_v_self->root);
  __Pyx_GIVEREF(__pyx_v_self->root);
  PyTuple_SET_ITEM(__pyx_t_1, 7, __pyx_v_self->root);
  __pyx_v_state = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self._is_overlapping, self._left_sorter, self._right_sorter, self.closed, self.dtype, self.left, self.right, self.root)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = (self._is_overlapping, self._left_sorter, self._right_sorter, self.closed, self.dtype, self.left, self.right, self.root)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict);
    __pyx_t_4 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_4));
    __pyx_t_4 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self._is_overlapping is not None or self._left_sorter is not None or self._right_sorter is not None or self.closed is not None or self.dtype is not None or self.left is not None or self.right is not None or self.root is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self._is_overlapping, self._left_sorter, self._right_sorter, self.closed, self.dtype, self.left, self.right, self.root)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self._is_overlapping is not None or self._left_sorter is not None or self._right_sorter is not None or self.closed is not None or self.dtype is not None or self.left is not None or self.right is not None or self.root is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_IntervalTree, (type(self), 0x05502fb, None), state
 */
  /*else*/ {
    __pyx_t_2 = (__pyx_v_self->_is_overlapping != Py_None);
    __pyx_t_5 = (__pyx_t_2 != 0);
    if (!__pyx_t_5) {
    } else {
      __pyx_t_3 = __pyx_t_5;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_5 = (__pyx_v_self->_left_sorter != Py_None);
    __pyx_t_2 = (__pyx_t_5 != 0);
    if (!__pyx_t_2) {
    } else {
      __pyx_t_3 = __pyx_t_2;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_2 = (__pyx_v_self->_right_sorter != Py_None);
    __pyx_t_5 = (__pyx_t_2 != 0);
    if (!__pyx_t_5) {
    } else {
      __pyx_t_3 = __pyx_t_5;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_5 = (__pyx_v_self->closed != ((PyObject*)Py_None));
    __pyx_t_2 = (__pyx_t_5 != 0);
    if (!__pyx_t_2) {
    } else {
      __pyx_t_3 = __pyx_t_2;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_2 = (__pyx_v_self->dtype != Py_None);
    __pyx_t_5 = (__pyx_t_2 != 0);
    if (!__pyx_t_5) {
    } else {
      __pyx_t_3 = __pyx_t_5;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_5 = (__pyx_v_self->left != Py_None);
    __pyx_t_2 = (__pyx_t_5 != 0);
    if (!__pyx_t_2) {
    } else {
      __pyx_t_3 = __pyx_t_2;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_2 = (__pyx_v_self->right != Py_None);
    __pyx_t_5 = (__pyx_t_2 != 0);
    if (!__pyx_t_5) {
    } else {
      __pyx_t_3 = __pyx_t_5;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_5 = (__pyx_v_self->root != Py_None);
    __pyx_t_2 = (__pyx_t_5 != 0);
    __pyx_t_3 = __pyx_t_2;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_3;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self._is_overlapping is not None or self._left_sorter is not None or self._right_sorter is not None or self.closed is not None or self.dtype is not None or self.left is not None or self.right is not None or self.root is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_IntervalTree, (type(self), 0x05502fb, None), state
 *     else:
 */
  __pyx_t_3 = (__pyx_v_use_setstate != 0);
  if (__pyx_t_3) {

    /* "(tree fragment)":13
 *         use_setstate = self._is_overlapping is not None or self._left_sorter is not None or self._right_sorter is not None or self.closed is not None or self.dtype is not None or self.left is not None or self.right is not None or self.root is not None
 *     if use_setstate:
 *         return __pyx_unpickle_IntervalTree, (type(self), 0x05502fb, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_IntervalTree, (type(self), 0x05502fb, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_pyx_unpickle_IntervalTree); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_5571323);
    __Pyx_GIVEREF(__pyx_int_5571323);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_5571323);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None);
    __pyx_t_6 = PyTuple_New(3); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_1);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_v_state);
    __pyx_t_4 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_6;
    __pyx_t_6 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self._is_overlapping is not None or self._left_sorter is not None or self._right_sorter is not None or self.closed is not None or self.dtype is not None or self.left is not None or self.right is not None or self.root is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_IntervalTree, (type(self), 0x05502fb, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_IntervalTree, (type(self), 0x05502fb, None), state
 *     else:
 *         return __pyx_unpickle_IntervalTree, (type(self), 0x05502fb, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_IntervalTree__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_pyx_unpickle_IntervalTree); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_5571323);
    __Pyx_GIVEREF(__pyx_int_5571323);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_5571323);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state);
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1);
    __pyx_t_6 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pandas._libs.interval.IntervalTree.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_IntervalTree, (type(self), 0x05502fb, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_IntervalTree__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_12IntervalTree_55__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_12IntervalTree_55__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_12IntervalTree_54__setstate_cython__(((struct __pyx_obj_6pandas_5_libs_8interval_IntervalTree *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_12IntervalTree_54__setstate_cython__(struct __pyx_obj_6pandas_5_libs_8interval_IntervalTree *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_IntervalTree, (type(self), 0x05502fb, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_IntervalTree__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(2, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_6pandas_5_libs_8interval___pyx_unpickle_IntervalTree__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_IntervalTree, (type(self), 0x05502fb, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_IntervalTree__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.IntervalTree.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":209
 * 
 * 
 * cdef take(ndarray source, ndarray indices):             # <<<<<<<<<<<<<<
 *     """Take the given positions from a 1D ndarray
 *     """
 */

static PyObject *__pyx_f_6pandas_5_libs_8interval_take(PyArrayObject *__pyx_v_source, PyArrayObject *__pyx_v_indices) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("take", 0);

  /* "pandas/_libs/intervaltree.pxi":212
 *     """Take the given positions from a 1D ndarray
 *     """
 *     return PyArray_Take(source, indices, 0)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_Take(__pyx_v_source, ((PyObject *)__pyx_v_indices), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 212, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pandas/_libs/intervaltree.pxi":209
 * 
 * 
 * cdef take(ndarray source, ndarray indices):             # <<<<<<<<<<<<<<
 *     """Take the given positions from a 1D ndarray
 *     """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.take", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":215
 * 
 * 
 * cdef sort_values_and_indices(all_values, all_indices, subset):             # <<<<<<<<<<<<<<
 *     indices = take(all_indices, subset)
 *     values = take(all_values, subset)
 */

static PyObject *__pyx_f_6pandas_5_libs_8interval_sort_values_and_indices(PyObject *__pyx_v_all_values, PyObject *__pyx_v_all_indices, PyObject *__pyx_v_subset) {
  PyObject *__pyx_v_indices = NULL;
  PyObject *__pyx_v_values = NULL;
  PyObject *__pyx_v_sorter = NULL;
  PyObject *__pyx_v_sorted_values = NULL;
  PyObject *__pyx_v_sorted_indices = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("sort_values_and_indices", 0);

  /* "pandas/_libs/intervaltree.pxi":216
 * 
 * cdef sort_values_and_indices(all_values, all_indices, subset):
 *     indices = take(all_indices, subset)             # <<<<<<<<<<<<<<
 *     values = take(all_values, subset)
 *     sorter = PyArray_ArgSort(values, 0, NPY_QUICKSORT)
 */
  if (!(likely(((__pyx_v_all_indices) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_all_indices, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(1, 216, __pyx_L1_error)
  if (!(likely(((__pyx_v_subset) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_subset, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(1, 216, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_6pandas_5_libs_8interval_take(((PyArrayObject *)__pyx_v_all_indices), ((PyArrayObject *)__pyx_v_subset)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 216, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_indices = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":217
 * cdef sort_values_and_indices(all_values, all_indices, subset):
 *     indices = take(all_indices, subset)
 *     values = take(all_values, subset)             # <<<<<<<<<<<<<<
 *     sorter = PyArray_ArgSort(values, 0, NPY_QUICKSORT)
 *     sorted_values = take(values, sorter)
 */
  if (!(likely(((__pyx_v_all_values) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_all_values, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(1, 217, __pyx_L1_error)
  if (!(likely(((__pyx_v_subset) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_subset, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(1, 217, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_6pandas_5_libs_8interval_take(((PyArrayObject *)__pyx_v_all_values), ((PyArrayObject *)__pyx_v_subset)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 217, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_values = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":218
 *     indices = take(all_indices, subset)
 *     values = take(all_values, subset)
 *     sorter = PyArray_ArgSort(values, 0, NPY_QUICKSORT)             # <<<<<<<<<<<<<<
 *     sorted_values = take(values, sorter)
 *     sorted_indices = take(indices, sorter)
 */
  if (!(likely(((__pyx_v_values) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_values, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(1, 218, __pyx_L1_error)
  __pyx_t_1 = PyArray_ArgSort(((PyArrayObject *)__pyx_v_values), 0, NPY_QUICKSORT); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 218, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_sorter = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":219
 *     values = take(all_values, subset)
 *     sorter = PyArray_ArgSort(values, 0, NPY_QUICKSORT)
 *     sorted_values = take(values, sorter)             # <<<<<<<<<<<<<<
 *     sorted_indices = take(indices, sorter)
 *     return sorted_values, sorted_indices
 */
  if (!(likely(((__pyx_v_values) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_values, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(1, 219, __pyx_L1_error)
  if (!(likely(((__pyx_v_sorter) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_sorter, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(1, 219, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_6pandas_5_libs_8interval_take(((PyArrayObject *)__pyx_v_values), ((PyArrayObject *)__pyx_v_sorter)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 219, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_sorted_values = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":220
 *     sorter = PyArray_ArgSort(values, 0, NPY_QUICKSORT)
 *     sorted_values = take(values, sorter)
 *     sorted_indices = take(indices, sorter)             # <<<<<<<<<<<<<<
 *     return sorted_values, sorted_indices
 * 
 */
  if (!(likely(((__pyx_v_indices) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_indices, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(1, 220, __pyx_L1_error)
  if (!(likely(((__pyx_v_sorter) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_sorter, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(1, 220, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_6pandas_5_libs_8interval_take(((PyArrayObject *)__pyx_v_indices), ((PyArrayObject *)__pyx_v_sorter)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 220, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_sorted_indices = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":221
 *     sorted_values = take(values, sorter)
 *     sorted_indices = take(indices, sorter)
 *     return sorted_values, sorted_indices             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 221, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_sorted_values);
  __Pyx_GIVEREF(__pyx_v_sorted_values);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_sorted_values);
  __Pyx_INCREF(__pyx_v_sorted_indices);
  __Pyx_GIVEREF(__pyx_v_sorted_indices);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_sorted_indices);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pandas/_libs/intervaltree.pxi":215
 * 
 * 
 * cdef sort_values_and_indices(all_values, all_indices, subset):             # <<<<<<<<<<<<<<
 *     indices = take(all_indices, subset)
 *     values = take(all_values, subset)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.sort_values_and_indices", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_indices);
  __Pyx_XDECREF(__pyx_v_values);
  __Pyx_XDECREF(__pyx_v_sorter);
  __Pyx_XDECREF(__pyx_v_sorted_values);
  __Pyx_XDECREF(__pyx_v_sorted_indices);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":248
 *         readonly bint is_leaf_node
 * 
 *     def __init__(self,             # <<<<<<<<<<<<<<
 *                  ndarray[float32_t, ndim=1] left,
 *                  ndarray[float32_t, ndim=1] right,
 */

/* Python wrapper */
static int __pyx_pw_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyArrayObject *__pyx_v_left = 0;
  PyArrayObject *__pyx_v_right = 0;
  PyArrayObject *__pyx_v_indices = 0;
  __pyx_t_5numpy_int64_t __pyx_v_leaf_size;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_left,&__pyx_n_s_right,&__pyx_n_s_indices,&__pyx_n_s_leaf_size,0};
    PyObject* values[4] = {0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_left)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_right)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 4, 4, 1); __PYX_ERR(1, 248, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_indices)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 4, 4, 2); __PYX_ERR(1, 248, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_leaf_size)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 4, 4, 3); __PYX_ERR(1, 248, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(1, 248, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 4) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
    }
    __pyx_v_left = ((PyArrayObject *)values[0]);
    __pyx_v_right = ((PyArrayObject *)values[1]);
    __pyx_v_indices = ((PyArrayObject *)values[2]);
    __pyx_v_leaf_size = __Pyx_PyInt_As_npy_int64(values[3]); if (unlikely((__pyx_v_leaf_size == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(1, 252, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 4, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 248, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Float32ClosedLeftIntervalNode.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_left), __pyx_ptype_5numpy_ndarray, 1, "left", 0))) __PYX_ERR(1, 249, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_right), __pyx_ptype_5numpy_ndarray, 1, "right", 0))) __PYX_ERR(1, 250, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_indices), __pyx_ptype_5numpy_ndarray, 1, "indices", 0))) __PYX_ERR(1, 251, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode___init__(((struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode *)__pyx_v_self), __pyx_v_left, __pyx_v_right, __pyx_v_indices, __pyx_v_leaf_size);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode___init__(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode *__pyx_v_self, PyArrayObject *__pyx_v_left, PyArrayObject *__pyx_v_right, PyArrayObject *__pyx_v_indices, __pyx_t_5numpy_int64_t __pyx_v_leaf_size) {
  PyObject *__pyx_v_left_set = NULL;
  PyObject *__pyx_v_right_set = NULL;
  PyObject *__pyx_v_center_set = NULL;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_indices;
  __Pyx_Buffer __pyx_pybuffer_indices;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_left;
  __Pyx_Buffer __pyx_pybuffer_left;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_right;
  __Pyx_Buffer __pyx_pybuffer_right;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __pyx_t_5numpy_float32_t __pyx_t_6;
  __Pyx_memviewslice __pyx_t_7 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_8 = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  __Pyx_memviewslice __pyx_t_11 = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *(*__pyx_t_12)(PyObject *);
  size_t __pyx_t_13;
  __Pyx_RefNannySetupContext("__init__", 0);
  __pyx_pybuffer_left.pybuffer.buf = NULL;
  __pyx_pybuffer_left.refcount = 0;
  __pyx_pybuffernd_left.data = NULL;
  __pyx_pybuffernd_left.rcbuffer = &__pyx_pybuffer_left;
  __pyx_pybuffer_right.pybuffer.buf = NULL;
  __pyx_pybuffer_right.refcount = 0;
  __pyx_pybuffernd_right.data = NULL;
  __pyx_pybuffernd_right.rcbuffer = &__pyx_pybuffer_right;
  __pyx_pybuffer_indices.pybuffer.buf = NULL;
  __pyx_pybuffer_indices.refcount = 0;
  __pyx_pybuffernd_indices.data = NULL;
  __pyx_pybuffernd_indices.rcbuffer = &__pyx_pybuffer_indices;
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_left.rcbuffer->pybuffer, (PyObject*)__pyx_v_left, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float32_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 248, __pyx_L1_error)
  }
  __pyx_pybuffernd_left.diminfo[0].strides = __pyx_pybuffernd_left.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_left.diminfo[0].shape = __pyx_pybuffernd_left.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_right.rcbuffer->pybuffer, (PyObject*)__pyx_v_right, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float32_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 248, __pyx_L1_error)
  }
  __pyx_pybuffernd_right.diminfo[0].strides = __pyx_pybuffernd_right.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_right.diminfo[0].shape = __pyx_pybuffernd_right.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_indices.rcbuffer->pybuffer, (PyObject*)__pyx_v_indices, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 248, __pyx_L1_error)
  }
  __pyx_pybuffernd_indices.diminfo[0].strides = __pyx_pybuffernd_indices.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_indices.diminfo[0].shape = __pyx_pybuffernd_indices.rcbuffer->pybuffer.shape[0];

  /* "pandas/_libs/intervaltree.pxi":254
 *                  int64_t leaf_size):
 * 
 *         self.n_elements = len(left)             # <<<<<<<<<<<<<<
 *         self.leaf_size = leaf_size
 * 
 */
  __pyx_t_1 = PyObject_Length(((PyObject *)__pyx_v_left)); if (unlikely(__pyx_t_1 == ((Py_ssize_t)-1))) __PYX_ERR(1, 254, __pyx_L1_error)
  __pyx_v_self->n_elements = __pyx_t_1;

  /* "pandas/_libs/intervaltree.pxi":255
 * 
 *         self.n_elements = len(left)
 *         self.leaf_size = leaf_size             # <<<<<<<<<<<<<<
 * 
 *         # min_left and min_right are used to speed-up query by skipping
 */
  __pyx_v_self->leaf_size = __pyx_v_leaf_size;

  /* "pandas/_libs/intervaltree.pxi":260
 *         # query on sub-nodes. If this node has size 0, query is cheap,
 *         # so these values don't matter.
 *         if left.size > 0:             # <<<<<<<<<<<<<<
 *             self.min_left = left.min()
 *             self.max_right = right.max()
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_left), __pyx_n_s_size); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 260, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyObject_RichCompare(__pyx_t_2, __pyx_int_0, Py_GT); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 260, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(1, 260, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_4) {

    /* "pandas/_libs/intervaltree.pxi":261
 *         # so these values don't matter.
 *         if left.size > 0:
 *             self.min_left = left.min()             # <<<<<<<<<<<<<<
 *             self.max_right = right.max()
 *         else:
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_left), __pyx_n_s_min); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 261, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 261, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_6 = __pyx_PyFloat_AsFloat(__pyx_t_3); if (unlikely((__pyx_t_6 == ((npy_float32)-1)) && PyErr_Occurred())) __PYX_ERR(1, 261, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_self->min_left = __pyx_t_6;

    /* "pandas/_libs/intervaltree.pxi":262
 *         if left.size > 0:
 *             self.min_left = left.min()
 *             self.max_right = right.max()             # <<<<<<<<<<<<<<
 *         else:
 *             self.min_left = 0
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_right), __pyx_n_s_max); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 262, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 262, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_6 = __pyx_PyFloat_AsFloat(__pyx_t_3); if (unlikely((__pyx_t_6 == ((npy_float32)-1)) && PyErr_Occurred())) __PYX_ERR(1, 262, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_self->max_right = __pyx_t_6;

    /* "pandas/_libs/intervaltree.pxi":260
 *         # query on sub-nodes. If this node has size 0, query is cheap,
 *         # so these values don't matter.
 *         if left.size > 0:             # <<<<<<<<<<<<<<
 *             self.min_left = left.min()
 *             self.max_right = right.max()
 */
    goto __pyx_L3;
  }

  /* "pandas/_libs/intervaltree.pxi":264
 *             self.max_right = right.max()
 *         else:
 *             self.min_left = 0             # <<<<<<<<<<<<<<
 *             self.max_right = 0
 * 
 */
  /*else*/ {
    __pyx_v_self->min_left = 0.0;

    /* "pandas/_libs/intervaltree.pxi":265
 *         else:
 *             self.min_left = 0
 *             self.max_right = 0             # <<<<<<<<<<<<<<
 * 
 *         if self.n_elements <= leaf_size:
 */
    __pyx_v_self->max_right = 0.0;
  }
  __pyx_L3:;

  /* "pandas/_libs/intervaltree.pxi":267
 *             self.max_right = 0
 * 
 *         if self.n_elements <= leaf_size:             # <<<<<<<<<<<<<<
 *             # make this a terminal (leaf) node
 *             self.is_leaf_node = True
 */
  __pyx_t_4 = ((__pyx_v_self->n_elements <= __pyx_v_leaf_size) != 0);
  if (__pyx_t_4) {

    /* "pandas/_libs/intervaltree.pxi":269
 *         if self.n_elements <= leaf_size:
 *             # make this a terminal (leaf) node
 *             self.is_leaf_node = True             # <<<<<<<<<<<<<<
 *             self.left = left
 *             self.right = right
 */
    __pyx_v_self->is_leaf_node = 1;

    /* "pandas/_libs/intervaltree.pxi":270
 *             # make this a terminal (leaf) node
 *             self.is_leaf_node = True
 *             self.left = left             # <<<<<<<<<<<<<<
 *             self.right = right
 *             self.indices = indices
 */
    __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_float32_t(((PyObject *)__pyx_v_left), PyBUF_WRITABLE); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(1, 270, __pyx_L1_error)
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->left, 0);
    __pyx_v_self->left = __pyx_t_7;
    __pyx_t_7.memview = NULL;
    __pyx_t_7.data = NULL;

    /* "pandas/_libs/intervaltree.pxi":271
 *             self.is_leaf_node = True
 *             self.left = left
 *             self.right = right             # <<<<<<<<<<<<<<
 *             self.indices = indices
 *             self.n_center = 0
 */
    __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_float32_t(((PyObject *)__pyx_v_right), PyBUF_WRITABLE); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(1, 271, __pyx_L1_error)
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->right, 0);
    __pyx_v_self->right = __pyx_t_7;
    __pyx_t_7.memview = NULL;
    __pyx_t_7.data = NULL;

    /* "pandas/_libs/intervaltree.pxi":272
 *             self.left = left
 *             self.right = right
 *             self.indices = indices             # <<<<<<<<<<<<<<
 *             self.n_center = 0
 *         else:
 */
    __pyx_t_8 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(((PyObject *)__pyx_v_indices), PyBUF_WRITABLE); if (unlikely(!__pyx_t_8.memview)) __PYX_ERR(1, 272, __pyx_L1_error)
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->indices, 0);
    __pyx_v_self->indices = __pyx_t_8;
    __pyx_t_8.memview = NULL;
    __pyx_t_8.data = NULL;

    /* "pandas/_libs/intervaltree.pxi":273
 *             self.right = right
 *             self.indices = indices
 *             self.n_center = 0             # <<<<<<<<<<<<<<
 *         else:
 *             # calculate a pivot so we can create child nodes
 */
    __pyx_v_self->n_center = 0;

    /* "pandas/_libs/intervaltree.pxi":267
 *             self.max_right = 0
 * 
 *         if self.n_elements <= leaf_size:             # <<<<<<<<<<<<<<
 *             # make this a terminal (leaf) node
 *             self.is_leaf_node = True
 */
    goto __pyx_L4;
  }

  /* "pandas/_libs/intervaltree.pxi":276
 *         else:
 *             # calculate a pivot so we can create child nodes
 *             self.is_leaf_node = False             # <<<<<<<<<<<<<<
 *             self.pivot = np.median(left / 2 + right / 2)
 *             left_set, right_set, center_set = self.classify_intervals(
 */
  /*else*/ {
    __pyx_v_self->is_leaf_node = 0;

    /* "pandas/_libs/intervaltree.pxi":277
 *             # calculate a pivot so we can create child nodes
 *             self.is_leaf_node = False
 *             self.pivot = np.median(left / 2 + right / 2)             # <<<<<<<<<<<<<<
 *             left_set, right_set, center_set = self.classify_intervals(
 *                 left, right)
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 277, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_median); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 277, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyNumber_Divide(((PyObject *)__pyx_v_left), __pyx_int_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 277, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_9 = __Pyx_PyNumber_Divide(((PyObject *)__pyx_v_right), __pyx_int_2); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 277, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_10 = PyNumber_Add(__pyx_t_2, __pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 277, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_9 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_3 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_9, __pyx_t_10) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_10);
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 277, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_6 = __pyx_PyFloat_AsFloat(__pyx_t_3); if (unlikely((__pyx_t_6 == ((npy_float32)-1)) && PyErr_Occurred())) __PYX_ERR(1, 277, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_self->pivot = __pyx_t_6;

    /* "pandas/_libs/intervaltree.pxi":279
 *             self.pivot = np.median(left / 2 + right / 2)
 *             left_set, right_set, center_set = self.classify_intervals(
 *                 left, right)             # <<<<<<<<<<<<<<
 * 
 *             self.left_node = self.new_child_node(left, right,
 */
    __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_float32_t(((PyObject *)__pyx_v_left), PyBUF_WRITABLE); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(1, 279, __pyx_L1_error)
    __pyx_t_11 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_float32_t(((PyObject *)__pyx_v_right), PyBUF_WRITABLE); if (unlikely(!__pyx_t_11.memview)) __PYX_ERR(1, 279, __pyx_L1_error)

    /* "pandas/_libs/intervaltree.pxi":278
 *             self.is_leaf_node = False
 *             self.pivot = np.median(left / 2 + right / 2)
 *             left_set, right_set, center_set = self.classify_intervals(             # <<<<<<<<<<<<<<
 *                 left, right)
 * 
 */
    __pyx_t_3 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode *)__pyx_v_self->__pyx_vtab)->classify_intervals(__pyx_v_self, __pyx_t_7, __pyx_t_11); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 278, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __PYX_XDEC_MEMVIEW(&__pyx_t_7, 1);
    __pyx_t_7.memview = NULL;
    __pyx_t_7.data = NULL;
    __PYX_XDEC_MEMVIEW(&__pyx_t_11, 1);
    __pyx_t_11.memview = NULL;
    __pyx_t_11.data = NULL;
    if ((likely(PyTuple_CheckExact(__pyx_t_3))) || (PyList_CheckExact(__pyx_t_3))) {
      PyObject* sequence = __pyx_t_3;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 3)) {
        if (size > 3) __Pyx_RaiseTooManyValuesError(3);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(1, 278, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_5 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_10 = PyTuple_GET_ITEM(sequence, 1); 
        __pyx_t_9 = PyTuple_GET_ITEM(sequence, 2); 
      } else {
        __pyx_t_5 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_10 = PyList_GET_ITEM(sequence, 1); 
        __pyx_t_9 = PyList_GET_ITEM(sequence, 2); 
      }
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_10);
      __Pyx_INCREF(__pyx_t_9);
      #else
      __pyx_t_5 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 278, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_10 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 278, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_9 = PySequence_ITEM(sequence, 2); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 278, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      #endif
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_2 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 278, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_12 = Py_TYPE(__pyx_t_2)->tp_iternext;
      index = 0; __pyx_t_5 = __pyx_t_12(__pyx_t_2); if (unlikely(!__pyx_t_5)) goto __pyx_L5_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_5);
      index = 1; __pyx_t_10 = __pyx_t_12(__pyx_t_2); if (unlikely(!__pyx_t_10)) goto __pyx_L5_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_10);
      index = 2; __pyx_t_9 = __pyx_t_12(__pyx_t_2); if (unlikely(!__pyx_t_9)) goto __pyx_L5_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_9);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_12(__pyx_t_2), 3) < 0) __PYX_ERR(1, 278, __pyx_L1_error)
      __pyx_t_12 = NULL;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      goto __pyx_L6_unpacking_done;
      __pyx_L5_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_12 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(1, 278, __pyx_L1_error)
      __pyx_L6_unpacking_done:;
    }
    __pyx_v_left_set = __pyx_t_5;
    __pyx_t_5 = 0;
    __pyx_v_right_set = __pyx_t_10;
    __pyx_t_10 = 0;
    __pyx_v_center_set = __pyx_t_9;
    __pyx_t_9 = 0;

    /* "pandas/_libs/intervaltree.pxi":282
 * 
 *             self.left_node = self.new_child_node(left, right,
 *                                                  indices, left_set)             # <<<<<<<<<<<<<<
 *             self.right_node = self.new_child_node(left, right,
 *                                                   indices, right_set)
 */
    if (!(likely(((__pyx_v_left_set) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_left_set, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(1, 282, __pyx_L1_error)

    /* "pandas/_libs/intervaltree.pxi":281
 *                 left, right)
 * 
 *             self.left_node = self.new_child_node(left, right,             # <<<<<<<<<<<<<<
 *                                                  indices, left_set)
 *             self.right_node = self.new_child_node(left, right,
 */
    __pyx_t_3 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode *)__pyx_v_self->__pyx_vtab)->new_child_node(__pyx_v_self, ((PyArrayObject *)__pyx_v_left), ((PyArrayObject *)__pyx_v_right), ((PyArrayObject *)__pyx_v_indices), ((PyArrayObject *)__pyx_v_left_set)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 281, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode))))) __PYX_ERR(1, 281, __pyx_L1_error)
    __Pyx_GIVEREF(__pyx_t_3);
    __Pyx_GOTREF(__pyx_v_self->left_node);
    __Pyx_DECREF(((PyObject *)__pyx_v_self->left_node));
    __pyx_v_self->left_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode *)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "pandas/_libs/intervaltree.pxi":284
 *                                                  indices, left_set)
 *             self.right_node = self.new_child_node(left, right,
 *                                                   indices, right_set)             # <<<<<<<<<<<<<<
 * 
 *             self.center_left_values, self.center_left_indices = \
 */
    if (!(likely(((__pyx_v_right_set) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_right_set, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(1, 284, __pyx_L1_error)

    /* "pandas/_libs/intervaltree.pxi":283
 *             self.left_node = self.new_child_node(left, right,
 *                                                  indices, left_set)
 *             self.right_node = self.new_child_node(left, right,             # <<<<<<<<<<<<<<
 *                                                   indices, right_set)
 * 
 */
    __pyx_t_3 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode *)__pyx_v_self->__pyx_vtab)->new_child_node(__pyx_v_self, ((PyArrayObject *)__pyx_v_left), ((PyArrayObject *)__pyx_v_right), ((PyArrayObject *)__pyx_v_indices), ((PyArrayObject *)__pyx_v_right_set)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 283, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode))))) __PYX_ERR(1, 283, __pyx_L1_error)
    __Pyx_GIVEREF(__pyx_t_3);
    __Pyx_GOTREF(__pyx_v_self->right_node);
    __Pyx_DECREF(((PyObject *)__pyx_v_self->right_node));
    __pyx_v_self->right_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode *)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "pandas/_libs/intervaltree.pxi":287
 * 
 *             self.center_left_values, self.center_left_indices = \
 *                 sort_values_and_indices(left, indices, center_set)             # <<<<<<<<<<<<<<
 *             self.center_right_values, self.center_right_indices = \
 *                 sort_values_and_indices(right, indices, center_set)
 */
    __pyx_t_3 = __pyx_f_6pandas_5_libs_8interval_sort_values_and_indices(((PyObject *)__pyx_v_left), ((PyObject *)__pyx_v_indices), __pyx_v_center_set); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 287, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if ((likely(PyTuple_CheckExact(__pyx_t_3))) || (PyList_CheckExact(__pyx_t_3))) {
      PyObject* sequence = __pyx_t_3;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(1, 286, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_9 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_10 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_9 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_10 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_9);
      __Pyx_INCREF(__pyx_t_10);
      #else
      __pyx_t_9 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 286, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_10 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 286, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      #endif
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_5 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 286, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_12 = Py_TYPE(__pyx_t_5)->tp_iternext;
      index = 0; __pyx_t_9 = __pyx_t_12(__pyx_t_5); if (unlikely(!__pyx_t_9)) goto __pyx_L7_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_9);
      index = 1; __pyx_t_10 = __pyx_t_12(__pyx_t_5); if (unlikely(!__pyx_t_10)) goto __pyx_L7_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_10);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_12(__pyx_t_5), 2) < 0) __PYX_ERR(1, 286, __pyx_L1_error)
      __pyx_t_12 = NULL;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      goto __pyx_L8_unpacking_done;
      __pyx_L7_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_12 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(1, 286, __pyx_L1_error)
      __pyx_L8_unpacking_done:;
    }

    /* "pandas/_libs/intervaltree.pxi":286
 *                                                   indices, right_set)
 * 
 *             self.center_left_values, self.center_left_indices = \             # <<<<<<<<<<<<<<
 *                 sort_values_and_indices(left, indices, center_set)
 *             self.center_right_values, self.center_right_indices = \
 */
    __pyx_t_11 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_float32_t(__pyx_t_9, PyBUF_WRITABLE); if (unlikely(!__pyx_t_11.memview)) __PYX_ERR(1, 286, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_8 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_t_10, PyBUF_WRITABLE); if (unlikely(!__pyx_t_8.memview)) __PYX_ERR(1, 286, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->center_left_values, 0);
    __pyx_v_self->center_left_values = __pyx_t_11;
    __pyx_t_11.memview = NULL;
    __pyx_t_11.data = NULL;
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->center_left_indices, 0);
    __pyx_v_self->center_left_indices = __pyx_t_8;
    __pyx_t_8.memview = NULL;
    __pyx_t_8.data = NULL;

    /* "pandas/_libs/intervaltree.pxi":289
 *                 sort_values_and_indices(left, indices, center_set)
 *             self.center_right_values, self.center_right_indices = \
 *                 sort_values_and_indices(right, indices, center_set)             # <<<<<<<<<<<<<<
 *             self.n_center = len(self.center_left_indices)
 * 
 */
    __pyx_t_3 = __pyx_f_6pandas_5_libs_8interval_sort_values_and_indices(((PyObject *)__pyx_v_right), ((PyObject *)__pyx_v_indices), __pyx_v_center_set); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 289, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if ((likely(PyTuple_CheckExact(__pyx_t_3))) || (PyList_CheckExact(__pyx_t_3))) {
      PyObject* sequence = __pyx_t_3;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(1, 288, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_10 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_9 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_10 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_9 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_10);
      __Pyx_INCREF(__pyx_t_9);
      #else
      __pyx_t_10 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 288, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_9 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 288, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      #endif
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_5 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 288, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_12 = Py_TYPE(__pyx_t_5)->tp_iternext;
      index = 0; __pyx_t_10 = __pyx_t_12(__pyx_t_5); if (unlikely(!__pyx_t_10)) goto __pyx_L9_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_10);
      index = 1; __pyx_t_9 = __pyx_t_12(__pyx_t_5); if (unlikely(!__pyx_t_9)) goto __pyx_L9_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_9);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_12(__pyx_t_5), 2) < 0) __PYX_ERR(1, 288, __pyx_L1_error)
      __pyx_t_12 = NULL;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      goto __pyx_L10_unpacking_done;
      __pyx_L9_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_12 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(1, 288, __pyx_L1_error)
      __pyx_L10_unpacking_done:;
    }

    /* "pandas/_libs/intervaltree.pxi":288
 *             self.center_left_values, self.center_left_indices = \
 *                 sort_values_and_indices(left, indices, center_set)
 *             self.center_right_values, self.center_right_indices = \             # <<<<<<<<<<<<<<
 *                 sort_values_and_indices(right, indices, center_set)
 *             self.n_center = len(self.center_left_indices)
 */
    __pyx_t_11 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_float32_t(__pyx_t_10, PyBUF_WRITABLE); if (unlikely(!__pyx_t_11.memview)) __PYX_ERR(1, 288, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_t_8 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_t_9, PyBUF_WRITABLE); if (unlikely(!__pyx_t_8.memview)) __PYX_ERR(1, 288, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->center_right_values, 0);
    __pyx_v_self->center_right_values = __pyx_t_11;
    __pyx_t_11.memview = NULL;
    __pyx_t_11.data = NULL;
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->center_right_indices, 0);
    __pyx_v_self->center_right_indices = __pyx_t_8;
    __pyx_t_8.memview = NULL;
    __pyx_t_8.data = NULL;

    /* "pandas/_libs/intervaltree.pxi":290
 *             self.center_right_values, self.center_right_indices = \
 *                 sort_values_and_indices(right, indices, center_set)
 *             self.n_center = len(self.center_left_indices)             # <<<<<<<<<<<<<<
 * 
 *     @cython.wraparound(False)
 */
    if (unlikely(!__pyx_v_self->center_left_indices.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(1, 290, __pyx_L1_error)}
    __pyx_t_13 = __Pyx_MemoryView_Len(__pyx_v_self->center_left_indices); 
    __pyx_v_self->n_center = __pyx_t_13;
  }
  __pyx_L4:;

  /* "pandas/_libs/intervaltree.pxi":248
 *         readonly bint is_leaf_node
 * 
 *     def __init__(self,             # <<<<<<<<<<<<<<
 *                  ndarray[float32_t, ndim=1] left,
 *                  ndarray[float32_t, ndim=1] right,
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __PYX_XDEC_MEMVIEW(&__pyx_t_7, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_8, 1);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __PYX_XDEC_MEMVIEW(&__pyx_t_11, 1);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indices.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_left.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_right.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("pandas._libs.interval.Float32ClosedLeftIntervalNode.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indices.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_left.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_right.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF(__pyx_v_left_set);
  __Pyx_XDECREF(__pyx_v_right_set);
  __Pyx_XDECREF(__pyx_v_center_set);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":294
 *     @cython.wraparound(False)
 *     @cython.boundscheck(False)
 *     cdef classify_intervals(self, float32_t[:] left, float32_t[:] right):             # <<<<<<<<<<<<<<
 *         """Classify the given intervals based upon whether they fall to the
 *         left, right, or overlap with this node's pivot.
 */

static PyObject *__pyx_f_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_classify_intervals(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode *__pyx_v_self, __Pyx_memviewslice __pyx_v_left, __Pyx_memviewslice __pyx_v_right) {
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_left_ind = 0;
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_right_ind = 0;
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_overlapping_ind = 0;
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __pyx_t_5numpy_int64_t __pyx_t_2;
  __pyx_t_5numpy_int64_t __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  int __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  __Pyx_RefNannySetupContext("classify_intervals", 0);

  /* "pandas/_libs/intervaltree.pxi":302
 *             Py_ssize_t i
 * 
 *         left_ind = Int64Vector()             # <<<<<<<<<<<<<<
 *         right_ind = Int64Vector()
 *         overlapping_ind = Int64Vector()
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 302, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_left_ind = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":303
 * 
 *         left_ind = Int64Vector()
 *         right_ind = Int64Vector()             # <<<<<<<<<<<<<<
 *         overlapping_ind = Int64Vector()
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 303, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_right_ind = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":304
 *         left_ind = Int64Vector()
 *         right_ind = Int64Vector()
 *         overlapping_ind = Int64Vector()             # <<<<<<<<<<<<<<
 * 
 *         for i in range(self.n_elements):
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 304, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_overlapping_ind = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":306
 *         overlapping_ind = Int64Vector()
 * 
 *         for i in range(self.n_elements):             # <<<<<<<<<<<<<<
 *             if right[i] <= self.pivot:
 *                 left_ind.append(i)
 */
  __pyx_t_2 = __pyx_v_self->n_elements;
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "pandas/_libs/intervaltree.pxi":307
 * 
 *         for i in range(self.n_elements):
 *             if right[i] <= self.pivot:             # <<<<<<<<<<<<<<
 *                 left_ind.append(i)
 *             elif self.pivot < left[i]:
 */
    __pyx_t_5 = __pyx_v_i;
    __pyx_t_6 = (((*((__pyx_t_5numpy_float32_t *) ( /* dim=0 */ (__pyx_v_right.data + __pyx_t_5 * __pyx_v_right.strides[0]) ))) <= __pyx_v_self->pivot) != 0);
    if (__pyx_t_6) {

      /* "pandas/_libs/intervaltree.pxi":308
 *         for i in range(self.n_elements):
 *             if right[i] <= self.pivot:
 *                 left_ind.append(i)             # <<<<<<<<<<<<<<
 *             elif self.pivot < left[i]:
 *                 right_ind.append(i)
 */
      ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_left_ind->__pyx_vtab)->append(__pyx_v_left_ind, __pyx_v_i);

      /* "pandas/_libs/intervaltree.pxi":307
 * 
 *         for i in range(self.n_elements):
 *             if right[i] <= self.pivot:             # <<<<<<<<<<<<<<
 *                 left_ind.append(i)
 *             elif self.pivot < left[i]:
 */
      goto __pyx_L5;
    }

    /* "pandas/_libs/intervaltree.pxi":309
 *             if right[i] <= self.pivot:
 *                 left_ind.append(i)
 *             elif self.pivot < left[i]:             # <<<<<<<<<<<<<<
 *                 right_ind.append(i)
 *             else:
 */
    __pyx_t_7 = __pyx_v_i;
    __pyx_t_6 = ((__pyx_v_self->pivot < (*((__pyx_t_5numpy_float32_t *) ( /* dim=0 */ (__pyx_v_left.data + __pyx_t_7 * __pyx_v_left.strides[0]) )))) != 0);
    if (__pyx_t_6) {

      /* "pandas/_libs/intervaltree.pxi":310
 *                 left_ind.append(i)
 *             elif self.pivot < left[i]:
 *                 right_ind.append(i)             # <<<<<<<<<<<<<<
 *             else:
 *                 overlapping_ind.append(i)
 */
      ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_right_ind->__pyx_vtab)->append(__pyx_v_right_ind, __pyx_v_i);

      /* "pandas/_libs/intervaltree.pxi":309
 *             if right[i] <= self.pivot:
 *                 left_ind.append(i)
 *             elif self.pivot < left[i]:             # <<<<<<<<<<<<<<
 *                 right_ind.append(i)
 *             else:
 */
      goto __pyx_L5;
    }

    /* "pandas/_libs/intervaltree.pxi":312
 *                 right_ind.append(i)
 *             else:
 *                 overlapping_ind.append(i)             # <<<<<<<<<<<<<<
 * 
 *         return (left_ind.to_array(),
 */
    /*else*/ {
      ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_overlapping_ind->__pyx_vtab)->append(__pyx_v_overlapping_ind, __pyx_v_i);
    }
    __pyx_L5:;
  }

  /* "pandas/_libs/intervaltree.pxi":314
 *                 overlapping_ind.append(i)
 * 
 *         return (left_ind.to_array(),             # <<<<<<<<<<<<<<
 *                 right_ind.to_array(),
 *                 overlapping_ind.to_array())
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_left_ind->__pyx_vtab)->to_array(__pyx_v_left_ind, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 314, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "pandas/_libs/intervaltree.pxi":315
 * 
 *         return (left_ind.to_array(),
 *                 right_ind.to_array(),             # <<<<<<<<<<<<<<
 *                 overlapping_ind.to_array())
 * 
 */
  __pyx_t_8 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_right_ind->__pyx_vtab)->to_array(__pyx_v_right_ind, 0); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 315, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);

  /* "pandas/_libs/intervaltree.pxi":316
 *         return (left_ind.to_array(),
 *                 right_ind.to_array(),
 *                 overlapping_ind.to_array())             # <<<<<<<<<<<<<<
 * 
 *     cdef new_child_node(self,
 */
  __pyx_t_9 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_overlapping_ind->__pyx_vtab)->to_array(__pyx_v_overlapping_ind, 0); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 316, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);

  /* "pandas/_libs/intervaltree.pxi":314
 *                 overlapping_ind.append(i)
 * 
 *         return (left_ind.to_array(),             # <<<<<<<<<<<<<<
 *                 right_ind.to_array(),
 *                 overlapping_ind.to_array())
 */
  __pyx_t_10 = PyTuple_New(3); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 314, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_t_8);
  __Pyx_GIVEREF(__pyx_t_9);
  PyTuple_SET_ITEM(__pyx_t_10, 2, __pyx_t_9);
  __pyx_t_1 = 0;
  __pyx_t_8 = 0;
  __pyx_t_9 = 0;
  __pyx_r = __pyx_t_10;
  __pyx_t_10 = 0;
  goto __pyx_L0;

  /* "pandas/_libs/intervaltree.pxi":294
 *     @cython.wraparound(False)
 *     @cython.boundscheck(False)
 *     cdef classify_intervals(self, float32_t[:] left, float32_t[:] right):             # <<<<<<<<<<<<<<
 *         """Classify the given intervals based upon whether they fall to the
 *         left, right, or overlap with this node's pivot.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("pandas._libs.interval.Float32ClosedLeftIntervalNode.classify_intervals", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_left_ind);
  __Pyx_XDECREF((PyObject *)__pyx_v_right_ind);
  __Pyx_XDECREF((PyObject *)__pyx_v_overlapping_ind);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":318
 *                 overlapping_ind.to_array())
 * 
 *     cdef new_child_node(self,             # <<<<<<<<<<<<<<
 *                         ndarray[float32_t, ndim=1] left,
 *                         ndarray[float32_t, ndim=1] right,
 */

static PyObject *__pyx_f_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_new_child_node(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode *__pyx_v_self, PyArrayObject *__pyx_v_left, PyArrayObject *__pyx_v_right, PyArrayObject *__pyx_v_indices, PyArrayObject *__pyx_v_subset) {
  __Pyx_LocalBuf_ND __pyx_pybuffernd_indices;
  __Pyx_Buffer __pyx_pybuffer_indices;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_left;
  __Pyx_Buffer __pyx_pybuffer_left;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_right;
  __Pyx_Buffer __pyx_pybuffer_right;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_subset;
  __Pyx_Buffer __pyx_pybuffer_subset;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyArrayObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyArrayObject *__pyx_t_7 = NULL;
  PyArrayObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  __Pyx_RefNannySetupContext("new_child_node", 0);
  __Pyx_INCREF((PyObject *)__pyx_v_left);
  __Pyx_INCREF((PyObject *)__pyx_v_right);
  __Pyx_INCREF((PyObject *)__pyx_v_indices);
  __pyx_pybuffer_left.pybuffer.buf = NULL;
  __pyx_pybuffer_left.refcount = 0;
  __pyx_pybuffernd_left.data = NULL;
  __pyx_pybuffernd_left.rcbuffer = &__pyx_pybuffer_left;
  __pyx_pybuffer_right.pybuffer.buf = NULL;
  __pyx_pybuffer_right.refcount = 0;
  __pyx_pybuffernd_right.data = NULL;
  __pyx_pybuffernd_right.rcbuffer = &__pyx_pybuffer_right;
  __pyx_pybuffer_indices.pybuffer.buf = NULL;
  __pyx_pybuffer_indices.refcount = 0;
  __pyx_pybuffernd_indices.data = NULL;
  __pyx_pybuffernd_indices.rcbuffer = &__pyx_pybuffer_indices;
  __pyx_pybuffer_subset.pybuffer.buf = NULL;
  __pyx_pybuffer_subset.refcount = 0;
  __pyx_pybuffernd_subset.data = NULL;
  __pyx_pybuffernd_subset.rcbuffer = &__pyx_pybuffer_subset;
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_left.rcbuffer->pybuffer, (PyObject*)__pyx_v_left, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float32_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 318, __pyx_L1_error)
  }
  __pyx_pybuffernd_left.diminfo[0].strides = __pyx_pybuffernd_left.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_left.diminfo[0].shape = __pyx_pybuffernd_left.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_right.rcbuffer->pybuffer, (PyObject*)__pyx_v_right, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float32_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 318, __pyx_L1_error)
  }
  __pyx_pybuffernd_right.diminfo[0].strides = __pyx_pybuffernd_right.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_right.diminfo[0].shape = __pyx_pybuffernd_right.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_indices.rcbuffer->pybuffer, (PyObject*)__pyx_v_indices, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 318, __pyx_L1_error)
  }
  __pyx_pybuffernd_indices.diminfo[0].strides = __pyx_pybuffernd_indices.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_indices.diminfo[0].shape = __pyx_pybuffernd_indices.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_subset.rcbuffer->pybuffer, (PyObject*)__pyx_v_subset, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 318, __pyx_L1_error)
  }
  __pyx_pybuffernd_subset.diminfo[0].strides = __pyx_pybuffernd_subset.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_subset.diminfo[0].shape = __pyx_pybuffernd_subset.rcbuffer->pybuffer.shape[0];

  /* "pandas/_libs/intervaltree.pxi":325
 *         """Create a new child node.
 *         """
 *         left = take(left, subset)             # <<<<<<<<<<<<<<
 *         right = take(right, subset)
 *         indices = take(indices, subset)
 */
  __pyx_t_1 = __pyx_f_6pandas_5_libs_8interval_take(((PyArrayObject *)__pyx_v_left), ((PyArrayObject *)__pyx_v_subset)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 325, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(1, 325, __pyx_L1_error)
  __pyx_t_2 = ((PyArrayObject *)__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_left.rcbuffer->pybuffer);
    __pyx_t_3 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_left.rcbuffer->pybuffer, (PyObject*)__pyx_t_2, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float32_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack);
    if (unlikely(__pyx_t_3 < 0)) {
      PyErr_Fetch(&__pyx_t_4, &__pyx_t_5, &__pyx_t_6);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_left.rcbuffer->pybuffer, (PyObject*)__pyx_v_left, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float32_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_4); Py_XDECREF(__pyx_t_5); Py_XDECREF(__pyx_t_6);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_4, __pyx_t_5, __pyx_t_6);
      }
      __pyx_t_4 = __pyx_t_5 = __pyx_t_6 = 0;
    }
    __pyx_pybuffernd_left.diminfo[0].strides = __pyx_pybuffernd_left.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_left.diminfo[0].shape = __pyx_pybuffernd_left.rcbuffer->pybuffer.shape[0];
    if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(1, 325, __pyx_L1_error)
  }
  __pyx_t_2 = 0;
  __Pyx_DECREF_SET(__pyx_v_left, ((PyArrayObject *)__pyx_t_1));
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":326
 *         """
 *         left = take(left, subset)
 *         right = take(right, subset)             # <<<<<<<<<<<<<<
 *         indices = take(indices, subset)
 *         return Float32ClosedLeftIntervalNode(
 */
  __pyx_t_1 = __pyx_f_6pandas_5_libs_8interval_take(((PyArrayObject *)__pyx_v_right), ((PyArrayObject *)__pyx_v_subset)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 326, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(1, 326, __pyx_L1_error)
  __pyx_t_7 = ((PyArrayObject *)__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_right.rcbuffer->pybuffer);
    __pyx_t_3 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_right.rcbuffer->pybuffer, (PyObject*)__pyx_t_7, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float32_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack);
    if (unlikely(__pyx_t_3 < 0)) {
      PyErr_Fetch(&__pyx_t_6, &__pyx_t_5, &__pyx_t_4);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_right.rcbuffer->pybuffer, (PyObject*)__pyx_v_right, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float32_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_6); Py_XDECREF(__pyx_t_5); Py_XDECREF(__pyx_t_4);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_6, __pyx_t_5, __pyx_t_4);
      }
      __pyx_t_6 = __pyx_t_5 = __pyx_t_4 = 0;
    }
    __pyx_pybuffernd_right.diminfo[0].strides = __pyx_pybuffernd_right.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_right.diminfo[0].shape = __pyx_pybuffernd_right.rcbuffer->pybuffer.shape[0];
    if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(1, 326, __pyx_L1_error)
  }
  __pyx_t_7 = 0;
  __Pyx_DECREF_SET(__pyx_v_right, ((PyArrayObject *)__pyx_t_1));
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":327
 *         left = take(left, subset)
 *         right = take(right, subset)
 *         indices = take(indices, subset)             # <<<<<<<<<<<<<<
 *         return Float32ClosedLeftIntervalNode(
 *             left, right, indices, self.leaf_size)
 */
  __pyx_t_1 = __pyx_f_6pandas_5_libs_8interval_take(((PyArrayObject *)__pyx_v_indices), ((PyArrayObject *)__pyx_v_subset)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 327, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(1, 327, __pyx_L1_error)
  __pyx_t_8 = ((PyArrayObject *)__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indices.rcbuffer->pybuffer);
    __pyx_t_3 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_indices.rcbuffer->pybuffer, (PyObject*)__pyx_t_8, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack);
    if (unlikely(__pyx_t_3 < 0)) {
      PyErr_Fetch(&__pyx_t_4, &__pyx_t_5, &__pyx_t_6);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_indices.rcbuffer->pybuffer, (PyObject*)__pyx_v_indices, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_4); Py_XDECREF(__pyx_t_5); Py_XDECREF(__pyx_t_6);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_4, __pyx_t_5, __pyx_t_6);
      }
      __pyx_t_4 = __pyx_t_5 = __pyx_t_6 = 0;
    }
    __pyx_pybuffernd_indices.diminfo[0].strides = __pyx_pybuffernd_indices.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_indices.diminfo[0].shape = __pyx_pybuffernd_indices.rcbuffer->pybuffer.shape[0];
    if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(1, 327, __pyx_L1_error)
  }
  __pyx_t_8 = 0;
  __Pyx_DECREF_SET(__pyx_v_indices, ((PyArrayObject *)__pyx_t_1));
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":328
 *         right = take(right, subset)
 *         indices = take(indices, subset)
 *         return Float32ClosedLeftIntervalNode(             # <<<<<<<<<<<<<<
 *             left, right, indices, self.leaf_size)
 * 
 */
  __Pyx_XDECREF(__pyx_r);

  /* "pandas/_libs/intervaltree.pxi":329
 *         indices = take(indices, subset)
 *         return Float32ClosedLeftIntervalNode(
 *             left, right, indices, self.leaf_size)             # <<<<<<<<<<<<<<
 * 
 *     @cython.wraparound(False)
 */
  __pyx_t_1 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->leaf_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 329, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "pandas/_libs/intervaltree.pxi":328
 *         right = take(right, subset)
 *         indices = take(indices, subset)
 *         return Float32ClosedLeftIntervalNode(             # <<<<<<<<<<<<<<
 *             left, right, indices, self.leaf_size)
 * 
 */
  __pyx_t_9 = PyTuple_New(4); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 328, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_INCREF(((PyObject *)__pyx_v_left));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_left));
  PyTuple_SET_ITEM(__pyx_t_9, 0, ((PyObject *)__pyx_v_left));
  __Pyx_INCREF(((PyObject *)__pyx_v_right));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_right));
  PyTuple_SET_ITEM(__pyx_t_9, 1, ((PyObject *)__pyx_v_right));
  __Pyx_INCREF(((PyObject *)__pyx_v_indices));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_indices));
  PyTuple_SET_ITEM(__pyx_t_9, 2, ((PyObject *)__pyx_v_indices));
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_9, 3, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode), __pyx_t_9, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 328, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pandas/_libs/intervaltree.pxi":318
 *                 overlapping_ind.to_array())
 * 
 *     cdef new_child_node(self,             # <<<<<<<<<<<<<<
 *                         ndarray[float32_t, ndim=1] left,
 *                         ndarray[float32_t, ndim=1] right,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_9);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indices.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_left.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_right.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_subset.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("pandas._libs.interval.Float32ClosedLeftIntervalNode.new_child_node", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indices.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_left.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_right.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_subset.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF((PyObject *)__pyx_v_left);
  __Pyx_XDECREF((PyObject *)__pyx_v_right);
  __Pyx_XDECREF((PyObject *)__pyx_v_indices);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":334
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_2query[] = "Recursively query this node and its sub-nodes for intervals that\n        overlap with the query point.\n        ";
static PyMethodDef __pyx_mdef_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_3query = {"query", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_3query, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_2query};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_args = 0;
  PyObject *__pyx_v_kwargs = 0;
  CYTHON_UNUSED PyObject *__pyx_v_defaults = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fused_cpdef (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_args,&__pyx_n_s_kwargs,&__pyx_n_s_defaults,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_args)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_kwargs)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fused_cpdef", 1, 3, 3, 1); __PYX_ERR(1, 334, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_defaults)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fused_cpdef", 1, 3, 3, 2); __PYX_ERR(1, 334, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fused_cpdef") < 0)) __PYX_ERR(1, 334, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_args = values[0];
    __pyx_v_kwargs = values[1];
    __pyx_v_defaults = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fused_cpdef", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 334, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Float32ClosedLeftIntervalNode.__pyx_fused_cpdef", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_2query(((struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode *)__pyx_v_signatures), __pyx_v_args, __pyx_v_kwargs, __pyx_v_defaults);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_2query(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode *__pyx_v_signatures, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs, CYTHON_UNUSED PyObject *__pyx_v_defaults) {
  PyObject *__pyx_v_dest_sig = NULL;
  Py_ssize_t __pyx_v_i;
  CYTHON_UNUSED PyTypeObject *__pyx_v_ndarray = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_v_itemsize;
  PyObject *__pyx_v_arg = NULL;
  PyObject *__pyx_v_candidates = NULL;
  PyObject *__pyx_v_sig = NULL;
  int __pyx_v_match_found;
  PyObject *__pyx_v_src_sig = NULL;
  PyObject *__pyx_v_dst_type = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  Py_ssize_t __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  Py_ssize_t __pyx_t_13;
  Py_ssize_t __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  int __pyx_t_16;
  __Pyx_RefNannySetupContext("query", 0);
  __Pyx_INCREF(__pyx_v_kwargs);
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 334, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  PyList_SET_ITEM(__pyx_t_1, 0, Py_None);
  __pyx_v_dest_sig = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_3 = (__pyx_v_kwargs != Py_None);
  __pyx_t_4 = (__pyx_t_3 != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_kwargs); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(1, 334, __pyx_L1_error)
  __pyx_t_3 = ((!__pyx_t_4) != 0);
  __pyx_t_2 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {
    __Pyx_INCREF(Py_None);
    __Pyx_DECREF_SET(__pyx_v_kwargs, Py_None);
  }
  __pyx_t_1 = ((PyObject *)__Pyx_ImportNumPyArrayTypeIfAvailable()); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 334, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_ndarray = ((PyTypeObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_v_itemsize = -1L;
  if (unlikely(__pyx_v_args == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 334, __pyx_L1_error)
  }
  __pyx_t_5 = PyTuple_GET_SIZE(((PyObject*)__pyx_v_args)); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(1, 334, __pyx_L1_error)
  __pyx_t_2 = ((2 < __pyx_t_5) != 0);
  if (__pyx_t_2) {
    if (unlikely(__pyx_v_args == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 334, __pyx_L1_error)
    }
    __pyx_t_1 = PyTuple_GET_ITEM(((PyObject*)__pyx_v_args), 2);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v_arg = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L6;
  }
  __pyx_t_3 = (__pyx_v_kwargs != Py_None);
  __pyx_t_4 = (__pyx_t_3 != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L7_bool_binop_done;
  }
  if (unlikely(__pyx_v_kwargs == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(1, 334, __pyx_L1_error)
  }
  __pyx_t_4 = (__Pyx_PyDict_ContainsTF(__pyx_n_s_point, ((PyObject*)__pyx_v_kwargs), Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(1, 334, __pyx_L1_error)
  __pyx_t_3 = (__pyx_t_4 != 0);
  __pyx_t_2 = __pyx_t_3;
  __pyx_L7_bool_binop_done:;
  if (__pyx_t_2) {
    if (unlikely(__pyx_v_kwargs == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 334, __pyx_L1_error)
    }
    __pyx_t_1 = __Pyx_PyDict_GetItem(((PyObject*)__pyx_v_kwargs), __pyx_n_s_point); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 334, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_arg = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L6;
  }
  /*else*/ {
    if (unlikely(__pyx_v_args == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
      __PYX_ERR(1, 334, __pyx_L1_error)
    }
    __pyx_t_5 = PyTuple_GET_SIZE(((PyObject*)__pyx_v_args)); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(1, 334, __pyx_L1_error)
    __pyx_t_1 = PyInt_FromSsize_t(__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 334, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_6 = PyTuple_New(3); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 334, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_INCREF(__pyx_int_3);
    __Pyx_GIVEREF(__pyx_int_3);
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_int_3);
    __Pyx_INCREF(__pyx_n_s_s);
    __Pyx_GIVEREF(__pyx_n_s_s);
    PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_n_s_s);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Expected_at_least_d_argument_s_g, __pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 334, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_1); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 334, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(1, 334, __pyx_L1_error)
  }
  __pyx_L6:;
  while (1) {
    __pyx_t_2 = PyFloat_Check(__pyx_v_arg); 
    __pyx_t_3 = (__pyx_t_2 != 0);
    if (__pyx_t_3) {
      if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, __pyx_n_s_float64_t, long, 1, __Pyx_PyInt_From_long, 1, 0, 0) < 0)) __PYX_ERR(1, 334, __pyx_L1_error)
      goto __pyx_L10_break;
    }
    __pyx_t_2 = PyInt_Check(__pyx_v_arg); 
    __pyx_t_4 = (__pyx_t_2 != 0);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_3 = __pyx_t_4;
      goto __pyx_L13_bool_binop_done;
    }
    __pyx_t_4 = PyLong_Check(__pyx_v_arg); 
    __pyx_t_2 = (__pyx_t_4 != 0);
    __pyx_t_3 = __pyx_t_2;
    __pyx_L13_bool_binop_done:;
    __pyx_t_2 = (__pyx_t_3 != 0);
    if (__pyx_t_2) {
      if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, __pyx_n_s_int64_t, long, 1, __Pyx_PyInt_From_long, 1, 0, 0) < 0)) __PYX_ERR(1, 334, __pyx_L1_error)
      goto __pyx_L10_break;
    }
    if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, Py_None, long, 1, __Pyx_PyInt_From_long, 1, 0, 0) < 0)) __PYX_ERR(1, 334, __pyx_L1_error)
    goto __pyx_L10_break;
  }
  __pyx_L10_break:;
  __pyx_t_6 = PyList_New(0); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 334, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_v_candidates = ((PyObject*)__pyx_t_6);
  __pyx_t_6 = 0;
  __pyx_t_5 = 0;
  if (unlikely(((PyObject *)__pyx_v_signatures) == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(1, 334, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_dict_iterator(((PyObject*)__pyx_v_signatures), 1, ((PyObject *)NULL), (&__pyx_t_7), (&__pyx_t_8)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 334, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __pyx_t_6 = __pyx_t_1;
  __pyx_t_1 = 0;
  while (1) {
    __pyx_t_9 = __Pyx_dict_iter_next(__pyx_t_6, __pyx_t_7, &__pyx_t_5, &__pyx_t_1, NULL, NULL, __pyx_t_8);
    if (unlikely(__pyx_t_9 == 0)) break;
    if (unlikely(__pyx_t_9 == -1)) __PYX_ERR(1, 334, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_sig, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_v_match_found = 0;
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_v_sig, __pyx_n_s_strip); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 334, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __pyx_t_12 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_11))) {
      __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_11);
      if (likely(__pyx_t_12)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
        __Pyx_INCREF(__pyx_t_12);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_11, function);
      }
    }
    __pyx_t_10 = (__pyx_t_12) ? __Pyx_PyObject_Call2Args(__pyx_t_11, __pyx_t_12, __pyx_kp_s__13) : __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_kp_s__13);
    __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
    if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 334, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_split); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 334, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_t_10 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_11))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_11);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_11, function);
      }
    }
    __pyx_t_1 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_11, __pyx_t_10, __pyx_kp_s__14) : __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_kp_s__14);
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 334, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_XDECREF_SET(__pyx_v_src_sig, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_13 = PyList_GET_SIZE(__pyx_v_dest_sig); if (unlikely(__pyx_t_13 == ((Py_ssize_t)-1))) __PYX_ERR(1, 334, __pyx_L1_error)
    __pyx_t_14 = __pyx_t_13;
    for (__pyx_t_15 = 0; __pyx_t_15 < __pyx_t_14; __pyx_t_15+=1) {
      __pyx_v_i = __pyx_t_15;
      __pyx_t_1 = PyList_GET_ITEM(__pyx_v_dest_sig, __pyx_v_i);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_XDECREF_SET(__pyx_v_dst_type, __pyx_t_1);
      __pyx_t_1 = 0;
      __pyx_t_2 = (__pyx_v_dst_type != Py_None);
      __pyx_t_3 = (__pyx_t_2 != 0);
      if (__pyx_t_3) {
        __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_src_sig, __pyx_v_i, Py_ssize_t, 1, PyInt_FromSsize_t, 0, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 334, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_11 = PyObject_RichCompare(__pyx_t_1, __pyx_v_dst_type, Py_EQ); __Pyx_XGOTREF(__pyx_t_11); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 334, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_11); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(1, 334, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        if (__pyx_t_3) {
          __pyx_v_match_found = 1;
          goto __pyx_L20;
        }
        /*else*/ {
          __pyx_v_match_found = 0;
          goto __pyx_L18_break;
        }
        __pyx_L20:;
      }
    }
    __pyx_L18_break:;
    __pyx_t_3 = (__pyx_v_match_found != 0);
    if (__pyx_t_3) {
      __pyx_t_16 = __Pyx_PyList_Append(__pyx_v_candidates, __pyx_v_sig); if (unlikely(__pyx_t_16 == ((int)-1))) __PYX_ERR(1, 334, __pyx_L1_error)
    }
  }
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_3 = (PyList_GET_SIZE(__pyx_v_candidates) != 0);
  __pyx_t_2 = ((!__pyx_t_3) != 0);
  if (__pyx_t_2) {
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__15, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 334, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(1, 334, __pyx_L1_error)
  }
  __pyx_t_7 = PyList_GET_SIZE(__pyx_v_candidates); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1))) __PYX_ERR(1, 334, __pyx_L1_error)
  __pyx_t_2 = ((__pyx_t_7 > 1) != 0);
  if (__pyx_t_2) {
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__16, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 334, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(1, 334, __pyx_L1_error)
  }
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    if (unlikely(((PyObject *)__pyx_v_signatures) == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 334, __pyx_L1_error)
    }
    __pyx_t_6 = __Pyx_PyDict_GetItem(((PyObject*)__pyx_v_signatures), PyList_GET_ITEM(__pyx_v_candidates, 0)); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 334, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_r = __pyx_t_6;
    __pyx_t_6 = 0;
    goto __pyx_L0;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("pandas._libs.interval.Float32ClosedLeftIntervalNode.__pyx_fused_cpdef", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_dest_sig);
  __Pyx_XDECREF(__pyx_v_ndarray);
  __Pyx_XDECREF(__pyx_v_arg);
  __Pyx_XDECREF(__pyx_v_candidates);
  __Pyx_XDECREF(__pyx_v_sig);
  __Pyx_XDECREF(__pyx_v_src_sig);
  __Pyx_XDECREF(__pyx_v_dst_type);
  __Pyx_XDECREF(__pyx_v_kwargs);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_9__pyx_fuse_0query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_fuse_0__pyx_f_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float64_t __pyx_v_point, int __pyx_skip_dispatch) {
  __Pyx_memviewslice __pyx_v_indices = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_values = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __pyx_t_5numpy_int64_t __pyx_t_9;
  __pyx_t_5numpy_int64_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  int __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  __Pyx_memviewslice __pyx_t_16 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_17 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  Py_ssize_t __pyx_t_21;
  __Pyx_RefNannySetupContext("__pyx_fuse_0query", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pyx_fuse_0query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 334, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_9__pyx_fuse_0query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = PyFloat_FromDouble(__pyx_v_point); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 334, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 334, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 334, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 334, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_5) {
            __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 334, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 334, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_3query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_4 = PyFloat_FromDouble(__pyx_v_point); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 334, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_7 = __pyx_t_1; __pyx_t_3 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 334, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 334, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 334, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (__pyx_t_3) {
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_t_4);
          __pyx_t_4 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 334, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pandas/_libs/intervaltree.pxi":343
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
  __pyx_t_8 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_8) {

    /* "pandas/_libs/intervaltree.pxi":347
 *             # continue the binary tree structure. Instead, we use linear
 *             # search.
 *             for i in range(self.n_elements):             # <<<<<<<<<<<<<<
 *                 if self.left[i] <= point < self.right[i]:
 *                     result.append(self.indices[i])
 */
    __pyx_t_9 = __pyx_v_self->n_elements;
    __pyx_t_10 = __pyx_t_9;
    for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
      __pyx_v_i = __pyx_t_11;

      /* "pandas/_libs/intervaltree.pxi":348
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point < self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      __pyx_t_12 = __pyx_v_i;
      __pyx_t_8 = ((*((__pyx_t_5numpy_float32_t *) ( /* dim=0 */ (__pyx_v_self->left.data + __pyx_t_12 * __pyx_v_self->left.strides[0]) ))) <= __pyx_v_point);
      if (__pyx_t_8) {
        __pyx_t_13 = __pyx_v_i;
        __pyx_t_8 = (__pyx_v_point < (*((__pyx_t_5numpy_float32_t *) ( /* dim=0 */ (__pyx_v_self->right.data + __pyx_t_13 * __pyx_v_self->right.strides[0]) ))));
      }
      __pyx_t_14 = (__pyx_t_8 != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":349
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point < self.right[i]:
 *                     result.append(self.indices[i])             # <<<<<<<<<<<<<<
 *         else:
 *             # There are child nodes. Based on comparing our query to the pivot,
 */
        __pyx_t_15 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->indices.data + __pyx_t_15 * __pyx_v_self->indices.strides[0]) ))));

        /* "pandas/_libs/intervaltree.pxi":348
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point < self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      }
    }

    /* "pandas/_libs/intervaltree.pxi":343
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
    goto __pyx_L3;
  }

  /* "pandas/_libs/intervaltree.pxi":353
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
  /*else*/ {
    __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":354
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:
 *                 values = self.center_left_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 */
      __pyx_t_16 = __pyx_v_self->center_left_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":355
 *             if point < self.pivot:
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:
 */
      __pyx_t_17 = __pyx_v_self->center_left_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":356
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):             # <<<<<<<<<<<<<<
 *                     if not values[i] <= point:
 *                         break
 */
      __pyx_t_9 = __pyx_v_self->n_center;
      __pyx_t_10 = __pyx_t_9;
      for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":357
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_18 = __pyx_v_i;
        __pyx_t_14 = ((!(((*((__pyx_t_5numpy_float32_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_18 * __pyx_v_values.strides[0]) ))) <= __pyx_v_point) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":358
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:
 */
          goto __pyx_L9_break;

          /* "pandas/_libs/intervaltree.pxi":357
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":359
 *                     if not values[i] <= point:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 */
        __pyx_t_19 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_19 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L9_break:;

      /* "pandas/_libs/intervaltree.pxi":360
 *                         break
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->left_node->max_right) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":361
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)             # <<<<<<<<<<<<<<
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode *)__pyx_v_self->left_node->__pyx_vtab)->__pyx_fuse_0query(__pyx_v_self->left_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 361, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":360
 *                         break
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":353
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":362
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
    __pyx_t_14 = ((__pyx_v_point > __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":363
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 *                 values = self.center_right_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 */
      __pyx_t_16 = __pyx_v_self->center_right_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":364
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:
 */
      __pyx_t_17 = __pyx_v_self->center_right_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":365
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):             # <<<<<<<<<<<<<<
 *                     if not point < values[i]:
 *                         break
 */
      for (__pyx_t_11 = (__pyx_v_self->n_center - 1); __pyx_t_11 > -1L; __pyx_t_11-=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":366
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_20 = __pyx_v_i;
        __pyx_t_14 = ((!((__pyx_v_point < (*((__pyx_t_5numpy_float32_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_20 * __pyx_v_values.strides[0]) )))) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":367
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:
 */
          goto __pyx_L13_break;

          /* "pandas/_libs/intervaltree.pxi":366
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":368
 *                     if not point < values[i]:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if self.right_node.min_left <= point:
 *                     self.right_node.query(result, point)
 */
        __pyx_t_21 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_21 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L13_break:;

      /* "pandas/_libs/intervaltree.pxi":369
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      __pyx_t_14 = ((__pyx_v_self->right_node->min_left <= __pyx_v_point) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":370
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:
 *                     self.right_node.query(result, point)             # <<<<<<<<<<<<<<
 *             else:
 *                 result.extend(self.center_left_indices)
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode *)__pyx_v_self->right_node->__pyx_vtab)->__pyx_fuse_0query(__pyx_v_self->right_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 370, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":369
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":362
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":372
 *                     self.right_node.query(result, point)
 *             else:
 *                 result.extend(self.center_left_indices)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
    /*else*/ {
      __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->extend(__pyx_v_result, __pyx_v_self->center_left_indices); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 372, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_L7:;
  }
  __pyx_L3:;

  /* "pandas/_libs/intervaltree.pxi":334
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __PYX_XDEC_MEMVIEW(&__pyx_t_16, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_17, 1);
  __Pyx_AddTraceback("pandas._libs.interval.Float32ClosedLeftIntervalNode.query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_indices, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_values, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_9__pyx_fuse_0query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_0__pyx_mdef_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_9__pyx_fuse_0query = {"__pyx_fuse_0query", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_9__pyx_fuse_0query, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_2query};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_9__pyx_fuse_0query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result = 0;
  __pyx_t_5numpy_float64_t __pyx_v_point;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fuse_0query (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_result,&__pyx_n_s_point,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_result)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_point)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_0query", 1, 2, 2, 1); __PYX_ERR(1, 334, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_0query") < 0)) __PYX_ERR(1, 334, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_result = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)values[0]);
    __pyx_v_point = __pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_point == ((npy_float64)-1)) && PyErr_Occurred())) __PYX_ERR(1, 334, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_0query", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 334, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Float32ClosedLeftIntervalNode.__pyx_fuse_0query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_result), __pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector, 1, "result", 0))) __PYX_ERR(1, 334, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_8__pyx_fuse_0query(((struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode *)__pyx_v_self), __pyx_v_result, __pyx_v_point);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_8__pyx_fuse_0query(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float64_t __pyx_v_point) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__pyx_fuse_0query", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_fuse_0__pyx_f_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_query(__pyx_v_self, __pyx_v_result, __pyx_v_point, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 334, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Float32ClosedLeftIntervalNode.__pyx_fuse_0query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_11__pyx_fuse_1query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_fuse_1__pyx_f_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float32_t __pyx_v_point, int __pyx_skip_dispatch) {
  __Pyx_memviewslice __pyx_v_indices = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_values = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __pyx_t_5numpy_int64_t __pyx_t_9;
  __pyx_t_5numpy_int64_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  int __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  __Pyx_memviewslice __pyx_t_16 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_17 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  Py_ssize_t __pyx_t_21;
  __Pyx_RefNannySetupContext("__pyx_fuse_1query", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pyx_fuse_1query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 334, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_11__pyx_fuse_1query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = PyFloat_FromDouble(__pyx_v_point); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 334, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 334, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 334, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 334, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_5) {
            __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 334, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 334, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_3query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_4 = PyFloat_FromDouble(__pyx_v_point); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 334, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_7 = __pyx_t_1; __pyx_t_3 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 334, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 334, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 334, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (__pyx_t_3) {
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_t_4);
          __pyx_t_4 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 334, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pandas/_libs/intervaltree.pxi":343
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
  __pyx_t_8 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_8) {

    /* "pandas/_libs/intervaltree.pxi":347
 *             # continue the binary tree structure. Instead, we use linear
 *             # search.
 *             for i in range(self.n_elements):             # <<<<<<<<<<<<<<
 *                 if self.left[i] <= point < self.right[i]:
 *                     result.append(self.indices[i])
 */
    __pyx_t_9 = __pyx_v_self->n_elements;
    __pyx_t_10 = __pyx_t_9;
    for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
      __pyx_v_i = __pyx_t_11;

      /* "pandas/_libs/intervaltree.pxi":348
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point < self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      __pyx_t_12 = __pyx_v_i;
      __pyx_t_8 = ((*((__pyx_t_5numpy_float32_t *) ( /* dim=0 */ (__pyx_v_self->left.data + __pyx_t_12 * __pyx_v_self->left.strides[0]) ))) <= __pyx_v_point);
      if (__pyx_t_8) {
        __pyx_t_13 = __pyx_v_i;
        __pyx_t_8 = (__pyx_v_point < (*((__pyx_t_5numpy_float32_t *) ( /* dim=0 */ (__pyx_v_self->right.data + __pyx_t_13 * __pyx_v_self->right.strides[0]) ))));
      }
      __pyx_t_14 = (__pyx_t_8 != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":349
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point < self.right[i]:
 *                     result.append(self.indices[i])             # <<<<<<<<<<<<<<
 *         else:
 *             # There are child nodes. Based on comparing our query to the pivot,
 */
        __pyx_t_15 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->indices.data + __pyx_t_15 * __pyx_v_self->indices.strides[0]) ))));

        /* "pandas/_libs/intervaltree.pxi":348
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point < self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      }
    }

    /* "pandas/_libs/intervaltree.pxi":343
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
    goto __pyx_L3;
  }

  /* "pandas/_libs/intervaltree.pxi":353
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
  /*else*/ {
    __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":354
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:
 *                 values = self.center_left_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 */
      __pyx_t_16 = __pyx_v_self->center_left_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":355
 *             if point < self.pivot:
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:
 */
      __pyx_t_17 = __pyx_v_self->center_left_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":356
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):             # <<<<<<<<<<<<<<
 *                     if not values[i] <= point:
 *                         break
 */
      __pyx_t_9 = __pyx_v_self->n_center;
      __pyx_t_10 = __pyx_t_9;
      for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":357
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_18 = __pyx_v_i;
        __pyx_t_14 = ((!(((*((__pyx_t_5numpy_float32_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_18 * __pyx_v_values.strides[0]) ))) <= __pyx_v_point) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":358
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:
 */
          goto __pyx_L9_break;

          /* "pandas/_libs/intervaltree.pxi":357
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":359
 *                     if not values[i] <= point:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 */
        __pyx_t_19 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_19 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L9_break:;

      /* "pandas/_libs/intervaltree.pxi":360
 *                         break
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->left_node->max_right) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":361
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)             # <<<<<<<<<<<<<<
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode *)__pyx_v_self->left_node->__pyx_vtab)->__pyx_fuse_1query(__pyx_v_self->left_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 361, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":360
 *                         break
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":353
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":362
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
    __pyx_t_14 = ((__pyx_v_point > __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":363
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 *                 values = self.center_right_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 */
      __pyx_t_16 = __pyx_v_self->center_right_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":364
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:
 */
      __pyx_t_17 = __pyx_v_self->center_right_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":365
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):             # <<<<<<<<<<<<<<
 *                     if not point < values[i]:
 *                         break
 */
      for (__pyx_t_11 = (__pyx_v_self->n_center - 1); __pyx_t_11 > -1L; __pyx_t_11-=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":366
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_20 = __pyx_v_i;
        __pyx_t_14 = ((!((__pyx_v_point < (*((__pyx_t_5numpy_float32_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_20 * __pyx_v_values.strides[0]) )))) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":367
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:
 */
          goto __pyx_L13_break;

          /* "pandas/_libs/intervaltree.pxi":366
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":368
 *                     if not point < values[i]:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if self.right_node.min_left <= point:
 *                     self.right_node.query(result, point)
 */
        __pyx_t_21 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_21 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L13_break:;

      /* "pandas/_libs/intervaltree.pxi":369
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      __pyx_t_14 = ((__pyx_v_self->right_node->min_left <= __pyx_v_point) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":370
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:
 *                     self.right_node.query(result, point)             # <<<<<<<<<<<<<<
 *             else:
 *                 result.extend(self.center_left_indices)
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode *)__pyx_v_self->right_node->__pyx_vtab)->__pyx_fuse_1query(__pyx_v_self->right_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 370, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":369
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":362
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":372
 *                     self.right_node.query(result, point)
 *             else:
 *                 result.extend(self.center_left_indices)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
    /*else*/ {
      __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->extend(__pyx_v_result, __pyx_v_self->center_left_indices); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 372, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_L7:;
  }
  __pyx_L3:;

  /* "pandas/_libs/intervaltree.pxi":334
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __PYX_XDEC_MEMVIEW(&__pyx_t_16, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_17, 1);
  __Pyx_AddTraceback("pandas._libs.interval.Float32ClosedLeftIntervalNode.query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_indices, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_values, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_11__pyx_fuse_1query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_1__pyx_mdef_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_11__pyx_fuse_1query = {"__pyx_fuse_1query", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_11__pyx_fuse_1query, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_2query};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_11__pyx_fuse_1query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result = 0;
  __pyx_t_5numpy_float32_t __pyx_v_point;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fuse_1query (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_result,&__pyx_n_s_point,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_result)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_point)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_1query", 1, 2, 2, 1); __PYX_ERR(1, 334, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_1query") < 0)) __PYX_ERR(1, 334, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_result = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)values[0]);
    __pyx_v_point = __pyx_PyFloat_AsFloat(values[1]); if (unlikely((__pyx_v_point == ((npy_float32)-1)) && PyErr_Occurred())) __PYX_ERR(1, 334, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_1query", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 334, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Float32ClosedLeftIntervalNode.__pyx_fuse_1query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_result), __pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector, 1, "result", 0))) __PYX_ERR(1, 334, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_10__pyx_fuse_1query(((struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode *)__pyx_v_self), __pyx_v_result, __pyx_v_point);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_10__pyx_fuse_1query(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float32_t __pyx_v_point) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__pyx_fuse_1query", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_fuse_1__pyx_f_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_query(__pyx_v_self, __pyx_v_result, __pyx_v_point, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 334, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Float32ClosedLeftIntervalNode.__pyx_fuse_1query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_13__pyx_fuse_2query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_fuse_2__pyx_f_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int64_t __pyx_v_point, int __pyx_skip_dispatch) {
  __Pyx_memviewslice __pyx_v_indices = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_values = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __pyx_t_5numpy_int64_t __pyx_t_9;
  __pyx_t_5numpy_int64_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  int __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  __Pyx_memviewslice __pyx_t_16 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_17 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  Py_ssize_t __pyx_t_21;
  __Pyx_RefNannySetupContext("__pyx_fuse_2query", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pyx_fuse_2query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 334, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_13__pyx_fuse_2query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = __Pyx_PyInt_From_npy_int64(__pyx_v_point); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 334, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 334, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 334, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 334, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_5) {
            __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 334, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 334, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_3query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_4 = __Pyx_PyInt_From_npy_int64(__pyx_v_point); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 334, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_7 = __pyx_t_1; __pyx_t_3 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 334, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 334, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 334, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (__pyx_t_3) {
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_t_4);
          __pyx_t_4 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 334, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pandas/_libs/intervaltree.pxi":343
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
  __pyx_t_8 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_8) {

    /* "pandas/_libs/intervaltree.pxi":347
 *             # continue the binary tree structure. Instead, we use linear
 *             # search.
 *             for i in range(self.n_elements):             # <<<<<<<<<<<<<<
 *                 if self.left[i] <= point < self.right[i]:
 *                     result.append(self.indices[i])
 */
    __pyx_t_9 = __pyx_v_self->n_elements;
    __pyx_t_10 = __pyx_t_9;
    for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
      __pyx_v_i = __pyx_t_11;

      /* "pandas/_libs/intervaltree.pxi":348
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point < self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      __pyx_t_12 = __pyx_v_i;
      __pyx_t_8 = ((*((__pyx_t_5numpy_float32_t *) ( /* dim=0 */ (__pyx_v_self->left.data + __pyx_t_12 * __pyx_v_self->left.strides[0]) ))) <= __pyx_v_point);
      if (__pyx_t_8) {
        __pyx_t_13 = __pyx_v_i;
        __pyx_t_8 = (__pyx_v_point < (*((__pyx_t_5numpy_float32_t *) ( /* dim=0 */ (__pyx_v_self->right.data + __pyx_t_13 * __pyx_v_self->right.strides[0]) ))));
      }
      __pyx_t_14 = (__pyx_t_8 != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":349
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point < self.right[i]:
 *                     result.append(self.indices[i])             # <<<<<<<<<<<<<<
 *         else:
 *             # There are child nodes. Based on comparing our query to the pivot,
 */
        __pyx_t_15 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->indices.data + __pyx_t_15 * __pyx_v_self->indices.strides[0]) ))));

        /* "pandas/_libs/intervaltree.pxi":348
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point < self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      }
    }

    /* "pandas/_libs/intervaltree.pxi":343
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
    goto __pyx_L3;
  }

  /* "pandas/_libs/intervaltree.pxi":353
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
  /*else*/ {
    __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":354
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:
 *                 values = self.center_left_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 */
      __pyx_t_16 = __pyx_v_self->center_left_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":355
 *             if point < self.pivot:
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:
 */
      __pyx_t_17 = __pyx_v_self->center_left_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":356
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):             # <<<<<<<<<<<<<<
 *                     if not values[i] <= point:
 *                         break
 */
      __pyx_t_9 = __pyx_v_self->n_center;
      __pyx_t_10 = __pyx_t_9;
      for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":357
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_18 = __pyx_v_i;
        __pyx_t_14 = ((!(((*((__pyx_t_5numpy_float32_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_18 * __pyx_v_values.strides[0]) ))) <= __pyx_v_point) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":358
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:
 */
          goto __pyx_L9_break;

          /* "pandas/_libs/intervaltree.pxi":357
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":359
 *                     if not values[i] <= point:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 */
        __pyx_t_19 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_19 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L9_break:;

      /* "pandas/_libs/intervaltree.pxi":360
 *                         break
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->left_node->max_right) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":361
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)             # <<<<<<<<<<<<<<
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode *)__pyx_v_self->left_node->__pyx_vtab)->__pyx_fuse_2query(__pyx_v_self->left_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 361, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":360
 *                         break
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":353
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":362
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
    __pyx_t_14 = ((__pyx_v_point > __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":363
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 *                 values = self.center_right_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 */
      __pyx_t_16 = __pyx_v_self->center_right_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":364
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:
 */
      __pyx_t_17 = __pyx_v_self->center_right_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":365
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):             # <<<<<<<<<<<<<<
 *                     if not point < values[i]:
 *                         break
 */
      for (__pyx_t_11 = (__pyx_v_self->n_center - 1); __pyx_t_11 > -1L; __pyx_t_11-=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":366
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_20 = __pyx_v_i;
        __pyx_t_14 = ((!((__pyx_v_point < (*((__pyx_t_5numpy_float32_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_20 * __pyx_v_values.strides[0]) )))) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":367
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:
 */
          goto __pyx_L13_break;

          /* "pandas/_libs/intervaltree.pxi":366
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":368
 *                     if not point < values[i]:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if self.right_node.min_left <= point:
 *                     self.right_node.query(result, point)
 */
        __pyx_t_21 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_21 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L13_break:;

      /* "pandas/_libs/intervaltree.pxi":369
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      __pyx_t_14 = ((__pyx_v_self->right_node->min_left <= __pyx_v_point) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":370
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:
 *                     self.right_node.query(result, point)             # <<<<<<<<<<<<<<
 *             else:
 *                 result.extend(self.center_left_indices)
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode *)__pyx_v_self->right_node->__pyx_vtab)->__pyx_fuse_2query(__pyx_v_self->right_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 370, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":369
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":362
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":372
 *                     self.right_node.query(result, point)
 *             else:
 *                 result.extend(self.center_left_indices)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
    /*else*/ {
      __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->extend(__pyx_v_result, __pyx_v_self->center_left_indices); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 372, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_L7:;
  }
  __pyx_L3:;

  /* "pandas/_libs/intervaltree.pxi":334
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __PYX_XDEC_MEMVIEW(&__pyx_t_16, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_17, 1);
  __Pyx_AddTraceback("pandas._libs.interval.Float32ClosedLeftIntervalNode.query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_indices, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_values, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_13__pyx_fuse_2query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_2__pyx_mdef_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_13__pyx_fuse_2query = {"__pyx_fuse_2query", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_13__pyx_fuse_2query, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_2query};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_13__pyx_fuse_2query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result = 0;
  __pyx_t_5numpy_int64_t __pyx_v_point;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fuse_2query (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_result,&__pyx_n_s_point,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_result)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_point)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_2query", 1, 2, 2, 1); __PYX_ERR(1, 334, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_2query") < 0)) __PYX_ERR(1, 334, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_result = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)values[0]);
    __pyx_v_point = __Pyx_PyInt_As_npy_int64(values[1]); if (unlikely((__pyx_v_point == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(1, 334, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_2query", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 334, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Float32ClosedLeftIntervalNode.__pyx_fuse_2query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_result), __pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector, 1, "result", 0))) __PYX_ERR(1, 334, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_12__pyx_fuse_2query(((struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode *)__pyx_v_self), __pyx_v_result, __pyx_v_point);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_12__pyx_fuse_2query(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int64_t __pyx_v_point) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__pyx_fuse_2query", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_fuse_2__pyx_f_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_query(__pyx_v_self, __pyx_v_result, __pyx_v_point, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 334, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Float32ClosedLeftIntervalNode.__pyx_fuse_2query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_15__pyx_fuse_3query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_fuse_3__pyx_f_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int32_t __pyx_v_point, int __pyx_skip_dispatch) {
  __Pyx_memviewslice __pyx_v_indices = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_values = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __pyx_t_5numpy_int64_t __pyx_t_9;
  __pyx_t_5numpy_int64_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  int __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  __Pyx_memviewslice __pyx_t_16 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_17 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  Py_ssize_t __pyx_t_21;
  __Pyx_RefNannySetupContext("__pyx_fuse_3query", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pyx_fuse_3query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 334, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_15__pyx_fuse_3query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = __Pyx_PyInt_From_npy_int32(__pyx_v_point); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 334, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 334, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 334, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 334, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_5) {
            __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 334, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 334, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_3query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_4 = __Pyx_PyInt_From_npy_int32(__pyx_v_point); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 334, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_7 = __pyx_t_1; __pyx_t_3 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 334, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 334, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 334, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (__pyx_t_3) {
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_t_4);
          __pyx_t_4 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 334, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pandas/_libs/intervaltree.pxi":343
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
  __pyx_t_8 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_8) {

    /* "pandas/_libs/intervaltree.pxi":347
 *             # continue the binary tree structure. Instead, we use linear
 *             # search.
 *             for i in range(self.n_elements):             # <<<<<<<<<<<<<<
 *                 if self.left[i] <= point < self.right[i]:
 *                     result.append(self.indices[i])
 */
    __pyx_t_9 = __pyx_v_self->n_elements;
    __pyx_t_10 = __pyx_t_9;
    for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
      __pyx_v_i = __pyx_t_11;

      /* "pandas/_libs/intervaltree.pxi":348
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point < self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      __pyx_t_12 = __pyx_v_i;
      __pyx_t_8 = ((*((__pyx_t_5numpy_float32_t *) ( /* dim=0 */ (__pyx_v_self->left.data + __pyx_t_12 * __pyx_v_self->left.strides[0]) ))) <= __pyx_v_point);
      if (__pyx_t_8) {
        __pyx_t_13 = __pyx_v_i;
        __pyx_t_8 = (__pyx_v_point < (*((__pyx_t_5numpy_float32_t *) ( /* dim=0 */ (__pyx_v_self->right.data + __pyx_t_13 * __pyx_v_self->right.strides[0]) ))));
      }
      __pyx_t_14 = (__pyx_t_8 != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":349
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point < self.right[i]:
 *                     result.append(self.indices[i])             # <<<<<<<<<<<<<<
 *         else:
 *             # There are child nodes. Based on comparing our query to the pivot,
 */
        __pyx_t_15 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->indices.data + __pyx_t_15 * __pyx_v_self->indices.strides[0]) ))));

        /* "pandas/_libs/intervaltree.pxi":348
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point < self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      }
    }

    /* "pandas/_libs/intervaltree.pxi":343
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
    goto __pyx_L3;
  }

  /* "pandas/_libs/intervaltree.pxi":353
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
  /*else*/ {
    __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":354
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:
 *                 values = self.center_left_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 */
      __pyx_t_16 = __pyx_v_self->center_left_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":355
 *             if point < self.pivot:
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:
 */
      __pyx_t_17 = __pyx_v_self->center_left_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":356
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):             # <<<<<<<<<<<<<<
 *                     if not values[i] <= point:
 *                         break
 */
      __pyx_t_9 = __pyx_v_self->n_center;
      __pyx_t_10 = __pyx_t_9;
      for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":357
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_18 = __pyx_v_i;
        __pyx_t_14 = ((!(((*((__pyx_t_5numpy_float32_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_18 * __pyx_v_values.strides[0]) ))) <= __pyx_v_point) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":358
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:
 */
          goto __pyx_L9_break;

          /* "pandas/_libs/intervaltree.pxi":357
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":359
 *                     if not values[i] <= point:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 */
        __pyx_t_19 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_19 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L9_break:;

      /* "pandas/_libs/intervaltree.pxi":360
 *                         break
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->left_node->max_right) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":361
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)             # <<<<<<<<<<<<<<
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode *)__pyx_v_self->left_node->__pyx_vtab)->__pyx_fuse_3query(__pyx_v_self->left_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 361, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":360
 *                         break
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":353
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":362
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
    __pyx_t_14 = ((__pyx_v_point > __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":363
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 *                 values = self.center_right_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 */
      __pyx_t_16 = __pyx_v_self->center_right_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":364
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:
 */
      __pyx_t_17 = __pyx_v_self->center_right_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":365
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):             # <<<<<<<<<<<<<<
 *                     if not point < values[i]:
 *                         break
 */
      for (__pyx_t_11 = (__pyx_v_self->n_center - 1); __pyx_t_11 > -1L; __pyx_t_11-=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":366
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_20 = __pyx_v_i;
        __pyx_t_14 = ((!((__pyx_v_point < (*((__pyx_t_5numpy_float32_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_20 * __pyx_v_values.strides[0]) )))) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":367
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:
 */
          goto __pyx_L13_break;

          /* "pandas/_libs/intervaltree.pxi":366
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":368
 *                     if not point < values[i]:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if self.right_node.min_left <= point:
 *                     self.right_node.query(result, point)
 */
        __pyx_t_21 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_21 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L13_break:;

      /* "pandas/_libs/intervaltree.pxi":369
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      __pyx_t_14 = ((__pyx_v_self->right_node->min_left <= __pyx_v_point) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":370
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:
 *                     self.right_node.query(result, point)             # <<<<<<<<<<<<<<
 *             else:
 *                 result.extend(self.center_left_indices)
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode *)__pyx_v_self->right_node->__pyx_vtab)->__pyx_fuse_3query(__pyx_v_self->right_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 370, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":369
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":362
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":372
 *                     self.right_node.query(result, point)
 *             else:
 *                 result.extend(self.center_left_indices)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
    /*else*/ {
      __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->extend(__pyx_v_result, __pyx_v_self->center_left_indices); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 372, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_L7:;
  }
  __pyx_L3:;

  /* "pandas/_libs/intervaltree.pxi":334
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __PYX_XDEC_MEMVIEW(&__pyx_t_16, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_17, 1);
  __Pyx_AddTraceback("pandas._libs.interval.Float32ClosedLeftIntervalNode.query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_indices, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_values, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_15__pyx_fuse_3query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_3__pyx_mdef_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_15__pyx_fuse_3query = {"__pyx_fuse_3query", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_15__pyx_fuse_3query, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_2query};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_15__pyx_fuse_3query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result = 0;
  __pyx_t_5numpy_int32_t __pyx_v_point;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fuse_3query (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_result,&__pyx_n_s_point,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_result)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_point)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_3query", 1, 2, 2, 1); __PYX_ERR(1, 334, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_3query") < 0)) __PYX_ERR(1, 334, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_result = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)values[0]);
    __pyx_v_point = __Pyx_PyInt_As_npy_int32(values[1]); if (unlikely((__pyx_v_point == ((npy_int32)-1)) && PyErr_Occurred())) __PYX_ERR(1, 334, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_3query", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 334, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Float32ClosedLeftIntervalNode.__pyx_fuse_3query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_result), __pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector, 1, "result", 0))) __PYX_ERR(1, 334, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_14__pyx_fuse_3query(((struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode *)__pyx_v_self), __pyx_v_result, __pyx_v_point);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_14__pyx_fuse_3query(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int32_t __pyx_v_point) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__pyx_fuse_3query", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_fuse_3__pyx_f_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_query(__pyx_v_self, __pyx_v_result, __pyx_v_point, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 334, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Float32ClosedLeftIntervalNode.__pyx_fuse_3query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_17__pyx_fuse_4query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_fuse_4__pyx_f_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_uint64_t __pyx_v_point, int __pyx_skip_dispatch) {
  __Pyx_memviewslice __pyx_v_indices = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_values = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __pyx_t_5numpy_int64_t __pyx_t_9;
  __pyx_t_5numpy_int64_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  int __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  __Pyx_memviewslice __pyx_t_16 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_17 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  Py_ssize_t __pyx_t_21;
  __Pyx_RefNannySetupContext("__pyx_fuse_4query", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pyx_fuse_4query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 334, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_17__pyx_fuse_4query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = __Pyx_PyInt_From_npy_uint64(__pyx_v_point); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 334, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 334, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 334, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 334, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_5) {
            __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 334, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 334, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_3query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_4 = __Pyx_PyInt_From_npy_uint64(__pyx_v_point); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 334, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_7 = __pyx_t_1; __pyx_t_3 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 334, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 334, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 334, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (__pyx_t_3) {
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_t_4);
          __pyx_t_4 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 334, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pandas/_libs/intervaltree.pxi":343
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
  __pyx_t_8 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_8) {

    /* "pandas/_libs/intervaltree.pxi":347
 *             # continue the binary tree structure. Instead, we use linear
 *             # search.
 *             for i in range(self.n_elements):             # <<<<<<<<<<<<<<
 *                 if self.left[i] <= point < self.right[i]:
 *                     result.append(self.indices[i])
 */
    __pyx_t_9 = __pyx_v_self->n_elements;
    __pyx_t_10 = __pyx_t_9;
    for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
      __pyx_v_i = __pyx_t_11;

      /* "pandas/_libs/intervaltree.pxi":348
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point < self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      __pyx_t_12 = __pyx_v_i;
      __pyx_t_8 = ((*((__pyx_t_5numpy_float32_t *) ( /* dim=0 */ (__pyx_v_self->left.data + __pyx_t_12 * __pyx_v_self->left.strides[0]) ))) <= __pyx_v_point);
      if (__pyx_t_8) {
        __pyx_t_13 = __pyx_v_i;
        __pyx_t_8 = (__pyx_v_point < (*((__pyx_t_5numpy_float32_t *) ( /* dim=0 */ (__pyx_v_self->right.data + __pyx_t_13 * __pyx_v_self->right.strides[0]) ))));
      }
      __pyx_t_14 = (__pyx_t_8 != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":349
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point < self.right[i]:
 *                     result.append(self.indices[i])             # <<<<<<<<<<<<<<
 *         else:
 *             # There are child nodes. Based on comparing our query to the pivot,
 */
        __pyx_t_15 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->indices.data + __pyx_t_15 * __pyx_v_self->indices.strides[0]) ))));

        /* "pandas/_libs/intervaltree.pxi":348
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point < self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      }
    }

    /* "pandas/_libs/intervaltree.pxi":343
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
    goto __pyx_L3;
  }

  /* "pandas/_libs/intervaltree.pxi":353
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
  /*else*/ {
    __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":354
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:
 *                 values = self.center_left_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 */
      __pyx_t_16 = __pyx_v_self->center_left_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":355
 *             if point < self.pivot:
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:
 */
      __pyx_t_17 = __pyx_v_self->center_left_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":356
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):             # <<<<<<<<<<<<<<
 *                     if not values[i] <= point:
 *                         break
 */
      __pyx_t_9 = __pyx_v_self->n_center;
      __pyx_t_10 = __pyx_t_9;
      for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":357
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_18 = __pyx_v_i;
        __pyx_t_14 = ((!(((*((__pyx_t_5numpy_float32_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_18 * __pyx_v_values.strides[0]) ))) <= __pyx_v_point) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":358
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:
 */
          goto __pyx_L9_break;

          /* "pandas/_libs/intervaltree.pxi":357
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":359
 *                     if not values[i] <= point:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 */
        __pyx_t_19 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_19 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L9_break:;

      /* "pandas/_libs/intervaltree.pxi":360
 *                         break
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->left_node->max_right) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":361
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)             # <<<<<<<<<<<<<<
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode *)__pyx_v_self->left_node->__pyx_vtab)->__pyx_fuse_4query(__pyx_v_self->left_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 361, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":360
 *                         break
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":353
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":362
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
    __pyx_t_14 = ((__pyx_v_point > __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":363
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 *                 values = self.center_right_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 */
      __pyx_t_16 = __pyx_v_self->center_right_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":364
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:
 */
      __pyx_t_17 = __pyx_v_self->center_right_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":365
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):             # <<<<<<<<<<<<<<
 *                     if not point < values[i]:
 *                         break
 */
      for (__pyx_t_11 = (__pyx_v_self->n_center - 1); __pyx_t_11 > -1L; __pyx_t_11-=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":366
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_20 = __pyx_v_i;
        __pyx_t_14 = ((!((__pyx_v_point < (*((__pyx_t_5numpy_float32_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_20 * __pyx_v_values.strides[0]) )))) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":367
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:
 */
          goto __pyx_L13_break;

          /* "pandas/_libs/intervaltree.pxi":366
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":368
 *                     if not point < values[i]:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if self.right_node.min_left <= point:
 *                     self.right_node.query(result, point)
 */
        __pyx_t_21 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_21 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L13_break:;

      /* "pandas/_libs/intervaltree.pxi":369
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      __pyx_t_14 = ((__pyx_v_self->right_node->min_left <= __pyx_v_point) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":370
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:
 *                     self.right_node.query(result, point)             # <<<<<<<<<<<<<<
 *             else:
 *                 result.extend(self.center_left_indices)
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode *)__pyx_v_self->right_node->__pyx_vtab)->__pyx_fuse_4query(__pyx_v_self->right_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 370, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":369
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":362
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":372
 *                     self.right_node.query(result, point)
 *             else:
 *                 result.extend(self.center_left_indices)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
    /*else*/ {
      __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->extend(__pyx_v_result, __pyx_v_self->center_left_indices); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 372, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_L7:;
  }
  __pyx_L3:;

  /* "pandas/_libs/intervaltree.pxi":334
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __PYX_XDEC_MEMVIEW(&__pyx_t_16, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_17, 1);
  __Pyx_AddTraceback("pandas._libs.interval.Float32ClosedLeftIntervalNode.query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_indices, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_values, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_17__pyx_fuse_4query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_4__pyx_mdef_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_17__pyx_fuse_4query = {"__pyx_fuse_4query", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_17__pyx_fuse_4query, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_2query};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_17__pyx_fuse_4query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result = 0;
  __pyx_t_5numpy_uint64_t __pyx_v_point;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fuse_4query (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_result,&__pyx_n_s_point,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_result)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_point)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_4query", 1, 2, 2, 1); __PYX_ERR(1, 334, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_4query") < 0)) __PYX_ERR(1, 334, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_result = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)values[0]);
    __pyx_v_point = __Pyx_PyInt_As_npy_uint64(values[1]); if (unlikely((__pyx_v_point == ((npy_uint64)-1)) && PyErr_Occurred())) __PYX_ERR(1, 334, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_4query", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 334, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Float32ClosedLeftIntervalNode.__pyx_fuse_4query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_result), __pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector, 1, "result", 0))) __PYX_ERR(1, 334, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_16__pyx_fuse_4query(((struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode *)__pyx_v_self), __pyx_v_result, __pyx_v_point);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_16__pyx_fuse_4query(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_uint64_t __pyx_v_point) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__pyx_fuse_4query", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_fuse_4__pyx_f_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_query(__pyx_v_self, __pyx_v_result, __pyx_v_point, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 334, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Float32ClosedLeftIntervalNode.__pyx_fuse_4query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":374
 *                 result.extend(self.center_left_indices)
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         if self.is_leaf_node:
 *             return ('<Float32ClosedLeftIntervalNode: '
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_5__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_5__repr__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_4__repr__(((struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_4__repr__(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode *__pyx_v_self) {
  PyObject *__pyx_v_n_left = NULL;
  PyObject *__pyx_v_n_right = NULL;
  PyObject *__pyx_v_n_center = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  Py_UCS4 __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "pandas/_libs/intervaltree.pxi":375
 * 
 *     def __repr__(self):
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             return ('<Float32ClosedLeftIntervalNode: '
 *                     '%s elements (terminal)>' % self.n_elements)
 */
  __pyx_t_1 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_1) {

    /* "pandas/_libs/intervaltree.pxi":376
 *     def __repr__(self):
 *         if self.is_leaf_node:
 *             return ('<Float32ClosedLeftIntervalNode: '             # <<<<<<<<<<<<<<
 *                     '%s elements (terminal)>' % self.n_elements)
 *         else:
 */
    __Pyx_XDECREF(__pyx_r);

    /* "pandas/_libs/intervaltree.pxi":377
 *         if self.is_leaf_node:
 *             return ('<Float32ClosedLeftIntervalNode: '
 *                     '%s elements (terminal)>' % self.n_elements)             # <<<<<<<<<<<<<<
 *         else:
 *             n_left = self.left_node.n_elements
 */
    __pyx_t_2 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_elements); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 377, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PyUnicode_Format(__pyx_kp_u_Float32ClosedLeftIntervalNode_s, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 377, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pandas/_libs/intervaltree.pxi":375
 * 
 *     def __repr__(self):
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             return ('<Float32ClosedLeftIntervalNode: '
 *                     '%s elements (terminal)>' % self.n_elements)
 */
  }

  /* "pandas/_libs/intervaltree.pxi":379
 *                     '%s elements (terminal)>' % self.n_elements)
 *         else:
 *             n_left = self.left_node.n_elements             # <<<<<<<<<<<<<<
 *             n_right = self.right_node.n_elements
 *             n_center = self.n_elements - n_left - n_right
 */
  /*else*/ {
    __pyx_t_3 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->left_node->n_elements); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 379, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_v_n_left = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "pandas/_libs/intervaltree.pxi":380
 *         else:
 *             n_left = self.left_node.n_elements
 *             n_right = self.right_node.n_elements             # <<<<<<<<<<<<<<
 *             n_center = self.n_elements - n_left - n_right
 *             return ('<Float32ClosedLeftIntervalNode: '
 */
    __pyx_t_3 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->right_node->n_elements); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 380, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_v_n_right = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "pandas/_libs/intervaltree.pxi":381
 *             n_left = self.left_node.n_elements
 *             n_right = self.right_node.n_elements
 *             n_center = self.n_elements - n_left - n_right             # <<<<<<<<<<<<<<
 *             return ('<Float32ClosedLeftIntervalNode: '
 *                     'pivot %s, %s elements (%s left, %s right, %s '
 */
    __pyx_t_3 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_elements); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 381, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = PyNumber_Subtract(__pyx_t_3, __pyx_v_n_left); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 381, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyNumber_Subtract(__pyx_t_2, __pyx_v_n_right); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 381, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_n_center = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "pandas/_libs/intervaltree.pxi":382
 *             n_right = self.right_node.n_elements
 *             n_center = self.n_elements - n_left - n_right
 *             return ('<Float32ClosedLeftIntervalNode: '             # <<<<<<<<<<<<<<
 *                     'pivot %s, %s elements (%s left, %s right, %s '
 *                     'overlapping)>' % (self.pivot, self.n_elements,
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = PyTuple_New(11); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 382, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = 0;
    __pyx_t_5 = 127;
    __Pyx_INCREF(__pyx_kp_u_Float32ClosedLeftIntervalNode_p);
    __pyx_t_4 += 38;
    __Pyx_GIVEREF(__pyx_kp_u_Float32ClosedLeftIntervalNode_p);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_kp_u_Float32ClosedLeftIntervalNode_p);

    /* "pandas/_libs/intervaltree.pxi":384
 *             return ('<Float32ClosedLeftIntervalNode: '
 *                     'pivot %s, %s elements (%s left, %s right, %s '
 *                     'overlapping)>' % (self.pivot, self.n_elements,             # <<<<<<<<<<<<<<
 *                                        n_left, n_right, n_center))
 * 
 */
    __pyx_t_2 = PyFloat_FromDouble(__pyx_v_self->pivot); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 384, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_t_2), __pyx_empty_unicode); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 384, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_5 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_6) > __pyx_t_5) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_6) : __pyx_t_5;
    __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_6);
    __pyx_t_6 = 0;
    __Pyx_INCREF(__pyx_kp_u__18);
    __pyx_t_4 += 2;
    __Pyx_GIVEREF(__pyx_kp_u__18);
    PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_kp_u__18);
    __pyx_t_6 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_elements); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 384, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_2 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_t_6), __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 384, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_5 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) > __pyx_t_5) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) : __pyx_t_5;
    __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_3, 3, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_elements);
    __pyx_t_4 += 11;
    __Pyx_GIVEREF(__pyx_kp_u_elements);
    PyTuple_SET_ITEM(__pyx_t_3, 4, __pyx_kp_u_elements);

    /* "pandas/_libs/intervaltree.pxi":385
 *                     'pivot %s, %s elements (%s left, %s right, %s '
 *                     'overlapping)>' % (self.pivot, self.n_elements,
 *                                        n_left, n_right, n_center))             # <<<<<<<<<<<<<<
 * 
 *     def counts(self):
 */
    __pyx_t_2 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_v_n_left), __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 385, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) > __pyx_t_5) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) : __pyx_t_5;
    __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_3, 5, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_left_2);
    __pyx_t_4 += 7;
    __Pyx_GIVEREF(__pyx_kp_u_left_2);
    PyTuple_SET_ITEM(__pyx_t_3, 6, __pyx_kp_u_left_2);
    __pyx_t_2 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_v_n_right), __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 385, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) > __pyx_t_5) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) : __pyx_t_5;
    __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_3, 7, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_right_2);
    __pyx_t_4 += 8;
    __Pyx_GIVEREF(__pyx_kp_u_right_2);
    PyTuple_SET_ITEM(__pyx_t_3, 8, __pyx_kp_u_right_2);
    __pyx_t_2 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_v_n_center), __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 385, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) > __pyx_t_5) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) : __pyx_t_5;
    __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_3, 9, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_overlapping);
    __pyx_t_4 += 14;
    __Pyx_GIVEREF(__pyx_kp_u_overlapping);
    PyTuple_SET_ITEM(__pyx_t_3, 10, __pyx_kp_u_overlapping);

    /* "pandas/_libs/intervaltree.pxi":382
 *             n_right = self.right_node.n_elements
 *             n_center = self.n_elements - n_left - n_right
 *             return ('<Float32ClosedLeftIntervalNode: '             # <<<<<<<<<<<<<<
 *                     'pivot %s, %s elements (%s left, %s right, %s '
 *                     'overlapping)>' % (self.pivot, self.n_elements,
 */
    __pyx_t_2 = __Pyx_PyUnicode_Join(__pyx_t_3, 11, __pyx_t_4, __pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 382, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;
  }

  /* "pandas/_libs/intervaltree.pxi":374
 *                 result.extend(self.center_left_indices)
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         if self.is_leaf_node:
 *             return ('<Float32ClosedLeftIntervalNode: '
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pandas._libs.interval.Float32ClosedLeftIntervalNode.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_n_left);
  __Pyx_XDECREF(__pyx_v_n_right);
  __Pyx_XDECREF(__pyx_v_n_center);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":387
 *                                        n_left, n_right, n_center))
 * 
 *     def counts(self):             # <<<<<<<<<<<<<<
 *         """
 *         Inspect counts on this node
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_7counts(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_6counts[] = "\n        Inspect counts on this node\n        useful for debugging purposes\n        ";
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_7counts(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("counts (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_6counts(((struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_6counts(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode *__pyx_v_self) {
  Py_ssize_t __pyx_v_m;
  PyObject *__pyx_v_l = NULL;
  PyObject *__pyx_v_r = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  size_t __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("counts", 0);

  /* "pandas/_libs/intervaltree.pxi":392
 *         useful for debugging purposes
 *         """
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             return self.n_elements
 *         else:
 */
  __pyx_t_1 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_1) {

    /* "pandas/_libs/intervaltree.pxi":393
 *         """
 *         if self.is_leaf_node:
 *             return self.n_elements             # <<<<<<<<<<<<<<
 *         else:
 *             m = len(self.center_left_values)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_elements); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 393, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "pandas/_libs/intervaltree.pxi":392
 *         useful for debugging purposes
 *         """
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             return self.n_elements
 *         else:
 */
  }

  /* "pandas/_libs/intervaltree.pxi":395
 *             return self.n_elements
 *         else:
 *             m = len(self.center_left_values)             # <<<<<<<<<<<<<<
 *             l = self.left_node.counts()
 *             r = self.right_node.counts()
 */
  /*else*/ {
    if (unlikely(!__pyx_v_self->center_left_values.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(1, 395, __pyx_L1_error)}
    __pyx_t_3 = __Pyx_MemoryView_Len(__pyx_v_self->center_left_values); 
    __pyx_v_m = __pyx_t_3;

    /* "pandas/_libs/intervaltree.pxi":396
 *         else:
 *             m = len(self.center_left_values)
 *             l = self.left_node.counts()             # <<<<<<<<<<<<<<
 *             r = self.right_node.counts()
 *             return (m, (l, r))
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->left_node), __pyx_n_s_counts); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 396, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 396, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_l = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "pandas/_libs/intervaltree.pxi":397
 *             m = len(self.center_left_values)
 *             l = self.left_node.counts()
 *             r = self.right_node.counts()             # <<<<<<<<<<<<<<
 *             return (m, (l, r))
 * 
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->right_node), __pyx_n_s_counts); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 397, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 397, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_r = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "pandas/_libs/intervaltree.pxi":398
 *             l = self.left_node.counts()
 *             r = self.right_node.counts()
 *             return (m, (l, r))             # <<<<<<<<<<<<<<
 * 
 * NODE_CLASSES['float32',
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = PyInt_FromSsize_t(__pyx_v_m); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 398, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 398, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_v_l);
    __Pyx_GIVEREF(__pyx_v_l);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_l);
    __Pyx_INCREF(__pyx_v_r);
    __Pyx_GIVEREF(__pyx_v_r);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_v_r);
    __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 398, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_4);
    __pyx_t_2 = 0;
    __pyx_t_4 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;
  }

  /* "pandas/_libs/intervaltree.pxi":387
 *                                        n_left, n_right, n_center))
 * 
 *     def counts(self):             # <<<<<<<<<<<<<<
 *         """
 *         Inspect counts on this node
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pandas._libs.interval.Float32ClosedLeftIntervalNode.counts", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_l);
  __Pyx_XDECREF(__pyx_v_r);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":244
 *         int64_t[:] center_left_indices, center_right_indices, indices
 *         float32_t min_left, max_right
 *         readonly float32_t pivot             # <<<<<<<<<<<<<<
 *         readonly int64_t n_elements, n_center, leaf_size
 *         readonly bint is_leaf_node
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_5pivot_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_5pivot_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_5pivot___get__(((struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_5pivot___get__(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->pivot); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 244, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Float32ClosedLeftIntervalNode.pivot.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":245
 *         float32_t min_left, max_right
 *         readonly float32_t pivot
 *         readonly int64_t n_elements, n_center, leaf_size             # <<<<<<<<<<<<<<
 *         readonly bint is_leaf_node
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_10n_elements_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_10n_elements_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_10n_elements___get__(((struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_10n_elements___get__(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_elements); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 245, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Float32ClosedLeftIntervalNode.n_elements.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_8n_center_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_8n_center_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_8n_center___get__(((struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_8n_center___get__(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_center); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 245, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Float32ClosedLeftIntervalNode.n_center.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_9leaf_size_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_9leaf_size_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_9leaf_size___get__(((struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_9leaf_size___get__(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->leaf_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 245, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Float32ClosedLeftIntervalNode.leaf_size.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":246
 *         readonly float32_t pivot
 *         readonly int64_t n_elements, n_center, leaf_size
 *         readonly bint is_leaf_node             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self,
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_12is_leaf_node_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_12is_leaf_node_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_12is_leaf_node___get__(((struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_12is_leaf_node___get__(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->is_leaf_node); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 246, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Float32ClosedLeftIntervalNode.is_leaf_node.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_21__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_21__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_20__reduce_cython__(((struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_20__reduce_cython__(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  int __pyx_t_16;
  int __pyx_t_17;
  int __pyx_t_18;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.center_left_indices, self.center_left_values, self.center_right_indices, self.center_right_values, self.indices, self.is_leaf_node, self.leaf_size, self.left, self.left_node, self.max_right, self.min_left, self.n_center, self.n_elements, self.pivot, self.right, self.right_node)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  if (unlikely(!__pyx_v_self->center_left_indices.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_1 = __pyx_memoryview_fromslice(__pyx_v_self->center_left_indices, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_int64_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_int64_t, 0);; if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (unlikely(!__pyx_v_self->center_left_values.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_2 = __pyx_memoryview_fromslice(__pyx_v_self->center_left_values, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_float32_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_float32_t, 0);; if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (unlikely(!__pyx_v_self->center_right_indices.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_3 = __pyx_memoryview_fromslice(__pyx_v_self->center_right_indices, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_int64_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_int64_t, 0);; if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely(!__pyx_v_self->center_right_values.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_4 = __pyx_memoryview_fromslice(__pyx_v_self->center_right_values, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_float32_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_float32_t, 0);; if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (unlikely(!__pyx_v_self->indices.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_5 = __pyx_memoryview_fromslice(__pyx_v_self->indices, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_int64_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_int64_t, 0);; if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyBool_FromLong(__pyx_v_self->is_leaf_node); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->leaf_size); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (unlikely(!__pyx_v_self->left.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_8 = __pyx_memoryview_fromslice(__pyx_v_self->left, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_float32_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_float32_t, 0);; if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = PyFloat_FromDouble(__pyx_v_self->max_right); if (unlikely(!__pyx_t_9)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_10 = PyFloat_FromDouble(__pyx_v_self->min_left); if (unlikely(!__pyx_t_10)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_11 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_center); if (unlikely(!__pyx_t_11)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_12 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_elements); if (unlikely(!__pyx_t_12)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __pyx_t_13 = PyFloat_FromDouble(__pyx_v_self->pivot); if (unlikely(!__pyx_t_13)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  if (unlikely(!__pyx_v_self->right.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_14 = __pyx_memoryview_fromslice(__pyx_v_self->right, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_float32_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_float32_t, 0);; if (unlikely(!__pyx_t_14)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __pyx_t_15 = PyTuple_New(16); if (unlikely(!__pyx_t_15)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_15, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_15, 1, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_15, 2, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_15, 3, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_15, 4, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_15, 5, __pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_15, 6, __pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_15, 7, __pyx_t_8);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->left_node));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self->left_node));
  PyTuple_SET_ITEM(__pyx_t_15, 8, ((PyObject *)__pyx_v_self->left_node));
  __Pyx_GIVEREF(__pyx_t_9);
  PyTuple_SET_ITEM(__pyx_t_15, 9, __pyx_t_9);
  __Pyx_GIVEREF(__pyx_t_10);
  PyTuple_SET_ITEM(__pyx_t_15, 10, __pyx_t_10);
  __Pyx_GIVEREF(__pyx_t_11);
  PyTuple_SET_ITEM(__pyx_t_15, 11, __pyx_t_11);
  __Pyx_GIVEREF(__pyx_t_12);
  PyTuple_SET_ITEM(__pyx_t_15, 12, __pyx_t_12);
  __Pyx_GIVEREF(__pyx_t_13);
  PyTuple_SET_ITEM(__pyx_t_15, 13, __pyx_t_13);
  __Pyx_GIVEREF(__pyx_t_14);
  PyTuple_SET_ITEM(__pyx_t_15, 14, __pyx_t_14);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->right_node));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self->right_node));
  PyTuple_SET_ITEM(__pyx_t_15, 15, ((PyObject *)__pyx_v_self->right_node));
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  __pyx_t_6 = 0;
  __pyx_t_7 = 0;
  __pyx_t_8 = 0;
  __pyx_t_9 = 0;
  __pyx_t_10 = 0;
  __pyx_t_11 = 0;
  __pyx_t_12 = 0;
  __pyx_t_13 = 0;
  __pyx_t_14 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_15);
  __pyx_t_15 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.center_left_indices, self.center_left_values, self.center_right_indices, self.center_right_values, self.indices, self.is_leaf_node, self.leaf_size, self.left, self.left_node, self.max_right, self.min_left, self.n_center, self.n_elements, self.pivot, self.right, self.right_node)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_15 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_15)) __PYX_ERR(2, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __pyx_v__dict = __pyx_t_15;
  __pyx_t_15 = 0;

  /* "(tree fragment)":7
 *     state = (self.center_left_indices, self.center_left_values, self.center_right_indices, self.center_right_values, self.indices, self.is_leaf_node, self.leaf_size, self.left, self.left_node, self.max_right, self.min_left, self.n_center, self.n_elements, self.pivot, self.right, self.right_node)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_16 = (__pyx_v__dict != Py_None);
  __pyx_t_17 = (__pyx_t_16 != 0);
  if (__pyx_t_17) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_15 = PyTuple_New(1); if (unlikely(!__pyx_t_15)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    PyTuple_SET_ITEM(__pyx_t_15, 0, __pyx_v__dict);
    __pyx_t_14 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_15); if (unlikely(!__pyx_t_14)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_14));
    __pyx_t_14 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.left_node is not None or self.right_node is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.center_left_indices, self.center_left_values, self.center_right_indices, self.center_right_values, self.indices, self.is_leaf_node, self.leaf_size, self.left, self.left_node, self.max_right, self.min_left, self.n_center, self.n_elements, self.pivot, self.right, self.right_node)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.left_node is not None or self.right_node is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_Float32ClosedLeftIntervalNode, (type(self), 0x7a104e5, None), state
 */
  /*else*/ {
    __pyx_t_16 = (((PyObject *)__pyx_v_self->left_node) != Py_None);
    __pyx_t_18 = (__pyx_t_16 != 0);
    if (!__pyx_t_18) {
    } else {
      __pyx_t_17 = __pyx_t_18;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_18 = (((PyObject *)__pyx_v_self->right_node) != Py_None);
    __pyx_t_16 = (__pyx_t_18 != 0);
    __pyx_t_17 = __pyx_t_16;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_17;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.left_node is not None or self.right_node is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Float32ClosedLeftIntervalNode, (type(self), 0x7a104e5, None), state
 *     else:
 */
  __pyx_t_17 = (__pyx_v_use_setstate != 0);
  if (__pyx_t_17) {

    /* "(tree fragment)":13
 *         use_setstate = self.left_node is not None or self.right_node is not None
 *     if use_setstate:
 *         return __pyx_unpickle_Float32ClosedLeftIntervalNode, (type(self), 0x7a104e5, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_Float32ClosedLeftIntervalNode, (type(self), 0x7a104e5, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_14, __pyx_n_s_pyx_unpickle_Float32ClosedLeft); if (unlikely(!__pyx_t_14)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __pyx_t_15 = PyTuple_New(3); if (unlikely(!__pyx_t_15)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_15, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_127993061);
    __Pyx_GIVEREF(__pyx_int_127993061);
    PyTuple_SET_ITEM(__pyx_t_15, 1, __pyx_int_127993061);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_15, 2, Py_None);
    __pyx_t_13 = PyTuple_New(3); if (unlikely(!__pyx_t_13)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __Pyx_GIVEREF(__pyx_t_14);
    PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_t_14);
    __Pyx_GIVEREF(__pyx_t_15);
    PyTuple_SET_ITEM(__pyx_t_13, 1, __pyx_t_15);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_13, 2, __pyx_v_state);
    __pyx_t_14 = 0;
    __pyx_t_15 = 0;
    __pyx_r = __pyx_t_13;
    __pyx_t_13 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.left_node is not None or self.right_node is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Float32ClosedLeftIntervalNode, (type(self), 0x7a104e5, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_Float32ClosedLeftIntervalNode, (type(self), 0x7a104e5, None), state
 *     else:
 *         return __pyx_unpickle_Float32ClosedLeftIntervalNode, (type(self), 0x7a104e5, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Float32ClosedLeftIntervalNode__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_13, __pyx_n_s_pyx_unpickle_Float32ClosedLeft); if (unlikely(!__pyx_t_13)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __pyx_t_15 = PyTuple_New(3); if (unlikely(!__pyx_t_15)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_15, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_127993061);
    __Pyx_GIVEREF(__pyx_int_127993061);
    PyTuple_SET_ITEM(__pyx_t_15, 1, __pyx_int_127993061);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_15, 2, __pyx_v_state);
    __pyx_t_14 = PyTuple_New(2); if (unlikely(!__pyx_t_14)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_GIVEREF(__pyx_t_13);
    PyTuple_SET_ITEM(__pyx_t_14, 0, __pyx_t_13);
    __Pyx_GIVEREF(__pyx_t_15);
    PyTuple_SET_ITEM(__pyx_t_14, 1, __pyx_t_15);
    __pyx_t_13 = 0;
    __pyx_t_15 = 0;
    __pyx_r = __pyx_t_14;
    __pyx_t_14 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_AddTraceback("pandas._libs.interval.Float32ClosedLeftIntervalNode.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Float32ClosedLeftIntervalNode, (type(self), 0x7a104e5, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Float32ClosedLeftIntervalNode__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_23__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_23__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_22__setstate_cython__(((struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_22__setstate_cython__(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_Float32ClosedLeftIntervalNode, (type(self), 0x7a104e5, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Float32ClosedLeftIntervalNode__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(2, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_6pandas_5_libs_8interval___pyx_unpickle_Float32ClosedLeftIntervalNode__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Float32ClosedLeftIntervalNode, (type(self), 0x7a104e5, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Float32ClosedLeftIntervalNode__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Float32ClosedLeftIntervalNode.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":418
 *         readonly bint is_leaf_node
 * 
 *     def __init__(self,             # <<<<<<<<<<<<<<
 *                  ndarray[float32_t, ndim=1] left,
 *                  ndarray[float32_t, ndim=1] right,
 */

/* Python wrapper */
static int __pyx_pw_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyArrayObject *__pyx_v_left = 0;
  PyArrayObject *__pyx_v_right = 0;
  PyArrayObject *__pyx_v_indices = 0;
  __pyx_t_5numpy_int64_t __pyx_v_leaf_size;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_left,&__pyx_n_s_right,&__pyx_n_s_indices,&__pyx_n_s_leaf_size,0};
    PyObject* values[4] = {0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_left)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_right)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 4, 4, 1); __PYX_ERR(1, 418, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_indices)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 4, 4, 2); __PYX_ERR(1, 418, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_leaf_size)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 4, 4, 3); __PYX_ERR(1, 418, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(1, 418, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 4) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
    }
    __pyx_v_left = ((PyArrayObject *)values[0]);
    __pyx_v_right = ((PyArrayObject *)values[1]);
    __pyx_v_indices = ((PyArrayObject *)values[2]);
    __pyx_v_leaf_size = __Pyx_PyInt_As_npy_int64(values[3]); if (unlikely((__pyx_v_leaf_size == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(1, 422, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 4, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 418, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Float32ClosedRightIntervalNode.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_left), __pyx_ptype_5numpy_ndarray, 1, "left", 0))) __PYX_ERR(1, 419, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_right), __pyx_ptype_5numpy_ndarray, 1, "right", 0))) __PYX_ERR(1, 420, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_indices), __pyx_ptype_5numpy_ndarray, 1, "indices", 0))) __PYX_ERR(1, 421, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode___init__(((struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode *)__pyx_v_self), __pyx_v_left, __pyx_v_right, __pyx_v_indices, __pyx_v_leaf_size);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode___init__(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode *__pyx_v_self, PyArrayObject *__pyx_v_left, PyArrayObject *__pyx_v_right, PyArrayObject *__pyx_v_indices, __pyx_t_5numpy_int64_t __pyx_v_leaf_size) {
  PyObject *__pyx_v_left_set = NULL;
  PyObject *__pyx_v_right_set = NULL;
  PyObject *__pyx_v_center_set = NULL;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_indices;
  __Pyx_Buffer __pyx_pybuffer_indices;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_left;
  __Pyx_Buffer __pyx_pybuffer_left;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_right;
  __Pyx_Buffer __pyx_pybuffer_right;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __pyx_t_5numpy_float32_t __pyx_t_6;
  __Pyx_memviewslice __pyx_t_7 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_8 = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  __Pyx_memviewslice __pyx_t_11 = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *(*__pyx_t_12)(PyObject *);
  size_t __pyx_t_13;
  __Pyx_RefNannySetupContext("__init__", 0);
  __pyx_pybuffer_left.pybuffer.buf = NULL;
  __pyx_pybuffer_left.refcount = 0;
  __pyx_pybuffernd_left.data = NULL;
  __pyx_pybuffernd_left.rcbuffer = &__pyx_pybuffer_left;
  __pyx_pybuffer_right.pybuffer.buf = NULL;
  __pyx_pybuffer_right.refcount = 0;
  __pyx_pybuffernd_right.data = NULL;
  __pyx_pybuffernd_right.rcbuffer = &__pyx_pybuffer_right;
  __pyx_pybuffer_indices.pybuffer.buf = NULL;
  __pyx_pybuffer_indices.refcount = 0;
  __pyx_pybuffernd_indices.data = NULL;
  __pyx_pybuffernd_indices.rcbuffer = &__pyx_pybuffer_indices;
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_left.rcbuffer->pybuffer, (PyObject*)__pyx_v_left, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float32_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 418, __pyx_L1_error)
  }
  __pyx_pybuffernd_left.diminfo[0].strides = __pyx_pybuffernd_left.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_left.diminfo[0].shape = __pyx_pybuffernd_left.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_right.rcbuffer->pybuffer, (PyObject*)__pyx_v_right, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float32_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 418, __pyx_L1_error)
  }
  __pyx_pybuffernd_right.diminfo[0].strides = __pyx_pybuffernd_right.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_right.diminfo[0].shape = __pyx_pybuffernd_right.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_indices.rcbuffer->pybuffer, (PyObject*)__pyx_v_indices, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 418, __pyx_L1_error)
  }
  __pyx_pybuffernd_indices.diminfo[0].strides = __pyx_pybuffernd_indices.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_indices.diminfo[0].shape = __pyx_pybuffernd_indices.rcbuffer->pybuffer.shape[0];

  /* "pandas/_libs/intervaltree.pxi":424
 *                  int64_t leaf_size):
 * 
 *         self.n_elements = len(left)             # <<<<<<<<<<<<<<
 *         self.leaf_size = leaf_size
 * 
 */
  __pyx_t_1 = PyObject_Length(((PyObject *)__pyx_v_left)); if (unlikely(__pyx_t_1 == ((Py_ssize_t)-1))) __PYX_ERR(1, 424, __pyx_L1_error)
  __pyx_v_self->n_elements = __pyx_t_1;

  /* "pandas/_libs/intervaltree.pxi":425
 * 
 *         self.n_elements = len(left)
 *         self.leaf_size = leaf_size             # <<<<<<<<<<<<<<
 * 
 *         # min_left and min_right are used to speed-up query by skipping
 */
  __pyx_v_self->leaf_size = __pyx_v_leaf_size;

  /* "pandas/_libs/intervaltree.pxi":430
 *         # query on sub-nodes. If this node has size 0, query is cheap,
 *         # so these values don't matter.
 *         if left.size > 0:             # <<<<<<<<<<<<<<
 *             self.min_left = left.min()
 *             self.max_right = right.max()
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_left), __pyx_n_s_size); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 430, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyObject_RichCompare(__pyx_t_2, __pyx_int_0, Py_GT); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 430, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(1, 430, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_4) {

    /* "pandas/_libs/intervaltree.pxi":431
 *         # so these values don't matter.
 *         if left.size > 0:
 *             self.min_left = left.min()             # <<<<<<<<<<<<<<
 *             self.max_right = right.max()
 *         else:
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_left), __pyx_n_s_min); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 431, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 431, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_6 = __pyx_PyFloat_AsFloat(__pyx_t_3); if (unlikely((__pyx_t_6 == ((npy_float32)-1)) && PyErr_Occurred())) __PYX_ERR(1, 431, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_self->min_left = __pyx_t_6;

    /* "pandas/_libs/intervaltree.pxi":432
 *         if left.size > 0:
 *             self.min_left = left.min()
 *             self.max_right = right.max()             # <<<<<<<<<<<<<<
 *         else:
 *             self.min_left = 0
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_right), __pyx_n_s_max); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 432, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 432, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_6 = __pyx_PyFloat_AsFloat(__pyx_t_3); if (unlikely((__pyx_t_6 == ((npy_float32)-1)) && PyErr_Occurred())) __PYX_ERR(1, 432, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_self->max_right = __pyx_t_6;

    /* "pandas/_libs/intervaltree.pxi":430
 *         # query on sub-nodes. If this node has size 0, query is cheap,
 *         # so these values don't matter.
 *         if left.size > 0:             # <<<<<<<<<<<<<<
 *             self.min_left = left.min()
 *             self.max_right = right.max()
 */
    goto __pyx_L3;
  }

  /* "pandas/_libs/intervaltree.pxi":434
 *             self.max_right = right.max()
 *         else:
 *             self.min_left = 0             # <<<<<<<<<<<<<<
 *             self.max_right = 0
 * 
 */
  /*else*/ {
    __pyx_v_self->min_left = 0.0;

    /* "pandas/_libs/intervaltree.pxi":435
 *         else:
 *             self.min_left = 0
 *             self.max_right = 0             # <<<<<<<<<<<<<<
 * 
 *         if self.n_elements <= leaf_size:
 */
    __pyx_v_self->max_right = 0.0;
  }
  __pyx_L3:;

  /* "pandas/_libs/intervaltree.pxi":437
 *             self.max_right = 0
 * 
 *         if self.n_elements <= leaf_size:             # <<<<<<<<<<<<<<
 *             # make this a terminal (leaf) node
 *             self.is_leaf_node = True
 */
  __pyx_t_4 = ((__pyx_v_self->n_elements <= __pyx_v_leaf_size) != 0);
  if (__pyx_t_4) {

    /* "pandas/_libs/intervaltree.pxi":439
 *         if self.n_elements <= leaf_size:
 *             # make this a terminal (leaf) node
 *             self.is_leaf_node = True             # <<<<<<<<<<<<<<
 *             self.left = left
 *             self.right = right
 */
    __pyx_v_self->is_leaf_node = 1;

    /* "pandas/_libs/intervaltree.pxi":440
 *             # make this a terminal (leaf) node
 *             self.is_leaf_node = True
 *             self.left = left             # <<<<<<<<<<<<<<
 *             self.right = right
 *             self.indices = indices
 */
    __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_float32_t(((PyObject *)__pyx_v_left), PyBUF_WRITABLE); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(1, 440, __pyx_L1_error)
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->left, 0);
    __pyx_v_self->left = __pyx_t_7;
    __pyx_t_7.memview = NULL;
    __pyx_t_7.data = NULL;

    /* "pandas/_libs/intervaltree.pxi":441
 *             self.is_leaf_node = True
 *             self.left = left
 *             self.right = right             # <<<<<<<<<<<<<<
 *             self.indices = indices
 *             self.n_center = 0
 */
    __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_float32_t(((PyObject *)__pyx_v_right), PyBUF_WRITABLE); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(1, 441, __pyx_L1_error)
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->right, 0);
    __pyx_v_self->right = __pyx_t_7;
    __pyx_t_7.memview = NULL;
    __pyx_t_7.data = NULL;

    /* "pandas/_libs/intervaltree.pxi":442
 *             self.left = left
 *             self.right = right
 *             self.indices = indices             # <<<<<<<<<<<<<<
 *             self.n_center = 0
 *         else:
 */
    __pyx_t_8 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(((PyObject *)__pyx_v_indices), PyBUF_WRITABLE); if (unlikely(!__pyx_t_8.memview)) __PYX_ERR(1, 442, __pyx_L1_error)
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->indices, 0);
    __pyx_v_self->indices = __pyx_t_8;
    __pyx_t_8.memview = NULL;
    __pyx_t_8.data = NULL;

    /* "pandas/_libs/intervaltree.pxi":443
 *             self.right = right
 *             self.indices = indices
 *             self.n_center = 0             # <<<<<<<<<<<<<<
 *         else:
 *             # calculate a pivot so we can create child nodes
 */
    __pyx_v_self->n_center = 0;

    /* "pandas/_libs/intervaltree.pxi":437
 *             self.max_right = 0
 * 
 *         if self.n_elements <= leaf_size:             # <<<<<<<<<<<<<<
 *             # make this a terminal (leaf) node
 *             self.is_leaf_node = True
 */
    goto __pyx_L4;
  }

  /* "pandas/_libs/intervaltree.pxi":446
 *         else:
 *             # calculate a pivot so we can create child nodes
 *             self.is_leaf_node = False             # <<<<<<<<<<<<<<
 *             self.pivot = np.median(left / 2 + right / 2)
 *             left_set, right_set, center_set = self.classify_intervals(
 */
  /*else*/ {
    __pyx_v_self->is_leaf_node = 0;

    /* "pandas/_libs/intervaltree.pxi":447
 *             # calculate a pivot so we can create child nodes
 *             self.is_leaf_node = False
 *             self.pivot = np.median(left / 2 + right / 2)             # <<<<<<<<<<<<<<
 *             left_set, right_set, center_set = self.classify_intervals(
 *                 left, right)
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 447, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_median); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 447, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyNumber_Divide(((PyObject *)__pyx_v_left), __pyx_int_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 447, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_9 = __Pyx_PyNumber_Divide(((PyObject *)__pyx_v_right), __pyx_int_2); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 447, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_10 = PyNumber_Add(__pyx_t_2, __pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 447, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_9 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_3 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_9, __pyx_t_10) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_10);
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 447, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_6 = __pyx_PyFloat_AsFloat(__pyx_t_3); if (unlikely((__pyx_t_6 == ((npy_float32)-1)) && PyErr_Occurred())) __PYX_ERR(1, 447, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_self->pivot = __pyx_t_6;

    /* "pandas/_libs/intervaltree.pxi":449
 *             self.pivot = np.median(left / 2 + right / 2)
 *             left_set, right_set, center_set = self.classify_intervals(
 *                 left, right)             # <<<<<<<<<<<<<<
 * 
 *             self.left_node = self.new_child_node(left, right,
 */
    __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_float32_t(((PyObject *)__pyx_v_left), PyBUF_WRITABLE); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(1, 449, __pyx_L1_error)
    __pyx_t_11 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_float32_t(((PyObject *)__pyx_v_right), PyBUF_WRITABLE); if (unlikely(!__pyx_t_11.memview)) __PYX_ERR(1, 449, __pyx_L1_error)

    /* "pandas/_libs/intervaltree.pxi":448
 *             self.is_leaf_node = False
 *             self.pivot = np.median(left / 2 + right / 2)
 *             left_set, right_set, center_set = self.classify_intervals(             # <<<<<<<<<<<<<<
 *                 left, right)
 * 
 */
    __pyx_t_3 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode *)__pyx_v_self->__pyx_vtab)->classify_intervals(__pyx_v_self, __pyx_t_7, __pyx_t_11); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 448, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __PYX_XDEC_MEMVIEW(&__pyx_t_7, 1);
    __pyx_t_7.memview = NULL;
    __pyx_t_7.data = NULL;
    __PYX_XDEC_MEMVIEW(&__pyx_t_11, 1);
    __pyx_t_11.memview = NULL;
    __pyx_t_11.data = NULL;
    if ((likely(PyTuple_CheckExact(__pyx_t_3))) || (PyList_CheckExact(__pyx_t_3))) {
      PyObject* sequence = __pyx_t_3;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 3)) {
        if (size > 3) __Pyx_RaiseTooManyValuesError(3);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(1, 448, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_5 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_10 = PyTuple_GET_ITEM(sequence, 1); 
        __pyx_t_9 = PyTuple_GET_ITEM(sequence, 2); 
      } else {
        __pyx_t_5 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_10 = PyList_GET_ITEM(sequence, 1); 
        __pyx_t_9 = PyList_GET_ITEM(sequence, 2); 
      }
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_10);
      __Pyx_INCREF(__pyx_t_9);
      #else
      __pyx_t_5 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 448, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_10 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 448, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_9 = PySequence_ITEM(sequence, 2); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 448, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      #endif
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_2 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 448, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_12 = Py_TYPE(__pyx_t_2)->tp_iternext;
      index = 0; __pyx_t_5 = __pyx_t_12(__pyx_t_2); if (unlikely(!__pyx_t_5)) goto __pyx_L5_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_5);
      index = 1; __pyx_t_10 = __pyx_t_12(__pyx_t_2); if (unlikely(!__pyx_t_10)) goto __pyx_L5_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_10);
      index = 2; __pyx_t_9 = __pyx_t_12(__pyx_t_2); if (unlikely(!__pyx_t_9)) goto __pyx_L5_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_9);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_12(__pyx_t_2), 3) < 0) __PYX_ERR(1, 448, __pyx_L1_error)
      __pyx_t_12 = NULL;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      goto __pyx_L6_unpacking_done;
      __pyx_L5_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_12 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(1, 448, __pyx_L1_error)
      __pyx_L6_unpacking_done:;
    }
    __pyx_v_left_set = __pyx_t_5;
    __pyx_t_5 = 0;
    __pyx_v_right_set = __pyx_t_10;
    __pyx_t_10 = 0;
    __pyx_v_center_set = __pyx_t_9;
    __pyx_t_9 = 0;

    /* "pandas/_libs/intervaltree.pxi":452
 * 
 *             self.left_node = self.new_child_node(left, right,
 *                                                  indices, left_set)             # <<<<<<<<<<<<<<
 *             self.right_node = self.new_child_node(left, right,
 *                                                   indices, right_set)
 */
    if (!(likely(((__pyx_v_left_set) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_left_set, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(1, 452, __pyx_L1_error)

    /* "pandas/_libs/intervaltree.pxi":451
 *                 left, right)
 * 
 *             self.left_node = self.new_child_node(left, right,             # <<<<<<<<<<<<<<
 *                                                  indices, left_set)
 *             self.right_node = self.new_child_node(left, right,
 */
    __pyx_t_3 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode *)__pyx_v_self->__pyx_vtab)->new_child_node(__pyx_v_self, ((PyArrayObject *)__pyx_v_left), ((PyArrayObject *)__pyx_v_right), ((PyArrayObject *)__pyx_v_indices), ((PyArrayObject *)__pyx_v_left_set)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 451, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode))))) __PYX_ERR(1, 451, __pyx_L1_error)
    __Pyx_GIVEREF(__pyx_t_3);
    __Pyx_GOTREF(__pyx_v_self->left_node);
    __Pyx_DECREF(((PyObject *)__pyx_v_self->left_node));
    __pyx_v_self->left_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode *)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "pandas/_libs/intervaltree.pxi":454
 *                                                  indices, left_set)
 *             self.right_node = self.new_child_node(left, right,
 *                                                   indices, right_set)             # <<<<<<<<<<<<<<
 * 
 *             self.center_left_values, self.center_left_indices = \
 */
    if (!(likely(((__pyx_v_right_set) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_right_set, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(1, 454, __pyx_L1_error)

    /* "pandas/_libs/intervaltree.pxi":453
 *             self.left_node = self.new_child_node(left, right,
 *                                                  indices, left_set)
 *             self.right_node = self.new_child_node(left, right,             # <<<<<<<<<<<<<<
 *                                                   indices, right_set)
 * 
 */
    __pyx_t_3 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode *)__pyx_v_self->__pyx_vtab)->new_child_node(__pyx_v_self, ((PyArrayObject *)__pyx_v_left), ((PyArrayObject *)__pyx_v_right), ((PyArrayObject *)__pyx_v_indices), ((PyArrayObject *)__pyx_v_right_set)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 453, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode))))) __PYX_ERR(1, 453, __pyx_L1_error)
    __Pyx_GIVEREF(__pyx_t_3);
    __Pyx_GOTREF(__pyx_v_self->right_node);
    __Pyx_DECREF(((PyObject *)__pyx_v_self->right_node));
    __pyx_v_self->right_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode *)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "pandas/_libs/intervaltree.pxi":457
 * 
 *             self.center_left_values, self.center_left_indices = \
 *                 sort_values_and_indices(left, indices, center_set)             # <<<<<<<<<<<<<<
 *             self.center_right_values, self.center_right_indices = \
 *                 sort_values_and_indices(right, indices, center_set)
 */
    __pyx_t_3 = __pyx_f_6pandas_5_libs_8interval_sort_values_and_indices(((PyObject *)__pyx_v_left), ((PyObject *)__pyx_v_indices), __pyx_v_center_set); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 457, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if ((likely(PyTuple_CheckExact(__pyx_t_3))) || (PyList_CheckExact(__pyx_t_3))) {
      PyObject* sequence = __pyx_t_3;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(1, 456, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_9 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_10 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_9 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_10 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_9);
      __Pyx_INCREF(__pyx_t_10);
      #else
      __pyx_t_9 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 456, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_10 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 456, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      #endif
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_5 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 456, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_12 = Py_TYPE(__pyx_t_5)->tp_iternext;
      index = 0; __pyx_t_9 = __pyx_t_12(__pyx_t_5); if (unlikely(!__pyx_t_9)) goto __pyx_L7_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_9);
      index = 1; __pyx_t_10 = __pyx_t_12(__pyx_t_5); if (unlikely(!__pyx_t_10)) goto __pyx_L7_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_10);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_12(__pyx_t_5), 2) < 0) __PYX_ERR(1, 456, __pyx_L1_error)
      __pyx_t_12 = NULL;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      goto __pyx_L8_unpacking_done;
      __pyx_L7_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_12 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(1, 456, __pyx_L1_error)
      __pyx_L8_unpacking_done:;
    }

    /* "pandas/_libs/intervaltree.pxi":456
 *                                                   indices, right_set)
 * 
 *             self.center_left_values, self.center_left_indices = \             # <<<<<<<<<<<<<<
 *                 sort_values_and_indices(left, indices, center_set)
 *             self.center_right_values, self.center_right_indices = \
 */
    __pyx_t_11 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_float32_t(__pyx_t_9, PyBUF_WRITABLE); if (unlikely(!__pyx_t_11.memview)) __PYX_ERR(1, 456, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_8 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_t_10, PyBUF_WRITABLE); if (unlikely(!__pyx_t_8.memview)) __PYX_ERR(1, 456, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->center_left_values, 0);
    __pyx_v_self->center_left_values = __pyx_t_11;
    __pyx_t_11.memview = NULL;
    __pyx_t_11.data = NULL;
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->center_left_indices, 0);
    __pyx_v_self->center_left_indices = __pyx_t_8;
    __pyx_t_8.memview = NULL;
    __pyx_t_8.data = NULL;

    /* "pandas/_libs/intervaltree.pxi":459
 *                 sort_values_and_indices(left, indices, center_set)
 *             self.center_right_values, self.center_right_indices = \
 *                 sort_values_and_indices(right, indices, center_set)             # <<<<<<<<<<<<<<
 *             self.n_center = len(self.center_left_indices)
 * 
 */
    __pyx_t_3 = __pyx_f_6pandas_5_libs_8interval_sort_values_and_indices(((PyObject *)__pyx_v_right), ((PyObject *)__pyx_v_indices), __pyx_v_center_set); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 459, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if ((likely(PyTuple_CheckExact(__pyx_t_3))) || (PyList_CheckExact(__pyx_t_3))) {
      PyObject* sequence = __pyx_t_3;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(1, 458, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_10 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_9 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_10 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_9 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_10);
      __Pyx_INCREF(__pyx_t_9);
      #else
      __pyx_t_10 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 458, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_9 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 458, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      #endif
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_5 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 458, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_12 = Py_TYPE(__pyx_t_5)->tp_iternext;
      index = 0; __pyx_t_10 = __pyx_t_12(__pyx_t_5); if (unlikely(!__pyx_t_10)) goto __pyx_L9_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_10);
      index = 1; __pyx_t_9 = __pyx_t_12(__pyx_t_5); if (unlikely(!__pyx_t_9)) goto __pyx_L9_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_9);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_12(__pyx_t_5), 2) < 0) __PYX_ERR(1, 458, __pyx_L1_error)
      __pyx_t_12 = NULL;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      goto __pyx_L10_unpacking_done;
      __pyx_L9_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_12 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(1, 458, __pyx_L1_error)
      __pyx_L10_unpacking_done:;
    }

    /* "pandas/_libs/intervaltree.pxi":458
 *             self.center_left_values, self.center_left_indices = \
 *                 sort_values_and_indices(left, indices, center_set)
 *             self.center_right_values, self.center_right_indices = \             # <<<<<<<<<<<<<<
 *                 sort_values_and_indices(right, indices, center_set)
 *             self.n_center = len(self.center_left_indices)
 */
    __pyx_t_11 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_float32_t(__pyx_t_10, PyBUF_WRITABLE); if (unlikely(!__pyx_t_11.memview)) __PYX_ERR(1, 458, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_t_8 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_t_9, PyBUF_WRITABLE); if (unlikely(!__pyx_t_8.memview)) __PYX_ERR(1, 458, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->center_right_values, 0);
    __pyx_v_self->center_right_values = __pyx_t_11;
    __pyx_t_11.memview = NULL;
    __pyx_t_11.data = NULL;
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->center_right_indices, 0);
    __pyx_v_self->center_right_indices = __pyx_t_8;
    __pyx_t_8.memview = NULL;
    __pyx_t_8.data = NULL;

    /* "pandas/_libs/intervaltree.pxi":460
 *             self.center_right_values, self.center_right_indices = \
 *                 sort_values_and_indices(right, indices, center_set)
 *             self.n_center = len(self.center_left_indices)             # <<<<<<<<<<<<<<
 * 
 *     @cython.wraparound(False)
 */
    if (unlikely(!__pyx_v_self->center_left_indices.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(1, 460, __pyx_L1_error)}
    __pyx_t_13 = __Pyx_MemoryView_Len(__pyx_v_self->center_left_indices); 
    __pyx_v_self->n_center = __pyx_t_13;
  }
  __pyx_L4:;

  /* "pandas/_libs/intervaltree.pxi":418
 *         readonly bint is_leaf_node
 * 
 *     def __init__(self,             # <<<<<<<<<<<<<<
 *                  ndarray[float32_t, ndim=1] left,
 *                  ndarray[float32_t, ndim=1] right,
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __PYX_XDEC_MEMVIEW(&__pyx_t_7, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_8, 1);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __PYX_XDEC_MEMVIEW(&__pyx_t_11, 1);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indices.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_left.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_right.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("pandas._libs.interval.Float32ClosedRightIntervalNode.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indices.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_left.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_right.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF(__pyx_v_left_set);
  __Pyx_XDECREF(__pyx_v_right_set);
  __Pyx_XDECREF(__pyx_v_center_set);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":464
 *     @cython.wraparound(False)
 *     @cython.boundscheck(False)
 *     cdef classify_intervals(self, float32_t[:] left, float32_t[:] right):             # <<<<<<<<<<<<<<
 *         """Classify the given intervals based upon whether they fall to the
 *         left, right, or overlap with this node's pivot.
 */

static PyObject *__pyx_f_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_classify_intervals(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode *__pyx_v_self, __Pyx_memviewslice __pyx_v_left, __Pyx_memviewslice __pyx_v_right) {
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_left_ind = 0;
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_right_ind = 0;
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_overlapping_ind = 0;
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __pyx_t_5numpy_int64_t __pyx_t_2;
  __pyx_t_5numpy_int64_t __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  int __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  __Pyx_RefNannySetupContext("classify_intervals", 0);

  /* "pandas/_libs/intervaltree.pxi":472
 *             Py_ssize_t i
 * 
 *         left_ind = Int64Vector()             # <<<<<<<<<<<<<<
 *         right_ind = Int64Vector()
 *         overlapping_ind = Int64Vector()
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 472, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_left_ind = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":473
 * 
 *         left_ind = Int64Vector()
 *         right_ind = Int64Vector()             # <<<<<<<<<<<<<<
 *         overlapping_ind = Int64Vector()
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 473, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_right_ind = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":474
 *         left_ind = Int64Vector()
 *         right_ind = Int64Vector()
 *         overlapping_ind = Int64Vector()             # <<<<<<<<<<<<<<
 * 
 *         for i in range(self.n_elements):
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 474, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_overlapping_ind = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":476
 *         overlapping_ind = Int64Vector()
 * 
 *         for i in range(self.n_elements):             # <<<<<<<<<<<<<<
 *             if right[i] < self.pivot:
 *                 left_ind.append(i)
 */
  __pyx_t_2 = __pyx_v_self->n_elements;
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "pandas/_libs/intervaltree.pxi":477
 * 
 *         for i in range(self.n_elements):
 *             if right[i] < self.pivot:             # <<<<<<<<<<<<<<
 *                 left_ind.append(i)
 *             elif self.pivot <= left[i]:
 */
    __pyx_t_5 = __pyx_v_i;
    __pyx_t_6 = (((*((__pyx_t_5numpy_float32_t *) ( /* dim=0 */ (__pyx_v_right.data + __pyx_t_5 * __pyx_v_right.strides[0]) ))) < __pyx_v_self->pivot) != 0);
    if (__pyx_t_6) {

      /* "pandas/_libs/intervaltree.pxi":478
 *         for i in range(self.n_elements):
 *             if right[i] < self.pivot:
 *                 left_ind.append(i)             # <<<<<<<<<<<<<<
 *             elif self.pivot <= left[i]:
 *                 right_ind.append(i)
 */
      ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_left_ind->__pyx_vtab)->append(__pyx_v_left_ind, __pyx_v_i);

      /* "pandas/_libs/intervaltree.pxi":477
 * 
 *         for i in range(self.n_elements):
 *             if right[i] < self.pivot:             # <<<<<<<<<<<<<<
 *                 left_ind.append(i)
 *             elif self.pivot <= left[i]:
 */
      goto __pyx_L5;
    }

    /* "pandas/_libs/intervaltree.pxi":479
 *             if right[i] < self.pivot:
 *                 left_ind.append(i)
 *             elif self.pivot <= left[i]:             # <<<<<<<<<<<<<<
 *                 right_ind.append(i)
 *             else:
 */
    __pyx_t_7 = __pyx_v_i;
    __pyx_t_6 = ((__pyx_v_self->pivot <= (*((__pyx_t_5numpy_float32_t *) ( /* dim=0 */ (__pyx_v_left.data + __pyx_t_7 * __pyx_v_left.strides[0]) )))) != 0);
    if (__pyx_t_6) {

      /* "pandas/_libs/intervaltree.pxi":480
 *                 left_ind.append(i)
 *             elif self.pivot <= left[i]:
 *                 right_ind.append(i)             # <<<<<<<<<<<<<<
 *             else:
 *                 overlapping_ind.append(i)
 */
      ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_right_ind->__pyx_vtab)->append(__pyx_v_right_ind, __pyx_v_i);

      /* "pandas/_libs/intervaltree.pxi":479
 *             if right[i] < self.pivot:
 *                 left_ind.append(i)
 *             elif self.pivot <= left[i]:             # <<<<<<<<<<<<<<
 *                 right_ind.append(i)
 *             else:
 */
      goto __pyx_L5;
    }

    /* "pandas/_libs/intervaltree.pxi":482
 *                 right_ind.append(i)
 *             else:
 *                 overlapping_ind.append(i)             # <<<<<<<<<<<<<<
 * 
 *         return (left_ind.to_array(),
 */
    /*else*/ {
      ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_overlapping_ind->__pyx_vtab)->append(__pyx_v_overlapping_ind, __pyx_v_i);
    }
    __pyx_L5:;
  }

  /* "pandas/_libs/intervaltree.pxi":484
 *                 overlapping_ind.append(i)
 * 
 *         return (left_ind.to_array(),             # <<<<<<<<<<<<<<
 *                 right_ind.to_array(),
 *                 overlapping_ind.to_array())
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_left_ind->__pyx_vtab)->to_array(__pyx_v_left_ind, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 484, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "pandas/_libs/intervaltree.pxi":485
 * 
 *         return (left_ind.to_array(),
 *                 right_ind.to_array(),             # <<<<<<<<<<<<<<
 *                 overlapping_ind.to_array())
 * 
 */
  __pyx_t_8 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_right_ind->__pyx_vtab)->to_array(__pyx_v_right_ind, 0); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 485, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);

  /* "pandas/_libs/intervaltree.pxi":486
 *         return (left_ind.to_array(),
 *                 right_ind.to_array(),
 *                 overlapping_ind.to_array())             # <<<<<<<<<<<<<<
 * 
 *     cdef new_child_node(self,
 */
  __pyx_t_9 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_overlapping_ind->__pyx_vtab)->to_array(__pyx_v_overlapping_ind, 0); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 486, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);

  /* "pandas/_libs/intervaltree.pxi":484
 *                 overlapping_ind.append(i)
 * 
 *         return (left_ind.to_array(),             # <<<<<<<<<<<<<<
 *                 right_ind.to_array(),
 *                 overlapping_ind.to_array())
 */
  __pyx_t_10 = PyTuple_New(3); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 484, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_t_8);
  __Pyx_GIVEREF(__pyx_t_9);
  PyTuple_SET_ITEM(__pyx_t_10, 2, __pyx_t_9);
  __pyx_t_1 = 0;
  __pyx_t_8 = 0;
  __pyx_t_9 = 0;
  __pyx_r = __pyx_t_10;
  __pyx_t_10 = 0;
  goto __pyx_L0;

  /* "pandas/_libs/intervaltree.pxi":464
 *     @cython.wraparound(False)
 *     @cython.boundscheck(False)
 *     cdef classify_intervals(self, float32_t[:] left, float32_t[:] right):             # <<<<<<<<<<<<<<
 *         """Classify the given intervals based upon whether they fall to the
 *         left, right, or overlap with this node's pivot.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("pandas._libs.interval.Float32ClosedRightIntervalNode.classify_intervals", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_left_ind);
  __Pyx_XDECREF((PyObject *)__pyx_v_right_ind);
  __Pyx_XDECREF((PyObject *)__pyx_v_overlapping_ind);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":488
 *                 overlapping_ind.to_array())
 * 
 *     cdef new_child_node(self,             # <<<<<<<<<<<<<<
 *                         ndarray[float32_t, ndim=1] left,
 *                         ndarray[float32_t, ndim=1] right,
 */

static PyObject *__pyx_f_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_new_child_node(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode *__pyx_v_self, PyArrayObject *__pyx_v_left, PyArrayObject *__pyx_v_right, PyArrayObject *__pyx_v_indices, PyArrayObject *__pyx_v_subset) {
  __Pyx_LocalBuf_ND __pyx_pybuffernd_indices;
  __Pyx_Buffer __pyx_pybuffer_indices;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_left;
  __Pyx_Buffer __pyx_pybuffer_left;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_right;
  __Pyx_Buffer __pyx_pybuffer_right;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_subset;
  __Pyx_Buffer __pyx_pybuffer_subset;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyArrayObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyArrayObject *__pyx_t_7 = NULL;
  PyArrayObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  __Pyx_RefNannySetupContext("new_child_node", 0);
  __Pyx_INCREF((PyObject *)__pyx_v_left);
  __Pyx_INCREF((PyObject *)__pyx_v_right);
  __Pyx_INCREF((PyObject *)__pyx_v_indices);
  __pyx_pybuffer_left.pybuffer.buf = NULL;
  __pyx_pybuffer_left.refcount = 0;
  __pyx_pybuffernd_left.data = NULL;
  __pyx_pybuffernd_left.rcbuffer = &__pyx_pybuffer_left;
  __pyx_pybuffer_right.pybuffer.buf = NULL;
  __pyx_pybuffer_right.refcount = 0;
  __pyx_pybuffernd_right.data = NULL;
  __pyx_pybuffernd_right.rcbuffer = &__pyx_pybuffer_right;
  __pyx_pybuffer_indices.pybuffer.buf = NULL;
  __pyx_pybuffer_indices.refcount = 0;
  __pyx_pybuffernd_indices.data = NULL;
  __pyx_pybuffernd_indices.rcbuffer = &__pyx_pybuffer_indices;
  __pyx_pybuffer_subset.pybuffer.buf = NULL;
  __pyx_pybuffer_subset.refcount = 0;
  __pyx_pybuffernd_subset.data = NULL;
  __pyx_pybuffernd_subset.rcbuffer = &__pyx_pybuffer_subset;
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_left.rcbuffer->pybuffer, (PyObject*)__pyx_v_left, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float32_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 488, __pyx_L1_error)
  }
  __pyx_pybuffernd_left.diminfo[0].strides = __pyx_pybuffernd_left.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_left.diminfo[0].shape = __pyx_pybuffernd_left.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_right.rcbuffer->pybuffer, (PyObject*)__pyx_v_right, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float32_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 488, __pyx_L1_error)
  }
  __pyx_pybuffernd_right.diminfo[0].strides = __pyx_pybuffernd_right.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_right.diminfo[0].shape = __pyx_pybuffernd_right.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_indices.rcbuffer->pybuffer, (PyObject*)__pyx_v_indices, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 488, __pyx_L1_error)
  }
  __pyx_pybuffernd_indices.diminfo[0].strides = __pyx_pybuffernd_indices.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_indices.diminfo[0].shape = __pyx_pybuffernd_indices.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_subset.rcbuffer->pybuffer, (PyObject*)__pyx_v_subset, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 488, __pyx_L1_error)
  }
  __pyx_pybuffernd_subset.diminfo[0].strides = __pyx_pybuffernd_subset.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_subset.diminfo[0].shape = __pyx_pybuffernd_subset.rcbuffer->pybuffer.shape[0];

  /* "pandas/_libs/intervaltree.pxi":495
 *         """Create a new child node.
 *         """
 *         left = take(left, subset)             # <<<<<<<<<<<<<<
 *         right = take(right, subset)
 *         indices = take(indices, subset)
 */
  __pyx_t_1 = __pyx_f_6pandas_5_libs_8interval_take(((PyArrayObject *)__pyx_v_left), ((PyArrayObject *)__pyx_v_subset)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 495, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(1, 495, __pyx_L1_error)
  __pyx_t_2 = ((PyArrayObject *)__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_left.rcbuffer->pybuffer);
    __pyx_t_3 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_left.rcbuffer->pybuffer, (PyObject*)__pyx_t_2, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float32_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack);
    if (unlikely(__pyx_t_3 < 0)) {
      PyErr_Fetch(&__pyx_t_4, &__pyx_t_5, &__pyx_t_6);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_left.rcbuffer->pybuffer, (PyObject*)__pyx_v_left, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float32_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_4); Py_XDECREF(__pyx_t_5); Py_XDECREF(__pyx_t_6);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_4, __pyx_t_5, __pyx_t_6);
      }
      __pyx_t_4 = __pyx_t_5 = __pyx_t_6 = 0;
    }
    __pyx_pybuffernd_left.diminfo[0].strides = __pyx_pybuffernd_left.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_left.diminfo[0].shape = __pyx_pybuffernd_left.rcbuffer->pybuffer.shape[0];
    if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(1, 495, __pyx_L1_error)
  }
  __pyx_t_2 = 0;
  __Pyx_DECREF_SET(__pyx_v_left, ((PyArrayObject *)__pyx_t_1));
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":496
 *         """
 *         left = take(left, subset)
 *         right = take(right, subset)             # <<<<<<<<<<<<<<
 *         indices = take(indices, subset)
 *         return Float32ClosedRightIntervalNode(
 */
  __pyx_t_1 = __pyx_f_6pandas_5_libs_8interval_take(((PyArrayObject *)__pyx_v_right), ((PyArrayObject *)__pyx_v_subset)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 496, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(1, 496, __pyx_L1_error)
  __pyx_t_7 = ((PyArrayObject *)__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_right.rcbuffer->pybuffer);
    __pyx_t_3 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_right.rcbuffer->pybuffer, (PyObject*)__pyx_t_7, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float32_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack);
    if (unlikely(__pyx_t_3 < 0)) {
      PyErr_Fetch(&__pyx_t_6, &__pyx_t_5, &__pyx_t_4);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_right.rcbuffer->pybuffer, (PyObject*)__pyx_v_right, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float32_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_6); Py_XDECREF(__pyx_t_5); Py_XDECREF(__pyx_t_4);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_6, __pyx_t_5, __pyx_t_4);
      }
      __pyx_t_6 = __pyx_t_5 = __pyx_t_4 = 0;
    }
    __pyx_pybuffernd_right.diminfo[0].strides = __pyx_pybuffernd_right.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_right.diminfo[0].shape = __pyx_pybuffernd_right.rcbuffer->pybuffer.shape[0];
    if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(1, 496, __pyx_L1_error)
  }
  __pyx_t_7 = 0;
  __Pyx_DECREF_SET(__pyx_v_right, ((PyArrayObject *)__pyx_t_1));
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":497
 *         left = take(left, subset)
 *         right = take(right, subset)
 *         indices = take(indices, subset)             # <<<<<<<<<<<<<<
 *         return Float32ClosedRightIntervalNode(
 *             left, right, indices, self.leaf_size)
 */
  __pyx_t_1 = __pyx_f_6pandas_5_libs_8interval_take(((PyArrayObject *)__pyx_v_indices), ((PyArrayObject *)__pyx_v_subset)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 497, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(1, 497, __pyx_L1_error)
  __pyx_t_8 = ((PyArrayObject *)__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indices.rcbuffer->pybuffer);
    __pyx_t_3 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_indices.rcbuffer->pybuffer, (PyObject*)__pyx_t_8, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack);
    if (unlikely(__pyx_t_3 < 0)) {
      PyErr_Fetch(&__pyx_t_4, &__pyx_t_5, &__pyx_t_6);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_indices.rcbuffer->pybuffer, (PyObject*)__pyx_v_indices, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_4); Py_XDECREF(__pyx_t_5); Py_XDECREF(__pyx_t_6);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_4, __pyx_t_5, __pyx_t_6);
      }
      __pyx_t_4 = __pyx_t_5 = __pyx_t_6 = 0;
    }
    __pyx_pybuffernd_indices.diminfo[0].strides = __pyx_pybuffernd_indices.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_indices.diminfo[0].shape = __pyx_pybuffernd_indices.rcbuffer->pybuffer.shape[0];
    if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(1, 497, __pyx_L1_error)
  }
  __pyx_t_8 = 0;
  __Pyx_DECREF_SET(__pyx_v_indices, ((PyArrayObject *)__pyx_t_1));
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":498
 *         right = take(right, subset)
 *         indices = take(indices, subset)
 *         return Float32ClosedRightIntervalNode(             # <<<<<<<<<<<<<<
 *             left, right, indices, self.leaf_size)
 * 
 */
  __Pyx_XDECREF(__pyx_r);

  /* "pandas/_libs/intervaltree.pxi":499
 *         indices = take(indices, subset)
 *         return Float32ClosedRightIntervalNode(
 *             left, right, indices, self.leaf_size)             # <<<<<<<<<<<<<<
 * 
 *     @cython.wraparound(False)
 */
  __pyx_t_1 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->leaf_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 499, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "pandas/_libs/intervaltree.pxi":498
 *         right = take(right, subset)
 *         indices = take(indices, subset)
 *         return Float32ClosedRightIntervalNode(             # <<<<<<<<<<<<<<
 *             left, right, indices, self.leaf_size)
 * 
 */
  __pyx_t_9 = PyTuple_New(4); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 498, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_INCREF(((PyObject *)__pyx_v_left));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_left));
  PyTuple_SET_ITEM(__pyx_t_9, 0, ((PyObject *)__pyx_v_left));
  __Pyx_INCREF(((PyObject *)__pyx_v_right));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_right));
  PyTuple_SET_ITEM(__pyx_t_9, 1, ((PyObject *)__pyx_v_right));
  __Pyx_INCREF(((PyObject *)__pyx_v_indices));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_indices));
  PyTuple_SET_ITEM(__pyx_t_9, 2, ((PyObject *)__pyx_v_indices));
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_9, 3, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode), __pyx_t_9, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 498, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pandas/_libs/intervaltree.pxi":488
 *                 overlapping_ind.to_array())
 * 
 *     cdef new_child_node(self,             # <<<<<<<<<<<<<<
 *                         ndarray[float32_t, ndim=1] left,
 *                         ndarray[float32_t, ndim=1] right,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_9);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indices.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_left.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_right.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_subset.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("pandas._libs.interval.Float32ClosedRightIntervalNode.new_child_node", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indices.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_left.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_right.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_subset.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF((PyObject *)__pyx_v_left);
  __Pyx_XDECREF((PyObject *)__pyx_v_right);
  __Pyx_XDECREF((PyObject *)__pyx_v_indices);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":504
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_2query[] = "Recursively query this node and its sub-nodes for intervals that\n        overlap with the query point.\n        ";
static PyMethodDef __pyx_mdef_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_3query = {"query", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_3query, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_2query};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_args = 0;
  PyObject *__pyx_v_kwargs = 0;
  CYTHON_UNUSED PyObject *__pyx_v_defaults = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fused_cpdef (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_args,&__pyx_n_s_kwargs,&__pyx_n_s_defaults,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_args)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_kwargs)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fused_cpdef", 1, 3, 3, 1); __PYX_ERR(1, 504, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_defaults)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fused_cpdef", 1, 3, 3, 2); __PYX_ERR(1, 504, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fused_cpdef") < 0)) __PYX_ERR(1, 504, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_args = values[0];
    __pyx_v_kwargs = values[1];
    __pyx_v_defaults = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fused_cpdef", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 504, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Float32ClosedRightIntervalNode.__pyx_fused_cpdef", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_2query(((struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode *)__pyx_v_signatures), __pyx_v_args, __pyx_v_kwargs, __pyx_v_defaults);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_2query(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode *__pyx_v_signatures, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs, CYTHON_UNUSED PyObject *__pyx_v_defaults) {
  PyObject *__pyx_v_dest_sig = NULL;
  Py_ssize_t __pyx_v_i;
  CYTHON_UNUSED PyTypeObject *__pyx_v_ndarray = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_v_itemsize;
  PyObject *__pyx_v_arg = NULL;
  PyObject *__pyx_v_candidates = NULL;
  PyObject *__pyx_v_sig = NULL;
  int __pyx_v_match_found;
  PyObject *__pyx_v_src_sig = NULL;
  PyObject *__pyx_v_dst_type = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  Py_ssize_t __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  Py_ssize_t __pyx_t_13;
  Py_ssize_t __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  int __pyx_t_16;
  __Pyx_RefNannySetupContext("query", 0);
  __Pyx_INCREF(__pyx_v_kwargs);
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 504, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  PyList_SET_ITEM(__pyx_t_1, 0, Py_None);
  __pyx_v_dest_sig = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_3 = (__pyx_v_kwargs != Py_None);
  __pyx_t_4 = (__pyx_t_3 != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_kwargs); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(1, 504, __pyx_L1_error)
  __pyx_t_3 = ((!__pyx_t_4) != 0);
  __pyx_t_2 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {
    __Pyx_INCREF(Py_None);
    __Pyx_DECREF_SET(__pyx_v_kwargs, Py_None);
  }
  __pyx_t_1 = ((PyObject *)__Pyx_ImportNumPyArrayTypeIfAvailable()); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 504, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_ndarray = ((PyTypeObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_v_itemsize = -1L;
  if (unlikely(__pyx_v_args == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 504, __pyx_L1_error)
  }
  __pyx_t_5 = PyTuple_GET_SIZE(((PyObject*)__pyx_v_args)); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(1, 504, __pyx_L1_error)
  __pyx_t_2 = ((2 < __pyx_t_5) != 0);
  if (__pyx_t_2) {
    if (unlikely(__pyx_v_args == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 504, __pyx_L1_error)
    }
    __pyx_t_1 = PyTuple_GET_ITEM(((PyObject*)__pyx_v_args), 2);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v_arg = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L6;
  }
  __pyx_t_3 = (__pyx_v_kwargs != Py_None);
  __pyx_t_4 = (__pyx_t_3 != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L7_bool_binop_done;
  }
  if (unlikely(__pyx_v_kwargs == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(1, 504, __pyx_L1_error)
  }
  __pyx_t_4 = (__Pyx_PyDict_ContainsTF(__pyx_n_s_point, ((PyObject*)__pyx_v_kwargs), Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(1, 504, __pyx_L1_error)
  __pyx_t_3 = (__pyx_t_4 != 0);
  __pyx_t_2 = __pyx_t_3;
  __pyx_L7_bool_binop_done:;
  if (__pyx_t_2) {
    if (unlikely(__pyx_v_kwargs == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 504, __pyx_L1_error)
    }
    __pyx_t_1 = __Pyx_PyDict_GetItem(((PyObject*)__pyx_v_kwargs), __pyx_n_s_point); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 504, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_arg = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L6;
  }
  /*else*/ {
    if (unlikely(__pyx_v_args == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
      __PYX_ERR(1, 504, __pyx_L1_error)
    }
    __pyx_t_5 = PyTuple_GET_SIZE(((PyObject*)__pyx_v_args)); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(1, 504, __pyx_L1_error)
    __pyx_t_1 = PyInt_FromSsize_t(__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 504, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_6 = PyTuple_New(3); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 504, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_INCREF(__pyx_int_3);
    __Pyx_GIVEREF(__pyx_int_3);
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_int_3);
    __Pyx_INCREF(__pyx_n_s_s);
    __Pyx_GIVEREF(__pyx_n_s_s);
    PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_n_s_s);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Expected_at_least_d_argument_s_g, __pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 504, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_1); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 504, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(1, 504, __pyx_L1_error)
  }
  __pyx_L6:;
  while (1) {
    __pyx_t_2 = PyFloat_Check(__pyx_v_arg); 
    __pyx_t_3 = (__pyx_t_2 != 0);
    if (__pyx_t_3) {
      if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, __pyx_n_s_float64_t, long, 1, __Pyx_PyInt_From_long, 1, 0, 0) < 0)) __PYX_ERR(1, 504, __pyx_L1_error)
      goto __pyx_L10_break;
    }
    __pyx_t_2 = PyInt_Check(__pyx_v_arg); 
    __pyx_t_4 = (__pyx_t_2 != 0);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_3 = __pyx_t_4;
      goto __pyx_L13_bool_binop_done;
    }
    __pyx_t_4 = PyLong_Check(__pyx_v_arg); 
    __pyx_t_2 = (__pyx_t_4 != 0);
    __pyx_t_3 = __pyx_t_2;
    __pyx_L13_bool_binop_done:;
    __pyx_t_2 = (__pyx_t_3 != 0);
    if (__pyx_t_2) {
      if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, __pyx_n_s_int64_t, long, 1, __Pyx_PyInt_From_long, 1, 0, 0) < 0)) __PYX_ERR(1, 504, __pyx_L1_error)
      goto __pyx_L10_break;
    }
    if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, Py_None, long, 1, __Pyx_PyInt_From_long, 1, 0, 0) < 0)) __PYX_ERR(1, 504, __pyx_L1_error)
    goto __pyx_L10_break;
  }
  __pyx_L10_break:;
  __pyx_t_6 = PyList_New(0); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 504, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_v_candidates = ((PyObject*)__pyx_t_6);
  __pyx_t_6 = 0;
  __pyx_t_5 = 0;
  if (unlikely(((PyObject *)__pyx_v_signatures) == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(1, 504, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_dict_iterator(((PyObject*)__pyx_v_signatures), 1, ((PyObject *)NULL), (&__pyx_t_7), (&__pyx_t_8)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 504, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __pyx_t_6 = __pyx_t_1;
  __pyx_t_1 = 0;
  while (1) {
    __pyx_t_9 = __Pyx_dict_iter_next(__pyx_t_6, __pyx_t_7, &__pyx_t_5, &__pyx_t_1, NULL, NULL, __pyx_t_8);
    if (unlikely(__pyx_t_9 == 0)) break;
    if (unlikely(__pyx_t_9 == -1)) __PYX_ERR(1, 504, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_sig, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_v_match_found = 0;
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_v_sig, __pyx_n_s_strip); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 504, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __pyx_t_12 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_11))) {
      __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_11);
      if (likely(__pyx_t_12)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
        __Pyx_INCREF(__pyx_t_12);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_11, function);
      }
    }
    __pyx_t_10 = (__pyx_t_12) ? __Pyx_PyObject_Call2Args(__pyx_t_11, __pyx_t_12, __pyx_kp_s__13) : __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_kp_s__13);
    __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
    if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 504, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_split); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 504, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_t_10 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_11))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_11);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_11, function);
      }
    }
    __pyx_t_1 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_11, __pyx_t_10, __pyx_kp_s__14) : __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_kp_s__14);
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 504, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_XDECREF_SET(__pyx_v_src_sig, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_13 = PyList_GET_SIZE(__pyx_v_dest_sig); if (unlikely(__pyx_t_13 == ((Py_ssize_t)-1))) __PYX_ERR(1, 504, __pyx_L1_error)
    __pyx_t_14 = __pyx_t_13;
    for (__pyx_t_15 = 0; __pyx_t_15 < __pyx_t_14; __pyx_t_15+=1) {
      __pyx_v_i = __pyx_t_15;
      __pyx_t_1 = PyList_GET_ITEM(__pyx_v_dest_sig, __pyx_v_i);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_XDECREF_SET(__pyx_v_dst_type, __pyx_t_1);
      __pyx_t_1 = 0;
      __pyx_t_2 = (__pyx_v_dst_type != Py_None);
      __pyx_t_3 = (__pyx_t_2 != 0);
      if (__pyx_t_3) {
        __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_src_sig, __pyx_v_i, Py_ssize_t, 1, PyInt_FromSsize_t, 0, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 504, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_11 = PyObject_RichCompare(__pyx_t_1, __pyx_v_dst_type, Py_EQ); __Pyx_XGOTREF(__pyx_t_11); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 504, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_11); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(1, 504, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        if (__pyx_t_3) {
          __pyx_v_match_found = 1;
          goto __pyx_L20;
        }
        /*else*/ {
          __pyx_v_match_found = 0;
          goto __pyx_L18_break;
        }
        __pyx_L20:;
      }
    }
    __pyx_L18_break:;
    __pyx_t_3 = (__pyx_v_match_found != 0);
    if (__pyx_t_3) {
      __pyx_t_16 = __Pyx_PyList_Append(__pyx_v_candidates, __pyx_v_sig); if (unlikely(__pyx_t_16 == ((int)-1))) __PYX_ERR(1, 504, __pyx_L1_error)
    }
  }
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_3 = (PyList_GET_SIZE(__pyx_v_candidates) != 0);
  __pyx_t_2 = ((!__pyx_t_3) != 0);
  if (__pyx_t_2) {
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__15, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 504, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(1, 504, __pyx_L1_error)
  }
  __pyx_t_7 = PyList_GET_SIZE(__pyx_v_candidates); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1))) __PYX_ERR(1, 504, __pyx_L1_error)
  __pyx_t_2 = ((__pyx_t_7 > 1) != 0);
  if (__pyx_t_2) {
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__16, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 504, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(1, 504, __pyx_L1_error)
  }
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    if (unlikely(((PyObject *)__pyx_v_signatures) == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 504, __pyx_L1_error)
    }
    __pyx_t_6 = __Pyx_PyDict_GetItem(((PyObject*)__pyx_v_signatures), PyList_GET_ITEM(__pyx_v_candidates, 0)); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 504, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_r = __pyx_t_6;
    __pyx_t_6 = 0;
    goto __pyx_L0;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("pandas._libs.interval.Float32ClosedRightIntervalNode.__pyx_fused_cpdef", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_dest_sig);
  __Pyx_XDECREF(__pyx_v_ndarray);
  __Pyx_XDECREF(__pyx_v_arg);
  __Pyx_XDECREF(__pyx_v_candidates);
  __Pyx_XDECREF(__pyx_v_sig);
  __Pyx_XDECREF(__pyx_v_src_sig);
  __Pyx_XDECREF(__pyx_v_dst_type);
  __Pyx_XDECREF(__pyx_v_kwargs);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_9__pyx_fuse_0query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_fuse_0__pyx_f_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float64_t __pyx_v_point, int __pyx_skip_dispatch) {
  __Pyx_memviewslice __pyx_v_indices = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_values = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __pyx_t_5numpy_int64_t __pyx_t_9;
  __pyx_t_5numpy_int64_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  int __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  __Pyx_memviewslice __pyx_t_16 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_17 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  Py_ssize_t __pyx_t_21;
  __Pyx_RefNannySetupContext("__pyx_fuse_0query", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pyx_fuse_0query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 504, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_9__pyx_fuse_0query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = PyFloat_FromDouble(__pyx_v_point); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 504, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 504, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 504, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 504, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_5) {
            __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 504, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 504, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_3query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_4 = PyFloat_FromDouble(__pyx_v_point); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 504, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_7 = __pyx_t_1; __pyx_t_3 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 504, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 504, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 504, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (__pyx_t_3) {
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_t_4);
          __pyx_t_4 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 504, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pandas/_libs/intervaltree.pxi":513
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
  __pyx_t_8 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_8) {

    /* "pandas/_libs/intervaltree.pxi":517
 *             # continue the binary tree structure. Instead, we use linear
 *             # search.
 *             for i in range(self.n_elements):             # <<<<<<<<<<<<<<
 *                 if self.left[i] < point <= self.right[i]:
 *                     result.append(self.indices[i])
 */
    __pyx_t_9 = __pyx_v_self->n_elements;
    __pyx_t_10 = __pyx_t_9;
    for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
      __pyx_v_i = __pyx_t_11;

      /* "pandas/_libs/intervaltree.pxi":518
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point <= self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      __pyx_t_12 = __pyx_v_i;
      __pyx_t_8 = ((*((__pyx_t_5numpy_float32_t *) ( /* dim=0 */ (__pyx_v_self->left.data + __pyx_t_12 * __pyx_v_self->left.strides[0]) ))) < __pyx_v_point);
      if (__pyx_t_8) {
        __pyx_t_13 = __pyx_v_i;
        __pyx_t_8 = (__pyx_v_point <= (*((__pyx_t_5numpy_float32_t *) ( /* dim=0 */ (__pyx_v_self->right.data + __pyx_t_13 * __pyx_v_self->right.strides[0]) ))));
      }
      __pyx_t_14 = (__pyx_t_8 != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":519
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point <= self.right[i]:
 *                     result.append(self.indices[i])             # <<<<<<<<<<<<<<
 *         else:
 *             # There are child nodes. Based on comparing our query to the pivot,
 */
        __pyx_t_15 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->indices.data + __pyx_t_15 * __pyx_v_self->indices.strides[0]) ))));

        /* "pandas/_libs/intervaltree.pxi":518
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point <= self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      }
    }

    /* "pandas/_libs/intervaltree.pxi":513
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
    goto __pyx_L3;
  }

  /* "pandas/_libs/intervaltree.pxi":523
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
  /*else*/ {
    __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":524
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:
 *                 values = self.center_left_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 */
      __pyx_t_16 = __pyx_v_self->center_left_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":525
 *             if point < self.pivot:
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:
 */
      __pyx_t_17 = __pyx_v_self->center_left_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":526
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):             # <<<<<<<<<<<<<<
 *                     if not values[i] < point:
 *                         break
 */
      __pyx_t_9 = __pyx_v_self->n_center;
      __pyx_t_10 = __pyx_t_9;
      for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":527
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_18 = __pyx_v_i;
        __pyx_t_14 = ((!(((*((__pyx_t_5numpy_float32_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_18 * __pyx_v_values.strides[0]) ))) < __pyx_v_point) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":528
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:
 */
          goto __pyx_L9_break;

          /* "pandas/_libs/intervaltree.pxi":527
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":529
 *                     if not values[i] < point:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 */
        __pyx_t_19 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_19 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L9_break:;

      /* "pandas/_libs/intervaltree.pxi":530
 *                         break
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      __pyx_t_14 = ((__pyx_v_point <= __pyx_v_self->left_node->max_right) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":531
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)             # <<<<<<<<<<<<<<
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode *)__pyx_v_self->left_node->__pyx_vtab)->__pyx_fuse_0query(__pyx_v_self->left_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 531, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":530
 *                         break
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":523
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":532
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
    __pyx_t_14 = ((__pyx_v_point > __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":533
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 *                 values = self.center_right_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 */
      __pyx_t_16 = __pyx_v_self->center_right_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":534
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:
 */
      __pyx_t_17 = __pyx_v_self->center_right_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":535
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):             # <<<<<<<<<<<<<<
 *                     if not point <= values[i]:
 *                         break
 */
      for (__pyx_t_11 = (__pyx_v_self->n_center - 1); __pyx_t_11 > -1L; __pyx_t_11-=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":536
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_20 = __pyx_v_i;
        __pyx_t_14 = ((!((__pyx_v_point <= (*((__pyx_t_5numpy_float32_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_20 * __pyx_v_values.strides[0]) )))) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":537
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:
 */
          goto __pyx_L13_break;

          /* "pandas/_libs/intervaltree.pxi":536
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":538
 *                     if not point <= values[i]:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if self.right_node.min_left < point:
 *                     self.right_node.query(result, point)
 */
        __pyx_t_21 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_21 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L13_break:;

      /* "pandas/_libs/intervaltree.pxi":539
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      __pyx_t_14 = ((__pyx_v_self->right_node->min_left < __pyx_v_point) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":540
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:
 *                     self.right_node.query(result, point)             # <<<<<<<<<<<<<<
 *             else:
 *                 result.extend(self.center_left_indices)
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode *)__pyx_v_self->right_node->__pyx_vtab)->__pyx_fuse_0query(__pyx_v_self->right_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 540, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":539
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":532
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":542
 *                     self.right_node.query(result, point)
 *             else:
 *                 result.extend(self.center_left_indices)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
    /*else*/ {
      __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->extend(__pyx_v_result, __pyx_v_self->center_left_indices); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 542, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_L7:;
  }
  __pyx_L3:;

  /* "pandas/_libs/intervaltree.pxi":504
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __PYX_XDEC_MEMVIEW(&__pyx_t_16, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_17, 1);
  __Pyx_AddTraceback("pandas._libs.interval.Float32ClosedRightIntervalNode.query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_indices, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_values, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_9__pyx_fuse_0query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_0__pyx_mdef_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_9__pyx_fuse_0query = {"__pyx_fuse_0query", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_9__pyx_fuse_0query, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_2query};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_9__pyx_fuse_0query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result = 0;
  __pyx_t_5numpy_float64_t __pyx_v_point;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fuse_0query (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_result,&__pyx_n_s_point,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_result)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_point)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_0query", 1, 2, 2, 1); __PYX_ERR(1, 504, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_0query") < 0)) __PYX_ERR(1, 504, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_result = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)values[0]);
    __pyx_v_point = __pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_point == ((npy_float64)-1)) && PyErr_Occurred())) __PYX_ERR(1, 504, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_0query", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 504, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Float32ClosedRightIntervalNode.__pyx_fuse_0query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_result), __pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector, 1, "result", 0))) __PYX_ERR(1, 504, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_8__pyx_fuse_0query(((struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode *)__pyx_v_self), __pyx_v_result, __pyx_v_point);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_8__pyx_fuse_0query(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float64_t __pyx_v_point) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__pyx_fuse_0query", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_fuse_0__pyx_f_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_query(__pyx_v_self, __pyx_v_result, __pyx_v_point, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 504, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Float32ClosedRightIntervalNode.__pyx_fuse_0query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_11__pyx_fuse_1query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_fuse_1__pyx_f_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float32_t __pyx_v_point, int __pyx_skip_dispatch) {
  __Pyx_memviewslice __pyx_v_indices = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_values = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __pyx_t_5numpy_int64_t __pyx_t_9;
  __pyx_t_5numpy_int64_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  int __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  __Pyx_memviewslice __pyx_t_16 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_17 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  Py_ssize_t __pyx_t_21;
  __Pyx_RefNannySetupContext("__pyx_fuse_1query", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pyx_fuse_1query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 504, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_11__pyx_fuse_1query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = PyFloat_FromDouble(__pyx_v_point); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 504, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 504, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 504, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 504, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_5) {
            __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 504, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 504, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_3query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_4 = PyFloat_FromDouble(__pyx_v_point); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 504, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_7 = __pyx_t_1; __pyx_t_3 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 504, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 504, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 504, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (__pyx_t_3) {
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_t_4);
          __pyx_t_4 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 504, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pandas/_libs/intervaltree.pxi":513
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
  __pyx_t_8 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_8) {

    /* "pandas/_libs/intervaltree.pxi":517
 *             # continue the binary tree structure. Instead, we use linear
 *             # search.
 *             for i in range(self.n_elements):             # <<<<<<<<<<<<<<
 *                 if self.left[i] < point <= self.right[i]:
 *                     result.append(self.indices[i])
 */
    __pyx_t_9 = __pyx_v_self->n_elements;
    __pyx_t_10 = __pyx_t_9;
    for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
      __pyx_v_i = __pyx_t_11;

      /* "pandas/_libs/intervaltree.pxi":518
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point <= self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      __pyx_t_12 = __pyx_v_i;
      __pyx_t_8 = ((*((__pyx_t_5numpy_float32_t *) ( /* dim=0 */ (__pyx_v_self->left.data + __pyx_t_12 * __pyx_v_self->left.strides[0]) ))) < __pyx_v_point);
      if (__pyx_t_8) {
        __pyx_t_13 = __pyx_v_i;
        __pyx_t_8 = (__pyx_v_point <= (*((__pyx_t_5numpy_float32_t *) ( /* dim=0 */ (__pyx_v_self->right.data + __pyx_t_13 * __pyx_v_self->right.strides[0]) ))));
      }
      __pyx_t_14 = (__pyx_t_8 != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":519
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point <= self.right[i]:
 *                     result.append(self.indices[i])             # <<<<<<<<<<<<<<
 *         else:
 *             # There are child nodes. Based on comparing our query to the pivot,
 */
        __pyx_t_15 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->indices.data + __pyx_t_15 * __pyx_v_self->indices.strides[0]) ))));

        /* "pandas/_libs/intervaltree.pxi":518
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point <= self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      }
    }

    /* "pandas/_libs/intervaltree.pxi":513
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
    goto __pyx_L3;
  }

  /* "pandas/_libs/intervaltree.pxi":523
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
  /*else*/ {
    __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":524
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:
 *                 values = self.center_left_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 */
      __pyx_t_16 = __pyx_v_self->center_left_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":525
 *             if point < self.pivot:
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:
 */
      __pyx_t_17 = __pyx_v_self->center_left_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":526
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):             # <<<<<<<<<<<<<<
 *                     if not values[i] < point:
 *                         break
 */
      __pyx_t_9 = __pyx_v_self->n_center;
      __pyx_t_10 = __pyx_t_9;
      for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":527
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_18 = __pyx_v_i;
        __pyx_t_14 = ((!(((*((__pyx_t_5numpy_float32_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_18 * __pyx_v_values.strides[0]) ))) < __pyx_v_point) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":528
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:
 */
          goto __pyx_L9_break;

          /* "pandas/_libs/intervaltree.pxi":527
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":529
 *                     if not values[i] < point:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 */
        __pyx_t_19 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_19 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L9_break:;

      /* "pandas/_libs/intervaltree.pxi":530
 *                         break
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      __pyx_t_14 = ((__pyx_v_point <= __pyx_v_self->left_node->max_right) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":531
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)             # <<<<<<<<<<<<<<
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode *)__pyx_v_self->left_node->__pyx_vtab)->__pyx_fuse_1query(__pyx_v_self->left_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 531, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":530
 *                         break
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":523
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":532
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
    __pyx_t_14 = ((__pyx_v_point > __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":533
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 *                 values = self.center_right_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 */
      __pyx_t_16 = __pyx_v_self->center_right_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":534
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:
 */
      __pyx_t_17 = __pyx_v_self->center_right_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":535
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):             # <<<<<<<<<<<<<<
 *                     if not point <= values[i]:
 *                         break
 */
      for (__pyx_t_11 = (__pyx_v_self->n_center - 1); __pyx_t_11 > -1L; __pyx_t_11-=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":536
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_20 = __pyx_v_i;
        __pyx_t_14 = ((!((__pyx_v_point <= (*((__pyx_t_5numpy_float32_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_20 * __pyx_v_values.strides[0]) )))) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":537
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:
 */
          goto __pyx_L13_break;

          /* "pandas/_libs/intervaltree.pxi":536
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":538
 *                     if not point <= values[i]:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if self.right_node.min_left < point:
 *                     self.right_node.query(result, point)
 */
        __pyx_t_21 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_21 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L13_break:;

      /* "pandas/_libs/intervaltree.pxi":539
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      __pyx_t_14 = ((__pyx_v_self->right_node->min_left < __pyx_v_point) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":540
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:
 *                     self.right_node.query(result, point)             # <<<<<<<<<<<<<<
 *             else:
 *                 result.extend(self.center_left_indices)
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode *)__pyx_v_self->right_node->__pyx_vtab)->__pyx_fuse_1query(__pyx_v_self->right_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 540, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":539
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":532
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":542
 *                     self.right_node.query(result, point)
 *             else:
 *                 result.extend(self.center_left_indices)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
    /*else*/ {
      __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->extend(__pyx_v_result, __pyx_v_self->center_left_indices); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 542, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_L7:;
  }
  __pyx_L3:;

  /* "pandas/_libs/intervaltree.pxi":504
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __PYX_XDEC_MEMVIEW(&__pyx_t_16, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_17, 1);
  __Pyx_AddTraceback("pandas._libs.interval.Float32ClosedRightIntervalNode.query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_indices, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_values, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_11__pyx_fuse_1query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_1__pyx_mdef_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_11__pyx_fuse_1query = {"__pyx_fuse_1query", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_11__pyx_fuse_1query, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_2query};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_11__pyx_fuse_1query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result = 0;
  __pyx_t_5numpy_float32_t __pyx_v_point;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fuse_1query (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_result,&__pyx_n_s_point,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_result)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_point)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_1query", 1, 2, 2, 1); __PYX_ERR(1, 504, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_1query") < 0)) __PYX_ERR(1, 504, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_result = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)values[0]);
    __pyx_v_point = __pyx_PyFloat_AsFloat(values[1]); if (unlikely((__pyx_v_point == ((npy_float32)-1)) && PyErr_Occurred())) __PYX_ERR(1, 504, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_1query", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 504, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Float32ClosedRightIntervalNode.__pyx_fuse_1query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_result), __pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector, 1, "result", 0))) __PYX_ERR(1, 504, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_10__pyx_fuse_1query(((struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode *)__pyx_v_self), __pyx_v_result, __pyx_v_point);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_10__pyx_fuse_1query(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float32_t __pyx_v_point) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__pyx_fuse_1query", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_fuse_1__pyx_f_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_query(__pyx_v_self, __pyx_v_result, __pyx_v_point, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 504, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Float32ClosedRightIntervalNode.__pyx_fuse_1query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_13__pyx_fuse_2query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_fuse_2__pyx_f_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int64_t __pyx_v_point, int __pyx_skip_dispatch) {
  __Pyx_memviewslice __pyx_v_indices = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_values = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __pyx_t_5numpy_int64_t __pyx_t_9;
  __pyx_t_5numpy_int64_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  int __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  __Pyx_memviewslice __pyx_t_16 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_17 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  Py_ssize_t __pyx_t_21;
  __Pyx_RefNannySetupContext("__pyx_fuse_2query", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pyx_fuse_2query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 504, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_13__pyx_fuse_2query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = __Pyx_PyInt_From_npy_int64(__pyx_v_point); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 504, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 504, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 504, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 504, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_5) {
            __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 504, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 504, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_3query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_4 = __Pyx_PyInt_From_npy_int64(__pyx_v_point); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 504, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_7 = __pyx_t_1; __pyx_t_3 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 504, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 504, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 504, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (__pyx_t_3) {
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_t_4);
          __pyx_t_4 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 504, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pandas/_libs/intervaltree.pxi":513
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
  __pyx_t_8 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_8) {

    /* "pandas/_libs/intervaltree.pxi":517
 *             # continue the binary tree structure. Instead, we use linear
 *             # search.
 *             for i in range(self.n_elements):             # <<<<<<<<<<<<<<
 *                 if self.left[i] < point <= self.right[i]:
 *                     result.append(self.indices[i])
 */
    __pyx_t_9 = __pyx_v_self->n_elements;
    __pyx_t_10 = __pyx_t_9;
    for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
      __pyx_v_i = __pyx_t_11;

      /* "pandas/_libs/intervaltree.pxi":518
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point <= self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      __pyx_t_12 = __pyx_v_i;
      __pyx_t_8 = ((*((__pyx_t_5numpy_float32_t *) ( /* dim=0 */ (__pyx_v_self->left.data + __pyx_t_12 * __pyx_v_self->left.strides[0]) ))) < __pyx_v_point);
      if (__pyx_t_8) {
        __pyx_t_13 = __pyx_v_i;
        __pyx_t_8 = (__pyx_v_point <= (*((__pyx_t_5numpy_float32_t *) ( /* dim=0 */ (__pyx_v_self->right.data + __pyx_t_13 * __pyx_v_self->right.strides[0]) ))));
      }
      __pyx_t_14 = (__pyx_t_8 != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":519
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point <= self.right[i]:
 *                     result.append(self.indices[i])             # <<<<<<<<<<<<<<
 *         else:
 *             # There are child nodes. Based on comparing our query to the pivot,
 */
        __pyx_t_15 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->indices.data + __pyx_t_15 * __pyx_v_self->indices.strides[0]) ))));

        /* "pandas/_libs/intervaltree.pxi":518
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point <= self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      }
    }

    /* "pandas/_libs/intervaltree.pxi":513
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
    goto __pyx_L3;
  }

  /* "pandas/_libs/intervaltree.pxi":523
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
  /*else*/ {
    __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":524
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:
 *                 values = self.center_left_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 */
      __pyx_t_16 = __pyx_v_self->center_left_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":525
 *             if point < self.pivot:
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:
 */
      __pyx_t_17 = __pyx_v_self->center_left_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":526
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):             # <<<<<<<<<<<<<<
 *                     if not values[i] < point:
 *                         break
 */
      __pyx_t_9 = __pyx_v_self->n_center;
      __pyx_t_10 = __pyx_t_9;
      for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":527
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_18 = __pyx_v_i;
        __pyx_t_14 = ((!(((*((__pyx_t_5numpy_float32_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_18 * __pyx_v_values.strides[0]) ))) < __pyx_v_point) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":528
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:
 */
          goto __pyx_L9_break;

          /* "pandas/_libs/intervaltree.pxi":527
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":529
 *                     if not values[i] < point:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 */
        __pyx_t_19 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_19 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L9_break:;

      /* "pandas/_libs/intervaltree.pxi":530
 *                         break
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      __pyx_t_14 = ((__pyx_v_point <= __pyx_v_self->left_node->max_right) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":531
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)             # <<<<<<<<<<<<<<
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode *)__pyx_v_self->left_node->__pyx_vtab)->__pyx_fuse_2query(__pyx_v_self->left_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 531, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":530
 *                         break
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":523
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":532
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
    __pyx_t_14 = ((__pyx_v_point > __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":533
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 *                 values = self.center_right_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 */
      __pyx_t_16 = __pyx_v_self->center_right_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":534
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:
 */
      __pyx_t_17 = __pyx_v_self->center_right_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":535
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):             # <<<<<<<<<<<<<<
 *                     if not point <= values[i]:
 *                         break
 */
      for (__pyx_t_11 = (__pyx_v_self->n_center - 1); __pyx_t_11 > -1L; __pyx_t_11-=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":536
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_20 = __pyx_v_i;
        __pyx_t_14 = ((!((__pyx_v_point <= (*((__pyx_t_5numpy_float32_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_20 * __pyx_v_values.strides[0]) )))) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":537
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:
 */
          goto __pyx_L13_break;

          /* "pandas/_libs/intervaltree.pxi":536
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":538
 *                     if not point <= values[i]:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if self.right_node.min_left < point:
 *                     self.right_node.query(result, point)
 */
        __pyx_t_21 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_21 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L13_break:;

      /* "pandas/_libs/intervaltree.pxi":539
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      __pyx_t_14 = ((__pyx_v_self->right_node->min_left < __pyx_v_point) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":540
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:
 *                     self.right_node.query(result, point)             # <<<<<<<<<<<<<<
 *             else:
 *                 result.extend(self.center_left_indices)
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode *)__pyx_v_self->right_node->__pyx_vtab)->__pyx_fuse_2query(__pyx_v_self->right_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 540, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":539
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":532
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":542
 *                     self.right_node.query(result, point)
 *             else:
 *                 result.extend(self.center_left_indices)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
    /*else*/ {
      __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->extend(__pyx_v_result, __pyx_v_self->center_left_indices); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 542, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_L7:;
  }
  __pyx_L3:;

  /* "pandas/_libs/intervaltree.pxi":504
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __PYX_XDEC_MEMVIEW(&__pyx_t_16, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_17, 1);
  __Pyx_AddTraceback("pandas._libs.interval.Float32ClosedRightIntervalNode.query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_indices, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_values, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_13__pyx_fuse_2query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_2__pyx_mdef_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_13__pyx_fuse_2query = {"__pyx_fuse_2query", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_13__pyx_fuse_2query, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_2query};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_13__pyx_fuse_2query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result = 0;
  __pyx_t_5numpy_int64_t __pyx_v_point;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fuse_2query (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_result,&__pyx_n_s_point,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_result)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_point)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_2query", 1, 2, 2, 1); __PYX_ERR(1, 504, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_2query") < 0)) __PYX_ERR(1, 504, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_result = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)values[0]);
    __pyx_v_point = __Pyx_PyInt_As_npy_int64(values[1]); if (unlikely((__pyx_v_point == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(1, 504, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_2query", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 504, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Float32ClosedRightIntervalNode.__pyx_fuse_2query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_result), __pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector, 1, "result", 0))) __PYX_ERR(1, 504, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_12__pyx_fuse_2query(((struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode *)__pyx_v_self), __pyx_v_result, __pyx_v_point);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_12__pyx_fuse_2query(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int64_t __pyx_v_point) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__pyx_fuse_2query", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_fuse_2__pyx_f_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_query(__pyx_v_self, __pyx_v_result, __pyx_v_point, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 504, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Float32ClosedRightIntervalNode.__pyx_fuse_2query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_15__pyx_fuse_3query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_fuse_3__pyx_f_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int32_t __pyx_v_point, int __pyx_skip_dispatch) {
  __Pyx_memviewslice __pyx_v_indices = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_values = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __pyx_t_5numpy_int64_t __pyx_t_9;
  __pyx_t_5numpy_int64_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  int __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  __Pyx_memviewslice __pyx_t_16 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_17 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  Py_ssize_t __pyx_t_21;
  __Pyx_RefNannySetupContext("__pyx_fuse_3query", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pyx_fuse_3query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 504, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_15__pyx_fuse_3query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = __Pyx_PyInt_From_npy_int32(__pyx_v_point); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 504, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 504, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 504, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 504, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_5) {
            __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 504, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 504, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_3query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_4 = __Pyx_PyInt_From_npy_int32(__pyx_v_point); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 504, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_7 = __pyx_t_1; __pyx_t_3 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 504, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 504, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 504, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (__pyx_t_3) {
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_t_4);
          __pyx_t_4 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 504, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pandas/_libs/intervaltree.pxi":513
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
  __pyx_t_8 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_8) {

    /* "pandas/_libs/intervaltree.pxi":517
 *             # continue the binary tree structure. Instead, we use linear
 *             # search.
 *             for i in range(self.n_elements):             # <<<<<<<<<<<<<<
 *                 if self.left[i] < point <= self.right[i]:
 *                     result.append(self.indices[i])
 */
    __pyx_t_9 = __pyx_v_self->n_elements;
    __pyx_t_10 = __pyx_t_9;
    for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
      __pyx_v_i = __pyx_t_11;

      /* "pandas/_libs/intervaltree.pxi":518
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point <= self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      __pyx_t_12 = __pyx_v_i;
      __pyx_t_8 = ((*((__pyx_t_5numpy_float32_t *) ( /* dim=0 */ (__pyx_v_self->left.data + __pyx_t_12 * __pyx_v_self->left.strides[0]) ))) < __pyx_v_point);
      if (__pyx_t_8) {
        __pyx_t_13 = __pyx_v_i;
        __pyx_t_8 = (__pyx_v_point <= (*((__pyx_t_5numpy_float32_t *) ( /* dim=0 */ (__pyx_v_self->right.data + __pyx_t_13 * __pyx_v_self->right.strides[0]) ))));
      }
      __pyx_t_14 = (__pyx_t_8 != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":519
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point <= self.right[i]:
 *                     result.append(self.indices[i])             # <<<<<<<<<<<<<<
 *         else:
 *             # There are child nodes. Based on comparing our query to the pivot,
 */
        __pyx_t_15 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->indices.data + __pyx_t_15 * __pyx_v_self->indices.strides[0]) ))));

        /* "pandas/_libs/intervaltree.pxi":518
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point <= self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      }
    }

    /* "pandas/_libs/intervaltree.pxi":513
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
    goto __pyx_L3;
  }

  /* "pandas/_libs/intervaltree.pxi":523
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
  /*else*/ {
    __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":524
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:
 *                 values = self.center_left_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 */
      __pyx_t_16 = __pyx_v_self->center_left_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":525
 *             if point < self.pivot:
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:
 */
      __pyx_t_17 = __pyx_v_self->center_left_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":526
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):             # <<<<<<<<<<<<<<
 *                     if not values[i] < point:
 *                         break
 */
      __pyx_t_9 = __pyx_v_self->n_center;
      __pyx_t_10 = __pyx_t_9;
      for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":527
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_18 = __pyx_v_i;
        __pyx_t_14 = ((!(((*((__pyx_t_5numpy_float32_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_18 * __pyx_v_values.strides[0]) ))) < __pyx_v_point) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":528
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:
 */
          goto __pyx_L9_break;

          /* "pandas/_libs/intervaltree.pxi":527
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":529
 *                     if not values[i] < point:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 */
        __pyx_t_19 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_19 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L9_break:;

      /* "pandas/_libs/intervaltree.pxi":530
 *                         break
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      __pyx_t_14 = ((__pyx_v_point <= __pyx_v_self->left_node->max_right) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":531
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)             # <<<<<<<<<<<<<<
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode *)__pyx_v_self->left_node->__pyx_vtab)->__pyx_fuse_3query(__pyx_v_self->left_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 531, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":530
 *                         break
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":523
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":532
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
    __pyx_t_14 = ((__pyx_v_point > __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":533
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 *                 values = self.center_right_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 */
      __pyx_t_16 = __pyx_v_self->center_right_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":534
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:
 */
      __pyx_t_17 = __pyx_v_self->center_right_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":535
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):             # <<<<<<<<<<<<<<
 *                     if not point <= values[i]:
 *                         break
 */
      for (__pyx_t_11 = (__pyx_v_self->n_center - 1); __pyx_t_11 > -1L; __pyx_t_11-=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":536
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_20 = __pyx_v_i;
        __pyx_t_14 = ((!((__pyx_v_point <= (*((__pyx_t_5numpy_float32_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_20 * __pyx_v_values.strides[0]) )))) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":537
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:
 */
          goto __pyx_L13_break;

          /* "pandas/_libs/intervaltree.pxi":536
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":538
 *                     if not point <= values[i]:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if self.right_node.min_left < point:
 *                     self.right_node.query(result, point)
 */
        __pyx_t_21 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_21 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L13_break:;

      /* "pandas/_libs/intervaltree.pxi":539
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      __pyx_t_14 = ((__pyx_v_self->right_node->min_left < __pyx_v_point) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":540
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:
 *                     self.right_node.query(result, point)             # <<<<<<<<<<<<<<
 *             else:
 *                 result.extend(self.center_left_indices)
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode *)__pyx_v_self->right_node->__pyx_vtab)->__pyx_fuse_3query(__pyx_v_self->right_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 540, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":539
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":532
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":542
 *                     self.right_node.query(result, point)
 *             else:
 *                 result.extend(self.center_left_indices)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
    /*else*/ {
      __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->extend(__pyx_v_result, __pyx_v_self->center_left_indices); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 542, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_L7:;
  }
  __pyx_L3:;

  /* "pandas/_libs/intervaltree.pxi":504
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __PYX_XDEC_MEMVIEW(&__pyx_t_16, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_17, 1);
  __Pyx_AddTraceback("pandas._libs.interval.Float32ClosedRightIntervalNode.query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_indices, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_values, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_15__pyx_fuse_3query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_3__pyx_mdef_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_15__pyx_fuse_3query = {"__pyx_fuse_3query", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_15__pyx_fuse_3query, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_2query};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_15__pyx_fuse_3query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result = 0;
  __pyx_t_5numpy_int32_t __pyx_v_point;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fuse_3query (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_result,&__pyx_n_s_point,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_result)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_point)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_3query", 1, 2, 2, 1); __PYX_ERR(1, 504, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_3query") < 0)) __PYX_ERR(1, 504, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_result = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)values[0]);
    __pyx_v_point = __Pyx_PyInt_As_npy_int32(values[1]); if (unlikely((__pyx_v_point == ((npy_int32)-1)) && PyErr_Occurred())) __PYX_ERR(1, 504, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_3query", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 504, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Float32ClosedRightIntervalNode.__pyx_fuse_3query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_result), __pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector, 1, "result", 0))) __PYX_ERR(1, 504, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_14__pyx_fuse_3query(((struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode *)__pyx_v_self), __pyx_v_result, __pyx_v_point);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_14__pyx_fuse_3query(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int32_t __pyx_v_point) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__pyx_fuse_3query", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_fuse_3__pyx_f_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_query(__pyx_v_self, __pyx_v_result, __pyx_v_point, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 504, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Float32ClosedRightIntervalNode.__pyx_fuse_3query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_17__pyx_fuse_4query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_fuse_4__pyx_f_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_uint64_t __pyx_v_point, int __pyx_skip_dispatch) {
  __Pyx_memviewslice __pyx_v_indices = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_values = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __pyx_t_5numpy_int64_t __pyx_t_9;
  __pyx_t_5numpy_int64_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  int __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  __Pyx_memviewslice __pyx_t_16 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_17 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  Py_ssize_t __pyx_t_21;
  __Pyx_RefNannySetupContext("__pyx_fuse_4query", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pyx_fuse_4query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 504, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_17__pyx_fuse_4query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = __Pyx_PyInt_From_npy_uint64(__pyx_v_point); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 504, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 504, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 504, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 504, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_5) {
            __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 504, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 504, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_3query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_4 = __Pyx_PyInt_From_npy_uint64(__pyx_v_point); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 504, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_7 = __pyx_t_1; __pyx_t_3 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 504, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 504, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 504, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (__pyx_t_3) {
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_t_4);
          __pyx_t_4 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 504, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pandas/_libs/intervaltree.pxi":513
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
  __pyx_t_8 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_8) {

    /* "pandas/_libs/intervaltree.pxi":517
 *             # continue the binary tree structure. Instead, we use linear
 *             # search.
 *             for i in range(self.n_elements):             # <<<<<<<<<<<<<<
 *                 if self.left[i] < point <= self.right[i]:
 *                     result.append(self.indices[i])
 */
    __pyx_t_9 = __pyx_v_self->n_elements;
    __pyx_t_10 = __pyx_t_9;
    for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
      __pyx_v_i = __pyx_t_11;

      /* "pandas/_libs/intervaltree.pxi":518
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point <= self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      __pyx_t_12 = __pyx_v_i;
      __pyx_t_8 = ((*((__pyx_t_5numpy_float32_t *) ( /* dim=0 */ (__pyx_v_self->left.data + __pyx_t_12 * __pyx_v_self->left.strides[0]) ))) < __pyx_v_point);
      if (__pyx_t_8) {
        __pyx_t_13 = __pyx_v_i;
        __pyx_t_8 = (__pyx_v_point <= (*((__pyx_t_5numpy_float32_t *) ( /* dim=0 */ (__pyx_v_self->right.data + __pyx_t_13 * __pyx_v_self->right.strides[0]) ))));
      }
      __pyx_t_14 = (__pyx_t_8 != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":519
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point <= self.right[i]:
 *                     result.append(self.indices[i])             # <<<<<<<<<<<<<<
 *         else:
 *             # There are child nodes. Based on comparing our query to the pivot,
 */
        __pyx_t_15 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->indices.data + __pyx_t_15 * __pyx_v_self->indices.strides[0]) ))));

        /* "pandas/_libs/intervaltree.pxi":518
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point <= self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      }
    }

    /* "pandas/_libs/intervaltree.pxi":513
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
    goto __pyx_L3;
  }

  /* "pandas/_libs/intervaltree.pxi":523
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
  /*else*/ {
    __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":524
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:
 *                 values = self.center_left_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 */
      __pyx_t_16 = __pyx_v_self->center_left_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":525
 *             if point < self.pivot:
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:
 */
      __pyx_t_17 = __pyx_v_self->center_left_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":526
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):             # <<<<<<<<<<<<<<
 *                     if not values[i] < point:
 *                         break
 */
      __pyx_t_9 = __pyx_v_self->n_center;
      __pyx_t_10 = __pyx_t_9;
      for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":527
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_18 = __pyx_v_i;
        __pyx_t_14 = ((!(((*((__pyx_t_5numpy_float32_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_18 * __pyx_v_values.strides[0]) ))) < __pyx_v_point) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":528
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:
 */
          goto __pyx_L9_break;

          /* "pandas/_libs/intervaltree.pxi":527
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":529
 *                     if not values[i] < point:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 */
        __pyx_t_19 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_19 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L9_break:;

      /* "pandas/_libs/intervaltree.pxi":530
 *                         break
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      __pyx_t_14 = ((__pyx_v_point <= __pyx_v_self->left_node->max_right) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":531
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)             # <<<<<<<<<<<<<<
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode *)__pyx_v_self->left_node->__pyx_vtab)->__pyx_fuse_4query(__pyx_v_self->left_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 531, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":530
 *                         break
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":523
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":532
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
    __pyx_t_14 = ((__pyx_v_point > __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":533
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 *                 values = self.center_right_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 */
      __pyx_t_16 = __pyx_v_self->center_right_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":534
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:
 */
      __pyx_t_17 = __pyx_v_self->center_right_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":535
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):             # <<<<<<<<<<<<<<
 *                     if not point <= values[i]:
 *                         break
 */
      for (__pyx_t_11 = (__pyx_v_self->n_center - 1); __pyx_t_11 > -1L; __pyx_t_11-=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":536
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_20 = __pyx_v_i;
        __pyx_t_14 = ((!((__pyx_v_point <= (*((__pyx_t_5numpy_float32_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_20 * __pyx_v_values.strides[0]) )))) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":537
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:
 */
          goto __pyx_L13_break;

          /* "pandas/_libs/intervaltree.pxi":536
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":538
 *                     if not point <= values[i]:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if self.right_node.min_left < point:
 *                     self.right_node.query(result, point)
 */
        __pyx_t_21 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_21 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L13_break:;

      /* "pandas/_libs/intervaltree.pxi":539
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      __pyx_t_14 = ((__pyx_v_self->right_node->min_left < __pyx_v_point) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":540
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:
 *                     self.right_node.query(result, point)             # <<<<<<<<<<<<<<
 *             else:
 *                 result.extend(self.center_left_indices)
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode *)__pyx_v_self->right_node->__pyx_vtab)->__pyx_fuse_4query(__pyx_v_self->right_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 540, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":539
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":532
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":542
 *                     self.right_node.query(result, point)
 *             else:
 *                 result.extend(self.center_left_indices)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
    /*else*/ {
      __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->extend(__pyx_v_result, __pyx_v_self->center_left_indices); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 542, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_L7:;
  }
  __pyx_L3:;

  /* "pandas/_libs/intervaltree.pxi":504
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __PYX_XDEC_MEMVIEW(&__pyx_t_16, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_17, 1);
  __Pyx_AddTraceback("pandas._libs.interval.Float32ClosedRightIntervalNode.query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_indices, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_values, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_17__pyx_fuse_4query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_4__pyx_mdef_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_17__pyx_fuse_4query = {"__pyx_fuse_4query", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_17__pyx_fuse_4query, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_2query};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_17__pyx_fuse_4query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result = 0;
  __pyx_t_5numpy_uint64_t __pyx_v_point;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fuse_4query (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_result,&__pyx_n_s_point,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_result)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_point)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_4query", 1, 2, 2, 1); __PYX_ERR(1, 504, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_4query") < 0)) __PYX_ERR(1, 504, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_result = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)values[0]);
    __pyx_v_point = __Pyx_PyInt_As_npy_uint64(values[1]); if (unlikely((__pyx_v_point == ((npy_uint64)-1)) && PyErr_Occurred())) __PYX_ERR(1, 504, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_4query", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 504, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Float32ClosedRightIntervalNode.__pyx_fuse_4query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_result), __pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector, 1, "result", 0))) __PYX_ERR(1, 504, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_16__pyx_fuse_4query(((struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode *)__pyx_v_self), __pyx_v_result, __pyx_v_point);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_16__pyx_fuse_4query(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_uint64_t __pyx_v_point) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__pyx_fuse_4query", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_fuse_4__pyx_f_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_query(__pyx_v_self, __pyx_v_result, __pyx_v_point, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 504, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Float32ClosedRightIntervalNode.__pyx_fuse_4query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":544
 *                 result.extend(self.center_left_indices)
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         if self.is_leaf_node:
 *             return ('<Float32ClosedRightIntervalNode: '
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_5__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_5__repr__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_4__repr__(((struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_4__repr__(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode *__pyx_v_self) {
  PyObject *__pyx_v_n_left = NULL;
  PyObject *__pyx_v_n_right = NULL;
  PyObject *__pyx_v_n_center = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  Py_UCS4 __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "pandas/_libs/intervaltree.pxi":545
 * 
 *     def __repr__(self):
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             return ('<Float32ClosedRightIntervalNode: '
 *                     '%s elements (terminal)>' % self.n_elements)
 */
  __pyx_t_1 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_1) {

    /* "pandas/_libs/intervaltree.pxi":546
 *     def __repr__(self):
 *         if self.is_leaf_node:
 *             return ('<Float32ClosedRightIntervalNode: '             # <<<<<<<<<<<<<<
 *                     '%s elements (terminal)>' % self.n_elements)
 *         else:
 */
    __Pyx_XDECREF(__pyx_r);

    /* "pandas/_libs/intervaltree.pxi":547
 *         if self.is_leaf_node:
 *             return ('<Float32ClosedRightIntervalNode: '
 *                     '%s elements (terminal)>' % self.n_elements)             # <<<<<<<<<<<<<<
 *         else:
 *             n_left = self.left_node.n_elements
 */
    __pyx_t_2 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_elements); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 547, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PyUnicode_Format(__pyx_kp_u_Float32ClosedRightIntervalNode, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 547, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pandas/_libs/intervaltree.pxi":545
 * 
 *     def __repr__(self):
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             return ('<Float32ClosedRightIntervalNode: '
 *                     '%s elements (terminal)>' % self.n_elements)
 */
  }

  /* "pandas/_libs/intervaltree.pxi":549
 *                     '%s elements (terminal)>' % self.n_elements)
 *         else:
 *             n_left = self.left_node.n_elements             # <<<<<<<<<<<<<<
 *             n_right = self.right_node.n_elements
 *             n_center = self.n_elements - n_left - n_right
 */
  /*else*/ {
    __pyx_t_3 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->left_node->n_elements); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 549, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_v_n_left = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "pandas/_libs/intervaltree.pxi":550
 *         else:
 *             n_left = self.left_node.n_elements
 *             n_right = self.right_node.n_elements             # <<<<<<<<<<<<<<
 *             n_center = self.n_elements - n_left - n_right
 *             return ('<Float32ClosedRightIntervalNode: '
 */
    __pyx_t_3 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->right_node->n_elements); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 550, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_v_n_right = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "pandas/_libs/intervaltree.pxi":551
 *             n_left = self.left_node.n_elements
 *             n_right = self.right_node.n_elements
 *             n_center = self.n_elements - n_left - n_right             # <<<<<<<<<<<<<<
 *             return ('<Float32ClosedRightIntervalNode: '
 *                     'pivot %s, %s elements (%s left, %s right, %s '
 */
    __pyx_t_3 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_elements); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 551, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = PyNumber_Subtract(__pyx_t_3, __pyx_v_n_left); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 551, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyNumber_Subtract(__pyx_t_2, __pyx_v_n_right); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 551, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_n_center = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "pandas/_libs/intervaltree.pxi":552
 *             n_right = self.right_node.n_elements
 *             n_center = self.n_elements - n_left - n_right
 *             return ('<Float32ClosedRightIntervalNode: '             # <<<<<<<<<<<<<<
 *                     'pivot %s, %s elements (%s left, %s right, %s '
 *                     'overlapping)>' % (self.pivot, self.n_elements,
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = PyTuple_New(11); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 552, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = 0;
    __pyx_t_5 = 127;
    __Pyx_INCREF(__pyx_kp_u_Float32ClosedRightIntervalNode_2);
    __pyx_t_4 += 39;
    __Pyx_GIVEREF(__pyx_kp_u_Float32ClosedRightIntervalNode_2);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_kp_u_Float32ClosedRightIntervalNode_2);

    /* "pandas/_libs/intervaltree.pxi":554
 *             return ('<Float32ClosedRightIntervalNode: '
 *                     'pivot %s, %s elements (%s left, %s right, %s '
 *                     'overlapping)>' % (self.pivot, self.n_elements,             # <<<<<<<<<<<<<<
 *                                        n_left, n_right, n_center))
 * 
 */
    __pyx_t_2 = PyFloat_FromDouble(__pyx_v_self->pivot); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 554, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_t_2), __pyx_empty_unicode); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 554, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_5 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_6) > __pyx_t_5) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_6) : __pyx_t_5;
    __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_6);
    __pyx_t_6 = 0;
    __Pyx_INCREF(__pyx_kp_u__18);
    __pyx_t_4 += 2;
    __Pyx_GIVEREF(__pyx_kp_u__18);
    PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_kp_u__18);
    __pyx_t_6 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_elements); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 554, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_2 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_t_6), __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 554, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_5 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) > __pyx_t_5) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) : __pyx_t_5;
    __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_3, 3, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_elements);
    __pyx_t_4 += 11;
    __Pyx_GIVEREF(__pyx_kp_u_elements);
    PyTuple_SET_ITEM(__pyx_t_3, 4, __pyx_kp_u_elements);

    /* "pandas/_libs/intervaltree.pxi":555
 *                     'pivot %s, %s elements (%s left, %s right, %s '
 *                     'overlapping)>' % (self.pivot, self.n_elements,
 *                                        n_left, n_right, n_center))             # <<<<<<<<<<<<<<
 * 
 *     def counts(self):
 */
    __pyx_t_2 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_v_n_left), __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 555, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) > __pyx_t_5) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) : __pyx_t_5;
    __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_3, 5, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_left_2);
    __pyx_t_4 += 7;
    __Pyx_GIVEREF(__pyx_kp_u_left_2);
    PyTuple_SET_ITEM(__pyx_t_3, 6, __pyx_kp_u_left_2);
    __pyx_t_2 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_v_n_right), __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 555, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) > __pyx_t_5) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) : __pyx_t_5;
    __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_3, 7, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_right_2);
    __pyx_t_4 += 8;
    __Pyx_GIVEREF(__pyx_kp_u_right_2);
    PyTuple_SET_ITEM(__pyx_t_3, 8, __pyx_kp_u_right_2);
    __pyx_t_2 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_v_n_center), __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 555, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) > __pyx_t_5) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) : __pyx_t_5;
    __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_3, 9, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_overlapping);
    __pyx_t_4 += 14;
    __Pyx_GIVEREF(__pyx_kp_u_overlapping);
    PyTuple_SET_ITEM(__pyx_t_3, 10, __pyx_kp_u_overlapping);

    /* "pandas/_libs/intervaltree.pxi":552
 *             n_right = self.right_node.n_elements
 *             n_center = self.n_elements - n_left - n_right
 *             return ('<Float32ClosedRightIntervalNode: '             # <<<<<<<<<<<<<<
 *                     'pivot %s, %s elements (%s left, %s right, %s '
 *                     'overlapping)>' % (self.pivot, self.n_elements,
 */
    __pyx_t_2 = __Pyx_PyUnicode_Join(__pyx_t_3, 11, __pyx_t_4, __pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 552, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;
  }

  /* "pandas/_libs/intervaltree.pxi":544
 *                 result.extend(self.center_left_indices)
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         if self.is_leaf_node:
 *             return ('<Float32ClosedRightIntervalNode: '
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pandas._libs.interval.Float32ClosedRightIntervalNode.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_n_left);
  __Pyx_XDECREF(__pyx_v_n_right);
  __Pyx_XDECREF(__pyx_v_n_center);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":557
 *                                        n_left, n_right, n_center))
 * 
 *     def counts(self):             # <<<<<<<<<<<<<<
 *         """
 *         Inspect counts on this node
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_7counts(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_6counts[] = "\n        Inspect counts on this node\n        useful for debugging purposes\n        ";
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_7counts(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("counts (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_6counts(((struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_6counts(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode *__pyx_v_self) {
  Py_ssize_t __pyx_v_m;
  PyObject *__pyx_v_l = NULL;
  PyObject *__pyx_v_r = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  size_t __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("counts", 0);

  /* "pandas/_libs/intervaltree.pxi":562
 *         useful for debugging purposes
 *         """
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             return self.n_elements
 *         else:
 */
  __pyx_t_1 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_1) {

    /* "pandas/_libs/intervaltree.pxi":563
 *         """
 *         if self.is_leaf_node:
 *             return self.n_elements             # <<<<<<<<<<<<<<
 *         else:
 *             m = len(self.center_left_values)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_elements); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 563, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "pandas/_libs/intervaltree.pxi":562
 *         useful for debugging purposes
 *         """
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             return self.n_elements
 *         else:
 */
  }

  /* "pandas/_libs/intervaltree.pxi":565
 *             return self.n_elements
 *         else:
 *             m = len(self.center_left_values)             # <<<<<<<<<<<<<<
 *             l = self.left_node.counts()
 *             r = self.right_node.counts()
 */
  /*else*/ {
    if (unlikely(!__pyx_v_self->center_left_values.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(1, 565, __pyx_L1_error)}
    __pyx_t_3 = __Pyx_MemoryView_Len(__pyx_v_self->center_left_values); 
    __pyx_v_m = __pyx_t_3;

    /* "pandas/_libs/intervaltree.pxi":566
 *         else:
 *             m = len(self.center_left_values)
 *             l = self.left_node.counts()             # <<<<<<<<<<<<<<
 *             r = self.right_node.counts()
 *             return (m, (l, r))
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->left_node), __pyx_n_s_counts); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 566, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 566, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_l = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "pandas/_libs/intervaltree.pxi":567
 *             m = len(self.center_left_values)
 *             l = self.left_node.counts()
 *             r = self.right_node.counts()             # <<<<<<<<<<<<<<
 *             return (m, (l, r))
 * 
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->right_node), __pyx_n_s_counts); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 567, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 567, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_r = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "pandas/_libs/intervaltree.pxi":568
 *             l = self.left_node.counts()
 *             r = self.right_node.counts()
 *             return (m, (l, r))             # <<<<<<<<<<<<<<
 * 
 * NODE_CLASSES['float32',
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = PyInt_FromSsize_t(__pyx_v_m); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 568, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 568, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_v_l);
    __Pyx_GIVEREF(__pyx_v_l);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_l);
    __Pyx_INCREF(__pyx_v_r);
    __Pyx_GIVEREF(__pyx_v_r);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_v_r);
    __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 568, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_4);
    __pyx_t_2 = 0;
    __pyx_t_4 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;
  }

  /* "pandas/_libs/intervaltree.pxi":557
 *                                        n_left, n_right, n_center))
 * 
 *     def counts(self):             # <<<<<<<<<<<<<<
 *         """
 *         Inspect counts on this node
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pandas._libs.interval.Float32ClosedRightIntervalNode.counts", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_l);
  __Pyx_XDECREF(__pyx_v_r);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":414
 *         int64_t[:] center_left_indices, center_right_indices, indices
 *         float32_t min_left, max_right
 *         readonly float32_t pivot             # <<<<<<<<<<<<<<
 *         readonly int64_t n_elements, n_center, leaf_size
 *         readonly bint is_leaf_node
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_5pivot_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_5pivot_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_5pivot___get__(((struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_5pivot___get__(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->pivot); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 414, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Float32ClosedRightIntervalNode.pivot.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":415
 *         float32_t min_left, max_right
 *         readonly float32_t pivot
 *         readonly int64_t n_elements, n_center, leaf_size             # <<<<<<<<<<<<<<
 *         readonly bint is_leaf_node
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_10n_elements_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_10n_elements_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_10n_elements___get__(((struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_10n_elements___get__(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_elements); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 415, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Float32ClosedRightIntervalNode.n_elements.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_8n_center_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_8n_center_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_8n_center___get__(((struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_8n_center___get__(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_center); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 415, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Float32ClosedRightIntervalNode.n_center.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_9leaf_size_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_9leaf_size_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_9leaf_size___get__(((struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_9leaf_size___get__(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->leaf_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 415, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Float32ClosedRightIntervalNode.leaf_size.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":416
 *         readonly float32_t pivot
 *         readonly int64_t n_elements, n_center, leaf_size
 *         readonly bint is_leaf_node             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self,
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_12is_leaf_node_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_12is_leaf_node_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_12is_leaf_node___get__(((struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_12is_leaf_node___get__(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->is_leaf_node); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 416, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Float32ClosedRightIntervalNode.is_leaf_node.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_21__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_21__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_20__reduce_cython__(((struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_20__reduce_cython__(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  int __pyx_t_16;
  int __pyx_t_17;
  int __pyx_t_18;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.center_left_indices, self.center_left_values, self.center_right_indices, self.center_right_values, self.indices, self.is_leaf_node, self.leaf_size, self.left, self.left_node, self.max_right, self.min_left, self.n_center, self.n_elements, self.pivot, self.right, self.right_node)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  if (unlikely(!__pyx_v_self->center_left_indices.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_1 = __pyx_memoryview_fromslice(__pyx_v_self->center_left_indices, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_int64_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_int64_t, 0);; if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (unlikely(!__pyx_v_self->center_left_values.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_2 = __pyx_memoryview_fromslice(__pyx_v_self->center_left_values, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_float32_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_float32_t, 0);; if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (unlikely(!__pyx_v_self->center_right_indices.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_3 = __pyx_memoryview_fromslice(__pyx_v_self->center_right_indices, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_int64_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_int64_t, 0);; if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely(!__pyx_v_self->center_right_values.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_4 = __pyx_memoryview_fromslice(__pyx_v_self->center_right_values, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_float32_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_float32_t, 0);; if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (unlikely(!__pyx_v_self->indices.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_5 = __pyx_memoryview_fromslice(__pyx_v_self->indices, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_int64_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_int64_t, 0);; if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyBool_FromLong(__pyx_v_self->is_leaf_node); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->leaf_size); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (unlikely(!__pyx_v_self->left.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_8 = __pyx_memoryview_fromslice(__pyx_v_self->left, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_float32_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_float32_t, 0);; if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = PyFloat_FromDouble(__pyx_v_self->max_right); if (unlikely(!__pyx_t_9)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_10 = PyFloat_FromDouble(__pyx_v_self->min_left); if (unlikely(!__pyx_t_10)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_11 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_center); if (unlikely(!__pyx_t_11)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_12 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_elements); if (unlikely(!__pyx_t_12)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __pyx_t_13 = PyFloat_FromDouble(__pyx_v_self->pivot); if (unlikely(!__pyx_t_13)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  if (unlikely(!__pyx_v_self->right.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_14 = __pyx_memoryview_fromslice(__pyx_v_self->right, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_float32_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_float32_t, 0);; if (unlikely(!__pyx_t_14)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __pyx_t_15 = PyTuple_New(16); if (unlikely(!__pyx_t_15)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_15, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_15, 1, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_15, 2, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_15, 3, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_15, 4, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_15, 5, __pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_15, 6, __pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_15, 7, __pyx_t_8);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->left_node));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self->left_node));
  PyTuple_SET_ITEM(__pyx_t_15, 8, ((PyObject *)__pyx_v_self->left_node));
  __Pyx_GIVEREF(__pyx_t_9);
  PyTuple_SET_ITEM(__pyx_t_15, 9, __pyx_t_9);
  __Pyx_GIVEREF(__pyx_t_10);
  PyTuple_SET_ITEM(__pyx_t_15, 10, __pyx_t_10);
  __Pyx_GIVEREF(__pyx_t_11);
  PyTuple_SET_ITEM(__pyx_t_15, 11, __pyx_t_11);
  __Pyx_GIVEREF(__pyx_t_12);
  PyTuple_SET_ITEM(__pyx_t_15, 12, __pyx_t_12);
  __Pyx_GIVEREF(__pyx_t_13);
  PyTuple_SET_ITEM(__pyx_t_15, 13, __pyx_t_13);
  __Pyx_GIVEREF(__pyx_t_14);
  PyTuple_SET_ITEM(__pyx_t_15, 14, __pyx_t_14);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->right_node));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self->right_node));
  PyTuple_SET_ITEM(__pyx_t_15, 15, ((PyObject *)__pyx_v_self->right_node));
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  __pyx_t_6 = 0;
  __pyx_t_7 = 0;
  __pyx_t_8 = 0;
  __pyx_t_9 = 0;
  __pyx_t_10 = 0;
  __pyx_t_11 = 0;
  __pyx_t_12 = 0;
  __pyx_t_13 = 0;
  __pyx_t_14 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_15);
  __pyx_t_15 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.center_left_indices, self.center_left_values, self.center_right_indices, self.center_right_values, self.indices, self.is_leaf_node, self.leaf_size, self.left, self.left_node, self.max_right, self.min_left, self.n_center, self.n_elements, self.pivot, self.right, self.right_node)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_15 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_15)) __PYX_ERR(2, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __pyx_v__dict = __pyx_t_15;
  __pyx_t_15 = 0;

  /* "(tree fragment)":7
 *     state = (self.center_left_indices, self.center_left_values, self.center_right_indices, self.center_right_values, self.indices, self.is_leaf_node, self.leaf_size, self.left, self.left_node, self.max_right, self.min_left, self.n_center, self.n_elements, self.pivot, self.right, self.right_node)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_16 = (__pyx_v__dict != Py_None);
  __pyx_t_17 = (__pyx_t_16 != 0);
  if (__pyx_t_17) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_15 = PyTuple_New(1); if (unlikely(!__pyx_t_15)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    PyTuple_SET_ITEM(__pyx_t_15, 0, __pyx_v__dict);
    __pyx_t_14 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_15); if (unlikely(!__pyx_t_14)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_14));
    __pyx_t_14 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.left_node is not None or self.right_node is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.center_left_indices, self.center_left_values, self.center_right_indices, self.center_right_values, self.indices, self.is_leaf_node, self.leaf_size, self.left, self.left_node, self.max_right, self.min_left, self.n_center, self.n_elements, self.pivot, self.right, self.right_node)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.left_node is not None or self.right_node is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_Float32ClosedRightIntervalNode, (type(self), 0x7a104e5, None), state
 */
  /*else*/ {
    __pyx_t_16 = (((PyObject *)__pyx_v_self->left_node) != Py_None);
    __pyx_t_18 = (__pyx_t_16 != 0);
    if (!__pyx_t_18) {
    } else {
      __pyx_t_17 = __pyx_t_18;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_18 = (((PyObject *)__pyx_v_self->right_node) != Py_None);
    __pyx_t_16 = (__pyx_t_18 != 0);
    __pyx_t_17 = __pyx_t_16;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_17;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.left_node is not None or self.right_node is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Float32ClosedRightIntervalNode, (type(self), 0x7a104e5, None), state
 *     else:
 */
  __pyx_t_17 = (__pyx_v_use_setstate != 0);
  if (__pyx_t_17) {

    /* "(tree fragment)":13
 *         use_setstate = self.left_node is not None or self.right_node is not None
 *     if use_setstate:
 *         return __pyx_unpickle_Float32ClosedRightIntervalNode, (type(self), 0x7a104e5, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_Float32ClosedRightIntervalNode, (type(self), 0x7a104e5, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_14, __pyx_n_s_pyx_unpickle_Float32ClosedRigh); if (unlikely(!__pyx_t_14)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __pyx_t_15 = PyTuple_New(3); if (unlikely(!__pyx_t_15)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_15, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_127993061);
    __Pyx_GIVEREF(__pyx_int_127993061);
    PyTuple_SET_ITEM(__pyx_t_15, 1, __pyx_int_127993061);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_15, 2, Py_None);
    __pyx_t_13 = PyTuple_New(3); if (unlikely(!__pyx_t_13)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __Pyx_GIVEREF(__pyx_t_14);
    PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_t_14);
    __Pyx_GIVEREF(__pyx_t_15);
    PyTuple_SET_ITEM(__pyx_t_13, 1, __pyx_t_15);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_13, 2, __pyx_v_state);
    __pyx_t_14 = 0;
    __pyx_t_15 = 0;
    __pyx_r = __pyx_t_13;
    __pyx_t_13 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.left_node is not None or self.right_node is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Float32ClosedRightIntervalNode, (type(self), 0x7a104e5, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_Float32ClosedRightIntervalNode, (type(self), 0x7a104e5, None), state
 *     else:
 *         return __pyx_unpickle_Float32ClosedRightIntervalNode, (type(self), 0x7a104e5, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Float32ClosedRightIntervalNode__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_13, __pyx_n_s_pyx_unpickle_Float32ClosedRigh); if (unlikely(!__pyx_t_13)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __pyx_t_15 = PyTuple_New(3); if (unlikely(!__pyx_t_15)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_15, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_127993061);
    __Pyx_GIVEREF(__pyx_int_127993061);
    PyTuple_SET_ITEM(__pyx_t_15, 1, __pyx_int_127993061);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_15, 2, __pyx_v_state);
    __pyx_t_14 = PyTuple_New(2); if (unlikely(!__pyx_t_14)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_GIVEREF(__pyx_t_13);
    PyTuple_SET_ITEM(__pyx_t_14, 0, __pyx_t_13);
    __Pyx_GIVEREF(__pyx_t_15);
    PyTuple_SET_ITEM(__pyx_t_14, 1, __pyx_t_15);
    __pyx_t_13 = 0;
    __pyx_t_15 = 0;
    __pyx_r = __pyx_t_14;
    __pyx_t_14 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_AddTraceback("pandas._libs.interval.Float32ClosedRightIntervalNode.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Float32ClosedRightIntervalNode, (type(self), 0x7a104e5, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Float32ClosedRightIntervalNode__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_23__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_23__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_22__setstate_cython__(((struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_22__setstate_cython__(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_Float32ClosedRightIntervalNode, (type(self), 0x7a104e5, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Float32ClosedRightIntervalNode__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(2, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_6pandas_5_libs_8interval___pyx_unpickle_Float32ClosedRightIntervalNode__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Float32ClosedRightIntervalNode, (type(self), 0x7a104e5, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Float32ClosedRightIntervalNode__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Float32ClosedRightIntervalNode.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":588
 *         readonly bint is_leaf_node
 * 
 *     def __init__(self,             # <<<<<<<<<<<<<<
 *                  ndarray[float32_t, ndim=1] left,
 *                  ndarray[float32_t, ndim=1] right,
 */

/* Python wrapper */
static int __pyx_pw_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyArrayObject *__pyx_v_left = 0;
  PyArrayObject *__pyx_v_right = 0;
  PyArrayObject *__pyx_v_indices = 0;
  __pyx_t_5numpy_int64_t __pyx_v_leaf_size;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_left,&__pyx_n_s_right,&__pyx_n_s_indices,&__pyx_n_s_leaf_size,0};
    PyObject* values[4] = {0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_left)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_right)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 4, 4, 1); __PYX_ERR(1, 588, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_indices)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 4, 4, 2); __PYX_ERR(1, 588, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_leaf_size)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 4, 4, 3); __PYX_ERR(1, 588, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(1, 588, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 4) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
    }
    __pyx_v_left = ((PyArrayObject *)values[0]);
    __pyx_v_right = ((PyArrayObject *)values[1]);
    __pyx_v_indices = ((PyArrayObject *)values[2]);
    __pyx_v_leaf_size = __Pyx_PyInt_As_npy_int64(values[3]); if (unlikely((__pyx_v_leaf_size == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(1, 592, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 4, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 588, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Float32ClosedBothIntervalNode.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_left), __pyx_ptype_5numpy_ndarray, 1, "left", 0))) __PYX_ERR(1, 589, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_right), __pyx_ptype_5numpy_ndarray, 1, "right", 0))) __PYX_ERR(1, 590, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_indices), __pyx_ptype_5numpy_ndarray, 1, "indices", 0))) __PYX_ERR(1, 591, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode___init__(((struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode *)__pyx_v_self), __pyx_v_left, __pyx_v_right, __pyx_v_indices, __pyx_v_leaf_size);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode___init__(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode *__pyx_v_self, PyArrayObject *__pyx_v_left, PyArrayObject *__pyx_v_right, PyArrayObject *__pyx_v_indices, __pyx_t_5numpy_int64_t __pyx_v_leaf_size) {
  PyObject *__pyx_v_left_set = NULL;
  PyObject *__pyx_v_right_set = NULL;
  PyObject *__pyx_v_center_set = NULL;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_indices;
  __Pyx_Buffer __pyx_pybuffer_indices;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_left;
  __Pyx_Buffer __pyx_pybuffer_left;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_right;
  __Pyx_Buffer __pyx_pybuffer_right;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __pyx_t_5numpy_float32_t __pyx_t_6;
  __Pyx_memviewslice __pyx_t_7 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_8 = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  __Pyx_memviewslice __pyx_t_11 = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *(*__pyx_t_12)(PyObject *);
  size_t __pyx_t_13;
  __Pyx_RefNannySetupContext("__init__", 0);
  __pyx_pybuffer_left.pybuffer.buf = NULL;
  __pyx_pybuffer_left.refcount = 0;
  __pyx_pybuffernd_left.data = NULL;
  __pyx_pybuffernd_left.rcbuffer = &__pyx_pybuffer_left;
  __pyx_pybuffer_right.pybuffer.buf = NULL;
  __pyx_pybuffer_right.refcount = 0;
  __pyx_pybuffernd_right.data = NULL;
  __pyx_pybuffernd_right.rcbuffer = &__pyx_pybuffer_right;
  __pyx_pybuffer_indices.pybuffer.buf = NULL;
  __pyx_pybuffer_indices.refcount = 0;
  __pyx_pybuffernd_indices.data = NULL;
  __pyx_pybuffernd_indices.rcbuffer = &__pyx_pybuffer_indices;
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_left.rcbuffer->pybuffer, (PyObject*)__pyx_v_left, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float32_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 588, __pyx_L1_error)
  }
  __pyx_pybuffernd_left.diminfo[0].strides = __pyx_pybuffernd_left.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_left.diminfo[0].shape = __pyx_pybuffernd_left.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_right.rcbuffer->pybuffer, (PyObject*)__pyx_v_right, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float32_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 588, __pyx_L1_error)
  }
  __pyx_pybuffernd_right.diminfo[0].strides = __pyx_pybuffernd_right.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_right.diminfo[0].shape = __pyx_pybuffernd_right.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_indices.rcbuffer->pybuffer, (PyObject*)__pyx_v_indices, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 588, __pyx_L1_error)
  }
  __pyx_pybuffernd_indices.diminfo[0].strides = __pyx_pybuffernd_indices.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_indices.diminfo[0].shape = __pyx_pybuffernd_indices.rcbuffer->pybuffer.shape[0];

  /* "pandas/_libs/intervaltree.pxi":594
 *                  int64_t leaf_size):
 * 
 *         self.n_elements = len(left)             # <<<<<<<<<<<<<<
 *         self.leaf_size = leaf_size
 * 
 */
  __pyx_t_1 = PyObject_Length(((PyObject *)__pyx_v_left)); if (unlikely(__pyx_t_1 == ((Py_ssize_t)-1))) __PYX_ERR(1, 594, __pyx_L1_error)
  __pyx_v_self->n_elements = __pyx_t_1;

  /* "pandas/_libs/intervaltree.pxi":595
 * 
 *         self.n_elements = len(left)
 *         self.leaf_size = leaf_size             # <<<<<<<<<<<<<<
 * 
 *         # min_left and min_right are used to speed-up query by skipping
 */
  __pyx_v_self->leaf_size = __pyx_v_leaf_size;

  /* "pandas/_libs/intervaltree.pxi":600
 *         # query on sub-nodes. If this node has size 0, query is cheap,
 *         # so these values don't matter.
 *         if left.size > 0:             # <<<<<<<<<<<<<<
 *             self.min_left = left.min()
 *             self.max_right = right.max()
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_left), __pyx_n_s_size); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 600, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyObject_RichCompare(__pyx_t_2, __pyx_int_0, Py_GT); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 600, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(1, 600, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_4) {

    /* "pandas/_libs/intervaltree.pxi":601
 *         # so these values don't matter.
 *         if left.size > 0:
 *             self.min_left = left.min()             # <<<<<<<<<<<<<<
 *             self.max_right = right.max()
 *         else:
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_left), __pyx_n_s_min); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 601, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 601, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_6 = __pyx_PyFloat_AsFloat(__pyx_t_3); if (unlikely((__pyx_t_6 == ((npy_float32)-1)) && PyErr_Occurred())) __PYX_ERR(1, 601, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_self->min_left = __pyx_t_6;

    /* "pandas/_libs/intervaltree.pxi":602
 *         if left.size > 0:
 *             self.min_left = left.min()
 *             self.max_right = right.max()             # <<<<<<<<<<<<<<
 *         else:
 *             self.min_left = 0
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_right), __pyx_n_s_max); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 602, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 602, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_6 = __pyx_PyFloat_AsFloat(__pyx_t_3); if (unlikely((__pyx_t_6 == ((npy_float32)-1)) && PyErr_Occurred())) __PYX_ERR(1, 602, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_self->max_right = __pyx_t_6;

    /* "pandas/_libs/intervaltree.pxi":600
 *         # query on sub-nodes. If this node has size 0, query is cheap,
 *         # so these values don't matter.
 *         if left.size > 0:             # <<<<<<<<<<<<<<
 *             self.min_left = left.min()
 *             self.max_right = right.max()
 */
    goto __pyx_L3;
  }

  /* "pandas/_libs/intervaltree.pxi":604
 *             self.max_right = right.max()
 *         else:
 *             self.min_left = 0             # <<<<<<<<<<<<<<
 *             self.max_right = 0
 * 
 */
  /*else*/ {
    __pyx_v_self->min_left = 0.0;

    /* "pandas/_libs/intervaltree.pxi":605
 *         else:
 *             self.min_left = 0
 *             self.max_right = 0             # <<<<<<<<<<<<<<
 * 
 *         if self.n_elements <= leaf_size:
 */
    __pyx_v_self->max_right = 0.0;
  }
  __pyx_L3:;

  /* "pandas/_libs/intervaltree.pxi":607
 *             self.max_right = 0
 * 
 *         if self.n_elements <= leaf_size:             # <<<<<<<<<<<<<<
 *             # make this a terminal (leaf) node
 *             self.is_leaf_node = True
 */
  __pyx_t_4 = ((__pyx_v_self->n_elements <= __pyx_v_leaf_size) != 0);
  if (__pyx_t_4) {

    /* "pandas/_libs/intervaltree.pxi":609
 *         if self.n_elements <= leaf_size:
 *             # make this a terminal (leaf) node
 *             self.is_leaf_node = True             # <<<<<<<<<<<<<<
 *             self.left = left
 *             self.right = right
 */
    __pyx_v_self->is_leaf_node = 1;

    /* "pandas/_libs/intervaltree.pxi":610
 *             # make this a terminal (leaf) node
 *             self.is_leaf_node = True
 *             self.left = left             # <<<<<<<<<<<<<<
 *             self.right = right
 *             self.indices = indices
 */
    __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_float32_t(((PyObject *)__pyx_v_left), PyBUF_WRITABLE); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(1, 610, __pyx_L1_error)
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->left, 0);
    __pyx_v_self->left = __pyx_t_7;
    __pyx_t_7.memview = NULL;
    __pyx_t_7.data = NULL;

    /* "pandas/_libs/intervaltree.pxi":611
 *             self.is_leaf_node = True
 *             self.left = left
 *             self.right = right             # <<<<<<<<<<<<<<
 *             self.indices = indices
 *             self.n_center = 0
 */
    __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_float32_t(((PyObject *)__pyx_v_right), PyBUF_WRITABLE); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(1, 611, __pyx_L1_error)
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->right, 0);
    __pyx_v_self->right = __pyx_t_7;
    __pyx_t_7.memview = NULL;
    __pyx_t_7.data = NULL;

    /* "pandas/_libs/intervaltree.pxi":612
 *             self.left = left
 *             self.right = right
 *             self.indices = indices             # <<<<<<<<<<<<<<
 *             self.n_center = 0
 *         else:
 */
    __pyx_t_8 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(((PyObject *)__pyx_v_indices), PyBUF_WRITABLE); if (unlikely(!__pyx_t_8.memview)) __PYX_ERR(1, 612, __pyx_L1_error)
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->indices, 0);
    __pyx_v_self->indices = __pyx_t_8;
    __pyx_t_8.memview = NULL;
    __pyx_t_8.data = NULL;

    /* "pandas/_libs/intervaltree.pxi":613
 *             self.right = right
 *             self.indices = indices
 *             self.n_center = 0             # <<<<<<<<<<<<<<
 *         else:
 *             # calculate a pivot so we can create child nodes
 */
    __pyx_v_self->n_center = 0;

    /* "pandas/_libs/intervaltree.pxi":607
 *             self.max_right = 0
 * 
 *         if self.n_elements <= leaf_size:             # <<<<<<<<<<<<<<
 *             # make this a terminal (leaf) node
 *             self.is_leaf_node = True
 */
    goto __pyx_L4;
  }

  /* "pandas/_libs/intervaltree.pxi":616
 *         else:
 *             # calculate a pivot so we can create child nodes
 *             self.is_leaf_node = False             # <<<<<<<<<<<<<<
 *             self.pivot = np.median(left / 2 + right / 2)
 *             left_set, right_set, center_set = self.classify_intervals(
 */
  /*else*/ {
    __pyx_v_self->is_leaf_node = 0;

    /* "pandas/_libs/intervaltree.pxi":617
 *             # calculate a pivot so we can create child nodes
 *             self.is_leaf_node = False
 *             self.pivot = np.median(left / 2 + right / 2)             # <<<<<<<<<<<<<<
 *             left_set, right_set, center_set = self.classify_intervals(
 *                 left, right)
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 617, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_median); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 617, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyNumber_Divide(((PyObject *)__pyx_v_left), __pyx_int_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 617, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_9 = __Pyx_PyNumber_Divide(((PyObject *)__pyx_v_right), __pyx_int_2); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 617, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_10 = PyNumber_Add(__pyx_t_2, __pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 617, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_9 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_3 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_9, __pyx_t_10) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_10);
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 617, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_6 = __pyx_PyFloat_AsFloat(__pyx_t_3); if (unlikely((__pyx_t_6 == ((npy_float32)-1)) && PyErr_Occurred())) __PYX_ERR(1, 617, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_self->pivot = __pyx_t_6;

    /* "pandas/_libs/intervaltree.pxi":619
 *             self.pivot = np.median(left / 2 + right / 2)
 *             left_set, right_set, center_set = self.classify_intervals(
 *                 left, right)             # <<<<<<<<<<<<<<
 * 
 *             self.left_node = self.new_child_node(left, right,
 */
    __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_float32_t(((PyObject *)__pyx_v_left), PyBUF_WRITABLE); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(1, 619, __pyx_L1_error)
    __pyx_t_11 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_float32_t(((PyObject *)__pyx_v_right), PyBUF_WRITABLE); if (unlikely(!__pyx_t_11.memview)) __PYX_ERR(1, 619, __pyx_L1_error)

    /* "pandas/_libs/intervaltree.pxi":618
 *             self.is_leaf_node = False
 *             self.pivot = np.median(left / 2 + right / 2)
 *             left_set, right_set, center_set = self.classify_intervals(             # <<<<<<<<<<<<<<
 *                 left, right)
 * 
 */
    __pyx_t_3 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode *)__pyx_v_self->__pyx_vtab)->classify_intervals(__pyx_v_self, __pyx_t_7, __pyx_t_11); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 618, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __PYX_XDEC_MEMVIEW(&__pyx_t_7, 1);
    __pyx_t_7.memview = NULL;
    __pyx_t_7.data = NULL;
    __PYX_XDEC_MEMVIEW(&__pyx_t_11, 1);
    __pyx_t_11.memview = NULL;
    __pyx_t_11.data = NULL;
    if ((likely(PyTuple_CheckExact(__pyx_t_3))) || (PyList_CheckExact(__pyx_t_3))) {
      PyObject* sequence = __pyx_t_3;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 3)) {
        if (size > 3) __Pyx_RaiseTooManyValuesError(3);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(1, 618, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_5 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_10 = PyTuple_GET_ITEM(sequence, 1); 
        __pyx_t_9 = PyTuple_GET_ITEM(sequence, 2); 
      } else {
        __pyx_t_5 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_10 = PyList_GET_ITEM(sequence, 1); 
        __pyx_t_9 = PyList_GET_ITEM(sequence, 2); 
      }
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_10);
      __Pyx_INCREF(__pyx_t_9);
      #else
      __pyx_t_5 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 618, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_10 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 618, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_9 = PySequence_ITEM(sequence, 2); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 618, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      #endif
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_2 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 618, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_12 = Py_TYPE(__pyx_t_2)->tp_iternext;
      index = 0; __pyx_t_5 = __pyx_t_12(__pyx_t_2); if (unlikely(!__pyx_t_5)) goto __pyx_L5_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_5);
      index = 1; __pyx_t_10 = __pyx_t_12(__pyx_t_2); if (unlikely(!__pyx_t_10)) goto __pyx_L5_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_10);
      index = 2; __pyx_t_9 = __pyx_t_12(__pyx_t_2); if (unlikely(!__pyx_t_9)) goto __pyx_L5_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_9);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_12(__pyx_t_2), 3) < 0) __PYX_ERR(1, 618, __pyx_L1_error)
      __pyx_t_12 = NULL;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      goto __pyx_L6_unpacking_done;
      __pyx_L5_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_12 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(1, 618, __pyx_L1_error)
      __pyx_L6_unpacking_done:;
    }
    __pyx_v_left_set = __pyx_t_5;
    __pyx_t_5 = 0;
    __pyx_v_right_set = __pyx_t_10;
    __pyx_t_10 = 0;
    __pyx_v_center_set = __pyx_t_9;
    __pyx_t_9 = 0;

    /* "pandas/_libs/intervaltree.pxi":622
 * 
 *             self.left_node = self.new_child_node(left, right,
 *                                                  indices, left_set)             # <<<<<<<<<<<<<<
 *             self.right_node = self.new_child_node(left, right,
 *                                                   indices, right_set)
 */
    if (!(likely(((__pyx_v_left_set) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_left_set, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(1, 622, __pyx_L1_error)

    /* "pandas/_libs/intervaltree.pxi":621
 *                 left, right)
 * 
 *             self.left_node = self.new_child_node(left, right,             # <<<<<<<<<<<<<<
 *                                                  indices, left_set)
 *             self.right_node = self.new_child_node(left, right,
 */
    __pyx_t_3 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode *)__pyx_v_self->__pyx_vtab)->new_child_node(__pyx_v_self, ((PyArrayObject *)__pyx_v_left), ((PyArrayObject *)__pyx_v_right), ((PyArrayObject *)__pyx_v_indices), ((PyArrayObject *)__pyx_v_left_set)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 621, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode))))) __PYX_ERR(1, 621, __pyx_L1_error)
    __Pyx_GIVEREF(__pyx_t_3);
    __Pyx_GOTREF(__pyx_v_self->left_node);
    __Pyx_DECREF(((PyObject *)__pyx_v_self->left_node));
    __pyx_v_self->left_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode *)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "pandas/_libs/intervaltree.pxi":624
 *                                                  indices, left_set)
 *             self.right_node = self.new_child_node(left, right,
 *                                                   indices, right_set)             # <<<<<<<<<<<<<<
 * 
 *             self.center_left_values, self.center_left_indices = \
 */
    if (!(likely(((__pyx_v_right_set) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_right_set, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(1, 624, __pyx_L1_error)

    /* "pandas/_libs/intervaltree.pxi":623
 *             self.left_node = self.new_child_node(left, right,
 *                                                  indices, left_set)
 *             self.right_node = self.new_child_node(left, right,             # <<<<<<<<<<<<<<
 *                                                   indices, right_set)
 * 
 */
    __pyx_t_3 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode *)__pyx_v_self->__pyx_vtab)->new_child_node(__pyx_v_self, ((PyArrayObject *)__pyx_v_left), ((PyArrayObject *)__pyx_v_right), ((PyArrayObject *)__pyx_v_indices), ((PyArrayObject *)__pyx_v_right_set)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 623, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode))))) __PYX_ERR(1, 623, __pyx_L1_error)
    __Pyx_GIVEREF(__pyx_t_3);
    __Pyx_GOTREF(__pyx_v_self->right_node);
    __Pyx_DECREF(((PyObject *)__pyx_v_self->right_node));
    __pyx_v_self->right_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode *)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "pandas/_libs/intervaltree.pxi":627
 * 
 *             self.center_left_values, self.center_left_indices = \
 *                 sort_values_and_indices(left, indices, center_set)             # <<<<<<<<<<<<<<
 *             self.center_right_values, self.center_right_indices = \
 *                 sort_values_and_indices(right, indices, center_set)
 */
    __pyx_t_3 = __pyx_f_6pandas_5_libs_8interval_sort_values_and_indices(((PyObject *)__pyx_v_left), ((PyObject *)__pyx_v_indices), __pyx_v_center_set); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 627, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if ((likely(PyTuple_CheckExact(__pyx_t_3))) || (PyList_CheckExact(__pyx_t_3))) {
      PyObject* sequence = __pyx_t_3;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(1, 626, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_9 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_10 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_9 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_10 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_9);
      __Pyx_INCREF(__pyx_t_10);
      #else
      __pyx_t_9 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 626, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_10 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 626, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      #endif
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_5 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 626, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_12 = Py_TYPE(__pyx_t_5)->tp_iternext;
      index = 0; __pyx_t_9 = __pyx_t_12(__pyx_t_5); if (unlikely(!__pyx_t_9)) goto __pyx_L7_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_9);
      index = 1; __pyx_t_10 = __pyx_t_12(__pyx_t_5); if (unlikely(!__pyx_t_10)) goto __pyx_L7_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_10);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_12(__pyx_t_5), 2) < 0) __PYX_ERR(1, 626, __pyx_L1_error)
      __pyx_t_12 = NULL;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      goto __pyx_L8_unpacking_done;
      __pyx_L7_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_12 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(1, 626, __pyx_L1_error)
      __pyx_L8_unpacking_done:;
    }

    /* "pandas/_libs/intervaltree.pxi":626
 *                                                   indices, right_set)
 * 
 *             self.center_left_values, self.center_left_indices = \             # <<<<<<<<<<<<<<
 *                 sort_values_and_indices(left, indices, center_set)
 *             self.center_right_values, self.center_right_indices = \
 */
    __pyx_t_11 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_float32_t(__pyx_t_9, PyBUF_WRITABLE); if (unlikely(!__pyx_t_11.memview)) __PYX_ERR(1, 626, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_8 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_t_10, PyBUF_WRITABLE); if (unlikely(!__pyx_t_8.memview)) __PYX_ERR(1, 626, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->center_left_values, 0);
    __pyx_v_self->center_left_values = __pyx_t_11;
    __pyx_t_11.memview = NULL;
    __pyx_t_11.data = NULL;
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->center_left_indices, 0);
    __pyx_v_self->center_left_indices = __pyx_t_8;
    __pyx_t_8.memview = NULL;
    __pyx_t_8.data = NULL;

    /* "pandas/_libs/intervaltree.pxi":629
 *                 sort_values_and_indices(left, indices, center_set)
 *             self.center_right_values, self.center_right_indices = \
 *                 sort_values_and_indices(right, indices, center_set)             # <<<<<<<<<<<<<<
 *             self.n_center = len(self.center_left_indices)
 * 
 */
    __pyx_t_3 = __pyx_f_6pandas_5_libs_8interval_sort_values_and_indices(((PyObject *)__pyx_v_right), ((PyObject *)__pyx_v_indices), __pyx_v_center_set); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 629, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if ((likely(PyTuple_CheckExact(__pyx_t_3))) || (PyList_CheckExact(__pyx_t_3))) {
      PyObject* sequence = __pyx_t_3;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(1, 628, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_10 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_9 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_10 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_9 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_10);
      __Pyx_INCREF(__pyx_t_9);
      #else
      __pyx_t_10 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 628, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_9 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 628, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      #endif
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_5 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 628, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_12 = Py_TYPE(__pyx_t_5)->tp_iternext;
      index = 0; __pyx_t_10 = __pyx_t_12(__pyx_t_5); if (unlikely(!__pyx_t_10)) goto __pyx_L9_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_10);
      index = 1; __pyx_t_9 = __pyx_t_12(__pyx_t_5); if (unlikely(!__pyx_t_9)) goto __pyx_L9_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_9);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_12(__pyx_t_5), 2) < 0) __PYX_ERR(1, 628, __pyx_L1_error)
      __pyx_t_12 = NULL;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      goto __pyx_L10_unpacking_done;
      __pyx_L9_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_12 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(1, 628, __pyx_L1_error)
      __pyx_L10_unpacking_done:;
    }

    /* "pandas/_libs/intervaltree.pxi":628
 *             self.center_left_values, self.center_left_indices = \
 *                 sort_values_and_indices(left, indices, center_set)
 *             self.center_right_values, self.center_right_indices = \             # <<<<<<<<<<<<<<
 *                 sort_values_and_indices(right, indices, center_set)
 *             self.n_center = len(self.center_left_indices)
 */
    __pyx_t_11 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_float32_t(__pyx_t_10, PyBUF_WRITABLE); if (unlikely(!__pyx_t_11.memview)) __PYX_ERR(1, 628, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_t_8 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_t_9, PyBUF_WRITABLE); if (unlikely(!__pyx_t_8.memview)) __PYX_ERR(1, 628, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->center_right_values, 0);
    __pyx_v_self->center_right_values = __pyx_t_11;
    __pyx_t_11.memview = NULL;
    __pyx_t_11.data = NULL;
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->center_right_indices, 0);
    __pyx_v_self->center_right_indices = __pyx_t_8;
    __pyx_t_8.memview = NULL;
    __pyx_t_8.data = NULL;

    /* "pandas/_libs/intervaltree.pxi":630
 *             self.center_right_values, self.center_right_indices = \
 *                 sort_values_and_indices(right, indices, center_set)
 *             self.n_center = len(self.center_left_indices)             # <<<<<<<<<<<<<<
 * 
 *     @cython.wraparound(False)
 */
    if (unlikely(!__pyx_v_self->center_left_indices.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(1, 630, __pyx_L1_error)}
    __pyx_t_13 = __Pyx_MemoryView_Len(__pyx_v_self->center_left_indices); 
    __pyx_v_self->n_center = __pyx_t_13;
  }
  __pyx_L4:;

  /* "pandas/_libs/intervaltree.pxi":588
 *         readonly bint is_leaf_node
 * 
 *     def __init__(self,             # <<<<<<<<<<<<<<
 *                  ndarray[float32_t, ndim=1] left,
 *                  ndarray[float32_t, ndim=1] right,
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __PYX_XDEC_MEMVIEW(&__pyx_t_7, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_8, 1);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __PYX_XDEC_MEMVIEW(&__pyx_t_11, 1);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indices.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_left.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_right.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("pandas._libs.interval.Float32ClosedBothIntervalNode.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indices.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_left.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_right.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF(__pyx_v_left_set);
  __Pyx_XDECREF(__pyx_v_right_set);
  __Pyx_XDECREF(__pyx_v_center_set);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":634
 *     @cython.wraparound(False)
 *     @cython.boundscheck(False)
 *     cdef classify_intervals(self, float32_t[:] left, float32_t[:] right):             # <<<<<<<<<<<<<<
 *         """Classify the given intervals based upon whether they fall to the
 *         left, right, or overlap with this node's pivot.
 */

static PyObject *__pyx_f_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_classify_intervals(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode *__pyx_v_self, __Pyx_memviewslice __pyx_v_left, __Pyx_memviewslice __pyx_v_right) {
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_left_ind = 0;
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_right_ind = 0;
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_overlapping_ind = 0;
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __pyx_t_5numpy_int64_t __pyx_t_2;
  __pyx_t_5numpy_int64_t __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  int __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  __Pyx_RefNannySetupContext("classify_intervals", 0);

  /* "pandas/_libs/intervaltree.pxi":642
 *             Py_ssize_t i
 * 
 *         left_ind = Int64Vector()             # <<<<<<<<<<<<<<
 *         right_ind = Int64Vector()
 *         overlapping_ind = Int64Vector()
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 642, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_left_ind = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":643
 * 
 *         left_ind = Int64Vector()
 *         right_ind = Int64Vector()             # <<<<<<<<<<<<<<
 *         overlapping_ind = Int64Vector()
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 643, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_right_ind = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":644
 *         left_ind = Int64Vector()
 *         right_ind = Int64Vector()
 *         overlapping_ind = Int64Vector()             # <<<<<<<<<<<<<<
 * 
 *         for i in range(self.n_elements):
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 644, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_overlapping_ind = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":646
 *         overlapping_ind = Int64Vector()
 * 
 *         for i in range(self.n_elements):             # <<<<<<<<<<<<<<
 *             if right[i] < self.pivot:
 *                 left_ind.append(i)
 */
  __pyx_t_2 = __pyx_v_self->n_elements;
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "pandas/_libs/intervaltree.pxi":647
 * 
 *         for i in range(self.n_elements):
 *             if right[i] < self.pivot:             # <<<<<<<<<<<<<<
 *                 left_ind.append(i)
 *             elif self.pivot < left[i]:
 */
    __pyx_t_5 = __pyx_v_i;
    __pyx_t_6 = (((*((__pyx_t_5numpy_float32_t *) ( /* dim=0 */ (__pyx_v_right.data + __pyx_t_5 * __pyx_v_right.strides[0]) ))) < __pyx_v_self->pivot) != 0);
    if (__pyx_t_6) {

      /* "pandas/_libs/intervaltree.pxi":648
 *         for i in range(self.n_elements):
 *             if right[i] < self.pivot:
 *                 left_ind.append(i)             # <<<<<<<<<<<<<<
 *             elif self.pivot < left[i]:
 *                 right_ind.append(i)
 */
      ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_left_ind->__pyx_vtab)->append(__pyx_v_left_ind, __pyx_v_i);

      /* "pandas/_libs/intervaltree.pxi":647
 * 
 *         for i in range(self.n_elements):
 *             if right[i] < self.pivot:             # <<<<<<<<<<<<<<
 *                 left_ind.append(i)
 *             elif self.pivot < left[i]:
 */
      goto __pyx_L5;
    }

    /* "pandas/_libs/intervaltree.pxi":649
 *             if right[i] < self.pivot:
 *                 left_ind.append(i)
 *             elif self.pivot < left[i]:             # <<<<<<<<<<<<<<
 *                 right_ind.append(i)
 *             else:
 */
    __pyx_t_7 = __pyx_v_i;
    __pyx_t_6 = ((__pyx_v_self->pivot < (*((__pyx_t_5numpy_float32_t *) ( /* dim=0 */ (__pyx_v_left.data + __pyx_t_7 * __pyx_v_left.strides[0]) )))) != 0);
    if (__pyx_t_6) {

      /* "pandas/_libs/intervaltree.pxi":650
 *                 left_ind.append(i)
 *             elif self.pivot < left[i]:
 *                 right_ind.append(i)             # <<<<<<<<<<<<<<
 *             else:
 *                 overlapping_ind.append(i)
 */
      ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_right_ind->__pyx_vtab)->append(__pyx_v_right_ind, __pyx_v_i);

      /* "pandas/_libs/intervaltree.pxi":649
 *             if right[i] < self.pivot:
 *                 left_ind.append(i)
 *             elif self.pivot < left[i]:             # <<<<<<<<<<<<<<
 *                 right_ind.append(i)
 *             else:
 */
      goto __pyx_L5;
    }

    /* "pandas/_libs/intervaltree.pxi":652
 *                 right_ind.append(i)
 *             else:
 *                 overlapping_ind.append(i)             # <<<<<<<<<<<<<<
 * 
 *         return (left_ind.to_array(),
 */
    /*else*/ {
      ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_overlapping_ind->__pyx_vtab)->append(__pyx_v_overlapping_ind, __pyx_v_i);
    }
    __pyx_L5:;
  }

  /* "pandas/_libs/intervaltree.pxi":654
 *                 overlapping_ind.append(i)
 * 
 *         return (left_ind.to_array(),             # <<<<<<<<<<<<<<
 *                 right_ind.to_array(),
 *                 overlapping_ind.to_array())
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_left_ind->__pyx_vtab)->to_array(__pyx_v_left_ind, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 654, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "pandas/_libs/intervaltree.pxi":655
 * 
 *         return (left_ind.to_array(),
 *                 right_ind.to_array(),             # <<<<<<<<<<<<<<
 *                 overlapping_ind.to_array())
 * 
 */
  __pyx_t_8 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_right_ind->__pyx_vtab)->to_array(__pyx_v_right_ind, 0); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 655, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);

  /* "pandas/_libs/intervaltree.pxi":656
 *         return (left_ind.to_array(),
 *                 right_ind.to_array(),
 *                 overlapping_ind.to_array())             # <<<<<<<<<<<<<<
 * 
 *     cdef new_child_node(self,
 */
  __pyx_t_9 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_overlapping_ind->__pyx_vtab)->to_array(__pyx_v_overlapping_ind, 0); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 656, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);

  /* "pandas/_libs/intervaltree.pxi":654
 *                 overlapping_ind.append(i)
 * 
 *         return (left_ind.to_array(),             # <<<<<<<<<<<<<<
 *                 right_ind.to_array(),
 *                 overlapping_ind.to_array())
 */
  __pyx_t_10 = PyTuple_New(3); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 654, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_t_8);
  __Pyx_GIVEREF(__pyx_t_9);
  PyTuple_SET_ITEM(__pyx_t_10, 2, __pyx_t_9);
  __pyx_t_1 = 0;
  __pyx_t_8 = 0;
  __pyx_t_9 = 0;
  __pyx_r = __pyx_t_10;
  __pyx_t_10 = 0;
  goto __pyx_L0;

  /* "pandas/_libs/intervaltree.pxi":634
 *     @cython.wraparound(False)
 *     @cython.boundscheck(False)
 *     cdef classify_intervals(self, float32_t[:] left, float32_t[:] right):             # <<<<<<<<<<<<<<
 *         """Classify the given intervals based upon whether they fall to the
 *         left, right, or overlap with this node's pivot.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("pandas._libs.interval.Float32ClosedBothIntervalNode.classify_intervals", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_left_ind);
  __Pyx_XDECREF((PyObject *)__pyx_v_right_ind);
  __Pyx_XDECREF((PyObject *)__pyx_v_overlapping_ind);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":658
 *                 overlapping_ind.to_array())
 * 
 *     cdef new_child_node(self,             # <<<<<<<<<<<<<<
 *                         ndarray[float32_t, ndim=1] left,
 *                         ndarray[float32_t, ndim=1] right,
 */

static PyObject *__pyx_f_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_new_child_node(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode *__pyx_v_self, PyArrayObject *__pyx_v_left, PyArrayObject *__pyx_v_right, PyArrayObject *__pyx_v_indices, PyArrayObject *__pyx_v_subset) {
  __Pyx_LocalBuf_ND __pyx_pybuffernd_indices;
  __Pyx_Buffer __pyx_pybuffer_indices;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_left;
  __Pyx_Buffer __pyx_pybuffer_left;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_right;
  __Pyx_Buffer __pyx_pybuffer_right;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_subset;
  __Pyx_Buffer __pyx_pybuffer_subset;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyArrayObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyArrayObject *__pyx_t_7 = NULL;
  PyArrayObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  __Pyx_RefNannySetupContext("new_child_node", 0);
  __Pyx_INCREF((PyObject *)__pyx_v_left);
  __Pyx_INCREF((PyObject *)__pyx_v_right);
  __Pyx_INCREF((PyObject *)__pyx_v_indices);
  __pyx_pybuffer_left.pybuffer.buf = NULL;
  __pyx_pybuffer_left.refcount = 0;
  __pyx_pybuffernd_left.data = NULL;
  __pyx_pybuffernd_left.rcbuffer = &__pyx_pybuffer_left;
  __pyx_pybuffer_right.pybuffer.buf = NULL;
  __pyx_pybuffer_right.refcount = 0;
  __pyx_pybuffernd_right.data = NULL;
  __pyx_pybuffernd_right.rcbuffer = &__pyx_pybuffer_right;
  __pyx_pybuffer_indices.pybuffer.buf = NULL;
  __pyx_pybuffer_indices.refcount = 0;
  __pyx_pybuffernd_indices.data = NULL;
  __pyx_pybuffernd_indices.rcbuffer = &__pyx_pybuffer_indices;
  __pyx_pybuffer_subset.pybuffer.buf = NULL;
  __pyx_pybuffer_subset.refcount = 0;
  __pyx_pybuffernd_subset.data = NULL;
  __pyx_pybuffernd_subset.rcbuffer = &__pyx_pybuffer_subset;
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_left.rcbuffer->pybuffer, (PyObject*)__pyx_v_left, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float32_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 658, __pyx_L1_error)
  }
  __pyx_pybuffernd_left.diminfo[0].strides = __pyx_pybuffernd_left.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_left.diminfo[0].shape = __pyx_pybuffernd_left.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_right.rcbuffer->pybuffer, (PyObject*)__pyx_v_right, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float32_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 658, __pyx_L1_error)
  }
  __pyx_pybuffernd_right.diminfo[0].strides = __pyx_pybuffernd_right.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_right.diminfo[0].shape = __pyx_pybuffernd_right.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_indices.rcbuffer->pybuffer, (PyObject*)__pyx_v_indices, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 658, __pyx_L1_error)
  }
  __pyx_pybuffernd_indices.diminfo[0].strides = __pyx_pybuffernd_indices.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_indices.diminfo[0].shape = __pyx_pybuffernd_indices.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_subset.rcbuffer->pybuffer, (PyObject*)__pyx_v_subset, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 658, __pyx_L1_error)
  }
  __pyx_pybuffernd_subset.diminfo[0].strides = __pyx_pybuffernd_subset.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_subset.diminfo[0].shape = __pyx_pybuffernd_subset.rcbuffer->pybuffer.shape[0];

  /* "pandas/_libs/intervaltree.pxi":665
 *         """Create a new child node.
 *         """
 *         left = take(left, subset)             # <<<<<<<<<<<<<<
 *         right = take(right, subset)
 *         indices = take(indices, subset)
 */
  __pyx_t_1 = __pyx_f_6pandas_5_libs_8interval_take(((PyArrayObject *)__pyx_v_left), ((PyArrayObject *)__pyx_v_subset)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 665, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(1, 665, __pyx_L1_error)
  __pyx_t_2 = ((PyArrayObject *)__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_left.rcbuffer->pybuffer);
    __pyx_t_3 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_left.rcbuffer->pybuffer, (PyObject*)__pyx_t_2, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float32_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack);
    if (unlikely(__pyx_t_3 < 0)) {
      PyErr_Fetch(&__pyx_t_4, &__pyx_t_5, &__pyx_t_6);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_left.rcbuffer->pybuffer, (PyObject*)__pyx_v_left, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float32_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_4); Py_XDECREF(__pyx_t_5); Py_XDECREF(__pyx_t_6);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_4, __pyx_t_5, __pyx_t_6);
      }
      __pyx_t_4 = __pyx_t_5 = __pyx_t_6 = 0;
    }
    __pyx_pybuffernd_left.diminfo[0].strides = __pyx_pybuffernd_left.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_left.diminfo[0].shape = __pyx_pybuffernd_left.rcbuffer->pybuffer.shape[0];
    if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(1, 665, __pyx_L1_error)
  }
  __pyx_t_2 = 0;
  __Pyx_DECREF_SET(__pyx_v_left, ((PyArrayObject *)__pyx_t_1));
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":666
 *         """
 *         left = take(left, subset)
 *         right = take(right, subset)             # <<<<<<<<<<<<<<
 *         indices = take(indices, subset)
 *         return Float32ClosedBothIntervalNode(
 */
  __pyx_t_1 = __pyx_f_6pandas_5_libs_8interval_take(((PyArrayObject *)__pyx_v_right), ((PyArrayObject *)__pyx_v_subset)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 666, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(1, 666, __pyx_L1_error)
  __pyx_t_7 = ((PyArrayObject *)__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_right.rcbuffer->pybuffer);
    __pyx_t_3 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_right.rcbuffer->pybuffer, (PyObject*)__pyx_t_7, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float32_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack);
    if (unlikely(__pyx_t_3 < 0)) {
      PyErr_Fetch(&__pyx_t_6, &__pyx_t_5, &__pyx_t_4);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_right.rcbuffer->pybuffer, (PyObject*)__pyx_v_right, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float32_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_6); Py_XDECREF(__pyx_t_5); Py_XDECREF(__pyx_t_4);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_6, __pyx_t_5, __pyx_t_4);
      }
      __pyx_t_6 = __pyx_t_5 = __pyx_t_4 = 0;
    }
    __pyx_pybuffernd_right.diminfo[0].strides = __pyx_pybuffernd_right.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_right.diminfo[0].shape = __pyx_pybuffernd_right.rcbuffer->pybuffer.shape[0];
    if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(1, 666, __pyx_L1_error)
  }
  __pyx_t_7 = 0;
  __Pyx_DECREF_SET(__pyx_v_right, ((PyArrayObject *)__pyx_t_1));
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":667
 *         left = take(left, subset)
 *         right = take(right, subset)
 *         indices = take(indices, subset)             # <<<<<<<<<<<<<<
 *         return Float32ClosedBothIntervalNode(
 *             left, right, indices, self.leaf_size)
 */
  __pyx_t_1 = __pyx_f_6pandas_5_libs_8interval_take(((PyArrayObject *)__pyx_v_indices), ((PyArrayObject *)__pyx_v_subset)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 667, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(1, 667, __pyx_L1_error)
  __pyx_t_8 = ((PyArrayObject *)__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indices.rcbuffer->pybuffer);
    __pyx_t_3 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_indices.rcbuffer->pybuffer, (PyObject*)__pyx_t_8, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack);
    if (unlikely(__pyx_t_3 < 0)) {
      PyErr_Fetch(&__pyx_t_4, &__pyx_t_5, &__pyx_t_6);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_indices.rcbuffer->pybuffer, (PyObject*)__pyx_v_indices, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_4); Py_XDECREF(__pyx_t_5); Py_XDECREF(__pyx_t_6);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_4, __pyx_t_5, __pyx_t_6);
      }
      __pyx_t_4 = __pyx_t_5 = __pyx_t_6 = 0;
    }
    __pyx_pybuffernd_indices.diminfo[0].strides = __pyx_pybuffernd_indices.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_indices.diminfo[0].shape = __pyx_pybuffernd_indices.rcbuffer->pybuffer.shape[0];
    if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(1, 667, __pyx_L1_error)
  }
  __pyx_t_8 = 0;
  __Pyx_DECREF_SET(__pyx_v_indices, ((PyArrayObject *)__pyx_t_1));
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":668
 *         right = take(right, subset)
 *         indices = take(indices, subset)
 *         return Float32ClosedBothIntervalNode(             # <<<<<<<<<<<<<<
 *             left, right, indices, self.leaf_size)
 * 
 */
  __Pyx_XDECREF(__pyx_r);

  /* "pandas/_libs/intervaltree.pxi":669
 *         indices = take(indices, subset)
 *         return Float32ClosedBothIntervalNode(
 *             left, right, indices, self.leaf_size)             # <<<<<<<<<<<<<<
 * 
 *     @cython.wraparound(False)
 */
  __pyx_t_1 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->leaf_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 669, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "pandas/_libs/intervaltree.pxi":668
 *         right = take(right, subset)
 *         indices = take(indices, subset)
 *         return Float32ClosedBothIntervalNode(             # <<<<<<<<<<<<<<
 *             left, right, indices, self.leaf_size)
 * 
 */
  __pyx_t_9 = PyTuple_New(4); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 668, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_INCREF(((PyObject *)__pyx_v_left));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_left));
  PyTuple_SET_ITEM(__pyx_t_9, 0, ((PyObject *)__pyx_v_left));
  __Pyx_INCREF(((PyObject *)__pyx_v_right));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_right));
  PyTuple_SET_ITEM(__pyx_t_9, 1, ((PyObject *)__pyx_v_right));
  __Pyx_INCREF(((PyObject *)__pyx_v_indices));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_indices));
  PyTuple_SET_ITEM(__pyx_t_9, 2, ((PyObject *)__pyx_v_indices));
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_9, 3, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode), __pyx_t_9, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 668, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pandas/_libs/intervaltree.pxi":658
 *                 overlapping_ind.to_array())
 * 
 *     cdef new_child_node(self,             # <<<<<<<<<<<<<<
 *                         ndarray[float32_t, ndim=1] left,
 *                         ndarray[float32_t, ndim=1] right,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_9);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indices.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_left.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_right.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_subset.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("pandas._libs.interval.Float32ClosedBothIntervalNode.new_child_node", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indices.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_left.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_right.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_subset.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF((PyObject *)__pyx_v_left);
  __Pyx_XDECREF((PyObject *)__pyx_v_right);
  __Pyx_XDECREF((PyObject *)__pyx_v_indices);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":674
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_2query[] = "Recursively query this node and its sub-nodes for intervals that\n        overlap with the query point.\n        ";
static PyMethodDef __pyx_mdef_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_3query = {"query", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_3query, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_2query};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_args = 0;
  PyObject *__pyx_v_kwargs = 0;
  CYTHON_UNUSED PyObject *__pyx_v_defaults = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fused_cpdef (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_args,&__pyx_n_s_kwargs,&__pyx_n_s_defaults,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_args)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_kwargs)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fused_cpdef", 1, 3, 3, 1); __PYX_ERR(1, 674, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_defaults)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fused_cpdef", 1, 3, 3, 2); __PYX_ERR(1, 674, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fused_cpdef") < 0)) __PYX_ERR(1, 674, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_args = values[0];
    __pyx_v_kwargs = values[1];
    __pyx_v_defaults = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fused_cpdef", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 674, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Float32ClosedBothIntervalNode.__pyx_fused_cpdef", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_2query(((struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode *)__pyx_v_signatures), __pyx_v_args, __pyx_v_kwargs, __pyx_v_defaults);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_2query(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode *__pyx_v_signatures, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs, CYTHON_UNUSED PyObject *__pyx_v_defaults) {
  PyObject *__pyx_v_dest_sig = NULL;
  Py_ssize_t __pyx_v_i;
  CYTHON_UNUSED PyTypeObject *__pyx_v_ndarray = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_v_itemsize;
  PyObject *__pyx_v_arg = NULL;
  PyObject *__pyx_v_candidates = NULL;
  PyObject *__pyx_v_sig = NULL;
  int __pyx_v_match_found;
  PyObject *__pyx_v_src_sig = NULL;
  PyObject *__pyx_v_dst_type = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  Py_ssize_t __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  Py_ssize_t __pyx_t_13;
  Py_ssize_t __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  int __pyx_t_16;
  __Pyx_RefNannySetupContext("query", 0);
  __Pyx_INCREF(__pyx_v_kwargs);
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 674, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  PyList_SET_ITEM(__pyx_t_1, 0, Py_None);
  __pyx_v_dest_sig = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_3 = (__pyx_v_kwargs != Py_None);
  __pyx_t_4 = (__pyx_t_3 != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_kwargs); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(1, 674, __pyx_L1_error)
  __pyx_t_3 = ((!__pyx_t_4) != 0);
  __pyx_t_2 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {
    __Pyx_INCREF(Py_None);
    __Pyx_DECREF_SET(__pyx_v_kwargs, Py_None);
  }
  __pyx_t_1 = ((PyObject *)__Pyx_ImportNumPyArrayTypeIfAvailable()); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 674, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_ndarray = ((PyTypeObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_v_itemsize = -1L;
  if (unlikely(__pyx_v_args == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 674, __pyx_L1_error)
  }
  __pyx_t_5 = PyTuple_GET_SIZE(((PyObject*)__pyx_v_args)); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(1, 674, __pyx_L1_error)
  __pyx_t_2 = ((2 < __pyx_t_5) != 0);
  if (__pyx_t_2) {
    if (unlikely(__pyx_v_args == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 674, __pyx_L1_error)
    }
    __pyx_t_1 = PyTuple_GET_ITEM(((PyObject*)__pyx_v_args), 2);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v_arg = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L6;
  }
  __pyx_t_3 = (__pyx_v_kwargs != Py_None);
  __pyx_t_4 = (__pyx_t_3 != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L7_bool_binop_done;
  }
  if (unlikely(__pyx_v_kwargs == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(1, 674, __pyx_L1_error)
  }
  __pyx_t_4 = (__Pyx_PyDict_ContainsTF(__pyx_n_s_point, ((PyObject*)__pyx_v_kwargs), Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(1, 674, __pyx_L1_error)
  __pyx_t_3 = (__pyx_t_4 != 0);
  __pyx_t_2 = __pyx_t_3;
  __pyx_L7_bool_binop_done:;
  if (__pyx_t_2) {
    if (unlikely(__pyx_v_kwargs == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 674, __pyx_L1_error)
    }
    __pyx_t_1 = __Pyx_PyDict_GetItem(((PyObject*)__pyx_v_kwargs), __pyx_n_s_point); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 674, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_arg = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L6;
  }
  /*else*/ {
    if (unlikely(__pyx_v_args == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
      __PYX_ERR(1, 674, __pyx_L1_error)
    }
    __pyx_t_5 = PyTuple_GET_SIZE(((PyObject*)__pyx_v_args)); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(1, 674, __pyx_L1_error)
    __pyx_t_1 = PyInt_FromSsize_t(__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 674, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_6 = PyTuple_New(3); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 674, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_INCREF(__pyx_int_3);
    __Pyx_GIVEREF(__pyx_int_3);
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_int_3);
    __Pyx_INCREF(__pyx_n_s_s);
    __Pyx_GIVEREF(__pyx_n_s_s);
    PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_n_s_s);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Expected_at_least_d_argument_s_g, __pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 674, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_1); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 674, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(1, 674, __pyx_L1_error)
  }
  __pyx_L6:;
  while (1) {
    __pyx_t_2 = PyFloat_Check(__pyx_v_arg); 
    __pyx_t_3 = (__pyx_t_2 != 0);
    if (__pyx_t_3) {
      if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, __pyx_n_s_float64_t, long, 1, __Pyx_PyInt_From_long, 1, 0, 0) < 0)) __PYX_ERR(1, 674, __pyx_L1_error)
      goto __pyx_L10_break;
    }
    __pyx_t_2 = PyInt_Check(__pyx_v_arg); 
    __pyx_t_4 = (__pyx_t_2 != 0);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_3 = __pyx_t_4;
      goto __pyx_L13_bool_binop_done;
    }
    __pyx_t_4 = PyLong_Check(__pyx_v_arg); 
    __pyx_t_2 = (__pyx_t_4 != 0);
    __pyx_t_3 = __pyx_t_2;
    __pyx_L13_bool_binop_done:;
    __pyx_t_2 = (__pyx_t_3 != 0);
    if (__pyx_t_2) {
      if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, __pyx_n_s_int64_t, long, 1, __Pyx_PyInt_From_long, 1, 0, 0) < 0)) __PYX_ERR(1, 674, __pyx_L1_error)
      goto __pyx_L10_break;
    }
    if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, Py_None, long, 1, __Pyx_PyInt_From_long, 1, 0, 0) < 0)) __PYX_ERR(1, 674, __pyx_L1_error)
    goto __pyx_L10_break;
  }
  __pyx_L10_break:;
  __pyx_t_6 = PyList_New(0); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 674, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_v_candidates = ((PyObject*)__pyx_t_6);
  __pyx_t_6 = 0;
  __pyx_t_5 = 0;
  if (unlikely(((PyObject *)__pyx_v_signatures) == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(1, 674, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_dict_iterator(((PyObject*)__pyx_v_signatures), 1, ((PyObject *)NULL), (&__pyx_t_7), (&__pyx_t_8)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 674, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __pyx_t_6 = __pyx_t_1;
  __pyx_t_1 = 0;
  while (1) {
    __pyx_t_9 = __Pyx_dict_iter_next(__pyx_t_6, __pyx_t_7, &__pyx_t_5, &__pyx_t_1, NULL, NULL, __pyx_t_8);
    if (unlikely(__pyx_t_9 == 0)) break;
    if (unlikely(__pyx_t_9 == -1)) __PYX_ERR(1, 674, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_sig, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_v_match_found = 0;
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_v_sig, __pyx_n_s_strip); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 674, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __pyx_t_12 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_11))) {
      __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_11);
      if (likely(__pyx_t_12)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
        __Pyx_INCREF(__pyx_t_12);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_11, function);
      }
    }
    __pyx_t_10 = (__pyx_t_12) ? __Pyx_PyObject_Call2Args(__pyx_t_11, __pyx_t_12, __pyx_kp_s__13) : __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_kp_s__13);
    __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
    if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 674, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_split); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 674, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_t_10 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_11))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_11);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_11, function);
      }
    }
    __pyx_t_1 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_11, __pyx_t_10, __pyx_kp_s__14) : __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_kp_s__14);
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 674, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_XDECREF_SET(__pyx_v_src_sig, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_13 = PyList_GET_SIZE(__pyx_v_dest_sig); if (unlikely(__pyx_t_13 == ((Py_ssize_t)-1))) __PYX_ERR(1, 674, __pyx_L1_error)
    __pyx_t_14 = __pyx_t_13;
    for (__pyx_t_15 = 0; __pyx_t_15 < __pyx_t_14; __pyx_t_15+=1) {
      __pyx_v_i = __pyx_t_15;
      __pyx_t_1 = PyList_GET_ITEM(__pyx_v_dest_sig, __pyx_v_i);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_XDECREF_SET(__pyx_v_dst_type, __pyx_t_1);
      __pyx_t_1 = 0;
      __pyx_t_2 = (__pyx_v_dst_type != Py_None);
      __pyx_t_3 = (__pyx_t_2 != 0);
      if (__pyx_t_3) {
        __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_src_sig, __pyx_v_i, Py_ssize_t, 1, PyInt_FromSsize_t, 0, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 674, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_11 = PyObject_RichCompare(__pyx_t_1, __pyx_v_dst_type, Py_EQ); __Pyx_XGOTREF(__pyx_t_11); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 674, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_11); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(1, 674, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        if (__pyx_t_3) {
          __pyx_v_match_found = 1;
          goto __pyx_L20;
        }
        /*else*/ {
          __pyx_v_match_found = 0;
          goto __pyx_L18_break;
        }
        __pyx_L20:;
      }
    }
    __pyx_L18_break:;
    __pyx_t_3 = (__pyx_v_match_found != 0);
    if (__pyx_t_3) {
      __pyx_t_16 = __Pyx_PyList_Append(__pyx_v_candidates, __pyx_v_sig); if (unlikely(__pyx_t_16 == ((int)-1))) __PYX_ERR(1, 674, __pyx_L1_error)
    }
  }
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_3 = (PyList_GET_SIZE(__pyx_v_candidates) != 0);
  __pyx_t_2 = ((!__pyx_t_3) != 0);
  if (__pyx_t_2) {
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__15, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 674, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(1, 674, __pyx_L1_error)
  }
  __pyx_t_7 = PyList_GET_SIZE(__pyx_v_candidates); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1))) __PYX_ERR(1, 674, __pyx_L1_error)
  __pyx_t_2 = ((__pyx_t_7 > 1) != 0);
  if (__pyx_t_2) {
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__16, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 674, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(1, 674, __pyx_L1_error)
  }
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    if (unlikely(((PyObject *)__pyx_v_signatures) == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 674, __pyx_L1_error)
    }
    __pyx_t_6 = __Pyx_PyDict_GetItem(((PyObject*)__pyx_v_signatures), PyList_GET_ITEM(__pyx_v_candidates, 0)); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 674, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_r = __pyx_t_6;
    __pyx_t_6 = 0;
    goto __pyx_L0;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("pandas._libs.interval.Float32ClosedBothIntervalNode.__pyx_fused_cpdef", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_dest_sig);
  __Pyx_XDECREF(__pyx_v_ndarray);
  __Pyx_XDECREF(__pyx_v_arg);
  __Pyx_XDECREF(__pyx_v_candidates);
  __Pyx_XDECREF(__pyx_v_sig);
  __Pyx_XDECREF(__pyx_v_src_sig);
  __Pyx_XDECREF(__pyx_v_dst_type);
  __Pyx_XDECREF(__pyx_v_kwargs);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_9__pyx_fuse_0query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_fuse_0__pyx_f_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float64_t __pyx_v_point, int __pyx_skip_dispatch) {
  __Pyx_memviewslice __pyx_v_indices = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_values = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __pyx_t_5numpy_int64_t __pyx_t_9;
  __pyx_t_5numpy_int64_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  int __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  __Pyx_memviewslice __pyx_t_16 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_17 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  Py_ssize_t __pyx_t_21;
  __Pyx_RefNannySetupContext("__pyx_fuse_0query", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pyx_fuse_0query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 674, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_9__pyx_fuse_0query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = PyFloat_FromDouble(__pyx_v_point); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 674, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 674, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 674, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 674, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_5) {
            __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 674, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 674, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_3query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_4 = PyFloat_FromDouble(__pyx_v_point); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 674, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_7 = __pyx_t_1; __pyx_t_3 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 674, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 674, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 674, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (__pyx_t_3) {
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_t_4);
          __pyx_t_4 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 674, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pandas/_libs/intervaltree.pxi":683
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
  __pyx_t_8 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_8) {

    /* "pandas/_libs/intervaltree.pxi":687
 *             # continue the binary tree structure. Instead, we use linear
 *             # search.
 *             for i in range(self.n_elements):             # <<<<<<<<<<<<<<
 *                 if self.left[i] <= point <= self.right[i]:
 *                     result.append(self.indices[i])
 */
    __pyx_t_9 = __pyx_v_self->n_elements;
    __pyx_t_10 = __pyx_t_9;
    for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
      __pyx_v_i = __pyx_t_11;

      /* "pandas/_libs/intervaltree.pxi":688
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point <= self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      __pyx_t_12 = __pyx_v_i;
      __pyx_t_8 = ((*((__pyx_t_5numpy_float32_t *) ( /* dim=0 */ (__pyx_v_self->left.data + __pyx_t_12 * __pyx_v_self->left.strides[0]) ))) <= __pyx_v_point);
      if (__pyx_t_8) {
        __pyx_t_13 = __pyx_v_i;
        __pyx_t_8 = (__pyx_v_point <= (*((__pyx_t_5numpy_float32_t *) ( /* dim=0 */ (__pyx_v_self->right.data + __pyx_t_13 * __pyx_v_self->right.strides[0]) ))));
      }
      __pyx_t_14 = (__pyx_t_8 != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":689
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point <= self.right[i]:
 *                     result.append(self.indices[i])             # <<<<<<<<<<<<<<
 *         else:
 *             # There are child nodes. Based on comparing our query to the pivot,
 */
        __pyx_t_15 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->indices.data + __pyx_t_15 * __pyx_v_self->indices.strides[0]) ))));

        /* "pandas/_libs/intervaltree.pxi":688
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point <= self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      }
    }

    /* "pandas/_libs/intervaltree.pxi":683
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
    goto __pyx_L3;
  }

  /* "pandas/_libs/intervaltree.pxi":693
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
  /*else*/ {
    __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":694
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:
 *                 values = self.center_left_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 */
      __pyx_t_16 = __pyx_v_self->center_left_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":695
 *             if point < self.pivot:
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:
 */
      __pyx_t_17 = __pyx_v_self->center_left_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":696
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):             # <<<<<<<<<<<<<<
 *                     if not values[i] <= point:
 *                         break
 */
      __pyx_t_9 = __pyx_v_self->n_center;
      __pyx_t_10 = __pyx_t_9;
      for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":697
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_18 = __pyx_v_i;
        __pyx_t_14 = ((!(((*((__pyx_t_5numpy_float32_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_18 * __pyx_v_values.strides[0]) ))) <= __pyx_v_point) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":698
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:
 */
          goto __pyx_L9_break;

          /* "pandas/_libs/intervaltree.pxi":697
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":699
 *                     if not values[i] <= point:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 */
        __pyx_t_19 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_19 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L9_break:;

      /* "pandas/_libs/intervaltree.pxi":700
 *                         break
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      __pyx_t_14 = ((__pyx_v_point <= __pyx_v_self->left_node->max_right) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":701
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)             # <<<<<<<<<<<<<<
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode *)__pyx_v_self->left_node->__pyx_vtab)->__pyx_fuse_0query(__pyx_v_self->left_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 701, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":700
 *                         break
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":693
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":702
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
    __pyx_t_14 = ((__pyx_v_point > __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":703
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 *                 values = self.center_right_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 */
      __pyx_t_16 = __pyx_v_self->center_right_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":704
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:
 */
      __pyx_t_17 = __pyx_v_self->center_right_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":705
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):             # <<<<<<<<<<<<<<
 *                     if not point <= values[i]:
 *                         break
 */
      for (__pyx_t_11 = (__pyx_v_self->n_center - 1); __pyx_t_11 > -1L; __pyx_t_11-=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":706
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_20 = __pyx_v_i;
        __pyx_t_14 = ((!((__pyx_v_point <= (*((__pyx_t_5numpy_float32_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_20 * __pyx_v_values.strides[0]) )))) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":707
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:
 */
          goto __pyx_L13_break;

          /* "pandas/_libs/intervaltree.pxi":706
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":708
 *                     if not point <= values[i]:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if self.right_node.min_left <= point:
 *                     self.right_node.query(result, point)
 */
        __pyx_t_21 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_21 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L13_break:;

      /* "pandas/_libs/intervaltree.pxi":709
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      __pyx_t_14 = ((__pyx_v_self->right_node->min_left <= __pyx_v_point) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":710
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:
 *                     self.right_node.query(result, point)             # <<<<<<<<<<<<<<
 *             else:
 *                 result.extend(self.center_left_indices)
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode *)__pyx_v_self->right_node->__pyx_vtab)->__pyx_fuse_0query(__pyx_v_self->right_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 710, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":709
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":702
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":712
 *                     self.right_node.query(result, point)
 *             else:
 *                 result.extend(self.center_left_indices)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
    /*else*/ {
      __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->extend(__pyx_v_result, __pyx_v_self->center_left_indices); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 712, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_L7:;
  }
  __pyx_L3:;

  /* "pandas/_libs/intervaltree.pxi":674
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __PYX_XDEC_MEMVIEW(&__pyx_t_16, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_17, 1);
  __Pyx_AddTraceback("pandas._libs.interval.Float32ClosedBothIntervalNode.query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_indices, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_values, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_9__pyx_fuse_0query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_0__pyx_mdef_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_9__pyx_fuse_0query = {"__pyx_fuse_0query", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_9__pyx_fuse_0query, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_2query};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_9__pyx_fuse_0query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result = 0;
  __pyx_t_5numpy_float64_t __pyx_v_point;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fuse_0query (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_result,&__pyx_n_s_point,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_result)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_point)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_0query", 1, 2, 2, 1); __PYX_ERR(1, 674, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_0query") < 0)) __PYX_ERR(1, 674, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_result = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)values[0]);
    __pyx_v_point = __pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_point == ((npy_float64)-1)) && PyErr_Occurred())) __PYX_ERR(1, 674, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_0query", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 674, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Float32ClosedBothIntervalNode.__pyx_fuse_0query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_result), __pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector, 1, "result", 0))) __PYX_ERR(1, 674, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_8__pyx_fuse_0query(((struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode *)__pyx_v_self), __pyx_v_result, __pyx_v_point);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_8__pyx_fuse_0query(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float64_t __pyx_v_point) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__pyx_fuse_0query", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_fuse_0__pyx_f_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_query(__pyx_v_self, __pyx_v_result, __pyx_v_point, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 674, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Float32ClosedBothIntervalNode.__pyx_fuse_0query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_11__pyx_fuse_1query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_fuse_1__pyx_f_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float32_t __pyx_v_point, int __pyx_skip_dispatch) {
  __Pyx_memviewslice __pyx_v_indices = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_values = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __pyx_t_5numpy_int64_t __pyx_t_9;
  __pyx_t_5numpy_int64_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  int __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  __Pyx_memviewslice __pyx_t_16 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_17 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  Py_ssize_t __pyx_t_21;
  __Pyx_RefNannySetupContext("__pyx_fuse_1query", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pyx_fuse_1query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 674, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_11__pyx_fuse_1query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = PyFloat_FromDouble(__pyx_v_point); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 674, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 674, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 674, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 674, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_5) {
            __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 674, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 674, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_3query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_4 = PyFloat_FromDouble(__pyx_v_point); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 674, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_7 = __pyx_t_1; __pyx_t_3 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 674, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 674, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 674, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (__pyx_t_3) {
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_t_4);
          __pyx_t_4 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 674, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pandas/_libs/intervaltree.pxi":683
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
  __pyx_t_8 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_8) {

    /* "pandas/_libs/intervaltree.pxi":687
 *             # continue the binary tree structure. Instead, we use linear
 *             # search.
 *             for i in range(self.n_elements):             # <<<<<<<<<<<<<<
 *                 if self.left[i] <= point <= self.right[i]:
 *                     result.append(self.indices[i])
 */
    __pyx_t_9 = __pyx_v_self->n_elements;
    __pyx_t_10 = __pyx_t_9;
    for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
      __pyx_v_i = __pyx_t_11;

      /* "pandas/_libs/intervaltree.pxi":688
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point <= self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      __pyx_t_12 = __pyx_v_i;
      __pyx_t_8 = ((*((__pyx_t_5numpy_float32_t *) ( /* dim=0 */ (__pyx_v_self->left.data + __pyx_t_12 * __pyx_v_self->left.strides[0]) ))) <= __pyx_v_point);
      if (__pyx_t_8) {
        __pyx_t_13 = __pyx_v_i;
        __pyx_t_8 = (__pyx_v_point <= (*((__pyx_t_5numpy_float32_t *) ( /* dim=0 */ (__pyx_v_self->right.data + __pyx_t_13 * __pyx_v_self->right.strides[0]) ))));
      }
      __pyx_t_14 = (__pyx_t_8 != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":689
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point <= self.right[i]:
 *                     result.append(self.indices[i])             # <<<<<<<<<<<<<<
 *         else:
 *             # There are child nodes. Based on comparing our query to the pivot,
 */
        __pyx_t_15 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->indices.data + __pyx_t_15 * __pyx_v_self->indices.strides[0]) ))));

        /* "pandas/_libs/intervaltree.pxi":688
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point <= self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      }
    }

    /* "pandas/_libs/intervaltree.pxi":683
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
    goto __pyx_L3;
  }

  /* "pandas/_libs/intervaltree.pxi":693
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
  /*else*/ {
    __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":694
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:
 *                 values = self.center_left_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 */
      __pyx_t_16 = __pyx_v_self->center_left_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":695
 *             if point < self.pivot:
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:
 */
      __pyx_t_17 = __pyx_v_self->center_left_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":696
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):             # <<<<<<<<<<<<<<
 *                     if not values[i] <= point:
 *                         break
 */
      __pyx_t_9 = __pyx_v_self->n_center;
      __pyx_t_10 = __pyx_t_9;
      for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":697
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_18 = __pyx_v_i;
        __pyx_t_14 = ((!(((*((__pyx_t_5numpy_float32_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_18 * __pyx_v_values.strides[0]) ))) <= __pyx_v_point) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":698
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:
 */
          goto __pyx_L9_break;

          /* "pandas/_libs/intervaltree.pxi":697
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":699
 *                     if not values[i] <= point:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 */
        __pyx_t_19 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_19 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L9_break:;

      /* "pandas/_libs/intervaltree.pxi":700
 *                         break
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      __pyx_t_14 = ((__pyx_v_point <= __pyx_v_self->left_node->max_right) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":701
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)             # <<<<<<<<<<<<<<
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode *)__pyx_v_self->left_node->__pyx_vtab)->__pyx_fuse_1query(__pyx_v_self->left_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 701, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":700
 *                         break
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":693
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":702
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
    __pyx_t_14 = ((__pyx_v_point > __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":703
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 *                 values = self.center_right_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 */
      __pyx_t_16 = __pyx_v_self->center_right_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":704
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:
 */
      __pyx_t_17 = __pyx_v_self->center_right_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":705
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):             # <<<<<<<<<<<<<<
 *                     if not point <= values[i]:
 *                         break
 */
      for (__pyx_t_11 = (__pyx_v_self->n_center - 1); __pyx_t_11 > -1L; __pyx_t_11-=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":706
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_20 = __pyx_v_i;
        __pyx_t_14 = ((!((__pyx_v_point <= (*((__pyx_t_5numpy_float32_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_20 * __pyx_v_values.strides[0]) )))) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":707
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:
 */
          goto __pyx_L13_break;

          /* "pandas/_libs/intervaltree.pxi":706
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":708
 *                     if not point <= values[i]:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if self.right_node.min_left <= point:
 *                     self.right_node.query(result, point)
 */
        __pyx_t_21 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_21 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L13_break:;

      /* "pandas/_libs/intervaltree.pxi":709
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      __pyx_t_14 = ((__pyx_v_self->right_node->min_left <= __pyx_v_point) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":710
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:
 *                     self.right_node.query(result, point)             # <<<<<<<<<<<<<<
 *             else:
 *                 result.extend(self.center_left_indices)
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode *)__pyx_v_self->right_node->__pyx_vtab)->__pyx_fuse_1query(__pyx_v_self->right_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 710, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":709
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":702
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":712
 *                     self.right_node.query(result, point)
 *             else:
 *                 result.extend(self.center_left_indices)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
    /*else*/ {
      __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->extend(__pyx_v_result, __pyx_v_self->center_left_indices); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 712, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_L7:;
  }
  __pyx_L3:;

  /* "pandas/_libs/intervaltree.pxi":674
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __PYX_XDEC_MEMVIEW(&__pyx_t_16, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_17, 1);
  __Pyx_AddTraceback("pandas._libs.interval.Float32ClosedBothIntervalNode.query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_indices, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_values, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_11__pyx_fuse_1query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_1__pyx_mdef_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_11__pyx_fuse_1query = {"__pyx_fuse_1query", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_11__pyx_fuse_1query, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_2query};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_11__pyx_fuse_1query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result = 0;
  __pyx_t_5numpy_float32_t __pyx_v_point;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fuse_1query (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_result,&__pyx_n_s_point,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_result)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_point)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_1query", 1, 2, 2, 1); __PYX_ERR(1, 674, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_1query") < 0)) __PYX_ERR(1, 674, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_result = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)values[0]);
    __pyx_v_point = __pyx_PyFloat_AsFloat(values[1]); if (unlikely((__pyx_v_point == ((npy_float32)-1)) && PyErr_Occurred())) __PYX_ERR(1, 674, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_1query", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 674, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Float32ClosedBothIntervalNode.__pyx_fuse_1query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_result), __pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector, 1, "result", 0))) __PYX_ERR(1, 674, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_10__pyx_fuse_1query(((struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode *)__pyx_v_self), __pyx_v_result, __pyx_v_point);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_10__pyx_fuse_1query(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float32_t __pyx_v_point) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__pyx_fuse_1query", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_fuse_1__pyx_f_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_query(__pyx_v_self, __pyx_v_result, __pyx_v_point, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 674, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Float32ClosedBothIntervalNode.__pyx_fuse_1query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_13__pyx_fuse_2query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_fuse_2__pyx_f_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int64_t __pyx_v_point, int __pyx_skip_dispatch) {
  __Pyx_memviewslice __pyx_v_indices = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_values = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __pyx_t_5numpy_int64_t __pyx_t_9;
  __pyx_t_5numpy_int64_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  int __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  __Pyx_memviewslice __pyx_t_16 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_17 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  Py_ssize_t __pyx_t_21;
  __Pyx_RefNannySetupContext("__pyx_fuse_2query", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pyx_fuse_2query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 674, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_13__pyx_fuse_2query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = __Pyx_PyInt_From_npy_int64(__pyx_v_point); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 674, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 674, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 674, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 674, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_5) {
            __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 674, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 674, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_3query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_4 = __Pyx_PyInt_From_npy_int64(__pyx_v_point); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 674, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_7 = __pyx_t_1; __pyx_t_3 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 674, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 674, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 674, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (__pyx_t_3) {
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_t_4);
          __pyx_t_4 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 674, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pandas/_libs/intervaltree.pxi":683
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
  __pyx_t_8 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_8) {

    /* "pandas/_libs/intervaltree.pxi":687
 *             # continue the binary tree structure. Instead, we use linear
 *             # search.
 *             for i in range(self.n_elements):             # <<<<<<<<<<<<<<
 *                 if self.left[i] <= point <= self.right[i]:
 *                     result.append(self.indices[i])
 */
    __pyx_t_9 = __pyx_v_self->n_elements;
    __pyx_t_10 = __pyx_t_9;
    for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
      __pyx_v_i = __pyx_t_11;

      /* "pandas/_libs/intervaltree.pxi":688
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point <= self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      __pyx_t_12 = __pyx_v_i;
      __pyx_t_8 = ((*((__pyx_t_5numpy_float32_t *) ( /* dim=0 */ (__pyx_v_self->left.data + __pyx_t_12 * __pyx_v_self->left.strides[0]) ))) <= __pyx_v_point);
      if (__pyx_t_8) {
        __pyx_t_13 = __pyx_v_i;
        __pyx_t_8 = (__pyx_v_point <= (*((__pyx_t_5numpy_float32_t *) ( /* dim=0 */ (__pyx_v_self->right.data + __pyx_t_13 * __pyx_v_self->right.strides[0]) ))));
      }
      __pyx_t_14 = (__pyx_t_8 != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":689
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point <= self.right[i]:
 *                     result.append(self.indices[i])             # <<<<<<<<<<<<<<
 *         else:
 *             # There are child nodes. Based on comparing our query to the pivot,
 */
        __pyx_t_15 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->indices.data + __pyx_t_15 * __pyx_v_self->indices.strides[0]) ))));

        /* "pandas/_libs/intervaltree.pxi":688
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point <= self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      }
    }

    /* "pandas/_libs/intervaltree.pxi":683
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
    goto __pyx_L3;
  }

  /* "pandas/_libs/intervaltree.pxi":693
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
  /*else*/ {
    __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":694
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:
 *                 values = self.center_left_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 */
      __pyx_t_16 = __pyx_v_self->center_left_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":695
 *             if point < self.pivot:
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:
 */
      __pyx_t_17 = __pyx_v_self->center_left_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":696
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):             # <<<<<<<<<<<<<<
 *                     if not values[i] <= point:
 *                         break
 */
      __pyx_t_9 = __pyx_v_self->n_center;
      __pyx_t_10 = __pyx_t_9;
      for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":697
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_18 = __pyx_v_i;
        __pyx_t_14 = ((!(((*((__pyx_t_5numpy_float32_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_18 * __pyx_v_values.strides[0]) ))) <= __pyx_v_point) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":698
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:
 */
          goto __pyx_L9_break;

          /* "pandas/_libs/intervaltree.pxi":697
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":699
 *                     if not values[i] <= point:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 */
        __pyx_t_19 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_19 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L9_break:;

      /* "pandas/_libs/intervaltree.pxi":700
 *                         break
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      __pyx_t_14 = ((__pyx_v_point <= __pyx_v_self->left_node->max_right) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":701
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)             # <<<<<<<<<<<<<<
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode *)__pyx_v_self->left_node->__pyx_vtab)->__pyx_fuse_2query(__pyx_v_self->left_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 701, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":700
 *                         break
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":693
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":702
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
    __pyx_t_14 = ((__pyx_v_point > __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":703
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 *                 values = self.center_right_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 */
      __pyx_t_16 = __pyx_v_self->center_right_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":704
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:
 */
      __pyx_t_17 = __pyx_v_self->center_right_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":705
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):             # <<<<<<<<<<<<<<
 *                     if not point <= values[i]:
 *                         break
 */
      for (__pyx_t_11 = (__pyx_v_self->n_center - 1); __pyx_t_11 > -1L; __pyx_t_11-=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":706
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_20 = __pyx_v_i;
        __pyx_t_14 = ((!((__pyx_v_point <= (*((__pyx_t_5numpy_float32_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_20 * __pyx_v_values.strides[0]) )))) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":707
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:
 */
          goto __pyx_L13_break;

          /* "pandas/_libs/intervaltree.pxi":706
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":708
 *                     if not point <= values[i]:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if self.right_node.min_left <= point:
 *                     self.right_node.query(result, point)
 */
        __pyx_t_21 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_21 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L13_break:;

      /* "pandas/_libs/intervaltree.pxi":709
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      __pyx_t_14 = ((__pyx_v_self->right_node->min_left <= __pyx_v_point) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":710
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:
 *                     self.right_node.query(result, point)             # <<<<<<<<<<<<<<
 *             else:
 *                 result.extend(self.center_left_indices)
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode *)__pyx_v_self->right_node->__pyx_vtab)->__pyx_fuse_2query(__pyx_v_self->right_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 710, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":709
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":702
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":712
 *                     self.right_node.query(result, point)
 *             else:
 *                 result.extend(self.center_left_indices)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
    /*else*/ {
      __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->extend(__pyx_v_result, __pyx_v_self->center_left_indices); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 712, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_L7:;
  }
  __pyx_L3:;

  /* "pandas/_libs/intervaltree.pxi":674
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __PYX_XDEC_MEMVIEW(&__pyx_t_16, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_17, 1);
  __Pyx_AddTraceback("pandas._libs.interval.Float32ClosedBothIntervalNode.query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_indices, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_values, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_13__pyx_fuse_2query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_2__pyx_mdef_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_13__pyx_fuse_2query = {"__pyx_fuse_2query", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_13__pyx_fuse_2query, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_2query};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_13__pyx_fuse_2query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result = 0;
  __pyx_t_5numpy_int64_t __pyx_v_point;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fuse_2query (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_result,&__pyx_n_s_point,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_result)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_point)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_2query", 1, 2, 2, 1); __PYX_ERR(1, 674, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_2query") < 0)) __PYX_ERR(1, 674, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_result = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)values[0]);
    __pyx_v_point = __Pyx_PyInt_As_npy_int64(values[1]); if (unlikely((__pyx_v_point == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(1, 674, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_2query", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 674, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Float32ClosedBothIntervalNode.__pyx_fuse_2query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_result), __pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector, 1, "result", 0))) __PYX_ERR(1, 674, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_12__pyx_fuse_2query(((struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode *)__pyx_v_self), __pyx_v_result, __pyx_v_point);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_12__pyx_fuse_2query(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int64_t __pyx_v_point) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__pyx_fuse_2query", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_fuse_2__pyx_f_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_query(__pyx_v_self, __pyx_v_result, __pyx_v_point, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 674, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Float32ClosedBothIntervalNode.__pyx_fuse_2query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_15__pyx_fuse_3query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_fuse_3__pyx_f_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int32_t __pyx_v_point, int __pyx_skip_dispatch) {
  __Pyx_memviewslice __pyx_v_indices = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_values = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __pyx_t_5numpy_int64_t __pyx_t_9;
  __pyx_t_5numpy_int64_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  int __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  __Pyx_memviewslice __pyx_t_16 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_17 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  Py_ssize_t __pyx_t_21;
  __Pyx_RefNannySetupContext("__pyx_fuse_3query", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pyx_fuse_3query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 674, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_15__pyx_fuse_3query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = __Pyx_PyInt_From_npy_int32(__pyx_v_point); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 674, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 674, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 674, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 674, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_5) {
            __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 674, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 674, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_3query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_4 = __Pyx_PyInt_From_npy_int32(__pyx_v_point); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 674, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_7 = __pyx_t_1; __pyx_t_3 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 674, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 674, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 674, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (__pyx_t_3) {
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_t_4);
          __pyx_t_4 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 674, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pandas/_libs/intervaltree.pxi":683
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
  __pyx_t_8 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_8) {

    /* "pandas/_libs/intervaltree.pxi":687
 *             # continue the binary tree structure. Instead, we use linear
 *             # search.
 *             for i in range(self.n_elements):             # <<<<<<<<<<<<<<
 *                 if self.left[i] <= point <= self.right[i]:
 *                     result.append(self.indices[i])
 */
    __pyx_t_9 = __pyx_v_self->n_elements;
    __pyx_t_10 = __pyx_t_9;
    for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
      __pyx_v_i = __pyx_t_11;

      /* "pandas/_libs/intervaltree.pxi":688
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point <= self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      __pyx_t_12 = __pyx_v_i;
      __pyx_t_8 = ((*((__pyx_t_5numpy_float32_t *) ( /* dim=0 */ (__pyx_v_self->left.data + __pyx_t_12 * __pyx_v_self->left.strides[0]) ))) <= __pyx_v_point);
      if (__pyx_t_8) {
        __pyx_t_13 = __pyx_v_i;
        __pyx_t_8 = (__pyx_v_point <= (*((__pyx_t_5numpy_float32_t *) ( /* dim=0 */ (__pyx_v_self->right.data + __pyx_t_13 * __pyx_v_self->right.strides[0]) ))));
      }
      __pyx_t_14 = (__pyx_t_8 != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":689
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point <= self.right[i]:
 *                     result.append(self.indices[i])             # <<<<<<<<<<<<<<
 *         else:
 *             # There are child nodes. Based on comparing our query to the pivot,
 */
        __pyx_t_15 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->indices.data + __pyx_t_15 * __pyx_v_self->indices.strides[0]) ))));

        /* "pandas/_libs/intervaltree.pxi":688
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point <= self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      }
    }

    /* "pandas/_libs/intervaltree.pxi":683
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
    goto __pyx_L3;
  }

  /* "pandas/_libs/intervaltree.pxi":693
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
  /*else*/ {
    __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":694
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:
 *                 values = self.center_left_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 */
      __pyx_t_16 = __pyx_v_self->center_left_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":695
 *             if point < self.pivot:
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:
 */
      __pyx_t_17 = __pyx_v_self->center_left_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":696
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):             # <<<<<<<<<<<<<<
 *                     if not values[i] <= point:
 *                         break
 */
      __pyx_t_9 = __pyx_v_self->n_center;
      __pyx_t_10 = __pyx_t_9;
      for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":697
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_18 = __pyx_v_i;
        __pyx_t_14 = ((!(((*((__pyx_t_5numpy_float32_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_18 * __pyx_v_values.strides[0]) ))) <= __pyx_v_point) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":698
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:
 */
          goto __pyx_L9_break;

          /* "pandas/_libs/intervaltree.pxi":697
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":699
 *                     if not values[i] <= point:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 */
        __pyx_t_19 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_19 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L9_break:;

      /* "pandas/_libs/intervaltree.pxi":700
 *                         break
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      __pyx_t_14 = ((__pyx_v_point <= __pyx_v_self->left_node->max_right) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":701
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)             # <<<<<<<<<<<<<<
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode *)__pyx_v_self->left_node->__pyx_vtab)->__pyx_fuse_3query(__pyx_v_self->left_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 701, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":700
 *                         break
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":693
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":702
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
    __pyx_t_14 = ((__pyx_v_point > __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":703
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 *                 values = self.center_right_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 */
      __pyx_t_16 = __pyx_v_self->center_right_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":704
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:
 */
      __pyx_t_17 = __pyx_v_self->center_right_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":705
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):             # <<<<<<<<<<<<<<
 *                     if not point <= values[i]:
 *                         break
 */
      for (__pyx_t_11 = (__pyx_v_self->n_center - 1); __pyx_t_11 > -1L; __pyx_t_11-=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":706
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_20 = __pyx_v_i;
        __pyx_t_14 = ((!((__pyx_v_point <= (*((__pyx_t_5numpy_float32_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_20 * __pyx_v_values.strides[0]) )))) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":707
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:
 */
          goto __pyx_L13_break;

          /* "pandas/_libs/intervaltree.pxi":706
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":708
 *                     if not point <= values[i]:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if self.right_node.min_left <= point:
 *                     self.right_node.query(result, point)
 */
        __pyx_t_21 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_21 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L13_break:;

      /* "pandas/_libs/intervaltree.pxi":709
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      __pyx_t_14 = ((__pyx_v_self->right_node->min_left <= __pyx_v_point) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":710
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:
 *                     self.right_node.query(result, point)             # <<<<<<<<<<<<<<
 *             else:
 *                 result.extend(self.center_left_indices)
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode *)__pyx_v_self->right_node->__pyx_vtab)->__pyx_fuse_3query(__pyx_v_self->right_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 710, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":709
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":702
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":712
 *                     self.right_node.query(result, point)
 *             else:
 *                 result.extend(self.center_left_indices)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
    /*else*/ {
      __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->extend(__pyx_v_result, __pyx_v_self->center_left_indices); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 712, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_L7:;
  }
  __pyx_L3:;

  /* "pandas/_libs/intervaltree.pxi":674
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __PYX_XDEC_MEMVIEW(&__pyx_t_16, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_17, 1);
  __Pyx_AddTraceback("pandas._libs.interval.Float32ClosedBothIntervalNode.query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_indices, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_values, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_15__pyx_fuse_3query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_3__pyx_mdef_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_15__pyx_fuse_3query = {"__pyx_fuse_3query", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_15__pyx_fuse_3query, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_2query};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_15__pyx_fuse_3query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result = 0;
  __pyx_t_5numpy_int32_t __pyx_v_point;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fuse_3query (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_result,&__pyx_n_s_point,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_result)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_point)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_3query", 1, 2, 2, 1); __PYX_ERR(1, 674, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_3query") < 0)) __PYX_ERR(1, 674, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_result = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)values[0]);
    __pyx_v_point = __Pyx_PyInt_As_npy_int32(values[1]); if (unlikely((__pyx_v_point == ((npy_int32)-1)) && PyErr_Occurred())) __PYX_ERR(1, 674, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_3query", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 674, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Float32ClosedBothIntervalNode.__pyx_fuse_3query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_result), __pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector, 1, "result", 0))) __PYX_ERR(1, 674, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_14__pyx_fuse_3query(((struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode *)__pyx_v_self), __pyx_v_result, __pyx_v_point);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_14__pyx_fuse_3query(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int32_t __pyx_v_point) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__pyx_fuse_3query", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_fuse_3__pyx_f_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_query(__pyx_v_self, __pyx_v_result, __pyx_v_point, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 674, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Float32ClosedBothIntervalNode.__pyx_fuse_3query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_17__pyx_fuse_4query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_fuse_4__pyx_f_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_uint64_t __pyx_v_point, int __pyx_skip_dispatch) {
  __Pyx_memviewslice __pyx_v_indices = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_values = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __pyx_t_5numpy_int64_t __pyx_t_9;
  __pyx_t_5numpy_int64_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  int __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  __Pyx_memviewslice __pyx_t_16 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_17 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  Py_ssize_t __pyx_t_21;
  __Pyx_RefNannySetupContext("__pyx_fuse_4query", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pyx_fuse_4query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 674, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_17__pyx_fuse_4query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = __Pyx_PyInt_From_npy_uint64(__pyx_v_point); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 674, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 674, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 674, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 674, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_5) {
            __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 674, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 674, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_3query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_4 = __Pyx_PyInt_From_npy_uint64(__pyx_v_point); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 674, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_7 = __pyx_t_1; __pyx_t_3 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 674, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 674, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 674, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (__pyx_t_3) {
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_t_4);
          __pyx_t_4 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 674, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pandas/_libs/intervaltree.pxi":683
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
  __pyx_t_8 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_8) {

    /* "pandas/_libs/intervaltree.pxi":687
 *             # continue the binary tree structure. Instead, we use linear
 *             # search.
 *             for i in range(self.n_elements):             # <<<<<<<<<<<<<<
 *                 if self.left[i] <= point <= self.right[i]:
 *                     result.append(self.indices[i])
 */
    __pyx_t_9 = __pyx_v_self->n_elements;
    __pyx_t_10 = __pyx_t_9;
    for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
      __pyx_v_i = __pyx_t_11;

      /* "pandas/_libs/intervaltree.pxi":688
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point <= self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      __pyx_t_12 = __pyx_v_i;
      __pyx_t_8 = ((*((__pyx_t_5numpy_float32_t *) ( /* dim=0 */ (__pyx_v_self->left.data + __pyx_t_12 * __pyx_v_self->left.strides[0]) ))) <= __pyx_v_point);
      if (__pyx_t_8) {
        __pyx_t_13 = __pyx_v_i;
        __pyx_t_8 = (__pyx_v_point <= (*((__pyx_t_5numpy_float32_t *) ( /* dim=0 */ (__pyx_v_self->right.data + __pyx_t_13 * __pyx_v_self->right.strides[0]) ))));
      }
      __pyx_t_14 = (__pyx_t_8 != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":689
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point <= self.right[i]:
 *                     result.append(self.indices[i])             # <<<<<<<<<<<<<<
 *         else:
 *             # There are child nodes. Based on comparing our query to the pivot,
 */
        __pyx_t_15 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->indices.data + __pyx_t_15 * __pyx_v_self->indices.strides[0]) ))));

        /* "pandas/_libs/intervaltree.pxi":688
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point <= self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      }
    }

    /* "pandas/_libs/intervaltree.pxi":683
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
    goto __pyx_L3;
  }

  /* "pandas/_libs/intervaltree.pxi":693
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
  /*else*/ {
    __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":694
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:
 *                 values = self.center_left_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 */
      __pyx_t_16 = __pyx_v_self->center_left_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":695
 *             if point < self.pivot:
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:
 */
      __pyx_t_17 = __pyx_v_self->center_left_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":696
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):             # <<<<<<<<<<<<<<
 *                     if not values[i] <= point:
 *                         break
 */
      __pyx_t_9 = __pyx_v_self->n_center;
      __pyx_t_10 = __pyx_t_9;
      for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":697
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_18 = __pyx_v_i;
        __pyx_t_14 = ((!(((*((__pyx_t_5numpy_float32_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_18 * __pyx_v_values.strides[0]) ))) <= __pyx_v_point) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":698
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:
 */
          goto __pyx_L9_break;

          /* "pandas/_libs/intervaltree.pxi":697
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":699
 *                     if not values[i] <= point:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 */
        __pyx_t_19 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_19 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L9_break:;

      /* "pandas/_libs/intervaltree.pxi":700
 *                         break
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      __pyx_t_14 = ((__pyx_v_point <= __pyx_v_self->left_node->max_right) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":701
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)             # <<<<<<<<<<<<<<
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode *)__pyx_v_self->left_node->__pyx_vtab)->__pyx_fuse_4query(__pyx_v_self->left_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 701, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":700
 *                         break
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":693
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":702
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
    __pyx_t_14 = ((__pyx_v_point > __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":703
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 *                 values = self.center_right_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 */
      __pyx_t_16 = __pyx_v_self->center_right_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":704
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:
 */
      __pyx_t_17 = __pyx_v_self->center_right_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":705
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):             # <<<<<<<<<<<<<<
 *                     if not point <= values[i]:
 *                         break
 */
      for (__pyx_t_11 = (__pyx_v_self->n_center - 1); __pyx_t_11 > -1L; __pyx_t_11-=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":706
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_20 = __pyx_v_i;
        __pyx_t_14 = ((!((__pyx_v_point <= (*((__pyx_t_5numpy_float32_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_20 * __pyx_v_values.strides[0]) )))) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":707
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:
 */
          goto __pyx_L13_break;

          /* "pandas/_libs/intervaltree.pxi":706
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":708
 *                     if not point <= values[i]:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if self.right_node.min_left <= point:
 *                     self.right_node.query(result, point)
 */
        __pyx_t_21 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_21 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L13_break:;

      /* "pandas/_libs/intervaltree.pxi":709
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      __pyx_t_14 = ((__pyx_v_self->right_node->min_left <= __pyx_v_point) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":710
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:
 *                     self.right_node.query(result, point)             # <<<<<<<<<<<<<<
 *             else:
 *                 result.extend(self.center_left_indices)
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode *)__pyx_v_self->right_node->__pyx_vtab)->__pyx_fuse_4query(__pyx_v_self->right_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 710, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":709
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":702
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":712
 *                     self.right_node.query(result, point)
 *             else:
 *                 result.extend(self.center_left_indices)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
    /*else*/ {
      __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->extend(__pyx_v_result, __pyx_v_self->center_left_indices); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 712, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_L7:;
  }
  __pyx_L3:;

  /* "pandas/_libs/intervaltree.pxi":674
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __PYX_XDEC_MEMVIEW(&__pyx_t_16, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_17, 1);
  __Pyx_AddTraceback("pandas._libs.interval.Float32ClosedBothIntervalNode.query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_indices, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_values, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_17__pyx_fuse_4query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_4__pyx_mdef_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_17__pyx_fuse_4query = {"__pyx_fuse_4query", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_17__pyx_fuse_4query, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_2query};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_17__pyx_fuse_4query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result = 0;
  __pyx_t_5numpy_uint64_t __pyx_v_point;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fuse_4query (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_result,&__pyx_n_s_point,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_result)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_point)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_4query", 1, 2, 2, 1); __PYX_ERR(1, 674, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_4query") < 0)) __PYX_ERR(1, 674, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_result = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)values[0]);
    __pyx_v_point = __Pyx_PyInt_As_npy_uint64(values[1]); if (unlikely((__pyx_v_point == ((npy_uint64)-1)) && PyErr_Occurred())) __PYX_ERR(1, 674, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_4query", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 674, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Float32ClosedBothIntervalNode.__pyx_fuse_4query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_result), __pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector, 1, "result", 0))) __PYX_ERR(1, 674, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_16__pyx_fuse_4query(((struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode *)__pyx_v_self), __pyx_v_result, __pyx_v_point);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_16__pyx_fuse_4query(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_uint64_t __pyx_v_point) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__pyx_fuse_4query", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_fuse_4__pyx_f_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_query(__pyx_v_self, __pyx_v_result, __pyx_v_point, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 674, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Float32ClosedBothIntervalNode.__pyx_fuse_4query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":714
 *                 result.extend(self.center_left_indices)
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         if self.is_leaf_node:
 *             return ('<Float32ClosedBothIntervalNode: '
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_5__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_5__repr__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_4__repr__(((struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_4__repr__(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode *__pyx_v_self) {
  PyObject *__pyx_v_n_left = NULL;
  PyObject *__pyx_v_n_right = NULL;
  PyObject *__pyx_v_n_center = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  Py_UCS4 __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "pandas/_libs/intervaltree.pxi":715
 * 
 *     def __repr__(self):
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             return ('<Float32ClosedBothIntervalNode: '
 *                     '%s elements (terminal)>' % self.n_elements)
 */
  __pyx_t_1 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_1) {

    /* "pandas/_libs/intervaltree.pxi":716
 *     def __repr__(self):
 *         if self.is_leaf_node:
 *             return ('<Float32ClosedBothIntervalNode: '             # <<<<<<<<<<<<<<
 *                     '%s elements (terminal)>' % self.n_elements)
 *         else:
 */
    __Pyx_XDECREF(__pyx_r);

    /* "pandas/_libs/intervaltree.pxi":717
 *         if self.is_leaf_node:
 *             return ('<Float32ClosedBothIntervalNode: '
 *                     '%s elements (terminal)>' % self.n_elements)             # <<<<<<<<<<<<<<
 *         else:
 *             n_left = self.left_node.n_elements
 */
    __pyx_t_2 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_elements); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 717, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PyUnicode_Format(__pyx_kp_u_Float32ClosedBothIntervalNode_s, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 717, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pandas/_libs/intervaltree.pxi":715
 * 
 *     def __repr__(self):
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             return ('<Float32ClosedBothIntervalNode: '
 *                     '%s elements (terminal)>' % self.n_elements)
 */
  }

  /* "pandas/_libs/intervaltree.pxi":719
 *                     '%s elements (terminal)>' % self.n_elements)
 *         else:
 *             n_left = self.left_node.n_elements             # <<<<<<<<<<<<<<
 *             n_right = self.right_node.n_elements
 *             n_center = self.n_elements - n_left - n_right
 */
  /*else*/ {
    __pyx_t_3 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->left_node->n_elements); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 719, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_v_n_left = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "pandas/_libs/intervaltree.pxi":720
 *         else:
 *             n_left = self.left_node.n_elements
 *             n_right = self.right_node.n_elements             # <<<<<<<<<<<<<<
 *             n_center = self.n_elements - n_left - n_right
 *             return ('<Float32ClosedBothIntervalNode: '
 */
    __pyx_t_3 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->right_node->n_elements); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 720, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_v_n_right = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "pandas/_libs/intervaltree.pxi":721
 *             n_left = self.left_node.n_elements
 *             n_right = self.right_node.n_elements
 *             n_center = self.n_elements - n_left - n_right             # <<<<<<<<<<<<<<
 *             return ('<Float32ClosedBothIntervalNode: '
 *                     'pivot %s, %s elements (%s left, %s right, %s '
 */
    __pyx_t_3 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_elements); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 721, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = PyNumber_Subtract(__pyx_t_3, __pyx_v_n_left); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 721, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyNumber_Subtract(__pyx_t_2, __pyx_v_n_right); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 721, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_n_center = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "pandas/_libs/intervaltree.pxi":722
 *             n_right = self.right_node.n_elements
 *             n_center = self.n_elements - n_left - n_right
 *             return ('<Float32ClosedBothIntervalNode: '             # <<<<<<<<<<<<<<
 *                     'pivot %s, %s elements (%s left, %s right, %s '
 *                     'overlapping)>' % (self.pivot, self.n_elements,
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = PyTuple_New(11); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 722, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = 0;
    __pyx_t_5 = 127;
    __Pyx_INCREF(__pyx_kp_u_Float32ClosedBothIntervalNode_p);
    __pyx_t_4 += 38;
    __Pyx_GIVEREF(__pyx_kp_u_Float32ClosedBothIntervalNode_p);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_kp_u_Float32ClosedBothIntervalNode_p);

    /* "pandas/_libs/intervaltree.pxi":724
 *             return ('<Float32ClosedBothIntervalNode: '
 *                     'pivot %s, %s elements (%s left, %s right, %s '
 *                     'overlapping)>' % (self.pivot, self.n_elements,             # <<<<<<<<<<<<<<
 *                                        n_left, n_right, n_center))
 * 
 */
    __pyx_t_2 = PyFloat_FromDouble(__pyx_v_self->pivot); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 724, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_t_2), __pyx_empty_unicode); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 724, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_5 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_6) > __pyx_t_5) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_6) : __pyx_t_5;
    __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_6);
    __pyx_t_6 = 0;
    __Pyx_INCREF(__pyx_kp_u__18);
    __pyx_t_4 += 2;
    __Pyx_GIVEREF(__pyx_kp_u__18);
    PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_kp_u__18);
    __pyx_t_6 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_elements); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 724, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_2 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_t_6), __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 724, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_5 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) > __pyx_t_5) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) : __pyx_t_5;
    __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_3, 3, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_elements);
    __pyx_t_4 += 11;
    __Pyx_GIVEREF(__pyx_kp_u_elements);
    PyTuple_SET_ITEM(__pyx_t_3, 4, __pyx_kp_u_elements);

    /* "pandas/_libs/intervaltree.pxi":725
 *                     'pivot %s, %s elements (%s left, %s right, %s '
 *                     'overlapping)>' % (self.pivot, self.n_elements,
 *                                        n_left, n_right, n_center))             # <<<<<<<<<<<<<<
 * 
 *     def counts(self):
 */
    __pyx_t_2 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_v_n_left), __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 725, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) > __pyx_t_5) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) : __pyx_t_5;
    __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_3, 5, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_left_2);
    __pyx_t_4 += 7;
    __Pyx_GIVEREF(__pyx_kp_u_left_2);
    PyTuple_SET_ITEM(__pyx_t_3, 6, __pyx_kp_u_left_2);
    __pyx_t_2 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_v_n_right), __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 725, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) > __pyx_t_5) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) : __pyx_t_5;
    __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_3, 7, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_right_2);
    __pyx_t_4 += 8;
    __Pyx_GIVEREF(__pyx_kp_u_right_2);
    PyTuple_SET_ITEM(__pyx_t_3, 8, __pyx_kp_u_right_2);
    __pyx_t_2 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_v_n_center), __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 725, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) > __pyx_t_5) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) : __pyx_t_5;
    __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_3, 9, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_overlapping);
    __pyx_t_4 += 14;
    __Pyx_GIVEREF(__pyx_kp_u_overlapping);
    PyTuple_SET_ITEM(__pyx_t_3, 10, __pyx_kp_u_overlapping);

    /* "pandas/_libs/intervaltree.pxi":722
 *             n_right = self.right_node.n_elements
 *             n_center = self.n_elements - n_left - n_right
 *             return ('<Float32ClosedBothIntervalNode: '             # <<<<<<<<<<<<<<
 *                     'pivot %s, %s elements (%s left, %s right, %s '
 *                     'overlapping)>' % (self.pivot, self.n_elements,
 */
    __pyx_t_2 = __Pyx_PyUnicode_Join(__pyx_t_3, 11, __pyx_t_4, __pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 722, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;
  }

  /* "pandas/_libs/intervaltree.pxi":714
 *                 result.extend(self.center_left_indices)
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         if self.is_leaf_node:
 *             return ('<Float32ClosedBothIntervalNode: '
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pandas._libs.interval.Float32ClosedBothIntervalNode.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_n_left);
  __Pyx_XDECREF(__pyx_v_n_right);
  __Pyx_XDECREF(__pyx_v_n_center);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":727
 *                                        n_left, n_right, n_center))
 * 
 *     def counts(self):             # <<<<<<<<<<<<<<
 *         """
 *         Inspect counts on this node
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_7counts(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_6counts[] = "\n        Inspect counts on this node\n        useful for debugging purposes\n        ";
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_7counts(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("counts (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_6counts(((struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_6counts(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode *__pyx_v_self) {
  Py_ssize_t __pyx_v_m;
  PyObject *__pyx_v_l = NULL;
  PyObject *__pyx_v_r = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  size_t __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("counts", 0);

  /* "pandas/_libs/intervaltree.pxi":732
 *         useful for debugging purposes
 *         """
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             return self.n_elements
 *         else:
 */
  __pyx_t_1 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_1) {

    /* "pandas/_libs/intervaltree.pxi":733
 *         """
 *         if self.is_leaf_node:
 *             return self.n_elements             # <<<<<<<<<<<<<<
 *         else:
 *             m = len(self.center_left_values)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_elements); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 733, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "pandas/_libs/intervaltree.pxi":732
 *         useful for debugging purposes
 *         """
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             return self.n_elements
 *         else:
 */
  }

  /* "pandas/_libs/intervaltree.pxi":735
 *             return self.n_elements
 *         else:
 *             m = len(self.center_left_values)             # <<<<<<<<<<<<<<
 *             l = self.left_node.counts()
 *             r = self.right_node.counts()
 */
  /*else*/ {
    if (unlikely(!__pyx_v_self->center_left_values.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(1, 735, __pyx_L1_error)}
    __pyx_t_3 = __Pyx_MemoryView_Len(__pyx_v_self->center_left_values); 
    __pyx_v_m = __pyx_t_3;

    /* "pandas/_libs/intervaltree.pxi":736
 *         else:
 *             m = len(self.center_left_values)
 *             l = self.left_node.counts()             # <<<<<<<<<<<<<<
 *             r = self.right_node.counts()
 *             return (m, (l, r))
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->left_node), __pyx_n_s_counts); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 736, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 736, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_l = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "pandas/_libs/intervaltree.pxi":737
 *             m = len(self.center_left_values)
 *             l = self.left_node.counts()
 *             r = self.right_node.counts()             # <<<<<<<<<<<<<<
 *             return (m, (l, r))
 * 
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->right_node), __pyx_n_s_counts); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 737, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 737, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_r = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "pandas/_libs/intervaltree.pxi":738
 *             l = self.left_node.counts()
 *             r = self.right_node.counts()
 *             return (m, (l, r))             # <<<<<<<<<<<<<<
 * 
 * NODE_CLASSES['float32',
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = PyInt_FromSsize_t(__pyx_v_m); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 738, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 738, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_v_l);
    __Pyx_GIVEREF(__pyx_v_l);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_l);
    __Pyx_INCREF(__pyx_v_r);
    __Pyx_GIVEREF(__pyx_v_r);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_v_r);
    __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 738, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_4);
    __pyx_t_2 = 0;
    __pyx_t_4 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;
  }

  /* "pandas/_libs/intervaltree.pxi":727
 *                                        n_left, n_right, n_center))
 * 
 *     def counts(self):             # <<<<<<<<<<<<<<
 *         """
 *         Inspect counts on this node
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pandas._libs.interval.Float32ClosedBothIntervalNode.counts", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_l);
  __Pyx_XDECREF(__pyx_v_r);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":584
 *         int64_t[:] center_left_indices, center_right_indices, indices
 *         float32_t min_left, max_right
 *         readonly float32_t pivot             # <<<<<<<<<<<<<<
 *         readonly int64_t n_elements, n_center, leaf_size
 *         readonly bint is_leaf_node
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_5pivot_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_5pivot_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_5pivot___get__(((struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_5pivot___get__(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->pivot); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 584, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Float32ClosedBothIntervalNode.pivot.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":585
 *         float32_t min_left, max_right
 *         readonly float32_t pivot
 *         readonly int64_t n_elements, n_center, leaf_size             # <<<<<<<<<<<<<<
 *         readonly bint is_leaf_node
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_10n_elements_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_10n_elements_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_10n_elements___get__(((struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_10n_elements___get__(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_elements); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 585, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Float32ClosedBothIntervalNode.n_elements.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_8n_center_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_8n_center_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_8n_center___get__(((struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_8n_center___get__(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_center); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 585, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Float32ClosedBothIntervalNode.n_center.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_9leaf_size_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_9leaf_size_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_9leaf_size___get__(((struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_9leaf_size___get__(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->leaf_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 585, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Float32ClosedBothIntervalNode.leaf_size.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":586
 *         readonly float32_t pivot
 *         readonly int64_t n_elements, n_center, leaf_size
 *         readonly bint is_leaf_node             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self,
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_12is_leaf_node_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_12is_leaf_node_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_12is_leaf_node___get__(((struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_12is_leaf_node___get__(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->is_leaf_node); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 586, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Float32ClosedBothIntervalNode.is_leaf_node.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_21__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_21__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_20__reduce_cython__(((struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_20__reduce_cython__(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  int __pyx_t_16;
  int __pyx_t_17;
  int __pyx_t_18;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.center_left_indices, self.center_left_values, self.center_right_indices, self.center_right_values, self.indices, self.is_leaf_node, self.leaf_size, self.left, self.left_node, self.max_right, self.min_left, self.n_center, self.n_elements, self.pivot, self.right, self.right_node)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  if (unlikely(!__pyx_v_self->center_left_indices.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_1 = __pyx_memoryview_fromslice(__pyx_v_self->center_left_indices, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_int64_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_int64_t, 0);; if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (unlikely(!__pyx_v_self->center_left_values.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_2 = __pyx_memoryview_fromslice(__pyx_v_self->center_left_values, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_float32_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_float32_t, 0);; if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (unlikely(!__pyx_v_self->center_right_indices.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_3 = __pyx_memoryview_fromslice(__pyx_v_self->center_right_indices, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_int64_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_int64_t, 0);; if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely(!__pyx_v_self->center_right_values.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_4 = __pyx_memoryview_fromslice(__pyx_v_self->center_right_values, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_float32_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_float32_t, 0);; if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (unlikely(!__pyx_v_self->indices.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_5 = __pyx_memoryview_fromslice(__pyx_v_self->indices, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_int64_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_int64_t, 0);; if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyBool_FromLong(__pyx_v_self->is_leaf_node); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->leaf_size); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (unlikely(!__pyx_v_self->left.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_8 = __pyx_memoryview_fromslice(__pyx_v_self->left, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_float32_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_float32_t, 0);; if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = PyFloat_FromDouble(__pyx_v_self->max_right); if (unlikely(!__pyx_t_9)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_10 = PyFloat_FromDouble(__pyx_v_self->min_left); if (unlikely(!__pyx_t_10)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_11 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_center); if (unlikely(!__pyx_t_11)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_12 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_elements); if (unlikely(!__pyx_t_12)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __pyx_t_13 = PyFloat_FromDouble(__pyx_v_self->pivot); if (unlikely(!__pyx_t_13)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  if (unlikely(!__pyx_v_self->right.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_14 = __pyx_memoryview_fromslice(__pyx_v_self->right, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_float32_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_float32_t, 0);; if (unlikely(!__pyx_t_14)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __pyx_t_15 = PyTuple_New(16); if (unlikely(!__pyx_t_15)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_15, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_15, 1, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_15, 2, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_15, 3, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_15, 4, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_15, 5, __pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_15, 6, __pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_15, 7, __pyx_t_8);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->left_node));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self->left_node));
  PyTuple_SET_ITEM(__pyx_t_15, 8, ((PyObject *)__pyx_v_self->left_node));
  __Pyx_GIVEREF(__pyx_t_9);
  PyTuple_SET_ITEM(__pyx_t_15, 9, __pyx_t_9);
  __Pyx_GIVEREF(__pyx_t_10);
  PyTuple_SET_ITEM(__pyx_t_15, 10, __pyx_t_10);
  __Pyx_GIVEREF(__pyx_t_11);
  PyTuple_SET_ITEM(__pyx_t_15, 11, __pyx_t_11);
  __Pyx_GIVEREF(__pyx_t_12);
  PyTuple_SET_ITEM(__pyx_t_15, 12, __pyx_t_12);
  __Pyx_GIVEREF(__pyx_t_13);
  PyTuple_SET_ITEM(__pyx_t_15, 13, __pyx_t_13);
  __Pyx_GIVEREF(__pyx_t_14);
  PyTuple_SET_ITEM(__pyx_t_15, 14, __pyx_t_14);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->right_node));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self->right_node));
  PyTuple_SET_ITEM(__pyx_t_15, 15, ((PyObject *)__pyx_v_self->right_node));
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  __pyx_t_6 = 0;
  __pyx_t_7 = 0;
  __pyx_t_8 = 0;
  __pyx_t_9 = 0;
  __pyx_t_10 = 0;
  __pyx_t_11 = 0;
  __pyx_t_12 = 0;
  __pyx_t_13 = 0;
  __pyx_t_14 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_15);
  __pyx_t_15 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.center_left_indices, self.center_left_values, self.center_right_indices, self.center_right_values, self.indices, self.is_leaf_node, self.leaf_size, self.left, self.left_node, self.max_right, self.min_left, self.n_center, self.n_elements, self.pivot, self.right, self.right_node)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_15 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_15)) __PYX_ERR(2, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __pyx_v__dict = __pyx_t_15;
  __pyx_t_15 = 0;

  /* "(tree fragment)":7
 *     state = (self.center_left_indices, self.center_left_values, self.center_right_indices, self.center_right_values, self.indices, self.is_leaf_node, self.leaf_size, self.left, self.left_node, self.max_right, self.min_left, self.n_center, self.n_elements, self.pivot, self.right, self.right_node)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_16 = (__pyx_v__dict != Py_None);
  __pyx_t_17 = (__pyx_t_16 != 0);
  if (__pyx_t_17) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_15 = PyTuple_New(1); if (unlikely(!__pyx_t_15)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    PyTuple_SET_ITEM(__pyx_t_15, 0, __pyx_v__dict);
    __pyx_t_14 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_15); if (unlikely(!__pyx_t_14)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_14));
    __pyx_t_14 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.left_node is not None or self.right_node is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.center_left_indices, self.center_left_values, self.center_right_indices, self.center_right_values, self.indices, self.is_leaf_node, self.leaf_size, self.left, self.left_node, self.max_right, self.min_left, self.n_center, self.n_elements, self.pivot, self.right, self.right_node)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.left_node is not None or self.right_node is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_Float32ClosedBothIntervalNode, (type(self), 0x7a104e5, None), state
 */
  /*else*/ {
    __pyx_t_16 = (((PyObject *)__pyx_v_self->left_node) != Py_None);
    __pyx_t_18 = (__pyx_t_16 != 0);
    if (!__pyx_t_18) {
    } else {
      __pyx_t_17 = __pyx_t_18;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_18 = (((PyObject *)__pyx_v_self->right_node) != Py_None);
    __pyx_t_16 = (__pyx_t_18 != 0);
    __pyx_t_17 = __pyx_t_16;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_17;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.left_node is not None or self.right_node is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Float32ClosedBothIntervalNode, (type(self), 0x7a104e5, None), state
 *     else:
 */
  __pyx_t_17 = (__pyx_v_use_setstate != 0);
  if (__pyx_t_17) {

    /* "(tree fragment)":13
 *         use_setstate = self.left_node is not None or self.right_node is not None
 *     if use_setstate:
 *         return __pyx_unpickle_Float32ClosedBothIntervalNode, (type(self), 0x7a104e5, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_Float32ClosedBothIntervalNode, (type(self), 0x7a104e5, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_14, __pyx_n_s_pyx_unpickle_Float32ClosedBoth); if (unlikely(!__pyx_t_14)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __pyx_t_15 = PyTuple_New(3); if (unlikely(!__pyx_t_15)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_15, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_127993061);
    __Pyx_GIVEREF(__pyx_int_127993061);
    PyTuple_SET_ITEM(__pyx_t_15, 1, __pyx_int_127993061);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_15, 2, Py_None);
    __pyx_t_13 = PyTuple_New(3); if (unlikely(!__pyx_t_13)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __Pyx_GIVEREF(__pyx_t_14);
    PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_t_14);
    __Pyx_GIVEREF(__pyx_t_15);
    PyTuple_SET_ITEM(__pyx_t_13, 1, __pyx_t_15);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_13, 2, __pyx_v_state);
    __pyx_t_14 = 0;
    __pyx_t_15 = 0;
    __pyx_r = __pyx_t_13;
    __pyx_t_13 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.left_node is not None or self.right_node is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Float32ClosedBothIntervalNode, (type(self), 0x7a104e5, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_Float32ClosedBothIntervalNode, (type(self), 0x7a104e5, None), state
 *     else:
 *         return __pyx_unpickle_Float32ClosedBothIntervalNode, (type(self), 0x7a104e5, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Float32ClosedBothIntervalNode__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_13, __pyx_n_s_pyx_unpickle_Float32ClosedBoth); if (unlikely(!__pyx_t_13)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __pyx_t_15 = PyTuple_New(3); if (unlikely(!__pyx_t_15)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_15, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_127993061);
    __Pyx_GIVEREF(__pyx_int_127993061);
    PyTuple_SET_ITEM(__pyx_t_15, 1, __pyx_int_127993061);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_15, 2, __pyx_v_state);
    __pyx_t_14 = PyTuple_New(2); if (unlikely(!__pyx_t_14)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_GIVEREF(__pyx_t_13);
    PyTuple_SET_ITEM(__pyx_t_14, 0, __pyx_t_13);
    __Pyx_GIVEREF(__pyx_t_15);
    PyTuple_SET_ITEM(__pyx_t_14, 1, __pyx_t_15);
    __pyx_t_13 = 0;
    __pyx_t_15 = 0;
    __pyx_r = __pyx_t_14;
    __pyx_t_14 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_AddTraceback("pandas._libs.interval.Float32ClosedBothIntervalNode.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Float32ClosedBothIntervalNode, (type(self), 0x7a104e5, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Float32ClosedBothIntervalNode__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_23__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_23__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_22__setstate_cython__(((struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_22__setstate_cython__(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_Float32ClosedBothIntervalNode, (type(self), 0x7a104e5, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Float32ClosedBothIntervalNode__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(2, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_6pandas_5_libs_8interval___pyx_unpickle_Float32ClosedBothIntervalNode__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Float32ClosedBothIntervalNode, (type(self), 0x7a104e5, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Float32ClosedBothIntervalNode__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Float32ClosedBothIntervalNode.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":758
 *         readonly bint is_leaf_node
 * 
 *     def __init__(self,             # <<<<<<<<<<<<<<
 *                  ndarray[float32_t, ndim=1] left,
 *                  ndarray[float32_t, ndim=1] right,
 */

/* Python wrapper */
static int __pyx_pw_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyArrayObject *__pyx_v_left = 0;
  PyArrayObject *__pyx_v_right = 0;
  PyArrayObject *__pyx_v_indices = 0;
  __pyx_t_5numpy_int64_t __pyx_v_leaf_size;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_left,&__pyx_n_s_right,&__pyx_n_s_indices,&__pyx_n_s_leaf_size,0};
    PyObject* values[4] = {0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_left)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_right)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 4, 4, 1); __PYX_ERR(1, 758, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_indices)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 4, 4, 2); __PYX_ERR(1, 758, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_leaf_size)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 4, 4, 3); __PYX_ERR(1, 758, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(1, 758, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 4) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
    }
    __pyx_v_left = ((PyArrayObject *)values[0]);
    __pyx_v_right = ((PyArrayObject *)values[1]);
    __pyx_v_indices = ((PyArrayObject *)values[2]);
    __pyx_v_leaf_size = __Pyx_PyInt_As_npy_int64(values[3]); if (unlikely((__pyx_v_leaf_size == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(1, 762, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 4, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 758, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Float32ClosedNeitherIntervalNode.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_left), __pyx_ptype_5numpy_ndarray, 1, "left", 0))) __PYX_ERR(1, 759, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_right), __pyx_ptype_5numpy_ndarray, 1, "right", 0))) __PYX_ERR(1, 760, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_indices), __pyx_ptype_5numpy_ndarray, 1, "indices", 0))) __PYX_ERR(1, 761, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode___init__(((struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode *)__pyx_v_self), __pyx_v_left, __pyx_v_right, __pyx_v_indices, __pyx_v_leaf_size);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode___init__(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode *__pyx_v_self, PyArrayObject *__pyx_v_left, PyArrayObject *__pyx_v_right, PyArrayObject *__pyx_v_indices, __pyx_t_5numpy_int64_t __pyx_v_leaf_size) {
  PyObject *__pyx_v_left_set = NULL;
  PyObject *__pyx_v_right_set = NULL;
  PyObject *__pyx_v_center_set = NULL;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_indices;
  __Pyx_Buffer __pyx_pybuffer_indices;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_left;
  __Pyx_Buffer __pyx_pybuffer_left;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_right;
  __Pyx_Buffer __pyx_pybuffer_right;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __pyx_t_5numpy_float32_t __pyx_t_6;
  __Pyx_memviewslice __pyx_t_7 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_8 = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  __Pyx_memviewslice __pyx_t_11 = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *(*__pyx_t_12)(PyObject *);
  size_t __pyx_t_13;
  __Pyx_RefNannySetupContext("__init__", 0);
  __pyx_pybuffer_left.pybuffer.buf = NULL;
  __pyx_pybuffer_left.refcount = 0;
  __pyx_pybuffernd_left.data = NULL;
  __pyx_pybuffernd_left.rcbuffer = &__pyx_pybuffer_left;
  __pyx_pybuffer_right.pybuffer.buf = NULL;
  __pyx_pybuffer_right.refcount = 0;
  __pyx_pybuffernd_right.data = NULL;
  __pyx_pybuffernd_right.rcbuffer = &__pyx_pybuffer_right;
  __pyx_pybuffer_indices.pybuffer.buf = NULL;
  __pyx_pybuffer_indices.refcount = 0;
  __pyx_pybuffernd_indices.data = NULL;
  __pyx_pybuffernd_indices.rcbuffer = &__pyx_pybuffer_indices;
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_left.rcbuffer->pybuffer, (PyObject*)__pyx_v_left, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float32_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 758, __pyx_L1_error)
  }
  __pyx_pybuffernd_left.diminfo[0].strides = __pyx_pybuffernd_left.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_left.diminfo[0].shape = __pyx_pybuffernd_left.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_right.rcbuffer->pybuffer, (PyObject*)__pyx_v_right, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float32_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 758, __pyx_L1_error)
  }
  __pyx_pybuffernd_right.diminfo[0].strides = __pyx_pybuffernd_right.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_right.diminfo[0].shape = __pyx_pybuffernd_right.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_indices.rcbuffer->pybuffer, (PyObject*)__pyx_v_indices, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 758, __pyx_L1_error)
  }
  __pyx_pybuffernd_indices.diminfo[0].strides = __pyx_pybuffernd_indices.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_indices.diminfo[0].shape = __pyx_pybuffernd_indices.rcbuffer->pybuffer.shape[0];

  /* "pandas/_libs/intervaltree.pxi":764
 *                  int64_t leaf_size):
 * 
 *         self.n_elements = len(left)             # <<<<<<<<<<<<<<
 *         self.leaf_size = leaf_size
 * 
 */
  __pyx_t_1 = PyObject_Length(((PyObject *)__pyx_v_left)); if (unlikely(__pyx_t_1 == ((Py_ssize_t)-1))) __PYX_ERR(1, 764, __pyx_L1_error)
  __pyx_v_self->n_elements = __pyx_t_1;

  /* "pandas/_libs/intervaltree.pxi":765
 * 
 *         self.n_elements = len(left)
 *         self.leaf_size = leaf_size             # <<<<<<<<<<<<<<
 * 
 *         # min_left and min_right are used to speed-up query by skipping
 */
  __pyx_v_self->leaf_size = __pyx_v_leaf_size;

  /* "pandas/_libs/intervaltree.pxi":770
 *         # query on sub-nodes. If this node has size 0, query is cheap,
 *         # so these values don't matter.
 *         if left.size > 0:             # <<<<<<<<<<<<<<
 *             self.min_left = left.min()
 *             self.max_right = right.max()
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_left), __pyx_n_s_size); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 770, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyObject_RichCompare(__pyx_t_2, __pyx_int_0, Py_GT); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 770, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(1, 770, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_4) {

    /* "pandas/_libs/intervaltree.pxi":771
 *         # so these values don't matter.
 *         if left.size > 0:
 *             self.min_left = left.min()             # <<<<<<<<<<<<<<
 *             self.max_right = right.max()
 *         else:
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_left), __pyx_n_s_min); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 771, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 771, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_6 = __pyx_PyFloat_AsFloat(__pyx_t_3); if (unlikely((__pyx_t_6 == ((npy_float32)-1)) && PyErr_Occurred())) __PYX_ERR(1, 771, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_self->min_left = __pyx_t_6;

    /* "pandas/_libs/intervaltree.pxi":772
 *         if left.size > 0:
 *             self.min_left = left.min()
 *             self.max_right = right.max()             # <<<<<<<<<<<<<<
 *         else:
 *             self.min_left = 0
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_right), __pyx_n_s_max); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 772, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 772, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_6 = __pyx_PyFloat_AsFloat(__pyx_t_3); if (unlikely((__pyx_t_6 == ((npy_float32)-1)) && PyErr_Occurred())) __PYX_ERR(1, 772, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_self->max_right = __pyx_t_6;

    /* "pandas/_libs/intervaltree.pxi":770
 *         # query on sub-nodes. If this node has size 0, query is cheap,
 *         # so these values don't matter.
 *         if left.size > 0:             # <<<<<<<<<<<<<<
 *             self.min_left = left.min()
 *             self.max_right = right.max()
 */
    goto __pyx_L3;
  }

  /* "pandas/_libs/intervaltree.pxi":774
 *             self.max_right = right.max()
 *         else:
 *             self.min_left = 0             # <<<<<<<<<<<<<<
 *             self.max_right = 0
 * 
 */
  /*else*/ {
    __pyx_v_self->min_left = 0.0;

    /* "pandas/_libs/intervaltree.pxi":775
 *         else:
 *             self.min_left = 0
 *             self.max_right = 0             # <<<<<<<<<<<<<<
 * 
 *         if self.n_elements <= leaf_size:
 */
    __pyx_v_self->max_right = 0.0;
  }
  __pyx_L3:;

  /* "pandas/_libs/intervaltree.pxi":777
 *             self.max_right = 0
 * 
 *         if self.n_elements <= leaf_size:             # <<<<<<<<<<<<<<
 *             # make this a terminal (leaf) node
 *             self.is_leaf_node = True
 */
  __pyx_t_4 = ((__pyx_v_self->n_elements <= __pyx_v_leaf_size) != 0);
  if (__pyx_t_4) {

    /* "pandas/_libs/intervaltree.pxi":779
 *         if self.n_elements <= leaf_size:
 *             # make this a terminal (leaf) node
 *             self.is_leaf_node = True             # <<<<<<<<<<<<<<
 *             self.left = left
 *             self.right = right
 */
    __pyx_v_self->is_leaf_node = 1;

    /* "pandas/_libs/intervaltree.pxi":780
 *             # make this a terminal (leaf) node
 *             self.is_leaf_node = True
 *             self.left = left             # <<<<<<<<<<<<<<
 *             self.right = right
 *             self.indices = indices
 */
    __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_float32_t(((PyObject *)__pyx_v_left), PyBUF_WRITABLE); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(1, 780, __pyx_L1_error)
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->left, 0);
    __pyx_v_self->left = __pyx_t_7;
    __pyx_t_7.memview = NULL;
    __pyx_t_7.data = NULL;

    /* "pandas/_libs/intervaltree.pxi":781
 *             self.is_leaf_node = True
 *             self.left = left
 *             self.right = right             # <<<<<<<<<<<<<<
 *             self.indices = indices
 *             self.n_center = 0
 */
    __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_float32_t(((PyObject *)__pyx_v_right), PyBUF_WRITABLE); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(1, 781, __pyx_L1_error)
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->right, 0);
    __pyx_v_self->right = __pyx_t_7;
    __pyx_t_7.memview = NULL;
    __pyx_t_7.data = NULL;

    /* "pandas/_libs/intervaltree.pxi":782
 *             self.left = left
 *             self.right = right
 *             self.indices = indices             # <<<<<<<<<<<<<<
 *             self.n_center = 0
 *         else:
 */
    __pyx_t_8 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(((PyObject *)__pyx_v_indices), PyBUF_WRITABLE); if (unlikely(!__pyx_t_8.memview)) __PYX_ERR(1, 782, __pyx_L1_error)
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->indices, 0);
    __pyx_v_self->indices = __pyx_t_8;
    __pyx_t_8.memview = NULL;
    __pyx_t_8.data = NULL;

    /* "pandas/_libs/intervaltree.pxi":783
 *             self.right = right
 *             self.indices = indices
 *             self.n_center = 0             # <<<<<<<<<<<<<<
 *         else:
 *             # calculate a pivot so we can create child nodes
 */
    __pyx_v_self->n_center = 0;

    /* "pandas/_libs/intervaltree.pxi":777
 *             self.max_right = 0
 * 
 *         if self.n_elements <= leaf_size:             # <<<<<<<<<<<<<<
 *             # make this a terminal (leaf) node
 *             self.is_leaf_node = True
 */
    goto __pyx_L4;
  }

  /* "pandas/_libs/intervaltree.pxi":786
 *         else:
 *             # calculate a pivot so we can create child nodes
 *             self.is_leaf_node = False             # <<<<<<<<<<<<<<
 *             self.pivot = np.median(left / 2 + right / 2)
 *             left_set, right_set, center_set = self.classify_intervals(
 */
  /*else*/ {
    __pyx_v_self->is_leaf_node = 0;

    /* "pandas/_libs/intervaltree.pxi":787
 *             # calculate a pivot so we can create child nodes
 *             self.is_leaf_node = False
 *             self.pivot = np.median(left / 2 + right / 2)             # <<<<<<<<<<<<<<
 *             left_set, right_set, center_set = self.classify_intervals(
 *                 left, right)
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 787, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_median); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 787, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyNumber_Divide(((PyObject *)__pyx_v_left), __pyx_int_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 787, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_9 = __Pyx_PyNumber_Divide(((PyObject *)__pyx_v_right), __pyx_int_2); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 787, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_10 = PyNumber_Add(__pyx_t_2, __pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 787, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_9 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_3 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_9, __pyx_t_10) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_10);
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 787, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_6 = __pyx_PyFloat_AsFloat(__pyx_t_3); if (unlikely((__pyx_t_6 == ((npy_float32)-1)) && PyErr_Occurred())) __PYX_ERR(1, 787, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_self->pivot = __pyx_t_6;

    /* "pandas/_libs/intervaltree.pxi":789
 *             self.pivot = np.median(left / 2 + right / 2)
 *             left_set, right_set, center_set = self.classify_intervals(
 *                 left, right)             # <<<<<<<<<<<<<<
 * 
 *             self.left_node = self.new_child_node(left, right,
 */
    __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_float32_t(((PyObject *)__pyx_v_left), PyBUF_WRITABLE); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(1, 789, __pyx_L1_error)
    __pyx_t_11 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_float32_t(((PyObject *)__pyx_v_right), PyBUF_WRITABLE); if (unlikely(!__pyx_t_11.memview)) __PYX_ERR(1, 789, __pyx_L1_error)

    /* "pandas/_libs/intervaltree.pxi":788
 *             self.is_leaf_node = False
 *             self.pivot = np.median(left / 2 + right / 2)
 *             left_set, right_set, center_set = self.classify_intervals(             # <<<<<<<<<<<<<<
 *                 left, right)
 * 
 */
    __pyx_t_3 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode *)__pyx_v_self->__pyx_vtab)->classify_intervals(__pyx_v_self, __pyx_t_7, __pyx_t_11); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 788, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __PYX_XDEC_MEMVIEW(&__pyx_t_7, 1);
    __pyx_t_7.memview = NULL;
    __pyx_t_7.data = NULL;
    __PYX_XDEC_MEMVIEW(&__pyx_t_11, 1);
    __pyx_t_11.memview = NULL;
    __pyx_t_11.data = NULL;
    if ((likely(PyTuple_CheckExact(__pyx_t_3))) || (PyList_CheckExact(__pyx_t_3))) {
      PyObject* sequence = __pyx_t_3;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 3)) {
        if (size > 3) __Pyx_RaiseTooManyValuesError(3);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(1, 788, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_5 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_10 = PyTuple_GET_ITEM(sequence, 1); 
        __pyx_t_9 = PyTuple_GET_ITEM(sequence, 2); 
      } else {
        __pyx_t_5 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_10 = PyList_GET_ITEM(sequence, 1); 
        __pyx_t_9 = PyList_GET_ITEM(sequence, 2); 
      }
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_10);
      __Pyx_INCREF(__pyx_t_9);
      #else
      __pyx_t_5 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 788, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_10 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 788, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_9 = PySequence_ITEM(sequence, 2); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 788, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      #endif
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_2 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 788, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_12 = Py_TYPE(__pyx_t_2)->tp_iternext;
      index = 0; __pyx_t_5 = __pyx_t_12(__pyx_t_2); if (unlikely(!__pyx_t_5)) goto __pyx_L5_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_5);
      index = 1; __pyx_t_10 = __pyx_t_12(__pyx_t_2); if (unlikely(!__pyx_t_10)) goto __pyx_L5_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_10);
      index = 2; __pyx_t_9 = __pyx_t_12(__pyx_t_2); if (unlikely(!__pyx_t_9)) goto __pyx_L5_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_9);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_12(__pyx_t_2), 3) < 0) __PYX_ERR(1, 788, __pyx_L1_error)
      __pyx_t_12 = NULL;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      goto __pyx_L6_unpacking_done;
      __pyx_L5_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_12 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(1, 788, __pyx_L1_error)
      __pyx_L6_unpacking_done:;
    }
    __pyx_v_left_set = __pyx_t_5;
    __pyx_t_5 = 0;
    __pyx_v_right_set = __pyx_t_10;
    __pyx_t_10 = 0;
    __pyx_v_center_set = __pyx_t_9;
    __pyx_t_9 = 0;

    /* "pandas/_libs/intervaltree.pxi":792
 * 
 *             self.left_node = self.new_child_node(left, right,
 *                                                  indices, left_set)             # <<<<<<<<<<<<<<
 *             self.right_node = self.new_child_node(left, right,
 *                                                   indices, right_set)
 */
    if (!(likely(((__pyx_v_left_set) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_left_set, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(1, 792, __pyx_L1_error)

    /* "pandas/_libs/intervaltree.pxi":791
 *                 left, right)
 * 
 *             self.left_node = self.new_child_node(left, right,             # <<<<<<<<<<<<<<
 *                                                  indices, left_set)
 *             self.right_node = self.new_child_node(left, right,
 */
    __pyx_t_3 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode *)__pyx_v_self->__pyx_vtab)->new_child_node(__pyx_v_self, ((PyArrayObject *)__pyx_v_left), ((PyArrayObject *)__pyx_v_right), ((PyArrayObject *)__pyx_v_indices), ((PyArrayObject *)__pyx_v_left_set)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 791, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode))))) __PYX_ERR(1, 791, __pyx_L1_error)
    __Pyx_GIVEREF(__pyx_t_3);
    __Pyx_GOTREF(__pyx_v_self->left_node);
    __Pyx_DECREF(((PyObject *)__pyx_v_self->left_node));
    __pyx_v_self->left_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode *)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "pandas/_libs/intervaltree.pxi":794
 *                                                  indices, left_set)
 *             self.right_node = self.new_child_node(left, right,
 *                                                   indices, right_set)             # <<<<<<<<<<<<<<
 * 
 *             self.center_left_values, self.center_left_indices = \
 */
    if (!(likely(((__pyx_v_right_set) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_right_set, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(1, 794, __pyx_L1_error)

    /* "pandas/_libs/intervaltree.pxi":793
 *             self.left_node = self.new_child_node(left, right,
 *                                                  indices, left_set)
 *             self.right_node = self.new_child_node(left, right,             # <<<<<<<<<<<<<<
 *                                                   indices, right_set)
 * 
 */
    __pyx_t_3 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode *)__pyx_v_self->__pyx_vtab)->new_child_node(__pyx_v_self, ((PyArrayObject *)__pyx_v_left), ((PyArrayObject *)__pyx_v_right), ((PyArrayObject *)__pyx_v_indices), ((PyArrayObject *)__pyx_v_right_set)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 793, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode))))) __PYX_ERR(1, 793, __pyx_L1_error)
    __Pyx_GIVEREF(__pyx_t_3);
    __Pyx_GOTREF(__pyx_v_self->right_node);
    __Pyx_DECREF(((PyObject *)__pyx_v_self->right_node));
    __pyx_v_self->right_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode *)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "pandas/_libs/intervaltree.pxi":797
 * 
 *             self.center_left_values, self.center_left_indices = \
 *                 sort_values_and_indices(left, indices, center_set)             # <<<<<<<<<<<<<<
 *             self.center_right_values, self.center_right_indices = \
 *                 sort_values_and_indices(right, indices, center_set)
 */
    __pyx_t_3 = __pyx_f_6pandas_5_libs_8interval_sort_values_and_indices(((PyObject *)__pyx_v_left), ((PyObject *)__pyx_v_indices), __pyx_v_center_set); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 797, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if ((likely(PyTuple_CheckExact(__pyx_t_3))) || (PyList_CheckExact(__pyx_t_3))) {
      PyObject* sequence = __pyx_t_3;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(1, 796, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_9 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_10 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_9 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_10 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_9);
      __Pyx_INCREF(__pyx_t_10);
      #else
      __pyx_t_9 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 796, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_10 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 796, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      #endif
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_5 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 796, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_12 = Py_TYPE(__pyx_t_5)->tp_iternext;
      index = 0; __pyx_t_9 = __pyx_t_12(__pyx_t_5); if (unlikely(!__pyx_t_9)) goto __pyx_L7_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_9);
      index = 1; __pyx_t_10 = __pyx_t_12(__pyx_t_5); if (unlikely(!__pyx_t_10)) goto __pyx_L7_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_10);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_12(__pyx_t_5), 2) < 0) __PYX_ERR(1, 796, __pyx_L1_error)
      __pyx_t_12 = NULL;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      goto __pyx_L8_unpacking_done;
      __pyx_L7_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_12 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(1, 796, __pyx_L1_error)
      __pyx_L8_unpacking_done:;
    }

    /* "pandas/_libs/intervaltree.pxi":796
 *                                                   indices, right_set)
 * 
 *             self.center_left_values, self.center_left_indices = \             # <<<<<<<<<<<<<<
 *                 sort_values_and_indices(left, indices, center_set)
 *             self.center_right_values, self.center_right_indices = \
 */
    __pyx_t_11 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_float32_t(__pyx_t_9, PyBUF_WRITABLE); if (unlikely(!__pyx_t_11.memview)) __PYX_ERR(1, 796, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_8 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_t_10, PyBUF_WRITABLE); if (unlikely(!__pyx_t_8.memview)) __PYX_ERR(1, 796, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->center_left_values, 0);
    __pyx_v_self->center_left_values = __pyx_t_11;
    __pyx_t_11.memview = NULL;
    __pyx_t_11.data = NULL;
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->center_left_indices, 0);
    __pyx_v_self->center_left_indices = __pyx_t_8;
    __pyx_t_8.memview = NULL;
    __pyx_t_8.data = NULL;

    /* "pandas/_libs/intervaltree.pxi":799
 *                 sort_values_and_indices(left, indices, center_set)
 *             self.center_right_values, self.center_right_indices = \
 *                 sort_values_and_indices(right, indices, center_set)             # <<<<<<<<<<<<<<
 *             self.n_center = len(self.center_left_indices)
 * 
 */
    __pyx_t_3 = __pyx_f_6pandas_5_libs_8interval_sort_values_and_indices(((PyObject *)__pyx_v_right), ((PyObject *)__pyx_v_indices), __pyx_v_center_set); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 799, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if ((likely(PyTuple_CheckExact(__pyx_t_3))) || (PyList_CheckExact(__pyx_t_3))) {
      PyObject* sequence = __pyx_t_3;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(1, 798, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_10 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_9 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_10 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_9 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_10);
      __Pyx_INCREF(__pyx_t_9);
      #else
      __pyx_t_10 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 798, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_9 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 798, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      #endif
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_5 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 798, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_12 = Py_TYPE(__pyx_t_5)->tp_iternext;
      index = 0; __pyx_t_10 = __pyx_t_12(__pyx_t_5); if (unlikely(!__pyx_t_10)) goto __pyx_L9_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_10);
      index = 1; __pyx_t_9 = __pyx_t_12(__pyx_t_5); if (unlikely(!__pyx_t_9)) goto __pyx_L9_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_9);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_12(__pyx_t_5), 2) < 0) __PYX_ERR(1, 798, __pyx_L1_error)
      __pyx_t_12 = NULL;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      goto __pyx_L10_unpacking_done;
      __pyx_L9_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_12 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(1, 798, __pyx_L1_error)
      __pyx_L10_unpacking_done:;
    }

    /* "pandas/_libs/intervaltree.pxi":798
 *             self.center_left_values, self.center_left_indices = \
 *                 sort_values_and_indices(left, indices, center_set)
 *             self.center_right_values, self.center_right_indices = \             # <<<<<<<<<<<<<<
 *                 sort_values_and_indices(right, indices, center_set)
 *             self.n_center = len(self.center_left_indices)
 */
    __pyx_t_11 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_float32_t(__pyx_t_10, PyBUF_WRITABLE); if (unlikely(!__pyx_t_11.memview)) __PYX_ERR(1, 798, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_t_8 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_t_9, PyBUF_WRITABLE); if (unlikely(!__pyx_t_8.memview)) __PYX_ERR(1, 798, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->center_right_values, 0);
    __pyx_v_self->center_right_values = __pyx_t_11;
    __pyx_t_11.memview = NULL;
    __pyx_t_11.data = NULL;
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->center_right_indices, 0);
    __pyx_v_self->center_right_indices = __pyx_t_8;
    __pyx_t_8.memview = NULL;
    __pyx_t_8.data = NULL;

    /* "pandas/_libs/intervaltree.pxi":800
 *             self.center_right_values, self.center_right_indices = \
 *                 sort_values_and_indices(right, indices, center_set)
 *             self.n_center = len(self.center_left_indices)             # <<<<<<<<<<<<<<
 * 
 *     @cython.wraparound(False)
 */
    if (unlikely(!__pyx_v_self->center_left_indices.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(1, 800, __pyx_L1_error)}
    __pyx_t_13 = __Pyx_MemoryView_Len(__pyx_v_self->center_left_indices); 
    __pyx_v_self->n_center = __pyx_t_13;
  }
  __pyx_L4:;

  /* "pandas/_libs/intervaltree.pxi":758
 *         readonly bint is_leaf_node
 * 
 *     def __init__(self,             # <<<<<<<<<<<<<<
 *                  ndarray[float32_t, ndim=1] left,
 *                  ndarray[float32_t, ndim=1] right,
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __PYX_XDEC_MEMVIEW(&__pyx_t_7, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_8, 1);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __PYX_XDEC_MEMVIEW(&__pyx_t_11, 1);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indices.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_left.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_right.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("pandas._libs.interval.Float32ClosedNeitherIntervalNode.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indices.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_left.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_right.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF(__pyx_v_left_set);
  __Pyx_XDECREF(__pyx_v_right_set);
  __Pyx_XDECREF(__pyx_v_center_set);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":804
 *     @cython.wraparound(False)
 *     @cython.boundscheck(False)
 *     cdef classify_intervals(self, float32_t[:] left, float32_t[:] right):             # <<<<<<<<<<<<<<
 *         """Classify the given intervals based upon whether they fall to the
 *         left, right, or overlap with this node's pivot.
 */

static PyObject *__pyx_f_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_classify_intervals(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode *__pyx_v_self, __Pyx_memviewslice __pyx_v_left, __Pyx_memviewslice __pyx_v_right) {
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_left_ind = 0;
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_right_ind = 0;
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_overlapping_ind = 0;
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __pyx_t_5numpy_int64_t __pyx_t_2;
  __pyx_t_5numpy_int64_t __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  int __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  __Pyx_RefNannySetupContext("classify_intervals", 0);

  /* "pandas/_libs/intervaltree.pxi":812
 *             Py_ssize_t i
 * 
 *         left_ind = Int64Vector()             # <<<<<<<<<<<<<<
 *         right_ind = Int64Vector()
 *         overlapping_ind = Int64Vector()
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 812, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_left_ind = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":813
 * 
 *         left_ind = Int64Vector()
 *         right_ind = Int64Vector()             # <<<<<<<<<<<<<<
 *         overlapping_ind = Int64Vector()
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 813, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_right_ind = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":814
 *         left_ind = Int64Vector()
 *         right_ind = Int64Vector()
 *         overlapping_ind = Int64Vector()             # <<<<<<<<<<<<<<
 * 
 *         for i in range(self.n_elements):
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 814, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_overlapping_ind = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":816
 *         overlapping_ind = Int64Vector()
 * 
 *         for i in range(self.n_elements):             # <<<<<<<<<<<<<<
 *             if right[i] <= self.pivot:
 *                 left_ind.append(i)
 */
  __pyx_t_2 = __pyx_v_self->n_elements;
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "pandas/_libs/intervaltree.pxi":817
 * 
 *         for i in range(self.n_elements):
 *             if right[i] <= self.pivot:             # <<<<<<<<<<<<<<
 *                 left_ind.append(i)
 *             elif self.pivot <= left[i]:
 */
    __pyx_t_5 = __pyx_v_i;
    __pyx_t_6 = (((*((__pyx_t_5numpy_float32_t *) ( /* dim=0 */ (__pyx_v_right.data + __pyx_t_5 * __pyx_v_right.strides[0]) ))) <= __pyx_v_self->pivot) != 0);
    if (__pyx_t_6) {

      /* "pandas/_libs/intervaltree.pxi":818
 *         for i in range(self.n_elements):
 *             if right[i] <= self.pivot:
 *                 left_ind.append(i)             # <<<<<<<<<<<<<<
 *             elif self.pivot <= left[i]:
 *                 right_ind.append(i)
 */
      ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_left_ind->__pyx_vtab)->append(__pyx_v_left_ind, __pyx_v_i);

      /* "pandas/_libs/intervaltree.pxi":817
 * 
 *         for i in range(self.n_elements):
 *             if right[i] <= self.pivot:             # <<<<<<<<<<<<<<
 *                 left_ind.append(i)
 *             elif self.pivot <= left[i]:
 */
      goto __pyx_L5;
    }

    /* "pandas/_libs/intervaltree.pxi":819
 *             if right[i] <= self.pivot:
 *                 left_ind.append(i)
 *             elif self.pivot <= left[i]:             # <<<<<<<<<<<<<<
 *                 right_ind.append(i)
 *             else:
 */
    __pyx_t_7 = __pyx_v_i;
    __pyx_t_6 = ((__pyx_v_self->pivot <= (*((__pyx_t_5numpy_float32_t *) ( /* dim=0 */ (__pyx_v_left.data + __pyx_t_7 * __pyx_v_left.strides[0]) )))) != 0);
    if (__pyx_t_6) {

      /* "pandas/_libs/intervaltree.pxi":820
 *                 left_ind.append(i)
 *             elif self.pivot <= left[i]:
 *                 right_ind.append(i)             # <<<<<<<<<<<<<<
 *             else:
 *                 overlapping_ind.append(i)
 */
      ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_right_ind->__pyx_vtab)->append(__pyx_v_right_ind, __pyx_v_i);

      /* "pandas/_libs/intervaltree.pxi":819
 *             if right[i] <= self.pivot:
 *                 left_ind.append(i)
 *             elif self.pivot <= left[i]:             # <<<<<<<<<<<<<<
 *                 right_ind.append(i)
 *             else:
 */
      goto __pyx_L5;
    }

    /* "pandas/_libs/intervaltree.pxi":822
 *                 right_ind.append(i)
 *             else:
 *                 overlapping_ind.append(i)             # <<<<<<<<<<<<<<
 * 
 *         return (left_ind.to_array(),
 */
    /*else*/ {
      ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_overlapping_ind->__pyx_vtab)->append(__pyx_v_overlapping_ind, __pyx_v_i);
    }
    __pyx_L5:;
  }

  /* "pandas/_libs/intervaltree.pxi":824
 *                 overlapping_ind.append(i)
 * 
 *         return (left_ind.to_array(),             # <<<<<<<<<<<<<<
 *                 right_ind.to_array(),
 *                 overlapping_ind.to_array())
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_left_ind->__pyx_vtab)->to_array(__pyx_v_left_ind, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 824, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "pandas/_libs/intervaltree.pxi":825
 * 
 *         return (left_ind.to_array(),
 *                 right_ind.to_array(),             # <<<<<<<<<<<<<<
 *                 overlapping_ind.to_array())
 * 
 */
  __pyx_t_8 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_right_ind->__pyx_vtab)->to_array(__pyx_v_right_ind, 0); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 825, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);

  /* "pandas/_libs/intervaltree.pxi":826
 *         return (left_ind.to_array(),
 *                 right_ind.to_array(),
 *                 overlapping_ind.to_array())             # <<<<<<<<<<<<<<
 * 
 *     cdef new_child_node(self,
 */
  __pyx_t_9 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_overlapping_ind->__pyx_vtab)->to_array(__pyx_v_overlapping_ind, 0); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 826, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);

  /* "pandas/_libs/intervaltree.pxi":824
 *                 overlapping_ind.append(i)
 * 
 *         return (left_ind.to_array(),             # <<<<<<<<<<<<<<
 *                 right_ind.to_array(),
 *                 overlapping_ind.to_array())
 */
  __pyx_t_10 = PyTuple_New(3); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 824, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_t_8);
  __Pyx_GIVEREF(__pyx_t_9);
  PyTuple_SET_ITEM(__pyx_t_10, 2, __pyx_t_9);
  __pyx_t_1 = 0;
  __pyx_t_8 = 0;
  __pyx_t_9 = 0;
  __pyx_r = __pyx_t_10;
  __pyx_t_10 = 0;
  goto __pyx_L0;

  /* "pandas/_libs/intervaltree.pxi":804
 *     @cython.wraparound(False)
 *     @cython.boundscheck(False)
 *     cdef classify_intervals(self, float32_t[:] left, float32_t[:] right):             # <<<<<<<<<<<<<<
 *         """Classify the given intervals based upon whether they fall to the
 *         left, right, or overlap with this node's pivot.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("pandas._libs.interval.Float32ClosedNeitherIntervalNode.classify_intervals", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_left_ind);
  __Pyx_XDECREF((PyObject *)__pyx_v_right_ind);
  __Pyx_XDECREF((PyObject *)__pyx_v_overlapping_ind);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":828
 *                 overlapping_ind.to_array())
 * 
 *     cdef new_child_node(self,             # <<<<<<<<<<<<<<
 *                         ndarray[float32_t, ndim=1] left,
 *                         ndarray[float32_t, ndim=1] right,
 */

static PyObject *__pyx_f_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_new_child_node(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode *__pyx_v_self, PyArrayObject *__pyx_v_left, PyArrayObject *__pyx_v_right, PyArrayObject *__pyx_v_indices, PyArrayObject *__pyx_v_subset) {
  __Pyx_LocalBuf_ND __pyx_pybuffernd_indices;
  __Pyx_Buffer __pyx_pybuffer_indices;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_left;
  __Pyx_Buffer __pyx_pybuffer_left;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_right;
  __Pyx_Buffer __pyx_pybuffer_right;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_subset;
  __Pyx_Buffer __pyx_pybuffer_subset;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyArrayObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyArrayObject *__pyx_t_7 = NULL;
  PyArrayObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  __Pyx_RefNannySetupContext("new_child_node", 0);
  __Pyx_INCREF((PyObject *)__pyx_v_left);
  __Pyx_INCREF((PyObject *)__pyx_v_right);
  __Pyx_INCREF((PyObject *)__pyx_v_indices);
  __pyx_pybuffer_left.pybuffer.buf = NULL;
  __pyx_pybuffer_left.refcount = 0;
  __pyx_pybuffernd_left.data = NULL;
  __pyx_pybuffernd_left.rcbuffer = &__pyx_pybuffer_left;
  __pyx_pybuffer_right.pybuffer.buf = NULL;
  __pyx_pybuffer_right.refcount = 0;
  __pyx_pybuffernd_right.data = NULL;
  __pyx_pybuffernd_right.rcbuffer = &__pyx_pybuffer_right;
  __pyx_pybuffer_indices.pybuffer.buf = NULL;
  __pyx_pybuffer_indices.refcount = 0;
  __pyx_pybuffernd_indices.data = NULL;
  __pyx_pybuffernd_indices.rcbuffer = &__pyx_pybuffer_indices;
  __pyx_pybuffer_subset.pybuffer.buf = NULL;
  __pyx_pybuffer_subset.refcount = 0;
  __pyx_pybuffernd_subset.data = NULL;
  __pyx_pybuffernd_subset.rcbuffer = &__pyx_pybuffer_subset;
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_left.rcbuffer->pybuffer, (PyObject*)__pyx_v_left, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float32_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 828, __pyx_L1_error)
  }
  __pyx_pybuffernd_left.diminfo[0].strides = __pyx_pybuffernd_left.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_left.diminfo[0].shape = __pyx_pybuffernd_left.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_right.rcbuffer->pybuffer, (PyObject*)__pyx_v_right, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float32_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 828, __pyx_L1_error)
  }
  __pyx_pybuffernd_right.diminfo[0].strides = __pyx_pybuffernd_right.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_right.diminfo[0].shape = __pyx_pybuffernd_right.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_indices.rcbuffer->pybuffer, (PyObject*)__pyx_v_indices, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 828, __pyx_L1_error)
  }
  __pyx_pybuffernd_indices.diminfo[0].strides = __pyx_pybuffernd_indices.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_indices.diminfo[0].shape = __pyx_pybuffernd_indices.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_subset.rcbuffer->pybuffer, (PyObject*)__pyx_v_subset, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 828, __pyx_L1_error)
  }
  __pyx_pybuffernd_subset.diminfo[0].strides = __pyx_pybuffernd_subset.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_subset.diminfo[0].shape = __pyx_pybuffernd_subset.rcbuffer->pybuffer.shape[0];

  /* "pandas/_libs/intervaltree.pxi":835
 *         """Create a new child node.
 *         """
 *         left = take(left, subset)             # <<<<<<<<<<<<<<
 *         right = take(right, subset)
 *         indices = take(indices, subset)
 */
  __pyx_t_1 = __pyx_f_6pandas_5_libs_8interval_take(((PyArrayObject *)__pyx_v_left), ((PyArrayObject *)__pyx_v_subset)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 835, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(1, 835, __pyx_L1_error)
  __pyx_t_2 = ((PyArrayObject *)__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_left.rcbuffer->pybuffer);
    __pyx_t_3 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_left.rcbuffer->pybuffer, (PyObject*)__pyx_t_2, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float32_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack);
    if (unlikely(__pyx_t_3 < 0)) {
      PyErr_Fetch(&__pyx_t_4, &__pyx_t_5, &__pyx_t_6);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_left.rcbuffer->pybuffer, (PyObject*)__pyx_v_left, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float32_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_4); Py_XDECREF(__pyx_t_5); Py_XDECREF(__pyx_t_6);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_4, __pyx_t_5, __pyx_t_6);
      }
      __pyx_t_4 = __pyx_t_5 = __pyx_t_6 = 0;
    }
    __pyx_pybuffernd_left.diminfo[0].strides = __pyx_pybuffernd_left.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_left.diminfo[0].shape = __pyx_pybuffernd_left.rcbuffer->pybuffer.shape[0];
    if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(1, 835, __pyx_L1_error)
  }
  __pyx_t_2 = 0;
  __Pyx_DECREF_SET(__pyx_v_left, ((PyArrayObject *)__pyx_t_1));
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":836
 *         """
 *         left = take(left, subset)
 *         right = take(right, subset)             # <<<<<<<<<<<<<<
 *         indices = take(indices, subset)
 *         return Float32ClosedNeitherIntervalNode(
 */
  __pyx_t_1 = __pyx_f_6pandas_5_libs_8interval_take(((PyArrayObject *)__pyx_v_right), ((PyArrayObject *)__pyx_v_subset)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 836, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(1, 836, __pyx_L1_error)
  __pyx_t_7 = ((PyArrayObject *)__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_right.rcbuffer->pybuffer);
    __pyx_t_3 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_right.rcbuffer->pybuffer, (PyObject*)__pyx_t_7, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float32_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack);
    if (unlikely(__pyx_t_3 < 0)) {
      PyErr_Fetch(&__pyx_t_6, &__pyx_t_5, &__pyx_t_4);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_right.rcbuffer->pybuffer, (PyObject*)__pyx_v_right, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float32_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_6); Py_XDECREF(__pyx_t_5); Py_XDECREF(__pyx_t_4);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_6, __pyx_t_5, __pyx_t_4);
      }
      __pyx_t_6 = __pyx_t_5 = __pyx_t_4 = 0;
    }
    __pyx_pybuffernd_right.diminfo[0].strides = __pyx_pybuffernd_right.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_right.diminfo[0].shape = __pyx_pybuffernd_right.rcbuffer->pybuffer.shape[0];
    if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(1, 836, __pyx_L1_error)
  }
  __pyx_t_7 = 0;
  __Pyx_DECREF_SET(__pyx_v_right, ((PyArrayObject *)__pyx_t_1));
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":837
 *         left = take(left, subset)
 *         right = take(right, subset)
 *         indices = take(indices, subset)             # <<<<<<<<<<<<<<
 *         return Float32ClosedNeitherIntervalNode(
 *             left, right, indices, self.leaf_size)
 */
  __pyx_t_1 = __pyx_f_6pandas_5_libs_8interval_take(((PyArrayObject *)__pyx_v_indices), ((PyArrayObject *)__pyx_v_subset)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 837, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(1, 837, __pyx_L1_error)
  __pyx_t_8 = ((PyArrayObject *)__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indices.rcbuffer->pybuffer);
    __pyx_t_3 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_indices.rcbuffer->pybuffer, (PyObject*)__pyx_t_8, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack);
    if (unlikely(__pyx_t_3 < 0)) {
      PyErr_Fetch(&__pyx_t_4, &__pyx_t_5, &__pyx_t_6);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_indices.rcbuffer->pybuffer, (PyObject*)__pyx_v_indices, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_4); Py_XDECREF(__pyx_t_5); Py_XDECREF(__pyx_t_6);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_4, __pyx_t_5, __pyx_t_6);
      }
      __pyx_t_4 = __pyx_t_5 = __pyx_t_6 = 0;
    }
    __pyx_pybuffernd_indices.diminfo[0].strides = __pyx_pybuffernd_indices.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_indices.diminfo[0].shape = __pyx_pybuffernd_indices.rcbuffer->pybuffer.shape[0];
    if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(1, 837, __pyx_L1_error)
  }
  __pyx_t_8 = 0;
  __Pyx_DECREF_SET(__pyx_v_indices, ((PyArrayObject *)__pyx_t_1));
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":838
 *         right = take(right, subset)
 *         indices = take(indices, subset)
 *         return Float32ClosedNeitherIntervalNode(             # <<<<<<<<<<<<<<
 *             left, right, indices, self.leaf_size)
 * 
 */
  __Pyx_XDECREF(__pyx_r);

  /* "pandas/_libs/intervaltree.pxi":839
 *         indices = take(indices, subset)
 *         return Float32ClosedNeitherIntervalNode(
 *             left, right, indices, self.leaf_size)             # <<<<<<<<<<<<<<
 * 
 *     @cython.wraparound(False)
 */
  __pyx_t_1 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->leaf_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 839, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "pandas/_libs/intervaltree.pxi":838
 *         right = take(right, subset)
 *         indices = take(indices, subset)
 *         return Float32ClosedNeitherIntervalNode(             # <<<<<<<<<<<<<<
 *             left, right, indices, self.leaf_size)
 * 
 */
  __pyx_t_9 = PyTuple_New(4); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 838, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_INCREF(((PyObject *)__pyx_v_left));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_left));
  PyTuple_SET_ITEM(__pyx_t_9, 0, ((PyObject *)__pyx_v_left));
  __Pyx_INCREF(((PyObject *)__pyx_v_right));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_right));
  PyTuple_SET_ITEM(__pyx_t_9, 1, ((PyObject *)__pyx_v_right));
  __Pyx_INCREF(((PyObject *)__pyx_v_indices));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_indices));
  PyTuple_SET_ITEM(__pyx_t_9, 2, ((PyObject *)__pyx_v_indices));
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_9, 3, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode), __pyx_t_9, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 838, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pandas/_libs/intervaltree.pxi":828
 *                 overlapping_ind.to_array())
 * 
 *     cdef new_child_node(self,             # <<<<<<<<<<<<<<
 *                         ndarray[float32_t, ndim=1] left,
 *                         ndarray[float32_t, ndim=1] right,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_9);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indices.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_left.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_right.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_subset.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("pandas._libs.interval.Float32ClosedNeitherIntervalNode.new_child_node", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indices.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_left.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_right.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_subset.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF((PyObject *)__pyx_v_left);
  __Pyx_XDECREF((PyObject *)__pyx_v_right);
  __Pyx_XDECREF((PyObject *)__pyx_v_indices);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":844
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_2query[] = "Recursively query this node and its sub-nodes for intervals that\n        overlap with the query point.\n        ";
static PyMethodDef __pyx_mdef_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_3query = {"query", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_3query, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_2query};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_args = 0;
  PyObject *__pyx_v_kwargs = 0;
  CYTHON_UNUSED PyObject *__pyx_v_defaults = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fused_cpdef (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_args,&__pyx_n_s_kwargs,&__pyx_n_s_defaults,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_args)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_kwargs)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fused_cpdef", 1, 3, 3, 1); __PYX_ERR(1, 844, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_defaults)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fused_cpdef", 1, 3, 3, 2); __PYX_ERR(1, 844, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fused_cpdef") < 0)) __PYX_ERR(1, 844, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_args = values[0];
    __pyx_v_kwargs = values[1];
    __pyx_v_defaults = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fused_cpdef", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 844, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Float32ClosedNeitherIntervalNode.__pyx_fused_cpdef", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_2query(((struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode *)__pyx_v_signatures), __pyx_v_args, __pyx_v_kwargs, __pyx_v_defaults);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_2query(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode *__pyx_v_signatures, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs, CYTHON_UNUSED PyObject *__pyx_v_defaults) {
  PyObject *__pyx_v_dest_sig = NULL;
  Py_ssize_t __pyx_v_i;
  CYTHON_UNUSED PyTypeObject *__pyx_v_ndarray = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_v_itemsize;
  PyObject *__pyx_v_arg = NULL;
  PyObject *__pyx_v_candidates = NULL;
  PyObject *__pyx_v_sig = NULL;
  int __pyx_v_match_found;
  PyObject *__pyx_v_src_sig = NULL;
  PyObject *__pyx_v_dst_type = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  Py_ssize_t __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  Py_ssize_t __pyx_t_13;
  Py_ssize_t __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  int __pyx_t_16;
  __Pyx_RefNannySetupContext("query", 0);
  __Pyx_INCREF(__pyx_v_kwargs);
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 844, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  PyList_SET_ITEM(__pyx_t_1, 0, Py_None);
  __pyx_v_dest_sig = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_3 = (__pyx_v_kwargs != Py_None);
  __pyx_t_4 = (__pyx_t_3 != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_kwargs); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(1, 844, __pyx_L1_error)
  __pyx_t_3 = ((!__pyx_t_4) != 0);
  __pyx_t_2 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {
    __Pyx_INCREF(Py_None);
    __Pyx_DECREF_SET(__pyx_v_kwargs, Py_None);
  }
  __pyx_t_1 = ((PyObject *)__Pyx_ImportNumPyArrayTypeIfAvailable()); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 844, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_ndarray = ((PyTypeObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_v_itemsize = -1L;
  if (unlikely(__pyx_v_args == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 844, __pyx_L1_error)
  }
  __pyx_t_5 = PyTuple_GET_SIZE(((PyObject*)__pyx_v_args)); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(1, 844, __pyx_L1_error)
  __pyx_t_2 = ((2 < __pyx_t_5) != 0);
  if (__pyx_t_2) {
    if (unlikely(__pyx_v_args == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 844, __pyx_L1_error)
    }
    __pyx_t_1 = PyTuple_GET_ITEM(((PyObject*)__pyx_v_args), 2);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v_arg = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L6;
  }
  __pyx_t_3 = (__pyx_v_kwargs != Py_None);
  __pyx_t_4 = (__pyx_t_3 != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L7_bool_binop_done;
  }
  if (unlikely(__pyx_v_kwargs == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(1, 844, __pyx_L1_error)
  }
  __pyx_t_4 = (__Pyx_PyDict_ContainsTF(__pyx_n_s_point, ((PyObject*)__pyx_v_kwargs), Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(1, 844, __pyx_L1_error)
  __pyx_t_3 = (__pyx_t_4 != 0);
  __pyx_t_2 = __pyx_t_3;
  __pyx_L7_bool_binop_done:;
  if (__pyx_t_2) {
    if (unlikely(__pyx_v_kwargs == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 844, __pyx_L1_error)
    }
    __pyx_t_1 = __Pyx_PyDict_GetItem(((PyObject*)__pyx_v_kwargs), __pyx_n_s_point); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 844, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_arg = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L6;
  }
  /*else*/ {
    if (unlikely(__pyx_v_args == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
      __PYX_ERR(1, 844, __pyx_L1_error)
    }
    __pyx_t_5 = PyTuple_GET_SIZE(((PyObject*)__pyx_v_args)); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(1, 844, __pyx_L1_error)
    __pyx_t_1 = PyInt_FromSsize_t(__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 844, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_6 = PyTuple_New(3); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 844, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_INCREF(__pyx_int_3);
    __Pyx_GIVEREF(__pyx_int_3);
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_int_3);
    __Pyx_INCREF(__pyx_n_s_s);
    __Pyx_GIVEREF(__pyx_n_s_s);
    PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_n_s_s);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Expected_at_least_d_argument_s_g, __pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 844, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_1); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 844, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(1, 844, __pyx_L1_error)
  }
  __pyx_L6:;
  while (1) {
    __pyx_t_2 = PyFloat_Check(__pyx_v_arg); 
    __pyx_t_3 = (__pyx_t_2 != 0);
    if (__pyx_t_3) {
      if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, __pyx_n_s_float64_t, long, 1, __Pyx_PyInt_From_long, 1, 0, 0) < 0)) __PYX_ERR(1, 844, __pyx_L1_error)
      goto __pyx_L10_break;
    }
    __pyx_t_2 = PyInt_Check(__pyx_v_arg); 
    __pyx_t_4 = (__pyx_t_2 != 0);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_3 = __pyx_t_4;
      goto __pyx_L13_bool_binop_done;
    }
    __pyx_t_4 = PyLong_Check(__pyx_v_arg); 
    __pyx_t_2 = (__pyx_t_4 != 0);
    __pyx_t_3 = __pyx_t_2;
    __pyx_L13_bool_binop_done:;
    __pyx_t_2 = (__pyx_t_3 != 0);
    if (__pyx_t_2) {
      if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, __pyx_n_s_int64_t, long, 1, __Pyx_PyInt_From_long, 1, 0, 0) < 0)) __PYX_ERR(1, 844, __pyx_L1_error)
      goto __pyx_L10_break;
    }
    if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, Py_None, long, 1, __Pyx_PyInt_From_long, 1, 0, 0) < 0)) __PYX_ERR(1, 844, __pyx_L1_error)
    goto __pyx_L10_break;
  }
  __pyx_L10_break:;
  __pyx_t_6 = PyList_New(0); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 844, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_v_candidates = ((PyObject*)__pyx_t_6);
  __pyx_t_6 = 0;
  __pyx_t_5 = 0;
  if (unlikely(((PyObject *)__pyx_v_signatures) == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(1, 844, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_dict_iterator(((PyObject*)__pyx_v_signatures), 1, ((PyObject *)NULL), (&__pyx_t_7), (&__pyx_t_8)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 844, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __pyx_t_6 = __pyx_t_1;
  __pyx_t_1 = 0;
  while (1) {
    __pyx_t_9 = __Pyx_dict_iter_next(__pyx_t_6, __pyx_t_7, &__pyx_t_5, &__pyx_t_1, NULL, NULL, __pyx_t_8);
    if (unlikely(__pyx_t_9 == 0)) break;
    if (unlikely(__pyx_t_9 == -1)) __PYX_ERR(1, 844, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_sig, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_v_match_found = 0;
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_v_sig, __pyx_n_s_strip); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 844, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __pyx_t_12 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_11))) {
      __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_11);
      if (likely(__pyx_t_12)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
        __Pyx_INCREF(__pyx_t_12);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_11, function);
      }
    }
    __pyx_t_10 = (__pyx_t_12) ? __Pyx_PyObject_Call2Args(__pyx_t_11, __pyx_t_12, __pyx_kp_s__13) : __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_kp_s__13);
    __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
    if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 844, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_split); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 844, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_t_10 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_11))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_11);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_11, function);
      }
    }
    __pyx_t_1 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_11, __pyx_t_10, __pyx_kp_s__14) : __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_kp_s__14);
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 844, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_XDECREF_SET(__pyx_v_src_sig, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_13 = PyList_GET_SIZE(__pyx_v_dest_sig); if (unlikely(__pyx_t_13 == ((Py_ssize_t)-1))) __PYX_ERR(1, 844, __pyx_L1_error)
    __pyx_t_14 = __pyx_t_13;
    for (__pyx_t_15 = 0; __pyx_t_15 < __pyx_t_14; __pyx_t_15+=1) {
      __pyx_v_i = __pyx_t_15;
      __pyx_t_1 = PyList_GET_ITEM(__pyx_v_dest_sig, __pyx_v_i);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_XDECREF_SET(__pyx_v_dst_type, __pyx_t_1);
      __pyx_t_1 = 0;
      __pyx_t_2 = (__pyx_v_dst_type != Py_None);
      __pyx_t_3 = (__pyx_t_2 != 0);
      if (__pyx_t_3) {
        __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_src_sig, __pyx_v_i, Py_ssize_t, 1, PyInt_FromSsize_t, 0, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 844, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_11 = PyObject_RichCompare(__pyx_t_1, __pyx_v_dst_type, Py_EQ); __Pyx_XGOTREF(__pyx_t_11); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 844, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_11); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(1, 844, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        if (__pyx_t_3) {
          __pyx_v_match_found = 1;
          goto __pyx_L20;
        }
        /*else*/ {
          __pyx_v_match_found = 0;
          goto __pyx_L18_break;
        }
        __pyx_L20:;
      }
    }
    __pyx_L18_break:;
    __pyx_t_3 = (__pyx_v_match_found != 0);
    if (__pyx_t_3) {
      __pyx_t_16 = __Pyx_PyList_Append(__pyx_v_candidates, __pyx_v_sig); if (unlikely(__pyx_t_16 == ((int)-1))) __PYX_ERR(1, 844, __pyx_L1_error)
    }
  }
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_3 = (PyList_GET_SIZE(__pyx_v_candidates) != 0);
  __pyx_t_2 = ((!__pyx_t_3) != 0);
  if (__pyx_t_2) {
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__15, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 844, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(1, 844, __pyx_L1_error)
  }
  __pyx_t_7 = PyList_GET_SIZE(__pyx_v_candidates); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1))) __PYX_ERR(1, 844, __pyx_L1_error)
  __pyx_t_2 = ((__pyx_t_7 > 1) != 0);
  if (__pyx_t_2) {
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__16, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 844, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(1, 844, __pyx_L1_error)
  }
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    if (unlikely(((PyObject *)__pyx_v_signatures) == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 844, __pyx_L1_error)
    }
    __pyx_t_6 = __Pyx_PyDict_GetItem(((PyObject*)__pyx_v_signatures), PyList_GET_ITEM(__pyx_v_candidates, 0)); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 844, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_r = __pyx_t_6;
    __pyx_t_6 = 0;
    goto __pyx_L0;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("pandas._libs.interval.Float32ClosedNeitherIntervalNode.__pyx_fused_cpdef", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_dest_sig);
  __Pyx_XDECREF(__pyx_v_ndarray);
  __Pyx_XDECREF(__pyx_v_arg);
  __Pyx_XDECREF(__pyx_v_candidates);
  __Pyx_XDECREF(__pyx_v_sig);
  __Pyx_XDECREF(__pyx_v_src_sig);
  __Pyx_XDECREF(__pyx_v_dst_type);
  __Pyx_XDECREF(__pyx_v_kwargs);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pw_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_9__pyx_fuse_0query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_fuse_0__pyx_f_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float64_t __pyx_v_point, int __pyx_skip_dispatch) {
  __Pyx_memviewslice __pyx_v_indices = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_values = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __pyx_t_5numpy_int64_t __pyx_t_9;
  __pyx_t_5numpy_int64_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  int __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  __Pyx_memviewslice __pyx_t_16 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_17 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  Py_ssize_t __pyx_t_21;
  __Pyx_RefNannySetupContext("__pyx_fuse_0query", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pyx_fuse_0query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 844, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_9__pyx_fuse_0query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = PyFloat_FromDouble(__pyx_v_point); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 844, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 844, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 844, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 844, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_5) {
            __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 844, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 844, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_3query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_4 = PyFloat_FromDouble(__pyx_v_point); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 844, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_7 = __pyx_t_1; __pyx_t_3 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 844, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 844, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 844, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (__pyx_t_3) {
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_t_4);
          __pyx_t_4 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 844, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pandas/_libs/intervaltree.pxi":853
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
  __pyx_t_8 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_8) {

    /* "pandas/_libs/intervaltree.pxi":857
 *             # continue the binary tree structure. Instead, we use linear
 *             # search.
 *             for i in range(self.n_elements):             # <<<<<<<<<<<<<<
 *                 if self.left[i] < point < self.right[i]:
 *                     result.append(self.indices[i])
 */
    __pyx_t_9 = __pyx_v_self->n_elements;
    __pyx_t_10 = __pyx_t_9;
    for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
      __pyx_v_i = __pyx_t_11;

      /* "pandas/_libs/intervaltree.pxi":858
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point < self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      __pyx_t_12 = __pyx_v_i;
      __pyx_t_8 = ((*((__pyx_t_5numpy_float32_t *) ( /* dim=0 */ (__pyx_v_self->left.data + __pyx_t_12 * __pyx_v_self->left.strides[0]) ))) < __pyx_v_point);
      if (__pyx_t_8) {
        __pyx_t_13 = __pyx_v_i;
        __pyx_t_8 = (__pyx_v_point < (*((__pyx_t_5numpy_float32_t *) ( /* dim=0 */ (__pyx_v_self->right.data + __pyx_t_13 * __pyx_v_self->right.strides[0]) ))));
      }
      __pyx_t_14 = (__pyx_t_8 != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":859
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point < self.right[i]:
 *                     result.append(self.indices[i])             # <<<<<<<<<<<<<<
 *         else:
 *             # There are child nodes. Based on comparing our query to the pivot,
 */
        __pyx_t_15 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->indices.data + __pyx_t_15 * __pyx_v_self->indices.strides[0]) ))));

        /* "pandas/_libs/intervaltree.pxi":858
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point < self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      }
    }

    /* "pandas/_libs/intervaltree.pxi":853
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
    goto __pyx_L3;
  }

  /* "pandas/_libs/intervaltree.pxi":863
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
  /*else*/ {
    __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":864
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:
 *                 values = self.center_left_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 */
      __pyx_t_16 = __pyx_v_self->center_left_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":865
 *             if point < self.pivot:
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:
 */
      __pyx_t_17 = __pyx_v_self->center_left_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":866
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):             # <<<<<<<<<<<<<<
 *                     if not values[i] < point:
 *                         break
 */
      __pyx_t_9 = __pyx_v_self->n_center;
      __pyx_t_10 = __pyx_t_9;
      for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":867
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_18 = __pyx_v_i;
        __pyx_t_14 = ((!(((*((__pyx_t_5numpy_float32_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_18 * __pyx_v_values.strides[0]) ))) < __pyx_v_point) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":868
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:
 */
          goto __pyx_L9_break;

          /* "pandas/_libs/intervaltree.pxi":867
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":869
 *                     if not values[i] < point:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 */
        __pyx_t_19 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_19 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L9_break:;

      /* "pandas/_libs/intervaltree.pxi":870
 *                         break
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->left_node->max_right) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":871
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)             # <<<<<<<<<<<<<<
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode *)__pyx_v_self->left_node->__pyx_vtab)->__pyx_fuse_0query(__pyx_v_self->left_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 871, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":870
 *                         break
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":863
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":872
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
    __pyx_t_14 = ((__pyx_v_point > __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":873
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 *                 values = self.center_right_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 */
      __pyx_t_16 = __pyx_v_self->center_right_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":874
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:
 */
      __pyx_t_17 = __pyx_v_self->center_right_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":875
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):             # <<<<<<<<<<<<<<
 *                     if not point < values[i]:
 *                         break
 */
      for (__pyx_t_11 = (__pyx_v_self->n_center - 1); __pyx_t_11 > -1L; __pyx_t_11-=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":876
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_20 = __pyx_v_i;
        __pyx_t_14 = ((!((__pyx_v_point < (*((__pyx_t_5numpy_float32_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_20 * __pyx_v_values.strides[0]) )))) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":877
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:
 */
          goto __pyx_L13_break;

          /* "pandas/_libs/intervaltree.pxi":876
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":878
 *                     if not point < values[i]:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if self.right_node.min_left < point:
 *                     self.right_node.query(result, point)
 */
        __pyx_t_21 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_21 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L13_break:;

      /* "pandas/_libs/intervaltree.pxi":879
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      __pyx_t_14 = ((__pyx_v_self->right_node->min_left < __pyx_v_point) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":880
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:
 *                     self.right_node.query(result, point)             # <<<<<<<<<<<<<<
 *             else:
 *                 result.extend(self.center_left_indices)
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode *)__pyx_v_self->right_node->__pyx_vtab)->__pyx_fuse_0query(__pyx_v_self->right_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 880, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":879
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":872
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":882
 *                     self.right_node.query(result, point)
 *             else:
 *                 result.extend(self.center_left_indices)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
    /*else*/ {
      __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->extend(__pyx_v_result, __pyx_v_self->center_left_indices); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 882, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_L7:;
  }
  __pyx_L3:;

  /* "pandas/_libs/intervaltree.pxi":844
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __PYX_XDEC_MEMVIEW(&__pyx_t_16, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_17, 1);
  __Pyx_AddTraceback("pandas._libs.interval.Float32ClosedNeitherIntervalNode.query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_indices, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_values, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_9__pyx_fuse_0query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_0__pyx_mdef_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_9__pyx_fuse_0query = {"__pyx_fuse_0query", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_9__pyx_fuse_0query, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_2query};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_9__pyx_fuse_0query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result = 0;
  __pyx_t_5numpy_float64_t __pyx_v_point;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fuse_0query (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_result,&__pyx_n_s_point,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_result)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_point)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_0query", 1, 2, 2, 1); __PYX_ERR(1, 844, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_0query") < 0)) __PYX_ERR(1, 844, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_result = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)values[0]);
    __pyx_v_point = __pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_point == ((npy_float64)-1)) && PyErr_Occurred())) __PYX_ERR(1, 844, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_0query", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 844, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Float32ClosedNeitherIntervalNode.__pyx_fuse_0query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_result), __pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector, 1, "result", 0))) __PYX_ERR(1, 844, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_8__pyx_fuse_0query(((struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode *)__pyx_v_self), __pyx_v_result, __pyx_v_point);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_8__pyx_fuse_0query(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float64_t __pyx_v_point) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__pyx_fuse_0query", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_fuse_0__pyx_f_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_query(__pyx_v_self, __pyx_v_result, __pyx_v_point, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 844, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Float32ClosedNeitherIntervalNode.__pyx_fuse_0query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pw_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_11__pyx_fuse_1query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_fuse_1__pyx_f_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float32_t __pyx_v_point, int __pyx_skip_dispatch) {
  __Pyx_memviewslice __pyx_v_indices = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_values = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __pyx_t_5numpy_int64_t __pyx_t_9;
  __pyx_t_5numpy_int64_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  int __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  __Pyx_memviewslice __pyx_t_16 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_17 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  Py_ssize_t __pyx_t_21;
  __Pyx_RefNannySetupContext("__pyx_fuse_1query", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pyx_fuse_1query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 844, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_11__pyx_fuse_1query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = PyFloat_FromDouble(__pyx_v_point); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 844, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 844, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 844, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 844, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_5) {
            __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 844, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 844, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_3query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_4 = PyFloat_FromDouble(__pyx_v_point); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 844, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_7 = __pyx_t_1; __pyx_t_3 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 844, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 844, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 844, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (__pyx_t_3) {
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_t_4);
          __pyx_t_4 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 844, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pandas/_libs/intervaltree.pxi":853
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
  __pyx_t_8 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_8) {

    /* "pandas/_libs/intervaltree.pxi":857
 *             # continue the binary tree structure. Instead, we use linear
 *             # search.
 *             for i in range(self.n_elements):             # <<<<<<<<<<<<<<
 *                 if self.left[i] < point < self.right[i]:
 *                     result.append(self.indices[i])
 */
    __pyx_t_9 = __pyx_v_self->n_elements;
    __pyx_t_10 = __pyx_t_9;
    for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
      __pyx_v_i = __pyx_t_11;

      /* "pandas/_libs/intervaltree.pxi":858
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point < self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      __pyx_t_12 = __pyx_v_i;
      __pyx_t_8 = ((*((__pyx_t_5numpy_float32_t *) ( /* dim=0 */ (__pyx_v_self->left.data + __pyx_t_12 * __pyx_v_self->left.strides[0]) ))) < __pyx_v_point);
      if (__pyx_t_8) {
        __pyx_t_13 = __pyx_v_i;
        __pyx_t_8 = (__pyx_v_point < (*((__pyx_t_5numpy_float32_t *) ( /* dim=0 */ (__pyx_v_self->right.data + __pyx_t_13 * __pyx_v_self->right.strides[0]) ))));
      }
      __pyx_t_14 = (__pyx_t_8 != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":859
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point < self.right[i]:
 *                     result.append(self.indices[i])             # <<<<<<<<<<<<<<
 *         else:
 *             # There are child nodes. Based on comparing our query to the pivot,
 */
        __pyx_t_15 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->indices.data + __pyx_t_15 * __pyx_v_self->indices.strides[0]) ))));

        /* "pandas/_libs/intervaltree.pxi":858
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point < self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      }
    }

    /* "pandas/_libs/intervaltree.pxi":853
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
    goto __pyx_L3;
  }

  /* "pandas/_libs/intervaltree.pxi":863
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
  /*else*/ {
    __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":864
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:
 *                 values = self.center_left_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 */
      __pyx_t_16 = __pyx_v_self->center_left_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":865
 *             if point < self.pivot:
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:
 */
      __pyx_t_17 = __pyx_v_self->center_left_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":866
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):             # <<<<<<<<<<<<<<
 *                     if not values[i] < point:
 *                         break
 */
      __pyx_t_9 = __pyx_v_self->n_center;
      __pyx_t_10 = __pyx_t_9;
      for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":867
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_18 = __pyx_v_i;
        __pyx_t_14 = ((!(((*((__pyx_t_5numpy_float32_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_18 * __pyx_v_values.strides[0]) ))) < __pyx_v_point) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":868
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:
 */
          goto __pyx_L9_break;

          /* "pandas/_libs/intervaltree.pxi":867
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":869
 *                     if not values[i] < point:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 */
        __pyx_t_19 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_19 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L9_break:;

      /* "pandas/_libs/intervaltree.pxi":870
 *                         break
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->left_node->max_right) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":871
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)             # <<<<<<<<<<<<<<
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode *)__pyx_v_self->left_node->__pyx_vtab)->__pyx_fuse_1query(__pyx_v_self->left_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 871, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":870
 *                         break
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":863
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":872
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
    __pyx_t_14 = ((__pyx_v_point > __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":873
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 *                 values = self.center_right_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 */
      __pyx_t_16 = __pyx_v_self->center_right_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":874
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:
 */
      __pyx_t_17 = __pyx_v_self->center_right_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":875
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):             # <<<<<<<<<<<<<<
 *                     if not point < values[i]:
 *                         break
 */
      for (__pyx_t_11 = (__pyx_v_self->n_center - 1); __pyx_t_11 > -1L; __pyx_t_11-=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":876
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_20 = __pyx_v_i;
        __pyx_t_14 = ((!((__pyx_v_point < (*((__pyx_t_5numpy_float32_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_20 * __pyx_v_values.strides[0]) )))) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":877
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:
 */
          goto __pyx_L13_break;

          /* "pandas/_libs/intervaltree.pxi":876
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":878
 *                     if not point < values[i]:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if self.right_node.min_left < point:
 *                     self.right_node.query(result, point)
 */
        __pyx_t_21 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_21 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L13_break:;

      /* "pandas/_libs/intervaltree.pxi":879
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      __pyx_t_14 = ((__pyx_v_self->right_node->min_left < __pyx_v_point) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":880
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:
 *                     self.right_node.query(result, point)             # <<<<<<<<<<<<<<
 *             else:
 *                 result.extend(self.center_left_indices)
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode *)__pyx_v_self->right_node->__pyx_vtab)->__pyx_fuse_1query(__pyx_v_self->right_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 880, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":879
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":872
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":882
 *                     self.right_node.query(result, point)
 *             else:
 *                 result.extend(self.center_left_indices)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
    /*else*/ {
      __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->extend(__pyx_v_result, __pyx_v_self->center_left_indices); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 882, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_L7:;
  }
  __pyx_L3:;

  /* "pandas/_libs/intervaltree.pxi":844
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __PYX_XDEC_MEMVIEW(&__pyx_t_16, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_17, 1);
  __Pyx_AddTraceback("pandas._libs.interval.Float32ClosedNeitherIntervalNode.query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_indices, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_values, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_11__pyx_fuse_1query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_1__pyx_mdef_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_11__pyx_fuse_1query = {"__pyx_fuse_1query", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_11__pyx_fuse_1query, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_2query};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_11__pyx_fuse_1query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result = 0;
  __pyx_t_5numpy_float32_t __pyx_v_point;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fuse_1query (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_result,&__pyx_n_s_point,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_result)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_point)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_1query", 1, 2, 2, 1); __PYX_ERR(1, 844, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_1query") < 0)) __PYX_ERR(1, 844, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_result = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)values[0]);
    __pyx_v_point = __pyx_PyFloat_AsFloat(values[1]); if (unlikely((__pyx_v_point == ((npy_float32)-1)) && PyErr_Occurred())) __PYX_ERR(1, 844, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_1query", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 844, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Float32ClosedNeitherIntervalNode.__pyx_fuse_1query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_result), __pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector, 1, "result", 0))) __PYX_ERR(1, 844, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_10__pyx_fuse_1query(((struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode *)__pyx_v_self), __pyx_v_result, __pyx_v_point);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_10__pyx_fuse_1query(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float32_t __pyx_v_point) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__pyx_fuse_1query", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_fuse_1__pyx_f_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_query(__pyx_v_self, __pyx_v_result, __pyx_v_point, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 844, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Float32ClosedNeitherIntervalNode.__pyx_fuse_1query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pw_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_13__pyx_fuse_2query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_fuse_2__pyx_f_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int64_t __pyx_v_point, int __pyx_skip_dispatch) {
  __Pyx_memviewslice __pyx_v_indices = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_values = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __pyx_t_5numpy_int64_t __pyx_t_9;
  __pyx_t_5numpy_int64_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  int __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  __Pyx_memviewslice __pyx_t_16 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_17 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  Py_ssize_t __pyx_t_21;
  __Pyx_RefNannySetupContext("__pyx_fuse_2query", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pyx_fuse_2query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 844, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_13__pyx_fuse_2query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = __Pyx_PyInt_From_npy_int64(__pyx_v_point); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 844, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 844, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 844, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 844, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_5) {
            __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 844, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 844, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_3query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_4 = __Pyx_PyInt_From_npy_int64(__pyx_v_point); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 844, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_7 = __pyx_t_1; __pyx_t_3 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 844, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 844, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 844, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (__pyx_t_3) {
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_t_4);
          __pyx_t_4 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 844, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pandas/_libs/intervaltree.pxi":853
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
  __pyx_t_8 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_8) {

    /* "pandas/_libs/intervaltree.pxi":857
 *             # continue the binary tree structure. Instead, we use linear
 *             # search.
 *             for i in range(self.n_elements):             # <<<<<<<<<<<<<<
 *                 if self.left[i] < point < self.right[i]:
 *                     result.append(self.indices[i])
 */
    __pyx_t_9 = __pyx_v_self->n_elements;
    __pyx_t_10 = __pyx_t_9;
    for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
      __pyx_v_i = __pyx_t_11;

      /* "pandas/_libs/intervaltree.pxi":858
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point < self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      __pyx_t_12 = __pyx_v_i;
      __pyx_t_8 = ((*((__pyx_t_5numpy_float32_t *) ( /* dim=0 */ (__pyx_v_self->left.data + __pyx_t_12 * __pyx_v_self->left.strides[0]) ))) < __pyx_v_point);
      if (__pyx_t_8) {
        __pyx_t_13 = __pyx_v_i;
        __pyx_t_8 = (__pyx_v_point < (*((__pyx_t_5numpy_float32_t *) ( /* dim=0 */ (__pyx_v_self->right.data + __pyx_t_13 * __pyx_v_self->right.strides[0]) ))));
      }
      __pyx_t_14 = (__pyx_t_8 != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":859
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point < self.right[i]:
 *                     result.append(self.indices[i])             # <<<<<<<<<<<<<<
 *         else:
 *             # There are child nodes. Based on comparing our query to the pivot,
 */
        __pyx_t_15 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->indices.data + __pyx_t_15 * __pyx_v_self->indices.strides[0]) ))));

        /* "pandas/_libs/intervaltree.pxi":858
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point < self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      }
    }

    /* "pandas/_libs/intervaltree.pxi":853
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
    goto __pyx_L3;
  }

  /* "pandas/_libs/intervaltree.pxi":863
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
  /*else*/ {
    __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":864
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:
 *                 values = self.center_left_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 */
      __pyx_t_16 = __pyx_v_self->center_left_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":865
 *             if point < self.pivot:
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:
 */
      __pyx_t_17 = __pyx_v_self->center_left_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":866
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):             # <<<<<<<<<<<<<<
 *                     if not values[i] < point:
 *                         break
 */
      __pyx_t_9 = __pyx_v_self->n_center;
      __pyx_t_10 = __pyx_t_9;
      for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":867
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_18 = __pyx_v_i;
        __pyx_t_14 = ((!(((*((__pyx_t_5numpy_float32_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_18 * __pyx_v_values.strides[0]) ))) < __pyx_v_point) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":868
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:
 */
          goto __pyx_L9_break;

          /* "pandas/_libs/intervaltree.pxi":867
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":869
 *                     if not values[i] < point:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 */
        __pyx_t_19 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_19 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L9_break:;

      /* "pandas/_libs/intervaltree.pxi":870
 *                         break
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->left_node->max_right) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":871
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)             # <<<<<<<<<<<<<<
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode *)__pyx_v_self->left_node->__pyx_vtab)->__pyx_fuse_2query(__pyx_v_self->left_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 871, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":870
 *                         break
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":863
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":872
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
    __pyx_t_14 = ((__pyx_v_point > __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":873
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 *                 values = self.center_right_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 */
      __pyx_t_16 = __pyx_v_self->center_right_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":874
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:
 */
      __pyx_t_17 = __pyx_v_self->center_right_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":875
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):             # <<<<<<<<<<<<<<
 *                     if not point < values[i]:
 *                         break
 */
      for (__pyx_t_11 = (__pyx_v_self->n_center - 1); __pyx_t_11 > -1L; __pyx_t_11-=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":876
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_20 = __pyx_v_i;
        __pyx_t_14 = ((!((__pyx_v_point < (*((__pyx_t_5numpy_float32_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_20 * __pyx_v_values.strides[0]) )))) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":877
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:
 */
          goto __pyx_L13_break;

          /* "pandas/_libs/intervaltree.pxi":876
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":878
 *                     if not point < values[i]:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if self.right_node.min_left < point:
 *                     self.right_node.query(result, point)
 */
        __pyx_t_21 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_21 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L13_break:;

      /* "pandas/_libs/intervaltree.pxi":879
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      __pyx_t_14 = ((__pyx_v_self->right_node->min_left < __pyx_v_point) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":880
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:
 *                     self.right_node.query(result, point)             # <<<<<<<<<<<<<<
 *             else:
 *                 result.extend(self.center_left_indices)
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode *)__pyx_v_self->right_node->__pyx_vtab)->__pyx_fuse_2query(__pyx_v_self->right_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 880, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":879
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":872
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":882
 *                     self.right_node.query(result, point)
 *             else:
 *                 result.extend(self.center_left_indices)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
    /*else*/ {
      __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->extend(__pyx_v_result, __pyx_v_self->center_left_indices); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 882, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_L7:;
  }
  __pyx_L3:;

  /* "pandas/_libs/intervaltree.pxi":844
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __PYX_XDEC_MEMVIEW(&__pyx_t_16, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_17, 1);
  __Pyx_AddTraceback("pandas._libs.interval.Float32ClosedNeitherIntervalNode.query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_indices, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_values, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_13__pyx_fuse_2query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_2__pyx_mdef_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_13__pyx_fuse_2query = {"__pyx_fuse_2query", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_13__pyx_fuse_2query, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_2query};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_13__pyx_fuse_2query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result = 0;
  __pyx_t_5numpy_int64_t __pyx_v_point;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fuse_2query (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_result,&__pyx_n_s_point,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_result)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_point)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_2query", 1, 2, 2, 1); __PYX_ERR(1, 844, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_2query") < 0)) __PYX_ERR(1, 844, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_result = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)values[0]);
    __pyx_v_point = __Pyx_PyInt_As_npy_int64(values[1]); if (unlikely((__pyx_v_point == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(1, 844, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_2query", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 844, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Float32ClosedNeitherIntervalNode.__pyx_fuse_2query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_result), __pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector, 1, "result", 0))) __PYX_ERR(1, 844, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_12__pyx_fuse_2query(((struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode *)__pyx_v_self), __pyx_v_result, __pyx_v_point);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_12__pyx_fuse_2query(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int64_t __pyx_v_point) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__pyx_fuse_2query", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_fuse_2__pyx_f_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_query(__pyx_v_self, __pyx_v_result, __pyx_v_point, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 844, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Float32ClosedNeitherIntervalNode.__pyx_fuse_2query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pw_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_15__pyx_fuse_3query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_fuse_3__pyx_f_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int32_t __pyx_v_point, int __pyx_skip_dispatch) {
  __Pyx_memviewslice __pyx_v_indices = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_values = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __pyx_t_5numpy_int64_t __pyx_t_9;
  __pyx_t_5numpy_int64_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  int __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  __Pyx_memviewslice __pyx_t_16 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_17 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  Py_ssize_t __pyx_t_21;
  __Pyx_RefNannySetupContext("__pyx_fuse_3query", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pyx_fuse_3query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 844, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_15__pyx_fuse_3query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = __Pyx_PyInt_From_npy_int32(__pyx_v_point); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 844, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 844, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 844, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 844, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_5) {
            __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 844, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 844, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_3query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_4 = __Pyx_PyInt_From_npy_int32(__pyx_v_point); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 844, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_7 = __pyx_t_1; __pyx_t_3 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 844, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 844, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 844, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (__pyx_t_3) {
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_t_4);
          __pyx_t_4 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 844, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pandas/_libs/intervaltree.pxi":853
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
  __pyx_t_8 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_8) {

    /* "pandas/_libs/intervaltree.pxi":857
 *             # continue the binary tree structure. Instead, we use linear
 *             # search.
 *             for i in range(self.n_elements):             # <<<<<<<<<<<<<<
 *                 if self.left[i] < point < self.right[i]:
 *                     result.append(self.indices[i])
 */
    __pyx_t_9 = __pyx_v_self->n_elements;
    __pyx_t_10 = __pyx_t_9;
    for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
      __pyx_v_i = __pyx_t_11;

      /* "pandas/_libs/intervaltree.pxi":858
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point < self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      __pyx_t_12 = __pyx_v_i;
      __pyx_t_8 = ((*((__pyx_t_5numpy_float32_t *) ( /* dim=0 */ (__pyx_v_self->left.data + __pyx_t_12 * __pyx_v_self->left.strides[0]) ))) < __pyx_v_point);
      if (__pyx_t_8) {
        __pyx_t_13 = __pyx_v_i;
        __pyx_t_8 = (__pyx_v_point < (*((__pyx_t_5numpy_float32_t *) ( /* dim=0 */ (__pyx_v_self->right.data + __pyx_t_13 * __pyx_v_self->right.strides[0]) ))));
      }
      __pyx_t_14 = (__pyx_t_8 != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":859
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point < self.right[i]:
 *                     result.append(self.indices[i])             # <<<<<<<<<<<<<<
 *         else:
 *             # There are child nodes. Based on comparing our query to the pivot,
 */
        __pyx_t_15 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->indices.data + __pyx_t_15 * __pyx_v_self->indices.strides[0]) ))));

        /* "pandas/_libs/intervaltree.pxi":858
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point < self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      }
    }

    /* "pandas/_libs/intervaltree.pxi":853
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
    goto __pyx_L3;
  }

  /* "pandas/_libs/intervaltree.pxi":863
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
  /*else*/ {
    __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":864
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:
 *                 values = self.center_left_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 */
      __pyx_t_16 = __pyx_v_self->center_left_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":865
 *             if point < self.pivot:
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:
 */
      __pyx_t_17 = __pyx_v_self->center_left_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":866
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):             # <<<<<<<<<<<<<<
 *                     if not values[i] < point:
 *                         break
 */
      __pyx_t_9 = __pyx_v_self->n_center;
      __pyx_t_10 = __pyx_t_9;
      for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":867
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_18 = __pyx_v_i;
        __pyx_t_14 = ((!(((*((__pyx_t_5numpy_float32_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_18 * __pyx_v_values.strides[0]) ))) < __pyx_v_point) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":868
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:
 */
          goto __pyx_L9_break;

          /* "pandas/_libs/intervaltree.pxi":867
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":869
 *                     if not values[i] < point:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 */
        __pyx_t_19 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_19 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L9_break:;

      /* "pandas/_libs/intervaltree.pxi":870
 *                         break
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->left_node->max_right) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":871
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)             # <<<<<<<<<<<<<<
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode *)__pyx_v_self->left_node->__pyx_vtab)->__pyx_fuse_3query(__pyx_v_self->left_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 871, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":870
 *                         break
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":863
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":872
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
    __pyx_t_14 = ((__pyx_v_point > __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":873
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 *                 values = self.center_right_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 */
      __pyx_t_16 = __pyx_v_self->center_right_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":874
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:
 */
      __pyx_t_17 = __pyx_v_self->center_right_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":875
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):             # <<<<<<<<<<<<<<
 *                     if not point < values[i]:
 *                         break
 */
      for (__pyx_t_11 = (__pyx_v_self->n_center - 1); __pyx_t_11 > -1L; __pyx_t_11-=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":876
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_20 = __pyx_v_i;
        __pyx_t_14 = ((!((__pyx_v_point < (*((__pyx_t_5numpy_float32_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_20 * __pyx_v_values.strides[0]) )))) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":877
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:
 */
          goto __pyx_L13_break;

          /* "pandas/_libs/intervaltree.pxi":876
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":878
 *                     if not point < values[i]:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if self.right_node.min_left < point:
 *                     self.right_node.query(result, point)
 */
        __pyx_t_21 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_21 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L13_break:;

      /* "pandas/_libs/intervaltree.pxi":879
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      __pyx_t_14 = ((__pyx_v_self->right_node->min_left < __pyx_v_point) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":880
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:
 *                     self.right_node.query(result, point)             # <<<<<<<<<<<<<<
 *             else:
 *                 result.extend(self.center_left_indices)
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode *)__pyx_v_self->right_node->__pyx_vtab)->__pyx_fuse_3query(__pyx_v_self->right_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 880, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":879
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":872
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":882
 *                     self.right_node.query(result, point)
 *             else:
 *                 result.extend(self.center_left_indices)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
    /*else*/ {
      __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->extend(__pyx_v_result, __pyx_v_self->center_left_indices); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 882, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_L7:;
  }
  __pyx_L3:;

  /* "pandas/_libs/intervaltree.pxi":844
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __PYX_XDEC_MEMVIEW(&__pyx_t_16, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_17, 1);
  __Pyx_AddTraceback("pandas._libs.interval.Float32ClosedNeitherIntervalNode.query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_indices, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_values, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_15__pyx_fuse_3query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_3__pyx_mdef_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_15__pyx_fuse_3query = {"__pyx_fuse_3query", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_15__pyx_fuse_3query, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_2query};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_15__pyx_fuse_3query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result = 0;
  __pyx_t_5numpy_int32_t __pyx_v_point;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fuse_3query (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_result,&__pyx_n_s_point,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_result)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_point)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_3query", 1, 2, 2, 1); __PYX_ERR(1, 844, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_3query") < 0)) __PYX_ERR(1, 844, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_result = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)values[0]);
    __pyx_v_point = __Pyx_PyInt_As_npy_int32(values[1]); if (unlikely((__pyx_v_point == ((npy_int32)-1)) && PyErr_Occurred())) __PYX_ERR(1, 844, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_3query", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 844, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Float32ClosedNeitherIntervalNode.__pyx_fuse_3query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_result), __pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector, 1, "result", 0))) __PYX_ERR(1, 844, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_14__pyx_fuse_3query(((struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode *)__pyx_v_self), __pyx_v_result, __pyx_v_point);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_14__pyx_fuse_3query(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int32_t __pyx_v_point) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__pyx_fuse_3query", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_fuse_3__pyx_f_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_query(__pyx_v_self, __pyx_v_result, __pyx_v_point, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 844, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Float32ClosedNeitherIntervalNode.__pyx_fuse_3query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pw_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_17__pyx_fuse_4query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_fuse_4__pyx_f_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_uint64_t __pyx_v_point, int __pyx_skip_dispatch) {
  __Pyx_memviewslice __pyx_v_indices = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_values = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __pyx_t_5numpy_int64_t __pyx_t_9;
  __pyx_t_5numpy_int64_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  int __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  __Pyx_memviewslice __pyx_t_16 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_17 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  Py_ssize_t __pyx_t_21;
  __Pyx_RefNannySetupContext("__pyx_fuse_4query", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pyx_fuse_4query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 844, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_17__pyx_fuse_4query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = __Pyx_PyInt_From_npy_uint64(__pyx_v_point); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 844, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 844, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 844, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 844, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_5) {
            __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 844, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 844, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_3query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_4 = __Pyx_PyInt_From_npy_uint64(__pyx_v_point); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 844, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_7 = __pyx_t_1; __pyx_t_3 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 844, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 844, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 844, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (__pyx_t_3) {
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_t_4);
          __pyx_t_4 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 844, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pandas/_libs/intervaltree.pxi":853
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
  __pyx_t_8 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_8) {

    /* "pandas/_libs/intervaltree.pxi":857
 *             # continue the binary tree structure. Instead, we use linear
 *             # search.
 *             for i in range(self.n_elements):             # <<<<<<<<<<<<<<
 *                 if self.left[i] < point < self.right[i]:
 *                     result.append(self.indices[i])
 */
    __pyx_t_9 = __pyx_v_self->n_elements;
    __pyx_t_10 = __pyx_t_9;
    for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
      __pyx_v_i = __pyx_t_11;

      /* "pandas/_libs/intervaltree.pxi":858
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point < self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      __pyx_t_12 = __pyx_v_i;
      __pyx_t_8 = ((*((__pyx_t_5numpy_float32_t *) ( /* dim=0 */ (__pyx_v_self->left.data + __pyx_t_12 * __pyx_v_self->left.strides[0]) ))) < __pyx_v_point);
      if (__pyx_t_8) {
        __pyx_t_13 = __pyx_v_i;
        __pyx_t_8 = (__pyx_v_point < (*((__pyx_t_5numpy_float32_t *) ( /* dim=0 */ (__pyx_v_self->right.data + __pyx_t_13 * __pyx_v_self->right.strides[0]) ))));
      }
      __pyx_t_14 = (__pyx_t_8 != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":859
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point < self.right[i]:
 *                     result.append(self.indices[i])             # <<<<<<<<<<<<<<
 *         else:
 *             # There are child nodes. Based on comparing our query to the pivot,
 */
        __pyx_t_15 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->indices.data + __pyx_t_15 * __pyx_v_self->indices.strides[0]) ))));

        /* "pandas/_libs/intervaltree.pxi":858
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point < self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      }
    }

    /* "pandas/_libs/intervaltree.pxi":853
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
    goto __pyx_L3;
  }

  /* "pandas/_libs/intervaltree.pxi":863
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
  /*else*/ {
    __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":864
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:
 *                 values = self.center_left_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 */
      __pyx_t_16 = __pyx_v_self->center_left_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":865
 *             if point < self.pivot:
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:
 */
      __pyx_t_17 = __pyx_v_self->center_left_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":866
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):             # <<<<<<<<<<<<<<
 *                     if not values[i] < point:
 *                         break
 */
      __pyx_t_9 = __pyx_v_self->n_center;
      __pyx_t_10 = __pyx_t_9;
      for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":867
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_18 = __pyx_v_i;
        __pyx_t_14 = ((!(((*((__pyx_t_5numpy_float32_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_18 * __pyx_v_values.strides[0]) ))) < __pyx_v_point) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":868
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:
 */
          goto __pyx_L9_break;

          /* "pandas/_libs/intervaltree.pxi":867
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":869
 *                     if not values[i] < point:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 */
        __pyx_t_19 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_19 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L9_break:;

      /* "pandas/_libs/intervaltree.pxi":870
 *                         break
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->left_node->max_right) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":871
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)             # <<<<<<<<<<<<<<
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode *)__pyx_v_self->left_node->__pyx_vtab)->__pyx_fuse_4query(__pyx_v_self->left_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 871, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":870
 *                         break
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":863
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":872
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
    __pyx_t_14 = ((__pyx_v_point > __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":873
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 *                 values = self.center_right_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 */
      __pyx_t_16 = __pyx_v_self->center_right_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":874
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:
 */
      __pyx_t_17 = __pyx_v_self->center_right_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":875
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):             # <<<<<<<<<<<<<<
 *                     if not point < values[i]:
 *                         break
 */
      for (__pyx_t_11 = (__pyx_v_self->n_center - 1); __pyx_t_11 > -1L; __pyx_t_11-=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":876
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_20 = __pyx_v_i;
        __pyx_t_14 = ((!((__pyx_v_point < (*((__pyx_t_5numpy_float32_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_20 * __pyx_v_values.strides[0]) )))) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":877
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:
 */
          goto __pyx_L13_break;

          /* "pandas/_libs/intervaltree.pxi":876
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":878
 *                     if not point < values[i]:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if self.right_node.min_left < point:
 *                     self.right_node.query(result, point)
 */
        __pyx_t_21 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_21 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L13_break:;

      /* "pandas/_libs/intervaltree.pxi":879
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      __pyx_t_14 = ((__pyx_v_self->right_node->min_left < __pyx_v_point) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":880
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:
 *                     self.right_node.query(result, point)             # <<<<<<<<<<<<<<
 *             else:
 *                 result.extend(self.center_left_indices)
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode *)__pyx_v_self->right_node->__pyx_vtab)->__pyx_fuse_4query(__pyx_v_self->right_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 880, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":879
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":872
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":882
 *                     self.right_node.query(result, point)
 *             else:
 *                 result.extend(self.center_left_indices)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
    /*else*/ {
      __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->extend(__pyx_v_result, __pyx_v_self->center_left_indices); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 882, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_L7:;
  }
  __pyx_L3:;

  /* "pandas/_libs/intervaltree.pxi":844
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __PYX_XDEC_MEMVIEW(&__pyx_t_16, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_17, 1);
  __Pyx_AddTraceback("pandas._libs.interval.Float32ClosedNeitherIntervalNode.query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_indices, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_values, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_17__pyx_fuse_4query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_4__pyx_mdef_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_17__pyx_fuse_4query = {"__pyx_fuse_4query", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_17__pyx_fuse_4query, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_2query};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_17__pyx_fuse_4query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result = 0;
  __pyx_t_5numpy_uint64_t __pyx_v_point;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fuse_4query (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_result,&__pyx_n_s_point,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_result)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_point)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_4query", 1, 2, 2, 1); __PYX_ERR(1, 844, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_4query") < 0)) __PYX_ERR(1, 844, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_result = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)values[0]);
    __pyx_v_point = __Pyx_PyInt_As_npy_uint64(values[1]); if (unlikely((__pyx_v_point == ((npy_uint64)-1)) && PyErr_Occurred())) __PYX_ERR(1, 844, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_4query", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 844, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Float32ClosedNeitherIntervalNode.__pyx_fuse_4query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_result), __pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector, 1, "result", 0))) __PYX_ERR(1, 844, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_16__pyx_fuse_4query(((struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode *)__pyx_v_self), __pyx_v_result, __pyx_v_point);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_16__pyx_fuse_4query(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_uint64_t __pyx_v_point) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__pyx_fuse_4query", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_fuse_4__pyx_f_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_query(__pyx_v_self, __pyx_v_result, __pyx_v_point, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 844, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Float32ClosedNeitherIntervalNode.__pyx_fuse_4query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":884
 *                 result.extend(self.center_left_indices)
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         if self.is_leaf_node:
 *             return ('<Float32ClosedNeitherIntervalNode: '
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_5__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_5__repr__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_4__repr__(((struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_4__repr__(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode *__pyx_v_self) {
  PyObject *__pyx_v_n_left = NULL;
  PyObject *__pyx_v_n_right = NULL;
  PyObject *__pyx_v_n_center = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  Py_UCS4 __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "pandas/_libs/intervaltree.pxi":885
 * 
 *     def __repr__(self):
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             return ('<Float32ClosedNeitherIntervalNode: '
 *                     '%s elements (terminal)>' % self.n_elements)
 */
  __pyx_t_1 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_1) {

    /* "pandas/_libs/intervaltree.pxi":886
 *     def __repr__(self):
 *         if self.is_leaf_node:
 *             return ('<Float32ClosedNeitherIntervalNode: '             # <<<<<<<<<<<<<<
 *                     '%s elements (terminal)>' % self.n_elements)
 *         else:
 */
    __Pyx_XDECREF(__pyx_r);

    /* "pandas/_libs/intervaltree.pxi":887
 *         if self.is_leaf_node:
 *             return ('<Float32ClosedNeitherIntervalNode: '
 *                     '%s elements (terminal)>' % self.n_elements)             # <<<<<<<<<<<<<<
 *         else:
 *             n_left = self.left_node.n_elements
 */
    __pyx_t_2 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_elements); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 887, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PyUnicode_Format(__pyx_kp_u_Float32ClosedNeitherIntervalNod, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 887, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pandas/_libs/intervaltree.pxi":885
 * 
 *     def __repr__(self):
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             return ('<Float32ClosedNeitherIntervalNode: '
 *                     '%s elements (terminal)>' % self.n_elements)
 */
  }

  /* "pandas/_libs/intervaltree.pxi":889
 *                     '%s elements (terminal)>' % self.n_elements)
 *         else:
 *             n_left = self.left_node.n_elements             # <<<<<<<<<<<<<<
 *             n_right = self.right_node.n_elements
 *             n_center = self.n_elements - n_left - n_right
 */
  /*else*/ {
    __pyx_t_3 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->left_node->n_elements); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 889, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_v_n_left = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "pandas/_libs/intervaltree.pxi":890
 *         else:
 *             n_left = self.left_node.n_elements
 *             n_right = self.right_node.n_elements             # <<<<<<<<<<<<<<
 *             n_center = self.n_elements - n_left - n_right
 *             return ('<Float32ClosedNeitherIntervalNode: '
 */
    __pyx_t_3 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->right_node->n_elements); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 890, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_v_n_right = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "pandas/_libs/intervaltree.pxi":891
 *             n_left = self.left_node.n_elements
 *             n_right = self.right_node.n_elements
 *             n_center = self.n_elements - n_left - n_right             # <<<<<<<<<<<<<<
 *             return ('<Float32ClosedNeitherIntervalNode: '
 *                     'pivot %s, %s elements (%s left, %s right, %s '
 */
    __pyx_t_3 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_elements); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 891, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = PyNumber_Subtract(__pyx_t_3, __pyx_v_n_left); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 891, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyNumber_Subtract(__pyx_t_2, __pyx_v_n_right); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 891, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_n_center = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "pandas/_libs/intervaltree.pxi":892
 *             n_right = self.right_node.n_elements
 *             n_center = self.n_elements - n_left - n_right
 *             return ('<Float32ClosedNeitherIntervalNode: '             # <<<<<<<<<<<<<<
 *                     'pivot %s, %s elements (%s left, %s right, %s '
 *                     'overlapping)>' % (self.pivot, self.n_elements,
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = PyTuple_New(11); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 892, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = 0;
    __pyx_t_5 = 127;
    __Pyx_INCREF(__pyx_kp_u_Float32ClosedNeitherIntervalNod_2);
    __pyx_t_4 += 41;
    __Pyx_GIVEREF(__pyx_kp_u_Float32ClosedNeitherIntervalNod_2);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_kp_u_Float32ClosedNeitherIntervalNod_2);

    /* "pandas/_libs/intervaltree.pxi":894
 *             return ('<Float32ClosedNeitherIntervalNode: '
 *                     'pivot %s, %s elements (%s left, %s right, %s '
 *                     'overlapping)>' % (self.pivot, self.n_elements,             # <<<<<<<<<<<<<<
 *                                        n_left, n_right, n_center))
 * 
 */
    __pyx_t_2 = PyFloat_FromDouble(__pyx_v_self->pivot); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 894, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_t_2), __pyx_empty_unicode); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 894, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_5 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_6) > __pyx_t_5) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_6) : __pyx_t_5;
    __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_6);
    __pyx_t_6 = 0;
    __Pyx_INCREF(__pyx_kp_u__18);
    __pyx_t_4 += 2;
    __Pyx_GIVEREF(__pyx_kp_u__18);
    PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_kp_u__18);
    __pyx_t_6 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_elements); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 894, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_2 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_t_6), __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 894, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_5 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) > __pyx_t_5) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) : __pyx_t_5;
    __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_3, 3, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_elements);
    __pyx_t_4 += 11;
    __Pyx_GIVEREF(__pyx_kp_u_elements);
    PyTuple_SET_ITEM(__pyx_t_3, 4, __pyx_kp_u_elements);

    /* "pandas/_libs/intervaltree.pxi":895
 *                     'pivot %s, %s elements (%s left, %s right, %s '
 *                     'overlapping)>' % (self.pivot, self.n_elements,
 *                                        n_left, n_right, n_center))             # <<<<<<<<<<<<<<
 * 
 *     def counts(self):
 */
    __pyx_t_2 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_v_n_left), __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 895, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) > __pyx_t_5) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) : __pyx_t_5;
    __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_3, 5, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_left_2);
    __pyx_t_4 += 7;
    __Pyx_GIVEREF(__pyx_kp_u_left_2);
    PyTuple_SET_ITEM(__pyx_t_3, 6, __pyx_kp_u_left_2);
    __pyx_t_2 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_v_n_right), __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 895, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) > __pyx_t_5) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) : __pyx_t_5;
    __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_3, 7, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_right_2);
    __pyx_t_4 += 8;
    __Pyx_GIVEREF(__pyx_kp_u_right_2);
    PyTuple_SET_ITEM(__pyx_t_3, 8, __pyx_kp_u_right_2);
    __pyx_t_2 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_v_n_center), __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 895, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) > __pyx_t_5) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) : __pyx_t_5;
    __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_3, 9, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_overlapping);
    __pyx_t_4 += 14;
    __Pyx_GIVEREF(__pyx_kp_u_overlapping);
    PyTuple_SET_ITEM(__pyx_t_3, 10, __pyx_kp_u_overlapping);

    /* "pandas/_libs/intervaltree.pxi":892
 *             n_right = self.right_node.n_elements
 *             n_center = self.n_elements - n_left - n_right
 *             return ('<Float32ClosedNeitherIntervalNode: '             # <<<<<<<<<<<<<<
 *                     'pivot %s, %s elements (%s left, %s right, %s '
 *                     'overlapping)>' % (self.pivot, self.n_elements,
 */
    __pyx_t_2 = __Pyx_PyUnicode_Join(__pyx_t_3, 11, __pyx_t_4, __pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 892, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;
  }

  /* "pandas/_libs/intervaltree.pxi":884
 *                 result.extend(self.center_left_indices)
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         if self.is_leaf_node:
 *             return ('<Float32ClosedNeitherIntervalNode: '
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pandas._libs.interval.Float32ClosedNeitherIntervalNode.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_n_left);
  __Pyx_XDECREF(__pyx_v_n_right);
  __Pyx_XDECREF(__pyx_v_n_center);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":897
 *                                        n_left, n_right, n_center))
 * 
 *     def counts(self):             # <<<<<<<<<<<<<<
 *         """
 *         Inspect counts on this node
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_7counts(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_6counts[] = "\n        Inspect counts on this node\n        useful for debugging purposes\n        ";
static PyObject *__pyx_pw_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_7counts(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("counts (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_6counts(((struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_6counts(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode *__pyx_v_self) {
  Py_ssize_t __pyx_v_m;
  PyObject *__pyx_v_l = NULL;
  PyObject *__pyx_v_r = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  size_t __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("counts", 0);

  /* "pandas/_libs/intervaltree.pxi":902
 *         useful for debugging purposes
 *         """
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             return self.n_elements
 *         else:
 */
  __pyx_t_1 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_1) {

    /* "pandas/_libs/intervaltree.pxi":903
 *         """
 *         if self.is_leaf_node:
 *             return self.n_elements             # <<<<<<<<<<<<<<
 *         else:
 *             m = len(self.center_left_values)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_elements); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 903, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "pandas/_libs/intervaltree.pxi":902
 *         useful for debugging purposes
 *         """
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             return self.n_elements
 *         else:
 */
  }

  /* "pandas/_libs/intervaltree.pxi":905
 *             return self.n_elements
 *         else:
 *             m = len(self.center_left_values)             # <<<<<<<<<<<<<<
 *             l = self.left_node.counts()
 *             r = self.right_node.counts()
 */
  /*else*/ {
    if (unlikely(!__pyx_v_self->center_left_values.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(1, 905, __pyx_L1_error)}
    __pyx_t_3 = __Pyx_MemoryView_Len(__pyx_v_self->center_left_values); 
    __pyx_v_m = __pyx_t_3;

    /* "pandas/_libs/intervaltree.pxi":906
 *         else:
 *             m = len(self.center_left_values)
 *             l = self.left_node.counts()             # <<<<<<<<<<<<<<
 *             r = self.right_node.counts()
 *             return (m, (l, r))
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->left_node), __pyx_n_s_counts); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 906, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 906, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_l = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "pandas/_libs/intervaltree.pxi":907
 *             m = len(self.center_left_values)
 *             l = self.left_node.counts()
 *             r = self.right_node.counts()             # <<<<<<<<<<<<<<
 *             return (m, (l, r))
 * 
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->right_node), __pyx_n_s_counts); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 907, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 907, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_r = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "pandas/_libs/intervaltree.pxi":908
 *             l = self.left_node.counts()
 *             r = self.right_node.counts()
 *             return (m, (l, r))             # <<<<<<<<<<<<<<
 * 
 * NODE_CLASSES['float32',
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = PyInt_FromSsize_t(__pyx_v_m); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 908, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 908, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_v_l);
    __Pyx_GIVEREF(__pyx_v_l);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_l);
    __Pyx_INCREF(__pyx_v_r);
    __Pyx_GIVEREF(__pyx_v_r);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_v_r);
    __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 908, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_4);
    __pyx_t_2 = 0;
    __pyx_t_4 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;
  }

  /* "pandas/_libs/intervaltree.pxi":897
 *                                        n_left, n_right, n_center))
 * 
 *     def counts(self):             # <<<<<<<<<<<<<<
 *         """
 *         Inspect counts on this node
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pandas._libs.interval.Float32ClosedNeitherIntervalNode.counts", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_l);
  __Pyx_XDECREF(__pyx_v_r);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":754
 *         int64_t[:] center_left_indices, center_right_indices, indices
 *         float32_t min_left, max_right
 *         readonly float32_t pivot             # <<<<<<<<<<<<<<
 *         readonly int64_t n_elements, n_center, leaf_size
 *         readonly bint is_leaf_node
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_5pivot_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_5pivot_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_5pivot___get__(((struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_5pivot___get__(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->pivot); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 754, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Float32ClosedNeitherIntervalNode.pivot.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":755
 *         float32_t min_left, max_right
 *         readonly float32_t pivot
 *         readonly int64_t n_elements, n_center, leaf_size             # <<<<<<<<<<<<<<
 *         readonly bint is_leaf_node
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_10n_elements_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_10n_elements_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_10n_elements___get__(((struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_10n_elements___get__(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_elements); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 755, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Float32ClosedNeitherIntervalNode.n_elements.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_8n_center_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_8n_center_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_8n_center___get__(((struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_8n_center___get__(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_center); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 755, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Float32ClosedNeitherIntervalNode.n_center.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_9leaf_size_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_9leaf_size_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_9leaf_size___get__(((struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_9leaf_size___get__(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->leaf_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 755, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Float32ClosedNeitherIntervalNode.leaf_size.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":756
 *         readonly float32_t pivot
 *         readonly int64_t n_elements, n_center, leaf_size
 *         readonly bint is_leaf_node             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self,
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_12is_leaf_node_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_12is_leaf_node_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_12is_leaf_node___get__(((struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_12is_leaf_node___get__(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->is_leaf_node); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 756, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Float32ClosedNeitherIntervalNode.is_leaf_node.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_21__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_21__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_20__reduce_cython__(((struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_20__reduce_cython__(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  int __pyx_t_16;
  int __pyx_t_17;
  int __pyx_t_18;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.center_left_indices, self.center_left_values, self.center_right_indices, self.center_right_values, self.indices, self.is_leaf_node, self.leaf_size, self.left, self.left_node, self.max_right, self.min_left, self.n_center, self.n_elements, self.pivot, self.right, self.right_node)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  if (unlikely(!__pyx_v_self->center_left_indices.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_1 = __pyx_memoryview_fromslice(__pyx_v_self->center_left_indices, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_int64_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_int64_t, 0);; if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (unlikely(!__pyx_v_self->center_left_values.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_2 = __pyx_memoryview_fromslice(__pyx_v_self->center_left_values, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_float32_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_float32_t, 0);; if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (unlikely(!__pyx_v_self->center_right_indices.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_3 = __pyx_memoryview_fromslice(__pyx_v_self->center_right_indices, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_int64_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_int64_t, 0);; if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely(!__pyx_v_self->center_right_values.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_4 = __pyx_memoryview_fromslice(__pyx_v_self->center_right_values, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_float32_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_float32_t, 0);; if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (unlikely(!__pyx_v_self->indices.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_5 = __pyx_memoryview_fromslice(__pyx_v_self->indices, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_int64_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_int64_t, 0);; if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyBool_FromLong(__pyx_v_self->is_leaf_node); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->leaf_size); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (unlikely(!__pyx_v_self->left.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_8 = __pyx_memoryview_fromslice(__pyx_v_self->left, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_float32_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_float32_t, 0);; if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = PyFloat_FromDouble(__pyx_v_self->max_right); if (unlikely(!__pyx_t_9)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_10 = PyFloat_FromDouble(__pyx_v_self->min_left); if (unlikely(!__pyx_t_10)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_11 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_center); if (unlikely(!__pyx_t_11)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_12 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_elements); if (unlikely(!__pyx_t_12)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __pyx_t_13 = PyFloat_FromDouble(__pyx_v_self->pivot); if (unlikely(!__pyx_t_13)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  if (unlikely(!__pyx_v_self->right.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_14 = __pyx_memoryview_fromslice(__pyx_v_self->right, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_float32_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_float32_t, 0);; if (unlikely(!__pyx_t_14)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __pyx_t_15 = PyTuple_New(16); if (unlikely(!__pyx_t_15)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_15, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_15, 1, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_15, 2, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_15, 3, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_15, 4, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_15, 5, __pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_15, 6, __pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_15, 7, __pyx_t_8);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->left_node));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self->left_node));
  PyTuple_SET_ITEM(__pyx_t_15, 8, ((PyObject *)__pyx_v_self->left_node));
  __Pyx_GIVEREF(__pyx_t_9);
  PyTuple_SET_ITEM(__pyx_t_15, 9, __pyx_t_9);
  __Pyx_GIVEREF(__pyx_t_10);
  PyTuple_SET_ITEM(__pyx_t_15, 10, __pyx_t_10);
  __Pyx_GIVEREF(__pyx_t_11);
  PyTuple_SET_ITEM(__pyx_t_15, 11, __pyx_t_11);
  __Pyx_GIVEREF(__pyx_t_12);
  PyTuple_SET_ITEM(__pyx_t_15, 12, __pyx_t_12);
  __Pyx_GIVEREF(__pyx_t_13);
  PyTuple_SET_ITEM(__pyx_t_15, 13, __pyx_t_13);
  __Pyx_GIVEREF(__pyx_t_14);
  PyTuple_SET_ITEM(__pyx_t_15, 14, __pyx_t_14);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->right_node));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self->right_node));
  PyTuple_SET_ITEM(__pyx_t_15, 15, ((PyObject *)__pyx_v_self->right_node));
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  __pyx_t_6 = 0;
  __pyx_t_7 = 0;
  __pyx_t_8 = 0;
  __pyx_t_9 = 0;
  __pyx_t_10 = 0;
  __pyx_t_11 = 0;
  __pyx_t_12 = 0;
  __pyx_t_13 = 0;
  __pyx_t_14 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_15);
  __pyx_t_15 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.center_left_indices, self.center_left_values, self.center_right_indices, self.center_right_values, self.indices, self.is_leaf_node, self.leaf_size, self.left, self.left_node, self.max_right, self.min_left, self.n_center, self.n_elements, self.pivot, self.right, self.right_node)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_15 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_15)) __PYX_ERR(2, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __pyx_v__dict = __pyx_t_15;
  __pyx_t_15 = 0;

  /* "(tree fragment)":7
 *     state = (self.center_left_indices, self.center_left_values, self.center_right_indices, self.center_right_values, self.indices, self.is_leaf_node, self.leaf_size, self.left, self.left_node, self.max_right, self.min_left, self.n_center, self.n_elements, self.pivot, self.right, self.right_node)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_16 = (__pyx_v__dict != Py_None);
  __pyx_t_17 = (__pyx_t_16 != 0);
  if (__pyx_t_17) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_15 = PyTuple_New(1); if (unlikely(!__pyx_t_15)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    PyTuple_SET_ITEM(__pyx_t_15, 0, __pyx_v__dict);
    __pyx_t_14 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_15); if (unlikely(!__pyx_t_14)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_14));
    __pyx_t_14 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.left_node is not None or self.right_node is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.center_left_indices, self.center_left_values, self.center_right_indices, self.center_right_values, self.indices, self.is_leaf_node, self.leaf_size, self.left, self.left_node, self.max_right, self.min_left, self.n_center, self.n_elements, self.pivot, self.right, self.right_node)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.left_node is not None or self.right_node is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_Float32ClosedNeitherIntervalNode, (type(self), 0x7a104e5, None), state
 */
  /*else*/ {
    __pyx_t_16 = (((PyObject *)__pyx_v_self->left_node) != Py_None);
    __pyx_t_18 = (__pyx_t_16 != 0);
    if (!__pyx_t_18) {
    } else {
      __pyx_t_17 = __pyx_t_18;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_18 = (((PyObject *)__pyx_v_self->right_node) != Py_None);
    __pyx_t_16 = (__pyx_t_18 != 0);
    __pyx_t_17 = __pyx_t_16;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_17;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.left_node is not None or self.right_node is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Float32ClosedNeitherIntervalNode, (type(self), 0x7a104e5, None), state
 *     else:
 */
  __pyx_t_17 = (__pyx_v_use_setstate != 0);
  if (__pyx_t_17) {

    /* "(tree fragment)":13
 *         use_setstate = self.left_node is not None or self.right_node is not None
 *     if use_setstate:
 *         return __pyx_unpickle_Float32ClosedNeitherIntervalNode, (type(self), 0x7a104e5, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_Float32ClosedNeitherIntervalNode, (type(self), 0x7a104e5, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_14, __pyx_n_s_pyx_unpickle_Float32ClosedNeit); if (unlikely(!__pyx_t_14)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __pyx_t_15 = PyTuple_New(3); if (unlikely(!__pyx_t_15)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_15, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_127993061);
    __Pyx_GIVEREF(__pyx_int_127993061);
    PyTuple_SET_ITEM(__pyx_t_15, 1, __pyx_int_127993061);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_15, 2, Py_None);
    __pyx_t_13 = PyTuple_New(3); if (unlikely(!__pyx_t_13)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __Pyx_GIVEREF(__pyx_t_14);
    PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_t_14);
    __Pyx_GIVEREF(__pyx_t_15);
    PyTuple_SET_ITEM(__pyx_t_13, 1, __pyx_t_15);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_13, 2, __pyx_v_state);
    __pyx_t_14 = 0;
    __pyx_t_15 = 0;
    __pyx_r = __pyx_t_13;
    __pyx_t_13 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.left_node is not None or self.right_node is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Float32ClosedNeitherIntervalNode, (type(self), 0x7a104e5, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_Float32ClosedNeitherIntervalNode, (type(self), 0x7a104e5, None), state
 *     else:
 *         return __pyx_unpickle_Float32ClosedNeitherIntervalNode, (type(self), 0x7a104e5, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Float32ClosedNeitherIntervalNode__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_13, __pyx_n_s_pyx_unpickle_Float32ClosedNeit); if (unlikely(!__pyx_t_13)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __pyx_t_15 = PyTuple_New(3); if (unlikely(!__pyx_t_15)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_15, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_127993061);
    __Pyx_GIVEREF(__pyx_int_127993061);
    PyTuple_SET_ITEM(__pyx_t_15, 1, __pyx_int_127993061);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_15, 2, __pyx_v_state);
    __pyx_t_14 = PyTuple_New(2); if (unlikely(!__pyx_t_14)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_GIVEREF(__pyx_t_13);
    PyTuple_SET_ITEM(__pyx_t_14, 0, __pyx_t_13);
    __Pyx_GIVEREF(__pyx_t_15);
    PyTuple_SET_ITEM(__pyx_t_14, 1, __pyx_t_15);
    __pyx_t_13 = 0;
    __pyx_t_15 = 0;
    __pyx_r = __pyx_t_14;
    __pyx_t_14 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_AddTraceback("pandas._libs.interval.Float32ClosedNeitherIntervalNode.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Float32ClosedNeitherIntervalNode, (type(self), 0x7a104e5, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Float32ClosedNeitherIntervalNode__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_23__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_23__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_22__setstate_cython__(((struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_22__setstate_cython__(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_Float32ClosedNeitherIntervalNode, (type(self), 0x7a104e5, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Float32ClosedNeitherIntervalNode__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(2, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_6pandas_5_libs_8interval___pyx_unpickle_Float32ClosedNeitherIntervalNode__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Float32ClosedNeitherIntervalNode, (type(self), 0x7a104e5, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Float32ClosedNeitherIntervalNode__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Float32ClosedNeitherIntervalNode.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":928
 *         readonly bint is_leaf_node
 * 
 *     def __init__(self,             # <<<<<<<<<<<<<<
 *                  ndarray[float64_t, ndim=1] left,
 *                  ndarray[float64_t, ndim=1] right,
 */

/* Python wrapper */
static int __pyx_pw_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyArrayObject *__pyx_v_left = 0;
  PyArrayObject *__pyx_v_right = 0;
  PyArrayObject *__pyx_v_indices = 0;
  __pyx_t_5numpy_int64_t __pyx_v_leaf_size;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_left,&__pyx_n_s_right,&__pyx_n_s_indices,&__pyx_n_s_leaf_size,0};
    PyObject* values[4] = {0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_left)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_right)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 4, 4, 1); __PYX_ERR(1, 928, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_indices)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 4, 4, 2); __PYX_ERR(1, 928, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_leaf_size)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 4, 4, 3); __PYX_ERR(1, 928, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(1, 928, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 4) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
    }
    __pyx_v_left = ((PyArrayObject *)values[0]);
    __pyx_v_right = ((PyArrayObject *)values[1]);
    __pyx_v_indices = ((PyArrayObject *)values[2]);
    __pyx_v_leaf_size = __Pyx_PyInt_As_npy_int64(values[3]); if (unlikely((__pyx_v_leaf_size == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(1, 932, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 4, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 928, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Float64ClosedLeftIntervalNode.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_left), __pyx_ptype_5numpy_ndarray, 1, "left", 0))) __PYX_ERR(1, 929, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_right), __pyx_ptype_5numpy_ndarray, 1, "right", 0))) __PYX_ERR(1, 930, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_indices), __pyx_ptype_5numpy_ndarray, 1, "indices", 0))) __PYX_ERR(1, 931, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode___init__(((struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode *)__pyx_v_self), __pyx_v_left, __pyx_v_right, __pyx_v_indices, __pyx_v_leaf_size);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode___init__(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode *__pyx_v_self, PyArrayObject *__pyx_v_left, PyArrayObject *__pyx_v_right, PyArrayObject *__pyx_v_indices, __pyx_t_5numpy_int64_t __pyx_v_leaf_size) {
  PyObject *__pyx_v_left_set = NULL;
  PyObject *__pyx_v_right_set = NULL;
  PyObject *__pyx_v_center_set = NULL;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_indices;
  __Pyx_Buffer __pyx_pybuffer_indices;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_left;
  __Pyx_Buffer __pyx_pybuffer_left;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_right;
  __Pyx_Buffer __pyx_pybuffer_right;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __pyx_t_5numpy_float64_t __pyx_t_6;
  __Pyx_memviewslice __pyx_t_7 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_8 = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  __Pyx_memviewslice __pyx_t_11 = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *(*__pyx_t_12)(PyObject *);
  size_t __pyx_t_13;
  __Pyx_RefNannySetupContext("__init__", 0);
  __pyx_pybuffer_left.pybuffer.buf = NULL;
  __pyx_pybuffer_left.refcount = 0;
  __pyx_pybuffernd_left.data = NULL;
  __pyx_pybuffernd_left.rcbuffer = &__pyx_pybuffer_left;
  __pyx_pybuffer_right.pybuffer.buf = NULL;
  __pyx_pybuffer_right.refcount = 0;
  __pyx_pybuffernd_right.data = NULL;
  __pyx_pybuffernd_right.rcbuffer = &__pyx_pybuffer_right;
  __pyx_pybuffer_indices.pybuffer.buf = NULL;
  __pyx_pybuffer_indices.refcount = 0;
  __pyx_pybuffernd_indices.data = NULL;
  __pyx_pybuffernd_indices.rcbuffer = &__pyx_pybuffer_indices;
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_left.rcbuffer->pybuffer, (PyObject*)__pyx_v_left, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 928, __pyx_L1_error)
  }
  __pyx_pybuffernd_left.diminfo[0].strides = __pyx_pybuffernd_left.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_left.diminfo[0].shape = __pyx_pybuffernd_left.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_right.rcbuffer->pybuffer, (PyObject*)__pyx_v_right, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 928, __pyx_L1_error)
  }
  __pyx_pybuffernd_right.diminfo[0].strides = __pyx_pybuffernd_right.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_right.diminfo[0].shape = __pyx_pybuffernd_right.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_indices.rcbuffer->pybuffer, (PyObject*)__pyx_v_indices, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 928, __pyx_L1_error)
  }
  __pyx_pybuffernd_indices.diminfo[0].strides = __pyx_pybuffernd_indices.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_indices.diminfo[0].shape = __pyx_pybuffernd_indices.rcbuffer->pybuffer.shape[0];

  /* "pandas/_libs/intervaltree.pxi":934
 *                  int64_t leaf_size):
 * 
 *         self.n_elements = len(left)             # <<<<<<<<<<<<<<
 *         self.leaf_size = leaf_size
 * 
 */
  __pyx_t_1 = PyObject_Length(((PyObject *)__pyx_v_left)); if (unlikely(__pyx_t_1 == ((Py_ssize_t)-1))) __PYX_ERR(1, 934, __pyx_L1_error)
  __pyx_v_self->n_elements = __pyx_t_1;

  /* "pandas/_libs/intervaltree.pxi":935
 * 
 *         self.n_elements = len(left)
 *         self.leaf_size = leaf_size             # <<<<<<<<<<<<<<
 * 
 *         # min_left and min_right are used to speed-up query by skipping
 */
  __pyx_v_self->leaf_size = __pyx_v_leaf_size;

  /* "pandas/_libs/intervaltree.pxi":940
 *         # query on sub-nodes. If this node has size 0, query is cheap,
 *         # so these values don't matter.
 *         if left.size > 0:             # <<<<<<<<<<<<<<
 *             self.min_left = left.min()
 *             self.max_right = right.max()
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_left), __pyx_n_s_size); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 940, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyObject_RichCompare(__pyx_t_2, __pyx_int_0, Py_GT); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 940, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(1, 940, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_4) {

    /* "pandas/_libs/intervaltree.pxi":941
 *         # so these values don't matter.
 *         if left.size > 0:
 *             self.min_left = left.min()             # <<<<<<<<<<<<<<
 *             self.max_right = right.max()
 *         else:
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_left), __pyx_n_s_min); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 941, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 941, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_6 = __pyx_PyFloat_AsDouble(__pyx_t_3); if (unlikely((__pyx_t_6 == ((npy_float64)-1)) && PyErr_Occurred())) __PYX_ERR(1, 941, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_self->min_left = __pyx_t_6;

    /* "pandas/_libs/intervaltree.pxi":942
 *         if left.size > 0:
 *             self.min_left = left.min()
 *             self.max_right = right.max()             # <<<<<<<<<<<<<<
 *         else:
 *             self.min_left = 0
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_right), __pyx_n_s_max); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 942, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 942, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_6 = __pyx_PyFloat_AsDouble(__pyx_t_3); if (unlikely((__pyx_t_6 == ((npy_float64)-1)) && PyErr_Occurred())) __PYX_ERR(1, 942, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_self->max_right = __pyx_t_6;

    /* "pandas/_libs/intervaltree.pxi":940
 *         # query on sub-nodes. If this node has size 0, query is cheap,
 *         # so these values don't matter.
 *         if left.size > 0:             # <<<<<<<<<<<<<<
 *             self.min_left = left.min()
 *             self.max_right = right.max()
 */
    goto __pyx_L3;
  }

  /* "pandas/_libs/intervaltree.pxi":944
 *             self.max_right = right.max()
 *         else:
 *             self.min_left = 0             # <<<<<<<<<<<<<<
 *             self.max_right = 0
 * 
 */
  /*else*/ {
    __pyx_v_self->min_left = 0.0;

    /* "pandas/_libs/intervaltree.pxi":945
 *         else:
 *             self.min_left = 0
 *             self.max_right = 0             # <<<<<<<<<<<<<<
 * 
 *         if self.n_elements <= leaf_size:
 */
    __pyx_v_self->max_right = 0.0;
  }
  __pyx_L3:;

  /* "pandas/_libs/intervaltree.pxi":947
 *             self.max_right = 0
 * 
 *         if self.n_elements <= leaf_size:             # <<<<<<<<<<<<<<
 *             # make this a terminal (leaf) node
 *             self.is_leaf_node = True
 */
  __pyx_t_4 = ((__pyx_v_self->n_elements <= __pyx_v_leaf_size) != 0);
  if (__pyx_t_4) {

    /* "pandas/_libs/intervaltree.pxi":949
 *         if self.n_elements <= leaf_size:
 *             # make this a terminal (leaf) node
 *             self.is_leaf_node = True             # <<<<<<<<<<<<<<
 *             self.left = left
 *             self.right = right
 */
    __pyx_v_self->is_leaf_node = 1;

    /* "pandas/_libs/intervaltree.pxi":950
 *             # make this a terminal (leaf) node
 *             self.is_leaf_node = True
 *             self.left = left             # <<<<<<<<<<<<<<
 *             self.right = right
 *             self.indices = indices
 */
    __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_float64_t(((PyObject *)__pyx_v_left), PyBUF_WRITABLE); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(1, 950, __pyx_L1_error)
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->left, 0);
    __pyx_v_self->left = __pyx_t_7;
    __pyx_t_7.memview = NULL;
    __pyx_t_7.data = NULL;

    /* "pandas/_libs/intervaltree.pxi":951
 *             self.is_leaf_node = True
 *             self.left = left
 *             self.right = right             # <<<<<<<<<<<<<<
 *             self.indices = indices
 *             self.n_center = 0
 */
    __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_float64_t(((PyObject *)__pyx_v_right), PyBUF_WRITABLE); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(1, 951, __pyx_L1_error)
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->right, 0);
    __pyx_v_self->right = __pyx_t_7;
    __pyx_t_7.memview = NULL;
    __pyx_t_7.data = NULL;

    /* "pandas/_libs/intervaltree.pxi":952
 *             self.left = left
 *             self.right = right
 *             self.indices = indices             # <<<<<<<<<<<<<<
 *             self.n_center = 0
 *         else:
 */
    __pyx_t_8 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(((PyObject *)__pyx_v_indices), PyBUF_WRITABLE); if (unlikely(!__pyx_t_8.memview)) __PYX_ERR(1, 952, __pyx_L1_error)
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->indices, 0);
    __pyx_v_self->indices = __pyx_t_8;
    __pyx_t_8.memview = NULL;
    __pyx_t_8.data = NULL;

    /* "pandas/_libs/intervaltree.pxi":953
 *             self.right = right
 *             self.indices = indices
 *             self.n_center = 0             # <<<<<<<<<<<<<<
 *         else:
 *             # calculate a pivot so we can create child nodes
 */
    __pyx_v_self->n_center = 0;

    /* "pandas/_libs/intervaltree.pxi":947
 *             self.max_right = 0
 * 
 *         if self.n_elements <= leaf_size:             # <<<<<<<<<<<<<<
 *             # make this a terminal (leaf) node
 *             self.is_leaf_node = True
 */
    goto __pyx_L4;
  }

  /* "pandas/_libs/intervaltree.pxi":956
 *         else:
 *             # calculate a pivot so we can create child nodes
 *             self.is_leaf_node = False             # <<<<<<<<<<<<<<
 *             self.pivot = np.median(left / 2 + right / 2)
 *             left_set, right_set, center_set = self.classify_intervals(
 */
  /*else*/ {
    __pyx_v_self->is_leaf_node = 0;

    /* "pandas/_libs/intervaltree.pxi":957
 *             # calculate a pivot so we can create child nodes
 *             self.is_leaf_node = False
 *             self.pivot = np.median(left / 2 + right / 2)             # <<<<<<<<<<<<<<
 *             left_set, right_set, center_set = self.classify_intervals(
 *                 left, right)
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 957, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_median); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 957, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyNumber_Divide(((PyObject *)__pyx_v_left), __pyx_int_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 957, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_9 = __Pyx_PyNumber_Divide(((PyObject *)__pyx_v_right), __pyx_int_2); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 957, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_10 = PyNumber_Add(__pyx_t_2, __pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 957, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_9 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_3 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_9, __pyx_t_10) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_10);
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 957, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_6 = __pyx_PyFloat_AsDouble(__pyx_t_3); if (unlikely((__pyx_t_6 == ((npy_float64)-1)) && PyErr_Occurred())) __PYX_ERR(1, 957, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_self->pivot = __pyx_t_6;

    /* "pandas/_libs/intervaltree.pxi":959
 *             self.pivot = np.median(left / 2 + right / 2)
 *             left_set, right_set, center_set = self.classify_intervals(
 *                 left, right)             # <<<<<<<<<<<<<<
 * 
 *             self.left_node = self.new_child_node(left, right,
 */
    __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_float64_t(((PyObject *)__pyx_v_left), PyBUF_WRITABLE); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(1, 959, __pyx_L1_error)
    __pyx_t_11 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_float64_t(((PyObject *)__pyx_v_right), PyBUF_WRITABLE); if (unlikely(!__pyx_t_11.memview)) __PYX_ERR(1, 959, __pyx_L1_error)

    /* "pandas/_libs/intervaltree.pxi":958
 *             self.is_leaf_node = False
 *             self.pivot = np.median(left / 2 + right / 2)
 *             left_set, right_set, center_set = self.classify_intervals(             # <<<<<<<<<<<<<<
 *                 left, right)
 * 
 */
    __pyx_t_3 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode *)__pyx_v_self->__pyx_vtab)->classify_intervals(__pyx_v_self, __pyx_t_7, __pyx_t_11); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 958, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __PYX_XDEC_MEMVIEW(&__pyx_t_7, 1);
    __pyx_t_7.memview = NULL;
    __pyx_t_7.data = NULL;
    __PYX_XDEC_MEMVIEW(&__pyx_t_11, 1);
    __pyx_t_11.memview = NULL;
    __pyx_t_11.data = NULL;
    if ((likely(PyTuple_CheckExact(__pyx_t_3))) || (PyList_CheckExact(__pyx_t_3))) {
      PyObject* sequence = __pyx_t_3;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 3)) {
        if (size > 3) __Pyx_RaiseTooManyValuesError(3);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(1, 958, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_5 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_10 = PyTuple_GET_ITEM(sequence, 1); 
        __pyx_t_9 = PyTuple_GET_ITEM(sequence, 2); 
      } else {
        __pyx_t_5 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_10 = PyList_GET_ITEM(sequence, 1); 
        __pyx_t_9 = PyList_GET_ITEM(sequence, 2); 
      }
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_10);
      __Pyx_INCREF(__pyx_t_9);
      #else
      __pyx_t_5 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 958, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_10 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 958, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_9 = PySequence_ITEM(sequence, 2); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 958, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      #endif
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_2 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 958, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_12 = Py_TYPE(__pyx_t_2)->tp_iternext;
      index = 0; __pyx_t_5 = __pyx_t_12(__pyx_t_2); if (unlikely(!__pyx_t_5)) goto __pyx_L5_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_5);
      index = 1; __pyx_t_10 = __pyx_t_12(__pyx_t_2); if (unlikely(!__pyx_t_10)) goto __pyx_L5_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_10);
      index = 2; __pyx_t_9 = __pyx_t_12(__pyx_t_2); if (unlikely(!__pyx_t_9)) goto __pyx_L5_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_9);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_12(__pyx_t_2), 3) < 0) __PYX_ERR(1, 958, __pyx_L1_error)
      __pyx_t_12 = NULL;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      goto __pyx_L6_unpacking_done;
      __pyx_L5_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_12 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(1, 958, __pyx_L1_error)
      __pyx_L6_unpacking_done:;
    }
    __pyx_v_left_set = __pyx_t_5;
    __pyx_t_5 = 0;
    __pyx_v_right_set = __pyx_t_10;
    __pyx_t_10 = 0;
    __pyx_v_center_set = __pyx_t_9;
    __pyx_t_9 = 0;

    /* "pandas/_libs/intervaltree.pxi":962
 * 
 *             self.left_node = self.new_child_node(left, right,
 *                                                  indices, left_set)             # <<<<<<<<<<<<<<
 *             self.right_node = self.new_child_node(left, right,
 *                                                   indices, right_set)
 */
    if (!(likely(((__pyx_v_left_set) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_left_set, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(1, 962, __pyx_L1_error)

    /* "pandas/_libs/intervaltree.pxi":961
 *                 left, right)
 * 
 *             self.left_node = self.new_child_node(left, right,             # <<<<<<<<<<<<<<
 *                                                  indices, left_set)
 *             self.right_node = self.new_child_node(left, right,
 */
    __pyx_t_3 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode *)__pyx_v_self->__pyx_vtab)->new_child_node(__pyx_v_self, ((PyArrayObject *)__pyx_v_left), ((PyArrayObject *)__pyx_v_right), ((PyArrayObject *)__pyx_v_indices), ((PyArrayObject *)__pyx_v_left_set)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 961, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode))))) __PYX_ERR(1, 961, __pyx_L1_error)
    __Pyx_GIVEREF(__pyx_t_3);
    __Pyx_GOTREF(__pyx_v_self->left_node);
    __Pyx_DECREF(((PyObject *)__pyx_v_self->left_node));
    __pyx_v_self->left_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode *)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "pandas/_libs/intervaltree.pxi":964
 *                                                  indices, left_set)
 *             self.right_node = self.new_child_node(left, right,
 *                                                   indices, right_set)             # <<<<<<<<<<<<<<
 * 
 *             self.center_left_values, self.center_left_indices = \
 */
    if (!(likely(((__pyx_v_right_set) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_right_set, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(1, 964, __pyx_L1_error)

    /* "pandas/_libs/intervaltree.pxi":963
 *             self.left_node = self.new_child_node(left, right,
 *                                                  indices, left_set)
 *             self.right_node = self.new_child_node(left, right,             # <<<<<<<<<<<<<<
 *                                                   indices, right_set)
 * 
 */
    __pyx_t_3 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode *)__pyx_v_self->__pyx_vtab)->new_child_node(__pyx_v_self, ((PyArrayObject *)__pyx_v_left), ((PyArrayObject *)__pyx_v_right), ((PyArrayObject *)__pyx_v_indices), ((PyArrayObject *)__pyx_v_right_set)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 963, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode))))) __PYX_ERR(1, 963, __pyx_L1_error)
    __Pyx_GIVEREF(__pyx_t_3);
    __Pyx_GOTREF(__pyx_v_self->right_node);
    __Pyx_DECREF(((PyObject *)__pyx_v_self->right_node));
    __pyx_v_self->right_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode *)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "pandas/_libs/intervaltree.pxi":967
 * 
 *             self.center_left_values, self.center_left_indices = \
 *                 sort_values_and_indices(left, indices, center_set)             # <<<<<<<<<<<<<<
 *             self.center_right_values, self.center_right_indices = \
 *                 sort_values_and_indices(right, indices, center_set)
 */
    __pyx_t_3 = __pyx_f_6pandas_5_libs_8interval_sort_values_and_indices(((PyObject *)__pyx_v_left), ((PyObject *)__pyx_v_indices), __pyx_v_center_set); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 967, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if ((likely(PyTuple_CheckExact(__pyx_t_3))) || (PyList_CheckExact(__pyx_t_3))) {
      PyObject* sequence = __pyx_t_3;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(1, 966, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_9 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_10 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_9 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_10 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_9);
      __Pyx_INCREF(__pyx_t_10);
      #else
      __pyx_t_9 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 966, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_10 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 966, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      #endif
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_5 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 966, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_12 = Py_TYPE(__pyx_t_5)->tp_iternext;
      index = 0; __pyx_t_9 = __pyx_t_12(__pyx_t_5); if (unlikely(!__pyx_t_9)) goto __pyx_L7_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_9);
      index = 1; __pyx_t_10 = __pyx_t_12(__pyx_t_5); if (unlikely(!__pyx_t_10)) goto __pyx_L7_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_10);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_12(__pyx_t_5), 2) < 0) __PYX_ERR(1, 966, __pyx_L1_error)
      __pyx_t_12 = NULL;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      goto __pyx_L8_unpacking_done;
      __pyx_L7_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_12 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(1, 966, __pyx_L1_error)
      __pyx_L8_unpacking_done:;
    }

    /* "pandas/_libs/intervaltree.pxi":966
 *                                                   indices, right_set)
 * 
 *             self.center_left_values, self.center_left_indices = \             # <<<<<<<<<<<<<<
 *                 sort_values_and_indices(left, indices, center_set)
 *             self.center_right_values, self.center_right_indices = \
 */
    __pyx_t_11 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_float64_t(__pyx_t_9, PyBUF_WRITABLE); if (unlikely(!__pyx_t_11.memview)) __PYX_ERR(1, 966, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_8 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_t_10, PyBUF_WRITABLE); if (unlikely(!__pyx_t_8.memview)) __PYX_ERR(1, 966, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->center_left_values, 0);
    __pyx_v_self->center_left_values = __pyx_t_11;
    __pyx_t_11.memview = NULL;
    __pyx_t_11.data = NULL;
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->center_left_indices, 0);
    __pyx_v_self->center_left_indices = __pyx_t_8;
    __pyx_t_8.memview = NULL;
    __pyx_t_8.data = NULL;

    /* "pandas/_libs/intervaltree.pxi":969
 *                 sort_values_and_indices(left, indices, center_set)
 *             self.center_right_values, self.center_right_indices = \
 *                 sort_values_and_indices(right, indices, center_set)             # <<<<<<<<<<<<<<
 *             self.n_center = len(self.center_left_indices)
 * 
 */
    __pyx_t_3 = __pyx_f_6pandas_5_libs_8interval_sort_values_and_indices(((PyObject *)__pyx_v_right), ((PyObject *)__pyx_v_indices), __pyx_v_center_set); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 969, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if ((likely(PyTuple_CheckExact(__pyx_t_3))) || (PyList_CheckExact(__pyx_t_3))) {
      PyObject* sequence = __pyx_t_3;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(1, 968, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_10 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_9 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_10 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_9 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_10);
      __Pyx_INCREF(__pyx_t_9);
      #else
      __pyx_t_10 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 968, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_9 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 968, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      #endif
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_5 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 968, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_12 = Py_TYPE(__pyx_t_5)->tp_iternext;
      index = 0; __pyx_t_10 = __pyx_t_12(__pyx_t_5); if (unlikely(!__pyx_t_10)) goto __pyx_L9_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_10);
      index = 1; __pyx_t_9 = __pyx_t_12(__pyx_t_5); if (unlikely(!__pyx_t_9)) goto __pyx_L9_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_9);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_12(__pyx_t_5), 2) < 0) __PYX_ERR(1, 968, __pyx_L1_error)
      __pyx_t_12 = NULL;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      goto __pyx_L10_unpacking_done;
      __pyx_L9_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_12 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(1, 968, __pyx_L1_error)
      __pyx_L10_unpacking_done:;
    }

    /* "pandas/_libs/intervaltree.pxi":968
 *             self.center_left_values, self.center_left_indices = \
 *                 sort_values_and_indices(left, indices, center_set)
 *             self.center_right_values, self.center_right_indices = \             # <<<<<<<<<<<<<<
 *                 sort_values_and_indices(right, indices, center_set)
 *             self.n_center = len(self.center_left_indices)
 */
    __pyx_t_11 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_float64_t(__pyx_t_10, PyBUF_WRITABLE); if (unlikely(!__pyx_t_11.memview)) __PYX_ERR(1, 968, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_t_8 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_t_9, PyBUF_WRITABLE); if (unlikely(!__pyx_t_8.memview)) __PYX_ERR(1, 968, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->center_right_values, 0);
    __pyx_v_self->center_right_values = __pyx_t_11;
    __pyx_t_11.memview = NULL;
    __pyx_t_11.data = NULL;
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->center_right_indices, 0);
    __pyx_v_self->center_right_indices = __pyx_t_8;
    __pyx_t_8.memview = NULL;
    __pyx_t_8.data = NULL;

    /* "pandas/_libs/intervaltree.pxi":970
 *             self.center_right_values, self.center_right_indices = \
 *                 sort_values_and_indices(right, indices, center_set)
 *             self.n_center = len(self.center_left_indices)             # <<<<<<<<<<<<<<
 * 
 *     @cython.wraparound(False)
 */
    if (unlikely(!__pyx_v_self->center_left_indices.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(1, 970, __pyx_L1_error)}
    __pyx_t_13 = __Pyx_MemoryView_Len(__pyx_v_self->center_left_indices); 
    __pyx_v_self->n_center = __pyx_t_13;
  }
  __pyx_L4:;

  /* "pandas/_libs/intervaltree.pxi":928
 *         readonly bint is_leaf_node
 * 
 *     def __init__(self,             # <<<<<<<<<<<<<<
 *                  ndarray[float64_t, ndim=1] left,
 *                  ndarray[float64_t, ndim=1] right,
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __PYX_XDEC_MEMVIEW(&__pyx_t_7, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_8, 1);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __PYX_XDEC_MEMVIEW(&__pyx_t_11, 1);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indices.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_left.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_right.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("pandas._libs.interval.Float64ClosedLeftIntervalNode.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indices.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_left.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_right.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF(__pyx_v_left_set);
  __Pyx_XDECREF(__pyx_v_right_set);
  __Pyx_XDECREF(__pyx_v_center_set);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":974
 *     @cython.wraparound(False)
 *     @cython.boundscheck(False)
 *     cdef classify_intervals(self, float64_t[:] left, float64_t[:] right):             # <<<<<<<<<<<<<<
 *         """Classify the given intervals based upon whether they fall to the
 *         left, right, or overlap with this node's pivot.
 */

static PyObject *__pyx_f_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_classify_intervals(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode *__pyx_v_self, __Pyx_memviewslice __pyx_v_left, __Pyx_memviewslice __pyx_v_right) {
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_left_ind = 0;
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_right_ind = 0;
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_overlapping_ind = 0;
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __pyx_t_5numpy_int64_t __pyx_t_2;
  __pyx_t_5numpy_int64_t __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  int __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  __Pyx_RefNannySetupContext("classify_intervals", 0);

  /* "pandas/_libs/intervaltree.pxi":982
 *             Py_ssize_t i
 * 
 *         left_ind = Int64Vector()             # <<<<<<<<<<<<<<
 *         right_ind = Int64Vector()
 *         overlapping_ind = Int64Vector()
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 982, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_left_ind = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":983
 * 
 *         left_ind = Int64Vector()
 *         right_ind = Int64Vector()             # <<<<<<<<<<<<<<
 *         overlapping_ind = Int64Vector()
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 983, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_right_ind = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":984
 *         left_ind = Int64Vector()
 *         right_ind = Int64Vector()
 *         overlapping_ind = Int64Vector()             # <<<<<<<<<<<<<<
 * 
 *         for i in range(self.n_elements):
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 984, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_overlapping_ind = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":986
 *         overlapping_ind = Int64Vector()
 * 
 *         for i in range(self.n_elements):             # <<<<<<<<<<<<<<
 *             if right[i] <= self.pivot:
 *                 left_ind.append(i)
 */
  __pyx_t_2 = __pyx_v_self->n_elements;
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "pandas/_libs/intervaltree.pxi":987
 * 
 *         for i in range(self.n_elements):
 *             if right[i] <= self.pivot:             # <<<<<<<<<<<<<<
 *                 left_ind.append(i)
 *             elif self.pivot < left[i]:
 */
    __pyx_t_5 = __pyx_v_i;
    __pyx_t_6 = (((*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_right.data + __pyx_t_5 * __pyx_v_right.strides[0]) ))) <= __pyx_v_self->pivot) != 0);
    if (__pyx_t_6) {

      /* "pandas/_libs/intervaltree.pxi":988
 *         for i in range(self.n_elements):
 *             if right[i] <= self.pivot:
 *                 left_ind.append(i)             # <<<<<<<<<<<<<<
 *             elif self.pivot < left[i]:
 *                 right_ind.append(i)
 */
      ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_left_ind->__pyx_vtab)->append(__pyx_v_left_ind, __pyx_v_i);

      /* "pandas/_libs/intervaltree.pxi":987
 * 
 *         for i in range(self.n_elements):
 *             if right[i] <= self.pivot:             # <<<<<<<<<<<<<<
 *                 left_ind.append(i)
 *             elif self.pivot < left[i]:
 */
      goto __pyx_L5;
    }

    /* "pandas/_libs/intervaltree.pxi":989
 *             if right[i] <= self.pivot:
 *                 left_ind.append(i)
 *             elif self.pivot < left[i]:             # <<<<<<<<<<<<<<
 *                 right_ind.append(i)
 *             else:
 */
    __pyx_t_7 = __pyx_v_i;
    __pyx_t_6 = ((__pyx_v_self->pivot < (*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_left.data + __pyx_t_7 * __pyx_v_left.strides[0]) )))) != 0);
    if (__pyx_t_6) {

      /* "pandas/_libs/intervaltree.pxi":990
 *                 left_ind.append(i)
 *             elif self.pivot < left[i]:
 *                 right_ind.append(i)             # <<<<<<<<<<<<<<
 *             else:
 *                 overlapping_ind.append(i)
 */
      ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_right_ind->__pyx_vtab)->append(__pyx_v_right_ind, __pyx_v_i);

      /* "pandas/_libs/intervaltree.pxi":989
 *             if right[i] <= self.pivot:
 *                 left_ind.append(i)
 *             elif self.pivot < left[i]:             # <<<<<<<<<<<<<<
 *                 right_ind.append(i)
 *             else:
 */
      goto __pyx_L5;
    }

    /* "pandas/_libs/intervaltree.pxi":992
 *                 right_ind.append(i)
 *             else:
 *                 overlapping_ind.append(i)             # <<<<<<<<<<<<<<
 * 
 *         return (left_ind.to_array(),
 */
    /*else*/ {
      ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_overlapping_ind->__pyx_vtab)->append(__pyx_v_overlapping_ind, __pyx_v_i);
    }
    __pyx_L5:;
  }

  /* "pandas/_libs/intervaltree.pxi":994
 *                 overlapping_ind.append(i)
 * 
 *         return (left_ind.to_array(),             # <<<<<<<<<<<<<<
 *                 right_ind.to_array(),
 *                 overlapping_ind.to_array())
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_left_ind->__pyx_vtab)->to_array(__pyx_v_left_ind, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 994, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "pandas/_libs/intervaltree.pxi":995
 * 
 *         return (left_ind.to_array(),
 *                 right_ind.to_array(),             # <<<<<<<<<<<<<<
 *                 overlapping_ind.to_array())
 * 
 */
  __pyx_t_8 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_right_ind->__pyx_vtab)->to_array(__pyx_v_right_ind, 0); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 995, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);

  /* "pandas/_libs/intervaltree.pxi":996
 *         return (left_ind.to_array(),
 *                 right_ind.to_array(),
 *                 overlapping_ind.to_array())             # <<<<<<<<<<<<<<
 * 
 *     cdef new_child_node(self,
 */
  __pyx_t_9 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_overlapping_ind->__pyx_vtab)->to_array(__pyx_v_overlapping_ind, 0); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 996, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);

  /* "pandas/_libs/intervaltree.pxi":994
 *                 overlapping_ind.append(i)
 * 
 *         return (left_ind.to_array(),             # <<<<<<<<<<<<<<
 *                 right_ind.to_array(),
 *                 overlapping_ind.to_array())
 */
  __pyx_t_10 = PyTuple_New(3); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 994, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_t_8);
  __Pyx_GIVEREF(__pyx_t_9);
  PyTuple_SET_ITEM(__pyx_t_10, 2, __pyx_t_9);
  __pyx_t_1 = 0;
  __pyx_t_8 = 0;
  __pyx_t_9 = 0;
  __pyx_r = __pyx_t_10;
  __pyx_t_10 = 0;
  goto __pyx_L0;

  /* "pandas/_libs/intervaltree.pxi":974
 *     @cython.wraparound(False)
 *     @cython.boundscheck(False)
 *     cdef classify_intervals(self, float64_t[:] left, float64_t[:] right):             # <<<<<<<<<<<<<<
 *         """Classify the given intervals based upon whether they fall to the
 *         left, right, or overlap with this node's pivot.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("pandas._libs.interval.Float64ClosedLeftIntervalNode.classify_intervals", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_left_ind);
  __Pyx_XDECREF((PyObject *)__pyx_v_right_ind);
  __Pyx_XDECREF((PyObject *)__pyx_v_overlapping_ind);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":998
 *                 overlapping_ind.to_array())
 * 
 *     cdef new_child_node(self,             # <<<<<<<<<<<<<<
 *                         ndarray[float64_t, ndim=1] left,
 *                         ndarray[float64_t, ndim=1] right,
 */

static PyObject *__pyx_f_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_new_child_node(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode *__pyx_v_self, PyArrayObject *__pyx_v_left, PyArrayObject *__pyx_v_right, PyArrayObject *__pyx_v_indices, PyArrayObject *__pyx_v_subset) {
  __Pyx_LocalBuf_ND __pyx_pybuffernd_indices;
  __Pyx_Buffer __pyx_pybuffer_indices;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_left;
  __Pyx_Buffer __pyx_pybuffer_left;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_right;
  __Pyx_Buffer __pyx_pybuffer_right;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_subset;
  __Pyx_Buffer __pyx_pybuffer_subset;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyArrayObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyArrayObject *__pyx_t_7 = NULL;
  PyArrayObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  __Pyx_RefNannySetupContext("new_child_node", 0);
  __Pyx_INCREF((PyObject *)__pyx_v_left);
  __Pyx_INCREF((PyObject *)__pyx_v_right);
  __Pyx_INCREF((PyObject *)__pyx_v_indices);
  __pyx_pybuffer_left.pybuffer.buf = NULL;
  __pyx_pybuffer_left.refcount = 0;
  __pyx_pybuffernd_left.data = NULL;
  __pyx_pybuffernd_left.rcbuffer = &__pyx_pybuffer_left;
  __pyx_pybuffer_right.pybuffer.buf = NULL;
  __pyx_pybuffer_right.refcount = 0;
  __pyx_pybuffernd_right.data = NULL;
  __pyx_pybuffernd_right.rcbuffer = &__pyx_pybuffer_right;
  __pyx_pybuffer_indices.pybuffer.buf = NULL;
  __pyx_pybuffer_indices.refcount = 0;
  __pyx_pybuffernd_indices.data = NULL;
  __pyx_pybuffernd_indices.rcbuffer = &__pyx_pybuffer_indices;
  __pyx_pybuffer_subset.pybuffer.buf = NULL;
  __pyx_pybuffer_subset.refcount = 0;
  __pyx_pybuffernd_subset.data = NULL;
  __pyx_pybuffernd_subset.rcbuffer = &__pyx_pybuffer_subset;
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_left.rcbuffer->pybuffer, (PyObject*)__pyx_v_left, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 998, __pyx_L1_error)
  }
  __pyx_pybuffernd_left.diminfo[0].strides = __pyx_pybuffernd_left.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_left.diminfo[0].shape = __pyx_pybuffernd_left.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_right.rcbuffer->pybuffer, (PyObject*)__pyx_v_right, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 998, __pyx_L1_error)
  }
  __pyx_pybuffernd_right.diminfo[0].strides = __pyx_pybuffernd_right.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_right.diminfo[0].shape = __pyx_pybuffernd_right.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_indices.rcbuffer->pybuffer, (PyObject*)__pyx_v_indices, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 998, __pyx_L1_error)
  }
  __pyx_pybuffernd_indices.diminfo[0].strides = __pyx_pybuffernd_indices.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_indices.diminfo[0].shape = __pyx_pybuffernd_indices.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_subset.rcbuffer->pybuffer, (PyObject*)__pyx_v_subset, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 998, __pyx_L1_error)
  }
  __pyx_pybuffernd_subset.diminfo[0].strides = __pyx_pybuffernd_subset.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_subset.diminfo[0].shape = __pyx_pybuffernd_subset.rcbuffer->pybuffer.shape[0];

  /* "pandas/_libs/intervaltree.pxi":1005
 *         """Create a new child node.
 *         """
 *         left = take(left, subset)             # <<<<<<<<<<<<<<
 *         right = take(right, subset)
 *         indices = take(indices, subset)
 */
  __pyx_t_1 = __pyx_f_6pandas_5_libs_8interval_take(((PyArrayObject *)__pyx_v_left), ((PyArrayObject *)__pyx_v_subset)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1005, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(1, 1005, __pyx_L1_error)
  __pyx_t_2 = ((PyArrayObject *)__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_left.rcbuffer->pybuffer);
    __pyx_t_3 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_left.rcbuffer->pybuffer, (PyObject*)__pyx_t_2, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack);
    if (unlikely(__pyx_t_3 < 0)) {
      PyErr_Fetch(&__pyx_t_4, &__pyx_t_5, &__pyx_t_6);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_left.rcbuffer->pybuffer, (PyObject*)__pyx_v_left, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_4); Py_XDECREF(__pyx_t_5); Py_XDECREF(__pyx_t_6);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_4, __pyx_t_5, __pyx_t_6);
      }
      __pyx_t_4 = __pyx_t_5 = __pyx_t_6 = 0;
    }
    __pyx_pybuffernd_left.diminfo[0].strides = __pyx_pybuffernd_left.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_left.diminfo[0].shape = __pyx_pybuffernd_left.rcbuffer->pybuffer.shape[0];
    if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(1, 1005, __pyx_L1_error)
  }
  __pyx_t_2 = 0;
  __Pyx_DECREF_SET(__pyx_v_left, ((PyArrayObject *)__pyx_t_1));
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":1006
 *         """
 *         left = take(left, subset)
 *         right = take(right, subset)             # <<<<<<<<<<<<<<
 *         indices = take(indices, subset)
 *         return Float64ClosedLeftIntervalNode(
 */
  __pyx_t_1 = __pyx_f_6pandas_5_libs_8interval_take(((PyArrayObject *)__pyx_v_right), ((PyArrayObject *)__pyx_v_subset)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1006, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(1, 1006, __pyx_L1_error)
  __pyx_t_7 = ((PyArrayObject *)__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_right.rcbuffer->pybuffer);
    __pyx_t_3 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_right.rcbuffer->pybuffer, (PyObject*)__pyx_t_7, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack);
    if (unlikely(__pyx_t_3 < 0)) {
      PyErr_Fetch(&__pyx_t_6, &__pyx_t_5, &__pyx_t_4);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_right.rcbuffer->pybuffer, (PyObject*)__pyx_v_right, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_6); Py_XDECREF(__pyx_t_5); Py_XDECREF(__pyx_t_4);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_6, __pyx_t_5, __pyx_t_4);
      }
      __pyx_t_6 = __pyx_t_5 = __pyx_t_4 = 0;
    }
    __pyx_pybuffernd_right.diminfo[0].strides = __pyx_pybuffernd_right.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_right.diminfo[0].shape = __pyx_pybuffernd_right.rcbuffer->pybuffer.shape[0];
    if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(1, 1006, __pyx_L1_error)
  }
  __pyx_t_7 = 0;
  __Pyx_DECREF_SET(__pyx_v_right, ((PyArrayObject *)__pyx_t_1));
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":1007
 *         left = take(left, subset)
 *         right = take(right, subset)
 *         indices = take(indices, subset)             # <<<<<<<<<<<<<<
 *         return Float64ClosedLeftIntervalNode(
 *             left, right, indices, self.leaf_size)
 */
  __pyx_t_1 = __pyx_f_6pandas_5_libs_8interval_take(((PyArrayObject *)__pyx_v_indices), ((PyArrayObject *)__pyx_v_subset)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1007, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(1, 1007, __pyx_L1_error)
  __pyx_t_8 = ((PyArrayObject *)__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indices.rcbuffer->pybuffer);
    __pyx_t_3 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_indices.rcbuffer->pybuffer, (PyObject*)__pyx_t_8, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack);
    if (unlikely(__pyx_t_3 < 0)) {
      PyErr_Fetch(&__pyx_t_4, &__pyx_t_5, &__pyx_t_6);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_indices.rcbuffer->pybuffer, (PyObject*)__pyx_v_indices, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_4); Py_XDECREF(__pyx_t_5); Py_XDECREF(__pyx_t_6);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_4, __pyx_t_5, __pyx_t_6);
      }
      __pyx_t_4 = __pyx_t_5 = __pyx_t_6 = 0;
    }
    __pyx_pybuffernd_indices.diminfo[0].strides = __pyx_pybuffernd_indices.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_indices.diminfo[0].shape = __pyx_pybuffernd_indices.rcbuffer->pybuffer.shape[0];
    if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(1, 1007, __pyx_L1_error)
  }
  __pyx_t_8 = 0;
  __Pyx_DECREF_SET(__pyx_v_indices, ((PyArrayObject *)__pyx_t_1));
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":1008
 *         right = take(right, subset)
 *         indices = take(indices, subset)
 *         return Float64ClosedLeftIntervalNode(             # <<<<<<<<<<<<<<
 *             left, right, indices, self.leaf_size)
 * 
 */
  __Pyx_XDECREF(__pyx_r);

  /* "pandas/_libs/intervaltree.pxi":1009
 *         indices = take(indices, subset)
 *         return Float64ClosedLeftIntervalNode(
 *             left, right, indices, self.leaf_size)             # <<<<<<<<<<<<<<
 * 
 *     @cython.wraparound(False)
 */
  __pyx_t_1 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->leaf_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1009, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "pandas/_libs/intervaltree.pxi":1008
 *         right = take(right, subset)
 *         indices = take(indices, subset)
 *         return Float64ClosedLeftIntervalNode(             # <<<<<<<<<<<<<<
 *             left, right, indices, self.leaf_size)
 * 
 */
  __pyx_t_9 = PyTuple_New(4); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 1008, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_INCREF(((PyObject *)__pyx_v_left));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_left));
  PyTuple_SET_ITEM(__pyx_t_9, 0, ((PyObject *)__pyx_v_left));
  __Pyx_INCREF(((PyObject *)__pyx_v_right));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_right));
  PyTuple_SET_ITEM(__pyx_t_9, 1, ((PyObject *)__pyx_v_right));
  __Pyx_INCREF(((PyObject *)__pyx_v_indices));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_indices));
  PyTuple_SET_ITEM(__pyx_t_9, 2, ((PyObject *)__pyx_v_indices));
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_9, 3, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode), __pyx_t_9, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1008, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pandas/_libs/intervaltree.pxi":998
 *                 overlapping_ind.to_array())
 * 
 *     cdef new_child_node(self,             # <<<<<<<<<<<<<<
 *                         ndarray[float64_t, ndim=1] left,
 *                         ndarray[float64_t, ndim=1] right,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_9);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indices.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_left.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_right.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_subset.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("pandas._libs.interval.Float64ClosedLeftIntervalNode.new_child_node", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indices.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_left.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_right.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_subset.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF((PyObject *)__pyx_v_left);
  __Pyx_XDECREF((PyObject *)__pyx_v_right);
  __Pyx_XDECREF((PyObject *)__pyx_v_indices);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":1014
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_2query[] = "Recursively query this node and its sub-nodes for intervals that\n        overlap with the query point.\n        ";
static PyMethodDef __pyx_mdef_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_3query = {"query", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_3query, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_2query};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_args = 0;
  PyObject *__pyx_v_kwargs = 0;
  CYTHON_UNUSED PyObject *__pyx_v_defaults = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fused_cpdef (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_args,&__pyx_n_s_kwargs,&__pyx_n_s_defaults,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_args)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_kwargs)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fused_cpdef", 1, 3, 3, 1); __PYX_ERR(1, 1014, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_defaults)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fused_cpdef", 1, 3, 3, 2); __PYX_ERR(1, 1014, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fused_cpdef") < 0)) __PYX_ERR(1, 1014, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_args = values[0];
    __pyx_v_kwargs = values[1];
    __pyx_v_defaults = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fused_cpdef", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 1014, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Float64ClosedLeftIntervalNode.__pyx_fused_cpdef", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_2query(((struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode *)__pyx_v_signatures), __pyx_v_args, __pyx_v_kwargs, __pyx_v_defaults);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_2query(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode *__pyx_v_signatures, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs, CYTHON_UNUSED PyObject *__pyx_v_defaults) {
  PyObject *__pyx_v_dest_sig = NULL;
  Py_ssize_t __pyx_v_i;
  CYTHON_UNUSED PyTypeObject *__pyx_v_ndarray = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_v_itemsize;
  PyObject *__pyx_v_arg = NULL;
  PyObject *__pyx_v_candidates = NULL;
  PyObject *__pyx_v_sig = NULL;
  int __pyx_v_match_found;
  PyObject *__pyx_v_src_sig = NULL;
  PyObject *__pyx_v_dst_type = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  Py_ssize_t __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  Py_ssize_t __pyx_t_13;
  Py_ssize_t __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  int __pyx_t_16;
  __Pyx_RefNannySetupContext("query", 0);
  __Pyx_INCREF(__pyx_v_kwargs);
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1014, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  PyList_SET_ITEM(__pyx_t_1, 0, Py_None);
  __pyx_v_dest_sig = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_3 = (__pyx_v_kwargs != Py_None);
  __pyx_t_4 = (__pyx_t_3 != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_kwargs); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(1, 1014, __pyx_L1_error)
  __pyx_t_3 = ((!__pyx_t_4) != 0);
  __pyx_t_2 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {
    __Pyx_INCREF(Py_None);
    __Pyx_DECREF_SET(__pyx_v_kwargs, Py_None);
  }
  __pyx_t_1 = ((PyObject *)__Pyx_ImportNumPyArrayTypeIfAvailable()); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1014, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_ndarray = ((PyTypeObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_v_itemsize = -1L;
  if (unlikely(__pyx_v_args == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 1014, __pyx_L1_error)
  }
  __pyx_t_5 = PyTuple_GET_SIZE(((PyObject*)__pyx_v_args)); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(1, 1014, __pyx_L1_error)
  __pyx_t_2 = ((2 < __pyx_t_5) != 0);
  if (__pyx_t_2) {
    if (unlikely(__pyx_v_args == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 1014, __pyx_L1_error)
    }
    __pyx_t_1 = PyTuple_GET_ITEM(((PyObject*)__pyx_v_args), 2);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v_arg = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L6;
  }
  __pyx_t_3 = (__pyx_v_kwargs != Py_None);
  __pyx_t_4 = (__pyx_t_3 != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L7_bool_binop_done;
  }
  if (unlikely(__pyx_v_kwargs == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(1, 1014, __pyx_L1_error)
  }
  __pyx_t_4 = (__Pyx_PyDict_ContainsTF(__pyx_n_s_point, ((PyObject*)__pyx_v_kwargs), Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(1, 1014, __pyx_L1_error)
  __pyx_t_3 = (__pyx_t_4 != 0);
  __pyx_t_2 = __pyx_t_3;
  __pyx_L7_bool_binop_done:;
  if (__pyx_t_2) {
    if (unlikely(__pyx_v_kwargs == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 1014, __pyx_L1_error)
    }
    __pyx_t_1 = __Pyx_PyDict_GetItem(((PyObject*)__pyx_v_kwargs), __pyx_n_s_point); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1014, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_arg = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L6;
  }
  /*else*/ {
    if (unlikely(__pyx_v_args == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
      __PYX_ERR(1, 1014, __pyx_L1_error)
    }
    __pyx_t_5 = PyTuple_GET_SIZE(((PyObject*)__pyx_v_args)); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(1, 1014, __pyx_L1_error)
    __pyx_t_1 = PyInt_FromSsize_t(__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1014, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_6 = PyTuple_New(3); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 1014, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_INCREF(__pyx_int_3);
    __Pyx_GIVEREF(__pyx_int_3);
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_int_3);
    __Pyx_INCREF(__pyx_n_s_s);
    __Pyx_GIVEREF(__pyx_n_s_s);
    PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_n_s_s);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Expected_at_least_d_argument_s_g, __pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1014, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_1); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 1014, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(1, 1014, __pyx_L1_error)
  }
  __pyx_L6:;
  while (1) {
    __pyx_t_2 = PyFloat_Check(__pyx_v_arg); 
    __pyx_t_3 = (__pyx_t_2 != 0);
    if (__pyx_t_3) {
      if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, __pyx_n_s_float64_t, long, 1, __Pyx_PyInt_From_long, 1, 0, 0) < 0)) __PYX_ERR(1, 1014, __pyx_L1_error)
      goto __pyx_L10_break;
    }
    __pyx_t_2 = PyInt_Check(__pyx_v_arg); 
    __pyx_t_4 = (__pyx_t_2 != 0);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_3 = __pyx_t_4;
      goto __pyx_L13_bool_binop_done;
    }
    __pyx_t_4 = PyLong_Check(__pyx_v_arg); 
    __pyx_t_2 = (__pyx_t_4 != 0);
    __pyx_t_3 = __pyx_t_2;
    __pyx_L13_bool_binop_done:;
    __pyx_t_2 = (__pyx_t_3 != 0);
    if (__pyx_t_2) {
      if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, __pyx_n_s_int64_t, long, 1, __Pyx_PyInt_From_long, 1, 0, 0) < 0)) __PYX_ERR(1, 1014, __pyx_L1_error)
      goto __pyx_L10_break;
    }
    if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, Py_None, long, 1, __Pyx_PyInt_From_long, 1, 0, 0) < 0)) __PYX_ERR(1, 1014, __pyx_L1_error)
    goto __pyx_L10_break;
  }
  __pyx_L10_break:;
  __pyx_t_6 = PyList_New(0); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 1014, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_v_candidates = ((PyObject*)__pyx_t_6);
  __pyx_t_6 = 0;
  __pyx_t_5 = 0;
  if (unlikely(((PyObject *)__pyx_v_signatures) == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(1, 1014, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_dict_iterator(((PyObject*)__pyx_v_signatures), 1, ((PyObject *)NULL), (&__pyx_t_7), (&__pyx_t_8)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1014, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __pyx_t_6 = __pyx_t_1;
  __pyx_t_1 = 0;
  while (1) {
    __pyx_t_9 = __Pyx_dict_iter_next(__pyx_t_6, __pyx_t_7, &__pyx_t_5, &__pyx_t_1, NULL, NULL, __pyx_t_8);
    if (unlikely(__pyx_t_9 == 0)) break;
    if (unlikely(__pyx_t_9 == -1)) __PYX_ERR(1, 1014, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_sig, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_v_match_found = 0;
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_v_sig, __pyx_n_s_strip); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 1014, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __pyx_t_12 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_11))) {
      __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_11);
      if (likely(__pyx_t_12)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
        __Pyx_INCREF(__pyx_t_12);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_11, function);
      }
    }
    __pyx_t_10 = (__pyx_t_12) ? __Pyx_PyObject_Call2Args(__pyx_t_11, __pyx_t_12, __pyx_kp_s__13) : __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_kp_s__13);
    __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
    if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 1014, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_split); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 1014, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_t_10 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_11))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_11);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_11, function);
      }
    }
    __pyx_t_1 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_11, __pyx_t_10, __pyx_kp_s__14) : __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_kp_s__14);
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1014, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_XDECREF_SET(__pyx_v_src_sig, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_13 = PyList_GET_SIZE(__pyx_v_dest_sig); if (unlikely(__pyx_t_13 == ((Py_ssize_t)-1))) __PYX_ERR(1, 1014, __pyx_L1_error)
    __pyx_t_14 = __pyx_t_13;
    for (__pyx_t_15 = 0; __pyx_t_15 < __pyx_t_14; __pyx_t_15+=1) {
      __pyx_v_i = __pyx_t_15;
      __pyx_t_1 = PyList_GET_ITEM(__pyx_v_dest_sig, __pyx_v_i);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_XDECREF_SET(__pyx_v_dst_type, __pyx_t_1);
      __pyx_t_1 = 0;
      __pyx_t_2 = (__pyx_v_dst_type != Py_None);
      __pyx_t_3 = (__pyx_t_2 != 0);
      if (__pyx_t_3) {
        __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_src_sig, __pyx_v_i, Py_ssize_t, 1, PyInt_FromSsize_t, 0, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1014, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_11 = PyObject_RichCompare(__pyx_t_1, __pyx_v_dst_type, Py_EQ); __Pyx_XGOTREF(__pyx_t_11); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 1014, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_11); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(1, 1014, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        if (__pyx_t_3) {
          __pyx_v_match_found = 1;
          goto __pyx_L20;
        }
        /*else*/ {
          __pyx_v_match_found = 0;
          goto __pyx_L18_break;
        }
        __pyx_L20:;
      }
    }
    __pyx_L18_break:;
    __pyx_t_3 = (__pyx_v_match_found != 0);
    if (__pyx_t_3) {
      __pyx_t_16 = __Pyx_PyList_Append(__pyx_v_candidates, __pyx_v_sig); if (unlikely(__pyx_t_16 == ((int)-1))) __PYX_ERR(1, 1014, __pyx_L1_error)
    }
  }
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_3 = (PyList_GET_SIZE(__pyx_v_candidates) != 0);
  __pyx_t_2 = ((!__pyx_t_3) != 0);
  if (__pyx_t_2) {
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__15, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 1014, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(1, 1014, __pyx_L1_error)
  }
  __pyx_t_7 = PyList_GET_SIZE(__pyx_v_candidates); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1))) __PYX_ERR(1, 1014, __pyx_L1_error)
  __pyx_t_2 = ((__pyx_t_7 > 1) != 0);
  if (__pyx_t_2) {
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__16, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 1014, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(1, 1014, __pyx_L1_error)
  }
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    if (unlikely(((PyObject *)__pyx_v_signatures) == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 1014, __pyx_L1_error)
    }
    __pyx_t_6 = __Pyx_PyDict_GetItem(((PyObject*)__pyx_v_signatures), PyList_GET_ITEM(__pyx_v_candidates, 0)); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 1014, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_r = __pyx_t_6;
    __pyx_t_6 = 0;
    goto __pyx_L0;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("pandas._libs.interval.Float64ClosedLeftIntervalNode.__pyx_fused_cpdef", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_dest_sig);
  __Pyx_XDECREF(__pyx_v_ndarray);
  __Pyx_XDECREF(__pyx_v_arg);
  __Pyx_XDECREF(__pyx_v_candidates);
  __Pyx_XDECREF(__pyx_v_sig);
  __Pyx_XDECREF(__pyx_v_src_sig);
  __Pyx_XDECREF(__pyx_v_dst_type);
  __Pyx_XDECREF(__pyx_v_kwargs);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_9__pyx_fuse_0query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_fuse_0__pyx_f_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float64_t __pyx_v_point, int __pyx_skip_dispatch) {
  __Pyx_memviewslice __pyx_v_indices = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_values = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __pyx_t_5numpy_int64_t __pyx_t_9;
  __pyx_t_5numpy_int64_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  int __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  __Pyx_memviewslice __pyx_t_16 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_17 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  Py_ssize_t __pyx_t_21;
  __Pyx_RefNannySetupContext("__pyx_fuse_0query", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pyx_fuse_0query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1014, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_9__pyx_fuse_0query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = PyFloat_FromDouble(__pyx_v_point); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1014, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1014, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1014, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 1014, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_5) {
            __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1014, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1014, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_3query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_4 = PyFloat_FromDouble(__pyx_v_point); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1014, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_7 = __pyx_t_1; __pyx_t_3 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1014, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1014, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1014, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (__pyx_t_3) {
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_t_4);
          __pyx_t_4 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1014, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pandas/_libs/intervaltree.pxi":1023
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
  __pyx_t_8 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_8) {

    /* "pandas/_libs/intervaltree.pxi":1027
 *             # continue the binary tree structure. Instead, we use linear
 *             # search.
 *             for i in range(self.n_elements):             # <<<<<<<<<<<<<<
 *                 if self.left[i] <= point < self.right[i]:
 *                     result.append(self.indices[i])
 */
    __pyx_t_9 = __pyx_v_self->n_elements;
    __pyx_t_10 = __pyx_t_9;
    for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
      __pyx_v_i = __pyx_t_11;

      /* "pandas/_libs/intervaltree.pxi":1028
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point < self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      __pyx_t_12 = __pyx_v_i;
      __pyx_t_8 = ((*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_self->left.data + __pyx_t_12 * __pyx_v_self->left.strides[0]) ))) <= __pyx_v_point);
      if (__pyx_t_8) {
        __pyx_t_13 = __pyx_v_i;
        __pyx_t_8 = (__pyx_v_point < (*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_self->right.data + __pyx_t_13 * __pyx_v_self->right.strides[0]) ))));
      }
      __pyx_t_14 = (__pyx_t_8 != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":1029
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point < self.right[i]:
 *                     result.append(self.indices[i])             # <<<<<<<<<<<<<<
 *         else:
 *             # There are child nodes. Based on comparing our query to the pivot,
 */
        __pyx_t_15 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->indices.data + __pyx_t_15 * __pyx_v_self->indices.strides[0]) ))));

        /* "pandas/_libs/intervaltree.pxi":1028
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point < self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      }
    }

    /* "pandas/_libs/intervaltree.pxi":1023
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
    goto __pyx_L3;
  }

  /* "pandas/_libs/intervaltree.pxi":1033
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
  /*else*/ {
    __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":1034
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:
 *                 values = self.center_left_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 */
      __pyx_t_16 = __pyx_v_self->center_left_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":1035
 *             if point < self.pivot:
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:
 */
      __pyx_t_17 = __pyx_v_self->center_left_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":1036
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):             # <<<<<<<<<<<<<<
 *                     if not values[i] <= point:
 *                         break
 */
      __pyx_t_9 = __pyx_v_self->n_center;
      __pyx_t_10 = __pyx_t_9;
      for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":1037
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_18 = __pyx_v_i;
        __pyx_t_14 = ((!(((*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_18 * __pyx_v_values.strides[0]) ))) <= __pyx_v_point) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":1038
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:
 */
          goto __pyx_L9_break;

          /* "pandas/_libs/intervaltree.pxi":1037
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":1039
 *                     if not values[i] <= point:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 */
        __pyx_t_19 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_19 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L9_break:;

      /* "pandas/_libs/intervaltree.pxi":1040
 *                         break
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->left_node->max_right) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":1041
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)             # <<<<<<<<<<<<<<
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode *)__pyx_v_self->left_node->__pyx_vtab)->__pyx_fuse_0query(__pyx_v_self->left_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1041, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":1040
 *                         break
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":1033
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":1042
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
    __pyx_t_14 = ((__pyx_v_point > __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":1043
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 *                 values = self.center_right_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 */
      __pyx_t_16 = __pyx_v_self->center_right_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":1044
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:
 */
      __pyx_t_17 = __pyx_v_self->center_right_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":1045
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):             # <<<<<<<<<<<<<<
 *                     if not point < values[i]:
 *                         break
 */
      for (__pyx_t_11 = (__pyx_v_self->n_center - 1); __pyx_t_11 > -1L; __pyx_t_11-=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":1046
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_20 = __pyx_v_i;
        __pyx_t_14 = ((!((__pyx_v_point < (*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_20 * __pyx_v_values.strides[0]) )))) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":1047
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:
 */
          goto __pyx_L13_break;

          /* "pandas/_libs/intervaltree.pxi":1046
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":1048
 *                     if not point < values[i]:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if self.right_node.min_left <= point:
 *                     self.right_node.query(result, point)
 */
        __pyx_t_21 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_21 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L13_break:;

      /* "pandas/_libs/intervaltree.pxi":1049
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      __pyx_t_14 = ((__pyx_v_self->right_node->min_left <= __pyx_v_point) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":1050
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:
 *                     self.right_node.query(result, point)             # <<<<<<<<<<<<<<
 *             else:
 *                 result.extend(self.center_left_indices)
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode *)__pyx_v_self->right_node->__pyx_vtab)->__pyx_fuse_0query(__pyx_v_self->right_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1050, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":1049
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":1042
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":1052
 *                     self.right_node.query(result, point)
 *             else:
 *                 result.extend(self.center_left_indices)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
    /*else*/ {
      __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->extend(__pyx_v_result, __pyx_v_self->center_left_indices); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1052, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_L7:;
  }
  __pyx_L3:;

  /* "pandas/_libs/intervaltree.pxi":1014
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __PYX_XDEC_MEMVIEW(&__pyx_t_16, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_17, 1);
  __Pyx_AddTraceback("pandas._libs.interval.Float64ClosedLeftIntervalNode.query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_indices, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_values, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_9__pyx_fuse_0query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_0__pyx_mdef_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_9__pyx_fuse_0query = {"__pyx_fuse_0query", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_9__pyx_fuse_0query, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_2query};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_9__pyx_fuse_0query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result = 0;
  __pyx_t_5numpy_float64_t __pyx_v_point;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fuse_0query (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_result,&__pyx_n_s_point,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_result)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_point)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_0query", 1, 2, 2, 1); __PYX_ERR(1, 1014, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_0query") < 0)) __PYX_ERR(1, 1014, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_result = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)values[0]);
    __pyx_v_point = __pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_point == ((npy_float64)-1)) && PyErr_Occurred())) __PYX_ERR(1, 1014, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_0query", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 1014, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Float64ClosedLeftIntervalNode.__pyx_fuse_0query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_result), __pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector, 1, "result", 0))) __PYX_ERR(1, 1014, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_8__pyx_fuse_0query(((struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode *)__pyx_v_self), __pyx_v_result, __pyx_v_point);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_8__pyx_fuse_0query(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float64_t __pyx_v_point) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__pyx_fuse_0query", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_fuse_0__pyx_f_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_query(__pyx_v_self, __pyx_v_result, __pyx_v_point, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1014, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Float64ClosedLeftIntervalNode.__pyx_fuse_0query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_11__pyx_fuse_1query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_fuse_1__pyx_f_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float32_t __pyx_v_point, int __pyx_skip_dispatch) {
  __Pyx_memviewslice __pyx_v_indices = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_values = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __pyx_t_5numpy_int64_t __pyx_t_9;
  __pyx_t_5numpy_int64_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  int __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  __Pyx_memviewslice __pyx_t_16 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_17 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  Py_ssize_t __pyx_t_21;
  __Pyx_RefNannySetupContext("__pyx_fuse_1query", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pyx_fuse_1query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1014, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_11__pyx_fuse_1query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = PyFloat_FromDouble(__pyx_v_point); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1014, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1014, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1014, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 1014, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_5) {
            __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1014, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1014, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_3query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_4 = PyFloat_FromDouble(__pyx_v_point); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1014, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_7 = __pyx_t_1; __pyx_t_3 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1014, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1014, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1014, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (__pyx_t_3) {
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_t_4);
          __pyx_t_4 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1014, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pandas/_libs/intervaltree.pxi":1023
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
  __pyx_t_8 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_8) {

    /* "pandas/_libs/intervaltree.pxi":1027
 *             # continue the binary tree structure. Instead, we use linear
 *             # search.
 *             for i in range(self.n_elements):             # <<<<<<<<<<<<<<
 *                 if self.left[i] <= point < self.right[i]:
 *                     result.append(self.indices[i])
 */
    __pyx_t_9 = __pyx_v_self->n_elements;
    __pyx_t_10 = __pyx_t_9;
    for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
      __pyx_v_i = __pyx_t_11;

      /* "pandas/_libs/intervaltree.pxi":1028
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point < self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      __pyx_t_12 = __pyx_v_i;
      __pyx_t_8 = ((*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_self->left.data + __pyx_t_12 * __pyx_v_self->left.strides[0]) ))) <= __pyx_v_point);
      if (__pyx_t_8) {
        __pyx_t_13 = __pyx_v_i;
        __pyx_t_8 = (__pyx_v_point < (*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_self->right.data + __pyx_t_13 * __pyx_v_self->right.strides[0]) ))));
      }
      __pyx_t_14 = (__pyx_t_8 != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":1029
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point < self.right[i]:
 *                     result.append(self.indices[i])             # <<<<<<<<<<<<<<
 *         else:
 *             # There are child nodes. Based on comparing our query to the pivot,
 */
        __pyx_t_15 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->indices.data + __pyx_t_15 * __pyx_v_self->indices.strides[0]) ))));

        /* "pandas/_libs/intervaltree.pxi":1028
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point < self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      }
    }

    /* "pandas/_libs/intervaltree.pxi":1023
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
    goto __pyx_L3;
  }

  /* "pandas/_libs/intervaltree.pxi":1033
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
  /*else*/ {
    __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":1034
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:
 *                 values = self.center_left_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 */
      __pyx_t_16 = __pyx_v_self->center_left_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":1035
 *             if point < self.pivot:
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:
 */
      __pyx_t_17 = __pyx_v_self->center_left_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":1036
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):             # <<<<<<<<<<<<<<
 *                     if not values[i] <= point:
 *                         break
 */
      __pyx_t_9 = __pyx_v_self->n_center;
      __pyx_t_10 = __pyx_t_9;
      for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":1037
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_18 = __pyx_v_i;
        __pyx_t_14 = ((!(((*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_18 * __pyx_v_values.strides[0]) ))) <= __pyx_v_point) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":1038
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:
 */
          goto __pyx_L9_break;

          /* "pandas/_libs/intervaltree.pxi":1037
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":1039
 *                     if not values[i] <= point:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 */
        __pyx_t_19 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_19 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L9_break:;

      /* "pandas/_libs/intervaltree.pxi":1040
 *                         break
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->left_node->max_right) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":1041
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)             # <<<<<<<<<<<<<<
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode *)__pyx_v_self->left_node->__pyx_vtab)->__pyx_fuse_1query(__pyx_v_self->left_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1041, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":1040
 *                         break
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":1033
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":1042
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
    __pyx_t_14 = ((__pyx_v_point > __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":1043
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 *                 values = self.center_right_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 */
      __pyx_t_16 = __pyx_v_self->center_right_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":1044
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:
 */
      __pyx_t_17 = __pyx_v_self->center_right_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":1045
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):             # <<<<<<<<<<<<<<
 *                     if not point < values[i]:
 *                         break
 */
      for (__pyx_t_11 = (__pyx_v_self->n_center - 1); __pyx_t_11 > -1L; __pyx_t_11-=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":1046
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_20 = __pyx_v_i;
        __pyx_t_14 = ((!((__pyx_v_point < (*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_20 * __pyx_v_values.strides[0]) )))) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":1047
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:
 */
          goto __pyx_L13_break;

          /* "pandas/_libs/intervaltree.pxi":1046
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":1048
 *                     if not point < values[i]:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if self.right_node.min_left <= point:
 *                     self.right_node.query(result, point)
 */
        __pyx_t_21 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_21 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L13_break:;

      /* "pandas/_libs/intervaltree.pxi":1049
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      __pyx_t_14 = ((__pyx_v_self->right_node->min_left <= __pyx_v_point) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":1050
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:
 *                     self.right_node.query(result, point)             # <<<<<<<<<<<<<<
 *             else:
 *                 result.extend(self.center_left_indices)
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode *)__pyx_v_self->right_node->__pyx_vtab)->__pyx_fuse_1query(__pyx_v_self->right_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1050, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":1049
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":1042
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":1052
 *                     self.right_node.query(result, point)
 *             else:
 *                 result.extend(self.center_left_indices)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
    /*else*/ {
      __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->extend(__pyx_v_result, __pyx_v_self->center_left_indices); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1052, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_L7:;
  }
  __pyx_L3:;

  /* "pandas/_libs/intervaltree.pxi":1014
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __PYX_XDEC_MEMVIEW(&__pyx_t_16, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_17, 1);
  __Pyx_AddTraceback("pandas._libs.interval.Float64ClosedLeftIntervalNode.query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_indices, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_values, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_11__pyx_fuse_1query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_1__pyx_mdef_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_11__pyx_fuse_1query = {"__pyx_fuse_1query", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_11__pyx_fuse_1query, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_2query};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_11__pyx_fuse_1query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result = 0;
  __pyx_t_5numpy_float32_t __pyx_v_point;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fuse_1query (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_result,&__pyx_n_s_point,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_result)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_point)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_1query", 1, 2, 2, 1); __PYX_ERR(1, 1014, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_1query") < 0)) __PYX_ERR(1, 1014, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_result = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)values[0]);
    __pyx_v_point = __pyx_PyFloat_AsFloat(values[1]); if (unlikely((__pyx_v_point == ((npy_float32)-1)) && PyErr_Occurred())) __PYX_ERR(1, 1014, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_1query", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 1014, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Float64ClosedLeftIntervalNode.__pyx_fuse_1query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_result), __pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector, 1, "result", 0))) __PYX_ERR(1, 1014, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_10__pyx_fuse_1query(((struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode *)__pyx_v_self), __pyx_v_result, __pyx_v_point);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_10__pyx_fuse_1query(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float32_t __pyx_v_point) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__pyx_fuse_1query", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_fuse_1__pyx_f_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_query(__pyx_v_self, __pyx_v_result, __pyx_v_point, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1014, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Float64ClosedLeftIntervalNode.__pyx_fuse_1query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_13__pyx_fuse_2query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_fuse_2__pyx_f_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int64_t __pyx_v_point, int __pyx_skip_dispatch) {
  __Pyx_memviewslice __pyx_v_indices = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_values = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __pyx_t_5numpy_int64_t __pyx_t_9;
  __pyx_t_5numpy_int64_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  int __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  __Pyx_memviewslice __pyx_t_16 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_17 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  Py_ssize_t __pyx_t_21;
  __Pyx_RefNannySetupContext("__pyx_fuse_2query", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pyx_fuse_2query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1014, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_13__pyx_fuse_2query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = __Pyx_PyInt_From_npy_int64(__pyx_v_point); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1014, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1014, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1014, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 1014, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_5) {
            __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1014, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1014, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_3query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_4 = __Pyx_PyInt_From_npy_int64(__pyx_v_point); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1014, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_7 = __pyx_t_1; __pyx_t_3 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1014, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1014, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1014, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (__pyx_t_3) {
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_t_4);
          __pyx_t_4 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1014, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pandas/_libs/intervaltree.pxi":1023
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
  __pyx_t_8 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_8) {

    /* "pandas/_libs/intervaltree.pxi":1027
 *             # continue the binary tree structure. Instead, we use linear
 *             # search.
 *             for i in range(self.n_elements):             # <<<<<<<<<<<<<<
 *                 if self.left[i] <= point < self.right[i]:
 *                     result.append(self.indices[i])
 */
    __pyx_t_9 = __pyx_v_self->n_elements;
    __pyx_t_10 = __pyx_t_9;
    for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
      __pyx_v_i = __pyx_t_11;

      /* "pandas/_libs/intervaltree.pxi":1028
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point < self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      __pyx_t_12 = __pyx_v_i;
      __pyx_t_8 = ((*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_self->left.data + __pyx_t_12 * __pyx_v_self->left.strides[0]) ))) <= __pyx_v_point);
      if (__pyx_t_8) {
        __pyx_t_13 = __pyx_v_i;
        __pyx_t_8 = (__pyx_v_point < (*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_self->right.data + __pyx_t_13 * __pyx_v_self->right.strides[0]) ))));
      }
      __pyx_t_14 = (__pyx_t_8 != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":1029
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point < self.right[i]:
 *                     result.append(self.indices[i])             # <<<<<<<<<<<<<<
 *         else:
 *             # There are child nodes. Based on comparing our query to the pivot,
 */
        __pyx_t_15 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->indices.data + __pyx_t_15 * __pyx_v_self->indices.strides[0]) ))));

        /* "pandas/_libs/intervaltree.pxi":1028
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point < self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      }
    }

    /* "pandas/_libs/intervaltree.pxi":1023
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
    goto __pyx_L3;
  }

  /* "pandas/_libs/intervaltree.pxi":1033
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
  /*else*/ {
    __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":1034
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:
 *                 values = self.center_left_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 */
      __pyx_t_16 = __pyx_v_self->center_left_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":1035
 *             if point < self.pivot:
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:
 */
      __pyx_t_17 = __pyx_v_self->center_left_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":1036
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):             # <<<<<<<<<<<<<<
 *                     if not values[i] <= point:
 *                         break
 */
      __pyx_t_9 = __pyx_v_self->n_center;
      __pyx_t_10 = __pyx_t_9;
      for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":1037
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_18 = __pyx_v_i;
        __pyx_t_14 = ((!(((*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_18 * __pyx_v_values.strides[0]) ))) <= __pyx_v_point) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":1038
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:
 */
          goto __pyx_L9_break;

          /* "pandas/_libs/intervaltree.pxi":1037
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":1039
 *                     if not values[i] <= point:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 */
        __pyx_t_19 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_19 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L9_break:;

      /* "pandas/_libs/intervaltree.pxi":1040
 *                         break
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->left_node->max_right) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":1041
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)             # <<<<<<<<<<<<<<
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode *)__pyx_v_self->left_node->__pyx_vtab)->__pyx_fuse_2query(__pyx_v_self->left_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1041, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":1040
 *                         break
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":1033
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":1042
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
    __pyx_t_14 = ((__pyx_v_point > __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":1043
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 *                 values = self.center_right_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 */
      __pyx_t_16 = __pyx_v_self->center_right_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":1044
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:
 */
      __pyx_t_17 = __pyx_v_self->center_right_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":1045
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):             # <<<<<<<<<<<<<<
 *                     if not point < values[i]:
 *                         break
 */
      for (__pyx_t_11 = (__pyx_v_self->n_center - 1); __pyx_t_11 > -1L; __pyx_t_11-=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":1046
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_20 = __pyx_v_i;
        __pyx_t_14 = ((!((__pyx_v_point < (*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_20 * __pyx_v_values.strides[0]) )))) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":1047
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:
 */
          goto __pyx_L13_break;

          /* "pandas/_libs/intervaltree.pxi":1046
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":1048
 *                     if not point < values[i]:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if self.right_node.min_left <= point:
 *                     self.right_node.query(result, point)
 */
        __pyx_t_21 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_21 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L13_break:;

      /* "pandas/_libs/intervaltree.pxi":1049
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      __pyx_t_14 = ((__pyx_v_self->right_node->min_left <= __pyx_v_point) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":1050
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:
 *                     self.right_node.query(result, point)             # <<<<<<<<<<<<<<
 *             else:
 *                 result.extend(self.center_left_indices)
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode *)__pyx_v_self->right_node->__pyx_vtab)->__pyx_fuse_2query(__pyx_v_self->right_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1050, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":1049
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":1042
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":1052
 *                     self.right_node.query(result, point)
 *             else:
 *                 result.extend(self.center_left_indices)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
    /*else*/ {
      __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->extend(__pyx_v_result, __pyx_v_self->center_left_indices); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1052, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_L7:;
  }
  __pyx_L3:;

  /* "pandas/_libs/intervaltree.pxi":1014
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __PYX_XDEC_MEMVIEW(&__pyx_t_16, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_17, 1);
  __Pyx_AddTraceback("pandas._libs.interval.Float64ClosedLeftIntervalNode.query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_indices, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_values, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_13__pyx_fuse_2query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_2__pyx_mdef_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_13__pyx_fuse_2query = {"__pyx_fuse_2query", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_13__pyx_fuse_2query, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_2query};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_13__pyx_fuse_2query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result = 0;
  __pyx_t_5numpy_int64_t __pyx_v_point;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fuse_2query (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_result,&__pyx_n_s_point,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_result)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_point)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_2query", 1, 2, 2, 1); __PYX_ERR(1, 1014, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_2query") < 0)) __PYX_ERR(1, 1014, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_result = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)values[0]);
    __pyx_v_point = __Pyx_PyInt_As_npy_int64(values[1]); if (unlikely((__pyx_v_point == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(1, 1014, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_2query", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 1014, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Float64ClosedLeftIntervalNode.__pyx_fuse_2query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_result), __pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector, 1, "result", 0))) __PYX_ERR(1, 1014, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_12__pyx_fuse_2query(((struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode *)__pyx_v_self), __pyx_v_result, __pyx_v_point);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_12__pyx_fuse_2query(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int64_t __pyx_v_point) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__pyx_fuse_2query", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_fuse_2__pyx_f_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_query(__pyx_v_self, __pyx_v_result, __pyx_v_point, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1014, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Float64ClosedLeftIntervalNode.__pyx_fuse_2query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_15__pyx_fuse_3query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_fuse_3__pyx_f_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int32_t __pyx_v_point, int __pyx_skip_dispatch) {
  __Pyx_memviewslice __pyx_v_indices = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_values = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __pyx_t_5numpy_int64_t __pyx_t_9;
  __pyx_t_5numpy_int64_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  int __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  __Pyx_memviewslice __pyx_t_16 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_17 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  Py_ssize_t __pyx_t_21;
  __Pyx_RefNannySetupContext("__pyx_fuse_3query", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pyx_fuse_3query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1014, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_15__pyx_fuse_3query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = __Pyx_PyInt_From_npy_int32(__pyx_v_point); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1014, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1014, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1014, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 1014, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_5) {
            __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1014, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1014, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_3query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_4 = __Pyx_PyInt_From_npy_int32(__pyx_v_point); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1014, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_7 = __pyx_t_1; __pyx_t_3 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1014, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1014, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1014, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (__pyx_t_3) {
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_t_4);
          __pyx_t_4 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1014, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pandas/_libs/intervaltree.pxi":1023
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
  __pyx_t_8 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_8) {

    /* "pandas/_libs/intervaltree.pxi":1027
 *             # continue the binary tree structure. Instead, we use linear
 *             # search.
 *             for i in range(self.n_elements):             # <<<<<<<<<<<<<<
 *                 if self.left[i] <= point < self.right[i]:
 *                     result.append(self.indices[i])
 */
    __pyx_t_9 = __pyx_v_self->n_elements;
    __pyx_t_10 = __pyx_t_9;
    for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
      __pyx_v_i = __pyx_t_11;

      /* "pandas/_libs/intervaltree.pxi":1028
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point < self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      __pyx_t_12 = __pyx_v_i;
      __pyx_t_8 = ((*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_self->left.data + __pyx_t_12 * __pyx_v_self->left.strides[0]) ))) <= __pyx_v_point);
      if (__pyx_t_8) {
        __pyx_t_13 = __pyx_v_i;
        __pyx_t_8 = (__pyx_v_point < (*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_self->right.data + __pyx_t_13 * __pyx_v_self->right.strides[0]) ))));
      }
      __pyx_t_14 = (__pyx_t_8 != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":1029
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point < self.right[i]:
 *                     result.append(self.indices[i])             # <<<<<<<<<<<<<<
 *         else:
 *             # There are child nodes. Based on comparing our query to the pivot,
 */
        __pyx_t_15 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->indices.data + __pyx_t_15 * __pyx_v_self->indices.strides[0]) ))));

        /* "pandas/_libs/intervaltree.pxi":1028
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point < self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      }
    }

    /* "pandas/_libs/intervaltree.pxi":1023
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
    goto __pyx_L3;
  }

  /* "pandas/_libs/intervaltree.pxi":1033
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
  /*else*/ {
    __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":1034
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:
 *                 values = self.center_left_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 */
      __pyx_t_16 = __pyx_v_self->center_left_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":1035
 *             if point < self.pivot:
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:
 */
      __pyx_t_17 = __pyx_v_self->center_left_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":1036
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):             # <<<<<<<<<<<<<<
 *                     if not values[i] <= point:
 *                         break
 */
      __pyx_t_9 = __pyx_v_self->n_center;
      __pyx_t_10 = __pyx_t_9;
      for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":1037
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_18 = __pyx_v_i;
        __pyx_t_14 = ((!(((*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_18 * __pyx_v_values.strides[0]) ))) <= __pyx_v_point) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":1038
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:
 */
          goto __pyx_L9_break;

          /* "pandas/_libs/intervaltree.pxi":1037
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":1039
 *                     if not values[i] <= point:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 */
        __pyx_t_19 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_19 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L9_break:;

      /* "pandas/_libs/intervaltree.pxi":1040
 *                         break
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->left_node->max_right) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":1041
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)             # <<<<<<<<<<<<<<
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode *)__pyx_v_self->left_node->__pyx_vtab)->__pyx_fuse_3query(__pyx_v_self->left_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1041, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":1040
 *                         break
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":1033
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":1042
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
    __pyx_t_14 = ((__pyx_v_point > __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":1043
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 *                 values = self.center_right_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 */
      __pyx_t_16 = __pyx_v_self->center_right_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":1044
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:
 */
      __pyx_t_17 = __pyx_v_self->center_right_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":1045
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):             # <<<<<<<<<<<<<<
 *                     if not point < values[i]:
 *                         break
 */
      for (__pyx_t_11 = (__pyx_v_self->n_center - 1); __pyx_t_11 > -1L; __pyx_t_11-=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":1046
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_20 = __pyx_v_i;
        __pyx_t_14 = ((!((__pyx_v_point < (*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_20 * __pyx_v_values.strides[0]) )))) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":1047
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:
 */
          goto __pyx_L13_break;

          /* "pandas/_libs/intervaltree.pxi":1046
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":1048
 *                     if not point < values[i]:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if self.right_node.min_left <= point:
 *                     self.right_node.query(result, point)
 */
        __pyx_t_21 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_21 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L13_break:;

      /* "pandas/_libs/intervaltree.pxi":1049
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      __pyx_t_14 = ((__pyx_v_self->right_node->min_left <= __pyx_v_point) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":1050
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:
 *                     self.right_node.query(result, point)             # <<<<<<<<<<<<<<
 *             else:
 *                 result.extend(self.center_left_indices)
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode *)__pyx_v_self->right_node->__pyx_vtab)->__pyx_fuse_3query(__pyx_v_self->right_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1050, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":1049
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":1042
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":1052
 *                     self.right_node.query(result, point)
 *             else:
 *                 result.extend(self.center_left_indices)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
    /*else*/ {
      __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->extend(__pyx_v_result, __pyx_v_self->center_left_indices); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1052, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_L7:;
  }
  __pyx_L3:;

  /* "pandas/_libs/intervaltree.pxi":1014
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __PYX_XDEC_MEMVIEW(&__pyx_t_16, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_17, 1);
  __Pyx_AddTraceback("pandas._libs.interval.Float64ClosedLeftIntervalNode.query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_indices, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_values, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_15__pyx_fuse_3query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_3__pyx_mdef_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_15__pyx_fuse_3query = {"__pyx_fuse_3query", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_15__pyx_fuse_3query, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_2query};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_15__pyx_fuse_3query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result = 0;
  __pyx_t_5numpy_int32_t __pyx_v_point;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fuse_3query (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_result,&__pyx_n_s_point,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_result)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_point)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_3query", 1, 2, 2, 1); __PYX_ERR(1, 1014, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_3query") < 0)) __PYX_ERR(1, 1014, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_result = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)values[0]);
    __pyx_v_point = __Pyx_PyInt_As_npy_int32(values[1]); if (unlikely((__pyx_v_point == ((npy_int32)-1)) && PyErr_Occurred())) __PYX_ERR(1, 1014, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_3query", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 1014, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Float64ClosedLeftIntervalNode.__pyx_fuse_3query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_result), __pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector, 1, "result", 0))) __PYX_ERR(1, 1014, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_14__pyx_fuse_3query(((struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode *)__pyx_v_self), __pyx_v_result, __pyx_v_point);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_14__pyx_fuse_3query(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int32_t __pyx_v_point) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__pyx_fuse_3query", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_fuse_3__pyx_f_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_query(__pyx_v_self, __pyx_v_result, __pyx_v_point, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1014, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Float64ClosedLeftIntervalNode.__pyx_fuse_3query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_17__pyx_fuse_4query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_fuse_4__pyx_f_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_uint64_t __pyx_v_point, int __pyx_skip_dispatch) {
  __Pyx_memviewslice __pyx_v_indices = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_values = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __pyx_t_5numpy_int64_t __pyx_t_9;
  __pyx_t_5numpy_int64_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  int __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  __Pyx_memviewslice __pyx_t_16 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_17 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  Py_ssize_t __pyx_t_21;
  __Pyx_RefNannySetupContext("__pyx_fuse_4query", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pyx_fuse_4query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1014, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_17__pyx_fuse_4query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = __Pyx_PyInt_From_npy_uint64(__pyx_v_point); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1014, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1014, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1014, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 1014, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_5) {
            __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1014, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1014, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_3query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_4 = __Pyx_PyInt_From_npy_uint64(__pyx_v_point); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1014, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_7 = __pyx_t_1; __pyx_t_3 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1014, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1014, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1014, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (__pyx_t_3) {
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_t_4);
          __pyx_t_4 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1014, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pandas/_libs/intervaltree.pxi":1023
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
  __pyx_t_8 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_8) {

    /* "pandas/_libs/intervaltree.pxi":1027
 *             # continue the binary tree structure. Instead, we use linear
 *             # search.
 *             for i in range(self.n_elements):             # <<<<<<<<<<<<<<
 *                 if self.left[i] <= point < self.right[i]:
 *                     result.append(self.indices[i])
 */
    __pyx_t_9 = __pyx_v_self->n_elements;
    __pyx_t_10 = __pyx_t_9;
    for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
      __pyx_v_i = __pyx_t_11;

      /* "pandas/_libs/intervaltree.pxi":1028
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point < self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      __pyx_t_12 = __pyx_v_i;
      __pyx_t_8 = ((*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_self->left.data + __pyx_t_12 * __pyx_v_self->left.strides[0]) ))) <= __pyx_v_point);
      if (__pyx_t_8) {
        __pyx_t_13 = __pyx_v_i;
        __pyx_t_8 = (__pyx_v_point < (*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_self->right.data + __pyx_t_13 * __pyx_v_self->right.strides[0]) ))));
      }
      __pyx_t_14 = (__pyx_t_8 != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":1029
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point < self.right[i]:
 *                     result.append(self.indices[i])             # <<<<<<<<<<<<<<
 *         else:
 *             # There are child nodes. Based on comparing our query to the pivot,
 */
        __pyx_t_15 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->indices.data + __pyx_t_15 * __pyx_v_self->indices.strides[0]) ))));

        /* "pandas/_libs/intervaltree.pxi":1028
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point < self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      }
    }

    /* "pandas/_libs/intervaltree.pxi":1023
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
    goto __pyx_L3;
  }

  /* "pandas/_libs/intervaltree.pxi":1033
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
  /*else*/ {
    __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":1034
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:
 *                 values = self.center_left_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 */
      __pyx_t_16 = __pyx_v_self->center_left_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":1035
 *             if point < self.pivot:
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:
 */
      __pyx_t_17 = __pyx_v_self->center_left_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":1036
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):             # <<<<<<<<<<<<<<
 *                     if not values[i] <= point:
 *                         break
 */
      __pyx_t_9 = __pyx_v_self->n_center;
      __pyx_t_10 = __pyx_t_9;
      for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":1037
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_18 = __pyx_v_i;
        __pyx_t_14 = ((!(((*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_18 * __pyx_v_values.strides[0]) ))) <= __pyx_v_point) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":1038
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:
 */
          goto __pyx_L9_break;

          /* "pandas/_libs/intervaltree.pxi":1037
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":1039
 *                     if not values[i] <= point:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 */
        __pyx_t_19 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_19 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L9_break:;

      /* "pandas/_libs/intervaltree.pxi":1040
 *                         break
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->left_node->max_right) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":1041
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)             # <<<<<<<<<<<<<<
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode *)__pyx_v_self->left_node->__pyx_vtab)->__pyx_fuse_4query(__pyx_v_self->left_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1041, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":1040
 *                         break
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":1033
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":1042
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
    __pyx_t_14 = ((__pyx_v_point > __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":1043
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 *                 values = self.center_right_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 */
      __pyx_t_16 = __pyx_v_self->center_right_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":1044
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:
 */
      __pyx_t_17 = __pyx_v_self->center_right_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":1045
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):             # <<<<<<<<<<<<<<
 *                     if not point < values[i]:
 *                         break
 */
      for (__pyx_t_11 = (__pyx_v_self->n_center - 1); __pyx_t_11 > -1L; __pyx_t_11-=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":1046
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_20 = __pyx_v_i;
        __pyx_t_14 = ((!((__pyx_v_point < (*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_20 * __pyx_v_values.strides[0]) )))) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":1047
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:
 */
          goto __pyx_L13_break;

          /* "pandas/_libs/intervaltree.pxi":1046
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":1048
 *                     if not point < values[i]:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if self.right_node.min_left <= point:
 *                     self.right_node.query(result, point)
 */
        __pyx_t_21 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_21 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L13_break:;

      /* "pandas/_libs/intervaltree.pxi":1049
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      __pyx_t_14 = ((__pyx_v_self->right_node->min_left <= __pyx_v_point) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":1050
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:
 *                     self.right_node.query(result, point)             # <<<<<<<<<<<<<<
 *             else:
 *                 result.extend(self.center_left_indices)
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode *)__pyx_v_self->right_node->__pyx_vtab)->__pyx_fuse_4query(__pyx_v_self->right_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1050, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":1049
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":1042
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":1052
 *                     self.right_node.query(result, point)
 *             else:
 *                 result.extend(self.center_left_indices)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
    /*else*/ {
      __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->extend(__pyx_v_result, __pyx_v_self->center_left_indices); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1052, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_L7:;
  }
  __pyx_L3:;

  /* "pandas/_libs/intervaltree.pxi":1014
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __PYX_XDEC_MEMVIEW(&__pyx_t_16, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_17, 1);
  __Pyx_AddTraceback("pandas._libs.interval.Float64ClosedLeftIntervalNode.query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_indices, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_values, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_17__pyx_fuse_4query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_4__pyx_mdef_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_17__pyx_fuse_4query = {"__pyx_fuse_4query", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_17__pyx_fuse_4query, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_2query};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_17__pyx_fuse_4query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result = 0;
  __pyx_t_5numpy_uint64_t __pyx_v_point;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fuse_4query (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_result,&__pyx_n_s_point,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_result)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_point)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_4query", 1, 2, 2, 1); __PYX_ERR(1, 1014, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_4query") < 0)) __PYX_ERR(1, 1014, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_result = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)values[0]);
    __pyx_v_point = __Pyx_PyInt_As_npy_uint64(values[1]); if (unlikely((__pyx_v_point == ((npy_uint64)-1)) && PyErr_Occurred())) __PYX_ERR(1, 1014, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_4query", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 1014, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Float64ClosedLeftIntervalNode.__pyx_fuse_4query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_result), __pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector, 1, "result", 0))) __PYX_ERR(1, 1014, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_16__pyx_fuse_4query(((struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode *)__pyx_v_self), __pyx_v_result, __pyx_v_point);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_16__pyx_fuse_4query(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_uint64_t __pyx_v_point) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__pyx_fuse_4query", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_fuse_4__pyx_f_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_query(__pyx_v_self, __pyx_v_result, __pyx_v_point, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1014, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Float64ClosedLeftIntervalNode.__pyx_fuse_4query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":1054
 *                 result.extend(self.center_left_indices)
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         if self.is_leaf_node:
 *             return ('<Float64ClosedLeftIntervalNode: '
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_5__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_5__repr__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_4__repr__(((struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_4__repr__(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode *__pyx_v_self) {
  PyObject *__pyx_v_n_left = NULL;
  PyObject *__pyx_v_n_right = NULL;
  PyObject *__pyx_v_n_center = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  Py_UCS4 __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "pandas/_libs/intervaltree.pxi":1055
 * 
 *     def __repr__(self):
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             return ('<Float64ClosedLeftIntervalNode: '
 *                     '%s elements (terminal)>' % self.n_elements)
 */
  __pyx_t_1 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_1) {

    /* "pandas/_libs/intervaltree.pxi":1056
 *     def __repr__(self):
 *         if self.is_leaf_node:
 *             return ('<Float64ClosedLeftIntervalNode: '             # <<<<<<<<<<<<<<
 *                     '%s elements (terminal)>' % self.n_elements)
 *         else:
 */
    __Pyx_XDECREF(__pyx_r);

    /* "pandas/_libs/intervaltree.pxi":1057
 *         if self.is_leaf_node:
 *             return ('<Float64ClosedLeftIntervalNode: '
 *                     '%s elements (terminal)>' % self.n_elements)             # <<<<<<<<<<<<<<
 *         else:
 *             n_left = self.left_node.n_elements
 */
    __pyx_t_2 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_elements); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1057, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PyUnicode_Format(__pyx_kp_u_Float64ClosedLeftIntervalNode_s, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1057, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pandas/_libs/intervaltree.pxi":1055
 * 
 *     def __repr__(self):
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             return ('<Float64ClosedLeftIntervalNode: '
 *                     '%s elements (terminal)>' % self.n_elements)
 */
  }

  /* "pandas/_libs/intervaltree.pxi":1059
 *                     '%s elements (terminal)>' % self.n_elements)
 *         else:
 *             n_left = self.left_node.n_elements             # <<<<<<<<<<<<<<
 *             n_right = self.right_node.n_elements
 *             n_center = self.n_elements - n_left - n_right
 */
  /*else*/ {
    __pyx_t_3 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->left_node->n_elements); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1059, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_v_n_left = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "pandas/_libs/intervaltree.pxi":1060
 *         else:
 *             n_left = self.left_node.n_elements
 *             n_right = self.right_node.n_elements             # <<<<<<<<<<<<<<
 *             n_center = self.n_elements - n_left - n_right
 *             return ('<Float64ClosedLeftIntervalNode: '
 */
    __pyx_t_3 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->right_node->n_elements); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1060, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_v_n_right = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "pandas/_libs/intervaltree.pxi":1061
 *             n_left = self.left_node.n_elements
 *             n_right = self.right_node.n_elements
 *             n_center = self.n_elements - n_left - n_right             # <<<<<<<<<<<<<<
 *             return ('<Float64ClosedLeftIntervalNode: '
 *                     'pivot %s, %s elements (%s left, %s right, %s '
 */
    __pyx_t_3 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_elements); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1061, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = PyNumber_Subtract(__pyx_t_3, __pyx_v_n_left); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1061, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyNumber_Subtract(__pyx_t_2, __pyx_v_n_right); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1061, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_n_center = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "pandas/_libs/intervaltree.pxi":1062
 *             n_right = self.right_node.n_elements
 *             n_center = self.n_elements - n_left - n_right
 *             return ('<Float64ClosedLeftIntervalNode: '             # <<<<<<<<<<<<<<
 *                     'pivot %s, %s elements (%s left, %s right, %s '
 *                     'overlapping)>' % (self.pivot, self.n_elements,
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = PyTuple_New(11); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1062, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = 0;
    __pyx_t_5 = 127;
    __Pyx_INCREF(__pyx_kp_u_Float64ClosedLeftIntervalNode_p);
    __pyx_t_4 += 38;
    __Pyx_GIVEREF(__pyx_kp_u_Float64ClosedLeftIntervalNode_p);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_kp_u_Float64ClosedLeftIntervalNode_p);

    /* "pandas/_libs/intervaltree.pxi":1064
 *             return ('<Float64ClosedLeftIntervalNode: '
 *                     'pivot %s, %s elements (%s left, %s right, %s '
 *                     'overlapping)>' % (self.pivot, self.n_elements,             # <<<<<<<<<<<<<<
 *                                        n_left, n_right, n_center))
 * 
 */
    __pyx_t_2 = PyFloat_FromDouble(__pyx_v_self->pivot); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1064, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_t_2), __pyx_empty_unicode); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 1064, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_5 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_6) > __pyx_t_5) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_6) : __pyx_t_5;
    __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_6);
    __pyx_t_6 = 0;
    __Pyx_INCREF(__pyx_kp_u__18);
    __pyx_t_4 += 2;
    __Pyx_GIVEREF(__pyx_kp_u__18);
    PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_kp_u__18);
    __pyx_t_6 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_elements); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 1064, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_2 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_t_6), __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1064, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_5 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) > __pyx_t_5) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) : __pyx_t_5;
    __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_3, 3, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_elements);
    __pyx_t_4 += 11;
    __Pyx_GIVEREF(__pyx_kp_u_elements);
    PyTuple_SET_ITEM(__pyx_t_3, 4, __pyx_kp_u_elements);

    /* "pandas/_libs/intervaltree.pxi":1065
 *                     'pivot %s, %s elements (%s left, %s right, %s '
 *                     'overlapping)>' % (self.pivot, self.n_elements,
 *                                        n_left, n_right, n_center))             # <<<<<<<<<<<<<<
 * 
 *     def counts(self):
 */
    __pyx_t_2 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_v_n_left), __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1065, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) > __pyx_t_5) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) : __pyx_t_5;
    __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_3, 5, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_left_2);
    __pyx_t_4 += 7;
    __Pyx_GIVEREF(__pyx_kp_u_left_2);
    PyTuple_SET_ITEM(__pyx_t_3, 6, __pyx_kp_u_left_2);
    __pyx_t_2 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_v_n_right), __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1065, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) > __pyx_t_5) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) : __pyx_t_5;
    __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_3, 7, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_right_2);
    __pyx_t_4 += 8;
    __Pyx_GIVEREF(__pyx_kp_u_right_2);
    PyTuple_SET_ITEM(__pyx_t_3, 8, __pyx_kp_u_right_2);
    __pyx_t_2 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_v_n_center), __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1065, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) > __pyx_t_5) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) : __pyx_t_5;
    __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_3, 9, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_overlapping);
    __pyx_t_4 += 14;
    __Pyx_GIVEREF(__pyx_kp_u_overlapping);
    PyTuple_SET_ITEM(__pyx_t_3, 10, __pyx_kp_u_overlapping);

    /* "pandas/_libs/intervaltree.pxi":1062
 *             n_right = self.right_node.n_elements
 *             n_center = self.n_elements - n_left - n_right
 *             return ('<Float64ClosedLeftIntervalNode: '             # <<<<<<<<<<<<<<
 *                     'pivot %s, %s elements (%s left, %s right, %s '
 *                     'overlapping)>' % (self.pivot, self.n_elements,
 */
    __pyx_t_2 = __Pyx_PyUnicode_Join(__pyx_t_3, 11, __pyx_t_4, __pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1062, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;
  }

  /* "pandas/_libs/intervaltree.pxi":1054
 *                 result.extend(self.center_left_indices)
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         if self.is_leaf_node:
 *             return ('<Float64ClosedLeftIntervalNode: '
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pandas._libs.interval.Float64ClosedLeftIntervalNode.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_n_left);
  __Pyx_XDECREF(__pyx_v_n_right);
  __Pyx_XDECREF(__pyx_v_n_center);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":1067
 *                                        n_left, n_right, n_center))
 * 
 *     def counts(self):             # <<<<<<<<<<<<<<
 *         """
 *         Inspect counts on this node
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_7counts(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_6counts[] = "\n        Inspect counts on this node\n        useful for debugging purposes\n        ";
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_7counts(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("counts (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_6counts(((struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_6counts(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode *__pyx_v_self) {
  Py_ssize_t __pyx_v_m;
  PyObject *__pyx_v_l = NULL;
  PyObject *__pyx_v_r = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  size_t __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("counts", 0);

  /* "pandas/_libs/intervaltree.pxi":1072
 *         useful for debugging purposes
 *         """
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             return self.n_elements
 *         else:
 */
  __pyx_t_1 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_1) {

    /* "pandas/_libs/intervaltree.pxi":1073
 *         """
 *         if self.is_leaf_node:
 *             return self.n_elements             # <<<<<<<<<<<<<<
 *         else:
 *             m = len(self.center_left_values)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_elements); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1073, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "pandas/_libs/intervaltree.pxi":1072
 *         useful for debugging purposes
 *         """
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             return self.n_elements
 *         else:
 */
  }

  /* "pandas/_libs/intervaltree.pxi":1075
 *             return self.n_elements
 *         else:
 *             m = len(self.center_left_values)             # <<<<<<<<<<<<<<
 *             l = self.left_node.counts()
 *             r = self.right_node.counts()
 */
  /*else*/ {
    if (unlikely(!__pyx_v_self->center_left_values.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(1, 1075, __pyx_L1_error)}
    __pyx_t_3 = __Pyx_MemoryView_Len(__pyx_v_self->center_left_values); 
    __pyx_v_m = __pyx_t_3;

    /* "pandas/_libs/intervaltree.pxi":1076
 *         else:
 *             m = len(self.center_left_values)
 *             l = self.left_node.counts()             # <<<<<<<<<<<<<<
 *             r = self.right_node.counts()
 *             return (m, (l, r))
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->left_node), __pyx_n_s_counts); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1076, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1076, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_l = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "pandas/_libs/intervaltree.pxi":1077
 *             m = len(self.center_left_values)
 *             l = self.left_node.counts()
 *             r = self.right_node.counts()             # <<<<<<<<<<<<<<
 *             return (m, (l, r))
 * 
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->right_node), __pyx_n_s_counts); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1077, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1077, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_r = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "pandas/_libs/intervaltree.pxi":1078
 *             l = self.left_node.counts()
 *             r = self.right_node.counts()
 *             return (m, (l, r))             # <<<<<<<<<<<<<<
 * 
 * NODE_CLASSES['float64',
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = PyInt_FromSsize_t(__pyx_v_m); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1078, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1078, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_v_l);
    __Pyx_GIVEREF(__pyx_v_l);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_l);
    __Pyx_INCREF(__pyx_v_r);
    __Pyx_GIVEREF(__pyx_v_r);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_v_r);
    __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1078, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_4);
    __pyx_t_2 = 0;
    __pyx_t_4 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;
  }

  /* "pandas/_libs/intervaltree.pxi":1067
 *                                        n_left, n_right, n_center))
 * 
 *     def counts(self):             # <<<<<<<<<<<<<<
 *         """
 *         Inspect counts on this node
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pandas._libs.interval.Float64ClosedLeftIntervalNode.counts", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_l);
  __Pyx_XDECREF(__pyx_v_r);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":924
 *         int64_t[:] center_left_indices, center_right_indices, indices
 *         float64_t min_left, max_right
 *         readonly float64_t pivot             # <<<<<<<<<<<<<<
 *         readonly int64_t n_elements, n_center, leaf_size
 *         readonly bint is_leaf_node
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_5pivot_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_5pivot_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_5pivot___get__(((struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_5pivot___get__(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->pivot); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 924, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Float64ClosedLeftIntervalNode.pivot.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":925
 *         float64_t min_left, max_right
 *         readonly float64_t pivot
 *         readonly int64_t n_elements, n_center, leaf_size             # <<<<<<<<<<<<<<
 *         readonly bint is_leaf_node
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_10n_elements_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_10n_elements_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_10n_elements___get__(((struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_10n_elements___get__(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_elements); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 925, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Float64ClosedLeftIntervalNode.n_elements.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_8n_center_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_8n_center_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_8n_center___get__(((struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_8n_center___get__(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_center); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 925, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Float64ClosedLeftIntervalNode.n_center.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_9leaf_size_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_9leaf_size_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_9leaf_size___get__(((struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_9leaf_size___get__(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->leaf_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 925, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Float64ClosedLeftIntervalNode.leaf_size.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":926
 *         readonly float64_t pivot
 *         readonly int64_t n_elements, n_center, leaf_size
 *         readonly bint is_leaf_node             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self,
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_12is_leaf_node_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_12is_leaf_node_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_12is_leaf_node___get__(((struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_12is_leaf_node___get__(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->is_leaf_node); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 926, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Float64ClosedLeftIntervalNode.is_leaf_node.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_21__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_21__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_20__reduce_cython__(((struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_20__reduce_cython__(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  int __pyx_t_16;
  int __pyx_t_17;
  int __pyx_t_18;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.center_left_indices, self.center_left_values, self.center_right_indices, self.center_right_values, self.indices, self.is_leaf_node, self.leaf_size, self.left, self.left_node, self.max_right, self.min_left, self.n_center, self.n_elements, self.pivot, self.right, self.right_node)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  if (unlikely(!__pyx_v_self->center_left_indices.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_1 = __pyx_memoryview_fromslice(__pyx_v_self->center_left_indices, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_int64_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_int64_t, 0);; if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (unlikely(!__pyx_v_self->center_left_values.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_2 = __pyx_memoryview_fromslice(__pyx_v_self->center_left_values, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_float64_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_float64_t, 0);; if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (unlikely(!__pyx_v_self->center_right_indices.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_3 = __pyx_memoryview_fromslice(__pyx_v_self->center_right_indices, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_int64_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_int64_t, 0);; if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely(!__pyx_v_self->center_right_values.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_4 = __pyx_memoryview_fromslice(__pyx_v_self->center_right_values, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_float64_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_float64_t, 0);; if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (unlikely(!__pyx_v_self->indices.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_5 = __pyx_memoryview_fromslice(__pyx_v_self->indices, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_int64_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_int64_t, 0);; if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyBool_FromLong(__pyx_v_self->is_leaf_node); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->leaf_size); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (unlikely(!__pyx_v_self->left.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_8 = __pyx_memoryview_fromslice(__pyx_v_self->left, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_float64_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_float64_t, 0);; if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = PyFloat_FromDouble(__pyx_v_self->max_right); if (unlikely(!__pyx_t_9)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_10 = PyFloat_FromDouble(__pyx_v_self->min_left); if (unlikely(!__pyx_t_10)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_11 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_center); if (unlikely(!__pyx_t_11)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_12 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_elements); if (unlikely(!__pyx_t_12)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __pyx_t_13 = PyFloat_FromDouble(__pyx_v_self->pivot); if (unlikely(!__pyx_t_13)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  if (unlikely(!__pyx_v_self->right.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_14 = __pyx_memoryview_fromslice(__pyx_v_self->right, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_float64_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_float64_t, 0);; if (unlikely(!__pyx_t_14)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __pyx_t_15 = PyTuple_New(16); if (unlikely(!__pyx_t_15)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_15, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_15, 1, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_15, 2, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_15, 3, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_15, 4, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_15, 5, __pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_15, 6, __pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_15, 7, __pyx_t_8);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->left_node));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self->left_node));
  PyTuple_SET_ITEM(__pyx_t_15, 8, ((PyObject *)__pyx_v_self->left_node));
  __Pyx_GIVEREF(__pyx_t_9);
  PyTuple_SET_ITEM(__pyx_t_15, 9, __pyx_t_9);
  __Pyx_GIVEREF(__pyx_t_10);
  PyTuple_SET_ITEM(__pyx_t_15, 10, __pyx_t_10);
  __Pyx_GIVEREF(__pyx_t_11);
  PyTuple_SET_ITEM(__pyx_t_15, 11, __pyx_t_11);
  __Pyx_GIVEREF(__pyx_t_12);
  PyTuple_SET_ITEM(__pyx_t_15, 12, __pyx_t_12);
  __Pyx_GIVEREF(__pyx_t_13);
  PyTuple_SET_ITEM(__pyx_t_15, 13, __pyx_t_13);
  __Pyx_GIVEREF(__pyx_t_14);
  PyTuple_SET_ITEM(__pyx_t_15, 14, __pyx_t_14);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->right_node));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self->right_node));
  PyTuple_SET_ITEM(__pyx_t_15, 15, ((PyObject *)__pyx_v_self->right_node));
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  __pyx_t_6 = 0;
  __pyx_t_7 = 0;
  __pyx_t_8 = 0;
  __pyx_t_9 = 0;
  __pyx_t_10 = 0;
  __pyx_t_11 = 0;
  __pyx_t_12 = 0;
  __pyx_t_13 = 0;
  __pyx_t_14 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_15);
  __pyx_t_15 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.center_left_indices, self.center_left_values, self.center_right_indices, self.center_right_values, self.indices, self.is_leaf_node, self.leaf_size, self.left, self.left_node, self.max_right, self.min_left, self.n_center, self.n_elements, self.pivot, self.right, self.right_node)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_15 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_15)) __PYX_ERR(2, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __pyx_v__dict = __pyx_t_15;
  __pyx_t_15 = 0;

  /* "(tree fragment)":7
 *     state = (self.center_left_indices, self.center_left_values, self.center_right_indices, self.center_right_values, self.indices, self.is_leaf_node, self.leaf_size, self.left, self.left_node, self.max_right, self.min_left, self.n_center, self.n_elements, self.pivot, self.right, self.right_node)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_16 = (__pyx_v__dict != Py_None);
  __pyx_t_17 = (__pyx_t_16 != 0);
  if (__pyx_t_17) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_15 = PyTuple_New(1); if (unlikely(!__pyx_t_15)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    PyTuple_SET_ITEM(__pyx_t_15, 0, __pyx_v__dict);
    __pyx_t_14 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_15); if (unlikely(!__pyx_t_14)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_14));
    __pyx_t_14 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.left_node is not None or self.right_node is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.center_left_indices, self.center_left_values, self.center_right_indices, self.center_right_values, self.indices, self.is_leaf_node, self.leaf_size, self.left, self.left_node, self.max_right, self.min_left, self.n_center, self.n_elements, self.pivot, self.right, self.right_node)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.left_node is not None or self.right_node is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_Float64ClosedLeftIntervalNode, (type(self), 0x7a104e5, None), state
 */
  /*else*/ {
    __pyx_t_16 = (((PyObject *)__pyx_v_self->left_node) != Py_None);
    __pyx_t_18 = (__pyx_t_16 != 0);
    if (!__pyx_t_18) {
    } else {
      __pyx_t_17 = __pyx_t_18;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_18 = (((PyObject *)__pyx_v_self->right_node) != Py_None);
    __pyx_t_16 = (__pyx_t_18 != 0);
    __pyx_t_17 = __pyx_t_16;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_17;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.left_node is not None or self.right_node is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Float64ClosedLeftIntervalNode, (type(self), 0x7a104e5, None), state
 *     else:
 */
  __pyx_t_17 = (__pyx_v_use_setstate != 0);
  if (__pyx_t_17) {

    /* "(tree fragment)":13
 *         use_setstate = self.left_node is not None or self.right_node is not None
 *     if use_setstate:
 *         return __pyx_unpickle_Float64ClosedLeftIntervalNode, (type(self), 0x7a104e5, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_Float64ClosedLeftIntervalNode, (type(self), 0x7a104e5, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_14, __pyx_n_s_pyx_unpickle_Float64ClosedLeft); if (unlikely(!__pyx_t_14)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __pyx_t_15 = PyTuple_New(3); if (unlikely(!__pyx_t_15)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_15, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_127993061);
    __Pyx_GIVEREF(__pyx_int_127993061);
    PyTuple_SET_ITEM(__pyx_t_15, 1, __pyx_int_127993061);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_15, 2, Py_None);
    __pyx_t_13 = PyTuple_New(3); if (unlikely(!__pyx_t_13)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __Pyx_GIVEREF(__pyx_t_14);
    PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_t_14);
    __Pyx_GIVEREF(__pyx_t_15);
    PyTuple_SET_ITEM(__pyx_t_13, 1, __pyx_t_15);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_13, 2, __pyx_v_state);
    __pyx_t_14 = 0;
    __pyx_t_15 = 0;
    __pyx_r = __pyx_t_13;
    __pyx_t_13 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.left_node is not None or self.right_node is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Float64ClosedLeftIntervalNode, (type(self), 0x7a104e5, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_Float64ClosedLeftIntervalNode, (type(self), 0x7a104e5, None), state
 *     else:
 *         return __pyx_unpickle_Float64ClosedLeftIntervalNode, (type(self), 0x7a104e5, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Float64ClosedLeftIntervalNode__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_13, __pyx_n_s_pyx_unpickle_Float64ClosedLeft); if (unlikely(!__pyx_t_13)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __pyx_t_15 = PyTuple_New(3); if (unlikely(!__pyx_t_15)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_15, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_127993061);
    __Pyx_GIVEREF(__pyx_int_127993061);
    PyTuple_SET_ITEM(__pyx_t_15, 1, __pyx_int_127993061);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_15, 2, __pyx_v_state);
    __pyx_t_14 = PyTuple_New(2); if (unlikely(!__pyx_t_14)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_GIVEREF(__pyx_t_13);
    PyTuple_SET_ITEM(__pyx_t_14, 0, __pyx_t_13);
    __Pyx_GIVEREF(__pyx_t_15);
    PyTuple_SET_ITEM(__pyx_t_14, 1, __pyx_t_15);
    __pyx_t_13 = 0;
    __pyx_t_15 = 0;
    __pyx_r = __pyx_t_14;
    __pyx_t_14 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_AddTraceback("pandas._libs.interval.Float64ClosedLeftIntervalNode.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Float64ClosedLeftIntervalNode, (type(self), 0x7a104e5, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Float64ClosedLeftIntervalNode__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_23__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_23__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_22__setstate_cython__(((struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_22__setstate_cython__(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_Float64ClosedLeftIntervalNode, (type(self), 0x7a104e5, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Float64ClosedLeftIntervalNode__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(2, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_6pandas_5_libs_8interval___pyx_unpickle_Float64ClosedLeftIntervalNode__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Float64ClosedLeftIntervalNode, (type(self), 0x7a104e5, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Float64ClosedLeftIntervalNode__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Float64ClosedLeftIntervalNode.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":1098
 *         readonly bint is_leaf_node
 * 
 *     def __init__(self,             # <<<<<<<<<<<<<<
 *                  ndarray[float64_t, ndim=1] left,
 *                  ndarray[float64_t, ndim=1] right,
 */

/* Python wrapper */
static int __pyx_pw_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyArrayObject *__pyx_v_left = 0;
  PyArrayObject *__pyx_v_right = 0;
  PyArrayObject *__pyx_v_indices = 0;
  __pyx_t_5numpy_int64_t __pyx_v_leaf_size;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_left,&__pyx_n_s_right,&__pyx_n_s_indices,&__pyx_n_s_leaf_size,0};
    PyObject* values[4] = {0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_left)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_right)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 4, 4, 1); __PYX_ERR(1, 1098, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_indices)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 4, 4, 2); __PYX_ERR(1, 1098, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_leaf_size)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 4, 4, 3); __PYX_ERR(1, 1098, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(1, 1098, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 4) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
    }
    __pyx_v_left = ((PyArrayObject *)values[0]);
    __pyx_v_right = ((PyArrayObject *)values[1]);
    __pyx_v_indices = ((PyArrayObject *)values[2]);
    __pyx_v_leaf_size = __Pyx_PyInt_As_npy_int64(values[3]); if (unlikely((__pyx_v_leaf_size == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(1, 1102, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 4, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 1098, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Float64ClosedRightIntervalNode.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_left), __pyx_ptype_5numpy_ndarray, 1, "left", 0))) __PYX_ERR(1, 1099, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_right), __pyx_ptype_5numpy_ndarray, 1, "right", 0))) __PYX_ERR(1, 1100, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_indices), __pyx_ptype_5numpy_ndarray, 1, "indices", 0))) __PYX_ERR(1, 1101, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode___init__(((struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode *)__pyx_v_self), __pyx_v_left, __pyx_v_right, __pyx_v_indices, __pyx_v_leaf_size);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode___init__(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode *__pyx_v_self, PyArrayObject *__pyx_v_left, PyArrayObject *__pyx_v_right, PyArrayObject *__pyx_v_indices, __pyx_t_5numpy_int64_t __pyx_v_leaf_size) {
  PyObject *__pyx_v_left_set = NULL;
  PyObject *__pyx_v_right_set = NULL;
  PyObject *__pyx_v_center_set = NULL;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_indices;
  __Pyx_Buffer __pyx_pybuffer_indices;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_left;
  __Pyx_Buffer __pyx_pybuffer_left;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_right;
  __Pyx_Buffer __pyx_pybuffer_right;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __pyx_t_5numpy_float64_t __pyx_t_6;
  __Pyx_memviewslice __pyx_t_7 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_8 = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  __Pyx_memviewslice __pyx_t_11 = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *(*__pyx_t_12)(PyObject *);
  size_t __pyx_t_13;
  __Pyx_RefNannySetupContext("__init__", 0);
  __pyx_pybuffer_left.pybuffer.buf = NULL;
  __pyx_pybuffer_left.refcount = 0;
  __pyx_pybuffernd_left.data = NULL;
  __pyx_pybuffernd_left.rcbuffer = &__pyx_pybuffer_left;
  __pyx_pybuffer_right.pybuffer.buf = NULL;
  __pyx_pybuffer_right.refcount = 0;
  __pyx_pybuffernd_right.data = NULL;
  __pyx_pybuffernd_right.rcbuffer = &__pyx_pybuffer_right;
  __pyx_pybuffer_indices.pybuffer.buf = NULL;
  __pyx_pybuffer_indices.refcount = 0;
  __pyx_pybuffernd_indices.data = NULL;
  __pyx_pybuffernd_indices.rcbuffer = &__pyx_pybuffer_indices;
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_left.rcbuffer->pybuffer, (PyObject*)__pyx_v_left, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 1098, __pyx_L1_error)
  }
  __pyx_pybuffernd_left.diminfo[0].strides = __pyx_pybuffernd_left.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_left.diminfo[0].shape = __pyx_pybuffernd_left.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_right.rcbuffer->pybuffer, (PyObject*)__pyx_v_right, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 1098, __pyx_L1_error)
  }
  __pyx_pybuffernd_right.diminfo[0].strides = __pyx_pybuffernd_right.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_right.diminfo[0].shape = __pyx_pybuffernd_right.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_indices.rcbuffer->pybuffer, (PyObject*)__pyx_v_indices, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 1098, __pyx_L1_error)
  }
  __pyx_pybuffernd_indices.diminfo[0].strides = __pyx_pybuffernd_indices.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_indices.diminfo[0].shape = __pyx_pybuffernd_indices.rcbuffer->pybuffer.shape[0];

  /* "pandas/_libs/intervaltree.pxi":1104
 *                  int64_t leaf_size):
 * 
 *         self.n_elements = len(left)             # <<<<<<<<<<<<<<
 *         self.leaf_size = leaf_size
 * 
 */
  __pyx_t_1 = PyObject_Length(((PyObject *)__pyx_v_left)); if (unlikely(__pyx_t_1 == ((Py_ssize_t)-1))) __PYX_ERR(1, 1104, __pyx_L1_error)
  __pyx_v_self->n_elements = __pyx_t_1;

  /* "pandas/_libs/intervaltree.pxi":1105
 * 
 *         self.n_elements = len(left)
 *         self.leaf_size = leaf_size             # <<<<<<<<<<<<<<
 * 
 *         # min_left and min_right are used to speed-up query by skipping
 */
  __pyx_v_self->leaf_size = __pyx_v_leaf_size;

  /* "pandas/_libs/intervaltree.pxi":1110
 *         # query on sub-nodes. If this node has size 0, query is cheap,
 *         # so these values don't matter.
 *         if left.size > 0:             # <<<<<<<<<<<<<<
 *             self.min_left = left.min()
 *             self.max_right = right.max()
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_left), __pyx_n_s_size); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1110, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyObject_RichCompare(__pyx_t_2, __pyx_int_0, Py_GT); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1110, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(1, 1110, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_4) {

    /* "pandas/_libs/intervaltree.pxi":1111
 *         # so these values don't matter.
 *         if left.size > 0:
 *             self.min_left = left.min()             # <<<<<<<<<<<<<<
 *             self.max_right = right.max()
 *         else:
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_left), __pyx_n_s_min); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1111, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1111, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_6 = __pyx_PyFloat_AsDouble(__pyx_t_3); if (unlikely((__pyx_t_6 == ((npy_float64)-1)) && PyErr_Occurred())) __PYX_ERR(1, 1111, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_self->min_left = __pyx_t_6;

    /* "pandas/_libs/intervaltree.pxi":1112
 *         if left.size > 0:
 *             self.min_left = left.min()
 *             self.max_right = right.max()             # <<<<<<<<<<<<<<
 *         else:
 *             self.min_left = 0
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_right), __pyx_n_s_max); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1112, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1112, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_6 = __pyx_PyFloat_AsDouble(__pyx_t_3); if (unlikely((__pyx_t_6 == ((npy_float64)-1)) && PyErr_Occurred())) __PYX_ERR(1, 1112, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_self->max_right = __pyx_t_6;

    /* "pandas/_libs/intervaltree.pxi":1110
 *         # query on sub-nodes. If this node has size 0, query is cheap,
 *         # so these values don't matter.
 *         if left.size > 0:             # <<<<<<<<<<<<<<
 *             self.min_left = left.min()
 *             self.max_right = right.max()
 */
    goto __pyx_L3;
  }

  /* "pandas/_libs/intervaltree.pxi":1114
 *             self.max_right = right.max()
 *         else:
 *             self.min_left = 0             # <<<<<<<<<<<<<<
 *             self.max_right = 0
 * 
 */
  /*else*/ {
    __pyx_v_self->min_left = 0.0;

    /* "pandas/_libs/intervaltree.pxi":1115
 *         else:
 *             self.min_left = 0
 *             self.max_right = 0             # <<<<<<<<<<<<<<
 * 
 *         if self.n_elements <= leaf_size:
 */
    __pyx_v_self->max_right = 0.0;
  }
  __pyx_L3:;

  /* "pandas/_libs/intervaltree.pxi":1117
 *             self.max_right = 0
 * 
 *         if self.n_elements <= leaf_size:             # <<<<<<<<<<<<<<
 *             # make this a terminal (leaf) node
 *             self.is_leaf_node = True
 */
  __pyx_t_4 = ((__pyx_v_self->n_elements <= __pyx_v_leaf_size) != 0);
  if (__pyx_t_4) {

    /* "pandas/_libs/intervaltree.pxi":1119
 *         if self.n_elements <= leaf_size:
 *             # make this a terminal (leaf) node
 *             self.is_leaf_node = True             # <<<<<<<<<<<<<<
 *             self.left = left
 *             self.right = right
 */
    __pyx_v_self->is_leaf_node = 1;

    /* "pandas/_libs/intervaltree.pxi":1120
 *             # make this a terminal (leaf) node
 *             self.is_leaf_node = True
 *             self.left = left             # <<<<<<<<<<<<<<
 *             self.right = right
 *             self.indices = indices
 */
    __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_float64_t(((PyObject *)__pyx_v_left), PyBUF_WRITABLE); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(1, 1120, __pyx_L1_error)
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->left, 0);
    __pyx_v_self->left = __pyx_t_7;
    __pyx_t_7.memview = NULL;
    __pyx_t_7.data = NULL;

    /* "pandas/_libs/intervaltree.pxi":1121
 *             self.is_leaf_node = True
 *             self.left = left
 *             self.right = right             # <<<<<<<<<<<<<<
 *             self.indices = indices
 *             self.n_center = 0
 */
    __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_float64_t(((PyObject *)__pyx_v_right), PyBUF_WRITABLE); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(1, 1121, __pyx_L1_error)
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->right, 0);
    __pyx_v_self->right = __pyx_t_7;
    __pyx_t_7.memview = NULL;
    __pyx_t_7.data = NULL;

    /* "pandas/_libs/intervaltree.pxi":1122
 *             self.left = left
 *             self.right = right
 *             self.indices = indices             # <<<<<<<<<<<<<<
 *             self.n_center = 0
 *         else:
 */
    __pyx_t_8 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(((PyObject *)__pyx_v_indices), PyBUF_WRITABLE); if (unlikely(!__pyx_t_8.memview)) __PYX_ERR(1, 1122, __pyx_L1_error)
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->indices, 0);
    __pyx_v_self->indices = __pyx_t_8;
    __pyx_t_8.memview = NULL;
    __pyx_t_8.data = NULL;

    /* "pandas/_libs/intervaltree.pxi":1123
 *             self.right = right
 *             self.indices = indices
 *             self.n_center = 0             # <<<<<<<<<<<<<<
 *         else:
 *             # calculate a pivot so we can create child nodes
 */
    __pyx_v_self->n_center = 0;

    /* "pandas/_libs/intervaltree.pxi":1117
 *             self.max_right = 0
 * 
 *         if self.n_elements <= leaf_size:             # <<<<<<<<<<<<<<
 *             # make this a terminal (leaf) node
 *             self.is_leaf_node = True
 */
    goto __pyx_L4;
  }

  /* "pandas/_libs/intervaltree.pxi":1126
 *         else:
 *             # calculate a pivot so we can create child nodes
 *             self.is_leaf_node = False             # <<<<<<<<<<<<<<
 *             self.pivot = np.median(left / 2 + right / 2)
 *             left_set, right_set, center_set = self.classify_intervals(
 */
  /*else*/ {
    __pyx_v_self->is_leaf_node = 0;

    /* "pandas/_libs/intervaltree.pxi":1127
 *             # calculate a pivot so we can create child nodes
 *             self.is_leaf_node = False
 *             self.pivot = np.median(left / 2 + right / 2)             # <<<<<<<<<<<<<<
 *             left_set, right_set, center_set = self.classify_intervals(
 *                 left, right)
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1127, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_median); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1127, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyNumber_Divide(((PyObject *)__pyx_v_left), __pyx_int_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1127, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_9 = __Pyx_PyNumber_Divide(((PyObject *)__pyx_v_right), __pyx_int_2); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 1127, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_10 = PyNumber_Add(__pyx_t_2, __pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 1127, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_9 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_3 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_9, __pyx_t_10) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_10);
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1127, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_6 = __pyx_PyFloat_AsDouble(__pyx_t_3); if (unlikely((__pyx_t_6 == ((npy_float64)-1)) && PyErr_Occurred())) __PYX_ERR(1, 1127, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_self->pivot = __pyx_t_6;

    /* "pandas/_libs/intervaltree.pxi":1129
 *             self.pivot = np.median(left / 2 + right / 2)
 *             left_set, right_set, center_set = self.classify_intervals(
 *                 left, right)             # <<<<<<<<<<<<<<
 * 
 *             self.left_node = self.new_child_node(left, right,
 */
    __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_float64_t(((PyObject *)__pyx_v_left), PyBUF_WRITABLE); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(1, 1129, __pyx_L1_error)
    __pyx_t_11 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_float64_t(((PyObject *)__pyx_v_right), PyBUF_WRITABLE); if (unlikely(!__pyx_t_11.memview)) __PYX_ERR(1, 1129, __pyx_L1_error)

    /* "pandas/_libs/intervaltree.pxi":1128
 *             self.is_leaf_node = False
 *             self.pivot = np.median(left / 2 + right / 2)
 *             left_set, right_set, center_set = self.classify_intervals(             # <<<<<<<<<<<<<<
 *                 left, right)
 * 
 */
    __pyx_t_3 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode *)__pyx_v_self->__pyx_vtab)->classify_intervals(__pyx_v_self, __pyx_t_7, __pyx_t_11); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1128, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __PYX_XDEC_MEMVIEW(&__pyx_t_7, 1);
    __pyx_t_7.memview = NULL;
    __pyx_t_7.data = NULL;
    __PYX_XDEC_MEMVIEW(&__pyx_t_11, 1);
    __pyx_t_11.memview = NULL;
    __pyx_t_11.data = NULL;
    if ((likely(PyTuple_CheckExact(__pyx_t_3))) || (PyList_CheckExact(__pyx_t_3))) {
      PyObject* sequence = __pyx_t_3;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 3)) {
        if (size > 3) __Pyx_RaiseTooManyValuesError(3);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(1, 1128, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_5 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_10 = PyTuple_GET_ITEM(sequence, 1); 
        __pyx_t_9 = PyTuple_GET_ITEM(sequence, 2); 
      } else {
        __pyx_t_5 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_10 = PyList_GET_ITEM(sequence, 1); 
        __pyx_t_9 = PyList_GET_ITEM(sequence, 2); 
      }
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_10);
      __Pyx_INCREF(__pyx_t_9);
      #else
      __pyx_t_5 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1128, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_10 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 1128, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_9 = PySequence_ITEM(sequence, 2); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 1128, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      #endif
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_2 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1128, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_12 = Py_TYPE(__pyx_t_2)->tp_iternext;
      index = 0; __pyx_t_5 = __pyx_t_12(__pyx_t_2); if (unlikely(!__pyx_t_5)) goto __pyx_L5_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_5);
      index = 1; __pyx_t_10 = __pyx_t_12(__pyx_t_2); if (unlikely(!__pyx_t_10)) goto __pyx_L5_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_10);
      index = 2; __pyx_t_9 = __pyx_t_12(__pyx_t_2); if (unlikely(!__pyx_t_9)) goto __pyx_L5_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_9);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_12(__pyx_t_2), 3) < 0) __PYX_ERR(1, 1128, __pyx_L1_error)
      __pyx_t_12 = NULL;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      goto __pyx_L6_unpacking_done;
      __pyx_L5_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_12 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(1, 1128, __pyx_L1_error)
      __pyx_L6_unpacking_done:;
    }
    __pyx_v_left_set = __pyx_t_5;
    __pyx_t_5 = 0;
    __pyx_v_right_set = __pyx_t_10;
    __pyx_t_10 = 0;
    __pyx_v_center_set = __pyx_t_9;
    __pyx_t_9 = 0;

    /* "pandas/_libs/intervaltree.pxi":1132
 * 
 *             self.left_node = self.new_child_node(left, right,
 *                                                  indices, left_set)             # <<<<<<<<<<<<<<
 *             self.right_node = self.new_child_node(left, right,
 *                                                   indices, right_set)
 */
    if (!(likely(((__pyx_v_left_set) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_left_set, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(1, 1132, __pyx_L1_error)

    /* "pandas/_libs/intervaltree.pxi":1131
 *                 left, right)
 * 
 *             self.left_node = self.new_child_node(left, right,             # <<<<<<<<<<<<<<
 *                                                  indices, left_set)
 *             self.right_node = self.new_child_node(left, right,
 */
    __pyx_t_3 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode *)__pyx_v_self->__pyx_vtab)->new_child_node(__pyx_v_self, ((PyArrayObject *)__pyx_v_left), ((PyArrayObject *)__pyx_v_right), ((PyArrayObject *)__pyx_v_indices), ((PyArrayObject *)__pyx_v_left_set)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1131, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode))))) __PYX_ERR(1, 1131, __pyx_L1_error)
    __Pyx_GIVEREF(__pyx_t_3);
    __Pyx_GOTREF(__pyx_v_self->left_node);
    __Pyx_DECREF(((PyObject *)__pyx_v_self->left_node));
    __pyx_v_self->left_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode *)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "pandas/_libs/intervaltree.pxi":1134
 *                                                  indices, left_set)
 *             self.right_node = self.new_child_node(left, right,
 *                                                   indices, right_set)             # <<<<<<<<<<<<<<
 * 
 *             self.center_left_values, self.center_left_indices = \
 */
    if (!(likely(((__pyx_v_right_set) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_right_set, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(1, 1134, __pyx_L1_error)

    /* "pandas/_libs/intervaltree.pxi":1133
 *             self.left_node = self.new_child_node(left, right,
 *                                                  indices, left_set)
 *             self.right_node = self.new_child_node(left, right,             # <<<<<<<<<<<<<<
 *                                                   indices, right_set)
 * 
 */
    __pyx_t_3 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode *)__pyx_v_self->__pyx_vtab)->new_child_node(__pyx_v_self, ((PyArrayObject *)__pyx_v_left), ((PyArrayObject *)__pyx_v_right), ((PyArrayObject *)__pyx_v_indices), ((PyArrayObject *)__pyx_v_right_set)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1133, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode))))) __PYX_ERR(1, 1133, __pyx_L1_error)
    __Pyx_GIVEREF(__pyx_t_3);
    __Pyx_GOTREF(__pyx_v_self->right_node);
    __Pyx_DECREF(((PyObject *)__pyx_v_self->right_node));
    __pyx_v_self->right_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode *)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "pandas/_libs/intervaltree.pxi":1137
 * 
 *             self.center_left_values, self.center_left_indices = \
 *                 sort_values_and_indices(left, indices, center_set)             # <<<<<<<<<<<<<<
 *             self.center_right_values, self.center_right_indices = \
 *                 sort_values_and_indices(right, indices, center_set)
 */
    __pyx_t_3 = __pyx_f_6pandas_5_libs_8interval_sort_values_and_indices(((PyObject *)__pyx_v_left), ((PyObject *)__pyx_v_indices), __pyx_v_center_set); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1137, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if ((likely(PyTuple_CheckExact(__pyx_t_3))) || (PyList_CheckExact(__pyx_t_3))) {
      PyObject* sequence = __pyx_t_3;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(1, 1136, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_9 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_10 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_9 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_10 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_9);
      __Pyx_INCREF(__pyx_t_10);
      #else
      __pyx_t_9 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 1136, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_10 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 1136, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      #endif
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_5 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1136, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_12 = Py_TYPE(__pyx_t_5)->tp_iternext;
      index = 0; __pyx_t_9 = __pyx_t_12(__pyx_t_5); if (unlikely(!__pyx_t_9)) goto __pyx_L7_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_9);
      index = 1; __pyx_t_10 = __pyx_t_12(__pyx_t_5); if (unlikely(!__pyx_t_10)) goto __pyx_L7_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_10);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_12(__pyx_t_5), 2) < 0) __PYX_ERR(1, 1136, __pyx_L1_error)
      __pyx_t_12 = NULL;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      goto __pyx_L8_unpacking_done;
      __pyx_L7_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_12 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(1, 1136, __pyx_L1_error)
      __pyx_L8_unpacking_done:;
    }

    /* "pandas/_libs/intervaltree.pxi":1136
 *                                                   indices, right_set)
 * 
 *             self.center_left_values, self.center_left_indices = \             # <<<<<<<<<<<<<<
 *                 sort_values_and_indices(left, indices, center_set)
 *             self.center_right_values, self.center_right_indices = \
 */
    __pyx_t_11 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_float64_t(__pyx_t_9, PyBUF_WRITABLE); if (unlikely(!__pyx_t_11.memview)) __PYX_ERR(1, 1136, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_8 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_t_10, PyBUF_WRITABLE); if (unlikely(!__pyx_t_8.memview)) __PYX_ERR(1, 1136, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->center_left_values, 0);
    __pyx_v_self->center_left_values = __pyx_t_11;
    __pyx_t_11.memview = NULL;
    __pyx_t_11.data = NULL;
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->center_left_indices, 0);
    __pyx_v_self->center_left_indices = __pyx_t_8;
    __pyx_t_8.memview = NULL;
    __pyx_t_8.data = NULL;

    /* "pandas/_libs/intervaltree.pxi":1139
 *                 sort_values_and_indices(left, indices, center_set)
 *             self.center_right_values, self.center_right_indices = \
 *                 sort_values_and_indices(right, indices, center_set)             # <<<<<<<<<<<<<<
 *             self.n_center = len(self.center_left_indices)
 * 
 */
    __pyx_t_3 = __pyx_f_6pandas_5_libs_8interval_sort_values_and_indices(((PyObject *)__pyx_v_right), ((PyObject *)__pyx_v_indices), __pyx_v_center_set); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1139, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if ((likely(PyTuple_CheckExact(__pyx_t_3))) || (PyList_CheckExact(__pyx_t_3))) {
      PyObject* sequence = __pyx_t_3;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(1, 1138, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_10 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_9 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_10 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_9 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_10);
      __Pyx_INCREF(__pyx_t_9);
      #else
      __pyx_t_10 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 1138, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_9 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 1138, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      #endif
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_5 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1138, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_12 = Py_TYPE(__pyx_t_5)->tp_iternext;
      index = 0; __pyx_t_10 = __pyx_t_12(__pyx_t_5); if (unlikely(!__pyx_t_10)) goto __pyx_L9_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_10);
      index = 1; __pyx_t_9 = __pyx_t_12(__pyx_t_5); if (unlikely(!__pyx_t_9)) goto __pyx_L9_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_9);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_12(__pyx_t_5), 2) < 0) __PYX_ERR(1, 1138, __pyx_L1_error)
      __pyx_t_12 = NULL;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      goto __pyx_L10_unpacking_done;
      __pyx_L9_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_12 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(1, 1138, __pyx_L1_error)
      __pyx_L10_unpacking_done:;
    }

    /* "pandas/_libs/intervaltree.pxi":1138
 *             self.center_left_values, self.center_left_indices = \
 *                 sort_values_and_indices(left, indices, center_set)
 *             self.center_right_values, self.center_right_indices = \             # <<<<<<<<<<<<<<
 *                 sort_values_and_indices(right, indices, center_set)
 *             self.n_center = len(self.center_left_indices)
 */
    __pyx_t_11 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_float64_t(__pyx_t_10, PyBUF_WRITABLE); if (unlikely(!__pyx_t_11.memview)) __PYX_ERR(1, 1138, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_t_8 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_t_9, PyBUF_WRITABLE); if (unlikely(!__pyx_t_8.memview)) __PYX_ERR(1, 1138, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->center_right_values, 0);
    __pyx_v_self->center_right_values = __pyx_t_11;
    __pyx_t_11.memview = NULL;
    __pyx_t_11.data = NULL;
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->center_right_indices, 0);
    __pyx_v_self->center_right_indices = __pyx_t_8;
    __pyx_t_8.memview = NULL;
    __pyx_t_8.data = NULL;

    /* "pandas/_libs/intervaltree.pxi":1140
 *             self.center_right_values, self.center_right_indices = \
 *                 sort_values_and_indices(right, indices, center_set)
 *             self.n_center = len(self.center_left_indices)             # <<<<<<<<<<<<<<
 * 
 *     @cython.wraparound(False)
 */
    if (unlikely(!__pyx_v_self->center_left_indices.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(1, 1140, __pyx_L1_error)}
    __pyx_t_13 = __Pyx_MemoryView_Len(__pyx_v_self->center_left_indices); 
    __pyx_v_self->n_center = __pyx_t_13;
  }
  __pyx_L4:;

  /* "pandas/_libs/intervaltree.pxi":1098
 *         readonly bint is_leaf_node
 * 
 *     def __init__(self,             # <<<<<<<<<<<<<<
 *                  ndarray[float64_t, ndim=1] left,
 *                  ndarray[float64_t, ndim=1] right,
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __PYX_XDEC_MEMVIEW(&__pyx_t_7, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_8, 1);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __PYX_XDEC_MEMVIEW(&__pyx_t_11, 1);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indices.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_left.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_right.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("pandas._libs.interval.Float64ClosedRightIntervalNode.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indices.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_left.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_right.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF(__pyx_v_left_set);
  __Pyx_XDECREF(__pyx_v_right_set);
  __Pyx_XDECREF(__pyx_v_center_set);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":1144
 *     @cython.wraparound(False)
 *     @cython.boundscheck(False)
 *     cdef classify_intervals(self, float64_t[:] left, float64_t[:] right):             # <<<<<<<<<<<<<<
 *         """Classify the given intervals based upon whether they fall to the
 *         left, right, or overlap with this node's pivot.
 */

static PyObject *__pyx_f_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_classify_intervals(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode *__pyx_v_self, __Pyx_memviewslice __pyx_v_left, __Pyx_memviewslice __pyx_v_right) {
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_left_ind = 0;
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_right_ind = 0;
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_overlapping_ind = 0;
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __pyx_t_5numpy_int64_t __pyx_t_2;
  __pyx_t_5numpy_int64_t __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  int __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  __Pyx_RefNannySetupContext("classify_intervals", 0);

  /* "pandas/_libs/intervaltree.pxi":1152
 *             Py_ssize_t i
 * 
 *         left_ind = Int64Vector()             # <<<<<<<<<<<<<<
 *         right_ind = Int64Vector()
 *         overlapping_ind = Int64Vector()
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1152, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_left_ind = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":1153
 * 
 *         left_ind = Int64Vector()
 *         right_ind = Int64Vector()             # <<<<<<<<<<<<<<
 *         overlapping_ind = Int64Vector()
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1153, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_right_ind = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":1154
 *         left_ind = Int64Vector()
 *         right_ind = Int64Vector()
 *         overlapping_ind = Int64Vector()             # <<<<<<<<<<<<<<
 * 
 *         for i in range(self.n_elements):
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1154, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_overlapping_ind = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":1156
 *         overlapping_ind = Int64Vector()
 * 
 *         for i in range(self.n_elements):             # <<<<<<<<<<<<<<
 *             if right[i] < self.pivot:
 *                 left_ind.append(i)
 */
  __pyx_t_2 = __pyx_v_self->n_elements;
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "pandas/_libs/intervaltree.pxi":1157
 * 
 *         for i in range(self.n_elements):
 *             if right[i] < self.pivot:             # <<<<<<<<<<<<<<
 *                 left_ind.append(i)
 *             elif self.pivot <= left[i]:
 */
    __pyx_t_5 = __pyx_v_i;
    __pyx_t_6 = (((*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_right.data + __pyx_t_5 * __pyx_v_right.strides[0]) ))) < __pyx_v_self->pivot) != 0);
    if (__pyx_t_6) {

      /* "pandas/_libs/intervaltree.pxi":1158
 *         for i in range(self.n_elements):
 *             if right[i] < self.pivot:
 *                 left_ind.append(i)             # <<<<<<<<<<<<<<
 *             elif self.pivot <= left[i]:
 *                 right_ind.append(i)
 */
      ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_left_ind->__pyx_vtab)->append(__pyx_v_left_ind, __pyx_v_i);

      /* "pandas/_libs/intervaltree.pxi":1157
 * 
 *         for i in range(self.n_elements):
 *             if right[i] < self.pivot:             # <<<<<<<<<<<<<<
 *                 left_ind.append(i)
 *             elif self.pivot <= left[i]:
 */
      goto __pyx_L5;
    }

    /* "pandas/_libs/intervaltree.pxi":1159
 *             if right[i] < self.pivot:
 *                 left_ind.append(i)
 *             elif self.pivot <= left[i]:             # <<<<<<<<<<<<<<
 *                 right_ind.append(i)
 *             else:
 */
    __pyx_t_7 = __pyx_v_i;
    __pyx_t_6 = ((__pyx_v_self->pivot <= (*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_left.data + __pyx_t_7 * __pyx_v_left.strides[0]) )))) != 0);
    if (__pyx_t_6) {

      /* "pandas/_libs/intervaltree.pxi":1160
 *                 left_ind.append(i)
 *             elif self.pivot <= left[i]:
 *                 right_ind.append(i)             # <<<<<<<<<<<<<<
 *             else:
 *                 overlapping_ind.append(i)
 */
      ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_right_ind->__pyx_vtab)->append(__pyx_v_right_ind, __pyx_v_i);

      /* "pandas/_libs/intervaltree.pxi":1159
 *             if right[i] < self.pivot:
 *                 left_ind.append(i)
 *             elif self.pivot <= left[i]:             # <<<<<<<<<<<<<<
 *                 right_ind.append(i)
 *             else:
 */
      goto __pyx_L5;
    }

    /* "pandas/_libs/intervaltree.pxi":1162
 *                 right_ind.append(i)
 *             else:
 *                 overlapping_ind.append(i)             # <<<<<<<<<<<<<<
 * 
 *         return (left_ind.to_array(),
 */
    /*else*/ {
      ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_overlapping_ind->__pyx_vtab)->append(__pyx_v_overlapping_ind, __pyx_v_i);
    }
    __pyx_L5:;
  }

  /* "pandas/_libs/intervaltree.pxi":1164
 *                 overlapping_ind.append(i)
 * 
 *         return (left_ind.to_array(),             # <<<<<<<<<<<<<<
 *                 right_ind.to_array(),
 *                 overlapping_ind.to_array())
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_left_ind->__pyx_vtab)->to_array(__pyx_v_left_ind, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1164, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "pandas/_libs/intervaltree.pxi":1165
 * 
 *         return (left_ind.to_array(),
 *                 right_ind.to_array(),             # <<<<<<<<<<<<<<
 *                 overlapping_ind.to_array())
 * 
 */
  __pyx_t_8 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_right_ind->__pyx_vtab)->to_array(__pyx_v_right_ind, 0); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 1165, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);

  /* "pandas/_libs/intervaltree.pxi":1166
 *         return (left_ind.to_array(),
 *                 right_ind.to_array(),
 *                 overlapping_ind.to_array())             # <<<<<<<<<<<<<<
 * 
 *     cdef new_child_node(self,
 */
  __pyx_t_9 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_overlapping_ind->__pyx_vtab)->to_array(__pyx_v_overlapping_ind, 0); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 1166, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);

  /* "pandas/_libs/intervaltree.pxi":1164
 *                 overlapping_ind.append(i)
 * 
 *         return (left_ind.to_array(),             # <<<<<<<<<<<<<<
 *                 right_ind.to_array(),
 *                 overlapping_ind.to_array())
 */
  __pyx_t_10 = PyTuple_New(3); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 1164, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_t_8);
  __Pyx_GIVEREF(__pyx_t_9);
  PyTuple_SET_ITEM(__pyx_t_10, 2, __pyx_t_9);
  __pyx_t_1 = 0;
  __pyx_t_8 = 0;
  __pyx_t_9 = 0;
  __pyx_r = __pyx_t_10;
  __pyx_t_10 = 0;
  goto __pyx_L0;

  /* "pandas/_libs/intervaltree.pxi":1144
 *     @cython.wraparound(False)
 *     @cython.boundscheck(False)
 *     cdef classify_intervals(self, float64_t[:] left, float64_t[:] right):             # <<<<<<<<<<<<<<
 *         """Classify the given intervals based upon whether they fall to the
 *         left, right, or overlap with this node's pivot.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("pandas._libs.interval.Float64ClosedRightIntervalNode.classify_intervals", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_left_ind);
  __Pyx_XDECREF((PyObject *)__pyx_v_right_ind);
  __Pyx_XDECREF((PyObject *)__pyx_v_overlapping_ind);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":1168
 *                 overlapping_ind.to_array())
 * 
 *     cdef new_child_node(self,             # <<<<<<<<<<<<<<
 *                         ndarray[float64_t, ndim=1] left,
 *                         ndarray[float64_t, ndim=1] right,
 */

static PyObject *__pyx_f_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_new_child_node(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode *__pyx_v_self, PyArrayObject *__pyx_v_left, PyArrayObject *__pyx_v_right, PyArrayObject *__pyx_v_indices, PyArrayObject *__pyx_v_subset) {
  __Pyx_LocalBuf_ND __pyx_pybuffernd_indices;
  __Pyx_Buffer __pyx_pybuffer_indices;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_left;
  __Pyx_Buffer __pyx_pybuffer_left;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_right;
  __Pyx_Buffer __pyx_pybuffer_right;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_subset;
  __Pyx_Buffer __pyx_pybuffer_subset;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyArrayObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyArrayObject *__pyx_t_7 = NULL;
  PyArrayObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  __Pyx_RefNannySetupContext("new_child_node", 0);
  __Pyx_INCREF((PyObject *)__pyx_v_left);
  __Pyx_INCREF((PyObject *)__pyx_v_right);
  __Pyx_INCREF((PyObject *)__pyx_v_indices);
  __pyx_pybuffer_left.pybuffer.buf = NULL;
  __pyx_pybuffer_left.refcount = 0;
  __pyx_pybuffernd_left.data = NULL;
  __pyx_pybuffernd_left.rcbuffer = &__pyx_pybuffer_left;
  __pyx_pybuffer_right.pybuffer.buf = NULL;
  __pyx_pybuffer_right.refcount = 0;
  __pyx_pybuffernd_right.data = NULL;
  __pyx_pybuffernd_right.rcbuffer = &__pyx_pybuffer_right;
  __pyx_pybuffer_indices.pybuffer.buf = NULL;
  __pyx_pybuffer_indices.refcount = 0;
  __pyx_pybuffernd_indices.data = NULL;
  __pyx_pybuffernd_indices.rcbuffer = &__pyx_pybuffer_indices;
  __pyx_pybuffer_subset.pybuffer.buf = NULL;
  __pyx_pybuffer_subset.refcount = 0;
  __pyx_pybuffernd_subset.data = NULL;
  __pyx_pybuffernd_subset.rcbuffer = &__pyx_pybuffer_subset;
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_left.rcbuffer->pybuffer, (PyObject*)__pyx_v_left, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 1168, __pyx_L1_error)
  }
  __pyx_pybuffernd_left.diminfo[0].strides = __pyx_pybuffernd_left.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_left.diminfo[0].shape = __pyx_pybuffernd_left.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_right.rcbuffer->pybuffer, (PyObject*)__pyx_v_right, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 1168, __pyx_L1_error)
  }
  __pyx_pybuffernd_right.diminfo[0].strides = __pyx_pybuffernd_right.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_right.diminfo[0].shape = __pyx_pybuffernd_right.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_indices.rcbuffer->pybuffer, (PyObject*)__pyx_v_indices, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 1168, __pyx_L1_error)
  }
  __pyx_pybuffernd_indices.diminfo[0].strides = __pyx_pybuffernd_indices.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_indices.diminfo[0].shape = __pyx_pybuffernd_indices.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_subset.rcbuffer->pybuffer, (PyObject*)__pyx_v_subset, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 1168, __pyx_L1_error)
  }
  __pyx_pybuffernd_subset.diminfo[0].strides = __pyx_pybuffernd_subset.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_subset.diminfo[0].shape = __pyx_pybuffernd_subset.rcbuffer->pybuffer.shape[0];

  /* "pandas/_libs/intervaltree.pxi":1175
 *         """Create a new child node.
 *         """
 *         left = take(left, subset)             # <<<<<<<<<<<<<<
 *         right = take(right, subset)
 *         indices = take(indices, subset)
 */
  __pyx_t_1 = __pyx_f_6pandas_5_libs_8interval_take(((PyArrayObject *)__pyx_v_left), ((PyArrayObject *)__pyx_v_subset)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1175, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(1, 1175, __pyx_L1_error)
  __pyx_t_2 = ((PyArrayObject *)__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_left.rcbuffer->pybuffer);
    __pyx_t_3 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_left.rcbuffer->pybuffer, (PyObject*)__pyx_t_2, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack);
    if (unlikely(__pyx_t_3 < 0)) {
      PyErr_Fetch(&__pyx_t_4, &__pyx_t_5, &__pyx_t_6);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_left.rcbuffer->pybuffer, (PyObject*)__pyx_v_left, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_4); Py_XDECREF(__pyx_t_5); Py_XDECREF(__pyx_t_6);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_4, __pyx_t_5, __pyx_t_6);
      }
      __pyx_t_4 = __pyx_t_5 = __pyx_t_6 = 0;
    }
    __pyx_pybuffernd_left.diminfo[0].strides = __pyx_pybuffernd_left.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_left.diminfo[0].shape = __pyx_pybuffernd_left.rcbuffer->pybuffer.shape[0];
    if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(1, 1175, __pyx_L1_error)
  }
  __pyx_t_2 = 0;
  __Pyx_DECREF_SET(__pyx_v_left, ((PyArrayObject *)__pyx_t_1));
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":1176
 *         """
 *         left = take(left, subset)
 *         right = take(right, subset)             # <<<<<<<<<<<<<<
 *         indices = take(indices, subset)
 *         return Float64ClosedRightIntervalNode(
 */
  __pyx_t_1 = __pyx_f_6pandas_5_libs_8interval_take(((PyArrayObject *)__pyx_v_right), ((PyArrayObject *)__pyx_v_subset)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1176, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(1, 1176, __pyx_L1_error)
  __pyx_t_7 = ((PyArrayObject *)__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_right.rcbuffer->pybuffer);
    __pyx_t_3 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_right.rcbuffer->pybuffer, (PyObject*)__pyx_t_7, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack);
    if (unlikely(__pyx_t_3 < 0)) {
      PyErr_Fetch(&__pyx_t_6, &__pyx_t_5, &__pyx_t_4);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_right.rcbuffer->pybuffer, (PyObject*)__pyx_v_right, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_6); Py_XDECREF(__pyx_t_5); Py_XDECREF(__pyx_t_4);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_6, __pyx_t_5, __pyx_t_4);
      }
      __pyx_t_6 = __pyx_t_5 = __pyx_t_4 = 0;
    }
    __pyx_pybuffernd_right.diminfo[0].strides = __pyx_pybuffernd_right.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_right.diminfo[0].shape = __pyx_pybuffernd_right.rcbuffer->pybuffer.shape[0];
    if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(1, 1176, __pyx_L1_error)
  }
  __pyx_t_7 = 0;
  __Pyx_DECREF_SET(__pyx_v_right, ((PyArrayObject *)__pyx_t_1));
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":1177
 *         left = take(left, subset)
 *         right = take(right, subset)
 *         indices = take(indices, subset)             # <<<<<<<<<<<<<<
 *         return Float64ClosedRightIntervalNode(
 *             left, right, indices, self.leaf_size)
 */
  __pyx_t_1 = __pyx_f_6pandas_5_libs_8interval_take(((PyArrayObject *)__pyx_v_indices), ((PyArrayObject *)__pyx_v_subset)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1177, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(1, 1177, __pyx_L1_error)
  __pyx_t_8 = ((PyArrayObject *)__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indices.rcbuffer->pybuffer);
    __pyx_t_3 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_indices.rcbuffer->pybuffer, (PyObject*)__pyx_t_8, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack);
    if (unlikely(__pyx_t_3 < 0)) {
      PyErr_Fetch(&__pyx_t_4, &__pyx_t_5, &__pyx_t_6);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_indices.rcbuffer->pybuffer, (PyObject*)__pyx_v_indices, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_4); Py_XDECREF(__pyx_t_5); Py_XDECREF(__pyx_t_6);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_4, __pyx_t_5, __pyx_t_6);
      }
      __pyx_t_4 = __pyx_t_5 = __pyx_t_6 = 0;
    }
    __pyx_pybuffernd_indices.diminfo[0].strides = __pyx_pybuffernd_indices.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_indices.diminfo[0].shape = __pyx_pybuffernd_indices.rcbuffer->pybuffer.shape[0];
    if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(1, 1177, __pyx_L1_error)
  }
  __pyx_t_8 = 0;
  __Pyx_DECREF_SET(__pyx_v_indices, ((PyArrayObject *)__pyx_t_1));
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":1178
 *         right = take(right, subset)
 *         indices = take(indices, subset)
 *         return Float64ClosedRightIntervalNode(             # <<<<<<<<<<<<<<
 *             left, right, indices, self.leaf_size)
 * 
 */
  __Pyx_XDECREF(__pyx_r);

  /* "pandas/_libs/intervaltree.pxi":1179
 *         indices = take(indices, subset)
 *         return Float64ClosedRightIntervalNode(
 *             left, right, indices, self.leaf_size)             # <<<<<<<<<<<<<<
 * 
 *     @cython.wraparound(False)
 */
  __pyx_t_1 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->leaf_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1179, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "pandas/_libs/intervaltree.pxi":1178
 *         right = take(right, subset)
 *         indices = take(indices, subset)
 *         return Float64ClosedRightIntervalNode(             # <<<<<<<<<<<<<<
 *             left, right, indices, self.leaf_size)
 * 
 */
  __pyx_t_9 = PyTuple_New(4); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 1178, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_INCREF(((PyObject *)__pyx_v_left));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_left));
  PyTuple_SET_ITEM(__pyx_t_9, 0, ((PyObject *)__pyx_v_left));
  __Pyx_INCREF(((PyObject *)__pyx_v_right));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_right));
  PyTuple_SET_ITEM(__pyx_t_9, 1, ((PyObject *)__pyx_v_right));
  __Pyx_INCREF(((PyObject *)__pyx_v_indices));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_indices));
  PyTuple_SET_ITEM(__pyx_t_9, 2, ((PyObject *)__pyx_v_indices));
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_9, 3, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode), __pyx_t_9, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1178, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pandas/_libs/intervaltree.pxi":1168
 *                 overlapping_ind.to_array())
 * 
 *     cdef new_child_node(self,             # <<<<<<<<<<<<<<
 *                         ndarray[float64_t, ndim=1] left,
 *                         ndarray[float64_t, ndim=1] right,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_9);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indices.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_left.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_right.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_subset.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("pandas._libs.interval.Float64ClosedRightIntervalNode.new_child_node", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indices.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_left.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_right.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_subset.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF((PyObject *)__pyx_v_left);
  __Pyx_XDECREF((PyObject *)__pyx_v_right);
  __Pyx_XDECREF((PyObject *)__pyx_v_indices);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":1184
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_2query[] = "Recursively query this node and its sub-nodes for intervals that\n        overlap with the query point.\n        ";
static PyMethodDef __pyx_mdef_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_3query = {"query", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_3query, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_2query};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_args = 0;
  PyObject *__pyx_v_kwargs = 0;
  CYTHON_UNUSED PyObject *__pyx_v_defaults = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fused_cpdef (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_args,&__pyx_n_s_kwargs,&__pyx_n_s_defaults,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_args)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_kwargs)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fused_cpdef", 1, 3, 3, 1); __PYX_ERR(1, 1184, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_defaults)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fused_cpdef", 1, 3, 3, 2); __PYX_ERR(1, 1184, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fused_cpdef") < 0)) __PYX_ERR(1, 1184, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_args = values[0];
    __pyx_v_kwargs = values[1];
    __pyx_v_defaults = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fused_cpdef", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 1184, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Float64ClosedRightIntervalNode.__pyx_fused_cpdef", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_2query(((struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode *)__pyx_v_signatures), __pyx_v_args, __pyx_v_kwargs, __pyx_v_defaults);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_2query(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode *__pyx_v_signatures, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs, CYTHON_UNUSED PyObject *__pyx_v_defaults) {
  PyObject *__pyx_v_dest_sig = NULL;
  Py_ssize_t __pyx_v_i;
  CYTHON_UNUSED PyTypeObject *__pyx_v_ndarray = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_v_itemsize;
  PyObject *__pyx_v_arg = NULL;
  PyObject *__pyx_v_candidates = NULL;
  PyObject *__pyx_v_sig = NULL;
  int __pyx_v_match_found;
  PyObject *__pyx_v_src_sig = NULL;
  PyObject *__pyx_v_dst_type = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  Py_ssize_t __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  Py_ssize_t __pyx_t_13;
  Py_ssize_t __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  int __pyx_t_16;
  __Pyx_RefNannySetupContext("query", 0);
  __Pyx_INCREF(__pyx_v_kwargs);
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1184, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  PyList_SET_ITEM(__pyx_t_1, 0, Py_None);
  __pyx_v_dest_sig = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_3 = (__pyx_v_kwargs != Py_None);
  __pyx_t_4 = (__pyx_t_3 != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_kwargs); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(1, 1184, __pyx_L1_error)
  __pyx_t_3 = ((!__pyx_t_4) != 0);
  __pyx_t_2 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {
    __Pyx_INCREF(Py_None);
    __Pyx_DECREF_SET(__pyx_v_kwargs, Py_None);
  }
  __pyx_t_1 = ((PyObject *)__Pyx_ImportNumPyArrayTypeIfAvailable()); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1184, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_ndarray = ((PyTypeObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_v_itemsize = -1L;
  if (unlikely(__pyx_v_args == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 1184, __pyx_L1_error)
  }
  __pyx_t_5 = PyTuple_GET_SIZE(((PyObject*)__pyx_v_args)); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(1, 1184, __pyx_L1_error)
  __pyx_t_2 = ((2 < __pyx_t_5) != 0);
  if (__pyx_t_2) {
    if (unlikely(__pyx_v_args == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 1184, __pyx_L1_error)
    }
    __pyx_t_1 = PyTuple_GET_ITEM(((PyObject*)__pyx_v_args), 2);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v_arg = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L6;
  }
  __pyx_t_3 = (__pyx_v_kwargs != Py_None);
  __pyx_t_4 = (__pyx_t_3 != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L7_bool_binop_done;
  }
  if (unlikely(__pyx_v_kwargs == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(1, 1184, __pyx_L1_error)
  }
  __pyx_t_4 = (__Pyx_PyDict_ContainsTF(__pyx_n_s_point, ((PyObject*)__pyx_v_kwargs), Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(1, 1184, __pyx_L1_error)
  __pyx_t_3 = (__pyx_t_4 != 0);
  __pyx_t_2 = __pyx_t_3;
  __pyx_L7_bool_binop_done:;
  if (__pyx_t_2) {
    if (unlikely(__pyx_v_kwargs == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 1184, __pyx_L1_error)
    }
    __pyx_t_1 = __Pyx_PyDict_GetItem(((PyObject*)__pyx_v_kwargs), __pyx_n_s_point); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1184, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_arg = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L6;
  }
  /*else*/ {
    if (unlikely(__pyx_v_args == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
      __PYX_ERR(1, 1184, __pyx_L1_error)
    }
    __pyx_t_5 = PyTuple_GET_SIZE(((PyObject*)__pyx_v_args)); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(1, 1184, __pyx_L1_error)
    __pyx_t_1 = PyInt_FromSsize_t(__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1184, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_6 = PyTuple_New(3); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 1184, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_INCREF(__pyx_int_3);
    __Pyx_GIVEREF(__pyx_int_3);
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_int_3);
    __Pyx_INCREF(__pyx_n_s_s);
    __Pyx_GIVEREF(__pyx_n_s_s);
    PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_n_s_s);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Expected_at_least_d_argument_s_g, __pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1184, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_1); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 1184, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(1, 1184, __pyx_L1_error)
  }
  __pyx_L6:;
  while (1) {
    __pyx_t_2 = PyFloat_Check(__pyx_v_arg); 
    __pyx_t_3 = (__pyx_t_2 != 0);
    if (__pyx_t_3) {
      if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, __pyx_n_s_float64_t, long, 1, __Pyx_PyInt_From_long, 1, 0, 0) < 0)) __PYX_ERR(1, 1184, __pyx_L1_error)
      goto __pyx_L10_break;
    }
    __pyx_t_2 = PyInt_Check(__pyx_v_arg); 
    __pyx_t_4 = (__pyx_t_2 != 0);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_3 = __pyx_t_4;
      goto __pyx_L13_bool_binop_done;
    }
    __pyx_t_4 = PyLong_Check(__pyx_v_arg); 
    __pyx_t_2 = (__pyx_t_4 != 0);
    __pyx_t_3 = __pyx_t_2;
    __pyx_L13_bool_binop_done:;
    __pyx_t_2 = (__pyx_t_3 != 0);
    if (__pyx_t_2) {
      if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, __pyx_n_s_int64_t, long, 1, __Pyx_PyInt_From_long, 1, 0, 0) < 0)) __PYX_ERR(1, 1184, __pyx_L1_error)
      goto __pyx_L10_break;
    }
    if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, Py_None, long, 1, __Pyx_PyInt_From_long, 1, 0, 0) < 0)) __PYX_ERR(1, 1184, __pyx_L1_error)
    goto __pyx_L10_break;
  }
  __pyx_L10_break:;
  __pyx_t_6 = PyList_New(0); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 1184, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_v_candidates = ((PyObject*)__pyx_t_6);
  __pyx_t_6 = 0;
  __pyx_t_5 = 0;
  if (unlikely(((PyObject *)__pyx_v_signatures) == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(1, 1184, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_dict_iterator(((PyObject*)__pyx_v_signatures), 1, ((PyObject *)NULL), (&__pyx_t_7), (&__pyx_t_8)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1184, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __pyx_t_6 = __pyx_t_1;
  __pyx_t_1 = 0;
  while (1) {
    __pyx_t_9 = __Pyx_dict_iter_next(__pyx_t_6, __pyx_t_7, &__pyx_t_5, &__pyx_t_1, NULL, NULL, __pyx_t_8);
    if (unlikely(__pyx_t_9 == 0)) break;
    if (unlikely(__pyx_t_9 == -1)) __PYX_ERR(1, 1184, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_sig, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_v_match_found = 0;
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_v_sig, __pyx_n_s_strip); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 1184, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __pyx_t_12 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_11))) {
      __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_11);
      if (likely(__pyx_t_12)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
        __Pyx_INCREF(__pyx_t_12);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_11, function);
      }
    }
    __pyx_t_10 = (__pyx_t_12) ? __Pyx_PyObject_Call2Args(__pyx_t_11, __pyx_t_12, __pyx_kp_s__13) : __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_kp_s__13);
    __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
    if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 1184, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_split); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 1184, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_t_10 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_11))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_11);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_11, function);
      }
    }
    __pyx_t_1 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_11, __pyx_t_10, __pyx_kp_s__14) : __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_kp_s__14);
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1184, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_XDECREF_SET(__pyx_v_src_sig, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_13 = PyList_GET_SIZE(__pyx_v_dest_sig); if (unlikely(__pyx_t_13 == ((Py_ssize_t)-1))) __PYX_ERR(1, 1184, __pyx_L1_error)
    __pyx_t_14 = __pyx_t_13;
    for (__pyx_t_15 = 0; __pyx_t_15 < __pyx_t_14; __pyx_t_15+=1) {
      __pyx_v_i = __pyx_t_15;
      __pyx_t_1 = PyList_GET_ITEM(__pyx_v_dest_sig, __pyx_v_i);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_XDECREF_SET(__pyx_v_dst_type, __pyx_t_1);
      __pyx_t_1 = 0;
      __pyx_t_2 = (__pyx_v_dst_type != Py_None);
      __pyx_t_3 = (__pyx_t_2 != 0);
      if (__pyx_t_3) {
        __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_src_sig, __pyx_v_i, Py_ssize_t, 1, PyInt_FromSsize_t, 0, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1184, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_11 = PyObject_RichCompare(__pyx_t_1, __pyx_v_dst_type, Py_EQ); __Pyx_XGOTREF(__pyx_t_11); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 1184, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_11); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(1, 1184, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        if (__pyx_t_3) {
          __pyx_v_match_found = 1;
          goto __pyx_L20;
        }
        /*else*/ {
          __pyx_v_match_found = 0;
          goto __pyx_L18_break;
        }
        __pyx_L20:;
      }
    }
    __pyx_L18_break:;
    __pyx_t_3 = (__pyx_v_match_found != 0);
    if (__pyx_t_3) {
      __pyx_t_16 = __Pyx_PyList_Append(__pyx_v_candidates, __pyx_v_sig); if (unlikely(__pyx_t_16 == ((int)-1))) __PYX_ERR(1, 1184, __pyx_L1_error)
    }
  }
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_3 = (PyList_GET_SIZE(__pyx_v_candidates) != 0);
  __pyx_t_2 = ((!__pyx_t_3) != 0);
  if (__pyx_t_2) {
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__15, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 1184, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(1, 1184, __pyx_L1_error)
  }
  __pyx_t_7 = PyList_GET_SIZE(__pyx_v_candidates); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1))) __PYX_ERR(1, 1184, __pyx_L1_error)
  __pyx_t_2 = ((__pyx_t_7 > 1) != 0);
  if (__pyx_t_2) {
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__16, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 1184, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(1, 1184, __pyx_L1_error)
  }
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    if (unlikely(((PyObject *)__pyx_v_signatures) == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 1184, __pyx_L1_error)
    }
    __pyx_t_6 = __Pyx_PyDict_GetItem(((PyObject*)__pyx_v_signatures), PyList_GET_ITEM(__pyx_v_candidates, 0)); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 1184, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_r = __pyx_t_6;
    __pyx_t_6 = 0;
    goto __pyx_L0;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("pandas._libs.interval.Float64ClosedRightIntervalNode.__pyx_fused_cpdef", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_dest_sig);
  __Pyx_XDECREF(__pyx_v_ndarray);
  __Pyx_XDECREF(__pyx_v_arg);
  __Pyx_XDECREF(__pyx_v_candidates);
  __Pyx_XDECREF(__pyx_v_sig);
  __Pyx_XDECREF(__pyx_v_src_sig);
  __Pyx_XDECREF(__pyx_v_dst_type);
  __Pyx_XDECREF(__pyx_v_kwargs);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_9__pyx_fuse_0query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_fuse_0__pyx_f_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float64_t __pyx_v_point, int __pyx_skip_dispatch) {
  __Pyx_memviewslice __pyx_v_indices = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_values = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __pyx_t_5numpy_int64_t __pyx_t_9;
  __pyx_t_5numpy_int64_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  int __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  __Pyx_memviewslice __pyx_t_16 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_17 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  Py_ssize_t __pyx_t_21;
  __Pyx_RefNannySetupContext("__pyx_fuse_0query", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pyx_fuse_0query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1184, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_9__pyx_fuse_0query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = PyFloat_FromDouble(__pyx_v_point); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1184, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1184, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1184, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 1184, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_5) {
            __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1184, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1184, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_3query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_4 = PyFloat_FromDouble(__pyx_v_point); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1184, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_7 = __pyx_t_1; __pyx_t_3 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1184, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1184, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1184, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (__pyx_t_3) {
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_t_4);
          __pyx_t_4 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1184, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pandas/_libs/intervaltree.pxi":1193
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
  __pyx_t_8 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_8) {

    /* "pandas/_libs/intervaltree.pxi":1197
 *             # continue the binary tree structure. Instead, we use linear
 *             # search.
 *             for i in range(self.n_elements):             # <<<<<<<<<<<<<<
 *                 if self.left[i] < point <= self.right[i]:
 *                     result.append(self.indices[i])
 */
    __pyx_t_9 = __pyx_v_self->n_elements;
    __pyx_t_10 = __pyx_t_9;
    for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
      __pyx_v_i = __pyx_t_11;

      /* "pandas/_libs/intervaltree.pxi":1198
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point <= self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      __pyx_t_12 = __pyx_v_i;
      __pyx_t_8 = ((*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_self->left.data + __pyx_t_12 * __pyx_v_self->left.strides[0]) ))) < __pyx_v_point);
      if (__pyx_t_8) {
        __pyx_t_13 = __pyx_v_i;
        __pyx_t_8 = (__pyx_v_point <= (*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_self->right.data + __pyx_t_13 * __pyx_v_self->right.strides[0]) ))));
      }
      __pyx_t_14 = (__pyx_t_8 != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":1199
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point <= self.right[i]:
 *                     result.append(self.indices[i])             # <<<<<<<<<<<<<<
 *         else:
 *             # There are child nodes. Based on comparing our query to the pivot,
 */
        __pyx_t_15 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->indices.data + __pyx_t_15 * __pyx_v_self->indices.strides[0]) ))));

        /* "pandas/_libs/intervaltree.pxi":1198
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point <= self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      }
    }

    /* "pandas/_libs/intervaltree.pxi":1193
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
    goto __pyx_L3;
  }

  /* "pandas/_libs/intervaltree.pxi":1203
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
  /*else*/ {
    __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":1204
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:
 *                 values = self.center_left_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 */
      __pyx_t_16 = __pyx_v_self->center_left_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":1205
 *             if point < self.pivot:
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:
 */
      __pyx_t_17 = __pyx_v_self->center_left_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":1206
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):             # <<<<<<<<<<<<<<
 *                     if not values[i] < point:
 *                         break
 */
      __pyx_t_9 = __pyx_v_self->n_center;
      __pyx_t_10 = __pyx_t_9;
      for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":1207
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_18 = __pyx_v_i;
        __pyx_t_14 = ((!(((*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_18 * __pyx_v_values.strides[0]) ))) < __pyx_v_point) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":1208
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:
 */
          goto __pyx_L9_break;

          /* "pandas/_libs/intervaltree.pxi":1207
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":1209
 *                     if not values[i] < point:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 */
        __pyx_t_19 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_19 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L9_break:;

      /* "pandas/_libs/intervaltree.pxi":1210
 *                         break
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      __pyx_t_14 = ((__pyx_v_point <= __pyx_v_self->left_node->max_right) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":1211
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)             # <<<<<<<<<<<<<<
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode *)__pyx_v_self->left_node->__pyx_vtab)->__pyx_fuse_0query(__pyx_v_self->left_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1211, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":1210
 *                         break
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":1203
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":1212
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
    __pyx_t_14 = ((__pyx_v_point > __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":1213
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 *                 values = self.center_right_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 */
      __pyx_t_16 = __pyx_v_self->center_right_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":1214
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:
 */
      __pyx_t_17 = __pyx_v_self->center_right_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":1215
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):             # <<<<<<<<<<<<<<
 *                     if not point <= values[i]:
 *                         break
 */
      for (__pyx_t_11 = (__pyx_v_self->n_center - 1); __pyx_t_11 > -1L; __pyx_t_11-=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":1216
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_20 = __pyx_v_i;
        __pyx_t_14 = ((!((__pyx_v_point <= (*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_20 * __pyx_v_values.strides[0]) )))) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":1217
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:
 */
          goto __pyx_L13_break;

          /* "pandas/_libs/intervaltree.pxi":1216
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":1218
 *                     if not point <= values[i]:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if self.right_node.min_left < point:
 *                     self.right_node.query(result, point)
 */
        __pyx_t_21 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_21 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L13_break:;

      /* "pandas/_libs/intervaltree.pxi":1219
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      __pyx_t_14 = ((__pyx_v_self->right_node->min_left < __pyx_v_point) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":1220
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:
 *                     self.right_node.query(result, point)             # <<<<<<<<<<<<<<
 *             else:
 *                 result.extend(self.center_left_indices)
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode *)__pyx_v_self->right_node->__pyx_vtab)->__pyx_fuse_0query(__pyx_v_self->right_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1220, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":1219
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":1212
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":1222
 *                     self.right_node.query(result, point)
 *             else:
 *                 result.extend(self.center_left_indices)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
    /*else*/ {
      __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->extend(__pyx_v_result, __pyx_v_self->center_left_indices); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1222, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_L7:;
  }
  __pyx_L3:;

  /* "pandas/_libs/intervaltree.pxi":1184
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __PYX_XDEC_MEMVIEW(&__pyx_t_16, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_17, 1);
  __Pyx_AddTraceback("pandas._libs.interval.Float64ClosedRightIntervalNode.query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_indices, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_values, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_9__pyx_fuse_0query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_0__pyx_mdef_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_9__pyx_fuse_0query = {"__pyx_fuse_0query", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_9__pyx_fuse_0query, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_2query};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_9__pyx_fuse_0query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result = 0;
  __pyx_t_5numpy_float64_t __pyx_v_point;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fuse_0query (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_result,&__pyx_n_s_point,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_result)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_point)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_0query", 1, 2, 2, 1); __PYX_ERR(1, 1184, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_0query") < 0)) __PYX_ERR(1, 1184, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_result = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)values[0]);
    __pyx_v_point = __pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_point == ((npy_float64)-1)) && PyErr_Occurred())) __PYX_ERR(1, 1184, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_0query", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 1184, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Float64ClosedRightIntervalNode.__pyx_fuse_0query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_result), __pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector, 1, "result", 0))) __PYX_ERR(1, 1184, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_8__pyx_fuse_0query(((struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode *)__pyx_v_self), __pyx_v_result, __pyx_v_point);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_8__pyx_fuse_0query(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float64_t __pyx_v_point) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__pyx_fuse_0query", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_fuse_0__pyx_f_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_query(__pyx_v_self, __pyx_v_result, __pyx_v_point, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1184, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Float64ClosedRightIntervalNode.__pyx_fuse_0query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_11__pyx_fuse_1query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_fuse_1__pyx_f_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float32_t __pyx_v_point, int __pyx_skip_dispatch) {
  __Pyx_memviewslice __pyx_v_indices = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_values = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __pyx_t_5numpy_int64_t __pyx_t_9;
  __pyx_t_5numpy_int64_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  int __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  __Pyx_memviewslice __pyx_t_16 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_17 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  Py_ssize_t __pyx_t_21;
  __Pyx_RefNannySetupContext("__pyx_fuse_1query", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pyx_fuse_1query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1184, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_11__pyx_fuse_1query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = PyFloat_FromDouble(__pyx_v_point); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1184, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1184, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1184, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 1184, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_5) {
            __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1184, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1184, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_3query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_4 = PyFloat_FromDouble(__pyx_v_point); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1184, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_7 = __pyx_t_1; __pyx_t_3 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1184, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1184, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1184, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (__pyx_t_3) {
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_t_4);
          __pyx_t_4 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1184, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pandas/_libs/intervaltree.pxi":1193
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
  __pyx_t_8 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_8) {

    /* "pandas/_libs/intervaltree.pxi":1197
 *             # continue the binary tree structure. Instead, we use linear
 *             # search.
 *             for i in range(self.n_elements):             # <<<<<<<<<<<<<<
 *                 if self.left[i] < point <= self.right[i]:
 *                     result.append(self.indices[i])
 */
    __pyx_t_9 = __pyx_v_self->n_elements;
    __pyx_t_10 = __pyx_t_9;
    for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
      __pyx_v_i = __pyx_t_11;

      /* "pandas/_libs/intervaltree.pxi":1198
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point <= self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      __pyx_t_12 = __pyx_v_i;
      __pyx_t_8 = ((*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_self->left.data + __pyx_t_12 * __pyx_v_self->left.strides[0]) ))) < __pyx_v_point);
      if (__pyx_t_8) {
        __pyx_t_13 = __pyx_v_i;
        __pyx_t_8 = (__pyx_v_point <= (*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_self->right.data + __pyx_t_13 * __pyx_v_self->right.strides[0]) ))));
      }
      __pyx_t_14 = (__pyx_t_8 != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":1199
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point <= self.right[i]:
 *                     result.append(self.indices[i])             # <<<<<<<<<<<<<<
 *         else:
 *             # There are child nodes. Based on comparing our query to the pivot,
 */
        __pyx_t_15 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->indices.data + __pyx_t_15 * __pyx_v_self->indices.strides[0]) ))));

        /* "pandas/_libs/intervaltree.pxi":1198
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point <= self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      }
    }

    /* "pandas/_libs/intervaltree.pxi":1193
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
    goto __pyx_L3;
  }

  /* "pandas/_libs/intervaltree.pxi":1203
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
  /*else*/ {
    __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":1204
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:
 *                 values = self.center_left_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 */
      __pyx_t_16 = __pyx_v_self->center_left_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":1205
 *             if point < self.pivot:
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:
 */
      __pyx_t_17 = __pyx_v_self->center_left_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":1206
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):             # <<<<<<<<<<<<<<
 *                     if not values[i] < point:
 *                         break
 */
      __pyx_t_9 = __pyx_v_self->n_center;
      __pyx_t_10 = __pyx_t_9;
      for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":1207
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_18 = __pyx_v_i;
        __pyx_t_14 = ((!(((*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_18 * __pyx_v_values.strides[0]) ))) < __pyx_v_point) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":1208
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:
 */
          goto __pyx_L9_break;

          /* "pandas/_libs/intervaltree.pxi":1207
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":1209
 *                     if not values[i] < point:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 */
        __pyx_t_19 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_19 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L9_break:;

      /* "pandas/_libs/intervaltree.pxi":1210
 *                         break
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      __pyx_t_14 = ((__pyx_v_point <= __pyx_v_self->left_node->max_right) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":1211
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)             # <<<<<<<<<<<<<<
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode *)__pyx_v_self->left_node->__pyx_vtab)->__pyx_fuse_1query(__pyx_v_self->left_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1211, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":1210
 *                         break
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":1203
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":1212
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
    __pyx_t_14 = ((__pyx_v_point > __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":1213
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 *                 values = self.center_right_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 */
      __pyx_t_16 = __pyx_v_self->center_right_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":1214
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:
 */
      __pyx_t_17 = __pyx_v_self->center_right_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":1215
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):             # <<<<<<<<<<<<<<
 *                     if not point <= values[i]:
 *                         break
 */
      for (__pyx_t_11 = (__pyx_v_self->n_center - 1); __pyx_t_11 > -1L; __pyx_t_11-=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":1216
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_20 = __pyx_v_i;
        __pyx_t_14 = ((!((__pyx_v_point <= (*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_20 * __pyx_v_values.strides[0]) )))) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":1217
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:
 */
          goto __pyx_L13_break;

          /* "pandas/_libs/intervaltree.pxi":1216
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":1218
 *                     if not point <= values[i]:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if self.right_node.min_left < point:
 *                     self.right_node.query(result, point)
 */
        __pyx_t_21 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_21 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L13_break:;

      /* "pandas/_libs/intervaltree.pxi":1219
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      __pyx_t_14 = ((__pyx_v_self->right_node->min_left < __pyx_v_point) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":1220
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:
 *                     self.right_node.query(result, point)             # <<<<<<<<<<<<<<
 *             else:
 *                 result.extend(self.center_left_indices)
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode *)__pyx_v_self->right_node->__pyx_vtab)->__pyx_fuse_1query(__pyx_v_self->right_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1220, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":1219
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":1212
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":1222
 *                     self.right_node.query(result, point)
 *             else:
 *                 result.extend(self.center_left_indices)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
    /*else*/ {
      __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->extend(__pyx_v_result, __pyx_v_self->center_left_indices); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1222, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_L7:;
  }
  __pyx_L3:;

  /* "pandas/_libs/intervaltree.pxi":1184
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __PYX_XDEC_MEMVIEW(&__pyx_t_16, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_17, 1);
  __Pyx_AddTraceback("pandas._libs.interval.Float64ClosedRightIntervalNode.query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_indices, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_values, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_11__pyx_fuse_1query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_1__pyx_mdef_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_11__pyx_fuse_1query = {"__pyx_fuse_1query", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_11__pyx_fuse_1query, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_2query};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_11__pyx_fuse_1query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result = 0;
  __pyx_t_5numpy_float32_t __pyx_v_point;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fuse_1query (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_result,&__pyx_n_s_point,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_result)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_point)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_1query", 1, 2, 2, 1); __PYX_ERR(1, 1184, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_1query") < 0)) __PYX_ERR(1, 1184, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_result = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)values[0]);
    __pyx_v_point = __pyx_PyFloat_AsFloat(values[1]); if (unlikely((__pyx_v_point == ((npy_float32)-1)) && PyErr_Occurred())) __PYX_ERR(1, 1184, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_1query", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 1184, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Float64ClosedRightIntervalNode.__pyx_fuse_1query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_result), __pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector, 1, "result", 0))) __PYX_ERR(1, 1184, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_10__pyx_fuse_1query(((struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode *)__pyx_v_self), __pyx_v_result, __pyx_v_point);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_10__pyx_fuse_1query(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float32_t __pyx_v_point) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__pyx_fuse_1query", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_fuse_1__pyx_f_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_query(__pyx_v_self, __pyx_v_result, __pyx_v_point, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1184, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Float64ClosedRightIntervalNode.__pyx_fuse_1query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_13__pyx_fuse_2query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_fuse_2__pyx_f_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int64_t __pyx_v_point, int __pyx_skip_dispatch) {
  __Pyx_memviewslice __pyx_v_indices = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_values = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __pyx_t_5numpy_int64_t __pyx_t_9;
  __pyx_t_5numpy_int64_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  int __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  __Pyx_memviewslice __pyx_t_16 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_17 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  Py_ssize_t __pyx_t_21;
  __Pyx_RefNannySetupContext("__pyx_fuse_2query", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pyx_fuse_2query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1184, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_13__pyx_fuse_2query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = __Pyx_PyInt_From_npy_int64(__pyx_v_point); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1184, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1184, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1184, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 1184, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_5) {
            __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1184, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1184, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_3query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_4 = __Pyx_PyInt_From_npy_int64(__pyx_v_point); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1184, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_7 = __pyx_t_1; __pyx_t_3 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1184, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1184, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1184, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (__pyx_t_3) {
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_t_4);
          __pyx_t_4 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1184, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pandas/_libs/intervaltree.pxi":1193
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
  __pyx_t_8 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_8) {

    /* "pandas/_libs/intervaltree.pxi":1197
 *             # continue the binary tree structure. Instead, we use linear
 *             # search.
 *             for i in range(self.n_elements):             # <<<<<<<<<<<<<<
 *                 if self.left[i] < point <= self.right[i]:
 *                     result.append(self.indices[i])
 */
    __pyx_t_9 = __pyx_v_self->n_elements;
    __pyx_t_10 = __pyx_t_9;
    for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
      __pyx_v_i = __pyx_t_11;

      /* "pandas/_libs/intervaltree.pxi":1198
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point <= self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      __pyx_t_12 = __pyx_v_i;
      __pyx_t_8 = ((*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_self->left.data + __pyx_t_12 * __pyx_v_self->left.strides[0]) ))) < __pyx_v_point);
      if (__pyx_t_8) {
        __pyx_t_13 = __pyx_v_i;
        __pyx_t_8 = (__pyx_v_point <= (*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_self->right.data + __pyx_t_13 * __pyx_v_self->right.strides[0]) ))));
      }
      __pyx_t_14 = (__pyx_t_8 != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":1199
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point <= self.right[i]:
 *                     result.append(self.indices[i])             # <<<<<<<<<<<<<<
 *         else:
 *             # There are child nodes. Based on comparing our query to the pivot,
 */
        __pyx_t_15 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->indices.data + __pyx_t_15 * __pyx_v_self->indices.strides[0]) ))));

        /* "pandas/_libs/intervaltree.pxi":1198
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point <= self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      }
    }

    /* "pandas/_libs/intervaltree.pxi":1193
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
    goto __pyx_L3;
  }

  /* "pandas/_libs/intervaltree.pxi":1203
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
  /*else*/ {
    __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":1204
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:
 *                 values = self.center_left_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 */
      __pyx_t_16 = __pyx_v_self->center_left_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":1205
 *             if point < self.pivot:
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:
 */
      __pyx_t_17 = __pyx_v_self->center_left_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":1206
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):             # <<<<<<<<<<<<<<
 *                     if not values[i] < point:
 *                         break
 */
      __pyx_t_9 = __pyx_v_self->n_center;
      __pyx_t_10 = __pyx_t_9;
      for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":1207
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_18 = __pyx_v_i;
        __pyx_t_14 = ((!(((*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_18 * __pyx_v_values.strides[0]) ))) < __pyx_v_point) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":1208
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:
 */
          goto __pyx_L9_break;

          /* "pandas/_libs/intervaltree.pxi":1207
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":1209
 *                     if not values[i] < point:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 */
        __pyx_t_19 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_19 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L9_break:;

      /* "pandas/_libs/intervaltree.pxi":1210
 *                         break
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      __pyx_t_14 = ((__pyx_v_point <= __pyx_v_self->left_node->max_right) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":1211
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)             # <<<<<<<<<<<<<<
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode *)__pyx_v_self->left_node->__pyx_vtab)->__pyx_fuse_2query(__pyx_v_self->left_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1211, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":1210
 *                         break
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":1203
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":1212
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
    __pyx_t_14 = ((__pyx_v_point > __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":1213
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 *                 values = self.center_right_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 */
      __pyx_t_16 = __pyx_v_self->center_right_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":1214
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:
 */
      __pyx_t_17 = __pyx_v_self->center_right_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":1215
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):             # <<<<<<<<<<<<<<
 *                     if not point <= values[i]:
 *                         break
 */
      for (__pyx_t_11 = (__pyx_v_self->n_center - 1); __pyx_t_11 > -1L; __pyx_t_11-=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":1216
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_20 = __pyx_v_i;
        __pyx_t_14 = ((!((__pyx_v_point <= (*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_20 * __pyx_v_values.strides[0]) )))) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":1217
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:
 */
          goto __pyx_L13_break;

          /* "pandas/_libs/intervaltree.pxi":1216
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":1218
 *                     if not point <= values[i]:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if self.right_node.min_left < point:
 *                     self.right_node.query(result, point)
 */
        __pyx_t_21 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_21 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L13_break:;

      /* "pandas/_libs/intervaltree.pxi":1219
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      __pyx_t_14 = ((__pyx_v_self->right_node->min_left < __pyx_v_point) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":1220
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:
 *                     self.right_node.query(result, point)             # <<<<<<<<<<<<<<
 *             else:
 *                 result.extend(self.center_left_indices)
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode *)__pyx_v_self->right_node->__pyx_vtab)->__pyx_fuse_2query(__pyx_v_self->right_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1220, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":1219
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":1212
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":1222
 *                     self.right_node.query(result, point)
 *             else:
 *                 result.extend(self.center_left_indices)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
    /*else*/ {
      __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->extend(__pyx_v_result, __pyx_v_self->center_left_indices); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1222, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_L7:;
  }
  __pyx_L3:;

  /* "pandas/_libs/intervaltree.pxi":1184
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __PYX_XDEC_MEMVIEW(&__pyx_t_16, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_17, 1);
  __Pyx_AddTraceback("pandas._libs.interval.Float64ClosedRightIntervalNode.query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_indices, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_values, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_13__pyx_fuse_2query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_2__pyx_mdef_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_13__pyx_fuse_2query = {"__pyx_fuse_2query", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_13__pyx_fuse_2query, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_2query};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_13__pyx_fuse_2query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result = 0;
  __pyx_t_5numpy_int64_t __pyx_v_point;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fuse_2query (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_result,&__pyx_n_s_point,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_result)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_point)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_2query", 1, 2, 2, 1); __PYX_ERR(1, 1184, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_2query") < 0)) __PYX_ERR(1, 1184, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_result = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)values[0]);
    __pyx_v_point = __Pyx_PyInt_As_npy_int64(values[1]); if (unlikely((__pyx_v_point == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(1, 1184, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_2query", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 1184, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Float64ClosedRightIntervalNode.__pyx_fuse_2query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_result), __pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector, 1, "result", 0))) __PYX_ERR(1, 1184, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_12__pyx_fuse_2query(((struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode *)__pyx_v_self), __pyx_v_result, __pyx_v_point);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_12__pyx_fuse_2query(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int64_t __pyx_v_point) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__pyx_fuse_2query", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_fuse_2__pyx_f_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_query(__pyx_v_self, __pyx_v_result, __pyx_v_point, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1184, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Float64ClosedRightIntervalNode.__pyx_fuse_2query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_15__pyx_fuse_3query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_fuse_3__pyx_f_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int32_t __pyx_v_point, int __pyx_skip_dispatch) {
  __Pyx_memviewslice __pyx_v_indices = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_values = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __pyx_t_5numpy_int64_t __pyx_t_9;
  __pyx_t_5numpy_int64_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  int __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  __Pyx_memviewslice __pyx_t_16 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_17 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  Py_ssize_t __pyx_t_21;
  __Pyx_RefNannySetupContext("__pyx_fuse_3query", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pyx_fuse_3query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1184, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_15__pyx_fuse_3query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = __Pyx_PyInt_From_npy_int32(__pyx_v_point); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1184, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1184, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1184, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 1184, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_5) {
            __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1184, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1184, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_3query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_4 = __Pyx_PyInt_From_npy_int32(__pyx_v_point); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1184, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_7 = __pyx_t_1; __pyx_t_3 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1184, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1184, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1184, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (__pyx_t_3) {
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_t_4);
          __pyx_t_4 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1184, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pandas/_libs/intervaltree.pxi":1193
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
  __pyx_t_8 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_8) {

    /* "pandas/_libs/intervaltree.pxi":1197
 *             # continue the binary tree structure. Instead, we use linear
 *             # search.
 *             for i in range(self.n_elements):             # <<<<<<<<<<<<<<
 *                 if self.left[i] < point <= self.right[i]:
 *                     result.append(self.indices[i])
 */
    __pyx_t_9 = __pyx_v_self->n_elements;
    __pyx_t_10 = __pyx_t_9;
    for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
      __pyx_v_i = __pyx_t_11;

      /* "pandas/_libs/intervaltree.pxi":1198
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point <= self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      __pyx_t_12 = __pyx_v_i;
      __pyx_t_8 = ((*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_self->left.data + __pyx_t_12 * __pyx_v_self->left.strides[0]) ))) < __pyx_v_point);
      if (__pyx_t_8) {
        __pyx_t_13 = __pyx_v_i;
        __pyx_t_8 = (__pyx_v_point <= (*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_self->right.data + __pyx_t_13 * __pyx_v_self->right.strides[0]) ))));
      }
      __pyx_t_14 = (__pyx_t_8 != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":1199
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point <= self.right[i]:
 *                     result.append(self.indices[i])             # <<<<<<<<<<<<<<
 *         else:
 *             # There are child nodes. Based on comparing our query to the pivot,
 */
        __pyx_t_15 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->indices.data + __pyx_t_15 * __pyx_v_self->indices.strides[0]) ))));

        /* "pandas/_libs/intervaltree.pxi":1198
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point <= self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      }
    }

    /* "pandas/_libs/intervaltree.pxi":1193
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
    goto __pyx_L3;
  }

  /* "pandas/_libs/intervaltree.pxi":1203
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
  /*else*/ {
    __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":1204
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:
 *                 values = self.center_left_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 */
      __pyx_t_16 = __pyx_v_self->center_left_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":1205
 *             if point < self.pivot:
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:
 */
      __pyx_t_17 = __pyx_v_self->center_left_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":1206
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):             # <<<<<<<<<<<<<<
 *                     if not values[i] < point:
 *                         break
 */
      __pyx_t_9 = __pyx_v_self->n_center;
      __pyx_t_10 = __pyx_t_9;
      for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":1207
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_18 = __pyx_v_i;
        __pyx_t_14 = ((!(((*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_18 * __pyx_v_values.strides[0]) ))) < __pyx_v_point) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":1208
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:
 */
          goto __pyx_L9_break;

          /* "pandas/_libs/intervaltree.pxi":1207
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":1209
 *                     if not values[i] < point:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 */
        __pyx_t_19 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_19 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L9_break:;

      /* "pandas/_libs/intervaltree.pxi":1210
 *                         break
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      __pyx_t_14 = ((__pyx_v_point <= __pyx_v_self->left_node->max_right) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":1211
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)             # <<<<<<<<<<<<<<
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode *)__pyx_v_self->left_node->__pyx_vtab)->__pyx_fuse_3query(__pyx_v_self->left_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1211, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":1210
 *                         break
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":1203
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":1212
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
    __pyx_t_14 = ((__pyx_v_point > __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":1213
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 *                 values = self.center_right_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 */
      __pyx_t_16 = __pyx_v_self->center_right_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":1214
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:
 */
      __pyx_t_17 = __pyx_v_self->center_right_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":1215
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):             # <<<<<<<<<<<<<<
 *                     if not point <= values[i]:
 *                         break
 */
      for (__pyx_t_11 = (__pyx_v_self->n_center - 1); __pyx_t_11 > -1L; __pyx_t_11-=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":1216
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_20 = __pyx_v_i;
        __pyx_t_14 = ((!((__pyx_v_point <= (*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_20 * __pyx_v_values.strides[0]) )))) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":1217
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:
 */
          goto __pyx_L13_break;

          /* "pandas/_libs/intervaltree.pxi":1216
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":1218
 *                     if not point <= values[i]:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if self.right_node.min_left < point:
 *                     self.right_node.query(result, point)
 */
        __pyx_t_21 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_21 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L13_break:;

      /* "pandas/_libs/intervaltree.pxi":1219
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      __pyx_t_14 = ((__pyx_v_self->right_node->min_left < __pyx_v_point) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":1220
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:
 *                     self.right_node.query(result, point)             # <<<<<<<<<<<<<<
 *             else:
 *                 result.extend(self.center_left_indices)
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode *)__pyx_v_self->right_node->__pyx_vtab)->__pyx_fuse_3query(__pyx_v_self->right_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1220, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":1219
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":1212
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":1222
 *                     self.right_node.query(result, point)
 *             else:
 *                 result.extend(self.center_left_indices)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
    /*else*/ {
      __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->extend(__pyx_v_result, __pyx_v_self->center_left_indices); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1222, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_L7:;
  }
  __pyx_L3:;

  /* "pandas/_libs/intervaltree.pxi":1184
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __PYX_XDEC_MEMVIEW(&__pyx_t_16, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_17, 1);
  __Pyx_AddTraceback("pandas._libs.interval.Float64ClosedRightIntervalNode.query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_indices, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_values, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_15__pyx_fuse_3query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_3__pyx_mdef_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_15__pyx_fuse_3query = {"__pyx_fuse_3query", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_15__pyx_fuse_3query, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_2query};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_15__pyx_fuse_3query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result = 0;
  __pyx_t_5numpy_int32_t __pyx_v_point;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fuse_3query (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_result,&__pyx_n_s_point,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_result)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_point)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_3query", 1, 2, 2, 1); __PYX_ERR(1, 1184, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_3query") < 0)) __PYX_ERR(1, 1184, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_result = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)values[0]);
    __pyx_v_point = __Pyx_PyInt_As_npy_int32(values[1]); if (unlikely((__pyx_v_point == ((npy_int32)-1)) && PyErr_Occurred())) __PYX_ERR(1, 1184, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_3query", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 1184, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Float64ClosedRightIntervalNode.__pyx_fuse_3query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_result), __pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector, 1, "result", 0))) __PYX_ERR(1, 1184, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_14__pyx_fuse_3query(((struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode *)__pyx_v_self), __pyx_v_result, __pyx_v_point);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_14__pyx_fuse_3query(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int32_t __pyx_v_point) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__pyx_fuse_3query", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_fuse_3__pyx_f_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_query(__pyx_v_self, __pyx_v_result, __pyx_v_point, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1184, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Float64ClosedRightIntervalNode.__pyx_fuse_3query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_17__pyx_fuse_4query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_fuse_4__pyx_f_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_uint64_t __pyx_v_point, int __pyx_skip_dispatch) {
  __Pyx_memviewslice __pyx_v_indices = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_values = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __pyx_t_5numpy_int64_t __pyx_t_9;
  __pyx_t_5numpy_int64_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  int __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  __Pyx_memviewslice __pyx_t_16 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_17 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  Py_ssize_t __pyx_t_21;
  __Pyx_RefNannySetupContext("__pyx_fuse_4query", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pyx_fuse_4query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1184, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_17__pyx_fuse_4query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = __Pyx_PyInt_From_npy_uint64(__pyx_v_point); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1184, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1184, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1184, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 1184, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_5) {
            __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1184, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1184, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_3query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_4 = __Pyx_PyInt_From_npy_uint64(__pyx_v_point); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1184, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_7 = __pyx_t_1; __pyx_t_3 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1184, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1184, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1184, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (__pyx_t_3) {
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_t_4);
          __pyx_t_4 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1184, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pandas/_libs/intervaltree.pxi":1193
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
  __pyx_t_8 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_8) {

    /* "pandas/_libs/intervaltree.pxi":1197
 *             # continue the binary tree structure. Instead, we use linear
 *             # search.
 *             for i in range(self.n_elements):             # <<<<<<<<<<<<<<
 *                 if self.left[i] < point <= self.right[i]:
 *                     result.append(self.indices[i])
 */
    __pyx_t_9 = __pyx_v_self->n_elements;
    __pyx_t_10 = __pyx_t_9;
    for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
      __pyx_v_i = __pyx_t_11;

      /* "pandas/_libs/intervaltree.pxi":1198
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point <= self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      __pyx_t_12 = __pyx_v_i;
      __pyx_t_8 = ((*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_self->left.data + __pyx_t_12 * __pyx_v_self->left.strides[0]) ))) < __pyx_v_point);
      if (__pyx_t_8) {
        __pyx_t_13 = __pyx_v_i;
        __pyx_t_8 = (__pyx_v_point <= (*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_self->right.data + __pyx_t_13 * __pyx_v_self->right.strides[0]) ))));
      }
      __pyx_t_14 = (__pyx_t_8 != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":1199
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point <= self.right[i]:
 *                     result.append(self.indices[i])             # <<<<<<<<<<<<<<
 *         else:
 *             # There are child nodes. Based on comparing our query to the pivot,
 */
        __pyx_t_15 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->indices.data + __pyx_t_15 * __pyx_v_self->indices.strides[0]) ))));

        /* "pandas/_libs/intervaltree.pxi":1198
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point <= self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      }
    }

    /* "pandas/_libs/intervaltree.pxi":1193
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
    goto __pyx_L3;
  }

  /* "pandas/_libs/intervaltree.pxi":1203
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
  /*else*/ {
    __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":1204
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:
 *                 values = self.center_left_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 */
      __pyx_t_16 = __pyx_v_self->center_left_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":1205
 *             if point < self.pivot:
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:
 */
      __pyx_t_17 = __pyx_v_self->center_left_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":1206
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):             # <<<<<<<<<<<<<<
 *                     if not values[i] < point:
 *                         break
 */
      __pyx_t_9 = __pyx_v_self->n_center;
      __pyx_t_10 = __pyx_t_9;
      for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":1207
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_18 = __pyx_v_i;
        __pyx_t_14 = ((!(((*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_18 * __pyx_v_values.strides[0]) ))) < __pyx_v_point) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":1208
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:
 */
          goto __pyx_L9_break;

          /* "pandas/_libs/intervaltree.pxi":1207
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":1209
 *                     if not values[i] < point:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 */
        __pyx_t_19 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_19 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L9_break:;

      /* "pandas/_libs/intervaltree.pxi":1210
 *                         break
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      __pyx_t_14 = ((__pyx_v_point <= __pyx_v_self->left_node->max_right) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":1211
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)             # <<<<<<<<<<<<<<
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode *)__pyx_v_self->left_node->__pyx_vtab)->__pyx_fuse_4query(__pyx_v_self->left_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1211, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":1210
 *                         break
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":1203
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":1212
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
    __pyx_t_14 = ((__pyx_v_point > __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":1213
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 *                 values = self.center_right_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 */
      __pyx_t_16 = __pyx_v_self->center_right_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":1214
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:
 */
      __pyx_t_17 = __pyx_v_self->center_right_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":1215
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):             # <<<<<<<<<<<<<<
 *                     if not point <= values[i]:
 *                         break
 */
      for (__pyx_t_11 = (__pyx_v_self->n_center - 1); __pyx_t_11 > -1L; __pyx_t_11-=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":1216
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_20 = __pyx_v_i;
        __pyx_t_14 = ((!((__pyx_v_point <= (*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_20 * __pyx_v_values.strides[0]) )))) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":1217
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:
 */
          goto __pyx_L13_break;

          /* "pandas/_libs/intervaltree.pxi":1216
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":1218
 *                     if not point <= values[i]:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if self.right_node.min_left < point:
 *                     self.right_node.query(result, point)
 */
        __pyx_t_21 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_21 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L13_break:;

      /* "pandas/_libs/intervaltree.pxi":1219
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      __pyx_t_14 = ((__pyx_v_self->right_node->min_left < __pyx_v_point) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":1220
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:
 *                     self.right_node.query(result, point)             # <<<<<<<<<<<<<<
 *             else:
 *                 result.extend(self.center_left_indices)
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode *)__pyx_v_self->right_node->__pyx_vtab)->__pyx_fuse_4query(__pyx_v_self->right_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1220, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":1219
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":1212
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":1222
 *                     self.right_node.query(result, point)
 *             else:
 *                 result.extend(self.center_left_indices)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
    /*else*/ {
      __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->extend(__pyx_v_result, __pyx_v_self->center_left_indices); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1222, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_L7:;
  }
  __pyx_L3:;

  /* "pandas/_libs/intervaltree.pxi":1184
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __PYX_XDEC_MEMVIEW(&__pyx_t_16, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_17, 1);
  __Pyx_AddTraceback("pandas._libs.interval.Float64ClosedRightIntervalNode.query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_indices, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_values, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_17__pyx_fuse_4query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_4__pyx_mdef_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_17__pyx_fuse_4query = {"__pyx_fuse_4query", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_17__pyx_fuse_4query, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_2query};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_17__pyx_fuse_4query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result = 0;
  __pyx_t_5numpy_uint64_t __pyx_v_point;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fuse_4query (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_result,&__pyx_n_s_point,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_result)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_point)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_4query", 1, 2, 2, 1); __PYX_ERR(1, 1184, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_4query") < 0)) __PYX_ERR(1, 1184, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_result = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)values[0]);
    __pyx_v_point = __Pyx_PyInt_As_npy_uint64(values[1]); if (unlikely((__pyx_v_point == ((npy_uint64)-1)) && PyErr_Occurred())) __PYX_ERR(1, 1184, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_4query", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 1184, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Float64ClosedRightIntervalNode.__pyx_fuse_4query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_result), __pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector, 1, "result", 0))) __PYX_ERR(1, 1184, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_16__pyx_fuse_4query(((struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode *)__pyx_v_self), __pyx_v_result, __pyx_v_point);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_16__pyx_fuse_4query(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_uint64_t __pyx_v_point) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__pyx_fuse_4query", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_fuse_4__pyx_f_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_query(__pyx_v_self, __pyx_v_result, __pyx_v_point, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1184, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Float64ClosedRightIntervalNode.__pyx_fuse_4query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":1224
 *                 result.extend(self.center_left_indices)
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         if self.is_leaf_node:
 *             return ('<Float64ClosedRightIntervalNode: '
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_5__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_5__repr__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_4__repr__(((struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_4__repr__(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode *__pyx_v_self) {
  PyObject *__pyx_v_n_left = NULL;
  PyObject *__pyx_v_n_right = NULL;
  PyObject *__pyx_v_n_center = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  Py_UCS4 __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "pandas/_libs/intervaltree.pxi":1225
 * 
 *     def __repr__(self):
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             return ('<Float64ClosedRightIntervalNode: '
 *                     '%s elements (terminal)>' % self.n_elements)
 */
  __pyx_t_1 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_1) {

    /* "pandas/_libs/intervaltree.pxi":1226
 *     def __repr__(self):
 *         if self.is_leaf_node:
 *             return ('<Float64ClosedRightIntervalNode: '             # <<<<<<<<<<<<<<
 *                     '%s elements (terminal)>' % self.n_elements)
 *         else:
 */
    __Pyx_XDECREF(__pyx_r);

    /* "pandas/_libs/intervaltree.pxi":1227
 *         if self.is_leaf_node:
 *             return ('<Float64ClosedRightIntervalNode: '
 *                     '%s elements (terminal)>' % self.n_elements)             # <<<<<<<<<<<<<<
 *         else:
 *             n_left = self.left_node.n_elements
 */
    __pyx_t_2 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_elements); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1227, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PyUnicode_Format(__pyx_kp_u_Float64ClosedRightIntervalNode, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1227, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pandas/_libs/intervaltree.pxi":1225
 * 
 *     def __repr__(self):
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             return ('<Float64ClosedRightIntervalNode: '
 *                     '%s elements (terminal)>' % self.n_elements)
 */
  }

  /* "pandas/_libs/intervaltree.pxi":1229
 *                     '%s elements (terminal)>' % self.n_elements)
 *         else:
 *             n_left = self.left_node.n_elements             # <<<<<<<<<<<<<<
 *             n_right = self.right_node.n_elements
 *             n_center = self.n_elements - n_left - n_right
 */
  /*else*/ {
    __pyx_t_3 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->left_node->n_elements); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1229, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_v_n_left = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "pandas/_libs/intervaltree.pxi":1230
 *         else:
 *             n_left = self.left_node.n_elements
 *             n_right = self.right_node.n_elements             # <<<<<<<<<<<<<<
 *             n_center = self.n_elements - n_left - n_right
 *             return ('<Float64ClosedRightIntervalNode: '
 */
    __pyx_t_3 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->right_node->n_elements); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1230, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_v_n_right = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "pandas/_libs/intervaltree.pxi":1231
 *             n_left = self.left_node.n_elements
 *             n_right = self.right_node.n_elements
 *             n_center = self.n_elements - n_left - n_right             # <<<<<<<<<<<<<<
 *             return ('<Float64ClosedRightIntervalNode: '
 *                     'pivot %s, %s elements (%s left, %s right, %s '
 */
    __pyx_t_3 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_elements); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1231, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = PyNumber_Subtract(__pyx_t_3, __pyx_v_n_left); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1231, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyNumber_Subtract(__pyx_t_2, __pyx_v_n_right); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1231, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_n_center = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "pandas/_libs/intervaltree.pxi":1232
 *             n_right = self.right_node.n_elements
 *             n_center = self.n_elements - n_left - n_right
 *             return ('<Float64ClosedRightIntervalNode: '             # <<<<<<<<<<<<<<
 *                     'pivot %s, %s elements (%s left, %s right, %s '
 *                     'overlapping)>' % (self.pivot, self.n_elements,
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = PyTuple_New(11); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1232, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = 0;
    __pyx_t_5 = 127;
    __Pyx_INCREF(__pyx_kp_u_Float64ClosedRightIntervalNode_2);
    __pyx_t_4 += 39;
    __Pyx_GIVEREF(__pyx_kp_u_Float64ClosedRightIntervalNode_2);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_kp_u_Float64ClosedRightIntervalNode_2);

    /* "pandas/_libs/intervaltree.pxi":1234
 *             return ('<Float64ClosedRightIntervalNode: '
 *                     'pivot %s, %s elements (%s left, %s right, %s '
 *                     'overlapping)>' % (self.pivot, self.n_elements,             # <<<<<<<<<<<<<<
 *                                        n_left, n_right, n_center))
 * 
 */
    __pyx_t_2 = PyFloat_FromDouble(__pyx_v_self->pivot); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1234, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_t_2), __pyx_empty_unicode); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 1234, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_5 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_6) > __pyx_t_5) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_6) : __pyx_t_5;
    __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_6);
    __pyx_t_6 = 0;
    __Pyx_INCREF(__pyx_kp_u__18);
    __pyx_t_4 += 2;
    __Pyx_GIVEREF(__pyx_kp_u__18);
    PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_kp_u__18);
    __pyx_t_6 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_elements); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 1234, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_2 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_t_6), __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1234, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_5 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) > __pyx_t_5) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) : __pyx_t_5;
    __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_3, 3, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_elements);
    __pyx_t_4 += 11;
    __Pyx_GIVEREF(__pyx_kp_u_elements);
    PyTuple_SET_ITEM(__pyx_t_3, 4, __pyx_kp_u_elements);

    /* "pandas/_libs/intervaltree.pxi":1235
 *                     'pivot %s, %s elements (%s left, %s right, %s '
 *                     'overlapping)>' % (self.pivot, self.n_elements,
 *                                        n_left, n_right, n_center))             # <<<<<<<<<<<<<<
 * 
 *     def counts(self):
 */
    __pyx_t_2 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_v_n_left), __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1235, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) > __pyx_t_5) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) : __pyx_t_5;
    __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_3, 5, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_left_2);
    __pyx_t_4 += 7;
    __Pyx_GIVEREF(__pyx_kp_u_left_2);
    PyTuple_SET_ITEM(__pyx_t_3, 6, __pyx_kp_u_left_2);
    __pyx_t_2 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_v_n_right), __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1235, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) > __pyx_t_5) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) : __pyx_t_5;
    __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_3, 7, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_right_2);
    __pyx_t_4 += 8;
    __Pyx_GIVEREF(__pyx_kp_u_right_2);
    PyTuple_SET_ITEM(__pyx_t_3, 8, __pyx_kp_u_right_2);
    __pyx_t_2 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_v_n_center), __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1235, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) > __pyx_t_5) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) : __pyx_t_5;
    __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_3, 9, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_overlapping);
    __pyx_t_4 += 14;
    __Pyx_GIVEREF(__pyx_kp_u_overlapping);
    PyTuple_SET_ITEM(__pyx_t_3, 10, __pyx_kp_u_overlapping);

    /* "pandas/_libs/intervaltree.pxi":1232
 *             n_right = self.right_node.n_elements
 *             n_center = self.n_elements - n_left - n_right
 *             return ('<Float64ClosedRightIntervalNode: '             # <<<<<<<<<<<<<<
 *                     'pivot %s, %s elements (%s left, %s right, %s '
 *                     'overlapping)>' % (self.pivot, self.n_elements,
 */
    __pyx_t_2 = __Pyx_PyUnicode_Join(__pyx_t_3, 11, __pyx_t_4, __pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1232, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;
  }

  /* "pandas/_libs/intervaltree.pxi":1224
 *                 result.extend(self.center_left_indices)
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         if self.is_leaf_node:
 *             return ('<Float64ClosedRightIntervalNode: '
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pandas._libs.interval.Float64ClosedRightIntervalNode.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_n_left);
  __Pyx_XDECREF(__pyx_v_n_right);
  __Pyx_XDECREF(__pyx_v_n_center);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":1237
 *                                        n_left, n_right, n_center))
 * 
 *     def counts(self):             # <<<<<<<<<<<<<<
 *         """
 *         Inspect counts on this node
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_7counts(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_6counts[] = "\n        Inspect counts on this node\n        useful for debugging purposes\n        ";
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_7counts(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("counts (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_6counts(((struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_6counts(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode *__pyx_v_self) {
  Py_ssize_t __pyx_v_m;
  PyObject *__pyx_v_l = NULL;
  PyObject *__pyx_v_r = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  size_t __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("counts", 0);

  /* "pandas/_libs/intervaltree.pxi":1242
 *         useful for debugging purposes
 *         """
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             return self.n_elements
 *         else:
 */
  __pyx_t_1 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_1) {

    /* "pandas/_libs/intervaltree.pxi":1243
 *         """
 *         if self.is_leaf_node:
 *             return self.n_elements             # <<<<<<<<<<<<<<
 *         else:
 *             m = len(self.center_left_values)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_elements); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1243, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "pandas/_libs/intervaltree.pxi":1242
 *         useful for debugging purposes
 *         """
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             return self.n_elements
 *         else:
 */
  }

  /* "pandas/_libs/intervaltree.pxi":1245
 *             return self.n_elements
 *         else:
 *             m = len(self.center_left_values)             # <<<<<<<<<<<<<<
 *             l = self.left_node.counts()
 *             r = self.right_node.counts()
 */
  /*else*/ {
    if (unlikely(!__pyx_v_self->center_left_values.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(1, 1245, __pyx_L1_error)}
    __pyx_t_3 = __Pyx_MemoryView_Len(__pyx_v_self->center_left_values); 
    __pyx_v_m = __pyx_t_3;

    /* "pandas/_libs/intervaltree.pxi":1246
 *         else:
 *             m = len(self.center_left_values)
 *             l = self.left_node.counts()             # <<<<<<<<<<<<<<
 *             r = self.right_node.counts()
 *             return (m, (l, r))
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->left_node), __pyx_n_s_counts); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1246, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1246, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_l = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "pandas/_libs/intervaltree.pxi":1247
 *             m = len(self.center_left_values)
 *             l = self.left_node.counts()
 *             r = self.right_node.counts()             # <<<<<<<<<<<<<<
 *             return (m, (l, r))
 * 
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->right_node), __pyx_n_s_counts); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1247, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1247, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_r = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "pandas/_libs/intervaltree.pxi":1248
 *             l = self.left_node.counts()
 *             r = self.right_node.counts()
 *             return (m, (l, r))             # <<<<<<<<<<<<<<
 * 
 * NODE_CLASSES['float64',
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = PyInt_FromSsize_t(__pyx_v_m); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1248, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1248, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_v_l);
    __Pyx_GIVEREF(__pyx_v_l);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_l);
    __Pyx_INCREF(__pyx_v_r);
    __Pyx_GIVEREF(__pyx_v_r);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_v_r);
    __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1248, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_4);
    __pyx_t_2 = 0;
    __pyx_t_4 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;
  }

  /* "pandas/_libs/intervaltree.pxi":1237
 *                                        n_left, n_right, n_center))
 * 
 *     def counts(self):             # <<<<<<<<<<<<<<
 *         """
 *         Inspect counts on this node
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pandas._libs.interval.Float64ClosedRightIntervalNode.counts", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_l);
  __Pyx_XDECREF(__pyx_v_r);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":1094
 *         int64_t[:] center_left_indices, center_right_indices, indices
 *         float64_t min_left, max_right
 *         readonly float64_t pivot             # <<<<<<<<<<<<<<
 *         readonly int64_t n_elements, n_center, leaf_size
 *         readonly bint is_leaf_node
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_5pivot_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_5pivot_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_5pivot___get__(((struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_5pivot___get__(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->pivot); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1094, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Float64ClosedRightIntervalNode.pivot.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":1095
 *         float64_t min_left, max_right
 *         readonly float64_t pivot
 *         readonly int64_t n_elements, n_center, leaf_size             # <<<<<<<<<<<<<<
 *         readonly bint is_leaf_node
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_10n_elements_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_10n_elements_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_10n_elements___get__(((struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_10n_elements___get__(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_elements); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1095, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Float64ClosedRightIntervalNode.n_elements.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_8n_center_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_8n_center_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_8n_center___get__(((struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_8n_center___get__(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_center); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1095, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Float64ClosedRightIntervalNode.n_center.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_9leaf_size_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_9leaf_size_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_9leaf_size___get__(((struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_9leaf_size___get__(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->leaf_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1095, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Float64ClosedRightIntervalNode.leaf_size.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":1096
 *         readonly float64_t pivot
 *         readonly int64_t n_elements, n_center, leaf_size
 *         readonly bint is_leaf_node             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self,
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_12is_leaf_node_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_12is_leaf_node_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_12is_leaf_node___get__(((struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_12is_leaf_node___get__(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->is_leaf_node); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1096, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Float64ClosedRightIntervalNode.is_leaf_node.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_21__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_21__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_20__reduce_cython__(((struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_20__reduce_cython__(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  int __pyx_t_16;
  int __pyx_t_17;
  int __pyx_t_18;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.center_left_indices, self.center_left_values, self.center_right_indices, self.center_right_values, self.indices, self.is_leaf_node, self.leaf_size, self.left, self.left_node, self.max_right, self.min_left, self.n_center, self.n_elements, self.pivot, self.right, self.right_node)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  if (unlikely(!__pyx_v_self->center_left_indices.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_1 = __pyx_memoryview_fromslice(__pyx_v_self->center_left_indices, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_int64_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_int64_t, 0);; if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (unlikely(!__pyx_v_self->center_left_values.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_2 = __pyx_memoryview_fromslice(__pyx_v_self->center_left_values, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_float64_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_float64_t, 0);; if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (unlikely(!__pyx_v_self->center_right_indices.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_3 = __pyx_memoryview_fromslice(__pyx_v_self->center_right_indices, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_int64_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_int64_t, 0);; if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely(!__pyx_v_self->center_right_values.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_4 = __pyx_memoryview_fromslice(__pyx_v_self->center_right_values, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_float64_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_float64_t, 0);; if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (unlikely(!__pyx_v_self->indices.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_5 = __pyx_memoryview_fromslice(__pyx_v_self->indices, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_int64_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_int64_t, 0);; if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyBool_FromLong(__pyx_v_self->is_leaf_node); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->leaf_size); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (unlikely(!__pyx_v_self->left.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_8 = __pyx_memoryview_fromslice(__pyx_v_self->left, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_float64_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_float64_t, 0);; if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = PyFloat_FromDouble(__pyx_v_self->max_right); if (unlikely(!__pyx_t_9)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_10 = PyFloat_FromDouble(__pyx_v_self->min_left); if (unlikely(!__pyx_t_10)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_11 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_center); if (unlikely(!__pyx_t_11)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_12 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_elements); if (unlikely(!__pyx_t_12)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __pyx_t_13 = PyFloat_FromDouble(__pyx_v_self->pivot); if (unlikely(!__pyx_t_13)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  if (unlikely(!__pyx_v_self->right.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_14 = __pyx_memoryview_fromslice(__pyx_v_self->right, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_float64_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_float64_t, 0);; if (unlikely(!__pyx_t_14)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __pyx_t_15 = PyTuple_New(16); if (unlikely(!__pyx_t_15)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_15, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_15, 1, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_15, 2, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_15, 3, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_15, 4, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_15, 5, __pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_15, 6, __pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_15, 7, __pyx_t_8);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->left_node));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self->left_node));
  PyTuple_SET_ITEM(__pyx_t_15, 8, ((PyObject *)__pyx_v_self->left_node));
  __Pyx_GIVEREF(__pyx_t_9);
  PyTuple_SET_ITEM(__pyx_t_15, 9, __pyx_t_9);
  __Pyx_GIVEREF(__pyx_t_10);
  PyTuple_SET_ITEM(__pyx_t_15, 10, __pyx_t_10);
  __Pyx_GIVEREF(__pyx_t_11);
  PyTuple_SET_ITEM(__pyx_t_15, 11, __pyx_t_11);
  __Pyx_GIVEREF(__pyx_t_12);
  PyTuple_SET_ITEM(__pyx_t_15, 12, __pyx_t_12);
  __Pyx_GIVEREF(__pyx_t_13);
  PyTuple_SET_ITEM(__pyx_t_15, 13, __pyx_t_13);
  __Pyx_GIVEREF(__pyx_t_14);
  PyTuple_SET_ITEM(__pyx_t_15, 14, __pyx_t_14);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->right_node));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self->right_node));
  PyTuple_SET_ITEM(__pyx_t_15, 15, ((PyObject *)__pyx_v_self->right_node));
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  __pyx_t_6 = 0;
  __pyx_t_7 = 0;
  __pyx_t_8 = 0;
  __pyx_t_9 = 0;
  __pyx_t_10 = 0;
  __pyx_t_11 = 0;
  __pyx_t_12 = 0;
  __pyx_t_13 = 0;
  __pyx_t_14 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_15);
  __pyx_t_15 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.center_left_indices, self.center_left_values, self.center_right_indices, self.center_right_values, self.indices, self.is_leaf_node, self.leaf_size, self.left, self.left_node, self.max_right, self.min_left, self.n_center, self.n_elements, self.pivot, self.right, self.right_node)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_15 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_15)) __PYX_ERR(2, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __pyx_v__dict = __pyx_t_15;
  __pyx_t_15 = 0;

  /* "(tree fragment)":7
 *     state = (self.center_left_indices, self.center_left_values, self.center_right_indices, self.center_right_values, self.indices, self.is_leaf_node, self.leaf_size, self.left, self.left_node, self.max_right, self.min_left, self.n_center, self.n_elements, self.pivot, self.right, self.right_node)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_16 = (__pyx_v__dict != Py_None);
  __pyx_t_17 = (__pyx_t_16 != 0);
  if (__pyx_t_17) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_15 = PyTuple_New(1); if (unlikely(!__pyx_t_15)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    PyTuple_SET_ITEM(__pyx_t_15, 0, __pyx_v__dict);
    __pyx_t_14 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_15); if (unlikely(!__pyx_t_14)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_14));
    __pyx_t_14 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.left_node is not None or self.right_node is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.center_left_indices, self.center_left_values, self.center_right_indices, self.center_right_values, self.indices, self.is_leaf_node, self.leaf_size, self.left, self.left_node, self.max_right, self.min_left, self.n_center, self.n_elements, self.pivot, self.right, self.right_node)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.left_node is not None or self.right_node is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_Float64ClosedRightIntervalNode, (type(self), 0x7a104e5, None), state
 */
  /*else*/ {
    __pyx_t_16 = (((PyObject *)__pyx_v_self->left_node) != Py_None);
    __pyx_t_18 = (__pyx_t_16 != 0);
    if (!__pyx_t_18) {
    } else {
      __pyx_t_17 = __pyx_t_18;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_18 = (((PyObject *)__pyx_v_self->right_node) != Py_None);
    __pyx_t_16 = (__pyx_t_18 != 0);
    __pyx_t_17 = __pyx_t_16;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_17;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.left_node is not None or self.right_node is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Float64ClosedRightIntervalNode, (type(self), 0x7a104e5, None), state
 *     else:
 */
  __pyx_t_17 = (__pyx_v_use_setstate != 0);
  if (__pyx_t_17) {

    /* "(tree fragment)":13
 *         use_setstate = self.left_node is not None or self.right_node is not None
 *     if use_setstate:
 *         return __pyx_unpickle_Float64ClosedRightIntervalNode, (type(self), 0x7a104e5, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_Float64ClosedRightIntervalNode, (type(self), 0x7a104e5, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_14, __pyx_n_s_pyx_unpickle_Float64ClosedRigh); if (unlikely(!__pyx_t_14)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __pyx_t_15 = PyTuple_New(3); if (unlikely(!__pyx_t_15)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_15, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_127993061);
    __Pyx_GIVEREF(__pyx_int_127993061);
    PyTuple_SET_ITEM(__pyx_t_15, 1, __pyx_int_127993061);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_15, 2, Py_None);
    __pyx_t_13 = PyTuple_New(3); if (unlikely(!__pyx_t_13)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __Pyx_GIVEREF(__pyx_t_14);
    PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_t_14);
    __Pyx_GIVEREF(__pyx_t_15);
    PyTuple_SET_ITEM(__pyx_t_13, 1, __pyx_t_15);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_13, 2, __pyx_v_state);
    __pyx_t_14 = 0;
    __pyx_t_15 = 0;
    __pyx_r = __pyx_t_13;
    __pyx_t_13 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.left_node is not None or self.right_node is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Float64ClosedRightIntervalNode, (type(self), 0x7a104e5, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_Float64ClosedRightIntervalNode, (type(self), 0x7a104e5, None), state
 *     else:
 *         return __pyx_unpickle_Float64ClosedRightIntervalNode, (type(self), 0x7a104e5, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Float64ClosedRightIntervalNode__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_13, __pyx_n_s_pyx_unpickle_Float64ClosedRigh); if (unlikely(!__pyx_t_13)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __pyx_t_15 = PyTuple_New(3); if (unlikely(!__pyx_t_15)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_15, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_127993061);
    __Pyx_GIVEREF(__pyx_int_127993061);
    PyTuple_SET_ITEM(__pyx_t_15, 1, __pyx_int_127993061);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_15, 2, __pyx_v_state);
    __pyx_t_14 = PyTuple_New(2); if (unlikely(!__pyx_t_14)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_GIVEREF(__pyx_t_13);
    PyTuple_SET_ITEM(__pyx_t_14, 0, __pyx_t_13);
    __Pyx_GIVEREF(__pyx_t_15);
    PyTuple_SET_ITEM(__pyx_t_14, 1, __pyx_t_15);
    __pyx_t_13 = 0;
    __pyx_t_15 = 0;
    __pyx_r = __pyx_t_14;
    __pyx_t_14 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_AddTraceback("pandas._libs.interval.Float64ClosedRightIntervalNode.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Float64ClosedRightIntervalNode, (type(self), 0x7a104e5, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Float64ClosedRightIntervalNode__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_23__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_23__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_22__setstate_cython__(((struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_22__setstate_cython__(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_Float64ClosedRightIntervalNode, (type(self), 0x7a104e5, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Float64ClosedRightIntervalNode__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(2, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_6pandas_5_libs_8interval___pyx_unpickle_Float64ClosedRightIntervalNode__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Float64ClosedRightIntervalNode, (type(self), 0x7a104e5, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Float64ClosedRightIntervalNode__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Float64ClosedRightIntervalNode.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":1268
 *         readonly bint is_leaf_node
 * 
 *     def __init__(self,             # <<<<<<<<<<<<<<
 *                  ndarray[float64_t, ndim=1] left,
 *                  ndarray[float64_t, ndim=1] right,
 */

/* Python wrapper */
static int __pyx_pw_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyArrayObject *__pyx_v_left = 0;
  PyArrayObject *__pyx_v_right = 0;
  PyArrayObject *__pyx_v_indices = 0;
  __pyx_t_5numpy_int64_t __pyx_v_leaf_size;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_left,&__pyx_n_s_right,&__pyx_n_s_indices,&__pyx_n_s_leaf_size,0};
    PyObject* values[4] = {0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_left)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_right)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 4, 4, 1); __PYX_ERR(1, 1268, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_indices)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 4, 4, 2); __PYX_ERR(1, 1268, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_leaf_size)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 4, 4, 3); __PYX_ERR(1, 1268, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(1, 1268, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 4) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
    }
    __pyx_v_left = ((PyArrayObject *)values[0]);
    __pyx_v_right = ((PyArrayObject *)values[1]);
    __pyx_v_indices = ((PyArrayObject *)values[2]);
    __pyx_v_leaf_size = __Pyx_PyInt_As_npy_int64(values[3]); if (unlikely((__pyx_v_leaf_size == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(1, 1272, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 4, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 1268, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Float64ClosedBothIntervalNode.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_left), __pyx_ptype_5numpy_ndarray, 1, "left", 0))) __PYX_ERR(1, 1269, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_right), __pyx_ptype_5numpy_ndarray, 1, "right", 0))) __PYX_ERR(1, 1270, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_indices), __pyx_ptype_5numpy_ndarray, 1, "indices", 0))) __PYX_ERR(1, 1271, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode___init__(((struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode *)__pyx_v_self), __pyx_v_left, __pyx_v_right, __pyx_v_indices, __pyx_v_leaf_size);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode___init__(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode *__pyx_v_self, PyArrayObject *__pyx_v_left, PyArrayObject *__pyx_v_right, PyArrayObject *__pyx_v_indices, __pyx_t_5numpy_int64_t __pyx_v_leaf_size) {
  PyObject *__pyx_v_left_set = NULL;
  PyObject *__pyx_v_right_set = NULL;
  PyObject *__pyx_v_center_set = NULL;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_indices;
  __Pyx_Buffer __pyx_pybuffer_indices;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_left;
  __Pyx_Buffer __pyx_pybuffer_left;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_right;
  __Pyx_Buffer __pyx_pybuffer_right;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __pyx_t_5numpy_float64_t __pyx_t_6;
  __Pyx_memviewslice __pyx_t_7 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_8 = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  __Pyx_memviewslice __pyx_t_11 = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *(*__pyx_t_12)(PyObject *);
  size_t __pyx_t_13;
  __Pyx_RefNannySetupContext("__init__", 0);
  __pyx_pybuffer_left.pybuffer.buf = NULL;
  __pyx_pybuffer_left.refcount = 0;
  __pyx_pybuffernd_left.data = NULL;
  __pyx_pybuffernd_left.rcbuffer = &__pyx_pybuffer_left;
  __pyx_pybuffer_right.pybuffer.buf = NULL;
  __pyx_pybuffer_right.refcount = 0;
  __pyx_pybuffernd_right.data = NULL;
  __pyx_pybuffernd_right.rcbuffer = &__pyx_pybuffer_right;
  __pyx_pybuffer_indices.pybuffer.buf = NULL;
  __pyx_pybuffer_indices.refcount = 0;
  __pyx_pybuffernd_indices.data = NULL;
  __pyx_pybuffernd_indices.rcbuffer = &__pyx_pybuffer_indices;
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_left.rcbuffer->pybuffer, (PyObject*)__pyx_v_left, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 1268, __pyx_L1_error)
  }
  __pyx_pybuffernd_left.diminfo[0].strides = __pyx_pybuffernd_left.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_left.diminfo[0].shape = __pyx_pybuffernd_left.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_right.rcbuffer->pybuffer, (PyObject*)__pyx_v_right, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 1268, __pyx_L1_error)
  }
  __pyx_pybuffernd_right.diminfo[0].strides = __pyx_pybuffernd_right.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_right.diminfo[0].shape = __pyx_pybuffernd_right.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_indices.rcbuffer->pybuffer, (PyObject*)__pyx_v_indices, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 1268, __pyx_L1_error)
  }
  __pyx_pybuffernd_indices.diminfo[0].strides = __pyx_pybuffernd_indices.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_indices.diminfo[0].shape = __pyx_pybuffernd_indices.rcbuffer->pybuffer.shape[0];

  /* "pandas/_libs/intervaltree.pxi":1274
 *                  int64_t leaf_size):
 * 
 *         self.n_elements = len(left)             # <<<<<<<<<<<<<<
 *         self.leaf_size = leaf_size
 * 
 */
  __pyx_t_1 = PyObject_Length(((PyObject *)__pyx_v_left)); if (unlikely(__pyx_t_1 == ((Py_ssize_t)-1))) __PYX_ERR(1, 1274, __pyx_L1_error)
  __pyx_v_self->n_elements = __pyx_t_1;

  /* "pandas/_libs/intervaltree.pxi":1275
 * 
 *         self.n_elements = len(left)
 *         self.leaf_size = leaf_size             # <<<<<<<<<<<<<<
 * 
 *         # min_left and min_right are used to speed-up query by skipping
 */
  __pyx_v_self->leaf_size = __pyx_v_leaf_size;

  /* "pandas/_libs/intervaltree.pxi":1280
 *         # query on sub-nodes. If this node has size 0, query is cheap,
 *         # so these values don't matter.
 *         if left.size > 0:             # <<<<<<<<<<<<<<
 *             self.min_left = left.min()
 *             self.max_right = right.max()
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_left), __pyx_n_s_size); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1280, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyObject_RichCompare(__pyx_t_2, __pyx_int_0, Py_GT); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1280, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(1, 1280, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_4) {

    /* "pandas/_libs/intervaltree.pxi":1281
 *         # so these values don't matter.
 *         if left.size > 0:
 *             self.min_left = left.min()             # <<<<<<<<<<<<<<
 *             self.max_right = right.max()
 *         else:
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_left), __pyx_n_s_min); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1281, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1281, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_6 = __pyx_PyFloat_AsDouble(__pyx_t_3); if (unlikely((__pyx_t_6 == ((npy_float64)-1)) && PyErr_Occurred())) __PYX_ERR(1, 1281, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_self->min_left = __pyx_t_6;

    /* "pandas/_libs/intervaltree.pxi":1282
 *         if left.size > 0:
 *             self.min_left = left.min()
 *             self.max_right = right.max()             # <<<<<<<<<<<<<<
 *         else:
 *             self.min_left = 0
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_right), __pyx_n_s_max); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1282, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1282, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_6 = __pyx_PyFloat_AsDouble(__pyx_t_3); if (unlikely((__pyx_t_6 == ((npy_float64)-1)) && PyErr_Occurred())) __PYX_ERR(1, 1282, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_self->max_right = __pyx_t_6;

    /* "pandas/_libs/intervaltree.pxi":1280
 *         # query on sub-nodes. If this node has size 0, query is cheap,
 *         # so these values don't matter.
 *         if left.size > 0:             # <<<<<<<<<<<<<<
 *             self.min_left = left.min()
 *             self.max_right = right.max()
 */
    goto __pyx_L3;
  }

  /* "pandas/_libs/intervaltree.pxi":1284
 *             self.max_right = right.max()
 *         else:
 *             self.min_left = 0             # <<<<<<<<<<<<<<
 *             self.max_right = 0
 * 
 */
  /*else*/ {
    __pyx_v_self->min_left = 0.0;

    /* "pandas/_libs/intervaltree.pxi":1285
 *         else:
 *             self.min_left = 0
 *             self.max_right = 0             # <<<<<<<<<<<<<<
 * 
 *         if self.n_elements <= leaf_size:
 */
    __pyx_v_self->max_right = 0.0;
  }
  __pyx_L3:;

  /* "pandas/_libs/intervaltree.pxi":1287
 *             self.max_right = 0
 * 
 *         if self.n_elements <= leaf_size:             # <<<<<<<<<<<<<<
 *             # make this a terminal (leaf) node
 *             self.is_leaf_node = True
 */
  __pyx_t_4 = ((__pyx_v_self->n_elements <= __pyx_v_leaf_size) != 0);
  if (__pyx_t_4) {

    /* "pandas/_libs/intervaltree.pxi":1289
 *         if self.n_elements <= leaf_size:
 *             # make this a terminal (leaf) node
 *             self.is_leaf_node = True             # <<<<<<<<<<<<<<
 *             self.left = left
 *             self.right = right
 */
    __pyx_v_self->is_leaf_node = 1;

    /* "pandas/_libs/intervaltree.pxi":1290
 *             # make this a terminal (leaf) node
 *             self.is_leaf_node = True
 *             self.left = left             # <<<<<<<<<<<<<<
 *             self.right = right
 *             self.indices = indices
 */
    __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_float64_t(((PyObject *)__pyx_v_left), PyBUF_WRITABLE); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(1, 1290, __pyx_L1_error)
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->left, 0);
    __pyx_v_self->left = __pyx_t_7;
    __pyx_t_7.memview = NULL;
    __pyx_t_7.data = NULL;

    /* "pandas/_libs/intervaltree.pxi":1291
 *             self.is_leaf_node = True
 *             self.left = left
 *             self.right = right             # <<<<<<<<<<<<<<
 *             self.indices = indices
 *             self.n_center = 0
 */
    __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_float64_t(((PyObject *)__pyx_v_right), PyBUF_WRITABLE); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(1, 1291, __pyx_L1_error)
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->right, 0);
    __pyx_v_self->right = __pyx_t_7;
    __pyx_t_7.memview = NULL;
    __pyx_t_7.data = NULL;

    /* "pandas/_libs/intervaltree.pxi":1292
 *             self.left = left
 *             self.right = right
 *             self.indices = indices             # <<<<<<<<<<<<<<
 *             self.n_center = 0
 *         else:
 */
    __pyx_t_8 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(((PyObject *)__pyx_v_indices), PyBUF_WRITABLE); if (unlikely(!__pyx_t_8.memview)) __PYX_ERR(1, 1292, __pyx_L1_error)
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->indices, 0);
    __pyx_v_self->indices = __pyx_t_8;
    __pyx_t_8.memview = NULL;
    __pyx_t_8.data = NULL;

    /* "pandas/_libs/intervaltree.pxi":1293
 *             self.right = right
 *             self.indices = indices
 *             self.n_center = 0             # <<<<<<<<<<<<<<
 *         else:
 *             # calculate a pivot so we can create child nodes
 */
    __pyx_v_self->n_center = 0;

    /* "pandas/_libs/intervaltree.pxi":1287
 *             self.max_right = 0
 * 
 *         if self.n_elements <= leaf_size:             # <<<<<<<<<<<<<<
 *             # make this a terminal (leaf) node
 *             self.is_leaf_node = True
 */
    goto __pyx_L4;
  }

  /* "pandas/_libs/intervaltree.pxi":1296
 *         else:
 *             # calculate a pivot so we can create child nodes
 *             self.is_leaf_node = False             # <<<<<<<<<<<<<<
 *             self.pivot = np.median(left / 2 + right / 2)
 *             left_set, right_set, center_set = self.classify_intervals(
 */
  /*else*/ {
    __pyx_v_self->is_leaf_node = 0;

    /* "pandas/_libs/intervaltree.pxi":1297
 *             # calculate a pivot so we can create child nodes
 *             self.is_leaf_node = False
 *             self.pivot = np.median(left / 2 + right / 2)             # <<<<<<<<<<<<<<
 *             left_set, right_set, center_set = self.classify_intervals(
 *                 left, right)
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1297, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_median); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1297, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyNumber_Divide(((PyObject *)__pyx_v_left), __pyx_int_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1297, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_9 = __Pyx_PyNumber_Divide(((PyObject *)__pyx_v_right), __pyx_int_2); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 1297, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_10 = PyNumber_Add(__pyx_t_2, __pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 1297, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_9 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_3 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_9, __pyx_t_10) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_10);
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1297, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_6 = __pyx_PyFloat_AsDouble(__pyx_t_3); if (unlikely((__pyx_t_6 == ((npy_float64)-1)) && PyErr_Occurred())) __PYX_ERR(1, 1297, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_self->pivot = __pyx_t_6;

    /* "pandas/_libs/intervaltree.pxi":1299
 *             self.pivot = np.median(left / 2 + right / 2)
 *             left_set, right_set, center_set = self.classify_intervals(
 *                 left, right)             # <<<<<<<<<<<<<<
 * 
 *             self.left_node = self.new_child_node(left, right,
 */
    __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_float64_t(((PyObject *)__pyx_v_left), PyBUF_WRITABLE); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(1, 1299, __pyx_L1_error)
    __pyx_t_11 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_float64_t(((PyObject *)__pyx_v_right), PyBUF_WRITABLE); if (unlikely(!__pyx_t_11.memview)) __PYX_ERR(1, 1299, __pyx_L1_error)

    /* "pandas/_libs/intervaltree.pxi":1298
 *             self.is_leaf_node = False
 *             self.pivot = np.median(left / 2 + right / 2)
 *             left_set, right_set, center_set = self.classify_intervals(             # <<<<<<<<<<<<<<
 *                 left, right)
 * 
 */
    __pyx_t_3 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode *)__pyx_v_self->__pyx_vtab)->classify_intervals(__pyx_v_self, __pyx_t_7, __pyx_t_11); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1298, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __PYX_XDEC_MEMVIEW(&__pyx_t_7, 1);
    __pyx_t_7.memview = NULL;
    __pyx_t_7.data = NULL;
    __PYX_XDEC_MEMVIEW(&__pyx_t_11, 1);
    __pyx_t_11.memview = NULL;
    __pyx_t_11.data = NULL;
    if ((likely(PyTuple_CheckExact(__pyx_t_3))) || (PyList_CheckExact(__pyx_t_3))) {
      PyObject* sequence = __pyx_t_3;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 3)) {
        if (size > 3) __Pyx_RaiseTooManyValuesError(3);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(1, 1298, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_5 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_10 = PyTuple_GET_ITEM(sequence, 1); 
        __pyx_t_9 = PyTuple_GET_ITEM(sequence, 2); 
      } else {
        __pyx_t_5 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_10 = PyList_GET_ITEM(sequence, 1); 
        __pyx_t_9 = PyList_GET_ITEM(sequence, 2); 
      }
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_10);
      __Pyx_INCREF(__pyx_t_9);
      #else
      __pyx_t_5 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1298, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_10 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 1298, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_9 = PySequence_ITEM(sequence, 2); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 1298, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      #endif
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_2 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1298, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_12 = Py_TYPE(__pyx_t_2)->tp_iternext;
      index = 0; __pyx_t_5 = __pyx_t_12(__pyx_t_2); if (unlikely(!__pyx_t_5)) goto __pyx_L5_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_5);
      index = 1; __pyx_t_10 = __pyx_t_12(__pyx_t_2); if (unlikely(!__pyx_t_10)) goto __pyx_L5_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_10);
      index = 2; __pyx_t_9 = __pyx_t_12(__pyx_t_2); if (unlikely(!__pyx_t_9)) goto __pyx_L5_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_9);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_12(__pyx_t_2), 3) < 0) __PYX_ERR(1, 1298, __pyx_L1_error)
      __pyx_t_12 = NULL;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      goto __pyx_L6_unpacking_done;
      __pyx_L5_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_12 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(1, 1298, __pyx_L1_error)
      __pyx_L6_unpacking_done:;
    }
    __pyx_v_left_set = __pyx_t_5;
    __pyx_t_5 = 0;
    __pyx_v_right_set = __pyx_t_10;
    __pyx_t_10 = 0;
    __pyx_v_center_set = __pyx_t_9;
    __pyx_t_9 = 0;

    /* "pandas/_libs/intervaltree.pxi":1302
 * 
 *             self.left_node = self.new_child_node(left, right,
 *                                                  indices, left_set)             # <<<<<<<<<<<<<<
 *             self.right_node = self.new_child_node(left, right,
 *                                                   indices, right_set)
 */
    if (!(likely(((__pyx_v_left_set) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_left_set, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(1, 1302, __pyx_L1_error)

    /* "pandas/_libs/intervaltree.pxi":1301
 *                 left, right)
 * 
 *             self.left_node = self.new_child_node(left, right,             # <<<<<<<<<<<<<<
 *                                                  indices, left_set)
 *             self.right_node = self.new_child_node(left, right,
 */
    __pyx_t_3 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode *)__pyx_v_self->__pyx_vtab)->new_child_node(__pyx_v_self, ((PyArrayObject *)__pyx_v_left), ((PyArrayObject *)__pyx_v_right), ((PyArrayObject *)__pyx_v_indices), ((PyArrayObject *)__pyx_v_left_set)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1301, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode))))) __PYX_ERR(1, 1301, __pyx_L1_error)
    __Pyx_GIVEREF(__pyx_t_3);
    __Pyx_GOTREF(__pyx_v_self->left_node);
    __Pyx_DECREF(((PyObject *)__pyx_v_self->left_node));
    __pyx_v_self->left_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode *)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "pandas/_libs/intervaltree.pxi":1304
 *                                                  indices, left_set)
 *             self.right_node = self.new_child_node(left, right,
 *                                                   indices, right_set)             # <<<<<<<<<<<<<<
 * 
 *             self.center_left_values, self.center_left_indices = \
 */
    if (!(likely(((__pyx_v_right_set) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_right_set, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(1, 1304, __pyx_L1_error)

    /* "pandas/_libs/intervaltree.pxi":1303
 *             self.left_node = self.new_child_node(left, right,
 *                                                  indices, left_set)
 *             self.right_node = self.new_child_node(left, right,             # <<<<<<<<<<<<<<
 *                                                   indices, right_set)
 * 
 */
    __pyx_t_3 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode *)__pyx_v_self->__pyx_vtab)->new_child_node(__pyx_v_self, ((PyArrayObject *)__pyx_v_left), ((PyArrayObject *)__pyx_v_right), ((PyArrayObject *)__pyx_v_indices), ((PyArrayObject *)__pyx_v_right_set)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1303, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode))))) __PYX_ERR(1, 1303, __pyx_L1_error)
    __Pyx_GIVEREF(__pyx_t_3);
    __Pyx_GOTREF(__pyx_v_self->right_node);
    __Pyx_DECREF(((PyObject *)__pyx_v_self->right_node));
    __pyx_v_self->right_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode *)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "pandas/_libs/intervaltree.pxi":1307
 * 
 *             self.center_left_values, self.center_left_indices = \
 *                 sort_values_and_indices(left, indices, center_set)             # <<<<<<<<<<<<<<
 *             self.center_right_values, self.center_right_indices = \
 *                 sort_values_and_indices(right, indices, center_set)
 */
    __pyx_t_3 = __pyx_f_6pandas_5_libs_8interval_sort_values_and_indices(((PyObject *)__pyx_v_left), ((PyObject *)__pyx_v_indices), __pyx_v_center_set); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1307, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if ((likely(PyTuple_CheckExact(__pyx_t_3))) || (PyList_CheckExact(__pyx_t_3))) {
      PyObject* sequence = __pyx_t_3;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(1, 1306, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_9 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_10 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_9 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_10 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_9);
      __Pyx_INCREF(__pyx_t_10);
      #else
      __pyx_t_9 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 1306, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_10 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 1306, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      #endif
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_5 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1306, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_12 = Py_TYPE(__pyx_t_5)->tp_iternext;
      index = 0; __pyx_t_9 = __pyx_t_12(__pyx_t_5); if (unlikely(!__pyx_t_9)) goto __pyx_L7_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_9);
      index = 1; __pyx_t_10 = __pyx_t_12(__pyx_t_5); if (unlikely(!__pyx_t_10)) goto __pyx_L7_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_10);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_12(__pyx_t_5), 2) < 0) __PYX_ERR(1, 1306, __pyx_L1_error)
      __pyx_t_12 = NULL;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      goto __pyx_L8_unpacking_done;
      __pyx_L7_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_12 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(1, 1306, __pyx_L1_error)
      __pyx_L8_unpacking_done:;
    }

    /* "pandas/_libs/intervaltree.pxi":1306
 *                                                   indices, right_set)
 * 
 *             self.center_left_values, self.center_left_indices = \             # <<<<<<<<<<<<<<
 *                 sort_values_and_indices(left, indices, center_set)
 *             self.center_right_values, self.center_right_indices = \
 */
    __pyx_t_11 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_float64_t(__pyx_t_9, PyBUF_WRITABLE); if (unlikely(!__pyx_t_11.memview)) __PYX_ERR(1, 1306, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_8 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_t_10, PyBUF_WRITABLE); if (unlikely(!__pyx_t_8.memview)) __PYX_ERR(1, 1306, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->center_left_values, 0);
    __pyx_v_self->center_left_values = __pyx_t_11;
    __pyx_t_11.memview = NULL;
    __pyx_t_11.data = NULL;
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->center_left_indices, 0);
    __pyx_v_self->center_left_indices = __pyx_t_8;
    __pyx_t_8.memview = NULL;
    __pyx_t_8.data = NULL;

    /* "pandas/_libs/intervaltree.pxi":1309
 *                 sort_values_and_indices(left, indices, center_set)
 *             self.center_right_values, self.center_right_indices = \
 *                 sort_values_and_indices(right, indices, center_set)             # <<<<<<<<<<<<<<
 *             self.n_center = len(self.center_left_indices)
 * 
 */
    __pyx_t_3 = __pyx_f_6pandas_5_libs_8interval_sort_values_and_indices(((PyObject *)__pyx_v_right), ((PyObject *)__pyx_v_indices), __pyx_v_center_set); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1309, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if ((likely(PyTuple_CheckExact(__pyx_t_3))) || (PyList_CheckExact(__pyx_t_3))) {
      PyObject* sequence = __pyx_t_3;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(1, 1308, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_10 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_9 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_10 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_9 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_10);
      __Pyx_INCREF(__pyx_t_9);
      #else
      __pyx_t_10 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 1308, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_9 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 1308, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      #endif
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_5 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1308, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_12 = Py_TYPE(__pyx_t_5)->tp_iternext;
      index = 0; __pyx_t_10 = __pyx_t_12(__pyx_t_5); if (unlikely(!__pyx_t_10)) goto __pyx_L9_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_10);
      index = 1; __pyx_t_9 = __pyx_t_12(__pyx_t_5); if (unlikely(!__pyx_t_9)) goto __pyx_L9_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_9);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_12(__pyx_t_5), 2) < 0) __PYX_ERR(1, 1308, __pyx_L1_error)
      __pyx_t_12 = NULL;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      goto __pyx_L10_unpacking_done;
      __pyx_L9_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_12 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(1, 1308, __pyx_L1_error)
      __pyx_L10_unpacking_done:;
    }

    /* "pandas/_libs/intervaltree.pxi":1308
 *             self.center_left_values, self.center_left_indices = \
 *                 sort_values_and_indices(left, indices, center_set)
 *             self.center_right_values, self.center_right_indices = \             # <<<<<<<<<<<<<<
 *                 sort_values_and_indices(right, indices, center_set)
 *             self.n_center = len(self.center_left_indices)
 */
    __pyx_t_11 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_float64_t(__pyx_t_10, PyBUF_WRITABLE); if (unlikely(!__pyx_t_11.memview)) __PYX_ERR(1, 1308, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_t_8 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_t_9, PyBUF_WRITABLE); if (unlikely(!__pyx_t_8.memview)) __PYX_ERR(1, 1308, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->center_right_values, 0);
    __pyx_v_self->center_right_values = __pyx_t_11;
    __pyx_t_11.memview = NULL;
    __pyx_t_11.data = NULL;
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->center_right_indices, 0);
    __pyx_v_self->center_right_indices = __pyx_t_8;
    __pyx_t_8.memview = NULL;
    __pyx_t_8.data = NULL;

    /* "pandas/_libs/intervaltree.pxi":1310
 *             self.center_right_values, self.center_right_indices = \
 *                 sort_values_and_indices(right, indices, center_set)
 *             self.n_center = len(self.center_left_indices)             # <<<<<<<<<<<<<<
 * 
 *     @cython.wraparound(False)
 */
    if (unlikely(!__pyx_v_self->center_left_indices.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(1, 1310, __pyx_L1_error)}
    __pyx_t_13 = __Pyx_MemoryView_Len(__pyx_v_self->center_left_indices); 
    __pyx_v_self->n_center = __pyx_t_13;
  }
  __pyx_L4:;

  /* "pandas/_libs/intervaltree.pxi":1268
 *         readonly bint is_leaf_node
 * 
 *     def __init__(self,             # <<<<<<<<<<<<<<
 *                  ndarray[float64_t, ndim=1] left,
 *                  ndarray[float64_t, ndim=1] right,
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __PYX_XDEC_MEMVIEW(&__pyx_t_7, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_8, 1);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __PYX_XDEC_MEMVIEW(&__pyx_t_11, 1);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indices.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_left.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_right.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("pandas._libs.interval.Float64ClosedBothIntervalNode.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indices.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_left.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_right.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF(__pyx_v_left_set);
  __Pyx_XDECREF(__pyx_v_right_set);
  __Pyx_XDECREF(__pyx_v_center_set);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":1314
 *     @cython.wraparound(False)
 *     @cython.boundscheck(False)
 *     cdef classify_intervals(self, float64_t[:] left, float64_t[:] right):             # <<<<<<<<<<<<<<
 *         """Classify the given intervals based upon whether they fall to the
 *         left, right, or overlap with this node's pivot.
 */

static PyObject *__pyx_f_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_classify_intervals(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode *__pyx_v_self, __Pyx_memviewslice __pyx_v_left, __Pyx_memviewslice __pyx_v_right) {
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_left_ind = 0;
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_right_ind = 0;
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_overlapping_ind = 0;
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __pyx_t_5numpy_int64_t __pyx_t_2;
  __pyx_t_5numpy_int64_t __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  int __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  __Pyx_RefNannySetupContext("classify_intervals", 0);

  /* "pandas/_libs/intervaltree.pxi":1322
 *             Py_ssize_t i
 * 
 *         left_ind = Int64Vector()             # <<<<<<<<<<<<<<
 *         right_ind = Int64Vector()
 *         overlapping_ind = Int64Vector()
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1322, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_left_ind = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":1323
 * 
 *         left_ind = Int64Vector()
 *         right_ind = Int64Vector()             # <<<<<<<<<<<<<<
 *         overlapping_ind = Int64Vector()
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1323, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_right_ind = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":1324
 *         left_ind = Int64Vector()
 *         right_ind = Int64Vector()
 *         overlapping_ind = Int64Vector()             # <<<<<<<<<<<<<<
 * 
 *         for i in range(self.n_elements):
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1324, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_overlapping_ind = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":1326
 *         overlapping_ind = Int64Vector()
 * 
 *         for i in range(self.n_elements):             # <<<<<<<<<<<<<<
 *             if right[i] < self.pivot:
 *                 left_ind.append(i)
 */
  __pyx_t_2 = __pyx_v_self->n_elements;
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "pandas/_libs/intervaltree.pxi":1327
 * 
 *         for i in range(self.n_elements):
 *             if right[i] < self.pivot:             # <<<<<<<<<<<<<<
 *                 left_ind.append(i)
 *             elif self.pivot < left[i]:
 */
    __pyx_t_5 = __pyx_v_i;
    __pyx_t_6 = (((*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_right.data + __pyx_t_5 * __pyx_v_right.strides[0]) ))) < __pyx_v_self->pivot) != 0);
    if (__pyx_t_6) {

      /* "pandas/_libs/intervaltree.pxi":1328
 *         for i in range(self.n_elements):
 *             if right[i] < self.pivot:
 *                 left_ind.append(i)             # <<<<<<<<<<<<<<
 *             elif self.pivot < left[i]:
 *                 right_ind.append(i)
 */
      ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_left_ind->__pyx_vtab)->append(__pyx_v_left_ind, __pyx_v_i);

      /* "pandas/_libs/intervaltree.pxi":1327
 * 
 *         for i in range(self.n_elements):
 *             if right[i] < self.pivot:             # <<<<<<<<<<<<<<
 *                 left_ind.append(i)
 *             elif self.pivot < left[i]:
 */
      goto __pyx_L5;
    }

    /* "pandas/_libs/intervaltree.pxi":1329
 *             if right[i] < self.pivot:
 *                 left_ind.append(i)
 *             elif self.pivot < left[i]:             # <<<<<<<<<<<<<<
 *                 right_ind.append(i)
 *             else:
 */
    __pyx_t_7 = __pyx_v_i;
    __pyx_t_6 = ((__pyx_v_self->pivot < (*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_left.data + __pyx_t_7 * __pyx_v_left.strides[0]) )))) != 0);
    if (__pyx_t_6) {

      /* "pandas/_libs/intervaltree.pxi":1330
 *                 left_ind.append(i)
 *             elif self.pivot < left[i]:
 *                 right_ind.append(i)             # <<<<<<<<<<<<<<
 *             else:
 *                 overlapping_ind.append(i)
 */
      ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_right_ind->__pyx_vtab)->append(__pyx_v_right_ind, __pyx_v_i);

      /* "pandas/_libs/intervaltree.pxi":1329
 *             if right[i] < self.pivot:
 *                 left_ind.append(i)
 *             elif self.pivot < left[i]:             # <<<<<<<<<<<<<<
 *                 right_ind.append(i)
 *             else:
 */
      goto __pyx_L5;
    }

    /* "pandas/_libs/intervaltree.pxi":1332
 *                 right_ind.append(i)
 *             else:
 *                 overlapping_ind.append(i)             # <<<<<<<<<<<<<<
 * 
 *         return (left_ind.to_array(),
 */
    /*else*/ {
      ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_overlapping_ind->__pyx_vtab)->append(__pyx_v_overlapping_ind, __pyx_v_i);
    }
    __pyx_L5:;
  }

  /* "pandas/_libs/intervaltree.pxi":1334
 *                 overlapping_ind.append(i)
 * 
 *         return (left_ind.to_array(),             # <<<<<<<<<<<<<<
 *                 right_ind.to_array(),
 *                 overlapping_ind.to_array())
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_left_ind->__pyx_vtab)->to_array(__pyx_v_left_ind, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1334, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "pandas/_libs/intervaltree.pxi":1335
 * 
 *         return (left_ind.to_array(),
 *                 right_ind.to_array(),             # <<<<<<<<<<<<<<
 *                 overlapping_ind.to_array())
 * 
 */
  __pyx_t_8 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_right_ind->__pyx_vtab)->to_array(__pyx_v_right_ind, 0); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 1335, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);

  /* "pandas/_libs/intervaltree.pxi":1336
 *         return (left_ind.to_array(),
 *                 right_ind.to_array(),
 *                 overlapping_ind.to_array())             # <<<<<<<<<<<<<<
 * 
 *     cdef new_child_node(self,
 */
  __pyx_t_9 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_overlapping_ind->__pyx_vtab)->to_array(__pyx_v_overlapping_ind, 0); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 1336, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);

  /* "pandas/_libs/intervaltree.pxi":1334
 *                 overlapping_ind.append(i)
 * 
 *         return (left_ind.to_array(),             # <<<<<<<<<<<<<<
 *                 right_ind.to_array(),
 *                 overlapping_ind.to_array())
 */
  __pyx_t_10 = PyTuple_New(3); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 1334, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_t_8);
  __Pyx_GIVEREF(__pyx_t_9);
  PyTuple_SET_ITEM(__pyx_t_10, 2, __pyx_t_9);
  __pyx_t_1 = 0;
  __pyx_t_8 = 0;
  __pyx_t_9 = 0;
  __pyx_r = __pyx_t_10;
  __pyx_t_10 = 0;
  goto __pyx_L0;

  /* "pandas/_libs/intervaltree.pxi":1314
 *     @cython.wraparound(False)
 *     @cython.boundscheck(False)
 *     cdef classify_intervals(self, float64_t[:] left, float64_t[:] right):             # <<<<<<<<<<<<<<
 *         """Classify the given intervals based upon whether they fall to the
 *         left, right, or overlap with this node's pivot.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("pandas._libs.interval.Float64ClosedBothIntervalNode.classify_intervals", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_left_ind);
  __Pyx_XDECREF((PyObject *)__pyx_v_right_ind);
  __Pyx_XDECREF((PyObject *)__pyx_v_overlapping_ind);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":1338
 *                 overlapping_ind.to_array())
 * 
 *     cdef new_child_node(self,             # <<<<<<<<<<<<<<
 *                         ndarray[float64_t, ndim=1] left,
 *                         ndarray[float64_t, ndim=1] right,
 */

static PyObject *__pyx_f_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_new_child_node(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode *__pyx_v_self, PyArrayObject *__pyx_v_left, PyArrayObject *__pyx_v_right, PyArrayObject *__pyx_v_indices, PyArrayObject *__pyx_v_subset) {
  __Pyx_LocalBuf_ND __pyx_pybuffernd_indices;
  __Pyx_Buffer __pyx_pybuffer_indices;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_left;
  __Pyx_Buffer __pyx_pybuffer_left;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_right;
  __Pyx_Buffer __pyx_pybuffer_right;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_subset;
  __Pyx_Buffer __pyx_pybuffer_subset;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyArrayObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyArrayObject *__pyx_t_7 = NULL;
  PyArrayObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  __Pyx_RefNannySetupContext("new_child_node", 0);
  __Pyx_INCREF((PyObject *)__pyx_v_left);
  __Pyx_INCREF((PyObject *)__pyx_v_right);
  __Pyx_INCREF((PyObject *)__pyx_v_indices);
  __pyx_pybuffer_left.pybuffer.buf = NULL;
  __pyx_pybuffer_left.refcount = 0;
  __pyx_pybuffernd_left.data = NULL;
  __pyx_pybuffernd_left.rcbuffer = &__pyx_pybuffer_left;
  __pyx_pybuffer_right.pybuffer.buf = NULL;
  __pyx_pybuffer_right.refcount = 0;
  __pyx_pybuffernd_right.data = NULL;
  __pyx_pybuffernd_right.rcbuffer = &__pyx_pybuffer_right;
  __pyx_pybuffer_indices.pybuffer.buf = NULL;
  __pyx_pybuffer_indices.refcount = 0;
  __pyx_pybuffernd_indices.data = NULL;
  __pyx_pybuffernd_indices.rcbuffer = &__pyx_pybuffer_indices;
  __pyx_pybuffer_subset.pybuffer.buf = NULL;
  __pyx_pybuffer_subset.refcount = 0;
  __pyx_pybuffernd_subset.data = NULL;
  __pyx_pybuffernd_subset.rcbuffer = &__pyx_pybuffer_subset;
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_left.rcbuffer->pybuffer, (PyObject*)__pyx_v_left, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 1338, __pyx_L1_error)
  }
  __pyx_pybuffernd_left.diminfo[0].strides = __pyx_pybuffernd_left.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_left.diminfo[0].shape = __pyx_pybuffernd_left.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_right.rcbuffer->pybuffer, (PyObject*)__pyx_v_right, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 1338, __pyx_L1_error)
  }
  __pyx_pybuffernd_right.diminfo[0].strides = __pyx_pybuffernd_right.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_right.diminfo[0].shape = __pyx_pybuffernd_right.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_indices.rcbuffer->pybuffer, (PyObject*)__pyx_v_indices, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 1338, __pyx_L1_error)
  }
  __pyx_pybuffernd_indices.diminfo[0].strides = __pyx_pybuffernd_indices.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_indices.diminfo[0].shape = __pyx_pybuffernd_indices.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_subset.rcbuffer->pybuffer, (PyObject*)__pyx_v_subset, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 1338, __pyx_L1_error)
  }
  __pyx_pybuffernd_subset.diminfo[0].strides = __pyx_pybuffernd_subset.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_subset.diminfo[0].shape = __pyx_pybuffernd_subset.rcbuffer->pybuffer.shape[0];

  /* "pandas/_libs/intervaltree.pxi":1345
 *         """Create a new child node.
 *         """
 *         left = take(left, subset)             # <<<<<<<<<<<<<<
 *         right = take(right, subset)
 *         indices = take(indices, subset)
 */
  __pyx_t_1 = __pyx_f_6pandas_5_libs_8interval_take(((PyArrayObject *)__pyx_v_left), ((PyArrayObject *)__pyx_v_subset)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1345, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(1, 1345, __pyx_L1_error)
  __pyx_t_2 = ((PyArrayObject *)__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_left.rcbuffer->pybuffer);
    __pyx_t_3 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_left.rcbuffer->pybuffer, (PyObject*)__pyx_t_2, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack);
    if (unlikely(__pyx_t_3 < 0)) {
      PyErr_Fetch(&__pyx_t_4, &__pyx_t_5, &__pyx_t_6);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_left.rcbuffer->pybuffer, (PyObject*)__pyx_v_left, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_4); Py_XDECREF(__pyx_t_5); Py_XDECREF(__pyx_t_6);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_4, __pyx_t_5, __pyx_t_6);
      }
      __pyx_t_4 = __pyx_t_5 = __pyx_t_6 = 0;
    }
    __pyx_pybuffernd_left.diminfo[0].strides = __pyx_pybuffernd_left.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_left.diminfo[0].shape = __pyx_pybuffernd_left.rcbuffer->pybuffer.shape[0];
    if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(1, 1345, __pyx_L1_error)
  }
  __pyx_t_2 = 0;
  __Pyx_DECREF_SET(__pyx_v_left, ((PyArrayObject *)__pyx_t_1));
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":1346
 *         """
 *         left = take(left, subset)
 *         right = take(right, subset)             # <<<<<<<<<<<<<<
 *         indices = take(indices, subset)
 *         return Float64ClosedBothIntervalNode(
 */
  __pyx_t_1 = __pyx_f_6pandas_5_libs_8interval_take(((PyArrayObject *)__pyx_v_right), ((PyArrayObject *)__pyx_v_subset)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1346, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(1, 1346, __pyx_L1_error)
  __pyx_t_7 = ((PyArrayObject *)__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_right.rcbuffer->pybuffer);
    __pyx_t_3 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_right.rcbuffer->pybuffer, (PyObject*)__pyx_t_7, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack);
    if (unlikely(__pyx_t_3 < 0)) {
      PyErr_Fetch(&__pyx_t_6, &__pyx_t_5, &__pyx_t_4);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_right.rcbuffer->pybuffer, (PyObject*)__pyx_v_right, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_6); Py_XDECREF(__pyx_t_5); Py_XDECREF(__pyx_t_4);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_6, __pyx_t_5, __pyx_t_4);
      }
      __pyx_t_6 = __pyx_t_5 = __pyx_t_4 = 0;
    }
    __pyx_pybuffernd_right.diminfo[0].strides = __pyx_pybuffernd_right.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_right.diminfo[0].shape = __pyx_pybuffernd_right.rcbuffer->pybuffer.shape[0];
    if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(1, 1346, __pyx_L1_error)
  }
  __pyx_t_7 = 0;
  __Pyx_DECREF_SET(__pyx_v_right, ((PyArrayObject *)__pyx_t_1));
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":1347
 *         left = take(left, subset)
 *         right = take(right, subset)
 *         indices = take(indices, subset)             # <<<<<<<<<<<<<<
 *         return Float64ClosedBothIntervalNode(
 *             left, right, indices, self.leaf_size)
 */
  __pyx_t_1 = __pyx_f_6pandas_5_libs_8interval_take(((PyArrayObject *)__pyx_v_indices), ((PyArrayObject *)__pyx_v_subset)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1347, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(1, 1347, __pyx_L1_error)
  __pyx_t_8 = ((PyArrayObject *)__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indices.rcbuffer->pybuffer);
    __pyx_t_3 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_indices.rcbuffer->pybuffer, (PyObject*)__pyx_t_8, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack);
    if (unlikely(__pyx_t_3 < 0)) {
      PyErr_Fetch(&__pyx_t_4, &__pyx_t_5, &__pyx_t_6);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_indices.rcbuffer->pybuffer, (PyObject*)__pyx_v_indices, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_4); Py_XDECREF(__pyx_t_5); Py_XDECREF(__pyx_t_6);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_4, __pyx_t_5, __pyx_t_6);
      }
      __pyx_t_4 = __pyx_t_5 = __pyx_t_6 = 0;
    }
    __pyx_pybuffernd_indices.diminfo[0].strides = __pyx_pybuffernd_indices.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_indices.diminfo[0].shape = __pyx_pybuffernd_indices.rcbuffer->pybuffer.shape[0];
    if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(1, 1347, __pyx_L1_error)
  }
  __pyx_t_8 = 0;
  __Pyx_DECREF_SET(__pyx_v_indices, ((PyArrayObject *)__pyx_t_1));
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":1348
 *         right = take(right, subset)
 *         indices = take(indices, subset)
 *         return Float64ClosedBothIntervalNode(             # <<<<<<<<<<<<<<
 *             left, right, indices, self.leaf_size)
 * 
 */
  __Pyx_XDECREF(__pyx_r);

  /* "pandas/_libs/intervaltree.pxi":1349
 *         indices = take(indices, subset)
 *         return Float64ClosedBothIntervalNode(
 *             left, right, indices, self.leaf_size)             # <<<<<<<<<<<<<<
 * 
 *     @cython.wraparound(False)
 */
  __pyx_t_1 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->leaf_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1349, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "pandas/_libs/intervaltree.pxi":1348
 *         right = take(right, subset)
 *         indices = take(indices, subset)
 *         return Float64ClosedBothIntervalNode(             # <<<<<<<<<<<<<<
 *             left, right, indices, self.leaf_size)
 * 
 */
  __pyx_t_9 = PyTuple_New(4); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 1348, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_INCREF(((PyObject *)__pyx_v_left));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_left));
  PyTuple_SET_ITEM(__pyx_t_9, 0, ((PyObject *)__pyx_v_left));
  __Pyx_INCREF(((PyObject *)__pyx_v_right));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_right));
  PyTuple_SET_ITEM(__pyx_t_9, 1, ((PyObject *)__pyx_v_right));
  __Pyx_INCREF(((PyObject *)__pyx_v_indices));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_indices));
  PyTuple_SET_ITEM(__pyx_t_9, 2, ((PyObject *)__pyx_v_indices));
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_9, 3, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode), __pyx_t_9, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1348, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pandas/_libs/intervaltree.pxi":1338
 *                 overlapping_ind.to_array())
 * 
 *     cdef new_child_node(self,             # <<<<<<<<<<<<<<
 *                         ndarray[float64_t, ndim=1] left,
 *                         ndarray[float64_t, ndim=1] right,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_9);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indices.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_left.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_right.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_subset.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("pandas._libs.interval.Float64ClosedBothIntervalNode.new_child_node", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indices.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_left.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_right.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_subset.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF((PyObject *)__pyx_v_left);
  __Pyx_XDECREF((PyObject *)__pyx_v_right);
  __Pyx_XDECREF((PyObject *)__pyx_v_indices);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":1354
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_2query[] = "Recursively query this node and its sub-nodes for intervals that\n        overlap with the query point.\n        ";
static PyMethodDef __pyx_mdef_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_3query = {"query", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_3query, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_2query};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_args = 0;
  PyObject *__pyx_v_kwargs = 0;
  CYTHON_UNUSED PyObject *__pyx_v_defaults = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fused_cpdef (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_args,&__pyx_n_s_kwargs,&__pyx_n_s_defaults,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_args)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_kwargs)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fused_cpdef", 1, 3, 3, 1); __PYX_ERR(1, 1354, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_defaults)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fused_cpdef", 1, 3, 3, 2); __PYX_ERR(1, 1354, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fused_cpdef") < 0)) __PYX_ERR(1, 1354, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_args = values[0];
    __pyx_v_kwargs = values[1];
    __pyx_v_defaults = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fused_cpdef", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 1354, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Float64ClosedBothIntervalNode.__pyx_fused_cpdef", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_2query(((struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode *)__pyx_v_signatures), __pyx_v_args, __pyx_v_kwargs, __pyx_v_defaults);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_2query(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode *__pyx_v_signatures, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs, CYTHON_UNUSED PyObject *__pyx_v_defaults) {
  PyObject *__pyx_v_dest_sig = NULL;
  Py_ssize_t __pyx_v_i;
  CYTHON_UNUSED PyTypeObject *__pyx_v_ndarray = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_v_itemsize;
  PyObject *__pyx_v_arg = NULL;
  PyObject *__pyx_v_candidates = NULL;
  PyObject *__pyx_v_sig = NULL;
  int __pyx_v_match_found;
  PyObject *__pyx_v_src_sig = NULL;
  PyObject *__pyx_v_dst_type = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  Py_ssize_t __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  Py_ssize_t __pyx_t_13;
  Py_ssize_t __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  int __pyx_t_16;
  __Pyx_RefNannySetupContext("query", 0);
  __Pyx_INCREF(__pyx_v_kwargs);
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1354, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  PyList_SET_ITEM(__pyx_t_1, 0, Py_None);
  __pyx_v_dest_sig = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_3 = (__pyx_v_kwargs != Py_None);
  __pyx_t_4 = (__pyx_t_3 != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_kwargs); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(1, 1354, __pyx_L1_error)
  __pyx_t_3 = ((!__pyx_t_4) != 0);
  __pyx_t_2 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {
    __Pyx_INCREF(Py_None);
    __Pyx_DECREF_SET(__pyx_v_kwargs, Py_None);
  }
  __pyx_t_1 = ((PyObject *)__Pyx_ImportNumPyArrayTypeIfAvailable()); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1354, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_ndarray = ((PyTypeObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_v_itemsize = -1L;
  if (unlikely(__pyx_v_args == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 1354, __pyx_L1_error)
  }
  __pyx_t_5 = PyTuple_GET_SIZE(((PyObject*)__pyx_v_args)); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(1, 1354, __pyx_L1_error)
  __pyx_t_2 = ((2 < __pyx_t_5) != 0);
  if (__pyx_t_2) {
    if (unlikely(__pyx_v_args == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 1354, __pyx_L1_error)
    }
    __pyx_t_1 = PyTuple_GET_ITEM(((PyObject*)__pyx_v_args), 2);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v_arg = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L6;
  }
  __pyx_t_3 = (__pyx_v_kwargs != Py_None);
  __pyx_t_4 = (__pyx_t_3 != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L7_bool_binop_done;
  }
  if (unlikely(__pyx_v_kwargs == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(1, 1354, __pyx_L1_error)
  }
  __pyx_t_4 = (__Pyx_PyDict_ContainsTF(__pyx_n_s_point, ((PyObject*)__pyx_v_kwargs), Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(1, 1354, __pyx_L1_error)
  __pyx_t_3 = (__pyx_t_4 != 0);
  __pyx_t_2 = __pyx_t_3;
  __pyx_L7_bool_binop_done:;
  if (__pyx_t_2) {
    if (unlikely(__pyx_v_kwargs == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 1354, __pyx_L1_error)
    }
    __pyx_t_1 = __Pyx_PyDict_GetItem(((PyObject*)__pyx_v_kwargs), __pyx_n_s_point); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1354, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_arg = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L6;
  }
  /*else*/ {
    if (unlikely(__pyx_v_args == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
      __PYX_ERR(1, 1354, __pyx_L1_error)
    }
    __pyx_t_5 = PyTuple_GET_SIZE(((PyObject*)__pyx_v_args)); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(1, 1354, __pyx_L1_error)
    __pyx_t_1 = PyInt_FromSsize_t(__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1354, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_6 = PyTuple_New(3); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 1354, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_INCREF(__pyx_int_3);
    __Pyx_GIVEREF(__pyx_int_3);
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_int_3);
    __Pyx_INCREF(__pyx_n_s_s);
    __Pyx_GIVEREF(__pyx_n_s_s);
    PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_n_s_s);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Expected_at_least_d_argument_s_g, __pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1354, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_1); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 1354, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(1, 1354, __pyx_L1_error)
  }
  __pyx_L6:;
  while (1) {
    __pyx_t_2 = PyFloat_Check(__pyx_v_arg); 
    __pyx_t_3 = (__pyx_t_2 != 0);
    if (__pyx_t_3) {
      if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, __pyx_n_s_float64_t, long, 1, __Pyx_PyInt_From_long, 1, 0, 0) < 0)) __PYX_ERR(1, 1354, __pyx_L1_error)
      goto __pyx_L10_break;
    }
    __pyx_t_2 = PyInt_Check(__pyx_v_arg); 
    __pyx_t_4 = (__pyx_t_2 != 0);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_3 = __pyx_t_4;
      goto __pyx_L13_bool_binop_done;
    }
    __pyx_t_4 = PyLong_Check(__pyx_v_arg); 
    __pyx_t_2 = (__pyx_t_4 != 0);
    __pyx_t_3 = __pyx_t_2;
    __pyx_L13_bool_binop_done:;
    __pyx_t_2 = (__pyx_t_3 != 0);
    if (__pyx_t_2) {
      if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, __pyx_n_s_int64_t, long, 1, __Pyx_PyInt_From_long, 1, 0, 0) < 0)) __PYX_ERR(1, 1354, __pyx_L1_error)
      goto __pyx_L10_break;
    }
    if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, Py_None, long, 1, __Pyx_PyInt_From_long, 1, 0, 0) < 0)) __PYX_ERR(1, 1354, __pyx_L1_error)
    goto __pyx_L10_break;
  }
  __pyx_L10_break:;
  __pyx_t_6 = PyList_New(0); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 1354, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_v_candidates = ((PyObject*)__pyx_t_6);
  __pyx_t_6 = 0;
  __pyx_t_5 = 0;
  if (unlikely(((PyObject *)__pyx_v_signatures) == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(1, 1354, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_dict_iterator(((PyObject*)__pyx_v_signatures), 1, ((PyObject *)NULL), (&__pyx_t_7), (&__pyx_t_8)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1354, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __pyx_t_6 = __pyx_t_1;
  __pyx_t_1 = 0;
  while (1) {
    __pyx_t_9 = __Pyx_dict_iter_next(__pyx_t_6, __pyx_t_7, &__pyx_t_5, &__pyx_t_1, NULL, NULL, __pyx_t_8);
    if (unlikely(__pyx_t_9 == 0)) break;
    if (unlikely(__pyx_t_9 == -1)) __PYX_ERR(1, 1354, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_sig, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_v_match_found = 0;
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_v_sig, __pyx_n_s_strip); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 1354, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __pyx_t_12 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_11))) {
      __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_11);
      if (likely(__pyx_t_12)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
        __Pyx_INCREF(__pyx_t_12);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_11, function);
      }
    }
    __pyx_t_10 = (__pyx_t_12) ? __Pyx_PyObject_Call2Args(__pyx_t_11, __pyx_t_12, __pyx_kp_s__13) : __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_kp_s__13);
    __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
    if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 1354, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_split); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 1354, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_t_10 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_11))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_11);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_11, function);
      }
    }
    __pyx_t_1 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_11, __pyx_t_10, __pyx_kp_s__14) : __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_kp_s__14);
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1354, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_XDECREF_SET(__pyx_v_src_sig, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_13 = PyList_GET_SIZE(__pyx_v_dest_sig); if (unlikely(__pyx_t_13 == ((Py_ssize_t)-1))) __PYX_ERR(1, 1354, __pyx_L1_error)
    __pyx_t_14 = __pyx_t_13;
    for (__pyx_t_15 = 0; __pyx_t_15 < __pyx_t_14; __pyx_t_15+=1) {
      __pyx_v_i = __pyx_t_15;
      __pyx_t_1 = PyList_GET_ITEM(__pyx_v_dest_sig, __pyx_v_i);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_XDECREF_SET(__pyx_v_dst_type, __pyx_t_1);
      __pyx_t_1 = 0;
      __pyx_t_2 = (__pyx_v_dst_type != Py_None);
      __pyx_t_3 = (__pyx_t_2 != 0);
      if (__pyx_t_3) {
        __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_src_sig, __pyx_v_i, Py_ssize_t, 1, PyInt_FromSsize_t, 0, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1354, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_11 = PyObject_RichCompare(__pyx_t_1, __pyx_v_dst_type, Py_EQ); __Pyx_XGOTREF(__pyx_t_11); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 1354, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_11); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(1, 1354, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        if (__pyx_t_3) {
          __pyx_v_match_found = 1;
          goto __pyx_L20;
        }
        /*else*/ {
          __pyx_v_match_found = 0;
          goto __pyx_L18_break;
        }
        __pyx_L20:;
      }
    }
    __pyx_L18_break:;
    __pyx_t_3 = (__pyx_v_match_found != 0);
    if (__pyx_t_3) {
      __pyx_t_16 = __Pyx_PyList_Append(__pyx_v_candidates, __pyx_v_sig); if (unlikely(__pyx_t_16 == ((int)-1))) __PYX_ERR(1, 1354, __pyx_L1_error)
    }
  }
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_3 = (PyList_GET_SIZE(__pyx_v_candidates) != 0);
  __pyx_t_2 = ((!__pyx_t_3) != 0);
  if (__pyx_t_2) {
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__15, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 1354, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(1, 1354, __pyx_L1_error)
  }
  __pyx_t_7 = PyList_GET_SIZE(__pyx_v_candidates); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1))) __PYX_ERR(1, 1354, __pyx_L1_error)
  __pyx_t_2 = ((__pyx_t_7 > 1) != 0);
  if (__pyx_t_2) {
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__16, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 1354, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(1, 1354, __pyx_L1_error)
  }
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    if (unlikely(((PyObject *)__pyx_v_signatures) == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 1354, __pyx_L1_error)
    }
    __pyx_t_6 = __Pyx_PyDict_GetItem(((PyObject*)__pyx_v_signatures), PyList_GET_ITEM(__pyx_v_candidates, 0)); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 1354, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_r = __pyx_t_6;
    __pyx_t_6 = 0;
    goto __pyx_L0;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("pandas._libs.interval.Float64ClosedBothIntervalNode.__pyx_fused_cpdef", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_dest_sig);
  __Pyx_XDECREF(__pyx_v_ndarray);
  __Pyx_XDECREF(__pyx_v_arg);
  __Pyx_XDECREF(__pyx_v_candidates);
  __Pyx_XDECREF(__pyx_v_sig);
  __Pyx_XDECREF(__pyx_v_src_sig);
  __Pyx_XDECREF(__pyx_v_dst_type);
  __Pyx_XDECREF(__pyx_v_kwargs);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_9__pyx_fuse_0query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_fuse_0__pyx_f_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float64_t __pyx_v_point, int __pyx_skip_dispatch) {
  __Pyx_memviewslice __pyx_v_indices = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_values = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __pyx_t_5numpy_int64_t __pyx_t_9;
  __pyx_t_5numpy_int64_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  int __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  __Pyx_memviewslice __pyx_t_16 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_17 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  Py_ssize_t __pyx_t_21;
  __Pyx_RefNannySetupContext("__pyx_fuse_0query", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pyx_fuse_0query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1354, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_9__pyx_fuse_0query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = PyFloat_FromDouble(__pyx_v_point); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1354, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1354, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1354, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 1354, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_5) {
            __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1354, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1354, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_3query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_4 = PyFloat_FromDouble(__pyx_v_point); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1354, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_7 = __pyx_t_1; __pyx_t_3 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1354, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1354, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1354, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (__pyx_t_3) {
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_t_4);
          __pyx_t_4 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1354, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pandas/_libs/intervaltree.pxi":1363
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
  __pyx_t_8 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_8) {

    /* "pandas/_libs/intervaltree.pxi":1367
 *             # continue the binary tree structure. Instead, we use linear
 *             # search.
 *             for i in range(self.n_elements):             # <<<<<<<<<<<<<<
 *                 if self.left[i] <= point <= self.right[i]:
 *                     result.append(self.indices[i])
 */
    __pyx_t_9 = __pyx_v_self->n_elements;
    __pyx_t_10 = __pyx_t_9;
    for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
      __pyx_v_i = __pyx_t_11;

      /* "pandas/_libs/intervaltree.pxi":1368
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point <= self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      __pyx_t_12 = __pyx_v_i;
      __pyx_t_8 = ((*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_self->left.data + __pyx_t_12 * __pyx_v_self->left.strides[0]) ))) <= __pyx_v_point);
      if (__pyx_t_8) {
        __pyx_t_13 = __pyx_v_i;
        __pyx_t_8 = (__pyx_v_point <= (*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_self->right.data + __pyx_t_13 * __pyx_v_self->right.strides[0]) ))));
      }
      __pyx_t_14 = (__pyx_t_8 != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":1369
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point <= self.right[i]:
 *                     result.append(self.indices[i])             # <<<<<<<<<<<<<<
 *         else:
 *             # There are child nodes. Based on comparing our query to the pivot,
 */
        __pyx_t_15 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->indices.data + __pyx_t_15 * __pyx_v_self->indices.strides[0]) ))));

        /* "pandas/_libs/intervaltree.pxi":1368
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point <= self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      }
    }

    /* "pandas/_libs/intervaltree.pxi":1363
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
    goto __pyx_L3;
  }

  /* "pandas/_libs/intervaltree.pxi":1373
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
  /*else*/ {
    __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":1374
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:
 *                 values = self.center_left_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 */
      __pyx_t_16 = __pyx_v_self->center_left_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":1375
 *             if point < self.pivot:
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:
 */
      __pyx_t_17 = __pyx_v_self->center_left_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":1376
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):             # <<<<<<<<<<<<<<
 *                     if not values[i] <= point:
 *                         break
 */
      __pyx_t_9 = __pyx_v_self->n_center;
      __pyx_t_10 = __pyx_t_9;
      for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":1377
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_18 = __pyx_v_i;
        __pyx_t_14 = ((!(((*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_18 * __pyx_v_values.strides[0]) ))) <= __pyx_v_point) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":1378
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:
 */
          goto __pyx_L9_break;

          /* "pandas/_libs/intervaltree.pxi":1377
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":1379
 *                     if not values[i] <= point:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 */
        __pyx_t_19 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_19 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L9_break:;

      /* "pandas/_libs/intervaltree.pxi":1380
 *                         break
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      __pyx_t_14 = ((__pyx_v_point <= __pyx_v_self->left_node->max_right) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":1381
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)             # <<<<<<<<<<<<<<
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode *)__pyx_v_self->left_node->__pyx_vtab)->__pyx_fuse_0query(__pyx_v_self->left_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1381, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":1380
 *                         break
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":1373
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":1382
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
    __pyx_t_14 = ((__pyx_v_point > __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":1383
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 *                 values = self.center_right_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 */
      __pyx_t_16 = __pyx_v_self->center_right_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":1384
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:
 */
      __pyx_t_17 = __pyx_v_self->center_right_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":1385
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):             # <<<<<<<<<<<<<<
 *                     if not point <= values[i]:
 *                         break
 */
      for (__pyx_t_11 = (__pyx_v_self->n_center - 1); __pyx_t_11 > -1L; __pyx_t_11-=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":1386
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_20 = __pyx_v_i;
        __pyx_t_14 = ((!((__pyx_v_point <= (*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_20 * __pyx_v_values.strides[0]) )))) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":1387
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:
 */
          goto __pyx_L13_break;

          /* "pandas/_libs/intervaltree.pxi":1386
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":1388
 *                     if not point <= values[i]:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if self.right_node.min_left <= point:
 *                     self.right_node.query(result, point)
 */
        __pyx_t_21 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_21 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L13_break:;

      /* "pandas/_libs/intervaltree.pxi":1389
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      __pyx_t_14 = ((__pyx_v_self->right_node->min_left <= __pyx_v_point) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":1390
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:
 *                     self.right_node.query(result, point)             # <<<<<<<<<<<<<<
 *             else:
 *                 result.extend(self.center_left_indices)
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode *)__pyx_v_self->right_node->__pyx_vtab)->__pyx_fuse_0query(__pyx_v_self->right_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1390, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":1389
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":1382
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":1392
 *                     self.right_node.query(result, point)
 *             else:
 *                 result.extend(self.center_left_indices)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
    /*else*/ {
      __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->extend(__pyx_v_result, __pyx_v_self->center_left_indices); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1392, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_L7:;
  }
  __pyx_L3:;

  /* "pandas/_libs/intervaltree.pxi":1354
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __PYX_XDEC_MEMVIEW(&__pyx_t_16, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_17, 1);
  __Pyx_AddTraceback("pandas._libs.interval.Float64ClosedBothIntervalNode.query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_indices, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_values, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_9__pyx_fuse_0query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_0__pyx_mdef_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_9__pyx_fuse_0query = {"__pyx_fuse_0query", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_9__pyx_fuse_0query, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_2query};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_9__pyx_fuse_0query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result = 0;
  __pyx_t_5numpy_float64_t __pyx_v_point;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fuse_0query (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_result,&__pyx_n_s_point,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_result)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_point)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_0query", 1, 2, 2, 1); __PYX_ERR(1, 1354, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_0query") < 0)) __PYX_ERR(1, 1354, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_result = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)values[0]);
    __pyx_v_point = __pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_point == ((npy_float64)-1)) && PyErr_Occurred())) __PYX_ERR(1, 1354, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_0query", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 1354, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Float64ClosedBothIntervalNode.__pyx_fuse_0query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_result), __pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector, 1, "result", 0))) __PYX_ERR(1, 1354, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_8__pyx_fuse_0query(((struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode *)__pyx_v_self), __pyx_v_result, __pyx_v_point);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_8__pyx_fuse_0query(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float64_t __pyx_v_point) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__pyx_fuse_0query", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_fuse_0__pyx_f_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_query(__pyx_v_self, __pyx_v_result, __pyx_v_point, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1354, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Float64ClosedBothIntervalNode.__pyx_fuse_0query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_11__pyx_fuse_1query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_fuse_1__pyx_f_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float32_t __pyx_v_point, int __pyx_skip_dispatch) {
  __Pyx_memviewslice __pyx_v_indices = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_values = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __pyx_t_5numpy_int64_t __pyx_t_9;
  __pyx_t_5numpy_int64_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  int __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  __Pyx_memviewslice __pyx_t_16 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_17 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  Py_ssize_t __pyx_t_21;
  __Pyx_RefNannySetupContext("__pyx_fuse_1query", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pyx_fuse_1query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1354, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_11__pyx_fuse_1query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = PyFloat_FromDouble(__pyx_v_point); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1354, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1354, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1354, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 1354, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_5) {
            __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1354, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1354, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_3query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_4 = PyFloat_FromDouble(__pyx_v_point); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1354, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_7 = __pyx_t_1; __pyx_t_3 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1354, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1354, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1354, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (__pyx_t_3) {
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_t_4);
          __pyx_t_4 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1354, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pandas/_libs/intervaltree.pxi":1363
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
  __pyx_t_8 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_8) {

    /* "pandas/_libs/intervaltree.pxi":1367
 *             # continue the binary tree structure. Instead, we use linear
 *             # search.
 *             for i in range(self.n_elements):             # <<<<<<<<<<<<<<
 *                 if self.left[i] <= point <= self.right[i]:
 *                     result.append(self.indices[i])
 */
    __pyx_t_9 = __pyx_v_self->n_elements;
    __pyx_t_10 = __pyx_t_9;
    for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
      __pyx_v_i = __pyx_t_11;

      /* "pandas/_libs/intervaltree.pxi":1368
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point <= self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      __pyx_t_12 = __pyx_v_i;
      __pyx_t_8 = ((*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_self->left.data + __pyx_t_12 * __pyx_v_self->left.strides[0]) ))) <= __pyx_v_point);
      if (__pyx_t_8) {
        __pyx_t_13 = __pyx_v_i;
        __pyx_t_8 = (__pyx_v_point <= (*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_self->right.data + __pyx_t_13 * __pyx_v_self->right.strides[0]) ))));
      }
      __pyx_t_14 = (__pyx_t_8 != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":1369
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point <= self.right[i]:
 *                     result.append(self.indices[i])             # <<<<<<<<<<<<<<
 *         else:
 *             # There are child nodes. Based on comparing our query to the pivot,
 */
        __pyx_t_15 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->indices.data + __pyx_t_15 * __pyx_v_self->indices.strides[0]) ))));

        /* "pandas/_libs/intervaltree.pxi":1368
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point <= self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      }
    }

    /* "pandas/_libs/intervaltree.pxi":1363
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
    goto __pyx_L3;
  }

  /* "pandas/_libs/intervaltree.pxi":1373
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
  /*else*/ {
    __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":1374
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:
 *                 values = self.center_left_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 */
      __pyx_t_16 = __pyx_v_self->center_left_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":1375
 *             if point < self.pivot:
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:
 */
      __pyx_t_17 = __pyx_v_self->center_left_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":1376
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):             # <<<<<<<<<<<<<<
 *                     if not values[i] <= point:
 *                         break
 */
      __pyx_t_9 = __pyx_v_self->n_center;
      __pyx_t_10 = __pyx_t_9;
      for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":1377
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_18 = __pyx_v_i;
        __pyx_t_14 = ((!(((*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_18 * __pyx_v_values.strides[0]) ))) <= __pyx_v_point) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":1378
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:
 */
          goto __pyx_L9_break;

          /* "pandas/_libs/intervaltree.pxi":1377
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":1379
 *                     if not values[i] <= point:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 */
        __pyx_t_19 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_19 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L9_break:;

      /* "pandas/_libs/intervaltree.pxi":1380
 *                         break
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      __pyx_t_14 = ((__pyx_v_point <= __pyx_v_self->left_node->max_right) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":1381
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)             # <<<<<<<<<<<<<<
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode *)__pyx_v_self->left_node->__pyx_vtab)->__pyx_fuse_1query(__pyx_v_self->left_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1381, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":1380
 *                         break
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":1373
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":1382
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
    __pyx_t_14 = ((__pyx_v_point > __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":1383
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 *                 values = self.center_right_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 */
      __pyx_t_16 = __pyx_v_self->center_right_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":1384
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:
 */
      __pyx_t_17 = __pyx_v_self->center_right_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":1385
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):             # <<<<<<<<<<<<<<
 *                     if not point <= values[i]:
 *                         break
 */
      for (__pyx_t_11 = (__pyx_v_self->n_center - 1); __pyx_t_11 > -1L; __pyx_t_11-=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":1386
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_20 = __pyx_v_i;
        __pyx_t_14 = ((!((__pyx_v_point <= (*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_20 * __pyx_v_values.strides[0]) )))) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":1387
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:
 */
          goto __pyx_L13_break;

          /* "pandas/_libs/intervaltree.pxi":1386
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":1388
 *                     if not point <= values[i]:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if self.right_node.min_left <= point:
 *                     self.right_node.query(result, point)
 */
        __pyx_t_21 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_21 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L13_break:;

      /* "pandas/_libs/intervaltree.pxi":1389
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      __pyx_t_14 = ((__pyx_v_self->right_node->min_left <= __pyx_v_point) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":1390
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:
 *                     self.right_node.query(result, point)             # <<<<<<<<<<<<<<
 *             else:
 *                 result.extend(self.center_left_indices)
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode *)__pyx_v_self->right_node->__pyx_vtab)->__pyx_fuse_1query(__pyx_v_self->right_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1390, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":1389
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":1382
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":1392
 *                     self.right_node.query(result, point)
 *             else:
 *                 result.extend(self.center_left_indices)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
    /*else*/ {
      __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->extend(__pyx_v_result, __pyx_v_self->center_left_indices); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1392, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_L7:;
  }
  __pyx_L3:;

  /* "pandas/_libs/intervaltree.pxi":1354
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __PYX_XDEC_MEMVIEW(&__pyx_t_16, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_17, 1);
  __Pyx_AddTraceback("pandas._libs.interval.Float64ClosedBothIntervalNode.query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_indices, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_values, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_11__pyx_fuse_1query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_1__pyx_mdef_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_11__pyx_fuse_1query = {"__pyx_fuse_1query", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_11__pyx_fuse_1query, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_2query};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_11__pyx_fuse_1query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result = 0;
  __pyx_t_5numpy_float32_t __pyx_v_point;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fuse_1query (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_result,&__pyx_n_s_point,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_result)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_point)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_1query", 1, 2, 2, 1); __PYX_ERR(1, 1354, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_1query") < 0)) __PYX_ERR(1, 1354, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_result = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)values[0]);
    __pyx_v_point = __pyx_PyFloat_AsFloat(values[1]); if (unlikely((__pyx_v_point == ((npy_float32)-1)) && PyErr_Occurred())) __PYX_ERR(1, 1354, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_1query", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 1354, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Float64ClosedBothIntervalNode.__pyx_fuse_1query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_result), __pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector, 1, "result", 0))) __PYX_ERR(1, 1354, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_10__pyx_fuse_1query(((struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode *)__pyx_v_self), __pyx_v_result, __pyx_v_point);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_10__pyx_fuse_1query(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float32_t __pyx_v_point) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__pyx_fuse_1query", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_fuse_1__pyx_f_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_query(__pyx_v_self, __pyx_v_result, __pyx_v_point, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1354, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Float64ClosedBothIntervalNode.__pyx_fuse_1query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_13__pyx_fuse_2query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_fuse_2__pyx_f_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int64_t __pyx_v_point, int __pyx_skip_dispatch) {
  __Pyx_memviewslice __pyx_v_indices = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_values = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __pyx_t_5numpy_int64_t __pyx_t_9;
  __pyx_t_5numpy_int64_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  int __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  __Pyx_memviewslice __pyx_t_16 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_17 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  Py_ssize_t __pyx_t_21;
  __Pyx_RefNannySetupContext("__pyx_fuse_2query", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pyx_fuse_2query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1354, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_13__pyx_fuse_2query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = __Pyx_PyInt_From_npy_int64(__pyx_v_point); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1354, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1354, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1354, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 1354, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_5) {
            __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1354, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1354, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_3query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_4 = __Pyx_PyInt_From_npy_int64(__pyx_v_point); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1354, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_7 = __pyx_t_1; __pyx_t_3 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1354, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1354, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1354, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (__pyx_t_3) {
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_t_4);
          __pyx_t_4 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1354, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pandas/_libs/intervaltree.pxi":1363
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
  __pyx_t_8 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_8) {

    /* "pandas/_libs/intervaltree.pxi":1367
 *             # continue the binary tree structure. Instead, we use linear
 *             # search.
 *             for i in range(self.n_elements):             # <<<<<<<<<<<<<<
 *                 if self.left[i] <= point <= self.right[i]:
 *                     result.append(self.indices[i])
 */
    __pyx_t_9 = __pyx_v_self->n_elements;
    __pyx_t_10 = __pyx_t_9;
    for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
      __pyx_v_i = __pyx_t_11;

      /* "pandas/_libs/intervaltree.pxi":1368
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point <= self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      __pyx_t_12 = __pyx_v_i;
      __pyx_t_8 = ((*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_self->left.data + __pyx_t_12 * __pyx_v_self->left.strides[0]) ))) <= __pyx_v_point);
      if (__pyx_t_8) {
        __pyx_t_13 = __pyx_v_i;
        __pyx_t_8 = (__pyx_v_point <= (*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_self->right.data + __pyx_t_13 * __pyx_v_self->right.strides[0]) ))));
      }
      __pyx_t_14 = (__pyx_t_8 != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":1369
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point <= self.right[i]:
 *                     result.append(self.indices[i])             # <<<<<<<<<<<<<<
 *         else:
 *             # There are child nodes. Based on comparing our query to the pivot,
 */
        __pyx_t_15 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->indices.data + __pyx_t_15 * __pyx_v_self->indices.strides[0]) ))));

        /* "pandas/_libs/intervaltree.pxi":1368
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point <= self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      }
    }

    /* "pandas/_libs/intervaltree.pxi":1363
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
    goto __pyx_L3;
  }

  /* "pandas/_libs/intervaltree.pxi":1373
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
  /*else*/ {
    __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":1374
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:
 *                 values = self.center_left_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 */
      __pyx_t_16 = __pyx_v_self->center_left_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":1375
 *             if point < self.pivot:
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:
 */
      __pyx_t_17 = __pyx_v_self->center_left_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":1376
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):             # <<<<<<<<<<<<<<
 *                     if not values[i] <= point:
 *                         break
 */
      __pyx_t_9 = __pyx_v_self->n_center;
      __pyx_t_10 = __pyx_t_9;
      for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":1377
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_18 = __pyx_v_i;
        __pyx_t_14 = ((!(((*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_18 * __pyx_v_values.strides[0]) ))) <= __pyx_v_point) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":1378
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:
 */
          goto __pyx_L9_break;

          /* "pandas/_libs/intervaltree.pxi":1377
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":1379
 *                     if not values[i] <= point:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 */
        __pyx_t_19 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_19 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L9_break:;

      /* "pandas/_libs/intervaltree.pxi":1380
 *                         break
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      __pyx_t_14 = ((__pyx_v_point <= __pyx_v_self->left_node->max_right) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":1381
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)             # <<<<<<<<<<<<<<
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode *)__pyx_v_self->left_node->__pyx_vtab)->__pyx_fuse_2query(__pyx_v_self->left_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1381, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":1380
 *                         break
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":1373
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":1382
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
    __pyx_t_14 = ((__pyx_v_point > __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":1383
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 *                 values = self.center_right_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 */
      __pyx_t_16 = __pyx_v_self->center_right_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":1384
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:
 */
      __pyx_t_17 = __pyx_v_self->center_right_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":1385
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):             # <<<<<<<<<<<<<<
 *                     if not point <= values[i]:
 *                         break
 */
      for (__pyx_t_11 = (__pyx_v_self->n_center - 1); __pyx_t_11 > -1L; __pyx_t_11-=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":1386
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_20 = __pyx_v_i;
        __pyx_t_14 = ((!((__pyx_v_point <= (*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_20 * __pyx_v_values.strides[0]) )))) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":1387
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:
 */
          goto __pyx_L13_break;

          /* "pandas/_libs/intervaltree.pxi":1386
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":1388
 *                     if not point <= values[i]:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if self.right_node.min_left <= point:
 *                     self.right_node.query(result, point)
 */
        __pyx_t_21 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_21 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L13_break:;

      /* "pandas/_libs/intervaltree.pxi":1389
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      __pyx_t_14 = ((__pyx_v_self->right_node->min_left <= __pyx_v_point) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":1390
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:
 *                     self.right_node.query(result, point)             # <<<<<<<<<<<<<<
 *             else:
 *                 result.extend(self.center_left_indices)
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode *)__pyx_v_self->right_node->__pyx_vtab)->__pyx_fuse_2query(__pyx_v_self->right_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1390, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":1389
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":1382
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":1392
 *                     self.right_node.query(result, point)
 *             else:
 *                 result.extend(self.center_left_indices)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
    /*else*/ {
      __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->extend(__pyx_v_result, __pyx_v_self->center_left_indices); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1392, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_L7:;
  }
  __pyx_L3:;

  /* "pandas/_libs/intervaltree.pxi":1354
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __PYX_XDEC_MEMVIEW(&__pyx_t_16, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_17, 1);
  __Pyx_AddTraceback("pandas._libs.interval.Float64ClosedBothIntervalNode.query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_indices, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_values, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_13__pyx_fuse_2query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_2__pyx_mdef_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_13__pyx_fuse_2query = {"__pyx_fuse_2query", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_13__pyx_fuse_2query, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_2query};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_13__pyx_fuse_2query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result = 0;
  __pyx_t_5numpy_int64_t __pyx_v_point;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fuse_2query (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_result,&__pyx_n_s_point,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_result)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_point)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_2query", 1, 2, 2, 1); __PYX_ERR(1, 1354, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_2query") < 0)) __PYX_ERR(1, 1354, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_result = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)values[0]);
    __pyx_v_point = __Pyx_PyInt_As_npy_int64(values[1]); if (unlikely((__pyx_v_point == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(1, 1354, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_2query", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 1354, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Float64ClosedBothIntervalNode.__pyx_fuse_2query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_result), __pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector, 1, "result", 0))) __PYX_ERR(1, 1354, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_12__pyx_fuse_2query(((struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode *)__pyx_v_self), __pyx_v_result, __pyx_v_point);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_12__pyx_fuse_2query(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int64_t __pyx_v_point) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__pyx_fuse_2query", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_fuse_2__pyx_f_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_query(__pyx_v_self, __pyx_v_result, __pyx_v_point, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1354, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Float64ClosedBothIntervalNode.__pyx_fuse_2query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_15__pyx_fuse_3query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_fuse_3__pyx_f_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int32_t __pyx_v_point, int __pyx_skip_dispatch) {
  __Pyx_memviewslice __pyx_v_indices = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_values = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __pyx_t_5numpy_int64_t __pyx_t_9;
  __pyx_t_5numpy_int64_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  int __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  __Pyx_memviewslice __pyx_t_16 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_17 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  Py_ssize_t __pyx_t_21;
  __Pyx_RefNannySetupContext("__pyx_fuse_3query", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pyx_fuse_3query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1354, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_15__pyx_fuse_3query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = __Pyx_PyInt_From_npy_int32(__pyx_v_point); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1354, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1354, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1354, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 1354, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_5) {
            __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1354, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1354, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_3query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_4 = __Pyx_PyInt_From_npy_int32(__pyx_v_point); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1354, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_7 = __pyx_t_1; __pyx_t_3 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1354, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1354, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1354, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (__pyx_t_3) {
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_t_4);
          __pyx_t_4 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1354, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pandas/_libs/intervaltree.pxi":1363
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
  __pyx_t_8 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_8) {

    /* "pandas/_libs/intervaltree.pxi":1367
 *             # continue the binary tree structure. Instead, we use linear
 *             # search.
 *             for i in range(self.n_elements):             # <<<<<<<<<<<<<<
 *                 if self.left[i] <= point <= self.right[i]:
 *                     result.append(self.indices[i])
 */
    __pyx_t_9 = __pyx_v_self->n_elements;
    __pyx_t_10 = __pyx_t_9;
    for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
      __pyx_v_i = __pyx_t_11;

      /* "pandas/_libs/intervaltree.pxi":1368
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point <= self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      __pyx_t_12 = __pyx_v_i;
      __pyx_t_8 = ((*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_self->left.data + __pyx_t_12 * __pyx_v_self->left.strides[0]) ))) <= __pyx_v_point);
      if (__pyx_t_8) {
        __pyx_t_13 = __pyx_v_i;
        __pyx_t_8 = (__pyx_v_point <= (*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_self->right.data + __pyx_t_13 * __pyx_v_self->right.strides[0]) ))));
      }
      __pyx_t_14 = (__pyx_t_8 != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":1369
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point <= self.right[i]:
 *                     result.append(self.indices[i])             # <<<<<<<<<<<<<<
 *         else:
 *             # There are child nodes. Based on comparing our query to the pivot,
 */
        __pyx_t_15 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->indices.data + __pyx_t_15 * __pyx_v_self->indices.strides[0]) ))));

        /* "pandas/_libs/intervaltree.pxi":1368
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point <= self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      }
    }

    /* "pandas/_libs/intervaltree.pxi":1363
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
    goto __pyx_L3;
  }

  /* "pandas/_libs/intervaltree.pxi":1373
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
  /*else*/ {
    __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":1374
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:
 *                 values = self.center_left_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 */
      __pyx_t_16 = __pyx_v_self->center_left_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":1375
 *             if point < self.pivot:
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:
 */
      __pyx_t_17 = __pyx_v_self->center_left_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":1376
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):             # <<<<<<<<<<<<<<
 *                     if not values[i] <= point:
 *                         break
 */
      __pyx_t_9 = __pyx_v_self->n_center;
      __pyx_t_10 = __pyx_t_9;
      for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":1377
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_18 = __pyx_v_i;
        __pyx_t_14 = ((!(((*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_18 * __pyx_v_values.strides[0]) ))) <= __pyx_v_point) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":1378
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:
 */
          goto __pyx_L9_break;

          /* "pandas/_libs/intervaltree.pxi":1377
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":1379
 *                     if not values[i] <= point:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 */
        __pyx_t_19 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_19 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L9_break:;

      /* "pandas/_libs/intervaltree.pxi":1380
 *                         break
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      __pyx_t_14 = ((__pyx_v_point <= __pyx_v_self->left_node->max_right) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":1381
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)             # <<<<<<<<<<<<<<
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode *)__pyx_v_self->left_node->__pyx_vtab)->__pyx_fuse_3query(__pyx_v_self->left_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1381, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":1380
 *                         break
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":1373
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":1382
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
    __pyx_t_14 = ((__pyx_v_point > __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":1383
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 *                 values = self.center_right_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 */
      __pyx_t_16 = __pyx_v_self->center_right_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":1384
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:
 */
      __pyx_t_17 = __pyx_v_self->center_right_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":1385
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):             # <<<<<<<<<<<<<<
 *                     if not point <= values[i]:
 *                         break
 */
      for (__pyx_t_11 = (__pyx_v_self->n_center - 1); __pyx_t_11 > -1L; __pyx_t_11-=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":1386
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_20 = __pyx_v_i;
        __pyx_t_14 = ((!((__pyx_v_point <= (*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_20 * __pyx_v_values.strides[0]) )))) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":1387
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:
 */
          goto __pyx_L13_break;

          /* "pandas/_libs/intervaltree.pxi":1386
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":1388
 *                     if not point <= values[i]:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if self.right_node.min_left <= point:
 *                     self.right_node.query(result, point)
 */
        __pyx_t_21 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_21 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L13_break:;

      /* "pandas/_libs/intervaltree.pxi":1389
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      __pyx_t_14 = ((__pyx_v_self->right_node->min_left <= __pyx_v_point) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":1390
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:
 *                     self.right_node.query(result, point)             # <<<<<<<<<<<<<<
 *             else:
 *                 result.extend(self.center_left_indices)
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode *)__pyx_v_self->right_node->__pyx_vtab)->__pyx_fuse_3query(__pyx_v_self->right_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1390, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":1389
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":1382
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":1392
 *                     self.right_node.query(result, point)
 *             else:
 *                 result.extend(self.center_left_indices)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
    /*else*/ {
      __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->extend(__pyx_v_result, __pyx_v_self->center_left_indices); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1392, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_L7:;
  }
  __pyx_L3:;

  /* "pandas/_libs/intervaltree.pxi":1354
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __PYX_XDEC_MEMVIEW(&__pyx_t_16, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_17, 1);
  __Pyx_AddTraceback("pandas._libs.interval.Float64ClosedBothIntervalNode.query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_indices, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_values, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_15__pyx_fuse_3query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_3__pyx_mdef_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_15__pyx_fuse_3query = {"__pyx_fuse_3query", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_15__pyx_fuse_3query, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_2query};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_15__pyx_fuse_3query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result = 0;
  __pyx_t_5numpy_int32_t __pyx_v_point;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fuse_3query (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_result,&__pyx_n_s_point,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_result)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_point)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_3query", 1, 2, 2, 1); __PYX_ERR(1, 1354, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_3query") < 0)) __PYX_ERR(1, 1354, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_result = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)values[0]);
    __pyx_v_point = __Pyx_PyInt_As_npy_int32(values[1]); if (unlikely((__pyx_v_point == ((npy_int32)-1)) && PyErr_Occurred())) __PYX_ERR(1, 1354, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_3query", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 1354, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Float64ClosedBothIntervalNode.__pyx_fuse_3query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_result), __pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector, 1, "result", 0))) __PYX_ERR(1, 1354, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_14__pyx_fuse_3query(((struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode *)__pyx_v_self), __pyx_v_result, __pyx_v_point);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_14__pyx_fuse_3query(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int32_t __pyx_v_point) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__pyx_fuse_3query", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_fuse_3__pyx_f_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_query(__pyx_v_self, __pyx_v_result, __pyx_v_point, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1354, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Float64ClosedBothIntervalNode.__pyx_fuse_3query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_17__pyx_fuse_4query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_fuse_4__pyx_f_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_uint64_t __pyx_v_point, int __pyx_skip_dispatch) {
  __Pyx_memviewslice __pyx_v_indices = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_values = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __pyx_t_5numpy_int64_t __pyx_t_9;
  __pyx_t_5numpy_int64_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  int __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  __Pyx_memviewslice __pyx_t_16 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_17 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  Py_ssize_t __pyx_t_21;
  __Pyx_RefNannySetupContext("__pyx_fuse_4query", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pyx_fuse_4query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1354, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_17__pyx_fuse_4query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = __Pyx_PyInt_From_npy_uint64(__pyx_v_point); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1354, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1354, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1354, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 1354, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_5) {
            __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1354, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1354, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_3query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_4 = __Pyx_PyInt_From_npy_uint64(__pyx_v_point); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1354, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_7 = __pyx_t_1; __pyx_t_3 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1354, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1354, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1354, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (__pyx_t_3) {
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_t_4);
          __pyx_t_4 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1354, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pandas/_libs/intervaltree.pxi":1363
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
  __pyx_t_8 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_8) {

    /* "pandas/_libs/intervaltree.pxi":1367
 *             # continue the binary tree structure. Instead, we use linear
 *             # search.
 *             for i in range(self.n_elements):             # <<<<<<<<<<<<<<
 *                 if self.left[i] <= point <= self.right[i]:
 *                     result.append(self.indices[i])
 */
    __pyx_t_9 = __pyx_v_self->n_elements;
    __pyx_t_10 = __pyx_t_9;
    for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
      __pyx_v_i = __pyx_t_11;

      /* "pandas/_libs/intervaltree.pxi":1368
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point <= self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      __pyx_t_12 = __pyx_v_i;
      __pyx_t_8 = ((*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_self->left.data + __pyx_t_12 * __pyx_v_self->left.strides[0]) ))) <= __pyx_v_point);
      if (__pyx_t_8) {
        __pyx_t_13 = __pyx_v_i;
        __pyx_t_8 = (__pyx_v_point <= (*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_self->right.data + __pyx_t_13 * __pyx_v_self->right.strides[0]) ))));
      }
      __pyx_t_14 = (__pyx_t_8 != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":1369
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point <= self.right[i]:
 *                     result.append(self.indices[i])             # <<<<<<<<<<<<<<
 *         else:
 *             # There are child nodes. Based on comparing our query to the pivot,
 */
        __pyx_t_15 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->indices.data + __pyx_t_15 * __pyx_v_self->indices.strides[0]) ))));

        /* "pandas/_libs/intervaltree.pxi":1368
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point <= self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      }
    }

    /* "pandas/_libs/intervaltree.pxi":1363
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
    goto __pyx_L3;
  }

  /* "pandas/_libs/intervaltree.pxi":1373
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
  /*else*/ {
    __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":1374
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:
 *                 values = self.center_left_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 */
      __pyx_t_16 = __pyx_v_self->center_left_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":1375
 *             if point < self.pivot:
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:
 */
      __pyx_t_17 = __pyx_v_self->center_left_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":1376
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):             # <<<<<<<<<<<<<<
 *                     if not values[i] <= point:
 *                         break
 */
      __pyx_t_9 = __pyx_v_self->n_center;
      __pyx_t_10 = __pyx_t_9;
      for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":1377
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_18 = __pyx_v_i;
        __pyx_t_14 = ((!(((*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_18 * __pyx_v_values.strides[0]) ))) <= __pyx_v_point) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":1378
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:
 */
          goto __pyx_L9_break;

          /* "pandas/_libs/intervaltree.pxi":1377
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":1379
 *                     if not values[i] <= point:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 */
        __pyx_t_19 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_19 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L9_break:;

      /* "pandas/_libs/intervaltree.pxi":1380
 *                         break
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      __pyx_t_14 = ((__pyx_v_point <= __pyx_v_self->left_node->max_right) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":1381
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)             # <<<<<<<<<<<<<<
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode *)__pyx_v_self->left_node->__pyx_vtab)->__pyx_fuse_4query(__pyx_v_self->left_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1381, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":1380
 *                         break
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":1373
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":1382
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
    __pyx_t_14 = ((__pyx_v_point > __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":1383
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 *                 values = self.center_right_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 */
      __pyx_t_16 = __pyx_v_self->center_right_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":1384
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:
 */
      __pyx_t_17 = __pyx_v_self->center_right_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":1385
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):             # <<<<<<<<<<<<<<
 *                     if not point <= values[i]:
 *                         break
 */
      for (__pyx_t_11 = (__pyx_v_self->n_center - 1); __pyx_t_11 > -1L; __pyx_t_11-=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":1386
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_20 = __pyx_v_i;
        __pyx_t_14 = ((!((__pyx_v_point <= (*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_20 * __pyx_v_values.strides[0]) )))) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":1387
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:
 */
          goto __pyx_L13_break;

          /* "pandas/_libs/intervaltree.pxi":1386
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":1388
 *                     if not point <= values[i]:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if self.right_node.min_left <= point:
 *                     self.right_node.query(result, point)
 */
        __pyx_t_21 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_21 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L13_break:;

      /* "pandas/_libs/intervaltree.pxi":1389
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      __pyx_t_14 = ((__pyx_v_self->right_node->min_left <= __pyx_v_point) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":1390
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:
 *                     self.right_node.query(result, point)             # <<<<<<<<<<<<<<
 *             else:
 *                 result.extend(self.center_left_indices)
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode *)__pyx_v_self->right_node->__pyx_vtab)->__pyx_fuse_4query(__pyx_v_self->right_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1390, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":1389
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":1382
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":1392
 *                     self.right_node.query(result, point)
 *             else:
 *                 result.extend(self.center_left_indices)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
    /*else*/ {
      __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->extend(__pyx_v_result, __pyx_v_self->center_left_indices); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1392, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_L7:;
  }
  __pyx_L3:;

  /* "pandas/_libs/intervaltree.pxi":1354
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __PYX_XDEC_MEMVIEW(&__pyx_t_16, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_17, 1);
  __Pyx_AddTraceback("pandas._libs.interval.Float64ClosedBothIntervalNode.query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_indices, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_values, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_17__pyx_fuse_4query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_4__pyx_mdef_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_17__pyx_fuse_4query = {"__pyx_fuse_4query", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_17__pyx_fuse_4query, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_2query};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_17__pyx_fuse_4query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result = 0;
  __pyx_t_5numpy_uint64_t __pyx_v_point;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fuse_4query (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_result,&__pyx_n_s_point,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_result)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_point)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_4query", 1, 2, 2, 1); __PYX_ERR(1, 1354, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_4query") < 0)) __PYX_ERR(1, 1354, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_result = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)values[0]);
    __pyx_v_point = __Pyx_PyInt_As_npy_uint64(values[1]); if (unlikely((__pyx_v_point == ((npy_uint64)-1)) && PyErr_Occurred())) __PYX_ERR(1, 1354, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_4query", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 1354, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Float64ClosedBothIntervalNode.__pyx_fuse_4query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_result), __pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector, 1, "result", 0))) __PYX_ERR(1, 1354, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_16__pyx_fuse_4query(((struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode *)__pyx_v_self), __pyx_v_result, __pyx_v_point);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_16__pyx_fuse_4query(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_uint64_t __pyx_v_point) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__pyx_fuse_4query", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_fuse_4__pyx_f_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_query(__pyx_v_self, __pyx_v_result, __pyx_v_point, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1354, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Float64ClosedBothIntervalNode.__pyx_fuse_4query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":1394
 *                 result.extend(self.center_left_indices)
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         if self.is_leaf_node:
 *             return ('<Float64ClosedBothIntervalNode: '
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_5__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_5__repr__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_4__repr__(((struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_4__repr__(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode *__pyx_v_self) {
  PyObject *__pyx_v_n_left = NULL;
  PyObject *__pyx_v_n_right = NULL;
  PyObject *__pyx_v_n_center = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  Py_UCS4 __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "pandas/_libs/intervaltree.pxi":1395
 * 
 *     def __repr__(self):
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             return ('<Float64ClosedBothIntervalNode: '
 *                     '%s elements (terminal)>' % self.n_elements)
 */
  __pyx_t_1 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_1) {

    /* "pandas/_libs/intervaltree.pxi":1396
 *     def __repr__(self):
 *         if self.is_leaf_node:
 *             return ('<Float64ClosedBothIntervalNode: '             # <<<<<<<<<<<<<<
 *                     '%s elements (terminal)>' % self.n_elements)
 *         else:
 */
    __Pyx_XDECREF(__pyx_r);

    /* "pandas/_libs/intervaltree.pxi":1397
 *         if self.is_leaf_node:
 *             return ('<Float64ClosedBothIntervalNode: '
 *                     '%s elements (terminal)>' % self.n_elements)             # <<<<<<<<<<<<<<
 *         else:
 *             n_left = self.left_node.n_elements
 */
    __pyx_t_2 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_elements); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1397, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PyUnicode_Format(__pyx_kp_u_Float64ClosedBothIntervalNode_s, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1397, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pandas/_libs/intervaltree.pxi":1395
 * 
 *     def __repr__(self):
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             return ('<Float64ClosedBothIntervalNode: '
 *                     '%s elements (terminal)>' % self.n_elements)
 */
  }

  /* "pandas/_libs/intervaltree.pxi":1399
 *                     '%s elements (terminal)>' % self.n_elements)
 *         else:
 *             n_left = self.left_node.n_elements             # <<<<<<<<<<<<<<
 *             n_right = self.right_node.n_elements
 *             n_center = self.n_elements - n_left - n_right
 */
  /*else*/ {
    __pyx_t_3 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->left_node->n_elements); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1399, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_v_n_left = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "pandas/_libs/intervaltree.pxi":1400
 *         else:
 *             n_left = self.left_node.n_elements
 *             n_right = self.right_node.n_elements             # <<<<<<<<<<<<<<
 *             n_center = self.n_elements - n_left - n_right
 *             return ('<Float64ClosedBothIntervalNode: '
 */
    __pyx_t_3 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->right_node->n_elements); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1400, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_v_n_right = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "pandas/_libs/intervaltree.pxi":1401
 *             n_left = self.left_node.n_elements
 *             n_right = self.right_node.n_elements
 *             n_center = self.n_elements - n_left - n_right             # <<<<<<<<<<<<<<
 *             return ('<Float64ClosedBothIntervalNode: '
 *                     'pivot %s, %s elements (%s left, %s right, %s '
 */
    __pyx_t_3 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_elements); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1401, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = PyNumber_Subtract(__pyx_t_3, __pyx_v_n_left); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1401, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyNumber_Subtract(__pyx_t_2, __pyx_v_n_right); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1401, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_n_center = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "pandas/_libs/intervaltree.pxi":1402
 *             n_right = self.right_node.n_elements
 *             n_center = self.n_elements - n_left - n_right
 *             return ('<Float64ClosedBothIntervalNode: '             # <<<<<<<<<<<<<<
 *                     'pivot %s, %s elements (%s left, %s right, %s '
 *                     'overlapping)>' % (self.pivot, self.n_elements,
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = PyTuple_New(11); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1402, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = 0;
    __pyx_t_5 = 127;
    __Pyx_INCREF(__pyx_kp_u_Float64ClosedBothIntervalNode_p);
    __pyx_t_4 += 38;
    __Pyx_GIVEREF(__pyx_kp_u_Float64ClosedBothIntervalNode_p);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_kp_u_Float64ClosedBothIntervalNode_p);

    /* "pandas/_libs/intervaltree.pxi":1404
 *             return ('<Float64ClosedBothIntervalNode: '
 *                     'pivot %s, %s elements (%s left, %s right, %s '
 *                     'overlapping)>' % (self.pivot, self.n_elements,             # <<<<<<<<<<<<<<
 *                                        n_left, n_right, n_center))
 * 
 */
    __pyx_t_2 = PyFloat_FromDouble(__pyx_v_self->pivot); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1404, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_t_2), __pyx_empty_unicode); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 1404, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_5 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_6) > __pyx_t_5) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_6) : __pyx_t_5;
    __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_6);
    __pyx_t_6 = 0;
    __Pyx_INCREF(__pyx_kp_u__18);
    __pyx_t_4 += 2;
    __Pyx_GIVEREF(__pyx_kp_u__18);
    PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_kp_u__18);
    __pyx_t_6 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_elements); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 1404, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_2 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_t_6), __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1404, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_5 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) > __pyx_t_5) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) : __pyx_t_5;
    __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_3, 3, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_elements);
    __pyx_t_4 += 11;
    __Pyx_GIVEREF(__pyx_kp_u_elements);
    PyTuple_SET_ITEM(__pyx_t_3, 4, __pyx_kp_u_elements);

    /* "pandas/_libs/intervaltree.pxi":1405
 *                     'pivot %s, %s elements (%s left, %s right, %s '
 *                     'overlapping)>' % (self.pivot, self.n_elements,
 *                                        n_left, n_right, n_center))             # <<<<<<<<<<<<<<
 * 
 *     def counts(self):
 */
    __pyx_t_2 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_v_n_left), __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1405, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) > __pyx_t_5) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) : __pyx_t_5;
    __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_3, 5, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_left_2);
    __pyx_t_4 += 7;
    __Pyx_GIVEREF(__pyx_kp_u_left_2);
    PyTuple_SET_ITEM(__pyx_t_3, 6, __pyx_kp_u_left_2);
    __pyx_t_2 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_v_n_right), __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1405, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) > __pyx_t_5) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) : __pyx_t_5;
    __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_3, 7, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_right_2);
    __pyx_t_4 += 8;
    __Pyx_GIVEREF(__pyx_kp_u_right_2);
    PyTuple_SET_ITEM(__pyx_t_3, 8, __pyx_kp_u_right_2);
    __pyx_t_2 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_v_n_center), __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1405, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) > __pyx_t_5) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) : __pyx_t_5;
    __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_3, 9, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_overlapping);
    __pyx_t_4 += 14;
    __Pyx_GIVEREF(__pyx_kp_u_overlapping);
    PyTuple_SET_ITEM(__pyx_t_3, 10, __pyx_kp_u_overlapping);

    /* "pandas/_libs/intervaltree.pxi":1402
 *             n_right = self.right_node.n_elements
 *             n_center = self.n_elements - n_left - n_right
 *             return ('<Float64ClosedBothIntervalNode: '             # <<<<<<<<<<<<<<
 *                     'pivot %s, %s elements (%s left, %s right, %s '
 *                     'overlapping)>' % (self.pivot, self.n_elements,
 */
    __pyx_t_2 = __Pyx_PyUnicode_Join(__pyx_t_3, 11, __pyx_t_4, __pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1402, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;
  }

  /* "pandas/_libs/intervaltree.pxi":1394
 *                 result.extend(self.center_left_indices)
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         if self.is_leaf_node:
 *             return ('<Float64ClosedBothIntervalNode: '
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pandas._libs.interval.Float64ClosedBothIntervalNode.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_n_left);
  __Pyx_XDECREF(__pyx_v_n_right);
  __Pyx_XDECREF(__pyx_v_n_center);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":1407
 *                                        n_left, n_right, n_center))
 * 
 *     def counts(self):             # <<<<<<<<<<<<<<
 *         """
 *         Inspect counts on this node
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_7counts(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_6counts[] = "\n        Inspect counts on this node\n        useful for debugging purposes\n        ";
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_7counts(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("counts (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_6counts(((struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_6counts(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode *__pyx_v_self) {
  Py_ssize_t __pyx_v_m;
  PyObject *__pyx_v_l = NULL;
  PyObject *__pyx_v_r = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  size_t __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("counts", 0);

  /* "pandas/_libs/intervaltree.pxi":1412
 *         useful for debugging purposes
 *         """
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             return self.n_elements
 *         else:
 */
  __pyx_t_1 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_1) {

    /* "pandas/_libs/intervaltree.pxi":1413
 *         """
 *         if self.is_leaf_node:
 *             return self.n_elements             # <<<<<<<<<<<<<<
 *         else:
 *             m = len(self.center_left_values)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_elements); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1413, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "pandas/_libs/intervaltree.pxi":1412
 *         useful for debugging purposes
 *         """
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             return self.n_elements
 *         else:
 */
  }

  /* "pandas/_libs/intervaltree.pxi":1415
 *             return self.n_elements
 *         else:
 *             m = len(self.center_left_values)             # <<<<<<<<<<<<<<
 *             l = self.left_node.counts()
 *             r = self.right_node.counts()
 */
  /*else*/ {
    if (unlikely(!__pyx_v_self->center_left_values.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(1, 1415, __pyx_L1_error)}
    __pyx_t_3 = __Pyx_MemoryView_Len(__pyx_v_self->center_left_values); 
    __pyx_v_m = __pyx_t_3;

    /* "pandas/_libs/intervaltree.pxi":1416
 *         else:
 *             m = len(self.center_left_values)
 *             l = self.left_node.counts()             # <<<<<<<<<<<<<<
 *             r = self.right_node.counts()
 *             return (m, (l, r))
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->left_node), __pyx_n_s_counts); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1416, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1416, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_l = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "pandas/_libs/intervaltree.pxi":1417
 *             m = len(self.center_left_values)
 *             l = self.left_node.counts()
 *             r = self.right_node.counts()             # <<<<<<<<<<<<<<
 *             return (m, (l, r))
 * 
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->right_node), __pyx_n_s_counts); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1417, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1417, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_r = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "pandas/_libs/intervaltree.pxi":1418
 *             l = self.left_node.counts()
 *             r = self.right_node.counts()
 *             return (m, (l, r))             # <<<<<<<<<<<<<<
 * 
 * NODE_CLASSES['float64',
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = PyInt_FromSsize_t(__pyx_v_m); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1418, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1418, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_v_l);
    __Pyx_GIVEREF(__pyx_v_l);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_l);
    __Pyx_INCREF(__pyx_v_r);
    __Pyx_GIVEREF(__pyx_v_r);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_v_r);
    __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1418, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_4);
    __pyx_t_2 = 0;
    __pyx_t_4 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;
  }

  /* "pandas/_libs/intervaltree.pxi":1407
 *                                        n_left, n_right, n_center))
 * 
 *     def counts(self):             # <<<<<<<<<<<<<<
 *         """
 *         Inspect counts on this node
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pandas._libs.interval.Float64ClosedBothIntervalNode.counts", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_l);
  __Pyx_XDECREF(__pyx_v_r);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":1264
 *         int64_t[:] center_left_indices, center_right_indices, indices
 *         float64_t min_left, max_right
 *         readonly float64_t pivot             # <<<<<<<<<<<<<<
 *         readonly int64_t n_elements, n_center, leaf_size
 *         readonly bint is_leaf_node
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_5pivot_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_5pivot_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_5pivot___get__(((struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_5pivot___get__(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->pivot); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1264, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Float64ClosedBothIntervalNode.pivot.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":1265
 *         float64_t min_left, max_right
 *         readonly float64_t pivot
 *         readonly int64_t n_elements, n_center, leaf_size             # <<<<<<<<<<<<<<
 *         readonly bint is_leaf_node
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_10n_elements_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_10n_elements_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_10n_elements___get__(((struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_10n_elements___get__(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_elements); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1265, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Float64ClosedBothIntervalNode.n_elements.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_8n_center_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_8n_center_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_8n_center___get__(((struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_8n_center___get__(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_center); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1265, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Float64ClosedBothIntervalNode.n_center.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_9leaf_size_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_9leaf_size_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_9leaf_size___get__(((struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_9leaf_size___get__(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->leaf_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1265, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Float64ClosedBothIntervalNode.leaf_size.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":1266
 *         readonly float64_t pivot
 *         readonly int64_t n_elements, n_center, leaf_size
 *         readonly bint is_leaf_node             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self,
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_12is_leaf_node_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_12is_leaf_node_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_12is_leaf_node___get__(((struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_12is_leaf_node___get__(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->is_leaf_node); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1266, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Float64ClosedBothIntervalNode.is_leaf_node.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_21__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_21__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_20__reduce_cython__(((struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_20__reduce_cython__(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  int __pyx_t_16;
  int __pyx_t_17;
  int __pyx_t_18;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.center_left_indices, self.center_left_values, self.center_right_indices, self.center_right_values, self.indices, self.is_leaf_node, self.leaf_size, self.left, self.left_node, self.max_right, self.min_left, self.n_center, self.n_elements, self.pivot, self.right, self.right_node)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  if (unlikely(!__pyx_v_self->center_left_indices.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_1 = __pyx_memoryview_fromslice(__pyx_v_self->center_left_indices, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_int64_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_int64_t, 0);; if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (unlikely(!__pyx_v_self->center_left_values.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_2 = __pyx_memoryview_fromslice(__pyx_v_self->center_left_values, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_float64_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_float64_t, 0);; if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (unlikely(!__pyx_v_self->center_right_indices.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_3 = __pyx_memoryview_fromslice(__pyx_v_self->center_right_indices, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_int64_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_int64_t, 0);; if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely(!__pyx_v_self->center_right_values.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_4 = __pyx_memoryview_fromslice(__pyx_v_self->center_right_values, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_float64_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_float64_t, 0);; if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (unlikely(!__pyx_v_self->indices.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_5 = __pyx_memoryview_fromslice(__pyx_v_self->indices, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_int64_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_int64_t, 0);; if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyBool_FromLong(__pyx_v_self->is_leaf_node); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->leaf_size); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (unlikely(!__pyx_v_self->left.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_8 = __pyx_memoryview_fromslice(__pyx_v_self->left, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_float64_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_float64_t, 0);; if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = PyFloat_FromDouble(__pyx_v_self->max_right); if (unlikely(!__pyx_t_9)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_10 = PyFloat_FromDouble(__pyx_v_self->min_left); if (unlikely(!__pyx_t_10)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_11 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_center); if (unlikely(!__pyx_t_11)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_12 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_elements); if (unlikely(!__pyx_t_12)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __pyx_t_13 = PyFloat_FromDouble(__pyx_v_self->pivot); if (unlikely(!__pyx_t_13)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  if (unlikely(!__pyx_v_self->right.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_14 = __pyx_memoryview_fromslice(__pyx_v_self->right, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_float64_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_float64_t, 0);; if (unlikely(!__pyx_t_14)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __pyx_t_15 = PyTuple_New(16); if (unlikely(!__pyx_t_15)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_15, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_15, 1, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_15, 2, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_15, 3, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_15, 4, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_15, 5, __pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_15, 6, __pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_15, 7, __pyx_t_8);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->left_node));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self->left_node));
  PyTuple_SET_ITEM(__pyx_t_15, 8, ((PyObject *)__pyx_v_self->left_node));
  __Pyx_GIVEREF(__pyx_t_9);
  PyTuple_SET_ITEM(__pyx_t_15, 9, __pyx_t_9);
  __Pyx_GIVEREF(__pyx_t_10);
  PyTuple_SET_ITEM(__pyx_t_15, 10, __pyx_t_10);
  __Pyx_GIVEREF(__pyx_t_11);
  PyTuple_SET_ITEM(__pyx_t_15, 11, __pyx_t_11);
  __Pyx_GIVEREF(__pyx_t_12);
  PyTuple_SET_ITEM(__pyx_t_15, 12, __pyx_t_12);
  __Pyx_GIVEREF(__pyx_t_13);
  PyTuple_SET_ITEM(__pyx_t_15, 13, __pyx_t_13);
  __Pyx_GIVEREF(__pyx_t_14);
  PyTuple_SET_ITEM(__pyx_t_15, 14, __pyx_t_14);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->right_node));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self->right_node));
  PyTuple_SET_ITEM(__pyx_t_15, 15, ((PyObject *)__pyx_v_self->right_node));
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  __pyx_t_6 = 0;
  __pyx_t_7 = 0;
  __pyx_t_8 = 0;
  __pyx_t_9 = 0;
  __pyx_t_10 = 0;
  __pyx_t_11 = 0;
  __pyx_t_12 = 0;
  __pyx_t_13 = 0;
  __pyx_t_14 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_15);
  __pyx_t_15 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.center_left_indices, self.center_left_values, self.center_right_indices, self.center_right_values, self.indices, self.is_leaf_node, self.leaf_size, self.left, self.left_node, self.max_right, self.min_left, self.n_center, self.n_elements, self.pivot, self.right, self.right_node)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_15 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_15)) __PYX_ERR(2, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __pyx_v__dict = __pyx_t_15;
  __pyx_t_15 = 0;

  /* "(tree fragment)":7
 *     state = (self.center_left_indices, self.center_left_values, self.center_right_indices, self.center_right_values, self.indices, self.is_leaf_node, self.leaf_size, self.left, self.left_node, self.max_right, self.min_left, self.n_center, self.n_elements, self.pivot, self.right, self.right_node)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_16 = (__pyx_v__dict != Py_None);
  __pyx_t_17 = (__pyx_t_16 != 0);
  if (__pyx_t_17) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_15 = PyTuple_New(1); if (unlikely(!__pyx_t_15)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    PyTuple_SET_ITEM(__pyx_t_15, 0, __pyx_v__dict);
    __pyx_t_14 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_15); if (unlikely(!__pyx_t_14)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_14));
    __pyx_t_14 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.left_node is not None or self.right_node is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.center_left_indices, self.center_left_values, self.center_right_indices, self.center_right_values, self.indices, self.is_leaf_node, self.leaf_size, self.left, self.left_node, self.max_right, self.min_left, self.n_center, self.n_elements, self.pivot, self.right, self.right_node)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.left_node is not None or self.right_node is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_Float64ClosedBothIntervalNode, (type(self), 0x7a104e5, None), state
 */
  /*else*/ {
    __pyx_t_16 = (((PyObject *)__pyx_v_self->left_node) != Py_None);
    __pyx_t_18 = (__pyx_t_16 != 0);
    if (!__pyx_t_18) {
    } else {
      __pyx_t_17 = __pyx_t_18;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_18 = (((PyObject *)__pyx_v_self->right_node) != Py_None);
    __pyx_t_16 = (__pyx_t_18 != 0);
    __pyx_t_17 = __pyx_t_16;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_17;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.left_node is not None or self.right_node is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Float64ClosedBothIntervalNode, (type(self), 0x7a104e5, None), state
 *     else:
 */
  __pyx_t_17 = (__pyx_v_use_setstate != 0);
  if (__pyx_t_17) {

    /* "(tree fragment)":13
 *         use_setstate = self.left_node is not None or self.right_node is not None
 *     if use_setstate:
 *         return __pyx_unpickle_Float64ClosedBothIntervalNode, (type(self), 0x7a104e5, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_Float64ClosedBothIntervalNode, (type(self), 0x7a104e5, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_14, __pyx_n_s_pyx_unpickle_Float64ClosedBoth); if (unlikely(!__pyx_t_14)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __pyx_t_15 = PyTuple_New(3); if (unlikely(!__pyx_t_15)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_15, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_127993061);
    __Pyx_GIVEREF(__pyx_int_127993061);
    PyTuple_SET_ITEM(__pyx_t_15, 1, __pyx_int_127993061);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_15, 2, Py_None);
    __pyx_t_13 = PyTuple_New(3); if (unlikely(!__pyx_t_13)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __Pyx_GIVEREF(__pyx_t_14);
    PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_t_14);
    __Pyx_GIVEREF(__pyx_t_15);
    PyTuple_SET_ITEM(__pyx_t_13, 1, __pyx_t_15);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_13, 2, __pyx_v_state);
    __pyx_t_14 = 0;
    __pyx_t_15 = 0;
    __pyx_r = __pyx_t_13;
    __pyx_t_13 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.left_node is not None or self.right_node is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Float64ClosedBothIntervalNode, (type(self), 0x7a104e5, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_Float64ClosedBothIntervalNode, (type(self), 0x7a104e5, None), state
 *     else:
 *         return __pyx_unpickle_Float64ClosedBothIntervalNode, (type(self), 0x7a104e5, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Float64ClosedBothIntervalNode__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_13, __pyx_n_s_pyx_unpickle_Float64ClosedBoth); if (unlikely(!__pyx_t_13)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __pyx_t_15 = PyTuple_New(3); if (unlikely(!__pyx_t_15)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_15, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_127993061);
    __Pyx_GIVEREF(__pyx_int_127993061);
    PyTuple_SET_ITEM(__pyx_t_15, 1, __pyx_int_127993061);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_15, 2, __pyx_v_state);
    __pyx_t_14 = PyTuple_New(2); if (unlikely(!__pyx_t_14)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_GIVEREF(__pyx_t_13);
    PyTuple_SET_ITEM(__pyx_t_14, 0, __pyx_t_13);
    __Pyx_GIVEREF(__pyx_t_15);
    PyTuple_SET_ITEM(__pyx_t_14, 1, __pyx_t_15);
    __pyx_t_13 = 0;
    __pyx_t_15 = 0;
    __pyx_r = __pyx_t_14;
    __pyx_t_14 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_AddTraceback("pandas._libs.interval.Float64ClosedBothIntervalNode.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Float64ClosedBothIntervalNode, (type(self), 0x7a104e5, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Float64ClosedBothIntervalNode__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_23__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_23__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_22__setstate_cython__(((struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_22__setstate_cython__(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_Float64ClosedBothIntervalNode, (type(self), 0x7a104e5, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Float64ClosedBothIntervalNode__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(2, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_6pandas_5_libs_8interval___pyx_unpickle_Float64ClosedBothIntervalNode__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Float64ClosedBothIntervalNode, (type(self), 0x7a104e5, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Float64ClosedBothIntervalNode__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Float64ClosedBothIntervalNode.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":1438
 *         readonly bint is_leaf_node
 * 
 *     def __init__(self,             # <<<<<<<<<<<<<<
 *                  ndarray[float64_t, ndim=1] left,
 *                  ndarray[float64_t, ndim=1] right,
 */

/* Python wrapper */
static int __pyx_pw_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyArrayObject *__pyx_v_left = 0;
  PyArrayObject *__pyx_v_right = 0;
  PyArrayObject *__pyx_v_indices = 0;
  __pyx_t_5numpy_int64_t __pyx_v_leaf_size;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_left,&__pyx_n_s_right,&__pyx_n_s_indices,&__pyx_n_s_leaf_size,0};
    PyObject* values[4] = {0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_left)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_right)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 4, 4, 1); __PYX_ERR(1, 1438, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_indices)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 4, 4, 2); __PYX_ERR(1, 1438, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_leaf_size)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 4, 4, 3); __PYX_ERR(1, 1438, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(1, 1438, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 4) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
    }
    __pyx_v_left = ((PyArrayObject *)values[0]);
    __pyx_v_right = ((PyArrayObject *)values[1]);
    __pyx_v_indices = ((PyArrayObject *)values[2]);
    __pyx_v_leaf_size = __Pyx_PyInt_As_npy_int64(values[3]); if (unlikely((__pyx_v_leaf_size == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(1, 1442, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 4, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 1438, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Float64ClosedNeitherIntervalNode.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_left), __pyx_ptype_5numpy_ndarray, 1, "left", 0))) __PYX_ERR(1, 1439, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_right), __pyx_ptype_5numpy_ndarray, 1, "right", 0))) __PYX_ERR(1, 1440, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_indices), __pyx_ptype_5numpy_ndarray, 1, "indices", 0))) __PYX_ERR(1, 1441, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode___init__(((struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode *)__pyx_v_self), __pyx_v_left, __pyx_v_right, __pyx_v_indices, __pyx_v_leaf_size);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode___init__(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode *__pyx_v_self, PyArrayObject *__pyx_v_left, PyArrayObject *__pyx_v_right, PyArrayObject *__pyx_v_indices, __pyx_t_5numpy_int64_t __pyx_v_leaf_size) {
  PyObject *__pyx_v_left_set = NULL;
  PyObject *__pyx_v_right_set = NULL;
  PyObject *__pyx_v_center_set = NULL;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_indices;
  __Pyx_Buffer __pyx_pybuffer_indices;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_left;
  __Pyx_Buffer __pyx_pybuffer_left;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_right;
  __Pyx_Buffer __pyx_pybuffer_right;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __pyx_t_5numpy_float64_t __pyx_t_6;
  __Pyx_memviewslice __pyx_t_7 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_8 = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  __Pyx_memviewslice __pyx_t_11 = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *(*__pyx_t_12)(PyObject *);
  size_t __pyx_t_13;
  __Pyx_RefNannySetupContext("__init__", 0);
  __pyx_pybuffer_left.pybuffer.buf = NULL;
  __pyx_pybuffer_left.refcount = 0;
  __pyx_pybuffernd_left.data = NULL;
  __pyx_pybuffernd_left.rcbuffer = &__pyx_pybuffer_left;
  __pyx_pybuffer_right.pybuffer.buf = NULL;
  __pyx_pybuffer_right.refcount = 0;
  __pyx_pybuffernd_right.data = NULL;
  __pyx_pybuffernd_right.rcbuffer = &__pyx_pybuffer_right;
  __pyx_pybuffer_indices.pybuffer.buf = NULL;
  __pyx_pybuffer_indices.refcount = 0;
  __pyx_pybuffernd_indices.data = NULL;
  __pyx_pybuffernd_indices.rcbuffer = &__pyx_pybuffer_indices;
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_left.rcbuffer->pybuffer, (PyObject*)__pyx_v_left, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 1438, __pyx_L1_error)
  }
  __pyx_pybuffernd_left.diminfo[0].strides = __pyx_pybuffernd_left.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_left.diminfo[0].shape = __pyx_pybuffernd_left.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_right.rcbuffer->pybuffer, (PyObject*)__pyx_v_right, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 1438, __pyx_L1_error)
  }
  __pyx_pybuffernd_right.diminfo[0].strides = __pyx_pybuffernd_right.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_right.diminfo[0].shape = __pyx_pybuffernd_right.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_indices.rcbuffer->pybuffer, (PyObject*)__pyx_v_indices, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 1438, __pyx_L1_error)
  }
  __pyx_pybuffernd_indices.diminfo[0].strides = __pyx_pybuffernd_indices.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_indices.diminfo[0].shape = __pyx_pybuffernd_indices.rcbuffer->pybuffer.shape[0];

  /* "pandas/_libs/intervaltree.pxi":1444
 *                  int64_t leaf_size):
 * 
 *         self.n_elements = len(left)             # <<<<<<<<<<<<<<
 *         self.leaf_size = leaf_size
 * 
 */
  __pyx_t_1 = PyObject_Length(((PyObject *)__pyx_v_left)); if (unlikely(__pyx_t_1 == ((Py_ssize_t)-1))) __PYX_ERR(1, 1444, __pyx_L1_error)
  __pyx_v_self->n_elements = __pyx_t_1;

  /* "pandas/_libs/intervaltree.pxi":1445
 * 
 *         self.n_elements = len(left)
 *         self.leaf_size = leaf_size             # <<<<<<<<<<<<<<
 * 
 *         # min_left and min_right are used to speed-up query by skipping
 */
  __pyx_v_self->leaf_size = __pyx_v_leaf_size;

  /* "pandas/_libs/intervaltree.pxi":1450
 *         # query on sub-nodes. If this node has size 0, query is cheap,
 *         # so these values don't matter.
 *         if left.size > 0:             # <<<<<<<<<<<<<<
 *             self.min_left = left.min()
 *             self.max_right = right.max()
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_left), __pyx_n_s_size); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1450, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyObject_RichCompare(__pyx_t_2, __pyx_int_0, Py_GT); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1450, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(1, 1450, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_4) {

    /* "pandas/_libs/intervaltree.pxi":1451
 *         # so these values don't matter.
 *         if left.size > 0:
 *             self.min_left = left.min()             # <<<<<<<<<<<<<<
 *             self.max_right = right.max()
 *         else:
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_left), __pyx_n_s_min); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1451, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1451, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_6 = __pyx_PyFloat_AsDouble(__pyx_t_3); if (unlikely((__pyx_t_6 == ((npy_float64)-1)) && PyErr_Occurred())) __PYX_ERR(1, 1451, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_self->min_left = __pyx_t_6;

    /* "pandas/_libs/intervaltree.pxi":1452
 *         if left.size > 0:
 *             self.min_left = left.min()
 *             self.max_right = right.max()             # <<<<<<<<<<<<<<
 *         else:
 *             self.min_left = 0
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_right), __pyx_n_s_max); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1452, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1452, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_6 = __pyx_PyFloat_AsDouble(__pyx_t_3); if (unlikely((__pyx_t_6 == ((npy_float64)-1)) && PyErr_Occurred())) __PYX_ERR(1, 1452, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_self->max_right = __pyx_t_6;

    /* "pandas/_libs/intervaltree.pxi":1450
 *         # query on sub-nodes. If this node has size 0, query is cheap,
 *         # so these values don't matter.
 *         if left.size > 0:             # <<<<<<<<<<<<<<
 *             self.min_left = left.min()
 *             self.max_right = right.max()
 */
    goto __pyx_L3;
  }

  /* "pandas/_libs/intervaltree.pxi":1454
 *             self.max_right = right.max()
 *         else:
 *             self.min_left = 0             # <<<<<<<<<<<<<<
 *             self.max_right = 0
 * 
 */
  /*else*/ {
    __pyx_v_self->min_left = 0.0;

    /* "pandas/_libs/intervaltree.pxi":1455
 *         else:
 *             self.min_left = 0
 *             self.max_right = 0             # <<<<<<<<<<<<<<
 * 
 *         if self.n_elements <= leaf_size:
 */
    __pyx_v_self->max_right = 0.0;
  }
  __pyx_L3:;

  /* "pandas/_libs/intervaltree.pxi":1457
 *             self.max_right = 0
 * 
 *         if self.n_elements <= leaf_size:             # <<<<<<<<<<<<<<
 *             # make this a terminal (leaf) node
 *             self.is_leaf_node = True
 */
  __pyx_t_4 = ((__pyx_v_self->n_elements <= __pyx_v_leaf_size) != 0);
  if (__pyx_t_4) {

    /* "pandas/_libs/intervaltree.pxi":1459
 *         if self.n_elements <= leaf_size:
 *             # make this a terminal (leaf) node
 *             self.is_leaf_node = True             # <<<<<<<<<<<<<<
 *             self.left = left
 *             self.right = right
 */
    __pyx_v_self->is_leaf_node = 1;

    /* "pandas/_libs/intervaltree.pxi":1460
 *             # make this a terminal (leaf) node
 *             self.is_leaf_node = True
 *             self.left = left             # <<<<<<<<<<<<<<
 *             self.right = right
 *             self.indices = indices
 */
    __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_float64_t(((PyObject *)__pyx_v_left), PyBUF_WRITABLE); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(1, 1460, __pyx_L1_error)
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->left, 0);
    __pyx_v_self->left = __pyx_t_7;
    __pyx_t_7.memview = NULL;
    __pyx_t_7.data = NULL;

    /* "pandas/_libs/intervaltree.pxi":1461
 *             self.is_leaf_node = True
 *             self.left = left
 *             self.right = right             # <<<<<<<<<<<<<<
 *             self.indices = indices
 *             self.n_center = 0
 */
    __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_float64_t(((PyObject *)__pyx_v_right), PyBUF_WRITABLE); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(1, 1461, __pyx_L1_error)
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->right, 0);
    __pyx_v_self->right = __pyx_t_7;
    __pyx_t_7.memview = NULL;
    __pyx_t_7.data = NULL;

    /* "pandas/_libs/intervaltree.pxi":1462
 *             self.left = left
 *             self.right = right
 *             self.indices = indices             # <<<<<<<<<<<<<<
 *             self.n_center = 0
 *         else:
 */
    __pyx_t_8 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(((PyObject *)__pyx_v_indices), PyBUF_WRITABLE); if (unlikely(!__pyx_t_8.memview)) __PYX_ERR(1, 1462, __pyx_L1_error)
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->indices, 0);
    __pyx_v_self->indices = __pyx_t_8;
    __pyx_t_8.memview = NULL;
    __pyx_t_8.data = NULL;

    /* "pandas/_libs/intervaltree.pxi":1463
 *             self.right = right
 *             self.indices = indices
 *             self.n_center = 0             # <<<<<<<<<<<<<<
 *         else:
 *             # calculate a pivot so we can create child nodes
 */
    __pyx_v_self->n_center = 0;

    /* "pandas/_libs/intervaltree.pxi":1457
 *             self.max_right = 0
 * 
 *         if self.n_elements <= leaf_size:             # <<<<<<<<<<<<<<
 *             # make this a terminal (leaf) node
 *             self.is_leaf_node = True
 */
    goto __pyx_L4;
  }

  /* "pandas/_libs/intervaltree.pxi":1466
 *         else:
 *             # calculate a pivot so we can create child nodes
 *             self.is_leaf_node = False             # <<<<<<<<<<<<<<
 *             self.pivot = np.median(left / 2 + right / 2)
 *             left_set, right_set, center_set = self.classify_intervals(
 */
  /*else*/ {
    __pyx_v_self->is_leaf_node = 0;

    /* "pandas/_libs/intervaltree.pxi":1467
 *             # calculate a pivot so we can create child nodes
 *             self.is_leaf_node = False
 *             self.pivot = np.median(left / 2 + right / 2)             # <<<<<<<<<<<<<<
 *             left_set, right_set, center_set = self.classify_intervals(
 *                 left, right)
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1467, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_median); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1467, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyNumber_Divide(((PyObject *)__pyx_v_left), __pyx_int_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1467, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_9 = __Pyx_PyNumber_Divide(((PyObject *)__pyx_v_right), __pyx_int_2); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 1467, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_10 = PyNumber_Add(__pyx_t_2, __pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 1467, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_9 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_3 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_9, __pyx_t_10) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_10);
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1467, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_6 = __pyx_PyFloat_AsDouble(__pyx_t_3); if (unlikely((__pyx_t_6 == ((npy_float64)-1)) && PyErr_Occurred())) __PYX_ERR(1, 1467, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_self->pivot = __pyx_t_6;

    /* "pandas/_libs/intervaltree.pxi":1469
 *             self.pivot = np.median(left / 2 + right / 2)
 *             left_set, right_set, center_set = self.classify_intervals(
 *                 left, right)             # <<<<<<<<<<<<<<
 * 
 *             self.left_node = self.new_child_node(left, right,
 */
    __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_float64_t(((PyObject *)__pyx_v_left), PyBUF_WRITABLE); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(1, 1469, __pyx_L1_error)
    __pyx_t_11 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_float64_t(((PyObject *)__pyx_v_right), PyBUF_WRITABLE); if (unlikely(!__pyx_t_11.memview)) __PYX_ERR(1, 1469, __pyx_L1_error)

    /* "pandas/_libs/intervaltree.pxi":1468
 *             self.is_leaf_node = False
 *             self.pivot = np.median(left / 2 + right / 2)
 *             left_set, right_set, center_set = self.classify_intervals(             # <<<<<<<<<<<<<<
 *                 left, right)
 * 
 */
    __pyx_t_3 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode *)__pyx_v_self->__pyx_vtab)->classify_intervals(__pyx_v_self, __pyx_t_7, __pyx_t_11); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1468, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __PYX_XDEC_MEMVIEW(&__pyx_t_7, 1);
    __pyx_t_7.memview = NULL;
    __pyx_t_7.data = NULL;
    __PYX_XDEC_MEMVIEW(&__pyx_t_11, 1);
    __pyx_t_11.memview = NULL;
    __pyx_t_11.data = NULL;
    if ((likely(PyTuple_CheckExact(__pyx_t_3))) || (PyList_CheckExact(__pyx_t_3))) {
      PyObject* sequence = __pyx_t_3;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 3)) {
        if (size > 3) __Pyx_RaiseTooManyValuesError(3);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(1, 1468, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_5 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_10 = PyTuple_GET_ITEM(sequence, 1); 
        __pyx_t_9 = PyTuple_GET_ITEM(sequence, 2); 
      } else {
        __pyx_t_5 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_10 = PyList_GET_ITEM(sequence, 1); 
        __pyx_t_9 = PyList_GET_ITEM(sequence, 2); 
      }
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_10);
      __Pyx_INCREF(__pyx_t_9);
      #else
      __pyx_t_5 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1468, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_10 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 1468, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_9 = PySequence_ITEM(sequence, 2); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 1468, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      #endif
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_2 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1468, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_12 = Py_TYPE(__pyx_t_2)->tp_iternext;
      index = 0; __pyx_t_5 = __pyx_t_12(__pyx_t_2); if (unlikely(!__pyx_t_5)) goto __pyx_L5_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_5);
      index = 1; __pyx_t_10 = __pyx_t_12(__pyx_t_2); if (unlikely(!__pyx_t_10)) goto __pyx_L5_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_10);
      index = 2; __pyx_t_9 = __pyx_t_12(__pyx_t_2); if (unlikely(!__pyx_t_9)) goto __pyx_L5_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_9);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_12(__pyx_t_2), 3) < 0) __PYX_ERR(1, 1468, __pyx_L1_error)
      __pyx_t_12 = NULL;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      goto __pyx_L6_unpacking_done;
      __pyx_L5_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_12 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(1, 1468, __pyx_L1_error)
      __pyx_L6_unpacking_done:;
    }
    __pyx_v_left_set = __pyx_t_5;
    __pyx_t_5 = 0;
    __pyx_v_right_set = __pyx_t_10;
    __pyx_t_10 = 0;
    __pyx_v_center_set = __pyx_t_9;
    __pyx_t_9 = 0;

    /* "pandas/_libs/intervaltree.pxi":1472
 * 
 *             self.left_node = self.new_child_node(left, right,
 *                                                  indices, left_set)             # <<<<<<<<<<<<<<
 *             self.right_node = self.new_child_node(left, right,
 *                                                   indices, right_set)
 */
    if (!(likely(((__pyx_v_left_set) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_left_set, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(1, 1472, __pyx_L1_error)

    /* "pandas/_libs/intervaltree.pxi":1471
 *                 left, right)
 * 
 *             self.left_node = self.new_child_node(left, right,             # <<<<<<<<<<<<<<
 *                                                  indices, left_set)
 *             self.right_node = self.new_child_node(left, right,
 */
    __pyx_t_3 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode *)__pyx_v_self->__pyx_vtab)->new_child_node(__pyx_v_self, ((PyArrayObject *)__pyx_v_left), ((PyArrayObject *)__pyx_v_right), ((PyArrayObject *)__pyx_v_indices), ((PyArrayObject *)__pyx_v_left_set)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1471, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode))))) __PYX_ERR(1, 1471, __pyx_L1_error)
    __Pyx_GIVEREF(__pyx_t_3);
    __Pyx_GOTREF(__pyx_v_self->left_node);
    __Pyx_DECREF(((PyObject *)__pyx_v_self->left_node));
    __pyx_v_self->left_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode *)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "pandas/_libs/intervaltree.pxi":1474
 *                                                  indices, left_set)
 *             self.right_node = self.new_child_node(left, right,
 *                                                   indices, right_set)             # <<<<<<<<<<<<<<
 * 
 *             self.center_left_values, self.center_left_indices = \
 */
    if (!(likely(((__pyx_v_right_set) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_right_set, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(1, 1474, __pyx_L1_error)

    /* "pandas/_libs/intervaltree.pxi":1473
 *             self.left_node = self.new_child_node(left, right,
 *                                                  indices, left_set)
 *             self.right_node = self.new_child_node(left, right,             # <<<<<<<<<<<<<<
 *                                                   indices, right_set)
 * 
 */
    __pyx_t_3 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode *)__pyx_v_self->__pyx_vtab)->new_child_node(__pyx_v_self, ((PyArrayObject *)__pyx_v_left), ((PyArrayObject *)__pyx_v_right), ((PyArrayObject *)__pyx_v_indices), ((PyArrayObject *)__pyx_v_right_set)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1473, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode))))) __PYX_ERR(1, 1473, __pyx_L1_error)
    __Pyx_GIVEREF(__pyx_t_3);
    __Pyx_GOTREF(__pyx_v_self->right_node);
    __Pyx_DECREF(((PyObject *)__pyx_v_self->right_node));
    __pyx_v_self->right_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode *)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "pandas/_libs/intervaltree.pxi":1477
 * 
 *             self.center_left_values, self.center_left_indices = \
 *                 sort_values_and_indices(left, indices, center_set)             # <<<<<<<<<<<<<<
 *             self.center_right_values, self.center_right_indices = \
 *                 sort_values_and_indices(right, indices, center_set)
 */
    __pyx_t_3 = __pyx_f_6pandas_5_libs_8interval_sort_values_and_indices(((PyObject *)__pyx_v_left), ((PyObject *)__pyx_v_indices), __pyx_v_center_set); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1477, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if ((likely(PyTuple_CheckExact(__pyx_t_3))) || (PyList_CheckExact(__pyx_t_3))) {
      PyObject* sequence = __pyx_t_3;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(1, 1476, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_9 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_10 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_9 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_10 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_9);
      __Pyx_INCREF(__pyx_t_10);
      #else
      __pyx_t_9 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 1476, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_10 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 1476, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      #endif
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_5 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1476, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_12 = Py_TYPE(__pyx_t_5)->tp_iternext;
      index = 0; __pyx_t_9 = __pyx_t_12(__pyx_t_5); if (unlikely(!__pyx_t_9)) goto __pyx_L7_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_9);
      index = 1; __pyx_t_10 = __pyx_t_12(__pyx_t_5); if (unlikely(!__pyx_t_10)) goto __pyx_L7_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_10);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_12(__pyx_t_5), 2) < 0) __PYX_ERR(1, 1476, __pyx_L1_error)
      __pyx_t_12 = NULL;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      goto __pyx_L8_unpacking_done;
      __pyx_L7_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_12 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(1, 1476, __pyx_L1_error)
      __pyx_L8_unpacking_done:;
    }

    /* "pandas/_libs/intervaltree.pxi":1476
 *                                                   indices, right_set)
 * 
 *             self.center_left_values, self.center_left_indices = \             # <<<<<<<<<<<<<<
 *                 sort_values_and_indices(left, indices, center_set)
 *             self.center_right_values, self.center_right_indices = \
 */
    __pyx_t_11 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_float64_t(__pyx_t_9, PyBUF_WRITABLE); if (unlikely(!__pyx_t_11.memview)) __PYX_ERR(1, 1476, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_8 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_t_10, PyBUF_WRITABLE); if (unlikely(!__pyx_t_8.memview)) __PYX_ERR(1, 1476, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->center_left_values, 0);
    __pyx_v_self->center_left_values = __pyx_t_11;
    __pyx_t_11.memview = NULL;
    __pyx_t_11.data = NULL;
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->center_left_indices, 0);
    __pyx_v_self->center_left_indices = __pyx_t_8;
    __pyx_t_8.memview = NULL;
    __pyx_t_8.data = NULL;

    /* "pandas/_libs/intervaltree.pxi":1479
 *                 sort_values_and_indices(left, indices, center_set)
 *             self.center_right_values, self.center_right_indices = \
 *                 sort_values_and_indices(right, indices, center_set)             # <<<<<<<<<<<<<<
 *             self.n_center = len(self.center_left_indices)
 * 
 */
    __pyx_t_3 = __pyx_f_6pandas_5_libs_8interval_sort_values_and_indices(((PyObject *)__pyx_v_right), ((PyObject *)__pyx_v_indices), __pyx_v_center_set); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1479, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if ((likely(PyTuple_CheckExact(__pyx_t_3))) || (PyList_CheckExact(__pyx_t_3))) {
      PyObject* sequence = __pyx_t_3;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(1, 1478, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_10 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_9 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_10 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_9 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_10);
      __Pyx_INCREF(__pyx_t_9);
      #else
      __pyx_t_10 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 1478, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_9 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 1478, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      #endif
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_5 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1478, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_12 = Py_TYPE(__pyx_t_5)->tp_iternext;
      index = 0; __pyx_t_10 = __pyx_t_12(__pyx_t_5); if (unlikely(!__pyx_t_10)) goto __pyx_L9_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_10);
      index = 1; __pyx_t_9 = __pyx_t_12(__pyx_t_5); if (unlikely(!__pyx_t_9)) goto __pyx_L9_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_9);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_12(__pyx_t_5), 2) < 0) __PYX_ERR(1, 1478, __pyx_L1_error)
      __pyx_t_12 = NULL;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      goto __pyx_L10_unpacking_done;
      __pyx_L9_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_12 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(1, 1478, __pyx_L1_error)
      __pyx_L10_unpacking_done:;
    }

    /* "pandas/_libs/intervaltree.pxi":1478
 *             self.center_left_values, self.center_left_indices = \
 *                 sort_values_and_indices(left, indices, center_set)
 *             self.center_right_values, self.center_right_indices = \             # <<<<<<<<<<<<<<
 *                 sort_values_and_indices(right, indices, center_set)
 *             self.n_center = len(self.center_left_indices)
 */
    __pyx_t_11 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_float64_t(__pyx_t_10, PyBUF_WRITABLE); if (unlikely(!__pyx_t_11.memview)) __PYX_ERR(1, 1478, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_t_8 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_t_9, PyBUF_WRITABLE); if (unlikely(!__pyx_t_8.memview)) __PYX_ERR(1, 1478, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->center_right_values, 0);
    __pyx_v_self->center_right_values = __pyx_t_11;
    __pyx_t_11.memview = NULL;
    __pyx_t_11.data = NULL;
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->center_right_indices, 0);
    __pyx_v_self->center_right_indices = __pyx_t_8;
    __pyx_t_8.memview = NULL;
    __pyx_t_8.data = NULL;

    /* "pandas/_libs/intervaltree.pxi":1480
 *             self.center_right_values, self.center_right_indices = \
 *                 sort_values_and_indices(right, indices, center_set)
 *             self.n_center = len(self.center_left_indices)             # <<<<<<<<<<<<<<
 * 
 *     @cython.wraparound(False)
 */
    if (unlikely(!__pyx_v_self->center_left_indices.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(1, 1480, __pyx_L1_error)}
    __pyx_t_13 = __Pyx_MemoryView_Len(__pyx_v_self->center_left_indices); 
    __pyx_v_self->n_center = __pyx_t_13;
  }
  __pyx_L4:;

  /* "pandas/_libs/intervaltree.pxi":1438
 *         readonly bint is_leaf_node
 * 
 *     def __init__(self,             # <<<<<<<<<<<<<<
 *                  ndarray[float64_t, ndim=1] left,
 *                  ndarray[float64_t, ndim=1] right,
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __PYX_XDEC_MEMVIEW(&__pyx_t_7, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_8, 1);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __PYX_XDEC_MEMVIEW(&__pyx_t_11, 1);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indices.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_left.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_right.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("pandas._libs.interval.Float64ClosedNeitherIntervalNode.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indices.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_left.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_right.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF(__pyx_v_left_set);
  __Pyx_XDECREF(__pyx_v_right_set);
  __Pyx_XDECREF(__pyx_v_center_set);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":1484
 *     @cython.wraparound(False)
 *     @cython.boundscheck(False)
 *     cdef classify_intervals(self, float64_t[:] left, float64_t[:] right):             # <<<<<<<<<<<<<<
 *         """Classify the given intervals based upon whether they fall to the
 *         left, right, or overlap with this node's pivot.
 */

static PyObject *__pyx_f_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_classify_intervals(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode *__pyx_v_self, __Pyx_memviewslice __pyx_v_left, __Pyx_memviewslice __pyx_v_right) {
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_left_ind = 0;
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_right_ind = 0;
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_overlapping_ind = 0;
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __pyx_t_5numpy_int64_t __pyx_t_2;
  __pyx_t_5numpy_int64_t __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  int __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  __Pyx_RefNannySetupContext("classify_intervals", 0);

  /* "pandas/_libs/intervaltree.pxi":1492
 *             Py_ssize_t i
 * 
 *         left_ind = Int64Vector()             # <<<<<<<<<<<<<<
 *         right_ind = Int64Vector()
 *         overlapping_ind = Int64Vector()
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1492, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_left_ind = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":1493
 * 
 *         left_ind = Int64Vector()
 *         right_ind = Int64Vector()             # <<<<<<<<<<<<<<
 *         overlapping_ind = Int64Vector()
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1493, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_right_ind = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":1494
 *         left_ind = Int64Vector()
 *         right_ind = Int64Vector()
 *         overlapping_ind = Int64Vector()             # <<<<<<<<<<<<<<
 * 
 *         for i in range(self.n_elements):
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1494, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_overlapping_ind = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":1496
 *         overlapping_ind = Int64Vector()
 * 
 *         for i in range(self.n_elements):             # <<<<<<<<<<<<<<
 *             if right[i] <= self.pivot:
 *                 left_ind.append(i)
 */
  __pyx_t_2 = __pyx_v_self->n_elements;
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "pandas/_libs/intervaltree.pxi":1497
 * 
 *         for i in range(self.n_elements):
 *             if right[i] <= self.pivot:             # <<<<<<<<<<<<<<
 *                 left_ind.append(i)
 *             elif self.pivot <= left[i]:
 */
    __pyx_t_5 = __pyx_v_i;
    __pyx_t_6 = (((*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_right.data + __pyx_t_5 * __pyx_v_right.strides[0]) ))) <= __pyx_v_self->pivot) != 0);
    if (__pyx_t_6) {

      /* "pandas/_libs/intervaltree.pxi":1498
 *         for i in range(self.n_elements):
 *             if right[i] <= self.pivot:
 *                 left_ind.append(i)             # <<<<<<<<<<<<<<
 *             elif self.pivot <= left[i]:
 *                 right_ind.append(i)
 */
      ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_left_ind->__pyx_vtab)->append(__pyx_v_left_ind, __pyx_v_i);

      /* "pandas/_libs/intervaltree.pxi":1497
 * 
 *         for i in range(self.n_elements):
 *             if right[i] <= self.pivot:             # <<<<<<<<<<<<<<
 *                 left_ind.append(i)
 *             elif self.pivot <= left[i]:
 */
      goto __pyx_L5;
    }

    /* "pandas/_libs/intervaltree.pxi":1499
 *             if right[i] <= self.pivot:
 *                 left_ind.append(i)
 *             elif self.pivot <= left[i]:             # <<<<<<<<<<<<<<
 *                 right_ind.append(i)
 *             else:
 */
    __pyx_t_7 = __pyx_v_i;
    __pyx_t_6 = ((__pyx_v_self->pivot <= (*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_left.data + __pyx_t_7 * __pyx_v_left.strides[0]) )))) != 0);
    if (__pyx_t_6) {

      /* "pandas/_libs/intervaltree.pxi":1500
 *                 left_ind.append(i)
 *             elif self.pivot <= left[i]:
 *                 right_ind.append(i)             # <<<<<<<<<<<<<<
 *             else:
 *                 overlapping_ind.append(i)
 */
      ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_right_ind->__pyx_vtab)->append(__pyx_v_right_ind, __pyx_v_i);

      /* "pandas/_libs/intervaltree.pxi":1499
 *             if right[i] <= self.pivot:
 *                 left_ind.append(i)
 *             elif self.pivot <= left[i]:             # <<<<<<<<<<<<<<
 *                 right_ind.append(i)
 *             else:
 */
      goto __pyx_L5;
    }

    /* "pandas/_libs/intervaltree.pxi":1502
 *                 right_ind.append(i)
 *             else:
 *                 overlapping_ind.append(i)             # <<<<<<<<<<<<<<
 * 
 *         return (left_ind.to_array(),
 */
    /*else*/ {
      ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_overlapping_ind->__pyx_vtab)->append(__pyx_v_overlapping_ind, __pyx_v_i);
    }
    __pyx_L5:;
  }

  /* "pandas/_libs/intervaltree.pxi":1504
 *                 overlapping_ind.append(i)
 * 
 *         return (left_ind.to_array(),             # <<<<<<<<<<<<<<
 *                 right_ind.to_array(),
 *                 overlapping_ind.to_array())
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_left_ind->__pyx_vtab)->to_array(__pyx_v_left_ind, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1504, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "pandas/_libs/intervaltree.pxi":1505
 * 
 *         return (left_ind.to_array(),
 *                 right_ind.to_array(),             # <<<<<<<<<<<<<<
 *                 overlapping_ind.to_array())
 * 
 */
  __pyx_t_8 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_right_ind->__pyx_vtab)->to_array(__pyx_v_right_ind, 0); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 1505, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);

  /* "pandas/_libs/intervaltree.pxi":1506
 *         return (left_ind.to_array(),
 *                 right_ind.to_array(),
 *                 overlapping_ind.to_array())             # <<<<<<<<<<<<<<
 * 
 *     cdef new_child_node(self,
 */
  __pyx_t_9 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_overlapping_ind->__pyx_vtab)->to_array(__pyx_v_overlapping_ind, 0); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 1506, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);

  /* "pandas/_libs/intervaltree.pxi":1504
 *                 overlapping_ind.append(i)
 * 
 *         return (left_ind.to_array(),             # <<<<<<<<<<<<<<
 *                 right_ind.to_array(),
 *                 overlapping_ind.to_array())
 */
  __pyx_t_10 = PyTuple_New(3); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 1504, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_t_8);
  __Pyx_GIVEREF(__pyx_t_9);
  PyTuple_SET_ITEM(__pyx_t_10, 2, __pyx_t_9);
  __pyx_t_1 = 0;
  __pyx_t_8 = 0;
  __pyx_t_9 = 0;
  __pyx_r = __pyx_t_10;
  __pyx_t_10 = 0;
  goto __pyx_L0;

  /* "pandas/_libs/intervaltree.pxi":1484
 *     @cython.wraparound(False)
 *     @cython.boundscheck(False)
 *     cdef classify_intervals(self, float64_t[:] left, float64_t[:] right):             # <<<<<<<<<<<<<<
 *         """Classify the given intervals based upon whether they fall to the
 *         left, right, or overlap with this node's pivot.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("pandas._libs.interval.Float64ClosedNeitherIntervalNode.classify_intervals", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_left_ind);
  __Pyx_XDECREF((PyObject *)__pyx_v_right_ind);
  __Pyx_XDECREF((PyObject *)__pyx_v_overlapping_ind);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":1508
 *                 overlapping_ind.to_array())
 * 
 *     cdef new_child_node(self,             # <<<<<<<<<<<<<<
 *                         ndarray[float64_t, ndim=1] left,
 *                         ndarray[float64_t, ndim=1] right,
 */

static PyObject *__pyx_f_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_new_child_node(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode *__pyx_v_self, PyArrayObject *__pyx_v_left, PyArrayObject *__pyx_v_right, PyArrayObject *__pyx_v_indices, PyArrayObject *__pyx_v_subset) {
  __Pyx_LocalBuf_ND __pyx_pybuffernd_indices;
  __Pyx_Buffer __pyx_pybuffer_indices;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_left;
  __Pyx_Buffer __pyx_pybuffer_left;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_right;
  __Pyx_Buffer __pyx_pybuffer_right;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_subset;
  __Pyx_Buffer __pyx_pybuffer_subset;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyArrayObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyArrayObject *__pyx_t_7 = NULL;
  PyArrayObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  __Pyx_RefNannySetupContext("new_child_node", 0);
  __Pyx_INCREF((PyObject *)__pyx_v_left);
  __Pyx_INCREF((PyObject *)__pyx_v_right);
  __Pyx_INCREF((PyObject *)__pyx_v_indices);
  __pyx_pybuffer_left.pybuffer.buf = NULL;
  __pyx_pybuffer_left.refcount = 0;
  __pyx_pybuffernd_left.data = NULL;
  __pyx_pybuffernd_left.rcbuffer = &__pyx_pybuffer_left;
  __pyx_pybuffer_right.pybuffer.buf = NULL;
  __pyx_pybuffer_right.refcount = 0;
  __pyx_pybuffernd_right.data = NULL;
  __pyx_pybuffernd_right.rcbuffer = &__pyx_pybuffer_right;
  __pyx_pybuffer_indices.pybuffer.buf = NULL;
  __pyx_pybuffer_indices.refcount = 0;
  __pyx_pybuffernd_indices.data = NULL;
  __pyx_pybuffernd_indices.rcbuffer = &__pyx_pybuffer_indices;
  __pyx_pybuffer_subset.pybuffer.buf = NULL;
  __pyx_pybuffer_subset.refcount = 0;
  __pyx_pybuffernd_subset.data = NULL;
  __pyx_pybuffernd_subset.rcbuffer = &__pyx_pybuffer_subset;
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_left.rcbuffer->pybuffer, (PyObject*)__pyx_v_left, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 1508, __pyx_L1_error)
  }
  __pyx_pybuffernd_left.diminfo[0].strides = __pyx_pybuffernd_left.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_left.diminfo[0].shape = __pyx_pybuffernd_left.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_right.rcbuffer->pybuffer, (PyObject*)__pyx_v_right, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 1508, __pyx_L1_error)
  }
  __pyx_pybuffernd_right.diminfo[0].strides = __pyx_pybuffernd_right.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_right.diminfo[0].shape = __pyx_pybuffernd_right.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_indices.rcbuffer->pybuffer, (PyObject*)__pyx_v_indices, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 1508, __pyx_L1_error)
  }
  __pyx_pybuffernd_indices.diminfo[0].strides = __pyx_pybuffernd_indices.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_indices.diminfo[0].shape = __pyx_pybuffernd_indices.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_subset.rcbuffer->pybuffer, (PyObject*)__pyx_v_subset, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 1508, __pyx_L1_error)
  }
  __pyx_pybuffernd_subset.diminfo[0].strides = __pyx_pybuffernd_subset.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_subset.diminfo[0].shape = __pyx_pybuffernd_subset.rcbuffer->pybuffer.shape[0];

  /* "pandas/_libs/intervaltree.pxi":1515
 *         """Create a new child node.
 *         """
 *         left = take(left, subset)             # <<<<<<<<<<<<<<
 *         right = take(right, subset)
 *         indices = take(indices, subset)
 */
  __pyx_t_1 = __pyx_f_6pandas_5_libs_8interval_take(((PyArrayObject *)__pyx_v_left), ((PyArrayObject *)__pyx_v_subset)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1515, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(1, 1515, __pyx_L1_error)
  __pyx_t_2 = ((PyArrayObject *)__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_left.rcbuffer->pybuffer);
    __pyx_t_3 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_left.rcbuffer->pybuffer, (PyObject*)__pyx_t_2, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack);
    if (unlikely(__pyx_t_3 < 0)) {
      PyErr_Fetch(&__pyx_t_4, &__pyx_t_5, &__pyx_t_6);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_left.rcbuffer->pybuffer, (PyObject*)__pyx_v_left, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_4); Py_XDECREF(__pyx_t_5); Py_XDECREF(__pyx_t_6);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_4, __pyx_t_5, __pyx_t_6);
      }
      __pyx_t_4 = __pyx_t_5 = __pyx_t_6 = 0;
    }
    __pyx_pybuffernd_left.diminfo[0].strides = __pyx_pybuffernd_left.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_left.diminfo[0].shape = __pyx_pybuffernd_left.rcbuffer->pybuffer.shape[0];
    if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(1, 1515, __pyx_L1_error)
  }
  __pyx_t_2 = 0;
  __Pyx_DECREF_SET(__pyx_v_left, ((PyArrayObject *)__pyx_t_1));
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":1516
 *         """
 *         left = take(left, subset)
 *         right = take(right, subset)             # <<<<<<<<<<<<<<
 *         indices = take(indices, subset)
 *         return Float64ClosedNeitherIntervalNode(
 */
  __pyx_t_1 = __pyx_f_6pandas_5_libs_8interval_take(((PyArrayObject *)__pyx_v_right), ((PyArrayObject *)__pyx_v_subset)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1516, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(1, 1516, __pyx_L1_error)
  __pyx_t_7 = ((PyArrayObject *)__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_right.rcbuffer->pybuffer);
    __pyx_t_3 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_right.rcbuffer->pybuffer, (PyObject*)__pyx_t_7, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack);
    if (unlikely(__pyx_t_3 < 0)) {
      PyErr_Fetch(&__pyx_t_6, &__pyx_t_5, &__pyx_t_4);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_right.rcbuffer->pybuffer, (PyObject*)__pyx_v_right, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_6); Py_XDECREF(__pyx_t_5); Py_XDECREF(__pyx_t_4);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_6, __pyx_t_5, __pyx_t_4);
      }
      __pyx_t_6 = __pyx_t_5 = __pyx_t_4 = 0;
    }
    __pyx_pybuffernd_right.diminfo[0].strides = __pyx_pybuffernd_right.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_right.diminfo[0].shape = __pyx_pybuffernd_right.rcbuffer->pybuffer.shape[0];
    if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(1, 1516, __pyx_L1_error)
  }
  __pyx_t_7 = 0;
  __Pyx_DECREF_SET(__pyx_v_right, ((PyArrayObject *)__pyx_t_1));
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":1517
 *         left = take(left, subset)
 *         right = take(right, subset)
 *         indices = take(indices, subset)             # <<<<<<<<<<<<<<
 *         return Float64ClosedNeitherIntervalNode(
 *             left, right, indices, self.leaf_size)
 */
  __pyx_t_1 = __pyx_f_6pandas_5_libs_8interval_take(((PyArrayObject *)__pyx_v_indices), ((PyArrayObject *)__pyx_v_subset)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1517, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(1, 1517, __pyx_L1_error)
  __pyx_t_8 = ((PyArrayObject *)__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indices.rcbuffer->pybuffer);
    __pyx_t_3 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_indices.rcbuffer->pybuffer, (PyObject*)__pyx_t_8, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack);
    if (unlikely(__pyx_t_3 < 0)) {
      PyErr_Fetch(&__pyx_t_4, &__pyx_t_5, &__pyx_t_6);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_indices.rcbuffer->pybuffer, (PyObject*)__pyx_v_indices, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_4); Py_XDECREF(__pyx_t_5); Py_XDECREF(__pyx_t_6);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_4, __pyx_t_5, __pyx_t_6);
      }
      __pyx_t_4 = __pyx_t_5 = __pyx_t_6 = 0;
    }
    __pyx_pybuffernd_indices.diminfo[0].strides = __pyx_pybuffernd_indices.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_indices.diminfo[0].shape = __pyx_pybuffernd_indices.rcbuffer->pybuffer.shape[0];
    if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(1, 1517, __pyx_L1_error)
  }
  __pyx_t_8 = 0;
  __Pyx_DECREF_SET(__pyx_v_indices, ((PyArrayObject *)__pyx_t_1));
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":1518
 *         right = take(right, subset)
 *         indices = take(indices, subset)
 *         return Float64ClosedNeitherIntervalNode(             # <<<<<<<<<<<<<<
 *             left, right, indices, self.leaf_size)
 * 
 */
  __Pyx_XDECREF(__pyx_r);

  /* "pandas/_libs/intervaltree.pxi":1519
 *         indices = take(indices, subset)
 *         return Float64ClosedNeitherIntervalNode(
 *             left, right, indices, self.leaf_size)             # <<<<<<<<<<<<<<
 * 
 *     @cython.wraparound(False)
 */
  __pyx_t_1 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->leaf_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1519, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "pandas/_libs/intervaltree.pxi":1518
 *         right = take(right, subset)
 *         indices = take(indices, subset)
 *         return Float64ClosedNeitherIntervalNode(             # <<<<<<<<<<<<<<
 *             left, right, indices, self.leaf_size)
 * 
 */
  __pyx_t_9 = PyTuple_New(4); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 1518, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_INCREF(((PyObject *)__pyx_v_left));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_left));
  PyTuple_SET_ITEM(__pyx_t_9, 0, ((PyObject *)__pyx_v_left));
  __Pyx_INCREF(((PyObject *)__pyx_v_right));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_right));
  PyTuple_SET_ITEM(__pyx_t_9, 1, ((PyObject *)__pyx_v_right));
  __Pyx_INCREF(((PyObject *)__pyx_v_indices));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_indices));
  PyTuple_SET_ITEM(__pyx_t_9, 2, ((PyObject *)__pyx_v_indices));
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_9, 3, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode), __pyx_t_9, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1518, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pandas/_libs/intervaltree.pxi":1508
 *                 overlapping_ind.to_array())
 * 
 *     cdef new_child_node(self,             # <<<<<<<<<<<<<<
 *                         ndarray[float64_t, ndim=1] left,
 *                         ndarray[float64_t, ndim=1] right,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_9);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indices.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_left.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_right.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_subset.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("pandas._libs.interval.Float64ClosedNeitherIntervalNode.new_child_node", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indices.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_left.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_right.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_subset.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF((PyObject *)__pyx_v_left);
  __Pyx_XDECREF((PyObject *)__pyx_v_right);
  __Pyx_XDECREF((PyObject *)__pyx_v_indices);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":1524
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_2query[] = "Recursively query this node and its sub-nodes for intervals that\n        overlap with the query point.\n        ";
static PyMethodDef __pyx_mdef_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_3query = {"query", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_3query, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_2query};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_args = 0;
  PyObject *__pyx_v_kwargs = 0;
  CYTHON_UNUSED PyObject *__pyx_v_defaults = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fused_cpdef (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_args,&__pyx_n_s_kwargs,&__pyx_n_s_defaults,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_args)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_kwargs)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fused_cpdef", 1, 3, 3, 1); __PYX_ERR(1, 1524, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_defaults)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fused_cpdef", 1, 3, 3, 2); __PYX_ERR(1, 1524, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fused_cpdef") < 0)) __PYX_ERR(1, 1524, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_args = values[0];
    __pyx_v_kwargs = values[1];
    __pyx_v_defaults = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fused_cpdef", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 1524, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Float64ClosedNeitherIntervalNode.__pyx_fused_cpdef", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_2query(((struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode *)__pyx_v_signatures), __pyx_v_args, __pyx_v_kwargs, __pyx_v_defaults);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_2query(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode *__pyx_v_signatures, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs, CYTHON_UNUSED PyObject *__pyx_v_defaults) {
  PyObject *__pyx_v_dest_sig = NULL;
  Py_ssize_t __pyx_v_i;
  CYTHON_UNUSED PyTypeObject *__pyx_v_ndarray = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_v_itemsize;
  PyObject *__pyx_v_arg = NULL;
  PyObject *__pyx_v_candidates = NULL;
  PyObject *__pyx_v_sig = NULL;
  int __pyx_v_match_found;
  PyObject *__pyx_v_src_sig = NULL;
  PyObject *__pyx_v_dst_type = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  Py_ssize_t __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  Py_ssize_t __pyx_t_13;
  Py_ssize_t __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  int __pyx_t_16;
  __Pyx_RefNannySetupContext("query", 0);
  __Pyx_INCREF(__pyx_v_kwargs);
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1524, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  PyList_SET_ITEM(__pyx_t_1, 0, Py_None);
  __pyx_v_dest_sig = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_3 = (__pyx_v_kwargs != Py_None);
  __pyx_t_4 = (__pyx_t_3 != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_kwargs); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(1, 1524, __pyx_L1_error)
  __pyx_t_3 = ((!__pyx_t_4) != 0);
  __pyx_t_2 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {
    __Pyx_INCREF(Py_None);
    __Pyx_DECREF_SET(__pyx_v_kwargs, Py_None);
  }
  __pyx_t_1 = ((PyObject *)__Pyx_ImportNumPyArrayTypeIfAvailable()); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1524, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_ndarray = ((PyTypeObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_v_itemsize = -1L;
  if (unlikely(__pyx_v_args == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 1524, __pyx_L1_error)
  }
  __pyx_t_5 = PyTuple_GET_SIZE(((PyObject*)__pyx_v_args)); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(1, 1524, __pyx_L1_error)
  __pyx_t_2 = ((2 < __pyx_t_5) != 0);
  if (__pyx_t_2) {
    if (unlikely(__pyx_v_args == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 1524, __pyx_L1_error)
    }
    __pyx_t_1 = PyTuple_GET_ITEM(((PyObject*)__pyx_v_args), 2);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v_arg = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L6;
  }
  __pyx_t_3 = (__pyx_v_kwargs != Py_None);
  __pyx_t_4 = (__pyx_t_3 != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L7_bool_binop_done;
  }
  if (unlikely(__pyx_v_kwargs == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(1, 1524, __pyx_L1_error)
  }
  __pyx_t_4 = (__Pyx_PyDict_ContainsTF(__pyx_n_s_point, ((PyObject*)__pyx_v_kwargs), Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(1, 1524, __pyx_L1_error)
  __pyx_t_3 = (__pyx_t_4 != 0);
  __pyx_t_2 = __pyx_t_3;
  __pyx_L7_bool_binop_done:;
  if (__pyx_t_2) {
    if (unlikely(__pyx_v_kwargs == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 1524, __pyx_L1_error)
    }
    __pyx_t_1 = __Pyx_PyDict_GetItem(((PyObject*)__pyx_v_kwargs), __pyx_n_s_point); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1524, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_arg = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L6;
  }
  /*else*/ {
    if (unlikely(__pyx_v_args == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
      __PYX_ERR(1, 1524, __pyx_L1_error)
    }
    __pyx_t_5 = PyTuple_GET_SIZE(((PyObject*)__pyx_v_args)); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(1, 1524, __pyx_L1_error)
    __pyx_t_1 = PyInt_FromSsize_t(__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1524, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_6 = PyTuple_New(3); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 1524, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_INCREF(__pyx_int_3);
    __Pyx_GIVEREF(__pyx_int_3);
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_int_3);
    __Pyx_INCREF(__pyx_n_s_s);
    __Pyx_GIVEREF(__pyx_n_s_s);
    PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_n_s_s);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Expected_at_least_d_argument_s_g, __pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1524, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_1); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 1524, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(1, 1524, __pyx_L1_error)
  }
  __pyx_L6:;
  while (1) {
    __pyx_t_2 = PyFloat_Check(__pyx_v_arg); 
    __pyx_t_3 = (__pyx_t_2 != 0);
    if (__pyx_t_3) {
      if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, __pyx_n_s_float64_t, long, 1, __Pyx_PyInt_From_long, 1, 0, 0) < 0)) __PYX_ERR(1, 1524, __pyx_L1_error)
      goto __pyx_L10_break;
    }
    __pyx_t_2 = PyInt_Check(__pyx_v_arg); 
    __pyx_t_4 = (__pyx_t_2 != 0);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_3 = __pyx_t_4;
      goto __pyx_L13_bool_binop_done;
    }
    __pyx_t_4 = PyLong_Check(__pyx_v_arg); 
    __pyx_t_2 = (__pyx_t_4 != 0);
    __pyx_t_3 = __pyx_t_2;
    __pyx_L13_bool_binop_done:;
    __pyx_t_2 = (__pyx_t_3 != 0);
    if (__pyx_t_2) {
      if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, __pyx_n_s_int64_t, long, 1, __Pyx_PyInt_From_long, 1, 0, 0) < 0)) __PYX_ERR(1, 1524, __pyx_L1_error)
      goto __pyx_L10_break;
    }
    if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, Py_None, long, 1, __Pyx_PyInt_From_long, 1, 0, 0) < 0)) __PYX_ERR(1, 1524, __pyx_L1_error)
    goto __pyx_L10_break;
  }
  __pyx_L10_break:;
  __pyx_t_6 = PyList_New(0); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 1524, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_v_candidates = ((PyObject*)__pyx_t_6);
  __pyx_t_6 = 0;
  __pyx_t_5 = 0;
  if (unlikely(((PyObject *)__pyx_v_signatures) == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(1, 1524, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_dict_iterator(((PyObject*)__pyx_v_signatures), 1, ((PyObject *)NULL), (&__pyx_t_7), (&__pyx_t_8)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1524, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __pyx_t_6 = __pyx_t_1;
  __pyx_t_1 = 0;
  while (1) {
    __pyx_t_9 = __Pyx_dict_iter_next(__pyx_t_6, __pyx_t_7, &__pyx_t_5, &__pyx_t_1, NULL, NULL, __pyx_t_8);
    if (unlikely(__pyx_t_9 == 0)) break;
    if (unlikely(__pyx_t_9 == -1)) __PYX_ERR(1, 1524, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_sig, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_v_match_found = 0;
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_v_sig, __pyx_n_s_strip); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 1524, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __pyx_t_12 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_11))) {
      __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_11);
      if (likely(__pyx_t_12)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
        __Pyx_INCREF(__pyx_t_12);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_11, function);
      }
    }
    __pyx_t_10 = (__pyx_t_12) ? __Pyx_PyObject_Call2Args(__pyx_t_11, __pyx_t_12, __pyx_kp_s__13) : __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_kp_s__13);
    __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
    if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 1524, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_split); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 1524, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_t_10 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_11))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_11);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_11, function);
      }
    }
    __pyx_t_1 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_11, __pyx_t_10, __pyx_kp_s__14) : __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_kp_s__14);
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1524, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_XDECREF_SET(__pyx_v_src_sig, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_13 = PyList_GET_SIZE(__pyx_v_dest_sig); if (unlikely(__pyx_t_13 == ((Py_ssize_t)-1))) __PYX_ERR(1, 1524, __pyx_L1_error)
    __pyx_t_14 = __pyx_t_13;
    for (__pyx_t_15 = 0; __pyx_t_15 < __pyx_t_14; __pyx_t_15+=1) {
      __pyx_v_i = __pyx_t_15;
      __pyx_t_1 = PyList_GET_ITEM(__pyx_v_dest_sig, __pyx_v_i);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_XDECREF_SET(__pyx_v_dst_type, __pyx_t_1);
      __pyx_t_1 = 0;
      __pyx_t_2 = (__pyx_v_dst_type != Py_None);
      __pyx_t_3 = (__pyx_t_2 != 0);
      if (__pyx_t_3) {
        __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_src_sig, __pyx_v_i, Py_ssize_t, 1, PyInt_FromSsize_t, 0, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1524, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_11 = PyObject_RichCompare(__pyx_t_1, __pyx_v_dst_type, Py_EQ); __Pyx_XGOTREF(__pyx_t_11); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 1524, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_11); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(1, 1524, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        if (__pyx_t_3) {
          __pyx_v_match_found = 1;
          goto __pyx_L20;
        }
        /*else*/ {
          __pyx_v_match_found = 0;
          goto __pyx_L18_break;
        }
        __pyx_L20:;
      }
    }
    __pyx_L18_break:;
    __pyx_t_3 = (__pyx_v_match_found != 0);
    if (__pyx_t_3) {
      __pyx_t_16 = __Pyx_PyList_Append(__pyx_v_candidates, __pyx_v_sig); if (unlikely(__pyx_t_16 == ((int)-1))) __PYX_ERR(1, 1524, __pyx_L1_error)
    }
  }
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_3 = (PyList_GET_SIZE(__pyx_v_candidates) != 0);
  __pyx_t_2 = ((!__pyx_t_3) != 0);
  if (__pyx_t_2) {
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__15, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 1524, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(1, 1524, __pyx_L1_error)
  }
  __pyx_t_7 = PyList_GET_SIZE(__pyx_v_candidates); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1))) __PYX_ERR(1, 1524, __pyx_L1_error)
  __pyx_t_2 = ((__pyx_t_7 > 1) != 0);
  if (__pyx_t_2) {
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__16, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 1524, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(1, 1524, __pyx_L1_error)
  }
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    if (unlikely(((PyObject *)__pyx_v_signatures) == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 1524, __pyx_L1_error)
    }
    __pyx_t_6 = __Pyx_PyDict_GetItem(((PyObject*)__pyx_v_signatures), PyList_GET_ITEM(__pyx_v_candidates, 0)); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 1524, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_r = __pyx_t_6;
    __pyx_t_6 = 0;
    goto __pyx_L0;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("pandas._libs.interval.Float64ClosedNeitherIntervalNode.__pyx_fused_cpdef", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_dest_sig);
  __Pyx_XDECREF(__pyx_v_ndarray);
  __Pyx_XDECREF(__pyx_v_arg);
  __Pyx_XDECREF(__pyx_v_candidates);
  __Pyx_XDECREF(__pyx_v_sig);
  __Pyx_XDECREF(__pyx_v_src_sig);
  __Pyx_XDECREF(__pyx_v_dst_type);
  __Pyx_XDECREF(__pyx_v_kwargs);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pw_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_9__pyx_fuse_0query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_fuse_0__pyx_f_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float64_t __pyx_v_point, int __pyx_skip_dispatch) {
  __Pyx_memviewslice __pyx_v_indices = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_values = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __pyx_t_5numpy_int64_t __pyx_t_9;
  __pyx_t_5numpy_int64_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  int __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  __Pyx_memviewslice __pyx_t_16 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_17 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  Py_ssize_t __pyx_t_21;
  __Pyx_RefNannySetupContext("__pyx_fuse_0query", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pyx_fuse_0query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1524, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_9__pyx_fuse_0query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = PyFloat_FromDouble(__pyx_v_point); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1524, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1524, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1524, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 1524, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_5) {
            __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1524, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1524, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_3query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_4 = PyFloat_FromDouble(__pyx_v_point); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1524, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_7 = __pyx_t_1; __pyx_t_3 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1524, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1524, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1524, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (__pyx_t_3) {
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_t_4);
          __pyx_t_4 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1524, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pandas/_libs/intervaltree.pxi":1533
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
  __pyx_t_8 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_8) {

    /* "pandas/_libs/intervaltree.pxi":1537
 *             # continue the binary tree structure. Instead, we use linear
 *             # search.
 *             for i in range(self.n_elements):             # <<<<<<<<<<<<<<
 *                 if self.left[i] < point < self.right[i]:
 *                     result.append(self.indices[i])
 */
    __pyx_t_9 = __pyx_v_self->n_elements;
    __pyx_t_10 = __pyx_t_9;
    for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
      __pyx_v_i = __pyx_t_11;

      /* "pandas/_libs/intervaltree.pxi":1538
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point < self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      __pyx_t_12 = __pyx_v_i;
      __pyx_t_8 = ((*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_self->left.data + __pyx_t_12 * __pyx_v_self->left.strides[0]) ))) < __pyx_v_point);
      if (__pyx_t_8) {
        __pyx_t_13 = __pyx_v_i;
        __pyx_t_8 = (__pyx_v_point < (*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_self->right.data + __pyx_t_13 * __pyx_v_self->right.strides[0]) ))));
      }
      __pyx_t_14 = (__pyx_t_8 != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":1539
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point < self.right[i]:
 *                     result.append(self.indices[i])             # <<<<<<<<<<<<<<
 *         else:
 *             # There are child nodes. Based on comparing our query to the pivot,
 */
        __pyx_t_15 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->indices.data + __pyx_t_15 * __pyx_v_self->indices.strides[0]) ))));

        /* "pandas/_libs/intervaltree.pxi":1538
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point < self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      }
    }

    /* "pandas/_libs/intervaltree.pxi":1533
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
    goto __pyx_L3;
  }

  /* "pandas/_libs/intervaltree.pxi":1543
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
  /*else*/ {
    __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":1544
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:
 *                 values = self.center_left_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 */
      __pyx_t_16 = __pyx_v_self->center_left_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":1545
 *             if point < self.pivot:
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:
 */
      __pyx_t_17 = __pyx_v_self->center_left_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":1546
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):             # <<<<<<<<<<<<<<
 *                     if not values[i] < point:
 *                         break
 */
      __pyx_t_9 = __pyx_v_self->n_center;
      __pyx_t_10 = __pyx_t_9;
      for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":1547
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_18 = __pyx_v_i;
        __pyx_t_14 = ((!(((*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_18 * __pyx_v_values.strides[0]) ))) < __pyx_v_point) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":1548
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:
 */
          goto __pyx_L9_break;

          /* "pandas/_libs/intervaltree.pxi":1547
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":1549
 *                     if not values[i] < point:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 */
        __pyx_t_19 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_19 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L9_break:;

      /* "pandas/_libs/intervaltree.pxi":1550
 *                         break
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->left_node->max_right) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":1551
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)             # <<<<<<<<<<<<<<
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode *)__pyx_v_self->left_node->__pyx_vtab)->__pyx_fuse_0query(__pyx_v_self->left_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1551, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":1550
 *                         break
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":1543
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":1552
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
    __pyx_t_14 = ((__pyx_v_point > __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":1553
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 *                 values = self.center_right_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 */
      __pyx_t_16 = __pyx_v_self->center_right_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":1554
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:
 */
      __pyx_t_17 = __pyx_v_self->center_right_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":1555
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):             # <<<<<<<<<<<<<<
 *                     if not point < values[i]:
 *                         break
 */
      for (__pyx_t_11 = (__pyx_v_self->n_center - 1); __pyx_t_11 > -1L; __pyx_t_11-=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":1556
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_20 = __pyx_v_i;
        __pyx_t_14 = ((!((__pyx_v_point < (*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_20 * __pyx_v_values.strides[0]) )))) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":1557
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:
 */
          goto __pyx_L13_break;

          /* "pandas/_libs/intervaltree.pxi":1556
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":1558
 *                     if not point < values[i]:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if self.right_node.min_left < point:
 *                     self.right_node.query(result, point)
 */
        __pyx_t_21 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_21 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L13_break:;

      /* "pandas/_libs/intervaltree.pxi":1559
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      __pyx_t_14 = ((__pyx_v_self->right_node->min_left < __pyx_v_point) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":1560
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:
 *                     self.right_node.query(result, point)             # <<<<<<<<<<<<<<
 *             else:
 *                 result.extend(self.center_left_indices)
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode *)__pyx_v_self->right_node->__pyx_vtab)->__pyx_fuse_0query(__pyx_v_self->right_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1560, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":1559
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":1552
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":1562
 *                     self.right_node.query(result, point)
 *             else:
 *                 result.extend(self.center_left_indices)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
    /*else*/ {
      __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->extend(__pyx_v_result, __pyx_v_self->center_left_indices); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1562, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_L7:;
  }
  __pyx_L3:;

  /* "pandas/_libs/intervaltree.pxi":1524
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __PYX_XDEC_MEMVIEW(&__pyx_t_16, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_17, 1);
  __Pyx_AddTraceback("pandas._libs.interval.Float64ClosedNeitherIntervalNode.query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_indices, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_values, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_9__pyx_fuse_0query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_0__pyx_mdef_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_9__pyx_fuse_0query = {"__pyx_fuse_0query", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_9__pyx_fuse_0query, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_2query};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_9__pyx_fuse_0query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result = 0;
  __pyx_t_5numpy_float64_t __pyx_v_point;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fuse_0query (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_result,&__pyx_n_s_point,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_result)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_point)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_0query", 1, 2, 2, 1); __PYX_ERR(1, 1524, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_0query") < 0)) __PYX_ERR(1, 1524, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_result = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)values[0]);
    __pyx_v_point = __pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_point == ((npy_float64)-1)) && PyErr_Occurred())) __PYX_ERR(1, 1524, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_0query", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 1524, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Float64ClosedNeitherIntervalNode.__pyx_fuse_0query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_result), __pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector, 1, "result", 0))) __PYX_ERR(1, 1524, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_8__pyx_fuse_0query(((struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode *)__pyx_v_self), __pyx_v_result, __pyx_v_point);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_8__pyx_fuse_0query(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float64_t __pyx_v_point) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__pyx_fuse_0query", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_fuse_0__pyx_f_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_query(__pyx_v_self, __pyx_v_result, __pyx_v_point, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1524, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Float64ClosedNeitherIntervalNode.__pyx_fuse_0query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pw_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_11__pyx_fuse_1query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_fuse_1__pyx_f_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float32_t __pyx_v_point, int __pyx_skip_dispatch) {
  __Pyx_memviewslice __pyx_v_indices = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_values = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __pyx_t_5numpy_int64_t __pyx_t_9;
  __pyx_t_5numpy_int64_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  int __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  __Pyx_memviewslice __pyx_t_16 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_17 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  Py_ssize_t __pyx_t_21;
  __Pyx_RefNannySetupContext("__pyx_fuse_1query", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pyx_fuse_1query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1524, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_11__pyx_fuse_1query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = PyFloat_FromDouble(__pyx_v_point); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1524, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1524, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1524, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 1524, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_5) {
            __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1524, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1524, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_3query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_4 = PyFloat_FromDouble(__pyx_v_point); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1524, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_7 = __pyx_t_1; __pyx_t_3 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1524, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1524, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1524, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (__pyx_t_3) {
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_t_4);
          __pyx_t_4 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1524, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pandas/_libs/intervaltree.pxi":1533
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
  __pyx_t_8 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_8) {

    /* "pandas/_libs/intervaltree.pxi":1537
 *             # continue the binary tree structure. Instead, we use linear
 *             # search.
 *             for i in range(self.n_elements):             # <<<<<<<<<<<<<<
 *                 if self.left[i] < point < self.right[i]:
 *                     result.append(self.indices[i])
 */
    __pyx_t_9 = __pyx_v_self->n_elements;
    __pyx_t_10 = __pyx_t_9;
    for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
      __pyx_v_i = __pyx_t_11;

      /* "pandas/_libs/intervaltree.pxi":1538
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point < self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      __pyx_t_12 = __pyx_v_i;
      __pyx_t_8 = ((*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_self->left.data + __pyx_t_12 * __pyx_v_self->left.strides[0]) ))) < __pyx_v_point);
      if (__pyx_t_8) {
        __pyx_t_13 = __pyx_v_i;
        __pyx_t_8 = (__pyx_v_point < (*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_self->right.data + __pyx_t_13 * __pyx_v_self->right.strides[0]) ))));
      }
      __pyx_t_14 = (__pyx_t_8 != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":1539
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point < self.right[i]:
 *                     result.append(self.indices[i])             # <<<<<<<<<<<<<<
 *         else:
 *             # There are child nodes. Based on comparing our query to the pivot,
 */
        __pyx_t_15 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->indices.data + __pyx_t_15 * __pyx_v_self->indices.strides[0]) ))));

        /* "pandas/_libs/intervaltree.pxi":1538
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point < self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      }
    }

    /* "pandas/_libs/intervaltree.pxi":1533
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
    goto __pyx_L3;
  }

  /* "pandas/_libs/intervaltree.pxi":1543
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
  /*else*/ {
    __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":1544
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:
 *                 values = self.center_left_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 */
      __pyx_t_16 = __pyx_v_self->center_left_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":1545
 *             if point < self.pivot:
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:
 */
      __pyx_t_17 = __pyx_v_self->center_left_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":1546
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):             # <<<<<<<<<<<<<<
 *                     if not values[i] < point:
 *                         break
 */
      __pyx_t_9 = __pyx_v_self->n_center;
      __pyx_t_10 = __pyx_t_9;
      for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":1547
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_18 = __pyx_v_i;
        __pyx_t_14 = ((!(((*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_18 * __pyx_v_values.strides[0]) ))) < __pyx_v_point) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":1548
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:
 */
          goto __pyx_L9_break;

          /* "pandas/_libs/intervaltree.pxi":1547
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":1549
 *                     if not values[i] < point:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 */
        __pyx_t_19 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_19 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L9_break:;

      /* "pandas/_libs/intervaltree.pxi":1550
 *                         break
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->left_node->max_right) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":1551
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)             # <<<<<<<<<<<<<<
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode *)__pyx_v_self->left_node->__pyx_vtab)->__pyx_fuse_1query(__pyx_v_self->left_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1551, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":1550
 *                         break
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":1543
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":1552
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
    __pyx_t_14 = ((__pyx_v_point > __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":1553
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 *                 values = self.center_right_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 */
      __pyx_t_16 = __pyx_v_self->center_right_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":1554
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:
 */
      __pyx_t_17 = __pyx_v_self->center_right_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":1555
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):             # <<<<<<<<<<<<<<
 *                     if not point < values[i]:
 *                         break
 */
      for (__pyx_t_11 = (__pyx_v_self->n_center - 1); __pyx_t_11 > -1L; __pyx_t_11-=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":1556
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_20 = __pyx_v_i;
        __pyx_t_14 = ((!((__pyx_v_point < (*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_20 * __pyx_v_values.strides[0]) )))) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":1557
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:
 */
          goto __pyx_L13_break;

          /* "pandas/_libs/intervaltree.pxi":1556
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":1558
 *                     if not point < values[i]:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if self.right_node.min_left < point:
 *                     self.right_node.query(result, point)
 */
        __pyx_t_21 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_21 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L13_break:;

      /* "pandas/_libs/intervaltree.pxi":1559
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      __pyx_t_14 = ((__pyx_v_self->right_node->min_left < __pyx_v_point) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":1560
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:
 *                     self.right_node.query(result, point)             # <<<<<<<<<<<<<<
 *             else:
 *                 result.extend(self.center_left_indices)
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode *)__pyx_v_self->right_node->__pyx_vtab)->__pyx_fuse_1query(__pyx_v_self->right_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1560, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":1559
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":1552
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":1562
 *                     self.right_node.query(result, point)
 *             else:
 *                 result.extend(self.center_left_indices)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
    /*else*/ {
      __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->extend(__pyx_v_result, __pyx_v_self->center_left_indices); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1562, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_L7:;
  }
  __pyx_L3:;

  /* "pandas/_libs/intervaltree.pxi":1524
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __PYX_XDEC_MEMVIEW(&__pyx_t_16, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_17, 1);
  __Pyx_AddTraceback("pandas._libs.interval.Float64ClosedNeitherIntervalNode.query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_indices, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_values, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_11__pyx_fuse_1query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_1__pyx_mdef_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_11__pyx_fuse_1query = {"__pyx_fuse_1query", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_11__pyx_fuse_1query, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_2query};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_11__pyx_fuse_1query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result = 0;
  __pyx_t_5numpy_float32_t __pyx_v_point;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fuse_1query (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_result,&__pyx_n_s_point,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_result)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_point)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_1query", 1, 2, 2, 1); __PYX_ERR(1, 1524, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_1query") < 0)) __PYX_ERR(1, 1524, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_result = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)values[0]);
    __pyx_v_point = __pyx_PyFloat_AsFloat(values[1]); if (unlikely((__pyx_v_point == ((npy_float32)-1)) && PyErr_Occurred())) __PYX_ERR(1, 1524, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_1query", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 1524, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Float64ClosedNeitherIntervalNode.__pyx_fuse_1query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_result), __pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector, 1, "result", 0))) __PYX_ERR(1, 1524, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_10__pyx_fuse_1query(((struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode *)__pyx_v_self), __pyx_v_result, __pyx_v_point);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_10__pyx_fuse_1query(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float32_t __pyx_v_point) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__pyx_fuse_1query", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_fuse_1__pyx_f_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_query(__pyx_v_self, __pyx_v_result, __pyx_v_point, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1524, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Float64ClosedNeitherIntervalNode.__pyx_fuse_1query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pw_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_13__pyx_fuse_2query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_fuse_2__pyx_f_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int64_t __pyx_v_point, int __pyx_skip_dispatch) {
  __Pyx_memviewslice __pyx_v_indices = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_values = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __pyx_t_5numpy_int64_t __pyx_t_9;
  __pyx_t_5numpy_int64_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  int __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  __Pyx_memviewslice __pyx_t_16 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_17 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  Py_ssize_t __pyx_t_21;
  __Pyx_RefNannySetupContext("__pyx_fuse_2query", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pyx_fuse_2query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1524, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_13__pyx_fuse_2query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = __Pyx_PyInt_From_npy_int64(__pyx_v_point); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1524, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1524, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1524, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 1524, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_5) {
            __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1524, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1524, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_3query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_4 = __Pyx_PyInt_From_npy_int64(__pyx_v_point); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1524, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_7 = __pyx_t_1; __pyx_t_3 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1524, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1524, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1524, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (__pyx_t_3) {
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_t_4);
          __pyx_t_4 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1524, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pandas/_libs/intervaltree.pxi":1533
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
  __pyx_t_8 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_8) {

    /* "pandas/_libs/intervaltree.pxi":1537
 *             # continue the binary tree structure. Instead, we use linear
 *             # search.
 *             for i in range(self.n_elements):             # <<<<<<<<<<<<<<
 *                 if self.left[i] < point < self.right[i]:
 *                     result.append(self.indices[i])
 */
    __pyx_t_9 = __pyx_v_self->n_elements;
    __pyx_t_10 = __pyx_t_9;
    for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
      __pyx_v_i = __pyx_t_11;

      /* "pandas/_libs/intervaltree.pxi":1538
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point < self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      __pyx_t_12 = __pyx_v_i;
      __pyx_t_8 = ((*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_self->left.data + __pyx_t_12 * __pyx_v_self->left.strides[0]) ))) < __pyx_v_point);
      if (__pyx_t_8) {
        __pyx_t_13 = __pyx_v_i;
        __pyx_t_8 = (__pyx_v_point < (*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_self->right.data + __pyx_t_13 * __pyx_v_self->right.strides[0]) ))));
      }
      __pyx_t_14 = (__pyx_t_8 != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":1539
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point < self.right[i]:
 *                     result.append(self.indices[i])             # <<<<<<<<<<<<<<
 *         else:
 *             # There are child nodes. Based on comparing our query to the pivot,
 */
        __pyx_t_15 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->indices.data + __pyx_t_15 * __pyx_v_self->indices.strides[0]) ))));

        /* "pandas/_libs/intervaltree.pxi":1538
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point < self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      }
    }

    /* "pandas/_libs/intervaltree.pxi":1533
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
    goto __pyx_L3;
  }

  /* "pandas/_libs/intervaltree.pxi":1543
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
  /*else*/ {
    __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":1544
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:
 *                 values = self.center_left_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 */
      __pyx_t_16 = __pyx_v_self->center_left_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":1545
 *             if point < self.pivot:
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:
 */
      __pyx_t_17 = __pyx_v_self->center_left_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":1546
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):             # <<<<<<<<<<<<<<
 *                     if not values[i] < point:
 *                         break
 */
      __pyx_t_9 = __pyx_v_self->n_center;
      __pyx_t_10 = __pyx_t_9;
      for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":1547
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_18 = __pyx_v_i;
        __pyx_t_14 = ((!(((*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_18 * __pyx_v_values.strides[0]) ))) < __pyx_v_point) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":1548
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:
 */
          goto __pyx_L9_break;

          /* "pandas/_libs/intervaltree.pxi":1547
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":1549
 *                     if not values[i] < point:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 */
        __pyx_t_19 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_19 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L9_break:;

      /* "pandas/_libs/intervaltree.pxi":1550
 *                         break
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->left_node->max_right) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":1551
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)             # <<<<<<<<<<<<<<
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode *)__pyx_v_self->left_node->__pyx_vtab)->__pyx_fuse_2query(__pyx_v_self->left_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1551, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":1550
 *                         break
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":1543
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":1552
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
    __pyx_t_14 = ((__pyx_v_point > __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":1553
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 *                 values = self.center_right_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 */
      __pyx_t_16 = __pyx_v_self->center_right_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":1554
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:
 */
      __pyx_t_17 = __pyx_v_self->center_right_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":1555
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):             # <<<<<<<<<<<<<<
 *                     if not point < values[i]:
 *                         break
 */
      for (__pyx_t_11 = (__pyx_v_self->n_center - 1); __pyx_t_11 > -1L; __pyx_t_11-=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":1556
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_20 = __pyx_v_i;
        __pyx_t_14 = ((!((__pyx_v_point < (*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_20 * __pyx_v_values.strides[0]) )))) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":1557
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:
 */
          goto __pyx_L13_break;

          /* "pandas/_libs/intervaltree.pxi":1556
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":1558
 *                     if not point < values[i]:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if self.right_node.min_left < point:
 *                     self.right_node.query(result, point)
 */
        __pyx_t_21 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_21 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L13_break:;

      /* "pandas/_libs/intervaltree.pxi":1559
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      __pyx_t_14 = ((__pyx_v_self->right_node->min_left < __pyx_v_point) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":1560
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:
 *                     self.right_node.query(result, point)             # <<<<<<<<<<<<<<
 *             else:
 *                 result.extend(self.center_left_indices)
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode *)__pyx_v_self->right_node->__pyx_vtab)->__pyx_fuse_2query(__pyx_v_self->right_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1560, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":1559
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":1552
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":1562
 *                     self.right_node.query(result, point)
 *             else:
 *                 result.extend(self.center_left_indices)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
    /*else*/ {
      __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->extend(__pyx_v_result, __pyx_v_self->center_left_indices); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1562, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_L7:;
  }
  __pyx_L3:;

  /* "pandas/_libs/intervaltree.pxi":1524
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __PYX_XDEC_MEMVIEW(&__pyx_t_16, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_17, 1);
  __Pyx_AddTraceback("pandas._libs.interval.Float64ClosedNeitherIntervalNode.query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_indices, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_values, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_13__pyx_fuse_2query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_2__pyx_mdef_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_13__pyx_fuse_2query = {"__pyx_fuse_2query", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_13__pyx_fuse_2query, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_2query};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_13__pyx_fuse_2query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result = 0;
  __pyx_t_5numpy_int64_t __pyx_v_point;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fuse_2query (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_result,&__pyx_n_s_point,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_result)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_point)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_2query", 1, 2, 2, 1); __PYX_ERR(1, 1524, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_2query") < 0)) __PYX_ERR(1, 1524, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_result = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)values[0]);
    __pyx_v_point = __Pyx_PyInt_As_npy_int64(values[1]); if (unlikely((__pyx_v_point == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(1, 1524, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_2query", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 1524, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Float64ClosedNeitherIntervalNode.__pyx_fuse_2query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_result), __pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector, 1, "result", 0))) __PYX_ERR(1, 1524, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_12__pyx_fuse_2query(((struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode *)__pyx_v_self), __pyx_v_result, __pyx_v_point);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_12__pyx_fuse_2query(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int64_t __pyx_v_point) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__pyx_fuse_2query", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_fuse_2__pyx_f_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_query(__pyx_v_self, __pyx_v_result, __pyx_v_point, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1524, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Float64ClosedNeitherIntervalNode.__pyx_fuse_2query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pw_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_15__pyx_fuse_3query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_fuse_3__pyx_f_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int32_t __pyx_v_point, int __pyx_skip_dispatch) {
  __Pyx_memviewslice __pyx_v_indices = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_values = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __pyx_t_5numpy_int64_t __pyx_t_9;
  __pyx_t_5numpy_int64_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  int __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  __Pyx_memviewslice __pyx_t_16 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_17 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  Py_ssize_t __pyx_t_21;
  __Pyx_RefNannySetupContext("__pyx_fuse_3query", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pyx_fuse_3query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1524, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_15__pyx_fuse_3query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = __Pyx_PyInt_From_npy_int32(__pyx_v_point); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1524, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1524, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1524, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 1524, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_5) {
            __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1524, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1524, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_3query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_4 = __Pyx_PyInt_From_npy_int32(__pyx_v_point); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1524, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_7 = __pyx_t_1; __pyx_t_3 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1524, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1524, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1524, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (__pyx_t_3) {
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_t_4);
          __pyx_t_4 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1524, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pandas/_libs/intervaltree.pxi":1533
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
  __pyx_t_8 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_8) {

    /* "pandas/_libs/intervaltree.pxi":1537
 *             # continue the binary tree structure. Instead, we use linear
 *             # search.
 *             for i in range(self.n_elements):             # <<<<<<<<<<<<<<
 *                 if self.left[i] < point < self.right[i]:
 *                     result.append(self.indices[i])
 */
    __pyx_t_9 = __pyx_v_self->n_elements;
    __pyx_t_10 = __pyx_t_9;
    for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
      __pyx_v_i = __pyx_t_11;

      /* "pandas/_libs/intervaltree.pxi":1538
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point < self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      __pyx_t_12 = __pyx_v_i;
      __pyx_t_8 = ((*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_self->left.data + __pyx_t_12 * __pyx_v_self->left.strides[0]) ))) < __pyx_v_point);
      if (__pyx_t_8) {
        __pyx_t_13 = __pyx_v_i;
        __pyx_t_8 = (__pyx_v_point < (*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_self->right.data + __pyx_t_13 * __pyx_v_self->right.strides[0]) ))));
      }
      __pyx_t_14 = (__pyx_t_8 != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":1539
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point < self.right[i]:
 *                     result.append(self.indices[i])             # <<<<<<<<<<<<<<
 *         else:
 *             # There are child nodes. Based on comparing our query to the pivot,
 */
        __pyx_t_15 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->indices.data + __pyx_t_15 * __pyx_v_self->indices.strides[0]) ))));

        /* "pandas/_libs/intervaltree.pxi":1538
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point < self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      }
    }

    /* "pandas/_libs/intervaltree.pxi":1533
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
    goto __pyx_L3;
  }

  /* "pandas/_libs/intervaltree.pxi":1543
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
  /*else*/ {
    __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":1544
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:
 *                 values = self.center_left_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 */
      __pyx_t_16 = __pyx_v_self->center_left_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":1545
 *             if point < self.pivot:
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:
 */
      __pyx_t_17 = __pyx_v_self->center_left_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":1546
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):             # <<<<<<<<<<<<<<
 *                     if not values[i] < point:
 *                         break
 */
      __pyx_t_9 = __pyx_v_self->n_center;
      __pyx_t_10 = __pyx_t_9;
      for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":1547
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_18 = __pyx_v_i;
        __pyx_t_14 = ((!(((*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_18 * __pyx_v_values.strides[0]) ))) < __pyx_v_point) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":1548
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:
 */
          goto __pyx_L9_break;

          /* "pandas/_libs/intervaltree.pxi":1547
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":1549
 *                     if not values[i] < point:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 */
        __pyx_t_19 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_19 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L9_break:;

      /* "pandas/_libs/intervaltree.pxi":1550
 *                         break
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->left_node->max_right) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":1551
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)             # <<<<<<<<<<<<<<
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode *)__pyx_v_self->left_node->__pyx_vtab)->__pyx_fuse_3query(__pyx_v_self->left_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1551, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":1550
 *                         break
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":1543
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":1552
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
    __pyx_t_14 = ((__pyx_v_point > __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":1553
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 *                 values = self.center_right_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 */
      __pyx_t_16 = __pyx_v_self->center_right_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":1554
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:
 */
      __pyx_t_17 = __pyx_v_self->center_right_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":1555
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):             # <<<<<<<<<<<<<<
 *                     if not point < values[i]:
 *                         break
 */
      for (__pyx_t_11 = (__pyx_v_self->n_center - 1); __pyx_t_11 > -1L; __pyx_t_11-=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":1556
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_20 = __pyx_v_i;
        __pyx_t_14 = ((!((__pyx_v_point < (*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_20 * __pyx_v_values.strides[0]) )))) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":1557
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:
 */
          goto __pyx_L13_break;

          /* "pandas/_libs/intervaltree.pxi":1556
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":1558
 *                     if not point < values[i]:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if self.right_node.min_left < point:
 *                     self.right_node.query(result, point)
 */
        __pyx_t_21 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_21 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L13_break:;

      /* "pandas/_libs/intervaltree.pxi":1559
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      __pyx_t_14 = ((__pyx_v_self->right_node->min_left < __pyx_v_point) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":1560
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:
 *                     self.right_node.query(result, point)             # <<<<<<<<<<<<<<
 *             else:
 *                 result.extend(self.center_left_indices)
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode *)__pyx_v_self->right_node->__pyx_vtab)->__pyx_fuse_3query(__pyx_v_self->right_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1560, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":1559
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":1552
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":1562
 *                     self.right_node.query(result, point)
 *             else:
 *                 result.extend(self.center_left_indices)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
    /*else*/ {
      __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->extend(__pyx_v_result, __pyx_v_self->center_left_indices); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1562, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_L7:;
  }
  __pyx_L3:;

  /* "pandas/_libs/intervaltree.pxi":1524
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __PYX_XDEC_MEMVIEW(&__pyx_t_16, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_17, 1);
  __Pyx_AddTraceback("pandas._libs.interval.Float64ClosedNeitherIntervalNode.query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_indices, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_values, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_15__pyx_fuse_3query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_3__pyx_mdef_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_15__pyx_fuse_3query = {"__pyx_fuse_3query", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_15__pyx_fuse_3query, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_2query};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_15__pyx_fuse_3query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result = 0;
  __pyx_t_5numpy_int32_t __pyx_v_point;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fuse_3query (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_result,&__pyx_n_s_point,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_result)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_point)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_3query", 1, 2, 2, 1); __PYX_ERR(1, 1524, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_3query") < 0)) __PYX_ERR(1, 1524, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_result = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)values[0]);
    __pyx_v_point = __Pyx_PyInt_As_npy_int32(values[1]); if (unlikely((__pyx_v_point == ((npy_int32)-1)) && PyErr_Occurred())) __PYX_ERR(1, 1524, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_3query", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 1524, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Float64ClosedNeitherIntervalNode.__pyx_fuse_3query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_result), __pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector, 1, "result", 0))) __PYX_ERR(1, 1524, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_14__pyx_fuse_3query(((struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode *)__pyx_v_self), __pyx_v_result, __pyx_v_point);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_14__pyx_fuse_3query(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int32_t __pyx_v_point) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__pyx_fuse_3query", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_fuse_3__pyx_f_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_query(__pyx_v_self, __pyx_v_result, __pyx_v_point, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1524, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Float64ClosedNeitherIntervalNode.__pyx_fuse_3query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pw_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_17__pyx_fuse_4query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_fuse_4__pyx_f_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_uint64_t __pyx_v_point, int __pyx_skip_dispatch) {
  __Pyx_memviewslice __pyx_v_indices = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_values = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __pyx_t_5numpy_int64_t __pyx_t_9;
  __pyx_t_5numpy_int64_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  int __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  __Pyx_memviewslice __pyx_t_16 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_17 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  Py_ssize_t __pyx_t_21;
  __Pyx_RefNannySetupContext("__pyx_fuse_4query", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pyx_fuse_4query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1524, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_17__pyx_fuse_4query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = __Pyx_PyInt_From_npy_uint64(__pyx_v_point); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1524, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1524, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1524, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 1524, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_5) {
            __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1524, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1524, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_3query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_4 = __Pyx_PyInt_From_npy_uint64(__pyx_v_point); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1524, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_7 = __pyx_t_1; __pyx_t_3 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1524, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1524, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1524, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (__pyx_t_3) {
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_t_4);
          __pyx_t_4 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1524, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pandas/_libs/intervaltree.pxi":1533
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
  __pyx_t_8 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_8) {

    /* "pandas/_libs/intervaltree.pxi":1537
 *             # continue the binary tree structure. Instead, we use linear
 *             # search.
 *             for i in range(self.n_elements):             # <<<<<<<<<<<<<<
 *                 if self.left[i] < point < self.right[i]:
 *                     result.append(self.indices[i])
 */
    __pyx_t_9 = __pyx_v_self->n_elements;
    __pyx_t_10 = __pyx_t_9;
    for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
      __pyx_v_i = __pyx_t_11;

      /* "pandas/_libs/intervaltree.pxi":1538
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point < self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      __pyx_t_12 = __pyx_v_i;
      __pyx_t_8 = ((*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_self->left.data + __pyx_t_12 * __pyx_v_self->left.strides[0]) ))) < __pyx_v_point);
      if (__pyx_t_8) {
        __pyx_t_13 = __pyx_v_i;
        __pyx_t_8 = (__pyx_v_point < (*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_self->right.data + __pyx_t_13 * __pyx_v_self->right.strides[0]) ))));
      }
      __pyx_t_14 = (__pyx_t_8 != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":1539
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point < self.right[i]:
 *                     result.append(self.indices[i])             # <<<<<<<<<<<<<<
 *         else:
 *             # There are child nodes. Based on comparing our query to the pivot,
 */
        __pyx_t_15 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->indices.data + __pyx_t_15 * __pyx_v_self->indices.strides[0]) ))));

        /* "pandas/_libs/intervaltree.pxi":1538
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point < self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      }
    }

    /* "pandas/_libs/intervaltree.pxi":1533
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
    goto __pyx_L3;
  }

  /* "pandas/_libs/intervaltree.pxi":1543
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
  /*else*/ {
    __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":1544
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:
 *                 values = self.center_left_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 */
      __pyx_t_16 = __pyx_v_self->center_left_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":1545
 *             if point < self.pivot:
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:
 */
      __pyx_t_17 = __pyx_v_self->center_left_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":1546
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):             # <<<<<<<<<<<<<<
 *                     if not values[i] < point:
 *                         break
 */
      __pyx_t_9 = __pyx_v_self->n_center;
      __pyx_t_10 = __pyx_t_9;
      for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":1547
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_18 = __pyx_v_i;
        __pyx_t_14 = ((!(((*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_18 * __pyx_v_values.strides[0]) ))) < __pyx_v_point) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":1548
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:
 */
          goto __pyx_L9_break;

          /* "pandas/_libs/intervaltree.pxi":1547
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":1549
 *                     if not values[i] < point:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 */
        __pyx_t_19 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_19 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L9_break:;

      /* "pandas/_libs/intervaltree.pxi":1550
 *                         break
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->left_node->max_right) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":1551
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)             # <<<<<<<<<<<<<<
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode *)__pyx_v_self->left_node->__pyx_vtab)->__pyx_fuse_4query(__pyx_v_self->left_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1551, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":1550
 *                         break
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":1543
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":1552
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
    __pyx_t_14 = ((__pyx_v_point > __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":1553
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 *                 values = self.center_right_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 */
      __pyx_t_16 = __pyx_v_self->center_right_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":1554
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:
 */
      __pyx_t_17 = __pyx_v_self->center_right_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":1555
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):             # <<<<<<<<<<<<<<
 *                     if not point < values[i]:
 *                         break
 */
      for (__pyx_t_11 = (__pyx_v_self->n_center - 1); __pyx_t_11 > -1L; __pyx_t_11-=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":1556
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_20 = __pyx_v_i;
        __pyx_t_14 = ((!((__pyx_v_point < (*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_20 * __pyx_v_values.strides[0]) )))) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":1557
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:
 */
          goto __pyx_L13_break;

          /* "pandas/_libs/intervaltree.pxi":1556
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":1558
 *                     if not point < values[i]:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if self.right_node.min_left < point:
 *                     self.right_node.query(result, point)
 */
        __pyx_t_21 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_21 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L13_break:;

      /* "pandas/_libs/intervaltree.pxi":1559
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      __pyx_t_14 = ((__pyx_v_self->right_node->min_left < __pyx_v_point) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":1560
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:
 *                     self.right_node.query(result, point)             # <<<<<<<<<<<<<<
 *             else:
 *                 result.extend(self.center_left_indices)
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode *)__pyx_v_self->right_node->__pyx_vtab)->__pyx_fuse_4query(__pyx_v_self->right_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1560, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":1559
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":1552
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":1562
 *                     self.right_node.query(result, point)
 *             else:
 *                 result.extend(self.center_left_indices)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
    /*else*/ {
      __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->extend(__pyx_v_result, __pyx_v_self->center_left_indices); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1562, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_L7:;
  }
  __pyx_L3:;

  /* "pandas/_libs/intervaltree.pxi":1524
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __PYX_XDEC_MEMVIEW(&__pyx_t_16, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_17, 1);
  __Pyx_AddTraceback("pandas._libs.interval.Float64ClosedNeitherIntervalNode.query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_indices, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_values, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_17__pyx_fuse_4query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_4__pyx_mdef_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_17__pyx_fuse_4query = {"__pyx_fuse_4query", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_17__pyx_fuse_4query, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_2query};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_17__pyx_fuse_4query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result = 0;
  __pyx_t_5numpy_uint64_t __pyx_v_point;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fuse_4query (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_result,&__pyx_n_s_point,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_result)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_point)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_4query", 1, 2, 2, 1); __PYX_ERR(1, 1524, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_4query") < 0)) __PYX_ERR(1, 1524, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_result = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)values[0]);
    __pyx_v_point = __Pyx_PyInt_As_npy_uint64(values[1]); if (unlikely((__pyx_v_point == ((npy_uint64)-1)) && PyErr_Occurred())) __PYX_ERR(1, 1524, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_4query", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 1524, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Float64ClosedNeitherIntervalNode.__pyx_fuse_4query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_result), __pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector, 1, "result", 0))) __PYX_ERR(1, 1524, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_16__pyx_fuse_4query(((struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode *)__pyx_v_self), __pyx_v_result, __pyx_v_point);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_16__pyx_fuse_4query(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_uint64_t __pyx_v_point) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__pyx_fuse_4query", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_fuse_4__pyx_f_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_query(__pyx_v_self, __pyx_v_result, __pyx_v_point, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1524, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Float64ClosedNeitherIntervalNode.__pyx_fuse_4query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":1564
 *                 result.extend(self.center_left_indices)
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         if self.is_leaf_node:
 *             return ('<Float64ClosedNeitherIntervalNode: '
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_5__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_5__repr__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_4__repr__(((struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_4__repr__(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode *__pyx_v_self) {
  PyObject *__pyx_v_n_left = NULL;
  PyObject *__pyx_v_n_right = NULL;
  PyObject *__pyx_v_n_center = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  Py_UCS4 __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "pandas/_libs/intervaltree.pxi":1565
 * 
 *     def __repr__(self):
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             return ('<Float64ClosedNeitherIntervalNode: '
 *                     '%s elements (terminal)>' % self.n_elements)
 */
  __pyx_t_1 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_1) {

    /* "pandas/_libs/intervaltree.pxi":1566
 *     def __repr__(self):
 *         if self.is_leaf_node:
 *             return ('<Float64ClosedNeitherIntervalNode: '             # <<<<<<<<<<<<<<
 *                     '%s elements (terminal)>' % self.n_elements)
 *         else:
 */
    __Pyx_XDECREF(__pyx_r);

    /* "pandas/_libs/intervaltree.pxi":1567
 *         if self.is_leaf_node:
 *             return ('<Float64ClosedNeitherIntervalNode: '
 *                     '%s elements (terminal)>' % self.n_elements)             # <<<<<<<<<<<<<<
 *         else:
 *             n_left = self.left_node.n_elements
 */
    __pyx_t_2 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_elements); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1567, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PyUnicode_Format(__pyx_kp_u_Float64ClosedNeitherIntervalNod, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1567, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pandas/_libs/intervaltree.pxi":1565
 * 
 *     def __repr__(self):
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             return ('<Float64ClosedNeitherIntervalNode: '
 *                     '%s elements (terminal)>' % self.n_elements)
 */
  }

  /* "pandas/_libs/intervaltree.pxi":1569
 *                     '%s elements (terminal)>' % self.n_elements)
 *         else:
 *             n_left = self.left_node.n_elements             # <<<<<<<<<<<<<<
 *             n_right = self.right_node.n_elements
 *             n_center = self.n_elements - n_left - n_right
 */
  /*else*/ {
    __pyx_t_3 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->left_node->n_elements); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1569, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_v_n_left = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "pandas/_libs/intervaltree.pxi":1570
 *         else:
 *             n_left = self.left_node.n_elements
 *             n_right = self.right_node.n_elements             # <<<<<<<<<<<<<<
 *             n_center = self.n_elements - n_left - n_right
 *             return ('<Float64ClosedNeitherIntervalNode: '
 */
    __pyx_t_3 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->right_node->n_elements); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1570, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_v_n_right = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "pandas/_libs/intervaltree.pxi":1571
 *             n_left = self.left_node.n_elements
 *             n_right = self.right_node.n_elements
 *             n_center = self.n_elements - n_left - n_right             # <<<<<<<<<<<<<<
 *             return ('<Float64ClosedNeitherIntervalNode: '
 *                     'pivot %s, %s elements (%s left, %s right, %s '
 */
    __pyx_t_3 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_elements); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1571, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = PyNumber_Subtract(__pyx_t_3, __pyx_v_n_left); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1571, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyNumber_Subtract(__pyx_t_2, __pyx_v_n_right); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1571, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_n_center = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "pandas/_libs/intervaltree.pxi":1572
 *             n_right = self.right_node.n_elements
 *             n_center = self.n_elements - n_left - n_right
 *             return ('<Float64ClosedNeitherIntervalNode: '             # <<<<<<<<<<<<<<
 *                     'pivot %s, %s elements (%s left, %s right, %s '
 *                     'overlapping)>' % (self.pivot, self.n_elements,
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = PyTuple_New(11); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1572, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = 0;
    __pyx_t_5 = 127;
    __Pyx_INCREF(__pyx_kp_u_Float64ClosedNeitherIntervalNod_2);
    __pyx_t_4 += 41;
    __Pyx_GIVEREF(__pyx_kp_u_Float64ClosedNeitherIntervalNod_2);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_kp_u_Float64ClosedNeitherIntervalNod_2);

    /* "pandas/_libs/intervaltree.pxi":1574
 *             return ('<Float64ClosedNeitherIntervalNode: '
 *                     'pivot %s, %s elements (%s left, %s right, %s '
 *                     'overlapping)>' % (self.pivot, self.n_elements,             # <<<<<<<<<<<<<<
 *                                        n_left, n_right, n_center))
 * 
 */
    __pyx_t_2 = PyFloat_FromDouble(__pyx_v_self->pivot); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1574, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_t_2), __pyx_empty_unicode); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 1574, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_5 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_6) > __pyx_t_5) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_6) : __pyx_t_5;
    __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_6);
    __pyx_t_6 = 0;
    __Pyx_INCREF(__pyx_kp_u__18);
    __pyx_t_4 += 2;
    __Pyx_GIVEREF(__pyx_kp_u__18);
    PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_kp_u__18);
    __pyx_t_6 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_elements); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 1574, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_2 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_t_6), __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1574, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_5 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) > __pyx_t_5) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) : __pyx_t_5;
    __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_3, 3, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_elements);
    __pyx_t_4 += 11;
    __Pyx_GIVEREF(__pyx_kp_u_elements);
    PyTuple_SET_ITEM(__pyx_t_3, 4, __pyx_kp_u_elements);

    /* "pandas/_libs/intervaltree.pxi":1575
 *                     'pivot %s, %s elements (%s left, %s right, %s '
 *                     'overlapping)>' % (self.pivot, self.n_elements,
 *                                        n_left, n_right, n_center))             # <<<<<<<<<<<<<<
 * 
 *     def counts(self):
 */
    __pyx_t_2 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_v_n_left), __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1575, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) > __pyx_t_5) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) : __pyx_t_5;
    __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_3, 5, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_left_2);
    __pyx_t_4 += 7;
    __Pyx_GIVEREF(__pyx_kp_u_left_2);
    PyTuple_SET_ITEM(__pyx_t_3, 6, __pyx_kp_u_left_2);
    __pyx_t_2 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_v_n_right), __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1575, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) > __pyx_t_5) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) : __pyx_t_5;
    __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_3, 7, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_right_2);
    __pyx_t_4 += 8;
    __Pyx_GIVEREF(__pyx_kp_u_right_2);
    PyTuple_SET_ITEM(__pyx_t_3, 8, __pyx_kp_u_right_2);
    __pyx_t_2 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_v_n_center), __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1575, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) > __pyx_t_5) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) : __pyx_t_5;
    __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_3, 9, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_overlapping);
    __pyx_t_4 += 14;
    __Pyx_GIVEREF(__pyx_kp_u_overlapping);
    PyTuple_SET_ITEM(__pyx_t_3, 10, __pyx_kp_u_overlapping);

    /* "pandas/_libs/intervaltree.pxi":1572
 *             n_right = self.right_node.n_elements
 *             n_center = self.n_elements - n_left - n_right
 *             return ('<Float64ClosedNeitherIntervalNode: '             # <<<<<<<<<<<<<<
 *                     'pivot %s, %s elements (%s left, %s right, %s '
 *                     'overlapping)>' % (self.pivot, self.n_elements,
 */
    __pyx_t_2 = __Pyx_PyUnicode_Join(__pyx_t_3, 11, __pyx_t_4, __pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1572, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;
  }

  /* "pandas/_libs/intervaltree.pxi":1564
 *                 result.extend(self.center_left_indices)
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         if self.is_leaf_node:
 *             return ('<Float64ClosedNeitherIntervalNode: '
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pandas._libs.interval.Float64ClosedNeitherIntervalNode.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_n_left);
  __Pyx_XDECREF(__pyx_v_n_right);
  __Pyx_XDECREF(__pyx_v_n_center);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":1577
 *                                        n_left, n_right, n_center))
 * 
 *     def counts(self):             # <<<<<<<<<<<<<<
 *         """
 *         Inspect counts on this node
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_7counts(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_6counts[] = "\n        Inspect counts on this node\n        useful for debugging purposes\n        ";
static PyObject *__pyx_pw_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_7counts(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("counts (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_6counts(((struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_6counts(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode *__pyx_v_self) {
  Py_ssize_t __pyx_v_m;
  PyObject *__pyx_v_l = NULL;
  PyObject *__pyx_v_r = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  size_t __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("counts", 0);

  /* "pandas/_libs/intervaltree.pxi":1582
 *         useful for debugging purposes
 *         """
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             return self.n_elements
 *         else:
 */
  __pyx_t_1 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_1) {

    /* "pandas/_libs/intervaltree.pxi":1583
 *         """
 *         if self.is_leaf_node:
 *             return self.n_elements             # <<<<<<<<<<<<<<
 *         else:
 *             m = len(self.center_left_values)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_elements); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1583, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "pandas/_libs/intervaltree.pxi":1582
 *         useful for debugging purposes
 *         """
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             return self.n_elements
 *         else:
 */
  }

  /* "pandas/_libs/intervaltree.pxi":1585
 *             return self.n_elements
 *         else:
 *             m = len(self.center_left_values)             # <<<<<<<<<<<<<<
 *             l = self.left_node.counts()
 *             r = self.right_node.counts()
 */
  /*else*/ {
    if (unlikely(!__pyx_v_self->center_left_values.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(1, 1585, __pyx_L1_error)}
    __pyx_t_3 = __Pyx_MemoryView_Len(__pyx_v_self->center_left_values); 
    __pyx_v_m = __pyx_t_3;

    /* "pandas/_libs/intervaltree.pxi":1586
 *         else:
 *             m = len(self.center_left_values)
 *             l = self.left_node.counts()             # <<<<<<<<<<<<<<
 *             r = self.right_node.counts()
 *             return (m, (l, r))
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->left_node), __pyx_n_s_counts); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1586, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1586, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_l = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "pandas/_libs/intervaltree.pxi":1587
 *             m = len(self.center_left_values)
 *             l = self.left_node.counts()
 *             r = self.right_node.counts()             # <<<<<<<<<<<<<<
 *             return (m, (l, r))
 * 
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->right_node), __pyx_n_s_counts); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1587, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1587, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_r = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "pandas/_libs/intervaltree.pxi":1588
 *             l = self.left_node.counts()
 *             r = self.right_node.counts()
 *             return (m, (l, r))             # <<<<<<<<<<<<<<
 * 
 * NODE_CLASSES['float64',
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = PyInt_FromSsize_t(__pyx_v_m); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1588, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1588, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_v_l);
    __Pyx_GIVEREF(__pyx_v_l);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_l);
    __Pyx_INCREF(__pyx_v_r);
    __Pyx_GIVEREF(__pyx_v_r);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_v_r);
    __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1588, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_4);
    __pyx_t_2 = 0;
    __pyx_t_4 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;
  }

  /* "pandas/_libs/intervaltree.pxi":1577
 *                                        n_left, n_right, n_center))
 * 
 *     def counts(self):             # <<<<<<<<<<<<<<
 *         """
 *         Inspect counts on this node
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pandas._libs.interval.Float64ClosedNeitherIntervalNode.counts", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_l);
  __Pyx_XDECREF(__pyx_v_r);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":1434
 *         int64_t[:] center_left_indices, center_right_indices, indices
 *         float64_t min_left, max_right
 *         readonly float64_t pivot             # <<<<<<<<<<<<<<
 *         readonly int64_t n_elements, n_center, leaf_size
 *         readonly bint is_leaf_node
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_5pivot_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_5pivot_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_5pivot___get__(((struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_5pivot___get__(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->pivot); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1434, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Float64ClosedNeitherIntervalNode.pivot.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":1435
 *         float64_t min_left, max_right
 *         readonly float64_t pivot
 *         readonly int64_t n_elements, n_center, leaf_size             # <<<<<<<<<<<<<<
 *         readonly bint is_leaf_node
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_10n_elements_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_10n_elements_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_10n_elements___get__(((struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_10n_elements___get__(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_elements); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1435, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Float64ClosedNeitherIntervalNode.n_elements.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_8n_center_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_8n_center_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_8n_center___get__(((struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_8n_center___get__(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_center); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1435, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Float64ClosedNeitherIntervalNode.n_center.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_9leaf_size_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_9leaf_size_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_9leaf_size___get__(((struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_9leaf_size___get__(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->leaf_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1435, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Float64ClosedNeitherIntervalNode.leaf_size.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":1436
 *         readonly float64_t pivot
 *         readonly int64_t n_elements, n_center, leaf_size
 *         readonly bint is_leaf_node             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self,
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_12is_leaf_node_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_12is_leaf_node_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_12is_leaf_node___get__(((struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_12is_leaf_node___get__(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->is_leaf_node); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1436, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Float64ClosedNeitherIntervalNode.is_leaf_node.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_21__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_21__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_20__reduce_cython__(((struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_20__reduce_cython__(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  int __pyx_t_16;
  int __pyx_t_17;
  int __pyx_t_18;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.center_left_indices, self.center_left_values, self.center_right_indices, self.center_right_values, self.indices, self.is_leaf_node, self.leaf_size, self.left, self.left_node, self.max_right, self.min_left, self.n_center, self.n_elements, self.pivot, self.right, self.right_node)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  if (unlikely(!__pyx_v_self->center_left_indices.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_1 = __pyx_memoryview_fromslice(__pyx_v_self->center_left_indices, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_int64_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_int64_t, 0);; if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (unlikely(!__pyx_v_self->center_left_values.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_2 = __pyx_memoryview_fromslice(__pyx_v_self->center_left_values, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_float64_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_float64_t, 0);; if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (unlikely(!__pyx_v_self->center_right_indices.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_3 = __pyx_memoryview_fromslice(__pyx_v_self->center_right_indices, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_int64_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_int64_t, 0);; if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely(!__pyx_v_self->center_right_values.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_4 = __pyx_memoryview_fromslice(__pyx_v_self->center_right_values, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_float64_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_float64_t, 0);; if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (unlikely(!__pyx_v_self->indices.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_5 = __pyx_memoryview_fromslice(__pyx_v_self->indices, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_int64_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_int64_t, 0);; if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyBool_FromLong(__pyx_v_self->is_leaf_node); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->leaf_size); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (unlikely(!__pyx_v_self->left.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_8 = __pyx_memoryview_fromslice(__pyx_v_self->left, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_float64_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_float64_t, 0);; if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = PyFloat_FromDouble(__pyx_v_self->max_right); if (unlikely(!__pyx_t_9)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_10 = PyFloat_FromDouble(__pyx_v_self->min_left); if (unlikely(!__pyx_t_10)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_11 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_center); if (unlikely(!__pyx_t_11)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_12 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_elements); if (unlikely(!__pyx_t_12)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __pyx_t_13 = PyFloat_FromDouble(__pyx_v_self->pivot); if (unlikely(!__pyx_t_13)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  if (unlikely(!__pyx_v_self->right.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_14 = __pyx_memoryview_fromslice(__pyx_v_self->right, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_float64_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_float64_t, 0);; if (unlikely(!__pyx_t_14)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __pyx_t_15 = PyTuple_New(16); if (unlikely(!__pyx_t_15)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_15, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_15, 1, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_15, 2, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_15, 3, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_15, 4, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_15, 5, __pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_15, 6, __pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_15, 7, __pyx_t_8);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->left_node));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self->left_node));
  PyTuple_SET_ITEM(__pyx_t_15, 8, ((PyObject *)__pyx_v_self->left_node));
  __Pyx_GIVEREF(__pyx_t_9);
  PyTuple_SET_ITEM(__pyx_t_15, 9, __pyx_t_9);
  __Pyx_GIVEREF(__pyx_t_10);
  PyTuple_SET_ITEM(__pyx_t_15, 10, __pyx_t_10);
  __Pyx_GIVEREF(__pyx_t_11);
  PyTuple_SET_ITEM(__pyx_t_15, 11, __pyx_t_11);
  __Pyx_GIVEREF(__pyx_t_12);
  PyTuple_SET_ITEM(__pyx_t_15, 12, __pyx_t_12);
  __Pyx_GIVEREF(__pyx_t_13);
  PyTuple_SET_ITEM(__pyx_t_15, 13, __pyx_t_13);
  __Pyx_GIVEREF(__pyx_t_14);
  PyTuple_SET_ITEM(__pyx_t_15, 14, __pyx_t_14);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->right_node));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self->right_node));
  PyTuple_SET_ITEM(__pyx_t_15, 15, ((PyObject *)__pyx_v_self->right_node));
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  __pyx_t_6 = 0;
  __pyx_t_7 = 0;
  __pyx_t_8 = 0;
  __pyx_t_9 = 0;
  __pyx_t_10 = 0;
  __pyx_t_11 = 0;
  __pyx_t_12 = 0;
  __pyx_t_13 = 0;
  __pyx_t_14 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_15);
  __pyx_t_15 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.center_left_indices, self.center_left_values, self.center_right_indices, self.center_right_values, self.indices, self.is_leaf_node, self.leaf_size, self.left, self.left_node, self.max_right, self.min_left, self.n_center, self.n_elements, self.pivot, self.right, self.right_node)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_15 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_15)) __PYX_ERR(2, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __pyx_v__dict = __pyx_t_15;
  __pyx_t_15 = 0;

  /* "(tree fragment)":7
 *     state = (self.center_left_indices, self.center_left_values, self.center_right_indices, self.center_right_values, self.indices, self.is_leaf_node, self.leaf_size, self.left, self.left_node, self.max_right, self.min_left, self.n_center, self.n_elements, self.pivot, self.right, self.right_node)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_16 = (__pyx_v__dict != Py_None);
  __pyx_t_17 = (__pyx_t_16 != 0);
  if (__pyx_t_17) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_15 = PyTuple_New(1); if (unlikely(!__pyx_t_15)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    PyTuple_SET_ITEM(__pyx_t_15, 0, __pyx_v__dict);
    __pyx_t_14 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_15); if (unlikely(!__pyx_t_14)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_14));
    __pyx_t_14 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.left_node is not None or self.right_node is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.center_left_indices, self.center_left_values, self.center_right_indices, self.center_right_values, self.indices, self.is_leaf_node, self.leaf_size, self.left, self.left_node, self.max_right, self.min_left, self.n_center, self.n_elements, self.pivot, self.right, self.right_node)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.left_node is not None or self.right_node is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_Float64ClosedNeitherIntervalNode, (type(self), 0x7a104e5, None), state
 */
  /*else*/ {
    __pyx_t_16 = (((PyObject *)__pyx_v_self->left_node) != Py_None);
    __pyx_t_18 = (__pyx_t_16 != 0);
    if (!__pyx_t_18) {
    } else {
      __pyx_t_17 = __pyx_t_18;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_18 = (((PyObject *)__pyx_v_self->right_node) != Py_None);
    __pyx_t_16 = (__pyx_t_18 != 0);
    __pyx_t_17 = __pyx_t_16;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_17;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.left_node is not None or self.right_node is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Float64ClosedNeitherIntervalNode, (type(self), 0x7a104e5, None), state
 *     else:
 */
  __pyx_t_17 = (__pyx_v_use_setstate != 0);
  if (__pyx_t_17) {

    /* "(tree fragment)":13
 *         use_setstate = self.left_node is not None or self.right_node is not None
 *     if use_setstate:
 *         return __pyx_unpickle_Float64ClosedNeitherIntervalNode, (type(self), 0x7a104e5, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_Float64ClosedNeitherIntervalNode, (type(self), 0x7a104e5, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_14, __pyx_n_s_pyx_unpickle_Float64ClosedNeit); if (unlikely(!__pyx_t_14)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __pyx_t_15 = PyTuple_New(3); if (unlikely(!__pyx_t_15)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_15, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_127993061);
    __Pyx_GIVEREF(__pyx_int_127993061);
    PyTuple_SET_ITEM(__pyx_t_15, 1, __pyx_int_127993061);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_15, 2, Py_None);
    __pyx_t_13 = PyTuple_New(3); if (unlikely(!__pyx_t_13)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __Pyx_GIVEREF(__pyx_t_14);
    PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_t_14);
    __Pyx_GIVEREF(__pyx_t_15);
    PyTuple_SET_ITEM(__pyx_t_13, 1, __pyx_t_15);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_13, 2, __pyx_v_state);
    __pyx_t_14 = 0;
    __pyx_t_15 = 0;
    __pyx_r = __pyx_t_13;
    __pyx_t_13 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.left_node is not None or self.right_node is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Float64ClosedNeitherIntervalNode, (type(self), 0x7a104e5, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_Float64ClosedNeitherIntervalNode, (type(self), 0x7a104e5, None), state
 *     else:
 *         return __pyx_unpickle_Float64ClosedNeitherIntervalNode, (type(self), 0x7a104e5, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Float64ClosedNeitherIntervalNode__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_13, __pyx_n_s_pyx_unpickle_Float64ClosedNeit); if (unlikely(!__pyx_t_13)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __pyx_t_15 = PyTuple_New(3); if (unlikely(!__pyx_t_15)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_15, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_127993061);
    __Pyx_GIVEREF(__pyx_int_127993061);
    PyTuple_SET_ITEM(__pyx_t_15, 1, __pyx_int_127993061);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_15, 2, __pyx_v_state);
    __pyx_t_14 = PyTuple_New(2); if (unlikely(!__pyx_t_14)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_GIVEREF(__pyx_t_13);
    PyTuple_SET_ITEM(__pyx_t_14, 0, __pyx_t_13);
    __Pyx_GIVEREF(__pyx_t_15);
    PyTuple_SET_ITEM(__pyx_t_14, 1, __pyx_t_15);
    __pyx_t_13 = 0;
    __pyx_t_15 = 0;
    __pyx_r = __pyx_t_14;
    __pyx_t_14 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_AddTraceback("pandas._libs.interval.Float64ClosedNeitherIntervalNode.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Float64ClosedNeitherIntervalNode, (type(self), 0x7a104e5, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Float64ClosedNeitherIntervalNode__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_23__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_23__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_22__setstate_cython__(((struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_22__setstate_cython__(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_Float64ClosedNeitherIntervalNode, (type(self), 0x7a104e5, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Float64ClosedNeitherIntervalNode__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(2, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_6pandas_5_libs_8interval___pyx_unpickle_Float64ClosedNeitherIntervalNode__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Float64ClosedNeitherIntervalNode, (type(self), 0x7a104e5, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Float64ClosedNeitherIntervalNode__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Float64ClosedNeitherIntervalNode.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":1608
 *         readonly bint is_leaf_node
 * 
 *     def __init__(self,             # <<<<<<<<<<<<<<
 *                  ndarray[int32_t, ndim=1] left,
 *                  ndarray[int32_t, ndim=1] right,
 */

/* Python wrapper */
static int __pyx_pw_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyArrayObject *__pyx_v_left = 0;
  PyArrayObject *__pyx_v_right = 0;
  PyArrayObject *__pyx_v_indices = 0;
  __pyx_t_5numpy_int64_t __pyx_v_leaf_size;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_left,&__pyx_n_s_right,&__pyx_n_s_indices,&__pyx_n_s_leaf_size,0};
    PyObject* values[4] = {0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_left)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_right)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 4, 4, 1); __PYX_ERR(1, 1608, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_indices)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 4, 4, 2); __PYX_ERR(1, 1608, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_leaf_size)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 4, 4, 3); __PYX_ERR(1, 1608, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(1, 1608, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 4) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
    }
    __pyx_v_left = ((PyArrayObject *)values[0]);
    __pyx_v_right = ((PyArrayObject *)values[1]);
    __pyx_v_indices = ((PyArrayObject *)values[2]);
    __pyx_v_leaf_size = __Pyx_PyInt_As_npy_int64(values[3]); if (unlikely((__pyx_v_leaf_size == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(1, 1612, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 4, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 1608, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Int32ClosedLeftIntervalNode.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_left), __pyx_ptype_5numpy_ndarray, 1, "left", 0))) __PYX_ERR(1, 1609, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_right), __pyx_ptype_5numpy_ndarray, 1, "right", 0))) __PYX_ERR(1, 1610, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_indices), __pyx_ptype_5numpy_ndarray, 1, "indices", 0))) __PYX_ERR(1, 1611, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode___init__(((struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode *)__pyx_v_self), __pyx_v_left, __pyx_v_right, __pyx_v_indices, __pyx_v_leaf_size);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode___init__(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode *__pyx_v_self, PyArrayObject *__pyx_v_left, PyArrayObject *__pyx_v_right, PyArrayObject *__pyx_v_indices, __pyx_t_5numpy_int64_t __pyx_v_leaf_size) {
  PyObject *__pyx_v_left_set = NULL;
  PyObject *__pyx_v_right_set = NULL;
  PyObject *__pyx_v_center_set = NULL;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_indices;
  __Pyx_Buffer __pyx_pybuffer_indices;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_left;
  __Pyx_Buffer __pyx_pybuffer_left;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_right;
  __Pyx_Buffer __pyx_pybuffer_right;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __pyx_t_5numpy_int32_t __pyx_t_6;
  __Pyx_memviewslice __pyx_t_7 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_8 = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  __Pyx_memviewslice __pyx_t_11 = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *(*__pyx_t_12)(PyObject *);
  size_t __pyx_t_13;
  __Pyx_RefNannySetupContext("__init__", 0);
  __pyx_pybuffer_left.pybuffer.buf = NULL;
  __pyx_pybuffer_left.refcount = 0;
  __pyx_pybuffernd_left.data = NULL;
  __pyx_pybuffernd_left.rcbuffer = &__pyx_pybuffer_left;
  __pyx_pybuffer_right.pybuffer.buf = NULL;
  __pyx_pybuffer_right.refcount = 0;
  __pyx_pybuffernd_right.data = NULL;
  __pyx_pybuffernd_right.rcbuffer = &__pyx_pybuffer_right;
  __pyx_pybuffer_indices.pybuffer.buf = NULL;
  __pyx_pybuffer_indices.refcount = 0;
  __pyx_pybuffernd_indices.data = NULL;
  __pyx_pybuffernd_indices.rcbuffer = &__pyx_pybuffer_indices;
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_left.rcbuffer->pybuffer, (PyObject*)__pyx_v_left, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int32_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 1608, __pyx_L1_error)
  }
  __pyx_pybuffernd_left.diminfo[0].strides = __pyx_pybuffernd_left.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_left.diminfo[0].shape = __pyx_pybuffernd_left.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_right.rcbuffer->pybuffer, (PyObject*)__pyx_v_right, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int32_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 1608, __pyx_L1_error)
  }
  __pyx_pybuffernd_right.diminfo[0].strides = __pyx_pybuffernd_right.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_right.diminfo[0].shape = __pyx_pybuffernd_right.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_indices.rcbuffer->pybuffer, (PyObject*)__pyx_v_indices, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 1608, __pyx_L1_error)
  }
  __pyx_pybuffernd_indices.diminfo[0].strides = __pyx_pybuffernd_indices.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_indices.diminfo[0].shape = __pyx_pybuffernd_indices.rcbuffer->pybuffer.shape[0];

  /* "pandas/_libs/intervaltree.pxi":1614
 *                  int64_t leaf_size):
 * 
 *         self.n_elements = len(left)             # <<<<<<<<<<<<<<
 *         self.leaf_size = leaf_size
 * 
 */
  __pyx_t_1 = PyObject_Length(((PyObject *)__pyx_v_left)); if (unlikely(__pyx_t_1 == ((Py_ssize_t)-1))) __PYX_ERR(1, 1614, __pyx_L1_error)
  __pyx_v_self->n_elements = __pyx_t_1;

  /* "pandas/_libs/intervaltree.pxi":1615
 * 
 *         self.n_elements = len(left)
 *         self.leaf_size = leaf_size             # <<<<<<<<<<<<<<
 * 
 *         # min_left and min_right are used to speed-up query by skipping
 */
  __pyx_v_self->leaf_size = __pyx_v_leaf_size;

  /* "pandas/_libs/intervaltree.pxi":1620
 *         # query on sub-nodes. If this node has size 0, query is cheap,
 *         # so these values don't matter.
 *         if left.size > 0:             # <<<<<<<<<<<<<<
 *             self.min_left = left.min()
 *             self.max_right = right.max()
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_left), __pyx_n_s_size); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1620, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyObject_RichCompare(__pyx_t_2, __pyx_int_0, Py_GT); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1620, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(1, 1620, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_4) {

    /* "pandas/_libs/intervaltree.pxi":1621
 *         # so these values don't matter.
 *         if left.size > 0:
 *             self.min_left = left.min()             # <<<<<<<<<<<<<<
 *             self.max_right = right.max()
 *         else:
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_left), __pyx_n_s_min); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1621, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1621, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_6 = __Pyx_PyInt_As_npy_int32(__pyx_t_3); if (unlikely((__pyx_t_6 == ((npy_int32)-1)) && PyErr_Occurred())) __PYX_ERR(1, 1621, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_self->min_left = __pyx_t_6;

    /* "pandas/_libs/intervaltree.pxi":1622
 *         if left.size > 0:
 *             self.min_left = left.min()
 *             self.max_right = right.max()             # <<<<<<<<<<<<<<
 *         else:
 *             self.min_left = 0
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_right), __pyx_n_s_max); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1622, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1622, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_6 = __Pyx_PyInt_As_npy_int32(__pyx_t_3); if (unlikely((__pyx_t_6 == ((npy_int32)-1)) && PyErr_Occurred())) __PYX_ERR(1, 1622, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_self->max_right = __pyx_t_6;

    /* "pandas/_libs/intervaltree.pxi":1620
 *         # query on sub-nodes. If this node has size 0, query is cheap,
 *         # so these values don't matter.
 *         if left.size > 0:             # <<<<<<<<<<<<<<
 *             self.min_left = left.min()
 *             self.max_right = right.max()
 */
    goto __pyx_L3;
  }

  /* "pandas/_libs/intervaltree.pxi":1624
 *             self.max_right = right.max()
 *         else:
 *             self.min_left = 0             # <<<<<<<<<<<<<<
 *             self.max_right = 0
 * 
 */
  /*else*/ {
    __pyx_v_self->min_left = 0;

    /* "pandas/_libs/intervaltree.pxi":1625
 *         else:
 *             self.min_left = 0
 *             self.max_right = 0             # <<<<<<<<<<<<<<
 * 
 *         if self.n_elements <= leaf_size:
 */
    __pyx_v_self->max_right = 0;
  }
  __pyx_L3:;

  /* "pandas/_libs/intervaltree.pxi":1627
 *             self.max_right = 0
 * 
 *         if self.n_elements <= leaf_size:             # <<<<<<<<<<<<<<
 *             # make this a terminal (leaf) node
 *             self.is_leaf_node = True
 */
  __pyx_t_4 = ((__pyx_v_self->n_elements <= __pyx_v_leaf_size) != 0);
  if (__pyx_t_4) {

    /* "pandas/_libs/intervaltree.pxi":1629
 *         if self.n_elements <= leaf_size:
 *             # make this a terminal (leaf) node
 *             self.is_leaf_node = True             # <<<<<<<<<<<<<<
 *             self.left = left
 *             self.right = right
 */
    __pyx_v_self->is_leaf_node = 1;

    /* "pandas/_libs/intervaltree.pxi":1630
 *             # make this a terminal (leaf) node
 *             self.is_leaf_node = True
 *             self.left = left             # <<<<<<<<<<<<<<
 *             self.right = right
 *             self.indices = indices
 */
    __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int32_t(((PyObject *)__pyx_v_left), PyBUF_WRITABLE); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(1, 1630, __pyx_L1_error)
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->left, 0);
    __pyx_v_self->left = __pyx_t_7;
    __pyx_t_7.memview = NULL;
    __pyx_t_7.data = NULL;

    /* "pandas/_libs/intervaltree.pxi":1631
 *             self.is_leaf_node = True
 *             self.left = left
 *             self.right = right             # <<<<<<<<<<<<<<
 *             self.indices = indices
 *             self.n_center = 0
 */
    __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int32_t(((PyObject *)__pyx_v_right), PyBUF_WRITABLE); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(1, 1631, __pyx_L1_error)
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->right, 0);
    __pyx_v_self->right = __pyx_t_7;
    __pyx_t_7.memview = NULL;
    __pyx_t_7.data = NULL;

    /* "pandas/_libs/intervaltree.pxi":1632
 *             self.left = left
 *             self.right = right
 *             self.indices = indices             # <<<<<<<<<<<<<<
 *             self.n_center = 0
 *         else:
 */
    __pyx_t_8 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(((PyObject *)__pyx_v_indices), PyBUF_WRITABLE); if (unlikely(!__pyx_t_8.memview)) __PYX_ERR(1, 1632, __pyx_L1_error)
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->indices, 0);
    __pyx_v_self->indices = __pyx_t_8;
    __pyx_t_8.memview = NULL;
    __pyx_t_8.data = NULL;

    /* "pandas/_libs/intervaltree.pxi":1633
 *             self.right = right
 *             self.indices = indices
 *             self.n_center = 0             # <<<<<<<<<<<<<<
 *         else:
 *             # calculate a pivot so we can create child nodes
 */
    __pyx_v_self->n_center = 0;

    /* "pandas/_libs/intervaltree.pxi":1627
 *             self.max_right = 0
 * 
 *         if self.n_elements <= leaf_size:             # <<<<<<<<<<<<<<
 *             # make this a terminal (leaf) node
 *             self.is_leaf_node = True
 */
    goto __pyx_L4;
  }

  /* "pandas/_libs/intervaltree.pxi":1636
 *         else:
 *             # calculate a pivot so we can create child nodes
 *             self.is_leaf_node = False             # <<<<<<<<<<<<<<
 *             self.pivot = np.median(left / 2 + right / 2)
 *             left_set, right_set, center_set = self.classify_intervals(
 */
  /*else*/ {
    __pyx_v_self->is_leaf_node = 0;

    /* "pandas/_libs/intervaltree.pxi":1637
 *             # calculate a pivot so we can create child nodes
 *             self.is_leaf_node = False
 *             self.pivot = np.median(left / 2 + right / 2)             # <<<<<<<<<<<<<<
 *             left_set, right_set, center_set = self.classify_intervals(
 *                 left, right)
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1637, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_median); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1637, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyNumber_Divide(((PyObject *)__pyx_v_left), __pyx_int_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1637, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_9 = __Pyx_PyNumber_Divide(((PyObject *)__pyx_v_right), __pyx_int_2); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 1637, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_10 = PyNumber_Add(__pyx_t_2, __pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 1637, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_9 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_3 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_9, __pyx_t_10) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_10);
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1637, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_6 = __Pyx_PyInt_As_npy_int32(__pyx_t_3); if (unlikely((__pyx_t_6 == ((npy_int32)-1)) && PyErr_Occurred())) __PYX_ERR(1, 1637, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_self->pivot = __pyx_t_6;

    /* "pandas/_libs/intervaltree.pxi":1639
 *             self.pivot = np.median(left / 2 + right / 2)
 *             left_set, right_set, center_set = self.classify_intervals(
 *                 left, right)             # <<<<<<<<<<<<<<
 * 
 *             self.left_node = self.new_child_node(left, right,
 */
    __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int32_t(((PyObject *)__pyx_v_left), PyBUF_WRITABLE); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(1, 1639, __pyx_L1_error)
    __pyx_t_11 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int32_t(((PyObject *)__pyx_v_right), PyBUF_WRITABLE); if (unlikely(!__pyx_t_11.memview)) __PYX_ERR(1, 1639, __pyx_L1_error)

    /* "pandas/_libs/intervaltree.pxi":1638
 *             self.is_leaf_node = False
 *             self.pivot = np.median(left / 2 + right / 2)
 *             left_set, right_set, center_set = self.classify_intervals(             # <<<<<<<<<<<<<<
 *                 left, right)
 * 
 */
    __pyx_t_3 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode *)__pyx_v_self->__pyx_vtab)->classify_intervals(__pyx_v_self, __pyx_t_7, __pyx_t_11); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1638, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __PYX_XDEC_MEMVIEW(&__pyx_t_7, 1);
    __pyx_t_7.memview = NULL;
    __pyx_t_7.data = NULL;
    __PYX_XDEC_MEMVIEW(&__pyx_t_11, 1);
    __pyx_t_11.memview = NULL;
    __pyx_t_11.data = NULL;
    if ((likely(PyTuple_CheckExact(__pyx_t_3))) || (PyList_CheckExact(__pyx_t_3))) {
      PyObject* sequence = __pyx_t_3;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 3)) {
        if (size > 3) __Pyx_RaiseTooManyValuesError(3);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(1, 1638, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_5 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_10 = PyTuple_GET_ITEM(sequence, 1); 
        __pyx_t_9 = PyTuple_GET_ITEM(sequence, 2); 
      } else {
        __pyx_t_5 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_10 = PyList_GET_ITEM(sequence, 1); 
        __pyx_t_9 = PyList_GET_ITEM(sequence, 2); 
      }
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_10);
      __Pyx_INCREF(__pyx_t_9);
      #else
      __pyx_t_5 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1638, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_10 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 1638, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_9 = PySequence_ITEM(sequence, 2); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 1638, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      #endif
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_2 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1638, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_12 = Py_TYPE(__pyx_t_2)->tp_iternext;
      index = 0; __pyx_t_5 = __pyx_t_12(__pyx_t_2); if (unlikely(!__pyx_t_5)) goto __pyx_L5_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_5);
      index = 1; __pyx_t_10 = __pyx_t_12(__pyx_t_2); if (unlikely(!__pyx_t_10)) goto __pyx_L5_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_10);
      index = 2; __pyx_t_9 = __pyx_t_12(__pyx_t_2); if (unlikely(!__pyx_t_9)) goto __pyx_L5_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_9);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_12(__pyx_t_2), 3) < 0) __PYX_ERR(1, 1638, __pyx_L1_error)
      __pyx_t_12 = NULL;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      goto __pyx_L6_unpacking_done;
      __pyx_L5_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_12 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(1, 1638, __pyx_L1_error)
      __pyx_L6_unpacking_done:;
    }
    __pyx_v_left_set = __pyx_t_5;
    __pyx_t_5 = 0;
    __pyx_v_right_set = __pyx_t_10;
    __pyx_t_10 = 0;
    __pyx_v_center_set = __pyx_t_9;
    __pyx_t_9 = 0;

    /* "pandas/_libs/intervaltree.pxi":1642
 * 
 *             self.left_node = self.new_child_node(left, right,
 *                                                  indices, left_set)             # <<<<<<<<<<<<<<
 *             self.right_node = self.new_child_node(left, right,
 *                                                   indices, right_set)
 */
    if (!(likely(((__pyx_v_left_set) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_left_set, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(1, 1642, __pyx_L1_error)

    /* "pandas/_libs/intervaltree.pxi":1641
 *                 left, right)
 * 
 *             self.left_node = self.new_child_node(left, right,             # <<<<<<<<<<<<<<
 *                                                  indices, left_set)
 *             self.right_node = self.new_child_node(left, right,
 */
    __pyx_t_3 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode *)__pyx_v_self->__pyx_vtab)->new_child_node(__pyx_v_self, ((PyArrayObject *)__pyx_v_left), ((PyArrayObject *)__pyx_v_right), ((PyArrayObject *)__pyx_v_indices), ((PyArrayObject *)__pyx_v_left_set)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1641, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode))))) __PYX_ERR(1, 1641, __pyx_L1_error)
    __Pyx_GIVEREF(__pyx_t_3);
    __Pyx_GOTREF(__pyx_v_self->left_node);
    __Pyx_DECREF(((PyObject *)__pyx_v_self->left_node));
    __pyx_v_self->left_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode *)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "pandas/_libs/intervaltree.pxi":1644
 *                                                  indices, left_set)
 *             self.right_node = self.new_child_node(left, right,
 *                                                   indices, right_set)             # <<<<<<<<<<<<<<
 * 
 *             self.center_left_values, self.center_left_indices = \
 */
    if (!(likely(((__pyx_v_right_set) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_right_set, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(1, 1644, __pyx_L1_error)

    /* "pandas/_libs/intervaltree.pxi":1643
 *             self.left_node = self.new_child_node(left, right,
 *                                                  indices, left_set)
 *             self.right_node = self.new_child_node(left, right,             # <<<<<<<<<<<<<<
 *                                                   indices, right_set)
 * 
 */
    __pyx_t_3 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode *)__pyx_v_self->__pyx_vtab)->new_child_node(__pyx_v_self, ((PyArrayObject *)__pyx_v_left), ((PyArrayObject *)__pyx_v_right), ((PyArrayObject *)__pyx_v_indices), ((PyArrayObject *)__pyx_v_right_set)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1643, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode))))) __PYX_ERR(1, 1643, __pyx_L1_error)
    __Pyx_GIVEREF(__pyx_t_3);
    __Pyx_GOTREF(__pyx_v_self->right_node);
    __Pyx_DECREF(((PyObject *)__pyx_v_self->right_node));
    __pyx_v_self->right_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode *)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "pandas/_libs/intervaltree.pxi":1647
 * 
 *             self.center_left_values, self.center_left_indices = \
 *                 sort_values_and_indices(left, indices, center_set)             # <<<<<<<<<<<<<<
 *             self.center_right_values, self.center_right_indices = \
 *                 sort_values_and_indices(right, indices, center_set)
 */
    __pyx_t_3 = __pyx_f_6pandas_5_libs_8interval_sort_values_and_indices(((PyObject *)__pyx_v_left), ((PyObject *)__pyx_v_indices), __pyx_v_center_set); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1647, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if ((likely(PyTuple_CheckExact(__pyx_t_3))) || (PyList_CheckExact(__pyx_t_3))) {
      PyObject* sequence = __pyx_t_3;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(1, 1646, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_9 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_10 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_9 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_10 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_9);
      __Pyx_INCREF(__pyx_t_10);
      #else
      __pyx_t_9 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 1646, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_10 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 1646, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      #endif
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_5 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1646, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_12 = Py_TYPE(__pyx_t_5)->tp_iternext;
      index = 0; __pyx_t_9 = __pyx_t_12(__pyx_t_5); if (unlikely(!__pyx_t_9)) goto __pyx_L7_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_9);
      index = 1; __pyx_t_10 = __pyx_t_12(__pyx_t_5); if (unlikely(!__pyx_t_10)) goto __pyx_L7_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_10);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_12(__pyx_t_5), 2) < 0) __PYX_ERR(1, 1646, __pyx_L1_error)
      __pyx_t_12 = NULL;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      goto __pyx_L8_unpacking_done;
      __pyx_L7_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_12 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(1, 1646, __pyx_L1_error)
      __pyx_L8_unpacking_done:;
    }

    /* "pandas/_libs/intervaltree.pxi":1646
 *                                                   indices, right_set)
 * 
 *             self.center_left_values, self.center_left_indices = \             # <<<<<<<<<<<<<<
 *                 sort_values_and_indices(left, indices, center_set)
 *             self.center_right_values, self.center_right_indices = \
 */
    __pyx_t_11 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int32_t(__pyx_t_9, PyBUF_WRITABLE); if (unlikely(!__pyx_t_11.memview)) __PYX_ERR(1, 1646, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_8 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_t_10, PyBUF_WRITABLE); if (unlikely(!__pyx_t_8.memview)) __PYX_ERR(1, 1646, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->center_left_values, 0);
    __pyx_v_self->center_left_values = __pyx_t_11;
    __pyx_t_11.memview = NULL;
    __pyx_t_11.data = NULL;
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->center_left_indices, 0);
    __pyx_v_self->center_left_indices = __pyx_t_8;
    __pyx_t_8.memview = NULL;
    __pyx_t_8.data = NULL;

    /* "pandas/_libs/intervaltree.pxi":1649
 *                 sort_values_and_indices(left, indices, center_set)
 *             self.center_right_values, self.center_right_indices = \
 *                 sort_values_and_indices(right, indices, center_set)             # <<<<<<<<<<<<<<
 *             self.n_center = len(self.center_left_indices)
 * 
 */
    __pyx_t_3 = __pyx_f_6pandas_5_libs_8interval_sort_values_and_indices(((PyObject *)__pyx_v_right), ((PyObject *)__pyx_v_indices), __pyx_v_center_set); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1649, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if ((likely(PyTuple_CheckExact(__pyx_t_3))) || (PyList_CheckExact(__pyx_t_3))) {
      PyObject* sequence = __pyx_t_3;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(1, 1648, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_10 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_9 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_10 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_9 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_10);
      __Pyx_INCREF(__pyx_t_9);
      #else
      __pyx_t_10 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 1648, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_9 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 1648, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      #endif
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_5 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1648, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_12 = Py_TYPE(__pyx_t_5)->tp_iternext;
      index = 0; __pyx_t_10 = __pyx_t_12(__pyx_t_5); if (unlikely(!__pyx_t_10)) goto __pyx_L9_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_10);
      index = 1; __pyx_t_9 = __pyx_t_12(__pyx_t_5); if (unlikely(!__pyx_t_9)) goto __pyx_L9_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_9);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_12(__pyx_t_5), 2) < 0) __PYX_ERR(1, 1648, __pyx_L1_error)
      __pyx_t_12 = NULL;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      goto __pyx_L10_unpacking_done;
      __pyx_L9_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_12 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(1, 1648, __pyx_L1_error)
      __pyx_L10_unpacking_done:;
    }

    /* "pandas/_libs/intervaltree.pxi":1648
 *             self.center_left_values, self.center_left_indices = \
 *                 sort_values_and_indices(left, indices, center_set)
 *             self.center_right_values, self.center_right_indices = \             # <<<<<<<<<<<<<<
 *                 sort_values_and_indices(right, indices, center_set)
 *             self.n_center = len(self.center_left_indices)
 */
    __pyx_t_11 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int32_t(__pyx_t_10, PyBUF_WRITABLE); if (unlikely(!__pyx_t_11.memview)) __PYX_ERR(1, 1648, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_t_8 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_t_9, PyBUF_WRITABLE); if (unlikely(!__pyx_t_8.memview)) __PYX_ERR(1, 1648, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->center_right_values, 0);
    __pyx_v_self->center_right_values = __pyx_t_11;
    __pyx_t_11.memview = NULL;
    __pyx_t_11.data = NULL;
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->center_right_indices, 0);
    __pyx_v_self->center_right_indices = __pyx_t_8;
    __pyx_t_8.memview = NULL;
    __pyx_t_8.data = NULL;

    /* "pandas/_libs/intervaltree.pxi":1650
 *             self.center_right_values, self.center_right_indices = \
 *                 sort_values_and_indices(right, indices, center_set)
 *             self.n_center = len(self.center_left_indices)             # <<<<<<<<<<<<<<
 * 
 *     @cython.wraparound(False)
 */
    if (unlikely(!__pyx_v_self->center_left_indices.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(1, 1650, __pyx_L1_error)}
    __pyx_t_13 = __Pyx_MemoryView_Len(__pyx_v_self->center_left_indices); 
    __pyx_v_self->n_center = __pyx_t_13;
  }
  __pyx_L4:;

  /* "pandas/_libs/intervaltree.pxi":1608
 *         readonly bint is_leaf_node
 * 
 *     def __init__(self,             # <<<<<<<<<<<<<<
 *                  ndarray[int32_t, ndim=1] left,
 *                  ndarray[int32_t, ndim=1] right,
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __PYX_XDEC_MEMVIEW(&__pyx_t_7, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_8, 1);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __PYX_XDEC_MEMVIEW(&__pyx_t_11, 1);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indices.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_left.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_right.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("pandas._libs.interval.Int32ClosedLeftIntervalNode.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indices.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_left.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_right.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF(__pyx_v_left_set);
  __Pyx_XDECREF(__pyx_v_right_set);
  __Pyx_XDECREF(__pyx_v_center_set);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":1654
 *     @cython.wraparound(False)
 *     @cython.boundscheck(False)
 *     cdef classify_intervals(self, int32_t[:] left, int32_t[:] right):             # <<<<<<<<<<<<<<
 *         """Classify the given intervals based upon whether they fall to the
 *         left, right, or overlap with this node's pivot.
 */

static PyObject *__pyx_f_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_classify_intervals(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode *__pyx_v_self, __Pyx_memviewslice __pyx_v_left, __Pyx_memviewslice __pyx_v_right) {
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_left_ind = 0;
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_right_ind = 0;
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_overlapping_ind = 0;
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __pyx_t_5numpy_int64_t __pyx_t_2;
  __pyx_t_5numpy_int64_t __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  int __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  __Pyx_RefNannySetupContext("classify_intervals", 0);

  /* "pandas/_libs/intervaltree.pxi":1662
 *             Py_ssize_t i
 * 
 *         left_ind = Int64Vector()             # <<<<<<<<<<<<<<
 *         right_ind = Int64Vector()
 *         overlapping_ind = Int64Vector()
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1662, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_left_ind = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":1663
 * 
 *         left_ind = Int64Vector()
 *         right_ind = Int64Vector()             # <<<<<<<<<<<<<<
 *         overlapping_ind = Int64Vector()
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1663, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_right_ind = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":1664
 *         left_ind = Int64Vector()
 *         right_ind = Int64Vector()
 *         overlapping_ind = Int64Vector()             # <<<<<<<<<<<<<<
 * 
 *         for i in range(self.n_elements):
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1664, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_overlapping_ind = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":1666
 *         overlapping_ind = Int64Vector()
 * 
 *         for i in range(self.n_elements):             # <<<<<<<<<<<<<<
 *             if right[i] <= self.pivot:
 *                 left_ind.append(i)
 */
  __pyx_t_2 = __pyx_v_self->n_elements;
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "pandas/_libs/intervaltree.pxi":1667
 * 
 *         for i in range(self.n_elements):
 *             if right[i] <= self.pivot:             # <<<<<<<<<<<<<<
 *                 left_ind.append(i)
 *             elif self.pivot < left[i]:
 */
    __pyx_t_5 = __pyx_v_i;
    __pyx_t_6 = (((*((__pyx_t_5numpy_int32_t *) ( /* dim=0 */ (__pyx_v_right.data + __pyx_t_5 * __pyx_v_right.strides[0]) ))) <= __pyx_v_self->pivot) != 0);
    if (__pyx_t_6) {

      /* "pandas/_libs/intervaltree.pxi":1668
 *         for i in range(self.n_elements):
 *             if right[i] <= self.pivot:
 *                 left_ind.append(i)             # <<<<<<<<<<<<<<
 *             elif self.pivot < left[i]:
 *                 right_ind.append(i)
 */
      ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_left_ind->__pyx_vtab)->append(__pyx_v_left_ind, __pyx_v_i);

      /* "pandas/_libs/intervaltree.pxi":1667
 * 
 *         for i in range(self.n_elements):
 *             if right[i] <= self.pivot:             # <<<<<<<<<<<<<<
 *                 left_ind.append(i)
 *             elif self.pivot < left[i]:
 */
      goto __pyx_L5;
    }

    /* "pandas/_libs/intervaltree.pxi":1669
 *             if right[i] <= self.pivot:
 *                 left_ind.append(i)
 *             elif self.pivot < left[i]:             # <<<<<<<<<<<<<<
 *                 right_ind.append(i)
 *             else:
 */
    __pyx_t_7 = __pyx_v_i;
    __pyx_t_6 = ((__pyx_v_self->pivot < (*((__pyx_t_5numpy_int32_t *) ( /* dim=0 */ (__pyx_v_left.data + __pyx_t_7 * __pyx_v_left.strides[0]) )))) != 0);
    if (__pyx_t_6) {

      /* "pandas/_libs/intervaltree.pxi":1670
 *                 left_ind.append(i)
 *             elif self.pivot < left[i]:
 *                 right_ind.append(i)             # <<<<<<<<<<<<<<
 *             else:
 *                 overlapping_ind.append(i)
 */
      ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_right_ind->__pyx_vtab)->append(__pyx_v_right_ind, __pyx_v_i);

      /* "pandas/_libs/intervaltree.pxi":1669
 *             if right[i] <= self.pivot:
 *                 left_ind.append(i)
 *             elif self.pivot < left[i]:             # <<<<<<<<<<<<<<
 *                 right_ind.append(i)
 *             else:
 */
      goto __pyx_L5;
    }

    /* "pandas/_libs/intervaltree.pxi":1672
 *                 right_ind.append(i)
 *             else:
 *                 overlapping_ind.append(i)             # <<<<<<<<<<<<<<
 * 
 *         return (left_ind.to_array(),
 */
    /*else*/ {
      ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_overlapping_ind->__pyx_vtab)->append(__pyx_v_overlapping_ind, __pyx_v_i);
    }
    __pyx_L5:;
  }

  /* "pandas/_libs/intervaltree.pxi":1674
 *                 overlapping_ind.append(i)
 * 
 *         return (left_ind.to_array(),             # <<<<<<<<<<<<<<
 *                 right_ind.to_array(),
 *                 overlapping_ind.to_array())
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_left_ind->__pyx_vtab)->to_array(__pyx_v_left_ind, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1674, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "pandas/_libs/intervaltree.pxi":1675
 * 
 *         return (left_ind.to_array(),
 *                 right_ind.to_array(),             # <<<<<<<<<<<<<<
 *                 overlapping_ind.to_array())
 * 
 */
  __pyx_t_8 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_right_ind->__pyx_vtab)->to_array(__pyx_v_right_ind, 0); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 1675, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);

  /* "pandas/_libs/intervaltree.pxi":1676
 *         return (left_ind.to_array(),
 *                 right_ind.to_array(),
 *                 overlapping_ind.to_array())             # <<<<<<<<<<<<<<
 * 
 *     cdef new_child_node(self,
 */
  __pyx_t_9 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_overlapping_ind->__pyx_vtab)->to_array(__pyx_v_overlapping_ind, 0); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 1676, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);

  /* "pandas/_libs/intervaltree.pxi":1674
 *                 overlapping_ind.append(i)
 * 
 *         return (left_ind.to_array(),             # <<<<<<<<<<<<<<
 *                 right_ind.to_array(),
 *                 overlapping_ind.to_array())
 */
  __pyx_t_10 = PyTuple_New(3); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 1674, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_t_8);
  __Pyx_GIVEREF(__pyx_t_9);
  PyTuple_SET_ITEM(__pyx_t_10, 2, __pyx_t_9);
  __pyx_t_1 = 0;
  __pyx_t_8 = 0;
  __pyx_t_9 = 0;
  __pyx_r = __pyx_t_10;
  __pyx_t_10 = 0;
  goto __pyx_L0;

  /* "pandas/_libs/intervaltree.pxi":1654
 *     @cython.wraparound(False)
 *     @cython.boundscheck(False)
 *     cdef classify_intervals(self, int32_t[:] left, int32_t[:] right):             # <<<<<<<<<<<<<<
 *         """Classify the given intervals based upon whether they fall to the
 *         left, right, or overlap with this node's pivot.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("pandas._libs.interval.Int32ClosedLeftIntervalNode.classify_intervals", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_left_ind);
  __Pyx_XDECREF((PyObject *)__pyx_v_right_ind);
  __Pyx_XDECREF((PyObject *)__pyx_v_overlapping_ind);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":1678
 *                 overlapping_ind.to_array())
 * 
 *     cdef new_child_node(self,             # <<<<<<<<<<<<<<
 *                         ndarray[int32_t, ndim=1] left,
 *                         ndarray[int32_t, ndim=1] right,
 */

static PyObject *__pyx_f_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_new_child_node(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode *__pyx_v_self, PyArrayObject *__pyx_v_left, PyArrayObject *__pyx_v_right, PyArrayObject *__pyx_v_indices, PyArrayObject *__pyx_v_subset) {
  __Pyx_LocalBuf_ND __pyx_pybuffernd_indices;
  __Pyx_Buffer __pyx_pybuffer_indices;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_left;
  __Pyx_Buffer __pyx_pybuffer_left;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_right;
  __Pyx_Buffer __pyx_pybuffer_right;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_subset;
  __Pyx_Buffer __pyx_pybuffer_subset;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyArrayObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyArrayObject *__pyx_t_7 = NULL;
  PyArrayObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  __Pyx_RefNannySetupContext("new_child_node", 0);
  __Pyx_INCREF((PyObject *)__pyx_v_left);
  __Pyx_INCREF((PyObject *)__pyx_v_right);
  __Pyx_INCREF((PyObject *)__pyx_v_indices);
  __pyx_pybuffer_left.pybuffer.buf = NULL;
  __pyx_pybuffer_left.refcount = 0;
  __pyx_pybuffernd_left.data = NULL;
  __pyx_pybuffernd_left.rcbuffer = &__pyx_pybuffer_left;
  __pyx_pybuffer_right.pybuffer.buf = NULL;
  __pyx_pybuffer_right.refcount = 0;
  __pyx_pybuffernd_right.data = NULL;
  __pyx_pybuffernd_right.rcbuffer = &__pyx_pybuffer_right;
  __pyx_pybuffer_indices.pybuffer.buf = NULL;
  __pyx_pybuffer_indices.refcount = 0;
  __pyx_pybuffernd_indices.data = NULL;
  __pyx_pybuffernd_indices.rcbuffer = &__pyx_pybuffer_indices;
  __pyx_pybuffer_subset.pybuffer.buf = NULL;
  __pyx_pybuffer_subset.refcount = 0;
  __pyx_pybuffernd_subset.data = NULL;
  __pyx_pybuffernd_subset.rcbuffer = &__pyx_pybuffer_subset;
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_left.rcbuffer->pybuffer, (PyObject*)__pyx_v_left, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int32_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 1678, __pyx_L1_error)
  }
  __pyx_pybuffernd_left.diminfo[0].strides = __pyx_pybuffernd_left.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_left.diminfo[0].shape = __pyx_pybuffernd_left.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_right.rcbuffer->pybuffer, (PyObject*)__pyx_v_right, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int32_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 1678, __pyx_L1_error)
  }
  __pyx_pybuffernd_right.diminfo[0].strides = __pyx_pybuffernd_right.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_right.diminfo[0].shape = __pyx_pybuffernd_right.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_indices.rcbuffer->pybuffer, (PyObject*)__pyx_v_indices, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 1678, __pyx_L1_error)
  }
  __pyx_pybuffernd_indices.diminfo[0].strides = __pyx_pybuffernd_indices.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_indices.diminfo[0].shape = __pyx_pybuffernd_indices.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_subset.rcbuffer->pybuffer, (PyObject*)__pyx_v_subset, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 1678, __pyx_L1_error)
  }
  __pyx_pybuffernd_subset.diminfo[0].strides = __pyx_pybuffernd_subset.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_subset.diminfo[0].shape = __pyx_pybuffernd_subset.rcbuffer->pybuffer.shape[0];

  /* "pandas/_libs/intervaltree.pxi":1685
 *         """Create a new child node.
 *         """
 *         left = take(left, subset)             # <<<<<<<<<<<<<<
 *         right = take(right, subset)
 *         indices = take(indices, subset)
 */
  __pyx_t_1 = __pyx_f_6pandas_5_libs_8interval_take(((PyArrayObject *)__pyx_v_left), ((PyArrayObject *)__pyx_v_subset)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1685, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(1, 1685, __pyx_L1_error)
  __pyx_t_2 = ((PyArrayObject *)__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_left.rcbuffer->pybuffer);
    __pyx_t_3 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_left.rcbuffer->pybuffer, (PyObject*)__pyx_t_2, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int32_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack);
    if (unlikely(__pyx_t_3 < 0)) {
      PyErr_Fetch(&__pyx_t_4, &__pyx_t_5, &__pyx_t_6);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_left.rcbuffer->pybuffer, (PyObject*)__pyx_v_left, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int32_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_4); Py_XDECREF(__pyx_t_5); Py_XDECREF(__pyx_t_6);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_4, __pyx_t_5, __pyx_t_6);
      }
      __pyx_t_4 = __pyx_t_5 = __pyx_t_6 = 0;
    }
    __pyx_pybuffernd_left.diminfo[0].strides = __pyx_pybuffernd_left.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_left.diminfo[0].shape = __pyx_pybuffernd_left.rcbuffer->pybuffer.shape[0];
    if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(1, 1685, __pyx_L1_error)
  }
  __pyx_t_2 = 0;
  __Pyx_DECREF_SET(__pyx_v_left, ((PyArrayObject *)__pyx_t_1));
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":1686
 *         """
 *         left = take(left, subset)
 *         right = take(right, subset)             # <<<<<<<<<<<<<<
 *         indices = take(indices, subset)
 *         return Int32ClosedLeftIntervalNode(
 */
  __pyx_t_1 = __pyx_f_6pandas_5_libs_8interval_take(((PyArrayObject *)__pyx_v_right), ((PyArrayObject *)__pyx_v_subset)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1686, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(1, 1686, __pyx_L1_error)
  __pyx_t_7 = ((PyArrayObject *)__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_right.rcbuffer->pybuffer);
    __pyx_t_3 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_right.rcbuffer->pybuffer, (PyObject*)__pyx_t_7, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int32_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack);
    if (unlikely(__pyx_t_3 < 0)) {
      PyErr_Fetch(&__pyx_t_6, &__pyx_t_5, &__pyx_t_4);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_right.rcbuffer->pybuffer, (PyObject*)__pyx_v_right, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int32_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_6); Py_XDECREF(__pyx_t_5); Py_XDECREF(__pyx_t_4);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_6, __pyx_t_5, __pyx_t_4);
      }
      __pyx_t_6 = __pyx_t_5 = __pyx_t_4 = 0;
    }
    __pyx_pybuffernd_right.diminfo[0].strides = __pyx_pybuffernd_right.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_right.diminfo[0].shape = __pyx_pybuffernd_right.rcbuffer->pybuffer.shape[0];
    if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(1, 1686, __pyx_L1_error)
  }
  __pyx_t_7 = 0;
  __Pyx_DECREF_SET(__pyx_v_right, ((PyArrayObject *)__pyx_t_1));
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":1687
 *         left = take(left, subset)
 *         right = take(right, subset)
 *         indices = take(indices, subset)             # <<<<<<<<<<<<<<
 *         return Int32ClosedLeftIntervalNode(
 *             left, right, indices, self.leaf_size)
 */
  __pyx_t_1 = __pyx_f_6pandas_5_libs_8interval_take(((PyArrayObject *)__pyx_v_indices), ((PyArrayObject *)__pyx_v_subset)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1687, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(1, 1687, __pyx_L1_error)
  __pyx_t_8 = ((PyArrayObject *)__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indices.rcbuffer->pybuffer);
    __pyx_t_3 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_indices.rcbuffer->pybuffer, (PyObject*)__pyx_t_8, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack);
    if (unlikely(__pyx_t_3 < 0)) {
      PyErr_Fetch(&__pyx_t_4, &__pyx_t_5, &__pyx_t_6);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_indices.rcbuffer->pybuffer, (PyObject*)__pyx_v_indices, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_4); Py_XDECREF(__pyx_t_5); Py_XDECREF(__pyx_t_6);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_4, __pyx_t_5, __pyx_t_6);
      }
      __pyx_t_4 = __pyx_t_5 = __pyx_t_6 = 0;
    }
    __pyx_pybuffernd_indices.diminfo[0].strides = __pyx_pybuffernd_indices.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_indices.diminfo[0].shape = __pyx_pybuffernd_indices.rcbuffer->pybuffer.shape[0];
    if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(1, 1687, __pyx_L1_error)
  }
  __pyx_t_8 = 0;
  __Pyx_DECREF_SET(__pyx_v_indices, ((PyArrayObject *)__pyx_t_1));
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":1688
 *         right = take(right, subset)
 *         indices = take(indices, subset)
 *         return Int32ClosedLeftIntervalNode(             # <<<<<<<<<<<<<<
 *             left, right, indices, self.leaf_size)
 * 
 */
  __Pyx_XDECREF(__pyx_r);

  /* "pandas/_libs/intervaltree.pxi":1689
 *         indices = take(indices, subset)
 *         return Int32ClosedLeftIntervalNode(
 *             left, right, indices, self.leaf_size)             # <<<<<<<<<<<<<<
 * 
 *     @cython.wraparound(False)
 */
  __pyx_t_1 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->leaf_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1689, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "pandas/_libs/intervaltree.pxi":1688
 *         right = take(right, subset)
 *         indices = take(indices, subset)
 *         return Int32ClosedLeftIntervalNode(             # <<<<<<<<<<<<<<
 *             left, right, indices, self.leaf_size)
 * 
 */
  __pyx_t_9 = PyTuple_New(4); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 1688, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_INCREF(((PyObject *)__pyx_v_left));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_left));
  PyTuple_SET_ITEM(__pyx_t_9, 0, ((PyObject *)__pyx_v_left));
  __Pyx_INCREF(((PyObject *)__pyx_v_right));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_right));
  PyTuple_SET_ITEM(__pyx_t_9, 1, ((PyObject *)__pyx_v_right));
  __Pyx_INCREF(((PyObject *)__pyx_v_indices));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_indices));
  PyTuple_SET_ITEM(__pyx_t_9, 2, ((PyObject *)__pyx_v_indices));
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_9, 3, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode), __pyx_t_9, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1688, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pandas/_libs/intervaltree.pxi":1678
 *                 overlapping_ind.to_array())
 * 
 *     cdef new_child_node(self,             # <<<<<<<<<<<<<<
 *                         ndarray[int32_t, ndim=1] left,
 *                         ndarray[int32_t, ndim=1] right,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_9);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indices.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_left.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_right.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_subset.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("pandas._libs.interval.Int32ClosedLeftIntervalNode.new_child_node", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indices.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_left.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_right.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_subset.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF((PyObject *)__pyx_v_left);
  __Pyx_XDECREF((PyObject *)__pyx_v_right);
  __Pyx_XDECREF((PyObject *)__pyx_v_indices);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":1694
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_2query[] = "Recursively query this node and its sub-nodes for intervals that\n        overlap with the query point.\n        ";
static PyMethodDef __pyx_mdef_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_3query = {"query", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_3query, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_2query};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_args = 0;
  PyObject *__pyx_v_kwargs = 0;
  CYTHON_UNUSED PyObject *__pyx_v_defaults = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fused_cpdef (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_args,&__pyx_n_s_kwargs,&__pyx_n_s_defaults,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_args)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_kwargs)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fused_cpdef", 1, 3, 3, 1); __PYX_ERR(1, 1694, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_defaults)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fused_cpdef", 1, 3, 3, 2); __PYX_ERR(1, 1694, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fused_cpdef") < 0)) __PYX_ERR(1, 1694, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_args = values[0];
    __pyx_v_kwargs = values[1];
    __pyx_v_defaults = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fused_cpdef", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 1694, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Int32ClosedLeftIntervalNode.__pyx_fused_cpdef", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_2query(((struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode *)__pyx_v_signatures), __pyx_v_args, __pyx_v_kwargs, __pyx_v_defaults);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_2query(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode *__pyx_v_signatures, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs, CYTHON_UNUSED PyObject *__pyx_v_defaults) {
  PyObject *__pyx_v_dest_sig = NULL;
  Py_ssize_t __pyx_v_i;
  CYTHON_UNUSED PyTypeObject *__pyx_v_ndarray = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_v_itemsize;
  PyObject *__pyx_v_arg = NULL;
  PyObject *__pyx_v_candidates = NULL;
  PyObject *__pyx_v_sig = NULL;
  int __pyx_v_match_found;
  PyObject *__pyx_v_src_sig = NULL;
  PyObject *__pyx_v_dst_type = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  Py_ssize_t __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  Py_ssize_t __pyx_t_13;
  Py_ssize_t __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  int __pyx_t_16;
  __Pyx_RefNannySetupContext("query", 0);
  __Pyx_INCREF(__pyx_v_kwargs);
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1694, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  PyList_SET_ITEM(__pyx_t_1, 0, Py_None);
  __pyx_v_dest_sig = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_3 = (__pyx_v_kwargs != Py_None);
  __pyx_t_4 = (__pyx_t_3 != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_kwargs); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(1, 1694, __pyx_L1_error)
  __pyx_t_3 = ((!__pyx_t_4) != 0);
  __pyx_t_2 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {
    __Pyx_INCREF(Py_None);
    __Pyx_DECREF_SET(__pyx_v_kwargs, Py_None);
  }
  __pyx_t_1 = ((PyObject *)__Pyx_ImportNumPyArrayTypeIfAvailable()); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1694, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_ndarray = ((PyTypeObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_v_itemsize = -1L;
  if (unlikely(__pyx_v_args == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 1694, __pyx_L1_error)
  }
  __pyx_t_5 = PyTuple_GET_SIZE(((PyObject*)__pyx_v_args)); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(1, 1694, __pyx_L1_error)
  __pyx_t_2 = ((2 < __pyx_t_5) != 0);
  if (__pyx_t_2) {
    if (unlikely(__pyx_v_args == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 1694, __pyx_L1_error)
    }
    __pyx_t_1 = PyTuple_GET_ITEM(((PyObject*)__pyx_v_args), 2);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v_arg = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L6;
  }
  __pyx_t_3 = (__pyx_v_kwargs != Py_None);
  __pyx_t_4 = (__pyx_t_3 != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L7_bool_binop_done;
  }
  if (unlikely(__pyx_v_kwargs == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(1, 1694, __pyx_L1_error)
  }
  __pyx_t_4 = (__Pyx_PyDict_ContainsTF(__pyx_n_s_point, ((PyObject*)__pyx_v_kwargs), Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(1, 1694, __pyx_L1_error)
  __pyx_t_3 = (__pyx_t_4 != 0);
  __pyx_t_2 = __pyx_t_3;
  __pyx_L7_bool_binop_done:;
  if (__pyx_t_2) {
    if (unlikely(__pyx_v_kwargs == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 1694, __pyx_L1_error)
    }
    __pyx_t_1 = __Pyx_PyDict_GetItem(((PyObject*)__pyx_v_kwargs), __pyx_n_s_point); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1694, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_arg = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L6;
  }
  /*else*/ {
    if (unlikely(__pyx_v_args == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
      __PYX_ERR(1, 1694, __pyx_L1_error)
    }
    __pyx_t_5 = PyTuple_GET_SIZE(((PyObject*)__pyx_v_args)); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(1, 1694, __pyx_L1_error)
    __pyx_t_1 = PyInt_FromSsize_t(__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1694, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_6 = PyTuple_New(3); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 1694, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_INCREF(__pyx_int_3);
    __Pyx_GIVEREF(__pyx_int_3);
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_int_3);
    __Pyx_INCREF(__pyx_n_s_s);
    __Pyx_GIVEREF(__pyx_n_s_s);
    PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_n_s_s);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Expected_at_least_d_argument_s_g, __pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1694, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_1); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 1694, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(1, 1694, __pyx_L1_error)
  }
  __pyx_L6:;
  while (1) {
    __pyx_t_2 = PyFloat_Check(__pyx_v_arg); 
    __pyx_t_3 = (__pyx_t_2 != 0);
    if (__pyx_t_3) {
      if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, __pyx_n_s_float64_t, long, 1, __Pyx_PyInt_From_long, 1, 0, 0) < 0)) __PYX_ERR(1, 1694, __pyx_L1_error)
      goto __pyx_L10_break;
    }
    __pyx_t_2 = PyInt_Check(__pyx_v_arg); 
    __pyx_t_4 = (__pyx_t_2 != 0);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_3 = __pyx_t_4;
      goto __pyx_L13_bool_binop_done;
    }
    __pyx_t_4 = PyLong_Check(__pyx_v_arg); 
    __pyx_t_2 = (__pyx_t_4 != 0);
    __pyx_t_3 = __pyx_t_2;
    __pyx_L13_bool_binop_done:;
    __pyx_t_2 = (__pyx_t_3 != 0);
    if (__pyx_t_2) {
      if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, __pyx_n_s_int64_t, long, 1, __Pyx_PyInt_From_long, 1, 0, 0) < 0)) __PYX_ERR(1, 1694, __pyx_L1_error)
      goto __pyx_L10_break;
    }
    if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, Py_None, long, 1, __Pyx_PyInt_From_long, 1, 0, 0) < 0)) __PYX_ERR(1, 1694, __pyx_L1_error)
    goto __pyx_L10_break;
  }
  __pyx_L10_break:;
  __pyx_t_6 = PyList_New(0); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 1694, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_v_candidates = ((PyObject*)__pyx_t_6);
  __pyx_t_6 = 0;
  __pyx_t_5 = 0;
  if (unlikely(((PyObject *)__pyx_v_signatures) == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(1, 1694, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_dict_iterator(((PyObject*)__pyx_v_signatures), 1, ((PyObject *)NULL), (&__pyx_t_7), (&__pyx_t_8)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1694, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __pyx_t_6 = __pyx_t_1;
  __pyx_t_1 = 0;
  while (1) {
    __pyx_t_9 = __Pyx_dict_iter_next(__pyx_t_6, __pyx_t_7, &__pyx_t_5, &__pyx_t_1, NULL, NULL, __pyx_t_8);
    if (unlikely(__pyx_t_9 == 0)) break;
    if (unlikely(__pyx_t_9 == -1)) __PYX_ERR(1, 1694, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_sig, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_v_match_found = 0;
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_v_sig, __pyx_n_s_strip); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 1694, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __pyx_t_12 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_11))) {
      __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_11);
      if (likely(__pyx_t_12)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
        __Pyx_INCREF(__pyx_t_12);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_11, function);
      }
    }
    __pyx_t_10 = (__pyx_t_12) ? __Pyx_PyObject_Call2Args(__pyx_t_11, __pyx_t_12, __pyx_kp_s__13) : __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_kp_s__13);
    __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
    if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 1694, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_split); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 1694, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_t_10 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_11))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_11);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_11, function);
      }
    }
    __pyx_t_1 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_11, __pyx_t_10, __pyx_kp_s__14) : __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_kp_s__14);
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1694, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_XDECREF_SET(__pyx_v_src_sig, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_13 = PyList_GET_SIZE(__pyx_v_dest_sig); if (unlikely(__pyx_t_13 == ((Py_ssize_t)-1))) __PYX_ERR(1, 1694, __pyx_L1_error)
    __pyx_t_14 = __pyx_t_13;
    for (__pyx_t_15 = 0; __pyx_t_15 < __pyx_t_14; __pyx_t_15+=1) {
      __pyx_v_i = __pyx_t_15;
      __pyx_t_1 = PyList_GET_ITEM(__pyx_v_dest_sig, __pyx_v_i);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_XDECREF_SET(__pyx_v_dst_type, __pyx_t_1);
      __pyx_t_1 = 0;
      __pyx_t_2 = (__pyx_v_dst_type != Py_None);
      __pyx_t_3 = (__pyx_t_2 != 0);
      if (__pyx_t_3) {
        __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_src_sig, __pyx_v_i, Py_ssize_t, 1, PyInt_FromSsize_t, 0, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1694, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_11 = PyObject_RichCompare(__pyx_t_1, __pyx_v_dst_type, Py_EQ); __Pyx_XGOTREF(__pyx_t_11); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 1694, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_11); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(1, 1694, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        if (__pyx_t_3) {
          __pyx_v_match_found = 1;
          goto __pyx_L20;
        }
        /*else*/ {
          __pyx_v_match_found = 0;
          goto __pyx_L18_break;
        }
        __pyx_L20:;
      }
    }
    __pyx_L18_break:;
    __pyx_t_3 = (__pyx_v_match_found != 0);
    if (__pyx_t_3) {
      __pyx_t_16 = __Pyx_PyList_Append(__pyx_v_candidates, __pyx_v_sig); if (unlikely(__pyx_t_16 == ((int)-1))) __PYX_ERR(1, 1694, __pyx_L1_error)
    }
  }
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_3 = (PyList_GET_SIZE(__pyx_v_candidates) != 0);
  __pyx_t_2 = ((!__pyx_t_3) != 0);
  if (__pyx_t_2) {
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__15, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 1694, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(1, 1694, __pyx_L1_error)
  }
  __pyx_t_7 = PyList_GET_SIZE(__pyx_v_candidates); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1))) __PYX_ERR(1, 1694, __pyx_L1_error)
  __pyx_t_2 = ((__pyx_t_7 > 1) != 0);
  if (__pyx_t_2) {
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__16, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 1694, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(1, 1694, __pyx_L1_error)
  }
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    if (unlikely(((PyObject *)__pyx_v_signatures) == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 1694, __pyx_L1_error)
    }
    __pyx_t_6 = __Pyx_PyDict_GetItem(((PyObject*)__pyx_v_signatures), PyList_GET_ITEM(__pyx_v_candidates, 0)); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 1694, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_r = __pyx_t_6;
    __pyx_t_6 = 0;
    goto __pyx_L0;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("pandas._libs.interval.Int32ClosedLeftIntervalNode.__pyx_fused_cpdef", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_dest_sig);
  __Pyx_XDECREF(__pyx_v_ndarray);
  __Pyx_XDECREF(__pyx_v_arg);
  __Pyx_XDECREF(__pyx_v_candidates);
  __Pyx_XDECREF(__pyx_v_sig);
  __Pyx_XDECREF(__pyx_v_src_sig);
  __Pyx_XDECREF(__pyx_v_dst_type);
  __Pyx_XDECREF(__pyx_v_kwargs);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_9__pyx_fuse_0query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_fuse_0__pyx_f_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float64_t __pyx_v_point, int __pyx_skip_dispatch) {
  __Pyx_memviewslice __pyx_v_indices = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_values = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __pyx_t_5numpy_int64_t __pyx_t_9;
  __pyx_t_5numpy_int64_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  int __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  __Pyx_memviewslice __pyx_t_16 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_17 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  Py_ssize_t __pyx_t_21;
  __Pyx_RefNannySetupContext("__pyx_fuse_0query", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pyx_fuse_0query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1694, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_9__pyx_fuse_0query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = PyFloat_FromDouble(__pyx_v_point); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1694, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1694, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1694, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 1694, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_5) {
            __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1694, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1694, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_3query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_4 = PyFloat_FromDouble(__pyx_v_point); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1694, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_7 = __pyx_t_1; __pyx_t_3 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1694, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1694, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1694, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (__pyx_t_3) {
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_t_4);
          __pyx_t_4 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1694, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pandas/_libs/intervaltree.pxi":1703
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
  __pyx_t_8 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_8) {

    /* "pandas/_libs/intervaltree.pxi":1707
 *             # continue the binary tree structure. Instead, we use linear
 *             # search.
 *             for i in range(self.n_elements):             # <<<<<<<<<<<<<<
 *                 if self.left[i] <= point < self.right[i]:
 *                     result.append(self.indices[i])
 */
    __pyx_t_9 = __pyx_v_self->n_elements;
    __pyx_t_10 = __pyx_t_9;
    for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
      __pyx_v_i = __pyx_t_11;

      /* "pandas/_libs/intervaltree.pxi":1708
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point < self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      __pyx_t_12 = __pyx_v_i;
      __pyx_t_8 = ((*((__pyx_t_5numpy_int32_t *) ( /* dim=0 */ (__pyx_v_self->left.data + __pyx_t_12 * __pyx_v_self->left.strides[0]) ))) <= __pyx_v_point);
      if (__pyx_t_8) {
        __pyx_t_13 = __pyx_v_i;
        __pyx_t_8 = (__pyx_v_point < (*((__pyx_t_5numpy_int32_t *) ( /* dim=0 */ (__pyx_v_self->right.data + __pyx_t_13 * __pyx_v_self->right.strides[0]) ))));
      }
      __pyx_t_14 = (__pyx_t_8 != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":1709
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point < self.right[i]:
 *                     result.append(self.indices[i])             # <<<<<<<<<<<<<<
 *         else:
 *             # There are child nodes. Based on comparing our query to the pivot,
 */
        __pyx_t_15 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->indices.data + __pyx_t_15 * __pyx_v_self->indices.strides[0]) ))));

        /* "pandas/_libs/intervaltree.pxi":1708
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point < self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      }
    }

    /* "pandas/_libs/intervaltree.pxi":1703
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
    goto __pyx_L3;
  }

  /* "pandas/_libs/intervaltree.pxi":1713
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
  /*else*/ {
    __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":1714
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:
 *                 values = self.center_left_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 */
      __pyx_t_16 = __pyx_v_self->center_left_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":1715
 *             if point < self.pivot:
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:
 */
      __pyx_t_17 = __pyx_v_self->center_left_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":1716
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):             # <<<<<<<<<<<<<<
 *                     if not values[i] <= point:
 *                         break
 */
      __pyx_t_9 = __pyx_v_self->n_center;
      __pyx_t_10 = __pyx_t_9;
      for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":1717
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_18 = __pyx_v_i;
        __pyx_t_14 = ((!(((*((__pyx_t_5numpy_int32_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_18 * __pyx_v_values.strides[0]) ))) <= __pyx_v_point) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":1718
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:
 */
          goto __pyx_L9_break;

          /* "pandas/_libs/intervaltree.pxi":1717
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":1719
 *                     if not values[i] <= point:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 */
        __pyx_t_19 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_19 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L9_break:;

      /* "pandas/_libs/intervaltree.pxi":1720
 *                         break
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->left_node->max_right) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":1721
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)             # <<<<<<<<<<<<<<
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode *)__pyx_v_self->left_node->__pyx_vtab)->__pyx_fuse_0query(__pyx_v_self->left_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1721, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":1720
 *                         break
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":1713
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":1722
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
    __pyx_t_14 = ((__pyx_v_point > __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":1723
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 *                 values = self.center_right_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 */
      __pyx_t_16 = __pyx_v_self->center_right_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":1724
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:
 */
      __pyx_t_17 = __pyx_v_self->center_right_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":1725
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):             # <<<<<<<<<<<<<<
 *                     if not point < values[i]:
 *                         break
 */
      for (__pyx_t_11 = (__pyx_v_self->n_center - 1); __pyx_t_11 > -1L; __pyx_t_11-=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":1726
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_20 = __pyx_v_i;
        __pyx_t_14 = ((!((__pyx_v_point < (*((__pyx_t_5numpy_int32_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_20 * __pyx_v_values.strides[0]) )))) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":1727
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:
 */
          goto __pyx_L13_break;

          /* "pandas/_libs/intervaltree.pxi":1726
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":1728
 *                     if not point < values[i]:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if self.right_node.min_left <= point:
 *                     self.right_node.query(result, point)
 */
        __pyx_t_21 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_21 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L13_break:;

      /* "pandas/_libs/intervaltree.pxi":1729
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      __pyx_t_14 = ((__pyx_v_self->right_node->min_left <= __pyx_v_point) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":1730
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:
 *                     self.right_node.query(result, point)             # <<<<<<<<<<<<<<
 *             else:
 *                 result.extend(self.center_left_indices)
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode *)__pyx_v_self->right_node->__pyx_vtab)->__pyx_fuse_0query(__pyx_v_self->right_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1730, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":1729
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":1722
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":1732
 *                     self.right_node.query(result, point)
 *             else:
 *                 result.extend(self.center_left_indices)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
    /*else*/ {
      __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->extend(__pyx_v_result, __pyx_v_self->center_left_indices); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1732, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_L7:;
  }
  __pyx_L3:;

  /* "pandas/_libs/intervaltree.pxi":1694
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __PYX_XDEC_MEMVIEW(&__pyx_t_16, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_17, 1);
  __Pyx_AddTraceback("pandas._libs.interval.Int32ClosedLeftIntervalNode.query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_indices, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_values, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_9__pyx_fuse_0query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_0__pyx_mdef_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_9__pyx_fuse_0query = {"__pyx_fuse_0query", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_9__pyx_fuse_0query, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_2query};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_9__pyx_fuse_0query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result = 0;
  __pyx_t_5numpy_float64_t __pyx_v_point;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fuse_0query (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_result,&__pyx_n_s_point,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_result)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_point)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_0query", 1, 2, 2, 1); __PYX_ERR(1, 1694, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_0query") < 0)) __PYX_ERR(1, 1694, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_result = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)values[0]);
    __pyx_v_point = __pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_point == ((npy_float64)-1)) && PyErr_Occurred())) __PYX_ERR(1, 1694, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_0query", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 1694, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Int32ClosedLeftIntervalNode.__pyx_fuse_0query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_result), __pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector, 1, "result", 0))) __PYX_ERR(1, 1694, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_8__pyx_fuse_0query(((struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode *)__pyx_v_self), __pyx_v_result, __pyx_v_point);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_8__pyx_fuse_0query(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float64_t __pyx_v_point) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__pyx_fuse_0query", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_fuse_0__pyx_f_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_query(__pyx_v_self, __pyx_v_result, __pyx_v_point, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1694, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Int32ClosedLeftIntervalNode.__pyx_fuse_0query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_11__pyx_fuse_1query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_fuse_1__pyx_f_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float32_t __pyx_v_point, int __pyx_skip_dispatch) {
  __Pyx_memviewslice __pyx_v_indices = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_values = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __pyx_t_5numpy_int64_t __pyx_t_9;
  __pyx_t_5numpy_int64_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  int __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  __Pyx_memviewslice __pyx_t_16 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_17 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  Py_ssize_t __pyx_t_21;
  __Pyx_RefNannySetupContext("__pyx_fuse_1query", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pyx_fuse_1query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1694, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_11__pyx_fuse_1query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = PyFloat_FromDouble(__pyx_v_point); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1694, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1694, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1694, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 1694, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_5) {
            __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1694, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1694, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_3query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_4 = PyFloat_FromDouble(__pyx_v_point); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1694, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_7 = __pyx_t_1; __pyx_t_3 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1694, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1694, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1694, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (__pyx_t_3) {
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_t_4);
          __pyx_t_4 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1694, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pandas/_libs/intervaltree.pxi":1703
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
  __pyx_t_8 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_8) {

    /* "pandas/_libs/intervaltree.pxi":1707
 *             # continue the binary tree structure. Instead, we use linear
 *             # search.
 *             for i in range(self.n_elements):             # <<<<<<<<<<<<<<
 *                 if self.left[i] <= point < self.right[i]:
 *                     result.append(self.indices[i])
 */
    __pyx_t_9 = __pyx_v_self->n_elements;
    __pyx_t_10 = __pyx_t_9;
    for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
      __pyx_v_i = __pyx_t_11;

      /* "pandas/_libs/intervaltree.pxi":1708
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point < self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      __pyx_t_12 = __pyx_v_i;
      __pyx_t_8 = ((*((__pyx_t_5numpy_int32_t *) ( /* dim=0 */ (__pyx_v_self->left.data + __pyx_t_12 * __pyx_v_self->left.strides[0]) ))) <= __pyx_v_point);
      if (__pyx_t_8) {
        __pyx_t_13 = __pyx_v_i;
        __pyx_t_8 = (__pyx_v_point < (*((__pyx_t_5numpy_int32_t *) ( /* dim=0 */ (__pyx_v_self->right.data + __pyx_t_13 * __pyx_v_self->right.strides[0]) ))));
      }
      __pyx_t_14 = (__pyx_t_8 != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":1709
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point < self.right[i]:
 *                     result.append(self.indices[i])             # <<<<<<<<<<<<<<
 *         else:
 *             # There are child nodes. Based on comparing our query to the pivot,
 */
        __pyx_t_15 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->indices.data + __pyx_t_15 * __pyx_v_self->indices.strides[0]) ))));

        /* "pandas/_libs/intervaltree.pxi":1708
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point < self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      }
    }

    /* "pandas/_libs/intervaltree.pxi":1703
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
    goto __pyx_L3;
  }

  /* "pandas/_libs/intervaltree.pxi":1713
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
  /*else*/ {
    __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":1714
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:
 *                 values = self.center_left_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 */
      __pyx_t_16 = __pyx_v_self->center_left_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":1715
 *             if point < self.pivot:
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:
 */
      __pyx_t_17 = __pyx_v_self->center_left_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":1716
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):             # <<<<<<<<<<<<<<
 *                     if not values[i] <= point:
 *                         break
 */
      __pyx_t_9 = __pyx_v_self->n_center;
      __pyx_t_10 = __pyx_t_9;
      for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":1717
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_18 = __pyx_v_i;
        __pyx_t_14 = ((!(((*((__pyx_t_5numpy_int32_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_18 * __pyx_v_values.strides[0]) ))) <= __pyx_v_point) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":1718
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:
 */
          goto __pyx_L9_break;

          /* "pandas/_libs/intervaltree.pxi":1717
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":1719
 *                     if not values[i] <= point:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 */
        __pyx_t_19 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_19 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L9_break:;

      /* "pandas/_libs/intervaltree.pxi":1720
 *                         break
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->left_node->max_right) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":1721
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)             # <<<<<<<<<<<<<<
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode *)__pyx_v_self->left_node->__pyx_vtab)->__pyx_fuse_1query(__pyx_v_self->left_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1721, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":1720
 *                         break
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":1713
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":1722
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
    __pyx_t_14 = ((__pyx_v_point > __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":1723
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 *                 values = self.center_right_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 */
      __pyx_t_16 = __pyx_v_self->center_right_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":1724
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:
 */
      __pyx_t_17 = __pyx_v_self->center_right_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":1725
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):             # <<<<<<<<<<<<<<
 *                     if not point < values[i]:
 *                         break
 */
      for (__pyx_t_11 = (__pyx_v_self->n_center - 1); __pyx_t_11 > -1L; __pyx_t_11-=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":1726
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_20 = __pyx_v_i;
        __pyx_t_14 = ((!((__pyx_v_point < (*((__pyx_t_5numpy_int32_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_20 * __pyx_v_values.strides[0]) )))) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":1727
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:
 */
          goto __pyx_L13_break;

          /* "pandas/_libs/intervaltree.pxi":1726
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":1728
 *                     if not point < values[i]:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if self.right_node.min_left <= point:
 *                     self.right_node.query(result, point)
 */
        __pyx_t_21 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_21 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L13_break:;

      /* "pandas/_libs/intervaltree.pxi":1729
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      __pyx_t_14 = ((__pyx_v_self->right_node->min_left <= __pyx_v_point) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":1730
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:
 *                     self.right_node.query(result, point)             # <<<<<<<<<<<<<<
 *             else:
 *                 result.extend(self.center_left_indices)
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode *)__pyx_v_self->right_node->__pyx_vtab)->__pyx_fuse_1query(__pyx_v_self->right_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1730, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":1729
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":1722
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":1732
 *                     self.right_node.query(result, point)
 *             else:
 *                 result.extend(self.center_left_indices)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
    /*else*/ {
      __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->extend(__pyx_v_result, __pyx_v_self->center_left_indices); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1732, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_L7:;
  }
  __pyx_L3:;

  /* "pandas/_libs/intervaltree.pxi":1694
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __PYX_XDEC_MEMVIEW(&__pyx_t_16, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_17, 1);
  __Pyx_AddTraceback("pandas._libs.interval.Int32ClosedLeftIntervalNode.query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_indices, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_values, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_11__pyx_fuse_1query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_1__pyx_mdef_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_11__pyx_fuse_1query = {"__pyx_fuse_1query", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_11__pyx_fuse_1query, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_2query};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_11__pyx_fuse_1query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result = 0;
  __pyx_t_5numpy_float32_t __pyx_v_point;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fuse_1query (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_result,&__pyx_n_s_point,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_result)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_point)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_1query", 1, 2, 2, 1); __PYX_ERR(1, 1694, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_1query") < 0)) __PYX_ERR(1, 1694, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_result = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)values[0]);
    __pyx_v_point = __pyx_PyFloat_AsFloat(values[1]); if (unlikely((__pyx_v_point == ((npy_float32)-1)) && PyErr_Occurred())) __PYX_ERR(1, 1694, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_1query", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 1694, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Int32ClosedLeftIntervalNode.__pyx_fuse_1query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_result), __pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector, 1, "result", 0))) __PYX_ERR(1, 1694, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_10__pyx_fuse_1query(((struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode *)__pyx_v_self), __pyx_v_result, __pyx_v_point);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_10__pyx_fuse_1query(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float32_t __pyx_v_point) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__pyx_fuse_1query", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_fuse_1__pyx_f_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_query(__pyx_v_self, __pyx_v_result, __pyx_v_point, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1694, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Int32ClosedLeftIntervalNode.__pyx_fuse_1query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_13__pyx_fuse_2query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_fuse_2__pyx_f_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int64_t __pyx_v_point, int __pyx_skip_dispatch) {
  __Pyx_memviewslice __pyx_v_indices = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_values = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __pyx_t_5numpy_int64_t __pyx_t_9;
  __pyx_t_5numpy_int64_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  int __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  __Pyx_memviewslice __pyx_t_16 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_17 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  Py_ssize_t __pyx_t_21;
  __Pyx_RefNannySetupContext("__pyx_fuse_2query", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pyx_fuse_2query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1694, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_13__pyx_fuse_2query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = __Pyx_PyInt_From_npy_int64(__pyx_v_point); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1694, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1694, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1694, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 1694, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_5) {
            __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1694, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1694, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_3query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_4 = __Pyx_PyInt_From_npy_int64(__pyx_v_point); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1694, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_7 = __pyx_t_1; __pyx_t_3 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1694, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1694, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1694, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (__pyx_t_3) {
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_t_4);
          __pyx_t_4 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1694, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pandas/_libs/intervaltree.pxi":1703
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
  __pyx_t_8 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_8) {

    /* "pandas/_libs/intervaltree.pxi":1707
 *             # continue the binary tree structure. Instead, we use linear
 *             # search.
 *             for i in range(self.n_elements):             # <<<<<<<<<<<<<<
 *                 if self.left[i] <= point < self.right[i]:
 *                     result.append(self.indices[i])
 */
    __pyx_t_9 = __pyx_v_self->n_elements;
    __pyx_t_10 = __pyx_t_9;
    for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
      __pyx_v_i = __pyx_t_11;

      /* "pandas/_libs/intervaltree.pxi":1708
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point < self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      __pyx_t_12 = __pyx_v_i;
      __pyx_t_8 = ((*((__pyx_t_5numpy_int32_t *) ( /* dim=0 */ (__pyx_v_self->left.data + __pyx_t_12 * __pyx_v_self->left.strides[0]) ))) <= __pyx_v_point);
      if (__pyx_t_8) {
        __pyx_t_13 = __pyx_v_i;
        __pyx_t_8 = (__pyx_v_point < (*((__pyx_t_5numpy_int32_t *) ( /* dim=0 */ (__pyx_v_self->right.data + __pyx_t_13 * __pyx_v_self->right.strides[0]) ))));
      }
      __pyx_t_14 = (__pyx_t_8 != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":1709
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point < self.right[i]:
 *                     result.append(self.indices[i])             # <<<<<<<<<<<<<<
 *         else:
 *             # There are child nodes. Based on comparing our query to the pivot,
 */
        __pyx_t_15 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->indices.data + __pyx_t_15 * __pyx_v_self->indices.strides[0]) ))));

        /* "pandas/_libs/intervaltree.pxi":1708
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point < self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      }
    }

    /* "pandas/_libs/intervaltree.pxi":1703
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
    goto __pyx_L3;
  }

  /* "pandas/_libs/intervaltree.pxi":1713
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
  /*else*/ {
    __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":1714
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:
 *                 values = self.center_left_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 */
      __pyx_t_16 = __pyx_v_self->center_left_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":1715
 *             if point < self.pivot:
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:
 */
      __pyx_t_17 = __pyx_v_self->center_left_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":1716
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):             # <<<<<<<<<<<<<<
 *                     if not values[i] <= point:
 *                         break
 */
      __pyx_t_9 = __pyx_v_self->n_center;
      __pyx_t_10 = __pyx_t_9;
      for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":1717
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_18 = __pyx_v_i;
        __pyx_t_14 = ((!(((*((__pyx_t_5numpy_int32_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_18 * __pyx_v_values.strides[0]) ))) <= __pyx_v_point) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":1718
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:
 */
          goto __pyx_L9_break;

          /* "pandas/_libs/intervaltree.pxi":1717
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":1719
 *                     if not values[i] <= point:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 */
        __pyx_t_19 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_19 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L9_break:;

      /* "pandas/_libs/intervaltree.pxi":1720
 *                         break
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->left_node->max_right) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":1721
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)             # <<<<<<<<<<<<<<
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode *)__pyx_v_self->left_node->__pyx_vtab)->__pyx_fuse_2query(__pyx_v_self->left_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1721, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":1720
 *                         break
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":1713
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":1722
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
    __pyx_t_14 = ((__pyx_v_point > __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":1723
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 *                 values = self.center_right_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 */
      __pyx_t_16 = __pyx_v_self->center_right_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":1724
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:
 */
      __pyx_t_17 = __pyx_v_self->center_right_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":1725
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):             # <<<<<<<<<<<<<<
 *                     if not point < values[i]:
 *                         break
 */
      for (__pyx_t_11 = (__pyx_v_self->n_center - 1); __pyx_t_11 > -1L; __pyx_t_11-=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":1726
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_20 = __pyx_v_i;
        __pyx_t_14 = ((!((__pyx_v_point < (*((__pyx_t_5numpy_int32_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_20 * __pyx_v_values.strides[0]) )))) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":1727
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:
 */
          goto __pyx_L13_break;

          /* "pandas/_libs/intervaltree.pxi":1726
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":1728
 *                     if not point < values[i]:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if self.right_node.min_left <= point:
 *                     self.right_node.query(result, point)
 */
        __pyx_t_21 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_21 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L13_break:;

      /* "pandas/_libs/intervaltree.pxi":1729
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      __pyx_t_14 = ((__pyx_v_self->right_node->min_left <= __pyx_v_point) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":1730
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:
 *                     self.right_node.query(result, point)             # <<<<<<<<<<<<<<
 *             else:
 *                 result.extend(self.center_left_indices)
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode *)__pyx_v_self->right_node->__pyx_vtab)->__pyx_fuse_2query(__pyx_v_self->right_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1730, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":1729
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":1722
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":1732
 *                     self.right_node.query(result, point)
 *             else:
 *                 result.extend(self.center_left_indices)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
    /*else*/ {
      __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->extend(__pyx_v_result, __pyx_v_self->center_left_indices); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1732, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_L7:;
  }
  __pyx_L3:;

  /* "pandas/_libs/intervaltree.pxi":1694
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __PYX_XDEC_MEMVIEW(&__pyx_t_16, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_17, 1);
  __Pyx_AddTraceback("pandas._libs.interval.Int32ClosedLeftIntervalNode.query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_indices, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_values, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_13__pyx_fuse_2query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_2__pyx_mdef_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_13__pyx_fuse_2query = {"__pyx_fuse_2query", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_13__pyx_fuse_2query, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_2query};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_13__pyx_fuse_2query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result = 0;
  __pyx_t_5numpy_int64_t __pyx_v_point;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fuse_2query (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_result,&__pyx_n_s_point,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_result)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_point)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_2query", 1, 2, 2, 1); __PYX_ERR(1, 1694, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_2query") < 0)) __PYX_ERR(1, 1694, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_result = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)values[0]);
    __pyx_v_point = __Pyx_PyInt_As_npy_int64(values[1]); if (unlikely((__pyx_v_point == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(1, 1694, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_2query", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 1694, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Int32ClosedLeftIntervalNode.__pyx_fuse_2query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_result), __pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector, 1, "result", 0))) __PYX_ERR(1, 1694, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_12__pyx_fuse_2query(((struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode *)__pyx_v_self), __pyx_v_result, __pyx_v_point);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_12__pyx_fuse_2query(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int64_t __pyx_v_point) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__pyx_fuse_2query", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_fuse_2__pyx_f_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_query(__pyx_v_self, __pyx_v_result, __pyx_v_point, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1694, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Int32ClosedLeftIntervalNode.__pyx_fuse_2query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_15__pyx_fuse_3query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_fuse_3__pyx_f_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int32_t __pyx_v_point, int __pyx_skip_dispatch) {
  __Pyx_memviewslice __pyx_v_indices = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_values = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __pyx_t_5numpy_int64_t __pyx_t_9;
  __pyx_t_5numpy_int64_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  int __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  __Pyx_memviewslice __pyx_t_16 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_17 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  Py_ssize_t __pyx_t_21;
  __Pyx_RefNannySetupContext("__pyx_fuse_3query", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pyx_fuse_3query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1694, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_15__pyx_fuse_3query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = __Pyx_PyInt_From_npy_int32(__pyx_v_point); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1694, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1694, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1694, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 1694, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_5) {
            __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1694, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1694, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_3query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_4 = __Pyx_PyInt_From_npy_int32(__pyx_v_point); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1694, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_7 = __pyx_t_1; __pyx_t_3 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1694, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1694, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1694, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (__pyx_t_3) {
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_t_4);
          __pyx_t_4 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1694, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pandas/_libs/intervaltree.pxi":1703
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
  __pyx_t_8 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_8) {

    /* "pandas/_libs/intervaltree.pxi":1707
 *             # continue the binary tree structure. Instead, we use linear
 *             # search.
 *             for i in range(self.n_elements):             # <<<<<<<<<<<<<<
 *                 if self.left[i] <= point < self.right[i]:
 *                     result.append(self.indices[i])
 */
    __pyx_t_9 = __pyx_v_self->n_elements;
    __pyx_t_10 = __pyx_t_9;
    for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
      __pyx_v_i = __pyx_t_11;

      /* "pandas/_libs/intervaltree.pxi":1708
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point < self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      __pyx_t_12 = __pyx_v_i;
      __pyx_t_8 = ((*((__pyx_t_5numpy_int32_t *) ( /* dim=0 */ (__pyx_v_self->left.data + __pyx_t_12 * __pyx_v_self->left.strides[0]) ))) <= __pyx_v_point);
      if (__pyx_t_8) {
        __pyx_t_13 = __pyx_v_i;
        __pyx_t_8 = (__pyx_v_point < (*((__pyx_t_5numpy_int32_t *) ( /* dim=0 */ (__pyx_v_self->right.data + __pyx_t_13 * __pyx_v_self->right.strides[0]) ))));
      }
      __pyx_t_14 = (__pyx_t_8 != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":1709
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point < self.right[i]:
 *                     result.append(self.indices[i])             # <<<<<<<<<<<<<<
 *         else:
 *             # There are child nodes. Based on comparing our query to the pivot,
 */
        __pyx_t_15 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->indices.data + __pyx_t_15 * __pyx_v_self->indices.strides[0]) ))));

        /* "pandas/_libs/intervaltree.pxi":1708
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point < self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      }
    }

    /* "pandas/_libs/intervaltree.pxi":1703
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
    goto __pyx_L3;
  }

  /* "pandas/_libs/intervaltree.pxi":1713
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
  /*else*/ {
    __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":1714
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:
 *                 values = self.center_left_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 */
      __pyx_t_16 = __pyx_v_self->center_left_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":1715
 *             if point < self.pivot:
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:
 */
      __pyx_t_17 = __pyx_v_self->center_left_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":1716
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):             # <<<<<<<<<<<<<<
 *                     if not values[i] <= point:
 *                         break
 */
      __pyx_t_9 = __pyx_v_self->n_center;
      __pyx_t_10 = __pyx_t_9;
      for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":1717
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_18 = __pyx_v_i;
        __pyx_t_14 = ((!(((*((__pyx_t_5numpy_int32_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_18 * __pyx_v_values.strides[0]) ))) <= __pyx_v_point) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":1718
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:
 */
          goto __pyx_L9_break;

          /* "pandas/_libs/intervaltree.pxi":1717
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":1719
 *                     if not values[i] <= point:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 */
        __pyx_t_19 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_19 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L9_break:;

      /* "pandas/_libs/intervaltree.pxi":1720
 *                         break
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->left_node->max_right) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":1721
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)             # <<<<<<<<<<<<<<
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode *)__pyx_v_self->left_node->__pyx_vtab)->__pyx_fuse_3query(__pyx_v_self->left_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1721, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":1720
 *                         break
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":1713
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":1722
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
    __pyx_t_14 = ((__pyx_v_point > __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":1723
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 *                 values = self.center_right_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 */
      __pyx_t_16 = __pyx_v_self->center_right_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":1724
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:
 */
      __pyx_t_17 = __pyx_v_self->center_right_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":1725
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):             # <<<<<<<<<<<<<<
 *                     if not point < values[i]:
 *                         break
 */
      for (__pyx_t_11 = (__pyx_v_self->n_center - 1); __pyx_t_11 > -1L; __pyx_t_11-=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":1726
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_20 = __pyx_v_i;
        __pyx_t_14 = ((!((__pyx_v_point < (*((__pyx_t_5numpy_int32_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_20 * __pyx_v_values.strides[0]) )))) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":1727
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:
 */
          goto __pyx_L13_break;

          /* "pandas/_libs/intervaltree.pxi":1726
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":1728
 *                     if not point < values[i]:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if self.right_node.min_left <= point:
 *                     self.right_node.query(result, point)
 */
        __pyx_t_21 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_21 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L13_break:;

      /* "pandas/_libs/intervaltree.pxi":1729
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      __pyx_t_14 = ((__pyx_v_self->right_node->min_left <= __pyx_v_point) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":1730
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:
 *                     self.right_node.query(result, point)             # <<<<<<<<<<<<<<
 *             else:
 *                 result.extend(self.center_left_indices)
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode *)__pyx_v_self->right_node->__pyx_vtab)->__pyx_fuse_3query(__pyx_v_self->right_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1730, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":1729
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":1722
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":1732
 *                     self.right_node.query(result, point)
 *             else:
 *                 result.extend(self.center_left_indices)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
    /*else*/ {
      __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->extend(__pyx_v_result, __pyx_v_self->center_left_indices); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1732, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_L7:;
  }
  __pyx_L3:;

  /* "pandas/_libs/intervaltree.pxi":1694
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __PYX_XDEC_MEMVIEW(&__pyx_t_16, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_17, 1);
  __Pyx_AddTraceback("pandas._libs.interval.Int32ClosedLeftIntervalNode.query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_indices, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_values, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_15__pyx_fuse_3query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_3__pyx_mdef_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_15__pyx_fuse_3query = {"__pyx_fuse_3query", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_15__pyx_fuse_3query, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_2query};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_15__pyx_fuse_3query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result = 0;
  __pyx_t_5numpy_int32_t __pyx_v_point;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fuse_3query (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_result,&__pyx_n_s_point,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_result)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_point)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_3query", 1, 2, 2, 1); __PYX_ERR(1, 1694, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_3query") < 0)) __PYX_ERR(1, 1694, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_result = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)values[0]);
    __pyx_v_point = __Pyx_PyInt_As_npy_int32(values[1]); if (unlikely((__pyx_v_point == ((npy_int32)-1)) && PyErr_Occurred())) __PYX_ERR(1, 1694, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_3query", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 1694, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Int32ClosedLeftIntervalNode.__pyx_fuse_3query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_result), __pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector, 1, "result", 0))) __PYX_ERR(1, 1694, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_14__pyx_fuse_3query(((struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode *)__pyx_v_self), __pyx_v_result, __pyx_v_point);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_14__pyx_fuse_3query(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int32_t __pyx_v_point) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__pyx_fuse_3query", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_fuse_3__pyx_f_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_query(__pyx_v_self, __pyx_v_result, __pyx_v_point, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1694, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Int32ClosedLeftIntervalNode.__pyx_fuse_3query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_17__pyx_fuse_4query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_fuse_4__pyx_f_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_uint64_t __pyx_v_point, int __pyx_skip_dispatch) {
  __Pyx_memviewslice __pyx_v_indices = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_values = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __pyx_t_5numpy_int64_t __pyx_t_9;
  __pyx_t_5numpy_int64_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  int __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  __Pyx_memviewslice __pyx_t_16 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_17 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  Py_ssize_t __pyx_t_21;
  __Pyx_RefNannySetupContext("__pyx_fuse_4query", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pyx_fuse_4query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1694, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_17__pyx_fuse_4query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = __Pyx_PyInt_From_npy_uint64(__pyx_v_point); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1694, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1694, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1694, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 1694, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_5) {
            __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1694, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1694, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_3query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_4 = __Pyx_PyInt_From_npy_uint64(__pyx_v_point); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1694, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_7 = __pyx_t_1; __pyx_t_3 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1694, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1694, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1694, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (__pyx_t_3) {
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_t_4);
          __pyx_t_4 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1694, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pandas/_libs/intervaltree.pxi":1703
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
  __pyx_t_8 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_8) {

    /* "pandas/_libs/intervaltree.pxi":1707
 *             # continue the binary tree structure. Instead, we use linear
 *             # search.
 *             for i in range(self.n_elements):             # <<<<<<<<<<<<<<
 *                 if self.left[i] <= point < self.right[i]:
 *                     result.append(self.indices[i])
 */
    __pyx_t_9 = __pyx_v_self->n_elements;
    __pyx_t_10 = __pyx_t_9;
    for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
      __pyx_v_i = __pyx_t_11;

      /* "pandas/_libs/intervaltree.pxi":1708
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point < self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      __pyx_t_12 = __pyx_v_i;
      __pyx_t_8 = ((*((__pyx_t_5numpy_int32_t *) ( /* dim=0 */ (__pyx_v_self->left.data + __pyx_t_12 * __pyx_v_self->left.strides[0]) ))) <= __pyx_v_point);
      if (__pyx_t_8) {
        __pyx_t_13 = __pyx_v_i;
        __pyx_t_8 = (__pyx_v_point < (*((__pyx_t_5numpy_int32_t *) ( /* dim=0 */ (__pyx_v_self->right.data + __pyx_t_13 * __pyx_v_self->right.strides[0]) ))));
      }
      __pyx_t_14 = (__pyx_t_8 != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":1709
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point < self.right[i]:
 *                     result.append(self.indices[i])             # <<<<<<<<<<<<<<
 *         else:
 *             # There are child nodes. Based on comparing our query to the pivot,
 */
        __pyx_t_15 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->indices.data + __pyx_t_15 * __pyx_v_self->indices.strides[0]) ))));

        /* "pandas/_libs/intervaltree.pxi":1708
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point < self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      }
    }

    /* "pandas/_libs/intervaltree.pxi":1703
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
    goto __pyx_L3;
  }

  /* "pandas/_libs/intervaltree.pxi":1713
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
  /*else*/ {
    __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":1714
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:
 *                 values = self.center_left_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 */
      __pyx_t_16 = __pyx_v_self->center_left_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":1715
 *             if point < self.pivot:
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:
 */
      __pyx_t_17 = __pyx_v_self->center_left_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":1716
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):             # <<<<<<<<<<<<<<
 *                     if not values[i] <= point:
 *                         break
 */
      __pyx_t_9 = __pyx_v_self->n_center;
      __pyx_t_10 = __pyx_t_9;
      for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":1717
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_18 = __pyx_v_i;
        __pyx_t_14 = ((!(((*((__pyx_t_5numpy_int32_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_18 * __pyx_v_values.strides[0]) ))) <= __pyx_v_point) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":1718
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:
 */
          goto __pyx_L9_break;

          /* "pandas/_libs/intervaltree.pxi":1717
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":1719
 *                     if not values[i] <= point:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 */
        __pyx_t_19 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_19 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L9_break:;

      /* "pandas/_libs/intervaltree.pxi":1720
 *                         break
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->left_node->max_right) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":1721
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)             # <<<<<<<<<<<<<<
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode *)__pyx_v_self->left_node->__pyx_vtab)->__pyx_fuse_4query(__pyx_v_self->left_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1721, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":1720
 *                         break
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":1713
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":1722
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
    __pyx_t_14 = ((__pyx_v_point > __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":1723
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 *                 values = self.center_right_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 */
      __pyx_t_16 = __pyx_v_self->center_right_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":1724
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:
 */
      __pyx_t_17 = __pyx_v_self->center_right_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":1725
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):             # <<<<<<<<<<<<<<
 *                     if not point < values[i]:
 *                         break
 */
      for (__pyx_t_11 = (__pyx_v_self->n_center - 1); __pyx_t_11 > -1L; __pyx_t_11-=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":1726
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_20 = __pyx_v_i;
        __pyx_t_14 = ((!((__pyx_v_point < (*((__pyx_t_5numpy_int32_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_20 * __pyx_v_values.strides[0]) )))) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":1727
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:
 */
          goto __pyx_L13_break;

          /* "pandas/_libs/intervaltree.pxi":1726
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":1728
 *                     if not point < values[i]:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if self.right_node.min_left <= point:
 *                     self.right_node.query(result, point)
 */
        __pyx_t_21 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_21 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L13_break:;

      /* "pandas/_libs/intervaltree.pxi":1729
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      __pyx_t_14 = ((__pyx_v_self->right_node->min_left <= __pyx_v_point) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":1730
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:
 *                     self.right_node.query(result, point)             # <<<<<<<<<<<<<<
 *             else:
 *                 result.extend(self.center_left_indices)
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode *)__pyx_v_self->right_node->__pyx_vtab)->__pyx_fuse_4query(__pyx_v_self->right_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1730, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":1729
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":1722
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":1732
 *                     self.right_node.query(result, point)
 *             else:
 *                 result.extend(self.center_left_indices)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
    /*else*/ {
      __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->extend(__pyx_v_result, __pyx_v_self->center_left_indices); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1732, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_L7:;
  }
  __pyx_L3:;

  /* "pandas/_libs/intervaltree.pxi":1694
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __PYX_XDEC_MEMVIEW(&__pyx_t_16, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_17, 1);
  __Pyx_AddTraceback("pandas._libs.interval.Int32ClosedLeftIntervalNode.query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_indices, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_values, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_17__pyx_fuse_4query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_4__pyx_mdef_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_17__pyx_fuse_4query = {"__pyx_fuse_4query", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_17__pyx_fuse_4query, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_2query};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_17__pyx_fuse_4query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result = 0;
  __pyx_t_5numpy_uint64_t __pyx_v_point;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fuse_4query (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_result,&__pyx_n_s_point,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_result)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_point)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_4query", 1, 2, 2, 1); __PYX_ERR(1, 1694, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_4query") < 0)) __PYX_ERR(1, 1694, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_result = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)values[0]);
    __pyx_v_point = __Pyx_PyInt_As_npy_uint64(values[1]); if (unlikely((__pyx_v_point == ((npy_uint64)-1)) && PyErr_Occurred())) __PYX_ERR(1, 1694, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_4query", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 1694, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Int32ClosedLeftIntervalNode.__pyx_fuse_4query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_result), __pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector, 1, "result", 0))) __PYX_ERR(1, 1694, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_16__pyx_fuse_4query(((struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode *)__pyx_v_self), __pyx_v_result, __pyx_v_point);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_16__pyx_fuse_4query(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_uint64_t __pyx_v_point) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__pyx_fuse_4query", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_fuse_4__pyx_f_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_query(__pyx_v_self, __pyx_v_result, __pyx_v_point, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1694, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Int32ClosedLeftIntervalNode.__pyx_fuse_4query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":1734
 *                 result.extend(self.center_left_indices)
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         if self.is_leaf_node:
 *             return ('<Int32ClosedLeftIntervalNode: '
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_5__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_5__repr__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_4__repr__(((struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_4__repr__(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode *__pyx_v_self) {
  PyObject *__pyx_v_n_left = NULL;
  PyObject *__pyx_v_n_right = NULL;
  PyObject *__pyx_v_n_center = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  Py_UCS4 __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "pandas/_libs/intervaltree.pxi":1735
 * 
 *     def __repr__(self):
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             return ('<Int32ClosedLeftIntervalNode: '
 *                     '%s elements (terminal)>' % self.n_elements)
 */
  __pyx_t_1 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_1) {

    /* "pandas/_libs/intervaltree.pxi":1736
 *     def __repr__(self):
 *         if self.is_leaf_node:
 *             return ('<Int32ClosedLeftIntervalNode: '             # <<<<<<<<<<<<<<
 *                     '%s elements (terminal)>' % self.n_elements)
 *         else:
 */
    __Pyx_XDECREF(__pyx_r);

    /* "pandas/_libs/intervaltree.pxi":1737
 *         if self.is_leaf_node:
 *             return ('<Int32ClosedLeftIntervalNode: '
 *                     '%s elements (terminal)>' % self.n_elements)             # <<<<<<<<<<<<<<
 *         else:
 *             n_left = self.left_node.n_elements
 */
    __pyx_t_2 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_elements); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1737, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PyUnicode_Format(__pyx_kp_u_Int32ClosedLeftIntervalNode_s_e, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1737, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pandas/_libs/intervaltree.pxi":1735
 * 
 *     def __repr__(self):
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             return ('<Int32ClosedLeftIntervalNode: '
 *                     '%s elements (terminal)>' % self.n_elements)
 */
  }

  /* "pandas/_libs/intervaltree.pxi":1739
 *                     '%s elements (terminal)>' % self.n_elements)
 *         else:
 *             n_left = self.left_node.n_elements             # <<<<<<<<<<<<<<
 *             n_right = self.right_node.n_elements
 *             n_center = self.n_elements - n_left - n_right
 */
  /*else*/ {
    __pyx_t_3 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->left_node->n_elements); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1739, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_v_n_left = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "pandas/_libs/intervaltree.pxi":1740
 *         else:
 *             n_left = self.left_node.n_elements
 *             n_right = self.right_node.n_elements             # <<<<<<<<<<<<<<
 *             n_center = self.n_elements - n_left - n_right
 *             return ('<Int32ClosedLeftIntervalNode: '
 */
    __pyx_t_3 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->right_node->n_elements); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1740, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_v_n_right = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "pandas/_libs/intervaltree.pxi":1741
 *             n_left = self.left_node.n_elements
 *             n_right = self.right_node.n_elements
 *             n_center = self.n_elements - n_left - n_right             # <<<<<<<<<<<<<<
 *             return ('<Int32ClosedLeftIntervalNode: '
 *                     'pivot %s, %s elements (%s left, %s right, %s '
 */
    __pyx_t_3 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_elements); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1741, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = PyNumber_Subtract(__pyx_t_3, __pyx_v_n_left); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1741, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyNumber_Subtract(__pyx_t_2, __pyx_v_n_right); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1741, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_n_center = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "pandas/_libs/intervaltree.pxi":1742
 *             n_right = self.right_node.n_elements
 *             n_center = self.n_elements - n_left - n_right
 *             return ('<Int32ClosedLeftIntervalNode: '             # <<<<<<<<<<<<<<
 *                     'pivot %s, %s elements (%s left, %s right, %s '
 *                     'overlapping)>' % (self.pivot, self.n_elements,
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = PyTuple_New(11); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1742, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = 0;
    __pyx_t_5 = 127;
    __Pyx_INCREF(__pyx_kp_u_Int32ClosedLeftIntervalNode_piv);
    __pyx_t_4 += 36;
    __Pyx_GIVEREF(__pyx_kp_u_Int32ClosedLeftIntervalNode_piv);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_kp_u_Int32ClosedLeftIntervalNode_piv);

    /* "pandas/_libs/intervaltree.pxi":1744
 *             return ('<Int32ClosedLeftIntervalNode: '
 *                     'pivot %s, %s elements (%s left, %s right, %s '
 *                     'overlapping)>' % (self.pivot, self.n_elements,             # <<<<<<<<<<<<<<
 *                                        n_left, n_right, n_center))
 * 
 */
    __pyx_t_2 = __Pyx_PyInt_From_npy_int32(__pyx_v_self->pivot); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1744, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_t_2), __pyx_empty_unicode); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 1744, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_5 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_6) > __pyx_t_5) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_6) : __pyx_t_5;
    __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_6);
    __pyx_t_6 = 0;
    __Pyx_INCREF(__pyx_kp_u__18);
    __pyx_t_4 += 2;
    __Pyx_GIVEREF(__pyx_kp_u__18);
    PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_kp_u__18);
    __pyx_t_6 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_elements); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 1744, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_2 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_t_6), __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1744, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_5 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) > __pyx_t_5) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) : __pyx_t_5;
    __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_3, 3, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_elements);
    __pyx_t_4 += 11;
    __Pyx_GIVEREF(__pyx_kp_u_elements);
    PyTuple_SET_ITEM(__pyx_t_3, 4, __pyx_kp_u_elements);

    /* "pandas/_libs/intervaltree.pxi":1745
 *                     'pivot %s, %s elements (%s left, %s right, %s '
 *                     'overlapping)>' % (self.pivot, self.n_elements,
 *                                        n_left, n_right, n_center))             # <<<<<<<<<<<<<<
 * 
 *     def counts(self):
 */
    __pyx_t_2 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_v_n_left), __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1745, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) > __pyx_t_5) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) : __pyx_t_5;
    __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_3, 5, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_left_2);
    __pyx_t_4 += 7;
    __Pyx_GIVEREF(__pyx_kp_u_left_2);
    PyTuple_SET_ITEM(__pyx_t_3, 6, __pyx_kp_u_left_2);
    __pyx_t_2 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_v_n_right), __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1745, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) > __pyx_t_5) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) : __pyx_t_5;
    __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_3, 7, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_right_2);
    __pyx_t_4 += 8;
    __Pyx_GIVEREF(__pyx_kp_u_right_2);
    PyTuple_SET_ITEM(__pyx_t_3, 8, __pyx_kp_u_right_2);
    __pyx_t_2 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_v_n_center), __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1745, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) > __pyx_t_5) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) : __pyx_t_5;
    __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_3, 9, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_overlapping);
    __pyx_t_4 += 14;
    __Pyx_GIVEREF(__pyx_kp_u_overlapping);
    PyTuple_SET_ITEM(__pyx_t_3, 10, __pyx_kp_u_overlapping);

    /* "pandas/_libs/intervaltree.pxi":1742
 *             n_right = self.right_node.n_elements
 *             n_center = self.n_elements - n_left - n_right
 *             return ('<Int32ClosedLeftIntervalNode: '             # <<<<<<<<<<<<<<
 *                     'pivot %s, %s elements (%s left, %s right, %s '
 *                     'overlapping)>' % (self.pivot, self.n_elements,
 */
    __pyx_t_2 = __Pyx_PyUnicode_Join(__pyx_t_3, 11, __pyx_t_4, __pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1742, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;
  }

  /* "pandas/_libs/intervaltree.pxi":1734
 *                 result.extend(self.center_left_indices)
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         if self.is_leaf_node:
 *             return ('<Int32ClosedLeftIntervalNode: '
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pandas._libs.interval.Int32ClosedLeftIntervalNode.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_n_left);
  __Pyx_XDECREF(__pyx_v_n_right);
  __Pyx_XDECREF(__pyx_v_n_center);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":1747
 *                                        n_left, n_right, n_center))
 * 
 *     def counts(self):             # <<<<<<<<<<<<<<
 *         """
 *         Inspect counts on this node
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_7counts(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_6counts[] = "\n        Inspect counts on this node\n        useful for debugging purposes\n        ";
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_7counts(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("counts (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_6counts(((struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_6counts(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode *__pyx_v_self) {
  Py_ssize_t __pyx_v_m;
  PyObject *__pyx_v_l = NULL;
  PyObject *__pyx_v_r = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  size_t __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("counts", 0);

  /* "pandas/_libs/intervaltree.pxi":1752
 *         useful for debugging purposes
 *         """
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             return self.n_elements
 *         else:
 */
  __pyx_t_1 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_1) {

    /* "pandas/_libs/intervaltree.pxi":1753
 *         """
 *         if self.is_leaf_node:
 *             return self.n_elements             # <<<<<<<<<<<<<<
 *         else:
 *             m = len(self.center_left_values)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_elements); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1753, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "pandas/_libs/intervaltree.pxi":1752
 *         useful for debugging purposes
 *         """
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             return self.n_elements
 *         else:
 */
  }

  /* "pandas/_libs/intervaltree.pxi":1755
 *             return self.n_elements
 *         else:
 *             m = len(self.center_left_values)             # <<<<<<<<<<<<<<
 *             l = self.left_node.counts()
 *             r = self.right_node.counts()
 */
  /*else*/ {
    if (unlikely(!__pyx_v_self->center_left_values.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(1, 1755, __pyx_L1_error)}
    __pyx_t_3 = __Pyx_MemoryView_Len(__pyx_v_self->center_left_values); 
    __pyx_v_m = __pyx_t_3;

    /* "pandas/_libs/intervaltree.pxi":1756
 *         else:
 *             m = len(self.center_left_values)
 *             l = self.left_node.counts()             # <<<<<<<<<<<<<<
 *             r = self.right_node.counts()
 *             return (m, (l, r))
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->left_node), __pyx_n_s_counts); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1756, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1756, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_l = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "pandas/_libs/intervaltree.pxi":1757
 *             m = len(self.center_left_values)
 *             l = self.left_node.counts()
 *             r = self.right_node.counts()             # <<<<<<<<<<<<<<
 *             return (m, (l, r))
 * 
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->right_node), __pyx_n_s_counts); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1757, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1757, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_r = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "pandas/_libs/intervaltree.pxi":1758
 *             l = self.left_node.counts()
 *             r = self.right_node.counts()
 *             return (m, (l, r))             # <<<<<<<<<<<<<<
 * 
 * NODE_CLASSES['int32',
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = PyInt_FromSsize_t(__pyx_v_m); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1758, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1758, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_v_l);
    __Pyx_GIVEREF(__pyx_v_l);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_l);
    __Pyx_INCREF(__pyx_v_r);
    __Pyx_GIVEREF(__pyx_v_r);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_v_r);
    __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1758, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_4);
    __pyx_t_2 = 0;
    __pyx_t_4 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;
  }

  /* "pandas/_libs/intervaltree.pxi":1747
 *                                        n_left, n_right, n_center))
 * 
 *     def counts(self):             # <<<<<<<<<<<<<<
 *         """
 *         Inspect counts on this node
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pandas._libs.interval.Int32ClosedLeftIntervalNode.counts", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_l);
  __Pyx_XDECREF(__pyx_v_r);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":1604
 *         int64_t[:] center_left_indices, center_right_indices, indices
 *         int32_t min_left, max_right
 *         readonly int32_t pivot             # <<<<<<<<<<<<<<
 *         readonly int64_t n_elements, n_center, leaf_size
 *         readonly bint is_leaf_node
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_5pivot_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_5pivot_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_5pivot___get__(((struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_5pivot___get__(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_npy_int32(__pyx_v_self->pivot); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1604, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Int32ClosedLeftIntervalNode.pivot.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":1605
 *         int32_t min_left, max_right
 *         readonly int32_t pivot
 *         readonly int64_t n_elements, n_center, leaf_size             # <<<<<<<<<<<<<<
 *         readonly bint is_leaf_node
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_10n_elements_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_10n_elements_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_10n_elements___get__(((struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_10n_elements___get__(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_elements); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1605, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Int32ClosedLeftIntervalNode.n_elements.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_8n_center_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_8n_center_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_8n_center___get__(((struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_8n_center___get__(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_center); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1605, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Int32ClosedLeftIntervalNode.n_center.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_9leaf_size_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_9leaf_size_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_9leaf_size___get__(((struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_9leaf_size___get__(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->leaf_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1605, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Int32ClosedLeftIntervalNode.leaf_size.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":1606
 *         readonly int32_t pivot
 *         readonly int64_t n_elements, n_center, leaf_size
 *         readonly bint is_leaf_node             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self,
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_12is_leaf_node_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_12is_leaf_node_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_12is_leaf_node___get__(((struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_12is_leaf_node___get__(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->is_leaf_node); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1606, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Int32ClosedLeftIntervalNode.is_leaf_node.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_21__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_21__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_20__reduce_cython__(((struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_20__reduce_cython__(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  int __pyx_t_16;
  int __pyx_t_17;
  int __pyx_t_18;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.center_left_indices, self.center_left_values, self.center_right_indices, self.center_right_values, self.indices, self.is_leaf_node, self.leaf_size, self.left, self.left_node, self.max_right, self.min_left, self.n_center, self.n_elements, self.pivot, self.right, self.right_node)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  if (unlikely(!__pyx_v_self->center_left_indices.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_1 = __pyx_memoryview_fromslice(__pyx_v_self->center_left_indices, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_int64_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_int64_t, 0);; if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (unlikely(!__pyx_v_self->center_left_values.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_2 = __pyx_memoryview_fromslice(__pyx_v_self->center_left_values, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_int32_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_int32_t, 0);; if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (unlikely(!__pyx_v_self->center_right_indices.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_3 = __pyx_memoryview_fromslice(__pyx_v_self->center_right_indices, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_int64_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_int64_t, 0);; if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely(!__pyx_v_self->center_right_values.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_4 = __pyx_memoryview_fromslice(__pyx_v_self->center_right_values, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_int32_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_int32_t, 0);; if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (unlikely(!__pyx_v_self->indices.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_5 = __pyx_memoryview_fromslice(__pyx_v_self->indices, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_int64_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_int64_t, 0);; if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyBool_FromLong(__pyx_v_self->is_leaf_node); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->leaf_size); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (unlikely(!__pyx_v_self->left.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_8 = __pyx_memoryview_fromslice(__pyx_v_self->left, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_int32_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_int32_t, 0);; if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_PyInt_From_npy_int32(__pyx_v_self->max_right); if (unlikely(!__pyx_t_9)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_10 = __Pyx_PyInt_From_npy_int32(__pyx_v_self->min_left); if (unlikely(!__pyx_t_10)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_11 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_center); if (unlikely(!__pyx_t_11)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_12 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_elements); if (unlikely(!__pyx_t_12)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __pyx_t_13 = __Pyx_PyInt_From_npy_int32(__pyx_v_self->pivot); if (unlikely(!__pyx_t_13)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  if (unlikely(!__pyx_v_self->right.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_14 = __pyx_memoryview_fromslice(__pyx_v_self->right, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_int32_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_int32_t, 0);; if (unlikely(!__pyx_t_14)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __pyx_t_15 = PyTuple_New(16); if (unlikely(!__pyx_t_15)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_15, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_15, 1, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_15, 2, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_15, 3, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_15, 4, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_15, 5, __pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_15, 6, __pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_15, 7, __pyx_t_8);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->left_node));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self->left_node));
  PyTuple_SET_ITEM(__pyx_t_15, 8, ((PyObject *)__pyx_v_self->left_node));
  __Pyx_GIVEREF(__pyx_t_9);
  PyTuple_SET_ITEM(__pyx_t_15, 9, __pyx_t_9);
  __Pyx_GIVEREF(__pyx_t_10);
  PyTuple_SET_ITEM(__pyx_t_15, 10, __pyx_t_10);
  __Pyx_GIVEREF(__pyx_t_11);
  PyTuple_SET_ITEM(__pyx_t_15, 11, __pyx_t_11);
  __Pyx_GIVEREF(__pyx_t_12);
  PyTuple_SET_ITEM(__pyx_t_15, 12, __pyx_t_12);
  __Pyx_GIVEREF(__pyx_t_13);
  PyTuple_SET_ITEM(__pyx_t_15, 13, __pyx_t_13);
  __Pyx_GIVEREF(__pyx_t_14);
  PyTuple_SET_ITEM(__pyx_t_15, 14, __pyx_t_14);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->right_node));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self->right_node));
  PyTuple_SET_ITEM(__pyx_t_15, 15, ((PyObject *)__pyx_v_self->right_node));
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  __pyx_t_6 = 0;
  __pyx_t_7 = 0;
  __pyx_t_8 = 0;
  __pyx_t_9 = 0;
  __pyx_t_10 = 0;
  __pyx_t_11 = 0;
  __pyx_t_12 = 0;
  __pyx_t_13 = 0;
  __pyx_t_14 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_15);
  __pyx_t_15 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.center_left_indices, self.center_left_values, self.center_right_indices, self.center_right_values, self.indices, self.is_leaf_node, self.leaf_size, self.left, self.left_node, self.max_right, self.min_left, self.n_center, self.n_elements, self.pivot, self.right, self.right_node)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_15 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_15)) __PYX_ERR(2, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __pyx_v__dict = __pyx_t_15;
  __pyx_t_15 = 0;

  /* "(tree fragment)":7
 *     state = (self.center_left_indices, self.center_left_values, self.center_right_indices, self.center_right_values, self.indices, self.is_leaf_node, self.leaf_size, self.left, self.left_node, self.max_right, self.min_left, self.n_center, self.n_elements, self.pivot, self.right, self.right_node)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_16 = (__pyx_v__dict != Py_None);
  __pyx_t_17 = (__pyx_t_16 != 0);
  if (__pyx_t_17) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_15 = PyTuple_New(1); if (unlikely(!__pyx_t_15)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    PyTuple_SET_ITEM(__pyx_t_15, 0, __pyx_v__dict);
    __pyx_t_14 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_15); if (unlikely(!__pyx_t_14)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_14));
    __pyx_t_14 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.left_node is not None or self.right_node is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.center_left_indices, self.center_left_values, self.center_right_indices, self.center_right_values, self.indices, self.is_leaf_node, self.leaf_size, self.left, self.left_node, self.max_right, self.min_left, self.n_center, self.n_elements, self.pivot, self.right, self.right_node)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.left_node is not None or self.right_node is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_Int32ClosedLeftIntervalNode, (type(self), 0x7a104e5, None), state
 */
  /*else*/ {
    __pyx_t_16 = (((PyObject *)__pyx_v_self->left_node) != Py_None);
    __pyx_t_18 = (__pyx_t_16 != 0);
    if (!__pyx_t_18) {
    } else {
      __pyx_t_17 = __pyx_t_18;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_18 = (((PyObject *)__pyx_v_self->right_node) != Py_None);
    __pyx_t_16 = (__pyx_t_18 != 0);
    __pyx_t_17 = __pyx_t_16;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_17;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.left_node is not None or self.right_node is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Int32ClosedLeftIntervalNode, (type(self), 0x7a104e5, None), state
 *     else:
 */
  __pyx_t_17 = (__pyx_v_use_setstate != 0);
  if (__pyx_t_17) {

    /* "(tree fragment)":13
 *         use_setstate = self.left_node is not None or self.right_node is not None
 *     if use_setstate:
 *         return __pyx_unpickle_Int32ClosedLeftIntervalNode, (type(self), 0x7a104e5, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_Int32ClosedLeftIntervalNode, (type(self), 0x7a104e5, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_14, __pyx_n_s_pyx_unpickle_Int32ClosedLeftIn); if (unlikely(!__pyx_t_14)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __pyx_t_15 = PyTuple_New(3); if (unlikely(!__pyx_t_15)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_15, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_127993061);
    __Pyx_GIVEREF(__pyx_int_127993061);
    PyTuple_SET_ITEM(__pyx_t_15, 1, __pyx_int_127993061);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_15, 2, Py_None);
    __pyx_t_13 = PyTuple_New(3); if (unlikely(!__pyx_t_13)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __Pyx_GIVEREF(__pyx_t_14);
    PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_t_14);
    __Pyx_GIVEREF(__pyx_t_15);
    PyTuple_SET_ITEM(__pyx_t_13, 1, __pyx_t_15);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_13, 2, __pyx_v_state);
    __pyx_t_14 = 0;
    __pyx_t_15 = 0;
    __pyx_r = __pyx_t_13;
    __pyx_t_13 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.left_node is not None or self.right_node is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Int32ClosedLeftIntervalNode, (type(self), 0x7a104e5, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_Int32ClosedLeftIntervalNode, (type(self), 0x7a104e5, None), state
 *     else:
 *         return __pyx_unpickle_Int32ClosedLeftIntervalNode, (type(self), 0x7a104e5, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Int32ClosedLeftIntervalNode__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_13, __pyx_n_s_pyx_unpickle_Int32ClosedLeftIn); if (unlikely(!__pyx_t_13)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __pyx_t_15 = PyTuple_New(3); if (unlikely(!__pyx_t_15)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_15, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_127993061);
    __Pyx_GIVEREF(__pyx_int_127993061);
    PyTuple_SET_ITEM(__pyx_t_15, 1, __pyx_int_127993061);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_15, 2, __pyx_v_state);
    __pyx_t_14 = PyTuple_New(2); if (unlikely(!__pyx_t_14)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_GIVEREF(__pyx_t_13);
    PyTuple_SET_ITEM(__pyx_t_14, 0, __pyx_t_13);
    __Pyx_GIVEREF(__pyx_t_15);
    PyTuple_SET_ITEM(__pyx_t_14, 1, __pyx_t_15);
    __pyx_t_13 = 0;
    __pyx_t_15 = 0;
    __pyx_r = __pyx_t_14;
    __pyx_t_14 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_AddTraceback("pandas._libs.interval.Int32ClosedLeftIntervalNode.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Int32ClosedLeftIntervalNode, (type(self), 0x7a104e5, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Int32ClosedLeftIntervalNode__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_23__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_23__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_22__setstate_cython__(((struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_22__setstate_cython__(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_Int32ClosedLeftIntervalNode, (type(self), 0x7a104e5, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Int32ClosedLeftIntervalNode__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(2, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_6pandas_5_libs_8interval___pyx_unpickle_Int32ClosedLeftIntervalNode__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Int32ClosedLeftIntervalNode, (type(self), 0x7a104e5, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Int32ClosedLeftIntervalNode__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Int32ClosedLeftIntervalNode.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":1778
 *         readonly bint is_leaf_node
 * 
 *     def __init__(self,             # <<<<<<<<<<<<<<
 *                  ndarray[int32_t, ndim=1] left,
 *                  ndarray[int32_t, ndim=1] right,
 */

/* Python wrapper */
static int __pyx_pw_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyArrayObject *__pyx_v_left = 0;
  PyArrayObject *__pyx_v_right = 0;
  PyArrayObject *__pyx_v_indices = 0;
  __pyx_t_5numpy_int64_t __pyx_v_leaf_size;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_left,&__pyx_n_s_right,&__pyx_n_s_indices,&__pyx_n_s_leaf_size,0};
    PyObject* values[4] = {0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_left)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_right)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 4, 4, 1); __PYX_ERR(1, 1778, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_indices)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 4, 4, 2); __PYX_ERR(1, 1778, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_leaf_size)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 4, 4, 3); __PYX_ERR(1, 1778, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(1, 1778, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 4) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
    }
    __pyx_v_left = ((PyArrayObject *)values[0]);
    __pyx_v_right = ((PyArrayObject *)values[1]);
    __pyx_v_indices = ((PyArrayObject *)values[2]);
    __pyx_v_leaf_size = __Pyx_PyInt_As_npy_int64(values[3]); if (unlikely((__pyx_v_leaf_size == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(1, 1782, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 4, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 1778, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Int32ClosedRightIntervalNode.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_left), __pyx_ptype_5numpy_ndarray, 1, "left", 0))) __PYX_ERR(1, 1779, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_right), __pyx_ptype_5numpy_ndarray, 1, "right", 0))) __PYX_ERR(1, 1780, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_indices), __pyx_ptype_5numpy_ndarray, 1, "indices", 0))) __PYX_ERR(1, 1781, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode___init__(((struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode *)__pyx_v_self), __pyx_v_left, __pyx_v_right, __pyx_v_indices, __pyx_v_leaf_size);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode___init__(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode *__pyx_v_self, PyArrayObject *__pyx_v_left, PyArrayObject *__pyx_v_right, PyArrayObject *__pyx_v_indices, __pyx_t_5numpy_int64_t __pyx_v_leaf_size) {
  PyObject *__pyx_v_left_set = NULL;
  PyObject *__pyx_v_right_set = NULL;
  PyObject *__pyx_v_center_set = NULL;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_indices;
  __Pyx_Buffer __pyx_pybuffer_indices;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_left;
  __Pyx_Buffer __pyx_pybuffer_left;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_right;
  __Pyx_Buffer __pyx_pybuffer_right;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __pyx_t_5numpy_int32_t __pyx_t_6;
  __Pyx_memviewslice __pyx_t_7 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_8 = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  __Pyx_memviewslice __pyx_t_11 = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *(*__pyx_t_12)(PyObject *);
  size_t __pyx_t_13;
  __Pyx_RefNannySetupContext("__init__", 0);
  __pyx_pybuffer_left.pybuffer.buf = NULL;
  __pyx_pybuffer_left.refcount = 0;
  __pyx_pybuffernd_left.data = NULL;
  __pyx_pybuffernd_left.rcbuffer = &__pyx_pybuffer_left;
  __pyx_pybuffer_right.pybuffer.buf = NULL;
  __pyx_pybuffer_right.refcount = 0;
  __pyx_pybuffernd_right.data = NULL;
  __pyx_pybuffernd_right.rcbuffer = &__pyx_pybuffer_right;
  __pyx_pybuffer_indices.pybuffer.buf = NULL;
  __pyx_pybuffer_indices.refcount = 0;
  __pyx_pybuffernd_indices.data = NULL;
  __pyx_pybuffernd_indices.rcbuffer = &__pyx_pybuffer_indices;
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_left.rcbuffer->pybuffer, (PyObject*)__pyx_v_left, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int32_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 1778, __pyx_L1_error)
  }
  __pyx_pybuffernd_left.diminfo[0].strides = __pyx_pybuffernd_left.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_left.diminfo[0].shape = __pyx_pybuffernd_left.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_right.rcbuffer->pybuffer, (PyObject*)__pyx_v_right, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int32_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 1778, __pyx_L1_error)
  }
  __pyx_pybuffernd_right.diminfo[0].strides = __pyx_pybuffernd_right.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_right.diminfo[0].shape = __pyx_pybuffernd_right.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_indices.rcbuffer->pybuffer, (PyObject*)__pyx_v_indices, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 1778, __pyx_L1_error)
  }
  __pyx_pybuffernd_indices.diminfo[0].strides = __pyx_pybuffernd_indices.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_indices.diminfo[0].shape = __pyx_pybuffernd_indices.rcbuffer->pybuffer.shape[0];

  /* "pandas/_libs/intervaltree.pxi":1784
 *                  int64_t leaf_size):
 * 
 *         self.n_elements = len(left)             # <<<<<<<<<<<<<<
 *         self.leaf_size = leaf_size
 * 
 */
  __pyx_t_1 = PyObject_Length(((PyObject *)__pyx_v_left)); if (unlikely(__pyx_t_1 == ((Py_ssize_t)-1))) __PYX_ERR(1, 1784, __pyx_L1_error)
  __pyx_v_self->n_elements = __pyx_t_1;

  /* "pandas/_libs/intervaltree.pxi":1785
 * 
 *         self.n_elements = len(left)
 *         self.leaf_size = leaf_size             # <<<<<<<<<<<<<<
 * 
 *         # min_left and min_right are used to speed-up query by skipping
 */
  __pyx_v_self->leaf_size = __pyx_v_leaf_size;

  /* "pandas/_libs/intervaltree.pxi":1790
 *         # query on sub-nodes. If this node has size 0, query is cheap,
 *         # so these values don't matter.
 *         if left.size > 0:             # <<<<<<<<<<<<<<
 *             self.min_left = left.min()
 *             self.max_right = right.max()
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_left), __pyx_n_s_size); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1790, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyObject_RichCompare(__pyx_t_2, __pyx_int_0, Py_GT); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1790, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(1, 1790, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_4) {

    /* "pandas/_libs/intervaltree.pxi":1791
 *         # so these values don't matter.
 *         if left.size > 0:
 *             self.min_left = left.min()             # <<<<<<<<<<<<<<
 *             self.max_right = right.max()
 *         else:
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_left), __pyx_n_s_min); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1791, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1791, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_6 = __Pyx_PyInt_As_npy_int32(__pyx_t_3); if (unlikely((__pyx_t_6 == ((npy_int32)-1)) && PyErr_Occurred())) __PYX_ERR(1, 1791, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_self->min_left = __pyx_t_6;

    /* "pandas/_libs/intervaltree.pxi":1792
 *         if left.size > 0:
 *             self.min_left = left.min()
 *             self.max_right = right.max()             # <<<<<<<<<<<<<<
 *         else:
 *             self.min_left = 0
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_right), __pyx_n_s_max); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1792, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1792, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_6 = __Pyx_PyInt_As_npy_int32(__pyx_t_3); if (unlikely((__pyx_t_6 == ((npy_int32)-1)) && PyErr_Occurred())) __PYX_ERR(1, 1792, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_self->max_right = __pyx_t_6;

    /* "pandas/_libs/intervaltree.pxi":1790
 *         # query on sub-nodes. If this node has size 0, query is cheap,
 *         # so these values don't matter.
 *         if left.size > 0:             # <<<<<<<<<<<<<<
 *             self.min_left = left.min()
 *             self.max_right = right.max()
 */
    goto __pyx_L3;
  }

  /* "pandas/_libs/intervaltree.pxi":1794
 *             self.max_right = right.max()
 *         else:
 *             self.min_left = 0             # <<<<<<<<<<<<<<
 *             self.max_right = 0
 * 
 */
  /*else*/ {
    __pyx_v_self->min_left = 0;

    /* "pandas/_libs/intervaltree.pxi":1795
 *         else:
 *             self.min_left = 0
 *             self.max_right = 0             # <<<<<<<<<<<<<<
 * 
 *         if self.n_elements <= leaf_size:
 */
    __pyx_v_self->max_right = 0;
  }
  __pyx_L3:;

  /* "pandas/_libs/intervaltree.pxi":1797
 *             self.max_right = 0
 * 
 *         if self.n_elements <= leaf_size:             # <<<<<<<<<<<<<<
 *             # make this a terminal (leaf) node
 *             self.is_leaf_node = True
 */
  __pyx_t_4 = ((__pyx_v_self->n_elements <= __pyx_v_leaf_size) != 0);
  if (__pyx_t_4) {

    /* "pandas/_libs/intervaltree.pxi":1799
 *         if self.n_elements <= leaf_size:
 *             # make this a terminal (leaf) node
 *             self.is_leaf_node = True             # <<<<<<<<<<<<<<
 *             self.left = left
 *             self.right = right
 */
    __pyx_v_self->is_leaf_node = 1;

    /* "pandas/_libs/intervaltree.pxi":1800
 *             # make this a terminal (leaf) node
 *             self.is_leaf_node = True
 *             self.left = left             # <<<<<<<<<<<<<<
 *             self.right = right
 *             self.indices = indices
 */
    __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int32_t(((PyObject *)__pyx_v_left), PyBUF_WRITABLE); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(1, 1800, __pyx_L1_error)
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->left, 0);
    __pyx_v_self->left = __pyx_t_7;
    __pyx_t_7.memview = NULL;
    __pyx_t_7.data = NULL;

    /* "pandas/_libs/intervaltree.pxi":1801
 *             self.is_leaf_node = True
 *             self.left = left
 *             self.right = right             # <<<<<<<<<<<<<<
 *             self.indices = indices
 *             self.n_center = 0
 */
    __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int32_t(((PyObject *)__pyx_v_right), PyBUF_WRITABLE); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(1, 1801, __pyx_L1_error)
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->right, 0);
    __pyx_v_self->right = __pyx_t_7;
    __pyx_t_7.memview = NULL;
    __pyx_t_7.data = NULL;

    /* "pandas/_libs/intervaltree.pxi":1802
 *             self.left = left
 *             self.right = right
 *             self.indices = indices             # <<<<<<<<<<<<<<
 *             self.n_center = 0
 *         else:
 */
    __pyx_t_8 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(((PyObject *)__pyx_v_indices), PyBUF_WRITABLE); if (unlikely(!__pyx_t_8.memview)) __PYX_ERR(1, 1802, __pyx_L1_error)
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->indices, 0);
    __pyx_v_self->indices = __pyx_t_8;
    __pyx_t_8.memview = NULL;
    __pyx_t_8.data = NULL;

    /* "pandas/_libs/intervaltree.pxi":1803
 *             self.right = right
 *             self.indices = indices
 *             self.n_center = 0             # <<<<<<<<<<<<<<
 *         else:
 *             # calculate a pivot so we can create child nodes
 */
    __pyx_v_self->n_center = 0;

    /* "pandas/_libs/intervaltree.pxi":1797
 *             self.max_right = 0
 * 
 *         if self.n_elements <= leaf_size:             # <<<<<<<<<<<<<<
 *             # make this a terminal (leaf) node
 *             self.is_leaf_node = True
 */
    goto __pyx_L4;
  }

  /* "pandas/_libs/intervaltree.pxi":1806
 *         else:
 *             # calculate a pivot so we can create child nodes
 *             self.is_leaf_node = False             # <<<<<<<<<<<<<<
 *             self.pivot = np.median(left / 2 + right / 2)
 *             left_set, right_set, center_set = self.classify_intervals(
 */
  /*else*/ {
    __pyx_v_self->is_leaf_node = 0;

    /* "pandas/_libs/intervaltree.pxi":1807
 *             # calculate a pivot so we can create child nodes
 *             self.is_leaf_node = False
 *             self.pivot = np.median(left / 2 + right / 2)             # <<<<<<<<<<<<<<
 *             left_set, right_set, center_set = self.classify_intervals(
 *                 left, right)
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1807, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_median); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1807, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyNumber_Divide(((PyObject *)__pyx_v_left), __pyx_int_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1807, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_9 = __Pyx_PyNumber_Divide(((PyObject *)__pyx_v_right), __pyx_int_2); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 1807, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_10 = PyNumber_Add(__pyx_t_2, __pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 1807, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_9 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_3 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_9, __pyx_t_10) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_10);
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1807, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_6 = __Pyx_PyInt_As_npy_int32(__pyx_t_3); if (unlikely((__pyx_t_6 == ((npy_int32)-1)) && PyErr_Occurred())) __PYX_ERR(1, 1807, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_self->pivot = __pyx_t_6;

    /* "pandas/_libs/intervaltree.pxi":1809
 *             self.pivot = np.median(left / 2 + right / 2)
 *             left_set, right_set, center_set = self.classify_intervals(
 *                 left, right)             # <<<<<<<<<<<<<<
 * 
 *             self.left_node = self.new_child_node(left, right,
 */
    __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int32_t(((PyObject *)__pyx_v_left), PyBUF_WRITABLE); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(1, 1809, __pyx_L1_error)
    __pyx_t_11 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int32_t(((PyObject *)__pyx_v_right), PyBUF_WRITABLE); if (unlikely(!__pyx_t_11.memview)) __PYX_ERR(1, 1809, __pyx_L1_error)

    /* "pandas/_libs/intervaltree.pxi":1808
 *             self.is_leaf_node = False
 *             self.pivot = np.median(left / 2 + right / 2)
 *             left_set, right_set, center_set = self.classify_intervals(             # <<<<<<<<<<<<<<
 *                 left, right)
 * 
 */
    __pyx_t_3 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode *)__pyx_v_self->__pyx_vtab)->classify_intervals(__pyx_v_self, __pyx_t_7, __pyx_t_11); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1808, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __PYX_XDEC_MEMVIEW(&__pyx_t_7, 1);
    __pyx_t_7.memview = NULL;
    __pyx_t_7.data = NULL;
    __PYX_XDEC_MEMVIEW(&__pyx_t_11, 1);
    __pyx_t_11.memview = NULL;
    __pyx_t_11.data = NULL;
    if ((likely(PyTuple_CheckExact(__pyx_t_3))) || (PyList_CheckExact(__pyx_t_3))) {
      PyObject* sequence = __pyx_t_3;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 3)) {
        if (size > 3) __Pyx_RaiseTooManyValuesError(3);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(1, 1808, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_5 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_10 = PyTuple_GET_ITEM(sequence, 1); 
        __pyx_t_9 = PyTuple_GET_ITEM(sequence, 2); 
      } else {
        __pyx_t_5 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_10 = PyList_GET_ITEM(sequence, 1); 
        __pyx_t_9 = PyList_GET_ITEM(sequence, 2); 
      }
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_10);
      __Pyx_INCREF(__pyx_t_9);
      #else
      __pyx_t_5 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1808, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_10 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 1808, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_9 = PySequence_ITEM(sequence, 2); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 1808, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      #endif
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_2 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1808, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_12 = Py_TYPE(__pyx_t_2)->tp_iternext;
      index = 0; __pyx_t_5 = __pyx_t_12(__pyx_t_2); if (unlikely(!__pyx_t_5)) goto __pyx_L5_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_5);
      index = 1; __pyx_t_10 = __pyx_t_12(__pyx_t_2); if (unlikely(!__pyx_t_10)) goto __pyx_L5_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_10);
      index = 2; __pyx_t_9 = __pyx_t_12(__pyx_t_2); if (unlikely(!__pyx_t_9)) goto __pyx_L5_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_9);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_12(__pyx_t_2), 3) < 0) __PYX_ERR(1, 1808, __pyx_L1_error)
      __pyx_t_12 = NULL;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      goto __pyx_L6_unpacking_done;
      __pyx_L5_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_12 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(1, 1808, __pyx_L1_error)
      __pyx_L6_unpacking_done:;
    }
    __pyx_v_left_set = __pyx_t_5;
    __pyx_t_5 = 0;
    __pyx_v_right_set = __pyx_t_10;
    __pyx_t_10 = 0;
    __pyx_v_center_set = __pyx_t_9;
    __pyx_t_9 = 0;

    /* "pandas/_libs/intervaltree.pxi":1812
 * 
 *             self.left_node = self.new_child_node(left, right,
 *                                                  indices, left_set)             # <<<<<<<<<<<<<<
 *             self.right_node = self.new_child_node(left, right,
 *                                                   indices, right_set)
 */
    if (!(likely(((__pyx_v_left_set) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_left_set, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(1, 1812, __pyx_L1_error)

    /* "pandas/_libs/intervaltree.pxi":1811
 *                 left, right)
 * 
 *             self.left_node = self.new_child_node(left, right,             # <<<<<<<<<<<<<<
 *                                                  indices, left_set)
 *             self.right_node = self.new_child_node(left, right,
 */
    __pyx_t_3 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode *)__pyx_v_self->__pyx_vtab)->new_child_node(__pyx_v_self, ((PyArrayObject *)__pyx_v_left), ((PyArrayObject *)__pyx_v_right), ((PyArrayObject *)__pyx_v_indices), ((PyArrayObject *)__pyx_v_left_set)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1811, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode))))) __PYX_ERR(1, 1811, __pyx_L1_error)
    __Pyx_GIVEREF(__pyx_t_3);
    __Pyx_GOTREF(__pyx_v_self->left_node);
    __Pyx_DECREF(((PyObject *)__pyx_v_self->left_node));
    __pyx_v_self->left_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode *)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "pandas/_libs/intervaltree.pxi":1814
 *                                                  indices, left_set)
 *             self.right_node = self.new_child_node(left, right,
 *                                                   indices, right_set)             # <<<<<<<<<<<<<<
 * 
 *             self.center_left_values, self.center_left_indices = \
 */
    if (!(likely(((__pyx_v_right_set) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_right_set, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(1, 1814, __pyx_L1_error)

    /* "pandas/_libs/intervaltree.pxi":1813
 *             self.left_node = self.new_child_node(left, right,
 *                                                  indices, left_set)
 *             self.right_node = self.new_child_node(left, right,             # <<<<<<<<<<<<<<
 *                                                   indices, right_set)
 * 
 */
    __pyx_t_3 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode *)__pyx_v_self->__pyx_vtab)->new_child_node(__pyx_v_self, ((PyArrayObject *)__pyx_v_left), ((PyArrayObject *)__pyx_v_right), ((PyArrayObject *)__pyx_v_indices), ((PyArrayObject *)__pyx_v_right_set)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1813, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode))))) __PYX_ERR(1, 1813, __pyx_L1_error)
    __Pyx_GIVEREF(__pyx_t_3);
    __Pyx_GOTREF(__pyx_v_self->right_node);
    __Pyx_DECREF(((PyObject *)__pyx_v_self->right_node));
    __pyx_v_self->right_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode *)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "pandas/_libs/intervaltree.pxi":1817
 * 
 *             self.center_left_values, self.center_left_indices = \
 *                 sort_values_and_indices(left, indices, center_set)             # <<<<<<<<<<<<<<
 *             self.center_right_values, self.center_right_indices = \
 *                 sort_values_and_indices(right, indices, center_set)
 */
    __pyx_t_3 = __pyx_f_6pandas_5_libs_8interval_sort_values_and_indices(((PyObject *)__pyx_v_left), ((PyObject *)__pyx_v_indices), __pyx_v_center_set); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1817, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if ((likely(PyTuple_CheckExact(__pyx_t_3))) || (PyList_CheckExact(__pyx_t_3))) {
      PyObject* sequence = __pyx_t_3;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(1, 1816, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_9 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_10 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_9 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_10 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_9);
      __Pyx_INCREF(__pyx_t_10);
      #else
      __pyx_t_9 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 1816, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_10 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 1816, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      #endif
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_5 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1816, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_12 = Py_TYPE(__pyx_t_5)->tp_iternext;
      index = 0; __pyx_t_9 = __pyx_t_12(__pyx_t_5); if (unlikely(!__pyx_t_9)) goto __pyx_L7_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_9);
      index = 1; __pyx_t_10 = __pyx_t_12(__pyx_t_5); if (unlikely(!__pyx_t_10)) goto __pyx_L7_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_10);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_12(__pyx_t_5), 2) < 0) __PYX_ERR(1, 1816, __pyx_L1_error)
      __pyx_t_12 = NULL;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      goto __pyx_L8_unpacking_done;
      __pyx_L7_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_12 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(1, 1816, __pyx_L1_error)
      __pyx_L8_unpacking_done:;
    }

    /* "pandas/_libs/intervaltree.pxi":1816
 *                                                   indices, right_set)
 * 
 *             self.center_left_values, self.center_left_indices = \             # <<<<<<<<<<<<<<
 *                 sort_values_and_indices(left, indices, center_set)
 *             self.center_right_values, self.center_right_indices = \
 */
    __pyx_t_11 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int32_t(__pyx_t_9, PyBUF_WRITABLE); if (unlikely(!__pyx_t_11.memview)) __PYX_ERR(1, 1816, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_8 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_t_10, PyBUF_WRITABLE); if (unlikely(!__pyx_t_8.memview)) __PYX_ERR(1, 1816, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->center_left_values, 0);
    __pyx_v_self->center_left_values = __pyx_t_11;
    __pyx_t_11.memview = NULL;
    __pyx_t_11.data = NULL;
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->center_left_indices, 0);
    __pyx_v_self->center_left_indices = __pyx_t_8;
    __pyx_t_8.memview = NULL;
    __pyx_t_8.data = NULL;

    /* "pandas/_libs/intervaltree.pxi":1819
 *                 sort_values_and_indices(left, indices, center_set)
 *             self.center_right_values, self.center_right_indices = \
 *                 sort_values_and_indices(right, indices, center_set)             # <<<<<<<<<<<<<<
 *             self.n_center = len(self.center_left_indices)
 * 
 */
    __pyx_t_3 = __pyx_f_6pandas_5_libs_8interval_sort_values_and_indices(((PyObject *)__pyx_v_right), ((PyObject *)__pyx_v_indices), __pyx_v_center_set); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1819, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if ((likely(PyTuple_CheckExact(__pyx_t_3))) || (PyList_CheckExact(__pyx_t_3))) {
      PyObject* sequence = __pyx_t_3;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(1, 1818, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_10 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_9 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_10 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_9 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_10);
      __Pyx_INCREF(__pyx_t_9);
      #else
      __pyx_t_10 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 1818, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_9 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 1818, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      #endif
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_5 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1818, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_12 = Py_TYPE(__pyx_t_5)->tp_iternext;
      index = 0; __pyx_t_10 = __pyx_t_12(__pyx_t_5); if (unlikely(!__pyx_t_10)) goto __pyx_L9_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_10);
      index = 1; __pyx_t_9 = __pyx_t_12(__pyx_t_5); if (unlikely(!__pyx_t_9)) goto __pyx_L9_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_9);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_12(__pyx_t_5), 2) < 0) __PYX_ERR(1, 1818, __pyx_L1_error)
      __pyx_t_12 = NULL;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      goto __pyx_L10_unpacking_done;
      __pyx_L9_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_12 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(1, 1818, __pyx_L1_error)
      __pyx_L10_unpacking_done:;
    }

    /* "pandas/_libs/intervaltree.pxi":1818
 *             self.center_left_values, self.center_left_indices = \
 *                 sort_values_and_indices(left, indices, center_set)
 *             self.center_right_values, self.center_right_indices = \             # <<<<<<<<<<<<<<
 *                 sort_values_and_indices(right, indices, center_set)
 *             self.n_center = len(self.center_left_indices)
 */
    __pyx_t_11 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int32_t(__pyx_t_10, PyBUF_WRITABLE); if (unlikely(!__pyx_t_11.memview)) __PYX_ERR(1, 1818, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_t_8 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_t_9, PyBUF_WRITABLE); if (unlikely(!__pyx_t_8.memview)) __PYX_ERR(1, 1818, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->center_right_values, 0);
    __pyx_v_self->center_right_values = __pyx_t_11;
    __pyx_t_11.memview = NULL;
    __pyx_t_11.data = NULL;
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->center_right_indices, 0);
    __pyx_v_self->center_right_indices = __pyx_t_8;
    __pyx_t_8.memview = NULL;
    __pyx_t_8.data = NULL;

    /* "pandas/_libs/intervaltree.pxi":1820
 *             self.center_right_values, self.center_right_indices = \
 *                 sort_values_and_indices(right, indices, center_set)
 *             self.n_center = len(self.center_left_indices)             # <<<<<<<<<<<<<<
 * 
 *     @cython.wraparound(False)
 */
    if (unlikely(!__pyx_v_self->center_left_indices.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(1, 1820, __pyx_L1_error)}
    __pyx_t_13 = __Pyx_MemoryView_Len(__pyx_v_self->center_left_indices); 
    __pyx_v_self->n_center = __pyx_t_13;
  }
  __pyx_L4:;

  /* "pandas/_libs/intervaltree.pxi":1778
 *         readonly bint is_leaf_node
 * 
 *     def __init__(self,             # <<<<<<<<<<<<<<
 *                  ndarray[int32_t, ndim=1] left,
 *                  ndarray[int32_t, ndim=1] right,
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __PYX_XDEC_MEMVIEW(&__pyx_t_7, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_8, 1);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __PYX_XDEC_MEMVIEW(&__pyx_t_11, 1);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indices.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_left.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_right.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("pandas._libs.interval.Int32ClosedRightIntervalNode.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indices.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_left.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_right.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF(__pyx_v_left_set);
  __Pyx_XDECREF(__pyx_v_right_set);
  __Pyx_XDECREF(__pyx_v_center_set);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":1824
 *     @cython.wraparound(False)
 *     @cython.boundscheck(False)
 *     cdef classify_intervals(self, int32_t[:] left, int32_t[:] right):             # <<<<<<<<<<<<<<
 *         """Classify the given intervals based upon whether they fall to the
 *         left, right, or overlap with this node's pivot.
 */

static PyObject *__pyx_f_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_classify_intervals(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode *__pyx_v_self, __Pyx_memviewslice __pyx_v_left, __Pyx_memviewslice __pyx_v_right) {
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_left_ind = 0;
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_right_ind = 0;
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_overlapping_ind = 0;
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __pyx_t_5numpy_int64_t __pyx_t_2;
  __pyx_t_5numpy_int64_t __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  int __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  __Pyx_RefNannySetupContext("classify_intervals", 0);

  /* "pandas/_libs/intervaltree.pxi":1832
 *             Py_ssize_t i
 * 
 *         left_ind = Int64Vector()             # <<<<<<<<<<<<<<
 *         right_ind = Int64Vector()
 *         overlapping_ind = Int64Vector()
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1832, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_left_ind = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":1833
 * 
 *         left_ind = Int64Vector()
 *         right_ind = Int64Vector()             # <<<<<<<<<<<<<<
 *         overlapping_ind = Int64Vector()
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1833, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_right_ind = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":1834
 *         left_ind = Int64Vector()
 *         right_ind = Int64Vector()
 *         overlapping_ind = Int64Vector()             # <<<<<<<<<<<<<<
 * 
 *         for i in range(self.n_elements):
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1834, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_overlapping_ind = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":1836
 *         overlapping_ind = Int64Vector()
 * 
 *         for i in range(self.n_elements):             # <<<<<<<<<<<<<<
 *             if right[i] < self.pivot:
 *                 left_ind.append(i)
 */
  __pyx_t_2 = __pyx_v_self->n_elements;
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "pandas/_libs/intervaltree.pxi":1837
 * 
 *         for i in range(self.n_elements):
 *             if right[i] < self.pivot:             # <<<<<<<<<<<<<<
 *                 left_ind.append(i)
 *             elif self.pivot <= left[i]:
 */
    __pyx_t_5 = __pyx_v_i;
    __pyx_t_6 = (((*((__pyx_t_5numpy_int32_t *) ( /* dim=0 */ (__pyx_v_right.data + __pyx_t_5 * __pyx_v_right.strides[0]) ))) < __pyx_v_self->pivot) != 0);
    if (__pyx_t_6) {

      /* "pandas/_libs/intervaltree.pxi":1838
 *         for i in range(self.n_elements):
 *             if right[i] < self.pivot:
 *                 left_ind.append(i)             # <<<<<<<<<<<<<<
 *             elif self.pivot <= left[i]:
 *                 right_ind.append(i)
 */
      ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_left_ind->__pyx_vtab)->append(__pyx_v_left_ind, __pyx_v_i);

      /* "pandas/_libs/intervaltree.pxi":1837
 * 
 *         for i in range(self.n_elements):
 *             if right[i] < self.pivot:             # <<<<<<<<<<<<<<
 *                 left_ind.append(i)
 *             elif self.pivot <= left[i]:
 */
      goto __pyx_L5;
    }

    /* "pandas/_libs/intervaltree.pxi":1839
 *             if right[i] < self.pivot:
 *                 left_ind.append(i)
 *             elif self.pivot <= left[i]:             # <<<<<<<<<<<<<<
 *                 right_ind.append(i)
 *             else:
 */
    __pyx_t_7 = __pyx_v_i;
    __pyx_t_6 = ((__pyx_v_self->pivot <= (*((__pyx_t_5numpy_int32_t *) ( /* dim=0 */ (__pyx_v_left.data + __pyx_t_7 * __pyx_v_left.strides[0]) )))) != 0);
    if (__pyx_t_6) {

      /* "pandas/_libs/intervaltree.pxi":1840
 *                 left_ind.append(i)
 *             elif self.pivot <= left[i]:
 *                 right_ind.append(i)             # <<<<<<<<<<<<<<
 *             else:
 *                 overlapping_ind.append(i)
 */
      ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_right_ind->__pyx_vtab)->append(__pyx_v_right_ind, __pyx_v_i);

      /* "pandas/_libs/intervaltree.pxi":1839
 *             if right[i] < self.pivot:
 *                 left_ind.append(i)
 *             elif self.pivot <= left[i]:             # <<<<<<<<<<<<<<
 *                 right_ind.append(i)
 *             else:
 */
      goto __pyx_L5;
    }

    /* "pandas/_libs/intervaltree.pxi":1842
 *                 right_ind.append(i)
 *             else:
 *                 overlapping_ind.append(i)             # <<<<<<<<<<<<<<
 * 
 *         return (left_ind.to_array(),
 */
    /*else*/ {
      ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_overlapping_ind->__pyx_vtab)->append(__pyx_v_overlapping_ind, __pyx_v_i);
    }
    __pyx_L5:;
  }

  /* "pandas/_libs/intervaltree.pxi":1844
 *                 overlapping_ind.append(i)
 * 
 *         return (left_ind.to_array(),             # <<<<<<<<<<<<<<
 *                 right_ind.to_array(),
 *                 overlapping_ind.to_array())
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_left_ind->__pyx_vtab)->to_array(__pyx_v_left_ind, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1844, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "pandas/_libs/intervaltree.pxi":1845
 * 
 *         return (left_ind.to_array(),
 *                 right_ind.to_array(),             # <<<<<<<<<<<<<<
 *                 overlapping_ind.to_array())
 * 
 */
  __pyx_t_8 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_right_ind->__pyx_vtab)->to_array(__pyx_v_right_ind, 0); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 1845, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);

  /* "pandas/_libs/intervaltree.pxi":1846
 *         return (left_ind.to_array(),
 *                 right_ind.to_array(),
 *                 overlapping_ind.to_array())             # <<<<<<<<<<<<<<
 * 
 *     cdef new_child_node(self,
 */
  __pyx_t_9 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_overlapping_ind->__pyx_vtab)->to_array(__pyx_v_overlapping_ind, 0); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 1846, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);

  /* "pandas/_libs/intervaltree.pxi":1844
 *                 overlapping_ind.append(i)
 * 
 *         return (left_ind.to_array(),             # <<<<<<<<<<<<<<
 *                 right_ind.to_array(),
 *                 overlapping_ind.to_array())
 */
  __pyx_t_10 = PyTuple_New(3); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 1844, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_t_8);
  __Pyx_GIVEREF(__pyx_t_9);
  PyTuple_SET_ITEM(__pyx_t_10, 2, __pyx_t_9);
  __pyx_t_1 = 0;
  __pyx_t_8 = 0;
  __pyx_t_9 = 0;
  __pyx_r = __pyx_t_10;
  __pyx_t_10 = 0;
  goto __pyx_L0;

  /* "pandas/_libs/intervaltree.pxi":1824
 *     @cython.wraparound(False)
 *     @cython.boundscheck(False)
 *     cdef classify_intervals(self, int32_t[:] left, int32_t[:] right):             # <<<<<<<<<<<<<<
 *         """Classify the given intervals based upon whether they fall to the
 *         left, right, or overlap with this node's pivot.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("pandas._libs.interval.Int32ClosedRightIntervalNode.classify_intervals", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_left_ind);
  __Pyx_XDECREF((PyObject *)__pyx_v_right_ind);
  __Pyx_XDECREF((PyObject *)__pyx_v_overlapping_ind);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":1848
 *                 overlapping_ind.to_array())
 * 
 *     cdef new_child_node(self,             # <<<<<<<<<<<<<<
 *                         ndarray[int32_t, ndim=1] left,
 *                         ndarray[int32_t, ndim=1] right,
 */

static PyObject *__pyx_f_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_new_child_node(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode *__pyx_v_self, PyArrayObject *__pyx_v_left, PyArrayObject *__pyx_v_right, PyArrayObject *__pyx_v_indices, PyArrayObject *__pyx_v_subset) {
  __Pyx_LocalBuf_ND __pyx_pybuffernd_indices;
  __Pyx_Buffer __pyx_pybuffer_indices;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_left;
  __Pyx_Buffer __pyx_pybuffer_left;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_right;
  __Pyx_Buffer __pyx_pybuffer_right;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_subset;
  __Pyx_Buffer __pyx_pybuffer_subset;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyArrayObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyArrayObject *__pyx_t_7 = NULL;
  PyArrayObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  __Pyx_RefNannySetupContext("new_child_node", 0);
  __Pyx_INCREF((PyObject *)__pyx_v_left);
  __Pyx_INCREF((PyObject *)__pyx_v_right);
  __Pyx_INCREF((PyObject *)__pyx_v_indices);
  __pyx_pybuffer_left.pybuffer.buf = NULL;
  __pyx_pybuffer_left.refcount = 0;
  __pyx_pybuffernd_left.data = NULL;
  __pyx_pybuffernd_left.rcbuffer = &__pyx_pybuffer_left;
  __pyx_pybuffer_right.pybuffer.buf = NULL;
  __pyx_pybuffer_right.refcount = 0;
  __pyx_pybuffernd_right.data = NULL;
  __pyx_pybuffernd_right.rcbuffer = &__pyx_pybuffer_right;
  __pyx_pybuffer_indices.pybuffer.buf = NULL;
  __pyx_pybuffer_indices.refcount = 0;
  __pyx_pybuffernd_indices.data = NULL;
  __pyx_pybuffernd_indices.rcbuffer = &__pyx_pybuffer_indices;
  __pyx_pybuffer_subset.pybuffer.buf = NULL;
  __pyx_pybuffer_subset.refcount = 0;
  __pyx_pybuffernd_subset.data = NULL;
  __pyx_pybuffernd_subset.rcbuffer = &__pyx_pybuffer_subset;
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_left.rcbuffer->pybuffer, (PyObject*)__pyx_v_left, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int32_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 1848, __pyx_L1_error)
  }
  __pyx_pybuffernd_left.diminfo[0].strides = __pyx_pybuffernd_left.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_left.diminfo[0].shape = __pyx_pybuffernd_left.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_right.rcbuffer->pybuffer, (PyObject*)__pyx_v_right, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int32_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 1848, __pyx_L1_error)
  }
  __pyx_pybuffernd_right.diminfo[0].strides = __pyx_pybuffernd_right.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_right.diminfo[0].shape = __pyx_pybuffernd_right.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_indices.rcbuffer->pybuffer, (PyObject*)__pyx_v_indices, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 1848, __pyx_L1_error)
  }
  __pyx_pybuffernd_indices.diminfo[0].strides = __pyx_pybuffernd_indices.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_indices.diminfo[0].shape = __pyx_pybuffernd_indices.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_subset.rcbuffer->pybuffer, (PyObject*)__pyx_v_subset, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 1848, __pyx_L1_error)
  }
  __pyx_pybuffernd_subset.diminfo[0].strides = __pyx_pybuffernd_subset.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_subset.diminfo[0].shape = __pyx_pybuffernd_subset.rcbuffer->pybuffer.shape[0];

  /* "pandas/_libs/intervaltree.pxi":1855
 *         """Create a new child node.
 *         """
 *         left = take(left, subset)             # <<<<<<<<<<<<<<
 *         right = take(right, subset)
 *         indices = take(indices, subset)
 */
  __pyx_t_1 = __pyx_f_6pandas_5_libs_8interval_take(((PyArrayObject *)__pyx_v_left), ((PyArrayObject *)__pyx_v_subset)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1855, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(1, 1855, __pyx_L1_error)
  __pyx_t_2 = ((PyArrayObject *)__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_left.rcbuffer->pybuffer);
    __pyx_t_3 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_left.rcbuffer->pybuffer, (PyObject*)__pyx_t_2, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int32_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack);
    if (unlikely(__pyx_t_3 < 0)) {
      PyErr_Fetch(&__pyx_t_4, &__pyx_t_5, &__pyx_t_6);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_left.rcbuffer->pybuffer, (PyObject*)__pyx_v_left, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int32_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_4); Py_XDECREF(__pyx_t_5); Py_XDECREF(__pyx_t_6);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_4, __pyx_t_5, __pyx_t_6);
      }
      __pyx_t_4 = __pyx_t_5 = __pyx_t_6 = 0;
    }
    __pyx_pybuffernd_left.diminfo[0].strides = __pyx_pybuffernd_left.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_left.diminfo[0].shape = __pyx_pybuffernd_left.rcbuffer->pybuffer.shape[0];
    if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(1, 1855, __pyx_L1_error)
  }
  __pyx_t_2 = 0;
  __Pyx_DECREF_SET(__pyx_v_left, ((PyArrayObject *)__pyx_t_1));
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":1856
 *         """
 *         left = take(left, subset)
 *         right = take(right, subset)             # <<<<<<<<<<<<<<
 *         indices = take(indices, subset)
 *         return Int32ClosedRightIntervalNode(
 */
  __pyx_t_1 = __pyx_f_6pandas_5_libs_8interval_take(((PyArrayObject *)__pyx_v_right), ((PyArrayObject *)__pyx_v_subset)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1856, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(1, 1856, __pyx_L1_error)
  __pyx_t_7 = ((PyArrayObject *)__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_right.rcbuffer->pybuffer);
    __pyx_t_3 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_right.rcbuffer->pybuffer, (PyObject*)__pyx_t_7, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int32_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack);
    if (unlikely(__pyx_t_3 < 0)) {
      PyErr_Fetch(&__pyx_t_6, &__pyx_t_5, &__pyx_t_4);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_right.rcbuffer->pybuffer, (PyObject*)__pyx_v_right, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int32_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_6); Py_XDECREF(__pyx_t_5); Py_XDECREF(__pyx_t_4);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_6, __pyx_t_5, __pyx_t_4);
      }
      __pyx_t_6 = __pyx_t_5 = __pyx_t_4 = 0;
    }
    __pyx_pybuffernd_right.diminfo[0].strides = __pyx_pybuffernd_right.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_right.diminfo[0].shape = __pyx_pybuffernd_right.rcbuffer->pybuffer.shape[0];
    if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(1, 1856, __pyx_L1_error)
  }
  __pyx_t_7 = 0;
  __Pyx_DECREF_SET(__pyx_v_right, ((PyArrayObject *)__pyx_t_1));
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":1857
 *         left = take(left, subset)
 *         right = take(right, subset)
 *         indices = take(indices, subset)             # <<<<<<<<<<<<<<
 *         return Int32ClosedRightIntervalNode(
 *             left, right, indices, self.leaf_size)
 */
  __pyx_t_1 = __pyx_f_6pandas_5_libs_8interval_take(((PyArrayObject *)__pyx_v_indices), ((PyArrayObject *)__pyx_v_subset)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1857, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(1, 1857, __pyx_L1_error)
  __pyx_t_8 = ((PyArrayObject *)__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indices.rcbuffer->pybuffer);
    __pyx_t_3 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_indices.rcbuffer->pybuffer, (PyObject*)__pyx_t_8, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack);
    if (unlikely(__pyx_t_3 < 0)) {
      PyErr_Fetch(&__pyx_t_4, &__pyx_t_5, &__pyx_t_6);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_indices.rcbuffer->pybuffer, (PyObject*)__pyx_v_indices, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_4); Py_XDECREF(__pyx_t_5); Py_XDECREF(__pyx_t_6);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_4, __pyx_t_5, __pyx_t_6);
      }
      __pyx_t_4 = __pyx_t_5 = __pyx_t_6 = 0;
    }
    __pyx_pybuffernd_indices.diminfo[0].strides = __pyx_pybuffernd_indices.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_indices.diminfo[0].shape = __pyx_pybuffernd_indices.rcbuffer->pybuffer.shape[0];
    if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(1, 1857, __pyx_L1_error)
  }
  __pyx_t_8 = 0;
  __Pyx_DECREF_SET(__pyx_v_indices, ((PyArrayObject *)__pyx_t_1));
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":1858
 *         right = take(right, subset)
 *         indices = take(indices, subset)
 *         return Int32ClosedRightIntervalNode(             # <<<<<<<<<<<<<<
 *             left, right, indices, self.leaf_size)
 * 
 */
  __Pyx_XDECREF(__pyx_r);

  /* "pandas/_libs/intervaltree.pxi":1859
 *         indices = take(indices, subset)
 *         return Int32ClosedRightIntervalNode(
 *             left, right, indices, self.leaf_size)             # <<<<<<<<<<<<<<
 * 
 *     @cython.wraparound(False)
 */
  __pyx_t_1 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->leaf_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1859, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "pandas/_libs/intervaltree.pxi":1858
 *         right = take(right, subset)
 *         indices = take(indices, subset)
 *         return Int32ClosedRightIntervalNode(             # <<<<<<<<<<<<<<
 *             left, right, indices, self.leaf_size)
 * 
 */
  __pyx_t_9 = PyTuple_New(4); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 1858, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_INCREF(((PyObject *)__pyx_v_left));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_left));
  PyTuple_SET_ITEM(__pyx_t_9, 0, ((PyObject *)__pyx_v_left));
  __Pyx_INCREF(((PyObject *)__pyx_v_right));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_right));
  PyTuple_SET_ITEM(__pyx_t_9, 1, ((PyObject *)__pyx_v_right));
  __Pyx_INCREF(((PyObject *)__pyx_v_indices));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_indices));
  PyTuple_SET_ITEM(__pyx_t_9, 2, ((PyObject *)__pyx_v_indices));
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_9, 3, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode), __pyx_t_9, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1858, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pandas/_libs/intervaltree.pxi":1848
 *                 overlapping_ind.to_array())
 * 
 *     cdef new_child_node(self,             # <<<<<<<<<<<<<<
 *                         ndarray[int32_t, ndim=1] left,
 *                         ndarray[int32_t, ndim=1] right,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_9);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indices.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_left.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_right.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_subset.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("pandas._libs.interval.Int32ClosedRightIntervalNode.new_child_node", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indices.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_left.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_right.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_subset.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF((PyObject *)__pyx_v_left);
  __Pyx_XDECREF((PyObject *)__pyx_v_right);
  __Pyx_XDECREF((PyObject *)__pyx_v_indices);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":1864
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_2query[] = "Recursively query this node and its sub-nodes for intervals that\n        overlap with the query point.\n        ";
static PyMethodDef __pyx_mdef_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_3query = {"query", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_3query, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_2query};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_args = 0;
  PyObject *__pyx_v_kwargs = 0;
  CYTHON_UNUSED PyObject *__pyx_v_defaults = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fused_cpdef (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_args,&__pyx_n_s_kwargs,&__pyx_n_s_defaults,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_args)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_kwargs)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fused_cpdef", 1, 3, 3, 1); __PYX_ERR(1, 1864, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_defaults)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fused_cpdef", 1, 3, 3, 2); __PYX_ERR(1, 1864, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fused_cpdef") < 0)) __PYX_ERR(1, 1864, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_args = values[0];
    __pyx_v_kwargs = values[1];
    __pyx_v_defaults = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fused_cpdef", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 1864, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Int32ClosedRightIntervalNode.__pyx_fused_cpdef", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_2query(((struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode *)__pyx_v_signatures), __pyx_v_args, __pyx_v_kwargs, __pyx_v_defaults);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_2query(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode *__pyx_v_signatures, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs, CYTHON_UNUSED PyObject *__pyx_v_defaults) {
  PyObject *__pyx_v_dest_sig = NULL;
  Py_ssize_t __pyx_v_i;
  CYTHON_UNUSED PyTypeObject *__pyx_v_ndarray = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_v_itemsize;
  PyObject *__pyx_v_arg = NULL;
  PyObject *__pyx_v_candidates = NULL;
  PyObject *__pyx_v_sig = NULL;
  int __pyx_v_match_found;
  PyObject *__pyx_v_src_sig = NULL;
  PyObject *__pyx_v_dst_type = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  Py_ssize_t __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  Py_ssize_t __pyx_t_13;
  Py_ssize_t __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  int __pyx_t_16;
  __Pyx_RefNannySetupContext("query", 0);
  __Pyx_INCREF(__pyx_v_kwargs);
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1864, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  PyList_SET_ITEM(__pyx_t_1, 0, Py_None);
  __pyx_v_dest_sig = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_3 = (__pyx_v_kwargs != Py_None);
  __pyx_t_4 = (__pyx_t_3 != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_kwargs); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(1, 1864, __pyx_L1_error)
  __pyx_t_3 = ((!__pyx_t_4) != 0);
  __pyx_t_2 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {
    __Pyx_INCREF(Py_None);
    __Pyx_DECREF_SET(__pyx_v_kwargs, Py_None);
  }
  __pyx_t_1 = ((PyObject *)__Pyx_ImportNumPyArrayTypeIfAvailable()); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1864, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_ndarray = ((PyTypeObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_v_itemsize = -1L;
  if (unlikely(__pyx_v_args == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 1864, __pyx_L1_error)
  }
  __pyx_t_5 = PyTuple_GET_SIZE(((PyObject*)__pyx_v_args)); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(1, 1864, __pyx_L1_error)
  __pyx_t_2 = ((2 < __pyx_t_5) != 0);
  if (__pyx_t_2) {
    if (unlikely(__pyx_v_args == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 1864, __pyx_L1_error)
    }
    __pyx_t_1 = PyTuple_GET_ITEM(((PyObject*)__pyx_v_args), 2);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v_arg = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L6;
  }
  __pyx_t_3 = (__pyx_v_kwargs != Py_None);
  __pyx_t_4 = (__pyx_t_3 != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L7_bool_binop_done;
  }
  if (unlikely(__pyx_v_kwargs == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(1, 1864, __pyx_L1_error)
  }
  __pyx_t_4 = (__Pyx_PyDict_ContainsTF(__pyx_n_s_point, ((PyObject*)__pyx_v_kwargs), Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(1, 1864, __pyx_L1_error)
  __pyx_t_3 = (__pyx_t_4 != 0);
  __pyx_t_2 = __pyx_t_3;
  __pyx_L7_bool_binop_done:;
  if (__pyx_t_2) {
    if (unlikely(__pyx_v_kwargs == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 1864, __pyx_L1_error)
    }
    __pyx_t_1 = __Pyx_PyDict_GetItem(((PyObject*)__pyx_v_kwargs), __pyx_n_s_point); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1864, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_arg = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L6;
  }
  /*else*/ {
    if (unlikely(__pyx_v_args == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
      __PYX_ERR(1, 1864, __pyx_L1_error)
    }
    __pyx_t_5 = PyTuple_GET_SIZE(((PyObject*)__pyx_v_args)); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(1, 1864, __pyx_L1_error)
    __pyx_t_1 = PyInt_FromSsize_t(__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1864, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_6 = PyTuple_New(3); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 1864, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_INCREF(__pyx_int_3);
    __Pyx_GIVEREF(__pyx_int_3);
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_int_3);
    __Pyx_INCREF(__pyx_n_s_s);
    __Pyx_GIVEREF(__pyx_n_s_s);
    PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_n_s_s);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Expected_at_least_d_argument_s_g, __pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1864, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_1); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 1864, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(1, 1864, __pyx_L1_error)
  }
  __pyx_L6:;
  while (1) {
    __pyx_t_2 = PyFloat_Check(__pyx_v_arg); 
    __pyx_t_3 = (__pyx_t_2 != 0);
    if (__pyx_t_3) {
      if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, __pyx_n_s_float64_t, long, 1, __Pyx_PyInt_From_long, 1, 0, 0) < 0)) __PYX_ERR(1, 1864, __pyx_L1_error)
      goto __pyx_L10_break;
    }
    __pyx_t_2 = PyInt_Check(__pyx_v_arg); 
    __pyx_t_4 = (__pyx_t_2 != 0);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_3 = __pyx_t_4;
      goto __pyx_L13_bool_binop_done;
    }
    __pyx_t_4 = PyLong_Check(__pyx_v_arg); 
    __pyx_t_2 = (__pyx_t_4 != 0);
    __pyx_t_3 = __pyx_t_2;
    __pyx_L13_bool_binop_done:;
    __pyx_t_2 = (__pyx_t_3 != 0);
    if (__pyx_t_2) {
      if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, __pyx_n_s_int64_t, long, 1, __Pyx_PyInt_From_long, 1, 0, 0) < 0)) __PYX_ERR(1, 1864, __pyx_L1_error)
      goto __pyx_L10_break;
    }
    if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, Py_None, long, 1, __Pyx_PyInt_From_long, 1, 0, 0) < 0)) __PYX_ERR(1, 1864, __pyx_L1_error)
    goto __pyx_L10_break;
  }
  __pyx_L10_break:;
  __pyx_t_6 = PyList_New(0); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 1864, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_v_candidates = ((PyObject*)__pyx_t_6);
  __pyx_t_6 = 0;
  __pyx_t_5 = 0;
  if (unlikely(((PyObject *)__pyx_v_signatures) == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(1, 1864, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_dict_iterator(((PyObject*)__pyx_v_signatures), 1, ((PyObject *)NULL), (&__pyx_t_7), (&__pyx_t_8)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1864, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __pyx_t_6 = __pyx_t_1;
  __pyx_t_1 = 0;
  while (1) {
    __pyx_t_9 = __Pyx_dict_iter_next(__pyx_t_6, __pyx_t_7, &__pyx_t_5, &__pyx_t_1, NULL, NULL, __pyx_t_8);
    if (unlikely(__pyx_t_9 == 0)) break;
    if (unlikely(__pyx_t_9 == -1)) __PYX_ERR(1, 1864, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_sig, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_v_match_found = 0;
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_v_sig, __pyx_n_s_strip); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 1864, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __pyx_t_12 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_11))) {
      __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_11);
      if (likely(__pyx_t_12)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
        __Pyx_INCREF(__pyx_t_12);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_11, function);
      }
    }
    __pyx_t_10 = (__pyx_t_12) ? __Pyx_PyObject_Call2Args(__pyx_t_11, __pyx_t_12, __pyx_kp_s__13) : __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_kp_s__13);
    __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
    if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 1864, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_split); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 1864, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_t_10 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_11))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_11);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_11, function);
      }
    }
    __pyx_t_1 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_11, __pyx_t_10, __pyx_kp_s__14) : __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_kp_s__14);
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1864, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_XDECREF_SET(__pyx_v_src_sig, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_13 = PyList_GET_SIZE(__pyx_v_dest_sig); if (unlikely(__pyx_t_13 == ((Py_ssize_t)-1))) __PYX_ERR(1, 1864, __pyx_L1_error)
    __pyx_t_14 = __pyx_t_13;
    for (__pyx_t_15 = 0; __pyx_t_15 < __pyx_t_14; __pyx_t_15+=1) {
      __pyx_v_i = __pyx_t_15;
      __pyx_t_1 = PyList_GET_ITEM(__pyx_v_dest_sig, __pyx_v_i);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_XDECREF_SET(__pyx_v_dst_type, __pyx_t_1);
      __pyx_t_1 = 0;
      __pyx_t_2 = (__pyx_v_dst_type != Py_None);
      __pyx_t_3 = (__pyx_t_2 != 0);
      if (__pyx_t_3) {
        __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_src_sig, __pyx_v_i, Py_ssize_t, 1, PyInt_FromSsize_t, 0, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1864, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_11 = PyObject_RichCompare(__pyx_t_1, __pyx_v_dst_type, Py_EQ); __Pyx_XGOTREF(__pyx_t_11); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 1864, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_11); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(1, 1864, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        if (__pyx_t_3) {
          __pyx_v_match_found = 1;
          goto __pyx_L20;
        }
        /*else*/ {
          __pyx_v_match_found = 0;
          goto __pyx_L18_break;
        }
        __pyx_L20:;
      }
    }
    __pyx_L18_break:;
    __pyx_t_3 = (__pyx_v_match_found != 0);
    if (__pyx_t_3) {
      __pyx_t_16 = __Pyx_PyList_Append(__pyx_v_candidates, __pyx_v_sig); if (unlikely(__pyx_t_16 == ((int)-1))) __PYX_ERR(1, 1864, __pyx_L1_error)
    }
  }
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_3 = (PyList_GET_SIZE(__pyx_v_candidates) != 0);
  __pyx_t_2 = ((!__pyx_t_3) != 0);
  if (__pyx_t_2) {
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__15, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 1864, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(1, 1864, __pyx_L1_error)
  }
  __pyx_t_7 = PyList_GET_SIZE(__pyx_v_candidates); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1))) __PYX_ERR(1, 1864, __pyx_L1_error)
  __pyx_t_2 = ((__pyx_t_7 > 1) != 0);
  if (__pyx_t_2) {
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__16, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 1864, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(1, 1864, __pyx_L1_error)
  }
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    if (unlikely(((PyObject *)__pyx_v_signatures) == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 1864, __pyx_L1_error)
    }
    __pyx_t_6 = __Pyx_PyDict_GetItem(((PyObject*)__pyx_v_signatures), PyList_GET_ITEM(__pyx_v_candidates, 0)); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 1864, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_r = __pyx_t_6;
    __pyx_t_6 = 0;
    goto __pyx_L0;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("pandas._libs.interval.Int32ClosedRightIntervalNode.__pyx_fused_cpdef", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_dest_sig);
  __Pyx_XDECREF(__pyx_v_ndarray);
  __Pyx_XDECREF(__pyx_v_arg);
  __Pyx_XDECREF(__pyx_v_candidates);
  __Pyx_XDECREF(__pyx_v_sig);
  __Pyx_XDECREF(__pyx_v_src_sig);
  __Pyx_XDECREF(__pyx_v_dst_type);
  __Pyx_XDECREF(__pyx_v_kwargs);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_9__pyx_fuse_0query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_fuse_0__pyx_f_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float64_t __pyx_v_point, int __pyx_skip_dispatch) {
  __Pyx_memviewslice __pyx_v_indices = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_values = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __pyx_t_5numpy_int64_t __pyx_t_9;
  __pyx_t_5numpy_int64_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  int __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  __Pyx_memviewslice __pyx_t_16 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_17 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  Py_ssize_t __pyx_t_21;
  __Pyx_RefNannySetupContext("__pyx_fuse_0query", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pyx_fuse_0query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1864, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_9__pyx_fuse_0query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = PyFloat_FromDouble(__pyx_v_point); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1864, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1864, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1864, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 1864, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_5) {
            __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1864, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1864, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_3query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_4 = PyFloat_FromDouble(__pyx_v_point); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1864, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_7 = __pyx_t_1; __pyx_t_3 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1864, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1864, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1864, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (__pyx_t_3) {
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_t_4);
          __pyx_t_4 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1864, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pandas/_libs/intervaltree.pxi":1873
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
  __pyx_t_8 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_8) {

    /* "pandas/_libs/intervaltree.pxi":1877
 *             # continue the binary tree structure. Instead, we use linear
 *             # search.
 *             for i in range(self.n_elements):             # <<<<<<<<<<<<<<
 *                 if self.left[i] < point <= self.right[i]:
 *                     result.append(self.indices[i])
 */
    __pyx_t_9 = __pyx_v_self->n_elements;
    __pyx_t_10 = __pyx_t_9;
    for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
      __pyx_v_i = __pyx_t_11;

      /* "pandas/_libs/intervaltree.pxi":1878
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point <= self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      __pyx_t_12 = __pyx_v_i;
      __pyx_t_8 = ((*((__pyx_t_5numpy_int32_t *) ( /* dim=0 */ (__pyx_v_self->left.data + __pyx_t_12 * __pyx_v_self->left.strides[0]) ))) < __pyx_v_point);
      if (__pyx_t_8) {
        __pyx_t_13 = __pyx_v_i;
        __pyx_t_8 = (__pyx_v_point <= (*((__pyx_t_5numpy_int32_t *) ( /* dim=0 */ (__pyx_v_self->right.data + __pyx_t_13 * __pyx_v_self->right.strides[0]) ))));
      }
      __pyx_t_14 = (__pyx_t_8 != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":1879
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point <= self.right[i]:
 *                     result.append(self.indices[i])             # <<<<<<<<<<<<<<
 *         else:
 *             # There are child nodes. Based on comparing our query to the pivot,
 */
        __pyx_t_15 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->indices.data + __pyx_t_15 * __pyx_v_self->indices.strides[0]) ))));

        /* "pandas/_libs/intervaltree.pxi":1878
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point <= self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      }
    }

    /* "pandas/_libs/intervaltree.pxi":1873
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
    goto __pyx_L3;
  }

  /* "pandas/_libs/intervaltree.pxi":1883
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
  /*else*/ {
    __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":1884
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:
 *                 values = self.center_left_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 */
      __pyx_t_16 = __pyx_v_self->center_left_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":1885
 *             if point < self.pivot:
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:
 */
      __pyx_t_17 = __pyx_v_self->center_left_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":1886
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):             # <<<<<<<<<<<<<<
 *                     if not values[i] < point:
 *                         break
 */
      __pyx_t_9 = __pyx_v_self->n_center;
      __pyx_t_10 = __pyx_t_9;
      for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":1887
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_18 = __pyx_v_i;
        __pyx_t_14 = ((!(((*((__pyx_t_5numpy_int32_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_18 * __pyx_v_values.strides[0]) ))) < __pyx_v_point) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":1888
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:
 */
          goto __pyx_L9_break;

          /* "pandas/_libs/intervaltree.pxi":1887
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":1889
 *                     if not values[i] < point:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 */
        __pyx_t_19 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_19 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L9_break:;

      /* "pandas/_libs/intervaltree.pxi":1890
 *                         break
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      __pyx_t_14 = ((__pyx_v_point <= __pyx_v_self->left_node->max_right) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":1891
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)             # <<<<<<<<<<<<<<
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode *)__pyx_v_self->left_node->__pyx_vtab)->__pyx_fuse_0query(__pyx_v_self->left_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1891, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":1890
 *                         break
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":1883
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":1892
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
    __pyx_t_14 = ((__pyx_v_point > __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":1893
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 *                 values = self.center_right_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 */
      __pyx_t_16 = __pyx_v_self->center_right_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":1894
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:
 */
      __pyx_t_17 = __pyx_v_self->center_right_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":1895
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):             # <<<<<<<<<<<<<<
 *                     if not point <= values[i]:
 *                         break
 */
      for (__pyx_t_11 = (__pyx_v_self->n_center - 1); __pyx_t_11 > -1L; __pyx_t_11-=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":1896
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_20 = __pyx_v_i;
        __pyx_t_14 = ((!((__pyx_v_point <= (*((__pyx_t_5numpy_int32_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_20 * __pyx_v_values.strides[0]) )))) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":1897
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:
 */
          goto __pyx_L13_break;

          /* "pandas/_libs/intervaltree.pxi":1896
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":1898
 *                     if not point <= values[i]:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if self.right_node.min_left < point:
 *                     self.right_node.query(result, point)
 */
        __pyx_t_21 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_21 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L13_break:;

      /* "pandas/_libs/intervaltree.pxi":1899
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      __pyx_t_14 = ((__pyx_v_self->right_node->min_left < __pyx_v_point) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":1900
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:
 *                     self.right_node.query(result, point)             # <<<<<<<<<<<<<<
 *             else:
 *                 result.extend(self.center_left_indices)
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode *)__pyx_v_self->right_node->__pyx_vtab)->__pyx_fuse_0query(__pyx_v_self->right_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1900, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":1899
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":1892
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":1902
 *                     self.right_node.query(result, point)
 *             else:
 *                 result.extend(self.center_left_indices)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
    /*else*/ {
      __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->extend(__pyx_v_result, __pyx_v_self->center_left_indices); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1902, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_L7:;
  }
  __pyx_L3:;

  /* "pandas/_libs/intervaltree.pxi":1864
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __PYX_XDEC_MEMVIEW(&__pyx_t_16, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_17, 1);
  __Pyx_AddTraceback("pandas._libs.interval.Int32ClosedRightIntervalNode.query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_indices, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_values, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_9__pyx_fuse_0query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_0__pyx_mdef_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_9__pyx_fuse_0query = {"__pyx_fuse_0query", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_9__pyx_fuse_0query, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_2query};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_9__pyx_fuse_0query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result = 0;
  __pyx_t_5numpy_float64_t __pyx_v_point;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fuse_0query (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_result,&__pyx_n_s_point,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_result)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_point)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_0query", 1, 2, 2, 1); __PYX_ERR(1, 1864, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_0query") < 0)) __PYX_ERR(1, 1864, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_result = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)values[0]);
    __pyx_v_point = __pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_point == ((npy_float64)-1)) && PyErr_Occurred())) __PYX_ERR(1, 1864, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_0query", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 1864, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Int32ClosedRightIntervalNode.__pyx_fuse_0query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_result), __pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector, 1, "result", 0))) __PYX_ERR(1, 1864, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_8__pyx_fuse_0query(((struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode *)__pyx_v_self), __pyx_v_result, __pyx_v_point);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_8__pyx_fuse_0query(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float64_t __pyx_v_point) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__pyx_fuse_0query", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_fuse_0__pyx_f_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_query(__pyx_v_self, __pyx_v_result, __pyx_v_point, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1864, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Int32ClosedRightIntervalNode.__pyx_fuse_0query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_11__pyx_fuse_1query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_fuse_1__pyx_f_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float32_t __pyx_v_point, int __pyx_skip_dispatch) {
  __Pyx_memviewslice __pyx_v_indices = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_values = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __pyx_t_5numpy_int64_t __pyx_t_9;
  __pyx_t_5numpy_int64_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  int __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  __Pyx_memviewslice __pyx_t_16 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_17 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  Py_ssize_t __pyx_t_21;
  __Pyx_RefNannySetupContext("__pyx_fuse_1query", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pyx_fuse_1query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1864, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_11__pyx_fuse_1query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = PyFloat_FromDouble(__pyx_v_point); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1864, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1864, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1864, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 1864, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_5) {
            __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1864, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1864, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_3query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_4 = PyFloat_FromDouble(__pyx_v_point); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1864, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_7 = __pyx_t_1; __pyx_t_3 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1864, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1864, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1864, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (__pyx_t_3) {
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_t_4);
          __pyx_t_4 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1864, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pandas/_libs/intervaltree.pxi":1873
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
  __pyx_t_8 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_8) {

    /* "pandas/_libs/intervaltree.pxi":1877
 *             # continue the binary tree structure. Instead, we use linear
 *             # search.
 *             for i in range(self.n_elements):             # <<<<<<<<<<<<<<
 *                 if self.left[i] < point <= self.right[i]:
 *                     result.append(self.indices[i])
 */
    __pyx_t_9 = __pyx_v_self->n_elements;
    __pyx_t_10 = __pyx_t_9;
    for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
      __pyx_v_i = __pyx_t_11;

      /* "pandas/_libs/intervaltree.pxi":1878
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point <= self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      __pyx_t_12 = __pyx_v_i;
      __pyx_t_8 = ((*((__pyx_t_5numpy_int32_t *) ( /* dim=0 */ (__pyx_v_self->left.data + __pyx_t_12 * __pyx_v_self->left.strides[0]) ))) < __pyx_v_point);
      if (__pyx_t_8) {
        __pyx_t_13 = __pyx_v_i;
        __pyx_t_8 = (__pyx_v_point <= (*((__pyx_t_5numpy_int32_t *) ( /* dim=0 */ (__pyx_v_self->right.data + __pyx_t_13 * __pyx_v_self->right.strides[0]) ))));
      }
      __pyx_t_14 = (__pyx_t_8 != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":1879
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point <= self.right[i]:
 *                     result.append(self.indices[i])             # <<<<<<<<<<<<<<
 *         else:
 *             # There are child nodes. Based on comparing our query to the pivot,
 */
        __pyx_t_15 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->indices.data + __pyx_t_15 * __pyx_v_self->indices.strides[0]) ))));

        /* "pandas/_libs/intervaltree.pxi":1878
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point <= self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      }
    }

    /* "pandas/_libs/intervaltree.pxi":1873
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
    goto __pyx_L3;
  }

  /* "pandas/_libs/intervaltree.pxi":1883
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
  /*else*/ {
    __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":1884
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:
 *                 values = self.center_left_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 */
      __pyx_t_16 = __pyx_v_self->center_left_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":1885
 *             if point < self.pivot:
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:
 */
      __pyx_t_17 = __pyx_v_self->center_left_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":1886
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):             # <<<<<<<<<<<<<<
 *                     if not values[i] < point:
 *                         break
 */
      __pyx_t_9 = __pyx_v_self->n_center;
      __pyx_t_10 = __pyx_t_9;
      for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":1887
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_18 = __pyx_v_i;
        __pyx_t_14 = ((!(((*((__pyx_t_5numpy_int32_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_18 * __pyx_v_values.strides[0]) ))) < __pyx_v_point) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":1888
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:
 */
          goto __pyx_L9_break;

          /* "pandas/_libs/intervaltree.pxi":1887
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":1889
 *                     if not values[i] < point:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 */
        __pyx_t_19 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_19 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L9_break:;

      /* "pandas/_libs/intervaltree.pxi":1890
 *                         break
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      __pyx_t_14 = ((__pyx_v_point <= __pyx_v_self->left_node->max_right) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":1891
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)             # <<<<<<<<<<<<<<
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode *)__pyx_v_self->left_node->__pyx_vtab)->__pyx_fuse_1query(__pyx_v_self->left_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1891, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":1890
 *                         break
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":1883
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":1892
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
    __pyx_t_14 = ((__pyx_v_point > __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":1893
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 *                 values = self.center_right_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 */
      __pyx_t_16 = __pyx_v_self->center_right_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":1894
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:
 */
      __pyx_t_17 = __pyx_v_self->center_right_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":1895
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):             # <<<<<<<<<<<<<<
 *                     if not point <= values[i]:
 *                         break
 */
      for (__pyx_t_11 = (__pyx_v_self->n_center - 1); __pyx_t_11 > -1L; __pyx_t_11-=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":1896
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_20 = __pyx_v_i;
        __pyx_t_14 = ((!((__pyx_v_point <= (*((__pyx_t_5numpy_int32_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_20 * __pyx_v_values.strides[0]) )))) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":1897
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:
 */
          goto __pyx_L13_break;

          /* "pandas/_libs/intervaltree.pxi":1896
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":1898
 *                     if not point <= values[i]:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if self.right_node.min_left < point:
 *                     self.right_node.query(result, point)
 */
        __pyx_t_21 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_21 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L13_break:;

      /* "pandas/_libs/intervaltree.pxi":1899
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      __pyx_t_14 = ((__pyx_v_self->right_node->min_left < __pyx_v_point) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":1900
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:
 *                     self.right_node.query(result, point)             # <<<<<<<<<<<<<<
 *             else:
 *                 result.extend(self.center_left_indices)
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode *)__pyx_v_self->right_node->__pyx_vtab)->__pyx_fuse_1query(__pyx_v_self->right_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1900, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":1899
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":1892
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":1902
 *                     self.right_node.query(result, point)
 *             else:
 *                 result.extend(self.center_left_indices)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
    /*else*/ {
      __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->extend(__pyx_v_result, __pyx_v_self->center_left_indices); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1902, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_L7:;
  }
  __pyx_L3:;

  /* "pandas/_libs/intervaltree.pxi":1864
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __PYX_XDEC_MEMVIEW(&__pyx_t_16, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_17, 1);
  __Pyx_AddTraceback("pandas._libs.interval.Int32ClosedRightIntervalNode.query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_indices, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_values, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_11__pyx_fuse_1query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_1__pyx_mdef_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_11__pyx_fuse_1query = {"__pyx_fuse_1query", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_11__pyx_fuse_1query, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_2query};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_11__pyx_fuse_1query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result = 0;
  __pyx_t_5numpy_float32_t __pyx_v_point;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fuse_1query (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_result,&__pyx_n_s_point,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_result)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_point)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_1query", 1, 2, 2, 1); __PYX_ERR(1, 1864, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_1query") < 0)) __PYX_ERR(1, 1864, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_result = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)values[0]);
    __pyx_v_point = __pyx_PyFloat_AsFloat(values[1]); if (unlikely((__pyx_v_point == ((npy_float32)-1)) && PyErr_Occurred())) __PYX_ERR(1, 1864, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_1query", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 1864, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Int32ClosedRightIntervalNode.__pyx_fuse_1query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_result), __pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector, 1, "result", 0))) __PYX_ERR(1, 1864, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_10__pyx_fuse_1query(((struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode *)__pyx_v_self), __pyx_v_result, __pyx_v_point);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_10__pyx_fuse_1query(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float32_t __pyx_v_point) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__pyx_fuse_1query", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_fuse_1__pyx_f_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_query(__pyx_v_self, __pyx_v_result, __pyx_v_point, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1864, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Int32ClosedRightIntervalNode.__pyx_fuse_1query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_13__pyx_fuse_2query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_fuse_2__pyx_f_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int64_t __pyx_v_point, int __pyx_skip_dispatch) {
  __Pyx_memviewslice __pyx_v_indices = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_values = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __pyx_t_5numpy_int64_t __pyx_t_9;
  __pyx_t_5numpy_int64_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  int __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  __Pyx_memviewslice __pyx_t_16 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_17 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  Py_ssize_t __pyx_t_21;
  __Pyx_RefNannySetupContext("__pyx_fuse_2query", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pyx_fuse_2query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1864, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_13__pyx_fuse_2query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = __Pyx_PyInt_From_npy_int64(__pyx_v_point); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1864, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1864, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1864, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 1864, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_5) {
            __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1864, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1864, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_3query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_4 = __Pyx_PyInt_From_npy_int64(__pyx_v_point); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1864, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_7 = __pyx_t_1; __pyx_t_3 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1864, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1864, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1864, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (__pyx_t_3) {
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_t_4);
          __pyx_t_4 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1864, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pandas/_libs/intervaltree.pxi":1873
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
  __pyx_t_8 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_8) {

    /* "pandas/_libs/intervaltree.pxi":1877
 *             # continue the binary tree structure. Instead, we use linear
 *             # search.
 *             for i in range(self.n_elements):             # <<<<<<<<<<<<<<
 *                 if self.left[i] < point <= self.right[i]:
 *                     result.append(self.indices[i])
 */
    __pyx_t_9 = __pyx_v_self->n_elements;
    __pyx_t_10 = __pyx_t_9;
    for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
      __pyx_v_i = __pyx_t_11;

      /* "pandas/_libs/intervaltree.pxi":1878
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point <= self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      __pyx_t_12 = __pyx_v_i;
      __pyx_t_8 = ((*((__pyx_t_5numpy_int32_t *) ( /* dim=0 */ (__pyx_v_self->left.data + __pyx_t_12 * __pyx_v_self->left.strides[0]) ))) < __pyx_v_point);
      if (__pyx_t_8) {
        __pyx_t_13 = __pyx_v_i;
        __pyx_t_8 = (__pyx_v_point <= (*((__pyx_t_5numpy_int32_t *) ( /* dim=0 */ (__pyx_v_self->right.data + __pyx_t_13 * __pyx_v_self->right.strides[0]) ))));
      }
      __pyx_t_14 = (__pyx_t_8 != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":1879
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point <= self.right[i]:
 *                     result.append(self.indices[i])             # <<<<<<<<<<<<<<
 *         else:
 *             # There are child nodes. Based on comparing our query to the pivot,
 */
        __pyx_t_15 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->indices.data + __pyx_t_15 * __pyx_v_self->indices.strides[0]) ))));

        /* "pandas/_libs/intervaltree.pxi":1878
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point <= self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      }
    }

    /* "pandas/_libs/intervaltree.pxi":1873
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
    goto __pyx_L3;
  }

  /* "pandas/_libs/intervaltree.pxi":1883
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
  /*else*/ {
    __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":1884
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:
 *                 values = self.center_left_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 */
      __pyx_t_16 = __pyx_v_self->center_left_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":1885
 *             if point < self.pivot:
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:
 */
      __pyx_t_17 = __pyx_v_self->center_left_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":1886
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):             # <<<<<<<<<<<<<<
 *                     if not values[i] < point:
 *                         break
 */
      __pyx_t_9 = __pyx_v_self->n_center;
      __pyx_t_10 = __pyx_t_9;
      for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":1887
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_18 = __pyx_v_i;
        __pyx_t_14 = ((!(((*((__pyx_t_5numpy_int32_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_18 * __pyx_v_values.strides[0]) ))) < __pyx_v_point) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":1888
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:
 */
          goto __pyx_L9_break;

          /* "pandas/_libs/intervaltree.pxi":1887
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":1889
 *                     if not values[i] < point:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 */
        __pyx_t_19 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_19 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L9_break:;

      /* "pandas/_libs/intervaltree.pxi":1890
 *                         break
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      __pyx_t_14 = ((__pyx_v_point <= __pyx_v_self->left_node->max_right) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":1891
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)             # <<<<<<<<<<<<<<
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode *)__pyx_v_self->left_node->__pyx_vtab)->__pyx_fuse_2query(__pyx_v_self->left_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1891, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":1890
 *                         break
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":1883
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":1892
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
    __pyx_t_14 = ((__pyx_v_point > __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":1893
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 *                 values = self.center_right_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 */
      __pyx_t_16 = __pyx_v_self->center_right_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":1894
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:
 */
      __pyx_t_17 = __pyx_v_self->center_right_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":1895
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):             # <<<<<<<<<<<<<<
 *                     if not point <= values[i]:
 *                         break
 */
      for (__pyx_t_11 = (__pyx_v_self->n_center - 1); __pyx_t_11 > -1L; __pyx_t_11-=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":1896
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_20 = __pyx_v_i;
        __pyx_t_14 = ((!((__pyx_v_point <= (*((__pyx_t_5numpy_int32_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_20 * __pyx_v_values.strides[0]) )))) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":1897
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:
 */
          goto __pyx_L13_break;

          /* "pandas/_libs/intervaltree.pxi":1896
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":1898
 *                     if not point <= values[i]:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if self.right_node.min_left < point:
 *                     self.right_node.query(result, point)
 */
        __pyx_t_21 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_21 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L13_break:;

      /* "pandas/_libs/intervaltree.pxi":1899
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      __pyx_t_14 = ((__pyx_v_self->right_node->min_left < __pyx_v_point) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":1900
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:
 *                     self.right_node.query(result, point)             # <<<<<<<<<<<<<<
 *             else:
 *                 result.extend(self.center_left_indices)
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode *)__pyx_v_self->right_node->__pyx_vtab)->__pyx_fuse_2query(__pyx_v_self->right_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1900, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":1899
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":1892
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":1902
 *                     self.right_node.query(result, point)
 *             else:
 *                 result.extend(self.center_left_indices)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
    /*else*/ {
      __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->extend(__pyx_v_result, __pyx_v_self->center_left_indices); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1902, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_L7:;
  }
  __pyx_L3:;

  /* "pandas/_libs/intervaltree.pxi":1864
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __PYX_XDEC_MEMVIEW(&__pyx_t_16, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_17, 1);
  __Pyx_AddTraceback("pandas._libs.interval.Int32ClosedRightIntervalNode.query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_indices, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_values, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_13__pyx_fuse_2query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_2__pyx_mdef_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_13__pyx_fuse_2query = {"__pyx_fuse_2query", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_13__pyx_fuse_2query, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_2query};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_13__pyx_fuse_2query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result = 0;
  __pyx_t_5numpy_int64_t __pyx_v_point;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fuse_2query (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_result,&__pyx_n_s_point,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_result)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_point)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_2query", 1, 2, 2, 1); __PYX_ERR(1, 1864, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_2query") < 0)) __PYX_ERR(1, 1864, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_result = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)values[0]);
    __pyx_v_point = __Pyx_PyInt_As_npy_int64(values[1]); if (unlikely((__pyx_v_point == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(1, 1864, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_2query", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 1864, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Int32ClosedRightIntervalNode.__pyx_fuse_2query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_result), __pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector, 1, "result", 0))) __PYX_ERR(1, 1864, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_12__pyx_fuse_2query(((struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode *)__pyx_v_self), __pyx_v_result, __pyx_v_point);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_12__pyx_fuse_2query(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int64_t __pyx_v_point) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__pyx_fuse_2query", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_fuse_2__pyx_f_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_query(__pyx_v_self, __pyx_v_result, __pyx_v_point, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1864, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Int32ClosedRightIntervalNode.__pyx_fuse_2query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_15__pyx_fuse_3query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_fuse_3__pyx_f_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int32_t __pyx_v_point, int __pyx_skip_dispatch) {
  __Pyx_memviewslice __pyx_v_indices = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_values = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __pyx_t_5numpy_int64_t __pyx_t_9;
  __pyx_t_5numpy_int64_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  int __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  __Pyx_memviewslice __pyx_t_16 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_17 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  Py_ssize_t __pyx_t_21;
  __Pyx_RefNannySetupContext("__pyx_fuse_3query", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pyx_fuse_3query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1864, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_15__pyx_fuse_3query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = __Pyx_PyInt_From_npy_int32(__pyx_v_point); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1864, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1864, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1864, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 1864, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_5) {
            __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1864, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1864, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_3query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_4 = __Pyx_PyInt_From_npy_int32(__pyx_v_point); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1864, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_7 = __pyx_t_1; __pyx_t_3 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1864, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1864, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1864, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (__pyx_t_3) {
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_t_4);
          __pyx_t_4 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1864, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pandas/_libs/intervaltree.pxi":1873
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
  __pyx_t_8 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_8) {

    /* "pandas/_libs/intervaltree.pxi":1877
 *             # continue the binary tree structure. Instead, we use linear
 *             # search.
 *             for i in range(self.n_elements):             # <<<<<<<<<<<<<<
 *                 if self.left[i] < point <= self.right[i]:
 *                     result.append(self.indices[i])
 */
    __pyx_t_9 = __pyx_v_self->n_elements;
    __pyx_t_10 = __pyx_t_9;
    for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
      __pyx_v_i = __pyx_t_11;

      /* "pandas/_libs/intervaltree.pxi":1878
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point <= self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      __pyx_t_12 = __pyx_v_i;
      __pyx_t_8 = ((*((__pyx_t_5numpy_int32_t *) ( /* dim=0 */ (__pyx_v_self->left.data + __pyx_t_12 * __pyx_v_self->left.strides[0]) ))) < __pyx_v_point);
      if (__pyx_t_8) {
        __pyx_t_13 = __pyx_v_i;
        __pyx_t_8 = (__pyx_v_point <= (*((__pyx_t_5numpy_int32_t *) ( /* dim=0 */ (__pyx_v_self->right.data + __pyx_t_13 * __pyx_v_self->right.strides[0]) ))));
      }
      __pyx_t_14 = (__pyx_t_8 != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":1879
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point <= self.right[i]:
 *                     result.append(self.indices[i])             # <<<<<<<<<<<<<<
 *         else:
 *             # There are child nodes. Based on comparing our query to the pivot,
 */
        __pyx_t_15 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->indices.data + __pyx_t_15 * __pyx_v_self->indices.strides[0]) ))));

        /* "pandas/_libs/intervaltree.pxi":1878
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point <= self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      }
    }

    /* "pandas/_libs/intervaltree.pxi":1873
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
    goto __pyx_L3;
  }

  /* "pandas/_libs/intervaltree.pxi":1883
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
  /*else*/ {
    __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":1884
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:
 *                 values = self.center_left_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 */
      __pyx_t_16 = __pyx_v_self->center_left_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":1885
 *             if point < self.pivot:
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:
 */
      __pyx_t_17 = __pyx_v_self->center_left_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":1886
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):             # <<<<<<<<<<<<<<
 *                     if not values[i] < point:
 *                         break
 */
      __pyx_t_9 = __pyx_v_self->n_center;
      __pyx_t_10 = __pyx_t_9;
      for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":1887
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_18 = __pyx_v_i;
        __pyx_t_14 = ((!(((*((__pyx_t_5numpy_int32_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_18 * __pyx_v_values.strides[0]) ))) < __pyx_v_point) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":1888
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:
 */
          goto __pyx_L9_break;

          /* "pandas/_libs/intervaltree.pxi":1887
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":1889
 *                     if not values[i] < point:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 */
        __pyx_t_19 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_19 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L9_break:;

      /* "pandas/_libs/intervaltree.pxi":1890
 *                         break
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      __pyx_t_14 = ((__pyx_v_point <= __pyx_v_self->left_node->max_right) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":1891
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)             # <<<<<<<<<<<<<<
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode *)__pyx_v_self->left_node->__pyx_vtab)->__pyx_fuse_3query(__pyx_v_self->left_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1891, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":1890
 *                         break
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":1883
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":1892
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
    __pyx_t_14 = ((__pyx_v_point > __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":1893
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 *                 values = self.center_right_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 */
      __pyx_t_16 = __pyx_v_self->center_right_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":1894
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:
 */
      __pyx_t_17 = __pyx_v_self->center_right_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":1895
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):             # <<<<<<<<<<<<<<
 *                     if not point <= values[i]:
 *                         break
 */
      for (__pyx_t_11 = (__pyx_v_self->n_center - 1); __pyx_t_11 > -1L; __pyx_t_11-=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":1896
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_20 = __pyx_v_i;
        __pyx_t_14 = ((!((__pyx_v_point <= (*((__pyx_t_5numpy_int32_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_20 * __pyx_v_values.strides[0]) )))) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":1897
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:
 */
          goto __pyx_L13_break;

          /* "pandas/_libs/intervaltree.pxi":1896
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":1898
 *                     if not point <= values[i]:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if self.right_node.min_left < point:
 *                     self.right_node.query(result, point)
 */
        __pyx_t_21 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_21 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L13_break:;

      /* "pandas/_libs/intervaltree.pxi":1899
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      __pyx_t_14 = ((__pyx_v_self->right_node->min_left < __pyx_v_point) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":1900
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:
 *                     self.right_node.query(result, point)             # <<<<<<<<<<<<<<
 *             else:
 *                 result.extend(self.center_left_indices)
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode *)__pyx_v_self->right_node->__pyx_vtab)->__pyx_fuse_3query(__pyx_v_self->right_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1900, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":1899
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":1892
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":1902
 *                     self.right_node.query(result, point)
 *             else:
 *                 result.extend(self.center_left_indices)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
    /*else*/ {
      __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->extend(__pyx_v_result, __pyx_v_self->center_left_indices); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1902, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_L7:;
  }
  __pyx_L3:;

  /* "pandas/_libs/intervaltree.pxi":1864
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __PYX_XDEC_MEMVIEW(&__pyx_t_16, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_17, 1);
  __Pyx_AddTraceback("pandas._libs.interval.Int32ClosedRightIntervalNode.query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_indices, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_values, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_15__pyx_fuse_3query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_3__pyx_mdef_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_15__pyx_fuse_3query = {"__pyx_fuse_3query", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_15__pyx_fuse_3query, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_2query};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_15__pyx_fuse_3query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result = 0;
  __pyx_t_5numpy_int32_t __pyx_v_point;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fuse_3query (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_result,&__pyx_n_s_point,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_result)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_point)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_3query", 1, 2, 2, 1); __PYX_ERR(1, 1864, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_3query") < 0)) __PYX_ERR(1, 1864, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_result = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)values[0]);
    __pyx_v_point = __Pyx_PyInt_As_npy_int32(values[1]); if (unlikely((__pyx_v_point == ((npy_int32)-1)) && PyErr_Occurred())) __PYX_ERR(1, 1864, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_3query", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 1864, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Int32ClosedRightIntervalNode.__pyx_fuse_3query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_result), __pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector, 1, "result", 0))) __PYX_ERR(1, 1864, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_14__pyx_fuse_3query(((struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode *)__pyx_v_self), __pyx_v_result, __pyx_v_point);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_14__pyx_fuse_3query(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int32_t __pyx_v_point) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__pyx_fuse_3query", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_fuse_3__pyx_f_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_query(__pyx_v_self, __pyx_v_result, __pyx_v_point, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1864, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Int32ClosedRightIntervalNode.__pyx_fuse_3query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_17__pyx_fuse_4query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_fuse_4__pyx_f_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_uint64_t __pyx_v_point, int __pyx_skip_dispatch) {
  __Pyx_memviewslice __pyx_v_indices = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_values = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __pyx_t_5numpy_int64_t __pyx_t_9;
  __pyx_t_5numpy_int64_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  int __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  __Pyx_memviewslice __pyx_t_16 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_17 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  Py_ssize_t __pyx_t_21;
  __Pyx_RefNannySetupContext("__pyx_fuse_4query", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pyx_fuse_4query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1864, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_17__pyx_fuse_4query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = __Pyx_PyInt_From_npy_uint64(__pyx_v_point); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1864, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1864, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1864, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 1864, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_5) {
            __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1864, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1864, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_3query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_4 = __Pyx_PyInt_From_npy_uint64(__pyx_v_point); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1864, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_7 = __pyx_t_1; __pyx_t_3 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1864, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1864, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1864, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (__pyx_t_3) {
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_t_4);
          __pyx_t_4 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1864, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pandas/_libs/intervaltree.pxi":1873
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
  __pyx_t_8 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_8) {

    /* "pandas/_libs/intervaltree.pxi":1877
 *             # continue the binary tree structure. Instead, we use linear
 *             # search.
 *             for i in range(self.n_elements):             # <<<<<<<<<<<<<<
 *                 if self.left[i] < point <= self.right[i]:
 *                     result.append(self.indices[i])
 */
    __pyx_t_9 = __pyx_v_self->n_elements;
    __pyx_t_10 = __pyx_t_9;
    for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
      __pyx_v_i = __pyx_t_11;

      /* "pandas/_libs/intervaltree.pxi":1878
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point <= self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      __pyx_t_12 = __pyx_v_i;
      __pyx_t_8 = ((*((__pyx_t_5numpy_int32_t *) ( /* dim=0 */ (__pyx_v_self->left.data + __pyx_t_12 * __pyx_v_self->left.strides[0]) ))) < __pyx_v_point);
      if (__pyx_t_8) {
        __pyx_t_13 = __pyx_v_i;
        __pyx_t_8 = (__pyx_v_point <= (*((__pyx_t_5numpy_int32_t *) ( /* dim=0 */ (__pyx_v_self->right.data + __pyx_t_13 * __pyx_v_self->right.strides[0]) ))));
      }
      __pyx_t_14 = (__pyx_t_8 != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":1879
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point <= self.right[i]:
 *                     result.append(self.indices[i])             # <<<<<<<<<<<<<<
 *         else:
 *             # There are child nodes. Based on comparing our query to the pivot,
 */
        __pyx_t_15 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->indices.data + __pyx_t_15 * __pyx_v_self->indices.strides[0]) ))));

        /* "pandas/_libs/intervaltree.pxi":1878
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point <= self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      }
    }

    /* "pandas/_libs/intervaltree.pxi":1873
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
    goto __pyx_L3;
  }

  /* "pandas/_libs/intervaltree.pxi":1883
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
  /*else*/ {
    __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":1884
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:
 *                 values = self.center_left_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 */
      __pyx_t_16 = __pyx_v_self->center_left_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":1885
 *             if point < self.pivot:
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:
 */
      __pyx_t_17 = __pyx_v_self->center_left_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":1886
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):             # <<<<<<<<<<<<<<
 *                     if not values[i] < point:
 *                         break
 */
      __pyx_t_9 = __pyx_v_self->n_center;
      __pyx_t_10 = __pyx_t_9;
      for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":1887
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_18 = __pyx_v_i;
        __pyx_t_14 = ((!(((*((__pyx_t_5numpy_int32_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_18 * __pyx_v_values.strides[0]) ))) < __pyx_v_point) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":1888
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:
 */
          goto __pyx_L9_break;

          /* "pandas/_libs/intervaltree.pxi":1887
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":1889
 *                     if not values[i] < point:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 */
        __pyx_t_19 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_19 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L9_break:;

      /* "pandas/_libs/intervaltree.pxi":1890
 *                         break
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      __pyx_t_14 = ((__pyx_v_point <= __pyx_v_self->left_node->max_right) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":1891
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)             # <<<<<<<<<<<<<<
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode *)__pyx_v_self->left_node->__pyx_vtab)->__pyx_fuse_4query(__pyx_v_self->left_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1891, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":1890
 *                         break
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":1883
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":1892
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
    __pyx_t_14 = ((__pyx_v_point > __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":1893
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 *                 values = self.center_right_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 */
      __pyx_t_16 = __pyx_v_self->center_right_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":1894
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:
 */
      __pyx_t_17 = __pyx_v_self->center_right_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":1895
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):             # <<<<<<<<<<<<<<
 *                     if not point <= values[i]:
 *                         break
 */
      for (__pyx_t_11 = (__pyx_v_self->n_center - 1); __pyx_t_11 > -1L; __pyx_t_11-=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":1896
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_20 = __pyx_v_i;
        __pyx_t_14 = ((!((__pyx_v_point <= (*((__pyx_t_5numpy_int32_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_20 * __pyx_v_values.strides[0]) )))) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":1897
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:
 */
          goto __pyx_L13_break;

          /* "pandas/_libs/intervaltree.pxi":1896
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":1898
 *                     if not point <= values[i]:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if self.right_node.min_left < point:
 *                     self.right_node.query(result, point)
 */
        __pyx_t_21 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_21 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L13_break:;

      /* "pandas/_libs/intervaltree.pxi":1899
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      __pyx_t_14 = ((__pyx_v_self->right_node->min_left < __pyx_v_point) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":1900
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:
 *                     self.right_node.query(result, point)             # <<<<<<<<<<<<<<
 *             else:
 *                 result.extend(self.center_left_indices)
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode *)__pyx_v_self->right_node->__pyx_vtab)->__pyx_fuse_4query(__pyx_v_self->right_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1900, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":1899
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":1892
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":1902
 *                     self.right_node.query(result, point)
 *             else:
 *                 result.extend(self.center_left_indices)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
    /*else*/ {
      __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->extend(__pyx_v_result, __pyx_v_self->center_left_indices); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1902, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_L7:;
  }
  __pyx_L3:;

  /* "pandas/_libs/intervaltree.pxi":1864
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __PYX_XDEC_MEMVIEW(&__pyx_t_16, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_17, 1);
  __Pyx_AddTraceback("pandas._libs.interval.Int32ClosedRightIntervalNode.query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_indices, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_values, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_17__pyx_fuse_4query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_4__pyx_mdef_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_17__pyx_fuse_4query = {"__pyx_fuse_4query", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_17__pyx_fuse_4query, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_2query};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_17__pyx_fuse_4query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result = 0;
  __pyx_t_5numpy_uint64_t __pyx_v_point;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fuse_4query (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_result,&__pyx_n_s_point,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_result)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_point)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_4query", 1, 2, 2, 1); __PYX_ERR(1, 1864, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_4query") < 0)) __PYX_ERR(1, 1864, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_result = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)values[0]);
    __pyx_v_point = __Pyx_PyInt_As_npy_uint64(values[1]); if (unlikely((__pyx_v_point == ((npy_uint64)-1)) && PyErr_Occurred())) __PYX_ERR(1, 1864, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_4query", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 1864, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Int32ClosedRightIntervalNode.__pyx_fuse_4query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_result), __pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector, 1, "result", 0))) __PYX_ERR(1, 1864, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_16__pyx_fuse_4query(((struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode *)__pyx_v_self), __pyx_v_result, __pyx_v_point);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_16__pyx_fuse_4query(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_uint64_t __pyx_v_point) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__pyx_fuse_4query", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_fuse_4__pyx_f_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_query(__pyx_v_self, __pyx_v_result, __pyx_v_point, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1864, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Int32ClosedRightIntervalNode.__pyx_fuse_4query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":1904
 *                 result.extend(self.center_left_indices)
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         if self.is_leaf_node:
 *             return ('<Int32ClosedRightIntervalNode: '
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_5__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_5__repr__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_4__repr__(((struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_4__repr__(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode *__pyx_v_self) {
  PyObject *__pyx_v_n_left = NULL;
  PyObject *__pyx_v_n_right = NULL;
  PyObject *__pyx_v_n_center = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  Py_UCS4 __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "pandas/_libs/intervaltree.pxi":1905
 * 
 *     def __repr__(self):
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             return ('<Int32ClosedRightIntervalNode: '
 *                     '%s elements (terminal)>' % self.n_elements)
 */
  __pyx_t_1 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_1) {

    /* "pandas/_libs/intervaltree.pxi":1906
 *     def __repr__(self):
 *         if self.is_leaf_node:
 *             return ('<Int32ClosedRightIntervalNode: '             # <<<<<<<<<<<<<<
 *                     '%s elements (terminal)>' % self.n_elements)
 *         else:
 */
    __Pyx_XDECREF(__pyx_r);

    /* "pandas/_libs/intervaltree.pxi":1907
 *         if self.is_leaf_node:
 *             return ('<Int32ClosedRightIntervalNode: '
 *                     '%s elements (terminal)>' % self.n_elements)             # <<<<<<<<<<<<<<
 *         else:
 *             n_left = self.left_node.n_elements
 */
    __pyx_t_2 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_elements); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1907, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PyUnicode_Format(__pyx_kp_u_Int32ClosedRightIntervalNode_s, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1907, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pandas/_libs/intervaltree.pxi":1905
 * 
 *     def __repr__(self):
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             return ('<Int32ClosedRightIntervalNode: '
 *                     '%s elements (terminal)>' % self.n_elements)
 */
  }

  /* "pandas/_libs/intervaltree.pxi":1909
 *                     '%s elements (terminal)>' % self.n_elements)
 *         else:
 *             n_left = self.left_node.n_elements             # <<<<<<<<<<<<<<
 *             n_right = self.right_node.n_elements
 *             n_center = self.n_elements - n_left - n_right
 */
  /*else*/ {
    __pyx_t_3 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->left_node->n_elements); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1909, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_v_n_left = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "pandas/_libs/intervaltree.pxi":1910
 *         else:
 *             n_left = self.left_node.n_elements
 *             n_right = self.right_node.n_elements             # <<<<<<<<<<<<<<
 *             n_center = self.n_elements - n_left - n_right
 *             return ('<Int32ClosedRightIntervalNode: '
 */
    __pyx_t_3 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->right_node->n_elements); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1910, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_v_n_right = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "pandas/_libs/intervaltree.pxi":1911
 *             n_left = self.left_node.n_elements
 *             n_right = self.right_node.n_elements
 *             n_center = self.n_elements - n_left - n_right             # <<<<<<<<<<<<<<
 *             return ('<Int32ClosedRightIntervalNode: '
 *                     'pivot %s, %s elements (%s left, %s right, %s '
 */
    __pyx_t_3 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_elements); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1911, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = PyNumber_Subtract(__pyx_t_3, __pyx_v_n_left); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1911, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyNumber_Subtract(__pyx_t_2, __pyx_v_n_right); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1911, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_n_center = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "pandas/_libs/intervaltree.pxi":1912
 *             n_right = self.right_node.n_elements
 *             n_center = self.n_elements - n_left - n_right
 *             return ('<Int32ClosedRightIntervalNode: '             # <<<<<<<<<<<<<<
 *                     'pivot %s, %s elements (%s left, %s right, %s '
 *                     'overlapping)>' % (self.pivot, self.n_elements,
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = PyTuple_New(11); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1912, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = 0;
    __pyx_t_5 = 127;
    __Pyx_INCREF(__pyx_kp_u_Int32ClosedRightIntervalNode_pi);
    __pyx_t_4 += 37;
    __Pyx_GIVEREF(__pyx_kp_u_Int32ClosedRightIntervalNode_pi);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_kp_u_Int32ClosedRightIntervalNode_pi);

    /* "pandas/_libs/intervaltree.pxi":1914
 *             return ('<Int32ClosedRightIntervalNode: '
 *                     'pivot %s, %s elements (%s left, %s right, %s '
 *                     'overlapping)>' % (self.pivot, self.n_elements,             # <<<<<<<<<<<<<<
 *                                        n_left, n_right, n_center))
 * 
 */
    __pyx_t_2 = __Pyx_PyInt_From_npy_int32(__pyx_v_self->pivot); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1914, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_t_2), __pyx_empty_unicode); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 1914, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_5 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_6) > __pyx_t_5) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_6) : __pyx_t_5;
    __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_6);
    __pyx_t_6 = 0;
    __Pyx_INCREF(__pyx_kp_u__18);
    __pyx_t_4 += 2;
    __Pyx_GIVEREF(__pyx_kp_u__18);
    PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_kp_u__18);
    __pyx_t_6 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_elements); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 1914, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_2 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_t_6), __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1914, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_5 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) > __pyx_t_5) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) : __pyx_t_5;
    __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_3, 3, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_elements);
    __pyx_t_4 += 11;
    __Pyx_GIVEREF(__pyx_kp_u_elements);
    PyTuple_SET_ITEM(__pyx_t_3, 4, __pyx_kp_u_elements);

    /* "pandas/_libs/intervaltree.pxi":1915
 *                     'pivot %s, %s elements (%s left, %s right, %s '
 *                     'overlapping)>' % (self.pivot, self.n_elements,
 *                                        n_left, n_right, n_center))             # <<<<<<<<<<<<<<
 * 
 *     def counts(self):
 */
    __pyx_t_2 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_v_n_left), __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1915, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) > __pyx_t_5) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) : __pyx_t_5;
    __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_3, 5, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_left_2);
    __pyx_t_4 += 7;
    __Pyx_GIVEREF(__pyx_kp_u_left_2);
    PyTuple_SET_ITEM(__pyx_t_3, 6, __pyx_kp_u_left_2);
    __pyx_t_2 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_v_n_right), __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1915, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) > __pyx_t_5) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) : __pyx_t_5;
    __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_3, 7, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_right_2);
    __pyx_t_4 += 8;
    __Pyx_GIVEREF(__pyx_kp_u_right_2);
    PyTuple_SET_ITEM(__pyx_t_3, 8, __pyx_kp_u_right_2);
    __pyx_t_2 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_v_n_center), __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1915, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) > __pyx_t_5) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) : __pyx_t_5;
    __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_3, 9, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_overlapping);
    __pyx_t_4 += 14;
    __Pyx_GIVEREF(__pyx_kp_u_overlapping);
    PyTuple_SET_ITEM(__pyx_t_3, 10, __pyx_kp_u_overlapping);

    /* "pandas/_libs/intervaltree.pxi":1912
 *             n_right = self.right_node.n_elements
 *             n_center = self.n_elements - n_left - n_right
 *             return ('<Int32ClosedRightIntervalNode: '             # <<<<<<<<<<<<<<
 *                     'pivot %s, %s elements (%s left, %s right, %s '
 *                     'overlapping)>' % (self.pivot, self.n_elements,
 */
    __pyx_t_2 = __Pyx_PyUnicode_Join(__pyx_t_3, 11, __pyx_t_4, __pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1912, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;
  }

  /* "pandas/_libs/intervaltree.pxi":1904
 *                 result.extend(self.center_left_indices)
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         if self.is_leaf_node:
 *             return ('<Int32ClosedRightIntervalNode: '
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pandas._libs.interval.Int32ClosedRightIntervalNode.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_n_left);
  __Pyx_XDECREF(__pyx_v_n_right);
  __Pyx_XDECREF(__pyx_v_n_center);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":1917
 *                                        n_left, n_right, n_center))
 * 
 *     def counts(self):             # <<<<<<<<<<<<<<
 *         """
 *         Inspect counts on this node
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_7counts(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_6counts[] = "\n        Inspect counts on this node\n        useful for debugging purposes\n        ";
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_7counts(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("counts (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_6counts(((struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_6counts(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode *__pyx_v_self) {
  Py_ssize_t __pyx_v_m;
  PyObject *__pyx_v_l = NULL;
  PyObject *__pyx_v_r = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  size_t __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("counts", 0);

  /* "pandas/_libs/intervaltree.pxi":1922
 *         useful for debugging purposes
 *         """
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             return self.n_elements
 *         else:
 */
  __pyx_t_1 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_1) {

    /* "pandas/_libs/intervaltree.pxi":1923
 *         """
 *         if self.is_leaf_node:
 *             return self.n_elements             # <<<<<<<<<<<<<<
 *         else:
 *             m = len(self.center_left_values)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_elements); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1923, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "pandas/_libs/intervaltree.pxi":1922
 *         useful for debugging purposes
 *         """
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             return self.n_elements
 *         else:
 */
  }

  /* "pandas/_libs/intervaltree.pxi":1925
 *             return self.n_elements
 *         else:
 *             m = len(self.center_left_values)             # <<<<<<<<<<<<<<
 *             l = self.left_node.counts()
 *             r = self.right_node.counts()
 */
  /*else*/ {
    if (unlikely(!__pyx_v_self->center_left_values.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(1, 1925, __pyx_L1_error)}
    __pyx_t_3 = __Pyx_MemoryView_Len(__pyx_v_self->center_left_values); 
    __pyx_v_m = __pyx_t_3;

    /* "pandas/_libs/intervaltree.pxi":1926
 *         else:
 *             m = len(self.center_left_values)
 *             l = self.left_node.counts()             # <<<<<<<<<<<<<<
 *             r = self.right_node.counts()
 *             return (m, (l, r))
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->left_node), __pyx_n_s_counts); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1926, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1926, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_l = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "pandas/_libs/intervaltree.pxi":1927
 *             m = len(self.center_left_values)
 *             l = self.left_node.counts()
 *             r = self.right_node.counts()             # <<<<<<<<<<<<<<
 *             return (m, (l, r))
 * 
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->right_node), __pyx_n_s_counts); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1927, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1927, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_r = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "pandas/_libs/intervaltree.pxi":1928
 *             l = self.left_node.counts()
 *             r = self.right_node.counts()
 *             return (m, (l, r))             # <<<<<<<<<<<<<<
 * 
 * NODE_CLASSES['int32',
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = PyInt_FromSsize_t(__pyx_v_m); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1928, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1928, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_v_l);
    __Pyx_GIVEREF(__pyx_v_l);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_l);
    __Pyx_INCREF(__pyx_v_r);
    __Pyx_GIVEREF(__pyx_v_r);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_v_r);
    __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1928, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_4);
    __pyx_t_2 = 0;
    __pyx_t_4 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;
  }

  /* "pandas/_libs/intervaltree.pxi":1917
 *                                        n_left, n_right, n_center))
 * 
 *     def counts(self):             # <<<<<<<<<<<<<<
 *         """
 *         Inspect counts on this node
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pandas._libs.interval.Int32ClosedRightIntervalNode.counts", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_l);
  __Pyx_XDECREF(__pyx_v_r);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":1774
 *         int64_t[:] center_left_indices, center_right_indices, indices
 *         int32_t min_left, max_right
 *         readonly int32_t pivot             # <<<<<<<<<<<<<<
 *         readonly int64_t n_elements, n_center, leaf_size
 *         readonly bint is_leaf_node
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_5pivot_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_5pivot_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_5pivot___get__(((struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_5pivot___get__(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_npy_int32(__pyx_v_self->pivot); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1774, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Int32ClosedRightIntervalNode.pivot.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":1775
 *         int32_t min_left, max_right
 *         readonly int32_t pivot
 *         readonly int64_t n_elements, n_center, leaf_size             # <<<<<<<<<<<<<<
 *         readonly bint is_leaf_node
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_10n_elements_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_10n_elements_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_10n_elements___get__(((struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_10n_elements___get__(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_elements); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1775, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Int32ClosedRightIntervalNode.n_elements.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_8n_center_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_8n_center_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_8n_center___get__(((struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_8n_center___get__(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_center); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1775, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Int32ClosedRightIntervalNode.n_center.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_9leaf_size_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_9leaf_size_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_9leaf_size___get__(((struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_9leaf_size___get__(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->leaf_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1775, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Int32ClosedRightIntervalNode.leaf_size.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":1776
 *         readonly int32_t pivot
 *         readonly int64_t n_elements, n_center, leaf_size
 *         readonly bint is_leaf_node             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self,
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_12is_leaf_node_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_12is_leaf_node_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_12is_leaf_node___get__(((struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_12is_leaf_node___get__(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->is_leaf_node); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1776, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Int32ClosedRightIntervalNode.is_leaf_node.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_21__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_21__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_20__reduce_cython__(((struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_20__reduce_cython__(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  int __pyx_t_16;
  int __pyx_t_17;
  int __pyx_t_18;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.center_left_indices, self.center_left_values, self.center_right_indices, self.center_right_values, self.indices, self.is_leaf_node, self.leaf_size, self.left, self.left_node, self.max_right, self.min_left, self.n_center, self.n_elements, self.pivot, self.right, self.right_node)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  if (unlikely(!__pyx_v_self->center_left_indices.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_1 = __pyx_memoryview_fromslice(__pyx_v_self->center_left_indices, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_int64_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_int64_t, 0);; if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (unlikely(!__pyx_v_self->center_left_values.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_2 = __pyx_memoryview_fromslice(__pyx_v_self->center_left_values, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_int32_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_int32_t, 0);; if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (unlikely(!__pyx_v_self->center_right_indices.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_3 = __pyx_memoryview_fromslice(__pyx_v_self->center_right_indices, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_int64_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_int64_t, 0);; if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely(!__pyx_v_self->center_right_values.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_4 = __pyx_memoryview_fromslice(__pyx_v_self->center_right_values, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_int32_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_int32_t, 0);; if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (unlikely(!__pyx_v_self->indices.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_5 = __pyx_memoryview_fromslice(__pyx_v_self->indices, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_int64_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_int64_t, 0);; if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyBool_FromLong(__pyx_v_self->is_leaf_node); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->leaf_size); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (unlikely(!__pyx_v_self->left.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_8 = __pyx_memoryview_fromslice(__pyx_v_self->left, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_int32_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_int32_t, 0);; if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_PyInt_From_npy_int32(__pyx_v_self->max_right); if (unlikely(!__pyx_t_9)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_10 = __Pyx_PyInt_From_npy_int32(__pyx_v_self->min_left); if (unlikely(!__pyx_t_10)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_11 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_center); if (unlikely(!__pyx_t_11)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_12 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_elements); if (unlikely(!__pyx_t_12)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __pyx_t_13 = __Pyx_PyInt_From_npy_int32(__pyx_v_self->pivot); if (unlikely(!__pyx_t_13)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  if (unlikely(!__pyx_v_self->right.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_14 = __pyx_memoryview_fromslice(__pyx_v_self->right, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_int32_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_int32_t, 0);; if (unlikely(!__pyx_t_14)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __pyx_t_15 = PyTuple_New(16); if (unlikely(!__pyx_t_15)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_15, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_15, 1, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_15, 2, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_15, 3, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_15, 4, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_15, 5, __pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_15, 6, __pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_15, 7, __pyx_t_8);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->left_node));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self->left_node));
  PyTuple_SET_ITEM(__pyx_t_15, 8, ((PyObject *)__pyx_v_self->left_node));
  __Pyx_GIVEREF(__pyx_t_9);
  PyTuple_SET_ITEM(__pyx_t_15, 9, __pyx_t_9);
  __Pyx_GIVEREF(__pyx_t_10);
  PyTuple_SET_ITEM(__pyx_t_15, 10, __pyx_t_10);
  __Pyx_GIVEREF(__pyx_t_11);
  PyTuple_SET_ITEM(__pyx_t_15, 11, __pyx_t_11);
  __Pyx_GIVEREF(__pyx_t_12);
  PyTuple_SET_ITEM(__pyx_t_15, 12, __pyx_t_12);
  __Pyx_GIVEREF(__pyx_t_13);
  PyTuple_SET_ITEM(__pyx_t_15, 13, __pyx_t_13);
  __Pyx_GIVEREF(__pyx_t_14);
  PyTuple_SET_ITEM(__pyx_t_15, 14, __pyx_t_14);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->right_node));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self->right_node));
  PyTuple_SET_ITEM(__pyx_t_15, 15, ((PyObject *)__pyx_v_self->right_node));
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  __pyx_t_6 = 0;
  __pyx_t_7 = 0;
  __pyx_t_8 = 0;
  __pyx_t_9 = 0;
  __pyx_t_10 = 0;
  __pyx_t_11 = 0;
  __pyx_t_12 = 0;
  __pyx_t_13 = 0;
  __pyx_t_14 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_15);
  __pyx_t_15 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.center_left_indices, self.center_left_values, self.center_right_indices, self.center_right_values, self.indices, self.is_leaf_node, self.leaf_size, self.left, self.left_node, self.max_right, self.min_left, self.n_center, self.n_elements, self.pivot, self.right, self.right_node)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_15 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_15)) __PYX_ERR(2, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __pyx_v__dict = __pyx_t_15;
  __pyx_t_15 = 0;

  /* "(tree fragment)":7
 *     state = (self.center_left_indices, self.center_left_values, self.center_right_indices, self.center_right_values, self.indices, self.is_leaf_node, self.leaf_size, self.left, self.left_node, self.max_right, self.min_left, self.n_center, self.n_elements, self.pivot, self.right, self.right_node)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_16 = (__pyx_v__dict != Py_None);
  __pyx_t_17 = (__pyx_t_16 != 0);
  if (__pyx_t_17) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_15 = PyTuple_New(1); if (unlikely(!__pyx_t_15)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    PyTuple_SET_ITEM(__pyx_t_15, 0, __pyx_v__dict);
    __pyx_t_14 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_15); if (unlikely(!__pyx_t_14)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_14));
    __pyx_t_14 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.left_node is not None or self.right_node is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.center_left_indices, self.center_left_values, self.center_right_indices, self.center_right_values, self.indices, self.is_leaf_node, self.leaf_size, self.left, self.left_node, self.max_right, self.min_left, self.n_center, self.n_elements, self.pivot, self.right, self.right_node)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.left_node is not None or self.right_node is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_Int32ClosedRightIntervalNode, (type(self), 0x7a104e5, None), state
 */
  /*else*/ {
    __pyx_t_16 = (((PyObject *)__pyx_v_self->left_node) != Py_None);
    __pyx_t_18 = (__pyx_t_16 != 0);
    if (!__pyx_t_18) {
    } else {
      __pyx_t_17 = __pyx_t_18;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_18 = (((PyObject *)__pyx_v_self->right_node) != Py_None);
    __pyx_t_16 = (__pyx_t_18 != 0);
    __pyx_t_17 = __pyx_t_16;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_17;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.left_node is not None or self.right_node is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Int32ClosedRightIntervalNode, (type(self), 0x7a104e5, None), state
 *     else:
 */
  __pyx_t_17 = (__pyx_v_use_setstate != 0);
  if (__pyx_t_17) {

    /* "(tree fragment)":13
 *         use_setstate = self.left_node is not None or self.right_node is not None
 *     if use_setstate:
 *         return __pyx_unpickle_Int32ClosedRightIntervalNode, (type(self), 0x7a104e5, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_Int32ClosedRightIntervalNode, (type(self), 0x7a104e5, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_14, __pyx_n_s_pyx_unpickle_Int32ClosedRightI); if (unlikely(!__pyx_t_14)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __pyx_t_15 = PyTuple_New(3); if (unlikely(!__pyx_t_15)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_15, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_127993061);
    __Pyx_GIVEREF(__pyx_int_127993061);
    PyTuple_SET_ITEM(__pyx_t_15, 1, __pyx_int_127993061);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_15, 2, Py_None);
    __pyx_t_13 = PyTuple_New(3); if (unlikely(!__pyx_t_13)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __Pyx_GIVEREF(__pyx_t_14);
    PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_t_14);
    __Pyx_GIVEREF(__pyx_t_15);
    PyTuple_SET_ITEM(__pyx_t_13, 1, __pyx_t_15);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_13, 2, __pyx_v_state);
    __pyx_t_14 = 0;
    __pyx_t_15 = 0;
    __pyx_r = __pyx_t_13;
    __pyx_t_13 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.left_node is not None or self.right_node is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Int32ClosedRightIntervalNode, (type(self), 0x7a104e5, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_Int32ClosedRightIntervalNode, (type(self), 0x7a104e5, None), state
 *     else:
 *         return __pyx_unpickle_Int32ClosedRightIntervalNode, (type(self), 0x7a104e5, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Int32ClosedRightIntervalNode__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_13, __pyx_n_s_pyx_unpickle_Int32ClosedRightI); if (unlikely(!__pyx_t_13)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __pyx_t_15 = PyTuple_New(3); if (unlikely(!__pyx_t_15)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_15, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_127993061);
    __Pyx_GIVEREF(__pyx_int_127993061);
    PyTuple_SET_ITEM(__pyx_t_15, 1, __pyx_int_127993061);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_15, 2, __pyx_v_state);
    __pyx_t_14 = PyTuple_New(2); if (unlikely(!__pyx_t_14)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_GIVEREF(__pyx_t_13);
    PyTuple_SET_ITEM(__pyx_t_14, 0, __pyx_t_13);
    __Pyx_GIVEREF(__pyx_t_15);
    PyTuple_SET_ITEM(__pyx_t_14, 1, __pyx_t_15);
    __pyx_t_13 = 0;
    __pyx_t_15 = 0;
    __pyx_r = __pyx_t_14;
    __pyx_t_14 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_AddTraceback("pandas._libs.interval.Int32ClosedRightIntervalNode.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Int32ClosedRightIntervalNode, (type(self), 0x7a104e5, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Int32ClosedRightIntervalNode__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_23__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_23__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_22__setstate_cython__(((struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_22__setstate_cython__(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_Int32ClosedRightIntervalNode, (type(self), 0x7a104e5, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Int32ClosedRightIntervalNode__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(2, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_6pandas_5_libs_8interval___pyx_unpickle_Int32ClosedRightIntervalNode__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Int32ClosedRightIntervalNode, (type(self), 0x7a104e5, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Int32ClosedRightIntervalNode__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Int32ClosedRightIntervalNode.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":1948
 *         readonly bint is_leaf_node
 * 
 *     def __init__(self,             # <<<<<<<<<<<<<<
 *                  ndarray[int32_t, ndim=1] left,
 *                  ndarray[int32_t, ndim=1] right,
 */

/* Python wrapper */
static int __pyx_pw_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyArrayObject *__pyx_v_left = 0;
  PyArrayObject *__pyx_v_right = 0;
  PyArrayObject *__pyx_v_indices = 0;
  __pyx_t_5numpy_int64_t __pyx_v_leaf_size;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_left,&__pyx_n_s_right,&__pyx_n_s_indices,&__pyx_n_s_leaf_size,0};
    PyObject* values[4] = {0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_left)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_right)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 4, 4, 1); __PYX_ERR(1, 1948, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_indices)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 4, 4, 2); __PYX_ERR(1, 1948, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_leaf_size)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 4, 4, 3); __PYX_ERR(1, 1948, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(1, 1948, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 4) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
    }
    __pyx_v_left = ((PyArrayObject *)values[0]);
    __pyx_v_right = ((PyArrayObject *)values[1]);
    __pyx_v_indices = ((PyArrayObject *)values[2]);
    __pyx_v_leaf_size = __Pyx_PyInt_As_npy_int64(values[3]); if (unlikely((__pyx_v_leaf_size == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(1, 1952, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 4, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 1948, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Int32ClosedBothIntervalNode.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_left), __pyx_ptype_5numpy_ndarray, 1, "left", 0))) __PYX_ERR(1, 1949, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_right), __pyx_ptype_5numpy_ndarray, 1, "right", 0))) __PYX_ERR(1, 1950, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_indices), __pyx_ptype_5numpy_ndarray, 1, "indices", 0))) __PYX_ERR(1, 1951, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode___init__(((struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode *)__pyx_v_self), __pyx_v_left, __pyx_v_right, __pyx_v_indices, __pyx_v_leaf_size);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode___init__(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode *__pyx_v_self, PyArrayObject *__pyx_v_left, PyArrayObject *__pyx_v_right, PyArrayObject *__pyx_v_indices, __pyx_t_5numpy_int64_t __pyx_v_leaf_size) {
  PyObject *__pyx_v_left_set = NULL;
  PyObject *__pyx_v_right_set = NULL;
  PyObject *__pyx_v_center_set = NULL;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_indices;
  __Pyx_Buffer __pyx_pybuffer_indices;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_left;
  __Pyx_Buffer __pyx_pybuffer_left;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_right;
  __Pyx_Buffer __pyx_pybuffer_right;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __pyx_t_5numpy_int32_t __pyx_t_6;
  __Pyx_memviewslice __pyx_t_7 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_8 = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  __Pyx_memviewslice __pyx_t_11 = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *(*__pyx_t_12)(PyObject *);
  size_t __pyx_t_13;
  __Pyx_RefNannySetupContext("__init__", 0);
  __pyx_pybuffer_left.pybuffer.buf = NULL;
  __pyx_pybuffer_left.refcount = 0;
  __pyx_pybuffernd_left.data = NULL;
  __pyx_pybuffernd_left.rcbuffer = &__pyx_pybuffer_left;
  __pyx_pybuffer_right.pybuffer.buf = NULL;
  __pyx_pybuffer_right.refcount = 0;
  __pyx_pybuffernd_right.data = NULL;
  __pyx_pybuffernd_right.rcbuffer = &__pyx_pybuffer_right;
  __pyx_pybuffer_indices.pybuffer.buf = NULL;
  __pyx_pybuffer_indices.refcount = 0;
  __pyx_pybuffernd_indices.data = NULL;
  __pyx_pybuffernd_indices.rcbuffer = &__pyx_pybuffer_indices;
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_left.rcbuffer->pybuffer, (PyObject*)__pyx_v_left, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int32_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 1948, __pyx_L1_error)
  }
  __pyx_pybuffernd_left.diminfo[0].strides = __pyx_pybuffernd_left.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_left.diminfo[0].shape = __pyx_pybuffernd_left.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_right.rcbuffer->pybuffer, (PyObject*)__pyx_v_right, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int32_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 1948, __pyx_L1_error)
  }
  __pyx_pybuffernd_right.diminfo[0].strides = __pyx_pybuffernd_right.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_right.diminfo[0].shape = __pyx_pybuffernd_right.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_indices.rcbuffer->pybuffer, (PyObject*)__pyx_v_indices, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 1948, __pyx_L1_error)
  }
  __pyx_pybuffernd_indices.diminfo[0].strides = __pyx_pybuffernd_indices.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_indices.diminfo[0].shape = __pyx_pybuffernd_indices.rcbuffer->pybuffer.shape[0];

  /* "pandas/_libs/intervaltree.pxi":1954
 *                  int64_t leaf_size):
 * 
 *         self.n_elements = len(left)             # <<<<<<<<<<<<<<
 *         self.leaf_size = leaf_size
 * 
 */
  __pyx_t_1 = PyObject_Length(((PyObject *)__pyx_v_left)); if (unlikely(__pyx_t_1 == ((Py_ssize_t)-1))) __PYX_ERR(1, 1954, __pyx_L1_error)
  __pyx_v_self->n_elements = __pyx_t_1;

  /* "pandas/_libs/intervaltree.pxi":1955
 * 
 *         self.n_elements = len(left)
 *         self.leaf_size = leaf_size             # <<<<<<<<<<<<<<
 * 
 *         # min_left and min_right are used to speed-up query by skipping
 */
  __pyx_v_self->leaf_size = __pyx_v_leaf_size;

  /* "pandas/_libs/intervaltree.pxi":1960
 *         # query on sub-nodes. If this node has size 0, query is cheap,
 *         # so these values don't matter.
 *         if left.size > 0:             # <<<<<<<<<<<<<<
 *             self.min_left = left.min()
 *             self.max_right = right.max()
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_left), __pyx_n_s_size); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1960, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyObject_RichCompare(__pyx_t_2, __pyx_int_0, Py_GT); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1960, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(1, 1960, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_4) {

    /* "pandas/_libs/intervaltree.pxi":1961
 *         # so these values don't matter.
 *         if left.size > 0:
 *             self.min_left = left.min()             # <<<<<<<<<<<<<<
 *             self.max_right = right.max()
 *         else:
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_left), __pyx_n_s_min); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1961, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1961, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_6 = __Pyx_PyInt_As_npy_int32(__pyx_t_3); if (unlikely((__pyx_t_6 == ((npy_int32)-1)) && PyErr_Occurred())) __PYX_ERR(1, 1961, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_self->min_left = __pyx_t_6;

    /* "pandas/_libs/intervaltree.pxi":1962
 *         if left.size > 0:
 *             self.min_left = left.min()
 *             self.max_right = right.max()             # <<<<<<<<<<<<<<
 *         else:
 *             self.min_left = 0
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_right), __pyx_n_s_max); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1962, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1962, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_6 = __Pyx_PyInt_As_npy_int32(__pyx_t_3); if (unlikely((__pyx_t_6 == ((npy_int32)-1)) && PyErr_Occurred())) __PYX_ERR(1, 1962, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_self->max_right = __pyx_t_6;

    /* "pandas/_libs/intervaltree.pxi":1960
 *         # query on sub-nodes. If this node has size 0, query is cheap,
 *         # so these values don't matter.
 *         if left.size > 0:             # <<<<<<<<<<<<<<
 *             self.min_left = left.min()
 *             self.max_right = right.max()
 */
    goto __pyx_L3;
  }

  /* "pandas/_libs/intervaltree.pxi":1964
 *             self.max_right = right.max()
 *         else:
 *             self.min_left = 0             # <<<<<<<<<<<<<<
 *             self.max_right = 0
 * 
 */
  /*else*/ {
    __pyx_v_self->min_left = 0;

    /* "pandas/_libs/intervaltree.pxi":1965
 *         else:
 *             self.min_left = 0
 *             self.max_right = 0             # <<<<<<<<<<<<<<
 * 
 *         if self.n_elements <= leaf_size:
 */
    __pyx_v_self->max_right = 0;
  }
  __pyx_L3:;

  /* "pandas/_libs/intervaltree.pxi":1967
 *             self.max_right = 0
 * 
 *         if self.n_elements <= leaf_size:             # <<<<<<<<<<<<<<
 *             # make this a terminal (leaf) node
 *             self.is_leaf_node = True
 */
  __pyx_t_4 = ((__pyx_v_self->n_elements <= __pyx_v_leaf_size) != 0);
  if (__pyx_t_4) {

    /* "pandas/_libs/intervaltree.pxi":1969
 *         if self.n_elements <= leaf_size:
 *             # make this a terminal (leaf) node
 *             self.is_leaf_node = True             # <<<<<<<<<<<<<<
 *             self.left = left
 *             self.right = right
 */
    __pyx_v_self->is_leaf_node = 1;

    /* "pandas/_libs/intervaltree.pxi":1970
 *             # make this a terminal (leaf) node
 *             self.is_leaf_node = True
 *             self.left = left             # <<<<<<<<<<<<<<
 *             self.right = right
 *             self.indices = indices
 */
    __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int32_t(((PyObject *)__pyx_v_left), PyBUF_WRITABLE); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(1, 1970, __pyx_L1_error)
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->left, 0);
    __pyx_v_self->left = __pyx_t_7;
    __pyx_t_7.memview = NULL;
    __pyx_t_7.data = NULL;

    /* "pandas/_libs/intervaltree.pxi":1971
 *             self.is_leaf_node = True
 *             self.left = left
 *             self.right = right             # <<<<<<<<<<<<<<
 *             self.indices = indices
 *             self.n_center = 0
 */
    __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int32_t(((PyObject *)__pyx_v_right), PyBUF_WRITABLE); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(1, 1971, __pyx_L1_error)
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->right, 0);
    __pyx_v_self->right = __pyx_t_7;
    __pyx_t_7.memview = NULL;
    __pyx_t_7.data = NULL;

    /* "pandas/_libs/intervaltree.pxi":1972
 *             self.left = left
 *             self.right = right
 *             self.indices = indices             # <<<<<<<<<<<<<<
 *             self.n_center = 0
 *         else:
 */
    __pyx_t_8 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(((PyObject *)__pyx_v_indices), PyBUF_WRITABLE); if (unlikely(!__pyx_t_8.memview)) __PYX_ERR(1, 1972, __pyx_L1_error)
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->indices, 0);
    __pyx_v_self->indices = __pyx_t_8;
    __pyx_t_8.memview = NULL;
    __pyx_t_8.data = NULL;

    /* "pandas/_libs/intervaltree.pxi":1973
 *             self.right = right
 *             self.indices = indices
 *             self.n_center = 0             # <<<<<<<<<<<<<<
 *         else:
 *             # calculate a pivot so we can create child nodes
 */
    __pyx_v_self->n_center = 0;

    /* "pandas/_libs/intervaltree.pxi":1967
 *             self.max_right = 0
 * 
 *         if self.n_elements <= leaf_size:             # <<<<<<<<<<<<<<
 *             # make this a terminal (leaf) node
 *             self.is_leaf_node = True
 */
    goto __pyx_L4;
  }

  /* "pandas/_libs/intervaltree.pxi":1976
 *         else:
 *             # calculate a pivot so we can create child nodes
 *             self.is_leaf_node = False             # <<<<<<<<<<<<<<
 *             self.pivot = np.median(left / 2 + right / 2)
 *             left_set, right_set, center_set = self.classify_intervals(
 */
  /*else*/ {
    __pyx_v_self->is_leaf_node = 0;

    /* "pandas/_libs/intervaltree.pxi":1977
 *             # calculate a pivot so we can create child nodes
 *             self.is_leaf_node = False
 *             self.pivot = np.median(left / 2 + right / 2)             # <<<<<<<<<<<<<<
 *             left_set, right_set, center_set = self.classify_intervals(
 *                 left, right)
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1977, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_median); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1977, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyNumber_Divide(((PyObject *)__pyx_v_left), __pyx_int_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1977, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_9 = __Pyx_PyNumber_Divide(((PyObject *)__pyx_v_right), __pyx_int_2); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 1977, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_10 = PyNumber_Add(__pyx_t_2, __pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 1977, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_9 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_3 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_9, __pyx_t_10) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_10);
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1977, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_6 = __Pyx_PyInt_As_npy_int32(__pyx_t_3); if (unlikely((__pyx_t_6 == ((npy_int32)-1)) && PyErr_Occurred())) __PYX_ERR(1, 1977, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_self->pivot = __pyx_t_6;

    /* "pandas/_libs/intervaltree.pxi":1979
 *             self.pivot = np.median(left / 2 + right / 2)
 *             left_set, right_set, center_set = self.classify_intervals(
 *                 left, right)             # <<<<<<<<<<<<<<
 * 
 *             self.left_node = self.new_child_node(left, right,
 */
    __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int32_t(((PyObject *)__pyx_v_left), PyBUF_WRITABLE); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(1, 1979, __pyx_L1_error)
    __pyx_t_11 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int32_t(((PyObject *)__pyx_v_right), PyBUF_WRITABLE); if (unlikely(!__pyx_t_11.memview)) __PYX_ERR(1, 1979, __pyx_L1_error)

    /* "pandas/_libs/intervaltree.pxi":1978
 *             self.is_leaf_node = False
 *             self.pivot = np.median(left / 2 + right / 2)
 *             left_set, right_set, center_set = self.classify_intervals(             # <<<<<<<<<<<<<<
 *                 left, right)
 * 
 */
    __pyx_t_3 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode *)__pyx_v_self->__pyx_vtab)->classify_intervals(__pyx_v_self, __pyx_t_7, __pyx_t_11); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1978, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __PYX_XDEC_MEMVIEW(&__pyx_t_7, 1);
    __pyx_t_7.memview = NULL;
    __pyx_t_7.data = NULL;
    __PYX_XDEC_MEMVIEW(&__pyx_t_11, 1);
    __pyx_t_11.memview = NULL;
    __pyx_t_11.data = NULL;
    if ((likely(PyTuple_CheckExact(__pyx_t_3))) || (PyList_CheckExact(__pyx_t_3))) {
      PyObject* sequence = __pyx_t_3;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 3)) {
        if (size > 3) __Pyx_RaiseTooManyValuesError(3);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(1, 1978, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_5 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_10 = PyTuple_GET_ITEM(sequence, 1); 
        __pyx_t_9 = PyTuple_GET_ITEM(sequence, 2); 
      } else {
        __pyx_t_5 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_10 = PyList_GET_ITEM(sequence, 1); 
        __pyx_t_9 = PyList_GET_ITEM(sequence, 2); 
      }
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_10);
      __Pyx_INCREF(__pyx_t_9);
      #else
      __pyx_t_5 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1978, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_10 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 1978, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_9 = PySequence_ITEM(sequence, 2); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 1978, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      #endif
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_2 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1978, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_12 = Py_TYPE(__pyx_t_2)->tp_iternext;
      index = 0; __pyx_t_5 = __pyx_t_12(__pyx_t_2); if (unlikely(!__pyx_t_5)) goto __pyx_L5_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_5);
      index = 1; __pyx_t_10 = __pyx_t_12(__pyx_t_2); if (unlikely(!__pyx_t_10)) goto __pyx_L5_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_10);
      index = 2; __pyx_t_9 = __pyx_t_12(__pyx_t_2); if (unlikely(!__pyx_t_9)) goto __pyx_L5_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_9);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_12(__pyx_t_2), 3) < 0) __PYX_ERR(1, 1978, __pyx_L1_error)
      __pyx_t_12 = NULL;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      goto __pyx_L6_unpacking_done;
      __pyx_L5_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_12 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(1, 1978, __pyx_L1_error)
      __pyx_L6_unpacking_done:;
    }
    __pyx_v_left_set = __pyx_t_5;
    __pyx_t_5 = 0;
    __pyx_v_right_set = __pyx_t_10;
    __pyx_t_10 = 0;
    __pyx_v_center_set = __pyx_t_9;
    __pyx_t_9 = 0;

    /* "pandas/_libs/intervaltree.pxi":1982
 * 
 *             self.left_node = self.new_child_node(left, right,
 *                                                  indices, left_set)             # <<<<<<<<<<<<<<
 *             self.right_node = self.new_child_node(left, right,
 *                                                   indices, right_set)
 */
    if (!(likely(((__pyx_v_left_set) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_left_set, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(1, 1982, __pyx_L1_error)

    /* "pandas/_libs/intervaltree.pxi":1981
 *                 left, right)
 * 
 *             self.left_node = self.new_child_node(left, right,             # <<<<<<<<<<<<<<
 *                                                  indices, left_set)
 *             self.right_node = self.new_child_node(left, right,
 */
    __pyx_t_3 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode *)__pyx_v_self->__pyx_vtab)->new_child_node(__pyx_v_self, ((PyArrayObject *)__pyx_v_left), ((PyArrayObject *)__pyx_v_right), ((PyArrayObject *)__pyx_v_indices), ((PyArrayObject *)__pyx_v_left_set)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1981, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode))))) __PYX_ERR(1, 1981, __pyx_L1_error)
    __Pyx_GIVEREF(__pyx_t_3);
    __Pyx_GOTREF(__pyx_v_self->left_node);
    __Pyx_DECREF(((PyObject *)__pyx_v_self->left_node));
    __pyx_v_self->left_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode *)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "pandas/_libs/intervaltree.pxi":1984
 *                                                  indices, left_set)
 *             self.right_node = self.new_child_node(left, right,
 *                                                   indices, right_set)             # <<<<<<<<<<<<<<
 * 
 *             self.center_left_values, self.center_left_indices = \
 */
    if (!(likely(((__pyx_v_right_set) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_right_set, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(1, 1984, __pyx_L1_error)

    /* "pandas/_libs/intervaltree.pxi":1983
 *             self.left_node = self.new_child_node(left, right,
 *                                                  indices, left_set)
 *             self.right_node = self.new_child_node(left, right,             # <<<<<<<<<<<<<<
 *                                                   indices, right_set)
 * 
 */
    __pyx_t_3 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode *)__pyx_v_self->__pyx_vtab)->new_child_node(__pyx_v_self, ((PyArrayObject *)__pyx_v_left), ((PyArrayObject *)__pyx_v_right), ((PyArrayObject *)__pyx_v_indices), ((PyArrayObject *)__pyx_v_right_set)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1983, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode))))) __PYX_ERR(1, 1983, __pyx_L1_error)
    __Pyx_GIVEREF(__pyx_t_3);
    __Pyx_GOTREF(__pyx_v_self->right_node);
    __Pyx_DECREF(((PyObject *)__pyx_v_self->right_node));
    __pyx_v_self->right_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode *)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "pandas/_libs/intervaltree.pxi":1987
 * 
 *             self.center_left_values, self.center_left_indices = \
 *                 sort_values_and_indices(left, indices, center_set)             # <<<<<<<<<<<<<<
 *             self.center_right_values, self.center_right_indices = \
 *                 sort_values_and_indices(right, indices, center_set)
 */
    __pyx_t_3 = __pyx_f_6pandas_5_libs_8interval_sort_values_and_indices(((PyObject *)__pyx_v_left), ((PyObject *)__pyx_v_indices), __pyx_v_center_set); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1987, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if ((likely(PyTuple_CheckExact(__pyx_t_3))) || (PyList_CheckExact(__pyx_t_3))) {
      PyObject* sequence = __pyx_t_3;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(1, 1986, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_9 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_10 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_9 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_10 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_9);
      __Pyx_INCREF(__pyx_t_10);
      #else
      __pyx_t_9 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 1986, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_10 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 1986, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      #endif
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_5 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1986, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_12 = Py_TYPE(__pyx_t_5)->tp_iternext;
      index = 0; __pyx_t_9 = __pyx_t_12(__pyx_t_5); if (unlikely(!__pyx_t_9)) goto __pyx_L7_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_9);
      index = 1; __pyx_t_10 = __pyx_t_12(__pyx_t_5); if (unlikely(!__pyx_t_10)) goto __pyx_L7_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_10);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_12(__pyx_t_5), 2) < 0) __PYX_ERR(1, 1986, __pyx_L1_error)
      __pyx_t_12 = NULL;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      goto __pyx_L8_unpacking_done;
      __pyx_L7_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_12 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(1, 1986, __pyx_L1_error)
      __pyx_L8_unpacking_done:;
    }

    /* "pandas/_libs/intervaltree.pxi":1986
 *                                                   indices, right_set)
 * 
 *             self.center_left_values, self.center_left_indices = \             # <<<<<<<<<<<<<<
 *                 sort_values_and_indices(left, indices, center_set)
 *             self.center_right_values, self.center_right_indices = \
 */
    __pyx_t_11 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int32_t(__pyx_t_9, PyBUF_WRITABLE); if (unlikely(!__pyx_t_11.memview)) __PYX_ERR(1, 1986, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_8 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_t_10, PyBUF_WRITABLE); if (unlikely(!__pyx_t_8.memview)) __PYX_ERR(1, 1986, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->center_left_values, 0);
    __pyx_v_self->center_left_values = __pyx_t_11;
    __pyx_t_11.memview = NULL;
    __pyx_t_11.data = NULL;
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->center_left_indices, 0);
    __pyx_v_self->center_left_indices = __pyx_t_8;
    __pyx_t_8.memview = NULL;
    __pyx_t_8.data = NULL;

    /* "pandas/_libs/intervaltree.pxi":1989
 *                 sort_values_and_indices(left, indices, center_set)
 *             self.center_right_values, self.center_right_indices = \
 *                 sort_values_and_indices(right, indices, center_set)             # <<<<<<<<<<<<<<
 *             self.n_center = len(self.center_left_indices)
 * 
 */
    __pyx_t_3 = __pyx_f_6pandas_5_libs_8interval_sort_values_and_indices(((PyObject *)__pyx_v_right), ((PyObject *)__pyx_v_indices), __pyx_v_center_set); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1989, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if ((likely(PyTuple_CheckExact(__pyx_t_3))) || (PyList_CheckExact(__pyx_t_3))) {
      PyObject* sequence = __pyx_t_3;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(1, 1988, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_10 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_9 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_10 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_9 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_10);
      __Pyx_INCREF(__pyx_t_9);
      #else
      __pyx_t_10 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 1988, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_9 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 1988, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      #endif
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_5 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1988, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_12 = Py_TYPE(__pyx_t_5)->tp_iternext;
      index = 0; __pyx_t_10 = __pyx_t_12(__pyx_t_5); if (unlikely(!__pyx_t_10)) goto __pyx_L9_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_10);
      index = 1; __pyx_t_9 = __pyx_t_12(__pyx_t_5); if (unlikely(!__pyx_t_9)) goto __pyx_L9_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_9);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_12(__pyx_t_5), 2) < 0) __PYX_ERR(1, 1988, __pyx_L1_error)
      __pyx_t_12 = NULL;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      goto __pyx_L10_unpacking_done;
      __pyx_L9_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_12 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(1, 1988, __pyx_L1_error)
      __pyx_L10_unpacking_done:;
    }

    /* "pandas/_libs/intervaltree.pxi":1988
 *             self.center_left_values, self.center_left_indices = \
 *                 sort_values_and_indices(left, indices, center_set)
 *             self.center_right_values, self.center_right_indices = \             # <<<<<<<<<<<<<<
 *                 sort_values_and_indices(right, indices, center_set)
 *             self.n_center = len(self.center_left_indices)
 */
    __pyx_t_11 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int32_t(__pyx_t_10, PyBUF_WRITABLE); if (unlikely(!__pyx_t_11.memview)) __PYX_ERR(1, 1988, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_t_8 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_t_9, PyBUF_WRITABLE); if (unlikely(!__pyx_t_8.memview)) __PYX_ERR(1, 1988, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->center_right_values, 0);
    __pyx_v_self->center_right_values = __pyx_t_11;
    __pyx_t_11.memview = NULL;
    __pyx_t_11.data = NULL;
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->center_right_indices, 0);
    __pyx_v_self->center_right_indices = __pyx_t_8;
    __pyx_t_8.memview = NULL;
    __pyx_t_8.data = NULL;

    /* "pandas/_libs/intervaltree.pxi":1990
 *             self.center_right_values, self.center_right_indices = \
 *                 sort_values_and_indices(right, indices, center_set)
 *             self.n_center = len(self.center_left_indices)             # <<<<<<<<<<<<<<
 * 
 *     @cython.wraparound(False)
 */
    if (unlikely(!__pyx_v_self->center_left_indices.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(1, 1990, __pyx_L1_error)}
    __pyx_t_13 = __Pyx_MemoryView_Len(__pyx_v_self->center_left_indices); 
    __pyx_v_self->n_center = __pyx_t_13;
  }
  __pyx_L4:;

  /* "pandas/_libs/intervaltree.pxi":1948
 *         readonly bint is_leaf_node
 * 
 *     def __init__(self,             # <<<<<<<<<<<<<<
 *                  ndarray[int32_t, ndim=1] left,
 *                  ndarray[int32_t, ndim=1] right,
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __PYX_XDEC_MEMVIEW(&__pyx_t_7, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_8, 1);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __PYX_XDEC_MEMVIEW(&__pyx_t_11, 1);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indices.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_left.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_right.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("pandas._libs.interval.Int32ClosedBothIntervalNode.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indices.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_left.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_right.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF(__pyx_v_left_set);
  __Pyx_XDECREF(__pyx_v_right_set);
  __Pyx_XDECREF(__pyx_v_center_set);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":1994
 *     @cython.wraparound(False)
 *     @cython.boundscheck(False)
 *     cdef classify_intervals(self, int32_t[:] left, int32_t[:] right):             # <<<<<<<<<<<<<<
 *         """Classify the given intervals based upon whether they fall to the
 *         left, right, or overlap with this node's pivot.
 */

static PyObject *__pyx_f_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_classify_intervals(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode *__pyx_v_self, __Pyx_memviewslice __pyx_v_left, __Pyx_memviewslice __pyx_v_right) {
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_left_ind = 0;
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_right_ind = 0;
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_overlapping_ind = 0;
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __pyx_t_5numpy_int64_t __pyx_t_2;
  __pyx_t_5numpy_int64_t __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  int __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  __Pyx_RefNannySetupContext("classify_intervals", 0);

  /* "pandas/_libs/intervaltree.pxi":2002
 *             Py_ssize_t i
 * 
 *         left_ind = Int64Vector()             # <<<<<<<<<<<<<<
 *         right_ind = Int64Vector()
 *         overlapping_ind = Int64Vector()
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2002, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_left_ind = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":2003
 * 
 *         left_ind = Int64Vector()
 *         right_ind = Int64Vector()             # <<<<<<<<<<<<<<
 *         overlapping_ind = Int64Vector()
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2003, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_right_ind = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":2004
 *         left_ind = Int64Vector()
 *         right_ind = Int64Vector()
 *         overlapping_ind = Int64Vector()             # <<<<<<<<<<<<<<
 * 
 *         for i in range(self.n_elements):
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2004, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_overlapping_ind = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":2006
 *         overlapping_ind = Int64Vector()
 * 
 *         for i in range(self.n_elements):             # <<<<<<<<<<<<<<
 *             if right[i] < self.pivot:
 *                 left_ind.append(i)
 */
  __pyx_t_2 = __pyx_v_self->n_elements;
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "pandas/_libs/intervaltree.pxi":2007
 * 
 *         for i in range(self.n_elements):
 *             if right[i] < self.pivot:             # <<<<<<<<<<<<<<
 *                 left_ind.append(i)
 *             elif self.pivot < left[i]:
 */
    __pyx_t_5 = __pyx_v_i;
    __pyx_t_6 = (((*((__pyx_t_5numpy_int32_t *) ( /* dim=0 */ (__pyx_v_right.data + __pyx_t_5 * __pyx_v_right.strides[0]) ))) < __pyx_v_self->pivot) != 0);
    if (__pyx_t_6) {

      /* "pandas/_libs/intervaltree.pxi":2008
 *         for i in range(self.n_elements):
 *             if right[i] < self.pivot:
 *                 left_ind.append(i)             # <<<<<<<<<<<<<<
 *             elif self.pivot < left[i]:
 *                 right_ind.append(i)
 */
      ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_left_ind->__pyx_vtab)->append(__pyx_v_left_ind, __pyx_v_i);

      /* "pandas/_libs/intervaltree.pxi":2007
 * 
 *         for i in range(self.n_elements):
 *             if right[i] < self.pivot:             # <<<<<<<<<<<<<<
 *                 left_ind.append(i)
 *             elif self.pivot < left[i]:
 */
      goto __pyx_L5;
    }

    /* "pandas/_libs/intervaltree.pxi":2009
 *             if right[i] < self.pivot:
 *                 left_ind.append(i)
 *             elif self.pivot < left[i]:             # <<<<<<<<<<<<<<
 *                 right_ind.append(i)
 *             else:
 */
    __pyx_t_7 = __pyx_v_i;
    __pyx_t_6 = ((__pyx_v_self->pivot < (*((__pyx_t_5numpy_int32_t *) ( /* dim=0 */ (__pyx_v_left.data + __pyx_t_7 * __pyx_v_left.strides[0]) )))) != 0);
    if (__pyx_t_6) {

      /* "pandas/_libs/intervaltree.pxi":2010
 *                 left_ind.append(i)
 *             elif self.pivot < left[i]:
 *                 right_ind.append(i)             # <<<<<<<<<<<<<<
 *             else:
 *                 overlapping_ind.append(i)
 */
      ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_right_ind->__pyx_vtab)->append(__pyx_v_right_ind, __pyx_v_i);

      /* "pandas/_libs/intervaltree.pxi":2009
 *             if right[i] < self.pivot:
 *                 left_ind.append(i)
 *             elif self.pivot < left[i]:             # <<<<<<<<<<<<<<
 *                 right_ind.append(i)
 *             else:
 */
      goto __pyx_L5;
    }

    /* "pandas/_libs/intervaltree.pxi":2012
 *                 right_ind.append(i)
 *             else:
 *                 overlapping_ind.append(i)             # <<<<<<<<<<<<<<
 * 
 *         return (left_ind.to_array(),
 */
    /*else*/ {
      ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_overlapping_ind->__pyx_vtab)->append(__pyx_v_overlapping_ind, __pyx_v_i);
    }
    __pyx_L5:;
  }

  /* "pandas/_libs/intervaltree.pxi":2014
 *                 overlapping_ind.append(i)
 * 
 *         return (left_ind.to_array(),             # <<<<<<<<<<<<<<
 *                 right_ind.to_array(),
 *                 overlapping_ind.to_array())
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_left_ind->__pyx_vtab)->to_array(__pyx_v_left_ind, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2014, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "pandas/_libs/intervaltree.pxi":2015
 * 
 *         return (left_ind.to_array(),
 *                 right_ind.to_array(),             # <<<<<<<<<<<<<<
 *                 overlapping_ind.to_array())
 * 
 */
  __pyx_t_8 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_right_ind->__pyx_vtab)->to_array(__pyx_v_right_ind, 0); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 2015, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);

  /* "pandas/_libs/intervaltree.pxi":2016
 *         return (left_ind.to_array(),
 *                 right_ind.to_array(),
 *                 overlapping_ind.to_array())             # <<<<<<<<<<<<<<
 * 
 *     cdef new_child_node(self,
 */
  __pyx_t_9 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_overlapping_ind->__pyx_vtab)->to_array(__pyx_v_overlapping_ind, 0); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 2016, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);

  /* "pandas/_libs/intervaltree.pxi":2014
 *                 overlapping_ind.append(i)
 * 
 *         return (left_ind.to_array(),             # <<<<<<<<<<<<<<
 *                 right_ind.to_array(),
 *                 overlapping_ind.to_array())
 */
  __pyx_t_10 = PyTuple_New(3); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 2014, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_t_8);
  __Pyx_GIVEREF(__pyx_t_9);
  PyTuple_SET_ITEM(__pyx_t_10, 2, __pyx_t_9);
  __pyx_t_1 = 0;
  __pyx_t_8 = 0;
  __pyx_t_9 = 0;
  __pyx_r = __pyx_t_10;
  __pyx_t_10 = 0;
  goto __pyx_L0;

  /* "pandas/_libs/intervaltree.pxi":1994
 *     @cython.wraparound(False)
 *     @cython.boundscheck(False)
 *     cdef classify_intervals(self, int32_t[:] left, int32_t[:] right):             # <<<<<<<<<<<<<<
 *         """Classify the given intervals based upon whether they fall to the
 *         left, right, or overlap with this node's pivot.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("pandas._libs.interval.Int32ClosedBothIntervalNode.classify_intervals", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_left_ind);
  __Pyx_XDECREF((PyObject *)__pyx_v_right_ind);
  __Pyx_XDECREF((PyObject *)__pyx_v_overlapping_ind);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":2018
 *                 overlapping_ind.to_array())
 * 
 *     cdef new_child_node(self,             # <<<<<<<<<<<<<<
 *                         ndarray[int32_t, ndim=1] left,
 *                         ndarray[int32_t, ndim=1] right,
 */

static PyObject *__pyx_f_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_new_child_node(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode *__pyx_v_self, PyArrayObject *__pyx_v_left, PyArrayObject *__pyx_v_right, PyArrayObject *__pyx_v_indices, PyArrayObject *__pyx_v_subset) {
  __Pyx_LocalBuf_ND __pyx_pybuffernd_indices;
  __Pyx_Buffer __pyx_pybuffer_indices;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_left;
  __Pyx_Buffer __pyx_pybuffer_left;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_right;
  __Pyx_Buffer __pyx_pybuffer_right;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_subset;
  __Pyx_Buffer __pyx_pybuffer_subset;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyArrayObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyArrayObject *__pyx_t_7 = NULL;
  PyArrayObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  __Pyx_RefNannySetupContext("new_child_node", 0);
  __Pyx_INCREF((PyObject *)__pyx_v_left);
  __Pyx_INCREF((PyObject *)__pyx_v_right);
  __Pyx_INCREF((PyObject *)__pyx_v_indices);
  __pyx_pybuffer_left.pybuffer.buf = NULL;
  __pyx_pybuffer_left.refcount = 0;
  __pyx_pybuffernd_left.data = NULL;
  __pyx_pybuffernd_left.rcbuffer = &__pyx_pybuffer_left;
  __pyx_pybuffer_right.pybuffer.buf = NULL;
  __pyx_pybuffer_right.refcount = 0;
  __pyx_pybuffernd_right.data = NULL;
  __pyx_pybuffernd_right.rcbuffer = &__pyx_pybuffer_right;
  __pyx_pybuffer_indices.pybuffer.buf = NULL;
  __pyx_pybuffer_indices.refcount = 0;
  __pyx_pybuffernd_indices.data = NULL;
  __pyx_pybuffernd_indices.rcbuffer = &__pyx_pybuffer_indices;
  __pyx_pybuffer_subset.pybuffer.buf = NULL;
  __pyx_pybuffer_subset.refcount = 0;
  __pyx_pybuffernd_subset.data = NULL;
  __pyx_pybuffernd_subset.rcbuffer = &__pyx_pybuffer_subset;
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_left.rcbuffer->pybuffer, (PyObject*)__pyx_v_left, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int32_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 2018, __pyx_L1_error)
  }
  __pyx_pybuffernd_left.diminfo[0].strides = __pyx_pybuffernd_left.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_left.diminfo[0].shape = __pyx_pybuffernd_left.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_right.rcbuffer->pybuffer, (PyObject*)__pyx_v_right, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int32_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 2018, __pyx_L1_error)
  }
  __pyx_pybuffernd_right.diminfo[0].strides = __pyx_pybuffernd_right.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_right.diminfo[0].shape = __pyx_pybuffernd_right.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_indices.rcbuffer->pybuffer, (PyObject*)__pyx_v_indices, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 2018, __pyx_L1_error)
  }
  __pyx_pybuffernd_indices.diminfo[0].strides = __pyx_pybuffernd_indices.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_indices.diminfo[0].shape = __pyx_pybuffernd_indices.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_subset.rcbuffer->pybuffer, (PyObject*)__pyx_v_subset, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 2018, __pyx_L1_error)
  }
  __pyx_pybuffernd_subset.diminfo[0].strides = __pyx_pybuffernd_subset.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_subset.diminfo[0].shape = __pyx_pybuffernd_subset.rcbuffer->pybuffer.shape[0];

  /* "pandas/_libs/intervaltree.pxi":2025
 *         """Create a new child node.
 *         """
 *         left = take(left, subset)             # <<<<<<<<<<<<<<
 *         right = take(right, subset)
 *         indices = take(indices, subset)
 */
  __pyx_t_1 = __pyx_f_6pandas_5_libs_8interval_take(((PyArrayObject *)__pyx_v_left), ((PyArrayObject *)__pyx_v_subset)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2025, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(1, 2025, __pyx_L1_error)
  __pyx_t_2 = ((PyArrayObject *)__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_left.rcbuffer->pybuffer);
    __pyx_t_3 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_left.rcbuffer->pybuffer, (PyObject*)__pyx_t_2, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int32_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack);
    if (unlikely(__pyx_t_3 < 0)) {
      PyErr_Fetch(&__pyx_t_4, &__pyx_t_5, &__pyx_t_6);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_left.rcbuffer->pybuffer, (PyObject*)__pyx_v_left, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int32_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_4); Py_XDECREF(__pyx_t_5); Py_XDECREF(__pyx_t_6);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_4, __pyx_t_5, __pyx_t_6);
      }
      __pyx_t_4 = __pyx_t_5 = __pyx_t_6 = 0;
    }
    __pyx_pybuffernd_left.diminfo[0].strides = __pyx_pybuffernd_left.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_left.diminfo[0].shape = __pyx_pybuffernd_left.rcbuffer->pybuffer.shape[0];
    if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(1, 2025, __pyx_L1_error)
  }
  __pyx_t_2 = 0;
  __Pyx_DECREF_SET(__pyx_v_left, ((PyArrayObject *)__pyx_t_1));
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":2026
 *         """
 *         left = take(left, subset)
 *         right = take(right, subset)             # <<<<<<<<<<<<<<
 *         indices = take(indices, subset)
 *         return Int32ClosedBothIntervalNode(
 */
  __pyx_t_1 = __pyx_f_6pandas_5_libs_8interval_take(((PyArrayObject *)__pyx_v_right), ((PyArrayObject *)__pyx_v_subset)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2026, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(1, 2026, __pyx_L1_error)
  __pyx_t_7 = ((PyArrayObject *)__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_right.rcbuffer->pybuffer);
    __pyx_t_3 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_right.rcbuffer->pybuffer, (PyObject*)__pyx_t_7, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int32_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack);
    if (unlikely(__pyx_t_3 < 0)) {
      PyErr_Fetch(&__pyx_t_6, &__pyx_t_5, &__pyx_t_4);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_right.rcbuffer->pybuffer, (PyObject*)__pyx_v_right, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int32_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_6); Py_XDECREF(__pyx_t_5); Py_XDECREF(__pyx_t_4);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_6, __pyx_t_5, __pyx_t_4);
      }
      __pyx_t_6 = __pyx_t_5 = __pyx_t_4 = 0;
    }
    __pyx_pybuffernd_right.diminfo[0].strides = __pyx_pybuffernd_right.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_right.diminfo[0].shape = __pyx_pybuffernd_right.rcbuffer->pybuffer.shape[0];
    if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(1, 2026, __pyx_L1_error)
  }
  __pyx_t_7 = 0;
  __Pyx_DECREF_SET(__pyx_v_right, ((PyArrayObject *)__pyx_t_1));
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":2027
 *         left = take(left, subset)
 *         right = take(right, subset)
 *         indices = take(indices, subset)             # <<<<<<<<<<<<<<
 *         return Int32ClosedBothIntervalNode(
 *             left, right, indices, self.leaf_size)
 */
  __pyx_t_1 = __pyx_f_6pandas_5_libs_8interval_take(((PyArrayObject *)__pyx_v_indices), ((PyArrayObject *)__pyx_v_subset)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2027, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(1, 2027, __pyx_L1_error)
  __pyx_t_8 = ((PyArrayObject *)__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indices.rcbuffer->pybuffer);
    __pyx_t_3 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_indices.rcbuffer->pybuffer, (PyObject*)__pyx_t_8, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack);
    if (unlikely(__pyx_t_3 < 0)) {
      PyErr_Fetch(&__pyx_t_4, &__pyx_t_5, &__pyx_t_6);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_indices.rcbuffer->pybuffer, (PyObject*)__pyx_v_indices, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_4); Py_XDECREF(__pyx_t_5); Py_XDECREF(__pyx_t_6);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_4, __pyx_t_5, __pyx_t_6);
      }
      __pyx_t_4 = __pyx_t_5 = __pyx_t_6 = 0;
    }
    __pyx_pybuffernd_indices.diminfo[0].strides = __pyx_pybuffernd_indices.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_indices.diminfo[0].shape = __pyx_pybuffernd_indices.rcbuffer->pybuffer.shape[0];
    if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(1, 2027, __pyx_L1_error)
  }
  __pyx_t_8 = 0;
  __Pyx_DECREF_SET(__pyx_v_indices, ((PyArrayObject *)__pyx_t_1));
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":2028
 *         right = take(right, subset)
 *         indices = take(indices, subset)
 *         return Int32ClosedBothIntervalNode(             # <<<<<<<<<<<<<<
 *             left, right, indices, self.leaf_size)
 * 
 */
  __Pyx_XDECREF(__pyx_r);

  /* "pandas/_libs/intervaltree.pxi":2029
 *         indices = take(indices, subset)
 *         return Int32ClosedBothIntervalNode(
 *             left, right, indices, self.leaf_size)             # <<<<<<<<<<<<<<
 * 
 *     @cython.wraparound(False)
 */
  __pyx_t_1 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->leaf_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2029, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "pandas/_libs/intervaltree.pxi":2028
 *         right = take(right, subset)
 *         indices = take(indices, subset)
 *         return Int32ClosedBothIntervalNode(             # <<<<<<<<<<<<<<
 *             left, right, indices, self.leaf_size)
 * 
 */
  __pyx_t_9 = PyTuple_New(4); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 2028, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_INCREF(((PyObject *)__pyx_v_left));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_left));
  PyTuple_SET_ITEM(__pyx_t_9, 0, ((PyObject *)__pyx_v_left));
  __Pyx_INCREF(((PyObject *)__pyx_v_right));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_right));
  PyTuple_SET_ITEM(__pyx_t_9, 1, ((PyObject *)__pyx_v_right));
  __Pyx_INCREF(((PyObject *)__pyx_v_indices));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_indices));
  PyTuple_SET_ITEM(__pyx_t_9, 2, ((PyObject *)__pyx_v_indices));
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_9, 3, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode), __pyx_t_9, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2028, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pandas/_libs/intervaltree.pxi":2018
 *                 overlapping_ind.to_array())
 * 
 *     cdef new_child_node(self,             # <<<<<<<<<<<<<<
 *                         ndarray[int32_t, ndim=1] left,
 *                         ndarray[int32_t, ndim=1] right,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_9);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indices.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_left.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_right.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_subset.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("pandas._libs.interval.Int32ClosedBothIntervalNode.new_child_node", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indices.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_left.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_right.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_subset.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF((PyObject *)__pyx_v_left);
  __Pyx_XDECREF((PyObject *)__pyx_v_right);
  __Pyx_XDECREF((PyObject *)__pyx_v_indices);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":2034
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_2query[] = "Recursively query this node and its sub-nodes for intervals that\n        overlap with the query point.\n        ";
static PyMethodDef __pyx_mdef_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_3query = {"query", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_3query, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_2query};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_args = 0;
  PyObject *__pyx_v_kwargs = 0;
  CYTHON_UNUSED PyObject *__pyx_v_defaults = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fused_cpdef (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_args,&__pyx_n_s_kwargs,&__pyx_n_s_defaults,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_args)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_kwargs)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fused_cpdef", 1, 3, 3, 1); __PYX_ERR(1, 2034, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_defaults)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fused_cpdef", 1, 3, 3, 2); __PYX_ERR(1, 2034, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fused_cpdef") < 0)) __PYX_ERR(1, 2034, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_args = values[0];
    __pyx_v_kwargs = values[1];
    __pyx_v_defaults = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fused_cpdef", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 2034, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Int32ClosedBothIntervalNode.__pyx_fused_cpdef", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_2query(((struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode *)__pyx_v_signatures), __pyx_v_args, __pyx_v_kwargs, __pyx_v_defaults);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_2query(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode *__pyx_v_signatures, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs, CYTHON_UNUSED PyObject *__pyx_v_defaults) {
  PyObject *__pyx_v_dest_sig = NULL;
  Py_ssize_t __pyx_v_i;
  CYTHON_UNUSED PyTypeObject *__pyx_v_ndarray = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_v_itemsize;
  PyObject *__pyx_v_arg = NULL;
  PyObject *__pyx_v_candidates = NULL;
  PyObject *__pyx_v_sig = NULL;
  int __pyx_v_match_found;
  PyObject *__pyx_v_src_sig = NULL;
  PyObject *__pyx_v_dst_type = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  Py_ssize_t __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  Py_ssize_t __pyx_t_13;
  Py_ssize_t __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  int __pyx_t_16;
  __Pyx_RefNannySetupContext("query", 0);
  __Pyx_INCREF(__pyx_v_kwargs);
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2034, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  PyList_SET_ITEM(__pyx_t_1, 0, Py_None);
  __pyx_v_dest_sig = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_3 = (__pyx_v_kwargs != Py_None);
  __pyx_t_4 = (__pyx_t_3 != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_kwargs); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(1, 2034, __pyx_L1_error)
  __pyx_t_3 = ((!__pyx_t_4) != 0);
  __pyx_t_2 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {
    __Pyx_INCREF(Py_None);
    __Pyx_DECREF_SET(__pyx_v_kwargs, Py_None);
  }
  __pyx_t_1 = ((PyObject *)__Pyx_ImportNumPyArrayTypeIfAvailable()); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2034, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_ndarray = ((PyTypeObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_v_itemsize = -1L;
  if (unlikely(__pyx_v_args == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 2034, __pyx_L1_error)
  }
  __pyx_t_5 = PyTuple_GET_SIZE(((PyObject*)__pyx_v_args)); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(1, 2034, __pyx_L1_error)
  __pyx_t_2 = ((2 < __pyx_t_5) != 0);
  if (__pyx_t_2) {
    if (unlikely(__pyx_v_args == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 2034, __pyx_L1_error)
    }
    __pyx_t_1 = PyTuple_GET_ITEM(((PyObject*)__pyx_v_args), 2);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v_arg = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L6;
  }
  __pyx_t_3 = (__pyx_v_kwargs != Py_None);
  __pyx_t_4 = (__pyx_t_3 != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L7_bool_binop_done;
  }
  if (unlikely(__pyx_v_kwargs == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(1, 2034, __pyx_L1_error)
  }
  __pyx_t_4 = (__Pyx_PyDict_ContainsTF(__pyx_n_s_point, ((PyObject*)__pyx_v_kwargs), Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(1, 2034, __pyx_L1_error)
  __pyx_t_3 = (__pyx_t_4 != 0);
  __pyx_t_2 = __pyx_t_3;
  __pyx_L7_bool_binop_done:;
  if (__pyx_t_2) {
    if (unlikely(__pyx_v_kwargs == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 2034, __pyx_L1_error)
    }
    __pyx_t_1 = __Pyx_PyDict_GetItem(((PyObject*)__pyx_v_kwargs), __pyx_n_s_point); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2034, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_arg = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L6;
  }
  /*else*/ {
    if (unlikely(__pyx_v_args == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
      __PYX_ERR(1, 2034, __pyx_L1_error)
    }
    __pyx_t_5 = PyTuple_GET_SIZE(((PyObject*)__pyx_v_args)); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(1, 2034, __pyx_L1_error)
    __pyx_t_1 = PyInt_FromSsize_t(__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2034, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_6 = PyTuple_New(3); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 2034, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_INCREF(__pyx_int_3);
    __Pyx_GIVEREF(__pyx_int_3);
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_int_3);
    __Pyx_INCREF(__pyx_n_s_s);
    __Pyx_GIVEREF(__pyx_n_s_s);
    PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_n_s_s);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Expected_at_least_d_argument_s_g, __pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2034, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_1); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 2034, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(1, 2034, __pyx_L1_error)
  }
  __pyx_L6:;
  while (1) {
    __pyx_t_2 = PyFloat_Check(__pyx_v_arg); 
    __pyx_t_3 = (__pyx_t_2 != 0);
    if (__pyx_t_3) {
      if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, __pyx_n_s_float64_t, long, 1, __Pyx_PyInt_From_long, 1, 0, 0) < 0)) __PYX_ERR(1, 2034, __pyx_L1_error)
      goto __pyx_L10_break;
    }
    __pyx_t_2 = PyInt_Check(__pyx_v_arg); 
    __pyx_t_4 = (__pyx_t_2 != 0);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_3 = __pyx_t_4;
      goto __pyx_L13_bool_binop_done;
    }
    __pyx_t_4 = PyLong_Check(__pyx_v_arg); 
    __pyx_t_2 = (__pyx_t_4 != 0);
    __pyx_t_3 = __pyx_t_2;
    __pyx_L13_bool_binop_done:;
    __pyx_t_2 = (__pyx_t_3 != 0);
    if (__pyx_t_2) {
      if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, __pyx_n_s_int64_t, long, 1, __Pyx_PyInt_From_long, 1, 0, 0) < 0)) __PYX_ERR(1, 2034, __pyx_L1_error)
      goto __pyx_L10_break;
    }
    if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, Py_None, long, 1, __Pyx_PyInt_From_long, 1, 0, 0) < 0)) __PYX_ERR(1, 2034, __pyx_L1_error)
    goto __pyx_L10_break;
  }
  __pyx_L10_break:;
  __pyx_t_6 = PyList_New(0); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 2034, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_v_candidates = ((PyObject*)__pyx_t_6);
  __pyx_t_6 = 0;
  __pyx_t_5 = 0;
  if (unlikely(((PyObject *)__pyx_v_signatures) == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(1, 2034, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_dict_iterator(((PyObject*)__pyx_v_signatures), 1, ((PyObject *)NULL), (&__pyx_t_7), (&__pyx_t_8)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2034, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __pyx_t_6 = __pyx_t_1;
  __pyx_t_1 = 0;
  while (1) {
    __pyx_t_9 = __Pyx_dict_iter_next(__pyx_t_6, __pyx_t_7, &__pyx_t_5, &__pyx_t_1, NULL, NULL, __pyx_t_8);
    if (unlikely(__pyx_t_9 == 0)) break;
    if (unlikely(__pyx_t_9 == -1)) __PYX_ERR(1, 2034, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_sig, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_v_match_found = 0;
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_v_sig, __pyx_n_s_strip); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 2034, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __pyx_t_12 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_11))) {
      __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_11);
      if (likely(__pyx_t_12)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
        __Pyx_INCREF(__pyx_t_12);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_11, function);
      }
    }
    __pyx_t_10 = (__pyx_t_12) ? __Pyx_PyObject_Call2Args(__pyx_t_11, __pyx_t_12, __pyx_kp_s__13) : __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_kp_s__13);
    __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
    if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 2034, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_split); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 2034, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_t_10 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_11))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_11);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_11, function);
      }
    }
    __pyx_t_1 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_11, __pyx_t_10, __pyx_kp_s__14) : __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_kp_s__14);
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2034, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_XDECREF_SET(__pyx_v_src_sig, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_13 = PyList_GET_SIZE(__pyx_v_dest_sig); if (unlikely(__pyx_t_13 == ((Py_ssize_t)-1))) __PYX_ERR(1, 2034, __pyx_L1_error)
    __pyx_t_14 = __pyx_t_13;
    for (__pyx_t_15 = 0; __pyx_t_15 < __pyx_t_14; __pyx_t_15+=1) {
      __pyx_v_i = __pyx_t_15;
      __pyx_t_1 = PyList_GET_ITEM(__pyx_v_dest_sig, __pyx_v_i);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_XDECREF_SET(__pyx_v_dst_type, __pyx_t_1);
      __pyx_t_1 = 0;
      __pyx_t_2 = (__pyx_v_dst_type != Py_None);
      __pyx_t_3 = (__pyx_t_2 != 0);
      if (__pyx_t_3) {
        __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_src_sig, __pyx_v_i, Py_ssize_t, 1, PyInt_FromSsize_t, 0, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2034, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_11 = PyObject_RichCompare(__pyx_t_1, __pyx_v_dst_type, Py_EQ); __Pyx_XGOTREF(__pyx_t_11); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 2034, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_11); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(1, 2034, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        if (__pyx_t_3) {
          __pyx_v_match_found = 1;
          goto __pyx_L20;
        }
        /*else*/ {
          __pyx_v_match_found = 0;
          goto __pyx_L18_break;
        }
        __pyx_L20:;
      }
    }
    __pyx_L18_break:;
    __pyx_t_3 = (__pyx_v_match_found != 0);
    if (__pyx_t_3) {
      __pyx_t_16 = __Pyx_PyList_Append(__pyx_v_candidates, __pyx_v_sig); if (unlikely(__pyx_t_16 == ((int)-1))) __PYX_ERR(1, 2034, __pyx_L1_error)
    }
  }
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_3 = (PyList_GET_SIZE(__pyx_v_candidates) != 0);
  __pyx_t_2 = ((!__pyx_t_3) != 0);
  if (__pyx_t_2) {
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__15, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 2034, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(1, 2034, __pyx_L1_error)
  }
  __pyx_t_7 = PyList_GET_SIZE(__pyx_v_candidates); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1))) __PYX_ERR(1, 2034, __pyx_L1_error)
  __pyx_t_2 = ((__pyx_t_7 > 1) != 0);
  if (__pyx_t_2) {
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__16, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 2034, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(1, 2034, __pyx_L1_error)
  }
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    if (unlikely(((PyObject *)__pyx_v_signatures) == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 2034, __pyx_L1_error)
    }
    __pyx_t_6 = __Pyx_PyDict_GetItem(((PyObject*)__pyx_v_signatures), PyList_GET_ITEM(__pyx_v_candidates, 0)); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 2034, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_r = __pyx_t_6;
    __pyx_t_6 = 0;
    goto __pyx_L0;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("pandas._libs.interval.Int32ClosedBothIntervalNode.__pyx_fused_cpdef", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_dest_sig);
  __Pyx_XDECREF(__pyx_v_ndarray);
  __Pyx_XDECREF(__pyx_v_arg);
  __Pyx_XDECREF(__pyx_v_candidates);
  __Pyx_XDECREF(__pyx_v_sig);
  __Pyx_XDECREF(__pyx_v_src_sig);
  __Pyx_XDECREF(__pyx_v_dst_type);
  __Pyx_XDECREF(__pyx_v_kwargs);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_9__pyx_fuse_0query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_fuse_0__pyx_f_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float64_t __pyx_v_point, int __pyx_skip_dispatch) {
  __Pyx_memviewslice __pyx_v_indices = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_values = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __pyx_t_5numpy_int64_t __pyx_t_9;
  __pyx_t_5numpy_int64_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  int __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  __Pyx_memviewslice __pyx_t_16 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_17 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  Py_ssize_t __pyx_t_21;
  __Pyx_RefNannySetupContext("__pyx_fuse_0query", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pyx_fuse_0query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2034, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_9__pyx_fuse_0query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = PyFloat_FromDouble(__pyx_v_point); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2034, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2034, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2034, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 2034, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_5) {
            __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2034, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2034, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_3query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_4 = PyFloat_FromDouble(__pyx_v_point); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 2034, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_7 = __pyx_t_1; __pyx_t_3 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2034, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2034, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 2034, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (__pyx_t_3) {
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_t_4);
          __pyx_t_4 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2034, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pandas/_libs/intervaltree.pxi":2043
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
  __pyx_t_8 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_8) {

    /* "pandas/_libs/intervaltree.pxi":2047
 *             # continue the binary tree structure. Instead, we use linear
 *             # search.
 *             for i in range(self.n_elements):             # <<<<<<<<<<<<<<
 *                 if self.left[i] <= point <= self.right[i]:
 *                     result.append(self.indices[i])
 */
    __pyx_t_9 = __pyx_v_self->n_elements;
    __pyx_t_10 = __pyx_t_9;
    for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
      __pyx_v_i = __pyx_t_11;

      /* "pandas/_libs/intervaltree.pxi":2048
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point <= self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      __pyx_t_12 = __pyx_v_i;
      __pyx_t_8 = ((*((__pyx_t_5numpy_int32_t *) ( /* dim=0 */ (__pyx_v_self->left.data + __pyx_t_12 * __pyx_v_self->left.strides[0]) ))) <= __pyx_v_point);
      if (__pyx_t_8) {
        __pyx_t_13 = __pyx_v_i;
        __pyx_t_8 = (__pyx_v_point <= (*((__pyx_t_5numpy_int32_t *) ( /* dim=0 */ (__pyx_v_self->right.data + __pyx_t_13 * __pyx_v_self->right.strides[0]) ))));
      }
      __pyx_t_14 = (__pyx_t_8 != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":2049
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point <= self.right[i]:
 *                     result.append(self.indices[i])             # <<<<<<<<<<<<<<
 *         else:
 *             # There are child nodes. Based on comparing our query to the pivot,
 */
        __pyx_t_15 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->indices.data + __pyx_t_15 * __pyx_v_self->indices.strides[0]) ))));

        /* "pandas/_libs/intervaltree.pxi":2048
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point <= self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      }
    }

    /* "pandas/_libs/intervaltree.pxi":2043
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
    goto __pyx_L3;
  }

  /* "pandas/_libs/intervaltree.pxi":2053
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
  /*else*/ {
    __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":2054
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:
 *                 values = self.center_left_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 */
      __pyx_t_16 = __pyx_v_self->center_left_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":2055
 *             if point < self.pivot:
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:
 */
      __pyx_t_17 = __pyx_v_self->center_left_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":2056
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):             # <<<<<<<<<<<<<<
 *                     if not values[i] <= point:
 *                         break
 */
      __pyx_t_9 = __pyx_v_self->n_center;
      __pyx_t_10 = __pyx_t_9;
      for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":2057
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_18 = __pyx_v_i;
        __pyx_t_14 = ((!(((*((__pyx_t_5numpy_int32_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_18 * __pyx_v_values.strides[0]) ))) <= __pyx_v_point) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":2058
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:
 */
          goto __pyx_L9_break;

          /* "pandas/_libs/intervaltree.pxi":2057
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":2059
 *                     if not values[i] <= point:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 */
        __pyx_t_19 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_19 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L9_break:;

      /* "pandas/_libs/intervaltree.pxi":2060
 *                         break
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      __pyx_t_14 = ((__pyx_v_point <= __pyx_v_self->left_node->max_right) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":2061
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)             # <<<<<<<<<<<<<<
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode *)__pyx_v_self->left_node->__pyx_vtab)->__pyx_fuse_0query(__pyx_v_self->left_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2061, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":2060
 *                         break
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":2053
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":2062
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
    __pyx_t_14 = ((__pyx_v_point > __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":2063
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 *                 values = self.center_right_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 */
      __pyx_t_16 = __pyx_v_self->center_right_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":2064
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:
 */
      __pyx_t_17 = __pyx_v_self->center_right_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":2065
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):             # <<<<<<<<<<<<<<
 *                     if not point <= values[i]:
 *                         break
 */
      for (__pyx_t_11 = (__pyx_v_self->n_center - 1); __pyx_t_11 > -1L; __pyx_t_11-=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":2066
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_20 = __pyx_v_i;
        __pyx_t_14 = ((!((__pyx_v_point <= (*((__pyx_t_5numpy_int32_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_20 * __pyx_v_values.strides[0]) )))) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":2067
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:
 */
          goto __pyx_L13_break;

          /* "pandas/_libs/intervaltree.pxi":2066
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":2068
 *                     if not point <= values[i]:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if self.right_node.min_left <= point:
 *                     self.right_node.query(result, point)
 */
        __pyx_t_21 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_21 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L13_break:;

      /* "pandas/_libs/intervaltree.pxi":2069
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      __pyx_t_14 = ((__pyx_v_self->right_node->min_left <= __pyx_v_point) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":2070
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:
 *                     self.right_node.query(result, point)             # <<<<<<<<<<<<<<
 *             else:
 *                 result.extend(self.center_left_indices)
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode *)__pyx_v_self->right_node->__pyx_vtab)->__pyx_fuse_0query(__pyx_v_self->right_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2070, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":2069
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":2062
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":2072
 *                     self.right_node.query(result, point)
 *             else:
 *                 result.extend(self.center_left_indices)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
    /*else*/ {
      __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->extend(__pyx_v_result, __pyx_v_self->center_left_indices); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2072, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_L7:;
  }
  __pyx_L3:;

  /* "pandas/_libs/intervaltree.pxi":2034
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __PYX_XDEC_MEMVIEW(&__pyx_t_16, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_17, 1);
  __Pyx_AddTraceback("pandas._libs.interval.Int32ClosedBothIntervalNode.query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_indices, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_values, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_9__pyx_fuse_0query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_0__pyx_mdef_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_9__pyx_fuse_0query = {"__pyx_fuse_0query", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_9__pyx_fuse_0query, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_2query};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_9__pyx_fuse_0query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result = 0;
  __pyx_t_5numpy_float64_t __pyx_v_point;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fuse_0query (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_result,&__pyx_n_s_point,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_result)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_point)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_0query", 1, 2, 2, 1); __PYX_ERR(1, 2034, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_0query") < 0)) __PYX_ERR(1, 2034, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_result = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)values[0]);
    __pyx_v_point = __pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_point == ((npy_float64)-1)) && PyErr_Occurred())) __PYX_ERR(1, 2034, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_0query", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 2034, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Int32ClosedBothIntervalNode.__pyx_fuse_0query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_result), __pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector, 1, "result", 0))) __PYX_ERR(1, 2034, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_8__pyx_fuse_0query(((struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode *)__pyx_v_self), __pyx_v_result, __pyx_v_point);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_8__pyx_fuse_0query(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float64_t __pyx_v_point) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__pyx_fuse_0query", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_fuse_0__pyx_f_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_query(__pyx_v_self, __pyx_v_result, __pyx_v_point, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2034, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Int32ClosedBothIntervalNode.__pyx_fuse_0query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_11__pyx_fuse_1query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_fuse_1__pyx_f_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float32_t __pyx_v_point, int __pyx_skip_dispatch) {
  __Pyx_memviewslice __pyx_v_indices = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_values = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __pyx_t_5numpy_int64_t __pyx_t_9;
  __pyx_t_5numpy_int64_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  int __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  __Pyx_memviewslice __pyx_t_16 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_17 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  Py_ssize_t __pyx_t_21;
  __Pyx_RefNannySetupContext("__pyx_fuse_1query", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pyx_fuse_1query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2034, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_11__pyx_fuse_1query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = PyFloat_FromDouble(__pyx_v_point); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2034, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2034, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2034, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 2034, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_5) {
            __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2034, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2034, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_3query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_4 = PyFloat_FromDouble(__pyx_v_point); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 2034, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_7 = __pyx_t_1; __pyx_t_3 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2034, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2034, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 2034, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (__pyx_t_3) {
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_t_4);
          __pyx_t_4 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2034, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pandas/_libs/intervaltree.pxi":2043
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
  __pyx_t_8 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_8) {

    /* "pandas/_libs/intervaltree.pxi":2047
 *             # continue the binary tree structure. Instead, we use linear
 *             # search.
 *             for i in range(self.n_elements):             # <<<<<<<<<<<<<<
 *                 if self.left[i] <= point <= self.right[i]:
 *                     result.append(self.indices[i])
 */
    __pyx_t_9 = __pyx_v_self->n_elements;
    __pyx_t_10 = __pyx_t_9;
    for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
      __pyx_v_i = __pyx_t_11;

      /* "pandas/_libs/intervaltree.pxi":2048
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point <= self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      __pyx_t_12 = __pyx_v_i;
      __pyx_t_8 = ((*((__pyx_t_5numpy_int32_t *) ( /* dim=0 */ (__pyx_v_self->left.data + __pyx_t_12 * __pyx_v_self->left.strides[0]) ))) <= __pyx_v_point);
      if (__pyx_t_8) {
        __pyx_t_13 = __pyx_v_i;
        __pyx_t_8 = (__pyx_v_point <= (*((__pyx_t_5numpy_int32_t *) ( /* dim=0 */ (__pyx_v_self->right.data + __pyx_t_13 * __pyx_v_self->right.strides[0]) ))));
      }
      __pyx_t_14 = (__pyx_t_8 != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":2049
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point <= self.right[i]:
 *                     result.append(self.indices[i])             # <<<<<<<<<<<<<<
 *         else:
 *             # There are child nodes. Based on comparing our query to the pivot,
 */
        __pyx_t_15 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->indices.data + __pyx_t_15 * __pyx_v_self->indices.strides[0]) ))));

        /* "pandas/_libs/intervaltree.pxi":2048
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point <= self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      }
    }

    /* "pandas/_libs/intervaltree.pxi":2043
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
    goto __pyx_L3;
  }

  /* "pandas/_libs/intervaltree.pxi":2053
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
  /*else*/ {
    __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":2054
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:
 *                 values = self.center_left_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 */
      __pyx_t_16 = __pyx_v_self->center_left_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":2055
 *             if point < self.pivot:
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:
 */
      __pyx_t_17 = __pyx_v_self->center_left_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":2056
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):             # <<<<<<<<<<<<<<
 *                     if not values[i] <= point:
 *                         break
 */
      __pyx_t_9 = __pyx_v_self->n_center;
      __pyx_t_10 = __pyx_t_9;
      for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":2057
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_18 = __pyx_v_i;
        __pyx_t_14 = ((!(((*((__pyx_t_5numpy_int32_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_18 * __pyx_v_values.strides[0]) ))) <= __pyx_v_point) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":2058
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:
 */
          goto __pyx_L9_break;

          /* "pandas/_libs/intervaltree.pxi":2057
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":2059
 *                     if not values[i] <= point:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 */
        __pyx_t_19 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_19 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L9_break:;

      /* "pandas/_libs/intervaltree.pxi":2060
 *                         break
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      __pyx_t_14 = ((__pyx_v_point <= __pyx_v_self->left_node->max_right) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":2061
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)             # <<<<<<<<<<<<<<
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode *)__pyx_v_self->left_node->__pyx_vtab)->__pyx_fuse_1query(__pyx_v_self->left_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2061, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":2060
 *                         break
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":2053
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":2062
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
    __pyx_t_14 = ((__pyx_v_point > __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":2063
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 *                 values = self.center_right_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 */
      __pyx_t_16 = __pyx_v_self->center_right_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":2064
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:
 */
      __pyx_t_17 = __pyx_v_self->center_right_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":2065
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):             # <<<<<<<<<<<<<<
 *                     if not point <= values[i]:
 *                         break
 */
      for (__pyx_t_11 = (__pyx_v_self->n_center - 1); __pyx_t_11 > -1L; __pyx_t_11-=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":2066
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_20 = __pyx_v_i;
        __pyx_t_14 = ((!((__pyx_v_point <= (*((__pyx_t_5numpy_int32_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_20 * __pyx_v_values.strides[0]) )))) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":2067
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:
 */
          goto __pyx_L13_break;

          /* "pandas/_libs/intervaltree.pxi":2066
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":2068
 *                     if not point <= values[i]:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if self.right_node.min_left <= point:
 *                     self.right_node.query(result, point)
 */
        __pyx_t_21 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_21 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L13_break:;

      /* "pandas/_libs/intervaltree.pxi":2069
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      __pyx_t_14 = ((__pyx_v_self->right_node->min_left <= __pyx_v_point) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":2070
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:
 *                     self.right_node.query(result, point)             # <<<<<<<<<<<<<<
 *             else:
 *                 result.extend(self.center_left_indices)
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode *)__pyx_v_self->right_node->__pyx_vtab)->__pyx_fuse_1query(__pyx_v_self->right_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2070, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":2069
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":2062
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":2072
 *                     self.right_node.query(result, point)
 *             else:
 *                 result.extend(self.center_left_indices)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
    /*else*/ {
      __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->extend(__pyx_v_result, __pyx_v_self->center_left_indices); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2072, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_L7:;
  }
  __pyx_L3:;

  /* "pandas/_libs/intervaltree.pxi":2034
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __PYX_XDEC_MEMVIEW(&__pyx_t_16, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_17, 1);
  __Pyx_AddTraceback("pandas._libs.interval.Int32ClosedBothIntervalNode.query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_indices, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_values, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_11__pyx_fuse_1query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_1__pyx_mdef_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_11__pyx_fuse_1query = {"__pyx_fuse_1query", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_11__pyx_fuse_1query, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_2query};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_11__pyx_fuse_1query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result = 0;
  __pyx_t_5numpy_float32_t __pyx_v_point;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fuse_1query (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_result,&__pyx_n_s_point,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_result)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_point)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_1query", 1, 2, 2, 1); __PYX_ERR(1, 2034, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_1query") < 0)) __PYX_ERR(1, 2034, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_result = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)values[0]);
    __pyx_v_point = __pyx_PyFloat_AsFloat(values[1]); if (unlikely((__pyx_v_point == ((npy_float32)-1)) && PyErr_Occurred())) __PYX_ERR(1, 2034, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_1query", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 2034, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Int32ClosedBothIntervalNode.__pyx_fuse_1query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_result), __pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector, 1, "result", 0))) __PYX_ERR(1, 2034, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_10__pyx_fuse_1query(((struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode *)__pyx_v_self), __pyx_v_result, __pyx_v_point);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_10__pyx_fuse_1query(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float32_t __pyx_v_point) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__pyx_fuse_1query", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_fuse_1__pyx_f_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_query(__pyx_v_self, __pyx_v_result, __pyx_v_point, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2034, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Int32ClosedBothIntervalNode.__pyx_fuse_1query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_13__pyx_fuse_2query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_fuse_2__pyx_f_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int64_t __pyx_v_point, int __pyx_skip_dispatch) {
  __Pyx_memviewslice __pyx_v_indices = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_values = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __pyx_t_5numpy_int64_t __pyx_t_9;
  __pyx_t_5numpy_int64_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  int __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  __Pyx_memviewslice __pyx_t_16 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_17 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  Py_ssize_t __pyx_t_21;
  __Pyx_RefNannySetupContext("__pyx_fuse_2query", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pyx_fuse_2query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2034, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_13__pyx_fuse_2query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = __Pyx_PyInt_From_npy_int64(__pyx_v_point); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2034, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2034, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2034, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 2034, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_5) {
            __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2034, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2034, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_3query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_4 = __Pyx_PyInt_From_npy_int64(__pyx_v_point); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 2034, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_7 = __pyx_t_1; __pyx_t_3 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2034, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2034, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 2034, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (__pyx_t_3) {
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_t_4);
          __pyx_t_4 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2034, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pandas/_libs/intervaltree.pxi":2043
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
  __pyx_t_8 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_8) {

    /* "pandas/_libs/intervaltree.pxi":2047
 *             # continue the binary tree structure. Instead, we use linear
 *             # search.
 *             for i in range(self.n_elements):             # <<<<<<<<<<<<<<
 *                 if self.left[i] <= point <= self.right[i]:
 *                     result.append(self.indices[i])
 */
    __pyx_t_9 = __pyx_v_self->n_elements;
    __pyx_t_10 = __pyx_t_9;
    for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
      __pyx_v_i = __pyx_t_11;

      /* "pandas/_libs/intervaltree.pxi":2048
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point <= self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      __pyx_t_12 = __pyx_v_i;
      __pyx_t_8 = ((*((__pyx_t_5numpy_int32_t *) ( /* dim=0 */ (__pyx_v_self->left.data + __pyx_t_12 * __pyx_v_self->left.strides[0]) ))) <= __pyx_v_point);
      if (__pyx_t_8) {
        __pyx_t_13 = __pyx_v_i;
        __pyx_t_8 = (__pyx_v_point <= (*((__pyx_t_5numpy_int32_t *) ( /* dim=0 */ (__pyx_v_self->right.data + __pyx_t_13 * __pyx_v_self->right.strides[0]) ))));
      }
      __pyx_t_14 = (__pyx_t_8 != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":2049
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point <= self.right[i]:
 *                     result.append(self.indices[i])             # <<<<<<<<<<<<<<
 *         else:
 *             # There are child nodes. Based on comparing our query to the pivot,
 */
        __pyx_t_15 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->indices.data + __pyx_t_15 * __pyx_v_self->indices.strides[0]) ))));

        /* "pandas/_libs/intervaltree.pxi":2048
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point <= self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      }
    }

    /* "pandas/_libs/intervaltree.pxi":2043
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
    goto __pyx_L3;
  }

  /* "pandas/_libs/intervaltree.pxi":2053
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
  /*else*/ {
    __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":2054
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:
 *                 values = self.center_left_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 */
      __pyx_t_16 = __pyx_v_self->center_left_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":2055
 *             if point < self.pivot:
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:
 */
      __pyx_t_17 = __pyx_v_self->center_left_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":2056
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):             # <<<<<<<<<<<<<<
 *                     if not values[i] <= point:
 *                         break
 */
      __pyx_t_9 = __pyx_v_self->n_center;
      __pyx_t_10 = __pyx_t_9;
      for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":2057
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_18 = __pyx_v_i;
        __pyx_t_14 = ((!(((*((__pyx_t_5numpy_int32_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_18 * __pyx_v_values.strides[0]) ))) <= __pyx_v_point) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":2058
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:
 */
          goto __pyx_L9_break;

          /* "pandas/_libs/intervaltree.pxi":2057
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":2059
 *                     if not values[i] <= point:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 */
        __pyx_t_19 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_19 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L9_break:;

      /* "pandas/_libs/intervaltree.pxi":2060
 *                         break
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      __pyx_t_14 = ((__pyx_v_point <= __pyx_v_self->left_node->max_right) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":2061
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)             # <<<<<<<<<<<<<<
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode *)__pyx_v_self->left_node->__pyx_vtab)->__pyx_fuse_2query(__pyx_v_self->left_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2061, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":2060
 *                         break
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":2053
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":2062
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
    __pyx_t_14 = ((__pyx_v_point > __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":2063
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 *                 values = self.center_right_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 */
      __pyx_t_16 = __pyx_v_self->center_right_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":2064
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:
 */
      __pyx_t_17 = __pyx_v_self->center_right_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":2065
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):             # <<<<<<<<<<<<<<
 *                     if not point <= values[i]:
 *                         break
 */
      for (__pyx_t_11 = (__pyx_v_self->n_center - 1); __pyx_t_11 > -1L; __pyx_t_11-=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":2066
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_20 = __pyx_v_i;
        __pyx_t_14 = ((!((__pyx_v_point <= (*((__pyx_t_5numpy_int32_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_20 * __pyx_v_values.strides[0]) )))) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":2067
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:
 */
          goto __pyx_L13_break;

          /* "pandas/_libs/intervaltree.pxi":2066
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":2068
 *                     if not point <= values[i]:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if self.right_node.min_left <= point:
 *                     self.right_node.query(result, point)
 */
        __pyx_t_21 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_21 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L13_break:;

      /* "pandas/_libs/intervaltree.pxi":2069
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      __pyx_t_14 = ((__pyx_v_self->right_node->min_left <= __pyx_v_point) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":2070
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:
 *                     self.right_node.query(result, point)             # <<<<<<<<<<<<<<
 *             else:
 *                 result.extend(self.center_left_indices)
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode *)__pyx_v_self->right_node->__pyx_vtab)->__pyx_fuse_2query(__pyx_v_self->right_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2070, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":2069
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":2062
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":2072
 *                     self.right_node.query(result, point)
 *             else:
 *                 result.extend(self.center_left_indices)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
    /*else*/ {
      __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->extend(__pyx_v_result, __pyx_v_self->center_left_indices); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2072, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_L7:;
  }
  __pyx_L3:;

  /* "pandas/_libs/intervaltree.pxi":2034
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __PYX_XDEC_MEMVIEW(&__pyx_t_16, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_17, 1);
  __Pyx_AddTraceback("pandas._libs.interval.Int32ClosedBothIntervalNode.query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_indices, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_values, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_13__pyx_fuse_2query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_2__pyx_mdef_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_13__pyx_fuse_2query = {"__pyx_fuse_2query", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_13__pyx_fuse_2query, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_2query};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_13__pyx_fuse_2query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result = 0;
  __pyx_t_5numpy_int64_t __pyx_v_point;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fuse_2query (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_result,&__pyx_n_s_point,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_result)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_point)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_2query", 1, 2, 2, 1); __PYX_ERR(1, 2034, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_2query") < 0)) __PYX_ERR(1, 2034, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_result = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)values[0]);
    __pyx_v_point = __Pyx_PyInt_As_npy_int64(values[1]); if (unlikely((__pyx_v_point == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(1, 2034, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_2query", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 2034, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Int32ClosedBothIntervalNode.__pyx_fuse_2query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_result), __pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector, 1, "result", 0))) __PYX_ERR(1, 2034, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_12__pyx_fuse_2query(((struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode *)__pyx_v_self), __pyx_v_result, __pyx_v_point);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_12__pyx_fuse_2query(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int64_t __pyx_v_point) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__pyx_fuse_2query", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_fuse_2__pyx_f_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_query(__pyx_v_self, __pyx_v_result, __pyx_v_point, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2034, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Int32ClosedBothIntervalNode.__pyx_fuse_2query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_15__pyx_fuse_3query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_fuse_3__pyx_f_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int32_t __pyx_v_point, int __pyx_skip_dispatch) {
  __Pyx_memviewslice __pyx_v_indices = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_values = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __pyx_t_5numpy_int64_t __pyx_t_9;
  __pyx_t_5numpy_int64_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  int __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  __Pyx_memviewslice __pyx_t_16 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_17 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  Py_ssize_t __pyx_t_21;
  __Pyx_RefNannySetupContext("__pyx_fuse_3query", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pyx_fuse_3query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2034, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_15__pyx_fuse_3query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = __Pyx_PyInt_From_npy_int32(__pyx_v_point); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2034, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2034, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2034, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 2034, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_5) {
            __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2034, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2034, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_3query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_4 = __Pyx_PyInt_From_npy_int32(__pyx_v_point); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 2034, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_7 = __pyx_t_1; __pyx_t_3 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2034, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2034, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 2034, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (__pyx_t_3) {
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_t_4);
          __pyx_t_4 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2034, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pandas/_libs/intervaltree.pxi":2043
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
  __pyx_t_8 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_8) {

    /* "pandas/_libs/intervaltree.pxi":2047
 *             # continue the binary tree structure. Instead, we use linear
 *             # search.
 *             for i in range(self.n_elements):             # <<<<<<<<<<<<<<
 *                 if self.left[i] <= point <= self.right[i]:
 *                     result.append(self.indices[i])
 */
    __pyx_t_9 = __pyx_v_self->n_elements;
    __pyx_t_10 = __pyx_t_9;
    for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
      __pyx_v_i = __pyx_t_11;

      /* "pandas/_libs/intervaltree.pxi":2048
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point <= self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      __pyx_t_12 = __pyx_v_i;
      __pyx_t_8 = ((*((__pyx_t_5numpy_int32_t *) ( /* dim=0 */ (__pyx_v_self->left.data + __pyx_t_12 * __pyx_v_self->left.strides[0]) ))) <= __pyx_v_point);
      if (__pyx_t_8) {
        __pyx_t_13 = __pyx_v_i;
        __pyx_t_8 = (__pyx_v_point <= (*((__pyx_t_5numpy_int32_t *) ( /* dim=0 */ (__pyx_v_self->right.data + __pyx_t_13 * __pyx_v_self->right.strides[0]) ))));
      }
      __pyx_t_14 = (__pyx_t_8 != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":2049
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point <= self.right[i]:
 *                     result.append(self.indices[i])             # <<<<<<<<<<<<<<
 *         else:
 *             # There are child nodes. Based on comparing our query to the pivot,
 */
        __pyx_t_15 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->indices.data + __pyx_t_15 * __pyx_v_self->indices.strides[0]) ))));

        /* "pandas/_libs/intervaltree.pxi":2048
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point <= self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      }
    }

    /* "pandas/_libs/intervaltree.pxi":2043
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
    goto __pyx_L3;
  }

  /* "pandas/_libs/intervaltree.pxi":2053
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
  /*else*/ {
    __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":2054
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:
 *                 values = self.center_left_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 */
      __pyx_t_16 = __pyx_v_self->center_left_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":2055
 *             if point < self.pivot:
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:
 */
      __pyx_t_17 = __pyx_v_self->center_left_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":2056
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):             # <<<<<<<<<<<<<<
 *                     if not values[i] <= point:
 *                         break
 */
      __pyx_t_9 = __pyx_v_self->n_center;
      __pyx_t_10 = __pyx_t_9;
      for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":2057
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_18 = __pyx_v_i;
        __pyx_t_14 = ((!(((*((__pyx_t_5numpy_int32_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_18 * __pyx_v_values.strides[0]) ))) <= __pyx_v_point) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":2058
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:
 */
          goto __pyx_L9_break;

          /* "pandas/_libs/intervaltree.pxi":2057
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":2059
 *                     if not values[i] <= point:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 */
        __pyx_t_19 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_19 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L9_break:;

      /* "pandas/_libs/intervaltree.pxi":2060
 *                         break
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      __pyx_t_14 = ((__pyx_v_point <= __pyx_v_self->left_node->max_right) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":2061
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)             # <<<<<<<<<<<<<<
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode *)__pyx_v_self->left_node->__pyx_vtab)->__pyx_fuse_3query(__pyx_v_self->left_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2061, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":2060
 *                         break
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":2053
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":2062
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
    __pyx_t_14 = ((__pyx_v_point > __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":2063
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 *                 values = self.center_right_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 */
      __pyx_t_16 = __pyx_v_self->center_right_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":2064
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:
 */
      __pyx_t_17 = __pyx_v_self->center_right_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":2065
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):             # <<<<<<<<<<<<<<
 *                     if not point <= values[i]:
 *                         break
 */
      for (__pyx_t_11 = (__pyx_v_self->n_center - 1); __pyx_t_11 > -1L; __pyx_t_11-=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":2066
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_20 = __pyx_v_i;
        __pyx_t_14 = ((!((__pyx_v_point <= (*((__pyx_t_5numpy_int32_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_20 * __pyx_v_values.strides[0]) )))) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":2067
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:
 */
          goto __pyx_L13_break;

          /* "pandas/_libs/intervaltree.pxi":2066
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":2068
 *                     if not point <= values[i]:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if self.right_node.min_left <= point:
 *                     self.right_node.query(result, point)
 */
        __pyx_t_21 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_21 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L13_break:;

      /* "pandas/_libs/intervaltree.pxi":2069
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      __pyx_t_14 = ((__pyx_v_self->right_node->min_left <= __pyx_v_point) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":2070
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:
 *                     self.right_node.query(result, point)             # <<<<<<<<<<<<<<
 *             else:
 *                 result.extend(self.center_left_indices)
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode *)__pyx_v_self->right_node->__pyx_vtab)->__pyx_fuse_3query(__pyx_v_self->right_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2070, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":2069
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":2062
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":2072
 *                     self.right_node.query(result, point)
 *             else:
 *                 result.extend(self.center_left_indices)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
    /*else*/ {
      __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->extend(__pyx_v_result, __pyx_v_self->center_left_indices); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2072, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_L7:;
  }
  __pyx_L3:;

  /* "pandas/_libs/intervaltree.pxi":2034
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __PYX_XDEC_MEMVIEW(&__pyx_t_16, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_17, 1);
  __Pyx_AddTraceback("pandas._libs.interval.Int32ClosedBothIntervalNode.query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_indices, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_values, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_15__pyx_fuse_3query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_3__pyx_mdef_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_15__pyx_fuse_3query = {"__pyx_fuse_3query", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_15__pyx_fuse_3query, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_2query};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_15__pyx_fuse_3query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result = 0;
  __pyx_t_5numpy_int32_t __pyx_v_point;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fuse_3query (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_result,&__pyx_n_s_point,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_result)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_point)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_3query", 1, 2, 2, 1); __PYX_ERR(1, 2034, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_3query") < 0)) __PYX_ERR(1, 2034, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_result = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)values[0]);
    __pyx_v_point = __Pyx_PyInt_As_npy_int32(values[1]); if (unlikely((__pyx_v_point == ((npy_int32)-1)) && PyErr_Occurred())) __PYX_ERR(1, 2034, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_3query", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 2034, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Int32ClosedBothIntervalNode.__pyx_fuse_3query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_result), __pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector, 1, "result", 0))) __PYX_ERR(1, 2034, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_14__pyx_fuse_3query(((struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode *)__pyx_v_self), __pyx_v_result, __pyx_v_point);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_14__pyx_fuse_3query(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int32_t __pyx_v_point) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__pyx_fuse_3query", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_fuse_3__pyx_f_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_query(__pyx_v_self, __pyx_v_result, __pyx_v_point, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2034, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Int32ClosedBothIntervalNode.__pyx_fuse_3query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_17__pyx_fuse_4query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_fuse_4__pyx_f_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_uint64_t __pyx_v_point, int __pyx_skip_dispatch) {
  __Pyx_memviewslice __pyx_v_indices = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_values = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __pyx_t_5numpy_int64_t __pyx_t_9;
  __pyx_t_5numpy_int64_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  int __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  __Pyx_memviewslice __pyx_t_16 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_17 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  Py_ssize_t __pyx_t_21;
  __Pyx_RefNannySetupContext("__pyx_fuse_4query", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pyx_fuse_4query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2034, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_17__pyx_fuse_4query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = __Pyx_PyInt_From_npy_uint64(__pyx_v_point); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2034, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2034, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2034, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 2034, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_5) {
            __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2034, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2034, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_3query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_4 = __Pyx_PyInt_From_npy_uint64(__pyx_v_point); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 2034, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_7 = __pyx_t_1; __pyx_t_3 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2034, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2034, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 2034, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (__pyx_t_3) {
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_t_4);
          __pyx_t_4 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2034, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pandas/_libs/intervaltree.pxi":2043
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
  __pyx_t_8 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_8) {

    /* "pandas/_libs/intervaltree.pxi":2047
 *             # continue the binary tree structure. Instead, we use linear
 *             # search.
 *             for i in range(self.n_elements):             # <<<<<<<<<<<<<<
 *                 if self.left[i] <= point <= self.right[i]:
 *                     result.append(self.indices[i])
 */
    __pyx_t_9 = __pyx_v_self->n_elements;
    __pyx_t_10 = __pyx_t_9;
    for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
      __pyx_v_i = __pyx_t_11;

      /* "pandas/_libs/intervaltree.pxi":2048
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point <= self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      __pyx_t_12 = __pyx_v_i;
      __pyx_t_8 = ((*((__pyx_t_5numpy_int32_t *) ( /* dim=0 */ (__pyx_v_self->left.data + __pyx_t_12 * __pyx_v_self->left.strides[0]) ))) <= __pyx_v_point);
      if (__pyx_t_8) {
        __pyx_t_13 = __pyx_v_i;
        __pyx_t_8 = (__pyx_v_point <= (*((__pyx_t_5numpy_int32_t *) ( /* dim=0 */ (__pyx_v_self->right.data + __pyx_t_13 * __pyx_v_self->right.strides[0]) ))));
      }
      __pyx_t_14 = (__pyx_t_8 != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":2049
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point <= self.right[i]:
 *                     result.append(self.indices[i])             # <<<<<<<<<<<<<<
 *         else:
 *             # There are child nodes. Based on comparing our query to the pivot,
 */
        __pyx_t_15 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->indices.data + __pyx_t_15 * __pyx_v_self->indices.strides[0]) ))));

        /* "pandas/_libs/intervaltree.pxi":2048
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point <= self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      }
    }

    /* "pandas/_libs/intervaltree.pxi":2043
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
    goto __pyx_L3;
  }

  /* "pandas/_libs/intervaltree.pxi":2053
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
  /*else*/ {
    __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":2054
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:
 *                 values = self.center_left_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 */
      __pyx_t_16 = __pyx_v_self->center_left_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":2055
 *             if point < self.pivot:
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:
 */
      __pyx_t_17 = __pyx_v_self->center_left_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":2056
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):             # <<<<<<<<<<<<<<
 *                     if not values[i] <= point:
 *                         break
 */
      __pyx_t_9 = __pyx_v_self->n_center;
      __pyx_t_10 = __pyx_t_9;
      for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":2057
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_18 = __pyx_v_i;
        __pyx_t_14 = ((!(((*((__pyx_t_5numpy_int32_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_18 * __pyx_v_values.strides[0]) ))) <= __pyx_v_point) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":2058
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:
 */
          goto __pyx_L9_break;

          /* "pandas/_libs/intervaltree.pxi":2057
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":2059
 *                     if not values[i] <= point:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 */
        __pyx_t_19 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_19 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L9_break:;

      /* "pandas/_libs/intervaltree.pxi":2060
 *                         break
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      __pyx_t_14 = ((__pyx_v_point <= __pyx_v_self->left_node->max_right) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":2061
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)             # <<<<<<<<<<<<<<
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode *)__pyx_v_self->left_node->__pyx_vtab)->__pyx_fuse_4query(__pyx_v_self->left_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2061, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":2060
 *                         break
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":2053
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":2062
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
    __pyx_t_14 = ((__pyx_v_point > __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":2063
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 *                 values = self.center_right_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 */
      __pyx_t_16 = __pyx_v_self->center_right_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":2064
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:
 */
      __pyx_t_17 = __pyx_v_self->center_right_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":2065
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):             # <<<<<<<<<<<<<<
 *                     if not point <= values[i]:
 *                         break
 */
      for (__pyx_t_11 = (__pyx_v_self->n_center - 1); __pyx_t_11 > -1L; __pyx_t_11-=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":2066
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_20 = __pyx_v_i;
        __pyx_t_14 = ((!((__pyx_v_point <= (*((__pyx_t_5numpy_int32_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_20 * __pyx_v_values.strides[0]) )))) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":2067
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:
 */
          goto __pyx_L13_break;

          /* "pandas/_libs/intervaltree.pxi":2066
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":2068
 *                     if not point <= values[i]:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if self.right_node.min_left <= point:
 *                     self.right_node.query(result, point)
 */
        __pyx_t_21 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_21 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L13_break:;

      /* "pandas/_libs/intervaltree.pxi":2069
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      __pyx_t_14 = ((__pyx_v_self->right_node->min_left <= __pyx_v_point) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":2070
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:
 *                     self.right_node.query(result, point)             # <<<<<<<<<<<<<<
 *             else:
 *                 result.extend(self.center_left_indices)
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode *)__pyx_v_self->right_node->__pyx_vtab)->__pyx_fuse_4query(__pyx_v_self->right_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2070, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":2069
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":2062
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":2072
 *                     self.right_node.query(result, point)
 *             else:
 *                 result.extend(self.center_left_indices)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
    /*else*/ {
      __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->extend(__pyx_v_result, __pyx_v_self->center_left_indices); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2072, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_L7:;
  }
  __pyx_L3:;

  /* "pandas/_libs/intervaltree.pxi":2034
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __PYX_XDEC_MEMVIEW(&__pyx_t_16, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_17, 1);
  __Pyx_AddTraceback("pandas._libs.interval.Int32ClosedBothIntervalNode.query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_indices, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_values, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_17__pyx_fuse_4query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_4__pyx_mdef_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_17__pyx_fuse_4query = {"__pyx_fuse_4query", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_17__pyx_fuse_4query, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_2query};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_17__pyx_fuse_4query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result = 0;
  __pyx_t_5numpy_uint64_t __pyx_v_point;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fuse_4query (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_result,&__pyx_n_s_point,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_result)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_point)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_4query", 1, 2, 2, 1); __PYX_ERR(1, 2034, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_4query") < 0)) __PYX_ERR(1, 2034, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_result = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)values[0]);
    __pyx_v_point = __Pyx_PyInt_As_npy_uint64(values[1]); if (unlikely((__pyx_v_point == ((npy_uint64)-1)) && PyErr_Occurred())) __PYX_ERR(1, 2034, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_4query", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 2034, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Int32ClosedBothIntervalNode.__pyx_fuse_4query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_result), __pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector, 1, "result", 0))) __PYX_ERR(1, 2034, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_16__pyx_fuse_4query(((struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode *)__pyx_v_self), __pyx_v_result, __pyx_v_point);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_16__pyx_fuse_4query(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_uint64_t __pyx_v_point) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__pyx_fuse_4query", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_fuse_4__pyx_f_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_query(__pyx_v_self, __pyx_v_result, __pyx_v_point, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2034, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Int32ClosedBothIntervalNode.__pyx_fuse_4query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":2074
 *                 result.extend(self.center_left_indices)
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         if self.is_leaf_node:
 *             return ('<Int32ClosedBothIntervalNode: '
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_5__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_5__repr__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_4__repr__(((struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_4__repr__(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode *__pyx_v_self) {
  PyObject *__pyx_v_n_left = NULL;
  PyObject *__pyx_v_n_right = NULL;
  PyObject *__pyx_v_n_center = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  Py_UCS4 __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "pandas/_libs/intervaltree.pxi":2075
 * 
 *     def __repr__(self):
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             return ('<Int32ClosedBothIntervalNode: '
 *                     '%s elements (terminal)>' % self.n_elements)
 */
  __pyx_t_1 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_1) {

    /* "pandas/_libs/intervaltree.pxi":2076
 *     def __repr__(self):
 *         if self.is_leaf_node:
 *             return ('<Int32ClosedBothIntervalNode: '             # <<<<<<<<<<<<<<
 *                     '%s elements (terminal)>' % self.n_elements)
 *         else:
 */
    __Pyx_XDECREF(__pyx_r);

    /* "pandas/_libs/intervaltree.pxi":2077
 *         if self.is_leaf_node:
 *             return ('<Int32ClosedBothIntervalNode: '
 *                     '%s elements (terminal)>' % self.n_elements)             # <<<<<<<<<<<<<<
 *         else:
 *             n_left = self.left_node.n_elements
 */
    __pyx_t_2 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_elements); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2077, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PyUnicode_Format(__pyx_kp_u_Int32ClosedBothIntervalNode_s_e, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2077, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pandas/_libs/intervaltree.pxi":2075
 * 
 *     def __repr__(self):
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             return ('<Int32ClosedBothIntervalNode: '
 *                     '%s elements (terminal)>' % self.n_elements)
 */
  }

  /* "pandas/_libs/intervaltree.pxi":2079
 *                     '%s elements (terminal)>' % self.n_elements)
 *         else:
 *             n_left = self.left_node.n_elements             # <<<<<<<<<<<<<<
 *             n_right = self.right_node.n_elements
 *             n_center = self.n_elements - n_left - n_right
 */
  /*else*/ {
    __pyx_t_3 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->left_node->n_elements); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2079, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_v_n_left = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "pandas/_libs/intervaltree.pxi":2080
 *         else:
 *             n_left = self.left_node.n_elements
 *             n_right = self.right_node.n_elements             # <<<<<<<<<<<<<<
 *             n_center = self.n_elements - n_left - n_right
 *             return ('<Int32ClosedBothIntervalNode: '
 */
    __pyx_t_3 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->right_node->n_elements); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2080, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_v_n_right = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "pandas/_libs/intervaltree.pxi":2081
 *             n_left = self.left_node.n_elements
 *             n_right = self.right_node.n_elements
 *             n_center = self.n_elements - n_left - n_right             # <<<<<<<<<<<<<<
 *             return ('<Int32ClosedBothIntervalNode: '
 *                     'pivot %s, %s elements (%s left, %s right, %s '
 */
    __pyx_t_3 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_elements); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2081, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = PyNumber_Subtract(__pyx_t_3, __pyx_v_n_left); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2081, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyNumber_Subtract(__pyx_t_2, __pyx_v_n_right); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2081, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_n_center = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "pandas/_libs/intervaltree.pxi":2082
 *             n_right = self.right_node.n_elements
 *             n_center = self.n_elements - n_left - n_right
 *             return ('<Int32ClosedBothIntervalNode: '             # <<<<<<<<<<<<<<
 *                     'pivot %s, %s elements (%s left, %s right, %s '
 *                     'overlapping)>' % (self.pivot, self.n_elements,
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = PyTuple_New(11); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2082, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = 0;
    __pyx_t_5 = 127;
    __Pyx_INCREF(__pyx_kp_u_Int32ClosedBothIntervalNode_piv);
    __pyx_t_4 += 36;
    __Pyx_GIVEREF(__pyx_kp_u_Int32ClosedBothIntervalNode_piv);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_kp_u_Int32ClosedBothIntervalNode_piv);

    /* "pandas/_libs/intervaltree.pxi":2084
 *             return ('<Int32ClosedBothIntervalNode: '
 *                     'pivot %s, %s elements (%s left, %s right, %s '
 *                     'overlapping)>' % (self.pivot, self.n_elements,             # <<<<<<<<<<<<<<
 *                                        n_left, n_right, n_center))
 * 
 */
    __pyx_t_2 = __Pyx_PyInt_From_npy_int32(__pyx_v_self->pivot); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2084, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_t_2), __pyx_empty_unicode); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 2084, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_5 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_6) > __pyx_t_5) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_6) : __pyx_t_5;
    __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_6);
    __pyx_t_6 = 0;
    __Pyx_INCREF(__pyx_kp_u__18);
    __pyx_t_4 += 2;
    __Pyx_GIVEREF(__pyx_kp_u__18);
    PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_kp_u__18);
    __pyx_t_6 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_elements); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 2084, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_2 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_t_6), __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2084, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_5 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) > __pyx_t_5) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) : __pyx_t_5;
    __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_3, 3, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_elements);
    __pyx_t_4 += 11;
    __Pyx_GIVEREF(__pyx_kp_u_elements);
    PyTuple_SET_ITEM(__pyx_t_3, 4, __pyx_kp_u_elements);

    /* "pandas/_libs/intervaltree.pxi":2085
 *                     'pivot %s, %s elements (%s left, %s right, %s '
 *                     'overlapping)>' % (self.pivot, self.n_elements,
 *                                        n_left, n_right, n_center))             # <<<<<<<<<<<<<<
 * 
 *     def counts(self):
 */
    __pyx_t_2 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_v_n_left), __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2085, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) > __pyx_t_5) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) : __pyx_t_5;
    __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_3, 5, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_left_2);
    __pyx_t_4 += 7;
    __Pyx_GIVEREF(__pyx_kp_u_left_2);
    PyTuple_SET_ITEM(__pyx_t_3, 6, __pyx_kp_u_left_2);
    __pyx_t_2 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_v_n_right), __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2085, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) > __pyx_t_5) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) : __pyx_t_5;
    __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_3, 7, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_right_2);
    __pyx_t_4 += 8;
    __Pyx_GIVEREF(__pyx_kp_u_right_2);
    PyTuple_SET_ITEM(__pyx_t_3, 8, __pyx_kp_u_right_2);
    __pyx_t_2 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_v_n_center), __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2085, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) > __pyx_t_5) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) : __pyx_t_5;
    __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_3, 9, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_overlapping);
    __pyx_t_4 += 14;
    __Pyx_GIVEREF(__pyx_kp_u_overlapping);
    PyTuple_SET_ITEM(__pyx_t_3, 10, __pyx_kp_u_overlapping);

    /* "pandas/_libs/intervaltree.pxi":2082
 *             n_right = self.right_node.n_elements
 *             n_center = self.n_elements - n_left - n_right
 *             return ('<Int32ClosedBothIntervalNode: '             # <<<<<<<<<<<<<<
 *                     'pivot %s, %s elements (%s left, %s right, %s '
 *                     'overlapping)>' % (self.pivot, self.n_elements,
 */
    __pyx_t_2 = __Pyx_PyUnicode_Join(__pyx_t_3, 11, __pyx_t_4, __pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2082, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;
  }

  /* "pandas/_libs/intervaltree.pxi":2074
 *                 result.extend(self.center_left_indices)
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         if self.is_leaf_node:
 *             return ('<Int32ClosedBothIntervalNode: '
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pandas._libs.interval.Int32ClosedBothIntervalNode.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_n_left);
  __Pyx_XDECREF(__pyx_v_n_right);
  __Pyx_XDECREF(__pyx_v_n_center);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":2087
 *                                        n_left, n_right, n_center))
 * 
 *     def counts(self):             # <<<<<<<<<<<<<<
 *         """
 *         Inspect counts on this node
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_7counts(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_6counts[] = "\n        Inspect counts on this node\n        useful for debugging purposes\n        ";
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_7counts(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("counts (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_6counts(((struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_6counts(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode *__pyx_v_self) {
  Py_ssize_t __pyx_v_m;
  PyObject *__pyx_v_l = NULL;
  PyObject *__pyx_v_r = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  size_t __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("counts", 0);

  /* "pandas/_libs/intervaltree.pxi":2092
 *         useful for debugging purposes
 *         """
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             return self.n_elements
 *         else:
 */
  __pyx_t_1 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_1) {

    /* "pandas/_libs/intervaltree.pxi":2093
 *         """
 *         if self.is_leaf_node:
 *             return self.n_elements             # <<<<<<<<<<<<<<
 *         else:
 *             m = len(self.center_left_values)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_elements); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2093, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "pandas/_libs/intervaltree.pxi":2092
 *         useful for debugging purposes
 *         """
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             return self.n_elements
 *         else:
 */
  }

  /* "pandas/_libs/intervaltree.pxi":2095
 *             return self.n_elements
 *         else:
 *             m = len(self.center_left_values)             # <<<<<<<<<<<<<<
 *             l = self.left_node.counts()
 *             r = self.right_node.counts()
 */
  /*else*/ {
    if (unlikely(!__pyx_v_self->center_left_values.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(1, 2095, __pyx_L1_error)}
    __pyx_t_3 = __Pyx_MemoryView_Len(__pyx_v_self->center_left_values); 
    __pyx_v_m = __pyx_t_3;

    /* "pandas/_libs/intervaltree.pxi":2096
 *         else:
 *             m = len(self.center_left_values)
 *             l = self.left_node.counts()             # <<<<<<<<<<<<<<
 *             r = self.right_node.counts()
 *             return (m, (l, r))
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->left_node), __pyx_n_s_counts); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 2096, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2096, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_l = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "pandas/_libs/intervaltree.pxi":2097
 *             m = len(self.center_left_values)
 *             l = self.left_node.counts()
 *             r = self.right_node.counts()             # <<<<<<<<<<<<<<
 *             return (m, (l, r))
 * 
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->right_node), __pyx_n_s_counts); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 2097, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2097, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_r = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "pandas/_libs/intervaltree.pxi":2098
 *             l = self.left_node.counts()
 *             r = self.right_node.counts()
 *             return (m, (l, r))             # <<<<<<<<<<<<<<
 * 
 * NODE_CLASSES['int32',
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = PyInt_FromSsize_t(__pyx_v_m); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2098, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 2098, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_v_l);
    __Pyx_GIVEREF(__pyx_v_l);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_l);
    __Pyx_INCREF(__pyx_v_r);
    __Pyx_GIVEREF(__pyx_v_r);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_v_r);
    __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 2098, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_4);
    __pyx_t_2 = 0;
    __pyx_t_4 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;
  }

  /* "pandas/_libs/intervaltree.pxi":2087
 *                                        n_left, n_right, n_center))
 * 
 *     def counts(self):             # <<<<<<<<<<<<<<
 *         """
 *         Inspect counts on this node
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pandas._libs.interval.Int32ClosedBothIntervalNode.counts", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_l);
  __Pyx_XDECREF(__pyx_v_r);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":1944
 *         int64_t[:] center_left_indices, center_right_indices, indices
 *         int32_t min_left, max_right
 *         readonly int32_t pivot             # <<<<<<<<<<<<<<
 *         readonly int64_t n_elements, n_center, leaf_size
 *         readonly bint is_leaf_node
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_5pivot_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_5pivot_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_5pivot___get__(((struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_5pivot___get__(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_npy_int32(__pyx_v_self->pivot); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1944, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Int32ClosedBothIntervalNode.pivot.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":1945
 *         int32_t min_left, max_right
 *         readonly int32_t pivot
 *         readonly int64_t n_elements, n_center, leaf_size             # <<<<<<<<<<<<<<
 *         readonly bint is_leaf_node
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_10n_elements_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_10n_elements_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_10n_elements___get__(((struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_10n_elements___get__(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_elements); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1945, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Int32ClosedBothIntervalNode.n_elements.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_8n_center_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_8n_center_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_8n_center___get__(((struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_8n_center___get__(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_center); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1945, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Int32ClosedBothIntervalNode.n_center.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_9leaf_size_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_9leaf_size_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_9leaf_size___get__(((struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_9leaf_size___get__(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->leaf_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1945, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Int32ClosedBothIntervalNode.leaf_size.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":1946
 *         readonly int32_t pivot
 *         readonly int64_t n_elements, n_center, leaf_size
 *         readonly bint is_leaf_node             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self,
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_12is_leaf_node_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_12is_leaf_node_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_12is_leaf_node___get__(((struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_12is_leaf_node___get__(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->is_leaf_node); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1946, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Int32ClosedBothIntervalNode.is_leaf_node.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_21__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_21__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_20__reduce_cython__(((struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_20__reduce_cython__(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  int __pyx_t_16;
  int __pyx_t_17;
  int __pyx_t_18;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.center_left_indices, self.center_left_values, self.center_right_indices, self.center_right_values, self.indices, self.is_leaf_node, self.leaf_size, self.left, self.left_node, self.max_right, self.min_left, self.n_center, self.n_elements, self.pivot, self.right, self.right_node)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  if (unlikely(!__pyx_v_self->center_left_indices.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_1 = __pyx_memoryview_fromslice(__pyx_v_self->center_left_indices, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_int64_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_int64_t, 0);; if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (unlikely(!__pyx_v_self->center_left_values.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_2 = __pyx_memoryview_fromslice(__pyx_v_self->center_left_values, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_int32_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_int32_t, 0);; if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (unlikely(!__pyx_v_self->center_right_indices.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_3 = __pyx_memoryview_fromslice(__pyx_v_self->center_right_indices, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_int64_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_int64_t, 0);; if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely(!__pyx_v_self->center_right_values.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_4 = __pyx_memoryview_fromslice(__pyx_v_self->center_right_values, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_int32_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_int32_t, 0);; if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (unlikely(!__pyx_v_self->indices.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_5 = __pyx_memoryview_fromslice(__pyx_v_self->indices, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_int64_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_int64_t, 0);; if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyBool_FromLong(__pyx_v_self->is_leaf_node); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->leaf_size); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (unlikely(!__pyx_v_self->left.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_8 = __pyx_memoryview_fromslice(__pyx_v_self->left, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_int32_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_int32_t, 0);; if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_PyInt_From_npy_int32(__pyx_v_self->max_right); if (unlikely(!__pyx_t_9)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_10 = __Pyx_PyInt_From_npy_int32(__pyx_v_self->min_left); if (unlikely(!__pyx_t_10)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_11 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_center); if (unlikely(!__pyx_t_11)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_12 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_elements); if (unlikely(!__pyx_t_12)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __pyx_t_13 = __Pyx_PyInt_From_npy_int32(__pyx_v_self->pivot); if (unlikely(!__pyx_t_13)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  if (unlikely(!__pyx_v_self->right.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_14 = __pyx_memoryview_fromslice(__pyx_v_self->right, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_int32_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_int32_t, 0);; if (unlikely(!__pyx_t_14)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __pyx_t_15 = PyTuple_New(16); if (unlikely(!__pyx_t_15)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_15, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_15, 1, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_15, 2, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_15, 3, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_15, 4, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_15, 5, __pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_15, 6, __pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_15, 7, __pyx_t_8);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->left_node));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self->left_node));
  PyTuple_SET_ITEM(__pyx_t_15, 8, ((PyObject *)__pyx_v_self->left_node));
  __Pyx_GIVEREF(__pyx_t_9);
  PyTuple_SET_ITEM(__pyx_t_15, 9, __pyx_t_9);
  __Pyx_GIVEREF(__pyx_t_10);
  PyTuple_SET_ITEM(__pyx_t_15, 10, __pyx_t_10);
  __Pyx_GIVEREF(__pyx_t_11);
  PyTuple_SET_ITEM(__pyx_t_15, 11, __pyx_t_11);
  __Pyx_GIVEREF(__pyx_t_12);
  PyTuple_SET_ITEM(__pyx_t_15, 12, __pyx_t_12);
  __Pyx_GIVEREF(__pyx_t_13);
  PyTuple_SET_ITEM(__pyx_t_15, 13, __pyx_t_13);
  __Pyx_GIVEREF(__pyx_t_14);
  PyTuple_SET_ITEM(__pyx_t_15, 14, __pyx_t_14);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->right_node));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self->right_node));
  PyTuple_SET_ITEM(__pyx_t_15, 15, ((PyObject *)__pyx_v_self->right_node));
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  __pyx_t_6 = 0;
  __pyx_t_7 = 0;
  __pyx_t_8 = 0;
  __pyx_t_9 = 0;
  __pyx_t_10 = 0;
  __pyx_t_11 = 0;
  __pyx_t_12 = 0;
  __pyx_t_13 = 0;
  __pyx_t_14 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_15);
  __pyx_t_15 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.center_left_indices, self.center_left_values, self.center_right_indices, self.center_right_values, self.indices, self.is_leaf_node, self.leaf_size, self.left, self.left_node, self.max_right, self.min_left, self.n_center, self.n_elements, self.pivot, self.right, self.right_node)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_15 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_15)) __PYX_ERR(2, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __pyx_v__dict = __pyx_t_15;
  __pyx_t_15 = 0;

  /* "(tree fragment)":7
 *     state = (self.center_left_indices, self.center_left_values, self.center_right_indices, self.center_right_values, self.indices, self.is_leaf_node, self.leaf_size, self.left, self.left_node, self.max_right, self.min_left, self.n_center, self.n_elements, self.pivot, self.right, self.right_node)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_16 = (__pyx_v__dict != Py_None);
  __pyx_t_17 = (__pyx_t_16 != 0);
  if (__pyx_t_17) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_15 = PyTuple_New(1); if (unlikely(!__pyx_t_15)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    PyTuple_SET_ITEM(__pyx_t_15, 0, __pyx_v__dict);
    __pyx_t_14 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_15); if (unlikely(!__pyx_t_14)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_14));
    __pyx_t_14 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.left_node is not None or self.right_node is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.center_left_indices, self.center_left_values, self.center_right_indices, self.center_right_values, self.indices, self.is_leaf_node, self.leaf_size, self.left, self.left_node, self.max_right, self.min_left, self.n_center, self.n_elements, self.pivot, self.right, self.right_node)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.left_node is not None or self.right_node is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_Int32ClosedBothIntervalNode, (type(self), 0x7a104e5, None), state
 */
  /*else*/ {
    __pyx_t_16 = (((PyObject *)__pyx_v_self->left_node) != Py_None);
    __pyx_t_18 = (__pyx_t_16 != 0);
    if (!__pyx_t_18) {
    } else {
      __pyx_t_17 = __pyx_t_18;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_18 = (((PyObject *)__pyx_v_self->right_node) != Py_None);
    __pyx_t_16 = (__pyx_t_18 != 0);
    __pyx_t_17 = __pyx_t_16;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_17;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.left_node is not None or self.right_node is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Int32ClosedBothIntervalNode, (type(self), 0x7a104e5, None), state
 *     else:
 */
  __pyx_t_17 = (__pyx_v_use_setstate != 0);
  if (__pyx_t_17) {

    /* "(tree fragment)":13
 *         use_setstate = self.left_node is not None or self.right_node is not None
 *     if use_setstate:
 *         return __pyx_unpickle_Int32ClosedBothIntervalNode, (type(self), 0x7a104e5, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_Int32ClosedBothIntervalNode, (type(self), 0x7a104e5, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_14, __pyx_n_s_pyx_unpickle_Int32ClosedBothIn); if (unlikely(!__pyx_t_14)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __pyx_t_15 = PyTuple_New(3); if (unlikely(!__pyx_t_15)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_15, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_127993061);
    __Pyx_GIVEREF(__pyx_int_127993061);
    PyTuple_SET_ITEM(__pyx_t_15, 1, __pyx_int_127993061);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_15, 2, Py_None);
    __pyx_t_13 = PyTuple_New(3); if (unlikely(!__pyx_t_13)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __Pyx_GIVEREF(__pyx_t_14);
    PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_t_14);
    __Pyx_GIVEREF(__pyx_t_15);
    PyTuple_SET_ITEM(__pyx_t_13, 1, __pyx_t_15);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_13, 2, __pyx_v_state);
    __pyx_t_14 = 0;
    __pyx_t_15 = 0;
    __pyx_r = __pyx_t_13;
    __pyx_t_13 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.left_node is not None or self.right_node is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Int32ClosedBothIntervalNode, (type(self), 0x7a104e5, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_Int32ClosedBothIntervalNode, (type(self), 0x7a104e5, None), state
 *     else:
 *         return __pyx_unpickle_Int32ClosedBothIntervalNode, (type(self), 0x7a104e5, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Int32ClosedBothIntervalNode__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_13, __pyx_n_s_pyx_unpickle_Int32ClosedBothIn); if (unlikely(!__pyx_t_13)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __pyx_t_15 = PyTuple_New(3); if (unlikely(!__pyx_t_15)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_15, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_127993061);
    __Pyx_GIVEREF(__pyx_int_127993061);
    PyTuple_SET_ITEM(__pyx_t_15, 1, __pyx_int_127993061);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_15, 2, __pyx_v_state);
    __pyx_t_14 = PyTuple_New(2); if (unlikely(!__pyx_t_14)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_GIVEREF(__pyx_t_13);
    PyTuple_SET_ITEM(__pyx_t_14, 0, __pyx_t_13);
    __Pyx_GIVEREF(__pyx_t_15);
    PyTuple_SET_ITEM(__pyx_t_14, 1, __pyx_t_15);
    __pyx_t_13 = 0;
    __pyx_t_15 = 0;
    __pyx_r = __pyx_t_14;
    __pyx_t_14 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_AddTraceback("pandas._libs.interval.Int32ClosedBothIntervalNode.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Int32ClosedBothIntervalNode, (type(self), 0x7a104e5, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Int32ClosedBothIntervalNode__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_23__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_23__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_22__setstate_cython__(((struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_22__setstate_cython__(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_Int32ClosedBothIntervalNode, (type(self), 0x7a104e5, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Int32ClosedBothIntervalNode__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(2, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_6pandas_5_libs_8interval___pyx_unpickle_Int32ClosedBothIntervalNode__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Int32ClosedBothIntervalNode, (type(self), 0x7a104e5, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Int32ClosedBothIntervalNode__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Int32ClosedBothIntervalNode.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":2118
 *         readonly bint is_leaf_node
 * 
 *     def __init__(self,             # <<<<<<<<<<<<<<
 *                  ndarray[int32_t, ndim=1] left,
 *                  ndarray[int32_t, ndim=1] right,
 */

/* Python wrapper */
static int __pyx_pw_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyArrayObject *__pyx_v_left = 0;
  PyArrayObject *__pyx_v_right = 0;
  PyArrayObject *__pyx_v_indices = 0;
  __pyx_t_5numpy_int64_t __pyx_v_leaf_size;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_left,&__pyx_n_s_right,&__pyx_n_s_indices,&__pyx_n_s_leaf_size,0};
    PyObject* values[4] = {0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_left)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_right)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 4, 4, 1); __PYX_ERR(1, 2118, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_indices)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 4, 4, 2); __PYX_ERR(1, 2118, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_leaf_size)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 4, 4, 3); __PYX_ERR(1, 2118, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(1, 2118, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 4) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
    }
    __pyx_v_left = ((PyArrayObject *)values[0]);
    __pyx_v_right = ((PyArrayObject *)values[1]);
    __pyx_v_indices = ((PyArrayObject *)values[2]);
    __pyx_v_leaf_size = __Pyx_PyInt_As_npy_int64(values[3]); if (unlikely((__pyx_v_leaf_size == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(1, 2122, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 4, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 2118, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Int32ClosedNeitherIntervalNode.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_left), __pyx_ptype_5numpy_ndarray, 1, "left", 0))) __PYX_ERR(1, 2119, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_right), __pyx_ptype_5numpy_ndarray, 1, "right", 0))) __PYX_ERR(1, 2120, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_indices), __pyx_ptype_5numpy_ndarray, 1, "indices", 0))) __PYX_ERR(1, 2121, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode___init__(((struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode *)__pyx_v_self), __pyx_v_left, __pyx_v_right, __pyx_v_indices, __pyx_v_leaf_size);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode___init__(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode *__pyx_v_self, PyArrayObject *__pyx_v_left, PyArrayObject *__pyx_v_right, PyArrayObject *__pyx_v_indices, __pyx_t_5numpy_int64_t __pyx_v_leaf_size) {
  PyObject *__pyx_v_left_set = NULL;
  PyObject *__pyx_v_right_set = NULL;
  PyObject *__pyx_v_center_set = NULL;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_indices;
  __Pyx_Buffer __pyx_pybuffer_indices;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_left;
  __Pyx_Buffer __pyx_pybuffer_left;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_right;
  __Pyx_Buffer __pyx_pybuffer_right;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __pyx_t_5numpy_int32_t __pyx_t_6;
  __Pyx_memviewslice __pyx_t_7 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_8 = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  __Pyx_memviewslice __pyx_t_11 = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *(*__pyx_t_12)(PyObject *);
  size_t __pyx_t_13;
  __Pyx_RefNannySetupContext("__init__", 0);
  __pyx_pybuffer_left.pybuffer.buf = NULL;
  __pyx_pybuffer_left.refcount = 0;
  __pyx_pybuffernd_left.data = NULL;
  __pyx_pybuffernd_left.rcbuffer = &__pyx_pybuffer_left;
  __pyx_pybuffer_right.pybuffer.buf = NULL;
  __pyx_pybuffer_right.refcount = 0;
  __pyx_pybuffernd_right.data = NULL;
  __pyx_pybuffernd_right.rcbuffer = &__pyx_pybuffer_right;
  __pyx_pybuffer_indices.pybuffer.buf = NULL;
  __pyx_pybuffer_indices.refcount = 0;
  __pyx_pybuffernd_indices.data = NULL;
  __pyx_pybuffernd_indices.rcbuffer = &__pyx_pybuffer_indices;
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_left.rcbuffer->pybuffer, (PyObject*)__pyx_v_left, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int32_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 2118, __pyx_L1_error)
  }
  __pyx_pybuffernd_left.diminfo[0].strides = __pyx_pybuffernd_left.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_left.diminfo[0].shape = __pyx_pybuffernd_left.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_right.rcbuffer->pybuffer, (PyObject*)__pyx_v_right, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int32_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 2118, __pyx_L1_error)
  }
  __pyx_pybuffernd_right.diminfo[0].strides = __pyx_pybuffernd_right.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_right.diminfo[0].shape = __pyx_pybuffernd_right.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_indices.rcbuffer->pybuffer, (PyObject*)__pyx_v_indices, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 2118, __pyx_L1_error)
  }
  __pyx_pybuffernd_indices.diminfo[0].strides = __pyx_pybuffernd_indices.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_indices.diminfo[0].shape = __pyx_pybuffernd_indices.rcbuffer->pybuffer.shape[0];

  /* "pandas/_libs/intervaltree.pxi":2124
 *                  int64_t leaf_size):
 * 
 *         self.n_elements = len(left)             # <<<<<<<<<<<<<<
 *         self.leaf_size = leaf_size
 * 
 */
  __pyx_t_1 = PyObject_Length(((PyObject *)__pyx_v_left)); if (unlikely(__pyx_t_1 == ((Py_ssize_t)-1))) __PYX_ERR(1, 2124, __pyx_L1_error)
  __pyx_v_self->n_elements = __pyx_t_1;

  /* "pandas/_libs/intervaltree.pxi":2125
 * 
 *         self.n_elements = len(left)
 *         self.leaf_size = leaf_size             # <<<<<<<<<<<<<<
 * 
 *         # min_left and min_right are used to speed-up query by skipping
 */
  __pyx_v_self->leaf_size = __pyx_v_leaf_size;

  /* "pandas/_libs/intervaltree.pxi":2130
 *         # query on sub-nodes. If this node has size 0, query is cheap,
 *         # so these values don't matter.
 *         if left.size > 0:             # <<<<<<<<<<<<<<
 *             self.min_left = left.min()
 *             self.max_right = right.max()
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_left), __pyx_n_s_size); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2130, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyObject_RichCompare(__pyx_t_2, __pyx_int_0, Py_GT); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2130, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(1, 2130, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_4) {

    /* "pandas/_libs/intervaltree.pxi":2131
 *         # so these values don't matter.
 *         if left.size > 0:
 *             self.min_left = left.min()             # <<<<<<<<<<<<<<
 *             self.max_right = right.max()
 *         else:
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_left), __pyx_n_s_min); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2131, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2131, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_6 = __Pyx_PyInt_As_npy_int32(__pyx_t_3); if (unlikely((__pyx_t_6 == ((npy_int32)-1)) && PyErr_Occurred())) __PYX_ERR(1, 2131, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_self->min_left = __pyx_t_6;

    /* "pandas/_libs/intervaltree.pxi":2132
 *         if left.size > 0:
 *             self.min_left = left.min()
 *             self.max_right = right.max()             # <<<<<<<<<<<<<<
 *         else:
 *             self.min_left = 0
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_right), __pyx_n_s_max); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2132, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2132, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_6 = __Pyx_PyInt_As_npy_int32(__pyx_t_3); if (unlikely((__pyx_t_6 == ((npy_int32)-1)) && PyErr_Occurred())) __PYX_ERR(1, 2132, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_self->max_right = __pyx_t_6;

    /* "pandas/_libs/intervaltree.pxi":2130
 *         # query on sub-nodes. If this node has size 0, query is cheap,
 *         # so these values don't matter.
 *         if left.size > 0:             # <<<<<<<<<<<<<<
 *             self.min_left = left.min()
 *             self.max_right = right.max()
 */
    goto __pyx_L3;
  }

  /* "pandas/_libs/intervaltree.pxi":2134
 *             self.max_right = right.max()
 *         else:
 *             self.min_left = 0             # <<<<<<<<<<<<<<
 *             self.max_right = 0
 * 
 */
  /*else*/ {
    __pyx_v_self->min_left = 0;

    /* "pandas/_libs/intervaltree.pxi":2135
 *         else:
 *             self.min_left = 0
 *             self.max_right = 0             # <<<<<<<<<<<<<<
 * 
 *         if self.n_elements <= leaf_size:
 */
    __pyx_v_self->max_right = 0;
  }
  __pyx_L3:;

  /* "pandas/_libs/intervaltree.pxi":2137
 *             self.max_right = 0
 * 
 *         if self.n_elements <= leaf_size:             # <<<<<<<<<<<<<<
 *             # make this a terminal (leaf) node
 *             self.is_leaf_node = True
 */
  __pyx_t_4 = ((__pyx_v_self->n_elements <= __pyx_v_leaf_size) != 0);
  if (__pyx_t_4) {

    /* "pandas/_libs/intervaltree.pxi":2139
 *         if self.n_elements <= leaf_size:
 *             # make this a terminal (leaf) node
 *             self.is_leaf_node = True             # <<<<<<<<<<<<<<
 *             self.left = left
 *             self.right = right
 */
    __pyx_v_self->is_leaf_node = 1;

    /* "pandas/_libs/intervaltree.pxi":2140
 *             # make this a terminal (leaf) node
 *             self.is_leaf_node = True
 *             self.left = left             # <<<<<<<<<<<<<<
 *             self.right = right
 *             self.indices = indices
 */
    __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int32_t(((PyObject *)__pyx_v_left), PyBUF_WRITABLE); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(1, 2140, __pyx_L1_error)
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->left, 0);
    __pyx_v_self->left = __pyx_t_7;
    __pyx_t_7.memview = NULL;
    __pyx_t_7.data = NULL;

    /* "pandas/_libs/intervaltree.pxi":2141
 *             self.is_leaf_node = True
 *             self.left = left
 *             self.right = right             # <<<<<<<<<<<<<<
 *             self.indices = indices
 *             self.n_center = 0
 */
    __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int32_t(((PyObject *)__pyx_v_right), PyBUF_WRITABLE); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(1, 2141, __pyx_L1_error)
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->right, 0);
    __pyx_v_self->right = __pyx_t_7;
    __pyx_t_7.memview = NULL;
    __pyx_t_7.data = NULL;

    /* "pandas/_libs/intervaltree.pxi":2142
 *             self.left = left
 *             self.right = right
 *             self.indices = indices             # <<<<<<<<<<<<<<
 *             self.n_center = 0
 *         else:
 */
    __pyx_t_8 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(((PyObject *)__pyx_v_indices), PyBUF_WRITABLE); if (unlikely(!__pyx_t_8.memview)) __PYX_ERR(1, 2142, __pyx_L1_error)
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->indices, 0);
    __pyx_v_self->indices = __pyx_t_8;
    __pyx_t_8.memview = NULL;
    __pyx_t_8.data = NULL;

    /* "pandas/_libs/intervaltree.pxi":2143
 *             self.right = right
 *             self.indices = indices
 *             self.n_center = 0             # <<<<<<<<<<<<<<
 *         else:
 *             # calculate a pivot so we can create child nodes
 */
    __pyx_v_self->n_center = 0;

    /* "pandas/_libs/intervaltree.pxi":2137
 *             self.max_right = 0
 * 
 *         if self.n_elements <= leaf_size:             # <<<<<<<<<<<<<<
 *             # make this a terminal (leaf) node
 *             self.is_leaf_node = True
 */
    goto __pyx_L4;
  }

  /* "pandas/_libs/intervaltree.pxi":2146
 *         else:
 *             # calculate a pivot so we can create child nodes
 *             self.is_leaf_node = False             # <<<<<<<<<<<<<<
 *             self.pivot = np.median(left / 2 + right / 2)
 *             left_set, right_set, center_set = self.classify_intervals(
 */
  /*else*/ {
    __pyx_v_self->is_leaf_node = 0;

    /* "pandas/_libs/intervaltree.pxi":2147
 *             # calculate a pivot so we can create child nodes
 *             self.is_leaf_node = False
 *             self.pivot = np.median(left / 2 + right / 2)             # <<<<<<<<<<<<<<
 *             left_set, right_set, center_set = self.classify_intervals(
 *                 left, right)
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2147, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_median); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 2147, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyNumber_Divide(((PyObject *)__pyx_v_left), __pyx_int_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2147, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_9 = __Pyx_PyNumber_Divide(((PyObject *)__pyx_v_right), __pyx_int_2); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 2147, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_10 = PyNumber_Add(__pyx_t_2, __pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 2147, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_9 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_3 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_9, __pyx_t_10) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_10);
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2147, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_6 = __Pyx_PyInt_As_npy_int32(__pyx_t_3); if (unlikely((__pyx_t_6 == ((npy_int32)-1)) && PyErr_Occurred())) __PYX_ERR(1, 2147, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_self->pivot = __pyx_t_6;

    /* "pandas/_libs/intervaltree.pxi":2149
 *             self.pivot = np.median(left / 2 + right / 2)
 *             left_set, right_set, center_set = self.classify_intervals(
 *                 left, right)             # <<<<<<<<<<<<<<
 * 
 *             self.left_node = self.new_child_node(left, right,
 */
    __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int32_t(((PyObject *)__pyx_v_left), PyBUF_WRITABLE); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(1, 2149, __pyx_L1_error)
    __pyx_t_11 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int32_t(((PyObject *)__pyx_v_right), PyBUF_WRITABLE); if (unlikely(!__pyx_t_11.memview)) __PYX_ERR(1, 2149, __pyx_L1_error)

    /* "pandas/_libs/intervaltree.pxi":2148
 *             self.is_leaf_node = False
 *             self.pivot = np.median(left / 2 + right / 2)
 *             left_set, right_set, center_set = self.classify_intervals(             # <<<<<<<<<<<<<<
 *                 left, right)
 * 
 */
    __pyx_t_3 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode *)__pyx_v_self->__pyx_vtab)->classify_intervals(__pyx_v_self, __pyx_t_7, __pyx_t_11); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2148, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __PYX_XDEC_MEMVIEW(&__pyx_t_7, 1);
    __pyx_t_7.memview = NULL;
    __pyx_t_7.data = NULL;
    __PYX_XDEC_MEMVIEW(&__pyx_t_11, 1);
    __pyx_t_11.memview = NULL;
    __pyx_t_11.data = NULL;
    if ((likely(PyTuple_CheckExact(__pyx_t_3))) || (PyList_CheckExact(__pyx_t_3))) {
      PyObject* sequence = __pyx_t_3;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 3)) {
        if (size > 3) __Pyx_RaiseTooManyValuesError(3);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(1, 2148, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_5 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_10 = PyTuple_GET_ITEM(sequence, 1); 
        __pyx_t_9 = PyTuple_GET_ITEM(sequence, 2); 
      } else {
        __pyx_t_5 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_10 = PyList_GET_ITEM(sequence, 1); 
        __pyx_t_9 = PyList_GET_ITEM(sequence, 2); 
      }
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_10);
      __Pyx_INCREF(__pyx_t_9);
      #else
      __pyx_t_5 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 2148, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_10 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 2148, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_9 = PySequence_ITEM(sequence, 2); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 2148, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      #endif
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_2 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2148, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_12 = Py_TYPE(__pyx_t_2)->tp_iternext;
      index = 0; __pyx_t_5 = __pyx_t_12(__pyx_t_2); if (unlikely(!__pyx_t_5)) goto __pyx_L5_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_5);
      index = 1; __pyx_t_10 = __pyx_t_12(__pyx_t_2); if (unlikely(!__pyx_t_10)) goto __pyx_L5_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_10);
      index = 2; __pyx_t_9 = __pyx_t_12(__pyx_t_2); if (unlikely(!__pyx_t_9)) goto __pyx_L5_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_9);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_12(__pyx_t_2), 3) < 0) __PYX_ERR(1, 2148, __pyx_L1_error)
      __pyx_t_12 = NULL;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      goto __pyx_L6_unpacking_done;
      __pyx_L5_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_12 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(1, 2148, __pyx_L1_error)
      __pyx_L6_unpacking_done:;
    }
    __pyx_v_left_set = __pyx_t_5;
    __pyx_t_5 = 0;
    __pyx_v_right_set = __pyx_t_10;
    __pyx_t_10 = 0;
    __pyx_v_center_set = __pyx_t_9;
    __pyx_t_9 = 0;

    /* "pandas/_libs/intervaltree.pxi":2152
 * 
 *             self.left_node = self.new_child_node(left, right,
 *                                                  indices, left_set)             # <<<<<<<<<<<<<<
 *             self.right_node = self.new_child_node(left, right,
 *                                                   indices, right_set)
 */
    if (!(likely(((__pyx_v_left_set) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_left_set, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(1, 2152, __pyx_L1_error)

    /* "pandas/_libs/intervaltree.pxi":2151
 *                 left, right)
 * 
 *             self.left_node = self.new_child_node(left, right,             # <<<<<<<<<<<<<<
 *                                                  indices, left_set)
 *             self.right_node = self.new_child_node(left, right,
 */
    __pyx_t_3 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode *)__pyx_v_self->__pyx_vtab)->new_child_node(__pyx_v_self, ((PyArrayObject *)__pyx_v_left), ((PyArrayObject *)__pyx_v_right), ((PyArrayObject *)__pyx_v_indices), ((PyArrayObject *)__pyx_v_left_set)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2151, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode))))) __PYX_ERR(1, 2151, __pyx_L1_error)
    __Pyx_GIVEREF(__pyx_t_3);
    __Pyx_GOTREF(__pyx_v_self->left_node);
    __Pyx_DECREF(((PyObject *)__pyx_v_self->left_node));
    __pyx_v_self->left_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode *)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "pandas/_libs/intervaltree.pxi":2154
 *                                                  indices, left_set)
 *             self.right_node = self.new_child_node(left, right,
 *                                                   indices, right_set)             # <<<<<<<<<<<<<<
 * 
 *             self.center_left_values, self.center_left_indices = \
 */
    if (!(likely(((__pyx_v_right_set) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_right_set, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(1, 2154, __pyx_L1_error)

    /* "pandas/_libs/intervaltree.pxi":2153
 *             self.left_node = self.new_child_node(left, right,
 *                                                  indices, left_set)
 *             self.right_node = self.new_child_node(left, right,             # <<<<<<<<<<<<<<
 *                                                   indices, right_set)
 * 
 */
    __pyx_t_3 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode *)__pyx_v_self->__pyx_vtab)->new_child_node(__pyx_v_self, ((PyArrayObject *)__pyx_v_left), ((PyArrayObject *)__pyx_v_right), ((PyArrayObject *)__pyx_v_indices), ((PyArrayObject *)__pyx_v_right_set)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2153, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode))))) __PYX_ERR(1, 2153, __pyx_L1_error)
    __Pyx_GIVEREF(__pyx_t_3);
    __Pyx_GOTREF(__pyx_v_self->right_node);
    __Pyx_DECREF(((PyObject *)__pyx_v_self->right_node));
    __pyx_v_self->right_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode *)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "pandas/_libs/intervaltree.pxi":2157
 * 
 *             self.center_left_values, self.center_left_indices = \
 *                 sort_values_and_indices(left, indices, center_set)             # <<<<<<<<<<<<<<
 *             self.center_right_values, self.center_right_indices = \
 *                 sort_values_and_indices(right, indices, center_set)
 */
    __pyx_t_3 = __pyx_f_6pandas_5_libs_8interval_sort_values_and_indices(((PyObject *)__pyx_v_left), ((PyObject *)__pyx_v_indices), __pyx_v_center_set); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2157, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if ((likely(PyTuple_CheckExact(__pyx_t_3))) || (PyList_CheckExact(__pyx_t_3))) {
      PyObject* sequence = __pyx_t_3;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(1, 2156, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_9 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_10 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_9 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_10 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_9);
      __Pyx_INCREF(__pyx_t_10);
      #else
      __pyx_t_9 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 2156, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_10 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 2156, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      #endif
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_5 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 2156, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_12 = Py_TYPE(__pyx_t_5)->tp_iternext;
      index = 0; __pyx_t_9 = __pyx_t_12(__pyx_t_5); if (unlikely(!__pyx_t_9)) goto __pyx_L7_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_9);
      index = 1; __pyx_t_10 = __pyx_t_12(__pyx_t_5); if (unlikely(!__pyx_t_10)) goto __pyx_L7_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_10);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_12(__pyx_t_5), 2) < 0) __PYX_ERR(1, 2156, __pyx_L1_error)
      __pyx_t_12 = NULL;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      goto __pyx_L8_unpacking_done;
      __pyx_L7_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_12 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(1, 2156, __pyx_L1_error)
      __pyx_L8_unpacking_done:;
    }

    /* "pandas/_libs/intervaltree.pxi":2156
 *                                                   indices, right_set)
 * 
 *             self.center_left_values, self.center_left_indices = \             # <<<<<<<<<<<<<<
 *                 sort_values_and_indices(left, indices, center_set)
 *             self.center_right_values, self.center_right_indices = \
 */
    __pyx_t_11 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int32_t(__pyx_t_9, PyBUF_WRITABLE); if (unlikely(!__pyx_t_11.memview)) __PYX_ERR(1, 2156, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_8 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_t_10, PyBUF_WRITABLE); if (unlikely(!__pyx_t_8.memview)) __PYX_ERR(1, 2156, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->center_left_values, 0);
    __pyx_v_self->center_left_values = __pyx_t_11;
    __pyx_t_11.memview = NULL;
    __pyx_t_11.data = NULL;
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->center_left_indices, 0);
    __pyx_v_self->center_left_indices = __pyx_t_8;
    __pyx_t_8.memview = NULL;
    __pyx_t_8.data = NULL;

    /* "pandas/_libs/intervaltree.pxi":2159
 *                 sort_values_and_indices(left, indices, center_set)
 *             self.center_right_values, self.center_right_indices = \
 *                 sort_values_and_indices(right, indices, center_set)             # <<<<<<<<<<<<<<
 *             self.n_center = len(self.center_left_indices)
 * 
 */
    __pyx_t_3 = __pyx_f_6pandas_5_libs_8interval_sort_values_and_indices(((PyObject *)__pyx_v_right), ((PyObject *)__pyx_v_indices), __pyx_v_center_set); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2159, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if ((likely(PyTuple_CheckExact(__pyx_t_3))) || (PyList_CheckExact(__pyx_t_3))) {
      PyObject* sequence = __pyx_t_3;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(1, 2158, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_10 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_9 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_10 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_9 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_10);
      __Pyx_INCREF(__pyx_t_9);
      #else
      __pyx_t_10 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 2158, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_9 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 2158, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      #endif
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_5 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 2158, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_12 = Py_TYPE(__pyx_t_5)->tp_iternext;
      index = 0; __pyx_t_10 = __pyx_t_12(__pyx_t_5); if (unlikely(!__pyx_t_10)) goto __pyx_L9_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_10);
      index = 1; __pyx_t_9 = __pyx_t_12(__pyx_t_5); if (unlikely(!__pyx_t_9)) goto __pyx_L9_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_9);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_12(__pyx_t_5), 2) < 0) __PYX_ERR(1, 2158, __pyx_L1_error)
      __pyx_t_12 = NULL;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      goto __pyx_L10_unpacking_done;
      __pyx_L9_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_12 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(1, 2158, __pyx_L1_error)
      __pyx_L10_unpacking_done:;
    }

    /* "pandas/_libs/intervaltree.pxi":2158
 *             self.center_left_values, self.center_left_indices = \
 *                 sort_values_and_indices(left, indices, center_set)
 *             self.center_right_values, self.center_right_indices = \             # <<<<<<<<<<<<<<
 *                 sort_values_and_indices(right, indices, center_set)
 *             self.n_center = len(self.center_left_indices)
 */
    __pyx_t_11 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int32_t(__pyx_t_10, PyBUF_WRITABLE); if (unlikely(!__pyx_t_11.memview)) __PYX_ERR(1, 2158, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_t_8 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_t_9, PyBUF_WRITABLE); if (unlikely(!__pyx_t_8.memview)) __PYX_ERR(1, 2158, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->center_right_values, 0);
    __pyx_v_self->center_right_values = __pyx_t_11;
    __pyx_t_11.memview = NULL;
    __pyx_t_11.data = NULL;
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->center_right_indices, 0);
    __pyx_v_self->center_right_indices = __pyx_t_8;
    __pyx_t_8.memview = NULL;
    __pyx_t_8.data = NULL;

    /* "pandas/_libs/intervaltree.pxi":2160
 *             self.center_right_values, self.center_right_indices = \
 *                 sort_values_and_indices(right, indices, center_set)
 *             self.n_center = len(self.center_left_indices)             # <<<<<<<<<<<<<<
 * 
 *     @cython.wraparound(False)
 */
    if (unlikely(!__pyx_v_self->center_left_indices.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(1, 2160, __pyx_L1_error)}
    __pyx_t_13 = __Pyx_MemoryView_Len(__pyx_v_self->center_left_indices); 
    __pyx_v_self->n_center = __pyx_t_13;
  }
  __pyx_L4:;

  /* "pandas/_libs/intervaltree.pxi":2118
 *         readonly bint is_leaf_node
 * 
 *     def __init__(self,             # <<<<<<<<<<<<<<
 *                  ndarray[int32_t, ndim=1] left,
 *                  ndarray[int32_t, ndim=1] right,
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __PYX_XDEC_MEMVIEW(&__pyx_t_7, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_8, 1);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __PYX_XDEC_MEMVIEW(&__pyx_t_11, 1);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indices.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_left.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_right.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("pandas._libs.interval.Int32ClosedNeitherIntervalNode.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indices.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_left.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_right.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF(__pyx_v_left_set);
  __Pyx_XDECREF(__pyx_v_right_set);
  __Pyx_XDECREF(__pyx_v_center_set);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":2164
 *     @cython.wraparound(False)
 *     @cython.boundscheck(False)
 *     cdef classify_intervals(self, int32_t[:] left, int32_t[:] right):             # <<<<<<<<<<<<<<
 *         """Classify the given intervals based upon whether they fall to the
 *         left, right, or overlap with this node's pivot.
 */

static PyObject *__pyx_f_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_classify_intervals(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode *__pyx_v_self, __Pyx_memviewslice __pyx_v_left, __Pyx_memviewslice __pyx_v_right) {
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_left_ind = 0;
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_right_ind = 0;
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_overlapping_ind = 0;
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __pyx_t_5numpy_int64_t __pyx_t_2;
  __pyx_t_5numpy_int64_t __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  int __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  __Pyx_RefNannySetupContext("classify_intervals", 0);

  /* "pandas/_libs/intervaltree.pxi":2172
 *             Py_ssize_t i
 * 
 *         left_ind = Int64Vector()             # <<<<<<<<<<<<<<
 *         right_ind = Int64Vector()
 *         overlapping_ind = Int64Vector()
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2172, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_left_ind = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":2173
 * 
 *         left_ind = Int64Vector()
 *         right_ind = Int64Vector()             # <<<<<<<<<<<<<<
 *         overlapping_ind = Int64Vector()
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2173, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_right_ind = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":2174
 *         left_ind = Int64Vector()
 *         right_ind = Int64Vector()
 *         overlapping_ind = Int64Vector()             # <<<<<<<<<<<<<<
 * 
 *         for i in range(self.n_elements):
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2174, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_overlapping_ind = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":2176
 *         overlapping_ind = Int64Vector()
 * 
 *         for i in range(self.n_elements):             # <<<<<<<<<<<<<<
 *             if right[i] <= self.pivot:
 *                 left_ind.append(i)
 */
  __pyx_t_2 = __pyx_v_self->n_elements;
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "pandas/_libs/intervaltree.pxi":2177
 * 
 *         for i in range(self.n_elements):
 *             if right[i] <= self.pivot:             # <<<<<<<<<<<<<<
 *                 left_ind.append(i)
 *             elif self.pivot <= left[i]:
 */
    __pyx_t_5 = __pyx_v_i;
    __pyx_t_6 = (((*((__pyx_t_5numpy_int32_t *) ( /* dim=0 */ (__pyx_v_right.data + __pyx_t_5 * __pyx_v_right.strides[0]) ))) <= __pyx_v_self->pivot) != 0);
    if (__pyx_t_6) {

      /* "pandas/_libs/intervaltree.pxi":2178
 *         for i in range(self.n_elements):
 *             if right[i] <= self.pivot:
 *                 left_ind.append(i)             # <<<<<<<<<<<<<<
 *             elif self.pivot <= left[i]:
 *                 right_ind.append(i)
 */
      ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_left_ind->__pyx_vtab)->append(__pyx_v_left_ind, __pyx_v_i);

      /* "pandas/_libs/intervaltree.pxi":2177
 * 
 *         for i in range(self.n_elements):
 *             if right[i] <= self.pivot:             # <<<<<<<<<<<<<<
 *                 left_ind.append(i)
 *             elif self.pivot <= left[i]:
 */
      goto __pyx_L5;
    }

    /* "pandas/_libs/intervaltree.pxi":2179
 *             if right[i] <= self.pivot:
 *                 left_ind.append(i)
 *             elif self.pivot <= left[i]:             # <<<<<<<<<<<<<<
 *                 right_ind.append(i)
 *             else:
 */
    __pyx_t_7 = __pyx_v_i;
    __pyx_t_6 = ((__pyx_v_self->pivot <= (*((__pyx_t_5numpy_int32_t *) ( /* dim=0 */ (__pyx_v_left.data + __pyx_t_7 * __pyx_v_left.strides[0]) )))) != 0);
    if (__pyx_t_6) {

      /* "pandas/_libs/intervaltree.pxi":2180
 *                 left_ind.append(i)
 *             elif self.pivot <= left[i]:
 *                 right_ind.append(i)             # <<<<<<<<<<<<<<
 *             else:
 *                 overlapping_ind.append(i)
 */
      ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_right_ind->__pyx_vtab)->append(__pyx_v_right_ind, __pyx_v_i);

      /* "pandas/_libs/intervaltree.pxi":2179
 *             if right[i] <= self.pivot:
 *                 left_ind.append(i)
 *             elif self.pivot <= left[i]:             # <<<<<<<<<<<<<<
 *                 right_ind.append(i)
 *             else:
 */
      goto __pyx_L5;
    }

    /* "pandas/_libs/intervaltree.pxi":2182
 *                 right_ind.append(i)
 *             else:
 *                 overlapping_ind.append(i)             # <<<<<<<<<<<<<<
 * 
 *         return (left_ind.to_array(),
 */
    /*else*/ {
      ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_overlapping_ind->__pyx_vtab)->append(__pyx_v_overlapping_ind, __pyx_v_i);
    }
    __pyx_L5:;
  }

  /* "pandas/_libs/intervaltree.pxi":2184
 *                 overlapping_ind.append(i)
 * 
 *         return (left_ind.to_array(),             # <<<<<<<<<<<<<<
 *                 right_ind.to_array(),
 *                 overlapping_ind.to_array())
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_left_ind->__pyx_vtab)->to_array(__pyx_v_left_ind, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2184, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "pandas/_libs/intervaltree.pxi":2185
 * 
 *         return (left_ind.to_array(),
 *                 right_ind.to_array(),             # <<<<<<<<<<<<<<
 *                 overlapping_ind.to_array())
 * 
 */
  __pyx_t_8 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_right_ind->__pyx_vtab)->to_array(__pyx_v_right_ind, 0); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 2185, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);

  /* "pandas/_libs/intervaltree.pxi":2186
 *         return (left_ind.to_array(),
 *                 right_ind.to_array(),
 *                 overlapping_ind.to_array())             # <<<<<<<<<<<<<<
 * 
 *     cdef new_child_node(self,
 */
  __pyx_t_9 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_overlapping_ind->__pyx_vtab)->to_array(__pyx_v_overlapping_ind, 0); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 2186, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);

  /* "pandas/_libs/intervaltree.pxi":2184
 *                 overlapping_ind.append(i)
 * 
 *         return (left_ind.to_array(),             # <<<<<<<<<<<<<<
 *                 right_ind.to_array(),
 *                 overlapping_ind.to_array())
 */
  __pyx_t_10 = PyTuple_New(3); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 2184, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_t_8);
  __Pyx_GIVEREF(__pyx_t_9);
  PyTuple_SET_ITEM(__pyx_t_10, 2, __pyx_t_9);
  __pyx_t_1 = 0;
  __pyx_t_8 = 0;
  __pyx_t_9 = 0;
  __pyx_r = __pyx_t_10;
  __pyx_t_10 = 0;
  goto __pyx_L0;

  /* "pandas/_libs/intervaltree.pxi":2164
 *     @cython.wraparound(False)
 *     @cython.boundscheck(False)
 *     cdef classify_intervals(self, int32_t[:] left, int32_t[:] right):             # <<<<<<<<<<<<<<
 *         """Classify the given intervals based upon whether they fall to the
 *         left, right, or overlap with this node's pivot.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("pandas._libs.interval.Int32ClosedNeitherIntervalNode.classify_intervals", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_left_ind);
  __Pyx_XDECREF((PyObject *)__pyx_v_right_ind);
  __Pyx_XDECREF((PyObject *)__pyx_v_overlapping_ind);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":2188
 *                 overlapping_ind.to_array())
 * 
 *     cdef new_child_node(self,             # <<<<<<<<<<<<<<
 *                         ndarray[int32_t, ndim=1] left,
 *                         ndarray[int32_t, ndim=1] right,
 */

static PyObject *__pyx_f_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_new_child_node(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode *__pyx_v_self, PyArrayObject *__pyx_v_left, PyArrayObject *__pyx_v_right, PyArrayObject *__pyx_v_indices, PyArrayObject *__pyx_v_subset) {
  __Pyx_LocalBuf_ND __pyx_pybuffernd_indices;
  __Pyx_Buffer __pyx_pybuffer_indices;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_left;
  __Pyx_Buffer __pyx_pybuffer_left;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_right;
  __Pyx_Buffer __pyx_pybuffer_right;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_subset;
  __Pyx_Buffer __pyx_pybuffer_subset;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyArrayObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyArrayObject *__pyx_t_7 = NULL;
  PyArrayObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  __Pyx_RefNannySetupContext("new_child_node", 0);
  __Pyx_INCREF((PyObject *)__pyx_v_left);
  __Pyx_INCREF((PyObject *)__pyx_v_right);
  __Pyx_INCREF((PyObject *)__pyx_v_indices);
  __pyx_pybuffer_left.pybuffer.buf = NULL;
  __pyx_pybuffer_left.refcount = 0;
  __pyx_pybuffernd_left.data = NULL;
  __pyx_pybuffernd_left.rcbuffer = &__pyx_pybuffer_left;
  __pyx_pybuffer_right.pybuffer.buf = NULL;
  __pyx_pybuffer_right.refcount = 0;
  __pyx_pybuffernd_right.data = NULL;
  __pyx_pybuffernd_right.rcbuffer = &__pyx_pybuffer_right;
  __pyx_pybuffer_indices.pybuffer.buf = NULL;
  __pyx_pybuffer_indices.refcount = 0;
  __pyx_pybuffernd_indices.data = NULL;
  __pyx_pybuffernd_indices.rcbuffer = &__pyx_pybuffer_indices;
  __pyx_pybuffer_subset.pybuffer.buf = NULL;
  __pyx_pybuffer_subset.refcount = 0;
  __pyx_pybuffernd_subset.data = NULL;
  __pyx_pybuffernd_subset.rcbuffer = &__pyx_pybuffer_subset;
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_left.rcbuffer->pybuffer, (PyObject*)__pyx_v_left, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int32_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 2188, __pyx_L1_error)
  }
  __pyx_pybuffernd_left.diminfo[0].strides = __pyx_pybuffernd_left.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_left.diminfo[0].shape = __pyx_pybuffernd_left.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_right.rcbuffer->pybuffer, (PyObject*)__pyx_v_right, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int32_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 2188, __pyx_L1_error)
  }
  __pyx_pybuffernd_right.diminfo[0].strides = __pyx_pybuffernd_right.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_right.diminfo[0].shape = __pyx_pybuffernd_right.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_indices.rcbuffer->pybuffer, (PyObject*)__pyx_v_indices, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 2188, __pyx_L1_error)
  }
  __pyx_pybuffernd_indices.diminfo[0].strides = __pyx_pybuffernd_indices.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_indices.diminfo[0].shape = __pyx_pybuffernd_indices.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_subset.rcbuffer->pybuffer, (PyObject*)__pyx_v_subset, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 2188, __pyx_L1_error)
  }
  __pyx_pybuffernd_subset.diminfo[0].strides = __pyx_pybuffernd_subset.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_subset.diminfo[0].shape = __pyx_pybuffernd_subset.rcbuffer->pybuffer.shape[0];

  /* "pandas/_libs/intervaltree.pxi":2195
 *         """Create a new child node.
 *         """
 *         left = take(left, subset)             # <<<<<<<<<<<<<<
 *         right = take(right, subset)
 *         indices = take(indices, subset)
 */
  __pyx_t_1 = __pyx_f_6pandas_5_libs_8interval_take(((PyArrayObject *)__pyx_v_left), ((PyArrayObject *)__pyx_v_subset)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2195, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(1, 2195, __pyx_L1_error)
  __pyx_t_2 = ((PyArrayObject *)__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_left.rcbuffer->pybuffer);
    __pyx_t_3 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_left.rcbuffer->pybuffer, (PyObject*)__pyx_t_2, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int32_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack);
    if (unlikely(__pyx_t_3 < 0)) {
      PyErr_Fetch(&__pyx_t_4, &__pyx_t_5, &__pyx_t_6);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_left.rcbuffer->pybuffer, (PyObject*)__pyx_v_left, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int32_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_4); Py_XDECREF(__pyx_t_5); Py_XDECREF(__pyx_t_6);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_4, __pyx_t_5, __pyx_t_6);
      }
      __pyx_t_4 = __pyx_t_5 = __pyx_t_6 = 0;
    }
    __pyx_pybuffernd_left.diminfo[0].strides = __pyx_pybuffernd_left.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_left.diminfo[0].shape = __pyx_pybuffernd_left.rcbuffer->pybuffer.shape[0];
    if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(1, 2195, __pyx_L1_error)
  }
  __pyx_t_2 = 0;
  __Pyx_DECREF_SET(__pyx_v_left, ((PyArrayObject *)__pyx_t_1));
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":2196
 *         """
 *         left = take(left, subset)
 *         right = take(right, subset)             # <<<<<<<<<<<<<<
 *         indices = take(indices, subset)
 *         return Int32ClosedNeitherIntervalNode(
 */
  __pyx_t_1 = __pyx_f_6pandas_5_libs_8interval_take(((PyArrayObject *)__pyx_v_right), ((PyArrayObject *)__pyx_v_subset)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2196, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(1, 2196, __pyx_L1_error)
  __pyx_t_7 = ((PyArrayObject *)__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_right.rcbuffer->pybuffer);
    __pyx_t_3 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_right.rcbuffer->pybuffer, (PyObject*)__pyx_t_7, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int32_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack);
    if (unlikely(__pyx_t_3 < 0)) {
      PyErr_Fetch(&__pyx_t_6, &__pyx_t_5, &__pyx_t_4);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_right.rcbuffer->pybuffer, (PyObject*)__pyx_v_right, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int32_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_6); Py_XDECREF(__pyx_t_5); Py_XDECREF(__pyx_t_4);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_6, __pyx_t_5, __pyx_t_4);
      }
      __pyx_t_6 = __pyx_t_5 = __pyx_t_4 = 0;
    }
    __pyx_pybuffernd_right.diminfo[0].strides = __pyx_pybuffernd_right.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_right.diminfo[0].shape = __pyx_pybuffernd_right.rcbuffer->pybuffer.shape[0];
    if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(1, 2196, __pyx_L1_error)
  }
  __pyx_t_7 = 0;
  __Pyx_DECREF_SET(__pyx_v_right, ((PyArrayObject *)__pyx_t_1));
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":2197
 *         left = take(left, subset)
 *         right = take(right, subset)
 *         indices = take(indices, subset)             # <<<<<<<<<<<<<<
 *         return Int32ClosedNeitherIntervalNode(
 *             left, right, indices, self.leaf_size)
 */
  __pyx_t_1 = __pyx_f_6pandas_5_libs_8interval_take(((PyArrayObject *)__pyx_v_indices), ((PyArrayObject *)__pyx_v_subset)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2197, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(1, 2197, __pyx_L1_error)
  __pyx_t_8 = ((PyArrayObject *)__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indices.rcbuffer->pybuffer);
    __pyx_t_3 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_indices.rcbuffer->pybuffer, (PyObject*)__pyx_t_8, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack);
    if (unlikely(__pyx_t_3 < 0)) {
      PyErr_Fetch(&__pyx_t_4, &__pyx_t_5, &__pyx_t_6);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_indices.rcbuffer->pybuffer, (PyObject*)__pyx_v_indices, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_4); Py_XDECREF(__pyx_t_5); Py_XDECREF(__pyx_t_6);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_4, __pyx_t_5, __pyx_t_6);
      }
      __pyx_t_4 = __pyx_t_5 = __pyx_t_6 = 0;
    }
    __pyx_pybuffernd_indices.diminfo[0].strides = __pyx_pybuffernd_indices.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_indices.diminfo[0].shape = __pyx_pybuffernd_indices.rcbuffer->pybuffer.shape[0];
    if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(1, 2197, __pyx_L1_error)
  }
  __pyx_t_8 = 0;
  __Pyx_DECREF_SET(__pyx_v_indices, ((PyArrayObject *)__pyx_t_1));
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":2198
 *         right = take(right, subset)
 *         indices = take(indices, subset)
 *         return Int32ClosedNeitherIntervalNode(             # <<<<<<<<<<<<<<
 *             left, right, indices, self.leaf_size)
 * 
 */
  __Pyx_XDECREF(__pyx_r);

  /* "pandas/_libs/intervaltree.pxi":2199
 *         indices = take(indices, subset)
 *         return Int32ClosedNeitherIntervalNode(
 *             left, right, indices, self.leaf_size)             # <<<<<<<<<<<<<<
 * 
 *     @cython.wraparound(False)
 */
  __pyx_t_1 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->leaf_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2199, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "pandas/_libs/intervaltree.pxi":2198
 *         right = take(right, subset)
 *         indices = take(indices, subset)
 *         return Int32ClosedNeitherIntervalNode(             # <<<<<<<<<<<<<<
 *             left, right, indices, self.leaf_size)
 * 
 */
  __pyx_t_9 = PyTuple_New(4); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 2198, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_INCREF(((PyObject *)__pyx_v_left));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_left));
  PyTuple_SET_ITEM(__pyx_t_9, 0, ((PyObject *)__pyx_v_left));
  __Pyx_INCREF(((PyObject *)__pyx_v_right));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_right));
  PyTuple_SET_ITEM(__pyx_t_9, 1, ((PyObject *)__pyx_v_right));
  __Pyx_INCREF(((PyObject *)__pyx_v_indices));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_indices));
  PyTuple_SET_ITEM(__pyx_t_9, 2, ((PyObject *)__pyx_v_indices));
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_9, 3, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode), __pyx_t_9, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2198, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pandas/_libs/intervaltree.pxi":2188
 *                 overlapping_ind.to_array())
 * 
 *     cdef new_child_node(self,             # <<<<<<<<<<<<<<
 *                         ndarray[int32_t, ndim=1] left,
 *                         ndarray[int32_t, ndim=1] right,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_9);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indices.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_left.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_right.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_subset.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("pandas._libs.interval.Int32ClosedNeitherIntervalNode.new_child_node", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indices.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_left.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_right.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_subset.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF((PyObject *)__pyx_v_left);
  __Pyx_XDECREF((PyObject *)__pyx_v_right);
  __Pyx_XDECREF((PyObject *)__pyx_v_indices);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":2204
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_2query[] = "Recursively query this node and its sub-nodes for intervals that\n        overlap with the query point.\n        ";
static PyMethodDef __pyx_mdef_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_3query = {"query", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_3query, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_2query};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_args = 0;
  PyObject *__pyx_v_kwargs = 0;
  CYTHON_UNUSED PyObject *__pyx_v_defaults = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fused_cpdef (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_args,&__pyx_n_s_kwargs,&__pyx_n_s_defaults,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_args)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_kwargs)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fused_cpdef", 1, 3, 3, 1); __PYX_ERR(1, 2204, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_defaults)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fused_cpdef", 1, 3, 3, 2); __PYX_ERR(1, 2204, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fused_cpdef") < 0)) __PYX_ERR(1, 2204, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_args = values[0];
    __pyx_v_kwargs = values[1];
    __pyx_v_defaults = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fused_cpdef", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 2204, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Int32ClosedNeitherIntervalNode.__pyx_fused_cpdef", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_2query(((struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode *)__pyx_v_signatures), __pyx_v_args, __pyx_v_kwargs, __pyx_v_defaults);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_2query(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode *__pyx_v_signatures, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs, CYTHON_UNUSED PyObject *__pyx_v_defaults) {
  PyObject *__pyx_v_dest_sig = NULL;
  Py_ssize_t __pyx_v_i;
  CYTHON_UNUSED PyTypeObject *__pyx_v_ndarray = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_v_itemsize;
  PyObject *__pyx_v_arg = NULL;
  PyObject *__pyx_v_candidates = NULL;
  PyObject *__pyx_v_sig = NULL;
  int __pyx_v_match_found;
  PyObject *__pyx_v_src_sig = NULL;
  PyObject *__pyx_v_dst_type = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  Py_ssize_t __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  Py_ssize_t __pyx_t_13;
  Py_ssize_t __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  int __pyx_t_16;
  __Pyx_RefNannySetupContext("query", 0);
  __Pyx_INCREF(__pyx_v_kwargs);
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2204, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  PyList_SET_ITEM(__pyx_t_1, 0, Py_None);
  __pyx_v_dest_sig = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_3 = (__pyx_v_kwargs != Py_None);
  __pyx_t_4 = (__pyx_t_3 != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_kwargs); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(1, 2204, __pyx_L1_error)
  __pyx_t_3 = ((!__pyx_t_4) != 0);
  __pyx_t_2 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {
    __Pyx_INCREF(Py_None);
    __Pyx_DECREF_SET(__pyx_v_kwargs, Py_None);
  }
  __pyx_t_1 = ((PyObject *)__Pyx_ImportNumPyArrayTypeIfAvailable()); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2204, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_ndarray = ((PyTypeObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_v_itemsize = -1L;
  if (unlikely(__pyx_v_args == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 2204, __pyx_L1_error)
  }
  __pyx_t_5 = PyTuple_GET_SIZE(((PyObject*)__pyx_v_args)); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(1, 2204, __pyx_L1_error)
  __pyx_t_2 = ((2 < __pyx_t_5) != 0);
  if (__pyx_t_2) {
    if (unlikely(__pyx_v_args == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 2204, __pyx_L1_error)
    }
    __pyx_t_1 = PyTuple_GET_ITEM(((PyObject*)__pyx_v_args), 2);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v_arg = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L6;
  }
  __pyx_t_3 = (__pyx_v_kwargs != Py_None);
  __pyx_t_4 = (__pyx_t_3 != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L7_bool_binop_done;
  }
  if (unlikely(__pyx_v_kwargs == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(1, 2204, __pyx_L1_error)
  }
  __pyx_t_4 = (__Pyx_PyDict_ContainsTF(__pyx_n_s_point, ((PyObject*)__pyx_v_kwargs), Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(1, 2204, __pyx_L1_error)
  __pyx_t_3 = (__pyx_t_4 != 0);
  __pyx_t_2 = __pyx_t_3;
  __pyx_L7_bool_binop_done:;
  if (__pyx_t_2) {
    if (unlikely(__pyx_v_kwargs == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 2204, __pyx_L1_error)
    }
    __pyx_t_1 = __Pyx_PyDict_GetItem(((PyObject*)__pyx_v_kwargs), __pyx_n_s_point); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2204, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_arg = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L6;
  }
  /*else*/ {
    if (unlikely(__pyx_v_args == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
      __PYX_ERR(1, 2204, __pyx_L1_error)
    }
    __pyx_t_5 = PyTuple_GET_SIZE(((PyObject*)__pyx_v_args)); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(1, 2204, __pyx_L1_error)
    __pyx_t_1 = PyInt_FromSsize_t(__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2204, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_6 = PyTuple_New(3); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 2204, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_INCREF(__pyx_int_3);
    __Pyx_GIVEREF(__pyx_int_3);
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_int_3);
    __Pyx_INCREF(__pyx_n_s_s);
    __Pyx_GIVEREF(__pyx_n_s_s);
    PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_n_s_s);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Expected_at_least_d_argument_s_g, __pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2204, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_1); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 2204, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(1, 2204, __pyx_L1_error)
  }
  __pyx_L6:;
  while (1) {
    __pyx_t_2 = PyFloat_Check(__pyx_v_arg); 
    __pyx_t_3 = (__pyx_t_2 != 0);
    if (__pyx_t_3) {
      if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, __pyx_n_s_float64_t, long, 1, __Pyx_PyInt_From_long, 1, 0, 0) < 0)) __PYX_ERR(1, 2204, __pyx_L1_error)
      goto __pyx_L10_break;
    }
    __pyx_t_2 = PyInt_Check(__pyx_v_arg); 
    __pyx_t_4 = (__pyx_t_2 != 0);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_3 = __pyx_t_4;
      goto __pyx_L13_bool_binop_done;
    }
    __pyx_t_4 = PyLong_Check(__pyx_v_arg); 
    __pyx_t_2 = (__pyx_t_4 != 0);
    __pyx_t_3 = __pyx_t_2;
    __pyx_L13_bool_binop_done:;
    __pyx_t_2 = (__pyx_t_3 != 0);
    if (__pyx_t_2) {
      if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, __pyx_n_s_int64_t, long, 1, __Pyx_PyInt_From_long, 1, 0, 0) < 0)) __PYX_ERR(1, 2204, __pyx_L1_error)
      goto __pyx_L10_break;
    }
    if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, Py_None, long, 1, __Pyx_PyInt_From_long, 1, 0, 0) < 0)) __PYX_ERR(1, 2204, __pyx_L1_error)
    goto __pyx_L10_break;
  }
  __pyx_L10_break:;
  __pyx_t_6 = PyList_New(0); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 2204, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_v_candidates = ((PyObject*)__pyx_t_6);
  __pyx_t_6 = 0;
  __pyx_t_5 = 0;
  if (unlikely(((PyObject *)__pyx_v_signatures) == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(1, 2204, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_dict_iterator(((PyObject*)__pyx_v_signatures), 1, ((PyObject *)NULL), (&__pyx_t_7), (&__pyx_t_8)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2204, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __pyx_t_6 = __pyx_t_1;
  __pyx_t_1 = 0;
  while (1) {
    __pyx_t_9 = __Pyx_dict_iter_next(__pyx_t_6, __pyx_t_7, &__pyx_t_5, &__pyx_t_1, NULL, NULL, __pyx_t_8);
    if (unlikely(__pyx_t_9 == 0)) break;
    if (unlikely(__pyx_t_9 == -1)) __PYX_ERR(1, 2204, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_sig, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_v_match_found = 0;
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_v_sig, __pyx_n_s_strip); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 2204, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __pyx_t_12 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_11))) {
      __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_11);
      if (likely(__pyx_t_12)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
        __Pyx_INCREF(__pyx_t_12);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_11, function);
      }
    }
    __pyx_t_10 = (__pyx_t_12) ? __Pyx_PyObject_Call2Args(__pyx_t_11, __pyx_t_12, __pyx_kp_s__13) : __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_kp_s__13);
    __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
    if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 2204, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_split); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 2204, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_t_10 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_11))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_11);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_11, function);
      }
    }
    __pyx_t_1 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_11, __pyx_t_10, __pyx_kp_s__14) : __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_kp_s__14);
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2204, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_XDECREF_SET(__pyx_v_src_sig, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_13 = PyList_GET_SIZE(__pyx_v_dest_sig); if (unlikely(__pyx_t_13 == ((Py_ssize_t)-1))) __PYX_ERR(1, 2204, __pyx_L1_error)
    __pyx_t_14 = __pyx_t_13;
    for (__pyx_t_15 = 0; __pyx_t_15 < __pyx_t_14; __pyx_t_15+=1) {
      __pyx_v_i = __pyx_t_15;
      __pyx_t_1 = PyList_GET_ITEM(__pyx_v_dest_sig, __pyx_v_i);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_XDECREF_SET(__pyx_v_dst_type, __pyx_t_1);
      __pyx_t_1 = 0;
      __pyx_t_2 = (__pyx_v_dst_type != Py_None);
      __pyx_t_3 = (__pyx_t_2 != 0);
      if (__pyx_t_3) {
        __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_src_sig, __pyx_v_i, Py_ssize_t, 1, PyInt_FromSsize_t, 0, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2204, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_11 = PyObject_RichCompare(__pyx_t_1, __pyx_v_dst_type, Py_EQ); __Pyx_XGOTREF(__pyx_t_11); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 2204, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_11); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(1, 2204, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        if (__pyx_t_3) {
          __pyx_v_match_found = 1;
          goto __pyx_L20;
        }
        /*else*/ {
          __pyx_v_match_found = 0;
          goto __pyx_L18_break;
        }
        __pyx_L20:;
      }
    }
    __pyx_L18_break:;
    __pyx_t_3 = (__pyx_v_match_found != 0);
    if (__pyx_t_3) {
      __pyx_t_16 = __Pyx_PyList_Append(__pyx_v_candidates, __pyx_v_sig); if (unlikely(__pyx_t_16 == ((int)-1))) __PYX_ERR(1, 2204, __pyx_L1_error)
    }
  }
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_3 = (PyList_GET_SIZE(__pyx_v_candidates) != 0);
  __pyx_t_2 = ((!__pyx_t_3) != 0);
  if (__pyx_t_2) {
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__15, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 2204, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(1, 2204, __pyx_L1_error)
  }
  __pyx_t_7 = PyList_GET_SIZE(__pyx_v_candidates); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1))) __PYX_ERR(1, 2204, __pyx_L1_error)
  __pyx_t_2 = ((__pyx_t_7 > 1) != 0);
  if (__pyx_t_2) {
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__16, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 2204, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(1, 2204, __pyx_L1_error)
  }
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    if (unlikely(((PyObject *)__pyx_v_signatures) == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 2204, __pyx_L1_error)
    }
    __pyx_t_6 = __Pyx_PyDict_GetItem(((PyObject*)__pyx_v_signatures), PyList_GET_ITEM(__pyx_v_candidates, 0)); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 2204, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_r = __pyx_t_6;
    __pyx_t_6 = 0;
    goto __pyx_L0;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("pandas._libs.interval.Int32ClosedNeitherIntervalNode.__pyx_fused_cpdef", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_dest_sig);
  __Pyx_XDECREF(__pyx_v_ndarray);
  __Pyx_XDECREF(__pyx_v_arg);
  __Pyx_XDECREF(__pyx_v_candidates);
  __Pyx_XDECREF(__pyx_v_sig);
  __Pyx_XDECREF(__pyx_v_src_sig);
  __Pyx_XDECREF(__pyx_v_dst_type);
  __Pyx_XDECREF(__pyx_v_kwargs);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_9__pyx_fuse_0query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_fuse_0__pyx_f_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float64_t __pyx_v_point, int __pyx_skip_dispatch) {
  __Pyx_memviewslice __pyx_v_indices = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_values = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __pyx_t_5numpy_int64_t __pyx_t_9;
  __pyx_t_5numpy_int64_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  int __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  __Pyx_memviewslice __pyx_t_16 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_17 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  Py_ssize_t __pyx_t_21;
  __Pyx_RefNannySetupContext("__pyx_fuse_0query", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pyx_fuse_0query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2204, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_9__pyx_fuse_0query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = PyFloat_FromDouble(__pyx_v_point); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2204, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2204, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2204, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 2204, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_5) {
            __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2204, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2204, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_3query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_4 = PyFloat_FromDouble(__pyx_v_point); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 2204, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_7 = __pyx_t_1; __pyx_t_3 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2204, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2204, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 2204, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (__pyx_t_3) {
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_t_4);
          __pyx_t_4 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2204, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pandas/_libs/intervaltree.pxi":2213
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
  __pyx_t_8 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_8) {

    /* "pandas/_libs/intervaltree.pxi":2217
 *             # continue the binary tree structure. Instead, we use linear
 *             # search.
 *             for i in range(self.n_elements):             # <<<<<<<<<<<<<<
 *                 if self.left[i] < point < self.right[i]:
 *                     result.append(self.indices[i])
 */
    __pyx_t_9 = __pyx_v_self->n_elements;
    __pyx_t_10 = __pyx_t_9;
    for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
      __pyx_v_i = __pyx_t_11;

      /* "pandas/_libs/intervaltree.pxi":2218
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point < self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      __pyx_t_12 = __pyx_v_i;
      __pyx_t_8 = ((*((__pyx_t_5numpy_int32_t *) ( /* dim=0 */ (__pyx_v_self->left.data + __pyx_t_12 * __pyx_v_self->left.strides[0]) ))) < __pyx_v_point);
      if (__pyx_t_8) {
        __pyx_t_13 = __pyx_v_i;
        __pyx_t_8 = (__pyx_v_point < (*((__pyx_t_5numpy_int32_t *) ( /* dim=0 */ (__pyx_v_self->right.data + __pyx_t_13 * __pyx_v_self->right.strides[0]) ))));
      }
      __pyx_t_14 = (__pyx_t_8 != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":2219
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point < self.right[i]:
 *                     result.append(self.indices[i])             # <<<<<<<<<<<<<<
 *         else:
 *             # There are child nodes. Based on comparing our query to the pivot,
 */
        __pyx_t_15 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->indices.data + __pyx_t_15 * __pyx_v_self->indices.strides[0]) ))));

        /* "pandas/_libs/intervaltree.pxi":2218
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point < self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      }
    }

    /* "pandas/_libs/intervaltree.pxi":2213
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
    goto __pyx_L3;
  }

  /* "pandas/_libs/intervaltree.pxi":2223
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
  /*else*/ {
    __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":2224
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:
 *                 values = self.center_left_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 */
      __pyx_t_16 = __pyx_v_self->center_left_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":2225
 *             if point < self.pivot:
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:
 */
      __pyx_t_17 = __pyx_v_self->center_left_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":2226
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):             # <<<<<<<<<<<<<<
 *                     if not values[i] < point:
 *                         break
 */
      __pyx_t_9 = __pyx_v_self->n_center;
      __pyx_t_10 = __pyx_t_9;
      for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":2227
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_18 = __pyx_v_i;
        __pyx_t_14 = ((!(((*((__pyx_t_5numpy_int32_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_18 * __pyx_v_values.strides[0]) ))) < __pyx_v_point) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":2228
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:
 */
          goto __pyx_L9_break;

          /* "pandas/_libs/intervaltree.pxi":2227
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":2229
 *                     if not values[i] < point:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 */
        __pyx_t_19 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_19 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L9_break:;

      /* "pandas/_libs/intervaltree.pxi":2230
 *                         break
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->left_node->max_right) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":2231
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)             # <<<<<<<<<<<<<<
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode *)__pyx_v_self->left_node->__pyx_vtab)->__pyx_fuse_0query(__pyx_v_self->left_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2231, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":2230
 *                         break
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":2223
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":2232
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
    __pyx_t_14 = ((__pyx_v_point > __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":2233
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 *                 values = self.center_right_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 */
      __pyx_t_16 = __pyx_v_self->center_right_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":2234
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:
 */
      __pyx_t_17 = __pyx_v_self->center_right_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":2235
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):             # <<<<<<<<<<<<<<
 *                     if not point < values[i]:
 *                         break
 */
      for (__pyx_t_11 = (__pyx_v_self->n_center - 1); __pyx_t_11 > -1L; __pyx_t_11-=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":2236
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_20 = __pyx_v_i;
        __pyx_t_14 = ((!((__pyx_v_point < (*((__pyx_t_5numpy_int32_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_20 * __pyx_v_values.strides[0]) )))) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":2237
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:
 */
          goto __pyx_L13_break;

          /* "pandas/_libs/intervaltree.pxi":2236
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":2238
 *                     if not point < values[i]:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if self.right_node.min_left < point:
 *                     self.right_node.query(result, point)
 */
        __pyx_t_21 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_21 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L13_break:;

      /* "pandas/_libs/intervaltree.pxi":2239
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      __pyx_t_14 = ((__pyx_v_self->right_node->min_left < __pyx_v_point) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":2240
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:
 *                     self.right_node.query(result, point)             # <<<<<<<<<<<<<<
 *             else:
 *                 result.extend(self.center_left_indices)
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode *)__pyx_v_self->right_node->__pyx_vtab)->__pyx_fuse_0query(__pyx_v_self->right_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2240, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":2239
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":2232
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":2242
 *                     self.right_node.query(result, point)
 *             else:
 *                 result.extend(self.center_left_indices)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
    /*else*/ {
      __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->extend(__pyx_v_result, __pyx_v_self->center_left_indices); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2242, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_L7:;
  }
  __pyx_L3:;

  /* "pandas/_libs/intervaltree.pxi":2204
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __PYX_XDEC_MEMVIEW(&__pyx_t_16, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_17, 1);
  __Pyx_AddTraceback("pandas._libs.interval.Int32ClosedNeitherIntervalNode.query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_indices, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_values, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_9__pyx_fuse_0query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_0__pyx_mdef_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_9__pyx_fuse_0query = {"__pyx_fuse_0query", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_9__pyx_fuse_0query, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_2query};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_9__pyx_fuse_0query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result = 0;
  __pyx_t_5numpy_float64_t __pyx_v_point;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fuse_0query (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_result,&__pyx_n_s_point,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_result)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_point)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_0query", 1, 2, 2, 1); __PYX_ERR(1, 2204, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_0query") < 0)) __PYX_ERR(1, 2204, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_result = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)values[0]);
    __pyx_v_point = __pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_point == ((npy_float64)-1)) && PyErr_Occurred())) __PYX_ERR(1, 2204, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_0query", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 2204, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Int32ClosedNeitherIntervalNode.__pyx_fuse_0query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_result), __pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector, 1, "result", 0))) __PYX_ERR(1, 2204, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_8__pyx_fuse_0query(((struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode *)__pyx_v_self), __pyx_v_result, __pyx_v_point);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_8__pyx_fuse_0query(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float64_t __pyx_v_point) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__pyx_fuse_0query", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_fuse_0__pyx_f_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_query(__pyx_v_self, __pyx_v_result, __pyx_v_point, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2204, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Int32ClosedNeitherIntervalNode.__pyx_fuse_0query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_11__pyx_fuse_1query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_fuse_1__pyx_f_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float32_t __pyx_v_point, int __pyx_skip_dispatch) {
  __Pyx_memviewslice __pyx_v_indices = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_values = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __pyx_t_5numpy_int64_t __pyx_t_9;
  __pyx_t_5numpy_int64_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  int __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  __Pyx_memviewslice __pyx_t_16 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_17 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  Py_ssize_t __pyx_t_21;
  __Pyx_RefNannySetupContext("__pyx_fuse_1query", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pyx_fuse_1query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2204, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_11__pyx_fuse_1query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = PyFloat_FromDouble(__pyx_v_point); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2204, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2204, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2204, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 2204, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_5) {
            __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2204, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2204, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_3query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_4 = PyFloat_FromDouble(__pyx_v_point); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 2204, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_7 = __pyx_t_1; __pyx_t_3 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2204, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2204, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 2204, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (__pyx_t_3) {
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_t_4);
          __pyx_t_4 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2204, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pandas/_libs/intervaltree.pxi":2213
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
  __pyx_t_8 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_8) {

    /* "pandas/_libs/intervaltree.pxi":2217
 *             # continue the binary tree structure. Instead, we use linear
 *             # search.
 *             for i in range(self.n_elements):             # <<<<<<<<<<<<<<
 *                 if self.left[i] < point < self.right[i]:
 *                     result.append(self.indices[i])
 */
    __pyx_t_9 = __pyx_v_self->n_elements;
    __pyx_t_10 = __pyx_t_9;
    for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
      __pyx_v_i = __pyx_t_11;

      /* "pandas/_libs/intervaltree.pxi":2218
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point < self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      __pyx_t_12 = __pyx_v_i;
      __pyx_t_8 = ((*((__pyx_t_5numpy_int32_t *) ( /* dim=0 */ (__pyx_v_self->left.data + __pyx_t_12 * __pyx_v_self->left.strides[0]) ))) < __pyx_v_point);
      if (__pyx_t_8) {
        __pyx_t_13 = __pyx_v_i;
        __pyx_t_8 = (__pyx_v_point < (*((__pyx_t_5numpy_int32_t *) ( /* dim=0 */ (__pyx_v_self->right.data + __pyx_t_13 * __pyx_v_self->right.strides[0]) ))));
      }
      __pyx_t_14 = (__pyx_t_8 != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":2219
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point < self.right[i]:
 *                     result.append(self.indices[i])             # <<<<<<<<<<<<<<
 *         else:
 *             # There are child nodes. Based on comparing our query to the pivot,
 */
        __pyx_t_15 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->indices.data + __pyx_t_15 * __pyx_v_self->indices.strides[0]) ))));

        /* "pandas/_libs/intervaltree.pxi":2218
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point < self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      }
    }

    /* "pandas/_libs/intervaltree.pxi":2213
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
    goto __pyx_L3;
  }

  /* "pandas/_libs/intervaltree.pxi":2223
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
  /*else*/ {
    __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":2224
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:
 *                 values = self.center_left_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 */
      __pyx_t_16 = __pyx_v_self->center_left_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":2225
 *             if point < self.pivot:
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:
 */
      __pyx_t_17 = __pyx_v_self->center_left_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":2226
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):             # <<<<<<<<<<<<<<
 *                     if not values[i] < point:
 *                         break
 */
      __pyx_t_9 = __pyx_v_self->n_center;
      __pyx_t_10 = __pyx_t_9;
      for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":2227
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_18 = __pyx_v_i;
        __pyx_t_14 = ((!(((*((__pyx_t_5numpy_int32_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_18 * __pyx_v_values.strides[0]) ))) < __pyx_v_point) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":2228
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:
 */
          goto __pyx_L9_break;

          /* "pandas/_libs/intervaltree.pxi":2227
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":2229
 *                     if not values[i] < point:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 */
        __pyx_t_19 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_19 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L9_break:;

      /* "pandas/_libs/intervaltree.pxi":2230
 *                         break
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->left_node->max_right) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":2231
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)             # <<<<<<<<<<<<<<
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode *)__pyx_v_self->left_node->__pyx_vtab)->__pyx_fuse_1query(__pyx_v_self->left_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2231, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":2230
 *                         break
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":2223
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":2232
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
    __pyx_t_14 = ((__pyx_v_point > __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":2233
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 *                 values = self.center_right_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 */
      __pyx_t_16 = __pyx_v_self->center_right_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":2234
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:
 */
      __pyx_t_17 = __pyx_v_self->center_right_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":2235
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):             # <<<<<<<<<<<<<<
 *                     if not point < values[i]:
 *                         break
 */
      for (__pyx_t_11 = (__pyx_v_self->n_center - 1); __pyx_t_11 > -1L; __pyx_t_11-=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":2236
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_20 = __pyx_v_i;
        __pyx_t_14 = ((!((__pyx_v_point < (*((__pyx_t_5numpy_int32_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_20 * __pyx_v_values.strides[0]) )))) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":2237
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:
 */
          goto __pyx_L13_break;

          /* "pandas/_libs/intervaltree.pxi":2236
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":2238
 *                     if not point < values[i]:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if self.right_node.min_left < point:
 *                     self.right_node.query(result, point)
 */
        __pyx_t_21 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_21 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L13_break:;

      /* "pandas/_libs/intervaltree.pxi":2239
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      __pyx_t_14 = ((__pyx_v_self->right_node->min_left < __pyx_v_point) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":2240
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:
 *                     self.right_node.query(result, point)             # <<<<<<<<<<<<<<
 *             else:
 *                 result.extend(self.center_left_indices)
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode *)__pyx_v_self->right_node->__pyx_vtab)->__pyx_fuse_1query(__pyx_v_self->right_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2240, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":2239
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":2232
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":2242
 *                     self.right_node.query(result, point)
 *             else:
 *                 result.extend(self.center_left_indices)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
    /*else*/ {
      __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->extend(__pyx_v_result, __pyx_v_self->center_left_indices); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2242, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_L7:;
  }
  __pyx_L3:;

  /* "pandas/_libs/intervaltree.pxi":2204
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __PYX_XDEC_MEMVIEW(&__pyx_t_16, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_17, 1);
  __Pyx_AddTraceback("pandas._libs.interval.Int32ClosedNeitherIntervalNode.query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_indices, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_values, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_11__pyx_fuse_1query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_1__pyx_mdef_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_11__pyx_fuse_1query = {"__pyx_fuse_1query", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_11__pyx_fuse_1query, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_2query};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_11__pyx_fuse_1query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result = 0;
  __pyx_t_5numpy_float32_t __pyx_v_point;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fuse_1query (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_result,&__pyx_n_s_point,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_result)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_point)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_1query", 1, 2, 2, 1); __PYX_ERR(1, 2204, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_1query") < 0)) __PYX_ERR(1, 2204, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_result = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)values[0]);
    __pyx_v_point = __pyx_PyFloat_AsFloat(values[1]); if (unlikely((__pyx_v_point == ((npy_float32)-1)) && PyErr_Occurred())) __PYX_ERR(1, 2204, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_1query", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 2204, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Int32ClosedNeitherIntervalNode.__pyx_fuse_1query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_result), __pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector, 1, "result", 0))) __PYX_ERR(1, 2204, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_10__pyx_fuse_1query(((struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode *)__pyx_v_self), __pyx_v_result, __pyx_v_point);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_10__pyx_fuse_1query(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float32_t __pyx_v_point) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__pyx_fuse_1query", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_fuse_1__pyx_f_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_query(__pyx_v_self, __pyx_v_result, __pyx_v_point, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2204, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Int32ClosedNeitherIntervalNode.__pyx_fuse_1query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_13__pyx_fuse_2query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_fuse_2__pyx_f_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int64_t __pyx_v_point, int __pyx_skip_dispatch) {
  __Pyx_memviewslice __pyx_v_indices = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_values = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __pyx_t_5numpy_int64_t __pyx_t_9;
  __pyx_t_5numpy_int64_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  int __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  __Pyx_memviewslice __pyx_t_16 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_17 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  Py_ssize_t __pyx_t_21;
  __Pyx_RefNannySetupContext("__pyx_fuse_2query", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pyx_fuse_2query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2204, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_13__pyx_fuse_2query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = __Pyx_PyInt_From_npy_int64(__pyx_v_point); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2204, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2204, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2204, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 2204, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_5) {
            __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2204, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2204, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_3query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_4 = __Pyx_PyInt_From_npy_int64(__pyx_v_point); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 2204, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_7 = __pyx_t_1; __pyx_t_3 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2204, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2204, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 2204, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (__pyx_t_3) {
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_t_4);
          __pyx_t_4 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2204, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pandas/_libs/intervaltree.pxi":2213
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
  __pyx_t_8 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_8) {

    /* "pandas/_libs/intervaltree.pxi":2217
 *             # continue the binary tree structure. Instead, we use linear
 *             # search.
 *             for i in range(self.n_elements):             # <<<<<<<<<<<<<<
 *                 if self.left[i] < point < self.right[i]:
 *                     result.append(self.indices[i])
 */
    __pyx_t_9 = __pyx_v_self->n_elements;
    __pyx_t_10 = __pyx_t_9;
    for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
      __pyx_v_i = __pyx_t_11;

      /* "pandas/_libs/intervaltree.pxi":2218
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point < self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      __pyx_t_12 = __pyx_v_i;
      __pyx_t_8 = ((*((__pyx_t_5numpy_int32_t *) ( /* dim=0 */ (__pyx_v_self->left.data + __pyx_t_12 * __pyx_v_self->left.strides[0]) ))) < __pyx_v_point);
      if (__pyx_t_8) {
        __pyx_t_13 = __pyx_v_i;
        __pyx_t_8 = (__pyx_v_point < (*((__pyx_t_5numpy_int32_t *) ( /* dim=0 */ (__pyx_v_self->right.data + __pyx_t_13 * __pyx_v_self->right.strides[0]) ))));
      }
      __pyx_t_14 = (__pyx_t_8 != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":2219
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point < self.right[i]:
 *                     result.append(self.indices[i])             # <<<<<<<<<<<<<<
 *         else:
 *             # There are child nodes. Based on comparing our query to the pivot,
 */
        __pyx_t_15 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->indices.data + __pyx_t_15 * __pyx_v_self->indices.strides[0]) ))));

        /* "pandas/_libs/intervaltree.pxi":2218
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point < self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      }
    }

    /* "pandas/_libs/intervaltree.pxi":2213
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
    goto __pyx_L3;
  }

  /* "pandas/_libs/intervaltree.pxi":2223
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
  /*else*/ {
    __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":2224
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:
 *                 values = self.center_left_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 */
      __pyx_t_16 = __pyx_v_self->center_left_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":2225
 *             if point < self.pivot:
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:
 */
      __pyx_t_17 = __pyx_v_self->center_left_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":2226
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):             # <<<<<<<<<<<<<<
 *                     if not values[i] < point:
 *                         break
 */
      __pyx_t_9 = __pyx_v_self->n_center;
      __pyx_t_10 = __pyx_t_9;
      for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":2227
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_18 = __pyx_v_i;
        __pyx_t_14 = ((!(((*((__pyx_t_5numpy_int32_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_18 * __pyx_v_values.strides[0]) ))) < __pyx_v_point) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":2228
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:
 */
          goto __pyx_L9_break;

          /* "pandas/_libs/intervaltree.pxi":2227
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":2229
 *                     if not values[i] < point:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 */
        __pyx_t_19 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_19 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L9_break:;

      /* "pandas/_libs/intervaltree.pxi":2230
 *                         break
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->left_node->max_right) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":2231
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)             # <<<<<<<<<<<<<<
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode *)__pyx_v_self->left_node->__pyx_vtab)->__pyx_fuse_2query(__pyx_v_self->left_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2231, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":2230
 *                         break
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":2223
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":2232
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
    __pyx_t_14 = ((__pyx_v_point > __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":2233
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 *                 values = self.center_right_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 */
      __pyx_t_16 = __pyx_v_self->center_right_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":2234
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:
 */
      __pyx_t_17 = __pyx_v_self->center_right_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":2235
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):             # <<<<<<<<<<<<<<
 *                     if not point < values[i]:
 *                         break
 */
      for (__pyx_t_11 = (__pyx_v_self->n_center - 1); __pyx_t_11 > -1L; __pyx_t_11-=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":2236
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_20 = __pyx_v_i;
        __pyx_t_14 = ((!((__pyx_v_point < (*((__pyx_t_5numpy_int32_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_20 * __pyx_v_values.strides[0]) )))) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":2237
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:
 */
          goto __pyx_L13_break;

          /* "pandas/_libs/intervaltree.pxi":2236
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":2238
 *                     if not point < values[i]:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if self.right_node.min_left < point:
 *                     self.right_node.query(result, point)
 */
        __pyx_t_21 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_21 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L13_break:;

      /* "pandas/_libs/intervaltree.pxi":2239
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      __pyx_t_14 = ((__pyx_v_self->right_node->min_left < __pyx_v_point) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":2240
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:
 *                     self.right_node.query(result, point)             # <<<<<<<<<<<<<<
 *             else:
 *                 result.extend(self.center_left_indices)
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode *)__pyx_v_self->right_node->__pyx_vtab)->__pyx_fuse_2query(__pyx_v_self->right_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2240, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":2239
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":2232
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":2242
 *                     self.right_node.query(result, point)
 *             else:
 *                 result.extend(self.center_left_indices)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
    /*else*/ {
      __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->extend(__pyx_v_result, __pyx_v_self->center_left_indices); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2242, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_L7:;
  }
  __pyx_L3:;

  /* "pandas/_libs/intervaltree.pxi":2204
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __PYX_XDEC_MEMVIEW(&__pyx_t_16, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_17, 1);
  __Pyx_AddTraceback("pandas._libs.interval.Int32ClosedNeitherIntervalNode.query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_indices, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_values, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_13__pyx_fuse_2query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_2__pyx_mdef_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_13__pyx_fuse_2query = {"__pyx_fuse_2query", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_13__pyx_fuse_2query, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_2query};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_13__pyx_fuse_2query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result = 0;
  __pyx_t_5numpy_int64_t __pyx_v_point;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fuse_2query (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_result,&__pyx_n_s_point,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_result)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_point)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_2query", 1, 2, 2, 1); __PYX_ERR(1, 2204, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_2query") < 0)) __PYX_ERR(1, 2204, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_result = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)values[0]);
    __pyx_v_point = __Pyx_PyInt_As_npy_int64(values[1]); if (unlikely((__pyx_v_point == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(1, 2204, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_2query", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 2204, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Int32ClosedNeitherIntervalNode.__pyx_fuse_2query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_result), __pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector, 1, "result", 0))) __PYX_ERR(1, 2204, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_12__pyx_fuse_2query(((struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode *)__pyx_v_self), __pyx_v_result, __pyx_v_point);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_12__pyx_fuse_2query(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int64_t __pyx_v_point) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__pyx_fuse_2query", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_fuse_2__pyx_f_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_query(__pyx_v_self, __pyx_v_result, __pyx_v_point, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2204, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Int32ClosedNeitherIntervalNode.__pyx_fuse_2query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_15__pyx_fuse_3query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_fuse_3__pyx_f_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int32_t __pyx_v_point, int __pyx_skip_dispatch) {
  __Pyx_memviewslice __pyx_v_indices = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_values = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __pyx_t_5numpy_int64_t __pyx_t_9;
  __pyx_t_5numpy_int64_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  int __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  __Pyx_memviewslice __pyx_t_16 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_17 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  Py_ssize_t __pyx_t_21;
  __Pyx_RefNannySetupContext("__pyx_fuse_3query", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pyx_fuse_3query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2204, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_15__pyx_fuse_3query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = __Pyx_PyInt_From_npy_int32(__pyx_v_point); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2204, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2204, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2204, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 2204, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_5) {
            __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2204, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2204, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_3query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_4 = __Pyx_PyInt_From_npy_int32(__pyx_v_point); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 2204, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_7 = __pyx_t_1; __pyx_t_3 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2204, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2204, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 2204, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (__pyx_t_3) {
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_t_4);
          __pyx_t_4 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2204, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pandas/_libs/intervaltree.pxi":2213
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
  __pyx_t_8 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_8) {

    /* "pandas/_libs/intervaltree.pxi":2217
 *             # continue the binary tree structure. Instead, we use linear
 *             # search.
 *             for i in range(self.n_elements):             # <<<<<<<<<<<<<<
 *                 if self.left[i] < point < self.right[i]:
 *                     result.append(self.indices[i])
 */
    __pyx_t_9 = __pyx_v_self->n_elements;
    __pyx_t_10 = __pyx_t_9;
    for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
      __pyx_v_i = __pyx_t_11;

      /* "pandas/_libs/intervaltree.pxi":2218
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point < self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      __pyx_t_12 = __pyx_v_i;
      __pyx_t_8 = ((*((__pyx_t_5numpy_int32_t *) ( /* dim=0 */ (__pyx_v_self->left.data + __pyx_t_12 * __pyx_v_self->left.strides[0]) ))) < __pyx_v_point);
      if (__pyx_t_8) {
        __pyx_t_13 = __pyx_v_i;
        __pyx_t_8 = (__pyx_v_point < (*((__pyx_t_5numpy_int32_t *) ( /* dim=0 */ (__pyx_v_self->right.data + __pyx_t_13 * __pyx_v_self->right.strides[0]) ))));
      }
      __pyx_t_14 = (__pyx_t_8 != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":2219
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point < self.right[i]:
 *                     result.append(self.indices[i])             # <<<<<<<<<<<<<<
 *         else:
 *             # There are child nodes. Based on comparing our query to the pivot,
 */
        __pyx_t_15 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->indices.data + __pyx_t_15 * __pyx_v_self->indices.strides[0]) ))));

        /* "pandas/_libs/intervaltree.pxi":2218
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point < self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      }
    }

    /* "pandas/_libs/intervaltree.pxi":2213
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
    goto __pyx_L3;
  }

  /* "pandas/_libs/intervaltree.pxi":2223
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
  /*else*/ {
    __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":2224
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:
 *                 values = self.center_left_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 */
      __pyx_t_16 = __pyx_v_self->center_left_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":2225
 *             if point < self.pivot:
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:
 */
      __pyx_t_17 = __pyx_v_self->center_left_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":2226
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):             # <<<<<<<<<<<<<<
 *                     if not values[i] < point:
 *                         break
 */
      __pyx_t_9 = __pyx_v_self->n_center;
      __pyx_t_10 = __pyx_t_9;
      for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":2227
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_18 = __pyx_v_i;
        __pyx_t_14 = ((!(((*((__pyx_t_5numpy_int32_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_18 * __pyx_v_values.strides[0]) ))) < __pyx_v_point) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":2228
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:
 */
          goto __pyx_L9_break;

          /* "pandas/_libs/intervaltree.pxi":2227
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":2229
 *                     if not values[i] < point:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 */
        __pyx_t_19 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_19 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L9_break:;

      /* "pandas/_libs/intervaltree.pxi":2230
 *                         break
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->left_node->max_right) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":2231
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)             # <<<<<<<<<<<<<<
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode *)__pyx_v_self->left_node->__pyx_vtab)->__pyx_fuse_3query(__pyx_v_self->left_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2231, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":2230
 *                         break
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":2223
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":2232
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
    __pyx_t_14 = ((__pyx_v_point > __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":2233
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 *                 values = self.center_right_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 */
      __pyx_t_16 = __pyx_v_self->center_right_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":2234
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:
 */
      __pyx_t_17 = __pyx_v_self->center_right_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":2235
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):             # <<<<<<<<<<<<<<
 *                     if not point < values[i]:
 *                         break
 */
      for (__pyx_t_11 = (__pyx_v_self->n_center - 1); __pyx_t_11 > -1L; __pyx_t_11-=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":2236
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_20 = __pyx_v_i;
        __pyx_t_14 = ((!((__pyx_v_point < (*((__pyx_t_5numpy_int32_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_20 * __pyx_v_values.strides[0]) )))) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":2237
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:
 */
          goto __pyx_L13_break;

          /* "pandas/_libs/intervaltree.pxi":2236
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":2238
 *                     if not point < values[i]:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if self.right_node.min_left < point:
 *                     self.right_node.query(result, point)
 */
        __pyx_t_21 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_21 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L13_break:;

      /* "pandas/_libs/intervaltree.pxi":2239
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      __pyx_t_14 = ((__pyx_v_self->right_node->min_left < __pyx_v_point) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":2240
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:
 *                     self.right_node.query(result, point)             # <<<<<<<<<<<<<<
 *             else:
 *                 result.extend(self.center_left_indices)
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode *)__pyx_v_self->right_node->__pyx_vtab)->__pyx_fuse_3query(__pyx_v_self->right_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2240, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":2239
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":2232
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":2242
 *                     self.right_node.query(result, point)
 *             else:
 *                 result.extend(self.center_left_indices)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
    /*else*/ {
      __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->extend(__pyx_v_result, __pyx_v_self->center_left_indices); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2242, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_L7:;
  }
  __pyx_L3:;

  /* "pandas/_libs/intervaltree.pxi":2204
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __PYX_XDEC_MEMVIEW(&__pyx_t_16, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_17, 1);
  __Pyx_AddTraceback("pandas._libs.interval.Int32ClosedNeitherIntervalNode.query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_indices, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_values, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_15__pyx_fuse_3query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_3__pyx_mdef_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_15__pyx_fuse_3query = {"__pyx_fuse_3query", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_15__pyx_fuse_3query, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_2query};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_15__pyx_fuse_3query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result = 0;
  __pyx_t_5numpy_int32_t __pyx_v_point;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fuse_3query (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_result,&__pyx_n_s_point,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_result)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_point)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_3query", 1, 2, 2, 1); __PYX_ERR(1, 2204, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_3query") < 0)) __PYX_ERR(1, 2204, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_result = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)values[0]);
    __pyx_v_point = __Pyx_PyInt_As_npy_int32(values[1]); if (unlikely((__pyx_v_point == ((npy_int32)-1)) && PyErr_Occurred())) __PYX_ERR(1, 2204, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_3query", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 2204, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Int32ClosedNeitherIntervalNode.__pyx_fuse_3query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_result), __pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector, 1, "result", 0))) __PYX_ERR(1, 2204, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_14__pyx_fuse_3query(((struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode *)__pyx_v_self), __pyx_v_result, __pyx_v_point);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_14__pyx_fuse_3query(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int32_t __pyx_v_point) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__pyx_fuse_3query", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_fuse_3__pyx_f_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_query(__pyx_v_self, __pyx_v_result, __pyx_v_point, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2204, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Int32ClosedNeitherIntervalNode.__pyx_fuse_3query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_17__pyx_fuse_4query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_fuse_4__pyx_f_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_uint64_t __pyx_v_point, int __pyx_skip_dispatch) {
  __Pyx_memviewslice __pyx_v_indices = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_values = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __pyx_t_5numpy_int64_t __pyx_t_9;
  __pyx_t_5numpy_int64_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  int __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  __Pyx_memviewslice __pyx_t_16 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_17 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  Py_ssize_t __pyx_t_21;
  __Pyx_RefNannySetupContext("__pyx_fuse_4query", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pyx_fuse_4query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2204, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_17__pyx_fuse_4query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = __Pyx_PyInt_From_npy_uint64(__pyx_v_point); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2204, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2204, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2204, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 2204, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_5) {
            __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2204, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2204, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_3query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_4 = __Pyx_PyInt_From_npy_uint64(__pyx_v_point); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 2204, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_7 = __pyx_t_1; __pyx_t_3 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2204, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2204, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 2204, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (__pyx_t_3) {
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_t_4);
          __pyx_t_4 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2204, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pandas/_libs/intervaltree.pxi":2213
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
  __pyx_t_8 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_8) {

    /* "pandas/_libs/intervaltree.pxi":2217
 *             # continue the binary tree structure. Instead, we use linear
 *             # search.
 *             for i in range(self.n_elements):             # <<<<<<<<<<<<<<
 *                 if self.left[i] < point < self.right[i]:
 *                     result.append(self.indices[i])
 */
    __pyx_t_9 = __pyx_v_self->n_elements;
    __pyx_t_10 = __pyx_t_9;
    for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
      __pyx_v_i = __pyx_t_11;

      /* "pandas/_libs/intervaltree.pxi":2218
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point < self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      __pyx_t_12 = __pyx_v_i;
      __pyx_t_8 = ((*((__pyx_t_5numpy_int32_t *) ( /* dim=0 */ (__pyx_v_self->left.data + __pyx_t_12 * __pyx_v_self->left.strides[0]) ))) < __pyx_v_point);
      if (__pyx_t_8) {
        __pyx_t_13 = __pyx_v_i;
        __pyx_t_8 = (__pyx_v_point < (*((__pyx_t_5numpy_int32_t *) ( /* dim=0 */ (__pyx_v_self->right.data + __pyx_t_13 * __pyx_v_self->right.strides[0]) ))));
      }
      __pyx_t_14 = (__pyx_t_8 != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":2219
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point < self.right[i]:
 *                     result.append(self.indices[i])             # <<<<<<<<<<<<<<
 *         else:
 *             # There are child nodes. Based on comparing our query to the pivot,
 */
        __pyx_t_15 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->indices.data + __pyx_t_15 * __pyx_v_self->indices.strides[0]) ))));

        /* "pandas/_libs/intervaltree.pxi":2218
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point < self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      }
    }

    /* "pandas/_libs/intervaltree.pxi":2213
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
    goto __pyx_L3;
  }

  /* "pandas/_libs/intervaltree.pxi":2223
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
  /*else*/ {
    __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":2224
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:
 *                 values = self.center_left_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 */
      __pyx_t_16 = __pyx_v_self->center_left_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":2225
 *             if point < self.pivot:
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:
 */
      __pyx_t_17 = __pyx_v_self->center_left_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":2226
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):             # <<<<<<<<<<<<<<
 *                     if not values[i] < point:
 *                         break
 */
      __pyx_t_9 = __pyx_v_self->n_center;
      __pyx_t_10 = __pyx_t_9;
      for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":2227
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_18 = __pyx_v_i;
        __pyx_t_14 = ((!(((*((__pyx_t_5numpy_int32_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_18 * __pyx_v_values.strides[0]) ))) < __pyx_v_point) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":2228
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:
 */
          goto __pyx_L9_break;

          /* "pandas/_libs/intervaltree.pxi":2227
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":2229
 *                     if not values[i] < point:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 */
        __pyx_t_19 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_19 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L9_break:;

      /* "pandas/_libs/intervaltree.pxi":2230
 *                         break
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->left_node->max_right) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":2231
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)             # <<<<<<<<<<<<<<
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode *)__pyx_v_self->left_node->__pyx_vtab)->__pyx_fuse_4query(__pyx_v_self->left_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2231, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":2230
 *                         break
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":2223
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":2232
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
    __pyx_t_14 = ((__pyx_v_point > __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":2233
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 *                 values = self.center_right_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 */
      __pyx_t_16 = __pyx_v_self->center_right_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":2234
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:
 */
      __pyx_t_17 = __pyx_v_self->center_right_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":2235
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):             # <<<<<<<<<<<<<<
 *                     if not point < values[i]:
 *                         break
 */
      for (__pyx_t_11 = (__pyx_v_self->n_center - 1); __pyx_t_11 > -1L; __pyx_t_11-=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":2236
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_20 = __pyx_v_i;
        __pyx_t_14 = ((!((__pyx_v_point < (*((__pyx_t_5numpy_int32_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_20 * __pyx_v_values.strides[0]) )))) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":2237
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:
 */
          goto __pyx_L13_break;

          /* "pandas/_libs/intervaltree.pxi":2236
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":2238
 *                     if not point < values[i]:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if self.right_node.min_left < point:
 *                     self.right_node.query(result, point)
 */
        __pyx_t_21 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_21 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L13_break:;

      /* "pandas/_libs/intervaltree.pxi":2239
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      __pyx_t_14 = ((__pyx_v_self->right_node->min_left < __pyx_v_point) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":2240
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:
 *                     self.right_node.query(result, point)             # <<<<<<<<<<<<<<
 *             else:
 *                 result.extend(self.center_left_indices)
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode *)__pyx_v_self->right_node->__pyx_vtab)->__pyx_fuse_4query(__pyx_v_self->right_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2240, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":2239
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":2232
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":2242
 *                     self.right_node.query(result, point)
 *             else:
 *                 result.extend(self.center_left_indices)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
    /*else*/ {
      __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->extend(__pyx_v_result, __pyx_v_self->center_left_indices); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2242, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_L7:;
  }
  __pyx_L3:;

  /* "pandas/_libs/intervaltree.pxi":2204
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __PYX_XDEC_MEMVIEW(&__pyx_t_16, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_17, 1);
  __Pyx_AddTraceback("pandas._libs.interval.Int32ClosedNeitherIntervalNode.query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_indices, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_values, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_17__pyx_fuse_4query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_4__pyx_mdef_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_17__pyx_fuse_4query = {"__pyx_fuse_4query", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_17__pyx_fuse_4query, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_2query};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_17__pyx_fuse_4query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result = 0;
  __pyx_t_5numpy_uint64_t __pyx_v_point;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fuse_4query (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_result,&__pyx_n_s_point,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_result)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_point)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_4query", 1, 2, 2, 1); __PYX_ERR(1, 2204, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_4query") < 0)) __PYX_ERR(1, 2204, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_result = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)values[0]);
    __pyx_v_point = __Pyx_PyInt_As_npy_uint64(values[1]); if (unlikely((__pyx_v_point == ((npy_uint64)-1)) && PyErr_Occurred())) __PYX_ERR(1, 2204, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_4query", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 2204, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Int32ClosedNeitherIntervalNode.__pyx_fuse_4query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_result), __pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector, 1, "result", 0))) __PYX_ERR(1, 2204, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_16__pyx_fuse_4query(((struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode *)__pyx_v_self), __pyx_v_result, __pyx_v_point);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_16__pyx_fuse_4query(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_uint64_t __pyx_v_point) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__pyx_fuse_4query", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_fuse_4__pyx_f_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_query(__pyx_v_self, __pyx_v_result, __pyx_v_point, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2204, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Int32ClosedNeitherIntervalNode.__pyx_fuse_4query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":2244
 *                 result.extend(self.center_left_indices)
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         if self.is_leaf_node:
 *             return ('<Int32ClosedNeitherIntervalNode: '
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_5__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_5__repr__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_4__repr__(((struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_4__repr__(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode *__pyx_v_self) {
  PyObject *__pyx_v_n_left = NULL;
  PyObject *__pyx_v_n_right = NULL;
  PyObject *__pyx_v_n_center = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  Py_UCS4 __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "pandas/_libs/intervaltree.pxi":2245
 * 
 *     def __repr__(self):
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             return ('<Int32ClosedNeitherIntervalNode: '
 *                     '%s elements (terminal)>' % self.n_elements)
 */
  __pyx_t_1 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_1) {

    /* "pandas/_libs/intervaltree.pxi":2246
 *     def __repr__(self):
 *         if self.is_leaf_node:
 *             return ('<Int32ClosedNeitherIntervalNode: '             # <<<<<<<<<<<<<<
 *                     '%s elements (terminal)>' % self.n_elements)
 *         else:
 */
    __Pyx_XDECREF(__pyx_r);

    /* "pandas/_libs/intervaltree.pxi":2247
 *         if self.is_leaf_node:
 *             return ('<Int32ClosedNeitherIntervalNode: '
 *                     '%s elements (terminal)>' % self.n_elements)             # <<<<<<<<<<<<<<
 *         else:
 *             n_left = self.left_node.n_elements
 */
    __pyx_t_2 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_elements); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2247, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PyUnicode_Format(__pyx_kp_u_Int32ClosedNeitherIntervalNode, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2247, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pandas/_libs/intervaltree.pxi":2245
 * 
 *     def __repr__(self):
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             return ('<Int32ClosedNeitherIntervalNode: '
 *                     '%s elements (terminal)>' % self.n_elements)
 */
  }

  /* "pandas/_libs/intervaltree.pxi":2249
 *                     '%s elements (terminal)>' % self.n_elements)
 *         else:
 *             n_left = self.left_node.n_elements             # <<<<<<<<<<<<<<
 *             n_right = self.right_node.n_elements
 *             n_center = self.n_elements - n_left - n_right
 */
  /*else*/ {
    __pyx_t_3 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->left_node->n_elements); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2249, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_v_n_left = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "pandas/_libs/intervaltree.pxi":2250
 *         else:
 *             n_left = self.left_node.n_elements
 *             n_right = self.right_node.n_elements             # <<<<<<<<<<<<<<
 *             n_center = self.n_elements - n_left - n_right
 *             return ('<Int32ClosedNeitherIntervalNode: '
 */
    __pyx_t_3 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->right_node->n_elements); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2250, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_v_n_right = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "pandas/_libs/intervaltree.pxi":2251
 *             n_left = self.left_node.n_elements
 *             n_right = self.right_node.n_elements
 *             n_center = self.n_elements - n_left - n_right             # <<<<<<<<<<<<<<
 *             return ('<Int32ClosedNeitherIntervalNode: '
 *                     'pivot %s, %s elements (%s left, %s right, %s '
 */
    __pyx_t_3 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_elements); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2251, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = PyNumber_Subtract(__pyx_t_3, __pyx_v_n_left); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2251, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyNumber_Subtract(__pyx_t_2, __pyx_v_n_right); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2251, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_n_center = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "pandas/_libs/intervaltree.pxi":2252
 *             n_right = self.right_node.n_elements
 *             n_center = self.n_elements - n_left - n_right
 *             return ('<Int32ClosedNeitherIntervalNode: '             # <<<<<<<<<<<<<<
 *                     'pivot %s, %s elements (%s left, %s right, %s '
 *                     'overlapping)>' % (self.pivot, self.n_elements,
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = PyTuple_New(11); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2252, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = 0;
    __pyx_t_5 = 127;
    __Pyx_INCREF(__pyx_kp_u_Int32ClosedNeitherIntervalNode_2);
    __pyx_t_4 += 39;
    __Pyx_GIVEREF(__pyx_kp_u_Int32ClosedNeitherIntervalNode_2);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_kp_u_Int32ClosedNeitherIntervalNode_2);

    /* "pandas/_libs/intervaltree.pxi":2254
 *             return ('<Int32ClosedNeitherIntervalNode: '
 *                     'pivot %s, %s elements (%s left, %s right, %s '
 *                     'overlapping)>' % (self.pivot, self.n_elements,             # <<<<<<<<<<<<<<
 *                                        n_left, n_right, n_center))
 * 
 */
    __pyx_t_2 = __Pyx_PyInt_From_npy_int32(__pyx_v_self->pivot); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2254, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_t_2), __pyx_empty_unicode); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 2254, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_5 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_6) > __pyx_t_5) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_6) : __pyx_t_5;
    __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_6);
    __pyx_t_6 = 0;
    __Pyx_INCREF(__pyx_kp_u__18);
    __pyx_t_4 += 2;
    __Pyx_GIVEREF(__pyx_kp_u__18);
    PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_kp_u__18);
    __pyx_t_6 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_elements); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 2254, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_2 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_t_6), __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2254, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_5 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) > __pyx_t_5) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) : __pyx_t_5;
    __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_3, 3, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_elements);
    __pyx_t_4 += 11;
    __Pyx_GIVEREF(__pyx_kp_u_elements);
    PyTuple_SET_ITEM(__pyx_t_3, 4, __pyx_kp_u_elements);

    /* "pandas/_libs/intervaltree.pxi":2255
 *                     'pivot %s, %s elements (%s left, %s right, %s '
 *                     'overlapping)>' % (self.pivot, self.n_elements,
 *                                        n_left, n_right, n_center))             # <<<<<<<<<<<<<<
 * 
 *     def counts(self):
 */
    __pyx_t_2 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_v_n_left), __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2255, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) > __pyx_t_5) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) : __pyx_t_5;
    __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_3, 5, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_left_2);
    __pyx_t_4 += 7;
    __Pyx_GIVEREF(__pyx_kp_u_left_2);
    PyTuple_SET_ITEM(__pyx_t_3, 6, __pyx_kp_u_left_2);
    __pyx_t_2 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_v_n_right), __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2255, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) > __pyx_t_5) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) : __pyx_t_5;
    __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_3, 7, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_right_2);
    __pyx_t_4 += 8;
    __Pyx_GIVEREF(__pyx_kp_u_right_2);
    PyTuple_SET_ITEM(__pyx_t_3, 8, __pyx_kp_u_right_2);
    __pyx_t_2 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_v_n_center), __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2255, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) > __pyx_t_5) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) : __pyx_t_5;
    __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_3, 9, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_overlapping);
    __pyx_t_4 += 14;
    __Pyx_GIVEREF(__pyx_kp_u_overlapping);
    PyTuple_SET_ITEM(__pyx_t_3, 10, __pyx_kp_u_overlapping);

    /* "pandas/_libs/intervaltree.pxi":2252
 *             n_right = self.right_node.n_elements
 *             n_center = self.n_elements - n_left - n_right
 *             return ('<Int32ClosedNeitherIntervalNode: '             # <<<<<<<<<<<<<<
 *                     'pivot %s, %s elements (%s left, %s right, %s '
 *                     'overlapping)>' % (self.pivot, self.n_elements,
 */
    __pyx_t_2 = __Pyx_PyUnicode_Join(__pyx_t_3, 11, __pyx_t_4, __pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2252, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;
  }

  /* "pandas/_libs/intervaltree.pxi":2244
 *                 result.extend(self.center_left_indices)
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         if self.is_leaf_node:
 *             return ('<Int32ClosedNeitherIntervalNode: '
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pandas._libs.interval.Int32ClosedNeitherIntervalNode.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_n_left);
  __Pyx_XDECREF(__pyx_v_n_right);
  __Pyx_XDECREF(__pyx_v_n_center);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":2257
 *                                        n_left, n_right, n_center))
 * 
 *     def counts(self):             # <<<<<<<<<<<<<<
 *         """
 *         Inspect counts on this node
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_7counts(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_6counts[] = "\n        Inspect counts on this node\n        useful for debugging purposes\n        ";
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_7counts(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("counts (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_6counts(((struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_6counts(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode *__pyx_v_self) {
  Py_ssize_t __pyx_v_m;
  PyObject *__pyx_v_l = NULL;
  PyObject *__pyx_v_r = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  size_t __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("counts", 0);

  /* "pandas/_libs/intervaltree.pxi":2262
 *         useful for debugging purposes
 *         """
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             return self.n_elements
 *         else:
 */
  __pyx_t_1 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_1) {

    /* "pandas/_libs/intervaltree.pxi":2263
 *         """
 *         if self.is_leaf_node:
 *             return self.n_elements             # <<<<<<<<<<<<<<
 *         else:
 *             m = len(self.center_left_values)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_elements); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2263, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "pandas/_libs/intervaltree.pxi":2262
 *         useful for debugging purposes
 *         """
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             return self.n_elements
 *         else:
 */
  }

  /* "pandas/_libs/intervaltree.pxi":2265
 *             return self.n_elements
 *         else:
 *             m = len(self.center_left_values)             # <<<<<<<<<<<<<<
 *             l = self.left_node.counts()
 *             r = self.right_node.counts()
 */
  /*else*/ {
    if (unlikely(!__pyx_v_self->center_left_values.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(1, 2265, __pyx_L1_error)}
    __pyx_t_3 = __Pyx_MemoryView_Len(__pyx_v_self->center_left_values); 
    __pyx_v_m = __pyx_t_3;

    /* "pandas/_libs/intervaltree.pxi":2266
 *         else:
 *             m = len(self.center_left_values)
 *             l = self.left_node.counts()             # <<<<<<<<<<<<<<
 *             r = self.right_node.counts()
 *             return (m, (l, r))
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->left_node), __pyx_n_s_counts); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 2266, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2266, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_l = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "pandas/_libs/intervaltree.pxi":2267
 *             m = len(self.center_left_values)
 *             l = self.left_node.counts()
 *             r = self.right_node.counts()             # <<<<<<<<<<<<<<
 *             return (m, (l, r))
 * 
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->right_node), __pyx_n_s_counts); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 2267, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2267, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_r = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "pandas/_libs/intervaltree.pxi":2268
 *             l = self.left_node.counts()
 *             r = self.right_node.counts()
 *             return (m, (l, r))             # <<<<<<<<<<<<<<
 * 
 * NODE_CLASSES['int32',
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = PyInt_FromSsize_t(__pyx_v_m); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2268, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 2268, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_v_l);
    __Pyx_GIVEREF(__pyx_v_l);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_l);
    __Pyx_INCREF(__pyx_v_r);
    __Pyx_GIVEREF(__pyx_v_r);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_v_r);
    __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 2268, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_4);
    __pyx_t_2 = 0;
    __pyx_t_4 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;
  }

  /* "pandas/_libs/intervaltree.pxi":2257
 *                                        n_left, n_right, n_center))
 * 
 *     def counts(self):             # <<<<<<<<<<<<<<
 *         """
 *         Inspect counts on this node
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pandas._libs.interval.Int32ClosedNeitherIntervalNode.counts", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_l);
  __Pyx_XDECREF(__pyx_v_r);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":2114
 *         int64_t[:] center_left_indices, center_right_indices, indices
 *         int32_t min_left, max_right
 *         readonly int32_t pivot             # <<<<<<<<<<<<<<
 *         readonly int64_t n_elements, n_center, leaf_size
 *         readonly bint is_leaf_node
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_5pivot_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_5pivot_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_5pivot___get__(((struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_5pivot___get__(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_npy_int32(__pyx_v_self->pivot); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2114, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Int32ClosedNeitherIntervalNode.pivot.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":2115
 *         int32_t min_left, max_right
 *         readonly int32_t pivot
 *         readonly int64_t n_elements, n_center, leaf_size             # <<<<<<<<<<<<<<
 *         readonly bint is_leaf_node
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_10n_elements_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_10n_elements_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_10n_elements___get__(((struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_10n_elements___get__(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_elements); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2115, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Int32ClosedNeitherIntervalNode.n_elements.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_8n_center_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_8n_center_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_8n_center___get__(((struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_8n_center___get__(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_center); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2115, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Int32ClosedNeitherIntervalNode.n_center.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_9leaf_size_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_9leaf_size_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_9leaf_size___get__(((struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_9leaf_size___get__(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->leaf_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2115, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Int32ClosedNeitherIntervalNode.leaf_size.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":2116
 *         readonly int32_t pivot
 *         readonly int64_t n_elements, n_center, leaf_size
 *         readonly bint is_leaf_node             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self,
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_12is_leaf_node_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_12is_leaf_node_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_12is_leaf_node___get__(((struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_12is_leaf_node___get__(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->is_leaf_node); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2116, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Int32ClosedNeitherIntervalNode.is_leaf_node.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_21__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_21__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_20__reduce_cython__(((struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_20__reduce_cython__(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  int __pyx_t_16;
  int __pyx_t_17;
  int __pyx_t_18;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.center_left_indices, self.center_left_values, self.center_right_indices, self.center_right_values, self.indices, self.is_leaf_node, self.leaf_size, self.left, self.left_node, self.max_right, self.min_left, self.n_center, self.n_elements, self.pivot, self.right, self.right_node)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  if (unlikely(!__pyx_v_self->center_left_indices.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_1 = __pyx_memoryview_fromslice(__pyx_v_self->center_left_indices, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_int64_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_int64_t, 0);; if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (unlikely(!__pyx_v_self->center_left_values.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_2 = __pyx_memoryview_fromslice(__pyx_v_self->center_left_values, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_int32_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_int32_t, 0);; if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (unlikely(!__pyx_v_self->center_right_indices.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_3 = __pyx_memoryview_fromslice(__pyx_v_self->center_right_indices, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_int64_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_int64_t, 0);; if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely(!__pyx_v_self->center_right_values.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_4 = __pyx_memoryview_fromslice(__pyx_v_self->center_right_values, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_int32_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_int32_t, 0);; if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (unlikely(!__pyx_v_self->indices.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_5 = __pyx_memoryview_fromslice(__pyx_v_self->indices, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_int64_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_int64_t, 0);; if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyBool_FromLong(__pyx_v_self->is_leaf_node); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->leaf_size); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (unlikely(!__pyx_v_self->left.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_8 = __pyx_memoryview_fromslice(__pyx_v_self->left, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_int32_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_int32_t, 0);; if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_PyInt_From_npy_int32(__pyx_v_self->max_right); if (unlikely(!__pyx_t_9)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_10 = __Pyx_PyInt_From_npy_int32(__pyx_v_self->min_left); if (unlikely(!__pyx_t_10)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_11 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_center); if (unlikely(!__pyx_t_11)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_12 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_elements); if (unlikely(!__pyx_t_12)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __pyx_t_13 = __Pyx_PyInt_From_npy_int32(__pyx_v_self->pivot); if (unlikely(!__pyx_t_13)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  if (unlikely(!__pyx_v_self->right.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_14 = __pyx_memoryview_fromslice(__pyx_v_self->right, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_int32_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_int32_t, 0);; if (unlikely(!__pyx_t_14)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __pyx_t_15 = PyTuple_New(16); if (unlikely(!__pyx_t_15)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_15, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_15, 1, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_15, 2, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_15, 3, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_15, 4, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_15, 5, __pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_15, 6, __pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_15, 7, __pyx_t_8);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->left_node));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self->left_node));
  PyTuple_SET_ITEM(__pyx_t_15, 8, ((PyObject *)__pyx_v_self->left_node));
  __Pyx_GIVEREF(__pyx_t_9);
  PyTuple_SET_ITEM(__pyx_t_15, 9, __pyx_t_9);
  __Pyx_GIVEREF(__pyx_t_10);
  PyTuple_SET_ITEM(__pyx_t_15, 10, __pyx_t_10);
  __Pyx_GIVEREF(__pyx_t_11);
  PyTuple_SET_ITEM(__pyx_t_15, 11, __pyx_t_11);
  __Pyx_GIVEREF(__pyx_t_12);
  PyTuple_SET_ITEM(__pyx_t_15, 12, __pyx_t_12);
  __Pyx_GIVEREF(__pyx_t_13);
  PyTuple_SET_ITEM(__pyx_t_15, 13, __pyx_t_13);
  __Pyx_GIVEREF(__pyx_t_14);
  PyTuple_SET_ITEM(__pyx_t_15, 14, __pyx_t_14);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->right_node));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self->right_node));
  PyTuple_SET_ITEM(__pyx_t_15, 15, ((PyObject *)__pyx_v_self->right_node));
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  __pyx_t_6 = 0;
  __pyx_t_7 = 0;
  __pyx_t_8 = 0;
  __pyx_t_9 = 0;
  __pyx_t_10 = 0;
  __pyx_t_11 = 0;
  __pyx_t_12 = 0;
  __pyx_t_13 = 0;
  __pyx_t_14 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_15);
  __pyx_t_15 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.center_left_indices, self.center_left_values, self.center_right_indices, self.center_right_values, self.indices, self.is_leaf_node, self.leaf_size, self.left, self.left_node, self.max_right, self.min_left, self.n_center, self.n_elements, self.pivot, self.right, self.right_node)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_15 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_15)) __PYX_ERR(2, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __pyx_v__dict = __pyx_t_15;
  __pyx_t_15 = 0;

  /* "(tree fragment)":7
 *     state = (self.center_left_indices, self.center_left_values, self.center_right_indices, self.center_right_values, self.indices, self.is_leaf_node, self.leaf_size, self.left, self.left_node, self.max_right, self.min_left, self.n_center, self.n_elements, self.pivot, self.right, self.right_node)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_16 = (__pyx_v__dict != Py_None);
  __pyx_t_17 = (__pyx_t_16 != 0);
  if (__pyx_t_17) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_15 = PyTuple_New(1); if (unlikely(!__pyx_t_15)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    PyTuple_SET_ITEM(__pyx_t_15, 0, __pyx_v__dict);
    __pyx_t_14 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_15); if (unlikely(!__pyx_t_14)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_14));
    __pyx_t_14 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.left_node is not None or self.right_node is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.center_left_indices, self.center_left_values, self.center_right_indices, self.center_right_values, self.indices, self.is_leaf_node, self.leaf_size, self.left, self.left_node, self.max_right, self.min_left, self.n_center, self.n_elements, self.pivot, self.right, self.right_node)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.left_node is not None or self.right_node is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_Int32ClosedNeitherIntervalNode, (type(self), 0x7a104e5, None), state
 */
  /*else*/ {
    __pyx_t_16 = (((PyObject *)__pyx_v_self->left_node) != Py_None);
    __pyx_t_18 = (__pyx_t_16 != 0);
    if (!__pyx_t_18) {
    } else {
      __pyx_t_17 = __pyx_t_18;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_18 = (((PyObject *)__pyx_v_self->right_node) != Py_None);
    __pyx_t_16 = (__pyx_t_18 != 0);
    __pyx_t_17 = __pyx_t_16;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_17;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.left_node is not None or self.right_node is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Int32ClosedNeitherIntervalNode, (type(self), 0x7a104e5, None), state
 *     else:
 */
  __pyx_t_17 = (__pyx_v_use_setstate != 0);
  if (__pyx_t_17) {

    /* "(tree fragment)":13
 *         use_setstate = self.left_node is not None or self.right_node is not None
 *     if use_setstate:
 *         return __pyx_unpickle_Int32ClosedNeitherIntervalNode, (type(self), 0x7a104e5, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_Int32ClosedNeitherIntervalNode, (type(self), 0x7a104e5, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_14, __pyx_n_s_pyx_unpickle_Int32ClosedNeithe); if (unlikely(!__pyx_t_14)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __pyx_t_15 = PyTuple_New(3); if (unlikely(!__pyx_t_15)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_15, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_127993061);
    __Pyx_GIVEREF(__pyx_int_127993061);
    PyTuple_SET_ITEM(__pyx_t_15, 1, __pyx_int_127993061);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_15, 2, Py_None);
    __pyx_t_13 = PyTuple_New(3); if (unlikely(!__pyx_t_13)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __Pyx_GIVEREF(__pyx_t_14);
    PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_t_14);
    __Pyx_GIVEREF(__pyx_t_15);
    PyTuple_SET_ITEM(__pyx_t_13, 1, __pyx_t_15);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_13, 2, __pyx_v_state);
    __pyx_t_14 = 0;
    __pyx_t_15 = 0;
    __pyx_r = __pyx_t_13;
    __pyx_t_13 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.left_node is not None or self.right_node is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Int32ClosedNeitherIntervalNode, (type(self), 0x7a104e5, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_Int32ClosedNeitherIntervalNode, (type(self), 0x7a104e5, None), state
 *     else:
 *         return __pyx_unpickle_Int32ClosedNeitherIntervalNode, (type(self), 0x7a104e5, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Int32ClosedNeitherIntervalNode__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_13, __pyx_n_s_pyx_unpickle_Int32ClosedNeithe); if (unlikely(!__pyx_t_13)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __pyx_t_15 = PyTuple_New(3); if (unlikely(!__pyx_t_15)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_15, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_127993061);
    __Pyx_GIVEREF(__pyx_int_127993061);
    PyTuple_SET_ITEM(__pyx_t_15, 1, __pyx_int_127993061);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_15, 2, __pyx_v_state);
    __pyx_t_14 = PyTuple_New(2); if (unlikely(!__pyx_t_14)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_GIVEREF(__pyx_t_13);
    PyTuple_SET_ITEM(__pyx_t_14, 0, __pyx_t_13);
    __Pyx_GIVEREF(__pyx_t_15);
    PyTuple_SET_ITEM(__pyx_t_14, 1, __pyx_t_15);
    __pyx_t_13 = 0;
    __pyx_t_15 = 0;
    __pyx_r = __pyx_t_14;
    __pyx_t_14 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_AddTraceback("pandas._libs.interval.Int32ClosedNeitherIntervalNode.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Int32ClosedNeitherIntervalNode, (type(self), 0x7a104e5, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Int32ClosedNeitherIntervalNode__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_23__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_23__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_22__setstate_cython__(((struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_22__setstate_cython__(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_Int32ClosedNeitherIntervalNode, (type(self), 0x7a104e5, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Int32ClosedNeitherIntervalNode__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(2, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_6pandas_5_libs_8interval___pyx_unpickle_Int32ClosedNeitherIntervalNode__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Int32ClosedNeitherIntervalNode, (type(self), 0x7a104e5, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Int32ClosedNeitherIntervalNode__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Int32ClosedNeitherIntervalNode.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":2288
 *         readonly bint is_leaf_node
 * 
 *     def __init__(self,             # <<<<<<<<<<<<<<
 *                  ndarray[int64_t, ndim=1] left,
 *                  ndarray[int64_t, ndim=1] right,
 */

/* Python wrapper */
static int __pyx_pw_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyArrayObject *__pyx_v_left = 0;
  PyArrayObject *__pyx_v_right = 0;
  PyArrayObject *__pyx_v_indices = 0;
  __pyx_t_5numpy_int64_t __pyx_v_leaf_size;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_left,&__pyx_n_s_right,&__pyx_n_s_indices,&__pyx_n_s_leaf_size,0};
    PyObject* values[4] = {0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_left)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_right)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 4, 4, 1); __PYX_ERR(1, 2288, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_indices)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 4, 4, 2); __PYX_ERR(1, 2288, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_leaf_size)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 4, 4, 3); __PYX_ERR(1, 2288, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(1, 2288, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 4) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
    }
    __pyx_v_left = ((PyArrayObject *)values[0]);
    __pyx_v_right = ((PyArrayObject *)values[1]);
    __pyx_v_indices = ((PyArrayObject *)values[2]);
    __pyx_v_leaf_size = __Pyx_PyInt_As_npy_int64(values[3]); if (unlikely((__pyx_v_leaf_size == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(1, 2292, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 4, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 2288, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Int64ClosedLeftIntervalNode.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_left), __pyx_ptype_5numpy_ndarray, 1, "left", 0))) __PYX_ERR(1, 2289, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_right), __pyx_ptype_5numpy_ndarray, 1, "right", 0))) __PYX_ERR(1, 2290, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_indices), __pyx_ptype_5numpy_ndarray, 1, "indices", 0))) __PYX_ERR(1, 2291, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode___init__(((struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode *)__pyx_v_self), __pyx_v_left, __pyx_v_right, __pyx_v_indices, __pyx_v_leaf_size);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode___init__(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode *__pyx_v_self, PyArrayObject *__pyx_v_left, PyArrayObject *__pyx_v_right, PyArrayObject *__pyx_v_indices, __pyx_t_5numpy_int64_t __pyx_v_leaf_size) {
  PyObject *__pyx_v_left_set = NULL;
  PyObject *__pyx_v_right_set = NULL;
  PyObject *__pyx_v_center_set = NULL;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_indices;
  __Pyx_Buffer __pyx_pybuffer_indices;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_left;
  __Pyx_Buffer __pyx_pybuffer_left;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_right;
  __Pyx_Buffer __pyx_pybuffer_right;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __pyx_t_5numpy_int64_t __pyx_t_6;
  __Pyx_memviewslice __pyx_t_7 = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  __Pyx_memviewslice __pyx_t_10 = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *(*__pyx_t_11)(PyObject *);
  size_t __pyx_t_12;
  __Pyx_RefNannySetupContext("__init__", 0);
  __pyx_pybuffer_left.pybuffer.buf = NULL;
  __pyx_pybuffer_left.refcount = 0;
  __pyx_pybuffernd_left.data = NULL;
  __pyx_pybuffernd_left.rcbuffer = &__pyx_pybuffer_left;
  __pyx_pybuffer_right.pybuffer.buf = NULL;
  __pyx_pybuffer_right.refcount = 0;
  __pyx_pybuffernd_right.data = NULL;
  __pyx_pybuffernd_right.rcbuffer = &__pyx_pybuffer_right;
  __pyx_pybuffer_indices.pybuffer.buf = NULL;
  __pyx_pybuffer_indices.refcount = 0;
  __pyx_pybuffernd_indices.data = NULL;
  __pyx_pybuffernd_indices.rcbuffer = &__pyx_pybuffer_indices;
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_left.rcbuffer->pybuffer, (PyObject*)__pyx_v_left, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 2288, __pyx_L1_error)
  }
  __pyx_pybuffernd_left.diminfo[0].strides = __pyx_pybuffernd_left.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_left.diminfo[0].shape = __pyx_pybuffernd_left.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_right.rcbuffer->pybuffer, (PyObject*)__pyx_v_right, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 2288, __pyx_L1_error)
  }
  __pyx_pybuffernd_right.diminfo[0].strides = __pyx_pybuffernd_right.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_right.diminfo[0].shape = __pyx_pybuffernd_right.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_indices.rcbuffer->pybuffer, (PyObject*)__pyx_v_indices, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 2288, __pyx_L1_error)
  }
  __pyx_pybuffernd_indices.diminfo[0].strides = __pyx_pybuffernd_indices.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_indices.diminfo[0].shape = __pyx_pybuffernd_indices.rcbuffer->pybuffer.shape[0];

  /* "pandas/_libs/intervaltree.pxi":2294
 *                  int64_t leaf_size):
 * 
 *         self.n_elements = len(left)             # <<<<<<<<<<<<<<
 *         self.leaf_size = leaf_size
 * 
 */
  __pyx_t_1 = PyObject_Length(((PyObject *)__pyx_v_left)); if (unlikely(__pyx_t_1 == ((Py_ssize_t)-1))) __PYX_ERR(1, 2294, __pyx_L1_error)
  __pyx_v_self->n_elements = __pyx_t_1;

  /* "pandas/_libs/intervaltree.pxi":2295
 * 
 *         self.n_elements = len(left)
 *         self.leaf_size = leaf_size             # <<<<<<<<<<<<<<
 * 
 *         # min_left and min_right are used to speed-up query by skipping
 */
  __pyx_v_self->leaf_size = __pyx_v_leaf_size;

  /* "pandas/_libs/intervaltree.pxi":2300
 *         # query on sub-nodes. If this node has size 0, query is cheap,
 *         # so these values don't matter.
 *         if left.size > 0:             # <<<<<<<<<<<<<<
 *             self.min_left = left.min()
 *             self.max_right = right.max()
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_left), __pyx_n_s_size); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2300, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyObject_RichCompare(__pyx_t_2, __pyx_int_0, Py_GT); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2300, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(1, 2300, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_4) {

    /* "pandas/_libs/intervaltree.pxi":2301
 *         # so these values don't matter.
 *         if left.size > 0:
 *             self.min_left = left.min()             # <<<<<<<<<<<<<<
 *             self.max_right = right.max()
 *         else:
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_left), __pyx_n_s_min); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2301, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2301, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_6 = __Pyx_PyInt_As_npy_int64(__pyx_t_3); if (unlikely((__pyx_t_6 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(1, 2301, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_self->min_left = __pyx_t_6;

    /* "pandas/_libs/intervaltree.pxi":2302
 *         if left.size > 0:
 *             self.min_left = left.min()
 *             self.max_right = right.max()             # <<<<<<<<<<<<<<
 *         else:
 *             self.min_left = 0
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_right), __pyx_n_s_max); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2302, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2302, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_6 = __Pyx_PyInt_As_npy_int64(__pyx_t_3); if (unlikely((__pyx_t_6 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(1, 2302, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_self->max_right = __pyx_t_6;

    /* "pandas/_libs/intervaltree.pxi":2300
 *         # query on sub-nodes. If this node has size 0, query is cheap,
 *         # so these values don't matter.
 *         if left.size > 0:             # <<<<<<<<<<<<<<
 *             self.min_left = left.min()
 *             self.max_right = right.max()
 */
    goto __pyx_L3;
  }

  /* "pandas/_libs/intervaltree.pxi":2304
 *             self.max_right = right.max()
 *         else:
 *             self.min_left = 0             # <<<<<<<<<<<<<<
 *             self.max_right = 0
 * 
 */
  /*else*/ {
    __pyx_v_self->min_left = 0;

    /* "pandas/_libs/intervaltree.pxi":2305
 *         else:
 *             self.min_left = 0
 *             self.max_right = 0             # <<<<<<<<<<<<<<
 * 
 *         if self.n_elements <= leaf_size:
 */
    __pyx_v_self->max_right = 0;
  }
  __pyx_L3:;

  /* "pandas/_libs/intervaltree.pxi":2307
 *             self.max_right = 0
 * 
 *         if self.n_elements <= leaf_size:             # <<<<<<<<<<<<<<
 *             # make this a terminal (leaf) node
 *             self.is_leaf_node = True
 */
  __pyx_t_4 = ((__pyx_v_self->n_elements <= __pyx_v_leaf_size) != 0);
  if (__pyx_t_4) {

    /* "pandas/_libs/intervaltree.pxi":2309
 *         if self.n_elements <= leaf_size:
 *             # make this a terminal (leaf) node
 *             self.is_leaf_node = True             # <<<<<<<<<<<<<<
 *             self.left = left
 *             self.right = right
 */
    __pyx_v_self->is_leaf_node = 1;

    /* "pandas/_libs/intervaltree.pxi":2310
 *             # make this a terminal (leaf) node
 *             self.is_leaf_node = True
 *             self.left = left             # <<<<<<<<<<<<<<
 *             self.right = right
 *             self.indices = indices
 */
    __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(((PyObject *)__pyx_v_left), PyBUF_WRITABLE); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(1, 2310, __pyx_L1_error)
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->left, 0);
    __pyx_v_self->left = __pyx_t_7;
    __pyx_t_7.memview = NULL;
    __pyx_t_7.data = NULL;

    /* "pandas/_libs/intervaltree.pxi":2311
 *             self.is_leaf_node = True
 *             self.left = left
 *             self.right = right             # <<<<<<<<<<<<<<
 *             self.indices = indices
 *             self.n_center = 0
 */
    __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(((PyObject *)__pyx_v_right), PyBUF_WRITABLE); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(1, 2311, __pyx_L1_error)
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->right, 0);
    __pyx_v_self->right = __pyx_t_7;
    __pyx_t_7.memview = NULL;
    __pyx_t_7.data = NULL;

    /* "pandas/_libs/intervaltree.pxi":2312
 *             self.left = left
 *             self.right = right
 *             self.indices = indices             # <<<<<<<<<<<<<<
 *             self.n_center = 0
 *         else:
 */
    __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(((PyObject *)__pyx_v_indices), PyBUF_WRITABLE); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(1, 2312, __pyx_L1_error)
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->indices, 0);
    __pyx_v_self->indices = __pyx_t_7;
    __pyx_t_7.memview = NULL;
    __pyx_t_7.data = NULL;

    /* "pandas/_libs/intervaltree.pxi":2313
 *             self.right = right
 *             self.indices = indices
 *             self.n_center = 0             # <<<<<<<<<<<<<<
 *         else:
 *             # calculate a pivot so we can create child nodes
 */
    __pyx_v_self->n_center = 0;

    /* "pandas/_libs/intervaltree.pxi":2307
 *             self.max_right = 0
 * 
 *         if self.n_elements <= leaf_size:             # <<<<<<<<<<<<<<
 *             # make this a terminal (leaf) node
 *             self.is_leaf_node = True
 */
    goto __pyx_L4;
  }

  /* "pandas/_libs/intervaltree.pxi":2316
 *         else:
 *             # calculate a pivot so we can create child nodes
 *             self.is_leaf_node = False             # <<<<<<<<<<<<<<
 *             self.pivot = np.median(left / 2 + right / 2)
 *             left_set, right_set, center_set = self.classify_intervals(
 */
  /*else*/ {
    __pyx_v_self->is_leaf_node = 0;

    /* "pandas/_libs/intervaltree.pxi":2317
 *             # calculate a pivot so we can create child nodes
 *             self.is_leaf_node = False
 *             self.pivot = np.median(left / 2 + right / 2)             # <<<<<<<<<<<<<<
 *             left_set, right_set, center_set = self.classify_intervals(
 *                 left, right)
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2317, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_median); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 2317, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyNumber_Divide(((PyObject *)__pyx_v_left), __pyx_int_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2317, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_8 = __Pyx_PyNumber_Divide(((PyObject *)__pyx_v_right), __pyx_int_2); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 2317, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_9 = PyNumber_Add(__pyx_t_2, __pyx_t_8); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 2317, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_3 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_8, __pyx_t_9) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_9);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2317, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_6 = __Pyx_PyInt_As_npy_int64(__pyx_t_3); if (unlikely((__pyx_t_6 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(1, 2317, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_self->pivot = __pyx_t_6;

    /* "pandas/_libs/intervaltree.pxi":2319
 *             self.pivot = np.median(left / 2 + right / 2)
 *             left_set, right_set, center_set = self.classify_intervals(
 *                 left, right)             # <<<<<<<<<<<<<<
 * 
 *             self.left_node = self.new_child_node(left, right,
 */
    __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(((PyObject *)__pyx_v_left), PyBUF_WRITABLE); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(1, 2319, __pyx_L1_error)
    __pyx_t_10 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(((PyObject *)__pyx_v_right), PyBUF_WRITABLE); if (unlikely(!__pyx_t_10.memview)) __PYX_ERR(1, 2319, __pyx_L1_error)

    /* "pandas/_libs/intervaltree.pxi":2318
 *             self.is_leaf_node = False
 *             self.pivot = np.median(left / 2 + right / 2)
 *             left_set, right_set, center_set = self.classify_intervals(             # <<<<<<<<<<<<<<
 *                 left, right)
 * 
 */
    __pyx_t_3 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode *)__pyx_v_self->__pyx_vtab)->classify_intervals(__pyx_v_self, __pyx_t_7, __pyx_t_10); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2318, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __PYX_XDEC_MEMVIEW(&__pyx_t_7, 1);
    __pyx_t_7.memview = NULL;
    __pyx_t_7.data = NULL;
    __PYX_XDEC_MEMVIEW(&__pyx_t_10, 1);
    __pyx_t_10.memview = NULL;
    __pyx_t_10.data = NULL;
    if ((likely(PyTuple_CheckExact(__pyx_t_3))) || (PyList_CheckExact(__pyx_t_3))) {
      PyObject* sequence = __pyx_t_3;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 3)) {
        if (size > 3) __Pyx_RaiseTooManyValuesError(3);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(1, 2318, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_5 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_9 = PyTuple_GET_ITEM(sequence, 1); 
        __pyx_t_8 = PyTuple_GET_ITEM(sequence, 2); 
      } else {
        __pyx_t_5 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_9 = PyList_GET_ITEM(sequence, 1); 
        __pyx_t_8 = PyList_GET_ITEM(sequence, 2); 
      }
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_9);
      __Pyx_INCREF(__pyx_t_8);
      #else
      __pyx_t_5 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 2318, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_9 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 2318, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_8 = PySequence_ITEM(sequence, 2); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 2318, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      #endif
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_2 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2318, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_11 = Py_TYPE(__pyx_t_2)->tp_iternext;
      index = 0; __pyx_t_5 = __pyx_t_11(__pyx_t_2); if (unlikely(!__pyx_t_5)) goto __pyx_L5_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_5);
      index = 1; __pyx_t_9 = __pyx_t_11(__pyx_t_2); if (unlikely(!__pyx_t_9)) goto __pyx_L5_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_9);
      index = 2; __pyx_t_8 = __pyx_t_11(__pyx_t_2); if (unlikely(!__pyx_t_8)) goto __pyx_L5_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_8);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_11(__pyx_t_2), 3) < 0) __PYX_ERR(1, 2318, __pyx_L1_error)
      __pyx_t_11 = NULL;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      goto __pyx_L6_unpacking_done;
      __pyx_L5_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_11 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(1, 2318, __pyx_L1_error)
      __pyx_L6_unpacking_done:;
    }
    __pyx_v_left_set = __pyx_t_5;
    __pyx_t_5 = 0;
    __pyx_v_right_set = __pyx_t_9;
    __pyx_t_9 = 0;
    __pyx_v_center_set = __pyx_t_8;
    __pyx_t_8 = 0;

    /* "pandas/_libs/intervaltree.pxi":2322
 * 
 *             self.left_node = self.new_child_node(left, right,
 *                                                  indices, left_set)             # <<<<<<<<<<<<<<
 *             self.right_node = self.new_child_node(left, right,
 *                                                   indices, right_set)
 */
    if (!(likely(((__pyx_v_left_set) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_left_set, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(1, 2322, __pyx_L1_error)

    /* "pandas/_libs/intervaltree.pxi":2321
 *                 left, right)
 * 
 *             self.left_node = self.new_child_node(left, right,             # <<<<<<<<<<<<<<
 *                                                  indices, left_set)
 *             self.right_node = self.new_child_node(left, right,
 */
    __pyx_t_3 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode *)__pyx_v_self->__pyx_vtab)->new_child_node(__pyx_v_self, ((PyArrayObject *)__pyx_v_left), ((PyArrayObject *)__pyx_v_right), ((PyArrayObject *)__pyx_v_indices), ((PyArrayObject *)__pyx_v_left_set)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2321, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode))))) __PYX_ERR(1, 2321, __pyx_L1_error)
    __Pyx_GIVEREF(__pyx_t_3);
    __Pyx_GOTREF(__pyx_v_self->left_node);
    __Pyx_DECREF(((PyObject *)__pyx_v_self->left_node));
    __pyx_v_self->left_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode *)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "pandas/_libs/intervaltree.pxi":2324
 *                                                  indices, left_set)
 *             self.right_node = self.new_child_node(left, right,
 *                                                   indices, right_set)             # <<<<<<<<<<<<<<
 * 
 *             self.center_left_values, self.center_left_indices = \
 */
    if (!(likely(((__pyx_v_right_set) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_right_set, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(1, 2324, __pyx_L1_error)

    /* "pandas/_libs/intervaltree.pxi":2323
 *             self.left_node = self.new_child_node(left, right,
 *                                                  indices, left_set)
 *             self.right_node = self.new_child_node(left, right,             # <<<<<<<<<<<<<<
 *                                                   indices, right_set)
 * 
 */
    __pyx_t_3 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode *)__pyx_v_self->__pyx_vtab)->new_child_node(__pyx_v_self, ((PyArrayObject *)__pyx_v_left), ((PyArrayObject *)__pyx_v_right), ((PyArrayObject *)__pyx_v_indices), ((PyArrayObject *)__pyx_v_right_set)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2323, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode))))) __PYX_ERR(1, 2323, __pyx_L1_error)
    __Pyx_GIVEREF(__pyx_t_3);
    __Pyx_GOTREF(__pyx_v_self->right_node);
    __Pyx_DECREF(((PyObject *)__pyx_v_self->right_node));
    __pyx_v_self->right_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode *)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "pandas/_libs/intervaltree.pxi":2327
 * 
 *             self.center_left_values, self.center_left_indices = \
 *                 sort_values_and_indices(left, indices, center_set)             # <<<<<<<<<<<<<<
 *             self.center_right_values, self.center_right_indices = \
 *                 sort_values_and_indices(right, indices, center_set)
 */
    __pyx_t_3 = __pyx_f_6pandas_5_libs_8interval_sort_values_and_indices(((PyObject *)__pyx_v_left), ((PyObject *)__pyx_v_indices), __pyx_v_center_set); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2327, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if ((likely(PyTuple_CheckExact(__pyx_t_3))) || (PyList_CheckExact(__pyx_t_3))) {
      PyObject* sequence = __pyx_t_3;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(1, 2326, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_8 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_9 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_8 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_9 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_8);
      __Pyx_INCREF(__pyx_t_9);
      #else
      __pyx_t_8 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 2326, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_9 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 2326, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      #endif
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_5 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 2326, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_11 = Py_TYPE(__pyx_t_5)->tp_iternext;
      index = 0; __pyx_t_8 = __pyx_t_11(__pyx_t_5); if (unlikely(!__pyx_t_8)) goto __pyx_L7_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_8);
      index = 1; __pyx_t_9 = __pyx_t_11(__pyx_t_5); if (unlikely(!__pyx_t_9)) goto __pyx_L7_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_9);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_11(__pyx_t_5), 2) < 0) __PYX_ERR(1, 2326, __pyx_L1_error)
      __pyx_t_11 = NULL;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      goto __pyx_L8_unpacking_done;
      __pyx_L7_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_11 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(1, 2326, __pyx_L1_error)
      __pyx_L8_unpacking_done:;
    }

    /* "pandas/_libs/intervaltree.pxi":2326
 *                                                   indices, right_set)
 * 
 *             self.center_left_values, self.center_left_indices = \             # <<<<<<<<<<<<<<
 *                 sort_values_and_indices(left, indices, center_set)
 *             self.center_right_values, self.center_right_indices = \
 */
    __pyx_t_10 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_t_8, PyBUF_WRITABLE); if (unlikely(!__pyx_t_10.memview)) __PYX_ERR(1, 2326, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_t_9, PyBUF_WRITABLE); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(1, 2326, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->center_left_values, 0);
    __pyx_v_self->center_left_values = __pyx_t_10;
    __pyx_t_10.memview = NULL;
    __pyx_t_10.data = NULL;
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->center_left_indices, 0);
    __pyx_v_self->center_left_indices = __pyx_t_7;
    __pyx_t_7.memview = NULL;
    __pyx_t_7.data = NULL;

    /* "pandas/_libs/intervaltree.pxi":2329
 *                 sort_values_and_indices(left, indices, center_set)
 *             self.center_right_values, self.center_right_indices = \
 *                 sort_values_and_indices(right, indices, center_set)             # <<<<<<<<<<<<<<
 *             self.n_center = len(self.center_left_indices)
 * 
 */
    __pyx_t_3 = __pyx_f_6pandas_5_libs_8interval_sort_values_and_indices(((PyObject *)__pyx_v_right), ((PyObject *)__pyx_v_indices), __pyx_v_center_set); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2329, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if ((likely(PyTuple_CheckExact(__pyx_t_3))) || (PyList_CheckExact(__pyx_t_3))) {
      PyObject* sequence = __pyx_t_3;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(1, 2328, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_9 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_8 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_9 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_8 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_9);
      __Pyx_INCREF(__pyx_t_8);
      #else
      __pyx_t_9 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 2328, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_8 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 2328, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      #endif
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_5 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 2328, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_11 = Py_TYPE(__pyx_t_5)->tp_iternext;
      index = 0; __pyx_t_9 = __pyx_t_11(__pyx_t_5); if (unlikely(!__pyx_t_9)) goto __pyx_L9_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_9);
      index = 1; __pyx_t_8 = __pyx_t_11(__pyx_t_5); if (unlikely(!__pyx_t_8)) goto __pyx_L9_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_8);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_11(__pyx_t_5), 2) < 0) __PYX_ERR(1, 2328, __pyx_L1_error)
      __pyx_t_11 = NULL;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      goto __pyx_L10_unpacking_done;
      __pyx_L9_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_11 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(1, 2328, __pyx_L1_error)
      __pyx_L10_unpacking_done:;
    }

    /* "pandas/_libs/intervaltree.pxi":2328
 *             self.center_left_values, self.center_left_indices = \
 *                 sort_values_and_indices(left, indices, center_set)
 *             self.center_right_values, self.center_right_indices = \             # <<<<<<<<<<<<<<
 *                 sort_values_and_indices(right, indices, center_set)
 *             self.n_center = len(self.center_left_indices)
 */
    __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_t_9, PyBUF_WRITABLE); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(1, 2328, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_10 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_t_8, PyBUF_WRITABLE); if (unlikely(!__pyx_t_10.memview)) __PYX_ERR(1, 2328, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->center_right_values, 0);
    __pyx_v_self->center_right_values = __pyx_t_7;
    __pyx_t_7.memview = NULL;
    __pyx_t_7.data = NULL;
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->center_right_indices, 0);
    __pyx_v_self->center_right_indices = __pyx_t_10;
    __pyx_t_10.memview = NULL;
    __pyx_t_10.data = NULL;

    /* "pandas/_libs/intervaltree.pxi":2330
 *             self.center_right_values, self.center_right_indices = \
 *                 sort_values_and_indices(right, indices, center_set)
 *             self.n_center = len(self.center_left_indices)             # <<<<<<<<<<<<<<
 * 
 *     @cython.wraparound(False)
 */
    if (unlikely(!__pyx_v_self->center_left_indices.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(1, 2330, __pyx_L1_error)}
    __pyx_t_12 = __Pyx_MemoryView_Len(__pyx_v_self->center_left_indices); 
    __pyx_v_self->n_center = __pyx_t_12;
  }
  __pyx_L4:;

  /* "pandas/_libs/intervaltree.pxi":2288
 *         readonly bint is_leaf_node
 * 
 *     def __init__(self,             # <<<<<<<<<<<<<<
 *                  ndarray[int64_t, ndim=1] left,
 *                  ndarray[int64_t, ndim=1] right,
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __PYX_XDEC_MEMVIEW(&__pyx_t_7, 1);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __PYX_XDEC_MEMVIEW(&__pyx_t_10, 1);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indices.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_left.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_right.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("pandas._libs.interval.Int64ClosedLeftIntervalNode.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indices.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_left.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_right.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF(__pyx_v_left_set);
  __Pyx_XDECREF(__pyx_v_right_set);
  __Pyx_XDECREF(__pyx_v_center_set);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":2334
 *     @cython.wraparound(False)
 *     @cython.boundscheck(False)
 *     cdef classify_intervals(self, int64_t[:] left, int64_t[:] right):             # <<<<<<<<<<<<<<
 *         """Classify the given intervals based upon whether they fall to the
 *         left, right, or overlap with this node's pivot.
 */

static PyObject *__pyx_f_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_classify_intervals(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode *__pyx_v_self, __Pyx_memviewslice __pyx_v_left, __Pyx_memviewslice __pyx_v_right) {
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_left_ind = 0;
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_right_ind = 0;
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_overlapping_ind = 0;
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __pyx_t_5numpy_int64_t __pyx_t_2;
  __pyx_t_5numpy_int64_t __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  int __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  __Pyx_RefNannySetupContext("classify_intervals", 0);

  /* "pandas/_libs/intervaltree.pxi":2342
 *             Py_ssize_t i
 * 
 *         left_ind = Int64Vector()             # <<<<<<<<<<<<<<
 *         right_ind = Int64Vector()
 *         overlapping_ind = Int64Vector()
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2342, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_left_ind = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":2343
 * 
 *         left_ind = Int64Vector()
 *         right_ind = Int64Vector()             # <<<<<<<<<<<<<<
 *         overlapping_ind = Int64Vector()
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2343, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_right_ind = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":2344
 *         left_ind = Int64Vector()
 *         right_ind = Int64Vector()
 *         overlapping_ind = Int64Vector()             # <<<<<<<<<<<<<<
 * 
 *         for i in range(self.n_elements):
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2344, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_overlapping_ind = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":2346
 *         overlapping_ind = Int64Vector()
 * 
 *         for i in range(self.n_elements):             # <<<<<<<<<<<<<<
 *             if right[i] <= self.pivot:
 *                 left_ind.append(i)
 */
  __pyx_t_2 = __pyx_v_self->n_elements;
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "pandas/_libs/intervaltree.pxi":2347
 * 
 *         for i in range(self.n_elements):
 *             if right[i] <= self.pivot:             # <<<<<<<<<<<<<<
 *                 left_ind.append(i)
 *             elif self.pivot < left[i]:
 */
    __pyx_t_5 = __pyx_v_i;
    __pyx_t_6 = (((*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_right.data + __pyx_t_5 * __pyx_v_right.strides[0]) ))) <= __pyx_v_self->pivot) != 0);
    if (__pyx_t_6) {

      /* "pandas/_libs/intervaltree.pxi":2348
 *         for i in range(self.n_elements):
 *             if right[i] <= self.pivot:
 *                 left_ind.append(i)             # <<<<<<<<<<<<<<
 *             elif self.pivot < left[i]:
 *                 right_ind.append(i)
 */
      ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_left_ind->__pyx_vtab)->append(__pyx_v_left_ind, __pyx_v_i);

      /* "pandas/_libs/intervaltree.pxi":2347
 * 
 *         for i in range(self.n_elements):
 *             if right[i] <= self.pivot:             # <<<<<<<<<<<<<<
 *                 left_ind.append(i)
 *             elif self.pivot < left[i]:
 */
      goto __pyx_L5;
    }

    /* "pandas/_libs/intervaltree.pxi":2349
 *             if right[i] <= self.pivot:
 *                 left_ind.append(i)
 *             elif self.pivot < left[i]:             # <<<<<<<<<<<<<<
 *                 right_ind.append(i)
 *             else:
 */
    __pyx_t_7 = __pyx_v_i;
    __pyx_t_6 = ((__pyx_v_self->pivot < (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_left.data + __pyx_t_7 * __pyx_v_left.strides[0]) )))) != 0);
    if (__pyx_t_6) {

      /* "pandas/_libs/intervaltree.pxi":2350
 *                 left_ind.append(i)
 *             elif self.pivot < left[i]:
 *                 right_ind.append(i)             # <<<<<<<<<<<<<<
 *             else:
 *                 overlapping_ind.append(i)
 */
      ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_right_ind->__pyx_vtab)->append(__pyx_v_right_ind, __pyx_v_i);

      /* "pandas/_libs/intervaltree.pxi":2349
 *             if right[i] <= self.pivot:
 *                 left_ind.append(i)
 *             elif self.pivot < left[i]:             # <<<<<<<<<<<<<<
 *                 right_ind.append(i)
 *             else:
 */
      goto __pyx_L5;
    }

    /* "pandas/_libs/intervaltree.pxi":2352
 *                 right_ind.append(i)
 *             else:
 *                 overlapping_ind.append(i)             # <<<<<<<<<<<<<<
 * 
 *         return (left_ind.to_array(),
 */
    /*else*/ {
      ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_overlapping_ind->__pyx_vtab)->append(__pyx_v_overlapping_ind, __pyx_v_i);
    }
    __pyx_L5:;
  }

  /* "pandas/_libs/intervaltree.pxi":2354
 *                 overlapping_ind.append(i)
 * 
 *         return (left_ind.to_array(),             # <<<<<<<<<<<<<<
 *                 right_ind.to_array(),
 *                 overlapping_ind.to_array())
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_left_ind->__pyx_vtab)->to_array(__pyx_v_left_ind, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2354, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "pandas/_libs/intervaltree.pxi":2355
 * 
 *         return (left_ind.to_array(),
 *                 right_ind.to_array(),             # <<<<<<<<<<<<<<
 *                 overlapping_ind.to_array())
 * 
 */
  __pyx_t_8 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_right_ind->__pyx_vtab)->to_array(__pyx_v_right_ind, 0); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 2355, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);

  /* "pandas/_libs/intervaltree.pxi":2356
 *         return (left_ind.to_array(),
 *                 right_ind.to_array(),
 *                 overlapping_ind.to_array())             # <<<<<<<<<<<<<<
 * 
 *     cdef new_child_node(self,
 */
  __pyx_t_9 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_overlapping_ind->__pyx_vtab)->to_array(__pyx_v_overlapping_ind, 0); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 2356, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);

  /* "pandas/_libs/intervaltree.pxi":2354
 *                 overlapping_ind.append(i)
 * 
 *         return (left_ind.to_array(),             # <<<<<<<<<<<<<<
 *                 right_ind.to_array(),
 *                 overlapping_ind.to_array())
 */
  __pyx_t_10 = PyTuple_New(3); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 2354, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_t_8);
  __Pyx_GIVEREF(__pyx_t_9);
  PyTuple_SET_ITEM(__pyx_t_10, 2, __pyx_t_9);
  __pyx_t_1 = 0;
  __pyx_t_8 = 0;
  __pyx_t_9 = 0;
  __pyx_r = __pyx_t_10;
  __pyx_t_10 = 0;
  goto __pyx_L0;

  /* "pandas/_libs/intervaltree.pxi":2334
 *     @cython.wraparound(False)
 *     @cython.boundscheck(False)
 *     cdef classify_intervals(self, int64_t[:] left, int64_t[:] right):             # <<<<<<<<<<<<<<
 *         """Classify the given intervals based upon whether they fall to the
 *         left, right, or overlap with this node's pivot.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("pandas._libs.interval.Int64ClosedLeftIntervalNode.classify_intervals", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_left_ind);
  __Pyx_XDECREF((PyObject *)__pyx_v_right_ind);
  __Pyx_XDECREF((PyObject *)__pyx_v_overlapping_ind);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":2358
 *                 overlapping_ind.to_array())
 * 
 *     cdef new_child_node(self,             # <<<<<<<<<<<<<<
 *                         ndarray[int64_t, ndim=1] left,
 *                         ndarray[int64_t, ndim=1] right,
 */

static PyObject *__pyx_f_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_new_child_node(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode *__pyx_v_self, PyArrayObject *__pyx_v_left, PyArrayObject *__pyx_v_right, PyArrayObject *__pyx_v_indices, PyArrayObject *__pyx_v_subset) {
  __Pyx_LocalBuf_ND __pyx_pybuffernd_indices;
  __Pyx_Buffer __pyx_pybuffer_indices;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_left;
  __Pyx_Buffer __pyx_pybuffer_left;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_right;
  __Pyx_Buffer __pyx_pybuffer_right;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_subset;
  __Pyx_Buffer __pyx_pybuffer_subset;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyArrayObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyArrayObject *__pyx_t_7 = NULL;
  PyArrayObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  __Pyx_RefNannySetupContext("new_child_node", 0);
  __Pyx_INCREF((PyObject *)__pyx_v_left);
  __Pyx_INCREF((PyObject *)__pyx_v_right);
  __Pyx_INCREF((PyObject *)__pyx_v_indices);
  __pyx_pybuffer_left.pybuffer.buf = NULL;
  __pyx_pybuffer_left.refcount = 0;
  __pyx_pybuffernd_left.data = NULL;
  __pyx_pybuffernd_left.rcbuffer = &__pyx_pybuffer_left;
  __pyx_pybuffer_right.pybuffer.buf = NULL;
  __pyx_pybuffer_right.refcount = 0;
  __pyx_pybuffernd_right.data = NULL;
  __pyx_pybuffernd_right.rcbuffer = &__pyx_pybuffer_right;
  __pyx_pybuffer_indices.pybuffer.buf = NULL;
  __pyx_pybuffer_indices.refcount = 0;
  __pyx_pybuffernd_indices.data = NULL;
  __pyx_pybuffernd_indices.rcbuffer = &__pyx_pybuffer_indices;
  __pyx_pybuffer_subset.pybuffer.buf = NULL;
  __pyx_pybuffer_subset.refcount = 0;
  __pyx_pybuffernd_subset.data = NULL;
  __pyx_pybuffernd_subset.rcbuffer = &__pyx_pybuffer_subset;
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_left.rcbuffer->pybuffer, (PyObject*)__pyx_v_left, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 2358, __pyx_L1_error)
  }
  __pyx_pybuffernd_left.diminfo[0].strides = __pyx_pybuffernd_left.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_left.diminfo[0].shape = __pyx_pybuffernd_left.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_right.rcbuffer->pybuffer, (PyObject*)__pyx_v_right, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 2358, __pyx_L1_error)
  }
  __pyx_pybuffernd_right.diminfo[0].strides = __pyx_pybuffernd_right.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_right.diminfo[0].shape = __pyx_pybuffernd_right.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_indices.rcbuffer->pybuffer, (PyObject*)__pyx_v_indices, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 2358, __pyx_L1_error)
  }
  __pyx_pybuffernd_indices.diminfo[0].strides = __pyx_pybuffernd_indices.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_indices.diminfo[0].shape = __pyx_pybuffernd_indices.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_subset.rcbuffer->pybuffer, (PyObject*)__pyx_v_subset, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 2358, __pyx_L1_error)
  }
  __pyx_pybuffernd_subset.diminfo[0].strides = __pyx_pybuffernd_subset.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_subset.diminfo[0].shape = __pyx_pybuffernd_subset.rcbuffer->pybuffer.shape[0];

  /* "pandas/_libs/intervaltree.pxi":2365
 *         """Create a new child node.
 *         """
 *         left = take(left, subset)             # <<<<<<<<<<<<<<
 *         right = take(right, subset)
 *         indices = take(indices, subset)
 */
  __pyx_t_1 = __pyx_f_6pandas_5_libs_8interval_take(((PyArrayObject *)__pyx_v_left), ((PyArrayObject *)__pyx_v_subset)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2365, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(1, 2365, __pyx_L1_error)
  __pyx_t_2 = ((PyArrayObject *)__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_left.rcbuffer->pybuffer);
    __pyx_t_3 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_left.rcbuffer->pybuffer, (PyObject*)__pyx_t_2, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack);
    if (unlikely(__pyx_t_3 < 0)) {
      PyErr_Fetch(&__pyx_t_4, &__pyx_t_5, &__pyx_t_6);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_left.rcbuffer->pybuffer, (PyObject*)__pyx_v_left, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_4); Py_XDECREF(__pyx_t_5); Py_XDECREF(__pyx_t_6);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_4, __pyx_t_5, __pyx_t_6);
      }
      __pyx_t_4 = __pyx_t_5 = __pyx_t_6 = 0;
    }
    __pyx_pybuffernd_left.diminfo[0].strides = __pyx_pybuffernd_left.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_left.diminfo[0].shape = __pyx_pybuffernd_left.rcbuffer->pybuffer.shape[0];
    if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(1, 2365, __pyx_L1_error)
  }
  __pyx_t_2 = 0;
  __Pyx_DECREF_SET(__pyx_v_left, ((PyArrayObject *)__pyx_t_1));
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":2366
 *         """
 *         left = take(left, subset)
 *         right = take(right, subset)             # <<<<<<<<<<<<<<
 *         indices = take(indices, subset)
 *         return Int64ClosedLeftIntervalNode(
 */
  __pyx_t_1 = __pyx_f_6pandas_5_libs_8interval_take(((PyArrayObject *)__pyx_v_right), ((PyArrayObject *)__pyx_v_subset)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2366, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(1, 2366, __pyx_L1_error)
  __pyx_t_7 = ((PyArrayObject *)__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_right.rcbuffer->pybuffer);
    __pyx_t_3 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_right.rcbuffer->pybuffer, (PyObject*)__pyx_t_7, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack);
    if (unlikely(__pyx_t_3 < 0)) {
      PyErr_Fetch(&__pyx_t_6, &__pyx_t_5, &__pyx_t_4);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_right.rcbuffer->pybuffer, (PyObject*)__pyx_v_right, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_6); Py_XDECREF(__pyx_t_5); Py_XDECREF(__pyx_t_4);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_6, __pyx_t_5, __pyx_t_4);
      }
      __pyx_t_6 = __pyx_t_5 = __pyx_t_4 = 0;
    }
    __pyx_pybuffernd_right.diminfo[0].strides = __pyx_pybuffernd_right.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_right.diminfo[0].shape = __pyx_pybuffernd_right.rcbuffer->pybuffer.shape[0];
    if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(1, 2366, __pyx_L1_error)
  }
  __pyx_t_7 = 0;
  __Pyx_DECREF_SET(__pyx_v_right, ((PyArrayObject *)__pyx_t_1));
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":2367
 *         left = take(left, subset)
 *         right = take(right, subset)
 *         indices = take(indices, subset)             # <<<<<<<<<<<<<<
 *         return Int64ClosedLeftIntervalNode(
 *             left, right, indices, self.leaf_size)
 */
  __pyx_t_1 = __pyx_f_6pandas_5_libs_8interval_take(((PyArrayObject *)__pyx_v_indices), ((PyArrayObject *)__pyx_v_subset)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2367, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(1, 2367, __pyx_L1_error)
  __pyx_t_8 = ((PyArrayObject *)__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indices.rcbuffer->pybuffer);
    __pyx_t_3 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_indices.rcbuffer->pybuffer, (PyObject*)__pyx_t_8, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack);
    if (unlikely(__pyx_t_3 < 0)) {
      PyErr_Fetch(&__pyx_t_4, &__pyx_t_5, &__pyx_t_6);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_indices.rcbuffer->pybuffer, (PyObject*)__pyx_v_indices, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_4); Py_XDECREF(__pyx_t_5); Py_XDECREF(__pyx_t_6);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_4, __pyx_t_5, __pyx_t_6);
      }
      __pyx_t_4 = __pyx_t_5 = __pyx_t_6 = 0;
    }
    __pyx_pybuffernd_indices.diminfo[0].strides = __pyx_pybuffernd_indices.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_indices.diminfo[0].shape = __pyx_pybuffernd_indices.rcbuffer->pybuffer.shape[0];
    if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(1, 2367, __pyx_L1_error)
  }
  __pyx_t_8 = 0;
  __Pyx_DECREF_SET(__pyx_v_indices, ((PyArrayObject *)__pyx_t_1));
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":2368
 *         right = take(right, subset)
 *         indices = take(indices, subset)
 *         return Int64ClosedLeftIntervalNode(             # <<<<<<<<<<<<<<
 *             left, right, indices, self.leaf_size)
 * 
 */
  __Pyx_XDECREF(__pyx_r);

  /* "pandas/_libs/intervaltree.pxi":2369
 *         indices = take(indices, subset)
 *         return Int64ClosedLeftIntervalNode(
 *             left, right, indices, self.leaf_size)             # <<<<<<<<<<<<<<
 * 
 *     @cython.wraparound(False)
 */
  __pyx_t_1 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->leaf_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2369, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "pandas/_libs/intervaltree.pxi":2368
 *         right = take(right, subset)
 *         indices = take(indices, subset)
 *         return Int64ClosedLeftIntervalNode(             # <<<<<<<<<<<<<<
 *             left, right, indices, self.leaf_size)
 * 
 */
  __pyx_t_9 = PyTuple_New(4); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 2368, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_INCREF(((PyObject *)__pyx_v_left));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_left));
  PyTuple_SET_ITEM(__pyx_t_9, 0, ((PyObject *)__pyx_v_left));
  __Pyx_INCREF(((PyObject *)__pyx_v_right));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_right));
  PyTuple_SET_ITEM(__pyx_t_9, 1, ((PyObject *)__pyx_v_right));
  __Pyx_INCREF(((PyObject *)__pyx_v_indices));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_indices));
  PyTuple_SET_ITEM(__pyx_t_9, 2, ((PyObject *)__pyx_v_indices));
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_9, 3, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode), __pyx_t_9, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2368, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pandas/_libs/intervaltree.pxi":2358
 *                 overlapping_ind.to_array())
 * 
 *     cdef new_child_node(self,             # <<<<<<<<<<<<<<
 *                         ndarray[int64_t, ndim=1] left,
 *                         ndarray[int64_t, ndim=1] right,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_9);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indices.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_left.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_right.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_subset.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("pandas._libs.interval.Int64ClosedLeftIntervalNode.new_child_node", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indices.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_left.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_right.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_subset.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF((PyObject *)__pyx_v_left);
  __Pyx_XDECREF((PyObject *)__pyx_v_right);
  __Pyx_XDECREF((PyObject *)__pyx_v_indices);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":2374
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_2query[] = "Recursively query this node and its sub-nodes for intervals that\n        overlap with the query point.\n        ";
static PyMethodDef __pyx_mdef_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_3query = {"query", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_3query, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_2query};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_args = 0;
  PyObject *__pyx_v_kwargs = 0;
  CYTHON_UNUSED PyObject *__pyx_v_defaults = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fused_cpdef (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_args,&__pyx_n_s_kwargs,&__pyx_n_s_defaults,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_args)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_kwargs)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fused_cpdef", 1, 3, 3, 1); __PYX_ERR(1, 2374, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_defaults)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fused_cpdef", 1, 3, 3, 2); __PYX_ERR(1, 2374, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fused_cpdef") < 0)) __PYX_ERR(1, 2374, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_args = values[0];
    __pyx_v_kwargs = values[1];
    __pyx_v_defaults = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fused_cpdef", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 2374, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Int64ClosedLeftIntervalNode.__pyx_fused_cpdef", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_2query(((struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode *)__pyx_v_signatures), __pyx_v_args, __pyx_v_kwargs, __pyx_v_defaults);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_2query(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode *__pyx_v_signatures, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs, CYTHON_UNUSED PyObject *__pyx_v_defaults) {
  PyObject *__pyx_v_dest_sig = NULL;
  Py_ssize_t __pyx_v_i;
  CYTHON_UNUSED PyTypeObject *__pyx_v_ndarray = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_v_itemsize;
  PyObject *__pyx_v_arg = NULL;
  PyObject *__pyx_v_candidates = NULL;
  PyObject *__pyx_v_sig = NULL;
  int __pyx_v_match_found;
  PyObject *__pyx_v_src_sig = NULL;
  PyObject *__pyx_v_dst_type = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  Py_ssize_t __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  Py_ssize_t __pyx_t_13;
  Py_ssize_t __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  int __pyx_t_16;
  __Pyx_RefNannySetupContext("query", 0);
  __Pyx_INCREF(__pyx_v_kwargs);
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2374, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  PyList_SET_ITEM(__pyx_t_1, 0, Py_None);
  __pyx_v_dest_sig = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_3 = (__pyx_v_kwargs != Py_None);
  __pyx_t_4 = (__pyx_t_3 != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_kwargs); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(1, 2374, __pyx_L1_error)
  __pyx_t_3 = ((!__pyx_t_4) != 0);
  __pyx_t_2 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {
    __Pyx_INCREF(Py_None);
    __Pyx_DECREF_SET(__pyx_v_kwargs, Py_None);
  }
  __pyx_t_1 = ((PyObject *)__Pyx_ImportNumPyArrayTypeIfAvailable()); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2374, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_ndarray = ((PyTypeObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_v_itemsize = -1L;
  if (unlikely(__pyx_v_args == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 2374, __pyx_L1_error)
  }
  __pyx_t_5 = PyTuple_GET_SIZE(((PyObject*)__pyx_v_args)); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(1, 2374, __pyx_L1_error)
  __pyx_t_2 = ((2 < __pyx_t_5) != 0);
  if (__pyx_t_2) {
    if (unlikely(__pyx_v_args == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 2374, __pyx_L1_error)
    }
    __pyx_t_1 = PyTuple_GET_ITEM(((PyObject*)__pyx_v_args), 2);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v_arg = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L6;
  }
  __pyx_t_3 = (__pyx_v_kwargs != Py_None);
  __pyx_t_4 = (__pyx_t_3 != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L7_bool_binop_done;
  }
  if (unlikely(__pyx_v_kwargs == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(1, 2374, __pyx_L1_error)
  }
  __pyx_t_4 = (__Pyx_PyDict_ContainsTF(__pyx_n_s_point, ((PyObject*)__pyx_v_kwargs), Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(1, 2374, __pyx_L1_error)
  __pyx_t_3 = (__pyx_t_4 != 0);
  __pyx_t_2 = __pyx_t_3;
  __pyx_L7_bool_binop_done:;
  if (__pyx_t_2) {
    if (unlikely(__pyx_v_kwargs == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 2374, __pyx_L1_error)
    }
    __pyx_t_1 = __Pyx_PyDict_GetItem(((PyObject*)__pyx_v_kwargs), __pyx_n_s_point); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2374, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_arg = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L6;
  }
  /*else*/ {
    if (unlikely(__pyx_v_args == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
      __PYX_ERR(1, 2374, __pyx_L1_error)
    }
    __pyx_t_5 = PyTuple_GET_SIZE(((PyObject*)__pyx_v_args)); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(1, 2374, __pyx_L1_error)
    __pyx_t_1 = PyInt_FromSsize_t(__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2374, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_6 = PyTuple_New(3); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 2374, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_INCREF(__pyx_int_3);
    __Pyx_GIVEREF(__pyx_int_3);
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_int_3);
    __Pyx_INCREF(__pyx_n_s_s);
    __Pyx_GIVEREF(__pyx_n_s_s);
    PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_n_s_s);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Expected_at_least_d_argument_s_g, __pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2374, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_1); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 2374, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(1, 2374, __pyx_L1_error)
  }
  __pyx_L6:;
  while (1) {
    __pyx_t_2 = PyFloat_Check(__pyx_v_arg); 
    __pyx_t_3 = (__pyx_t_2 != 0);
    if (__pyx_t_3) {
      if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, __pyx_n_s_float64_t, long, 1, __Pyx_PyInt_From_long, 1, 0, 0) < 0)) __PYX_ERR(1, 2374, __pyx_L1_error)
      goto __pyx_L10_break;
    }
    __pyx_t_2 = PyInt_Check(__pyx_v_arg); 
    __pyx_t_4 = (__pyx_t_2 != 0);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_3 = __pyx_t_4;
      goto __pyx_L13_bool_binop_done;
    }
    __pyx_t_4 = PyLong_Check(__pyx_v_arg); 
    __pyx_t_2 = (__pyx_t_4 != 0);
    __pyx_t_3 = __pyx_t_2;
    __pyx_L13_bool_binop_done:;
    __pyx_t_2 = (__pyx_t_3 != 0);
    if (__pyx_t_2) {
      if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, __pyx_n_s_int64_t, long, 1, __Pyx_PyInt_From_long, 1, 0, 0) < 0)) __PYX_ERR(1, 2374, __pyx_L1_error)
      goto __pyx_L10_break;
    }
    if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, Py_None, long, 1, __Pyx_PyInt_From_long, 1, 0, 0) < 0)) __PYX_ERR(1, 2374, __pyx_L1_error)
    goto __pyx_L10_break;
  }
  __pyx_L10_break:;
  __pyx_t_6 = PyList_New(0); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 2374, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_v_candidates = ((PyObject*)__pyx_t_6);
  __pyx_t_6 = 0;
  __pyx_t_5 = 0;
  if (unlikely(((PyObject *)__pyx_v_signatures) == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(1, 2374, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_dict_iterator(((PyObject*)__pyx_v_signatures), 1, ((PyObject *)NULL), (&__pyx_t_7), (&__pyx_t_8)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2374, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __pyx_t_6 = __pyx_t_1;
  __pyx_t_1 = 0;
  while (1) {
    __pyx_t_9 = __Pyx_dict_iter_next(__pyx_t_6, __pyx_t_7, &__pyx_t_5, &__pyx_t_1, NULL, NULL, __pyx_t_8);
    if (unlikely(__pyx_t_9 == 0)) break;
    if (unlikely(__pyx_t_9 == -1)) __PYX_ERR(1, 2374, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_sig, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_v_match_found = 0;
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_v_sig, __pyx_n_s_strip); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 2374, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __pyx_t_12 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_11))) {
      __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_11);
      if (likely(__pyx_t_12)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
        __Pyx_INCREF(__pyx_t_12);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_11, function);
      }
    }
    __pyx_t_10 = (__pyx_t_12) ? __Pyx_PyObject_Call2Args(__pyx_t_11, __pyx_t_12, __pyx_kp_s__13) : __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_kp_s__13);
    __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
    if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 2374, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_split); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 2374, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_t_10 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_11))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_11);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_11, function);
      }
    }
    __pyx_t_1 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_11, __pyx_t_10, __pyx_kp_s__14) : __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_kp_s__14);
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2374, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_XDECREF_SET(__pyx_v_src_sig, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_13 = PyList_GET_SIZE(__pyx_v_dest_sig); if (unlikely(__pyx_t_13 == ((Py_ssize_t)-1))) __PYX_ERR(1, 2374, __pyx_L1_error)
    __pyx_t_14 = __pyx_t_13;
    for (__pyx_t_15 = 0; __pyx_t_15 < __pyx_t_14; __pyx_t_15+=1) {
      __pyx_v_i = __pyx_t_15;
      __pyx_t_1 = PyList_GET_ITEM(__pyx_v_dest_sig, __pyx_v_i);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_XDECREF_SET(__pyx_v_dst_type, __pyx_t_1);
      __pyx_t_1 = 0;
      __pyx_t_2 = (__pyx_v_dst_type != Py_None);
      __pyx_t_3 = (__pyx_t_2 != 0);
      if (__pyx_t_3) {
        __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_src_sig, __pyx_v_i, Py_ssize_t, 1, PyInt_FromSsize_t, 0, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2374, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_11 = PyObject_RichCompare(__pyx_t_1, __pyx_v_dst_type, Py_EQ); __Pyx_XGOTREF(__pyx_t_11); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 2374, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_11); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(1, 2374, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        if (__pyx_t_3) {
          __pyx_v_match_found = 1;
          goto __pyx_L20;
        }
        /*else*/ {
          __pyx_v_match_found = 0;
          goto __pyx_L18_break;
        }
        __pyx_L20:;
      }
    }
    __pyx_L18_break:;
    __pyx_t_3 = (__pyx_v_match_found != 0);
    if (__pyx_t_3) {
      __pyx_t_16 = __Pyx_PyList_Append(__pyx_v_candidates, __pyx_v_sig); if (unlikely(__pyx_t_16 == ((int)-1))) __PYX_ERR(1, 2374, __pyx_L1_error)
    }
  }
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_3 = (PyList_GET_SIZE(__pyx_v_candidates) != 0);
  __pyx_t_2 = ((!__pyx_t_3) != 0);
  if (__pyx_t_2) {
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__15, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 2374, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(1, 2374, __pyx_L1_error)
  }
  __pyx_t_7 = PyList_GET_SIZE(__pyx_v_candidates); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1))) __PYX_ERR(1, 2374, __pyx_L1_error)
  __pyx_t_2 = ((__pyx_t_7 > 1) != 0);
  if (__pyx_t_2) {
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__16, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 2374, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(1, 2374, __pyx_L1_error)
  }
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    if (unlikely(((PyObject *)__pyx_v_signatures) == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 2374, __pyx_L1_error)
    }
    __pyx_t_6 = __Pyx_PyDict_GetItem(((PyObject*)__pyx_v_signatures), PyList_GET_ITEM(__pyx_v_candidates, 0)); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 2374, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_r = __pyx_t_6;
    __pyx_t_6 = 0;
    goto __pyx_L0;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("pandas._libs.interval.Int64ClosedLeftIntervalNode.__pyx_fused_cpdef", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_dest_sig);
  __Pyx_XDECREF(__pyx_v_ndarray);
  __Pyx_XDECREF(__pyx_v_arg);
  __Pyx_XDECREF(__pyx_v_candidates);
  __Pyx_XDECREF(__pyx_v_sig);
  __Pyx_XDECREF(__pyx_v_src_sig);
  __Pyx_XDECREF(__pyx_v_dst_type);
  __Pyx_XDECREF(__pyx_v_kwargs);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_9__pyx_fuse_0query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_fuse_0__pyx_f_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float64_t __pyx_v_point, int __pyx_skip_dispatch) {
  __Pyx_memviewslice __pyx_v_indices = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_values = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __pyx_t_5numpy_int64_t __pyx_t_9;
  __pyx_t_5numpy_int64_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  int __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  __Pyx_memviewslice __pyx_t_16 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_17;
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  __Pyx_RefNannySetupContext("__pyx_fuse_0query", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pyx_fuse_0query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2374, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_9__pyx_fuse_0query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = PyFloat_FromDouble(__pyx_v_point); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2374, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2374, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2374, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 2374, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_5) {
            __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2374, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2374, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_3query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_4 = PyFloat_FromDouble(__pyx_v_point); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 2374, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_7 = __pyx_t_1; __pyx_t_3 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2374, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2374, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 2374, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (__pyx_t_3) {
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_t_4);
          __pyx_t_4 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2374, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pandas/_libs/intervaltree.pxi":2383
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
  __pyx_t_8 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_8) {

    /* "pandas/_libs/intervaltree.pxi":2387
 *             # continue the binary tree structure. Instead, we use linear
 *             # search.
 *             for i in range(self.n_elements):             # <<<<<<<<<<<<<<
 *                 if self.left[i] <= point < self.right[i]:
 *                     result.append(self.indices[i])
 */
    __pyx_t_9 = __pyx_v_self->n_elements;
    __pyx_t_10 = __pyx_t_9;
    for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
      __pyx_v_i = __pyx_t_11;

      /* "pandas/_libs/intervaltree.pxi":2388
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point < self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      __pyx_t_12 = __pyx_v_i;
      __pyx_t_8 = ((*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->left.data + __pyx_t_12 * __pyx_v_self->left.strides[0]) ))) <= __pyx_v_point);
      if (__pyx_t_8) {
        __pyx_t_13 = __pyx_v_i;
        __pyx_t_8 = (__pyx_v_point < (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->right.data + __pyx_t_13 * __pyx_v_self->right.strides[0]) ))));
      }
      __pyx_t_14 = (__pyx_t_8 != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":2389
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point < self.right[i]:
 *                     result.append(self.indices[i])             # <<<<<<<<<<<<<<
 *         else:
 *             # There are child nodes. Based on comparing our query to the pivot,
 */
        __pyx_t_15 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->indices.data + __pyx_t_15 * __pyx_v_self->indices.strides[0]) ))));

        /* "pandas/_libs/intervaltree.pxi":2388
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point < self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      }
    }

    /* "pandas/_libs/intervaltree.pxi":2383
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
    goto __pyx_L3;
  }

  /* "pandas/_libs/intervaltree.pxi":2393
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
  /*else*/ {
    __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":2394
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:
 *                 values = self.center_left_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 */
      __pyx_t_16 = __pyx_v_self->center_left_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":2395
 *             if point < self.pivot:
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:
 */
      __pyx_t_16 = __pyx_v_self->center_left_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_indices = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":2396
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):             # <<<<<<<<<<<<<<
 *                     if not values[i] <= point:
 *                         break
 */
      __pyx_t_9 = __pyx_v_self->n_center;
      __pyx_t_10 = __pyx_t_9;
      for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":2397
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_17 = __pyx_v_i;
        __pyx_t_14 = ((!(((*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_17 * __pyx_v_values.strides[0]) ))) <= __pyx_v_point) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":2398
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:
 */
          goto __pyx_L9_break;

          /* "pandas/_libs/intervaltree.pxi":2397
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":2399
 *                     if not values[i] <= point:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 */
        __pyx_t_18 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_18 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L9_break:;

      /* "pandas/_libs/intervaltree.pxi":2400
 *                         break
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->left_node->max_right) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":2401
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)             # <<<<<<<<<<<<<<
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode *)__pyx_v_self->left_node->__pyx_vtab)->__pyx_fuse_0query(__pyx_v_self->left_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2401, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":2400
 *                         break
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":2393
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":2402
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
    __pyx_t_14 = ((__pyx_v_point > __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":2403
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 *                 values = self.center_right_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 */
      __pyx_t_16 = __pyx_v_self->center_right_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":2404
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:
 */
      __pyx_t_16 = __pyx_v_self->center_right_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_indices = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":2405
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):             # <<<<<<<<<<<<<<
 *                     if not point < values[i]:
 *                         break
 */
      for (__pyx_t_11 = (__pyx_v_self->n_center - 1); __pyx_t_11 > -1L; __pyx_t_11-=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":2406
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_19 = __pyx_v_i;
        __pyx_t_14 = ((!((__pyx_v_point < (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_19 * __pyx_v_values.strides[0]) )))) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":2407
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:
 */
          goto __pyx_L13_break;

          /* "pandas/_libs/intervaltree.pxi":2406
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":2408
 *                     if not point < values[i]:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if self.right_node.min_left <= point:
 *                     self.right_node.query(result, point)
 */
        __pyx_t_20 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_20 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L13_break:;

      /* "pandas/_libs/intervaltree.pxi":2409
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      __pyx_t_14 = ((__pyx_v_self->right_node->min_left <= __pyx_v_point) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":2410
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:
 *                     self.right_node.query(result, point)             # <<<<<<<<<<<<<<
 *             else:
 *                 result.extend(self.center_left_indices)
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode *)__pyx_v_self->right_node->__pyx_vtab)->__pyx_fuse_0query(__pyx_v_self->right_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2410, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":2409
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":2402
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":2412
 *                     self.right_node.query(result, point)
 *             else:
 *                 result.extend(self.center_left_indices)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
    /*else*/ {
      __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->extend(__pyx_v_result, __pyx_v_self->center_left_indices); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2412, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_L7:;
  }
  __pyx_L3:;

  /* "pandas/_libs/intervaltree.pxi":2374
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __PYX_XDEC_MEMVIEW(&__pyx_t_16, 1);
  __Pyx_AddTraceback("pandas._libs.interval.Int64ClosedLeftIntervalNode.query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_indices, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_values, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_9__pyx_fuse_0query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_0__pyx_mdef_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_9__pyx_fuse_0query = {"__pyx_fuse_0query", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_9__pyx_fuse_0query, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_2query};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_9__pyx_fuse_0query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result = 0;
  __pyx_t_5numpy_float64_t __pyx_v_point;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fuse_0query (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_result,&__pyx_n_s_point,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_result)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_point)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_0query", 1, 2, 2, 1); __PYX_ERR(1, 2374, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_0query") < 0)) __PYX_ERR(1, 2374, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_result = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)values[0]);
    __pyx_v_point = __pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_point == ((npy_float64)-1)) && PyErr_Occurred())) __PYX_ERR(1, 2374, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_0query", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 2374, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Int64ClosedLeftIntervalNode.__pyx_fuse_0query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_result), __pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector, 1, "result", 0))) __PYX_ERR(1, 2374, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_8__pyx_fuse_0query(((struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode *)__pyx_v_self), __pyx_v_result, __pyx_v_point);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_8__pyx_fuse_0query(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float64_t __pyx_v_point) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__pyx_fuse_0query", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_fuse_0__pyx_f_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_query(__pyx_v_self, __pyx_v_result, __pyx_v_point, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2374, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Int64ClosedLeftIntervalNode.__pyx_fuse_0query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_11__pyx_fuse_1query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_fuse_1__pyx_f_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float32_t __pyx_v_point, int __pyx_skip_dispatch) {
  __Pyx_memviewslice __pyx_v_indices = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_values = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __pyx_t_5numpy_int64_t __pyx_t_9;
  __pyx_t_5numpy_int64_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  int __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  __Pyx_memviewslice __pyx_t_16 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_17;
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  __Pyx_RefNannySetupContext("__pyx_fuse_1query", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pyx_fuse_1query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2374, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_11__pyx_fuse_1query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = PyFloat_FromDouble(__pyx_v_point); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2374, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2374, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2374, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 2374, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_5) {
            __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2374, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2374, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_3query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_4 = PyFloat_FromDouble(__pyx_v_point); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 2374, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_7 = __pyx_t_1; __pyx_t_3 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2374, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2374, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 2374, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (__pyx_t_3) {
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_t_4);
          __pyx_t_4 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2374, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pandas/_libs/intervaltree.pxi":2383
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
  __pyx_t_8 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_8) {

    /* "pandas/_libs/intervaltree.pxi":2387
 *             # continue the binary tree structure. Instead, we use linear
 *             # search.
 *             for i in range(self.n_elements):             # <<<<<<<<<<<<<<
 *                 if self.left[i] <= point < self.right[i]:
 *                     result.append(self.indices[i])
 */
    __pyx_t_9 = __pyx_v_self->n_elements;
    __pyx_t_10 = __pyx_t_9;
    for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
      __pyx_v_i = __pyx_t_11;

      /* "pandas/_libs/intervaltree.pxi":2388
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point < self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      __pyx_t_12 = __pyx_v_i;
      __pyx_t_8 = ((*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->left.data + __pyx_t_12 * __pyx_v_self->left.strides[0]) ))) <= __pyx_v_point);
      if (__pyx_t_8) {
        __pyx_t_13 = __pyx_v_i;
        __pyx_t_8 = (__pyx_v_point < (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->right.data + __pyx_t_13 * __pyx_v_self->right.strides[0]) ))));
      }
      __pyx_t_14 = (__pyx_t_8 != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":2389
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point < self.right[i]:
 *                     result.append(self.indices[i])             # <<<<<<<<<<<<<<
 *         else:
 *             # There are child nodes. Based on comparing our query to the pivot,
 */
        __pyx_t_15 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->indices.data + __pyx_t_15 * __pyx_v_self->indices.strides[0]) ))));

        /* "pandas/_libs/intervaltree.pxi":2388
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point < self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      }
    }

    /* "pandas/_libs/intervaltree.pxi":2383
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
    goto __pyx_L3;
  }

  /* "pandas/_libs/intervaltree.pxi":2393
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
  /*else*/ {
    __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":2394
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:
 *                 values = self.center_left_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 */
      __pyx_t_16 = __pyx_v_self->center_left_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":2395
 *             if point < self.pivot:
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:
 */
      __pyx_t_16 = __pyx_v_self->center_left_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_indices = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":2396
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):             # <<<<<<<<<<<<<<
 *                     if not values[i] <= point:
 *                         break
 */
      __pyx_t_9 = __pyx_v_self->n_center;
      __pyx_t_10 = __pyx_t_9;
      for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":2397
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_17 = __pyx_v_i;
        __pyx_t_14 = ((!(((*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_17 * __pyx_v_values.strides[0]) ))) <= __pyx_v_point) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":2398
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:
 */
          goto __pyx_L9_break;

          /* "pandas/_libs/intervaltree.pxi":2397
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":2399
 *                     if not values[i] <= point:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 */
        __pyx_t_18 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_18 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L9_break:;

      /* "pandas/_libs/intervaltree.pxi":2400
 *                         break
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->left_node->max_right) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":2401
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)             # <<<<<<<<<<<<<<
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode *)__pyx_v_self->left_node->__pyx_vtab)->__pyx_fuse_1query(__pyx_v_self->left_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2401, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":2400
 *                         break
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":2393
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":2402
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
    __pyx_t_14 = ((__pyx_v_point > __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":2403
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 *                 values = self.center_right_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 */
      __pyx_t_16 = __pyx_v_self->center_right_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":2404
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:
 */
      __pyx_t_16 = __pyx_v_self->center_right_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_indices = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":2405
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):             # <<<<<<<<<<<<<<
 *                     if not point < values[i]:
 *                         break
 */
      for (__pyx_t_11 = (__pyx_v_self->n_center - 1); __pyx_t_11 > -1L; __pyx_t_11-=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":2406
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_19 = __pyx_v_i;
        __pyx_t_14 = ((!((__pyx_v_point < (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_19 * __pyx_v_values.strides[0]) )))) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":2407
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:
 */
          goto __pyx_L13_break;

          /* "pandas/_libs/intervaltree.pxi":2406
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":2408
 *                     if not point < values[i]:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if self.right_node.min_left <= point:
 *                     self.right_node.query(result, point)
 */
        __pyx_t_20 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_20 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L13_break:;

      /* "pandas/_libs/intervaltree.pxi":2409
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      __pyx_t_14 = ((__pyx_v_self->right_node->min_left <= __pyx_v_point) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":2410
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:
 *                     self.right_node.query(result, point)             # <<<<<<<<<<<<<<
 *             else:
 *                 result.extend(self.center_left_indices)
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode *)__pyx_v_self->right_node->__pyx_vtab)->__pyx_fuse_1query(__pyx_v_self->right_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2410, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":2409
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":2402
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":2412
 *                     self.right_node.query(result, point)
 *             else:
 *                 result.extend(self.center_left_indices)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
    /*else*/ {
      __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->extend(__pyx_v_result, __pyx_v_self->center_left_indices); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2412, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_L7:;
  }
  __pyx_L3:;

  /* "pandas/_libs/intervaltree.pxi":2374
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __PYX_XDEC_MEMVIEW(&__pyx_t_16, 1);
  __Pyx_AddTraceback("pandas._libs.interval.Int64ClosedLeftIntervalNode.query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_indices, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_values, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_11__pyx_fuse_1query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_1__pyx_mdef_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_11__pyx_fuse_1query = {"__pyx_fuse_1query", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_11__pyx_fuse_1query, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_2query};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_11__pyx_fuse_1query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result = 0;
  __pyx_t_5numpy_float32_t __pyx_v_point;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fuse_1query (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_result,&__pyx_n_s_point,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_result)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_point)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_1query", 1, 2, 2, 1); __PYX_ERR(1, 2374, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_1query") < 0)) __PYX_ERR(1, 2374, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_result = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)values[0]);
    __pyx_v_point = __pyx_PyFloat_AsFloat(values[1]); if (unlikely((__pyx_v_point == ((npy_float32)-1)) && PyErr_Occurred())) __PYX_ERR(1, 2374, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_1query", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 2374, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Int64ClosedLeftIntervalNode.__pyx_fuse_1query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_result), __pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector, 1, "result", 0))) __PYX_ERR(1, 2374, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_10__pyx_fuse_1query(((struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode *)__pyx_v_self), __pyx_v_result, __pyx_v_point);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_10__pyx_fuse_1query(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float32_t __pyx_v_point) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__pyx_fuse_1query", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_fuse_1__pyx_f_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_query(__pyx_v_self, __pyx_v_result, __pyx_v_point, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2374, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Int64ClosedLeftIntervalNode.__pyx_fuse_1query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_13__pyx_fuse_2query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_fuse_2__pyx_f_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int64_t __pyx_v_point, int __pyx_skip_dispatch) {
  __Pyx_memviewslice __pyx_v_indices = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_values = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __pyx_t_5numpy_int64_t __pyx_t_9;
  __pyx_t_5numpy_int64_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  int __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  __Pyx_memviewslice __pyx_t_16 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_17;
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  __Pyx_RefNannySetupContext("__pyx_fuse_2query", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pyx_fuse_2query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2374, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_13__pyx_fuse_2query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = __Pyx_PyInt_From_npy_int64(__pyx_v_point); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2374, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2374, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2374, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 2374, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_5) {
            __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2374, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2374, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_3query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_4 = __Pyx_PyInt_From_npy_int64(__pyx_v_point); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 2374, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_7 = __pyx_t_1; __pyx_t_3 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2374, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2374, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 2374, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (__pyx_t_3) {
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_t_4);
          __pyx_t_4 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2374, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pandas/_libs/intervaltree.pxi":2383
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
  __pyx_t_8 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_8) {

    /* "pandas/_libs/intervaltree.pxi":2387
 *             # continue the binary tree structure. Instead, we use linear
 *             # search.
 *             for i in range(self.n_elements):             # <<<<<<<<<<<<<<
 *                 if self.left[i] <= point < self.right[i]:
 *                     result.append(self.indices[i])
 */
    __pyx_t_9 = __pyx_v_self->n_elements;
    __pyx_t_10 = __pyx_t_9;
    for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
      __pyx_v_i = __pyx_t_11;

      /* "pandas/_libs/intervaltree.pxi":2388
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point < self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      __pyx_t_12 = __pyx_v_i;
      __pyx_t_8 = ((*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->left.data + __pyx_t_12 * __pyx_v_self->left.strides[0]) ))) <= __pyx_v_point);
      if (__pyx_t_8) {
        __pyx_t_13 = __pyx_v_i;
        __pyx_t_8 = (__pyx_v_point < (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->right.data + __pyx_t_13 * __pyx_v_self->right.strides[0]) ))));
      }
      __pyx_t_14 = (__pyx_t_8 != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":2389
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point < self.right[i]:
 *                     result.append(self.indices[i])             # <<<<<<<<<<<<<<
 *         else:
 *             # There are child nodes. Based on comparing our query to the pivot,
 */
        __pyx_t_15 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->indices.data + __pyx_t_15 * __pyx_v_self->indices.strides[0]) ))));

        /* "pandas/_libs/intervaltree.pxi":2388
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point < self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      }
    }

    /* "pandas/_libs/intervaltree.pxi":2383
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
    goto __pyx_L3;
  }

  /* "pandas/_libs/intervaltree.pxi":2393
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
  /*else*/ {
    __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":2394
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:
 *                 values = self.center_left_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 */
      __pyx_t_16 = __pyx_v_self->center_left_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":2395
 *             if point < self.pivot:
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:
 */
      __pyx_t_16 = __pyx_v_self->center_left_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_indices = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":2396
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):             # <<<<<<<<<<<<<<
 *                     if not values[i] <= point:
 *                         break
 */
      __pyx_t_9 = __pyx_v_self->n_center;
      __pyx_t_10 = __pyx_t_9;
      for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":2397
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_17 = __pyx_v_i;
        __pyx_t_14 = ((!(((*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_17 * __pyx_v_values.strides[0]) ))) <= __pyx_v_point) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":2398
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:
 */
          goto __pyx_L9_break;

          /* "pandas/_libs/intervaltree.pxi":2397
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":2399
 *                     if not values[i] <= point:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 */
        __pyx_t_18 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_18 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L9_break:;

      /* "pandas/_libs/intervaltree.pxi":2400
 *                         break
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->left_node->max_right) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":2401
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)             # <<<<<<<<<<<<<<
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode *)__pyx_v_self->left_node->__pyx_vtab)->__pyx_fuse_2query(__pyx_v_self->left_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2401, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":2400
 *                         break
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":2393
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":2402
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
    __pyx_t_14 = ((__pyx_v_point > __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":2403
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 *                 values = self.center_right_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 */
      __pyx_t_16 = __pyx_v_self->center_right_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":2404
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:
 */
      __pyx_t_16 = __pyx_v_self->center_right_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_indices = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":2405
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):             # <<<<<<<<<<<<<<
 *                     if not point < values[i]:
 *                         break
 */
      for (__pyx_t_11 = (__pyx_v_self->n_center - 1); __pyx_t_11 > -1L; __pyx_t_11-=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":2406
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_19 = __pyx_v_i;
        __pyx_t_14 = ((!((__pyx_v_point < (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_19 * __pyx_v_values.strides[0]) )))) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":2407
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:
 */
          goto __pyx_L13_break;

          /* "pandas/_libs/intervaltree.pxi":2406
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":2408
 *                     if not point < values[i]:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if self.right_node.min_left <= point:
 *                     self.right_node.query(result, point)
 */
        __pyx_t_20 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_20 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L13_break:;

      /* "pandas/_libs/intervaltree.pxi":2409
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      __pyx_t_14 = ((__pyx_v_self->right_node->min_left <= __pyx_v_point) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":2410
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:
 *                     self.right_node.query(result, point)             # <<<<<<<<<<<<<<
 *             else:
 *                 result.extend(self.center_left_indices)
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode *)__pyx_v_self->right_node->__pyx_vtab)->__pyx_fuse_2query(__pyx_v_self->right_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2410, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":2409
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":2402
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":2412
 *                     self.right_node.query(result, point)
 *             else:
 *                 result.extend(self.center_left_indices)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
    /*else*/ {
      __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->extend(__pyx_v_result, __pyx_v_self->center_left_indices); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2412, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_L7:;
  }
  __pyx_L3:;

  /* "pandas/_libs/intervaltree.pxi":2374
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __PYX_XDEC_MEMVIEW(&__pyx_t_16, 1);
  __Pyx_AddTraceback("pandas._libs.interval.Int64ClosedLeftIntervalNode.query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_indices, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_values, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_13__pyx_fuse_2query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_2__pyx_mdef_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_13__pyx_fuse_2query = {"__pyx_fuse_2query", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_13__pyx_fuse_2query, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_2query};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_13__pyx_fuse_2query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result = 0;
  __pyx_t_5numpy_int64_t __pyx_v_point;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fuse_2query (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_result,&__pyx_n_s_point,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_result)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_point)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_2query", 1, 2, 2, 1); __PYX_ERR(1, 2374, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_2query") < 0)) __PYX_ERR(1, 2374, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_result = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)values[0]);
    __pyx_v_point = __Pyx_PyInt_As_npy_int64(values[1]); if (unlikely((__pyx_v_point == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(1, 2374, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_2query", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 2374, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Int64ClosedLeftIntervalNode.__pyx_fuse_2query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_result), __pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector, 1, "result", 0))) __PYX_ERR(1, 2374, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_12__pyx_fuse_2query(((struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode *)__pyx_v_self), __pyx_v_result, __pyx_v_point);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_12__pyx_fuse_2query(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int64_t __pyx_v_point) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__pyx_fuse_2query", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_fuse_2__pyx_f_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_query(__pyx_v_self, __pyx_v_result, __pyx_v_point, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2374, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Int64ClosedLeftIntervalNode.__pyx_fuse_2query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_15__pyx_fuse_3query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_fuse_3__pyx_f_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int32_t __pyx_v_point, int __pyx_skip_dispatch) {
  __Pyx_memviewslice __pyx_v_indices = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_values = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __pyx_t_5numpy_int64_t __pyx_t_9;
  __pyx_t_5numpy_int64_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  int __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  __Pyx_memviewslice __pyx_t_16 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_17;
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  __Pyx_RefNannySetupContext("__pyx_fuse_3query", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pyx_fuse_3query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2374, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_15__pyx_fuse_3query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = __Pyx_PyInt_From_npy_int32(__pyx_v_point); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2374, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2374, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2374, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 2374, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_5) {
            __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2374, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2374, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_3query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_4 = __Pyx_PyInt_From_npy_int32(__pyx_v_point); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 2374, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_7 = __pyx_t_1; __pyx_t_3 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2374, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2374, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 2374, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (__pyx_t_3) {
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_t_4);
          __pyx_t_4 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2374, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pandas/_libs/intervaltree.pxi":2383
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
  __pyx_t_8 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_8) {

    /* "pandas/_libs/intervaltree.pxi":2387
 *             # continue the binary tree structure. Instead, we use linear
 *             # search.
 *             for i in range(self.n_elements):             # <<<<<<<<<<<<<<
 *                 if self.left[i] <= point < self.right[i]:
 *                     result.append(self.indices[i])
 */
    __pyx_t_9 = __pyx_v_self->n_elements;
    __pyx_t_10 = __pyx_t_9;
    for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
      __pyx_v_i = __pyx_t_11;

      /* "pandas/_libs/intervaltree.pxi":2388
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point < self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      __pyx_t_12 = __pyx_v_i;
      __pyx_t_8 = ((*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->left.data + __pyx_t_12 * __pyx_v_self->left.strides[0]) ))) <= __pyx_v_point);
      if (__pyx_t_8) {
        __pyx_t_13 = __pyx_v_i;
        __pyx_t_8 = (__pyx_v_point < (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->right.data + __pyx_t_13 * __pyx_v_self->right.strides[0]) ))));
      }
      __pyx_t_14 = (__pyx_t_8 != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":2389
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point < self.right[i]:
 *                     result.append(self.indices[i])             # <<<<<<<<<<<<<<
 *         else:
 *             # There are child nodes. Based on comparing our query to the pivot,
 */
        __pyx_t_15 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->indices.data + __pyx_t_15 * __pyx_v_self->indices.strides[0]) ))));

        /* "pandas/_libs/intervaltree.pxi":2388
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point < self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      }
    }

    /* "pandas/_libs/intervaltree.pxi":2383
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
    goto __pyx_L3;
  }

  /* "pandas/_libs/intervaltree.pxi":2393
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
  /*else*/ {
    __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":2394
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:
 *                 values = self.center_left_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 */
      __pyx_t_16 = __pyx_v_self->center_left_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":2395
 *             if point < self.pivot:
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:
 */
      __pyx_t_16 = __pyx_v_self->center_left_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_indices = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":2396
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):             # <<<<<<<<<<<<<<
 *                     if not values[i] <= point:
 *                         break
 */
      __pyx_t_9 = __pyx_v_self->n_center;
      __pyx_t_10 = __pyx_t_9;
      for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":2397
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_17 = __pyx_v_i;
        __pyx_t_14 = ((!(((*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_17 * __pyx_v_values.strides[0]) ))) <= __pyx_v_point) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":2398
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:
 */
          goto __pyx_L9_break;

          /* "pandas/_libs/intervaltree.pxi":2397
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":2399
 *                     if not values[i] <= point:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 */
        __pyx_t_18 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_18 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L9_break:;

      /* "pandas/_libs/intervaltree.pxi":2400
 *                         break
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->left_node->max_right) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":2401
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)             # <<<<<<<<<<<<<<
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode *)__pyx_v_self->left_node->__pyx_vtab)->__pyx_fuse_3query(__pyx_v_self->left_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2401, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":2400
 *                         break
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":2393
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":2402
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
    __pyx_t_14 = ((__pyx_v_point > __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":2403
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 *                 values = self.center_right_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 */
      __pyx_t_16 = __pyx_v_self->center_right_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":2404
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:
 */
      __pyx_t_16 = __pyx_v_self->center_right_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_indices = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":2405
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):             # <<<<<<<<<<<<<<
 *                     if not point < values[i]:
 *                         break
 */
      for (__pyx_t_11 = (__pyx_v_self->n_center - 1); __pyx_t_11 > -1L; __pyx_t_11-=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":2406
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_19 = __pyx_v_i;
        __pyx_t_14 = ((!((__pyx_v_point < (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_19 * __pyx_v_values.strides[0]) )))) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":2407
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:
 */
          goto __pyx_L13_break;

          /* "pandas/_libs/intervaltree.pxi":2406
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":2408
 *                     if not point < values[i]:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if self.right_node.min_left <= point:
 *                     self.right_node.query(result, point)
 */
        __pyx_t_20 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_20 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L13_break:;

      /* "pandas/_libs/intervaltree.pxi":2409
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      __pyx_t_14 = ((__pyx_v_self->right_node->min_left <= __pyx_v_point) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":2410
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:
 *                     self.right_node.query(result, point)             # <<<<<<<<<<<<<<
 *             else:
 *                 result.extend(self.center_left_indices)
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode *)__pyx_v_self->right_node->__pyx_vtab)->__pyx_fuse_3query(__pyx_v_self->right_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2410, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":2409
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":2402
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":2412
 *                     self.right_node.query(result, point)
 *             else:
 *                 result.extend(self.center_left_indices)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
    /*else*/ {
      __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->extend(__pyx_v_result, __pyx_v_self->center_left_indices); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2412, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_L7:;
  }
  __pyx_L3:;

  /* "pandas/_libs/intervaltree.pxi":2374
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __PYX_XDEC_MEMVIEW(&__pyx_t_16, 1);
  __Pyx_AddTraceback("pandas._libs.interval.Int64ClosedLeftIntervalNode.query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_indices, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_values, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_15__pyx_fuse_3query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_3__pyx_mdef_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_15__pyx_fuse_3query = {"__pyx_fuse_3query", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_15__pyx_fuse_3query, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_2query};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_15__pyx_fuse_3query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result = 0;
  __pyx_t_5numpy_int32_t __pyx_v_point;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fuse_3query (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_result,&__pyx_n_s_point,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_result)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_point)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_3query", 1, 2, 2, 1); __PYX_ERR(1, 2374, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_3query") < 0)) __PYX_ERR(1, 2374, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_result = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)values[0]);
    __pyx_v_point = __Pyx_PyInt_As_npy_int32(values[1]); if (unlikely((__pyx_v_point == ((npy_int32)-1)) && PyErr_Occurred())) __PYX_ERR(1, 2374, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_3query", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 2374, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Int64ClosedLeftIntervalNode.__pyx_fuse_3query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_result), __pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector, 1, "result", 0))) __PYX_ERR(1, 2374, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_14__pyx_fuse_3query(((struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode *)__pyx_v_self), __pyx_v_result, __pyx_v_point);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_14__pyx_fuse_3query(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int32_t __pyx_v_point) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__pyx_fuse_3query", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_fuse_3__pyx_f_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_query(__pyx_v_self, __pyx_v_result, __pyx_v_point, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2374, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Int64ClosedLeftIntervalNode.__pyx_fuse_3query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_17__pyx_fuse_4query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_fuse_4__pyx_f_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_uint64_t __pyx_v_point, int __pyx_skip_dispatch) {
  __Pyx_memviewslice __pyx_v_indices = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_values = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __pyx_t_5numpy_int64_t __pyx_t_9;
  __pyx_t_5numpy_int64_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  int __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  __Pyx_memviewslice __pyx_t_16 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_17;
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  __Pyx_RefNannySetupContext("__pyx_fuse_4query", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pyx_fuse_4query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2374, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_17__pyx_fuse_4query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = __Pyx_PyInt_From_npy_uint64(__pyx_v_point); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2374, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2374, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2374, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 2374, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_5) {
            __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2374, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2374, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_3query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_4 = __Pyx_PyInt_From_npy_uint64(__pyx_v_point); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 2374, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_7 = __pyx_t_1; __pyx_t_3 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2374, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2374, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 2374, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (__pyx_t_3) {
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_t_4);
          __pyx_t_4 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2374, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pandas/_libs/intervaltree.pxi":2383
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
  __pyx_t_8 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_8) {

    /* "pandas/_libs/intervaltree.pxi":2387
 *             # continue the binary tree structure. Instead, we use linear
 *             # search.
 *             for i in range(self.n_elements):             # <<<<<<<<<<<<<<
 *                 if self.left[i] <= point < self.right[i]:
 *                     result.append(self.indices[i])
 */
    __pyx_t_9 = __pyx_v_self->n_elements;
    __pyx_t_10 = __pyx_t_9;
    for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
      __pyx_v_i = __pyx_t_11;

      /* "pandas/_libs/intervaltree.pxi":2388
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point < self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      __pyx_t_12 = __pyx_v_i;
      __pyx_t_8 = ((*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->left.data + __pyx_t_12 * __pyx_v_self->left.strides[0]) ))) <= __pyx_v_point);
      if (__pyx_t_8) {
        __pyx_t_13 = __pyx_v_i;
        __pyx_t_8 = (__pyx_v_point < (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->right.data + __pyx_t_13 * __pyx_v_self->right.strides[0]) ))));
      }
      __pyx_t_14 = (__pyx_t_8 != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":2389
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point < self.right[i]:
 *                     result.append(self.indices[i])             # <<<<<<<<<<<<<<
 *         else:
 *             # There are child nodes. Based on comparing our query to the pivot,
 */
        __pyx_t_15 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->indices.data + __pyx_t_15 * __pyx_v_self->indices.strides[0]) ))));

        /* "pandas/_libs/intervaltree.pxi":2388
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point < self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      }
    }

    /* "pandas/_libs/intervaltree.pxi":2383
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
    goto __pyx_L3;
  }

  /* "pandas/_libs/intervaltree.pxi":2393
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
  /*else*/ {
    __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":2394
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:
 *                 values = self.center_left_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 */
      __pyx_t_16 = __pyx_v_self->center_left_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":2395
 *             if point < self.pivot:
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:
 */
      __pyx_t_16 = __pyx_v_self->center_left_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_indices = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":2396
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):             # <<<<<<<<<<<<<<
 *                     if not values[i] <= point:
 *                         break
 */
      __pyx_t_9 = __pyx_v_self->n_center;
      __pyx_t_10 = __pyx_t_9;
      for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":2397
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_17 = __pyx_v_i;
        __pyx_t_14 = ((!(((*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_17 * __pyx_v_values.strides[0]) ))) <= __pyx_v_point) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":2398
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:
 */
          goto __pyx_L9_break;

          /* "pandas/_libs/intervaltree.pxi":2397
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":2399
 *                     if not values[i] <= point:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 */
        __pyx_t_18 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_18 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L9_break:;

      /* "pandas/_libs/intervaltree.pxi":2400
 *                         break
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->left_node->max_right) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":2401
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)             # <<<<<<<<<<<<<<
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode *)__pyx_v_self->left_node->__pyx_vtab)->__pyx_fuse_4query(__pyx_v_self->left_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2401, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":2400
 *                         break
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":2393
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":2402
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
    __pyx_t_14 = ((__pyx_v_point > __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":2403
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 *                 values = self.center_right_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 */
      __pyx_t_16 = __pyx_v_self->center_right_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":2404
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:
 */
      __pyx_t_16 = __pyx_v_self->center_right_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_indices = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":2405
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):             # <<<<<<<<<<<<<<
 *                     if not point < values[i]:
 *                         break
 */
      for (__pyx_t_11 = (__pyx_v_self->n_center - 1); __pyx_t_11 > -1L; __pyx_t_11-=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":2406
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_19 = __pyx_v_i;
        __pyx_t_14 = ((!((__pyx_v_point < (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_19 * __pyx_v_values.strides[0]) )))) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":2407
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:
 */
          goto __pyx_L13_break;

          /* "pandas/_libs/intervaltree.pxi":2406
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":2408
 *                     if not point < values[i]:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if self.right_node.min_left <= point:
 *                     self.right_node.query(result, point)
 */
        __pyx_t_20 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_20 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L13_break:;

      /* "pandas/_libs/intervaltree.pxi":2409
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      __pyx_t_14 = ((__pyx_v_self->right_node->min_left <= __pyx_v_point) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":2410
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:
 *                     self.right_node.query(result, point)             # <<<<<<<<<<<<<<
 *             else:
 *                 result.extend(self.center_left_indices)
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode *)__pyx_v_self->right_node->__pyx_vtab)->__pyx_fuse_4query(__pyx_v_self->right_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2410, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":2409
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":2402
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":2412
 *                     self.right_node.query(result, point)
 *             else:
 *                 result.extend(self.center_left_indices)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
    /*else*/ {
      __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->extend(__pyx_v_result, __pyx_v_self->center_left_indices); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2412, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_L7:;
  }
  __pyx_L3:;

  /* "pandas/_libs/intervaltree.pxi":2374
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __PYX_XDEC_MEMVIEW(&__pyx_t_16, 1);
  __Pyx_AddTraceback("pandas._libs.interval.Int64ClosedLeftIntervalNode.query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_indices, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_values, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_17__pyx_fuse_4query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_4__pyx_mdef_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_17__pyx_fuse_4query = {"__pyx_fuse_4query", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_17__pyx_fuse_4query, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_2query};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_17__pyx_fuse_4query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result = 0;
  __pyx_t_5numpy_uint64_t __pyx_v_point;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fuse_4query (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_result,&__pyx_n_s_point,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_result)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_point)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_4query", 1, 2, 2, 1); __PYX_ERR(1, 2374, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_4query") < 0)) __PYX_ERR(1, 2374, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_result = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)values[0]);
    __pyx_v_point = __Pyx_PyInt_As_npy_uint64(values[1]); if (unlikely((__pyx_v_point == ((npy_uint64)-1)) && PyErr_Occurred())) __PYX_ERR(1, 2374, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_4query", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 2374, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Int64ClosedLeftIntervalNode.__pyx_fuse_4query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_result), __pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector, 1, "result", 0))) __PYX_ERR(1, 2374, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_16__pyx_fuse_4query(((struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode *)__pyx_v_self), __pyx_v_result, __pyx_v_point);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_16__pyx_fuse_4query(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_uint64_t __pyx_v_point) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__pyx_fuse_4query", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_fuse_4__pyx_f_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_query(__pyx_v_self, __pyx_v_result, __pyx_v_point, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2374, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Int64ClosedLeftIntervalNode.__pyx_fuse_4query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":2414
 *                 result.extend(self.center_left_indices)
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         if self.is_leaf_node:
 *             return ('<Int64ClosedLeftIntervalNode: '
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_5__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_5__repr__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_4__repr__(((struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_4__repr__(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode *__pyx_v_self) {
  PyObject *__pyx_v_n_left = NULL;
  PyObject *__pyx_v_n_right = NULL;
  PyObject *__pyx_v_n_center = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  Py_UCS4 __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "pandas/_libs/intervaltree.pxi":2415
 * 
 *     def __repr__(self):
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             return ('<Int64ClosedLeftIntervalNode: '
 *                     '%s elements (terminal)>' % self.n_elements)
 */
  __pyx_t_1 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_1) {

    /* "pandas/_libs/intervaltree.pxi":2416
 *     def __repr__(self):
 *         if self.is_leaf_node:
 *             return ('<Int64ClosedLeftIntervalNode: '             # <<<<<<<<<<<<<<
 *                     '%s elements (terminal)>' % self.n_elements)
 *         else:
 */
    __Pyx_XDECREF(__pyx_r);

    /* "pandas/_libs/intervaltree.pxi":2417
 *         if self.is_leaf_node:
 *             return ('<Int64ClosedLeftIntervalNode: '
 *                     '%s elements (terminal)>' % self.n_elements)             # <<<<<<<<<<<<<<
 *         else:
 *             n_left = self.left_node.n_elements
 */
    __pyx_t_2 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_elements); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2417, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PyUnicode_Format(__pyx_kp_u_Int64ClosedLeftIntervalNode_s_e, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2417, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pandas/_libs/intervaltree.pxi":2415
 * 
 *     def __repr__(self):
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             return ('<Int64ClosedLeftIntervalNode: '
 *                     '%s elements (terminal)>' % self.n_elements)
 */
  }

  /* "pandas/_libs/intervaltree.pxi":2419
 *                     '%s elements (terminal)>' % self.n_elements)
 *         else:
 *             n_left = self.left_node.n_elements             # <<<<<<<<<<<<<<
 *             n_right = self.right_node.n_elements
 *             n_center = self.n_elements - n_left - n_right
 */
  /*else*/ {
    __pyx_t_3 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->left_node->n_elements); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2419, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_v_n_left = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "pandas/_libs/intervaltree.pxi":2420
 *         else:
 *             n_left = self.left_node.n_elements
 *             n_right = self.right_node.n_elements             # <<<<<<<<<<<<<<
 *             n_center = self.n_elements - n_left - n_right
 *             return ('<Int64ClosedLeftIntervalNode: '
 */
    __pyx_t_3 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->right_node->n_elements); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2420, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_v_n_right = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "pandas/_libs/intervaltree.pxi":2421
 *             n_left = self.left_node.n_elements
 *             n_right = self.right_node.n_elements
 *             n_center = self.n_elements - n_left - n_right             # <<<<<<<<<<<<<<
 *             return ('<Int64ClosedLeftIntervalNode: '
 *                     'pivot %s, %s elements (%s left, %s right, %s '
 */
    __pyx_t_3 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_elements); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2421, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = PyNumber_Subtract(__pyx_t_3, __pyx_v_n_left); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2421, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyNumber_Subtract(__pyx_t_2, __pyx_v_n_right); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2421, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_n_center = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "pandas/_libs/intervaltree.pxi":2422
 *             n_right = self.right_node.n_elements
 *             n_center = self.n_elements - n_left - n_right
 *             return ('<Int64ClosedLeftIntervalNode: '             # <<<<<<<<<<<<<<
 *                     'pivot %s, %s elements (%s left, %s right, %s '
 *                     'overlapping)>' % (self.pivot, self.n_elements,
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = PyTuple_New(11); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2422, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = 0;
    __pyx_t_5 = 127;
    __Pyx_INCREF(__pyx_kp_u_Int64ClosedLeftIntervalNode_piv);
    __pyx_t_4 += 36;
    __Pyx_GIVEREF(__pyx_kp_u_Int64ClosedLeftIntervalNode_piv);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_kp_u_Int64ClosedLeftIntervalNode_piv);

    /* "pandas/_libs/intervaltree.pxi":2424
 *             return ('<Int64ClosedLeftIntervalNode: '
 *                     'pivot %s, %s elements (%s left, %s right, %s '
 *                     'overlapping)>' % (self.pivot, self.n_elements,             # <<<<<<<<<<<<<<
 *                                        n_left, n_right, n_center))
 * 
 */
    __pyx_t_2 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->pivot); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2424, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_t_2), __pyx_empty_unicode); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 2424, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_5 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_6) > __pyx_t_5) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_6) : __pyx_t_5;
    __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_6);
    __pyx_t_6 = 0;
    __Pyx_INCREF(__pyx_kp_u__18);
    __pyx_t_4 += 2;
    __Pyx_GIVEREF(__pyx_kp_u__18);
    PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_kp_u__18);
    __pyx_t_6 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_elements); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 2424, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_2 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_t_6), __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2424, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_5 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) > __pyx_t_5) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) : __pyx_t_5;
    __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_3, 3, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_elements);
    __pyx_t_4 += 11;
    __Pyx_GIVEREF(__pyx_kp_u_elements);
    PyTuple_SET_ITEM(__pyx_t_3, 4, __pyx_kp_u_elements);

    /* "pandas/_libs/intervaltree.pxi":2425
 *                     'pivot %s, %s elements (%s left, %s right, %s '
 *                     'overlapping)>' % (self.pivot, self.n_elements,
 *                                        n_left, n_right, n_center))             # <<<<<<<<<<<<<<
 * 
 *     def counts(self):
 */
    __pyx_t_2 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_v_n_left), __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2425, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) > __pyx_t_5) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) : __pyx_t_5;
    __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_3, 5, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_left_2);
    __pyx_t_4 += 7;
    __Pyx_GIVEREF(__pyx_kp_u_left_2);
    PyTuple_SET_ITEM(__pyx_t_3, 6, __pyx_kp_u_left_2);
    __pyx_t_2 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_v_n_right), __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2425, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) > __pyx_t_5) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) : __pyx_t_5;
    __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_3, 7, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_right_2);
    __pyx_t_4 += 8;
    __Pyx_GIVEREF(__pyx_kp_u_right_2);
    PyTuple_SET_ITEM(__pyx_t_3, 8, __pyx_kp_u_right_2);
    __pyx_t_2 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_v_n_center), __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2425, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) > __pyx_t_5) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) : __pyx_t_5;
    __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_3, 9, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_overlapping);
    __pyx_t_4 += 14;
    __Pyx_GIVEREF(__pyx_kp_u_overlapping);
    PyTuple_SET_ITEM(__pyx_t_3, 10, __pyx_kp_u_overlapping);

    /* "pandas/_libs/intervaltree.pxi":2422
 *             n_right = self.right_node.n_elements
 *             n_center = self.n_elements - n_left - n_right
 *             return ('<Int64ClosedLeftIntervalNode: '             # <<<<<<<<<<<<<<
 *                     'pivot %s, %s elements (%s left, %s right, %s '
 *                     'overlapping)>' % (self.pivot, self.n_elements,
 */
    __pyx_t_2 = __Pyx_PyUnicode_Join(__pyx_t_3, 11, __pyx_t_4, __pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2422, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;
  }

  /* "pandas/_libs/intervaltree.pxi":2414
 *                 result.extend(self.center_left_indices)
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         if self.is_leaf_node:
 *             return ('<Int64ClosedLeftIntervalNode: '
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pandas._libs.interval.Int64ClosedLeftIntervalNode.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_n_left);
  __Pyx_XDECREF(__pyx_v_n_right);
  __Pyx_XDECREF(__pyx_v_n_center);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":2427
 *                                        n_left, n_right, n_center))
 * 
 *     def counts(self):             # <<<<<<<<<<<<<<
 *         """
 *         Inspect counts on this node
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_7counts(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_6counts[] = "\n        Inspect counts on this node\n        useful for debugging purposes\n        ";
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_7counts(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("counts (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_6counts(((struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_6counts(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode *__pyx_v_self) {
  Py_ssize_t __pyx_v_m;
  PyObject *__pyx_v_l = NULL;
  PyObject *__pyx_v_r = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  size_t __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("counts", 0);

  /* "pandas/_libs/intervaltree.pxi":2432
 *         useful for debugging purposes
 *         """
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             return self.n_elements
 *         else:
 */
  __pyx_t_1 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_1) {

    /* "pandas/_libs/intervaltree.pxi":2433
 *         """
 *         if self.is_leaf_node:
 *             return self.n_elements             # <<<<<<<<<<<<<<
 *         else:
 *             m = len(self.center_left_values)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_elements); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2433, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "pandas/_libs/intervaltree.pxi":2432
 *         useful for debugging purposes
 *         """
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             return self.n_elements
 *         else:
 */
  }

  /* "pandas/_libs/intervaltree.pxi":2435
 *             return self.n_elements
 *         else:
 *             m = len(self.center_left_values)             # <<<<<<<<<<<<<<
 *             l = self.left_node.counts()
 *             r = self.right_node.counts()
 */
  /*else*/ {
    if (unlikely(!__pyx_v_self->center_left_values.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(1, 2435, __pyx_L1_error)}
    __pyx_t_3 = __Pyx_MemoryView_Len(__pyx_v_self->center_left_values); 
    __pyx_v_m = __pyx_t_3;

    /* "pandas/_libs/intervaltree.pxi":2436
 *         else:
 *             m = len(self.center_left_values)
 *             l = self.left_node.counts()             # <<<<<<<<<<<<<<
 *             r = self.right_node.counts()
 *             return (m, (l, r))
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->left_node), __pyx_n_s_counts); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 2436, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2436, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_l = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "pandas/_libs/intervaltree.pxi":2437
 *             m = len(self.center_left_values)
 *             l = self.left_node.counts()
 *             r = self.right_node.counts()             # <<<<<<<<<<<<<<
 *             return (m, (l, r))
 * 
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->right_node), __pyx_n_s_counts); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 2437, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2437, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_r = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "pandas/_libs/intervaltree.pxi":2438
 *             l = self.left_node.counts()
 *             r = self.right_node.counts()
 *             return (m, (l, r))             # <<<<<<<<<<<<<<
 * 
 * NODE_CLASSES['int64',
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = PyInt_FromSsize_t(__pyx_v_m); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2438, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 2438, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_v_l);
    __Pyx_GIVEREF(__pyx_v_l);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_l);
    __Pyx_INCREF(__pyx_v_r);
    __Pyx_GIVEREF(__pyx_v_r);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_v_r);
    __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 2438, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_4);
    __pyx_t_2 = 0;
    __pyx_t_4 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;
  }

  /* "pandas/_libs/intervaltree.pxi":2427
 *                                        n_left, n_right, n_center))
 * 
 *     def counts(self):             # <<<<<<<<<<<<<<
 *         """
 *         Inspect counts on this node
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pandas._libs.interval.Int64ClosedLeftIntervalNode.counts", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_l);
  __Pyx_XDECREF(__pyx_v_r);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":2284
 *         int64_t[:] center_left_indices, center_right_indices, indices
 *         int64_t min_left, max_right
 *         readonly int64_t pivot             # <<<<<<<<<<<<<<
 *         readonly int64_t n_elements, n_center, leaf_size
 *         readonly bint is_leaf_node
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_5pivot_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_5pivot_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_5pivot___get__(((struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_5pivot___get__(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->pivot); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2284, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Int64ClosedLeftIntervalNode.pivot.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":2285
 *         int64_t min_left, max_right
 *         readonly int64_t pivot
 *         readonly int64_t n_elements, n_center, leaf_size             # <<<<<<<<<<<<<<
 *         readonly bint is_leaf_node
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_10n_elements_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_10n_elements_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_10n_elements___get__(((struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_10n_elements___get__(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_elements); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2285, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Int64ClosedLeftIntervalNode.n_elements.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_8n_center_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_8n_center_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_8n_center___get__(((struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_8n_center___get__(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_center); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2285, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Int64ClosedLeftIntervalNode.n_center.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_9leaf_size_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_9leaf_size_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_9leaf_size___get__(((struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_9leaf_size___get__(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->leaf_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2285, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Int64ClosedLeftIntervalNode.leaf_size.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":2286
 *         readonly int64_t pivot
 *         readonly int64_t n_elements, n_center, leaf_size
 *         readonly bint is_leaf_node             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self,
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_12is_leaf_node_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_12is_leaf_node_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_12is_leaf_node___get__(((struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_12is_leaf_node___get__(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->is_leaf_node); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2286, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Int64ClosedLeftIntervalNode.is_leaf_node.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_21__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_21__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_20__reduce_cython__(((struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_20__reduce_cython__(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  int __pyx_t_16;
  int __pyx_t_17;
  int __pyx_t_18;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.center_left_indices, self.center_left_values, self.center_right_indices, self.center_right_values, self.indices, self.is_leaf_node, self.leaf_size, self.left, self.left_node, self.max_right, self.min_left, self.n_center, self.n_elements, self.pivot, self.right, self.right_node)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  if (unlikely(!__pyx_v_self->center_left_indices.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_1 = __pyx_memoryview_fromslice(__pyx_v_self->center_left_indices, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_int64_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_int64_t, 0);; if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (unlikely(!__pyx_v_self->center_left_values.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_2 = __pyx_memoryview_fromslice(__pyx_v_self->center_left_values, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_int64_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_int64_t, 0);; if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (unlikely(!__pyx_v_self->center_right_indices.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_3 = __pyx_memoryview_fromslice(__pyx_v_self->center_right_indices, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_int64_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_int64_t, 0);; if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely(!__pyx_v_self->center_right_values.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_4 = __pyx_memoryview_fromslice(__pyx_v_self->center_right_values, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_int64_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_int64_t, 0);; if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (unlikely(!__pyx_v_self->indices.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_5 = __pyx_memoryview_fromslice(__pyx_v_self->indices, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_int64_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_int64_t, 0);; if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyBool_FromLong(__pyx_v_self->is_leaf_node); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->leaf_size); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (unlikely(!__pyx_v_self->left.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_8 = __pyx_memoryview_fromslice(__pyx_v_self->left, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_int64_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_int64_t, 0);; if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->max_right); if (unlikely(!__pyx_t_9)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_10 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->min_left); if (unlikely(!__pyx_t_10)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_11 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_center); if (unlikely(!__pyx_t_11)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_12 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_elements); if (unlikely(!__pyx_t_12)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __pyx_t_13 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->pivot); if (unlikely(!__pyx_t_13)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  if (unlikely(!__pyx_v_self->right.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_14 = __pyx_memoryview_fromslice(__pyx_v_self->right, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_int64_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_int64_t, 0);; if (unlikely(!__pyx_t_14)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __pyx_t_15 = PyTuple_New(16); if (unlikely(!__pyx_t_15)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_15, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_15, 1, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_15, 2, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_15, 3, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_15, 4, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_15, 5, __pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_15, 6, __pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_15, 7, __pyx_t_8);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->left_node));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self->left_node));
  PyTuple_SET_ITEM(__pyx_t_15, 8, ((PyObject *)__pyx_v_self->left_node));
  __Pyx_GIVEREF(__pyx_t_9);
  PyTuple_SET_ITEM(__pyx_t_15, 9, __pyx_t_9);
  __Pyx_GIVEREF(__pyx_t_10);
  PyTuple_SET_ITEM(__pyx_t_15, 10, __pyx_t_10);
  __Pyx_GIVEREF(__pyx_t_11);
  PyTuple_SET_ITEM(__pyx_t_15, 11, __pyx_t_11);
  __Pyx_GIVEREF(__pyx_t_12);
  PyTuple_SET_ITEM(__pyx_t_15, 12, __pyx_t_12);
  __Pyx_GIVEREF(__pyx_t_13);
  PyTuple_SET_ITEM(__pyx_t_15, 13, __pyx_t_13);
  __Pyx_GIVEREF(__pyx_t_14);
  PyTuple_SET_ITEM(__pyx_t_15, 14, __pyx_t_14);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->right_node));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self->right_node));
  PyTuple_SET_ITEM(__pyx_t_15, 15, ((PyObject *)__pyx_v_self->right_node));
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  __pyx_t_6 = 0;
  __pyx_t_7 = 0;
  __pyx_t_8 = 0;
  __pyx_t_9 = 0;
  __pyx_t_10 = 0;
  __pyx_t_11 = 0;
  __pyx_t_12 = 0;
  __pyx_t_13 = 0;
  __pyx_t_14 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_15);
  __pyx_t_15 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.center_left_indices, self.center_left_values, self.center_right_indices, self.center_right_values, self.indices, self.is_leaf_node, self.leaf_size, self.left, self.left_node, self.max_right, self.min_left, self.n_center, self.n_elements, self.pivot, self.right, self.right_node)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_15 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_15)) __PYX_ERR(2, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __pyx_v__dict = __pyx_t_15;
  __pyx_t_15 = 0;

  /* "(tree fragment)":7
 *     state = (self.center_left_indices, self.center_left_values, self.center_right_indices, self.center_right_values, self.indices, self.is_leaf_node, self.leaf_size, self.left, self.left_node, self.max_right, self.min_left, self.n_center, self.n_elements, self.pivot, self.right, self.right_node)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_16 = (__pyx_v__dict != Py_None);
  __pyx_t_17 = (__pyx_t_16 != 0);
  if (__pyx_t_17) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_15 = PyTuple_New(1); if (unlikely(!__pyx_t_15)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    PyTuple_SET_ITEM(__pyx_t_15, 0, __pyx_v__dict);
    __pyx_t_14 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_15); if (unlikely(!__pyx_t_14)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_14));
    __pyx_t_14 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.left_node is not None or self.right_node is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.center_left_indices, self.center_left_values, self.center_right_indices, self.center_right_values, self.indices, self.is_leaf_node, self.leaf_size, self.left, self.left_node, self.max_right, self.min_left, self.n_center, self.n_elements, self.pivot, self.right, self.right_node)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.left_node is not None or self.right_node is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_Int64ClosedLeftIntervalNode, (type(self), 0x7a104e5, None), state
 */
  /*else*/ {
    __pyx_t_16 = (((PyObject *)__pyx_v_self->left_node) != Py_None);
    __pyx_t_18 = (__pyx_t_16 != 0);
    if (!__pyx_t_18) {
    } else {
      __pyx_t_17 = __pyx_t_18;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_18 = (((PyObject *)__pyx_v_self->right_node) != Py_None);
    __pyx_t_16 = (__pyx_t_18 != 0);
    __pyx_t_17 = __pyx_t_16;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_17;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.left_node is not None or self.right_node is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Int64ClosedLeftIntervalNode, (type(self), 0x7a104e5, None), state
 *     else:
 */
  __pyx_t_17 = (__pyx_v_use_setstate != 0);
  if (__pyx_t_17) {

    /* "(tree fragment)":13
 *         use_setstate = self.left_node is not None or self.right_node is not None
 *     if use_setstate:
 *         return __pyx_unpickle_Int64ClosedLeftIntervalNode, (type(self), 0x7a104e5, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_Int64ClosedLeftIntervalNode, (type(self), 0x7a104e5, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_14, __pyx_n_s_pyx_unpickle_Int64ClosedLeftIn); if (unlikely(!__pyx_t_14)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __pyx_t_15 = PyTuple_New(3); if (unlikely(!__pyx_t_15)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_15, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_127993061);
    __Pyx_GIVEREF(__pyx_int_127993061);
    PyTuple_SET_ITEM(__pyx_t_15, 1, __pyx_int_127993061);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_15, 2, Py_None);
    __pyx_t_13 = PyTuple_New(3); if (unlikely(!__pyx_t_13)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __Pyx_GIVEREF(__pyx_t_14);
    PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_t_14);
    __Pyx_GIVEREF(__pyx_t_15);
    PyTuple_SET_ITEM(__pyx_t_13, 1, __pyx_t_15);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_13, 2, __pyx_v_state);
    __pyx_t_14 = 0;
    __pyx_t_15 = 0;
    __pyx_r = __pyx_t_13;
    __pyx_t_13 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.left_node is not None or self.right_node is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Int64ClosedLeftIntervalNode, (type(self), 0x7a104e5, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_Int64ClosedLeftIntervalNode, (type(self), 0x7a104e5, None), state
 *     else:
 *         return __pyx_unpickle_Int64ClosedLeftIntervalNode, (type(self), 0x7a104e5, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Int64ClosedLeftIntervalNode__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_13, __pyx_n_s_pyx_unpickle_Int64ClosedLeftIn); if (unlikely(!__pyx_t_13)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __pyx_t_15 = PyTuple_New(3); if (unlikely(!__pyx_t_15)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_15, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_127993061);
    __Pyx_GIVEREF(__pyx_int_127993061);
    PyTuple_SET_ITEM(__pyx_t_15, 1, __pyx_int_127993061);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_15, 2, __pyx_v_state);
    __pyx_t_14 = PyTuple_New(2); if (unlikely(!__pyx_t_14)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_GIVEREF(__pyx_t_13);
    PyTuple_SET_ITEM(__pyx_t_14, 0, __pyx_t_13);
    __Pyx_GIVEREF(__pyx_t_15);
    PyTuple_SET_ITEM(__pyx_t_14, 1, __pyx_t_15);
    __pyx_t_13 = 0;
    __pyx_t_15 = 0;
    __pyx_r = __pyx_t_14;
    __pyx_t_14 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_AddTraceback("pandas._libs.interval.Int64ClosedLeftIntervalNode.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Int64ClosedLeftIntervalNode, (type(self), 0x7a104e5, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Int64ClosedLeftIntervalNode__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_23__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_23__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_22__setstate_cython__(((struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_22__setstate_cython__(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_Int64ClosedLeftIntervalNode, (type(self), 0x7a104e5, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Int64ClosedLeftIntervalNode__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(2, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_6pandas_5_libs_8interval___pyx_unpickle_Int64ClosedLeftIntervalNode__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Int64ClosedLeftIntervalNode, (type(self), 0x7a104e5, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Int64ClosedLeftIntervalNode__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Int64ClosedLeftIntervalNode.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":2458
 *         readonly bint is_leaf_node
 * 
 *     def __init__(self,             # <<<<<<<<<<<<<<
 *                  ndarray[int64_t, ndim=1] left,
 *                  ndarray[int64_t, ndim=1] right,
 */

/* Python wrapper */
static int __pyx_pw_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyArrayObject *__pyx_v_left = 0;
  PyArrayObject *__pyx_v_right = 0;
  PyArrayObject *__pyx_v_indices = 0;
  __pyx_t_5numpy_int64_t __pyx_v_leaf_size;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_left,&__pyx_n_s_right,&__pyx_n_s_indices,&__pyx_n_s_leaf_size,0};
    PyObject* values[4] = {0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_left)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_right)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 4, 4, 1); __PYX_ERR(1, 2458, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_indices)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 4, 4, 2); __PYX_ERR(1, 2458, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_leaf_size)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 4, 4, 3); __PYX_ERR(1, 2458, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(1, 2458, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 4) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
    }
    __pyx_v_left = ((PyArrayObject *)values[0]);
    __pyx_v_right = ((PyArrayObject *)values[1]);
    __pyx_v_indices = ((PyArrayObject *)values[2]);
    __pyx_v_leaf_size = __Pyx_PyInt_As_npy_int64(values[3]); if (unlikely((__pyx_v_leaf_size == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(1, 2462, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 4, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 2458, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Int64ClosedRightIntervalNode.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_left), __pyx_ptype_5numpy_ndarray, 1, "left", 0))) __PYX_ERR(1, 2459, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_right), __pyx_ptype_5numpy_ndarray, 1, "right", 0))) __PYX_ERR(1, 2460, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_indices), __pyx_ptype_5numpy_ndarray, 1, "indices", 0))) __PYX_ERR(1, 2461, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode___init__(((struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode *)__pyx_v_self), __pyx_v_left, __pyx_v_right, __pyx_v_indices, __pyx_v_leaf_size);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode___init__(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode *__pyx_v_self, PyArrayObject *__pyx_v_left, PyArrayObject *__pyx_v_right, PyArrayObject *__pyx_v_indices, __pyx_t_5numpy_int64_t __pyx_v_leaf_size) {
  PyObject *__pyx_v_left_set = NULL;
  PyObject *__pyx_v_right_set = NULL;
  PyObject *__pyx_v_center_set = NULL;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_indices;
  __Pyx_Buffer __pyx_pybuffer_indices;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_left;
  __Pyx_Buffer __pyx_pybuffer_left;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_right;
  __Pyx_Buffer __pyx_pybuffer_right;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __pyx_t_5numpy_int64_t __pyx_t_6;
  __Pyx_memviewslice __pyx_t_7 = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  __Pyx_memviewslice __pyx_t_10 = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *(*__pyx_t_11)(PyObject *);
  size_t __pyx_t_12;
  __Pyx_RefNannySetupContext("__init__", 0);
  __pyx_pybuffer_left.pybuffer.buf = NULL;
  __pyx_pybuffer_left.refcount = 0;
  __pyx_pybuffernd_left.data = NULL;
  __pyx_pybuffernd_left.rcbuffer = &__pyx_pybuffer_left;
  __pyx_pybuffer_right.pybuffer.buf = NULL;
  __pyx_pybuffer_right.refcount = 0;
  __pyx_pybuffernd_right.data = NULL;
  __pyx_pybuffernd_right.rcbuffer = &__pyx_pybuffer_right;
  __pyx_pybuffer_indices.pybuffer.buf = NULL;
  __pyx_pybuffer_indices.refcount = 0;
  __pyx_pybuffernd_indices.data = NULL;
  __pyx_pybuffernd_indices.rcbuffer = &__pyx_pybuffer_indices;
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_left.rcbuffer->pybuffer, (PyObject*)__pyx_v_left, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 2458, __pyx_L1_error)
  }
  __pyx_pybuffernd_left.diminfo[0].strides = __pyx_pybuffernd_left.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_left.diminfo[0].shape = __pyx_pybuffernd_left.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_right.rcbuffer->pybuffer, (PyObject*)__pyx_v_right, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 2458, __pyx_L1_error)
  }
  __pyx_pybuffernd_right.diminfo[0].strides = __pyx_pybuffernd_right.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_right.diminfo[0].shape = __pyx_pybuffernd_right.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_indices.rcbuffer->pybuffer, (PyObject*)__pyx_v_indices, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 2458, __pyx_L1_error)
  }
  __pyx_pybuffernd_indices.diminfo[0].strides = __pyx_pybuffernd_indices.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_indices.diminfo[0].shape = __pyx_pybuffernd_indices.rcbuffer->pybuffer.shape[0];

  /* "pandas/_libs/intervaltree.pxi":2464
 *                  int64_t leaf_size):
 * 
 *         self.n_elements = len(left)             # <<<<<<<<<<<<<<
 *         self.leaf_size = leaf_size
 * 
 */
  __pyx_t_1 = PyObject_Length(((PyObject *)__pyx_v_left)); if (unlikely(__pyx_t_1 == ((Py_ssize_t)-1))) __PYX_ERR(1, 2464, __pyx_L1_error)
  __pyx_v_self->n_elements = __pyx_t_1;

  /* "pandas/_libs/intervaltree.pxi":2465
 * 
 *         self.n_elements = len(left)
 *         self.leaf_size = leaf_size             # <<<<<<<<<<<<<<
 * 
 *         # min_left and min_right are used to speed-up query by skipping
 */
  __pyx_v_self->leaf_size = __pyx_v_leaf_size;

  /* "pandas/_libs/intervaltree.pxi":2470
 *         # query on sub-nodes. If this node has size 0, query is cheap,
 *         # so these values don't matter.
 *         if left.size > 0:             # <<<<<<<<<<<<<<
 *             self.min_left = left.min()
 *             self.max_right = right.max()
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_left), __pyx_n_s_size); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2470, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyObject_RichCompare(__pyx_t_2, __pyx_int_0, Py_GT); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2470, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(1, 2470, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_4) {

    /* "pandas/_libs/intervaltree.pxi":2471
 *         # so these values don't matter.
 *         if left.size > 0:
 *             self.min_left = left.min()             # <<<<<<<<<<<<<<
 *             self.max_right = right.max()
 *         else:
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_left), __pyx_n_s_min); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2471, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2471, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_6 = __Pyx_PyInt_As_npy_int64(__pyx_t_3); if (unlikely((__pyx_t_6 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(1, 2471, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_self->min_left = __pyx_t_6;

    /* "pandas/_libs/intervaltree.pxi":2472
 *         if left.size > 0:
 *             self.min_left = left.min()
 *             self.max_right = right.max()             # <<<<<<<<<<<<<<
 *         else:
 *             self.min_left = 0
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_right), __pyx_n_s_max); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2472, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2472, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_6 = __Pyx_PyInt_As_npy_int64(__pyx_t_3); if (unlikely((__pyx_t_6 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(1, 2472, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_self->max_right = __pyx_t_6;

    /* "pandas/_libs/intervaltree.pxi":2470
 *         # query on sub-nodes. If this node has size 0, query is cheap,
 *         # so these values don't matter.
 *         if left.size > 0:             # <<<<<<<<<<<<<<
 *             self.min_left = left.min()
 *             self.max_right = right.max()
 */
    goto __pyx_L3;
  }

  /* "pandas/_libs/intervaltree.pxi":2474
 *             self.max_right = right.max()
 *         else:
 *             self.min_left = 0             # <<<<<<<<<<<<<<
 *             self.max_right = 0
 * 
 */
  /*else*/ {
    __pyx_v_self->min_left = 0;

    /* "pandas/_libs/intervaltree.pxi":2475
 *         else:
 *             self.min_left = 0
 *             self.max_right = 0             # <<<<<<<<<<<<<<
 * 
 *         if self.n_elements <= leaf_size:
 */
    __pyx_v_self->max_right = 0;
  }
  __pyx_L3:;

  /* "pandas/_libs/intervaltree.pxi":2477
 *             self.max_right = 0
 * 
 *         if self.n_elements <= leaf_size:             # <<<<<<<<<<<<<<
 *             # make this a terminal (leaf) node
 *             self.is_leaf_node = True
 */
  __pyx_t_4 = ((__pyx_v_self->n_elements <= __pyx_v_leaf_size) != 0);
  if (__pyx_t_4) {

    /* "pandas/_libs/intervaltree.pxi":2479
 *         if self.n_elements <= leaf_size:
 *             # make this a terminal (leaf) node
 *             self.is_leaf_node = True             # <<<<<<<<<<<<<<
 *             self.left = left
 *             self.right = right
 */
    __pyx_v_self->is_leaf_node = 1;

    /* "pandas/_libs/intervaltree.pxi":2480
 *             # make this a terminal (leaf) node
 *             self.is_leaf_node = True
 *             self.left = left             # <<<<<<<<<<<<<<
 *             self.right = right
 *             self.indices = indices
 */
    __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(((PyObject *)__pyx_v_left), PyBUF_WRITABLE); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(1, 2480, __pyx_L1_error)
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->left, 0);
    __pyx_v_self->left = __pyx_t_7;
    __pyx_t_7.memview = NULL;
    __pyx_t_7.data = NULL;

    /* "pandas/_libs/intervaltree.pxi":2481
 *             self.is_leaf_node = True
 *             self.left = left
 *             self.right = right             # <<<<<<<<<<<<<<
 *             self.indices = indices
 *             self.n_center = 0
 */
    __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(((PyObject *)__pyx_v_right), PyBUF_WRITABLE); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(1, 2481, __pyx_L1_error)
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->right, 0);
    __pyx_v_self->right = __pyx_t_7;
    __pyx_t_7.memview = NULL;
    __pyx_t_7.data = NULL;

    /* "pandas/_libs/intervaltree.pxi":2482
 *             self.left = left
 *             self.right = right
 *             self.indices = indices             # <<<<<<<<<<<<<<
 *             self.n_center = 0
 *         else:
 */
    __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(((PyObject *)__pyx_v_indices), PyBUF_WRITABLE); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(1, 2482, __pyx_L1_error)
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->indices, 0);
    __pyx_v_self->indices = __pyx_t_7;
    __pyx_t_7.memview = NULL;
    __pyx_t_7.data = NULL;

    /* "pandas/_libs/intervaltree.pxi":2483
 *             self.right = right
 *             self.indices = indices
 *             self.n_center = 0             # <<<<<<<<<<<<<<
 *         else:
 *             # calculate a pivot so we can create child nodes
 */
    __pyx_v_self->n_center = 0;

    /* "pandas/_libs/intervaltree.pxi":2477
 *             self.max_right = 0
 * 
 *         if self.n_elements <= leaf_size:             # <<<<<<<<<<<<<<
 *             # make this a terminal (leaf) node
 *             self.is_leaf_node = True
 */
    goto __pyx_L4;
  }

  /* "pandas/_libs/intervaltree.pxi":2486
 *         else:
 *             # calculate a pivot so we can create child nodes
 *             self.is_leaf_node = False             # <<<<<<<<<<<<<<
 *             self.pivot = np.median(left / 2 + right / 2)
 *             left_set, right_set, center_set = self.classify_intervals(
 */
  /*else*/ {
    __pyx_v_self->is_leaf_node = 0;

    /* "pandas/_libs/intervaltree.pxi":2487
 *             # calculate a pivot so we can create child nodes
 *             self.is_leaf_node = False
 *             self.pivot = np.median(left / 2 + right / 2)             # <<<<<<<<<<<<<<
 *             left_set, right_set, center_set = self.classify_intervals(
 *                 left, right)
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2487, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_median); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 2487, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyNumber_Divide(((PyObject *)__pyx_v_left), __pyx_int_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2487, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_8 = __Pyx_PyNumber_Divide(((PyObject *)__pyx_v_right), __pyx_int_2); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 2487, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_9 = PyNumber_Add(__pyx_t_2, __pyx_t_8); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 2487, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_3 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_8, __pyx_t_9) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_9);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2487, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_6 = __Pyx_PyInt_As_npy_int64(__pyx_t_3); if (unlikely((__pyx_t_6 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(1, 2487, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_self->pivot = __pyx_t_6;

    /* "pandas/_libs/intervaltree.pxi":2489
 *             self.pivot = np.median(left / 2 + right / 2)
 *             left_set, right_set, center_set = self.classify_intervals(
 *                 left, right)             # <<<<<<<<<<<<<<
 * 
 *             self.left_node = self.new_child_node(left, right,
 */
    __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(((PyObject *)__pyx_v_left), PyBUF_WRITABLE); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(1, 2489, __pyx_L1_error)
    __pyx_t_10 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(((PyObject *)__pyx_v_right), PyBUF_WRITABLE); if (unlikely(!__pyx_t_10.memview)) __PYX_ERR(1, 2489, __pyx_L1_error)

    /* "pandas/_libs/intervaltree.pxi":2488
 *             self.is_leaf_node = False
 *             self.pivot = np.median(left / 2 + right / 2)
 *             left_set, right_set, center_set = self.classify_intervals(             # <<<<<<<<<<<<<<
 *                 left, right)
 * 
 */
    __pyx_t_3 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode *)__pyx_v_self->__pyx_vtab)->classify_intervals(__pyx_v_self, __pyx_t_7, __pyx_t_10); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2488, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __PYX_XDEC_MEMVIEW(&__pyx_t_7, 1);
    __pyx_t_7.memview = NULL;
    __pyx_t_7.data = NULL;
    __PYX_XDEC_MEMVIEW(&__pyx_t_10, 1);
    __pyx_t_10.memview = NULL;
    __pyx_t_10.data = NULL;
    if ((likely(PyTuple_CheckExact(__pyx_t_3))) || (PyList_CheckExact(__pyx_t_3))) {
      PyObject* sequence = __pyx_t_3;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 3)) {
        if (size > 3) __Pyx_RaiseTooManyValuesError(3);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(1, 2488, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_5 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_9 = PyTuple_GET_ITEM(sequence, 1); 
        __pyx_t_8 = PyTuple_GET_ITEM(sequence, 2); 
      } else {
        __pyx_t_5 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_9 = PyList_GET_ITEM(sequence, 1); 
        __pyx_t_8 = PyList_GET_ITEM(sequence, 2); 
      }
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_9);
      __Pyx_INCREF(__pyx_t_8);
      #else
      __pyx_t_5 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 2488, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_9 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 2488, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_8 = PySequence_ITEM(sequence, 2); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 2488, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      #endif
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_2 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2488, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_11 = Py_TYPE(__pyx_t_2)->tp_iternext;
      index = 0; __pyx_t_5 = __pyx_t_11(__pyx_t_2); if (unlikely(!__pyx_t_5)) goto __pyx_L5_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_5);
      index = 1; __pyx_t_9 = __pyx_t_11(__pyx_t_2); if (unlikely(!__pyx_t_9)) goto __pyx_L5_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_9);
      index = 2; __pyx_t_8 = __pyx_t_11(__pyx_t_2); if (unlikely(!__pyx_t_8)) goto __pyx_L5_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_8);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_11(__pyx_t_2), 3) < 0) __PYX_ERR(1, 2488, __pyx_L1_error)
      __pyx_t_11 = NULL;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      goto __pyx_L6_unpacking_done;
      __pyx_L5_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_11 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(1, 2488, __pyx_L1_error)
      __pyx_L6_unpacking_done:;
    }
    __pyx_v_left_set = __pyx_t_5;
    __pyx_t_5 = 0;
    __pyx_v_right_set = __pyx_t_9;
    __pyx_t_9 = 0;
    __pyx_v_center_set = __pyx_t_8;
    __pyx_t_8 = 0;

    /* "pandas/_libs/intervaltree.pxi":2492
 * 
 *             self.left_node = self.new_child_node(left, right,
 *                                                  indices, left_set)             # <<<<<<<<<<<<<<
 *             self.right_node = self.new_child_node(left, right,
 *                                                   indices, right_set)
 */
    if (!(likely(((__pyx_v_left_set) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_left_set, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(1, 2492, __pyx_L1_error)

    /* "pandas/_libs/intervaltree.pxi":2491
 *                 left, right)
 * 
 *             self.left_node = self.new_child_node(left, right,             # <<<<<<<<<<<<<<
 *                                                  indices, left_set)
 *             self.right_node = self.new_child_node(left, right,
 */
    __pyx_t_3 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode *)__pyx_v_self->__pyx_vtab)->new_child_node(__pyx_v_self, ((PyArrayObject *)__pyx_v_left), ((PyArrayObject *)__pyx_v_right), ((PyArrayObject *)__pyx_v_indices), ((PyArrayObject *)__pyx_v_left_set)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2491, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode))))) __PYX_ERR(1, 2491, __pyx_L1_error)
    __Pyx_GIVEREF(__pyx_t_3);
    __Pyx_GOTREF(__pyx_v_self->left_node);
    __Pyx_DECREF(((PyObject *)__pyx_v_self->left_node));
    __pyx_v_self->left_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode *)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "pandas/_libs/intervaltree.pxi":2494
 *                                                  indices, left_set)
 *             self.right_node = self.new_child_node(left, right,
 *                                                   indices, right_set)             # <<<<<<<<<<<<<<
 * 
 *             self.center_left_values, self.center_left_indices = \
 */
    if (!(likely(((__pyx_v_right_set) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_right_set, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(1, 2494, __pyx_L1_error)

    /* "pandas/_libs/intervaltree.pxi":2493
 *             self.left_node = self.new_child_node(left, right,
 *                                                  indices, left_set)
 *             self.right_node = self.new_child_node(left, right,             # <<<<<<<<<<<<<<
 *                                                   indices, right_set)
 * 
 */
    __pyx_t_3 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode *)__pyx_v_self->__pyx_vtab)->new_child_node(__pyx_v_self, ((PyArrayObject *)__pyx_v_left), ((PyArrayObject *)__pyx_v_right), ((PyArrayObject *)__pyx_v_indices), ((PyArrayObject *)__pyx_v_right_set)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2493, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode))))) __PYX_ERR(1, 2493, __pyx_L1_error)
    __Pyx_GIVEREF(__pyx_t_3);
    __Pyx_GOTREF(__pyx_v_self->right_node);
    __Pyx_DECREF(((PyObject *)__pyx_v_self->right_node));
    __pyx_v_self->right_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode *)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "pandas/_libs/intervaltree.pxi":2497
 * 
 *             self.center_left_values, self.center_left_indices = \
 *                 sort_values_and_indices(left, indices, center_set)             # <<<<<<<<<<<<<<
 *             self.center_right_values, self.center_right_indices = \
 *                 sort_values_and_indices(right, indices, center_set)
 */
    __pyx_t_3 = __pyx_f_6pandas_5_libs_8interval_sort_values_and_indices(((PyObject *)__pyx_v_left), ((PyObject *)__pyx_v_indices), __pyx_v_center_set); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2497, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if ((likely(PyTuple_CheckExact(__pyx_t_3))) || (PyList_CheckExact(__pyx_t_3))) {
      PyObject* sequence = __pyx_t_3;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(1, 2496, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_8 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_9 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_8 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_9 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_8);
      __Pyx_INCREF(__pyx_t_9);
      #else
      __pyx_t_8 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 2496, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_9 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 2496, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      #endif
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_5 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 2496, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_11 = Py_TYPE(__pyx_t_5)->tp_iternext;
      index = 0; __pyx_t_8 = __pyx_t_11(__pyx_t_5); if (unlikely(!__pyx_t_8)) goto __pyx_L7_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_8);
      index = 1; __pyx_t_9 = __pyx_t_11(__pyx_t_5); if (unlikely(!__pyx_t_9)) goto __pyx_L7_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_9);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_11(__pyx_t_5), 2) < 0) __PYX_ERR(1, 2496, __pyx_L1_error)
      __pyx_t_11 = NULL;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      goto __pyx_L8_unpacking_done;
      __pyx_L7_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_11 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(1, 2496, __pyx_L1_error)
      __pyx_L8_unpacking_done:;
    }

    /* "pandas/_libs/intervaltree.pxi":2496
 *                                                   indices, right_set)
 * 
 *             self.center_left_values, self.center_left_indices = \             # <<<<<<<<<<<<<<
 *                 sort_values_and_indices(left, indices, center_set)
 *             self.center_right_values, self.center_right_indices = \
 */
    __pyx_t_10 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_t_8, PyBUF_WRITABLE); if (unlikely(!__pyx_t_10.memview)) __PYX_ERR(1, 2496, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_t_9, PyBUF_WRITABLE); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(1, 2496, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->center_left_values, 0);
    __pyx_v_self->center_left_values = __pyx_t_10;
    __pyx_t_10.memview = NULL;
    __pyx_t_10.data = NULL;
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->center_left_indices, 0);
    __pyx_v_self->center_left_indices = __pyx_t_7;
    __pyx_t_7.memview = NULL;
    __pyx_t_7.data = NULL;

    /* "pandas/_libs/intervaltree.pxi":2499
 *                 sort_values_and_indices(left, indices, center_set)
 *             self.center_right_values, self.center_right_indices = \
 *                 sort_values_and_indices(right, indices, center_set)             # <<<<<<<<<<<<<<
 *             self.n_center = len(self.center_left_indices)
 * 
 */
    __pyx_t_3 = __pyx_f_6pandas_5_libs_8interval_sort_values_and_indices(((PyObject *)__pyx_v_right), ((PyObject *)__pyx_v_indices), __pyx_v_center_set); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2499, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if ((likely(PyTuple_CheckExact(__pyx_t_3))) || (PyList_CheckExact(__pyx_t_3))) {
      PyObject* sequence = __pyx_t_3;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(1, 2498, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_9 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_8 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_9 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_8 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_9);
      __Pyx_INCREF(__pyx_t_8);
      #else
      __pyx_t_9 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 2498, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_8 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 2498, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      #endif
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_5 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 2498, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_11 = Py_TYPE(__pyx_t_5)->tp_iternext;
      index = 0; __pyx_t_9 = __pyx_t_11(__pyx_t_5); if (unlikely(!__pyx_t_9)) goto __pyx_L9_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_9);
      index = 1; __pyx_t_8 = __pyx_t_11(__pyx_t_5); if (unlikely(!__pyx_t_8)) goto __pyx_L9_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_8);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_11(__pyx_t_5), 2) < 0) __PYX_ERR(1, 2498, __pyx_L1_error)
      __pyx_t_11 = NULL;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      goto __pyx_L10_unpacking_done;
      __pyx_L9_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_11 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(1, 2498, __pyx_L1_error)
      __pyx_L10_unpacking_done:;
    }

    /* "pandas/_libs/intervaltree.pxi":2498
 *             self.center_left_values, self.center_left_indices = \
 *                 sort_values_and_indices(left, indices, center_set)
 *             self.center_right_values, self.center_right_indices = \             # <<<<<<<<<<<<<<
 *                 sort_values_and_indices(right, indices, center_set)
 *             self.n_center = len(self.center_left_indices)
 */
    __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_t_9, PyBUF_WRITABLE); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(1, 2498, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_10 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_t_8, PyBUF_WRITABLE); if (unlikely(!__pyx_t_10.memview)) __PYX_ERR(1, 2498, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->center_right_values, 0);
    __pyx_v_self->center_right_values = __pyx_t_7;
    __pyx_t_7.memview = NULL;
    __pyx_t_7.data = NULL;
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->center_right_indices, 0);
    __pyx_v_self->center_right_indices = __pyx_t_10;
    __pyx_t_10.memview = NULL;
    __pyx_t_10.data = NULL;

    /* "pandas/_libs/intervaltree.pxi":2500
 *             self.center_right_values, self.center_right_indices = \
 *                 sort_values_and_indices(right, indices, center_set)
 *             self.n_center = len(self.center_left_indices)             # <<<<<<<<<<<<<<
 * 
 *     @cython.wraparound(False)
 */
    if (unlikely(!__pyx_v_self->center_left_indices.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(1, 2500, __pyx_L1_error)}
    __pyx_t_12 = __Pyx_MemoryView_Len(__pyx_v_self->center_left_indices); 
    __pyx_v_self->n_center = __pyx_t_12;
  }
  __pyx_L4:;

  /* "pandas/_libs/intervaltree.pxi":2458
 *         readonly bint is_leaf_node
 * 
 *     def __init__(self,             # <<<<<<<<<<<<<<
 *                  ndarray[int64_t, ndim=1] left,
 *                  ndarray[int64_t, ndim=1] right,
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __PYX_XDEC_MEMVIEW(&__pyx_t_7, 1);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __PYX_XDEC_MEMVIEW(&__pyx_t_10, 1);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indices.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_left.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_right.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("pandas._libs.interval.Int64ClosedRightIntervalNode.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indices.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_left.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_right.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF(__pyx_v_left_set);
  __Pyx_XDECREF(__pyx_v_right_set);
  __Pyx_XDECREF(__pyx_v_center_set);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":2504
 *     @cython.wraparound(False)
 *     @cython.boundscheck(False)
 *     cdef classify_intervals(self, int64_t[:] left, int64_t[:] right):             # <<<<<<<<<<<<<<
 *         """Classify the given intervals based upon whether they fall to the
 *         left, right, or overlap with this node's pivot.
 */

static PyObject *__pyx_f_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_classify_intervals(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode *__pyx_v_self, __Pyx_memviewslice __pyx_v_left, __Pyx_memviewslice __pyx_v_right) {
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_left_ind = 0;
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_right_ind = 0;
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_overlapping_ind = 0;
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __pyx_t_5numpy_int64_t __pyx_t_2;
  __pyx_t_5numpy_int64_t __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  int __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  __Pyx_RefNannySetupContext("classify_intervals", 0);

  /* "pandas/_libs/intervaltree.pxi":2512
 *             Py_ssize_t i
 * 
 *         left_ind = Int64Vector()             # <<<<<<<<<<<<<<
 *         right_ind = Int64Vector()
 *         overlapping_ind = Int64Vector()
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2512, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_left_ind = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":2513
 * 
 *         left_ind = Int64Vector()
 *         right_ind = Int64Vector()             # <<<<<<<<<<<<<<
 *         overlapping_ind = Int64Vector()
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2513, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_right_ind = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":2514
 *         left_ind = Int64Vector()
 *         right_ind = Int64Vector()
 *         overlapping_ind = Int64Vector()             # <<<<<<<<<<<<<<
 * 
 *         for i in range(self.n_elements):
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2514, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_overlapping_ind = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":2516
 *         overlapping_ind = Int64Vector()
 * 
 *         for i in range(self.n_elements):             # <<<<<<<<<<<<<<
 *             if right[i] < self.pivot:
 *                 left_ind.append(i)
 */
  __pyx_t_2 = __pyx_v_self->n_elements;
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "pandas/_libs/intervaltree.pxi":2517
 * 
 *         for i in range(self.n_elements):
 *             if right[i] < self.pivot:             # <<<<<<<<<<<<<<
 *                 left_ind.append(i)
 *             elif self.pivot <= left[i]:
 */
    __pyx_t_5 = __pyx_v_i;
    __pyx_t_6 = (((*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_right.data + __pyx_t_5 * __pyx_v_right.strides[0]) ))) < __pyx_v_self->pivot) != 0);
    if (__pyx_t_6) {

      /* "pandas/_libs/intervaltree.pxi":2518
 *         for i in range(self.n_elements):
 *             if right[i] < self.pivot:
 *                 left_ind.append(i)             # <<<<<<<<<<<<<<
 *             elif self.pivot <= left[i]:
 *                 right_ind.append(i)
 */
      ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_left_ind->__pyx_vtab)->append(__pyx_v_left_ind, __pyx_v_i);

      /* "pandas/_libs/intervaltree.pxi":2517
 * 
 *         for i in range(self.n_elements):
 *             if right[i] < self.pivot:             # <<<<<<<<<<<<<<
 *                 left_ind.append(i)
 *             elif self.pivot <= left[i]:
 */
      goto __pyx_L5;
    }

    /* "pandas/_libs/intervaltree.pxi":2519
 *             if right[i] < self.pivot:
 *                 left_ind.append(i)
 *             elif self.pivot <= left[i]:             # <<<<<<<<<<<<<<
 *                 right_ind.append(i)
 *             else:
 */
    __pyx_t_7 = __pyx_v_i;
    __pyx_t_6 = ((__pyx_v_self->pivot <= (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_left.data + __pyx_t_7 * __pyx_v_left.strides[0]) )))) != 0);
    if (__pyx_t_6) {

      /* "pandas/_libs/intervaltree.pxi":2520
 *                 left_ind.append(i)
 *             elif self.pivot <= left[i]:
 *                 right_ind.append(i)             # <<<<<<<<<<<<<<
 *             else:
 *                 overlapping_ind.append(i)
 */
      ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_right_ind->__pyx_vtab)->append(__pyx_v_right_ind, __pyx_v_i);

      /* "pandas/_libs/intervaltree.pxi":2519
 *             if right[i] < self.pivot:
 *                 left_ind.append(i)
 *             elif self.pivot <= left[i]:             # <<<<<<<<<<<<<<
 *                 right_ind.append(i)
 *             else:
 */
      goto __pyx_L5;
    }

    /* "pandas/_libs/intervaltree.pxi":2522
 *                 right_ind.append(i)
 *             else:
 *                 overlapping_ind.append(i)             # <<<<<<<<<<<<<<
 * 
 *         return (left_ind.to_array(),
 */
    /*else*/ {
      ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_overlapping_ind->__pyx_vtab)->append(__pyx_v_overlapping_ind, __pyx_v_i);
    }
    __pyx_L5:;
  }

  /* "pandas/_libs/intervaltree.pxi":2524
 *                 overlapping_ind.append(i)
 * 
 *         return (left_ind.to_array(),             # <<<<<<<<<<<<<<
 *                 right_ind.to_array(),
 *                 overlapping_ind.to_array())
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_left_ind->__pyx_vtab)->to_array(__pyx_v_left_ind, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2524, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "pandas/_libs/intervaltree.pxi":2525
 * 
 *         return (left_ind.to_array(),
 *                 right_ind.to_array(),             # <<<<<<<<<<<<<<
 *                 overlapping_ind.to_array())
 * 
 */
  __pyx_t_8 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_right_ind->__pyx_vtab)->to_array(__pyx_v_right_ind, 0); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 2525, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);

  /* "pandas/_libs/intervaltree.pxi":2526
 *         return (left_ind.to_array(),
 *                 right_ind.to_array(),
 *                 overlapping_ind.to_array())             # <<<<<<<<<<<<<<
 * 
 *     cdef new_child_node(self,
 */
  __pyx_t_9 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_overlapping_ind->__pyx_vtab)->to_array(__pyx_v_overlapping_ind, 0); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 2526, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);

  /* "pandas/_libs/intervaltree.pxi":2524
 *                 overlapping_ind.append(i)
 * 
 *         return (left_ind.to_array(),             # <<<<<<<<<<<<<<
 *                 right_ind.to_array(),
 *                 overlapping_ind.to_array())
 */
  __pyx_t_10 = PyTuple_New(3); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 2524, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_t_8);
  __Pyx_GIVEREF(__pyx_t_9);
  PyTuple_SET_ITEM(__pyx_t_10, 2, __pyx_t_9);
  __pyx_t_1 = 0;
  __pyx_t_8 = 0;
  __pyx_t_9 = 0;
  __pyx_r = __pyx_t_10;
  __pyx_t_10 = 0;
  goto __pyx_L0;

  /* "pandas/_libs/intervaltree.pxi":2504
 *     @cython.wraparound(False)
 *     @cython.boundscheck(False)
 *     cdef classify_intervals(self, int64_t[:] left, int64_t[:] right):             # <<<<<<<<<<<<<<
 *         """Classify the given intervals based upon whether they fall to the
 *         left, right, or overlap with this node's pivot.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("pandas._libs.interval.Int64ClosedRightIntervalNode.classify_intervals", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_left_ind);
  __Pyx_XDECREF((PyObject *)__pyx_v_right_ind);
  __Pyx_XDECREF((PyObject *)__pyx_v_overlapping_ind);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":2528
 *                 overlapping_ind.to_array())
 * 
 *     cdef new_child_node(self,             # <<<<<<<<<<<<<<
 *                         ndarray[int64_t, ndim=1] left,
 *                         ndarray[int64_t, ndim=1] right,
 */

static PyObject *__pyx_f_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_new_child_node(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode *__pyx_v_self, PyArrayObject *__pyx_v_left, PyArrayObject *__pyx_v_right, PyArrayObject *__pyx_v_indices, PyArrayObject *__pyx_v_subset) {
  __Pyx_LocalBuf_ND __pyx_pybuffernd_indices;
  __Pyx_Buffer __pyx_pybuffer_indices;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_left;
  __Pyx_Buffer __pyx_pybuffer_left;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_right;
  __Pyx_Buffer __pyx_pybuffer_right;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_subset;
  __Pyx_Buffer __pyx_pybuffer_subset;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyArrayObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyArrayObject *__pyx_t_7 = NULL;
  PyArrayObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  __Pyx_RefNannySetupContext("new_child_node", 0);
  __Pyx_INCREF((PyObject *)__pyx_v_left);
  __Pyx_INCREF((PyObject *)__pyx_v_right);
  __Pyx_INCREF((PyObject *)__pyx_v_indices);
  __pyx_pybuffer_left.pybuffer.buf = NULL;
  __pyx_pybuffer_left.refcount = 0;
  __pyx_pybuffernd_left.data = NULL;
  __pyx_pybuffernd_left.rcbuffer = &__pyx_pybuffer_left;
  __pyx_pybuffer_right.pybuffer.buf = NULL;
  __pyx_pybuffer_right.refcount = 0;
  __pyx_pybuffernd_right.data = NULL;
  __pyx_pybuffernd_right.rcbuffer = &__pyx_pybuffer_right;
  __pyx_pybuffer_indices.pybuffer.buf = NULL;
  __pyx_pybuffer_indices.refcount = 0;
  __pyx_pybuffernd_indices.data = NULL;
  __pyx_pybuffernd_indices.rcbuffer = &__pyx_pybuffer_indices;
  __pyx_pybuffer_subset.pybuffer.buf = NULL;
  __pyx_pybuffer_subset.refcount = 0;
  __pyx_pybuffernd_subset.data = NULL;
  __pyx_pybuffernd_subset.rcbuffer = &__pyx_pybuffer_subset;
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_left.rcbuffer->pybuffer, (PyObject*)__pyx_v_left, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 2528, __pyx_L1_error)
  }
  __pyx_pybuffernd_left.diminfo[0].strides = __pyx_pybuffernd_left.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_left.diminfo[0].shape = __pyx_pybuffernd_left.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_right.rcbuffer->pybuffer, (PyObject*)__pyx_v_right, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 2528, __pyx_L1_error)
  }
  __pyx_pybuffernd_right.diminfo[0].strides = __pyx_pybuffernd_right.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_right.diminfo[0].shape = __pyx_pybuffernd_right.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_indices.rcbuffer->pybuffer, (PyObject*)__pyx_v_indices, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 2528, __pyx_L1_error)
  }
  __pyx_pybuffernd_indices.diminfo[0].strides = __pyx_pybuffernd_indices.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_indices.diminfo[0].shape = __pyx_pybuffernd_indices.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_subset.rcbuffer->pybuffer, (PyObject*)__pyx_v_subset, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 2528, __pyx_L1_error)
  }
  __pyx_pybuffernd_subset.diminfo[0].strides = __pyx_pybuffernd_subset.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_subset.diminfo[0].shape = __pyx_pybuffernd_subset.rcbuffer->pybuffer.shape[0];

  /* "pandas/_libs/intervaltree.pxi":2535
 *         """Create a new child node.
 *         """
 *         left = take(left, subset)             # <<<<<<<<<<<<<<
 *         right = take(right, subset)
 *         indices = take(indices, subset)
 */
  __pyx_t_1 = __pyx_f_6pandas_5_libs_8interval_take(((PyArrayObject *)__pyx_v_left), ((PyArrayObject *)__pyx_v_subset)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2535, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(1, 2535, __pyx_L1_error)
  __pyx_t_2 = ((PyArrayObject *)__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_left.rcbuffer->pybuffer);
    __pyx_t_3 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_left.rcbuffer->pybuffer, (PyObject*)__pyx_t_2, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack);
    if (unlikely(__pyx_t_3 < 0)) {
      PyErr_Fetch(&__pyx_t_4, &__pyx_t_5, &__pyx_t_6);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_left.rcbuffer->pybuffer, (PyObject*)__pyx_v_left, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_4); Py_XDECREF(__pyx_t_5); Py_XDECREF(__pyx_t_6);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_4, __pyx_t_5, __pyx_t_6);
      }
      __pyx_t_4 = __pyx_t_5 = __pyx_t_6 = 0;
    }
    __pyx_pybuffernd_left.diminfo[0].strides = __pyx_pybuffernd_left.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_left.diminfo[0].shape = __pyx_pybuffernd_left.rcbuffer->pybuffer.shape[0];
    if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(1, 2535, __pyx_L1_error)
  }
  __pyx_t_2 = 0;
  __Pyx_DECREF_SET(__pyx_v_left, ((PyArrayObject *)__pyx_t_1));
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":2536
 *         """
 *         left = take(left, subset)
 *         right = take(right, subset)             # <<<<<<<<<<<<<<
 *         indices = take(indices, subset)
 *         return Int64ClosedRightIntervalNode(
 */
  __pyx_t_1 = __pyx_f_6pandas_5_libs_8interval_take(((PyArrayObject *)__pyx_v_right), ((PyArrayObject *)__pyx_v_subset)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2536, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(1, 2536, __pyx_L1_error)
  __pyx_t_7 = ((PyArrayObject *)__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_right.rcbuffer->pybuffer);
    __pyx_t_3 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_right.rcbuffer->pybuffer, (PyObject*)__pyx_t_7, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack);
    if (unlikely(__pyx_t_3 < 0)) {
      PyErr_Fetch(&__pyx_t_6, &__pyx_t_5, &__pyx_t_4);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_right.rcbuffer->pybuffer, (PyObject*)__pyx_v_right, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_6); Py_XDECREF(__pyx_t_5); Py_XDECREF(__pyx_t_4);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_6, __pyx_t_5, __pyx_t_4);
      }
      __pyx_t_6 = __pyx_t_5 = __pyx_t_4 = 0;
    }
    __pyx_pybuffernd_right.diminfo[0].strides = __pyx_pybuffernd_right.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_right.diminfo[0].shape = __pyx_pybuffernd_right.rcbuffer->pybuffer.shape[0];
    if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(1, 2536, __pyx_L1_error)
  }
  __pyx_t_7 = 0;
  __Pyx_DECREF_SET(__pyx_v_right, ((PyArrayObject *)__pyx_t_1));
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":2537
 *         left = take(left, subset)
 *         right = take(right, subset)
 *         indices = take(indices, subset)             # <<<<<<<<<<<<<<
 *         return Int64ClosedRightIntervalNode(
 *             left, right, indices, self.leaf_size)
 */
  __pyx_t_1 = __pyx_f_6pandas_5_libs_8interval_take(((PyArrayObject *)__pyx_v_indices), ((PyArrayObject *)__pyx_v_subset)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2537, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(1, 2537, __pyx_L1_error)
  __pyx_t_8 = ((PyArrayObject *)__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indices.rcbuffer->pybuffer);
    __pyx_t_3 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_indices.rcbuffer->pybuffer, (PyObject*)__pyx_t_8, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack);
    if (unlikely(__pyx_t_3 < 0)) {
      PyErr_Fetch(&__pyx_t_4, &__pyx_t_5, &__pyx_t_6);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_indices.rcbuffer->pybuffer, (PyObject*)__pyx_v_indices, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_4); Py_XDECREF(__pyx_t_5); Py_XDECREF(__pyx_t_6);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_4, __pyx_t_5, __pyx_t_6);
      }
      __pyx_t_4 = __pyx_t_5 = __pyx_t_6 = 0;
    }
    __pyx_pybuffernd_indices.diminfo[0].strides = __pyx_pybuffernd_indices.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_indices.diminfo[0].shape = __pyx_pybuffernd_indices.rcbuffer->pybuffer.shape[0];
    if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(1, 2537, __pyx_L1_error)
  }
  __pyx_t_8 = 0;
  __Pyx_DECREF_SET(__pyx_v_indices, ((PyArrayObject *)__pyx_t_1));
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":2538
 *         right = take(right, subset)
 *         indices = take(indices, subset)
 *         return Int64ClosedRightIntervalNode(             # <<<<<<<<<<<<<<
 *             left, right, indices, self.leaf_size)
 * 
 */
  __Pyx_XDECREF(__pyx_r);

  /* "pandas/_libs/intervaltree.pxi":2539
 *         indices = take(indices, subset)
 *         return Int64ClosedRightIntervalNode(
 *             left, right, indices, self.leaf_size)             # <<<<<<<<<<<<<<
 * 
 *     @cython.wraparound(False)
 */
  __pyx_t_1 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->leaf_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2539, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "pandas/_libs/intervaltree.pxi":2538
 *         right = take(right, subset)
 *         indices = take(indices, subset)
 *         return Int64ClosedRightIntervalNode(             # <<<<<<<<<<<<<<
 *             left, right, indices, self.leaf_size)
 * 
 */
  __pyx_t_9 = PyTuple_New(4); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 2538, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_INCREF(((PyObject *)__pyx_v_left));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_left));
  PyTuple_SET_ITEM(__pyx_t_9, 0, ((PyObject *)__pyx_v_left));
  __Pyx_INCREF(((PyObject *)__pyx_v_right));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_right));
  PyTuple_SET_ITEM(__pyx_t_9, 1, ((PyObject *)__pyx_v_right));
  __Pyx_INCREF(((PyObject *)__pyx_v_indices));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_indices));
  PyTuple_SET_ITEM(__pyx_t_9, 2, ((PyObject *)__pyx_v_indices));
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_9, 3, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode), __pyx_t_9, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2538, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pandas/_libs/intervaltree.pxi":2528
 *                 overlapping_ind.to_array())
 * 
 *     cdef new_child_node(self,             # <<<<<<<<<<<<<<
 *                         ndarray[int64_t, ndim=1] left,
 *                         ndarray[int64_t, ndim=1] right,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_9);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indices.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_left.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_right.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_subset.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("pandas._libs.interval.Int64ClosedRightIntervalNode.new_child_node", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indices.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_left.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_right.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_subset.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF((PyObject *)__pyx_v_left);
  __Pyx_XDECREF((PyObject *)__pyx_v_right);
  __Pyx_XDECREF((PyObject *)__pyx_v_indices);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":2544
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_2query[] = "Recursively query this node and its sub-nodes for intervals that\n        overlap with the query point.\n        ";
static PyMethodDef __pyx_mdef_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_3query = {"query", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_3query, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_2query};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_args = 0;
  PyObject *__pyx_v_kwargs = 0;
  CYTHON_UNUSED PyObject *__pyx_v_defaults = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fused_cpdef (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_args,&__pyx_n_s_kwargs,&__pyx_n_s_defaults,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_args)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_kwargs)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fused_cpdef", 1, 3, 3, 1); __PYX_ERR(1, 2544, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_defaults)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fused_cpdef", 1, 3, 3, 2); __PYX_ERR(1, 2544, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fused_cpdef") < 0)) __PYX_ERR(1, 2544, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_args = values[0];
    __pyx_v_kwargs = values[1];
    __pyx_v_defaults = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fused_cpdef", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 2544, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Int64ClosedRightIntervalNode.__pyx_fused_cpdef", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_2query(((struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode *)__pyx_v_signatures), __pyx_v_args, __pyx_v_kwargs, __pyx_v_defaults);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_2query(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode *__pyx_v_signatures, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs, CYTHON_UNUSED PyObject *__pyx_v_defaults) {
  PyObject *__pyx_v_dest_sig = NULL;
  Py_ssize_t __pyx_v_i;
  CYTHON_UNUSED PyTypeObject *__pyx_v_ndarray = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_v_itemsize;
  PyObject *__pyx_v_arg = NULL;
  PyObject *__pyx_v_candidates = NULL;
  PyObject *__pyx_v_sig = NULL;
  int __pyx_v_match_found;
  PyObject *__pyx_v_src_sig = NULL;
  PyObject *__pyx_v_dst_type = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  Py_ssize_t __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  Py_ssize_t __pyx_t_13;
  Py_ssize_t __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  int __pyx_t_16;
  __Pyx_RefNannySetupContext("query", 0);
  __Pyx_INCREF(__pyx_v_kwargs);
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2544, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  PyList_SET_ITEM(__pyx_t_1, 0, Py_None);
  __pyx_v_dest_sig = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_3 = (__pyx_v_kwargs != Py_None);
  __pyx_t_4 = (__pyx_t_3 != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_kwargs); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(1, 2544, __pyx_L1_error)
  __pyx_t_3 = ((!__pyx_t_4) != 0);
  __pyx_t_2 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {
    __Pyx_INCREF(Py_None);
    __Pyx_DECREF_SET(__pyx_v_kwargs, Py_None);
  }
  __pyx_t_1 = ((PyObject *)__Pyx_ImportNumPyArrayTypeIfAvailable()); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2544, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_ndarray = ((PyTypeObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_v_itemsize = -1L;
  if (unlikely(__pyx_v_args == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 2544, __pyx_L1_error)
  }
  __pyx_t_5 = PyTuple_GET_SIZE(((PyObject*)__pyx_v_args)); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(1, 2544, __pyx_L1_error)
  __pyx_t_2 = ((2 < __pyx_t_5) != 0);
  if (__pyx_t_2) {
    if (unlikely(__pyx_v_args == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 2544, __pyx_L1_error)
    }
    __pyx_t_1 = PyTuple_GET_ITEM(((PyObject*)__pyx_v_args), 2);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v_arg = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L6;
  }
  __pyx_t_3 = (__pyx_v_kwargs != Py_None);
  __pyx_t_4 = (__pyx_t_3 != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L7_bool_binop_done;
  }
  if (unlikely(__pyx_v_kwargs == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(1, 2544, __pyx_L1_error)
  }
  __pyx_t_4 = (__Pyx_PyDict_ContainsTF(__pyx_n_s_point, ((PyObject*)__pyx_v_kwargs), Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(1, 2544, __pyx_L1_error)
  __pyx_t_3 = (__pyx_t_4 != 0);
  __pyx_t_2 = __pyx_t_3;
  __pyx_L7_bool_binop_done:;
  if (__pyx_t_2) {
    if (unlikely(__pyx_v_kwargs == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 2544, __pyx_L1_error)
    }
    __pyx_t_1 = __Pyx_PyDict_GetItem(((PyObject*)__pyx_v_kwargs), __pyx_n_s_point); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2544, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_arg = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L6;
  }
  /*else*/ {
    if (unlikely(__pyx_v_args == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
      __PYX_ERR(1, 2544, __pyx_L1_error)
    }
    __pyx_t_5 = PyTuple_GET_SIZE(((PyObject*)__pyx_v_args)); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(1, 2544, __pyx_L1_error)
    __pyx_t_1 = PyInt_FromSsize_t(__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2544, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_6 = PyTuple_New(3); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 2544, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_INCREF(__pyx_int_3);
    __Pyx_GIVEREF(__pyx_int_3);
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_int_3);
    __Pyx_INCREF(__pyx_n_s_s);
    __Pyx_GIVEREF(__pyx_n_s_s);
    PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_n_s_s);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Expected_at_least_d_argument_s_g, __pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2544, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_1); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 2544, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(1, 2544, __pyx_L1_error)
  }
  __pyx_L6:;
  while (1) {
    __pyx_t_2 = PyFloat_Check(__pyx_v_arg); 
    __pyx_t_3 = (__pyx_t_2 != 0);
    if (__pyx_t_3) {
      if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, __pyx_n_s_float64_t, long, 1, __Pyx_PyInt_From_long, 1, 0, 0) < 0)) __PYX_ERR(1, 2544, __pyx_L1_error)
      goto __pyx_L10_break;
    }
    __pyx_t_2 = PyInt_Check(__pyx_v_arg); 
    __pyx_t_4 = (__pyx_t_2 != 0);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_3 = __pyx_t_4;
      goto __pyx_L13_bool_binop_done;
    }
    __pyx_t_4 = PyLong_Check(__pyx_v_arg); 
    __pyx_t_2 = (__pyx_t_4 != 0);
    __pyx_t_3 = __pyx_t_2;
    __pyx_L13_bool_binop_done:;
    __pyx_t_2 = (__pyx_t_3 != 0);
    if (__pyx_t_2) {
      if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, __pyx_n_s_int64_t, long, 1, __Pyx_PyInt_From_long, 1, 0, 0) < 0)) __PYX_ERR(1, 2544, __pyx_L1_error)
      goto __pyx_L10_break;
    }
    if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, Py_None, long, 1, __Pyx_PyInt_From_long, 1, 0, 0) < 0)) __PYX_ERR(1, 2544, __pyx_L1_error)
    goto __pyx_L10_break;
  }
  __pyx_L10_break:;
  __pyx_t_6 = PyList_New(0); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 2544, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_v_candidates = ((PyObject*)__pyx_t_6);
  __pyx_t_6 = 0;
  __pyx_t_5 = 0;
  if (unlikely(((PyObject *)__pyx_v_signatures) == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(1, 2544, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_dict_iterator(((PyObject*)__pyx_v_signatures), 1, ((PyObject *)NULL), (&__pyx_t_7), (&__pyx_t_8)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2544, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __pyx_t_6 = __pyx_t_1;
  __pyx_t_1 = 0;
  while (1) {
    __pyx_t_9 = __Pyx_dict_iter_next(__pyx_t_6, __pyx_t_7, &__pyx_t_5, &__pyx_t_1, NULL, NULL, __pyx_t_8);
    if (unlikely(__pyx_t_9 == 0)) break;
    if (unlikely(__pyx_t_9 == -1)) __PYX_ERR(1, 2544, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_sig, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_v_match_found = 0;
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_v_sig, __pyx_n_s_strip); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 2544, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __pyx_t_12 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_11))) {
      __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_11);
      if (likely(__pyx_t_12)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
        __Pyx_INCREF(__pyx_t_12);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_11, function);
      }
    }
    __pyx_t_10 = (__pyx_t_12) ? __Pyx_PyObject_Call2Args(__pyx_t_11, __pyx_t_12, __pyx_kp_s__13) : __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_kp_s__13);
    __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
    if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 2544, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_split); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 2544, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_t_10 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_11))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_11);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_11, function);
      }
    }
    __pyx_t_1 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_11, __pyx_t_10, __pyx_kp_s__14) : __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_kp_s__14);
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2544, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_XDECREF_SET(__pyx_v_src_sig, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_13 = PyList_GET_SIZE(__pyx_v_dest_sig); if (unlikely(__pyx_t_13 == ((Py_ssize_t)-1))) __PYX_ERR(1, 2544, __pyx_L1_error)
    __pyx_t_14 = __pyx_t_13;
    for (__pyx_t_15 = 0; __pyx_t_15 < __pyx_t_14; __pyx_t_15+=1) {
      __pyx_v_i = __pyx_t_15;
      __pyx_t_1 = PyList_GET_ITEM(__pyx_v_dest_sig, __pyx_v_i);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_XDECREF_SET(__pyx_v_dst_type, __pyx_t_1);
      __pyx_t_1 = 0;
      __pyx_t_2 = (__pyx_v_dst_type != Py_None);
      __pyx_t_3 = (__pyx_t_2 != 0);
      if (__pyx_t_3) {
        __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_src_sig, __pyx_v_i, Py_ssize_t, 1, PyInt_FromSsize_t, 0, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2544, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_11 = PyObject_RichCompare(__pyx_t_1, __pyx_v_dst_type, Py_EQ); __Pyx_XGOTREF(__pyx_t_11); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 2544, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_11); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(1, 2544, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        if (__pyx_t_3) {
          __pyx_v_match_found = 1;
          goto __pyx_L20;
        }
        /*else*/ {
          __pyx_v_match_found = 0;
          goto __pyx_L18_break;
        }
        __pyx_L20:;
      }
    }
    __pyx_L18_break:;
    __pyx_t_3 = (__pyx_v_match_found != 0);
    if (__pyx_t_3) {
      __pyx_t_16 = __Pyx_PyList_Append(__pyx_v_candidates, __pyx_v_sig); if (unlikely(__pyx_t_16 == ((int)-1))) __PYX_ERR(1, 2544, __pyx_L1_error)
    }
  }
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_3 = (PyList_GET_SIZE(__pyx_v_candidates) != 0);
  __pyx_t_2 = ((!__pyx_t_3) != 0);
  if (__pyx_t_2) {
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__15, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 2544, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(1, 2544, __pyx_L1_error)
  }
  __pyx_t_7 = PyList_GET_SIZE(__pyx_v_candidates); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1))) __PYX_ERR(1, 2544, __pyx_L1_error)
  __pyx_t_2 = ((__pyx_t_7 > 1) != 0);
  if (__pyx_t_2) {
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__16, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 2544, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(1, 2544, __pyx_L1_error)
  }
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    if (unlikely(((PyObject *)__pyx_v_signatures) == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 2544, __pyx_L1_error)
    }
    __pyx_t_6 = __Pyx_PyDict_GetItem(((PyObject*)__pyx_v_signatures), PyList_GET_ITEM(__pyx_v_candidates, 0)); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 2544, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_r = __pyx_t_6;
    __pyx_t_6 = 0;
    goto __pyx_L0;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("pandas._libs.interval.Int64ClosedRightIntervalNode.__pyx_fused_cpdef", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_dest_sig);
  __Pyx_XDECREF(__pyx_v_ndarray);
  __Pyx_XDECREF(__pyx_v_arg);
  __Pyx_XDECREF(__pyx_v_candidates);
  __Pyx_XDECREF(__pyx_v_sig);
  __Pyx_XDECREF(__pyx_v_src_sig);
  __Pyx_XDECREF(__pyx_v_dst_type);
  __Pyx_XDECREF(__pyx_v_kwargs);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_9__pyx_fuse_0query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_fuse_0__pyx_f_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float64_t __pyx_v_point, int __pyx_skip_dispatch) {
  __Pyx_memviewslice __pyx_v_indices = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_values = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __pyx_t_5numpy_int64_t __pyx_t_9;
  __pyx_t_5numpy_int64_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  int __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  __Pyx_memviewslice __pyx_t_16 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_17;
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  __Pyx_RefNannySetupContext("__pyx_fuse_0query", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pyx_fuse_0query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2544, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_9__pyx_fuse_0query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = PyFloat_FromDouble(__pyx_v_point); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2544, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2544, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2544, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 2544, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_5) {
            __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2544, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2544, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_3query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_4 = PyFloat_FromDouble(__pyx_v_point); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 2544, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_7 = __pyx_t_1; __pyx_t_3 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2544, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2544, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 2544, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (__pyx_t_3) {
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_t_4);
          __pyx_t_4 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2544, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pandas/_libs/intervaltree.pxi":2553
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
  __pyx_t_8 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_8) {

    /* "pandas/_libs/intervaltree.pxi":2557
 *             # continue the binary tree structure. Instead, we use linear
 *             # search.
 *             for i in range(self.n_elements):             # <<<<<<<<<<<<<<
 *                 if self.left[i] < point <= self.right[i]:
 *                     result.append(self.indices[i])
 */
    __pyx_t_9 = __pyx_v_self->n_elements;
    __pyx_t_10 = __pyx_t_9;
    for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
      __pyx_v_i = __pyx_t_11;

      /* "pandas/_libs/intervaltree.pxi":2558
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point <= self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      __pyx_t_12 = __pyx_v_i;
      __pyx_t_8 = ((*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->left.data + __pyx_t_12 * __pyx_v_self->left.strides[0]) ))) < __pyx_v_point);
      if (__pyx_t_8) {
        __pyx_t_13 = __pyx_v_i;
        __pyx_t_8 = (__pyx_v_point <= (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->right.data + __pyx_t_13 * __pyx_v_self->right.strides[0]) ))));
      }
      __pyx_t_14 = (__pyx_t_8 != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":2559
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point <= self.right[i]:
 *                     result.append(self.indices[i])             # <<<<<<<<<<<<<<
 *         else:
 *             # There are child nodes. Based on comparing our query to the pivot,
 */
        __pyx_t_15 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->indices.data + __pyx_t_15 * __pyx_v_self->indices.strides[0]) ))));

        /* "pandas/_libs/intervaltree.pxi":2558
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point <= self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      }
    }

    /* "pandas/_libs/intervaltree.pxi":2553
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
    goto __pyx_L3;
  }

  /* "pandas/_libs/intervaltree.pxi":2563
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
  /*else*/ {
    __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":2564
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:
 *                 values = self.center_left_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 */
      __pyx_t_16 = __pyx_v_self->center_left_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":2565
 *             if point < self.pivot:
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:
 */
      __pyx_t_16 = __pyx_v_self->center_left_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_indices = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":2566
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):             # <<<<<<<<<<<<<<
 *                     if not values[i] < point:
 *                         break
 */
      __pyx_t_9 = __pyx_v_self->n_center;
      __pyx_t_10 = __pyx_t_9;
      for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":2567
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_17 = __pyx_v_i;
        __pyx_t_14 = ((!(((*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_17 * __pyx_v_values.strides[0]) ))) < __pyx_v_point) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":2568
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:
 */
          goto __pyx_L9_break;

          /* "pandas/_libs/intervaltree.pxi":2567
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":2569
 *                     if not values[i] < point:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 */
        __pyx_t_18 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_18 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L9_break:;

      /* "pandas/_libs/intervaltree.pxi":2570
 *                         break
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      __pyx_t_14 = ((__pyx_v_point <= __pyx_v_self->left_node->max_right) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":2571
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)             # <<<<<<<<<<<<<<
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode *)__pyx_v_self->left_node->__pyx_vtab)->__pyx_fuse_0query(__pyx_v_self->left_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2571, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":2570
 *                         break
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":2563
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":2572
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
    __pyx_t_14 = ((__pyx_v_point > __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":2573
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 *                 values = self.center_right_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 */
      __pyx_t_16 = __pyx_v_self->center_right_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":2574
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:
 */
      __pyx_t_16 = __pyx_v_self->center_right_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_indices = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":2575
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):             # <<<<<<<<<<<<<<
 *                     if not point <= values[i]:
 *                         break
 */
      for (__pyx_t_11 = (__pyx_v_self->n_center - 1); __pyx_t_11 > -1L; __pyx_t_11-=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":2576
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_19 = __pyx_v_i;
        __pyx_t_14 = ((!((__pyx_v_point <= (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_19 * __pyx_v_values.strides[0]) )))) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":2577
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:
 */
          goto __pyx_L13_break;

          /* "pandas/_libs/intervaltree.pxi":2576
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":2578
 *                     if not point <= values[i]:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if self.right_node.min_left < point:
 *                     self.right_node.query(result, point)
 */
        __pyx_t_20 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_20 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L13_break:;

      /* "pandas/_libs/intervaltree.pxi":2579
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      __pyx_t_14 = ((__pyx_v_self->right_node->min_left < __pyx_v_point) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":2580
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:
 *                     self.right_node.query(result, point)             # <<<<<<<<<<<<<<
 *             else:
 *                 result.extend(self.center_left_indices)
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode *)__pyx_v_self->right_node->__pyx_vtab)->__pyx_fuse_0query(__pyx_v_self->right_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2580, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":2579
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":2572
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":2582
 *                     self.right_node.query(result, point)
 *             else:
 *                 result.extend(self.center_left_indices)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
    /*else*/ {
      __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->extend(__pyx_v_result, __pyx_v_self->center_left_indices); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2582, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_L7:;
  }
  __pyx_L3:;

  /* "pandas/_libs/intervaltree.pxi":2544
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __PYX_XDEC_MEMVIEW(&__pyx_t_16, 1);
  __Pyx_AddTraceback("pandas._libs.interval.Int64ClosedRightIntervalNode.query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_indices, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_values, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_9__pyx_fuse_0query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_0__pyx_mdef_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_9__pyx_fuse_0query = {"__pyx_fuse_0query", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_9__pyx_fuse_0query, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_2query};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_9__pyx_fuse_0query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result = 0;
  __pyx_t_5numpy_float64_t __pyx_v_point;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fuse_0query (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_result,&__pyx_n_s_point,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_result)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_point)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_0query", 1, 2, 2, 1); __PYX_ERR(1, 2544, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_0query") < 0)) __PYX_ERR(1, 2544, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_result = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)values[0]);
    __pyx_v_point = __pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_point == ((npy_float64)-1)) && PyErr_Occurred())) __PYX_ERR(1, 2544, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_0query", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 2544, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Int64ClosedRightIntervalNode.__pyx_fuse_0query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_result), __pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector, 1, "result", 0))) __PYX_ERR(1, 2544, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_8__pyx_fuse_0query(((struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode *)__pyx_v_self), __pyx_v_result, __pyx_v_point);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_8__pyx_fuse_0query(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float64_t __pyx_v_point) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__pyx_fuse_0query", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_fuse_0__pyx_f_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_query(__pyx_v_self, __pyx_v_result, __pyx_v_point, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2544, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Int64ClosedRightIntervalNode.__pyx_fuse_0query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_11__pyx_fuse_1query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_fuse_1__pyx_f_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float32_t __pyx_v_point, int __pyx_skip_dispatch) {
  __Pyx_memviewslice __pyx_v_indices = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_values = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __pyx_t_5numpy_int64_t __pyx_t_9;
  __pyx_t_5numpy_int64_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  int __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  __Pyx_memviewslice __pyx_t_16 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_17;
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  __Pyx_RefNannySetupContext("__pyx_fuse_1query", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pyx_fuse_1query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2544, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_11__pyx_fuse_1query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = PyFloat_FromDouble(__pyx_v_point); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2544, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2544, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2544, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 2544, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_5) {
            __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2544, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2544, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_3query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_4 = PyFloat_FromDouble(__pyx_v_point); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 2544, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_7 = __pyx_t_1; __pyx_t_3 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2544, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2544, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 2544, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (__pyx_t_3) {
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_t_4);
          __pyx_t_4 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2544, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pandas/_libs/intervaltree.pxi":2553
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
  __pyx_t_8 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_8) {

    /* "pandas/_libs/intervaltree.pxi":2557
 *             # continue the binary tree structure. Instead, we use linear
 *             # search.
 *             for i in range(self.n_elements):             # <<<<<<<<<<<<<<
 *                 if self.left[i] < point <= self.right[i]:
 *                     result.append(self.indices[i])
 */
    __pyx_t_9 = __pyx_v_self->n_elements;
    __pyx_t_10 = __pyx_t_9;
    for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
      __pyx_v_i = __pyx_t_11;

      /* "pandas/_libs/intervaltree.pxi":2558
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point <= self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      __pyx_t_12 = __pyx_v_i;
      __pyx_t_8 = ((*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->left.data + __pyx_t_12 * __pyx_v_self->left.strides[0]) ))) < __pyx_v_point);
      if (__pyx_t_8) {
        __pyx_t_13 = __pyx_v_i;
        __pyx_t_8 = (__pyx_v_point <= (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->right.data + __pyx_t_13 * __pyx_v_self->right.strides[0]) ))));
      }
      __pyx_t_14 = (__pyx_t_8 != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":2559
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point <= self.right[i]:
 *                     result.append(self.indices[i])             # <<<<<<<<<<<<<<
 *         else:
 *             # There are child nodes. Based on comparing our query to the pivot,
 */
        __pyx_t_15 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->indices.data + __pyx_t_15 * __pyx_v_self->indices.strides[0]) ))));

        /* "pandas/_libs/intervaltree.pxi":2558
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point <= self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      }
    }

    /* "pandas/_libs/intervaltree.pxi":2553
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
    goto __pyx_L3;
  }

  /* "pandas/_libs/intervaltree.pxi":2563
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
  /*else*/ {
    __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":2564
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:
 *                 values = self.center_left_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 */
      __pyx_t_16 = __pyx_v_self->center_left_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":2565
 *             if point < self.pivot:
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:
 */
      __pyx_t_16 = __pyx_v_self->center_left_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_indices = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":2566
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):             # <<<<<<<<<<<<<<
 *                     if not values[i] < point:
 *                         break
 */
      __pyx_t_9 = __pyx_v_self->n_center;
      __pyx_t_10 = __pyx_t_9;
      for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":2567
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_17 = __pyx_v_i;
        __pyx_t_14 = ((!(((*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_17 * __pyx_v_values.strides[0]) ))) < __pyx_v_point) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":2568
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:
 */
          goto __pyx_L9_break;

          /* "pandas/_libs/intervaltree.pxi":2567
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":2569
 *                     if not values[i] < point:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 */
        __pyx_t_18 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_18 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L9_break:;

      /* "pandas/_libs/intervaltree.pxi":2570
 *                         break
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      __pyx_t_14 = ((__pyx_v_point <= __pyx_v_self->left_node->max_right) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":2571
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)             # <<<<<<<<<<<<<<
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode *)__pyx_v_self->left_node->__pyx_vtab)->__pyx_fuse_1query(__pyx_v_self->left_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2571, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":2570
 *                         break
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":2563
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":2572
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
    __pyx_t_14 = ((__pyx_v_point > __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":2573
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 *                 values = self.center_right_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 */
      __pyx_t_16 = __pyx_v_self->center_right_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":2574
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:
 */
      __pyx_t_16 = __pyx_v_self->center_right_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_indices = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":2575
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):             # <<<<<<<<<<<<<<
 *                     if not point <= values[i]:
 *                         break
 */
      for (__pyx_t_11 = (__pyx_v_self->n_center - 1); __pyx_t_11 > -1L; __pyx_t_11-=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":2576
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_19 = __pyx_v_i;
        __pyx_t_14 = ((!((__pyx_v_point <= (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_19 * __pyx_v_values.strides[0]) )))) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":2577
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:
 */
          goto __pyx_L13_break;

          /* "pandas/_libs/intervaltree.pxi":2576
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":2578
 *                     if not point <= values[i]:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if self.right_node.min_left < point:
 *                     self.right_node.query(result, point)
 */
        __pyx_t_20 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_20 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L13_break:;

      /* "pandas/_libs/intervaltree.pxi":2579
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      __pyx_t_14 = ((__pyx_v_self->right_node->min_left < __pyx_v_point) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":2580
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:
 *                     self.right_node.query(result, point)             # <<<<<<<<<<<<<<
 *             else:
 *                 result.extend(self.center_left_indices)
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode *)__pyx_v_self->right_node->__pyx_vtab)->__pyx_fuse_1query(__pyx_v_self->right_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2580, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":2579
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":2572
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":2582
 *                     self.right_node.query(result, point)
 *             else:
 *                 result.extend(self.center_left_indices)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
    /*else*/ {
      __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->extend(__pyx_v_result, __pyx_v_self->center_left_indices); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2582, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_L7:;
  }
  __pyx_L3:;

  /* "pandas/_libs/intervaltree.pxi":2544
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __PYX_XDEC_MEMVIEW(&__pyx_t_16, 1);
  __Pyx_AddTraceback("pandas._libs.interval.Int64ClosedRightIntervalNode.query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_indices, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_values, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_11__pyx_fuse_1query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_1__pyx_mdef_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_11__pyx_fuse_1query = {"__pyx_fuse_1query", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_11__pyx_fuse_1query, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_2query};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_11__pyx_fuse_1query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result = 0;
  __pyx_t_5numpy_float32_t __pyx_v_point;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fuse_1query (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_result,&__pyx_n_s_point,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_result)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_point)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_1query", 1, 2, 2, 1); __PYX_ERR(1, 2544, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_1query") < 0)) __PYX_ERR(1, 2544, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_result = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)values[0]);
    __pyx_v_point = __pyx_PyFloat_AsFloat(values[1]); if (unlikely((__pyx_v_point == ((npy_float32)-1)) && PyErr_Occurred())) __PYX_ERR(1, 2544, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_1query", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 2544, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Int64ClosedRightIntervalNode.__pyx_fuse_1query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_result), __pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector, 1, "result", 0))) __PYX_ERR(1, 2544, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_10__pyx_fuse_1query(((struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode *)__pyx_v_self), __pyx_v_result, __pyx_v_point);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_10__pyx_fuse_1query(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float32_t __pyx_v_point) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__pyx_fuse_1query", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_fuse_1__pyx_f_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_query(__pyx_v_self, __pyx_v_result, __pyx_v_point, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2544, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Int64ClosedRightIntervalNode.__pyx_fuse_1query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_13__pyx_fuse_2query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_fuse_2__pyx_f_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int64_t __pyx_v_point, int __pyx_skip_dispatch) {
  __Pyx_memviewslice __pyx_v_indices = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_values = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __pyx_t_5numpy_int64_t __pyx_t_9;
  __pyx_t_5numpy_int64_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  int __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  __Pyx_memviewslice __pyx_t_16 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_17;
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  __Pyx_RefNannySetupContext("__pyx_fuse_2query", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pyx_fuse_2query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2544, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_13__pyx_fuse_2query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = __Pyx_PyInt_From_npy_int64(__pyx_v_point); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2544, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2544, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2544, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 2544, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_5) {
            __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2544, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2544, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_3query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_4 = __Pyx_PyInt_From_npy_int64(__pyx_v_point); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 2544, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_7 = __pyx_t_1; __pyx_t_3 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2544, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2544, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 2544, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (__pyx_t_3) {
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_t_4);
          __pyx_t_4 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2544, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pandas/_libs/intervaltree.pxi":2553
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
  __pyx_t_8 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_8) {

    /* "pandas/_libs/intervaltree.pxi":2557
 *             # continue the binary tree structure. Instead, we use linear
 *             # search.
 *             for i in range(self.n_elements):             # <<<<<<<<<<<<<<
 *                 if self.left[i] < point <= self.right[i]:
 *                     result.append(self.indices[i])
 */
    __pyx_t_9 = __pyx_v_self->n_elements;
    __pyx_t_10 = __pyx_t_9;
    for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
      __pyx_v_i = __pyx_t_11;

      /* "pandas/_libs/intervaltree.pxi":2558
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point <= self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      __pyx_t_12 = __pyx_v_i;
      __pyx_t_8 = ((*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->left.data + __pyx_t_12 * __pyx_v_self->left.strides[0]) ))) < __pyx_v_point);
      if (__pyx_t_8) {
        __pyx_t_13 = __pyx_v_i;
        __pyx_t_8 = (__pyx_v_point <= (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->right.data + __pyx_t_13 * __pyx_v_self->right.strides[0]) ))));
      }
      __pyx_t_14 = (__pyx_t_8 != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":2559
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point <= self.right[i]:
 *                     result.append(self.indices[i])             # <<<<<<<<<<<<<<
 *         else:
 *             # There are child nodes. Based on comparing our query to the pivot,
 */
        __pyx_t_15 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->indices.data + __pyx_t_15 * __pyx_v_self->indices.strides[0]) ))));

        /* "pandas/_libs/intervaltree.pxi":2558
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point <= self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      }
    }

    /* "pandas/_libs/intervaltree.pxi":2553
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
    goto __pyx_L3;
  }

  /* "pandas/_libs/intervaltree.pxi":2563
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
  /*else*/ {
    __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":2564
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:
 *                 values = self.center_left_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 */
      __pyx_t_16 = __pyx_v_self->center_left_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":2565
 *             if point < self.pivot:
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:
 */
      __pyx_t_16 = __pyx_v_self->center_left_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_indices = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":2566
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):             # <<<<<<<<<<<<<<
 *                     if not values[i] < point:
 *                         break
 */
      __pyx_t_9 = __pyx_v_self->n_center;
      __pyx_t_10 = __pyx_t_9;
      for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":2567
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_17 = __pyx_v_i;
        __pyx_t_14 = ((!(((*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_17 * __pyx_v_values.strides[0]) ))) < __pyx_v_point) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":2568
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:
 */
          goto __pyx_L9_break;

          /* "pandas/_libs/intervaltree.pxi":2567
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":2569
 *                     if not values[i] < point:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 */
        __pyx_t_18 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_18 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L9_break:;

      /* "pandas/_libs/intervaltree.pxi":2570
 *                         break
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      __pyx_t_14 = ((__pyx_v_point <= __pyx_v_self->left_node->max_right) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":2571
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)             # <<<<<<<<<<<<<<
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode *)__pyx_v_self->left_node->__pyx_vtab)->__pyx_fuse_2query(__pyx_v_self->left_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2571, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":2570
 *                         break
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":2563
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":2572
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
    __pyx_t_14 = ((__pyx_v_point > __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":2573
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 *                 values = self.center_right_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 */
      __pyx_t_16 = __pyx_v_self->center_right_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":2574
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:
 */
      __pyx_t_16 = __pyx_v_self->center_right_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_indices = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":2575
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):             # <<<<<<<<<<<<<<
 *                     if not point <= values[i]:
 *                         break
 */
      for (__pyx_t_11 = (__pyx_v_self->n_center - 1); __pyx_t_11 > -1L; __pyx_t_11-=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":2576
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_19 = __pyx_v_i;
        __pyx_t_14 = ((!((__pyx_v_point <= (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_19 * __pyx_v_values.strides[0]) )))) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":2577
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:
 */
          goto __pyx_L13_break;

          /* "pandas/_libs/intervaltree.pxi":2576
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":2578
 *                     if not point <= values[i]:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if self.right_node.min_left < point:
 *                     self.right_node.query(result, point)
 */
        __pyx_t_20 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_20 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L13_break:;

      /* "pandas/_libs/intervaltree.pxi":2579
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      __pyx_t_14 = ((__pyx_v_self->right_node->min_left < __pyx_v_point) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":2580
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:
 *                     self.right_node.query(result, point)             # <<<<<<<<<<<<<<
 *             else:
 *                 result.extend(self.center_left_indices)
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode *)__pyx_v_self->right_node->__pyx_vtab)->__pyx_fuse_2query(__pyx_v_self->right_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2580, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":2579
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":2572
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":2582
 *                     self.right_node.query(result, point)
 *             else:
 *                 result.extend(self.center_left_indices)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
    /*else*/ {
      __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->extend(__pyx_v_result, __pyx_v_self->center_left_indices); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2582, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_L7:;
  }
  __pyx_L3:;

  /* "pandas/_libs/intervaltree.pxi":2544
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __PYX_XDEC_MEMVIEW(&__pyx_t_16, 1);
  __Pyx_AddTraceback("pandas._libs.interval.Int64ClosedRightIntervalNode.query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_indices, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_values, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_13__pyx_fuse_2query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_2__pyx_mdef_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_13__pyx_fuse_2query = {"__pyx_fuse_2query", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_13__pyx_fuse_2query, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_2query};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_13__pyx_fuse_2query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result = 0;
  __pyx_t_5numpy_int64_t __pyx_v_point;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fuse_2query (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_result,&__pyx_n_s_point,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_result)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_point)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_2query", 1, 2, 2, 1); __PYX_ERR(1, 2544, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_2query") < 0)) __PYX_ERR(1, 2544, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_result = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)values[0]);
    __pyx_v_point = __Pyx_PyInt_As_npy_int64(values[1]); if (unlikely((__pyx_v_point == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(1, 2544, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_2query", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 2544, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Int64ClosedRightIntervalNode.__pyx_fuse_2query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_result), __pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector, 1, "result", 0))) __PYX_ERR(1, 2544, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_12__pyx_fuse_2query(((struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode *)__pyx_v_self), __pyx_v_result, __pyx_v_point);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_12__pyx_fuse_2query(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int64_t __pyx_v_point) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__pyx_fuse_2query", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_fuse_2__pyx_f_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_query(__pyx_v_self, __pyx_v_result, __pyx_v_point, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2544, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Int64ClosedRightIntervalNode.__pyx_fuse_2query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_15__pyx_fuse_3query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_fuse_3__pyx_f_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int32_t __pyx_v_point, int __pyx_skip_dispatch) {
  __Pyx_memviewslice __pyx_v_indices = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_values = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __pyx_t_5numpy_int64_t __pyx_t_9;
  __pyx_t_5numpy_int64_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  int __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  __Pyx_memviewslice __pyx_t_16 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_17;
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  __Pyx_RefNannySetupContext("__pyx_fuse_3query", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pyx_fuse_3query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2544, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_15__pyx_fuse_3query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = __Pyx_PyInt_From_npy_int32(__pyx_v_point); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2544, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2544, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2544, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 2544, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_5) {
            __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2544, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2544, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_3query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_4 = __Pyx_PyInt_From_npy_int32(__pyx_v_point); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 2544, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_7 = __pyx_t_1; __pyx_t_3 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2544, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2544, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 2544, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (__pyx_t_3) {
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_t_4);
          __pyx_t_4 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2544, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pandas/_libs/intervaltree.pxi":2553
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
  __pyx_t_8 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_8) {

    /* "pandas/_libs/intervaltree.pxi":2557
 *             # continue the binary tree structure. Instead, we use linear
 *             # search.
 *             for i in range(self.n_elements):             # <<<<<<<<<<<<<<
 *                 if self.left[i] < point <= self.right[i]:
 *                     result.append(self.indices[i])
 */
    __pyx_t_9 = __pyx_v_self->n_elements;
    __pyx_t_10 = __pyx_t_9;
    for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
      __pyx_v_i = __pyx_t_11;

      /* "pandas/_libs/intervaltree.pxi":2558
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point <= self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      __pyx_t_12 = __pyx_v_i;
      __pyx_t_8 = ((*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->left.data + __pyx_t_12 * __pyx_v_self->left.strides[0]) ))) < __pyx_v_point);
      if (__pyx_t_8) {
        __pyx_t_13 = __pyx_v_i;
        __pyx_t_8 = (__pyx_v_point <= (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->right.data + __pyx_t_13 * __pyx_v_self->right.strides[0]) ))));
      }
      __pyx_t_14 = (__pyx_t_8 != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":2559
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point <= self.right[i]:
 *                     result.append(self.indices[i])             # <<<<<<<<<<<<<<
 *         else:
 *             # There are child nodes. Based on comparing our query to the pivot,
 */
        __pyx_t_15 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->indices.data + __pyx_t_15 * __pyx_v_self->indices.strides[0]) ))));

        /* "pandas/_libs/intervaltree.pxi":2558
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point <= self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      }
    }

    /* "pandas/_libs/intervaltree.pxi":2553
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
    goto __pyx_L3;
  }

  /* "pandas/_libs/intervaltree.pxi":2563
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
  /*else*/ {
    __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":2564
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:
 *                 values = self.center_left_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 */
      __pyx_t_16 = __pyx_v_self->center_left_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":2565
 *             if point < self.pivot:
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:
 */
      __pyx_t_16 = __pyx_v_self->center_left_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_indices = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":2566
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):             # <<<<<<<<<<<<<<
 *                     if not values[i] < point:
 *                         break
 */
      __pyx_t_9 = __pyx_v_self->n_center;
      __pyx_t_10 = __pyx_t_9;
      for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":2567
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_17 = __pyx_v_i;
        __pyx_t_14 = ((!(((*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_17 * __pyx_v_values.strides[0]) ))) < __pyx_v_point) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":2568
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:
 */
          goto __pyx_L9_break;

          /* "pandas/_libs/intervaltree.pxi":2567
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":2569
 *                     if not values[i] < point:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 */
        __pyx_t_18 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_18 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L9_break:;

      /* "pandas/_libs/intervaltree.pxi":2570
 *                         break
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      __pyx_t_14 = ((__pyx_v_point <= __pyx_v_self->left_node->max_right) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":2571
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)             # <<<<<<<<<<<<<<
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode *)__pyx_v_self->left_node->__pyx_vtab)->__pyx_fuse_3query(__pyx_v_self->left_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2571, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":2570
 *                         break
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":2563
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":2572
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
    __pyx_t_14 = ((__pyx_v_point > __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":2573
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 *                 values = self.center_right_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 */
      __pyx_t_16 = __pyx_v_self->center_right_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":2574
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:
 */
      __pyx_t_16 = __pyx_v_self->center_right_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_indices = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":2575
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):             # <<<<<<<<<<<<<<
 *                     if not point <= values[i]:
 *                         break
 */
      for (__pyx_t_11 = (__pyx_v_self->n_center - 1); __pyx_t_11 > -1L; __pyx_t_11-=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":2576
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_19 = __pyx_v_i;
        __pyx_t_14 = ((!((__pyx_v_point <= (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_19 * __pyx_v_values.strides[0]) )))) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":2577
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:
 */
          goto __pyx_L13_break;

          /* "pandas/_libs/intervaltree.pxi":2576
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":2578
 *                     if not point <= values[i]:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if self.right_node.min_left < point:
 *                     self.right_node.query(result, point)
 */
        __pyx_t_20 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_20 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L13_break:;

      /* "pandas/_libs/intervaltree.pxi":2579
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      __pyx_t_14 = ((__pyx_v_self->right_node->min_left < __pyx_v_point) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":2580
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:
 *                     self.right_node.query(result, point)             # <<<<<<<<<<<<<<
 *             else:
 *                 result.extend(self.center_left_indices)
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode *)__pyx_v_self->right_node->__pyx_vtab)->__pyx_fuse_3query(__pyx_v_self->right_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2580, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":2579
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":2572
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":2582
 *                     self.right_node.query(result, point)
 *             else:
 *                 result.extend(self.center_left_indices)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
    /*else*/ {
      __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->extend(__pyx_v_result, __pyx_v_self->center_left_indices); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2582, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_L7:;
  }
  __pyx_L3:;

  /* "pandas/_libs/intervaltree.pxi":2544
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __PYX_XDEC_MEMVIEW(&__pyx_t_16, 1);
  __Pyx_AddTraceback("pandas._libs.interval.Int64ClosedRightIntervalNode.query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_indices, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_values, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_15__pyx_fuse_3query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_3__pyx_mdef_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_15__pyx_fuse_3query = {"__pyx_fuse_3query", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_15__pyx_fuse_3query, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_2query};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_15__pyx_fuse_3query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result = 0;
  __pyx_t_5numpy_int32_t __pyx_v_point;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fuse_3query (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_result,&__pyx_n_s_point,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_result)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_point)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_3query", 1, 2, 2, 1); __PYX_ERR(1, 2544, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_3query") < 0)) __PYX_ERR(1, 2544, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_result = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)values[0]);
    __pyx_v_point = __Pyx_PyInt_As_npy_int32(values[1]); if (unlikely((__pyx_v_point == ((npy_int32)-1)) && PyErr_Occurred())) __PYX_ERR(1, 2544, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_3query", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 2544, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Int64ClosedRightIntervalNode.__pyx_fuse_3query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_result), __pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector, 1, "result", 0))) __PYX_ERR(1, 2544, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_14__pyx_fuse_3query(((struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode *)__pyx_v_self), __pyx_v_result, __pyx_v_point);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_14__pyx_fuse_3query(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int32_t __pyx_v_point) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__pyx_fuse_3query", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_fuse_3__pyx_f_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_query(__pyx_v_self, __pyx_v_result, __pyx_v_point, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2544, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Int64ClosedRightIntervalNode.__pyx_fuse_3query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_17__pyx_fuse_4query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_fuse_4__pyx_f_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_uint64_t __pyx_v_point, int __pyx_skip_dispatch) {
  __Pyx_memviewslice __pyx_v_indices = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_values = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __pyx_t_5numpy_int64_t __pyx_t_9;
  __pyx_t_5numpy_int64_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  int __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  __Pyx_memviewslice __pyx_t_16 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_17;
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  __Pyx_RefNannySetupContext("__pyx_fuse_4query", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pyx_fuse_4query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2544, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_17__pyx_fuse_4query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = __Pyx_PyInt_From_npy_uint64(__pyx_v_point); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2544, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2544, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2544, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 2544, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_5) {
            __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2544, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2544, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_3query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_4 = __Pyx_PyInt_From_npy_uint64(__pyx_v_point); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 2544, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_7 = __pyx_t_1; __pyx_t_3 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2544, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2544, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 2544, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (__pyx_t_3) {
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_t_4);
          __pyx_t_4 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2544, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pandas/_libs/intervaltree.pxi":2553
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
  __pyx_t_8 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_8) {

    /* "pandas/_libs/intervaltree.pxi":2557
 *             # continue the binary tree structure. Instead, we use linear
 *             # search.
 *             for i in range(self.n_elements):             # <<<<<<<<<<<<<<
 *                 if self.left[i] < point <= self.right[i]:
 *                     result.append(self.indices[i])
 */
    __pyx_t_9 = __pyx_v_self->n_elements;
    __pyx_t_10 = __pyx_t_9;
    for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
      __pyx_v_i = __pyx_t_11;

      /* "pandas/_libs/intervaltree.pxi":2558
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point <= self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      __pyx_t_12 = __pyx_v_i;
      __pyx_t_8 = ((*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->left.data + __pyx_t_12 * __pyx_v_self->left.strides[0]) ))) < __pyx_v_point);
      if (__pyx_t_8) {
        __pyx_t_13 = __pyx_v_i;
        __pyx_t_8 = (__pyx_v_point <= (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->right.data + __pyx_t_13 * __pyx_v_self->right.strides[0]) ))));
      }
      __pyx_t_14 = (__pyx_t_8 != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":2559
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point <= self.right[i]:
 *                     result.append(self.indices[i])             # <<<<<<<<<<<<<<
 *         else:
 *             # There are child nodes. Based on comparing our query to the pivot,
 */
        __pyx_t_15 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->indices.data + __pyx_t_15 * __pyx_v_self->indices.strides[0]) ))));

        /* "pandas/_libs/intervaltree.pxi":2558
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point <= self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      }
    }

    /* "pandas/_libs/intervaltree.pxi":2553
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
    goto __pyx_L3;
  }

  /* "pandas/_libs/intervaltree.pxi":2563
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
  /*else*/ {
    __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":2564
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:
 *                 values = self.center_left_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 */
      __pyx_t_16 = __pyx_v_self->center_left_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":2565
 *             if point < self.pivot:
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:
 */
      __pyx_t_16 = __pyx_v_self->center_left_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_indices = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":2566
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):             # <<<<<<<<<<<<<<
 *                     if not values[i] < point:
 *                         break
 */
      __pyx_t_9 = __pyx_v_self->n_center;
      __pyx_t_10 = __pyx_t_9;
      for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":2567
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_17 = __pyx_v_i;
        __pyx_t_14 = ((!(((*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_17 * __pyx_v_values.strides[0]) ))) < __pyx_v_point) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":2568
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:
 */
          goto __pyx_L9_break;

          /* "pandas/_libs/intervaltree.pxi":2567
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":2569
 *                     if not values[i] < point:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 */
        __pyx_t_18 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_18 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L9_break:;

      /* "pandas/_libs/intervaltree.pxi":2570
 *                         break
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      __pyx_t_14 = ((__pyx_v_point <= __pyx_v_self->left_node->max_right) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":2571
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)             # <<<<<<<<<<<<<<
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode *)__pyx_v_self->left_node->__pyx_vtab)->__pyx_fuse_4query(__pyx_v_self->left_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2571, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":2570
 *                         break
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":2563
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":2572
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
    __pyx_t_14 = ((__pyx_v_point > __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":2573
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 *                 values = self.center_right_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 */
      __pyx_t_16 = __pyx_v_self->center_right_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":2574
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:
 */
      __pyx_t_16 = __pyx_v_self->center_right_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_indices = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":2575
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):             # <<<<<<<<<<<<<<
 *                     if not point <= values[i]:
 *                         break
 */
      for (__pyx_t_11 = (__pyx_v_self->n_center - 1); __pyx_t_11 > -1L; __pyx_t_11-=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":2576
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_19 = __pyx_v_i;
        __pyx_t_14 = ((!((__pyx_v_point <= (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_19 * __pyx_v_values.strides[0]) )))) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":2577
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:
 */
          goto __pyx_L13_break;

          /* "pandas/_libs/intervaltree.pxi":2576
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":2578
 *                     if not point <= values[i]:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if self.right_node.min_left < point:
 *                     self.right_node.query(result, point)
 */
        __pyx_t_20 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_20 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L13_break:;

      /* "pandas/_libs/intervaltree.pxi":2579
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      __pyx_t_14 = ((__pyx_v_self->right_node->min_left < __pyx_v_point) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":2580
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:
 *                     self.right_node.query(result, point)             # <<<<<<<<<<<<<<
 *             else:
 *                 result.extend(self.center_left_indices)
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode *)__pyx_v_self->right_node->__pyx_vtab)->__pyx_fuse_4query(__pyx_v_self->right_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2580, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":2579
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":2572
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":2582
 *                     self.right_node.query(result, point)
 *             else:
 *                 result.extend(self.center_left_indices)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
    /*else*/ {
      __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->extend(__pyx_v_result, __pyx_v_self->center_left_indices); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2582, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_L7:;
  }
  __pyx_L3:;

  /* "pandas/_libs/intervaltree.pxi":2544
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __PYX_XDEC_MEMVIEW(&__pyx_t_16, 1);
  __Pyx_AddTraceback("pandas._libs.interval.Int64ClosedRightIntervalNode.query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_indices, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_values, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_17__pyx_fuse_4query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_4__pyx_mdef_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_17__pyx_fuse_4query = {"__pyx_fuse_4query", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_17__pyx_fuse_4query, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_2query};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_17__pyx_fuse_4query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result = 0;
  __pyx_t_5numpy_uint64_t __pyx_v_point;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fuse_4query (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_result,&__pyx_n_s_point,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_result)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_point)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_4query", 1, 2, 2, 1); __PYX_ERR(1, 2544, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_4query") < 0)) __PYX_ERR(1, 2544, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_result = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)values[0]);
    __pyx_v_point = __Pyx_PyInt_As_npy_uint64(values[1]); if (unlikely((__pyx_v_point == ((npy_uint64)-1)) && PyErr_Occurred())) __PYX_ERR(1, 2544, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_4query", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 2544, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Int64ClosedRightIntervalNode.__pyx_fuse_4query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_result), __pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector, 1, "result", 0))) __PYX_ERR(1, 2544, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_16__pyx_fuse_4query(((struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode *)__pyx_v_self), __pyx_v_result, __pyx_v_point);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_16__pyx_fuse_4query(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_uint64_t __pyx_v_point) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__pyx_fuse_4query", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_fuse_4__pyx_f_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_query(__pyx_v_self, __pyx_v_result, __pyx_v_point, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2544, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Int64ClosedRightIntervalNode.__pyx_fuse_4query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":2584
 *                 result.extend(self.center_left_indices)
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         if self.is_leaf_node:
 *             return ('<Int64ClosedRightIntervalNode: '
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_5__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_5__repr__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_4__repr__(((struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_4__repr__(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode *__pyx_v_self) {
  PyObject *__pyx_v_n_left = NULL;
  PyObject *__pyx_v_n_right = NULL;
  PyObject *__pyx_v_n_center = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  Py_UCS4 __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "pandas/_libs/intervaltree.pxi":2585
 * 
 *     def __repr__(self):
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             return ('<Int64ClosedRightIntervalNode: '
 *                     '%s elements (terminal)>' % self.n_elements)
 */
  __pyx_t_1 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_1) {

    /* "pandas/_libs/intervaltree.pxi":2586
 *     def __repr__(self):
 *         if self.is_leaf_node:
 *             return ('<Int64ClosedRightIntervalNode: '             # <<<<<<<<<<<<<<
 *                     '%s elements (terminal)>' % self.n_elements)
 *         else:
 */
    __Pyx_XDECREF(__pyx_r);

    /* "pandas/_libs/intervaltree.pxi":2587
 *         if self.is_leaf_node:
 *             return ('<Int64ClosedRightIntervalNode: '
 *                     '%s elements (terminal)>' % self.n_elements)             # <<<<<<<<<<<<<<
 *         else:
 *             n_left = self.left_node.n_elements
 */
    __pyx_t_2 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_elements); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2587, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PyUnicode_Format(__pyx_kp_u_Int64ClosedRightIntervalNode_s, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2587, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pandas/_libs/intervaltree.pxi":2585
 * 
 *     def __repr__(self):
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             return ('<Int64ClosedRightIntervalNode: '
 *                     '%s elements (terminal)>' % self.n_elements)
 */
  }

  /* "pandas/_libs/intervaltree.pxi":2589
 *                     '%s elements (terminal)>' % self.n_elements)
 *         else:
 *             n_left = self.left_node.n_elements             # <<<<<<<<<<<<<<
 *             n_right = self.right_node.n_elements
 *             n_center = self.n_elements - n_left - n_right
 */
  /*else*/ {
    __pyx_t_3 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->left_node->n_elements); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2589, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_v_n_left = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "pandas/_libs/intervaltree.pxi":2590
 *         else:
 *             n_left = self.left_node.n_elements
 *             n_right = self.right_node.n_elements             # <<<<<<<<<<<<<<
 *             n_center = self.n_elements - n_left - n_right
 *             return ('<Int64ClosedRightIntervalNode: '
 */
    __pyx_t_3 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->right_node->n_elements); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2590, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_v_n_right = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "pandas/_libs/intervaltree.pxi":2591
 *             n_left = self.left_node.n_elements
 *             n_right = self.right_node.n_elements
 *             n_center = self.n_elements - n_left - n_right             # <<<<<<<<<<<<<<
 *             return ('<Int64ClosedRightIntervalNode: '
 *                     'pivot %s, %s elements (%s left, %s right, %s '
 */
    __pyx_t_3 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_elements); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2591, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = PyNumber_Subtract(__pyx_t_3, __pyx_v_n_left); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2591, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyNumber_Subtract(__pyx_t_2, __pyx_v_n_right); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2591, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_n_center = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "pandas/_libs/intervaltree.pxi":2592
 *             n_right = self.right_node.n_elements
 *             n_center = self.n_elements - n_left - n_right
 *             return ('<Int64ClosedRightIntervalNode: '             # <<<<<<<<<<<<<<
 *                     'pivot %s, %s elements (%s left, %s right, %s '
 *                     'overlapping)>' % (self.pivot, self.n_elements,
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = PyTuple_New(11); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2592, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = 0;
    __pyx_t_5 = 127;
    __Pyx_INCREF(__pyx_kp_u_Int64ClosedRightIntervalNode_pi);
    __pyx_t_4 += 37;
    __Pyx_GIVEREF(__pyx_kp_u_Int64ClosedRightIntervalNode_pi);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_kp_u_Int64ClosedRightIntervalNode_pi);

    /* "pandas/_libs/intervaltree.pxi":2594
 *             return ('<Int64ClosedRightIntervalNode: '
 *                     'pivot %s, %s elements (%s left, %s right, %s '
 *                     'overlapping)>' % (self.pivot, self.n_elements,             # <<<<<<<<<<<<<<
 *                                        n_left, n_right, n_center))
 * 
 */
    __pyx_t_2 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->pivot); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2594, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_t_2), __pyx_empty_unicode); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 2594, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_5 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_6) > __pyx_t_5) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_6) : __pyx_t_5;
    __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_6);
    __pyx_t_6 = 0;
    __Pyx_INCREF(__pyx_kp_u__18);
    __pyx_t_4 += 2;
    __Pyx_GIVEREF(__pyx_kp_u__18);
    PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_kp_u__18);
    __pyx_t_6 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_elements); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 2594, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_2 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_t_6), __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2594, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_5 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) > __pyx_t_5) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) : __pyx_t_5;
    __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_3, 3, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_elements);
    __pyx_t_4 += 11;
    __Pyx_GIVEREF(__pyx_kp_u_elements);
    PyTuple_SET_ITEM(__pyx_t_3, 4, __pyx_kp_u_elements);

    /* "pandas/_libs/intervaltree.pxi":2595
 *                     'pivot %s, %s elements (%s left, %s right, %s '
 *                     'overlapping)>' % (self.pivot, self.n_elements,
 *                                        n_left, n_right, n_center))             # <<<<<<<<<<<<<<
 * 
 *     def counts(self):
 */
    __pyx_t_2 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_v_n_left), __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2595, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) > __pyx_t_5) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) : __pyx_t_5;
    __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_3, 5, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_left_2);
    __pyx_t_4 += 7;
    __Pyx_GIVEREF(__pyx_kp_u_left_2);
    PyTuple_SET_ITEM(__pyx_t_3, 6, __pyx_kp_u_left_2);
    __pyx_t_2 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_v_n_right), __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2595, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) > __pyx_t_5) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) : __pyx_t_5;
    __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_3, 7, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_right_2);
    __pyx_t_4 += 8;
    __Pyx_GIVEREF(__pyx_kp_u_right_2);
    PyTuple_SET_ITEM(__pyx_t_3, 8, __pyx_kp_u_right_2);
    __pyx_t_2 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_v_n_center), __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2595, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) > __pyx_t_5) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) : __pyx_t_5;
    __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_3, 9, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_overlapping);
    __pyx_t_4 += 14;
    __Pyx_GIVEREF(__pyx_kp_u_overlapping);
    PyTuple_SET_ITEM(__pyx_t_3, 10, __pyx_kp_u_overlapping);

    /* "pandas/_libs/intervaltree.pxi":2592
 *             n_right = self.right_node.n_elements
 *             n_center = self.n_elements - n_left - n_right
 *             return ('<Int64ClosedRightIntervalNode: '             # <<<<<<<<<<<<<<
 *                     'pivot %s, %s elements (%s left, %s right, %s '
 *                     'overlapping)>' % (self.pivot, self.n_elements,
 */
    __pyx_t_2 = __Pyx_PyUnicode_Join(__pyx_t_3, 11, __pyx_t_4, __pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2592, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;
  }

  /* "pandas/_libs/intervaltree.pxi":2584
 *                 result.extend(self.center_left_indices)
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         if self.is_leaf_node:
 *             return ('<Int64ClosedRightIntervalNode: '
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pandas._libs.interval.Int64ClosedRightIntervalNode.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_n_left);
  __Pyx_XDECREF(__pyx_v_n_right);
  __Pyx_XDECREF(__pyx_v_n_center);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":2597
 *                                        n_left, n_right, n_center))
 * 
 *     def counts(self):             # <<<<<<<<<<<<<<
 *         """
 *         Inspect counts on this node
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_7counts(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_6counts[] = "\n        Inspect counts on this node\n        useful for debugging purposes\n        ";
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_7counts(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("counts (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_6counts(((struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_6counts(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode *__pyx_v_self) {
  Py_ssize_t __pyx_v_m;
  PyObject *__pyx_v_l = NULL;
  PyObject *__pyx_v_r = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  size_t __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("counts", 0);

  /* "pandas/_libs/intervaltree.pxi":2602
 *         useful for debugging purposes
 *         """
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             return self.n_elements
 *         else:
 */
  __pyx_t_1 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_1) {

    /* "pandas/_libs/intervaltree.pxi":2603
 *         """
 *         if self.is_leaf_node:
 *             return self.n_elements             # <<<<<<<<<<<<<<
 *         else:
 *             m = len(self.center_left_values)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_elements); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2603, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "pandas/_libs/intervaltree.pxi":2602
 *         useful for debugging purposes
 *         """
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             return self.n_elements
 *         else:
 */
  }

  /* "pandas/_libs/intervaltree.pxi":2605
 *             return self.n_elements
 *         else:
 *             m = len(self.center_left_values)             # <<<<<<<<<<<<<<
 *             l = self.left_node.counts()
 *             r = self.right_node.counts()
 */
  /*else*/ {
    if (unlikely(!__pyx_v_self->center_left_values.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(1, 2605, __pyx_L1_error)}
    __pyx_t_3 = __Pyx_MemoryView_Len(__pyx_v_self->center_left_values); 
    __pyx_v_m = __pyx_t_3;

    /* "pandas/_libs/intervaltree.pxi":2606
 *         else:
 *             m = len(self.center_left_values)
 *             l = self.left_node.counts()             # <<<<<<<<<<<<<<
 *             r = self.right_node.counts()
 *             return (m, (l, r))
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->left_node), __pyx_n_s_counts); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 2606, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2606, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_l = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "pandas/_libs/intervaltree.pxi":2607
 *             m = len(self.center_left_values)
 *             l = self.left_node.counts()
 *             r = self.right_node.counts()             # <<<<<<<<<<<<<<
 *             return (m, (l, r))
 * 
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->right_node), __pyx_n_s_counts); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 2607, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2607, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_r = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "pandas/_libs/intervaltree.pxi":2608
 *             l = self.left_node.counts()
 *             r = self.right_node.counts()
 *             return (m, (l, r))             # <<<<<<<<<<<<<<
 * 
 * NODE_CLASSES['int64',
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = PyInt_FromSsize_t(__pyx_v_m); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2608, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 2608, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_v_l);
    __Pyx_GIVEREF(__pyx_v_l);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_l);
    __Pyx_INCREF(__pyx_v_r);
    __Pyx_GIVEREF(__pyx_v_r);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_v_r);
    __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 2608, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_4);
    __pyx_t_2 = 0;
    __pyx_t_4 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;
  }

  /* "pandas/_libs/intervaltree.pxi":2597
 *                                        n_left, n_right, n_center))
 * 
 *     def counts(self):             # <<<<<<<<<<<<<<
 *         """
 *         Inspect counts on this node
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pandas._libs.interval.Int64ClosedRightIntervalNode.counts", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_l);
  __Pyx_XDECREF(__pyx_v_r);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":2454
 *         int64_t[:] center_left_indices, center_right_indices, indices
 *         int64_t min_left, max_right
 *         readonly int64_t pivot             # <<<<<<<<<<<<<<
 *         readonly int64_t n_elements, n_center, leaf_size
 *         readonly bint is_leaf_node
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_5pivot_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_5pivot_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_5pivot___get__(((struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_5pivot___get__(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->pivot); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2454, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Int64ClosedRightIntervalNode.pivot.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":2455
 *         int64_t min_left, max_right
 *         readonly int64_t pivot
 *         readonly int64_t n_elements, n_center, leaf_size             # <<<<<<<<<<<<<<
 *         readonly bint is_leaf_node
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_10n_elements_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_10n_elements_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_10n_elements___get__(((struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_10n_elements___get__(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_elements); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2455, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Int64ClosedRightIntervalNode.n_elements.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_8n_center_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_8n_center_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_8n_center___get__(((struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_8n_center___get__(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_center); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2455, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Int64ClosedRightIntervalNode.n_center.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_9leaf_size_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_9leaf_size_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_9leaf_size___get__(((struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_9leaf_size___get__(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->leaf_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2455, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Int64ClosedRightIntervalNode.leaf_size.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":2456
 *         readonly int64_t pivot
 *         readonly int64_t n_elements, n_center, leaf_size
 *         readonly bint is_leaf_node             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self,
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_12is_leaf_node_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_12is_leaf_node_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_12is_leaf_node___get__(((struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_12is_leaf_node___get__(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->is_leaf_node); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2456, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Int64ClosedRightIntervalNode.is_leaf_node.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_21__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_21__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_20__reduce_cython__(((struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_20__reduce_cython__(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  int __pyx_t_16;
  int __pyx_t_17;
  int __pyx_t_18;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.center_left_indices, self.center_left_values, self.center_right_indices, self.center_right_values, self.indices, self.is_leaf_node, self.leaf_size, self.left, self.left_node, self.max_right, self.min_left, self.n_center, self.n_elements, self.pivot, self.right, self.right_node)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  if (unlikely(!__pyx_v_self->center_left_indices.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_1 = __pyx_memoryview_fromslice(__pyx_v_self->center_left_indices, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_int64_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_int64_t, 0);; if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (unlikely(!__pyx_v_self->center_left_values.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_2 = __pyx_memoryview_fromslice(__pyx_v_self->center_left_values, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_int64_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_int64_t, 0);; if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (unlikely(!__pyx_v_self->center_right_indices.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_3 = __pyx_memoryview_fromslice(__pyx_v_self->center_right_indices, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_int64_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_int64_t, 0);; if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely(!__pyx_v_self->center_right_values.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_4 = __pyx_memoryview_fromslice(__pyx_v_self->center_right_values, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_int64_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_int64_t, 0);; if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (unlikely(!__pyx_v_self->indices.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_5 = __pyx_memoryview_fromslice(__pyx_v_self->indices, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_int64_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_int64_t, 0);; if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyBool_FromLong(__pyx_v_self->is_leaf_node); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->leaf_size); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (unlikely(!__pyx_v_self->left.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_8 = __pyx_memoryview_fromslice(__pyx_v_self->left, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_int64_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_int64_t, 0);; if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->max_right); if (unlikely(!__pyx_t_9)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_10 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->min_left); if (unlikely(!__pyx_t_10)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_11 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_center); if (unlikely(!__pyx_t_11)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_12 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_elements); if (unlikely(!__pyx_t_12)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __pyx_t_13 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->pivot); if (unlikely(!__pyx_t_13)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  if (unlikely(!__pyx_v_self->right.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_14 = __pyx_memoryview_fromslice(__pyx_v_self->right, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_int64_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_int64_t, 0);; if (unlikely(!__pyx_t_14)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __pyx_t_15 = PyTuple_New(16); if (unlikely(!__pyx_t_15)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_15, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_15, 1, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_15, 2, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_15, 3, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_15, 4, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_15, 5, __pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_15, 6, __pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_15, 7, __pyx_t_8);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->left_node));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self->left_node));
  PyTuple_SET_ITEM(__pyx_t_15, 8, ((PyObject *)__pyx_v_self->left_node));
  __Pyx_GIVEREF(__pyx_t_9);
  PyTuple_SET_ITEM(__pyx_t_15, 9, __pyx_t_9);
  __Pyx_GIVEREF(__pyx_t_10);
  PyTuple_SET_ITEM(__pyx_t_15, 10, __pyx_t_10);
  __Pyx_GIVEREF(__pyx_t_11);
  PyTuple_SET_ITEM(__pyx_t_15, 11, __pyx_t_11);
  __Pyx_GIVEREF(__pyx_t_12);
  PyTuple_SET_ITEM(__pyx_t_15, 12, __pyx_t_12);
  __Pyx_GIVEREF(__pyx_t_13);
  PyTuple_SET_ITEM(__pyx_t_15, 13, __pyx_t_13);
  __Pyx_GIVEREF(__pyx_t_14);
  PyTuple_SET_ITEM(__pyx_t_15, 14, __pyx_t_14);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->right_node));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self->right_node));
  PyTuple_SET_ITEM(__pyx_t_15, 15, ((PyObject *)__pyx_v_self->right_node));
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  __pyx_t_6 = 0;
  __pyx_t_7 = 0;
  __pyx_t_8 = 0;
  __pyx_t_9 = 0;
  __pyx_t_10 = 0;
  __pyx_t_11 = 0;
  __pyx_t_12 = 0;
  __pyx_t_13 = 0;
  __pyx_t_14 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_15);
  __pyx_t_15 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.center_left_indices, self.center_left_values, self.center_right_indices, self.center_right_values, self.indices, self.is_leaf_node, self.leaf_size, self.left, self.left_node, self.max_right, self.min_left, self.n_center, self.n_elements, self.pivot, self.right, self.right_node)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_15 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_15)) __PYX_ERR(2, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __pyx_v__dict = __pyx_t_15;
  __pyx_t_15 = 0;

  /* "(tree fragment)":7
 *     state = (self.center_left_indices, self.center_left_values, self.center_right_indices, self.center_right_values, self.indices, self.is_leaf_node, self.leaf_size, self.left, self.left_node, self.max_right, self.min_left, self.n_center, self.n_elements, self.pivot, self.right, self.right_node)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_16 = (__pyx_v__dict != Py_None);
  __pyx_t_17 = (__pyx_t_16 != 0);
  if (__pyx_t_17) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_15 = PyTuple_New(1); if (unlikely(!__pyx_t_15)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    PyTuple_SET_ITEM(__pyx_t_15, 0, __pyx_v__dict);
    __pyx_t_14 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_15); if (unlikely(!__pyx_t_14)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_14));
    __pyx_t_14 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.left_node is not None or self.right_node is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.center_left_indices, self.center_left_values, self.center_right_indices, self.center_right_values, self.indices, self.is_leaf_node, self.leaf_size, self.left, self.left_node, self.max_right, self.min_left, self.n_center, self.n_elements, self.pivot, self.right, self.right_node)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.left_node is not None or self.right_node is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_Int64ClosedRightIntervalNode, (type(self), 0x7a104e5, None), state
 */
  /*else*/ {
    __pyx_t_16 = (((PyObject *)__pyx_v_self->left_node) != Py_None);
    __pyx_t_18 = (__pyx_t_16 != 0);
    if (!__pyx_t_18) {
    } else {
      __pyx_t_17 = __pyx_t_18;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_18 = (((PyObject *)__pyx_v_self->right_node) != Py_None);
    __pyx_t_16 = (__pyx_t_18 != 0);
    __pyx_t_17 = __pyx_t_16;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_17;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.left_node is not None or self.right_node is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Int64ClosedRightIntervalNode, (type(self), 0x7a104e5, None), state
 *     else:
 */
  __pyx_t_17 = (__pyx_v_use_setstate != 0);
  if (__pyx_t_17) {

    /* "(tree fragment)":13
 *         use_setstate = self.left_node is not None or self.right_node is not None
 *     if use_setstate:
 *         return __pyx_unpickle_Int64ClosedRightIntervalNode, (type(self), 0x7a104e5, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_Int64ClosedRightIntervalNode, (type(self), 0x7a104e5, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_14, __pyx_n_s_pyx_unpickle_Int64ClosedRightI); if (unlikely(!__pyx_t_14)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __pyx_t_15 = PyTuple_New(3); if (unlikely(!__pyx_t_15)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_15, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_127993061);
    __Pyx_GIVEREF(__pyx_int_127993061);
    PyTuple_SET_ITEM(__pyx_t_15, 1, __pyx_int_127993061);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_15, 2, Py_None);
    __pyx_t_13 = PyTuple_New(3); if (unlikely(!__pyx_t_13)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __Pyx_GIVEREF(__pyx_t_14);
    PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_t_14);
    __Pyx_GIVEREF(__pyx_t_15);
    PyTuple_SET_ITEM(__pyx_t_13, 1, __pyx_t_15);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_13, 2, __pyx_v_state);
    __pyx_t_14 = 0;
    __pyx_t_15 = 0;
    __pyx_r = __pyx_t_13;
    __pyx_t_13 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.left_node is not None or self.right_node is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Int64ClosedRightIntervalNode, (type(self), 0x7a104e5, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_Int64ClosedRightIntervalNode, (type(self), 0x7a104e5, None), state
 *     else:
 *         return __pyx_unpickle_Int64ClosedRightIntervalNode, (type(self), 0x7a104e5, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Int64ClosedRightIntervalNode__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_13, __pyx_n_s_pyx_unpickle_Int64ClosedRightI); if (unlikely(!__pyx_t_13)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __pyx_t_15 = PyTuple_New(3); if (unlikely(!__pyx_t_15)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_15, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_127993061);
    __Pyx_GIVEREF(__pyx_int_127993061);
    PyTuple_SET_ITEM(__pyx_t_15, 1, __pyx_int_127993061);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_15, 2, __pyx_v_state);
    __pyx_t_14 = PyTuple_New(2); if (unlikely(!__pyx_t_14)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_GIVEREF(__pyx_t_13);
    PyTuple_SET_ITEM(__pyx_t_14, 0, __pyx_t_13);
    __Pyx_GIVEREF(__pyx_t_15);
    PyTuple_SET_ITEM(__pyx_t_14, 1, __pyx_t_15);
    __pyx_t_13 = 0;
    __pyx_t_15 = 0;
    __pyx_r = __pyx_t_14;
    __pyx_t_14 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_AddTraceback("pandas._libs.interval.Int64ClosedRightIntervalNode.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Int64ClosedRightIntervalNode, (type(self), 0x7a104e5, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Int64ClosedRightIntervalNode__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_23__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_23__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_22__setstate_cython__(((struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_22__setstate_cython__(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_Int64ClosedRightIntervalNode, (type(self), 0x7a104e5, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Int64ClosedRightIntervalNode__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(2, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_6pandas_5_libs_8interval___pyx_unpickle_Int64ClosedRightIntervalNode__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Int64ClosedRightIntervalNode, (type(self), 0x7a104e5, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Int64ClosedRightIntervalNode__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Int64ClosedRightIntervalNode.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":2628
 *         readonly bint is_leaf_node
 * 
 *     def __init__(self,             # <<<<<<<<<<<<<<
 *                  ndarray[int64_t, ndim=1] left,
 *                  ndarray[int64_t, ndim=1] right,
 */

/* Python wrapper */
static int __pyx_pw_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyArrayObject *__pyx_v_left = 0;
  PyArrayObject *__pyx_v_right = 0;
  PyArrayObject *__pyx_v_indices = 0;
  __pyx_t_5numpy_int64_t __pyx_v_leaf_size;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_left,&__pyx_n_s_right,&__pyx_n_s_indices,&__pyx_n_s_leaf_size,0};
    PyObject* values[4] = {0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_left)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_right)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 4, 4, 1); __PYX_ERR(1, 2628, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_indices)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 4, 4, 2); __PYX_ERR(1, 2628, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_leaf_size)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 4, 4, 3); __PYX_ERR(1, 2628, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(1, 2628, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 4) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
    }
    __pyx_v_left = ((PyArrayObject *)values[0]);
    __pyx_v_right = ((PyArrayObject *)values[1]);
    __pyx_v_indices = ((PyArrayObject *)values[2]);
    __pyx_v_leaf_size = __Pyx_PyInt_As_npy_int64(values[3]); if (unlikely((__pyx_v_leaf_size == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(1, 2632, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 4, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 2628, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Int64ClosedBothIntervalNode.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_left), __pyx_ptype_5numpy_ndarray, 1, "left", 0))) __PYX_ERR(1, 2629, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_right), __pyx_ptype_5numpy_ndarray, 1, "right", 0))) __PYX_ERR(1, 2630, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_indices), __pyx_ptype_5numpy_ndarray, 1, "indices", 0))) __PYX_ERR(1, 2631, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode___init__(((struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode *)__pyx_v_self), __pyx_v_left, __pyx_v_right, __pyx_v_indices, __pyx_v_leaf_size);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode___init__(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode *__pyx_v_self, PyArrayObject *__pyx_v_left, PyArrayObject *__pyx_v_right, PyArrayObject *__pyx_v_indices, __pyx_t_5numpy_int64_t __pyx_v_leaf_size) {
  PyObject *__pyx_v_left_set = NULL;
  PyObject *__pyx_v_right_set = NULL;
  PyObject *__pyx_v_center_set = NULL;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_indices;
  __Pyx_Buffer __pyx_pybuffer_indices;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_left;
  __Pyx_Buffer __pyx_pybuffer_left;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_right;
  __Pyx_Buffer __pyx_pybuffer_right;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __pyx_t_5numpy_int64_t __pyx_t_6;
  __Pyx_memviewslice __pyx_t_7 = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  __Pyx_memviewslice __pyx_t_10 = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *(*__pyx_t_11)(PyObject *);
  size_t __pyx_t_12;
  __Pyx_RefNannySetupContext("__init__", 0);
  __pyx_pybuffer_left.pybuffer.buf = NULL;
  __pyx_pybuffer_left.refcount = 0;
  __pyx_pybuffernd_left.data = NULL;
  __pyx_pybuffernd_left.rcbuffer = &__pyx_pybuffer_left;
  __pyx_pybuffer_right.pybuffer.buf = NULL;
  __pyx_pybuffer_right.refcount = 0;
  __pyx_pybuffernd_right.data = NULL;
  __pyx_pybuffernd_right.rcbuffer = &__pyx_pybuffer_right;
  __pyx_pybuffer_indices.pybuffer.buf = NULL;
  __pyx_pybuffer_indices.refcount = 0;
  __pyx_pybuffernd_indices.data = NULL;
  __pyx_pybuffernd_indices.rcbuffer = &__pyx_pybuffer_indices;
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_left.rcbuffer->pybuffer, (PyObject*)__pyx_v_left, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 2628, __pyx_L1_error)
  }
  __pyx_pybuffernd_left.diminfo[0].strides = __pyx_pybuffernd_left.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_left.diminfo[0].shape = __pyx_pybuffernd_left.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_right.rcbuffer->pybuffer, (PyObject*)__pyx_v_right, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 2628, __pyx_L1_error)
  }
  __pyx_pybuffernd_right.diminfo[0].strides = __pyx_pybuffernd_right.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_right.diminfo[0].shape = __pyx_pybuffernd_right.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_indices.rcbuffer->pybuffer, (PyObject*)__pyx_v_indices, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 2628, __pyx_L1_error)
  }
  __pyx_pybuffernd_indices.diminfo[0].strides = __pyx_pybuffernd_indices.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_indices.diminfo[0].shape = __pyx_pybuffernd_indices.rcbuffer->pybuffer.shape[0];

  /* "pandas/_libs/intervaltree.pxi":2634
 *                  int64_t leaf_size):
 * 
 *         self.n_elements = len(left)             # <<<<<<<<<<<<<<
 *         self.leaf_size = leaf_size
 * 
 */
  __pyx_t_1 = PyObject_Length(((PyObject *)__pyx_v_left)); if (unlikely(__pyx_t_1 == ((Py_ssize_t)-1))) __PYX_ERR(1, 2634, __pyx_L1_error)
  __pyx_v_self->n_elements = __pyx_t_1;

  /* "pandas/_libs/intervaltree.pxi":2635
 * 
 *         self.n_elements = len(left)
 *         self.leaf_size = leaf_size             # <<<<<<<<<<<<<<
 * 
 *         # min_left and min_right are used to speed-up query by skipping
 */
  __pyx_v_self->leaf_size = __pyx_v_leaf_size;

  /* "pandas/_libs/intervaltree.pxi":2640
 *         # query on sub-nodes. If this node has size 0, query is cheap,
 *         # so these values don't matter.
 *         if left.size > 0:             # <<<<<<<<<<<<<<
 *             self.min_left = left.min()
 *             self.max_right = right.max()
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_left), __pyx_n_s_size); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2640, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyObject_RichCompare(__pyx_t_2, __pyx_int_0, Py_GT); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2640, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(1, 2640, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_4) {

    /* "pandas/_libs/intervaltree.pxi":2641
 *         # so these values don't matter.
 *         if left.size > 0:
 *             self.min_left = left.min()             # <<<<<<<<<<<<<<
 *             self.max_right = right.max()
 *         else:
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_left), __pyx_n_s_min); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2641, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2641, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_6 = __Pyx_PyInt_As_npy_int64(__pyx_t_3); if (unlikely((__pyx_t_6 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(1, 2641, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_self->min_left = __pyx_t_6;

    /* "pandas/_libs/intervaltree.pxi":2642
 *         if left.size > 0:
 *             self.min_left = left.min()
 *             self.max_right = right.max()             # <<<<<<<<<<<<<<
 *         else:
 *             self.min_left = 0
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_right), __pyx_n_s_max); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2642, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2642, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_6 = __Pyx_PyInt_As_npy_int64(__pyx_t_3); if (unlikely((__pyx_t_6 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(1, 2642, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_self->max_right = __pyx_t_6;

    /* "pandas/_libs/intervaltree.pxi":2640
 *         # query on sub-nodes. If this node has size 0, query is cheap,
 *         # so these values don't matter.
 *         if left.size > 0:             # <<<<<<<<<<<<<<
 *             self.min_left = left.min()
 *             self.max_right = right.max()
 */
    goto __pyx_L3;
  }

  /* "pandas/_libs/intervaltree.pxi":2644
 *             self.max_right = right.max()
 *         else:
 *             self.min_left = 0             # <<<<<<<<<<<<<<
 *             self.max_right = 0
 * 
 */
  /*else*/ {
    __pyx_v_self->min_left = 0;

    /* "pandas/_libs/intervaltree.pxi":2645
 *         else:
 *             self.min_left = 0
 *             self.max_right = 0             # <<<<<<<<<<<<<<
 * 
 *         if self.n_elements <= leaf_size:
 */
    __pyx_v_self->max_right = 0;
  }
  __pyx_L3:;

  /* "pandas/_libs/intervaltree.pxi":2647
 *             self.max_right = 0
 * 
 *         if self.n_elements <= leaf_size:             # <<<<<<<<<<<<<<
 *             # make this a terminal (leaf) node
 *             self.is_leaf_node = True
 */
  __pyx_t_4 = ((__pyx_v_self->n_elements <= __pyx_v_leaf_size) != 0);
  if (__pyx_t_4) {

    /* "pandas/_libs/intervaltree.pxi":2649
 *         if self.n_elements <= leaf_size:
 *             # make this a terminal (leaf) node
 *             self.is_leaf_node = True             # <<<<<<<<<<<<<<
 *             self.left = left
 *             self.right = right
 */
    __pyx_v_self->is_leaf_node = 1;

    /* "pandas/_libs/intervaltree.pxi":2650
 *             # make this a terminal (leaf) node
 *             self.is_leaf_node = True
 *             self.left = left             # <<<<<<<<<<<<<<
 *             self.right = right
 *             self.indices = indices
 */
    __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(((PyObject *)__pyx_v_left), PyBUF_WRITABLE); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(1, 2650, __pyx_L1_error)
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->left, 0);
    __pyx_v_self->left = __pyx_t_7;
    __pyx_t_7.memview = NULL;
    __pyx_t_7.data = NULL;

    /* "pandas/_libs/intervaltree.pxi":2651
 *             self.is_leaf_node = True
 *             self.left = left
 *             self.right = right             # <<<<<<<<<<<<<<
 *             self.indices = indices
 *             self.n_center = 0
 */
    __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(((PyObject *)__pyx_v_right), PyBUF_WRITABLE); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(1, 2651, __pyx_L1_error)
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->right, 0);
    __pyx_v_self->right = __pyx_t_7;
    __pyx_t_7.memview = NULL;
    __pyx_t_7.data = NULL;

    /* "pandas/_libs/intervaltree.pxi":2652
 *             self.left = left
 *             self.right = right
 *             self.indices = indices             # <<<<<<<<<<<<<<
 *             self.n_center = 0
 *         else:
 */
    __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(((PyObject *)__pyx_v_indices), PyBUF_WRITABLE); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(1, 2652, __pyx_L1_error)
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->indices, 0);
    __pyx_v_self->indices = __pyx_t_7;
    __pyx_t_7.memview = NULL;
    __pyx_t_7.data = NULL;

    /* "pandas/_libs/intervaltree.pxi":2653
 *             self.right = right
 *             self.indices = indices
 *             self.n_center = 0             # <<<<<<<<<<<<<<
 *         else:
 *             # calculate a pivot so we can create child nodes
 */
    __pyx_v_self->n_center = 0;

    /* "pandas/_libs/intervaltree.pxi":2647
 *             self.max_right = 0
 * 
 *         if self.n_elements <= leaf_size:             # <<<<<<<<<<<<<<
 *             # make this a terminal (leaf) node
 *             self.is_leaf_node = True
 */
    goto __pyx_L4;
  }

  /* "pandas/_libs/intervaltree.pxi":2656
 *         else:
 *             # calculate a pivot so we can create child nodes
 *             self.is_leaf_node = False             # <<<<<<<<<<<<<<
 *             self.pivot = np.median(left / 2 + right / 2)
 *             left_set, right_set, center_set = self.classify_intervals(
 */
  /*else*/ {
    __pyx_v_self->is_leaf_node = 0;

    /* "pandas/_libs/intervaltree.pxi":2657
 *             # calculate a pivot so we can create child nodes
 *             self.is_leaf_node = False
 *             self.pivot = np.median(left / 2 + right / 2)             # <<<<<<<<<<<<<<
 *             left_set, right_set, center_set = self.classify_intervals(
 *                 left, right)
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2657, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_median); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 2657, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyNumber_Divide(((PyObject *)__pyx_v_left), __pyx_int_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2657, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_8 = __Pyx_PyNumber_Divide(((PyObject *)__pyx_v_right), __pyx_int_2); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 2657, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_9 = PyNumber_Add(__pyx_t_2, __pyx_t_8); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 2657, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_3 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_8, __pyx_t_9) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_9);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2657, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_6 = __Pyx_PyInt_As_npy_int64(__pyx_t_3); if (unlikely((__pyx_t_6 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(1, 2657, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_self->pivot = __pyx_t_6;

    /* "pandas/_libs/intervaltree.pxi":2659
 *             self.pivot = np.median(left / 2 + right / 2)
 *             left_set, right_set, center_set = self.classify_intervals(
 *                 left, right)             # <<<<<<<<<<<<<<
 * 
 *             self.left_node = self.new_child_node(left, right,
 */
    __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(((PyObject *)__pyx_v_left), PyBUF_WRITABLE); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(1, 2659, __pyx_L1_error)
    __pyx_t_10 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(((PyObject *)__pyx_v_right), PyBUF_WRITABLE); if (unlikely(!__pyx_t_10.memview)) __PYX_ERR(1, 2659, __pyx_L1_error)

    /* "pandas/_libs/intervaltree.pxi":2658
 *             self.is_leaf_node = False
 *             self.pivot = np.median(left / 2 + right / 2)
 *             left_set, right_set, center_set = self.classify_intervals(             # <<<<<<<<<<<<<<
 *                 left, right)
 * 
 */
    __pyx_t_3 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode *)__pyx_v_self->__pyx_vtab)->classify_intervals(__pyx_v_self, __pyx_t_7, __pyx_t_10); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2658, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __PYX_XDEC_MEMVIEW(&__pyx_t_7, 1);
    __pyx_t_7.memview = NULL;
    __pyx_t_7.data = NULL;
    __PYX_XDEC_MEMVIEW(&__pyx_t_10, 1);
    __pyx_t_10.memview = NULL;
    __pyx_t_10.data = NULL;
    if ((likely(PyTuple_CheckExact(__pyx_t_3))) || (PyList_CheckExact(__pyx_t_3))) {
      PyObject* sequence = __pyx_t_3;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 3)) {
        if (size > 3) __Pyx_RaiseTooManyValuesError(3);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(1, 2658, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_5 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_9 = PyTuple_GET_ITEM(sequence, 1); 
        __pyx_t_8 = PyTuple_GET_ITEM(sequence, 2); 
      } else {
        __pyx_t_5 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_9 = PyList_GET_ITEM(sequence, 1); 
        __pyx_t_8 = PyList_GET_ITEM(sequence, 2); 
      }
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_9);
      __Pyx_INCREF(__pyx_t_8);
      #else
      __pyx_t_5 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 2658, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_9 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 2658, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_8 = PySequence_ITEM(sequence, 2); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 2658, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      #endif
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_2 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2658, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_11 = Py_TYPE(__pyx_t_2)->tp_iternext;
      index = 0; __pyx_t_5 = __pyx_t_11(__pyx_t_2); if (unlikely(!__pyx_t_5)) goto __pyx_L5_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_5);
      index = 1; __pyx_t_9 = __pyx_t_11(__pyx_t_2); if (unlikely(!__pyx_t_9)) goto __pyx_L5_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_9);
      index = 2; __pyx_t_8 = __pyx_t_11(__pyx_t_2); if (unlikely(!__pyx_t_8)) goto __pyx_L5_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_8);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_11(__pyx_t_2), 3) < 0) __PYX_ERR(1, 2658, __pyx_L1_error)
      __pyx_t_11 = NULL;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      goto __pyx_L6_unpacking_done;
      __pyx_L5_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_11 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(1, 2658, __pyx_L1_error)
      __pyx_L6_unpacking_done:;
    }
    __pyx_v_left_set = __pyx_t_5;
    __pyx_t_5 = 0;
    __pyx_v_right_set = __pyx_t_9;
    __pyx_t_9 = 0;
    __pyx_v_center_set = __pyx_t_8;
    __pyx_t_8 = 0;

    /* "pandas/_libs/intervaltree.pxi":2662
 * 
 *             self.left_node = self.new_child_node(left, right,
 *                                                  indices, left_set)             # <<<<<<<<<<<<<<
 *             self.right_node = self.new_child_node(left, right,
 *                                                   indices, right_set)
 */
    if (!(likely(((__pyx_v_left_set) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_left_set, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(1, 2662, __pyx_L1_error)

    /* "pandas/_libs/intervaltree.pxi":2661
 *                 left, right)
 * 
 *             self.left_node = self.new_child_node(left, right,             # <<<<<<<<<<<<<<
 *                                                  indices, left_set)
 *             self.right_node = self.new_child_node(left, right,
 */
    __pyx_t_3 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode *)__pyx_v_self->__pyx_vtab)->new_child_node(__pyx_v_self, ((PyArrayObject *)__pyx_v_left), ((PyArrayObject *)__pyx_v_right), ((PyArrayObject *)__pyx_v_indices), ((PyArrayObject *)__pyx_v_left_set)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2661, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode))))) __PYX_ERR(1, 2661, __pyx_L1_error)
    __Pyx_GIVEREF(__pyx_t_3);
    __Pyx_GOTREF(__pyx_v_self->left_node);
    __Pyx_DECREF(((PyObject *)__pyx_v_self->left_node));
    __pyx_v_self->left_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode *)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "pandas/_libs/intervaltree.pxi":2664
 *                                                  indices, left_set)
 *             self.right_node = self.new_child_node(left, right,
 *                                                   indices, right_set)             # <<<<<<<<<<<<<<
 * 
 *             self.center_left_values, self.center_left_indices = \
 */
    if (!(likely(((__pyx_v_right_set) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_right_set, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(1, 2664, __pyx_L1_error)

    /* "pandas/_libs/intervaltree.pxi":2663
 *             self.left_node = self.new_child_node(left, right,
 *                                                  indices, left_set)
 *             self.right_node = self.new_child_node(left, right,             # <<<<<<<<<<<<<<
 *                                                   indices, right_set)
 * 
 */
    __pyx_t_3 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode *)__pyx_v_self->__pyx_vtab)->new_child_node(__pyx_v_self, ((PyArrayObject *)__pyx_v_left), ((PyArrayObject *)__pyx_v_right), ((PyArrayObject *)__pyx_v_indices), ((PyArrayObject *)__pyx_v_right_set)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2663, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode))))) __PYX_ERR(1, 2663, __pyx_L1_error)
    __Pyx_GIVEREF(__pyx_t_3);
    __Pyx_GOTREF(__pyx_v_self->right_node);
    __Pyx_DECREF(((PyObject *)__pyx_v_self->right_node));
    __pyx_v_self->right_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode *)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "pandas/_libs/intervaltree.pxi":2667
 * 
 *             self.center_left_values, self.center_left_indices = \
 *                 sort_values_and_indices(left, indices, center_set)             # <<<<<<<<<<<<<<
 *             self.center_right_values, self.center_right_indices = \
 *                 sort_values_and_indices(right, indices, center_set)
 */
    __pyx_t_3 = __pyx_f_6pandas_5_libs_8interval_sort_values_and_indices(((PyObject *)__pyx_v_left), ((PyObject *)__pyx_v_indices), __pyx_v_center_set); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2667, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if ((likely(PyTuple_CheckExact(__pyx_t_3))) || (PyList_CheckExact(__pyx_t_3))) {
      PyObject* sequence = __pyx_t_3;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(1, 2666, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_8 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_9 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_8 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_9 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_8);
      __Pyx_INCREF(__pyx_t_9);
      #else
      __pyx_t_8 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 2666, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_9 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 2666, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      #endif
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_5 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 2666, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_11 = Py_TYPE(__pyx_t_5)->tp_iternext;
      index = 0; __pyx_t_8 = __pyx_t_11(__pyx_t_5); if (unlikely(!__pyx_t_8)) goto __pyx_L7_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_8);
      index = 1; __pyx_t_9 = __pyx_t_11(__pyx_t_5); if (unlikely(!__pyx_t_9)) goto __pyx_L7_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_9);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_11(__pyx_t_5), 2) < 0) __PYX_ERR(1, 2666, __pyx_L1_error)
      __pyx_t_11 = NULL;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      goto __pyx_L8_unpacking_done;
      __pyx_L7_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_11 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(1, 2666, __pyx_L1_error)
      __pyx_L8_unpacking_done:;
    }

    /* "pandas/_libs/intervaltree.pxi":2666
 *                                                   indices, right_set)
 * 
 *             self.center_left_values, self.center_left_indices = \             # <<<<<<<<<<<<<<
 *                 sort_values_and_indices(left, indices, center_set)
 *             self.center_right_values, self.center_right_indices = \
 */
    __pyx_t_10 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_t_8, PyBUF_WRITABLE); if (unlikely(!__pyx_t_10.memview)) __PYX_ERR(1, 2666, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_t_9, PyBUF_WRITABLE); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(1, 2666, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->center_left_values, 0);
    __pyx_v_self->center_left_values = __pyx_t_10;
    __pyx_t_10.memview = NULL;
    __pyx_t_10.data = NULL;
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->center_left_indices, 0);
    __pyx_v_self->center_left_indices = __pyx_t_7;
    __pyx_t_7.memview = NULL;
    __pyx_t_7.data = NULL;

    /* "pandas/_libs/intervaltree.pxi":2669
 *                 sort_values_and_indices(left, indices, center_set)
 *             self.center_right_values, self.center_right_indices = \
 *                 sort_values_and_indices(right, indices, center_set)             # <<<<<<<<<<<<<<
 *             self.n_center = len(self.center_left_indices)
 * 
 */
    __pyx_t_3 = __pyx_f_6pandas_5_libs_8interval_sort_values_and_indices(((PyObject *)__pyx_v_right), ((PyObject *)__pyx_v_indices), __pyx_v_center_set); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2669, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if ((likely(PyTuple_CheckExact(__pyx_t_3))) || (PyList_CheckExact(__pyx_t_3))) {
      PyObject* sequence = __pyx_t_3;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(1, 2668, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_9 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_8 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_9 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_8 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_9);
      __Pyx_INCREF(__pyx_t_8);
      #else
      __pyx_t_9 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 2668, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_8 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 2668, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      #endif
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_5 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 2668, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_11 = Py_TYPE(__pyx_t_5)->tp_iternext;
      index = 0; __pyx_t_9 = __pyx_t_11(__pyx_t_5); if (unlikely(!__pyx_t_9)) goto __pyx_L9_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_9);
      index = 1; __pyx_t_8 = __pyx_t_11(__pyx_t_5); if (unlikely(!__pyx_t_8)) goto __pyx_L9_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_8);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_11(__pyx_t_5), 2) < 0) __PYX_ERR(1, 2668, __pyx_L1_error)
      __pyx_t_11 = NULL;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      goto __pyx_L10_unpacking_done;
      __pyx_L9_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_11 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(1, 2668, __pyx_L1_error)
      __pyx_L10_unpacking_done:;
    }

    /* "pandas/_libs/intervaltree.pxi":2668
 *             self.center_left_values, self.center_left_indices = \
 *                 sort_values_and_indices(left, indices, center_set)
 *             self.center_right_values, self.center_right_indices = \             # <<<<<<<<<<<<<<
 *                 sort_values_and_indices(right, indices, center_set)
 *             self.n_center = len(self.center_left_indices)
 */
    __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_t_9, PyBUF_WRITABLE); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(1, 2668, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_10 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_t_8, PyBUF_WRITABLE); if (unlikely(!__pyx_t_10.memview)) __PYX_ERR(1, 2668, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->center_right_values, 0);
    __pyx_v_self->center_right_values = __pyx_t_7;
    __pyx_t_7.memview = NULL;
    __pyx_t_7.data = NULL;
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->center_right_indices, 0);
    __pyx_v_self->center_right_indices = __pyx_t_10;
    __pyx_t_10.memview = NULL;
    __pyx_t_10.data = NULL;

    /* "pandas/_libs/intervaltree.pxi":2670
 *             self.center_right_values, self.center_right_indices = \
 *                 sort_values_and_indices(right, indices, center_set)
 *             self.n_center = len(self.center_left_indices)             # <<<<<<<<<<<<<<
 * 
 *     @cython.wraparound(False)
 */
    if (unlikely(!__pyx_v_self->center_left_indices.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(1, 2670, __pyx_L1_error)}
    __pyx_t_12 = __Pyx_MemoryView_Len(__pyx_v_self->center_left_indices); 
    __pyx_v_self->n_center = __pyx_t_12;
  }
  __pyx_L4:;

  /* "pandas/_libs/intervaltree.pxi":2628
 *         readonly bint is_leaf_node
 * 
 *     def __init__(self,             # <<<<<<<<<<<<<<
 *                  ndarray[int64_t, ndim=1] left,
 *                  ndarray[int64_t, ndim=1] right,
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __PYX_XDEC_MEMVIEW(&__pyx_t_7, 1);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __PYX_XDEC_MEMVIEW(&__pyx_t_10, 1);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indices.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_left.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_right.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("pandas._libs.interval.Int64ClosedBothIntervalNode.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indices.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_left.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_right.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF(__pyx_v_left_set);
  __Pyx_XDECREF(__pyx_v_right_set);
  __Pyx_XDECREF(__pyx_v_center_set);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":2674
 *     @cython.wraparound(False)
 *     @cython.boundscheck(False)
 *     cdef classify_intervals(self, int64_t[:] left, int64_t[:] right):             # <<<<<<<<<<<<<<
 *         """Classify the given intervals based upon whether they fall to the
 *         left, right, or overlap with this node's pivot.
 */

static PyObject *__pyx_f_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_classify_intervals(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode *__pyx_v_self, __Pyx_memviewslice __pyx_v_left, __Pyx_memviewslice __pyx_v_right) {
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_left_ind = 0;
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_right_ind = 0;
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_overlapping_ind = 0;
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __pyx_t_5numpy_int64_t __pyx_t_2;
  __pyx_t_5numpy_int64_t __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  int __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  __Pyx_RefNannySetupContext("classify_intervals", 0);

  /* "pandas/_libs/intervaltree.pxi":2682
 *             Py_ssize_t i
 * 
 *         left_ind = Int64Vector()             # <<<<<<<<<<<<<<
 *         right_ind = Int64Vector()
 *         overlapping_ind = Int64Vector()
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2682, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_left_ind = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":2683
 * 
 *         left_ind = Int64Vector()
 *         right_ind = Int64Vector()             # <<<<<<<<<<<<<<
 *         overlapping_ind = Int64Vector()
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2683, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_right_ind = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":2684
 *         left_ind = Int64Vector()
 *         right_ind = Int64Vector()
 *         overlapping_ind = Int64Vector()             # <<<<<<<<<<<<<<
 * 
 *         for i in range(self.n_elements):
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2684, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_overlapping_ind = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":2686
 *         overlapping_ind = Int64Vector()
 * 
 *         for i in range(self.n_elements):             # <<<<<<<<<<<<<<
 *             if right[i] < self.pivot:
 *                 left_ind.append(i)
 */
  __pyx_t_2 = __pyx_v_self->n_elements;
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "pandas/_libs/intervaltree.pxi":2687
 * 
 *         for i in range(self.n_elements):
 *             if right[i] < self.pivot:             # <<<<<<<<<<<<<<
 *                 left_ind.append(i)
 *             elif self.pivot < left[i]:
 */
    __pyx_t_5 = __pyx_v_i;
    __pyx_t_6 = (((*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_right.data + __pyx_t_5 * __pyx_v_right.strides[0]) ))) < __pyx_v_self->pivot) != 0);
    if (__pyx_t_6) {

      /* "pandas/_libs/intervaltree.pxi":2688
 *         for i in range(self.n_elements):
 *             if right[i] < self.pivot:
 *                 left_ind.append(i)             # <<<<<<<<<<<<<<
 *             elif self.pivot < left[i]:
 *                 right_ind.append(i)
 */
      ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_left_ind->__pyx_vtab)->append(__pyx_v_left_ind, __pyx_v_i);

      /* "pandas/_libs/intervaltree.pxi":2687
 * 
 *         for i in range(self.n_elements):
 *             if right[i] < self.pivot:             # <<<<<<<<<<<<<<
 *                 left_ind.append(i)
 *             elif self.pivot < left[i]:
 */
      goto __pyx_L5;
    }

    /* "pandas/_libs/intervaltree.pxi":2689
 *             if right[i] < self.pivot:
 *                 left_ind.append(i)
 *             elif self.pivot < left[i]:             # <<<<<<<<<<<<<<
 *                 right_ind.append(i)
 *             else:
 */
    __pyx_t_7 = __pyx_v_i;
    __pyx_t_6 = ((__pyx_v_self->pivot < (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_left.data + __pyx_t_7 * __pyx_v_left.strides[0]) )))) != 0);
    if (__pyx_t_6) {

      /* "pandas/_libs/intervaltree.pxi":2690
 *                 left_ind.append(i)
 *             elif self.pivot < left[i]:
 *                 right_ind.append(i)             # <<<<<<<<<<<<<<
 *             else:
 *                 overlapping_ind.append(i)
 */
      ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_right_ind->__pyx_vtab)->append(__pyx_v_right_ind, __pyx_v_i);

      /* "pandas/_libs/intervaltree.pxi":2689
 *             if right[i] < self.pivot:
 *                 left_ind.append(i)
 *             elif self.pivot < left[i]:             # <<<<<<<<<<<<<<
 *                 right_ind.append(i)
 *             else:
 */
      goto __pyx_L5;
    }

    /* "pandas/_libs/intervaltree.pxi":2692
 *                 right_ind.append(i)
 *             else:
 *                 overlapping_ind.append(i)             # <<<<<<<<<<<<<<
 * 
 *         return (left_ind.to_array(),
 */
    /*else*/ {
      ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_overlapping_ind->__pyx_vtab)->append(__pyx_v_overlapping_ind, __pyx_v_i);
    }
    __pyx_L5:;
  }

  /* "pandas/_libs/intervaltree.pxi":2694
 *                 overlapping_ind.append(i)
 * 
 *         return (left_ind.to_array(),             # <<<<<<<<<<<<<<
 *                 right_ind.to_array(),
 *                 overlapping_ind.to_array())
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_left_ind->__pyx_vtab)->to_array(__pyx_v_left_ind, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2694, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "pandas/_libs/intervaltree.pxi":2695
 * 
 *         return (left_ind.to_array(),
 *                 right_ind.to_array(),             # <<<<<<<<<<<<<<
 *                 overlapping_ind.to_array())
 * 
 */
  __pyx_t_8 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_right_ind->__pyx_vtab)->to_array(__pyx_v_right_ind, 0); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 2695, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);

  /* "pandas/_libs/intervaltree.pxi":2696
 *         return (left_ind.to_array(),
 *                 right_ind.to_array(),
 *                 overlapping_ind.to_array())             # <<<<<<<<<<<<<<
 * 
 *     cdef new_child_node(self,
 */
  __pyx_t_9 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_overlapping_ind->__pyx_vtab)->to_array(__pyx_v_overlapping_ind, 0); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 2696, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);

  /* "pandas/_libs/intervaltree.pxi":2694
 *                 overlapping_ind.append(i)
 * 
 *         return (left_ind.to_array(),             # <<<<<<<<<<<<<<
 *                 right_ind.to_array(),
 *                 overlapping_ind.to_array())
 */
  __pyx_t_10 = PyTuple_New(3); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 2694, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_t_8);
  __Pyx_GIVEREF(__pyx_t_9);
  PyTuple_SET_ITEM(__pyx_t_10, 2, __pyx_t_9);
  __pyx_t_1 = 0;
  __pyx_t_8 = 0;
  __pyx_t_9 = 0;
  __pyx_r = __pyx_t_10;
  __pyx_t_10 = 0;
  goto __pyx_L0;

  /* "pandas/_libs/intervaltree.pxi":2674
 *     @cython.wraparound(False)
 *     @cython.boundscheck(False)
 *     cdef classify_intervals(self, int64_t[:] left, int64_t[:] right):             # <<<<<<<<<<<<<<
 *         """Classify the given intervals based upon whether they fall to the
 *         left, right, or overlap with this node's pivot.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("pandas._libs.interval.Int64ClosedBothIntervalNode.classify_intervals", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_left_ind);
  __Pyx_XDECREF((PyObject *)__pyx_v_right_ind);
  __Pyx_XDECREF((PyObject *)__pyx_v_overlapping_ind);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":2698
 *                 overlapping_ind.to_array())
 * 
 *     cdef new_child_node(self,             # <<<<<<<<<<<<<<
 *                         ndarray[int64_t, ndim=1] left,
 *                         ndarray[int64_t, ndim=1] right,
 */

static PyObject *__pyx_f_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_new_child_node(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode *__pyx_v_self, PyArrayObject *__pyx_v_left, PyArrayObject *__pyx_v_right, PyArrayObject *__pyx_v_indices, PyArrayObject *__pyx_v_subset) {
  __Pyx_LocalBuf_ND __pyx_pybuffernd_indices;
  __Pyx_Buffer __pyx_pybuffer_indices;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_left;
  __Pyx_Buffer __pyx_pybuffer_left;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_right;
  __Pyx_Buffer __pyx_pybuffer_right;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_subset;
  __Pyx_Buffer __pyx_pybuffer_subset;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyArrayObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyArrayObject *__pyx_t_7 = NULL;
  PyArrayObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  __Pyx_RefNannySetupContext("new_child_node", 0);
  __Pyx_INCREF((PyObject *)__pyx_v_left);
  __Pyx_INCREF((PyObject *)__pyx_v_right);
  __Pyx_INCREF((PyObject *)__pyx_v_indices);
  __pyx_pybuffer_left.pybuffer.buf = NULL;
  __pyx_pybuffer_left.refcount = 0;
  __pyx_pybuffernd_left.data = NULL;
  __pyx_pybuffernd_left.rcbuffer = &__pyx_pybuffer_left;
  __pyx_pybuffer_right.pybuffer.buf = NULL;
  __pyx_pybuffer_right.refcount = 0;
  __pyx_pybuffernd_right.data = NULL;
  __pyx_pybuffernd_right.rcbuffer = &__pyx_pybuffer_right;
  __pyx_pybuffer_indices.pybuffer.buf = NULL;
  __pyx_pybuffer_indices.refcount = 0;
  __pyx_pybuffernd_indices.data = NULL;
  __pyx_pybuffernd_indices.rcbuffer = &__pyx_pybuffer_indices;
  __pyx_pybuffer_subset.pybuffer.buf = NULL;
  __pyx_pybuffer_subset.refcount = 0;
  __pyx_pybuffernd_subset.data = NULL;
  __pyx_pybuffernd_subset.rcbuffer = &__pyx_pybuffer_subset;
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_left.rcbuffer->pybuffer, (PyObject*)__pyx_v_left, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 2698, __pyx_L1_error)
  }
  __pyx_pybuffernd_left.diminfo[0].strides = __pyx_pybuffernd_left.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_left.diminfo[0].shape = __pyx_pybuffernd_left.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_right.rcbuffer->pybuffer, (PyObject*)__pyx_v_right, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 2698, __pyx_L1_error)
  }
  __pyx_pybuffernd_right.diminfo[0].strides = __pyx_pybuffernd_right.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_right.diminfo[0].shape = __pyx_pybuffernd_right.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_indices.rcbuffer->pybuffer, (PyObject*)__pyx_v_indices, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 2698, __pyx_L1_error)
  }
  __pyx_pybuffernd_indices.diminfo[0].strides = __pyx_pybuffernd_indices.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_indices.diminfo[0].shape = __pyx_pybuffernd_indices.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_subset.rcbuffer->pybuffer, (PyObject*)__pyx_v_subset, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 2698, __pyx_L1_error)
  }
  __pyx_pybuffernd_subset.diminfo[0].strides = __pyx_pybuffernd_subset.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_subset.diminfo[0].shape = __pyx_pybuffernd_subset.rcbuffer->pybuffer.shape[0];

  /* "pandas/_libs/intervaltree.pxi":2705
 *         """Create a new child node.
 *         """
 *         left = take(left, subset)             # <<<<<<<<<<<<<<
 *         right = take(right, subset)
 *         indices = take(indices, subset)
 */
  __pyx_t_1 = __pyx_f_6pandas_5_libs_8interval_take(((PyArrayObject *)__pyx_v_left), ((PyArrayObject *)__pyx_v_subset)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2705, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(1, 2705, __pyx_L1_error)
  __pyx_t_2 = ((PyArrayObject *)__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_left.rcbuffer->pybuffer);
    __pyx_t_3 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_left.rcbuffer->pybuffer, (PyObject*)__pyx_t_2, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack);
    if (unlikely(__pyx_t_3 < 0)) {
      PyErr_Fetch(&__pyx_t_4, &__pyx_t_5, &__pyx_t_6);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_left.rcbuffer->pybuffer, (PyObject*)__pyx_v_left, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_4); Py_XDECREF(__pyx_t_5); Py_XDECREF(__pyx_t_6);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_4, __pyx_t_5, __pyx_t_6);
      }
      __pyx_t_4 = __pyx_t_5 = __pyx_t_6 = 0;
    }
    __pyx_pybuffernd_left.diminfo[0].strides = __pyx_pybuffernd_left.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_left.diminfo[0].shape = __pyx_pybuffernd_left.rcbuffer->pybuffer.shape[0];
    if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(1, 2705, __pyx_L1_error)
  }
  __pyx_t_2 = 0;
  __Pyx_DECREF_SET(__pyx_v_left, ((PyArrayObject *)__pyx_t_1));
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":2706
 *         """
 *         left = take(left, subset)
 *         right = take(right, subset)             # <<<<<<<<<<<<<<
 *         indices = take(indices, subset)
 *         return Int64ClosedBothIntervalNode(
 */
  __pyx_t_1 = __pyx_f_6pandas_5_libs_8interval_take(((PyArrayObject *)__pyx_v_right), ((PyArrayObject *)__pyx_v_subset)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2706, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(1, 2706, __pyx_L1_error)
  __pyx_t_7 = ((PyArrayObject *)__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_right.rcbuffer->pybuffer);
    __pyx_t_3 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_right.rcbuffer->pybuffer, (PyObject*)__pyx_t_7, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack);
    if (unlikely(__pyx_t_3 < 0)) {
      PyErr_Fetch(&__pyx_t_6, &__pyx_t_5, &__pyx_t_4);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_right.rcbuffer->pybuffer, (PyObject*)__pyx_v_right, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_6); Py_XDECREF(__pyx_t_5); Py_XDECREF(__pyx_t_4);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_6, __pyx_t_5, __pyx_t_4);
      }
      __pyx_t_6 = __pyx_t_5 = __pyx_t_4 = 0;
    }
    __pyx_pybuffernd_right.diminfo[0].strides = __pyx_pybuffernd_right.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_right.diminfo[0].shape = __pyx_pybuffernd_right.rcbuffer->pybuffer.shape[0];
    if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(1, 2706, __pyx_L1_error)
  }
  __pyx_t_7 = 0;
  __Pyx_DECREF_SET(__pyx_v_right, ((PyArrayObject *)__pyx_t_1));
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":2707
 *         left = take(left, subset)
 *         right = take(right, subset)
 *         indices = take(indices, subset)             # <<<<<<<<<<<<<<
 *         return Int64ClosedBothIntervalNode(
 *             left, right, indices, self.leaf_size)
 */
  __pyx_t_1 = __pyx_f_6pandas_5_libs_8interval_take(((PyArrayObject *)__pyx_v_indices), ((PyArrayObject *)__pyx_v_subset)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2707, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(1, 2707, __pyx_L1_error)
  __pyx_t_8 = ((PyArrayObject *)__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indices.rcbuffer->pybuffer);
    __pyx_t_3 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_indices.rcbuffer->pybuffer, (PyObject*)__pyx_t_8, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack);
    if (unlikely(__pyx_t_3 < 0)) {
      PyErr_Fetch(&__pyx_t_4, &__pyx_t_5, &__pyx_t_6);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_indices.rcbuffer->pybuffer, (PyObject*)__pyx_v_indices, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_4); Py_XDECREF(__pyx_t_5); Py_XDECREF(__pyx_t_6);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_4, __pyx_t_5, __pyx_t_6);
      }
      __pyx_t_4 = __pyx_t_5 = __pyx_t_6 = 0;
    }
    __pyx_pybuffernd_indices.diminfo[0].strides = __pyx_pybuffernd_indices.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_indices.diminfo[0].shape = __pyx_pybuffernd_indices.rcbuffer->pybuffer.shape[0];
    if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(1, 2707, __pyx_L1_error)
  }
  __pyx_t_8 = 0;
  __Pyx_DECREF_SET(__pyx_v_indices, ((PyArrayObject *)__pyx_t_1));
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":2708
 *         right = take(right, subset)
 *         indices = take(indices, subset)
 *         return Int64ClosedBothIntervalNode(             # <<<<<<<<<<<<<<
 *             left, right, indices, self.leaf_size)
 * 
 */
  __Pyx_XDECREF(__pyx_r);

  /* "pandas/_libs/intervaltree.pxi":2709
 *         indices = take(indices, subset)
 *         return Int64ClosedBothIntervalNode(
 *             left, right, indices, self.leaf_size)             # <<<<<<<<<<<<<<
 * 
 *     @cython.wraparound(False)
 */
  __pyx_t_1 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->leaf_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2709, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "pandas/_libs/intervaltree.pxi":2708
 *         right = take(right, subset)
 *         indices = take(indices, subset)
 *         return Int64ClosedBothIntervalNode(             # <<<<<<<<<<<<<<
 *             left, right, indices, self.leaf_size)
 * 
 */
  __pyx_t_9 = PyTuple_New(4); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 2708, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_INCREF(((PyObject *)__pyx_v_left));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_left));
  PyTuple_SET_ITEM(__pyx_t_9, 0, ((PyObject *)__pyx_v_left));
  __Pyx_INCREF(((PyObject *)__pyx_v_right));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_right));
  PyTuple_SET_ITEM(__pyx_t_9, 1, ((PyObject *)__pyx_v_right));
  __Pyx_INCREF(((PyObject *)__pyx_v_indices));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_indices));
  PyTuple_SET_ITEM(__pyx_t_9, 2, ((PyObject *)__pyx_v_indices));
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_9, 3, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode), __pyx_t_9, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2708, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pandas/_libs/intervaltree.pxi":2698
 *                 overlapping_ind.to_array())
 * 
 *     cdef new_child_node(self,             # <<<<<<<<<<<<<<
 *                         ndarray[int64_t, ndim=1] left,
 *                         ndarray[int64_t, ndim=1] right,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_9);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indices.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_left.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_right.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_subset.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("pandas._libs.interval.Int64ClosedBothIntervalNode.new_child_node", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indices.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_left.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_right.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_subset.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF((PyObject *)__pyx_v_left);
  __Pyx_XDECREF((PyObject *)__pyx_v_right);
  __Pyx_XDECREF((PyObject *)__pyx_v_indices);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":2714
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_2query[] = "Recursively query this node and its sub-nodes for intervals that\n        overlap with the query point.\n        ";
static PyMethodDef __pyx_mdef_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_3query = {"query", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_3query, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_2query};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_args = 0;
  PyObject *__pyx_v_kwargs = 0;
  CYTHON_UNUSED PyObject *__pyx_v_defaults = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fused_cpdef (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_args,&__pyx_n_s_kwargs,&__pyx_n_s_defaults,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_args)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_kwargs)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fused_cpdef", 1, 3, 3, 1); __PYX_ERR(1, 2714, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_defaults)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fused_cpdef", 1, 3, 3, 2); __PYX_ERR(1, 2714, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fused_cpdef") < 0)) __PYX_ERR(1, 2714, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_args = values[0];
    __pyx_v_kwargs = values[1];
    __pyx_v_defaults = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fused_cpdef", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 2714, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Int64ClosedBothIntervalNode.__pyx_fused_cpdef", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_2query(((struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode *)__pyx_v_signatures), __pyx_v_args, __pyx_v_kwargs, __pyx_v_defaults);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_2query(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode *__pyx_v_signatures, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs, CYTHON_UNUSED PyObject *__pyx_v_defaults) {
  PyObject *__pyx_v_dest_sig = NULL;
  Py_ssize_t __pyx_v_i;
  CYTHON_UNUSED PyTypeObject *__pyx_v_ndarray = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_v_itemsize;
  PyObject *__pyx_v_arg = NULL;
  PyObject *__pyx_v_candidates = NULL;
  PyObject *__pyx_v_sig = NULL;
  int __pyx_v_match_found;
  PyObject *__pyx_v_src_sig = NULL;
  PyObject *__pyx_v_dst_type = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  Py_ssize_t __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  Py_ssize_t __pyx_t_13;
  Py_ssize_t __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  int __pyx_t_16;
  __Pyx_RefNannySetupContext("query", 0);
  __Pyx_INCREF(__pyx_v_kwargs);
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2714, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  PyList_SET_ITEM(__pyx_t_1, 0, Py_None);
  __pyx_v_dest_sig = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_3 = (__pyx_v_kwargs != Py_None);
  __pyx_t_4 = (__pyx_t_3 != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_kwargs); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(1, 2714, __pyx_L1_error)
  __pyx_t_3 = ((!__pyx_t_4) != 0);
  __pyx_t_2 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {
    __Pyx_INCREF(Py_None);
    __Pyx_DECREF_SET(__pyx_v_kwargs, Py_None);
  }
  __pyx_t_1 = ((PyObject *)__Pyx_ImportNumPyArrayTypeIfAvailable()); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2714, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_ndarray = ((PyTypeObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_v_itemsize = -1L;
  if (unlikely(__pyx_v_args == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 2714, __pyx_L1_error)
  }
  __pyx_t_5 = PyTuple_GET_SIZE(((PyObject*)__pyx_v_args)); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(1, 2714, __pyx_L1_error)
  __pyx_t_2 = ((2 < __pyx_t_5) != 0);
  if (__pyx_t_2) {
    if (unlikely(__pyx_v_args == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 2714, __pyx_L1_error)
    }
    __pyx_t_1 = PyTuple_GET_ITEM(((PyObject*)__pyx_v_args), 2);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v_arg = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L6;
  }
  __pyx_t_3 = (__pyx_v_kwargs != Py_None);
  __pyx_t_4 = (__pyx_t_3 != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L7_bool_binop_done;
  }
  if (unlikely(__pyx_v_kwargs == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(1, 2714, __pyx_L1_error)
  }
  __pyx_t_4 = (__Pyx_PyDict_ContainsTF(__pyx_n_s_point, ((PyObject*)__pyx_v_kwargs), Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(1, 2714, __pyx_L1_error)
  __pyx_t_3 = (__pyx_t_4 != 0);
  __pyx_t_2 = __pyx_t_3;
  __pyx_L7_bool_binop_done:;
  if (__pyx_t_2) {
    if (unlikely(__pyx_v_kwargs == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 2714, __pyx_L1_error)
    }
    __pyx_t_1 = __Pyx_PyDict_GetItem(((PyObject*)__pyx_v_kwargs), __pyx_n_s_point); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2714, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_arg = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L6;
  }
  /*else*/ {
    if (unlikely(__pyx_v_args == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
      __PYX_ERR(1, 2714, __pyx_L1_error)
    }
    __pyx_t_5 = PyTuple_GET_SIZE(((PyObject*)__pyx_v_args)); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(1, 2714, __pyx_L1_error)
    __pyx_t_1 = PyInt_FromSsize_t(__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2714, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_6 = PyTuple_New(3); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 2714, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_INCREF(__pyx_int_3);
    __Pyx_GIVEREF(__pyx_int_3);
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_int_3);
    __Pyx_INCREF(__pyx_n_s_s);
    __Pyx_GIVEREF(__pyx_n_s_s);
    PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_n_s_s);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Expected_at_least_d_argument_s_g, __pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2714, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_1); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 2714, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(1, 2714, __pyx_L1_error)
  }
  __pyx_L6:;
  while (1) {
    __pyx_t_2 = PyFloat_Check(__pyx_v_arg); 
    __pyx_t_3 = (__pyx_t_2 != 0);
    if (__pyx_t_3) {
      if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, __pyx_n_s_float64_t, long, 1, __Pyx_PyInt_From_long, 1, 0, 0) < 0)) __PYX_ERR(1, 2714, __pyx_L1_error)
      goto __pyx_L10_break;
    }
    __pyx_t_2 = PyInt_Check(__pyx_v_arg); 
    __pyx_t_4 = (__pyx_t_2 != 0);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_3 = __pyx_t_4;
      goto __pyx_L13_bool_binop_done;
    }
    __pyx_t_4 = PyLong_Check(__pyx_v_arg); 
    __pyx_t_2 = (__pyx_t_4 != 0);
    __pyx_t_3 = __pyx_t_2;
    __pyx_L13_bool_binop_done:;
    __pyx_t_2 = (__pyx_t_3 != 0);
    if (__pyx_t_2) {
      if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, __pyx_n_s_int64_t, long, 1, __Pyx_PyInt_From_long, 1, 0, 0) < 0)) __PYX_ERR(1, 2714, __pyx_L1_error)
      goto __pyx_L10_break;
    }
    if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, Py_None, long, 1, __Pyx_PyInt_From_long, 1, 0, 0) < 0)) __PYX_ERR(1, 2714, __pyx_L1_error)
    goto __pyx_L10_break;
  }
  __pyx_L10_break:;
  __pyx_t_6 = PyList_New(0); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 2714, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_v_candidates = ((PyObject*)__pyx_t_6);
  __pyx_t_6 = 0;
  __pyx_t_5 = 0;
  if (unlikely(((PyObject *)__pyx_v_signatures) == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(1, 2714, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_dict_iterator(((PyObject*)__pyx_v_signatures), 1, ((PyObject *)NULL), (&__pyx_t_7), (&__pyx_t_8)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2714, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __pyx_t_6 = __pyx_t_1;
  __pyx_t_1 = 0;
  while (1) {
    __pyx_t_9 = __Pyx_dict_iter_next(__pyx_t_6, __pyx_t_7, &__pyx_t_5, &__pyx_t_1, NULL, NULL, __pyx_t_8);
    if (unlikely(__pyx_t_9 == 0)) break;
    if (unlikely(__pyx_t_9 == -1)) __PYX_ERR(1, 2714, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_sig, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_v_match_found = 0;
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_v_sig, __pyx_n_s_strip); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 2714, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __pyx_t_12 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_11))) {
      __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_11);
      if (likely(__pyx_t_12)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
        __Pyx_INCREF(__pyx_t_12);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_11, function);
      }
    }
    __pyx_t_10 = (__pyx_t_12) ? __Pyx_PyObject_Call2Args(__pyx_t_11, __pyx_t_12, __pyx_kp_s__13) : __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_kp_s__13);
    __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
    if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 2714, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_split); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 2714, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_t_10 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_11))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_11);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_11, function);
      }
    }
    __pyx_t_1 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_11, __pyx_t_10, __pyx_kp_s__14) : __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_kp_s__14);
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2714, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_XDECREF_SET(__pyx_v_src_sig, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_13 = PyList_GET_SIZE(__pyx_v_dest_sig); if (unlikely(__pyx_t_13 == ((Py_ssize_t)-1))) __PYX_ERR(1, 2714, __pyx_L1_error)
    __pyx_t_14 = __pyx_t_13;
    for (__pyx_t_15 = 0; __pyx_t_15 < __pyx_t_14; __pyx_t_15+=1) {
      __pyx_v_i = __pyx_t_15;
      __pyx_t_1 = PyList_GET_ITEM(__pyx_v_dest_sig, __pyx_v_i);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_XDECREF_SET(__pyx_v_dst_type, __pyx_t_1);
      __pyx_t_1 = 0;
      __pyx_t_2 = (__pyx_v_dst_type != Py_None);
      __pyx_t_3 = (__pyx_t_2 != 0);
      if (__pyx_t_3) {
        __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_src_sig, __pyx_v_i, Py_ssize_t, 1, PyInt_FromSsize_t, 0, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2714, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_11 = PyObject_RichCompare(__pyx_t_1, __pyx_v_dst_type, Py_EQ); __Pyx_XGOTREF(__pyx_t_11); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 2714, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_11); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(1, 2714, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        if (__pyx_t_3) {
          __pyx_v_match_found = 1;
          goto __pyx_L20;
        }
        /*else*/ {
          __pyx_v_match_found = 0;
          goto __pyx_L18_break;
        }
        __pyx_L20:;
      }
    }
    __pyx_L18_break:;
    __pyx_t_3 = (__pyx_v_match_found != 0);
    if (__pyx_t_3) {
      __pyx_t_16 = __Pyx_PyList_Append(__pyx_v_candidates, __pyx_v_sig); if (unlikely(__pyx_t_16 == ((int)-1))) __PYX_ERR(1, 2714, __pyx_L1_error)
    }
  }
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_3 = (PyList_GET_SIZE(__pyx_v_candidates) != 0);
  __pyx_t_2 = ((!__pyx_t_3) != 0);
  if (__pyx_t_2) {
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__15, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 2714, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(1, 2714, __pyx_L1_error)
  }
  __pyx_t_7 = PyList_GET_SIZE(__pyx_v_candidates); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1))) __PYX_ERR(1, 2714, __pyx_L1_error)
  __pyx_t_2 = ((__pyx_t_7 > 1) != 0);
  if (__pyx_t_2) {
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__16, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 2714, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(1, 2714, __pyx_L1_error)
  }
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    if (unlikely(((PyObject *)__pyx_v_signatures) == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 2714, __pyx_L1_error)
    }
    __pyx_t_6 = __Pyx_PyDict_GetItem(((PyObject*)__pyx_v_signatures), PyList_GET_ITEM(__pyx_v_candidates, 0)); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 2714, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_r = __pyx_t_6;
    __pyx_t_6 = 0;
    goto __pyx_L0;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("pandas._libs.interval.Int64ClosedBothIntervalNode.__pyx_fused_cpdef", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_dest_sig);
  __Pyx_XDECREF(__pyx_v_ndarray);
  __Pyx_XDECREF(__pyx_v_arg);
  __Pyx_XDECREF(__pyx_v_candidates);
  __Pyx_XDECREF(__pyx_v_sig);
  __Pyx_XDECREF(__pyx_v_src_sig);
  __Pyx_XDECREF(__pyx_v_dst_type);
  __Pyx_XDECREF(__pyx_v_kwargs);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_9__pyx_fuse_0query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_fuse_0__pyx_f_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float64_t __pyx_v_point, int __pyx_skip_dispatch) {
  __Pyx_memviewslice __pyx_v_indices = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_values = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __pyx_t_5numpy_int64_t __pyx_t_9;
  __pyx_t_5numpy_int64_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  int __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  __Pyx_memviewslice __pyx_t_16 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_17;
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  __Pyx_RefNannySetupContext("__pyx_fuse_0query", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pyx_fuse_0query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2714, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_9__pyx_fuse_0query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = PyFloat_FromDouble(__pyx_v_point); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2714, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2714, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2714, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 2714, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_5) {
            __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2714, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2714, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_3query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_4 = PyFloat_FromDouble(__pyx_v_point); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 2714, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_7 = __pyx_t_1; __pyx_t_3 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2714, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2714, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 2714, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (__pyx_t_3) {
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_t_4);
          __pyx_t_4 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2714, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pandas/_libs/intervaltree.pxi":2723
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
  __pyx_t_8 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_8) {

    /* "pandas/_libs/intervaltree.pxi":2727
 *             # continue the binary tree structure. Instead, we use linear
 *             # search.
 *             for i in range(self.n_elements):             # <<<<<<<<<<<<<<
 *                 if self.left[i] <= point <= self.right[i]:
 *                     result.append(self.indices[i])
 */
    __pyx_t_9 = __pyx_v_self->n_elements;
    __pyx_t_10 = __pyx_t_9;
    for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
      __pyx_v_i = __pyx_t_11;

      /* "pandas/_libs/intervaltree.pxi":2728
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point <= self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      __pyx_t_12 = __pyx_v_i;
      __pyx_t_8 = ((*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->left.data + __pyx_t_12 * __pyx_v_self->left.strides[0]) ))) <= __pyx_v_point);
      if (__pyx_t_8) {
        __pyx_t_13 = __pyx_v_i;
        __pyx_t_8 = (__pyx_v_point <= (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->right.data + __pyx_t_13 * __pyx_v_self->right.strides[0]) ))));
      }
      __pyx_t_14 = (__pyx_t_8 != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":2729
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point <= self.right[i]:
 *                     result.append(self.indices[i])             # <<<<<<<<<<<<<<
 *         else:
 *             # There are child nodes. Based on comparing our query to the pivot,
 */
        __pyx_t_15 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->indices.data + __pyx_t_15 * __pyx_v_self->indices.strides[0]) ))));

        /* "pandas/_libs/intervaltree.pxi":2728
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point <= self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      }
    }

    /* "pandas/_libs/intervaltree.pxi":2723
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
    goto __pyx_L3;
  }

  /* "pandas/_libs/intervaltree.pxi":2733
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
  /*else*/ {
    __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":2734
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:
 *                 values = self.center_left_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 */
      __pyx_t_16 = __pyx_v_self->center_left_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":2735
 *             if point < self.pivot:
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:
 */
      __pyx_t_16 = __pyx_v_self->center_left_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_indices = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":2736
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):             # <<<<<<<<<<<<<<
 *                     if not values[i] <= point:
 *                         break
 */
      __pyx_t_9 = __pyx_v_self->n_center;
      __pyx_t_10 = __pyx_t_9;
      for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":2737
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_17 = __pyx_v_i;
        __pyx_t_14 = ((!(((*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_17 * __pyx_v_values.strides[0]) ))) <= __pyx_v_point) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":2738
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:
 */
          goto __pyx_L9_break;

          /* "pandas/_libs/intervaltree.pxi":2737
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":2739
 *                     if not values[i] <= point:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 */
        __pyx_t_18 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_18 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L9_break:;

      /* "pandas/_libs/intervaltree.pxi":2740
 *                         break
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      __pyx_t_14 = ((__pyx_v_point <= __pyx_v_self->left_node->max_right) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":2741
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)             # <<<<<<<<<<<<<<
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode *)__pyx_v_self->left_node->__pyx_vtab)->__pyx_fuse_0query(__pyx_v_self->left_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2741, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":2740
 *                         break
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":2733
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":2742
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
    __pyx_t_14 = ((__pyx_v_point > __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":2743
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 *                 values = self.center_right_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 */
      __pyx_t_16 = __pyx_v_self->center_right_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":2744
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:
 */
      __pyx_t_16 = __pyx_v_self->center_right_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_indices = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":2745
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):             # <<<<<<<<<<<<<<
 *                     if not point <= values[i]:
 *                         break
 */
      for (__pyx_t_11 = (__pyx_v_self->n_center - 1); __pyx_t_11 > -1L; __pyx_t_11-=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":2746
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_19 = __pyx_v_i;
        __pyx_t_14 = ((!((__pyx_v_point <= (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_19 * __pyx_v_values.strides[0]) )))) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":2747
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:
 */
          goto __pyx_L13_break;

          /* "pandas/_libs/intervaltree.pxi":2746
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":2748
 *                     if not point <= values[i]:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if self.right_node.min_left <= point:
 *                     self.right_node.query(result, point)
 */
        __pyx_t_20 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_20 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L13_break:;

      /* "pandas/_libs/intervaltree.pxi":2749
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      __pyx_t_14 = ((__pyx_v_self->right_node->min_left <= __pyx_v_point) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":2750
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:
 *                     self.right_node.query(result, point)             # <<<<<<<<<<<<<<
 *             else:
 *                 result.extend(self.center_left_indices)
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode *)__pyx_v_self->right_node->__pyx_vtab)->__pyx_fuse_0query(__pyx_v_self->right_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2750, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":2749
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":2742
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":2752
 *                     self.right_node.query(result, point)
 *             else:
 *                 result.extend(self.center_left_indices)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
    /*else*/ {
      __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->extend(__pyx_v_result, __pyx_v_self->center_left_indices); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2752, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_L7:;
  }
  __pyx_L3:;

  /* "pandas/_libs/intervaltree.pxi":2714
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __PYX_XDEC_MEMVIEW(&__pyx_t_16, 1);
  __Pyx_AddTraceback("pandas._libs.interval.Int64ClosedBothIntervalNode.query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_indices, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_values, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_9__pyx_fuse_0query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_0__pyx_mdef_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_9__pyx_fuse_0query = {"__pyx_fuse_0query", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_9__pyx_fuse_0query, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_2query};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_9__pyx_fuse_0query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result = 0;
  __pyx_t_5numpy_float64_t __pyx_v_point;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fuse_0query (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_result,&__pyx_n_s_point,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_result)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_point)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_0query", 1, 2, 2, 1); __PYX_ERR(1, 2714, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_0query") < 0)) __PYX_ERR(1, 2714, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_result = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)values[0]);
    __pyx_v_point = __pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_point == ((npy_float64)-1)) && PyErr_Occurred())) __PYX_ERR(1, 2714, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_0query", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 2714, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Int64ClosedBothIntervalNode.__pyx_fuse_0query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_result), __pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector, 1, "result", 0))) __PYX_ERR(1, 2714, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_8__pyx_fuse_0query(((struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode *)__pyx_v_self), __pyx_v_result, __pyx_v_point);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_8__pyx_fuse_0query(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float64_t __pyx_v_point) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__pyx_fuse_0query", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_fuse_0__pyx_f_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_query(__pyx_v_self, __pyx_v_result, __pyx_v_point, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2714, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Int64ClosedBothIntervalNode.__pyx_fuse_0query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_11__pyx_fuse_1query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_fuse_1__pyx_f_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float32_t __pyx_v_point, int __pyx_skip_dispatch) {
  __Pyx_memviewslice __pyx_v_indices = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_values = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __pyx_t_5numpy_int64_t __pyx_t_9;
  __pyx_t_5numpy_int64_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  int __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  __Pyx_memviewslice __pyx_t_16 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_17;
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  __Pyx_RefNannySetupContext("__pyx_fuse_1query", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pyx_fuse_1query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2714, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_11__pyx_fuse_1query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = PyFloat_FromDouble(__pyx_v_point); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2714, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2714, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2714, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 2714, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_5) {
            __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2714, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2714, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_3query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_4 = PyFloat_FromDouble(__pyx_v_point); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 2714, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_7 = __pyx_t_1; __pyx_t_3 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2714, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2714, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 2714, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (__pyx_t_3) {
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_t_4);
          __pyx_t_4 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2714, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pandas/_libs/intervaltree.pxi":2723
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
  __pyx_t_8 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_8) {

    /* "pandas/_libs/intervaltree.pxi":2727
 *             # continue the binary tree structure. Instead, we use linear
 *             # search.
 *             for i in range(self.n_elements):             # <<<<<<<<<<<<<<
 *                 if self.left[i] <= point <= self.right[i]:
 *                     result.append(self.indices[i])
 */
    __pyx_t_9 = __pyx_v_self->n_elements;
    __pyx_t_10 = __pyx_t_9;
    for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
      __pyx_v_i = __pyx_t_11;

      /* "pandas/_libs/intervaltree.pxi":2728
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point <= self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      __pyx_t_12 = __pyx_v_i;
      __pyx_t_8 = ((*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->left.data + __pyx_t_12 * __pyx_v_self->left.strides[0]) ))) <= __pyx_v_point);
      if (__pyx_t_8) {
        __pyx_t_13 = __pyx_v_i;
        __pyx_t_8 = (__pyx_v_point <= (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->right.data + __pyx_t_13 * __pyx_v_self->right.strides[0]) ))));
      }
      __pyx_t_14 = (__pyx_t_8 != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":2729
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point <= self.right[i]:
 *                     result.append(self.indices[i])             # <<<<<<<<<<<<<<
 *         else:
 *             # There are child nodes. Based on comparing our query to the pivot,
 */
        __pyx_t_15 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->indices.data + __pyx_t_15 * __pyx_v_self->indices.strides[0]) ))));

        /* "pandas/_libs/intervaltree.pxi":2728
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point <= self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      }
    }

    /* "pandas/_libs/intervaltree.pxi":2723
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
    goto __pyx_L3;
  }

  /* "pandas/_libs/intervaltree.pxi":2733
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
  /*else*/ {
    __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":2734
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:
 *                 values = self.center_left_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 */
      __pyx_t_16 = __pyx_v_self->center_left_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":2735
 *             if point < self.pivot:
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:
 */
      __pyx_t_16 = __pyx_v_self->center_left_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_indices = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":2736
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):             # <<<<<<<<<<<<<<
 *                     if not values[i] <= point:
 *                         break
 */
      __pyx_t_9 = __pyx_v_self->n_center;
      __pyx_t_10 = __pyx_t_9;
      for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":2737
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_17 = __pyx_v_i;
        __pyx_t_14 = ((!(((*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_17 * __pyx_v_values.strides[0]) ))) <= __pyx_v_point) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":2738
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:
 */
          goto __pyx_L9_break;

          /* "pandas/_libs/intervaltree.pxi":2737
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":2739
 *                     if not values[i] <= point:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 */
        __pyx_t_18 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_18 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L9_break:;

      /* "pandas/_libs/intervaltree.pxi":2740
 *                         break
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      __pyx_t_14 = ((__pyx_v_point <= __pyx_v_self->left_node->max_right) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":2741
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)             # <<<<<<<<<<<<<<
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode *)__pyx_v_self->left_node->__pyx_vtab)->__pyx_fuse_1query(__pyx_v_self->left_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2741, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":2740
 *                         break
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":2733
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":2742
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
    __pyx_t_14 = ((__pyx_v_point > __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":2743
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 *                 values = self.center_right_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 */
      __pyx_t_16 = __pyx_v_self->center_right_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":2744
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:
 */
      __pyx_t_16 = __pyx_v_self->center_right_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_indices = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":2745
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):             # <<<<<<<<<<<<<<
 *                     if not point <= values[i]:
 *                         break
 */
      for (__pyx_t_11 = (__pyx_v_self->n_center - 1); __pyx_t_11 > -1L; __pyx_t_11-=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":2746
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_19 = __pyx_v_i;
        __pyx_t_14 = ((!((__pyx_v_point <= (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_19 * __pyx_v_values.strides[0]) )))) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":2747
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:
 */
          goto __pyx_L13_break;

          /* "pandas/_libs/intervaltree.pxi":2746
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":2748
 *                     if not point <= values[i]:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if self.right_node.min_left <= point:
 *                     self.right_node.query(result, point)
 */
        __pyx_t_20 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_20 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L13_break:;

      /* "pandas/_libs/intervaltree.pxi":2749
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      __pyx_t_14 = ((__pyx_v_self->right_node->min_left <= __pyx_v_point) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":2750
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:
 *                     self.right_node.query(result, point)             # <<<<<<<<<<<<<<
 *             else:
 *                 result.extend(self.center_left_indices)
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode *)__pyx_v_self->right_node->__pyx_vtab)->__pyx_fuse_1query(__pyx_v_self->right_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2750, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":2749
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":2742
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":2752
 *                     self.right_node.query(result, point)
 *             else:
 *                 result.extend(self.center_left_indices)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
    /*else*/ {
      __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->extend(__pyx_v_result, __pyx_v_self->center_left_indices); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2752, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_L7:;
  }
  __pyx_L3:;

  /* "pandas/_libs/intervaltree.pxi":2714
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __PYX_XDEC_MEMVIEW(&__pyx_t_16, 1);
  __Pyx_AddTraceback("pandas._libs.interval.Int64ClosedBothIntervalNode.query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_indices, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_values, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_11__pyx_fuse_1query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_1__pyx_mdef_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_11__pyx_fuse_1query = {"__pyx_fuse_1query", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_11__pyx_fuse_1query, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_2query};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_11__pyx_fuse_1query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result = 0;
  __pyx_t_5numpy_float32_t __pyx_v_point;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fuse_1query (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_result,&__pyx_n_s_point,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_result)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_point)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_1query", 1, 2, 2, 1); __PYX_ERR(1, 2714, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_1query") < 0)) __PYX_ERR(1, 2714, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_result = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)values[0]);
    __pyx_v_point = __pyx_PyFloat_AsFloat(values[1]); if (unlikely((__pyx_v_point == ((npy_float32)-1)) && PyErr_Occurred())) __PYX_ERR(1, 2714, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_1query", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 2714, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Int64ClosedBothIntervalNode.__pyx_fuse_1query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_result), __pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector, 1, "result", 0))) __PYX_ERR(1, 2714, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_10__pyx_fuse_1query(((struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode *)__pyx_v_self), __pyx_v_result, __pyx_v_point);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_10__pyx_fuse_1query(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float32_t __pyx_v_point) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__pyx_fuse_1query", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_fuse_1__pyx_f_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_query(__pyx_v_self, __pyx_v_result, __pyx_v_point, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2714, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Int64ClosedBothIntervalNode.__pyx_fuse_1query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_13__pyx_fuse_2query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_fuse_2__pyx_f_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int64_t __pyx_v_point, int __pyx_skip_dispatch) {
  __Pyx_memviewslice __pyx_v_indices = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_values = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __pyx_t_5numpy_int64_t __pyx_t_9;
  __pyx_t_5numpy_int64_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  int __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  __Pyx_memviewslice __pyx_t_16 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_17;
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  __Pyx_RefNannySetupContext("__pyx_fuse_2query", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pyx_fuse_2query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2714, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_13__pyx_fuse_2query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = __Pyx_PyInt_From_npy_int64(__pyx_v_point); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2714, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2714, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2714, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 2714, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_5) {
            __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2714, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2714, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_3query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_4 = __Pyx_PyInt_From_npy_int64(__pyx_v_point); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 2714, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_7 = __pyx_t_1; __pyx_t_3 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2714, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2714, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 2714, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (__pyx_t_3) {
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_t_4);
          __pyx_t_4 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2714, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pandas/_libs/intervaltree.pxi":2723
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
  __pyx_t_8 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_8) {

    /* "pandas/_libs/intervaltree.pxi":2727
 *             # continue the binary tree structure. Instead, we use linear
 *             # search.
 *             for i in range(self.n_elements):             # <<<<<<<<<<<<<<
 *                 if self.left[i] <= point <= self.right[i]:
 *                     result.append(self.indices[i])
 */
    __pyx_t_9 = __pyx_v_self->n_elements;
    __pyx_t_10 = __pyx_t_9;
    for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
      __pyx_v_i = __pyx_t_11;

      /* "pandas/_libs/intervaltree.pxi":2728
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point <= self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      __pyx_t_12 = __pyx_v_i;
      __pyx_t_8 = ((*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->left.data + __pyx_t_12 * __pyx_v_self->left.strides[0]) ))) <= __pyx_v_point);
      if (__pyx_t_8) {
        __pyx_t_13 = __pyx_v_i;
        __pyx_t_8 = (__pyx_v_point <= (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->right.data + __pyx_t_13 * __pyx_v_self->right.strides[0]) ))));
      }
      __pyx_t_14 = (__pyx_t_8 != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":2729
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point <= self.right[i]:
 *                     result.append(self.indices[i])             # <<<<<<<<<<<<<<
 *         else:
 *             # There are child nodes. Based on comparing our query to the pivot,
 */
        __pyx_t_15 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->indices.data + __pyx_t_15 * __pyx_v_self->indices.strides[0]) ))));

        /* "pandas/_libs/intervaltree.pxi":2728
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point <= self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      }
    }

    /* "pandas/_libs/intervaltree.pxi":2723
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
    goto __pyx_L3;
  }

  /* "pandas/_libs/intervaltree.pxi":2733
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
  /*else*/ {
    __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":2734
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:
 *                 values = self.center_left_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 */
      __pyx_t_16 = __pyx_v_self->center_left_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":2735
 *             if point < self.pivot:
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:
 */
      __pyx_t_16 = __pyx_v_self->center_left_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_indices = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":2736
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):             # <<<<<<<<<<<<<<
 *                     if not values[i] <= point:
 *                         break
 */
      __pyx_t_9 = __pyx_v_self->n_center;
      __pyx_t_10 = __pyx_t_9;
      for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":2737
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_17 = __pyx_v_i;
        __pyx_t_14 = ((!(((*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_17 * __pyx_v_values.strides[0]) ))) <= __pyx_v_point) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":2738
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:
 */
          goto __pyx_L9_break;

          /* "pandas/_libs/intervaltree.pxi":2737
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":2739
 *                     if not values[i] <= point:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 */
        __pyx_t_18 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_18 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L9_break:;

      /* "pandas/_libs/intervaltree.pxi":2740
 *                         break
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      __pyx_t_14 = ((__pyx_v_point <= __pyx_v_self->left_node->max_right) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":2741
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)             # <<<<<<<<<<<<<<
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode *)__pyx_v_self->left_node->__pyx_vtab)->__pyx_fuse_2query(__pyx_v_self->left_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2741, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":2740
 *                         break
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":2733
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":2742
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
    __pyx_t_14 = ((__pyx_v_point > __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":2743
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 *                 values = self.center_right_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 */
      __pyx_t_16 = __pyx_v_self->center_right_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":2744
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:
 */
      __pyx_t_16 = __pyx_v_self->center_right_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_indices = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":2745
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):             # <<<<<<<<<<<<<<
 *                     if not point <= values[i]:
 *                         break
 */
      for (__pyx_t_11 = (__pyx_v_self->n_center - 1); __pyx_t_11 > -1L; __pyx_t_11-=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":2746
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_19 = __pyx_v_i;
        __pyx_t_14 = ((!((__pyx_v_point <= (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_19 * __pyx_v_values.strides[0]) )))) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":2747
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:
 */
          goto __pyx_L13_break;

          /* "pandas/_libs/intervaltree.pxi":2746
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":2748
 *                     if not point <= values[i]:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if self.right_node.min_left <= point:
 *                     self.right_node.query(result, point)
 */
        __pyx_t_20 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_20 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L13_break:;

      /* "pandas/_libs/intervaltree.pxi":2749
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      __pyx_t_14 = ((__pyx_v_self->right_node->min_left <= __pyx_v_point) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":2750
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:
 *                     self.right_node.query(result, point)             # <<<<<<<<<<<<<<
 *             else:
 *                 result.extend(self.center_left_indices)
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode *)__pyx_v_self->right_node->__pyx_vtab)->__pyx_fuse_2query(__pyx_v_self->right_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2750, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":2749
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":2742
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":2752
 *                     self.right_node.query(result, point)
 *             else:
 *                 result.extend(self.center_left_indices)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
    /*else*/ {
      __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->extend(__pyx_v_result, __pyx_v_self->center_left_indices); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2752, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_L7:;
  }
  __pyx_L3:;

  /* "pandas/_libs/intervaltree.pxi":2714
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __PYX_XDEC_MEMVIEW(&__pyx_t_16, 1);
  __Pyx_AddTraceback("pandas._libs.interval.Int64ClosedBothIntervalNode.query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_indices, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_values, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_13__pyx_fuse_2query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_2__pyx_mdef_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_13__pyx_fuse_2query = {"__pyx_fuse_2query", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_13__pyx_fuse_2query, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_2query};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_13__pyx_fuse_2query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result = 0;
  __pyx_t_5numpy_int64_t __pyx_v_point;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fuse_2query (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_result,&__pyx_n_s_point,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_result)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_point)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_2query", 1, 2, 2, 1); __PYX_ERR(1, 2714, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_2query") < 0)) __PYX_ERR(1, 2714, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_result = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)values[0]);
    __pyx_v_point = __Pyx_PyInt_As_npy_int64(values[1]); if (unlikely((__pyx_v_point == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(1, 2714, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_2query", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 2714, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Int64ClosedBothIntervalNode.__pyx_fuse_2query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_result), __pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector, 1, "result", 0))) __PYX_ERR(1, 2714, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_12__pyx_fuse_2query(((struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode *)__pyx_v_self), __pyx_v_result, __pyx_v_point);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_12__pyx_fuse_2query(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int64_t __pyx_v_point) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__pyx_fuse_2query", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_fuse_2__pyx_f_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_query(__pyx_v_self, __pyx_v_result, __pyx_v_point, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2714, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Int64ClosedBothIntervalNode.__pyx_fuse_2query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_15__pyx_fuse_3query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_fuse_3__pyx_f_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int32_t __pyx_v_point, int __pyx_skip_dispatch) {
  __Pyx_memviewslice __pyx_v_indices = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_values = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __pyx_t_5numpy_int64_t __pyx_t_9;
  __pyx_t_5numpy_int64_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  int __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  __Pyx_memviewslice __pyx_t_16 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_17;
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  __Pyx_RefNannySetupContext("__pyx_fuse_3query", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pyx_fuse_3query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2714, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_15__pyx_fuse_3query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = __Pyx_PyInt_From_npy_int32(__pyx_v_point); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2714, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2714, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2714, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 2714, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_5) {
            __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2714, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2714, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_3query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_4 = __Pyx_PyInt_From_npy_int32(__pyx_v_point); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 2714, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_7 = __pyx_t_1; __pyx_t_3 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2714, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2714, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 2714, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (__pyx_t_3) {
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_t_4);
          __pyx_t_4 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2714, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pandas/_libs/intervaltree.pxi":2723
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
  __pyx_t_8 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_8) {

    /* "pandas/_libs/intervaltree.pxi":2727
 *             # continue the binary tree structure. Instead, we use linear
 *             # search.
 *             for i in range(self.n_elements):             # <<<<<<<<<<<<<<
 *                 if self.left[i] <= point <= self.right[i]:
 *                     result.append(self.indices[i])
 */
    __pyx_t_9 = __pyx_v_self->n_elements;
    __pyx_t_10 = __pyx_t_9;
    for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
      __pyx_v_i = __pyx_t_11;

      /* "pandas/_libs/intervaltree.pxi":2728
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point <= self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      __pyx_t_12 = __pyx_v_i;
      __pyx_t_8 = ((*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->left.data + __pyx_t_12 * __pyx_v_self->left.strides[0]) ))) <= __pyx_v_point);
      if (__pyx_t_8) {
        __pyx_t_13 = __pyx_v_i;
        __pyx_t_8 = (__pyx_v_point <= (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->right.data + __pyx_t_13 * __pyx_v_self->right.strides[0]) ))));
      }
      __pyx_t_14 = (__pyx_t_8 != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":2729
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point <= self.right[i]:
 *                     result.append(self.indices[i])             # <<<<<<<<<<<<<<
 *         else:
 *             # There are child nodes. Based on comparing our query to the pivot,
 */
        __pyx_t_15 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->indices.data + __pyx_t_15 * __pyx_v_self->indices.strides[0]) ))));

        /* "pandas/_libs/intervaltree.pxi":2728
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point <= self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      }
    }

    /* "pandas/_libs/intervaltree.pxi":2723
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
    goto __pyx_L3;
  }

  /* "pandas/_libs/intervaltree.pxi":2733
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
  /*else*/ {
    __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":2734
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:
 *                 values = self.center_left_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 */
      __pyx_t_16 = __pyx_v_self->center_left_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":2735
 *             if point < self.pivot:
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:
 */
      __pyx_t_16 = __pyx_v_self->center_left_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_indices = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":2736
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):             # <<<<<<<<<<<<<<
 *                     if not values[i] <= point:
 *                         break
 */
      __pyx_t_9 = __pyx_v_self->n_center;
      __pyx_t_10 = __pyx_t_9;
      for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":2737
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_17 = __pyx_v_i;
        __pyx_t_14 = ((!(((*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_17 * __pyx_v_values.strides[0]) ))) <= __pyx_v_point) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":2738
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:
 */
          goto __pyx_L9_break;

          /* "pandas/_libs/intervaltree.pxi":2737
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":2739
 *                     if not values[i] <= point:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 */
        __pyx_t_18 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_18 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L9_break:;

      /* "pandas/_libs/intervaltree.pxi":2740
 *                         break
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      __pyx_t_14 = ((__pyx_v_point <= __pyx_v_self->left_node->max_right) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":2741
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)             # <<<<<<<<<<<<<<
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode *)__pyx_v_self->left_node->__pyx_vtab)->__pyx_fuse_3query(__pyx_v_self->left_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2741, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":2740
 *                         break
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":2733
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":2742
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
    __pyx_t_14 = ((__pyx_v_point > __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":2743
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 *                 values = self.center_right_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 */
      __pyx_t_16 = __pyx_v_self->center_right_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":2744
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:
 */
      __pyx_t_16 = __pyx_v_self->center_right_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_indices = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":2745
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):             # <<<<<<<<<<<<<<
 *                     if not point <= values[i]:
 *                         break
 */
      for (__pyx_t_11 = (__pyx_v_self->n_center - 1); __pyx_t_11 > -1L; __pyx_t_11-=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":2746
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_19 = __pyx_v_i;
        __pyx_t_14 = ((!((__pyx_v_point <= (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_19 * __pyx_v_values.strides[0]) )))) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":2747
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:
 */
          goto __pyx_L13_break;

          /* "pandas/_libs/intervaltree.pxi":2746
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":2748
 *                     if not point <= values[i]:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if self.right_node.min_left <= point:
 *                     self.right_node.query(result, point)
 */
        __pyx_t_20 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_20 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L13_break:;

      /* "pandas/_libs/intervaltree.pxi":2749
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      __pyx_t_14 = ((__pyx_v_self->right_node->min_left <= __pyx_v_point) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":2750
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:
 *                     self.right_node.query(result, point)             # <<<<<<<<<<<<<<
 *             else:
 *                 result.extend(self.center_left_indices)
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode *)__pyx_v_self->right_node->__pyx_vtab)->__pyx_fuse_3query(__pyx_v_self->right_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2750, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":2749
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":2742
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":2752
 *                     self.right_node.query(result, point)
 *             else:
 *                 result.extend(self.center_left_indices)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
    /*else*/ {
      __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->extend(__pyx_v_result, __pyx_v_self->center_left_indices); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2752, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_L7:;
  }
  __pyx_L3:;

  /* "pandas/_libs/intervaltree.pxi":2714
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __PYX_XDEC_MEMVIEW(&__pyx_t_16, 1);
  __Pyx_AddTraceback("pandas._libs.interval.Int64ClosedBothIntervalNode.query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_indices, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_values, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_15__pyx_fuse_3query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_3__pyx_mdef_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_15__pyx_fuse_3query = {"__pyx_fuse_3query", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_15__pyx_fuse_3query, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_2query};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_15__pyx_fuse_3query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result = 0;
  __pyx_t_5numpy_int32_t __pyx_v_point;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fuse_3query (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_result,&__pyx_n_s_point,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_result)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_point)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_3query", 1, 2, 2, 1); __PYX_ERR(1, 2714, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_3query") < 0)) __PYX_ERR(1, 2714, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_result = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)values[0]);
    __pyx_v_point = __Pyx_PyInt_As_npy_int32(values[1]); if (unlikely((__pyx_v_point == ((npy_int32)-1)) && PyErr_Occurred())) __PYX_ERR(1, 2714, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_3query", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 2714, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Int64ClosedBothIntervalNode.__pyx_fuse_3query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_result), __pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector, 1, "result", 0))) __PYX_ERR(1, 2714, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_14__pyx_fuse_3query(((struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode *)__pyx_v_self), __pyx_v_result, __pyx_v_point);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_14__pyx_fuse_3query(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int32_t __pyx_v_point) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__pyx_fuse_3query", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_fuse_3__pyx_f_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_query(__pyx_v_self, __pyx_v_result, __pyx_v_point, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2714, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Int64ClosedBothIntervalNode.__pyx_fuse_3query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_17__pyx_fuse_4query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_fuse_4__pyx_f_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_uint64_t __pyx_v_point, int __pyx_skip_dispatch) {
  __Pyx_memviewslice __pyx_v_indices = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_values = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __pyx_t_5numpy_int64_t __pyx_t_9;
  __pyx_t_5numpy_int64_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  int __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  __Pyx_memviewslice __pyx_t_16 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_17;
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  __Pyx_RefNannySetupContext("__pyx_fuse_4query", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pyx_fuse_4query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2714, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_17__pyx_fuse_4query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = __Pyx_PyInt_From_npy_uint64(__pyx_v_point); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2714, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2714, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2714, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 2714, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_5) {
            __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2714, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2714, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_3query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_4 = __Pyx_PyInt_From_npy_uint64(__pyx_v_point); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 2714, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_7 = __pyx_t_1; __pyx_t_3 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2714, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2714, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 2714, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (__pyx_t_3) {
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_t_4);
          __pyx_t_4 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2714, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pandas/_libs/intervaltree.pxi":2723
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
  __pyx_t_8 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_8) {

    /* "pandas/_libs/intervaltree.pxi":2727
 *             # continue the binary tree structure. Instead, we use linear
 *             # search.
 *             for i in range(self.n_elements):             # <<<<<<<<<<<<<<
 *                 if self.left[i] <= point <= self.right[i]:
 *                     result.append(self.indices[i])
 */
    __pyx_t_9 = __pyx_v_self->n_elements;
    __pyx_t_10 = __pyx_t_9;
    for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
      __pyx_v_i = __pyx_t_11;

      /* "pandas/_libs/intervaltree.pxi":2728
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point <= self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      __pyx_t_12 = __pyx_v_i;
      __pyx_t_8 = ((*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->left.data + __pyx_t_12 * __pyx_v_self->left.strides[0]) ))) <= __pyx_v_point);
      if (__pyx_t_8) {
        __pyx_t_13 = __pyx_v_i;
        __pyx_t_8 = (__pyx_v_point <= (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->right.data + __pyx_t_13 * __pyx_v_self->right.strides[0]) ))));
      }
      __pyx_t_14 = (__pyx_t_8 != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":2729
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point <= self.right[i]:
 *                     result.append(self.indices[i])             # <<<<<<<<<<<<<<
 *         else:
 *             # There are child nodes. Based on comparing our query to the pivot,
 */
        __pyx_t_15 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->indices.data + __pyx_t_15 * __pyx_v_self->indices.strides[0]) ))));

        /* "pandas/_libs/intervaltree.pxi":2728
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point <= self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      }
    }

    /* "pandas/_libs/intervaltree.pxi":2723
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
    goto __pyx_L3;
  }

  /* "pandas/_libs/intervaltree.pxi":2733
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
  /*else*/ {
    __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":2734
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:
 *                 values = self.center_left_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 */
      __pyx_t_16 = __pyx_v_self->center_left_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":2735
 *             if point < self.pivot:
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:
 */
      __pyx_t_16 = __pyx_v_self->center_left_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_indices = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":2736
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):             # <<<<<<<<<<<<<<
 *                     if not values[i] <= point:
 *                         break
 */
      __pyx_t_9 = __pyx_v_self->n_center;
      __pyx_t_10 = __pyx_t_9;
      for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":2737
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_17 = __pyx_v_i;
        __pyx_t_14 = ((!(((*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_17 * __pyx_v_values.strides[0]) ))) <= __pyx_v_point) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":2738
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:
 */
          goto __pyx_L9_break;

          /* "pandas/_libs/intervaltree.pxi":2737
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":2739
 *                     if not values[i] <= point:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 */
        __pyx_t_18 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_18 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L9_break:;

      /* "pandas/_libs/intervaltree.pxi":2740
 *                         break
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      __pyx_t_14 = ((__pyx_v_point <= __pyx_v_self->left_node->max_right) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":2741
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)             # <<<<<<<<<<<<<<
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode *)__pyx_v_self->left_node->__pyx_vtab)->__pyx_fuse_4query(__pyx_v_self->left_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2741, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":2740
 *                         break
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":2733
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":2742
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
    __pyx_t_14 = ((__pyx_v_point > __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":2743
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 *                 values = self.center_right_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 */
      __pyx_t_16 = __pyx_v_self->center_right_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":2744
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:
 */
      __pyx_t_16 = __pyx_v_self->center_right_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_indices = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":2745
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):             # <<<<<<<<<<<<<<
 *                     if not point <= values[i]:
 *                         break
 */
      for (__pyx_t_11 = (__pyx_v_self->n_center - 1); __pyx_t_11 > -1L; __pyx_t_11-=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":2746
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_19 = __pyx_v_i;
        __pyx_t_14 = ((!((__pyx_v_point <= (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_19 * __pyx_v_values.strides[0]) )))) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":2747
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:
 */
          goto __pyx_L13_break;

          /* "pandas/_libs/intervaltree.pxi":2746
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":2748
 *                     if not point <= values[i]:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if self.right_node.min_left <= point:
 *                     self.right_node.query(result, point)
 */
        __pyx_t_20 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_20 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L13_break:;

      /* "pandas/_libs/intervaltree.pxi":2749
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      __pyx_t_14 = ((__pyx_v_self->right_node->min_left <= __pyx_v_point) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":2750
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:
 *                     self.right_node.query(result, point)             # <<<<<<<<<<<<<<
 *             else:
 *                 result.extend(self.center_left_indices)
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode *)__pyx_v_self->right_node->__pyx_vtab)->__pyx_fuse_4query(__pyx_v_self->right_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2750, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":2749
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":2742
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":2752
 *                     self.right_node.query(result, point)
 *             else:
 *                 result.extend(self.center_left_indices)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
    /*else*/ {
      __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->extend(__pyx_v_result, __pyx_v_self->center_left_indices); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2752, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_L7:;
  }
  __pyx_L3:;

  /* "pandas/_libs/intervaltree.pxi":2714
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __PYX_XDEC_MEMVIEW(&__pyx_t_16, 1);
  __Pyx_AddTraceback("pandas._libs.interval.Int64ClosedBothIntervalNode.query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_indices, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_values, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_17__pyx_fuse_4query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_4__pyx_mdef_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_17__pyx_fuse_4query = {"__pyx_fuse_4query", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_17__pyx_fuse_4query, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_2query};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_17__pyx_fuse_4query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result = 0;
  __pyx_t_5numpy_uint64_t __pyx_v_point;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fuse_4query (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_result,&__pyx_n_s_point,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_result)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_point)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_4query", 1, 2, 2, 1); __PYX_ERR(1, 2714, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_4query") < 0)) __PYX_ERR(1, 2714, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_result = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)values[0]);
    __pyx_v_point = __Pyx_PyInt_As_npy_uint64(values[1]); if (unlikely((__pyx_v_point == ((npy_uint64)-1)) && PyErr_Occurred())) __PYX_ERR(1, 2714, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_4query", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 2714, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Int64ClosedBothIntervalNode.__pyx_fuse_4query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_result), __pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector, 1, "result", 0))) __PYX_ERR(1, 2714, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_16__pyx_fuse_4query(((struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode *)__pyx_v_self), __pyx_v_result, __pyx_v_point);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_16__pyx_fuse_4query(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_uint64_t __pyx_v_point) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__pyx_fuse_4query", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_fuse_4__pyx_f_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_query(__pyx_v_self, __pyx_v_result, __pyx_v_point, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2714, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Int64ClosedBothIntervalNode.__pyx_fuse_4query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":2754
 *                 result.extend(self.center_left_indices)
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         if self.is_leaf_node:
 *             return ('<Int64ClosedBothIntervalNode: '
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_5__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_5__repr__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_4__repr__(((struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_4__repr__(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode *__pyx_v_self) {
  PyObject *__pyx_v_n_left = NULL;
  PyObject *__pyx_v_n_right = NULL;
  PyObject *__pyx_v_n_center = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  Py_UCS4 __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "pandas/_libs/intervaltree.pxi":2755
 * 
 *     def __repr__(self):
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             return ('<Int64ClosedBothIntervalNode: '
 *                     '%s elements (terminal)>' % self.n_elements)
 */
  __pyx_t_1 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_1) {

    /* "pandas/_libs/intervaltree.pxi":2756
 *     def __repr__(self):
 *         if self.is_leaf_node:
 *             return ('<Int64ClosedBothIntervalNode: '             # <<<<<<<<<<<<<<
 *                     '%s elements (terminal)>' % self.n_elements)
 *         else:
 */
    __Pyx_XDECREF(__pyx_r);

    /* "pandas/_libs/intervaltree.pxi":2757
 *         if self.is_leaf_node:
 *             return ('<Int64ClosedBothIntervalNode: '
 *                     '%s elements (terminal)>' % self.n_elements)             # <<<<<<<<<<<<<<
 *         else:
 *             n_left = self.left_node.n_elements
 */
    __pyx_t_2 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_elements); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2757, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PyUnicode_Format(__pyx_kp_u_Int64ClosedBothIntervalNode_s_e, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2757, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pandas/_libs/intervaltree.pxi":2755
 * 
 *     def __repr__(self):
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             return ('<Int64ClosedBothIntervalNode: '
 *                     '%s elements (terminal)>' % self.n_elements)
 */
  }

  /* "pandas/_libs/intervaltree.pxi":2759
 *                     '%s elements (terminal)>' % self.n_elements)
 *         else:
 *             n_left = self.left_node.n_elements             # <<<<<<<<<<<<<<
 *             n_right = self.right_node.n_elements
 *             n_center = self.n_elements - n_left - n_right
 */
  /*else*/ {
    __pyx_t_3 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->left_node->n_elements); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2759, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_v_n_left = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "pandas/_libs/intervaltree.pxi":2760
 *         else:
 *             n_left = self.left_node.n_elements
 *             n_right = self.right_node.n_elements             # <<<<<<<<<<<<<<
 *             n_center = self.n_elements - n_left - n_right
 *             return ('<Int64ClosedBothIntervalNode: '
 */
    __pyx_t_3 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->right_node->n_elements); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2760, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_v_n_right = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "pandas/_libs/intervaltree.pxi":2761
 *             n_left = self.left_node.n_elements
 *             n_right = self.right_node.n_elements
 *             n_center = self.n_elements - n_left - n_right             # <<<<<<<<<<<<<<
 *             return ('<Int64ClosedBothIntervalNode: '
 *                     'pivot %s, %s elements (%s left, %s right, %s '
 */
    __pyx_t_3 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_elements); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2761, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = PyNumber_Subtract(__pyx_t_3, __pyx_v_n_left); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2761, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyNumber_Subtract(__pyx_t_2, __pyx_v_n_right); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2761, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_n_center = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "pandas/_libs/intervaltree.pxi":2762
 *             n_right = self.right_node.n_elements
 *             n_center = self.n_elements - n_left - n_right
 *             return ('<Int64ClosedBothIntervalNode: '             # <<<<<<<<<<<<<<
 *                     'pivot %s, %s elements (%s left, %s right, %s '
 *                     'overlapping)>' % (self.pivot, self.n_elements,
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = PyTuple_New(11); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2762, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = 0;
    __pyx_t_5 = 127;
    __Pyx_INCREF(__pyx_kp_u_Int64ClosedBothIntervalNode_piv);
    __pyx_t_4 += 36;
    __Pyx_GIVEREF(__pyx_kp_u_Int64ClosedBothIntervalNode_piv);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_kp_u_Int64ClosedBothIntervalNode_piv);

    /* "pandas/_libs/intervaltree.pxi":2764
 *             return ('<Int64ClosedBothIntervalNode: '
 *                     'pivot %s, %s elements (%s left, %s right, %s '
 *                     'overlapping)>' % (self.pivot, self.n_elements,             # <<<<<<<<<<<<<<
 *                                        n_left, n_right, n_center))
 * 
 */
    __pyx_t_2 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->pivot); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2764, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_t_2), __pyx_empty_unicode); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 2764, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_5 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_6) > __pyx_t_5) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_6) : __pyx_t_5;
    __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_6);
    __pyx_t_6 = 0;
    __Pyx_INCREF(__pyx_kp_u__18);
    __pyx_t_4 += 2;
    __Pyx_GIVEREF(__pyx_kp_u__18);
    PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_kp_u__18);
    __pyx_t_6 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_elements); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 2764, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_2 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_t_6), __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2764, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_5 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) > __pyx_t_5) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) : __pyx_t_5;
    __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_3, 3, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_elements);
    __pyx_t_4 += 11;
    __Pyx_GIVEREF(__pyx_kp_u_elements);
    PyTuple_SET_ITEM(__pyx_t_3, 4, __pyx_kp_u_elements);

    /* "pandas/_libs/intervaltree.pxi":2765
 *                     'pivot %s, %s elements (%s left, %s right, %s '
 *                     'overlapping)>' % (self.pivot, self.n_elements,
 *                                        n_left, n_right, n_center))             # <<<<<<<<<<<<<<
 * 
 *     def counts(self):
 */
    __pyx_t_2 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_v_n_left), __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2765, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) > __pyx_t_5) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) : __pyx_t_5;
    __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_3, 5, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_left_2);
    __pyx_t_4 += 7;
    __Pyx_GIVEREF(__pyx_kp_u_left_2);
    PyTuple_SET_ITEM(__pyx_t_3, 6, __pyx_kp_u_left_2);
    __pyx_t_2 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_v_n_right), __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2765, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) > __pyx_t_5) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) : __pyx_t_5;
    __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_3, 7, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_right_2);
    __pyx_t_4 += 8;
    __Pyx_GIVEREF(__pyx_kp_u_right_2);
    PyTuple_SET_ITEM(__pyx_t_3, 8, __pyx_kp_u_right_2);
    __pyx_t_2 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_v_n_center), __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2765, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) > __pyx_t_5) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) : __pyx_t_5;
    __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_3, 9, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_overlapping);
    __pyx_t_4 += 14;
    __Pyx_GIVEREF(__pyx_kp_u_overlapping);
    PyTuple_SET_ITEM(__pyx_t_3, 10, __pyx_kp_u_overlapping);

    /* "pandas/_libs/intervaltree.pxi":2762
 *             n_right = self.right_node.n_elements
 *             n_center = self.n_elements - n_left - n_right
 *             return ('<Int64ClosedBothIntervalNode: '             # <<<<<<<<<<<<<<
 *                     'pivot %s, %s elements (%s left, %s right, %s '
 *                     'overlapping)>' % (self.pivot, self.n_elements,
 */
    __pyx_t_2 = __Pyx_PyUnicode_Join(__pyx_t_3, 11, __pyx_t_4, __pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2762, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;
  }

  /* "pandas/_libs/intervaltree.pxi":2754
 *                 result.extend(self.center_left_indices)
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         if self.is_leaf_node:
 *             return ('<Int64ClosedBothIntervalNode: '
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pandas._libs.interval.Int64ClosedBothIntervalNode.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_n_left);
  __Pyx_XDECREF(__pyx_v_n_right);
  __Pyx_XDECREF(__pyx_v_n_center);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":2767
 *                                        n_left, n_right, n_center))
 * 
 *     def counts(self):             # <<<<<<<<<<<<<<
 *         """
 *         Inspect counts on this node
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_7counts(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_6counts[] = "\n        Inspect counts on this node\n        useful for debugging purposes\n        ";
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_7counts(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("counts (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_6counts(((struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_6counts(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode *__pyx_v_self) {
  Py_ssize_t __pyx_v_m;
  PyObject *__pyx_v_l = NULL;
  PyObject *__pyx_v_r = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  size_t __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("counts", 0);

  /* "pandas/_libs/intervaltree.pxi":2772
 *         useful for debugging purposes
 *         """
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             return self.n_elements
 *         else:
 */
  __pyx_t_1 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_1) {

    /* "pandas/_libs/intervaltree.pxi":2773
 *         """
 *         if self.is_leaf_node:
 *             return self.n_elements             # <<<<<<<<<<<<<<
 *         else:
 *             m = len(self.center_left_values)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_elements); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2773, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "pandas/_libs/intervaltree.pxi":2772
 *         useful for debugging purposes
 *         """
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             return self.n_elements
 *         else:
 */
  }

  /* "pandas/_libs/intervaltree.pxi":2775
 *             return self.n_elements
 *         else:
 *             m = len(self.center_left_values)             # <<<<<<<<<<<<<<
 *             l = self.left_node.counts()
 *             r = self.right_node.counts()
 */
  /*else*/ {
    if (unlikely(!__pyx_v_self->center_left_values.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(1, 2775, __pyx_L1_error)}
    __pyx_t_3 = __Pyx_MemoryView_Len(__pyx_v_self->center_left_values); 
    __pyx_v_m = __pyx_t_3;

    /* "pandas/_libs/intervaltree.pxi":2776
 *         else:
 *             m = len(self.center_left_values)
 *             l = self.left_node.counts()             # <<<<<<<<<<<<<<
 *             r = self.right_node.counts()
 *             return (m, (l, r))
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->left_node), __pyx_n_s_counts); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 2776, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2776, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_l = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "pandas/_libs/intervaltree.pxi":2777
 *             m = len(self.center_left_values)
 *             l = self.left_node.counts()
 *             r = self.right_node.counts()             # <<<<<<<<<<<<<<
 *             return (m, (l, r))
 * 
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->right_node), __pyx_n_s_counts); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 2777, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2777, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_r = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "pandas/_libs/intervaltree.pxi":2778
 *             l = self.left_node.counts()
 *             r = self.right_node.counts()
 *             return (m, (l, r))             # <<<<<<<<<<<<<<
 * 
 * NODE_CLASSES['int64',
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = PyInt_FromSsize_t(__pyx_v_m); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2778, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 2778, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_v_l);
    __Pyx_GIVEREF(__pyx_v_l);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_l);
    __Pyx_INCREF(__pyx_v_r);
    __Pyx_GIVEREF(__pyx_v_r);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_v_r);
    __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 2778, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_4);
    __pyx_t_2 = 0;
    __pyx_t_4 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;
  }

  /* "pandas/_libs/intervaltree.pxi":2767
 *                                        n_left, n_right, n_center))
 * 
 *     def counts(self):             # <<<<<<<<<<<<<<
 *         """
 *         Inspect counts on this node
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pandas._libs.interval.Int64ClosedBothIntervalNode.counts", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_l);
  __Pyx_XDECREF(__pyx_v_r);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":2624
 *         int64_t[:] center_left_indices, center_right_indices, indices
 *         int64_t min_left, max_right
 *         readonly int64_t pivot             # <<<<<<<<<<<<<<
 *         readonly int64_t n_elements, n_center, leaf_size
 *         readonly bint is_leaf_node
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_5pivot_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_5pivot_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_5pivot___get__(((struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_5pivot___get__(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->pivot); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2624, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Int64ClosedBothIntervalNode.pivot.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":2625
 *         int64_t min_left, max_right
 *         readonly int64_t pivot
 *         readonly int64_t n_elements, n_center, leaf_size             # <<<<<<<<<<<<<<
 *         readonly bint is_leaf_node
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_10n_elements_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_10n_elements_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_10n_elements___get__(((struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_10n_elements___get__(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_elements); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2625, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Int64ClosedBothIntervalNode.n_elements.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_8n_center_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_8n_center_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_8n_center___get__(((struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_8n_center___get__(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_center); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2625, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Int64ClosedBothIntervalNode.n_center.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_9leaf_size_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_9leaf_size_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_9leaf_size___get__(((struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_9leaf_size___get__(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->leaf_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2625, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Int64ClosedBothIntervalNode.leaf_size.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":2626
 *         readonly int64_t pivot
 *         readonly int64_t n_elements, n_center, leaf_size
 *         readonly bint is_leaf_node             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self,
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_12is_leaf_node_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_12is_leaf_node_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_12is_leaf_node___get__(((struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_12is_leaf_node___get__(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->is_leaf_node); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2626, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Int64ClosedBothIntervalNode.is_leaf_node.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_21__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_21__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_20__reduce_cython__(((struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_20__reduce_cython__(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  int __pyx_t_16;
  int __pyx_t_17;
  int __pyx_t_18;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.center_left_indices, self.center_left_values, self.center_right_indices, self.center_right_values, self.indices, self.is_leaf_node, self.leaf_size, self.left, self.left_node, self.max_right, self.min_left, self.n_center, self.n_elements, self.pivot, self.right, self.right_node)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  if (unlikely(!__pyx_v_self->center_left_indices.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_1 = __pyx_memoryview_fromslice(__pyx_v_self->center_left_indices, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_int64_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_int64_t, 0);; if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (unlikely(!__pyx_v_self->center_left_values.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_2 = __pyx_memoryview_fromslice(__pyx_v_self->center_left_values, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_int64_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_int64_t, 0);; if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (unlikely(!__pyx_v_self->center_right_indices.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_3 = __pyx_memoryview_fromslice(__pyx_v_self->center_right_indices, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_int64_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_int64_t, 0);; if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely(!__pyx_v_self->center_right_values.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_4 = __pyx_memoryview_fromslice(__pyx_v_self->center_right_values, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_int64_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_int64_t, 0);; if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (unlikely(!__pyx_v_self->indices.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_5 = __pyx_memoryview_fromslice(__pyx_v_self->indices, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_int64_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_int64_t, 0);; if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyBool_FromLong(__pyx_v_self->is_leaf_node); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->leaf_size); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (unlikely(!__pyx_v_self->left.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_8 = __pyx_memoryview_fromslice(__pyx_v_self->left, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_int64_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_int64_t, 0);; if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->max_right); if (unlikely(!__pyx_t_9)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_10 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->min_left); if (unlikely(!__pyx_t_10)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_11 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_center); if (unlikely(!__pyx_t_11)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_12 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_elements); if (unlikely(!__pyx_t_12)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __pyx_t_13 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->pivot); if (unlikely(!__pyx_t_13)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  if (unlikely(!__pyx_v_self->right.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_14 = __pyx_memoryview_fromslice(__pyx_v_self->right, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_int64_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_int64_t, 0);; if (unlikely(!__pyx_t_14)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __pyx_t_15 = PyTuple_New(16); if (unlikely(!__pyx_t_15)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_15, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_15, 1, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_15, 2, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_15, 3, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_15, 4, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_15, 5, __pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_15, 6, __pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_15, 7, __pyx_t_8);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->left_node));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self->left_node));
  PyTuple_SET_ITEM(__pyx_t_15, 8, ((PyObject *)__pyx_v_self->left_node));
  __Pyx_GIVEREF(__pyx_t_9);
  PyTuple_SET_ITEM(__pyx_t_15, 9, __pyx_t_9);
  __Pyx_GIVEREF(__pyx_t_10);
  PyTuple_SET_ITEM(__pyx_t_15, 10, __pyx_t_10);
  __Pyx_GIVEREF(__pyx_t_11);
  PyTuple_SET_ITEM(__pyx_t_15, 11, __pyx_t_11);
  __Pyx_GIVEREF(__pyx_t_12);
  PyTuple_SET_ITEM(__pyx_t_15, 12, __pyx_t_12);
  __Pyx_GIVEREF(__pyx_t_13);
  PyTuple_SET_ITEM(__pyx_t_15, 13, __pyx_t_13);
  __Pyx_GIVEREF(__pyx_t_14);
  PyTuple_SET_ITEM(__pyx_t_15, 14, __pyx_t_14);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->right_node));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self->right_node));
  PyTuple_SET_ITEM(__pyx_t_15, 15, ((PyObject *)__pyx_v_self->right_node));
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  __pyx_t_6 = 0;
  __pyx_t_7 = 0;
  __pyx_t_8 = 0;
  __pyx_t_9 = 0;
  __pyx_t_10 = 0;
  __pyx_t_11 = 0;
  __pyx_t_12 = 0;
  __pyx_t_13 = 0;
  __pyx_t_14 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_15);
  __pyx_t_15 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.center_left_indices, self.center_left_values, self.center_right_indices, self.center_right_values, self.indices, self.is_leaf_node, self.leaf_size, self.left, self.left_node, self.max_right, self.min_left, self.n_center, self.n_elements, self.pivot, self.right, self.right_node)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_15 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_15)) __PYX_ERR(2, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __pyx_v__dict = __pyx_t_15;
  __pyx_t_15 = 0;

  /* "(tree fragment)":7
 *     state = (self.center_left_indices, self.center_left_values, self.center_right_indices, self.center_right_values, self.indices, self.is_leaf_node, self.leaf_size, self.left, self.left_node, self.max_right, self.min_left, self.n_center, self.n_elements, self.pivot, self.right, self.right_node)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_16 = (__pyx_v__dict != Py_None);
  __pyx_t_17 = (__pyx_t_16 != 0);
  if (__pyx_t_17) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_15 = PyTuple_New(1); if (unlikely(!__pyx_t_15)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    PyTuple_SET_ITEM(__pyx_t_15, 0, __pyx_v__dict);
    __pyx_t_14 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_15); if (unlikely(!__pyx_t_14)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_14));
    __pyx_t_14 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.left_node is not None or self.right_node is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.center_left_indices, self.center_left_values, self.center_right_indices, self.center_right_values, self.indices, self.is_leaf_node, self.leaf_size, self.left, self.left_node, self.max_right, self.min_left, self.n_center, self.n_elements, self.pivot, self.right, self.right_node)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.left_node is not None or self.right_node is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_Int64ClosedBothIntervalNode, (type(self), 0x7a104e5, None), state
 */
  /*else*/ {
    __pyx_t_16 = (((PyObject *)__pyx_v_self->left_node) != Py_None);
    __pyx_t_18 = (__pyx_t_16 != 0);
    if (!__pyx_t_18) {
    } else {
      __pyx_t_17 = __pyx_t_18;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_18 = (((PyObject *)__pyx_v_self->right_node) != Py_None);
    __pyx_t_16 = (__pyx_t_18 != 0);
    __pyx_t_17 = __pyx_t_16;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_17;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.left_node is not None or self.right_node is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Int64ClosedBothIntervalNode, (type(self), 0x7a104e5, None), state
 *     else:
 */
  __pyx_t_17 = (__pyx_v_use_setstate != 0);
  if (__pyx_t_17) {

    /* "(tree fragment)":13
 *         use_setstate = self.left_node is not None or self.right_node is not None
 *     if use_setstate:
 *         return __pyx_unpickle_Int64ClosedBothIntervalNode, (type(self), 0x7a104e5, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_Int64ClosedBothIntervalNode, (type(self), 0x7a104e5, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_14, __pyx_n_s_pyx_unpickle_Int64ClosedBothIn); if (unlikely(!__pyx_t_14)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __pyx_t_15 = PyTuple_New(3); if (unlikely(!__pyx_t_15)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_15, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_127993061);
    __Pyx_GIVEREF(__pyx_int_127993061);
    PyTuple_SET_ITEM(__pyx_t_15, 1, __pyx_int_127993061);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_15, 2, Py_None);
    __pyx_t_13 = PyTuple_New(3); if (unlikely(!__pyx_t_13)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __Pyx_GIVEREF(__pyx_t_14);
    PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_t_14);
    __Pyx_GIVEREF(__pyx_t_15);
    PyTuple_SET_ITEM(__pyx_t_13, 1, __pyx_t_15);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_13, 2, __pyx_v_state);
    __pyx_t_14 = 0;
    __pyx_t_15 = 0;
    __pyx_r = __pyx_t_13;
    __pyx_t_13 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.left_node is not None or self.right_node is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Int64ClosedBothIntervalNode, (type(self), 0x7a104e5, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_Int64ClosedBothIntervalNode, (type(self), 0x7a104e5, None), state
 *     else:
 *         return __pyx_unpickle_Int64ClosedBothIntervalNode, (type(self), 0x7a104e5, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Int64ClosedBothIntervalNode__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_13, __pyx_n_s_pyx_unpickle_Int64ClosedBothIn); if (unlikely(!__pyx_t_13)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __pyx_t_15 = PyTuple_New(3); if (unlikely(!__pyx_t_15)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_15, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_127993061);
    __Pyx_GIVEREF(__pyx_int_127993061);
    PyTuple_SET_ITEM(__pyx_t_15, 1, __pyx_int_127993061);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_15, 2, __pyx_v_state);
    __pyx_t_14 = PyTuple_New(2); if (unlikely(!__pyx_t_14)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_GIVEREF(__pyx_t_13);
    PyTuple_SET_ITEM(__pyx_t_14, 0, __pyx_t_13);
    __Pyx_GIVEREF(__pyx_t_15);
    PyTuple_SET_ITEM(__pyx_t_14, 1, __pyx_t_15);
    __pyx_t_13 = 0;
    __pyx_t_15 = 0;
    __pyx_r = __pyx_t_14;
    __pyx_t_14 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_AddTraceback("pandas._libs.interval.Int64ClosedBothIntervalNode.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Int64ClosedBothIntervalNode, (type(self), 0x7a104e5, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Int64ClosedBothIntervalNode__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_23__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_23__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_22__setstate_cython__(((struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_22__setstate_cython__(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_Int64ClosedBothIntervalNode, (type(self), 0x7a104e5, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Int64ClosedBothIntervalNode__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(2, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_6pandas_5_libs_8interval___pyx_unpickle_Int64ClosedBothIntervalNode__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Int64ClosedBothIntervalNode, (type(self), 0x7a104e5, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Int64ClosedBothIntervalNode__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Int64ClosedBothIntervalNode.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":2798
 *         readonly bint is_leaf_node
 * 
 *     def __init__(self,             # <<<<<<<<<<<<<<
 *                  ndarray[int64_t, ndim=1] left,
 *                  ndarray[int64_t, ndim=1] right,
 */

/* Python wrapper */
static int __pyx_pw_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyArrayObject *__pyx_v_left = 0;
  PyArrayObject *__pyx_v_right = 0;
  PyArrayObject *__pyx_v_indices = 0;
  __pyx_t_5numpy_int64_t __pyx_v_leaf_size;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_left,&__pyx_n_s_right,&__pyx_n_s_indices,&__pyx_n_s_leaf_size,0};
    PyObject* values[4] = {0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_left)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_right)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 4, 4, 1); __PYX_ERR(1, 2798, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_indices)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 4, 4, 2); __PYX_ERR(1, 2798, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_leaf_size)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 4, 4, 3); __PYX_ERR(1, 2798, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(1, 2798, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 4) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
    }
    __pyx_v_left = ((PyArrayObject *)values[0]);
    __pyx_v_right = ((PyArrayObject *)values[1]);
    __pyx_v_indices = ((PyArrayObject *)values[2]);
    __pyx_v_leaf_size = __Pyx_PyInt_As_npy_int64(values[3]); if (unlikely((__pyx_v_leaf_size == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(1, 2802, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 4, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 2798, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Int64ClosedNeitherIntervalNode.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_left), __pyx_ptype_5numpy_ndarray, 1, "left", 0))) __PYX_ERR(1, 2799, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_right), __pyx_ptype_5numpy_ndarray, 1, "right", 0))) __PYX_ERR(1, 2800, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_indices), __pyx_ptype_5numpy_ndarray, 1, "indices", 0))) __PYX_ERR(1, 2801, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode___init__(((struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode *)__pyx_v_self), __pyx_v_left, __pyx_v_right, __pyx_v_indices, __pyx_v_leaf_size);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode___init__(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode *__pyx_v_self, PyArrayObject *__pyx_v_left, PyArrayObject *__pyx_v_right, PyArrayObject *__pyx_v_indices, __pyx_t_5numpy_int64_t __pyx_v_leaf_size) {
  PyObject *__pyx_v_left_set = NULL;
  PyObject *__pyx_v_right_set = NULL;
  PyObject *__pyx_v_center_set = NULL;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_indices;
  __Pyx_Buffer __pyx_pybuffer_indices;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_left;
  __Pyx_Buffer __pyx_pybuffer_left;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_right;
  __Pyx_Buffer __pyx_pybuffer_right;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __pyx_t_5numpy_int64_t __pyx_t_6;
  __Pyx_memviewslice __pyx_t_7 = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  __Pyx_memviewslice __pyx_t_10 = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *(*__pyx_t_11)(PyObject *);
  size_t __pyx_t_12;
  __Pyx_RefNannySetupContext("__init__", 0);
  __pyx_pybuffer_left.pybuffer.buf = NULL;
  __pyx_pybuffer_left.refcount = 0;
  __pyx_pybuffernd_left.data = NULL;
  __pyx_pybuffernd_left.rcbuffer = &__pyx_pybuffer_left;
  __pyx_pybuffer_right.pybuffer.buf = NULL;
  __pyx_pybuffer_right.refcount = 0;
  __pyx_pybuffernd_right.data = NULL;
  __pyx_pybuffernd_right.rcbuffer = &__pyx_pybuffer_right;
  __pyx_pybuffer_indices.pybuffer.buf = NULL;
  __pyx_pybuffer_indices.refcount = 0;
  __pyx_pybuffernd_indices.data = NULL;
  __pyx_pybuffernd_indices.rcbuffer = &__pyx_pybuffer_indices;
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_left.rcbuffer->pybuffer, (PyObject*)__pyx_v_left, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 2798, __pyx_L1_error)
  }
  __pyx_pybuffernd_left.diminfo[0].strides = __pyx_pybuffernd_left.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_left.diminfo[0].shape = __pyx_pybuffernd_left.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_right.rcbuffer->pybuffer, (PyObject*)__pyx_v_right, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 2798, __pyx_L1_error)
  }
  __pyx_pybuffernd_right.diminfo[0].strides = __pyx_pybuffernd_right.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_right.diminfo[0].shape = __pyx_pybuffernd_right.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_indices.rcbuffer->pybuffer, (PyObject*)__pyx_v_indices, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 2798, __pyx_L1_error)
  }
  __pyx_pybuffernd_indices.diminfo[0].strides = __pyx_pybuffernd_indices.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_indices.diminfo[0].shape = __pyx_pybuffernd_indices.rcbuffer->pybuffer.shape[0];

  /* "pandas/_libs/intervaltree.pxi":2804
 *                  int64_t leaf_size):
 * 
 *         self.n_elements = len(left)             # <<<<<<<<<<<<<<
 *         self.leaf_size = leaf_size
 * 
 */
  __pyx_t_1 = PyObject_Length(((PyObject *)__pyx_v_left)); if (unlikely(__pyx_t_1 == ((Py_ssize_t)-1))) __PYX_ERR(1, 2804, __pyx_L1_error)
  __pyx_v_self->n_elements = __pyx_t_1;

  /* "pandas/_libs/intervaltree.pxi":2805
 * 
 *         self.n_elements = len(left)
 *         self.leaf_size = leaf_size             # <<<<<<<<<<<<<<
 * 
 *         # min_left and min_right are used to speed-up query by skipping
 */
  __pyx_v_self->leaf_size = __pyx_v_leaf_size;

  /* "pandas/_libs/intervaltree.pxi":2810
 *         # query on sub-nodes. If this node has size 0, query is cheap,
 *         # so these values don't matter.
 *         if left.size > 0:             # <<<<<<<<<<<<<<
 *             self.min_left = left.min()
 *             self.max_right = right.max()
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_left), __pyx_n_s_size); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2810, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyObject_RichCompare(__pyx_t_2, __pyx_int_0, Py_GT); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2810, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(1, 2810, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_4) {

    /* "pandas/_libs/intervaltree.pxi":2811
 *         # so these values don't matter.
 *         if left.size > 0:
 *             self.min_left = left.min()             # <<<<<<<<<<<<<<
 *             self.max_right = right.max()
 *         else:
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_left), __pyx_n_s_min); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2811, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2811, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_6 = __Pyx_PyInt_As_npy_int64(__pyx_t_3); if (unlikely((__pyx_t_6 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(1, 2811, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_self->min_left = __pyx_t_6;

    /* "pandas/_libs/intervaltree.pxi":2812
 *         if left.size > 0:
 *             self.min_left = left.min()
 *             self.max_right = right.max()             # <<<<<<<<<<<<<<
 *         else:
 *             self.min_left = 0
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_right), __pyx_n_s_max); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2812, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2812, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_6 = __Pyx_PyInt_As_npy_int64(__pyx_t_3); if (unlikely((__pyx_t_6 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(1, 2812, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_self->max_right = __pyx_t_6;

    /* "pandas/_libs/intervaltree.pxi":2810
 *         # query on sub-nodes. If this node has size 0, query is cheap,
 *         # so these values don't matter.
 *         if left.size > 0:             # <<<<<<<<<<<<<<
 *             self.min_left = left.min()
 *             self.max_right = right.max()
 */
    goto __pyx_L3;
  }

  /* "pandas/_libs/intervaltree.pxi":2814
 *             self.max_right = right.max()
 *         else:
 *             self.min_left = 0             # <<<<<<<<<<<<<<
 *             self.max_right = 0
 * 
 */
  /*else*/ {
    __pyx_v_self->min_left = 0;

    /* "pandas/_libs/intervaltree.pxi":2815
 *         else:
 *             self.min_left = 0
 *             self.max_right = 0             # <<<<<<<<<<<<<<
 * 
 *         if self.n_elements <= leaf_size:
 */
    __pyx_v_self->max_right = 0;
  }
  __pyx_L3:;

  /* "pandas/_libs/intervaltree.pxi":2817
 *             self.max_right = 0
 * 
 *         if self.n_elements <= leaf_size:             # <<<<<<<<<<<<<<
 *             # make this a terminal (leaf) node
 *             self.is_leaf_node = True
 */
  __pyx_t_4 = ((__pyx_v_self->n_elements <= __pyx_v_leaf_size) != 0);
  if (__pyx_t_4) {

    /* "pandas/_libs/intervaltree.pxi":2819
 *         if self.n_elements <= leaf_size:
 *             # make this a terminal (leaf) node
 *             self.is_leaf_node = True             # <<<<<<<<<<<<<<
 *             self.left = left
 *             self.right = right
 */
    __pyx_v_self->is_leaf_node = 1;

    /* "pandas/_libs/intervaltree.pxi":2820
 *             # make this a terminal (leaf) node
 *             self.is_leaf_node = True
 *             self.left = left             # <<<<<<<<<<<<<<
 *             self.right = right
 *             self.indices = indices
 */
    __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(((PyObject *)__pyx_v_left), PyBUF_WRITABLE); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(1, 2820, __pyx_L1_error)
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->left, 0);
    __pyx_v_self->left = __pyx_t_7;
    __pyx_t_7.memview = NULL;
    __pyx_t_7.data = NULL;

    /* "pandas/_libs/intervaltree.pxi":2821
 *             self.is_leaf_node = True
 *             self.left = left
 *             self.right = right             # <<<<<<<<<<<<<<
 *             self.indices = indices
 *             self.n_center = 0
 */
    __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(((PyObject *)__pyx_v_right), PyBUF_WRITABLE); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(1, 2821, __pyx_L1_error)
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->right, 0);
    __pyx_v_self->right = __pyx_t_7;
    __pyx_t_7.memview = NULL;
    __pyx_t_7.data = NULL;

    /* "pandas/_libs/intervaltree.pxi":2822
 *             self.left = left
 *             self.right = right
 *             self.indices = indices             # <<<<<<<<<<<<<<
 *             self.n_center = 0
 *         else:
 */
    __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(((PyObject *)__pyx_v_indices), PyBUF_WRITABLE); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(1, 2822, __pyx_L1_error)
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->indices, 0);
    __pyx_v_self->indices = __pyx_t_7;
    __pyx_t_7.memview = NULL;
    __pyx_t_7.data = NULL;

    /* "pandas/_libs/intervaltree.pxi":2823
 *             self.right = right
 *             self.indices = indices
 *             self.n_center = 0             # <<<<<<<<<<<<<<
 *         else:
 *             # calculate a pivot so we can create child nodes
 */
    __pyx_v_self->n_center = 0;

    /* "pandas/_libs/intervaltree.pxi":2817
 *             self.max_right = 0
 * 
 *         if self.n_elements <= leaf_size:             # <<<<<<<<<<<<<<
 *             # make this a terminal (leaf) node
 *             self.is_leaf_node = True
 */
    goto __pyx_L4;
  }

  /* "pandas/_libs/intervaltree.pxi":2826
 *         else:
 *             # calculate a pivot so we can create child nodes
 *             self.is_leaf_node = False             # <<<<<<<<<<<<<<
 *             self.pivot = np.median(left / 2 + right / 2)
 *             left_set, right_set, center_set = self.classify_intervals(
 */
  /*else*/ {
    __pyx_v_self->is_leaf_node = 0;

    /* "pandas/_libs/intervaltree.pxi":2827
 *             # calculate a pivot so we can create child nodes
 *             self.is_leaf_node = False
 *             self.pivot = np.median(left / 2 + right / 2)             # <<<<<<<<<<<<<<
 *             left_set, right_set, center_set = self.classify_intervals(
 *                 left, right)
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2827, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_median); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 2827, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyNumber_Divide(((PyObject *)__pyx_v_left), __pyx_int_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2827, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_8 = __Pyx_PyNumber_Divide(((PyObject *)__pyx_v_right), __pyx_int_2); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 2827, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_9 = PyNumber_Add(__pyx_t_2, __pyx_t_8); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 2827, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_3 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_8, __pyx_t_9) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_9);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2827, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_6 = __Pyx_PyInt_As_npy_int64(__pyx_t_3); if (unlikely((__pyx_t_6 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(1, 2827, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_self->pivot = __pyx_t_6;

    /* "pandas/_libs/intervaltree.pxi":2829
 *             self.pivot = np.median(left / 2 + right / 2)
 *             left_set, right_set, center_set = self.classify_intervals(
 *                 left, right)             # <<<<<<<<<<<<<<
 * 
 *             self.left_node = self.new_child_node(left, right,
 */
    __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(((PyObject *)__pyx_v_left), PyBUF_WRITABLE); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(1, 2829, __pyx_L1_error)
    __pyx_t_10 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(((PyObject *)__pyx_v_right), PyBUF_WRITABLE); if (unlikely(!__pyx_t_10.memview)) __PYX_ERR(1, 2829, __pyx_L1_error)

    /* "pandas/_libs/intervaltree.pxi":2828
 *             self.is_leaf_node = False
 *             self.pivot = np.median(left / 2 + right / 2)
 *             left_set, right_set, center_set = self.classify_intervals(             # <<<<<<<<<<<<<<
 *                 left, right)
 * 
 */
    __pyx_t_3 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode *)__pyx_v_self->__pyx_vtab)->classify_intervals(__pyx_v_self, __pyx_t_7, __pyx_t_10); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2828, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __PYX_XDEC_MEMVIEW(&__pyx_t_7, 1);
    __pyx_t_7.memview = NULL;
    __pyx_t_7.data = NULL;
    __PYX_XDEC_MEMVIEW(&__pyx_t_10, 1);
    __pyx_t_10.memview = NULL;
    __pyx_t_10.data = NULL;
    if ((likely(PyTuple_CheckExact(__pyx_t_3))) || (PyList_CheckExact(__pyx_t_3))) {
      PyObject* sequence = __pyx_t_3;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 3)) {
        if (size > 3) __Pyx_RaiseTooManyValuesError(3);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(1, 2828, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_5 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_9 = PyTuple_GET_ITEM(sequence, 1); 
        __pyx_t_8 = PyTuple_GET_ITEM(sequence, 2); 
      } else {
        __pyx_t_5 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_9 = PyList_GET_ITEM(sequence, 1); 
        __pyx_t_8 = PyList_GET_ITEM(sequence, 2); 
      }
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_9);
      __Pyx_INCREF(__pyx_t_8);
      #else
      __pyx_t_5 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 2828, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_9 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 2828, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_8 = PySequence_ITEM(sequence, 2); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 2828, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      #endif
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_2 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2828, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_11 = Py_TYPE(__pyx_t_2)->tp_iternext;
      index = 0; __pyx_t_5 = __pyx_t_11(__pyx_t_2); if (unlikely(!__pyx_t_5)) goto __pyx_L5_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_5);
      index = 1; __pyx_t_9 = __pyx_t_11(__pyx_t_2); if (unlikely(!__pyx_t_9)) goto __pyx_L5_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_9);
      index = 2; __pyx_t_8 = __pyx_t_11(__pyx_t_2); if (unlikely(!__pyx_t_8)) goto __pyx_L5_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_8);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_11(__pyx_t_2), 3) < 0) __PYX_ERR(1, 2828, __pyx_L1_error)
      __pyx_t_11 = NULL;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      goto __pyx_L6_unpacking_done;
      __pyx_L5_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_11 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(1, 2828, __pyx_L1_error)
      __pyx_L6_unpacking_done:;
    }
    __pyx_v_left_set = __pyx_t_5;
    __pyx_t_5 = 0;
    __pyx_v_right_set = __pyx_t_9;
    __pyx_t_9 = 0;
    __pyx_v_center_set = __pyx_t_8;
    __pyx_t_8 = 0;

    /* "pandas/_libs/intervaltree.pxi":2832
 * 
 *             self.left_node = self.new_child_node(left, right,
 *                                                  indices, left_set)             # <<<<<<<<<<<<<<
 *             self.right_node = self.new_child_node(left, right,
 *                                                   indices, right_set)
 */
    if (!(likely(((__pyx_v_left_set) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_left_set, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(1, 2832, __pyx_L1_error)

    /* "pandas/_libs/intervaltree.pxi":2831
 *                 left, right)
 * 
 *             self.left_node = self.new_child_node(left, right,             # <<<<<<<<<<<<<<
 *                                                  indices, left_set)
 *             self.right_node = self.new_child_node(left, right,
 */
    __pyx_t_3 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode *)__pyx_v_self->__pyx_vtab)->new_child_node(__pyx_v_self, ((PyArrayObject *)__pyx_v_left), ((PyArrayObject *)__pyx_v_right), ((PyArrayObject *)__pyx_v_indices), ((PyArrayObject *)__pyx_v_left_set)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2831, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode))))) __PYX_ERR(1, 2831, __pyx_L1_error)
    __Pyx_GIVEREF(__pyx_t_3);
    __Pyx_GOTREF(__pyx_v_self->left_node);
    __Pyx_DECREF(((PyObject *)__pyx_v_self->left_node));
    __pyx_v_self->left_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode *)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "pandas/_libs/intervaltree.pxi":2834
 *                                                  indices, left_set)
 *             self.right_node = self.new_child_node(left, right,
 *                                                   indices, right_set)             # <<<<<<<<<<<<<<
 * 
 *             self.center_left_values, self.center_left_indices = \
 */
    if (!(likely(((__pyx_v_right_set) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_right_set, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(1, 2834, __pyx_L1_error)

    /* "pandas/_libs/intervaltree.pxi":2833
 *             self.left_node = self.new_child_node(left, right,
 *                                                  indices, left_set)
 *             self.right_node = self.new_child_node(left, right,             # <<<<<<<<<<<<<<
 *                                                   indices, right_set)
 * 
 */
    __pyx_t_3 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode *)__pyx_v_self->__pyx_vtab)->new_child_node(__pyx_v_self, ((PyArrayObject *)__pyx_v_left), ((PyArrayObject *)__pyx_v_right), ((PyArrayObject *)__pyx_v_indices), ((PyArrayObject *)__pyx_v_right_set)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2833, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode))))) __PYX_ERR(1, 2833, __pyx_L1_error)
    __Pyx_GIVEREF(__pyx_t_3);
    __Pyx_GOTREF(__pyx_v_self->right_node);
    __Pyx_DECREF(((PyObject *)__pyx_v_self->right_node));
    __pyx_v_self->right_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode *)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "pandas/_libs/intervaltree.pxi":2837
 * 
 *             self.center_left_values, self.center_left_indices = \
 *                 sort_values_and_indices(left, indices, center_set)             # <<<<<<<<<<<<<<
 *             self.center_right_values, self.center_right_indices = \
 *                 sort_values_and_indices(right, indices, center_set)
 */
    __pyx_t_3 = __pyx_f_6pandas_5_libs_8interval_sort_values_and_indices(((PyObject *)__pyx_v_left), ((PyObject *)__pyx_v_indices), __pyx_v_center_set); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2837, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if ((likely(PyTuple_CheckExact(__pyx_t_3))) || (PyList_CheckExact(__pyx_t_3))) {
      PyObject* sequence = __pyx_t_3;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(1, 2836, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_8 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_9 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_8 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_9 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_8);
      __Pyx_INCREF(__pyx_t_9);
      #else
      __pyx_t_8 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 2836, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_9 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 2836, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      #endif
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_5 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 2836, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_11 = Py_TYPE(__pyx_t_5)->tp_iternext;
      index = 0; __pyx_t_8 = __pyx_t_11(__pyx_t_5); if (unlikely(!__pyx_t_8)) goto __pyx_L7_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_8);
      index = 1; __pyx_t_9 = __pyx_t_11(__pyx_t_5); if (unlikely(!__pyx_t_9)) goto __pyx_L7_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_9);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_11(__pyx_t_5), 2) < 0) __PYX_ERR(1, 2836, __pyx_L1_error)
      __pyx_t_11 = NULL;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      goto __pyx_L8_unpacking_done;
      __pyx_L7_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_11 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(1, 2836, __pyx_L1_error)
      __pyx_L8_unpacking_done:;
    }

    /* "pandas/_libs/intervaltree.pxi":2836
 *                                                   indices, right_set)
 * 
 *             self.center_left_values, self.center_left_indices = \             # <<<<<<<<<<<<<<
 *                 sort_values_and_indices(left, indices, center_set)
 *             self.center_right_values, self.center_right_indices = \
 */
    __pyx_t_10 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_t_8, PyBUF_WRITABLE); if (unlikely(!__pyx_t_10.memview)) __PYX_ERR(1, 2836, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_t_9, PyBUF_WRITABLE); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(1, 2836, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->center_left_values, 0);
    __pyx_v_self->center_left_values = __pyx_t_10;
    __pyx_t_10.memview = NULL;
    __pyx_t_10.data = NULL;
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->center_left_indices, 0);
    __pyx_v_self->center_left_indices = __pyx_t_7;
    __pyx_t_7.memview = NULL;
    __pyx_t_7.data = NULL;

    /* "pandas/_libs/intervaltree.pxi":2839
 *                 sort_values_and_indices(left, indices, center_set)
 *             self.center_right_values, self.center_right_indices = \
 *                 sort_values_and_indices(right, indices, center_set)             # <<<<<<<<<<<<<<
 *             self.n_center = len(self.center_left_indices)
 * 
 */
    __pyx_t_3 = __pyx_f_6pandas_5_libs_8interval_sort_values_and_indices(((PyObject *)__pyx_v_right), ((PyObject *)__pyx_v_indices), __pyx_v_center_set); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2839, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if ((likely(PyTuple_CheckExact(__pyx_t_3))) || (PyList_CheckExact(__pyx_t_3))) {
      PyObject* sequence = __pyx_t_3;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(1, 2838, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_9 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_8 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_9 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_8 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_9);
      __Pyx_INCREF(__pyx_t_8);
      #else
      __pyx_t_9 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 2838, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_8 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 2838, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      #endif
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_5 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 2838, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_11 = Py_TYPE(__pyx_t_5)->tp_iternext;
      index = 0; __pyx_t_9 = __pyx_t_11(__pyx_t_5); if (unlikely(!__pyx_t_9)) goto __pyx_L9_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_9);
      index = 1; __pyx_t_8 = __pyx_t_11(__pyx_t_5); if (unlikely(!__pyx_t_8)) goto __pyx_L9_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_8);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_11(__pyx_t_5), 2) < 0) __PYX_ERR(1, 2838, __pyx_L1_error)
      __pyx_t_11 = NULL;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      goto __pyx_L10_unpacking_done;
      __pyx_L9_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_11 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(1, 2838, __pyx_L1_error)
      __pyx_L10_unpacking_done:;
    }

    /* "pandas/_libs/intervaltree.pxi":2838
 *             self.center_left_values, self.center_left_indices = \
 *                 sort_values_and_indices(left, indices, center_set)
 *             self.center_right_values, self.center_right_indices = \             # <<<<<<<<<<<<<<
 *                 sort_values_and_indices(right, indices, center_set)
 *             self.n_center = len(self.center_left_indices)
 */
    __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_t_9, PyBUF_WRITABLE); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(1, 2838, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_10 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_t_8, PyBUF_WRITABLE); if (unlikely(!__pyx_t_10.memview)) __PYX_ERR(1, 2838, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->center_right_values, 0);
    __pyx_v_self->center_right_values = __pyx_t_7;
    __pyx_t_7.memview = NULL;
    __pyx_t_7.data = NULL;
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->center_right_indices, 0);
    __pyx_v_self->center_right_indices = __pyx_t_10;
    __pyx_t_10.memview = NULL;
    __pyx_t_10.data = NULL;

    /* "pandas/_libs/intervaltree.pxi":2840
 *             self.center_right_values, self.center_right_indices = \
 *                 sort_values_and_indices(right, indices, center_set)
 *             self.n_center = len(self.center_left_indices)             # <<<<<<<<<<<<<<
 * 
 *     @cython.wraparound(False)
 */
    if (unlikely(!__pyx_v_self->center_left_indices.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(1, 2840, __pyx_L1_error)}
    __pyx_t_12 = __Pyx_MemoryView_Len(__pyx_v_self->center_left_indices); 
    __pyx_v_self->n_center = __pyx_t_12;
  }
  __pyx_L4:;

  /* "pandas/_libs/intervaltree.pxi":2798
 *         readonly bint is_leaf_node
 * 
 *     def __init__(self,             # <<<<<<<<<<<<<<
 *                  ndarray[int64_t, ndim=1] left,
 *                  ndarray[int64_t, ndim=1] right,
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __PYX_XDEC_MEMVIEW(&__pyx_t_7, 1);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __PYX_XDEC_MEMVIEW(&__pyx_t_10, 1);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indices.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_left.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_right.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("pandas._libs.interval.Int64ClosedNeitherIntervalNode.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indices.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_left.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_right.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF(__pyx_v_left_set);
  __Pyx_XDECREF(__pyx_v_right_set);
  __Pyx_XDECREF(__pyx_v_center_set);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":2844
 *     @cython.wraparound(False)
 *     @cython.boundscheck(False)
 *     cdef classify_intervals(self, int64_t[:] left, int64_t[:] right):             # <<<<<<<<<<<<<<
 *         """Classify the given intervals based upon whether they fall to the
 *         left, right, or overlap with this node's pivot.
 */

static PyObject *__pyx_f_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_classify_intervals(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode *__pyx_v_self, __Pyx_memviewslice __pyx_v_left, __Pyx_memviewslice __pyx_v_right) {
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_left_ind = 0;
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_right_ind = 0;
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_overlapping_ind = 0;
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __pyx_t_5numpy_int64_t __pyx_t_2;
  __pyx_t_5numpy_int64_t __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  int __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  __Pyx_RefNannySetupContext("classify_intervals", 0);

  /* "pandas/_libs/intervaltree.pxi":2852
 *             Py_ssize_t i
 * 
 *         left_ind = Int64Vector()             # <<<<<<<<<<<<<<
 *         right_ind = Int64Vector()
 *         overlapping_ind = Int64Vector()
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2852, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_left_ind = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":2853
 * 
 *         left_ind = Int64Vector()
 *         right_ind = Int64Vector()             # <<<<<<<<<<<<<<
 *         overlapping_ind = Int64Vector()
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2853, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_right_ind = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":2854
 *         left_ind = Int64Vector()
 *         right_ind = Int64Vector()
 *         overlapping_ind = Int64Vector()             # <<<<<<<<<<<<<<
 * 
 *         for i in range(self.n_elements):
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2854, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_overlapping_ind = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":2856
 *         overlapping_ind = Int64Vector()
 * 
 *         for i in range(self.n_elements):             # <<<<<<<<<<<<<<
 *             if right[i] <= self.pivot:
 *                 left_ind.append(i)
 */
  __pyx_t_2 = __pyx_v_self->n_elements;
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "pandas/_libs/intervaltree.pxi":2857
 * 
 *         for i in range(self.n_elements):
 *             if right[i] <= self.pivot:             # <<<<<<<<<<<<<<
 *                 left_ind.append(i)
 *             elif self.pivot <= left[i]:
 */
    __pyx_t_5 = __pyx_v_i;
    __pyx_t_6 = (((*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_right.data + __pyx_t_5 * __pyx_v_right.strides[0]) ))) <= __pyx_v_self->pivot) != 0);
    if (__pyx_t_6) {

      /* "pandas/_libs/intervaltree.pxi":2858
 *         for i in range(self.n_elements):
 *             if right[i] <= self.pivot:
 *                 left_ind.append(i)             # <<<<<<<<<<<<<<
 *             elif self.pivot <= left[i]:
 *                 right_ind.append(i)
 */
      ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_left_ind->__pyx_vtab)->append(__pyx_v_left_ind, __pyx_v_i);

      /* "pandas/_libs/intervaltree.pxi":2857
 * 
 *         for i in range(self.n_elements):
 *             if right[i] <= self.pivot:             # <<<<<<<<<<<<<<
 *                 left_ind.append(i)
 *             elif self.pivot <= left[i]:
 */
      goto __pyx_L5;
    }

    /* "pandas/_libs/intervaltree.pxi":2859
 *             if right[i] <= self.pivot:
 *                 left_ind.append(i)
 *             elif self.pivot <= left[i]:             # <<<<<<<<<<<<<<
 *                 right_ind.append(i)
 *             else:
 */
    __pyx_t_7 = __pyx_v_i;
    __pyx_t_6 = ((__pyx_v_self->pivot <= (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_left.data + __pyx_t_7 * __pyx_v_left.strides[0]) )))) != 0);
    if (__pyx_t_6) {

      /* "pandas/_libs/intervaltree.pxi":2860
 *                 left_ind.append(i)
 *             elif self.pivot <= left[i]:
 *                 right_ind.append(i)             # <<<<<<<<<<<<<<
 *             else:
 *                 overlapping_ind.append(i)
 */
      ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_right_ind->__pyx_vtab)->append(__pyx_v_right_ind, __pyx_v_i);

      /* "pandas/_libs/intervaltree.pxi":2859
 *             if right[i] <= self.pivot:
 *                 left_ind.append(i)
 *             elif self.pivot <= left[i]:             # <<<<<<<<<<<<<<
 *                 right_ind.append(i)
 *             else:
 */
      goto __pyx_L5;
    }

    /* "pandas/_libs/intervaltree.pxi":2862
 *                 right_ind.append(i)
 *             else:
 *                 overlapping_ind.append(i)             # <<<<<<<<<<<<<<
 * 
 *         return (left_ind.to_array(),
 */
    /*else*/ {
      ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_overlapping_ind->__pyx_vtab)->append(__pyx_v_overlapping_ind, __pyx_v_i);
    }
    __pyx_L5:;
  }

  /* "pandas/_libs/intervaltree.pxi":2864
 *                 overlapping_ind.append(i)
 * 
 *         return (left_ind.to_array(),             # <<<<<<<<<<<<<<
 *                 right_ind.to_array(),
 *                 overlapping_ind.to_array())
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_left_ind->__pyx_vtab)->to_array(__pyx_v_left_ind, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2864, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "pandas/_libs/intervaltree.pxi":2865
 * 
 *         return (left_ind.to_array(),
 *                 right_ind.to_array(),             # <<<<<<<<<<<<<<
 *                 overlapping_ind.to_array())
 * 
 */
  __pyx_t_8 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_right_ind->__pyx_vtab)->to_array(__pyx_v_right_ind, 0); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 2865, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);

  /* "pandas/_libs/intervaltree.pxi":2866
 *         return (left_ind.to_array(),
 *                 right_ind.to_array(),
 *                 overlapping_ind.to_array())             # <<<<<<<<<<<<<<
 * 
 *     cdef new_child_node(self,
 */
  __pyx_t_9 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_overlapping_ind->__pyx_vtab)->to_array(__pyx_v_overlapping_ind, 0); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 2866, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);

  /* "pandas/_libs/intervaltree.pxi":2864
 *                 overlapping_ind.append(i)
 * 
 *         return (left_ind.to_array(),             # <<<<<<<<<<<<<<
 *                 right_ind.to_array(),
 *                 overlapping_ind.to_array())
 */
  __pyx_t_10 = PyTuple_New(3); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 2864, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_t_8);
  __Pyx_GIVEREF(__pyx_t_9);
  PyTuple_SET_ITEM(__pyx_t_10, 2, __pyx_t_9);
  __pyx_t_1 = 0;
  __pyx_t_8 = 0;
  __pyx_t_9 = 0;
  __pyx_r = __pyx_t_10;
  __pyx_t_10 = 0;
  goto __pyx_L0;

  /* "pandas/_libs/intervaltree.pxi":2844
 *     @cython.wraparound(False)
 *     @cython.boundscheck(False)
 *     cdef classify_intervals(self, int64_t[:] left, int64_t[:] right):             # <<<<<<<<<<<<<<
 *         """Classify the given intervals based upon whether they fall to the
 *         left, right, or overlap with this node's pivot.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("pandas._libs.interval.Int64ClosedNeitherIntervalNode.classify_intervals", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_left_ind);
  __Pyx_XDECREF((PyObject *)__pyx_v_right_ind);
  __Pyx_XDECREF((PyObject *)__pyx_v_overlapping_ind);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":2868
 *                 overlapping_ind.to_array())
 * 
 *     cdef new_child_node(self,             # <<<<<<<<<<<<<<
 *                         ndarray[int64_t, ndim=1] left,
 *                         ndarray[int64_t, ndim=1] right,
 */

static PyObject *__pyx_f_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_new_child_node(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode *__pyx_v_self, PyArrayObject *__pyx_v_left, PyArrayObject *__pyx_v_right, PyArrayObject *__pyx_v_indices, PyArrayObject *__pyx_v_subset) {
  __Pyx_LocalBuf_ND __pyx_pybuffernd_indices;
  __Pyx_Buffer __pyx_pybuffer_indices;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_left;
  __Pyx_Buffer __pyx_pybuffer_left;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_right;
  __Pyx_Buffer __pyx_pybuffer_right;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_subset;
  __Pyx_Buffer __pyx_pybuffer_subset;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyArrayObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyArrayObject *__pyx_t_7 = NULL;
  PyArrayObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  __Pyx_RefNannySetupContext("new_child_node", 0);
  __Pyx_INCREF((PyObject *)__pyx_v_left);
  __Pyx_INCREF((PyObject *)__pyx_v_right);
  __Pyx_INCREF((PyObject *)__pyx_v_indices);
  __pyx_pybuffer_left.pybuffer.buf = NULL;
  __pyx_pybuffer_left.refcount = 0;
  __pyx_pybuffernd_left.data = NULL;
  __pyx_pybuffernd_left.rcbuffer = &__pyx_pybuffer_left;
  __pyx_pybuffer_right.pybuffer.buf = NULL;
  __pyx_pybuffer_right.refcount = 0;
  __pyx_pybuffernd_right.data = NULL;
  __pyx_pybuffernd_right.rcbuffer = &__pyx_pybuffer_right;
  __pyx_pybuffer_indices.pybuffer.buf = NULL;
  __pyx_pybuffer_indices.refcount = 0;
  __pyx_pybuffernd_indices.data = NULL;
  __pyx_pybuffernd_indices.rcbuffer = &__pyx_pybuffer_indices;
  __pyx_pybuffer_subset.pybuffer.buf = NULL;
  __pyx_pybuffer_subset.refcount = 0;
  __pyx_pybuffernd_subset.data = NULL;
  __pyx_pybuffernd_subset.rcbuffer = &__pyx_pybuffer_subset;
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_left.rcbuffer->pybuffer, (PyObject*)__pyx_v_left, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 2868, __pyx_L1_error)
  }
  __pyx_pybuffernd_left.diminfo[0].strides = __pyx_pybuffernd_left.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_left.diminfo[0].shape = __pyx_pybuffernd_left.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_right.rcbuffer->pybuffer, (PyObject*)__pyx_v_right, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 2868, __pyx_L1_error)
  }
  __pyx_pybuffernd_right.diminfo[0].strides = __pyx_pybuffernd_right.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_right.diminfo[0].shape = __pyx_pybuffernd_right.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_indices.rcbuffer->pybuffer, (PyObject*)__pyx_v_indices, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 2868, __pyx_L1_error)
  }
  __pyx_pybuffernd_indices.diminfo[0].strides = __pyx_pybuffernd_indices.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_indices.diminfo[0].shape = __pyx_pybuffernd_indices.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_subset.rcbuffer->pybuffer, (PyObject*)__pyx_v_subset, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 2868, __pyx_L1_error)
  }
  __pyx_pybuffernd_subset.diminfo[0].strides = __pyx_pybuffernd_subset.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_subset.diminfo[0].shape = __pyx_pybuffernd_subset.rcbuffer->pybuffer.shape[0];

  /* "pandas/_libs/intervaltree.pxi":2875
 *         """Create a new child node.
 *         """
 *         left = take(left, subset)             # <<<<<<<<<<<<<<
 *         right = take(right, subset)
 *         indices = take(indices, subset)
 */
  __pyx_t_1 = __pyx_f_6pandas_5_libs_8interval_take(((PyArrayObject *)__pyx_v_left), ((PyArrayObject *)__pyx_v_subset)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2875, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(1, 2875, __pyx_L1_error)
  __pyx_t_2 = ((PyArrayObject *)__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_left.rcbuffer->pybuffer);
    __pyx_t_3 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_left.rcbuffer->pybuffer, (PyObject*)__pyx_t_2, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack);
    if (unlikely(__pyx_t_3 < 0)) {
      PyErr_Fetch(&__pyx_t_4, &__pyx_t_5, &__pyx_t_6);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_left.rcbuffer->pybuffer, (PyObject*)__pyx_v_left, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_4); Py_XDECREF(__pyx_t_5); Py_XDECREF(__pyx_t_6);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_4, __pyx_t_5, __pyx_t_6);
      }
      __pyx_t_4 = __pyx_t_5 = __pyx_t_6 = 0;
    }
    __pyx_pybuffernd_left.diminfo[0].strides = __pyx_pybuffernd_left.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_left.diminfo[0].shape = __pyx_pybuffernd_left.rcbuffer->pybuffer.shape[0];
    if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(1, 2875, __pyx_L1_error)
  }
  __pyx_t_2 = 0;
  __Pyx_DECREF_SET(__pyx_v_left, ((PyArrayObject *)__pyx_t_1));
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":2876
 *         """
 *         left = take(left, subset)
 *         right = take(right, subset)             # <<<<<<<<<<<<<<
 *         indices = take(indices, subset)
 *         return Int64ClosedNeitherIntervalNode(
 */
  __pyx_t_1 = __pyx_f_6pandas_5_libs_8interval_take(((PyArrayObject *)__pyx_v_right), ((PyArrayObject *)__pyx_v_subset)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2876, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(1, 2876, __pyx_L1_error)
  __pyx_t_7 = ((PyArrayObject *)__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_right.rcbuffer->pybuffer);
    __pyx_t_3 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_right.rcbuffer->pybuffer, (PyObject*)__pyx_t_7, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack);
    if (unlikely(__pyx_t_3 < 0)) {
      PyErr_Fetch(&__pyx_t_6, &__pyx_t_5, &__pyx_t_4);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_right.rcbuffer->pybuffer, (PyObject*)__pyx_v_right, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_6); Py_XDECREF(__pyx_t_5); Py_XDECREF(__pyx_t_4);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_6, __pyx_t_5, __pyx_t_4);
      }
      __pyx_t_6 = __pyx_t_5 = __pyx_t_4 = 0;
    }
    __pyx_pybuffernd_right.diminfo[0].strides = __pyx_pybuffernd_right.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_right.diminfo[0].shape = __pyx_pybuffernd_right.rcbuffer->pybuffer.shape[0];
    if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(1, 2876, __pyx_L1_error)
  }
  __pyx_t_7 = 0;
  __Pyx_DECREF_SET(__pyx_v_right, ((PyArrayObject *)__pyx_t_1));
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":2877
 *         left = take(left, subset)
 *         right = take(right, subset)
 *         indices = take(indices, subset)             # <<<<<<<<<<<<<<
 *         return Int64ClosedNeitherIntervalNode(
 *             left, right, indices, self.leaf_size)
 */
  __pyx_t_1 = __pyx_f_6pandas_5_libs_8interval_take(((PyArrayObject *)__pyx_v_indices), ((PyArrayObject *)__pyx_v_subset)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2877, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(1, 2877, __pyx_L1_error)
  __pyx_t_8 = ((PyArrayObject *)__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indices.rcbuffer->pybuffer);
    __pyx_t_3 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_indices.rcbuffer->pybuffer, (PyObject*)__pyx_t_8, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack);
    if (unlikely(__pyx_t_3 < 0)) {
      PyErr_Fetch(&__pyx_t_4, &__pyx_t_5, &__pyx_t_6);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_indices.rcbuffer->pybuffer, (PyObject*)__pyx_v_indices, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_4); Py_XDECREF(__pyx_t_5); Py_XDECREF(__pyx_t_6);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_4, __pyx_t_5, __pyx_t_6);
      }
      __pyx_t_4 = __pyx_t_5 = __pyx_t_6 = 0;
    }
    __pyx_pybuffernd_indices.diminfo[0].strides = __pyx_pybuffernd_indices.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_indices.diminfo[0].shape = __pyx_pybuffernd_indices.rcbuffer->pybuffer.shape[0];
    if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(1, 2877, __pyx_L1_error)
  }
  __pyx_t_8 = 0;
  __Pyx_DECREF_SET(__pyx_v_indices, ((PyArrayObject *)__pyx_t_1));
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":2878
 *         right = take(right, subset)
 *         indices = take(indices, subset)
 *         return Int64ClosedNeitherIntervalNode(             # <<<<<<<<<<<<<<
 *             left, right, indices, self.leaf_size)
 * 
 */
  __Pyx_XDECREF(__pyx_r);

  /* "pandas/_libs/intervaltree.pxi":2879
 *         indices = take(indices, subset)
 *         return Int64ClosedNeitherIntervalNode(
 *             left, right, indices, self.leaf_size)             # <<<<<<<<<<<<<<
 * 
 *     @cython.wraparound(False)
 */
  __pyx_t_1 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->leaf_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2879, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "pandas/_libs/intervaltree.pxi":2878
 *         right = take(right, subset)
 *         indices = take(indices, subset)
 *         return Int64ClosedNeitherIntervalNode(             # <<<<<<<<<<<<<<
 *             left, right, indices, self.leaf_size)
 * 
 */
  __pyx_t_9 = PyTuple_New(4); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 2878, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_INCREF(((PyObject *)__pyx_v_left));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_left));
  PyTuple_SET_ITEM(__pyx_t_9, 0, ((PyObject *)__pyx_v_left));
  __Pyx_INCREF(((PyObject *)__pyx_v_right));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_right));
  PyTuple_SET_ITEM(__pyx_t_9, 1, ((PyObject *)__pyx_v_right));
  __Pyx_INCREF(((PyObject *)__pyx_v_indices));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_indices));
  PyTuple_SET_ITEM(__pyx_t_9, 2, ((PyObject *)__pyx_v_indices));
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_9, 3, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode), __pyx_t_9, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2878, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pandas/_libs/intervaltree.pxi":2868
 *                 overlapping_ind.to_array())
 * 
 *     cdef new_child_node(self,             # <<<<<<<<<<<<<<
 *                         ndarray[int64_t, ndim=1] left,
 *                         ndarray[int64_t, ndim=1] right,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_9);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indices.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_left.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_right.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_subset.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("pandas._libs.interval.Int64ClosedNeitherIntervalNode.new_child_node", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indices.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_left.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_right.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_subset.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF((PyObject *)__pyx_v_left);
  __Pyx_XDECREF((PyObject *)__pyx_v_right);
  __Pyx_XDECREF((PyObject *)__pyx_v_indices);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":2884
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_2query[] = "Recursively query this node and its sub-nodes for intervals that\n        overlap with the query point.\n        ";
static PyMethodDef __pyx_mdef_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_3query = {"query", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_3query, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_2query};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_args = 0;
  PyObject *__pyx_v_kwargs = 0;
  CYTHON_UNUSED PyObject *__pyx_v_defaults = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fused_cpdef (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_args,&__pyx_n_s_kwargs,&__pyx_n_s_defaults,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_args)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_kwargs)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fused_cpdef", 1, 3, 3, 1); __PYX_ERR(1, 2884, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_defaults)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fused_cpdef", 1, 3, 3, 2); __PYX_ERR(1, 2884, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fused_cpdef") < 0)) __PYX_ERR(1, 2884, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_args = values[0];
    __pyx_v_kwargs = values[1];
    __pyx_v_defaults = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fused_cpdef", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 2884, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Int64ClosedNeitherIntervalNode.__pyx_fused_cpdef", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_2query(((struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode *)__pyx_v_signatures), __pyx_v_args, __pyx_v_kwargs, __pyx_v_defaults);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_2query(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode *__pyx_v_signatures, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs, CYTHON_UNUSED PyObject *__pyx_v_defaults) {
  PyObject *__pyx_v_dest_sig = NULL;
  Py_ssize_t __pyx_v_i;
  CYTHON_UNUSED PyTypeObject *__pyx_v_ndarray = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_v_itemsize;
  PyObject *__pyx_v_arg = NULL;
  PyObject *__pyx_v_candidates = NULL;
  PyObject *__pyx_v_sig = NULL;
  int __pyx_v_match_found;
  PyObject *__pyx_v_src_sig = NULL;
  PyObject *__pyx_v_dst_type = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  Py_ssize_t __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  Py_ssize_t __pyx_t_13;
  Py_ssize_t __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  int __pyx_t_16;
  __Pyx_RefNannySetupContext("query", 0);
  __Pyx_INCREF(__pyx_v_kwargs);
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2884, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  PyList_SET_ITEM(__pyx_t_1, 0, Py_None);
  __pyx_v_dest_sig = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_3 = (__pyx_v_kwargs != Py_None);
  __pyx_t_4 = (__pyx_t_3 != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_kwargs); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(1, 2884, __pyx_L1_error)
  __pyx_t_3 = ((!__pyx_t_4) != 0);
  __pyx_t_2 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {
    __Pyx_INCREF(Py_None);
    __Pyx_DECREF_SET(__pyx_v_kwargs, Py_None);
  }
  __pyx_t_1 = ((PyObject *)__Pyx_ImportNumPyArrayTypeIfAvailable()); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2884, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_ndarray = ((PyTypeObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_v_itemsize = -1L;
  if (unlikely(__pyx_v_args == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 2884, __pyx_L1_error)
  }
  __pyx_t_5 = PyTuple_GET_SIZE(((PyObject*)__pyx_v_args)); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(1, 2884, __pyx_L1_error)
  __pyx_t_2 = ((2 < __pyx_t_5) != 0);
  if (__pyx_t_2) {
    if (unlikely(__pyx_v_args == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 2884, __pyx_L1_error)
    }
    __pyx_t_1 = PyTuple_GET_ITEM(((PyObject*)__pyx_v_args), 2);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v_arg = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L6;
  }
  __pyx_t_3 = (__pyx_v_kwargs != Py_None);
  __pyx_t_4 = (__pyx_t_3 != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L7_bool_binop_done;
  }
  if (unlikely(__pyx_v_kwargs == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(1, 2884, __pyx_L1_error)
  }
  __pyx_t_4 = (__Pyx_PyDict_ContainsTF(__pyx_n_s_point, ((PyObject*)__pyx_v_kwargs), Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(1, 2884, __pyx_L1_error)
  __pyx_t_3 = (__pyx_t_4 != 0);
  __pyx_t_2 = __pyx_t_3;
  __pyx_L7_bool_binop_done:;
  if (__pyx_t_2) {
    if (unlikely(__pyx_v_kwargs == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 2884, __pyx_L1_error)
    }
    __pyx_t_1 = __Pyx_PyDict_GetItem(((PyObject*)__pyx_v_kwargs), __pyx_n_s_point); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2884, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_arg = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L6;
  }
  /*else*/ {
    if (unlikely(__pyx_v_args == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
      __PYX_ERR(1, 2884, __pyx_L1_error)
    }
    __pyx_t_5 = PyTuple_GET_SIZE(((PyObject*)__pyx_v_args)); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(1, 2884, __pyx_L1_error)
    __pyx_t_1 = PyInt_FromSsize_t(__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2884, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_6 = PyTuple_New(3); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 2884, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_INCREF(__pyx_int_3);
    __Pyx_GIVEREF(__pyx_int_3);
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_int_3);
    __Pyx_INCREF(__pyx_n_s_s);
    __Pyx_GIVEREF(__pyx_n_s_s);
    PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_n_s_s);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Expected_at_least_d_argument_s_g, __pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2884, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_1); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 2884, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(1, 2884, __pyx_L1_error)
  }
  __pyx_L6:;
  while (1) {
    __pyx_t_2 = PyFloat_Check(__pyx_v_arg); 
    __pyx_t_3 = (__pyx_t_2 != 0);
    if (__pyx_t_3) {
      if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, __pyx_n_s_float64_t, long, 1, __Pyx_PyInt_From_long, 1, 0, 0) < 0)) __PYX_ERR(1, 2884, __pyx_L1_error)
      goto __pyx_L10_break;
    }
    __pyx_t_2 = PyInt_Check(__pyx_v_arg); 
    __pyx_t_4 = (__pyx_t_2 != 0);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_3 = __pyx_t_4;
      goto __pyx_L13_bool_binop_done;
    }
    __pyx_t_4 = PyLong_Check(__pyx_v_arg); 
    __pyx_t_2 = (__pyx_t_4 != 0);
    __pyx_t_3 = __pyx_t_2;
    __pyx_L13_bool_binop_done:;
    __pyx_t_2 = (__pyx_t_3 != 0);
    if (__pyx_t_2) {
      if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, __pyx_n_s_int64_t, long, 1, __Pyx_PyInt_From_long, 1, 0, 0) < 0)) __PYX_ERR(1, 2884, __pyx_L1_error)
      goto __pyx_L10_break;
    }
    if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, Py_None, long, 1, __Pyx_PyInt_From_long, 1, 0, 0) < 0)) __PYX_ERR(1, 2884, __pyx_L1_error)
    goto __pyx_L10_break;
  }
  __pyx_L10_break:;
  __pyx_t_6 = PyList_New(0); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 2884, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_v_candidates = ((PyObject*)__pyx_t_6);
  __pyx_t_6 = 0;
  __pyx_t_5 = 0;
  if (unlikely(((PyObject *)__pyx_v_signatures) == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(1, 2884, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_dict_iterator(((PyObject*)__pyx_v_signatures), 1, ((PyObject *)NULL), (&__pyx_t_7), (&__pyx_t_8)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2884, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __pyx_t_6 = __pyx_t_1;
  __pyx_t_1 = 0;
  while (1) {
    __pyx_t_9 = __Pyx_dict_iter_next(__pyx_t_6, __pyx_t_7, &__pyx_t_5, &__pyx_t_1, NULL, NULL, __pyx_t_8);
    if (unlikely(__pyx_t_9 == 0)) break;
    if (unlikely(__pyx_t_9 == -1)) __PYX_ERR(1, 2884, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_sig, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_v_match_found = 0;
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_v_sig, __pyx_n_s_strip); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 2884, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __pyx_t_12 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_11))) {
      __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_11);
      if (likely(__pyx_t_12)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
        __Pyx_INCREF(__pyx_t_12);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_11, function);
      }
    }
    __pyx_t_10 = (__pyx_t_12) ? __Pyx_PyObject_Call2Args(__pyx_t_11, __pyx_t_12, __pyx_kp_s__13) : __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_kp_s__13);
    __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
    if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 2884, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_split); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 2884, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_t_10 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_11))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_11);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_11, function);
      }
    }
    __pyx_t_1 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_11, __pyx_t_10, __pyx_kp_s__14) : __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_kp_s__14);
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2884, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_XDECREF_SET(__pyx_v_src_sig, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_13 = PyList_GET_SIZE(__pyx_v_dest_sig); if (unlikely(__pyx_t_13 == ((Py_ssize_t)-1))) __PYX_ERR(1, 2884, __pyx_L1_error)
    __pyx_t_14 = __pyx_t_13;
    for (__pyx_t_15 = 0; __pyx_t_15 < __pyx_t_14; __pyx_t_15+=1) {
      __pyx_v_i = __pyx_t_15;
      __pyx_t_1 = PyList_GET_ITEM(__pyx_v_dest_sig, __pyx_v_i);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_XDECREF_SET(__pyx_v_dst_type, __pyx_t_1);
      __pyx_t_1 = 0;
      __pyx_t_2 = (__pyx_v_dst_type != Py_None);
      __pyx_t_3 = (__pyx_t_2 != 0);
      if (__pyx_t_3) {
        __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_src_sig, __pyx_v_i, Py_ssize_t, 1, PyInt_FromSsize_t, 0, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2884, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_11 = PyObject_RichCompare(__pyx_t_1, __pyx_v_dst_type, Py_EQ); __Pyx_XGOTREF(__pyx_t_11); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 2884, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_11); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(1, 2884, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        if (__pyx_t_3) {
          __pyx_v_match_found = 1;
          goto __pyx_L20;
        }
        /*else*/ {
          __pyx_v_match_found = 0;
          goto __pyx_L18_break;
        }
        __pyx_L20:;
      }
    }
    __pyx_L18_break:;
    __pyx_t_3 = (__pyx_v_match_found != 0);
    if (__pyx_t_3) {
      __pyx_t_16 = __Pyx_PyList_Append(__pyx_v_candidates, __pyx_v_sig); if (unlikely(__pyx_t_16 == ((int)-1))) __PYX_ERR(1, 2884, __pyx_L1_error)
    }
  }
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_3 = (PyList_GET_SIZE(__pyx_v_candidates) != 0);
  __pyx_t_2 = ((!__pyx_t_3) != 0);
  if (__pyx_t_2) {
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__15, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 2884, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(1, 2884, __pyx_L1_error)
  }
  __pyx_t_7 = PyList_GET_SIZE(__pyx_v_candidates); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1))) __PYX_ERR(1, 2884, __pyx_L1_error)
  __pyx_t_2 = ((__pyx_t_7 > 1) != 0);
  if (__pyx_t_2) {
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__16, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 2884, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(1, 2884, __pyx_L1_error)
  }
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    if (unlikely(((PyObject *)__pyx_v_signatures) == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 2884, __pyx_L1_error)
    }
    __pyx_t_6 = __Pyx_PyDict_GetItem(((PyObject*)__pyx_v_signatures), PyList_GET_ITEM(__pyx_v_candidates, 0)); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 2884, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_r = __pyx_t_6;
    __pyx_t_6 = 0;
    goto __pyx_L0;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("pandas._libs.interval.Int64ClosedNeitherIntervalNode.__pyx_fused_cpdef", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_dest_sig);
  __Pyx_XDECREF(__pyx_v_ndarray);
  __Pyx_XDECREF(__pyx_v_arg);
  __Pyx_XDECREF(__pyx_v_candidates);
  __Pyx_XDECREF(__pyx_v_sig);
  __Pyx_XDECREF(__pyx_v_src_sig);
  __Pyx_XDECREF(__pyx_v_dst_type);
  __Pyx_XDECREF(__pyx_v_kwargs);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_9__pyx_fuse_0query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_fuse_0__pyx_f_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float64_t __pyx_v_point, int __pyx_skip_dispatch) {
  __Pyx_memviewslice __pyx_v_indices = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_values = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __pyx_t_5numpy_int64_t __pyx_t_9;
  __pyx_t_5numpy_int64_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  int __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  __Pyx_memviewslice __pyx_t_16 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_17;
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  __Pyx_RefNannySetupContext("__pyx_fuse_0query", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pyx_fuse_0query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2884, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_9__pyx_fuse_0query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = PyFloat_FromDouble(__pyx_v_point); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2884, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2884, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2884, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 2884, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_5) {
            __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2884, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2884, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_3query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_4 = PyFloat_FromDouble(__pyx_v_point); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 2884, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_7 = __pyx_t_1; __pyx_t_3 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2884, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2884, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 2884, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (__pyx_t_3) {
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_t_4);
          __pyx_t_4 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2884, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pandas/_libs/intervaltree.pxi":2893
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
  __pyx_t_8 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_8) {

    /* "pandas/_libs/intervaltree.pxi":2897
 *             # continue the binary tree structure. Instead, we use linear
 *             # search.
 *             for i in range(self.n_elements):             # <<<<<<<<<<<<<<
 *                 if self.left[i] < point < self.right[i]:
 *                     result.append(self.indices[i])
 */
    __pyx_t_9 = __pyx_v_self->n_elements;
    __pyx_t_10 = __pyx_t_9;
    for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
      __pyx_v_i = __pyx_t_11;

      /* "pandas/_libs/intervaltree.pxi":2898
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point < self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      __pyx_t_12 = __pyx_v_i;
      __pyx_t_8 = ((*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->left.data + __pyx_t_12 * __pyx_v_self->left.strides[0]) ))) < __pyx_v_point);
      if (__pyx_t_8) {
        __pyx_t_13 = __pyx_v_i;
        __pyx_t_8 = (__pyx_v_point < (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->right.data + __pyx_t_13 * __pyx_v_self->right.strides[0]) ))));
      }
      __pyx_t_14 = (__pyx_t_8 != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":2899
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point < self.right[i]:
 *                     result.append(self.indices[i])             # <<<<<<<<<<<<<<
 *         else:
 *             # There are child nodes. Based on comparing our query to the pivot,
 */
        __pyx_t_15 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->indices.data + __pyx_t_15 * __pyx_v_self->indices.strides[0]) ))));

        /* "pandas/_libs/intervaltree.pxi":2898
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point < self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      }
    }

    /* "pandas/_libs/intervaltree.pxi":2893
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
    goto __pyx_L3;
  }

  /* "pandas/_libs/intervaltree.pxi":2903
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
  /*else*/ {
    __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":2904
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:
 *                 values = self.center_left_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 */
      __pyx_t_16 = __pyx_v_self->center_left_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":2905
 *             if point < self.pivot:
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:
 */
      __pyx_t_16 = __pyx_v_self->center_left_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_indices = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":2906
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):             # <<<<<<<<<<<<<<
 *                     if not values[i] < point:
 *                         break
 */
      __pyx_t_9 = __pyx_v_self->n_center;
      __pyx_t_10 = __pyx_t_9;
      for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":2907
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_17 = __pyx_v_i;
        __pyx_t_14 = ((!(((*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_17 * __pyx_v_values.strides[0]) ))) < __pyx_v_point) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":2908
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:
 */
          goto __pyx_L9_break;

          /* "pandas/_libs/intervaltree.pxi":2907
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":2909
 *                     if not values[i] < point:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 */
        __pyx_t_18 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_18 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L9_break:;

      /* "pandas/_libs/intervaltree.pxi":2910
 *                         break
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->left_node->max_right) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":2911
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)             # <<<<<<<<<<<<<<
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode *)__pyx_v_self->left_node->__pyx_vtab)->__pyx_fuse_0query(__pyx_v_self->left_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2911, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":2910
 *                         break
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":2903
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":2912
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
    __pyx_t_14 = ((__pyx_v_point > __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":2913
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 *                 values = self.center_right_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 */
      __pyx_t_16 = __pyx_v_self->center_right_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":2914
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:
 */
      __pyx_t_16 = __pyx_v_self->center_right_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_indices = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":2915
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):             # <<<<<<<<<<<<<<
 *                     if not point < values[i]:
 *                         break
 */
      for (__pyx_t_11 = (__pyx_v_self->n_center - 1); __pyx_t_11 > -1L; __pyx_t_11-=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":2916
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_19 = __pyx_v_i;
        __pyx_t_14 = ((!((__pyx_v_point < (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_19 * __pyx_v_values.strides[0]) )))) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":2917
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:
 */
          goto __pyx_L13_break;

          /* "pandas/_libs/intervaltree.pxi":2916
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":2918
 *                     if not point < values[i]:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if self.right_node.min_left < point:
 *                     self.right_node.query(result, point)
 */
        __pyx_t_20 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_20 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L13_break:;

      /* "pandas/_libs/intervaltree.pxi":2919
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      __pyx_t_14 = ((__pyx_v_self->right_node->min_left < __pyx_v_point) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":2920
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:
 *                     self.right_node.query(result, point)             # <<<<<<<<<<<<<<
 *             else:
 *                 result.extend(self.center_left_indices)
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode *)__pyx_v_self->right_node->__pyx_vtab)->__pyx_fuse_0query(__pyx_v_self->right_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2920, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":2919
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":2912
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":2922
 *                     self.right_node.query(result, point)
 *             else:
 *                 result.extend(self.center_left_indices)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
    /*else*/ {
      __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->extend(__pyx_v_result, __pyx_v_self->center_left_indices); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2922, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_L7:;
  }
  __pyx_L3:;

  /* "pandas/_libs/intervaltree.pxi":2884
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __PYX_XDEC_MEMVIEW(&__pyx_t_16, 1);
  __Pyx_AddTraceback("pandas._libs.interval.Int64ClosedNeitherIntervalNode.query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_indices, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_values, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_9__pyx_fuse_0query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_0__pyx_mdef_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_9__pyx_fuse_0query = {"__pyx_fuse_0query", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_9__pyx_fuse_0query, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_2query};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_9__pyx_fuse_0query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result = 0;
  __pyx_t_5numpy_float64_t __pyx_v_point;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fuse_0query (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_result,&__pyx_n_s_point,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_result)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_point)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_0query", 1, 2, 2, 1); __PYX_ERR(1, 2884, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_0query") < 0)) __PYX_ERR(1, 2884, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_result = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)values[0]);
    __pyx_v_point = __pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_point == ((npy_float64)-1)) && PyErr_Occurred())) __PYX_ERR(1, 2884, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_0query", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 2884, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Int64ClosedNeitherIntervalNode.__pyx_fuse_0query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_result), __pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector, 1, "result", 0))) __PYX_ERR(1, 2884, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_8__pyx_fuse_0query(((struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode *)__pyx_v_self), __pyx_v_result, __pyx_v_point);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_8__pyx_fuse_0query(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float64_t __pyx_v_point) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__pyx_fuse_0query", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_fuse_0__pyx_f_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_query(__pyx_v_self, __pyx_v_result, __pyx_v_point, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2884, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Int64ClosedNeitherIntervalNode.__pyx_fuse_0query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_11__pyx_fuse_1query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_fuse_1__pyx_f_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float32_t __pyx_v_point, int __pyx_skip_dispatch) {
  __Pyx_memviewslice __pyx_v_indices = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_values = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __pyx_t_5numpy_int64_t __pyx_t_9;
  __pyx_t_5numpy_int64_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  int __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  __Pyx_memviewslice __pyx_t_16 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_17;
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  __Pyx_RefNannySetupContext("__pyx_fuse_1query", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pyx_fuse_1query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2884, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_11__pyx_fuse_1query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = PyFloat_FromDouble(__pyx_v_point); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2884, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2884, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2884, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 2884, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_5) {
            __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2884, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2884, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_3query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_4 = PyFloat_FromDouble(__pyx_v_point); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 2884, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_7 = __pyx_t_1; __pyx_t_3 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2884, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2884, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 2884, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (__pyx_t_3) {
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_t_4);
          __pyx_t_4 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2884, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pandas/_libs/intervaltree.pxi":2893
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
  __pyx_t_8 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_8) {

    /* "pandas/_libs/intervaltree.pxi":2897
 *             # continue the binary tree structure. Instead, we use linear
 *             # search.
 *             for i in range(self.n_elements):             # <<<<<<<<<<<<<<
 *                 if self.left[i] < point < self.right[i]:
 *                     result.append(self.indices[i])
 */
    __pyx_t_9 = __pyx_v_self->n_elements;
    __pyx_t_10 = __pyx_t_9;
    for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
      __pyx_v_i = __pyx_t_11;

      /* "pandas/_libs/intervaltree.pxi":2898
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point < self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      __pyx_t_12 = __pyx_v_i;
      __pyx_t_8 = ((*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->left.data + __pyx_t_12 * __pyx_v_self->left.strides[0]) ))) < __pyx_v_point);
      if (__pyx_t_8) {
        __pyx_t_13 = __pyx_v_i;
        __pyx_t_8 = (__pyx_v_point < (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->right.data + __pyx_t_13 * __pyx_v_self->right.strides[0]) ))));
      }
      __pyx_t_14 = (__pyx_t_8 != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":2899
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point < self.right[i]:
 *                     result.append(self.indices[i])             # <<<<<<<<<<<<<<
 *         else:
 *             # There are child nodes. Based on comparing our query to the pivot,
 */
        __pyx_t_15 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->indices.data + __pyx_t_15 * __pyx_v_self->indices.strides[0]) ))));

        /* "pandas/_libs/intervaltree.pxi":2898
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point < self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      }
    }

    /* "pandas/_libs/intervaltree.pxi":2893
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
    goto __pyx_L3;
  }

  /* "pandas/_libs/intervaltree.pxi":2903
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
  /*else*/ {
    __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":2904
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:
 *                 values = self.center_left_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 */
      __pyx_t_16 = __pyx_v_self->center_left_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":2905
 *             if point < self.pivot:
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:
 */
      __pyx_t_16 = __pyx_v_self->center_left_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_indices = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":2906
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):             # <<<<<<<<<<<<<<
 *                     if not values[i] < point:
 *                         break
 */
      __pyx_t_9 = __pyx_v_self->n_center;
      __pyx_t_10 = __pyx_t_9;
      for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":2907
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_17 = __pyx_v_i;
        __pyx_t_14 = ((!(((*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_17 * __pyx_v_values.strides[0]) ))) < __pyx_v_point) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":2908
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:
 */
          goto __pyx_L9_break;

          /* "pandas/_libs/intervaltree.pxi":2907
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":2909
 *                     if not values[i] < point:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 */
        __pyx_t_18 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_18 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L9_break:;

      /* "pandas/_libs/intervaltree.pxi":2910
 *                         break
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->left_node->max_right) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":2911
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)             # <<<<<<<<<<<<<<
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode *)__pyx_v_self->left_node->__pyx_vtab)->__pyx_fuse_1query(__pyx_v_self->left_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2911, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":2910
 *                         break
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":2903
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":2912
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
    __pyx_t_14 = ((__pyx_v_point > __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":2913
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 *                 values = self.center_right_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 */
      __pyx_t_16 = __pyx_v_self->center_right_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":2914
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:
 */
      __pyx_t_16 = __pyx_v_self->center_right_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_indices = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":2915
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):             # <<<<<<<<<<<<<<
 *                     if not point < values[i]:
 *                         break
 */
      for (__pyx_t_11 = (__pyx_v_self->n_center - 1); __pyx_t_11 > -1L; __pyx_t_11-=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":2916
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_19 = __pyx_v_i;
        __pyx_t_14 = ((!((__pyx_v_point < (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_19 * __pyx_v_values.strides[0]) )))) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":2917
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:
 */
          goto __pyx_L13_break;

          /* "pandas/_libs/intervaltree.pxi":2916
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":2918
 *                     if not point < values[i]:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if self.right_node.min_left < point:
 *                     self.right_node.query(result, point)
 */
        __pyx_t_20 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_20 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L13_break:;

      /* "pandas/_libs/intervaltree.pxi":2919
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      __pyx_t_14 = ((__pyx_v_self->right_node->min_left < __pyx_v_point) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":2920
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:
 *                     self.right_node.query(result, point)             # <<<<<<<<<<<<<<
 *             else:
 *                 result.extend(self.center_left_indices)
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode *)__pyx_v_self->right_node->__pyx_vtab)->__pyx_fuse_1query(__pyx_v_self->right_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2920, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":2919
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":2912
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":2922
 *                     self.right_node.query(result, point)
 *             else:
 *                 result.extend(self.center_left_indices)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
    /*else*/ {
      __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->extend(__pyx_v_result, __pyx_v_self->center_left_indices); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2922, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_L7:;
  }
  __pyx_L3:;

  /* "pandas/_libs/intervaltree.pxi":2884
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __PYX_XDEC_MEMVIEW(&__pyx_t_16, 1);
  __Pyx_AddTraceback("pandas._libs.interval.Int64ClosedNeitherIntervalNode.query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_indices, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_values, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_11__pyx_fuse_1query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_1__pyx_mdef_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_11__pyx_fuse_1query = {"__pyx_fuse_1query", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_11__pyx_fuse_1query, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_2query};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_11__pyx_fuse_1query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result = 0;
  __pyx_t_5numpy_float32_t __pyx_v_point;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fuse_1query (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_result,&__pyx_n_s_point,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_result)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_point)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_1query", 1, 2, 2, 1); __PYX_ERR(1, 2884, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_1query") < 0)) __PYX_ERR(1, 2884, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_result = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)values[0]);
    __pyx_v_point = __pyx_PyFloat_AsFloat(values[1]); if (unlikely((__pyx_v_point == ((npy_float32)-1)) && PyErr_Occurred())) __PYX_ERR(1, 2884, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_1query", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 2884, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Int64ClosedNeitherIntervalNode.__pyx_fuse_1query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_result), __pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector, 1, "result", 0))) __PYX_ERR(1, 2884, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_10__pyx_fuse_1query(((struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode *)__pyx_v_self), __pyx_v_result, __pyx_v_point);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_10__pyx_fuse_1query(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float32_t __pyx_v_point) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__pyx_fuse_1query", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_fuse_1__pyx_f_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_query(__pyx_v_self, __pyx_v_result, __pyx_v_point, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2884, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Int64ClosedNeitherIntervalNode.__pyx_fuse_1query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_13__pyx_fuse_2query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_fuse_2__pyx_f_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int64_t __pyx_v_point, int __pyx_skip_dispatch) {
  __Pyx_memviewslice __pyx_v_indices = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_values = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __pyx_t_5numpy_int64_t __pyx_t_9;
  __pyx_t_5numpy_int64_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  int __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  __Pyx_memviewslice __pyx_t_16 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_17;
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  __Pyx_RefNannySetupContext("__pyx_fuse_2query", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pyx_fuse_2query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2884, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_13__pyx_fuse_2query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = __Pyx_PyInt_From_npy_int64(__pyx_v_point); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2884, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2884, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2884, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 2884, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_5) {
            __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2884, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2884, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_3query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_4 = __Pyx_PyInt_From_npy_int64(__pyx_v_point); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 2884, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_7 = __pyx_t_1; __pyx_t_3 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2884, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2884, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 2884, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (__pyx_t_3) {
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_t_4);
          __pyx_t_4 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2884, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pandas/_libs/intervaltree.pxi":2893
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
  __pyx_t_8 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_8) {

    /* "pandas/_libs/intervaltree.pxi":2897
 *             # continue the binary tree structure. Instead, we use linear
 *             # search.
 *             for i in range(self.n_elements):             # <<<<<<<<<<<<<<
 *                 if self.left[i] < point < self.right[i]:
 *                     result.append(self.indices[i])
 */
    __pyx_t_9 = __pyx_v_self->n_elements;
    __pyx_t_10 = __pyx_t_9;
    for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
      __pyx_v_i = __pyx_t_11;

      /* "pandas/_libs/intervaltree.pxi":2898
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point < self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      __pyx_t_12 = __pyx_v_i;
      __pyx_t_8 = ((*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->left.data + __pyx_t_12 * __pyx_v_self->left.strides[0]) ))) < __pyx_v_point);
      if (__pyx_t_8) {
        __pyx_t_13 = __pyx_v_i;
        __pyx_t_8 = (__pyx_v_point < (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->right.data + __pyx_t_13 * __pyx_v_self->right.strides[0]) ))));
      }
      __pyx_t_14 = (__pyx_t_8 != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":2899
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point < self.right[i]:
 *                     result.append(self.indices[i])             # <<<<<<<<<<<<<<
 *         else:
 *             # There are child nodes. Based on comparing our query to the pivot,
 */
        __pyx_t_15 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->indices.data + __pyx_t_15 * __pyx_v_self->indices.strides[0]) ))));

        /* "pandas/_libs/intervaltree.pxi":2898
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point < self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      }
    }

    /* "pandas/_libs/intervaltree.pxi":2893
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
    goto __pyx_L3;
  }

  /* "pandas/_libs/intervaltree.pxi":2903
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
  /*else*/ {
    __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":2904
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:
 *                 values = self.center_left_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 */
      __pyx_t_16 = __pyx_v_self->center_left_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":2905
 *             if point < self.pivot:
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:
 */
      __pyx_t_16 = __pyx_v_self->center_left_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_indices = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":2906
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):             # <<<<<<<<<<<<<<
 *                     if not values[i] < point:
 *                         break
 */
      __pyx_t_9 = __pyx_v_self->n_center;
      __pyx_t_10 = __pyx_t_9;
      for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":2907
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_17 = __pyx_v_i;
        __pyx_t_14 = ((!(((*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_17 * __pyx_v_values.strides[0]) ))) < __pyx_v_point) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":2908
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:
 */
          goto __pyx_L9_break;

          /* "pandas/_libs/intervaltree.pxi":2907
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":2909
 *                     if not values[i] < point:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 */
        __pyx_t_18 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_18 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L9_break:;

      /* "pandas/_libs/intervaltree.pxi":2910
 *                         break
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->left_node->max_right) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":2911
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)             # <<<<<<<<<<<<<<
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode *)__pyx_v_self->left_node->__pyx_vtab)->__pyx_fuse_2query(__pyx_v_self->left_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2911, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":2910
 *                         break
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":2903
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":2912
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
    __pyx_t_14 = ((__pyx_v_point > __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":2913
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 *                 values = self.center_right_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 */
      __pyx_t_16 = __pyx_v_self->center_right_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":2914
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:
 */
      __pyx_t_16 = __pyx_v_self->center_right_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_indices = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":2915
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):             # <<<<<<<<<<<<<<
 *                     if not point < values[i]:
 *                         break
 */
      for (__pyx_t_11 = (__pyx_v_self->n_center - 1); __pyx_t_11 > -1L; __pyx_t_11-=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":2916
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_19 = __pyx_v_i;
        __pyx_t_14 = ((!((__pyx_v_point < (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_19 * __pyx_v_values.strides[0]) )))) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":2917
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:
 */
          goto __pyx_L13_break;

          /* "pandas/_libs/intervaltree.pxi":2916
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":2918
 *                     if not point < values[i]:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if self.right_node.min_left < point:
 *                     self.right_node.query(result, point)
 */
        __pyx_t_20 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_20 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L13_break:;

      /* "pandas/_libs/intervaltree.pxi":2919
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      __pyx_t_14 = ((__pyx_v_self->right_node->min_left < __pyx_v_point) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":2920
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:
 *                     self.right_node.query(result, point)             # <<<<<<<<<<<<<<
 *             else:
 *                 result.extend(self.center_left_indices)
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode *)__pyx_v_self->right_node->__pyx_vtab)->__pyx_fuse_2query(__pyx_v_self->right_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2920, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":2919
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":2912
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":2922
 *                     self.right_node.query(result, point)
 *             else:
 *                 result.extend(self.center_left_indices)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
    /*else*/ {
      __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->extend(__pyx_v_result, __pyx_v_self->center_left_indices); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2922, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_L7:;
  }
  __pyx_L3:;

  /* "pandas/_libs/intervaltree.pxi":2884
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __PYX_XDEC_MEMVIEW(&__pyx_t_16, 1);
  __Pyx_AddTraceback("pandas._libs.interval.Int64ClosedNeitherIntervalNode.query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_indices, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_values, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_13__pyx_fuse_2query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_2__pyx_mdef_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_13__pyx_fuse_2query = {"__pyx_fuse_2query", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_13__pyx_fuse_2query, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_2query};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_13__pyx_fuse_2query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result = 0;
  __pyx_t_5numpy_int64_t __pyx_v_point;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fuse_2query (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_result,&__pyx_n_s_point,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_result)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_point)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_2query", 1, 2, 2, 1); __PYX_ERR(1, 2884, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_2query") < 0)) __PYX_ERR(1, 2884, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_result = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)values[0]);
    __pyx_v_point = __Pyx_PyInt_As_npy_int64(values[1]); if (unlikely((__pyx_v_point == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(1, 2884, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_2query", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 2884, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Int64ClosedNeitherIntervalNode.__pyx_fuse_2query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_result), __pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector, 1, "result", 0))) __PYX_ERR(1, 2884, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_12__pyx_fuse_2query(((struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode *)__pyx_v_self), __pyx_v_result, __pyx_v_point);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_12__pyx_fuse_2query(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int64_t __pyx_v_point) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__pyx_fuse_2query", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_fuse_2__pyx_f_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_query(__pyx_v_self, __pyx_v_result, __pyx_v_point, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2884, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Int64ClosedNeitherIntervalNode.__pyx_fuse_2query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_15__pyx_fuse_3query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_fuse_3__pyx_f_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int32_t __pyx_v_point, int __pyx_skip_dispatch) {
  __Pyx_memviewslice __pyx_v_indices = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_values = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __pyx_t_5numpy_int64_t __pyx_t_9;
  __pyx_t_5numpy_int64_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  int __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  __Pyx_memviewslice __pyx_t_16 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_17;
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  __Pyx_RefNannySetupContext("__pyx_fuse_3query", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pyx_fuse_3query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2884, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_15__pyx_fuse_3query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = __Pyx_PyInt_From_npy_int32(__pyx_v_point); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2884, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2884, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2884, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 2884, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_5) {
            __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2884, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2884, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_3query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_4 = __Pyx_PyInt_From_npy_int32(__pyx_v_point); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 2884, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_7 = __pyx_t_1; __pyx_t_3 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2884, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2884, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 2884, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (__pyx_t_3) {
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_t_4);
          __pyx_t_4 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2884, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pandas/_libs/intervaltree.pxi":2893
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
  __pyx_t_8 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_8) {

    /* "pandas/_libs/intervaltree.pxi":2897
 *             # continue the binary tree structure. Instead, we use linear
 *             # search.
 *             for i in range(self.n_elements):             # <<<<<<<<<<<<<<
 *                 if self.left[i] < point < self.right[i]:
 *                     result.append(self.indices[i])
 */
    __pyx_t_9 = __pyx_v_self->n_elements;
    __pyx_t_10 = __pyx_t_9;
    for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
      __pyx_v_i = __pyx_t_11;

      /* "pandas/_libs/intervaltree.pxi":2898
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point < self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      __pyx_t_12 = __pyx_v_i;
      __pyx_t_8 = ((*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->left.data + __pyx_t_12 * __pyx_v_self->left.strides[0]) ))) < __pyx_v_point);
      if (__pyx_t_8) {
        __pyx_t_13 = __pyx_v_i;
        __pyx_t_8 = (__pyx_v_point < (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->right.data + __pyx_t_13 * __pyx_v_self->right.strides[0]) ))));
      }
      __pyx_t_14 = (__pyx_t_8 != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":2899
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point < self.right[i]:
 *                     result.append(self.indices[i])             # <<<<<<<<<<<<<<
 *         else:
 *             # There are child nodes. Based on comparing our query to the pivot,
 */
        __pyx_t_15 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->indices.data + __pyx_t_15 * __pyx_v_self->indices.strides[0]) ))));

        /* "pandas/_libs/intervaltree.pxi":2898
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point < self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      }
    }

    /* "pandas/_libs/intervaltree.pxi":2893
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
    goto __pyx_L3;
  }

  /* "pandas/_libs/intervaltree.pxi":2903
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
  /*else*/ {
    __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":2904
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:
 *                 values = self.center_left_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 */
      __pyx_t_16 = __pyx_v_self->center_left_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":2905
 *             if point < self.pivot:
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:
 */
      __pyx_t_16 = __pyx_v_self->center_left_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_indices = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":2906
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):             # <<<<<<<<<<<<<<
 *                     if not values[i] < point:
 *                         break
 */
      __pyx_t_9 = __pyx_v_self->n_center;
      __pyx_t_10 = __pyx_t_9;
      for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":2907
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_17 = __pyx_v_i;
        __pyx_t_14 = ((!(((*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_17 * __pyx_v_values.strides[0]) ))) < __pyx_v_point) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":2908
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:
 */
          goto __pyx_L9_break;

          /* "pandas/_libs/intervaltree.pxi":2907
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":2909
 *                     if not values[i] < point:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 */
        __pyx_t_18 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_18 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L9_break:;

      /* "pandas/_libs/intervaltree.pxi":2910
 *                         break
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->left_node->max_right) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":2911
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)             # <<<<<<<<<<<<<<
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode *)__pyx_v_self->left_node->__pyx_vtab)->__pyx_fuse_3query(__pyx_v_self->left_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2911, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":2910
 *                         break
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":2903
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":2912
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
    __pyx_t_14 = ((__pyx_v_point > __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":2913
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 *                 values = self.center_right_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 */
      __pyx_t_16 = __pyx_v_self->center_right_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":2914
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:
 */
      __pyx_t_16 = __pyx_v_self->center_right_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_indices = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":2915
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):             # <<<<<<<<<<<<<<
 *                     if not point < values[i]:
 *                         break
 */
      for (__pyx_t_11 = (__pyx_v_self->n_center - 1); __pyx_t_11 > -1L; __pyx_t_11-=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":2916
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_19 = __pyx_v_i;
        __pyx_t_14 = ((!((__pyx_v_point < (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_19 * __pyx_v_values.strides[0]) )))) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":2917
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:
 */
          goto __pyx_L13_break;

          /* "pandas/_libs/intervaltree.pxi":2916
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":2918
 *                     if not point < values[i]:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if self.right_node.min_left < point:
 *                     self.right_node.query(result, point)
 */
        __pyx_t_20 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_20 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L13_break:;

      /* "pandas/_libs/intervaltree.pxi":2919
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      __pyx_t_14 = ((__pyx_v_self->right_node->min_left < __pyx_v_point) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":2920
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:
 *                     self.right_node.query(result, point)             # <<<<<<<<<<<<<<
 *             else:
 *                 result.extend(self.center_left_indices)
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode *)__pyx_v_self->right_node->__pyx_vtab)->__pyx_fuse_3query(__pyx_v_self->right_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2920, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":2919
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":2912
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":2922
 *                     self.right_node.query(result, point)
 *             else:
 *                 result.extend(self.center_left_indices)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
    /*else*/ {
      __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->extend(__pyx_v_result, __pyx_v_self->center_left_indices); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2922, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_L7:;
  }
  __pyx_L3:;

  /* "pandas/_libs/intervaltree.pxi":2884
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __PYX_XDEC_MEMVIEW(&__pyx_t_16, 1);
  __Pyx_AddTraceback("pandas._libs.interval.Int64ClosedNeitherIntervalNode.query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_indices, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_values, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_15__pyx_fuse_3query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_3__pyx_mdef_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_15__pyx_fuse_3query = {"__pyx_fuse_3query", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_15__pyx_fuse_3query, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_2query};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_15__pyx_fuse_3query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result = 0;
  __pyx_t_5numpy_int32_t __pyx_v_point;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fuse_3query (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_result,&__pyx_n_s_point,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_result)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_point)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_3query", 1, 2, 2, 1); __PYX_ERR(1, 2884, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_3query") < 0)) __PYX_ERR(1, 2884, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_result = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)values[0]);
    __pyx_v_point = __Pyx_PyInt_As_npy_int32(values[1]); if (unlikely((__pyx_v_point == ((npy_int32)-1)) && PyErr_Occurred())) __PYX_ERR(1, 2884, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_3query", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 2884, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Int64ClosedNeitherIntervalNode.__pyx_fuse_3query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_result), __pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector, 1, "result", 0))) __PYX_ERR(1, 2884, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_14__pyx_fuse_3query(((struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode *)__pyx_v_self), __pyx_v_result, __pyx_v_point);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_14__pyx_fuse_3query(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int32_t __pyx_v_point) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__pyx_fuse_3query", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_fuse_3__pyx_f_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_query(__pyx_v_self, __pyx_v_result, __pyx_v_point, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2884, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Int64ClosedNeitherIntervalNode.__pyx_fuse_3query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_17__pyx_fuse_4query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_fuse_4__pyx_f_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_uint64_t __pyx_v_point, int __pyx_skip_dispatch) {
  __Pyx_memviewslice __pyx_v_indices = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_values = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __pyx_t_5numpy_int64_t __pyx_t_9;
  __pyx_t_5numpy_int64_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  int __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  __Pyx_memviewslice __pyx_t_16 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_17;
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  __Pyx_RefNannySetupContext("__pyx_fuse_4query", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pyx_fuse_4query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2884, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_17__pyx_fuse_4query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = __Pyx_PyInt_From_npy_uint64(__pyx_v_point); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2884, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2884, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2884, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 2884, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_5) {
            __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2884, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2884, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_3query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_4 = __Pyx_PyInt_From_npy_uint64(__pyx_v_point); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 2884, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_7 = __pyx_t_1; __pyx_t_3 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2884, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2884, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 2884, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (__pyx_t_3) {
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_t_4);
          __pyx_t_4 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2884, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pandas/_libs/intervaltree.pxi":2893
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
  __pyx_t_8 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_8) {

    /* "pandas/_libs/intervaltree.pxi":2897
 *             # continue the binary tree structure. Instead, we use linear
 *             # search.
 *             for i in range(self.n_elements):             # <<<<<<<<<<<<<<
 *                 if self.left[i] < point < self.right[i]:
 *                     result.append(self.indices[i])
 */
    __pyx_t_9 = __pyx_v_self->n_elements;
    __pyx_t_10 = __pyx_t_9;
    for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
      __pyx_v_i = __pyx_t_11;

      /* "pandas/_libs/intervaltree.pxi":2898
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point < self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      __pyx_t_12 = __pyx_v_i;
      __pyx_t_8 = ((*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->left.data + __pyx_t_12 * __pyx_v_self->left.strides[0]) ))) < __pyx_v_point);
      if (__pyx_t_8) {
        __pyx_t_13 = __pyx_v_i;
        __pyx_t_8 = (__pyx_v_point < (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->right.data + __pyx_t_13 * __pyx_v_self->right.strides[0]) ))));
      }
      __pyx_t_14 = (__pyx_t_8 != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":2899
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point < self.right[i]:
 *                     result.append(self.indices[i])             # <<<<<<<<<<<<<<
 *         else:
 *             # There are child nodes. Based on comparing our query to the pivot,
 */
        __pyx_t_15 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->indices.data + __pyx_t_15 * __pyx_v_self->indices.strides[0]) ))));

        /* "pandas/_libs/intervaltree.pxi":2898
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point < self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      }
    }

    /* "pandas/_libs/intervaltree.pxi":2893
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
    goto __pyx_L3;
  }

  /* "pandas/_libs/intervaltree.pxi":2903
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
  /*else*/ {
    __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":2904
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:
 *                 values = self.center_left_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 */
      __pyx_t_16 = __pyx_v_self->center_left_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":2905
 *             if point < self.pivot:
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:
 */
      __pyx_t_16 = __pyx_v_self->center_left_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_indices = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":2906
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):             # <<<<<<<<<<<<<<
 *                     if not values[i] < point:
 *                         break
 */
      __pyx_t_9 = __pyx_v_self->n_center;
      __pyx_t_10 = __pyx_t_9;
      for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":2907
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_17 = __pyx_v_i;
        __pyx_t_14 = ((!(((*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_17 * __pyx_v_values.strides[0]) ))) < __pyx_v_point) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":2908
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:
 */
          goto __pyx_L9_break;

          /* "pandas/_libs/intervaltree.pxi":2907
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":2909
 *                     if not values[i] < point:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 */
        __pyx_t_18 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_18 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L9_break:;

      /* "pandas/_libs/intervaltree.pxi":2910
 *                         break
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->left_node->max_right) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":2911
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)             # <<<<<<<<<<<<<<
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode *)__pyx_v_self->left_node->__pyx_vtab)->__pyx_fuse_4query(__pyx_v_self->left_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2911, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":2910
 *                         break
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":2903
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":2912
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
    __pyx_t_14 = ((__pyx_v_point > __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":2913
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 *                 values = self.center_right_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 */
      __pyx_t_16 = __pyx_v_self->center_right_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":2914
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:
 */
      __pyx_t_16 = __pyx_v_self->center_right_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_indices = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":2915
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):             # <<<<<<<<<<<<<<
 *                     if not point < values[i]:
 *                         break
 */
      for (__pyx_t_11 = (__pyx_v_self->n_center - 1); __pyx_t_11 > -1L; __pyx_t_11-=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":2916
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_19 = __pyx_v_i;
        __pyx_t_14 = ((!((__pyx_v_point < (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_19 * __pyx_v_values.strides[0]) )))) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":2917
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:
 */
          goto __pyx_L13_break;

          /* "pandas/_libs/intervaltree.pxi":2916
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":2918
 *                     if not point < values[i]:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if self.right_node.min_left < point:
 *                     self.right_node.query(result, point)
 */
        __pyx_t_20 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_20 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L13_break:;

      /* "pandas/_libs/intervaltree.pxi":2919
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      __pyx_t_14 = ((__pyx_v_self->right_node->min_left < __pyx_v_point) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":2920
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:
 *                     self.right_node.query(result, point)             # <<<<<<<<<<<<<<
 *             else:
 *                 result.extend(self.center_left_indices)
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode *)__pyx_v_self->right_node->__pyx_vtab)->__pyx_fuse_4query(__pyx_v_self->right_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2920, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":2919
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":2912
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":2922
 *                     self.right_node.query(result, point)
 *             else:
 *                 result.extend(self.center_left_indices)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
    /*else*/ {
      __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->extend(__pyx_v_result, __pyx_v_self->center_left_indices); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2922, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_L7:;
  }
  __pyx_L3:;

  /* "pandas/_libs/intervaltree.pxi":2884
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __PYX_XDEC_MEMVIEW(&__pyx_t_16, 1);
  __Pyx_AddTraceback("pandas._libs.interval.Int64ClosedNeitherIntervalNode.query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_indices, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_values, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_17__pyx_fuse_4query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_4__pyx_mdef_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_17__pyx_fuse_4query = {"__pyx_fuse_4query", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_17__pyx_fuse_4query, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_2query};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_17__pyx_fuse_4query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result = 0;
  __pyx_t_5numpy_uint64_t __pyx_v_point;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fuse_4query (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_result,&__pyx_n_s_point,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_result)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_point)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_4query", 1, 2, 2, 1); __PYX_ERR(1, 2884, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_4query") < 0)) __PYX_ERR(1, 2884, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_result = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)values[0]);
    __pyx_v_point = __Pyx_PyInt_As_npy_uint64(values[1]); if (unlikely((__pyx_v_point == ((npy_uint64)-1)) && PyErr_Occurred())) __PYX_ERR(1, 2884, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_4query", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 2884, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Int64ClosedNeitherIntervalNode.__pyx_fuse_4query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_result), __pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector, 1, "result", 0))) __PYX_ERR(1, 2884, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_16__pyx_fuse_4query(((struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode *)__pyx_v_self), __pyx_v_result, __pyx_v_point);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_16__pyx_fuse_4query(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_uint64_t __pyx_v_point) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__pyx_fuse_4query", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_fuse_4__pyx_f_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_query(__pyx_v_self, __pyx_v_result, __pyx_v_point, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2884, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Int64ClosedNeitherIntervalNode.__pyx_fuse_4query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":2924
 *                 result.extend(self.center_left_indices)
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         if self.is_leaf_node:
 *             return ('<Int64ClosedNeitherIntervalNode: '
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_5__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_5__repr__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_4__repr__(((struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_4__repr__(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode *__pyx_v_self) {
  PyObject *__pyx_v_n_left = NULL;
  PyObject *__pyx_v_n_right = NULL;
  PyObject *__pyx_v_n_center = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  Py_UCS4 __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "pandas/_libs/intervaltree.pxi":2925
 * 
 *     def __repr__(self):
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             return ('<Int64ClosedNeitherIntervalNode: '
 *                     '%s elements (terminal)>' % self.n_elements)
 */
  __pyx_t_1 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_1) {

    /* "pandas/_libs/intervaltree.pxi":2926
 *     def __repr__(self):
 *         if self.is_leaf_node:
 *             return ('<Int64ClosedNeitherIntervalNode: '             # <<<<<<<<<<<<<<
 *                     '%s elements (terminal)>' % self.n_elements)
 *         else:
 */
    __Pyx_XDECREF(__pyx_r);

    /* "pandas/_libs/intervaltree.pxi":2927
 *         if self.is_leaf_node:
 *             return ('<Int64ClosedNeitherIntervalNode: '
 *                     '%s elements (terminal)>' % self.n_elements)             # <<<<<<<<<<<<<<
 *         else:
 *             n_left = self.left_node.n_elements
 */
    __pyx_t_2 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_elements); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2927, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PyUnicode_Format(__pyx_kp_u_Int64ClosedNeitherIntervalNode, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2927, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pandas/_libs/intervaltree.pxi":2925
 * 
 *     def __repr__(self):
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             return ('<Int64ClosedNeitherIntervalNode: '
 *                     '%s elements (terminal)>' % self.n_elements)
 */
  }

  /* "pandas/_libs/intervaltree.pxi":2929
 *                     '%s elements (terminal)>' % self.n_elements)
 *         else:
 *             n_left = self.left_node.n_elements             # <<<<<<<<<<<<<<
 *             n_right = self.right_node.n_elements
 *             n_center = self.n_elements - n_left - n_right
 */
  /*else*/ {
    __pyx_t_3 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->left_node->n_elements); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2929, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_v_n_left = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "pandas/_libs/intervaltree.pxi":2930
 *         else:
 *             n_left = self.left_node.n_elements
 *             n_right = self.right_node.n_elements             # <<<<<<<<<<<<<<
 *             n_center = self.n_elements - n_left - n_right
 *             return ('<Int64ClosedNeitherIntervalNode: '
 */
    __pyx_t_3 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->right_node->n_elements); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2930, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_v_n_right = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "pandas/_libs/intervaltree.pxi":2931
 *             n_left = self.left_node.n_elements
 *             n_right = self.right_node.n_elements
 *             n_center = self.n_elements - n_left - n_right             # <<<<<<<<<<<<<<
 *             return ('<Int64ClosedNeitherIntervalNode: '
 *                     'pivot %s, %s elements (%s left, %s right, %s '
 */
    __pyx_t_3 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_elements); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2931, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = PyNumber_Subtract(__pyx_t_3, __pyx_v_n_left); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2931, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyNumber_Subtract(__pyx_t_2, __pyx_v_n_right); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2931, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_n_center = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "pandas/_libs/intervaltree.pxi":2932
 *             n_right = self.right_node.n_elements
 *             n_center = self.n_elements - n_left - n_right
 *             return ('<Int64ClosedNeitherIntervalNode: '             # <<<<<<<<<<<<<<
 *                     'pivot %s, %s elements (%s left, %s right, %s '
 *                     'overlapping)>' % (self.pivot, self.n_elements,
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = PyTuple_New(11); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2932, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = 0;
    __pyx_t_5 = 127;
    __Pyx_INCREF(__pyx_kp_u_Int64ClosedNeitherIntervalNode_2);
    __pyx_t_4 += 39;
    __Pyx_GIVEREF(__pyx_kp_u_Int64ClosedNeitherIntervalNode_2);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_kp_u_Int64ClosedNeitherIntervalNode_2);

    /* "pandas/_libs/intervaltree.pxi":2934
 *             return ('<Int64ClosedNeitherIntervalNode: '
 *                     'pivot %s, %s elements (%s left, %s right, %s '
 *                     'overlapping)>' % (self.pivot, self.n_elements,             # <<<<<<<<<<<<<<
 *                                        n_left, n_right, n_center))
 * 
 */
    __pyx_t_2 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->pivot); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2934, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_t_2), __pyx_empty_unicode); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 2934, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_5 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_6) > __pyx_t_5) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_6) : __pyx_t_5;
    __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_6);
    __pyx_t_6 = 0;
    __Pyx_INCREF(__pyx_kp_u__18);
    __pyx_t_4 += 2;
    __Pyx_GIVEREF(__pyx_kp_u__18);
    PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_kp_u__18);
    __pyx_t_6 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_elements); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 2934, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_2 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_t_6), __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2934, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_5 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) > __pyx_t_5) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) : __pyx_t_5;
    __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_3, 3, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_elements);
    __pyx_t_4 += 11;
    __Pyx_GIVEREF(__pyx_kp_u_elements);
    PyTuple_SET_ITEM(__pyx_t_3, 4, __pyx_kp_u_elements);

    /* "pandas/_libs/intervaltree.pxi":2935
 *                     'pivot %s, %s elements (%s left, %s right, %s '
 *                     'overlapping)>' % (self.pivot, self.n_elements,
 *                                        n_left, n_right, n_center))             # <<<<<<<<<<<<<<
 * 
 *     def counts(self):
 */
    __pyx_t_2 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_v_n_left), __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2935, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) > __pyx_t_5) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) : __pyx_t_5;
    __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_3, 5, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_left_2);
    __pyx_t_4 += 7;
    __Pyx_GIVEREF(__pyx_kp_u_left_2);
    PyTuple_SET_ITEM(__pyx_t_3, 6, __pyx_kp_u_left_2);
    __pyx_t_2 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_v_n_right), __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2935, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) > __pyx_t_5) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) : __pyx_t_5;
    __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_3, 7, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_right_2);
    __pyx_t_4 += 8;
    __Pyx_GIVEREF(__pyx_kp_u_right_2);
    PyTuple_SET_ITEM(__pyx_t_3, 8, __pyx_kp_u_right_2);
    __pyx_t_2 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_v_n_center), __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2935, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) > __pyx_t_5) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) : __pyx_t_5;
    __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_3, 9, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_overlapping);
    __pyx_t_4 += 14;
    __Pyx_GIVEREF(__pyx_kp_u_overlapping);
    PyTuple_SET_ITEM(__pyx_t_3, 10, __pyx_kp_u_overlapping);

    /* "pandas/_libs/intervaltree.pxi":2932
 *             n_right = self.right_node.n_elements
 *             n_center = self.n_elements - n_left - n_right
 *             return ('<Int64ClosedNeitherIntervalNode: '             # <<<<<<<<<<<<<<
 *                     'pivot %s, %s elements (%s left, %s right, %s '
 *                     'overlapping)>' % (self.pivot, self.n_elements,
 */
    __pyx_t_2 = __Pyx_PyUnicode_Join(__pyx_t_3, 11, __pyx_t_4, __pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2932, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;
  }

  /* "pandas/_libs/intervaltree.pxi":2924
 *                 result.extend(self.center_left_indices)
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         if self.is_leaf_node:
 *             return ('<Int64ClosedNeitherIntervalNode: '
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pandas._libs.interval.Int64ClosedNeitherIntervalNode.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_n_left);
  __Pyx_XDECREF(__pyx_v_n_right);
  __Pyx_XDECREF(__pyx_v_n_center);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":2937
 *                                        n_left, n_right, n_center))
 * 
 *     def counts(self):             # <<<<<<<<<<<<<<
 *         """
 *         Inspect counts on this node
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_7counts(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_6counts[] = "\n        Inspect counts on this node\n        useful for debugging purposes\n        ";
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_7counts(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("counts (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_6counts(((struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_6counts(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode *__pyx_v_self) {
  Py_ssize_t __pyx_v_m;
  PyObject *__pyx_v_l = NULL;
  PyObject *__pyx_v_r = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  size_t __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("counts", 0);

  /* "pandas/_libs/intervaltree.pxi":2942
 *         useful for debugging purposes
 *         """
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             return self.n_elements
 *         else:
 */
  __pyx_t_1 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_1) {

    /* "pandas/_libs/intervaltree.pxi":2943
 *         """
 *         if self.is_leaf_node:
 *             return self.n_elements             # <<<<<<<<<<<<<<
 *         else:
 *             m = len(self.center_left_values)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_elements); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2943, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "pandas/_libs/intervaltree.pxi":2942
 *         useful for debugging purposes
 *         """
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             return self.n_elements
 *         else:
 */
  }

  /* "pandas/_libs/intervaltree.pxi":2945
 *             return self.n_elements
 *         else:
 *             m = len(self.center_left_values)             # <<<<<<<<<<<<<<
 *             l = self.left_node.counts()
 *             r = self.right_node.counts()
 */
  /*else*/ {
    if (unlikely(!__pyx_v_self->center_left_values.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(1, 2945, __pyx_L1_error)}
    __pyx_t_3 = __Pyx_MemoryView_Len(__pyx_v_self->center_left_values); 
    __pyx_v_m = __pyx_t_3;

    /* "pandas/_libs/intervaltree.pxi":2946
 *         else:
 *             m = len(self.center_left_values)
 *             l = self.left_node.counts()             # <<<<<<<<<<<<<<
 *             r = self.right_node.counts()
 *             return (m, (l, r))
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->left_node), __pyx_n_s_counts); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 2946, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2946, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_l = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "pandas/_libs/intervaltree.pxi":2947
 *             m = len(self.center_left_values)
 *             l = self.left_node.counts()
 *             r = self.right_node.counts()             # <<<<<<<<<<<<<<
 *             return (m, (l, r))
 * 
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->right_node), __pyx_n_s_counts); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 2947, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2947, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_r = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "pandas/_libs/intervaltree.pxi":2948
 *             l = self.left_node.counts()
 *             r = self.right_node.counts()
 *             return (m, (l, r))             # <<<<<<<<<<<<<<
 * 
 * NODE_CLASSES['int64',
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = PyInt_FromSsize_t(__pyx_v_m); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2948, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 2948, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_v_l);
    __Pyx_GIVEREF(__pyx_v_l);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_l);
    __Pyx_INCREF(__pyx_v_r);
    __Pyx_GIVEREF(__pyx_v_r);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_v_r);
    __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 2948, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_4);
    __pyx_t_2 = 0;
    __pyx_t_4 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;
  }

  /* "pandas/_libs/intervaltree.pxi":2937
 *                                        n_left, n_right, n_center))
 * 
 *     def counts(self):             # <<<<<<<<<<<<<<
 *         """
 *         Inspect counts on this node
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pandas._libs.interval.Int64ClosedNeitherIntervalNode.counts", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_l);
  __Pyx_XDECREF(__pyx_v_r);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":2794
 *         int64_t[:] center_left_indices, center_right_indices, indices
 *         int64_t min_left, max_right
 *         readonly int64_t pivot             # <<<<<<<<<<<<<<
 *         readonly int64_t n_elements, n_center, leaf_size
 *         readonly bint is_leaf_node
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_5pivot_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_5pivot_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_5pivot___get__(((struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_5pivot___get__(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->pivot); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2794, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Int64ClosedNeitherIntervalNode.pivot.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":2795
 *         int64_t min_left, max_right
 *         readonly int64_t pivot
 *         readonly int64_t n_elements, n_center, leaf_size             # <<<<<<<<<<<<<<
 *         readonly bint is_leaf_node
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_10n_elements_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_10n_elements_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_10n_elements___get__(((struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_10n_elements___get__(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_elements); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2795, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Int64ClosedNeitherIntervalNode.n_elements.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_8n_center_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_8n_center_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_8n_center___get__(((struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_8n_center___get__(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_center); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2795, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Int64ClosedNeitherIntervalNode.n_center.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_9leaf_size_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_9leaf_size_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_9leaf_size___get__(((struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_9leaf_size___get__(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->leaf_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2795, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Int64ClosedNeitherIntervalNode.leaf_size.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":2796
 *         readonly int64_t pivot
 *         readonly int64_t n_elements, n_center, leaf_size
 *         readonly bint is_leaf_node             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self,
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_12is_leaf_node_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_12is_leaf_node_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_12is_leaf_node___get__(((struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_12is_leaf_node___get__(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->is_leaf_node); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2796, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Int64ClosedNeitherIntervalNode.is_leaf_node.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_21__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_21__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_20__reduce_cython__(((struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_20__reduce_cython__(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  int __pyx_t_16;
  int __pyx_t_17;
  int __pyx_t_18;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.center_left_indices, self.center_left_values, self.center_right_indices, self.center_right_values, self.indices, self.is_leaf_node, self.leaf_size, self.left, self.left_node, self.max_right, self.min_left, self.n_center, self.n_elements, self.pivot, self.right, self.right_node)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  if (unlikely(!__pyx_v_self->center_left_indices.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_1 = __pyx_memoryview_fromslice(__pyx_v_self->center_left_indices, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_int64_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_int64_t, 0);; if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (unlikely(!__pyx_v_self->center_left_values.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_2 = __pyx_memoryview_fromslice(__pyx_v_self->center_left_values, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_int64_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_int64_t, 0);; if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (unlikely(!__pyx_v_self->center_right_indices.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_3 = __pyx_memoryview_fromslice(__pyx_v_self->center_right_indices, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_int64_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_int64_t, 0);; if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely(!__pyx_v_self->center_right_values.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_4 = __pyx_memoryview_fromslice(__pyx_v_self->center_right_values, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_int64_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_int64_t, 0);; if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (unlikely(!__pyx_v_self->indices.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_5 = __pyx_memoryview_fromslice(__pyx_v_self->indices, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_int64_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_int64_t, 0);; if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyBool_FromLong(__pyx_v_self->is_leaf_node); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->leaf_size); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (unlikely(!__pyx_v_self->left.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_8 = __pyx_memoryview_fromslice(__pyx_v_self->left, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_int64_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_int64_t, 0);; if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->max_right); if (unlikely(!__pyx_t_9)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_10 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->min_left); if (unlikely(!__pyx_t_10)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_11 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_center); if (unlikely(!__pyx_t_11)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_12 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_elements); if (unlikely(!__pyx_t_12)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __pyx_t_13 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->pivot); if (unlikely(!__pyx_t_13)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  if (unlikely(!__pyx_v_self->right.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_14 = __pyx_memoryview_fromslice(__pyx_v_self->right, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_int64_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_int64_t, 0);; if (unlikely(!__pyx_t_14)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __pyx_t_15 = PyTuple_New(16); if (unlikely(!__pyx_t_15)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_15, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_15, 1, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_15, 2, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_15, 3, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_15, 4, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_15, 5, __pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_15, 6, __pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_15, 7, __pyx_t_8);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->left_node));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self->left_node));
  PyTuple_SET_ITEM(__pyx_t_15, 8, ((PyObject *)__pyx_v_self->left_node));
  __Pyx_GIVEREF(__pyx_t_9);
  PyTuple_SET_ITEM(__pyx_t_15, 9, __pyx_t_9);
  __Pyx_GIVEREF(__pyx_t_10);
  PyTuple_SET_ITEM(__pyx_t_15, 10, __pyx_t_10);
  __Pyx_GIVEREF(__pyx_t_11);
  PyTuple_SET_ITEM(__pyx_t_15, 11, __pyx_t_11);
  __Pyx_GIVEREF(__pyx_t_12);
  PyTuple_SET_ITEM(__pyx_t_15, 12, __pyx_t_12);
  __Pyx_GIVEREF(__pyx_t_13);
  PyTuple_SET_ITEM(__pyx_t_15, 13, __pyx_t_13);
  __Pyx_GIVEREF(__pyx_t_14);
  PyTuple_SET_ITEM(__pyx_t_15, 14, __pyx_t_14);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->right_node));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self->right_node));
  PyTuple_SET_ITEM(__pyx_t_15, 15, ((PyObject *)__pyx_v_self->right_node));
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  __pyx_t_6 = 0;
  __pyx_t_7 = 0;
  __pyx_t_8 = 0;
  __pyx_t_9 = 0;
  __pyx_t_10 = 0;
  __pyx_t_11 = 0;
  __pyx_t_12 = 0;
  __pyx_t_13 = 0;
  __pyx_t_14 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_15);
  __pyx_t_15 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.center_left_indices, self.center_left_values, self.center_right_indices, self.center_right_values, self.indices, self.is_leaf_node, self.leaf_size, self.left, self.left_node, self.max_right, self.min_left, self.n_center, self.n_elements, self.pivot, self.right, self.right_node)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_15 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_15)) __PYX_ERR(2, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __pyx_v__dict = __pyx_t_15;
  __pyx_t_15 = 0;

  /* "(tree fragment)":7
 *     state = (self.center_left_indices, self.center_left_values, self.center_right_indices, self.center_right_values, self.indices, self.is_leaf_node, self.leaf_size, self.left, self.left_node, self.max_right, self.min_left, self.n_center, self.n_elements, self.pivot, self.right, self.right_node)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_16 = (__pyx_v__dict != Py_None);
  __pyx_t_17 = (__pyx_t_16 != 0);
  if (__pyx_t_17) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_15 = PyTuple_New(1); if (unlikely(!__pyx_t_15)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    PyTuple_SET_ITEM(__pyx_t_15, 0, __pyx_v__dict);
    __pyx_t_14 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_15); if (unlikely(!__pyx_t_14)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_14));
    __pyx_t_14 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.left_node is not None or self.right_node is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.center_left_indices, self.center_left_values, self.center_right_indices, self.center_right_values, self.indices, self.is_leaf_node, self.leaf_size, self.left, self.left_node, self.max_right, self.min_left, self.n_center, self.n_elements, self.pivot, self.right, self.right_node)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.left_node is not None or self.right_node is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_Int64ClosedNeitherIntervalNode, (type(self), 0x7a104e5, None), state
 */
  /*else*/ {
    __pyx_t_16 = (((PyObject *)__pyx_v_self->left_node) != Py_None);
    __pyx_t_18 = (__pyx_t_16 != 0);
    if (!__pyx_t_18) {
    } else {
      __pyx_t_17 = __pyx_t_18;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_18 = (((PyObject *)__pyx_v_self->right_node) != Py_None);
    __pyx_t_16 = (__pyx_t_18 != 0);
    __pyx_t_17 = __pyx_t_16;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_17;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.left_node is not None or self.right_node is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Int64ClosedNeitherIntervalNode, (type(self), 0x7a104e5, None), state
 *     else:
 */
  __pyx_t_17 = (__pyx_v_use_setstate != 0);
  if (__pyx_t_17) {

    /* "(tree fragment)":13
 *         use_setstate = self.left_node is not None or self.right_node is not None
 *     if use_setstate:
 *         return __pyx_unpickle_Int64ClosedNeitherIntervalNode, (type(self), 0x7a104e5, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_Int64ClosedNeitherIntervalNode, (type(self), 0x7a104e5, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_14, __pyx_n_s_pyx_unpickle_Int64ClosedNeithe); if (unlikely(!__pyx_t_14)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __pyx_t_15 = PyTuple_New(3); if (unlikely(!__pyx_t_15)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_15, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_127993061);
    __Pyx_GIVEREF(__pyx_int_127993061);
    PyTuple_SET_ITEM(__pyx_t_15, 1, __pyx_int_127993061);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_15, 2, Py_None);
    __pyx_t_13 = PyTuple_New(3); if (unlikely(!__pyx_t_13)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __Pyx_GIVEREF(__pyx_t_14);
    PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_t_14);
    __Pyx_GIVEREF(__pyx_t_15);
    PyTuple_SET_ITEM(__pyx_t_13, 1, __pyx_t_15);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_13, 2, __pyx_v_state);
    __pyx_t_14 = 0;
    __pyx_t_15 = 0;
    __pyx_r = __pyx_t_13;
    __pyx_t_13 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.left_node is not None or self.right_node is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Int64ClosedNeitherIntervalNode, (type(self), 0x7a104e5, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_Int64ClosedNeitherIntervalNode, (type(self), 0x7a104e5, None), state
 *     else:
 *         return __pyx_unpickle_Int64ClosedNeitherIntervalNode, (type(self), 0x7a104e5, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Int64ClosedNeitherIntervalNode__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_13, __pyx_n_s_pyx_unpickle_Int64ClosedNeithe); if (unlikely(!__pyx_t_13)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __pyx_t_15 = PyTuple_New(3); if (unlikely(!__pyx_t_15)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_15, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_127993061);
    __Pyx_GIVEREF(__pyx_int_127993061);
    PyTuple_SET_ITEM(__pyx_t_15, 1, __pyx_int_127993061);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_15, 2, __pyx_v_state);
    __pyx_t_14 = PyTuple_New(2); if (unlikely(!__pyx_t_14)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_GIVEREF(__pyx_t_13);
    PyTuple_SET_ITEM(__pyx_t_14, 0, __pyx_t_13);
    __Pyx_GIVEREF(__pyx_t_15);
    PyTuple_SET_ITEM(__pyx_t_14, 1, __pyx_t_15);
    __pyx_t_13 = 0;
    __pyx_t_15 = 0;
    __pyx_r = __pyx_t_14;
    __pyx_t_14 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_AddTraceback("pandas._libs.interval.Int64ClosedNeitherIntervalNode.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Int64ClosedNeitherIntervalNode, (type(self), 0x7a104e5, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Int64ClosedNeitherIntervalNode__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_23__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_23__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_22__setstate_cython__(((struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_22__setstate_cython__(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_Int64ClosedNeitherIntervalNode, (type(self), 0x7a104e5, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Int64ClosedNeitherIntervalNode__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(2, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_6pandas_5_libs_8interval___pyx_unpickle_Int64ClosedNeitherIntervalNode__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Int64ClosedNeitherIntervalNode, (type(self), 0x7a104e5, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Int64ClosedNeitherIntervalNode__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Int64ClosedNeitherIntervalNode.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":2968
 *         readonly bint is_leaf_node
 * 
 *     def __init__(self,             # <<<<<<<<<<<<<<
 *                  ndarray[uint64_t, ndim=1] left,
 *                  ndarray[uint64_t, ndim=1] right,
 */

/* Python wrapper */
static int __pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyArrayObject *__pyx_v_left = 0;
  PyArrayObject *__pyx_v_right = 0;
  PyArrayObject *__pyx_v_indices = 0;
  __pyx_t_5numpy_int64_t __pyx_v_leaf_size;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_left,&__pyx_n_s_right,&__pyx_n_s_indices,&__pyx_n_s_leaf_size,0};
    PyObject* values[4] = {0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_left)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_right)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 4, 4, 1); __PYX_ERR(1, 2968, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_indices)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 4, 4, 2); __PYX_ERR(1, 2968, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_leaf_size)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 4, 4, 3); __PYX_ERR(1, 2968, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(1, 2968, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 4) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
    }
    __pyx_v_left = ((PyArrayObject *)values[0]);
    __pyx_v_right = ((PyArrayObject *)values[1]);
    __pyx_v_indices = ((PyArrayObject *)values[2]);
    __pyx_v_leaf_size = __Pyx_PyInt_As_npy_int64(values[3]); if (unlikely((__pyx_v_leaf_size == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(1, 2972, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 4, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 2968, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Uint64ClosedLeftIntervalNode.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_left), __pyx_ptype_5numpy_ndarray, 1, "left", 0))) __PYX_ERR(1, 2969, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_right), __pyx_ptype_5numpy_ndarray, 1, "right", 0))) __PYX_ERR(1, 2970, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_indices), __pyx_ptype_5numpy_ndarray, 1, "indices", 0))) __PYX_ERR(1, 2971, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode___init__(((struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode *)__pyx_v_self), __pyx_v_left, __pyx_v_right, __pyx_v_indices, __pyx_v_leaf_size);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode___init__(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode *__pyx_v_self, PyArrayObject *__pyx_v_left, PyArrayObject *__pyx_v_right, PyArrayObject *__pyx_v_indices, __pyx_t_5numpy_int64_t __pyx_v_leaf_size) {
  PyObject *__pyx_v_left_set = NULL;
  PyObject *__pyx_v_right_set = NULL;
  PyObject *__pyx_v_center_set = NULL;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_indices;
  __Pyx_Buffer __pyx_pybuffer_indices;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_left;
  __Pyx_Buffer __pyx_pybuffer_left;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_right;
  __Pyx_Buffer __pyx_pybuffer_right;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __pyx_t_5numpy_uint64_t __pyx_t_6;
  __Pyx_memviewslice __pyx_t_7 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_8 = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  __Pyx_memviewslice __pyx_t_11 = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *(*__pyx_t_12)(PyObject *);
  size_t __pyx_t_13;
  __Pyx_RefNannySetupContext("__init__", 0);
  __pyx_pybuffer_left.pybuffer.buf = NULL;
  __pyx_pybuffer_left.refcount = 0;
  __pyx_pybuffernd_left.data = NULL;
  __pyx_pybuffernd_left.rcbuffer = &__pyx_pybuffer_left;
  __pyx_pybuffer_right.pybuffer.buf = NULL;
  __pyx_pybuffer_right.refcount = 0;
  __pyx_pybuffernd_right.data = NULL;
  __pyx_pybuffernd_right.rcbuffer = &__pyx_pybuffer_right;
  __pyx_pybuffer_indices.pybuffer.buf = NULL;
  __pyx_pybuffer_indices.refcount = 0;
  __pyx_pybuffernd_indices.data = NULL;
  __pyx_pybuffernd_indices.rcbuffer = &__pyx_pybuffer_indices;
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_left.rcbuffer->pybuffer, (PyObject*)__pyx_v_left, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 2968, __pyx_L1_error)
  }
  __pyx_pybuffernd_left.diminfo[0].strides = __pyx_pybuffernd_left.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_left.diminfo[0].shape = __pyx_pybuffernd_left.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_right.rcbuffer->pybuffer, (PyObject*)__pyx_v_right, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 2968, __pyx_L1_error)
  }
  __pyx_pybuffernd_right.diminfo[0].strides = __pyx_pybuffernd_right.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_right.diminfo[0].shape = __pyx_pybuffernd_right.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_indices.rcbuffer->pybuffer, (PyObject*)__pyx_v_indices, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 2968, __pyx_L1_error)
  }
  __pyx_pybuffernd_indices.diminfo[0].strides = __pyx_pybuffernd_indices.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_indices.diminfo[0].shape = __pyx_pybuffernd_indices.rcbuffer->pybuffer.shape[0];

  /* "pandas/_libs/intervaltree.pxi":2974
 *                  int64_t leaf_size):
 * 
 *         self.n_elements = len(left)             # <<<<<<<<<<<<<<
 *         self.leaf_size = leaf_size
 * 
 */
  __pyx_t_1 = PyObject_Length(((PyObject *)__pyx_v_left)); if (unlikely(__pyx_t_1 == ((Py_ssize_t)-1))) __PYX_ERR(1, 2974, __pyx_L1_error)
  __pyx_v_self->n_elements = __pyx_t_1;

  /* "pandas/_libs/intervaltree.pxi":2975
 * 
 *         self.n_elements = len(left)
 *         self.leaf_size = leaf_size             # <<<<<<<<<<<<<<
 * 
 *         # min_left and min_right are used to speed-up query by skipping
 */
  __pyx_v_self->leaf_size = __pyx_v_leaf_size;

  /* "pandas/_libs/intervaltree.pxi":2980
 *         # query on sub-nodes. If this node has size 0, query is cheap,
 *         # so these values don't matter.
 *         if left.size > 0:             # <<<<<<<<<<<<<<
 *             self.min_left = left.min()
 *             self.max_right = right.max()
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_left), __pyx_n_s_size); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2980, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyObject_RichCompare(__pyx_t_2, __pyx_int_0, Py_GT); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2980, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(1, 2980, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_4) {

    /* "pandas/_libs/intervaltree.pxi":2981
 *         # so these values don't matter.
 *         if left.size > 0:
 *             self.min_left = left.min()             # <<<<<<<<<<<<<<
 *             self.max_right = right.max()
 *         else:
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_left), __pyx_n_s_min); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2981, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2981, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_6 = __Pyx_PyInt_As_npy_uint64(__pyx_t_3); if (unlikely((__pyx_t_6 == ((npy_uint64)-1)) && PyErr_Occurred())) __PYX_ERR(1, 2981, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_self->min_left = __pyx_t_6;

    /* "pandas/_libs/intervaltree.pxi":2982
 *         if left.size > 0:
 *             self.min_left = left.min()
 *             self.max_right = right.max()             # <<<<<<<<<<<<<<
 *         else:
 *             self.min_left = 0
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_right), __pyx_n_s_max); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2982, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2982, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_6 = __Pyx_PyInt_As_npy_uint64(__pyx_t_3); if (unlikely((__pyx_t_6 == ((npy_uint64)-1)) && PyErr_Occurred())) __PYX_ERR(1, 2982, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_self->max_right = __pyx_t_6;

    /* "pandas/_libs/intervaltree.pxi":2980
 *         # query on sub-nodes. If this node has size 0, query is cheap,
 *         # so these values don't matter.
 *         if left.size > 0:             # <<<<<<<<<<<<<<
 *             self.min_left = left.min()
 *             self.max_right = right.max()
 */
    goto __pyx_L3;
  }

  /* "pandas/_libs/intervaltree.pxi":2984
 *             self.max_right = right.max()
 *         else:
 *             self.min_left = 0             # <<<<<<<<<<<<<<
 *             self.max_right = 0
 * 
 */
  /*else*/ {
    __pyx_v_self->min_left = 0;

    /* "pandas/_libs/intervaltree.pxi":2985
 *         else:
 *             self.min_left = 0
 *             self.max_right = 0             # <<<<<<<<<<<<<<
 * 
 *         if self.n_elements <= leaf_size:
 */
    __pyx_v_self->max_right = 0;
  }
  __pyx_L3:;

  /* "pandas/_libs/intervaltree.pxi":2987
 *             self.max_right = 0
 * 
 *         if self.n_elements <= leaf_size:             # <<<<<<<<<<<<<<
 *             # make this a terminal (leaf) node
 *             self.is_leaf_node = True
 */
  __pyx_t_4 = ((__pyx_v_self->n_elements <= __pyx_v_leaf_size) != 0);
  if (__pyx_t_4) {

    /* "pandas/_libs/intervaltree.pxi":2989
 *         if self.n_elements <= leaf_size:
 *             # make this a terminal (leaf) node
 *             self.is_leaf_node = True             # <<<<<<<<<<<<<<
 *             self.left = left
 *             self.right = right
 */
    __pyx_v_self->is_leaf_node = 1;

    /* "pandas/_libs/intervaltree.pxi":2990
 *             # make this a terminal (leaf) node
 *             self.is_leaf_node = True
 *             self.left = left             # <<<<<<<<<<<<<<
 *             self.right = right
 *             self.indices = indices
 */
    __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_uint64_t(((PyObject *)__pyx_v_left), PyBUF_WRITABLE); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(1, 2990, __pyx_L1_error)
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->left, 0);
    __pyx_v_self->left = __pyx_t_7;
    __pyx_t_7.memview = NULL;
    __pyx_t_7.data = NULL;

    /* "pandas/_libs/intervaltree.pxi":2991
 *             self.is_leaf_node = True
 *             self.left = left
 *             self.right = right             # <<<<<<<<<<<<<<
 *             self.indices = indices
 *             self.n_center = 0
 */
    __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_uint64_t(((PyObject *)__pyx_v_right), PyBUF_WRITABLE); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(1, 2991, __pyx_L1_error)
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->right, 0);
    __pyx_v_self->right = __pyx_t_7;
    __pyx_t_7.memview = NULL;
    __pyx_t_7.data = NULL;

    /* "pandas/_libs/intervaltree.pxi":2992
 *             self.left = left
 *             self.right = right
 *             self.indices = indices             # <<<<<<<<<<<<<<
 *             self.n_center = 0
 *         else:
 */
    __pyx_t_8 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(((PyObject *)__pyx_v_indices), PyBUF_WRITABLE); if (unlikely(!__pyx_t_8.memview)) __PYX_ERR(1, 2992, __pyx_L1_error)
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->indices, 0);
    __pyx_v_self->indices = __pyx_t_8;
    __pyx_t_8.memview = NULL;
    __pyx_t_8.data = NULL;

    /* "pandas/_libs/intervaltree.pxi":2993
 *             self.right = right
 *             self.indices = indices
 *             self.n_center = 0             # <<<<<<<<<<<<<<
 *         else:
 *             # calculate a pivot so we can create child nodes
 */
    __pyx_v_self->n_center = 0;

    /* "pandas/_libs/intervaltree.pxi":2987
 *             self.max_right = 0
 * 
 *         if self.n_elements <= leaf_size:             # <<<<<<<<<<<<<<
 *             # make this a terminal (leaf) node
 *             self.is_leaf_node = True
 */
    goto __pyx_L4;
  }

  /* "pandas/_libs/intervaltree.pxi":2996
 *         else:
 *             # calculate a pivot so we can create child nodes
 *             self.is_leaf_node = False             # <<<<<<<<<<<<<<
 *             self.pivot = np.median(left / 2 + right / 2)
 *             left_set, right_set, center_set = self.classify_intervals(
 */
  /*else*/ {
    __pyx_v_self->is_leaf_node = 0;

    /* "pandas/_libs/intervaltree.pxi":2997
 *             # calculate a pivot so we can create child nodes
 *             self.is_leaf_node = False
 *             self.pivot = np.median(left / 2 + right / 2)             # <<<<<<<<<<<<<<
 *             left_set, right_set, center_set = self.classify_intervals(
 *                 left, right)
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2997, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_median); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 2997, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyNumber_Divide(((PyObject *)__pyx_v_left), __pyx_int_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2997, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_9 = __Pyx_PyNumber_Divide(((PyObject *)__pyx_v_right), __pyx_int_2); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 2997, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_10 = PyNumber_Add(__pyx_t_2, __pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 2997, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_9 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_3 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_9, __pyx_t_10) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_10);
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2997, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_6 = __Pyx_PyInt_As_npy_uint64(__pyx_t_3); if (unlikely((__pyx_t_6 == ((npy_uint64)-1)) && PyErr_Occurred())) __PYX_ERR(1, 2997, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_self->pivot = __pyx_t_6;

    /* "pandas/_libs/intervaltree.pxi":2999
 *             self.pivot = np.median(left / 2 + right / 2)
 *             left_set, right_set, center_set = self.classify_intervals(
 *                 left, right)             # <<<<<<<<<<<<<<
 * 
 *             self.left_node = self.new_child_node(left, right,
 */
    __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_uint64_t(((PyObject *)__pyx_v_left), PyBUF_WRITABLE); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(1, 2999, __pyx_L1_error)
    __pyx_t_11 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_uint64_t(((PyObject *)__pyx_v_right), PyBUF_WRITABLE); if (unlikely(!__pyx_t_11.memview)) __PYX_ERR(1, 2999, __pyx_L1_error)

    /* "pandas/_libs/intervaltree.pxi":2998
 *             self.is_leaf_node = False
 *             self.pivot = np.median(left / 2 + right / 2)
 *             left_set, right_set, center_set = self.classify_intervals(             # <<<<<<<<<<<<<<
 *                 left, right)
 * 
 */
    __pyx_t_3 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode *)__pyx_v_self->__pyx_vtab)->classify_intervals(__pyx_v_self, __pyx_t_7, __pyx_t_11); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 2998, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __PYX_XDEC_MEMVIEW(&__pyx_t_7, 1);
    __pyx_t_7.memview = NULL;
    __pyx_t_7.data = NULL;
    __PYX_XDEC_MEMVIEW(&__pyx_t_11, 1);
    __pyx_t_11.memview = NULL;
    __pyx_t_11.data = NULL;
    if ((likely(PyTuple_CheckExact(__pyx_t_3))) || (PyList_CheckExact(__pyx_t_3))) {
      PyObject* sequence = __pyx_t_3;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 3)) {
        if (size > 3) __Pyx_RaiseTooManyValuesError(3);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(1, 2998, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_5 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_10 = PyTuple_GET_ITEM(sequence, 1); 
        __pyx_t_9 = PyTuple_GET_ITEM(sequence, 2); 
      } else {
        __pyx_t_5 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_10 = PyList_GET_ITEM(sequence, 1); 
        __pyx_t_9 = PyList_GET_ITEM(sequence, 2); 
      }
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_10);
      __Pyx_INCREF(__pyx_t_9);
      #else
      __pyx_t_5 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 2998, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_10 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 2998, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_9 = PySequence_ITEM(sequence, 2); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 2998, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      #endif
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_2 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2998, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_12 = Py_TYPE(__pyx_t_2)->tp_iternext;
      index = 0; __pyx_t_5 = __pyx_t_12(__pyx_t_2); if (unlikely(!__pyx_t_5)) goto __pyx_L5_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_5);
      index = 1; __pyx_t_10 = __pyx_t_12(__pyx_t_2); if (unlikely(!__pyx_t_10)) goto __pyx_L5_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_10);
      index = 2; __pyx_t_9 = __pyx_t_12(__pyx_t_2); if (unlikely(!__pyx_t_9)) goto __pyx_L5_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_9);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_12(__pyx_t_2), 3) < 0) __PYX_ERR(1, 2998, __pyx_L1_error)
      __pyx_t_12 = NULL;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      goto __pyx_L6_unpacking_done;
      __pyx_L5_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_12 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(1, 2998, __pyx_L1_error)
      __pyx_L6_unpacking_done:;
    }
    __pyx_v_left_set = __pyx_t_5;
    __pyx_t_5 = 0;
    __pyx_v_right_set = __pyx_t_10;
    __pyx_t_10 = 0;
    __pyx_v_center_set = __pyx_t_9;
    __pyx_t_9 = 0;

    /* "pandas/_libs/intervaltree.pxi":3002
 * 
 *             self.left_node = self.new_child_node(left, right,
 *                                                  indices, left_set)             # <<<<<<<<<<<<<<
 *             self.right_node = self.new_child_node(left, right,
 *                                                   indices, right_set)
 */
    if (!(likely(((__pyx_v_left_set) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_left_set, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(1, 3002, __pyx_L1_error)

    /* "pandas/_libs/intervaltree.pxi":3001
 *                 left, right)
 * 
 *             self.left_node = self.new_child_node(left, right,             # <<<<<<<<<<<<<<
 *                                                  indices, left_set)
 *             self.right_node = self.new_child_node(left, right,
 */
    __pyx_t_3 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode *)__pyx_v_self->__pyx_vtab)->new_child_node(__pyx_v_self, ((PyArrayObject *)__pyx_v_left), ((PyArrayObject *)__pyx_v_right), ((PyArrayObject *)__pyx_v_indices), ((PyArrayObject *)__pyx_v_left_set)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 3001, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode))))) __PYX_ERR(1, 3001, __pyx_L1_error)
    __Pyx_GIVEREF(__pyx_t_3);
    __Pyx_GOTREF(__pyx_v_self->left_node);
    __Pyx_DECREF(((PyObject *)__pyx_v_self->left_node));
    __pyx_v_self->left_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode *)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "pandas/_libs/intervaltree.pxi":3004
 *                                                  indices, left_set)
 *             self.right_node = self.new_child_node(left, right,
 *                                                   indices, right_set)             # <<<<<<<<<<<<<<
 * 
 *             self.center_left_values, self.center_left_indices = \
 */
    if (!(likely(((__pyx_v_right_set) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_right_set, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(1, 3004, __pyx_L1_error)

    /* "pandas/_libs/intervaltree.pxi":3003
 *             self.left_node = self.new_child_node(left, right,
 *                                                  indices, left_set)
 *             self.right_node = self.new_child_node(left, right,             # <<<<<<<<<<<<<<
 *                                                   indices, right_set)
 * 
 */
    __pyx_t_3 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode *)__pyx_v_self->__pyx_vtab)->new_child_node(__pyx_v_self, ((PyArrayObject *)__pyx_v_left), ((PyArrayObject *)__pyx_v_right), ((PyArrayObject *)__pyx_v_indices), ((PyArrayObject *)__pyx_v_right_set)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 3003, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode))))) __PYX_ERR(1, 3003, __pyx_L1_error)
    __Pyx_GIVEREF(__pyx_t_3);
    __Pyx_GOTREF(__pyx_v_self->right_node);
    __Pyx_DECREF(((PyObject *)__pyx_v_self->right_node));
    __pyx_v_self->right_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode *)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "pandas/_libs/intervaltree.pxi":3007
 * 
 *             self.center_left_values, self.center_left_indices = \
 *                 sort_values_and_indices(left, indices, center_set)             # <<<<<<<<<<<<<<
 *             self.center_right_values, self.center_right_indices = \
 *                 sort_values_and_indices(right, indices, center_set)
 */
    __pyx_t_3 = __pyx_f_6pandas_5_libs_8interval_sort_values_and_indices(((PyObject *)__pyx_v_left), ((PyObject *)__pyx_v_indices), __pyx_v_center_set); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 3007, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if ((likely(PyTuple_CheckExact(__pyx_t_3))) || (PyList_CheckExact(__pyx_t_3))) {
      PyObject* sequence = __pyx_t_3;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(1, 3006, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_9 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_10 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_9 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_10 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_9);
      __Pyx_INCREF(__pyx_t_10);
      #else
      __pyx_t_9 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 3006, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_10 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 3006, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      #endif
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_5 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 3006, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_12 = Py_TYPE(__pyx_t_5)->tp_iternext;
      index = 0; __pyx_t_9 = __pyx_t_12(__pyx_t_5); if (unlikely(!__pyx_t_9)) goto __pyx_L7_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_9);
      index = 1; __pyx_t_10 = __pyx_t_12(__pyx_t_5); if (unlikely(!__pyx_t_10)) goto __pyx_L7_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_10);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_12(__pyx_t_5), 2) < 0) __PYX_ERR(1, 3006, __pyx_L1_error)
      __pyx_t_12 = NULL;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      goto __pyx_L8_unpacking_done;
      __pyx_L7_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_12 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(1, 3006, __pyx_L1_error)
      __pyx_L8_unpacking_done:;
    }

    /* "pandas/_libs/intervaltree.pxi":3006
 *                                                   indices, right_set)
 * 
 *             self.center_left_values, self.center_left_indices = \             # <<<<<<<<<<<<<<
 *                 sort_values_and_indices(left, indices, center_set)
 *             self.center_right_values, self.center_right_indices = \
 */
    __pyx_t_11 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_uint64_t(__pyx_t_9, PyBUF_WRITABLE); if (unlikely(!__pyx_t_11.memview)) __PYX_ERR(1, 3006, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_8 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_t_10, PyBUF_WRITABLE); if (unlikely(!__pyx_t_8.memview)) __PYX_ERR(1, 3006, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->center_left_values, 0);
    __pyx_v_self->center_left_values = __pyx_t_11;
    __pyx_t_11.memview = NULL;
    __pyx_t_11.data = NULL;
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->center_left_indices, 0);
    __pyx_v_self->center_left_indices = __pyx_t_8;
    __pyx_t_8.memview = NULL;
    __pyx_t_8.data = NULL;

    /* "pandas/_libs/intervaltree.pxi":3009
 *                 sort_values_and_indices(left, indices, center_set)
 *             self.center_right_values, self.center_right_indices = \
 *                 sort_values_and_indices(right, indices, center_set)             # <<<<<<<<<<<<<<
 *             self.n_center = len(self.center_left_indices)
 * 
 */
    __pyx_t_3 = __pyx_f_6pandas_5_libs_8interval_sort_values_and_indices(((PyObject *)__pyx_v_right), ((PyObject *)__pyx_v_indices), __pyx_v_center_set); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 3009, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if ((likely(PyTuple_CheckExact(__pyx_t_3))) || (PyList_CheckExact(__pyx_t_3))) {
      PyObject* sequence = __pyx_t_3;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(1, 3008, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_10 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_9 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_10 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_9 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_10);
      __Pyx_INCREF(__pyx_t_9);
      #else
      __pyx_t_10 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 3008, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_9 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 3008, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      #endif
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_5 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 3008, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_12 = Py_TYPE(__pyx_t_5)->tp_iternext;
      index = 0; __pyx_t_10 = __pyx_t_12(__pyx_t_5); if (unlikely(!__pyx_t_10)) goto __pyx_L9_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_10);
      index = 1; __pyx_t_9 = __pyx_t_12(__pyx_t_5); if (unlikely(!__pyx_t_9)) goto __pyx_L9_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_9);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_12(__pyx_t_5), 2) < 0) __PYX_ERR(1, 3008, __pyx_L1_error)
      __pyx_t_12 = NULL;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      goto __pyx_L10_unpacking_done;
      __pyx_L9_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_12 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(1, 3008, __pyx_L1_error)
      __pyx_L10_unpacking_done:;
    }

    /* "pandas/_libs/intervaltree.pxi":3008
 *             self.center_left_values, self.center_left_indices = \
 *                 sort_values_and_indices(left, indices, center_set)
 *             self.center_right_values, self.center_right_indices = \             # <<<<<<<<<<<<<<
 *                 sort_values_and_indices(right, indices, center_set)
 *             self.n_center = len(self.center_left_indices)
 */
    __pyx_t_11 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_uint64_t(__pyx_t_10, PyBUF_WRITABLE); if (unlikely(!__pyx_t_11.memview)) __PYX_ERR(1, 3008, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_t_8 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_t_9, PyBUF_WRITABLE); if (unlikely(!__pyx_t_8.memview)) __PYX_ERR(1, 3008, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->center_right_values, 0);
    __pyx_v_self->center_right_values = __pyx_t_11;
    __pyx_t_11.memview = NULL;
    __pyx_t_11.data = NULL;
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->center_right_indices, 0);
    __pyx_v_self->center_right_indices = __pyx_t_8;
    __pyx_t_8.memview = NULL;
    __pyx_t_8.data = NULL;

    /* "pandas/_libs/intervaltree.pxi":3010
 *             self.center_right_values, self.center_right_indices = \
 *                 sort_values_and_indices(right, indices, center_set)
 *             self.n_center = len(self.center_left_indices)             # <<<<<<<<<<<<<<
 * 
 *     @cython.wraparound(False)
 */
    if (unlikely(!__pyx_v_self->center_left_indices.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(1, 3010, __pyx_L1_error)}
    __pyx_t_13 = __Pyx_MemoryView_Len(__pyx_v_self->center_left_indices); 
    __pyx_v_self->n_center = __pyx_t_13;
  }
  __pyx_L4:;

  /* "pandas/_libs/intervaltree.pxi":2968
 *         readonly bint is_leaf_node
 * 
 *     def __init__(self,             # <<<<<<<<<<<<<<
 *                  ndarray[uint64_t, ndim=1] left,
 *                  ndarray[uint64_t, ndim=1] right,
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __PYX_XDEC_MEMVIEW(&__pyx_t_7, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_8, 1);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __PYX_XDEC_MEMVIEW(&__pyx_t_11, 1);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indices.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_left.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_right.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("pandas._libs.interval.Uint64ClosedLeftIntervalNode.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indices.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_left.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_right.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF(__pyx_v_left_set);
  __Pyx_XDECREF(__pyx_v_right_set);
  __Pyx_XDECREF(__pyx_v_center_set);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":3014
 *     @cython.wraparound(False)
 *     @cython.boundscheck(False)
 *     cdef classify_intervals(self, uint64_t[:] left, uint64_t[:] right):             # <<<<<<<<<<<<<<
 *         """Classify the given intervals based upon whether they fall to the
 *         left, right, or overlap with this node's pivot.
 */

static PyObject *__pyx_f_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_classify_intervals(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode *__pyx_v_self, __Pyx_memviewslice __pyx_v_left, __Pyx_memviewslice __pyx_v_right) {
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_left_ind = 0;
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_right_ind = 0;
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_overlapping_ind = 0;
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __pyx_t_5numpy_int64_t __pyx_t_2;
  __pyx_t_5numpy_int64_t __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  int __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  __Pyx_RefNannySetupContext("classify_intervals", 0);

  /* "pandas/_libs/intervaltree.pxi":3022
 *             Py_ssize_t i
 * 
 *         left_ind = Int64Vector()             # <<<<<<<<<<<<<<
 *         right_ind = Int64Vector()
 *         overlapping_ind = Int64Vector()
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3022, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_left_ind = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":3023
 * 
 *         left_ind = Int64Vector()
 *         right_ind = Int64Vector()             # <<<<<<<<<<<<<<
 *         overlapping_ind = Int64Vector()
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3023, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_right_ind = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":3024
 *         left_ind = Int64Vector()
 *         right_ind = Int64Vector()
 *         overlapping_ind = Int64Vector()             # <<<<<<<<<<<<<<
 * 
 *         for i in range(self.n_elements):
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3024, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_overlapping_ind = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":3026
 *         overlapping_ind = Int64Vector()
 * 
 *         for i in range(self.n_elements):             # <<<<<<<<<<<<<<
 *             if right[i] <= self.pivot:
 *                 left_ind.append(i)
 */
  __pyx_t_2 = __pyx_v_self->n_elements;
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "pandas/_libs/intervaltree.pxi":3027
 * 
 *         for i in range(self.n_elements):
 *             if right[i] <= self.pivot:             # <<<<<<<<<<<<<<
 *                 left_ind.append(i)
 *             elif self.pivot < left[i]:
 */
    __pyx_t_5 = __pyx_v_i;
    __pyx_t_6 = (((*((__pyx_t_5numpy_uint64_t *) ( /* dim=0 */ (__pyx_v_right.data + __pyx_t_5 * __pyx_v_right.strides[0]) ))) <= __pyx_v_self->pivot) != 0);
    if (__pyx_t_6) {

      /* "pandas/_libs/intervaltree.pxi":3028
 *         for i in range(self.n_elements):
 *             if right[i] <= self.pivot:
 *                 left_ind.append(i)             # <<<<<<<<<<<<<<
 *             elif self.pivot < left[i]:
 *                 right_ind.append(i)
 */
      ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_left_ind->__pyx_vtab)->append(__pyx_v_left_ind, __pyx_v_i);

      /* "pandas/_libs/intervaltree.pxi":3027
 * 
 *         for i in range(self.n_elements):
 *             if right[i] <= self.pivot:             # <<<<<<<<<<<<<<
 *                 left_ind.append(i)
 *             elif self.pivot < left[i]:
 */
      goto __pyx_L5;
    }

    /* "pandas/_libs/intervaltree.pxi":3029
 *             if right[i] <= self.pivot:
 *                 left_ind.append(i)
 *             elif self.pivot < left[i]:             # <<<<<<<<<<<<<<
 *                 right_ind.append(i)
 *             else:
 */
    __pyx_t_7 = __pyx_v_i;
    __pyx_t_6 = ((__pyx_v_self->pivot < (*((__pyx_t_5numpy_uint64_t *) ( /* dim=0 */ (__pyx_v_left.data + __pyx_t_7 * __pyx_v_left.strides[0]) )))) != 0);
    if (__pyx_t_6) {

      /* "pandas/_libs/intervaltree.pxi":3030
 *                 left_ind.append(i)
 *             elif self.pivot < left[i]:
 *                 right_ind.append(i)             # <<<<<<<<<<<<<<
 *             else:
 *                 overlapping_ind.append(i)
 */
      ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_right_ind->__pyx_vtab)->append(__pyx_v_right_ind, __pyx_v_i);

      /* "pandas/_libs/intervaltree.pxi":3029
 *             if right[i] <= self.pivot:
 *                 left_ind.append(i)
 *             elif self.pivot < left[i]:             # <<<<<<<<<<<<<<
 *                 right_ind.append(i)
 *             else:
 */
      goto __pyx_L5;
    }

    /* "pandas/_libs/intervaltree.pxi":3032
 *                 right_ind.append(i)
 *             else:
 *                 overlapping_ind.append(i)             # <<<<<<<<<<<<<<
 * 
 *         return (left_ind.to_array(),
 */
    /*else*/ {
      ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_overlapping_ind->__pyx_vtab)->append(__pyx_v_overlapping_ind, __pyx_v_i);
    }
    __pyx_L5:;
  }

  /* "pandas/_libs/intervaltree.pxi":3034
 *                 overlapping_ind.append(i)
 * 
 *         return (left_ind.to_array(),             # <<<<<<<<<<<<<<
 *                 right_ind.to_array(),
 *                 overlapping_ind.to_array())
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_left_ind->__pyx_vtab)->to_array(__pyx_v_left_ind, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3034, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "pandas/_libs/intervaltree.pxi":3035
 * 
 *         return (left_ind.to_array(),
 *                 right_ind.to_array(),             # <<<<<<<<<<<<<<
 *                 overlapping_ind.to_array())
 * 
 */
  __pyx_t_8 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_right_ind->__pyx_vtab)->to_array(__pyx_v_right_ind, 0); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 3035, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);

  /* "pandas/_libs/intervaltree.pxi":3036
 *         return (left_ind.to_array(),
 *                 right_ind.to_array(),
 *                 overlapping_ind.to_array())             # <<<<<<<<<<<<<<
 * 
 *     cdef new_child_node(self,
 */
  __pyx_t_9 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_overlapping_ind->__pyx_vtab)->to_array(__pyx_v_overlapping_ind, 0); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 3036, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);

  /* "pandas/_libs/intervaltree.pxi":3034
 *                 overlapping_ind.append(i)
 * 
 *         return (left_ind.to_array(),             # <<<<<<<<<<<<<<
 *                 right_ind.to_array(),
 *                 overlapping_ind.to_array())
 */
  __pyx_t_10 = PyTuple_New(3); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 3034, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_t_8);
  __Pyx_GIVEREF(__pyx_t_9);
  PyTuple_SET_ITEM(__pyx_t_10, 2, __pyx_t_9);
  __pyx_t_1 = 0;
  __pyx_t_8 = 0;
  __pyx_t_9 = 0;
  __pyx_r = __pyx_t_10;
  __pyx_t_10 = 0;
  goto __pyx_L0;

  /* "pandas/_libs/intervaltree.pxi":3014
 *     @cython.wraparound(False)
 *     @cython.boundscheck(False)
 *     cdef classify_intervals(self, uint64_t[:] left, uint64_t[:] right):             # <<<<<<<<<<<<<<
 *         """Classify the given intervals based upon whether they fall to the
 *         left, right, or overlap with this node's pivot.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("pandas._libs.interval.Uint64ClosedLeftIntervalNode.classify_intervals", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_left_ind);
  __Pyx_XDECREF((PyObject *)__pyx_v_right_ind);
  __Pyx_XDECREF((PyObject *)__pyx_v_overlapping_ind);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":3038
 *                 overlapping_ind.to_array())
 * 
 *     cdef new_child_node(self,             # <<<<<<<<<<<<<<
 *                         ndarray[uint64_t, ndim=1] left,
 *                         ndarray[uint64_t, ndim=1] right,
 */

static PyObject *__pyx_f_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_new_child_node(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode *__pyx_v_self, PyArrayObject *__pyx_v_left, PyArrayObject *__pyx_v_right, PyArrayObject *__pyx_v_indices, PyArrayObject *__pyx_v_subset) {
  __Pyx_LocalBuf_ND __pyx_pybuffernd_indices;
  __Pyx_Buffer __pyx_pybuffer_indices;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_left;
  __Pyx_Buffer __pyx_pybuffer_left;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_right;
  __Pyx_Buffer __pyx_pybuffer_right;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_subset;
  __Pyx_Buffer __pyx_pybuffer_subset;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyArrayObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyArrayObject *__pyx_t_7 = NULL;
  PyArrayObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  __Pyx_RefNannySetupContext("new_child_node", 0);
  __Pyx_INCREF((PyObject *)__pyx_v_left);
  __Pyx_INCREF((PyObject *)__pyx_v_right);
  __Pyx_INCREF((PyObject *)__pyx_v_indices);
  __pyx_pybuffer_left.pybuffer.buf = NULL;
  __pyx_pybuffer_left.refcount = 0;
  __pyx_pybuffernd_left.data = NULL;
  __pyx_pybuffernd_left.rcbuffer = &__pyx_pybuffer_left;
  __pyx_pybuffer_right.pybuffer.buf = NULL;
  __pyx_pybuffer_right.refcount = 0;
  __pyx_pybuffernd_right.data = NULL;
  __pyx_pybuffernd_right.rcbuffer = &__pyx_pybuffer_right;
  __pyx_pybuffer_indices.pybuffer.buf = NULL;
  __pyx_pybuffer_indices.refcount = 0;
  __pyx_pybuffernd_indices.data = NULL;
  __pyx_pybuffernd_indices.rcbuffer = &__pyx_pybuffer_indices;
  __pyx_pybuffer_subset.pybuffer.buf = NULL;
  __pyx_pybuffer_subset.refcount = 0;
  __pyx_pybuffernd_subset.data = NULL;
  __pyx_pybuffernd_subset.rcbuffer = &__pyx_pybuffer_subset;
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_left.rcbuffer->pybuffer, (PyObject*)__pyx_v_left, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 3038, __pyx_L1_error)
  }
  __pyx_pybuffernd_left.diminfo[0].strides = __pyx_pybuffernd_left.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_left.diminfo[0].shape = __pyx_pybuffernd_left.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_right.rcbuffer->pybuffer, (PyObject*)__pyx_v_right, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 3038, __pyx_L1_error)
  }
  __pyx_pybuffernd_right.diminfo[0].strides = __pyx_pybuffernd_right.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_right.diminfo[0].shape = __pyx_pybuffernd_right.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_indices.rcbuffer->pybuffer, (PyObject*)__pyx_v_indices, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 3038, __pyx_L1_error)
  }
  __pyx_pybuffernd_indices.diminfo[0].strides = __pyx_pybuffernd_indices.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_indices.diminfo[0].shape = __pyx_pybuffernd_indices.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_subset.rcbuffer->pybuffer, (PyObject*)__pyx_v_subset, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 3038, __pyx_L1_error)
  }
  __pyx_pybuffernd_subset.diminfo[0].strides = __pyx_pybuffernd_subset.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_subset.diminfo[0].shape = __pyx_pybuffernd_subset.rcbuffer->pybuffer.shape[0];

  /* "pandas/_libs/intervaltree.pxi":3045
 *         """Create a new child node.
 *         """
 *         left = take(left, subset)             # <<<<<<<<<<<<<<
 *         right = take(right, subset)
 *         indices = take(indices, subset)
 */
  __pyx_t_1 = __pyx_f_6pandas_5_libs_8interval_take(((PyArrayObject *)__pyx_v_left), ((PyArrayObject *)__pyx_v_subset)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3045, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(1, 3045, __pyx_L1_error)
  __pyx_t_2 = ((PyArrayObject *)__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_left.rcbuffer->pybuffer);
    __pyx_t_3 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_left.rcbuffer->pybuffer, (PyObject*)__pyx_t_2, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack);
    if (unlikely(__pyx_t_3 < 0)) {
      PyErr_Fetch(&__pyx_t_4, &__pyx_t_5, &__pyx_t_6);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_left.rcbuffer->pybuffer, (PyObject*)__pyx_v_left, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_4); Py_XDECREF(__pyx_t_5); Py_XDECREF(__pyx_t_6);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_4, __pyx_t_5, __pyx_t_6);
      }
      __pyx_t_4 = __pyx_t_5 = __pyx_t_6 = 0;
    }
    __pyx_pybuffernd_left.diminfo[0].strides = __pyx_pybuffernd_left.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_left.diminfo[0].shape = __pyx_pybuffernd_left.rcbuffer->pybuffer.shape[0];
    if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(1, 3045, __pyx_L1_error)
  }
  __pyx_t_2 = 0;
  __Pyx_DECREF_SET(__pyx_v_left, ((PyArrayObject *)__pyx_t_1));
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":3046
 *         """
 *         left = take(left, subset)
 *         right = take(right, subset)             # <<<<<<<<<<<<<<
 *         indices = take(indices, subset)
 *         return Uint64ClosedLeftIntervalNode(
 */
  __pyx_t_1 = __pyx_f_6pandas_5_libs_8interval_take(((PyArrayObject *)__pyx_v_right), ((PyArrayObject *)__pyx_v_subset)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3046, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(1, 3046, __pyx_L1_error)
  __pyx_t_7 = ((PyArrayObject *)__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_right.rcbuffer->pybuffer);
    __pyx_t_3 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_right.rcbuffer->pybuffer, (PyObject*)__pyx_t_7, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack);
    if (unlikely(__pyx_t_3 < 0)) {
      PyErr_Fetch(&__pyx_t_6, &__pyx_t_5, &__pyx_t_4);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_right.rcbuffer->pybuffer, (PyObject*)__pyx_v_right, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_6); Py_XDECREF(__pyx_t_5); Py_XDECREF(__pyx_t_4);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_6, __pyx_t_5, __pyx_t_4);
      }
      __pyx_t_6 = __pyx_t_5 = __pyx_t_4 = 0;
    }
    __pyx_pybuffernd_right.diminfo[0].strides = __pyx_pybuffernd_right.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_right.diminfo[0].shape = __pyx_pybuffernd_right.rcbuffer->pybuffer.shape[0];
    if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(1, 3046, __pyx_L1_error)
  }
  __pyx_t_7 = 0;
  __Pyx_DECREF_SET(__pyx_v_right, ((PyArrayObject *)__pyx_t_1));
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":3047
 *         left = take(left, subset)
 *         right = take(right, subset)
 *         indices = take(indices, subset)             # <<<<<<<<<<<<<<
 *         return Uint64ClosedLeftIntervalNode(
 *             left, right, indices, self.leaf_size)
 */
  __pyx_t_1 = __pyx_f_6pandas_5_libs_8interval_take(((PyArrayObject *)__pyx_v_indices), ((PyArrayObject *)__pyx_v_subset)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3047, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(1, 3047, __pyx_L1_error)
  __pyx_t_8 = ((PyArrayObject *)__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indices.rcbuffer->pybuffer);
    __pyx_t_3 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_indices.rcbuffer->pybuffer, (PyObject*)__pyx_t_8, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack);
    if (unlikely(__pyx_t_3 < 0)) {
      PyErr_Fetch(&__pyx_t_4, &__pyx_t_5, &__pyx_t_6);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_indices.rcbuffer->pybuffer, (PyObject*)__pyx_v_indices, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_4); Py_XDECREF(__pyx_t_5); Py_XDECREF(__pyx_t_6);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_4, __pyx_t_5, __pyx_t_6);
      }
      __pyx_t_4 = __pyx_t_5 = __pyx_t_6 = 0;
    }
    __pyx_pybuffernd_indices.diminfo[0].strides = __pyx_pybuffernd_indices.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_indices.diminfo[0].shape = __pyx_pybuffernd_indices.rcbuffer->pybuffer.shape[0];
    if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(1, 3047, __pyx_L1_error)
  }
  __pyx_t_8 = 0;
  __Pyx_DECREF_SET(__pyx_v_indices, ((PyArrayObject *)__pyx_t_1));
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":3048
 *         right = take(right, subset)
 *         indices = take(indices, subset)
 *         return Uint64ClosedLeftIntervalNode(             # <<<<<<<<<<<<<<
 *             left, right, indices, self.leaf_size)
 * 
 */
  __Pyx_XDECREF(__pyx_r);

  /* "pandas/_libs/intervaltree.pxi":3049
 *         indices = take(indices, subset)
 *         return Uint64ClosedLeftIntervalNode(
 *             left, right, indices, self.leaf_size)             # <<<<<<<<<<<<<<
 * 
 *     @cython.wraparound(False)
 */
  __pyx_t_1 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->leaf_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3049, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "pandas/_libs/intervaltree.pxi":3048
 *         right = take(right, subset)
 *         indices = take(indices, subset)
 *         return Uint64ClosedLeftIntervalNode(             # <<<<<<<<<<<<<<
 *             left, right, indices, self.leaf_size)
 * 
 */
  __pyx_t_9 = PyTuple_New(4); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 3048, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_INCREF(((PyObject *)__pyx_v_left));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_left));
  PyTuple_SET_ITEM(__pyx_t_9, 0, ((PyObject *)__pyx_v_left));
  __Pyx_INCREF(((PyObject *)__pyx_v_right));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_right));
  PyTuple_SET_ITEM(__pyx_t_9, 1, ((PyObject *)__pyx_v_right));
  __Pyx_INCREF(((PyObject *)__pyx_v_indices));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_indices));
  PyTuple_SET_ITEM(__pyx_t_9, 2, ((PyObject *)__pyx_v_indices));
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_9, 3, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode), __pyx_t_9, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3048, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pandas/_libs/intervaltree.pxi":3038
 *                 overlapping_ind.to_array())
 * 
 *     cdef new_child_node(self,             # <<<<<<<<<<<<<<
 *                         ndarray[uint64_t, ndim=1] left,
 *                         ndarray[uint64_t, ndim=1] right,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_9);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indices.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_left.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_right.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_subset.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("pandas._libs.interval.Uint64ClosedLeftIntervalNode.new_child_node", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indices.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_left.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_right.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_subset.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF((PyObject *)__pyx_v_left);
  __Pyx_XDECREF((PyObject *)__pyx_v_right);
  __Pyx_XDECREF((PyObject *)__pyx_v_indices);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":3054
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_2query[] = "Recursively query this node and its sub-nodes for intervals that\n        overlap with the query point.\n        ";
static PyMethodDef __pyx_mdef_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_3query = {"query", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_3query, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_2query};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_args = 0;
  PyObject *__pyx_v_kwargs = 0;
  CYTHON_UNUSED PyObject *__pyx_v_defaults = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fused_cpdef (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_args,&__pyx_n_s_kwargs,&__pyx_n_s_defaults,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_args)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_kwargs)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fused_cpdef", 1, 3, 3, 1); __PYX_ERR(1, 3054, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_defaults)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fused_cpdef", 1, 3, 3, 2); __PYX_ERR(1, 3054, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fused_cpdef") < 0)) __PYX_ERR(1, 3054, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_args = values[0];
    __pyx_v_kwargs = values[1];
    __pyx_v_defaults = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fused_cpdef", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 3054, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Uint64ClosedLeftIntervalNode.__pyx_fused_cpdef", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_2query(((struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode *)__pyx_v_signatures), __pyx_v_args, __pyx_v_kwargs, __pyx_v_defaults);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_2query(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode *__pyx_v_signatures, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs, CYTHON_UNUSED PyObject *__pyx_v_defaults) {
  PyObject *__pyx_v_dest_sig = NULL;
  Py_ssize_t __pyx_v_i;
  CYTHON_UNUSED PyTypeObject *__pyx_v_ndarray = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_v_itemsize;
  PyObject *__pyx_v_arg = NULL;
  PyObject *__pyx_v_candidates = NULL;
  PyObject *__pyx_v_sig = NULL;
  int __pyx_v_match_found;
  PyObject *__pyx_v_src_sig = NULL;
  PyObject *__pyx_v_dst_type = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  Py_ssize_t __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  Py_ssize_t __pyx_t_13;
  Py_ssize_t __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  int __pyx_t_16;
  __Pyx_RefNannySetupContext("query", 0);
  __Pyx_INCREF(__pyx_v_kwargs);
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3054, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  PyList_SET_ITEM(__pyx_t_1, 0, Py_None);
  __pyx_v_dest_sig = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_3 = (__pyx_v_kwargs != Py_None);
  __pyx_t_4 = (__pyx_t_3 != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_kwargs); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(1, 3054, __pyx_L1_error)
  __pyx_t_3 = ((!__pyx_t_4) != 0);
  __pyx_t_2 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {
    __Pyx_INCREF(Py_None);
    __Pyx_DECREF_SET(__pyx_v_kwargs, Py_None);
  }
  __pyx_t_1 = ((PyObject *)__Pyx_ImportNumPyArrayTypeIfAvailable()); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3054, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_ndarray = ((PyTypeObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_v_itemsize = -1L;
  if (unlikely(__pyx_v_args == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 3054, __pyx_L1_error)
  }
  __pyx_t_5 = PyTuple_GET_SIZE(((PyObject*)__pyx_v_args)); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(1, 3054, __pyx_L1_error)
  __pyx_t_2 = ((2 < __pyx_t_5) != 0);
  if (__pyx_t_2) {
    if (unlikely(__pyx_v_args == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 3054, __pyx_L1_error)
    }
    __pyx_t_1 = PyTuple_GET_ITEM(((PyObject*)__pyx_v_args), 2);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v_arg = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L6;
  }
  __pyx_t_3 = (__pyx_v_kwargs != Py_None);
  __pyx_t_4 = (__pyx_t_3 != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L7_bool_binop_done;
  }
  if (unlikely(__pyx_v_kwargs == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(1, 3054, __pyx_L1_error)
  }
  __pyx_t_4 = (__Pyx_PyDict_ContainsTF(__pyx_n_s_point, ((PyObject*)__pyx_v_kwargs), Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(1, 3054, __pyx_L1_error)
  __pyx_t_3 = (__pyx_t_4 != 0);
  __pyx_t_2 = __pyx_t_3;
  __pyx_L7_bool_binop_done:;
  if (__pyx_t_2) {
    if (unlikely(__pyx_v_kwargs == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 3054, __pyx_L1_error)
    }
    __pyx_t_1 = __Pyx_PyDict_GetItem(((PyObject*)__pyx_v_kwargs), __pyx_n_s_point); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3054, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_arg = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L6;
  }
  /*else*/ {
    if (unlikely(__pyx_v_args == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
      __PYX_ERR(1, 3054, __pyx_L1_error)
    }
    __pyx_t_5 = PyTuple_GET_SIZE(((PyObject*)__pyx_v_args)); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(1, 3054, __pyx_L1_error)
    __pyx_t_1 = PyInt_FromSsize_t(__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3054, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_6 = PyTuple_New(3); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 3054, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_INCREF(__pyx_int_3);
    __Pyx_GIVEREF(__pyx_int_3);
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_int_3);
    __Pyx_INCREF(__pyx_n_s_s);
    __Pyx_GIVEREF(__pyx_n_s_s);
    PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_n_s_s);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Expected_at_least_d_argument_s_g, __pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3054, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_1); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 3054, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(1, 3054, __pyx_L1_error)
  }
  __pyx_L6:;
  while (1) {
    __pyx_t_2 = PyFloat_Check(__pyx_v_arg); 
    __pyx_t_3 = (__pyx_t_2 != 0);
    if (__pyx_t_3) {
      if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, __pyx_n_s_float64_t, long, 1, __Pyx_PyInt_From_long, 1, 0, 0) < 0)) __PYX_ERR(1, 3054, __pyx_L1_error)
      goto __pyx_L10_break;
    }
    __pyx_t_2 = PyInt_Check(__pyx_v_arg); 
    __pyx_t_4 = (__pyx_t_2 != 0);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_3 = __pyx_t_4;
      goto __pyx_L13_bool_binop_done;
    }
    __pyx_t_4 = PyLong_Check(__pyx_v_arg); 
    __pyx_t_2 = (__pyx_t_4 != 0);
    __pyx_t_3 = __pyx_t_2;
    __pyx_L13_bool_binop_done:;
    __pyx_t_2 = (__pyx_t_3 != 0);
    if (__pyx_t_2) {
      if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, __pyx_n_s_int64_t, long, 1, __Pyx_PyInt_From_long, 1, 0, 0) < 0)) __PYX_ERR(1, 3054, __pyx_L1_error)
      goto __pyx_L10_break;
    }
    if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, Py_None, long, 1, __Pyx_PyInt_From_long, 1, 0, 0) < 0)) __PYX_ERR(1, 3054, __pyx_L1_error)
    goto __pyx_L10_break;
  }
  __pyx_L10_break:;
  __pyx_t_6 = PyList_New(0); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 3054, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_v_candidates = ((PyObject*)__pyx_t_6);
  __pyx_t_6 = 0;
  __pyx_t_5 = 0;
  if (unlikely(((PyObject *)__pyx_v_signatures) == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(1, 3054, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_dict_iterator(((PyObject*)__pyx_v_signatures), 1, ((PyObject *)NULL), (&__pyx_t_7), (&__pyx_t_8)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3054, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __pyx_t_6 = __pyx_t_1;
  __pyx_t_1 = 0;
  while (1) {
    __pyx_t_9 = __Pyx_dict_iter_next(__pyx_t_6, __pyx_t_7, &__pyx_t_5, &__pyx_t_1, NULL, NULL, __pyx_t_8);
    if (unlikely(__pyx_t_9 == 0)) break;
    if (unlikely(__pyx_t_9 == -1)) __PYX_ERR(1, 3054, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_sig, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_v_match_found = 0;
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_v_sig, __pyx_n_s_strip); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 3054, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __pyx_t_12 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_11))) {
      __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_11);
      if (likely(__pyx_t_12)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
        __Pyx_INCREF(__pyx_t_12);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_11, function);
      }
    }
    __pyx_t_10 = (__pyx_t_12) ? __Pyx_PyObject_Call2Args(__pyx_t_11, __pyx_t_12, __pyx_kp_s__13) : __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_kp_s__13);
    __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
    if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 3054, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_split); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 3054, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_t_10 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_11))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_11);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_11, function);
      }
    }
    __pyx_t_1 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_11, __pyx_t_10, __pyx_kp_s__14) : __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_kp_s__14);
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3054, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_XDECREF_SET(__pyx_v_src_sig, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_13 = PyList_GET_SIZE(__pyx_v_dest_sig); if (unlikely(__pyx_t_13 == ((Py_ssize_t)-1))) __PYX_ERR(1, 3054, __pyx_L1_error)
    __pyx_t_14 = __pyx_t_13;
    for (__pyx_t_15 = 0; __pyx_t_15 < __pyx_t_14; __pyx_t_15+=1) {
      __pyx_v_i = __pyx_t_15;
      __pyx_t_1 = PyList_GET_ITEM(__pyx_v_dest_sig, __pyx_v_i);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_XDECREF_SET(__pyx_v_dst_type, __pyx_t_1);
      __pyx_t_1 = 0;
      __pyx_t_2 = (__pyx_v_dst_type != Py_None);
      __pyx_t_3 = (__pyx_t_2 != 0);
      if (__pyx_t_3) {
        __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_src_sig, __pyx_v_i, Py_ssize_t, 1, PyInt_FromSsize_t, 0, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3054, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_11 = PyObject_RichCompare(__pyx_t_1, __pyx_v_dst_type, Py_EQ); __Pyx_XGOTREF(__pyx_t_11); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 3054, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_11); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(1, 3054, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        if (__pyx_t_3) {
          __pyx_v_match_found = 1;
          goto __pyx_L20;
        }
        /*else*/ {
          __pyx_v_match_found = 0;
          goto __pyx_L18_break;
        }
        __pyx_L20:;
      }
    }
    __pyx_L18_break:;
    __pyx_t_3 = (__pyx_v_match_found != 0);
    if (__pyx_t_3) {
      __pyx_t_16 = __Pyx_PyList_Append(__pyx_v_candidates, __pyx_v_sig); if (unlikely(__pyx_t_16 == ((int)-1))) __PYX_ERR(1, 3054, __pyx_L1_error)
    }
  }
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_3 = (PyList_GET_SIZE(__pyx_v_candidates) != 0);
  __pyx_t_2 = ((!__pyx_t_3) != 0);
  if (__pyx_t_2) {
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__15, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 3054, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(1, 3054, __pyx_L1_error)
  }
  __pyx_t_7 = PyList_GET_SIZE(__pyx_v_candidates); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1))) __PYX_ERR(1, 3054, __pyx_L1_error)
  __pyx_t_2 = ((__pyx_t_7 > 1) != 0);
  if (__pyx_t_2) {
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__16, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 3054, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(1, 3054, __pyx_L1_error)
  }
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    if (unlikely(((PyObject *)__pyx_v_signatures) == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 3054, __pyx_L1_error)
    }
    __pyx_t_6 = __Pyx_PyDict_GetItem(((PyObject*)__pyx_v_signatures), PyList_GET_ITEM(__pyx_v_candidates, 0)); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 3054, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_r = __pyx_t_6;
    __pyx_t_6 = 0;
    goto __pyx_L0;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("pandas._libs.interval.Uint64ClosedLeftIntervalNode.__pyx_fused_cpdef", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_dest_sig);
  __Pyx_XDECREF(__pyx_v_ndarray);
  __Pyx_XDECREF(__pyx_v_arg);
  __Pyx_XDECREF(__pyx_v_candidates);
  __Pyx_XDECREF(__pyx_v_sig);
  __Pyx_XDECREF(__pyx_v_src_sig);
  __Pyx_XDECREF(__pyx_v_dst_type);
  __Pyx_XDECREF(__pyx_v_kwargs);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_9__pyx_fuse_0query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_fuse_0__pyx_f_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float64_t __pyx_v_point, int __pyx_skip_dispatch) {
  __Pyx_memviewslice __pyx_v_indices = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_values = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __pyx_t_5numpy_int64_t __pyx_t_9;
  __pyx_t_5numpy_int64_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  int __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  __Pyx_memviewslice __pyx_t_16 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_17 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  Py_ssize_t __pyx_t_21;
  __Pyx_RefNannySetupContext("__pyx_fuse_0query", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pyx_fuse_0query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3054, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_9__pyx_fuse_0query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = PyFloat_FromDouble(__pyx_v_point); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 3054, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3054, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3054, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 3054, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_5) {
            __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3054, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3054, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_3query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_4 = PyFloat_FromDouble(__pyx_v_point); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 3054, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_7 = __pyx_t_1; __pyx_t_3 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3054, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3054, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 3054, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (__pyx_t_3) {
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_t_4);
          __pyx_t_4 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3054, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pandas/_libs/intervaltree.pxi":3063
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
  __pyx_t_8 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_8) {

    /* "pandas/_libs/intervaltree.pxi":3067
 *             # continue the binary tree structure. Instead, we use linear
 *             # search.
 *             for i in range(self.n_elements):             # <<<<<<<<<<<<<<
 *                 if self.left[i] <= point < self.right[i]:
 *                     result.append(self.indices[i])
 */
    __pyx_t_9 = __pyx_v_self->n_elements;
    __pyx_t_10 = __pyx_t_9;
    for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
      __pyx_v_i = __pyx_t_11;

      /* "pandas/_libs/intervaltree.pxi":3068
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point < self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      __pyx_t_12 = __pyx_v_i;
      __pyx_t_8 = ((*((__pyx_t_5numpy_uint64_t *) ( /* dim=0 */ (__pyx_v_self->left.data + __pyx_t_12 * __pyx_v_self->left.strides[0]) ))) <= __pyx_v_point);
      if (__pyx_t_8) {
        __pyx_t_13 = __pyx_v_i;
        __pyx_t_8 = (__pyx_v_point < (*((__pyx_t_5numpy_uint64_t *) ( /* dim=0 */ (__pyx_v_self->right.data + __pyx_t_13 * __pyx_v_self->right.strides[0]) ))));
      }
      __pyx_t_14 = (__pyx_t_8 != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":3069
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point < self.right[i]:
 *                     result.append(self.indices[i])             # <<<<<<<<<<<<<<
 *         else:
 *             # There are child nodes. Based on comparing our query to the pivot,
 */
        __pyx_t_15 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->indices.data + __pyx_t_15 * __pyx_v_self->indices.strides[0]) ))));

        /* "pandas/_libs/intervaltree.pxi":3068
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point < self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      }
    }

    /* "pandas/_libs/intervaltree.pxi":3063
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
    goto __pyx_L3;
  }

  /* "pandas/_libs/intervaltree.pxi":3073
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
  /*else*/ {
    __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":3074
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:
 *                 values = self.center_left_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 */
      __pyx_t_16 = __pyx_v_self->center_left_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":3075
 *             if point < self.pivot:
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:
 */
      __pyx_t_17 = __pyx_v_self->center_left_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":3076
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):             # <<<<<<<<<<<<<<
 *                     if not values[i] <= point:
 *                         break
 */
      __pyx_t_9 = __pyx_v_self->n_center;
      __pyx_t_10 = __pyx_t_9;
      for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":3077
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_18 = __pyx_v_i;
        __pyx_t_14 = ((!(((*((__pyx_t_5numpy_uint64_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_18 * __pyx_v_values.strides[0]) ))) <= __pyx_v_point) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":3078
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:
 */
          goto __pyx_L9_break;

          /* "pandas/_libs/intervaltree.pxi":3077
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":3079
 *                     if not values[i] <= point:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 */
        __pyx_t_19 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_19 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L9_break:;

      /* "pandas/_libs/intervaltree.pxi":3080
 *                         break
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->left_node->max_right) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":3081
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)             # <<<<<<<<<<<<<<
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode *)__pyx_v_self->left_node->__pyx_vtab)->__pyx_fuse_0query(__pyx_v_self->left_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3081, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":3080
 *                         break
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":3073
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":3082
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
    __pyx_t_14 = ((__pyx_v_point > __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":3083
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 *                 values = self.center_right_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 */
      __pyx_t_16 = __pyx_v_self->center_right_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":3084
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:
 */
      __pyx_t_17 = __pyx_v_self->center_right_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":3085
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):             # <<<<<<<<<<<<<<
 *                     if not point < values[i]:
 *                         break
 */
      for (__pyx_t_11 = (__pyx_v_self->n_center - 1); __pyx_t_11 > -1L; __pyx_t_11-=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":3086
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_20 = __pyx_v_i;
        __pyx_t_14 = ((!((__pyx_v_point < (*((__pyx_t_5numpy_uint64_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_20 * __pyx_v_values.strides[0]) )))) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":3087
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:
 */
          goto __pyx_L13_break;

          /* "pandas/_libs/intervaltree.pxi":3086
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":3088
 *                     if not point < values[i]:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if self.right_node.min_left <= point:
 *                     self.right_node.query(result, point)
 */
        __pyx_t_21 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_21 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L13_break:;

      /* "pandas/_libs/intervaltree.pxi":3089
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      __pyx_t_14 = ((__pyx_v_self->right_node->min_left <= __pyx_v_point) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":3090
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:
 *                     self.right_node.query(result, point)             # <<<<<<<<<<<<<<
 *             else:
 *                 result.extend(self.center_left_indices)
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode *)__pyx_v_self->right_node->__pyx_vtab)->__pyx_fuse_0query(__pyx_v_self->right_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3090, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":3089
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":3082
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":3092
 *                     self.right_node.query(result, point)
 *             else:
 *                 result.extend(self.center_left_indices)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
    /*else*/ {
      __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->extend(__pyx_v_result, __pyx_v_self->center_left_indices); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3092, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_L7:;
  }
  __pyx_L3:;

  /* "pandas/_libs/intervaltree.pxi":3054
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __PYX_XDEC_MEMVIEW(&__pyx_t_16, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_17, 1);
  __Pyx_AddTraceback("pandas._libs.interval.Uint64ClosedLeftIntervalNode.query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_indices, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_values, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_9__pyx_fuse_0query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_0__pyx_mdef_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_9__pyx_fuse_0query = {"__pyx_fuse_0query", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_9__pyx_fuse_0query, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_2query};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_9__pyx_fuse_0query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result = 0;
  __pyx_t_5numpy_float64_t __pyx_v_point;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fuse_0query (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_result,&__pyx_n_s_point,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_result)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_point)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_0query", 1, 2, 2, 1); __PYX_ERR(1, 3054, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_0query") < 0)) __PYX_ERR(1, 3054, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_result = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)values[0]);
    __pyx_v_point = __pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_point == ((npy_float64)-1)) && PyErr_Occurred())) __PYX_ERR(1, 3054, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_0query", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 3054, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Uint64ClosedLeftIntervalNode.__pyx_fuse_0query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_result), __pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector, 1, "result", 0))) __PYX_ERR(1, 3054, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_8__pyx_fuse_0query(((struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode *)__pyx_v_self), __pyx_v_result, __pyx_v_point);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_8__pyx_fuse_0query(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float64_t __pyx_v_point) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__pyx_fuse_0query", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_fuse_0__pyx_f_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_query(__pyx_v_self, __pyx_v_result, __pyx_v_point, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3054, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Uint64ClosedLeftIntervalNode.__pyx_fuse_0query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_11__pyx_fuse_1query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_fuse_1__pyx_f_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float32_t __pyx_v_point, int __pyx_skip_dispatch) {
  __Pyx_memviewslice __pyx_v_indices = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_values = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __pyx_t_5numpy_int64_t __pyx_t_9;
  __pyx_t_5numpy_int64_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  int __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  __Pyx_memviewslice __pyx_t_16 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_17 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  Py_ssize_t __pyx_t_21;
  __Pyx_RefNannySetupContext("__pyx_fuse_1query", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pyx_fuse_1query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3054, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_11__pyx_fuse_1query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = PyFloat_FromDouble(__pyx_v_point); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 3054, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3054, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3054, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 3054, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_5) {
            __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3054, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3054, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_3query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_4 = PyFloat_FromDouble(__pyx_v_point); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 3054, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_7 = __pyx_t_1; __pyx_t_3 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3054, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3054, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 3054, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (__pyx_t_3) {
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_t_4);
          __pyx_t_4 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3054, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pandas/_libs/intervaltree.pxi":3063
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
  __pyx_t_8 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_8) {

    /* "pandas/_libs/intervaltree.pxi":3067
 *             # continue the binary tree structure. Instead, we use linear
 *             # search.
 *             for i in range(self.n_elements):             # <<<<<<<<<<<<<<
 *                 if self.left[i] <= point < self.right[i]:
 *                     result.append(self.indices[i])
 */
    __pyx_t_9 = __pyx_v_self->n_elements;
    __pyx_t_10 = __pyx_t_9;
    for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
      __pyx_v_i = __pyx_t_11;

      /* "pandas/_libs/intervaltree.pxi":3068
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point < self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      __pyx_t_12 = __pyx_v_i;
      __pyx_t_8 = ((*((__pyx_t_5numpy_uint64_t *) ( /* dim=0 */ (__pyx_v_self->left.data + __pyx_t_12 * __pyx_v_self->left.strides[0]) ))) <= __pyx_v_point);
      if (__pyx_t_8) {
        __pyx_t_13 = __pyx_v_i;
        __pyx_t_8 = (__pyx_v_point < (*((__pyx_t_5numpy_uint64_t *) ( /* dim=0 */ (__pyx_v_self->right.data + __pyx_t_13 * __pyx_v_self->right.strides[0]) ))));
      }
      __pyx_t_14 = (__pyx_t_8 != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":3069
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point < self.right[i]:
 *                     result.append(self.indices[i])             # <<<<<<<<<<<<<<
 *         else:
 *             # There are child nodes. Based on comparing our query to the pivot,
 */
        __pyx_t_15 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->indices.data + __pyx_t_15 * __pyx_v_self->indices.strides[0]) ))));

        /* "pandas/_libs/intervaltree.pxi":3068
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point < self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      }
    }

    /* "pandas/_libs/intervaltree.pxi":3063
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
    goto __pyx_L3;
  }

  /* "pandas/_libs/intervaltree.pxi":3073
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
  /*else*/ {
    __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":3074
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:
 *                 values = self.center_left_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 */
      __pyx_t_16 = __pyx_v_self->center_left_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":3075
 *             if point < self.pivot:
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:
 */
      __pyx_t_17 = __pyx_v_self->center_left_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":3076
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):             # <<<<<<<<<<<<<<
 *                     if not values[i] <= point:
 *                         break
 */
      __pyx_t_9 = __pyx_v_self->n_center;
      __pyx_t_10 = __pyx_t_9;
      for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":3077
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_18 = __pyx_v_i;
        __pyx_t_14 = ((!(((*((__pyx_t_5numpy_uint64_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_18 * __pyx_v_values.strides[0]) ))) <= __pyx_v_point) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":3078
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:
 */
          goto __pyx_L9_break;

          /* "pandas/_libs/intervaltree.pxi":3077
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":3079
 *                     if not values[i] <= point:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 */
        __pyx_t_19 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_19 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L9_break:;

      /* "pandas/_libs/intervaltree.pxi":3080
 *                         break
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->left_node->max_right) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":3081
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)             # <<<<<<<<<<<<<<
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode *)__pyx_v_self->left_node->__pyx_vtab)->__pyx_fuse_1query(__pyx_v_self->left_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3081, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":3080
 *                         break
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":3073
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":3082
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
    __pyx_t_14 = ((__pyx_v_point > __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":3083
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 *                 values = self.center_right_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 */
      __pyx_t_16 = __pyx_v_self->center_right_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":3084
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:
 */
      __pyx_t_17 = __pyx_v_self->center_right_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":3085
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):             # <<<<<<<<<<<<<<
 *                     if not point < values[i]:
 *                         break
 */
      for (__pyx_t_11 = (__pyx_v_self->n_center - 1); __pyx_t_11 > -1L; __pyx_t_11-=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":3086
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_20 = __pyx_v_i;
        __pyx_t_14 = ((!((__pyx_v_point < (*((__pyx_t_5numpy_uint64_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_20 * __pyx_v_values.strides[0]) )))) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":3087
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:
 */
          goto __pyx_L13_break;

          /* "pandas/_libs/intervaltree.pxi":3086
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":3088
 *                     if not point < values[i]:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if self.right_node.min_left <= point:
 *                     self.right_node.query(result, point)
 */
        __pyx_t_21 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_21 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L13_break:;

      /* "pandas/_libs/intervaltree.pxi":3089
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      __pyx_t_14 = ((__pyx_v_self->right_node->min_left <= __pyx_v_point) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":3090
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:
 *                     self.right_node.query(result, point)             # <<<<<<<<<<<<<<
 *             else:
 *                 result.extend(self.center_left_indices)
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode *)__pyx_v_self->right_node->__pyx_vtab)->__pyx_fuse_1query(__pyx_v_self->right_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3090, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":3089
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":3082
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":3092
 *                     self.right_node.query(result, point)
 *             else:
 *                 result.extend(self.center_left_indices)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
    /*else*/ {
      __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->extend(__pyx_v_result, __pyx_v_self->center_left_indices); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3092, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_L7:;
  }
  __pyx_L3:;

  /* "pandas/_libs/intervaltree.pxi":3054
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __PYX_XDEC_MEMVIEW(&__pyx_t_16, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_17, 1);
  __Pyx_AddTraceback("pandas._libs.interval.Uint64ClosedLeftIntervalNode.query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_indices, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_values, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_11__pyx_fuse_1query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_1__pyx_mdef_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_11__pyx_fuse_1query = {"__pyx_fuse_1query", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_11__pyx_fuse_1query, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_2query};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_11__pyx_fuse_1query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result = 0;
  __pyx_t_5numpy_float32_t __pyx_v_point;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fuse_1query (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_result,&__pyx_n_s_point,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_result)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_point)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_1query", 1, 2, 2, 1); __PYX_ERR(1, 3054, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_1query") < 0)) __PYX_ERR(1, 3054, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_result = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)values[0]);
    __pyx_v_point = __pyx_PyFloat_AsFloat(values[1]); if (unlikely((__pyx_v_point == ((npy_float32)-1)) && PyErr_Occurred())) __PYX_ERR(1, 3054, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_1query", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 3054, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Uint64ClosedLeftIntervalNode.__pyx_fuse_1query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_result), __pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector, 1, "result", 0))) __PYX_ERR(1, 3054, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_10__pyx_fuse_1query(((struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode *)__pyx_v_self), __pyx_v_result, __pyx_v_point);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_10__pyx_fuse_1query(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float32_t __pyx_v_point) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__pyx_fuse_1query", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_fuse_1__pyx_f_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_query(__pyx_v_self, __pyx_v_result, __pyx_v_point, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3054, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Uint64ClosedLeftIntervalNode.__pyx_fuse_1query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_13__pyx_fuse_2query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_fuse_2__pyx_f_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int64_t __pyx_v_point, int __pyx_skip_dispatch) {
  __Pyx_memviewslice __pyx_v_indices = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_values = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __pyx_t_5numpy_int64_t __pyx_t_9;
  __pyx_t_5numpy_int64_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  int __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  __Pyx_memviewslice __pyx_t_16 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_17 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  Py_ssize_t __pyx_t_21;
  __Pyx_RefNannySetupContext("__pyx_fuse_2query", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pyx_fuse_2query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3054, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_13__pyx_fuse_2query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = __Pyx_PyInt_From_npy_int64(__pyx_v_point); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 3054, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3054, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3054, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 3054, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_5) {
            __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3054, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3054, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_3query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_4 = __Pyx_PyInt_From_npy_int64(__pyx_v_point); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 3054, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_7 = __pyx_t_1; __pyx_t_3 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3054, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3054, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 3054, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (__pyx_t_3) {
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_t_4);
          __pyx_t_4 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3054, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pandas/_libs/intervaltree.pxi":3063
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
  __pyx_t_8 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_8) {

    /* "pandas/_libs/intervaltree.pxi":3067
 *             # continue the binary tree structure. Instead, we use linear
 *             # search.
 *             for i in range(self.n_elements):             # <<<<<<<<<<<<<<
 *                 if self.left[i] <= point < self.right[i]:
 *                     result.append(self.indices[i])
 */
    __pyx_t_9 = __pyx_v_self->n_elements;
    __pyx_t_10 = __pyx_t_9;
    for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
      __pyx_v_i = __pyx_t_11;

      /* "pandas/_libs/intervaltree.pxi":3068
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point < self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      __pyx_t_12 = __pyx_v_i;
      __pyx_t_8 = ((*((__pyx_t_5numpy_uint64_t *) ( /* dim=0 */ (__pyx_v_self->left.data + __pyx_t_12 * __pyx_v_self->left.strides[0]) ))) <= __pyx_v_point);
      if (__pyx_t_8) {
        __pyx_t_13 = __pyx_v_i;
        __pyx_t_8 = (__pyx_v_point < (*((__pyx_t_5numpy_uint64_t *) ( /* dim=0 */ (__pyx_v_self->right.data + __pyx_t_13 * __pyx_v_self->right.strides[0]) ))));
      }
      __pyx_t_14 = (__pyx_t_8 != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":3069
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point < self.right[i]:
 *                     result.append(self.indices[i])             # <<<<<<<<<<<<<<
 *         else:
 *             # There are child nodes. Based on comparing our query to the pivot,
 */
        __pyx_t_15 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->indices.data + __pyx_t_15 * __pyx_v_self->indices.strides[0]) ))));

        /* "pandas/_libs/intervaltree.pxi":3068
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point < self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      }
    }

    /* "pandas/_libs/intervaltree.pxi":3063
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
    goto __pyx_L3;
  }

  /* "pandas/_libs/intervaltree.pxi":3073
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
  /*else*/ {
    __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":3074
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:
 *                 values = self.center_left_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 */
      __pyx_t_16 = __pyx_v_self->center_left_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":3075
 *             if point < self.pivot:
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:
 */
      __pyx_t_17 = __pyx_v_self->center_left_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":3076
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):             # <<<<<<<<<<<<<<
 *                     if not values[i] <= point:
 *                         break
 */
      __pyx_t_9 = __pyx_v_self->n_center;
      __pyx_t_10 = __pyx_t_9;
      for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":3077
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_18 = __pyx_v_i;
        __pyx_t_14 = ((!(((*((__pyx_t_5numpy_uint64_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_18 * __pyx_v_values.strides[0]) ))) <= __pyx_v_point) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":3078
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:
 */
          goto __pyx_L9_break;

          /* "pandas/_libs/intervaltree.pxi":3077
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":3079
 *                     if not values[i] <= point:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 */
        __pyx_t_19 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_19 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L9_break:;

      /* "pandas/_libs/intervaltree.pxi":3080
 *                         break
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->left_node->max_right) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":3081
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)             # <<<<<<<<<<<<<<
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode *)__pyx_v_self->left_node->__pyx_vtab)->__pyx_fuse_2query(__pyx_v_self->left_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3081, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":3080
 *                         break
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":3073
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":3082
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
    __pyx_t_14 = ((__pyx_v_point > __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":3083
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 *                 values = self.center_right_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 */
      __pyx_t_16 = __pyx_v_self->center_right_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":3084
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:
 */
      __pyx_t_17 = __pyx_v_self->center_right_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":3085
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):             # <<<<<<<<<<<<<<
 *                     if not point < values[i]:
 *                         break
 */
      for (__pyx_t_11 = (__pyx_v_self->n_center - 1); __pyx_t_11 > -1L; __pyx_t_11-=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":3086
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_20 = __pyx_v_i;
        __pyx_t_14 = ((!((__pyx_v_point < (*((__pyx_t_5numpy_uint64_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_20 * __pyx_v_values.strides[0]) )))) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":3087
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:
 */
          goto __pyx_L13_break;

          /* "pandas/_libs/intervaltree.pxi":3086
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":3088
 *                     if not point < values[i]:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if self.right_node.min_left <= point:
 *                     self.right_node.query(result, point)
 */
        __pyx_t_21 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_21 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L13_break:;

      /* "pandas/_libs/intervaltree.pxi":3089
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      __pyx_t_14 = ((__pyx_v_self->right_node->min_left <= __pyx_v_point) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":3090
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:
 *                     self.right_node.query(result, point)             # <<<<<<<<<<<<<<
 *             else:
 *                 result.extend(self.center_left_indices)
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode *)__pyx_v_self->right_node->__pyx_vtab)->__pyx_fuse_2query(__pyx_v_self->right_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3090, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":3089
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":3082
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":3092
 *                     self.right_node.query(result, point)
 *             else:
 *                 result.extend(self.center_left_indices)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
    /*else*/ {
      __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->extend(__pyx_v_result, __pyx_v_self->center_left_indices); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3092, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_L7:;
  }
  __pyx_L3:;

  /* "pandas/_libs/intervaltree.pxi":3054
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __PYX_XDEC_MEMVIEW(&__pyx_t_16, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_17, 1);
  __Pyx_AddTraceback("pandas._libs.interval.Uint64ClosedLeftIntervalNode.query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_indices, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_values, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_13__pyx_fuse_2query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_2__pyx_mdef_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_13__pyx_fuse_2query = {"__pyx_fuse_2query", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_13__pyx_fuse_2query, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_2query};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_13__pyx_fuse_2query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result = 0;
  __pyx_t_5numpy_int64_t __pyx_v_point;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fuse_2query (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_result,&__pyx_n_s_point,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_result)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_point)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_2query", 1, 2, 2, 1); __PYX_ERR(1, 3054, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_2query") < 0)) __PYX_ERR(1, 3054, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_result = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)values[0]);
    __pyx_v_point = __Pyx_PyInt_As_npy_int64(values[1]); if (unlikely((__pyx_v_point == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(1, 3054, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_2query", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 3054, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Uint64ClosedLeftIntervalNode.__pyx_fuse_2query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_result), __pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector, 1, "result", 0))) __PYX_ERR(1, 3054, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_12__pyx_fuse_2query(((struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode *)__pyx_v_self), __pyx_v_result, __pyx_v_point);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_12__pyx_fuse_2query(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int64_t __pyx_v_point) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__pyx_fuse_2query", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_fuse_2__pyx_f_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_query(__pyx_v_self, __pyx_v_result, __pyx_v_point, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3054, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Uint64ClosedLeftIntervalNode.__pyx_fuse_2query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_15__pyx_fuse_3query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_fuse_3__pyx_f_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int32_t __pyx_v_point, int __pyx_skip_dispatch) {
  __Pyx_memviewslice __pyx_v_indices = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_values = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __pyx_t_5numpy_int64_t __pyx_t_9;
  __pyx_t_5numpy_int64_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  int __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  __Pyx_memviewslice __pyx_t_16 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_17 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  Py_ssize_t __pyx_t_21;
  __Pyx_RefNannySetupContext("__pyx_fuse_3query", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pyx_fuse_3query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3054, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_15__pyx_fuse_3query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = __Pyx_PyInt_From_npy_int32(__pyx_v_point); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 3054, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3054, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3054, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 3054, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_5) {
            __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3054, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3054, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_3query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_4 = __Pyx_PyInt_From_npy_int32(__pyx_v_point); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 3054, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_7 = __pyx_t_1; __pyx_t_3 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3054, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3054, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 3054, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (__pyx_t_3) {
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_t_4);
          __pyx_t_4 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3054, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pandas/_libs/intervaltree.pxi":3063
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
  __pyx_t_8 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_8) {

    /* "pandas/_libs/intervaltree.pxi":3067
 *             # continue the binary tree structure. Instead, we use linear
 *             # search.
 *             for i in range(self.n_elements):             # <<<<<<<<<<<<<<
 *                 if self.left[i] <= point < self.right[i]:
 *                     result.append(self.indices[i])
 */
    __pyx_t_9 = __pyx_v_self->n_elements;
    __pyx_t_10 = __pyx_t_9;
    for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
      __pyx_v_i = __pyx_t_11;

      /* "pandas/_libs/intervaltree.pxi":3068
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point < self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      __pyx_t_12 = __pyx_v_i;
      __pyx_t_8 = ((*((__pyx_t_5numpy_uint64_t *) ( /* dim=0 */ (__pyx_v_self->left.data + __pyx_t_12 * __pyx_v_self->left.strides[0]) ))) <= __pyx_v_point);
      if (__pyx_t_8) {
        __pyx_t_13 = __pyx_v_i;
        __pyx_t_8 = (__pyx_v_point < (*((__pyx_t_5numpy_uint64_t *) ( /* dim=0 */ (__pyx_v_self->right.data + __pyx_t_13 * __pyx_v_self->right.strides[0]) ))));
      }
      __pyx_t_14 = (__pyx_t_8 != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":3069
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point < self.right[i]:
 *                     result.append(self.indices[i])             # <<<<<<<<<<<<<<
 *         else:
 *             # There are child nodes. Based on comparing our query to the pivot,
 */
        __pyx_t_15 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->indices.data + __pyx_t_15 * __pyx_v_self->indices.strides[0]) ))));

        /* "pandas/_libs/intervaltree.pxi":3068
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point < self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      }
    }

    /* "pandas/_libs/intervaltree.pxi":3063
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
    goto __pyx_L3;
  }

  /* "pandas/_libs/intervaltree.pxi":3073
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
  /*else*/ {
    __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":3074
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:
 *                 values = self.center_left_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 */
      __pyx_t_16 = __pyx_v_self->center_left_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":3075
 *             if point < self.pivot:
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:
 */
      __pyx_t_17 = __pyx_v_self->center_left_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":3076
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):             # <<<<<<<<<<<<<<
 *                     if not values[i] <= point:
 *                         break
 */
      __pyx_t_9 = __pyx_v_self->n_center;
      __pyx_t_10 = __pyx_t_9;
      for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":3077
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_18 = __pyx_v_i;
        __pyx_t_14 = ((!(((*((__pyx_t_5numpy_uint64_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_18 * __pyx_v_values.strides[0]) ))) <= __pyx_v_point) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":3078
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:
 */
          goto __pyx_L9_break;

          /* "pandas/_libs/intervaltree.pxi":3077
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":3079
 *                     if not values[i] <= point:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 */
        __pyx_t_19 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_19 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L9_break:;

      /* "pandas/_libs/intervaltree.pxi":3080
 *                         break
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->left_node->max_right) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":3081
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)             # <<<<<<<<<<<<<<
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode *)__pyx_v_self->left_node->__pyx_vtab)->__pyx_fuse_3query(__pyx_v_self->left_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3081, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":3080
 *                         break
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":3073
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":3082
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
    __pyx_t_14 = ((__pyx_v_point > __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":3083
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 *                 values = self.center_right_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 */
      __pyx_t_16 = __pyx_v_self->center_right_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":3084
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:
 */
      __pyx_t_17 = __pyx_v_self->center_right_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":3085
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):             # <<<<<<<<<<<<<<
 *                     if not point < values[i]:
 *                         break
 */
      for (__pyx_t_11 = (__pyx_v_self->n_center - 1); __pyx_t_11 > -1L; __pyx_t_11-=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":3086
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_20 = __pyx_v_i;
        __pyx_t_14 = ((!((__pyx_v_point < (*((__pyx_t_5numpy_uint64_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_20 * __pyx_v_values.strides[0]) )))) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":3087
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:
 */
          goto __pyx_L13_break;

          /* "pandas/_libs/intervaltree.pxi":3086
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":3088
 *                     if not point < values[i]:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if self.right_node.min_left <= point:
 *                     self.right_node.query(result, point)
 */
        __pyx_t_21 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_21 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L13_break:;

      /* "pandas/_libs/intervaltree.pxi":3089
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      __pyx_t_14 = ((__pyx_v_self->right_node->min_left <= __pyx_v_point) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":3090
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:
 *                     self.right_node.query(result, point)             # <<<<<<<<<<<<<<
 *             else:
 *                 result.extend(self.center_left_indices)
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode *)__pyx_v_self->right_node->__pyx_vtab)->__pyx_fuse_3query(__pyx_v_self->right_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3090, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":3089
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":3082
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":3092
 *                     self.right_node.query(result, point)
 *             else:
 *                 result.extend(self.center_left_indices)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
    /*else*/ {
      __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->extend(__pyx_v_result, __pyx_v_self->center_left_indices); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3092, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_L7:;
  }
  __pyx_L3:;

  /* "pandas/_libs/intervaltree.pxi":3054
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __PYX_XDEC_MEMVIEW(&__pyx_t_16, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_17, 1);
  __Pyx_AddTraceback("pandas._libs.interval.Uint64ClosedLeftIntervalNode.query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_indices, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_values, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_15__pyx_fuse_3query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_3__pyx_mdef_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_15__pyx_fuse_3query = {"__pyx_fuse_3query", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_15__pyx_fuse_3query, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_2query};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_15__pyx_fuse_3query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result = 0;
  __pyx_t_5numpy_int32_t __pyx_v_point;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fuse_3query (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_result,&__pyx_n_s_point,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_result)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_point)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_3query", 1, 2, 2, 1); __PYX_ERR(1, 3054, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_3query") < 0)) __PYX_ERR(1, 3054, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_result = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)values[0]);
    __pyx_v_point = __Pyx_PyInt_As_npy_int32(values[1]); if (unlikely((__pyx_v_point == ((npy_int32)-1)) && PyErr_Occurred())) __PYX_ERR(1, 3054, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_3query", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 3054, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Uint64ClosedLeftIntervalNode.__pyx_fuse_3query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_result), __pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector, 1, "result", 0))) __PYX_ERR(1, 3054, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_14__pyx_fuse_3query(((struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode *)__pyx_v_self), __pyx_v_result, __pyx_v_point);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_14__pyx_fuse_3query(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int32_t __pyx_v_point) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__pyx_fuse_3query", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_fuse_3__pyx_f_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_query(__pyx_v_self, __pyx_v_result, __pyx_v_point, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3054, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Uint64ClosedLeftIntervalNode.__pyx_fuse_3query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_17__pyx_fuse_4query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_fuse_4__pyx_f_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_uint64_t __pyx_v_point, int __pyx_skip_dispatch) {
  __Pyx_memviewslice __pyx_v_indices = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_values = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __pyx_t_5numpy_int64_t __pyx_t_9;
  __pyx_t_5numpy_int64_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  int __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  __Pyx_memviewslice __pyx_t_16 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_17 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  Py_ssize_t __pyx_t_21;
  __Pyx_RefNannySetupContext("__pyx_fuse_4query", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pyx_fuse_4query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3054, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_17__pyx_fuse_4query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = __Pyx_PyInt_From_npy_uint64(__pyx_v_point); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 3054, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3054, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3054, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 3054, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_5) {
            __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3054, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3054, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_3query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_4 = __Pyx_PyInt_From_npy_uint64(__pyx_v_point); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 3054, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_7 = __pyx_t_1; __pyx_t_3 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3054, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3054, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 3054, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (__pyx_t_3) {
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_t_4);
          __pyx_t_4 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3054, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pandas/_libs/intervaltree.pxi":3063
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
  __pyx_t_8 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_8) {

    /* "pandas/_libs/intervaltree.pxi":3067
 *             # continue the binary tree structure. Instead, we use linear
 *             # search.
 *             for i in range(self.n_elements):             # <<<<<<<<<<<<<<
 *                 if self.left[i] <= point < self.right[i]:
 *                     result.append(self.indices[i])
 */
    __pyx_t_9 = __pyx_v_self->n_elements;
    __pyx_t_10 = __pyx_t_9;
    for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
      __pyx_v_i = __pyx_t_11;

      /* "pandas/_libs/intervaltree.pxi":3068
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point < self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      __pyx_t_12 = __pyx_v_i;
      __pyx_t_8 = ((*((__pyx_t_5numpy_uint64_t *) ( /* dim=0 */ (__pyx_v_self->left.data + __pyx_t_12 * __pyx_v_self->left.strides[0]) ))) <= __pyx_v_point);
      if (__pyx_t_8) {
        __pyx_t_13 = __pyx_v_i;
        __pyx_t_8 = (__pyx_v_point < (*((__pyx_t_5numpy_uint64_t *) ( /* dim=0 */ (__pyx_v_self->right.data + __pyx_t_13 * __pyx_v_self->right.strides[0]) ))));
      }
      __pyx_t_14 = (__pyx_t_8 != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":3069
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point < self.right[i]:
 *                     result.append(self.indices[i])             # <<<<<<<<<<<<<<
 *         else:
 *             # There are child nodes. Based on comparing our query to the pivot,
 */
        __pyx_t_15 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->indices.data + __pyx_t_15 * __pyx_v_self->indices.strides[0]) ))));

        /* "pandas/_libs/intervaltree.pxi":3068
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point < self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      }
    }

    /* "pandas/_libs/intervaltree.pxi":3063
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
    goto __pyx_L3;
  }

  /* "pandas/_libs/intervaltree.pxi":3073
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
  /*else*/ {
    __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":3074
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:
 *                 values = self.center_left_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 */
      __pyx_t_16 = __pyx_v_self->center_left_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":3075
 *             if point < self.pivot:
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:
 */
      __pyx_t_17 = __pyx_v_self->center_left_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":3076
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):             # <<<<<<<<<<<<<<
 *                     if not values[i] <= point:
 *                         break
 */
      __pyx_t_9 = __pyx_v_self->n_center;
      __pyx_t_10 = __pyx_t_9;
      for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":3077
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_18 = __pyx_v_i;
        __pyx_t_14 = ((!(((*((__pyx_t_5numpy_uint64_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_18 * __pyx_v_values.strides[0]) ))) <= __pyx_v_point) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":3078
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:
 */
          goto __pyx_L9_break;

          /* "pandas/_libs/intervaltree.pxi":3077
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":3079
 *                     if not values[i] <= point:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 */
        __pyx_t_19 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_19 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L9_break:;

      /* "pandas/_libs/intervaltree.pxi":3080
 *                         break
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->left_node->max_right) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":3081
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)             # <<<<<<<<<<<<<<
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode *)__pyx_v_self->left_node->__pyx_vtab)->__pyx_fuse_4query(__pyx_v_self->left_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3081, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":3080
 *                         break
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":3073
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":3082
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
    __pyx_t_14 = ((__pyx_v_point > __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":3083
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 *                 values = self.center_right_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 */
      __pyx_t_16 = __pyx_v_self->center_right_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":3084
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:
 */
      __pyx_t_17 = __pyx_v_self->center_right_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":3085
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):             # <<<<<<<<<<<<<<
 *                     if not point < values[i]:
 *                         break
 */
      for (__pyx_t_11 = (__pyx_v_self->n_center - 1); __pyx_t_11 > -1L; __pyx_t_11-=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":3086
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_20 = __pyx_v_i;
        __pyx_t_14 = ((!((__pyx_v_point < (*((__pyx_t_5numpy_uint64_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_20 * __pyx_v_values.strides[0]) )))) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":3087
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:
 */
          goto __pyx_L13_break;

          /* "pandas/_libs/intervaltree.pxi":3086
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":3088
 *                     if not point < values[i]:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if self.right_node.min_left <= point:
 *                     self.right_node.query(result, point)
 */
        __pyx_t_21 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_21 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L13_break:;

      /* "pandas/_libs/intervaltree.pxi":3089
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      __pyx_t_14 = ((__pyx_v_self->right_node->min_left <= __pyx_v_point) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":3090
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:
 *                     self.right_node.query(result, point)             # <<<<<<<<<<<<<<
 *             else:
 *                 result.extend(self.center_left_indices)
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode *)__pyx_v_self->right_node->__pyx_vtab)->__pyx_fuse_4query(__pyx_v_self->right_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3090, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":3089
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":3082
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":3092
 *                     self.right_node.query(result, point)
 *             else:
 *                 result.extend(self.center_left_indices)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
    /*else*/ {
      __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->extend(__pyx_v_result, __pyx_v_self->center_left_indices); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3092, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_L7:;
  }
  __pyx_L3:;

  /* "pandas/_libs/intervaltree.pxi":3054
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __PYX_XDEC_MEMVIEW(&__pyx_t_16, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_17, 1);
  __Pyx_AddTraceback("pandas._libs.interval.Uint64ClosedLeftIntervalNode.query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_indices, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_values, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_17__pyx_fuse_4query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_4__pyx_mdef_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_17__pyx_fuse_4query = {"__pyx_fuse_4query", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_17__pyx_fuse_4query, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_2query};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_17__pyx_fuse_4query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result = 0;
  __pyx_t_5numpy_uint64_t __pyx_v_point;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fuse_4query (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_result,&__pyx_n_s_point,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_result)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_point)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_4query", 1, 2, 2, 1); __PYX_ERR(1, 3054, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_4query") < 0)) __PYX_ERR(1, 3054, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_result = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)values[0]);
    __pyx_v_point = __Pyx_PyInt_As_npy_uint64(values[1]); if (unlikely((__pyx_v_point == ((npy_uint64)-1)) && PyErr_Occurred())) __PYX_ERR(1, 3054, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_4query", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 3054, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Uint64ClosedLeftIntervalNode.__pyx_fuse_4query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_result), __pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector, 1, "result", 0))) __PYX_ERR(1, 3054, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_16__pyx_fuse_4query(((struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode *)__pyx_v_self), __pyx_v_result, __pyx_v_point);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_16__pyx_fuse_4query(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_uint64_t __pyx_v_point) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__pyx_fuse_4query", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_fuse_4__pyx_f_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_query(__pyx_v_self, __pyx_v_result, __pyx_v_point, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3054, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Uint64ClosedLeftIntervalNode.__pyx_fuse_4query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":3094
 *                 result.extend(self.center_left_indices)
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         if self.is_leaf_node:
 *             return ('<Uint64ClosedLeftIntervalNode: '
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_5__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_5__repr__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_4__repr__(((struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_4__repr__(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode *__pyx_v_self) {
  PyObject *__pyx_v_n_left = NULL;
  PyObject *__pyx_v_n_right = NULL;
  PyObject *__pyx_v_n_center = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  Py_UCS4 __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "pandas/_libs/intervaltree.pxi":3095
 * 
 *     def __repr__(self):
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             return ('<Uint64ClosedLeftIntervalNode: '
 *                     '%s elements (terminal)>' % self.n_elements)
 */
  __pyx_t_1 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_1) {

    /* "pandas/_libs/intervaltree.pxi":3096
 *     def __repr__(self):
 *         if self.is_leaf_node:
 *             return ('<Uint64ClosedLeftIntervalNode: '             # <<<<<<<<<<<<<<
 *                     '%s elements (terminal)>' % self.n_elements)
 *         else:
 */
    __Pyx_XDECREF(__pyx_r);

    /* "pandas/_libs/intervaltree.pxi":3097
 *         if self.is_leaf_node:
 *             return ('<Uint64ClosedLeftIntervalNode: '
 *                     '%s elements (terminal)>' % self.n_elements)             # <<<<<<<<<<<<<<
 *         else:
 *             n_left = self.left_node.n_elements
 */
    __pyx_t_2 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_elements); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3097, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PyUnicode_Format(__pyx_kp_u_Uint64ClosedLeftIntervalNode_s, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 3097, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pandas/_libs/intervaltree.pxi":3095
 * 
 *     def __repr__(self):
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             return ('<Uint64ClosedLeftIntervalNode: '
 *                     '%s elements (terminal)>' % self.n_elements)
 */
  }

  /* "pandas/_libs/intervaltree.pxi":3099
 *                     '%s elements (terminal)>' % self.n_elements)
 *         else:
 *             n_left = self.left_node.n_elements             # <<<<<<<<<<<<<<
 *             n_right = self.right_node.n_elements
 *             n_center = self.n_elements - n_left - n_right
 */
  /*else*/ {
    __pyx_t_3 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->left_node->n_elements); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 3099, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_v_n_left = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "pandas/_libs/intervaltree.pxi":3100
 *         else:
 *             n_left = self.left_node.n_elements
 *             n_right = self.right_node.n_elements             # <<<<<<<<<<<<<<
 *             n_center = self.n_elements - n_left - n_right
 *             return ('<Uint64ClosedLeftIntervalNode: '
 */
    __pyx_t_3 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->right_node->n_elements); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 3100, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_v_n_right = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "pandas/_libs/intervaltree.pxi":3101
 *             n_left = self.left_node.n_elements
 *             n_right = self.right_node.n_elements
 *             n_center = self.n_elements - n_left - n_right             # <<<<<<<<<<<<<<
 *             return ('<Uint64ClosedLeftIntervalNode: '
 *                     'pivot %s, %s elements (%s left, %s right, %s '
 */
    __pyx_t_3 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_elements); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 3101, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = PyNumber_Subtract(__pyx_t_3, __pyx_v_n_left); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3101, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyNumber_Subtract(__pyx_t_2, __pyx_v_n_right); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 3101, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_n_center = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "pandas/_libs/intervaltree.pxi":3102
 *             n_right = self.right_node.n_elements
 *             n_center = self.n_elements - n_left - n_right
 *             return ('<Uint64ClosedLeftIntervalNode: '             # <<<<<<<<<<<<<<
 *                     'pivot %s, %s elements (%s left, %s right, %s '
 *                     'overlapping)>' % (self.pivot, self.n_elements,
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = PyTuple_New(11); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 3102, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = 0;
    __pyx_t_5 = 127;
    __Pyx_INCREF(__pyx_kp_u_Uint64ClosedLeftIntervalNode_pi);
    __pyx_t_4 += 37;
    __Pyx_GIVEREF(__pyx_kp_u_Uint64ClosedLeftIntervalNode_pi);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_kp_u_Uint64ClosedLeftIntervalNode_pi);

    /* "pandas/_libs/intervaltree.pxi":3104
 *             return ('<Uint64ClosedLeftIntervalNode: '
 *                     'pivot %s, %s elements (%s left, %s right, %s '
 *                     'overlapping)>' % (self.pivot, self.n_elements,             # <<<<<<<<<<<<<<
 *                                        n_left, n_right, n_center))
 * 
 */
    __pyx_t_2 = __Pyx_PyInt_From_npy_uint64(__pyx_v_self->pivot); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3104, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_t_2), __pyx_empty_unicode); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 3104, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_5 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_6) > __pyx_t_5) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_6) : __pyx_t_5;
    __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_6);
    __pyx_t_6 = 0;
    __Pyx_INCREF(__pyx_kp_u__18);
    __pyx_t_4 += 2;
    __Pyx_GIVEREF(__pyx_kp_u__18);
    PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_kp_u__18);
    __pyx_t_6 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_elements); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 3104, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_2 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_t_6), __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3104, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_5 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) > __pyx_t_5) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) : __pyx_t_5;
    __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_3, 3, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_elements);
    __pyx_t_4 += 11;
    __Pyx_GIVEREF(__pyx_kp_u_elements);
    PyTuple_SET_ITEM(__pyx_t_3, 4, __pyx_kp_u_elements);

    /* "pandas/_libs/intervaltree.pxi":3105
 *                     'pivot %s, %s elements (%s left, %s right, %s '
 *                     'overlapping)>' % (self.pivot, self.n_elements,
 *                                        n_left, n_right, n_center))             # <<<<<<<<<<<<<<
 * 
 *     def counts(self):
 */
    __pyx_t_2 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_v_n_left), __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3105, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) > __pyx_t_5) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) : __pyx_t_5;
    __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_3, 5, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_left_2);
    __pyx_t_4 += 7;
    __Pyx_GIVEREF(__pyx_kp_u_left_2);
    PyTuple_SET_ITEM(__pyx_t_3, 6, __pyx_kp_u_left_2);
    __pyx_t_2 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_v_n_right), __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3105, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) > __pyx_t_5) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) : __pyx_t_5;
    __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_3, 7, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_right_2);
    __pyx_t_4 += 8;
    __Pyx_GIVEREF(__pyx_kp_u_right_2);
    PyTuple_SET_ITEM(__pyx_t_3, 8, __pyx_kp_u_right_2);
    __pyx_t_2 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_v_n_center), __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3105, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) > __pyx_t_5) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) : __pyx_t_5;
    __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_3, 9, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_overlapping);
    __pyx_t_4 += 14;
    __Pyx_GIVEREF(__pyx_kp_u_overlapping);
    PyTuple_SET_ITEM(__pyx_t_3, 10, __pyx_kp_u_overlapping);

    /* "pandas/_libs/intervaltree.pxi":3102
 *             n_right = self.right_node.n_elements
 *             n_center = self.n_elements - n_left - n_right
 *             return ('<Uint64ClosedLeftIntervalNode: '             # <<<<<<<<<<<<<<
 *                     'pivot %s, %s elements (%s left, %s right, %s '
 *                     'overlapping)>' % (self.pivot, self.n_elements,
 */
    __pyx_t_2 = __Pyx_PyUnicode_Join(__pyx_t_3, 11, __pyx_t_4, __pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3102, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;
  }

  /* "pandas/_libs/intervaltree.pxi":3094
 *                 result.extend(self.center_left_indices)
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         if self.is_leaf_node:
 *             return ('<Uint64ClosedLeftIntervalNode: '
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pandas._libs.interval.Uint64ClosedLeftIntervalNode.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_n_left);
  __Pyx_XDECREF(__pyx_v_n_right);
  __Pyx_XDECREF(__pyx_v_n_center);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":3107
 *                                        n_left, n_right, n_center))
 * 
 *     def counts(self):             # <<<<<<<<<<<<<<
 *         """
 *         Inspect counts on this node
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_7counts(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_6counts[] = "\n        Inspect counts on this node\n        useful for debugging purposes\n        ";
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_7counts(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("counts (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_6counts(((struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_6counts(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode *__pyx_v_self) {
  Py_ssize_t __pyx_v_m;
  PyObject *__pyx_v_l = NULL;
  PyObject *__pyx_v_r = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  size_t __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("counts", 0);

  /* "pandas/_libs/intervaltree.pxi":3112
 *         useful for debugging purposes
 *         """
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             return self.n_elements
 *         else:
 */
  __pyx_t_1 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_1) {

    /* "pandas/_libs/intervaltree.pxi":3113
 *         """
 *         if self.is_leaf_node:
 *             return self.n_elements             # <<<<<<<<<<<<<<
 *         else:
 *             m = len(self.center_left_values)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_elements); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3113, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "pandas/_libs/intervaltree.pxi":3112
 *         useful for debugging purposes
 *         """
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             return self.n_elements
 *         else:
 */
  }

  /* "pandas/_libs/intervaltree.pxi":3115
 *             return self.n_elements
 *         else:
 *             m = len(self.center_left_values)             # <<<<<<<<<<<<<<
 *             l = self.left_node.counts()
 *             r = self.right_node.counts()
 */
  /*else*/ {
    if (unlikely(!__pyx_v_self->center_left_values.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(1, 3115, __pyx_L1_error)}
    __pyx_t_3 = __Pyx_MemoryView_Len(__pyx_v_self->center_left_values); 
    __pyx_v_m = __pyx_t_3;

    /* "pandas/_libs/intervaltree.pxi":3116
 *         else:
 *             m = len(self.center_left_values)
 *             l = self.left_node.counts()             # <<<<<<<<<<<<<<
 *             r = self.right_node.counts()
 *             return (m, (l, r))
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->left_node), __pyx_n_s_counts); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 3116, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3116, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_l = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "pandas/_libs/intervaltree.pxi":3117
 *             m = len(self.center_left_values)
 *             l = self.left_node.counts()
 *             r = self.right_node.counts()             # <<<<<<<<<<<<<<
 *             return (m, (l, r))
 * 
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->right_node), __pyx_n_s_counts); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 3117, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3117, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_r = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "pandas/_libs/intervaltree.pxi":3118
 *             l = self.left_node.counts()
 *             r = self.right_node.counts()
 *             return (m, (l, r))             # <<<<<<<<<<<<<<
 * 
 * NODE_CLASSES['uint64',
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = PyInt_FromSsize_t(__pyx_v_m); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3118, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 3118, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_v_l);
    __Pyx_GIVEREF(__pyx_v_l);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_l);
    __Pyx_INCREF(__pyx_v_r);
    __Pyx_GIVEREF(__pyx_v_r);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_v_r);
    __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 3118, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_4);
    __pyx_t_2 = 0;
    __pyx_t_4 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;
  }

  /* "pandas/_libs/intervaltree.pxi":3107
 *                                        n_left, n_right, n_center))
 * 
 *     def counts(self):             # <<<<<<<<<<<<<<
 *         """
 *         Inspect counts on this node
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pandas._libs.interval.Uint64ClosedLeftIntervalNode.counts", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_l);
  __Pyx_XDECREF(__pyx_v_r);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":2964
 *         int64_t[:] center_left_indices, center_right_indices, indices
 *         uint64_t min_left, max_right
 *         readonly uint64_t pivot             # <<<<<<<<<<<<<<
 *         readonly int64_t n_elements, n_center, leaf_size
 *         readonly bint is_leaf_node
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_5pivot_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_5pivot_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_5pivot___get__(((struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_5pivot___get__(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_npy_uint64(__pyx_v_self->pivot); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2964, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Uint64ClosedLeftIntervalNode.pivot.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":2965
 *         uint64_t min_left, max_right
 *         readonly uint64_t pivot
 *         readonly int64_t n_elements, n_center, leaf_size             # <<<<<<<<<<<<<<
 *         readonly bint is_leaf_node
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_10n_elements_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_10n_elements_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_10n_elements___get__(((struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_10n_elements___get__(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_elements); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2965, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Uint64ClosedLeftIntervalNode.n_elements.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_8n_center_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_8n_center_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_8n_center___get__(((struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_8n_center___get__(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_center); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2965, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Uint64ClosedLeftIntervalNode.n_center.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_9leaf_size_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_9leaf_size_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_9leaf_size___get__(((struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_9leaf_size___get__(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->leaf_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2965, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Uint64ClosedLeftIntervalNode.leaf_size.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":2966
 *         readonly uint64_t pivot
 *         readonly int64_t n_elements, n_center, leaf_size
 *         readonly bint is_leaf_node             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self,
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_12is_leaf_node_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_12is_leaf_node_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_12is_leaf_node___get__(((struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_12is_leaf_node___get__(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->is_leaf_node); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2966, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Uint64ClosedLeftIntervalNode.is_leaf_node.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_21__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_21__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_20__reduce_cython__(((struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_20__reduce_cython__(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  int __pyx_t_16;
  int __pyx_t_17;
  int __pyx_t_18;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.center_left_indices, self.center_left_values, self.center_right_indices, self.center_right_values, self.indices, self.is_leaf_node, self.leaf_size, self.left, self.left_node, self.max_right, self.min_left, self.n_center, self.n_elements, self.pivot, self.right, self.right_node)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  if (unlikely(!__pyx_v_self->center_left_indices.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_1 = __pyx_memoryview_fromslice(__pyx_v_self->center_left_indices, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_int64_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_int64_t, 0);; if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (unlikely(!__pyx_v_self->center_left_values.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_2 = __pyx_memoryview_fromslice(__pyx_v_self->center_left_values, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_uint64_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_uint64_t, 0);; if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (unlikely(!__pyx_v_self->center_right_indices.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_3 = __pyx_memoryview_fromslice(__pyx_v_self->center_right_indices, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_int64_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_int64_t, 0);; if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely(!__pyx_v_self->center_right_values.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_4 = __pyx_memoryview_fromslice(__pyx_v_self->center_right_values, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_uint64_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_uint64_t, 0);; if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (unlikely(!__pyx_v_self->indices.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_5 = __pyx_memoryview_fromslice(__pyx_v_self->indices, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_int64_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_int64_t, 0);; if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyBool_FromLong(__pyx_v_self->is_leaf_node); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->leaf_size); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (unlikely(!__pyx_v_self->left.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_8 = __pyx_memoryview_fromslice(__pyx_v_self->left, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_uint64_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_uint64_t, 0);; if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_PyInt_From_npy_uint64(__pyx_v_self->max_right); if (unlikely(!__pyx_t_9)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_10 = __Pyx_PyInt_From_npy_uint64(__pyx_v_self->min_left); if (unlikely(!__pyx_t_10)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_11 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_center); if (unlikely(!__pyx_t_11)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_12 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_elements); if (unlikely(!__pyx_t_12)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __pyx_t_13 = __Pyx_PyInt_From_npy_uint64(__pyx_v_self->pivot); if (unlikely(!__pyx_t_13)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  if (unlikely(!__pyx_v_self->right.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_14 = __pyx_memoryview_fromslice(__pyx_v_self->right, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_uint64_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_uint64_t, 0);; if (unlikely(!__pyx_t_14)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __pyx_t_15 = PyTuple_New(16); if (unlikely(!__pyx_t_15)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_15, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_15, 1, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_15, 2, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_15, 3, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_15, 4, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_15, 5, __pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_15, 6, __pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_15, 7, __pyx_t_8);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->left_node));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self->left_node));
  PyTuple_SET_ITEM(__pyx_t_15, 8, ((PyObject *)__pyx_v_self->left_node));
  __Pyx_GIVEREF(__pyx_t_9);
  PyTuple_SET_ITEM(__pyx_t_15, 9, __pyx_t_9);
  __Pyx_GIVEREF(__pyx_t_10);
  PyTuple_SET_ITEM(__pyx_t_15, 10, __pyx_t_10);
  __Pyx_GIVEREF(__pyx_t_11);
  PyTuple_SET_ITEM(__pyx_t_15, 11, __pyx_t_11);
  __Pyx_GIVEREF(__pyx_t_12);
  PyTuple_SET_ITEM(__pyx_t_15, 12, __pyx_t_12);
  __Pyx_GIVEREF(__pyx_t_13);
  PyTuple_SET_ITEM(__pyx_t_15, 13, __pyx_t_13);
  __Pyx_GIVEREF(__pyx_t_14);
  PyTuple_SET_ITEM(__pyx_t_15, 14, __pyx_t_14);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->right_node));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self->right_node));
  PyTuple_SET_ITEM(__pyx_t_15, 15, ((PyObject *)__pyx_v_self->right_node));
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  __pyx_t_6 = 0;
  __pyx_t_7 = 0;
  __pyx_t_8 = 0;
  __pyx_t_9 = 0;
  __pyx_t_10 = 0;
  __pyx_t_11 = 0;
  __pyx_t_12 = 0;
  __pyx_t_13 = 0;
  __pyx_t_14 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_15);
  __pyx_t_15 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.center_left_indices, self.center_left_values, self.center_right_indices, self.center_right_values, self.indices, self.is_leaf_node, self.leaf_size, self.left, self.left_node, self.max_right, self.min_left, self.n_center, self.n_elements, self.pivot, self.right, self.right_node)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_15 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_15)) __PYX_ERR(2, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __pyx_v__dict = __pyx_t_15;
  __pyx_t_15 = 0;

  /* "(tree fragment)":7
 *     state = (self.center_left_indices, self.center_left_values, self.center_right_indices, self.center_right_values, self.indices, self.is_leaf_node, self.leaf_size, self.left, self.left_node, self.max_right, self.min_left, self.n_center, self.n_elements, self.pivot, self.right, self.right_node)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_16 = (__pyx_v__dict != Py_None);
  __pyx_t_17 = (__pyx_t_16 != 0);
  if (__pyx_t_17) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_15 = PyTuple_New(1); if (unlikely(!__pyx_t_15)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    PyTuple_SET_ITEM(__pyx_t_15, 0, __pyx_v__dict);
    __pyx_t_14 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_15); if (unlikely(!__pyx_t_14)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_14));
    __pyx_t_14 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.left_node is not None or self.right_node is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.center_left_indices, self.center_left_values, self.center_right_indices, self.center_right_values, self.indices, self.is_leaf_node, self.leaf_size, self.left, self.left_node, self.max_right, self.min_left, self.n_center, self.n_elements, self.pivot, self.right, self.right_node)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.left_node is not None or self.right_node is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_Uint64ClosedLeftIntervalNode, (type(self), 0x7a104e5, None), state
 */
  /*else*/ {
    __pyx_t_16 = (((PyObject *)__pyx_v_self->left_node) != Py_None);
    __pyx_t_18 = (__pyx_t_16 != 0);
    if (!__pyx_t_18) {
    } else {
      __pyx_t_17 = __pyx_t_18;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_18 = (((PyObject *)__pyx_v_self->right_node) != Py_None);
    __pyx_t_16 = (__pyx_t_18 != 0);
    __pyx_t_17 = __pyx_t_16;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_17;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.left_node is not None or self.right_node is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Uint64ClosedLeftIntervalNode, (type(self), 0x7a104e5, None), state
 *     else:
 */
  __pyx_t_17 = (__pyx_v_use_setstate != 0);
  if (__pyx_t_17) {

    /* "(tree fragment)":13
 *         use_setstate = self.left_node is not None or self.right_node is not None
 *     if use_setstate:
 *         return __pyx_unpickle_Uint64ClosedLeftIntervalNode, (type(self), 0x7a104e5, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_Uint64ClosedLeftIntervalNode, (type(self), 0x7a104e5, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_14, __pyx_n_s_pyx_unpickle_Uint64ClosedLeftI); if (unlikely(!__pyx_t_14)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __pyx_t_15 = PyTuple_New(3); if (unlikely(!__pyx_t_15)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_15, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_127993061);
    __Pyx_GIVEREF(__pyx_int_127993061);
    PyTuple_SET_ITEM(__pyx_t_15, 1, __pyx_int_127993061);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_15, 2, Py_None);
    __pyx_t_13 = PyTuple_New(3); if (unlikely(!__pyx_t_13)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __Pyx_GIVEREF(__pyx_t_14);
    PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_t_14);
    __Pyx_GIVEREF(__pyx_t_15);
    PyTuple_SET_ITEM(__pyx_t_13, 1, __pyx_t_15);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_13, 2, __pyx_v_state);
    __pyx_t_14 = 0;
    __pyx_t_15 = 0;
    __pyx_r = __pyx_t_13;
    __pyx_t_13 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.left_node is not None or self.right_node is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Uint64ClosedLeftIntervalNode, (type(self), 0x7a104e5, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_Uint64ClosedLeftIntervalNode, (type(self), 0x7a104e5, None), state
 *     else:
 *         return __pyx_unpickle_Uint64ClosedLeftIntervalNode, (type(self), 0x7a104e5, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Uint64ClosedLeftIntervalNode__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_13, __pyx_n_s_pyx_unpickle_Uint64ClosedLeftI); if (unlikely(!__pyx_t_13)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __pyx_t_15 = PyTuple_New(3); if (unlikely(!__pyx_t_15)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_15, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_127993061);
    __Pyx_GIVEREF(__pyx_int_127993061);
    PyTuple_SET_ITEM(__pyx_t_15, 1, __pyx_int_127993061);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_15, 2, __pyx_v_state);
    __pyx_t_14 = PyTuple_New(2); if (unlikely(!__pyx_t_14)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_GIVEREF(__pyx_t_13);
    PyTuple_SET_ITEM(__pyx_t_14, 0, __pyx_t_13);
    __Pyx_GIVEREF(__pyx_t_15);
    PyTuple_SET_ITEM(__pyx_t_14, 1, __pyx_t_15);
    __pyx_t_13 = 0;
    __pyx_t_15 = 0;
    __pyx_r = __pyx_t_14;
    __pyx_t_14 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_AddTraceback("pandas._libs.interval.Uint64ClosedLeftIntervalNode.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Uint64ClosedLeftIntervalNode, (type(self), 0x7a104e5, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Uint64ClosedLeftIntervalNode__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_23__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_23__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_22__setstate_cython__(((struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_22__setstate_cython__(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_Uint64ClosedLeftIntervalNode, (type(self), 0x7a104e5, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Uint64ClosedLeftIntervalNode__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(2, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_6pandas_5_libs_8interval___pyx_unpickle_Uint64ClosedLeftIntervalNode__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Uint64ClosedLeftIntervalNode, (type(self), 0x7a104e5, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Uint64ClosedLeftIntervalNode__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Uint64ClosedLeftIntervalNode.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":3138
 *         readonly bint is_leaf_node
 * 
 *     def __init__(self,             # <<<<<<<<<<<<<<
 *                  ndarray[uint64_t, ndim=1] left,
 *                  ndarray[uint64_t, ndim=1] right,
 */

/* Python wrapper */
static int __pyx_pw_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyArrayObject *__pyx_v_left = 0;
  PyArrayObject *__pyx_v_right = 0;
  PyArrayObject *__pyx_v_indices = 0;
  __pyx_t_5numpy_int64_t __pyx_v_leaf_size;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_left,&__pyx_n_s_right,&__pyx_n_s_indices,&__pyx_n_s_leaf_size,0};
    PyObject* values[4] = {0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_left)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_right)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 4, 4, 1); __PYX_ERR(1, 3138, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_indices)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 4, 4, 2); __PYX_ERR(1, 3138, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_leaf_size)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 4, 4, 3); __PYX_ERR(1, 3138, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(1, 3138, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 4) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
    }
    __pyx_v_left = ((PyArrayObject *)values[0]);
    __pyx_v_right = ((PyArrayObject *)values[1]);
    __pyx_v_indices = ((PyArrayObject *)values[2]);
    __pyx_v_leaf_size = __Pyx_PyInt_As_npy_int64(values[3]); if (unlikely((__pyx_v_leaf_size == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(1, 3142, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 4, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 3138, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Uint64ClosedRightIntervalNode.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_left), __pyx_ptype_5numpy_ndarray, 1, "left", 0))) __PYX_ERR(1, 3139, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_right), __pyx_ptype_5numpy_ndarray, 1, "right", 0))) __PYX_ERR(1, 3140, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_indices), __pyx_ptype_5numpy_ndarray, 1, "indices", 0))) __PYX_ERR(1, 3141, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode___init__(((struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode *)__pyx_v_self), __pyx_v_left, __pyx_v_right, __pyx_v_indices, __pyx_v_leaf_size);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode___init__(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode *__pyx_v_self, PyArrayObject *__pyx_v_left, PyArrayObject *__pyx_v_right, PyArrayObject *__pyx_v_indices, __pyx_t_5numpy_int64_t __pyx_v_leaf_size) {
  PyObject *__pyx_v_left_set = NULL;
  PyObject *__pyx_v_right_set = NULL;
  PyObject *__pyx_v_center_set = NULL;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_indices;
  __Pyx_Buffer __pyx_pybuffer_indices;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_left;
  __Pyx_Buffer __pyx_pybuffer_left;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_right;
  __Pyx_Buffer __pyx_pybuffer_right;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __pyx_t_5numpy_uint64_t __pyx_t_6;
  __Pyx_memviewslice __pyx_t_7 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_8 = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  __Pyx_memviewslice __pyx_t_11 = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *(*__pyx_t_12)(PyObject *);
  size_t __pyx_t_13;
  __Pyx_RefNannySetupContext("__init__", 0);
  __pyx_pybuffer_left.pybuffer.buf = NULL;
  __pyx_pybuffer_left.refcount = 0;
  __pyx_pybuffernd_left.data = NULL;
  __pyx_pybuffernd_left.rcbuffer = &__pyx_pybuffer_left;
  __pyx_pybuffer_right.pybuffer.buf = NULL;
  __pyx_pybuffer_right.refcount = 0;
  __pyx_pybuffernd_right.data = NULL;
  __pyx_pybuffernd_right.rcbuffer = &__pyx_pybuffer_right;
  __pyx_pybuffer_indices.pybuffer.buf = NULL;
  __pyx_pybuffer_indices.refcount = 0;
  __pyx_pybuffernd_indices.data = NULL;
  __pyx_pybuffernd_indices.rcbuffer = &__pyx_pybuffer_indices;
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_left.rcbuffer->pybuffer, (PyObject*)__pyx_v_left, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 3138, __pyx_L1_error)
  }
  __pyx_pybuffernd_left.diminfo[0].strides = __pyx_pybuffernd_left.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_left.diminfo[0].shape = __pyx_pybuffernd_left.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_right.rcbuffer->pybuffer, (PyObject*)__pyx_v_right, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 3138, __pyx_L1_error)
  }
  __pyx_pybuffernd_right.diminfo[0].strides = __pyx_pybuffernd_right.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_right.diminfo[0].shape = __pyx_pybuffernd_right.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_indices.rcbuffer->pybuffer, (PyObject*)__pyx_v_indices, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 3138, __pyx_L1_error)
  }
  __pyx_pybuffernd_indices.diminfo[0].strides = __pyx_pybuffernd_indices.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_indices.diminfo[0].shape = __pyx_pybuffernd_indices.rcbuffer->pybuffer.shape[0];

  /* "pandas/_libs/intervaltree.pxi":3144
 *                  int64_t leaf_size):
 * 
 *         self.n_elements = len(left)             # <<<<<<<<<<<<<<
 *         self.leaf_size = leaf_size
 * 
 */
  __pyx_t_1 = PyObject_Length(((PyObject *)__pyx_v_left)); if (unlikely(__pyx_t_1 == ((Py_ssize_t)-1))) __PYX_ERR(1, 3144, __pyx_L1_error)
  __pyx_v_self->n_elements = __pyx_t_1;

  /* "pandas/_libs/intervaltree.pxi":3145
 * 
 *         self.n_elements = len(left)
 *         self.leaf_size = leaf_size             # <<<<<<<<<<<<<<
 * 
 *         # min_left and min_right are used to speed-up query by skipping
 */
  __pyx_v_self->leaf_size = __pyx_v_leaf_size;

  /* "pandas/_libs/intervaltree.pxi":3150
 *         # query on sub-nodes. If this node has size 0, query is cheap,
 *         # so these values don't matter.
 *         if left.size > 0:             # <<<<<<<<<<<<<<
 *             self.min_left = left.min()
 *             self.max_right = right.max()
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_left), __pyx_n_s_size); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3150, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyObject_RichCompare(__pyx_t_2, __pyx_int_0, Py_GT); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 3150, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(1, 3150, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_4) {

    /* "pandas/_libs/intervaltree.pxi":3151
 *         # so these values don't matter.
 *         if left.size > 0:
 *             self.min_left = left.min()             # <<<<<<<<<<<<<<
 *             self.max_right = right.max()
 *         else:
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_left), __pyx_n_s_min); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3151, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 3151, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_6 = __Pyx_PyInt_As_npy_uint64(__pyx_t_3); if (unlikely((__pyx_t_6 == ((npy_uint64)-1)) && PyErr_Occurred())) __PYX_ERR(1, 3151, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_self->min_left = __pyx_t_6;

    /* "pandas/_libs/intervaltree.pxi":3152
 *         if left.size > 0:
 *             self.min_left = left.min()
 *             self.max_right = right.max()             # <<<<<<<<<<<<<<
 *         else:
 *             self.min_left = 0
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_right), __pyx_n_s_max); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3152, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 3152, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_6 = __Pyx_PyInt_As_npy_uint64(__pyx_t_3); if (unlikely((__pyx_t_6 == ((npy_uint64)-1)) && PyErr_Occurred())) __PYX_ERR(1, 3152, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_self->max_right = __pyx_t_6;

    /* "pandas/_libs/intervaltree.pxi":3150
 *         # query on sub-nodes. If this node has size 0, query is cheap,
 *         # so these values don't matter.
 *         if left.size > 0:             # <<<<<<<<<<<<<<
 *             self.min_left = left.min()
 *             self.max_right = right.max()
 */
    goto __pyx_L3;
  }

  /* "pandas/_libs/intervaltree.pxi":3154
 *             self.max_right = right.max()
 *         else:
 *             self.min_left = 0             # <<<<<<<<<<<<<<
 *             self.max_right = 0
 * 
 */
  /*else*/ {
    __pyx_v_self->min_left = 0;

    /* "pandas/_libs/intervaltree.pxi":3155
 *         else:
 *             self.min_left = 0
 *             self.max_right = 0             # <<<<<<<<<<<<<<
 * 
 *         if self.n_elements <= leaf_size:
 */
    __pyx_v_self->max_right = 0;
  }
  __pyx_L3:;

  /* "pandas/_libs/intervaltree.pxi":3157
 *             self.max_right = 0
 * 
 *         if self.n_elements <= leaf_size:             # <<<<<<<<<<<<<<
 *             # make this a terminal (leaf) node
 *             self.is_leaf_node = True
 */
  __pyx_t_4 = ((__pyx_v_self->n_elements <= __pyx_v_leaf_size) != 0);
  if (__pyx_t_4) {

    /* "pandas/_libs/intervaltree.pxi":3159
 *         if self.n_elements <= leaf_size:
 *             # make this a terminal (leaf) node
 *             self.is_leaf_node = True             # <<<<<<<<<<<<<<
 *             self.left = left
 *             self.right = right
 */
    __pyx_v_self->is_leaf_node = 1;

    /* "pandas/_libs/intervaltree.pxi":3160
 *             # make this a terminal (leaf) node
 *             self.is_leaf_node = True
 *             self.left = left             # <<<<<<<<<<<<<<
 *             self.right = right
 *             self.indices = indices
 */
    __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_uint64_t(((PyObject *)__pyx_v_left), PyBUF_WRITABLE); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(1, 3160, __pyx_L1_error)
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->left, 0);
    __pyx_v_self->left = __pyx_t_7;
    __pyx_t_7.memview = NULL;
    __pyx_t_7.data = NULL;

    /* "pandas/_libs/intervaltree.pxi":3161
 *             self.is_leaf_node = True
 *             self.left = left
 *             self.right = right             # <<<<<<<<<<<<<<
 *             self.indices = indices
 *             self.n_center = 0
 */
    __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_uint64_t(((PyObject *)__pyx_v_right), PyBUF_WRITABLE); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(1, 3161, __pyx_L1_error)
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->right, 0);
    __pyx_v_self->right = __pyx_t_7;
    __pyx_t_7.memview = NULL;
    __pyx_t_7.data = NULL;

    /* "pandas/_libs/intervaltree.pxi":3162
 *             self.left = left
 *             self.right = right
 *             self.indices = indices             # <<<<<<<<<<<<<<
 *             self.n_center = 0
 *         else:
 */
    __pyx_t_8 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(((PyObject *)__pyx_v_indices), PyBUF_WRITABLE); if (unlikely(!__pyx_t_8.memview)) __PYX_ERR(1, 3162, __pyx_L1_error)
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->indices, 0);
    __pyx_v_self->indices = __pyx_t_8;
    __pyx_t_8.memview = NULL;
    __pyx_t_8.data = NULL;

    /* "pandas/_libs/intervaltree.pxi":3163
 *             self.right = right
 *             self.indices = indices
 *             self.n_center = 0             # <<<<<<<<<<<<<<
 *         else:
 *             # calculate a pivot so we can create child nodes
 */
    __pyx_v_self->n_center = 0;

    /* "pandas/_libs/intervaltree.pxi":3157
 *             self.max_right = 0
 * 
 *         if self.n_elements <= leaf_size:             # <<<<<<<<<<<<<<
 *             # make this a terminal (leaf) node
 *             self.is_leaf_node = True
 */
    goto __pyx_L4;
  }

  /* "pandas/_libs/intervaltree.pxi":3166
 *         else:
 *             # calculate a pivot so we can create child nodes
 *             self.is_leaf_node = False             # <<<<<<<<<<<<<<
 *             self.pivot = np.median(left / 2 + right / 2)
 *             left_set, right_set, center_set = self.classify_intervals(
 */
  /*else*/ {
    __pyx_v_self->is_leaf_node = 0;

    /* "pandas/_libs/intervaltree.pxi":3167
 *             # calculate a pivot so we can create child nodes
 *             self.is_leaf_node = False
 *             self.pivot = np.median(left / 2 + right / 2)             # <<<<<<<<<<<<<<
 *             left_set, right_set, center_set = self.classify_intervals(
 *                 left, right)
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3167, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_median); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 3167, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyNumber_Divide(((PyObject *)__pyx_v_left), __pyx_int_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3167, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_9 = __Pyx_PyNumber_Divide(((PyObject *)__pyx_v_right), __pyx_int_2); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 3167, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_10 = PyNumber_Add(__pyx_t_2, __pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 3167, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_9 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_3 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_9, __pyx_t_10) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_10);
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 3167, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_6 = __Pyx_PyInt_As_npy_uint64(__pyx_t_3); if (unlikely((__pyx_t_6 == ((npy_uint64)-1)) && PyErr_Occurred())) __PYX_ERR(1, 3167, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_self->pivot = __pyx_t_6;

    /* "pandas/_libs/intervaltree.pxi":3169
 *             self.pivot = np.median(left / 2 + right / 2)
 *             left_set, right_set, center_set = self.classify_intervals(
 *                 left, right)             # <<<<<<<<<<<<<<
 * 
 *             self.left_node = self.new_child_node(left, right,
 */
    __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_uint64_t(((PyObject *)__pyx_v_left), PyBUF_WRITABLE); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(1, 3169, __pyx_L1_error)
    __pyx_t_11 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_uint64_t(((PyObject *)__pyx_v_right), PyBUF_WRITABLE); if (unlikely(!__pyx_t_11.memview)) __PYX_ERR(1, 3169, __pyx_L1_error)

    /* "pandas/_libs/intervaltree.pxi":3168
 *             self.is_leaf_node = False
 *             self.pivot = np.median(left / 2 + right / 2)
 *             left_set, right_set, center_set = self.classify_intervals(             # <<<<<<<<<<<<<<
 *                 left, right)
 * 
 */
    __pyx_t_3 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode *)__pyx_v_self->__pyx_vtab)->classify_intervals(__pyx_v_self, __pyx_t_7, __pyx_t_11); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 3168, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __PYX_XDEC_MEMVIEW(&__pyx_t_7, 1);
    __pyx_t_7.memview = NULL;
    __pyx_t_7.data = NULL;
    __PYX_XDEC_MEMVIEW(&__pyx_t_11, 1);
    __pyx_t_11.memview = NULL;
    __pyx_t_11.data = NULL;
    if ((likely(PyTuple_CheckExact(__pyx_t_3))) || (PyList_CheckExact(__pyx_t_3))) {
      PyObject* sequence = __pyx_t_3;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 3)) {
        if (size > 3) __Pyx_RaiseTooManyValuesError(3);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(1, 3168, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_5 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_10 = PyTuple_GET_ITEM(sequence, 1); 
        __pyx_t_9 = PyTuple_GET_ITEM(sequence, 2); 
      } else {
        __pyx_t_5 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_10 = PyList_GET_ITEM(sequence, 1); 
        __pyx_t_9 = PyList_GET_ITEM(sequence, 2); 
      }
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_10);
      __Pyx_INCREF(__pyx_t_9);
      #else
      __pyx_t_5 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 3168, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_10 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 3168, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_9 = PySequence_ITEM(sequence, 2); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 3168, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      #endif
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_2 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3168, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_12 = Py_TYPE(__pyx_t_2)->tp_iternext;
      index = 0; __pyx_t_5 = __pyx_t_12(__pyx_t_2); if (unlikely(!__pyx_t_5)) goto __pyx_L5_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_5);
      index = 1; __pyx_t_10 = __pyx_t_12(__pyx_t_2); if (unlikely(!__pyx_t_10)) goto __pyx_L5_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_10);
      index = 2; __pyx_t_9 = __pyx_t_12(__pyx_t_2); if (unlikely(!__pyx_t_9)) goto __pyx_L5_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_9);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_12(__pyx_t_2), 3) < 0) __PYX_ERR(1, 3168, __pyx_L1_error)
      __pyx_t_12 = NULL;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      goto __pyx_L6_unpacking_done;
      __pyx_L5_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_12 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(1, 3168, __pyx_L1_error)
      __pyx_L6_unpacking_done:;
    }
    __pyx_v_left_set = __pyx_t_5;
    __pyx_t_5 = 0;
    __pyx_v_right_set = __pyx_t_10;
    __pyx_t_10 = 0;
    __pyx_v_center_set = __pyx_t_9;
    __pyx_t_9 = 0;

    /* "pandas/_libs/intervaltree.pxi":3172
 * 
 *             self.left_node = self.new_child_node(left, right,
 *                                                  indices, left_set)             # <<<<<<<<<<<<<<
 *             self.right_node = self.new_child_node(left, right,
 *                                                   indices, right_set)
 */
    if (!(likely(((__pyx_v_left_set) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_left_set, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(1, 3172, __pyx_L1_error)

    /* "pandas/_libs/intervaltree.pxi":3171
 *                 left, right)
 * 
 *             self.left_node = self.new_child_node(left, right,             # <<<<<<<<<<<<<<
 *                                                  indices, left_set)
 *             self.right_node = self.new_child_node(left, right,
 */
    __pyx_t_3 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode *)__pyx_v_self->__pyx_vtab)->new_child_node(__pyx_v_self, ((PyArrayObject *)__pyx_v_left), ((PyArrayObject *)__pyx_v_right), ((PyArrayObject *)__pyx_v_indices), ((PyArrayObject *)__pyx_v_left_set)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 3171, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode))))) __PYX_ERR(1, 3171, __pyx_L1_error)
    __Pyx_GIVEREF(__pyx_t_3);
    __Pyx_GOTREF(__pyx_v_self->left_node);
    __Pyx_DECREF(((PyObject *)__pyx_v_self->left_node));
    __pyx_v_self->left_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode *)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "pandas/_libs/intervaltree.pxi":3174
 *                                                  indices, left_set)
 *             self.right_node = self.new_child_node(left, right,
 *                                                   indices, right_set)             # <<<<<<<<<<<<<<
 * 
 *             self.center_left_values, self.center_left_indices = \
 */
    if (!(likely(((__pyx_v_right_set) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_right_set, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(1, 3174, __pyx_L1_error)

    /* "pandas/_libs/intervaltree.pxi":3173
 *             self.left_node = self.new_child_node(left, right,
 *                                                  indices, left_set)
 *             self.right_node = self.new_child_node(left, right,             # <<<<<<<<<<<<<<
 *                                                   indices, right_set)
 * 
 */
    __pyx_t_3 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode *)__pyx_v_self->__pyx_vtab)->new_child_node(__pyx_v_self, ((PyArrayObject *)__pyx_v_left), ((PyArrayObject *)__pyx_v_right), ((PyArrayObject *)__pyx_v_indices), ((PyArrayObject *)__pyx_v_right_set)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 3173, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode))))) __PYX_ERR(1, 3173, __pyx_L1_error)
    __Pyx_GIVEREF(__pyx_t_3);
    __Pyx_GOTREF(__pyx_v_self->right_node);
    __Pyx_DECREF(((PyObject *)__pyx_v_self->right_node));
    __pyx_v_self->right_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode *)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "pandas/_libs/intervaltree.pxi":3177
 * 
 *             self.center_left_values, self.center_left_indices = \
 *                 sort_values_and_indices(left, indices, center_set)             # <<<<<<<<<<<<<<
 *             self.center_right_values, self.center_right_indices = \
 *                 sort_values_and_indices(right, indices, center_set)
 */
    __pyx_t_3 = __pyx_f_6pandas_5_libs_8interval_sort_values_and_indices(((PyObject *)__pyx_v_left), ((PyObject *)__pyx_v_indices), __pyx_v_center_set); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 3177, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if ((likely(PyTuple_CheckExact(__pyx_t_3))) || (PyList_CheckExact(__pyx_t_3))) {
      PyObject* sequence = __pyx_t_3;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(1, 3176, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_9 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_10 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_9 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_10 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_9);
      __Pyx_INCREF(__pyx_t_10);
      #else
      __pyx_t_9 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 3176, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_10 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 3176, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      #endif
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_5 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 3176, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_12 = Py_TYPE(__pyx_t_5)->tp_iternext;
      index = 0; __pyx_t_9 = __pyx_t_12(__pyx_t_5); if (unlikely(!__pyx_t_9)) goto __pyx_L7_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_9);
      index = 1; __pyx_t_10 = __pyx_t_12(__pyx_t_5); if (unlikely(!__pyx_t_10)) goto __pyx_L7_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_10);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_12(__pyx_t_5), 2) < 0) __PYX_ERR(1, 3176, __pyx_L1_error)
      __pyx_t_12 = NULL;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      goto __pyx_L8_unpacking_done;
      __pyx_L7_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_12 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(1, 3176, __pyx_L1_error)
      __pyx_L8_unpacking_done:;
    }

    /* "pandas/_libs/intervaltree.pxi":3176
 *                                                   indices, right_set)
 * 
 *             self.center_left_values, self.center_left_indices = \             # <<<<<<<<<<<<<<
 *                 sort_values_and_indices(left, indices, center_set)
 *             self.center_right_values, self.center_right_indices = \
 */
    __pyx_t_11 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_uint64_t(__pyx_t_9, PyBUF_WRITABLE); if (unlikely(!__pyx_t_11.memview)) __PYX_ERR(1, 3176, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_8 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_t_10, PyBUF_WRITABLE); if (unlikely(!__pyx_t_8.memview)) __PYX_ERR(1, 3176, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->center_left_values, 0);
    __pyx_v_self->center_left_values = __pyx_t_11;
    __pyx_t_11.memview = NULL;
    __pyx_t_11.data = NULL;
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->center_left_indices, 0);
    __pyx_v_self->center_left_indices = __pyx_t_8;
    __pyx_t_8.memview = NULL;
    __pyx_t_8.data = NULL;

    /* "pandas/_libs/intervaltree.pxi":3179
 *                 sort_values_and_indices(left, indices, center_set)
 *             self.center_right_values, self.center_right_indices = \
 *                 sort_values_and_indices(right, indices, center_set)             # <<<<<<<<<<<<<<
 *             self.n_center = len(self.center_left_indices)
 * 
 */
    __pyx_t_3 = __pyx_f_6pandas_5_libs_8interval_sort_values_and_indices(((PyObject *)__pyx_v_right), ((PyObject *)__pyx_v_indices), __pyx_v_center_set); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 3179, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if ((likely(PyTuple_CheckExact(__pyx_t_3))) || (PyList_CheckExact(__pyx_t_3))) {
      PyObject* sequence = __pyx_t_3;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(1, 3178, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_10 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_9 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_10 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_9 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_10);
      __Pyx_INCREF(__pyx_t_9);
      #else
      __pyx_t_10 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 3178, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_9 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 3178, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      #endif
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_5 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 3178, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_12 = Py_TYPE(__pyx_t_5)->tp_iternext;
      index = 0; __pyx_t_10 = __pyx_t_12(__pyx_t_5); if (unlikely(!__pyx_t_10)) goto __pyx_L9_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_10);
      index = 1; __pyx_t_9 = __pyx_t_12(__pyx_t_5); if (unlikely(!__pyx_t_9)) goto __pyx_L9_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_9);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_12(__pyx_t_5), 2) < 0) __PYX_ERR(1, 3178, __pyx_L1_error)
      __pyx_t_12 = NULL;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      goto __pyx_L10_unpacking_done;
      __pyx_L9_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_12 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(1, 3178, __pyx_L1_error)
      __pyx_L10_unpacking_done:;
    }

    /* "pandas/_libs/intervaltree.pxi":3178
 *             self.center_left_values, self.center_left_indices = \
 *                 sort_values_and_indices(left, indices, center_set)
 *             self.center_right_values, self.center_right_indices = \             # <<<<<<<<<<<<<<
 *                 sort_values_and_indices(right, indices, center_set)
 *             self.n_center = len(self.center_left_indices)
 */
    __pyx_t_11 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_uint64_t(__pyx_t_10, PyBUF_WRITABLE); if (unlikely(!__pyx_t_11.memview)) __PYX_ERR(1, 3178, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_t_8 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_t_9, PyBUF_WRITABLE); if (unlikely(!__pyx_t_8.memview)) __PYX_ERR(1, 3178, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->center_right_values, 0);
    __pyx_v_self->center_right_values = __pyx_t_11;
    __pyx_t_11.memview = NULL;
    __pyx_t_11.data = NULL;
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->center_right_indices, 0);
    __pyx_v_self->center_right_indices = __pyx_t_8;
    __pyx_t_8.memview = NULL;
    __pyx_t_8.data = NULL;

    /* "pandas/_libs/intervaltree.pxi":3180
 *             self.center_right_values, self.center_right_indices = \
 *                 sort_values_and_indices(right, indices, center_set)
 *             self.n_center = len(self.center_left_indices)             # <<<<<<<<<<<<<<
 * 
 *     @cython.wraparound(False)
 */
    if (unlikely(!__pyx_v_self->center_left_indices.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(1, 3180, __pyx_L1_error)}
    __pyx_t_13 = __Pyx_MemoryView_Len(__pyx_v_self->center_left_indices); 
    __pyx_v_self->n_center = __pyx_t_13;
  }
  __pyx_L4:;

  /* "pandas/_libs/intervaltree.pxi":3138
 *         readonly bint is_leaf_node
 * 
 *     def __init__(self,             # <<<<<<<<<<<<<<
 *                  ndarray[uint64_t, ndim=1] left,
 *                  ndarray[uint64_t, ndim=1] right,
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __PYX_XDEC_MEMVIEW(&__pyx_t_7, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_8, 1);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __PYX_XDEC_MEMVIEW(&__pyx_t_11, 1);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indices.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_left.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_right.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("pandas._libs.interval.Uint64ClosedRightIntervalNode.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indices.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_left.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_right.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF(__pyx_v_left_set);
  __Pyx_XDECREF(__pyx_v_right_set);
  __Pyx_XDECREF(__pyx_v_center_set);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":3184
 *     @cython.wraparound(False)
 *     @cython.boundscheck(False)
 *     cdef classify_intervals(self, uint64_t[:] left, uint64_t[:] right):             # <<<<<<<<<<<<<<
 *         """Classify the given intervals based upon whether they fall to the
 *         left, right, or overlap with this node's pivot.
 */

static PyObject *__pyx_f_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_classify_intervals(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode *__pyx_v_self, __Pyx_memviewslice __pyx_v_left, __Pyx_memviewslice __pyx_v_right) {
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_left_ind = 0;
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_right_ind = 0;
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_overlapping_ind = 0;
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __pyx_t_5numpy_int64_t __pyx_t_2;
  __pyx_t_5numpy_int64_t __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  int __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  __Pyx_RefNannySetupContext("classify_intervals", 0);

  /* "pandas/_libs/intervaltree.pxi":3192
 *             Py_ssize_t i
 * 
 *         left_ind = Int64Vector()             # <<<<<<<<<<<<<<
 *         right_ind = Int64Vector()
 *         overlapping_ind = Int64Vector()
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3192, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_left_ind = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":3193
 * 
 *         left_ind = Int64Vector()
 *         right_ind = Int64Vector()             # <<<<<<<<<<<<<<
 *         overlapping_ind = Int64Vector()
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3193, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_right_ind = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":3194
 *         left_ind = Int64Vector()
 *         right_ind = Int64Vector()
 *         overlapping_ind = Int64Vector()             # <<<<<<<<<<<<<<
 * 
 *         for i in range(self.n_elements):
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3194, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_overlapping_ind = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":3196
 *         overlapping_ind = Int64Vector()
 * 
 *         for i in range(self.n_elements):             # <<<<<<<<<<<<<<
 *             if right[i] < self.pivot:
 *                 left_ind.append(i)
 */
  __pyx_t_2 = __pyx_v_self->n_elements;
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "pandas/_libs/intervaltree.pxi":3197
 * 
 *         for i in range(self.n_elements):
 *             if right[i] < self.pivot:             # <<<<<<<<<<<<<<
 *                 left_ind.append(i)
 *             elif self.pivot <= left[i]:
 */
    __pyx_t_5 = __pyx_v_i;
    __pyx_t_6 = (((*((__pyx_t_5numpy_uint64_t *) ( /* dim=0 */ (__pyx_v_right.data + __pyx_t_5 * __pyx_v_right.strides[0]) ))) < __pyx_v_self->pivot) != 0);
    if (__pyx_t_6) {

      /* "pandas/_libs/intervaltree.pxi":3198
 *         for i in range(self.n_elements):
 *             if right[i] < self.pivot:
 *                 left_ind.append(i)             # <<<<<<<<<<<<<<
 *             elif self.pivot <= left[i]:
 *                 right_ind.append(i)
 */
      ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_left_ind->__pyx_vtab)->append(__pyx_v_left_ind, __pyx_v_i);

      /* "pandas/_libs/intervaltree.pxi":3197
 * 
 *         for i in range(self.n_elements):
 *             if right[i] < self.pivot:             # <<<<<<<<<<<<<<
 *                 left_ind.append(i)
 *             elif self.pivot <= left[i]:
 */
      goto __pyx_L5;
    }

    /* "pandas/_libs/intervaltree.pxi":3199
 *             if right[i] < self.pivot:
 *                 left_ind.append(i)
 *             elif self.pivot <= left[i]:             # <<<<<<<<<<<<<<
 *                 right_ind.append(i)
 *             else:
 */
    __pyx_t_7 = __pyx_v_i;
    __pyx_t_6 = ((__pyx_v_self->pivot <= (*((__pyx_t_5numpy_uint64_t *) ( /* dim=0 */ (__pyx_v_left.data + __pyx_t_7 * __pyx_v_left.strides[0]) )))) != 0);
    if (__pyx_t_6) {

      /* "pandas/_libs/intervaltree.pxi":3200
 *                 left_ind.append(i)
 *             elif self.pivot <= left[i]:
 *                 right_ind.append(i)             # <<<<<<<<<<<<<<
 *             else:
 *                 overlapping_ind.append(i)
 */
      ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_right_ind->__pyx_vtab)->append(__pyx_v_right_ind, __pyx_v_i);

      /* "pandas/_libs/intervaltree.pxi":3199
 *             if right[i] < self.pivot:
 *                 left_ind.append(i)
 *             elif self.pivot <= left[i]:             # <<<<<<<<<<<<<<
 *                 right_ind.append(i)
 *             else:
 */
      goto __pyx_L5;
    }

    /* "pandas/_libs/intervaltree.pxi":3202
 *                 right_ind.append(i)
 *             else:
 *                 overlapping_ind.append(i)             # <<<<<<<<<<<<<<
 * 
 *         return (left_ind.to_array(),
 */
    /*else*/ {
      ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_overlapping_ind->__pyx_vtab)->append(__pyx_v_overlapping_ind, __pyx_v_i);
    }
    __pyx_L5:;
  }

  /* "pandas/_libs/intervaltree.pxi":3204
 *                 overlapping_ind.append(i)
 * 
 *         return (left_ind.to_array(),             # <<<<<<<<<<<<<<
 *                 right_ind.to_array(),
 *                 overlapping_ind.to_array())
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_left_ind->__pyx_vtab)->to_array(__pyx_v_left_ind, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3204, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "pandas/_libs/intervaltree.pxi":3205
 * 
 *         return (left_ind.to_array(),
 *                 right_ind.to_array(),             # <<<<<<<<<<<<<<
 *                 overlapping_ind.to_array())
 * 
 */
  __pyx_t_8 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_right_ind->__pyx_vtab)->to_array(__pyx_v_right_ind, 0); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 3205, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);

  /* "pandas/_libs/intervaltree.pxi":3206
 *         return (left_ind.to_array(),
 *                 right_ind.to_array(),
 *                 overlapping_ind.to_array())             # <<<<<<<<<<<<<<
 * 
 *     cdef new_child_node(self,
 */
  __pyx_t_9 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_overlapping_ind->__pyx_vtab)->to_array(__pyx_v_overlapping_ind, 0); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 3206, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);

  /* "pandas/_libs/intervaltree.pxi":3204
 *                 overlapping_ind.append(i)
 * 
 *         return (left_ind.to_array(),             # <<<<<<<<<<<<<<
 *                 right_ind.to_array(),
 *                 overlapping_ind.to_array())
 */
  __pyx_t_10 = PyTuple_New(3); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 3204, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_t_8);
  __Pyx_GIVEREF(__pyx_t_9);
  PyTuple_SET_ITEM(__pyx_t_10, 2, __pyx_t_9);
  __pyx_t_1 = 0;
  __pyx_t_8 = 0;
  __pyx_t_9 = 0;
  __pyx_r = __pyx_t_10;
  __pyx_t_10 = 0;
  goto __pyx_L0;

  /* "pandas/_libs/intervaltree.pxi":3184
 *     @cython.wraparound(False)
 *     @cython.boundscheck(False)
 *     cdef classify_intervals(self, uint64_t[:] left, uint64_t[:] right):             # <<<<<<<<<<<<<<
 *         """Classify the given intervals based upon whether they fall to the
 *         left, right, or overlap with this node's pivot.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("pandas._libs.interval.Uint64ClosedRightIntervalNode.classify_intervals", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_left_ind);
  __Pyx_XDECREF((PyObject *)__pyx_v_right_ind);
  __Pyx_XDECREF((PyObject *)__pyx_v_overlapping_ind);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":3208
 *                 overlapping_ind.to_array())
 * 
 *     cdef new_child_node(self,             # <<<<<<<<<<<<<<
 *                         ndarray[uint64_t, ndim=1] left,
 *                         ndarray[uint64_t, ndim=1] right,
 */

static PyObject *__pyx_f_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_new_child_node(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode *__pyx_v_self, PyArrayObject *__pyx_v_left, PyArrayObject *__pyx_v_right, PyArrayObject *__pyx_v_indices, PyArrayObject *__pyx_v_subset) {
  __Pyx_LocalBuf_ND __pyx_pybuffernd_indices;
  __Pyx_Buffer __pyx_pybuffer_indices;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_left;
  __Pyx_Buffer __pyx_pybuffer_left;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_right;
  __Pyx_Buffer __pyx_pybuffer_right;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_subset;
  __Pyx_Buffer __pyx_pybuffer_subset;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyArrayObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyArrayObject *__pyx_t_7 = NULL;
  PyArrayObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  __Pyx_RefNannySetupContext("new_child_node", 0);
  __Pyx_INCREF((PyObject *)__pyx_v_left);
  __Pyx_INCREF((PyObject *)__pyx_v_right);
  __Pyx_INCREF((PyObject *)__pyx_v_indices);
  __pyx_pybuffer_left.pybuffer.buf = NULL;
  __pyx_pybuffer_left.refcount = 0;
  __pyx_pybuffernd_left.data = NULL;
  __pyx_pybuffernd_left.rcbuffer = &__pyx_pybuffer_left;
  __pyx_pybuffer_right.pybuffer.buf = NULL;
  __pyx_pybuffer_right.refcount = 0;
  __pyx_pybuffernd_right.data = NULL;
  __pyx_pybuffernd_right.rcbuffer = &__pyx_pybuffer_right;
  __pyx_pybuffer_indices.pybuffer.buf = NULL;
  __pyx_pybuffer_indices.refcount = 0;
  __pyx_pybuffernd_indices.data = NULL;
  __pyx_pybuffernd_indices.rcbuffer = &__pyx_pybuffer_indices;
  __pyx_pybuffer_subset.pybuffer.buf = NULL;
  __pyx_pybuffer_subset.refcount = 0;
  __pyx_pybuffernd_subset.data = NULL;
  __pyx_pybuffernd_subset.rcbuffer = &__pyx_pybuffer_subset;
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_left.rcbuffer->pybuffer, (PyObject*)__pyx_v_left, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 3208, __pyx_L1_error)
  }
  __pyx_pybuffernd_left.diminfo[0].strides = __pyx_pybuffernd_left.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_left.diminfo[0].shape = __pyx_pybuffernd_left.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_right.rcbuffer->pybuffer, (PyObject*)__pyx_v_right, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 3208, __pyx_L1_error)
  }
  __pyx_pybuffernd_right.diminfo[0].strides = __pyx_pybuffernd_right.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_right.diminfo[0].shape = __pyx_pybuffernd_right.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_indices.rcbuffer->pybuffer, (PyObject*)__pyx_v_indices, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 3208, __pyx_L1_error)
  }
  __pyx_pybuffernd_indices.diminfo[0].strides = __pyx_pybuffernd_indices.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_indices.diminfo[0].shape = __pyx_pybuffernd_indices.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_subset.rcbuffer->pybuffer, (PyObject*)__pyx_v_subset, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 3208, __pyx_L1_error)
  }
  __pyx_pybuffernd_subset.diminfo[0].strides = __pyx_pybuffernd_subset.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_subset.diminfo[0].shape = __pyx_pybuffernd_subset.rcbuffer->pybuffer.shape[0];

  /* "pandas/_libs/intervaltree.pxi":3215
 *         """Create a new child node.
 *         """
 *         left = take(left, subset)             # <<<<<<<<<<<<<<
 *         right = take(right, subset)
 *         indices = take(indices, subset)
 */
  __pyx_t_1 = __pyx_f_6pandas_5_libs_8interval_take(((PyArrayObject *)__pyx_v_left), ((PyArrayObject *)__pyx_v_subset)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3215, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(1, 3215, __pyx_L1_error)
  __pyx_t_2 = ((PyArrayObject *)__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_left.rcbuffer->pybuffer);
    __pyx_t_3 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_left.rcbuffer->pybuffer, (PyObject*)__pyx_t_2, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack);
    if (unlikely(__pyx_t_3 < 0)) {
      PyErr_Fetch(&__pyx_t_4, &__pyx_t_5, &__pyx_t_6);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_left.rcbuffer->pybuffer, (PyObject*)__pyx_v_left, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_4); Py_XDECREF(__pyx_t_5); Py_XDECREF(__pyx_t_6);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_4, __pyx_t_5, __pyx_t_6);
      }
      __pyx_t_4 = __pyx_t_5 = __pyx_t_6 = 0;
    }
    __pyx_pybuffernd_left.diminfo[0].strides = __pyx_pybuffernd_left.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_left.diminfo[0].shape = __pyx_pybuffernd_left.rcbuffer->pybuffer.shape[0];
    if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(1, 3215, __pyx_L1_error)
  }
  __pyx_t_2 = 0;
  __Pyx_DECREF_SET(__pyx_v_left, ((PyArrayObject *)__pyx_t_1));
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":3216
 *         """
 *         left = take(left, subset)
 *         right = take(right, subset)             # <<<<<<<<<<<<<<
 *         indices = take(indices, subset)
 *         return Uint64ClosedRightIntervalNode(
 */
  __pyx_t_1 = __pyx_f_6pandas_5_libs_8interval_take(((PyArrayObject *)__pyx_v_right), ((PyArrayObject *)__pyx_v_subset)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3216, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(1, 3216, __pyx_L1_error)
  __pyx_t_7 = ((PyArrayObject *)__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_right.rcbuffer->pybuffer);
    __pyx_t_3 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_right.rcbuffer->pybuffer, (PyObject*)__pyx_t_7, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack);
    if (unlikely(__pyx_t_3 < 0)) {
      PyErr_Fetch(&__pyx_t_6, &__pyx_t_5, &__pyx_t_4);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_right.rcbuffer->pybuffer, (PyObject*)__pyx_v_right, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_6); Py_XDECREF(__pyx_t_5); Py_XDECREF(__pyx_t_4);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_6, __pyx_t_5, __pyx_t_4);
      }
      __pyx_t_6 = __pyx_t_5 = __pyx_t_4 = 0;
    }
    __pyx_pybuffernd_right.diminfo[0].strides = __pyx_pybuffernd_right.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_right.diminfo[0].shape = __pyx_pybuffernd_right.rcbuffer->pybuffer.shape[0];
    if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(1, 3216, __pyx_L1_error)
  }
  __pyx_t_7 = 0;
  __Pyx_DECREF_SET(__pyx_v_right, ((PyArrayObject *)__pyx_t_1));
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":3217
 *         left = take(left, subset)
 *         right = take(right, subset)
 *         indices = take(indices, subset)             # <<<<<<<<<<<<<<
 *         return Uint64ClosedRightIntervalNode(
 *             left, right, indices, self.leaf_size)
 */
  __pyx_t_1 = __pyx_f_6pandas_5_libs_8interval_take(((PyArrayObject *)__pyx_v_indices), ((PyArrayObject *)__pyx_v_subset)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3217, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(1, 3217, __pyx_L1_error)
  __pyx_t_8 = ((PyArrayObject *)__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indices.rcbuffer->pybuffer);
    __pyx_t_3 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_indices.rcbuffer->pybuffer, (PyObject*)__pyx_t_8, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack);
    if (unlikely(__pyx_t_3 < 0)) {
      PyErr_Fetch(&__pyx_t_4, &__pyx_t_5, &__pyx_t_6);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_indices.rcbuffer->pybuffer, (PyObject*)__pyx_v_indices, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_4); Py_XDECREF(__pyx_t_5); Py_XDECREF(__pyx_t_6);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_4, __pyx_t_5, __pyx_t_6);
      }
      __pyx_t_4 = __pyx_t_5 = __pyx_t_6 = 0;
    }
    __pyx_pybuffernd_indices.diminfo[0].strides = __pyx_pybuffernd_indices.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_indices.diminfo[0].shape = __pyx_pybuffernd_indices.rcbuffer->pybuffer.shape[0];
    if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(1, 3217, __pyx_L1_error)
  }
  __pyx_t_8 = 0;
  __Pyx_DECREF_SET(__pyx_v_indices, ((PyArrayObject *)__pyx_t_1));
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":3218
 *         right = take(right, subset)
 *         indices = take(indices, subset)
 *         return Uint64ClosedRightIntervalNode(             # <<<<<<<<<<<<<<
 *             left, right, indices, self.leaf_size)
 * 
 */
  __Pyx_XDECREF(__pyx_r);

  /* "pandas/_libs/intervaltree.pxi":3219
 *         indices = take(indices, subset)
 *         return Uint64ClosedRightIntervalNode(
 *             left, right, indices, self.leaf_size)             # <<<<<<<<<<<<<<
 * 
 *     @cython.wraparound(False)
 */
  __pyx_t_1 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->leaf_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3219, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "pandas/_libs/intervaltree.pxi":3218
 *         right = take(right, subset)
 *         indices = take(indices, subset)
 *         return Uint64ClosedRightIntervalNode(             # <<<<<<<<<<<<<<
 *             left, right, indices, self.leaf_size)
 * 
 */
  __pyx_t_9 = PyTuple_New(4); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 3218, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_INCREF(((PyObject *)__pyx_v_left));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_left));
  PyTuple_SET_ITEM(__pyx_t_9, 0, ((PyObject *)__pyx_v_left));
  __Pyx_INCREF(((PyObject *)__pyx_v_right));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_right));
  PyTuple_SET_ITEM(__pyx_t_9, 1, ((PyObject *)__pyx_v_right));
  __Pyx_INCREF(((PyObject *)__pyx_v_indices));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_indices));
  PyTuple_SET_ITEM(__pyx_t_9, 2, ((PyObject *)__pyx_v_indices));
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_9, 3, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode), __pyx_t_9, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3218, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pandas/_libs/intervaltree.pxi":3208
 *                 overlapping_ind.to_array())
 * 
 *     cdef new_child_node(self,             # <<<<<<<<<<<<<<
 *                         ndarray[uint64_t, ndim=1] left,
 *                         ndarray[uint64_t, ndim=1] right,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_9);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indices.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_left.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_right.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_subset.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("pandas._libs.interval.Uint64ClosedRightIntervalNode.new_child_node", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indices.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_left.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_right.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_subset.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF((PyObject *)__pyx_v_left);
  __Pyx_XDECREF((PyObject *)__pyx_v_right);
  __Pyx_XDECREF((PyObject *)__pyx_v_indices);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":3224
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_2query[] = "Recursively query this node and its sub-nodes for intervals that\n        overlap with the query point.\n        ";
static PyMethodDef __pyx_mdef_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_3query = {"query", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_3query, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_2query};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_args = 0;
  PyObject *__pyx_v_kwargs = 0;
  CYTHON_UNUSED PyObject *__pyx_v_defaults = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fused_cpdef (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_args,&__pyx_n_s_kwargs,&__pyx_n_s_defaults,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_args)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_kwargs)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fused_cpdef", 1, 3, 3, 1); __PYX_ERR(1, 3224, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_defaults)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fused_cpdef", 1, 3, 3, 2); __PYX_ERR(1, 3224, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fused_cpdef") < 0)) __PYX_ERR(1, 3224, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_args = values[0];
    __pyx_v_kwargs = values[1];
    __pyx_v_defaults = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fused_cpdef", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 3224, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Uint64ClosedRightIntervalNode.__pyx_fused_cpdef", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_2query(((struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode *)__pyx_v_signatures), __pyx_v_args, __pyx_v_kwargs, __pyx_v_defaults);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_2query(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode *__pyx_v_signatures, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs, CYTHON_UNUSED PyObject *__pyx_v_defaults) {
  PyObject *__pyx_v_dest_sig = NULL;
  Py_ssize_t __pyx_v_i;
  CYTHON_UNUSED PyTypeObject *__pyx_v_ndarray = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_v_itemsize;
  PyObject *__pyx_v_arg = NULL;
  PyObject *__pyx_v_candidates = NULL;
  PyObject *__pyx_v_sig = NULL;
  int __pyx_v_match_found;
  PyObject *__pyx_v_src_sig = NULL;
  PyObject *__pyx_v_dst_type = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  Py_ssize_t __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  Py_ssize_t __pyx_t_13;
  Py_ssize_t __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  int __pyx_t_16;
  __Pyx_RefNannySetupContext("query", 0);
  __Pyx_INCREF(__pyx_v_kwargs);
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3224, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  PyList_SET_ITEM(__pyx_t_1, 0, Py_None);
  __pyx_v_dest_sig = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_3 = (__pyx_v_kwargs != Py_None);
  __pyx_t_4 = (__pyx_t_3 != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_kwargs); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(1, 3224, __pyx_L1_error)
  __pyx_t_3 = ((!__pyx_t_4) != 0);
  __pyx_t_2 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {
    __Pyx_INCREF(Py_None);
    __Pyx_DECREF_SET(__pyx_v_kwargs, Py_None);
  }
  __pyx_t_1 = ((PyObject *)__Pyx_ImportNumPyArrayTypeIfAvailable()); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3224, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_ndarray = ((PyTypeObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_v_itemsize = -1L;
  if (unlikely(__pyx_v_args == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 3224, __pyx_L1_error)
  }
  __pyx_t_5 = PyTuple_GET_SIZE(((PyObject*)__pyx_v_args)); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(1, 3224, __pyx_L1_error)
  __pyx_t_2 = ((2 < __pyx_t_5) != 0);
  if (__pyx_t_2) {
    if (unlikely(__pyx_v_args == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 3224, __pyx_L1_error)
    }
    __pyx_t_1 = PyTuple_GET_ITEM(((PyObject*)__pyx_v_args), 2);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v_arg = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L6;
  }
  __pyx_t_3 = (__pyx_v_kwargs != Py_None);
  __pyx_t_4 = (__pyx_t_3 != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L7_bool_binop_done;
  }
  if (unlikely(__pyx_v_kwargs == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(1, 3224, __pyx_L1_error)
  }
  __pyx_t_4 = (__Pyx_PyDict_ContainsTF(__pyx_n_s_point, ((PyObject*)__pyx_v_kwargs), Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(1, 3224, __pyx_L1_error)
  __pyx_t_3 = (__pyx_t_4 != 0);
  __pyx_t_2 = __pyx_t_3;
  __pyx_L7_bool_binop_done:;
  if (__pyx_t_2) {
    if (unlikely(__pyx_v_kwargs == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 3224, __pyx_L1_error)
    }
    __pyx_t_1 = __Pyx_PyDict_GetItem(((PyObject*)__pyx_v_kwargs), __pyx_n_s_point); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3224, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_arg = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L6;
  }
  /*else*/ {
    if (unlikely(__pyx_v_args == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
      __PYX_ERR(1, 3224, __pyx_L1_error)
    }
    __pyx_t_5 = PyTuple_GET_SIZE(((PyObject*)__pyx_v_args)); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(1, 3224, __pyx_L1_error)
    __pyx_t_1 = PyInt_FromSsize_t(__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3224, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_6 = PyTuple_New(3); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 3224, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_INCREF(__pyx_int_3);
    __Pyx_GIVEREF(__pyx_int_3);
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_int_3);
    __Pyx_INCREF(__pyx_n_s_s);
    __Pyx_GIVEREF(__pyx_n_s_s);
    PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_n_s_s);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Expected_at_least_d_argument_s_g, __pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3224, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_1); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 3224, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(1, 3224, __pyx_L1_error)
  }
  __pyx_L6:;
  while (1) {
    __pyx_t_2 = PyFloat_Check(__pyx_v_arg); 
    __pyx_t_3 = (__pyx_t_2 != 0);
    if (__pyx_t_3) {
      if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, __pyx_n_s_float64_t, long, 1, __Pyx_PyInt_From_long, 1, 0, 0) < 0)) __PYX_ERR(1, 3224, __pyx_L1_error)
      goto __pyx_L10_break;
    }
    __pyx_t_2 = PyInt_Check(__pyx_v_arg); 
    __pyx_t_4 = (__pyx_t_2 != 0);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_3 = __pyx_t_4;
      goto __pyx_L13_bool_binop_done;
    }
    __pyx_t_4 = PyLong_Check(__pyx_v_arg); 
    __pyx_t_2 = (__pyx_t_4 != 0);
    __pyx_t_3 = __pyx_t_2;
    __pyx_L13_bool_binop_done:;
    __pyx_t_2 = (__pyx_t_3 != 0);
    if (__pyx_t_2) {
      if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, __pyx_n_s_int64_t, long, 1, __Pyx_PyInt_From_long, 1, 0, 0) < 0)) __PYX_ERR(1, 3224, __pyx_L1_error)
      goto __pyx_L10_break;
    }
    if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, Py_None, long, 1, __Pyx_PyInt_From_long, 1, 0, 0) < 0)) __PYX_ERR(1, 3224, __pyx_L1_error)
    goto __pyx_L10_break;
  }
  __pyx_L10_break:;
  __pyx_t_6 = PyList_New(0); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 3224, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_v_candidates = ((PyObject*)__pyx_t_6);
  __pyx_t_6 = 0;
  __pyx_t_5 = 0;
  if (unlikely(((PyObject *)__pyx_v_signatures) == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(1, 3224, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_dict_iterator(((PyObject*)__pyx_v_signatures), 1, ((PyObject *)NULL), (&__pyx_t_7), (&__pyx_t_8)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3224, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __pyx_t_6 = __pyx_t_1;
  __pyx_t_1 = 0;
  while (1) {
    __pyx_t_9 = __Pyx_dict_iter_next(__pyx_t_6, __pyx_t_7, &__pyx_t_5, &__pyx_t_1, NULL, NULL, __pyx_t_8);
    if (unlikely(__pyx_t_9 == 0)) break;
    if (unlikely(__pyx_t_9 == -1)) __PYX_ERR(1, 3224, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_sig, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_v_match_found = 0;
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_v_sig, __pyx_n_s_strip); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 3224, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __pyx_t_12 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_11))) {
      __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_11);
      if (likely(__pyx_t_12)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
        __Pyx_INCREF(__pyx_t_12);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_11, function);
      }
    }
    __pyx_t_10 = (__pyx_t_12) ? __Pyx_PyObject_Call2Args(__pyx_t_11, __pyx_t_12, __pyx_kp_s__13) : __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_kp_s__13);
    __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
    if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 3224, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_split); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 3224, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_t_10 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_11))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_11);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_11, function);
      }
    }
    __pyx_t_1 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_11, __pyx_t_10, __pyx_kp_s__14) : __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_kp_s__14);
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3224, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_XDECREF_SET(__pyx_v_src_sig, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_13 = PyList_GET_SIZE(__pyx_v_dest_sig); if (unlikely(__pyx_t_13 == ((Py_ssize_t)-1))) __PYX_ERR(1, 3224, __pyx_L1_error)
    __pyx_t_14 = __pyx_t_13;
    for (__pyx_t_15 = 0; __pyx_t_15 < __pyx_t_14; __pyx_t_15+=1) {
      __pyx_v_i = __pyx_t_15;
      __pyx_t_1 = PyList_GET_ITEM(__pyx_v_dest_sig, __pyx_v_i);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_XDECREF_SET(__pyx_v_dst_type, __pyx_t_1);
      __pyx_t_1 = 0;
      __pyx_t_2 = (__pyx_v_dst_type != Py_None);
      __pyx_t_3 = (__pyx_t_2 != 0);
      if (__pyx_t_3) {
        __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_src_sig, __pyx_v_i, Py_ssize_t, 1, PyInt_FromSsize_t, 0, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3224, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_11 = PyObject_RichCompare(__pyx_t_1, __pyx_v_dst_type, Py_EQ); __Pyx_XGOTREF(__pyx_t_11); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 3224, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_11); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(1, 3224, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        if (__pyx_t_3) {
          __pyx_v_match_found = 1;
          goto __pyx_L20;
        }
        /*else*/ {
          __pyx_v_match_found = 0;
          goto __pyx_L18_break;
        }
        __pyx_L20:;
      }
    }
    __pyx_L18_break:;
    __pyx_t_3 = (__pyx_v_match_found != 0);
    if (__pyx_t_3) {
      __pyx_t_16 = __Pyx_PyList_Append(__pyx_v_candidates, __pyx_v_sig); if (unlikely(__pyx_t_16 == ((int)-1))) __PYX_ERR(1, 3224, __pyx_L1_error)
    }
  }
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_3 = (PyList_GET_SIZE(__pyx_v_candidates) != 0);
  __pyx_t_2 = ((!__pyx_t_3) != 0);
  if (__pyx_t_2) {
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__15, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 3224, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(1, 3224, __pyx_L1_error)
  }
  __pyx_t_7 = PyList_GET_SIZE(__pyx_v_candidates); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1))) __PYX_ERR(1, 3224, __pyx_L1_error)
  __pyx_t_2 = ((__pyx_t_7 > 1) != 0);
  if (__pyx_t_2) {
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__16, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 3224, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(1, 3224, __pyx_L1_error)
  }
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    if (unlikely(((PyObject *)__pyx_v_signatures) == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 3224, __pyx_L1_error)
    }
    __pyx_t_6 = __Pyx_PyDict_GetItem(((PyObject*)__pyx_v_signatures), PyList_GET_ITEM(__pyx_v_candidates, 0)); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 3224, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_r = __pyx_t_6;
    __pyx_t_6 = 0;
    goto __pyx_L0;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("pandas._libs.interval.Uint64ClosedRightIntervalNode.__pyx_fused_cpdef", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_dest_sig);
  __Pyx_XDECREF(__pyx_v_ndarray);
  __Pyx_XDECREF(__pyx_v_arg);
  __Pyx_XDECREF(__pyx_v_candidates);
  __Pyx_XDECREF(__pyx_v_sig);
  __Pyx_XDECREF(__pyx_v_src_sig);
  __Pyx_XDECREF(__pyx_v_dst_type);
  __Pyx_XDECREF(__pyx_v_kwargs);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_9__pyx_fuse_0query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_fuse_0__pyx_f_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float64_t __pyx_v_point, int __pyx_skip_dispatch) {
  __Pyx_memviewslice __pyx_v_indices = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_values = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __pyx_t_5numpy_int64_t __pyx_t_9;
  __pyx_t_5numpy_int64_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  int __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  __Pyx_memviewslice __pyx_t_16 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_17 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  Py_ssize_t __pyx_t_21;
  __Pyx_RefNannySetupContext("__pyx_fuse_0query", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pyx_fuse_0query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3224, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_9__pyx_fuse_0query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = PyFloat_FromDouble(__pyx_v_point); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 3224, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3224, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3224, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 3224, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_5) {
            __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3224, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3224, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_3query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_4 = PyFloat_FromDouble(__pyx_v_point); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 3224, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_7 = __pyx_t_1; __pyx_t_3 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3224, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3224, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 3224, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (__pyx_t_3) {
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_t_4);
          __pyx_t_4 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3224, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pandas/_libs/intervaltree.pxi":3233
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
  __pyx_t_8 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_8) {

    /* "pandas/_libs/intervaltree.pxi":3237
 *             # continue the binary tree structure. Instead, we use linear
 *             # search.
 *             for i in range(self.n_elements):             # <<<<<<<<<<<<<<
 *                 if self.left[i] < point <= self.right[i]:
 *                     result.append(self.indices[i])
 */
    __pyx_t_9 = __pyx_v_self->n_elements;
    __pyx_t_10 = __pyx_t_9;
    for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
      __pyx_v_i = __pyx_t_11;

      /* "pandas/_libs/intervaltree.pxi":3238
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point <= self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      __pyx_t_12 = __pyx_v_i;
      __pyx_t_8 = ((*((__pyx_t_5numpy_uint64_t *) ( /* dim=0 */ (__pyx_v_self->left.data + __pyx_t_12 * __pyx_v_self->left.strides[0]) ))) < __pyx_v_point);
      if (__pyx_t_8) {
        __pyx_t_13 = __pyx_v_i;
        __pyx_t_8 = (__pyx_v_point <= (*((__pyx_t_5numpy_uint64_t *) ( /* dim=0 */ (__pyx_v_self->right.data + __pyx_t_13 * __pyx_v_self->right.strides[0]) ))));
      }
      __pyx_t_14 = (__pyx_t_8 != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":3239
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point <= self.right[i]:
 *                     result.append(self.indices[i])             # <<<<<<<<<<<<<<
 *         else:
 *             # There are child nodes. Based on comparing our query to the pivot,
 */
        __pyx_t_15 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->indices.data + __pyx_t_15 * __pyx_v_self->indices.strides[0]) ))));

        /* "pandas/_libs/intervaltree.pxi":3238
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point <= self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      }
    }

    /* "pandas/_libs/intervaltree.pxi":3233
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
    goto __pyx_L3;
  }

  /* "pandas/_libs/intervaltree.pxi":3243
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
  /*else*/ {
    __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":3244
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:
 *                 values = self.center_left_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 */
      __pyx_t_16 = __pyx_v_self->center_left_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":3245
 *             if point < self.pivot:
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:
 */
      __pyx_t_17 = __pyx_v_self->center_left_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":3246
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):             # <<<<<<<<<<<<<<
 *                     if not values[i] < point:
 *                         break
 */
      __pyx_t_9 = __pyx_v_self->n_center;
      __pyx_t_10 = __pyx_t_9;
      for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":3247
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_18 = __pyx_v_i;
        __pyx_t_14 = ((!(((*((__pyx_t_5numpy_uint64_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_18 * __pyx_v_values.strides[0]) ))) < __pyx_v_point) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":3248
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:
 */
          goto __pyx_L9_break;

          /* "pandas/_libs/intervaltree.pxi":3247
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":3249
 *                     if not values[i] < point:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 */
        __pyx_t_19 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_19 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L9_break:;

      /* "pandas/_libs/intervaltree.pxi":3250
 *                         break
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      __pyx_t_14 = ((__pyx_v_point <= __pyx_v_self->left_node->max_right) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":3251
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)             # <<<<<<<<<<<<<<
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode *)__pyx_v_self->left_node->__pyx_vtab)->__pyx_fuse_0query(__pyx_v_self->left_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3251, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":3250
 *                         break
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":3243
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":3252
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
    __pyx_t_14 = ((__pyx_v_point > __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":3253
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 *                 values = self.center_right_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 */
      __pyx_t_16 = __pyx_v_self->center_right_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":3254
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:
 */
      __pyx_t_17 = __pyx_v_self->center_right_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":3255
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):             # <<<<<<<<<<<<<<
 *                     if not point <= values[i]:
 *                         break
 */
      for (__pyx_t_11 = (__pyx_v_self->n_center - 1); __pyx_t_11 > -1L; __pyx_t_11-=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":3256
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_20 = __pyx_v_i;
        __pyx_t_14 = ((!((__pyx_v_point <= (*((__pyx_t_5numpy_uint64_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_20 * __pyx_v_values.strides[0]) )))) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":3257
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:
 */
          goto __pyx_L13_break;

          /* "pandas/_libs/intervaltree.pxi":3256
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":3258
 *                     if not point <= values[i]:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if self.right_node.min_left < point:
 *                     self.right_node.query(result, point)
 */
        __pyx_t_21 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_21 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L13_break:;

      /* "pandas/_libs/intervaltree.pxi":3259
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      __pyx_t_14 = ((__pyx_v_self->right_node->min_left < __pyx_v_point) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":3260
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:
 *                     self.right_node.query(result, point)             # <<<<<<<<<<<<<<
 *             else:
 *                 result.extend(self.center_left_indices)
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode *)__pyx_v_self->right_node->__pyx_vtab)->__pyx_fuse_0query(__pyx_v_self->right_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3260, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":3259
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":3252
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":3262
 *                     self.right_node.query(result, point)
 *             else:
 *                 result.extend(self.center_left_indices)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
    /*else*/ {
      __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->extend(__pyx_v_result, __pyx_v_self->center_left_indices); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3262, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_L7:;
  }
  __pyx_L3:;

  /* "pandas/_libs/intervaltree.pxi":3224
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __PYX_XDEC_MEMVIEW(&__pyx_t_16, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_17, 1);
  __Pyx_AddTraceback("pandas._libs.interval.Uint64ClosedRightIntervalNode.query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_indices, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_values, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_9__pyx_fuse_0query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_0__pyx_mdef_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_9__pyx_fuse_0query = {"__pyx_fuse_0query", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_9__pyx_fuse_0query, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_2query};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_9__pyx_fuse_0query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result = 0;
  __pyx_t_5numpy_float64_t __pyx_v_point;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fuse_0query (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_result,&__pyx_n_s_point,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_result)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_point)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_0query", 1, 2, 2, 1); __PYX_ERR(1, 3224, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_0query") < 0)) __PYX_ERR(1, 3224, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_result = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)values[0]);
    __pyx_v_point = __pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_point == ((npy_float64)-1)) && PyErr_Occurred())) __PYX_ERR(1, 3224, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_0query", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 3224, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Uint64ClosedRightIntervalNode.__pyx_fuse_0query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_result), __pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector, 1, "result", 0))) __PYX_ERR(1, 3224, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_8__pyx_fuse_0query(((struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode *)__pyx_v_self), __pyx_v_result, __pyx_v_point);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_8__pyx_fuse_0query(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float64_t __pyx_v_point) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__pyx_fuse_0query", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_fuse_0__pyx_f_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_query(__pyx_v_self, __pyx_v_result, __pyx_v_point, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3224, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Uint64ClosedRightIntervalNode.__pyx_fuse_0query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_11__pyx_fuse_1query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_fuse_1__pyx_f_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float32_t __pyx_v_point, int __pyx_skip_dispatch) {
  __Pyx_memviewslice __pyx_v_indices = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_values = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __pyx_t_5numpy_int64_t __pyx_t_9;
  __pyx_t_5numpy_int64_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  int __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  __Pyx_memviewslice __pyx_t_16 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_17 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  Py_ssize_t __pyx_t_21;
  __Pyx_RefNannySetupContext("__pyx_fuse_1query", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pyx_fuse_1query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3224, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_11__pyx_fuse_1query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = PyFloat_FromDouble(__pyx_v_point); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 3224, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3224, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3224, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 3224, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_5) {
            __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3224, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3224, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_3query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_4 = PyFloat_FromDouble(__pyx_v_point); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 3224, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_7 = __pyx_t_1; __pyx_t_3 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3224, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3224, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 3224, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (__pyx_t_3) {
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_t_4);
          __pyx_t_4 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3224, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pandas/_libs/intervaltree.pxi":3233
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
  __pyx_t_8 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_8) {

    /* "pandas/_libs/intervaltree.pxi":3237
 *             # continue the binary tree structure. Instead, we use linear
 *             # search.
 *             for i in range(self.n_elements):             # <<<<<<<<<<<<<<
 *                 if self.left[i] < point <= self.right[i]:
 *                     result.append(self.indices[i])
 */
    __pyx_t_9 = __pyx_v_self->n_elements;
    __pyx_t_10 = __pyx_t_9;
    for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
      __pyx_v_i = __pyx_t_11;

      /* "pandas/_libs/intervaltree.pxi":3238
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point <= self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      __pyx_t_12 = __pyx_v_i;
      __pyx_t_8 = ((*((__pyx_t_5numpy_uint64_t *) ( /* dim=0 */ (__pyx_v_self->left.data + __pyx_t_12 * __pyx_v_self->left.strides[0]) ))) < __pyx_v_point);
      if (__pyx_t_8) {
        __pyx_t_13 = __pyx_v_i;
        __pyx_t_8 = (__pyx_v_point <= (*((__pyx_t_5numpy_uint64_t *) ( /* dim=0 */ (__pyx_v_self->right.data + __pyx_t_13 * __pyx_v_self->right.strides[0]) ))));
      }
      __pyx_t_14 = (__pyx_t_8 != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":3239
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point <= self.right[i]:
 *                     result.append(self.indices[i])             # <<<<<<<<<<<<<<
 *         else:
 *             # There are child nodes. Based on comparing our query to the pivot,
 */
        __pyx_t_15 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->indices.data + __pyx_t_15 * __pyx_v_self->indices.strides[0]) ))));

        /* "pandas/_libs/intervaltree.pxi":3238
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point <= self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      }
    }

    /* "pandas/_libs/intervaltree.pxi":3233
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
    goto __pyx_L3;
  }

  /* "pandas/_libs/intervaltree.pxi":3243
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
  /*else*/ {
    __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":3244
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:
 *                 values = self.center_left_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 */
      __pyx_t_16 = __pyx_v_self->center_left_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":3245
 *             if point < self.pivot:
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:
 */
      __pyx_t_17 = __pyx_v_self->center_left_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":3246
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):             # <<<<<<<<<<<<<<
 *                     if not values[i] < point:
 *                         break
 */
      __pyx_t_9 = __pyx_v_self->n_center;
      __pyx_t_10 = __pyx_t_9;
      for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":3247
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_18 = __pyx_v_i;
        __pyx_t_14 = ((!(((*((__pyx_t_5numpy_uint64_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_18 * __pyx_v_values.strides[0]) ))) < __pyx_v_point) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":3248
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:
 */
          goto __pyx_L9_break;

          /* "pandas/_libs/intervaltree.pxi":3247
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":3249
 *                     if not values[i] < point:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 */
        __pyx_t_19 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_19 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L9_break:;

      /* "pandas/_libs/intervaltree.pxi":3250
 *                         break
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      __pyx_t_14 = ((__pyx_v_point <= __pyx_v_self->left_node->max_right) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":3251
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)             # <<<<<<<<<<<<<<
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode *)__pyx_v_self->left_node->__pyx_vtab)->__pyx_fuse_1query(__pyx_v_self->left_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3251, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":3250
 *                         break
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":3243
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":3252
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
    __pyx_t_14 = ((__pyx_v_point > __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":3253
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 *                 values = self.center_right_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 */
      __pyx_t_16 = __pyx_v_self->center_right_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":3254
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:
 */
      __pyx_t_17 = __pyx_v_self->center_right_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":3255
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):             # <<<<<<<<<<<<<<
 *                     if not point <= values[i]:
 *                         break
 */
      for (__pyx_t_11 = (__pyx_v_self->n_center - 1); __pyx_t_11 > -1L; __pyx_t_11-=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":3256
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_20 = __pyx_v_i;
        __pyx_t_14 = ((!((__pyx_v_point <= (*((__pyx_t_5numpy_uint64_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_20 * __pyx_v_values.strides[0]) )))) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":3257
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:
 */
          goto __pyx_L13_break;

          /* "pandas/_libs/intervaltree.pxi":3256
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":3258
 *                     if not point <= values[i]:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if self.right_node.min_left < point:
 *                     self.right_node.query(result, point)
 */
        __pyx_t_21 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_21 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L13_break:;

      /* "pandas/_libs/intervaltree.pxi":3259
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      __pyx_t_14 = ((__pyx_v_self->right_node->min_left < __pyx_v_point) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":3260
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:
 *                     self.right_node.query(result, point)             # <<<<<<<<<<<<<<
 *             else:
 *                 result.extend(self.center_left_indices)
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode *)__pyx_v_self->right_node->__pyx_vtab)->__pyx_fuse_1query(__pyx_v_self->right_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3260, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":3259
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":3252
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":3262
 *                     self.right_node.query(result, point)
 *             else:
 *                 result.extend(self.center_left_indices)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
    /*else*/ {
      __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->extend(__pyx_v_result, __pyx_v_self->center_left_indices); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3262, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_L7:;
  }
  __pyx_L3:;

  /* "pandas/_libs/intervaltree.pxi":3224
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __PYX_XDEC_MEMVIEW(&__pyx_t_16, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_17, 1);
  __Pyx_AddTraceback("pandas._libs.interval.Uint64ClosedRightIntervalNode.query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_indices, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_values, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_11__pyx_fuse_1query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_1__pyx_mdef_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_11__pyx_fuse_1query = {"__pyx_fuse_1query", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_11__pyx_fuse_1query, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_2query};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_11__pyx_fuse_1query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result = 0;
  __pyx_t_5numpy_float32_t __pyx_v_point;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fuse_1query (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_result,&__pyx_n_s_point,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_result)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_point)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_1query", 1, 2, 2, 1); __PYX_ERR(1, 3224, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_1query") < 0)) __PYX_ERR(1, 3224, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_result = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)values[0]);
    __pyx_v_point = __pyx_PyFloat_AsFloat(values[1]); if (unlikely((__pyx_v_point == ((npy_float32)-1)) && PyErr_Occurred())) __PYX_ERR(1, 3224, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_1query", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 3224, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Uint64ClosedRightIntervalNode.__pyx_fuse_1query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_result), __pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector, 1, "result", 0))) __PYX_ERR(1, 3224, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_10__pyx_fuse_1query(((struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode *)__pyx_v_self), __pyx_v_result, __pyx_v_point);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_10__pyx_fuse_1query(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float32_t __pyx_v_point) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__pyx_fuse_1query", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_fuse_1__pyx_f_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_query(__pyx_v_self, __pyx_v_result, __pyx_v_point, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3224, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Uint64ClosedRightIntervalNode.__pyx_fuse_1query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_13__pyx_fuse_2query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_fuse_2__pyx_f_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int64_t __pyx_v_point, int __pyx_skip_dispatch) {
  __Pyx_memviewslice __pyx_v_indices = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_values = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __pyx_t_5numpy_int64_t __pyx_t_9;
  __pyx_t_5numpy_int64_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  int __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  __Pyx_memviewslice __pyx_t_16 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_17 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  Py_ssize_t __pyx_t_21;
  __Pyx_RefNannySetupContext("__pyx_fuse_2query", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pyx_fuse_2query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3224, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_13__pyx_fuse_2query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = __Pyx_PyInt_From_npy_int64(__pyx_v_point); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 3224, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3224, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3224, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 3224, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_5) {
            __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3224, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3224, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_3query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_4 = __Pyx_PyInt_From_npy_int64(__pyx_v_point); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 3224, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_7 = __pyx_t_1; __pyx_t_3 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3224, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3224, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 3224, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (__pyx_t_3) {
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_t_4);
          __pyx_t_4 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3224, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pandas/_libs/intervaltree.pxi":3233
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
  __pyx_t_8 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_8) {

    /* "pandas/_libs/intervaltree.pxi":3237
 *             # continue the binary tree structure. Instead, we use linear
 *             # search.
 *             for i in range(self.n_elements):             # <<<<<<<<<<<<<<
 *                 if self.left[i] < point <= self.right[i]:
 *                     result.append(self.indices[i])
 */
    __pyx_t_9 = __pyx_v_self->n_elements;
    __pyx_t_10 = __pyx_t_9;
    for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
      __pyx_v_i = __pyx_t_11;

      /* "pandas/_libs/intervaltree.pxi":3238
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point <= self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      __pyx_t_12 = __pyx_v_i;
      __pyx_t_8 = ((*((__pyx_t_5numpy_uint64_t *) ( /* dim=0 */ (__pyx_v_self->left.data + __pyx_t_12 * __pyx_v_self->left.strides[0]) ))) < __pyx_v_point);
      if (__pyx_t_8) {
        __pyx_t_13 = __pyx_v_i;
        __pyx_t_8 = (__pyx_v_point <= (*((__pyx_t_5numpy_uint64_t *) ( /* dim=0 */ (__pyx_v_self->right.data + __pyx_t_13 * __pyx_v_self->right.strides[0]) ))));
      }
      __pyx_t_14 = (__pyx_t_8 != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":3239
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point <= self.right[i]:
 *                     result.append(self.indices[i])             # <<<<<<<<<<<<<<
 *         else:
 *             # There are child nodes. Based on comparing our query to the pivot,
 */
        __pyx_t_15 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->indices.data + __pyx_t_15 * __pyx_v_self->indices.strides[0]) ))));

        /* "pandas/_libs/intervaltree.pxi":3238
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point <= self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      }
    }

    /* "pandas/_libs/intervaltree.pxi":3233
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
    goto __pyx_L3;
  }

  /* "pandas/_libs/intervaltree.pxi":3243
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
  /*else*/ {
    __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":3244
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:
 *                 values = self.center_left_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 */
      __pyx_t_16 = __pyx_v_self->center_left_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":3245
 *             if point < self.pivot:
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:
 */
      __pyx_t_17 = __pyx_v_self->center_left_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":3246
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):             # <<<<<<<<<<<<<<
 *                     if not values[i] < point:
 *                         break
 */
      __pyx_t_9 = __pyx_v_self->n_center;
      __pyx_t_10 = __pyx_t_9;
      for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":3247
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_18 = __pyx_v_i;
        __pyx_t_14 = ((!(((*((__pyx_t_5numpy_uint64_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_18 * __pyx_v_values.strides[0]) ))) < __pyx_v_point) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":3248
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:
 */
          goto __pyx_L9_break;

          /* "pandas/_libs/intervaltree.pxi":3247
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":3249
 *                     if not values[i] < point:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 */
        __pyx_t_19 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_19 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L9_break:;

      /* "pandas/_libs/intervaltree.pxi":3250
 *                         break
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      __pyx_t_14 = ((__pyx_v_point <= __pyx_v_self->left_node->max_right) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":3251
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)             # <<<<<<<<<<<<<<
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode *)__pyx_v_self->left_node->__pyx_vtab)->__pyx_fuse_2query(__pyx_v_self->left_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3251, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":3250
 *                         break
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":3243
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":3252
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
    __pyx_t_14 = ((__pyx_v_point > __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":3253
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 *                 values = self.center_right_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 */
      __pyx_t_16 = __pyx_v_self->center_right_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":3254
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:
 */
      __pyx_t_17 = __pyx_v_self->center_right_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":3255
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):             # <<<<<<<<<<<<<<
 *                     if not point <= values[i]:
 *                         break
 */
      for (__pyx_t_11 = (__pyx_v_self->n_center - 1); __pyx_t_11 > -1L; __pyx_t_11-=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":3256
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_20 = __pyx_v_i;
        __pyx_t_14 = ((!((__pyx_v_point <= (*((__pyx_t_5numpy_uint64_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_20 * __pyx_v_values.strides[0]) )))) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":3257
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:
 */
          goto __pyx_L13_break;

          /* "pandas/_libs/intervaltree.pxi":3256
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":3258
 *                     if not point <= values[i]:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if self.right_node.min_left < point:
 *                     self.right_node.query(result, point)
 */
        __pyx_t_21 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_21 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L13_break:;

      /* "pandas/_libs/intervaltree.pxi":3259
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      __pyx_t_14 = ((__pyx_v_self->right_node->min_left < __pyx_v_point) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":3260
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:
 *                     self.right_node.query(result, point)             # <<<<<<<<<<<<<<
 *             else:
 *                 result.extend(self.center_left_indices)
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode *)__pyx_v_self->right_node->__pyx_vtab)->__pyx_fuse_2query(__pyx_v_self->right_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3260, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":3259
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":3252
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":3262
 *                     self.right_node.query(result, point)
 *             else:
 *                 result.extend(self.center_left_indices)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
    /*else*/ {
      __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->extend(__pyx_v_result, __pyx_v_self->center_left_indices); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3262, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_L7:;
  }
  __pyx_L3:;

  /* "pandas/_libs/intervaltree.pxi":3224
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __PYX_XDEC_MEMVIEW(&__pyx_t_16, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_17, 1);
  __Pyx_AddTraceback("pandas._libs.interval.Uint64ClosedRightIntervalNode.query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_indices, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_values, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_13__pyx_fuse_2query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_2__pyx_mdef_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_13__pyx_fuse_2query = {"__pyx_fuse_2query", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_13__pyx_fuse_2query, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_2query};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_13__pyx_fuse_2query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result = 0;
  __pyx_t_5numpy_int64_t __pyx_v_point;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fuse_2query (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_result,&__pyx_n_s_point,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_result)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_point)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_2query", 1, 2, 2, 1); __PYX_ERR(1, 3224, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_2query") < 0)) __PYX_ERR(1, 3224, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_result = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)values[0]);
    __pyx_v_point = __Pyx_PyInt_As_npy_int64(values[1]); if (unlikely((__pyx_v_point == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(1, 3224, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_2query", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 3224, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Uint64ClosedRightIntervalNode.__pyx_fuse_2query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_result), __pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector, 1, "result", 0))) __PYX_ERR(1, 3224, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_12__pyx_fuse_2query(((struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode *)__pyx_v_self), __pyx_v_result, __pyx_v_point);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_12__pyx_fuse_2query(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int64_t __pyx_v_point) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__pyx_fuse_2query", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_fuse_2__pyx_f_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_query(__pyx_v_self, __pyx_v_result, __pyx_v_point, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3224, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Uint64ClosedRightIntervalNode.__pyx_fuse_2query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_15__pyx_fuse_3query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_fuse_3__pyx_f_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int32_t __pyx_v_point, int __pyx_skip_dispatch) {
  __Pyx_memviewslice __pyx_v_indices = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_values = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __pyx_t_5numpy_int64_t __pyx_t_9;
  __pyx_t_5numpy_int64_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  int __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  __Pyx_memviewslice __pyx_t_16 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_17 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  Py_ssize_t __pyx_t_21;
  __Pyx_RefNannySetupContext("__pyx_fuse_3query", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pyx_fuse_3query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3224, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_15__pyx_fuse_3query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = __Pyx_PyInt_From_npy_int32(__pyx_v_point); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 3224, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3224, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3224, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 3224, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_5) {
            __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3224, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3224, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_3query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_4 = __Pyx_PyInt_From_npy_int32(__pyx_v_point); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 3224, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_7 = __pyx_t_1; __pyx_t_3 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3224, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3224, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 3224, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (__pyx_t_3) {
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_t_4);
          __pyx_t_4 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3224, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pandas/_libs/intervaltree.pxi":3233
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
  __pyx_t_8 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_8) {

    /* "pandas/_libs/intervaltree.pxi":3237
 *             # continue the binary tree structure. Instead, we use linear
 *             # search.
 *             for i in range(self.n_elements):             # <<<<<<<<<<<<<<
 *                 if self.left[i] < point <= self.right[i]:
 *                     result.append(self.indices[i])
 */
    __pyx_t_9 = __pyx_v_self->n_elements;
    __pyx_t_10 = __pyx_t_9;
    for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
      __pyx_v_i = __pyx_t_11;

      /* "pandas/_libs/intervaltree.pxi":3238
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point <= self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      __pyx_t_12 = __pyx_v_i;
      __pyx_t_8 = ((*((__pyx_t_5numpy_uint64_t *) ( /* dim=0 */ (__pyx_v_self->left.data + __pyx_t_12 * __pyx_v_self->left.strides[0]) ))) < __pyx_v_point);
      if (__pyx_t_8) {
        __pyx_t_13 = __pyx_v_i;
        __pyx_t_8 = (__pyx_v_point <= (*((__pyx_t_5numpy_uint64_t *) ( /* dim=0 */ (__pyx_v_self->right.data + __pyx_t_13 * __pyx_v_self->right.strides[0]) ))));
      }
      __pyx_t_14 = (__pyx_t_8 != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":3239
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point <= self.right[i]:
 *                     result.append(self.indices[i])             # <<<<<<<<<<<<<<
 *         else:
 *             # There are child nodes. Based on comparing our query to the pivot,
 */
        __pyx_t_15 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->indices.data + __pyx_t_15 * __pyx_v_self->indices.strides[0]) ))));

        /* "pandas/_libs/intervaltree.pxi":3238
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point <= self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      }
    }

    /* "pandas/_libs/intervaltree.pxi":3233
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
    goto __pyx_L3;
  }

  /* "pandas/_libs/intervaltree.pxi":3243
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
  /*else*/ {
    __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":3244
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:
 *                 values = self.center_left_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 */
      __pyx_t_16 = __pyx_v_self->center_left_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":3245
 *             if point < self.pivot:
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:
 */
      __pyx_t_17 = __pyx_v_self->center_left_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":3246
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):             # <<<<<<<<<<<<<<
 *                     if not values[i] < point:
 *                         break
 */
      __pyx_t_9 = __pyx_v_self->n_center;
      __pyx_t_10 = __pyx_t_9;
      for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":3247
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_18 = __pyx_v_i;
        __pyx_t_14 = ((!(((*((__pyx_t_5numpy_uint64_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_18 * __pyx_v_values.strides[0]) ))) < __pyx_v_point) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":3248
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:
 */
          goto __pyx_L9_break;

          /* "pandas/_libs/intervaltree.pxi":3247
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":3249
 *                     if not values[i] < point:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 */
        __pyx_t_19 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_19 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L9_break:;

      /* "pandas/_libs/intervaltree.pxi":3250
 *                         break
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      __pyx_t_14 = ((__pyx_v_point <= __pyx_v_self->left_node->max_right) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":3251
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)             # <<<<<<<<<<<<<<
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode *)__pyx_v_self->left_node->__pyx_vtab)->__pyx_fuse_3query(__pyx_v_self->left_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3251, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":3250
 *                         break
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":3243
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":3252
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
    __pyx_t_14 = ((__pyx_v_point > __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":3253
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 *                 values = self.center_right_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 */
      __pyx_t_16 = __pyx_v_self->center_right_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":3254
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:
 */
      __pyx_t_17 = __pyx_v_self->center_right_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":3255
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):             # <<<<<<<<<<<<<<
 *                     if not point <= values[i]:
 *                         break
 */
      for (__pyx_t_11 = (__pyx_v_self->n_center - 1); __pyx_t_11 > -1L; __pyx_t_11-=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":3256
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_20 = __pyx_v_i;
        __pyx_t_14 = ((!((__pyx_v_point <= (*((__pyx_t_5numpy_uint64_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_20 * __pyx_v_values.strides[0]) )))) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":3257
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:
 */
          goto __pyx_L13_break;

          /* "pandas/_libs/intervaltree.pxi":3256
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":3258
 *                     if not point <= values[i]:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if self.right_node.min_left < point:
 *                     self.right_node.query(result, point)
 */
        __pyx_t_21 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_21 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L13_break:;

      /* "pandas/_libs/intervaltree.pxi":3259
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      __pyx_t_14 = ((__pyx_v_self->right_node->min_left < __pyx_v_point) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":3260
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:
 *                     self.right_node.query(result, point)             # <<<<<<<<<<<<<<
 *             else:
 *                 result.extend(self.center_left_indices)
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode *)__pyx_v_self->right_node->__pyx_vtab)->__pyx_fuse_3query(__pyx_v_self->right_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3260, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":3259
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":3252
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":3262
 *                     self.right_node.query(result, point)
 *             else:
 *                 result.extend(self.center_left_indices)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
    /*else*/ {
      __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->extend(__pyx_v_result, __pyx_v_self->center_left_indices); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3262, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_L7:;
  }
  __pyx_L3:;

  /* "pandas/_libs/intervaltree.pxi":3224
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __PYX_XDEC_MEMVIEW(&__pyx_t_16, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_17, 1);
  __Pyx_AddTraceback("pandas._libs.interval.Uint64ClosedRightIntervalNode.query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_indices, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_values, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_15__pyx_fuse_3query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_3__pyx_mdef_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_15__pyx_fuse_3query = {"__pyx_fuse_3query", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_15__pyx_fuse_3query, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_2query};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_15__pyx_fuse_3query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result = 0;
  __pyx_t_5numpy_int32_t __pyx_v_point;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fuse_3query (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_result,&__pyx_n_s_point,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_result)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_point)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_3query", 1, 2, 2, 1); __PYX_ERR(1, 3224, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_3query") < 0)) __PYX_ERR(1, 3224, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_result = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)values[0]);
    __pyx_v_point = __Pyx_PyInt_As_npy_int32(values[1]); if (unlikely((__pyx_v_point == ((npy_int32)-1)) && PyErr_Occurred())) __PYX_ERR(1, 3224, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_3query", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 3224, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Uint64ClosedRightIntervalNode.__pyx_fuse_3query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_result), __pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector, 1, "result", 0))) __PYX_ERR(1, 3224, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_14__pyx_fuse_3query(((struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode *)__pyx_v_self), __pyx_v_result, __pyx_v_point);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_14__pyx_fuse_3query(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int32_t __pyx_v_point) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__pyx_fuse_3query", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_fuse_3__pyx_f_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_query(__pyx_v_self, __pyx_v_result, __pyx_v_point, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3224, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Uint64ClosedRightIntervalNode.__pyx_fuse_3query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_17__pyx_fuse_4query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_fuse_4__pyx_f_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_uint64_t __pyx_v_point, int __pyx_skip_dispatch) {
  __Pyx_memviewslice __pyx_v_indices = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_values = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __pyx_t_5numpy_int64_t __pyx_t_9;
  __pyx_t_5numpy_int64_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  int __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  __Pyx_memviewslice __pyx_t_16 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_17 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  Py_ssize_t __pyx_t_21;
  __Pyx_RefNannySetupContext("__pyx_fuse_4query", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pyx_fuse_4query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3224, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_17__pyx_fuse_4query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = __Pyx_PyInt_From_npy_uint64(__pyx_v_point); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 3224, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3224, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3224, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 3224, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_5) {
            __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3224, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3224, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_3query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_4 = __Pyx_PyInt_From_npy_uint64(__pyx_v_point); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 3224, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_7 = __pyx_t_1; __pyx_t_3 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3224, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3224, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 3224, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (__pyx_t_3) {
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_t_4);
          __pyx_t_4 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3224, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pandas/_libs/intervaltree.pxi":3233
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
  __pyx_t_8 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_8) {

    /* "pandas/_libs/intervaltree.pxi":3237
 *             # continue the binary tree structure. Instead, we use linear
 *             # search.
 *             for i in range(self.n_elements):             # <<<<<<<<<<<<<<
 *                 if self.left[i] < point <= self.right[i]:
 *                     result.append(self.indices[i])
 */
    __pyx_t_9 = __pyx_v_self->n_elements;
    __pyx_t_10 = __pyx_t_9;
    for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
      __pyx_v_i = __pyx_t_11;

      /* "pandas/_libs/intervaltree.pxi":3238
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point <= self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      __pyx_t_12 = __pyx_v_i;
      __pyx_t_8 = ((*((__pyx_t_5numpy_uint64_t *) ( /* dim=0 */ (__pyx_v_self->left.data + __pyx_t_12 * __pyx_v_self->left.strides[0]) ))) < __pyx_v_point);
      if (__pyx_t_8) {
        __pyx_t_13 = __pyx_v_i;
        __pyx_t_8 = (__pyx_v_point <= (*((__pyx_t_5numpy_uint64_t *) ( /* dim=0 */ (__pyx_v_self->right.data + __pyx_t_13 * __pyx_v_self->right.strides[0]) ))));
      }
      __pyx_t_14 = (__pyx_t_8 != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":3239
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point <= self.right[i]:
 *                     result.append(self.indices[i])             # <<<<<<<<<<<<<<
 *         else:
 *             # There are child nodes. Based on comparing our query to the pivot,
 */
        __pyx_t_15 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->indices.data + __pyx_t_15 * __pyx_v_self->indices.strides[0]) ))));

        /* "pandas/_libs/intervaltree.pxi":3238
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point <= self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      }
    }

    /* "pandas/_libs/intervaltree.pxi":3233
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
    goto __pyx_L3;
  }

  /* "pandas/_libs/intervaltree.pxi":3243
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
  /*else*/ {
    __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":3244
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:
 *                 values = self.center_left_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 */
      __pyx_t_16 = __pyx_v_self->center_left_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":3245
 *             if point < self.pivot:
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:
 */
      __pyx_t_17 = __pyx_v_self->center_left_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":3246
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):             # <<<<<<<<<<<<<<
 *                     if not values[i] < point:
 *                         break
 */
      __pyx_t_9 = __pyx_v_self->n_center;
      __pyx_t_10 = __pyx_t_9;
      for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":3247
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_18 = __pyx_v_i;
        __pyx_t_14 = ((!(((*((__pyx_t_5numpy_uint64_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_18 * __pyx_v_values.strides[0]) ))) < __pyx_v_point) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":3248
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:
 */
          goto __pyx_L9_break;

          /* "pandas/_libs/intervaltree.pxi":3247
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":3249
 *                     if not values[i] < point:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 */
        __pyx_t_19 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_19 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L9_break:;

      /* "pandas/_libs/intervaltree.pxi":3250
 *                         break
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      __pyx_t_14 = ((__pyx_v_point <= __pyx_v_self->left_node->max_right) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":3251
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)             # <<<<<<<<<<<<<<
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode *)__pyx_v_self->left_node->__pyx_vtab)->__pyx_fuse_4query(__pyx_v_self->left_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3251, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":3250
 *                         break
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":3243
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":3252
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
    __pyx_t_14 = ((__pyx_v_point > __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":3253
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 *                 values = self.center_right_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 */
      __pyx_t_16 = __pyx_v_self->center_right_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":3254
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:
 */
      __pyx_t_17 = __pyx_v_self->center_right_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":3255
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):             # <<<<<<<<<<<<<<
 *                     if not point <= values[i]:
 *                         break
 */
      for (__pyx_t_11 = (__pyx_v_self->n_center - 1); __pyx_t_11 > -1L; __pyx_t_11-=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":3256
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_20 = __pyx_v_i;
        __pyx_t_14 = ((!((__pyx_v_point <= (*((__pyx_t_5numpy_uint64_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_20 * __pyx_v_values.strides[0]) )))) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":3257
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:
 */
          goto __pyx_L13_break;

          /* "pandas/_libs/intervaltree.pxi":3256
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":3258
 *                     if not point <= values[i]:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if self.right_node.min_left < point:
 *                     self.right_node.query(result, point)
 */
        __pyx_t_21 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_21 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L13_break:;

      /* "pandas/_libs/intervaltree.pxi":3259
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      __pyx_t_14 = ((__pyx_v_self->right_node->min_left < __pyx_v_point) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":3260
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:
 *                     self.right_node.query(result, point)             # <<<<<<<<<<<<<<
 *             else:
 *                 result.extend(self.center_left_indices)
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode *)__pyx_v_self->right_node->__pyx_vtab)->__pyx_fuse_4query(__pyx_v_self->right_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3260, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":3259
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":3252
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":3262
 *                     self.right_node.query(result, point)
 *             else:
 *                 result.extend(self.center_left_indices)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
    /*else*/ {
      __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->extend(__pyx_v_result, __pyx_v_self->center_left_indices); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3262, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_L7:;
  }
  __pyx_L3:;

  /* "pandas/_libs/intervaltree.pxi":3224
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __PYX_XDEC_MEMVIEW(&__pyx_t_16, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_17, 1);
  __Pyx_AddTraceback("pandas._libs.interval.Uint64ClosedRightIntervalNode.query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_indices, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_values, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_17__pyx_fuse_4query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_4__pyx_mdef_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_17__pyx_fuse_4query = {"__pyx_fuse_4query", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_17__pyx_fuse_4query, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_2query};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_17__pyx_fuse_4query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result = 0;
  __pyx_t_5numpy_uint64_t __pyx_v_point;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fuse_4query (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_result,&__pyx_n_s_point,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_result)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_point)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_4query", 1, 2, 2, 1); __PYX_ERR(1, 3224, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_4query") < 0)) __PYX_ERR(1, 3224, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_result = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)values[0]);
    __pyx_v_point = __Pyx_PyInt_As_npy_uint64(values[1]); if (unlikely((__pyx_v_point == ((npy_uint64)-1)) && PyErr_Occurred())) __PYX_ERR(1, 3224, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_4query", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 3224, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Uint64ClosedRightIntervalNode.__pyx_fuse_4query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_result), __pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector, 1, "result", 0))) __PYX_ERR(1, 3224, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_16__pyx_fuse_4query(((struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode *)__pyx_v_self), __pyx_v_result, __pyx_v_point);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_16__pyx_fuse_4query(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_uint64_t __pyx_v_point) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__pyx_fuse_4query", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_fuse_4__pyx_f_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_query(__pyx_v_self, __pyx_v_result, __pyx_v_point, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3224, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Uint64ClosedRightIntervalNode.__pyx_fuse_4query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":3264
 *                 result.extend(self.center_left_indices)
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         if self.is_leaf_node:
 *             return ('<Uint64ClosedRightIntervalNode: '
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_5__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_5__repr__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_4__repr__(((struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_4__repr__(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode *__pyx_v_self) {
  PyObject *__pyx_v_n_left = NULL;
  PyObject *__pyx_v_n_right = NULL;
  PyObject *__pyx_v_n_center = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  Py_UCS4 __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "pandas/_libs/intervaltree.pxi":3265
 * 
 *     def __repr__(self):
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             return ('<Uint64ClosedRightIntervalNode: '
 *                     '%s elements (terminal)>' % self.n_elements)
 */
  __pyx_t_1 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_1) {

    /* "pandas/_libs/intervaltree.pxi":3266
 *     def __repr__(self):
 *         if self.is_leaf_node:
 *             return ('<Uint64ClosedRightIntervalNode: '             # <<<<<<<<<<<<<<
 *                     '%s elements (terminal)>' % self.n_elements)
 *         else:
 */
    __Pyx_XDECREF(__pyx_r);

    /* "pandas/_libs/intervaltree.pxi":3267
 *         if self.is_leaf_node:
 *             return ('<Uint64ClosedRightIntervalNode: '
 *                     '%s elements (terminal)>' % self.n_elements)             # <<<<<<<<<<<<<<
 *         else:
 *             n_left = self.left_node.n_elements
 */
    __pyx_t_2 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_elements); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3267, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PyUnicode_Format(__pyx_kp_u_Uint64ClosedRightIntervalNode_s, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 3267, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pandas/_libs/intervaltree.pxi":3265
 * 
 *     def __repr__(self):
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             return ('<Uint64ClosedRightIntervalNode: '
 *                     '%s elements (terminal)>' % self.n_elements)
 */
  }

  /* "pandas/_libs/intervaltree.pxi":3269
 *                     '%s elements (terminal)>' % self.n_elements)
 *         else:
 *             n_left = self.left_node.n_elements             # <<<<<<<<<<<<<<
 *             n_right = self.right_node.n_elements
 *             n_center = self.n_elements - n_left - n_right
 */
  /*else*/ {
    __pyx_t_3 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->left_node->n_elements); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 3269, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_v_n_left = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "pandas/_libs/intervaltree.pxi":3270
 *         else:
 *             n_left = self.left_node.n_elements
 *             n_right = self.right_node.n_elements             # <<<<<<<<<<<<<<
 *             n_center = self.n_elements - n_left - n_right
 *             return ('<Uint64ClosedRightIntervalNode: '
 */
    __pyx_t_3 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->right_node->n_elements); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 3270, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_v_n_right = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "pandas/_libs/intervaltree.pxi":3271
 *             n_left = self.left_node.n_elements
 *             n_right = self.right_node.n_elements
 *             n_center = self.n_elements - n_left - n_right             # <<<<<<<<<<<<<<
 *             return ('<Uint64ClosedRightIntervalNode: '
 *                     'pivot %s, %s elements (%s left, %s right, %s '
 */
    __pyx_t_3 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_elements); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 3271, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = PyNumber_Subtract(__pyx_t_3, __pyx_v_n_left); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3271, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyNumber_Subtract(__pyx_t_2, __pyx_v_n_right); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 3271, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_n_center = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "pandas/_libs/intervaltree.pxi":3272
 *             n_right = self.right_node.n_elements
 *             n_center = self.n_elements - n_left - n_right
 *             return ('<Uint64ClosedRightIntervalNode: '             # <<<<<<<<<<<<<<
 *                     'pivot %s, %s elements (%s left, %s right, %s '
 *                     'overlapping)>' % (self.pivot, self.n_elements,
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = PyTuple_New(11); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 3272, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = 0;
    __pyx_t_5 = 127;
    __Pyx_INCREF(__pyx_kp_u_Uint64ClosedRightIntervalNode_p);
    __pyx_t_4 += 38;
    __Pyx_GIVEREF(__pyx_kp_u_Uint64ClosedRightIntervalNode_p);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_kp_u_Uint64ClosedRightIntervalNode_p);

    /* "pandas/_libs/intervaltree.pxi":3274
 *             return ('<Uint64ClosedRightIntervalNode: '
 *                     'pivot %s, %s elements (%s left, %s right, %s '
 *                     'overlapping)>' % (self.pivot, self.n_elements,             # <<<<<<<<<<<<<<
 *                                        n_left, n_right, n_center))
 * 
 */
    __pyx_t_2 = __Pyx_PyInt_From_npy_uint64(__pyx_v_self->pivot); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3274, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_t_2), __pyx_empty_unicode); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 3274, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_5 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_6) > __pyx_t_5) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_6) : __pyx_t_5;
    __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_6);
    __pyx_t_6 = 0;
    __Pyx_INCREF(__pyx_kp_u__18);
    __pyx_t_4 += 2;
    __Pyx_GIVEREF(__pyx_kp_u__18);
    PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_kp_u__18);
    __pyx_t_6 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_elements); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 3274, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_2 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_t_6), __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3274, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_5 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) > __pyx_t_5) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) : __pyx_t_5;
    __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_3, 3, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_elements);
    __pyx_t_4 += 11;
    __Pyx_GIVEREF(__pyx_kp_u_elements);
    PyTuple_SET_ITEM(__pyx_t_3, 4, __pyx_kp_u_elements);

    /* "pandas/_libs/intervaltree.pxi":3275
 *                     'pivot %s, %s elements (%s left, %s right, %s '
 *                     'overlapping)>' % (self.pivot, self.n_elements,
 *                                        n_left, n_right, n_center))             # <<<<<<<<<<<<<<
 * 
 *     def counts(self):
 */
    __pyx_t_2 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_v_n_left), __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3275, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) > __pyx_t_5) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) : __pyx_t_5;
    __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_3, 5, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_left_2);
    __pyx_t_4 += 7;
    __Pyx_GIVEREF(__pyx_kp_u_left_2);
    PyTuple_SET_ITEM(__pyx_t_3, 6, __pyx_kp_u_left_2);
    __pyx_t_2 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_v_n_right), __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3275, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) > __pyx_t_5) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) : __pyx_t_5;
    __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_3, 7, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_right_2);
    __pyx_t_4 += 8;
    __Pyx_GIVEREF(__pyx_kp_u_right_2);
    PyTuple_SET_ITEM(__pyx_t_3, 8, __pyx_kp_u_right_2);
    __pyx_t_2 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_v_n_center), __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3275, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) > __pyx_t_5) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) : __pyx_t_5;
    __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_3, 9, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_overlapping);
    __pyx_t_4 += 14;
    __Pyx_GIVEREF(__pyx_kp_u_overlapping);
    PyTuple_SET_ITEM(__pyx_t_3, 10, __pyx_kp_u_overlapping);

    /* "pandas/_libs/intervaltree.pxi":3272
 *             n_right = self.right_node.n_elements
 *             n_center = self.n_elements - n_left - n_right
 *             return ('<Uint64ClosedRightIntervalNode: '             # <<<<<<<<<<<<<<
 *                     'pivot %s, %s elements (%s left, %s right, %s '
 *                     'overlapping)>' % (self.pivot, self.n_elements,
 */
    __pyx_t_2 = __Pyx_PyUnicode_Join(__pyx_t_3, 11, __pyx_t_4, __pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3272, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;
  }

  /* "pandas/_libs/intervaltree.pxi":3264
 *                 result.extend(self.center_left_indices)
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         if self.is_leaf_node:
 *             return ('<Uint64ClosedRightIntervalNode: '
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pandas._libs.interval.Uint64ClosedRightIntervalNode.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_n_left);
  __Pyx_XDECREF(__pyx_v_n_right);
  __Pyx_XDECREF(__pyx_v_n_center);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":3277
 *                                        n_left, n_right, n_center))
 * 
 *     def counts(self):             # <<<<<<<<<<<<<<
 *         """
 *         Inspect counts on this node
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_7counts(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_6counts[] = "\n        Inspect counts on this node\n        useful for debugging purposes\n        ";
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_7counts(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("counts (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_6counts(((struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_6counts(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode *__pyx_v_self) {
  Py_ssize_t __pyx_v_m;
  PyObject *__pyx_v_l = NULL;
  PyObject *__pyx_v_r = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  size_t __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("counts", 0);

  /* "pandas/_libs/intervaltree.pxi":3282
 *         useful for debugging purposes
 *         """
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             return self.n_elements
 *         else:
 */
  __pyx_t_1 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_1) {

    /* "pandas/_libs/intervaltree.pxi":3283
 *         """
 *         if self.is_leaf_node:
 *             return self.n_elements             # <<<<<<<<<<<<<<
 *         else:
 *             m = len(self.center_left_values)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_elements); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3283, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "pandas/_libs/intervaltree.pxi":3282
 *         useful for debugging purposes
 *         """
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             return self.n_elements
 *         else:
 */
  }

  /* "pandas/_libs/intervaltree.pxi":3285
 *             return self.n_elements
 *         else:
 *             m = len(self.center_left_values)             # <<<<<<<<<<<<<<
 *             l = self.left_node.counts()
 *             r = self.right_node.counts()
 */
  /*else*/ {
    if (unlikely(!__pyx_v_self->center_left_values.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(1, 3285, __pyx_L1_error)}
    __pyx_t_3 = __Pyx_MemoryView_Len(__pyx_v_self->center_left_values); 
    __pyx_v_m = __pyx_t_3;

    /* "pandas/_libs/intervaltree.pxi":3286
 *         else:
 *             m = len(self.center_left_values)
 *             l = self.left_node.counts()             # <<<<<<<<<<<<<<
 *             r = self.right_node.counts()
 *             return (m, (l, r))
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->left_node), __pyx_n_s_counts); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 3286, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3286, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_l = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "pandas/_libs/intervaltree.pxi":3287
 *             m = len(self.center_left_values)
 *             l = self.left_node.counts()
 *             r = self.right_node.counts()             # <<<<<<<<<<<<<<
 *             return (m, (l, r))
 * 
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->right_node), __pyx_n_s_counts); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 3287, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3287, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_r = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "pandas/_libs/intervaltree.pxi":3288
 *             l = self.left_node.counts()
 *             r = self.right_node.counts()
 *             return (m, (l, r))             # <<<<<<<<<<<<<<
 * 
 * NODE_CLASSES['uint64',
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = PyInt_FromSsize_t(__pyx_v_m); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3288, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 3288, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_v_l);
    __Pyx_GIVEREF(__pyx_v_l);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_l);
    __Pyx_INCREF(__pyx_v_r);
    __Pyx_GIVEREF(__pyx_v_r);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_v_r);
    __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 3288, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_4);
    __pyx_t_2 = 0;
    __pyx_t_4 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;
  }

  /* "pandas/_libs/intervaltree.pxi":3277
 *                                        n_left, n_right, n_center))
 * 
 *     def counts(self):             # <<<<<<<<<<<<<<
 *         """
 *         Inspect counts on this node
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pandas._libs.interval.Uint64ClosedRightIntervalNode.counts", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_l);
  __Pyx_XDECREF(__pyx_v_r);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":3134
 *         int64_t[:] center_left_indices, center_right_indices, indices
 *         uint64_t min_left, max_right
 *         readonly uint64_t pivot             # <<<<<<<<<<<<<<
 *         readonly int64_t n_elements, n_center, leaf_size
 *         readonly bint is_leaf_node
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_5pivot_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_5pivot_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_5pivot___get__(((struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_5pivot___get__(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_npy_uint64(__pyx_v_self->pivot); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3134, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Uint64ClosedRightIntervalNode.pivot.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":3135
 *         uint64_t min_left, max_right
 *         readonly uint64_t pivot
 *         readonly int64_t n_elements, n_center, leaf_size             # <<<<<<<<<<<<<<
 *         readonly bint is_leaf_node
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_10n_elements_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_10n_elements_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_10n_elements___get__(((struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_10n_elements___get__(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_elements); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3135, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Uint64ClosedRightIntervalNode.n_elements.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_8n_center_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_8n_center_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_8n_center___get__(((struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_8n_center___get__(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_center); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3135, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Uint64ClosedRightIntervalNode.n_center.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_9leaf_size_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_9leaf_size_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_9leaf_size___get__(((struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_9leaf_size___get__(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->leaf_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3135, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Uint64ClosedRightIntervalNode.leaf_size.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":3136
 *         readonly uint64_t pivot
 *         readonly int64_t n_elements, n_center, leaf_size
 *         readonly bint is_leaf_node             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self,
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_12is_leaf_node_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_12is_leaf_node_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_12is_leaf_node___get__(((struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_12is_leaf_node___get__(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->is_leaf_node); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3136, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Uint64ClosedRightIntervalNode.is_leaf_node.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_21__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_21__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_20__reduce_cython__(((struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_20__reduce_cython__(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  int __pyx_t_16;
  int __pyx_t_17;
  int __pyx_t_18;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.center_left_indices, self.center_left_values, self.center_right_indices, self.center_right_values, self.indices, self.is_leaf_node, self.leaf_size, self.left, self.left_node, self.max_right, self.min_left, self.n_center, self.n_elements, self.pivot, self.right, self.right_node)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  if (unlikely(!__pyx_v_self->center_left_indices.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_1 = __pyx_memoryview_fromslice(__pyx_v_self->center_left_indices, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_int64_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_int64_t, 0);; if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (unlikely(!__pyx_v_self->center_left_values.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_2 = __pyx_memoryview_fromslice(__pyx_v_self->center_left_values, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_uint64_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_uint64_t, 0);; if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (unlikely(!__pyx_v_self->center_right_indices.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_3 = __pyx_memoryview_fromslice(__pyx_v_self->center_right_indices, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_int64_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_int64_t, 0);; if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely(!__pyx_v_self->center_right_values.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_4 = __pyx_memoryview_fromslice(__pyx_v_self->center_right_values, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_uint64_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_uint64_t, 0);; if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (unlikely(!__pyx_v_self->indices.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_5 = __pyx_memoryview_fromslice(__pyx_v_self->indices, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_int64_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_int64_t, 0);; if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyBool_FromLong(__pyx_v_self->is_leaf_node); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->leaf_size); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (unlikely(!__pyx_v_self->left.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_8 = __pyx_memoryview_fromslice(__pyx_v_self->left, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_uint64_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_uint64_t, 0);; if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_PyInt_From_npy_uint64(__pyx_v_self->max_right); if (unlikely(!__pyx_t_9)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_10 = __Pyx_PyInt_From_npy_uint64(__pyx_v_self->min_left); if (unlikely(!__pyx_t_10)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_11 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_center); if (unlikely(!__pyx_t_11)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_12 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_elements); if (unlikely(!__pyx_t_12)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __pyx_t_13 = __Pyx_PyInt_From_npy_uint64(__pyx_v_self->pivot); if (unlikely(!__pyx_t_13)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  if (unlikely(!__pyx_v_self->right.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_14 = __pyx_memoryview_fromslice(__pyx_v_self->right, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_uint64_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_uint64_t, 0);; if (unlikely(!__pyx_t_14)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __pyx_t_15 = PyTuple_New(16); if (unlikely(!__pyx_t_15)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_15, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_15, 1, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_15, 2, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_15, 3, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_15, 4, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_15, 5, __pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_15, 6, __pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_15, 7, __pyx_t_8);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->left_node));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self->left_node));
  PyTuple_SET_ITEM(__pyx_t_15, 8, ((PyObject *)__pyx_v_self->left_node));
  __Pyx_GIVEREF(__pyx_t_9);
  PyTuple_SET_ITEM(__pyx_t_15, 9, __pyx_t_9);
  __Pyx_GIVEREF(__pyx_t_10);
  PyTuple_SET_ITEM(__pyx_t_15, 10, __pyx_t_10);
  __Pyx_GIVEREF(__pyx_t_11);
  PyTuple_SET_ITEM(__pyx_t_15, 11, __pyx_t_11);
  __Pyx_GIVEREF(__pyx_t_12);
  PyTuple_SET_ITEM(__pyx_t_15, 12, __pyx_t_12);
  __Pyx_GIVEREF(__pyx_t_13);
  PyTuple_SET_ITEM(__pyx_t_15, 13, __pyx_t_13);
  __Pyx_GIVEREF(__pyx_t_14);
  PyTuple_SET_ITEM(__pyx_t_15, 14, __pyx_t_14);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->right_node));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self->right_node));
  PyTuple_SET_ITEM(__pyx_t_15, 15, ((PyObject *)__pyx_v_self->right_node));
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  __pyx_t_6 = 0;
  __pyx_t_7 = 0;
  __pyx_t_8 = 0;
  __pyx_t_9 = 0;
  __pyx_t_10 = 0;
  __pyx_t_11 = 0;
  __pyx_t_12 = 0;
  __pyx_t_13 = 0;
  __pyx_t_14 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_15);
  __pyx_t_15 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.center_left_indices, self.center_left_values, self.center_right_indices, self.center_right_values, self.indices, self.is_leaf_node, self.leaf_size, self.left, self.left_node, self.max_right, self.min_left, self.n_center, self.n_elements, self.pivot, self.right, self.right_node)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_15 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_15)) __PYX_ERR(2, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __pyx_v__dict = __pyx_t_15;
  __pyx_t_15 = 0;

  /* "(tree fragment)":7
 *     state = (self.center_left_indices, self.center_left_values, self.center_right_indices, self.center_right_values, self.indices, self.is_leaf_node, self.leaf_size, self.left, self.left_node, self.max_right, self.min_left, self.n_center, self.n_elements, self.pivot, self.right, self.right_node)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_16 = (__pyx_v__dict != Py_None);
  __pyx_t_17 = (__pyx_t_16 != 0);
  if (__pyx_t_17) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_15 = PyTuple_New(1); if (unlikely(!__pyx_t_15)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    PyTuple_SET_ITEM(__pyx_t_15, 0, __pyx_v__dict);
    __pyx_t_14 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_15); if (unlikely(!__pyx_t_14)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_14));
    __pyx_t_14 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.left_node is not None or self.right_node is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.center_left_indices, self.center_left_values, self.center_right_indices, self.center_right_values, self.indices, self.is_leaf_node, self.leaf_size, self.left, self.left_node, self.max_right, self.min_left, self.n_center, self.n_elements, self.pivot, self.right, self.right_node)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.left_node is not None or self.right_node is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_Uint64ClosedRightIntervalNode, (type(self), 0x7a104e5, None), state
 */
  /*else*/ {
    __pyx_t_16 = (((PyObject *)__pyx_v_self->left_node) != Py_None);
    __pyx_t_18 = (__pyx_t_16 != 0);
    if (!__pyx_t_18) {
    } else {
      __pyx_t_17 = __pyx_t_18;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_18 = (((PyObject *)__pyx_v_self->right_node) != Py_None);
    __pyx_t_16 = (__pyx_t_18 != 0);
    __pyx_t_17 = __pyx_t_16;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_17;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.left_node is not None or self.right_node is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Uint64ClosedRightIntervalNode, (type(self), 0x7a104e5, None), state
 *     else:
 */
  __pyx_t_17 = (__pyx_v_use_setstate != 0);
  if (__pyx_t_17) {

    /* "(tree fragment)":13
 *         use_setstate = self.left_node is not None or self.right_node is not None
 *     if use_setstate:
 *         return __pyx_unpickle_Uint64ClosedRightIntervalNode, (type(self), 0x7a104e5, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_Uint64ClosedRightIntervalNode, (type(self), 0x7a104e5, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_14, __pyx_n_s_pyx_unpickle_Uint64ClosedRight); if (unlikely(!__pyx_t_14)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __pyx_t_15 = PyTuple_New(3); if (unlikely(!__pyx_t_15)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_15, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_127993061);
    __Pyx_GIVEREF(__pyx_int_127993061);
    PyTuple_SET_ITEM(__pyx_t_15, 1, __pyx_int_127993061);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_15, 2, Py_None);
    __pyx_t_13 = PyTuple_New(3); if (unlikely(!__pyx_t_13)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __Pyx_GIVEREF(__pyx_t_14);
    PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_t_14);
    __Pyx_GIVEREF(__pyx_t_15);
    PyTuple_SET_ITEM(__pyx_t_13, 1, __pyx_t_15);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_13, 2, __pyx_v_state);
    __pyx_t_14 = 0;
    __pyx_t_15 = 0;
    __pyx_r = __pyx_t_13;
    __pyx_t_13 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.left_node is not None or self.right_node is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Uint64ClosedRightIntervalNode, (type(self), 0x7a104e5, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_Uint64ClosedRightIntervalNode, (type(self), 0x7a104e5, None), state
 *     else:
 *         return __pyx_unpickle_Uint64ClosedRightIntervalNode, (type(self), 0x7a104e5, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Uint64ClosedRightIntervalNode__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_13, __pyx_n_s_pyx_unpickle_Uint64ClosedRight); if (unlikely(!__pyx_t_13)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __pyx_t_15 = PyTuple_New(3); if (unlikely(!__pyx_t_15)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_15, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_127993061);
    __Pyx_GIVEREF(__pyx_int_127993061);
    PyTuple_SET_ITEM(__pyx_t_15, 1, __pyx_int_127993061);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_15, 2, __pyx_v_state);
    __pyx_t_14 = PyTuple_New(2); if (unlikely(!__pyx_t_14)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_GIVEREF(__pyx_t_13);
    PyTuple_SET_ITEM(__pyx_t_14, 0, __pyx_t_13);
    __Pyx_GIVEREF(__pyx_t_15);
    PyTuple_SET_ITEM(__pyx_t_14, 1, __pyx_t_15);
    __pyx_t_13 = 0;
    __pyx_t_15 = 0;
    __pyx_r = __pyx_t_14;
    __pyx_t_14 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_AddTraceback("pandas._libs.interval.Uint64ClosedRightIntervalNode.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Uint64ClosedRightIntervalNode, (type(self), 0x7a104e5, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Uint64ClosedRightIntervalNode__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_23__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_23__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_22__setstate_cython__(((struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_22__setstate_cython__(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_Uint64ClosedRightIntervalNode, (type(self), 0x7a104e5, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Uint64ClosedRightIntervalNode__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(2, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_6pandas_5_libs_8interval___pyx_unpickle_Uint64ClosedRightIntervalNode__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Uint64ClosedRightIntervalNode, (type(self), 0x7a104e5, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Uint64ClosedRightIntervalNode__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Uint64ClosedRightIntervalNode.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":3308
 *         readonly bint is_leaf_node
 * 
 *     def __init__(self,             # <<<<<<<<<<<<<<
 *                  ndarray[uint64_t, ndim=1] left,
 *                  ndarray[uint64_t, ndim=1] right,
 */

/* Python wrapper */
static int __pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyArrayObject *__pyx_v_left = 0;
  PyArrayObject *__pyx_v_right = 0;
  PyArrayObject *__pyx_v_indices = 0;
  __pyx_t_5numpy_int64_t __pyx_v_leaf_size;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_left,&__pyx_n_s_right,&__pyx_n_s_indices,&__pyx_n_s_leaf_size,0};
    PyObject* values[4] = {0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_left)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_right)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 4, 4, 1); __PYX_ERR(1, 3308, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_indices)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 4, 4, 2); __PYX_ERR(1, 3308, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_leaf_size)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 4, 4, 3); __PYX_ERR(1, 3308, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(1, 3308, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 4) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
    }
    __pyx_v_left = ((PyArrayObject *)values[0]);
    __pyx_v_right = ((PyArrayObject *)values[1]);
    __pyx_v_indices = ((PyArrayObject *)values[2]);
    __pyx_v_leaf_size = __Pyx_PyInt_As_npy_int64(values[3]); if (unlikely((__pyx_v_leaf_size == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(1, 3312, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 4, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 3308, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Uint64ClosedBothIntervalNode.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_left), __pyx_ptype_5numpy_ndarray, 1, "left", 0))) __PYX_ERR(1, 3309, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_right), __pyx_ptype_5numpy_ndarray, 1, "right", 0))) __PYX_ERR(1, 3310, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_indices), __pyx_ptype_5numpy_ndarray, 1, "indices", 0))) __PYX_ERR(1, 3311, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode___init__(((struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode *)__pyx_v_self), __pyx_v_left, __pyx_v_right, __pyx_v_indices, __pyx_v_leaf_size);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode___init__(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode *__pyx_v_self, PyArrayObject *__pyx_v_left, PyArrayObject *__pyx_v_right, PyArrayObject *__pyx_v_indices, __pyx_t_5numpy_int64_t __pyx_v_leaf_size) {
  PyObject *__pyx_v_left_set = NULL;
  PyObject *__pyx_v_right_set = NULL;
  PyObject *__pyx_v_center_set = NULL;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_indices;
  __Pyx_Buffer __pyx_pybuffer_indices;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_left;
  __Pyx_Buffer __pyx_pybuffer_left;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_right;
  __Pyx_Buffer __pyx_pybuffer_right;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __pyx_t_5numpy_uint64_t __pyx_t_6;
  __Pyx_memviewslice __pyx_t_7 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_8 = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  __Pyx_memviewslice __pyx_t_11 = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *(*__pyx_t_12)(PyObject *);
  size_t __pyx_t_13;
  __Pyx_RefNannySetupContext("__init__", 0);
  __pyx_pybuffer_left.pybuffer.buf = NULL;
  __pyx_pybuffer_left.refcount = 0;
  __pyx_pybuffernd_left.data = NULL;
  __pyx_pybuffernd_left.rcbuffer = &__pyx_pybuffer_left;
  __pyx_pybuffer_right.pybuffer.buf = NULL;
  __pyx_pybuffer_right.refcount = 0;
  __pyx_pybuffernd_right.data = NULL;
  __pyx_pybuffernd_right.rcbuffer = &__pyx_pybuffer_right;
  __pyx_pybuffer_indices.pybuffer.buf = NULL;
  __pyx_pybuffer_indices.refcount = 0;
  __pyx_pybuffernd_indices.data = NULL;
  __pyx_pybuffernd_indices.rcbuffer = &__pyx_pybuffer_indices;
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_left.rcbuffer->pybuffer, (PyObject*)__pyx_v_left, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 3308, __pyx_L1_error)
  }
  __pyx_pybuffernd_left.diminfo[0].strides = __pyx_pybuffernd_left.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_left.diminfo[0].shape = __pyx_pybuffernd_left.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_right.rcbuffer->pybuffer, (PyObject*)__pyx_v_right, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 3308, __pyx_L1_error)
  }
  __pyx_pybuffernd_right.diminfo[0].strides = __pyx_pybuffernd_right.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_right.diminfo[0].shape = __pyx_pybuffernd_right.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_indices.rcbuffer->pybuffer, (PyObject*)__pyx_v_indices, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 3308, __pyx_L1_error)
  }
  __pyx_pybuffernd_indices.diminfo[0].strides = __pyx_pybuffernd_indices.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_indices.diminfo[0].shape = __pyx_pybuffernd_indices.rcbuffer->pybuffer.shape[0];

  /* "pandas/_libs/intervaltree.pxi":3314
 *                  int64_t leaf_size):
 * 
 *         self.n_elements = len(left)             # <<<<<<<<<<<<<<
 *         self.leaf_size = leaf_size
 * 
 */
  __pyx_t_1 = PyObject_Length(((PyObject *)__pyx_v_left)); if (unlikely(__pyx_t_1 == ((Py_ssize_t)-1))) __PYX_ERR(1, 3314, __pyx_L1_error)
  __pyx_v_self->n_elements = __pyx_t_1;

  /* "pandas/_libs/intervaltree.pxi":3315
 * 
 *         self.n_elements = len(left)
 *         self.leaf_size = leaf_size             # <<<<<<<<<<<<<<
 * 
 *         # min_left and min_right are used to speed-up query by skipping
 */
  __pyx_v_self->leaf_size = __pyx_v_leaf_size;

  /* "pandas/_libs/intervaltree.pxi":3320
 *         # query on sub-nodes. If this node has size 0, query is cheap,
 *         # so these values don't matter.
 *         if left.size > 0:             # <<<<<<<<<<<<<<
 *             self.min_left = left.min()
 *             self.max_right = right.max()
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_left), __pyx_n_s_size); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3320, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyObject_RichCompare(__pyx_t_2, __pyx_int_0, Py_GT); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 3320, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(1, 3320, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_4) {

    /* "pandas/_libs/intervaltree.pxi":3321
 *         # so these values don't matter.
 *         if left.size > 0:
 *             self.min_left = left.min()             # <<<<<<<<<<<<<<
 *             self.max_right = right.max()
 *         else:
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_left), __pyx_n_s_min); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3321, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 3321, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_6 = __Pyx_PyInt_As_npy_uint64(__pyx_t_3); if (unlikely((__pyx_t_6 == ((npy_uint64)-1)) && PyErr_Occurred())) __PYX_ERR(1, 3321, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_self->min_left = __pyx_t_6;

    /* "pandas/_libs/intervaltree.pxi":3322
 *         if left.size > 0:
 *             self.min_left = left.min()
 *             self.max_right = right.max()             # <<<<<<<<<<<<<<
 *         else:
 *             self.min_left = 0
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_right), __pyx_n_s_max); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3322, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 3322, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_6 = __Pyx_PyInt_As_npy_uint64(__pyx_t_3); if (unlikely((__pyx_t_6 == ((npy_uint64)-1)) && PyErr_Occurred())) __PYX_ERR(1, 3322, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_self->max_right = __pyx_t_6;

    /* "pandas/_libs/intervaltree.pxi":3320
 *         # query on sub-nodes. If this node has size 0, query is cheap,
 *         # so these values don't matter.
 *         if left.size > 0:             # <<<<<<<<<<<<<<
 *             self.min_left = left.min()
 *             self.max_right = right.max()
 */
    goto __pyx_L3;
  }

  /* "pandas/_libs/intervaltree.pxi":3324
 *             self.max_right = right.max()
 *         else:
 *             self.min_left = 0             # <<<<<<<<<<<<<<
 *             self.max_right = 0
 * 
 */
  /*else*/ {
    __pyx_v_self->min_left = 0;

    /* "pandas/_libs/intervaltree.pxi":3325
 *         else:
 *             self.min_left = 0
 *             self.max_right = 0             # <<<<<<<<<<<<<<
 * 
 *         if self.n_elements <= leaf_size:
 */
    __pyx_v_self->max_right = 0;
  }
  __pyx_L3:;

  /* "pandas/_libs/intervaltree.pxi":3327
 *             self.max_right = 0
 * 
 *         if self.n_elements <= leaf_size:             # <<<<<<<<<<<<<<
 *             # make this a terminal (leaf) node
 *             self.is_leaf_node = True
 */
  __pyx_t_4 = ((__pyx_v_self->n_elements <= __pyx_v_leaf_size) != 0);
  if (__pyx_t_4) {

    /* "pandas/_libs/intervaltree.pxi":3329
 *         if self.n_elements <= leaf_size:
 *             # make this a terminal (leaf) node
 *             self.is_leaf_node = True             # <<<<<<<<<<<<<<
 *             self.left = left
 *             self.right = right
 */
    __pyx_v_self->is_leaf_node = 1;

    /* "pandas/_libs/intervaltree.pxi":3330
 *             # make this a terminal (leaf) node
 *             self.is_leaf_node = True
 *             self.left = left             # <<<<<<<<<<<<<<
 *             self.right = right
 *             self.indices = indices
 */
    __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_uint64_t(((PyObject *)__pyx_v_left), PyBUF_WRITABLE); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(1, 3330, __pyx_L1_error)
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->left, 0);
    __pyx_v_self->left = __pyx_t_7;
    __pyx_t_7.memview = NULL;
    __pyx_t_7.data = NULL;

    /* "pandas/_libs/intervaltree.pxi":3331
 *             self.is_leaf_node = True
 *             self.left = left
 *             self.right = right             # <<<<<<<<<<<<<<
 *             self.indices = indices
 *             self.n_center = 0
 */
    __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_uint64_t(((PyObject *)__pyx_v_right), PyBUF_WRITABLE); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(1, 3331, __pyx_L1_error)
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->right, 0);
    __pyx_v_self->right = __pyx_t_7;
    __pyx_t_7.memview = NULL;
    __pyx_t_7.data = NULL;

    /* "pandas/_libs/intervaltree.pxi":3332
 *             self.left = left
 *             self.right = right
 *             self.indices = indices             # <<<<<<<<<<<<<<
 *             self.n_center = 0
 *         else:
 */
    __pyx_t_8 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(((PyObject *)__pyx_v_indices), PyBUF_WRITABLE); if (unlikely(!__pyx_t_8.memview)) __PYX_ERR(1, 3332, __pyx_L1_error)
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->indices, 0);
    __pyx_v_self->indices = __pyx_t_8;
    __pyx_t_8.memview = NULL;
    __pyx_t_8.data = NULL;

    /* "pandas/_libs/intervaltree.pxi":3333
 *             self.right = right
 *             self.indices = indices
 *             self.n_center = 0             # <<<<<<<<<<<<<<
 *         else:
 *             # calculate a pivot so we can create child nodes
 */
    __pyx_v_self->n_center = 0;

    /* "pandas/_libs/intervaltree.pxi":3327
 *             self.max_right = 0
 * 
 *         if self.n_elements <= leaf_size:             # <<<<<<<<<<<<<<
 *             # make this a terminal (leaf) node
 *             self.is_leaf_node = True
 */
    goto __pyx_L4;
  }

  /* "pandas/_libs/intervaltree.pxi":3336
 *         else:
 *             # calculate a pivot so we can create child nodes
 *             self.is_leaf_node = False             # <<<<<<<<<<<<<<
 *             self.pivot = np.median(left / 2 + right / 2)
 *             left_set, right_set, center_set = self.classify_intervals(
 */
  /*else*/ {
    __pyx_v_self->is_leaf_node = 0;

    /* "pandas/_libs/intervaltree.pxi":3337
 *             # calculate a pivot so we can create child nodes
 *             self.is_leaf_node = False
 *             self.pivot = np.median(left / 2 + right / 2)             # <<<<<<<<<<<<<<
 *             left_set, right_set, center_set = self.classify_intervals(
 *                 left, right)
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3337, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_median); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 3337, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyNumber_Divide(((PyObject *)__pyx_v_left), __pyx_int_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3337, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_9 = __Pyx_PyNumber_Divide(((PyObject *)__pyx_v_right), __pyx_int_2); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 3337, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_10 = PyNumber_Add(__pyx_t_2, __pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 3337, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_9 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_3 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_9, __pyx_t_10) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_10);
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 3337, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_6 = __Pyx_PyInt_As_npy_uint64(__pyx_t_3); if (unlikely((__pyx_t_6 == ((npy_uint64)-1)) && PyErr_Occurred())) __PYX_ERR(1, 3337, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_self->pivot = __pyx_t_6;

    /* "pandas/_libs/intervaltree.pxi":3339
 *             self.pivot = np.median(left / 2 + right / 2)
 *             left_set, right_set, center_set = self.classify_intervals(
 *                 left, right)             # <<<<<<<<<<<<<<
 * 
 *             self.left_node = self.new_child_node(left, right,
 */
    __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_uint64_t(((PyObject *)__pyx_v_left), PyBUF_WRITABLE); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(1, 3339, __pyx_L1_error)
    __pyx_t_11 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_uint64_t(((PyObject *)__pyx_v_right), PyBUF_WRITABLE); if (unlikely(!__pyx_t_11.memview)) __PYX_ERR(1, 3339, __pyx_L1_error)

    /* "pandas/_libs/intervaltree.pxi":3338
 *             self.is_leaf_node = False
 *             self.pivot = np.median(left / 2 + right / 2)
 *             left_set, right_set, center_set = self.classify_intervals(             # <<<<<<<<<<<<<<
 *                 left, right)
 * 
 */
    __pyx_t_3 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode *)__pyx_v_self->__pyx_vtab)->classify_intervals(__pyx_v_self, __pyx_t_7, __pyx_t_11); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 3338, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __PYX_XDEC_MEMVIEW(&__pyx_t_7, 1);
    __pyx_t_7.memview = NULL;
    __pyx_t_7.data = NULL;
    __PYX_XDEC_MEMVIEW(&__pyx_t_11, 1);
    __pyx_t_11.memview = NULL;
    __pyx_t_11.data = NULL;
    if ((likely(PyTuple_CheckExact(__pyx_t_3))) || (PyList_CheckExact(__pyx_t_3))) {
      PyObject* sequence = __pyx_t_3;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 3)) {
        if (size > 3) __Pyx_RaiseTooManyValuesError(3);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(1, 3338, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_5 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_10 = PyTuple_GET_ITEM(sequence, 1); 
        __pyx_t_9 = PyTuple_GET_ITEM(sequence, 2); 
      } else {
        __pyx_t_5 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_10 = PyList_GET_ITEM(sequence, 1); 
        __pyx_t_9 = PyList_GET_ITEM(sequence, 2); 
      }
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_10);
      __Pyx_INCREF(__pyx_t_9);
      #else
      __pyx_t_5 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 3338, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_10 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 3338, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_9 = PySequence_ITEM(sequence, 2); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 3338, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      #endif
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_2 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3338, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_12 = Py_TYPE(__pyx_t_2)->tp_iternext;
      index = 0; __pyx_t_5 = __pyx_t_12(__pyx_t_2); if (unlikely(!__pyx_t_5)) goto __pyx_L5_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_5);
      index = 1; __pyx_t_10 = __pyx_t_12(__pyx_t_2); if (unlikely(!__pyx_t_10)) goto __pyx_L5_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_10);
      index = 2; __pyx_t_9 = __pyx_t_12(__pyx_t_2); if (unlikely(!__pyx_t_9)) goto __pyx_L5_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_9);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_12(__pyx_t_2), 3) < 0) __PYX_ERR(1, 3338, __pyx_L1_error)
      __pyx_t_12 = NULL;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      goto __pyx_L6_unpacking_done;
      __pyx_L5_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_12 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(1, 3338, __pyx_L1_error)
      __pyx_L6_unpacking_done:;
    }
    __pyx_v_left_set = __pyx_t_5;
    __pyx_t_5 = 0;
    __pyx_v_right_set = __pyx_t_10;
    __pyx_t_10 = 0;
    __pyx_v_center_set = __pyx_t_9;
    __pyx_t_9 = 0;

    /* "pandas/_libs/intervaltree.pxi":3342
 * 
 *             self.left_node = self.new_child_node(left, right,
 *                                                  indices, left_set)             # <<<<<<<<<<<<<<
 *             self.right_node = self.new_child_node(left, right,
 *                                                   indices, right_set)
 */
    if (!(likely(((__pyx_v_left_set) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_left_set, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(1, 3342, __pyx_L1_error)

    /* "pandas/_libs/intervaltree.pxi":3341
 *                 left, right)
 * 
 *             self.left_node = self.new_child_node(left, right,             # <<<<<<<<<<<<<<
 *                                                  indices, left_set)
 *             self.right_node = self.new_child_node(left, right,
 */
    __pyx_t_3 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode *)__pyx_v_self->__pyx_vtab)->new_child_node(__pyx_v_self, ((PyArrayObject *)__pyx_v_left), ((PyArrayObject *)__pyx_v_right), ((PyArrayObject *)__pyx_v_indices), ((PyArrayObject *)__pyx_v_left_set)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 3341, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode))))) __PYX_ERR(1, 3341, __pyx_L1_error)
    __Pyx_GIVEREF(__pyx_t_3);
    __Pyx_GOTREF(__pyx_v_self->left_node);
    __Pyx_DECREF(((PyObject *)__pyx_v_self->left_node));
    __pyx_v_self->left_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode *)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "pandas/_libs/intervaltree.pxi":3344
 *                                                  indices, left_set)
 *             self.right_node = self.new_child_node(left, right,
 *                                                   indices, right_set)             # <<<<<<<<<<<<<<
 * 
 *             self.center_left_values, self.center_left_indices = \
 */
    if (!(likely(((__pyx_v_right_set) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_right_set, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(1, 3344, __pyx_L1_error)

    /* "pandas/_libs/intervaltree.pxi":3343
 *             self.left_node = self.new_child_node(left, right,
 *                                                  indices, left_set)
 *             self.right_node = self.new_child_node(left, right,             # <<<<<<<<<<<<<<
 *                                                   indices, right_set)
 * 
 */
    __pyx_t_3 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode *)__pyx_v_self->__pyx_vtab)->new_child_node(__pyx_v_self, ((PyArrayObject *)__pyx_v_left), ((PyArrayObject *)__pyx_v_right), ((PyArrayObject *)__pyx_v_indices), ((PyArrayObject *)__pyx_v_right_set)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 3343, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode))))) __PYX_ERR(1, 3343, __pyx_L1_error)
    __Pyx_GIVEREF(__pyx_t_3);
    __Pyx_GOTREF(__pyx_v_self->right_node);
    __Pyx_DECREF(((PyObject *)__pyx_v_self->right_node));
    __pyx_v_self->right_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode *)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "pandas/_libs/intervaltree.pxi":3347
 * 
 *             self.center_left_values, self.center_left_indices = \
 *                 sort_values_and_indices(left, indices, center_set)             # <<<<<<<<<<<<<<
 *             self.center_right_values, self.center_right_indices = \
 *                 sort_values_and_indices(right, indices, center_set)
 */
    __pyx_t_3 = __pyx_f_6pandas_5_libs_8interval_sort_values_and_indices(((PyObject *)__pyx_v_left), ((PyObject *)__pyx_v_indices), __pyx_v_center_set); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 3347, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if ((likely(PyTuple_CheckExact(__pyx_t_3))) || (PyList_CheckExact(__pyx_t_3))) {
      PyObject* sequence = __pyx_t_3;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(1, 3346, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_9 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_10 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_9 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_10 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_9);
      __Pyx_INCREF(__pyx_t_10);
      #else
      __pyx_t_9 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 3346, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_10 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 3346, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      #endif
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_5 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 3346, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_12 = Py_TYPE(__pyx_t_5)->tp_iternext;
      index = 0; __pyx_t_9 = __pyx_t_12(__pyx_t_5); if (unlikely(!__pyx_t_9)) goto __pyx_L7_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_9);
      index = 1; __pyx_t_10 = __pyx_t_12(__pyx_t_5); if (unlikely(!__pyx_t_10)) goto __pyx_L7_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_10);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_12(__pyx_t_5), 2) < 0) __PYX_ERR(1, 3346, __pyx_L1_error)
      __pyx_t_12 = NULL;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      goto __pyx_L8_unpacking_done;
      __pyx_L7_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_12 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(1, 3346, __pyx_L1_error)
      __pyx_L8_unpacking_done:;
    }

    /* "pandas/_libs/intervaltree.pxi":3346
 *                                                   indices, right_set)
 * 
 *             self.center_left_values, self.center_left_indices = \             # <<<<<<<<<<<<<<
 *                 sort_values_and_indices(left, indices, center_set)
 *             self.center_right_values, self.center_right_indices = \
 */
    __pyx_t_11 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_uint64_t(__pyx_t_9, PyBUF_WRITABLE); if (unlikely(!__pyx_t_11.memview)) __PYX_ERR(1, 3346, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_8 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_t_10, PyBUF_WRITABLE); if (unlikely(!__pyx_t_8.memview)) __PYX_ERR(1, 3346, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->center_left_values, 0);
    __pyx_v_self->center_left_values = __pyx_t_11;
    __pyx_t_11.memview = NULL;
    __pyx_t_11.data = NULL;
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->center_left_indices, 0);
    __pyx_v_self->center_left_indices = __pyx_t_8;
    __pyx_t_8.memview = NULL;
    __pyx_t_8.data = NULL;

    /* "pandas/_libs/intervaltree.pxi":3349
 *                 sort_values_and_indices(left, indices, center_set)
 *             self.center_right_values, self.center_right_indices = \
 *                 sort_values_and_indices(right, indices, center_set)             # <<<<<<<<<<<<<<
 *             self.n_center = len(self.center_left_indices)
 * 
 */
    __pyx_t_3 = __pyx_f_6pandas_5_libs_8interval_sort_values_and_indices(((PyObject *)__pyx_v_right), ((PyObject *)__pyx_v_indices), __pyx_v_center_set); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 3349, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if ((likely(PyTuple_CheckExact(__pyx_t_3))) || (PyList_CheckExact(__pyx_t_3))) {
      PyObject* sequence = __pyx_t_3;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(1, 3348, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_10 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_9 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_10 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_9 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_10);
      __Pyx_INCREF(__pyx_t_9);
      #else
      __pyx_t_10 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 3348, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_9 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 3348, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      #endif
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_5 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 3348, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_12 = Py_TYPE(__pyx_t_5)->tp_iternext;
      index = 0; __pyx_t_10 = __pyx_t_12(__pyx_t_5); if (unlikely(!__pyx_t_10)) goto __pyx_L9_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_10);
      index = 1; __pyx_t_9 = __pyx_t_12(__pyx_t_5); if (unlikely(!__pyx_t_9)) goto __pyx_L9_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_9);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_12(__pyx_t_5), 2) < 0) __PYX_ERR(1, 3348, __pyx_L1_error)
      __pyx_t_12 = NULL;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      goto __pyx_L10_unpacking_done;
      __pyx_L9_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_12 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(1, 3348, __pyx_L1_error)
      __pyx_L10_unpacking_done:;
    }

    /* "pandas/_libs/intervaltree.pxi":3348
 *             self.center_left_values, self.center_left_indices = \
 *                 sort_values_and_indices(left, indices, center_set)
 *             self.center_right_values, self.center_right_indices = \             # <<<<<<<<<<<<<<
 *                 sort_values_and_indices(right, indices, center_set)
 *             self.n_center = len(self.center_left_indices)
 */
    __pyx_t_11 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_uint64_t(__pyx_t_10, PyBUF_WRITABLE); if (unlikely(!__pyx_t_11.memview)) __PYX_ERR(1, 3348, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_t_8 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_t_9, PyBUF_WRITABLE); if (unlikely(!__pyx_t_8.memview)) __PYX_ERR(1, 3348, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->center_right_values, 0);
    __pyx_v_self->center_right_values = __pyx_t_11;
    __pyx_t_11.memview = NULL;
    __pyx_t_11.data = NULL;
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->center_right_indices, 0);
    __pyx_v_self->center_right_indices = __pyx_t_8;
    __pyx_t_8.memview = NULL;
    __pyx_t_8.data = NULL;

    /* "pandas/_libs/intervaltree.pxi":3350
 *             self.center_right_values, self.center_right_indices = \
 *                 sort_values_and_indices(right, indices, center_set)
 *             self.n_center = len(self.center_left_indices)             # <<<<<<<<<<<<<<
 * 
 *     @cython.wraparound(False)
 */
    if (unlikely(!__pyx_v_self->center_left_indices.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(1, 3350, __pyx_L1_error)}
    __pyx_t_13 = __Pyx_MemoryView_Len(__pyx_v_self->center_left_indices); 
    __pyx_v_self->n_center = __pyx_t_13;
  }
  __pyx_L4:;

  /* "pandas/_libs/intervaltree.pxi":3308
 *         readonly bint is_leaf_node
 * 
 *     def __init__(self,             # <<<<<<<<<<<<<<
 *                  ndarray[uint64_t, ndim=1] left,
 *                  ndarray[uint64_t, ndim=1] right,
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __PYX_XDEC_MEMVIEW(&__pyx_t_7, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_8, 1);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __PYX_XDEC_MEMVIEW(&__pyx_t_11, 1);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indices.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_left.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_right.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("pandas._libs.interval.Uint64ClosedBothIntervalNode.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indices.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_left.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_right.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF(__pyx_v_left_set);
  __Pyx_XDECREF(__pyx_v_right_set);
  __Pyx_XDECREF(__pyx_v_center_set);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":3354
 *     @cython.wraparound(False)
 *     @cython.boundscheck(False)
 *     cdef classify_intervals(self, uint64_t[:] left, uint64_t[:] right):             # <<<<<<<<<<<<<<
 *         """Classify the given intervals based upon whether they fall to the
 *         left, right, or overlap with this node's pivot.
 */

static PyObject *__pyx_f_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_classify_intervals(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode *__pyx_v_self, __Pyx_memviewslice __pyx_v_left, __Pyx_memviewslice __pyx_v_right) {
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_left_ind = 0;
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_right_ind = 0;
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_overlapping_ind = 0;
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __pyx_t_5numpy_int64_t __pyx_t_2;
  __pyx_t_5numpy_int64_t __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  int __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  __Pyx_RefNannySetupContext("classify_intervals", 0);

  /* "pandas/_libs/intervaltree.pxi":3362
 *             Py_ssize_t i
 * 
 *         left_ind = Int64Vector()             # <<<<<<<<<<<<<<
 *         right_ind = Int64Vector()
 *         overlapping_ind = Int64Vector()
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3362, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_left_ind = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":3363
 * 
 *         left_ind = Int64Vector()
 *         right_ind = Int64Vector()             # <<<<<<<<<<<<<<
 *         overlapping_ind = Int64Vector()
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3363, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_right_ind = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":3364
 *         left_ind = Int64Vector()
 *         right_ind = Int64Vector()
 *         overlapping_ind = Int64Vector()             # <<<<<<<<<<<<<<
 * 
 *         for i in range(self.n_elements):
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3364, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_overlapping_ind = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":3366
 *         overlapping_ind = Int64Vector()
 * 
 *         for i in range(self.n_elements):             # <<<<<<<<<<<<<<
 *             if right[i] < self.pivot:
 *                 left_ind.append(i)
 */
  __pyx_t_2 = __pyx_v_self->n_elements;
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "pandas/_libs/intervaltree.pxi":3367
 * 
 *         for i in range(self.n_elements):
 *             if right[i] < self.pivot:             # <<<<<<<<<<<<<<
 *                 left_ind.append(i)
 *             elif self.pivot < left[i]:
 */
    __pyx_t_5 = __pyx_v_i;
    __pyx_t_6 = (((*((__pyx_t_5numpy_uint64_t *) ( /* dim=0 */ (__pyx_v_right.data + __pyx_t_5 * __pyx_v_right.strides[0]) ))) < __pyx_v_self->pivot) != 0);
    if (__pyx_t_6) {

      /* "pandas/_libs/intervaltree.pxi":3368
 *         for i in range(self.n_elements):
 *             if right[i] < self.pivot:
 *                 left_ind.append(i)             # <<<<<<<<<<<<<<
 *             elif self.pivot < left[i]:
 *                 right_ind.append(i)
 */
      ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_left_ind->__pyx_vtab)->append(__pyx_v_left_ind, __pyx_v_i);

      /* "pandas/_libs/intervaltree.pxi":3367
 * 
 *         for i in range(self.n_elements):
 *             if right[i] < self.pivot:             # <<<<<<<<<<<<<<
 *                 left_ind.append(i)
 *             elif self.pivot < left[i]:
 */
      goto __pyx_L5;
    }

    /* "pandas/_libs/intervaltree.pxi":3369
 *             if right[i] < self.pivot:
 *                 left_ind.append(i)
 *             elif self.pivot < left[i]:             # <<<<<<<<<<<<<<
 *                 right_ind.append(i)
 *             else:
 */
    __pyx_t_7 = __pyx_v_i;
    __pyx_t_6 = ((__pyx_v_self->pivot < (*((__pyx_t_5numpy_uint64_t *) ( /* dim=0 */ (__pyx_v_left.data + __pyx_t_7 * __pyx_v_left.strides[0]) )))) != 0);
    if (__pyx_t_6) {

      /* "pandas/_libs/intervaltree.pxi":3370
 *                 left_ind.append(i)
 *             elif self.pivot < left[i]:
 *                 right_ind.append(i)             # <<<<<<<<<<<<<<
 *             else:
 *                 overlapping_ind.append(i)
 */
      ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_right_ind->__pyx_vtab)->append(__pyx_v_right_ind, __pyx_v_i);

      /* "pandas/_libs/intervaltree.pxi":3369
 *             if right[i] < self.pivot:
 *                 left_ind.append(i)
 *             elif self.pivot < left[i]:             # <<<<<<<<<<<<<<
 *                 right_ind.append(i)
 *             else:
 */
      goto __pyx_L5;
    }

    /* "pandas/_libs/intervaltree.pxi":3372
 *                 right_ind.append(i)
 *             else:
 *                 overlapping_ind.append(i)             # <<<<<<<<<<<<<<
 * 
 *         return (left_ind.to_array(),
 */
    /*else*/ {
      ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_overlapping_ind->__pyx_vtab)->append(__pyx_v_overlapping_ind, __pyx_v_i);
    }
    __pyx_L5:;
  }

  /* "pandas/_libs/intervaltree.pxi":3374
 *                 overlapping_ind.append(i)
 * 
 *         return (left_ind.to_array(),             # <<<<<<<<<<<<<<
 *                 right_ind.to_array(),
 *                 overlapping_ind.to_array())
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_left_ind->__pyx_vtab)->to_array(__pyx_v_left_ind, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3374, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "pandas/_libs/intervaltree.pxi":3375
 * 
 *         return (left_ind.to_array(),
 *                 right_ind.to_array(),             # <<<<<<<<<<<<<<
 *                 overlapping_ind.to_array())
 * 
 */
  __pyx_t_8 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_right_ind->__pyx_vtab)->to_array(__pyx_v_right_ind, 0); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 3375, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);

  /* "pandas/_libs/intervaltree.pxi":3376
 *         return (left_ind.to_array(),
 *                 right_ind.to_array(),
 *                 overlapping_ind.to_array())             # <<<<<<<<<<<<<<
 * 
 *     cdef new_child_node(self,
 */
  __pyx_t_9 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_overlapping_ind->__pyx_vtab)->to_array(__pyx_v_overlapping_ind, 0); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 3376, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);

  /* "pandas/_libs/intervaltree.pxi":3374
 *                 overlapping_ind.append(i)
 * 
 *         return (left_ind.to_array(),             # <<<<<<<<<<<<<<
 *                 right_ind.to_array(),
 *                 overlapping_ind.to_array())
 */
  __pyx_t_10 = PyTuple_New(3); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 3374, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_t_8);
  __Pyx_GIVEREF(__pyx_t_9);
  PyTuple_SET_ITEM(__pyx_t_10, 2, __pyx_t_9);
  __pyx_t_1 = 0;
  __pyx_t_8 = 0;
  __pyx_t_9 = 0;
  __pyx_r = __pyx_t_10;
  __pyx_t_10 = 0;
  goto __pyx_L0;

  /* "pandas/_libs/intervaltree.pxi":3354
 *     @cython.wraparound(False)
 *     @cython.boundscheck(False)
 *     cdef classify_intervals(self, uint64_t[:] left, uint64_t[:] right):             # <<<<<<<<<<<<<<
 *         """Classify the given intervals based upon whether they fall to the
 *         left, right, or overlap with this node's pivot.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("pandas._libs.interval.Uint64ClosedBothIntervalNode.classify_intervals", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_left_ind);
  __Pyx_XDECREF((PyObject *)__pyx_v_right_ind);
  __Pyx_XDECREF((PyObject *)__pyx_v_overlapping_ind);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":3378
 *                 overlapping_ind.to_array())
 * 
 *     cdef new_child_node(self,             # <<<<<<<<<<<<<<
 *                         ndarray[uint64_t, ndim=1] left,
 *                         ndarray[uint64_t, ndim=1] right,
 */

static PyObject *__pyx_f_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_new_child_node(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode *__pyx_v_self, PyArrayObject *__pyx_v_left, PyArrayObject *__pyx_v_right, PyArrayObject *__pyx_v_indices, PyArrayObject *__pyx_v_subset) {
  __Pyx_LocalBuf_ND __pyx_pybuffernd_indices;
  __Pyx_Buffer __pyx_pybuffer_indices;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_left;
  __Pyx_Buffer __pyx_pybuffer_left;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_right;
  __Pyx_Buffer __pyx_pybuffer_right;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_subset;
  __Pyx_Buffer __pyx_pybuffer_subset;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyArrayObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyArrayObject *__pyx_t_7 = NULL;
  PyArrayObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  __Pyx_RefNannySetupContext("new_child_node", 0);
  __Pyx_INCREF((PyObject *)__pyx_v_left);
  __Pyx_INCREF((PyObject *)__pyx_v_right);
  __Pyx_INCREF((PyObject *)__pyx_v_indices);
  __pyx_pybuffer_left.pybuffer.buf = NULL;
  __pyx_pybuffer_left.refcount = 0;
  __pyx_pybuffernd_left.data = NULL;
  __pyx_pybuffernd_left.rcbuffer = &__pyx_pybuffer_left;
  __pyx_pybuffer_right.pybuffer.buf = NULL;
  __pyx_pybuffer_right.refcount = 0;
  __pyx_pybuffernd_right.data = NULL;
  __pyx_pybuffernd_right.rcbuffer = &__pyx_pybuffer_right;
  __pyx_pybuffer_indices.pybuffer.buf = NULL;
  __pyx_pybuffer_indices.refcount = 0;
  __pyx_pybuffernd_indices.data = NULL;
  __pyx_pybuffernd_indices.rcbuffer = &__pyx_pybuffer_indices;
  __pyx_pybuffer_subset.pybuffer.buf = NULL;
  __pyx_pybuffer_subset.refcount = 0;
  __pyx_pybuffernd_subset.data = NULL;
  __pyx_pybuffernd_subset.rcbuffer = &__pyx_pybuffer_subset;
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_left.rcbuffer->pybuffer, (PyObject*)__pyx_v_left, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 3378, __pyx_L1_error)
  }
  __pyx_pybuffernd_left.diminfo[0].strides = __pyx_pybuffernd_left.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_left.diminfo[0].shape = __pyx_pybuffernd_left.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_right.rcbuffer->pybuffer, (PyObject*)__pyx_v_right, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 3378, __pyx_L1_error)
  }
  __pyx_pybuffernd_right.diminfo[0].strides = __pyx_pybuffernd_right.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_right.diminfo[0].shape = __pyx_pybuffernd_right.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_indices.rcbuffer->pybuffer, (PyObject*)__pyx_v_indices, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 3378, __pyx_L1_error)
  }
  __pyx_pybuffernd_indices.diminfo[0].strides = __pyx_pybuffernd_indices.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_indices.diminfo[0].shape = __pyx_pybuffernd_indices.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_subset.rcbuffer->pybuffer, (PyObject*)__pyx_v_subset, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 3378, __pyx_L1_error)
  }
  __pyx_pybuffernd_subset.diminfo[0].strides = __pyx_pybuffernd_subset.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_subset.diminfo[0].shape = __pyx_pybuffernd_subset.rcbuffer->pybuffer.shape[0];

  /* "pandas/_libs/intervaltree.pxi":3385
 *         """Create a new child node.
 *         """
 *         left = take(left, subset)             # <<<<<<<<<<<<<<
 *         right = take(right, subset)
 *         indices = take(indices, subset)
 */
  __pyx_t_1 = __pyx_f_6pandas_5_libs_8interval_take(((PyArrayObject *)__pyx_v_left), ((PyArrayObject *)__pyx_v_subset)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3385, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(1, 3385, __pyx_L1_error)
  __pyx_t_2 = ((PyArrayObject *)__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_left.rcbuffer->pybuffer);
    __pyx_t_3 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_left.rcbuffer->pybuffer, (PyObject*)__pyx_t_2, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack);
    if (unlikely(__pyx_t_3 < 0)) {
      PyErr_Fetch(&__pyx_t_4, &__pyx_t_5, &__pyx_t_6);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_left.rcbuffer->pybuffer, (PyObject*)__pyx_v_left, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_4); Py_XDECREF(__pyx_t_5); Py_XDECREF(__pyx_t_6);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_4, __pyx_t_5, __pyx_t_6);
      }
      __pyx_t_4 = __pyx_t_5 = __pyx_t_6 = 0;
    }
    __pyx_pybuffernd_left.diminfo[0].strides = __pyx_pybuffernd_left.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_left.diminfo[0].shape = __pyx_pybuffernd_left.rcbuffer->pybuffer.shape[0];
    if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(1, 3385, __pyx_L1_error)
  }
  __pyx_t_2 = 0;
  __Pyx_DECREF_SET(__pyx_v_left, ((PyArrayObject *)__pyx_t_1));
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":3386
 *         """
 *         left = take(left, subset)
 *         right = take(right, subset)             # <<<<<<<<<<<<<<
 *         indices = take(indices, subset)
 *         return Uint64ClosedBothIntervalNode(
 */
  __pyx_t_1 = __pyx_f_6pandas_5_libs_8interval_take(((PyArrayObject *)__pyx_v_right), ((PyArrayObject *)__pyx_v_subset)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3386, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(1, 3386, __pyx_L1_error)
  __pyx_t_7 = ((PyArrayObject *)__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_right.rcbuffer->pybuffer);
    __pyx_t_3 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_right.rcbuffer->pybuffer, (PyObject*)__pyx_t_7, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack);
    if (unlikely(__pyx_t_3 < 0)) {
      PyErr_Fetch(&__pyx_t_6, &__pyx_t_5, &__pyx_t_4);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_right.rcbuffer->pybuffer, (PyObject*)__pyx_v_right, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_6); Py_XDECREF(__pyx_t_5); Py_XDECREF(__pyx_t_4);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_6, __pyx_t_5, __pyx_t_4);
      }
      __pyx_t_6 = __pyx_t_5 = __pyx_t_4 = 0;
    }
    __pyx_pybuffernd_right.diminfo[0].strides = __pyx_pybuffernd_right.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_right.diminfo[0].shape = __pyx_pybuffernd_right.rcbuffer->pybuffer.shape[0];
    if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(1, 3386, __pyx_L1_error)
  }
  __pyx_t_7 = 0;
  __Pyx_DECREF_SET(__pyx_v_right, ((PyArrayObject *)__pyx_t_1));
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":3387
 *         left = take(left, subset)
 *         right = take(right, subset)
 *         indices = take(indices, subset)             # <<<<<<<<<<<<<<
 *         return Uint64ClosedBothIntervalNode(
 *             left, right, indices, self.leaf_size)
 */
  __pyx_t_1 = __pyx_f_6pandas_5_libs_8interval_take(((PyArrayObject *)__pyx_v_indices), ((PyArrayObject *)__pyx_v_subset)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3387, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(1, 3387, __pyx_L1_error)
  __pyx_t_8 = ((PyArrayObject *)__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indices.rcbuffer->pybuffer);
    __pyx_t_3 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_indices.rcbuffer->pybuffer, (PyObject*)__pyx_t_8, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack);
    if (unlikely(__pyx_t_3 < 0)) {
      PyErr_Fetch(&__pyx_t_4, &__pyx_t_5, &__pyx_t_6);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_indices.rcbuffer->pybuffer, (PyObject*)__pyx_v_indices, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_4); Py_XDECREF(__pyx_t_5); Py_XDECREF(__pyx_t_6);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_4, __pyx_t_5, __pyx_t_6);
      }
      __pyx_t_4 = __pyx_t_5 = __pyx_t_6 = 0;
    }
    __pyx_pybuffernd_indices.diminfo[0].strides = __pyx_pybuffernd_indices.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_indices.diminfo[0].shape = __pyx_pybuffernd_indices.rcbuffer->pybuffer.shape[0];
    if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(1, 3387, __pyx_L1_error)
  }
  __pyx_t_8 = 0;
  __Pyx_DECREF_SET(__pyx_v_indices, ((PyArrayObject *)__pyx_t_1));
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":3388
 *         right = take(right, subset)
 *         indices = take(indices, subset)
 *         return Uint64ClosedBothIntervalNode(             # <<<<<<<<<<<<<<
 *             left, right, indices, self.leaf_size)
 * 
 */
  __Pyx_XDECREF(__pyx_r);

  /* "pandas/_libs/intervaltree.pxi":3389
 *         indices = take(indices, subset)
 *         return Uint64ClosedBothIntervalNode(
 *             left, right, indices, self.leaf_size)             # <<<<<<<<<<<<<<
 * 
 *     @cython.wraparound(False)
 */
  __pyx_t_1 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->leaf_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3389, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "pandas/_libs/intervaltree.pxi":3388
 *         right = take(right, subset)
 *         indices = take(indices, subset)
 *         return Uint64ClosedBothIntervalNode(             # <<<<<<<<<<<<<<
 *             left, right, indices, self.leaf_size)
 * 
 */
  __pyx_t_9 = PyTuple_New(4); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 3388, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_INCREF(((PyObject *)__pyx_v_left));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_left));
  PyTuple_SET_ITEM(__pyx_t_9, 0, ((PyObject *)__pyx_v_left));
  __Pyx_INCREF(((PyObject *)__pyx_v_right));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_right));
  PyTuple_SET_ITEM(__pyx_t_9, 1, ((PyObject *)__pyx_v_right));
  __Pyx_INCREF(((PyObject *)__pyx_v_indices));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_indices));
  PyTuple_SET_ITEM(__pyx_t_9, 2, ((PyObject *)__pyx_v_indices));
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_9, 3, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode), __pyx_t_9, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3388, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pandas/_libs/intervaltree.pxi":3378
 *                 overlapping_ind.to_array())
 * 
 *     cdef new_child_node(self,             # <<<<<<<<<<<<<<
 *                         ndarray[uint64_t, ndim=1] left,
 *                         ndarray[uint64_t, ndim=1] right,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_9);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indices.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_left.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_right.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_subset.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("pandas._libs.interval.Uint64ClosedBothIntervalNode.new_child_node", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indices.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_left.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_right.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_subset.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF((PyObject *)__pyx_v_left);
  __Pyx_XDECREF((PyObject *)__pyx_v_right);
  __Pyx_XDECREF((PyObject *)__pyx_v_indices);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":3394
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_2query[] = "Recursively query this node and its sub-nodes for intervals that\n        overlap with the query point.\n        ";
static PyMethodDef __pyx_mdef_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_3query = {"query", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_3query, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_2query};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_args = 0;
  PyObject *__pyx_v_kwargs = 0;
  CYTHON_UNUSED PyObject *__pyx_v_defaults = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fused_cpdef (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_args,&__pyx_n_s_kwargs,&__pyx_n_s_defaults,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_args)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_kwargs)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fused_cpdef", 1, 3, 3, 1); __PYX_ERR(1, 3394, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_defaults)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fused_cpdef", 1, 3, 3, 2); __PYX_ERR(1, 3394, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fused_cpdef") < 0)) __PYX_ERR(1, 3394, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_args = values[0];
    __pyx_v_kwargs = values[1];
    __pyx_v_defaults = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fused_cpdef", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 3394, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Uint64ClosedBothIntervalNode.__pyx_fused_cpdef", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_2query(((struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode *)__pyx_v_signatures), __pyx_v_args, __pyx_v_kwargs, __pyx_v_defaults);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_2query(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode *__pyx_v_signatures, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs, CYTHON_UNUSED PyObject *__pyx_v_defaults) {
  PyObject *__pyx_v_dest_sig = NULL;
  Py_ssize_t __pyx_v_i;
  CYTHON_UNUSED PyTypeObject *__pyx_v_ndarray = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_v_itemsize;
  PyObject *__pyx_v_arg = NULL;
  PyObject *__pyx_v_candidates = NULL;
  PyObject *__pyx_v_sig = NULL;
  int __pyx_v_match_found;
  PyObject *__pyx_v_src_sig = NULL;
  PyObject *__pyx_v_dst_type = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  Py_ssize_t __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  Py_ssize_t __pyx_t_13;
  Py_ssize_t __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  int __pyx_t_16;
  __Pyx_RefNannySetupContext("query", 0);
  __Pyx_INCREF(__pyx_v_kwargs);
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3394, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  PyList_SET_ITEM(__pyx_t_1, 0, Py_None);
  __pyx_v_dest_sig = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_3 = (__pyx_v_kwargs != Py_None);
  __pyx_t_4 = (__pyx_t_3 != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_kwargs); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(1, 3394, __pyx_L1_error)
  __pyx_t_3 = ((!__pyx_t_4) != 0);
  __pyx_t_2 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {
    __Pyx_INCREF(Py_None);
    __Pyx_DECREF_SET(__pyx_v_kwargs, Py_None);
  }
  __pyx_t_1 = ((PyObject *)__Pyx_ImportNumPyArrayTypeIfAvailable()); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3394, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_ndarray = ((PyTypeObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_v_itemsize = -1L;
  if (unlikely(__pyx_v_args == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 3394, __pyx_L1_error)
  }
  __pyx_t_5 = PyTuple_GET_SIZE(((PyObject*)__pyx_v_args)); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(1, 3394, __pyx_L1_error)
  __pyx_t_2 = ((2 < __pyx_t_5) != 0);
  if (__pyx_t_2) {
    if (unlikely(__pyx_v_args == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 3394, __pyx_L1_error)
    }
    __pyx_t_1 = PyTuple_GET_ITEM(((PyObject*)__pyx_v_args), 2);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v_arg = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L6;
  }
  __pyx_t_3 = (__pyx_v_kwargs != Py_None);
  __pyx_t_4 = (__pyx_t_3 != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L7_bool_binop_done;
  }
  if (unlikely(__pyx_v_kwargs == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(1, 3394, __pyx_L1_error)
  }
  __pyx_t_4 = (__Pyx_PyDict_ContainsTF(__pyx_n_s_point, ((PyObject*)__pyx_v_kwargs), Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(1, 3394, __pyx_L1_error)
  __pyx_t_3 = (__pyx_t_4 != 0);
  __pyx_t_2 = __pyx_t_3;
  __pyx_L7_bool_binop_done:;
  if (__pyx_t_2) {
    if (unlikely(__pyx_v_kwargs == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 3394, __pyx_L1_error)
    }
    __pyx_t_1 = __Pyx_PyDict_GetItem(((PyObject*)__pyx_v_kwargs), __pyx_n_s_point); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3394, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_arg = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L6;
  }
  /*else*/ {
    if (unlikely(__pyx_v_args == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
      __PYX_ERR(1, 3394, __pyx_L1_error)
    }
    __pyx_t_5 = PyTuple_GET_SIZE(((PyObject*)__pyx_v_args)); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(1, 3394, __pyx_L1_error)
    __pyx_t_1 = PyInt_FromSsize_t(__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3394, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_6 = PyTuple_New(3); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 3394, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_INCREF(__pyx_int_3);
    __Pyx_GIVEREF(__pyx_int_3);
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_int_3);
    __Pyx_INCREF(__pyx_n_s_s);
    __Pyx_GIVEREF(__pyx_n_s_s);
    PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_n_s_s);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Expected_at_least_d_argument_s_g, __pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3394, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_1); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 3394, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(1, 3394, __pyx_L1_error)
  }
  __pyx_L6:;
  while (1) {
    __pyx_t_2 = PyFloat_Check(__pyx_v_arg); 
    __pyx_t_3 = (__pyx_t_2 != 0);
    if (__pyx_t_3) {
      if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, __pyx_n_s_float64_t, long, 1, __Pyx_PyInt_From_long, 1, 0, 0) < 0)) __PYX_ERR(1, 3394, __pyx_L1_error)
      goto __pyx_L10_break;
    }
    __pyx_t_2 = PyInt_Check(__pyx_v_arg); 
    __pyx_t_4 = (__pyx_t_2 != 0);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_3 = __pyx_t_4;
      goto __pyx_L13_bool_binop_done;
    }
    __pyx_t_4 = PyLong_Check(__pyx_v_arg); 
    __pyx_t_2 = (__pyx_t_4 != 0);
    __pyx_t_3 = __pyx_t_2;
    __pyx_L13_bool_binop_done:;
    __pyx_t_2 = (__pyx_t_3 != 0);
    if (__pyx_t_2) {
      if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, __pyx_n_s_int64_t, long, 1, __Pyx_PyInt_From_long, 1, 0, 0) < 0)) __PYX_ERR(1, 3394, __pyx_L1_error)
      goto __pyx_L10_break;
    }
    if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, Py_None, long, 1, __Pyx_PyInt_From_long, 1, 0, 0) < 0)) __PYX_ERR(1, 3394, __pyx_L1_error)
    goto __pyx_L10_break;
  }
  __pyx_L10_break:;
  __pyx_t_6 = PyList_New(0); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 3394, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_v_candidates = ((PyObject*)__pyx_t_6);
  __pyx_t_6 = 0;
  __pyx_t_5 = 0;
  if (unlikely(((PyObject *)__pyx_v_signatures) == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(1, 3394, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_dict_iterator(((PyObject*)__pyx_v_signatures), 1, ((PyObject *)NULL), (&__pyx_t_7), (&__pyx_t_8)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3394, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __pyx_t_6 = __pyx_t_1;
  __pyx_t_1 = 0;
  while (1) {
    __pyx_t_9 = __Pyx_dict_iter_next(__pyx_t_6, __pyx_t_7, &__pyx_t_5, &__pyx_t_1, NULL, NULL, __pyx_t_8);
    if (unlikely(__pyx_t_9 == 0)) break;
    if (unlikely(__pyx_t_9 == -1)) __PYX_ERR(1, 3394, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_sig, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_v_match_found = 0;
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_v_sig, __pyx_n_s_strip); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 3394, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __pyx_t_12 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_11))) {
      __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_11);
      if (likely(__pyx_t_12)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
        __Pyx_INCREF(__pyx_t_12);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_11, function);
      }
    }
    __pyx_t_10 = (__pyx_t_12) ? __Pyx_PyObject_Call2Args(__pyx_t_11, __pyx_t_12, __pyx_kp_s__13) : __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_kp_s__13);
    __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
    if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 3394, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_split); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 3394, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_t_10 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_11))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_11);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_11, function);
      }
    }
    __pyx_t_1 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_11, __pyx_t_10, __pyx_kp_s__14) : __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_kp_s__14);
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3394, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_XDECREF_SET(__pyx_v_src_sig, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_13 = PyList_GET_SIZE(__pyx_v_dest_sig); if (unlikely(__pyx_t_13 == ((Py_ssize_t)-1))) __PYX_ERR(1, 3394, __pyx_L1_error)
    __pyx_t_14 = __pyx_t_13;
    for (__pyx_t_15 = 0; __pyx_t_15 < __pyx_t_14; __pyx_t_15+=1) {
      __pyx_v_i = __pyx_t_15;
      __pyx_t_1 = PyList_GET_ITEM(__pyx_v_dest_sig, __pyx_v_i);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_XDECREF_SET(__pyx_v_dst_type, __pyx_t_1);
      __pyx_t_1 = 0;
      __pyx_t_2 = (__pyx_v_dst_type != Py_None);
      __pyx_t_3 = (__pyx_t_2 != 0);
      if (__pyx_t_3) {
        __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_src_sig, __pyx_v_i, Py_ssize_t, 1, PyInt_FromSsize_t, 0, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3394, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_11 = PyObject_RichCompare(__pyx_t_1, __pyx_v_dst_type, Py_EQ); __Pyx_XGOTREF(__pyx_t_11); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 3394, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_11); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(1, 3394, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        if (__pyx_t_3) {
          __pyx_v_match_found = 1;
          goto __pyx_L20;
        }
        /*else*/ {
          __pyx_v_match_found = 0;
          goto __pyx_L18_break;
        }
        __pyx_L20:;
      }
    }
    __pyx_L18_break:;
    __pyx_t_3 = (__pyx_v_match_found != 0);
    if (__pyx_t_3) {
      __pyx_t_16 = __Pyx_PyList_Append(__pyx_v_candidates, __pyx_v_sig); if (unlikely(__pyx_t_16 == ((int)-1))) __PYX_ERR(1, 3394, __pyx_L1_error)
    }
  }
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_3 = (PyList_GET_SIZE(__pyx_v_candidates) != 0);
  __pyx_t_2 = ((!__pyx_t_3) != 0);
  if (__pyx_t_2) {
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__15, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 3394, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(1, 3394, __pyx_L1_error)
  }
  __pyx_t_7 = PyList_GET_SIZE(__pyx_v_candidates); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1))) __PYX_ERR(1, 3394, __pyx_L1_error)
  __pyx_t_2 = ((__pyx_t_7 > 1) != 0);
  if (__pyx_t_2) {
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__16, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 3394, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(1, 3394, __pyx_L1_error)
  }
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    if (unlikely(((PyObject *)__pyx_v_signatures) == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 3394, __pyx_L1_error)
    }
    __pyx_t_6 = __Pyx_PyDict_GetItem(((PyObject*)__pyx_v_signatures), PyList_GET_ITEM(__pyx_v_candidates, 0)); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 3394, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_r = __pyx_t_6;
    __pyx_t_6 = 0;
    goto __pyx_L0;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("pandas._libs.interval.Uint64ClosedBothIntervalNode.__pyx_fused_cpdef", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_dest_sig);
  __Pyx_XDECREF(__pyx_v_ndarray);
  __Pyx_XDECREF(__pyx_v_arg);
  __Pyx_XDECREF(__pyx_v_candidates);
  __Pyx_XDECREF(__pyx_v_sig);
  __Pyx_XDECREF(__pyx_v_src_sig);
  __Pyx_XDECREF(__pyx_v_dst_type);
  __Pyx_XDECREF(__pyx_v_kwargs);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_9__pyx_fuse_0query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_fuse_0__pyx_f_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float64_t __pyx_v_point, int __pyx_skip_dispatch) {
  __Pyx_memviewslice __pyx_v_indices = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_values = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __pyx_t_5numpy_int64_t __pyx_t_9;
  __pyx_t_5numpy_int64_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  int __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  __Pyx_memviewslice __pyx_t_16 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_17 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  Py_ssize_t __pyx_t_21;
  __Pyx_RefNannySetupContext("__pyx_fuse_0query", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pyx_fuse_0query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3394, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_9__pyx_fuse_0query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = PyFloat_FromDouble(__pyx_v_point); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 3394, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3394, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3394, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 3394, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_5) {
            __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3394, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3394, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_3query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_4 = PyFloat_FromDouble(__pyx_v_point); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 3394, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_7 = __pyx_t_1; __pyx_t_3 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3394, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3394, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 3394, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (__pyx_t_3) {
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_t_4);
          __pyx_t_4 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3394, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pandas/_libs/intervaltree.pxi":3403
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
  __pyx_t_8 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_8) {

    /* "pandas/_libs/intervaltree.pxi":3407
 *             # continue the binary tree structure. Instead, we use linear
 *             # search.
 *             for i in range(self.n_elements):             # <<<<<<<<<<<<<<
 *                 if self.left[i] <= point <= self.right[i]:
 *                     result.append(self.indices[i])
 */
    __pyx_t_9 = __pyx_v_self->n_elements;
    __pyx_t_10 = __pyx_t_9;
    for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
      __pyx_v_i = __pyx_t_11;

      /* "pandas/_libs/intervaltree.pxi":3408
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point <= self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      __pyx_t_12 = __pyx_v_i;
      __pyx_t_8 = ((*((__pyx_t_5numpy_uint64_t *) ( /* dim=0 */ (__pyx_v_self->left.data + __pyx_t_12 * __pyx_v_self->left.strides[0]) ))) <= __pyx_v_point);
      if (__pyx_t_8) {
        __pyx_t_13 = __pyx_v_i;
        __pyx_t_8 = (__pyx_v_point <= (*((__pyx_t_5numpy_uint64_t *) ( /* dim=0 */ (__pyx_v_self->right.data + __pyx_t_13 * __pyx_v_self->right.strides[0]) ))));
      }
      __pyx_t_14 = (__pyx_t_8 != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":3409
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point <= self.right[i]:
 *                     result.append(self.indices[i])             # <<<<<<<<<<<<<<
 *         else:
 *             # There are child nodes. Based on comparing our query to the pivot,
 */
        __pyx_t_15 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->indices.data + __pyx_t_15 * __pyx_v_self->indices.strides[0]) ))));

        /* "pandas/_libs/intervaltree.pxi":3408
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point <= self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      }
    }

    /* "pandas/_libs/intervaltree.pxi":3403
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
    goto __pyx_L3;
  }

  /* "pandas/_libs/intervaltree.pxi":3413
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
  /*else*/ {
    __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":3414
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:
 *                 values = self.center_left_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 */
      __pyx_t_16 = __pyx_v_self->center_left_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":3415
 *             if point < self.pivot:
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:
 */
      __pyx_t_17 = __pyx_v_self->center_left_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":3416
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):             # <<<<<<<<<<<<<<
 *                     if not values[i] <= point:
 *                         break
 */
      __pyx_t_9 = __pyx_v_self->n_center;
      __pyx_t_10 = __pyx_t_9;
      for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":3417
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_18 = __pyx_v_i;
        __pyx_t_14 = ((!(((*((__pyx_t_5numpy_uint64_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_18 * __pyx_v_values.strides[0]) ))) <= __pyx_v_point) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":3418
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:
 */
          goto __pyx_L9_break;

          /* "pandas/_libs/intervaltree.pxi":3417
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":3419
 *                     if not values[i] <= point:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 */
        __pyx_t_19 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_19 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L9_break:;

      /* "pandas/_libs/intervaltree.pxi":3420
 *                         break
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      __pyx_t_14 = ((__pyx_v_point <= __pyx_v_self->left_node->max_right) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":3421
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)             # <<<<<<<<<<<<<<
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode *)__pyx_v_self->left_node->__pyx_vtab)->__pyx_fuse_0query(__pyx_v_self->left_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3421, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":3420
 *                         break
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":3413
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":3422
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
    __pyx_t_14 = ((__pyx_v_point > __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":3423
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 *                 values = self.center_right_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 */
      __pyx_t_16 = __pyx_v_self->center_right_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":3424
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:
 */
      __pyx_t_17 = __pyx_v_self->center_right_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":3425
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):             # <<<<<<<<<<<<<<
 *                     if not point <= values[i]:
 *                         break
 */
      for (__pyx_t_11 = (__pyx_v_self->n_center - 1); __pyx_t_11 > -1L; __pyx_t_11-=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":3426
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_20 = __pyx_v_i;
        __pyx_t_14 = ((!((__pyx_v_point <= (*((__pyx_t_5numpy_uint64_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_20 * __pyx_v_values.strides[0]) )))) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":3427
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:
 */
          goto __pyx_L13_break;

          /* "pandas/_libs/intervaltree.pxi":3426
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":3428
 *                     if not point <= values[i]:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if self.right_node.min_left <= point:
 *                     self.right_node.query(result, point)
 */
        __pyx_t_21 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_21 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L13_break:;

      /* "pandas/_libs/intervaltree.pxi":3429
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      __pyx_t_14 = ((__pyx_v_self->right_node->min_left <= __pyx_v_point) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":3430
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:
 *                     self.right_node.query(result, point)             # <<<<<<<<<<<<<<
 *             else:
 *                 result.extend(self.center_left_indices)
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode *)__pyx_v_self->right_node->__pyx_vtab)->__pyx_fuse_0query(__pyx_v_self->right_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3430, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":3429
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":3422
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":3432
 *                     self.right_node.query(result, point)
 *             else:
 *                 result.extend(self.center_left_indices)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
    /*else*/ {
      __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->extend(__pyx_v_result, __pyx_v_self->center_left_indices); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3432, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_L7:;
  }
  __pyx_L3:;

  /* "pandas/_libs/intervaltree.pxi":3394
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __PYX_XDEC_MEMVIEW(&__pyx_t_16, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_17, 1);
  __Pyx_AddTraceback("pandas._libs.interval.Uint64ClosedBothIntervalNode.query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_indices, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_values, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_9__pyx_fuse_0query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_0__pyx_mdef_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_9__pyx_fuse_0query = {"__pyx_fuse_0query", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_9__pyx_fuse_0query, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_2query};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_9__pyx_fuse_0query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result = 0;
  __pyx_t_5numpy_float64_t __pyx_v_point;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fuse_0query (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_result,&__pyx_n_s_point,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_result)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_point)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_0query", 1, 2, 2, 1); __PYX_ERR(1, 3394, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_0query") < 0)) __PYX_ERR(1, 3394, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_result = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)values[0]);
    __pyx_v_point = __pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_point == ((npy_float64)-1)) && PyErr_Occurred())) __PYX_ERR(1, 3394, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_0query", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 3394, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Uint64ClosedBothIntervalNode.__pyx_fuse_0query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_result), __pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector, 1, "result", 0))) __PYX_ERR(1, 3394, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_8__pyx_fuse_0query(((struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode *)__pyx_v_self), __pyx_v_result, __pyx_v_point);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_8__pyx_fuse_0query(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float64_t __pyx_v_point) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__pyx_fuse_0query", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_fuse_0__pyx_f_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_query(__pyx_v_self, __pyx_v_result, __pyx_v_point, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3394, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Uint64ClosedBothIntervalNode.__pyx_fuse_0query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_11__pyx_fuse_1query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_fuse_1__pyx_f_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float32_t __pyx_v_point, int __pyx_skip_dispatch) {
  __Pyx_memviewslice __pyx_v_indices = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_values = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __pyx_t_5numpy_int64_t __pyx_t_9;
  __pyx_t_5numpy_int64_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  int __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  __Pyx_memviewslice __pyx_t_16 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_17 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  Py_ssize_t __pyx_t_21;
  __Pyx_RefNannySetupContext("__pyx_fuse_1query", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pyx_fuse_1query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3394, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_11__pyx_fuse_1query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = PyFloat_FromDouble(__pyx_v_point); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 3394, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3394, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3394, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 3394, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_5) {
            __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3394, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3394, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_3query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_4 = PyFloat_FromDouble(__pyx_v_point); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 3394, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_7 = __pyx_t_1; __pyx_t_3 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3394, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3394, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 3394, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (__pyx_t_3) {
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_t_4);
          __pyx_t_4 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3394, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pandas/_libs/intervaltree.pxi":3403
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
  __pyx_t_8 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_8) {

    /* "pandas/_libs/intervaltree.pxi":3407
 *             # continue the binary tree structure. Instead, we use linear
 *             # search.
 *             for i in range(self.n_elements):             # <<<<<<<<<<<<<<
 *                 if self.left[i] <= point <= self.right[i]:
 *                     result.append(self.indices[i])
 */
    __pyx_t_9 = __pyx_v_self->n_elements;
    __pyx_t_10 = __pyx_t_9;
    for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
      __pyx_v_i = __pyx_t_11;

      /* "pandas/_libs/intervaltree.pxi":3408
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point <= self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      __pyx_t_12 = __pyx_v_i;
      __pyx_t_8 = ((*((__pyx_t_5numpy_uint64_t *) ( /* dim=0 */ (__pyx_v_self->left.data + __pyx_t_12 * __pyx_v_self->left.strides[0]) ))) <= __pyx_v_point);
      if (__pyx_t_8) {
        __pyx_t_13 = __pyx_v_i;
        __pyx_t_8 = (__pyx_v_point <= (*((__pyx_t_5numpy_uint64_t *) ( /* dim=0 */ (__pyx_v_self->right.data + __pyx_t_13 * __pyx_v_self->right.strides[0]) ))));
      }
      __pyx_t_14 = (__pyx_t_8 != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":3409
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point <= self.right[i]:
 *                     result.append(self.indices[i])             # <<<<<<<<<<<<<<
 *         else:
 *             # There are child nodes. Based on comparing our query to the pivot,
 */
        __pyx_t_15 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->indices.data + __pyx_t_15 * __pyx_v_self->indices.strides[0]) ))));

        /* "pandas/_libs/intervaltree.pxi":3408
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point <= self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      }
    }

    /* "pandas/_libs/intervaltree.pxi":3403
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
    goto __pyx_L3;
  }

  /* "pandas/_libs/intervaltree.pxi":3413
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
  /*else*/ {
    __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":3414
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:
 *                 values = self.center_left_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 */
      __pyx_t_16 = __pyx_v_self->center_left_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":3415
 *             if point < self.pivot:
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:
 */
      __pyx_t_17 = __pyx_v_self->center_left_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":3416
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):             # <<<<<<<<<<<<<<
 *                     if not values[i] <= point:
 *                         break
 */
      __pyx_t_9 = __pyx_v_self->n_center;
      __pyx_t_10 = __pyx_t_9;
      for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":3417
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_18 = __pyx_v_i;
        __pyx_t_14 = ((!(((*((__pyx_t_5numpy_uint64_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_18 * __pyx_v_values.strides[0]) ))) <= __pyx_v_point) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":3418
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:
 */
          goto __pyx_L9_break;

          /* "pandas/_libs/intervaltree.pxi":3417
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":3419
 *                     if not values[i] <= point:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 */
        __pyx_t_19 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_19 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L9_break:;

      /* "pandas/_libs/intervaltree.pxi":3420
 *                         break
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      __pyx_t_14 = ((__pyx_v_point <= __pyx_v_self->left_node->max_right) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":3421
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)             # <<<<<<<<<<<<<<
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode *)__pyx_v_self->left_node->__pyx_vtab)->__pyx_fuse_1query(__pyx_v_self->left_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3421, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":3420
 *                         break
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":3413
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":3422
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
    __pyx_t_14 = ((__pyx_v_point > __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":3423
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 *                 values = self.center_right_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 */
      __pyx_t_16 = __pyx_v_self->center_right_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":3424
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:
 */
      __pyx_t_17 = __pyx_v_self->center_right_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":3425
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):             # <<<<<<<<<<<<<<
 *                     if not point <= values[i]:
 *                         break
 */
      for (__pyx_t_11 = (__pyx_v_self->n_center - 1); __pyx_t_11 > -1L; __pyx_t_11-=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":3426
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_20 = __pyx_v_i;
        __pyx_t_14 = ((!((__pyx_v_point <= (*((__pyx_t_5numpy_uint64_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_20 * __pyx_v_values.strides[0]) )))) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":3427
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:
 */
          goto __pyx_L13_break;

          /* "pandas/_libs/intervaltree.pxi":3426
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":3428
 *                     if not point <= values[i]:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if self.right_node.min_left <= point:
 *                     self.right_node.query(result, point)
 */
        __pyx_t_21 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_21 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L13_break:;

      /* "pandas/_libs/intervaltree.pxi":3429
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      __pyx_t_14 = ((__pyx_v_self->right_node->min_left <= __pyx_v_point) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":3430
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:
 *                     self.right_node.query(result, point)             # <<<<<<<<<<<<<<
 *             else:
 *                 result.extend(self.center_left_indices)
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode *)__pyx_v_self->right_node->__pyx_vtab)->__pyx_fuse_1query(__pyx_v_self->right_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3430, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":3429
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":3422
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":3432
 *                     self.right_node.query(result, point)
 *             else:
 *                 result.extend(self.center_left_indices)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
    /*else*/ {
      __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->extend(__pyx_v_result, __pyx_v_self->center_left_indices); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3432, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_L7:;
  }
  __pyx_L3:;

  /* "pandas/_libs/intervaltree.pxi":3394
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __PYX_XDEC_MEMVIEW(&__pyx_t_16, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_17, 1);
  __Pyx_AddTraceback("pandas._libs.interval.Uint64ClosedBothIntervalNode.query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_indices, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_values, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_11__pyx_fuse_1query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_1__pyx_mdef_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_11__pyx_fuse_1query = {"__pyx_fuse_1query", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_11__pyx_fuse_1query, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_2query};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_11__pyx_fuse_1query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result = 0;
  __pyx_t_5numpy_float32_t __pyx_v_point;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fuse_1query (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_result,&__pyx_n_s_point,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_result)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_point)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_1query", 1, 2, 2, 1); __PYX_ERR(1, 3394, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_1query") < 0)) __PYX_ERR(1, 3394, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_result = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)values[0]);
    __pyx_v_point = __pyx_PyFloat_AsFloat(values[1]); if (unlikely((__pyx_v_point == ((npy_float32)-1)) && PyErr_Occurred())) __PYX_ERR(1, 3394, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_1query", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 3394, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Uint64ClosedBothIntervalNode.__pyx_fuse_1query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_result), __pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector, 1, "result", 0))) __PYX_ERR(1, 3394, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_10__pyx_fuse_1query(((struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode *)__pyx_v_self), __pyx_v_result, __pyx_v_point);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_10__pyx_fuse_1query(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float32_t __pyx_v_point) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__pyx_fuse_1query", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_fuse_1__pyx_f_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_query(__pyx_v_self, __pyx_v_result, __pyx_v_point, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3394, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Uint64ClosedBothIntervalNode.__pyx_fuse_1query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_13__pyx_fuse_2query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_fuse_2__pyx_f_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int64_t __pyx_v_point, int __pyx_skip_dispatch) {
  __Pyx_memviewslice __pyx_v_indices = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_values = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __pyx_t_5numpy_int64_t __pyx_t_9;
  __pyx_t_5numpy_int64_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  int __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  __Pyx_memviewslice __pyx_t_16 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_17 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  Py_ssize_t __pyx_t_21;
  __Pyx_RefNannySetupContext("__pyx_fuse_2query", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pyx_fuse_2query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3394, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_13__pyx_fuse_2query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = __Pyx_PyInt_From_npy_int64(__pyx_v_point); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 3394, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3394, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3394, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 3394, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_5) {
            __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3394, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3394, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_3query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_4 = __Pyx_PyInt_From_npy_int64(__pyx_v_point); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 3394, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_7 = __pyx_t_1; __pyx_t_3 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3394, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3394, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 3394, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (__pyx_t_3) {
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_t_4);
          __pyx_t_4 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3394, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pandas/_libs/intervaltree.pxi":3403
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
  __pyx_t_8 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_8) {

    /* "pandas/_libs/intervaltree.pxi":3407
 *             # continue the binary tree structure. Instead, we use linear
 *             # search.
 *             for i in range(self.n_elements):             # <<<<<<<<<<<<<<
 *                 if self.left[i] <= point <= self.right[i]:
 *                     result.append(self.indices[i])
 */
    __pyx_t_9 = __pyx_v_self->n_elements;
    __pyx_t_10 = __pyx_t_9;
    for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
      __pyx_v_i = __pyx_t_11;

      /* "pandas/_libs/intervaltree.pxi":3408
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point <= self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      __pyx_t_12 = __pyx_v_i;
      __pyx_t_8 = ((*((__pyx_t_5numpy_uint64_t *) ( /* dim=0 */ (__pyx_v_self->left.data + __pyx_t_12 * __pyx_v_self->left.strides[0]) ))) <= __pyx_v_point);
      if (__pyx_t_8) {
        __pyx_t_13 = __pyx_v_i;
        __pyx_t_8 = (__pyx_v_point <= (*((__pyx_t_5numpy_uint64_t *) ( /* dim=0 */ (__pyx_v_self->right.data + __pyx_t_13 * __pyx_v_self->right.strides[0]) ))));
      }
      __pyx_t_14 = (__pyx_t_8 != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":3409
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point <= self.right[i]:
 *                     result.append(self.indices[i])             # <<<<<<<<<<<<<<
 *         else:
 *             # There are child nodes. Based on comparing our query to the pivot,
 */
        __pyx_t_15 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->indices.data + __pyx_t_15 * __pyx_v_self->indices.strides[0]) ))));

        /* "pandas/_libs/intervaltree.pxi":3408
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point <= self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      }
    }

    /* "pandas/_libs/intervaltree.pxi":3403
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
    goto __pyx_L3;
  }

  /* "pandas/_libs/intervaltree.pxi":3413
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
  /*else*/ {
    __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":3414
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:
 *                 values = self.center_left_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 */
      __pyx_t_16 = __pyx_v_self->center_left_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":3415
 *             if point < self.pivot:
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:
 */
      __pyx_t_17 = __pyx_v_self->center_left_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":3416
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):             # <<<<<<<<<<<<<<
 *                     if not values[i] <= point:
 *                         break
 */
      __pyx_t_9 = __pyx_v_self->n_center;
      __pyx_t_10 = __pyx_t_9;
      for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":3417
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_18 = __pyx_v_i;
        __pyx_t_14 = ((!(((*((__pyx_t_5numpy_uint64_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_18 * __pyx_v_values.strides[0]) ))) <= __pyx_v_point) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":3418
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:
 */
          goto __pyx_L9_break;

          /* "pandas/_libs/intervaltree.pxi":3417
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":3419
 *                     if not values[i] <= point:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 */
        __pyx_t_19 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_19 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L9_break:;

      /* "pandas/_libs/intervaltree.pxi":3420
 *                         break
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      __pyx_t_14 = ((__pyx_v_point <= __pyx_v_self->left_node->max_right) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":3421
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)             # <<<<<<<<<<<<<<
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode *)__pyx_v_self->left_node->__pyx_vtab)->__pyx_fuse_2query(__pyx_v_self->left_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3421, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":3420
 *                         break
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":3413
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":3422
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
    __pyx_t_14 = ((__pyx_v_point > __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":3423
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 *                 values = self.center_right_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 */
      __pyx_t_16 = __pyx_v_self->center_right_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":3424
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:
 */
      __pyx_t_17 = __pyx_v_self->center_right_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":3425
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):             # <<<<<<<<<<<<<<
 *                     if not point <= values[i]:
 *                         break
 */
      for (__pyx_t_11 = (__pyx_v_self->n_center - 1); __pyx_t_11 > -1L; __pyx_t_11-=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":3426
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_20 = __pyx_v_i;
        __pyx_t_14 = ((!((__pyx_v_point <= (*((__pyx_t_5numpy_uint64_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_20 * __pyx_v_values.strides[0]) )))) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":3427
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:
 */
          goto __pyx_L13_break;

          /* "pandas/_libs/intervaltree.pxi":3426
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":3428
 *                     if not point <= values[i]:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if self.right_node.min_left <= point:
 *                     self.right_node.query(result, point)
 */
        __pyx_t_21 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_21 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L13_break:;

      /* "pandas/_libs/intervaltree.pxi":3429
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      __pyx_t_14 = ((__pyx_v_self->right_node->min_left <= __pyx_v_point) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":3430
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:
 *                     self.right_node.query(result, point)             # <<<<<<<<<<<<<<
 *             else:
 *                 result.extend(self.center_left_indices)
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode *)__pyx_v_self->right_node->__pyx_vtab)->__pyx_fuse_2query(__pyx_v_self->right_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3430, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":3429
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":3422
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":3432
 *                     self.right_node.query(result, point)
 *             else:
 *                 result.extend(self.center_left_indices)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
    /*else*/ {
      __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->extend(__pyx_v_result, __pyx_v_self->center_left_indices); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3432, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_L7:;
  }
  __pyx_L3:;

  /* "pandas/_libs/intervaltree.pxi":3394
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __PYX_XDEC_MEMVIEW(&__pyx_t_16, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_17, 1);
  __Pyx_AddTraceback("pandas._libs.interval.Uint64ClosedBothIntervalNode.query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_indices, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_values, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_13__pyx_fuse_2query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_2__pyx_mdef_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_13__pyx_fuse_2query = {"__pyx_fuse_2query", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_13__pyx_fuse_2query, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_2query};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_13__pyx_fuse_2query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result = 0;
  __pyx_t_5numpy_int64_t __pyx_v_point;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fuse_2query (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_result,&__pyx_n_s_point,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_result)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_point)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_2query", 1, 2, 2, 1); __PYX_ERR(1, 3394, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_2query") < 0)) __PYX_ERR(1, 3394, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_result = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)values[0]);
    __pyx_v_point = __Pyx_PyInt_As_npy_int64(values[1]); if (unlikely((__pyx_v_point == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(1, 3394, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_2query", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 3394, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Uint64ClosedBothIntervalNode.__pyx_fuse_2query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_result), __pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector, 1, "result", 0))) __PYX_ERR(1, 3394, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_12__pyx_fuse_2query(((struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode *)__pyx_v_self), __pyx_v_result, __pyx_v_point);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_12__pyx_fuse_2query(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int64_t __pyx_v_point) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__pyx_fuse_2query", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_fuse_2__pyx_f_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_query(__pyx_v_self, __pyx_v_result, __pyx_v_point, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3394, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Uint64ClosedBothIntervalNode.__pyx_fuse_2query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_15__pyx_fuse_3query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_fuse_3__pyx_f_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int32_t __pyx_v_point, int __pyx_skip_dispatch) {
  __Pyx_memviewslice __pyx_v_indices = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_values = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __pyx_t_5numpy_int64_t __pyx_t_9;
  __pyx_t_5numpy_int64_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  int __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  __Pyx_memviewslice __pyx_t_16 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_17 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  Py_ssize_t __pyx_t_21;
  __Pyx_RefNannySetupContext("__pyx_fuse_3query", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pyx_fuse_3query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3394, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_15__pyx_fuse_3query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = __Pyx_PyInt_From_npy_int32(__pyx_v_point); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 3394, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3394, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3394, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 3394, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_5) {
            __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3394, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3394, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_3query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_4 = __Pyx_PyInt_From_npy_int32(__pyx_v_point); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 3394, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_7 = __pyx_t_1; __pyx_t_3 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3394, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3394, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 3394, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (__pyx_t_3) {
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_t_4);
          __pyx_t_4 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3394, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pandas/_libs/intervaltree.pxi":3403
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
  __pyx_t_8 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_8) {

    /* "pandas/_libs/intervaltree.pxi":3407
 *             # continue the binary tree structure. Instead, we use linear
 *             # search.
 *             for i in range(self.n_elements):             # <<<<<<<<<<<<<<
 *                 if self.left[i] <= point <= self.right[i]:
 *                     result.append(self.indices[i])
 */
    __pyx_t_9 = __pyx_v_self->n_elements;
    __pyx_t_10 = __pyx_t_9;
    for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
      __pyx_v_i = __pyx_t_11;

      /* "pandas/_libs/intervaltree.pxi":3408
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point <= self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      __pyx_t_12 = __pyx_v_i;
      __pyx_t_8 = ((*((__pyx_t_5numpy_uint64_t *) ( /* dim=0 */ (__pyx_v_self->left.data + __pyx_t_12 * __pyx_v_self->left.strides[0]) ))) <= __pyx_v_point);
      if (__pyx_t_8) {
        __pyx_t_13 = __pyx_v_i;
        __pyx_t_8 = (__pyx_v_point <= (*((__pyx_t_5numpy_uint64_t *) ( /* dim=0 */ (__pyx_v_self->right.data + __pyx_t_13 * __pyx_v_self->right.strides[0]) ))));
      }
      __pyx_t_14 = (__pyx_t_8 != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":3409
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point <= self.right[i]:
 *                     result.append(self.indices[i])             # <<<<<<<<<<<<<<
 *         else:
 *             # There are child nodes. Based on comparing our query to the pivot,
 */
        __pyx_t_15 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->indices.data + __pyx_t_15 * __pyx_v_self->indices.strides[0]) ))));

        /* "pandas/_libs/intervaltree.pxi":3408
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point <= self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      }
    }

    /* "pandas/_libs/intervaltree.pxi":3403
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
    goto __pyx_L3;
  }

  /* "pandas/_libs/intervaltree.pxi":3413
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
  /*else*/ {
    __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":3414
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:
 *                 values = self.center_left_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 */
      __pyx_t_16 = __pyx_v_self->center_left_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":3415
 *             if point < self.pivot:
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:
 */
      __pyx_t_17 = __pyx_v_self->center_left_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":3416
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):             # <<<<<<<<<<<<<<
 *                     if not values[i] <= point:
 *                         break
 */
      __pyx_t_9 = __pyx_v_self->n_center;
      __pyx_t_10 = __pyx_t_9;
      for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":3417
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_18 = __pyx_v_i;
        __pyx_t_14 = ((!(((*((__pyx_t_5numpy_uint64_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_18 * __pyx_v_values.strides[0]) ))) <= __pyx_v_point) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":3418
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:
 */
          goto __pyx_L9_break;

          /* "pandas/_libs/intervaltree.pxi":3417
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":3419
 *                     if not values[i] <= point:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 */
        __pyx_t_19 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_19 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L9_break:;

      /* "pandas/_libs/intervaltree.pxi":3420
 *                         break
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      __pyx_t_14 = ((__pyx_v_point <= __pyx_v_self->left_node->max_right) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":3421
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)             # <<<<<<<<<<<<<<
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode *)__pyx_v_self->left_node->__pyx_vtab)->__pyx_fuse_3query(__pyx_v_self->left_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3421, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":3420
 *                         break
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":3413
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":3422
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
    __pyx_t_14 = ((__pyx_v_point > __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":3423
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 *                 values = self.center_right_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 */
      __pyx_t_16 = __pyx_v_self->center_right_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":3424
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:
 */
      __pyx_t_17 = __pyx_v_self->center_right_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":3425
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):             # <<<<<<<<<<<<<<
 *                     if not point <= values[i]:
 *                         break
 */
      for (__pyx_t_11 = (__pyx_v_self->n_center - 1); __pyx_t_11 > -1L; __pyx_t_11-=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":3426
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_20 = __pyx_v_i;
        __pyx_t_14 = ((!((__pyx_v_point <= (*((__pyx_t_5numpy_uint64_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_20 * __pyx_v_values.strides[0]) )))) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":3427
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:
 */
          goto __pyx_L13_break;

          /* "pandas/_libs/intervaltree.pxi":3426
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":3428
 *                     if not point <= values[i]:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if self.right_node.min_left <= point:
 *                     self.right_node.query(result, point)
 */
        __pyx_t_21 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_21 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L13_break:;

      /* "pandas/_libs/intervaltree.pxi":3429
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      __pyx_t_14 = ((__pyx_v_self->right_node->min_left <= __pyx_v_point) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":3430
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:
 *                     self.right_node.query(result, point)             # <<<<<<<<<<<<<<
 *             else:
 *                 result.extend(self.center_left_indices)
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode *)__pyx_v_self->right_node->__pyx_vtab)->__pyx_fuse_3query(__pyx_v_self->right_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3430, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":3429
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":3422
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":3432
 *                     self.right_node.query(result, point)
 *             else:
 *                 result.extend(self.center_left_indices)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
    /*else*/ {
      __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->extend(__pyx_v_result, __pyx_v_self->center_left_indices); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3432, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_L7:;
  }
  __pyx_L3:;

  /* "pandas/_libs/intervaltree.pxi":3394
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __PYX_XDEC_MEMVIEW(&__pyx_t_16, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_17, 1);
  __Pyx_AddTraceback("pandas._libs.interval.Uint64ClosedBothIntervalNode.query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_indices, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_values, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_15__pyx_fuse_3query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_3__pyx_mdef_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_15__pyx_fuse_3query = {"__pyx_fuse_3query", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_15__pyx_fuse_3query, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_2query};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_15__pyx_fuse_3query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result = 0;
  __pyx_t_5numpy_int32_t __pyx_v_point;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fuse_3query (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_result,&__pyx_n_s_point,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_result)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_point)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_3query", 1, 2, 2, 1); __PYX_ERR(1, 3394, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_3query") < 0)) __PYX_ERR(1, 3394, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_result = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)values[0]);
    __pyx_v_point = __Pyx_PyInt_As_npy_int32(values[1]); if (unlikely((__pyx_v_point == ((npy_int32)-1)) && PyErr_Occurred())) __PYX_ERR(1, 3394, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_3query", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 3394, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Uint64ClosedBothIntervalNode.__pyx_fuse_3query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_result), __pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector, 1, "result", 0))) __PYX_ERR(1, 3394, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_14__pyx_fuse_3query(((struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode *)__pyx_v_self), __pyx_v_result, __pyx_v_point);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_14__pyx_fuse_3query(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int32_t __pyx_v_point) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__pyx_fuse_3query", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_fuse_3__pyx_f_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_query(__pyx_v_self, __pyx_v_result, __pyx_v_point, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3394, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Uint64ClosedBothIntervalNode.__pyx_fuse_3query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_17__pyx_fuse_4query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_fuse_4__pyx_f_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_uint64_t __pyx_v_point, int __pyx_skip_dispatch) {
  __Pyx_memviewslice __pyx_v_indices = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_values = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __pyx_t_5numpy_int64_t __pyx_t_9;
  __pyx_t_5numpy_int64_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  int __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  __Pyx_memviewslice __pyx_t_16 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_17 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  Py_ssize_t __pyx_t_21;
  __Pyx_RefNannySetupContext("__pyx_fuse_4query", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pyx_fuse_4query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3394, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_17__pyx_fuse_4query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = __Pyx_PyInt_From_npy_uint64(__pyx_v_point); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 3394, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3394, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3394, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 3394, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_5) {
            __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3394, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3394, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_3query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_4 = __Pyx_PyInt_From_npy_uint64(__pyx_v_point); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 3394, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_7 = __pyx_t_1; __pyx_t_3 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3394, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3394, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 3394, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (__pyx_t_3) {
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_t_4);
          __pyx_t_4 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3394, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pandas/_libs/intervaltree.pxi":3403
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
  __pyx_t_8 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_8) {

    /* "pandas/_libs/intervaltree.pxi":3407
 *             # continue the binary tree structure. Instead, we use linear
 *             # search.
 *             for i in range(self.n_elements):             # <<<<<<<<<<<<<<
 *                 if self.left[i] <= point <= self.right[i]:
 *                     result.append(self.indices[i])
 */
    __pyx_t_9 = __pyx_v_self->n_elements;
    __pyx_t_10 = __pyx_t_9;
    for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
      __pyx_v_i = __pyx_t_11;

      /* "pandas/_libs/intervaltree.pxi":3408
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point <= self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      __pyx_t_12 = __pyx_v_i;
      __pyx_t_8 = ((*((__pyx_t_5numpy_uint64_t *) ( /* dim=0 */ (__pyx_v_self->left.data + __pyx_t_12 * __pyx_v_self->left.strides[0]) ))) <= __pyx_v_point);
      if (__pyx_t_8) {
        __pyx_t_13 = __pyx_v_i;
        __pyx_t_8 = (__pyx_v_point <= (*((__pyx_t_5numpy_uint64_t *) ( /* dim=0 */ (__pyx_v_self->right.data + __pyx_t_13 * __pyx_v_self->right.strides[0]) ))));
      }
      __pyx_t_14 = (__pyx_t_8 != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":3409
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point <= self.right[i]:
 *                     result.append(self.indices[i])             # <<<<<<<<<<<<<<
 *         else:
 *             # There are child nodes. Based on comparing our query to the pivot,
 */
        __pyx_t_15 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->indices.data + __pyx_t_15 * __pyx_v_self->indices.strides[0]) ))));

        /* "pandas/_libs/intervaltree.pxi":3408
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] <= point <= self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      }
    }

    /* "pandas/_libs/intervaltree.pxi":3403
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
    goto __pyx_L3;
  }

  /* "pandas/_libs/intervaltree.pxi":3413
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
  /*else*/ {
    __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":3414
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:
 *                 values = self.center_left_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 */
      __pyx_t_16 = __pyx_v_self->center_left_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":3415
 *             if point < self.pivot:
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:
 */
      __pyx_t_17 = __pyx_v_self->center_left_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":3416
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):             # <<<<<<<<<<<<<<
 *                     if not values[i] <= point:
 *                         break
 */
      __pyx_t_9 = __pyx_v_self->n_center;
      __pyx_t_10 = __pyx_t_9;
      for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":3417
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_18 = __pyx_v_i;
        __pyx_t_14 = ((!(((*((__pyx_t_5numpy_uint64_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_18 * __pyx_v_values.strides[0]) ))) <= __pyx_v_point) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":3418
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:
 */
          goto __pyx_L9_break;

          /* "pandas/_libs/intervaltree.pxi":3417
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] <= point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":3419
 *                     if not values[i] <= point:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 */
        __pyx_t_19 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_19 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L9_break:;

      /* "pandas/_libs/intervaltree.pxi":3420
 *                         break
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      __pyx_t_14 = ((__pyx_v_point <= __pyx_v_self->left_node->max_right) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":3421
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)             # <<<<<<<<<<<<<<
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode *)__pyx_v_self->left_node->__pyx_vtab)->__pyx_fuse_4query(__pyx_v_self->left_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3421, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":3420
 *                         break
 *                     result.append(indices[i])
 *                 if point <= self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":3413
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":3422
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
    __pyx_t_14 = ((__pyx_v_point > __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":3423
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 *                 values = self.center_right_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 */
      __pyx_t_16 = __pyx_v_self->center_right_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":3424
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:
 */
      __pyx_t_17 = __pyx_v_self->center_right_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":3425
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):             # <<<<<<<<<<<<<<
 *                     if not point <= values[i]:
 *                         break
 */
      for (__pyx_t_11 = (__pyx_v_self->n_center - 1); __pyx_t_11 > -1L; __pyx_t_11-=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":3426
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_20 = __pyx_v_i;
        __pyx_t_14 = ((!((__pyx_v_point <= (*((__pyx_t_5numpy_uint64_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_20 * __pyx_v_values.strides[0]) )))) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":3427
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:
 */
          goto __pyx_L13_break;

          /* "pandas/_libs/intervaltree.pxi":3426
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point <= values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":3428
 *                     if not point <= values[i]:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if self.right_node.min_left <= point:
 *                     self.right_node.query(result, point)
 */
        __pyx_t_21 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_21 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L13_break:;

      /* "pandas/_libs/intervaltree.pxi":3429
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      __pyx_t_14 = ((__pyx_v_self->right_node->min_left <= __pyx_v_point) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":3430
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:
 *                     self.right_node.query(result, point)             # <<<<<<<<<<<<<<
 *             else:
 *                 result.extend(self.center_left_indices)
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode *)__pyx_v_self->right_node->__pyx_vtab)->__pyx_fuse_4query(__pyx_v_self->right_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3430, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":3429
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left <= point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":3422
 *                 if point <= self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":3432
 *                     self.right_node.query(result, point)
 *             else:
 *                 result.extend(self.center_left_indices)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
    /*else*/ {
      __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->extend(__pyx_v_result, __pyx_v_self->center_left_indices); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3432, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_L7:;
  }
  __pyx_L3:;

  /* "pandas/_libs/intervaltree.pxi":3394
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __PYX_XDEC_MEMVIEW(&__pyx_t_16, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_17, 1);
  __Pyx_AddTraceback("pandas._libs.interval.Uint64ClosedBothIntervalNode.query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_indices, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_values, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_17__pyx_fuse_4query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_4__pyx_mdef_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_17__pyx_fuse_4query = {"__pyx_fuse_4query", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_17__pyx_fuse_4query, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_2query};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_17__pyx_fuse_4query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result = 0;
  __pyx_t_5numpy_uint64_t __pyx_v_point;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fuse_4query (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_result,&__pyx_n_s_point,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_result)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_point)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_4query", 1, 2, 2, 1); __PYX_ERR(1, 3394, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_4query") < 0)) __PYX_ERR(1, 3394, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_result = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)values[0]);
    __pyx_v_point = __Pyx_PyInt_As_npy_uint64(values[1]); if (unlikely((__pyx_v_point == ((npy_uint64)-1)) && PyErr_Occurred())) __PYX_ERR(1, 3394, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_4query", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 3394, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Uint64ClosedBothIntervalNode.__pyx_fuse_4query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_result), __pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector, 1, "result", 0))) __PYX_ERR(1, 3394, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_16__pyx_fuse_4query(((struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode *)__pyx_v_self), __pyx_v_result, __pyx_v_point);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_16__pyx_fuse_4query(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_uint64_t __pyx_v_point) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__pyx_fuse_4query", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_fuse_4__pyx_f_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_query(__pyx_v_self, __pyx_v_result, __pyx_v_point, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3394, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Uint64ClosedBothIntervalNode.__pyx_fuse_4query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":3434
 *                 result.extend(self.center_left_indices)
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         if self.is_leaf_node:
 *             return ('<Uint64ClosedBothIntervalNode: '
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_5__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_5__repr__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_4__repr__(((struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_4__repr__(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode *__pyx_v_self) {
  PyObject *__pyx_v_n_left = NULL;
  PyObject *__pyx_v_n_right = NULL;
  PyObject *__pyx_v_n_center = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  Py_UCS4 __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "pandas/_libs/intervaltree.pxi":3435
 * 
 *     def __repr__(self):
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             return ('<Uint64ClosedBothIntervalNode: '
 *                     '%s elements (terminal)>' % self.n_elements)
 */
  __pyx_t_1 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_1) {

    /* "pandas/_libs/intervaltree.pxi":3436
 *     def __repr__(self):
 *         if self.is_leaf_node:
 *             return ('<Uint64ClosedBothIntervalNode: '             # <<<<<<<<<<<<<<
 *                     '%s elements (terminal)>' % self.n_elements)
 *         else:
 */
    __Pyx_XDECREF(__pyx_r);

    /* "pandas/_libs/intervaltree.pxi":3437
 *         if self.is_leaf_node:
 *             return ('<Uint64ClosedBothIntervalNode: '
 *                     '%s elements (terminal)>' % self.n_elements)             # <<<<<<<<<<<<<<
 *         else:
 *             n_left = self.left_node.n_elements
 */
    __pyx_t_2 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_elements); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3437, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PyUnicode_Format(__pyx_kp_u_Uint64ClosedBothIntervalNode_s, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 3437, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pandas/_libs/intervaltree.pxi":3435
 * 
 *     def __repr__(self):
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             return ('<Uint64ClosedBothIntervalNode: '
 *                     '%s elements (terminal)>' % self.n_elements)
 */
  }

  /* "pandas/_libs/intervaltree.pxi":3439
 *                     '%s elements (terminal)>' % self.n_elements)
 *         else:
 *             n_left = self.left_node.n_elements             # <<<<<<<<<<<<<<
 *             n_right = self.right_node.n_elements
 *             n_center = self.n_elements - n_left - n_right
 */
  /*else*/ {
    __pyx_t_3 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->left_node->n_elements); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 3439, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_v_n_left = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "pandas/_libs/intervaltree.pxi":3440
 *         else:
 *             n_left = self.left_node.n_elements
 *             n_right = self.right_node.n_elements             # <<<<<<<<<<<<<<
 *             n_center = self.n_elements - n_left - n_right
 *             return ('<Uint64ClosedBothIntervalNode: '
 */
    __pyx_t_3 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->right_node->n_elements); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 3440, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_v_n_right = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "pandas/_libs/intervaltree.pxi":3441
 *             n_left = self.left_node.n_elements
 *             n_right = self.right_node.n_elements
 *             n_center = self.n_elements - n_left - n_right             # <<<<<<<<<<<<<<
 *             return ('<Uint64ClosedBothIntervalNode: '
 *                     'pivot %s, %s elements (%s left, %s right, %s '
 */
    __pyx_t_3 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_elements); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 3441, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = PyNumber_Subtract(__pyx_t_3, __pyx_v_n_left); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3441, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyNumber_Subtract(__pyx_t_2, __pyx_v_n_right); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 3441, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_n_center = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "pandas/_libs/intervaltree.pxi":3442
 *             n_right = self.right_node.n_elements
 *             n_center = self.n_elements - n_left - n_right
 *             return ('<Uint64ClosedBothIntervalNode: '             # <<<<<<<<<<<<<<
 *                     'pivot %s, %s elements (%s left, %s right, %s '
 *                     'overlapping)>' % (self.pivot, self.n_elements,
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = PyTuple_New(11); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 3442, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = 0;
    __pyx_t_5 = 127;
    __Pyx_INCREF(__pyx_kp_u_Uint64ClosedBothIntervalNode_pi);
    __pyx_t_4 += 37;
    __Pyx_GIVEREF(__pyx_kp_u_Uint64ClosedBothIntervalNode_pi);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_kp_u_Uint64ClosedBothIntervalNode_pi);

    /* "pandas/_libs/intervaltree.pxi":3444
 *             return ('<Uint64ClosedBothIntervalNode: '
 *                     'pivot %s, %s elements (%s left, %s right, %s '
 *                     'overlapping)>' % (self.pivot, self.n_elements,             # <<<<<<<<<<<<<<
 *                                        n_left, n_right, n_center))
 * 
 */
    __pyx_t_2 = __Pyx_PyInt_From_npy_uint64(__pyx_v_self->pivot); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3444, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_t_2), __pyx_empty_unicode); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 3444, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_5 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_6) > __pyx_t_5) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_6) : __pyx_t_5;
    __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_6);
    __pyx_t_6 = 0;
    __Pyx_INCREF(__pyx_kp_u__18);
    __pyx_t_4 += 2;
    __Pyx_GIVEREF(__pyx_kp_u__18);
    PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_kp_u__18);
    __pyx_t_6 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_elements); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 3444, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_2 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_t_6), __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3444, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_5 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) > __pyx_t_5) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) : __pyx_t_5;
    __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_3, 3, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_elements);
    __pyx_t_4 += 11;
    __Pyx_GIVEREF(__pyx_kp_u_elements);
    PyTuple_SET_ITEM(__pyx_t_3, 4, __pyx_kp_u_elements);

    /* "pandas/_libs/intervaltree.pxi":3445
 *                     'pivot %s, %s elements (%s left, %s right, %s '
 *                     'overlapping)>' % (self.pivot, self.n_elements,
 *                                        n_left, n_right, n_center))             # <<<<<<<<<<<<<<
 * 
 *     def counts(self):
 */
    __pyx_t_2 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_v_n_left), __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3445, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) > __pyx_t_5) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) : __pyx_t_5;
    __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_3, 5, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_left_2);
    __pyx_t_4 += 7;
    __Pyx_GIVEREF(__pyx_kp_u_left_2);
    PyTuple_SET_ITEM(__pyx_t_3, 6, __pyx_kp_u_left_2);
    __pyx_t_2 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_v_n_right), __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3445, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) > __pyx_t_5) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) : __pyx_t_5;
    __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_3, 7, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_right_2);
    __pyx_t_4 += 8;
    __Pyx_GIVEREF(__pyx_kp_u_right_2);
    PyTuple_SET_ITEM(__pyx_t_3, 8, __pyx_kp_u_right_2);
    __pyx_t_2 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_v_n_center), __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3445, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) > __pyx_t_5) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) : __pyx_t_5;
    __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_3, 9, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_overlapping);
    __pyx_t_4 += 14;
    __Pyx_GIVEREF(__pyx_kp_u_overlapping);
    PyTuple_SET_ITEM(__pyx_t_3, 10, __pyx_kp_u_overlapping);

    /* "pandas/_libs/intervaltree.pxi":3442
 *             n_right = self.right_node.n_elements
 *             n_center = self.n_elements - n_left - n_right
 *             return ('<Uint64ClosedBothIntervalNode: '             # <<<<<<<<<<<<<<
 *                     'pivot %s, %s elements (%s left, %s right, %s '
 *                     'overlapping)>' % (self.pivot, self.n_elements,
 */
    __pyx_t_2 = __Pyx_PyUnicode_Join(__pyx_t_3, 11, __pyx_t_4, __pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3442, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;
  }

  /* "pandas/_libs/intervaltree.pxi":3434
 *                 result.extend(self.center_left_indices)
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         if self.is_leaf_node:
 *             return ('<Uint64ClosedBothIntervalNode: '
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pandas._libs.interval.Uint64ClosedBothIntervalNode.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_n_left);
  __Pyx_XDECREF(__pyx_v_n_right);
  __Pyx_XDECREF(__pyx_v_n_center);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":3447
 *                                        n_left, n_right, n_center))
 * 
 *     def counts(self):             # <<<<<<<<<<<<<<
 *         """
 *         Inspect counts on this node
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_7counts(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_6counts[] = "\n        Inspect counts on this node\n        useful for debugging purposes\n        ";
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_7counts(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("counts (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_6counts(((struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_6counts(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode *__pyx_v_self) {
  Py_ssize_t __pyx_v_m;
  PyObject *__pyx_v_l = NULL;
  PyObject *__pyx_v_r = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  size_t __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("counts", 0);

  /* "pandas/_libs/intervaltree.pxi":3452
 *         useful for debugging purposes
 *         """
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             return self.n_elements
 *         else:
 */
  __pyx_t_1 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_1) {

    /* "pandas/_libs/intervaltree.pxi":3453
 *         """
 *         if self.is_leaf_node:
 *             return self.n_elements             # <<<<<<<<<<<<<<
 *         else:
 *             m = len(self.center_left_values)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_elements); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3453, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "pandas/_libs/intervaltree.pxi":3452
 *         useful for debugging purposes
 *         """
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             return self.n_elements
 *         else:
 */
  }

  /* "pandas/_libs/intervaltree.pxi":3455
 *             return self.n_elements
 *         else:
 *             m = len(self.center_left_values)             # <<<<<<<<<<<<<<
 *             l = self.left_node.counts()
 *             r = self.right_node.counts()
 */
  /*else*/ {
    if (unlikely(!__pyx_v_self->center_left_values.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(1, 3455, __pyx_L1_error)}
    __pyx_t_3 = __Pyx_MemoryView_Len(__pyx_v_self->center_left_values); 
    __pyx_v_m = __pyx_t_3;

    /* "pandas/_libs/intervaltree.pxi":3456
 *         else:
 *             m = len(self.center_left_values)
 *             l = self.left_node.counts()             # <<<<<<<<<<<<<<
 *             r = self.right_node.counts()
 *             return (m, (l, r))
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->left_node), __pyx_n_s_counts); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 3456, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3456, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_l = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "pandas/_libs/intervaltree.pxi":3457
 *             m = len(self.center_left_values)
 *             l = self.left_node.counts()
 *             r = self.right_node.counts()             # <<<<<<<<<<<<<<
 *             return (m, (l, r))
 * 
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->right_node), __pyx_n_s_counts); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 3457, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3457, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_r = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "pandas/_libs/intervaltree.pxi":3458
 *             l = self.left_node.counts()
 *             r = self.right_node.counts()
 *             return (m, (l, r))             # <<<<<<<<<<<<<<
 * 
 * NODE_CLASSES['uint64',
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = PyInt_FromSsize_t(__pyx_v_m); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3458, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 3458, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_v_l);
    __Pyx_GIVEREF(__pyx_v_l);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_l);
    __Pyx_INCREF(__pyx_v_r);
    __Pyx_GIVEREF(__pyx_v_r);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_v_r);
    __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 3458, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_4);
    __pyx_t_2 = 0;
    __pyx_t_4 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;
  }

  /* "pandas/_libs/intervaltree.pxi":3447
 *                                        n_left, n_right, n_center))
 * 
 *     def counts(self):             # <<<<<<<<<<<<<<
 *         """
 *         Inspect counts on this node
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pandas._libs.interval.Uint64ClosedBothIntervalNode.counts", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_l);
  __Pyx_XDECREF(__pyx_v_r);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":3304
 *         int64_t[:] center_left_indices, center_right_indices, indices
 *         uint64_t min_left, max_right
 *         readonly uint64_t pivot             # <<<<<<<<<<<<<<
 *         readonly int64_t n_elements, n_center, leaf_size
 *         readonly bint is_leaf_node
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_5pivot_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_5pivot_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_5pivot___get__(((struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_5pivot___get__(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_npy_uint64(__pyx_v_self->pivot); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3304, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Uint64ClosedBothIntervalNode.pivot.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":3305
 *         uint64_t min_left, max_right
 *         readonly uint64_t pivot
 *         readonly int64_t n_elements, n_center, leaf_size             # <<<<<<<<<<<<<<
 *         readonly bint is_leaf_node
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_10n_elements_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_10n_elements_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_10n_elements___get__(((struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_10n_elements___get__(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_elements); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3305, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Uint64ClosedBothIntervalNode.n_elements.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_8n_center_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_8n_center_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_8n_center___get__(((struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_8n_center___get__(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_center); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3305, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Uint64ClosedBothIntervalNode.n_center.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_9leaf_size_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_9leaf_size_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_9leaf_size___get__(((struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_9leaf_size___get__(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->leaf_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3305, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Uint64ClosedBothIntervalNode.leaf_size.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":3306
 *         readonly uint64_t pivot
 *         readonly int64_t n_elements, n_center, leaf_size
 *         readonly bint is_leaf_node             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self,
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_12is_leaf_node_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_12is_leaf_node_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_12is_leaf_node___get__(((struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_12is_leaf_node___get__(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->is_leaf_node); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3306, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Uint64ClosedBothIntervalNode.is_leaf_node.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_21__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_21__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_20__reduce_cython__(((struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_20__reduce_cython__(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  int __pyx_t_16;
  int __pyx_t_17;
  int __pyx_t_18;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.center_left_indices, self.center_left_values, self.center_right_indices, self.center_right_values, self.indices, self.is_leaf_node, self.leaf_size, self.left, self.left_node, self.max_right, self.min_left, self.n_center, self.n_elements, self.pivot, self.right, self.right_node)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  if (unlikely(!__pyx_v_self->center_left_indices.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_1 = __pyx_memoryview_fromslice(__pyx_v_self->center_left_indices, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_int64_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_int64_t, 0);; if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (unlikely(!__pyx_v_self->center_left_values.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_2 = __pyx_memoryview_fromslice(__pyx_v_self->center_left_values, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_uint64_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_uint64_t, 0);; if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (unlikely(!__pyx_v_self->center_right_indices.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_3 = __pyx_memoryview_fromslice(__pyx_v_self->center_right_indices, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_int64_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_int64_t, 0);; if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely(!__pyx_v_self->center_right_values.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_4 = __pyx_memoryview_fromslice(__pyx_v_self->center_right_values, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_uint64_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_uint64_t, 0);; if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (unlikely(!__pyx_v_self->indices.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_5 = __pyx_memoryview_fromslice(__pyx_v_self->indices, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_int64_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_int64_t, 0);; if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyBool_FromLong(__pyx_v_self->is_leaf_node); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->leaf_size); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (unlikely(!__pyx_v_self->left.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_8 = __pyx_memoryview_fromslice(__pyx_v_self->left, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_uint64_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_uint64_t, 0);; if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_PyInt_From_npy_uint64(__pyx_v_self->max_right); if (unlikely(!__pyx_t_9)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_10 = __Pyx_PyInt_From_npy_uint64(__pyx_v_self->min_left); if (unlikely(!__pyx_t_10)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_11 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_center); if (unlikely(!__pyx_t_11)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_12 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_elements); if (unlikely(!__pyx_t_12)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __pyx_t_13 = __Pyx_PyInt_From_npy_uint64(__pyx_v_self->pivot); if (unlikely(!__pyx_t_13)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  if (unlikely(!__pyx_v_self->right.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_14 = __pyx_memoryview_fromslice(__pyx_v_self->right, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_uint64_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_uint64_t, 0);; if (unlikely(!__pyx_t_14)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __pyx_t_15 = PyTuple_New(16); if (unlikely(!__pyx_t_15)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_15, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_15, 1, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_15, 2, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_15, 3, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_15, 4, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_15, 5, __pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_15, 6, __pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_15, 7, __pyx_t_8);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->left_node));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self->left_node));
  PyTuple_SET_ITEM(__pyx_t_15, 8, ((PyObject *)__pyx_v_self->left_node));
  __Pyx_GIVEREF(__pyx_t_9);
  PyTuple_SET_ITEM(__pyx_t_15, 9, __pyx_t_9);
  __Pyx_GIVEREF(__pyx_t_10);
  PyTuple_SET_ITEM(__pyx_t_15, 10, __pyx_t_10);
  __Pyx_GIVEREF(__pyx_t_11);
  PyTuple_SET_ITEM(__pyx_t_15, 11, __pyx_t_11);
  __Pyx_GIVEREF(__pyx_t_12);
  PyTuple_SET_ITEM(__pyx_t_15, 12, __pyx_t_12);
  __Pyx_GIVEREF(__pyx_t_13);
  PyTuple_SET_ITEM(__pyx_t_15, 13, __pyx_t_13);
  __Pyx_GIVEREF(__pyx_t_14);
  PyTuple_SET_ITEM(__pyx_t_15, 14, __pyx_t_14);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->right_node));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self->right_node));
  PyTuple_SET_ITEM(__pyx_t_15, 15, ((PyObject *)__pyx_v_self->right_node));
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  __pyx_t_6 = 0;
  __pyx_t_7 = 0;
  __pyx_t_8 = 0;
  __pyx_t_9 = 0;
  __pyx_t_10 = 0;
  __pyx_t_11 = 0;
  __pyx_t_12 = 0;
  __pyx_t_13 = 0;
  __pyx_t_14 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_15);
  __pyx_t_15 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.center_left_indices, self.center_left_values, self.center_right_indices, self.center_right_values, self.indices, self.is_leaf_node, self.leaf_size, self.left, self.left_node, self.max_right, self.min_left, self.n_center, self.n_elements, self.pivot, self.right, self.right_node)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_15 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_15)) __PYX_ERR(2, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __pyx_v__dict = __pyx_t_15;
  __pyx_t_15 = 0;

  /* "(tree fragment)":7
 *     state = (self.center_left_indices, self.center_left_values, self.center_right_indices, self.center_right_values, self.indices, self.is_leaf_node, self.leaf_size, self.left, self.left_node, self.max_right, self.min_left, self.n_center, self.n_elements, self.pivot, self.right, self.right_node)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_16 = (__pyx_v__dict != Py_None);
  __pyx_t_17 = (__pyx_t_16 != 0);
  if (__pyx_t_17) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_15 = PyTuple_New(1); if (unlikely(!__pyx_t_15)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    PyTuple_SET_ITEM(__pyx_t_15, 0, __pyx_v__dict);
    __pyx_t_14 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_15); if (unlikely(!__pyx_t_14)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_14));
    __pyx_t_14 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.left_node is not None or self.right_node is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.center_left_indices, self.center_left_values, self.center_right_indices, self.center_right_values, self.indices, self.is_leaf_node, self.leaf_size, self.left, self.left_node, self.max_right, self.min_left, self.n_center, self.n_elements, self.pivot, self.right, self.right_node)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.left_node is not None or self.right_node is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_Uint64ClosedBothIntervalNode, (type(self), 0x7a104e5, None), state
 */
  /*else*/ {
    __pyx_t_16 = (((PyObject *)__pyx_v_self->left_node) != Py_None);
    __pyx_t_18 = (__pyx_t_16 != 0);
    if (!__pyx_t_18) {
    } else {
      __pyx_t_17 = __pyx_t_18;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_18 = (((PyObject *)__pyx_v_self->right_node) != Py_None);
    __pyx_t_16 = (__pyx_t_18 != 0);
    __pyx_t_17 = __pyx_t_16;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_17;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.left_node is not None or self.right_node is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Uint64ClosedBothIntervalNode, (type(self), 0x7a104e5, None), state
 *     else:
 */
  __pyx_t_17 = (__pyx_v_use_setstate != 0);
  if (__pyx_t_17) {

    /* "(tree fragment)":13
 *         use_setstate = self.left_node is not None or self.right_node is not None
 *     if use_setstate:
 *         return __pyx_unpickle_Uint64ClosedBothIntervalNode, (type(self), 0x7a104e5, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_Uint64ClosedBothIntervalNode, (type(self), 0x7a104e5, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_14, __pyx_n_s_pyx_unpickle_Uint64ClosedBothI); if (unlikely(!__pyx_t_14)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __pyx_t_15 = PyTuple_New(3); if (unlikely(!__pyx_t_15)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_15, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_127993061);
    __Pyx_GIVEREF(__pyx_int_127993061);
    PyTuple_SET_ITEM(__pyx_t_15, 1, __pyx_int_127993061);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_15, 2, Py_None);
    __pyx_t_13 = PyTuple_New(3); if (unlikely(!__pyx_t_13)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __Pyx_GIVEREF(__pyx_t_14);
    PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_t_14);
    __Pyx_GIVEREF(__pyx_t_15);
    PyTuple_SET_ITEM(__pyx_t_13, 1, __pyx_t_15);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_13, 2, __pyx_v_state);
    __pyx_t_14 = 0;
    __pyx_t_15 = 0;
    __pyx_r = __pyx_t_13;
    __pyx_t_13 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.left_node is not None or self.right_node is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Uint64ClosedBothIntervalNode, (type(self), 0x7a104e5, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_Uint64ClosedBothIntervalNode, (type(self), 0x7a104e5, None), state
 *     else:
 *         return __pyx_unpickle_Uint64ClosedBothIntervalNode, (type(self), 0x7a104e5, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Uint64ClosedBothIntervalNode__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_13, __pyx_n_s_pyx_unpickle_Uint64ClosedBothI); if (unlikely(!__pyx_t_13)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __pyx_t_15 = PyTuple_New(3); if (unlikely(!__pyx_t_15)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_15, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_127993061);
    __Pyx_GIVEREF(__pyx_int_127993061);
    PyTuple_SET_ITEM(__pyx_t_15, 1, __pyx_int_127993061);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_15, 2, __pyx_v_state);
    __pyx_t_14 = PyTuple_New(2); if (unlikely(!__pyx_t_14)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_GIVEREF(__pyx_t_13);
    PyTuple_SET_ITEM(__pyx_t_14, 0, __pyx_t_13);
    __Pyx_GIVEREF(__pyx_t_15);
    PyTuple_SET_ITEM(__pyx_t_14, 1, __pyx_t_15);
    __pyx_t_13 = 0;
    __pyx_t_15 = 0;
    __pyx_r = __pyx_t_14;
    __pyx_t_14 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_AddTraceback("pandas._libs.interval.Uint64ClosedBothIntervalNode.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Uint64ClosedBothIntervalNode, (type(self), 0x7a104e5, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Uint64ClosedBothIntervalNode__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_23__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_23__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_22__setstate_cython__(((struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_22__setstate_cython__(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_Uint64ClosedBothIntervalNode, (type(self), 0x7a104e5, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Uint64ClosedBothIntervalNode__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(2, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_6pandas_5_libs_8interval___pyx_unpickle_Uint64ClosedBothIntervalNode__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Uint64ClosedBothIntervalNode, (type(self), 0x7a104e5, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Uint64ClosedBothIntervalNode__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Uint64ClosedBothIntervalNode.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":3478
 *         readonly bint is_leaf_node
 * 
 *     def __init__(self,             # <<<<<<<<<<<<<<
 *                  ndarray[uint64_t, ndim=1] left,
 *                  ndarray[uint64_t, ndim=1] right,
 */

/* Python wrapper */
static int __pyx_pw_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyArrayObject *__pyx_v_left = 0;
  PyArrayObject *__pyx_v_right = 0;
  PyArrayObject *__pyx_v_indices = 0;
  __pyx_t_5numpy_int64_t __pyx_v_leaf_size;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_left,&__pyx_n_s_right,&__pyx_n_s_indices,&__pyx_n_s_leaf_size,0};
    PyObject* values[4] = {0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_left)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_right)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 4, 4, 1); __PYX_ERR(1, 3478, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_indices)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 4, 4, 2); __PYX_ERR(1, 3478, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_leaf_size)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 4, 4, 3); __PYX_ERR(1, 3478, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(1, 3478, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 4) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
    }
    __pyx_v_left = ((PyArrayObject *)values[0]);
    __pyx_v_right = ((PyArrayObject *)values[1]);
    __pyx_v_indices = ((PyArrayObject *)values[2]);
    __pyx_v_leaf_size = __Pyx_PyInt_As_npy_int64(values[3]); if (unlikely((__pyx_v_leaf_size == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(1, 3482, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 4, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 3478, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Uint64ClosedNeitherIntervalNode.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_left), __pyx_ptype_5numpy_ndarray, 1, "left", 0))) __PYX_ERR(1, 3479, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_right), __pyx_ptype_5numpy_ndarray, 1, "right", 0))) __PYX_ERR(1, 3480, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_indices), __pyx_ptype_5numpy_ndarray, 1, "indices", 0))) __PYX_ERR(1, 3481, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode___init__(((struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode *)__pyx_v_self), __pyx_v_left, __pyx_v_right, __pyx_v_indices, __pyx_v_leaf_size);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode___init__(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode *__pyx_v_self, PyArrayObject *__pyx_v_left, PyArrayObject *__pyx_v_right, PyArrayObject *__pyx_v_indices, __pyx_t_5numpy_int64_t __pyx_v_leaf_size) {
  PyObject *__pyx_v_left_set = NULL;
  PyObject *__pyx_v_right_set = NULL;
  PyObject *__pyx_v_center_set = NULL;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_indices;
  __Pyx_Buffer __pyx_pybuffer_indices;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_left;
  __Pyx_Buffer __pyx_pybuffer_left;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_right;
  __Pyx_Buffer __pyx_pybuffer_right;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __pyx_t_5numpy_uint64_t __pyx_t_6;
  __Pyx_memviewslice __pyx_t_7 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_8 = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  __Pyx_memviewslice __pyx_t_11 = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *(*__pyx_t_12)(PyObject *);
  size_t __pyx_t_13;
  __Pyx_RefNannySetupContext("__init__", 0);
  __pyx_pybuffer_left.pybuffer.buf = NULL;
  __pyx_pybuffer_left.refcount = 0;
  __pyx_pybuffernd_left.data = NULL;
  __pyx_pybuffernd_left.rcbuffer = &__pyx_pybuffer_left;
  __pyx_pybuffer_right.pybuffer.buf = NULL;
  __pyx_pybuffer_right.refcount = 0;
  __pyx_pybuffernd_right.data = NULL;
  __pyx_pybuffernd_right.rcbuffer = &__pyx_pybuffer_right;
  __pyx_pybuffer_indices.pybuffer.buf = NULL;
  __pyx_pybuffer_indices.refcount = 0;
  __pyx_pybuffernd_indices.data = NULL;
  __pyx_pybuffernd_indices.rcbuffer = &__pyx_pybuffer_indices;
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_left.rcbuffer->pybuffer, (PyObject*)__pyx_v_left, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 3478, __pyx_L1_error)
  }
  __pyx_pybuffernd_left.diminfo[0].strides = __pyx_pybuffernd_left.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_left.diminfo[0].shape = __pyx_pybuffernd_left.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_right.rcbuffer->pybuffer, (PyObject*)__pyx_v_right, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 3478, __pyx_L1_error)
  }
  __pyx_pybuffernd_right.diminfo[0].strides = __pyx_pybuffernd_right.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_right.diminfo[0].shape = __pyx_pybuffernd_right.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_indices.rcbuffer->pybuffer, (PyObject*)__pyx_v_indices, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 3478, __pyx_L1_error)
  }
  __pyx_pybuffernd_indices.diminfo[0].strides = __pyx_pybuffernd_indices.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_indices.diminfo[0].shape = __pyx_pybuffernd_indices.rcbuffer->pybuffer.shape[0];

  /* "pandas/_libs/intervaltree.pxi":3484
 *                  int64_t leaf_size):
 * 
 *         self.n_elements = len(left)             # <<<<<<<<<<<<<<
 *         self.leaf_size = leaf_size
 * 
 */
  __pyx_t_1 = PyObject_Length(((PyObject *)__pyx_v_left)); if (unlikely(__pyx_t_1 == ((Py_ssize_t)-1))) __PYX_ERR(1, 3484, __pyx_L1_error)
  __pyx_v_self->n_elements = __pyx_t_1;

  /* "pandas/_libs/intervaltree.pxi":3485
 * 
 *         self.n_elements = len(left)
 *         self.leaf_size = leaf_size             # <<<<<<<<<<<<<<
 * 
 *         # min_left and min_right are used to speed-up query by skipping
 */
  __pyx_v_self->leaf_size = __pyx_v_leaf_size;

  /* "pandas/_libs/intervaltree.pxi":3490
 *         # query on sub-nodes. If this node has size 0, query is cheap,
 *         # so these values don't matter.
 *         if left.size > 0:             # <<<<<<<<<<<<<<
 *             self.min_left = left.min()
 *             self.max_right = right.max()
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_left), __pyx_n_s_size); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3490, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyObject_RichCompare(__pyx_t_2, __pyx_int_0, Py_GT); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 3490, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(1, 3490, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_4) {

    /* "pandas/_libs/intervaltree.pxi":3491
 *         # so these values don't matter.
 *         if left.size > 0:
 *             self.min_left = left.min()             # <<<<<<<<<<<<<<
 *             self.max_right = right.max()
 *         else:
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_left), __pyx_n_s_min); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3491, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 3491, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_6 = __Pyx_PyInt_As_npy_uint64(__pyx_t_3); if (unlikely((__pyx_t_6 == ((npy_uint64)-1)) && PyErr_Occurred())) __PYX_ERR(1, 3491, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_self->min_left = __pyx_t_6;

    /* "pandas/_libs/intervaltree.pxi":3492
 *         if left.size > 0:
 *             self.min_left = left.min()
 *             self.max_right = right.max()             # <<<<<<<<<<<<<<
 *         else:
 *             self.min_left = 0
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_right), __pyx_n_s_max); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3492, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 3492, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_6 = __Pyx_PyInt_As_npy_uint64(__pyx_t_3); if (unlikely((__pyx_t_6 == ((npy_uint64)-1)) && PyErr_Occurred())) __PYX_ERR(1, 3492, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_self->max_right = __pyx_t_6;

    /* "pandas/_libs/intervaltree.pxi":3490
 *         # query on sub-nodes. If this node has size 0, query is cheap,
 *         # so these values don't matter.
 *         if left.size > 0:             # <<<<<<<<<<<<<<
 *             self.min_left = left.min()
 *             self.max_right = right.max()
 */
    goto __pyx_L3;
  }

  /* "pandas/_libs/intervaltree.pxi":3494
 *             self.max_right = right.max()
 *         else:
 *             self.min_left = 0             # <<<<<<<<<<<<<<
 *             self.max_right = 0
 * 
 */
  /*else*/ {
    __pyx_v_self->min_left = 0;

    /* "pandas/_libs/intervaltree.pxi":3495
 *         else:
 *             self.min_left = 0
 *             self.max_right = 0             # <<<<<<<<<<<<<<
 * 
 *         if self.n_elements <= leaf_size:
 */
    __pyx_v_self->max_right = 0;
  }
  __pyx_L3:;

  /* "pandas/_libs/intervaltree.pxi":3497
 *             self.max_right = 0
 * 
 *         if self.n_elements <= leaf_size:             # <<<<<<<<<<<<<<
 *             # make this a terminal (leaf) node
 *             self.is_leaf_node = True
 */
  __pyx_t_4 = ((__pyx_v_self->n_elements <= __pyx_v_leaf_size) != 0);
  if (__pyx_t_4) {

    /* "pandas/_libs/intervaltree.pxi":3499
 *         if self.n_elements <= leaf_size:
 *             # make this a terminal (leaf) node
 *             self.is_leaf_node = True             # <<<<<<<<<<<<<<
 *             self.left = left
 *             self.right = right
 */
    __pyx_v_self->is_leaf_node = 1;

    /* "pandas/_libs/intervaltree.pxi":3500
 *             # make this a terminal (leaf) node
 *             self.is_leaf_node = True
 *             self.left = left             # <<<<<<<<<<<<<<
 *             self.right = right
 *             self.indices = indices
 */
    __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_uint64_t(((PyObject *)__pyx_v_left), PyBUF_WRITABLE); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(1, 3500, __pyx_L1_error)
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->left, 0);
    __pyx_v_self->left = __pyx_t_7;
    __pyx_t_7.memview = NULL;
    __pyx_t_7.data = NULL;

    /* "pandas/_libs/intervaltree.pxi":3501
 *             self.is_leaf_node = True
 *             self.left = left
 *             self.right = right             # <<<<<<<<<<<<<<
 *             self.indices = indices
 *             self.n_center = 0
 */
    __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_uint64_t(((PyObject *)__pyx_v_right), PyBUF_WRITABLE); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(1, 3501, __pyx_L1_error)
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->right, 0);
    __pyx_v_self->right = __pyx_t_7;
    __pyx_t_7.memview = NULL;
    __pyx_t_7.data = NULL;

    /* "pandas/_libs/intervaltree.pxi":3502
 *             self.left = left
 *             self.right = right
 *             self.indices = indices             # <<<<<<<<<<<<<<
 *             self.n_center = 0
 *         else:
 */
    __pyx_t_8 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(((PyObject *)__pyx_v_indices), PyBUF_WRITABLE); if (unlikely(!__pyx_t_8.memview)) __PYX_ERR(1, 3502, __pyx_L1_error)
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->indices, 0);
    __pyx_v_self->indices = __pyx_t_8;
    __pyx_t_8.memview = NULL;
    __pyx_t_8.data = NULL;

    /* "pandas/_libs/intervaltree.pxi":3503
 *             self.right = right
 *             self.indices = indices
 *             self.n_center = 0             # <<<<<<<<<<<<<<
 *         else:
 *             # calculate a pivot so we can create child nodes
 */
    __pyx_v_self->n_center = 0;

    /* "pandas/_libs/intervaltree.pxi":3497
 *             self.max_right = 0
 * 
 *         if self.n_elements <= leaf_size:             # <<<<<<<<<<<<<<
 *             # make this a terminal (leaf) node
 *             self.is_leaf_node = True
 */
    goto __pyx_L4;
  }

  /* "pandas/_libs/intervaltree.pxi":3506
 *         else:
 *             # calculate a pivot so we can create child nodes
 *             self.is_leaf_node = False             # <<<<<<<<<<<<<<
 *             self.pivot = np.median(left / 2 + right / 2)
 *             left_set, right_set, center_set = self.classify_intervals(
 */
  /*else*/ {
    __pyx_v_self->is_leaf_node = 0;

    /* "pandas/_libs/intervaltree.pxi":3507
 *             # calculate a pivot so we can create child nodes
 *             self.is_leaf_node = False
 *             self.pivot = np.median(left / 2 + right / 2)             # <<<<<<<<<<<<<<
 *             left_set, right_set, center_set = self.classify_intervals(
 *                 left, right)
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3507, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_median); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 3507, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyNumber_Divide(((PyObject *)__pyx_v_left), __pyx_int_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3507, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_9 = __Pyx_PyNumber_Divide(((PyObject *)__pyx_v_right), __pyx_int_2); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 3507, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_10 = PyNumber_Add(__pyx_t_2, __pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 3507, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_9 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_3 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_9, __pyx_t_10) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_10);
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 3507, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_6 = __Pyx_PyInt_As_npy_uint64(__pyx_t_3); if (unlikely((__pyx_t_6 == ((npy_uint64)-1)) && PyErr_Occurred())) __PYX_ERR(1, 3507, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_self->pivot = __pyx_t_6;

    /* "pandas/_libs/intervaltree.pxi":3509
 *             self.pivot = np.median(left / 2 + right / 2)
 *             left_set, right_set, center_set = self.classify_intervals(
 *                 left, right)             # <<<<<<<<<<<<<<
 * 
 *             self.left_node = self.new_child_node(left, right,
 */
    __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_uint64_t(((PyObject *)__pyx_v_left), PyBUF_WRITABLE); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(1, 3509, __pyx_L1_error)
    __pyx_t_11 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_uint64_t(((PyObject *)__pyx_v_right), PyBUF_WRITABLE); if (unlikely(!__pyx_t_11.memview)) __PYX_ERR(1, 3509, __pyx_L1_error)

    /* "pandas/_libs/intervaltree.pxi":3508
 *             self.is_leaf_node = False
 *             self.pivot = np.median(left / 2 + right / 2)
 *             left_set, right_set, center_set = self.classify_intervals(             # <<<<<<<<<<<<<<
 *                 left, right)
 * 
 */
    __pyx_t_3 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode *)__pyx_v_self->__pyx_vtab)->classify_intervals(__pyx_v_self, __pyx_t_7, __pyx_t_11); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 3508, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __PYX_XDEC_MEMVIEW(&__pyx_t_7, 1);
    __pyx_t_7.memview = NULL;
    __pyx_t_7.data = NULL;
    __PYX_XDEC_MEMVIEW(&__pyx_t_11, 1);
    __pyx_t_11.memview = NULL;
    __pyx_t_11.data = NULL;
    if ((likely(PyTuple_CheckExact(__pyx_t_3))) || (PyList_CheckExact(__pyx_t_3))) {
      PyObject* sequence = __pyx_t_3;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 3)) {
        if (size > 3) __Pyx_RaiseTooManyValuesError(3);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(1, 3508, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_5 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_10 = PyTuple_GET_ITEM(sequence, 1); 
        __pyx_t_9 = PyTuple_GET_ITEM(sequence, 2); 
      } else {
        __pyx_t_5 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_10 = PyList_GET_ITEM(sequence, 1); 
        __pyx_t_9 = PyList_GET_ITEM(sequence, 2); 
      }
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_10);
      __Pyx_INCREF(__pyx_t_9);
      #else
      __pyx_t_5 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 3508, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_10 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 3508, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_9 = PySequence_ITEM(sequence, 2); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 3508, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      #endif
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_2 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3508, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_12 = Py_TYPE(__pyx_t_2)->tp_iternext;
      index = 0; __pyx_t_5 = __pyx_t_12(__pyx_t_2); if (unlikely(!__pyx_t_5)) goto __pyx_L5_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_5);
      index = 1; __pyx_t_10 = __pyx_t_12(__pyx_t_2); if (unlikely(!__pyx_t_10)) goto __pyx_L5_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_10);
      index = 2; __pyx_t_9 = __pyx_t_12(__pyx_t_2); if (unlikely(!__pyx_t_9)) goto __pyx_L5_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_9);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_12(__pyx_t_2), 3) < 0) __PYX_ERR(1, 3508, __pyx_L1_error)
      __pyx_t_12 = NULL;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      goto __pyx_L6_unpacking_done;
      __pyx_L5_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_12 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(1, 3508, __pyx_L1_error)
      __pyx_L6_unpacking_done:;
    }
    __pyx_v_left_set = __pyx_t_5;
    __pyx_t_5 = 0;
    __pyx_v_right_set = __pyx_t_10;
    __pyx_t_10 = 0;
    __pyx_v_center_set = __pyx_t_9;
    __pyx_t_9 = 0;

    /* "pandas/_libs/intervaltree.pxi":3512
 * 
 *             self.left_node = self.new_child_node(left, right,
 *                                                  indices, left_set)             # <<<<<<<<<<<<<<
 *             self.right_node = self.new_child_node(left, right,
 *                                                   indices, right_set)
 */
    if (!(likely(((__pyx_v_left_set) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_left_set, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(1, 3512, __pyx_L1_error)

    /* "pandas/_libs/intervaltree.pxi":3511
 *                 left, right)
 * 
 *             self.left_node = self.new_child_node(left, right,             # <<<<<<<<<<<<<<
 *                                                  indices, left_set)
 *             self.right_node = self.new_child_node(left, right,
 */
    __pyx_t_3 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode *)__pyx_v_self->__pyx_vtab)->new_child_node(__pyx_v_self, ((PyArrayObject *)__pyx_v_left), ((PyArrayObject *)__pyx_v_right), ((PyArrayObject *)__pyx_v_indices), ((PyArrayObject *)__pyx_v_left_set)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 3511, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode))))) __PYX_ERR(1, 3511, __pyx_L1_error)
    __Pyx_GIVEREF(__pyx_t_3);
    __Pyx_GOTREF(__pyx_v_self->left_node);
    __Pyx_DECREF(((PyObject *)__pyx_v_self->left_node));
    __pyx_v_self->left_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode *)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "pandas/_libs/intervaltree.pxi":3514
 *                                                  indices, left_set)
 *             self.right_node = self.new_child_node(left, right,
 *                                                   indices, right_set)             # <<<<<<<<<<<<<<
 * 
 *             self.center_left_values, self.center_left_indices = \
 */
    if (!(likely(((__pyx_v_right_set) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_right_set, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(1, 3514, __pyx_L1_error)

    /* "pandas/_libs/intervaltree.pxi":3513
 *             self.left_node = self.new_child_node(left, right,
 *                                                  indices, left_set)
 *             self.right_node = self.new_child_node(left, right,             # <<<<<<<<<<<<<<
 *                                                   indices, right_set)
 * 
 */
    __pyx_t_3 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode *)__pyx_v_self->__pyx_vtab)->new_child_node(__pyx_v_self, ((PyArrayObject *)__pyx_v_left), ((PyArrayObject *)__pyx_v_right), ((PyArrayObject *)__pyx_v_indices), ((PyArrayObject *)__pyx_v_right_set)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 3513, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode))))) __PYX_ERR(1, 3513, __pyx_L1_error)
    __Pyx_GIVEREF(__pyx_t_3);
    __Pyx_GOTREF(__pyx_v_self->right_node);
    __Pyx_DECREF(((PyObject *)__pyx_v_self->right_node));
    __pyx_v_self->right_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode *)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "pandas/_libs/intervaltree.pxi":3517
 * 
 *             self.center_left_values, self.center_left_indices = \
 *                 sort_values_and_indices(left, indices, center_set)             # <<<<<<<<<<<<<<
 *             self.center_right_values, self.center_right_indices = \
 *                 sort_values_and_indices(right, indices, center_set)
 */
    __pyx_t_3 = __pyx_f_6pandas_5_libs_8interval_sort_values_and_indices(((PyObject *)__pyx_v_left), ((PyObject *)__pyx_v_indices), __pyx_v_center_set); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 3517, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if ((likely(PyTuple_CheckExact(__pyx_t_3))) || (PyList_CheckExact(__pyx_t_3))) {
      PyObject* sequence = __pyx_t_3;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(1, 3516, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_9 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_10 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_9 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_10 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_9);
      __Pyx_INCREF(__pyx_t_10);
      #else
      __pyx_t_9 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 3516, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_10 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 3516, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      #endif
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_5 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 3516, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_12 = Py_TYPE(__pyx_t_5)->tp_iternext;
      index = 0; __pyx_t_9 = __pyx_t_12(__pyx_t_5); if (unlikely(!__pyx_t_9)) goto __pyx_L7_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_9);
      index = 1; __pyx_t_10 = __pyx_t_12(__pyx_t_5); if (unlikely(!__pyx_t_10)) goto __pyx_L7_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_10);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_12(__pyx_t_5), 2) < 0) __PYX_ERR(1, 3516, __pyx_L1_error)
      __pyx_t_12 = NULL;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      goto __pyx_L8_unpacking_done;
      __pyx_L7_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_12 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(1, 3516, __pyx_L1_error)
      __pyx_L8_unpacking_done:;
    }

    /* "pandas/_libs/intervaltree.pxi":3516
 *                                                   indices, right_set)
 * 
 *             self.center_left_values, self.center_left_indices = \             # <<<<<<<<<<<<<<
 *                 sort_values_and_indices(left, indices, center_set)
 *             self.center_right_values, self.center_right_indices = \
 */
    __pyx_t_11 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_uint64_t(__pyx_t_9, PyBUF_WRITABLE); if (unlikely(!__pyx_t_11.memview)) __PYX_ERR(1, 3516, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_8 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_t_10, PyBUF_WRITABLE); if (unlikely(!__pyx_t_8.memview)) __PYX_ERR(1, 3516, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->center_left_values, 0);
    __pyx_v_self->center_left_values = __pyx_t_11;
    __pyx_t_11.memview = NULL;
    __pyx_t_11.data = NULL;
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->center_left_indices, 0);
    __pyx_v_self->center_left_indices = __pyx_t_8;
    __pyx_t_8.memview = NULL;
    __pyx_t_8.data = NULL;

    /* "pandas/_libs/intervaltree.pxi":3519
 *                 sort_values_and_indices(left, indices, center_set)
 *             self.center_right_values, self.center_right_indices = \
 *                 sort_values_and_indices(right, indices, center_set)             # <<<<<<<<<<<<<<
 *             self.n_center = len(self.center_left_indices)
 * 
 */
    __pyx_t_3 = __pyx_f_6pandas_5_libs_8interval_sort_values_and_indices(((PyObject *)__pyx_v_right), ((PyObject *)__pyx_v_indices), __pyx_v_center_set); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 3519, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if ((likely(PyTuple_CheckExact(__pyx_t_3))) || (PyList_CheckExact(__pyx_t_3))) {
      PyObject* sequence = __pyx_t_3;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(1, 3518, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_10 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_9 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_10 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_9 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_10);
      __Pyx_INCREF(__pyx_t_9);
      #else
      __pyx_t_10 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 3518, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_9 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 3518, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      #endif
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_5 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 3518, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_12 = Py_TYPE(__pyx_t_5)->tp_iternext;
      index = 0; __pyx_t_10 = __pyx_t_12(__pyx_t_5); if (unlikely(!__pyx_t_10)) goto __pyx_L9_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_10);
      index = 1; __pyx_t_9 = __pyx_t_12(__pyx_t_5); if (unlikely(!__pyx_t_9)) goto __pyx_L9_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_9);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_12(__pyx_t_5), 2) < 0) __PYX_ERR(1, 3518, __pyx_L1_error)
      __pyx_t_12 = NULL;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      goto __pyx_L10_unpacking_done;
      __pyx_L9_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_12 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(1, 3518, __pyx_L1_error)
      __pyx_L10_unpacking_done:;
    }

    /* "pandas/_libs/intervaltree.pxi":3518
 *             self.center_left_values, self.center_left_indices = \
 *                 sort_values_and_indices(left, indices, center_set)
 *             self.center_right_values, self.center_right_indices = \             # <<<<<<<<<<<<<<
 *                 sort_values_and_indices(right, indices, center_set)
 *             self.n_center = len(self.center_left_indices)
 */
    __pyx_t_11 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_uint64_t(__pyx_t_10, PyBUF_WRITABLE); if (unlikely(!__pyx_t_11.memview)) __PYX_ERR(1, 3518, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_t_8 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_t_9, PyBUF_WRITABLE); if (unlikely(!__pyx_t_8.memview)) __PYX_ERR(1, 3518, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->center_right_values, 0);
    __pyx_v_self->center_right_values = __pyx_t_11;
    __pyx_t_11.memview = NULL;
    __pyx_t_11.data = NULL;
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->center_right_indices, 0);
    __pyx_v_self->center_right_indices = __pyx_t_8;
    __pyx_t_8.memview = NULL;
    __pyx_t_8.data = NULL;

    /* "pandas/_libs/intervaltree.pxi":3520
 *             self.center_right_values, self.center_right_indices = \
 *                 sort_values_and_indices(right, indices, center_set)
 *             self.n_center = len(self.center_left_indices)             # <<<<<<<<<<<<<<
 * 
 *     @cython.wraparound(False)
 */
    if (unlikely(!__pyx_v_self->center_left_indices.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(1, 3520, __pyx_L1_error)}
    __pyx_t_13 = __Pyx_MemoryView_Len(__pyx_v_self->center_left_indices); 
    __pyx_v_self->n_center = __pyx_t_13;
  }
  __pyx_L4:;

  /* "pandas/_libs/intervaltree.pxi":3478
 *         readonly bint is_leaf_node
 * 
 *     def __init__(self,             # <<<<<<<<<<<<<<
 *                  ndarray[uint64_t, ndim=1] left,
 *                  ndarray[uint64_t, ndim=1] right,
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __PYX_XDEC_MEMVIEW(&__pyx_t_7, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_8, 1);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __PYX_XDEC_MEMVIEW(&__pyx_t_11, 1);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indices.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_left.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_right.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("pandas._libs.interval.Uint64ClosedNeitherIntervalNode.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indices.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_left.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_right.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF(__pyx_v_left_set);
  __Pyx_XDECREF(__pyx_v_right_set);
  __Pyx_XDECREF(__pyx_v_center_set);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":3524
 *     @cython.wraparound(False)
 *     @cython.boundscheck(False)
 *     cdef classify_intervals(self, uint64_t[:] left, uint64_t[:] right):             # <<<<<<<<<<<<<<
 *         """Classify the given intervals based upon whether they fall to the
 *         left, right, or overlap with this node's pivot.
 */

static PyObject *__pyx_f_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_classify_intervals(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode *__pyx_v_self, __Pyx_memviewslice __pyx_v_left, __Pyx_memviewslice __pyx_v_right) {
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_left_ind = 0;
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_right_ind = 0;
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_overlapping_ind = 0;
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __pyx_t_5numpy_int64_t __pyx_t_2;
  __pyx_t_5numpy_int64_t __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  int __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  __Pyx_RefNannySetupContext("classify_intervals", 0);

  /* "pandas/_libs/intervaltree.pxi":3532
 *             Py_ssize_t i
 * 
 *         left_ind = Int64Vector()             # <<<<<<<<<<<<<<
 *         right_ind = Int64Vector()
 *         overlapping_ind = Int64Vector()
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3532, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_left_ind = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":3533
 * 
 *         left_ind = Int64Vector()
 *         right_ind = Int64Vector()             # <<<<<<<<<<<<<<
 *         overlapping_ind = Int64Vector()
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3533, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_right_ind = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":3534
 *         left_ind = Int64Vector()
 *         right_ind = Int64Vector()
 *         overlapping_ind = Int64Vector()             # <<<<<<<<<<<<<<
 * 
 *         for i in range(self.n_elements):
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3534, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_overlapping_ind = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":3536
 *         overlapping_ind = Int64Vector()
 * 
 *         for i in range(self.n_elements):             # <<<<<<<<<<<<<<
 *             if right[i] <= self.pivot:
 *                 left_ind.append(i)
 */
  __pyx_t_2 = __pyx_v_self->n_elements;
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "pandas/_libs/intervaltree.pxi":3537
 * 
 *         for i in range(self.n_elements):
 *             if right[i] <= self.pivot:             # <<<<<<<<<<<<<<
 *                 left_ind.append(i)
 *             elif self.pivot <= left[i]:
 */
    __pyx_t_5 = __pyx_v_i;
    __pyx_t_6 = (((*((__pyx_t_5numpy_uint64_t *) ( /* dim=0 */ (__pyx_v_right.data + __pyx_t_5 * __pyx_v_right.strides[0]) ))) <= __pyx_v_self->pivot) != 0);
    if (__pyx_t_6) {

      /* "pandas/_libs/intervaltree.pxi":3538
 *         for i in range(self.n_elements):
 *             if right[i] <= self.pivot:
 *                 left_ind.append(i)             # <<<<<<<<<<<<<<
 *             elif self.pivot <= left[i]:
 *                 right_ind.append(i)
 */
      ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_left_ind->__pyx_vtab)->append(__pyx_v_left_ind, __pyx_v_i);

      /* "pandas/_libs/intervaltree.pxi":3537
 * 
 *         for i in range(self.n_elements):
 *             if right[i] <= self.pivot:             # <<<<<<<<<<<<<<
 *                 left_ind.append(i)
 *             elif self.pivot <= left[i]:
 */
      goto __pyx_L5;
    }

    /* "pandas/_libs/intervaltree.pxi":3539
 *             if right[i] <= self.pivot:
 *                 left_ind.append(i)
 *             elif self.pivot <= left[i]:             # <<<<<<<<<<<<<<
 *                 right_ind.append(i)
 *             else:
 */
    __pyx_t_7 = __pyx_v_i;
    __pyx_t_6 = ((__pyx_v_self->pivot <= (*((__pyx_t_5numpy_uint64_t *) ( /* dim=0 */ (__pyx_v_left.data + __pyx_t_7 * __pyx_v_left.strides[0]) )))) != 0);
    if (__pyx_t_6) {

      /* "pandas/_libs/intervaltree.pxi":3540
 *                 left_ind.append(i)
 *             elif self.pivot <= left[i]:
 *                 right_ind.append(i)             # <<<<<<<<<<<<<<
 *             else:
 *                 overlapping_ind.append(i)
 */
      ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_right_ind->__pyx_vtab)->append(__pyx_v_right_ind, __pyx_v_i);

      /* "pandas/_libs/intervaltree.pxi":3539
 *             if right[i] <= self.pivot:
 *                 left_ind.append(i)
 *             elif self.pivot <= left[i]:             # <<<<<<<<<<<<<<
 *                 right_ind.append(i)
 *             else:
 */
      goto __pyx_L5;
    }

    /* "pandas/_libs/intervaltree.pxi":3542
 *                 right_ind.append(i)
 *             else:
 *                 overlapping_ind.append(i)             # <<<<<<<<<<<<<<
 * 
 *         return (left_ind.to_array(),
 */
    /*else*/ {
      ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_overlapping_ind->__pyx_vtab)->append(__pyx_v_overlapping_ind, __pyx_v_i);
    }
    __pyx_L5:;
  }

  /* "pandas/_libs/intervaltree.pxi":3544
 *                 overlapping_ind.append(i)
 * 
 *         return (left_ind.to_array(),             # <<<<<<<<<<<<<<
 *                 right_ind.to_array(),
 *                 overlapping_ind.to_array())
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_left_ind->__pyx_vtab)->to_array(__pyx_v_left_ind, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3544, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "pandas/_libs/intervaltree.pxi":3545
 * 
 *         return (left_ind.to_array(),
 *                 right_ind.to_array(),             # <<<<<<<<<<<<<<
 *                 overlapping_ind.to_array())
 * 
 */
  __pyx_t_8 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_right_ind->__pyx_vtab)->to_array(__pyx_v_right_ind, 0); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 3545, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);

  /* "pandas/_libs/intervaltree.pxi":3546
 *         return (left_ind.to_array(),
 *                 right_ind.to_array(),
 *                 overlapping_ind.to_array())             # <<<<<<<<<<<<<<
 * 
 *     cdef new_child_node(self,
 */
  __pyx_t_9 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_overlapping_ind->__pyx_vtab)->to_array(__pyx_v_overlapping_ind, 0); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 3546, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);

  /* "pandas/_libs/intervaltree.pxi":3544
 *                 overlapping_ind.append(i)
 * 
 *         return (left_ind.to_array(),             # <<<<<<<<<<<<<<
 *                 right_ind.to_array(),
 *                 overlapping_ind.to_array())
 */
  __pyx_t_10 = PyTuple_New(3); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 3544, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_t_8);
  __Pyx_GIVEREF(__pyx_t_9);
  PyTuple_SET_ITEM(__pyx_t_10, 2, __pyx_t_9);
  __pyx_t_1 = 0;
  __pyx_t_8 = 0;
  __pyx_t_9 = 0;
  __pyx_r = __pyx_t_10;
  __pyx_t_10 = 0;
  goto __pyx_L0;

  /* "pandas/_libs/intervaltree.pxi":3524
 *     @cython.wraparound(False)
 *     @cython.boundscheck(False)
 *     cdef classify_intervals(self, uint64_t[:] left, uint64_t[:] right):             # <<<<<<<<<<<<<<
 *         """Classify the given intervals based upon whether they fall to the
 *         left, right, or overlap with this node's pivot.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("pandas._libs.interval.Uint64ClosedNeitherIntervalNode.classify_intervals", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_left_ind);
  __Pyx_XDECREF((PyObject *)__pyx_v_right_ind);
  __Pyx_XDECREF((PyObject *)__pyx_v_overlapping_ind);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":3548
 *                 overlapping_ind.to_array())
 * 
 *     cdef new_child_node(self,             # <<<<<<<<<<<<<<
 *                         ndarray[uint64_t, ndim=1] left,
 *                         ndarray[uint64_t, ndim=1] right,
 */

static PyObject *__pyx_f_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_new_child_node(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode *__pyx_v_self, PyArrayObject *__pyx_v_left, PyArrayObject *__pyx_v_right, PyArrayObject *__pyx_v_indices, PyArrayObject *__pyx_v_subset) {
  __Pyx_LocalBuf_ND __pyx_pybuffernd_indices;
  __Pyx_Buffer __pyx_pybuffer_indices;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_left;
  __Pyx_Buffer __pyx_pybuffer_left;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_right;
  __Pyx_Buffer __pyx_pybuffer_right;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_subset;
  __Pyx_Buffer __pyx_pybuffer_subset;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyArrayObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyArrayObject *__pyx_t_7 = NULL;
  PyArrayObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  __Pyx_RefNannySetupContext("new_child_node", 0);
  __Pyx_INCREF((PyObject *)__pyx_v_left);
  __Pyx_INCREF((PyObject *)__pyx_v_right);
  __Pyx_INCREF((PyObject *)__pyx_v_indices);
  __pyx_pybuffer_left.pybuffer.buf = NULL;
  __pyx_pybuffer_left.refcount = 0;
  __pyx_pybuffernd_left.data = NULL;
  __pyx_pybuffernd_left.rcbuffer = &__pyx_pybuffer_left;
  __pyx_pybuffer_right.pybuffer.buf = NULL;
  __pyx_pybuffer_right.refcount = 0;
  __pyx_pybuffernd_right.data = NULL;
  __pyx_pybuffernd_right.rcbuffer = &__pyx_pybuffer_right;
  __pyx_pybuffer_indices.pybuffer.buf = NULL;
  __pyx_pybuffer_indices.refcount = 0;
  __pyx_pybuffernd_indices.data = NULL;
  __pyx_pybuffernd_indices.rcbuffer = &__pyx_pybuffer_indices;
  __pyx_pybuffer_subset.pybuffer.buf = NULL;
  __pyx_pybuffer_subset.refcount = 0;
  __pyx_pybuffernd_subset.data = NULL;
  __pyx_pybuffernd_subset.rcbuffer = &__pyx_pybuffer_subset;
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_left.rcbuffer->pybuffer, (PyObject*)__pyx_v_left, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 3548, __pyx_L1_error)
  }
  __pyx_pybuffernd_left.diminfo[0].strides = __pyx_pybuffernd_left.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_left.diminfo[0].shape = __pyx_pybuffernd_left.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_right.rcbuffer->pybuffer, (PyObject*)__pyx_v_right, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 3548, __pyx_L1_error)
  }
  __pyx_pybuffernd_right.diminfo[0].strides = __pyx_pybuffernd_right.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_right.diminfo[0].shape = __pyx_pybuffernd_right.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_indices.rcbuffer->pybuffer, (PyObject*)__pyx_v_indices, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 3548, __pyx_L1_error)
  }
  __pyx_pybuffernd_indices.diminfo[0].strides = __pyx_pybuffernd_indices.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_indices.diminfo[0].shape = __pyx_pybuffernd_indices.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_subset.rcbuffer->pybuffer, (PyObject*)__pyx_v_subset, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(1, 3548, __pyx_L1_error)
  }
  __pyx_pybuffernd_subset.diminfo[0].strides = __pyx_pybuffernd_subset.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_subset.diminfo[0].shape = __pyx_pybuffernd_subset.rcbuffer->pybuffer.shape[0];

  /* "pandas/_libs/intervaltree.pxi":3555
 *         """Create a new child node.
 *         """
 *         left = take(left, subset)             # <<<<<<<<<<<<<<
 *         right = take(right, subset)
 *         indices = take(indices, subset)
 */
  __pyx_t_1 = __pyx_f_6pandas_5_libs_8interval_take(((PyArrayObject *)__pyx_v_left), ((PyArrayObject *)__pyx_v_subset)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3555, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(1, 3555, __pyx_L1_error)
  __pyx_t_2 = ((PyArrayObject *)__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_left.rcbuffer->pybuffer);
    __pyx_t_3 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_left.rcbuffer->pybuffer, (PyObject*)__pyx_t_2, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack);
    if (unlikely(__pyx_t_3 < 0)) {
      PyErr_Fetch(&__pyx_t_4, &__pyx_t_5, &__pyx_t_6);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_left.rcbuffer->pybuffer, (PyObject*)__pyx_v_left, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_4); Py_XDECREF(__pyx_t_5); Py_XDECREF(__pyx_t_6);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_4, __pyx_t_5, __pyx_t_6);
      }
      __pyx_t_4 = __pyx_t_5 = __pyx_t_6 = 0;
    }
    __pyx_pybuffernd_left.diminfo[0].strides = __pyx_pybuffernd_left.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_left.diminfo[0].shape = __pyx_pybuffernd_left.rcbuffer->pybuffer.shape[0];
    if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(1, 3555, __pyx_L1_error)
  }
  __pyx_t_2 = 0;
  __Pyx_DECREF_SET(__pyx_v_left, ((PyArrayObject *)__pyx_t_1));
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":3556
 *         """
 *         left = take(left, subset)
 *         right = take(right, subset)             # <<<<<<<<<<<<<<
 *         indices = take(indices, subset)
 *         return Uint64ClosedNeitherIntervalNode(
 */
  __pyx_t_1 = __pyx_f_6pandas_5_libs_8interval_take(((PyArrayObject *)__pyx_v_right), ((PyArrayObject *)__pyx_v_subset)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3556, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(1, 3556, __pyx_L1_error)
  __pyx_t_7 = ((PyArrayObject *)__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_right.rcbuffer->pybuffer);
    __pyx_t_3 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_right.rcbuffer->pybuffer, (PyObject*)__pyx_t_7, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack);
    if (unlikely(__pyx_t_3 < 0)) {
      PyErr_Fetch(&__pyx_t_6, &__pyx_t_5, &__pyx_t_4);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_right.rcbuffer->pybuffer, (PyObject*)__pyx_v_right, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_6); Py_XDECREF(__pyx_t_5); Py_XDECREF(__pyx_t_4);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_6, __pyx_t_5, __pyx_t_4);
      }
      __pyx_t_6 = __pyx_t_5 = __pyx_t_4 = 0;
    }
    __pyx_pybuffernd_right.diminfo[0].strides = __pyx_pybuffernd_right.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_right.diminfo[0].shape = __pyx_pybuffernd_right.rcbuffer->pybuffer.shape[0];
    if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(1, 3556, __pyx_L1_error)
  }
  __pyx_t_7 = 0;
  __Pyx_DECREF_SET(__pyx_v_right, ((PyArrayObject *)__pyx_t_1));
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":3557
 *         left = take(left, subset)
 *         right = take(right, subset)
 *         indices = take(indices, subset)             # <<<<<<<<<<<<<<
 *         return Uint64ClosedNeitherIntervalNode(
 *             left, right, indices, self.leaf_size)
 */
  __pyx_t_1 = __pyx_f_6pandas_5_libs_8interval_take(((PyArrayObject *)__pyx_v_indices), ((PyArrayObject *)__pyx_v_subset)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3557, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(1, 3557, __pyx_L1_error)
  __pyx_t_8 = ((PyArrayObject *)__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indices.rcbuffer->pybuffer);
    __pyx_t_3 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_indices.rcbuffer->pybuffer, (PyObject*)__pyx_t_8, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack);
    if (unlikely(__pyx_t_3 < 0)) {
      PyErr_Fetch(&__pyx_t_4, &__pyx_t_5, &__pyx_t_6);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_indices.rcbuffer->pybuffer, (PyObject*)__pyx_v_indices, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_4); Py_XDECREF(__pyx_t_5); Py_XDECREF(__pyx_t_6);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_4, __pyx_t_5, __pyx_t_6);
      }
      __pyx_t_4 = __pyx_t_5 = __pyx_t_6 = 0;
    }
    __pyx_pybuffernd_indices.diminfo[0].strides = __pyx_pybuffernd_indices.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_indices.diminfo[0].shape = __pyx_pybuffernd_indices.rcbuffer->pybuffer.shape[0];
    if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(1, 3557, __pyx_L1_error)
  }
  __pyx_t_8 = 0;
  __Pyx_DECREF_SET(__pyx_v_indices, ((PyArrayObject *)__pyx_t_1));
  __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":3558
 *         right = take(right, subset)
 *         indices = take(indices, subset)
 *         return Uint64ClosedNeitherIntervalNode(             # <<<<<<<<<<<<<<
 *             left, right, indices, self.leaf_size)
 * 
 */
  __Pyx_XDECREF(__pyx_r);

  /* "pandas/_libs/intervaltree.pxi":3559
 *         indices = take(indices, subset)
 *         return Uint64ClosedNeitherIntervalNode(
 *             left, right, indices, self.leaf_size)             # <<<<<<<<<<<<<<
 * 
 *     @cython.wraparound(False)
 */
  __pyx_t_1 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->leaf_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3559, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "pandas/_libs/intervaltree.pxi":3558
 *         right = take(right, subset)
 *         indices = take(indices, subset)
 *         return Uint64ClosedNeitherIntervalNode(             # <<<<<<<<<<<<<<
 *             left, right, indices, self.leaf_size)
 * 
 */
  __pyx_t_9 = PyTuple_New(4); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 3558, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_INCREF(((PyObject *)__pyx_v_left));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_left));
  PyTuple_SET_ITEM(__pyx_t_9, 0, ((PyObject *)__pyx_v_left));
  __Pyx_INCREF(((PyObject *)__pyx_v_right));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_right));
  PyTuple_SET_ITEM(__pyx_t_9, 1, ((PyObject *)__pyx_v_right));
  __Pyx_INCREF(((PyObject *)__pyx_v_indices));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_indices));
  PyTuple_SET_ITEM(__pyx_t_9, 2, ((PyObject *)__pyx_v_indices));
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_9, 3, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode), __pyx_t_9, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3558, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pandas/_libs/intervaltree.pxi":3548
 *                 overlapping_ind.to_array())
 * 
 *     cdef new_child_node(self,             # <<<<<<<<<<<<<<
 *                         ndarray[uint64_t, ndim=1] left,
 *                         ndarray[uint64_t, ndim=1] right,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_9);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indices.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_left.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_right.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_subset.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("pandas._libs.interval.Uint64ClosedNeitherIntervalNode.new_child_node", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indices.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_left.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_right.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_subset.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF((PyObject *)__pyx_v_left);
  __Pyx_XDECREF((PyObject *)__pyx_v_right);
  __Pyx_XDECREF((PyObject *)__pyx_v_indices);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":3564
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_2query[] = "Recursively query this node and its sub-nodes for intervals that\n        overlap with the query point.\n        ";
static PyMethodDef __pyx_mdef_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_3query = {"query", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_3query, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_2query};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_args = 0;
  PyObject *__pyx_v_kwargs = 0;
  CYTHON_UNUSED PyObject *__pyx_v_defaults = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fused_cpdef (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_args,&__pyx_n_s_kwargs,&__pyx_n_s_defaults,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_args)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_kwargs)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fused_cpdef", 1, 3, 3, 1); __PYX_ERR(1, 3564, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_defaults)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fused_cpdef", 1, 3, 3, 2); __PYX_ERR(1, 3564, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fused_cpdef") < 0)) __PYX_ERR(1, 3564, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_args = values[0];
    __pyx_v_kwargs = values[1];
    __pyx_v_defaults = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fused_cpdef", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 3564, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Uint64ClosedNeitherIntervalNode.__pyx_fused_cpdef", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_2query(((struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode *)__pyx_v_signatures), __pyx_v_args, __pyx_v_kwargs, __pyx_v_defaults);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_2query(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode *__pyx_v_signatures, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs, CYTHON_UNUSED PyObject *__pyx_v_defaults) {
  PyObject *__pyx_v_dest_sig = NULL;
  Py_ssize_t __pyx_v_i;
  CYTHON_UNUSED PyTypeObject *__pyx_v_ndarray = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_v_itemsize;
  PyObject *__pyx_v_arg = NULL;
  PyObject *__pyx_v_candidates = NULL;
  PyObject *__pyx_v_sig = NULL;
  int __pyx_v_match_found;
  PyObject *__pyx_v_src_sig = NULL;
  PyObject *__pyx_v_dst_type = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  Py_ssize_t __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  Py_ssize_t __pyx_t_13;
  Py_ssize_t __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  int __pyx_t_16;
  __Pyx_RefNannySetupContext("query", 0);
  __Pyx_INCREF(__pyx_v_kwargs);
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3564, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  PyList_SET_ITEM(__pyx_t_1, 0, Py_None);
  __pyx_v_dest_sig = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_3 = (__pyx_v_kwargs != Py_None);
  __pyx_t_4 = (__pyx_t_3 != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_kwargs); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(1, 3564, __pyx_L1_error)
  __pyx_t_3 = ((!__pyx_t_4) != 0);
  __pyx_t_2 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {
    __Pyx_INCREF(Py_None);
    __Pyx_DECREF_SET(__pyx_v_kwargs, Py_None);
  }
  __pyx_t_1 = ((PyObject *)__Pyx_ImportNumPyArrayTypeIfAvailable()); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3564, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_ndarray = ((PyTypeObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_v_itemsize = -1L;
  if (unlikely(__pyx_v_args == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 3564, __pyx_L1_error)
  }
  __pyx_t_5 = PyTuple_GET_SIZE(((PyObject*)__pyx_v_args)); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(1, 3564, __pyx_L1_error)
  __pyx_t_2 = ((2 < __pyx_t_5) != 0);
  if (__pyx_t_2) {
    if (unlikely(__pyx_v_args == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 3564, __pyx_L1_error)
    }
    __pyx_t_1 = PyTuple_GET_ITEM(((PyObject*)__pyx_v_args), 2);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v_arg = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L6;
  }
  __pyx_t_3 = (__pyx_v_kwargs != Py_None);
  __pyx_t_4 = (__pyx_t_3 != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L7_bool_binop_done;
  }
  if (unlikely(__pyx_v_kwargs == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(1, 3564, __pyx_L1_error)
  }
  __pyx_t_4 = (__Pyx_PyDict_ContainsTF(__pyx_n_s_point, ((PyObject*)__pyx_v_kwargs), Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(1, 3564, __pyx_L1_error)
  __pyx_t_3 = (__pyx_t_4 != 0);
  __pyx_t_2 = __pyx_t_3;
  __pyx_L7_bool_binop_done:;
  if (__pyx_t_2) {
    if (unlikely(__pyx_v_kwargs == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 3564, __pyx_L1_error)
    }
    __pyx_t_1 = __Pyx_PyDict_GetItem(((PyObject*)__pyx_v_kwargs), __pyx_n_s_point); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3564, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_arg = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L6;
  }
  /*else*/ {
    if (unlikely(__pyx_v_args == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
      __PYX_ERR(1, 3564, __pyx_L1_error)
    }
    __pyx_t_5 = PyTuple_GET_SIZE(((PyObject*)__pyx_v_args)); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(1, 3564, __pyx_L1_error)
    __pyx_t_1 = PyInt_FromSsize_t(__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3564, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_6 = PyTuple_New(3); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 3564, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_INCREF(__pyx_int_3);
    __Pyx_GIVEREF(__pyx_int_3);
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_int_3);
    __Pyx_INCREF(__pyx_n_s_s);
    __Pyx_GIVEREF(__pyx_n_s_s);
    PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_n_s_s);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Expected_at_least_d_argument_s_g, __pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3564, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_1); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 3564, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(1, 3564, __pyx_L1_error)
  }
  __pyx_L6:;
  while (1) {
    __pyx_t_2 = PyFloat_Check(__pyx_v_arg); 
    __pyx_t_3 = (__pyx_t_2 != 0);
    if (__pyx_t_3) {
      if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, __pyx_n_s_float64_t, long, 1, __Pyx_PyInt_From_long, 1, 0, 0) < 0)) __PYX_ERR(1, 3564, __pyx_L1_error)
      goto __pyx_L10_break;
    }
    __pyx_t_2 = PyInt_Check(__pyx_v_arg); 
    __pyx_t_4 = (__pyx_t_2 != 0);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_3 = __pyx_t_4;
      goto __pyx_L13_bool_binop_done;
    }
    __pyx_t_4 = PyLong_Check(__pyx_v_arg); 
    __pyx_t_2 = (__pyx_t_4 != 0);
    __pyx_t_3 = __pyx_t_2;
    __pyx_L13_bool_binop_done:;
    __pyx_t_2 = (__pyx_t_3 != 0);
    if (__pyx_t_2) {
      if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, __pyx_n_s_int64_t, long, 1, __Pyx_PyInt_From_long, 1, 0, 0) < 0)) __PYX_ERR(1, 3564, __pyx_L1_error)
      goto __pyx_L10_break;
    }
    if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, Py_None, long, 1, __Pyx_PyInt_From_long, 1, 0, 0) < 0)) __PYX_ERR(1, 3564, __pyx_L1_error)
    goto __pyx_L10_break;
  }
  __pyx_L10_break:;
  __pyx_t_6 = PyList_New(0); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 3564, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_v_candidates = ((PyObject*)__pyx_t_6);
  __pyx_t_6 = 0;
  __pyx_t_5 = 0;
  if (unlikely(((PyObject *)__pyx_v_signatures) == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(1, 3564, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_dict_iterator(((PyObject*)__pyx_v_signatures), 1, ((PyObject *)NULL), (&__pyx_t_7), (&__pyx_t_8)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3564, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __pyx_t_6 = __pyx_t_1;
  __pyx_t_1 = 0;
  while (1) {
    __pyx_t_9 = __Pyx_dict_iter_next(__pyx_t_6, __pyx_t_7, &__pyx_t_5, &__pyx_t_1, NULL, NULL, __pyx_t_8);
    if (unlikely(__pyx_t_9 == 0)) break;
    if (unlikely(__pyx_t_9 == -1)) __PYX_ERR(1, 3564, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_sig, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_v_match_found = 0;
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_v_sig, __pyx_n_s_strip); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 3564, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __pyx_t_12 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_11))) {
      __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_11);
      if (likely(__pyx_t_12)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
        __Pyx_INCREF(__pyx_t_12);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_11, function);
      }
    }
    __pyx_t_10 = (__pyx_t_12) ? __Pyx_PyObject_Call2Args(__pyx_t_11, __pyx_t_12, __pyx_kp_s__13) : __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_kp_s__13);
    __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
    if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 3564, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_split); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 3564, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_t_10 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_11))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_11);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_11, function);
      }
    }
    __pyx_t_1 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_11, __pyx_t_10, __pyx_kp_s__14) : __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_kp_s__14);
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3564, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_XDECREF_SET(__pyx_v_src_sig, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_13 = PyList_GET_SIZE(__pyx_v_dest_sig); if (unlikely(__pyx_t_13 == ((Py_ssize_t)-1))) __PYX_ERR(1, 3564, __pyx_L1_error)
    __pyx_t_14 = __pyx_t_13;
    for (__pyx_t_15 = 0; __pyx_t_15 < __pyx_t_14; __pyx_t_15+=1) {
      __pyx_v_i = __pyx_t_15;
      __pyx_t_1 = PyList_GET_ITEM(__pyx_v_dest_sig, __pyx_v_i);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_XDECREF_SET(__pyx_v_dst_type, __pyx_t_1);
      __pyx_t_1 = 0;
      __pyx_t_2 = (__pyx_v_dst_type != Py_None);
      __pyx_t_3 = (__pyx_t_2 != 0);
      if (__pyx_t_3) {
        __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_src_sig, __pyx_v_i, Py_ssize_t, 1, PyInt_FromSsize_t, 0, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3564, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_11 = PyObject_RichCompare(__pyx_t_1, __pyx_v_dst_type, Py_EQ); __Pyx_XGOTREF(__pyx_t_11); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 3564, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_11); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(1, 3564, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        if (__pyx_t_3) {
          __pyx_v_match_found = 1;
          goto __pyx_L20;
        }
        /*else*/ {
          __pyx_v_match_found = 0;
          goto __pyx_L18_break;
        }
        __pyx_L20:;
      }
    }
    __pyx_L18_break:;
    __pyx_t_3 = (__pyx_v_match_found != 0);
    if (__pyx_t_3) {
      __pyx_t_16 = __Pyx_PyList_Append(__pyx_v_candidates, __pyx_v_sig); if (unlikely(__pyx_t_16 == ((int)-1))) __PYX_ERR(1, 3564, __pyx_L1_error)
    }
  }
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_3 = (PyList_GET_SIZE(__pyx_v_candidates) != 0);
  __pyx_t_2 = ((!__pyx_t_3) != 0);
  if (__pyx_t_2) {
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__15, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 3564, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(1, 3564, __pyx_L1_error)
  }
  __pyx_t_7 = PyList_GET_SIZE(__pyx_v_candidates); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1))) __PYX_ERR(1, 3564, __pyx_L1_error)
  __pyx_t_2 = ((__pyx_t_7 > 1) != 0);
  if (__pyx_t_2) {
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__16, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 3564, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(1, 3564, __pyx_L1_error)
  }
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    if (unlikely(((PyObject *)__pyx_v_signatures) == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 3564, __pyx_L1_error)
    }
    __pyx_t_6 = __Pyx_PyDict_GetItem(((PyObject*)__pyx_v_signatures), PyList_GET_ITEM(__pyx_v_candidates, 0)); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 3564, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_r = __pyx_t_6;
    __pyx_t_6 = 0;
    goto __pyx_L0;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("pandas._libs.interval.Uint64ClosedNeitherIntervalNode.__pyx_fused_cpdef", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_dest_sig);
  __Pyx_XDECREF(__pyx_v_ndarray);
  __Pyx_XDECREF(__pyx_v_arg);
  __Pyx_XDECREF(__pyx_v_candidates);
  __Pyx_XDECREF(__pyx_v_sig);
  __Pyx_XDECREF(__pyx_v_src_sig);
  __Pyx_XDECREF(__pyx_v_dst_type);
  __Pyx_XDECREF(__pyx_v_kwargs);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pw_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_9__pyx_fuse_0query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_fuse_0__pyx_f_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float64_t __pyx_v_point, int __pyx_skip_dispatch) {
  __Pyx_memviewslice __pyx_v_indices = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_values = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __pyx_t_5numpy_int64_t __pyx_t_9;
  __pyx_t_5numpy_int64_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  int __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  __Pyx_memviewslice __pyx_t_16 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_17 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  Py_ssize_t __pyx_t_21;
  __Pyx_RefNannySetupContext("__pyx_fuse_0query", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pyx_fuse_0query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3564, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_9__pyx_fuse_0query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = PyFloat_FromDouble(__pyx_v_point); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 3564, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3564, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3564, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 3564, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_5) {
            __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3564, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3564, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_3query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_4 = PyFloat_FromDouble(__pyx_v_point); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 3564, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_7 = __pyx_t_1; __pyx_t_3 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3564, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3564, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 3564, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (__pyx_t_3) {
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_t_4);
          __pyx_t_4 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3564, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pandas/_libs/intervaltree.pxi":3573
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
  __pyx_t_8 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_8) {

    /* "pandas/_libs/intervaltree.pxi":3577
 *             # continue the binary tree structure. Instead, we use linear
 *             # search.
 *             for i in range(self.n_elements):             # <<<<<<<<<<<<<<
 *                 if self.left[i] < point < self.right[i]:
 *                     result.append(self.indices[i])
 */
    __pyx_t_9 = __pyx_v_self->n_elements;
    __pyx_t_10 = __pyx_t_9;
    for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
      __pyx_v_i = __pyx_t_11;

      /* "pandas/_libs/intervaltree.pxi":3578
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point < self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      __pyx_t_12 = __pyx_v_i;
      __pyx_t_8 = ((*((__pyx_t_5numpy_uint64_t *) ( /* dim=0 */ (__pyx_v_self->left.data + __pyx_t_12 * __pyx_v_self->left.strides[0]) ))) < __pyx_v_point);
      if (__pyx_t_8) {
        __pyx_t_13 = __pyx_v_i;
        __pyx_t_8 = (__pyx_v_point < (*((__pyx_t_5numpy_uint64_t *) ( /* dim=0 */ (__pyx_v_self->right.data + __pyx_t_13 * __pyx_v_self->right.strides[0]) ))));
      }
      __pyx_t_14 = (__pyx_t_8 != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":3579
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point < self.right[i]:
 *                     result.append(self.indices[i])             # <<<<<<<<<<<<<<
 *         else:
 *             # There are child nodes. Based on comparing our query to the pivot,
 */
        __pyx_t_15 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->indices.data + __pyx_t_15 * __pyx_v_self->indices.strides[0]) ))));

        /* "pandas/_libs/intervaltree.pxi":3578
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point < self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      }
    }

    /* "pandas/_libs/intervaltree.pxi":3573
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
    goto __pyx_L3;
  }

  /* "pandas/_libs/intervaltree.pxi":3583
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
  /*else*/ {
    __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":3584
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:
 *                 values = self.center_left_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 */
      __pyx_t_16 = __pyx_v_self->center_left_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":3585
 *             if point < self.pivot:
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:
 */
      __pyx_t_17 = __pyx_v_self->center_left_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":3586
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):             # <<<<<<<<<<<<<<
 *                     if not values[i] < point:
 *                         break
 */
      __pyx_t_9 = __pyx_v_self->n_center;
      __pyx_t_10 = __pyx_t_9;
      for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":3587
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_18 = __pyx_v_i;
        __pyx_t_14 = ((!(((*((__pyx_t_5numpy_uint64_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_18 * __pyx_v_values.strides[0]) ))) < __pyx_v_point) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":3588
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:
 */
          goto __pyx_L9_break;

          /* "pandas/_libs/intervaltree.pxi":3587
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":3589
 *                     if not values[i] < point:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 */
        __pyx_t_19 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_19 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L9_break:;

      /* "pandas/_libs/intervaltree.pxi":3590
 *                         break
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->left_node->max_right) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":3591
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)             # <<<<<<<<<<<<<<
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode *)__pyx_v_self->left_node->__pyx_vtab)->__pyx_fuse_0query(__pyx_v_self->left_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3591, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":3590
 *                         break
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":3583
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":3592
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
    __pyx_t_14 = ((__pyx_v_point > __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":3593
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 *                 values = self.center_right_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 */
      __pyx_t_16 = __pyx_v_self->center_right_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":3594
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:
 */
      __pyx_t_17 = __pyx_v_self->center_right_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":3595
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):             # <<<<<<<<<<<<<<
 *                     if not point < values[i]:
 *                         break
 */
      for (__pyx_t_11 = (__pyx_v_self->n_center - 1); __pyx_t_11 > -1L; __pyx_t_11-=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":3596
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_20 = __pyx_v_i;
        __pyx_t_14 = ((!((__pyx_v_point < (*((__pyx_t_5numpy_uint64_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_20 * __pyx_v_values.strides[0]) )))) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":3597
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:
 */
          goto __pyx_L13_break;

          /* "pandas/_libs/intervaltree.pxi":3596
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":3598
 *                     if not point < values[i]:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if self.right_node.min_left < point:
 *                     self.right_node.query(result, point)
 */
        __pyx_t_21 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_21 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L13_break:;

      /* "pandas/_libs/intervaltree.pxi":3599
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      __pyx_t_14 = ((__pyx_v_self->right_node->min_left < __pyx_v_point) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":3600
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:
 *                     self.right_node.query(result, point)             # <<<<<<<<<<<<<<
 *             else:
 *                 result.extend(self.center_left_indices)
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode *)__pyx_v_self->right_node->__pyx_vtab)->__pyx_fuse_0query(__pyx_v_self->right_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3600, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":3599
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":3592
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":3602
 *                     self.right_node.query(result, point)
 *             else:
 *                 result.extend(self.center_left_indices)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
    /*else*/ {
      __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->extend(__pyx_v_result, __pyx_v_self->center_left_indices); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3602, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_L7:;
  }
  __pyx_L3:;

  /* "pandas/_libs/intervaltree.pxi":3564
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __PYX_XDEC_MEMVIEW(&__pyx_t_16, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_17, 1);
  __Pyx_AddTraceback("pandas._libs.interval.Uint64ClosedNeitherIntervalNode.query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_indices, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_values, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_9__pyx_fuse_0query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_0__pyx_mdef_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_9__pyx_fuse_0query = {"__pyx_fuse_0query", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_9__pyx_fuse_0query, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_2query};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_9__pyx_fuse_0query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result = 0;
  __pyx_t_5numpy_float64_t __pyx_v_point;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fuse_0query (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_result,&__pyx_n_s_point,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_result)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_point)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_0query", 1, 2, 2, 1); __PYX_ERR(1, 3564, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_0query") < 0)) __PYX_ERR(1, 3564, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_result = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)values[0]);
    __pyx_v_point = __pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_point == ((npy_float64)-1)) && PyErr_Occurred())) __PYX_ERR(1, 3564, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_0query", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 3564, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Uint64ClosedNeitherIntervalNode.__pyx_fuse_0query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_result), __pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector, 1, "result", 0))) __PYX_ERR(1, 3564, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_8__pyx_fuse_0query(((struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode *)__pyx_v_self), __pyx_v_result, __pyx_v_point);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_8__pyx_fuse_0query(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float64_t __pyx_v_point) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__pyx_fuse_0query", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_fuse_0__pyx_f_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_query(__pyx_v_self, __pyx_v_result, __pyx_v_point, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3564, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Uint64ClosedNeitherIntervalNode.__pyx_fuse_0query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pw_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_11__pyx_fuse_1query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_fuse_1__pyx_f_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float32_t __pyx_v_point, int __pyx_skip_dispatch) {
  __Pyx_memviewslice __pyx_v_indices = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_values = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __pyx_t_5numpy_int64_t __pyx_t_9;
  __pyx_t_5numpy_int64_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  int __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  __Pyx_memviewslice __pyx_t_16 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_17 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  Py_ssize_t __pyx_t_21;
  __Pyx_RefNannySetupContext("__pyx_fuse_1query", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pyx_fuse_1query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3564, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_11__pyx_fuse_1query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = PyFloat_FromDouble(__pyx_v_point); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 3564, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3564, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3564, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 3564, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_5) {
            __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3564, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3564, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_3query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_4 = PyFloat_FromDouble(__pyx_v_point); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 3564, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_7 = __pyx_t_1; __pyx_t_3 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3564, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3564, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 3564, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (__pyx_t_3) {
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_t_4);
          __pyx_t_4 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3564, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pandas/_libs/intervaltree.pxi":3573
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
  __pyx_t_8 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_8) {

    /* "pandas/_libs/intervaltree.pxi":3577
 *             # continue the binary tree structure. Instead, we use linear
 *             # search.
 *             for i in range(self.n_elements):             # <<<<<<<<<<<<<<
 *                 if self.left[i] < point < self.right[i]:
 *                     result.append(self.indices[i])
 */
    __pyx_t_9 = __pyx_v_self->n_elements;
    __pyx_t_10 = __pyx_t_9;
    for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
      __pyx_v_i = __pyx_t_11;

      /* "pandas/_libs/intervaltree.pxi":3578
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point < self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      __pyx_t_12 = __pyx_v_i;
      __pyx_t_8 = ((*((__pyx_t_5numpy_uint64_t *) ( /* dim=0 */ (__pyx_v_self->left.data + __pyx_t_12 * __pyx_v_self->left.strides[0]) ))) < __pyx_v_point);
      if (__pyx_t_8) {
        __pyx_t_13 = __pyx_v_i;
        __pyx_t_8 = (__pyx_v_point < (*((__pyx_t_5numpy_uint64_t *) ( /* dim=0 */ (__pyx_v_self->right.data + __pyx_t_13 * __pyx_v_self->right.strides[0]) ))));
      }
      __pyx_t_14 = (__pyx_t_8 != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":3579
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point < self.right[i]:
 *                     result.append(self.indices[i])             # <<<<<<<<<<<<<<
 *         else:
 *             # There are child nodes. Based on comparing our query to the pivot,
 */
        __pyx_t_15 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->indices.data + __pyx_t_15 * __pyx_v_self->indices.strides[0]) ))));

        /* "pandas/_libs/intervaltree.pxi":3578
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point < self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      }
    }

    /* "pandas/_libs/intervaltree.pxi":3573
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
    goto __pyx_L3;
  }

  /* "pandas/_libs/intervaltree.pxi":3583
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
  /*else*/ {
    __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":3584
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:
 *                 values = self.center_left_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 */
      __pyx_t_16 = __pyx_v_self->center_left_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":3585
 *             if point < self.pivot:
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:
 */
      __pyx_t_17 = __pyx_v_self->center_left_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":3586
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):             # <<<<<<<<<<<<<<
 *                     if not values[i] < point:
 *                         break
 */
      __pyx_t_9 = __pyx_v_self->n_center;
      __pyx_t_10 = __pyx_t_9;
      for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":3587
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_18 = __pyx_v_i;
        __pyx_t_14 = ((!(((*((__pyx_t_5numpy_uint64_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_18 * __pyx_v_values.strides[0]) ))) < __pyx_v_point) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":3588
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:
 */
          goto __pyx_L9_break;

          /* "pandas/_libs/intervaltree.pxi":3587
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":3589
 *                     if not values[i] < point:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 */
        __pyx_t_19 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_19 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L9_break:;

      /* "pandas/_libs/intervaltree.pxi":3590
 *                         break
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->left_node->max_right) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":3591
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)             # <<<<<<<<<<<<<<
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode *)__pyx_v_self->left_node->__pyx_vtab)->__pyx_fuse_1query(__pyx_v_self->left_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3591, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":3590
 *                         break
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":3583
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":3592
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
    __pyx_t_14 = ((__pyx_v_point > __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":3593
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 *                 values = self.center_right_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 */
      __pyx_t_16 = __pyx_v_self->center_right_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":3594
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:
 */
      __pyx_t_17 = __pyx_v_self->center_right_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":3595
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):             # <<<<<<<<<<<<<<
 *                     if not point < values[i]:
 *                         break
 */
      for (__pyx_t_11 = (__pyx_v_self->n_center - 1); __pyx_t_11 > -1L; __pyx_t_11-=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":3596
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_20 = __pyx_v_i;
        __pyx_t_14 = ((!((__pyx_v_point < (*((__pyx_t_5numpy_uint64_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_20 * __pyx_v_values.strides[0]) )))) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":3597
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:
 */
          goto __pyx_L13_break;

          /* "pandas/_libs/intervaltree.pxi":3596
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":3598
 *                     if not point < values[i]:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if self.right_node.min_left < point:
 *                     self.right_node.query(result, point)
 */
        __pyx_t_21 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_21 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L13_break:;

      /* "pandas/_libs/intervaltree.pxi":3599
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      __pyx_t_14 = ((__pyx_v_self->right_node->min_left < __pyx_v_point) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":3600
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:
 *                     self.right_node.query(result, point)             # <<<<<<<<<<<<<<
 *             else:
 *                 result.extend(self.center_left_indices)
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode *)__pyx_v_self->right_node->__pyx_vtab)->__pyx_fuse_1query(__pyx_v_self->right_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3600, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":3599
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":3592
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":3602
 *                     self.right_node.query(result, point)
 *             else:
 *                 result.extend(self.center_left_indices)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
    /*else*/ {
      __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->extend(__pyx_v_result, __pyx_v_self->center_left_indices); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3602, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_L7:;
  }
  __pyx_L3:;

  /* "pandas/_libs/intervaltree.pxi":3564
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __PYX_XDEC_MEMVIEW(&__pyx_t_16, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_17, 1);
  __Pyx_AddTraceback("pandas._libs.interval.Uint64ClosedNeitherIntervalNode.query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_indices, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_values, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_11__pyx_fuse_1query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_1__pyx_mdef_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_11__pyx_fuse_1query = {"__pyx_fuse_1query", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_11__pyx_fuse_1query, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_2query};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_11__pyx_fuse_1query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result = 0;
  __pyx_t_5numpy_float32_t __pyx_v_point;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fuse_1query (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_result,&__pyx_n_s_point,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_result)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_point)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_1query", 1, 2, 2, 1); __PYX_ERR(1, 3564, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_1query") < 0)) __PYX_ERR(1, 3564, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_result = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)values[0]);
    __pyx_v_point = __pyx_PyFloat_AsFloat(values[1]); if (unlikely((__pyx_v_point == ((npy_float32)-1)) && PyErr_Occurred())) __PYX_ERR(1, 3564, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_1query", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 3564, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Uint64ClosedNeitherIntervalNode.__pyx_fuse_1query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_result), __pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector, 1, "result", 0))) __PYX_ERR(1, 3564, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_10__pyx_fuse_1query(((struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode *)__pyx_v_self), __pyx_v_result, __pyx_v_point);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_10__pyx_fuse_1query(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_float32_t __pyx_v_point) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__pyx_fuse_1query", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_fuse_1__pyx_f_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_query(__pyx_v_self, __pyx_v_result, __pyx_v_point, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3564, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Uint64ClosedNeitherIntervalNode.__pyx_fuse_1query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pw_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_13__pyx_fuse_2query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_fuse_2__pyx_f_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int64_t __pyx_v_point, int __pyx_skip_dispatch) {
  __Pyx_memviewslice __pyx_v_indices = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_values = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __pyx_t_5numpy_int64_t __pyx_t_9;
  __pyx_t_5numpy_int64_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  int __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  __Pyx_memviewslice __pyx_t_16 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_17 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  Py_ssize_t __pyx_t_21;
  __Pyx_RefNannySetupContext("__pyx_fuse_2query", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pyx_fuse_2query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3564, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_13__pyx_fuse_2query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = __Pyx_PyInt_From_npy_int64(__pyx_v_point); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 3564, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3564, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3564, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 3564, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_5) {
            __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3564, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3564, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_3query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_4 = __Pyx_PyInt_From_npy_int64(__pyx_v_point); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 3564, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_7 = __pyx_t_1; __pyx_t_3 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3564, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3564, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 3564, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (__pyx_t_3) {
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_t_4);
          __pyx_t_4 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3564, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pandas/_libs/intervaltree.pxi":3573
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
  __pyx_t_8 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_8) {

    /* "pandas/_libs/intervaltree.pxi":3577
 *             # continue the binary tree structure. Instead, we use linear
 *             # search.
 *             for i in range(self.n_elements):             # <<<<<<<<<<<<<<
 *                 if self.left[i] < point < self.right[i]:
 *                     result.append(self.indices[i])
 */
    __pyx_t_9 = __pyx_v_self->n_elements;
    __pyx_t_10 = __pyx_t_9;
    for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
      __pyx_v_i = __pyx_t_11;

      /* "pandas/_libs/intervaltree.pxi":3578
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point < self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      __pyx_t_12 = __pyx_v_i;
      __pyx_t_8 = ((*((__pyx_t_5numpy_uint64_t *) ( /* dim=0 */ (__pyx_v_self->left.data + __pyx_t_12 * __pyx_v_self->left.strides[0]) ))) < __pyx_v_point);
      if (__pyx_t_8) {
        __pyx_t_13 = __pyx_v_i;
        __pyx_t_8 = (__pyx_v_point < (*((__pyx_t_5numpy_uint64_t *) ( /* dim=0 */ (__pyx_v_self->right.data + __pyx_t_13 * __pyx_v_self->right.strides[0]) ))));
      }
      __pyx_t_14 = (__pyx_t_8 != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":3579
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point < self.right[i]:
 *                     result.append(self.indices[i])             # <<<<<<<<<<<<<<
 *         else:
 *             # There are child nodes. Based on comparing our query to the pivot,
 */
        __pyx_t_15 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->indices.data + __pyx_t_15 * __pyx_v_self->indices.strides[0]) ))));

        /* "pandas/_libs/intervaltree.pxi":3578
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point < self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      }
    }

    /* "pandas/_libs/intervaltree.pxi":3573
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
    goto __pyx_L3;
  }

  /* "pandas/_libs/intervaltree.pxi":3583
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
  /*else*/ {
    __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":3584
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:
 *                 values = self.center_left_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 */
      __pyx_t_16 = __pyx_v_self->center_left_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":3585
 *             if point < self.pivot:
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:
 */
      __pyx_t_17 = __pyx_v_self->center_left_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":3586
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):             # <<<<<<<<<<<<<<
 *                     if not values[i] < point:
 *                         break
 */
      __pyx_t_9 = __pyx_v_self->n_center;
      __pyx_t_10 = __pyx_t_9;
      for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":3587
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_18 = __pyx_v_i;
        __pyx_t_14 = ((!(((*((__pyx_t_5numpy_uint64_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_18 * __pyx_v_values.strides[0]) ))) < __pyx_v_point) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":3588
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:
 */
          goto __pyx_L9_break;

          /* "pandas/_libs/intervaltree.pxi":3587
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":3589
 *                     if not values[i] < point:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 */
        __pyx_t_19 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_19 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L9_break:;

      /* "pandas/_libs/intervaltree.pxi":3590
 *                         break
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->left_node->max_right) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":3591
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)             # <<<<<<<<<<<<<<
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode *)__pyx_v_self->left_node->__pyx_vtab)->__pyx_fuse_2query(__pyx_v_self->left_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3591, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":3590
 *                         break
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":3583
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":3592
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
    __pyx_t_14 = ((__pyx_v_point > __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":3593
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 *                 values = self.center_right_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 */
      __pyx_t_16 = __pyx_v_self->center_right_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":3594
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:
 */
      __pyx_t_17 = __pyx_v_self->center_right_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":3595
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):             # <<<<<<<<<<<<<<
 *                     if not point < values[i]:
 *                         break
 */
      for (__pyx_t_11 = (__pyx_v_self->n_center - 1); __pyx_t_11 > -1L; __pyx_t_11-=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":3596
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_20 = __pyx_v_i;
        __pyx_t_14 = ((!((__pyx_v_point < (*((__pyx_t_5numpy_uint64_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_20 * __pyx_v_values.strides[0]) )))) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":3597
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:
 */
          goto __pyx_L13_break;

          /* "pandas/_libs/intervaltree.pxi":3596
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":3598
 *                     if not point < values[i]:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if self.right_node.min_left < point:
 *                     self.right_node.query(result, point)
 */
        __pyx_t_21 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_21 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L13_break:;

      /* "pandas/_libs/intervaltree.pxi":3599
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      __pyx_t_14 = ((__pyx_v_self->right_node->min_left < __pyx_v_point) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":3600
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:
 *                     self.right_node.query(result, point)             # <<<<<<<<<<<<<<
 *             else:
 *                 result.extend(self.center_left_indices)
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode *)__pyx_v_self->right_node->__pyx_vtab)->__pyx_fuse_2query(__pyx_v_self->right_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3600, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":3599
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":3592
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":3602
 *                     self.right_node.query(result, point)
 *             else:
 *                 result.extend(self.center_left_indices)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
    /*else*/ {
      __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->extend(__pyx_v_result, __pyx_v_self->center_left_indices); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3602, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_L7:;
  }
  __pyx_L3:;

  /* "pandas/_libs/intervaltree.pxi":3564
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __PYX_XDEC_MEMVIEW(&__pyx_t_16, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_17, 1);
  __Pyx_AddTraceback("pandas._libs.interval.Uint64ClosedNeitherIntervalNode.query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_indices, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_values, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_13__pyx_fuse_2query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_2__pyx_mdef_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_13__pyx_fuse_2query = {"__pyx_fuse_2query", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_13__pyx_fuse_2query, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_2query};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_13__pyx_fuse_2query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result = 0;
  __pyx_t_5numpy_int64_t __pyx_v_point;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fuse_2query (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_result,&__pyx_n_s_point,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_result)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_point)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_2query", 1, 2, 2, 1); __PYX_ERR(1, 3564, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_2query") < 0)) __PYX_ERR(1, 3564, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_result = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)values[0]);
    __pyx_v_point = __Pyx_PyInt_As_npy_int64(values[1]); if (unlikely((__pyx_v_point == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(1, 3564, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_2query", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 3564, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Uint64ClosedNeitherIntervalNode.__pyx_fuse_2query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_result), __pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector, 1, "result", 0))) __PYX_ERR(1, 3564, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_12__pyx_fuse_2query(((struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode *)__pyx_v_self), __pyx_v_result, __pyx_v_point);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_12__pyx_fuse_2query(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int64_t __pyx_v_point) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__pyx_fuse_2query", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_fuse_2__pyx_f_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_query(__pyx_v_self, __pyx_v_result, __pyx_v_point, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3564, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Uint64ClosedNeitherIntervalNode.__pyx_fuse_2query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pw_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_15__pyx_fuse_3query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_fuse_3__pyx_f_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int32_t __pyx_v_point, int __pyx_skip_dispatch) {
  __Pyx_memviewslice __pyx_v_indices = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_values = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __pyx_t_5numpy_int64_t __pyx_t_9;
  __pyx_t_5numpy_int64_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  int __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  __Pyx_memviewslice __pyx_t_16 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_17 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  Py_ssize_t __pyx_t_21;
  __Pyx_RefNannySetupContext("__pyx_fuse_3query", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pyx_fuse_3query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3564, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_15__pyx_fuse_3query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = __Pyx_PyInt_From_npy_int32(__pyx_v_point); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 3564, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3564, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3564, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 3564, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_5) {
            __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3564, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3564, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_3query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_4 = __Pyx_PyInt_From_npy_int32(__pyx_v_point); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 3564, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_7 = __pyx_t_1; __pyx_t_3 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3564, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3564, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 3564, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (__pyx_t_3) {
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_t_4);
          __pyx_t_4 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3564, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pandas/_libs/intervaltree.pxi":3573
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
  __pyx_t_8 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_8) {

    /* "pandas/_libs/intervaltree.pxi":3577
 *             # continue the binary tree structure. Instead, we use linear
 *             # search.
 *             for i in range(self.n_elements):             # <<<<<<<<<<<<<<
 *                 if self.left[i] < point < self.right[i]:
 *                     result.append(self.indices[i])
 */
    __pyx_t_9 = __pyx_v_self->n_elements;
    __pyx_t_10 = __pyx_t_9;
    for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
      __pyx_v_i = __pyx_t_11;

      /* "pandas/_libs/intervaltree.pxi":3578
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point < self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      __pyx_t_12 = __pyx_v_i;
      __pyx_t_8 = ((*((__pyx_t_5numpy_uint64_t *) ( /* dim=0 */ (__pyx_v_self->left.data + __pyx_t_12 * __pyx_v_self->left.strides[0]) ))) < __pyx_v_point);
      if (__pyx_t_8) {
        __pyx_t_13 = __pyx_v_i;
        __pyx_t_8 = (__pyx_v_point < (*((__pyx_t_5numpy_uint64_t *) ( /* dim=0 */ (__pyx_v_self->right.data + __pyx_t_13 * __pyx_v_self->right.strides[0]) ))));
      }
      __pyx_t_14 = (__pyx_t_8 != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":3579
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point < self.right[i]:
 *                     result.append(self.indices[i])             # <<<<<<<<<<<<<<
 *         else:
 *             # There are child nodes. Based on comparing our query to the pivot,
 */
        __pyx_t_15 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->indices.data + __pyx_t_15 * __pyx_v_self->indices.strides[0]) ))));

        /* "pandas/_libs/intervaltree.pxi":3578
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point < self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      }
    }

    /* "pandas/_libs/intervaltree.pxi":3573
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
    goto __pyx_L3;
  }

  /* "pandas/_libs/intervaltree.pxi":3583
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
  /*else*/ {
    __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":3584
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:
 *                 values = self.center_left_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 */
      __pyx_t_16 = __pyx_v_self->center_left_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":3585
 *             if point < self.pivot:
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:
 */
      __pyx_t_17 = __pyx_v_self->center_left_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":3586
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):             # <<<<<<<<<<<<<<
 *                     if not values[i] < point:
 *                         break
 */
      __pyx_t_9 = __pyx_v_self->n_center;
      __pyx_t_10 = __pyx_t_9;
      for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":3587
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_18 = __pyx_v_i;
        __pyx_t_14 = ((!(((*((__pyx_t_5numpy_uint64_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_18 * __pyx_v_values.strides[0]) ))) < __pyx_v_point) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":3588
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:
 */
          goto __pyx_L9_break;

          /* "pandas/_libs/intervaltree.pxi":3587
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":3589
 *                     if not values[i] < point:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 */
        __pyx_t_19 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_19 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L9_break:;

      /* "pandas/_libs/intervaltree.pxi":3590
 *                         break
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->left_node->max_right) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":3591
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)             # <<<<<<<<<<<<<<
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode *)__pyx_v_self->left_node->__pyx_vtab)->__pyx_fuse_3query(__pyx_v_self->left_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3591, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":3590
 *                         break
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":3583
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":3592
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
    __pyx_t_14 = ((__pyx_v_point > __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":3593
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 *                 values = self.center_right_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 */
      __pyx_t_16 = __pyx_v_self->center_right_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":3594
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:
 */
      __pyx_t_17 = __pyx_v_self->center_right_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":3595
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):             # <<<<<<<<<<<<<<
 *                     if not point < values[i]:
 *                         break
 */
      for (__pyx_t_11 = (__pyx_v_self->n_center - 1); __pyx_t_11 > -1L; __pyx_t_11-=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":3596
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_20 = __pyx_v_i;
        __pyx_t_14 = ((!((__pyx_v_point < (*((__pyx_t_5numpy_uint64_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_20 * __pyx_v_values.strides[0]) )))) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":3597
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:
 */
          goto __pyx_L13_break;

          /* "pandas/_libs/intervaltree.pxi":3596
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":3598
 *                     if not point < values[i]:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if self.right_node.min_left < point:
 *                     self.right_node.query(result, point)
 */
        __pyx_t_21 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_21 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L13_break:;

      /* "pandas/_libs/intervaltree.pxi":3599
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      __pyx_t_14 = ((__pyx_v_self->right_node->min_left < __pyx_v_point) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":3600
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:
 *                     self.right_node.query(result, point)             # <<<<<<<<<<<<<<
 *             else:
 *                 result.extend(self.center_left_indices)
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode *)__pyx_v_self->right_node->__pyx_vtab)->__pyx_fuse_3query(__pyx_v_self->right_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3600, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":3599
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":3592
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":3602
 *                     self.right_node.query(result, point)
 *             else:
 *                 result.extend(self.center_left_indices)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
    /*else*/ {
      __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->extend(__pyx_v_result, __pyx_v_self->center_left_indices); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3602, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_L7:;
  }
  __pyx_L3:;

  /* "pandas/_libs/intervaltree.pxi":3564
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __PYX_XDEC_MEMVIEW(&__pyx_t_16, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_17, 1);
  __Pyx_AddTraceback("pandas._libs.interval.Uint64ClosedNeitherIntervalNode.query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_indices, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_values, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_15__pyx_fuse_3query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_3__pyx_mdef_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_15__pyx_fuse_3query = {"__pyx_fuse_3query", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_15__pyx_fuse_3query, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_2query};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_15__pyx_fuse_3query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result = 0;
  __pyx_t_5numpy_int32_t __pyx_v_point;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fuse_3query (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_result,&__pyx_n_s_point,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_result)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_point)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_3query", 1, 2, 2, 1); __PYX_ERR(1, 3564, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_3query") < 0)) __PYX_ERR(1, 3564, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_result = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)values[0]);
    __pyx_v_point = __Pyx_PyInt_As_npy_int32(values[1]); if (unlikely((__pyx_v_point == ((npy_int32)-1)) && PyErr_Occurred())) __PYX_ERR(1, 3564, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_3query", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 3564, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Uint64ClosedNeitherIntervalNode.__pyx_fuse_3query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_result), __pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector, 1, "result", 0))) __PYX_ERR(1, 3564, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_14__pyx_fuse_3query(((struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode *)__pyx_v_self), __pyx_v_result, __pyx_v_point);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_14__pyx_fuse_3query(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_int32_t __pyx_v_point) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__pyx_fuse_3query", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_fuse_3__pyx_f_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_query(__pyx_v_self, __pyx_v_result, __pyx_v_point, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3564, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Uint64ClosedNeitherIntervalNode.__pyx_fuse_3query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pw_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_17__pyx_fuse_4query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_3query(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_fuse_4__pyx_f_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_query(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_uint64_t __pyx_v_point, int __pyx_skip_dispatch) {
  __Pyx_memviewslice __pyx_v_indices = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_values = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __pyx_t_5numpy_int64_t __pyx_t_9;
  __pyx_t_5numpy_int64_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  int __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  __Pyx_memviewslice __pyx_t_16 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_17 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  Py_ssize_t __pyx_t_21;
  __Pyx_RefNannySetupContext("__pyx_fuse_4query", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pyx_fuse_4query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3564, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_17__pyx_fuse_4query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = __Pyx_PyInt_From_npy_uint64(__pyx_v_point); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 3564, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3564, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_result), __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3564, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 3564, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_5) {
            __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3564, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_query); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3564, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_3query)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_4 = __Pyx_PyInt_From_npy_uint64(__pyx_v_point); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 3564, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_7 = __pyx_t_1; __pyx_t_3 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3564, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, ((PyObject *)__pyx_v_result), __pyx_t_4};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3564, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 3564, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          if (__pyx_t_3) {
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
          }
          __Pyx_INCREF(((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_result));
          PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, ((PyObject *)__pyx_v_result));
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_t_4);
          __pyx_t_4 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3564, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pandas/_libs/intervaltree.pxi":3573
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
  __pyx_t_8 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_8) {

    /* "pandas/_libs/intervaltree.pxi":3577
 *             # continue the binary tree structure. Instead, we use linear
 *             # search.
 *             for i in range(self.n_elements):             # <<<<<<<<<<<<<<
 *                 if self.left[i] < point < self.right[i]:
 *                     result.append(self.indices[i])
 */
    __pyx_t_9 = __pyx_v_self->n_elements;
    __pyx_t_10 = __pyx_t_9;
    for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
      __pyx_v_i = __pyx_t_11;

      /* "pandas/_libs/intervaltree.pxi":3578
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point < self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      __pyx_t_12 = __pyx_v_i;
      __pyx_t_8 = ((*((__pyx_t_5numpy_uint64_t *) ( /* dim=0 */ (__pyx_v_self->left.data + __pyx_t_12 * __pyx_v_self->left.strides[0]) ))) < __pyx_v_point);
      if (__pyx_t_8) {
        __pyx_t_13 = __pyx_v_i;
        __pyx_t_8 = (__pyx_v_point < (*((__pyx_t_5numpy_uint64_t *) ( /* dim=0 */ (__pyx_v_self->right.data + __pyx_t_13 * __pyx_v_self->right.strides[0]) ))));
      }
      __pyx_t_14 = (__pyx_t_8 != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":3579
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point < self.right[i]:
 *                     result.append(self.indices[i])             # <<<<<<<<<<<<<<
 *         else:
 *             # There are child nodes. Based on comparing our query to the pivot,
 */
        __pyx_t_15 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_self->indices.data + __pyx_t_15 * __pyx_v_self->indices.strides[0]) ))));

        /* "pandas/_libs/intervaltree.pxi":3578
 *             # search.
 *             for i in range(self.n_elements):
 *                 if self.left[i] < point < self.right[i]:             # <<<<<<<<<<<<<<
 *                     result.append(self.indices[i])
 *         else:
 */
      }
    }

    /* "pandas/_libs/intervaltree.pxi":3573
 *             Py_ssize_t i
 * 
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             # Once we get down to a certain size, it doesn't make sense to
 *             # continue the binary tree structure. Instead, we use linear
 */
    goto __pyx_L3;
  }

  /* "pandas/_libs/intervaltree.pxi":3583
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
  /*else*/ {
    __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":3584
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:
 *                 values = self.center_left_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 */
      __pyx_t_16 = __pyx_v_self->center_left_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":3585
 *             if point < self.pivot:
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:
 */
      __pyx_t_17 = __pyx_v_self->center_left_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":3586
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):             # <<<<<<<<<<<<<<
 *                     if not values[i] < point:
 *                         break
 */
      __pyx_t_9 = __pyx_v_self->n_center;
      __pyx_t_10 = __pyx_t_9;
      for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":3587
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_18 = __pyx_v_i;
        __pyx_t_14 = ((!(((*((__pyx_t_5numpy_uint64_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_18 * __pyx_v_values.strides[0]) ))) < __pyx_v_point) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":3588
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:
 */
          goto __pyx_L9_break;

          /* "pandas/_libs/intervaltree.pxi":3587
 *                 indices = self.center_left_indices
 *                 for i in range(self.n_center):
 *                     if not values[i] < point:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":3589
 *                     if not values[i] < point:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 */
        __pyx_t_19 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_19 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L9_break:;

      /* "pandas/_libs/intervaltree.pxi":3590
 *                         break
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      __pyx_t_14 = ((__pyx_v_point < __pyx_v_self->left_node->max_right) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":3591
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)             # <<<<<<<<<<<<<<
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode *)__pyx_v_self->left_node->__pyx_vtab)->__pyx_fuse_4query(__pyx_v_self->left_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3591, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":3590
 *                         break
 *                     result.append(indices[i])
 *                 if point < self.left_node.max_right:             # <<<<<<<<<<<<<<
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":3583
 *             # There are child nodes. Based on comparing our query to the pivot,
 *             # look at the center values, then go to the relevant child.
 *             if point < self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_left_values
 *                 indices = self.center_left_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":3592
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
    __pyx_t_14 = ((__pyx_v_point > __pyx_v_self->pivot) != 0);
    if (__pyx_t_14) {

      /* "pandas/_libs/intervaltree.pxi":3593
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:
 *                 values = self.center_right_values             # <<<<<<<<<<<<<<
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 */
      __pyx_t_16 = __pyx_v_self->center_right_values;
      __PYX_INC_MEMVIEW(&__pyx_t_16, 1);
      __pyx_v_values = __pyx_t_16;
      __pyx_t_16.memview = NULL;
      __pyx_t_16.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":3594
 *             elif point > self.pivot:
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices             # <<<<<<<<<<<<<<
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:
 */
      __pyx_t_17 = __pyx_v_self->center_right_indices;
      __PYX_INC_MEMVIEW(&__pyx_t_17, 1);
      __pyx_v_indices = __pyx_t_17;
      __pyx_t_17.memview = NULL;
      __pyx_t_17.data = NULL;

      /* "pandas/_libs/intervaltree.pxi":3595
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):             # <<<<<<<<<<<<<<
 *                     if not point < values[i]:
 *                         break
 */
      for (__pyx_t_11 = (__pyx_v_self->n_center - 1); __pyx_t_11 > -1L; __pyx_t_11-=1) {
        __pyx_v_i = __pyx_t_11;

        /* "pandas/_libs/intervaltree.pxi":3596
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        __pyx_t_20 = __pyx_v_i;
        __pyx_t_14 = ((!((__pyx_v_point < (*((__pyx_t_5numpy_uint64_t *) ( /* dim=0 */ (__pyx_v_values.data + __pyx_t_20 * __pyx_v_values.strides[0]) )))) != 0)) != 0);
        if (__pyx_t_14) {

          /* "pandas/_libs/intervaltree.pxi":3597
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:
 *                         break             # <<<<<<<<<<<<<<
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:
 */
          goto __pyx_L13_break;

          /* "pandas/_libs/intervaltree.pxi":3596
 *                 indices = self.center_right_indices
 *                 for i in range(self.n_center - 1, -1, -1):
 *                     if not point < values[i]:             # <<<<<<<<<<<<<<
 *                         break
 *                     result.append(indices[i])
 */
        }

        /* "pandas/_libs/intervaltree.pxi":3598
 *                     if not point < values[i]:
 *                         break
 *                     result.append(indices[i])             # <<<<<<<<<<<<<<
 *                 if self.right_node.min_left < point:
 *                     self.right_node.query(result, point)
 */
        __pyx_t_21 = __pyx_v_i;
        ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->append(__pyx_v_result, (*((__pyx_t_5numpy_int64_t *) ( /* dim=0 */ (__pyx_v_indices.data + __pyx_t_21 * __pyx_v_indices.strides[0]) ))));
      }
      __pyx_L13_break:;

      /* "pandas/_libs/intervaltree.pxi":3599
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      __pyx_t_14 = ((__pyx_v_self->right_node->min_left < __pyx_v_point) != 0);
      if (__pyx_t_14) {

        /* "pandas/_libs/intervaltree.pxi":3600
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:
 *                     self.right_node.query(result, point)             # <<<<<<<<<<<<<<
 *             else:
 *                 result.extend(self.center_left_indices)
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode *)__pyx_v_self->right_node->__pyx_vtab)->__pyx_fuse_4query(__pyx_v_self->right_node, __pyx_v_result, __pyx_v_point, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3600, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pandas/_libs/intervaltree.pxi":3599
 *                         break
 *                     result.append(indices[i])
 *                 if self.right_node.min_left < point:             # <<<<<<<<<<<<<<
 *                     self.right_node.query(result, point)
 *             else:
 */
      }

      /* "pandas/_libs/intervaltree.pxi":3592
 *                 if point < self.left_node.max_right:
 *                     self.left_node.query(result, point)
 *             elif point > self.pivot:             # <<<<<<<<<<<<<<
 *                 values = self.center_right_values
 *                 indices = self.center_right_indices
 */
      goto __pyx_L7;
    }

    /* "pandas/_libs/intervaltree.pxi":3602
 *                     self.right_node.query(result, point)
 *             else:
 *                 result.extend(self.center_left_indices)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
    /*else*/ {
      __pyx_t_1 = ((struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector *)__pyx_v_result->__pyx_vtab)->extend(__pyx_v_result, __pyx_v_self->center_left_indices); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3602, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_L7:;
  }
  __pyx_L3:;

  /* "pandas/_libs/intervaltree.pxi":3564
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __PYX_XDEC_MEMVIEW(&__pyx_t_16, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_17, 1);
  __Pyx_AddTraceback("pandas._libs.interval.Uint64ClosedNeitherIntervalNode.query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_indices, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_values, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_17__pyx_fuse_4query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_4__pyx_mdef_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_17__pyx_fuse_4query = {"__pyx_fuse_4query", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_17__pyx_fuse_4query, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_2query};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_17__pyx_fuse_4query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result = 0;
  __pyx_t_5numpy_uint64_t __pyx_v_point;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fuse_4query (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_result,&__pyx_n_s_point,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_result)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_point)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fuse_4query", 1, 2, 2, 1); __PYX_ERR(1, 3564, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fuse_4query") < 0)) __PYX_ERR(1, 3564, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_result = ((struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *)values[0]);
    __pyx_v_point = __Pyx_PyInt_As_npy_uint64(values[1]); if (unlikely((__pyx_v_point == ((npy_uint64)-1)) && PyErr_Occurred())) __PYX_ERR(1, 3564, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fuse_4query", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 3564, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.Uint64ClosedNeitherIntervalNode.__pyx_fuse_4query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_result), __pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector, 1, "result", 0))) __PYX_ERR(1, 3564, __pyx_L1_error)
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_16__pyx_fuse_4query(((struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode *)__pyx_v_self), __pyx_v_result, __pyx_v_point);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_16__pyx_fuse_4query(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode *__pyx_v_self, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *__pyx_v_result, __pyx_t_5numpy_uint64_t __pyx_v_point) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__pyx_fuse_4query", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_fuse_4__pyx_f_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_query(__pyx_v_self, __pyx_v_result, __pyx_v_point, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3564, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Uint64ClosedNeitherIntervalNode.__pyx_fuse_4query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":3604
 *                 result.extend(self.center_left_indices)
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         if self.is_leaf_node:
 *             return ('<Uint64ClosedNeitherIntervalNode: '
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_5__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_5__repr__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_4__repr__(((struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_4__repr__(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode *__pyx_v_self) {
  PyObject *__pyx_v_n_left = NULL;
  PyObject *__pyx_v_n_right = NULL;
  PyObject *__pyx_v_n_center = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  Py_UCS4 __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "pandas/_libs/intervaltree.pxi":3605
 * 
 *     def __repr__(self):
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             return ('<Uint64ClosedNeitherIntervalNode: '
 *                     '%s elements (terminal)>' % self.n_elements)
 */
  __pyx_t_1 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_1) {

    /* "pandas/_libs/intervaltree.pxi":3606
 *     def __repr__(self):
 *         if self.is_leaf_node:
 *             return ('<Uint64ClosedNeitherIntervalNode: '             # <<<<<<<<<<<<<<
 *                     '%s elements (terminal)>' % self.n_elements)
 *         else:
 */
    __Pyx_XDECREF(__pyx_r);

    /* "pandas/_libs/intervaltree.pxi":3607
 *         if self.is_leaf_node:
 *             return ('<Uint64ClosedNeitherIntervalNode: '
 *                     '%s elements (terminal)>' % self.n_elements)             # <<<<<<<<<<<<<<
 *         else:
 *             n_left = self.left_node.n_elements
 */
    __pyx_t_2 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_elements); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3607, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PyUnicode_Format(__pyx_kp_u_Uint64ClosedNeitherIntervalNode, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 3607, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "pandas/_libs/intervaltree.pxi":3605
 * 
 *     def __repr__(self):
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             return ('<Uint64ClosedNeitherIntervalNode: '
 *                     '%s elements (terminal)>' % self.n_elements)
 */
  }

  /* "pandas/_libs/intervaltree.pxi":3609
 *                     '%s elements (terminal)>' % self.n_elements)
 *         else:
 *             n_left = self.left_node.n_elements             # <<<<<<<<<<<<<<
 *             n_right = self.right_node.n_elements
 *             n_center = self.n_elements - n_left - n_right
 */
  /*else*/ {
    __pyx_t_3 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->left_node->n_elements); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 3609, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_v_n_left = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "pandas/_libs/intervaltree.pxi":3610
 *         else:
 *             n_left = self.left_node.n_elements
 *             n_right = self.right_node.n_elements             # <<<<<<<<<<<<<<
 *             n_center = self.n_elements - n_left - n_right
 *             return ('<Uint64ClosedNeitherIntervalNode: '
 */
    __pyx_t_3 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->right_node->n_elements); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 3610, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_v_n_right = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "pandas/_libs/intervaltree.pxi":3611
 *             n_left = self.left_node.n_elements
 *             n_right = self.right_node.n_elements
 *             n_center = self.n_elements - n_left - n_right             # <<<<<<<<<<<<<<
 *             return ('<Uint64ClosedNeitherIntervalNode: '
 *                     'pivot %s, %s elements (%s left, %s right, %s '
 */
    __pyx_t_3 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_elements); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 3611, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = PyNumber_Subtract(__pyx_t_3, __pyx_v_n_left); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3611, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyNumber_Subtract(__pyx_t_2, __pyx_v_n_right); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 3611, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_n_center = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "pandas/_libs/intervaltree.pxi":3612
 *             n_right = self.right_node.n_elements
 *             n_center = self.n_elements - n_left - n_right
 *             return ('<Uint64ClosedNeitherIntervalNode: '             # <<<<<<<<<<<<<<
 *                     'pivot %s, %s elements (%s left, %s right, %s '
 *                     'overlapping)>' % (self.pivot, self.n_elements,
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = PyTuple_New(11); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 3612, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = 0;
    __pyx_t_5 = 127;
    __Pyx_INCREF(__pyx_kp_u_Uint64ClosedNeitherIntervalNode_2);
    __pyx_t_4 += 40;
    __Pyx_GIVEREF(__pyx_kp_u_Uint64ClosedNeitherIntervalNode_2);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_kp_u_Uint64ClosedNeitherIntervalNode_2);

    /* "pandas/_libs/intervaltree.pxi":3614
 *             return ('<Uint64ClosedNeitherIntervalNode: '
 *                     'pivot %s, %s elements (%s left, %s right, %s '
 *                     'overlapping)>' % (self.pivot, self.n_elements,             # <<<<<<<<<<<<<<
 *                                        n_left, n_right, n_center))
 * 
 */
    __pyx_t_2 = __Pyx_PyInt_From_npy_uint64(__pyx_v_self->pivot); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3614, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_t_2), __pyx_empty_unicode); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 3614, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_5 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_6) > __pyx_t_5) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_6) : __pyx_t_5;
    __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_6);
    __pyx_t_6 = 0;
    __Pyx_INCREF(__pyx_kp_u__18);
    __pyx_t_4 += 2;
    __Pyx_GIVEREF(__pyx_kp_u__18);
    PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_kp_u__18);
    __pyx_t_6 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_elements); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 3614, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_2 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_t_6), __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3614, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_5 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) > __pyx_t_5) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) : __pyx_t_5;
    __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_3, 3, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_elements);
    __pyx_t_4 += 11;
    __Pyx_GIVEREF(__pyx_kp_u_elements);
    PyTuple_SET_ITEM(__pyx_t_3, 4, __pyx_kp_u_elements);

    /* "pandas/_libs/intervaltree.pxi":3615
 *                     'pivot %s, %s elements (%s left, %s right, %s '
 *                     'overlapping)>' % (self.pivot, self.n_elements,
 *                                        n_left, n_right, n_center))             # <<<<<<<<<<<<<<
 * 
 *     def counts(self):
 */
    __pyx_t_2 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_v_n_left), __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3615, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) > __pyx_t_5) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) : __pyx_t_5;
    __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_3, 5, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_left_2);
    __pyx_t_4 += 7;
    __Pyx_GIVEREF(__pyx_kp_u_left_2);
    PyTuple_SET_ITEM(__pyx_t_3, 6, __pyx_kp_u_left_2);
    __pyx_t_2 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_v_n_right), __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3615, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) > __pyx_t_5) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) : __pyx_t_5;
    __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_3, 7, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_right_2);
    __pyx_t_4 += 8;
    __Pyx_GIVEREF(__pyx_kp_u_right_2);
    PyTuple_SET_ITEM(__pyx_t_3, 8, __pyx_kp_u_right_2);
    __pyx_t_2 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_v_n_center), __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3615, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) > __pyx_t_5) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) : __pyx_t_5;
    __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_3, 9, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_overlapping);
    __pyx_t_4 += 14;
    __Pyx_GIVEREF(__pyx_kp_u_overlapping);
    PyTuple_SET_ITEM(__pyx_t_3, 10, __pyx_kp_u_overlapping);

    /* "pandas/_libs/intervaltree.pxi":3612
 *             n_right = self.right_node.n_elements
 *             n_center = self.n_elements - n_left - n_right
 *             return ('<Uint64ClosedNeitherIntervalNode: '             # <<<<<<<<<<<<<<
 *                     'pivot %s, %s elements (%s left, %s right, %s '
 *                     'overlapping)>' % (self.pivot, self.n_elements,
 */
    __pyx_t_2 = __Pyx_PyUnicode_Join(__pyx_t_3, 11, __pyx_t_4, __pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3612, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;
  }

  /* "pandas/_libs/intervaltree.pxi":3604
 *                 result.extend(self.center_left_indices)
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         if self.is_leaf_node:
 *             return ('<Uint64ClosedNeitherIntervalNode: '
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pandas._libs.interval.Uint64ClosedNeitherIntervalNode.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_n_left);
  __Pyx_XDECREF(__pyx_v_n_right);
  __Pyx_XDECREF(__pyx_v_n_center);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":3617
 *                                        n_left, n_right, n_center))
 * 
 *     def counts(self):             # <<<<<<<<<<<<<<
 *         """
 *         Inspect counts on this node
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_7counts(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_6counts[] = "\n        Inspect counts on this node\n        useful for debugging purposes\n        ";
static PyObject *__pyx_pw_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_7counts(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("counts (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_6counts(((struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_6counts(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode *__pyx_v_self) {
  Py_ssize_t __pyx_v_m;
  PyObject *__pyx_v_l = NULL;
  PyObject *__pyx_v_r = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  size_t __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("counts", 0);

  /* "pandas/_libs/intervaltree.pxi":3622
 *         useful for debugging purposes
 *         """
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             return self.n_elements
 *         else:
 */
  __pyx_t_1 = (__pyx_v_self->is_leaf_node != 0);
  if (__pyx_t_1) {

    /* "pandas/_libs/intervaltree.pxi":3623
 *         """
 *         if self.is_leaf_node:
 *             return self.n_elements             # <<<<<<<<<<<<<<
 *         else:
 *             m = len(self.center_left_values)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_elements); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3623, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "pandas/_libs/intervaltree.pxi":3622
 *         useful for debugging purposes
 *         """
 *         if self.is_leaf_node:             # <<<<<<<<<<<<<<
 *             return self.n_elements
 *         else:
 */
  }

  /* "pandas/_libs/intervaltree.pxi":3625
 *             return self.n_elements
 *         else:
 *             m = len(self.center_left_values)             # <<<<<<<<<<<<<<
 *             l = self.left_node.counts()
 *             r = self.right_node.counts()
 */
  /*else*/ {
    if (unlikely(!__pyx_v_self->center_left_values.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(1, 3625, __pyx_L1_error)}
    __pyx_t_3 = __Pyx_MemoryView_Len(__pyx_v_self->center_left_values); 
    __pyx_v_m = __pyx_t_3;

    /* "pandas/_libs/intervaltree.pxi":3626
 *         else:
 *             m = len(self.center_left_values)
 *             l = self.left_node.counts()             # <<<<<<<<<<<<<<
 *             r = self.right_node.counts()
 *             return (m, (l, r))
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->left_node), __pyx_n_s_counts); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 3626, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3626, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_l = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "pandas/_libs/intervaltree.pxi":3627
 *             m = len(self.center_left_values)
 *             l = self.left_node.counts()
 *             r = self.right_node.counts()             # <<<<<<<<<<<<<<
 *             return (m, (l, r))
 * 
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->right_node), __pyx_n_s_counts); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 3627, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3627, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_r = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "pandas/_libs/intervaltree.pxi":3628
 *             l = self.left_node.counts()
 *             r = self.right_node.counts()
 *             return (m, (l, r))             # <<<<<<<<<<<<<<
 * 
 * NODE_CLASSES['uint64',
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = PyInt_FromSsize_t(__pyx_v_m); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3628, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 3628, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_v_l);
    __Pyx_GIVEREF(__pyx_v_l);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_l);
    __Pyx_INCREF(__pyx_v_r);
    __Pyx_GIVEREF(__pyx_v_r);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_v_r);
    __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 3628, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_4);
    __pyx_t_2 = 0;
    __pyx_t_4 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;
  }

  /* "pandas/_libs/intervaltree.pxi":3617
 *                                        n_left, n_right, n_center))
 * 
 *     def counts(self):             # <<<<<<<<<<<<<<
 *         """
 *         Inspect counts on this node
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pandas._libs.interval.Uint64ClosedNeitherIntervalNode.counts", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_l);
  __Pyx_XDECREF(__pyx_v_r);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":3474
 *         int64_t[:] center_left_indices, center_right_indices, indices
 *         uint64_t min_left, max_right
 *         readonly uint64_t pivot             # <<<<<<<<<<<<<<
 *         readonly int64_t n_elements, n_center, leaf_size
 *         readonly bint is_leaf_node
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_5pivot_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_5pivot_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_5pivot___get__(((struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_5pivot___get__(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_npy_uint64(__pyx_v_self->pivot); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3474, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Uint64ClosedNeitherIntervalNode.pivot.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":3475
 *         uint64_t min_left, max_right
 *         readonly uint64_t pivot
 *         readonly int64_t n_elements, n_center, leaf_size             # <<<<<<<<<<<<<<
 *         readonly bint is_leaf_node
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_10n_elements_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_10n_elements_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_10n_elements___get__(((struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_10n_elements___get__(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_elements); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3475, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Uint64ClosedNeitherIntervalNode.n_elements.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_8n_center_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_8n_center_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_8n_center___get__(((struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_8n_center___get__(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_center); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3475, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Uint64ClosedNeitherIntervalNode.n_center.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_9leaf_size_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_9leaf_size_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_9leaf_size___get__(((struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_9leaf_size___get__(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->leaf_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3475, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Uint64ClosedNeitherIntervalNode.leaf_size.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/intervaltree.pxi":3476
 *         readonly uint64_t pivot
 *         readonly int64_t n_elements, n_center, leaf_size
 *         readonly bint is_leaf_node             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self,
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_12is_leaf_node_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_12is_leaf_node_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_12is_leaf_node___get__(((struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_12is_leaf_node___get__(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->is_leaf_node); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3476, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Uint64ClosedNeitherIntervalNode.is_leaf_node.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_21__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_21__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_20__reduce_cython__(((struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_20__reduce_cython__(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  int __pyx_t_16;
  int __pyx_t_17;
  int __pyx_t_18;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.center_left_indices, self.center_left_values, self.center_right_indices, self.center_right_values, self.indices, self.is_leaf_node, self.leaf_size, self.left, self.left_node, self.max_right, self.min_left, self.n_center, self.n_elements, self.pivot, self.right, self.right_node)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  if (unlikely(!__pyx_v_self->center_left_indices.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_1 = __pyx_memoryview_fromslice(__pyx_v_self->center_left_indices, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_int64_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_int64_t, 0);; if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (unlikely(!__pyx_v_self->center_left_values.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_2 = __pyx_memoryview_fromslice(__pyx_v_self->center_left_values, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_uint64_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_uint64_t, 0);; if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (unlikely(!__pyx_v_self->center_right_indices.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_3 = __pyx_memoryview_fromslice(__pyx_v_self->center_right_indices, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_int64_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_int64_t, 0);; if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely(!__pyx_v_self->center_right_values.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_4 = __pyx_memoryview_fromslice(__pyx_v_self->center_right_values, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_uint64_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_uint64_t, 0);; if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (unlikely(!__pyx_v_self->indices.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_5 = __pyx_memoryview_fromslice(__pyx_v_self->indices, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_int64_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_int64_t, 0);; if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyBool_FromLong(__pyx_v_self->is_leaf_node); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->leaf_size); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (unlikely(!__pyx_v_self->left.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_8 = __pyx_memoryview_fromslice(__pyx_v_self->left, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_uint64_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_uint64_t, 0);; if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_PyInt_From_npy_uint64(__pyx_v_self->max_right); if (unlikely(!__pyx_t_9)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_10 = __Pyx_PyInt_From_npy_uint64(__pyx_v_self->min_left); if (unlikely(!__pyx_t_10)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_11 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_center); if (unlikely(!__pyx_t_11)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_12 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->n_elements); if (unlikely(!__pyx_t_12)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __pyx_t_13 = __Pyx_PyInt_From_npy_uint64(__pyx_v_self->pivot); if (unlikely(!__pyx_t_13)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  if (unlikely(!__pyx_v_self->right.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 5, __pyx_L1_error)}
  __pyx_t_14 = __pyx_memoryview_fromslice(__pyx_v_self->right, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_uint64_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_uint64_t, 0);; if (unlikely(!__pyx_t_14)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __pyx_t_15 = PyTuple_New(16); if (unlikely(!__pyx_t_15)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_15, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_15, 1, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_15, 2, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_15, 3, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_15, 4, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_15, 5, __pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_15, 6, __pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_15, 7, __pyx_t_8);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->left_node));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self->left_node));
  PyTuple_SET_ITEM(__pyx_t_15, 8, ((PyObject *)__pyx_v_self->left_node));
  __Pyx_GIVEREF(__pyx_t_9);
  PyTuple_SET_ITEM(__pyx_t_15, 9, __pyx_t_9);
  __Pyx_GIVEREF(__pyx_t_10);
  PyTuple_SET_ITEM(__pyx_t_15, 10, __pyx_t_10);
  __Pyx_GIVEREF(__pyx_t_11);
  PyTuple_SET_ITEM(__pyx_t_15, 11, __pyx_t_11);
  __Pyx_GIVEREF(__pyx_t_12);
  PyTuple_SET_ITEM(__pyx_t_15, 12, __pyx_t_12);
  __Pyx_GIVEREF(__pyx_t_13);
  PyTuple_SET_ITEM(__pyx_t_15, 13, __pyx_t_13);
  __Pyx_GIVEREF(__pyx_t_14);
  PyTuple_SET_ITEM(__pyx_t_15, 14, __pyx_t_14);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->right_node));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self->right_node));
  PyTuple_SET_ITEM(__pyx_t_15, 15, ((PyObject *)__pyx_v_self->right_node));
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  __pyx_t_6 = 0;
  __pyx_t_7 = 0;
  __pyx_t_8 = 0;
  __pyx_t_9 = 0;
  __pyx_t_10 = 0;
  __pyx_t_11 = 0;
  __pyx_t_12 = 0;
  __pyx_t_13 = 0;
  __pyx_t_14 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_15);
  __pyx_t_15 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.center_left_indices, self.center_left_values, self.center_right_indices, self.center_right_values, self.indices, self.is_leaf_node, self.leaf_size, self.left, self.left_node, self.max_right, self.min_left, self.n_center, self.n_elements, self.pivot, self.right, self.right_node)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_15 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_15)) __PYX_ERR(2, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __pyx_v__dict = __pyx_t_15;
  __pyx_t_15 = 0;

  /* "(tree fragment)":7
 *     state = (self.center_left_indices, self.center_left_values, self.center_right_indices, self.center_right_values, self.indices, self.is_leaf_node, self.leaf_size, self.left, self.left_node, self.max_right, self.min_left, self.n_center, self.n_elements, self.pivot, self.right, self.right_node)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_16 = (__pyx_v__dict != Py_None);
  __pyx_t_17 = (__pyx_t_16 != 0);
  if (__pyx_t_17) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_15 = PyTuple_New(1); if (unlikely(!__pyx_t_15)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    PyTuple_SET_ITEM(__pyx_t_15, 0, __pyx_v__dict);
    __pyx_t_14 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_15); if (unlikely(!__pyx_t_14)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_14));
    __pyx_t_14 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.left_node is not None or self.right_node is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.center_left_indices, self.center_left_values, self.center_right_indices, self.center_right_values, self.indices, self.is_leaf_node, self.leaf_size, self.left, self.left_node, self.max_right, self.min_left, self.n_center, self.n_elements, self.pivot, self.right, self.right_node)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.left_node is not None or self.right_node is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_Uint64ClosedNeitherIntervalNode, (type(self), 0x7a104e5, None), state
 */
  /*else*/ {
    __pyx_t_16 = (((PyObject *)__pyx_v_self->left_node) != Py_None);
    __pyx_t_18 = (__pyx_t_16 != 0);
    if (!__pyx_t_18) {
    } else {
      __pyx_t_17 = __pyx_t_18;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_18 = (((PyObject *)__pyx_v_self->right_node) != Py_None);
    __pyx_t_16 = (__pyx_t_18 != 0);
    __pyx_t_17 = __pyx_t_16;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_17;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.left_node is not None or self.right_node is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Uint64ClosedNeitherIntervalNode, (type(self), 0x7a104e5, None), state
 *     else:
 */
  __pyx_t_17 = (__pyx_v_use_setstate != 0);
  if (__pyx_t_17) {

    /* "(tree fragment)":13
 *         use_setstate = self.left_node is not None or self.right_node is not None
 *     if use_setstate:
 *         return __pyx_unpickle_Uint64ClosedNeitherIntervalNode, (type(self), 0x7a104e5, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_Uint64ClosedNeitherIntervalNode, (type(self), 0x7a104e5, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_14, __pyx_n_s_pyx_unpickle_Uint64ClosedNeith); if (unlikely(!__pyx_t_14)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __pyx_t_15 = PyTuple_New(3); if (unlikely(!__pyx_t_15)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_15, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_127993061);
    __Pyx_GIVEREF(__pyx_int_127993061);
    PyTuple_SET_ITEM(__pyx_t_15, 1, __pyx_int_127993061);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_15, 2, Py_None);
    __pyx_t_13 = PyTuple_New(3); if (unlikely(!__pyx_t_13)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __Pyx_GIVEREF(__pyx_t_14);
    PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_t_14);
    __Pyx_GIVEREF(__pyx_t_15);
    PyTuple_SET_ITEM(__pyx_t_13, 1, __pyx_t_15);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_13, 2, __pyx_v_state);
    __pyx_t_14 = 0;
    __pyx_t_15 = 0;
    __pyx_r = __pyx_t_13;
    __pyx_t_13 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.left_node is not None or self.right_node is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Uint64ClosedNeitherIntervalNode, (type(self), 0x7a104e5, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_Uint64ClosedNeitherIntervalNode, (type(self), 0x7a104e5, None), state
 *     else:
 *         return __pyx_unpickle_Uint64ClosedNeitherIntervalNode, (type(self), 0x7a104e5, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Uint64ClosedNeitherIntervalNode__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_13, __pyx_n_s_pyx_unpickle_Uint64ClosedNeith); if (unlikely(!__pyx_t_13)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __pyx_t_15 = PyTuple_New(3); if (unlikely(!__pyx_t_15)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_15, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_127993061);
    __Pyx_GIVEREF(__pyx_int_127993061);
    PyTuple_SET_ITEM(__pyx_t_15, 1, __pyx_int_127993061);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_15, 2, __pyx_v_state);
    __pyx_t_14 = PyTuple_New(2); if (unlikely(!__pyx_t_14)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_GIVEREF(__pyx_t_13);
    PyTuple_SET_ITEM(__pyx_t_14, 0, __pyx_t_13);
    __Pyx_GIVEREF(__pyx_t_15);
    PyTuple_SET_ITEM(__pyx_t_14, 1, __pyx_t_15);
    __pyx_t_13 = 0;
    __pyx_t_15 = 0;
    __pyx_r = __pyx_t_14;
    __pyx_t_14 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_AddTraceback("pandas._libs.interval.Uint64ClosedNeitherIntervalNode.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Uint64ClosedNeitherIntervalNode, (type(self), 0x7a104e5, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Uint64ClosedNeitherIntervalNode__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_23__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_23__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_22__setstate_cython__(((struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_22__setstate_cython__(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_Uint64ClosedNeitherIntervalNode, (type(self), 0x7a104e5, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Uint64ClosedNeitherIntervalNode__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(2, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_6pandas_5_libs_8interval___pyx_unpickle_Uint64ClosedNeitherIntervalNode__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Uint64ClosedNeitherIntervalNode, (type(self), 0x7a104e5, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Uint64ClosedNeitherIntervalNode__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pandas._libs.interval.Uint64ClosedNeitherIntervalNode.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_IntervalMixin(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_3__pyx_unpickle_IntervalMixin(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pandas_5_libs_8interval_3__pyx_unpickle_IntervalMixin = {"__pyx_unpickle_IntervalMixin", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_3__pyx_unpickle_IntervalMixin, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_3__pyx_unpickle_IntervalMixin(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_IntervalMixin (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_type)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_checksum)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_IntervalMixin", 1, 3, 3, 1); __PYX_ERR(2, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_IntervalMixin", 1, 3, 3, 2); __PYX_ERR(2, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_unpickle_IntervalMixin") < 0)) __PYX_ERR(2, 1, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_IntervalMixin", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(2, 1, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.__pyx_unpickle_IntervalMixin", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_2__pyx_unpickle_IntervalMixin(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_2__pyx_unpickle_IntervalMixin(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  __Pyx_RefNannySetupContext("__pyx_unpickle_IntervalMixin", 0);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0xd41d8cd:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xd41d8cd = ())" % __pyx_checksum)
 */
  __pyx_t_1 = ((__pyx_v___pyx_checksum != 0xd41d8cd) != 0);
  if (__pyx_t_1) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0xd41d8cd:
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xd41d8cd = ())" % __pyx_checksum)
 *     __pyx_result = IntervalMixin.__new__(__pyx_type)
 */
    __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_PickleError);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_2);
    __pyx_v___pyx_PickleError = __pyx_t_2;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum != 0xd41d8cd:
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xd41d8cd = ())" % __pyx_checksum)             # <<<<<<<<<<<<<<
 *     __pyx_result = IntervalMixin.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_2 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_s_vs_0xd4, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_v___pyx_PickleError);
    __pyx_t_2 = __pyx_v___pyx_PickleError; __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_5, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(2, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0xd41d8cd:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xd41d8cd = ())" % __pyx_checksum)
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xd41d8cd = ())" % __pyx_checksum)
 *     __pyx_result = IntervalMixin.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_IntervalMixin__set_state(<IntervalMixin> __pyx_result, __pyx_state)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6pandas_5_libs_8interval_IntervalMixin), __pyx_n_s_new); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_v___pyx_type) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v___pyx_type);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v___pyx_result = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xd41d8cd = ())" % __pyx_checksum)
 *     __pyx_result = IntervalMixin.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_IntervalMixin__set_state(<IntervalMixin> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_1 = (__pyx_v___pyx_state != Py_None);
  __pyx_t_6 = (__pyx_t_1 != 0);
  if (__pyx_t_6) {

    /* "(tree fragment)":9
 *     __pyx_result = IntervalMixin.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_IntervalMixin__set_state(<IntervalMixin> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_IntervalMixin__set_state(IntervalMixin __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(2, 9, __pyx_L1_error)
    __pyx_t_3 = __pyx_f_6pandas_5_libs_8interval___pyx_unpickle_IntervalMixin__set_state(((struct __pyx_obj_6pandas_5_libs_8interval_IntervalMixin *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xd41d8cd = ())" % __pyx_checksum)
 *     __pyx_result = IntervalMixin.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_IntervalMixin__set_state(<IntervalMixin> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_IntervalMixin__set_state(<IntervalMixin> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_IntervalMixin__set_state(IntervalMixin __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_IntervalMixin(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pandas._libs.interval.__pyx_unpickle_IntervalMixin", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_IntervalMixin__set_state(<IntervalMixin> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_IntervalMixin__set_state(IntervalMixin __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */

static PyObject *__pyx_f_6pandas_5_libs_8interval___pyx_unpickle_IntervalMixin__set_state(struct __pyx_obj_6pandas_5_libs_8interval_IntervalMixin *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  __Pyx_RefNannySetupContext("__pyx_unpickle_IntervalMixin__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_IntervalMixin__set_state(IntervalMixin __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_2 = PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(2, 12, __pyx_L1_error)
  __pyx_t_3 = ((__pyx_t_2 > 0) != 0);
  if (__pyx_t_3) {
  } else {
    __pyx_t_1 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(2, 12, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_3 != 0);
  __pyx_t_1 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_IntervalMixin__set_state(IntervalMixin __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])             # <<<<<<<<<<<<<<
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_update); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(2, 13, __pyx_L1_error)
    }
    __pyx_t_6 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
      }
    }
    __pyx_t_5 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_8, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_6);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_IntervalMixin__set_state(IntervalMixin __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_IntervalMixin__set_state(<IntervalMixin> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_IntervalMixin__set_state(IntervalMixin __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("pandas._libs.interval.__pyx_unpickle_IntervalMixin__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_IntervalTree(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_5__pyx_unpickle_IntervalTree(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pandas_5_libs_8interval_5__pyx_unpickle_IntervalTree = {"__pyx_unpickle_IntervalTree", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_5__pyx_unpickle_IntervalTree, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_5__pyx_unpickle_IntervalTree(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_IntervalTree (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_type)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_checksum)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_IntervalTree", 1, 3, 3, 1); __PYX_ERR(2, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_IntervalTree", 1, 3, 3, 2); __PYX_ERR(2, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_unpickle_IntervalTree") < 0)) __PYX_ERR(2, 1, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_IntervalTree", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(2, 1, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.__pyx_unpickle_IntervalTree", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_4__pyx_unpickle_IntervalTree(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_4__pyx_unpickle_IntervalTree(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  __Pyx_RefNannySetupContext("__pyx_unpickle_IntervalTree", 0);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x05502fb:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x05502fb = (_is_overlapping, _left_sorter, _right_sorter, closed, dtype, left, right, root))" % __pyx_checksum)
 */
  __pyx_t_1 = ((__pyx_v___pyx_checksum != 0x05502fb) != 0);
  if (__pyx_t_1) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x05502fb:
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x05502fb = (_is_overlapping, _left_sorter, _right_sorter, closed, dtype, left, right, root))" % __pyx_checksum)
 *     __pyx_result = IntervalTree.__new__(__pyx_type)
 */
    __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_PickleError);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_2);
    __pyx_v___pyx_PickleError = __pyx_t_2;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum != 0x05502fb:
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x05502fb = (_is_overlapping, _left_sorter, _right_sorter, closed, dtype, left, right, root))" % __pyx_checksum)             # <<<<<<<<<<<<<<
 *     __pyx_result = IntervalTree.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_2 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_s_vs_0x05, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_v___pyx_PickleError);
    __pyx_t_2 = __pyx_v___pyx_PickleError; __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_5, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(2, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x05502fb:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x05502fb = (_is_overlapping, _left_sorter, _right_sorter, closed, dtype, left, right, root))" % __pyx_checksum)
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x05502fb = (_is_overlapping, _left_sorter, _right_sorter, closed, dtype, left, right, root))" % __pyx_checksum)
 *     __pyx_result = IntervalTree.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_IntervalTree__set_state(<IntervalTree> __pyx_result, __pyx_state)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6pandas_5_libs_8interval_IntervalTree), __pyx_n_s_new); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_v___pyx_type) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v___pyx_type);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v___pyx_result = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x05502fb = (_is_overlapping, _left_sorter, _right_sorter, closed, dtype, left, right, root))" % __pyx_checksum)
 *     __pyx_result = IntervalTree.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_IntervalTree__set_state(<IntervalTree> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_1 = (__pyx_v___pyx_state != Py_None);
  __pyx_t_6 = (__pyx_t_1 != 0);
  if (__pyx_t_6) {

    /* "(tree fragment)":9
 *     __pyx_result = IntervalTree.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_IntervalTree__set_state(<IntervalTree> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_IntervalTree__set_state(IntervalTree __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(2, 9, __pyx_L1_error)
    __pyx_t_3 = __pyx_f_6pandas_5_libs_8interval___pyx_unpickle_IntervalTree__set_state(((struct __pyx_obj_6pandas_5_libs_8interval_IntervalTree *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x05502fb = (_is_overlapping, _left_sorter, _right_sorter, closed, dtype, left, right, root))" % __pyx_checksum)
 *     __pyx_result = IntervalTree.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_IntervalTree__set_state(<IntervalTree> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_IntervalTree__set_state(<IntervalTree> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_IntervalTree__set_state(IntervalTree __pyx_result, tuple __pyx_state):
 *     __pyx_result._is_overlapping = __pyx_state[0]; __pyx_result._left_sorter = __pyx_state[1]; __pyx_result._right_sorter = __pyx_state[2]; __pyx_result.closed = __pyx_state[3]; __pyx_result.dtype = __pyx_state[4]; __pyx_result.left = __pyx_state[5]; __pyx_result.right = __pyx_state[6]; __pyx_result.root = __pyx_state[7]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_IntervalTree(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pandas._libs.interval.__pyx_unpickle_IntervalTree", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_IntervalTree__set_state(<IntervalTree> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_IntervalTree__set_state(IntervalTree __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result._is_overlapping = __pyx_state[0]; __pyx_result._left_sorter = __pyx_state[1]; __pyx_result._right_sorter = __pyx_state[2]; __pyx_result.closed = __pyx_state[3]; __pyx_result.dtype = __pyx_state[4]; __pyx_result.left = __pyx_state[5]; __pyx_result.right = __pyx_state[6]; __pyx_result.root = __pyx_state[7]
 *     if len(__pyx_state) > 8 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_6pandas_5_libs_8interval___pyx_unpickle_IntervalTree__set_state(struct __pyx_obj_6pandas_5_libs_8interval_IntervalTree *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  __Pyx_RefNannySetupContext("__pyx_unpickle_IntervalTree__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_IntervalTree__set_state(IntervalTree __pyx_result, tuple __pyx_state):
 *     __pyx_result._is_overlapping = __pyx_state[0]; __pyx_result._left_sorter = __pyx_state[1]; __pyx_result._right_sorter = __pyx_state[2]; __pyx_result.closed = __pyx_state[3]; __pyx_result.dtype = __pyx_state[4]; __pyx_result.left = __pyx_state[5]; __pyx_result.right = __pyx_state[6]; __pyx_result.root = __pyx_state[7]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 8 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[8])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->_is_overlapping);
  __Pyx_DECREF(__pyx_v___pyx_result->_is_overlapping);
  __pyx_v___pyx_result->_is_overlapping = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->_left_sorter);
  __Pyx_DECREF(__pyx_v___pyx_result->_left_sorter);
  __pyx_v___pyx_result->_left_sorter = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->_right_sorter);
  __Pyx_DECREF(__pyx_v___pyx_result->_right_sorter);
  __pyx_v___pyx_result->_right_sorter = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyUnicode_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "unicode", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->closed);
  __Pyx_DECREF(__pyx_v___pyx_result->closed);
  __pyx_v___pyx_result->closed = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 4, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->dtype);
  __Pyx_DECREF(__pyx_v___pyx_result->dtype);
  __pyx_v___pyx_result->dtype = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 5, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->left);
  __Pyx_DECREF(__pyx_v___pyx_result->left);
  __pyx_v___pyx_result->left = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 6, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->right);
  __Pyx_DECREF(__pyx_v___pyx_result->right);
  __pyx_v___pyx_result->right = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 7, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->root);
  __Pyx_DECREF(__pyx_v___pyx_result->root);
  __pyx_v___pyx_result->root = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_IntervalTree__set_state(IntervalTree __pyx_result, tuple __pyx_state):
 *     __pyx_result._is_overlapping = __pyx_state[0]; __pyx_result._left_sorter = __pyx_state[1]; __pyx_result._right_sorter = __pyx_state[2]; __pyx_result.closed = __pyx_state[3]; __pyx_result.dtype = __pyx_state[4]; __pyx_result.left = __pyx_state[5]; __pyx_result.right = __pyx_state[6]; __pyx_result.root = __pyx_state[7]
 *     if len(__pyx_state) > 8 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[8])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(2, 13, __pyx_L1_error)
  }
  __pyx_t_3 = PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_4 = ((__pyx_t_3 > 8) != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_5 = (__pyx_t_4 != 0);
  __pyx_t_2 = __pyx_t_5;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result._is_overlapping = __pyx_state[0]; __pyx_result._left_sorter = __pyx_state[1]; __pyx_result._right_sorter = __pyx_state[2]; __pyx_result.closed = __pyx_state[3]; __pyx_result.dtype = __pyx_state[4]; __pyx_result.left = __pyx_state[5]; __pyx_result.right = __pyx_state[6]; __pyx_result.root = __pyx_state[7]
 *     if len(__pyx_state) > 8 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[8])             # <<<<<<<<<<<<<<
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_update); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(2, 14, __pyx_L1_error)
    }
    __pyx_t_6 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 8, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
      }
    }
    __pyx_t_1 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_8, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_6);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_IntervalTree__set_state(IntervalTree __pyx_result, tuple __pyx_state):
 *     __pyx_result._is_overlapping = __pyx_state[0]; __pyx_result._left_sorter = __pyx_state[1]; __pyx_result._right_sorter = __pyx_state[2]; __pyx_result.closed = __pyx_state[3]; __pyx_result.dtype = __pyx_state[4]; __pyx_result.left = __pyx_state[5]; __pyx_result.right = __pyx_state[6]; __pyx_result.root = __pyx_state[7]
 *     if len(__pyx_state) > 8 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[8])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_IntervalTree__set_state(<IntervalTree> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_IntervalTree__set_state(IntervalTree __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result._is_overlapping = __pyx_state[0]; __pyx_result._left_sorter = __pyx_state[1]; __pyx_result._right_sorter = __pyx_state[2]; __pyx_result.closed = __pyx_state[3]; __pyx_result.dtype = __pyx_state[4]; __pyx_result.left = __pyx_state[5]; __pyx_result.right = __pyx_state[6]; __pyx_result.root = __pyx_state[7]
 *     if len(__pyx_state) > 8 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("pandas._libs.interval.__pyx_unpickle_IntervalTree__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_Float32ClosedLeftIntervalNode(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_7__pyx_unpickle_Float32ClosedLeftIntervalNode(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pandas_5_libs_8interval_7__pyx_unpickle_Float32ClosedLeftIntervalNode = {"__pyx_unpickle_Float32ClosedLeftIntervalNode", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_7__pyx_unpickle_Float32ClosedLeftIntervalNode, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_7__pyx_unpickle_Float32ClosedLeftIntervalNode(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_Float32ClosedLeftIntervalNode (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_type)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_checksum)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Float32ClosedLeftIntervalNode", 1, 3, 3, 1); __PYX_ERR(2, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Float32ClosedLeftIntervalNode", 1, 3, 3, 2); __PYX_ERR(2, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_unpickle_Float32ClosedLeftIntervalNode") < 0)) __PYX_ERR(2, 1, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Float32ClosedLeftIntervalNode", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(2, 1, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.__pyx_unpickle_Float32ClosedLeftIntervalNode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_6__pyx_unpickle_Float32ClosedLeftIntervalNode(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_6__pyx_unpickle_Float32ClosedLeftIntervalNode(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Float32ClosedLeftIntervalNode", 0);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x7a104e5:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)
 */
  __pyx_t_1 = ((__pyx_v___pyx_checksum != 0x7a104e5) != 0);
  if (__pyx_t_1) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x7a104e5:
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)
 *     __pyx_result = Float32ClosedLeftIntervalNode.__new__(__pyx_type)
 */
    __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_PickleError);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_2);
    __pyx_v___pyx_PickleError = __pyx_t_2;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum != 0x7a104e5:
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)             # <<<<<<<<<<<<<<
 *     __pyx_result = Float32ClosedLeftIntervalNode.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_2 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_s_vs_0x7a, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_v___pyx_PickleError);
    __pyx_t_2 = __pyx_v___pyx_PickleError; __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_5, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(2, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x7a104e5:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)
 *     __pyx_result = Float32ClosedLeftIntervalNode.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Float32ClosedLeftIntervalNode__set_state(<Float32ClosedLeftIntervalNode> __pyx_result, __pyx_state)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode), __pyx_n_s_new); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_v___pyx_type) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v___pyx_type);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v___pyx_result = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)
 *     __pyx_result = Float32ClosedLeftIntervalNode.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Float32ClosedLeftIntervalNode__set_state(<Float32ClosedLeftIntervalNode> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_1 = (__pyx_v___pyx_state != Py_None);
  __pyx_t_6 = (__pyx_t_1 != 0);
  if (__pyx_t_6) {

    /* "(tree fragment)":9
 *     __pyx_result = Float32ClosedLeftIntervalNode.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Float32ClosedLeftIntervalNode__set_state(<Float32ClosedLeftIntervalNode> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_Float32ClosedLeftIntervalNode__set_state(Float32ClosedLeftIntervalNode __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(2, 9, __pyx_L1_error)
    __pyx_t_3 = __pyx_f_6pandas_5_libs_8interval___pyx_unpickle_Float32ClosedLeftIntervalNode__set_state(((struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)
 *     __pyx_result = Float32ClosedLeftIntervalNode.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Float32ClosedLeftIntervalNode__set_state(<Float32ClosedLeftIntervalNode> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Float32ClosedLeftIntervalNode__set_state(<Float32ClosedLeftIntervalNode> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_Float32ClosedLeftIntervalNode__set_state(Float32ClosedLeftIntervalNode __pyx_result, tuple __pyx_state):
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_Float32ClosedLeftIntervalNode(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pandas._libs.interval.__pyx_unpickle_Float32ClosedLeftIntervalNode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_Float32ClosedLeftIntervalNode__set_state(<Float32ClosedLeftIntervalNode> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Float32ClosedLeftIntervalNode__set_state(Float32ClosedLeftIntervalNode __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_6pandas_5_libs_8interval___pyx_unpickle_Float32ClosedLeftIntervalNode__set_state(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_memviewslice __pyx_t_2 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_3 = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_t_4;
  __pyx_t_5numpy_int64_t __pyx_t_5;
  __pyx_t_5numpy_float32_t __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Float32ClosedLeftIntervalNode__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_Float32ClosedLeftIntervalNode__set_state(Float32ClosedLeftIntervalNode __pyx_result, tuple __pyx_state):
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[16])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_2.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->center_left_indices, 0);
  __pyx_v___pyx_result->center_left_indices = __pyx_t_2;
  __pyx_t_2.memview = NULL;
  __pyx_t_2.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_float32_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_3.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->center_left_values, 0);
  __pyx_v___pyx_result->center_left_values = __pyx_t_3;
  __pyx_t_3.memview = NULL;
  __pyx_t_3.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_2.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->center_right_indices, 0);
  __pyx_v___pyx_result->center_right_indices = __pyx_t_2;
  __pyx_t_2.memview = NULL;
  __pyx_t_2.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_float32_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_3.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->center_right_values, 0);
  __pyx_v___pyx_result->center_right_values = __pyx_t_3;
  __pyx_t_3.memview = NULL;
  __pyx_t_3.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 4, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_2.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->indices, 0);
  __pyx_v___pyx_result->indices = __pyx_t_2;
  __pyx_t_2.memview = NULL;
  __pyx_t_2.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 5, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->is_leaf_node = __pyx_t_4;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 6, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyInt_As_npy_int64(__pyx_t_1); if (unlikely((__pyx_t_5 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->leaf_size = __pyx_t_5;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 7, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_float32_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_3.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->left, 0);
  __pyx_v___pyx_result->left = __pyx_t_3;
  __pyx_t_3.memview = NULL;
  __pyx_t_3.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 8, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode))))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->left_node);
  __Pyx_DECREF(((PyObject *)__pyx_v___pyx_result->left_node));
  __pyx_v___pyx_result->left_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 9, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = __pyx_PyFloat_AsFloat(__pyx_t_1); if (unlikely((__pyx_t_6 == ((npy_float32)-1)) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->max_right = __pyx_t_6;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 10, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = __pyx_PyFloat_AsFloat(__pyx_t_1); if (unlikely((__pyx_t_6 == ((npy_float32)-1)) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->min_left = __pyx_t_6;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 11, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyInt_As_npy_int64(__pyx_t_1); if (unlikely((__pyx_t_5 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->n_center = __pyx_t_5;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 12, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyInt_As_npy_int64(__pyx_t_1); if (unlikely((__pyx_t_5 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->n_elements = __pyx_t_5;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 13, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = __pyx_PyFloat_AsFloat(__pyx_t_1); if (unlikely((__pyx_t_6 == ((npy_float32)-1)) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->pivot = __pyx_t_6;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 14, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_float32_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_3.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->right, 0);
  __pyx_v___pyx_result->right = __pyx_t_3;
  __pyx_t_3.memview = NULL;
  __pyx_t_3.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 15, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode))))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->right_node);
  __Pyx_DECREF(((PyObject *)__pyx_v___pyx_result->right_node));
  __pyx_v___pyx_result->right_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_Float32ClosedLeftIntervalNode__set_state(Float32ClosedLeftIntervalNode __pyx_result, tuple __pyx_state):
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[16])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(2, 13, __pyx_L1_error)
  }
  __pyx_t_7 = PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_8 = ((__pyx_t_7 > 16) != 0);
  if (__pyx_t_8) {
  } else {
    __pyx_t_4 = __pyx_t_8;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_8 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_8 == ((int)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_9 = (__pyx_t_8 != 0);
  __pyx_t_4 = __pyx_t_9;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_4) {

    /* "(tree fragment)":14
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[16])             # <<<<<<<<<<<<<<
 */
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_10)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_update); if (unlikely(!__pyx_t_11)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(2, 14, __pyx_L1_error)
    }
    __pyx_t_10 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 16, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_10)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_12 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_11))) {
      __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_11);
      if (likely(__pyx_t_12)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
        __Pyx_INCREF(__pyx_t_12);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_11, function);
      }
    }
    __pyx_t_1 = (__pyx_t_12) ? __Pyx_PyObject_Call2Args(__pyx_t_11, __pyx_t_12, __pyx_t_10) : __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_t_10);
    __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_Float32ClosedLeftIntervalNode__set_state(Float32ClosedLeftIntervalNode __pyx_result, tuple __pyx_state):
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[16])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_Float32ClosedLeftIntervalNode__set_state(<Float32ClosedLeftIntervalNode> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Float32ClosedLeftIntervalNode__set_state(Float32ClosedLeftIntervalNode __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_2, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_3, 1);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("pandas._libs.interval.__pyx_unpickle_Float32ClosedLeftIntervalNode__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_Float32ClosedRightIntervalNode(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_9__pyx_unpickle_Float32ClosedRightIntervalNode(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pandas_5_libs_8interval_9__pyx_unpickle_Float32ClosedRightIntervalNode = {"__pyx_unpickle_Float32ClosedRightIntervalNode", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_9__pyx_unpickle_Float32ClosedRightIntervalNode, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_9__pyx_unpickle_Float32ClosedRightIntervalNode(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_Float32ClosedRightIntervalNode (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_type)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_checksum)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Float32ClosedRightIntervalNode", 1, 3, 3, 1); __PYX_ERR(2, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Float32ClosedRightIntervalNode", 1, 3, 3, 2); __PYX_ERR(2, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_unpickle_Float32ClosedRightIntervalNode") < 0)) __PYX_ERR(2, 1, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Float32ClosedRightIntervalNode", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(2, 1, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.__pyx_unpickle_Float32ClosedRightIntervalNode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_8__pyx_unpickle_Float32ClosedRightIntervalNode(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_8__pyx_unpickle_Float32ClosedRightIntervalNode(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Float32ClosedRightIntervalNode", 0);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x7a104e5:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)
 */
  __pyx_t_1 = ((__pyx_v___pyx_checksum != 0x7a104e5) != 0);
  if (__pyx_t_1) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x7a104e5:
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)
 *     __pyx_result = Float32ClosedRightIntervalNode.__new__(__pyx_type)
 */
    __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_PickleError);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_2);
    __pyx_v___pyx_PickleError = __pyx_t_2;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum != 0x7a104e5:
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)             # <<<<<<<<<<<<<<
 *     __pyx_result = Float32ClosedRightIntervalNode.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_2 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_s_vs_0x7a, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_v___pyx_PickleError);
    __pyx_t_2 = __pyx_v___pyx_PickleError; __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_5, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(2, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x7a104e5:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)
 *     __pyx_result = Float32ClosedRightIntervalNode.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Float32ClosedRightIntervalNode__set_state(<Float32ClosedRightIntervalNode> __pyx_result, __pyx_state)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode), __pyx_n_s_new); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_v___pyx_type) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v___pyx_type);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v___pyx_result = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)
 *     __pyx_result = Float32ClosedRightIntervalNode.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Float32ClosedRightIntervalNode__set_state(<Float32ClosedRightIntervalNode> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_1 = (__pyx_v___pyx_state != Py_None);
  __pyx_t_6 = (__pyx_t_1 != 0);
  if (__pyx_t_6) {

    /* "(tree fragment)":9
 *     __pyx_result = Float32ClosedRightIntervalNode.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Float32ClosedRightIntervalNode__set_state(<Float32ClosedRightIntervalNode> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_Float32ClosedRightIntervalNode__set_state(Float32ClosedRightIntervalNode __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(2, 9, __pyx_L1_error)
    __pyx_t_3 = __pyx_f_6pandas_5_libs_8interval___pyx_unpickle_Float32ClosedRightIntervalNode__set_state(((struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)
 *     __pyx_result = Float32ClosedRightIntervalNode.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Float32ClosedRightIntervalNode__set_state(<Float32ClosedRightIntervalNode> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Float32ClosedRightIntervalNode__set_state(<Float32ClosedRightIntervalNode> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_Float32ClosedRightIntervalNode__set_state(Float32ClosedRightIntervalNode __pyx_result, tuple __pyx_state):
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_Float32ClosedRightIntervalNode(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pandas._libs.interval.__pyx_unpickle_Float32ClosedRightIntervalNode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_Float32ClosedRightIntervalNode__set_state(<Float32ClosedRightIntervalNode> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Float32ClosedRightIntervalNode__set_state(Float32ClosedRightIntervalNode __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_6pandas_5_libs_8interval___pyx_unpickle_Float32ClosedRightIntervalNode__set_state(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_memviewslice __pyx_t_2 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_3 = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_t_4;
  __pyx_t_5numpy_int64_t __pyx_t_5;
  __pyx_t_5numpy_float32_t __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Float32ClosedRightIntervalNode__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_Float32ClosedRightIntervalNode__set_state(Float32ClosedRightIntervalNode __pyx_result, tuple __pyx_state):
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[16])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_2.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->center_left_indices, 0);
  __pyx_v___pyx_result->center_left_indices = __pyx_t_2;
  __pyx_t_2.memview = NULL;
  __pyx_t_2.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_float32_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_3.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->center_left_values, 0);
  __pyx_v___pyx_result->center_left_values = __pyx_t_3;
  __pyx_t_3.memview = NULL;
  __pyx_t_3.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_2.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->center_right_indices, 0);
  __pyx_v___pyx_result->center_right_indices = __pyx_t_2;
  __pyx_t_2.memview = NULL;
  __pyx_t_2.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_float32_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_3.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->center_right_values, 0);
  __pyx_v___pyx_result->center_right_values = __pyx_t_3;
  __pyx_t_3.memview = NULL;
  __pyx_t_3.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 4, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_2.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->indices, 0);
  __pyx_v___pyx_result->indices = __pyx_t_2;
  __pyx_t_2.memview = NULL;
  __pyx_t_2.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 5, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->is_leaf_node = __pyx_t_4;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 6, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyInt_As_npy_int64(__pyx_t_1); if (unlikely((__pyx_t_5 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->leaf_size = __pyx_t_5;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 7, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_float32_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_3.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->left, 0);
  __pyx_v___pyx_result->left = __pyx_t_3;
  __pyx_t_3.memview = NULL;
  __pyx_t_3.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 8, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode))))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->left_node);
  __Pyx_DECREF(((PyObject *)__pyx_v___pyx_result->left_node));
  __pyx_v___pyx_result->left_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 9, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = __pyx_PyFloat_AsFloat(__pyx_t_1); if (unlikely((__pyx_t_6 == ((npy_float32)-1)) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->max_right = __pyx_t_6;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 10, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = __pyx_PyFloat_AsFloat(__pyx_t_1); if (unlikely((__pyx_t_6 == ((npy_float32)-1)) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->min_left = __pyx_t_6;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 11, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyInt_As_npy_int64(__pyx_t_1); if (unlikely((__pyx_t_5 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->n_center = __pyx_t_5;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 12, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyInt_As_npy_int64(__pyx_t_1); if (unlikely((__pyx_t_5 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->n_elements = __pyx_t_5;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 13, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = __pyx_PyFloat_AsFloat(__pyx_t_1); if (unlikely((__pyx_t_6 == ((npy_float32)-1)) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->pivot = __pyx_t_6;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 14, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_float32_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_3.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->right, 0);
  __pyx_v___pyx_result->right = __pyx_t_3;
  __pyx_t_3.memview = NULL;
  __pyx_t_3.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 15, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode))))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->right_node);
  __Pyx_DECREF(((PyObject *)__pyx_v___pyx_result->right_node));
  __pyx_v___pyx_result->right_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_Float32ClosedRightIntervalNode__set_state(Float32ClosedRightIntervalNode __pyx_result, tuple __pyx_state):
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[16])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(2, 13, __pyx_L1_error)
  }
  __pyx_t_7 = PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_8 = ((__pyx_t_7 > 16) != 0);
  if (__pyx_t_8) {
  } else {
    __pyx_t_4 = __pyx_t_8;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_8 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_8 == ((int)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_9 = (__pyx_t_8 != 0);
  __pyx_t_4 = __pyx_t_9;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_4) {

    /* "(tree fragment)":14
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[16])             # <<<<<<<<<<<<<<
 */
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_10)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_update); if (unlikely(!__pyx_t_11)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(2, 14, __pyx_L1_error)
    }
    __pyx_t_10 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 16, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_10)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_12 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_11))) {
      __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_11);
      if (likely(__pyx_t_12)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
        __Pyx_INCREF(__pyx_t_12);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_11, function);
      }
    }
    __pyx_t_1 = (__pyx_t_12) ? __Pyx_PyObject_Call2Args(__pyx_t_11, __pyx_t_12, __pyx_t_10) : __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_t_10);
    __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_Float32ClosedRightIntervalNode__set_state(Float32ClosedRightIntervalNode __pyx_result, tuple __pyx_state):
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[16])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_Float32ClosedRightIntervalNode__set_state(<Float32ClosedRightIntervalNode> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Float32ClosedRightIntervalNode__set_state(Float32ClosedRightIntervalNode __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_2, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_3, 1);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("pandas._libs.interval.__pyx_unpickle_Float32ClosedRightIntervalNode__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_Float32ClosedBothIntervalNode(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_11__pyx_unpickle_Float32ClosedBothIntervalNode(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pandas_5_libs_8interval_11__pyx_unpickle_Float32ClosedBothIntervalNode = {"__pyx_unpickle_Float32ClosedBothIntervalNode", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_11__pyx_unpickle_Float32ClosedBothIntervalNode, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_11__pyx_unpickle_Float32ClosedBothIntervalNode(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_Float32ClosedBothIntervalNode (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_type)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_checksum)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Float32ClosedBothIntervalNode", 1, 3, 3, 1); __PYX_ERR(2, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Float32ClosedBothIntervalNode", 1, 3, 3, 2); __PYX_ERR(2, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_unpickle_Float32ClosedBothIntervalNode") < 0)) __PYX_ERR(2, 1, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Float32ClosedBothIntervalNode", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(2, 1, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.__pyx_unpickle_Float32ClosedBothIntervalNode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_10__pyx_unpickle_Float32ClosedBothIntervalNode(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_10__pyx_unpickle_Float32ClosedBothIntervalNode(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Float32ClosedBothIntervalNode", 0);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x7a104e5:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)
 */
  __pyx_t_1 = ((__pyx_v___pyx_checksum != 0x7a104e5) != 0);
  if (__pyx_t_1) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x7a104e5:
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)
 *     __pyx_result = Float32ClosedBothIntervalNode.__new__(__pyx_type)
 */
    __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_PickleError);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_2);
    __pyx_v___pyx_PickleError = __pyx_t_2;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum != 0x7a104e5:
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)             # <<<<<<<<<<<<<<
 *     __pyx_result = Float32ClosedBothIntervalNode.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_2 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_s_vs_0x7a, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_v___pyx_PickleError);
    __pyx_t_2 = __pyx_v___pyx_PickleError; __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_5, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(2, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x7a104e5:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)
 *     __pyx_result = Float32ClosedBothIntervalNode.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Float32ClosedBothIntervalNode__set_state(<Float32ClosedBothIntervalNode> __pyx_result, __pyx_state)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode), __pyx_n_s_new); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_v___pyx_type) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v___pyx_type);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v___pyx_result = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)
 *     __pyx_result = Float32ClosedBothIntervalNode.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Float32ClosedBothIntervalNode__set_state(<Float32ClosedBothIntervalNode> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_1 = (__pyx_v___pyx_state != Py_None);
  __pyx_t_6 = (__pyx_t_1 != 0);
  if (__pyx_t_6) {

    /* "(tree fragment)":9
 *     __pyx_result = Float32ClosedBothIntervalNode.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Float32ClosedBothIntervalNode__set_state(<Float32ClosedBothIntervalNode> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_Float32ClosedBothIntervalNode__set_state(Float32ClosedBothIntervalNode __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(2, 9, __pyx_L1_error)
    __pyx_t_3 = __pyx_f_6pandas_5_libs_8interval___pyx_unpickle_Float32ClosedBothIntervalNode__set_state(((struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)
 *     __pyx_result = Float32ClosedBothIntervalNode.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Float32ClosedBothIntervalNode__set_state(<Float32ClosedBothIntervalNode> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Float32ClosedBothIntervalNode__set_state(<Float32ClosedBothIntervalNode> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_Float32ClosedBothIntervalNode__set_state(Float32ClosedBothIntervalNode __pyx_result, tuple __pyx_state):
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_Float32ClosedBothIntervalNode(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pandas._libs.interval.__pyx_unpickle_Float32ClosedBothIntervalNode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_Float32ClosedBothIntervalNode__set_state(<Float32ClosedBothIntervalNode> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Float32ClosedBothIntervalNode__set_state(Float32ClosedBothIntervalNode __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_6pandas_5_libs_8interval___pyx_unpickle_Float32ClosedBothIntervalNode__set_state(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_memviewslice __pyx_t_2 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_3 = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_t_4;
  __pyx_t_5numpy_int64_t __pyx_t_5;
  __pyx_t_5numpy_float32_t __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Float32ClosedBothIntervalNode__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_Float32ClosedBothIntervalNode__set_state(Float32ClosedBothIntervalNode __pyx_result, tuple __pyx_state):
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[16])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_2.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->center_left_indices, 0);
  __pyx_v___pyx_result->center_left_indices = __pyx_t_2;
  __pyx_t_2.memview = NULL;
  __pyx_t_2.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_float32_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_3.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->center_left_values, 0);
  __pyx_v___pyx_result->center_left_values = __pyx_t_3;
  __pyx_t_3.memview = NULL;
  __pyx_t_3.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_2.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->center_right_indices, 0);
  __pyx_v___pyx_result->center_right_indices = __pyx_t_2;
  __pyx_t_2.memview = NULL;
  __pyx_t_2.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_float32_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_3.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->center_right_values, 0);
  __pyx_v___pyx_result->center_right_values = __pyx_t_3;
  __pyx_t_3.memview = NULL;
  __pyx_t_3.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 4, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_2.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->indices, 0);
  __pyx_v___pyx_result->indices = __pyx_t_2;
  __pyx_t_2.memview = NULL;
  __pyx_t_2.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 5, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->is_leaf_node = __pyx_t_4;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 6, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyInt_As_npy_int64(__pyx_t_1); if (unlikely((__pyx_t_5 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->leaf_size = __pyx_t_5;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 7, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_float32_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_3.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->left, 0);
  __pyx_v___pyx_result->left = __pyx_t_3;
  __pyx_t_3.memview = NULL;
  __pyx_t_3.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 8, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode))))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->left_node);
  __Pyx_DECREF(((PyObject *)__pyx_v___pyx_result->left_node));
  __pyx_v___pyx_result->left_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 9, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = __pyx_PyFloat_AsFloat(__pyx_t_1); if (unlikely((__pyx_t_6 == ((npy_float32)-1)) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->max_right = __pyx_t_6;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 10, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = __pyx_PyFloat_AsFloat(__pyx_t_1); if (unlikely((__pyx_t_6 == ((npy_float32)-1)) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->min_left = __pyx_t_6;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 11, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyInt_As_npy_int64(__pyx_t_1); if (unlikely((__pyx_t_5 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->n_center = __pyx_t_5;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 12, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyInt_As_npy_int64(__pyx_t_1); if (unlikely((__pyx_t_5 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->n_elements = __pyx_t_5;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 13, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = __pyx_PyFloat_AsFloat(__pyx_t_1); if (unlikely((__pyx_t_6 == ((npy_float32)-1)) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->pivot = __pyx_t_6;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 14, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_float32_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_3.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->right, 0);
  __pyx_v___pyx_result->right = __pyx_t_3;
  __pyx_t_3.memview = NULL;
  __pyx_t_3.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 15, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode))))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->right_node);
  __Pyx_DECREF(((PyObject *)__pyx_v___pyx_result->right_node));
  __pyx_v___pyx_result->right_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_Float32ClosedBothIntervalNode__set_state(Float32ClosedBothIntervalNode __pyx_result, tuple __pyx_state):
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[16])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(2, 13, __pyx_L1_error)
  }
  __pyx_t_7 = PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_8 = ((__pyx_t_7 > 16) != 0);
  if (__pyx_t_8) {
  } else {
    __pyx_t_4 = __pyx_t_8;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_8 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_8 == ((int)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_9 = (__pyx_t_8 != 0);
  __pyx_t_4 = __pyx_t_9;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_4) {

    /* "(tree fragment)":14
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[16])             # <<<<<<<<<<<<<<
 */
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_10)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_update); if (unlikely(!__pyx_t_11)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(2, 14, __pyx_L1_error)
    }
    __pyx_t_10 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 16, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_10)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_12 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_11))) {
      __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_11);
      if (likely(__pyx_t_12)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
        __Pyx_INCREF(__pyx_t_12);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_11, function);
      }
    }
    __pyx_t_1 = (__pyx_t_12) ? __Pyx_PyObject_Call2Args(__pyx_t_11, __pyx_t_12, __pyx_t_10) : __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_t_10);
    __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_Float32ClosedBothIntervalNode__set_state(Float32ClosedBothIntervalNode __pyx_result, tuple __pyx_state):
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[16])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_Float32ClosedBothIntervalNode__set_state(<Float32ClosedBothIntervalNode> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Float32ClosedBothIntervalNode__set_state(Float32ClosedBothIntervalNode __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_2, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_3, 1);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("pandas._libs.interval.__pyx_unpickle_Float32ClosedBothIntervalNode__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_Float32ClosedNeitherIntervalNode(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_13__pyx_unpickle_Float32ClosedNeitherIntervalNode(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pandas_5_libs_8interval_13__pyx_unpickle_Float32ClosedNeitherIntervalNode = {"__pyx_unpickle_Float32ClosedNeitherIntervalNode", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_13__pyx_unpickle_Float32ClosedNeitherIntervalNode, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_13__pyx_unpickle_Float32ClosedNeitherIntervalNode(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_Float32ClosedNeitherIntervalNode (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_type)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_checksum)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Float32ClosedNeitherIntervalNode", 1, 3, 3, 1); __PYX_ERR(2, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Float32ClosedNeitherIntervalNode", 1, 3, 3, 2); __PYX_ERR(2, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_unpickle_Float32ClosedNeitherIntervalNode") < 0)) __PYX_ERR(2, 1, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Float32ClosedNeitherIntervalNode", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(2, 1, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.__pyx_unpickle_Float32ClosedNeitherIntervalNode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_12__pyx_unpickle_Float32ClosedNeitherIntervalNode(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_12__pyx_unpickle_Float32ClosedNeitherIntervalNode(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Float32ClosedNeitherIntervalNode", 0);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x7a104e5:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)
 */
  __pyx_t_1 = ((__pyx_v___pyx_checksum != 0x7a104e5) != 0);
  if (__pyx_t_1) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x7a104e5:
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)
 *     __pyx_result = Float32ClosedNeitherIntervalNode.__new__(__pyx_type)
 */
    __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_PickleError);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_2);
    __pyx_v___pyx_PickleError = __pyx_t_2;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum != 0x7a104e5:
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)             # <<<<<<<<<<<<<<
 *     __pyx_result = Float32ClosedNeitherIntervalNode.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_2 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_s_vs_0x7a, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_v___pyx_PickleError);
    __pyx_t_2 = __pyx_v___pyx_PickleError; __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_5, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(2, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x7a104e5:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)
 *     __pyx_result = Float32ClosedNeitherIntervalNode.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Float32ClosedNeitherIntervalNode__set_state(<Float32ClosedNeitherIntervalNode> __pyx_result, __pyx_state)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode), __pyx_n_s_new); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_v___pyx_type) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v___pyx_type);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v___pyx_result = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)
 *     __pyx_result = Float32ClosedNeitherIntervalNode.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Float32ClosedNeitherIntervalNode__set_state(<Float32ClosedNeitherIntervalNode> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_1 = (__pyx_v___pyx_state != Py_None);
  __pyx_t_6 = (__pyx_t_1 != 0);
  if (__pyx_t_6) {

    /* "(tree fragment)":9
 *     __pyx_result = Float32ClosedNeitherIntervalNode.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Float32ClosedNeitherIntervalNode__set_state(<Float32ClosedNeitherIntervalNode> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_Float32ClosedNeitherIntervalNode__set_state(Float32ClosedNeitherIntervalNode __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(2, 9, __pyx_L1_error)
    __pyx_t_3 = __pyx_f_6pandas_5_libs_8interval___pyx_unpickle_Float32ClosedNeitherIntervalNode__set_state(((struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)
 *     __pyx_result = Float32ClosedNeitherIntervalNode.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Float32ClosedNeitherIntervalNode__set_state(<Float32ClosedNeitherIntervalNode> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Float32ClosedNeitherIntervalNode__set_state(<Float32ClosedNeitherIntervalNode> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_Float32ClosedNeitherIntervalNode__set_state(Float32ClosedNeitherIntervalNode __pyx_result, tuple __pyx_state):
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_Float32ClosedNeitherIntervalNode(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pandas._libs.interval.__pyx_unpickle_Float32ClosedNeitherIntervalNode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_Float32ClosedNeitherIntervalNode__set_state(<Float32ClosedNeitherIntervalNode> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Float32ClosedNeitherIntervalNode__set_state(Float32ClosedNeitherIntervalNode __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_6pandas_5_libs_8interval___pyx_unpickle_Float32ClosedNeitherIntervalNode__set_state(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_memviewslice __pyx_t_2 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_3 = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_t_4;
  __pyx_t_5numpy_int64_t __pyx_t_5;
  __pyx_t_5numpy_float32_t __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Float32ClosedNeitherIntervalNode__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_Float32ClosedNeitherIntervalNode__set_state(Float32ClosedNeitherIntervalNode __pyx_result, tuple __pyx_state):
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[16])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_2.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->center_left_indices, 0);
  __pyx_v___pyx_result->center_left_indices = __pyx_t_2;
  __pyx_t_2.memview = NULL;
  __pyx_t_2.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_float32_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_3.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->center_left_values, 0);
  __pyx_v___pyx_result->center_left_values = __pyx_t_3;
  __pyx_t_3.memview = NULL;
  __pyx_t_3.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_2.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->center_right_indices, 0);
  __pyx_v___pyx_result->center_right_indices = __pyx_t_2;
  __pyx_t_2.memview = NULL;
  __pyx_t_2.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_float32_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_3.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->center_right_values, 0);
  __pyx_v___pyx_result->center_right_values = __pyx_t_3;
  __pyx_t_3.memview = NULL;
  __pyx_t_3.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 4, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_2.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->indices, 0);
  __pyx_v___pyx_result->indices = __pyx_t_2;
  __pyx_t_2.memview = NULL;
  __pyx_t_2.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 5, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->is_leaf_node = __pyx_t_4;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 6, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyInt_As_npy_int64(__pyx_t_1); if (unlikely((__pyx_t_5 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->leaf_size = __pyx_t_5;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 7, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_float32_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_3.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->left, 0);
  __pyx_v___pyx_result->left = __pyx_t_3;
  __pyx_t_3.memview = NULL;
  __pyx_t_3.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 8, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode))))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->left_node);
  __Pyx_DECREF(((PyObject *)__pyx_v___pyx_result->left_node));
  __pyx_v___pyx_result->left_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 9, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = __pyx_PyFloat_AsFloat(__pyx_t_1); if (unlikely((__pyx_t_6 == ((npy_float32)-1)) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->max_right = __pyx_t_6;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 10, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = __pyx_PyFloat_AsFloat(__pyx_t_1); if (unlikely((__pyx_t_6 == ((npy_float32)-1)) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->min_left = __pyx_t_6;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 11, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyInt_As_npy_int64(__pyx_t_1); if (unlikely((__pyx_t_5 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->n_center = __pyx_t_5;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 12, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyInt_As_npy_int64(__pyx_t_1); if (unlikely((__pyx_t_5 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->n_elements = __pyx_t_5;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 13, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = __pyx_PyFloat_AsFloat(__pyx_t_1); if (unlikely((__pyx_t_6 == ((npy_float32)-1)) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->pivot = __pyx_t_6;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 14, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_float32_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_3.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->right, 0);
  __pyx_v___pyx_result->right = __pyx_t_3;
  __pyx_t_3.memview = NULL;
  __pyx_t_3.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 15, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode))))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->right_node);
  __Pyx_DECREF(((PyObject *)__pyx_v___pyx_result->right_node));
  __pyx_v___pyx_result->right_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_Float32ClosedNeitherIntervalNode__set_state(Float32ClosedNeitherIntervalNode __pyx_result, tuple __pyx_state):
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[16])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(2, 13, __pyx_L1_error)
  }
  __pyx_t_7 = PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_8 = ((__pyx_t_7 > 16) != 0);
  if (__pyx_t_8) {
  } else {
    __pyx_t_4 = __pyx_t_8;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_8 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_8 == ((int)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_9 = (__pyx_t_8 != 0);
  __pyx_t_4 = __pyx_t_9;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_4) {

    /* "(tree fragment)":14
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[16])             # <<<<<<<<<<<<<<
 */
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_10)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_update); if (unlikely(!__pyx_t_11)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(2, 14, __pyx_L1_error)
    }
    __pyx_t_10 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 16, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_10)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_12 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_11))) {
      __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_11);
      if (likely(__pyx_t_12)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
        __Pyx_INCREF(__pyx_t_12);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_11, function);
      }
    }
    __pyx_t_1 = (__pyx_t_12) ? __Pyx_PyObject_Call2Args(__pyx_t_11, __pyx_t_12, __pyx_t_10) : __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_t_10);
    __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_Float32ClosedNeitherIntervalNode__set_state(Float32ClosedNeitherIntervalNode __pyx_result, tuple __pyx_state):
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[16])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_Float32ClosedNeitherIntervalNode__set_state(<Float32ClosedNeitherIntervalNode> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Float32ClosedNeitherIntervalNode__set_state(Float32ClosedNeitherIntervalNode __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_2, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_3, 1);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("pandas._libs.interval.__pyx_unpickle_Float32ClosedNeitherIntervalNode__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_Float64ClosedLeftIntervalNode(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_15__pyx_unpickle_Float64ClosedLeftIntervalNode(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pandas_5_libs_8interval_15__pyx_unpickle_Float64ClosedLeftIntervalNode = {"__pyx_unpickle_Float64ClosedLeftIntervalNode", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_15__pyx_unpickle_Float64ClosedLeftIntervalNode, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_15__pyx_unpickle_Float64ClosedLeftIntervalNode(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_Float64ClosedLeftIntervalNode (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_type)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_checksum)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Float64ClosedLeftIntervalNode", 1, 3, 3, 1); __PYX_ERR(2, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Float64ClosedLeftIntervalNode", 1, 3, 3, 2); __PYX_ERR(2, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_unpickle_Float64ClosedLeftIntervalNode") < 0)) __PYX_ERR(2, 1, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Float64ClosedLeftIntervalNode", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(2, 1, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.__pyx_unpickle_Float64ClosedLeftIntervalNode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_14__pyx_unpickle_Float64ClosedLeftIntervalNode(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_14__pyx_unpickle_Float64ClosedLeftIntervalNode(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Float64ClosedLeftIntervalNode", 0);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x7a104e5:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)
 */
  __pyx_t_1 = ((__pyx_v___pyx_checksum != 0x7a104e5) != 0);
  if (__pyx_t_1) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x7a104e5:
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)
 *     __pyx_result = Float64ClosedLeftIntervalNode.__new__(__pyx_type)
 */
    __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_PickleError);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_2);
    __pyx_v___pyx_PickleError = __pyx_t_2;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum != 0x7a104e5:
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)             # <<<<<<<<<<<<<<
 *     __pyx_result = Float64ClosedLeftIntervalNode.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_2 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_s_vs_0x7a, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_v___pyx_PickleError);
    __pyx_t_2 = __pyx_v___pyx_PickleError; __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_5, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(2, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x7a104e5:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)
 *     __pyx_result = Float64ClosedLeftIntervalNode.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Float64ClosedLeftIntervalNode__set_state(<Float64ClosedLeftIntervalNode> __pyx_result, __pyx_state)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode), __pyx_n_s_new); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_v___pyx_type) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v___pyx_type);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v___pyx_result = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)
 *     __pyx_result = Float64ClosedLeftIntervalNode.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Float64ClosedLeftIntervalNode__set_state(<Float64ClosedLeftIntervalNode> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_1 = (__pyx_v___pyx_state != Py_None);
  __pyx_t_6 = (__pyx_t_1 != 0);
  if (__pyx_t_6) {

    /* "(tree fragment)":9
 *     __pyx_result = Float64ClosedLeftIntervalNode.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Float64ClosedLeftIntervalNode__set_state(<Float64ClosedLeftIntervalNode> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_Float64ClosedLeftIntervalNode__set_state(Float64ClosedLeftIntervalNode __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(2, 9, __pyx_L1_error)
    __pyx_t_3 = __pyx_f_6pandas_5_libs_8interval___pyx_unpickle_Float64ClosedLeftIntervalNode__set_state(((struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)
 *     __pyx_result = Float64ClosedLeftIntervalNode.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Float64ClosedLeftIntervalNode__set_state(<Float64ClosedLeftIntervalNode> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Float64ClosedLeftIntervalNode__set_state(<Float64ClosedLeftIntervalNode> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_Float64ClosedLeftIntervalNode__set_state(Float64ClosedLeftIntervalNode __pyx_result, tuple __pyx_state):
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_Float64ClosedLeftIntervalNode(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pandas._libs.interval.__pyx_unpickle_Float64ClosedLeftIntervalNode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_Float64ClosedLeftIntervalNode__set_state(<Float64ClosedLeftIntervalNode> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Float64ClosedLeftIntervalNode__set_state(Float64ClosedLeftIntervalNode __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_6pandas_5_libs_8interval___pyx_unpickle_Float64ClosedLeftIntervalNode__set_state(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_memviewslice __pyx_t_2 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_3 = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_t_4;
  __pyx_t_5numpy_int64_t __pyx_t_5;
  __pyx_t_5numpy_float64_t __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Float64ClosedLeftIntervalNode__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_Float64ClosedLeftIntervalNode__set_state(Float64ClosedLeftIntervalNode __pyx_result, tuple __pyx_state):
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[16])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_2.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->center_left_indices, 0);
  __pyx_v___pyx_result->center_left_indices = __pyx_t_2;
  __pyx_t_2.memview = NULL;
  __pyx_t_2.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_float64_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_3.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->center_left_values, 0);
  __pyx_v___pyx_result->center_left_values = __pyx_t_3;
  __pyx_t_3.memview = NULL;
  __pyx_t_3.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_2.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->center_right_indices, 0);
  __pyx_v___pyx_result->center_right_indices = __pyx_t_2;
  __pyx_t_2.memview = NULL;
  __pyx_t_2.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_float64_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_3.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->center_right_values, 0);
  __pyx_v___pyx_result->center_right_values = __pyx_t_3;
  __pyx_t_3.memview = NULL;
  __pyx_t_3.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 4, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_2.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->indices, 0);
  __pyx_v___pyx_result->indices = __pyx_t_2;
  __pyx_t_2.memview = NULL;
  __pyx_t_2.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 5, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->is_leaf_node = __pyx_t_4;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 6, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyInt_As_npy_int64(__pyx_t_1); if (unlikely((__pyx_t_5 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->leaf_size = __pyx_t_5;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 7, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_float64_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_3.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->left, 0);
  __pyx_v___pyx_result->left = __pyx_t_3;
  __pyx_t_3.memview = NULL;
  __pyx_t_3.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 8, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode))))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->left_node);
  __Pyx_DECREF(((PyObject *)__pyx_v___pyx_result->left_node));
  __pyx_v___pyx_result->left_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 9, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_6 == ((npy_float64)-1)) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->max_right = __pyx_t_6;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 10, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_6 == ((npy_float64)-1)) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->min_left = __pyx_t_6;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 11, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyInt_As_npy_int64(__pyx_t_1); if (unlikely((__pyx_t_5 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->n_center = __pyx_t_5;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 12, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyInt_As_npy_int64(__pyx_t_1); if (unlikely((__pyx_t_5 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->n_elements = __pyx_t_5;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 13, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_6 == ((npy_float64)-1)) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->pivot = __pyx_t_6;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 14, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_float64_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_3.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->right, 0);
  __pyx_v___pyx_result->right = __pyx_t_3;
  __pyx_t_3.memview = NULL;
  __pyx_t_3.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 15, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode))))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->right_node);
  __Pyx_DECREF(((PyObject *)__pyx_v___pyx_result->right_node));
  __pyx_v___pyx_result->right_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_Float64ClosedLeftIntervalNode__set_state(Float64ClosedLeftIntervalNode __pyx_result, tuple __pyx_state):
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[16])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(2, 13, __pyx_L1_error)
  }
  __pyx_t_7 = PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_8 = ((__pyx_t_7 > 16) != 0);
  if (__pyx_t_8) {
  } else {
    __pyx_t_4 = __pyx_t_8;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_8 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_8 == ((int)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_9 = (__pyx_t_8 != 0);
  __pyx_t_4 = __pyx_t_9;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_4) {

    /* "(tree fragment)":14
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[16])             # <<<<<<<<<<<<<<
 */
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_10)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_update); if (unlikely(!__pyx_t_11)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(2, 14, __pyx_L1_error)
    }
    __pyx_t_10 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 16, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_10)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_12 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_11))) {
      __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_11);
      if (likely(__pyx_t_12)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
        __Pyx_INCREF(__pyx_t_12);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_11, function);
      }
    }
    __pyx_t_1 = (__pyx_t_12) ? __Pyx_PyObject_Call2Args(__pyx_t_11, __pyx_t_12, __pyx_t_10) : __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_t_10);
    __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_Float64ClosedLeftIntervalNode__set_state(Float64ClosedLeftIntervalNode __pyx_result, tuple __pyx_state):
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[16])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_Float64ClosedLeftIntervalNode__set_state(<Float64ClosedLeftIntervalNode> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Float64ClosedLeftIntervalNode__set_state(Float64ClosedLeftIntervalNode __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_2, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_3, 1);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("pandas._libs.interval.__pyx_unpickle_Float64ClosedLeftIntervalNode__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_Float64ClosedRightIntervalNode(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_17__pyx_unpickle_Float64ClosedRightIntervalNode(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pandas_5_libs_8interval_17__pyx_unpickle_Float64ClosedRightIntervalNode = {"__pyx_unpickle_Float64ClosedRightIntervalNode", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_17__pyx_unpickle_Float64ClosedRightIntervalNode, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_17__pyx_unpickle_Float64ClosedRightIntervalNode(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_Float64ClosedRightIntervalNode (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_type)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_checksum)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Float64ClosedRightIntervalNode", 1, 3, 3, 1); __PYX_ERR(2, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Float64ClosedRightIntervalNode", 1, 3, 3, 2); __PYX_ERR(2, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_unpickle_Float64ClosedRightIntervalNode") < 0)) __PYX_ERR(2, 1, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Float64ClosedRightIntervalNode", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(2, 1, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.__pyx_unpickle_Float64ClosedRightIntervalNode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_16__pyx_unpickle_Float64ClosedRightIntervalNode(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_16__pyx_unpickle_Float64ClosedRightIntervalNode(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Float64ClosedRightIntervalNode", 0);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x7a104e5:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)
 */
  __pyx_t_1 = ((__pyx_v___pyx_checksum != 0x7a104e5) != 0);
  if (__pyx_t_1) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x7a104e5:
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)
 *     __pyx_result = Float64ClosedRightIntervalNode.__new__(__pyx_type)
 */
    __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_PickleError);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_2);
    __pyx_v___pyx_PickleError = __pyx_t_2;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum != 0x7a104e5:
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)             # <<<<<<<<<<<<<<
 *     __pyx_result = Float64ClosedRightIntervalNode.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_2 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_s_vs_0x7a, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_v___pyx_PickleError);
    __pyx_t_2 = __pyx_v___pyx_PickleError; __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_5, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(2, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x7a104e5:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)
 *     __pyx_result = Float64ClosedRightIntervalNode.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Float64ClosedRightIntervalNode__set_state(<Float64ClosedRightIntervalNode> __pyx_result, __pyx_state)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode), __pyx_n_s_new); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_v___pyx_type) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v___pyx_type);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v___pyx_result = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)
 *     __pyx_result = Float64ClosedRightIntervalNode.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Float64ClosedRightIntervalNode__set_state(<Float64ClosedRightIntervalNode> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_1 = (__pyx_v___pyx_state != Py_None);
  __pyx_t_6 = (__pyx_t_1 != 0);
  if (__pyx_t_6) {

    /* "(tree fragment)":9
 *     __pyx_result = Float64ClosedRightIntervalNode.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Float64ClosedRightIntervalNode__set_state(<Float64ClosedRightIntervalNode> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_Float64ClosedRightIntervalNode__set_state(Float64ClosedRightIntervalNode __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(2, 9, __pyx_L1_error)
    __pyx_t_3 = __pyx_f_6pandas_5_libs_8interval___pyx_unpickle_Float64ClosedRightIntervalNode__set_state(((struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)
 *     __pyx_result = Float64ClosedRightIntervalNode.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Float64ClosedRightIntervalNode__set_state(<Float64ClosedRightIntervalNode> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Float64ClosedRightIntervalNode__set_state(<Float64ClosedRightIntervalNode> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_Float64ClosedRightIntervalNode__set_state(Float64ClosedRightIntervalNode __pyx_result, tuple __pyx_state):
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_Float64ClosedRightIntervalNode(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pandas._libs.interval.__pyx_unpickle_Float64ClosedRightIntervalNode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_Float64ClosedRightIntervalNode__set_state(<Float64ClosedRightIntervalNode> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Float64ClosedRightIntervalNode__set_state(Float64ClosedRightIntervalNode __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_6pandas_5_libs_8interval___pyx_unpickle_Float64ClosedRightIntervalNode__set_state(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_memviewslice __pyx_t_2 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_3 = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_t_4;
  __pyx_t_5numpy_int64_t __pyx_t_5;
  __pyx_t_5numpy_float64_t __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Float64ClosedRightIntervalNode__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_Float64ClosedRightIntervalNode__set_state(Float64ClosedRightIntervalNode __pyx_result, tuple __pyx_state):
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[16])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_2.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->center_left_indices, 0);
  __pyx_v___pyx_result->center_left_indices = __pyx_t_2;
  __pyx_t_2.memview = NULL;
  __pyx_t_2.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_float64_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_3.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->center_left_values, 0);
  __pyx_v___pyx_result->center_left_values = __pyx_t_3;
  __pyx_t_3.memview = NULL;
  __pyx_t_3.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_2.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->center_right_indices, 0);
  __pyx_v___pyx_result->center_right_indices = __pyx_t_2;
  __pyx_t_2.memview = NULL;
  __pyx_t_2.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_float64_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_3.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->center_right_values, 0);
  __pyx_v___pyx_result->center_right_values = __pyx_t_3;
  __pyx_t_3.memview = NULL;
  __pyx_t_3.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 4, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_2.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->indices, 0);
  __pyx_v___pyx_result->indices = __pyx_t_2;
  __pyx_t_2.memview = NULL;
  __pyx_t_2.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 5, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->is_leaf_node = __pyx_t_4;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 6, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyInt_As_npy_int64(__pyx_t_1); if (unlikely((__pyx_t_5 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->leaf_size = __pyx_t_5;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 7, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_float64_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_3.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->left, 0);
  __pyx_v___pyx_result->left = __pyx_t_3;
  __pyx_t_3.memview = NULL;
  __pyx_t_3.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 8, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode))))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->left_node);
  __Pyx_DECREF(((PyObject *)__pyx_v___pyx_result->left_node));
  __pyx_v___pyx_result->left_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 9, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_6 == ((npy_float64)-1)) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->max_right = __pyx_t_6;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 10, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_6 == ((npy_float64)-1)) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->min_left = __pyx_t_6;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 11, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyInt_As_npy_int64(__pyx_t_1); if (unlikely((__pyx_t_5 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->n_center = __pyx_t_5;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 12, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyInt_As_npy_int64(__pyx_t_1); if (unlikely((__pyx_t_5 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->n_elements = __pyx_t_5;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 13, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_6 == ((npy_float64)-1)) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->pivot = __pyx_t_6;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 14, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_float64_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_3.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->right, 0);
  __pyx_v___pyx_result->right = __pyx_t_3;
  __pyx_t_3.memview = NULL;
  __pyx_t_3.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 15, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode))))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->right_node);
  __Pyx_DECREF(((PyObject *)__pyx_v___pyx_result->right_node));
  __pyx_v___pyx_result->right_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_Float64ClosedRightIntervalNode__set_state(Float64ClosedRightIntervalNode __pyx_result, tuple __pyx_state):
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[16])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(2, 13, __pyx_L1_error)
  }
  __pyx_t_7 = PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_8 = ((__pyx_t_7 > 16) != 0);
  if (__pyx_t_8) {
  } else {
    __pyx_t_4 = __pyx_t_8;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_8 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_8 == ((int)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_9 = (__pyx_t_8 != 0);
  __pyx_t_4 = __pyx_t_9;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_4) {

    /* "(tree fragment)":14
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[16])             # <<<<<<<<<<<<<<
 */
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_10)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_update); if (unlikely(!__pyx_t_11)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(2, 14, __pyx_L1_error)
    }
    __pyx_t_10 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 16, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_10)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_12 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_11))) {
      __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_11);
      if (likely(__pyx_t_12)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
        __Pyx_INCREF(__pyx_t_12);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_11, function);
      }
    }
    __pyx_t_1 = (__pyx_t_12) ? __Pyx_PyObject_Call2Args(__pyx_t_11, __pyx_t_12, __pyx_t_10) : __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_t_10);
    __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_Float64ClosedRightIntervalNode__set_state(Float64ClosedRightIntervalNode __pyx_result, tuple __pyx_state):
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[16])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_Float64ClosedRightIntervalNode__set_state(<Float64ClosedRightIntervalNode> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Float64ClosedRightIntervalNode__set_state(Float64ClosedRightIntervalNode __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_2, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_3, 1);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("pandas._libs.interval.__pyx_unpickle_Float64ClosedRightIntervalNode__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_Float64ClosedBothIntervalNode(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_19__pyx_unpickle_Float64ClosedBothIntervalNode(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pandas_5_libs_8interval_19__pyx_unpickle_Float64ClosedBothIntervalNode = {"__pyx_unpickle_Float64ClosedBothIntervalNode", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_19__pyx_unpickle_Float64ClosedBothIntervalNode, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_19__pyx_unpickle_Float64ClosedBothIntervalNode(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_Float64ClosedBothIntervalNode (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_type)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_checksum)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Float64ClosedBothIntervalNode", 1, 3, 3, 1); __PYX_ERR(2, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Float64ClosedBothIntervalNode", 1, 3, 3, 2); __PYX_ERR(2, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_unpickle_Float64ClosedBothIntervalNode") < 0)) __PYX_ERR(2, 1, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Float64ClosedBothIntervalNode", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(2, 1, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.__pyx_unpickle_Float64ClosedBothIntervalNode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_18__pyx_unpickle_Float64ClosedBothIntervalNode(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_18__pyx_unpickle_Float64ClosedBothIntervalNode(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Float64ClosedBothIntervalNode", 0);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x7a104e5:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)
 */
  __pyx_t_1 = ((__pyx_v___pyx_checksum != 0x7a104e5) != 0);
  if (__pyx_t_1) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x7a104e5:
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)
 *     __pyx_result = Float64ClosedBothIntervalNode.__new__(__pyx_type)
 */
    __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_PickleError);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_2);
    __pyx_v___pyx_PickleError = __pyx_t_2;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum != 0x7a104e5:
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)             # <<<<<<<<<<<<<<
 *     __pyx_result = Float64ClosedBothIntervalNode.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_2 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_s_vs_0x7a, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_v___pyx_PickleError);
    __pyx_t_2 = __pyx_v___pyx_PickleError; __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_5, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(2, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x7a104e5:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)
 *     __pyx_result = Float64ClosedBothIntervalNode.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Float64ClosedBothIntervalNode__set_state(<Float64ClosedBothIntervalNode> __pyx_result, __pyx_state)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode), __pyx_n_s_new); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_v___pyx_type) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v___pyx_type);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v___pyx_result = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)
 *     __pyx_result = Float64ClosedBothIntervalNode.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Float64ClosedBothIntervalNode__set_state(<Float64ClosedBothIntervalNode> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_1 = (__pyx_v___pyx_state != Py_None);
  __pyx_t_6 = (__pyx_t_1 != 0);
  if (__pyx_t_6) {

    /* "(tree fragment)":9
 *     __pyx_result = Float64ClosedBothIntervalNode.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Float64ClosedBothIntervalNode__set_state(<Float64ClosedBothIntervalNode> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_Float64ClosedBothIntervalNode__set_state(Float64ClosedBothIntervalNode __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(2, 9, __pyx_L1_error)
    __pyx_t_3 = __pyx_f_6pandas_5_libs_8interval___pyx_unpickle_Float64ClosedBothIntervalNode__set_state(((struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)
 *     __pyx_result = Float64ClosedBothIntervalNode.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Float64ClosedBothIntervalNode__set_state(<Float64ClosedBothIntervalNode> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Float64ClosedBothIntervalNode__set_state(<Float64ClosedBothIntervalNode> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_Float64ClosedBothIntervalNode__set_state(Float64ClosedBothIntervalNode __pyx_result, tuple __pyx_state):
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_Float64ClosedBothIntervalNode(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pandas._libs.interval.__pyx_unpickle_Float64ClosedBothIntervalNode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_Float64ClosedBothIntervalNode__set_state(<Float64ClosedBothIntervalNode> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Float64ClosedBothIntervalNode__set_state(Float64ClosedBothIntervalNode __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_6pandas_5_libs_8interval___pyx_unpickle_Float64ClosedBothIntervalNode__set_state(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_memviewslice __pyx_t_2 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_3 = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_t_4;
  __pyx_t_5numpy_int64_t __pyx_t_5;
  __pyx_t_5numpy_float64_t __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Float64ClosedBothIntervalNode__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_Float64ClosedBothIntervalNode__set_state(Float64ClosedBothIntervalNode __pyx_result, tuple __pyx_state):
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[16])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_2.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->center_left_indices, 0);
  __pyx_v___pyx_result->center_left_indices = __pyx_t_2;
  __pyx_t_2.memview = NULL;
  __pyx_t_2.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_float64_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_3.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->center_left_values, 0);
  __pyx_v___pyx_result->center_left_values = __pyx_t_3;
  __pyx_t_3.memview = NULL;
  __pyx_t_3.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_2.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->center_right_indices, 0);
  __pyx_v___pyx_result->center_right_indices = __pyx_t_2;
  __pyx_t_2.memview = NULL;
  __pyx_t_2.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_float64_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_3.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->center_right_values, 0);
  __pyx_v___pyx_result->center_right_values = __pyx_t_3;
  __pyx_t_3.memview = NULL;
  __pyx_t_3.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 4, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_2.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->indices, 0);
  __pyx_v___pyx_result->indices = __pyx_t_2;
  __pyx_t_2.memview = NULL;
  __pyx_t_2.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 5, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->is_leaf_node = __pyx_t_4;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 6, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyInt_As_npy_int64(__pyx_t_1); if (unlikely((__pyx_t_5 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->leaf_size = __pyx_t_5;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 7, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_float64_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_3.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->left, 0);
  __pyx_v___pyx_result->left = __pyx_t_3;
  __pyx_t_3.memview = NULL;
  __pyx_t_3.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 8, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode))))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->left_node);
  __Pyx_DECREF(((PyObject *)__pyx_v___pyx_result->left_node));
  __pyx_v___pyx_result->left_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 9, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_6 == ((npy_float64)-1)) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->max_right = __pyx_t_6;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 10, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_6 == ((npy_float64)-1)) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->min_left = __pyx_t_6;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 11, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyInt_As_npy_int64(__pyx_t_1); if (unlikely((__pyx_t_5 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->n_center = __pyx_t_5;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 12, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyInt_As_npy_int64(__pyx_t_1); if (unlikely((__pyx_t_5 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->n_elements = __pyx_t_5;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 13, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_6 == ((npy_float64)-1)) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->pivot = __pyx_t_6;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 14, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_float64_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_3.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->right, 0);
  __pyx_v___pyx_result->right = __pyx_t_3;
  __pyx_t_3.memview = NULL;
  __pyx_t_3.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 15, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode))))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->right_node);
  __Pyx_DECREF(((PyObject *)__pyx_v___pyx_result->right_node));
  __pyx_v___pyx_result->right_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_Float64ClosedBothIntervalNode__set_state(Float64ClosedBothIntervalNode __pyx_result, tuple __pyx_state):
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[16])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(2, 13, __pyx_L1_error)
  }
  __pyx_t_7 = PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_8 = ((__pyx_t_7 > 16) != 0);
  if (__pyx_t_8) {
  } else {
    __pyx_t_4 = __pyx_t_8;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_8 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_8 == ((int)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_9 = (__pyx_t_8 != 0);
  __pyx_t_4 = __pyx_t_9;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_4) {

    /* "(tree fragment)":14
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[16])             # <<<<<<<<<<<<<<
 */
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_10)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_update); if (unlikely(!__pyx_t_11)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(2, 14, __pyx_L1_error)
    }
    __pyx_t_10 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 16, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_10)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_12 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_11))) {
      __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_11);
      if (likely(__pyx_t_12)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
        __Pyx_INCREF(__pyx_t_12);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_11, function);
      }
    }
    __pyx_t_1 = (__pyx_t_12) ? __Pyx_PyObject_Call2Args(__pyx_t_11, __pyx_t_12, __pyx_t_10) : __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_t_10);
    __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_Float64ClosedBothIntervalNode__set_state(Float64ClosedBothIntervalNode __pyx_result, tuple __pyx_state):
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[16])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_Float64ClosedBothIntervalNode__set_state(<Float64ClosedBothIntervalNode> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Float64ClosedBothIntervalNode__set_state(Float64ClosedBothIntervalNode __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_2, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_3, 1);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("pandas._libs.interval.__pyx_unpickle_Float64ClosedBothIntervalNode__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_Float64ClosedNeitherIntervalNode(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_21__pyx_unpickle_Float64ClosedNeitherIntervalNode(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pandas_5_libs_8interval_21__pyx_unpickle_Float64ClosedNeitherIntervalNode = {"__pyx_unpickle_Float64ClosedNeitherIntervalNode", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_21__pyx_unpickle_Float64ClosedNeitherIntervalNode, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_21__pyx_unpickle_Float64ClosedNeitherIntervalNode(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_Float64ClosedNeitherIntervalNode (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_type)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_checksum)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Float64ClosedNeitherIntervalNode", 1, 3, 3, 1); __PYX_ERR(2, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Float64ClosedNeitherIntervalNode", 1, 3, 3, 2); __PYX_ERR(2, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_unpickle_Float64ClosedNeitherIntervalNode") < 0)) __PYX_ERR(2, 1, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Float64ClosedNeitherIntervalNode", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(2, 1, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.__pyx_unpickle_Float64ClosedNeitherIntervalNode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_20__pyx_unpickle_Float64ClosedNeitherIntervalNode(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_20__pyx_unpickle_Float64ClosedNeitherIntervalNode(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Float64ClosedNeitherIntervalNode", 0);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x7a104e5:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)
 */
  __pyx_t_1 = ((__pyx_v___pyx_checksum != 0x7a104e5) != 0);
  if (__pyx_t_1) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x7a104e5:
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)
 *     __pyx_result = Float64ClosedNeitherIntervalNode.__new__(__pyx_type)
 */
    __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_PickleError);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_2);
    __pyx_v___pyx_PickleError = __pyx_t_2;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum != 0x7a104e5:
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)             # <<<<<<<<<<<<<<
 *     __pyx_result = Float64ClosedNeitherIntervalNode.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_2 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_s_vs_0x7a, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_v___pyx_PickleError);
    __pyx_t_2 = __pyx_v___pyx_PickleError; __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_5, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(2, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x7a104e5:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)
 *     __pyx_result = Float64ClosedNeitherIntervalNode.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Float64ClosedNeitherIntervalNode__set_state(<Float64ClosedNeitherIntervalNode> __pyx_result, __pyx_state)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode), __pyx_n_s_new); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_v___pyx_type) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v___pyx_type);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v___pyx_result = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)
 *     __pyx_result = Float64ClosedNeitherIntervalNode.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Float64ClosedNeitherIntervalNode__set_state(<Float64ClosedNeitherIntervalNode> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_1 = (__pyx_v___pyx_state != Py_None);
  __pyx_t_6 = (__pyx_t_1 != 0);
  if (__pyx_t_6) {

    /* "(tree fragment)":9
 *     __pyx_result = Float64ClosedNeitherIntervalNode.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Float64ClosedNeitherIntervalNode__set_state(<Float64ClosedNeitherIntervalNode> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_Float64ClosedNeitherIntervalNode__set_state(Float64ClosedNeitherIntervalNode __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(2, 9, __pyx_L1_error)
    __pyx_t_3 = __pyx_f_6pandas_5_libs_8interval___pyx_unpickle_Float64ClosedNeitherIntervalNode__set_state(((struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)
 *     __pyx_result = Float64ClosedNeitherIntervalNode.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Float64ClosedNeitherIntervalNode__set_state(<Float64ClosedNeitherIntervalNode> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Float64ClosedNeitherIntervalNode__set_state(<Float64ClosedNeitherIntervalNode> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_Float64ClosedNeitherIntervalNode__set_state(Float64ClosedNeitherIntervalNode __pyx_result, tuple __pyx_state):
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_Float64ClosedNeitherIntervalNode(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pandas._libs.interval.__pyx_unpickle_Float64ClosedNeitherIntervalNode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_Float64ClosedNeitherIntervalNode__set_state(<Float64ClosedNeitherIntervalNode> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Float64ClosedNeitherIntervalNode__set_state(Float64ClosedNeitherIntervalNode __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_6pandas_5_libs_8interval___pyx_unpickle_Float64ClosedNeitherIntervalNode__set_state(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_memviewslice __pyx_t_2 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_3 = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_t_4;
  __pyx_t_5numpy_int64_t __pyx_t_5;
  __pyx_t_5numpy_float64_t __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Float64ClosedNeitherIntervalNode__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_Float64ClosedNeitherIntervalNode__set_state(Float64ClosedNeitherIntervalNode __pyx_result, tuple __pyx_state):
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[16])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_2.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->center_left_indices, 0);
  __pyx_v___pyx_result->center_left_indices = __pyx_t_2;
  __pyx_t_2.memview = NULL;
  __pyx_t_2.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_float64_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_3.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->center_left_values, 0);
  __pyx_v___pyx_result->center_left_values = __pyx_t_3;
  __pyx_t_3.memview = NULL;
  __pyx_t_3.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_2.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->center_right_indices, 0);
  __pyx_v___pyx_result->center_right_indices = __pyx_t_2;
  __pyx_t_2.memview = NULL;
  __pyx_t_2.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_float64_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_3.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->center_right_values, 0);
  __pyx_v___pyx_result->center_right_values = __pyx_t_3;
  __pyx_t_3.memview = NULL;
  __pyx_t_3.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 4, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_2.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->indices, 0);
  __pyx_v___pyx_result->indices = __pyx_t_2;
  __pyx_t_2.memview = NULL;
  __pyx_t_2.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 5, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->is_leaf_node = __pyx_t_4;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 6, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyInt_As_npy_int64(__pyx_t_1); if (unlikely((__pyx_t_5 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->leaf_size = __pyx_t_5;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 7, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_float64_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_3.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->left, 0);
  __pyx_v___pyx_result->left = __pyx_t_3;
  __pyx_t_3.memview = NULL;
  __pyx_t_3.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 8, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode))))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->left_node);
  __Pyx_DECREF(((PyObject *)__pyx_v___pyx_result->left_node));
  __pyx_v___pyx_result->left_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 9, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_6 == ((npy_float64)-1)) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->max_right = __pyx_t_6;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 10, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_6 == ((npy_float64)-1)) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->min_left = __pyx_t_6;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 11, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyInt_As_npy_int64(__pyx_t_1); if (unlikely((__pyx_t_5 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->n_center = __pyx_t_5;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 12, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyInt_As_npy_int64(__pyx_t_1); if (unlikely((__pyx_t_5 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->n_elements = __pyx_t_5;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 13, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_6 == ((npy_float64)-1)) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->pivot = __pyx_t_6;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 14, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_float64_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_3.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->right, 0);
  __pyx_v___pyx_result->right = __pyx_t_3;
  __pyx_t_3.memview = NULL;
  __pyx_t_3.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 15, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode))))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->right_node);
  __Pyx_DECREF(((PyObject *)__pyx_v___pyx_result->right_node));
  __pyx_v___pyx_result->right_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_Float64ClosedNeitherIntervalNode__set_state(Float64ClosedNeitherIntervalNode __pyx_result, tuple __pyx_state):
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[16])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(2, 13, __pyx_L1_error)
  }
  __pyx_t_7 = PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_8 = ((__pyx_t_7 > 16) != 0);
  if (__pyx_t_8) {
  } else {
    __pyx_t_4 = __pyx_t_8;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_8 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_8 == ((int)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_9 = (__pyx_t_8 != 0);
  __pyx_t_4 = __pyx_t_9;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_4) {

    /* "(tree fragment)":14
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[16])             # <<<<<<<<<<<<<<
 */
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_10)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_update); if (unlikely(!__pyx_t_11)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(2, 14, __pyx_L1_error)
    }
    __pyx_t_10 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 16, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_10)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_12 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_11))) {
      __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_11);
      if (likely(__pyx_t_12)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
        __Pyx_INCREF(__pyx_t_12);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_11, function);
      }
    }
    __pyx_t_1 = (__pyx_t_12) ? __Pyx_PyObject_Call2Args(__pyx_t_11, __pyx_t_12, __pyx_t_10) : __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_t_10);
    __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_Float64ClosedNeitherIntervalNode__set_state(Float64ClosedNeitherIntervalNode __pyx_result, tuple __pyx_state):
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[16])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_Float64ClosedNeitherIntervalNode__set_state(<Float64ClosedNeitherIntervalNode> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Float64ClosedNeitherIntervalNode__set_state(Float64ClosedNeitherIntervalNode __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_2, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_3, 1);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("pandas._libs.interval.__pyx_unpickle_Float64ClosedNeitherIntervalNode__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_Int32ClosedLeftIntervalNode(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_23__pyx_unpickle_Int32ClosedLeftIntervalNode(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pandas_5_libs_8interval_23__pyx_unpickle_Int32ClosedLeftIntervalNode = {"__pyx_unpickle_Int32ClosedLeftIntervalNode", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_23__pyx_unpickle_Int32ClosedLeftIntervalNode, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_23__pyx_unpickle_Int32ClosedLeftIntervalNode(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_Int32ClosedLeftIntervalNode (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_type)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_checksum)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Int32ClosedLeftIntervalNode", 1, 3, 3, 1); __PYX_ERR(2, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Int32ClosedLeftIntervalNode", 1, 3, 3, 2); __PYX_ERR(2, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_unpickle_Int32ClosedLeftIntervalNode") < 0)) __PYX_ERR(2, 1, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Int32ClosedLeftIntervalNode", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(2, 1, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.__pyx_unpickle_Int32ClosedLeftIntervalNode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_22__pyx_unpickle_Int32ClosedLeftIntervalNode(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_22__pyx_unpickle_Int32ClosedLeftIntervalNode(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Int32ClosedLeftIntervalNode", 0);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x7a104e5:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)
 */
  __pyx_t_1 = ((__pyx_v___pyx_checksum != 0x7a104e5) != 0);
  if (__pyx_t_1) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x7a104e5:
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)
 *     __pyx_result = Int32ClosedLeftIntervalNode.__new__(__pyx_type)
 */
    __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_PickleError);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_2);
    __pyx_v___pyx_PickleError = __pyx_t_2;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum != 0x7a104e5:
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)             # <<<<<<<<<<<<<<
 *     __pyx_result = Int32ClosedLeftIntervalNode.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_2 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_s_vs_0x7a, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_v___pyx_PickleError);
    __pyx_t_2 = __pyx_v___pyx_PickleError; __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_5, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(2, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x7a104e5:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)
 *     __pyx_result = Int32ClosedLeftIntervalNode.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Int32ClosedLeftIntervalNode__set_state(<Int32ClosedLeftIntervalNode> __pyx_result, __pyx_state)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode), __pyx_n_s_new); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_v___pyx_type) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v___pyx_type);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v___pyx_result = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)
 *     __pyx_result = Int32ClosedLeftIntervalNode.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Int32ClosedLeftIntervalNode__set_state(<Int32ClosedLeftIntervalNode> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_1 = (__pyx_v___pyx_state != Py_None);
  __pyx_t_6 = (__pyx_t_1 != 0);
  if (__pyx_t_6) {

    /* "(tree fragment)":9
 *     __pyx_result = Int32ClosedLeftIntervalNode.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Int32ClosedLeftIntervalNode__set_state(<Int32ClosedLeftIntervalNode> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_Int32ClosedLeftIntervalNode__set_state(Int32ClosedLeftIntervalNode __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(2, 9, __pyx_L1_error)
    __pyx_t_3 = __pyx_f_6pandas_5_libs_8interval___pyx_unpickle_Int32ClosedLeftIntervalNode__set_state(((struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)
 *     __pyx_result = Int32ClosedLeftIntervalNode.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Int32ClosedLeftIntervalNode__set_state(<Int32ClosedLeftIntervalNode> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Int32ClosedLeftIntervalNode__set_state(<Int32ClosedLeftIntervalNode> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_Int32ClosedLeftIntervalNode__set_state(Int32ClosedLeftIntervalNode __pyx_result, tuple __pyx_state):
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_Int32ClosedLeftIntervalNode(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pandas._libs.interval.__pyx_unpickle_Int32ClosedLeftIntervalNode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_Int32ClosedLeftIntervalNode__set_state(<Int32ClosedLeftIntervalNode> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Int32ClosedLeftIntervalNode__set_state(Int32ClosedLeftIntervalNode __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_6pandas_5_libs_8interval___pyx_unpickle_Int32ClosedLeftIntervalNode__set_state(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_memviewslice __pyx_t_2 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_3 = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_t_4;
  __pyx_t_5numpy_int64_t __pyx_t_5;
  __pyx_t_5numpy_int32_t __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Int32ClosedLeftIntervalNode__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_Int32ClosedLeftIntervalNode__set_state(Int32ClosedLeftIntervalNode __pyx_result, tuple __pyx_state):
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[16])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_2.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->center_left_indices, 0);
  __pyx_v___pyx_result->center_left_indices = __pyx_t_2;
  __pyx_t_2.memview = NULL;
  __pyx_t_2.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int32_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_3.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->center_left_values, 0);
  __pyx_v___pyx_result->center_left_values = __pyx_t_3;
  __pyx_t_3.memview = NULL;
  __pyx_t_3.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_2.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->center_right_indices, 0);
  __pyx_v___pyx_result->center_right_indices = __pyx_t_2;
  __pyx_t_2.memview = NULL;
  __pyx_t_2.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int32_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_3.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->center_right_values, 0);
  __pyx_v___pyx_result->center_right_values = __pyx_t_3;
  __pyx_t_3.memview = NULL;
  __pyx_t_3.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 4, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_2.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->indices, 0);
  __pyx_v___pyx_result->indices = __pyx_t_2;
  __pyx_t_2.memview = NULL;
  __pyx_t_2.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 5, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->is_leaf_node = __pyx_t_4;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 6, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyInt_As_npy_int64(__pyx_t_1); if (unlikely((__pyx_t_5 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->leaf_size = __pyx_t_5;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 7, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int32_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_3.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->left, 0);
  __pyx_v___pyx_result->left = __pyx_t_3;
  __pyx_t_3.memview = NULL;
  __pyx_t_3.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 8, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode))))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->left_node);
  __Pyx_DECREF(((PyObject *)__pyx_v___pyx_result->left_node));
  __pyx_v___pyx_result->left_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 9, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = __Pyx_PyInt_As_npy_int32(__pyx_t_1); if (unlikely((__pyx_t_6 == ((npy_int32)-1)) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->max_right = __pyx_t_6;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 10, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = __Pyx_PyInt_As_npy_int32(__pyx_t_1); if (unlikely((__pyx_t_6 == ((npy_int32)-1)) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->min_left = __pyx_t_6;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 11, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyInt_As_npy_int64(__pyx_t_1); if (unlikely((__pyx_t_5 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->n_center = __pyx_t_5;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 12, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyInt_As_npy_int64(__pyx_t_1); if (unlikely((__pyx_t_5 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->n_elements = __pyx_t_5;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 13, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = __Pyx_PyInt_As_npy_int32(__pyx_t_1); if (unlikely((__pyx_t_6 == ((npy_int32)-1)) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->pivot = __pyx_t_6;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 14, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int32_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_3.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->right, 0);
  __pyx_v___pyx_result->right = __pyx_t_3;
  __pyx_t_3.memview = NULL;
  __pyx_t_3.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 15, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode))))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->right_node);
  __Pyx_DECREF(((PyObject *)__pyx_v___pyx_result->right_node));
  __pyx_v___pyx_result->right_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_Int32ClosedLeftIntervalNode__set_state(Int32ClosedLeftIntervalNode __pyx_result, tuple __pyx_state):
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[16])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(2, 13, __pyx_L1_error)
  }
  __pyx_t_7 = PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_8 = ((__pyx_t_7 > 16) != 0);
  if (__pyx_t_8) {
  } else {
    __pyx_t_4 = __pyx_t_8;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_8 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_8 == ((int)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_9 = (__pyx_t_8 != 0);
  __pyx_t_4 = __pyx_t_9;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_4) {

    /* "(tree fragment)":14
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[16])             # <<<<<<<<<<<<<<
 */
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_10)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_update); if (unlikely(!__pyx_t_11)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(2, 14, __pyx_L1_error)
    }
    __pyx_t_10 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 16, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_10)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_12 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_11))) {
      __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_11);
      if (likely(__pyx_t_12)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
        __Pyx_INCREF(__pyx_t_12);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_11, function);
      }
    }
    __pyx_t_1 = (__pyx_t_12) ? __Pyx_PyObject_Call2Args(__pyx_t_11, __pyx_t_12, __pyx_t_10) : __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_t_10);
    __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_Int32ClosedLeftIntervalNode__set_state(Int32ClosedLeftIntervalNode __pyx_result, tuple __pyx_state):
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[16])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_Int32ClosedLeftIntervalNode__set_state(<Int32ClosedLeftIntervalNode> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Int32ClosedLeftIntervalNode__set_state(Int32ClosedLeftIntervalNode __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_2, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_3, 1);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("pandas._libs.interval.__pyx_unpickle_Int32ClosedLeftIntervalNode__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_Int32ClosedRightIntervalNode(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_25__pyx_unpickle_Int32ClosedRightIntervalNode(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pandas_5_libs_8interval_25__pyx_unpickle_Int32ClosedRightIntervalNode = {"__pyx_unpickle_Int32ClosedRightIntervalNode", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_25__pyx_unpickle_Int32ClosedRightIntervalNode, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_25__pyx_unpickle_Int32ClosedRightIntervalNode(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_Int32ClosedRightIntervalNode (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_type)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_checksum)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Int32ClosedRightIntervalNode", 1, 3, 3, 1); __PYX_ERR(2, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Int32ClosedRightIntervalNode", 1, 3, 3, 2); __PYX_ERR(2, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_unpickle_Int32ClosedRightIntervalNode") < 0)) __PYX_ERR(2, 1, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Int32ClosedRightIntervalNode", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(2, 1, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.__pyx_unpickle_Int32ClosedRightIntervalNode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_24__pyx_unpickle_Int32ClosedRightIntervalNode(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_24__pyx_unpickle_Int32ClosedRightIntervalNode(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Int32ClosedRightIntervalNode", 0);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x7a104e5:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)
 */
  __pyx_t_1 = ((__pyx_v___pyx_checksum != 0x7a104e5) != 0);
  if (__pyx_t_1) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x7a104e5:
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)
 *     __pyx_result = Int32ClosedRightIntervalNode.__new__(__pyx_type)
 */
    __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_PickleError);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_2);
    __pyx_v___pyx_PickleError = __pyx_t_2;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum != 0x7a104e5:
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)             # <<<<<<<<<<<<<<
 *     __pyx_result = Int32ClosedRightIntervalNode.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_2 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_s_vs_0x7a, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_v___pyx_PickleError);
    __pyx_t_2 = __pyx_v___pyx_PickleError; __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_5, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(2, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x7a104e5:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)
 *     __pyx_result = Int32ClosedRightIntervalNode.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Int32ClosedRightIntervalNode__set_state(<Int32ClosedRightIntervalNode> __pyx_result, __pyx_state)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode), __pyx_n_s_new); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_v___pyx_type) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v___pyx_type);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v___pyx_result = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)
 *     __pyx_result = Int32ClosedRightIntervalNode.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Int32ClosedRightIntervalNode__set_state(<Int32ClosedRightIntervalNode> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_1 = (__pyx_v___pyx_state != Py_None);
  __pyx_t_6 = (__pyx_t_1 != 0);
  if (__pyx_t_6) {

    /* "(tree fragment)":9
 *     __pyx_result = Int32ClosedRightIntervalNode.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Int32ClosedRightIntervalNode__set_state(<Int32ClosedRightIntervalNode> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_Int32ClosedRightIntervalNode__set_state(Int32ClosedRightIntervalNode __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(2, 9, __pyx_L1_error)
    __pyx_t_3 = __pyx_f_6pandas_5_libs_8interval___pyx_unpickle_Int32ClosedRightIntervalNode__set_state(((struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)
 *     __pyx_result = Int32ClosedRightIntervalNode.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Int32ClosedRightIntervalNode__set_state(<Int32ClosedRightIntervalNode> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Int32ClosedRightIntervalNode__set_state(<Int32ClosedRightIntervalNode> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_Int32ClosedRightIntervalNode__set_state(Int32ClosedRightIntervalNode __pyx_result, tuple __pyx_state):
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_Int32ClosedRightIntervalNode(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pandas._libs.interval.__pyx_unpickle_Int32ClosedRightIntervalNode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_Int32ClosedRightIntervalNode__set_state(<Int32ClosedRightIntervalNode> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Int32ClosedRightIntervalNode__set_state(Int32ClosedRightIntervalNode __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_6pandas_5_libs_8interval___pyx_unpickle_Int32ClosedRightIntervalNode__set_state(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_memviewslice __pyx_t_2 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_3 = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_t_4;
  __pyx_t_5numpy_int64_t __pyx_t_5;
  __pyx_t_5numpy_int32_t __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Int32ClosedRightIntervalNode__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_Int32ClosedRightIntervalNode__set_state(Int32ClosedRightIntervalNode __pyx_result, tuple __pyx_state):
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[16])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_2.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->center_left_indices, 0);
  __pyx_v___pyx_result->center_left_indices = __pyx_t_2;
  __pyx_t_2.memview = NULL;
  __pyx_t_2.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int32_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_3.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->center_left_values, 0);
  __pyx_v___pyx_result->center_left_values = __pyx_t_3;
  __pyx_t_3.memview = NULL;
  __pyx_t_3.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_2.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->center_right_indices, 0);
  __pyx_v___pyx_result->center_right_indices = __pyx_t_2;
  __pyx_t_2.memview = NULL;
  __pyx_t_2.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int32_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_3.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->center_right_values, 0);
  __pyx_v___pyx_result->center_right_values = __pyx_t_3;
  __pyx_t_3.memview = NULL;
  __pyx_t_3.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 4, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_2.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->indices, 0);
  __pyx_v___pyx_result->indices = __pyx_t_2;
  __pyx_t_2.memview = NULL;
  __pyx_t_2.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 5, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->is_leaf_node = __pyx_t_4;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 6, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyInt_As_npy_int64(__pyx_t_1); if (unlikely((__pyx_t_5 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->leaf_size = __pyx_t_5;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 7, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int32_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_3.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->left, 0);
  __pyx_v___pyx_result->left = __pyx_t_3;
  __pyx_t_3.memview = NULL;
  __pyx_t_3.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 8, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode))))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->left_node);
  __Pyx_DECREF(((PyObject *)__pyx_v___pyx_result->left_node));
  __pyx_v___pyx_result->left_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 9, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = __Pyx_PyInt_As_npy_int32(__pyx_t_1); if (unlikely((__pyx_t_6 == ((npy_int32)-1)) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->max_right = __pyx_t_6;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 10, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = __Pyx_PyInt_As_npy_int32(__pyx_t_1); if (unlikely((__pyx_t_6 == ((npy_int32)-1)) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->min_left = __pyx_t_6;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 11, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyInt_As_npy_int64(__pyx_t_1); if (unlikely((__pyx_t_5 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->n_center = __pyx_t_5;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 12, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyInt_As_npy_int64(__pyx_t_1); if (unlikely((__pyx_t_5 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->n_elements = __pyx_t_5;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 13, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = __Pyx_PyInt_As_npy_int32(__pyx_t_1); if (unlikely((__pyx_t_6 == ((npy_int32)-1)) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->pivot = __pyx_t_6;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 14, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int32_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_3.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->right, 0);
  __pyx_v___pyx_result->right = __pyx_t_3;
  __pyx_t_3.memview = NULL;
  __pyx_t_3.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 15, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode))))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->right_node);
  __Pyx_DECREF(((PyObject *)__pyx_v___pyx_result->right_node));
  __pyx_v___pyx_result->right_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_Int32ClosedRightIntervalNode__set_state(Int32ClosedRightIntervalNode __pyx_result, tuple __pyx_state):
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[16])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(2, 13, __pyx_L1_error)
  }
  __pyx_t_7 = PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_8 = ((__pyx_t_7 > 16) != 0);
  if (__pyx_t_8) {
  } else {
    __pyx_t_4 = __pyx_t_8;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_8 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_8 == ((int)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_9 = (__pyx_t_8 != 0);
  __pyx_t_4 = __pyx_t_9;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_4) {

    /* "(tree fragment)":14
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[16])             # <<<<<<<<<<<<<<
 */
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_10)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_update); if (unlikely(!__pyx_t_11)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(2, 14, __pyx_L1_error)
    }
    __pyx_t_10 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 16, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_10)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_12 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_11))) {
      __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_11);
      if (likely(__pyx_t_12)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
        __Pyx_INCREF(__pyx_t_12);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_11, function);
      }
    }
    __pyx_t_1 = (__pyx_t_12) ? __Pyx_PyObject_Call2Args(__pyx_t_11, __pyx_t_12, __pyx_t_10) : __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_t_10);
    __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_Int32ClosedRightIntervalNode__set_state(Int32ClosedRightIntervalNode __pyx_result, tuple __pyx_state):
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[16])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_Int32ClosedRightIntervalNode__set_state(<Int32ClosedRightIntervalNode> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Int32ClosedRightIntervalNode__set_state(Int32ClosedRightIntervalNode __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_2, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_3, 1);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("pandas._libs.interval.__pyx_unpickle_Int32ClosedRightIntervalNode__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_Int32ClosedBothIntervalNode(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27__pyx_unpickle_Int32ClosedBothIntervalNode(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pandas_5_libs_8interval_27__pyx_unpickle_Int32ClosedBothIntervalNode = {"__pyx_unpickle_Int32ClosedBothIntervalNode", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_27__pyx_unpickle_Int32ClosedBothIntervalNode, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_27__pyx_unpickle_Int32ClosedBothIntervalNode(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_Int32ClosedBothIntervalNode (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_type)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_checksum)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Int32ClosedBothIntervalNode", 1, 3, 3, 1); __PYX_ERR(2, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Int32ClosedBothIntervalNode", 1, 3, 3, 2); __PYX_ERR(2, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_unpickle_Int32ClosedBothIntervalNode") < 0)) __PYX_ERR(2, 1, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Int32ClosedBothIntervalNode", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(2, 1, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.__pyx_unpickle_Int32ClosedBothIntervalNode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_26__pyx_unpickle_Int32ClosedBothIntervalNode(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_26__pyx_unpickle_Int32ClosedBothIntervalNode(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Int32ClosedBothIntervalNode", 0);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x7a104e5:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)
 */
  __pyx_t_1 = ((__pyx_v___pyx_checksum != 0x7a104e5) != 0);
  if (__pyx_t_1) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x7a104e5:
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)
 *     __pyx_result = Int32ClosedBothIntervalNode.__new__(__pyx_type)
 */
    __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_PickleError);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_2);
    __pyx_v___pyx_PickleError = __pyx_t_2;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum != 0x7a104e5:
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)             # <<<<<<<<<<<<<<
 *     __pyx_result = Int32ClosedBothIntervalNode.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_2 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_s_vs_0x7a, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_v___pyx_PickleError);
    __pyx_t_2 = __pyx_v___pyx_PickleError; __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_5, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(2, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x7a104e5:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)
 *     __pyx_result = Int32ClosedBothIntervalNode.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Int32ClosedBothIntervalNode__set_state(<Int32ClosedBothIntervalNode> __pyx_result, __pyx_state)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode), __pyx_n_s_new); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_v___pyx_type) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v___pyx_type);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v___pyx_result = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)
 *     __pyx_result = Int32ClosedBothIntervalNode.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Int32ClosedBothIntervalNode__set_state(<Int32ClosedBothIntervalNode> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_1 = (__pyx_v___pyx_state != Py_None);
  __pyx_t_6 = (__pyx_t_1 != 0);
  if (__pyx_t_6) {

    /* "(tree fragment)":9
 *     __pyx_result = Int32ClosedBothIntervalNode.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Int32ClosedBothIntervalNode__set_state(<Int32ClosedBothIntervalNode> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_Int32ClosedBothIntervalNode__set_state(Int32ClosedBothIntervalNode __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(2, 9, __pyx_L1_error)
    __pyx_t_3 = __pyx_f_6pandas_5_libs_8interval___pyx_unpickle_Int32ClosedBothIntervalNode__set_state(((struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)
 *     __pyx_result = Int32ClosedBothIntervalNode.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Int32ClosedBothIntervalNode__set_state(<Int32ClosedBothIntervalNode> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Int32ClosedBothIntervalNode__set_state(<Int32ClosedBothIntervalNode> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_Int32ClosedBothIntervalNode__set_state(Int32ClosedBothIntervalNode __pyx_result, tuple __pyx_state):
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_Int32ClosedBothIntervalNode(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pandas._libs.interval.__pyx_unpickle_Int32ClosedBothIntervalNode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_Int32ClosedBothIntervalNode__set_state(<Int32ClosedBothIntervalNode> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Int32ClosedBothIntervalNode__set_state(Int32ClosedBothIntervalNode __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_6pandas_5_libs_8interval___pyx_unpickle_Int32ClosedBothIntervalNode__set_state(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_memviewslice __pyx_t_2 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_3 = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_t_4;
  __pyx_t_5numpy_int64_t __pyx_t_5;
  __pyx_t_5numpy_int32_t __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Int32ClosedBothIntervalNode__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_Int32ClosedBothIntervalNode__set_state(Int32ClosedBothIntervalNode __pyx_result, tuple __pyx_state):
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[16])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_2.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->center_left_indices, 0);
  __pyx_v___pyx_result->center_left_indices = __pyx_t_2;
  __pyx_t_2.memview = NULL;
  __pyx_t_2.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int32_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_3.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->center_left_values, 0);
  __pyx_v___pyx_result->center_left_values = __pyx_t_3;
  __pyx_t_3.memview = NULL;
  __pyx_t_3.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_2.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->center_right_indices, 0);
  __pyx_v___pyx_result->center_right_indices = __pyx_t_2;
  __pyx_t_2.memview = NULL;
  __pyx_t_2.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int32_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_3.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->center_right_values, 0);
  __pyx_v___pyx_result->center_right_values = __pyx_t_3;
  __pyx_t_3.memview = NULL;
  __pyx_t_3.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 4, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_2.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->indices, 0);
  __pyx_v___pyx_result->indices = __pyx_t_2;
  __pyx_t_2.memview = NULL;
  __pyx_t_2.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 5, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->is_leaf_node = __pyx_t_4;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 6, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyInt_As_npy_int64(__pyx_t_1); if (unlikely((__pyx_t_5 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->leaf_size = __pyx_t_5;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 7, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int32_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_3.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->left, 0);
  __pyx_v___pyx_result->left = __pyx_t_3;
  __pyx_t_3.memview = NULL;
  __pyx_t_3.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 8, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode))))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->left_node);
  __Pyx_DECREF(((PyObject *)__pyx_v___pyx_result->left_node));
  __pyx_v___pyx_result->left_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 9, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = __Pyx_PyInt_As_npy_int32(__pyx_t_1); if (unlikely((__pyx_t_6 == ((npy_int32)-1)) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->max_right = __pyx_t_6;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 10, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = __Pyx_PyInt_As_npy_int32(__pyx_t_1); if (unlikely((__pyx_t_6 == ((npy_int32)-1)) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->min_left = __pyx_t_6;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 11, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyInt_As_npy_int64(__pyx_t_1); if (unlikely((__pyx_t_5 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->n_center = __pyx_t_5;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 12, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyInt_As_npy_int64(__pyx_t_1); if (unlikely((__pyx_t_5 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->n_elements = __pyx_t_5;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 13, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = __Pyx_PyInt_As_npy_int32(__pyx_t_1); if (unlikely((__pyx_t_6 == ((npy_int32)-1)) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->pivot = __pyx_t_6;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 14, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int32_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_3.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->right, 0);
  __pyx_v___pyx_result->right = __pyx_t_3;
  __pyx_t_3.memview = NULL;
  __pyx_t_3.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 15, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode))))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->right_node);
  __Pyx_DECREF(((PyObject *)__pyx_v___pyx_result->right_node));
  __pyx_v___pyx_result->right_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_Int32ClosedBothIntervalNode__set_state(Int32ClosedBothIntervalNode __pyx_result, tuple __pyx_state):
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[16])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(2, 13, __pyx_L1_error)
  }
  __pyx_t_7 = PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_8 = ((__pyx_t_7 > 16) != 0);
  if (__pyx_t_8) {
  } else {
    __pyx_t_4 = __pyx_t_8;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_8 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_8 == ((int)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_9 = (__pyx_t_8 != 0);
  __pyx_t_4 = __pyx_t_9;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_4) {

    /* "(tree fragment)":14
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[16])             # <<<<<<<<<<<<<<
 */
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_10)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_update); if (unlikely(!__pyx_t_11)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(2, 14, __pyx_L1_error)
    }
    __pyx_t_10 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 16, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_10)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_12 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_11))) {
      __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_11);
      if (likely(__pyx_t_12)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
        __Pyx_INCREF(__pyx_t_12);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_11, function);
      }
    }
    __pyx_t_1 = (__pyx_t_12) ? __Pyx_PyObject_Call2Args(__pyx_t_11, __pyx_t_12, __pyx_t_10) : __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_t_10);
    __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_Int32ClosedBothIntervalNode__set_state(Int32ClosedBothIntervalNode __pyx_result, tuple __pyx_state):
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[16])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_Int32ClosedBothIntervalNode__set_state(<Int32ClosedBothIntervalNode> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Int32ClosedBothIntervalNode__set_state(Int32ClosedBothIntervalNode __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_2, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_3, 1);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("pandas._libs.interval.__pyx_unpickle_Int32ClosedBothIntervalNode__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_Int32ClosedNeitherIntervalNode(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29__pyx_unpickle_Int32ClosedNeitherIntervalNode(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pandas_5_libs_8interval_29__pyx_unpickle_Int32ClosedNeitherIntervalNode = {"__pyx_unpickle_Int32ClosedNeitherIntervalNode", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_29__pyx_unpickle_Int32ClosedNeitherIntervalNode, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_29__pyx_unpickle_Int32ClosedNeitherIntervalNode(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_Int32ClosedNeitherIntervalNode (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_type)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_checksum)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Int32ClosedNeitherIntervalNode", 1, 3, 3, 1); __PYX_ERR(2, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Int32ClosedNeitherIntervalNode", 1, 3, 3, 2); __PYX_ERR(2, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_unpickle_Int32ClosedNeitherIntervalNode") < 0)) __PYX_ERR(2, 1, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Int32ClosedNeitherIntervalNode", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(2, 1, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.__pyx_unpickle_Int32ClosedNeitherIntervalNode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_28__pyx_unpickle_Int32ClosedNeitherIntervalNode(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_28__pyx_unpickle_Int32ClosedNeitherIntervalNode(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Int32ClosedNeitherIntervalNode", 0);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x7a104e5:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)
 */
  __pyx_t_1 = ((__pyx_v___pyx_checksum != 0x7a104e5) != 0);
  if (__pyx_t_1) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x7a104e5:
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)
 *     __pyx_result = Int32ClosedNeitherIntervalNode.__new__(__pyx_type)
 */
    __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_PickleError);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_2);
    __pyx_v___pyx_PickleError = __pyx_t_2;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum != 0x7a104e5:
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)             # <<<<<<<<<<<<<<
 *     __pyx_result = Int32ClosedNeitherIntervalNode.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_2 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_s_vs_0x7a, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_v___pyx_PickleError);
    __pyx_t_2 = __pyx_v___pyx_PickleError; __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_5, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(2, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x7a104e5:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)
 *     __pyx_result = Int32ClosedNeitherIntervalNode.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Int32ClosedNeitherIntervalNode__set_state(<Int32ClosedNeitherIntervalNode> __pyx_result, __pyx_state)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode), __pyx_n_s_new); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_v___pyx_type) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v___pyx_type);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v___pyx_result = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)
 *     __pyx_result = Int32ClosedNeitherIntervalNode.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Int32ClosedNeitherIntervalNode__set_state(<Int32ClosedNeitherIntervalNode> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_1 = (__pyx_v___pyx_state != Py_None);
  __pyx_t_6 = (__pyx_t_1 != 0);
  if (__pyx_t_6) {

    /* "(tree fragment)":9
 *     __pyx_result = Int32ClosedNeitherIntervalNode.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Int32ClosedNeitherIntervalNode__set_state(<Int32ClosedNeitherIntervalNode> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_Int32ClosedNeitherIntervalNode__set_state(Int32ClosedNeitherIntervalNode __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(2, 9, __pyx_L1_error)
    __pyx_t_3 = __pyx_f_6pandas_5_libs_8interval___pyx_unpickle_Int32ClosedNeitherIntervalNode__set_state(((struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)
 *     __pyx_result = Int32ClosedNeitherIntervalNode.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Int32ClosedNeitherIntervalNode__set_state(<Int32ClosedNeitherIntervalNode> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Int32ClosedNeitherIntervalNode__set_state(<Int32ClosedNeitherIntervalNode> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_Int32ClosedNeitherIntervalNode__set_state(Int32ClosedNeitherIntervalNode __pyx_result, tuple __pyx_state):
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_Int32ClosedNeitherIntervalNode(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pandas._libs.interval.__pyx_unpickle_Int32ClosedNeitherIntervalNode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_Int32ClosedNeitherIntervalNode__set_state(<Int32ClosedNeitherIntervalNode> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Int32ClosedNeitherIntervalNode__set_state(Int32ClosedNeitherIntervalNode __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_6pandas_5_libs_8interval___pyx_unpickle_Int32ClosedNeitherIntervalNode__set_state(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_memviewslice __pyx_t_2 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_3 = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_t_4;
  __pyx_t_5numpy_int64_t __pyx_t_5;
  __pyx_t_5numpy_int32_t __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Int32ClosedNeitherIntervalNode__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_Int32ClosedNeitherIntervalNode__set_state(Int32ClosedNeitherIntervalNode __pyx_result, tuple __pyx_state):
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[16])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_2.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->center_left_indices, 0);
  __pyx_v___pyx_result->center_left_indices = __pyx_t_2;
  __pyx_t_2.memview = NULL;
  __pyx_t_2.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int32_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_3.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->center_left_values, 0);
  __pyx_v___pyx_result->center_left_values = __pyx_t_3;
  __pyx_t_3.memview = NULL;
  __pyx_t_3.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_2.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->center_right_indices, 0);
  __pyx_v___pyx_result->center_right_indices = __pyx_t_2;
  __pyx_t_2.memview = NULL;
  __pyx_t_2.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int32_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_3.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->center_right_values, 0);
  __pyx_v___pyx_result->center_right_values = __pyx_t_3;
  __pyx_t_3.memview = NULL;
  __pyx_t_3.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 4, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_2.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->indices, 0);
  __pyx_v___pyx_result->indices = __pyx_t_2;
  __pyx_t_2.memview = NULL;
  __pyx_t_2.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 5, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->is_leaf_node = __pyx_t_4;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 6, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyInt_As_npy_int64(__pyx_t_1); if (unlikely((__pyx_t_5 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->leaf_size = __pyx_t_5;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 7, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int32_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_3.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->left, 0);
  __pyx_v___pyx_result->left = __pyx_t_3;
  __pyx_t_3.memview = NULL;
  __pyx_t_3.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 8, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode))))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->left_node);
  __Pyx_DECREF(((PyObject *)__pyx_v___pyx_result->left_node));
  __pyx_v___pyx_result->left_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 9, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = __Pyx_PyInt_As_npy_int32(__pyx_t_1); if (unlikely((__pyx_t_6 == ((npy_int32)-1)) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->max_right = __pyx_t_6;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 10, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = __Pyx_PyInt_As_npy_int32(__pyx_t_1); if (unlikely((__pyx_t_6 == ((npy_int32)-1)) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->min_left = __pyx_t_6;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 11, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyInt_As_npy_int64(__pyx_t_1); if (unlikely((__pyx_t_5 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->n_center = __pyx_t_5;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 12, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyInt_As_npy_int64(__pyx_t_1); if (unlikely((__pyx_t_5 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->n_elements = __pyx_t_5;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 13, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = __Pyx_PyInt_As_npy_int32(__pyx_t_1); if (unlikely((__pyx_t_6 == ((npy_int32)-1)) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->pivot = __pyx_t_6;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 14, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int32_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_3.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->right, 0);
  __pyx_v___pyx_result->right = __pyx_t_3;
  __pyx_t_3.memview = NULL;
  __pyx_t_3.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 15, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode))))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->right_node);
  __Pyx_DECREF(((PyObject *)__pyx_v___pyx_result->right_node));
  __pyx_v___pyx_result->right_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_Int32ClosedNeitherIntervalNode__set_state(Int32ClosedNeitherIntervalNode __pyx_result, tuple __pyx_state):
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[16])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(2, 13, __pyx_L1_error)
  }
  __pyx_t_7 = PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_8 = ((__pyx_t_7 > 16) != 0);
  if (__pyx_t_8) {
  } else {
    __pyx_t_4 = __pyx_t_8;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_8 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_8 == ((int)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_9 = (__pyx_t_8 != 0);
  __pyx_t_4 = __pyx_t_9;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_4) {

    /* "(tree fragment)":14
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[16])             # <<<<<<<<<<<<<<
 */
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_10)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_update); if (unlikely(!__pyx_t_11)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(2, 14, __pyx_L1_error)
    }
    __pyx_t_10 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 16, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_10)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_12 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_11))) {
      __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_11);
      if (likely(__pyx_t_12)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
        __Pyx_INCREF(__pyx_t_12);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_11, function);
      }
    }
    __pyx_t_1 = (__pyx_t_12) ? __Pyx_PyObject_Call2Args(__pyx_t_11, __pyx_t_12, __pyx_t_10) : __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_t_10);
    __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_Int32ClosedNeitherIntervalNode__set_state(Int32ClosedNeitherIntervalNode __pyx_result, tuple __pyx_state):
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[16])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_Int32ClosedNeitherIntervalNode__set_state(<Int32ClosedNeitherIntervalNode> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Int32ClosedNeitherIntervalNode__set_state(Int32ClosedNeitherIntervalNode __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_2, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_3, 1);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("pandas._libs.interval.__pyx_unpickle_Int32ClosedNeitherIntervalNode__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_Int64ClosedLeftIntervalNode(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_31__pyx_unpickle_Int64ClosedLeftIntervalNode(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pandas_5_libs_8interval_31__pyx_unpickle_Int64ClosedLeftIntervalNode = {"__pyx_unpickle_Int64ClosedLeftIntervalNode", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_31__pyx_unpickle_Int64ClosedLeftIntervalNode, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_31__pyx_unpickle_Int64ClosedLeftIntervalNode(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_Int64ClosedLeftIntervalNode (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_type)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_checksum)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Int64ClosedLeftIntervalNode", 1, 3, 3, 1); __PYX_ERR(2, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Int64ClosedLeftIntervalNode", 1, 3, 3, 2); __PYX_ERR(2, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_unpickle_Int64ClosedLeftIntervalNode") < 0)) __PYX_ERR(2, 1, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Int64ClosedLeftIntervalNode", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(2, 1, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.__pyx_unpickle_Int64ClosedLeftIntervalNode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_30__pyx_unpickle_Int64ClosedLeftIntervalNode(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_30__pyx_unpickle_Int64ClosedLeftIntervalNode(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Int64ClosedLeftIntervalNode", 0);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x7a104e5:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)
 */
  __pyx_t_1 = ((__pyx_v___pyx_checksum != 0x7a104e5) != 0);
  if (__pyx_t_1) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x7a104e5:
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)
 *     __pyx_result = Int64ClosedLeftIntervalNode.__new__(__pyx_type)
 */
    __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_PickleError);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_2);
    __pyx_v___pyx_PickleError = __pyx_t_2;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum != 0x7a104e5:
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)             # <<<<<<<<<<<<<<
 *     __pyx_result = Int64ClosedLeftIntervalNode.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_2 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_s_vs_0x7a, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_v___pyx_PickleError);
    __pyx_t_2 = __pyx_v___pyx_PickleError; __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_5, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(2, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x7a104e5:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)
 *     __pyx_result = Int64ClosedLeftIntervalNode.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Int64ClosedLeftIntervalNode__set_state(<Int64ClosedLeftIntervalNode> __pyx_result, __pyx_state)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode), __pyx_n_s_new); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_v___pyx_type) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v___pyx_type);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v___pyx_result = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)
 *     __pyx_result = Int64ClosedLeftIntervalNode.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Int64ClosedLeftIntervalNode__set_state(<Int64ClosedLeftIntervalNode> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_1 = (__pyx_v___pyx_state != Py_None);
  __pyx_t_6 = (__pyx_t_1 != 0);
  if (__pyx_t_6) {

    /* "(tree fragment)":9
 *     __pyx_result = Int64ClosedLeftIntervalNode.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Int64ClosedLeftIntervalNode__set_state(<Int64ClosedLeftIntervalNode> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_Int64ClosedLeftIntervalNode__set_state(Int64ClosedLeftIntervalNode __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(2, 9, __pyx_L1_error)
    __pyx_t_3 = __pyx_f_6pandas_5_libs_8interval___pyx_unpickle_Int64ClosedLeftIntervalNode__set_state(((struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)
 *     __pyx_result = Int64ClosedLeftIntervalNode.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Int64ClosedLeftIntervalNode__set_state(<Int64ClosedLeftIntervalNode> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Int64ClosedLeftIntervalNode__set_state(<Int64ClosedLeftIntervalNode> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_Int64ClosedLeftIntervalNode__set_state(Int64ClosedLeftIntervalNode __pyx_result, tuple __pyx_state):
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_Int64ClosedLeftIntervalNode(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pandas._libs.interval.__pyx_unpickle_Int64ClosedLeftIntervalNode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_Int64ClosedLeftIntervalNode__set_state(<Int64ClosedLeftIntervalNode> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Int64ClosedLeftIntervalNode__set_state(Int64ClosedLeftIntervalNode __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_6pandas_5_libs_8interval___pyx_unpickle_Int64ClosedLeftIntervalNode__set_state(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_memviewslice __pyx_t_2 = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_t_3;
  __pyx_t_5numpy_int64_t __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Int64ClosedLeftIntervalNode__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_Int64ClosedLeftIntervalNode__set_state(Int64ClosedLeftIntervalNode __pyx_result, tuple __pyx_state):
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[16])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_2.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->center_left_indices, 0);
  __pyx_v___pyx_result->center_left_indices = __pyx_t_2;
  __pyx_t_2.memview = NULL;
  __pyx_t_2.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_2.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->center_left_values, 0);
  __pyx_v___pyx_result->center_left_values = __pyx_t_2;
  __pyx_t_2.memview = NULL;
  __pyx_t_2.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_2.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->center_right_indices, 0);
  __pyx_v___pyx_result->center_right_indices = __pyx_t_2;
  __pyx_t_2.memview = NULL;
  __pyx_t_2.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_2.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->center_right_values, 0);
  __pyx_v___pyx_result->center_right_values = __pyx_t_2;
  __pyx_t_2.memview = NULL;
  __pyx_t_2.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 4, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_2.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->indices, 0);
  __pyx_v___pyx_result->indices = __pyx_t_2;
  __pyx_t_2.memview = NULL;
  __pyx_t_2.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 5, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_3 == (int)-1) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->is_leaf_node = __pyx_t_3;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 6, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyInt_As_npy_int64(__pyx_t_1); if (unlikely((__pyx_t_4 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->leaf_size = __pyx_t_4;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 7, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_2.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->left, 0);
  __pyx_v___pyx_result->left = __pyx_t_2;
  __pyx_t_2.memview = NULL;
  __pyx_t_2.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 8, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode))))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->left_node);
  __Pyx_DECREF(((PyObject *)__pyx_v___pyx_result->left_node));
  __pyx_v___pyx_result->left_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 9, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyInt_As_npy_int64(__pyx_t_1); if (unlikely((__pyx_t_4 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->max_right = __pyx_t_4;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 10, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyInt_As_npy_int64(__pyx_t_1); if (unlikely((__pyx_t_4 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->min_left = __pyx_t_4;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 11, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyInt_As_npy_int64(__pyx_t_1); if (unlikely((__pyx_t_4 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->n_center = __pyx_t_4;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 12, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyInt_As_npy_int64(__pyx_t_1); if (unlikely((__pyx_t_4 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->n_elements = __pyx_t_4;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 13, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyInt_As_npy_int64(__pyx_t_1); if (unlikely((__pyx_t_4 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->pivot = __pyx_t_4;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 14, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_2.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->right, 0);
  __pyx_v___pyx_result->right = __pyx_t_2;
  __pyx_t_2.memview = NULL;
  __pyx_t_2.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 15, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode))))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->right_node);
  __Pyx_DECREF(((PyObject *)__pyx_v___pyx_result->right_node));
  __pyx_v___pyx_result->right_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_Int64ClosedLeftIntervalNode__set_state(Int64ClosedLeftIntervalNode __pyx_result, tuple __pyx_state):
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[16])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(2, 13, __pyx_L1_error)
  }
  __pyx_t_5 = PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_6 = ((__pyx_t_5 > 16) != 0);
  if (__pyx_t_6) {
  } else {
    __pyx_t_3 = __pyx_t_6;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_6 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_7 = (__pyx_t_6 != 0);
  __pyx_t_3 = __pyx_t_7;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_3) {

    /* "(tree fragment)":14
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[16])             # <<<<<<<<<<<<<<
 */
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_update); if (unlikely(!__pyx_t_9)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(2, 14, __pyx_L1_error)
    }
    __pyx_t_8 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 16, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_10 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_9))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_9);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_9, function);
      }
    }
    __pyx_t_1 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_9, __pyx_t_10, __pyx_t_8) : __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_t_8);
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_Int64ClosedLeftIntervalNode__set_state(Int64ClosedLeftIntervalNode __pyx_result, tuple __pyx_state):
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[16])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_Int64ClosedLeftIntervalNode__set_state(<Int64ClosedLeftIntervalNode> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Int64ClosedLeftIntervalNode__set_state(Int64ClosedLeftIntervalNode __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_2, 1);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("pandas._libs.interval.__pyx_unpickle_Int64ClosedLeftIntervalNode__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_Int64ClosedRightIntervalNode(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_33__pyx_unpickle_Int64ClosedRightIntervalNode(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pandas_5_libs_8interval_33__pyx_unpickle_Int64ClosedRightIntervalNode = {"__pyx_unpickle_Int64ClosedRightIntervalNode", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_33__pyx_unpickle_Int64ClosedRightIntervalNode, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_33__pyx_unpickle_Int64ClosedRightIntervalNode(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_Int64ClosedRightIntervalNode (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_type)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_checksum)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Int64ClosedRightIntervalNode", 1, 3, 3, 1); __PYX_ERR(2, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Int64ClosedRightIntervalNode", 1, 3, 3, 2); __PYX_ERR(2, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_unpickle_Int64ClosedRightIntervalNode") < 0)) __PYX_ERR(2, 1, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Int64ClosedRightIntervalNode", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(2, 1, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.__pyx_unpickle_Int64ClosedRightIntervalNode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_32__pyx_unpickle_Int64ClosedRightIntervalNode(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_32__pyx_unpickle_Int64ClosedRightIntervalNode(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Int64ClosedRightIntervalNode", 0);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x7a104e5:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)
 */
  __pyx_t_1 = ((__pyx_v___pyx_checksum != 0x7a104e5) != 0);
  if (__pyx_t_1) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x7a104e5:
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)
 *     __pyx_result = Int64ClosedRightIntervalNode.__new__(__pyx_type)
 */
    __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_PickleError);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_2);
    __pyx_v___pyx_PickleError = __pyx_t_2;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum != 0x7a104e5:
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)             # <<<<<<<<<<<<<<
 *     __pyx_result = Int64ClosedRightIntervalNode.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_2 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_s_vs_0x7a, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_v___pyx_PickleError);
    __pyx_t_2 = __pyx_v___pyx_PickleError; __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_5, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(2, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x7a104e5:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)
 *     __pyx_result = Int64ClosedRightIntervalNode.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Int64ClosedRightIntervalNode__set_state(<Int64ClosedRightIntervalNode> __pyx_result, __pyx_state)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode), __pyx_n_s_new); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_v___pyx_type) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v___pyx_type);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v___pyx_result = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)
 *     __pyx_result = Int64ClosedRightIntervalNode.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Int64ClosedRightIntervalNode__set_state(<Int64ClosedRightIntervalNode> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_1 = (__pyx_v___pyx_state != Py_None);
  __pyx_t_6 = (__pyx_t_1 != 0);
  if (__pyx_t_6) {

    /* "(tree fragment)":9
 *     __pyx_result = Int64ClosedRightIntervalNode.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Int64ClosedRightIntervalNode__set_state(<Int64ClosedRightIntervalNode> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_Int64ClosedRightIntervalNode__set_state(Int64ClosedRightIntervalNode __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(2, 9, __pyx_L1_error)
    __pyx_t_3 = __pyx_f_6pandas_5_libs_8interval___pyx_unpickle_Int64ClosedRightIntervalNode__set_state(((struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)
 *     __pyx_result = Int64ClosedRightIntervalNode.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Int64ClosedRightIntervalNode__set_state(<Int64ClosedRightIntervalNode> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Int64ClosedRightIntervalNode__set_state(<Int64ClosedRightIntervalNode> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_Int64ClosedRightIntervalNode__set_state(Int64ClosedRightIntervalNode __pyx_result, tuple __pyx_state):
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_Int64ClosedRightIntervalNode(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pandas._libs.interval.__pyx_unpickle_Int64ClosedRightIntervalNode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_Int64ClosedRightIntervalNode__set_state(<Int64ClosedRightIntervalNode> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Int64ClosedRightIntervalNode__set_state(Int64ClosedRightIntervalNode __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_6pandas_5_libs_8interval___pyx_unpickle_Int64ClosedRightIntervalNode__set_state(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_memviewslice __pyx_t_2 = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_t_3;
  __pyx_t_5numpy_int64_t __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Int64ClosedRightIntervalNode__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_Int64ClosedRightIntervalNode__set_state(Int64ClosedRightIntervalNode __pyx_result, tuple __pyx_state):
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[16])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_2.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->center_left_indices, 0);
  __pyx_v___pyx_result->center_left_indices = __pyx_t_2;
  __pyx_t_2.memview = NULL;
  __pyx_t_2.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_2.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->center_left_values, 0);
  __pyx_v___pyx_result->center_left_values = __pyx_t_2;
  __pyx_t_2.memview = NULL;
  __pyx_t_2.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_2.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->center_right_indices, 0);
  __pyx_v___pyx_result->center_right_indices = __pyx_t_2;
  __pyx_t_2.memview = NULL;
  __pyx_t_2.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_2.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->center_right_values, 0);
  __pyx_v___pyx_result->center_right_values = __pyx_t_2;
  __pyx_t_2.memview = NULL;
  __pyx_t_2.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 4, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_2.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->indices, 0);
  __pyx_v___pyx_result->indices = __pyx_t_2;
  __pyx_t_2.memview = NULL;
  __pyx_t_2.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 5, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_3 == (int)-1) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->is_leaf_node = __pyx_t_3;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 6, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyInt_As_npy_int64(__pyx_t_1); if (unlikely((__pyx_t_4 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->leaf_size = __pyx_t_4;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 7, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_2.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->left, 0);
  __pyx_v___pyx_result->left = __pyx_t_2;
  __pyx_t_2.memview = NULL;
  __pyx_t_2.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 8, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode))))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->left_node);
  __Pyx_DECREF(((PyObject *)__pyx_v___pyx_result->left_node));
  __pyx_v___pyx_result->left_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 9, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyInt_As_npy_int64(__pyx_t_1); if (unlikely((__pyx_t_4 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->max_right = __pyx_t_4;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 10, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyInt_As_npy_int64(__pyx_t_1); if (unlikely((__pyx_t_4 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->min_left = __pyx_t_4;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 11, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyInt_As_npy_int64(__pyx_t_1); if (unlikely((__pyx_t_4 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->n_center = __pyx_t_4;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 12, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyInt_As_npy_int64(__pyx_t_1); if (unlikely((__pyx_t_4 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->n_elements = __pyx_t_4;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 13, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyInt_As_npy_int64(__pyx_t_1); if (unlikely((__pyx_t_4 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->pivot = __pyx_t_4;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 14, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_2.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->right, 0);
  __pyx_v___pyx_result->right = __pyx_t_2;
  __pyx_t_2.memview = NULL;
  __pyx_t_2.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 15, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode))))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->right_node);
  __Pyx_DECREF(((PyObject *)__pyx_v___pyx_result->right_node));
  __pyx_v___pyx_result->right_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_Int64ClosedRightIntervalNode__set_state(Int64ClosedRightIntervalNode __pyx_result, tuple __pyx_state):
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[16])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(2, 13, __pyx_L1_error)
  }
  __pyx_t_5 = PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_6 = ((__pyx_t_5 > 16) != 0);
  if (__pyx_t_6) {
  } else {
    __pyx_t_3 = __pyx_t_6;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_6 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_7 = (__pyx_t_6 != 0);
  __pyx_t_3 = __pyx_t_7;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_3) {

    /* "(tree fragment)":14
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[16])             # <<<<<<<<<<<<<<
 */
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_update); if (unlikely(!__pyx_t_9)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(2, 14, __pyx_L1_error)
    }
    __pyx_t_8 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 16, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_10 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_9))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_9);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_9, function);
      }
    }
    __pyx_t_1 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_9, __pyx_t_10, __pyx_t_8) : __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_t_8);
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_Int64ClosedRightIntervalNode__set_state(Int64ClosedRightIntervalNode __pyx_result, tuple __pyx_state):
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[16])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_Int64ClosedRightIntervalNode__set_state(<Int64ClosedRightIntervalNode> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Int64ClosedRightIntervalNode__set_state(Int64ClosedRightIntervalNode __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_2, 1);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("pandas._libs.interval.__pyx_unpickle_Int64ClosedRightIntervalNode__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_Int64ClosedBothIntervalNode(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_35__pyx_unpickle_Int64ClosedBothIntervalNode(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pandas_5_libs_8interval_35__pyx_unpickle_Int64ClosedBothIntervalNode = {"__pyx_unpickle_Int64ClosedBothIntervalNode", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_35__pyx_unpickle_Int64ClosedBothIntervalNode, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_35__pyx_unpickle_Int64ClosedBothIntervalNode(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_Int64ClosedBothIntervalNode (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_type)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_checksum)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Int64ClosedBothIntervalNode", 1, 3, 3, 1); __PYX_ERR(2, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Int64ClosedBothIntervalNode", 1, 3, 3, 2); __PYX_ERR(2, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_unpickle_Int64ClosedBothIntervalNode") < 0)) __PYX_ERR(2, 1, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Int64ClosedBothIntervalNode", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(2, 1, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.__pyx_unpickle_Int64ClosedBothIntervalNode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_34__pyx_unpickle_Int64ClosedBothIntervalNode(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_34__pyx_unpickle_Int64ClosedBothIntervalNode(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Int64ClosedBothIntervalNode", 0);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x7a104e5:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)
 */
  __pyx_t_1 = ((__pyx_v___pyx_checksum != 0x7a104e5) != 0);
  if (__pyx_t_1) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x7a104e5:
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)
 *     __pyx_result = Int64ClosedBothIntervalNode.__new__(__pyx_type)
 */
    __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_PickleError);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_2);
    __pyx_v___pyx_PickleError = __pyx_t_2;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum != 0x7a104e5:
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)             # <<<<<<<<<<<<<<
 *     __pyx_result = Int64ClosedBothIntervalNode.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_2 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_s_vs_0x7a, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_v___pyx_PickleError);
    __pyx_t_2 = __pyx_v___pyx_PickleError; __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_5, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(2, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x7a104e5:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)
 *     __pyx_result = Int64ClosedBothIntervalNode.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Int64ClosedBothIntervalNode__set_state(<Int64ClosedBothIntervalNode> __pyx_result, __pyx_state)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode), __pyx_n_s_new); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_v___pyx_type) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v___pyx_type);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v___pyx_result = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)
 *     __pyx_result = Int64ClosedBothIntervalNode.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Int64ClosedBothIntervalNode__set_state(<Int64ClosedBothIntervalNode> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_1 = (__pyx_v___pyx_state != Py_None);
  __pyx_t_6 = (__pyx_t_1 != 0);
  if (__pyx_t_6) {

    /* "(tree fragment)":9
 *     __pyx_result = Int64ClosedBothIntervalNode.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Int64ClosedBothIntervalNode__set_state(<Int64ClosedBothIntervalNode> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_Int64ClosedBothIntervalNode__set_state(Int64ClosedBothIntervalNode __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(2, 9, __pyx_L1_error)
    __pyx_t_3 = __pyx_f_6pandas_5_libs_8interval___pyx_unpickle_Int64ClosedBothIntervalNode__set_state(((struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)
 *     __pyx_result = Int64ClosedBothIntervalNode.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Int64ClosedBothIntervalNode__set_state(<Int64ClosedBothIntervalNode> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Int64ClosedBothIntervalNode__set_state(<Int64ClosedBothIntervalNode> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_Int64ClosedBothIntervalNode__set_state(Int64ClosedBothIntervalNode __pyx_result, tuple __pyx_state):
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_Int64ClosedBothIntervalNode(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pandas._libs.interval.__pyx_unpickle_Int64ClosedBothIntervalNode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_Int64ClosedBothIntervalNode__set_state(<Int64ClosedBothIntervalNode> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Int64ClosedBothIntervalNode__set_state(Int64ClosedBothIntervalNode __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_6pandas_5_libs_8interval___pyx_unpickle_Int64ClosedBothIntervalNode__set_state(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_memviewslice __pyx_t_2 = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_t_3;
  __pyx_t_5numpy_int64_t __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Int64ClosedBothIntervalNode__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_Int64ClosedBothIntervalNode__set_state(Int64ClosedBothIntervalNode __pyx_result, tuple __pyx_state):
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[16])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_2.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->center_left_indices, 0);
  __pyx_v___pyx_result->center_left_indices = __pyx_t_2;
  __pyx_t_2.memview = NULL;
  __pyx_t_2.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_2.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->center_left_values, 0);
  __pyx_v___pyx_result->center_left_values = __pyx_t_2;
  __pyx_t_2.memview = NULL;
  __pyx_t_2.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_2.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->center_right_indices, 0);
  __pyx_v___pyx_result->center_right_indices = __pyx_t_2;
  __pyx_t_2.memview = NULL;
  __pyx_t_2.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_2.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->center_right_values, 0);
  __pyx_v___pyx_result->center_right_values = __pyx_t_2;
  __pyx_t_2.memview = NULL;
  __pyx_t_2.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 4, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_2.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->indices, 0);
  __pyx_v___pyx_result->indices = __pyx_t_2;
  __pyx_t_2.memview = NULL;
  __pyx_t_2.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 5, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_3 == (int)-1) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->is_leaf_node = __pyx_t_3;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 6, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyInt_As_npy_int64(__pyx_t_1); if (unlikely((__pyx_t_4 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->leaf_size = __pyx_t_4;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 7, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_2.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->left, 0);
  __pyx_v___pyx_result->left = __pyx_t_2;
  __pyx_t_2.memview = NULL;
  __pyx_t_2.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 8, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode))))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->left_node);
  __Pyx_DECREF(((PyObject *)__pyx_v___pyx_result->left_node));
  __pyx_v___pyx_result->left_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 9, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyInt_As_npy_int64(__pyx_t_1); if (unlikely((__pyx_t_4 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->max_right = __pyx_t_4;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 10, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyInt_As_npy_int64(__pyx_t_1); if (unlikely((__pyx_t_4 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->min_left = __pyx_t_4;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 11, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyInt_As_npy_int64(__pyx_t_1); if (unlikely((__pyx_t_4 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->n_center = __pyx_t_4;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 12, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyInt_As_npy_int64(__pyx_t_1); if (unlikely((__pyx_t_4 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->n_elements = __pyx_t_4;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 13, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyInt_As_npy_int64(__pyx_t_1); if (unlikely((__pyx_t_4 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->pivot = __pyx_t_4;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 14, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_2.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->right, 0);
  __pyx_v___pyx_result->right = __pyx_t_2;
  __pyx_t_2.memview = NULL;
  __pyx_t_2.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 15, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode))))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->right_node);
  __Pyx_DECREF(((PyObject *)__pyx_v___pyx_result->right_node));
  __pyx_v___pyx_result->right_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_Int64ClosedBothIntervalNode__set_state(Int64ClosedBothIntervalNode __pyx_result, tuple __pyx_state):
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[16])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(2, 13, __pyx_L1_error)
  }
  __pyx_t_5 = PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_6 = ((__pyx_t_5 > 16) != 0);
  if (__pyx_t_6) {
  } else {
    __pyx_t_3 = __pyx_t_6;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_6 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_7 = (__pyx_t_6 != 0);
  __pyx_t_3 = __pyx_t_7;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_3) {

    /* "(tree fragment)":14
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[16])             # <<<<<<<<<<<<<<
 */
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_update); if (unlikely(!__pyx_t_9)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(2, 14, __pyx_L1_error)
    }
    __pyx_t_8 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 16, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_10 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_9))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_9);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_9, function);
      }
    }
    __pyx_t_1 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_9, __pyx_t_10, __pyx_t_8) : __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_t_8);
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_Int64ClosedBothIntervalNode__set_state(Int64ClosedBothIntervalNode __pyx_result, tuple __pyx_state):
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[16])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_Int64ClosedBothIntervalNode__set_state(<Int64ClosedBothIntervalNode> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Int64ClosedBothIntervalNode__set_state(Int64ClosedBothIntervalNode __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_2, 1);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("pandas._libs.interval.__pyx_unpickle_Int64ClosedBothIntervalNode__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_Int64ClosedNeitherIntervalNode(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_37__pyx_unpickle_Int64ClosedNeitherIntervalNode(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pandas_5_libs_8interval_37__pyx_unpickle_Int64ClosedNeitherIntervalNode = {"__pyx_unpickle_Int64ClosedNeitherIntervalNode", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_37__pyx_unpickle_Int64ClosedNeitherIntervalNode, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_37__pyx_unpickle_Int64ClosedNeitherIntervalNode(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_Int64ClosedNeitherIntervalNode (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_type)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_checksum)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Int64ClosedNeitherIntervalNode", 1, 3, 3, 1); __PYX_ERR(2, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Int64ClosedNeitherIntervalNode", 1, 3, 3, 2); __PYX_ERR(2, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_unpickle_Int64ClosedNeitherIntervalNode") < 0)) __PYX_ERR(2, 1, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Int64ClosedNeitherIntervalNode", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(2, 1, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.__pyx_unpickle_Int64ClosedNeitherIntervalNode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_36__pyx_unpickle_Int64ClosedNeitherIntervalNode(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_36__pyx_unpickle_Int64ClosedNeitherIntervalNode(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Int64ClosedNeitherIntervalNode", 0);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x7a104e5:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)
 */
  __pyx_t_1 = ((__pyx_v___pyx_checksum != 0x7a104e5) != 0);
  if (__pyx_t_1) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x7a104e5:
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)
 *     __pyx_result = Int64ClosedNeitherIntervalNode.__new__(__pyx_type)
 */
    __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_PickleError);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_2);
    __pyx_v___pyx_PickleError = __pyx_t_2;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum != 0x7a104e5:
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)             # <<<<<<<<<<<<<<
 *     __pyx_result = Int64ClosedNeitherIntervalNode.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_2 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_s_vs_0x7a, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_v___pyx_PickleError);
    __pyx_t_2 = __pyx_v___pyx_PickleError; __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_5, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(2, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x7a104e5:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)
 *     __pyx_result = Int64ClosedNeitherIntervalNode.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Int64ClosedNeitherIntervalNode__set_state(<Int64ClosedNeitherIntervalNode> __pyx_result, __pyx_state)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode), __pyx_n_s_new); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_v___pyx_type) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v___pyx_type);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v___pyx_result = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)
 *     __pyx_result = Int64ClosedNeitherIntervalNode.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Int64ClosedNeitherIntervalNode__set_state(<Int64ClosedNeitherIntervalNode> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_1 = (__pyx_v___pyx_state != Py_None);
  __pyx_t_6 = (__pyx_t_1 != 0);
  if (__pyx_t_6) {

    /* "(tree fragment)":9
 *     __pyx_result = Int64ClosedNeitherIntervalNode.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Int64ClosedNeitherIntervalNode__set_state(<Int64ClosedNeitherIntervalNode> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_Int64ClosedNeitherIntervalNode__set_state(Int64ClosedNeitherIntervalNode __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(2, 9, __pyx_L1_error)
    __pyx_t_3 = __pyx_f_6pandas_5_libs_8interval___pyx_unpickle_Int64ClosedNeitherIntervalNode__set_state(((struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)
 *     __pyx_result = Int64ClosedNeitherIntervalNode.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Int64ClosedNeitherIntervalNode__set_state(<Int64ClosedNeitherIntervalNode> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Int64ClosedNeitherIntervalNode__set_state(<Int64ClosedNeitherIntervalNode> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_Int64ClosedNeitherIntervalNode__set_state(Int64ClosedNeitherIntervalNode __pyx_result, tuple __pyx_state):
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_Int64ClosedNeitherIntervalNode(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pandas._libs.interval.__pyx_unpickle_Int64ClosedNeitherIntervalNode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_Int64ClosedNeitherIntervalNode__set_state(<Int64ClosedNeitherIntervalNode> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Int64ClosedNeitherIntervalNode__set_state(Int64ClosedNeitherIntervalNode __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_6pandas_5_libs_8interval___pyx_unpickle_Int64ClosedNeitherIntervalNode__set_state(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_memviewslice __pyx_t_2 = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_t_3;
  __pyx_t_5numpy_int64_t __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Int64ClosedNeitherIntervalNode__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_Int64ClosedNeitherIntervalNode__set_state(Int64ClosedNeitherIntervalNode __pyx_result, tuple __pyx_state):
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[16])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_2.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->center_left_indices, 0);
  __pyx_v___pyx_result->center_left_indices = __pyx_t_2;
  __pyx_t_2.memview = NULL;
  __pyx_t_2.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_2.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->center_left_values, 0);
  __pyx_v___pyx_result->center_left_values = __pyx_t_2;
  __pyx_t_2.memview = NULL;
  __pyx_t_2.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_2.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->center_right_indices, 0);
  __pyx_v___pyx_result->center_right_indices = __pyx_t_2;
  __pyx_t_2.memview = NULL;
  __pyx_t_2.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_2.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->center_right_values, 0);
  __pyx_v___pyx_result->center_right_values = __pyx_t_2;
  __pyx_t_2.memview = NULL;
  __pyx_t_2.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 4, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_2.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->indices, 0);
  __pyx_v___pyx_result->indices = __pyx_t_2;
  __pyx_t_2.memview = NULL;
  __pyx_t_2.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 5, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_3 == (int)-1) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->is_leaf_node = __pyx_t_3;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 6, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyInt_As_npy_int64(__pyx_t_1); if (unlikely((__pyx_t_4 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->leaf_size = __pyx_t_4;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 7, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_2.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->left, 0);
  __pyx_v___pyx_result->left = __pyx_t_2;
  __pyx_t_2.memview = NULL;
  __pyx_t_2.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 8, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode))))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->left_node);
  __Pyx_DECREF(((PyObject *)__pyx_v___pyx_result->left_node));
  __pyx_v___pyx_result->left_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 9, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyInt_As_npy_int64(__pyx_t_1); if (unlikely((__pyx_t_4 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->max_right = __pyx_t_4;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 10, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyInt_As_npy_int64(__pyx_t_1); if (unlikely((__pyx_t_4 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->min_left = __pyx_t_4;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 11, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyInt_As_npy_int64(__pyx_t_1); if (unlikely((__pyx_t_4 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->n_center = __pyx_t_4;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 12, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyInt_As_npy_int64(__pyx_t_1); if (unlikely((__pyx_t_4 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->n_elements = __pyx_t_4;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 13, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyInt_As_npy_int64(__pyx_t_1); if (unlikely((__pyx_t_4 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->pivot = __pyx_t_4;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 14, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_2.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->right, 0);
  __pyx_v___pyx_result->right = __pyx_t_2;
  __pyx_t_2.memview = NULL;
  __pyx_t_2.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 15, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode))))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->right_node);
  __Pyx_DECREF(((PyObject *)__pyx_v___pyx_result->right_node));
  __pyx_v___pyx_result->right_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_Int64ClosedNeitherIntervalNode__set_state(Int64ClosedNeitherIntervalNode __pyx_result, tuple __pyx_state):
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[16])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(2, 13, __pyx_L1_error)
  }
  __pyx_t_5 = PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_6 = ((__pyx_t_5 > 16) != 0);
  if (__pyx_t_6) {
  } else {
    __pyx_t_3 = __pyx_t_6;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_6 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_7 = (__pyx_t_6 != 0);
  __pyx_t_3 = __pyx_t_7;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_3) {

    /* "(tree fragment)":14
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[16])             # <<<<<<<<<<<<<<
 */
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_update); if (unlikely(!__pyx_t_9)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(2, 14, __pyx_L1_error)
    }
    __pyx_t_8 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 16, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_10 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_9))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_9);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_9, function);
      }
    }
    __pyx_t_1 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_9, __pyx_t_10, __pyx_t_8) : __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_t_8);
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_Int64ClosedNeitherIntervalNode__set_state(Int64ClosedNeitherIntervalNode __pyx_result, tuple __pyx_state):
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[16])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_Int64ClosedNeitherIntervalNode__set_state(<Int64ClosedNeitherIntervalNode> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Int64ClosedNeitherIntervalNode__set_state(Int64ClosedNeitherIntervalNode __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_2, 1);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("pandas._libs.interval.__pyx_unpickle_Int64ClosedNeitherIntervalNode__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_Uint64ClosedLeftIntervalNode(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_39__pyx_unpickle_Uint64ClosedLeftIntervalNode(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pandas_5_libs_8interval_39__pyx_unpickle_Uint64ClosedLeftIntervalNode = {"__pyx_unpickle_Uint64ClosedLeftIntervalNode", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_39__pyx_unpickle_Uint64ClosedLeftIntervalNode, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_39__pyx_unpickle_Uint64ClosedLeftIntervalNode(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_Uint64ClosedLeftIntervalNode (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_type)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_checksum)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Uint64ClosedLeftIntervalNode", 1, 3, 3, 1); __PYX_ERR(2, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Uint64ClosedLeftIntervalNode", 1, 3, 3, 2); __PYX_ERR(2, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_unpickle_Uint64ClosedLeftIntervalNode") < 0)) __PYX_ERR(2, 1, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Uint64ClosedLeftIntervalNode", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(2, 1, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.__pyx_unpickle_Uint64ClosedLeftIntervalNode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_38__pyx_unpickle_Uint64ClosedLeftIntervalNode(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_38__pyx_unpickle_Uint64ClosedLeftIntervalNode(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Uint64ClosedLeftIntervalNode", 0);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x7a104e5:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)
 */
  __pyx_t_1 = ((__pyx_v___pyx_checksum != 0x7a104e5) != 0);
  if (__pyx_t_1) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x7a104e5:
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)
 *     __pyx_result = Uint64ClosedLeftIntervalNode.__new__(__pyx_type)
 */
    __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_PickleError);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_2);
    __pyx_v___pyx_PickleError = __pyx_t_2;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum != 0x7a104e5:
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)             # <<<<<<<<<<<<<<
 *     __pyx_result = Uint64ClosedLeftIntervalNode.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_2 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_s_vs_0x7a, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_v___pyx_PickleError);
    __pyx_t_2 = __pyx_v___pyx_PickleError; __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_5, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(2, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x7a104e5:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)
 *     __pyx_result = Uint64ClosedLeftIntervalNode.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Uint64ClosedLeftIntervalNode__set_state(<Uint64ClosedLeftIntervalNode> __pyx_result, __pyx_state)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode), __pyx_n_s_new); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_v___pyx_type) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v___pyx_type);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v___pyx_result = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)
 *     __pyx_result = Uint64ClosedLeftIntervalNode.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Uint64ClosedLeftIntervalNode__set_state(<Uint64ClosedLeftIntervalNode> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_1 = (__pyx_v___pyx_state != Py_None);
  __pyx_t_6 = (__pyx_t_1 != 0);
  if (__pyx_t_6) {

    /* "(tree fragment)":9
 *     __pyx_result = Uint64ClosedLeftIntervalNode.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Uint64ClosedLeftIntervalNode__set_state(<Uint64ClosedLeftIntervalNode> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_Uint64ClosedLeftIntervalNode__set_state(Uint64ClosedLeftIntervalNode __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(2, 9, __pyx_L1_error)
    __pyx_t_3 = __pyx_f_6pandas_5_libs_8interval___pyx_unpickle_Uint64ClosedLeftIntervalNode__set_state(((struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)
 *     __pyx_result = Uint64ClosedLeftIntervalNode.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Uint64ClosedLeftIntervalNode__set_state(<Uint64ClosedLeftIntervalNode> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Uint64ClosedLeftIntervalNode__set_state(<Uint64ClosedLeftIntervalNode> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_Uint64ClosedLeftIntervalNode__set_state(Uint64ClosedLeftIntervalNode __pyx_result, tuple __pyx_state):
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_Uint64ClosedLeftIntervalNode(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pandas._libs.interval.__pyx_unpickle_Uint64ClosedLeftIntervalNode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_Uint64ClosedLeftIntervalNode__set_state(<Uint64ClosedLeftIntervalNode> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Uint64ClosedLeftIntervalNode__set_state(Uint64ClosedLeftIntervalNode __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_6pandas_5_libs_8interval___pyx_unpickle_Uint64ClosedLeftIntervalNode__set_state(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_memviewslice __pyx_t_2 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_3 = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_t_4;
  __pyx_t_5numpy_int64_t __pyx_t_5;
  __pyx_t_5numpy_uint64_t __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Uint64ClosedLeftIntervalNode__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_Uint64ClosedLeftIntervalNode__set_state(Uint64ClosedLeftIntervalNode __pyx_result, tuple __pyx_state):
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[16])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_2.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->center_left_indices, 0);
  __pyx_v___pyx_result->center_left_indices = __pyx_t_2;
  __pyx_t_2.memview = NULL;
  __pyx_t_2.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_uint64_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_3.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->center_left_values, 0);
  __pyx_v___pyx_result->center_left_values = __pyx_t_3;
  __pyx_t_3.memview = NULL;
  __pyx_t_3.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_2.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->center_right_indices, 0);
  __pyx_v___pyx_result->center_right_indices = __pyx_t_2;
  __pyx_t_2.memview = NULL;
  __pyx_t_2.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_uint64_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_3.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->center_right_values, 0);
  __pyx_v___pyx_result->center_right_values = __pyx_t_3;
  __pyx_t_3.memview = NULL;
  __pyx_t_3.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 4, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_2.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->indices, 0);
  __pyx_v___pyx_result->indices = __pyx_t_2;
  __pyx_t_2.memview = NULL;
  __pyx_t_2.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 5, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->is_leaf_node = __pyx_t_4;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 6, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyInt_As_npy_int64(__pyx_t_1); if (unlikely((__pyx_t_5 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->leaf_size = __pyx_t_5;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 7, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_uint64_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_3.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->left, 0);
  __pyx_v___pyx_result->left = __pyx_t_3;
  __pyx_t_3.memview = NULL;
  __pyx_t_3.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 8, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode))))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->left_node);
  __Pyx_DECREF(((PyObject *)__pyx_v___pyx_result->left_node));
  __pyx_v___pyx_result->left_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 9, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = __Pyx_PyInt_As_npy_uint64(__pyx_t_1); if (unlikely((__pyx_t_6 == ((npy_uint64)-1)) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->max_right = __pyx_t_6;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 10, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = __Pyx_PyInt_As_npy_uint64(__pyx_t_1); if (unlikely((__pyx_t_6 == ((npy_uint64)-1)) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->min_left = __pyx_t_6;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 11, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyInt_As_npy_int64(__pyx_t_1); if (unlikely((__pyx_t_5 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->n_center = __pyx_t_5;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 12, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyInt_As_npy_int64(__pyx_t_1); if (unlikely((__pyx_t_5 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->n_elements = __pyx_t_5;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 13, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = __Pyx_PyInt_As_npy_uint64(__pyx_t_1); if (unlikely((__pyx_t_6 == ((npy_uint64)-1)) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->pivot = __pyx_t_6;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 14, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_uint64_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_3.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->right, 0);
  __pyx_v___pyx_result->right = __pyx_t_3;
  __pyx_t_3.memview = NULL;
  __pyx_t_3.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 15, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode))))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->right_node);
  __Pyx_DECREF(((PyObject *)__pyx_v___pyx_result->right_node));
  __pyx_v___pyx_result->right_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_Uint64ClosedLeftIntervalNode__set_state(Uint64ClosedLeftIntervalNode __pyx_result, tuple __pyx_state):
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[16])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(2, 13, __pyx_L1_error)
  }
  __pyx_t_7 = PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_8 = ((__pyx_t_7 > 16) != 0);
  if (__pyx_t_8) {
  } else {
    __pyx_t_4 = __pyx_t_8;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_8 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_8 == ((int)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_9 = (__pyx_t_8 != 0);
  __pyx_t_4 = __pyx_t_9;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_4) {

    /* "(tree fragment)":14
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[16])             # <<<<<<<<<<<<<<
 */
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_10)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_update); if (unlikely(!__pyx_t_11)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(2, 14, __pyx_L1_error)
    }
    __pyx_t_10 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 16, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_10)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_12 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_11))) {
      __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_11);
      if (likely(__pyx_t_12)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
        __Pyx_INCREF(__pyx_t_12);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_11, function);
      }
    }
    __pyx_t_1 = (__pyx_t_12) ? __Pyx_PyObject_Call2Args(__pyx_t_11, __pyx_t_12, __pyx_t_10) : __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_t_10);
    __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_Uint64ClosedLeftIntervalNode__set_state(Uint64ClosedLeftIntervalNode __pyx_result, tuple __pyx_state):
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[16])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_Uint64ClosedLeftIntervalNode__set_state(<Uint64ClosedLeftIntervalNode> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Uint64ClosedLeftIntervalNode__set_state(Uint64ClosedLeftIntervalNode __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_2, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_3, 1);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("pandas._libs.interval.__pyx_unpickle_Uint64ClosedLeftIntervalNode__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_Uint64ClosedRightIntervalNode(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_41__pyx_unpickle_Uint64ClosedRightIntervalNode(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pandas_5_libs_8interval_41__pyx_unpickle_Uint64ClosedRightIntervalNode = {"__pyx_unpickle_Uint64ClosedRightIntervalNode", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_41__pyx_unpickle_Uint64ClosedRightIntervalNode, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_41__pyx_unpickle_Uint64ClosedRightIntervalNode(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_Uint64ClosedRightIntervalNode (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_type)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_checksum)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Uint64ClosedRightIntervalNode", 1, 3, 3, 1); __PYX_ERR(2, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Uint64ClosedRightIntervalNode", 1, 3, 3, 2); __PYX_ERR(2, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_unpickle_Uint64ClosedRightIntervalNode") < 0)) __PYX_ERR(2, 1, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Uint64ClosedRightIntervalNode", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(2, 1, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.__pyx_unpickle_Uint64ClosedRightIntervalNode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_40__pyx_unpickle_Uint64ClosedRightIntervalNode(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_40__pyx_unpickle_Uint64ClosedRightIntervalNode(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Uint64ClosedRightIntervalNode", 0);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x7a104e5:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)
 */
  __pyx_t_1 = ((__pyx_v___pyx_checksum != 0x7a104e5) != 0);
  if (__pyx_t_1) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x7a104e5:
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)
 *     __pyx_result = Uint64ClosedRightIntervalNode.__new__(__pyx_type)
 */
    __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_PickleError);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_2);
    __pyx_v___pyx_PickleError = __pyx_t_2;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum != 0x7a104e5:
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)             # <<<<<<<<<<<<<<
 *     __pyx_result = Uint64ClosedRightIntervalNode.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_2 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_s_vs_0x7a, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_v___pyx_PickleError);
    __pyx_t_2 = __pyx_v___pyx_PickleError; __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_5, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(2, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x7a104e5:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)
 *     __pyx_result = Uint64ClosedRightIntervalNode.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Uint64ClosedRightIntervalNode__set_state(<Uint64ClosedRightIntervalNode> __pyx_result, __pyx_state)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode), __pyx_n_s_new); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_v___pyx_type) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v___pyx_type);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v___pyx_result = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)
 *     __pyx_result = Uint64ClosedRightIntervalNode.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Uint64ClosedRightIntervalNode__set_state(<Uint64ClosedRightIntervalNode> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_1 = (__pyx_v___pyx_state != Py_None);
  __pyx_t_6 = (__pyx_t_1 != 0);
  if (__pyx_t_6) {

    /* "(tree fragment)":9
 *     __pyx_result = Uint64ClosedRightIntervalNode.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Uint64ClosedRightIntervalNode__set_state(<Uint64ClosedRightIntervalNode> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_Uint64ClosedRightIntervalNode__set_state(Uint64ClosedRightIntervalNode __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(2, 9, __pyx_L1_error)
    __pyx_t_3 = __pyx_f_6pandas_5_libs_8interval___pyx_unpickle_Uint64ClosedRightIntervalNode__set_state(((struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)
 *     __pyx_result = Uint64ClosedRightIntervalNode.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Uint64ClosedRightIntervalNode__set_state(<Uint64ClosedRightIntervalNode> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Uint64ClosedRightIntervalNode__set_state(<Uint64ClosedRightIntervalNode> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_Uint64ClosedRightIntervalNode__set_state(Uint64ClosedRightIntervalNode __pyx_result, tuple __pyx_state):
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_Uint64ClosedRightIntervalNode(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pandas._libs.interval.__pyx_unpickle_Uint64ClosedRightIntervalNode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_Uint64ClosedRightIntervalNode__set_state(<Uint64ClosedRightIntervalNode> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Uint64ClosedRightIntervalNode__set_state(Uint64ClosedRightIntervalNode __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_6pandas_5_libs_8interval___pyx_unpickle_Uint64ClosedRightIntervalNode__set_state(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_memviewslice __pyx_t_2 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_3 = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_t_4;
  __pyx_t_5numpy_int64_t __pyx_t_5;
  __pyx_t_5numpy_uint64_t __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Uint64ClosedRightIntervalNode__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_Uint64ClosedRightIntervalNode__set_state(Uint64ClosedRightIntervalNode __pyx_result, tuple __pyx_state):
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[16])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_2.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->center_left_indices, 0);
  __pyx_v___pyx_result->center_left_indices = __pyx_t_2;
  __pyx_t_2.memview = NULL;
  __pyx_t_2.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_uint64_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_3.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->center_left_values, 0);
  __pyx_v___pyx_result->center_left_values = __pyx_t_3;
  __pyx_t_3.memview = NULL;
  __pyx_t_3.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_2.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->center_right_indices, 0);
  __pyx_v___pyx_result->center_right_indices = __pyx_t_2;
  __pyx_t_2.memview = NULL;
  __pyx_t_2.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_uint64_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_3.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->center_right_values, 0);
  __pyx_v___pyx_result->center_right_values = __pyx_t_3;
  __pyx_t_3.memview = NULL;
  __pyx_t_3.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 4, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_2.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->indices, 0);
  __pyx_v___pyx_result->indices = __pyx_t_2;
  __pyx_t_2.memview = NULL;
  __pyx_t_2.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 5, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->is_leaf_node = __pyx_t_4;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 6, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyInt_As_npy_int64(__pyx_t_1); if (unlikely((__pyx_t_5 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->leaf_size = __pyx_t_5;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 7, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_uint64_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_3.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->left, 0);
  __pyx_v___pyx_result->left = __pyx_t_3;
  __pyx_t_3.memview = NULL;
  __pyx_t_3.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 8, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode))))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->left_node);
  __Pyx_DECREF(((PyObject *)__pyx_v___pyx_result->left_node));
  __pyx_v___pyx_result->left_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 9, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = __Pyx_PyInt_As_npy_uint64(__pyx_t_1); if (unlikely((__pyx_t_6 == ((npy_uint64)-1)) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->max_right = __pyx_t_6;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 10, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = __Pyx_PyInt_As_npy_uint64(__pyx_t_1); if (unlikely((__pyx_t_6 == ((npy_uint64)-1)) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->min_left = __pyx_t_6;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 11, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyInt_As_npy_int64(__pyx_t_1); if (unlikely((__pyx_t_5 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->n_center = __pyx_t_5;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 12, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyInt_As_npy_int64(__pyx_t_1); if (unlikely((__pyx_t_5 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->n_elements = __pyx_t_5;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 13, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = __Pyx_PyInt_As_npy_uint64(__pyx_t_1); if (unlikely((__pyx_t_6 == ((npy_uint64)-1)) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->pivot = __pyx_t_6;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 14, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_uint64_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_3.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->right, 0);
  __pyx_v___pyx_result->right = __pyx_t_3;
  __pyx_t_3.memview = NULL;
  __pyx_t_3.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 15, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode))))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->right_node);
  __Pyx_DECREF(((PyObject *)__pyx_v___pyx_result->right_node));
  __pyx_v___pyx_result->right_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_Uint64ClosedRightIntervalNode__set_state(Uint64ClosedRightIntervalNode __pyx_result, tuple __pyx_state):
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[16])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(2, 13, __pyx_L1_error)
  }
  __pyx_t_7 = PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_8 = ((__pyx_t_7 > 16) != 0);
  if (__pyx_t_8) {
  } else {
    __pyx_t_4 = __pyx_t_8;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_8 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_8 == ((int)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_9 = (__pyx_t_8 != 0);
  __pyx_t_4 = __pyx_t_9;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_4) {

    /* "(tree fragment)":14
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[16])             # <<<<<<<<<<<<<<
 */
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_10)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_update); if (unlikely(!__pyx_t_11)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(2, 14, __pyx_L1_error)
    }
    __pyx_t_10 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 16, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_10)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_12 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_11))) {
      __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_11);
      if (likely(__pyx_t_12)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
        __Pyx_INCREF(__pyx_t_12);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_11, function);
      }
    }
    __pyx_t_1 = (__pyx_t_12) ? __Pyx_PyObject_Call2Args(__pyx_t_11, __pyx_t_12, __pyx_t_10) : __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_t_10);
    __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_Uint64ClosedRightIntervalNode__set_state(Uint64ClosedRightIntervalNode __pyx_result, tuple __pyx_state):
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[16])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_Uint64ClosedRightIntervalNode__set_state(<Uint64ClosedRightIntervalNode> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Uint64ClosedRightIntervalNode__set_state(Uint64ClosedRightIntervalNode __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_2, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_3, 1);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("pandas._libs.interval.__pyx_unpickle_Uint64ClosedRightIntervalNode__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_Uint64ClosedBothIntervalNode(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_43__pyx_unpickle_Uint64ClosedBothIntervalNode(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pandas_5_libs_8interval_43__pyx_unpickle_Uint64ClosedBothIntervalNode = {"__pyx_unpickle_Uint64ClosedBothIntervalNode", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_43__pyx_unpickle_Uint64ClosedBothIntervalNode, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_43__pyx_unpickle_Uint64ClosedBothIntervalNode(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_Uint64ClosedBothIntervalNode (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_type)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_checksum)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Uint64ClosedBothIntervalNode", 1, 3, 3, 1); __PYX_ERR(2, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Uint64ClosedBothIntervalNode", 1, 3, 3, 2); __PYX_ERR(2, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_unpickle_Uint64ClosedBothIntervalNode") < 0)) __PYX_ERR(2, 1, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Uint64ClosedBothIntervalNode", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(2, 1, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.__pyx_unpickle_Uint64ClosedBothIntervalNode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_42__pyx_unpickle_Uint64ClosedBothIntervalNode(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_42__pyx_unpickle_Uint64ClosedBothIntervalNode(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Uint64ClosedBothIntervalNode", 0);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x7a104e5:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)
 */
  __pyx_t_1 = ((__pyx_v___pyx_checksum != 0x7a104e5) != 0);
  if (__pyx_t_1) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x7a104e5:
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)
 *     __pyx_result = Uint64ClosedBothIntervalNode.__new__(__pyx_type)
 */
    __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_PickleError);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_2);
    __pyx_v___pyx_PickleError = __pyx_t_2;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum != 0x7a104e5:
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)             # <<<<<<<<<<<<<<
 *     __pyx_result = Uint64ClosedBothIntervalNode.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_2 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_s_vs_0x7a, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_v___pyx_PickleError);
    __pyx_t_2 = __pyx_v___pyx_PickleError; __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_5, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(2, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x7a104e5:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)
 *     __pyx_result = Uint64ClosedBothIntervalNode.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Uint64ClosedBothIntervalNode__set_state(<Uint64ClosedBothIntervalNode> __pyx_result, __pyx_state)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode), __pyx_n_s_new); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_v___pyx_type) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v___pyx_type);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v___pyx_result = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)
 *     __pyx_result = Uint64ClosedBothIntervalNode.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Uint64ClosedBothIntervalNode__set_state(<Uint64ClosedBothIntervalNode> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_1 = (__pyx_v___pyx_state != Py_None);
  __pyx_t_6 = (__pyx_t_1 != 0);
  if (__pyx_t_6) {

    /* "(tree fragment)":9
 *     __pyx_result = Uint64ClosedBothIntervalNode.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Uint64ClosedBothIntervalNode__set_state(<Uint64ClosedBothIntervalNode> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_Uint64ClosedBothIntervalNode__set_state(Uint64ClosedBothIntervalNode __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(2, 9, __pyx_L1_error)
    __pyx_t_3 = __pyx_f_6pandas_5_libs_8interval___pyx_unpickle_Uint64ClosedBothIntervalNode__set_state(((struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)
 *     __pyx_result = Uint64ClosedBothIntervalNode.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Uint64ClosedBothIntervalNode__set_state(<Uint64ClosedBothIntervalNode> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Uint64ClosedBothIntervalNode__set_state(<Uint64ClosedBothIntervalNode> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_Uint64ClosedBothIntervalNode__set_state(Uint64ClosedBothIntervalNode __pyx_result, tuple __pyx_state):
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_Uint64ClosedBothIntervalNode(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pandas._libs.interval.__pyx_unpickle_Uint64ClosedBothIntervalNode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_Uint64ClosedBothIntervalNode__set_state(<Uint64ClosedBothIntervalNode> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Uint64ClosedBothIntervalNode__set_state(Uint64ClosedBothIntervalNode __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_6pandas_5_libs_8interval___pyx_unpickle_Uint64ClosedBothIntervalNode__set_state(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_memviewslice __pyx_t_2 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_3 = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_t_4;
  __pyx_t_5numpy_int64_t __pyx_t_5;
  __pyx_t_5numpy_uint64_t __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Uint64ClosedBothIntervalNode__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_Uint64ClosedBothIntervalNode__set_state(Uint64ClosedBothIntervalNode __pyx_result, tuple __pyx_state):
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[16])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_2.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->center_left_indices, 0);
  __pyx_v___pyx_result->center_left_indices = __pyx_t_2;
  __pyx_t_2.memview = NULL;
  __pyx_t_2.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_uint64_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_3.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->center_left_values, 0);
  __pyx_v___pyx_result->center_left_values = __pyx_t_3;
  __pyx_t_3.memview = NULL;
  __pyx_t_3.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_2.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->center_right_indices, 0);
  __pyx_v___pyx_result->center_right_indices = __pyx_t_2;
  __pyx_t_2.memview = NULL;
  __pyx_t_2.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_uint64_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_3.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->center_right_values, 0);
  __pyx_v___pyx_result->center_right_values = __pyx_t_3;
  __pyx_t_3.memview = NULL;
  __pyx_t_3.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 4, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_2.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->indices, 0);
  __pyx_v___pyx_result->indices = __pyx_t_2;
  __pyx_t_2.memview = NULL;
  __pyx_t_2.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 5, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->is_leaf_node = __pyx_t_4;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 6, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyInt_As_npy_int64(__pyx_t_1); if (unlikely((__pyx_t_5 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->leaf_size = __pyx_t_5;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 7, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_uint64_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_3.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->left, 0);
  __pyx_v___pyx_result->left = __pyx_t_3;
  __pyx_t_3.memview = NULL;
  __pyx_t_3.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 8, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode))))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->left_node);
  __Pyx_DECREF(((PyObject *)__pyx_v___pyx_result->left_node));
  __pyx_v___pyx_result->left_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 9, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = __Pyx_PyInt_As_npy_uint64(__pyx_t_1); if (unlikely((__pyx_t_6 == ((npy_uint64)-1)) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->max_right = __pyx_t_6;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 10, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = __Pyx_PyInt_As_npy_uint64(__pyx_t_1); if (unlikely((__pyx_t_6 == ((npy_uint64)-1)) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->min_left = __pyx_t_6;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 11, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyInt_As_npy_int64(__pyx_t_1); if (unlikely((__pyx_t_5 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->n_center = __pyx_t_5;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 12, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyInt_As_npy_int64(__pyx_t_1); if (unlikely((__pyx_t_5 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->n_elements = __pyx_t_5;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 13, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = __Pyx_PyInt_As_npy_uint64(__pyx_t_1); if (unlikely((__pyx_t_6 == ((npy_uint64)-1)) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->pivot = __pyx_t_6;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 14, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_uint64_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_3.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->right, 0);
  __pyx_v___pyx_result->right = __pyx_t_3;
  __pyx_t_3.memview = NULL;
  __pyx_t_3.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 15, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode))))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->right_node);
  __Pyx_DECREF(((PyObject *)__pyx_v___pyx_result->right_node));
  __pyx_v___pyx_result->right_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_Uint64ClosedBothIntervalNode__set_state(Uint64ClosedBothIntervalNode __pyx_result, tuple __pyx_state):
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[16])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(2, 13, __pyx_L1_error)
  }
  __pyx_t_7 = PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_8 = ((__pyx_t_7 > 16) != 0);
  if (__pyx_t_8) {
  } else {
    __pyx_t_4 = __pyx_t_8;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_8 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_8 == ((int)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_9 = (__pyx_t_8 != 0);
  __pyx_t_4 = __pyx_t_9;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_4) {

    /* "(tree fragment)":14
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[16])             # <<<<<<<<<<<<<<
 */
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_10)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_update); if (unlikely(!__pyx_t_11)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(2, 14, __pyx_L1_error)
    }
    __pyx_t_10 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 16, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_10)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_12 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_11))) {
      __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_11);
      if (likely(__pyx_t_12)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
        __Pyx_INCREF(__pyx_t_12);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_11, function);
      }
    }
    __pyx_t_1 = (__pyx_t_12) ? __Pyx_PyObject_Call2Args(__pyx_t_11, __pyx_t_12, __pyx_t_10) : __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_t_10);
    __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_Uint64ClosedBothIntervalNode__set_state(Uint64ClosedBothIntervalNode __pyx_result, tuple __pyx_state):
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[16])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_Uint64ClosedBothIntervalNode__set_state(<Uint64ClosedBothIntervalNode> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Uint64ClosedBothIntervalNode__set_state(Uint64ClosedBothIntervalNode __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_2, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_3, 1);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("pandas._libs.interval.__pyx_unpickle_Uint64ClosedBothIntervalNode__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_Uint64ClosedNeitherIntervalNode(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pandas_5_libs_8interval_45__pyx_unpickle_Uint64ClosedNeitherIntervalNode(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pandas_5_libs_8interval_45__pyx_unpickle_Uint64ClosedNeitherIntervalNode = {"__pyx_unpickle_Uint64ClosedNeitherIntervalNode", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_45__pyx_unpickle_Uint64ClosedNeitherIntervalNode, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pandas_5_libs_8interval_45__pyx_unpickle_Uint64ClosedNeitherIntervalNode(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_Uint64ClosedNeitherIntervalNode (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_type)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_checksum)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Uint64ClosedNeitherIntervalNode", 1, 3, 3, 1); __PYX_ERR(2, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Uint64ClosedNeitherIntervalNode", 1, 3, 3, 2); __PYX_ERR(2, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_unpickle_Uint64ClosedNeitherIntervalNode") < 0)) __PYX_ERR(2, 1, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Uint64ClosedNeitherIntervalNode", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(2, 1, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pandas._libs.interval.__pyx_unpickle_Uint64ClosedNeitherIntervalNode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pandas_5_libs_8interval_44__pyx_unpickle_Uint64ClosedNeitherIntervalNode(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pandas_5_libs_8interval_44__pyx_unpickle_Uint64ClosedNeitherIntervalNode(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Uint64ClosedNeitherIntervalNode", 0);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x7a104e5:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)
 */
  __pyx_t_1 = ((__pyx_v___pyx_checksum != 0x7a104e5) != 0);
  if (__pyx_t_1) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x7a104e5:
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)
 *     __pyx_result = Uint64ClosedNeitherIntervalNode.__new__(__pyx_type)
 */
    __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_PickleError);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_2);
    __pyx_v___pyx_PickleError = __pyx_t_2;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum != 0x7a104e5:
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)             # <<<<<<<<<<<<<<
 *     __pyx_result = Uint64ClosedNeitherIntervalNode.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_2 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_s_vs_0x7a, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_v___pyx_PickleError);
    __pyx_t_2 = __pyx_v___pyx_PickleError; __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_5, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(2, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x7a104e5:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)
 *     __pyx_result = Uint64ClosedNeitherIntervalNode.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Uint64ClosedNeitherIntervalNode__set_state(<Uint64ClosedNeitherIntervalNode> __pyx_result, __pyx_state)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode), __pyx_n_s_new); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_v___pyx_type) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v___pyx_type);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v___pyx_result = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)
 *     __pyx_result = Uint64ClosedNeitherIntervalNode.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Uint64ClosedNeitherIntervalNode__set_state(<Uint64ClosedNeitherIntervalNode> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_1 = (__pyx_v___pyx_state != Py_None);
  __pyx_t_6 = (__pyx_t_1 != 0);
  if (__pyx_t_6) {

    /* "(tree fragment)":9
 *     __pyx_result = Uint64ClosedNeitherIntervalNode.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Uint64ClosedNeitherIntervalNode__set_state(<Uint64ClosedNeitherIntervalNode> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_Uint64ClosedNeitherIntervalNode__set_state(Uint64ClosedNeitherIntervalNode __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(2, 9, __pyx_L1_error)
    __pyx_t_3 = __pyx_f_6pandas_5_libs_8interval___pyx_unpickle_Uint64ClosedNeitherIntervalNode__set_state(((struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7a104e5 = (center_left_indices, center_left_values, center_right_indices, center_right_values, indices, is_leaf_node, leaf_size, left, left_node, max_right, min_left, n_center, n_elements, pivot, right, right_node))" % __pyx_checksum)
 *     __pyx_result = Uint64ClosedNeitherIntervalNode.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Uint64ClosedNeitherIntervalNode__set_state(<Uint64ClosedNeitherIntervalNode> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Uint64ClosedNeitherIntervalNode__set_state(<Uint64ClosedNeitherIntervalNode> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_Uint64ClosedNeitherIntervalNode__set_state(Uint64ClosedNeitherIntervalNode __pyx_result, tuple __pyx_state):
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_Uint64ClosedNeitherIntervalNode(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pandas._libs.interval.__pyx_unpickle_Uint64ClosedNeitherIntervalNode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_Uint64ClosedNeitherIntervalNode__set_state(<Uint64ClosedNeitherIntervalNode> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Uint64ClosedNeitherIntervalNode__set_state(Uint64ClosedNeitherIntervalNode __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_6pandas_5_libs_8interval___pyx_unpickle_Uint64ClosedNeitherIntervalNode__set_state(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_memviewslice __pyx_t_2 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_3 = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_t_4;
  __pyx_t_5numpy_int64_t __pyx_t_5;
  __pyx_t_5numpy_uint64_t __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Uint64ClosedNeitherIntervalNode__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_Uint64ClosedNeitherIntervalNode__set_state(Uint64ClosedNeitherIntervalNode __pyx_result, tuple __pyx_state):
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[16])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_2.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->center_left_indices, 0);
  __pyx_v___pyx_result->center_left_indices = __pyx_t_2;
  __pyx_t_2.memview = NULL;
  __pyx_t_2.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_uint64_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_3.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->center_left_values, 0);
  __pyx_v___pyx_result->center_left_values = __pyx_t_3;
  __pyx_t_3.memview = NULL;
  __pyx_t_3.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_2.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->center_right_indices, 0);
  __pyx_v___pyx_result->center_right_indices = __pyx_t_2;
  __pyx_t_2.memview = NULL;
  __pyx_t_2.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_uint64_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_3.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->center_right_values, 0);
  __pyx_v___pyx_result->center_right_values = __pyx_t_3;
  __pyx_t_3.memview = NULL;
  __pyx_t_3.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 4, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_2.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->indices, 0);
  __pyx_v___pyx_result->indices = __pyx_t_2;
  __pyx_t_2.memview = NULL;
  __pyx_t_2.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 5, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->is_leaf_node = __pyx_t_4;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 6, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyInt_As_npy_int64(__pyx_t_1); if (unlikely((__pyx_t_5 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->leaf_size = __pyx_t_5;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 7, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_uint64_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_3.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->left, 0);
  __pyx_v___pyx_result->left = __pyx_t_3;
  __pyx_t_3.memview = NULL;
  __pyx_t_3.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 8, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode))))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->left_node);
  __Pyx_DECREF(((PyObject *)__pyx_v___pyx_result->left_node));
  __pyx_v___pyx_result->left_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 9, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = __Pyx_PyInt_As_npy_uint64(__pyx_t_1); if (unlikely((__pyx_t_6 == ((npy_uint64)-1)) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->max_right = __pyx_t_6;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 10, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = __Pyx_PyInt_As_npy_uint64(__pyx_t_1); if (unlikely((__pyx_t_6 == ((npy_uint64)-1)) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->min_left = __pyx_t_6;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 11, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyInt_As_npy_int64(__pyx_t_1); if (unlikely((__pyx_t_5 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->n_center = __pyx_t_5;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 12, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyInt_As_npy_int64(__pyx_t_1); if (unlikely((__pyx_t_5 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->n_elements = __pyx_t_5;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 13, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = __Pyx_PyInt_As_npy_uint64(__pyx_t_1); if (unlikely((__pyx_t_6 == ((npy_uint64)-1)) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->pivot = __pyx_t_6;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 14, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_uint64_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_3.memview)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->right, 0);
  __pyx_v___pyx_result->right = __pyx_t_3;
  __pyx_t_3.memview = NULL;
  __pyx_t_3.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 15, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode))))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->right_node);
  __Pyx_DECREF(((PyObject *)__pyx_v___pyx_result->right_node));
  __pyx_v___pyx_result->right_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_Uint64ClosedNeitherIntervalNode__set_state(Uint64ClosedNeitherIntervalNode __pyx_result, tuple __pyx_state):
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[16])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(2, 13, __pyx_L1_error)
  }
  __pyx_t_7 = PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_8 = ((__pyx_t_7 > 16) != 0);
  if (__pyx_t_8) {
  } else {
    __pyx_t_4 = __pyx_t_8;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_8 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_8 == ((int)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_9 = (__pyx_t_8 != 0);
  __pyx_t_4 = __pyx_t_9;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_4) {

    /* "(tree fragment)":14
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[16])             # <<<<<<<<<<<<<<
 */
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_10)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_update); if (unlikely(!__pyx_t_11)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(2, 14, __pyx_L1_error)
    }
    __pyx_t_10 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 16, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_10)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_12 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_11))) {
      __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_11);
      if (likely(__pyx_t_12)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
        __Pyx_INCREF(__pyx_t_12);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_11, function);
      }
    }
    __pyx_t_1 = (__pyx_t_12) ? __Pyx_PyObject_Call2Args(__pyx_t_11, __pyx_t_12, __pyx_t_10) : __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_t_10);
    __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_Uint64ClosedNeitherIntervalNode__set_state(Uint64ClosedNeitherIntervalNode __pyx_result, tuple __pyx_state):
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[16])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_Uint64ClosedNeitherIntervalNode__set_state(<Uint64ClosedNeitherIntervalNode> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Uint64ClosedNeitherIntervalNode__set_state(Uint64ClosedNeitherIntervalNode __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_2, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_3, 1);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("pandas._libs.interval.__pyx_unpickle_Uint64ClosedNeitherIntervalNode__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":258
 *         # experimental exception made for __getbuffer__ and __releasebuffer__
 *         # -- the details of this may change.
 *         def __getbuffer__(ndarray self, Py_buffer* info, int flags):             # <<<<<<<<<<<<<<
 *             # This implementation of getbuffer is geared towards Cython
 *             # requirements, and does not yet fulfill the PEP.
 */

/* Python wrapper */
static CYTHON_UNUSED int __pyx_pw_5numpy_7ndarray_1__getbuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /*proto*/
static CYTHON_UNUSED int __pyx_pw_5numpy_7ndarray_1__getbuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getbuffer__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5numpy_7ndarray___getbuffer__(((PyArrayObject *)__pyx_v_self), ((Py_buffer *)__pyx_v_info), ((int)__pyx_v_flags));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5numpy_7ndarray___getbuffer__(PyArrayObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags) {
  int __pyx_v_i;
  int __pyx_v_ndim;
  int __pyx_v_endian_detector;
  int __pyx_v_little_endian;
  int __pyx_v_t;
  char *__pyx_v_f;
  PyArray_Descr *__pyx_v_descr = 0;
  int __pyx_v_offset;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  PyArray_Descr *__pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  char *__pyx_t_9;
  if (__pyx_v_info == NULL) {
    PyErr_SetString(PyExc_BufferError, "PyObject_GetBuffer: view==NULL argument is obsolete");
    return -1;
  }
  __Pyx_RefNannySetupContext("__getbuffer__", 0);
  __pyx_v_info->obj = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(__pyx_v_info->obj);

  /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":265
 * 
 *             cdef int i, ndim
 *             cdef int endian_detector = 1             # <<<<<<<<<<<<<<
 *             cdef bint little_endian = ((<char*>&endian_detector)[0] != 0)
 * 
 */
  __pyx_v_endian_detector = 1;

  /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":266
 *             cdef int i, ndim
 *             cdef int endian_detector = 1
 *             cdef bint little_endian = ((<char*>&endian_detector)[0] != 0)             # <<<<<<<<<<<<<<
 * 
 *             ndim = PyArray_NDIM(self)
 */
  __pyx_v_little_endian = ((((char *)(&__pyx_v_endian_detector))[0]) != 0);

  /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":268
 *             cdef bint little_endian = ((<char*>&endian_detector)[0] != 0)
 * 
 *             ndim = PyArray_NDIM(self)             # <<<<<<<<<<<<<<
 * 
 *             if ((flags & pybuf.PyBUF_C_CONTIGUOUS == pybuf.PyBUF_C_CONTIGUOUS)
 */
  __pyx_v_ndim = PyArray_NDIM(__pyx_v_self);

  /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":270
 *             ndim = PyArray_NDIM(self)
 * 
 *             if ((flags & pybuf.PyBUF_C_CONTIGUOUS == pybuf.PyBUF_C_CONTIGUOUS)             # <<<<<<<<<<<<<<
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_C_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not C contiguous")
 */
  __pyx_t_2 = (((__pyx_v_flags & PyBUF_C_CONTIGUOUS) == PyBUF_C_CONTIGUOUS) != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }

  /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":271
 * 
 *             if ((flags & pybuf.PyBUF_C_CONTIGUOUS == pybuf.PyBUF_C_CONTIGUOUS)
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_C_CONTIGUOUS)):             # <<<<<<<<<<<<<<
 *                 raise ValueError(u"ndarray is not C contiguous")
 * 
 */
  __pyx_t_2 = ((!(PyArray_CHKFLAGS(__pyx_v_self, NPY_ARRAY_C_CONTIGUOUS) != 0)) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;

  /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":270
 *             ndim = PyArray_NDIM(self)
 * 
 *             if ((flags & pybuf.PyBUF_C_CONTIGUOUS == pybuf.PyBUF_C_CONTIGUOUS)             # <<<<<<<<<<<<<<
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_C_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not C contiguous")
 */
  if (unlikely(__pyx_t_1)) {

    /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":272
 *             if ((flags & pybuf.PyBUF_C_CONTIGUOUS == pybuf.PyBUF_C_CONTIGUOUS)
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_C_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not C contiguous")             # <<<<<<<<<<<<<<
 * 
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__19, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 272, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(3, 272, __pyx_L1_error)

    /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":270
 *             ndim = PyArray_NDIM(self)
 * 
 *             if ((flags & pybuf.PyBUF_C_CONTIGUOUS == pybuf.PyBUF_C_CONTIGUOUS)             # <<<<<<<<<<<<<<
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_C_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not C contiguous")
 */
  }

  /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":274
 *                 raise ValueError(u"ndarray is not C contiguous")
 * 
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)             # <<<<<<<<<<<<<<
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_F_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not Fortran contiguous")
 */
  __pyx_t_2 = (((__pyx_v_flags & PyBUF_F_CONTIGUOUS) == PyBUF_F_CONTIGUOUS) != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L7_bool_binop_done;
  }

  /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":275
 * 
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_F_CONTIGUOUS)):             # <<<<<<<<<<<<<<
 *                 raise ValueError(u"ndarray is not Fortran contiguous")
 * 
 */
  __pyx_t_2 = ((!(PyArray_CHKFLAGS(__pyx_v_self, NPY_ARRAY_F_CONTIGUOUS) != 0)) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L7_bool_binop_done:;

  /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":274
 *                 raise ValueError(u"ndarray is not C contiguous")
 * 
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)             # <<<<<<<<<<<<<<
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_F_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not Fortran contiguous")
 */
  if (unlikely(__pyx_t_1)) {

    /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":276
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_F_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not Fortran contiguous")             # <<<<<<<<<<<<<<
 * 
 *             info.buf = PyArray_DATA(self)
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__20, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 276, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(3, 276, __pyx_L1_error)

    /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":274
 *                 raise ValueError(u"ndarray is not C contiguous")
 * 
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)             # <<<<<<<<<<<<<<
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_F_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not Fortran contiguous")
 */
  }

  /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":278
 *                 raise ValueError(u"ndarray is not Fortran contiguous")
 * 
 *             info.buf = PyArray_DATA(self)             # <<<<<<<<<<<<<<
 *             info.ndim = ndim
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):
 */
  __pyx_v_info->buf = PyArray_DATA(__pyx_v_self);

  /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":279
 * 
 *             info.buf = PyArray_DATA(self)
 *             info.ndim = ndim             # <<<<<<<<<<<<<<
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):
 *                 # Allocate new buffer for strides and shape info.
 */
  __pyx_v_info->ndim = __pyx_v_ndim;

  /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":280
 *             info.buf = PyArray_DATA(self)
 *             info.ndim = ndim
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):             # <<<<<<<<<<<<<<
 *                 # Allocate new buffer for strides and shape info.
 *                 # This is allocated as one block, strides first.
 */
  __pyx_t_1 = (((sizeof(npy_intp)) != (sizeof(Py_ssize_t))) != 0);
  if (__pyx_t_1) {

    /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":283
 *                 # Allocate new buffer for strides and shape info.
 *                 # This is allocated as one block, strides first.
 *                 info.strides = <Py_ssize_t*>PyObject_Malloc(sizeof(Py_ssize_t) * 2 * <size_t>ndim)             # <<<<<<<<<<<<<<
 *                 info.shape = info.strides + ndim
 *                 for i in range(ndim):
 */
    __pyx_v_info->strides = ((Py_ssize_t *)PyObject_Malloc((((sizeof(Py_ssize_t)) * 2) * ((size_t)__pyx_v_ndim))));

    /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":284
 *                 # This is allocated as one block, strides first.
 *                 info.strides = <Py_ssize_t*>PyObject_Malloc(sizeof(Py_ssize_t) * 2 * <size_t>ndim)
 *                 info.shape = info.strides + ndim             # <<<<<<<<<<<<<<
 *                 for i in range(ndim):
 *                     info.strides[i] = PyArray_STRIDES(self)[i]
 */
    __pyx_v_info->shape = (__pyx_v_info->strides + __pyx_v_ndim);

    /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":285
 *                 info.strides = <Py_ssize_t*>PyObject_Malloc(sizeof(Py_ssize_t) * 2 * <size_t>ndim)
 *                 info.shape = info.strides + ndim
 *                 for i in range(ndim):             # <<<<<<<<<<<<<<
 *                     info.strides[i] = PyArray_STRIDES(self)[i]
 *                     info.shape[i] = PyArray_DIMS(self)[i]
 */
    __pyx_t_4 = __pyx_v_ndim;
    __pyx_t_5 = __pyx_t_4;
    for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
      __pyx_v_i = __pyx_t_6;

      /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":286
 *                 info.shape = info.strides + ndim
 *                 for i in range(ndim):
 *                     info.strides[i] = PyArray_STRIDES(self)[i]             # <<<<<<<<<<<<<<
 *                     info.shape[i] = PyArray_DIMS(self)[i]
 *             else:
 */
      (__pyx_v_info->strides[__pyx_v_i]) = (PyArray_STRIDES(__pyx_v_self)[__pyx_v_i]);

      /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":287
 *                 for i in range(ndim):
 *                     info.strides[i] = PyArray_STRIDES(self)[i]
 *                     info.shape[i] = PyArray_DIMS(self)[i]             # <<<<<<<<<<<<<<
 *             else:
 *                 info.strides = <Py_ssize_t*>PyArray_STRIDES(self)
 */
      (__pyx_v_info->shape[__pyx_v_i]) = (PyArray_DIMS(__pyx_v_self)[__pyx_v_i]);
    }

    /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":280
 *             info.buf = PyArray_DATA(self)
 *             info.ndim = ndim
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):             # <<<<<<<<<<<<<<
 *                 # Allocate new buffer for strides and shape info.
 *                 # This is allocated as one block, strides first.
 */
    goto __pyx_L9;
  }

  /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":289
 *                     info.shape[i] = PyArray_DIMS(self)[i]
 *             else:
 *                 info.strides = <Py_ssize_t*>PyArray_STRIDES(self)             # <<<<<<<<<<<<<<
 *                 info.shape = <Py_ssize_t*>PyArray_DIMS(self)
 *             info.suboffsets = NULL
 */
  /*else*/ {
    __pyx_v_info->strides = ((Py_ssize_t *)PyArray_STRIDES(__pyx_v_self));

    /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":290
 *             else:
 *                 info.strides = <Py_ssize_t*>PyArray_STRIDES(self)
 *                 info.shape = <Py_ssize_t*>PyArray_DIMS(self)             # <<<<<<<<<<<<<<
 *             info.suboffsets = NULL
 *             info.itemsize = PyArray_ITEMSIZE(self)
 */
    __pyx_v_info->shape = ((Py_ssize_t *)PyArray_DIMS(__pyx_v_self));
  }
  __pyx_L9:;

  /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":291
 *                 info.strides = <Py_ssize_t*>PyArray_STRIDES(self)
 *                 info.shape = <Py_ssize_t*>PyArray_DIMS(self)
 *             info.suboffsets = NULL             # <<<<<<<<<<<<<<
 *             info.itemsize = PyArray_ITEMSIZE(self)
 *             info.readonly = not PyArray_ISWRITEABLE(self)
 */
  __pyx_v_info->suboffsets = NULL;

  /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":292
 *                 info.shape = <Py_ssize_t*>PyArray_DIMS(self)
 *             info.suboffsets = NULL
 *             info.itemsize = PyArray_ITEMSIZE(self)             # <<<<<<<<<<<<<<
 *             info.readonly = not PyArray_ISWRITEABLE(self)
 * 
 */
  __pyx_v_info->itemsize = PyArray_ITEMSIZE(__pyx_v_self);

  /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":293
 *             info.suboffsets = NULL
 *             info.itemsize = PyArray_ITEMSIZE(self)
 *             info.readonly = not PyArray_ISWRITEABLE(self)             # <<<<<<<<<<<<<<
 * 
 *             cdef int t
 */
  __pyx_v_info->readonly = (!(PyArray_ISWRITEABLE(__pyx_v_self) != 0));

  /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":296
 * 
 *             cdef int t
 *             cdef char* f = NULL             # <<<<<<<<<<<<<<
 *             cdef dtype descr = <dtype>PyArray_DESCR(self)
 *             cdef int offset
 */
  __pyx_v_f = NULL;

  /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":297
 *             cdef int t
 *             cdef char* f = NULL
 *             cdef dtype descr = <dtype>PyArray_DESCR(self)             # <<<<<<<<<<<<<<
 *             cdef int offset
 * 
 */
  __pyx_t_7 = PyArray_DESCR(__pyx_v_self);
  __pyx_t_3 = ((PyObject *)__pyx_t_7);
  __Pyx_INCREF(__pyx_t_3);
  __pyx_v_descr = ((PyArray_Descr *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":300
 *             cdef int offset
 * 
 *             info.obj = self             # <<<<<<<<<<<<<<
 * 
 *             if not PyDataType_HASFIELDS(descr):
 */
  __Pyx_INCREF(((PyObject *)__pyx_v_self));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
  __Pyx_GOTREF(__pyx_v_info->obj);
  __Pyx_DECREF(__pyx_v_info->obj);
  __pyx_v_info->obj = ((PyObject *)__pyx_v_self);

  /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":302
 *             info.obj = self
 * 
 *             if not PyDataType_HASFIELDS(descr):             # <<<<<<<<<<<<<<
 *                 t = descr.type_num
 *                 if ((descr.byteorder == c'>' and little_endian) or
 */
  __pyx_t_1 = ((!(PyDataType_HASFIELDS(__pyx_v_descr) != 0)) != 0);
  if (__pyx_t_1) {

    /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":303
 * 
 *             if not PyDataType_HASFIELDS(descr):
 *                 t = descr.type_num             # <<<<<<<<<<<<<<
 *                 if ((descr.byteorder == c'>' and little_endian) or
 *                     (descr.byteorder == c'<' and not little_endian)):
 */
    __pyx_t_4 = __pyx_v_descr->type_num;
    __pyx_v_t = __pyx_t_4;

    /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":304
 *             if not PyDataType_HASFIELDS(descr):
 *                 t = descr.type_num
 *                 if ((descr.byteorder == c'>' and little_endian) or             # <<<<<<<<<<<<<<
 *                     (descr.byteorder == c'<' and not little_endian)):
 *                     raise ValueError(u"Non-native byte order not supported")
 */
    __pyx_t_2 = ((__pyx_v_descr->byteorder == '>') != 0);
    if (!__pyx_t_2) {
      goto __pyx_L15_next_or;
    } else {
    }
    __pyx_t_2 = (__pyx_v_little_endian != 0);
    if (!__pyx_t_2) {
    } else {
      __pyx_t_1 = __pyx_t_2;
      goto __pyx_L14_bool_binop_done;
    }
    __pyx_L15_next_or:;

    /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":305
 *                 t = descr.type_num
 *                 if ((descr.byteorder == c'>' and little_endian) or
 *                     (descr.byteorder == c'<' and not little_endian)):             # <<<<<<<<<<<<<<
 *                     raise ValueError(u"Non-native byte order not supported")
 *                 if   t == NPY_BYTE:        f = "b"
 */
    __pyx_t_2 = ((__pyx_v_descr->byteorder == '<') != 0);
    if (__pyx_t_2) {
    } else {
      __pyx_t_1 = __pyx_t_2;
      goto __pyx_L14_bool_binop_done;
    }
    __pyx_t_2 = ((!(__pyx_v_little_endian != 0)) != 0);
    __pyx_t_1 = __pyx_t_2;
    __pyx_L14_bool_binop_done:;

    /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":304
 *             if not PyDataType_HASFIELDS(descr):
 *                 t = descr.type_num
 *                 if ((descr.byteorder == c'>' and little_endian) or             # <<<<<<<<<<<<<<
 *                     (descr.byteorder == c'<' and not little_endian)):
 *                     raise ValueError(u"Non-native byte order not supported")
 */
    if (unlikely(__pyx_t_1)) {

      /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":306
 *                 if ((descr.byteorder == c'>' and little_endian) or
 *                     (descr.byteorder == c'<' and not little_endian)):
 *                     raise ValueError(u"Non-native byte order not supported")             # <<<<<<<<<<<<<<
 *                 if   t == NPY_BYTE:        f = "b"
 *                 elif t == NPY_UBYTE:       f = "B"
 */
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__21, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 306, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(3, 306, __pyx_L1_error)

      /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":304
 *             if not PyDataType_HASFIELDS(descr):
 *                 t = descr.type_num
 *                 if ((descr.byteorder == c'>' and little_endian) or             # <<<<<<<<<<<<<<
 *                     (descr.byteorder == c'<' and not little_endian)):
 *                     raise ValueError(u"Non-native byte order not supported")
 */
    }

    /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":307
 *                     (descr.byteorder == c'<' and not little_endian)):
 *                     raise ValueError(u"Non-native byte order not supported")
 *                 if   t == NPY_BYTE:        f = "b"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_UBYTE:       f = "B"
 *                 elif t == NPY_SHORT:       f = "h"
 */
    switch (__pyx_v_t) {
      case NPY_BYTE:
      __pyx_v_f = ((char *)"b");
      break;
      case NPY_UBYTE:

      /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":308
 *                     raise ValueError(u"Non-native byte order not supported")
 *                 if   t == NPY_BYTE:        f = "b"
 *                 elif t == NPY_UBYTE:       f = "B"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_SHORT:       f = "h"
 *                 elif t == NPY_USHORT:      f = "H"
 */
      __pyx_v_f = ((char *)"B");
      break;
      case NPY_SHORT:

      /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":309
 *                 if   t == NPY_BYTE:        f = "b"
 *                 elif t == NPY_UBYTE:       f = "B"
 *                 elif t == NPY_SHORT:       f = "h"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_USHORT:      f = "H"
 *                 elif t == NPY_INT:         f = "i"
 */
      __pyx_v_f = ((char *)"h");
      break;
      case NPY_USHORT:

      /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":310
 *                 elif t == NPY_UBYTE:       f = "B"
 *                 elif t == NPY_SHORT:       f = "h"
 *                 elif t == NPY_USHORT:      f = "H"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_INT:         f = "i"
 *                 elif t == NPY_UINT:        f = "I"
 */
      __pyx_v_f = ((char *)"H");
      break;
      case NPY_INT:

      /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":311
 *                 elif t == NPY_SHORT:       f = "h"
 *                 elif t == NPY_USHORT:      f = "H"
 *                 elif t == NPY_INT:         f = "i"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_UINT:        f = "I"
 *                 elif t == NPY_LONG:        f = "l"
 */
      __pyx_v_f = ((char *)"i");
      break;
      case NPY_UINT:

      /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":312
 *                 elif t == NPY_USHORT:      f = "H"
 *                 elif t == NPY_INT:         f = "i"
 *                 elif t == NPY_UINT:        f = "I"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_LONG:        f = "l"
 *                 elif t == NPY_ULONG:       f = "L"
 */
      __pyx_v_f = ((char *)"I");
      break;
      case NPY_LONG:

      /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":313
 *                 elif t == NPY_INT:         f = "i"
 *                 elif t == NPY_UINT:        f = "I"
 *                 elif t == NPY_LONG:        f = "l"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_ULONG:       f = "L"
 *                 elif t == NPY_LONGLONG:    f = "q"
 */
      __pyx_v_f = ((char *)"l");
      break;
      case NPY_ULONG:

      /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":314
 *                 elif t == NPY_UINT:        f = "I"
 *                 elif t == NPY_LONG:        f = "l"
 *                 elif t == NPY_ULONG:       f = "L"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_LONGLONG:    f = "q"
 *                 elif t == NPY_ULONGLONG:   f = "Q"
 */
      __pyx_v_f = ((char *)"L");
      break;
      case NPY_LONGLONG:

      /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":315
 *                 elif t == NPY_LONG:        f = "l"
 *                 elif t == NPY_ULONG:       f = "L"
 *                 elif t == NPY_LONGLONG:    f = "q"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_ULONGLONG:   f = "Q"
 *                 elif t == NPY_FLOAT:       f = "f"
 */
      __pyx_v_f = ((char *)"q");
      break;
      case NPY_ULONGLONG:

      /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":316
 *                 elif t == NPY_ULONG:       f = "L"
 *                 elif t == NPY_LONGLONG:    f = "q"
 *                 elif t == NPY_ULONGLONG:   f = "Q"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_FLOAT:       f = "f"
 *                 elif t == NPY_DOUBLE:      f = "d"
 */
      __pyx_v_f = ((char *)"Q");
      break;
      case NPY_FLOAT:

      /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":317
 *                 elif t == NPY_LONGLONG:    f = "q"
 *                 elif t == NPY_ULONGLONG:   f = "Q"
 *                 elif t == NPY_FLOAT:       f = "f"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_DOUBLE:      f = "d"
 *                 elif t == NPY_LONGDOUBLE:  f = "g"
 */
      __pyx_v_f = ((char *)"f");
      break;
      case NPY_DOUBLE:

      /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":318
 *                 elif t == NPY_ULONGLONG:   f = "Q"
 *                 elif t == NPY_FLOAT:       f = "f"
 *                 elif t == NPY_DOUBLE:      f = "d"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_LONGDOUBLE:  f = "g"
 *                 elif t == NPY_CFLOAT:      f = "Zf"
 */
      __pyx_v_f = ((char *)"d");
      break;
      case NPY_LONGDOUBLE:

      /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":319
 *                 elif t == NPY_FLOAT:       f = "f"
 *                 elif t == NPY_DOUBLE:      f = "d"
 *                 elif t == NPY_LONGDOUBLE:  f = "g"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_CFLOAT:      f = "Zf"
 *                 elif t == NPY_CDOUBLE:     f = "Zd"
 */
      __pyx_v_f = ((char *)"g");
      break;
      case NPY_CFLOAT:

      /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":320
 *                 elif t == NPY_DOUBLE:      f = "d"
 *                 elif t == NPY_LONGDOUBLE:  f = "g"
 *                 elif t == NPY_CFLOAT:      f = "Zf"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_CDOUBLE:     f = "Zd"
 *                 elif t == NPY_CLONGDOUBLE: f = "Zg"
 */
      __pyx_v_f = ((char *)"Zf");
      break;
      case NPY_CDOUBLE:

      /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":321
 *                 elif t == NPY_LONGDOUBLE:  f = "g"
 *                 elif t == NPY_CFLOAT:      f = "Zf"
 *                 elif t == NPY_CDOUBLE:     f = "Zd"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_CLONGDOUBLE: f = "Zg"
 *                 elif t == NPY_OBJECT:      f = "O"
 */
      __pyx_v_f = ((char *)"Zd");
      break;
      case NPY_CLONGDOUBLE:

      /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":322
 *                 elif t == NPY_CFLOAT:      f = "Zf"
 *                 elif t == NPY_CDOUBLE:     f = "Zd"
 *                 elif t == NPY_CLONGDOUBLE: f = "Zg"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_OBJECT:      f = "O"
 *                 else:
 */
      __pyx_v_f = ((char *)"Zg");
      break;
      case NPY_OBJECT:

      /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":323
 *                 elif t == NPY_CDOUBLE:     f = "Zd"
 *                 elif t == NPY_CLONGDOUBLE: f = "Zg"
 *                 elif t == NPY_OBJECT:      f = "O"             # <<<<<<<<<<<<<<
 *                 else:
 *                     raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)
 */
      __pyx_v_f = ((char *)"O");
      break;
      default:

      /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":325
 *                 elif t == NPY_OBJECT:      f = "O"
 *                 else:
 *                     raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)             # <<<<<<<<<<<<<<
 *                 info.format = f
 *                 return
 */
      __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_t); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 325, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_8 = PyUnicode_Format(__pyx_kp_u_unknown_dtype_code_in_numpy_pxd, __pyx_t_3); if (unlikely(!__pyx_t_8)) __PYX_ERR(3, 325, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 325, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(3, 325, __pyx_L1_error)
      break;
    }

    /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":326
 *                 else:
 *                     raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)
 *                 info.format = f             # <<<<<<<<<<<<<<
 *                 return
 *             else:
 */
    __pyx_v_info->format = __pyx_v_f;

    /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":327
 *                     raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)
 *                 info.format = f
 *                 return             # <<<<<<<<<<<<<<
 *             else:
 *                 info.format = <char*>PyObject_Malloc(_buffer_format_string_len)
 */
    __pyx_r = 0;
    goto __pyx_L0;

    /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":302
 *             info.obj = self
 * 
 *             if not PyDataType_HASFIELDS(descr):             # <<<<<<<<<<<<<<
 *                 t = descr.type_num
 *                 if ((descr.byteorder == c'>' and little_endian) or
 */
  }

  /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":329
 *                 return
 *             else:
 *                 info.format = <char*>PyObject_Malloc(_buffer_format_string_len)             # <<<<<<<<<<<<<<
 *                 info.format[0] = c'^' # Native data types, manual alignment
 *                 offset = 0
 */
  /*else*/ {
    __pyx_v_info->format = ((char *)PyObject_Malloc(0xFF));

    /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":330
 *             else:
 *                 info.format = <char*>PyObject_Malloc(_buffer_format_string_len)
 *                 info.format[0] = c'^' # Native data types, manual alignment             # <<<<<<<<<<<<<<
 *                 offset = 0
 *                 f = _util_dtypestring(descr, info.format + 1,
 */
    (__pyx_v_info->format[0]) = '^';

    /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":331
 *                 info.format = <char*>PyObject_Malloc(_buffer_format_string_len)
 *                 info.format[0] = c'^' # Native data types, manual alignment
 *                 offset = 0             # <<<<<<<<<<<<<<
 *                 f = _util_dtypestring(descr, info.format + 1,
 *                                       info.format + _buffer_format_string_len,
 */
    __pyx_v_offset = 0;

    /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":332
 *                 info.format[0] = c'^' # Native data types, manual alignment
 *                 offset = 0
 *                 f = _util_dtypestring(descr, info.format + 1,             # <<<<<<<<<<<<<<
 *                                       info.format + _buffer_format_string_len,
 *                                       &offset)
 */
    __pyx_t_9 = __pyx_f_5numpy__util_dtypestring(__pyx_v_descr, (__pyx_v_info->format + 1), (__pyx_v_info->format + 0xFF), (&__pyx_v_offset)); if (unlikely(__pyx_t_9 == ((char *)NULL))) __PYX_ERR(3, 332, __pyx_L1_error)
    __pyx_v_f = __pyx_t_9;

    /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":335
 *                                       info.format + _buffer_format_string_len,
 *                                       &offset)
 *                 f[0] = c'\0' # Terminate format string             # <<<<<<<<<<<<<<
 * 
 *         def __releasebuffer__(ndarray self, Py_buffer* info):
 */
    (__pyx_v_f[0]) = '\x00';
  }

  /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":258
 *         # experimental exception made for __getbuffer__ and __releasebuffer__
 *         # -- the details of this may change.
 *         def __getbuffer__(ndarray self, Py_buffer* info, int flags):             # <<<<<<<<<<<<<<
 *             # This implementation of getbuffer is geared towards Cython
 *             # requirements, and does not yet fulfill the PEP.
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("numpy.ndarray.__getbuffer__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  if (__pyx_v_info->obj != NULL) {
    __Pyx_GOTREF(__pyx_v_info->obj);
    __Pyx_DECREF(__pyx_v_info->obj); __pyx_v_info->obj = 0;
  }
  goto __pyx_L2;
  __pyx_L0:;
  if (__pyx_v_info->obj == Py_None) {
    __Pyx_GOTREF(__pyx_v_info->obj);
    __Pyx_DECREF(__pyx_v_info->obj); __pyx_v_info->obj = 0;
  }
  __pyx_L2:;
  __Pyx_XDECREF((PyObject *)__pyx_v_descr);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":337
 *                 f[0] = c'\0' # Terminate format string
 * 
 *         def __releasebuffer__(ndarray self, Py_buffer* info):             # <<<<<<<<<<<<<<
 *             if PyArray_HASFIELDS(self):
 *                 PyObject_Free(info.format)
 */

/* Python wrapper */
static CYTHON_UNUSED void __pyx_pw_5numpy_7ndarray_3__releasebuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info); /*proto*/
static CYTHON_UNUSED void __pyx_pw_5numpy_7ndarray_3__releasebuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__releasebuffer__ (wrapper)", 0);
  __pyx_pf_5numpy_7ndarray_2__releasebuffer__(((PyArrayObject *)__pyx_v_self), ((Py_buffer *)__pyx_v_info));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_5numpy_7ndarray_2__releasebuffer__(PyArrayObject *__pyx_v_self, Py_buffer *__pyx_v_info) {
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("__releasebuffer__", 0);

  /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":338
 * 
 *         def __releasebuffer__(ndarray self, Py_buffer* info):
 *             if PyArray_HASFIELDS(self):             # <<<<<<<<<<<<<<
 *                 PyObject_Free(info.format)
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):
 */
  __pyx_t_1 = (PyArray_HASFIELDS(__pyx_v_self) != 0);
  if (__pyx_t_1) {

    /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":339
 *         def __releasebuffer__(ndarray self, Py_buffer* info):
 *             if PyArray_HASFIELDS(self):
 *                 PyObject_Free(info.format)             # <<<<<<<<<<<<<<
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):
 *                 PyObject_Free(info.strides)
 */
    PyObject_Free(__pyx_v_info->format);

    /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":338
 * 
 *         def __releasebuffer__(ndarray self, Py_buffer* info):
 *             if PyArray_HASFIELDS(self):             # <<<<<<<<<<<<<<
 *                 PyObject_Free(info.format)
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):
 */
  }

  /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":340
 *             if PyArray_HASFIELDS(self):
 *                 PyObject_Free(info.format)
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):             # <<<<<<<<<<<<<<
 *                 PyObject_Free(info.strides)
 *                 # info.shape was stored after info.strides in the same block
 */
  __pyx_t_1 = (((sizeof(npy_intp)) != (sizeof(Py_ssize_t))) != 0);
  if (__pyx_t_1) {

    /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":341
 *                 PyObject_Free(info.format)
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):
 *                 PyObject_Free(info.strides)             # <<<<<<<<<<<<<<
 *                 # info.shape was stored after info.strides in the same block
 * 
 */
    PyObject_Free(__pyx_v_info->strides);

    /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":340
 *             if PyArray_HASFIELDS(self):
 *                 PyObject_Free(info.format)
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):             # <<<<<<<<<<<<<<
 *                 PyObject_Free(info.strides)
 *                 # info.shape was stored after info.strides in the same block
 */
  }

  /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":337
 *                 f[0] = c'\0' # Terminate format string
 * 
 *         def __releasebuffer__(ndarray self, Py_buffer* info):             # <<<<<<<<<<<<<<
 *             if PyArray_HASFIELDS(self):
 *                 PyObject_Free(info.format)
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":821
 * ctypedef npy_cdouble     complex_t
 * 
 * cdef inline object PyArray_MultiIterNew1(a):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew1(PyObject *__pyx_v_a) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew1", 0);

  /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":822
 * 
 * cdef inline object PyArray_MultiIterNew1(a):
 *     return PyArray_MultiIterNew(1, <void*>a)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(1, ((void *)__pyx_v_a)); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 822, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":821
 * ctypedef npy_cdouble     complex_t
 * 
 * cdef inline object PyArray_MultiIterNew1(a):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew1", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":824
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew2(PyObject *__pyx_v_a, PyObject *__pyx_v_b) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew2", 0);

  /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":825
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(2, ((void *)__pyx_v_a), ((void *)__pyx_v_b)); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 825, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":824
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew2", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":827
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew3(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew3", 0);

  /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":828
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(3, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c)); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 828, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":827
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew3", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":830
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew4(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c, PyObject *__pyx_v_d) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew4", 0);

  /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":831
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(4, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c), ((void *)__pyx_v_d)); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 831, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":830
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew4", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":833
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew5(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c, PyObject *__pyx_v_d, PyObject *__pyx_v_e) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew5", 0);

  /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":834
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)             # <<<<<<<<<<<<<<
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(5, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c), ((void *)__pyx_v_d), ((void *)__pyx_v_e)); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 834, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":833
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew5", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":836
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):             # <<<<<<<<<<<<<<
 *     if PyDataType_HASSUBARRAY(d):
 *         return <tuple>d.subarray.shape
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyDataType_SHAPE(PyArray_Descr *__pyx_v_d) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("PyDataType_SHAPE", 0);

  /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":837
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 *     if PyDataType_HASSUBARRAY(d):             # <<<<<<<<<<<<<<
 *         return <tuple>d.subarray.shape
 *     else:
 */
  __pyx_t_1 = (PyDataType_HASSUBARRAY(__pyx_v_d) != 0);
  if (__pyx_t_1) {

    /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":838
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 *     if PyDataType_HASSUBARRAY(d):
 *         return <tuple>d.subarray.shape             # <<<<<<<<<<<<<<
 *     else:
 *         return ()
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(((PyObject*)__pyx_v_d->subarray->shape));
    __pyx_r = ((PyObject*)__pyx_v_d->subarray->shape);
    goto __pyx_L0;

    /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":837
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 *     if PyDataType_HASSUBARRAY(d):             # <<<<<<<<<<<<<<
 *         return <tuple>d.subarray.shape
 *     else:
 */
  }

  /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":840
 *         return <tuple>d.subarray.shape
 *     else:
 *         return ()             # <<<<<<<<<<<<<<
 * 
 * cdef inline char* _util_dtypestring(dtype descr, char* f, char* end, int* offset) except NULL:
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_empty_tuple);
    __pyx_r = __pyx_empty_tuple;
    goto __pyx_L0;
  }

  /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":836
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):             # <<<<<<<<<<<<<<
 *     if PyDataType_HASSUBARRAY(d):
 *         return <tuple>d.subarray.shape
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":842
 *         return ()
 * 
 * cdef inline char* _util_dtypestring(dtype descr, char* f, char* end, int* offset) except NULL:             # <<<<<<<<<<<<<<
 *     # Recursive utility function used in __getbuffer__ to get format
 *     # string. The new location in the format string is returned.
 */

static CYTHON_INLINE char *__pyx_f_5numpy__util_dtypestring(PyArray_Descr *__pyx_v_descr, char *__pyx_v_f, char *__pyx_v_end, int *__pyx_v_offset) {
  PyArray_Descr *__pyx_v_child = 0;
  int __pyx_v_endian_detector;
  int __pyx_v_little_endian;
  PyObject *__pyx_v_fields = 0;
  PyObject *__pyx_v_childname = NULL;
  PyObject *__pyx_v_new_offset = NULL;
  PyObject *__pyx_v_t = NULL;
  char *__pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  long __pyx_t_8;
  char *__pyx_t_9;
  __Pyx_RefNannySetupContext("_util_dtypestring", 0);

  /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":847
 * 
 *     cdef dtype child
 *     cdef int endian_detector = 1             # <<<<<<<<<<<<<<
 *     cdef bint little_endian = ((<char*>&endian_detector)[0] != 0)
 *     cdef tuple fields
 */
  __pyx_v_endian_detector = 1;

  /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":848
 *     cdef dtype child
 *     cdef int endian_detector = 1
 *     cdef bint little_endian = ((<char*>&endian_detector)[0] != 0)             # <<<<<<<<<<<<<<
 *     cdef tuple fields
 * 
 */
  __pyx_v_little_endian = ((((char *)(&__pyx_v_endian_detector))[0]) != 0);

  /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":851
 *     cdef tuple fields
 * 
 *     for childname in descr.names:             # <<<<<<<<<<<<<<
 *         fields = descr.fields[childname]
 *         child, new_offset = fields
 */
  if (unlikely(__pyx_v_descr->names == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(3, 851, __pyx_L1_error)
  }
  __pyx_t_1 = __pyx_v_descr->names; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
  for (;;) {
    if (__pyx_t_2 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_3); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(3, 851, __pyx_L1_error)
    #else
    __pyx_t_3 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 851, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    __Pyx_XDECREF_SET(__pyx_v_childname, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":852
 * 
 *     for childname in descr.names:
 *         fields = descr.fields[childname]             # <<<<<<<<<<<<<<
 *         child, new_offset = fields
 * 
 */
    if (unlikely(__pyx_v_descr->fields == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(3, 852, __pyx_L1_error)
    }
    __pyx_t_3 = __Pyx_PyDict_GetItem(__pyx_v_descr->fields, __pyx_v_childname); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 852, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (!(likely(PyTuple_CheckExact(__pyx_t_3))||((__pyx_t_3) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_t_3)->tp_name), 0))) __PYX_ERR(3, 852, __pyx_L1_error)
    __Pyx_XDECREF_SET(__pyx_v_fields, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":853
 *     for childname in descr.names:
 *         fields = descr.fields[childname]
 *         child, new_offset = fields             # <<<<<<<<<<<<<<
 * 
 *         if (end - f) - <int>(new_offset - offset[0]) < 15:
 */
    if (likely(__pyx_v_fields != Py_None)) {
      PyObject* sequence = __pyx_v_fields;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(3, 853, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      __pyx_t_3 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_4 = PyTuple_GET_ITEM(sequence, 1); 
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      #else
      __pyx_t_3 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 853, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 853, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      #endif
    } else {
      __Pyx_RaiseNoneNotIterableError(); __PYX_ERR(3, 853, __pyx_L1_error)
    }
    if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_5numpy_dtype))))) __PYX_ERR(3, 853, __pyx_L1_error)
    __Pyx_XDECREF_SET(__pyx_v_child, ((PyArray_Descr *)__pyx_t_3));
    __pyx_t_3 = 0;
    __Pyx_XDECREF_SET(__pyx_v_new_offset, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":855
 *         child, new_offset = fields
 * 
 *         if (end - f) - <int>(new_offset - offset[0]) < 15:             # <<<<<<<<<<<<<<
 *             raise RuntimeError(u"Format string allocated too short, see comment in numpy.pxd")
 * 
 */
    __pyx_t_4 = __Pyx_PyInt_From_int((__pyx_v_offset[0])); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 855, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = PyNumber_Subtract(__pyx_v_new_offset, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 855, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_5 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(3, 855, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_6 = ((((__pyx_v_end - __pyx_v_f) - ((int)__pyx_t_5)) < 15) != 0);
    if (unlikely(__pyx_t_6)) {

      /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":856
 * 
 *         if (end - f) - <int>(new_offset - offset[0]) < 15:
 *             raise RuntimeError(u"Format string allocated too short, see comment in numpy.pxd")             # <<<<<<<<<<<<<<
 * 
 *         if ((child.byteorder == c'>' and little_endian) or
 */
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple__22, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 856, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(3, 856, __pyx_L1_error)

      /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":855
 *         child, new_offset = fields
 * 
 *         if (end - f) - <int>(new_offset - offset[0]) < 15:             # <<<<<<<<<<<<<<
 *             raise RuntimeError(u"Format string allocated too short, see comment in numpy.pxd")
 * 
 */
    }

    /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":858
 *             raise RuntimeError(u"Format string allocated too short, see comment in numpy.pxd")
 * 
 *         if ((child.byteorder == c'>' and little_endian) or             # <<<<<<<<<<<<<<
 *             (child.byteorder == c'<' and not little_endian)):
 *             raise ValueError(u"Non-native byte order not supported")
 */
    __pyx_t_7 = ((__pyx_v_child->byteorder == '>') != 0);
    if (!__pyx_t_7) {
      goto __pyx_L8_next_or;
    } else {
    }
    __pyx_t_7 = (__pyx_v_little_endian != 0);
    if (!__pyx_t_7) {
    } else {
      __pyx_t_6 = __pyx_t_7;
      goto __pyx_L7_bool_binop_done;
    }
    __pyx_L8_next_or:;

    /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":859
 * 
 *         if ((child.byteorder == c'>' and little_endian) or
 *             (child.byteorder == c'<' and not little_endian)):             # <<<<<<<<<<<<<<
 *             raise ValueError(u"Non-native byte order not supported")
 *             # One could encode it in the format string and have Cython
 */
    __pyx_t_7 = ((__pyx_v_child->byteorder == '<') != 0);
    if (__pyx_t_7) {
    } else {
      __pyx_t_6 = __pyx_t_7;
      goto __pyx_L7_bool_binop_done;
    }
    __pyx_t_7 = ((!(__pyx_v_little_endian != 0)) != 0);
    __pyx_t_6 = __pyx_t_7;
    __pyx_L7_bool_binop_done:;

    /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":858
 *             raise RuntimeError(u"Format string allocated too short, see comment in numpy.pxd")
 * 
 *         if ((child.byteorder == c'>' and little_endian) or             # <<<<<<<<<<<<<<
 *             (child.byteorder == c'<' and not little_endian)):
 *             raise ValueError(u"Non-native byte order not supported")
 */
    if (unlikely(__pyx_t_6)) {

      /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":860
 *         if ((child.byteorder == c'>' and little_endian) or
 *             (child.byteorder == c'<' and not little_endian)):
 *             raise ValueError(u"Non-native byte order not supported")             # <<<<<<<<<<<<<<
 *             # One could encode it in the format string and have Cython
 *             # complain instead, BUT: < and > in format strings also imply
 */
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__21, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 860, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(3, 860, __pyx_L1_error)

      /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":858
 *             raise RuntimeError(u"Format string allocated too short, see comment in numpy.pxd")
 * 
 *         if ((child.byteorder == c'>' and little_endian) or             # <<<<<<<<<<<<<<
 *             (child.byteorder == c'<' and not little_endian)):
 *             raise ValueError(u"Non-native byte order not supported")
 */
    }

    /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":870
 * 
 *         # Output padding bytes
 *         while offset[0] < new_offset:             # <<<<<<<<<<<<<<
 *             f[0] = 120 # "x"; pad byte
 *             f += 1
 */
    while (1) {
      __pyx_t_3 = __Pyx_PyInt_From_int((__pyx_v_offset[0])); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 870, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_t_3, __pyx_v_new_offset, Py_LT); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 870, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(3, 870, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (!__pyx_t_6) break;

      /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":871
 *         # Output padding bytes
 *         while offset[0] < new_offset:
 *             f[0] = 120 # "x"; pad byte             # <<<<<<<<<<<<<<
 *             f += 1
 *             offset[0] += 1
 */
      (__pyx_v_f[0]) = 0x78;

      /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":872
 *         while offset[0] < new_offset:
 *             f[0] = 120 # "x"; pad byte
 *             f += 1             # <<<<<<<<<<<<<<
 *             offset[0] += 1
 * 
 */
      __pyx_v_f = (__pyx_v_f + 1);

      /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":873
 *             f[0] = 120 # "x"; pad byte
 *             f += 1
 *             offset[0] += 1             # <<<<<<<<<<<<<<
 * 
 *         offset[0] += child.itemsize
 */
      __pyx_t_8 = 0;
      (__pyx_v_offset[__pyx_t_8]) = ((__pyx_v_offset[__pyx_t_8]) + 1);
    }

    /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":875
 *             offset[0] += 1
 * 
 *         offset[0] += child.itemsize             # <<<<<<<<<<<<<<
 * 
 *         if not PyDataType_HASFIELDS(child):
 */
    __pyx_t_8 = 0;
    (__pyx_v_offset[__pyx_t_8]) = ((__pyx_v_offset[__pyx_t_8]) + __pyx_v_child->elsize);

    /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":877
 *         offset[0] += child.itemsize
 * 
 *         if not PyDataType_HASFIELDS(child):             # <<<<<<<<<<<<<<
 *             t = child.type_num
 *             if end - f < 5:
 */
    __pyx_t_6 = ((!(PyDataType_HASFIELDS(__pyx_v_child) != 0)) != 0);
    if (__pyx_t_6) {

      /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":878
 * 
 *         if not PyDataType_HASFIELDS(child):
 *             t = child.type_num             # <<<<<<<<<<<<<<
 *             if end - f < 5:
 *                 raise RuntimeError(u"Format string allocated too short.")
 */
      __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_child->type_num); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 878, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_XDECREF_SET(__pyx_v_t, __pyx_t_4);
      __pyx_t_4 = 0;

      /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":879
 *         if not PyDataType_HASFIELDS(child):
 *             t = child.type_num
 *             if end - f < 5:             # <<<<<<<<<<<<<<
 *                 raise RuntimeError(u"Format string allocated too short.")
 * 
 */
      __pyx_t_6 = (((__pyx_v_end - __pyx_v_f) < 5) != 0);
      if (unlikely(__pyx_t_6)) {

        /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":880
 *             t = child.type_num
 *             if end - f < 5:
 *                 raise RuntimeError(u"Format string allocated too short.")             # <<<<<<<<<<<<<<
 * 
 *             # Until ticket #99 is fixed, use integers to avoid warnings
 */
        __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple__23, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 880, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_Raise(__pyx_t_4, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __PYX_ERR(3, 880, __pyx_L1_error)

        /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":879
 *         if not PyDataType_HASFIELDS(child):
 *             t = child.type_num
 *             if end - f < 5:             # <<<<<<<<<<<<<<
 *                 raise RuntimeError(u"Format string allocated too short.")
 * 
 */
      }

      /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":883
 * 
 *             # Until ticket #99 is fixed, use integers to avoid warnings
 *             if   t == NPY_BYTE:        f[0] =  98 #"b"             # <<<<<<<<<<<<<<
 *             elif t == NPY_UBYTE:       f[0] =  66 #"B"
 *             elif t == NPY_SHORT:       f[0] = 104 #"h"
 */
      __pyx_t_4 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_BYTE); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 883, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 883, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(3, 883, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 98;
        goto __pyx_L15;
      }

      /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":884
 *             # Until ticket #99 is fixed, use integers to avoid warnings
 *             if   t == NPY_BYTE:        f[0] =  98 #"b"
 *             elif t == NPY_UBYTE:       f[0] =  66 #"B"             # <<<<<<<<<<<<<<
 *             elif t == NPY_SHORT:       f[0] = 104 #"h"
 *             elif t == NPY_USHORT:      f[0] =  72 #"H"
 */
      __pyx_t_3 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_UBYTE); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 884, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 884, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(3, 884, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 66;
        goto __pyx_L15;
      }

      /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":885
 *             if   t == NPY_BYTE:        f[0] =  98 #"b"
 *             elif t == NPY_UBYTE:       f[0] =  66 #"B"
 *             elif t == NPY_SHORT:       f[0] = 104 #"h"             # <<<<<<<<<<<<<<
 *             elif t == NPY_USHORT:      f[0] =  72 #"H"
 *             elif t == NPY_INT:         f[0] = 105 #"i"
 */
      __pyx_t_4 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_SHORT); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 885, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 885, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(3, 885, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 0x68;
        goto __pyx_L15;
      }

      /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":886
 *             elif t == NPY_UBYTE:       f[0] =  66 #"B"
 *             elif t == NPY_SHORT:       f[0] = 104 #"h"
 *             elif t == NPY_USHORT:      f[0] =  72 #"H"             # <<<<<<<<<<<<<<
 *             elif t == NPY_INT:         f[0] = 105 #"i"
 *             elif t == NPY_UINT:        f[0] =  73 #"I"
 */
      __pyx_t_3 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_USHORT); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 886, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 886, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(3, 886, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 72;
        goto __pyx_L15;
      }

      /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":887
 *             elif t == NPY_SHORT:       f[0] = 104 #"h"
 *             elif t == NPY_USHORT:      f[0] =  72 #"H"
 *             elif t == NPY_INT:         f[0] = 105 #"i"             # <<<<<<<<<<<<<<
 *             elif t == NPY_UINT:        f[0] =  73 #"I"
 *             elif t == NPY_LONG:        f[0] = 108 #"l"
 */
      __pyx_t_4 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_INT); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 887, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 887, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(3, 887, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 0x69;
        goto __pyx_L15;
      }

      /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":888
 *             elif t == NPY_USHORT:      f[0] =  72 #"H"
 *             elif t == NPY_INT:         f[0] = 105 #"i"
 *             elif t == NPY_UINT:        f[0] =  73 #"I"             # <<<<<<<<<<<<<<
 *             elif t == NPY_LONG:        f[0] = 108 #"l"
 *             elif t == NPY_ULONG:       f[0] = 76  #"L"
 */
      __pyx_t_3 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_UINT); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 888, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 888, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(3, 888, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 73;
        goto __pyx_L15;
      }

      /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":889
 *             elif t == NPY_INT:         f[0] = 105 #"i"
 *             elif t == NPY_UINT:        f[0] =  73 #"I"
 *             elif t == NPY_LONG:        f[0] = 108 #"l"             # <<<<<<<<<<<<<<
 *             elif t == NPY_ULONG:       f[0] = 76  #"L"
 *             elif t == NPY_LONGLONG:    f[0] = 113 #"q"
 */
      __pyx_t_4 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_LONG); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 889, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 889, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(3, 889, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 0x6C;
        goto __pyx_L15;
      }

      /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":890
 *             elif t == NPY_UINT:        f[0] =  73 #"I"
 *             elif t == NPY_LONG:        f[0] = 108 #"l"
 *             elif t == NPY_ULONG:       f[0] = 76  #"L"             # <<<<<<<<<<<<<<
 *             elif t == NPY_LONGLONG:    f[0] = 113 #"q"
 *             elif t == NPY_ULONGLONG:   f[0] = 81  #"Q"
 */
      __pyx_t_3 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_ULONG); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 890, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 890, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(3, 890, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 76;
        goto __pyx_L15;
      }

      /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":891
 *             elif t == NPY_LONG:        f[0] = 108 #"l"
 *             elif t == NPY_ULONG:       f[0] = 76  #"L"
 *             elif t == NPY_LONGLONG:    f[0] = 113 #"q"             # <<<<<<<<<<<<<<
 *             elif t == NPY_ULONGLONG:   f[0] = 81  #"Q"
 *             elif t == NPY_FLOAT:       f[0] = 102 #"f"
 */
      __pyx_t_4 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_LONGLONG); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 891, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 891, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(3, 891, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 0x71;
        goto __pyx_L15;
      }

      /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":892
 *             elif t == NPY_ULONG:       f[0] = 76  #"L"
 *             elif t == NPY_LONGLONG:    f[0] = 113 #"q"
 *             elif t == NPY_ULONGLONG:   f[0] = 81  #"Q"             # <<<<<<<<<<<<<<
 *             elif t == NPY_FLOAT:       f[0] = 102 #"f"
 *             elif t == NPY_DOUBLE:      f[0] = 100 #"d"
 */
      __pyx_t_3 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_ULONGLONG); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 892, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 892, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(3, 892, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 81;
        goto __pyx_L15;
      }

      /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":893
 *             elif t == NPY_LONGLONG:    f[0] = 113 #"q"
 *             elif t == NPY_ULONGLONG:   f[0] = 81  #"Q"
 *             elif t == NPY_FLOAT:       f[0] = 102 #"f"             # <<<<<<<<<<<<<<
 *             elif t == NPY_DOUBLE:      f[0] = 100 #"d"
 *             elif t == NPY_LONGDOUBLE:  f[0] = 103 #"g"
 */
      __pyx_t_4 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_FLOAT); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 893, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 893, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(3, 893, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 0x66;
        goto __pyx_L15;
      }

      /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":894
 *             elif t == NPY_ULONGLONG:   f[0] = 81  #"Q"
 *             elif t == NPY_FLOAT:       f[0] = 102 #"f"
 *             elif t == NPY_DOUBLE:      f[0] = 100 #"d"             # <<<<<<<<<<<<<<
 *             elif t == NPY_LONGDOUBLE:  f[0] = 103 #"g"
 *             elif t == NPY_CFLOAT:      f[0] = 90; f[1] = 102; f += 1 # Zf
 */
      __pyx_t_3 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_DOUBLE); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 894, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 894, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(3, 894, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 0x64;
        goto __pyx_L15;
      }

      /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":895
 *             elif t == NPY_FLOAT:       f[0] = 102 #"f"
 *             elif t == NPY_DOUBLE:      f[0] = 100 #"d"
 *             elif t == NPY_LONGDOUBLE:  f[0] = 103 #"g"             # <<<<<<<<<<<<<<
 *             elif t == NPY_CFLOAT:      f[0] = 90; f[1] = 102; f += 1 # Zf
 *             elif t == NPY_CDOUBLE:     f[0] = 90; f[1] = 100; f += 1 # Zd
 */
      __pyx_t_4 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_LONGDOUBLE); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 895, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 895, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(3, 895, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 0x67;
        goto __pyx_L15;
      }

      /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":896
 *             elif t == NPY_DOUBLE:      f[0] = 100 #"d"
 *             elif t == NPY_LONGDOUBLE:  f[0] = 103 #"g"
 *             elif t == NPY_CFLOAT:      f[0] = 90; f[1] = 102; f += 1 # Zf             # <<<<<<<<<<<<<<
 *             elif t == NPY_CDOUBLE:     f[0] = 90; f[1] = 100; f += 1 # Zd
 *             elif t == NPY_CLONGDOUBLE: f[0] = 90; f[1] = 103; f += 1 # Zg
 */
      __pyx_t_3 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_CFLOAT); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 896, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 896, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(3, 896, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 90;
        (__pyx_v_f[1]) = 0x66;
        __pyx_v_f = (__pyx_v_f + 1);
        goto __pyx_L15;
      }

      /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":897
 *             elif t == NPY_LONGDOUBLE:  f[0] = 103 #"g"
 *             elif t == NPY_CFLOAT:      f[0] = 90; f[1] = 102; f += 1 # Zf
 *             elif t == NPY_CDOUBLE:     f[0] = 90; f[1] = 100; f += 1 # Zd             # <<<<<<<<<<<<<<
 *             elif t == NPY_CLONGDOUBLE: f[0] = 90; f[1] = 103; f += 1 # Zg
 *             elif t == NPY_OBJECT:      f[0] = 79 #"O"
 */
      __pyx_t_4 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_CDOUBLE); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 897, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 897, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(3, 897, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 90;
        (__pyx_v_f[1]) = 0x64;
        __pyx_v_f = (__pyx_v_f + 1);
        goto __pyx_L15;
      }

      /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":898
 *             elif t == NPY_CFLOAT:      f[0] = 90; f[1] = 102; f += 1 # Zf
 *             elif t == NPY_CDOUBLE:     f[0] = 90; f[1] = 100; f += 1 # Zd
 *             elif t == NPY_CLONGDOUBLE: f[0] = 90; f[1] = 103; f += 1 # Zg             # <<<<<<<<<<<<<<
 *             elif t == NPY_OBJECT:      f[0] = 79 #"O"
 *             else:
 */
      __pyx_t_3 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_CLONGDOUBLE); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 898, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 898, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(3, 898, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 90;
        (__pyx_v_f[1]) = 0x67;
        __pyx_v_f = (__pyx_v_f + 1);
        goto __pyx_L15;
      }

      /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":899
 *             elif t == NPY_CDOUBLE:     f[0] = 90; f[1] = 100; f += 1 # Zd
 *             elif t == NPY_CLONGDOUBLE: f[0] = 90; f[1] = 103; f += 1 # Zg
 *             elif t == NPY_OBJECT:      f[0] = 79 #"O"             # <<<<<<<<<<<<<<
 *             else:
 *                 raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)
 */
      __pyx_t_4 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_OBJECT); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 899, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 899, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(3, 899, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (likely(__pyx_t_6)) {
        (__pyx_v_f[0]) = 79;
        goto __pyx_L15;
      }

      /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":901
 *             elif t == NPY_OBJECT:      f[0] = 79 #"O"
 *             else:
 *                 raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)             # <<<<<<<<<<<<<<
 *             f += 1
 *         else:
 */
      /*else*/ {
        __pyx_t_3 = __Pyx_PyUnicode_FormatSafe(__pyx_kp_u_unknown_dtype_code_in_numpy_pxd, __pyx_v_t); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 901, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 901, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_Raise(__pyx_t_4, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __PYX_ERR(3, 901, __pyx_L1_error)
      }
      __pyx_L15:;

      /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":902
 *             else:
 *                 raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)
 *             f += 1             # <<<<<<<<<<<<<<
 *         else:
 *             # Cython ignores struct boundary information ("T{...}"),
 */
      __pyx_v_f = (__pyx_v_f + 1);

      /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":877
 *         offset[0] += child.itemsize
 * 
 *         if not PyDataType_HASFIELDS(child):             # <<<<<<<<<<<<<<
 *             t = child.type_num
 *             if end - f < 5:
 */
      goto __pyx_L13;
    }

    /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":906
 *             # Cython ignores struct boundary information ("T{...}"),
 *             # so don't output it
 *             f = _util_dtypestring(child, f, end, offset)             # <<<<<<<<<<<<<<
 *     return f
 * 
 */
    /*else*/ {
      __pyx_t_9 = __pyx_f_5numpy__util_dtypestring(__pyx_v_child, __pyx_v_f, __pyx_v_end, __pyx_v_offset); if (unlikely(__pyx_t_9 == ((char *)NULL))) __PYX_ERR(3, 906, __pyx_L1_error)
      __pyx_v_f = __pyx_t_9;
    }
    __pyx_L13:;

    /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":851
 *     cdef tuple fields
 * 
 *     for childname in descr.names:             # <<<<<<<<<<<<<<
 *         fields = descr.fields[childname]
 *         child, new_offset = fields
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":907
 *             # so don't output it
 *             f = _util_dtypestring(child, f, end, offset)
 *     return f             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_f;
  goto __pyx_L0;

  /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":842
 *         return ()
 * 
 * cdef inline char* _util_dtypestring(dtype descr, char* f, char* end, int* offset) except NULL:             # <<<<<<<<<<<<<<
 *     # Recursive utility function used in __getbuffer__ to get format
 *     # string. The new location in the format string is returned.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("numpy._util_dtypestring", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_child);
  __Pyx_XDECREF(__pyx_v_fields);
  __Pyx_XDECREF(__pyx_v_childname);
  __Pyx_XDECREF(__pyx_v_new_offset);
  __Pyx_XDECREF(__pyx_v_t);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":1022
 *     int _import_umath() except -1
 * 
 * cdef inline void set_array_base(ndarray arr, object base):             # <<<<<<<<<<<<<<
 *     Py_INCREF(base) # important to do this before stealing the reference below!
 *     PyArray_SetBaseObject(arr, base)
 */

static CYTHON_INLINE void __pyx_f_5numpy_set_array_base(PyArrayObject *__pyx_v_arr, PyObject *__pyx_v_base) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_array_base", 0);

  /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":1023
 * 
 * cdef inline void set_array_base(ndarray arr, object base):
 *     Py_INCREF(base) # important to do this before stealing the reference below!             # <<<<<<<<<<<<<<
 *     PyArray_SetBaseObject(arr, base)
 * 
 */
  Py_INCREF(__pyx_v_base);

  /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":1024
 * cdef inline void set_array_base(ndarray arr, object base):
 *     Py_INCREF(base) # important to do this before stealing the reference below!
 *     PyArray_SetBaseObject(arr, base)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object get_array_base(ndarray arr):
 */
  (void)(PyArray_SetBaseObject(__pyx_v_arr, __pyx_v_base));

  /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":1022
 *     int _import_umath() except -1
 * 
 * cdef inline void set_array_base(ndarray arr, object base):             # <<<<<<<<<<<<<<
 *     Py_INCREF(base) # important to do this before stealing the reference below!
 *     PyArray_SetBaseObject(arr, base)
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":1026
 *     PyArray_SetBaseObject(arr, base)
 * 
 * cdef inline object get_array_base(ndarray arr):             # <<<<<<<<<<<<<<
 *     base = PyArray_BASE(arr)
 *     if base is NULL:
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_get_array_base(PyArrayObject *__pyx_v_arr) {
  PyObject *__pyx_v_base;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("get_array_base", 0);

  /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":1027
 * 
 * cdef inline object get_array_base(ndarray arr):
 *     base = PyArray_BASE(arr)             # <<<<<<<<<<<<<<
 *     if base is NULL:
 *         return None
 */
  __pyx_v_base = PyArray_BASE(__pyx_v_arr);

  /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":1028
 * cdef inline object get_array_base(ndarray arr):
 *     base = PyArray_BASE(arr)
 *     if base is NULL:             # <<<<<<<<<<<<<<
 *         return None
 *     return <object>base
 */
  __pyx_t_1 = ((__pyx_v_base == NULL) != 0);
  if (__pyx_t_1) {

    /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":1029
 *     base = PyArray_BASE(arr)
 *     if base is NULL:
 *         return None             # <<<<<<<<<<<<<<
 *     return <object>base
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":1028
 * cdef inline object get_array_base(ndarray arr):
 *     base = PyArray_BASE(arr)
 *     if base is NULL:             # <<<<<<<<<<<<<<
 *         return None
 *     return <object>base
 */
  }

  /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":1030
 *     if base is NULL:
 *         return None
 *     return <object>base             # <<<<<<<<<<<<<<
 * 
 * # Versions of the import_* functions which are more suitable for
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_base));
  __pyx_r = ((PyObject *)__pyx_v_base);
  goto __pyx_L0;

  /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":1026
 *     PyArray_SetBaseObject(arr, base)
 * 
 * cdef inline object get_array_base(ndarray arr):             # <<<<<<<<<<<<<<
 *     base = PyArray_BASE(arr)
 *     if base is NULL:
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":1034
 * # Versions of the import_* functions which are more suitable for
 * # Cython code.
 * cdef inline int import_array() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_array()
 */

static CYTHON_INLINE int __pyx_f_5numpy_import_array(void) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  __Pyx_RefNannySetupContext("import_array", 0);

  /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":1035
 * # Cython code.
 * cdef inline int import_array() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_array()
 *     except Exception:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":1036
 * cdef inline int import_array() except -1:
 *     try:
 *         _import_array()             # <<<<<<<<<<<<<<
 *     except Exception:
 *         raise ImportError("numpy.core.multiarray failed to import")
 */
      __pyx_t_4 = _import_array(); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(3, 1036, __pyx_L3_error)

      /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":1035
 * # Cython code.
 * cdef inline int import_array() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_array()
 *     except Exception:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;

    /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":1037
 *     try:
 *         _import_array()
 *     except Exception:             # <<<<<<<<<<<<<<
 *         raise ImportError("numpy.core.multiarray failed to import")
 * 
 */
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_4) {
      __Pyx_AddTraceback("numpy.import_array", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(3, 1037, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_7);

      /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":1038
 *         _import_array()
 *     except Exception:
 *         raise ImportError("numpy.core.multiarray failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_umath() except -1:
 */
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_ImportError, __pyx_tuple__24, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(3, 1038, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __PYX_ERR(3, 1038, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":1035
 * # Cython code.
 * cdef inline int import_array() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_array()
 *     except Exception:
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L8_try_end:;
  }

  /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":1034
 * # Versions of the import_* functions which are more suitable for
 * # Cython code.
 * cdef inline int import_array() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_array()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("numpy.import_array", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":1040
 *         raise ImportError("numpy.core.multiarray failed to import")
 * 
 * cdef inline int import_umath() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

static CYTHON_INLINE int __pyx_f_5numpy_import_umath(void) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  __Pyx_RefNannySetupContext("import_umath", 0);

  /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":1041
 * 
 * cdef inline int import_umath() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":1042
 * cdef inline int import_umath() except -1:
 *     try:
 *         _import_umath()             # <<<<<<<<<<<<<<
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")
 */
      __pyx_t_4 = _import_umath(); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(3, 1042, __pyx_L3_error)

      /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":1041
 * 
 * cdef inline int import_umath() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;

    /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":1043
 *     try:
 *         _import_umath()
 *     except Exception:             # <<<<<<<<<<<<<<
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 */
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_4) {
      __Pyx_AddTraceback("numpy.import_umath", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(3, 1043, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_7);

      /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":1044
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_ufunc() except -1:
 */
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_ImportError, __pyx_tuple__25, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(3, 1044, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __PYX_ERR(3, 1044, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":1041
 * 
 * cdef inline int import_umath() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L8_try_end:;
  }

  /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":1040
 *         raise ImportError("numpy.core.multiarray failed to import")
 * 
 * cdef inline int import_umath() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("numpy.import_umath", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":1046
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 * cdef inline int import_ufunc() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

static CYTHON_INLINE int __pyx_f_5numpy_import_ufunc(void) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  __Pyx_RefNannySetupContext("import_ufunc", 0);

  /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":1047
 * 
 * cdef inline int import_ufunc() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":1048
 * cdef inline int import_ufunc() except -1:
 *     try:
 *         _import_umath()             # <<<<<<<<<<<<<<
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")
 */
      __pyx_t_4 = _import_umath(); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(3, 1048, __pyx_L3_error)

      /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":1047
 * 
 * cdef inline int import_ufunc() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;

    /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":1049
 *     try:
 *         _import_umath()
 *     except Exception:             # <<<<<<<<<<<<<<
 *         raise ImportError("numpy.core.umath failed to import")
 */
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_4) {
      __Pyx_AddTraceback("numpy.import_ufunc", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(3, 1049, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_7);

      /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":1050
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")             # <<<<<<<<<<<<<<
 */
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_ImportError, __pyx_tuple__25, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(3, 1050, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __PYX_ERR(3, 1050, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":1047
 * 
 * cdef inline int import_ufunc() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L8_try_end:;
  }

  /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":1046
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 * cdef inline int import_ufunc() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("numpy.import_ufunc", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/tslibs/util.pxd":54
 * 
 * 
 * cdef inline int64_t get_nat():             # <<<<<<<<<<<<<<
 *     return NPY_MIN_INT64
 * 
 */

static CYTHON_INLINE __pyx_t_5numpy_int64_t __pyx_f_6pandas_5_libs_6tslibs_4util_get_nat(void) {
  __pyx_t_5numpy_int64_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_nat", 0);

  /* "pandas/_libs/tslibs/util.pxd":55
 * 
 * cdef inline int64_t get_nat():
 *     return NPY_MIN_INT64             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = NPY_MIN_INT64;
  goto __pyx_L0;

  /* "pandas/_libs/tslibs/util.pxd":54
 * 
 * 
 * cdef inline int64_t get_nat():             # <<<<<<<<<<<<<<
 *     return NPY_MIN_INT64
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/tslibs/util.pxd":61
 * # Type Checking
 * 
 * cdef inline bint is_integer_object(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     Cython equivalent of
 */

static CYTHON_INLINE int __pyx_f_6pandas_5_libs_6tslibs_4util_is_integer_object(PyObject *__pyx_v_obj) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_t_2;

  /* "pandas/_libs/tslibs/util.pxd":79
 *     This counts np.timedelta64 objects as integers.
 *     """
 *     return (not PyBool_Check(obj) and PyArray_IsIntegerScalar(obj)             # <<<<<<<<<<<<<<
 *             and not is_timedelta64_object(obj))
 * 
 */
  __pyx_t_2 = ((!(PyBool_Check(__pyx_v_obj) != 0)) != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L3_bool_binop_done;
  }

  /* "pandas/_libs/tslibs/util.pxd":80
 *     """
 *     return (not PyBool_Check(obj) and PyArray_IsIntegerScalar(obj)
 *             and not is_timedelta64_object(obj))             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = (PyArray_IsIntegerScalar(__pyx_v_obj) != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L3_bool_binop_done;
  }
  __pyx_t_2 = ((!(__pyx_f_6pandas_5_libs_6tslibs_4util_is_timedelta64_object(__pyx_v_obj) != 0)) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L3_bool_binop_done:;
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "pandas/_libs/tslibs/util.pxd":61
 * # Type Checking
 * 
 * cdef inline bint is_integer_object(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     Cython equivalent of
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pandas/_libs/tslibs/util.pxd":83
 * 
 * 
 * cdef inline bint is_float_object(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     Cython equivalent of `isinstance(val, (float, np.complex_))`
 */

static CYTHON_INLINE int __pyx_f_6pandas_5_libs_6tslibs_4util_is_float_object(PyObject *__pyx_v_obj) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_t_2;

  /* "pandas/_libs/tslibs/util.pxd":95
 *     is_float : bool
 *     """
 *     return (PyFloat_Check(obj) or             # <<<<<<<<<<<<<<
 *             (PyObject_TypeCheck(obj, &PyFloatingArrType_Type)))
 * 
 */
  __pyx_t_2 = (PyFloat_Check(__pyx_v_obj) != 0);
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L3_bool_binop_done;
  }

  /* "pandas/_libs/tslibs/util.pxd":96
 *     """
 *     return (PyFloat_Check(obj) or
 *             (PyObject_TypeCheck(obj, &PyFloatingArrType_Type)))             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = (PyObject_TypeCheck(__pyx_v_obj, (&PyFloatingArrType_Type)) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L3_bool_binop_done:;
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "pandas/_libs/tslibs/util.pxd":83
 * 
 * 
 * cdef inline bint is_float_object(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     Cython equivalent of `isinstance(val, (float, np.complex_))`
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pandas/_libs/tslibs/util.pxd":99
 * 
 * 
 * cdef inline bint is_complex_object(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     Cython equivalent of `isinstance(val, (complex, np.complex_))`
 */

static CYTHON_INLINE int __pyx_f_6pandas_5_libs_6tslibs_4util_is_complex_object(PyObject *__pyx_v_obj) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_t_2;

  /* "pandas/_libs/tslibs/util.pxd":111
 *     is_complex : bool
 *     """
 *     return (PyComplex_Check(obj) or             # <<<<<<<<<<<<<<
 *             PyObject_TypeCheck(obj, &PyComplexFloatingArrType_Type))
 * 
 */
  __pyx_t_2 = (PyComplex_Check(__pyx_v_obj) != 0);
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L3_bool_binop_done;
  }

  /* "pandas/_libs/tslibs/util.pxd":112
 *     """
 *     return (PyComplex_Check(obj) or
 *             PyObject_TypeCheck(obj, &PyComplexFloatingArrType_Type))             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = (PyObject_TypeCheck(__pyx_v_obj, (&PyComplexFloatingArrType_Type)) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L3_bool_binop_done:;
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "pandas/_libs/tslibs/util.pxd":99
 * 
 * 
 * cdef inline bint is_complex_object(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     Cython equivalent of `isinstance(val, (complex, np.complex_))`
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pandas/_libs/tslibs/util.pxd":115
 * 
 * 
 * cdef inline bint is_bool_object(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     Cython equivalent of `isinstance(val, (bool, np.bool_))`
 */

static CYTHON_INLINE int __pyx_f_6pandas_5_libs_6tslibs_4util_is_bool_object(PyObject *__pyx_v_obj) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_t_2;

  /* "pandas/_libs/tslibs/util.pxd":127
 *     is_bool : bool
 *     """
 *     return (PyBool_Check(obj) or             # <<<<<<<<<<<<<<
 *             PyObject_TypeCheck(obj, &PyBoolArrType_Type))
 * 
 */
  __pyx_t_2 = (PyBool_Check(__pyx_v_obj) != 0);
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L3_bool_binop_done;
  }

  /* "pandas/_libs/tslibs/util.pxd":128
 *     """
 *     return (PyBool_Check(obj) or
 *             PyObject_TypeCheck(obj, &PyBoolArrType_Type))             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = (PyObject_TypeCheck(__pyx_v_obj, (&PyBoolArrType_Type)) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L3_bool_binop_done:;
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "pandas/_libs/tslibs/util.pxd":115
 * 
 * 
 * cdef inline bint is_bool_object(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     Cython equivalent of `isinstance(val, (bool, np.bool_))`
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pandas/_libs/tslibs/util.pxd":131
 * 
 * 
 * cdef inline bint is_timedelta64_object(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     Cython equivalent of `isinstance(val, np.timedelta64)`
 */

static CYTHON_INLINE int __pyx_f_6pandas_5_libs_6tslibs_4util_is_timedelta64_object(PyObject *__pyx_v_obj) {
  int __pyx_r;

  /* "pandas/_libs/tslibs/util.pxd":143
 *     is_timedelta64 : bool
 *     """
 *     return PyObject_TypeCheck(obj, &PyTimedeltaArrType_Type)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = PyObject_TypeCheck(__pyx_v_obj, (&PyTimedeltaArrType_Type));
  goto __pyx_L0;

  /* "pandas/_libs/tslibs/util.pxd":131
 * 
 * 
 * cdef inline bint is_timedelta64_object(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     Cython equivalent of `isinstance(val, np.timedelta64)`
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pandas/_libs/tslibs/util.pxd":146
 * 
 * 
 * cdef inline bint is_datetime64_object(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     Cython equivalent of `isinstance(val, np.datetime64)`
 */

static CYTHON_INLINE int __pyx_f_6pandas_5_libs_6tslibs_4util_is_datetime64_object(PyObject *__pyx_v_obj) {
  int __pyx_r;

  /* "pandas/_libs/tslibs/util.pxd":158
 *     is_datetime64 : bool
 *     """
 *     return PyObject_TypeCheck(obj, &PyDatetimeArrType_Type)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = PyObject_TypeCheck(__pyx_v_obj, (&PyDatetimeArrType_Type));
  goto __pyx_L0;

  /* "pandas/_libs/tslibs/util.pxd":146
 * 
 * 
 * cdef inline bint is_datetime64_object(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     Cython equivalent of `isinstance(val, np.datetime64)`
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "pandas/_libs/tslibs/util.pxd":161
 * 
 * 
 * cdef inline bint is_array(object val):             # <<<<<<<<<<<<<<
 *     """
 *     Cython equivalent of `isinstance(val, np.ndarray)`
 */

static CYTHON_INLINE int __pyx_f_6pandas_5_libs_6tslibs_4util_is_array(PyObject *__pyx_v_val) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("is_array", 0);

  /* "pandas/_libs/tslibs/util.pxd":173
 *     is_ndarray : bool
 *     """
 *     return PyArray_Check(val)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = PyArray_Check(__pyx_v_val);
  goto __pyx_L0;

  /* "pandas/_libs/tslibs/util.pxd":161
 * 
 * 
 * cdef inline bint is_array(object val):             # <<<<<<<<<<<<<<
 *     """
 *     Cython equivalent of `isinstance(val, np.ndarray)`
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/tslibs/util.pxd":176
 * 
 * 
 * cdef inline bint is_period_object(object val):             # <<<<<<<<<<<<<<
 *     """
 *     Cython equivalent of `isinstance(val, pd.Period)`
 */

static CYTHON_INLINE int __pyx_f_6pandas_5_libs_6tslibs_4util_is_period_object(PyObject *__pyx_v_val) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  __Pyx_RefNannySetupContext("is_period_object", 0);

  /* "pandas/_libs/tslibs/util.pxd":188
 *     is_period : bool
 *     """
 *     return getattr(val, '_typ', '_typ') == 'period'             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __Pyx_GetAttr3(__pyx_v_val, __pyx_n_u_typ_2, __pyx_n_u_typ_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 188, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_RichCompare(__pyx_t_1, __pyx_n_u_period, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 188, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_3 == (int)-1) && PyErr_Occurred())) __PYX_ERR(4, 188, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  goto __pyx_L0;

  /* "pandas/_libs/tslibs/util.pxd":176
 * 
 * 
 * cdef inline bint is_period_object(object val):             # <<<<<<<<<<<<<<
 *     """
 *     Cython equivalent of `isinstance(val, pd.Period)`
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_WriteUnraisable("pandas._libs.tslibs.util.is_period_object", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/tslibs/util.pxd":191
 * 
 * 
 * cdef inline bint is_offset_object(object val):             # <<<<<<<<<<<<<<
 *     """
 *     Check if an object is a DateOffset object.
 */

static CYTHON_INLINE int __pyx_f_6pandas_5_libs_6tslibs_4util_is_offset_object(PyObject *__pyx_v_val) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  __Pyx_RefNannySetupContext("is_offset_object", 0);

  /* "pandas/_libs/tslibs/util.pxd":203
 *     is_date_offset : bool
 *     """
 *     return getattr(val, '_typ', None) == "dateoffset"             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __Pyx_GetAttr3(__pyx_v_val, __pyx_n_u_typ_2, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 203, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_RichCompare(__pyx_t_1, __pyx_n_u_dateoffset, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 203, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_3 == (int)-1) && PyErr_Occurred())) __PYX_ERR(4, 203, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  goto __pyx_L0;

  /* "pandas/_libs/tslibs/util.pxd":191
 * 
 * 
 * cdef inline bint is_offset_object(object val):             # <<<<<<<<<<<<<<
 *     """
 *     Check if an object is a DateOffset object.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_WriteUnraisable("pandas._libs.tslibs.util.is_offset_object", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/tslibs/util.pxd":206
 * 
 * 
 * cdef inline bint is_nan(object val):             # <<<<<<<<<<<<<<
 *     """
 *     Check if val is a Not-A-Number float or complex, including
 */

static CYTHON_INLINE int __pyx_f_6pandas_5_libs_6tslibs_4util_is_nan(PyObject *__pyx_v_val) {
  __pyx_t_5numpy_float64_t __pyx_v_fval;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __pyx_t_5numpy_float64_t __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("is_nan", 0);

  /* "pandas/_libs/tslibs/util.pxd":220
 *     """
 *     cdef float64_t fval
 *     if is_float_object(val):             # <<<<<<<<<<<<<<
 *         fval = val
 *         return fval != fval
 */
  __pyx_t_1 = (__pyx_f_6pandas_5_libs_6tslibs_4util_is_float_object(__pyx_v_val) != 0);
  if (__pyx_t_1) {

    /* "pandas/_libs/tslibs/util.pxd":221
 *     cdef float64_t fval
 *     if is_float_object(val):
 *         fval = val             # <<<<<<<<<<<<<<
 *         return fval != fval
 *     return is_complex_object(val) and val != val
 */
    __pyx_t_2 = __pyx_PyFloat_AsDouble(__pyx_v_val); if (unlikely((__pyx_t_2 == ((npy_float64)-1)) && PyErr_Occurred())) __PYX_ERR(4, 221, __pyx_L1_error)
    __pyx_v_fval = __pyx_t_2;

    /* "pandas/_libs/tslibs/util.pxd":222
 *     if is_float_object(val):
 *         fval = val
 *         return fval != fval             # <<<<<<<<<<<<<<
 *     return is_complex_object(val) and val != val
 * 
 */
    __pyx_r = (__pyx_v_fval != __pyx_v_fval);
    goto __pyx_L0;

    /* "pandas/_libs/tslibs/util.pxd":220
 *     """
 *     cdef float64_t fval
 *     if is_float_object(val):             # <<<<<<<<<<<<<<
 *         fval = val
 *         return fval != fval
 */
  }

  /* "pandas/_libs/tslibs/util.pxd":223
 *         fval = val
 *         return fval != fval
 *     return is_complex_object(val) and val != val             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_3 = (__pyx_f_6pandas_5_libs_6tslibs_4util_is_complex_object(__pyx_v_val) != 0);
  if (__pyx_t_3) {
  } else {
    __pyx_t_1 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = PyObject_RichCompare(__pyx_v_val, __pyx_v_val, Py_NE); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 223, __pyx_L1_error)
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(4, 223, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_1 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "pandas/_libs/tslibs/util.pxd":206
 * 
 * 
 * cdef inline bint is_nan(object val):             # <<<<<<<<<<<<<<
 *     """
 *     Check if val is a Not-A-Number float or complex, including
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_WriteUnraisable("pandas._libs.tslibs.util.is_nan", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/tslibs/util.pxd":226
 * 
 * 
 * cdef inline const char* get_c_string_buf_and_size(object py_string,             # <<<<<<<<<<<<<<
 *                                                   Py_ssize_t *length):
 *     """
 */

static CYTHON_INLINE char const *__pyx_f_6pandas_5_libs_6tslibs_4util_get_c_string_buf_and_size(PyObject *__pyx_v_py_string, Py_ssize_t *__pyx_v_length) {
  char const *__pyx_v_buf;
  char const *__pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  char const *__pyx_t_2;
  __Pyx_RefNannySetupContext("get_c_string_buf_and_size", 0);

  /* "pandas/_libs/tslibs/util.pxd":249
 *         const char *buf
 * 
 *     if PyUnicode_Check(py_string):             # <<<<<<<<<<<<<<
 *         buf = PyUnicode_AsUTF8AndSize(py_string, length)
 *     else:
 */
  __pyx_t_1 = (PyUnicode_Check(__pyx_v_py_string) != 0);
  if (__pyx_t_1) {

    /* "pandas/_libs/tslibs/util.pxd":250
 * 
 *     if PyUnicode_Check(py_string):
 *         buf = PyUnicode_AsUTF8AndSize(py_string, length)             # <<<<<<<<<<<<<<
 *     else:
 *         PyBytes_AsStringAndSize(py_string, <char**>&buf, length)
 */
    __pyx_t_2 = PyUnicode_AsUTF8AndSize(__pyx_v_py_string, __pyx_v_length); if (unlikely(__pyx_t_2 == ((char const *)NULL))) __PYX_ERR(4, 250, __pyx_L1_error)
    __pyx_v_buf = __pyx_t_2;

    /* "pandas/_libs/tslibs/util.pxd":249
 *         const char *buf
 * 
 *     if PyUnicode_Check(py_string):             # <<<<<<<<<<<<<<
 *         buf = PyUnicode_AsUTF8AndSize(py_string, length)
 *     else:
 */
    goto __pyx_L3;
  }

  /* "pandas/_libs/tslibs/util.pxd":252
 *         buf = PyUnicode_AsUTF8AndSize(py_string, length)
 *     else:
 *         PyBytes_AsStringAndSize(py_string, <char**>&buf, length)             # <<<<<<<<<<<<<<
 *     return buf
 * 
 */
  /*else*/ {
    __pyx_t_1 = PyBytes_AsStringAndSize(__pyx_v_py_string, ((char **)(&__pyx_v_buf)), __pyx_v_length); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(4, 252, __pyx_L1_error)
  }
  __pyx_L3:;

  /* "pandas/_libs/tslibs/util.pxd":253
 *     else:
 *         PyBytes_AsStringAndSize(py_string, <char**>&buf, length)
 *     return buf             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_buf;
  goto __pyx_L0;

  /* "pandas/_libs/tslibs/util.pxd":226
 * 
 * 
 * cdef inline const char* get_c_string_buf_and_size(object py_string,             # <<<<<<<<<<<<<<
 *                                                   Py_ssize_t *length):
 *     """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("pandas._libs.tslibs.util.get_c_string_buf_and_size", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/tslibs/util.pxd":256
 * 
 * 
 * cdef inline const char* get_c_string(object py_string):             # <<<<<<<<<<<<<<
 *     return get_c_string_buf_and_size(py_string, NULL)
 */

static CYTHON_INLINE char const *__pyx_f_6pandas_5_libs_6tslibs_4util_get_c_string(PyObject *__pyx_v_py_string) {
  char const *__pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_c_string", 0);

  /* "pandas/_libs/tslibs/util.pxd":257
 * 
 * cdef inline const char* get_c_string(object py_string):
 *     return get_c_string_buf_and_size(py_string, NULL)             # <<<<<<<<<<<<<<
 */
  __pyx_r = __pyx_f_6pandas_5_libs_6tslibs_4util_get_c_string_buf_and_size(__pyx_v_py_string, NULL);
  goto __pyx_L0;

  /* "pandas/_libs/tslibs/util.pxd":256
 * 
 * 
 * cdef inline const char* get_c_string(object py_string):             # <<<<<<<<<<<<<<
 *     return get_c_string_buf_and_size(py_string, NULL)
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/util.pxd":48
 * 
 * 
 * cdef inline void set_array_not_contiguous(ndarray ao) nogil:             # <<<<<<<<<<<<<<
 *     # Numpy>=1.8-compliant equivalent to:
 *     # ao->flags &= ~(NPY_ARRAY_C_CONTIGUOUS | NPY_ARRAY_F_CONTIGUOUS);
 */

static CYTHON_INLINE void __pyx_f_6pandas_5_libs_4util_set_array_not_contiguous(PyArrayObject *__pyx_v_ao) {

  /* "pandas/_libs/util.pxd":51
 *     # Numpy>=1.8-compliant equivalent to:
 *     # ao->flags &= ~(NPY_ARRAY_C_CONTIGUOUS | NPY_ARRAY_F_CONTIGUOUS);
 *     PyArray_CLEARFLAGS(ao,             # <<<<<<<<<<<<<<
 *                        (NPY_ARRAY_C_CONTIGUOUS | NPY_ARRAY_F_CONTIGUOUS))
 * 
 */
  PyArray_CLEARFLAGS(__pyx_v_ao, (NPY_ARRAY_C_CONTIGUOUS | NPY_ARRAY_F_CONTIGUOUS));

  /* "pandas/_libs/util.pxd":48
 * 
 * 
 * cdef inline void set_array_not_contiguous(ndarray ao) nogil:             # <<<<<<<<<<<<<<
 *     # Numpy>=1.8-compliant equivalent to:
 *     # ao->flags &= ~(NPY_ARRAY_C_CONTIGUOUS | NPY_ARRAY_F_CONTIGUOUS);
 */

  /* function exit code */
}

/* "pandas/_libs/util.pxd":55
 * 
 * 
 * cdef inline Py_ssize_t validate_indexer(ndarray arr, object loc) except -1:             # <<<<<<<<<<<<<<
 *     """
 *     Cast the given indexer `loc` to an integer.  If it is negative, i.e. a
 */

static CYTHON_INLINE Py_ssize_t __pyx_f_6pandas_5_libs_4util_validate_indexer(PyArrayObject *__pyx_v_arr, PyObject *__pyx_v_loc) {
  Py_ssize_t __pyx_v_idx;
  Py_ssize_t __pyx_v_size;
  int __pyx_v_casted;
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  Py_ssize_t __pyx_t_5;
  int __pyx_t_6;
  __Pyx_RefNannySetupContext("validate_indexer", 0);
  __Pyx_INCREF(__pyx_v_loc);

  /* "pandas/_libs/util.pxd":78
 *         int casted
 * 
 *     if is_float_object(loc):             # <<<<<<<<<<<<<<
 *         casted = int(loc)
 *         if casted == loc:
 */
  __pyx_t_1 = (__pyx_f_6pandas_5_libs_6tslibs_4util_is_float_object(__pyx_v_loc) != 0);
  if (__pyx_t_1) {

    /* "pandas/_libs/util.pxd":79
 * 
 *     if is_float_object(loc):
 *         casted = int(loc)             # <<<<<<<<<<<<<<
 *         if casted == loc:
 *             loc = casted
 */
    __pyx_t_2 = __Pyx_PyNumber_Int(__pyx_v_loc); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 79, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyInt_As_int(__pyx_t_2); if (unlikely((__pyx_t_3 == (int)-1) && PyErr_Occurred())) __PYX_ERR(5, 79, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_casted = __pyx_t_3;

    /* "pandas/_libs/util.pxd":80
 *     if is_float_object(loc):
 *         casted = int(loc)
 *         if casted == loc:             # <<<<<<<<<<<<<<
 *             loc = casted
 * 
 */
    __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_casted); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 80, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = PyObject_RichCompare(__pyx_t_2, __pyx_v_loc, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(5, 80, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(5, 80, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (__pyx_t_1) {

      /* "pandas/_libs/util.pxd":81
 *         casted = int(loc)
 *         if casted == loc:
 *             loc = casted             # <<<<<<<<<<<<<<
 * 
 *     idx = <Py_ssize_t>loc
 */
      __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_casted); if (unlikely(!__pyx_t_4)) __PYX_ERR(5, 81, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF_SET(__pyx_v_loc, __pyx_t_4);
      __pyx_t_4 = 0;

      /* "pandas/_libs/util.pxd":80
 *     if is_float_object(loc):
 *         casted = int(loc)
 *         if casted == loc:             # <<<<<<<<<<<<<<
 *             loc = casted
 * 
 */
    }

    /* "pandas/_libs/util.pxd":78
 *         int casted
 * 
 *     if is_float_object(loc):             # <<<<<<<<<<<<<<
 *         casted = int(loc)
 *         if casted == loc:
 */
  }

  /* "pandas/_libs/util.pxd":83
 *             loc = casted
 * 
 *     idx = <Py_ssize_t>loc             # <<<<<<<<<<<<<<
 *     size = cnp.PyArray_SIZE(arr)
 * 
 */
  __pyx_t_5 = __Pyx_PyIndex_AsSsize_t(__pyx_v_loc); if (unlikely((__pyx_t_5 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(5, 83, __pyx_L1_error)
  __pyx_v_idx = ((Py_ssize_t)__pyx_t_5);

  /* "pandas/_libs/util.pxd":84
 * 
 *     idx = <Py_ssize_t>loc
 *     size = cnp.PyArray_SIZE(arr)             # <<<<<<<<<<<<<<
 * 
 *     if idx < 0 and size > 0:
 */
  __pyx_v_size = PyArray_SIZE(__pyx_v_arr);

  /* "pandas/_libs/util.pxd":86
 *     size = cnp.PyArray_SIZE(arr)
 * 
 *     if idx < 0 and size > 0:             # <<<<<<<<<<<<<<
 *         idx += size
 *     if idx >= size or size == 0 or idx < 0:
 */
  __pyx_t_6 = ((__pyx_v_idx < 0) != 0);
  if (__pyx_t_6) {
  } else {
    __pyx_t_1 = __pyx_t_6;
    goto __pyx_L6_bool_binop_done;
  }
  __pyx_t_6 = ((__pyx_v_size > 0) != 0);
  __pyx_t_1 = __pyx_t_6;
  __pyx_L6_bool_binop_done:;
  if (__pyx_t_1) {

    /* "pandas/_libs/util.pxd":87
 * 
 *     if idx < 0 and size > 0:
 *         idx += size             # <<<<<<<<<<<<<<
 *     if idx >= size or size == 0 or idx < 0:
 *         raise IndexError('index out of bounds')
 */
    __pyx_v_idx = (__pyx_v_idx + __pyx_v_size);

    /* "pandas/_libs/util.pxd":86
 *     size = cnp.PyArray_SIZE(arr)
 * 
 *     if idx < 0 and size > 0:             # <<<<<<<<<<<<<<
 *         idx += size
 *     if idx >= size or size == 0 or idx < 0:
 */
  }

  /* "pandas/_libs/util.pxd":88
 *     if idx < 0 and size > 0:
 *         idx += size
 *     if idx >= size or size == 0 or idx < 0:             # <<<<<<<<<<<<<<
 *         raise IndexError('index out of bounds')
 * 
 */
  __pyx_t_6 = ((__pyx_v_idx >= __pyx_v_size) != 0);
  if (!__pyx_t_6) {
  } else {
    __pyx_t_1 = __pyx_t_6;
    goto __pyx_L9_bool_binop_done;
  }
  __pyx_t_6 = ((__pyx_v_size == 0) != 0);
  if (!__pyx_t_6) {
  } else {
    __pyx_t_1 = __pyx_t_6;
    goto __pyx_L9_bool_binop_done;
  }
  __pyx_t_6 = ((__pyx_v_idx < 0) != 0);
  __pyx_t_1 = __pyx_t_6;
  __pyx_L9_bool_binop_done:;
  if (unlikely(__pyx_t_1)) {

    /* "pandas/_libs/util.pxd":89
 *         idx += size
 *     if idx >= size or size == 0 or idx < 0:
 *         raise IndexError('index out of bounds')             # <<<<<<<<<<<<<<
 * 
 *     return idx
 */
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_IndexError, __pyx_tuple__26, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(5, 89, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __PYX_ERR(5, 89, __pyx_L1_error)

    /* "pandas/_libs/util.pxd":88
 *     if idx < 0 and size > 0:
 *         idx += size
 *     if idx >= size or size == 0 or idx < 0:             # <<<<<<<<<<<<<<
 *         raise IndexError('index out of bounds')
 * 
 */
  }

  /* "pandas/_libs/util.pxd":91
 *         raise IndexError('index out of bounds')
 * 
 *     return idx             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_idx;
  goto __pyx_L0;

  /* "pandas/_libs/util.pxd":55
 * 
 * 
 * cdef inline Py_ssize_t validate_indexer(ndarray arr, object loc) except -1:             # <<<<<<<<<<<<<<
 *     """
 *     Cast the given indexer `loc` to an integer.  If it is negative, i.e. a
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pandas._libs.util.validate_indexer", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1L;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_loc);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pandas/_libs/util.pxd":94
 * 
 * 
 * cdef inline object get_value_at(ndarray arr, object loc):             # <<<<<<<<<<<<<<
 *     cdef:
 *         Py_ssize_t i
 */

static CYTHON_INLINE PyObject *__pyx_f_6pandas_5_libs_4util_get_value_at(PyArrayObject *__pyx_v_arr, PyObject *__pyx_v_loc) {
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("get_value_at", 0);

  /* "pandas/_libs/util.pxd":98
 *         Py_ssize_t i
 * 
 *     i = validate_indexer(arr, loc)             # <<<<<<<<<<<<<<
 *     return arr[i]
 */
  __pyx_t_1 = __pyx_f_6pandas_5_libs_4util_validate_indexer(__pyx_v_arr, __pyx_v_loc); if (unlikely(__pyx_t_1 == ((Py_ssize_t)-1L))) __PYX_ERR(5, 98, __pyx_L1_error)
  __pyx_v_i = __pyx_t_1;

  /* "pandas/_libs/util.pxd":99
 * 
 *     i = validate_indexer(arr, loc)
 *     return arr[i]             # <<<<<<<<<<<<<<
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_GetItemInt(((PyObject *)__pyx_v_arr), __pyx_v_i, Py_ssize_t, 1, PyInt_FromSsize_t, 0, 1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 99, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pandas/_libs/util.pxd":94
 * 
 * 
 * cdef inline object get_value_at(ndarray arr, object loc):             # <<<<<<<<<<<<<<
 *     cdef:
 *         Py_ssize_t i
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pandas._libs.util.get_value_at", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":122
 *         cdef bint dtype_is_object
 * 
 *     def __cinit__(array self, tuple shape, Py_ssize_t itemsize, format not None,             # <<<<<<<<<<<<<<
 *                   mode="c", bint allocate_buffer=True):
 * 
 */

/* Python wrapper */
static int __pyx_array___cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_array___cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_shape = 0;
  Py_ssize_t __pyx_v_itemsize;
  PyObject *__pyx_v_format = 0;
  PyObject *__pyx_v_mode = 0;
  int __pyx_v_allocate_buffer;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_shape,&__pyx_n_s_itemsize,&__pyx_n_s_format,&__pyx_n_s_mode,&__pyx_n_s_allocate_buffer,0};
    PyObject* values[5] = {0,0,0,0,0};
    values[3] = ((PyObject *)__pyx_n_s_c);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_shape)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_itemsize)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 3, 5, 1); __PYX_ERR(2, 122, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_format)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 3, 5, 2); __PYX_ERR(2, 122, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_mode);
          if (value) { values[3] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_allocate_buffer);
          if (value) { values[4] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__cinit__") < 0)) __PYX_ERR(2, 122, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_shape = ((PyObject*)values[0]);
    __pyx_v_itemsize = __Pyx_PyIndex_AsSsize_t(values[1]); if (unlikely((__pyx_v_itemsize == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(2, 122, __pyx_L3_error)
    __pyx_v_format = values[2];
    __pyx_v_mode = values[3];
    if (values[4]) {
      __pyx_v_allocate_buffer = __Pyx_PyObject_IsTrue(values[4]); if (unlikely((__pyx_v_allocate_buffer == (int)-1) && PyErr_Occurred())) __PYX_ERR(2, 123, __pyx_L3_error)
    } else {

      /* "View.MemoryView":123
 * 
 *     def __cinit__(array self, tuple shape, Py_ssize_t itemsize, format not None,
 *                   mode="c", bint allocate_buffer=True):             # <<<<<<<<<<<<<<
 * 
 *         cdef int idx
 */
      __pyx_v_allocate_buffer = ((int)1);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 3, 5, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(2, 122, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("View.MemoryView.array.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_shape), (&PyTuple_Type), 1, "shape", 1))) __PYX_ERR(2, 122, __pyx_L1_error)
  if (unlikely(((PyObject *)__pyx_v_format) == Py_None)) {
    PyErr_Format(PyExc_TypeError, "Argument '%.200s' must not be None", "format"); __PYX_ERR(2, 122, __pyx_L1_error)
  }
  __pyx_r = __pyx_array___pyx_pf_15View_dot_MemoryView_5array___cinit__(((struct __pyx_array_obj *)__pyx_v_self), __pyx_v_shape, __pyx_v_itemsize, __pyx_v_format, __pyx_v_mode, __pyx_v_allocate_buffer);

  /* "View.MemoryView":122
 *         cdef bint dtype_is_object
 * 
 *     def __cinit__(array self, tuple shape, Py_ssize_t itemsize, format not None,             # <<<<<<<<<<<<<<
 *                   mode="c", bint allocate_buffer=True):
 * 
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_array___pyx_pf_15View_dot_MemoryView_5array___cinit__(struct __pyx_array_obj *__pyx_v_self, PyObject *__pyx_v_shape, Py_ssize_t __pyx_v_itemsize, PyObject *__pyx_v_format, PyObject *__pyx_v_mode, int __pyx_v_allocate_buffer) {
  int __pyx_v_idx;
  Py_ssize_t __pyx_v_i;
  Py_ssize_t __pyx_v_dim;
  PyObject **__pyx_v_p;
  char __pyx_v_order;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  char *__pyx_t_7;
  int __pyx_t_8;
  Py_ssize_t __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  Py_ssize_t __pyx_t_11;
  __Pyx_RefNannySetupContext("__cinit__", 0);
  __Pyx_INCREF(__pyx_v_format);

  /* "View.MemoryView":129
 *         cdef PyObject **p
 * 
 *         self.ndim = <int> len(shape)             # <<<<<<<<<<<<<<
 *         self.itemsize = itemsize
 * 
 */
  if (unlikely(__pyx_v_shape == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(2, 129, __pyx_L1_error)
  }
  __pyx_t_1 = PyTuple_GET_SIZE(__pyx_v_shape); if (unlikely(__pyx_t_1 == ((Py_ssize_t)-1))) __PYX_ERR(2, 129, __pyx_L1_error)
  __pyx_v_self->ndim = ((int)__pyx_t_1);

  /* "View.MemoryView":130
 * 
 *         self.ndim = <int> len(shape)
 *         self.itemsize = itemsize             # <<<<<<<<<<<<<<
 * 
 *         if not self.ndim:
 */
  __pyx_v_self->itemsize = __pyx_v_itemsize;

  /* "View.MemoryView":132
 *         self.itemsize = itemsize
 * 
 *         if not self.ndim:             # <<<<<<<<<<<<<<
 *             raise ValueError("Empty shape tuple for cython.array")
 * 
 */
  __pyx_t_2 = ((!(__pyx_v_self->ndim != 0)) != 0);
  if (unlikely(__pyx_t_2)) {

    /* "View.MemoryView":133
 * 
 *         if not self.ndim:
 *             raise ValueError("Empty shape tuple for cython.array")             # <<<<<<<<<<<<<<
 * 
 *         if itemsize <= 0:
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__27, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 133, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(2, 133, __pyx_L1_error)

    /* "View.MemoryView":132
 *         self.itemsize = itemsize
 * 
 *         if not self.ndim:             # <<<<<<<<<<<<<<
 *             raise ValueError("Empty shape tuple for cython.array")
 * 
 */
  }

  /* "View.MemoryView":135
 *             raise ValueError("Empty shape tuple for cython.array")
 * 
 *         if itemsize <= 0:             # <<<<<<<<<<<<<<
 *             raise ValueError("itemsize <= 0 for cython.array")
 * 
 */
  __pyx_t_2 = ((__pyx_v_itemsize <= 0) != 0);
  if (unlikely(__pyx_t_2)) {

    /* "View.MemoryView":136
 * 
 *         if itemsize <= 0:
 *             raise ValueError("itemsize <= 0 for cython.array")             # <<<<<<<<<<<<<<
 * 
 *         if not isinstance(format, bytes):
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__28, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 136, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(2, 136, __pyx_L1_error)

    /* "View.MemoryView":135
 *             raise ValueError("Empty shape tuple for cython.array")
 * 
 *         if itemsize <= 0:             # <<<<<<<<<<<<<<
 *             raise ValueError("itemsize <= 0 for cython.array")
 * 
 */
  }

  /* "View.MemoryView":138
 *             raise ValueError("itemsize <= 0 for cython.array")
 * 
 *         if not isinstance(format, bytes):             # <<<<<<<<<<<<<<
 *             format = format.encode('ASCII')
 *         self._format = format  # keep a reference to the byte string
 */
  __pyx_t_2 = PyBytes_Check(__pyx_v_format); 
  __pyx_t_4 = ((!(__pyx_t_2 != 0)) != 0);
  if (__pyx_t_4) {

    /* "View.MemoryView":139
 * 
 *         if not isinstance(format, bytes):
 *             format = format.encode('ASCII')             # <<<<<<<<<<<<<<
 *         self._format = format  # keep a reference to the byte string
 *         self.format = self._format
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_format, __pyx_n_s_encode); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 139, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_3 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_6, __pyx_n_s_ASCII) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_n_s_ASCII);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 139, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF_SET(__pyx_v_format, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "View.MemoryView":138
 *             raise ValueError("itemsize <= 0 for cython.array")
 * 
 *         if not isinstance(format, bytes):             # <<<<<<<<<<<<<<
 *             format = format.encode('ASCII')
 *         self._format = format  # keep a reference to the byte string
 */
  }

  /* "View.MemoryView":140
 *         if not isinstance(format, bytes):
 *             format = format.encode('ASCII')
 *         self._format = format  # keep a reference to the byte string             # <<<<<<<<<<<<<<
 *         self.format = self._format
 * 
 */
  if (!(likely(PyBytes_CheckExact(__pyx_v_format))||((__pyx_v_format) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_v_format)->tp_name), 0))) __PYX_ERR(2, 140, __pyx_L1_error)
  __pyx_t_3 = __pyx_v_format;
  __Pyx_INCREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __Pyx_GOTREF(__pyx_v_self->_format);
  __Pyx_DECREF(__pyx_v_self->_format);
  __pyx_v_self->_format = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "View.MemoryView":141
 *             format = format.encode('ASCII')
 *         self._format = format  # keep a reference to the byte string
 *         self.format = self._format             # <<<<<<<<<<<<<<
 * 
 * 
 */
  if (unlikely(__pyx_v_self->_format == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
    __PYX_ERR(2, 141, __pyx_L1_error)
  }
  __pyx_t_7 = __Pyx_PyBytes_AsWritableString(__pyx_v_self->_format); if (unlikely((!__pyx_t_7) && PyErr_Occurred())) __PYX_ERR(2, 141, __pyx_L1_error)
  __pyx_v_self->format = __pyx_t_7;

  /* "View.MemoryView":144
 * 
 * 
 *         self._shape = <Py_ssize_t *> PyObject_Malloc(sizeof(Py_ssize_t)*self.ndim*2)             # <<<<<<<<<<<<<<
 *         self._strides = self._shape + self.ndim
 * 
 */
  __pyx_v_self->_shape = ((Py_ssize_t *)PyObject_Malloc((((sizeof(Py_ssize_t)) * __pyx_v_self->ndim) * 2)));

  /* "View.MemoryView":145
 * 
 *         self._shape = <Py_ssize_t *> PyObject_Malloc(sizeof(Py_ssize_t)*self.ndim*2)
 *         self._strides = self._shape + self.ndim             # <<<<<<<<<<<<<<
 * 
 *         if not self._shape:
 */
  __pyx_v_self->_strides = (__pyx_v_self->_shape + __pyx_v_self->ndim);

  /* "View.MemoryView":147
 *         self._strides = self._shape + self.ndim
 * 
 *         if not self._shape:             # <<<<<<<<<<<<<<
 *             raise MemoryError("unable to allocate shape and strides.")
 * 
 */
  __pyx_t_4 = ((!(__pyx_v_self->_shape != 0)) != 0);
  if (unlikely(__pyx_t_4)) {

    /* "View.MemoryView":148
 * 
 *         if not self._shape:
 *             raise MemoryError("unable to allocate shape and strides.")             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_MemoryError, __pyx_tuple__29, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 148, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(2, 148, __pyx_L1_error)

    /* "View.MemoryView":147
 *         self._strides = self._shape + self.ndim
 * 
 *         if not self._shape:             # <<<<<<<<<<<<<<
 *             raise MemoryError("unable to allocate shape and strides.")
 * 
 */
  }

  /* "View.MemoryView":151
 * 
 * 
 *         for idx, dim in enumerate(shape):             # <<<<<<<<<<<<<<
 *             if dim <= 0:
 *                 raise ValueError("Invalid shape in axis %d: %d." % (idx, dim))
 */
  __pyx_t_8 = 0;
  __pyx_t_3 = __pyx_v_shape; __Pyx_INCREF(__pyx_t_3); __pyx_t_1 = 0;
  for (;;) {
    if (__pyx_t_1 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_5 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_1); __Pyx_INCREF(__pyx_t_5); __pyx_t_1++; if (unlikely(0 < 0)) __PYX_ERR(2, 151, __pyx_L1_error)
    #else
    __pyx_t_5 = PySequence_ITEM(__pyx_t_3, __pyx_t_1); __pyx_t_1++; if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 151, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    #endif
    __pyx_t_9 = __Pyx_PyIndex_AsSsize_t(__pyx_t_5); if (unlikely((__pyx_t_9 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(2, 151, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_v_dim = __pyx_t_9;
    __pyx_v_idx = __pyx_t_8;
    __pyx_t_8 = (__pyx_t_8 + 1);

    /* "View.MemoryView":152
 * 
 *         for idx, dim in enumerate(shape):
 *             if dim <= 0:             # <<<<<<<<<<<<<<
 *                 raise ValueError("Invalid shape in axis %d: %d." % (idx, dim))
 *             self._shape[idx] = dim
 */
    __pyx_t_4 = ((__pyx_v_dim <= 0) != 0);
    if (unlikely(__pyx_t_4)) {

      /* "View.MemoryView":153
 *         for idx, dim in enumerate(shape):
 *             if dim <= 0:
 *                 raise ValueError("Invalid shape in axis %d: %d." % (idx, dim))             # <<<<<<<<<<<<<<
 *             self._shape[idx] = dim
 * 
 */
      __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_idx); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 153, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = PyInt_FromSsize_t(__pyx_v_dim); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 153, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_10 = PyTuple_New(2); if (unlikely(!__pyx_t_10)) __PYX_ERR(2, 153, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_GIVEREF(__pyx_t_5);
      PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_5);
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_t_6);
      __pyx_t_5 = 0;
      __pyx_t_6 = 0;
      __pyx_t_6 = __Pyx_PyString_Format(__pyx_kp_s_Invalid_shape_in_axis_d_d, __pyx_t_10); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 153, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __pyx_t_10 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_6); if (unlikely(!__pyx_t_10)) __PYX_ERR(2, 153, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_Raise(__pyx_t_10, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __PYX_ERR(2, 153, __pyx_L1_error)

      /* "View.MemoryView":152
 * 
 *         for idx, dim in enumerate(shape):
 *             if dim <= 0:             # <<<<<<<<<<<<<<
 *                 raise ValueError("Invalid shape in axis %d: %d." % (idx, dim))
 *             self._shape[idx] = dim
 */
    }

    /* "View.MemoryView":154
 *             if dim <= 0:
 *                 raise ValueError("Invalid shape in axis %d: %d." % (idx, dim))
 *             self._shape[idx] = dim             # <<<<<<<<<<<<<<
 * 
 *         cdef char order
 */
    (__pyx_v_self->_shape[__pyx_v_idx]) = __pyx_v_dim;

    /* "View.MemoryView":151
 * 
 * 
 *         for idx, dim in enumerate(shape):             # <<<<<<<<<<<<<<
 *             if dim <= 0:
 *                 raise ValueError("Invalid shape in axis %d: %d." % (idx, dim))
 */
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "View.MemoryView":157
 * 
 *         cdef char order
 *         if mode == 'fortran':             # <<<<<<<<<<<<<<
 *             order = b'F'
 *             self.mode = u'fortran'
 */
  __pyx_t_4 = (__Pyx_PyString_Equals(__pyx_v_mode, __pyx_n_s_fortran, Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(2, 157, __pyx_L1_error)
  if (__pyx_t_4) {

    /* "View.MemoryView":158
 *         cdef char order
 *         if mode == 'fortran':
 *             order = b'F'             # <<<<<<<<<<<<<<
 *             self.mode = u'fortran'
 *         elif mode == 'c':
 */
    __pyx_v_order = 'F';

    /* "View.MemoryView":159
 *         if mode == 'fortran':
 *             order = b'F'
 *             self.mode = u'fortran'             # <<<<<<<<<<<<<<
 *         elif mode == 'c':
 *             order = b'C'
 */
    __Pyx_INCREF(__pyx_n_u_fortran);
    __Pyx_GIVEREF(__pyx_n_u_fortran);
    __Pyx_GOTREF(__pyx_v_self->mode);
    __Pyx_DECREF(__pyx_v_self->mode);
    __pyx_v_self->mode = __pyx_n_u_fortran;

    /* "View.MemoryView":157
 * 
 *         cdef char order
 *         if mode == 'fortran':             # <<<<<<<<<<<<<<
 *             order = b'F'
 *             self.mode = u'fortran'
 */
    goto __pyx_L10;
  }

  /* "View.MemoryView":160
 *             order = b'F'
 *             self.mode = u'fortran'
 *         elif mode == 'c':             # <<<<<<<<<<<<<<
 *             order = b'C'
 *             self.mode = u'c'
 */
  __pyx_t_4 = (__Pyx_PyString_Equals(__pyx_v_mode, __pyx_n_s_c, Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(2, 160, __pyx_L1_error)
  if (likely(__pyx_t_4)) {

    /* "View.MemoryView":161
 *             self.mode = u'fortran'
 *         elif mode == 'c':
 *             order = b'C'             # <<<<<<<<<<<<<<
 *             self.mode = u'c'
 *         else:
 */
    __pyx_v_order = 'C';

    /* "View.MemoryView":162
 *         elif mode == 'c':
 *             order = b'C'
 *             self.mode = u'c'             # <<<<<<<<<<<<<<
 *         else:
 *             raise ValueError("Invalid mode, expected 'c' or 'fortran', got %s" % mode)
 */
    __Pyx_INCREF(__pyx_n_u_c);
    __Pyx_GIVEREF(__pyx_n_u_c);
    __Pyx_GOTREF(__pyx_v_self->mode);
    __Pyx_DECREF(__pyx_v_self->mode);
    __pyx_v_self->mode = __pyx_n_u_c;

    /* "View.MemoryView":160
 *             order = b'F'
 *             self.mode = u'fortran'
 *         elif mode == 'c':             # <<<<<<<<<<<<<<
 *             order = b'C'
 *             self.mode = u'c'
 */
    goto __pyx_L10;
  }

  /* "View.MemoryView":164
 *             self.mode = u'c'
 *         else:
 *             raise ValueError("Invalid mode, expected 'c' or 'fortran', got %s" % mode)             # <<<<<<<<<<<<<<
 * 
 *         self.len = fill_contig_strides_array(self._shape, self._strides,
 */
  /*else*/ {
    __pyx_t_3 = __Pyx_PyString_FormatSafe(__pyx_kp_s_Invalid_mode_expected_c_or_fortr, __pyx_v_mode); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 164, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_10 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_3); if (unlikely(!__pyx_t_10)) __PYX_ERR(2, 164, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_t_10, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __PYX_ERR(2, 164, __pyx_L1_error)
  }
  __pyx_L10:;

  /* "View.MemoryView":166
 *             raise ValueError("Invalid mode, expected 'c' or 'fortran', got %s" % mode)
 * 
 *         self.len = fill_contig_strides_array(self._shape, self._strides,             # <<<<<<<<<<<<<<
 *                                              itemsize, self.ndim, order)
 * 
 */
  __pyx_v_self->len = __pyx_fill_contig_strides_array(__pyx_v_self->_shape, __pyx_v_self->_strides, __pyx_v_itemsize, __pyx_v_self->ndim, __pyx_v_order);

  /* "View.MemoryView":169
 *                                              itemsize, self.ndim, order)
 * 
 *         self.free_data = allocate_buffer             # <<<<<<<<<<<<<<
 *         self.dtype_is_object = format == b'O'
 *         if allocate_buffer:
 */
  __pyx_v_self->free_data = __pyx_v_allocate_buffer;

  /* "View.MemoryView":170
 * 
 *         self.free_data = allocate_buffer
 *         self.dtype_is_object = format == b'O'             # <<<<<<<<<<<<<<
 *         if allocate_buffer:
 * 
 */
  __pyx_t_10 = PyObject_RichCompare(__pyx_v_format, __pyx_n_b_O, Py_EQ); __Pyx_XGOTREF(__pyx_t_10); if (unlikely(!__pyx_t_10)) __PYX_ERR(2, 170, __pyx_L1_error)
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_10); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(2, 170, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __pyx_v_self->dtype_is_object = __pyx_t_4;

  /* "View.MemoryView":171
 *         self.free_data = allocate_buffer
 *         self.dtype_is_object = format == b'O'
 *         if allocate_buffer:             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_4 = (__pyx_v_allocate_buffer != 0);
  if (__pyx_t_4) {

    /* "View.MemoryView":174
 * 
 * 
 *             self.data = <char *>malloc(self.len)             # <<<<<<<<<<<<<<
 *             if not self.data:
 *                 raise MemoryError("unable to allocate array data.")
 */
    __pyx_v_self->data = ((char *)malloc(__pyx_v_self->len));

    /* "View.MemoryView":175
 * 
 *             self.data = <char *>malloc(self.len)
 *             if not self.data:             # <<<<<<<<<<<<<<
 *                 raise MemoryError("unable to allocate array data.")
 * 
 */
    __pyx_t_4 = ((!(__pyx_v_self->data != 0)) != 0);
    if (unlikely(__pyx_t_4)) {

      /* "View.MemoryView":176
 *             self.data = <char *>malloc(self.len)
 *             if not self.data:
 *                 raise MemoryError("unable to allocate array data.")             # <<<<<<<<<<<<<<
 * 
 *             if self.dtype_is_object:
 */
      __pyx_t_10 = __Pyx_PyObject_Call(__pyx_builtin_MemoryError, __pyx_tuple__30, NULL); if (unlikely(!__pyx_t_10)) __PYX_ERR(2, 176, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_Raise(__pyx_t_10, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __PYX_ERR(2, 176, __pyx_L1_error)

      /* "View.MemoryView":175
 * 
 *             self.data = <char *>malloc(self.len)
 *             if not self.data:             # <<<<<<<<<<<<<<
 *                 raise MemoryError("unable to allocate array data.")
 * 
 */
    }

    /* "View.MemoryView":178
 *                 raise MemoryError("unable to allocate array data.")
 * 
 *             if self.dtype_is_object:             # <<<<<<<<<<<<<<
 *                 p = <PyObject **> self.data
 *                 for i in range(self.len / itemsize):
 */
    __pyx_t_4 = (__pyx_v_self->dtype_is_object != 0);
    if (__pyx_t_4) {

      /* "View.MemoryView":179
 * 
 *             if self.dtype_is_object:
 *                 p = <PyObject **> self.data             # <<<<<<<<<<<<<<
 *                 for i in range(self.len / itemsize):
 *                     p[i] = Py_None
 */
      __pyx_v_p = ((PyObject **)__pyx_v_self->data);

      /* "View.MemoryView":180
 *             if self.dtype_is_object:
 *                 p = <PyObject **> self.data
 *                 for i in range(self.len / itemsize):             # <<<<<<<<<<<<<<
 *                     p[i] = Py_None
 *                     Py_INCREF(Py_None)
 */
      if (unlikely(__pyx_v_itemsize == 0)) {
        PyErr_SetString(PyExc_ZeroDivisionError, "integer division or modulo by zero");
        __PYX_ERR(2, 180, __pyx_L1_error)
      }
      else if (sizeof(Py_ssize_t) == sizeof(long) && (!(((Py_ssize_t)-1) > 0)) && unlikely(__pyx_v_itemsize == (Py_ssize_t)-1)  && unlikely(UNARY_NEG_WOULD_OVERFLOW(__pyx_v_self->len))) {
        PyErr_SetString(PyExc_OverflowError, "value too large to perform division");
        __PYX_ERR(2, 180, __pyx_L1_error)
      }
      __pyx_t_1 = __Pyx_div_Py_ssize_t(__pyx_v_self->len, __pyx_v_itemsize);
      __pyx_t_9 = __pyx_t_1;
      for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_9; __pyx_t_11+=1) {
        __pyx_v_i = __pyx_t_11;

        /* "View.MemoryView":181
 *                 p = <PyObject **> self.data
 *                 for i in range(self.len / itemsize):
 *                     p[i] = Py_None             # <<<<<<<<<<<<<<
 *                     Py_INCREF(Py_None)
 * 
 */
        (__pyx_v_p[__pyx_v_i]) = Py_None;

        /* "View.MemoryView":182
 *                 for i in range(self.len / itemsize):
 *                     p[i] = Py_None
 *                     Py_INCREF(Py_None)             # <<<<<<<<<<<<<<
 * 
 *     @cname('getbuffer')
 */
        Py_INCREF(Py_None);
      }

      /* "View.MemoryView":178
 *                 raise MemoryError("unable to allocate array data.")
 * 
 *             if self.dtype_is_object:             # <<<<<<<<<<<<<<
 *                 p = <PyObject **> self.data
 *                 for i in range(self.len / itemsize):
 */
    }

    /* "View.MemoryView":171
 *         self.free_data = allocate_buffer
 *         self.dtype_is_object = format == b'O'
 *         if allocate_buffer:             # <<<<<<<<<<<<<<
 * 
 * 
 */
  }

  /* "View.MemoryView":122
 *         cdef bint dtype_is_object
 * 
 *     def __cinit__(array self, tuple shape, Py_ssize_t itemsize, format not None,             # <<<<<<<<<<<<<<
 *                   mode="c", bint allocate_buffer=True):
 * 
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("View.MemoryView.array.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_format);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":185
 * 
 *     @cname('getbuffer')
 *     def __getbuffer__(self, Py_buffer *info, int flags):             # <<<<<<<<<<<<<<
 *         cdef int bufmode = -1
 *         if self.mode == u"c":
 */

/* Python wrapper */
static CYTHON_UNUSED int __pyx_array_getbuffer(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /*proto*/
static CYTHON_UNUSED int __pyx_array_getbuffer(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getbuffer__ (wrapper)", 0);
  __pyx_r = __pyx_array___pyx_pf_15View_dot_MemoryView_5array_2__getbuffer__(((struct __pyx_array_obj *)__pyx_v_self), ((Py_buffer *)__pyx_v_info), ((int)__pyx_v_flags));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_array___pyx_pf_15View_dot_MemoryView_5array_2__getbuffer__(struct __pyx_array_obj *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags) {
  int __pyx_v_bufmode;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  char *__pyx_t_4;
  Py_ssize_t __pyx_t_5;
  int __pyx_t_6;
  Py_ssize_t *__pyx_t_7;
  if (__pyx_v_info == NULL) {
    PyErr_SetString(PyExc_BufferError, "PyObject_GetBuffer: view==NULL argument is obsolete");
    return -1;
  }
  __Pyx_RefNannySetupContext("__getbuffer__", 0);
  __pyx_v_info->obj = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(__pyx_v_info->obj);

  /* "View.MemoryView":186
 *     @cname('getbuffer')
 *     def __getbuffer__(self, Py_buffer *info, int flags):
 *         cdef int bufmode = -1             # <<<<<<<<<<<<<<
 *         if self.mode == u"c":
 *             bufmode = PyBUF_C_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 */
  __pyx_v_bufmode = -1;

  /* "View.MemoryView":187
 *     def __getbuffer__(self, Py_buffer *info, int flags):
 *         cdef int bufmode = -1
 *         if self.mode == u"c":             # <<<<<<<<<<<<<<
 *             bufmode = PyBUF_C_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         elif self.mode == u"fortran":
 */
  __pyx_t_1 = (__Pyx_PyUnicode_Equals(__pyx_v_self->mode, __pyx_n_u_c, Py_EQ)); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(2, 187, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":188
 *         cdef int bufmode = -1
 *         if self.mode == u"c":
 *             bufmode = PyBUF_C_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS             # <<<<<<<<<<<<<<
 *         elif self.mode == u"fortran":
 *             bufmode = PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 */
    __pyx_v_bufmode = (PyBUF_C_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS);

    /* "View.MemoryView":187
 *     def __getbuffer__(self, Py_buffer *info, int flags):
 *         cdef int bufmode = -1
 *         if self.mode == u"c":             # <<<<<<<<<<<<<<
 *             bufmode = PyBUF_C_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         elif self.mode == u"fortran":
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":189
 *         if self.mode == u"c":
 *             bufmode = PyBUF_C_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         elif self.mode == u"fortran":             # <<<<<<<<<<<<<<
 *             bufmode = PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         if not (flags & bufmode):
 */
  __pyx_t_2 = (__Pyx_PyUnicode_Equals(__pyx_v_self->mode, __pyx_n_u_fortran, Py_EQ)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(2, 189, __pyx_L1_error)
  __pyx_t_1 = (__pyx_t_2 != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":190
 *             bufmode = PyBUF_C_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         elif self.mode == u"fortran":
 *             bufmode = PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS             # <<<<<<<<<<<<<<
 *         if not (flags & bufmode):
 *             raise ValueError("Can only create a buffer that is contiguous in memory.")
 */
    __pyx_v_bufmode = (PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS);

    /* "View.MemoryView":189
 *         if self.mode == u"c":
 *             bufmode = PyBUF_C_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         elif self.mode == u"fortran":             # <<<<<<<<<<<<<<
 *             bufmode = PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         if not (flags & bufmode):
 */
  }
  __pyx_L3:;

  /* "View.MemoryView":191
 *         elif self.mode == u"fortran":
 *             bufmode = PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         if not (flags & bufmode):             # <<<<<<<<<<<<<<
 *             raise ValueError("Can only create a buffer that is contiguous in memory.")
 *         info.buf = self.data
 */
  __pyx_t_1 = ((!((__pyx_v_flags & __pyx_v_bufmode) != 0)) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "View.MemoryView":192
 *             bufmode = PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         if not (flags & bufmode):
 *             raise ValueError("Can only create a buffer that is contiguous in memory.")             # <<<<<<<<<<<<<<
 *         info.buf = self.data
 *         info.len = self.len
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__31, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 192, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(2, 192, __pyx_L1_error)

    /* "View.MemoryView":191
 *         elif self.mode == u"fortran":
 *             bufmode = PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         if not (flags & bufmode):             # <<<<<<<<<<<<<<
 *             raise ValueError("Can only create a buffer that is contiguous in memory.")
 *         info.buf = self.data
 */
  }

  /* "View.MemoryView":193
 *         if not (flags & bufmode):
 *             raise ValueError("Can only create a buffer that is contiguous in memory.")
 *         info.buf = self.data             # <<<<<<<<<<<<<<
 *         info.len = self.len
 *         info.ndim = self.ndim
 */
  __pyx_t_4 = __pyx_v_self->data;
  __pyx_v_info->buf = __pyx_t_4;

  /* "View.MemoryView":194
 *             raise ValueError("Can only create a buffer that is contiguous in memory.")
 *         info.buf = self.data
 *         info.len = self.len             # <<<<<<<<<<<<<<
 *         info.ndim = self.ndim
 *         info.shape = self._shape
 */
  __pyx_t_5 = __pyx_v_self->len;
  __pyx_v_info->len = __pyx_t_5;

  /* "View.MemoryView":195
 *         info.buf = self.data
 *         info.len = self.len
 *         info.ndim = self.ndim             # <<<<<<<<<<<<<<
 *         info.shape = self._shape
 *         info.strides = self._strides
 */
  __pyx_t_6 = __pyx_v_self->ndim;
  __pyx_v_info->ndim = __pyx_t_6;

  /* "View.MemoryView":196
 *         info.len = self.len
 *         info.ndim = self.ndim
 *         info.shape = self._shape             # <<<<<<<<<<<<<<
 *         info.strides = self._strides
 *         info.suboffsets = NULL
 */
  __pyx_t_7 = __pyx_v_self->_shape;
  __pyx_v_info->shape = __pyx_t_7;

  /* "View.MemoryView":197
 *         info.ndim = self.ndim
 *         info.shape = self._shape
 *         info.strides = self._strides             # <<<<<<<<<<<<<<
 *         info.suboffsets = NULL
 *         info.itemsize = self.itemsize
 */
  __pyx_t_7 = __pyx_v_self->_strides;
  __pyx_v_info->strides = __pyx_t_7;

  /* "View.MemoryView":198
 *         info.shape = self._shape
 *         info.strides = self._strides
 *         info.suboffsets = NULL             # <<<<<<<<<<<<<<
 *         info.itemsize = self.itemsize
 *         info.readonly = 0
 */
  __pyx_v_info->suboffsets = NULL;

  /* "View.MemoryView":199
 *         info.strides = self._strides
 *         info.suboffsets = NULL
 *         info.itemsize = self.itemsize             # <<<<<<<<<<<<<<
 *         info.readonly = 0
 * 
 */
  __pyx_t_5 = __pyx_v_self->itemsize;
  __pyx_v_info->itemsize = __pyx_t_5;

  /* "View.MemoryView":200
 *         info.suboffsets = NULL
 *         info.itemsize = self.itemsize
 *         info.readonly = 0             # <<<<<<<<<<<<<<
 * 
 *         if flags & PyBUF_FORMAT:
 */
  __pyx_v_info->readonly = 0;

  /* "View.MemoryView":202
 *         info.readonly = 0
 * 
 *         if flags & PyBUF_FORMAT:             # <<<<<<<<<<<<<<
 *             info.format = self.format
 *         else:
 */
  __pyx_t_1 = ((__pyx_v_flags & PyBUF_FORMAT) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":203
 * 
 *         if flags & PyBUF_FORMAT:
 *             info.format = self.format             # <<<<<<<<<<<<<<
 *         else:
 *             info.format = NULL
 */
    __pyx_t_4 = __pyx_v_self->format;
    __pyx_v_info->format = __pyx_t_4;

    /* "View.MemoryView":202
 *         info.readonly = 0
 * 
 *         if flags & PyBUF_FORMAT:             # <<<<<<<<<<<<<<
 *             info.format = self.format
 *         else:
 */
    goto __pyx_L5;
  }

  /* "View.MemoryView":205
 *             info.format = self.format
 *         else:
 *             info.format = NULL             # <<<<<<<<<<<<<<
 * 
 *         info.obj = self
 */
  /*else*/ {
    __pyx_v_info->format = NULL;
  }
  __pyx_L5:;

  /* "View.MemoryView":207
 *             info.format = NULL
 * 
 *         info.obj = self             # <<<<<<<<<<<<<<
 * 
 *     __pyx_getbuffer = capsule(<void *> &__pyx_array_getbuffer, "getbuffer(obj, view, flags)")
 */
  __Pyx_INCREF(((PyObject *)__pyx_v_self));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
  __Pyx_GOTREF(__pyx_v_info->obj);
  __Pyx_DECREF(__pyx_v_info->obj);
  __pyx_v_info->obj = ((PyObject *)__pyx_v_self);

  /* "View.MemoryView":185
 * 
 *     @cname('getbuffer')
 *     def __getbuffer__(self, Py_buffer *info, int flags):             # <<<<<<<<<<<<<<
 *         cdef int bufmode = -1
 *         if self.mode == u"c":
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView.array.__getbuffer__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  if (__pyx_v_info->obj != NULL) {
    __Pyx_GOTREF(__pyx_v_info->obj);
    __Pyx_DECREF(__pyx_v_info->obj); __pyx_v_info->obj = 0;
  }
  goto __pyx_L2;
  __pyx_L0:;
  if (__pyx_v_info->obj == Py_None) {
    __Pyx_GOTREF(__pyx_v_info->obj);
    __Pyx_DECREF(__pyx_v_info->obj); __pyx_v_info->obj = 0;
  }
  __pyx_L2:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":211
 *     __pyx_getbuffer = capsule(<void *> &__pyx_array_getbuffer, "getbuffer(obj, view, flags)")
 * 
 *     def __dealloc__(array self):             # <<<<<<<<<<<<<<
 *         if self.callback_free_data != NULL:
 *             self.callback_free_data(self.data)
 */

/* Python wrapper */
static void __pyx_array___dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_array___dealloc__(PyObject *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_array___pyx_pf_15View_dot_MemoryView_5array_4__dealloc__(((struct __pyx_array_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_array___pyx_pf_15View_dot_MemoryView_5array_4__dealloc__(struct __pyx_array_obj *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "View.MemoryView":212
 * 
 *     def __dealloc__(array self):
 *         if self.callback_free_data != NULL:             # <<<<<<<<<<<<<<
 *             self.callback_free_data(self.data)
 *         elif self.free_data:
 */
  __pyx_t_1 = ((__pyx_v_self->callback_free_data != NULL) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":213
 *     def __dealloc__(array self):
 *         if self.callback_free_data != NULL:
 *             self.callback_free_data(self.data)             # <<<<<<<<<<<<<<
 *         elif self.free_data:
 *             if self.dtype_is_object:
 */
    __pyx_v_self->callback_free_data(__pyx_v_self->data);

    /* "View.MemoryView":212
 * 
 *     def __dealloc__(array self):
 *         if self.callback_free_data != NULL:             # <<<<<<<<<<<<<<
 *             self.callback_free_data(self.data)
 *         elif self.free_data:
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":214
 *         if self.callback_free_data != NULL:
 *             self.callback_free_data(self.data)
 *         elif self.free_data:             # <<<<<<<<<<<<<<
 *             if self.dtype_is_object:
 *                 refcount_objects_in_slice(self.data, self._shape,
 */
  __pyx_t_1 = (__pyx_v_self->free_data != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":215
 *             self.callback_free_data(self.data)
 *         elif self.free_data:
 *             if self.dtype_is_object:             # <<<<<<<<<<<<<<
 *                 refcount_objects_in_slice(self.data, self._shape,
 *                                           self._strides, self.ndim, False)
 */
    __pyx_t_1 = (__pyx_v_self->dtype_is_object != 0);
    if (__pyx_t_1) {

      /* "View.MemoryView":216
 *         elif self.free_data:
 *             if self.dtype_is_object:
 *                 refcount_objects_in_slice(self.data, self._shape,             # <<<<<<<<<<<<<<
 *                                           self._strides, self.ndim, False)
 *             free(self.data)
 */
      __pyx_memoryview_refcount_objects_in_slice(__pyx_v_self->data, __pyx_v_self->_shape, __pyx_v_self->_strides, __pyx_v_self->ndim, 0);

      /* "View.MemoryView":215
 *             self.callback_free_data(self.data)
 *         elif self.free_data:
 *             if self.dtype_is_object:             # <<<<<<<<<<<<<<
 *                 refcount_objects_in_slice(self.data, self._shape,
 *                                           self._strides, self.ndim, False)
 */
    }

    /* "View.MemoryView":218
 *                 refcount_objects_in_slice(self.data, self._shape,
 *                                           self._strides, self.ndim, False)
 *             free(self.data)             # <<<<<<<<<<<<<<
 *         PyObject_Free(self._shape)
 * 
 */
    free(__pyx_v_self->data);

    /* "View.MemoryView":214
 *         if self.callback_free_data != NULL:
 *             self.callback_free_data(self.data)
 *         elif self.free_data:             # <<<<<<<<<<<<<<
 *             if self.dtype_is_object:
 *                 refcount_objects_in_slice(self.data, self._shape,
 */
  }
  __pyx_L3:;

  /* "View.MemoryView":219
 *                                           self._strides, self.ndim, False)
 *             free(self.data)
 *         PyObject_Free(self._shape)             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  PyObject_Free(__pyx_v_self->_shape);

  /* "View.MemoryView":211
 *     __pyx_getbuffer = capsule(<void *> &__pyx_array_getbuffer, "getbuffer(obj, view, flags)")
 * 
 *     def __dealloc__(array self):             # <<<<<<<<<<<<<<
 *         if self.callback_free_data != NULL:
 *             self.callback_free_data(self.data)
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "View.MemoryView":222
 * 
 *     @property
 *     def memview(self):             # <<<<<<<<<<<<<<
 *         return self.get_memview()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_5array_7memview_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_5array_7memview_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_5array_7memview___get__(((struct __pyx_array_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_5array_7memview___get__(struct __pyx_array_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":223
 *     @property
 *     def memview(self):
 *         return self.get_memview()             # <<<<<<<<<<<<<<
 * 
 *     @cname('get_memview')
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((struct __pyx_vtabstruct_array *)__pyx_v_self->__pyx_vtab)->get_memview(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 223, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":222
 * 
 *     @property
 *     def memview(self):             # <<<<<<<<<<<<<<
 *         return self.get_memview()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.array.memview.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":226
 * 
 *     @cname('get_memview')
 *     cdef get_memview(self):             # <<<<<<<<<<<<<<
 *         flags =  PyBUF_ANY_CONTIGUOUS|PyBUF_FORMAT|PyBUF_WRITABLE
 *         return  memoryview(self, flags, self.dtype_is_object)
 */

static PyObject *__pyx_array_get_memview(struct __pyx_array_obj *__pyx_v_self) {
  int __pyx_v_flags;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("get_memview", 0);

  /* "View.MemoryView":227
 *     @cname('get_memview')
 *     cdef get_memview(self):
 *         flags =  PyBUF_ANY_CONTIGUOUS|PyBUF_FORMAT|PyBUF_WRITABLE             # <<<<<<<<<<<<<<
 *         return  memoryview(self, flags, self.dtype_is_object)
 * 
 */
  __pyx_v_flags = ((PyBUF_ANY_CONTIGUOUS | PyBUF_FORMAT) | PyBUF_WRITABLE);

  /* "View.MemoryView":228
 *     cdef get_memview(self):
 *         flags =  PyBUF_ANY_CONTIGUOUS|PyBUF_FORMAT|PyBUF_WRITABLE
 *         return  memoryview(self, flags, self.dtype_is_object)             # <<<<<<<<<<<<<<
 * 
 *     def __len__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_flags); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 228, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_v_self->dtype_is_object); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 228, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 228, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(((PyObject *)__pyx_v_self));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
  PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)__pyx_v_self));
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_t_2);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_memoryview_type), __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 228, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":226
 * 
 *     @cname('get_memview')
 *     cdef get_memview(self):             # <<<<<<<<<<<<<<
 *         flags =  PyBUF_ANY_CONTIGUOUS|PyBUF_FORMAT|PyBUF_WRITABLE
 *         return  memoryview(self, flags, self.dtype_is_object)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView.array.get_memview", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":230
 *         return  memoryview(self, flags, self.dtype_is_object)
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return self._shape[0]
 * 
 */

/* Python wrapper */
static Py_ssize_t __pyx_array___len__(PyObject *__pyx_v_self); /*proto*/
static Py_ssize_t __pyx_array___len__(PyObject *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__ (wrapper)", 0);
  __pyx_r = __pyx_array___pyx_pf_15View_dot_MemoryView_5array_6__len__(((struct __pyx_array_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static Py_ssize_t __pyx_array___pyx_pf_15View_dot_MemoryView_5array_6__len__(struct __pyx_array_obj *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__", 0);

  /* "View.MemoryView":231
 * 
 *     def __len__(self):
 *         return self._shape[0]             # <<<<<<<<<<<<<<
 * 
 *     def __getattr__(self, attr):
 */
  __pyx_r = (__pyx_v_self->_shape[0]);
  goto __pyx_L0;

  /* "View.MemoryView":230
 *         return  memoryview(self, flags, self.dtype_is_object)
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return self._shape[0]
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":233
 *         return self._shape[0]
 * 
 *     def __getattr__(self, attr):             # <<<<<<<<<<<<<<
 *         return getattr(self.memview, attr)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_array___getattr__(PyObject *__pyx_v_self, PyObject *__pyx_v_attr); /*proto*/
static PyObject *__pyx_array___getattr__(PyObject *__pyx_v_self, PyObject *__pyx_v_attr) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getattr__ (wrapper)", 0);
  __pyx_r = __pyx_array___pyx_pf_15View_dot_MemoryView_5array_8__getattr__(((struct __pyx_array_obj *)__pyx_v_self), ((PyObject *)__pyx_v_attr));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_array___pyx_pf_15View_dot_MemoryView_5array_8__getattr__(struct __pyx_array_obj *__pyx_v_self, PyObject *__pyx_v_attr) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("__getattr__", 0);

  /* "View.MemoryView":234
 * 
 *     def __getattr__(self, attr):
 *         return getattr(self.memview, attr)             # <<<<<<<<<<<<<<
 * 
 *     def __getitem__(self, item):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_memview); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 234, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_GetAttr(__pyx_t_1, __pyx_v_attr); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 234, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":233
 *         return self._shape[0]
 * 
 *     def __getattr__(self, attr):             # <<<<<<<<<<<<<<
 *         return getattr(self.memview, attr)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("View.MemoryView.array.__getattr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":236
 *         return getattr(self.memview, attr)
 * 
 *     def __getitem__(self, item):             # <<<<<<<<<<<<<<
 *         return self.memview[item]
 * 
 */

/* Python wrapper */
static PyObject *__pyx_array___getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_item); /*proto*/
static PyObject *__pyx_array___getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_item) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getitem__ (wrapper)", 0);
  __pyx_r = __pyx_array___pyx_pf_15View_dot_MemoryView_5array_10__getitem__(((struct __pyx_array_obj *)__pyx_v_self), ((PyObject *)__pyx_v_item));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_array___pyx_pf_15View_dot_MemoryView_5array_10__getitem__(struct __pyx_array_obj *__pyx_v_self, PyObject *__pyx_v_item) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("__getitem__", 0);

  /* "View.MemoryView":237
 * 
 *     def __getitem__(self, item):
 *         return self.memview[item]             # <<<<<<<<<<<<<<
 * 
 *     def __setitem__(self, item, value):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_memview); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 237, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetItem(__pyx_t_1, __pyx_v_item); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 237, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":236
 *         return getattr(self.memview, attr)
 * 
 *     def __getitem__(self, item):             # <<<<<<<<<<<<<<
 *         return self.memview[item]
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("View.MemoryView.array.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":239
 *         return self.memview[item]
 * 
 *     def __setitem__(self, item, value):             # <<<<<<<<<<<<<<
 *         self.memview[item] = value
 * 
 */

/* Python wrapper */
static int __pyx_array___setitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_item, PyObject *__pyx_v_value); /*proto*/
static int __pyx_array___setitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_item, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setitem__ (wrapper)", 0);
  __pyx_r = __pyx_array___pyx_pf_15View_dot_MemoryView_5array_12__setitem__(((struct __pyx_array_obj *)__pyx_v_self), ((PyObject *)__pyx_v_item), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_array___pyx_pf_15View_dot_MemoryView_5array_12__setitem__(struct __pyx_array_obj *__pyx_v_self, PyObject *__pyx_v_item, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setitem__", 0);

  /* "View.MemoryView":240
 * 
 *     def __setitem__(self, item, value):
 *         self.memview[item] = value             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_memview); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 240, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (unlikely(PyObject_SetItem(__pyx_t_1, __pyx_v_item, __pyx_v_value) < 0)) __PYX_ERR(2, 240, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "View.MemoryView":239
 *         return self.memview[item]
 * 
 *     def __setitem__(self, item, value):             # <<<<<<<<<<<<<<
 *         self.memview[item] = value
 * 
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.array.__setitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw___pyx_array_1__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw___pyx_array_1__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf___pyx_array___reduce_cython__(((struct __pyx_array_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf___pyx_array___reduce_cython__(CYTHON_UNUSED struct __pyx_array_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__32, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(2, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.array.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw___pyx_array_3__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw___pyx_array_3__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf___pyx_array_2__setstate_cython__(((struct __pyx_array_obj *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf___pyx_array_2__setstate_cython__(CYTHON_UNUSED struct __pyx_array_obj *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__33, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(2, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.array.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":244
 * 
 * @cname("__pyx_array_new")
 * cdef array array_cwrapper(tuple shape, Py_ssize_t itemsize, char *format,             # <<<<<<<<<<<<<<
 *                           char *mode, char *buf):
 *     cdef array result
 */

static struct __pyx_array_obj *__pyx_array_new(PyObject *__pyx_v_shape, Py_ssize_t __pyx_v_itemsize, char *__pyx_v_format, char *__pyx_v_mode, char *__pyx_v_buf) {
  struct __pyx_array_obj *__pyx_v_result = 0;
  struct __pyx_array_obj *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("array_cwrapper", 0);

  /* "View.MemoryView":248
 *     cdef array result
 * 
 *     if buf == NULL:             # <<<<<<<<<<<<<<
 *         result = array(shape, itemsize, format, mode.decode('ASCII'))
 *     else:
 */
  __pyx_t_1 = ((__pyx_v_buf == NULL) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":249
 * 
 *     if buf == NULL:
 *         result = array(shape, itemsize, format, mode.decode('ASCII'))             # <<<<<<<<<<<<<<
 *     else:
 *         result = array(shape, itemsize, format, mode.decode('ASCII'),
 */
    __pyx_t_2 = PyInt_FromSsize_t(__pyx_v_itemsize); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 249, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyBytes_FromString(__pyx_v_format); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 249, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_decode_c_string(__pyx_v_mode, 0, strlen(__pyx_v_mode), NULL, NULL, PyUnicode_DecodeASCII); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 249, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = PyTuple_New(4); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 249, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_INCREF(__pyx_v_shape);
    __Pyx_GIVEREF(__pyx_v_shape);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v_shape);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_5, 3, __pyx_t_4);
    __pyx_t_2 = 0;
    __pyx_t_3 = 0;
    __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyObject_Call(((PyObject *)__pyx_array_type), __pyx_t_5, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 249, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_v_result = ((struct __pyx_array_obj *)__pyx_t_4);
    __pyx_t_4 = 0;

    /* "View.MemoryView":248
 *     cdef array result
 * 
 *     if buf == NULL:             # <<<<<<<<<<<<<<
 *         result = array(shape, itemsize, format, mode.decode('ASCII'))
 *     else:
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":251
 *         result = array(shape, itemsize, format, mode.decode('ASCII'))
 *     else:
 *         result = array(shape, itemsize, format, mode.decode('ASCII'),             # <<<<<<<<<<<<<<
 *                        allocate_buffer=False)
 *         result.data = buf
 */
  /*else*/ {
    __pyx_t_4 = PyInt_FromSsize_t(__pyx_v_itemsize); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 251, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyBytes_FromString(__pyx_v_format); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 251, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_3 = __Pyx_decode_c_string(__pyx_v_mode, 0, strlen(__pyx_v_mode), NULL, NULL, PyUnicode_DecodeASCII); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 251, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = PyTuple_New(4); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 251, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_v_shape);
    __Pyx_GIVEREF(__pyx_v_shape);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_shape);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_2, 3, __pyx_t_3);
    __pyx_t_4 = 0;
    __pyx_t_5 = 0;
    __pyx_t_3 = 0;

    /* "View.MemoryView":252
 *     else:
 *         result = array(shape, itemsize, format, mode.decode('ASCII'),
 *                        allocate_buffer=False)             # <<<<<<<<<<<<<<
 *         result.data = buf
 * 
 */
    __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 252, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_allocate_buffer, Py_False) < 0) __PYX_ERR(2, 252, __pyx_L1_error)

    /* "View.MemoryView":251
 *         result = array(shape, itemsize, format, mode.decode('ASCII'))
 *     else:
 *         result = array(shape, itemsize, format, mode.decode('ASCII'),             # <<<<<<<<<<<<<<
 *                        allocate_buffer=False)
 *         result.data = buf
 */
    __pyx_t_5 = __Pyx_PyObject_Call(((PyObject *)__pyx_array_type), __pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 251, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_result = ((struct __pyx_array_obj *)__pyx_t_5);
    __pyx_t_5 = 0;

    /* "View.MemoryView":253
 *         result = array(shape, itemsize, format, mode.decode('ASCII'),
 *                        allocate_buffer=False)
 *         result.data = buf             # <<<<<<<<<<<<<<
 * 
 *     return result
 */
    __pyx_v_result->data = __pyx_v_buf;
  }
  __pyx_L3:;

  /* "View.MemoryView":255
 *         result.data = buf
 * 
 *     return result             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __Pyx_INCREF(((PyObject *)__pyx_v_result));
  __pyx_r = __pyx_v_result;
  goto __pyx_L0;

  /* "View.MemoryView":244
 * 
 * @cname("__pyx_array_new")
 * cdef array array_cwrapper(tuple shape, Py_ssize_t itemsize, char *format,             # <<<<<<<<<<<<<<
 *                           char *mode, char *buf):
 *     cdef array result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("View.MemoryView.array_cwrapper", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_result);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":281
 * cdef class Enum(object):
 *     cdef object name
 *     def __init__(self, name):             # <<<<<<<<<<<<<<
 *         self.name = name
 *     def __repr__(self):
 */

/* Python wrapper */
static int __pyx_MemviewEnum___init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_MemviewEnum___init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_name = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_name,0};
    PyObject* values[1] = {0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_name)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(2, 281, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 1) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
    }
    __pyx_v_name = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(2, 281, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("View.MemoryView.Enum.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_MemviewEnum___pyx_pf_15View_dot_MemoryView_4Enum___init__(((struct __pyx_MemviewEnum_obj *)__pyx_v_self), __pyx_v_name);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_MemviewEnum___pyx_pf_15View_dot_MemoryView_4Enum___init__(struct __pyx_MemviewEnum_obj *__pyx_v_self, PyObject *__pyx_v_name) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "View.MemoryView":282
 *     cdef object name
 *     def __init__(self, name):
 *         self.name = name             # <<<<<<<<<<<<<<
 *     def __repr__(self):
 *         return self.name
 */
  __Pyx_INCREF(__pyx_v_name);
  __Pyx_GIVEREF(__pyx_v_name);
  __Pyx_GOTREF(__pyx_v_self->name);
  __Pyx_DECREF(__pyx_v_self->name);
  __pyx_v_self->name = __pyx_v_name;

  /* "View.MemoryView":281
 * cdef class Enum(object):
 *     cdef object name
 *     def __init__(self, name):             # <<<<<<<<<<<<<<
 *         self.name = name
 *     def __repr__(self):
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":283
 *     def __init__(self, name):
 *         self.name = name
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return self.name
 * 
 */

/* Python wrapper */
static PyObject *__pyx_MemviewEnum___repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_MemviewEnum___repr__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_MemviewEnum___pyx_pf_15View_dot_MemoryView_4Enum_2__repr__(((struct __pyx_MemviewEnum_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_MemviewEnum___pyx_pf_15View_dot_MemoryView_4Enum_2__repr__(struct __pyx_MemviewEnum_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "View.MemoryView":284
 *         self.name = name
 *     def __repr__(self):
 *         return self.name             # <<<<<<<<<<<<<<
 * 
 * cdef generic = Enum("<strided and direct or indirect>")
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->name);
  __pyx_r = __pyx_v_self->name;
  goto __pyx_L0;

  /* "View.MemoryView":283
 *     def __init__(self, name):
 *         self.name = name
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return self.name
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw___pyx_MemviewEnum_1__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw___pyx_MemviewEnum_1__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf___pyx_MemviewEnum___reduce_cython__(((struct __pyx_MemviewEnum_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf___pyx_MemviewEnum___reduce_cython__(struct __pyx_MemviewEnum_obj *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.name,)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_self->name);
  __Pyx_GIVEREF(__pyx_v_self->name);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_self->name);
  __pyx_v_state = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.name,)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = (self.name,)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict);
    __pyx_t_4 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_4));
    __pyx_t_4 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.name is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.name,)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.name is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_Enum, (type(self), 0xb068931, None), state
 */
  /*else*/ {
    __pyx_t_3 = (__pyx_v_self->name != Py_None);
    __pyx_v_use_setstate = __pyx_t_3;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.name is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Enum, (type(self), 0xb068931, None), state
 *     else:
 */
  __pyx_t_3 = (__pyx_v_use_setstate != 0);
  if (__pyx_t_3) {

    /* "(tree fragment)":13
 *         use_setstate = self.name is not None
 *     if use_setstate:
 *         return __pyx_unpickle_Enum, (type(self), 0xb068931, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_Enum, (type(self), 0xb068931, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_pyx_unpickle_Enum); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_184977713);
    __Pyx_GIVEREF(__pyx_int_184977713);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_184977713);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state);
    __pyx_t_4 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.name is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Enum, (type(self), 0xb068931, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_Enum, (type(self), 0xb068931, None), state
 *     else:
 *         return __pyx_unpickle_Enum, (type(self), 0xb068931, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Enum__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_pyx_unpickle_Enum); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_184977713);
    __Pyx_GIVEREF(__pyx_int_184977713);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_184977713);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state);
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1);
    __pyx_t_5 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("View.MemoryView.Enum.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Enum, (type(self), 0xb068931, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Enum__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw___pyx_MemviewEnum_3__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw___pyx_MemviewEnum_3__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf___pyx_MemviewEnum_2__setstate_cython__(((struct __pyx_MemviewEnum_obj *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf___pyx_MemviewEnum_2__setstate_cython__(struct __pyx_MemviewEnum_obj *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_Enum, (type(self), 0xb068931, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Enum__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(2, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_unpickle_Enum__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Enum, (type(self), 0xb068931, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Enum__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.Enum.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":298
 * 
 * @cname('__pyx_align_pointer')
 * cdef void *align_pointer(void *memory, size_t alignment) nogil:             # <<<<<<<<<<<<<<
 *     "Align pointer memory on a given boundary"
 *     cdef Py_intptr_t aligned_p = <Py_intptr_t> memory
 */

static void *__pyx_align_pointer(void *__pyx_v_memory, size_t __pyx_v_alignment) {
  Py_intptr_t __pyx_v_aligned_p;
  size_t __pyx_v_offset;
  void *__pyx_r;
  int __pyx_t_1;

  /* "View.MemoryView":300
 * cdef void *align_pointer(void *memory, size_t alignment) nogil:
 *     "Align pointer memory on a given boundary"
 *     cdef Py_intptr_t aligned_p = <Py_intptr_t> memory             # <<<<<<<<<<<<<<
 *     cdef size_t offset
 * 
 */
  __pyx_v_aligned_p = ((Py_intptr_t)__pyx_v_memory);

  /* "View.MemoryView":304
 * 
 *     with cython.cdivision(True):
 *         offset = aligned_p % alignment             # <<<<<<<<<<<<<<
 * 
 *     if offset > 0:
 */
  __pyx_v_offset = (__pyx_v_aligned_p % __pyx_v_alignment);

  /* "View.MemoryView":306
 *         offset = aligned_p % alignment
 * 
 *     if offset > 0:             # <<<<<<<<<<<<<<
 *         aligned_p += alignment - offset
 * 
 */
  __pyx_t_1 = ((__pyx_v_offset > 0) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":307
 * 
 *     if offset > 0:
 *         aligned_p += alignment - offset             # <<<<<<<<<<<<<<
 * 
 *     return <void *> aligned_p
 */
    __pyx_v_aligned_p = (__pyx_v_aligned_p + (__pyx_v_alignment - __pyx_v_offset));

    /* "View.MemoryView":306
 *         offset = aligned_p % alignment
 * 
 *     if offset > 0:             # <<<<<<<<<<<<<<
 *         aligned_p += alignment - offset
 * 
 */
  }

  /* "View.MemoryView":309
 *         aligned_p += alignment - offset
 * 
 *     return <void *> aligned_p             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = ((void *)__pyx_v_aligned_p);
  goto __pyx_L0;

  /* "View.MemoryView":298
 * 
 * @cname('__pyx_align_pointer')
 * cdef void *align_pointer(void *memory, size_t alignment) nogil:             # <<<<<<<<<<<<<<
 *     "Align pointer memory on a given boundary"
 *     cdef Py_intptr_t aligned_p = <Py_intptr_t> memory
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "View.MemoryView":345
 *     cdef __Pyx_TypeInfo *typeinfo
 * 
 *     def __cinit__(memoryview self, object obj, int flags, bint dtype_is_object=False):             # <<<<<<<<<<<<<<
 *         self.obj = obj
 *         self.flags = flags
 */

/* Python wrapper */
static int __pyx_memoryview___cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_memoryview___cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_obj = 0;
  int __pyx_v_flags;
  int __pyx_v_dtype_is_object;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_obj,&__pyx_n_s_flags,&__pyx_n_s_dtype_is_object,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_obj)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_flags)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 2, 3, 1); __PYX_ERR(2, 345, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_dtype_is_object);
          if (value) { values[2] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__cinit__") < 0)) __PYX_ERR(2, 345, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_obj = values[0];
    __pyx_v_flags = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_flags == (int)-1) && PyErr_Occurred())) __PYX_ERR(2, 345, __pyx_L3_error)
    if (values[2]) {
      __pyx_v_dtype_is_object = __Pyx_PyObject_IsTrue(values[2]); if (unlikely((__pyx_v_dtype_is_object == (int)-1) && PyErr_Occurred())) __PYX_ERR(2, 345, __pyx_L3_error)
    } else {
      __pyx_v_dtype_is_object = ((int)0);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 2, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(2, 345, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("View.MemoryView.memoryview.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview___cinit__(((struct __pyx_memoryview_obj *)__pyx_v_self), __pyx_v_obj, __pyx_v_flags, __pyx_v_dtype_is_object);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview___cinit__(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_obj, int __pyx_v_flags, int __pyx_v_dtype_is_object) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "View.MemoryView":346
 * 
 *     def __cinit__(memoryview self, object obj, int flags, bint dtype_is_object=False):
 *         self.obj = obj             # <<<<<<<<<<<<<<
 *         self.flags = flags
 *         if type(self) is memoryview or obj is not None:
 */
  __Pyx_INCREF(__pyx_v_obj);
  __Pyx_GIVEREF(__pyx_v_obj);
  __Pyx_GOTREF(__pyx_v_self->obj);
  __Pyx_DECREF(__pyx_v_self->obj);
  __pyx_v_self->obj = __pyx_v_obj;

  /* "View.MemoryView":347
 *     def __cinit__(memoryview self, object obj, int flags, bint dtype_is_object=False):
 *         self.obj = obj
 *         self.flags = flags             # <<<<<<<<<<<<<<
 *         if type(self) is memoryview or obj is not None:
 *             __Pyx_GetBuffer(obj, &self.view, flags)
 */
  __pyx_v_self->flags = __pyx_v_flags;

  /* "View.MemoryView":348
 *         self.obj = obj
 *         self.flags = flags
 *         if type(self) is memoryview or obj is not None:             # <<<<<<<<<<<<<<
 *             __Pyx_GetBuffer(obj, &self.view, flags)
 *             if <PyObject *> self.view.obj == NULL:
 */
  __pyx_t_2 = (((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))) == ((PyObject *)__pyx_memoryview_type));
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (!__pyx_t_3) {
  } else {
    __pyx_t_1 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = (__pyx_v_obj != Py_None);
  __pyx_t_2 = (__pyx_t_3 != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "View.MemoryView":349
 *         self.flags = flags
 *         if type(self) is memoryview or obj is not None:
 *             __Pyx_GetBuffer(obj, &self.view, flags)             # <<<<<<<<<<<<<<
 *             if <PyObject *> self.view.obj == NULL:
 *                 (<__pyx_buffer *> &self.view).obj = Py_None
 */
    __pyx_t_4 = __Pyx_GetBuffer(__pyx_v_obj, (&__pyx_v_self->view), __pyx_v_flags); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(2, 349, __pyx_L1_error)

    /* "View.MemoryView":350
 *         if type(self) is memoryview or obj is not None:
 *             __Pyx_GetBuffer(obj, &self.view, flags)
 *             if <PyObject *> self.view.obj == NULL:             # <<<<<<<<<<<<<<
 *                 (<__pyx_buffer *> &self.view).obj = Py_None
 *                 Py_INCREF(Py_None)
 */
    __pyx_t_1 = ((((PyObject *)__pyx_v_self->view.obj) == NULL) != 0);
    if (__pyx_t_1) {

      /* "View.MemoryView":351
 *             __Pyx_GetBuffer(obj, &self.view, flags)
 *             if <PyObject *> self.view.obj == NULL:
 *                 (<__pyx_buffer *> &self.view).obj = Py_None             # <<<<<<<<<<<<<<
 *                 Py_INCREF(Py_None)
 * 
 */
      ((Py_buffer *)(&__pyx_v_self->view))->obj = Py_None;

      /* "View.MemoryView":352
 *             if <PyObject *> self.view.obj == NULL:
 *                 (<__pyx_buffer *> &self.view).obj = Py_None
 *                 Py_INCREF(Py_None)             # <<<<<<<<<<<<<<
 * 
 *         global __pyx_memoryview_thread_locks_used
 */
      Py_INCREF(Py_None);

      /* "View.MemoryView":350
 *         if type(self) is memoryview or obj is not None:
 *             __Pyx_GetBuffer(obj, &self.view, flags)
 *             if <PyObject *> self.view.obj == NULL:             # <<<<<<<<<<<<<<
 *                 (<__pyx_buffer *> &self.view).obj = Py_None
 *                 Py_INCREF(Py_None)
 */
    }

    /* "View.MemoryView":348
 *         self.obj = obj
 *         self.flags = flags
 *         if type(self) is memoryview or obj is not None:             # <<<<<<<<<<<<<<
 *             __Pyx_GetBuffer(obj, &self.view, flags)
 *             if <PyObject *> self.view.obj == NULL:
 */
  }

  /* "View.MemoryView":355
 * 
 *         global __pyx_memoryview_thread_locks_used
 *         if __pyx_memoryview_thread_locks_used < THREAD_LOCKS_PREALLOCATED:             # <<<<<<<<<<<<<<
 *             self.lock = __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]
 *             __pyx_memoryview_thread_locks_used += 1
 */
  __pyx_t_1 = ((__pyx_memoryview_thread_locks_used < 8) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":356
 *         global __pyx_memoryview_thread_locks_used
 *         if __pyx_memoryview_thread_locks_used < THREAD_LOCKS_PREALLOCATED:
 *             self.lock = __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]             # <<<<<<<<<<<<<<
 *             __pyx_memoryview_thread_locks_used += 1
 *         if self.lock is NULL:
 */
    __pyx_v_self->lock = (__pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]);

    /* "View.MemoryView":357
 *         if __pyx_memoryview_thread_locks_used < THREAD_LOCKS_PREALLOCATED:
 *             self.lock = __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]
 *             __pyx_memoryview_thread_locks_used += 1             # <<<<<<<<<<<<<<
 *         if self.lock is NULL:
 *             self.lock = PyThread_allocate_lock()
 */
    __pyx_memoryview_thread_locks_used = (__pyx_memoryview_thread_locks_used + 1);

    /* "View.MemoryView":355
 * 
 *         global __pyx_memoryview_thread_locks_used
 *         if __pyx_memoryview_thread_locks_used < THREAD_LOCKS_PREALLOCATED:             # <<<<<<<<<<<<<<
 *             self.lock = __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]
 *             __pyx_memoryview_thread_locks_used += 1
 */
  }

  /* "View.MemoryView":358
 *             self.lock = __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]
 *             __pyx_memoryview_thread_locks_used += 1
 *         if self.lock is NULL:             # <<<<<<<<<<<<<<
 *             self.lock = PyThread_allocate_lock()
 *             if self.lock is NULL:
 */
  __pyx_t_1 = ((__pyx_v_self->lock == NULL) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":359
 *             __pyx_memoryview_thread_locks_used += 1
 *         if self.lock is NULL:
 *             self.lock = PyThread_allocate_lock()             # <<<<<<<<<<<<<<
 *             if self.lock is NULL:
 *                 raise MemoryError
 */
    __pyx_v_self->lock = PyThread_allocate_lock();

    /* "View.MemoryView":360
 *         if self.lock is NULL:
 *             self.lock = PyThread_allocate_lock()
 *             if self.lock is NULL:             # <<<<<<<<<<<<<<
 *                 raise MemoryError
 * 
 */
    __pyx_t_1 = ((__pyx_v_self->lock == NULL) != 0);
    if (unlikely(__pyx_t_1)) {

      /* "View.MemoryView":361
 *             self.lock = PyThread_allocate_lock()
 *             if self.lock is NULL:
 *                 raise MemoryError             # <<<<<<<<<<<<<<
 * 
 *         if flags & PyBUF_FORMAT:
 */
      PyErr_NoMemory(); __PYX_ERR(2, 361, __pyx_L1_error)

      /* "View.MemoryView":360
 *         if self.lock is NULL:
 *             self.lock = PyThread_allocate_lock()
 *             if self.lock is NULL:             # <<<<<<<<<<<<<<
 *                 raise MemoryError
 * 
 */
    }

    /* "View.MemoryView":358
 *             self.lock = __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]
 *             __pyx_memoryview_thread_locks_used += 1
 *         if self.lock is NULL:             # <<<<<<<<<<<<<<
 *             self.lock = PyThread_allocate_lock()
 *             if self.lock is NULL:
 */
  }

  /* "View.MemoryView":363
 *                 raise MemoryError
 * 
 *         if flags & PyBUF_FORMAT:             # <<<<<<<<<<<<<<
 *             self.dtype_is_object = (self.view.format[0] == b'O' and self.view.format[1] == b'\0')
 *         else:
 */
  __pyx_t_1 = ((__pyx_v_flags & PyBUF_FORMAT) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":364
 * 
 *         if flags & PyBUF_FORMAT:
 *             self.dtype_is_object = (self.view.format[0] == b'O' and self.view.format[1] == b'\0')             # <<<<<<<<<<<<<<
 *         else:
 *             self.dtype_is_object = dtype_is_object
 */
    __pyx_t_2 = (((__pyx_v_self->view.format[0]) == 'O') != 0);
    if (__pyx_t_2) {
    } else {
      __pyx_t_1 = __pyx_t_2;
      goto __pyx_L11_bool_binop_done;
    }
    __pyx_t_2 = (((__pyx_v_self->view.format[1]) == '\x00') != 0);
    __pyx_t_1 = __pyx_t_2;
    __pyx_L11_bool_binop_done:;
    __pyx_v_self->dtype_is_object = __pyx_t_1;

    /* "View.MemoryView":363
 *                 raise MemoryError
 * 
 *         if flags & PyBUF_FORMAT:             # <<<<<<<<<<<<<<
 *             self.dtype_is_object = (self.view.format[0] == b'O' and self.view.format[1] == b'\0')
 *         else:
 */
    goto __pyx_L10;
  }

  /* "View.MemoryView":366
 *             self.dtype_is_object = (self.view.format[0] == b'O' and self.view.format[1] == b'\0')
 *         else:
 *             self.dtype_is_object = dtype_is_object             # <<<<<<<<<<<<<<
 * 
 *         self.acquisition_count_aligned_p = <__pyx_atomic_int *> align_pointer(
 */
  /*else*/ {
    __pyx_v_self->dtype_is_object = __pyx_v_dtype_is_object;
  }
  __pyx_L10:;

  /* "View.MemoryView":368
 *             self.dtype_is_object = dtype_is_object
 * 
 *         self.acquisition_count_aligned_p = <__pyx_atomic_int *> align_pointer(             # <<<<<<<<<<<<<<
 *                   <void *> &self.acquisition_count[0], sizeof(__pyx_atomic_int))
 *         self.typeinfo = NULL
 */
  __pyx_v_self->acquisition_count_aligned_p = ((__pyx_atomic_int *)__pyx_align_pointer(((void *)(&(__pyx_v_self->acquisition_count[0]))), (sizeof(__pyx_atomic_int))));

  /* "View.MemoryView":370
 *         self.acquisition_count_aligned_p = <__pyx_atomic_int *> align_pointer(
 *                   <void *> &self.acquisition_count[0], sizeof(__pyx_atomic_int))
 *         self.typeinfo = NULL             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(memoryview self):
 */
  __pyx_v_self->typeinfo = NULL;

  /* "View.MemoryView":345
 *     cdef __Pyx_TypeInfo *typeinfo
 * 
 *     def __cinit__(memoryview self, object obj, int flags, bint dtype_is_object=False):             # <<<<<<<<<<<<<<
 *         self.obj = obj
 *         self.flags = flags
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("View.MemoryView.memoryview.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":372
 *         self.typeinfo = NULL
 * 
 *     def __dealloc__(memoryview self):             # <<<<<<<<<<<<<<
 *         if self.obj is not None:
 *             __Pyx_ReleaseBuffer(&self.view)
 */

/* Python wrapper */
static void __pyx_memoryview___dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_memoryview___dealloc__(PyObject *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_2__dealloc__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_2__dealloc__(struct __pyx_memoryview_obj *__pyx_v_self) {
  int __pyx_v_i;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  PyThread_type_lock __pyx_t_6;
  PyThread_type_lock __pyx_t_7;
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "View.MemoryView":373
 * 
 *     def __dealloc__(memoryview self):
 *         if self.obj is not None:             # <<<<<<<<<<<<<<
 *             __Pyx_ReleaseBuffer(&self.view)
 *         elif (<__pyx_buffer *> &self.view).obj == Py_None:
 */
  __pyx_t_1 = (__pyx_v_self->obj != Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":374
 *     def __dealloc__(memoryview self):
 *         if self.obj is not None:
 *             __Pyx_ReleaseBuffer(&self.view)             # <<<<<<<<<<<<<<
 *         elif (<__pyx_buffer *> &self.view).obj == Py_None:
 * 
 */
    __Pyx_ReleaseBuffer((&__pyx_v_self->view));

    /* "View.MemoryView":373
 * 
 *     def __dealloc__(memoryview self):
 *         if self.obj is not None:             # <<<<<<<<<<<<<<
 *             __Pyx_ReleaseBuffer(&self.view)
 *         elif (<__pyx_buffer *> &self.view).obj == Py_None:
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":375
 *         if self.obj is not None:
 *             __Pyx_ReleaseBuffer(&self.view)
 *         elif (<__pyx_buffer *> &self.view).obj == Py_None:             # <<<<<<<<<<<<<<
 * 
 *             (<__pyx_buffer *> &self.view).obj = NULL
 */
  __pyx_t_2 = ((((Py_buffer *)(&__pyx_v_self->view))->obj == Py_None) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":377
 *         elif (<__pyx_buffer *> &self.view).obj == Py_None:
 * 
 *             (<__pyx_buffer *> &self.view).obj = NULL             # <<<<<<<<<<<<<<
 *             Py_DECREF(Py_None)
 * 
 */
    ((Py_buffer *)(&__pyx_v_self->view))->obj = NULL;

    /* "View.MemoryView":378
 * 
 *             (<__pyx_buffer *> &self.view).obj = NULL
 *             Py_DECREF(Py_None)             # <<<<<<<<<<<<<<
 * 
 *         cdef int i
 */
    Py_DECREF(Py_None);

    /* "View.MemoryView":375
 *         if self.obj is not None:
 *             __Pyx_ReleaseBuffer(&self.view)
 *         elif (<__pyx_buffer *> &self.view).obj == Py_None:             # <<<<<<<<<<<<<<
 * 
 *             (<__pyx_buffer *> &self.view).obj = NULL
 */
  }
  __pyx_L3:;

  /* "View.MemoryView":382
 *         cdef int i
 *         global __pyx_memoryview_thread_locks_used
 *         if self.lock != NULL:             # <<<<<<<<<<<<<<
 *             for i in range(__pyx_memoryview_thread_locks_used):
 *                 if __pyx_memoryview_thread_locks[i] is self.lock:
 */
  __pyx_t_2 = ((__pyx_v_self->lock != NULL) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":383
 *         global __pyx_memoryview_thread_locks_used
 *         if self.lock != NULL:
 *             for i in range(__pyx_memoryview_thread_locks_used):             # <<<<<<<<<<<<<<
 *                 if __pyx_memoryview_thread_locks[i] is self.lock:
 *                     __pyx_memoryview_thread_locks_used -= 1
 */
    __pyx_t_3 = __pyx_memoryview_thread_locks_used;
    __pyx_t_4 = __pyx_t_3;
    for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_4; __pyx_t_5+=1) {
      __pyx_v_i = __pyx_t_5;

      /* "View.MemoryView":384
 *         if self.lock != NULL:
 *             for i in range(__pyx_memoryview_thread_locks_used):
 *                 if __pyx_memoryview_thread_locks[i] is self.lock:             # <<<<<<<<<<<<<<
 *                     __pyx_memoryview_thread_locks_used -= 1
 *                     if i != __pyx_memoryview_thread_locks_used:
 */
      __pyx_t_2 = (((__pyx_memoryview_thread_locks[__pyx_v_i]) == __pyx_v_self->lock) != 0);
      if (__pyx_t_2) {

        /* "View.MemoryView":385
 *             for i in range(__pyx_memoryview_thread_locks_used):
 *                 if __pyx_memoryview_thread_locks[i] is self.lock:
 *                     __pyx_memoryview_thread_locks_used -= 1             # <<<<<<<<<<<<<<
 *                     if i != __pyx_memoryview_thread_locks_used:
 *                         __pyx_memoryview_thread_locks[i], __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used] = (
 */
        __pyx_memoryview_thread_locks_used = (__pyx_memoryview_thread_locks_used - 1);

        /* "View.MemoryView":386
 *                 if __pyx_memoryview_thread_locks[i] is self.lock:
 *                     __pyx_memoryview_thread_locks_used -= 1
 *                     if i != __pyx_memoryview_thread_locks_used:             # <<<<<<<<<<<<<<
 *                         __pyx_memoryview_thread_locks[i], __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used] = (
 *                             __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used], __pyx_memoryview_thread_locks[i])
 */
        __pyx_t_2 = ((__pyx_v_i != __pyx_memoryview_thread_locks_used) != 0);
        if (__pyx_t_2) {

          /* "View.MemoryView":388
 *                     if i != __pyx_memoryview_thread_locks_used:
 *                         __pyx_memoryview_thread_locks[i], __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used] = (
 *                             __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used], __pyx_memoryview_thread_locks[i])             # <<<<<<<<<<<<<<
 *                     break
 *             else:
 */
          __pyx_t_6 = (__pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]);
          __pyx_t_7 = (__pyx_memoryview_thread_locks[__pyx_v_i]);

          /* "View.MemoryView":387
 *                     __pyx_memoryview_thread_locks_used -= 1
 *                     if i != __pyx_memoryview_thread_locks_used:
 *                         __pyx_memoryview_thread_locks[i], __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used] = (             # <<<<<<<<<<<<<<
 *                             __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used], __pyx_memoryview_thread_locks[i])
 *                     break
 */
          (__pyx_memoryview_thread_locks[__pyx_v_i]) = __pyx_t_6;
          (__pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]) = __pyx_t_7;

          /* "View.MemoryView":386
 *                 if __pyx_memoryview_thread_locks[i] is self.lock:
 *                     __pyx_memoryview_thread_locks_used -= 1
 *                     if i != __pyx_memoryview_thread_locks_used:             # <<<<<<<<<<<<<<
 *                         __pyx_memoryview_thread_locks[i], __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used] = (
 *                             __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used], __pyx_memoryview_thread_locks[i])
 */
        }

        /* "View.MemoryView":389
 *                         __pyx_memoryview_thread_locks[i], __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used] = (
 *                             __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used], __pyx_memoryview_thread_locks[i])
 *                     break             # <<<<<<<<<<<<<<
 *             else:
 *                 PyThread_free_lock(self.lock)
 */
        goto __pyx_L6_break;

        /* "View.MemoryView":384
 *         if self.lock != NULL:
 *             for i in range(__pyx_memoryview_thread_locks_used):
 *                 if __pyx_memoryview_thread_locks[i] is self.lock:             # <<<<<<<<<<<<<<
 *                     __pyx_memoryview_thread_locks_used -= 1
 *                     if i != __pyx_memoryview_thread_locks_used:
 */
      }
    }
    /*else*/ {

      /* "View.MemoryView":391
 *                     break
 *             else:
 *                 PyThread_free_lock(self.lock)             # <<<<<<<<<<<<<<
 * 
 *     cdef char *get_item_pointer(memoryview self, object index) except NULL:
 */
      PyThread_free_lock(__pyx_v_self->lock);
    }
    __pyx_L6_break:;

    /* "View.MemoryView":382
 *         cdef int i
 *         global __pyx_memoryview_thread_locks_used
 *         if self.lock != NULL:             # <<<<<<<<<<<<<<
 *             for i in range(__pyx_memoryview_thread_locks_used):
 *                 if __pyx_memoryview_thread_locks[i] is self.lock:
 */
  }

  /* "View.MemoryView":372
 *         self.typeinfo = NULL
 * 
 *     def __dealloc__(memoryview self):             # <<<<<<<<<<<<<<
 *         if self.obj is not None:
 *             __Pyx_ReleaseBuffer(&self.view)
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "View.MemoryView":393
 *                 PyThread_free_lock(self.lock)
 * 
 *     cdef char *get_item_pointer(memoryview self, object index) except NULL:             # <<<<<<<<<<<<<<
 *         cdef Py_ssize_t dim
 *         cdef char *itemp = <char *> self.view.buf
 */

static char *__pyx_memoryview_get_item_pointer(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_index) {
  Py_ssize_t __pyx_v_dim;
  char *__pyx_v_itemp;
  PyObject *__pyx_v_idx = NULL;
  char *__pyx_r;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  PyObject *(*__pyx_t_4)(PyObject *);
  PyObject *__pyx_t_5 = NULL;
  Py_ssize_t __pyx_t_6;
  char *__pyx_t_7;
  __Pyx_RefNannySetupContext("get_item_pointer", 0);

  /* "View.MemoryView":395
 *     cdef char *get_item_pointer(memoryview self, object index) except NULL:
 *         cdef Py_ssize_t dim
 *         cdef char *itemp = <char *> self.view.buf             # <<<<<<<<<<<<<<
 * 
 *         for dim, idx in enumerate(index):
 */
  __pyx_v_itemp = ((char *)__pyx_v_self->view.buf);

  /* "View.MemoryView":397
 *         cdef char *itemp = <char *> self.view.buf
 * 
 *         for dim, idx in enumerate(index):             # <<<<<<<<<<<<<<
 *             itemp = pybuffer_index(&self.view, itemp, idx, dim)
 * 
 */
  __pyx_t_1 = 0;
  if (likely(PyList_CheckExact(__pyx_v_index)) || PyTuple_CheckExact(__pyx_v_index)) {
    __pyx_t_2 = __pyx_v_index; __Pyx_INCREF(__pyx_t_2); __pyx_t_3 = 0;
    __pyx_t_4 = NULL;
  } else {
    __pyx_t_3 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_v_index); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 397, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 397, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_4)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_3 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_5 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_5); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(2, 397, __pyx_L1_error)
        #else
        __pyx_t_5 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 397, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        #endif
      } else {
        if (__pyx_t_3 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_5 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_5); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(2, 397, __pyx_L1_error)
        #else
        __pyx_t_5 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 397, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        #endif
      }
    } else {
      __pyx_t_5 = __pyx_t_4(__pyx_t_2);
      if (unlikely(!__pyx_t_5)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(2, 397, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_5);
    }
    __Pyx_XDECREF_SET(__pyx_v_idx, __pyx_t_5);
    __pyx_t_5 = 0;
    __pyx_v_dim = __pyx_t_1;
    __pyx_t_1 = (__pyx_t_1 + 1);

    /* "View.MemoryView":398
 * 
 *         for dim, idx in enumerate(index):
 *             itemp = pybuffer_index(&self.view, itemp, idx, dim)             # <<<<<<<<<<<<<<
 * 
 *         return itemp
 */
    __pyx_t_6 = __Pyx_PyIndex_AsSsize_t(__pyx_v_idx); if (unlikely((__pyx_t_6 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(2, 398, __pyx_L1_error)
    __pyx_t_7 = __pyx_pybuffer_index((&__pyx_v_self->view), __pyx_v_itemp, __pyx_t_6, __pyx_v_dim); if (unlikely(__pyx_t_7 == ((char *)NULL))) __PYX_ERR(2, 398, __pyx_L1_error)
    __pyx_v_itemp = __pyx_t_7;

    /* "View.MemoryView":397
 *         cdef char *itemp = <char *> self.view.buf
 * 
 *         for dim, idx in enumerate(index):             # <<<<<<<<<<<<<<
 *             itemp = pybuffer_index(&self.view, itemp, idx, dim)
 * 
 */
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "View.MemoryView":400
 *             itemp = pybuffer_index(&self.view, itemp, idx, dim)
 * 
 *         return itemp             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_itemp;
  goto __pyx_L0;

  /* "View.MemoryView":393
 *                 PyThread_free_lock(self.lock)
 * 
 *     cdef char *get_item_pointer(memoryview self, object index) except NULL:             # <<<<<<<<<<<<<<
 *         cdef Py_ssize_t dim
 *         cdef char *itemp = <char *> self.view.buf
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("View.MemoryView.memoryview.get_item_pointer", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_idx);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":403
 * 
 * 
 *     def __getitem__(memoryview self, object index):             # <<<<<<<<<<<<<<
 *         if index is Ellipsis:
 *             return self
 */

/* Python wrapper */
static PyObject *__pyx_memoryview___getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_index); /*proto*/
static PyObject *__pyx_memoryview___getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_index) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getitem__ (wrapper)", 0);
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_4__getitem__(((struct __pyx_memoryview_obj *)__pyx_v_self), ((PyObject *)__pyx_v_index));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_4__getitem__(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_index) {
  PyObject *__pyx_v_have_slices = NULL;
  PyObject *__pyx_v_indices = NULL;
  char *__pyx_v_itemp;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  char *__pyx_t_6;
  __Pyx_RefNannySetupContext("__getitem__", 0);

  /* "View.MemoryView":404
 * 
 *     def __getitem__(memoryview self, object index):
 *         if index is Ellipsis:             # <<<<<<<<<<<<<<
 *             return self
 * 
 */
  __pyx_t_1 = (__pyx_v_index == __pyx_builtin_Ellipsis);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":405
 *     def __getitem__(memoryview self, object index):
 *         if index is Ellipsis:
 *             return self             # <<<<<<<<<<<<<<
 * 
 *         have_slices, indices = _unellipsify(index, self.view.ndim)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(((PyObject *)__pyx_v_self));
    __pyx_r = ((PyObject *)__pyx_v_self);
    goto __pyx_L0;

    /* "View.MemoryView":404
 * 
 *     def __getitem__(memoryview self, object index):
 *         if index is Ellipsis:             # <<<<<<<<<<<<<<
 *             return self
 * 
 */
  }

  /* "View.MemoryView":407
 *             return self
 * 
 *         have_slices, indices = _unellipsify(index, self.view.ndim)             # <<<<<<<<<<<<<<
 * 
 *         cdef char *itemp
 */
  __pyx_t_3 = _unellipsify(__pyx_v_index, __pyx_v_self->view.ndim); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 407, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (likely(__pyx_t_3 != Py_None)) {
    PyObject* sequence = __pyx_t_3;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 2)) {
      if (size > 2) __Pyx_RaiseTooManyValuesError(2);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(2, 407, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_4 = PyTuple_GET_ITEM(sequence, 0); 
    __pyx_t_5 = PyTuple_GET_ITEM(sequence, 1); 
    __Pyx_INCREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_t_5);
    #else
    __pyx_t_4 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 407, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 407, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    #endif
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else {
    __Pyx_RaiseNoneNotIterableError(); __PYX_ERR(2, 407, __pyx_L1_error)
  }
  __pyx_v_have_slices = __pyx_t_4;
  __pyx_t_4 = 0;
  __pyx_v_indices = __pyx_t_5;
  __pyx_t_5 = 0;

  /* "View.MemoryView":410
 * 
 *         cdef char *itemp
 *         if have_slices:             # <<<<<<<<<<<<<<
 *             return memview_slice(self, indices)
 *         else:
 */
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_have_slices); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(2, 410, __pyx_L1_error)
  if (__pyx_t_2) {

    /* "View.MemoryView":411
 *         cdef char *itemp
 *         if have_slices:
 *             return memview_slice(self, indices)             # <<<<<<<<<<<<<<
 *         else:
 *             itemp = self.get_item_pointer(indices)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = ((PyObject *)__pyx_memview_slice(__pyx_v_self, __pyx_v_indices)); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 411, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "View.MemoryView":410
 * 
 *         cdef char *itemp
 *         if have_slices:             # <<<<<<<<<<<<<<
 *             return memview_slice(self, indices)
 *         else:
 */
  }

  /* "View.MemoryView":413
 *             return memview_slice(self, indices)
 *         else:
 *             itemp = self.get_item_pointer(indices)             # <<<<<<<<<<<<<<
 *             return self.convert_item_to_object(itemp)
 * 
 */
  /*else*/ {
    __pyx_t_6 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->get_item_pointer(__pyx_v_self, __pyx_v_indices); if (unlikely(__pyx_t_6 == ((char *)NULL))) __PYX_ERR(2, 413, __pyx_L1_error)
    __pyx_v_itemp = __pyx_t_6;

    /* "View.MemoryView":414
 *         else:
 *             itemp = self.get_item_pointer(indices)
 *             return self.convert_item_to_object(itemp)             # <<<<<<<<<<<<<<
 * 
 *     def __setitem__(memoryview self, object index, object value):
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->convert_item_to_object(__pyx_v_self, __pyx_v_itemp); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 414, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "View.MemoryView":403
 * 
 * 
 *     def __getitem__(memoryview self, object index):             # <<<<<<<<<<<<<<
 *         if index is Ellipsis:
 *             return self
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("View.MemoryView.memoryview.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_have_slices);
  __Pyx_XDECREF(__pyx_v_indices);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":416
 *             return self.convert_item_to_object(itemp)
 * 
 *     def __setitem__(memoryview self, object index, object value):             # <<<<<<<<<<<<<<
 *         if self.view.readonly:
 *             raise TypeError("Cannot assign to read-only memoryview")
 */

/* Python wrapper */
static int __pyx_memoryview___setitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_index, PyObject *__pyx_v_value); /*proto*/
static int __pyx_memoryview___setitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_index, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setitem__ (wrapper)", 0);
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_6__setitem__(((struct __pyx_memoryview_obj *)__pyx_v_self), ((PyObject *)__pyx_v_index), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_6__setitem__(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_index, PyObject *__pyx_v_value) {
  PyObject *__pyx_v_have_slices = NULL;
  PyObject *__pyx_v_obj = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("__setitem__", 0);
  __Pyx_INCREF(__pyx_v_index);

  /* "View.MemoryView":417
 * 
 *     def __setitem__(memoryview self, object index, object value):
 *         if self.view.readonly:             # <<<<<<<<<<<<<<
 *             raise TypeError("Cannot assign to read-only memoryview")
 * 
 */
  __pyx_t_1 = (__pyx_v_self->view.readonly != 0);
  if (unlikely(__pyx_t_1)) {

    /* "View.MemoryView":418
 *     def __setitem__(memoryview self, object index, object value):
 *         if self.view.readonly:
 *             raise TypeError("Cannot assign to read-only memoryview")             # <<<<<<<<<<<<<<
 * 
 *         have_slices, index = _unellipsify(index, self.view.ndim)
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__34, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 418, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(2, 418, __pyx_L1_error)

    /* "View.MemoryView":417
 * 
 *     def __setitem__(memoryview self, object index, object value):
 *         if self.view.readonly:             # <<<<<<<<<<<<<<
 *             raise TypeError("Cannot assign to read-only memoryview")
 * 
 */
  }

  /* "View.MemoryView":420
 *             raise TypeError("Cannot assign to read-only memoryview")
 * 
 *         have_slices, index = _unellipsify(index, self.view.ndim)             # <<<<<<<<<<<<<<
 * 
 *         if have_slices:
 */
  __pyx_t_2 = _unellipsify(__pyx_v_index, __pyx_v_self->view.ndim); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 420, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (likely(__pyx_t_2 != Py_None)) {
    PyObject* sequence = __pyx_t_2;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 2)) {
      if (size > 2) __Pyx_RaiseTooManyValuesError(2);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(2, 420, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_3 = PyTuple_GET_ITEM(sequence, 0); 
    __pyx_t_4 = PyTuple_GET_ITEM(sequence, 1); 
    __Pyx_INCREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_t_4);
    #else
    __pyx_t_3 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 420, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 420, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    #endif
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  } else {
    __Pyx_RaiseNoneNotIterableError(); __PYX_ERR(2, 420, __pyx_L1_error)
  }
  __pyx_v_have_slices = __pyx_t_3;
  __pyx_t_3 = 0;
  __Pyx_DECREF_SET(__pyx_v_index, __pyx_t_4);
  __pyx_t_4 = 0;

  /* "View.MemoryView":422
 *         have_slices, index = _unellipsify(index, self.view.ndim)
 * 
 *         if have_slices:             # <<<<<<<<<<<<<<
 *             obj = self.is_slice(value)
 *             if obj:
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_have_slices); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(2, 422, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "View.MemoryView":423
 * 
 *         if have_slices:
 *             obj = self.is_slice(value)             # <<<<<<<<<<<<<<
 *             if obj:
 *                 self.setitem_slice_assignment(self[index], obj)
 */
    __pyx_t_2 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->is_slice(__pyx_v_self, __pyx_v_value); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 423, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_obj = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "View.MemoryView":424
 *         if have_slices:
 *             obj = self.is_slice(value)
 *             if obj:             # <<<<<<<<<<<<<<
 *                 self.setitem_slice_assignment(self[index], obj)
 *             else:
 */
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_obj); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(2, 424, __pyx_L1_error)
    if (__pyx_t_1) {

      /* "View.MemoryView":425
 *             obj = self.is_slice(value)
 *             if obj:
 *                 self.setitem_slice_assignment(self[index], obj)             # <<<<<<<<<<<<<<
 *             else:
 *                 self.setitem_slice_assign_scalar(self[index], value)
 */
      __pyx_t_2 = __Pyx_PyObject_GetItem(((PyObject *)__pyx_v_self), __pyx_v_index); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 425, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_4 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->setitem_slice_assignment(__pyx_v_self, __pyx_t_2, __pyx_v_obj); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 425, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "View.MemoryView":424
 *         if have_slices:
 *             obj = self.is_slice(value)
 *             if obj:             # <<<<<<<<<<<<<<
 *                 self.setitem_slice_assignment(self[index], obj)
 *             else:
 */
      goto __pyx_L5;
    }

    /* "View.MemoryView":427
 *                 self.setitem_slice_assignment(self[index], obj)
 *             else:
 *                 self.setitem_slice_assign_scalar(self[index], value)             # <<<<<<<<<<<<<<
 *         else:
 *             self.setitem_indexed(index, value)
 */
    /*else*/ {
      __pyx_t_4 = __Pyx_PyObject_GetItem(((PyObject *)__pyx_v_self), __pyx_v_index); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 427, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      if (!(likely(((__pyx_t_4) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_4, __pyx_memoryview_type))))) __PYX_ERR(2, 427, __pyx_L1_error)
      __pyx_t_2 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->setitem_slice_assign_scalar(__pyx_v_self, ((struct __pyx_memoryview_obj *)__pyx_t_4), __pyx_v_value); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 427, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
    __pyx_L5:;

    /* "View.MemoryView":422
 *         have_slices, index = _unellipsify(index, self.view.ndim)
 * 
 *         if have_slices:             # <<<<<<<<<<<<<<
 *             obj = self.is_slice(value)
 *             if obj:
 */
    goto __pyx_L4;
  }

  /* "View.MemoryView":429
 *                 self.setitem_slice_assign_scalar(self[index], value)
 *         else:
 *             self.setitem_indexed(index, value)             # <<<<<<<<<<<<<<
 * 
 *     cdef is_slice(self, obj):
 */
  /*else*/ {
    __pyx_t_2 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->setitem_indexed(__pyx_v_self, __pyx_v_index, __pyx_v_value); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 429, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_L4:;

  /* "View.MemoryView":416
 *             return self.convert_item_to_object(itemp)
 * 
 *     def __setitem__(memoryview self, object index, object value):             # <<<<<<<<<<<<<<
 *         if self.view.readonly:
 *             raise TypeError("Cannot assign to read-only memoryview")
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("View.MemoryView.memoryview.__setitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_have_slices);
  __Pyx_XDECREF(__pyx_v_obj);
  __Pyx_XDECREF(__pyx_v_index);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":431
 *             self.setitem_indexed(index, value)
 * 
 *     cdef is_slice(self, obj):             # <<<<<<<<<<<<<<
 *         if not isinstance(obj, memoryview):
 *             try:
 */

static PyObject *__pyx_memoryview_is_slice(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_obj) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  __Pyx_RefNannySetupContext("is_slice", 0);
  __Pyx_INCREF(__pyx_v_obj);

  /* "View.MemoryView":432
 * 
 *     cdef is_slice(self, obj):
 *         if not isinstance(obj, memoryview):             # <<<<<<<<<<<<<<
 *             try:
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,
 */
  __pyx_t_1 = __Pyx_TypeCheck(__pyx_v_obj, __pyx_memoryview_type); 
  __pyx_t_2 = ((!(__pyx_t_1 != 0)) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":433
 *     cdef is_slice(self, obj):
 *         if not isinstance(obj, memoryview):
 *             try:             # <<<<<<<<<<<<<<
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,
 *                                  self.dtype_is_object)
 */
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_3, &__pyx_t_4, &__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_3);
      __Pyx_XGOTREF(__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_5);
      /*try:*/ {

        /* "View.MemoryView":434
 *         if not isinstance(obj, memoryview):
 *             try:
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,             # <<<<<<<<<<<<<<
 *                                  self.dtype_is_object)
 *             except TypeError:
 */
        __pyx_t_6 = __Pyx_PyInt_From_int(((__pyx_v_self->flags & (~PyBUF_WRITABLE)) | PyBUF_ANY_CONTIGUOUS)); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 434, __pyx_L4_error)
        __Pyx_GOTREF(__pyx_t_6);

        /* "View.MemoryView":435
 *             try:
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,
 *                                  self.dtype_is_object)             # <<<<<<<<<<<<<<
 *             except TypeError:
 *                 return None
 */
        __pyx_t_7 = __Pyx_PyBool_FromLong(__pyx_v_self->dtype_is_object); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 435, __pyx_L4_error)
        __Pyx_GOTREF(__pyx_t_7);

        /* "View.MemoryView":434
 *         if not isinstance(obj, memoryview):
 *             try:
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,             # <<<<<<<<<<<<<<
 *                                  self.dtype_is_object)
 *             except TypeError:
 */
        __pyx_t_8 = PyTuple_New(3); if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 434, __pyx_L4_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_INCREF(__pyx_v_obj);
        __Pyx_GIVEREF(__pyx_v_obj);
        PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_v_obj);
        __Pyx_GIVEREF(__pyx_t_6);
        PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_7);
        PyTuple_SET_ITEM(__pyx_t_8, 2, __pyx_t_7);
        __pyx_t_6 = 0;
        __pyx_t_7 = 0;
        __pyx_t_7 = __Pyx_PyObject_Call(((PyObject *)__pyx_memoryview_type), __pyx_t_8, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 434, __pyx_L4_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_DECREF_SET(__pyx_v_obj, __pyx_t_7);
        __pyx_t_7 = 0;

        /* "View.MemoryView":433
 *     cdef is_slice(self, obj):
 *         if not isinstance(obj, memoryview):
 *             try:             # <<<<<<<<<<<<<<
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,
 *                                  self.dtype_is_object)
 */
      }
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      goto __pyx_L9_try_end;
      __pyx_L4_error:;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;

      /* "View.MemoryView":436
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,
 *                                  self.dtype_is_object)
 *             except TypeError:             # <<<<<<<<<<<<<<
 *                 return None
 * 
 */
      __pyx_t_9 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_TypeError);
      if (__pyx_t_9) {
        __Pyx_AddTraceback("View.MemoryView.memoryview.is_slice", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_7, &__pyx_t_8, &__pyx_t_6) < 0) __PYX_ERR(2, 436, __pyx_L6_except_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_GOTREF(__pyx_t_6);

        /* "View.MemoryView":437
 *                                  self.dtype_is_object)
 *             except TypeError:
 *                 return None             # <<<<<<<<<<<<<<
 * 
 *         return obj
 */
        __Pyx_XDECREF(__pyx_r);
        __pyx_r = Py_None; __Pyx_INCREF(Py_None);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        goto __pyx_L7_except_return;
      }
      goto __pyx_L6_except_error;
      __pyx_L6_except_error:;

      /* "View.MemoryView":433
 *     cdef is_slice(self, obj):
 *         if not isinstance(obj, memoryview):
 *             try:             # <<<<<<<<<<<<<<
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,
 *                                  self.dtype_is_object)
 */
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_4, __pyx_t_5);
      goto __pyx_L1_error;
      __pyx_L7_except_return:;
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_4, __pyx_t_5);
      goto __pyx_L0;
      __pyx_L9_try_end:;
    }

    /* "View.MemoryView":432
 * 
 *     cdef is_slice(self, obj):
 *         if not isinstance(obj, memoryview):             # <<<<<<<<<<<<<<
 *             try:
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,
 */
  }

  /* "View.MemoryView":439
 *                 return None
 * 
 *         return obj             # <<<<<<<<<<<<<<
 * 
 *     cdef setitem_slice_assignment(self, dst, src):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_obj);
  __pyx_r = __pyx_v_obj;
  goto __pyx_L0;

  /* "View.MemoryView":431
 *             self.setitem_indexed(index, value)
 * 
 *     cdef is_slice(self, obj):             # <<<<<<<<<<<<<<
 *         if not isinstance(obj, memoryview):
 *             try:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("View.MemoryView.memoryview.is_slice", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_obj);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":441
 *         return obj
 * 
 *     cdef setitem_slice_assignment(self, dst, src):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice dst_slice
 *         cdef __Pyx_memviewslice src_slice
 */

static PyObject *__pyx_memoryview_setitem_slice_assignment(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_dst, PyObject *__pyx_v_src) {
  __Pyx_memviewslice __pyx_v_dst_slice;
  __Pyx_memviewslice __pyx_v_src_slice;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  __Pyx_RefNannySetupContext("setitem_slice_assignment", 0);

  /* "View.MemoryView":445
 *         cdef __Pyx_memviewslice src_slice
 * 
 *         memoryview_copy_contents(get_slice_from_memview(src, &src_slice)[0],             # <<<<<<<<<<<<<<
 *                                  get_slice_from_memview(dst, &dst_slice)[0],
 *                                  src.ndim, dst.ndim, self.dtype_is_object)
 */
  if (!(likely(((__pyx_v_src) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_src, __pyx_memoryview_type))))) __PYX_ERR(2, 445, __pyx_L1_error)

  /* "View.MemoryView":446
 * 
 *         memoryview_copy_contents(get_slice_from_memview(src, &src_slice)[0],
 *                                  get_slice_from_memview(dst, &dst_slice)[0],             # <<<<<<<<<<<<<<
 *                                  src.ndim, dst.ndim, self.dtype_is_object)
 * 
 */
  if (!(likely(((__pyx_v_dst) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_dst, __pyx_memoryview_type))))) __PYX_ERR(2, 446, __pyx_L1_error)

  /* "View.MemoryView":447
 *         memoryview_copy_contents(get_slice_from_memview(src, &src_slice)[0],
 *                                  get_slice_from_memview(dst, &dst_slice)[0],
 *                                  src.ndim, dst.ndim, self.dtype_is_object)             # <<<<<<<<<<<<<<
 * 
 *     cdef setitem_slice_assign_scalar(self, memoryview dst, value):
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_src, __pyx_n_s_ndim); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 447, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(2, 447, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_dst, __pyx_n_s_ndim); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 447, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_3 == (int)-1) && PyErr_Occurred())) __PYX_ERR(2, 447, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "View.MemoryView":445
 *         cdef __Pyx_memviewslice src_slice
 * 
 *         memoryview_copy_contents(get_slice_from_memview(src, &src_slice)[0],             # <<<<<<<<<<<<<<
 *                                  get_slice_from_memview(dst, &dst_slice)[0],
 *                                  src.ndim, dst.ndim, self.dtype_is_object)
 */
  __pyx_t_4 = __pyx_memoryview_copy_contents((__pyx_memoryview_get_slice_from_memoryview(((struct __pyx_memoryview_obj *)__pyx_v_src), (&__pyx_v_src_slice))[0]), (__pyx_memoryview_get_slice_from_memoryview(((struct __pyx_memoryview_obj *)__pyx_v_dst), (&__pyx_v_dst_slice))[0]), __pyx_t_2, __pyx_t_3, __pyx_v_self->dtype_is_object); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(2, 445, __pyx_L1_error)

  /* "View.MemoryView":441
 *         return obj
 * 
 *     cdef setitem_slice_assignment(self, dst, src):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice dst_slice
 *         cdef __Pyx_memviewslice src_slice
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.memoryview.setitem_slice_assignment", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":449
 *                                  src.ndim, dst.ndim, self.dtype_is_object)
 * 
 *     cdef setitem_slice_assign_scalar(self, memoryview dst, value):             # <<<<<<<<<<<<<<
 *         cdef int array[128]
 *         cdef void *tmp = NULL
 */

static PyObject *__pyx_memoryview_setitem_slice_assign_scalar(struct __pyx_memoryview_obj *__pyx_v_self, struct __pyx_memoryview_obj *__pyx_v_dst, PyObject *__pyx_v_value) {
  int __pyx_v_array[0x80];
  void *__pyx_v_tmp;
  void *__pyx_v_item;
  __Pyx_memviewslice *__pyx_v_dst_slice;
  __Pyx_memviewslice __pyx_v_tmp_slice;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_t_4;
  char const *__pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  __Pyx_RefNannySetupContext("setitem_slice_assign_scalar", 0);

  /* "View.MemoryView":451
 *     cdef setitem_slice_assign_scalar(self, memoryview dst, value):
 *         cdef int array[128]
 *         cdef void *tmp = NULL             # <<<<<<<<<<<<<<
 *         cdef void *item
 * 
 */
  __pyx_v_tmp = NULL;

  /* "View.MemoryView":456
 *         cdef __Pyx_memviewslice *dst_slice
 *         cdef __Pyx_memviewslice tmp_slice
 *         dst_slice = get_slice_from_memview(dst, &tmp_slice)             # <<<<<<<<<<<<<<
 * 
 *         if <size_t>self.view.itemsize > sizeof(array):
 */
  __pyx_v_dst_slice = __pyx_memoryview_get_slice_from_memoryview(__pyx_v_dst, (&__pyx_v_tmp_slice));

  /* "View.MemoryView":458
 *         dst_slice = get_slice_from_memview(dst, &tmp_slice)
 * 
 *         if <size_t>self.view.itemsize > sizeof(array):             # <<<<<<<<<<<<<<
 *             tmp = PyMem_Malloc(self.view.itemsize)
 *             if tmp == NULL:
 */
  __pyx_t_1 = ((((size_t)__pyx_v_self->view.itemsize) > (sizeof(__pyx_v_array))) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":459
 * 
 *         if <size_t>self.view.itemsize > sizeof(array):
 *             tmp = PyMem_Malloc(self.view.itemsize)             # <<<<<<<<<<<<<<
 *             if tmp == NULL:
 *                 raise MemoryError
 */
    __pyx_v_tmp = PyMem_Malloc(__pyx_v_self->view.itemsize);

    /* "View.MemoryView":460
 *         if <size_t>self.view.itemsize > sizeof(array):
 *             tmp = PyMem_Malloc(self.view.itemsize)
 *             if tmp == NULL:             # <<<<<<<<<<<<<<
 *                 raise MemoryError
 *             item = tmp
 */
    __pyx_t_1 = ((__pyx_v_tmp == NULL) != 0);
    if (unlikely(__pyx_t_1)) {

      /* "View.MemoryView":461
 *             tmp = PyMem_Malloc(self.view.itemsize)
 *             if tmp == NULL:
 *                 raise MemoryError             # <<<<<<<<<<<<<<
 *             item = tmp
 *         else:
 */
      PyErr_NoMemory(); __PYX_ERR(2, 461, __pyx_L1_error)

      /* "View.MemoryView":460
 *         if <size_t>self.view.itemsize > sizeof(array):
 *             tmp = PyMem_Malloc(self.view.itemsize)
 *             if tmp == NULL:             # <<<<<<<<<<<<<<
 *                 raise MemoryError
 *             item = tmp
 */
    }

    /* "View.MemoryView":462
 *             if tmp == NULL:
 *                 raise MemoryError
 *             item = tmp             # <<<<<<<<<<<<<<
 *         else:
 *             item = <void *> array
 */
    __pyx_v_item = __pyx_v_tmp;

    /* "View.MemoryView":458
 *         dst_slice = get_slice_from_memview(dst, &tmp_slice)
 * 
 *         if <size_t>self.view.itemsize > sizeof(array):             # <<<<<<<<<<<<<<
 *             tmp = PyMem_Malloc(self.view.itemsize)
 *             if tmp == NULL:
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":464
 *             item = tmp
 *         else:
 *             item = <void *> array             # <<<<<<<<<<<<<<
 * 
 *         try:
 */
  /*else*/ {
    __pyx_v_item = ((void *)__pyx_v_array);
  }
  __pyx_L3:;

  /* "View.MemoryView":466
 *             item = <void *> array
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             if self.dtype_is_object:
 *                 (<PyObject **> item)[0] = <PyObject *> value
 */
  /*try:*/ {

    /* "View.MemoryView":467
 * 
 *         try:
 *             if self.dtype_is_object:             # <<<<<<<<<<<<<<
 *                 (<PyObject **> item)[0] = <PyObject *> value
 *             else:
 */
    __pyx_t_1 = (__pyx_v_self->dtype_is_object != 0);
    if (__pyx_t_1) {

      /* "View.MemoryView":468
 *         try:
 *             if self.dtype_is_object:
 *                 (<PyObject **> item)[0] = <PyObject *> value             # <<<<<<<<<<<<<<
 *             else:
 *                 self.assign_item_from_object(<char *> item, value)
 */
      (((PyObject **)__pyx_v_item)[0]) = ((PyObject *)__pyx_v_value);

      /* "View.MemoryView":467
 * 
 *         try:
 *             if self.dtype_is_object:             # <<<<<<<<<<<<<<
 *                 (<PyObject **> item)[0] = <PyObject *> value
 *             else:
 */
      goto __pyx_L8;
    }

    /* "View.MemoryView":470
 *                 (<PyObject **> item)[0] = <PyObject *> value
 *             else:
 *                 self.assign_item_from_object(<char *> item, value)             # <<<<<<<<<<<<<<
 * 
 * 
 */
    /*else*/ {
      __pyx_t_2 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->assign_item_from_object(__pyx_v_self, ((char *)__pyx_v_item), __pyx_v_value); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 470, __pyx_L6_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
    __pyx_L8:;

    /* "View.MemoryView":474
 * 
 * 
 *             if self.view.suboffsets != NULL:             # <<<<<<<<<<<<<<
 *                 assert_direct_dimensions(self.view.suboffsets, self.view.ndim)
 *             slice_assign_scalar(dst_slice, dst.view.ndim, self.view.itemsize,
 */
    __pyx_t_1 = ((__pyx_v_self->view.suboffsets != NULL) != 0);
    if (__pyx_t_1) {

      /* "View.MemoryView":475
 * 
 *             if self.view.suboffsets != NULL:
 *                 assert_direct_dimensions(self.view.suboffsets, self.view.ndim)             # <<<<<<<<<<<<<<
 *             slice_assign_scalar(dst_slice, dst.view.ndim, self.view.itemsize,
 *                                 item, self.dtype_is_object)
 */
      __pyx_t_2 = assert_direct_dimensions(__pyx_v_self->view.suboffsets, __pyx_v_self->view.ndim); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 475, __pyx_L6_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "View.MemoryView":474
 * 
 * 
 *             if self.view.suboffsets != NULL:             # <<<<<<<<<<<<<<
 *                 assert_direct_dimensions(self.view.suboffsets, self.view.ndim)
 *             slice_assign_scalar(dst_slice, dst.view.ndim, self.view.itemsize,
 */
    }

    /* "View.MemoryView":476
 *             if self.view.suboffsets != NULL:
 *                 assert_direct_dimensions(self.view.suboffsets, self.view.ndim)
 *             slice_assign_scalar(dst_slice, dst.view.ndim, self.view.itemsize,             # <<<<<<<<<<<<<<
 *                                 item, self.dtype_is_object)
 *         finally:
 */
    __pyx_memoryview_slice_assign_scalar(__pyx_v_dst_slice, __pyx_v_dst->view.ndim, __pyx_v_self->view.itemsize, __pyx_v_item, __pyx_v_self->dtype_is_object);
  }

  /* "View.MemoryView":479
 *                                 item, self.dtype_is_object)
 *         finally:
 *             PyMem_Free(tmp)             # <<<<<<<<<<<<<<
 * 
 *     cdef setitem_indexed(self, index, value):
 */
  /*finally:*/ {
    /*normal exit:*/{
      PyMem_Free(__pyx_v_tmp);
      goto __pyx_L7;
    }
    __pyx_L6_error:;
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __pyx_t_6 = 0; __pyx_t_7 = 0; __pyx_t_8 = 0; __pyx_t_9 = 0; __pyx_t_10 = 0; __pyx_t_11 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_9, &__pyx_t_10, &__pyx_t_11);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_6, &__pyx_t_7, &__pyx_t_8) < 0)) __Pyx_ErrFetch(&__pyx_t_6, &__pyx_t_7, &__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_7);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_11);
      __pyx_t_3 = __pyx_lineno; __pyx_t_4 = __pyx_clineno; __pyx_t_5 = __pyx_filename;
      {
        PyMem_Free(__pyx_v_tmp);
      }
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_9);
        __Pyx_XGIVEREF(__pyx_t_10);
        __Pyx_XGIVEREF(__pyx_t_11);
        __Pyx_ExceptionReset(__pyx_t_9, __pyx_t_10, __pyx_t_11);
      }
      __Pyx_XGIVEREF(__pyx_t_6);
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_ErrRestore(__pyx_t_6, __pyx_t_7, __pyx_t_8);
      __pyx_t_6 = 0; __pyx_t_7 = 0; __pyx_t_8 = 0; __pyx_t_9 = 0; __pyx_t_10 = 0; __pyx_t_11 = 0;
      __pyx_lineno = __pyx_t_3; __pyx_clineno = __pyx_t_4; __pyx_filename = __pyx_t_5;
      goto __pyx_L1_error;
    }
    __pyx_L7:;
  }

  /* "View.MemoryView":449
 *                                  src.ndim, dst.ndim, self.dtype_is_object)
 * 
 *     cdef setitem_slice_assign_scalar(self, memoryview dst, value):             # <<<<<<<<<<<<<<
 *         cdef int array[128]
 *         cdef void *tmp = NULL
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("View.MemoryView.memoryview.setitem_slice_assign_scalar", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":481
 *             PyMem_Free(tmp)
 * 
 *     cdef setitem_indexed(self, index, value):             # <<<<<<<<<<<<<<
 *         cdef char *itemp = self.get_item_pointer(index)
 *         self.assign_item_from_object(itemp, value)
 */

static PyObject *__pyx_memoryview_setitem_indexed(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_index, PyObject *__pyx_v_value) {
  char *__pyx_v_itemp;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  char *__pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("setitem_indexed", 0);

  /* "View.MemoryView":482
 * 
 *     cdef setitem_indexed(self, index, value):
 *         cdef char *itemp = self.get_item_pointer(index)             # <<<<<<<<<<<<<<
 *         self.assign_item_from_object(itemp, value)
 * 
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->get_item_pointer(__pyx_v_self, __pyx_v_index); if (unlikely(__pyx_t_1 == ((char *)NULL))) __PYX_ERR(2, 482, __pyx_L1_error)
  __pyx_v_itemp = __pyx_t_1;

  /* "View.MemoryView":483
 *     cdef setitem_indexed(self, index, value):
 *         cdef char *itemp = self.get_item_pointer(index)
 *         self.assign_item_from_object(itemp, value)             # <<<<<<<<<<<<<<
 * 
 *     cdef convert_item_to_object(self, char *itemp):
 */
  __pyx_t_2 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->assign_item_from_object(__pyx_v_self, __pyx_v_itemp, __pyx_v_value); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 483, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "View.MemoryView":481
 *             PyMem_Free(tmp)
 * 
 *     cdef setitem_indexed(self, index, value):             # <<<<<<<<<<<<<<
 *         cdef char *itemp = self.get_item_pointer(index)
 *         self.assign_item_from_object(itemp, value)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("View.MemoryView.memoryview.setitem_indexed", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":485
 *         self.assign_item_from_object(itemp, value)
 * 
 *     cdef convert_item_to_object(self, char *itemp):             # <<<<<<<<<<<<<<
 *         """Only used if instantiated manually by the user, or if Cython doesn't
 *         know how to convert the type"""
 */

static PyObject *__pyx_memoryview_convert_item_to_object(struct __pyx_memoryview_obj *__pyx_v_self, char *__pyx_v_itemp) {
  PyObject *__pyx_v_struct = NULL;
  PyObject *__pyx_v_bytesitem = 0;
  PyObject *__pyx_v_result = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  size_t __pyx_t_10;
  int __pyx_t_11;
  __Pyx_RefNannySetupContext("convert_item_to_object", 0);

  /* "View.MemoryView":488
 *         """Only used if instantiated manually by the user, or if Cython doesn't
 *         know how to convert the type"""
 *         import struct             # <<<<<<<<<<<<<<
 *         cdef bytes bytesitem
 * 
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_struct, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 488, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_struct = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "View.MemoryView":491
 *         cdef bytes bytesitem
 * 
 *         bytesitem = itemp[:self.view.itemsize]             # <<<<<<<<<<<<<<
 *         try:
 *             result = struct.unpack(self.view.format, bytesitem)
 */
  __pyx_t_1 = __Pyx_PyBytes_FromStringAndSize(__pyx_v_itemp + 0, __pyx_v_self->view.itemsize - 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 491, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_bytesitem = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "View.MemoryView":492
 * 
 *         bytesitem = itemp[:self.view.itemsize]
 *         try:             # <<<<<<<<<<<<<<
 *             result = struct.unpack(self.view.format, bytesitem)
 *         except struct.error:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_2, &__pyx_t_3, &__pyx_t_4);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_4);
    /*try:*/ {

      /* "View.MemoryView":493
 *         bytesitem = itemp[:self.view.itemsize]
 *         try:
 *             result = struct.unpack(self.view.format, bytesitem)             # <<<<<<<<<<<<<<
 *         except struct.error:
 *             raise ValueError("Unable to convert item to object")
 */
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_struct, __pyx_n_s_unpack); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 493, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = __Pyx_PyBytes_FromString(__pyx_v_self->view.format); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 493, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_7 = NULL;
      __pyx_t_8 = 0;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_7)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_7);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
          __pyx_t_8 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_5)) {
        PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_6, __pyx_v_bytesitem};
        __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 493, __pyx_L3_error)
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
        PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_6, __pyx_v_bytesitem};
        __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 493, __pyx_L3_error)
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      {
        __pyx_t_9 = PyTuple_New(2+__pyx_t_8); if (unlikely(!__pyx_t_9)) __PYX_ERR(2, 493, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_9);
        if (__pyx_t_7) {
          __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_7); __pyx_t_7 = NULL;
        }
        __Pyx_GIVEREF(__pyx_t_6);
        PyTuple_SET_ITEM(__pyx_t_9, 0+__pyx_t_8, __pyx_t_6);
        __Pyx_INCREF(__pyx_v_bytesitem);
        __Pyx_GIVEREF(__pyx_v_bytesitem);
        PyTuple_SET_ITEM(__pyx_t_9, 1+__pyx_t_8, __pyx_v_bytesitem);
        __pyx_t_6 = 0;
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_9, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 493, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      }
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_v_result = __pyx_t_1;
      __pyx_t_1 = 0;

      /* "View.MemoryView":492
 * 
 *         bytesitem = itemp[:self.view.itemsize]
 *         try:             # <<<<<<<<<<<<<<
 *             result = struct.unpack(self.view.format, bytesitem)
 *         except struct.error:
 */
    }

    /* "View.MemoryView":497
 *             raise ValueError("Unable to convert item to object")
 *         else:
 *             if len(self.view.format) == 1:             # <<<<<<<<<<<<<<
 *                 return result[0]
 *             return result
 */
    /*else:*/ {
      __pyx_t_10 = strlen(__pyx_v_self->view.format); 
      __pyx_t_11 = ((__pyx_t_10 == 1) != 0);
      if (__pyx_t_11) {

        /* "View.MemoryView":498
 *         else:
 *             if len(self.view.format) == 1:
 *                 return result[0]             # <<<<<<<<<<<<<<
 *             return result
 * 
 */
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_result, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 498, __pyx_L5_except_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_r = __pyx_t_1;
        __pyx_t_1 = 0;
        goto __pyx_L6_except_return;

        /* "View.MemoryView":497
 *             raise ValueError("Unable to convert item to object")
 *         else:
 *             if len(self.view.format) == 1:             # <<<<<<<<<<<<<<
 *                 return result[0]
 *             return result
 */
      }

      /* "View.MemoryView":499
 *             if len(self.view.format) == 1:
 *                 return result[0]
 *             return result             # <<<<<<<<<<<<<<
 * 
 *     cdef assign_item_from_object(self, char *itemp, object value):
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_v_result);
      __pyx_r = __pyx_v_result;
      goto __pyx_L6_except_return;
    }
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

    /* "View.MemoryView":494
 *         try:
 *             result = struct.unpack(self.view.format, bytesitem)
 *         except struct.error:             # <<<<<<<<<<<<<<
 *             raise ValueError("Unable to convert item to object")
 *         else:
 */
    __Pyx_ErrFetch(&__pyx_t_1, &__pyx_t_5, &__pyx_t_9);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_struct, __pyx_n_s_error); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 494, __pyx_L5_except_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_8 = __Pyx_PyErr_GivenExceptionMatches(__pyx_t_1, __pyx_t_6);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_ErrRestore(__pyx_t_1, __pyx_t_5, __pyx_t_9);
    __pyx_t_1 = 0; __pyx_t_5 = 0; __pyx_t_9 = 0;
    if (__pyx_t_8) {
      __Pyx_AddTraceback("View.MemoryView.memoryview.convert_item_to_object", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_9, &__pyx_t_5, &__pyx_t_1) < 0) __PYX_ERR(2, 494, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_1);

      /* "View.MemoryView":495
 *             result = struct.unpack(self.view.format, bytesitem)
 *         except struct.error:
 *             raise ValueError("Unable to convert item to object")             # <<<<<<<<<<<<<<
 *         else:
 *             if len(self.view.format) == 1:
 */
      __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__35, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 495, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_Raise(__pyx_t_6, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __PYX_ERR(2, 495, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "View.MemoryView":492
 * 
 *         bytesitem = itemp[:self.view.itemsize]
 *         try:             # <<<<<<<<<<<<<<
 *             result = struct.unpack(self.view.format, bytesitem)
 *         except struct.error:
 */
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_ExceptionReset(__pyx_t_2, __pyx_t_3, __pyx_t_4);
    goto __pyx_L1_error;
    __pyx_L6_except_return:;
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_ExceptionReset(__pyx_t_2, __pyx_t_3, __pyx_t_4);
    goto __pyx_L0;
  }

  /* "View.MemoryView":485
 *         self.assign_item_from_object(itemp, value)
 * 
 *     cdef convert_item_to_object(self, char *itemp):             # <<<<<<<<<<<<<<
 *         """Only used if instantiated manually by the user, or if Cython doesn't
 *         know how to convert the type"""
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("View.MemoryView.memoryview.convert_item_to_object", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_struct);
  __Pyx_XDECREF(__pyx_v_bytesitem);
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":501
 *             return result
 * 
 *     cdef assign_item_from_object(self, char *itemp, object value):             # <<<<<<<<<<<<<<
 *         """Only used if instantiated manually by the user, or if Cython doesn't
 *         know how to convert the type"""
 */

static PyObject *__pyx_memoryview_assign_item_from_object(struct __pyx_memoryview_obj *__pyx_v_self, char *__pyx_v_itemp, PyObject *__pyx_v_value) {
  PyObject *__pyx_v_struct = NULL;
  char __pyx_v_c;
  PyObject *__pyx_v_bytesvalue = 0;
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  Py_ssize_t __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  char *__pyx_t_11;
  char *__pyx_t_12;
  char *__pyx_t_13;
  char *__pyx_t_14;
  __Pyx_RefNannySetupContext("assign_item_from_object", 0);

  /* "View.MemoryView":504
 *         """Only used if instantiated manually by the user, or if Cython doesn't
 *         know how to convert the type"""
 *         import struct             # <<<<<<<<<<<<<<
 *         cdef char c
 *         cdef bytes bytesvalue
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_struct, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 504, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_struct = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "View.MemoryView":509
 *         cdef Py_ssize_t i
 * 
 *         if isinstance(value, tuple):             # <<<<<<<<<<<<<<
 *             bytesvalue = struct.pack(self.view.format, *value)
 *         else:
 */
  __pyx_t_2 = PyTuple_Check(__pyx_v_value); 
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "View.MemoryView":510
 * 
 *         if isinstance(value, tuple):
 *             bytesvalue = struct.pack(self.view.format, *value)             # <<<<<<<<<<<<<<
 *         else:
 *             bytesvalue = struct.pack(self.view.format, value)
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_struct, __pyx_n_s_pack); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 510, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = __Pyx_PyBytes_FromString(__pyx_v_self->view.format); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 510, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 510, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4);
    __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PySequence_Tuple(__pyx_v_value); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 510, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = PyNumber_Add(__pyx_t_5, __pyx_t_4); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 510, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_6, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 510, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (!(likely(PyBytes_CheckExact(__pyx_t_4))||((__pyx_t_4) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_t_4)->tp_name), 0))) __PYX_ERR(2, 510, __pyx_L1_error)
    __pyx_v_bytesvalue = ((PyObject*)__pyx_t_4);
    __pyx_t_4 = 0;

    /* "View.MemoryView":509
 *         cdef Py_ssize_t i
 * 
 *         if isinstance(value, tuple):             # <<<<<<<<<<<<<<
 *             bytesvalue = struct.pack(self.view.format, *value)
 *         else:
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":512
 *             bytesvalue = struct.pack(self.view.format, *value)
 *         else:
 *             bytesvalue = struct.pack(self.view.format, value)             # <<<<<<<<<<<<<<
 * 
 *         for i, c in enumerate(bytesvalue):
 */
  /*else*/ {
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_struct, __pyx_n_s_pack); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 512, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_1 = __Pyx_PyBytes_FromString(__pyx_v_self->view.format); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 512, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = NULL;
    __pyx_t_7 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_7 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_6)) {
      PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_t_1, __pyx_v_value};
      __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 512, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_6)) {
      PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_t_1, __pyx_v_value};
      __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 512, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    } else
    #endif
    {
      __pyx_t_8 = PyTuple_New(2+__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 512, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      if (__pyx_t_5) {
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_5); __pyx_t_5 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_1);
      PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_7, __pyx_t_1);
      __Pyx_INCREF(__pyx_v_value);
      __Pyx_GIVEREF(__pyx_v_value);
      PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_7, __pyx_v_value);
      __pyx_t_1 = 0;
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_8, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 512, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    }
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (!(likely(PyBytes_CheckExact(__pyx_t_4))||((__pyx_t_4) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_t_4)->tp_name), 0))) __PYX_ERR(2, 512, __pyx_L1_error)
    __pyx_v_bytesvalue = ((PyObject*)__pyx_t_4);
    __pyx_t_4 = 0;
  }
  __pyx_L3:;

  /* "View.MemoryView":514
 *             bytesvalue = struct.pack(self.view.format, value)
 * 
 *         for i, c in enumerate(bytesvalue):             # <<<<<<<<<<<<<<
 *             itemp[i] = c
 * 
 */
  __pyx_t_9 = 0;
  if (unlikely(__pyx_v_bytesvalue == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' is not iterable");
    __PYX_ERR(2, 514, __pyx_L1_error)
  }
  __Pyx_INCREF(__pyx_v_bytesvalue);
  __pyx_t_10 = __pyx_v_bytesvalue;
  __pyx_t_12 = PyBytes_AS_STRING(__pyx_t_10);
  __pyx_t_13 = (__pyx_t_12 + PyBytes_GET_SIZE(__pyx_t_10));
  for (__pyx_t_14 = __pyx_t_12; __pyx_t_14 < __pyx_t_13; __pyx_t_14++) {
    __pyx_t_11 = __pyx_t_14;
    __pyx_v_c = (__pyx_t_11[0]);

    /* "View.MemoryView":515
 * 
 *         for i, c in enumerate(bytesvalue):
 *             itemp[i] = c             # <<<<<<<<<<<<<<
 * 
 *     @cname('getbuffer')
 */
    __pyx_v_i = __pyx_t_9;

    /* "View.MemoryView":514
 *             bytesvalue = struct.pack(self.view.format, value)
 * 
 *         for i, c in enumerate(bytesvalue):             # <<<<<<<<<<<<<<
 *             itemp[i] = c
 * 
 */
    __pyx_t_9 = (__pyx_t_9 + 1);

    /* "View.MemoryView":515
 * 
 *         for i, c in enumerate(bytesvalue):
 *             itemp[i] = c             # <<<<<<<<<<<<<<
 * 
 *     @cname('getbuffer')
 */
    (__pyx_v_itemp[__pyx_v_i]) = __pyx_v_c;
  }
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

  /* "View.MemoryView":501
 *             return result
 * 
 *     cdef assign_item_from_object(self, char *itemp, object value):             # <<<<<<<<<<<<<<
 *         """Only used if instantiated manually by the user, or if Cython doesn't
 *         know how to convert the type"""
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("View.MemoryView.memoryview.assign_item_from_object", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_struct);
  __Pyx_XDECREF(__pyx_v_bytesvalue);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":518
 * 
 *     @cname('getbuffer')
 *     def __getbuffer__(self, Py_buffer *info, int flags):             # <<<<<<<<<<<<<<
 *         if flags & PyBUF_WRITABLE and self.view.readonly:
 *             raise ValueError("Cannot create writable memory view from read-only memoryview")
 */

/* Python wrapper */
static CYTHON_UNUSED int __pyx_memoryview_getbuffer(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /*proto*/
static CYTHON_UNUSED int __pyx_memoryview_getbuffer(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getbuffer__ (wrapper)", 0);
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_8__getbuffer__(((struct __pyx_memoryview_obj *)__pyx_v_self), ((Py_buffer *)__pyx_v_info), ((int)__pyx_v_flags));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_8__getbuffer__(struct __pyx_memoryview_obj *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t *__pyx_t_4;
  char *__pyx_t_5;
  void *__pyx_t_6;
  int __pyx_t_7;
  Py_ssize_t __pyx_t_8;
  if (__pyx_v_info == NULL) {
    PyErr_SetString(PyExc_BufferError, "PyObject_GetBuffer: view==NULL argument is obsolete");
    return -1;
  }
  __Pyx_RefNannySetupContext("__getbuffer__", 0);
  __pyx_v_info->obj = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(__pyx_v_info->obj);

  /* "View.MemoryView":519
 *     @cname('getbuffer')
 *     def __getbuffer__(self, Py_buffer *info, int flags):
 *         if flags & PyBUF_WRITABLE and self.view.readonly:             # <<<<<<<<<<<<<<
 *             raise ValueError("Cannot create writable memory view from read-only memoryview")
 * 
 */
  __pyx_t_2 = ((__pyx_v_flags & PyBUF_WRITABLE) != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = (__pyx_v_self->view.readonly != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (unlikely(__pyx_t_1)) {

    /* "View.MemoryView":520
 *     def __getbuffer__(self, Py_buffer *info, int flags):
 *         if flags & PyBUF_WRITABLE and self.view.readonly:
 *             raise ValueError("Cannot create writable memory view from read-only memoryview")             # <<<<<<<<<<<<<<
 * 
 *         if flags & PyBUF_ND:
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__36, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 520, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(2, 520, __pyx_L1_error)

    /* "View.MemoryView":519
 *     @cname('getbuffer')
 *     def __getbuffer__(self, Py_buffer *info, int flags):
 *         if flags & PyBUF_WRITABLE and self.view.readonly:             # <<<<<<<<<<<<<<
 *             raise ValueError("Cannot create writable memory view from read-only memoryview")
 * 
 */
  }

  /* "View.MemoryView":522
 *             raise ValueError("Cannot create writable memory view from read-only memoryview")
 * 
 *         if flags & PyBUF_ND:             # <<<<<<<<<<<<<<
 *             info.shape = self.view.shape
 *         else:
 */
  __pyx_t_1 = ((__pyx_v_flags & PyBUF_ND) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":523
 * 
 *         if flags & PyBUF_ND:
 *             info.shape = self.view.shape             # <<<<<<<<<<<<<<
 *         else:
 *             info.shape = NULL
 */
    __pyx_t_4 = __pyx_v_self->view.shape;
    __pyx_v_info->shape = __pyx_t_4;

    /* "View.MemoryView":522
 *             raise ValueError("Cannot create writable memory view from read-only memoryview")
 * 
 *         if flags & PyBUF_ND:             # <<<<<<<<<<<<<<
 *             info.shape = self.view.shape
 *         else:
 */
    goto __pyx_L6;
  }

  /* "View.MemoryView":525
 *             info.shape = self.view.shape
 *         else:
 *             info.shape = NULL             # <<<<<<<<<<<<<<
 * 
 *         if flags & PyBUF_STRIDES:
 */
  /*else*/ {
    __pyx_v_info->shape = NULL;
  }
  __pyx_L6:;

  /* "View.MemoryView":527
 *             info.shape = NULL
 * 
 *         if flags & PyBUF_STRIDES:             # <<<<<<<<<<<<<<
 *             info.strides = self.view.strides
 *         else:
 */
  __pyx_t_1 = ((__pyx_v_flags & PyBUF_STRIDES) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":528
 * 
 *         if flags & PyBUF_STRIDES:
 *             info.strides = self.view.strides             # <<<<<<<<<<<<<<
 *         else:
 *             info.strides = NULL
 */
    __pyx_t_4 = __pyx_v_self->view.strides;
    __pyx_v_info->strides = __pyx_t_4;

    /* "View.MemoryView":527
 *             info.shape = NULL
 * 
 *         if flags & PyBUF_STRIDES:             # <<<<<<<<<<<<<<
 *             info.strides = self.view.strides
 *         else:
 */
    goto __pyx_L7;
  }

  /* "View.MemoryView":530
 *             info.strides = self.view.strides
 *         else:
 *             info.strides = NULL             # <<<<<<<<<<<<<<
 * 
 *         if flags & PyBUF_INDIRECT:
 */
  /*else*/ {
    __pyx_v_info->strides = NULL;
  }
  __pyx_L7:;

  /* "View.MemoryView":532
 *             info.strides = NULL
 * 
 *         if flags & PyBUF_INDIRECT:             # <<<<<<<<<<<<<<
 *             info.suboffsets = self.view.suboffsets
 *         else:
 */
  __pyx_t_1 = ((__pyx_v_flags & PyBUF_INDIRECT) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":533
 * 
 *         if flags & PyBUF_INDIRECT:
 *             info.suboffsets = self.view.suboffsets             # <<<<<<<<<<<<<<
 *         else:
 *             info.suboffsets = NULL
 */
    __pyx_t_4 = __pyx_v_self->view.suboffsets;
    __pyx_v_info->suboffsets = __pyx_t_4;

    /* "View.MemoryView":532
 *             info.strides = NULL
 * 
 *         if flags & PyBUF_INDIRECT:             # <<<<<<<<<<<<<<
 *             info.suboffsets = self.view.suboffsets
 *         else:
 */
    goto __pyx_L8;
  }

  /* "View.MemoryView":535
 *             info.suboffsets = self.view.suboffsets
 *         else:
 *             info.suboffsets = NULL             # <<<<<<<<<<<<<<
 * 
 *         if flags & PyBUF_FORMAT:
 */
  /*else*/ {
    __pyx_v_info->suboffsets = NULL;
  }
  __pyx_L8:;

  /* "View.MemoryView":537
 *             info.suboffsets = NULL
 * 
 *         if flags & PyBUF_FORMAT:             # <<<<<<<<<<<<<<
 *             info.format = self.view.format
 *         else:
 */
  __pyx_t_1 = ((__pyx_v_flags & PyBUF_FORMAT) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":538
 * 
 *         if flags & PyBUF_FORMAT:
 *             info.format = self.view.format             # <<<<<<<<<<<<<<
 *         else:
 *             info.format = NULL
 */
    __pyx_t_5 = __pyx_v_self->view.format;
    __pyx_v_info->format = __pyx_t_5;

    /* "View.MemoryView":537
 *             info.suboffsets = NULL
 * 
 *         if flags & PyBUF_FORMAT:             # <<<<<<<<<<<<<<
 *             info.format = self.view.format
 *         else:
 */
    goto __pyx_L9;
  }

  /* "View.MemoryView":540
 *             info.format = self.view.format
 *         else:
 *             info.format = NULL             # <<<<<<<<<<<<<<
 * 
 *         info.buf = self.view.buf
 */
  /*else*/ {
    __pyx_v_info->format = NULL;
  }
  __pyx_L9:;

  /* "View.MemoryView":542
 *             info.format = NULL
 * 
 *         info.buf = self.view.buf             # <<<<<<<<<<<<<<
 *         info.ndim = self.view.ndim
 *         info.itemsize = self.view.itemsize
 */
  __pyx_t_6 = __pyx_v_self->view.buf;
  __pyx_v_info->buf = __pyx_t_6;

  /* "View.MemoryView":543
 * 
 *         info.buf = self.view.buf
 *         info.ndim = self.view.ndim             # <<<<<<<<<<<<<<
 *         info.itemsize = self.view.itemsize
 *         info.len = self.view.len
 */
  __pyx_t_7 = __pyx_v_self->view.ndim;
  __pyx_v_info->ndim = __pyx_t_7;

  /* "View.MemoryView":544
 *         info.buf = self.view.buf
 *         info.ndim = self.view.ndim
 *         info.itemsize = self.view.itemsize             # <<<<<<<<<<<<<<
 *         info.len = self.view.len
 *         info.readonly = self.view.readonly
 */
  __pyx_t_8 = __pyx_v_self->view.itemsize;
  __pyx_v_info->itemsize = __pyx_t_8;

  /* "View.MemoryView":545
 *         info.ndim = self.view.ndim
 *         info.itemsize = self.view.itemsize
 *         info.len = self.view.len             # <<<<<<<<<<<<<<
 *         info.readonly = self.view.readonly
 *         info.obj = self
 */
  __pyx_t_8 = __pyx_v_self->view.len;
  __pyx_v_info->len = __pyx_t_8;

  /* "View.MemoryView":546
 *         info.itemsize = self.view.itemsize
 *         info.len = self.view.len
 *         info.readonly = self.view.readonly             # <<<<<<<<<<<<<<
 *         info.obj = self
 * 
 */
  __pyx_t_1 = __pyx_v_self->view.readonly;
  __pyx_v_info->readonly = __pyx_t_1;

  /* "View.MemoryView":547
 *         info.len = self.view.len
 *         info.readonly = self.view.readonly
 *         info.obj = self             # <<<<<<<<<<<<<<
 * 
 *     __pyx_getbuffer = capsule(<void *> &__pyx_memoryview_getbuffer, "getbuffer(obj, view, flags)")
 */
  __Pyx_INCREF(((PyObject *)__pyx_v_self));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
  __Pyx_GOTREF(__pyx_v_info->obj);
  __Pyx_DECREF(__pyx_v_info->obj);
  __pyx_v_info->obj = ((PyObject *)__pyx_v_self);

  /* "View.MemoryView":518
 * 
 *     @cname('getbuffer')
 *     def __getbuffer__(self, Py_buffer *info, int flags):             # <<<<<<<<<<<<<<
 *         if flags & PyBUF_WRITABLE and self.view.readonly:
 *             raise ValueError("Cannot create writable memory view from read-only memoryview")
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView.memoryview.__getbuffer__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  if (__pyx_v_info->obj != NULL) {
    __Pyx_GOTREF(__pyx_v_info->obj);
    __Pyx_DECREF(__pyx_v_info->obj); __pyx_v_info->obj = 0;
  }
  goto __pyx_L2;
  __pyx_L0:;
  if (__pyx_v_info->obj == Py_None) {
    __Pyx_GOTREF(__pyx_v_info->obj);
    __Pyx_DECREF(__pyx_v_info->obj); __pyx_v_info->obj = 0;
  }
  __pyx_L2:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":553
 * 
 *     @property
 *     def T(self):             # <<<<<<<<<<<<<<
 *         cdef _memoryviewslice result = memoryview_copy(self)
 *         transpose_memslice(&result.from_slice)
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_1T_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_1T_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_10memoryview_1T___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_1T___get__(struct __pyx_memoryview_obj *__pyx_v_self) {
  struct __pyx_memoryviewslice_obj *__pyx_v_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":554
 *     @property
 *     def T(self):
 *         cdef _memoryviewslice result = memoryview_copy(self)             # <<<<<<<<<<<<<<
 *         transpose_memslice(&result.from_slice)
 *         return result
 */
  __pyx_t_1 = __pyx_memoryview_copy_object(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 554, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_memoryviewslice_type))))) __PYX_ERR(2, 554, __pyx_L1_error)
  __pyx_v_result = ((struct __pyx_memoryviewslice_obj *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "View.MemoryView":555
 *     def T(self):
 *         cdef _memoryviewslice result = memoryview_copy(self)
 *         transpose_memslice(&result.from_slice)             # <<<<<<<<<<<<<<
 *         return result
 * 
 */
  __pyx_t_2 = __pyx_memslice_transpose((&__pyx_v_result->from_slice)); if (unlikely(__pyx_t_2 == ((int)0))) __PYX_ERR(2, 555, __pyx_L1_error)

  /* "View.MemoryView":556
 *         cdef _memoryviewslice result = memoryview_copy(self)
 *         transpose_memslice(&result.from_slice)
 *         return result             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_result));
  __pyx_r = ((PyObject *)__pyx_v_result);
  goto __pyx_L0;

  /* "View.MemoryView":553
 * 
 *     @property
 *     def T(self):             # <<<<<<<<<<<<<<
 *         cdef _memoryviewslice result = memoryview_copy(self)
 *         transpose_memslice(&result.from_slice)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.memoryview.T.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":559
 * 
 *     @property
 *     def base(self):             # <<<<<<<<<<<<<<
 *         return self.obj
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_4base_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_4base_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_10memoryview_4base___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_4base___get__(struct __pyx_memoryview_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":560
 *     @property
 *     def base(self):
 *         return self.obj             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->obj);
  __pyx_r = __pyx_v_self->obj;
  goto __pyx_L0;

  /* "View.MemoryView":559
 * 
 *     @property
 *     def base(self):             # <<<<<<<<<<<<<<
 *         return self.obj
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":563
 * 
 *     @property
 *     def shape(self):             # <<<<<<<<<<<<<<
 *         return tuple([length for length in self.view.shape[:self.view.ndim]])
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_5shape_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_5shape_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_10memoryview_5shape___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_5shape___get__(struct __pyx_memoryview_obj *__pyx_v_self) {
  Py_ssize_t __pyx_v_length;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t *__pyx_t_2;
  Py_ssize_t *__pyx_t_3;
  Py_ssize_t *__pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":564
 *     @property
 *     def shape(self):
 *         return tuple([length for length in self.view.shape[:self.view.ndim]])             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 564, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = (__pyx_v_self->view.shape + __pyx_v_self->view.ndim);
  for (__pyx_t_4 = __pyx_v_self->view.shape; __pyx_t_4 < __pyx_t_3; __pyx_t_4++) {
    __pyx_t_2 = __pyx_t_4;
    __pyx_v_length = (__pyx_t_2[0]);
    __pyx_t_5 = PyInt_FromSsize_t(__pyx_v_length); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 564, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_5))) __PYX_ERR(2, 564, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __pyx_t_5 = PyList_AsTuple(((PyObject*)__pyx_t_1)); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 564, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_5;
  __pyx_t_5 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":563
 * 
 *     @property
 *     def shape(self):             # <<<<<<<<<<<<<<
 *         return tuple([length for length in self.view.shape[:self.view.ndim]])
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("View.MemoryView.memoryview.shape.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":567
 * 
 *     @property
 *     def strides(self):             # <<<<<<<<<<<<<<
 *         if self.view.strides == NULL:
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_7strides_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_7strides_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_10memoryview_7strides___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_7strides___get__(struct __pyx_memoryview_obj *__pyx_v_self) {
  Py_ssize_t __pyx_v_stride;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t *__pyx_t_3;
  Py_ssize_t *__pyx_t_4;
  Py_ssize_t *__pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":568
 *     @property
 *     def strides(self):
 *         if self.view.strides == NULL:             # <<<<<<<<<<<<<<
 * 
 *             raise ValueError("Buffer view does not expose strides")
 */
  __pyx_t_1 = ((__pyx_v_self->view.strides == NULL) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "View.MemoryView":570
 *         if self.view.strides == NULL:
 * 
 *             raise ValueError("Buffer view does not expose strides")             # <<<<<<<<<<<<<<
 * 
 *         return tuple([stride for stride in self.view.strides[:self.view.ndim]])
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__37, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 570, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(2, 570, __pyx_L1_error)

    /* "View.MemoryView":568
 *     @property
 *     def strides(self):
 *         if self.view.strides == NULL:             # <<<<<<<<<<<<<<
 * 
 *             raise ValueError("Buffer view does not expose strides")
 */
  }

  /* "View.MemoryView":572
 *             raise ValueError("Buffer view does not expose strides")
 * 
 *         return tuple([stride for stride in self.view.strides[:self.view.ndim]])             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 572, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = (__pyx_v_self->view.strides + __pyx_v_self->view.ndim);
  for (__pyx_t_5 = __pyx_v_self->view.strides; __pyx_t_5 < __pyx_t_4; __pyx_t_5++) {
    __pyx_t_3 = __pyx_t_5;
    __pyx_v_stride = (__pyx_t_3[0]);
    __pyx_t_6 = PyInt_FromSsize_t(__pyx_v_stride); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 572, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (unlikely(__Pyx_ListComp_Append(__pyx_t_2, (PyObject*)__pyx_t_6))) __PYX_ERR(2, 572, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __pyx_t_6 = PyList_AsTuple(((PyObject*)__pyx_t_2)); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 572, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_6;
  __pyx_t_6 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":567
 * 
 *     @property
 *     def strides(self):             # <<<<<<<<<<<<<<
 *         if self.view.strides == NULL:
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("View.MemoryView.memoryview.strides.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":575
 * 
 *     @property
 *     def suboffsets(self):             # <<<<<<<<<<<<<<
 *         if self.view.suboffsets == NULL:
 *             return (-1,) * self.view.ndim
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_10suboffsets_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_10suboffsets_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_10memoryview_10suboffsets___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_10suboffsets___get__(struct __pyx_memoryview_obj *__pyx_v_self) {
  Py_ssize_t __pyx_v_suboffset;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t *__pyx_t_4;
  Py_ssize_t *__pyx_t_5;
  Py_ssize_t *__pyx_t_6;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":576
 *     @property
 *     def suboffsets(self):
 *         if self.view.suboffsets == NULL:             # <<<<<<<<<<<<<<
 *             return (-1,) * self.view.ndim
 * 
 */
  __pyx_t_1 = ((__pyx_v_self->view.suboffsets == NULL) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":577
 *     def suboffsets(self):
 *         if self.view.suboffsets == NULL:
 *             return (-1,) * self.view.ndim             # <<<<<<<<<<<<<<
 * 
 *         return tuple([suboffset for suboffset in self.view.suboffsets[:self.view.ndim]])
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_self->view.ndim); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 577, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PyNumber_Multiply(__pyx_tuple__38, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 577, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "View.MemoryView":576
 *     @property
 *     def suboffsets(self):
 *         if self.view.suboffsets == NULL:             # <<<<<<<<<<<<<<
 *             return (-1,) * self.view.ndim
 * 
 */
  }

  /* "View.MemoryView":579
 *             return (-1,) * self.view.ndim
 * 
 *         return tuple([suboffset for suboffset in self.view.suboffsets[:self.view.ndim]])             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 579, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = (__pyx_v_self->view.suboffsets + __pyx_v_self->view.ndim);
  for (__pyx_t_6 = __pyx_v_self->view.suboffsets; __pyx_t_6 < __pyx_t_5; __pyx_t_6++) {
    __pyx_t_4 = __pyx_t_6;
    __pyx_v_suboffset = (__pyx_t_4[0]);
    __pyx_t_2 = PyInt_FromSsize_t(__pyx_v_suboffset); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 579, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (unlikely(__Pyx_ListComp_Append(__pyx_t_3, (PyObject*)__pyx_t_2))) __PYX_ERR(2, 579, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_t_2 = PyList_AsTuple(((PyObject*)__pyx_t_3)); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 579, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":575
 * 
 *     @property
 *     def suboffsets(self):             # <<<<<<<<<<<<<<
 *         if self.view.suboffsets == NULL:
 *             return (-1,) * self.view.ndim
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView.memoryview.suboffsets.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":582
 * 
 *     @property
 *     def ndim(self):             # <<<<<<<<<<<<<<
 *         return self.view.ndim
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_4ndim_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_4ndim_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_10memoryview_4ndim___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_4ndim___get__(struct __pyx_memoryview_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":583
 *     @property
 *     def ndim(self):
 *         return self.view.ndim             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->view.ndim); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 583, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":582
 * 
 *     @property
 *     def ndim(self):             # <<<<<<<<<<<<<<
 *         return self.view.ndim
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.memoryview.ndim.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":586
 * 
 *     @property
 *     def itemsize(self):             # <<<<<<<<<<<<<<
 *         return self.view.itemsize
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_8itemsize_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_8itemsize_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_10memoryview_8itemsize___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_8itemsize___get__(struct __pyx_memoryview_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":587
 *     @property
 *     def itemsize(self):
 *         return self.view.itemsize             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyInt_FromSsize_t(__pyx_v_self->view.itemsize); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 587, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":586
 * 
 *     @property
 *     def itemsize(self):             # <<<<<<<<<<<<<<
 *         return self.view.itemsize
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.memoryview.itemsize.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":590
 * 
 *     @property
 *     def nbytes(self):             # <<<<<<<<<<<<<<
 *         return self.size * self.view.itemsize
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_6nbytes_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_6nbytes_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_10memoryview_6nbytes___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_6nbytes___get__(struct __pyx_memoryview_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":591
 *     @property
 *     def nbytes(self):
 *         return self.size * self.view.itemsize             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 591, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyInt_FromSsize_t(__pyx_v_self->view.itemsize); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 591, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyNumber_Multiply(__pyx_t_1, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 591, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":590
 * 
 *     @property
 *     def nbytes(self):             # <<<<<<<<<<<<<<
 *         return self.size * self.view.itemsize
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView.memoryview.nbytes.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":594
 * 
 *     @property
 *     def size(self):             # <<<<<<<<<<<<<<
 *         if self._size is None:
 *             result = 1
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_4size_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_4size_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_10memoryview_4size___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_4size___get__(struct __pyx_memoryview_obj *__pyx_v_self) {
  PyObject *__pyx_v_result = NULL;
  PyObject *__pyx_v_length = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  Py_ssize_t *__pyx_t_3;
  Py_ssize_t *__pyx_t_4;
  Py_ssize_t *__pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":595
 *     @property
 *     def size(self):
 *         if self._size is None:             # <<<<<<<<<<<<<<
 *             result = 1
 * 
 */
  __pyx_t_1 = (__pyx_v_self->_size == Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":596
 *     def size(self):
 *         if self._size is None:
 *             result = 1             # <<<<<<<<<<<<<<
 * 
 *             for length in self.view.shape[:self.view.ndim]:
 */
    __Pyx_INCREF(__pyx_int_1);
    __pyx_v_result = __pyx_int_1;

    /* "View.MemoryView":598
 *             result = 1
 * 
 *             for length in self.view.shape[:self.view.ndim]:             # <<<<<<<<<<<<<<
 *                 result *= length
 * 
 */
    __pyx_t_4 = (__pyx_v_self->view.shape + __pyx_v_self->view.ndim);
    for (__pyx_t_5 = __pyx_v_self->view.shape; __pyx_t_5 < __pyx_t_4; __pyx_t_5++) {
      __pyx_t_3 = __pyx_t_5;
      __pyx_t_6 = PyInt_FromSsize_t((__pyx_t_3[0])); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 598, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_XDECREF_SET(__pyx_v_length, __pyx_t_6);
      __pyx_t_6 = 0;

      /* "View.MemoryView":599
 * 
 *             for length in self.view.shape[:self.view.ndim]:
 *                 result *= length             # <<<<<<<<<<<<<<
 * 
 *             self._size = result
 */
      __pyx_t_6 = PyNumber_InPlaceMultiply(__pyx_v_result, __pyx_v_length); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 599, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF_SET(__pyx_v_result, __pyx_t_6);
      __pyx_t_6 = 0;
    }

    /* "View.MemoryView":601
 *                 result *= length
 * 
 *             self._size = result             # <<<<<<<<<<<<<<
 * 
 *         return self._size
 */
    __Pyx_INCREF(__pyx_v_result);
    __Pyx_GIVEREF(__pyx_v_result);
    __Pyx_GOTREF(__pyx_v_self->_size);
    __Pyx_DECREF(__pyx_v_self->_size);
    __pyx_v_self->_size = __pyx_v_result;

    /* "View.MemoryView":595
 *     @property
 *     def size(self):
 *         if self._size is None:             # <<<<<<<<<<<<<<
 *             result = 1
 * 
 */
  }

  /* "View.MemoryView":603
 *             self._size = result
 * 
 *         return self._size             # <<<<<<<<<<<<<<
 * 
 *     def __len__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_size);
  __pyx_r = __pyx_v_self->_size;
  goto __pyx_L0;

  /* "View.MemoryView":594
 * 
 *     @property
 *     def size(self):             # <<<<<<<<<<<<<<
 *         if self._size is None:
 *             result = 1
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("View.MemoryView.memoryview.size.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XDECREF(__pyx_v_length);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":605
 *         return self._size
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         if self.view.ndim >= 1:
 *             return self.view.shape[0]
 */

/* Python wrapper */
static Py_ssize_t __pyx_memoryview___len__(PyObject *__pyx_v_self); /*proto*/
static Py_ssize_t __pyx_memoryview___len__(PyObject *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__ (wrapper)", 0);
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_10__len__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static Py_ssize_t __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_10__len__(struct __pyx_memoryview_obj *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("__len__", 0);

  /* "View.MemoryView":606
 * 
 *     def __len__(self):
 *         if self.view.ndim >= 1:             # <<<<<<<<<<<<<<
 *             return self.view.shape[0]
 * 
 */
  __pyx_t_1 = ((__pyx_v_self->view.ndim >= 1) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":607
 *     def __len__(self):
 *         if self.view.ndim >= 1:
 *             return self.view.shape[0]             # <<<<<<<<<<<<<<
 * 
 *         return 0
 */
    __pyx_r = (__pyx_v_self->view.shape[0]);
    goto __pyx_L0;

    /* "View.MemoryView":606
 * 
 *     def __len__(self):
 *         if self.view.ndim >= 1:             # <<<<<<<<<<<<<<
 *             return self.view.shape[0]
 * 
 */
  }

  /* "View.MemoryView":609
 *             return self.view.shape[0]
 * 
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "View.MemoryView":605
 *         return self._size
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         if self.view.ndim >= 1:
 *             return self.view.shape[0]
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":611
 *         return 0
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return "<MemoryView of %r at 0x%x>" % (self.base.__class__.__name__,
 *                                                id(self))
 */

/* Python wrapper */
static PyObject *__pyx_memoryview___repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_memoryview___repr__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_12__repr__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_12__repr__(struct __pyx_memoryview_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "View.MemoryView":612
 * 
 *     def __repr__(self):
 *         return "<MemoryView of %r at 0x%x>" % (self.base.__class__.__name__,             # <<<<<<<<<<<<<<
 *                                                id(self))
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_base); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 612, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_class); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 612, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_name_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 612, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "View.MemoryView":613
 *     def __repr__(self):
 *         return "<MemoryView of %r at 0x%x>" % (self.base.__class__.__name__,
 *                                                id(self))             # <<<<<<<<<<<<<<
 * 
 *     def __str__(self):
 */
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_id, ((PyObject *)__pyx_v_self)); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 613, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "View.MemoryView":612
 * 
 *     def __repr__(self):
 *         return "<MemoryView of %r at 0x%x>" % (self.base.__class__.__name__,             # <<<<<<<<<<<<<<
 *                                                id(self))
 * 
 */
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 612, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_2);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyString_Format(__pyx_kp_s_MemoryView_of_r_at_0x_x, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 612, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":611
 *         return 0
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return "<MemoryView of %r at 0x%x>" % (self.base.__class__.__name__,
 *                                                id(self))
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView.memoryview.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":615
 *                                                id(self))
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return "<MemoryView of %r object>" % (self.base.__class__.__name__,)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_memoryview___str__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_memoryview___str__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__str__ (wrapper)", 0);
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_14__str__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_14__str__(struct __pyx_memoryview_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("__str__", 0);

  /* "View.MemoryView":616
 * 
 *     def __str__(self):
 *         return "<MemoryView of %r object>" % (self.base.__class__.__name__,)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_base); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 616, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_class); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 616, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_name_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 616, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 616, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_MemoryView_of_r_object, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 616, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":615
 *                                                id(self))
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return "<MemoryView of %r object>" % (self.base.__class__.__name__,)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("View.MemoryView.memoryview.__str__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":619
 * 
 * 
 *     def is_c_contig(self):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice *mslice
 *         cdef __Pyx_memviewslice tmp
 */

/* Python wrapper */
static PyObject *__pyx_memoryview_is_c_contig(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_memoryview_is_c_contig(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("is_c_contig (wrapper)", 0);
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_16is_c_contig(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_16is_c_contig(struct __pyx_memoryview_obj *__pyx_v_self) {
  __Pyx_memviewslice *__pyx_v_mslice;
  __Pyx_memviewslice __pyx_v_tmp;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("is_c_contig", 0);

  /* "View.MemoryView":622
 *         cdef __Pyx_memviewslice *mslice
 *         cdef __Pyx_memviewslice tmp
 *         mslice = get_slice_from_memview(self, &tmp)             # <<<<<<<<<<<<<<
 *         return slice_is_contig(mslice[0], 'C', self.view.ndim)
 * 
 */
  __pyx_v_mslice = __pyx_memoryview_get_slice_from_memoryview(__pyx_v_self, (&__pyx_v_tmp));

  /* "View.MemoryView":623
 *         cdef __Pyx_memviewslice tmp
 *         mslice = get_slice_from_memview(self, &tmp)
 *         return slice_is_contig(mslice[0], 'C', self.view.ndim)             # <<<<<<<<<<<<<<
 * 
 *     def is_f_contig(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_memviewslice_is_contig((__pyx_v_mslice[0]), 'C', __pyx_v_self->view.ndim)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 623, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":619
 * 
 * 
 *     def is_c_contig(self):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice *mslice
 *         cdef __Pyx_memviewslice tmp
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.memoryview.is_c_contig", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":625
 *         return slice_is_contig(mslice[0], 'C', self.view.ndim)
 * 
 *     def is_f_contig(self):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice *mslice
 *         cdef __Pyx_memviewslice tmp
 */

/* Python wrapper */
static PyObject *__pyx_memoryview_is_f_contig(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_memoryview_is_f_contig(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("is_f_contig (wrapper)", 0);
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_18is_f_contig(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_18is_f_contig(struct __pyx_memoryview_obj *__pyx_v_self) {
  __Pyx_memviewslice *__pyx_v_mslice;
  __Pyx_memviewslice __pyx_v_tmp;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("is_f_contig", 0);

  /* "View.MemoryView":628
 *         cdef __Pyx_memviewslice *mslice
 *         cdef __Pyx_memviewslice tmp
 *         mslice = get_slice_from_memview(self, &tmp)             # <<<<<<<<<<<<<<
 *         return slice_is_contig(mslice[0], 'F', self.view.ndim)
 * 
 */
  __pyx_v_mslice = __pyx_memoryview_get_slice_from_memoryview(__pyx_v_self, (&__pyx_v_tmp));

  /* "View.MemoryView":629
 *         cdef __Pyx_memviewslice tmp
 *         mslice = get_slice_from_memview(self, &tmp)
 *         return slice_is_contig(mslice[0], 'F', self.view.ndim)             # <<<<<<<<<<<<<<
 * 
 *     def copy(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_memviewslice_is_contig((__pyx_v_mslice[0]), 'F', __pyx_v_self->view.ndim)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 629, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":625
 *         return slice_is_contig(mslice[0], 'C', self.view.ndim)
 * 
 *     def is_f_contig(self):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice *mslice
 *         cdef __Pyx_memviewslice tmp
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.memoryview.is_f_contig", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":631
 *         return slice_is_contig(mslice[0], 'F', self.view.ndim)
 * 
 *     def copy(self):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice mslice
 *         cdef int flags = self.flags & ~PyBUF_F_CONTIGUOUS
 */

/* Python wrapper */
static PyObject *__pyx_memoryview_copy(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_memoryview_copy(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("copy (wrapper)", 0);
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_20copy(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_20copy(struct __pyx_memoryview_obj *__pyx_v_self) {
  __Pyx_memviewslice __pyx_v_mslice;
  int __pyx_v_flags;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_memviewslice __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("copy", 0);

  /* "View.MemoryView":633
 *     def copy(self):
 *         cdef __Pyx_memviewslice mslice
 *         cdef int flags = self.flags & ~PyBUF_F_CONTIGUOUS             # <<<<<<<<<<<<<<
 * 
 *         slice_copy(self, &mslice)
 */
  __pyx_v_flags = (__pyx_v_self->flags & (~PyBUF_F_CONTIGUOUS));

  /* "View.MemoryView":635
 *         cdef int flags = self.flags & ~PyBUF_F_CONTIGUOUS
 * 
 *         slice_copy(self, &mslice)             # <<<<<<<<<<<<<<
 *         mslice = slice_copy_contig(&mslice, "c", self.view.ndim,
 *                                    self.view.itemsize,
 */
  __pyx_memoryview_slice_copy(__pyx_v_self, (&__pyx_v_mslice));

  /* "View.MemoryView":636
 * 
 *         slice_copy(self, &mslice)
 *         mslice = slice_copy_contig(&mslice, "c", self.view.ndim,             # <<<<<<<<<<<<<<
 *                                    self.view.itemsize,
 *                                    flags|PyBUF_C_CONTIGUOUS,
 */
  __pyx_t_1 = __pyx_memoryview_copy_new_contig((&__pyx_v_mslice), ((char *)"c"), __pyx_v_self->view.ndim, __pyx_v_self->view.itemsize, (__pyx_v_flags | PyBUF_C_CONTIGUOUS), __pyx_v_self->dtype_is_object); if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 636, __pyx_L1_error)
  __pyx_v_mslice = __pyx_t_1;

  /* "View.MemoryView":641
 *                                    self.dtype_is_object)
 * 
 *         return memoryview_copy_from_slice(self, &mslice)             # <<<<<<<<<<<<<<
 * 
 *     def copy_fortran(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __pyx_memoryview_copy_object_from_slice(__pyx_v_self, (&__pyx_v_mslice)); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 641, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":631
 *         return slice_is_contig(mslice[0], 'F', self.view.ndim)
 * 
 *     def copy(self):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice mslice
 *         cdef int flags = self.flags & ~PyBUF_F_CONTIGUOUS
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("View.MemoryView.memoryview.copy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":643
 *         return memoryview_copy_from_slice(self, &mslice)
 * 
 *     def copy_fortran(self):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice src, dst
 *         cdef int flags = self.flags & ~PyBUF_C_CONTIGUOUS
 */

/* Python wrapper */
static PyObject *__pyx_memoryview_copy_fortran(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_memoryview_copy_fortran(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("copy_fortran (wrapper)", 0);
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_22copy_fortran(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_22copy_fortran(struct __pyx_memoryview_obj *__pyx_v_self) {
  __Pyx_memviewslice __pyx_v_src;
  __Pyx_memviewslice __pyx_v_dst;
  int __pyx_v_flags;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_memviewslice __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("copy_fortran", 0);

  /* "View.MemoryView":645
 *     def copy_fortran(self):
 *         cdef __Pyx_memviewslice src, dst
 *         cdef int flags = self.flags & ~PyBUF_C_CONTIGUOUS             # <<<<<<<<<<<<<<
 * 
 *         slice_copy(self, &src)
 */
  __pyx_v_flags = (__pyx_v_self->flags & (~PyBUF_C_CONTIGUOUS));

  /* "View.MemoryView":647
 *         cdef int flags = self.flags & ~PyBUF_C_CONTIGUOUS
 * 
 *         slice_copy(self, &src)             # <<<<<<<<<<<<<<
 *         dst = slice_copy_contig(&src, "fortran", self.view.ndim,
 *                                 self.view.itemsize,
 */
  __pyx_memoryview_slice_copy(__pyx_v_self, (&__pyx_v_src));

  /* "View.MemoryView":648
 * 
 *         slice_copy(self, &src)
 *         dst = slice_copy_contig(&src, "fortran", self.view.ndim,             # <<<<<<<<<<<<<<
 *                                 self.view.itemsize,
 *                                 flags|PyBUF_F_CONTIGUOUS,
 */
  __pyx_t_1 = __pyx_memoryview_copy_new_contig((&__pyx_v_src), ((char *)"fortran"), __pyx_v_self->view.ndim, __pyx_v_self->view.itemsize, (__pyx_v_flags | PyBUF_F_CONTIGUOUS), __pyx_v_self->dtype_is_object); if (unlikely(PyErr_Occurred())) __PYX_ERR(2, 648, __pyx_L1_error)
  __pyx_v_dst = __pyx_t_1;

  /* "View.MemoryView":653
 *                                 self.dtype_is_object)
 * 
 *         return memoryview_copy_from_slice(self, &dst)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __pyx_memoryview_copy_object_from_slice(__pyx_v_self, (&__pyx_v_dst)); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 653, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":643
 *         return memoryview_copy_from_slice(self, &mslice)
 * 
 *     def copy_fortran(self):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice src, dst
 *         cdef int flags = self.flags & ~PyBUF_C_CONTIGUOUS
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("View.MemoryView.memoryview.copy_fortran", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw___pyx_memoryview_1__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw___pyx_memoryview_1__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf___pyx_memoryview___reduce_cython__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf___pyx_memoryview___reduce_cython__(CYTHON_UNUSED struct __pyx_memoryview_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__39, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(2, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.memoryview.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw___pyx_memoryview_3__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw___pyx_memoryview_3__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf___pyx_memoryview_2__setstate_cython__(((struct __pyx_memoryview_obj *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf___pyx_memoryview_2__setstate_cython__(CYTHON_UNUSED struct __pyx_memoryview_obj *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__40, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(2, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.memoryview.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":657
 * 
 * @cname('__pyx_memoryview_new')
 * cdef memoryview_cwrapper(object o, int flags, bint dtype_is_object, __Pyx_TypeInfo *typeinfo):             # <<<<<<<<<<<<<<
 *     cdef memoryview result = memoryview(o, flags, dtype_is_object)
 *     result.typeinfo = typeinfo
 */

static PyObject *__pyx_memoryview_new(PyObject *__pyx_v_o, int __pyx_v_flags, int __pyx_v_dtype_is_object, __Pyx_TypeInfo *__pyx_v_typeinfo) {
  struct __pyx_memoryview_obj *__pyx_v_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("memoryview_cwrapper", 0);

  /* "View.MemoryView":658
 * @cname('__pyx_memoryview_new')
 * cdef memoryview_cwrapper(object o, int flags, bint dtype_is_object, __Pyx_TypeInfo *typeinfo):
 *     cdef memoryview result = memoryview(o, flags, dtype_is_object)             # <<<<<<<<<<<<<<
 *     result.typeinfo = typeinfo
 *     return result
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_flags); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 658, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_v_dtype_is_object); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 658, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 658, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_v_o);
  __Pyx_GIVEREF(__pyx_v_o);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_o);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_t_2);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_memoryview_type), __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 658, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_result = ((struct __pyx_memoryview_obj *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "View.MemoryView":659
 * cdef memoryview_cwrapper(object o, int flags, bint dtype_is_object, __Pyx_TypeInfo *typeinfo):
 *     cdef memoryview result = memoryview(o, flags, dtype_is_object)
 *     result.typeinfo = typeinfo             # <<<<<<<<<<<<<<
 *     return result
 * 
 */
  __pyx_v_result->typeinfo = __pyx_v_typeinfo;

  /* "View.MemoryView":660
 *     cdef memoryview result = memoryview(o, flags, dtype_is_object)
 *     result.typeinfo = typeinfo
 *     return result             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_check')
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_result));
  __pyx_r = ((PyObject *)__pyx_v_result);
  goto __pyx_L0;

  /* "View.MemoryView":657
 * 
 * @cname('__pyx_memoryview_new')
 * cdef memoryview_cwrapper(object o, int flags, bint dtype_is_object, __Pyx_TypeInfo *typeinfo):             # <<<<<<<<<<<<<<
 *     cdef memoryview result = memoryview(o, flags, dtype_is_object)
 *     result.typeinfo = typeinfo
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView.memoryview_cwrapper", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":663
 * 
 * @cname('__pyx_memoryview_check')
 * cdef inline bint memoryview_check(object o):             # <<<<<<<<<<<<<<
 *     return isinstance(o, memoryview)
 * 
 */

static CYTHON_INLINE int __pyx_memoryview_check(PyObject *__pyx_v_o) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("memoryview_check", 0);

  /* "View.MemoryView":664
 * @cname('__pyx_memoryview_check')
 * cdef inline bint memoryview_check(object o):
 *     return isinstance(o, memoryview)             # <<<<<<<<<<<<<<
 * 
 * cdef tuple _unellipsify(object index, int ndim):
 */
  __pyx_t_1 = __Pyx_TypeCheck(__pyx_v_o, __pyx_memoryview_type); 
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "View.MemoryView":663
 * 
 * @cname('__pyx_memoryview_check')
 * cdef inline bint memoryview_check(object o):             # <<<<<<<<<<<<<<
 *     return isinstance(o, memoryview)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":666
 *     return isinstance(o, memoryview)
 * 
 * cdef tuple _unellipsify(object index, int ndim):             # <<<<<<<<<<<<<<
 *     """
 *     Replace all ellipses with full slices and fill incomplete indices with
 */

static PyObject *_unellipsify(PyObject *__pyx_v_index, int __pyx_v_ndim) {
  PyObject *__pyx_v_tup = NULL;
  PyObject *__pyx_v_result = NULL;
  int __pyx_v_have_slices;
  int __pyx_v_seen_ellipsis;
  CYTHON_UNUSED PyObject *__pyx_v_idx = NULL;
  PyObject *__pyx_v_item = NULL;
  Py_ssize_t __pyx_v_nslices;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  Py_ssize_t __pyx_t_5;
  PyObject *(*__pyx_t_6)(PyObject *);
  PyObject *__pyx_t_7 = NULL;
  Py_ssize_t __pyx_t_8;
  int __pyx_t_9;
  int __pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  __Pyx_RefNannySetupContext("_unellipsify", 0);

  /* "View.MemoryView":671
 *     full slices.
 *     """
 *     if not isinstance(index, tuple):             # <<<<<<<<<<<<<<
 *         tup = (index,)
 *     else:
 */
  __pyx_t_1 = PyTuple_Check(__pyx_v_index); 
  __pyx_t_2 = ((!(__pyx_t_1 != 0)) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":672
 *     """
 *     if not isinstance(index, tuple):
 *         tup = (index,)             # <<<<<<<<<<<<<<
 *     else:
 *         tup = index
 */
    __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 672, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_v_index);
    __Pyx_GIVEREF(__pyx_v_index);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_index);
    __pyx_v_tup = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "View.MemoryView":671
 *     full slices.
 *     """
 *     if not isinstance(index, tuple):             # <<<<<<<<<<<<<<
 *         tup = (index,)
 *     else:
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":674
 *         tup = (index,)
 *     else:
 *         tup = index             # <<<<<<<<<<<<<<
 * 
 *     result = []
 */
  /*else*/ {
    __Pyx_INCREF(__pyx_v_index);
    __pyx_v_tup = __pyx_v_index;
  }
  __pyx_L3:;

  /* "View.MemoryView":676
 *         tup = index
 * 
 *     result = []             # <<<<<<<<<<<<<<
 *     have_slices = False
 *     seen_ellipsis = False
 */
  __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 676, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v_result = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "View.MemoryView":677
 * 
 *     result = []
 *     have_slices = False             # <<<<<<<<<<<<<<
 *     seen_ellipsis = False
 *     for idx, item in enumerate(tup):
 */
  __pyx_v_have_slices = 0;

  /* "View.MemoryView":678
 *     result = []
 *     have_slices = False
 *     seen_ellipsis = False             # <<<<<<<<<<<<<<
 *     for idx, item in enumerate(tup):
 *         if item is Ellipsis:
 */
  __pyx_v_seen_ellipsis = 0;

  /* "View.MemoryView":679
 *     have_slices = False
 *     seen_ellipsis = False
 *     for idx, item in enumerate(tup):             # <<<<<<<<<<<<<<
 *         if item is Ellipsis:
 *             if not seen_ellipsis:
 */
  __Pyx_INCREF(__pyx_int_0);
  __pyx_t_3 = __pyx_int_0;
  if (likely(PyList_CheckExact(__pyx_v_tup)) || PyTuple_CheckExact(__pyx_v_tup)) {
    __pyx_t_4 = __pyx_v_tup; __Pyx_INCREF(__pyx_t_4); __pyx_t_5 = 0;
    __pyx_t_6 = NULL;
  } else {
    __pyx_t_5 = -1; __pyx_t_4 = PyObject_GetIter(__pyx_v_tup); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 679, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = Py_TYPE(__pyx_t_4)->tp_iternext; if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 679, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_6)) {
      if (likely(PyList_CheckExact(__pyx_t_4))) {
        if (__pyx_t_5 >= PyList_GET_SIZE(__pyx_t_4)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_7 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_5); __Pyx_INCREF(__pyx_t_7); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(2, 679, __pyx_L1_error)
        #else
        __pyx_t_7 = PySequence_ITEM(__pyx_t_4, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 679, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        #endif
      } else {
        if (__pyx_t_5 >= PyTuple_GET_SIZE(__pyx_t_4)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_7 = PyTuple_GET_ITEM(__pyx_t_4, __pyx_t_5); __Pyx_INCREF(__pyx_t_7); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(2, 679, __pyx_L1_error)
        #else
        __pyx_t_7 = PySequence_ITEM(__pyx_t_4, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 679, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        #endif
      }
    } else {
      __pyx_t_7 = __pyx_t_6(__pyx_t_4);
      if (unlikely(!__pyx_t_7)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(2, 679, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_7);
    }
    __Pyx_XDECREF_SET(__pyx_v_item, __pyx_t_7);
    __pyx_t_7 = 0;
    __Pyx_INCREF(__pyx_t_3);
    __Pyx_XDECREF_SET(__pyx_v_idx, __pyx_t_3);
    __pyx_t_7 = __Pyx_PyInt_AddObjC(__pyx_t_3, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 679, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_3);
    __pyx_t_3 = __pyx_t_7;
    __pyx_t_7 = 0;

    /* "View.MemoryView":680
 *     seen_ellipsis = False
 *     for idx, item in enumerate(tup):
 *         if item is Ellipsis:             # <<<<<<<<<<<<<<
 *             if not seen_ellipsis:
 *                 result.extend([slice(None)] * (ndim - len(tup) + 1))
 */
    __pyx_t_2 = (__pyx_v_item == __pyx_builtin_Ellipsis);
    __pyx_t_1 = (__pyx_t_2 != 0);
    if (__pyx_t_1) {

      /* "View.MemoryView":681
 *     for idx, item in enumerate(tup):
 *         if item is Ellipsis:
 *             if not seen_ellipsis:             # <<<<<<<<<<<<<<
 *                 result.extend([slice(None)] * (ndim - len(tup) + 1))
 *                 seen_ellipsis = True
 */
      __pyx_t_1 = ((!(__pyx_v_seen_ellipsis != 0)) != 0);
      if (__pyx_t_1) {

        /* "View.MemoryView":682
 *         if item is Ellipsis:
 *             if not seen_ellipsis:
 *                 result.extend([slice(None)] * (ndim - len(tup) + 1))             # <<<<<<<<<<<<<<
 *                 seen_ellipsis = True
 *             else:
 */
        __pyx_t_8 = PyObject_Length(__pyx_v_tup); if (unlikely(__pyx_t_8 == ((Py_ssize_t)-1))) __PYX_ERR(2, 682, __pyx_L1_error)
        __pyx_t_7 = PyList_New(1 * ((((__pyx_v_ndim - __pyx_t_8) + 1)<0) ? 0:((__pyx_v_ndim - __pyx_t_8) + 1))); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 682, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        { Py_ssize_t __pyx_temp;
          for (__pyx_temp=0; __pyx_temp < ((__pyx_v_ndim - __pyx_t_8) + 1); __pyx_temp++) {
            __Pyx_INCREF(__pyx_slice__41);
            __Pyx_GIVEREF(__pyx_slice__41);
            PyList_SET_ITEM(__pyx_t_7, __pyx_temp, __pyx_slice__41);
          }
        }
        __pyx_t_9 = __Pyx_PyList_Extend(__pyx_v_result, __pyx_t_7); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(2, 682, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

        /* "View.MemoryView":683
 *             if not seen_ellipsis:
 *                 result.extend([slice(None)] * (ndim - len(tup) + 1))
 *                 seen_ellipsis = True             # <<<<<<<<<<<<<<
 *             else:
 *                 result.append(slice(None))
 */
        __pyx_v_seen_ellipsis = 1;

        /* "View.MemoryView":681
 *     for idx, item in enumerate(tup):
 *         if item is Ellipsis:
 *             if not seen_ellipsis:             # <<<<<<<<<<<<<<
 *                 result.extend([slice(None)] * (ndim - len(tup) + 1))
 *                 seen_ellipsis = True
 */
        goto __pyx_L7;
      }

      /* "View.MemoryView":685
 *                 seen_ellipsis = True
 *             else:
 *                 result.append(slice(None))             # <<<<<<<<<<<<<<
 *             have_slices = True
 *         else:
 */
      /*else*/ {
        __pyx_t_9 = __Pyx_PyList_Append(__pyx_v_result, __pyx_slice__41); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(2, 685, __pyx_L1_error)
      }
      __pyx_L7:;

      /* "View.MemoryView":686
 *             else:
 *                 result.append(slice(None))
 *             have_slices = True             # <<<<<<<<<<<<<<
 *         else:
 *             if not isinstance(item, slice) and not PyIndex_Check(item):
 */
      __pyx_v_have_slices = 1;

      /* "View.MemoryView":680
 *     seen_ellipsis = False
 *     for idx, item in enumerate(tup):
 *         if item is Ellipsis:             # <<<<<<<<<<<<<<
 *             if not seen_ellipsis:
 *                 result.extend([slice(None)] * (ndim - len(tup) + 1))
 */
      goto __pyx_L6;
    }

    /* "View.MemoryView":688
 *             have_slices = True
 *         else:
 *             if not isinstance(item, slice) and not PyIndex_Check(item):             # <<<<<<<<<<<<<<
 *                 raise TypeError("Cannot index with type '%s'" % type(item))
 * 
 */
    /*else*/ {
      __pyx_t_2 = PySlice_Check(__pyx_v_item); 
      __pyx_t_10 = ((!(__pyx_t_2 != 0)) != 0);
      if (__pyx_t_10) {
      } else {
        __pyx_t_1 = __pyx_t_10;
        goto __pyx_L9_bool_binop_done;
      }
      __pyx_t_10 = ((!(PyIndex_Check(__pyx_v_item) != 0)) != 0);
      __pyx_t_1 = __pyx_t_10;
      __pyx_L9_bool_binop_done:;
      if (unlikely(__pyx_t_1)) {

        /* "View.MemoryView":689
 *         else:
 *             if not isinstance(item, slice) and not PyIndex_Check(item):
 *                 raise TypeError("Cannot index with type '%s'" % type(item))             # <<<<<<<<<<<<<<
 * 
 *             have_slices = have_slices or isinstance(item, slice)
 */
        __pyx_t_7 = __Pyx_PyString_FormatSafe(__pyx_kp_s_Cannot_index_with_type_s, ((PyObject *)Py_TYPE(__pyx_v_item))); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 689, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_11 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_7); if (unlikely(!__pyx_t_11)) __PYX_ERR(2, 689, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_11);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_Raise(__pyx_t_11, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        __PYX_ERR(2, 689, __pyx_L1_error)

        /* "View.MemoryView":688
 *             have_slices = True
 *         else:
 *             if not isinstance(item, slice) and not PyIndex_Check(item):             # <<<<<<<<<<<<<<
 *                 raise TypeError("Cannot index with type '%s'" % type(item))
 * 
 */
      }

      /* "View.MemoryView":691
 *                 raise TypeError("Cannot index with type '%s'" % type(item))
 * 
 *             have_slices = have_slices or isinstance(item, slice)             # <<<<<<<<<<<<<<
 *             result.append(item)
 * 
 */
      __pyx_t_10 = (__pyx_v_have_slices != 0);
      if (!__pyx_t_10) {
      } else {
        __pyx_t_1 = __pyx_t_10;
        goto __pyx_L11_bool_binop_done;
      }
      __pyx_t_10 = PySlice_Check(__pyx_v_item); 
      __pyx_t_2 = (__pyx_t_10 != 0);
      __pyx_t_1 = __pyx_t_2;
      __pyx_L11_bool_binop_done:;
      __pyx_v_have_slices = __pyx_t_1;

      /* "View.MemoryView":692
 * 
 *             have_slices = have_slices or isinstance(item, slice)
 *             result.append(item)             # <<<<<<<<<<<<<<
 * 
 *     nslices = ndim - len(result)
 */
      __pyx_t_9 = __Pyx_PyList_Append(__pyx_v_result, __pyx_v_item); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(2, 692, __pyx_L1_error)
    }
    __pyx_L6:;

    /* "View.MemoryView":679
 *     have_slices = False
 *     seen_ellipsis = False
 *     for idx, item in enumerate(tup):             # <<<<<<<<<<<<<<
 *         if item is Ellipsis:
 *             if not seen_ellipsis:
 */
  }
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "View.MemoryView":694
 *             result.append(item)
 * 
 *     nslices = ndim - len(result)             # <<<<<<<<<<<<<<
 *     if nslices:
 *         result.extend([slice(None)] * nslices)
 */
  __pyx_t_5 = PyList_GET_SIZE(__pyx_v_result); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(2, 694, __pyx_L1_error)
  __pyx_v_nslices = (__pyx_v_ndim - __pyx_t_5);

  /* "View.MemoryView":695
 * 
 *     nslices = ndim - len(result)
 *     if nslices:             # <<<<<<<<<<<<<<
 *         result.extend([slice(None)] * nslices)
 * 
 */
  __pyx_t_1 = (__pyx_v_nslices != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":696
 *     nslices = ndim - len(result)
 *     if nslices:
 *         result.extend([slice(None)] * nslices)             # <<<<<<<<<<<<<<
 * 
 *     return have_slices or nslices, tuple(result)
 */
    __pyx_t_3 = PyList_New(1 * ((__pyx_v_nslices<0) ? 0:__pyx_v_nslices)); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 696, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    { Py_ssize_t __pyx_temp;
      for (__pyx_temp=0; __pyx_temp < __pyx_v_nslices; __pyx_temp++) {
        __Pyx_INCREF(__pyx_slice__41);
        __Pyx_GIVEREF(__pyx_slice__41);
        PyList_SET_ITEM(__pyx_t_3, __pyx_temp, __pyx_slice__41);
      }
    }
    __pyx_t_9 = __Pyx_PyList_Extend(__pyx_v_result, __pyx_t_3); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(2, 696, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "View.MemoryView":695
 * 
 *     nslices = ndim - len(result)
 *     if nslices:             # <<<<<<<<<<<<<<
 *         result.extend([slice(None)] * nslices)
 * 
 */
  }

  /* "View.MemoryView":698
 *         result.extend([slice(None)] * nslices)
 * 
 *     return have_slices or nslices, tuple(result)             # <<<<<<<<<<<<<<
 * 
 * cdef assert_direct_dimensions(Py_ssize_t *suboffsets, int ndim):
 */
  __Pyx_XDECREF(__pyx_r);
  if (!__pyx_v_have_slices) {
  } else {
    __pyx_t_4 = __Pyx_PyBool_FromLong(__pyx_v_have_slices); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 698, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L14_bool_binop_done;
  }
  __pyx_t_4 = PyInt_FromSsize_t(__pyx_v_nslices); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 698, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = __pyx_t_4;
  __pyx_t_4 = 0;
  __pyx_L14_bool_binop_done:;
  __pyx_t_4 = PyList_AsTuple(__pyx_v_result); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 698, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_11 = PyTuple_New(2); if (unlikely(!__pyx_t_11)) __PYX_ERR(2, 698, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_11, 1, __pyx_t_4);
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_r = ((PyObject*)__pyx_t_11);
  __pyx_t_11 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":666
 *     return isinstance(o, memoryview)
 * 
 * cdef tuple _unellipsify(object index, int ndim):             # <<<<<<<<<<<<<<
 *     """
 *     Replace all ellipses with full slices and fill incomplete indices with
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("View.MemoryView._unellipsify", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_tup);
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XDECREF(__pyx_v_idx);
  __Pyx_XDECREF(__pyx_v_item);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":700
 *     return have_slices or nslices, tuple(result)
 * 
 * cdef assert_direct_dimensions(Py_ssize_t *suboffsets, int ndim):             # <<<<<<<<<<<<<<
 *     for suboffset in suboffsets[:ndim]:
 *         if suboffset >= 0:
 */

static PyObject *assert_direct_dimensions(Py_ssize_t *__pyx_v_suboffsets, int __pyx_v_ndim) {
  Py_ssize_t __pyx_v_suboffset;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  Py_ssize_t *__pyx_t_1;
  Py_ssize_t *__pyx_t_2;
  Py_ssize_t *__pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("assert_direct_dimensions", 0);

  /* "View.MemoryView":701
 * 
 * cdef assert_direct_dimensions(Py_ssize_t *suboffsets, int ndim):
 *     for suboffset in suboffsets[:ndim]:             # <<<<<<<<<<<<<<
 *         if suboffset >= 0:
 *             raise ValueError("Indirect dimensions not supported")
 */
  __pyx_t_2 = (__pyx_v_suboffsets + __pyx_v_ndim);
  for (__pyx_t_3 = __pyx_v_suboffsets; __pyx_t_3 < __pyx_t_2; __pyx_t_3++) {
    __pyx_t_1 = __pyx_t_3;
    __pyx_v_suboffset = (__pyx_t_1[0]);

    /* "View.MemoryView":702
 * cdef assert_direct_dimensions(Py_ssize_t *suboffsets, int ndim):
 *     for suboffset in suboffsets[:ndim]:
 *         if suboffset >= 0:             # <<<<<<<<<<<<<<
 *             raise ValueError("Indirect dimensions not supported")
 * 
 */
    __pyx_t_4 = ((__pyx_v_suboffset >= 0) != 0);
    if (unlikely(__pyx_t_4)) {

      /* "View.MemoryView":703
 *     for suboffset in suboffsets[:ndim]:
 *         if suboffset >= 0:
 *             raise ValueError("Indirect dimensions not supported")             # <<<<<<<<<<<<<<
 * 
 * 
 */
      __pyx_t_5 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__42, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 703, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_Raise(__pyx_t_5, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __PYX_ERR(2, 703, __pyx_L1_error)

      /* "View.MemoryView":702
 * cdef assert_direct_dimensions(Py_ssize_t *suboffsets, int ndim):
 *     for suboffset in suboffsets[:ndim]:
 *         if suboffset >= 0:             # <<<<<<<<<<<<<<
 *             raise ValueError("Indirect dimensions not supported")
 * 
 */
    }
  }

  /* "View.MemoryView":700
 *     return have_slices or nslices, tuple(result)
 * 
 * cdef assert_direct_dimensions(Py_ssize_t *suboffsets, int ndim):             # <<<<<<<<<<<<<<
 *     for suboffset in suboffsets[:ndim]:
 *         if suboffset >= 0:
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("View.MemoryView.assert_direct_dimensions", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":710
 * 
 * @cname('__pyx_memview_slice')
 * cdef memoryview memview_slice(memoryview memview, object indices):             # <<<<<<<<<<<<<<
 *     cdef int new_ndim = 0, suboffset_dim = -1, dim
 *     cdef bint negative_step
 */

static struct __pyx_memoryview_obj *__pyx_memview_slice(struct __pyx_memoryview_obj *__pyx_v_memview, PyObject *__pyx_v_indices) {
  int __pyx_v_new_ndim;
  int __pyx_v_suboffset_dim;
  int __pyx_v_dim;
  __Pyx_memviewslice __pyx_v_src;
  __Pyx_memviewslice __pyx_v_dst;
  __Pyx_memviewslice *__pyx_v_p_src;
  struct __pyx_memoryviewslice_obj *__pyx_v_memviewsliceobj = 0;
  __Pyx_memviewslice *__pyx_v_p_dst;
  int *__pyx_v_p_suboffset_dim;
  Py_ssize_t __pyx_v_start;
  Py_ssize_t __pyx_v_stop;
  Py_ssize_t __pyx_v_step;
  int __pyx_v_have_start;
  int __pyx_v_have_stop;
  int __pyx_v_have_step;
  PyObject *__pyx_v_index = NULL;
  struct __pyx_memoryview_obj *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  struct __pyx_memoryview_obj *__pyx_t_4;
  char *__pyx_t_5;
  int __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  PyObject *(*__pyx_t_8)(PyObject *);
  PyObject *__pyx_t_9 = NULL;
  Py_ssize_t __pyx_t_10;
  int __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  __Pyx_RefNannySetupContext("memview_slice", 0);

  /* "View.MemoryView":711
 * @cname('__pyx_memview_slice')
 * cdef memoryview memview_slice(memoryview memview, object indices):
 *     cdef int new_ndim = 0, suboffset_dim = -1, dim             # <<<<<<<<<<<<<<
 *     cdef bint negative_step
 *     cdef __Pyx_memviewslice src, dst
 */
  __pyx_v_new_ndim = 0;
  __pyx_v_suboffset_dim = -1;

  /* "View.MemoryView":718
 * 
 * 
 *     memset(&dst, 0, sizeof(dst))             # <<<<<<<<<<<<<<
 * 
 *     cdef _memoryviewslice memviewsliceobj
 */
  (void)(memset((&__pyx_v_dst), 0, (sizeof(__pyx_v_dst))));

  /* "View.MemoryView":722
 *     cdef _memoryviewslice memviewsliceobj
 * 
 *     assert memview.view.ndim > 0             # <<<<<<<<<<<<<<
 * 
 *     if isinstance(memview, _memoryviewslice):
 */
  #ifndef CYTHON_WITHOUT_ASSERTIONS
  if (unlikely(!Py_OptimizeFlag)) {
    if (unlikely(!((__pyx_v_memview->view.ndim > 0) != 0))) {
      PyErr_SetNone(PyExc_AssertionError);
      __PYX_ERR(2, 722, __pyx_L1_error)
    }
  }
  #endif

  /* "View.MemoryView":724
 *     assert memview.view.ndim > 0
 * 
 *     if isinstance(memview, _memoryviewslice):             # <<<<<<<<<<<<<<
 *         memviewsliceobj = memview
 *         p_src = &memviewsliceobj.from_slice
 */
  __pyx_t_1 = __Pyx_TypeCheck(((PyObject *)__pyx_v_memview), __pyx_memoryviewslice_type); 
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":725
 * 
 *     if isinstance(memview, _memoryviewslice):
 *         memviewsliceobj = memview             # <<<<<<<<<<<<<<
 *         p_src = &memviewsliceobj.from_slice
 *     else:
 */
    if (!(likely(((((PyObject *)__pyx_v_memview)) == Py_None) || likely(__Pyx_TypeTest(((PyObject *)__pyx_v_memview), __pyx_memoryviewslice_type))))) __PYX_ERR(2, 725, __pyx_L1_error)
    __pyx_t_3 = ((PyObject *)__pyx_v_memview);
    __Pyx_INCREF(__pyx_t_3);
    __pyx_v_memviewsliceobj = ((struct __pyx_memoryviewslice_obj *)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "View.MemoryView":726
 *     if isinstance(memview, _memoryviewslice):
 *         memviewsliceobj = memview
 *         p_src = &memviewsliceobj.from_slice             # <<<<<<<<<<<<<<
 *     else:
 *         slice_copy(memview, &src)
 */
    __pyx_v_p_src = (&__pyx_v_memviewsliceobj->from_slice);

    /* "View.MemoryView":724
 *     assert memview.view.ndim > 0
 * 
 *     if isinstance(memview, _memoryviewslice):             # <<<<<<<<<<<<<<
 *         memviewsliceobj = memview
 *         p_src = &memviewsliceobj.from_slice
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":728
 *         p_src = &memviewsliceobj.from_slice
 *     else:
 *         slice_copy(memview, &src)             # <<<<<<<<<<<<<<
 *         p_src = &src
 * 
 */
  /*else*/ {
    __pyx_memoryview_slice_copy(__pyx_v_memview, (&__pyx_v_src));

    /* "View.MemoryView":729
 *     else:
 *         slice_copy(memview, &src)
 *         p_src = &src             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_v_p_src = (&__pyx_v_src);
  }
  __pyx_L3:;

  /* "View.MemoryView":735
 * 
 * 
 *     dst.memview = p_src.memview             # <<<<<<<<<<<<<<
 *     dst.data = p_src.data
 * 
 */
  __pyx_t_4 = __pyx_v_p_src->memview;
  __pyx_v_dst.memview = __pyx_t_4;

  /* "View.MemoryView":736
 * 
 *     dst.memview = p_src.memview
 *     dst.data = p_src.data             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_5 = __pyx_v_p_src->data;
  __pyx_v_dst.data = __pyx_t_5;

  /* "View.MemoryView":741
 * 
 * 
 *     cdef __Pyx_memviewslice *p_dst = &dst             # <<<<<<<<<<<<<<
 *     cdef int *p_suboffset_dim = &suboffset_dim
 *     cdef Py_ssize_t start, stop, step
 */
  __pyx_v_p_dst = (&__pyx_v_dst);

  /* "View.MemoryView":742
 * 
 *     cdef __Pyx_memviewslice *p_dst = &dst
 *     cdef int *p_suboffset_dim = &suboffset_dim             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t start, stop, step
 *     cdef bint have_start, have_stop, have_step
 */
  __pyx_v_p_suboffset_dim = (&__pyx_v_suboffset_dim);

  /* "View.MemoryView":746
 *     cdef bint have_start, have_stop, have_step
 * 
 *     for dim, index in enumerate(indices):             # <<<<<<<<<<<<<<
 *         if PyIndex_Check(index):
 *             slice_memviewslice(
 */
  __pyx_t_6 = 0;
  if (likely(PyList_CheckExact(__pyx_v_indices)) || PyTuple_CheckExact(__pyx_v_indices)) {
    __pyx_t_3 = __pyx_v_indices; __Pyx_INCREF(__pyx_t_3); __pyx_t_7 = 0;
    __pyx_t_8 = NULL;
  } else {
    __pyx_t_7 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_v_indices); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 746, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_8 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 746, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_8)) {
      if (likely(PyList_CheckExact(__pyx_t_3))) {
        if (__pyx_t_7 >= PyList_GET_SIZE(__pyx_t_3)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_9 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_7); __Pyx_INCREF(__pyx_t_9); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(2, 746, __pyx_L1_error)
        #else
        __pyx_t_9 = PySequence_ITEM(__pyx_t_3, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_9)) __PYX_ERR(2, 746, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        #endif
      } else {
        if (__pyx_t_7 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_9 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_7); __Pyx_INCREF(__pyx_t_9); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(2, 746, __pyx_L1_error)
        #else
        __pyx_t_9 = PySequence_ITEM(__pyx_t_3, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_9)) __PYX_ERR(2, 746, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        #endif
      }
    } else {
      __pyx_t_9 = __pyx_t_8(__pyx_t_3);
      if (unlikely(!__pyx_t_9)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(2, 746, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_9);
    }
    __Pyx_XDECREF_SET(__pyx_v_index, __pyx_t_9);
    __pyx_t_9 = 0;
    __pyx_v_dim = __pyx_t_6;
    __pyx_t_6 = (__pyx_t_6 + 1);

    /* "View.MemoryView":747
 * 
 *     for dim, index in enumerate(indices):
 *         if PyIndex_Check(index):             # <<<<<<<<<<<<<<
 *             slice_memviewslice(
 *                 p_dst, p_src.shape[dim], p_src.strides[dim], p_src.suboffsets[dim],
 */
    __pyx_t_2 = (PyIndex_Check(__pyx_v_index) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":751
 *                 p_dst, p_src.shape[dim], p_src.strides[dim], p_src.suboffsets[dim],
 *                 dim, new_ndim, p_suboffset_dim,
 *                 index, 0, 0, # start, stop, step             # <<<<<<<<<<<<<<
 *                 0, 0, 0, # have_{start,stop,step}
 *                 False)
 */
      __pyx_t_10 = __Pyx_PyIndex_AsSsize_t(__pyx_v_index); if (unlikely((__pyx_t_10 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(2, 751, __pyx_L1_error)

      /* "View.MemoryView":748
 *     for dim, index in enumerate(indices):
 *         if PyIndex_Check(index):
 *             slice_memviewslice(             # <<<<<<<<<<<<<<
 *                 p_dst, p_src.shape[dim], p_src.strides[dim], p_src.suboffsets[dim],
 *                 dim, new_ndim, p_suboffset_dim,
 */
      __pyx_t_11 = __pyx_memoryview_slice_memviewslice(__pyx_v_p_dst, (__pyx_v_p_src->shape[__pyx_v_dim]), (__pyx_v_p_src->strides[__pyx_v_dim]), (__pyx_v_p_src->suboffsets[__pyx_v_dim]), __pyx_v_dim, __pyx_v_new_ndim, __pyx_v_p_suboffset_dim, __pyx_t_10, 0, 0, 0, 0, 0, 0); if (unlikely(__pyx_t_11 == ((int)-1))) __PYX_ERR(2, 748, __pyx_L1_error)

      /* "View.MemoryView":747
 * 
 *     for dim, index in enumerate(indices):
 *         if PyIndex_Check(index):             # <<<<<<<<<<<<<<
 *             slice_memviewslice(
 *                 p_dst, p_src.shape[dim], p_src.strides[dim], p_src.suboffsets[dim],
 */
      goto __pyx_L6;
    }

    /* "View.MemoryView":754
 *                 0, 0, 0, # have_{start,stop,step}
 *                 False)
 *         elif index is None:             # <<<<<<<<<<<<<<
 *             p_dst.shape[new_ndim] = 1
 *             p_dst.strides[new_ndim] = 0
 */
    __pyx_t_2 = (__pyx_v_index == Py_None);
    __pyx_t_1 = (__pyx_t_2 != 0);
    if (__pyx_t_1) {

      /* "View.MemoryView":755
 *                 False)
 *         elif index is None:
 *             p_dst.shape[new_ndim] = 1             # <<<<<<<<<<<<<<
 *             p_dst.strides[new_ndim] = 0
 *             p_dst.suboffsets[new_ndim] = -1
 */
      (__pyx_v_p_dst->shape[__pyx_v_new_ndim]) = 1;

      /* "View.MemoryView":756
 *         elif index is None:
 *             p_dst.shape[new_ndim] = 1
 *             p_dst.strides[new_ndim] = 0             # <<<<<<<<<<<<<<
 *             p_dst.suboffsets[new_ndim] = -1
 *             new_ndim += 1
 */
      (__pyx_v_p_dst->strides[__pyx_v_new_ndim]) = 0;

      /* "View.MemoryView":757
 *             p_dst.shape[new_ndim] = 1
 *             p_dst.strides[new_ndim] = 0
 *             p_dst.suboffsets[new_ndim] = -1             # <<<<<<<<<<<<<<
 *             new_ndim += 1
 *         else:
 */
      (__pyx_v_p_dst->suboffsets[__pyx_v_new_ndim]) = -1L;

      /* "View.MemoryView":758
 *             p_dst.strides[new_ndim] = 0
 *             p_dst.suboffsets[new_ndim] = -1
 *             new_ndim += 1             # <<<<<<<<<<<<<<
 *         else:
 *             start = index.start or 0
 */
      __pyx_v_new_ndim = (__pyx_v_new_ndim + 1);

      /* "View.MemoryView":754
 *                 0, 0, 0, # have_{start,stop,step}
 *                 False)
 *         elif index is None:             # <<<<<<<<<<<<<<
 *             p_dst.shape[new_ndim] = 1
 *             p_dst.strides[new_ndim] = 0
 */
      goto __pyx_L6;
    }

    /* "View.MemoryView":760
 *             new_ndim += 1
 *         else:
 *             start = index.start or 0             # <<<<<<<<<<<<<<
 *             stop = index.stop or 0
 *             step = index.step or 0
 */
    /*else*/ {
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_index, __pyx_n_s_start); if (unlikely(!__pyx_t_9)) __PYX_ERR(2, 760, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_9); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(2, 760, __pyx_L1_error)
      if (!__pyx_t_1) {
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      } else {
        __pyx_t_12 = __Pyx_PyIndex_AsSsize_t(__pyx_t_9); if (unlikely((__pyx_t_12 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(2, 760, __pyx_L1_error)
        __pyx_t_10 = __pyx_t_12;
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        goto __pyx_L7_bool_binop_done;
      }
      __pyx_t_10 = 0;
      __pyx_L7_bool_binop_done:;
      __pyx_v_start = __pyx_t_10;

      /* "View.MemoryView":761
 *         else:
 *             start = index.start or 0
 *             stop = index.stop or 0             # <<<<<<<<<<<<<<
 *             step = index.step or 0
 * 
 */
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_index, __pyx_n_s_stop); if (unlikely(!__pyx_t_9)) __PYX_ERR(2, 761, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_9); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(2, 761, __pyx_L1_error)
      if (!__pyx_t_1) {
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      } else {
        __pyx_t_12 = __Pyx_PyIndex_AsSsize_t(__pyx_t_9); if (unlikely((__pyx_t_12 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(2, 761, __pyx_L1_error)
        __pyx_t_10 = __pyx_t_12;
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        goto __pyx_L9_bool_binop_done;
      }
      __pyx_t_10 = 0;
      __pyx_L9_bool_binop_done:;
      __pyx_v_stop = __pyx_t_10;

      /* "View.MemoryView":762
 *             start = index.start or 0
 *             stop = index.stop or 0
 *             step = index.step or 0             # <<<<<<<<<<<<<<
 * 
 *             have_start = index.start is not None
 */
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_index, __pyx_n_s_step); if (unlikely(!__pyx_t_9)) __PYX_ERR(2, 762, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_9); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(2, 762, __pyx_L1_error)
      if (!__pyx_t_1) {
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      } else {
        __pyx_t_12 = __Pyx_PyIndex_AsSsize_t(__pyx_t_9); if (unlikely((__pyx_t_12 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(2, 762, __pyx_L1_error)
        __pyx_t_10 = __pyx_t_12;
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        goto __pyx_L11_bool_binop_done;
      }
      __pyx_t_10 = 0;
      __pyx_L11_bool_binop_done:;
      __pyx_v_step = __pyx_t_10;

      /* "View.MemoryView":764
 *             step = index.step or 0
 * 
 *             have_start = index.start is not None             # <<<<<<<<<<<<<<
 *             have_stop = index.stop is not None
 *             have_step = index.step is not None
 */
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_index, __pyx_n_s_start); if (unlikely(!__pyx_t_9)) __PYX_ERR(2, 764, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_1 = (__pyx_t_9 != Py_None);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_v_have_start = __pyx_t_1;

      /* "View.MemoryView":765
 * 
 *             have_start = index.start is not None
 *             have_stop = index.stop is not None             # <<<<<<<<<<<<<<
 *             have_step = index.step is not None
 * 
 */
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_index, __pyx_n_s_stop); if (unlikely(!__pyx_t_9)) __PYX_ERR(2, 765, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_1 = (__pyx_t_9 != Py_None);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_v_have_stop = __pyx_t_1;

      /* "View.MemoryView":766
 *             have_start = index.start is not None
 *             have_stop = index.stop is not None
 *             have_step = index.step is not None             # <<<<<<<<<<<<<<
 * 
 *             slice_memviewslice(
 */
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_index, __pyx_n_s_step); if (unlikely(!__pyx_t_9)) __PYX_ERR(2, 766, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_1 = (__pyx_t_9 != Py_None);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_v_have_step = __pyx_t_1;

      /* "View.MemoryView":768
 *             have_step = index.step is not None
 * 
 *             slice_memviewslice(             # <<<<<<<<<<<<<<
 *                 p_dst, p_src.shape[dim], p_src.strides[dim], p_src.suboffsets[dim],
 *                 dim, new_ndim, p_suboffset_dim,
 */
      __pyx_t_11 = __pyx_memoryview_slice_memviewslice(__pyx_v_p_dst, (__pyx_v_p_src->shape[__pyx_v_dim]), (__pyx_v_p_src->strides[__pyx_v_dim]), (__pyx_v_p_src->suboffsets[__pyx_v_dim]), __pyx_v_dim, __pyx_v_new_ndim, __pyx_v_p_suboffset_dim, __pyx_v_start, __pyx_v_stop, __pyx_v_step, __pyx_v_have_start, __pyx_v_have_stop, __pyx_v_have_step, 1); if (unlikely(__pyx_t_11 == ((int)-1))) __PYX_ERR(2, 768, __pyx_L1_error)

      /* "View.MemoryView":774
 *                 have_start, have_stop, have_step,
 *                 True)
 *             new_ndim += 1             # <<<<<<<<<<<<<<
 * 
 *     if isinstance(memview, _memoryviewslice):
 */
      __pyx_v_new_ndim = (__pyx_v_new_ndim + 1);
    }
    __pyx_L6:;

    /* "View.MemoryView":746
 *     cdef bint have_start, have_stop, have_step
 * 
 *     for dim, index in enumerate(indices):             # <<<<<<<<<<<<<<
 *         if PyIndex_Check(index):
 *             slice_memviewslice(
 */
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "View.MemoryView":776
 *             new_ndim += 1
 * 
 *     if isinstance(memview, _memoryviewslice):             # <<<<<<<<<<<<<<
 *         return memoryview_fromslice(dst, new_ndim,
 *                                     memviewsliceobj.to_object_func,
 */
  __pyx_t_1 = __Pyx_TypeCheck(((PyObject *)__pyx_v_memview), __pyx_memoryviewslice_type); 
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":777
 * 
 *     if isinstance(memview, _memoryviewslice):
 *         return memoryview_fromslice(dst, new_ndim,             # <<<<<<<<<<<<<<
 *                                     memviewsliceobj.to_object_func,
 *                                     memviewsliceobj.to_dtype_func,
 */
    __Pyx_XDECREF(((PyObject *)__pyx_r));

    /* "View.MemoryView":778
 *     if isinstance(memview, _memoryviewslice):
 *         return memoryview_fromslice(dst, new_ndim,
 *                                     memviewsliceobj.to_object_func,             # <<<<<<<<<<<<<<
 *                                     memviewsliceobj.to_dtype_func,
 *                                     memview.dtype_is_object)
 */
    if (unlikely(!__pyx_v_memviewsliceobj)) { __Pyx_RaiseUnboundLocalError("memviewsliceobj"); __PYX_ERR(2, 778, __pyx_L1_error) }

    /* "View.MemoryView":779
 *         return memoryview_fromslice(dst, new_ndim,
 *                                     memviewsliceobj.to_object_func,
 *                                     memviewsliceobj.to_dtype_func,             # <<<<<<<<<<<<<<
 *                                     memview.dtype_is_object)
 *     else:
 */
    if (unlikely(!__pyx_v_memviewsliceobj)) { __Pyx_RaiseUnboundLocalError("memviewsliceobj"); __PYX_ERR(2, 779, __pyx_L1_error) }

    /* "View.MemoryView":777
 * 
 *     if isinstance(memview, _memoryviewslice):
 *         return memoryview_fromslice(dst, new_ndim,             # <<<<<<<<<<<<<<
 *                                     memviewsliceobj.to_object_func,
 *                                     memviewsliceobj.to_dtype_func,
 */
    __pyx_t_3 = __pyx_memoryview_fromslice(__pyx_v_dst, __pyx_v_new_ndim, __pyx_v_memviewsliceobj->to_object_func, __pyx_v_memviewsliceobj->to_dtype_func, __pyx_v_memview->dtype_is_object); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 777, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_memoryview_type))))) __PYX_ERR(2, 777, __pyx_L1_error)
    __pyx_r = ((struct __pyx_memoryview_obj *)__pyx_t_3);
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "View.MemoryView":776
 *             new_ndim += 1
 * 
 *     if isinstance(memview, _memoryviewslice):             # <<<<<<<<<<<<<<
 *         return memoryview_fromslice(dst, new_ndim,
 *                                     memviewsliceobj.to_object_func,
 */
  }

  /* "View.MemoryView":782
 *                                     memview.dtype_is_object)
 *     else:
 *         return memoryview_fromslice(dst, new_ndim, NULL, NULL,             # <<<<<<<<<<<<<<
 *                                     memview.dtype_is_object)
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(((PyObject *)__pyx_r));

    /* "View.MemoryView":783
 *     else:
 *         return memoryview_fromslice(dst, new_ndim, NULL, NULL,
 *                                     memview.dtype_is_object)             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_t_3 = __pyx_memoryview_fromslice(__pyx_v_dst, __pyx_v_new_ndim, NULL, NULL, __pyx_v_memview->dtype_is_object); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 782, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);

    /* "View.MemoryView":782
 *                                     memview.dtype_is_object)
 *     else:
 *         return memoryview_fromslice(dst, new_ndim, NULL, NULL,             # <<<<<<<<<<<<<<
 *                                     memview.dtype_is_object)
 * 
 */
    if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_memoryview_type))))) __PYX_ERR(2, 782, __pyx_L1_error)
    __pyx_r = ((struct __pyx_memoryview_obj *)__pyx_t_3);
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "View.MemoryView":710
 * 
 * @cname('__pyx_memview_slice')
 * cdef memoryview memview_slice(memoryview memview, object indices):             # <<<<<<<<<<<<<<
 *     cdef int new_ndim = 0, suboffset_dim = -1, dim
 *     cdef bint negative_step
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("View.MemoryView.memview_slice", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_memviewsliceobj);
  __Pyx_XDECREF(__pyx_v_index);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":807
 * 
 * @cname('__pyx_memoryview_slice_memviewslice')
 * cdef int slice_memviewslice(             # <<<<<<<<<<<<<<
 *         __Pyx_memviewslice *dst,
 *         Py_ssize_t shape, Py_ssize_t stride, Py_ssize_t suboffset,
 */

static int __pyx_memoryview_slice_memviewslice(__Pyx_memviewslice *__pyx_v_dst, Py_ssize_t __pyx_v_shape, Py_ssize_t __pyx_v_stride, Py_ssize_t __pyx_v_suboffset, int __pyx_v_dim, int __pyx_v_new_ndim, int *__pyx_v_suboffset_dim, Py_ssize_t __pyx_v_start, Py_ssize_t __pyx_v_stop, Py_ssize_t __pyx_v_step, int __pyx_v_have_start, int __pyx_v_have_stop, int __pyx_v_have_step, int __pyx_v_is_slice) {
  Py_ssize_t __pyx_v_new_shape;
  int __pyx_v_negative_step;
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;

  /* "View.MemoryView":827
 *     cdef bint negative_step
 * 
 *     if not is_slice:             # <<<<<<<<<<<<<<
 * 
 *         if start < 0:
 */
  __pyx_t_1 = ((!(__pyx_v_is_slice != 0)) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":829
 *     if not is_slice:
 * 
 *         if start < 0:             # <<<<<<<<<<<<<<
 *             start += shape
 *         if not 0 <= start < shape:
 */
    __pyx_t_1 = ((__pyx_v_start < 0) != 0);
    if (__pyx_t_1) {

      /* "View.MemoryView":830
 * 
 *         if start < 0:
 *             start += shape             # <<<<<<<<<<<<<<
 *         if not 0 <= start < shape:
 *             _err_dim(IndexError, "Index out of bounds (axis %d)", dim)
 */
      __pyx_v_start = (__pyx_v_start + __pyx_v_shape);

      /* "View.MemoryView":829
 *     if not is_slice:
 * 
 *         if start < 0:             # <<<<<<<<<<<<<<
 *             start += shape
 *         if not 0 <= start < shape:
 */
    }

    /* "View.MemoryView":831
 *         if start < 0:
 *             start += shape
 *         if not 0 <= start < shape:             # <<<<<<<<<<<<<<
 *             _err_dim(IndexError, "Index out of bounds (axis %d)", dim)
 *     else:
 */
    __pyx_t_1 = (0 <= __pyx_v_start);
    if (__pyx_t_1) {
      __pyx_t_1 = (__pyx_v_start < __pyx_v_shape);
    }
    __pyx_t_2 = ((!(__pyx_t_1 != 0)) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":832
 *             start += shape
 *         if not 0 <= start < shape:
 *             _err_dim(IndexError, "Index out of bounds (axis %d)", dim)             # <<<<<<<<<<<<<<
 *     else:
 * 
 */
      __pyx_t_3 = __pyx_memoryview_err_dim(__pyx_builtin_IndexError, ((char *)"Index out of bounds (axis %d)"), __pyx_v_dim); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(2, 832, __pyx_L1_error)

      /* "View.MemoryView":831
 *         if start < 0:
 *             start += shape
 *         if not 0 <= start < shape:             # <<<<<<<<<<<<<<
 *             _err_dim(IndexError, "Index out of bounds (axis %d)", dim)
 *     else:
 */
    }

    /* "View.MemoryView":827
 *     cdef bint negative_step
 * 
 *     if not is_slice:             # <<<<<<<<<<<<<<
 * 
 *         if start < 0:
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":835
 *     else:
 * 
 *         negative_step = have_step != 0 and step < 0             # <<<<<<<<<<<<<<
 * 
 *         if have_step and step == 0:
 */
  /*else*/ {
    __pyx_t_1 = ((__pyx_v_have_step != 0) != 0);
    if (__pyx_t_1) {
    } else {
      __pyx_t_2 = __pyx_t_1;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_1 = ((__pyx_v_step < 0) != 0);
    __pyx_t_2 = __pyx_t_1;
    __pyx_L6_bool_binop_done:;
    __pyx_v_negative_step = __pyx_t_2;

    /* "View.MemoryView":837
 *         negative_step = have_step != 0 and step < 0
 * 
 *         if have_step and step == 0:             # <<<<<<<<<<<<<<
 *             _err_dim(ValueError, "Step may not be zero (axis %d)", dim)
 * 
 */
    __pyx_t_1 = (__pyx_v_have_step != 0);
    if (__pyx_t_1) {
    } else {
      __pyx_t_2 = __pyx_t_1;
      goto __pyx_L9_bool_binop_done;
    }
    __pyx_t_1 = ((__pyx_v_step == 0) != 0);
    __pyx_t_2 = __pyx_t_1;
    __pyx_L9_bool_binop_done:;
    if (__pyx_t_2) {

      /* "View.MemoryView":838
 * 
 *         if have_step and step == 0:
 *             _err_dim(ValueError, "Step may not be zero (axis %d)", dim)             # <<<<<<<<<<<<<<
 * 
 * 
 */
      __pyx_t_3 = __pyx_memoryview_err_dim(__pyx_builtin_ValueError, ((char *)"Step may not be zero (axis %d)"), __pyx_v_dim); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(2, 838, __pyx_L1_error)

      /* "View.MemoryView":837
 *         negative_step = have_step != 0 and step < 0
 * 
 *         if have_step and step == 0:             # <<<<<<<<<<<<<<
 *             _err_dim(ValueError, "Step may not be zero (axis %d)", dim)
 * 
 */
    }

    /* "View.MemoryView":841
 * 
 * 
 *         if have_start:             # <<<<<<<<<<<<<<
 *             if start < 0:
 *                 start += shape
 */
    __pyx_t_2 = (__pyx_v_have_start != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":842
 * 
 *         if have_start:
 *             if start < 0:             # <<<<<<<<<<<<<<
 *                 start += shape
 *                 if start < 0:
 */
      __pyx_t_2 = ((__pyx_v_start < 0) != 0);
      if (__pyx_t_2) {

        /* "View.MemoryView":843
 *         if have_start:
 *             if start < 0:
 *                 start += shape             # <<<<<<<<<<<<<<
 *                 if start < 0:
 *                     start = 0
 */
        __pyx_v_start = (__pyx_v_start + __pyx_v_shape);

        /* "View.MemoryView":844
 *             if start < 0:
 *                 start += shape
 *                 if start < 0:             # <<<<<<<<<<<<<<
 *                     start = 0
 *             elif start >= shape:
 */
        __pyx_t_2 = ((__pyx_v_start < 0) != 0);
        if (__pyx_t_2) {

          /* "View.MemoryView":845
 *                 start += shape
 *                 if start < 0:
 *                     start = 0             # <<<<<<<<<<<<<<
 *             elif start >= shape:
 *                 if negative_step:
 */
          __pyx_v_start = 0;

          /* "View.MemoryView":844
 *             if start < 0:
 *                 start += shape
 *                 if start < 0:             # <<<<<<<<<<<<<<
 *                     start = 0
 *             elif start >= shape:
 */
        }

        /* "View.MemoryView":842
 * 
 *         if have_start:
 *             if start < 0:             # <<<<<<<<<<<<<<
 *                 start += shape
 *                 if start < 0:
 */
        goto __pyx_L12;
      }

      /* "View.MemoryView":846
 *                 if start < 0:
 *                     start = 0
 *             elif start >= shape:             # <<<<<<<<<<<<<<
 *                 if negative_step:
 *                     start = shape - 1
 */
      __pyx_t_2 = ((__pyx_v_start >= __pyx_v_shape) != 0);
      if (__pyx_t_2) {

        /* "View.MemoryView":847
 *                     start = 0
 *             elif start >= shape:
 *                 if negative_step:             # <<<<<<<<<<<<<<
 *                     start = shape - 1
 *                 else:
 */
        __pyx_t_2 = (__pyx_v_negative_step != 0);
        if (__pyx_t_2) {

          /* "View.MemoryView":848
 *             elif start >= shape:
 *                 if negative_step:
 *                     start = shape - 1             # <<<<<<<<<<<<<<
 *                 else:
 *                     start = shape
 */
          __pyx_v_start = (__pyx_v_shape - 1);

          /* "View.MemoryView":847
 *                     start = 0
 *             elif start >= shape:
 *                 if negative_step:             # <<<<<<<<<<<<<<
 *                     start = shape - 1
 *                 else:
 */
          goto __pyx_L14;
        }

        /* "View.MemoryView":850
 *                     start = shape - 1
 *                 else:
 *                     start = shape             # <<<<<<<<<<<<<<
 *         else:
 *             if negative_step:
 */
        /*else*/ {
          __pyx_v_start = __pyx_v_shape;
        }
        __pyx_L14:;

        /* "View.MemoryView":846
 *                 if start < 0:
 *                     start = 0
 *             elif start >= shape:             # <<<<<<<<<<<<<<
 *                 if negative_step:
 *                     start = shape - 1
 */
      }
      __pyx_L12:;

      /* "View.MemoryView":841
 * 
 * 
 *         if have_start:             # <<<<<<<<<<<<<<
 *             if start < 0:
 *                 start += shape
 */
      goto __pyx_L11;
    }

    /* "View.MemoryView":852
 *                     start = shape
 *         else:
 *             if negative_step:             # <<<<<<<<<<<<<<
 *                 start = shape - 1
 *             else:
 */
    /*else*/ {
      __pyx_t_2 = (__pyx_v_negative_step != 0);
      if (__pyx_t_2) {

        /* "View.MemoryView":853
 *         else:
 *             if negative_step:
 *                 start = shape - 1             # <<<<<<<<<<<<<<
 *             else:
 *                 start = 0
 */
        __pyx_v_start = (__pyx_v_shape - 1);

        /* "View.MemoryView":852
 *                     start = shape
 *         else:
 *             if negative_step:             # <<<<<<<<<<<<<<
 *                 start = shape - 1
 *             else:
 */
        goto __pyx_L15;
      }

      /* "View.MemoryView":855
 *                 start = shape - 1
 *             else:
 *                 start = 0             # <<<<<<<<<<<<<<
 * 
 *         if have_stop:
 */
      /*else*/ {
        __pyx_v_start = 0;
      }
      __pyx_L15:;
    }
    __pyx_L11:;

    /* "View.MemoryView":857
 *                 start = 0
 * 
 *         if have_stop:             # <<<<<<<<<<<<<<
 *             if stop < 0:
 *                 stop += shape
 */
    __pyx_t_2 = (__pyx_v_have_stop != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":858
 * 
 *         if have_stop:
 *             if stop < 0:             # <<<<<<<<<<<<<<
 *                 stop += shape
 *                 if stop < 0:
 */
      __pyx_t_2 = ((__pyx_v_stop < 0) != 0);
      if (__pyx_t_2) {

        /* "View.MemoryView":859
 *         if have_stop:
 *             if stop < 0:
 *                 stop += shape             # <<<<<<<<<<<<<<
 *                 if stop < 0:
 *                     stop = 0
 */
        __pyx_v_stop = (__pyx_v_stop + __pyx_v_shape);

        /* "View.MemoryView":860
 *             if stop < 0:
 *                 stop += shape
 *                 if stop < 0:             # <<<<<<<<<<<<<<
 *                     stop = 0
 *             elif stop > shape:
 */
        __pyx_t_2 = ((__pyx_v_stop < 0) != 0);
        if (__pyx_t_2) {

          /* "View.MemoryView":861
 *                 stop += shape
 *                 if stop < 0:
 *                     stop = 0             # <<<<<<<<<<<<<<
 *             elif stop > shape:
 *                 stop = shape
 */
          __pyx_v_stop = 0;

          /* "View.MemoryView":860
 *             if stop < 0:
 *                 stop += shape
 *                 if stop < 0:             # <<<<<<<<<<<<<<
 *                     stop = 0
 *             elif stop > shape:
 */
        }

        /* "View.MemoryView":858
 * 
 *         if have_stop:
 *             if stop < 0:             # <<<<<<<<<<<<<<
 *                 stop += shape
 *                 if stop < 0:
 */
        goto __pyx_L17;
      }

      /* "View.MemoryView":862
 *                 if stop < 0:
 *                     stop = 0
 *             elif stop > shape:             # <<<<<<<<<<<<<<
 *                 stop = shape
 *         else:
 */
      __pyx_t_2 = ((__pyx_v_stop > __pyx_v_shape) != 0);
      if (__pyx_t_2) {

        /* "View.MemoryView":863
 *                     stop = 0
 *             elif stop > shape:
 *                 stop = shape             # <<<<<<<<<<<<<<
 *         else:
 *             if negative_step:
 */
        __pyx_v_stop = __pyx_v_shape;

        /* "View.MemoryView":862
 *                 if stop < 0:
 *                     stop = 0
 *             elif stop > shape:             # <<<<<<<<<<<<<<
 *                 stop = shape
 *         else:
 */
      }
      __pyx_L17:;

      /* "View.MemoryView":857
 *                 start = 0
 * 
 *         if have_stop:             # <<<<<<<<<<<<<<
 *             if stop < 0:
 *                 stop += shape
 */
      goto __pyx_L16;
    }

    /* "View.MemoryView":865
 *                 stop = shape
 *         else:
 *             if negative_step:             # <<<<<<<<<<<<<<
 *                 stop = -1
 *             else:
 */
    /*else*/ {
      __pyx_t_2 = (__pyx_v_negative_step != 0);
      if (__pyx_t_2) {

        /* "View.MemoryView":866
 *         else:
 *             if negative_step:
 *                 stop = -1             # <<<<<<<<<<<<<<
 *             else:
 *                 stop = shape
 */
        __pyx_v_stop = -1L;

        /* "View.MemoryView":865
 *                 stop = shape
 *         else:
 *             if negative_step:             # <<<<<<<<<<<<<<
 *                 stop = -1
 *             else:
 */
        goto __pyx_L19;
      }

      /* "View.MemoryView":868
 *                 stop = -1
 *             else:
 *                 stop = shape             # <<<<<<<<<<<<<<
 * 
 *         if not have_step:
 */
      /*else*/ {
        __pyx_v_stop = __pyx_v_shape;
      }
      __pyx_L19:;
    }
    __pyx_L16:;

    /* "View.MemoryView":870
 *                 stop = shape
 * 
 *         if not have_step:             # <<<<<<<<<<<<<<
 *             step = 1
 * 
 */
    __pyx_t_2 = ((!(__pyx_v_have_step != 0)) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":871
 * 
 *         if not have_step:
 *             step = 1             # <<<<<<<<<<<<<<
 * 
 * 
 */
      __pyx_v_step = 1;

      /* "View.MemoryView":870
 *                 stop = shape
 * 
 *         if not have_step:             # <<<<<<<<<<<<<<
 *             step = 1
 * 
 */
    }

    /* "View.MemoryView":875
 * 
 *         with cython.cdivision(True):
 *             new_shape = (stop - start) // step             # <<<<<<<<<<<<<<
 * 
 *             if (stop - start) - step * new_shape:
 */
    __pyx_v_new_shape = ((__pyx_v_stop - __pyx_v_start) / __pyx_v_step);

    /* "View.MemoryView":877
 *             new_shape = (stop - start) // step
 * 
 *             if (stop - start) - step * new_shape:             # <<<<<<<<<<<<<<
 *                 new_shape += 1
 * 
 */
    __pyx_t_2 = (((__pyx_v_stop - __pyx_v_start) - (__pyx_v_step * __pyx_v_new_shape)) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":878
 * 
 *             if (stop - start) - step * new_shape:
 *                 new_shape += 1             # <<<<<<<<<<<<<<
 * 
 *         if new_shape < 0:
 */
      __pyx_v_new_shape = (__pyx_v_new_shape + 1);

      /* "View.MemoryView":877
 *             new_shape = (stop - start) // step
 * 
 *             if (stop - start) - step * new_shape:             # <<<<<<<<<<<<<<
 *                 new_shape += 1
 * 
 */
    }

    /* "View.MemoryView":880
 *                 new_shape += 1
 * 
 *         if new_shape < 0:             # <<<<<<<<<<<<<<
 *             new_shape = 0
 * 
 */
    __pyx_t_2 = ((__pyx_v_new_shape < 0) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":881
 * 
 *         if new_shape < 0:
 *             new_shape = 0             # <<<<<<<<<<<<<<
 * 
 * 
 */
      __pyx_v_new_shape = 0;

      /* "View.MemoryView":880
 *                 new_shape += 1
 * 
 *         if new_shape < 0:             # <<<<<<<<<<<<<<
 *             new_shape = 0
 * 
 */
    }

    /* "View.MemoryView":884
 * 
 * 
 *         dst.strides[new_ndim] = stride * step             # <<<<<<<<<<<<<<
 *         dst.shape[new_ndim] = new_shape
 *         dst.suboffsets[new_ndim] = suboffset
 */
    (__pyx_v_dst->strides[__pyx_v_new_ndim]) = (__pyx_v_stride * __pyx_v_step);

    /* "View.MemoryView":885
 * 
 *         dst.strides[new_ndim] = stride * step
 *         dst.shape[new_ndim] = new_shape             # <<<<<<<<<<<<<<
 *         dst.suboffsets[new_ndim] = suboffset
 * 
 */
    (__pyx_v_dst->shape[__pyx_v_new_ndim]) = __pyx_v_new_shape;

    /* "View.MemoryView":886
 *         dst.strides[new_ndim] = stride * step
 *         dst.shape[new_ndim] = new_shape
 *         dst.suboffsets[new_ndim] = suboffset             # <<<<<<<<<<<<<<
 * 
 * 
 */
    (__pyx_v_dst->suboffsets[__pyx_v_new_ndim]) = __pyx_v_suboffset;
  }
  __pyx_L3:;

  /* "View.MemoryView":889
 * 
 * 
 *     if suboffset_dim[0] < 0:             # <<<<<<<<<<<<<<
 *         dst.data += start * stride
 *     else:
 */
  __pyx_t_2 = (((__pyx_v_suboffset_dim[0]) < 0) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":890
 * 
 *     if suboffset_dim[0] < 0:
 *         dst.data += start * stride             # <<<<<<<<<<<<<<
 *     else:
 *         dst.suboffsets[suboffset_dim[0]] += start * stride
 */
    __pyx_v_dst->data = (__pyx_v_dst->data + (__pyx_v_start * __pyx_v_stride));

    /* "View.MemoryView":889
 * 
 * 
 *     if suboffset_dim[0] < 0:             # <<<<<<<<<<<<<<
 *         dst.data += start * stride
 *     else:
 */
    goto __pyx_L23;
  }

  /* "View.MemoryView":892
 *         dst.data += start * stride
 *     else:
 *         dst.suboffsets[suboffset_dim[0]] += start * stride             # <<<<<<<<<<<<<<
 * 
 *     if suboffset >= 0:
 */
  /*else*/ {
    __pyx_t_3 = (__pyx_v_suboffset_dim[0]);
    (__pyx_v_dst->suboffsets[__pyx_t_3]) = ((__pyx_v_dst->suboffsets[__pyx_t_3]) + (__pyx_v_start * __pyx_v_stride));
  }
  __pyx_L23:;

  /* "View.MemoryView":894
 *         dst.suboffsets[suboffset_dim[0]] += start * stride
 * 
 *     if suboffset >= 0:             # <<<<<<<<<<<<<<
 *         if not is_slice:
 *             if new_ndim == 0:
 */
  __pyx_t_2 = ((__pyx_v_suboffset >= 0) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":895
 * 
 *     if suboffset >= 0:
 *         if not is_slice:             # <<<<<<<<<<<<<<
 *             if new_ndim == 0:
 *                 dst.data = (<char **> dst.data)[0] + suboffset
 */
    __pyx_t_2 = ((!(__pyx_v_is_slice != 0)) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":896
 *     if suboffset >= 0:
 *         if not is_slice:
 *             if new_ndim == 0:             # <<<<<<<<<<<<<<
 *                 dst.data = (<char **> dst.data)[0] + suboffset
 *             else:
 */
      __pyx_t_2 = ((__pyx_v_new_ndim == 0) != 0);
      if (__pyx_t_2) {

        /* "View.MemoryView":897
 *         if not is_slice:
 *             if new_ndim == 0:
 *                 dst.data = (<char **> dst.data)[0] + suboffset             # <<<<<<<<<<<<<<
 *             else:
 *                 _err_dim(IndexError, "All dimensions preceding dimension %d "
 */
        __pyx_v_dst->data = ((((char **)__pyx_v_dst->data)[0]) + __pyx_v_suboffset);

        /* "View.MemoryView":896
 *     if suboffset >= 0:
 *         if not is_slice:
 *             if new_ndim == 0:             # <<<<<<<<<<<<<<
 *                 dst.data = (<char **> dst.data)[0] + suboffset
 *             else:
 */
        goto __pyx_L26;
      }

      /* "View.MemoryView":899
 *                 dst.data = (<char **> dst.data)[0] + suboffset
 *             else:
 *                 _err_dim(IndexError, "All dimensions preceding dimension %d "             # <<<<<<<<<<<<<<
 *                                      "must be indexed and not sliced", dim)
 *         else:
 */
      /*else*/ {

        /* "View.MemoryView":900
 *             else:
 *                 _err_dim(IndexError, "All dimensions preceding dimension %d "
 *                                      "must be indexed and not sliced", dim)             # <<<<<<<<<<<<<<
 *         else:
 *             suboffset_dim[0] = new_ndim
 */
        __pyx_t_3 = __pyx_memoryview_err_dim(__pyx_builtin_IndexError, ((char *)"All dimensions preceding dimension %d must be indexed and not sliced"), __pyx_v_dim); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(2, 899, __pyx_L1_error)
      }
      __pyx_L26:;

      /* "View.MemoryView":895
 * 
 *     if suboffset >= 0:
 *         if not is_slice:             # <<<<<<<<<<<<<<
 *             if new_ndim == 0:
 *                 dst.data = (<char **> dst.data)[0] + suboffset
 */
      goto __pyx_L25;
    }

    /* "View.MemoryView":902
 *                                      "must be indexed and not sliced", dim)
 *         else:
 *             suboffset_dim[0] = new_ndim             # <<<<<<<<<<<<<<
 * 
 *     return 0
 */
    /*else*/ {
      (__pyx_v_suboffset_dim[0]) = __pyx_v_new_ndim;
    }
    __pyx_L25:;

    /* "View.MemoryView":894
 *         dst.suboffsets[suboffset_dim[0]] += start * stride
 * 
 *     if suboffset >= 0:             # <<<<<<<<<<<<<<
 *         if not is_slice:
 *             if new_ndim == 0:
 */
  }

  /* "View.MemoryView":904
 *             suboffset_dim[0] = new_ndim
 * 
 *     return 0             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "View.MemoryView":807
 * 
 * @cname('__pyx_memoryview_slice_memviewslice')
 * cdef int slice_memviewslice(             # <<<<<<<<<<<<<<
 *         __Pyx_memviewslice *dst,
 *         Py_ssize_t shape, Py_ssize_t stride, Py_ssize_t suboffset,
 */

  /* function exit code */
  __pyx_L1_error:;
  {
    #ifdef WITH_THREAD
    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
    #endif
    __Pyx_AddTraceback("View.MemoryView.slice_memviewslice", __pyx_clineno, __pyx_lineno, __pyx_filename);
    #ifdef WITH_THREAD
    __Pyx_PyGILState_Release(__pyx_gilstate_save);
    #endif
  }
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "View.MemoryView":910
 * 
 * @cname('__pyx_pybuffer_index')
 * cdef char *pybuffer_index(Py_buffer *view, char *bufp, Py_ssize_t index,             # <<<<<<<<<<<<<<
 *                           Py_ssize_t dim) except NULL:
 *     cdef Py_ssize_t shape, stride, suboffset = -1
 */

static char *__pyx_pybuffer_index(Py_buffer *__pyx_v_view, char *__pyx_v_bufp, Py_ssize_t __pyx_v_index, Py_ssize_t __pyx_v_dim) {
  Py_ssize_t __pyx_v_shape;
  Py_ssize_t __pyx_v_stride;
  Py_ssize_t __pyx_v_suboffset;
  Py_ssize_t __pyx_v_itemsize;
  char *__pyx_v_resultp;
  char *__pyx_r;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("pybuffer_index", 0);

  /* "View.MemoryView":912
 * cdef char *pybuffer_index(Py_buffer *view, char *bufp, Py_ssize_t index,
 *                           Py_ssize_t dim) except NULL:
 *     cdef Py_ssize_t shape, stride, suboffset = -1             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t itemsize = view.itemsize
 *     cdef char *resultp
 */
  __pyx_v_suboffset = -1L;

  /* "View.MemoryView":913
 *                           Py_ssize_t dim) except NULL:
 *     cdef Py_ssize_t shape, stride, suboffset = -1
 *     cdef Py_ssize_t itemsize = view.itemsize             # <<<<<<<<<<<<<<
 *     cdef char *resultp
 * 
 */
  __pyx_t_1 = __pyx_v_view->itemsize;
  __pyx_v_itemsize = __pyx_t_1;

  /* "View.MemoryView":916
 *     cdef char *resultp
 * 
 *     if view.ndim == 0:             # <<<<<<<<<<<<<<
 *         shape = view.len / itemsize
 *         stride = itemsize
 */
  __pyx_t_2 = ((__pyx_v_view->ndim == 0) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":917
 * 
 *     if view.ndim == 0:
 *         shape = view.len / itemsize             # <<<<<<<<<<<<<<
 *         stride = itemsize
 *     else:
 */
    if (unlikely(__pyx_v_itemsize == 0)) {
      PyErr_SetString(PyExc_ZeroDivisionError, "integer division or modulo by zero");
      __PYX_ERR(2, 917, __pyx_L1_error)
    }
    else if (sizeof(Py_ssize_t) == sizeof(long) && (!(((Py_ssize_t)-1) > 0)) && unlikely(__pyx_v_itemsize == (Py_ssize_t)-1)  && unlikely(UNARY_NEG_WOULD_OVERFLOW(__pyx_v_view->len))) {
      PyErr_SetString(PyExc_OverflowError, "value too large to perform division");
      __PYX_ERR(2, 917, __pyx_L1_error)
    }
    __pyx_v_shape = __Pyx_div_Py_ssize_t(__pyx_v_view->len, __pyx_v_itemsize);

    /* "View.MemoryView":918
 *     if view.ndim == 0:
 *         shape = view.len / itemsize
 *         stride = itemsize             # <<<<<<<<<<<<<<
 *     else:
 *         shape = view.shape[dim]
 */
    __pyx_v_stride = __pyx_v_itemsize;

    /* "View.MemoryView":916
 *     cdef char *resultp
 * 
 *     if view.ndim == 0:             # <<<<<<<<<<<<<<
 *         shape = view.len / itemsize
 *         stride = itemsize
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":920
 *         stride = itemsize
 *     else:
 *         shape = view.shape[dim]             # <<<<<<<<<<<<<<
 *         stride = view.strides[dim]
 *         if view.suboffsets != NULL:
 */
  /*else*/ {
    __pyx_v_shape = (__pyx_v_view->shape[__pyx_v_dim]);

    /* "View.MemoryView":921
 *     else:
 *         shape = view.shape[dim]
 *         stride = view.strides[dim]             # <<<<<<<<<<<<<<
 *         if view.suboffsets != NULL:
 *             suboffset = view.suboffsets[dim]
 */
    __pyx_v_stride = (__pyx_v_view->strides[__pyx_v_dim]);

    /* "View.MemoryView":922
 *         shape = view.shape[dim]
 *         stride = view.strides[dim]
 *         if view.suboffsets != NULL:             # <<<<<<<<<<<<<<
 *             suboffset = view.suboffsets[dim]
 * 
 */
    __pyx_t_2 = ((__pyx_v_view->suboffsets != NULL) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":923
 *         stride = view.strides[dim]
 *         if view.suboffsets != NULL:
 *             suboffset = view.suboffsets[dim]             # <<<<<<<<<<<<<<
 * 
 *     if index < 0:
 */
      __pyx_v_suboffset = (__pyx_v_view->suboffsets[__pyx_v_dim]);

      /* "View.MemoryView":922
 *         shape = view.shape[dim]
 *         stride = view.strides[dim]
 *         if view.suboffsets != NULL:             # <<<<<<<<<<<<<<
 *             suboffset = view.suboffsets[dim]
 * 
 */
    }
  }
  __pyx_L3:;

  /* "View.MemoryView":925
 *             suboffset = view.suboffsets[dim]
 * 
 *     if index < 0:             # <<<<<<<<<<<<<<
 *         index += view.shape[dim]
 *         if index < 0:
 */
  __pyx_t_2 = ((__pyx_v_index < 0) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":926
 * 
 *     if index < 0:
 *         index += view.shape[dim]             # <<<<<<<<<<<<<<
 *         if index < 0:
 *             raise IndexError("Out of bounds on buffer access (axis %d)" % dim)
 */
    __pyx_v_index = (__pyx_v_index + (__pyx_v_view->shape[__pyx_v_dim]));

    /* "View.MemoryView":927
 *     if index < 0:
 *         index += view.shape[dim]
 *         if index < 0:             # <<<<<<<<<<<<<<
 *             raise IndexError("Out of bounds on buffer access (axis %d)" % dim)
 * 
 */
    __pyx_t_2 = ((__pyx_v_index < 0) != 0);
    if (unlikely(__pyx_t_2)) {

      /* "View.MemoryView":928
 *         index += view.shape[dim]
 *         if index < 0:
 *             raise IndexError("Out of bounds on buffer access (axis %d)" % dim)             # <<<<<<<<<<<<<<
 * 
 *     if index >= shape:
 */
      __pyx_t_3 = PyInt_FromSsize_t(__pyx_v_dim); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 928, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_Out_of_bounds_on_buffer_access_a, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 928, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_IndexError, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 928, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(2, 928, __pyx_L1_error)

      /* "View.MemoryView":927
 *     if index < 0:
 *         index += view.shape[dim]
 *         if index < 0:             # <<<<<<<<<<<<<<
 *             raise IndexError("Out of bounds on buffer access (axis %d)" % dim)
 * 
 */
    }

    /* "View.MemoryView":925
 *             suboffset = view.suboffsets[dim]
 * 
 *     if index < 0:             # <<<<<<<<<<<<<<
 *         index += view.shape[dim]
 *         if index < 0:
 */
  }

  /* "View.MemoryView":930
 *             raise IndexError("Out of bounds on buffer access (axis %d)" % dim)
 * 
 *     if index >= shape:             # <<<<<<<<<<<<<<
 *         raise IndexError("Out of bounds on buffer access (axis %d)" % dim)
 * 
 */
  __pyx_t_2 = ((__pyx_v_index >= __pyx_v_shape) != 0);
  if (unlikely(__pyx_t_2)) {

    /* "View.MemoryView":931
 * 
 *     if index >= shape:
 *         raise IndexError("Out of bounds on buffer access (axis %d)" % dim)             # <<<<<<<<<<<<<<
 * 
 *     resultp = bufp + index * stride
 */
    __pyx_t_3 = PyInt_FromSsize_t(__pyx_v_dim); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 931, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_Out_of_bounds_on_buffer_access_a, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 931, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_IndexError, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 931, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(2, 931, __pyx_L1_error)

    /* "View.MemoryView":930
 *             raise IndexError("Out of bounds on buffer access (axis %d)" % dim)
 * 
 *     if index >= shape:             # <<<<<<<<<<<<<<
 *         raise IndexError("Out of bounds on buffer access (axis %d)" % dim)
 * 
 */
  }

  /* "View.MemoryView":933
 *         raise IndexError("Out of bounds on buffer access (axis %d)" % dim)
 * 
 *     resultp = bufp + index * stride             # <<<<<<<<<<<<<<
 *     if suboffset >= 0:
 *         resultp = (<char **> resultp)[0] + suboffset
 */
  __pyx_v_resultp = (__pyx_v_bufp + (__pyx_v_index * __pyx_v_stride));

  /* "View.MemoryView":934
 * 
 *     resultp = bufp + index * stride
 *     if suboffset >= 0:             # <<<<<<<<<<<<<<
 *         resultp = (<char **> resultp)[0] + suboffset
 * 
 */
  __pyx_t_2 = ((__pyx_v_suboffset >= 0) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":935
 *     resultp = bufp + index * stride
 *     if suboffset >= 0:
 *         resultp = (<char **> resultp)[0] + suboffset             # <<<<<<<<<<<<<<
 * 
 *     return resultp
 */
    __pyx_v_resultp = ((((char **)__pyx_v_resultp)[0]) + __pyx_v_suboffset);

    /* "View.MemoryView":934
 * 
 *     resultp = bufp + index * stride
 *     if suboffset >= 0:             # <<<<<<<<<<<<<<
 *         resultp = (<char **> resultp)[0] + suboffset
 * 
 */
  }

  /* "View.MemoryView":937
 *         resultp = (<char **> resultp)[0] + suboffset
 * 
 *     return resultp             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_resultp;
  goto __pyx_L0;

  /* "View.MemoryView":910
 * 
 * @cname('__pyx_pybuffer_index')
 * cdef char *pybuffer_index(Py_buffer *view, char *bufp, Py_ssize_t index,             # <<<<<<<<<<<<<<
 *                           Py_ssize_t dim) except NULL:
 *     cdef Py_ssize_t shape, stride, suboffset = -1
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("View.MemoryView.pybuffer_index", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":943
 * 
 * @cname('__pyx_memslice_transpose')
 * cdef int transpose_memslice(__Pyx_memviewslice *memslice) nogil except 0:             # <<<<<<<<<<<<<<
 *     cdef int ndim = memslice.memview.view.ndim
 * 
 */

static int __pyx_memslice_transpose(__Pyx_memviewslice *__pyx_v_memslice) {
  int __pyx_v_ndim;
  Py_ssize_t *__pyx_v_shape;
  Py_ssize_t *__pyx_v_strides;
  int __pyx_v_i;
  int __pyx_v_j;
  int __pyx_r;
  int __pyx_t_1;
  Py_ssize_t *__pyx_t_2;
  long __pyx_t_3;
  long __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;

  /* "View.MemoryView":944
 * @cname('__pyx_memslice_transpose')
 * cdef int transpose_memslice(__Pyx_memviewslice *memslice) nogil except 0:
 *     cdef int ndim = memslice.memview.view.ndim             # <<<<<<<<<<<<<<
 * 
 *     cdef Py_ssize_t *shape = memslice.shape
 */
  __pyx_t_1 = __pyx_v_memslice->memview->view.ndim;
  __pyx_v_ndim = __pyx_t_1;

  /* "View.MemoryView":946
 *     cdef int ndim = memslice.memview.view.ndim
 * 
 *     cdef Py_ssize_t *shape = memslice.shape             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t *strides = memslice.strides
 * 
 */
  __pyx_t_2 = __pyx_v_memslice->shape;
  __pyx_v_shape = __pyx_t_2;

  /* "View.MemoryView":947
 * 
 *     cdef Py_ssize_t *shape = memslice.shape
 *     cdef Py_ssize_t *strides = memslice.strides             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = __pyx_v_memslice->strides;
  __pyx_v_strides = __pyx_t_2;

  /* "View.MemoryView":951
 * 
 *     cdef int i, j
 *     for i in range(ndim / 2):             # <<<<<<<<<<<<<<
 *         j = ndim - 1 - i
 *         strides[i], strides[j] = strides[j], strides[i]
 */
  __pyx_t_3 = __Pyx_div_long(__pyx_v_ndim, 2);
  __pyx_t_4 = __pyx_t_3;
  for (__pyx_t_1 = 0; __pyx_t_1 < __pyx_t_4; __pyx_t_1+=1) {
    __pyx_v_i = __pyx_t_1;

    /* "View.MemoryView":952
 *     cdef int i, j
 *     for i in range(ndim / 2):
 *         j = ndim - 1 - i             # <<<<<<<<<<<<<<
 *         strides[i], strides[j] = strides[j], strides[i]
 *         shape[i], shape[j] = shape[j], shape[i]
 */
    __pyx_v_j = ((__pyx_v_ndim - 1) - __pyx_v_i);

    /* "View.MemoryView":953
 *     for i in range(ndim / 2):
 *         j = ndim - 1 - i
 *         strides[i], strides[j] = strides[j], strides[i]             # <<<<<<<<<<<<<<
 *         shape[i], shape[j] = shape[j], shape[i]
 * 
 */
    __pyx_t_5 = (__pyx_v_strides[__pyx_v_j]);
    __pyx_t_6 = (__pyx_v_strides[__pyx_v_i]);
    (__pyx_v_strides[__pyx_v_i]) = __pyx_t_5;
    (__pyx_v_strides[__pyx_v_j]) = __pyx_t_6;

    /* "View.MemoryView":954
 *         j = ndim - 1 - i
 *         strides[i], strides[j] = strides[j], strides[i]
 *         shape[i], shape[j] = shape[j], shape[i]             # <<<<<<<<<<<<<<
 * 
 *         if memslice.suboffsets[i] >= 0 or memslice.suboffsets[j] >= 0:
 */
    __pyx_t_6 = (__pyx_v_shape[__pyx_v_j]);
    __pyx_t_5 = (__pyx_v_shape[__pyx_v_i]);
    (__pyx_v_shape[__pyx_v_i]) = __pyx_t_6;
    (__pyx_v_shape[__pyx_v_j]) = __pyx_t_5;

    /* "View.MemoryView":956
 *         shape[i], shape[j] = shape[j], shape[i]
 * 
 *         if memslice.suboffsets[i] >= 0 or memslice.suboffsets[j] >= 0:             # <<<<<<<<<<<<<<
 *             _err(ValueError, "Cannot transpose memoryview with indirect dimensions")
 * 
 */
    __pyx_t_8 = (((__pyx_v_memslice->suboffsets[__pyx_v_i]) >= 0) != 0);
    if (!__pyx_t_8) {
    } else {
      __pyx_t_7 = __pyx_t_8;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_8 = (((__pyx_v_memslice->suboffsets[__pyx_v_j]) >= 0) != 0);
    __pyx_t_7 = __pyx_t_8;
    __pyx_L6_bool_binop_done:;
    if (__pyx_t_7) {

      /* "View.MemoryView":957
 * 
 *         if memslice.suboffsets[i] >= 0 or memslice.suboffsets[j] >= 0:
 *             _err(ValueError, "Cannot transpose memoryview with indirect dimensions")             # <<<<<<<<<<<<<<
 * 
 *     return 1
 */
      __pyx_t_9 = __pyx_memoryview_err(__pyx_builtin_ValueError, ((char *)"Cannot transpose memoryview with indirect dimensions")); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(2, 957, __pyx_L1_error)

      /* "View.MemoryView":956
 *         shape[i], shape[j] = shape[j], shape[i]
 * 
 *         if memslice.suboffsets[i] >= 0 or memslice.suboffsets[j] >= 0:             # <<<<<<<<<<<<<<
 *             _err(ValueError, "Cannot transpose memoryview with indirect dimensions")
 * 
 */
    }
  }

  /* "View.MemoryView":959
 *             _err(ValueError, "Cannot transpose memoryview with indirect dimensions")
 * 
 *     return 1             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 1;
  goto __pyx_L0;

  /* "View.MemoryView":943
 * 
 * @cname('__pyx_memslice_transpose')
 * cdef int transpose_memslice(__Pyx_memviewslice *memslice) nogil except 0:             # <<<<<<<<<<<<<<
 *     cdef int ndim = memslice.memview.view.ndim
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  {
    #ifdef WITH_THREAD
    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
    #endif
    __Pyx_AddTraceback("View.MemoryView.transpose_memslice", __pyx_clineno, __pyx_lineno, __pyx_filename);
    #ifdef WITH_THREAD
    __Pyx_PyGILState_Release(__pyx_gilstate_save);
    #endif
  }
  __pyx_r = 0;
  __pyx_L0:;
  return __pyx_r;
}

/* "View.MemoryView":976
 *     cdef int (*to_dtype_func)(char *, object) except 0
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         __PYX_XDEC_MEMVIEW(&self.from_slice, 1)
 * 
 */

/* Python wrapper */
static void __pyx_memoryviewslice___dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_memoryviewslice___dealloc__(PyObject *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_memoryviewslice___pyx_pf_15View_dot_MemoryView_16_memoryviewslice___dealloc__(((struct __pyx_memoryviewslice_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_memoryviewslice___pyx_pf_15View_dot_MemoryView_16_memoryviewslice___dealloc__(struct __pyx_memoryviewslice_obj *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "View.MemoryView":977
 * 
 *     def __dealloc__(self):
 *         __PYX_XDEC_MEMVIEW(&self.from_slice, 1)             # <<<<<<<<<<<<<<
 * 
 *     cdef convert_item_to_object(self, char *itemp):
 */
  __PYX_XDEC_MEMVIEW((&__pyx_v_self->from_slice), 1);

  /* "View.MemoryView":976
 *     cdef int (*to_dtype_func)(char *, object) except 0
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         __PYX_XDEC_MEMVIEW(&self.from_slice, 1)
 * 
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "View.MemoryView":979
 *         __PYX_XDEC_MEMVIEW(&self.from_slice, 1)
 * 
 *     cdef convert_item_to_object(self, char *itemp):             # <<<<<<<<<<<<<<
 *         if self.to_object_func != NULL:
 *             return self.to_object_func(itemp)
 */

static PyObject *__pyx_memoryviewslice_convert_item_to_object(struct __pyx_memoryviewslice_obj *__pyx_v_self, char *__pyx_v_itemp) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("convert_item_to_object", 0);

  /* "View.MemoryView":980
 * 
 *     cdef convert_item_to_object(self, char *itemp):
 *         if self.to_object_func != NULL:             # <<<<<<<<<<<<<<
 *             return self.to_object_func(itemp)
 *         else:
 */
  __pyx_t_1 = ((__pyx_v_self->to_object_func != NULL) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":981
 *     cdef convert_item_to_object(self, char *itemp):
 *         if self.to_object_func != NULL:
 *             return self.to_object_func(itemp)             # <<<<<<<<<<<<<<
 *         else:
 *             return memoryview.convert_item_to_object(self, itemp)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __pyx_v_self->to_object_func(__pyx_v_itemp); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 981, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "View.MemoryView":980
 * 
 *     cdef convert_item_to_object(self, char *itemp):
 *         if self.to_object_func != NULL:             # <<<<<<<<<<<<<<
 *             return self.to_object_func(itemp)
 *         else:
 */
  }

  /* "View.MemoryView":983
 *             return self.to_object_func(itemp)
 *         else:
 *             return memoryview.convert_item_to_object(self, itemp)             # <<<<<<<<<<<<<<
 * 
 *     cdef assign_item_from_object(self, char *itemp, object value):
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __pyx_memoryview_convert_item_to_object(((struct __pyx_memoryview_obj *)__pyx_v_self), __pyx_v_itemp); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 983, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;
  }

  /* "View.MemoryView":979
 *         __PYX_XDEC_MEMVIEW(&self.from_slice, 1)
 * 
 *     cdef convert_item_to_object(self, char *itemp):             # <<<<<<<<<<<<<<
 *         if self.to_object_func != NULL:
 *             return self.to_object_func(itemp)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("View.MemoryView._memoryviewslice.convert_item_to_object", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":985
 *             return memoryview.convert_item_to_object(self, itemp)
 * 
 *     cdef assign_item_from_object(self, char *itemp, object value):             # <<<<<<<<<<<<<<
 *         if self.to_dtype_func != NULL:
 *             self.to_dtype_func(itemp, value)
 */

static PyObject *__pyx_memoryviewslice_assign_item_from_object(struct __pyx_memoryviewslice_obj *__pyx_v_self, char *__pyx_v_itemp, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("assign_item_from_object", 0);

  /* "View.MemoryView":986
 * 
 *     cdef assign_item_from_object(self, char *itemp, object value):
 *         if self.to_dtype_func != NULL:             # <<<<<<<<<<<<<<
 *             self.to_dtype_func(itemp, value)
 *         else:
 */
  __pyx_t_1 = ((__pyx_v_self->to_dtype_func != NULL) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":987
 *     cdef assign_item_from_object(self, char *itemp, object value):
 *         if self.to_dtype_func != NULL:
 *             self.to_dtype_func(itemp, value)             # <<<<<<<<<<<<<<
 *         else:
 *             memoryview.assign_item_from_object(self, itemp, value)
 */
    __pyx_t_2 = __pyx_v_self->to_dtype_func(__pyx_v_itemp, __pyx_v_value); if (unlikely(__pyx_t_2 == ((int)0))) __PYX_ERR(2, 987, __pyx_L1_error)

    /* "View.MemoryView":986
 * 
 *     cdef assign_item_from_object(self, char *itemp, object value):
 *         if self.to_dtype_func != NULL:             # <<<<<<<<<<<<<<
 *             self.to_dtype_func(itemp, value)
 *         else:
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":989
 *             self.to_dtype_func(itemp, value)
 *         else:
 *             memoryview.assign_item_from_object(self, itemp, value)             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  /*else*/ {
    __pyx_t_3 = __pyx_memoryview_assign_item_from_object(((struct __pyx_memoryview_obj *)__pyx_v_self), __pyx_v_itemp, __pyx_v_value); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 989, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_L3:;

  /* "View.MemoryView":985
 *             return memoryview.convert_item_to_object(self, itemp)
 * 
 *     cdef assign_item_from_object(self, char *itemp, object value):             # <<<<<<<<<<<<<<
 *         if self.to_dtype_func != NULL:
 *             self.to_dtype_func(itemp, value)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView._memoryviewslice.assign_item_from_object", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":992
 * 
 *     @property
 *     def base(self):             # <<<<<<<<<<<<<<
 *         return self.from_object
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_16_memoryviewslice_4base_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_16_memoryviewslice_4base_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_16_memoryviewslice_4base___get__(((struct __pyx_memoryviewslice_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_16_memoryviewslice_4base___get__(struct __pyx_memoryviewslice_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":993
 *     @property
 *     def base(self):
 *         return self.from_object             # <<<<<<<<<<<<<<
 * 
 *     __pyx_getbuffer = capsule(<void *> &__pyx_memoryview_getbuffer, "getbuffer(obj, view, flags)")
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->from_object);
  __pyx_r = __pyx_v_self->from_object;
  goto __pyx_L0;

  /* "View.MemoryView":992
 * 
 *     @property
 *     def base(self):             # <<<<<<<<<<<<<<
 *         return self.from_object
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw___pyx_memoryviewslice_1__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw___pyx_memoryviewslice_1__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf___pyx_memoryviewslice___reduce_cython__(((struct __pyx_memoryviewslice_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf___pyx_memoryviewslice___reduce_cython__(CYTHON_UNUSED struct __pyx_memoryviewslice_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__43, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(2, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView._memoryviewslice.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw___pyx_memoryviewslice_3__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw___pyx_memoryviewslice_3__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf___pyx_memoryviewslice_2__setstate_cython__(((struct __pyx_memoryviewslice_obj *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf___pyx_memoryviewslice_2__setstate_cython__(CYTHON_UNUSED struct __pyx_memoryviewslice_obj *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__44, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(2, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView._memoryviewslice.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":999
 * 
 * @cname('__pyx_memoryview_fromslice')
 * cdef memoryview_fromslice(__Pyx_memviewslice memviewslice,             # <<<<<<<<<<<<<<
 *                           int ndim,
 *                           object (*to_object_func)(char *),
 */

static PyObject *__pyx_memoryview_fromslice(__Pyx_memviewslice __pyx_v_memviewslice, int __pyx_v_ndim, PyObject *(*__pyx_v_to_object_func)(char *), int (*__pyx_v_to_dtype_func)(char *, PyObject *), int __pyx_v_dtype_is_object) {
  struct __pyx_memoryviewslice_obj *__pyx_v_result = 0;
  Py_ssize_t __pyx_v_suboffset;
  PyObject *__pyx_v_length = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_TypeInfo *__pyx_t_4;
  Py_buffer __pyx_t_5;
  Py_ssize_t *__pyx_t_6;
  Py_ssize_t *__pyx_t_7;
  Py_ssize_t *__pyx_t_8;
  Py_ssize_t __pyx_t_9;
  __Pyx_RefNannySetupContext("memoryview_fromslice", 0);

  /* "View.MemoryView":1007
 *     cdef _memoryviewslice result
 * 
 *     if <PyObject *> memviewslice.memview == Py_None:             # <<<<<<<<<<<<<<
 *         return None
 * 
 */
  __pyx_t_1 = ((((PyObject *)__pyx_v_memviewslice.memview) == Py_None) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":1008
 * 
 *     if <PyObject *> memviewslice.memview == Py_None:
 *         return None             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "View.MemoryView":1007
 *     cdef _memoryviewslice result
 * 
 *     if <PyObject *> memviewslice.memview == Py_None:             # <<<<<<<<<<<<<<
 *         return None
 * 
 */
  }

  /* "View.MemoryView":1013
 * 
 * 
 *     result = _memoryviewslice(None, 0, dtype_is_object)             # <<<<<<<<<<<<<<
 * 
 *     result.from_slice = memviewslice
 */
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_v_dtype_is_object); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 1013, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 1013, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  PyTuple_SET_ITEM(__pyx_t_3, 0, Py_None);
  __Pyx_INCREF(__pyx_int_0);
  __Pyx_GIVEREF(__pyx_int_0);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_int_0);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_memoryviewslice_type), __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 1013, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_result = ((struct __pyx_memoryviewslice_obj *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "View.MemoryView":1015
 *     result = _memoryviewslice(None, 0, dtype_is_object)
 * 
 *     result.from_slice = memviewslice             # <<<<<<<<<<<<<<
 *     __PYX_INC_MEMVIEW(&memviewslice, 1)
 * 
 */
  __pyx_v_result->from_slice = __pyx_v_memviewslice;

  /* "View.MemoryView":1016
 * 
 *     result.from_slice = memviewslice
 *     __PYX_INC_MEMVIEW(&memviewslice, 1)             # <<<<<<<<<<<<<<
 * 
 *     result.from_object = (<memoryview> memviewslice.memview).base
 */
  __PYX_INC_MEMVIEW((&__pyx_v_memviewslice), 1);

  /* "View.MemoryView":1018
 *     __PYX_INC_MEMVIEW(&memviewslice, 1)
 * 
 *     result.from_object = (<memoryview> memviewslice.memview).base             # <<<<<<<<<<<<<<
 *     result.typeinfo = memviewslice.memview.typeinfo
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_memviewslice.memview), __pyx_n_s_base); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 1018, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF(__pyx_v_result->from_object);
  __Pyx_DECREF(__pyx_v_result->from_object);
  __pyx_v_result->from_object = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "View.MemoryView":1019
 * 
 *     result.from_object = (<memoryview> memviewslice.memview).base
 *     result.typeinfo = memviewslice.memview.typeinfo             # <<<<<<<<<<<<<<
 * 
 *     result.view = memviewslice.memview.view
 */
  __pyx_t_4 = __pyx_v_memviewslice.memview->typeinfo;
  __pyx_v_result->__pyx_base.typeinfo = __pyx_t_4;

  /* "View.MemoryView":1021
 *     result.typeinfo = memviewslice.memview.typeinfo
 * 
 *     result.view = memviewslice.memview.view             # <<<<<<<<<<<<<<
 *     result.view.buf = <void *> memviewslice.data
 *     result.view.ndim = ndim
 */
  __pyx_t_5 = __pyx_v_memviewslice.memview->view;
  __pyx_v_result->__pyx_base.view = __pyx_t_5;

  /* "View.MemoryView":1022
 * 
 *     result.view = memviewslice.memview.view
 *     result.view.buf = <void *> memviewslice.data             # <<<<<<<<<<<<<<
 *     result.view.ndim = ndim
 *     (<__pyx_buffer *> &result.view).obj = Py_None
 */
  __pyx_v_result->__pyx_base.view.buf = ((void *)__pyx_v_memviewslice.data);

  /* "View.MemoryView":1023
 *     result.view = memviewslice.memview.view
 *     result.view.buf = <void *> memviewslice.data
 *     result.view.ndim = ndim             # <<<<<<<<<<<<<<
 *     (<__pyx_buffer *> &result.view).obj = Py_None
 *     Py_INCREF(Py_None)
 */
  __pyx_v_result->__pyx_base.view.ndim = __pyx_v_ndim;

  /* "View.MemoryView":1024
 *     result.view.buf = <void *> memviewslice.data
 *     result.view.ndim = ndim
 *     (<__pyx_buffer *> &result.view).obj = Py_None             # <<<<<<<<<<<<<<
 *     Py_INCREF(Py_None)
 * 
 */
  ((Py_buffer *)(&__pyx_v_result->__pyx_base.view))->obj = Py_None;

  /* "View.MemoryView":1025
 *     result.view.ndim = ndim
 *     (<__pyx_buffer *> &result.view).obj = Py_None
 *     Py_INCREF(Py_None)             # <<<<<<<<<<<<<<
 * 
 *     if (<memoryview>memviewslice.memview).flags & PyBUF_WRITABLE:
 */
  Py_INCREF(Py_None);

  /* "View.MemoryView":1027
 *     Py_INCREF(Py_None)
 * 
 *     if (<memoryview>memviewslice.memview).flags & PyBUF_WRITABLE:             # <<<<<<<<<<<<<<
 *         result.flags = PyBUF_RECORDS
 *     else:
 */
  __pyx_t_1 = ((((struct __pyx_memoryview_obj *)__pyx_v_memviewslice.memview)->flags & PyBUF_WRITABLE) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":1028
 * 
 *     if (<memoryview>memviewslice.memview).flags & PyBUF_WRITABLE:
 *         result.flags = PyBUF_RECORDS             # <<<<<<<<<<<<<<
 *     else:
 *         result.flags = PyBUF_RECORDS_RO
 */
    __pyx_v_result->__pyx_base.flags = PyBUF_RECORDS;

    /* "View.MemoryView":1027
 *     Py_INCREF(Py_None)
 * 
 *     if (<memoryview>memviewslice.memview).flags & PyBUF_WRITABLE:             # <<<<<<<<<<<<<<
 *         result.flags = PyBUF_RECORDS
 *     else:
 */
    goto __pyx_L4;
  }

  /* "View.MemoryView":1030
 *         result.flags = PyBUF_RECORDS
 *     else:
 *         result.flags = PyBUF_RECORDS_RO             # <<<<<<<<<<<<<<
 * 
 *     result.view.shape = <Py_ssize_t *> result.from_slice.shape
 */
  /*else*/ {
    __pyx_v_result->__pyx_base.flags = PyBUF_RECORDS_RO;
  }
  __pyx_L4:;

  /* "View.MemoryView":1032
 *         result.flags = PyBUF_RECORDS_RO
 * 
 *     result.view.shape = <Py_ssize_t *> result.from_slice.shape             # <<<<<<<<<<<<<<
 *     result.view.strides = <Py_ssize_t *> result.from_slice.strides
 * 
 */
  __pyx_v_result->__pyx_base.view.shape = ((Py_ssize_t *)__pyx_v_result->from_slice.shape);

  /* "View.MemoryView":1033
 * 
 *     result.view.shape = <Py_ssize_t *> result.from_slice.shape
 *     result.view.strides = <Py_ssize_t *> result.from_slice.strides             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_v_result->__pyx_base.view.strides = ((Py_ssize_t *)__pyx_v_result->from_slice.strides);

  /* "View.MemoryView":1036
 * 
 * 
 *     result.view.suboffsets = NULL             # <<<<<<<<<<<<<<
 *     for suboffset in result.from_slice.suboffsets[:ndim]:
 *         if suboffset >= 0:
 */
  __pyx_v_result->__pyx_base.view.suboffsets = NULL;

  /* "View.MemoryView":1037
 * 
 *     result.view.suboffsets = NULL
 *     for suboffset in result.from_slice.suboffsets[:ndim]:             # <<<<<<<<<<<<<<
 *         if suboffset >= 0:
 *             result.view.suboffsets = <Py_ssize_t *> result.from_slice.suboffsets
 */
  __pyx_t_7 = (__pyx_v_result->from_slice.suboffsets + __pyx_v_ndim);
  for (__pyx_t_8 = __pyx_v_result->from_slice.suboffsets; __pyx_t_8 < __pyx_t_7; __pyx_t_8++) {
    __pyx_t_6 = __pyx_t_8;
    __pyx_v_suboffset = (__pyx_t_6[0]);

    /* "View.MemoryView":1038
 *     result.view.suboffsets = NULL
 *     for suboffset in result.from_slice.suboffsets[:ndim]:
 *         if suboffset >= 0:             # <<<<<<<<<<<<<<
 *             result.view.suboffsets = <Py_ssize_t *> result.from_slice.suboffsets
 *             break
 */
    __pyx_t_1 = ((__pyx_v_suboffset >= 0) != 0);
    if (__pyx_t_1) {

      /* "View.MemoryView":1039
 *     for suboffset in result.from_slice.suboffsets[:ndim]:
 *         if suboffset >= 0:
 *             result.view.suboffsets = <Py_ssize_t *> result.from_slice.suboffsets             # <<<<<<<<<<<<<<
 *             break
 * 
 */
      __pyx_v_result->__pyx_base.view.suboffsets = ((Py_ssize_t *)__pyx_v_result->from_slice.suboffsets);

      /* "View.MemoryView":1040
 *         if suboffset >= 0:
 *             result.view.suboffsets = <Py_ssize_t *> result.from_slice.suboffsets
 *             break             # <<<<<<<<<<<<<<
 * 
 *     result.view.len = result.view.itemsize
 */
      goto __pyx_L6_break;

      /* "View.MemoryView":1038
 *     result.view.suboffsets = NULL
 *     for suboffset in result.from_slice.suboffsets[:ndim]:
 *         if suboffset >= 0:             # <<<<<<<<<<<<<<
 *             result.view.suboffsets = <Py_ssize_t *> result.from_slice.suboffsets
 *             break
 */
    }
  }
  __pyx_L6_break:;

  /* "View.MemoryView":1042
 *             break
 * 
 *     result.view.len = result.view.itemsize             # <<<<<<<<<<<<<<
 *     for length in result.view.shape[:ndim]:
 *         result.view.len *= length
 */
  __pyx_t_9 = __pyx_v_result->__pyx_base.view.itemsize;
  __pyx_v_result->__pyx_base.view.len = __pyx_t_9;

  /* "View.MemoryView":1043
 * 
 *     result.view.len = result.view.itemsize
 *     for length in result.view.shape[:ndim]:             # <<<<<<<<<<<<<<
 *         result.view.len *= length
 * 
 */
  __pyx_t_7 = (__pyx_v_result->__pyx_base.view.shape + __pyx_v_ndim);
  for (__pyx_t_8 = __pyx_v_result->__pyx_base.view.shape; __pyx_t_8 < __pyx_t_7; __pyx_t_8++) {
    __pyx_t_6 = __pyx_t_8;
    __pyx_t_2 = PyInt_FromSsize_t((__pyx_t_6[0])); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 1043, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_XDECREF_SET(__pyx_v_length, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "View.MemoryView":1044
 *     result.view.len = result.view.itemsize
 *     for length in result.view.shape[:ndim]:
 *         result.view.len *= length             # <<<<<<<<<<<<<<
 * 
 *     result.to_object_func = to_object_func
 */
    __pyx_t_2 = PyInt_FromSsize_t(__pyx_v_result->__pyx_base.view.len); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 1044, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PyNumber_InPlaceMultiply(__pyx_t_2, __pyx_v_length); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 1044, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_9 = __Pyx_PyIndex_AsSsize_t(__pyx_t_3); if (unlikely((__pyx_t_9 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(2, 1044, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_result->__pyx_base.view.len = __pyx_t_9;
  }

  /* "View.MemoryView":1046
 *         result.view.len *= length
 * 
 *     result.to_object_func = to_object_func             # <<<<<<<<<<<<<<
 *     result.to_dtype_func = to_dtype_func
 * 
 */
  __pyx_v_result->to_object_func = __pyx_v_to_object_func;

  /* "View.MemoryView":1047
 * 
 *     result.to_object_func = to_object_func
 *     result.to_dtype_func = to_dtype_func             # <<<<<<<<<<<<<<
 * 
 *     return result
 */
  __pyx_v_result->to_dtype_func = __pyx_v_to_dtype_func;

  /* "View.MemoryView":1049
 *     result.to_dtype_func = to_dtype_func
 * 
 *     return result             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_get_slice_from_memoryview')
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_result));
  __pyx_r = ((PyObject *)__pyx_v_result);
  goto __pyx_L0;

  /* "View.MemoryView":999
 * 
 * @cname('__pyx_memoryview_fromslice')
 * cdef memoryview_fromslice(__Pyx_memviewslice memviewslice,             # <<<<<<<<<<<<<<
 *                           int ndim,
 *                           object (*to_object_func)(char *),
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView.memoryview_fromslice", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_result);
  __Pyx_XDECREF(__pyx_v_length);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":1052
 * 
 * @cname('__pyx_memoryview_get_slice_from_memoryview')
 * cdef __Pyx_memviewslice *get_slice_from_memview(memoryview memview,             # <<<<<<<<<<<<<<
 *                                                    __Pyx_memviewslice *mslice):
 *     cdef _memoryviewslice obj
 */

static __Pyx_memviewslice *__pyx_memoryview_get_slice_from_memoryview(struct __pyx_memoryview_obj *__pyx_v_memview, __Pyx_memviewslice *__pyx_v_mslice) {
  struct __pyx_memoryviewslice_obj *__pyx_v_obj = 0;
  __Pyx_memviewslice *__pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("get_slice_from_memview", 0);

  /* "View.MemoryView":1055
 *                                                    __Pyx_memviewslice *mslice):
 *     cdef _memoryviewslice obj
 *     if isinstance(memview, _memoryviewslice):             # <<<<<<<<<<<<<<
 *         obj = memview
 *         return &obj.from_slice
 */
  __pyx_t_1 = __Pyx_TypeCheck(((PyObject *)__pyx_v_memview), __pyx_memoryviewslice_type); 
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":1056
 *     cdef _memoryviewslice obj
 *     if isinstance(memview, _memoryviewslice):
 *         obj = memview             # <<<<<<<<<<<<<<
 *         return &obj.from_slice
 *     else:
 */
    if (!(likely(((((PyObject *)__pyx_v_memview)) == Py_None) || likely(__Pyx_TypeTest(((PyObject *)__pyx_v_memview), __pyx_memoryviewslice_type))))) __PYX_ERR(2, 1056, __pyx_L1_error)
    __pyx_t_3 = ((PyObject *)__pyx_v_memview);
    __Pyx_INCREF(__pyx_t_3);
    __pyx_v_obj = ((struct __pyx_memoryviewslice_obj *)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "View.MemoryView":1057
 *     if isinstance(memview, _memoryviewslice):
 *         obj = memview
 *         return &obj.from_slice             # <<<<<<<<<<<<<<
 *     else:
 *         slice_copy(memview, mslice)
 */
    __pyx_r = (&__pyx_v_obj->from_slice);
    goto __pyx_L0;

    /* "View.MemoryView":1055
 *                                                    __Pyx_memviewslice *mslice):
 *     cdef _memoryviewslice obj
 *     if isinstance(memview, _memoryviewslice):             # <<<<<<<<<<<<<<
 *         obj = memview
 *         return &obj.from_slice
 */
  }

  /* "View.MemoryView":1059
 *         return &obj.from_slice
 *     else:
 *         slice_copy(memview, mslice)             # <<<<<<<<<<<<<<
 *         return mslice
 * 
 */
  /*else*/ {
    __pyx_memoryview_slice_copy(__pyx_v_memview, __pyx_v_mslice);

    /* "View.MemoryView":1060
 *     else:
 *         slice_copy(memview, mslice)
 *         return mslice             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_slice_copy')
 */
    __pyx_r = __pyx_v_mslice;
    goto __pyx_L0;
  }

  /* "View.MemoryView":1052
 * 
 * @cname('__pyx_memoryview_get_slice_from_memoryview')
 * cdef __Pyx_memviewslice *get_slice_from_memview(memoryview memview,             # <<<<<<<<<<<<<<
 *                                                    __Pyx_memviewslice *mslice):
 *     cdef _memoryviewslice obj
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_WriteUnraisable("View.MemoryView.get_slice_from_memview", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_obj);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":1063
 * 
 * @cname('__pyx_memoryview_slice_copy')
 * cdef void slice_copy(memoryview memview, __Pyx_memviewslice *dst):             # <<<<<<<<<<<<<<
 *     cdef int dim
 *     cdef (Py_ssize_t*) shape, strides, suboffsets
 */

static void __pyx_memoryview_slice_copy(struct __pyx_memoryview_obj *__pyx_v_memview, __Pyx_memviewslice *__pyx_v_dst) {
  int __pyx_v_dim;
  Py_ssize_t *__pyx_v_shape;
  Py_ssize_t *__pyx_v_strides;
  Py_ssize_t *__pyx_v_suboffsets;
  __Pyx_RefNannyDeclarations
  Py_ssize_t *__pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  __Pyx_RefNannySetupContext("slice_copy", 0);

  /* "View.MemoryView":1067
 *     cdef (Py_ssize_t*) shape, strides, suboffsets
 * 
 *     shape = memview.view.shape             # <<<<<<<<<<<<<<
 *     strides = memview.view.strides
 *     suboffsets = memview.view.suboffsets
 */
  __pyx_t_1 = __pyx_v_memview->view.shape;
  __pyx_v_shape = __pyx_t_1;

  /* "View.MemoryView":1068
 * 
 *     shape = memview.view.shape
 *     strides = memview.view.strides             # <<<<<<<<<<<<<<
 *     suboffsets = memview.view.suboffsets
 * 
 */
  __pyx_t_1 = __pyx_v_memview->view.strides;
  __pyx_v_strides = __pyx_t_1;

  /* "View.MemoryView":1069
 *     shape = memview.view.shape
 *     strides = memview.view.strides
 *     suboffsets = memview.view.suboffsets             # <<<<<<<<<<<<<<
 * 
 *     dst.memview = <__pyx_memoryview *> memview
 */
  __pyx_t_1 = __pyx_v_memview->view.suboffsets;
  __pyx_v_suboffsets = __pyx_t_1;

  /* "View.MemoryView":1071
 *     suboffsets = memview.view.suboffsets
 * 
 *     dst.memview = <__pyx_memoryview *> memview             # <<<<<<<<<<<<<<
 *     dst.data = <char *> memview.view.buf
 * 
 */
  __pyx_v_dst->memview = ((struct __pyx_memoryview_obj *)__pyx_v_memview);

  /* "View.MemoryView":1072
 * 
 *     dst.memview = <__pyx_memoryview *> memview
 *     dst.data = <char *> memview.view.buf             # <<<<<<<<<<<<<<
 * 
 *     for dim in range(memview.view.ndim):
 */
  __pyx_v_dst->data = ((char *)__pyx_v_memview->view.buf);

  /* "View.MemoryView":1074
 *     dst.data = <char *> memview.view.buf
 * 
 *     for dim in range(memview.view.ndim):             # <<<<<<<<<<<<<<
 *         dst.shape[dim] = shape[dim]
 *         dst.strides[dim] = strides[dim]
 */
  __pyx_t_2 = __pyx_v_memview->view.ndim;
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_dim = __pyx_t_4;

    /* "View.MemoryView":1075
 * 
 *     for dim in range(memview.view.ndim):
 *         dst.shape[dim] = shape[dim]             # <<<<<<<<<<<<<<
 *         dst.strides[dim] = strides[dim]
 *         dst.suboffsets[dim] = suboffsets[dim] if suboffsets else -1
 */
    (__pyx_v_dst->shape[__pyx_v_dim]) = (__pyx_v_shape[__pyx_v_dim]);

    /* "View.MemoryView":1076
 *     for dim in range(memview.view.ndim):
 *         dst.shape[dim] = shape[dim]
 *         dst.strides[dim] = strides[dim]             # <<<<<<<<<<<<<<
 *         dst.suboffsets[dim] = suboffsets[dim] if suboffsets else -1
 * 
 */
    (__pyx_v_dst->strides[__pyx_v_dim]) = (__pyx_v_strides[__pyx_v_dim]);

    /* "View.MemoryView":1077
 *         dst.shape[dim] = shape[dim]
 *         dst.strides[dim] = strides[dim]
 *         dst.suboffsets[dim] = suboffsets[dim] if suboffsets else -1             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_copy_object')
 */
    if ((__pyx_v_suboffsets != 0)) {
      __pyx_t_5 = (__pyx_v_suboffsets[__pyx_v_dim]);
    } else {
      __pyx_t_5 = -1L;
    }
    (__pyx_v_dst->suboffsets[__pyx_v_dim]) = __pyx_t_5;
  }

  /* "View.MemoryView":1063
 * 
 * @cname('__pyx_memoryview_slice_copy')
 * cdef void slice_copy(memoryview memview, __Pyx_memviewslice *dst):             # <<<<<<<<<<<<<<
 *     cdef int dim
 *     cdef (Py_ssize_t*) shape, strides, suboffsets
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "View.MemoryView":1080
 * 
 * @cname('__pyx_memoryview_copy_object')
 * cdef memoryview_copy(memoryview memview):             # <<<<<<<<<<<<<<
 *     "Create a new memoryview object"
 *     cdef __Pyx_memviewslice memviewslice
 */

static PyObject *__pyx_memoryview_copy_object(struct __pyx_memoryview_obj *__pyx_v_memview) {
  __Pyx_memviewslice __pyx_v_memviewslice;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("memoryview_copy", 0);

  /* "View.MemoryView":1083
 *     "Create a new memoryview object"
 *     cdef __Pyx_memviewslice memviewslice
 *     slice_copy(memview, &memviewslice)             # <<<<<<<<<<<<<<
 *     return memoryview_copy_from_slice(memview, &memviewslice)
 * 
 */
  __pyx_memoryview_slice_copy(__pyx_v_memview, (&__pyx_v_memviewslice));

  /* "View.MemoryView":1084
 *     cdef __Pyx_memviewslice memviewslice
 *     slice_copy(memview, &memviewslice)
 *     return memoryview_copy_from_slice(memview, &memviewslice)             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_copy_object_from_slice')
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_memoryview_copy_object_from_slice(__pyx_v_memview, (&__pyx_v_memviewslice)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 1084, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":1080
 * 
 * @cname('__pyx_memoryview_copy_object')
 * cdef memoryview_copy(memoryview memview):             # <<<<<<<<<<<<<<
 *     "Create a new memoryview object"
 *     cdef __Pyx_memviewslice memviewslice
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.memoryview_copy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":1087
 * 
 * @cname('__pyx_memoryview_copy_object_from_slice')
 * cdef memoryview_copy_from_slice(memoryview memview, __Pyx_memviewslice *memviewslice):             # <<<<<<<<<<<<<<
 *     """
 *     Create a new memoryview object from a given memoryview object and slice.
 */

static PyObject *__pyx_memoryview_copy_object_from_slice(struct __pyx_memoryview_obj *__pyx_v_memview, __Pyx_memviewslice *__pyx_v_memviewslice) {
  PyObject *(*__pyx_v_to_object_func)(char *);
  int (*__pyx_v_to_dtype_func)(char *, PyObject *);
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *(*__pyx_t_3)(char *);
  int (*__pyx_t_4)(char *, PyObject *);
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("memoryview_copy_from_slice", 0);

  /* "View.MemoryView":1094
 *     cdef int (*to_dtype_func)(char *, object) except 0
 * 
 *     if isinstance(memview, _memoryviewslice):             # <<<<<<<<<<<<<<
 *         to_object_func = (<_memoryviewslice> memview).to_object_func
 *         to_dtype_func = (<_memoryviewslice> memview).to_dtype_func
 */
  __pyx_t_1 = __Pyx_TypeCheck(((PyObject *)__pyx_v_memview), __pyx_memoryviewslice_type); 
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":1095
 * 
 *     if isinstance(memview, _memoryviewslice):
 *         to_object_func = (<_memoryviewslice> memview).to_object_func             # <<<<<<<<<<<<<<
 *         to_dtype_func = (<_memoryviewslice> memview).to_dtype_func
 *     else:
 */
    __pyx_t_3 = ((struct __pyx_memoryviewslice_obj *)__pyx_v_memview)->to_object_func;
    __pyx_v_to_object_func = __pyx_t_3;

    /* "View.MemoryView":1096
 *     if isinstance(memview, _memoryviewslice):
 *         to_object_func = (<_memoryviewslice> memview).to_object_func
 *         to_dtype_func = (<_memoryviewslice> memview).to_dtype_func             # <<<<<<<<<<<<<<
 *     else:
 *         to_object_func = NULL
 */
    __pyx_t_4 = ((struct __pyx_memoryviewslice_obj *)__pyx_v_memview)->to_dtype_func;
    __pyx_v_to_dtype_func = __pyx_t_4;

    /* "View.MemoryView":1094
 *     cdef int (*to_dtype_func)(char *, object) except 0
 * 
 *     if isinstance(memview, _memoryviewslice):             # <<<<<<<<<<<<<<
 *         to_object_func = (<_memoryviewslice> memview).to_object_func
 *         to_dtype_func = (<_memoryviewslice> memview).to_dtype_func
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":1098
 *         to_dtype_func = (<_memoryviewslice> memview).to_dtype_func
 *     else:
 *         to_object_func = NULL             # <<<<<<<<<<<<<<
 *         to_dtype_func = NULL
 * 
 */
  /*else*/ {
    __pyx_v_to_object_func = NULL;

    /* "View.MemoryView":1099
 *     else:
 *         to_object_func = NULL
 *         to_dtype_func = NULL             # <<<<<<<<<<<<<<
 * 
 *     return memoryview_fromslice(memviewslice[0], memview.view.ndim,
 */
    __pyx_v_to_dtype_func = NULL;
  }
  __pyx_L3:;

  /* "View.MemoryView":1101
 *         to_dtype_func = NULL
 * 
 *     return memoryview_fromslice(memviewslice[0], memview.view.ndim,             # <<<<<<<<<<<<<<
 *                                 to_object_func, to_dtype_func,
 *                                 memview.dtype_is_object)
 */
  __Pyx_XDECREF(__pyx_r);

  /* "View.MemoryView":1103
 *     return memoryview_fromslice(memviewslice[0], memview.view.ndim,
 *                                 to_object_func, to_dtype_func,
 *                                 memview.dtype_is_object)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_5 = __pyx_memoryview_fromslice((__pyx_v_memviewslice[0]), __pyx_v_memview->view.ndim, __pyx_v_to_object_func, __pyx_v_to_dtype_func, __pyx_v_memview->dtype_is_object); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 1101, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_r = __pyx_t_5;
  __pyx_t_5 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":1087
 * 
 * @cname('__pyx_memoryview_copy_object_from_slice')
 * cdef memoryview_copy_from_slice(memoryview memview, __Pyx_memviewslice *memviewslice):             # <<<<<<<<<<<<<<
 *     """
 *     Create a new memoryview object from a given memoryview object and slice.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("View.MemoryView.memoryview_copy_from_slice", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":1109
 * 
 * 
 * cdef Py_ssize_t abs_py_ssize_t(Py_ssize_t arg) nogil:             # <<<<<<<<<<<<<<
 *     if arg < 0:
 *         return -arg
 */

static Py_ssize_t abs_py_ssize_t(Py_ssize_t __pyx_v_arg) {
  Py_ssize_t __pyx_r;
  int __pyx_t_1;

  /* "View.MemoryView":1110
 * 
 * cdef Py_ssize_t abs_py_ssize_t(Py_ssize_t arg) nogil:
 *     if arg < 0:             # <<<<<<<<<<<<<<
 *         return -arg
 *     else:
 */
  __pyx_t_1 = ((__pyx_v_arg < 0) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":1111
 * cdef Py_ssize_t abs_py_ssize_t(Py_ssize_t arg) nogil:
 *     if arg < 0:
 *         return -arg             # <<<<<<<<<<<<<<
 *     else:
 *         return arg
 */
    __pyx_r = (-__pyx_v_arg);
    goto __pyx_L0;

    /* "View.MemoryView":1110
 * 
 * cdef Py_ssize_t abs_py_ssize_t(Py_ssize_t arg) nogil:
 *     if arg < 0:             # <<<<<<<<<<<<<<
 *         return -arg
 *     else:
 */
  }

  /* "View.MemoryView":1113
 *         return -arg
 *     else:
 *         return arg             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_get_best_slice_order')
 */
  /*else*/ {
    __pyx_r = __pyx_v_arg;
    goto __pyx_L0;
  }

  /* "View.MemoryView":1109
 * 
 * 
 * cdef Py_ssize_t abs_py_ssize_t(Py_ssize_t arg) nogil:             # <<<<<<<<<<<<<<
 *     if arg < 0:
 *         return -arg
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "View.MemoryView":1116
 * 
 * @cname('__pyx_get_best_slice_order')
 * cdef char get_best_order(__Pyx_memviewslice *mslice, int ndim) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     Figure out the best memory access order for a given slice.
 */

static char __pyx_get_best_slice_order(__Pyx_memviewslice *__pyx_v_mslice, int __pyx_v_ndim) {
  int __pyx_v_i;
  Py_ssize_t __pyx_v_c_stride;
  Py_ssize_t __pyx_v_f_stride;
  char __pyx_r;
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;

  /* "View.MemoryView":1121
 *     """
 *     cdef int i
 *     cdef Py_ssize_t c_stride = 0             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t f_stride = 0
 * 
 */
  __pyx_v_c_stride = 0;

  /* "View.MemoryView":1122
 *     cdef int i
 *     cdef Py_ssize_t c_stride = 0
 *     cdef Py_ssize_t f_stride = 0             # <<<<<<<<<<<<<<
 * 
 *     for i in range(ndim - 1, -1, -1):
 */
  __pyx_v_f_stride = 0;

  /* "View.MemoryView":1124
 *     cdef Py_ssize_t f_stride = 0
 * 
 *     for i in range(ndim - 1, -1, -1):             # <<<<<<<<<<<<<<
 *         if mslice.shape[i] > 1:
 *             c_stride = mslice.strides[i]
 */
  for (__pyx_t_1 = (__pyx_v_ndim - 1); __pyx_t_1 > -1; __pyx_t_1-=1) {
    __pyx_v_i = __pyx_t_1;

    /* "View.MemoryView":1125
 * 
 *     for i in range(ndim - 1, -1, -1):
 *         if mslice.shape[i] > 1:             # <<<<<<<<<<<<<<
 *             c_stride = mslice.strides[i]
 *             break
 */
    __pyx_t_2 = (((__pyx_v_mslice->shape[__pyx_v_i]) > 1) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":1126
 *     for i in range(ndim - 1, -1, -1):
 *         if mslice.shape[i] > 1:
 *             c_stride = mslice.strides[i]             # <<<<<<<<<<<<<<
 *             break
 * 
 */
      __pyx_v_c_stride = (__pyx_v_mslice->strides[__pyx_v_i]);

      /* "View.MemoryView":1127
 *         if mslice.shape[i] > 1:
 *             c_stride = mslice.strides[i]
 *             break             # <<<<<<<<<<<<<<
 * 
 *     for i in range(ndim):
 */
      goto __pyx_L4_break;

      /* "View.MemoryView":1125
 * 
 *     for i in range(ndim - 1, -1, -1):
 *         if mslice.shape[i] > 1:             # <<<<<<<<<<<<<<
 *             c_stride = mslice.strides[i]
 *             break
 */
    }
  }
  __pyx_L4_break:;

  /* "View.MemoryView":1129
 *             break
 * 
 *     for i in range(ndim):             # <<<<<<<<<<<<<<
 *         if mslice.shape[i] > 1:
 *             f_stride = mslice.strides[i]
 */
  __pyx_t_1 = __pyx_v_ndim;
  __pyx_t_3 = __pyx_t_1;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "View.MemoryView":1130
 * 
 *     for i in range(ndim):
 *         if mslice.shape[i] > 1:             # <<<<<<<<<<<<<<
 *             f_stride = mslice.strides[i]
 *             break
 */
    __pyx_t_2 = (((__pyx_v_mslice->shape[__pyx_v_i]) > 1) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":1131
 *     for i in range(ndim):
 *         if mslice.shape[i] > 1:
 *             f_stride = mslice.strides[i]             # <<<<<<<<<<<<<<
 *             break
 * 
 */
      __pyx_v_f_stride = (__pyx_v_mslice->strides[__pyx_v_i]);

      /* "View.MemoryView":1132
 *         if mslice.shape[i] > 1:
 *             f_stride = mslice.strides[i]
 *             break             # <<<<<<<<<<<<<<
 * 
 *     if abs_py_ssize_t(c_stride) <= abs_py_ssize_t(f_stride):
 */
      goto __pyx_L7_break;

      /* "View.MemoryView":1130
 * 
 *     for i in range(ndim):
 *         if mslice.shape[i] > 1:             # <<<<<<<<<<<<<<
 *             f_stride = mslice.strides[i]
 *             break
 */
    }
  }
  __pyx_L7_break:;

  /* "View.MemoryView":1134
 *             break
 * 
 *     if abs_py_ssize_t(c_stride) <= abs_py_ssize_t(f_stride):             # <<<<<<<<<<<<<<
 *         return 'C'
 *     else:
 */
  __pyx_t_2 = ((abs_py_ssize_t(__pyx_v_c_stride) <= abs_py_ssize_t(__pyx_v_f_stride)) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":1135
 * 
 *     if abs_py_ssize_t(c_stride) <= abs_py_ssize_t(f_stride):
 *         return 'C'             # <<<<<<<<<<<<<<
 *     else:
 *         return 'F'
 */
    __pyx_r = 'C';
    goto __pyx_L0;

    /* "View.MemoryView":1134
 *             break
 * 
 *     if abs_py_ssize_t(c_stride) <= abs_py_ssize_t(f_stride):             # <<<<<<<<<<<<<<
 *         return 'C'
 *     else:
 */
  }

  /* "View.MemoryView":1137
 *         return 'C'
 *     else:
 *         return 'F'             # <<<<<<<<<<<<<<
 * 
 * @cython.cdivision(True)
 */
  /*else*/ {
    __pyx_r = 'F';
    goto __pyx_L0;
  }

  /* "View.MemoryView":1116
 * 
 * @cname('__pyx_get_best_slice_order')
 * cdef char get_best_order(__Pyx_memviewslice *mslice, int ndim) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     Figure out the best memory access order for a given slice.
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "View.MemoryView":1140
 * 
 * @cython.cdivision(True)
 * cdef void _copy_strided_to_strided(char *src_data, Py_ssize_t *src_strides,             # <<<<<<<<<<<<<<
 *                                    char *dst_data, Py_ssize_t *dst_strides,
 *                                    Py_ssize_t *src_shape, Py_ssize_t *dst_shape,
 */

static void _copy_strided_to_strided(char *__pyx_v_src_data, Py_ssize_t *__pyx_v_src_strides, char *__pyx_v_dst_data, Py_ssize_t *__pyx_v_dst_strides, Py_ssize_t *__pyx_v_src_shape, Py_ssize_t *__pyx_v_dst_shape, int __pyx_v_ndim, size_t __pyx_v_itemsize) {
  CYTHON_UNUSED Py_ssize_t __pyx_v_i;
  CYTHON_UNUSED Py_ssize_t __pyx_v_src_extent;
  Py_ssize_t __pyx_v_dst_extent;
  Py_ssize_t __pyx_v_src_stride;
  Py_ssize_t __pyx_v_dst_stride;
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  Py_ssize_t __pyx_t_6;

  /* "View.MemoryView":1147
 * 
 *     cdef Py_ssize_t i
 *     cdef Py_ssize_t src_extent = src_shape[0]             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t dst_extent = dst_shape[0]
 *     cdef Py_ssize_t src_stride = src_strides[0]
 */
  __pyx_v_src_extent = (__pyx_v_src_shape[0]);

  /* "View.MemoryView":1148
 *     cdef Py_ssize_t i
 *     cdef Py_ssize_t src_extent = src_shape[0]
 *     cdef Py_ssize_t dst_extent = dst_shape[0]             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t src_stride = src_strides[0]
 *     cdef Py_ssize_t dst_stride = dst_strides[0]
 */
  __pyx_v_dst_extent = (__pyx_v_dst_shape[0]);

  /* "View.MemoryView":1149
 *     cdef Py_ssize_t src_extent = src_shape[0]
 *     cdef Py_ssize_t dst_extent = dst_shape[0]
 *     cdef Py_ssize_t src_stride = src_strides[0]             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t dst_stride = dst_strides[0]
 * 
 */
  __pyx_v_src_stride = (__pyx_v_src_strides[0]);

  /* "View.MemoryView":1150
 *     cdef Py_ssize_t dst_extent = dst_shape[0]
 *     cdef Py_ssize_t src_stride = src_strides[0]
 *     cdef Py_ssize_t dst_stride = dst_strides[0]             # <<<<<<<<<<<<<<
 * 
 *     if ndim == 1:
 */
  __pyx_v_dst_stride = (__pyx_v_dst_strides[0]);

  /* "View.MemoryView":1152
 *     cdef Py_ssize_t dst_stride = dst_strides[0]
 * 
 *     if ndim == 1:             # <<<<<<<<<<<<<<
 *        if (src_stride > 0 and dst_stride > 0 and
 *            <size_t> src_stride == itemsize == <size_t> dst_stride):
 */
  __pyx_t_1 = ((__pyx_v_ndim == 1) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":1153
 * 
 *     if ndim == 1:
 *        if (src_stride > 0 and dst_stride > 0 and             # <<<<<<<<<<<<<<
 *            <size_t> src_stride == itemsize == <size_t> dst_stride):
 *            memcpy(dst_data, src_data, itemsize * dst_extent)
 */
    __pyx_t_2 = ((__pyx_v_src_stride > 0) != 0);
    if (__pyx_t_2) {
    } else {
      __pyx_t_1 = __pyx_t_2;
      goto __pyx_L5_bool_binop_done;
    }
    __pyx_t_2 = ((__pyx_v_dst_stride > 0) != 0);
    if (__pyx_t_2) {
    } else {
      __pyx_t_1 = __pyx_t_2;
      goto __pyx_L5_bool_binop_done;
    }

    /* "View.MemoryView":1154
 *     if ndim == 1:
 *        if (src_stride > 0 and dst_stride > 0 and
 *            <size_t> src_stride == itemsize == <size_t> dst_stride):             # <<<<<<<<<<<<<<
 *            memcpy(dst_data, src_data, itemsize * dst_extent)
 *        else:
 */
    __pyx_t_2 = (((size_t)__pyx_v_src_stride) == __pyx_v_itemsize);
    if (__pyx_t_2) {
      __pyx_t_2 = (__pyx_v_itemsize == ((size_t)__pyx_v_dst_stride));
    }
    __pyx_t_3 = (__pyx_t_2 != 0);
    __pyx_t_1 = __pyx_t_3;
    __pyx_L5_bool_binop_done:;

    /* "View.MemoryView":1153
 * 
 *     if ndim == 1:
 *        if (src_stride > 0 and dst_stride > 0 and             # <<<<<<<<<<<<<<
 *            <size_t> src_stride == itemsize == <size_t> dst_stride):
 *            memcpy(dst_data, src_data, itemsize * dst_extent)
 */
    if (__pyx_t_1) {

      /* "View.MemoryView":1155
 *        if (src_stride > 0 and dst_stride > 0 and
 *            <size_t> src_stride == itemsize == <size_t> dst_stride):
 *            memcpy(dst_data, src_data, itemsize * dst_extent)             # <<<<<<<<<<<<<<
 *        else:
 *            for i in range(dst_extent):
 */
      (void)(memcpy(__pyx_v_dst_data, __pyx_v_src_data, (__pyx_v_itemsize * __pyx_v_dst_extent)));

      /* "View.MemoryView":1153
 * 
 *     if ndim == 1:
 *        if (src_stride > 0 and dst_stride > 0 and             # <<<<<<<<<<<<<<
 *            <size_t> src_stride == itemsize == <size_t> dst_stride):
 *            memcpy(dst_data, src_data, itemsize * dst_extent)
 */
      goto __pyx_L4;
    }

    /* "View.MemoryView":1157
 *            memcpy(dst_data, src_data, itemsize * dst_extent)
 *        else:
 *            for i in range(dst_extent):             # <<<<<<<<<<<<<<
 *                memcpy(dst_data, src_data, itemsize)
 *                src_data += src_stride
 */
    /*else*/ {
      __pyx_t_4 = __pyx_v_dst_extent;
      __pyx_t_5 = __pyx_t_4;
      for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
        __pyx_v_i = __pyx_t_6;

        /* "View.MemoryView":1158
 *        else:
 *            for i in range(dst_extent):
 *                memcpy(dst_data, src_data, itemsize)             # <<<<<<<<<<<<<<
 *                src_data += src_stride
 *                dst_data += dst_stride
 */
        (void)(memcpy(__pyx_v_dst_data, __pyx_v_src_data, __pyx_v_itemsize));

        /* "View.MemoryView":1159
 *            for i in range(dst_extent):
 *                memcpy(dst_data, src_data, itemsize)
 *                src_data += src_stride             # <<<<<<<<<<<<<<
 *                dst_data += dst_stride
 *     else:
 */
        __pyx_v_src_data = (__pyx_v_src_data + __pyx_v_src_stride);

        /* "View.MemoryView":1160
 *                memcpy(dst_data, src_data, itemsize)
 *                src_data += src_stride
 *                dst_data += dst_stride             # <<<<<<<<<<<<<<
 *     else:
 *         for i in range(dst_extent):
 */
        __pyx_v_dst_data = (__pyx_v_dst_data + __pyx_v_dst_stride);
      }
    }
    __pyx_L4:;

    /* "View.MemoryView":1152
 *     cdef Py_ssize_t dst_stride = dst_strides[0]
 * 
 *     if ndim == 1:             # <<<<<<<<<<<<<<
 *        if (src_stride > 0 and dst_stride > 0 and
 *            <size_t> src_stride == itemsize == <size_t> dst_stride):
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":1162
 *                dst_data += dst_stride
 *     else:
 *         for i in range(dst_extent):             # <<<<<<<<<<<<<<
 *             _copy_strided_to_strided(src_data, src_strides + 1,
 *                                      dst_data, dst_strides + 1,
 */
  /*else*/ {
    __pyx_t_4 = __pyx_v_dst_extent;
    __pyx_t_5 = __pyx_t_4;
    for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
      __pyx_v_i = __pyx_t_6;

      /* "View.MemoryView":1163
 *     else:
 *         for i in range(dst_extent):
 *             _copy_strided_to_strided(src_data, src_strides + 1,             # <<<<<<<<<<<<<<
 *                                      dst_data, dst_strides + 1,
 *                                      src_shape + 1, dst_shape + 1,
 */
      _copy_strided_to_strided(__pyx_v_src_data, (__pyx_v_src_strides + 1), __pyx_v_dst_data, (__pyx_v_dst_strides + 1), (__pyx_v_src_shape + 1), (__pyx_v_dst_shape + 1), (__pyx_v_ndim - 1), __pyx_v_itemsize);

      /* "View.MemoryView":1167
 *                                      src_shape + 1, dst_shape + 1,
 *                                      ndim - 1, itemsize)
 *             src_data += src_stride             # <<<<<<<<<<<<<<
 *             dst_data += dst_stride
 * 
 */
      __pyx_v_src_data = (__pyx_v_src_data + __pyx_v_src_stride);

      /* "View.MemoryView":1168
 *                                      ndim - 1, itemsize)
 *             src_data += src_stride
 *             dst_data += dst_stride             # <<<<<<<<<<<<<<
 * 
 * cdef void copy_strided_to_strided(__Pyx_memviewslice *src,
 */
      __pyx_v_dst_data = (__pyx_v_dst_data + __pyx_v_dst_stride);
    }
  }
  __pyx_L3:;

  /* "View.MemoryView":1140
 * 
 * @cython.cdivision(True)
 * cdef void _copy_strided_to_strided(char *src_data, Py_ssize_t *src_strides,             # <<<<<<<<<<<<<<
 *                                    char *dst_data, Py_ssize_t *dst_strides,
 *                                    Py_ssize_t *src_shape, Py_ssize_t *dst_shape,
 */

  /* function exit code */
}

/* "View.MemoryView":1170
 *             dst_data += dst_stride
 * 
 * cdef void copy_strided_to_strided(__Pyx_memviewslice *src,             # <<<<<<<<<<<<<<
 *                                   __Pyx_memviewslice *dst,
 *                                   int ndim, size_t itemsize) nogil:
 */

static void copy_strided_to_strided(__Pyx_memviewslice *__pyx_v_src, __Pyx_memviewslice *__pyx_v_dst, int __pyx_v_ndim, size_t __pyx_v_itemsize) {

  /* "View.MemoryView":1173
 *                                   __Pyx_memviewslice *dst,
 *                                   int ndim, size_t itemsize) nogil:
 *     _copy_strided_to_strided(src.data, src.strides, dst.data, dst.strides,             # <<<<<<<<<<<<<<
 *                              src.shape, dst.shape, ndim, itemsize)
 * 
 */
  _copy_strided_to_strided(__pyx_v_src->data, __pyx_v_src->strides, __pyx_v_dst->data, __pyx_v_dst->strides, __pyx_v_src->shape, __pyx_v_dst->shape, __pyx_v_ndim, __pyx_v_itemsize);

  /* "View.MemoryView":1170
 *             dst_data += dst_stride
 * 
 * cdef void copy_strided_to_strided(__Pyx_memviewslice *src,             # <<<<<<<<<<<<<<
 *                                   __Pyx_memviewslice *dst,
 *                                   int ndim, size_t itemsize) nogil:
 */

  /* function exit code */
}

/* "View.MemoryView":1177
 * 
 * @cname('__pyx_memoryview_slice_get_size')
 * cdef Py_ssize_t slice_get_size(__Pyx_memviewslice *src, int ndim) nogil:             # <<<<<<<<<<<<<<
 *     "Return the size of the memory occupied by the slice in number of bytes"
 *     cdef int i
 */

static Py_ssize_t __pyx_memoryview_slice_get_size(__Pyx_memviewslice *__pyx_v_src, int __pyx_v_ndim) {
  int __pyx_v_i;
  Py_ssize_t __pyx_v_size;
  Py_ssize_t __pyx_r;
  Py_ssize_t __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;

  /* "View.MemoryView":1180
 *     "Return the size of the memory occupied by the slice in number of bytes"
 *     cdef int i
 *     cdef Py_ssize_t size = src.memview.view.itemsize             # <<<<<<<<<<<<<<
 * 
 *     for i in range(ndim):
 */
  __pyx_t_1 = __pyx_v_src->memview->view.itemsize;
  __pyx_v_size = __pyx_t_1;

  /* "View.MemoryView":1182
 *     cdef Py_ssize_t size = src.memview.view.itemsize
 * 
 *     for i in range(ndim):             # <<<<<<<<<<<<<<
 *         size *= src.shape[i]
 * 
 */
  __pyx_t_2 = __pyx_v_ndim;
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "View.MemoryView":1183
 * 
 *     for i in range(ndim):
 *         size *= src.shape[i]             # <<<<<<<<<<<<<<
 * 
 *     return size
 */
    __pyx_v_size = (__pyx_v_size * (__pyx_v_src->shape[__pyx_v_i]));
  }

  /* "View.MemoryView":1185
 *         size *= src.shape[i]
 * 
 *     return size             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_fill_contig_strides_array')
 */
  __pyx_r = __pyx_v_size;
  goto __pyx_L0;

  /* "View.MemoryView":1177
 * 
 * @cname('__pyx_memoryview_slice_get_size')
 * cdef Py_ssize_t slice_get_size(__Pyx_memviewslice *src, int ndim) nogil:             # <<<<<<<<<<<<<<
 *     "Return the size of the memory occupied by the slice in number of bytes"
 *     cdef int i
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "View.MemoryView":1188
 * 
 * @cname('__pyx_fill_contig_strides_array')
 * cdef Py_ssize_t fill_contig_strides_array(             # <<<<<<<<<<<<<<
 *                 Py_ssize_t *shape, Py_ssize_t *strides, Py_ssize_t stride,
 *                 int ndim, char order) nogil:
 */

static Py_ssize_t __pyx_fill_contig_strides_array(Py_ssize_t *__pyx_v_shape, Py_ssize_t *__pyx_v_strides, Py_ssize_t __pyx_v_stride, int __pyx_v_ndim, char __pyx_v_order) {
  int __pyx_v_idx;
  Py_ssize_t __pyx_r;
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;

  /* "View.MemoryView":1197
 *     cdef int idx
 * 
 *     if order == 'F':             # <<<<<<<<<<<<<<
 *         for idx in range(ndim):
 *             strides[idx] = stride
 */
  __pyx_t_1 = ((__pyx_v_order == 'F') != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":1198
 * 
 *     if order == 'F':
 *         for idx in range(ndim):             # <<<<<<<<<<<<<<
 *             strides[idx] = stride
 *             stride = stride * shape[idx]
 */
    __pyx_t_2 = __pyx_v_ndim;
    __pyx_t_3 = __pyx_t_2;
    for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
      __pyx_v_idx = __pyx_t_4;

      /* "View.MemoryView":1199
 *     if order == 'F':
 *         for idx in range(ndim):
 *             strides[idx] = stride             # <<<<<<<<<<<<<<
 *             stride = stride * shape[idx]
 *     else:
 */
      (__pyx_v_strides[__pyx_v_idx]) = __pyx_v_stride;

      /* "View.MemoryView":1200
 *         for idx in range(ndim):
 *             strides[idx] = stride
 *             stride = stride * shape[idx]             # <<<<<<<<<<<<<<
 *     else:
 *         for idx in range(ndim - 1, -1, -1):
 */
      __pyx_v_stride = (__pyx_v_stride * (__pyx_v_shape[__pyx_v_idx]));
    }

    /* "View.MemoryView":1197
 *     cdef int idx
 * 
 *     if order == 'F':             # <<<<<<<<<<<<<<
 *         for idx in range(ndim):
 *             strides[idx] = stride
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":1202
 *             stride = stride * shape[idx]
 *     else:
 *         for idx in range(ndim - 1, -1, -1):             # <<<<<<<<<<<<<<
 *             strides[idx] = stride
 *             stride = stride * shape[idx]
 */
  /*else*/ {
    for (__pyx_t_2 = (__pyx_v_ndim - 1); __pyx_t_2 > -1; __pyx_t_2-=1) {
      __pyx_v_idx = __pyx_t_2;

      /* "View.MemoryView":1203
 *     else:
 *         for idx in range(ndim - 1, -1, -1):
 *             strides[idx] = stride             # <<<<<<<<<<<<<<
 *             stride = stride * shape[idx]
 * 
 */
      (__pyx_v_strides[__pyx_v_idx]) = __pyx_v_stride;

      /* "View.MemoryView":1204
 *         for idx in range(ndim - 1, -1, -1):
 *             strides[idx] = stride
 *             stride = stride * shape[idx]             # <<<<<<<<<<<<<<
 * 
 *     return stride
 */
      __pyx_v_stride = (__pyx_v_stride * (__pyx_v_shape[__pyx_v_idx]));
    }
  }
  __pyx_L3:;

  /* "View.MemoryView":1206
 *             stride = stride * shape[idx]
 * 
 *     return stride             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_copy_data_to_temp')
 */
  __pyx_r = __pyx_v_stride;
  goto __pyx_L0;

  /* "View.MemoryView":1188
 * 
 * @cname('__pyx_fill_contig_strides_array')
 * cdef Py_ssize_t fill_contig_strides_array(             # <<<<<<<<<<<<<<
 *                 Py_ssize_t *shape, Py_ssize_t *strides, Py_ssize_t stride,
 *                 int ndim, char order) nogil:
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "View.MemoryView":1209
 * 
 * @cname('__pyx_memoryview_copy_data_to_temp')
 * cdef void *copy_data_to_temp(__Pyx_memviewslice *src,             # <<<<<<<<<<<<<<
 *                              __Pyx_memviewslice *tmpslice,
 *                              char order,
 */

static void *__pyx_memoryview_copy_data_to_temp(__Pyx_memviewslice *__pyx_v_src, __Pyx_memviewslice *__pyx_v_tmpslice, char __pyx_v_order, int __pyx_v_ndim) {
  int __pyx_v_i;
  void *__pyx_v_result;
  size_t __pyx_v_itemsize;
  size_t __pyx_v_size;
  void *__pyx_r;
  Py_ssize_t __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  struct __pyx_memoryview_obj *__pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;

  /* "View.MemoryView":1220
 *     cdef void *result
 * 
 *     cdef size_t itemsize = src.memview.view.itemsize             # <<<<<<<<<<<<<<
 *     cdef size_t size = slice_get_size(src, ndim)
 * 
 */
  __pyx_t_1 = __pyx_v_src->memview->view.itemsize;
  __pyx_v_itemsize = __pyx_t_1;

  /* "View.MemoryView":1221
 * 
 *     cdef size_t itemsize = src.memview.view.itemsize
 *     cdef size_t size = slice_get_size(src, ndim)             # <<<<<<<<<<<<<<
 * 
 *     result = malloc(size)
 */
  __pyx_v_size = __pyx_memoryview_slice_get_size(__pyx_v_src, __pyx_v_ndim);

  /* "View.MemoryView":1223
 *     cdef size_t size = slice_get_size(src, ndim)
 * 
 *     result = malloc(size)             # <<<<<<<<<<<<<<
 *     if not result:
 *         _err(MemoryError, NULL)
 */
  __pyx_v_result = malloc(__pyx_v_size);

  /* "View.MemoryView":1224
 * 
 *     result = malloc(size)
 *     if not result:             # <<<<<<<<<<<<<<
 *         _err(MemoryError, NULL)
 * 
 */
  __pyx_t_2 = ((!(__pyx_v_result != 0)) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":1225
 *     result = malloc(size)
 *     if not result:
 *         _err(MemoryError, NULL)             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_t_3 = __pyx_memoryview_err(__pyx_builtin_MemoryError, NULL); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(2, 1225, __pyx_L1_error)

    /* "View.MemoryView":1224
 * 
 *     result = malloc(size)
 *     if not result:             # <<<<<<<<<<<<<<
 *         _err(MemoryError, NULL)
 * 
 */
  }

  /* "View.MemoryView":1228
 * 
 * 
 *     tmpslice.data = <char *> result             # <<<<<<<<<<<<<<
 *     tmpslice.memview = src.memview
 *     for i in range(ndim):
 */
  __pyx_v_tmpslice->data = ((char *)__pyx_v_result);

  /* "View.MemoryView":1229
 * 
 *     tmpslice.data = <char *> result
 *     tmpslice.memview = src.memview             # <<<<<<<<<<<<<<
 *     for i in range(ndim):
 *         tmpslice.shape[i] = src.shape[i]
 */
  __pyx_t_4 = __pyx_v_src->memview;
  __pyx_v_tmpslice->memview = __pyx_t_4;

  /* "View.MemoryView":1230
 *     tmpslice.data = <char *> result
 *     tmpslice.memview = src.memview
 *     for i in range(ndim):             # <<<<<<<<<<<<<<
 *         tmpslice.shape[i] = src.shape[i]
 *         tmpslice.suboffsets[i] = -1
 */
  __pyx_t_3 = __pyx_v_ndim;
  __pyx_t_5 = __pyx_t_3;
  for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
    __pyx_v_i = __pyx_t_6;

    /* "View.MemoryView":1231
 *     tmpslice.memview = src.memview
 *     for i in range(ndim):
 *         tmpslice.shape[i] = src.shape[i]             # <<<<<<<<<<<<<<
 *         tmpslice.suboffsets[i] = -1
 * 
 */
    (__pyx_v_tmpslice->shape[__pyx_v_i]) = (__pyx_v_src->shape[__pyx_v_i]);

    /* "View.MemoryView":1232
 *     for i in range(ndim):
 *         tmpslice.shape[i] = src.shape[i]
 *         tmpslice.suboffsets[i] = -1             # <<<<<<<<<<<<<<
 * 
 *     fill_contig_strides_array(&tmpslice.shape[0], &tmpslice.strides[0], itemsize,
 */
    (__pyx_v_tmpslice->suboffsets[__pyx_v_i]) = -1L;
  }

  /* "View.MemoryView":1234
 *         tmpslice.suboffsets[i] = -1
 * 
 *     fill_contig_strides_array(&tmpslice.shape[0], &tmpslice.strides[0], itemsize,             # <<<<<<<<<<<<<<
 *                               ndim, order)
 * 
 */
  (void)(__pyx_fill_contig_strides_array((&(__pyx_v_tmpslice->shape[0])), (&(__pyx_v_tmpslice->strides[0])), __pyx_v_itemsize, __pyx_v_ndim, __pyx_v_order));

  /* "View.MemoryView":1238
 * 
 * 
 *     for i in range(ndim):             # <<<<<<<<<<<<<<
 *         if tmpslice.shape[i] == 1:
 *             tmpslice.strides[i] = 0
 */
  __pyx_t_3 = __pyx_v_ndim;
  __pyx_t_5 = __pyx_t_3;
  for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
    __pyx_v_i = __pyx_t_6;

    /* "View.MemoryView":1239
 * 
 *     for i in range(ndim):
 *         if tmpslice.shape[i] == 1:             # <<<<<<<<<<<<<<
 *             tmpslice.strides[i] = 0
 * 
 */
    __pyx_t_2 = (((__pyx_v_tmpslice->shape[__pyx_v_i]) == 1) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":1240
 *     for i in range(ndim):
 *         if tmpslice.shape[i] == 1:
 *             tmpslice.strides[i] = 0             # <<<<<<<<<<<<<<
 * 
 *     if slice_is_contig(src[0], order, ndim):
 */
      (__pyx_v_tmpslice->strides[__pyx_v_i]) = 0;

      /* "View.MemoryView":1239
 * 
 *     for i in range(ndim):
 *         if tmpslice.shape[i] == 1:             # <<<<<<<<<<<<<<
 *             tmpslice.strides[i] = 0
 * 
 */
    }
  }

  /* "View.MemoryView":1242
 *             tmpslice.strides[i] = 0
 * 
 *     if slice_is_contig(src[0], order, ndim):             # <<<<<<<<<<<<<<
 *         memcpy(result, src.data, size)
 *     else:
 */
  __pyx_t_2 = (__pyx_memviewslice_is_contig((__pyx_v_src[0]), __pyx_v_order, __pyx_v_ndim) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":1243
 * 
 *     if slice_is_contig(src[0], order, ndim):
 *         memcpy(result, src.data, size)             # <<<<<<<<<<<<<<
 *     else:
 *         copy_strided_to_strided(src, tmpslice, ndim, itemsize)
 */
    (void)(memcpy(__pyx_v_result, __pyx_v_src->data, __pyx_v_size));

    /* "View.MemoryView":1242
 *             tmpslice.strides[i] = 0
 * 
 *     if slice_is_contig(src[0], order, ndim):             # <<<<<<<<<<<<<<
 *         memcpy(result, src.data, size)
 *     else:
 */
    goto __pyx_L9;
  }

  /* "View.MemoryView":1245
 *         memcpy(result, src.data, size)
 *     else:
 *         copy_strided_to_strided(src, tmpslice, ndim, itemsize)             # <<<<<<<<<<<<<<
 * 
 *     return result
 */
  /*else*/ {
    copy_strided_to_strided(__pyx_v_src, __pyx_v_tmpslice, __pyx_v_ndim, __pyx_v_itemsize);
  }
  __pyx_L9:;

  /* "View.MemoryView":1247
 *         copy_strided_to_strided(src, tmpslice, ndim, itemsize)
 * 
 *     return result             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_result;
  goto __pyx_L0;

  /* "View.MemoryView":1209
 * 
 * @cname('__pyx_memoryview_copy_data_to_temp')
 * cdef void *copy_data_to_temp(__Pyx_memviewslice *src,             # <<<<<<<<<<<<<<
 *                              __Pyx_memviewslice *tmpslice,
 *                              char order,
 */

  /* function exit code */
  __pyx_L1_error:;
  {
    #ifdef WITH_THREAD
    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
    #endif
    __Pyx_AddTraceback("View.MemoryView.copy_data_to_temp", __pyx_clineno, __pyx_lineno, __pyx_filename);
    #ifdef WITH_THREAD
    __Pyx_PyGILState_Release(__pyx_gilstate_save);
    #endif
  }
  __pyx_r = NULL;
  __pyx_L0:;
  return __pyx_r;
}

/* "View.MemoryView":1252
 * 
 * @cname('__pyx_memoryview_err_extents')
 * cdef int _err_extents(int i, Py_ssize_t extent1,             # <<<<<<<<<<<<<<
 *                              Py_ssize_t extent2) except -1 with gil:
 *     raise ValueError("got differing extents in dimension %d (got %d and %d)" %
 */

static int __pyx_memoryview_err_extents(int __pyx_v_i, Py_ssize_t __pyx_v_extent1, Py_ssize_t __pyx_v_extent2) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  #endif
  __Pyx_RefNannySetupContext("_err_extents", 0);

  /* "View.MemoryView":1255
 *                              Py_ssize_t extent2) except -1 with gil:
 *     raise ValueError("got differing extents in dimension %d (got %d and %d)" %
 *                                                         (i, extent1, extent2))             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_err_dim')
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_i); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 1255, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyInt_FromSsize_t(__pyx_v_extent1); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 1255, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyInt_FromSsize_t(__pyx_v_extent2); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 1255, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 1255, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_t_3);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;

  /* "View.MemoryView":1254
 * cdef int _err_extents(int i, Py_ssize_t extent1,
 *                              Py_ssize_t extent2) except -1 with gil:
 *     raise ValueError("got differing extents in dimension %d (got %d and %d)" %             # <<<<<<<<<<<<<<
 *                                                         (i, extent1, extent2))
 * 
 */
  __pyx_t_3 = __Pyx_PyString_Format(__pyx_kp_s_got_differing_extents_in_dimensi, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 1254, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 1254, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_Raise(__pyx_t_4, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __PYX_ERR(2, 1254, __pyx_L1_error)

  /* "View.MemoryView":1252
 * 
 * @cname('__pyx_memoryview_err_extents')
 * cdef int _err_extents(int i, Py_ssize_t extent1,             # <<<<<<<<<<<<<<
 *                              Py_ssize_t extent2) except -1 with gil:
 *     raise ValueError("got differing extents in dimension %d (got %d and %d)" %
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("View.MemoryView._err_extents", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_RefNannyFinishContext();
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
  return __pyx_r;
}

/* "View.MemoryView":1258
 * 
 * @cname('__pyx_memoryview_err_dim')
 * cdef int _err_dim(object error, char *msg, int dim) except -1 with gil:             # <<<<<<<<<<<<<<
 *     raise error(msg.decode('ascii') % dim)
 * 
 */

static int __pyx_memoryview_err_dim(PyObject *__pyx_v_error, char *__pyx_v_msg, int __pyx_v_dim) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  #endif
  __Pyx_RefNannySetupContext("_err_dim", 0);
  __Pyx_INCREF(__pyx_v_error);

  /* "View.MemoryView":1259
 * @cname('__pyx_memoryview_err_dim')
 * cdef int _err_dim(object error, char *msg, int dim) except -1 with gil:
 *     raise error(msg.decode('ascii') % dim)             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_err')
 */
  __pyx_t_2 = __Pyx_decode_c_string(__pyx_v_msg, 0, strlen(__pyx_v_msg), NULL, NULL, PyUnicode_DecodeASCII); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 1259, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_dim); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 1259, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyUnicode_Format(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 1259, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_INCREF(__pyx_v_error);
  __pyx_t_3 = __pyx_v_error; __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 1259, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(2, 1259, __pyx_L1_error)

  /* "View.MemoryView":1258
 * 
 * @cname('__pyx_memoryview_err_dim')
 * cdef int _err_dim(object error, char *msg, int dim) except -1 with gil:             # <<<<<<<<<<<<<<
 *     raise error(msg.decode('ascii') % dim)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("View.MemoryView._err_dim", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_XDECREF(__pyx_v_error);
  __Pyx_RefNannyFinishContext();
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
  return __pyx_r;
}

/* "View.MemoryView":1262
 * 
 * @cname('__pyx_memoryview_err')
 * cdef int _err(object error, char *msg) except -1 with gil:             # <<<<<<<<<<<<<<
 *     if msg != NULL:
 *         raise error(msg.decode('ascii'))
 */

static int __pyx_memoryview_err(PyObject *__pyx_v_error, char *__pyx_v_msg) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  #endif
  __Pyx_RefNannySetupContext("_err", 0);
  __Pyx_INCREF(__pyx_v_error);

  /* "View.MemoryView":1263
 * @cname('__pyx_memoryview_err')
 * cdef int _err(object error, char *msg) except -1 with gil:
 *     if msg != NULL:             # <<<<<<<<<<<<<<
 *         raise error(msg.decode('ascii'))
 *     else:
 */
  __pyx_t_1 = ((__pyx_v_msg != NULL) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "View.MemoryView":1264
 * cdef int _err(object error, char *msg) except -1 with gil:
 *     if msg != NULL:
 *         raise error(msg.decode('ascii'))             # <<<<<<<<<<<<<<
 *     else:
 *         raise error
 */
    __pyx_t_3 = __Pyx_decode_c_string(__pyx_v_msg, 0, strlen(__pyx_v_msg), NULL, NULL, PyUnicode_DecodeASCII); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 1264, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_v_error);
    __pyx_t_4 = __pyx_v_error; __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 1264, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(2, 1264, __pyx_L1_error)

    /* "View.MemoryView":1263
 * @cname('__pyx_memoryview_err')
 * cdef int _err(object error, char *msg) except -1 with gil:
 *     if msg != NULL:             # <<<<<<<<<<<<<<
 *         raise error(msg.decode('ascii'))
 *     else:
 */
  }

  /* "View.MemoryView":1266
 *         raise error(msg.decode('ascii'))
 *     else:
 *         raise error             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_copy_contents')
 */
  /*else*/ {
    __Pyx_Raise(__pyx_v_error, 0, 0, 0);
    __PYX_ERR(2, 1266, __pyx_L1_error)
  }

  /* "View.MemoryView":1262
 * 
 * @cname('__pyx_memoryview_err')
 * cdef int _err(object error, char *msg) except -1 with gil:             # <<<<<<<<<<<<<<
 *     if msg != NULL:
 *         raise error(msg.decode('ascii'))
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("View.MemoryView._err", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_XDECREF(__pyx_v_error);
  __Pyx_RefNannyFinishContext();
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
  return __pyx_r;
}

/* "View.MemoryView":1269
 * 
 * @cname('__pyx_memoryview_copy_contents')
 * cdef int memoryview_copy_contents(__Pyx_memviewslice src,             # <<<<<<<<<<<<<<
 *                                   __Pyx_memviewslice dst,
 *                                   int src_ndim, int dst_ndim,
 */

static int __pyx_memoryview_copy_contents(__Pyx_memviewslice __pyx_v_src, __Pyx_memviewslice __pyx_v_dst, int __pyx_v_src_ndim, int __pyx_v_dst_ndim, int __pyx_v_dtype_is_object) {
  void *__pyx_v_tmpdata;
  size_t __pyx_v_itemsize;
  int __pyx_v_i;
  char __pyx_v_order;
  int __pyx_v_broadcasting;
  int __pyx_v_direct_copy;
  __Pyx_memviewslice __pyx_v_tmp;
  int __pyx_v_ndim;
  int __pyx_r;
  Py_ssize_t __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  void *__pyx_t_7;
  int __pyx_t_8;

  /* "View.MemoryView":1277
 *     Check for overlapping memory and verify the shapes.
 *     """
 *     cdef void *tmpdata = NULL             # <<<<<<<<<<<<<<
 *     cdef size_t itemsize = src.memview.view.itemsize
 *     cdef int i
 */
  __pyx_v_tmpdata = NULL;

  /* "View.MemoryView":1278
 *     """
 *     cdef void *tmpdata = NULL
 *     cdef size_t itemsize = src.memview.view.itemsize             # <<<<<<<<<<<<<<
 *     cdef int i
 *     cdef char order = get_best_order(&src, src_ndim)
 */
  __pyx_t_1 = __pyx_v_src.memview->view.itemsize;
  __pyx_v_itemsize = __pyx_t_1;

  /* "View.MemoryView":1280
 *     cdef size_t itemsize = src.memview.view.itemsize
 *     cdef int i
 *     cdef char order = get_best_order(&src, src_ndim)             # <<<<<<<<<<<<<<
 *     cdef bint broadcasting = False
 *     cdef bint direct_copy = False
 */
  __pyx_v_order = __pyx_get_best_slice_order((&__pyx_v_src), __pyx_v_src_ndim);

  /* "View.MemoryView":1281
 *     cdef int i
 *     cdef char order = get_best_order(&src, src_ndim)
 *     cdef bint broadcasting = False             # <<<<<<<<<<<<<<
 *     cdef bint direct_copy = False
 *     cdef __Pyx_memviewslice tmp
 */
  __pyx_v_broadcasting = 0;

  /* "View.MemoryView":1282
 *     cdef char order = get_best_order(&src, src_ndim)
 *     cdef bint broadcasting = False
 *     cdef bint direct_copy = False             # <<<<<<<<<<<<<<
 *     cdef __Pyx_memviewslice tmp
 * 
 */
  __pyx_v_direct_copy = 0;

  /* "View.MemoryView":1285
 *     cdef __Pyx_memviewslice tmp
 * 
 *     if src_ndim < dst_ndim:             # <<<<<<<<<<<<<<
 *         broadcast_leading(&src, src_ndim, dst_ndim)
 *     elif dst_ndim < src_ndim:
 */
  __pyx_t_2 = ((__pyx_v_src_ndim < __pyx_v_dst_ndim) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":1286
 * 
 *     if src_ndim < dst_ndim:
 *         broadcast_leading(&src, src_ndim, dst_ndim)             # <<<<<<<<<<<<<<
 *     elif dst_ndim < src_ndim:
 *         broadcast_leading(&dst, dst_ndim, src_ndim)
 */
    __pyx_memoryview_broadcast_leading((&__pyx_v_src), __pyx_v_src_ndim, __pyx_v_dst_ndim);

    /* "View.MemoryView":1285
 *     cdef __Pyx_memviewslice tmp
 * 
 *     if src_ndim < dst_ndim:             # <<<<<<<<<<<<<<
 *         broadcast_leading(&src, src_ndim, dst_ndim)
 *     elif dst_ndim < src_ndim:
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":1287
 *     if src_ndim < dst_ndim:
 *         broadcast_leading(&src, src_ndim, dst_ndim)
 *     elif dst_ndim < src_ndim:             # <<<<<<<<<<<<<<
 *         broadcast_leading(&dst, dst_ndim, src_ndim)
 * 
 */
  __pyx_t_2 = ((__pyx_v_dst_ndim < __pyx_v_src_ndim) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":1288
 *         broadcast_leading(&src, src_ndim, dst_ndim)
 *     elif dst_ndim < src_ndim:
 *         broadcast_leading(&dst, dst_ndim, src_ndim)             # <<<<<<<<<<<<<<
 * 
 *     cdef int ndim = max(src_ndim, dst_ndim)
 */
    __pyx_memoryview_broadcast_leading((&__pyx_v_dst), __pyx_v_dst_ndim, __pyx_v_src_ndim);

    /* "View.MemoryView":1287
 *     if src_ndim < dst_ndim:
 *         broadcast_leading(&src, src_ndim, dst_ndim)
 *     elif dst_ndim < src_ndim:             # <<<<<<<<<<<<<<
 *         broadcast_leading(&dst, dst_ndim, src_ndim)
 * 
 */
  }
  __pyx_L3:;

  /* "View.MemoryView":1290
 *         broadcast_leading(&dst, dst_ndim, src_ndim)
 * 
 *     cdef int ndim = max(src_ndim, dst_ndim)             # <<<<<<<<<<<<<<
 * 
 *     for i in range(ndim):
 */
  __pyx_t_3 = __pyx_v_dst_ndim;
  __pyx_t_4 = __pyx_v_src_ndim;
  if (((__pyx_t_3 > __pyx_t_4) != 0)) {
    __pyx_t_5 = __pyx_t_3;
  } else {
    __pyx_t_5 = __pyx_t_4;
  }
  __pyx_v_ndim = __pyx_t_5;

  /* "View.MemoryView":1292
 *     cdef int ndim = max(src_ndim, dst_ndim)
 * 
 *     for i in range(ndim):             # <<<<<<<<<<<<<<
 *         if src.shape[i] != dst.shape[i]:
 *             if src.shape[i] == 1:
 */
  __pyx_t_5 = __pyx_v_ndim;
  __pyx_t_3 = __pyx_t_5;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "View.MemoryView":1293
 * 
 *     for i in range(ndim):
 *         if src.shape[i] != dst.shape[i]:             # <<<<<<<<<<<<<<
 *             if src.shape[i] == 1:
 *                 broadcasting = True
 */
    __pyx_t_2 = (((__pyx_v_src.shape[__pyx_v_i]) != (__pyx_v_dst.shape[__pyx_v_i])) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":1294
 *     for i in range(ndim):
 *         if src.shape[i] != dst.shape[i]:
 *             if src.shape[i] == 1:             # <<<<<<<<<<<<<<
 *                 broadcasting = True
 *                 src.strides[i] = 0
 */
      __pyx_t_2 = (((__pyx_v_src.shape[__pyx_v_i]) == 1) != 0);
      if (__pyx_t_2) {

        /* "View.MemoryView":1295
 *         if src.shape[i] != dst.shape[i]:
 *             if src.shape[i] == 1:
 *                 broadcasting = True             # <<<<<<<<<<<<<<
 *                 src.strides[i] = 0
 *             else:
 */
        __pyx_v_broadcasting = 1;

        /* "View.MemoryView":1296
 *             if src.shape[i] == 1:
 *                 broadcasting = True
 *                 src.strides[i] = 0             # <<<<<<<<<<<<<<
 *             else:
 *                 _err_extents(i, dst.shape[i], src.shape[i])
 */
        (__pyx_v_src.strides[__pyx_v_i]) = 0;

        /* "View.MemoryView":1294
 *     for i in range(ndim):
 *         if src.shape[i] != dst.shape[i]:
 *             if src.shape[i] == 1:             # <<<<<<<<<<<<<<
 *                 broadcasting = True
 *                 src.strides[i] = 0
 */
        goto __pyx_L7;
      }

      /* "View.MemoryView":1298
 *                 src.strides[i] = 0
 *             else:
 *                 _err_extents(i, dst.shape[i], src.shape[i])             # <<<<<<<<<<<<<<
 * 
 *         if src.suboffsets[i] >= 0:
 */
      /*else*/ {
        __pyx_t_6 = __pyx_memoryview_err_extents(__pyx_v_i, (__pyx_v_dst.shape[__pyx_v_i]), (__pyx_v_src.shape[__pyx_v_i])); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(2, 1298, __pyx_L1_error)
      }
      __pyx_L7:;

      /* "View.MemoryView":1293
 * 
 *     for i in range(ndim):
 *         if src.shape[i] != dst.shape[i]:             # <<<<<<<<<<<<<<
 *             if src.shape[i] == 1:
 *                 broadcasting = True
 */
    }

    /* "View.MemoryView":1300
 *                 _err_extents(i, dst.shape[i], src.shape[i])
 * 
 *         if src.suboffsets[i] >= 0:             # <<<<<<<<<<<<<<
 *             _err_dim(ValueError, "Dimension %d is not direct", i)
 * 
 */
    __pyx_t_2 = (((__pyx_v_src.suboffsets[__pyx_v_i]) >= 0) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":1301
 * 
 *         if src.suboffsets[i] >= 0:
 *             _err_dim(ValueError, "Dimension %d is not direct", i)             # <<<<<<<<<<<<<<
 * 
 *     if slices_overlap(&src, &dst, ndim, itemsize):
 */
      __pyx_t_6 = __pyx_memoryview_err_dim(__pyx_builtin_ValueError, ((char *)"Dimension %d is not direct"), __pyx_v_i); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(2, 1301, __pyx_L1_error)

      /* "View.MemoryView":1300
 *                 _err_extents(i, dst.shape[i], src.shape[i])
 * 
 *         if src.suboffsets[i] >= 0:             # <<<<<<<<<<<<<<
 *             _err_dim(ValueError, "Dimension %d is not direct", i)
 * 
 */
    }
  }

  /* "View.MemoryView":1303
 *             _err_dim(ValueError, "Dimension %d is not direct", i)
 * 
 *     if slices_overlap(&src, &dst, ndim, itemsize):             # <<<<<<<<<<<<<<
 * 
 *         if not slice_is_contig(src, order, ndim):
 */
  __pyx_t_2 = (__pyx_slices_overlap((&__pyx_v_src), (&__pyx_v_dst), __pyx_v_ndim, __pyx_v_itemsize) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":1305
 *     if slices_overlap(&src, &dst, ndim, itemsize):
 * 
 *         if not slice_is_contig(src, order, ndim):             # <<<<<<<<<<<<<<
 *             order = get_best_order(&dst, ndim)
 * 
 */
    __pyx_t_2 = ((!(__pyx_memviewslice_is_contig(__pyx_v_src, __pyx_v_order, __pyx_v_ndim) != 0)) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":1306
 * 
 *         if not slice_is_contig(src, order, ndim):
 *             order = get_best_order(&dst, ndim)             # <<<<<<<<<<<<<<
 * 
 *         tmpdata = copy_data_to_temp(&src, &tmp, order, ndim)
 */
      __pyx_v_order = __pyx_get_best_slice_order((&__pyx_v_dst), __pyx_v_ndim);

      /* "View.MemoryView":1305
 *     if slices_overlap(&src, &dst, ndim, itemsize):
 * 
 *         if not slice_is_contig(src, order, ndim):             # <<<<<<<<<<<<<<
 *             order = get_best_order(&dst, ndim)
 * 
 */
    }

    /* "View.MemoryView":1308
 *             order = get_best_order(&dst, ndim)
 * 
 *         tmpdata = copy_data_to_temp(&src, &tmp, order, ndim)             # <<<<<<<<<<<<<<
 *         src = tmp
 * 
 */
    __pyx_t_7 = __pyx_memoryview_copy_data_to_temp((&__pyx_v_src), (&__pyx_v_tmp), __pyx_v_order, __pyx_v_ndim); if (unlikely(__pyx_t_7 == ((void *)NULL))) __PYX_ERR(2, 1308, __pyx_L1_error)
    __pyx_v_tmpdata = __pyx_t_7;

    /* "View.MemoryView":1309
 * 
 *         tmpdata = copy_data_to_temp(&src, &tmp, order, ndim)
 *         src = tmp             # <<<<<<<<<<<<<<
 * 
 *     if not broadcasting:
 */
    __pyx_v_src = __pyx_v_tmp;

    /* "View.MemoryView":1303
 *             _err_dim(ValueError, "Dimension %d is not direct", i)
 * 
 *     if slices_overlap(&src, &dst, ndim, itemsize):             # <<<<<<<<<<<<<<
 * 
 *         if not slice_is_contig(src, order, ndim):
 */
  }

  /* "View.MemoryView":1311
 *         src = tmp
 * 
 *     if not broadcasting:             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = ((!(__pyx_v_broadcasting != 0)) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":1314
 * 
 * 
 *         if slice_is_contig(src, 'C', ndim):             # <<<<<<<<<<<<<<
 *             direct_copy = slice_is_contig(dst, 'C', ndim)
 *         elif slice_is_contig(src, 'F', ndim):
 */
    __pyx_t_2 = (__pyx_memviewslice_is_contig(__pyx_v_src, 'C', __pyx_v_ndim) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":1315
 * 
 *         if slice_is_contig(src, 'C', ndim):
 *             direct_copy = slice_is_contig(dst, 'C', ndim)             # <<<<<<<<<<<<<<
 *         elif slice_is_contig(src, 'F', ndim):
 *             direct_copy = slice_is_contig(dst, 'F', ndim)
 */
      __pyx_v_direct_copy = __pyx_memviewslice_is_contig(__pyx_v_dst, 'C', __pyx_v_ndim);

      /* "View.MemoryView":1314
 * 
 * 
 *         if slice_is_contig(src, 'C', ndim):             # <<<<<<<<<<<<<<
 *             direct_copy = slice_is_contig(dst, 'C', ndim)
 *         elif slice_is_contig(src, 'F', ndim):
 */
      goto __pyx_L12;
    }

    /* "View.MemoryView":1316
 *         if slice_is_contig(src, 'C', ndim):
 *             direct_copy = slice_is_contig(dst, 'C', ndim)
 *         elif slice_is_contig(src, 'F', ndim):             # <<<<<<<<<<<<<<
 *             direct_copy = slice_is_contig(dst, 'F', ndim)
 * 
 */
    __pyx_t_2 = (__pyx_memviewslice_is_contig(__pyx_v_src, 'F', __pyx_v_ndim) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":1317
 *             direct_copy = slice_is_contig(dst, 'C', ndim)
 *         elif slice_is_contig(src, 'F', ndim):
 *             direct_copy = slice_is_contig(dst, 'F', ndim)             # <<<<<<<<<<<<<<
 * 
 *         if direct_copy:
 */
      __pyx_v_direct_copy = __pyx_memviewslice_is_contig(__pyx_v_dst, 'F', __pyx_v_ndim);

      /* "View.MemoryView":1316
 *         if slice_is_contig(src, 'C', ndim):
 *             direct_copy = slice_is_contig(dst, 'C', ndim)
 *         elif slice_is_contig(src, 'F', ndim):             # <<<<<<<<<<<<<<
 *             direct_copy = slice_is_contig(dst, 'F', ndim)
 * 
 */
    }
    __pyx_L12:;

    /* "View.MemoryView":1319
 *             direct_copy = slice_is_contig(dst, 'F', ndim)
 * 
 *         if direct_copy:             # <<<<<<<<<<<<<<
 * 
 *             refcount_copying(&dst, dtype_is_object, ndim, False)
 */
    __pyx_t_2 = (__pyx_v_direct_copy != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":1321
 *         if direct_copy:
 * 
 *             refcount_copying(&dst, dtype_is_object, ndim, False)             # <<<<<<<<<<<<<<
 *             memcpy(dst.data, src.data, slice_get_size(&src, ndim))
 *             refcount_copying(&dst, dtype_is_object, ndim, True)
 */
      __pyx_memoryview_refcount_copying((&__pyx_v_dst), __pyx_v_dtype_is_object, __pyx_v_ndim, 0);

      /* "View.MemoryView":1322
 * 
 *             refcount_copying(&dst, dtype_is_object, ndim, False)
 *             memcpy(dst.data, src.data, slice_get_size(&src, ndim))             # <<<<<<<<<<<<<<
 *             refcount_copying(&dst, dtype_is_object, ndim, True)
 *             free(tmpdata)
 */
      (void)(memcpy(__pyx_v_dst.data, __pyx_v_src.data, __pyx_memoryview_slice_get_size((&__pyx_v_src), __pyx_v_ndim)));

      /* "View.MemoryView":1323
 *             refcount_copying(&dst, dtype_is_object, ndim, False)
 *             memcpy(dst.data, src.data, slice_get_size(&src, ndim))
 *             refcount_copying(&dst, dtype_is_object, ndim, True)             # <<<<<<<<<<<<<<
 *             free(tmpdata)
 *             return 0
 */
      __pyx_memoryview_refcount_copying((&__pyx_v_dst), __pyx_v_dtype_is_object, __pyx_v_ndim, 1);

      /* "View.MemoryView":1324
 *             memcpy(dst.data, src.data, slice_get_size(&src, ndim))
 *             refcount_copying(&dst, dtype_is_object, ndim, True)
 *             free(tmpdata)             # <<<<<<<<<<<<<<
 *             return 0
 * 
 */
      free(__pyx_v_tmpdata);

      /* "View.MemoryView":1325
 *             refcount_copying(&dst, dtype_is_object, ndim, True)
 *             free(tmpdata)
 *             return 0             # <<<<<<<<<<<<<<
 * 
 *     if order == 'F' == get_best_order(&dst, ndim):
 */
      __pyx_r = 0;
      goto __pyx_L0;

      /* "View.MemoryView":1319
 *             direct_copy = slice_is_contig(dst, 'F', ndim)
 * 
 *         if direct_copy:             # <<<<<<<<<<<<<<
 * 
 *             refcount_copying(&dst, dtype_is_object, ndim, False)
 */
    }

    /* "View.MemoryView":1311
 *         src = tmp
 * 
 *     if not broadcasting:             # <<<<<<<<<<<<<<
 * 
 * 
 */
  }

  /* "View.MemoryView":1327
 *             return 0
 * 
 *     if order == 'F' == get_best_order(&dst, ndim):             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = (__pyx_v_order == 'F');
  if (__pyx_t_2) {
    __pyx_t_2 = ('F' == __pyx_get_best_slice_order((&__pyx_v_dst), __pyx_v_ndim));
  }
  __pyx_t_8 = (__pyx_t_2 != 0);
  if (__pyx_t_8) {

    /* "View.MemoryView":1330
 * 
 * 
 *         transpose_memslice(&src)             # <<<<<<<<<<<<<<
 *         transpose_memslice(&dst)
 * 
 */
    __pyx_t_5 = __pyx_memslice_transpose((&__pyx_v_src)); if (unlikely(__pyx_t_5 == ((int)0))) __PYX_ERR(2, 1330, __pyx_L1_error)

    /* "View.MemoryView":1331
 * 
 *         transpose_memslice(&src)
 *         transpose_memslice(&dst)             # <<<<<<<<<<<<<<
 * 
 *     refcount_copying(&dst, dtype_is_object, ndim, False)
 */
    __pyx_t_5 = __pyx_memslice_transpose((&__pyx_v_dst)); if (unlikely(__pyx_t_5 == ((int)0))) __PYX_ERR(2, 1331, __pyx_L1_error)

    /* "View.MemoryView":1327
 *             return 0
 * 
 *     if order == 'F' == get_best_order(&dst, ndim):             # <<<<<<<<<<<<<<
 * 
 * 
 */
  }

  /* "View.MemoryView":1333
 *         transpose_memslice(&dst)
 * 
 *     refcount_copying(&dst, dtype_is_object, ndim, False)             # <<<<<<<<<<<<<<
 *     copy_strided_to_strided(&src, &dst, ndim, itemsize)
 *     refcount_copying(&dst, dtype_is_object, ndim, True)
 */
  __pyx_memoryview_refcount_copying((&__pyx_v_dst), __pyx_v_dtype_is_object, __pyx_v_ndim, 0);

  /* "View.MemoryView":1334
 * 
 *     refcount_copying(&dst, dtype_is_object, ndim, False)
 *     copy_strided_to_strided(&src, &dst, ndim, itemsize)             # <<<<<<<<<<<<<<
 *     refcount_copying(&dst, dtype_is_object, ndim, True)
 * 
 */
  copy_strided_to_strided((&__pyx_v_src), (&__pyx_v_dst), __pyx_v_ndim, __pyx_v_itemsize);

  /* "View.MemoryView":1335
 *     refcount_copying(&dst, dtype_is_object, ndim, False)
 *     copy_strided_to_strided(&src, &dst, ndim, itemsize)
 *     refcount_copying(&dst, dtype_is_object, ndim, True)             # <<<<<<<<<<<<<<
 * 
 *     free(tmpdata)
 */
  __pyx_memoryview_refcount_copying((&__pyx_v_dst), __pyx_v_dtype_is_object, __pyx_v_ndim, 1);

  /* "View.MemoryView":1337
 *     refcount_copying(&dst, dtype_is_object, ndim, True)
 * 
 *     free(tmpdata)             # <<<<<<<<<<<<<<
 *     return 0
 * 
 */
  free(__pyx_v_tmpdata);

  /* "View.MemoryView":1338
 * 
 *     free(tmpdata)
 *     return 0             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_broadcast_leading')
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "View.MemoryView":1269
 * 
 * @cname('__pyx_memoryview_copy_contents')
 * cdef int memoryview_copy_contents(__Pyx_memviewslice src,             # <<<<<<<<<<<<<<
 *                                   __Pyx_memviewslice dst,
 *                                   int src_ndim, int dst_ndim,
 */

  /* function exit code */
  __pyx_L1_error:;
  {
    #ifdef WITH_THREAD
    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
    #endif
    __Pyx_AddTraceback("View.MemoryView.memoryview_copy_contents", __pyx_clineno, __pyx_lineno, __pyx_filename);
    #ifdef WITH_THREAD
    __Pyx_PyGILState_Release(__pyx_gilstate_save);
    #endif
  }
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "View.MemoryView":1341
 * 
 * @cname('__pyx_memoryview_broadcast_leading')
 * cdef void broadcast_leading(__Pyx_memviewslice *mslice,             # <<<<<<<<<<<<<<
 *                             int ndim,
 *                             int ndim_other) nogil:
 */

static void __pyx_memoryview_broadcast_leading(__Pyx_memviewslice *__pyx_v_mslice, int __pyx_v_ndim, int __pyx_v_ndim_other) {
  int __pyx_v_i;
  int __pyx_v_offset;
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;

  /* "View.MemoryView":1345
 *                             int ndim_other) nogil:
 *     cdef int i
 *     cdef int offset = ndim_other - ndim             # <<<<<<<<<<<<<<
 * 
 *     for i in range(ndim - 1, -1, -1):
 */
  __pyx_v_offset = (__pyx_v_ndim_other - __pyx_v_ndim);

  /* "View.MemoryView":1347
 *     cdef int offset = ndim_other - ndim
 * 
 *     for i in range(ndim - 1, -1, -1):             # <<<<<<<<<<<<<<
 *         mslice.shape[i + offset] = mslice.shape[i]
 *         mslice.strides[i + offset] = mslice.strides[i]
 */
  for (__pyx_t_1 = (__pyx_v_ndim - 1); __pyx_t_1 > -1; __pyx_t_1-=1) {
    __pyx_v_i = __pyx_t_1;

    /* "View.MemoryView":1348
 * 
 *     for i in range(ndim - 1, -1, -1):
 *         mslice.shape[i + offset] = mslice.shape[i]             # <<<<<<<<<<<<<<
 *         mslice.strides[i + offset] = mslice.strides[i]
 *         mslice.suboffsets[i + offset] = mslice.suboffsets[i]
 */
    (__pyx_v_mslice->shape[(__pyx_v_i + __pyx_v_offset)]) = (__pyx_v_mslice->shape[__pyx_v_i]);

    /* "View.MemoryView":1349
 *     for i in range(ndim - 1, -1, -1):
 *         mslice.shape[i + offset] = mslice.shape[i]
 *         mslice.strides[i + offset] = mslice.strides[i]             # <<<<<<<<<<<<<<
 *         mslice.suboffsets[i + offset] = mslice.suboffsets[i]
 * 
 */
    (__pyx_v_mslice->strides[(__pyx_v_i + __pyx_v_offset)]) = (__pyx_v_mslice->strides[__pyx_v_i]);

    /* "View.MemoryView":1350
 *         mslice.shape[i + offset] = mslice.shape[i]
 *         mslice.strides[i + offset] = mslice.strides[i]
 *         mslice.suboffsets[i + offset] = mslice.suboffsets[i]             # <<<<<<<<<<<<<<
 * 
 *     for i in range(offset):
 */
    (__pyx_v_mslice->suboffsets[(__pyx_v_i + __pyx_v_offset)]) = (__pyx_v_mslice->suboffsets[__pyx_v_i]);
  }

  /* "View.MemoryView":1352
 *         mslice.suboffsets[i + offset] = mslice.suboffsets[i]
 * 
 *     for i in range(offset):             # <<<<<<<<<<<<<<
 *         mslice.shape[i] = 1
 *         mslice.strides[i] = mslice.strides[0]
 */
  __pyx_t_1 = __pyx_v_offset;
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "View.MemoryView":1353
 * 
 *     for i in range(offset):
 *         mslice.shape[i] = 1             # <<<<<<<<<<<<<<
 *         mslice.strides[i] = mslice.strides[0]
 *         mslice.suboffsets[i] = -1
 */
    (__pyx_v_mslice->shape[__pyx_v_i]) = 1;

    /* "View.MemoryView":1354
 *     for i in range(offset):
 *         mslice.shape[i] = 1
 *         mslice.strides[i] = mslice.strides[0]             # <<<<<<<<<<<<<<
 *         mslice.suboffsets[i] = -1
 * 
 */
    (__pyx_v_mslice->strides[__pyx_v_i]) = (__pyx_v_mslice->strides[0]);

    /* "View.MemoryView":1355
 *         mslice.shape[i] = 1
 *         mslice.strides[i] = mslice.strides[0]
 *         mslice.suboffsets[i] = -1             # <<<<<<<<<<<<<<
 * 
 * 
 */
    (__pyx_v_mslice->suboffsets[__pyx_v_i]) = -1L;
  }

  /* "View.MemoryView":1341
 * 
 * @cname('__pyx_memoryview_broadcast_leading')
 * cdef void broadcast_leading(__Pyx_memviewslice *mslice,             # <<<<<<<<<<<<<<
 *                             int ndim,
 *                             int ndim_other) nogil:
 */

  /* function exit code */
}

/* "View.MemoryView":1363
 * 
 * @cname('__pyx_memoryview_refcount_copying')
 * cdef void refcount_copying(__Pyx_memviewslice *dst, bint dtype_is_object,             # <<<<<<<<<<<<<<
 *                            int ndim, bint inc) nogil:
 * 
 */

static void __pyx_memoryview_refcount_copying(__Pyx_memviewslice *__pyx_v_dst, int __pyx_v_dtype_is_object, int __pyx_v_ndim, int __pyx_v_inc) {
  int __pyx_t_1;

  /* "View.MemoryView":1367
 * 
 * 
 *     if dtype_is_object:             # <<<<<<<<<<<<<<
 *         refcount_objects_in_slice_with_gil(dst.data, dst.shape,
 *                                            dst.strides, ndim, inc)
 */
  __pyx_t_1 = (__pyx_v_dtype_is_object != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":1368
 * 
 *     if dtype_is_object:
 *         refcount_objects_in_slice_with_gil(dst.data, dst.shape,             # <<<<<<<<<<<<<<
 *                                            dst.strides, ndim, inc)
 * 
 */
    __pyx_memoryview_refcount_objects_in_slice_with_gil(__pyx_v_dst->data, __pyx_v_dst->shape, __pyx_v_dst->strides, __pyx_v_ndim, __pyx_v_inc);

    /* "View.MemoryView":1367
 * 
 * 
 *     if dtype_is_object:             # <<<<<<<<<<<<<<
 *         refcount_objects_in_slice_with_gil(dst.data, dst.shape,
 *                                            dst.strides, ndim, inc)
 */
  }

  /* "View.MemoryView":1363
 * 
 * @cname('__pyx_memoryview_refcount_copying')
 * cdef void refcount_copying(__Pyx_memviewslice *dst, bint dtype_is_object,             # <<<<<<<<<<<<<<
 *                            int ndim, bint inc) nogil:
 * 
 */

  /* function exit code */
}

/* "View.MemoryView":1372
 * 
 * @cname('__pyx_memoryview_refcount_objects_in_slice_with_gil')
 * cdef void refcount_objects_in_slice_with_gil(char *data, Py_ssize_t *shape,             # <<<<<<<<<<<<<<
 *                                              Py_ssize_t *strides, int ndim,
 *                                              bint inc) with gil:
 */

static void __pyx_memoryview_refcount_objects_in_slice_with_gil(char *__pyx_v_data, Py_ssize_t *__pyx_v_shape, Py_ssize_t *__pyx_v_strides, int __pyx_v_ndim, int __pyx_v_inc) {
  __Pyx_RefNannyDeclarations
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  #endif
  __Pyx_RefNannySetupContext("refcount_objects_in_slice_with_gil", 0);

  /* "View.MemoryView":1375
 *                                              Py_ssize_t *strides, int ndim,
 *                                              bint inc) with gil:
 *     refcount_objects_in_slice(data, shape, strides, ndim, inc)             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_refcount_objects_in_slice')
 */
  __pyx_memoryview_refcount_objects_in_slice(__pyx_v_data, __pyx_v_shape, __pyx_v_strides, __pyx_v_ndim, __pyx_v_inc);

  /* "View.MemoryView":1372
 * 
 * @cname('__pyx_memoryview_refcount_objects_in_slice_with_gil')
 * cdef void refcount_objects_in_slice_with_gil(char *data, Py_ssize_t *shape,             # <<<<<<<<<<<<<<
 *                                              Py_ssize_t *strides, int ndim,
 *                                              bint inc) with gil:
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
}

/* "View.MemoryView":1378
 * 
 * @cname('__pyx_memoryview_refcount_objects_in_slice')
 * cdef void refcount_objects_in_slice(char *data, Py_ssize_t *shape,             # <<<<<<<<<<<<<<
 *                                     Py_ssize_t *strides, int ndim, bint inc):
 *     cdef Py_ssize_t i
 */

static void __pyx_memoryview_refcount_objects_in_slice(char *__pyx_v_data, Py_ssize_t *__pyx_v_shape, Py_ssize_t *__pyx_v_strides, int __pyx_v_ndim, int __pyx_v_inc) {
  CYTHON_UNUSED Py_ssize_t __pyx_v_i;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  __Pyx_RefNannySetupContext("refcount_objects_in_slice", 0);

  /* "View.MemoryView":1382
 *     cdef Py_ssize_t i
 * 
 *     for i in range(shape[0]):             # <<<<<<<<<<<<<<
 *         if ndim == 1:
 *             if inc:
 */
  __pyx_t_1 = (__pyx_v_shape[0]);
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "View.MemoryView":1383
 * 
 *     for i in range(shape[0]):
 *         if ndim == 1:             # <<<<<<<<<<<<<<
 *             if inc:
 *                 Py_INCREF((<PyObject **> data)[0])
 */
    __pyx_t_4 = ((__pyx_v_ndim == 1) != 0);
    if (__pyx_t_4) {

      /* "View.MemoryView":1384
 *     for i in range(shape[0]):
 *         if ndim == 1:
 *             if inc:             # <<<<<<<<<<<<<<
 *                 Py_INCREF((<PyObject **> data)[0])
 *             else:
 */
      __pyx_t_4 = (__pyx_v_inc != 0);
      if (__pyx_t_4) {

        /* "View.MemoryView":1385
 *         if ndim == 1:
 *             if inc:
 *                 Py_INCREF((<PyObject **> data)[0])             # <<<<<<<<<<<<<<
 *             else:
 *                 Py_DECREF((<PyObject **> data)[0])
 */
        Py_INCREF((((PyObject **)__pyx_v_data)[0]));

        /* "View.MemoryView":1384
 *     for i in range(shape[0]):
 *         if ndim == 1:
 *             if inc:             # <<<<<<<<<<<<<<
 *                 Py_INCREF((<PyObject **> data)[0])
 *             else:
 */
        goto __pyx_L6;
      }

      /* "View.MemoryView":1387
 *                 Py_INCREF((<PyObject **> data)[0])
 *             else:
 *                 Py_DECREF((<PyObject **> data)[0])             # <<<<<<<<<<<<<<
 *         else:
 *             refcount_objects_in_slice(data, shape + 1, strides + 1,
 */
      /*else*/ {
        Py_DECREF((((PyObject **)__pyx_v_data)[0]));
      }
      __pyx_L6:;

      /* "View.MemoryView":1383
 * 
 *     for i in range(shape[0]):
 *         if ndim == 1:             # <<<<<<<<<<<<<<
 *             if inc:
 *                 Py_INCREF((<PyObject **> data)[0])
 */
      goto __pyx_L5;
    }

    /* "View.MemoryView":1389
 *                 Py_DECREF((<PyObject **> data)[0])
 *         else:
 *             refcount_objects_in_slice(data, shape + 1, strides + 1,             # <<<<<<<<<<<<<<
 *                                       ndim - 1, inc)
 * 
 */
    /*else*/ {

      /* "View.MemoryView":1390
 *         else:
 *             refcount_objects_in_slice(data, shape + 1, strides + 1,
 *                                       ndim - 1, inc)             # <<<<<<<<<<<<<<
 * 
 *         data += strides[0]
 */
      __pyx_memoryview_refcount_objects_in_slice(__pyx_v_data, (__pyx_v_shape + 1), (__pyx_v_strides + 1), (__pyx_v_ndim - 1), __pyx_v_inc);
    }
    __pyx_L5:;

    /* "View.MemoryView":1392
 *                                       ndim - 1, inc)
 * 
 *         data += strides[0]             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_v_data = (__pyx_v_data + (__pyx_v_strides[0]));
  }

  /* "View.MemoryView":1378
 * 
 * @cname('__pyx_memoryview_refcount_objects_in_slice')
 * cdef void refcount_objects_in_slice(char *data, Py_ssize_t *shape,             # <<<<<<<<<<<<<<
 *                                     Py_ssize_t *strides, int ndim, bint inc):
 *     cdef Py_ssize_t i
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "View.MemoryView":1398
 * 
 * @cname('__pyx_memoryview_slice_assign_scalar')
 * cdef void slice_assign_scalar(__Pyx_memviewslice *dst, int ndim,             # <<<<<<<<<<<<<<
 *                               size_t itemsize, void *item,
 *                               bint dtype_is_object) nogil:
 */

static void __pyx_memoryview_slice_assign_scalar(__Pyx_memviewslice *__pyx_v_dst, int __pyx_v_ndim, size_t __pyx_v_itemsize, void *__pyx_v_item, int __pyx_v_dtype_is_object) {

  /* "View.MemoryView":1401
 *                               size_t itemsize, void *item,
 *                               bint dtype_is_object) nogil:
 *     refcount_copying(dst, dtype_is_object, ndim, False)             # <<<<<<<<<<<<<<
 *     _slice_assign_scalar(dst.data, dst.shape, dst.strides, ndim,
 *                          itemsize, item)
 */
  __pyx_memoryview_refcount_copying(__pyx_v_dst, __pyx_v_dtype_is_object, __pyx_v_ndim, 0);

  /* "View.MemoryView":1402
 *                               bint dtype_is_object) nogil:
 *     refcount_copying(dst, dtype_is_object, ndim, False)
 *     _slice_assign_scalar(dst.data, dst.shape, dst.strides, ndim,             # <<<<<<<<<<<<<<
 *                          itemsize, item)
 *     refcount_copying(dst, dtype_is_object, ndim, True)
 */
  __pyx_memoryview__slice_assign_scalar(__pyx_v_dst->data, __pyx_v_dst->shape, __pyx_v_dst->strides, __pyx_v_ndim, __pyx_v_itemsize, __pyx_v_item);

  /* "View.MemoryView":1404
 *     _slice_assign_scalar(dst.data, dst.shape, dst.strides, ndim,
 *                          itemsize, item)
 *     refcount_copying(dst, dtype_is_object, ndim, True)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_memoryview_refcount_copying(__pyx_v_dst, __pyx_v_dtype_is_object, __pyx_v_ndim, 1);

  /* "View.MemoryView":1398
 * 
 * @cname('__pyx_memoryview_slice_assign_scalar')
 * cdef void slice_assign_scalar(__Pyx_memviewslice *dst, int ndim,             # <<<<<<<<<<<<<<
 *                               size_t itemsize, void *item,
 *                               bint dtype_is_object) nogil:
 */

  /* function exit code */
}

/* "View.MemoryView":1408
 * 
 * @cname('__pyx_memoryview__slice_assign_scalar')
 * cdef void _slice_assign_scalar(char *data, Py_ssize_t *shape,             # <<<<<<<<<<<<<<
 *                               Py_ssize_t *strides, int ndim,
 *                               size_t itemsize, void *item) nogil:
 */

static void __pyx_memoryview__slice_assign_scalar(char *__pyx_v_data, Py_ssize_t *__pyx_v_shape, Py_ssize_t *__pyx_v_strides, int __pyx_v_ndim, size_t __pyx_v_itemsize, void *__pyx_v_item) {
  CYTHON_UNUSED Py_ssize_t __pyx_v_i;
  Py_ssize_t __pyx_v_stride;
  Py_ssize_t __pyx_v_extent;
  int __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  Py_ssize_t __pyx_t_4;

  /* "View.MemoryView":1412
 *                               size_t itemsize, void *item) nogil:
 *     cdef Py_ssize_t i
 *     cdef Py_ssize_t stride = strides[0]             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t extent = shape[0]
 * 
 */
  __pyx_v_stride = (__pyx_v_strides[0]);

  /* "View.MemoryView":1413
 *     cdef Py_ssize_t i
 *     cdef Py_ssize_t stride = strides[0]
 *     cdef Py_ssize_t extent = shape[0]             # <<<<<<<<<<<<<<
 * 
 *     if ndim == 1:
 */
  __pyx_v_extent = (__pyx_v_shape[0]);

  /* "View.MemoryView":1415
 *     cdef Py_ssize_t extent = shape[0]
 * 
 *     if ndim == 1:             # <<<<<<<<<<<<<<
 *         for i in range(extent):
 *             memcpy(data, item, itemsize)
 */
  __pyx_t_1 = ((__pyx_v_ndim == 1) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":1416
 * 
 *     if ndim == 1:
 *         for i in range(extent):             # <<<<<<<<<<<<<<
 *             memcpy(data, item, itemsize)
 *             data += stride
 */
    __pyx_t_2 = __pyx_v_extent;
    __pyx_t_3 = __pyx_t_2;
    for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
      __pyx_v_i = __pyx_t_4;

      /* "View.MemoryView":1417
 *     if ndim == 1:
 *         for i in range(extent):
 *             memcpy(data, item, itemsize)             # <<<<<<<<<<<<<<
 *             data += stride
 *     else:
 */
      (void)(memcpy(__pyx_v_data, __pyx_v_item, __pyx_v_itemsize));

      /* "View.MemoryView":1418
 *         for i in range(extent):
 *             memcpy(data, item, itemsize)
 *             data += stride             # <<<<<<<<<<<<<<
 *     else:
 *         for i in range(extent):
 */
      __pyx_v_data = (__pyx_v_data + __pyx_v_stride);
    }

    /* "View.MemoryView":1415
 *     cdef Py_ssize_t extent = shape[0]
 * 
 *     if ndim == 1:             # <<<<<<<<<<<<<<
 *         for i in range(extent):
 *             memcpy(data, item, itemsize)
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":1420
 *             data += stride
 *     else:
 *         for i in range(extent):             # <<<<<<<<<<<<<<
 *             _slice_assign_scalar(data, shape + 1, strides + 1,
 *                                 ndim - 1, itemsize, item)
 */
  /*else*/ {
    __pyx_t_2 = __pyx_v_extent;
    __pyx_t_3 = __pyx_t_2;
    for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
      __pyx_v_i = __pyx_t_4;

      /* "View.MemoryView":1421
 *     else:
 *         for i in range(extent):
 *             _slice_assign_scalar(data, shape + 1, strides + 1,             # <<<<<<<<<<<<<<
 *                                 ndim - 1, itemsize, item)
 *             data += stride
 */
      __pyx_memoryview__slice_assign_scalar(__pyx_v_data, (__pyx_v_shape + 1), (__pyx_v_strides + 1), (__pyx_v_ndim - 1), __pyx_v_itemsize, __pyx_v_item);

      /* "View.MemoryView":1423
 *             _slice_assign_scalar(data, shape + 1, strides + 1,
 *                                 ndim - 1, itemsize, item)
 *             data += stride             # <<<<<<<<<<<<<<
 * 
 * 
 */
      __pyx_v_data = (__pyx_v_data + __pyx_v_stride);
    }
  }
  __pyx_L3:;

  /* "View.MemoryView":1408
 * 
 * @cname('__pyx_memoryview__slice_assign_scalar')
 * cdef void _slice_assign_scalar(char *data, Py_ssize_t *shape,             # <<<<<<<<<<<<<<
 *                               Py_ssize_t *strides, int ndim,
 *                               size_t itemsize, void *item) nogil:
 */

  /* function exit code */
}

/* "(tree fragment)":1
 * def __pyx_unpickle_Enum(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_1__pyx_unpickle_Enum(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_15View_dot_MemoryView_1__pyx_unpickle_Enum = {"__pyx_unpickle_Enum", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_15View_dot_MemoryView_1__pyx_unpickle_Enum, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_15View_dot_MemoryView_1__pyx_unpickle_Enum(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_Enum (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_type)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_checksum)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Enum", 1, 3, 3, 1); __PYX_ERR(2, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Enum", 1, 3, 3, 2); __PYX_ERR(2, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_unpickle_Enum") < 0)) __PYX_ERR(2, 1, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Enum", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(2, 1, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("View.MemoryView.__pyx_unpickle_Enum", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_15View_dot_MemoryView___pyx_unpickle_Enum(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView___pyx_unpickle_Enum(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Enum", 0);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0xb068931:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xb068931 = (name))" % __pyx_checksum)
 */
  __pyx_t_1 = ((__pyx_v___pyx_checksum != 0xb068931) != 0);
  if (__pyx_t_1) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0xb068931:
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xb068931 = (name))" % __pyx_checksum)
 *     __pyx_result = Enum.__new__(__pyx_type)
 */
    __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_PickleError);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_2);
    __pyx_v___pyx_PickleError = __pyx_t_2;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum != 0xb068931:
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xb068931 = (name))" % __pyx_checksum)             # <<<<<<<<<<<<<<
 *     __pyx_result = Enum.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_2 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_s_vs_0xb0, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_v___pyx_PickleError);
    __pyx_t_2 = __pyx_v___pyx_PickleError; __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_5, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(2, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0xb068931:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xb068931 = (name))" % __pyx_checksum)
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xb068931 = (name))" % __pyx_checksum)
 *     __pyx_result = Enum.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Enum__set_state(<Enum> __pyx_result, __pyx_state)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_MemviewEnum_type), __pyx_n_s_new); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_v___pyx_type) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v___pyx_type);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v___pyx_result = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xb068931 = (name))" % __pyx_checksum)
 *     __pyx_result = Enum.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Enum__set_state(<Enum> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_1 = (__pyx_v___pyx_state != Py_None);
  __pyx_t_6 = (__pyx_t_1 != 0);
  if (__pyx_t_6) {

    /* "(tree fragment)":9
 *     __pyx_result = Enum.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Enum__set_state(<Enum> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_Enum__set_state(Enum __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(2, 9, __pyx_L1_error)
    __pyx_t_3 = __pyx_unpickle_Enum__set_state(((struct __pyx_MemviewEnum_obj *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xb068931 = (name))" % __pyx_checksum)
 *     __pyx_result = Enum.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Enum__set_state(<Enum> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Enum__set_state(<Enum> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_Enum__set_state(Enum __pyx_result, tuple __pyx_state):
 *     __pyx_result.name = __pyx_state[0]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_Enum(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("View.MemoryView.__pyx_unpickle_Enum", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_Enum__set_state(<Enum> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Enum__set_state(Enum __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.name = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_unpickle_Enum__set_state(struct __pyx_MemviewEnum_obj *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Enum__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_Enum__set_state(Enum __pyx_result, tuple __pyx_state):
 *     __pyx_result.name = __pyx_state[0]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->name);
  __Pyx_DECREF(__pyx_v___pyx_result->name);
  __pyx_v___pyx_result->name = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_Enum__set_state(Enum __pyx_result, tuple __pyx_state):
 *     __pyx_result.name = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(2, 13, __pyx_L1_error)
  }
  __pyx_t_3 = PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_4 = ((__pyx_t_3 > 1) != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_5 = (__pyx_t_4 != 0);
  __pyx_t_2 = __pyx_t_5;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.name = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[1])             # <<<<<<<<<<<<<<
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_update); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(2, 14, __pyx_L1_error)
    }
    __pyx_t_6 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
      }
    }
    __pyx_t_1 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_8, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_6);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_Enum__set_state(Enum __pyx_result, tuple __pyx_state):
 *     __pyx_result.name = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_Enum__set_state(<Enum> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Enum__set_state(Enum __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.name = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("View.MemoryView.__pyx_unpickle_Enum__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_tp_new_6pandas_5_libs_8interval_IntervalMixin(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  return o;
}

static void __pyx_tp_dealloc_6pandas_5_libs_8interval_IntervalMixin(PyObject *o) {
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && (!PyType_IS_GC(Py_TYPE(o)) || !_PyGC_FINALIZED(o))) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  (*Py_TYPE(o)->tp_free)(o);
}

static PyObject *__pyx_getprop_6pandas_5_libs_8interval_13IntervalMixin_closed_left(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pandas_5_libs_8interval_13IntervalMixin_11closed_left_1__get__(o);
}

static PyObject *__pyx_getprop_6pandas_5_libs_8interval_13IntervalMixin_closed_right(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pandas_5_libs_8interval_13IntervalMixin_12closed_right_1__get__(o);
}

static PyObject *__pyx_getprop_6pandas_5_libs_8interval_13IntervalMixin_open_left(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pandas_5_libs_8interval_13IntervalMixin_9open_left_1__get__(o);
}

static PyObject *__pyx_getprop_6pandas_5_libs_8interval_13IntervalMixin_open_right(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pandas_5_libs_8interval_13IntervalMixin_10open_right_1__get__(o);
}

static PyObject *__pyx_getprop_6pandas_5_libs_8interval_13IntervalMixin_mid(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pandas_5_libs_8interval_13IntervalMixin_3mid_1__get__(o);
}

static PyObject *__pyx_getprop_6pandas_5_libs_8interval_13IntervalMixin_length(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pandas_5_libs_8interval_13IntervalMixin_6length_1__get__(o);
}

static PyObject *__pyx_getprop_6pandas_5_libs_8interval_13IntervalMixin_is_empty(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pandas_5_libs_8interval_13IntervalMixin_8is_empty_1__get__(o);
}

static PyMethodDef __pyx_methods_6pandas_5_libs_8interval_IntervalMixin[] = {
  {"_check_closed_matches", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_13IntervalMixin_1_check_closed_matches, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_13IntervalMixin__check_closed_matches},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_6pandas_5_libs_8interval_13IntervalMixin_3__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_6pandas_5_libs_8interval_13IntervalMixin_5__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_6pandas_5_libs_8interval_IntervalMixin[] = {
  {(char *)"closed_left", __pyx_getprop_6pandas_5_libs_8interval_13IntervalMixin_closed_left, 0, (char *)"\n        Check if the interval is closed on the left side.\n\n        For the meaning of `closed` and `open` see :class:`~pandas.Interval`.\n\n        Returns\n        -------\n        bool\n            ``True`` if the Interval is closed on the left-side, else\n            ``False``.\n        ", 0},
  {(char *)"closed_right", __pyx_getprop_6pandas_5_libs_8interval_13IntervalMixin_closed_right, 0, (char *)"\n        Check if the interval is closed on the right side.\n\n        For the meaning of `closed` and `open` see :class:`~pandas.Interval`.\n\n        Returns\n        -------\n        bool\n            ``True`` if the Interval is closed on the left-side, else\n            ``False``.\n        ", 0},
  {(char *)"open_left", __pyx_getprop_6pandas_5_libs_8interval_13IntervalMixin_open_left, 0, (char *)"\n        Check if the interval is open on the left side.\n\n        For the meaning of `closed` and `open` see :class:`~pandas.Interval`.\n\n        Returns\n        -------\n        bool\n            ``True`` if the Interval is closed on the left-side, else\n            ``False``.\n        ", 0},
  {(char *)"open_right", __pyx_getprop_6pandas_5_libs_8interval_13IntervalMixin_open_right, 0, (char *)"\n        Check if the interval is open on the right side.\n\n        For the meaning of `closed` and `open` see :class:`~pandas.Interval`.\n\n        Returns\n        -------\n        bool\n            ``True`` if the Interval is closed on the left-side, else\n            ``False``.\n        ", 0},
  {(char *)"mid", __pyx_getprop_6pandas_5_libs_8interval_13IntervalMixin_mid, 0, (char *)"\n        Return the midpoint of the Interval\n        ", 0},
  {(char *)"length", __pyx_getprop_6pandas_5_libs_8interval_13IntervalMixin_length, 0, (char *)"Return the length of the Interval", 0},
  {(char *)"is_empty", __pyx_getprop_6pandas_5_libs_8interval_13IntervalMixin_is_empty, 0, (char *)"\n        Indicates if an interval is empty, meaning it contains no points.\n\n        .. versionadded:: 0.25.0\n\n        Returns\n        -------\n        bool or ndarray\n            A boolean indicating if a scalar :class:`Interval` is empty, or a\n            boolean ``ndarray`` positionally indicating if an ``Interval`` in\n            an :class:`~arrays.IntervalArray` or :class:`IntervalIndex` is\n            empty.\n\n        Examples\n        --------\n        An :class:`Interval` that contains points is not empty:\n\n        >>> pd.Interval(0, 1, closed='right').is_empty\n        False\n\n        An ``Interval`` that does not contain any points is empty:\n\n        >>> pd.Interval(0, 0, closed='right').is_empty\n        True\n        >>> pd.Interval(0, 0, closed='left').is_empty\n        True\n        >>> pd.Interval(0, 0, closed='neither').is_empty\n        True\n\n        An ``Interval`` that contains a single point is not empty:\n\n        >>> pd.Interval(0, 0, closed='both').is_empty\n        False\n\n        An :class:`~arrays.IntervalArray` or :class:`IntervalIndex` returns a\n        boolean ``ndarray`` positionally indicating if an ``Interval`` is\n        empty:\n\n        >>> ivs = [pd.Interval(0, 0, closed='neither'),\n        ...        pd.Interval(1, 2, closed='neither')]\n        >>> pd.arrays.IntervalArray(ivs).is_empty\n        array([ True, False])\n\n        Missing values are not considered empty:\n\n        >>> ivs = [pd.Interval(0, 0, closed='neither'), np.nan]\n        >>> pd.IntervalIndex(ivs).is_empty\n        array([ True, False])\n        ", 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type_6pandas_5_libs_8interval_IntervalMixin = {
  PyVarObject_HEAD_INIT(0, 0)
  "pandas._libs.interval.IntervalMixin", /*tp_name*/
  sizeof(struct __pyx_obj_6pandas_5_libs_8interval_IntervalMixin), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6pandas_5_libs_8interval_IntervalMixin, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6pandas_5_libs_8interval_IntervalMixin, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_6pandas_5_libs_8interval_IntervalMixin, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6pandas_5_libs_8interval_IntervalMixin, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
};

static PyObject *__pyx_tp_new_6pandas_5_libs_8interval_Interval(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_6pandas_5_libs_8interval_Interval *p;
  PyObject *o = __pyx_tp_new_6pandas_5_libs_8interval_IntervalMixin(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_6pandas_5_libs_8interval_Interval *)o);
  p->left = Py_None; Py_INCREF(Py_None);
  p->right = Py_None; Py_INCREF(Py_None);
  p->closed = ((PyObject*)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_6pandas_5_libs_8interval_Interval(PyObject *o) {
  struct __pyx_obj_6pandas_5_libs_8interval_Interval *p = (struct __pyx_obj_6pandas_5_libs_8interval_Interval *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->left);
  Py_CLEAR(p->right);
  Py_CLEAR(p->closed);
  #if CYTHON_USE_TYPE_SLOTS
  if (PyType_IS_GC(Py_TYPE(o)->tp_base))
  #endif
  PyObject_GC_Track(o);
  __pyx_tp_dealloc_6pandas_5_libs_8interval_IntervalMixin(o);
}

static int __pyx_tp_traverse_6pandas_5_libs_8interval_Interval(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6pandas_5_libs_8interval_Interval *p = (struct __pyx_obj_6pandas_5_libs_8interval_Interval *)o;
  e = ((likely(__pyx_ptype_6pandas_5_libs_8interval_IntervalMixin)) ? ((__pyx_ptype_6pandas_5_libs_8interval_IntervalMixin->tp_traverse) ? __pyx_ptype_6pandas_5_libs_8interval_IntervalMixin->tp_traverse(o, v, a) : 0) : __Pyx_call_next_tp_traverse(o, v, a, __pyx_tp_traverse_6pandas_5_libs_8interval_Interval)); if (e) return e;
  if (p->left) {
    e = (*v)(p->left, a); if (e) return e;
  }
  if (p->right) {
    e = (*v)(p->right, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6pandas_5_libs_8interval_Interval(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6pandas_5_libs_8interval_Interval *p = (struct __pyx_obj_6pandas_5_libs_8interval_Interval *)o;
  if (likely(__pyx_ptype_6pandas_5_libs_8interval_IntervalMixin)) { if (__pyx_ptype_6pandas_5_libs_8interval_IntervalMixin->tp_clear) __pyx_ptype_6pandas_5_libs_8interval_IntervalMixin->tp_clear(o); } else __Pyx_call_next_tp_clear(o, __pyx_tp_clear_6pandas_5_libs_8interval_Interval);
  tmp = ((PyObject*)p->left);
  p->left = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->right);
  p->right = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_6pandas_5_libs_8interval_8Interval_left(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pandas_5_libs_8interval_8Interval_4left_1__get__(o);
}

static PyObject *__pyx_getprop_6pandas_5_libs_8interval_8Interval_right(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pandas_5_libs_8interval_8Interval_5right_1__get__(o);
}

static PyObject *__pyx_getprop_6pandas_5_libs_8interval_8Interval_closed(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pandas_5_libs_8interval_8Interval_6closed_1__get__(o);
}

static PyMethodDef __pyx_methods_6pandas_5_libs_8interval_Interval[] = {
  {"_validate_endpoint", (PyCFunction)__pyx_pw_6pandas_5_libs_8interval_8Interval_3_validate_endpoint, METH_O, 0},
  {"__reduce__", (PyCFunction)__pyx_pw_6pandas_5_libs_8interval_8Interval_11__reduce__, METH_NOARGS, 0},
  {"_repr_base", (PyCFunction)__pyx_pw_6pandas_5_libs_8interval_8Interval_13_repr_base, METH_NOARGS, 0},
  {"overlaps", (PyCFunction)__pyx_pw_6pandas_5_libs_8interval_8Interval_31overlaps, METH_O, __pyx_doc_6pandas_5_libs_8interval_8Interval_30overlaps},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_6pandas_5_libs_8interval_Interval[] = {
  {(char *)"left", __pyx_getprop_6pandas_5_libs_8interval_8Interval_left, 0, (char *)"Left bound for the interval", 0},
  {(char *)"right", __pyx_getprop_6pandas_5_libs_8interval_8Interval_right, 0, (char *)"Right bound for the interval", 0},
  {(char *)"closed", __pyx_getprop_6pandas_5_libs_8interval_8Interval_closed, 0, (char *)"\n    Whether the interval is closed on the left-side, right-side, both or\n    neither\n    ", 0},
  {0, 0, 0, 0, 0}
};

static PyNumberMethods __pyx_tp_as_number_Interval = {
  __pyx_pw_6pandas_5_libs_8interval_8Interval_19__add__, /*nb_add*/
  __pyx_pw_6pandas_5_libs_8interval_8Interval_21__sub__, /*nb_subtract*/
  __pyx_pw_6pandas_5_libs_8interval_8Interval_23__mul__, /*nb_multiply*/
  #if PY_MAJOR_VERSION < 3 || (CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x03050000)
  __pyx_pw_6pandas_5_libs_8interval_8Interval_25__div__, /*nb_divide*/
  #endif
  0, /*nb_remainder*/
  0, /*nb_divmod*/
  0, /*nb_power*/
  0, /*nb_negative*/
  0, /*nb_positive*/
  0, /*nb_absolute*/
  0, /*nb_nonzero*/
  0, /*nb_invert*/
  0, /*nb_lshift*/
  0, /*nb_rshift*/
  0, /*nb_and*/
  0, /*nb_xor*/
  0, /*nb_or*/
  #if PY_MAJOR_VERSION < 3 || (CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x03050000)
  0, /*nb_coerce*/
  #endif
  0, /*nb_int*/
  #if PY_MAJOR_VERSION < 3
  0, /*nb_long*/
  #else
  0, /*reserved*/
  #endif
  0, /*nb_float*/
  #if PY_MAJOR_VERSION < 3 || (CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x03050000)
  0, /*nb_oct*/
  #endif
  #if PY_MAJOR_VERSION < 3 || (CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x03050000)
  0, /*nb_hex*/
  #endif
  0, /*nb_inplace_add*/
  0, /*nb_inplace_subtract*/
  0, /*nb_inplace_multiply*/
  #if PY_MAJOR_VERSION < 3 || (CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x03050000)
  0, /*nb_inplace_divide*/
  #endif
  0, /*nb_inplace_remainder*/
  0, /*nb_inplace_power*/
  0, /*nb_inplace_lshift*/
  0, /*nb_inplace_rshift*/
  0, /*nb_inplace_and*/
  0, /*nb_inplace_xor*/
  0, /*nb_inplace_or*/
  __pyx_pw_6pandas_5_libs_8interval_8Interval_29__floordiv__, /*nb_floor_divide*/
  __pyx_pw_6pandas_5_libs_8interval_8Interval_27__truediv__, /*nb_true_divide*/
  0, /*nb_inplace_floor_divide*/
  0, /*nb_inplace_true_divide*/
  0, /*nb_index*/
  #if PY_VERSION_HEX >= 0x03050000
  0, /*nb_matrix_multiply*/
  #endif
  #if PY_VERSION_HEX >= 0x03050000
  0, /*nb_inplace_matrix_multiply*/
  #endif
};

static PySequenceMethods __pyx_tp_as_sequence_Interval = {
  0, /*sq_length*/
  0, /*sq_concat*/
  0, /*sq_repeat*/
  0, /*sq_item*/
  0, /*sq_slice*/
  0, /*sq_ass_item*/
  0, /*sq_ass_slice*/
  __pyx_pw_6pandas_5_libs_8interval_8Interval_7__contains__, /*sq_contains*/
  0, /*sq_inplace_concat*/
  0, /*sq_inplace_repeat*/
};

static PyTypeObject __pyx_type_6pandas_5_libs_8interval_Interval = {
  PyVarObject_HEAD_INIT(0, 0)
  "pandas._libs.interval.Interval", /*tp_name*/
  sizeof(struct __pyx_obj_6pandas_5_libs_8interval_Interval), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6pandas_5_libs_8interval_Interval, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_6pandas_5_libs_8interval_8Interval_15__repr__, /*tp_repr*/
  &__pyx_tp_as_number_Interval, /*tp_as_number*/
  &__pyx_tp_as_sequence_Interval, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  __pyx_pw_6pandas_5_libs_8interval_8Interval_5__hash__, /*tp_hash*/
  0, /*tp_call*/
  __pyx_pw_6pandas_5_libs_8interval_8Interval_17__str__, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  "\n    Immutable object implementing an Interval, a bounded slice-like interval.\n\n    .. versionadded:: 0.20.0\n\n    Parameters\n    ----------\n    left : orderable scalar\n        Left bound for the interval.\n    right : orderable scalar\n        Right bound for the interval.\n    closed : {'right', 'left', 'both', 'neither'}, default 'right'\n        Whether the interval is closed on the left-side, right-side, both or\n        neither. See the Notes for more detailed explanation.\n\n    See Also\n    --------\n    IntervalIndex : An Index of Interval objects that are all closed on the\n        same side.\n    cut : Convert continuous data into discrete bins (Categorical\n        of Interval objects).\n    qcut : Convert continuous data into bins (Categorical of Interval objects)\n        based on quantiles.\n    Period : Represents a period of time.\n\n    Notes\n    -----\n    The parameters `left` and `right` must be from the same type, you must be\n    able to compare them and they must satisfy ``left <= right``.\n\n    A closed interval (in mathematics denoted by square brackets) contains\n    its endpoints, i.e. the closed interval ``[0, 5]`` is characterized by the\n    conditions ``0 <= x <= 5``. This is what ``closed='both'`` stands for.\n    An open interval (in mathematics denoted by parentheses) does not contain\n    its endpoints, i.e. the open interval ``(0, 5)`` is characterized by the\n    conditions ``0 < x < 5``. This is what ``closed='neither'`` stands for.\n    Intervals can also be half-open or half-closed, i.e. ``[0, 5)`` is\n    described by ``0 <= x < 5`` (``closed='left'``) and ``(0, 5]`` is\n    described by ``0 < x <= 5`` (``closed='right'``).\n\n    Examples\n    --------\n    It is possible to build Intervals of different types, like numeric ones:\n\n    >>> iv = pd.Interval(left=0, right=5)\n    >>> iv\n    Interval(0, 5, closed='right')\n\n    You can check if an element belongs to it\n\n    >>> 2.5 in iv\n    True\n\n    You can ""test the bounds (``closed='right'``, so ``0 < x <= 5``):\n\n    >>> 0 in iv\n    False\n    >>> 5 in iv\n    True\n    >>> 0.0001 in iv\n    True\n\n    Calculate its length\n\n    >>> iv.length\n    5\n\n    You can operate with `+` and `*` over an Interval and the operation\n    is applied to each of its bounds, so the result depends on the type\n    of the bound elements\n\n    >>> shifted_iv = iv + 3\n    >>> shifted_iv\n    Interval(3, 8, closed='right')\n    >>> extended_iv = iv * 10.0\n    >>> extended_iv\n    Interval(0.0, 50.0, closed='right')\n\n    To create a time interval you can use Timestamps as the bounds\n\n    >>> year_2017 = pd.Interval(pd.Timestamp('2017-01-01 00:00:00'),\n    ...                         pd.Timestamp('2018-01-01 00:00:00'),\n    ...                         closed='left')\n    >>> pd.Timestamp('2017-01-01 00:00') in year_2017\n    True\n    >>> year_2017.length\n    Timedelta('365 days 00:00:00')\n\n    And also you can create string intervals\n\n    >>> volume_1 = pd.Interval('Ant', 'Dog', closed='both')\n    >>> 'Bee' in volume_1\n    True\n    ", /*tp_doc*/
  __pyx_tp_traverse_6pandas_5_libs_8interval_Interval, /*tp_traverse*/
  __pyx_tp_clear_6pandas_5_libs_8interval_Interval, /*tp_clear*/
  __pyx_pw_6pandas_5_libs_8interval_8Interval_9__richcmp__, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6pandas_5_libs_8interval_Interval, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_6pandas_5_libs_8interval_Interval, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_6pandas_5_libs_8interval_8Interval_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6pandas_5_libs_8interval_Interval, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
};

static PyObject *__pyx_tp_new_6pandas_5_libs_8interval_IntervalTree(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_6pandas_5_libs_8interval_IntervalTree *p;
  PyObject *o = __pyx_tp_new_6pandas_5_libs_8interval_IntervalMixin(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_6pandas_5_libs_8interval_IntervalTree *)o);
  p->left = Py_None; Py_INCREF(Py_None);
  p->right = Py_None; Py_INCREF(Py_None);
  p->root = Py_None; Py_INCREF(Py_None);
  p->dtype = Py_None; Py_INCREF(Py_None);
  p->closed = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->_is_overlapping = Py_None; Py_INCREF(Py_None);
  p->_left_sorter = Py_None; Py_INCREF(Py_None);
  p->_right_sorter = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_6pandas_5_libs_8interval_IntervalTree(PyObject *o) {
  struct __pyx_obj_6pandas_5_libs_8interval_IntervalTree *p = (struct __pyx_obj_6pandas_5_libs_8interval_IntervalTree *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->left);
  Py_CLEAR(p->right);
  Py_CLEAR(p->root);
  Py_CLEAR(p->dtype);
  Py_CLEAR(p->closed);
  Py_CLEAR(p->_is_overlapping);
  Py_CLEAR(p->_left_sorter);
  Py_CLEAR(p->_right_sorter);
  #if CYTHON_USE_TYPE_SLOTS
  if (PyType_IS_GC(Py_TYPE(o)->tp_base))
  #endif
  PyObject_GC_Track(o);
  __pyx_tp_dealloc_6pandas_5_libs_8interval_IntervalMixin(o);
}

static int __pyx_tp_traverse_6pandas_5_libs_8interval_IntervalTree(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6pandas_5_libs_8interval_IntervalTree *p = (struct __pyx_obj_6pandas_5_libs_8interval_IntervalTree *)o;
  e = ((likely(__pyx_ptype_6pandas_5_libs_8interval_IntervalMixin)) ? ((__pyx_ptype_6pandas_5_libs_8interval_IntervalMixin->tp_traverse) ? __pyx_ptype_6pandas_5_libs_8interval_IntervalMixin->tp_traverse(o, v, a) : 0) : __Pyx_call_next_tp_traverse(o, v, a, __pyx_tp_traverse_6pandas_5_libs_8interval_IntervalTree)); if (e) return e;
  if (p->left) {
    e = (*v)(p->left, a); if (e) return e;
  }
  if (p->right) {
    e = (*v)(p->right, a); if (e) return e;
  }
  if (p->root) {
    e = (*v)(p->root, a); if (e) return e;
  }
  if (p->dtype) {
    e = (*v)(p->dtype, a); if (e) return e;
  }
  if (p->_is_overlapping) {
    e = (*v)(p->_is_overlapping, a); if (e) return e;
  }
  if (p->_left_sorter) {
    e = (*v)(p->_left_sorter, a); if (e) return e;
  }
  if (p->_right_sorter) {
    e = (*v)(p->_right_sorter, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6pandas_5_libs_8interval_IntervalTree(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6pandas_5_libs_8interval_IntervalTree *p = (struct __pyx_obj_6pandas_5_libs_8interval_IntervalTree *)o;
  if (likely(__pyx_ptype_6pandas_5_libs_8interval_IntervalMixin)) { if (__pyx_ptype_6pandas_5_libs_8interval_IntervalMixin->tp_clear) __pyx_ptype_6pandas_5_libs_8interval_IntervalMixin->tp_clear(o); } else __Pyx_call_next_tp_clear(o, __pyx_tp_clear_6pandas_5_libs_8interval_IntervalTree);
  tmp = ((PyObject*)p->left);
  p->left = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->right);
  p->right = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->root);
  p->root = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->dtype);
  p->dtype = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_is_overlapping);
  p->_is_overlapping = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_left_sorter);
  p->_left_sorter = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_right_sorter);
  p->_right_sorter = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_6pandas_5_libs_8interval_12IntervalTree_left_sorter(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pandas_5_libs_8interval_12IntervalTree_11left_sorter_1__get__(o);
}

static PyObject *__pyx_getprop_6pandas_5_libs_8interval_12IntervalTree_right_sorter(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pandas_5_libs_8interval_12IntervalTree_12right_sorter_1__get__(o);
}

static PyObject *__pyx_getprop_6pandas_5_libs_8interval_12IntervalTree_is_overlapping(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pandas_5_libs_8interval_12IntervalTree_14is_overlapping_1__get__(o);
}

static PyObject *__pyx_getprop_6pandas_5_libs_8interval_12IntervalTree_is_monotonic_increasing(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pandas_5_libs_8interval_12IntervalTree_23is_monotonic_increasing_1__get__(o);
}

static PyObject *__pyx_getprop_6pandas_5_libs_8interval_12IntervalTree_left(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pandas_5_libs_8interval_12IntervalTree_4left_1__get__(o);
}

static PyObject *__pyx_getprop_6pandas_5_libs_8interval_12IntervalTree_right(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pandas_5_libs_8interval_12IntervalTree_5right_1__get__(o);
}

static PyObject *__pyx_getprop_6pandas_5_libs_8interval_12IntervalTree_root(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pandas_5_libs_8interval_12IntervalTree_4root_1__get__(o);
}

static PyObject *__pyx_getprop_6pandas_5_libs_8interval_12IntervalTree_dtype(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pandas_5_libs_8interval_12IntervalTree_5dtype_1__get__(o);
}

static PyObject *__pyx_getprop_6pandas_5_libs_8interval_12IntervalTree_closed(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pandas_5_libs_8interval_12IntervalTree_6closed_1__get__(o);
}

static PyMethodDef __pyx_methods_6pandas_5_libs_8interval_IntervalTree[] = {
  {"_get_partial_overlap", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_12IntervalTree_5_get_partial_overlap, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_12IntervalTree_4_get_partial_overlap},
  {"get_loc_interval", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6pandas_5_libs_8interval_12IntervalTree_7get_loc_interval, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6pandas_5_libs_8interval_12IntervalTree_6get_loc_interval},
  {"clear_mapping", (PyCFunction)__pyx_pw_6pandas_5_libs_8interval_12IntervalTree_15clear_mapping, METH_NOARGS, 0},
  {"__pyx_fuse_0get_loc", (PyCFunction)__pyx_fuse_0__pyx_pw_6pandas_5_libs_8interval_12IntervalTree_17get_loc, METH_O, __pyx_doc_6pandas_5_libs_8interval_12IntervalTree_2get_loc},
  {"__pyx_fuse_1get_loc", (PyCFunction)__pyx_fuse_1__pyx_pw_6pandas_5_libs_8interval_12IntervalTree_19get_loc, METH_O, __pyx_doc_6pandas_5_libs_8interval_12IntervalTree_2get_loc},
  {"__pyx_fuse_2get_loc", (PyCFunction)__pyx_fuse_2__pyx_pw_6pandas_5_libs_8interval_12IntervalTree_21get_loc, METH_O, __pyx_doc_6pandas_5_libs_8interval_12IntervalTree_2get_loc},
  {"__pyx_fuse_3get_loc", (PyCFunction)__pyx_fuse_3__pyx_pw_6pandas_5_libs_8interval_12IntervalTree_23get_loc, METH_O, __pyx_doc_6pandas_5_libs_8interval_12IntervalTree_2get_loc},
  {"__pyx_fuse_4get_loc", (PyCFunction)__pyx_fuse_4__pyx_pw_6pandas_5_libs_8interval_12IntervalTree_25get_loc, METH_O, __pyx_doc_6pandas_5_libs_8interval_12IntervalTree_2get_loc},
  {"__pyx_fuse_0get_indexer", (PyCFunction)__pyx_fuse_0__pyx_pw_6pandas_5_libs_8interval_12IntervalTree_29get_indexer, METH_O, __pyx_doc_6pandas_5_libs_8interval_12IntervalTree_8get_indexer},
  {"__pyx_fuse_1get_indexer", (PyCFunction)__pyx_fuse_1__pyx_pw_6pandas_5_libs_8interval_12IntervalTree_31get_indexer, METH_O, __pyx_doc_6pandas_5_libs_8interval_12IntervalTree_8get_indexer},
  {"__pyx_fuse_2get_indexer", (PyCFunction)__pyx_fuse_2__pyx_pw_6pandas_5_libs_8interval_12IntervalTree_33get_indexer, METH_O, __pyx_doc_6pandas_5_libs_8interval_12IntervalTree_8get_indexer},
  {"__pyx_fuse_3get_indexer", (PyCFunction)__pyx_fuse_3__pyx_pw_6pandas_5_libs_8interval_12IntervalTree_35get_indexer, METH_O, __pyx_doc_6pandas_5_libs_8interval_12IntervalTree_8get_indexer},
  {"__pyx_fuse_4get_indexer", (PyCFunction)__pyx_fuse_4__pyx_pw_6pandas_5_libs_8interval_12IntervalTree_37get_indexer, METH_O, __pyx_doc_6pandas_5_libs_8interval_12IntervalTree_8get_indexer},
  {"__pyx_fuse_0get_indexer_non_unique", (PyCFunction)__pyx_fuse_0__pyx_pw_6pandas_5_libs_8interval_12IntervalTree_41get_indexer_non_unique, METH_O, __pyx_doc_6pandas_5_libs_8interval_12IntervalTree_10get_indexer_non_unique},
  {"__pyx_fuse_1get_indexer_non_unique", (PyCFunction)__pyx_fuse_1__pyx_pw_6pandas_5_libs_8interval_12IntervalTree_43get_indexer_non_unique, METH_O, __pyx_doc_6pandas_5_libs_8interval_12IntervalTree_10get_indexer_non_unique},
  {"__pyx_fuse_2get_indexer_non_unique", (PyCFunction)__pyx_fuse_2__pyx_pw_6pandas_5_libs_8interval_12IntervalTree_45get_indexer_non_unique, METH_O, __pyx_doc_6pandas_5_libs_8interval_12IntervalTree_10get_indexer_non_unique},
  {"__pyx_fuse_3get_indexer_non_unique", (PyCFunction)__pyx_fuse_3__pyx_pw_6pandas_5_libs_8interval_12IntervalTree_47get_indexer_non_unique, METH_O, __pyx_doc_6pandas_5_libs_8interval_12IntervalTree_10get_indexer_non_unique},
  {"__pyx_fuse_4get_indexer_non_unique", (PyCFunction)__pyx_fuse_4__pyx_pw_6pandas_5_libs_8interval_12IntervalTree_49get_indexer_non_unique, METH_O, __pyx_doc_6pandas_5_libs_8interval_12IntervalTree_10get_indexer_non_unique},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_6pandas_5_libs_8interval_12IntervalTree_53__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_6pandas_5_libs_8interval_12IntervalTree_55__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_6pandas_5_libs_8interval_IntervalTree[] = {
  {(char *)"left_sorter", __pyx_getprop_6pandas_5_libs_8interval_12IntervalTree_left_sorter, 0, (char *)"How to sort the left labels; this is used for binary search\n        ", 0},
  {(char *)"right_sorter", __pyx_getprop_6pandas_5_libs_8interval_12IntervalTree_right_sorter, 0, (char *)"How to sort the right labels\n        ", 0},
  {(char *)"is_overlapping", __pyx_getprop_6pandas_5_libs_8interval_12IntervalTree_is_overlapping, 0, (char *)"\n        Determine if the IntervalTree contains overlapping intervals.\n        Cached as self._is_overlapping.\n        ", 0},
  {(char *)"is_monotonic_increasing", __pyx_getprop_6pandas_5_libs_8interval_12IntervalTree_is_monotonic_increasing, 0, (char *)"\n        Return True if the IntervalTree is monotonic increasing (only equal or\n        increasing values), else False\n        ", 0},
  {(char *)"left", __pyx_getprop_6pandas_5_libs_8interval_12IntervalTree_left, 0, (char *)0, 0},
  {(char *)"right", __pyx_getprop_6pandas_5_libs_8interval_12IntervalTree_right, 0, (char *)0, 0},
  {(char *)"root", __pyx_getprop_6pandas_5_libs_8interval_12IntervalTree_root, 0, (char *)0, 0},
  {(char *)"dtype", __pyx_getprop_6pandas_5_libs_8interval_12IntervalTree_dtype, 0, (char *)0, 0},
  {(char *)"closed", __pyx_getprop_6pandas_5_libs_8interval_12IntervalTree_closed, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type_6pandas_5_libs_8interval_IntervalTree = {
  PyVarObject_HEAD_INIT(0, 0)
  "pandas._libs.interval.IntervalTree", /*tp_name*/
  sizeof(struct __pyx_obj_6pandas_5_libs_8interval_IntervalTree), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6pandas_5_libs_8interval_IntervalTree, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_6pandas_5_libs_8interval_12IntervalTree_13__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  "A centered interval tree\n\n    Based off the algorithm described on Wikipedia:\n    http://en.wikipedia.org/wiki/Interval_tree\n\n    we are emulating the IndexEngine interface\n    ", /*tp_doc*/
  __pyx_tp_traverse_6pandas_5_libs_8interval_IntervalTree, /*tp_traverse*/
  __pyx_tp_clear_6pandas_5_libs_8interval_IntervalTree, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6pandas_5_libs_8interval_IntervalTree, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_6pandas_5_libs_8interval_IntervalTree, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_6pandas_5_libs_8interval_12IntervalTree_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6pandas_5_libs_8interval_IntervalTree, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
};
static struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode __pyx_vtable_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode;

static PyObject *__pyx_tp_new_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode *)o);
  p->__pyx_vtab = __pyx_vtabptr_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode;
  p->left_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode *)Py_None); Py_INCREF(Py_None);
  p->right_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode *)Py_None); Py_INCREF(Py_None);
  p->center_left_values.data = NULL;
  p->center_left_values.memview = NULL;
  p->center_right_values.data = NULL;
  p->center_right_values.memview = NULL;
  p->left.data = NULL;
  p->left.memview = NULL;
  p->right.data = NULL;
  p->right.memview = NULL;
  p->center_left_indices.data = NULL;
  p->center_left_indices.memview = NULL;
  p->center_right_indices.data = NULL;
  p->center_right_indices.memview = NULL;
  p->indices.data = NULL;
  p->indices.memview = NULL;
  return o;
}

static void __pyx_tp_dealloc_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode(PyObject *o) {
  struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode *p = (struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->left_node);
  Py_CLEAR(p->right_node);
  __PYX_XDEC_MEMVIEW(&p->center_left_values, 1);
  __PYX_XDEC_MEMVIEW(&p->center_right_values, 1);
  __PYX_XDEC_MEMVIEW(&p->left, 1);
  __PYX_XDEC_MEMVIEW(&p->right, 1);
  __PYX_XDEC_MEMVIEW(&p->center_left_indices, 1);
  __PYX_XDEC_MEMVIEW(&p->center_right_indices, 1);
  __PYX_XDEC_MEMVIEW(&p->indices, 1);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode *p = (struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode *)o;
  if (p->left_node) {
    e = (*v)(((PyObject *)p->left_node), a); if (e) return e;
  }
  if (p->right_node) {
    e = (*v)(((PyObject *)p->right_node), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode *p = (struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode *)o;
  tmp = ((PyObject*)p->left_node);
  p->left_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->right_node);
  p->right_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_pivot(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_5pivot_1__get__(o);
}

static PyObject *__pyx_getprop_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_n_elements(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_10n_elements_1__get__(o);
}

static PyObject *__pyx_getprop_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_n_center(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_8n_center_1__get__(o);
}

static PyObject *__pyx_getprop_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_leaf_size(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_9leaf_size_1__get__(o);
}

static PyObject *__pyx_getprop_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_is_leaf_node(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_12is_leaf_node_1__get__(o);
}

static PyMethodDef __pyx_methods_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode[] = {
  {"counts", (PyCFunction)__pyx_pw_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_7counts, METH_NOARGS, __pyx_doc_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_6counts},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_21__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_23__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode[] = {
  {(char *)"pivot", __pyx_getprop_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_pivot, 0, (char *)0, 0},
  {(char *)"n_elements", __pyx_getprop_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_n_elements, 0, (char *)0, 0},
  {(char *)"n_center", __pyx_getprop_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_n_center, 0, (char *)0, 0},
  {(char *)"leaf_size", __pyx_getprop_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_leaf_size, 0, (char *)0, 0},
  {(char *)"is_leaf_node", __pyx_getprop_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_is_leaf_node, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode = {
  PyVarObject_HEAD_INIT(0, 0)
  "pandas._libs.interval.Float32ClosedLeftIntervalNode", /*tp_name*/
  sizeof(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_5__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  "Non-terminal node for an IntervalTree\n\n    Categorizes intervals by those that fall to the left, those that fall to\n    the right, and those that overlap with the pivot.\n    ", /*tp_doc*/
  __pyx_tp_traverse_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode, /*tp_traverse*/
  __pyx_tp_clear_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
};
static struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode __pyx_vtable_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode;

static PyObject *__pyx_tp_new_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode *)o);
  p->__pyx_vtab = __pyx_vtabptr_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode;
  p->left_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode *)Py_None); Py_INCREF(Py_None);
  p->right_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode *)Py_None); Py_INCREF(Py_None);
  p->center_left_values.data = NULL;
  p->center_left_values.memview = NULL;
  p->center_right_values.data = NULL;
  p->center_right_values.memview = NULL;
  p->left.data = NULL;
  p->left.memview = NULL;
  p->right.data = NULL;
  p->right.memview = NULL;
  p->center_left_indices.data = NULL;
  p->center_left_indices.memview = NULL;
  p->center_right_indices.data = NULL;
  p->center_right_indices.memview = NULL;
  p->indices.data = NULL;
  p->indices.memview = NULL;
  return o;
}

static void __pyx_tp_dealloc_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode(PyObject *o) {
  struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode *p = (struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->left_node);
  Py_CLEAR(p->right_node);
  __PYX_XDEC_MEMVIEW(&p->center_left_values, 1);
  __PYX_XDEC_MEMVIEW(&p->center_right_values, 1);
  __PYX_XDEC_MEMVIEW(&p->left, 1);
  __PYX_XDEC_MEMVIEW(&p->right, 1);
  __PYX_XDEC_MEMVIEW(&p->center_left_indices, 1);
  __PYX_XDEC_MEMVIEW(&p->center_right_indices, 1);
  __PYX_XDEC_MEMVIEW(&p->indices, 1);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode *p = (struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode *)o;
  if (p->left_node) {
    e = (*v)(((PyObject *)p->left_node), a); if (e) return e;
  }
  if (p->right_node) {
    e = (*v)(((PyObject *)p->right_node), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode *p = (struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode *)o;
  tmp = ((PyObject*)p->left_node);
  p->left_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->right_node);
  p->right_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_pivot(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_5pivot_1__get__(o);
}

static PyObject *__pyx_getprop_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_n_elements(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_10n_elements_1__get__(o);
}

static PyObject *__pyx_getprop_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_n_center(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_8n_center_1__get__(o);
}

static PyObject *__pyx_getprop_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_leaf_size(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_9leaf_size_1__get__(o);
}

static PyObject *__pyx_getprop_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_is_leaf_node(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_12is_leaf_node_1__get__(o);
}

static PyMethodDef __pyx_methods_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode[] = {
  {"counts", (PyCFunction)__pyx_pw_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_7counts, METH_NOARGS, __pyx_doc_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_6counts},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_21__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_23__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode[] = {
  {(char *)"pivot", __pyx_getprop_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_pivot, 0, (char *)0, 0},
  {(char *)"n_elements", __pyx_getprop_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_n_elements, 0, (char *)0, 0},
  {(char *)"n_center", __pyx_getprop_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_n_center, 0, (char *)0, 0},
  {(char *)"leaf_size", __pyx_getprop_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_leaf_size, 0, (char *)0, 0},
  {(char *)"is_leaf_node", __pyx_getprop_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_is_leaf_node, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode = {
  PyVarObject_HEAD_INIT(0, 0)
  "pandas._libs.interval.Float32ClosedRightIntervalNode", /*tp_name*/
  sizeof(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_5__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  "Non-terminal node for an IntervalTree\n\n    Categorizes intervals by those that fall to the left, those that fall to\n    the right, and those that overlap with the pivot.\n    ", /*tp_doc*/
  __pyx_tp_traverse_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode, /*tp_traverse*/
  __pyx_tp_clear_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
};
static struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode __pyx_vtable_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode;

static PyObject *__pyx_tp_new_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode *)o);
  p->__pyx_vtab = __pyx_vtabptr_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode;
  p->left_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode *)Py_None); Py_INCREF(Py_None);
  p->right_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode *)Py_None); Py_INCREF(Py_None);
  p->center_left_values.data = NULL;
  p->center_left_values.memview = NULL;
  p->center_right_values.data = NULL;
  p->center_right_values.memview = NULL;
  p->left.data = NULL;
  p->left.memview = NULL;
  p->right.data = NULL;
  p->right.memview = NULL;
  p->center_left_indices.data = NULL;
  p->center_left_indices.memview = NULL;
  p->center_right_indices.data = NULL;
  p->center_right_indices.memview = NULL;
  p->indices.data = NULL;
  p->indices.memview = NULL;
  return o;
}

static void __pyx_tp_dealloc_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode(PyObject *o) {
  struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode *p = (struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->left_node);
  Py_CLEAR(p->right_node);
  __PYX_XDEC_MEMVIEW(&p->center_left_values, 1);
  __PYX_XDEC_MEMVIEW(&p->center_right_values, 1);
  __PYX_XDEC_MEMVIEW(&p->left, 1);
  __PYX_XDEC_MEMVIEW(&p->right, 1);
  __PYX_XDEC_MEMVIEW(&p->center_left_indices, 1);
  __PYX_XDEC_MEMVIEW(&p->center_right_indices, 1);
  __PYX_XDEC_MEMVIEW(&p->indices, 1);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode *p = (struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode *)o;
  if (p->left_node) {
    e = (*v)(((PyObject *)p->left_node), a); if (e) return e;
  }
  if (p->right_node) {
    e = (*v)(((PyObject *)p->right_node), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode *p = (struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode *)o;
  tmp = ((PyObject*)p->left_node);
  p->left_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->right_node);
  p->right_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_pivot(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_5pivot_1__get__(o);
}

static PyObject *__pyx_getprop_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_n_elements(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_10n_elements_1__get__(o);
}

static PyObject *__pyx_getprop_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_n_center(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_8n_center_1__get__(o);
}

static PyObject *__pyx_getprop_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_leaf_size(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_9leaf_size_1__get__(o);
}

static PyObject *__pyx_getprop_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_is_leaf_node(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_12is_leaf_node_1__get__(o);
}

static PyMethodDef __pyx_methods_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode[] = {
  {"counts", (PyCFunction)__pyx_pw_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_7counts, METH_NOARGS, __pyx_doc_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_6counts},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_21__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_23__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode[] = {
  {(char *)"pivot", __pyx_getprop_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_pivot, 0, (char *)0, 0},
  {(char *)"n_elements", __pyx_getprop_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_n_elements, 0, (char *)0, 0},
  {(char *)"n_center", __pyx_getprop_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_n_center, 0, (char *)0, 0},
  {(char *)"leaf_size", __pyx_getprop_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_leaf_size, 0, (char *)0, 0},
  {(char *)"is_leaf_node", __pyx_getprop_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_is_leaf_node, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode = {
  PyVarObject_HEAD_INIT(0, 0)
  "pandas._libs.interval.Float32ClosedBothIntervalNode", /*tp_name*/
  sizeof(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_5__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  "Non-terminal node for an IntervalTree\n\n    Categorizes intervals by those that fall to the left, those that fall to\n    the right, and those that overlap with the pivot.\n    ", /*tp_doc*/
  __pyx_tp_traverse_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode, /*tp_traverse*/
  __pyx_tp_clear_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
};
static struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode __pyx_vtable_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode;

static PyObject *__pyx_tp_new_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode *)o);
  p->__pyx_vtab = __pyx_vtabptr_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode;
  p->left_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode *)Py_None); Py_INCREF(Py_None);
  p->right_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode *)Py_None); Py_INCREF(Py_None);
  p->center_left_values.data = NULL;
  p->center_left_values.memview = NULL;
  p->center_right_values.data = NULL;
  p->center_right_values.memview = NULL;
  p->left.data = NULL;
  p->left.memview = NULL;
  p->right.data = NULL;
  p->right.memview = NULL;
  p->center_left_indices.data = NULL;
  p->center_left_indices.memview = NULL;
  p->center_right_indices.data = NULL;
  p->center_right_indices.memview = NULL;
  p->indices.data = NULL;
  p->indices.memview = NULL;
  return o;
}

static void __pyx_tp_dealloc_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode(PyObject *o) {
  struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode *p = (struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->left_node);
  Py_CLEAR(p->right_node);
  __PYX_XDEC_MEMVIEW(&p->center_left_values, 1);
  __PYX_XDEC_MEMVIEW(&p->center_right_values, 1);
  __PYX_XDEC_MEMVIEW(&p->left, 1);
  __PYX_XDEC_MEMVIEW(&p->right, 1);
  __PYX_XDEC_MEMVIEW(&p->center_left_indices, 1);
  __PYX_XDEC_MEMVIEW(&p->center_right_indices, 1);
  __PYX_XDEC_MEMVIEW(&p->indices, 1);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode *p = (struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode *)o;
  if (p->left_node) {
    e = (*v)(((PyObject *)p->left_node), a); if (e) return e;
  }
  if (p->right_node) {
    e = (*v)(((PyObject *)p->right_node), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode *p = (struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode *)o;
  tmp = ((PyObject*)p->left_node);
  p->left_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->right_node);
  p->right_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_pivot(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_5pivot_1__get__(o);
}

static PyObject *__pyx_getprop_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_n_elements(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_10n_elements_1__get__(o);
}

static PyObject *__pyx_getprop_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_n_center(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_8n_center_1__get__(o);
}

static PyObject *__pyx_getprop_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_leaf_size(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_9leaf_size_1__get__(o);
}

static PyObject *__pyx_getprop_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_is_leaf_node(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_12is_leaf_node_1__get__(o);
}

static PyMethodDef __pyx_methods_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode[] = {
  {"counts", (PyCFunction)__pyx_pw_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_7counts, METH_NOARGS, __pyx_doc_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_6counts},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_21__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_23__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode[] = {
  {(char *)"pivot", __pyx_getprop_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_pivot, 0, (char *)0, 0},
  {(char *)"n_elements", __pyx_getprop_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_n_elements, 0, (char *)0, 0},
  {(char *)"n_center", __pyx_getprop_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_n_center, 0, (char *)0, 0},
  {(char *)"leaf_size", __pyx_getprop_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_leaf_size, 0, (char *)0, 0},
  {(char *)"is_leaf_node", __pyx_getprop_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_is_leaf_node, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode = {
  PyVarObject_HEAD_INIT(0, 0)
  "pandas._libs.interval.Float32ClosedNeitherIntervalNode", /*tp_name*/
  sizeof(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_5__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  "Non-terminal node for an IntervalTree\n\n    Categorizes intervals by those that fall to the left, those that fall to\n    the right, and those that overlap with the pivot.\n    ", /*tp_doc*/
  __pyx_tp_traverse_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode, /*tp_traverse*/
  __pyx_tp_clear_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
};
static struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode __pyx_vtable_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode;

static PyObject *__pyx_tp_new_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode *)o);
  p->__pyx_vtab = __pyx_vtabptr_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode;
  p->left_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode *)Py_None); Py_INCREF(Py_None);
  p->right_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode *)Py_None); Py_INCREF(Py_None);
  p->center_left_values.data = NULL;
  p->center_left_values.memview = NULL;
  p->center_right_values.data = NULL;
  p->center_right_values.memview = NULL;
  p->left.data = NULL;
  p->left.memview = NULL;
  p->right.data = NULL;
  p->right.memview = NULL;
  p->center_left_indices.data = NULL;
  p->center_left_indices.memview = NULL;
  p->center_right_indices.data = NULL;
  p->center_right_indices.memview = NULL;
  p->indices.data = NULL;
  p->indices.memview = NULL;
  return o;
}

static void __pyx_tp_dealloc_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode(PyObject *o) {
  struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode *p = (struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->left_node);
  Py_CLEAR(p->right_node);
  __PYX_XDEC_MEMVIEW(&p->center_left_values, 1);
  __PYX_XDEC_MEMVIEW(&p->center_right_values, 1);
  __PYX_XDEC_MEMVIEW(&p->left, 1);
  __PYX_XDEC_MEMVIEW(&p->right, 1);
  __PYX_XDEC_MEMVIEW(&p->center_left_indices, 1);
  __PYX_XDEC_MEMVIEW(&p->center_right_indices, 1);
  __PYX_XDEC_MEMVIEW(&p->indices, 1);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode *p = (struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode *)o;
  if (p->left_node) {
    e = (*v)(((PyObject *)p->left_node), a); if (e) return e;
  }
  if (p->right_node) {
    e = (*v)(((PyObject *)p->right_node), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode *p = (struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode *)o;
  tmp = ((PyObject*)p->left_node);
  p->left_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->right_node);
  p->right_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_pivot(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_5pivot_1__get__(o);
}

static PyObject *__pyx_getprop_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_n_elements(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_10n_elements_1__get__(o);
}

static PyObject *__pyx_getprop_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_n_center(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_8n_center_1__get__(o);
}

static PyObject *__pyx_getprop_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_leaf_size(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_9leaf_size_1__get__(o);
}

static PyObject *__pyx_getprop_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_is_leaf_node(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_12is_leaf_node_1__get__(o);
}

static PyMethodDef __pyx_methods_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode[] = {
  {"counts", (PyCFunction)__pyx_pw_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_7counts, METH_NOARGS, __pyx_doc_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_6counts},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_21__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_23__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode[] = {
  {(char *)"pivot", __pyx_getprop_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_pivot, 0, (char *)0, 0},
  {(char *)"n_elements", __pyx_getprop_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_n_elements, 0, (char *)0, 0},
  {(char *)"n_center", __pyx_getprop_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_n_center, 0, (char *)0, 0},
  {(char *)"leaf_size", __pyx_getprop_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_leaf_size, 0, (char *)0, 0},
  {(char *)"is_leaf_node", __pyx_getprop_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_is_leaf_node, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode = {
  PyVarObject_HEAD_INIT(0, 0)
  "pandas._libs.interval.Float64ClosedLeftIntervalNode", /*tp_name*/
  sizeof(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_5__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  "Non-terminal node for an IntervalTree\n\n    Categorizes intervals by those that fall to the left, those that fall to\n    the right, and those that overlap with the pivot.\n    ", /*tp_doc*/
  __pyx_tp_traverse_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode, /*tp_traverse*/
  __pyx_tp_clear_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
};
static struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode __pyx_vtable_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode;

static PyObject *__pyx_tp_new_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode *)o);
  p->__pyx_vtab = __pyx_vtabptr_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode;
  p->left_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode *)Py_None); Py_INCREF(Py_None);
  p->right_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode *)Py_None); Py_INCREF(Py_None);
  p->center_left_values.data = NULL;
  p->center_left_values.memview = NULL;
  p->center_right_values.data = NULL;
  p->center_right_values.memview = NULL;
  p->left.data = NULL;
  p->left.memview = NULL;
  p->right.data = NULL;
  p->right.memview = NULL;
  p->center_left_indices.data = NULL;
  p->center_left_indices.memview = NULL;
  p->center_right_indices.data = NULL;
  p->center_right_indices.memview = NULL;
  p->indices.data = NULL;
  p->indices.memview = NULL;
  return o;
}

static void __pyx_tp_dealloc_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode(PyObject *o) {
  struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode *p = (struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->left_node);
  Py_CLEAR(p->right_node);
  __PYX_XDEC_MEMVIEW(&p->center_left_values, 1);
  __PYX_XDEC_MEMVIEW(&p->center_right_values, 1);
  __PYX_XDEC_MEMVIEW(&p->left, 1);
  __PYX_XDEC_MEMVIEW(&p->right, 1);
  __PYX_XDEC_MEMVIEW(&p->center_left_indices, 1);
  __PYX_XDEC_MEMVIEW(&p->center_right_indices, 1);
  __PYX_XDEC_MEMVIEW(&p->indices, 1);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode *p = (struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode *)o;
  if (p->left_node) {
    e = (*v)(((PyObject *)p->left_node), a); if (e) return e;
  }
  if (p->right_node) {
    e = (*v)(((PyObject *)p->right_node), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode *p = (struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode *)o;
  tmp = ((PyObject*)p->left_node);
  p->left_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->right_node);
  p->right_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_pivot(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_5pivot_1__get__(o);
}

static PyObject *__pyx_getprop_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_n_elements(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_10n_elements_1__get__(o);
}

static PyObject *__pyx_getprop_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_n_center(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_8n_center_1__get__(o);
}

static PyObject *__pyx_getprop_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_leaf_size(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_9leaf_size_1__get__(o);
}

static PyObject *__pyx_getprop_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_is_leaf_node(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_12is_leaf_node_1__get__(o);
}

static PyMethodDef __pyx_methods_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode[] = {
  {"counts", (PyCFunction)__pyx_pw_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_7counts, METH_NOARGS, __pyx_doc_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_6counts},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_21__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_23__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode[] = {
  {(char *)"pivot", __pyx_getprop_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_pivot, 0, (char *)0, 0},
  {(char *)"n_elements", __pyx_getprop_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_n_elements, 0, (char *)0, 0},
  {(char *)"n_center", __pyx_getprop_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_n_center, 0, (char *)0, 0},
  {(char *)"leaf_size", __pyx_getprop_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_leaf_size, 0, (char *)0, 0},
  {(char *)"is_leaf_node", __pyx_getprop_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_is_leaf_node, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode = {
  PyVarObject_HEAD_INIT(0, 0)
  "pandas._libs.interval.Float64ClosedRightIntervalNode", /*tp_name*/
  sizeof(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_5__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  "Non-terminal node for an IntervalTree\n\n    Categorizes intervals by those that fall to the left, those that fall to\n    the right, and those that overlap with the pivot.\n    ", /*tp_doc*/
  __pyx_tp_traverse_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode, /*tp_traverse*/
  __pyx_tp_clear_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
};
static struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode __pyx_vtable_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode;

static PyObject *__pyx_tp_new_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode *)o);
  p->__pyx_vtab = __pyx_vtabptr_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode;
  p->left_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode *)Py_None); Py_INCREF(Py_None);
  p->right_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode *)Py_None); Py_INCREF(Py_None);
  p->center_left_values.data = NULL;
  p->center_left_values.memview = NULL;
  p->center_right_values.data = NULL;
  p->center_right_values.memview = NULL;
  p->left.data = NULL;
  p->left.memview = NULL;
  p->right.data = NULL;
  p->right.memview = NULL;
  p->center_left_indices.data = NULL;
  p->center_left_indices.memview = NULL;
  p->center_right_indices.data = NULL;
  p->center_right_indices.memview = NULL;
  p->indices.data = NULL;
  p->indices.memview = NULL;
  return o;
}

static void __pyx_tp_dealloc_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode(PyObject *o) {
  struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode *p = (struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->left_node);
  Py_CLEAR(p->right_node);
  __PYX_XDEC_MEMVIEW(&p->center_left_values, 1);
  __PYX_XDEC_MEMVIEW(&p->center_right_values, 1);
  __PYX_XDEC_MEMVIEW(&p->left, 1);
  __PYX_XDEC_MEMVIEW(&p->right, 1);
  __PYX_XDEC_MEMVIEW(&p->center_left_indices, 1);
  __PYX_XDEC_MEMVIEW(&p->center_right_indices, 1);
  __PYX_XDEC_MEMVIEW(&p->indices, 1);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode *p = (struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode *)o;
  if (p->left_node) {
    e = (*v)(((PyObject *)p->left_node), a); if (e) return e;
  }
  if (p->right_node) {
    e = (*v)(((PyObject *)p->right_node), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode *p = (struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode *)o;
  tmp = ((PyObject*)p->left_node);
  p->left_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->right_node);
  p->right_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_pivot(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_5pivot_1__get__(o);
}

static PyObject *__pyx_getprop_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_n_elements(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_10n_elements_1__get__(o);
}

static PyObject *__pyx_getprop_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_n_center(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_8n_center_1__get__(o);
}

static PyObject *__pyx_getprop_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_leaf_size(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_9leaf_size_1__get__(o);
}

static PyObject *__pyx_getprop_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_is_leaf_node(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_12is_leaf_node_1__get__(o);
}

static PyMethodDef __pyx_methods_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode[] = {
  {"counts", (PyCFunction)__pyx_pw_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_7counts, METH_NOARGS, __pyx_doc_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_6counts},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_21__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_23__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode[] = {
  {(char *)"pivot", __pyx_getprop_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_pivot, 0, (char *)0, 0},
  {(char *)"n_elements", __pyx_getprop_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_n_elements, 0, (char *)0, 0},
  {(char *)"n_center", __pyx_getprop_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_n_center, 0, (char *)0, 0},
  {(char *)"leaf_size", __pyx_getprop_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_leaf_size, 0, (char *)0, 0},
  {(char *)"is_leaf_node", __pyx_getprop_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_is_leaf_node, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode = {
  PyVarObject_HEAD_INIT(0, 0)
  "pandas._libs.interval.Float64ClosedBothIntervalNode", /*tp_name*/
  sizeof(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_5__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  "Non-terminal node for an IntervalTree\n\n    Categorizes intervals by those that fall to the left, those that fall to\n    the right, and those that overlap with the pivot.\n    ", /*tp_doc*/
  __pyx_tp_traverse_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode, /*tp_traverse*/
  __pyx_tp_clear_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
};
static struct __pyx_vtabstruct_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode __pyx_vtable_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode;

static PyObject *__pyx_tp_new_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode *)o);
  p->__pyx_vtab = __pyx_vtabptr_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode;
  p->left_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode *)Py_None); Py_INCREF(Py_None);
  p->right_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode *)Py_None); Py_INCREF(Py_None);
  p->center_left_values.data = NULL;
  p->center_left_values.memview = NULL;
  p->center_right_values.data = NULL;
  p->center_right_values.memview = NULL;
  p->left.data = NULL;
  p->left.memview = NULL;
  p->right.data = NULL;
  p->right.memview = NULL;
  p->center_left_indices.data = NULL;
  p->center_left_indices.memview = NULL;
  p->center_right_indices.data = NULL;
  p->center_right_indices.memview = NULL;
  p->indices.data = NULL;
  p->indices.memview = NULL;
  return o;
}

static void __pyx_tp_dealloc_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode(PyObject *o) {
  struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode *p = (struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->left_node);
  Py_CLEAR(p->right_node);
  __PYX_XDEC_MEMVIEW(&p->center_left_values, 1);
  __PYX_XDEC_MEMVIEW(&p->center_right_values, 1);
  __PYX_XDEC_MEMVIEW(&p->left, 1);
  __PYX_XDEC_MEMVIEW(&p->right, 1);
  __PYX_XDEC_MEMVIEW(&p->center_left_indices, 1);
  __PYX_XDEC_MEMVIEW(&p->center_right_indices, 1);
  __PYX_XDEC_MEMVIEW(&p->indices, 1);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode *p = (struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode *)o;
  if (p->left_node) {
    e = (*v)(((PyObject *)p->left_node), a); if (e) return e;
  }
  if (p->right_node) {
    e = (*v)(((PyObject *)p->right_node), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode *p = (struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode *)o;
  tmp = ((PyObject*)p->left_node);
  p->left_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->right_node);
  p->right_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_pivot(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_5pivot_1__get__(o);
}

static PyObject *__pyx_getprop_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_n_elements(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_10n_elements_1__get__(o);
}

static PyObject *__pyx_getprop_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_n_center(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_8n_center_1__get__(o);
}

static PyObject *__pyx_getprop_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_leaf_size(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_9leaf_size_1__get__(o);
}

static PyObject *__pyx_getprop_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_is_leaf_node(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_12is_leaf_node_1__get__(o);
}

static PyMethodDef __pyx_methods_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode[] = {
  {"counts", (PyCFunction)__pyx_pw_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_7counts, METH_NOARGS, __pyx_doc_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_6counts},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_21__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_23__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode[] = {
  {(char *)"pivot", __pyx_getprop_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_pivot, 0, (char *)0, 0},
  {(char *)"n_elements", __pyx_getprop_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_n_elements, 0, (char *)0, 0},
  {(char *)"n_center", __pyx_getprop_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_n_center, 0, (char *)0, 0},
  {(char *)"leaf_size", __pyx_getprop_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_leaf_size, 0, (char *)0, 0},
  {(char *)"is_leaf_node", __pyx_getprop_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_is_leaf_node, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode = {
  PyVarObject_HEAD_INIT(0, 0)
  "pandas._libs.interval.Float64ClosedNeitherIntervalNode", /*tp_name*/
  sizeof(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_5__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  "Non-terminal node for an IntervalTree\n\n    Categorizes intervals by those that fall to the left, those that fall to\n    the right, and those that overlap with the pivot.\n    ", /*tp_doc*/
  __pyx_tp_traverse_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode, /*tp_traverse*/
  __pyx_tp_clear_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
};
static struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode __pyx_vtable_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode;

static PyObject *__pyx_tp_new_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode *)o);
  p->__pyx_vtab = __pyx_vtabptr_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode;
  p->left_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode *)Py_None); Py_INCREF(Py_None);
  p->right_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode *)Py_None); Py_INCREF(Py_None);
  p->center_left_values.data = NULL;
  p->center_left_values.memview = NULL;
  p->center_right_values.data = NULL;
  p->center_right_values.memview = NULL;
  p->left.data = NULL;
  p->left.memview = NULL;
  p->right.data = NULL;
  p->right.memview = NULL;
  p->center_left_indices.data = NULL;
  p->center_left_indices.memview = NULL;
  p->center_right_indices.data = NULL;
  p->center_right_indices.memview = NULL;
  p->indices.data = NULL;
  p->indices.memview = NULL;
  return o;
}

static void __pyx_tp_dealloc_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode(PyObject *o) {
  struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode *p = (struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->left_node);
  Py_CLEAR(p->right_node);
  __PYX_XDEC_MEMVIEW(&p->center_left_values, 1);
  __PYX_XDEC_MEMVIEW(&p->center_right_values, 1);
  __PYX_XDEC_MEMVIEW(&p->left, 1);
  __PYX_XDEC_MEMVIEW(&p->right, 1);
  __PYX_XDEC_MEMVIEW(&p->center_left_indices, 1);
  __PYX_XDEC_MEMVIEW(&p->center_right_indices, 1);
  __PYX_XDEC_MEMVIEW(&p->indices, 1);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode *p = (struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode *)o;
  if (p->left_node) {
    e = (*v)(((PyObject *)p->left_node), a); if (e) return e;
  }
  if (p->right_node) {
    e = (*v)(((PyObject *)p->right_node), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode *p = (struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode *)o;
  tmp = ((PyObject*)p->left_node);
  p->left_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->right_node);
  p->right_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_pivot(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_5pivot_1__get__(o);
}

static PyObject *__pyx_getprop_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_n_elements(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_10n_elements_1__get__(o);
}

static PyObject *__pyx_getprop_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_n_center(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_8n_center_1__get__(o);
}

static PyObject *__pyx_getprop_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_leaf_size(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_9leaf_size_1__get__(o);
}

static PyObject *__pyx_getprop_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_is_leaf_node(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_12is_leaf_node_1__get__(o);
}

static PyMethodDef __pyx_methods_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode[] = {
  {"counts", (PyCFunction)__pyx_pw_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_7counts, METH_NOARGS, __pyx_doc_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_6counts},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_21__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_23__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode[] = {
  {(char *)"pivot", __pyx_getprop_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_pivot, 0, (char *)0, 0},
  {(char *)"n_elements", __pyx_getprop_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_n_elements, 0, (char *)0, 0},
  {(char *)"n_center", __pyx_getprop_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_n_center, 0, (char *)0, 0},
  {(char *)"leaf_size", __pyx_getprop_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_leaf_size, 0, (char *)0, 0},
  {(char *)"is_leaf_node", __pyx_getprop_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_is_leaf_node, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode = {
  PyVarObject_HEAD_INIT(0, 0)
  "pandas._libs.interval.Int32ClosedLeftIntervalNode", /*tp_name*/
  sizeof(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_5__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  "Non-terminal node for an IntervalTree\n\n    Categorizes intervals by those that fall to the left, those that fall to\n    the right, and those that overlap with the pivot.\n    ", /*tp_doc*/
  __pyx_tp_traverse_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode, /*tp_traverse*/
  __pyx_tp_clear_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
};
static struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode __pyx_vtable_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode;

static PyObject *__pyx_tp_new_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode *)o);
  p->__pyx_vtab = __pyx_vtabptr_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode;
  p->left_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode *)Py_None); Py_INCREF(Py_None);
  p->right_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode *)Py_None); Py_INCREF(Py_None);
  p->center_left_values.data = NULL;
  p->center_left_values.memview = NULL;
  p->center_right_values.data = NULL;
  p->center_right_values.memview = NULL;
  p->left.data = NULL;
  p->left.memview = NULL;
  p->right.data = NULL;
  p->right.memview = NULL;
  p->center_left_indices.data = NULL;
  p->center_left_indices.memview = NULL;
  p->center_right_indices.data = NULL;
  p->center_right_indices.memview = NULL;
  p->indices.data = NULL;
  p->indices.memview = NULL;
  return o;
}

static void __pyx_tp_dealloc_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode(PyObject *o) {
  struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode *p = (struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->left_node);
  Py_CLEAR(p->right_node);
  __PYX_XDEC_MEMVIEW(&p->center_left_values, 1);
  __PYX_XDEC_MEMVIEW(&p->center_right_values, 1);
  __PYX_XDEC_MEMVIEW(&p->left, 1);
  __PYX_XDEC_MEMVIEW(&p->right, 1);
  __PYX_XDEC_MEMVIEW(&p->center_left_indices, 1);
  __PYX_XDEC_MEMVIEW(&p->center_right_indices, 1);
  __PYX_XDEC_MEMVIEW(&p->indices, 1);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode *p = (struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode *)o;
  if (p->left_node) {
    e = (*v)(((PyObject *)p->left_node), a); if (e) return e;
  }
  if (p->right_node) {
    e = (*v)(((PyObject *)p->right_node), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode *p = (struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode *)o;
  tmp = ((PyObject*)p->left_node);
  p->left_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->right_node);
  p->right_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_pivot(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_5pivot_1__get__(o);
}

static PyObject *__pyx_getprop_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_n_elements(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_10n_elements_1__get__(o);
}

static PyObject *__pyx_getprop_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_n_center(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_8n_center_1__get__(o);
}

static PyObject *__pyx_getprop_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_leaf_size(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_9leaf_size_1__get__(o);
}

static PyObject *__pyx_getprop_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_is_leaf_node(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_12is_leaf_node_1__get__(o);
}

static PyMethodDef __pyx_methods_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode[] = {
  {"counts", (PyCFunction)__pyx_pw_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_7counts, METH_NOARGS, __pyx_doc_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_6counts},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_21__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_23__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode[] = {
  {(char *)"pivot", __pyx_getprop_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_pivot, 0, (char *)0, 0},
  {(char *)"n_elements", __pyx_getprop_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_n_elements, 0, (char *)0, 0},
  {(char *)"n_center", __pyx_getprop_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_n_center, 0, (char *)0, 0},
  {(char *)"leaf_size", __pyx_getprop_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_leaf_size, 0, (char *)0, 0},
  {(char *)"is_leaf_node", __pyx_getprop_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_is_leaf_node, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode = {
  PyVarObject_HEAD_INIT(0, 0)
  "pandas._libs.interval.Int32ClosedRightIntervalNode", /*tp_name*/
  sizeof(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_5__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  "Non-terminal node for an IntervalTree\n\n    Categorizes intervals by those that fall to the left, those that fall to\n    the right, and those that overlap with the pivot.\n    ", /*tp_doc*/
  __pyx_tp_traverse_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode, /*tp_traverse*/
  __pyx_tp_clear_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
};
static struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode __pyx_vtable_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode;

static PyObject *__pyx_tp_new_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode *)o);
  p->__pyx_vtab = __pyx_vtabptr_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode;
  p->left_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode *)Py_None); Py_INCREF(Py_None);
  p->right_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode *)Py_None); Py_INCREF(Py_None);
  p->center_left_values.data = NULL;
  p->center_left_values.memview = NULL;
  p->center_right_values.data = NULL;
  p->center_right_values.memview = NULL;
  p->left.data = NULL;
  p->left.memview = NULL;
  p->right.data = NULL;
  p->right.memview = NULL;
  p->center_left_indices.data = NULL;
  p->center_left_indices.memview = NULL;
  p->center_right_indices.data = NULL;
  p->center_right_indices.memview = NULL;
  p->indices.data = NULL;
  p->indices.memview = NULL;
  return o;
}

static void __pyx_tp_dealloc_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode(PyObject *o) {
  struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode *p = (struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->left_node);
  Py_CLEAR(p->right_node);
  __PYX_XDEC_MEMVIEW(&p->center_left_values, 1);
  __PYX_XDEC_MEMVIEW(&p->center_right_values, 1);
  __PYX_XDEC_MEMVIEW(&p->left, 1);
  __PYX_XDEC_MEMVIEW(&p->right, 1);
  __PYX_XDEC_MEMVIEW(&p->center_left_indices, 1);
  __PYX_XDEC_MEMVIEW(&p->center_right_indices, 1);
  __PYX_XDEC_MEMVIEW(&p->indices, 1);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode *p = (struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode *)o;
  if (p->left_node) {
    e = (*v)(((PyObject *)p->left_node), a); if (e) return e;
  }
  if (p->right_node) {
    e = (*v)(((PyObject *)p->right_node), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode *p = (struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode *)o;
  tmp = ((PyObject*)p->left_node);
  p->left_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->right_node);
  p->right_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_pivot(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_5pivot_1__get__(o);
}

static PyObject *__pyx_getprop_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_n_elements(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_10n_elements_1__get__(o);
}

static PyObject *__pyx_getprop_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_n_center(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_8n_center_1__get__(o);
}

static PyObject *__pyx_getprop_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_leaf_size(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_9leaf_size_1__get__(o);
}

static PyObject *__pyx_getprop_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_is_leaf_node(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_12is_leaf_node_1__get__(o);
}

static PyMethodDef __pyx_methods_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode[] = {
  {"counts", (PyCFunction)__pyx_pw_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_7counts, METH_NOARGS, __pyx_doc_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_6counts},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_21__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_23__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode[] = {
  {(char *)"pivot", __pyx_getprop_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_pivot, 0, (char *)0, 0},
  {(char *)"n_elements", __pyx_getprop_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_n_elements, 0, (char *)0, 0},
  {(char *)"n_center", __pyx_getprop_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_n_center, 0, (char *)0, 0},
  {(char *)"leaf_size", __pyx_getprop_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_leaf_size, 0, (char *)0, 0},
  {(char *)"is_leaf_node", __pyx_getprop_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_is_leaf_node, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode = {
  PyVarObject_HEAD_INIT(0, 0)
  "pandas._libs.interval.Int32ClosedBothIntervalNode", /*tp_name*/
  sizeof(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_5__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  "Non-terminal node for an IntervalTree\n\n    Categorizes intervals by those that fall to the left, those that fall to\n    the right, and those that overlap with the pivot.\n    ", /*tp_doc*/
  __pyx_tp_traverse_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode, /*tp_traverse*/
  __pyx_tp_clear_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
};
static struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode __pyx_vtable_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode;

static PyObject *__pyx_tp_new_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode *)o);
  p->__pyx_vtab = __pyx_vtabptr_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode;
  p->left_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode *)Py_None); Py_INCREF(Py_None);
  p->right_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode *)Py_None); Py_INCREF(Py_None);
  p->center_left_values.data = NULL;
  p->center_left_values.memview = NULL;
  p->center_right_values.data = NULL;
  p->center_right_values.memview = NULL;
  p->left.data = NULL;
  p->left.memview = NULL;
  p->right.data = NULL;
  p->right.memview = NULL;
  p->center_left_indices.data = NULL;
  p->center_left_indices.memview = NULL;
  p->center_right_indices.data = NULL;
  p->center_right_indices.memview = NULL;
  p->indices.data = NULL;
  p->indices.memview = NULL;
  return o;
}

static void __pyx_tp_dealloc_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode(PyObject *o) {
  struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode *p = (struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->left_node);
  Py_CLEAR(p->right_node);
  __PYX_XDEC_MEMVIEW(&p->center_left_values, 1);
  __PYX_XDEC_MEMVIEW(&p->center_right_values, 1);
  __PYX_XDEC_MEMVIEW(&p->left, 1);
  __PYX_XDEC_MEMVIEW(&p->right, 1);
  __PYX_XDEC_MEMVIEW(&p->center_left_indices, 1);
  __PYX_XDEC_MEMVIEW(&p->center_right_indices, 1);
  __PYX_XDEC_MEMVIEW(&p->indices, 1);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode *p = (struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode *)o;
  if (p->left_node) {
    e = (*v)(((PyObject *)p->left_node), a); if (e) return e;
  }
  if (p->right_node) {
    e = (*v)(((PyObject *)p->right_node), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode *p = (struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode *)o;
  tmp = ((PyObject*)p->left_node);
  p->left_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->right_node);
  p->right_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_pivot(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_5pivot_1__get__(o);
}

static PyObject *__pyx_getprop_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_n_elements(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_10n_elements_1__get__(o);
}

static PyObject *__pyx_getprop_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_n_center(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_8n_center_1__get__(o);
}

static PyObject *__pyx_getprop_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_leaf_size(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_9leaf_size_1__get__(o);
}

static PyObject *__pyx_getprop_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_is_leaf_node(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_12is_leaf_node_1__get__(o);
}

static PyMethodDef __pyx_methods_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode[] = {
  {"counts", (PyCFunction)__pyx_pw_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_7counts, METH_NOARGS, __pyx_doc_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_6counts},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_21__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_23__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode[] = {
  {(char *)"pivot", __pyx_getprop_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_pivot, 0, (char *)0, 0},
  {(char *)"n_elements", __pyx_getprop_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_n_elements, 0, (char *)0, 0},
  {(char *)"n_center", __pyx_getprop_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_n_center, 0, (char *)0, 0},
  {(char *)"leaf_size", __pyx_getprop_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_leaf_size, 0, (char *)0, 0},
  {(char *)"is_leaf_node", __pyx_getprop_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_is_leaf_node, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode = {
  PyVarObject_HEAD_INIT(0, 0)
  "pandas._libs.interval.Int32ClosedNeitherIntervalNode", /*tp_name*/
  sizeof(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_5__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  "Non-terminal node for an IntervalTree\n\n    Categorizes intervals by those that fall to the left, those that fall to\n    the right, and those that overlap with the pivot.\n    ", /*tp_doc*/
  __pyx_tp_traverse_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode, /*tp_traverse*/
  __pyx_tp_clear_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
};
static struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode __pyx_vtable_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode;

static PyObject *__pyx_tp_new_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode *)o);
  p->__pyx_vtab = __pyx_vtabptr_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode;
  p->left_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode *)Py_None); Py_INCREF(Py_None);
  p->right_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode *)Py_None); Py_INCREF(Py_None);
  p->center_left_values.data = NULL;
  p->center_left_values.memview = NULL;
  p->center_right_values.data = NULL;
  p->center_right_values.memview = NULL;
  p->left.data = NULL;
  p->left.memview = NULL;
  p->right.data = NULL;
  p->right.memview = NULL;
  p->center_left_indices.data = NULL;
  p->center_left_indices.memview = NULL;
  p->center_right_indices.data = NULL;
  p->center_right_indices.memview = NULL;
  p->indices.data = NULL;
  p->indices.memview = NULL;
  return o;
}

static void __pyx_tp_dealloc_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode(PyObject *o) {
  struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode *p = (struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->left_node);
  Py_CLEAR(p->right_node);
  __PYX_XDEC_MEMVIEW(&p->center_left_values, 1);
  __PYX_XDEC_MEMVIEW(&p->center_right_values, 1);
  __PYX_XDEC_MEMVIEW(&p->left, 1);
  __PYX_XDEC_MEMVIEW(&p->right, 1);
  __PYX_XDEC_MEMVIEW(&p->center_left_indices, 1);
  __PYX_XDEC_MEMVIEW(&p->center_right_indices, 1);
  __PYX_XDEC_MEMVIEW(&p->indices, 1);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode *p = (struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode *)o;
  if (p->left_node) {
    e = (*v)(((PyObject *)p->left_node), a); if (e) return e;
  }
  if (p->right_node) {
    e = (*v)(((PyObject *)p->right_node), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode *p = (struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode *)o;
  tmp = ((PyObject*)p->left_node);
  p->left_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->right_node);
  p->right_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_pivot(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_5pivot_1__get__(o);
}

static PyObject *__pyx_getprop_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_n_elements(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_10n_elements_1__get__(o);
}

static PyObject *__pyx_getprop_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_n_center(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_8n_center_1__get__(o);
}

static PyObject *__pyx_getprop_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_leaf_size(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_9leaf_size_1__get__(o);
}

static PyObject *__pyx_getprop_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_is_leaf_node(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_12is_leaf_node_1__get__(o);
}

static PyMethodDef __pyx_methods_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode[] = {
  {"counts", (PyCFunction)__pyx_pw_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_7counts, METH_NOARGS, __pyx_doc_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_6counts},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_21__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_23__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode[] = {
  {(char *)"pivot", __pyx_getprop_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_pivot, 0, (char *)0, 0},
  {(char *)"n_elements", __pyx_getprop_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_n_elements, 0, (char *)0, 0},
  {(char *)"n_center", __pyx_getprop_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_n_center, 0, (char *)0, 0},
  {(char *)"leaf_size", __pyx_getprop_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_leaf_size, 0, (char *)0, 0},
  {(char *)"is_leaf_node", __pyx_getprop_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_is_leaf_node, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode = {
  PyVarObject_HEAD_INIT(0, 0)
  "pandas._libs.interval.Int64ClosedLeftIntervalNode", /*tp_name*/
  sizeof(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_5__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  "Non-terminal node for an IntervalTree\n\n    Categorizes intervals by those that fall to the left, those that fall to\n    the right, and those that overlap with the pivot.\n    ", /*tp_doc*/
  __pyx_tp_traverse_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode, /*tp_traverse*/
  __pyx_tp_clear_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
};
static struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode __pyx_vtable_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode;

static PyObject *__pyx_tp_new_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode *)o);
  p->__pyx_vtab = __pyx_vtabptr_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode;
  p->left_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode *)Py_None); Py_INCREF(Py_None);
  p->right_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode *)Py_None); Py_INCREF(Py_None);
  p->center_left_values.data = NULL;
  p->center_left_values.memview = NULL;
  p->center_right_values.data = NULL;
  p->center_right_values.memview = NULL;
  p->left.data = NULL;
  p->left.memview = NULL;
  p->right.data = NULL;
  p->right.memview = NULL;
  p->center_left_indices.data = NULL;
  p->center_left_indices.memview = NULL;
  p->center_right_indices.data = NULL;
  p->center_right_indices.memview = NULL;
  p->indices.data = NULL;
  p->indices.memview = NULL;
  return o;
}

static void __pyx_tp_dealloc_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode(PyObject *o) {
  struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode *p = (struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->left_node);
  Py_CLEAR(p->right_node);
  __PYX_XDEC_MEMVIEW(&p->center_left_values, 1);
  __PYX_XDEC_MEMVIEW(&p->center_right_values, 1);
  __PYX_XDEC_MEMVIEW(&p->left, 1);
  __PYX_XDEC_MEMVIEW(&p->right, 1);
  __PYX_XDEC_MEMVIEW(&p->center_left_indices, 1);
  __PYX_XDEC_MEMVIEW(&p->center_right_indices, 1);
  __PYX_XDEC_MEMVIEW(&p->indices, 1);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode *p = (struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode *)o;
  if (p->left_node) {
    e = (*v)(((PyObject *)p->left_node), a); if (e) return e;
  }
  if (p->right_node) {
    e = (*v)(((PyObject *)p->right_node), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode *p = (struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode *)o;
  tmp = ((PyObject*)p->left_node);
  p->left_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->right_node);
  p->right_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_pivot(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_5pivot_1__get__(o);
}

static PyObject *__pyx_getprop_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_n_elements(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_10n_elements_1__get__(o);
}

static PyObject *__pyx_getprop_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_n_center(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_8n_center_1__get__(o);
}

static PyObject *__pyx_getprop_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_leaf_size(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_9leaf_size_1__get__(o);
}

static PyObject *__pyx_getprop_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_is_leaf_node(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_12is_leaf_node_1__get__(o);
}

static PyMethodDef __pyx_methods_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode[] = {
  {"counts", (PyCFunction)__pyx_pw_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_7counts, METH_NOARGS, __pyx_doc_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_6counts},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_21__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_23__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode[] = {
  {(char *)"pivot", __pyx_getprop_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_pivot, 0, (char *)0, 0},
  {(char *)"n_elements", __pyx_getprop_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_n_elements, 0, (char *)0, 0},
  {(char *)"n_center", __pyx_getprop_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_n_center, 0, (char *)0, 0},
  {(char *)"leaf_size", __pyx_getprop_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_leaf_size, 0, (char *)0, 0},
  {(char *)"is_leaf_node", __pyx_getprop_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_is_leaf_node, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode = {
  PyVarObject_HEAD_INIT(0, 0)
  "pandas._libs.interval.Int64ClosedRightIntervalNode", /*tp_name*/
  sizeof(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_5__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  "Non-terminal node for an IntervalTree\n\n    Categorizes intervals by those that fall to the left, those that fall to\n    the right, and those that overlap with the pivot.\n    ", /*tp_doc*/
  __pyx_tp_traverse_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode, /*tp_traverse*/
  __pyx_tp_clear_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
};
static struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode __pyx_vtable_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode;

static PyObject *__pyx_tp_new_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode *)o);
  p->__pyx_vtab = __pyx_vtabptr_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode;
  p->left_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode *)Py_None); Py_INCREF(Py_None);
  p->right_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode *)Py_None); Py_INCREF(Py_None);
  p->center_left_values.data = NULL;
  p->center_left_values.memview = NULL;
  p->center_right_values.data = NULL;
  p->center_right_values.memview = NULL;
  p->left.data = NULL;
  p->left.memview = NULL;
  p->right.data = NULL;
  p->right.memview = NULL;
  p->center_left_indices.data = NULL;
  p->center_left_indices.memview = NULL;
  p->center_right_indices.data = NULL;
  p->center_right_indices.memview = NULL;
  p->indices.data = NULL;
  p->indices.memview = NULL;
  return o;
}

static void __pyx_tp_dealloc_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode(PyObject *o) {
  struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode *p = (struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->left_node);
  Py_CLEAR(p->right_node);
  __PYX_XDEC_MEMVIEW(&p->center_left_values, 1);
  __PYX_XDEC_MEMVIEW(&p->center_right_values, 1);
  __PYX_XDEC_MEMVIEW(&p->left, 1);
  __PYX_XDEC_MEMVIEW(&p->right, 1);
  __PYX_XDEC_MEMVIEW(&p->center_left_indices, 1);
  __PYX_XDEC_MEMVIEW(&p->center_right_indices, 1);
  __PYX_XDEC_MEMVIEW(&p->indices, 1);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode *p = (struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode *)o;
  if (p->left_node) {
    e = (*v)(((PyObject *)p->left_node), a); if (e) return e;
  }
  if (p->right_node) {
    e = (*v)(((PyObject *)p->right_node), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode *p = (struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode *)o;
  tmp = ((PyObject*)p->left_node);
  p->left_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->right_node);
  p->right_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_pivot(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_5pivot_1__get__(o);
}

static PyObject *__pyx_getprop_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_n_elements(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_10n_elements_1__get__(o);
}

static PyObject *__pyx_getprop_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_n_center(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_8n_center_1__get__(o);
}

static PyObject *__pyx_getprop_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_leaf_size(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_9leaf_size_1__get__(o);
}

static PyObject *__pyx_getprop_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_is_leaf_node(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_12is_leaf_node_1__get__(o);
}

static PyMethodDef __pyx_methods_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode[] = {
  {"counts", (PyCFunction)__pyx_pw_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_7counts, METH_NOARGS, __pyx_doc_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_6counts},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_21__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_23__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode[] = {
  {(char *)"pivot", __pyx_getprop_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_pivot, 0, (char *)0, 0},
  {(char *)"n_elements", __pyx_getprop_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_n_elements, 0, (char *)0, 0},
  {(char *)"n_center", __pyx_getprop_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_n_center, 0, (char *)0, 0},
  {(char *)"leaf_size", __pyx_getprop_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_leaf_size, 0, (char *)0, 0},
  {(char *)"is_leaf_node", __pyx_getprop_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_is_leaf_node, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode = {
  PyVarObject_HEAD_INIT(0, 0)
  "pandas._libs.interval.Int64ClosedBothIntervalNode", /*tp_name*/
  sizeof(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_5__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  "Non-terminal node for an IntervalTree\n\n    Categorizes intervals by those that fall to the left, those that fall to\n    the right, and those that overlap with the pivot.\n    ", /*tp_doc*/
  __pyx_tp_traverse_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode, /*tp_traverse*/
  __pyx_tp_clear_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
};
static struct __pyx_vtabstruct_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode __pyx_vtable_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode;

static PyObject *__pyx_tp_new_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode *)o);
  p->__pyx_vtab = __pyx_vtabptr_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode;
  p->left_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode *)Py_None); Py_INCREF(Py_None);
  p->right_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode *)Py_None); Py_INCREF(Py_None);
  p->center_left_values.data = NULL;
  p->center_left_values.memview = NULL;
  p->center_right_values.data = NULL;
  p->center_right_values.memview = NULL;
  p->left.data = NULL;
  p->left.memview = NULL;
  p->right.data = NULL;
  p->right.memview = NULL;
  p->center_left_indices.data = NULL;
  p->center_left_indices.memview = NULL;
  p->center_right_indices.data = NULL;
  p->center_right_indices.memview = NULL;
  p->indices.data = NULL;
  p->indices.memview = NULL;
  return o;
}

static void __pyx_tp_dealloc_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode(PyObject *o) {
  struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode *p = (struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->left_node);
  Py_CLEAR(p->right_node);
  __PYX_XDEC_MEMVIEW(&p->center_left_values, 1);
  __PYX_XDEC_MEMVIEW(&p->center_right_values, 1);
  __PYX_XDEC_MEMVIEW(&p->left, 1);
  __PYX_XDEC_MEMVIEW(&p->right, 1);
  __PYX_XDEC_MEMVIEW(&p->center_left_indices, 1);
  __PYX_XDEC_MEMVIEW(&p->center_right_indices, 1);
  __PYX_XDEC_MEMVIEW(&p->indices, 1);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode *p = (struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode *)o;
  if (p->left_node) {
    e = (*v)(((PyObject *)p->left_node), a); if (e) return e;
  }
  if (p->right_node) {
    e = (*v)(((PyObject *)p->right_node), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode *p = (struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode *)o;
  tmp = ((PyObject*)p->left_node);
  p->left_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->right_node);
  p->right_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_pivot(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_5pivot_1__get__(o);
}

static PyObject *__pyx_getprop_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_n_elements(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_10n_elements_1__get__(o);
}

static PyObject *__pyx_getprop_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_n_center(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_8n_center_1__get__(o);
}

static PyObject *__pyx_getprop_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_leaf_size(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_9leaf_size_1__get__(o);
}

static PyObject *__pyx_getprop_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_is_leaf_node(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_12is_leaf_node_1__get__(o);
}

static PyMethodDef __pyx_methods_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode[] = {
  {"counts", (PyCFunction)__pyx_pw_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_7counts, METH_NOARGS, __pyx_doc_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_6counts},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_21__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_23__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode[] = {
  {(char *)"pivot", __pyx_getprop_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_pivot, 0, (char *)0, 0},
  {(char *)"n_elements", __pyx_getprop_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_n_elements, 0, (char *)0, 0},
  {(char *)"n_center", __pyx_getprop_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_n_center, 0, (char *)0, 0},
  {(char *)"leaf_size", __pyx_getprop_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_leaf_size, 0, (char *)0, 0},
  {(char *)"is_leaf_node", __pyx_getprop_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_is_leaf_node, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode = {
  PyVarObject_HEAD_INIT(0, 0)
  "pandas._libs.interval.Int64ClosedNeitherIntervalNode", /*tp_name*/
  sizeof(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_5__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  "Non-terminal node for an IntervalTree\n\n    Categorizes intervals by those that fall to the left, those that fall to\n    the right, and those that overlap with the pivot.\n    ", /*tp_doc*/
  __pyx_tp_traverse_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode, /*tp_traverse*/
  __pyx_tp_clear_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
};
static struct __pyx_vtabstruct_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode __pyx_vtable_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode;

static PyObject *__pyx_tp_new_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode *)o);
  p->__pyx_vtab = __pyx_vtabptr_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode;
  p->left_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode *)Py_None); Py_INCREF(Py_None);
  p->right_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode *)Py_None); Py_INCREF(Py_None);
  p->center_left_values.data = NULL;
  p->center_left_values.memview = NULL;
  p->center_right_values.data = NULL;
  p->center_right_values.memview = NULL;
  p->left.data = NULL;
  p->left.memview = NULL;
  p->right.data = NULL;
  p->right.memview = NULL;
  p->center_left_indices.data = NULL;
  p->center_left_indices.memview = NULL;
  p->center_right_indices.data = NULL;
  p->center_right_indices.memview = NULL;
  p->indices.data = NULL;
  p->indices.memview = NULL;
  return o;
}

static void __pyx_tp_dealloc_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode(PyObject *o) {
  struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode *p = (struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->left_node);
  Py_CLEAR(p->right_node);
  __PYX_XDEC_MEMVIEW(&p->center_left_values, 1);
  __PYX_XDEC_MEMVIEW(&p->center_right_values, 1);
  __PYX_XDEC_MEMVIEW(&p->left, 1);
  __PYX_XDEC_MEMVIEW(&p->right, 1);
  __PYX_XDEC_MEMVIEW(&p->center_left_indices, 1);
  __PYX_XDEC_MEMVIEW(&p->center_right_indices, 1);
  __PYX_XDEC_MEMVIEW(&p->indices, 1);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode *p = (struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode *)o;
  if (p->left_node) {
    e = (*v)(((PyObject *)p->left_node), a); if (e) return e;
  }
  if (p->right_node) {
    e = (*v)(((PyObject *)p->right_node), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode *p = (struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode *)o;
  tmp = ((PyObject*)p->left_node);
  p->left_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->right_node);
  p->right_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_pivot(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_5pivot_1__get__(o);
}

static PyObject *__pyx_getprop_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_n_elements(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_10n_elements_1__get__(o);
}

static PyObject *__pyx_getprop_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_n_center(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_8n_center_1__get__(o);
}

static PyObject *__pyx_getprop_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_leaf_size(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_9leaf_size_1__get__(o);
}

static PyObject *__pyx_getprop_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_is_leaf_node(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_12is_leaf_node_1__get__(o);
}

static PyMethodDef __pyx_methods_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode[] = {
  {"counts", (PyCFunction)__pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_7counts, METH_NOARGS, __pyx_doc_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_6counts},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_21__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_23__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode[] = {
  {(char *)"pivot", __pyx_getprop_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_pivot, 0, (char *)0, 0},
  {(char *)"n_elements", __pyx_getprop_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_n_elements, 0, (char *)0, 0},
  {(char *)"n_center", __pyx_getprop_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_n_center, 0, (char *)0, 0},
  {(char *)"leaf_size", __pyx_getprop_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_leaf_size, 0, (char *)0, 0},
  {(char *)"is_leaf_node", __pyx_getprop_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_is_leaf_node, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode = {
  PyVarObject_HEAD_INIT(0, 0)
  "pandas._libs.interval.Uint64ClosedLeftIntervalNode", /*tp_name*/
  sizeof(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_5__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  "Non-terminal node for an IntervalTree\n\n    Categorizes intervals by those that fall to the left, those that fall to\n    the right, and those that overlap with the pivot.\n    ", /*tp_doc*/
  __pyx_tp_traverse_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode, /*tp_traverse*/
  __pyx_tp_clear_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
};
static struct __pyx_vtabstruct_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode __pyx_vtable_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode;

static PyObject *__pyx_tp_new_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode *)o);
  p->__pyx_vtab = __pyx_vtabptr_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode;
  p->left_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode *)Py_None); Py_INCREF(Py_None);
  p->right_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode *)Py_None); Py_INCREF(Py_None);
  p->center_left_values.data = NULL;
  p->center_left_values.memview = NULL;
  p->center_right_values.data = NULL;
  p->center_right_values.memview = NULL;
  p->left.data = NULL;
  p->left.memview = NULL;
  p->right.data = NULL;
  p->right.memview = NULL;
  p->center_left_indices.data = NULL;
  p->center_left_indices.memview = NULL;
  p->center_right_indices.data = NULL;
  p->center_right_indices.memview = NULL;
  p->indices.data = NULL;
  p->indices.memview = NULL;
  return o;
}

static void __pyx_tp_dealloc_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode(PyObject *o) {
  struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode *p = (struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->left_node);
  Py_CLEAR(p->right_node);
  __PYX_XDEC_MEMVIEW(&p->center_left_values, 1);
  __PYX_XDEC_MEMVIEW(&p->center_right_values, 1);
  __PYX_XDEC_MEMVIEW(&p->left, 1);
  __PYX_XDEC_MEMVIEW(&p->right, 1);
  __PYX_XDEC_MEMVIEW(&p->center_left_indices, 1);
  __PYX_XDEC_MEMVIEW(&p->center_right_indices, 1);
  __PYX_XDEC_MEMVIEW(&p->indices, 1);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode *p = (struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode *)o;
  if (p->left_node) {
    e = (*v)(((PyObject *)p->left_node), a); if (e) return e;
  }
  if (p->right_node) {
    e = (*v)(((PyObject *)p->right_node), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode *p = (struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode *)o;
  tmp = ((PyObject*)p->left_node);
  p->left_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->right_node);
  p->right_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_pivot(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_5pivot_1__get__(o);
}

static PyObject *__pyx_getprop_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_n_elements(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_10n_elements_1__get__(o);
}

static PyObject *__pyx_getprop_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_n_center(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_8n_center_1__get__(o);
}

static PyObject *__pyx_getprop_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_leaf_size(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_9leaf_size_1__get__(o);
}

static PyObject *__pyx_getprop_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_is_leaf_node(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_12is_leaf_node_1__get__(o);
}

static PyMethodDef __pyx_methods_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode[] = {
  {"counts", (PyCFunction)__pyx_pw_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_7counts, METH_NOARGS, __pyx_doc_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_6counts},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_21__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_23__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode[] = {
  {(char *)"pivot", __pyx_getprop_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_pivot, 0, (char *)0, 0},
  {(char *)"n_elements", __pyx_getprop_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_n_elements, 0, (char *)0, 0},
  {(char *)"n_center", __pyx_getprop_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_n_center, 0, (char *)0, 0},
  {(char *)"leaf_size", __pyx_getprop_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_leaf_size, 0, (char *)0, 0},
  {(char *)"is_leaf_node", __pyx_getprop_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_is_leaf_node, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode = {
  PyVarObject_HEAD_INIT(0, 0)
  "pandas._libs.interval.Uint64ClosedRightIntervalNode", /*tp_name*/
  sizeof(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_5__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  "Non-terminal node for an IntervalTree\n\n    Categorizes intervals by those that fall to the left, those that fall to\n    the right, and those that overlap with the pivot.\n    ", /*tp_doc*/
  __pyx_tp_traverse_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode, /*tp_traverse*/
  __pyx_tp_clear_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
};
static struct __pyx_vtabstruct_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode __pyx_vtable_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode;

static PyObject *__pyx_tp_new_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode *)o);
  p->__pyx_vtab = __pyx_vtabptr_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode;
  p->left_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode *)Py_None); Py_INCREF(Py_None);
  p->right_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode *)Py_None); Py_INCREF(Py_None);
  p->center_left_values.data = NULL;
  p->center_left_values.memview = NULL;
  p->center_right_values.data = NULL;
  p->center_right_values.memview = NULL;
  p->left.data = NULL;
  p->left.memview = NULL;
  p->right.data = NULL;
  p->right.memview = NULL;
  p->center_left_indices.data = NULL;
  p->center_left_indices.memview = NULL;
  p->center_right_indices.data = NULL;
  p->center_right_indices.memview = NULL;
  p->indices.data = NULL;
  p->indices.memview = NULL;
  return o;
}

static void __pyx_tp_dealloc_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode(PyObject *o) {
  struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode *p = (struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->left_node);
  Py_CLEAR(p->right_node);
  __PYX_XDEC_MEMVIEW(&p->center_left_values, 1);
  __PYX_XDEC_MEMVIEW(&p->center_right_values, 1);
  __PYX_XDEC_MEMVIEW(&p->left, 1);
  __PYX_XDEC_MEMVIEW(&p->right, 1);
  __PYX_XDEC_MEMVIEW(&p->center_left_indices, 1);
  __PYX_XDEC_MEMVIEW(&p->center_right_indices, 1);
  __PYX_XDEC_MEMVIEW(&p->indices, 1);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode *p = (struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode *)o;
  if (p->left_node) {
    e = (*v)(((PyObject *)p->left_node), a); if (e) return e;
  }
  if (p->right_node) {
    e = (*v)(((PyObject *)p->right_node), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode *p = (struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode *)o;
  tmp = ((PyObject*)p->left_node);
  p->left_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->right_node);
  p->right_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_pivot(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_5pivot_1__get__(o);
}

static PyObject *__pyx_getprop_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_n_elements(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_10n_elements_1__get__(o);
}

static PyObject *__pyx_getprop_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_n_center(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_8n_center_1__get__(o);
}

static PyObject *__pyx_getprop_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_leaf_size(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_9leaf_size_1__get__(o);
}

static PyObject *__pyx_getprop_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_is_leaf_node(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_12is_leaf_node_1__get__(o);
}

static PyMethodDef __pyx_methods_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode[] = {
  {"counts", (PyCFunction)__pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_7counts, METH_NOARGS, __pyx_doc_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_6counts},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_21__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_23__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode[] = {
  {(char *)"pivot", __pyx_getprop_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_pivot, 0, (char *)0, 0},
  {(char *)"n_elements", __pyx_getprop_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_n_elements, 0, (char *)0, 0},
  {(char *)"n_center", __pyx_getprop_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_n_center, 0, (char *)0, 0},
  {(char *)"leaf_size", __pyx_getprop_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_leaf_size, 0, (char *)0, 0},
  {(char *)"is_leaf_node", __pyx_getprop_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_is_leaf_node, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode = {
  PyVarObject_HEAD_INIT(0, 0)
  "pandas._libs.interval.Uint64ClosedBothIntervalNode", /*tp_name*/
  sizeof(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_5__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  "Non-terminal node for an IntervalTree\n\n    Categorizes intervals by those that fall to the left, those that fall to\n    the right, and those that overlap with the pivot.\n    ", /*tp_doc*/
  __pyx_tp_traverse_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode, /*tp_traverse*/
  __pyx_tp_clear_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
};
static struct __pyx_vtabstruct_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode __pyx_vtable_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode;

static PyObject *__pyx_tp_new_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode *)o);
  p->__pyx_vtab = __pyx_vtabptr_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode;
  p->left_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode *)Py_None); Py_INCREF(Py_None);
  p->right_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode *)Py_None); Py_INCREF(Py_None);
  p->center_left_values.data = NULL;
  p->center_left_values.memview = NULL;
  p->center_right_values.data = NULL;
  p->center_right_values.memview = NULL;
  p->left.data = NULL;
  p->left.memview = NULL;
  p->right.data = NULL;
  p->right.memview = NULL;
  p->center_left_indices.data = NULL;
  p->center_left_indices.memview = NULL;
  p->center_right_indices.data = NULL;
  p->center_right_indices.memview = NULL;
  p->indices.data = NULL;
  p->indices.memview = NULL;
  return o;
}

static void __pyx_tp_dealloc_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode(PyObject *o) {
  struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode *p = (struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->left_node);
  Py_CLEAR(p->right_node);
  __PYX_XDEC_MEMVIEW(&p->center_left_values, 1);
  __PYX_XDEC_MEMVIEW(&p->center_right_values, 1);
  __PYX_XDEC_MEMVIEW(&p->left, 1);
  __PYX_XDEC_MEMVIEW(&p->right, 1);
  __PYX_XDEC_MEMVIEW(&p->center_left_indices, 1);
  __PYX_XDEC_MEMVIEW(&p->center_right_indices, 1);
  __PYX_XDEC_MEMVIEW(&p->indices, 1);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode *p = (struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode *)o;
  if (p->left_node) {
    e = (*v)(((PyObject *)p->left_node), a); if (e) return e;
  }
  if (p->right_node) {
    e = (*v)(((PyObject *)p->right_node), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode *p = (struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode *)o;
  tmp = ((PyObject*)p->left_node);
  p->left_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->right_node);
  p->right_node = ((struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_pivot(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_5pivot_1__get__(o);
}

static PyObject *__pyx_getprop_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_n_elements(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_10n_elements_1__get__(o);
}

static PyObject *__pyx_getprop_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_n_center(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_8n_center_1__get__(o);
}

static PyObject *__pyx_getprop_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_leaf_size(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_9leaf_size_1__get__(o);
}

static PyObject *__pyx_getprop_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_is_leaf_node(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_12is_leaf_node_1__get__(o);
}

static PyMethodDef __pyx_methods_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode[] = {
  {"counts", (PyCFunction)__pyx_pw_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_7counts, METH_NOARGS, __pyx_doc_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_6counts},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_21__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_23__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode[] = {
  {(char *)"pivot", __pyx_getprop_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_pivot, 0, (char *)0, 0},
  {(char *)"n_elements", __pyx_getprop_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_n_elements, 0, (char *)0, 0},
  {(char *)"n_center", __pyx_getprop_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_n_center, 0, (char *)0, 0},
  {(char *)"leaf_size", __pyx_getprop_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_leaf_size, 0, (char *)0, 0},
  {(char *)"is_leaf_node", __pyx_getprop_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_is_leaf_node, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode = {
  PyVarObject_HEAD_INIT(0, 0)
  "pandas._libs.interval.Uint64ClosedNeitherIntervalNode", /*tp_name*/
  sizeof(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_5__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  "Non-terminal node for an IntervalTree\n\n    Categorizes intervals by those that fall to the left, those that fall to\n    the right, and those that overlap with the pivot.\n    ", /*tp_doc*/
  __pyx_tp_traverse_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode, /*tp_traverse*/
  __pyx_tp_clear_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
};
static struct __pyx_vtabstruct_array __pyx_vtable_array;

static PyObject *__pyx_tp_new_array(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_array_obj *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_array_obj *)o);
  p->__pyx_vtab = __pyx_vtabptr_array;
  p->mode = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->_format = ((PyObject*)Py_None); Py_INCREF(Py_None);
  if (unlikely(__pyx_array___cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_array(PyObject *o) {
  struct __pyx_array_obj *p = (struct __pyx_array_obj *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && (!PyType_IS_GC(Py_TYPE(o)) || !_PyGC_FINALIZED(o))) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    ++Py_REFCNT(o);
    __pyx_array___dealloc__(o);
    --Py_REFCNT(o);
    PyErr_Restore(etype, eval, etb);
  }
  Py_CLEAR(p->mode);
  Py_CLEAR(p->_format);
  (*Py_TYPE(o)->tp_free)(o);
}
static PyObject *__pyx_sq_item_array(PyObject *o, Py_ssize_t i) {
  PyObject *r;
  PyObject *x = PyInt_FromSsize_t(i); if(!x) return 0;
  r = Py_TYPE(o)->tp_as_mapping->mp_subscript(o, x);
  Py_DECREF(x);
  return r;
}

static int __pyx_mp_ass_subscript_array(PyObject *o, PyObject *i, PyObject *v) {
  if (v) {
    return __pyx_array___setitem__(o, i, v);
  }
  else {
    PyErr_Format(PyExc_NotImplementedError,
      "Subscript deletion not supported by %.200s", Py_TYPE(o)->tp_name);
    return -1;
  }
}

static PyObject *__pyx_tp_getattro_array(PyObject *o, PyObject *n) {
  PyObject *v = __Pyx_PyObject_GenericGetAttr(o, n);
  if (!v && PyErr_ExceptionMatches(PyExc_AttributeError)) {
    PyErr_Clear();
    v = __pyx_array___getattr__(o, n);
  }
  return v;
}

static PyObject *__pyx_getprop___pyx_array_memview(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_5array_7memview_1__get__(o);
}

static PyMethodDef __pyx_methods_array[] = {
  {"__getattr__", (PyCFunction)__pyx_array___getattr__, METH_O|METH_COEXIST, 0},
  {"__reduce_cython__", (PyCFunction)__pyx_pw___pyx_array_1__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw___pyx_array_3__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_array[] = {
  {(char *)"memview", __pyx_getprop___pyx_array_memview, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PySequenceMethods __pyx_tp_as_sequence_array = {
  __pyx_array___len__, /*sq_length*/
  0, /*sq_concat*/
  0, /*sq_repeat*/
  __pyx_sq_item_array, /*sq_item*/
  0, /*sq_slice*/
  0, /*sq_ass_item*/
  0, /*sq_ass_slice*/
  0, /*sq_contains*/
  0, /*sq_inplace_concat*/
  0, /*sq_inplace_repeat*/
};

static PyMappingMethods __pyx_tp_as_mapping_array = {
  __pyx_array___len__, /*mp_length*/
  __pyx_array___getitem__, /*mp_subscript*/
  __pyx_mp_ass_subscript_array, /*mp_ass_subscript*/
};

static PyBufferProcs __pyx_tp_as_buffer_array = {
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getreadbuffer*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getwritebuffer*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getsegcount*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getcharbuffer*/
  #endif
  __pyx_array_getbuffer, /*bf_getbuffer*/
  0, /*bf_releasebuffer*/
};

static PyTypeObject __pyx_type___pyx_array = {
  PyVarObject_HEAD_INIT(0, 0)
  "pandas._libs.interval.array", /*tp_name*/
  sizeof(struct __pyx_array_obj), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_array, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  &__pyx_tp_as_sequence_array, /*tp_as_sequence*/
  &__pyx_tp_as_mapping_array, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  __pyx_tp_getattro_array, /*tp_getattro*/
  0, /*tp_setattro*/
  &__pyx_tp_as_buffer_array, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_array, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_array, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_array, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
};

static PyObject *__pyx_tp_new_Enum(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_MemviewEnum_obj *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_MemviewEnum_obj *)o);
  p->name = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_Enum(PyObject *o) {
  struct __pyx_MemviewEnum_obj *p = (struct __pyx_MemviewEnum_obj *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->name);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_Enum(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_MemviewEnum_obj *p = (struct __pyx_MemviewEnum_obj *)o;
  if (p->name) {
    e = (*v)(p->name, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_Enum(PyObject *o) {
  PyObject* tmp;
  struct __pyx_MemviewEnum_obj *p = (struct __pyx_MemviewEnum_obj *)o;
  tmp = ((PyObject*)p->name);
  p->name = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyMethodDef __pyx_methods_Enum[] = {
  {"__reduce_cython__", (PyCFunction)__pyx_pw___pyx_MemviewEnum_1__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw___pyx_MemviewEnum_3__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type___pyx_MemviewEnum = {
  PyVarObject_HEAD_INIT(0, 0)
  "pandas._libs.interval.Enum", /*tp_name*/
  sizeof(struct __pyx_MemviewEnum_obj), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_Enum, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_MemviewEnum___repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_Enum, /*tp_traverse*/
  __pyx_tp_clear_Enum, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_Enum, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_MemviewEnum___init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_Enum, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
};
static struct __pyx_vtabstruct_memoryview __pyx_vtable_memoryview;

static PyObject *__pyx_tp_new_memoryview(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_memoryview_obj *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_memoryview_obj *)o);
  p->__pyx_vtab = __pyx_vtabptr_memoryview;
  p->obj = Py_None; Py_INCREF(Py_None);
  p->_size = Py_None; Py_INCREF(Py_None);
  p->_array_interface = Py_None; Py_INCREF(Py_None);
  p->view.obj = NULL;
  if (unlikely(__pyx_memoryview___cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_memoryview(PyObject *o) {
  struct __pyx_memoryview_obj *p = (struct __pyx_memoryview_obj *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    ++Py_REFCNT(o);
    __pyx_memoryview___dealloc__(o);
    --Py_REFCNT(o);
    PyErr_Restore(etype, eval, etb);
  }
  Py_CLEAR(p->obj);
  Py_CLEAR(p->_size);
  Py_CLEAR(p->_array_interface);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_memoryview(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_memoryview_obj *p = (struct __pyx_memoryview_obj *)o;
  if (p->obj) {
    e = (*v)(p->obj, a); if (e) return e;
  }
  if (p->_size) {
    e = (*v)(p->_size, a); if (e) return e;
  }
  if (p->_array_interface) {
    e = (*v)(p->_array_interface, a); if (e) return e;
  }
  if (p->view.obj) {
    e = (*v)(p->view.obj, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_memoryview(PyObject *o) {
  PyObject* tmp;
  struct __pyx_memoryview_obj *p = (struct __pyx_memoryview_obj *)o;
  tmp = ((PyObject*)p->obj);
  p->obj = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_size);
  p->_size = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_array_interface);
  p->_array_interface = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  Py_CLEAR(p->view.obj);
  return 0;
}
static PyObject *__pyx_sq_item_memoryview(PyObject *o, Py_ssize_t i) {
  PyObject *r;
  PyObject *x = PyInt_FromSsize_t(i); if(!x) return 0;
  r = Py_TYPE(o)->tp_as_mapping->mp_subscript(o, x);
  Py_DECREF(x);
  return r;
}

static int __pyx_mp_ass_subscript_memoryview(PyObject *o, PyObject *i, PyObject *v) {
  if (v) {
    return __pyx_memoryview___setitem__(o, i, v);
  }
  else {
    PyErr_Format(PyExc_NotImplementedError,
      "Subscript deletion not supported by %.200s", Py_TYPE(o)->tp_name);
    return -1;
  }
}

static PyObject *__pyx_getprop___pyx_memoryview_T(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_10memoryview_1T_1__get__(o);
}

static PyObject *__pyx_getprop___pyx_memoryview_base(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_10memoryview_4base_1__get__(o);
}

static PyObject *__pyx_getprop___pyx_memoryview_shape(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_10memoryview_5shape_1__get__(o);
}

static PyObject *__pyx_getprop___pyx_memoryview_strides(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_10memoryview_7strides_1__get__(o);
}

static PyObject *__pyx_getprop___pyx_memoryview_suboffsets(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_10memoryview_10suboffsets_1__get__(o);
}

static PyObject *__pyx_getprop___pyx_memoryview_ndim(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_10memoryview_4ndim_1__get__(o);
}

static PyObject *__pyx_getprop___pyx_memoryview_itemsize(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_10memoryview_8itemsize_1__get__(o);
}

static PyObject *__pyx_getprop___pyx_memoryview_nbytes(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_10memoryview_6nbytes_1__get__(o);
}

static PyObject *__pyx_getprop___pyx_memoryview_size(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_10memoryview_4size_1__get__(o);
}

static PyMethodDef __pyx_methods_memoryview[] = {
  {"is_c_contig", (PyCFunction)__pyx_memoryview_is_c_contig, METH_NOARGS, 0},
  {"is_f_contig", (PyCFunction)__pyx_memoryview_is_f_contig, METH_NOARGS, 0},
  {"copy", (PyCFunction)__pyx_memoryview_copy, METH_NOARGS, 0},
  {"copy_fortran", (PyCFunction)__pyx_memoryview_copy_fortran, METH_NOARGS, 0},
  {"__reduce_cython__", (PyCFunction)__pyx_pw___pyx_memoryview_1__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw___pyx_memoryview_3__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_memoryview[] = {
  {(char *)"T", __pyx_getprop___pyx_memoryview_T, 0, (char *)0, 0},
  {(char *)"base", __pyx_getprop___pyx_memoryview_base, 0, (char *)0, 0},
  {(char *)"shape", __pyx_getprop___pyx_memoryview_shape, 0, (char *)0, 0},
  {(char *)"strides", __pyx_getprop___pyx_memoryview_strides, 0, (char *)0, 0},
  {(char *)"suboffsets", __pyx_getprop___pyx_memoryview_suboffsets, 0, (char *)0, 0},
  {(char *)"ndim", __pyx_getprop___pyx_memoryview_ndim, 0, (char *)0, 0},
  {(char *)"itemsize", __pyx_getprop___pyx_memoryview_itemsize, 0, (char *)0, 0},
  {(char *)"nbytes", __pyx_getprop___pyx_memoryview_nbytes, 0, (char *)0, 0},
  {(char *)"size", __pyx_getprop___pyx_memoryview_size, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PySequenceMethods __pyx_tp_as_sequence_memoryview = {
  __pyx_memoryview___len__, /*sq_length*/
  0, /*sq_concat*/
  0, /*sq_repeat*/
  __pyx_sq_item_memoryview, /*sq_item*/
  0, /*sq_slice*/
  0, /*sq_ass_item*/
  0, /*sq_ass_slice*/
  0, /*sq_contains*/
  0, /*sq_inplace_concat*/
  0, /*sq_inplace_repeat*/
};

static PyMappingMethods __pyx_tp_as_mapping_memoryview = {
  __pyx_memoryview___len__, /*mp_length*/
  __pyx_memoryview___getitem__, /*mp_subscript*/
  __pyx_mp_ass_subscript_memoryview, /*mp_ass_subscript*/
};

static PyBufferProcs __pyx_tp_as_buffer_memoryview = {
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getreadbuffer*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getwritebuffer*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getsegcount*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getcharbuffer*/
  #endif
  __pyx_memoryview_getbuffer, /*bf_getbuffer*/
  0, /*bf_releasebuffer*/
};

static PyTypeObject __pyx_type___pyx_memoryview = {
  PyVarObject_HEAD_INIT(0, 0)
  "pandas._libs.interval.memoryview", /*tp_name*/
  sizeof(struct __pyx_memoryview_obj), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_memoryview, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_memoryview___repr__, /*tp_repr*/
  0, /*tp_as_number*/
  &__pyx_tp_as_sequence_memoryview, /*tp_as_sequence*/
  &__pyx_tp_as_mapping_memoryview, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  __pyx_memoryview___str__, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  &__pyx_tp_as_buffer_memoryview, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_memoryview, /*tp_traverse*/
  __pyx_tp_clear_memoryview, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_memoryview, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_memoryview, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_memoryview, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
};
static struct __pyx_vtabstruct__memoryviewslice __pyx_vtable__memoryviewslice;

static PyObject *__pyx_tp_new__memoryviewslice(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_memoryviewslice_obj *p;
  PyObject *o = __pyx_tp_new_memoryview(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_memoryviewslice_obj *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_memoryview*)__pyx_vtabptr__memoryviewslice;
  p->from_object = Py_None; Py_INCREF(Py_None);
  p->from_slice.memview = NULL;
  return o;
}

static void __pyx_tp_dealloc__memoryviewslice(PyObject *o) {
  struct __pyx_memoryviewslice_obj *p = (struct __pyx_memoryviewslice_obj *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    ++Py_REFCNT(o);
    __pyx_memoryviewslice___dealloc__(o);
    --Py_REFCNT(o);
    PyErr_Restore(etype, eval, etb);
  }
  Py_CLEAR(p->from_object);
  PyObject_GC_Track(o);
  __pyx_tp_dealloc_memoryview(o);
}

static int __pyx_tp_traverse__memoryviewslice(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_memoryviewslice_obj *p = (struct __pyx_memoryviewslice_obj *)o;
  e = __pyx_tp_traverse_memoryview(o, v, a); if (e) return e;
  if (p->from_object) {
    e = (*v)(p->from_object, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear__memoryviewslice(PyObject *o) {
  PyObject* tmp;
  struct __pyx_memoryviewslice_obj *p = (struct __pyx_memoryviewslice_obj *)o;
  __pyx_tp_clear_memoryview(o);
  tmp = ((PyObject*)p->from_object);
  p->from_object = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  __PYX_XDEC_MEMVIEW(&p->from_slice, 1);
  return 0;
}

static PyObject *__pyx_getprop___pyx_memoryviewslice_base(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_16_memoryviewslice_4base_1__get__(o);
}

static PyMethodDef __pyx_methods__memoryviewslice[] = {
  {"__reduce_cython__", (PyCFunction)__pyx_pw___pyx_memoryviewslice_1__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw___pyx_memoryviewslice_3__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets__memoryviewslice[] = {
  {(char *)"base", __pyx_getprop___pyx_memoryviewslice_base, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type___pyx_memoryviewslice = {
  PyVarObject_HEAD_INIT(0, 0)
  "pandas._libs.interval._memoryviewslice", /*tp_name*/
  sizeof(struct __pyx_memoryviewslice_obj), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc__memoryviewslice, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY
  __pyx_memoryview___repr__, /*tp_repr*/
  #else
  0, /*tp_repr*/
  #endif
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  #if CYTHON_COMPILING_IN_PYPY
  __pyx_memoryview___str__, /*tp_str*/
  #else
  0, /*tp_str*/
  #endif
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  "Internal class for passing memoryview slices to Python", /*tp_doc*/
  __pyx_tp_traverse__memoryviewslice, /*tp_traverse*/
  __pyx_tp_clear__memoryviewslice, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods__memoryviewslice, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets__memoryviewslice, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new__memoryviewslice, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
};

static PyMethodDef __pyx_methods[] = {
  {0, 0, 0, 0}
};

#if PY_MAJOR_VERSION >= 3
#if CYTHON_PEP489_MULTI_PHASE_INIT
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/
static int __pyx_pymod_exec_interval(PyObject* module); /*proto*/
static PyModuleDef_Slot __pyx_moduledef_slots[] = {
  {Py_mod_create, (void*)__pyx_pymod_create},
  {Py_mod_exec, (void*)__pyx_pymod_exec_interval},
  {0, NULL}
};
#endif

static struct PyModuleDef __pyx_moduledef = {
    PyModuleDef_HEAD_INIT,
    "interval",
    0, /* m_doc */
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    0, /* m_size */
  #else
    -1, /* m_size */
  #endif
    __pyx_methods /* m_methods */,
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    __pyx_moduledef_slots, /* m_slots */
  #else
    NULL, /* m_reload */
  #endif
    NULL, /* m_traverse */
    NULL, /* m_clear */
    NULL /* m_free */
};
#endif
#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 3))
    #define CYTHON_SMALL_CODE __attribute__((cold))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif

static __Pyx_StringTabEntry __pyx_string_tab[] = {
  {&__pyx_n_s_ASCII, __pyx_k_ASCII, sizeof(__pyx_k_ASCII), 0, 0, 1, 1},
  {&__pyx_kp_s_Buffer_view_does_not_expose_stri, __pyx_k_Buffer_view_does_not_expose_stri, sizeof(__pyx_k_Buffer_view_does_not_expose_stri), 0, 0, 1, 0},
  {&__pyx_kp_s_Can_only_create_a_buffer_that_is, __pyx_k_Can_only_create_a_buffer_that_is, sizeof(__pyx_k_Can_only_create_a_buffer_that_is), 0, 0, 1, 0},
  {&__pyx_kp_s_Cannot_assign_to_read_only_memor, __pyx_k_Cannot_assign_to_read_only_memor, sizeof(__pyx_k_Cannot_assign_to_read_only_memor), 0, 0, 1, 0},
  {&__pyx_kp_s_Cannot_create_writable_memory_vi, __pyx_k_Cannot_create_writable_memory_vi, sizeof(__pyx_k_Cannot_create_writable_memory_vi), 0, 0, 1, 0},
  {&__pyx_kp_s_Cannot_index_with_type_s, __pyx_k_Cannot_index_with_type_s, sizeof(__pyx_k_Cannot_index_with_type_s), 0, 0, 1, 0},
  {&__pyx_kp_u_Check_whether_two_Interval_obje, __pyx_k_Check_whether_two_Interval_obje, sizeof(__pyx_k_Check_whether_two_Interval_obje), 0, 1, 0, 0},
  {&__pyx_n_s_Ellipsis, __pyx_k_Ellipsis, sizeof(__pyx_k_Ellipsis), 0, 0, 1, 1},
  {&__pyx_kp_s_Empty_shape_tuple_for_cython_arr, __pyx_k_Empty_shape_tuple_for_cython_arr, sizeof(__pyx_k_Empty_shape_tuple_for_cython_arr), 0, 0, 1, 0},
  {&__pyx_kp_s_Expected_at_least_d_argument_s_g, __pyx_k_Expected_at_least_d_argument_s_g, sizeof(__pyx_k_Expected_at_least_d_argument_s_g), 0, 0, 1, 0},
  {&__pyx_n_s_Float32ClosedBothIntervalNode, __pyx_k_Float32ClosedBothIntervalNode, sizeof(__pyx_k_Float32ClosedBothIntervalNode), 0, 0, 1, 1},
  {&__pyx_n_s_Float32ClosedBothIntervalNode_2, __pyx_k_Float32ClosedBothIntervalNode_2, sizeof(__pyx_k_Float32ClosedBothIntervalNode_2), 0, 0, 1, 1},
  {&__pyx_n_s_Float32ClosedBothIntervalNode_3, __pyx_k_Float32ClosedBothIntervalNode_3, sizeof(__pyx_k_Float32ClosedBothIntervalNode_3), 0, 0, 1, 1},
  {&__pyx_n_s_Float32ClosedBothIntervalNode_4, __pyx_k_Float32ClosedBothIntervalNode_4, sizeof(__pyx_k_Float32ClosedBothIntervalNode_4), 0, 0, 1, 1},
  {&__pyx_n_s_Float32ClosedBothIntervalNode_5, __pyx_k_Float32ClosedBothIntervalNode_5, sizeof(__pyx_k_Float32ClosedBothIntervalNode_5), 0, 0, 1, 1},
  {&__pyx_n_s_Float32ClosedBothIntervalNode_6, __pyx_k_Float32ClosedBothIntervalNode_6, sizeof(__pyx_k_Float32ClosedBothIntervalNode_6), 0, 0, 1, 1},
  {&__pyx_kp_u_Float32ClosedBothIntervalNode_p, __pyx_k_Float32ClosedBothIntervalNode_p, sizeof(__pyx_k_Float32ClosedBothIntervalNode_p), 0, 1, 0, 0},
  {&__pyx_n_s_Float32ClosedBothIntervalNode_qu, __pyx_k_Float32ClosedBothIntervalNode_qu, sizeof(__pyx_k_Float32ClosedBothIntervalNode_qu), 0, 0, 1, 1},
  {&__pyx_kp_u_Float32ClosedBothIntervalNode_s, __pyx_k_Float32ClosedBothIntervalNode_s, sizeof(__pyx_k_Float32ClosedBothIntervalNode_s), 0, 1, 0, 0},
  {&__pyx_n_s_Float32ClosedLeftIntervalNode, __pyx_k_Float32ClosedLeftIntervalNode, sizeof(__pyx_k_Float32ClosedLeftIntervalNode), 0, 0, 1, 1},
  {&__pyx_n_s_Float32ClosedLeftIntervalNode_2, __pyx_k_Float32ClosedLeftIntervalNode_2, sizeof(__pyx_k_Float32ClosedLeftIntervalNode_2), 0, 0, 1, 1},
  {&__pyx_n_s_Float32ClosedLeftIntervalNode_3, __pyx_k_Float32ClosedLeftIntervalNode_3, sizeof(__pyx_k_Float32ClosedLeftIntervalNode_3), 0, 0, 1, 1},
  {&__pyx_n_s_Float32ClosedLeftIntervalNode_4, __pyx_k_Float32ClosedLeftIntervalNode_4, sizeof(__pyx_k_Float32ClosedLeftIntervalNode_4), 0, 0, 1, 1},
  {&__pyx_n_s_Float32ClosedLeftIntervalNode_5, __pyx_k_Float32ClosedLeftIntervalNode_5, sizeof(__pyx_k_Float32ClosedLeftIntervalNode_5), 0, 0, 1, 1},
  {&__pyx_n_s_Float32ClosedLeftIntervalNode_6, __pyx_k_Float32ClosedLeftIntervalNode_6, sizeof(__pyx_k_Float32ClosedLeftIntervalNode_6), 0, 0, 1, 1},
  {&__pyx_kp_u_Float32ClosedLeftIntervalNode_p, __pyx_k_Float32ClosedLeftIntervalNode_p, sizeof(__pyx_k_Float32ClosedLeftIntervalNode_p), 0, 1, 0, 0},
  {&__pyx_n_s_Float32ClosedLeftIntervalNode_qu, __pyx_k_Float32ClosedLeftIntervalNode_qu, sizeof(__pyx_k_Float32ClosedLeftIntervalNode_qu), 0, 0, 1, 1},
  {&__pyx_kp_u_Float32ClosedLeftIntervalNode_s, __pyx_k_Float32ClosedLeftIntervalNode_s, sizeof(__pyx_k_Float32ClosedLeftIntervalNode_s), 0, 1, 0, 0},
  {&__pyx_kp_u_Float32ClosedNeitherIntervalNod, __pyx_k_Float32ClosedNeitherIntervalNod, sizeof(__pyx_k_Float32ClosedNeitherIntervalNod), 0, 1, 0, 0},
  {&__pyx_kp_u_Float32ClosedNeitherIntervalNod_2, __pyx_k_Float32ClosedNeitherIntervalNod_2, sizeof(__pyx_k_Float32ClosedNeitherIntervalNod_2), 0, 1, 0, 0},
  {&__pyx_n_s_Float32ClosedNeitherIntervalNode, __pyx_k_Float32ClosedNeitherIntervalNode, sizeof(__pyx_k_Float32ClosedNeitherIntervalNode), 0, 0, 1, 1},
  {&__pyx_n_s_Float32ClosedNeitherIntervalNode_2, __pyx_k_Float32ClosedNeitherIntervalNode_2, sizeof(__pyx_k_Float32ClosedNeitherIntervalNode_2), 0, 0, 1, 1},
  {&__pyx_n_s_Float32ClosedNeitherIntervalNode_3, __pyx_k_Float32ClosedNeitherIntervalNode_3, sizeof(__pyx_k_Float32ClosedNeitherIntervalNode_3), 0, 0, 1, 1},
  {&__pyx_n_s_Float32ClosedNeitherIntervalNode_4, __pyx_k_Float32ClosedNeitherIntervalNode_4, sizeof(__pyx_k_Float32ClosedNeitherIntervalNode_4), 0, 0, 1, 1},
  {&__pyx_n_s_Float32ClosedNeitherIntervalNode_5, __pyx_k_Float32ClosedNeitherIntervalNode_5, sizeof(__pyx_k_Float32ClosedNeitherIntervalNode_5), 0, 0, 1, 1},
  {&__pyx_n_s_Float32ClosedNeitherIntervalNode_6, __pyx_k_Float32ClosedNeitherIntervalNode_6, sizeof(__pyx_k_Float32ClosedNeitherIntervalNode_6), 0, 0, 1, 1},
  {&__pyx_n_s_Float32ClosedNeitherIntervalNode_7, __pyx_k_Float32ClosedNeitherIntervalNode_7, sizeof(__pyx_k_Float32ClosedNeitherIntervalNode_7), 0, 0, 1, 1},
  {&__pyx_kp_u_Float32ClosedRightIntervalNode, __pyx_k_Float32ClosedRightIntervalNode, sizeof(__pyx_k_Float32ClosedRightIntervalNode), 0, 1, 0, 0},
  {&__pyx_kp_u_Float32ClosedRightIntervalNode_2, __pyx_k_Float32ClosedRightIntervalNode_2, sizeof(__pyx_k_Float32ClosedRightIntervalNode_2), 0, 1, 0, 0},
  {&__pyx_n_s_Float32ClosedRightIntervalNode_3, __pyx_k_Float32ClosedRightIntervalNode_3, sizeof(__pyx_k_Float32ClosedRightIntervalNode_3), 0, 0, 1, 1},
  {&__pyx_n_s_Float32ClosedRightIntervalNode_4, __pyx_k_Float32ClosedRightIntervalNode_4, sizeof(__pyx_k_Float32ClosedRightIntervalNode_4), 0, 0, 1, 1},
  {&__pyx_n_s_Float32ClosedRightIntervalNode_5, __pyx_k_Float32ClosedRightIntervalNode_5, sizeof(__pyx_k_Float32ClosedRightIntervalNode_5), 0, 0, 1, 1},
  {&__pyx_n_s_Float32ClosedRightIntervalNode_6, __pyx_k_Float32ClosedRightIntervalNode_6, sizeof(__pyx_k_Float32ClosedRightIntervalNode_6), 0, 0, 1, 1},
  {&__pyx_n_s_Float32ClosedRightIntervalNode_7, __pyx_k_Float32ClosedRightIntervalNode_7, sizeof(__pyx_k_Float32ClosedRightIntervalNode_7), 0, 0, 1, 1},
  {&__pyx_n_s_Float32ClosedRightIntervalNode_8, __pyx_k_Float32ClosedRightIntervalNode_8, sizeof(__pyx_k_Float32ClosedRightIntervalNode_8), 0, 0, 1, 1},
  {&__pyx_n_s_Float32ClosedRightIntervalNode_q, __pyx_k_Float32ClosedRightIntervalNode_q, sizeof(__pyx_k_Float32ClosedRightIntervalNode_q), 0, 0, 1, 1},
  {&__pyx_n_s_Float64ClosedBothIntervalNode, __pyx_k_Float64ClosedBothIntervalNode, sizeof(__pyx_k_Float64ClosedBothIntervalNode), 0, 0, 1, 1},
  {&__pyx_n_s_Float64ClosedBothIntervalNode_2, __pyx_k_Float64ClosedBothIntervalNode_2, sizeof(__pyx_k_Float64ClosedBothIntervalNode_2), 0, 0, 1, 1},
  {&__pyx_n_s_Float64ClosedBothIntervalNode_3, __pyx_k_Float64ClosedBothIntervalNode_3, sizeof(__pyx_k_Float64ClosedBothIntervalNode_3), 0, 0, 1, 1},
  {&__pyx_n_s_Float64ClosedBothIntervalNode_4, __pyx_k_Float64ClosedBothIntervalNode_4, sizeof(__pyx_k_Float64ClosedBothIntervalNode_4), 0, 0, 1, 1},
  {&__pyx_n_s_Float64ClosedBothIntervalNode_5, __pyx_k_Float64ClosedBothIntervalNode_5, sizeof(__pyx_k_Float64ClosedBothIntervalNode_5), 0, 0, 1, 1},
  {&__pyx_n_s_Float64ClosedBothIntervalNode_6, __pyx_k_Float64ClosedBothIntervalNode_6, sizeof(__pyx_k_Float64ClosedBothIntervalNode_6), 0, 0, 1, 1},
  {&__pyx_kp_u_Float64ClosedBothIntervalNode_p, __pyx_k_Float64ClosedBothIntervalNode_p, sizeof(__pyx_k_Float64ClosedBothIntervalNode_p), 0, 1, 0, 0},
  {&__pyx_n_s_Float64ClosedBothIntervalNode_qu, __pyx_k_Float64ClosedBothIntervalNode_qu, sizeof(__pyx_k_Float64ClosedBothIntervalNode_qu), 0, 0, 1, 1},
  {&__pyx_kp_u_Float64ClosedBothIntervalNode_s, __pyx_k_Float64ClosedBothIntervalNode_s, sizeof(__pyx_k_Float64ClosedBothIntervalNode_s), 0, 1, 0, 0},
  {&__pyx_n_s_Float64ClosedLeftIntervalNode, __pyx_k_Float64ClosedLeftIntervalNode, sizeof(__pyx_k_Float64ClosedLeftIntervalNode), 0, 0, 1, 1},
  {&__pyx_n_s_Float64ClosedLeftIntervalNode_2, __pyx_k_Float64ClosedLeftIntervalNode_2, sizeof(__pyx_k_Float64ClosedLeftIntervalNode_2), 0, 0, 1, 1},
  {&__pyx_n_s_Float64ClosedLeftIntervalNode_3, __pyx_k_Float64ClosedLeftIntervalNode_3, sizeof(__pyx_k_Float64ClosedLeftIntervalNode_3), 0, 0, 1, 1},
  {&__pyx_n_s_Float64ClosedLeftIntervalNode_4, __pyx_k_Float64ClosedLeftIntervalNode_4, sizeof(__pyx_k_Float64ClosedLeftIntervalNode_4), 0, 0, 1, 1},
  {&__pyx_n_s_Float64ClosedLeftIntervalNode_5, __pyx_k_Float64ClosedLeftIntervalNode_5, sizeof(__pyx_k_Float64ClosedLeftIntervalNode_5), 0, 0, 1, 1},
  {&__pyx_n_s_Float64ClosedLeftIntervalNode_6, __pyx_k_Float64ClosedLeftIntervalNode_6, sizeof(__pyx_k_Float64ClosedLeftIntervalNode_6), 0, 0, 1, 1},
  {&__pyx_kp_u_Float64ClosedLeftIntervalNode_p, __pyx_k_Float64ClosedLeftIntervalNode_p, sizeof(__pyx_k_Float64ClosedLeftIntervalNode_p), 0, 1, 0, 0},
  {&__pyx_n_s_Float64ClosedLeftIntervalNode_qu, __pyx_k_Float64ClosedLeftIntervalNode_qu, sizeof(__pyx_k_Float64ClosedLeftIntervalNode_qu), 0, 0, 1, 1},
  {&__pyx_kp_u_Float64ClosedLeftIntervalNode_s, __pyx_k_Float64ClosedLeftIntervalNode_s, sizeof(__pyx_k_Float64ClosedLeftIntervalNode_s), 0, 1, 0, 0},
  {&__pyx_kp_u_Float64ClosedNeitherIntervalNod, __pyx_k_Float64ClosedNeitherIntervalNod, sizeof(__pyx_k_Float64ClosedNeitherIntervalNod), 0, 1, 0, 0},
  {&__pyx_kp_u_Float64ClosedNeitherIntervalNod_2, __pyx_k_Float64ClosedNeitherIntervalNod_2, sizeof(__pyx_k_Float64ClosedNeitherIntervalNod_2), 0, 1, 0, 0},
  {&__pyx_n_s_Float64ClosedNeitherIntervalNode, __pyx_k_Float64ClosedNeitherIntervalNode, sizeof(__pyx_k_Float64ClosedNeitherIntervalNode), 0, 0, 1, 1},
  {&__pyx_n_s_Float64ClosedNeitherIntervalNode_2, __pyx_k_Float64ClosedNeitherIntervalNode_2, sizeof(__pyx_k_Float64ClosedNeitherIntervalNode_2), 0, 0, 1, 1},
  {&__pyx_n_s_Float64ClosedNeitherIntervalNode_3, __pyx_k_Float64ClosedNeitherIntervalNode_3, sizeof(__pyx_k_Float64ClosedNeitherIntervalNode_3), 0, 0, 1, 1},
  {&__pyx_n_s_Float64ClosedNeitherIntervalNode_4, __pyx_k_Float64ClosedNeitherIntervalNode_4, sizeof(__pyx_k_Float64ClosedNeitherIntervalNode_4), 0, 0, 1, 1},
  {&__pyx_n_s_Float64ClosedNeitherIntervalNode_5, __pyx_k_Float64ClosedNeitherIntervalNode_5, sizeof(__pyx_k_Float64ClosedNeitherIntervalNode_5), 0, 0, 1, 1},
  {&__pyx_n_s_Float64ClosedNeitherIntervalNode_6, __pyx_k_Float64ClosedNeitherIntervalNode_6, sizeof(__pyx_k_Float64ClosedNeitherIntervalNode_6), 0, 0, 1, 1},
  {&__pyx_n_s_Float64ClosedNeitherIntervalNode_7, __pyx_k_Float64ClosedNeitherIntervalNode_7, sizeof(__pyx_k_Float64ClosedNeitherIntervalNode_7), 0, 0, 1, 1},
  {&__pyx_kp_u_Float64ClosedRightIntervalNode, __pyx_k_Float64ClosedRightIntervalNode, sizeof(__pyx_k_Float64ClosedRightIntervalNode), 0, 1, 0, 0},
  {&__pyx_kp_u_Float64ClosedRightIntervalNode_2, __pyx_k_Float64ClosedRightIntervalNode_2, sizeof(__pyx_k_Float64ClosedRightIntervalNode_2), 0, 1, 0, 0},
  {&__pyx_n_s_Float64ClosedRightIntervalNode_3, __pyx_k_Float64ClosedRightIntervalNode_3, sizeof(__pyx_k_Float64ClosedRightIntervalNode_3), 0, 0, 1, 1},
  {&__pyx_n_s_Float64ClosedRightIntervalNode_4, __pyx_k_Float64ClosedRightIntervalNode_4, sizeof(__pyx_k_Float64ClosedRightIntervalNode_4), 0, 0, 1, 1},
  {&__pyx_n_s_Float64ClosedRightIntervalNode_5, __pyx_k_Float64ClosedRightIntervalNode_5, sizeof(__pyx_k_Float64ClosedRightIntervalNode_5), 0, 0, 1, 1},
  {&__pyx_n_s_Float64ClosedRightIntervalNode_6, __pyx_k_Float64ClosedRightIntervalNode_6, sizeof(__pyx_k_Float64ClosedRightIntervalNode_6), 0, 0, 1, 1},
  {&__pyx_n_s_Float64ClosedRightIntervalNode_7, __pyx_k_Float64ClosedRightIntervalNode_7, sizeof(__pyx_k_Float64ClosedRightIntervalNode_7), 0, 0, 1, 1},
  {&__pyx_n_s_Float64ClosedRightIntervalNode_8, __pyx_k_Float64ClosedRightIntervalNode_8, sizeof(__pyx_k_Float64ClosedRightIntervalNode_8), 0, 0, 1, 1},
  {&__pyx_n_s_Float64ClosedRightIntervalNode_q, __pyx_k_Float64ClosedRightIntervalNode_q, sizeof(__pyx_k_Float64ClosedRightIntervalNode_q), 0, 0, 1, 1},
  {&__pyx_kp_u_Format_string_allocated_too_shor, __pyx_k_Format_string_allocated_too_shor, sizeof(__pyx_k_Format_string_allocated_too_shor), 0, 1, 0, 0},
  {&__pyx_kp_u_Format_string_allocated_too_shor_2, __pyx_k_Format_string_allocated_too_shor_2, sizeof(__pyx_k_Format_string_allocated_too_shor_2), 0, 1, 0, 0},
  {&__pyx_kp_s_Function_call_with_ambiguous_arg, __pyx_k_Function_call_with_ambiguous_arg, sizeof(__pyx_k_Function_call_with_ambiguous_arg), 0, 0, 1, 0},
  {&__pyx_n_s_ImportError, __pyx_k_ImportError, sizeof(__pyx_k_ImportError), 0, 0, 1, 1},
  {&__pyx_kp_s_Incompatible_checksums_s_vs_0x05, __pyx_k_Incompatible_checksums_s_vs_0x05, sizeof(__pyx_k_Incompatible_checksums_s_vs_0x05), 0, 0, 1, 0},
  {&__pyx_kp_s_Incompatible_checksums_s_vs_0x7a, __pyx_k_Incompatible_checksums_s_vs_0x7a, sizeof(__pyx_k_Incompatible_checksums_s_vs_0x7a), 0, 0, 1, 0},
  {&__pyx_kp_s_Incompatible_checksums_s_vs_0xb0, __pyx_k_Incompatible_checksums_s_vs_0xb0, sizeof(__pyx_k_Incompatible_checksums_s_vs_0xb0), 0, 0, 1, 0},
  {&__pyx_kp_s_Incompatible_checksums_s_vs_0xd4, __pyx_k_Incompatible_checksums_s_vs_0xd4, sizeof(__pyx_k_Incompatible_checksums_s_vs_0xd4), 0, 0, 1, 0},
  {&__pyx_n_s_IndexError, __pyx_k_IndexError, sizeof(__pyx_k_IndexError), 0, 0, 1, 1},
  {&__pyx_kp_u_Indicates_if_an_interval_is_emp, __pyx_k_Indicates_if_an_interval_is_emp, sizeof(__pyx_k_Indicates_if_an_interval_is_emp), 0, 1, 0, 0},
  {&__pyx_kp_s_Indirect_dimensions_not_supporte, __pyx_k_Indirect_dimensions_not_supporte, sizeof(__pyx_k_Indirect_dimensions_not_supporte), 0, 0, 1, 0},
  {&__pyx_n_s_Int32ClosedBothIntervalNode, __pyx_k_Int32ClosedBothIntervalNode, sizeof(__pyx_k_Int32ClosedBothIntervalNode), 0, 0, 1, 1},
  {&__pyx_n_s_Int32ClosedBothIntervalNode___py, __pyx_k_Int32ClosedBothIntervalNode___py, sizeof(__pyx_k_Int32ClosedBothIntervalNode___py), 0, 0, 1, 1},
  {&__pyx_n_s_Int32ClosedBothIntervalNode___py_2, __pyx_k_Int32ClosedBothIntervalNode___py_2, sizeof(__pyx_k_Int32ClosedBothIntervalNode___py_2), 0, 0, 1, 1},
  {&__pyx_n_s_Int32ClosedBothIntervalNode___py_3, __pyx_k_Int32ClosedBothIntervalNode___py_3, sizeof(__pyx_k_Int32ClosedBothIntervalNode___py_3), 0, 0, 1, 1},
  {&__pyx_n_s_Int32ClosedBothIntervalNode___py_4, __pyx_k_Int32ClosedBothIntervalNode___py_4, sizeof(__pyx_k_Int32ClosedBothIntervalNode___py_4), 0, 0, 1, 1},
  {&__pyx_n_s_Int32ClosedBothIntervalNode___py_5, __pyx_k_Int32ClosedBothIntervalNode___py_5, sizeof(__pyx_k_Int32ClosedBothIntervalNode___py_5), 0, 0, 1, 1},
  {&__pyx_kp_u_Int32ClosedBothIntervalNode_piv, __pyx_k_Int32ClosedBothIntervalNode_piv, sizeof(__pyx_k_Int32ClosedBothIntervalNode_piv), 0, 1, 0, 0},
  {&__pyx_n_s_Int32ClosedBothIntervalNode_quer, __pyx_k_Int32ClosedBothIntervalNode_quer, sizeof(__pyx_k_Int32ClosedBothIntervalNode_quer), 0, 0, 1, 1},
  {&__pyx_kp_u_Int32ClosedBothIntervalNode_s_e, __pyx_k_Int32ClosedBothIntervalNode_s_e, sizeof(__pyx_k_Int32ClosedBothIntervalNode_s_e), 0, 1, 0, 0},
  {&__pyx_n_s_Int32ClosedLeftIntervalNode, __pyx_k_Int32ClosedLeftIntervalNode, sizeof(__pyx_k_Int32ClosedLeftIntervalNode), 0, 0, 1, 1},
  {&__pyx_n_s_Int32ClosedLeftIntervalNode___py, __pyx_k_Int32ClosedLeftIntervalNode___py, sizeof(__pyx_k_Int32ClosedLeftIntervalNode___py), 0, 0, 1, 1},
  {&__pyx_n_s_Int32ClosedLeftIntervalNode___py_2, __pyx_k_Int32ClosedLeftIntervalNode___py_2, sizeof(__pyx_k_Int32ClosedLeftIntervalNode___py_2), 0, 0, 1, 1},
  {&__pyx_n_s_Int32ClosedLeftIntervalNode___py_3, __pyx_k_Int32ClosedLeftIntervalNode___py_3, sizeof(__pyx_k_Int32ClosedLeftIntervalNode___py_3), 0, 0, 1, 1},
  {&__pyx_n_s_Int32ClosedLeftIntervalNode___py_4, __pyx_k_Int32ClosedLeftIntervalNode___py_4, sizeof(__pyx_k_Int32ClosedLeftIntervalNode___py_4), 0, 0, 1, 1},
  {&__pyx_n_s_Int32ClosedLeftIntervalNode___py_5, __pyx_k_Int32ClosedLeftIntervalNode___py_5, sizeof(__pyx_k_Int32ClosedLeftIntervalNode___py_5), 0, 0, 1, 1},
  {&__pyx_kp_u_Int32ClosedLeftIntervalNode_piv, __pyx_k_Int32ClosedLeftIntervalNode_piv, sizeof(__pyx_k_Int32ClosedLeftIntervalNode_piv), 0, 1, 0, 0},
  {&__pyx_n_s_Int32ClosedLeftIntervalNode_quer, __pyx_k_Int32ClosedLeftIntervalNode_quer, sizeof(__pyx_k_Int32ClosedLeftIntervalNode_quer), 0, 0, 1, 1},
  {&__pyx_kp_u_Int32ClosedLeftIntervalNode_s_e, __pyx_k_Int32ClosedLeftIntervalNode_s_e, sizeof(__pyx_k_Int32ClosedLeftIntervalNode_s_e), 0, 1, 0, 0},
  {&__pyx_kp_u_Int32ClosedNeitherIntervalNode, __pyx_k_Int32ClosedNeitherIntervalNode, sizeof(__pyx_k_Int32ClosedNeitherIntervalNode), 0, 1, 0, 0},
  {&__pyx_kp_u_Int32ClosedNeitherIntervalNode_2, __pyx_k_Int32ClosedNeitherIntervalNode_2, sizeof(__pyx_k_Int32ClosedNeitherIntervalNode_2), 0, 1, 0, 0},
  {&__pyx_n_s_Int32ClosedNeitherIntervalNode_3, __pyx_k_Int32ClosedNeitherIntervalNode_3, sizeof(__pyx_k_Int32ClosedNeitherIntervalNode_3), 0, 0, 1, 1},
  {&__pyx_n_s_Int32ClosedNeitherIntervalNode_4, __pyx_k_Int32ClosedNeitherIntervalNode_4, sizeof(__pyx_k_Int32ClosedNeitherIntervalNode_4), 0, 0, 1, 1},
  {&__pyx_n_s_Int32ClosedNeitherIntervalNode_5, __pyx_k_Int32ClosedNeitherIntervalNode_5, sizeof(__pyx_k_Int32ClosedNeitherIntervalNode_5), 0, 0, 1, 1},
  {&__pyx_n_s_Int32ClosedNeitherIntervalNode_6, __pyx_k_Int32ClosedNeitherIntervalNode_6, sizeof(__pyx_k_Int32ClosedNeitherIntervalNode_6), 0, 0, 1, 1},
  {&__pyx_n_s_Int32ClosedNeitherIntervalNode_7, __pyx_k_Int32ClosedNeitherIntervalNode_7, sizeof(__pyx_k_Int32ClosedNeitherIntervalNode_7), 0, 0, 1, 1},
  {&__pyx_n_s_Int32ClosedNeitherIntervalNode_8, __pyx_k_Int32ClosedNeitherIntervalNode_8, sizeof(__pyx_k_Int32ClosedNeitherIntervalNode_8), 0, 0, 1, 1},
  {&__pyx_n_s_Int32ClosedNeitherIntervalNode_q, __pyx_k_Int32ClosedNeitherIntervalNode_q, sizeof(__pyx_k_Int32ClosedNeitherIntervalNode_q), 0, 0, 1, 1},
  {&__pyx_n_s_Int32ClosedRightIntervalNode, __pyx_k_Int32ClosedRightIntervalNode, sizeof(__pyx_k_Int32ClosedRightIntervalNode), 0, 0, 1, 1},
  {&__pyx_n_s_Int32ClosedRightIntervalNode___p, __pyx_k_Int32ClosedRightIntervalNode___p, sizeof(__pyx_k_Int32ClosedRightIntervalNode___p), 0, 0, 1, 1},
  {&__pyx_n_s_Int32ClosedRightIntervalNode___p_2, __pyx_k_Int32ClosedRightIntervalNode___p_2, sizeof(__pyx_k_Int32ClosedRightIntervalNode___p_2), 0, 0, 1, 1},
  {&__pyx_n_s_Int32ClosedRightIntervalNode___p_3, __pyx_k_Int32ClosedRightIntervalNode___p_3, sizeof(__pyx_k_Int32ClosedRightIntervalNode___p_3), 0, 0, 1, 1},
  {&__pyx_n_s_Int32ClosedRightIntervalNode___p_4, __pyx_k_Int32ClosedRightIntervalNode___p_4, sizeof(__pyx_k_Int32ClosedRightIntervalNode___p_4), 0, 0, 1, 1},
  {&__pyx_n_s_Int32ClosedRightIntervalNode___p_5, __pyx_k_Int32ClosedRightIntervalNode___p_5, sizeof(__pyx_k_Int32ClosedRightIntervalNode___p_5), 0, 0, 1, 1},
  {&__pyx_kp_u_Int32ClosedRightIntervalNode_pi, __pyx_k_Int32ClosedRightIntervalNode_pi, sizeof(__pyx_k_Int32ClosedRightIntervalNode_pi), 0, 1, 0, 0},
  {&__pyx_n_s_Int32ClosedRightIntervalNode_que, __pyx_k_Int32ClosedRightIntervalNode_que, sizeof(__pyx_k_Int32ClosedRightIntervalNode_que), 0, 0, 1, 1},
  {&__pyx_kp_u_Int32ClosedRightIntervalNode_s, __pyx_k_Int32ClosedRightIntervalNode_s, sizeof(__pyx_k_Int32ClosedRightIntervalNode_s), 0, 1, 0, 0},
  {&__pyx_n_s_Int64ClosedBothIntervalNode, __pyx_k_Int64ClosedBothIntervalNode, sizeof(__pyx_k_Int64ClosedBothIntervalNode), 0, 0, 1, 1},
  {&__pyx_n_s_Int64ClosedBothIntervalNode___py, __pyx_k_Int64ClosedBothIntervalNode___py, sizeof(__pyx_k_Int64ClosedBothIntervalNode___py), 0, 0, 1, 1},
  {&__pyx_n_s_Int64ClosedBothIntervalNode___py_2, __pyx_k_Int64ClosedBothIntervalNode___py_2, sizeof(__pyx_k_Int64ClosedBothIntervalNode___py_2), 0, 0, 1, 1},
  {&__pyx_n_s_Int64ClosedBothIntervalNode___py_3, __pyx_k_Int64ClosedBothIntervalNode___py_3, sizeof(__pyx_k_Int64ClosedBothIntervalNode___py_3), 0, 0, 1, 1},
  {&__pyx_n_s_Int64ClosedBothIntervalNode___py_4, __pyx_k_Int64ClosedBothIntervalNode___py_4, sizeof(__pyx_k_Int64ClosedBothIntervalNode___py_4), 0, 0, 1, 1},
  {&__pyx_n_s_Int64ClosedBothIntervalNode___py_5, __pyx_k_Int64ClosedBothIntervalNode___py_5, sizeof(__pyx_k_Int64ClosedBothIntervalNode___py_5), 0, 0, 1, 1},
  {&__pyx_kp_u_Int64ClosedBothIntervalNode_piv, __pyx_k_Int64ClosedBothIntervalNode_piv, sizeof(__pyx_k_Int64ClosedBothIntervalNode_piv), 0, 1, 0, 0},
  {&__pyx_n_s_Int64ClosedBothIntervalNode_quer, __pyx_k_Int64ClosedBothIntervalNode_quer, sizeof(__pyx_k_Int64ClosedBothIntervalNode_quer), 0, 0, 1, 1},
  {&__pyx_kp_u_Int64ClosedBothIntervalNode_s_e, __pyx_k_Int64ClosedBothIntervalNode_s_e, sizeof(__pyx_k_Int64ClosedBothIntervalNode_s_e), 0, 1, 0, 0},
  {&__pyx_n_s_Int64ClosedLeftIntervalNode, __pyx_k_Int64ClosedLeftIntervalNode, sizeof(__pyx_k_Int64ClosedLeftIntervalNode), 0, 0, 1, 1},
  {&__pyx_n_s_Int64ClosedLeftIntervalNode___py, __pyx_k_Int64ClosedLeftIntervalNode___py, sizeof(__pyx_k_Int64ClosedLeftIntervalNode___py), 0, 0, 1, 1},
  {&__pyx_n_s_Int64ClosedLeftIntervalNode___py_2, __pyx_k_Int64ClosedLeftIntervalNode___py_2, sizeof(__pyx_k_Int64ClosedLeftIntervalNode___py_2), 0, 0, 1, 1},
  {&__pyx_n_s_Int64ClosedLeftIntervalNode___py_3, __pyx_k_Int64ClosedLeftIntervalNode___py_3, sizeof(__pyx_k_Int64ClosedLeftIntervalNode___py_3), 0, 0, 1, 1},
  {&__pyx_n_s_Int64ClosedLeftIntervalNode___py_4, __pyx_k_Int64ClosedLeftIntervalNode___py_4, sizeof(__pyx_k_Int64ClosedLeftIntervalNode___py_4), 0, 0, 1, 1},
  {&__pyx_n_s_Int64ClosedLeftIntervalNode___py_5, __pyx_k_Int64ClosedLeftIntervalNode___py_5, sizeof(__pyx_k_Int64ClosedLeftIntervalNode___py_5), 0, 0, 1, 1},
  {&__pyx_kp_u_Int64ClosedLeftIntervalNode_piv, __pyx_k_Int64ClosedLeftIntervalNode_piv, sizeof(__pyx_k_Int64ClosedLeftIntervalNode_piv), 0, 1, 0, 0},
  {&__pyx_n_s_Int64ClosedLeftIntervalNode_quer, __pyx_k_Int64ClosedLeftIntervalNode_quer, sizeof(__pyx_k_Int64ClosedLeftIntervalNode_quer), 0, 0, 1, 1},
  {&__pyx_kp_u_Int64ClosedLeftIntervalNode_s_e, __pyx_k_Int64ClosedLeftIntervalNode_s_e, sizeof(__pyx_k_Int64ClosedLeftIntervalNode_s_e), 0, 1, 0, 0},
  {&__pyx_kp_u_Int64ClosedNeitherIntervalNode, __pyx_k_Int64ClosedNeitherIntervalNode, sizeof(__pyx_k_Int64ClosedNeitherIntervalNode), 0, 1, 0, 0},
  {&__pyx_kp_u_Int64ClosedNeitherIntervalNode_2, __pyx_k_Int64ClosedNeitherIntervalNode_2, sizeof(__pyx_k_Int64ClosedNeitherIntervalNode_2), 0, 1, 0, 0},
  {&__pyx_n_s_Int64ClosedNeitherIntervalNode_3, __pyx_k_Int64ClosedNeitherIntervalNode_3, sizeof(__pyx_k_Int64ClosedNeitherIntervalNode_3), 0, 0, 1, 1},
  {&__pyx_n_s_Int64ClosedNeitherIntervalNode_4, __pyx_k_Int64ClosedNeitherIntervalNode_4, sizeof(__pyx_k_Int64ClosedNeitherIntervalNode_4), 0, 0, 1, 1},
  {&__pyx_n_s_Int64ClosedNeitherIntervalNode_5, __pyx_k_Int64ClosedNeitherIntervalNode_5, sizeof(__pyx_k_Int64ClosedNeitherIntervalNode_5), 0, 0, 1, 1},
  {&__pyx_n_s_Int64ClosedNeitherIntervalNode_6, __pyx_k_Int64ClosedNeitherIntervalNode_6, sizeof(__pyx_k_Int64ClosedNeitherIntervalNode_6), 0, 0, 1, 1},
  {&__pyx_n_s_Int64ClosedNeitherIntervalNode_7, __pyx_k_Int64ClosedNeitherIntervalNode_7, sizeof(__pyx_k_Int64ClosedNeitherIntervalNode_7), 0, 0, 1, 1},
  {&__pyx_n_s_Int64ClosedNeitherIntervalNode_8, __pyx_k_Int64ClosedNeitherIntervalNode_8, sizeof(__pyx_k_Int64ClosedNeitherIntervalNode_8), 0, 0, 1, 1},
  {&__pyx_n_s_Int64ClosedNeitherIntervalNode_q, __pyx_k_Int64ClosedNeitherIntervalNode_q, sizeof(__pyx_k_Int64ClosedNeitherIntervalNode_q), 0, 0, 1, 1},
  {&__pyx_n_s_Int64ClosedRightIntervalNode, __pyx_k_Int64ClosedRightIntervalNode, sizeof(__pyx_k_Int64ClosedRightIntervalNode), 0, 0, 1, 1},
  {&__pyx_n_s_Int64ClosedRightIntervalNode___p, __pyx_k_Int64ClosedRightIntervalNode___p, sizeof(__pyx_k_Int64ClosedRightIntervalNode___p), 0, 0, 1, 1},
  {&__pyx_n_s_Int64ClosedRightIntervalNode___p_2, __pyx_k_Int64ClosedRightIntervalNode___p_2, sizeof(__pyx_k_Int64ClosedRightIntervalNode___p_2), 0, 0, 1, 1},
  {&__pyx_n_s_Int64ClosedRightIntervalNode___p_3, __pyx_k_Int64ClosedRightIntervalNode___p_3, sizeof(__pyx_k_Int64ClosedRightIntervalNode___p_3), 0, 0, 1, 1},
  {&__pyx_n_s_Int64ClosedRightIntervalNode___p_4, __pyx_k_Int64ClosedRightIntervalNode___p_4, sizeof(__pyx_k_Int64ClosedRightIntervalNode___p_4), 0, 0, 1, 1},
  {&__pyx_n_s_Int64ClosedRightIntervalNode___p_5, __pyx_k_Int64ClosedRightIntervalNode___p_5, sizeof(__pyx_k_Int64ClosedRightIntervalNode___p_5), 0, 0, 1, 1},
  {&__pyx_kp_u_Int64ClosedRightIntervalNode_pi, __pyx_k_Int64ClosedRightIntervalNode_pi, sizeof(__pyx_k_Int64ClosedRightIntervalNode_pi), 0, 1, 0, 0},
  {&__pyx_n_s_Int64ClosedRightIntervalNode_que, __pyx_k_Int64ClosedRightIntervalNode_que, sizeof(__pyx_k_Int64ClosedRightIntervalNode_que), 0, 0, 1, 1},
  {&__pyx_kp_u_Int64ClosedRightIntervalNode_s, __pyx_k_Int64ClosedRightIntervalNode_s, sizeof(__pyx_k_Int64ClosedRightIntervalNode_s), 0, 1, 0, 0},
  {&__pyx_n_s_Interval, __pyx_k_Interval, sizeof(__pyx_k_Interval), 0, 0, 1, 1},
  {&__pyx_n_s_IntervalMixin, __pyx_k_IntervalMixin, sizeof(__pyx_k_IntervalMixin), 0, 0, 1, 1},
  {&__pyx_kp_u_IntervalMixin_is_empty___get___l, __pyx_k_IntervalMixin_is_empty___get___l, sizeof(__pyx_k_IntervalMixin_is_empty___get___l), 0, 1, 0, 0},
  {&__pyx_n_s_IntervalTree, __pyx_k_IntervalTree, sizeof(__pyx_k_IntervalTree), 0, 0, 1, 1},
  {&__pyx_kp_u_IntervalTree_dtype_closed_n_ele, __pyx_k_IntervalTree_dtype_closed_n_ele, sizeof(__pyx_k_IntervalTree_dtype_closed_n_ele), 0, 1, 0, 0},
  {&__pyx_n_s_IntervalTree_get_indexer, __pyx_k_IntervalTree_get_indexer, sizeof(__pyx_k_IntervalTree_get_indexer), 0, 0, 1, 1},
  {&__pyx_n_s_IntervalTree_get_indexer_non_uni, __pyx_k_IntervalTree_get_indexer_non_uni, sizeof(__pyx_k_IntervalTree_get_indexer_non_uni), 0, 0, 1, 1},
  {&__pyx_n_s_IntervalTree_get_loc, __pyx_k_IntervalTree_get_loc, sizeof(__pyx_k_IntervalTree_get_loc), 0, 0, 1, 1},
  {&__pyx_kp_u_Interval_overlaps_line_427, __pyx_k_Interval_overlaps_line_427, sizeof(__pyx_k_Interval_overlaps_line_427), 0, 1, 0, 0},
  {&__pyx_kp_s_Invalid_mode_expected_c_or_fortr, __pyx_k_Invalid_mode_expected_c_or_fortr, sizeof(__pyx_k_Invalid_mode_expected_c_or_fortr), 0, 0, 1, 0},
  {&__pyx_kp_s_Invalid_shape_in_axis_d_d, __pyx_k_Invalid_shape_in_axis_d_d, sizeof(__pyx_k_Invalid_shape_in_axis_d_d), 0, 0, 1, 0},
  {&__pyx_n_s_KeyError, __pyx_k_KeyError, sizeof(__pyx_k_KeyError), 0, 0, 1, 1},
  {&__pyx_n_s_MemoryError, __pyx_k_MemoryError, sizeof(__pyx_k_MemoryError), 0, 0, 1, 1},
  {&__pyx_kp_s_MemoryView_of_r_at_0x_x, __pyx_k_MemoryView_of_r_at_0x_x, sizeof(__pyx_k_MemoryView_of_r_at_0x_x), 0, 0, 1, 0},
  {&__pyx_kp_s_MemoryView_of_r_object, __pyx_k_MemoryView_of_r_object, sizeof(__pyx_k_MemoryView_of_r_object), 0, 0, 1, 0},
  {&__pyx_n_s_NODE_CLASSES, __pyx_k_NODE_CLASSES, sizeof(__pyx_k_NODE_CLASSES), 0, 0, 1, 1},
  {&__pyx_kp_s_No_matching_signature_found, __pyx_k_No_matching_signature_found, sizeof(__pyx_k_No_matching_signature_found), 0, 0, 1, 0},
  {&__pyx_kp_u_Non_native_byte_order_not_suppor, __pyx_k_Non_native_byte_order_not_suppor, sizeof(__pyx_k_Non_native_byte_order_not_suppor), 0, 1, 0, 0},
  {&__pyx_n_s_NotImplemented, __pyx_k_NotImplemented, sizeof(__pyx_k_NotImplemented), 0, 0, 1, 1},
  {&__pyx_n_s_Number, __pyx_k_Number, sizeof(__pyx_k_Number), 0, 0, 1, 1},
  {&__pyx_n_b_O, __pyx_k_O, sizeof(__pyx_k_O), 0, 0, 0, 1},
  {&__pyx_kp_u_Only_numeric_Timestamp_and_Timed, __pyx_k_Only_numeric_Timestamp_and_Timed, sizeof(__pyx_k_Only_numeric_Timestamp_and_Timed), 0, 1, 0, 0},
  {&__pyx_kp_s_Out_of_bounds_on_buffer_access_a, __pyx_k_Out_of_bounds_on_buffer_access_a, sizeof(__pyx_k_Out_of_bounds_on_buffer_access_a), 0, 0, 1, 0},
  {&__pyx_n_s_PickleError, __pyx_k_PickleError, sizeof(__pyx_k_PickleError), 0, 0, 1, 1},
  {&__pyx_n_s_RuntimeError, __pyx_k_RuntimeError, sizeof(__pyx_k_RuntimeError), 0, 0, 1, 1},
  {&__pyx_n_s_Timedelta, __pyx_k_Timedelta, sizeof(__pyx_k_Timedelta), 0, 0, 1, 1},
  {&__pyx_n_s_Timestamp, __pyx_k_Timestamp, sizeof(__pyx_k_Timestamp), 0, 0, 1, 1},
  {&__pyx_n_s_TypeError, __pyx_k_TypeError, sizeof(__pyx_k_TypeError), 0, 0, 1, 1},
  {&__pyx_n_s_Uint64ClosedBothIntervalNode, __pyx_k_Uint64ClosedBothIntervalNode, sizeof(__pyx_k_Uint64ClosedBothIntervalNode), 0, 0, 1, 1},
  {&__pyx_n_s_Uint64ClosedBothIntervalNode___p, __pyx_k_Uint64ClosedBothIntervalNode___p, sizeof(__pyx_k_Uint64ClosedBothIntervalNode___p), 0, 0, 1, 1},
  {&__pyx_n_s_Uint64ClosedBothIntervalNode___p_2, __pyx_k_Uint64ClosedBothIntervalNode___p_2, sizeof(__pyx_k_Uint64ClosedBothIntervalNode___p_2), 0, 0, 1, 1},
  {&__pyx_n_s_Uint64ClosedBothIntervalNode___p_3, __pyx_k_Uint64ClosedBothIntervalNode___p_3, sizeof(__pyx_k_Uint64ClosedBothIntervalNode___p_3), 0, 0, 1, 1},
  {&__pyx_n_s_Uint64ClosedBothIntervalNode___p_4, __pyx_k_Uint64ClosedBothIntervalNode___p_4, sizeof(__pyx_k_Uint64ClosedBothIntervalNode___p_4), 0, 0, 1, 1},
  {&__pyx_n_s_Uint64ClosedBothIntervalNode___p_5, __pyx_k_Uint64ClosedBothIntervalNode___p_5, sizeof(__pyx_k_Uint64ClosedBothIntervalNode___p_5), 0, 0, 1, 1},
  {&__pyx_kp_u_Uint64ClosedBothIntervalNode_pi, __pyx_k_Uint64ClosedBothIntervalNode_pi, sizeof(__pyx_k_Uint64ClosedBothIntervalNode_pi), 0, 1, 0, 0},
  {&__pyx_n_s_Uint64ClosedBothIntervalNode_que, __pyx_k_Uint64ClosedBothIntervalNode_que, sizeof(__pyx_k_Uint64ClosedBothIntervalNode_que), 0, 0, 1, 1},
  {&__pyx_kp_u_Uint64ClosedBothIntervalNode_s, __pyx_k_Uint64ClosedBothIntervalNode_s, sizeof(__pyx_k_Uint64ClosedBothIntervalNode_s), 0, 1, 0, 0},
  {&__pyx_n_s_Uint64ClosedLeftIntervalNode, __pyx_k_Uint64ClosedLeftIntervalNode, sizeof(__pyx_k_Uint64ClosedLeftIntervalNode), 0, 0, 1, 1},
  {&__pyx_n_s_Uint64ClosedLeftIntervalNode___p, __pyx_k_Uint64ClosedLeftIntervalNode___p, sizeof(__pyx_k_Uint64ClosedLeftIntervalNode___p), 0, 0, 1, 1},
  {&__pyx_n_s_Uint64ClosedLeftIntervalNode___p_2, __pyx_k_Uint64ClosedLeftIntervalNode___p_2, sizeof(__pyx_k_Uint64ClosedLeftIntervalNode___p_2), 0, 0, 1, 1},
  {&__pyx_n_s_Uint64ClosedLeftIntervalNode___p_3, __pyx_k_Uint64ClosedLeftIntervalNode___p_3, sizeof(__pyx_k_Uint64ClosedLeftIntervalNode___p_3), 0, 0, 1, 1},
  {&__pyx_n_s_Uint64ClosedLeftIntervalNode___p_4, __pyx_k_Uint64ClosedLeftIntervalNode___p_4, sizeof(__pyx_k_Uint64ClosedLeftIntervalNode___p_4), 0, 0, 1, 1},
  {&__pyx_n_s_Uint64ClosedLeftIntervalNode___p_5, __pyx_k_Uint64ClosedLeftIntervalNode___p_5, sizeof(__pyx_k_Uint64ClosedLeftIntervalNode___p_5), 0, 0, 1, 1},
  {&__pyx_kp_u_Uint64ClosedLeftIntervalNode_pi, __pyx_k_Uint64ClosedLeftIntervalNode_pi, sizeof(__pyx_k_Uint64ClosedLeftIntervalNode_pi), 0, 1, 0, 0},
  {&__pyx_n_s_Uint64ClosedLeftIntervalNode_que, __pyx_k_Uint64ClosedLeftIntervalNode_que, sizeof(__pyx_k_Uint64ClosedLeftIntervalNode_que), 0, 0, 1, 1},
  {&__pyx_kp_u_Uint64ClosedLeftIntervalNode_s, __pyx_k_Uint64ClosedLeftIntervalNode_s, sizeof(__pyx_k_Uint64ClosedLeftIntervalNode_s), 0, 1, 0, 0},
  {&__pyx_kp_u_Uint64ClosedNeitherIntervalNode, __pyx_k_Uint64ClosedNeitherIntervalNode, sizeof(__pyx_k_Uint64ClosedNeitherIntervalNode), 0, 1, 0, 0},
  {&__pyx_kp_u_Uint64ClosedNeitherIntervalNode_2, __pyx_k_Uint64ClosedNeitherIntervalNode_2, sizeof(__pyx_k_Uint64ClosedNeitherIntervalNode_2), 0, 1, 0, 0},
  {&__pyx_n_s_Uint64ClosedNeitherIntervalNode_3, __pyx_k_Uint64ClosedNeitherIntervalNode_3, sizeof(__pyx_k_Uint64ClosedNeitherIntervalNode_3), 0, 0, 1, 1},
  {&__pyx_n_s_Uint64ClosedNeitherIntervalNode_4, __pyx_k_Uint64ClosedNeitherIntervalNode_4, sizeof(__pyx_k_Uint64ClosedNeitherIntervalNode_4), 0, 0, 1, 1},
  {&__pyx_n_s_Uint64ClosedNeitherIntervalNode_5, __pyx_k_Uint64ClosedNeitherIntervalNode_5, sizeof(__pyx_k_Uint64ClosedNeitherIntervalNode_5), 0, 0, 1, 1},
  {&__pyx_n_s_Uint64ClosedNeitherIntervalNode_6, __pyx_k_Uint64ClosedNeitherIntervalNode_6, sizeof(__pyx_k_Uint64ClosedNeitherIntervalNode_6), 0, 0, 1, 1},
  {&__pyx_n_s_Uint64ClosedNeitherIntervalNode_7, __pyx_k_Uint64ClosedNeitherIntervalNode_7, sizeof(__pyx_k_Uint64ClosedNeitherIntervalNode_7), 0, 0, 1, 1},
  {&__pyx_n_s_Uint64ClosedNeitherIntervalNode_8, __pyx_k_Uint64ClosedNeitherIntervalNode_8, sizeof(__pyx_k_Uint64ClosedNeitherIntervalNode_8), 0, 0, 1, 1},
  {&__pyx_n_s_Uint64ClosedNeitherIntervalNode_9, __pyx_k_Uint64ClosedNeitherIntervalNode_9, sizeof(__pyx_k_Uint64ClosedNeitherIntervalNode_9), 0, 0, 1, 1},
  {&__pyx_n_s_Uint64ClosedRightIntervalNode, __pyx_k_Uint64ClosedRightIntervalNode, sizeof(__pyx_k_Uint64ClosedRightIntervalNode), 0, 0, 1, 1},
  {&__pyx_n_s_Uint64ClosedRightIntervalNode_2, __pyx_k_Uint64ClosedRightIntervalNode_2, sizeof(__pyx_k_Uint64ClosedRightIntervalNode_2), 0, 0, 1, 1},
  {&__pyx_n_s_Uint64ClosedRightIntervalNode_3, __pyx_k_Uint64ClosedRightIntervalNode_3, sizeof(__pyx_k_Uint64ClosedRightIntervalNode_3), 0, 0, 1, 1},
  {&__pyx_n_s_Uint64ClosedRightIntervalNode_4, __pyx_k_Uint64ClosedRightIntervalNode_4, sizeof(__pyx_k_Uint64ClosedRightIntervalNode_4), 0, 0, 1, 1},
  {&__pyx_n_s_Uint64ClosedRightIntervalNode_5, __pyx_k_Uint64ClosedRightIntervalNode_5, sizeof(__pyx_k_Uint64ClosedRightIntervalNode_5), 0, 0, 1, 1},
  {&__pyx_n_s_Uint64ClosedRightIntervalNode_6, __pyx_k_Uint64ClosedRightIntervalNode_6, sizeof(__pyx_k_Uint64ClosedRightIntervalNode_6), 0, 0, 1, 1},
  {&__pyx_kp_u_Uint64ClosedRightIntervalNode_p, __pyx_k_Uint64ClosedRightIntervalNode_p, sizeof(__pyx_k_Uint64ClosedRightIntervalNode_p), 0, 1, 0, 0},
  {&__pyx_n_s_Uint64ClosedRightIntervalNode_qu, __pyx_k_Uint64ClosedRightIntervalNode_qu, sizeof(__pyx_k_Uint64ClosedRightIntervalNode_qu), 0, 0, 1, 1},
  {&__pyx_kp_u_Uint64ClosedRightIntervalNode_s, __pyx_k_Uint64ClosedRightIntervalNode_s, sizeof(__pyx_k_Uint64ClosedRightIntervalNode_s), 0, 1, 0, 0},
  {&__pyx_kp_s_Unable_to_convert_item_to_object, __pyx_k_Unable_to_convert_item_to_object, sizeof(__pyx_k_Unable_to_convert_item_to_object), 0, 0, 1, 0},
  {&__pyx_n_s_VALID_CLOSED, __pyx_k_VALID_CLOSED, sizeof(__pyx_k_VALID_CLOSED), 0, 0, 1, 1},
  {&__pyx_n_s_ValueError, __pyx_k_ValueError, sizeof(__pyx_k_ValueError), 0, 0, 1, 1},
  {&__pyx_n_s_View_MemoryView, __pyx_k_View_MemoryView, sizeof(__pyx_k_View_MemoryView), 0, 0, 1, 1},
  {&__pyx_kp_u__10, __pyx_k__10, sizeof(__pyx_k__10), 0, 1, 0, 0},
  {&__pyx_kp_s__13, __pyx_k__13, sizeof(__pyx_k__13), 0, 0, 1, 0},
  {&__pyx_kp_s__14, __pyx_k__14, sizeof(__pyx_k__14), 0, 0, 1, 0},
  {&__pyx_kp_u__18, __pyx_k__18, sizeof(__pyx_k__18), 0, 1, 0, 0},
  {&__pyx_kp_u__3, __pyx_k__3, sizeof(__pyx_k__3), 0, 1, 0, 0},
  {&__pyx_kp_u__4, __pyx_k__4, sizeof(__pyx_k__4), 0, 1, 0, 0},
  {&__pyx_kp_u__5, __pyx_k__5, sizeof(__pyx_k__5), 0, 1, 0, 0},
  {&__pyx_kp_u__6, __pyx_k__6, sizeof(__pyx_k__6), 0, 1, 0, 0},
  {&__pyx_kp_u__7, __pyx_k__7, sizeof(__pyx_k__7), 0, 1, 0, 0},
  {&__pyx_kp_u__8, __pyx_k__8, sizeof(__pyx_k__8), 0, 1, 0, 0},
  {&__pyx_kp_u__9, __pyx_k__9, sizeof(__pyx_k__9), 0, 1, 0, 0},
  {&__pyx_n_s_allocate_buffer, __pyx_k_allocate_buffer, sizeof(__pyx_k_allocate_buffer), 0, 0, 1, 1},
  {&__pyx_n_s_any, __pyx_k_any, sizeof(__pyx_k_any), 0, 0, 1, 1},
  {&__pyx_n_s_arange, __pyx_k_arange, sizeof(__pyx_k_arange), 0, 0, 1, 1},
  {&__pyx_n_s_args, __pyx_k_args, sizeof(__pyx_k_args), 0, 0, 1, 1},
  {&__pyx_n_s_argsort, __pyx_k_argsort, sizeof(__pyx_k_argsort), 0, 0, 1, 1},
  {&__pyx_n_s_asarray, __pyx_k_asarray, sizeof(__pyx_k_asarray), 0, 0, 1, 1},
  {&__pyx_n_s_astype, __pyx_k_astype, sizeof(__pyx_k_astype), 0, 0, 1, 1},
  {&__pyx_n_s_base, __pyx_k_base, sizeof(__pyx_k_base), 0, 0, 1, 1},
  {&__pyx_n_u_both, __pyx_k_both, sizeof(__pyx_k_both), 0, 1, 0, 1},
  {&__pyx_n_s_c, __pyx_k_c, sizeof(__pyx_k_c), 0, 0, 1, 1},
  {&__pyx_n_u_c, __pyx_k_c, sizeof(__pyx_k_c), 0, 1, 0, 1},
  {&__pyx_n_s_class, __pyx_k_class, sizeof(__pyx_k_class), 0, 0, 1, 1},
  {&__pyx_n_s_cline_in_traceback, __pyx_k_cline_in_traceback, sizeof(__pyx_k_cline_in_traceback), 0, 0, 1, 1},
  {&__pyx_n_s_closed, __pyx_k_closed, sizeof(__pyx_k_closed), 0, 0, 1, 1},
  {&__pyx_n_u_closed, __pyx_k_closed, sizeof(__pyx_k_closed), 0, 1, 0, 1},
  {&__pyx_kp_u_closed_is_expected, __pyx_k_closed_is_expected, sizeof(__pyx_k_closed_is_expected), 0, 1, 0, 0},
  {&__pyx_n_s_closed_left, __pyx_k_closed_left, sizeof(__pyx_k_closed_left), 0, 0, 1, 1},
  {&__pyx_n_s_closed_right, __pyx_k_closed_right, sizeof(__pyx_k_closed_right), 0, 0, 1, 1},
  {&__pyx_n_s_concatenate, __pyx_k_concatenate, sizeof(__pyx_k_concatenate), 0, 0, 1, 1},
  {&__pyx_kp_u_contains___not_defined_for_two, __pyx_k_contains___not_defined_for_two, sizeof(__pyx_k_contains___not_defined_for_two), 0, 1, 0, 0},
  {&__pyx_kp_s_contiguous_and_direct, __pyx_k_contiguous_and_direct, sizeof(__pyx_k_contiguous_and_direct), 0, 0, 1, 0},
  {&__pyx_kp_s_contiguous_and_indirect, __pyx_k_contiguous_and_indirect, sizeof(__pyx_k_contiguous_and_indirect), 0, 0, 1, 0},
  {&__pyx_n_s_counts, __pyx_k_counts, sizeof(__pyx_k_counts), 0, 0, 1, 1},
  {&__pyx_n_u_dateoffset, __pyx_k_dateoffset, sizeof(__pyx_k_dateoffset), 0, 1, 0, 1},
  {&__pyx_n_s_defaults, __pyx_k_defaults, sizeof(__pyx_k_defaults), 0, 0, 1, 1},
  {&__pyx_n_s_dict, __pyx_k_dict, sizeof(__pyx_k_dict), 0, 0, 1, 1},
  {&__pyx_n_s_dtype, __pyx_k_dtype, sizeof(__pyx_k_dtype), 0, 0, 1, 1},
  {&__pyx_n_s_dtype_is_object, __pyx_k_dtype_is_object, sizeof(__pyx_k_dtype_is_object), 0, 0, 1, 1},
  {&__pyx_kp_u_elements, __pyx_k_elements, sizeof(__pyx_k_elements), 0, 1, 0, 0},
  {&__pyx_n_s_empty, __pyx_k_empty, sizeof(__pyx_k_empty), 0, 0, 1, 1},
  {&__pyx_n_s_encode, __pyx_k_encode, sizeof(__pyx_k_encode), 0, 0, 1, 1},
  {&__pyx_n_s_end, __pyx_k_end, sizeof(__pyx_k_end), 0, 0, 1, 1},
  {&__pyx_n_s_enumerate, __pyx_k_enumerate, sizeof(__pyx_k_enumerate), 0, 0, 1, 1},
  {&__pyx_n_s_error, __pyx_k_error, sizeof(__pyx_k_error), 0, 0, 1, 1},
  {&__pyx_n_s_flags, __pyx_k_flags, sizeof(__pyx_k_flags), 0, 0, 1, 1},
  {&__pyx_n_u_float32, __pyx_k_float32, sizeof(__pyx_k_float32), 0, 1, 0, 1},
  {&__pyx_n_s_float32_t, __pyx_k_float32_t, sizeof(__pyx_k_float32_t), 0, 0, 1, 1},
  {&__pyx_n_u_float64, __pyx_k_float64, sizeof(__pyx_k_float64), 0, 1, 0, 1},
  {&__pyx_n_s_float64_t, __pyx_k_float64_t, sizeof(__pyx_k_float64_t), 0, 0, 1, 1},
  {&__pyx_n_s_format, __pyx_k_format, sizeof(__pyx_k_format), 0, 0, 1, 1},
  {&__pyx_n_s_fortran, __pyx_k_fortran, sizeof(__pyx_k_fortran), 0, 0, 1, 1},
  {&__pyx_n_u_fortran, __pyx_k_fortran, sizeof(__pyx_k_fortran), 0, 1, 0, 1},
  {&__pyx_n_s_get_indexer, __pyx_k_get_indexer, sizeof(__pyx_k_get_indexer), 0, 0, 1, 1},
  {&__pyx_n_s_get_indexer_non_unique, __pyx_k_get_indexer_non_unique, sizeof(__pyx_k_get_indexer_non_unique), 0, 0, 1, 1},
  {&__pyx_n_s_get_loc, __pyx_k_get_loc, sizeof(__pyx_k_get_loc), 0, 0, 1, 1},
  {&__pyx_n_s_get_partial_overlap, __pyx_k_get_partial_overlap, sizeof(__pyx_k_get_partial_overlap), 0, 0, 1, 1},
  {&__pyx_n_s_getstate, __pyx_k_getstate, sizeof(__pyx_k_getstate), 0, 0, 1, 1},
  {&__pyx_kp_s_got_differing_extents_in_dimensi, __pyx_k_got_differing_extents_in_dimensi, sizeof(__pyx_k_got_differing_extents_in_dimensi), 0, 0, 1, 0},
  {&__pyx_n_s_i, __pyx_k_i, sizeof(__pyx_k_i), 0, 0, 1, 1},
  {&__pyx_n_s_id, __pyx_k_id, sizeof(__pyx_k_id), 0, 0, 1, 1},
  {&__pyx_n_s_import, __pyx_k_import, sizeof(__pyx_k_import), 0, 0, 1, 1},
  {&__pyx_kp_u_index_out_of_bounds, __pyx_k_index_out_of_bounds, sizeof(__pyx_k_index_out_of_bounds), 0, 1, 0, 0},
  {&__pyx_kp_u_indexer_does_not_intersect_a_uni, __pyx_k_indexer_does_not_intersect_a_uni, sizeof(__pyx_k_indexer_does_not_intersect_a_uni), 0, 1, 0, 0},
  {&__pyx_n_s_indices, __pyx_k_indices, sizeof(__pyx_k_indices), 0, 0, 1, 1},
  {&__pyx_n_u_int32, __pyx_k_int32, sizeof(__pyx_k_int32), 0, 1, 0, 1},
  {&__pyx_n_s_int32_t, __pyx_k_int32_t, sizeof(__pyx_k_int32_t), 0, 0, 1, 1},
  {&__pyx_n_u_int64, __pyx_k_int64, sizeof(__pyx_k_int64), 0, 1, 0, 1},
  {&__pyx_n_s_int64_t, __pyx_k_int64_t, sizeof(__pyx_k_int64_t), 0, 0, 1, 1},
  {&__pyx_n_s_interval, __pyx_k_interval, sizeof(__pyx_k_interval), 0, 0, 1, 1},
  {&__pyx_n_u_interval, __pyx_k_interval, sizeof(__pyx_k_interval), 0, 1, 0, 1},
  {&__pyx_n_s_intervals, __pyx_k_intervals, sizeof(__pyx_k_intervals), 0, 0, 1, 1},
  {&__pyx_kp_u_intervals_must_all_be_closed_on, __pyx_k_intervals_must_all_be_closed_on, sizeof(__pyx_k_intervals_must_all_be_closed_on), 0, 1, 0, 0},
  {&__pyx_n_s_intervals_to_interval_bounds, __pyx_k_intervals_to_interval_bounds, sizeof(__pyx_k_intervals_to_interval_bounds), 0, 0, 1, 1},
  {&__pyx_n_u_intp, __pyx_k_intp, sizeof(__pyx_k_intp), 0, 1, 0, 1},
  {&__pyx_kp_u_invalid_option_for_closed_closed, __pyx_k_invalid_option_for_closed_closed, sizeof(__pyx_k_invalid_option_for_closed_closed), 0, 1, 0, 0},
  {&__pyx_kp_u_invalid_option_for_closed_s, __pyx_k_invalid_option_for_closed_s, sizeof(__pyx_k_invalid_option_for_closed_s), 0, 1, 0, 0},
  {&__pyx_n_s_is_monotonic, __pyx_k_is_monotonic, sizeof(__pyx_k_is_monotonic), 0, 0, 1, 1},
  {&__pyx_n_s_isnan, __pyx_k_isnan, sizeof(__pyx_k_isnan), 0, 0, 1, 1},
  {&__pyx_n_s_itemsize, __pyx_k_itemsize, sizeof(__pyx_k_itemsize), 0, 0, 1, 1},
  {&__pyx_kp_s_itemsize_0_for_cython_array, __pyx_k_itemsize_0_for_cython_array, sizeof(__pyx_k_itemsize_0_for_cython_array), 0, 0, 1, 0},
  {&__pyx_n_s_iv, __pyx_k_iv, sizeof(__pyx_k_iv), 0, 0, 1, 1},
  {&__pyx_n_s_key, __pyx_k_key, sizeof(__pyx_k_key), 0, 0, 1, 1},
  {&__pyx_n_s_key_left, __pyx_k_key_left, sizeof(__pyx_k_key_left), 0, 0, 1, 1},
  {&__pyx_n_s_key_right, __pyx_k_key_right, sizeof(__pyx_k_key_right), 0, 0, 1, 1},
  {&__pyx_n_s_kind, __pyx_k_kind, sizeof(__pyx_k_kind), 0, 0, 1, 1},
  {&__pyx_n_s_kwargs, __pyx_k_kwargs, sizeof(__pyx_k_kwargs), 0, 0, 1, 1},
  {&__pyx_n_s_le, __pyx_k_le, sizeof(__pyx_k_le), 0, 0, 1, 1},
  {&__pyx_n_s_leaf_size, __pyx_k_leaf_size, sizeof(__pyx_k_leaf_size), 0, 0, 1, 1},
  {&__pyx_n_s_left, __pyx_k_left, sizeof(__pyx_k_left), 0, 0, 1, 1},
  {&__pyx_n_u_left, __pyx_k_left, sizeof(__pyx_k_left), 0, 1, 0, 1},
  {&__pyx_kp_u_left_2, __pyx_k_left_2, sizeof(__pyx_k_left_2), 0, 1, 0, 0},
  {&__pyx_kp_u_left_and_right_must_have_the_sam, __pyx_k_left_and_right_must_have_the_sam, sizeof(__pyx_k_left_and_right_must_have_the_sam), 0, 1, 0, 0},
  {&__pyx_kp_u_left_side_of_interval_must_be_ri, __pyx_k_left_side_of_interval_must_be_ri, sizeof(__pyx_k_left_side_of_interval_must_be_ri), 0, 1, 0, 0},
  {&__pyx_n_s_left_sorter, __pyx_k_left_sorter, sizeof(__pyx_k_left_sorter), 0, 0, 1, 1},
  {&__pyx_n_s_left_tz, __pyx_k_left_tz, sizeof(__pyx_k_left_tz), 0, 0, 1, 1},
  {&__pyx_n_s_length, __pyx_k_length, sizeof(__pyx_k_length), 0, 0, 1, 1},
  {&__pyx_n_s_lexsort, __pyx_k_lexsort, sizeof(__pyx_k_lexsort), 0, 0, 1, 1},
  {&__pyx_n_s_lt, __pyx_k_lt, sizeof(__pyx_k_lt), 0, 0, 1, 1},
  {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
  {&__pyx_n_s_max, __pyx_k_max, sizeof(__pyx_k_max), 0, 0, 1, 1},
  {&__pyx_n_s_median, __pyx_k_median, sizeof(__pyx_k_median), 0, 0, 1, 1},
  {&__pyx_n_s_memview, __pyx_k_memview, sizeof(__pyx_k_memview), 0, 0, 1, 1},
  {&__pyx_n_s_min, __pyx_k_min, sizeof(__pyx_k_min), 0, 0, 1, 1},
  {&__pyx_n_s_missing, __pyx_k_missing, sizeof(__pyx_k_missing), 0, 0, 1, 1},
  {&__pyx_n_s_mode, __pyx_k_mode, sizeof(__pyx_k_mode), 0, 0, 1, 1},
  {&__pyx_n_s_msg, __pyx_k_msg, sizeof(__pyx_k_msg), 0, 0, 1, 1},
  {&__pyx_n_s_n, __pyx_k_n, sizeof(__pyx_k_n), 0, 0, 1, 1},
  {&__pyx_n_s_n_elements, __pyx_k_n_elements, sizeof(__pyx_k_n_elements), 0, 0, 1, 1},
  {&__pyx_n_s_name, __pyx_k_name, sizeof(__pyx_k_name), 0, 0, 1, 1},
  {&__pyx_n_s_name_2, __pyx_k_name_2, sizeof(__pyx_k_name_2), 0, 0, 1, 1},
  {&__pyx_kp_u_name_left_r_right_r_closed_clos, __pyx_k_name_left_r_right_r_closed_clos, sizeof(__pyx_k_name_left_r_right_r_closed_clos), 0, 1, 0, 0},
  {&__pyx_n_s_nan, __pyx_k_nan, sizeof(__pyx_k_nan), 0, 0, 1, 1},
  {&__pyx_kp_u_ndarray_is_not_C_contiguous, __pyx_k_ndarray_is_not_C_contiguous, sizeof(__pyx_k_ndarray_is_not_C_contiguous), 0, 1, 0, 0},
  {&__pyx_kp_u_ndarray_is_not_Fortran_contiguou, __pyx_k_ndarray_is_not_Fortran_contiguou, sizeof(__pyx_k_ndarray_is_not_Fortran_contiguou), 0, 1, 0, 0},
  {&__pyx_n_s_ndim, __pyx_k_ndim, sizeof(__pyx_k_ndim), 0, 0, 1, 1},
  {&__pyx_n_u_ndim, __pyx_k_ndim, sizeof(__pyx_k_ndim), 0, 1, 0, 1},
  {&__pyx_n_u_neither, __pyx_k_neither, sizeof(__pyx_k_neither), 0, 1, 0, 1},
  {&__pyx_n_s_new, __pyx_k_new, sizeof(__pyx_k_new), 0, 0, 1, 1},
  {&__pyx_kp_s_no_default___reduce___due_to_non, __pyx_k_no_default___reduce___due_to_non, sizeof(__pyx_k_no_default___reduce___due_to_non), 0, 0, 1, 0},
  {&__pyx_n_s_np, __pyx_k_np, sizeof(__pyx_k_np), 0, 0, 1, 1},
  {&__pyx_n_s_numbers, __pyx_k_numbers, sizeof(__pyx_k_numbers), 0, 0, 1, 1},
  {&__pyx_n_s_numpy, __pyx_k_numpy, sizeof(__pyx_k_numpy), 0, 0, 1, 1},
  {&__pyx_kp_u_numpy_core_multiarray_failed_to, __pyx_k_numpy_core_multiarray_failed_to, sizeof(__pyx_k_numpy_core_multiarray_failed_to), 0, 1, 0, 0},
  {&__pyx_kp_u_numpy_core_umath_failed_to_impor, __pyx_k_numpy_core_umath_failed_to_impor, sizeof(__pyx_k_numpy_core_umath_failed_to_impor), 0, 1, 0, 0},
  {&__pyx_n_s_obj, __pyx_k_obj, sizeof(__pyx_k_obj), 0, 0, 1, 1},
  {&__pyx_n_s_old_len, __pyx_k_old_len, sizeof(__pyx_k_old_len), 0, 0, 1, 1},
  {&__pyx_n_s_op, __pyx_k_op, sizeof(__pyx_k_op), 0, 0, 1, 1},
  {&__pyx_n_s_open_left, __pyx_k_open_left, sizeof(__pyx_k_open_left), 0, 0, 1, 1},
  {&__pyx_n_s_open_right, __pyx_k_open_right, sizeof(__pyx_k_open_right), 0, 0, 1, 1},
  {&__pyx_n_s_operator, __pyx_k_operator, sizeof(__pyx_k_operator), 0, 0, 1, 1},
  {&__pyx_n_s_other, __pyx_k_other, sizeof(__pyx_k_other), 0, 0, 1, 1},
  {&__pyx_n_u_other, __pyx_k_other, sizeof(__pyx_k_other), 0, 1, 0, 1},
  {&__pyx_kp_u_other_must_be_an_Interval_got_o, __pyx_k_other_must_be_an_Interval_got_o, sizeof(__pyx_k_other_must_be_an_Interval_got_o), 0, 1, 0, 0},
  {&__pyx_kp_u_overlapping, __pyx_k_overlapping, sizeof(__pyx_k_overlapping), 0, 1, 0, 0},
  {&__pyx_n_s_pack, __pyx_k_pack, sizeof(__pyx_k_pack), 0, 0, 1, 1},
  {&__pyx_n_s_pandas, __pyx_k_pandas, sizeof(__pyx_k_pandas), 0, 0, 1, 1},
  {&__pyx_n_s_pandas__libs_algos, __pyx_k_pandas__libs_algos, sizeof(__pyx_k_pandas__libs_algos), 0, 0, 1, 1},
  {&__pyx_n_s_pandas__libs_interval, __pyx_k_pandas__libs_interval, sizeof(__pyx_k_pandas__libs_interval), 0, 0, 1, 1},
  {&__pyx_kp_s_pandas__libs_interval_pyx, __pyx_k_pandas__libs_interval_pyx, sizeof(__pyx_k_pandas__libs_interval_pyx), 0, 0, 1, 0},
  {&__pyx_kp_s_pandas__libs_intervaltree_pxi, __pyx_k_pandas__libs_intervaltree_pxi, sizeof(__pyx_k_pandas__libs_intervaltree_pxi), 0, 0, 1, 0},
  {&__pyx_n_s_pandas__libs_tslibs, __pyx_k_pandas__libs_tslibs, sizeof(__pyx_k_pandas__libs_tslibs), 0, 0, 1, 1},
  {&__pyx_n_s_pandas__libs_tslibs_timedeltas, __pyx_k_pandas__libs_tslibs_timedeltas, sizeof(__pyx_k_pandas__libs_tslibs_timedeltas), 0, 0, 1, 1},
  {&__pyx_n_u_period, __pyx_k_period, sizeof(__pyx_k_period), 0, 1, 0, 1},
  {&__pyx_n_s_pickle, __pyx_k_pickle, sizeof(__pyx_k_pickle), 0, 0, 1, 1},
  {&__pyx_n_s_point, __pyx_k_point, sizeof(__pyx_k_point), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_PickleError, __pyx_k_pyx_PickleError, sizeof(__pyx_k_pyx_PickleError), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_checksum, __pyx_k_pyx_checksum, sizeof(__pyx_k_pyx_checksum), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_fuse_0query, __pyx_k_pyx_fuse_0query, sizeof(__pyx_k_pyx_fuse_0query), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_fuse_1query, __pyx_k_pyx_fuse_1query, sizeof(__pyx_k_pyx_fuse_1query), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_fuse_2query, __pyx_k_pyx_fuse_2query, sizeof(__pyx_k_pyx_fuse_2query), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_fuse_3query, __pyx_k_pyx_fuse_3query, sizeof(__pyx_k_pyx_fuse_3query), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_fuse_4query, __pyx_k_pyx_fuse_4query, sizeof(__pyx_k_pyx_fuse_4query), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_getbuffer, __pyx_k_pyx_getbuffer, sizeof(__pyx_k_pyx_getbuffer), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_result, __pyx_k_pyx_result, sizeof(__pyx_k_pyx_result), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_state, __pyx_k_pyx_state, sizeof(__pyx_k_pyx_state), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_type, __pyx_k_pyx_type, sizeof(__pyx_k_pyx_type), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_unpickle_Enum, __pyx_k_pyx_unpickle_Enum, sizeof(__pyx_k_pyx_unpickle_Enum), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_unpickle_Float32ClosedBoth, __pyx_k_pyx_unpickle_Float32ClosedBoth, sizeof(__pyx_k_pyx_unpickle_Float32ClosedBoth), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_unpickle_Float32ClosedLeft, __pyx_k_pyx_unpickle_Float32ClosedLeft, sizeof(__pyx_k_pyx_unpickle_Float32ClosedLeft), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_unpickle_Float32ClosedNeit, __pyx_k_pyx_unpickle_Float32ClosedNeit, sizeof(__pyx_k_pyx_unpickle_Float32ClosedNeit), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_unpickle_Float32ClosedRigh, __pyx_k_pyx_unpickle_Float32ClosedRigh, sizeof(__pyx_k_pyx_unpickle_Float32ClosedRigh), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_unpickle_Float64ClosedBoth, __pyx_k_pyx_unpickle_Float64ClosedBoth, sizeof(__pyx_k_pyx_unpickle_Float64ClosedBoth), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_unpickle_Float64ClosedLeft, __pyx_k_pyx_unpickle_Float64ClosedLeft, sizeof(__pyx_k_pyx_unpickle_Float64ClosedLeft), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_unpickle_Float64ClosedNeit, __pyx_k_pyx_unpickle_Float64ClosedNeit, sizeof(__pyx_k_pyx_unpickle_Float64ClosedNeit), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_unpickle_Float64ClosedRigh, __pyx_k_pyx_unpickle_Float64ClosedRigh, sizeof(__pyx_k_pyx_unpickle_Float64ClosedRigh), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_unpickle_Int32ClosedBothIn, __pyx_k_pyx_unpickle_Int32ClosedBothIn, sizeof(__pyx_k_pyx_unpickle_Int32ClosedBothIn), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_unpickle_Int32ClosedLeftIn, __pyx_k_pyx_unpickle_Int32ClosedLeftIn, sizeof(__pyx_k_pyx_unpickle_Int32ClosedLeftIn), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_unpickle_Int32ClosedNeithe, __pyx_k_pyx_unpickle_Int32ClosedNeithe, sizeof(__pyx_k_pyx_unpickle_Int32ClosedNeithe), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_unpickle_Int32ClosedRightI, __pyx_k_pyx_unpickle_Int32ClosedRightI, sizeof(__pyx_k_pyx_unpickle_Int32ClosedRightI), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_unpickle_Int64ClosedBothIn, __pyx_k_pyx_unpickle_Int64ClosedBothIn, sizeof(__pyx_k_pyx_unpickle_Int64ClosedBothIn), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_unpickle_Int64ClosedLeftIn, __pyx_k_pyx_unpickle_Int64ClosedLeftIn, sizeof(__pyx_k_pyx_unpickle_Int64ClosedLeftIn), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_unpickle_Int64ClosedNeithe, __pyx_k_pyx_unpickle_Int64ClosedNeithe, sizeof(__pyx_k_pyx_unpickle_Int64ClosedNeithe), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_unpickle_Int64ClosedRightI, __pyx_k_pyx_unpickle_Int64ClosedRightI, sizeof(__pyx_k_pyx_unpickle_Int64ClosedRightI), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_unpickle_IntervalMixin, __pyx_k_pyx_unpickle_IntervalMixin, sizeof(__pyx_k_pyx_unpickle_IntervalMixin), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_unpickle_IntervalTree, __pyx_k_pyx_unpickle_IntervalTree, sizeof(__pyx_k_pyx_unpickle_IntervalTree), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_unpickle_Uint64ClosedBothI, __pyx_k_pyx_unpickle_Uint64ClosedBothI, sizeof(__pyx_k_pyx_unpickle_Uint64ClosedBothI), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_unpickle_Uint64ClosedLeftI, __pyx_k_pyx_unpickle_Uint64ClosedLeftI, sizeof(__pyx_k_pyx_unpickle_Uint64ClosedLeftI), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_unpickle_Uint64ClosedNeith, __pyx_k_pyx_unpickle_Uint64ClosedNeith, sizeof(__pyx_k_pyx_unpickle_Uint64ClosedNeith), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_unpickle_Uint64ClosedRight, __pyx_k_pyx_unpickle_Uint64ClosedRight, sizeof(__pyx_k_pyx_unpickle_Uint64ClosedRight), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_vtable, __pyx_k_pyx_vtable, sizeof(__pyx_k_pyx_vtable), 0, 0, 1, 1},
  {&__pyx_n_s_query, __pyx_k_query, sizeof(__pyx_k_query), 0, 0, 1, 1},
  {&__pyx_n_s_range, __pyx_k_range, sizeof(__pyx_k_range), 0, 0, 1, 1},
  {&__pyx_n_s_reduce, __pyx_k_reduce, sizeof(__pyx_k_reduce), 0, 0, 1, 1},
  {&__pyx_n_s_reduce_cython, __pyx_k_reduce_cython, sizeof(__pyx_k_reduce_cython), 0, 0, 1, 1},
  {&__pyx_n_s_reduce_ex, __pyx_k_reduce_ex, sizeof(__pyx_k_reduce_ex), 0, 0, 1, 1},
  {&__pyx_n_s_repr_base, __pyx_k_repr_base, sizeof(__pyx_k_repr_base), 0, 0, 1, 1},
  {&__pyx_n_s_result, __pyx_k_result, sizeof(__pyx_k_result), 0, 0, 1, 1},
  {&__pyx_n_s_result_type, __pyx_k_result_type, sizeof(__pyx_k_result_type), 0, 0, 1, 1},
  {&__pyx_n_s_right, __pyx_k_right, sizeof(__pyx_k_right), 0, 0, 1, 1},
  {&__pyx_n_u_right, __pyx_k_right, sizeof(__pyx_k_right), 0, 1, 0, 1},
  {&__pyx_kp_u_right_2, __pyx_k_right_2, sizeof(__pyx_k_right_2), 0, 1, 0, 0},
  {&__pyx_n_s_right_sorter, __pyx_k_right_sorter, sizeof(__pyx_k_right_sorter), 0, 0, 1, 1},
  {&__pyx_n_s_right_tz, __pyx_k_right_tz, sizeof(__pyx_k_right_tz), 0, 0, 1, 1},
  {&__pyx_n_s_s, __pyx_k_s, sizeof(__pyx_k_s), 0, 0, 1, 1},
  {&__pyx_n_s_searchsorted, __pyx_k_searchsorted, sizeof(__pyx_k_searchsorted), 0, 0, 1, 1},
  {&__pyx_n_s_seen_closed, __pyx_k_seen_closed, sizeof(__pyx_k_seen_closed), 0, 0, 1, 1},
  {&__pyx_n_s_self, __pyx_k_self, sizeof(__pyx_k_self), 0, 0, 1, 1},
  {&__pyx_n_s_setstate, __pyx_k_setstate, sizeof(__pyx_k_setstate), 0, 0, 1, 1},
  {&__pyx_n_s_setstate_cython, __pyx_k_setstate_cython, sizeof(__pyx_k_setstate_cython), 0, 0, 1, 1},
  {&__pyx_n_s_shape, __pyx_k_shape, sizeof(__pyx_k_shape), 0, 0, 1, 1},
  {&__pyx_n_s_short_repr, __pyx_k_short_repr, sizeof(__pyx_k_short_repr), 0, 0, 1, 1},
  {&__pyx_n_s_side, __pyx_k_side, sizeof(__pyx_k_side), 0, 0, 1, 1},
  {&__pyx_n_s_size, __pyx_k_size, sizeof(__pyx_k_size), 0, 0, 1, 1},
  {&__pyx_n_s_sorter, __pyx_k_sorter, sizeof(__pyx_k_sorter), 0, 0, 1, 1},
  {&__pyx_n_s_split, __pyx_k_split, sizeof(__pyx_k_split), 0, 0, 1, 1},
  {&__pyx_n_s_start, __pyx_k_start, sizeof(__pyx_k_start), 0, 0, 1, 1},
  {&__pyx_kp_u_start_left_right_end, __pyx_k_start_left_right_end, sizeof(__pyx_k_start_left_right_end), 0, 1, 0, 0},
  {&__pyx_n_s_step, __pyx_k_step, sizeof(__pyx_k_step), 0, 0, 1, 1},
  {&__pyx_n_s_stop, __pyx_k_stop, sizeof(__pyx_k_stop), 0, 0, 1, 1},
  {&__pyx_kp_s_strided_and_direct, __pyx_k_strided_and_direct, sizeof(__pyx_k_strided_and_direct), 0, 0, 1, 0},
  {&__pyx_kp_s_strided_and_direct_or_indirect, __pyx_k_strided_and_direct_or_indirect, sizeof(__pyx_k_strided_and_direct_or_indirect), 0, 0, 1, 0},
  {&__pyx_kp_s_strided_and_indirect, __pyx_k_strided_and_indirect, sizeof(__pyx_k_strided_and_indirect), 0, 0, 1, 0},
  {&__pyx_kp_s_stringsource, __pyx_k_stringsource, sizeof(__pyx_k_stringsource), 0, 0, 1, 0},
  {&__pyx_n_s_strip, __pyx_k_strip, sizeof(__pyx_k_strip), 0, 0, 1, 1},
  {&__pyx_n_s_struct, __pyx_k_struct, sizeof(__pyx_k_struct), 0, 0, 1, 1},
  {&__pyx_n_s_target, __pyx_k_target, sizeof(__pyx_k_target), 0, 0, 1, 1},
  {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
  {&__pyx_n_s_typ, __pyx_k_typ, sizeof(__pyx_k_typ), 0, 0, 1, 1},
  {&__pyx_n_s_typ_2, __pyx_k_typ_2, sizeof(__pyx_k_typ_2), 0, 0, 1, 1},
  {&__pyx_n_u_typ_2, __pyx_k_typ_2, sizeof(__pyx_k_typ_2), 0, 1, 0, 1},
  {&__pyx_kp_u_type_typ_with_value_iv_is_not_an, __pyx_k_type_typ_with_value_iv_is_not_an, sizeof(__pyx_k_type_typ_with_value_iv_is_not_an), 0, 1, 0, 0},
  {&__pyx_n_s_tzinfo, __pyx_k_tzinfo, sizeof(__pyx_k_tzinfo), 0, 0, 1, 1},
  {&__pyx_n_u_uint64, __pyx_k_uint64, sizeof(__pyx_k_uint64), 0, 1, 0, 1},
  {&__pyx_n_s_uint64_t, __pyx_k_uint64_t, sizeof(__pyx_k_uint64_t), 0, 0, 1, 1},
  {&__pyx_kp_s_unable_to_allocate_array_data, __pyx_k_unable_to_allocate_array_data, sizeof(__pyx_k_unable_to_allocate_array_data), 0, 0, 1, 0},
  {&__pyx_kp_s_unable_to_allocate_shape_and_str, __pyx_k_unable_to_allocate_shape_and_str, sizeof(__pyx_k_unable_to_allocate_shape_and_str), 0, 0, 1, 0},
  {&__pyx_n_s_unique, __pyx_k_unique, sizeof(__pyx_k_unique), 0, 0, 1, 1},
  {&__pyx_kp_u_unknown_dtype_code_in_numpy_pxd, __pyx_k_unknown_dtype_code_in_numpy_pxd, sizeof(__pyx_k_unknown_dtype_code_in_numpy_pxd), 0, 1, 0, 0},
  {&__pyx_kp_u_unorderable_types_name_op_other, __pyx_k_unorderable_types_name_op_other, sizeof(__pyx_k_unorderable_types_name_op_other), 0, 1, 0, 0},
  {&__pyx_n_s_unpack, __pyx_k_unpack, sizeof(__pyx_k_unpack), 0, 0, 1, 1},
  {&__pyx_n_s_update, __pyx_k_update, sizeof(__pyx_k_update), 0, 0, 1, 1},
  {&__pyx_n_s_validate_closed, __pyx_k_validate_closed, sizeof(__pyx_k_validate_closed), 0, 0, 1, 1},
  {&__pyx_n_s_validate_endpoint, __pyx_k_validate_endpoint, sizeof(__pyx_k_validate_endpoint), 0, 0, 1, 1},
  {0, 0, 0, 0, 0, 0, 0}
};
static CYTHON_SMALL_CODE int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_TypeError = __Pyx_GetBuiltinName(__pyx_n_s_TypeError); if (!__pyx_builtin_TypeError) __PYX_ERR(0, 101, __pyx_L1_error)
  __pyx_builtin_ValueError = __Pyx_GetBuiltinName(__pyx_n_s_ValueError); if (!__pyx_builtin_ValueError) __PYX_ERR(0, 181, __pyx_L1_error)
  __pyx_builtin_NotImplemented = __Pyx_GetBuiltinName(__pyx_n_s_NotImplemented); if (!__pyx_builtin_NotImplemented) __PYX_ERR(0, 342, __pyx_L1_error)
  __pyx_builtin_range = __Pyx_GetBuiltinName(__pyx_n_s_range); if (!__pyx_builtin_range) __PYX_ERR(0, 520, __pyx_L1_error)
  __pyx_builtin_KeyError = __Pyx_GetBuiltinName(__pyx_n_s_KeyError); if (!__pyx_builtin_KeyError) __PYX_ERR(1, 124, __pyx_L1_error)
  __pyx_builtin_RuntimeError = __Pyx_GetBuiltinName(__pyx_n_s_RuntimeError); if (!__pyx_builtin_RuntimeError) __PYX_ERR(3, 856, __pyx_L1_error)
  __pyx_builtin_ImportError = __Pyx_GetBuiltinName(__pyx_n_s_ImportError); if (!__pyx_builtin_ImportError) __PYX_ERR(3, 1038, __pyx_L1_error)
  __pyx_builtin_IndexError = __Pyx_GetBuiltinName(__pyx_n_s_IndexError); if (!__pyx_builtin_IndexError) __PYX_ERR(5, 89, __pyx_L1_error)
  __pyx_builtin_MemoryError = __Pyx_GetBuiltinName(__pyx_n_s_MemoryError); if (!__pyx_builtin_MemoryError) __PYX_ERR(2, 148, __pyx_L1_error)
  __pyx_builtin_enumerate = __Pyx_GetBuiltinName(__pyx_n_s_enumerate); if (!__pyx_builtin_enumerate) __PYX_ERR(2, 151, __pyx_L1_error)
  __pyx_builtin_Ellipsis = __Pyx_GetBuiltinName(__pyx_n_s_Ellipsis); if (!__pyx_builtin_Ellipsis) __PYX_ERR(2, 404, __pyx_L1_error)
  __pyx_builtin_id = __Pyx_GetBuiltinName(__pyx_n_s_id); if (!__pyx_builtin_id) __PYX_ERR(2, 613, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "pandas/_libs/interval.pyx":310
 *             raise ValueError(msg)
 *         if not left <= right:
 *             raise ValueError('left side of interval must be <= right side')             # <<<<<<<<<<<<<<
 *         if (isinstance(left, Timestamp) and
 *                 not tz_compare(left.tzinfo, right.tzinfo)):
 */
  __pyx_tuple_ = PyTuple_Pack(1, __pyx_kp_u_left_side_of_interval_must_be_ri); if (unlikely(!__pyx_tuple_)) __PYX_ERR(0, 310, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple_);
  __Pyx_GIVEREF(__pyx_tuple_);

  /* "pandas/_libs/interval.pyx":335
 *     def __contains__(self, key):
 *         if _interval_like(key):
 *             raise TypeError('__contains__ not defined for two intervals')             # <<<<<<<<<<<<<<
 *         return ((self.left < key if self.open_left else self.left <= key) and
 *                 (key < self.right if self.open_right else key <= self.right))
 */
  __pyx_tuple__2 = PyTuple_Pack(1, __pyx_kp_u_contains___not_defined_for_two); if (unlikely(!__pyx_tuple__2)) __PYX_ERR(0, 335, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2);
  __Pyx_GIVEREF(__pyx_tuple__2);

  /* "pandas/_libs/intervaltree.pxi":100
 *         # overlap if start of current interval < end of previous interval
 *         # (current and previous in terms of sorted order by left/start side)
 *         current = self.left[self.left_sorter[1:]]             # <<<<<<<<<<<<<<
 *         previous = self.right[self.left_sorter[:-1]]
 *         self._is_overlapping = bool(op(current, previous).any())
 */
  __pyx_slice__11 = PySlice_New(__pyx_int_1, Py_None, Py_None); if (unlikely(!__pyx_slice__11)) __PYX_ERR(1, 100, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_slice__11);
  __Pyx_GIVEREF(__pyx_slice__11);

  /* "pandas/_libs/intervaltree.pxi":101
 *         # (current and previous in terms of sorted order by left/start side)
 *         current = self.left[self.left_sorter[1:]]
 *         previous = self.right[self.left_sorter[:-1]]             # <<<<<<<<<<<<<<
 *         self._is_overlapping = bool(op(current, previous).any())
 * 
 */
  __pyx_slice__12 = PySlice_New(Py_None, __pyx_int_neg_1, Py_None); if (unlikely(!__pyx_slice__12)) __PYX_ERR(1, 101, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_slice__12);
  __Pyx_GIVEREF(__pyx_slice__12);

  /* "pandas/_libs/intervaltree.pxi":117
 *         return is_monotonic(sort_order, False)[0]
 * 
 *     def get_loc(self, scalar_t key):             # <<<<<<<<<<<<<<
 *         """Return all positions corresponding to intervals that overlap with
 *         the given scalar key
 */
  __pyx_tuple__15 = PyTuple_Pack(1, __pyx_kp_s_No_matching_signature_found); if (unlikely(!__pyx_tuple__15)) __PYX_ERR(1, 117, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__15);
  __Pyx_GIVEREF(__pyx_tuple__15);
  __pyx_tuple__16 = PyTuple_Pack(1, __pyx_kp_s_Function_call_with_ambiguous_arg); if (unlikely(!__pyx_tuple__16)) __PYX_ERR(1, 117, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__16);
  __Pyx_GIVEREF(__pyx_tuple__16);

  /* "pandas/_libs/intervaltree.pxi":172
 *                 result.append(-1)
 *             elif result.data.n > old_len + 1:
 *                 raise KeyError(             # <<<<<<<<<<<<<<
 *                     'indexer does not intersect a unique set of intervals')
 *             old_len = result.data.n
 */
  __pyx_tuple__17 = PyTuple_Pack(1, __pyx_kp_u_indexer_does_not_intersect_a_uni); if (unlikely(!__pyx_tuple__17)) __PYX_ERR(1, 172, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__17);
  __Pyx_GIVEREF(__pyx_tuple__17);

  /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":272
 *             if ((flags & pybuf.PyBUF_C_CONTIGUOUS == pybuf.PyBUF_C_CONTIGUOUS)
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_C_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not C contiguous")             # <<<<<<<<<<<<<<
 * 
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)
 */
  __pyx_tuple__19 = PyTuple_Pack(1, __pyx_kp_u_ndarray_is_not_C_contiguous); if (unlikely(!__pyx_tuple__19)) __PYX_ERR(3, 272, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__19);
  __Pyx_GIVEREF(__pyx_tuple__19);

  /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":276
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_F_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not Fortran contiguous")             # <<<<<<<<<<<<<<
 * 
 *             info.buf = PyArray_DATA(self)
 */
  __pyx_tuple__20 = PyTuple_Pack(1, __pyx_kp_u_ndarray_is_not_Fortran_contiguou); if (unlikely(!__pyx_tuple__20)) __PYX_ERR(3, 276, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__20);
  __Pyx_GIVEREF(__pyx_tuple__20);

  /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":306
 *                 if ((descr.byteorder == c'>' and little_endian) or
 *                     (descr.byteorder == c'<' and not little_endian)):
 *                     raise ValueError(u"Non-native byte order not supported")             # <<<<<<<<<<<<<<
 *                 if   t == NPY_BYTE:        f = "b"
 *                 elif t == NPY_UBYTE:       f = "B"
 */
  __pyx_tuple__21 = PyTuple_Pack(1, __pyx_kp_u_Non_native_byte_order_not_suppor); if (unlikely(!__pyx_tuple__21)) __PYX_ERR(3, 306, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__21);
  __Pyx_GIVEREF(__pyx_tuple__21);

  /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":856
 * 
 *         if (end - f) - <int>(new_offset - offset[0]) < 15:
 *             raise RuntimeError(u"Format string allocated too short, see comment in numpy.pxd")             # <<<<<<<<<<<<<<
 * 
 *         if ((child.byteorder == c'>' and little_endian) or
 */
  __pyx_tuple__22 = PyTuple_Pack(1, __pyx_kp_u_Format_string_allocated_too_shor); if (unlikely(!__pyx_tuple__22)) __PYX_ERR(3, 856, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__22);
  __Pyx_GIVEREF(__pyx_tuple__22);

  /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":880
 *             t = child.type_num
 *             if end - f < 5:
 *                 raise RuntimeError(u"Format string allocated too short.")             # <<<<<<<<<<<<<<
 * 
 *             # Until ticket #99 is fixed, use integers to avoid warnings
 */
  __pyx_tuple__23 = PyTuple_Pack(1, __pyx_kp_u_Format_string_allocated_too_shor_2); if (unlikely(!__pyx_tuple__23)) __PYX_ERR(3, 880, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__23);
  __Pyx_GIVEREF(__pyx_tuple__23);

  /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":1038
 *         _import_array()
 *     except Exception:
 *         raise ImportError("numpy.core.multiarray failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_umath() except -1:
 */
  __pyx_tuple__24 = PyTuple_Pack(1, __pyx_kp_u_numpy_core_multiarray_failed_to); if (unlikely(!__pyx_tuple__24)) __PYX_ERR(3, 1038, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__24);
  __Pyx_GIVEREF(__pyx_tuple__24);

  /* "../opt/conda/lib/python3.7/site-packages/Cython/Includes/numpy/__init__.pxd":1044
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_ufunc() except -1:
 */
  __pyx_tuple__25 = PyTuple_Pack(1, __pyx_kp_u_numpy_core_umath_failed_to_impor); if (unlikely(!__pyx_tuple__25)) __PYX_ERR(3, 1044, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__25);
  __Pyx_GIVEREF(__pyx_tuple__25);

  /* "pandas/_libs/util.pxd":89
 *         idx += size
 *     if idx >= size or size == 0 or idx < 0:
 *         raise IndexError('index out of bounds')             # <<<<<<<<<<<<<<
 * 
 *     return idx
 */
  __pyx_tuple__26 = PyTuple_Pack(1, __pyx_kp_u_index_out_of_bounds); if (unlikely(!__pyx_tuple__26)) __PYX_ERR(5, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__26);
  __Pyx_GIVEREF(__pyx_tuple__26);

  /* "View.MemoryView":133
 * 
 *         if not self.ndim:
 *             raise ValueError("Empty shape tuple for cython.array")             # <<<<<<<<<<<<<<
 * 
 *         if itemsize <= 0:
 */
  __pyx_tuple__27 = PyTuple_Pack(1, __pyx_kp_s_Empty_shape_tuple_for_cython_arr); if (unlikely(!__pyx_tuple__27)) __PYX_ERR(2, 133, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__27);
  __Pyx_GIVEREF(__pyx_tuple__27);

  /* "View.MemoryView":136
 * 
 *         if itemsize <= 0:
 *             raise ValueError("itemsize <= 0 for cython.array")             # <<<<<<<<<<<<<<
 * 
 *         if not isinstance(format, bytes):
 */
  __pyx_tuple__28 = PyTuple_Pack(1, __pyx_kp_s_itemsize_0_for_cython_array); if (unlikely(!__pyx_tuple__28)) __PYX_ERR(2, 136, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__28);
  __Pyx_GIVEREF(__pyx_tuple__28);

  /* "View.MemoryView":148
 * 
 *         if not self._shape:
 *             raise MemoryError("unable to allocate shape and strides.")             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_tuple__29 = PyTuple_Pack(1, __pyx_kp_s_unable_to_allocate_shape_and_str); if (unlikely(!__pyx_tuple__29)) __PYX_ERR(2, 148, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__29);
  __Pyx_GIVEREF(__pyx_tuple__29);

  /* "View.MemoryView":176
 *             self.data = <char *>malloc(self.len)
 *             if not self.data:
 *                 raise MemoryError("unable to allocate array data.")             # <<<<<<<<<<<<<<
 * 
 *             if self.dtype_is_object:
 */
  __pyx_tuple__30 = PyTuple_Pack(1, __pyx_kp_s_unable_to_allocate_array_data); if (unlikely(!__pyx_tuple__30)) __PYX_ERR(2, 176, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__30);
  __Pyx_GIVEREF(__pyx_tuple__30);

  /* "View.MemoryView":192
 *             bufmode = PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         if not (flags & bufmode):
 *             raise ValueError("Can only create a buffer that is contiguous in memory.")             # <<<<<<<<<<<<<<
 *         info.buf = self.data
 *         info.len = self.len
 */
  __pyx_tuple__31 = PyTuple_Pack(1, __pyx_kp_s_Can_only_create_a_buffer_that_is); if (unlikely(!__pyx_tuple__31)) __PYX_ERR(2, 192, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__31);
  __Pyx_GIVEREF(__pyx_tuple__31);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__32 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__32)) __PYX_ERR(2, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__32);
  __Pyx_GIVEREF(__pyx_tuple__32);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__33 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__33)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__33);
  __Pyx_GIVEREF(__pyx_tuple__33);

  /* "View.MemoryView":418
 *     def __setitem__(memoryview self, object index, object value):
 *         if self.view.readonly:
 *             raise TypeError("Cannot assign to read-only memoryview")             # <<<<<<<<<<<<<<
 * 
 *         have_slices, index = _unellipsify(index, self.view.ndim)
 */
  __pyx_tuple__34 = PyTuple_Pack(1, __pyx_kp_s_Cannot_assign_to_read_only_memor); if (unlikely(!__pyx_tuple__34)) __PYX_ERR(2, 418, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__34);
  __Pyx_GIVEREF(__pyx_tuple__34);

  /* "View.MemoryView":495
 *             result = struct.unpack(self.view.format, bytesitem)
 *         except struct.error:
 *             raise ValueError("Unable to convert item to object")             # <<<<<<<<<<<<<<
 *         else:
 *             if len(self.view.format) == 1:
 */
  __pyx_tuple__35 = PyTuple_Pack(1, __pyx_kp_s_Unable_to_convert_item_to_object); if (unlikely(!__pyx_tuple__35)) __PYX_ERR(2, 495, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__35);
  __Pyx_GIVEREF(__pyx_tuple__35);

  /* "View.MemoryView":520
 *     def __getbuffer__(self, Py_buffer *info, int flags):
 *         if flags & PyBUF_WRITABLE and self.view.readonly:
 *             raise ValueError("Cannot create writable memory view from read-only memoryview")             # <<<<<<<<<<<<<<
 * 
 *         if flags & PyBUF_ND:
 */
  __pyx_tuple__36 = PyTuple_Pack(1, __pyx_kp_s_Cannot_create_writable_memory_vi); if (unlikely(!__pyx_tuple__36)) __PYX_ERR(2, 520, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__36);
  __Pyx_GIVEREF(__pyx_tuple__36);

  /* "View.MemoryView":570
 *         if self.view.strides == NULL:
 * 
 *             raise ValueError("Buffer view does not expose strides")             # <<<<<<<<<<<<<<
 * 
 *         return tuple([stride for stride in self.view.strides[:self.view.ndim]])
 */
  __pyx_tuple__37 = PyTuple_Pack(1, __pyx_kp_s_Buffer_view_does_not_expose_stri); if (unlikely(!__pyx_tuple__37)) __PYX_ERR(2, 570, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__37);
  __Pyx_GIVEREF(__pyx_tuple__37);

  /* "View.MemoryView":577
 *     def suboffsets(self):
 *         if self.view.suboffsets == NULL:
 *             return (-1,) * self.view.ndim             # <<<<<<<<<<<<<<
 * 
 *         return tuple([suboffset for suboffset in self.view.suboffsets[:self.view.ndim]])
 */
  __pyx_tuple__38 = PyTuple_New(1); if (unlikely(!__pyx_tuple__38)) __PYX_ERR(2, 577, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__38);
  __Pyx_INCREF(__pyx_int_neg_1);
  __Pyx_GIVEREF(__pyx_int_neg_1);
  PyTuple_SET_ITEM(__pyx_tuple__38, 0, __pyx_int_neg_1);
  __Pyx_GIVEREF(__pyx_tuple__38);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__39 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__39)) __PYX_ERR(2, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__39);
  __Pyx_GIVEREF(__pyx_tuple__39);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__40 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__40)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__40);
  __Pyx_GIVEREF(__pyx_tuple__40);

  /* "View.MemoryView":682
 *         if item is Ellipsis:
 *             if not seen_ellipsis:
 *                 result.extend([slice(None)] * (ndim - len(tup) + 1))             # <<<<<<<<<<<<<<
 *                 seen_ellipsis = True
 *             else:
 */
  __pyx_slice__41 = PySlice_New(Py_None, Py_None, Py_None); if (unlikely(!__pyx_slice__41)) __PYX_ERR(2, 682, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_slice__41);
  __Pyx_GIVEREF(__pyx_slice__41);

  /* "View.MemoryView":703
 *     for suboffset in suboffsets[:ndim]:
 *         if suboffset >= 0:
 *             raise ValueError("Indirect dimensions not supported")             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_tuple__42 = PyTuple_Pack(1, __pyx_kp_s_Indirect_dimensions_not_supporte); if (unlikely(!__pyx_tuple__42)) __PYX_ERR(2, 703, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__42);
  __Pyx_GIVEREF(__pyx_tuple__42);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__43 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__43)) __PYX_ERR(2, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__43);
  __Pyx_GIVEREF(__pyx_tuple__43);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__44 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__44)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__44);
  __Pyx_GIVEREF(__pyx_tuple__44);

  /* "pandas/_libs/interval.pyx":29
 * 
 * 
 * _VALID_CLOSED = frozenset(['left', 'right', 'both', 'neither'])             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_tuple__45 = PyTuple_Pack(4, __pyx_n_u_left, __pyx_n_u_right, __pyx_n_u_both, __pyx_n_u_neither); if (unlikely(!__pyx_tuple__45)) __PYX_ERR(0, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__45);
  __Pyx_GIVEREF(__pyx_tuple__45);

  /* "pandas/_libs/interval.pyx":491
 * @cython.wraparound(False)
 * @cython.boundscheck(False)
 * def intervals_to_interval_bounds(ndarray intervals,             # <<<<<<<<<<<<<<
 *                                  bint validate_closed=True):
 *     """
 */
  __pyx_tuple__46 = PyTuple_Pack(10, __pyx_n_s_intervals, __pyx_n_s_validate_closed, __pyx_n_s_closed, __pyx_n_s_interval, __pyx_n_s_n, __pyx_n_s_left, __pyx_n_s_right, __pyx_n_s_seen_closed, __pyx_n_s_i, __pyx_n_s_msg); if (unlikely(!__pyx_tuple__46)) __PYX_ERR(0, 491, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__46);
  __Pyx_GIVEREF(__pyx_tuple__46);
  __pyx_codeobj__47 = (PyObject*)__Pyx_PyCode_New(2, 0, 10, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__46, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_pandas__libs_interval_pyx, __pyx_n_s_intervals_to_interval_bounds, 491, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__47)) __PYX_ERR(0, 491, __pyx_L1_error)

  /* "pandas/_libs/intervaltree.pxi":117
 *         return is_monotonic(sort_order, False)[0]
 * 
 *     def get_loc(self, scalar_t key):             # <<<<<<<<<<<<<<
 *         """Return all positions corresponding to intervals that overlap with
 *         the given scalar key
 */
  __pyx_tuple__48 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_key, __pyx_n_s_key, __pyx_n_s_result); if (unlikely(!__pyx_tuple__48)) __PYX_ERR(1, 117, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__48);
  __Pyx_GIVEREF(__pyx_tuple__48);
  __pyx_codeobj__49 = (PyObject*)__Pyx_PyCode_New(2, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__48, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_pandas__libs_intervaltree_pxi, __pyx_n_s_get_loc, 117, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__49)) __PYX_ERR(1, 117, __pyx_L1_error)

  /* "pandas/_libs/intervaltree.pxi":154
 *         return uniques.astype('intp')
 * 
 *     def get_indexer(self, scalar_t[:] target):             # <<<<<<<<<<<<<<
 *         """Return the positions corresponding to unique intervals that overlap
 *         with the given array of scalar targets.
 */
  __pyx_tuple__50 = PyTuple_Pack(6, __pyx_n_s_self, __pyx_n_s_target, __pyx_n_s_target, __pyx_n_s_old_len, __pyx_n_s_i, __pyx_n_s_result); if (unlikely(!__pyx_tuple__50)) __PYX_ERR(1, 154, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__50);
  __Pyx_GIVEREF(__pyx_tuple__50);
  __pyx_codeobj__51 = (PyObject*)__Pyx_PyCode_New(2, 0, 6, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__50, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_pandas__libs_intervaltree_pxi, __pyx_n_s_get_indexer, 154, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__51)) __PYX_ERR(1, 154, __pyx_L1_error)

  /* "pandas/_libs/intervaltree.pxi":177
 *         return result.to_array().astype('intp')
 * 
 *     def get_indexer_non_unique(self, scalar_t[:] target):             # <<<<<<<<<<<<<<
 *         """Return the positions corresponding to intervals that overlap with
 *         the given array of scalar targets. Non-unique positions are repeated.
 */
  __pyx_tuple__52 = PyTuple_Pack(7, __pyx_n_s_self, __pyx_n_s_target, __pyx_n_s_target, __pyx_n_s_old_len, __pyx_n_s_i, __pyx_n_s_result, __pyx_n_s_missing); if (unlikely(!__pyx_tuple__52)) __PYX_ERR(1, 177, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__52);
  __Pyx_GIVEREF(__pyx_tuple__52);
  __pyx_codeobj__53 = (PyObject*)__Pyx_PyCode_New(2, 0, 7, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__52, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_pandas__libs_intervaltree_pxi, __pyx_n_s_get_indexer_non_unique, 177, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__53)) __PYX_ERR(1, 177, __pyx_L1_error)

  /* "pandas/_libs/intervaltree.pxi":334
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */
  __pyx_tuple__54 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_result, __pyx_n_s_point); if (unlikely(!__pyx_tuple__54)) __PYX_ERR(1, 334, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__54);
  __Pyx_GIVEREF(__pyx_tuple__54);
  __pyx_codeobj__55 = (PyObject*)__Pyx_PyCode_New(3, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__54, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_pandas__libs_intervaltree_pxi, __pyx_n_s_pyx_fuse_0query, 334, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__55)) __PYX_ERR(1, 334, __pyx_L1_error)

  /* "pandas/_libs/intervaltree.pxi":400
 *             return (m, (l, r))
 * 
 * NODE_CLASSES['float32',             # <<<<<<<<<<<<<<
 *              'left'] = Float32ClosedLeftIntervalNode
 * 
 */
  __pyx_tuple__56 = PyTuple_Pack(2, __pyx_n_u_float32, __pyx_n_u_left); if (unlikely(!__pyx_tuple__56)) __PYX_ERR(1, 400, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__56);
  __Pyx_GIVEREF(__pyx_tuple__56);

  /* "pandas/_libs/intervaltree.pxi":504
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */
  __pyx_tuple__57 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_result, __pyx_n_s_point); if (unlikely(!__pyx_tuple__57)) __PYX_ERR(1, 504, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__57);
  __Pyx_GIVEREF(__pyx_tuple__57);
  __pyx_codeobj__58 = (PyObject*)__Pyx_PyCode_New(3, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__57, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_pandas__libs_intervaltree_pxi, __pyx_n_s_pyx_fuse_0query, 504, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__58)) __PYX_ERR(1, 504, __pyx_L1_error)

  /* "pandas/_libs/intervaltree.pxi":570
 *             return (m, (l, r))
 * 
 * NODE_CLASSES['float32',             # <<<<<<<<<<<<<<
 *              'right'] = Float32ClosedRightIntervalNode
 * 
 */
  __pyx_tuple__59 = PyTuple_Pack(2, __pyx_n_u_float32, __pyx_n_u_right); if (unlikely(!__pyx_tuple__59)) __PYX_ERR(1, 570, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__59);
  __Pyx_GIVEREF(__pyx_tuple__59);

  /* "pandas/_libs/intervaltree.pxi":674
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */
  __pyx_tuple__60 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_result, __pyx_n_s_point); if (unlikely(!__pyx_tuple__60)) __PYX_ERR(1, 674, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__60);
  __Pyx_GIVEREF(__pyx_tuple__60);
  __pyx_codeobj__61 = (PyObject*)__Pyx_PyCode_New(3, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__60, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_pandas__libs_intervaltree_pxi, __pyx_n_s_pyx_fuse_0query, 674, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__61)) __PYX_ERR(1, 674, __pyx_L1_error)

  /* "pandas/_libs/intervaltree.pxi":740
 *             return (m, (l, r))
 * 
 * NODE_CLASSES['float32',             # <<<<<<<<<<<<<<
 *              'both'] = Float32ClosedBothIntervalNode
 * 
 */
  __pyx_tuple__62 = PyTuple_Pack(2, __pyx_n_u_float32, __pyx_n_u_both); if (unlikely(!__pyx_tuple__62)) __PYX_ERR(1, 740, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__62);
  __Pyx_GIVEREF(__pyx_tuple__62);

  /* "pandas/_libs/intervaltree.pxi":844
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */
  __pyx_tuple__63 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_result, __pyx_n_s_point); if (unlikely(!__pyx_tuple__63)) __PYX_ERR(1, 844, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__63);
  __Pyx_GIVEREF(__pyx_tuple__63);
  __pyx_codeobj__64 = (PyObject*)__Pyx_PyCode_New(3, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__63, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_pandas__libs_intervaltree_pxi, __pyx_n_s_pyx_fuse_0query, 844, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__64)) __PYX_ERR(1, 844, __pyx_L1_error)

  /* "pandas/_libs/intervaltree.pxi":910
 *             return (m, (l, r))
 * 
 * NODE_CLASSES['float32',             # <<<<<<<<<<<<<<
 *              'neither'] = Float32ClosedNeitherIntervalNode
 * 
 */
  __pyx_tuple__65 = PyTuple_Pack(2, __pyx_n_u_float32, __pyx_n_u_neither); if (unlikely(!__pyx_tuple__65)) __PYX_ERR(1, 910, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__65);
  __Pyx_GIVEREF(__pyx_tuple__65);

  /* "pandas/_libs/intervaltree.pxi":1014
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */
  __pyx_tuple__66 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_result, __pyx_n_s_point); if (unlikely(!__pyx_tuple__66)) __PYX_ERR(1, 1014, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__66);
  __Pyx_GIVEREF(__pyx_tuple__66);
  __pyx_codeobj__67 = (PyObject*)__Pyx_PyCode_New(3, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__66, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_pandas__libs_intervaltree_pxi, __pyx_n_s_pyx_fuse_0query, 1014, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__67)) __PYX_ERR(1, 1014, __pyx_L1_error)

  /* "pandas/_libs/intervaltree.pxi":1080
 *             return (m, (l, r))
 * 
 * NODE_CLASSES['float64',             # <<<<<<<<<<<<<<
 *              'left'] = Float64ClosedLeftIntervalNode
 * 
 */
  __pyx_tuple__68 = PyTuple_Pack(2, __pyx_n_u_float64, __pyx_n_u_left); if (unlikely(!__pyx_tuple__68)) __PYX_ERR(1, 1080, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__68);
  __Pyx_GIVEREF(__pyx_tuple__68);

  /* "pandas/_libs/intervaltree.pxi":1184
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */
  __pyx_tuple__69 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_result, __pyx_n_s_point); if (unlikely(!__pyx_tuple__69)) __PYX_ERR(1, 1184, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__69);
  __Pyx_GIVEREF(__pyx_tuple__69);
  __pyx_codeobj__70 = (PyObject*)__Pyx_PyCode_New(3, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__69, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_pandas__libs_intervaltree_pxi, __pyx_n_s_pyx_fuse_0query, 1184, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__70)) __PYX_ERR(1, 1184, __pyx_L1_error)

  /* "pandas/_libs/intervaltree.pxi":1250
 *             return (m, (l, r))
 * 
 * NODE_CLASSES['float64',             # <<<<<<<<<<<<<<
 *              'right'] = Float64ClosedRightIntervalNode
 * 
 */
  __pyx_tuple__71 = PyTuple_Pack(2, __pyx_n_u_float64, __pyx_n_u_right); if (unlikely(!__pyx_tuple__71)) __PYX_ERR(1, 1250, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__71);
  __Pyx_GIVEREF(__pyx_tuple__71);

  /* "pandas/_libs/intervaltree.pxi":1354
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */
  __pyx_tuple__72 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_result, __pyx_n_s_point); if (unlikely(!__pyx_tuple__72)) __PYX_ERR(1, 1354, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__72);
  __Pyx_GIVEREF(__pyx_tuple__72);
  __pyx_codeobj__73 = (PyObject*)__Pyx_PyCode_New(3, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__72, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_pandas__libs_intervaltree_pxi, __pyx_n_s_pyx_fuse_0query, 1354, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__73)) __PYX_ERR(1, 1354, __pyx_L1_error)

  /* "pandas/_libs/intervaltree.pxi":1420
 *             return (m, (l, r))
 * 
 * NODE_CLASSES['float64',             # <<<<<<<<<<<<<<
 *              'both'] = Float64ClosedBothIntervalNode
 * 
 */
  __pyx_tuple__74 = PyTuple_Pack(2, __pyx_n_u_float64, __pyx_n_u_both); if (unlikely(!__pyx_tuple__74)) __PYX_ERR(1, 1420, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__74);
  __Pyx_GIVEREF(__pyx_tuple__74);

  /* "pandas/_libs/intervaltree.pxi":1524
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */
  __pyx_tuple__75 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_result, __pyx_n_s_point); if (unlikely(!__pyx_tuple__75)) __PYX_ERR(1, 1524, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__75);
  __Pyx_GIVEREF(__pyx_tuple__75);
  __pyx_codeobj__76 = (PyObject*)__Pyx_PyCode_New(3, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__75, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_pandas__libs_intervaltree_pxi, __pyx_n_s_pyx_fuse_0query, 1524, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__76)) __PYX_ERR(1, 1524, __pyx_L1_error)

  /* "pandas/_libs/intervaltree.pxi":1590
 *             return (m, (l, r))
 * 
 * NODE_CLASSES['float64',             # <<<<<<<<<<<<<<
 *              'neither'] = Float64ClosedNeitherIntervalNode
 * 
 */
  __pyx_tuple__77 = PyTuple_Pack(2, __pyx_n_u_float64, __pyx_n_u_neither); if (unlikely(!__pyx_tuple__77)) __PYX_ERR(1, 1590, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__77);
  __Pyx_GIVEREF(__pyx_tuple__77);

  /* "pandas/_libs/intervaltree.pxi":1694
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */
  __pyx_tuple__78 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_result, __pyx_n_s_point); if (unlikely(!__pyx_tuple__78)) __PYX_ERR(1, 1694, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__78);
  __Pyx_GIVEREF(__pyx_tuple__78);
  __pyx_codeobj__79 = (PyObject*)__Pyx_PyCode_New(3, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__78, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_pandas__libs_intervaltree_pxi, __pyx_n_s_pyx_fuse_0query, 1694, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__79)) __PYX_ERR(1, 1694, __pyx_L1_error)

  /* "pandas/_libs/intervaltree.pxi":1760
 *             return (m, (l, r))
 * 
 * NODE_CLASSES['int32',             # <<<<<<<<<<<<<<
 *              'left'] = Int32ClosedLeftIntervalNode
 * 
 */
  __pyx_tuple__80 = PyTuple_Pack(2, __pyx_n_u_int32, __pyx_n_u_left); if (unlikely(!__pyx_tuple__80)) __PYX_ERR(1, 1760, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__80);
  __Pyx_GIVEREF(__pyx_tuple__80);

  /* "pandas/_libs/intervaltree.pxi":1864
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */
  __pyx_tuple__81 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_result, __pyx_n_s_point); if (unlikely(!__pyx_tuple__81)) __PYX_ERR(1, 1864, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__81);
  __Pyx_GIVEREF(__pyx_tuple__81);
  __pyx_codeobj__82 = (PyObject*)__Pyx_PyCode_New(3, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__81, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_pandas__libs_intervaltree_pxi, __pyx_n_s_pyx_fuse_0query, 1864, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__82)) __PYX_ERR(1, 1864, __pyx_L1_error)

  /* "pandas/_libs/intervaltree.pxi":1930
 *             return (m, (l, r))
 * 
 * NODE_CLASSES['int32',             # <<<<<<<<<<<<<<
 *              'right'] = Int32ClosedRightIntervalNode
 * 
 */
  __pyx_tuple__83 = PyTuple_Pack(2, __pyx_n_u_int32, __pyx_n_u_right); if (unlikely(!__pyx_tuple__83)) __PYX_ERR(1, 1930, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__83);
  __Pyx_GIVEREF(__pyx_tuple__83);

  /* "pandas/_libs/intervaltree.pxi":2034
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */
  __pyx_tuple__84 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_result, __pyx_n_s_point); if (unlikely(!__pyx_tuple__84)) __PYX_ERR(1, 2034, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__84);
  __Pyx_GIVEREF(__pyx_tuple__84);
  __pyx_codeobj__85 = (PyObject*)__Pyx_PyCode_New(3, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__84, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_pandas__libs_intervaltree_pxi, __pyx_n_s_pyx_fuse_0query, 2034, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__85)) __PYX_ERR(1, 2034, __pyx_L1_error)

  /* "pandas/_libs/intervaltree.pxi":2100
 *             return (m, (l, r))
 * 
 * NODE_CLASSES['int32',             # <<<<<<<<<<<<<<
 *              'both'] = Int32ClosedBothIntervalNode
 * 
 */
  __pyx_tuple__86 = PyTuple_Pack(2, __pyx_n_u_int32, __pyx_n_u_both); if (unlikely(!__pyx_tuple__86)) __PYX_ERR(1, 2100, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__86);
  __Pyx_GIVEREF(__pyx_tuple__86);

  /* "pandas/_libs/intervaltree.pxi":2204
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */
  __pyx_tuple__87 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_result, __pyx_n_s_point); if (unlikely(!__pyx_tuple__87)) __PYX_ERR(1, 2204, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__87);
  __Pyx_GIVEREF(__pyx_tuple__87);
  __pyx_codeobj__88 = (PyObject*)__Pyx_PyCode_New(3, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__87, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_pandas__libs_intervaltree_pxi, __pyx_n_s_pyx_fuse_0query, 2204, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__88)) __PYX_ERR(1, 2204, __pyx_L1_error)

  /* "pandas/_libs/intervaltree.pxi":2270
 *             return (m, (l, r))
 * 
 * NODE_CLASSES['int32',             # <<<<<<<<<<<<<<
 *              'neither'] = Int32ClosedNeitherIntervalNode
 * 
 */
  __pyx_tuple__89 = PyTuple_Pack(2, __pyx_n_u_int32, __pyx_n_u_neither); if (unlikely(!__pyx_tuple__89)) __PYX_ERR(1, 2270, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__89);
  __Pyx_GIVEREF(__pyx_tuple__89);

  /* "pandas/_libs/intervaltree.pxi":2374
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */
  __pyx_tuple__90 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_result, __pyx_n_s_point); if (unlikely(!__pyx_tuple__90)) __PYX_ERR(1, 2374, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__90);
  __Pyx_GIVEREF(__pyx_tuple__90);
  __pyx_codeobj__91 = (PyObject*)__Pyx_PyCode_New(3, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__90, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_pandas__libs_intervaltree_pxi, __pyx_n_s_pyx_fuse_0query, 2374, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__91)) __PYX_ERR(1, 2374, __pyx_L1_error)

  /* "pandas/_libs/intervaltree.pxi":2440
 *             return (m, (l, r))
 * 
 * NODE_CLASSES['int64',             # <<<<<<<<<<<<<<
 *              'left'] = Int64ClosedLeftIntervalNode
 * 
 */
  __pyx_tuple__92 = PyTuple_Pack(2, __pyx_n_u_int64, __pyx_n_u_left); if (unlikely(!__pyx_tuple__92)) __PYX_ERR(1, 2440, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__92);
  __Pyx_GIVEREF(__pyx_tuple__92);

  /* "pandas/_libs/intervaltree.pxi":2544
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */
  __pyx_tuple__93 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_result, __pyx_n_s_point); if (unlikely(!__pyx_tuple__93)) __PYX_ERR(1, 2544, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__93);
  __Pyx_GIVEREF(__pyx_tuple__93);
  __pyx_codeobj__94 = (PyObject*)__Pyx_PyCode_New(3, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__93, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_pandas__libs_intervaltree_pxi, __pyx_n_s_pyx_fuse_0query, 2544, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__94)) __PYX_ERR(1, 2544, __pyx_L1_error)

  /* "pandas/_libs/intervaltree.pxi":2610
 *             return (m, (l, r))
 * 
 * NODE_CLASSES['int64',             # <<<<<<<<<<<<<<
 *              'right'] = Int64ClosedRightIntervalNode
 * 
 */
  __pyx_tuple__95 = PyTuple_Pack(2, __pyx_n_u_int64, __pyx_n_u_right); if (unlikely(!__pyx_tuple__95)) __PYX_ERR(1, 2610, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__95);
  __Pyx_GIVEREF(__pyx_tuple__95);

  /* "pandas/_libs/intervaltree.pxi":2714
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */
  __pyx_tuple__96 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_result, __pyx_n_s_point); if (unlikely(!__pyx_tuple__96)) __PYX_ERR(1, 2714, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__96);
  __Pyx_GIVEREF(__pyx_tuple__96);
  __pyx_codeobj__97 = (PyObject*)__Pyx_PyCode_New(3, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__96, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_pandas__libs_intervaltree_pxi, __pyx_n_s_pyx_fuse_0query, 2714, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__97)) __PYX_ERR(1, 2714, __pyx_L1_error)

  /* "pandas/_libs/intervaltree.pxi":2780
 *             return (m, (l, r))
 * 
 * NODE_CLASSES['int64',             # <<<<<<<<<<<<<<
 *              'both'] = Int64ClosedBothIntervalNode
 * 
 */
  __pyx_tuple__98 = PyTuple_Pack(2, __pyx_n_u_int64, __pyx_n_u_both); if (unlikely(!__pyx_tuple__98)) __PYX_ERR(1, 2780, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__98);
  __Pyx_GIVEREF(__pyx_tuple__98);

  /* "pandas/_libs/intervaltree.pxi":2884
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */
  __pyx_tuple__99 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_result, __pyx_n_s_point); if (unlikely(!__pyx_tuple__99)) __PYX_ERR(1, 2884, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__99);
  __Pyx_GIVEREF(__pyx_tuple__99);
  __pyx_codeobj__100 = (PyObject*)__Pyx_PyCode_New(3, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__99, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_pandas__libs_intervaltree_pxi, __pyx_n_s_pyx_fuse_0query, 2884, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__100)) __PYX_ERR(1, 2884, __pyx_L1_error)

  /* "pandas/_libs/intervaltree.pxi":2950
 *             return (m, (l, r))
 * 
 * NODE_CLASSES['int64',             # <<<<<<<<<<<<<<
 *              'neither'] = Int64ClosedNeitherIntervalNode
 * 
 */
  __pyx_tuple__101 = PyTuple_Pack(2, __pyx_n_u_int64, __pyx_n_u_neither); if (unlikely(!__pyx_tuple__101)) __PYX_ERR(1, 2950, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__101);
  __Pyx_GIVEREF(__pyx_tuple__101);

  /* "pandas/_libs/intervaltree.pxi":3054
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */
  __pyx_tuple__102 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_result, __pyx_n_s_point); if (unlikely(!__pyx_tuple__102)) __PYX_ERR(1, 3054, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__102);
  __Pyx_GIVEREF(__pyx_tuple__102);
  __pyx_codeobj__103 = (PyObject*)__Pyx_PyCode_New(3, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__102, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_pandas__libs_intervaltree_pxi, __pyx_n_s_pyx_fuse_0query, 3054, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__103)) __PYX_ERR(1, 3054, __pyx_L1_error)

  /* "pandas/_libs/intervaltree.pxi":3120
 *             return (m, (l, r))
 * 
 * NODE_CLASSES['uint64',             # <<<<<<<<<<<<<<
 *              'left'] = Uint64ClosedLeftIntervalNode
 * 
 */
  __pyx_tuple__104 = PyTuple_Pack(2, __pyx_n_u_uint64, __pyx_n_u_left); if (unlikely(!__pyx_tuple__104)) __PYX_ERR(1, 3120, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__104);
  __Pyx_GIVEREF(__pyx_tuple__104);

  /* "pandas/_libs/intervaltree.pxi":3224
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */
  __pyx_tuple__105 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_result, __pyx_n_s_point); if (unlikely(!__pyx_tuple__105)) __PYX_ERR(1, 3224, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__105);
  __Pyx_GIVEREF(__pyx_tuple__105);
  __pyx_codeobj__106 = (PyObject*)__Pyx_PyCode_New(3, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__105, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_pandas__libs_intervaltree_pxi, __pyx_n_s_pyx_fuse_0query, 3224, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__106)) __PYX_ERR(1, 3224, __pyx_L1_error)

  /* "pandas/_libs/intervaltree.pxi":3290
 *             return (m, (l, r))
 * 
 * NODE_CLASSES['uint64',             # <<<<<<<<<<<<<<
 *              'right'] = Uint64ClosedRightIntervalNode
 * 
 */
  __pyx_tuple__107 = PyTuple_Pack(2, __pyx_n_u_uint64, __pyx_n_u_right); if (unlikely(!__pyx_tuple__107)) __PYX_ERR(1, 3290, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__107);
  __Pyx_GIVEREF(__pyx_tuple__107);

  /* "pandas/_libs/intervaltree.pxi":3394
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */
  __pyx_tuple__108 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_result, __pyx_n_s_point); if (unlikely(!__pyx_tuple__108)) __PYX_ERR(1, 3394, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__108);
  __Pyx_GIVEREF(__pyx_tuple__108);
  __pyx_codeobj__109 = (PyObject*)__Pyx_PyCode_New(3, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__108, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_pandas__libs_intervaltree_pxi, __pyx_n_s_pyx_fuse_0query, 3394, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__109)) __PYX_ERR(1, 3394, __pyx_L1_error)

  /* "pandas/_libs/intervaltree.pxi":3460
 *             return (m, (l, r))
 * 
 * NODE_CLASSES['uint64',             # <<<<<<<<<<<<<<
 *              'both'] = Uint64ClosedBothIntervalNode
 * 
 */
  __pyx_tuple__110 = PyTuple_Pack(2, __pyx_n_u_uint64, __pyx_n_u_both); if (unlikely(!__pyx_tuple__110)) __PYX_ERR(1, 3460, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__110);
  __Pyx_GIVEREF(__pyx_tuple__110);

  /* "pandas/_libs/intervaltree.pxi":3564
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */
  __pyx_tuple__111 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_result, __pyx_n_s_point); if (unlikely(!__pyx_tuple__111)) __PYX_ERR(1, 3564, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__111);
  __Pyx_GIVEREF(__pyx_tuple__111);
  __pyx_codeobj__112 = (PyObject*)__Pyx_PyCode_New(3, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__111, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_pandas__libs_intervaltree_pxi, __pyx_n_s_pyx_fuse_0query, 3564, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__112)) __PYX_ERR(1, 3564, __pyx_L1_error)

  /* "pandas/_libs/intervaltree.pxi":3630
 *             return (m, (l, r))
 * 
 * NODE_CLASSES['uint64',             # <<<<<<<<<<<<<<
 *              'neither'] = Uint64ClosedNeitherIntervalNode
 */
  __pyx_tuple__113 = PyTuple_Pack(2, __pyx_n_u_uint64, __pyx_n_u_neither); if (unlikely(!__pyx_tuple__113)) __PYX_ERR(1, 3630, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__113);
  __Pyx_GIVEREF(__pyx_tuple__113);

  /* "(tree fragment)":1
 * def __pyx_unpickle_IntervalMixin(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_tuple__114 = PyTuple_Pack(5, __pyx_n_s_pyx_type, __pyx_n_s_pyx_checksum, __pyx_n_s_pyx_state, __pyx_n_s_pyx_PickleError, __pyx_n_s_pyx_result); if (unlikely(!__pyx_tuple__114)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__114);
  __Pyx_GIVEREF(__pyx_tuple__114);
  __pyx_codeobj__115 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__114, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_IntervalMixin, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__115)) __PYX_ERR(2, 1, __pyx_L1_error)
  __pyx_tuple__116 = PyTuple_Pack(5, __pyx_n_s_pyx_type, __pyx_n_s_pyx_checksum, __pyx_n_s_pyx_state, __pyx_n_s_pyx_PickleError, __pyx_n_s_pyx_result); if (unlikely(!__pyx_tuple__116)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__116);
  __Pyx_GIVEREF(__pyx_tuple__116);
  __pyx_codeobj__117 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__116, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_IntervalTree, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__117)) __PYX_ERR(2, 1, __pyx_L1_error)
  __pyx_tuple__118 = PyTuple_Pack(5, __pyx_n_s_pyx_type, __pyx_n_s_pyx_checksum, __pyx_n_s_pyx_state, __pyx_n_s_pyx_PickleError, __pyx_n_s_pyx_result); if (unlikely(!__pyx_tuple__118)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__118);
  __Pyx_GIVEREF(__pyx_tuple__118);
  __pyx_codeobj__119 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__118, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_Float32ClosedLeft, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__119)) __PYX_ERR(2, 1, __pyx_L1_error)
  __pyx_tuple__120 = PyTuple_Pack(5, __pyx_n_s_pyx_type, __pyx_n_s_pyx_checksum, __pyx_n_s_pyx_state, __pyx_n_s_pyx_PickleError, __pyx_n_s_pyx_result); if (unlikely(!__pyx_tuple__120)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__120);
  __Pyx_GIVEREF(__pyx_tuple__120);
  __pyx_codeobj__121 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__120, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_Float32ClosedRigh, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__121)) __PYX_ERR(2, 1, __pyx_L1_error)
  __pyx_tuple__122 = PyTuple_Pack(5, __pyx_n_s_pyx_type, __pyx_n_s_pyx_checksum, __pyx_n_s_pyx_state, __pyx_n_s_pyx_PickleError, __pyx_n_s_pyx_result); if (unlikely(!__pyx_tuple__122)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__122);
  __Pyx_GIVEREF(__pyx_tuple__122);
  __pyx_codeobj__123 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__122, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_Float32ClosedBoth, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__123)) __PYX_ERR(2, 1, __pyx_L1_error)
  __pyx_tuple__124 = PyTuple_Pack(5, __pyx_n_s_pyx_type, __pyx_n_s_pyx_checksum, __pyx_n_s_pyx_state, __pyx_n_s_pyx_PickleError, __pyx_n_s_pyx_result); if (unlikely(!__pyx_tuple__124)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__124);
  __Pyx_GIVEREF(__pyx_tuple__124);
  __pyx_codeobj__125 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__124, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_Float32ClosedNeit, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__125)) __PYX_ERR(2, 1, __pyx_L1_error)
  __pyx_tuple__126 = PyTuple_Pack(5, __pyx_n_s_pyx_type, __pyx_n_s_pyx_checksum, __pyx_n_s_pyx_state, __pyx_n_s_pyx_PickleError, __pyx_n_s_pyx_result); if (unlikely(!__pyx_tuple__126)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__126);
  __Pyx_GIVEREF(__pyx_tuple__126);
  __pyx_codeobj__127 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__126, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_Float64ClosedLeft, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__127)) __PYX_ERR(2, 1, __pyx_L1_error)
  __pyx_tuple__128 = PyTuple_Pack(5, __pyx_n_s_pyx_type, __pyx_n_s_pyx_checksum, __pyx_n_s_pyx_state, __pyx_n_s_pyx_PickleError, __pyx_n_s_pyx_result); if (unlikely(!__pyx_tuple__128)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__128);
  __Pyx_GIVEREF(__pyx_tuple__128);
  __pyx_codeobj__129 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__128, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_Float64ClosedRigh, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__129)) __PYX_ERR(2, 1, __pyx_L1_error)
  __pyx_tuple__130 = PyTuple_Pack(5, __pyx_n_s_pyx_type, __pyx_n_s_pyx_checksum, __pyx_n_s_pyx_state, __pyx_n_s_pyx_PickleError, __pyx_n_s_pyx_result); if (unlikely(!__pyx_tuple__130)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__130);
  __Pyx_GIVEREF(__pyx_tuple__130);
  __pyx_codeobj__131 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__130, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_Float64ClosedBoth, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__131)) __PYX_ERR(2, 1, __pyx_L1_error)
  __pyx_tuple__132 = PyTuple_Pack(5, __pyx_n_s_pyx_type, __pyx_n_s_pyx_checksum, __pyx_n_s_pyx_state, __pyx_n_s_pyx_PickleError, __pyx_n_s_pyx_result); if (unlikely(!__pyx_tuple__132)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__132);
  __Pyx_GIVEREF(__pyx_tuple__132);
  __pyx_codeobj__133 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__132, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_Float64ClosedNeit, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__133)) __PYX_ERR(2, 1, __pyx_L1_error)
  __pyx_tuple__134 = PyTuple_Pack(5, __pyx_n_s_pyx_type, __pyx_n_s_pyx_checksum, __pyx_n_s_pyx_state, __pyx_n_s_pyx_PickleError, __pyx_n_s_pyx_result); if (unlikely(!__pyx_tuple__134)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__134);
  __Pyx_GIVEREF(__pyx_tuple__134);
  __pyx_codeobj__135 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__134, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_Int32ClosedLeftIn, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__135)) __PYX_ERR(2, 1, __pyx_L1_error)
  __pyx_tuple__136 = PyTuple_Pack(5, __pyx_n_s_pyx_type, __pyx_n_s_pyx_checksum, __pyx_n_s_pyx_state, __pyx_n_s_pyx_PickleError, __pyx_n_s_pyx_result); if (unlikely(!__pyx_tuple__136)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__136);
  __Pyx_GIVEREF(__pyx_tuple__136);
  __pyx_codeobj__137 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__136, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_Int32ClosedRightI, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__137)) __PYX_ERR(2, 1, __pyx_L1_error)
  __pyx_tuple__138 = PyTuple_Pack(5, __pyx_n_s_pyx_type, __pyx_n_s_pyx_checksum, __pyx_n_s_pyx_state, __pyx_n_s_pyx_PickleError, __pyx_n_s_pyx_result); if (unlikely(!__pyx_tuple__138)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__138);
  __Pyx_GIVEREF(__pyx_tuple__138);
  __pyx_codeobj__139 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__138, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_Int32ClosedBothIn, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__139)) __PYX_ERR(2, 1, __pyx_L1_error)
  __pyx_tuple__140 = PyTuple_Pack(5, __pyx_n_s_pyx_type, __pyx_n_s_pyx_checksum, __pyx_n_s_pyx_state, __pyx_n_s_pyx_PickleError, __pyx_n_s_pyx_result); if (unlikely(!__pyx_tuple__140)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__140);
  __Pyx_GIVEREF(__pyx_tuple__140);
  __pyx_codeobj__141 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__140, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_Int32ClosedNeithe, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__141)) __PYX_ERR(2, 1, __pyx_L1_error)
  __pyx_tuple__142 = PyTuple_Pack(5, __pyx_n_s_pyx_type, __pyx_n_s_pyx_checksum, __pyx_n_s_pyx_state, __pyx_n_s_pyx_PickleError, __pyx_n_s_pyx_result); if (unlikely(!__pyx_tuple__142)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__142);
  __Pyx_GIVEREF(__pyx_tuple__142);
  __pyx_codeobj__143 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__142, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_Int64ClosedLeftIn, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__143)) __PYX_ERR(2, 1, __pyx_L1_error)
  __pyx_tuple__144 = PyTuple_Pack(5, __pyx_n_s_pyx_type, __pyx_n_s_pyx_checksum, __pyx_n_s_pyx_state, __pyx_n_s_pyx_PickleError, __pyx_n_s_pyx_result); if (unlikely(!__pyx_tuple__144)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__144);
  __Pyx_GIVEREF(__pyx_tuple__144);
  __pyx_codeobj__145 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__144, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_Int64ClosedRightI, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__145)) __PYX_ERR(2, 1, __pyx_L1_error)
  __pyx_tuple__146 = PyTuple_Pack(5, __pyx_n_s_pyx_type, __pyx_n_s_pyx_checksum, __pyx_n_s_pyx_state, __pyx_n_s_pyx_PickleError, __pyx_n_s_pyx_result); if (unlikely(!__pyx_tuple__146)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__146);
  __Pyx_GIVEREF(__pyx_tuple__146);
  __pyx_codeobj__147 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__146, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_Int64ClosedBothIn, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__147)) __PYX_ERR(2, 1, __pyx_L1_error)
  __pyx_tuple__148 = PyTuple_Pack(5, __pyx_n_s_pyx_type, __pyx_n_s_pyx_checksum, __pyx_n_s_pyx_state, __pyx_n_s_pyx_PickleError, __pyx_n_s_pyx_result); if (unlikely(!__pyx_tuple__148)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__148);
  __Pyx_GIVEREF(__pyx_tuple__148);
  __pyx_codeobj__149 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__148, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_Int64ClosedNeithe, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__149)) __PYX_ERR(2, 1, __pyx_L1_error)
  __pyx_tuple__150 = PyTuple_Pack(5, __pyx_n_s_pyx_type, __pyx_n_s_pyx_checksum, __pyx_n_s_pyx_state, __pyx_n_s_pyx_PickleError, __pyx_n_s_pyx_result); if (unlikely(!__pyx_tuple__150)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__150);
  __Pyx_GIVEREF(__pyx_tuple__150);
  __pyx_codeobj__151 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__150, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_Uint64ClosedLeftI, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__151)) __PYX_ERR(2, 1, __pyx_L1_error)
  __pyx_tuple__152 = PyTuple_Pack(5, __pyx_n_s_pyx_type, __pyx_n_s_pyx_checksum, __pyx_n_s_pyx_state, __pyx_n_s_pyx_PickleError, __pyx_n_s_pyx_result); if (unlikely(!__pyx_tuple__152)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__152);
  __Pyx_GIVEREF(__pyx_tuple__152);
  __pyx_codeobj__153 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__152, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_Uint64ClosedRight, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__153)) __PYX_ERR(2, 1, __pyx_L1_error)
  __pyx_tuple__154 = PyTuple_Pack(5, __pyx_n_s_pyx_type, __pyx_n_s_pyx_checksum, __pyx_n_s_pyx_state, __pyx_n_s_pyx_PickleError, __pyx_n_s_pyx_result); if (unlikely(!__pyx_tuple__154)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__154);
  __Pyx_GIVEREF(__pyx_tuple__154);
  __pyx_codeobj__155 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__154, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_Uint64ClosedBothI, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__155)) __PYX_ERR(2, 1, __pyx_L1_error)
  __pyx_tuple__156 = PyTuple_Pack(5, __pyx_n_s_pyx_type, __pyx_n_s_pyx_checksum, __pyx_n_s_pyx_state, __pyx_n_s_pyx_PickleError, __pyx_n_s_pyx_result); if (unlikely(!__pyx_tuple__156)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__156);
  __Pyx_GIVEREF(__pyx_tuple__156);
  __pyx_codeobj__157 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__156, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_Uint64ClosedNeith, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__157)) __PYX_ERR(2, 1, __pyx_L1_error)

  /* "View.MemoryView":286
 *         return self.name
 * 
 * cdef generic = Enum("<strided and direct or indirect>")             # <<<<<<<<<<<<<<
 * cdef strided = Enum("<strided and direct>") # default
 * cdef indirect = Enum("<strided and indirect>")
 */
  __pyx_tuple__158 = PyTuple_Pack(1, __pyx_kp_s_strided_and_direct_or_indirect); if (unlikely(!__pyx_tuple__158)) __PYX_ERR(2, 286, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__158);
  __Pyx_GIVEREF(__pyx_tuple__158);

  /* "View.MemoryView":287
 * 
 * cdef generic = Enum("<strided and direct or indirect>")
 * cdef strided = Enum("<strided and direct>") # default             # <<<<<<<<<<<<<<
 * cdef indirect = Enum("<strided and indirect>")
 * 
 */
  __pyx_tuple__159 = PyTuple_Pack(1, __pyx_kp_s_strided_and_direct); if (unlikely(!__pyx_tuple__159)) __PYX_ERR(2, 287, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__159);
  __Pyx_GIVEREF(__pyx_tuple__159);

  /* "View.MemoryView":288
 * cdef generic = Enum("<strided and direct or indirect>")
 * cdef strided = Enum("<strided and direct>") # default
 * cdef indirect = Enum("<strided and indirect>")             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_tuple__160 = PyTuple_Pack(1, __pyx_kp_s_strided_and_indirect); if (unlikely(!__pyx_tuple__160)) __PYX_ERR(2, 288, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__160);
  __Pyx_GIVEREF(__pyx_tuple__160);

  /* "View.MemoryView":291
 * 
 * 
 * cdef contiguous = Enum("<contiguous and direct>")             # <<<<<<<<<<<<<<
 * cdef indirect_contiguous = Enum("<contiguous and indirect>")
 * 
 */
  __pyx_tuple__161 = PyTuple_Pack(1, __pyx_kp_s_contiguous_and_direct); if (unlikely(!__pyx_tuple__161)) __PYX_ERR(2, 291, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__161);
  __Pyx_GIVEREF(__pyx_tuple__161);

  /* "View.MemoryView":292
 * 
 * cdef contiguous = Enum("<contiguous and direct>")
 * cdef indirect_contiguous = Enum("<contiguous and indirect>")             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_tuple__162 = PyTuple_Pack(1, __pyx_kp_s_contiguous_and_indirect); if (unlikely(!__pyx_tuple__162)) __PYX_ERR(2, 292, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__162);
  __Pyx_GIVEREF(__pyx_tuple__162);

  /* "(tree fragment)":1
 * def __pyx_unpickle_Enum(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_tuple__163 = PyTuple_Pack(5, __pyx_n_s_pyx_type, __pyx_n_s_pyx_checksum, __pyx_n_s_pyx_state, __pyx_n_s_pyx_PickleError, __pyx_n_s_pyx_result); if (unlikely(!__pyx_tuple__163)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__163);
  __Pyx_GIVEREF(__pyx_tuple__163);
  __pyx_codeobj__164 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__163, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_Enum, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__164)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitGlobals(void) {
  if (__Pyx_InitStrings(__pyx_string_tab) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  __pyx_float_0_5 = PyFloat_FromDouble(0.5); if (unlikely(!__pyx_float_0_5)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_0 = PyInt_FromLong(0); if (unlikely(!__pyx_int_0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_1 = PyInt_FromLong(1); if (unlikely(!__pyx_int_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_2 = PyInt_FromLong(2); if (unlikely(!__pyx_int_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_3 = PyInt_FromLong(3); if (unlikely(!__pyx_int_3)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_100 = PyInt_FromLong(100); if (unlikely(!__pyx_int_100)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_5571323 = PyInt_FromLong(5571323L); if (unlikely(!__pyx_int_5571323)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_127993061 = PyInt_FromLong(127993061L); if (unlikely(!__pyx_int_127993061)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_184977713 = PyInt_FromLong(184977713L); if (unlikely(!__pyx_int_184977713)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_222419149 = PyInt_FromLong(222419149L); if (unlikely(!__pyx_int_222419149)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_neg_1 = PyInt_FromLong(-1); if (unlikely(!__pyx_int_neg_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_modinit_global_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_import_code(void); /*proto*/

static int __Pyx_modinit_global_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  generic = Py_None; Py_INCREF(Py_None);
  strided = Py_None; Py_INCREF(Py_None);
  indirect = Py_None; Py_INCREF(Py_None);
  contiguous = Py_None; Py_INCREF(Py_None);
  indirect_contiguous = Py_None; Py_INCREF(Py_None);
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_type_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  if (PyType_Ready(&__pyx_type_6pandas_5_libs_8interval_IntervalMixin) < 0) __PYX_ERR(0, 32, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6pandas_5_libs_8interval_IntervalMixin.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6pandas_5_libs_8interval_IntervalMixin.tp_dictoffset && __pyx_type_6pandas_5_libs_8interval_IntervalMixin.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6pandas_5_libs_8interval_IntervalMixin.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_IntervalMixin, (PyObject *)&__pyx_type_6pandas_5_libs_8interval_IntervalMixin) < 0) __PYX_ERR(0, 32, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_6pandas_5_libs_8interval_IntervalMixin) < 0) __PYX_ERR(0, 32, __pyx_L1_error)
  __pyx_ptype_6pandas_5_libs_8interval_IntervalMixin = &__pyx_type_6pandas_5_libs_8interval_IntervalMixin;
  __pyx_type_6pandas_5_libs_8interval_Interval.tp_base = __pyx_ptype_6pandas_5_libs_8interval_IntervalMixin;
  if (PyType_Ready(&__pyx_type_6pandas_5_libs_8interval_Interval) < 0) __PYX_ERR(0, 190, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6pandas_5_libs_8interval_Interval.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6pandas_5_libs_8interval_Interval.tp_dictoffset && __pyx_type_6pandas_5_libs_8interval_Interval.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6pandas_5_libs_8interval_Interval.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Interval, (PyObject *)&__pyx_type_6pandas_5_libs_8interval_Interval) < 0) __PYX_ERR(0, 190, __pyx_L1_error)
  __pyx_ptype_6pandas_5_libs_8interval_Interval = &__pyx_type_6pandas_5_libs_8interval_Interval;
  __pyx_type_6pandas_5_libs_8interval_IntervalTree.tp_base = __pyx_ptype_6pandas_5_libs_8interval_IntervalMixin;
  if (PyType_Ready(&__pyx_type_6pandas_5_libs_8interval_IntervalTree) < 0) __PYX_ERR(1, 20, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6pandas_5_libs_8interval_IntervalTree.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6pandas_5_libs_8interval_IntervalTree.tp_dictoffset && __pyx_type_6pandas_5_libs_8interval_IntervalTree.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6pandas_5_libs_8interval_IntervalTree.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #if CYTHON_COMPILING_IN_CPYTHON
  {
    PyObject *wrapper = PyObject_GetAttrString((PyObject *)&__pyx_type_6pandas_5_libs_8interval_IntervalTree, "__init__"); if (unlikely(!wrapper)) __PYX_ERR(1, 20, __pyx_L1_error)
    if (Py_TYPE(wrapper) == &PyWrapperDescr_Type) {
      __pyx_wrapperbase_6pandas_5_libs_8interval_12IntervalTree___init__ = *((PyWrapperDescrObject *)wrapper)->d_base;
      __pyx_wrapperbase_6pandas_5_libs_8interval_12IntervalTree___init__.doc = __pyx_doc_6pandas_5_libs_8interval_12IntervalTree___init__;
      ((PyWrapperDescrObject *)wrapper)->d_base = &__pyx_wrapperbase_6pandas_5_libs_8interval_12IntervalTree___init__;
    }
  }
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_IntervalTree, (PyObject *)&__pyx_type_6pandas_5_libs_8interval_IntervalTree) < 0) __PYX_ERR(1, 20, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_6pandas_5_libs_8interval_IntervalTree) < 0) __PYX_ERR(1, 20, __pyx_L1_error)
  __pyx_ptype_6pandas_5_libs_8interval_IntervalTree = &__pyx_type_6pandas_5_libs_8interval_IntervalTree;
  __pyx_vtabptr_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode = &__pyx_vtable_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode;
  __pyx_vtable_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode.classify_intervals = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode *, __Pyx_memviewslice, __Pyx_memviewslice))__pyx_f_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_classify_intervals;
  __pyx_vtable_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode.new_child_node = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode *, PyArrayObject *, PyArrayObject *, PyArrayObject *, PyArrayObject *))__pyx_f_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_new_child_node;
  __pyx_vtable_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode.__pyx_fuse_0query = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_float64_t, int __pyx_skip_dispatch))__pyx_fuse_0__pyx_f_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_query;
  __pyx_vtable_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode.__pyx_fuse_1query = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_float32_t, int __pyx_skip_dispatch))__pyx_fuse_1__pyx_f_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_query;
  __pyx_vtable_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode.__pyx_fuse_2query = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_int64_t, int __pyx_skip_dispatch))__pyx_fuse_2__pyx_f_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_query;
  __pyx_vtable_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode.__pyx_fuse_3query = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_int32_t, int __pyx_skip_dispatch))__pyx_fuse_3__pyx_f_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_query;
  __pyx_vtable_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode.__pyx_fuse_4query = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_uint64_t, int __pyx_skip_dispatch))__pyx_fuse_4__pyx_f_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_query;
  if (PyType_Ready(&__pyx_type_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode) < 0) __PYX_ERR(1, 233, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode.tp_dictoffset && __pyx_type_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode.tp_dict, __pyx_vtabptr_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode) < 0) __PYX_ERR(1, 233, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Float32ClosedLeftIntervalNode, (PyObject *)&__pyx_type_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode) < 0) __PYX_ERR(1, 233, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode) < 0) __PYX_ERR(1, 233, __pyx_L1_error)
  __pyx_ptype_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode = &__pyx_type_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode;
  __pyx_vtabptr_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode = &__pyx_vtable_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode;
  __pyx_vtable_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode.classify_intervals = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode *, __Pyx_memviewslice, __Pyx_memviewslice))__pyx_f_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_classify_intervals;
  __pyx_vtable_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode.new_child_node = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode *, PyArrayObject *, PyArrayObject *, PyArrayObject *, PyArrayObject *))__pyx_f_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_new_child_node;
  __pyx_vtable_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode.__pyx_fuse_0query = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_float64_t, int __pyx_skip_dispatch))__pyx_fuse_0__pyx_f_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_query;
  __pyx_vtable_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode.__pyx_fuse_1query = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_float32_t, int __pyx_skip_dispatch))__pyx_fuse_1__pyx_f_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_query;
  __pyx_vtable_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode.__pyx_fuse_2query = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_int64_t, int __pyx_skip_dispatch))__pyx_fuse_2__pyx_f_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_query;
  __pyx_vtable_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode.__pyx_fuse_3query = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_int32_t, int __pyx_skip_dispatch))__pyx_fuse_3__pyx_f_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_query;
  __pyx_vtable_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode.__pyx_fuse_4query = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_uint64_t, int __pyx_skip_dispatch))__pyx_fuse_4__pyx_f_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_query;
  if (PyType_Ready(&__pyx_type_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode) < 0) __PYX_ERR(1, 403, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode.tp_dictoffset && __pyx_type_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode.tp_dict, __pyx_vtabptr_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode) < 0) __PYX_ERR(1, 403, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Float32ClosedRightIntervalNode_3, (PyObject *)&__pyx_type_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode) < 0) __PYX_ERR(1, 403, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode) < 0) __PYX_ERR(1, 403, __pyx_L1_error)
  __pyx_ptype_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode = &__pyx_type_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode;
  __pyx_vtabptr_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode = &__pyx_vtable_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode;
  __pyx_vtable_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode.classify_intervals = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode *, __Pyx_memviewslice, __Pyx_memviewslice))__pyx_f_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_classify_intervals;
  __pyx_vtable_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode.new_child_node = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode *, PyArrayObject *, PyArrayObject *, PyArrayObject *, PyArrayObject *))__pyx_f_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_new_child_node;
  __pyx_vtable_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode.__pyx_fuse_0query = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_float64_t, int __pyx_skip_dispatch))__pyx_fuse_0__pyx_f_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_query;
  __pyx_vtable_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode.__pyx_fuse_1query = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_float32_t, int __pyx_skip_dispatch))__pyx_fuse_1__pyx_f_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_query;
  __pyx_vtable_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode.__pyx_fuse_2query = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_int64_t, int __pyx_skip_dispatch))__pyx_fuse_2__pyx_f_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_query;
  __pyx_vtable_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode.__pyx_fuse_3query = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_int32_t, int __pyx_skip_dispatch))__pyx_fuse_3__pyx_f_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_query;
  __pyx_vtable_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode.__pyx_fuse_4query = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_uint64_t, int __pyx_skip_dispatch))__pyx_fuse_4__pyx_f_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_query;
  if (PyType_Ready(&__pyx_type_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode) < 0) __PYX_ERR(1, 573, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode.tp_dictoffset && __pyx_type_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode.tp_dict, __pyx_vtabptr_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode) < 0) __PYX_ERR(1, 573, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Float32ClosedBothIntervalNode, (PyObject *)&__pyx_type_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode) < 0) __PYX_ERR(1, 573, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode) < 0) __PYX_ERR(1, 573, __pyx_L1_error)
  __pyx_ptype_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode = &__pyx_type_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode;
  __pyx_vtabptr_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode = &__pyx_vtable_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode;
  __pyx_vtable_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode.classify_intervals = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode *, __Pyx_memviewslice, __Pyx_memviewslice))__pyx_f_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_classify_intervals;
  __pyx_vtable_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode.new_child_node = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode *, PyArrayObject *, PyArrayObject *, PyArrayObject *, PyArrayObject *))__pyx_f_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_new_child_node;
  __pyx_vtable_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode.__pyx_fuse_0query = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_float64_t, int __pyx_skip_dispatch))__pyx_fuse_0__pyx_f_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_query;
  __pyx_vtable_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode.__pyx_fuse_1query = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_float32_t, int __pyx_skip_dispatch))__pyx_fuse_1__pyx_f_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_query;
  __pyx_vtable_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode.__pyx_fuse_2query = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_int64_t, int __pyx_skip_dispatch))__pyx_fuse_2__pyx_f_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_query;
  __pyx_vtable_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode.__pyx_fuse_3query = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_int32_t, int __pyx_skip_dispatch))__pyx_fuse_3__pyx_f_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_query;
  __pyx_vtable_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode.__pyx_fuse_4query = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_uint64_t, int __pyx_skip_dispatch))__pyx_fuse_4__pyx_f_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_query;
  if (PyType_Ready(&__pyx_type_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode) < 0) __PYX_ERR(1, 743, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode.tp_dictoffset && __pyx_type_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode.tp_dict, __pyx_vtabptr_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode) < 0) __PYX_ERR(1, 743, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Float32ClosedNeitherIntervalNode, (PyObject *)&__pyx_type_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode) < 0) __PYX_ERR(1, 743, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode) < 0) __PYX_ERR(1, 743, __pyx_L1_error)
  __pyx_ptype_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode = &__pyx_type_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode;
  __pyx_vtabptr_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode = &__pyx_vtable_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode;
  __pyx_vtable_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode.classify_intervals = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode *, __Pyx_memviewslice, __Pyx_memviewslice))__pyx_f_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_classify_intervals;
  __pyx_vtable_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode.new_child_node = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode *, PyArrayObject *, PyArrayObject *, PyArrayObject *, PyArrayObject *))__pyx_f_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_new_child_node;
  __pyx_vtable_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode.__pyx_fuse_0query = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_float64_t, int __pyx_skip_dispatch))__pyx_fuse_0__pyx_f_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_query;
  __pyx_vtable_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode.__pyx_fuse_1query = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_float32_t, int __pyx_skip_dispatch))__pyx_fuse_1__pyx_f_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_query;
  __pyx_vtable_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode.__pyx_fuse_2query = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_int64_t, int __pyx_skip_dispatch))__pyx_fuse_2__pyx_f_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_query;
  __pyx_vtable_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode.__pyx_fuse_3query = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_int32_t, int __pyx_skip_dispatch))__pyx_fuse_3__pyx_f_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_query;
  __pyx_vtable_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode.__pyx_fuse_4query = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_uint64_t, int __pyx_skip_dispatch))__pyx_fuse_4__pyx_f_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_query;
  if (PyType_Ready(&__pyx_type_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode) < 0) __PYX_ERR(1, 913, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode.tp_dictoffset && __pyx_type_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode.tp_dict, __pyx_vtabptr_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode) < 0) __PYX_ERR(1, 913, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Float64ClosedLeftIntervalNode, (PyObject *)&__pyx_type_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode) < 0) __PYX_ERR(1, 913, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode) < 0) __PYX_ERR(1, 913, __pyx_L1_error)
  __pyx_ptype_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode = &__pyx_type_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode;
  __pyx_vtabptr_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode = &__pyx_vtable_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode;
  __pyx_vtable_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode.classify_intervals = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode *, __Pyx_memviewslice, __Pyx_memviewslice))__pyx_f_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_classify_intervals;
  __pyx_vtable_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode.new_child_node = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode *, PyArrayObject *, PyArrayObject *, PyArrayObject *, PyArrayObject *))__pyx_f_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_new_child_node;
  __pyx_vtable_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode.__pyx_fuse_0query = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_float64_t, int __pyx_skip_dispatch))__pyx_fuse_0__pyx_f_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_query;
  __pyx_vtable_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode.__pyx_fuse_1query = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_float32_t, int __pyx_skip_dispatch))__pyx_fuse_1__pyx_f_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_query;
  __pyx_vtable_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode.__pyx_fuse_2query = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_int64_t, int __pyx_skip_dispatch))__pyx_fuse_2__pyx_f_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_query;
  __pyx_vtable_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode.__pyx_fuse_3query = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_int32_t, int __pyx_skip_dispatch))__pyx_fuse_3__pyx_f_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_query;
  __pyx_vtable_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode.__pyx_fuse_4query = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_uint64_t, int __pyx_skip_dispatch))__pyx_fuse_4__pyx_f_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_query;
  if (PyType_Ready(&__pyx_type_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode) < 0) __PYX_ERR(1, 1083, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode.tp_dictoffset && __pyx_type_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode.tp_dict, __pyx_vtabptr_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode) < 0) __PYX_ERR(1, 1083, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Float64ClosedRightIntervalNode_3, (PyObject *)&__pyx_type_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode) < 0) __PYX_ERR(1, 1083, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode) < 0) __PYX_ERR(1, 1083, __pyx_L1_error)
  __pyx_ptype_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode = &__pyx_type_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode;
  __pyx_vtabptr_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode = &__pyx_vtable_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode;
  __pyx_vtable_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode.classify_intervals = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode *, __Pyx_memviewslice, __Pyx_memviewslice))__pyx_f_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_classify_intervals;
  __pyx_vtable_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode.new_child_node = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode *, PyArrayObject *, PyArrayObject *, PyArrayObject *, PyArrayObject *))__pyx_f_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_new_child_node;
  __pyx_vtable_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode.__pyx_fuse_0query = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_float64_t, int __pyx_skip_dispatch))__pyx_fuse_0__pyx_f_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_query;
  __pyx_vtable_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode.__pyx_fuse_1query = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_float32_t, int __pyx_skip_dispatch))__pyx_fuse_1__pyx_f_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_query;
  __pyx_vtable_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode.__pyx_fuse_2query = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_int64_t, int __pyx_skip_dispatch))__pyx_fuse_2__pyx_f_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_query;
  __pyx_vtable_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode.__pyx_fuse_3query = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_int32_t, int __pyx_skip_dispatch))__pyx_fuse_3__pyx_f_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_query;
  __pyx_vtable_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode.__pyx_fuse_4query = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_uint64_t, int __pyx_skip_dispatch))__pyx_fuse_4__pyx_f_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_query;
  if (PyType_Ready(&__pyx_type_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode) < 0) __PYX_ERR(1, 1253, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode.tp_dictoffset && __pyx_type_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode.tp_dict, __pyx_vtabptr_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode) < 0) __PYX_ERR(1, 1253, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Float64ClosedBothIntervalNode, (PyObject *)&__pyx_type_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode) < 0) __PYX_ERR(1, 1253, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode) < 0) __PYX_ERR(1, 1253, __pyx_L1_error)
  __pyx_ptype_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode = &__pyx_type_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode;
  __pyx_vtabptr_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode = &__pyx_vtable_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode;
  __pyx_vtable_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode.classify_intervals = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode *, __Pyx_memviewslice, __Pyx_memviewslice))__pyx_f_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_classify_intervals;
  __pyx_vtable_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode.new_child_node = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode *, PyArrayObject *, PyArrayObject *, PyArrayObject *, PyArrayObject *))__pyx_f_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_new_child_node;
  __pyx_vtable_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode.__pyx_fuse_0query = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_float64_t, int __pyx_skip_dispatch))__pyx_fuse_0__pyx_f_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_query;
  __pyx_vtable_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode.__pyx_fuse_1query = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_float32_t, int __pyx_skip_dispatch))__pyx_fuse_1__pyx_f_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_query;
  __pyx_vtable_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode.__pyx_fuse_2query = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_int64_t, int __pyx_skip_dispatch))__pyx_fuse_2__pyx_f_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_query;
  __pyx_vtable_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode.__pyx_fuse_3query = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_int32_t, int __pyx_skip_dispatch))__pyx_fuse_3__pyx_f_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_query;
  __pyx_vtable_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode.__pyx_fuse_4query = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_uint64_t, int __pyx_skip_dispatch))__pyx_fuse_4__pyx_f_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_query;
  if (PyType_Ready(&__pyx_type_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode) < 0) __PYX_ERR(1, 1423, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode.tp_dictoffset && __pyx_type_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode.tp_dict, __pyx_vtabptr_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode) < 0) __PYX_ERR(1, 1423, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Float64ClosedNeitherIntervalNode, (PyObject *)&__pyx_type_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode) < 0) __PYX_ERR(1, 1423, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode) < 0) __PYX_ERR(1, 1423, __pyx_L1_error)
  __pyx_ptype_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode = &__pyx_type_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode;
  __pyx_vtabptr_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode = &__pyx_vtable_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode;
  __pyx_vtable_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode.classify_intervals = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode *, __Pyx_memviewslice, __Pyx_memviewslice))__pyx_f_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_classify_intervals;
  __pyx_vtable_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode.new_child_node = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode *, PyArrayObject *, PyArrayObject *, PyArrayObject *, PyArrayObject *))__pyx_f_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_new_child_node;
  __pyx_vtable_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode.__pyx_fuse_0query = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_float64_t, int __pyx_skip_dispatch))__pyx_fuse_0__pyx_f_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_query;
  __pyx_vtable_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode.__pyx_fuse_1query = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_float32_t, int __pyx_skip_dispatch))__pyx_fuse_1__pyx_f_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_query;
  __pyx_vtable_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode.__pyx_fuse_2query = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_int64_t, int __pyx_skip_dispatch))__pyx_fuse_2__pyx_f_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_query;
  __pyx_vtable_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode.__pyx_fuse_3query = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_int32_t, int __pyx_skip_dispatch))__pyx_fuse_3__pyx_f_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_query;
  __pyx_vtable_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode.__pyx_fuse_4query = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_uint64_t, int __pyx_skip_dispatch))__pyx_fuse_4__pyx_f_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_query;
  if (PyType_Ready(&__pyx_type_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode) < 0) __PYX_ERR(1, 1593, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode.tp_dictoffset && __pyx_type_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode.tp_dict, __pyx_vtabptr_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode) < 0) __PYX_ERR(1, 1593, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Int32ClosedLeftIntervalNode, (PyObject *)&__pyx_type_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode) < 0) __PYX_ERR(1, 1593, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode) < 0) __PYX_ERR(1, 1593, __pyx_L1_error)
  __pyx_ptype_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode = &__pyx_type_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode;
  __pyx_vtabptr_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode = &__pyx_vtable_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode;
  __pyx_vtable_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode.classify_intervals = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode *, __Pyx_memviewslice, __Pyx_memviewslice))__pyx_f_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_classify_intervals;
  __pyx_vtable_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode.new_child_node = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode *, PyArrayObject *, PyArrayObject *, PyArrayObject *, PyArrayObject *))__pyx_f_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_new_child_node;
  __pyx_vtable_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode.__pyx_fuse_0query = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_float64_t, int __pyx_skip_dispatch))__pyx_fuse_0__pyx_f_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_query;
  __pyx_vtable_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode.__pyx_fuse_1query = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_float32_t, int __pyx_skip_dispatch))__pyx_fuse_1__pyx_f_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_query;
  __pyx_vtable_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode.__pyx_fuse_2query = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_int64_t, int __pyx_skip_dispatch))__pyx_fuse_2__pyx_f_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_query;
  __pyx_vtable_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode.__pyx_fuse_3query = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_int32_t, int __pyx_skip_dispatch))__pyx_fuse_3__pyx_f_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_query;
  __pyx_vtable_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode.__pyx_fuse_4query = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_uint64_t, int __pyx_skip_dispatch))__pyx_fuse_4__pyx_f_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_query;
  if (PyType_Ready(&__pyx_type_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode) < 0) __PYX_ERR(1, 1763, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode.tp_dictoffset && __pyx_type_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode.tp_dict, __pyx_vtabptr_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode) < 0) __PYX_ERR(1, 1763, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Int32ClosedRightIntervalNode, (PyObject *)&__pyx_type_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode) < 0) __PYX_ERR(1, 1763, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode) < 0) __PYX_ERR(1, 1763, __pyx_L1_error)
  __pyx_ptype_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode = &__pyx_type_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode;
  __pyx_vtabptr_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode = &__pyx_vtable_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode;
  __pyx_vtable_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode.classify_intervals = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode *, __Pyx_memviewslice, __Pyx_memviewslice))__pyx_f_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_classify_intervals;
  __pyx_vtable_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode.new_child_node = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode *, PyArrayObject *, PyArrayObject *, PyArrayObject *, PyArrayObject *))__pyx_f_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_new_child_node;
  __pyx_vtable_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode.__pyx_fuse_0query = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_float64_t, int __pyx_skip_dispatch))__pyx_fuse_0__pyx_f_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_query;
  __pyx_vtable_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode.__pyx_fuse_1query = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_float32_t, int __pyx_skip_dispatch))__pyx_fuse_1__pyx_f_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_query;
  __pyx_vtable_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode.__pyx_fuse_2query = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_int64_t, int __pyx_skip_dispatch))__pyx_fuse_2__pyx_f_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_query;
  __pyx_vtable_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode.__pyx_fuse_3query = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_int32_t, int __pyx_skip_dispatch))__pyx_fuse_3__pyx_f_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_query;
  __pyx_vtable_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode.__pyx_fuse_4query = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_uint64_t, int __pyx_skip_dispatch))__pyx_fuse_4__pyx_f_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_query;
  if (PyType_Ready(&__pyx_type_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode) < 0) __PYX_ERR(1, 1933, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode.tp_dictoffset && __pyx_type_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode.tp_dict, __pyx_vtabptr_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode) < 0) __PYX_ERR(1, 1933, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Int32ClosedBothIntervalNode, (PyObject *)&__pyx_type_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode) < 0) __PYX_ERR(1, 1933, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode) < 0) __PYX_ERR(1, 1933, __pyx_L1_error)
  __pyx_ptype_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode = &__pyx_type_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode;
  __pyx_vtabptr_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode = &__pyx_vtable_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode;
  __pyx_vtable_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode.classify_intervals = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode *, __Pyx_memviewslice, __Pyx_memviewslice))__pyx_f_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_classify_intervals;
  __pyx_vtable_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode.new_child_node = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode *, PyArrayObject *, PyArrayObject *, PyArrayObject *, PyArrayObject *))__pyx_f_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_new_child_node;
  __pyx_vtable_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode.__pyx_fuse_0query = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_float64_t, int __pyx_skip_dispatch))__pyx_fuse_0__pyx_f_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_query;
  __pyx_vtable_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode.__pyx_fuse_1query = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_float32_t, int __pyx_skip_dispatch))__pyx_fuse_1__pyx_f_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_query;
  __pyx_vtable_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode.__pyx_fuse_2query = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_int64_t, int __pyx_skip_dispatch))__pyx_fuse_2__pyx_f_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_query;
  __pyx_vtable_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode.__pyx_fuse_3query = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_int32_t, int __pyx_skip_dispatch))__pyx_fuse_3__pyx_f_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_query;
  __pyx_vtable_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode.__pyx_fuse_4query = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_uint64_t, int __pyx_skip_dispatch))__pyx_fuse_4__pyx_f_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_query;
  if (PyType_Ready(&__pyx_type_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode) < 0) __PYX_ERR(1, 2103, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode.tp_dictoffset && __pyx_type_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode.tp_dict, __pyx_vtabptr_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode) < 0) __PYX_ERR(1, 2103, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Int32ClosedNeitherIntervalNode_3, (PyObject *)&__pyx_type_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode) < 0) __PYX_ERR(1, 2103, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode) < 0) __PYX_ERR(1, 2103, __pyx_L1_error)
  __pyx_ptype_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode = &__pyx_type_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode;
  __pyx_vtabptr_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode = &__pyx_vtable_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode;
  __pyx_vtable_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode.classify_intervals = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode *, __Pyx_memviewslice, __Pyx_memviewslice))__pyx_f_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_classify_intervals;
  __pyx_vtable_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode.new_child_node = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode *, PyArrayObject *, PyArrayObject *, PyArrayObject *, PyArrayObject *))__pyx_f_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_new_child_node;
  __pyx_vtable_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode.__pyx_fuse_0query = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_float64_t, int __pyx_skip_dispatch))__pyx_fuse_0__pyx_f_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_query;
  __pyx_vtable_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode.__pyx_fuse_1query = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_float32_t, int __pyx_skip_dispatch))__pyx_fuse_1__pyx_f_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_query;
  __pyx_vtable_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode.__pyx_fuse_2query = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_int64_t, int __pyx_skip_dispatch))__pyx_fuse_2__pyx_f_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_query;
  __pyx_vtable_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode.__pyx_fuse_3query = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_int32_t, int __pyx_skip_dispatch))__pyx_fuse_3__pyx_f_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_query;
  __pyx_vtable_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode.__pyx_fuse_4query = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_uint64_t, int __pyx_skip_dispatch))__pyx_fuse_4__pyx_f_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_query;
  if (PyType_Ready(&__pyx_type_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode) < 0) __PYX_ERR(1, 2273, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode.tp_dictoffset && __pyx_type_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode.tp_dict, __pyx_vtabptr_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode) < 0) __PYX_ERR(1, 2273, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Int64ClosedLeftIntervalNode, (PyObject *)&__pyx_type_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode) < 0) __PYX_ERR(1, 2273, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode) < 0) __PYX_ERR(1, 2273, __pyx_L1_error)
  __pyx_ptype_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode = &__pyx_type_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode;
  __pyx_vtabptr_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode = &__pyx_vtable_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode;
  __pyx_vtable_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode.classify_intervals = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode *, __Pyx_memviewslice, __Pyx_memviewslice))__pyx_f_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_classify_intervals;
  __pyx_vtable_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode.new_child_node = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode *, PyArrayObject *, PyArrayObject *, PyArrayObject *, PyArrayObject *))__pyx_f_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_new_child_node;
  __pyx_vtable_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode.__pyx_fuse_0query = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_float64_t, int __pyx_skip_dispatch))__pyx_fuse_0__pyx_f_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_query;
  __pyx_vtable_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode.__pyx_fuse_1query = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_float32_t, int __pyx_skip_dispatch))__pyx_fuse_1__pyx_f_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_query;
  __pyx_vtable_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode.__pyx_fuse_2query = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_int64_t, int __pyx_skip_dispatch))__pyx_fuse_2__pyx_f_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_query;
  __pyx_vtable_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode.__pyx_fuse_3query = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_int32_t, int __pyx_skip_dispatch))__pyx_fuse_3__pyx_f_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_query;
  __pyx_vtable_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode.__pyx_fuse_4query = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_uint64_t, int __pyx_skip_dispatch))__pyx_fuse_4__pyx_f_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_query;
  if (PyType_Ready(&__pyx_type_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode) < 0) __PYX_ERR(1, 2443, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode.tp_dictoffset && __pyx_type_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode.tp_dict, __pyx_vtabptr_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode) < 0) __PYX_ERR(1, 2443, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Int64ClosedRightIntervalNode, (PyObject *)&__pyx_type_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode) < 0) __PYX_ERR(1, 2443, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode) < 0) __PYX_ERR(1, 2443, __pyx_L1_error)
  __pyx_ptype_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode = &__pyx_type_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode;
  __pyx_vtabptr_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode = &__pyx_vtable_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode;
  __pyx_vtable_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode.classify_intervals = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode *, __Pyx_memviewslice, __Pyx_memviewslice))__pyx_f_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_classify_intervals;
  __pyx_vtable_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode.new_child_node = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode *, PyArrayObject *, PyArrayObject *, PyArrayObject *, PyArrayObject *))__pyx_f_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_new_child_node;
  __pyx_vtable_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode.__pyx_fuse_0query = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_float64_t, int __pyx_skip_dispatch))__pyx_fuse_0__pyx_f_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_query;
  __pyx_vtable_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode.__pyx_fuse_1query = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_float32_t, int __pyx_skip_dispatch))__pyx_fuse_1__pyx_f_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_query;
  __pyx_vtable_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode.__pyx_fuse_2query = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_int64_t, int __pyx_skip_dispatch))__pyx_fuse_2__pyx_f_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_query;
  __pyx_vtable_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode.__pyx_fuse_3query = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_int32_t, int __pyx_skip_dispatch))__pyx_fuse_3__pyx_f_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_query;
  __pyx_vtable_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode.__pyx_fuse_4query = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_uint64_t, int __pyx_skip_dispatch))__pyx_fuse_4__pyx_f_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_query;
  if (PyType_Ready(&__pyx_type_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode) < 0) __PYX_ERR(1, 2613, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode.tp_dictoffset && __pyx_type_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode.tp_dict, __pyx_vtabptr_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode) < 0) __PYX_ERR(1, 2613, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Int64ClosedBothIntervalNode, (PyObject *)&__pyx_type_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode) < 0) __PYX_ERR(1, 2613, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode) < 0) __PYX_ERR(1, 2613, __pyx_L1_error)
  __pyx_ptype_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode = &__pyx_type_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode;
  __pyx_vtabptr_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode = &__pyx_vtable_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode;
  __pyx_vtable_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode.classify_intervals = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode *, __Pyx_memviewslice, __Pyx_memviewslice))__pyx_f_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_classify_intervals;
  __pyx_vtable_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode.new_child_node = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode *, PyArrayObject *, PyArrayObject *, PyArrayObject *, PyArrayObject *))__pyx_f_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_new_child_node;
  __pyx_vtable_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode.__pyx_fuse_0query = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_float64_t, int __pyx_skip_dispatch))__pyx_fuse_0__pyx_f_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_query;
  __pyx_vtable_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode.__pyx_fuse_1query = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_float32_t, int __pyx_skip_dispatch))__pyx_fuse_1__pyx_f_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_query;
  __pyx_vtable_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode.__pyx_fuse_2query = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_int64_t, int __pyx_skip_dispatch))__pyx_fuse_2__pyx_f_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_query;
  __pyx_vtable_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode.__pyx_fuse_3query = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_int32_t, int __pyx_skip_dispatch))__pyx_fuse_3__pyx_f_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_query;
  __pyx_vtable_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode.__pyx_fuse_4query = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_uint64_t, int __pyx_skip_dispatch))__pyx_fuse_4__pyx_f_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_query;
  if (PyType_Ready(&__pyx_type_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode) < 0) __PYX_ERR(1, 2783, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode.tp_dictoffset && __pyx_type_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode.tp_dict, __pyx_vtabptr_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode) < 0) __PYX_ERR(1, 2783, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Int64ClosedNeitherIntervalNode_3, (PyObject *)&__pyx_type_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode) < 0) __PYX_ERR(1, 2783, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode) < 0) __PYX_ERR(1, 2783, __pyx_L1_error)
  __pyx_ptype_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode = &__pyx_type_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode;
  __pyx_vtabptr_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode = &__pyx_vtable_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode;
  __pyx_vtable_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode.classify_intervals = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode *, __Pyx_memviewslice, __Pyx_memviewslice))__pyx_f_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_classify_intervals;
  __pyx_vtable_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode.new_child_node = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode *, PyArrayObject *, PyArrayObject *, PyArrayObject *, PyArrayObject *))__pyx_f_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_new_child_node;
  __pyx_vtable_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode.__pyx_fuse_0query = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_float64_t, int __pyx_skip_dispatch))__pyx_fuse_0__pyx_f_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_query;
  __pyx_vtable_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode.__pyx_fuse_1query = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_float32_t, int __pyx_skip_dispatch))__pyx_fuse_1__pyx_f_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_query;
  __pyx_vtable_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode.__pyx_fuse_2query = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_int64_t, int __pyx_skip_dispatch))__pyx_fuse_2__pyx_f_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_query;
  __pyx_vtable_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode.__pyx_fuse_3query = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_int32_t, int __pyx_skip_dispatch))__pyx_fuse_3__pyx_f_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_query;
  __pyx_vtable_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode.__pyx_fuse_4query = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_uint64_t, int __pyx_skip_dispatch))__pyx_fuse_4__pyx_f_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_query;
  if (PyType_Ready(&__pyx_type_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode) < 0) __PYX_ERR(1, 2953, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode.tp_dictoffset && __pyx_type_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode.tp_dict, __pyx_vtabptr_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode) < 0) __PYX_ERR(1, 2953, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Uint64ClosedLeftIntervalNode, (PyObject *)&__pyx_type_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode) < 0) __PYX_ERR(1, 2953, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode) < 0) __PYX_ERR(1, 2953, __pyx_L1_error)
  __pyx_ptype_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode = &__pyx_type_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode;
  __pyx_vtabptr_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode = &__pyx_vtable_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode;
  __pyx_vtable_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode.classify_intervals = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode *, __Pyx_memviewslice, __Pyx_memviewslice))__pyx_f_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_classify_intervals;
  __pyx_vtable_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode.new_child_node = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode *, PyArrayObject *, PyArrayObject *, PyArrayObject *, PyArrayObject *))__pyx_f_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_new_child_node;
  __pyx_vtable_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode.__pyx_fuse_0query = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_float64_t, int __pyx_skip_dispatch))__pyx_fuse_0__pyx_f_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_query;
  __pyx_vtable_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode.__pyx_fuse_1query = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_float32_t, int __pyx_skip_dispatch))__pyx_fuse_1__pyx_f_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_query;
  __pyx_vtable_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode.__pyx_fuse_2query = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_int64_t, int __pyx_skip_dispatch))__pyx_fuse_2__pyx_f_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_query;
  __pyx_vtable_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode.__pyx_fuse_3query = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_int32_t, int __pyx_skip_dispatch))__pyx_fuse_3__pyx_f_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_query;
  __pyx_vtable_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode.__pyx_fuse_4query = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_uint64_t, int __pyx_skip_dispatch))__pyx_fuse_4__pyx_f_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_query;
  if (PyType_Ready(&__pyx_type_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode) < 0) __PYX_ERR(1, 3123, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode.tp_dictoffset && __pyx_type_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode.tp_dict, __pyx_vtabptr_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode) < 0) __PYX_ERR(1, 3123, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Uint64ClosedRightIntervalNode, (PyObject *)&__pyx_type_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode) < 0) __PYX_ERR(1, 3123, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode) < 0) __PYX_ERR(1, 3123, __pyx_L1_error)
  __pyx_ptype_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode = &__pyx_type_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode;
  __pyx_vtabptr_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode = &__pyx_vtable_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode;
  __pyx_vtable_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode.classify_intervals = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode *, __Pyx_memviewslice, __Pyx_memviewslice))__pyx_f_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_classify_intervals;
  __pyx_vtable_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode.new_child_node = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode *, PyArrayObject *, PyArrayObject *, PyArrayObject *, PyArrayObject *))__pyx_f_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_new_child_node;
  __pyx_vtable_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode.__pyx_fuse_0query = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_float64_t, int __pyx_skip_dispatch))__pyx_fuse_0__pyx_f_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_query;
  __pyx_vtable_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode.__pyx_fuse_1query = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_float32_t, int __pyx_skip_dispatch))__pyx_fuse_1__pyx_f_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_query;
  __pyx_vtable_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode.__pyx_fuse_2query = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_int64_t, int __pyx_skip_dispatch))__pyx_fuse_2__pyx_f_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_query;
  __pyx_vtable_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode.__pyx_fuse_3query = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_int32_t, int __pyx_skip_dispatch))__pyx_fuse_3__pyx_f_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_query;
  __pyx_vtable_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode.__pyx_fuse_4query = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_uint64_t, int __pyx_skip_dispatch))__pyx_fuse_4__pyx_f_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_query;
  if (PyType_Ready(&__pyx_type_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode) < 0) __PYX_ERR(1, 3293, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode.tp_dictoffset && __pyx_type_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode.tp_dict, __pyx_vtabptr_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode) < 0) __PYX_ERR(1, 3293, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Uint64ClosedBothIntervalNode, (PyObject *)&__pyx_type_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode) < 0) __PYX_ERR(1, 3293, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode) < 0) __PYX_ERR(1, 3293, __pyx_L1_error)
  __pyx_ptype_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode = &__pyx_type_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode;
  __pyx_vtabptr_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode = &__pyx_vtable_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode;
  __pyx_vtable_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode.classify_intervals = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode *, __Pyx_memviewslice, __Pyx_memviewslice))__pyx_f_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_classify_intervals;
  __pyx_vtable_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode.new_child_node = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode *, PyArrayObject *, PyArrayObject *, PyArrayObject *, PyArrayObject *))__pyx_f_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_new_child_node;
  __pyx_vtable_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode.__pyx_fuse_0query = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_float64_t, int __pyx_skip_dispatch))__pyx_fuse_0__pyx_f_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_query;
  __pyx_vtable_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode.__pyx_fuse_1query = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_float32_t, int __pyx_skip_dispatch))__pyx_fuse_1__pyx_f_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_query;
  __pyx_vtable_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode.__pyx_fuse_2query = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_int64_t, int __pyx_skip_dispatch))__pyx_fuse_2__pyx_f_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_query;
  __pyx_vtable_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode.__pyx_fuse_3query = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_int32_t, int __pyx_skip_dispatch))__pyx_fuse_3__pyx_f_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_query;
  __pyx_vtable_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode.__pyx_fuse_4query = (PyObject *(*)(struct __pyx_obj_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode *, struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector *, __pyx_t_5numpy_uint64_t, int __pyx_skip_dispatch))__pyx_fuse_4__pyx_f_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_query;
  if (PyType_Ready(&__pyx_type_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode) < 0) __PYX_ERR(1, 3463, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode.tp_dictoffset && __pyx_type_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode.tp_dict, __pyx_vtabptr_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode) < 0) __PYX_ERR(1, 3463, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Uint64ClosedNeitherIntervalNode_3, (PyObject *)&__pyx_type_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode) < 0) __PYX_ERR(1, 3463, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode) < 0) __PYX_ERR(1, 3463, __pyx_L1_error)
  __pyx_ptype_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode = &__pyx_type_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode;
  __pyx_vtabptr_array = &__pyx_vtable_array;
  __pyx_vtable_array.get_memview = (PyObject *(*)(struct __pyx_array_obj *))__pyx_array_get_memview;
  if (PyType_Ready(&__pyx_type___pyx_array) < 0) __PYX_ERR(2, 105, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type___pyx_array.tp_print = 0;
  #endif
  if (__Pyx_SetVtable(__pyx_type___pyx_array.tp_dict, __pyx_vtabptr_array) < 0) __PYX_ERR(2, 105, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type___pyx_array) < 0) __PYX_ERR(2, 105, __pyx_L1_error)
  __pyx_array_type = &__pyx_type___pyx_array;
  if (PyType_Ready(&__pyx_type___pyx_MemviewEnum) < 0) __PYX_ERR(2, 279, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type___pyx_MemviewEnum.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type___pyx_MemviewEnum.tp_dictoffset && __pyx_type___pyx_MemviewEnum.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type___pyx_MemviewEnum.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type___pyx_MemviewEnum) < 0) __PYX_ERR(2, 279, __pyx_L1_error)
  __pyx_MemviewEnum_type = &__pyx_type___pyx_MemviewEnum;
  __pyx_vtabptr_memoryview = &__pyx_vtable_memoryview;
  __pyx_vtable_memoryview.get_item_pointer = (char *(*)(struct __pyx_memoryview_obj *, PyObject *))__pyx_memoryview_get_item_pointer;
  __pyx_vtable_memoryview.is_slice = (PyObject *(*)(struct __pyx_memoryview_obj *, PyObject *))__pyx_memoryview_is_slice;
  __pyx_vtable_memoryview.setitem_slice_assignment = (PyObject *(*)(struct __pyx_memoryview_obj *, PyObject *, PyObject *))__pyx_memoryview_setitem_slice_assignment;
  __pyx_vtable_memoryview.setitem_slice_assign_scalar = (PyObject *(*)(struct __pyx_memoryview_obj *, struct __pyx_memoryview_obj *, PyObject *))__pyx_memoryview_setitem_slice_assign_scalar;
  __pyx_vtable_memoryview.setitem_indexed = (PyObject *(*)(struct __pyx_memoryview_obj *, PyObject *, PyObject *))__pyx_memoryview_setitem_indexed;
  __pyx_vtable_memoryview.convert_item_to_object = (PyObject *(*)(struct __pyx_memoryview_obj *, char *))__pyx_memoryview_convert_item_to_object;
  __pyx_vtable_memoryview.assign_item_from_object = (PyObject *(*)(struct __pyx_memoryview_obj *, char *, PyObject *))__pyx_memoryview_assign_item_from_object;
  if (PyType_Ready(&__pyx_type___pyx_memoryview) < 0) __PYX_ERR(2, 330, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type___pyx_memoryview.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type___pyx_memoryview.tp_dictoffset && __pyx_type___pyx_memoryview.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type___pyx_memoryview.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type___pyx_memoryview.tp_dict, __pyx_vtabptr_memoryview) < 0) __PYX_ERR(2, 330, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type___pyx_memoryview) < 0) __PYX_ERR(2, 330, __pyx_L1_error)
  __pyx_memoryview_type = &__pyx_type___pyx_memoryview;
  __pyx_vtabptr__memoryviewslice = &__pyx_vtable__memoryviewslice;
  __pyx_vtable__memoryviewslice.__pyx_base = *__pyx_vtabptr_memoryview;
  __pyx_vtable__memoryviewslice.__pyx_base.convert_item_to_object = (PyObject *(*)(struct __pyx_memoryview_obj *, char *))__pyx_memoryviewslice_convert_item_to_object;
  __pyx_vtable__memoryviewslice.__pyx_base.assign_item_from_object = (PyObject *(*)(struct __pyx_memoryview_obj *, char *, PyObject *))__pyx_memoryviewslice_assign_item_from_object;
  __pyx_type___pyx_memoryviewslice.tp_base = __pyx_memoryview_type;
  if (PyType_Ready(&__pyx_type___pyx_memoryviewslice) < 0) __PYX_ERR(2, 965, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type___pyx_memoryviewslice.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type___pyx_memoryviewslice.tp_dictoffset && __pyx_type___pyx_memoryviewslice.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type___pyx_memoryviewslice.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type___pyx_memoryviewslice.tp_dict, __pyx_vtabptr__memoryviewslice) < 0) __PYX_ERR(2, 965, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type___pyx_memoryviewslice) < 0) __PYX_ERR(2, 965, __pyx_L1_error)
  __pyx_memoryviewslice_type = &__pyx_type___pyx_memoryviewslice;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_4type_type = __Pyx_ImportType(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "type", 
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyTypeObject),
  #else
  sizeof(PyHeapTypeObject),
  #endif
  __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7cpython_4type_type) __PYX_ERR(6, 9, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(7, 8, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_4bool_bool = __Pyx_ImportType(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "bool", sizeof(PyBoolObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7cpython_4bool_bool) __PYX_ERR(7, 8, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 15, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_7complex_complex = __Pyx_ImportType(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "complex", sizeof(PyComplexObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7cpython_7complex_complex) __PYX_ERR(8, 15, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("numpy"); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 206, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_5numpy_dtype = __Pyx_ImportType(__pyx_t_1, "numpy", "dtype", sizeof(PyArray_Descr), __Pyx_ImportType_CheckSize_Ignore);
   if (!__pyx_ptype_5numpy_dtype) __PYX_ERR(3, 206, __pyx_L1_error)
  __pyx_ptype_5numpy_flatiter = __Pyx_ImportType(__pyx_t_1, "numpy", "flatiter", sizeof(PyArrayIterObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_flatiter) __PYX_ERR(3, 229, __pyx_L1_error)
  __pyx_ptype_5numpy_broadcast = __Pyx_ImportType(__pyx_t_1, "numpy", "broadcast", sizeof(PyArrayMultiIterObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_broadcast) __PYX_ERR(3, 233, __pyx_L1_error)
  __pyx_ptype_5numpy_ndarray = __Pyx_ImportType(__pyx_t_1, "numpy", "ndarray", sizeof(PyArrayObject), __Pyx_ImportType_CheckSize_Ignore);
   if (!__pyx_ptype_5numpy_ndarray) __PYX_ERR(3, 242, __pyx_L1_error)
  __pyx_ptype_5numpy_ufunc = __Pyx_ImportType(__pyx_t_1, "numpy", "ufunc", sizeof(PyUFuncObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_ufunc) __PYX_ERR(3, 918, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("pandas._libs.hashtable"); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 8, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_6pandas_5_libs_9hashtable_HashTable = __Pyx_ImportType(__pyx_t_1, "pandas._libs.hashtable", "HashTable", sizeof(struct __pyx_obj_6pandas_5_libs_9hashtable_HashTable), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_6pandas_5_libs_9hashtable_HashTable) __PYX_ERR(9, 8, __pyx_L1_error)
  __pyx_ptype_6pandas_5_libs_9hashtable_UInt64HashTable = __Pyx_ImportType(__pyx_t_1, "pandas._libs.hashtable", "UInt64HashTable", sizeof(struct __pyx_obj_6pandas_5_libs_9hashtable_UInt64HashTable), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_6pandas_5_libs_9hashtable_UInt64HashTable) __PYX_ERR(9, 11, __pyx_L1_error)
  __pyx_vtabptr_6pandas_5_libs_9hashtable_UInt64HashTable = (struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_UInt64HashTable*)__Pyx_GetVtable(__pyx_ptype_6pandas_5_libs_9hashtable_UInt64HashTable->tp_dict); if (unlikely(!__pyx_vtabptr_6pandas_5_libs_9hashtable_UInt64HashTable)) __PYX_ERR(9, 11, __pyx_L1_error)
  __pyx_ptype_6pandas_5_libs_9hashtable_Int64HashTable = __Pyx_ImportType(__pyx_t_1, "pandas._libs.hashtable", "Int64HashTable", sizeof(struct __pyx_obj_6pandas_5_libs_9hashtable_Int64HashTable), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_6pandas_5_libs_9hashtable_Int64HashTable) __PYX_ERR(9, 17, __pyx_L1_error)
  __pyx_vtabptr_6pandas_5_libs_9hashtable_Int64HashTable = (struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64HashTable*)__Pyx_GetVtable(__pyx_ptype_6pandas_5_libs_9hashtable_Int64HashTable->tp_dict); if (unlikely(!__pyx_vtabptr_6pandas_5_libs_9hashtable_Int64HashTable)) __PYX_ERR(9, 17, __pyx_L1_error)
  __pyx_ptype_6pandas_5_libs_9hashtable_Float64HashTable = __Pyx_ImportType(__pyx_t_1, "pandas._libs.hashtable", "Float64HashTable", sizeof(struct __pyx_obj_6pandas_5_libs_9hashtable_Float64HashTable), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_6pandas_5_libs_9hashtable_Float64HashTable) __PYX_ERR(9, 23, __pyx_L1_error)
  __pyx_vtabptr_6pandas_5_libs_9hashtable_Float64HashTable = (struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Float64HashTable*)__Pyx_GetVtable(__pyx_ptype_6pandas_5_libs_9hashtable_Float64HashTable->tp_dict); if (unlikely(!__pyx_vtabptr_6pandas_5_libs_9hashtable_Float64HashTable)) __PYX_ERR(9, 23, __pyx_L1_error)
  __pyx_ptype_6pandas_5_libs_9hashtable_PyObjectHashTable = __Pyx_ImportType(__pyx_t_1, "pandas._libs.hashtable", "PyObjectHashTable", sizeof(struct __pyx_obj_6pandas_5_libs_9hashtable_PyObjectHashTable), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_6pandas_5_libs_9hashtable_PyObjectHashTable) __PYX_ERR(9, 29, __pyx_L1_error)
  __pyx_vtabptr_6pandas_5_libs_9hashtable_PyObjectHashTable = (struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_PyObjectHashTable*)__Pyx_GetVtable(__pyx_ptype_6pandas_5_libs_9hashtable_PyObjectHashTable->tp_dict); if (unlikely(!__pyx_vtabptr_6pandas_5_libs_9hashtable_PyObjectHashTable)) __PYX_ERR(9, 29, __pyx_L1_error)
  __pyx_ptype_6pandas_5_libs_9hashtable_StringHashTable = __Pyx_ImportType(__pyx_t_1, "pandas._libs.hashtable", "StringHashTable", sizeof(struct __pyx_obj_6pandas_5_libs_9hashtable_StringHashTable), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_6pandas_5_libs_9hashtable_StringHashTable) __PYX_ERR(9, 36, __pyx_L1_error)
  __pyx_vtabptr_6pandas_5_libs_9hashtable_StringHashTable = (struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_StringHashTable*)__Pyx_GetVtable(__pyx_ptype_6pandas_5_libs_9hashtable_StringHashTable->tp_dict); if (unlikely(!__pyx_vtabptr_6pandas_5_libs_9hashtable_StringHashTable)) __PYX_ERR(9, 36, __pyx_L1_error)
  __pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector = __Pyx_ImportType(__pyx_t_1, "pandas._libs.hashtable", "Int64Vector", sizeof(struct __pyx_obj_6pandas_5_libs_9hashtable_Int64Vector), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector) __PYX_ERR(9, 46, __pyx_L1_error)
  __pyx_vtabptr_6pandas_5_libs_9hashtable_Int64Vector = (struct __pyx_vtabstruct_6pandas_5_libs_9hashtable_Int64Vector*)__Pyx_GetVtable(__pyx_ptype_6pandas_5_libs_9hashtable_Int64Vector->tp_dict); if (unlikely(!__pyx_vtabptr_6pandas_5_libs_9hashtable_Int64Vector)) __PYX_ERR(9, 46, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_variable_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  __pyx_t_1 = PyImport_ImportModule("pandas._libs.tslibs.timezones"); if (!__pyx_t_1) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_1, "tz_compare", (void (**)(void))&__pyx_f_6pandas_5_libs_6tslibs_9timezones_tz_compare, "int (PyObject *, PyObject *, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}


#if PY_MAJOR_VERSION < 3
#ifdef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC void
#else
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#endif
#else
#ifdef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC PyObject *
#else
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#endif
#endif


#if PY_MAJOR_VERSION < 3
__Pyx_PyMODINIT_FUNC initinterval(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC initinterval(void)
#else
__Pyx_PyMODINIT_FUNC PyInit_interval(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit_interval(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
static CYTHON_SMALL_CODE int __Pyx_check_single_interpreter(void) {
    #if PY_VERSION_HEX >= 0x030700A1
    static PY_INT64_T main_interpreter_id = -1;
    PY_INT64_T current_id = PyInterpreterState_GetID(PyThreadState_Get()->interp);
    if (main_interpreter_id == -1) {
        main_interpreter_id = current_id;
        return (unlikely(current_id == -1)) ? -1 : 0;
    } else if (unlikely(main_interpreter_id != current_id))
    #else
    static PyInterpreterState *main_interpreter = NULL;
    PyInterpreterState *current_interpreter = PyThreadState_Get()->interp;
    if (!main_interpreter) {
        main_interpreter = current_interpreter;
    } else if (unlikely(main_interpreter != current_interpreter))
    #endif
    {
        PyErr_SetString(
            PyExc_ImportError,
            "Interpreter change detected - this module can only be loaded into one interpreter per process.");
        return -1;
    }
    return 0;
}
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name, int allow_none) {
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        if (allow_none || value != Py_None) {
            result = PyDict_SetItemString(moddict, to_name, value);
        }
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static CYTHON_SMALL_CODE PyObject* __pyx_pymod_create(PyObject *spec, CYTHON_UNUSED PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    if (__Pyx_check_single_interpreter())
        return NULL;
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__", 0) < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static CYTHON_SMALL_CODE int __pyx_pymod_exec_interval(PyObject *__pyx_pyinit_module)
#endif
#endif
{
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  PyObject *__pyx_t_21 = NULL;
  PyObject *__pyx_t_22 = NULL;
  PyObject *__pyx_t_23 = NULL;
  PyObject *__pyx_t_24 = NULL;
  PyObject *__pyx_t_25 = NULL;
  PyObject *__pyx_t_26 = NULL;
  PyObject *__pyx_t_27 = NULL;
  PyObject *__pyx_t_28 = NULL;
  PyObject *__pyx_t_29 = NULL;
  PyObject *__pyx_t_30 = NULL;
  PyObject *__pyx_t_31 = NULL;
  PyObject *__pyx_t_32 = NULL;
  PyObject *__pyx_t_33 = NULL;
  PyObject *__pyx_t_34 = NULL;
  PyObject *__pyx_t_35 = NULL;
  PyObject *__pyx_t_36 = NULL;
  PyObject *__pyx_t_37 = NULL;
  PyObject *__pyx_t_38 = NULL;
  PyObject *__pyx_t_39 = NULL;
  PyObject *__pyx_t_40 = NULL;
  PyObject *__pyx_t_41 = NULL;
  PyObject *__pyx_t_42 = NULL;
  PyObject *__pyx_t_43 = NULL;
  PyObject *__pyx_t_44 = NULL;
  PyObject *__pyx_t_45 = NULL;
  PyObject *__pyx_t_46 = NULL;
  PyObject *__pyx_t_47 = NULL;
  PyObject *__pyx_t_48 = NULL;
  static PyThread_type_lock __pyx_t_49[8];
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m) {
    if (__pyx_m == __pyx_pyinit_module) return 0;
    PyErr_SetString(PyExc_RuntimeError, "Module 'interval' has already been imported. Re-initialisation is not supported.");
    return -1;
  }
  #elif PY_MAJOR_VERSION >= 3
  if (__pyx_m) return __Pyx_NewRef(__pyx_m);
  #endif
  #if CYTHON_REFNANNY
__Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
if (!__Pyx_RefNanny) {
  PyErr_Clear();
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
  if (!__Pyx_RefNanny)
      Py_FatalError("failed to import 'refnanny' module");
}
#endif
  __Pyx_RefNannySetupContext("__Pyx_PyMODINIT_FUNC PyInit_interval(void)", 0);
  if (__Pyx_check_binary_version() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pxy_PyFrame_Initialize_Offsets
  __Pxy_PyFrame_Initialize_Offsets();
  #endif
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_AsyncGen_USED
  if (__pyx_AsyncGen_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  #ifdef WITH_THREAD /* Python build with threading support? */
  PyEval_InitThreads();
  #endif
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_m = __pyx_pyinit_module;
  Py_INCREF(__pyx_m);
  #else
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("interval", __pyx_methods, 0, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  #endif
  if (unlikely(!__pyx_m)) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_d);
  __pyx_b = PyImport_AddModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_b);
  __pyx_cython_runtime = PyImport_AddModule((char *) "cython_runtime"); if (unlikely(!__pyx_cython_runtime)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_cython_runtime);
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  if (__pyx_module_is_main_pandas___libs__interval) {
    if (PyObject_SetAttr(__pyx_m, __pyx_n_s_name_2, __pyx_n_s_main) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "pandas._libs.interval")) {
      if (unlikely(PyDict_SetItemString(modules, "pandas._libs.interval", __pyx_m) < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) goto __pyx_L1_error;
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) goto __pyx_L1_error;
  /*--- Global type/function init code ---*/
  (void)__Pyx_modinit_global_init_code();
  (void)__Pyx_modinit_variable_export_code();
  (void)__Pyx_modinit_function_export_code();
  if (unlikely(__Pyx_modinit_type_init_code() != 0)) goto __pyx_L1_error;
  if (unlikely(__Pyx_modinit_type_import_code() != 0)) goto __pyx_L1_error;
  (void)__Pyx_modinit_variable_import_code();
  if (unlikely(__Pyx_modinit_function_import_code() != 0)) goto __pyx_L1_error;
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif

  /* "pandas/_libs/interval.pyx":1
 * import numbers             # <<<<<<<<<<<<<<
 * from operator import le, lt
 * 
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_numbers, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_numbers, __pyx_t_1) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pandas/_libs/interval.pyx":2
 * import numbers
 * from operator import le, lt             # <<<<<<<<<<<<<<
 * 
 * from cpython.object cimport (Py_EQ, Py_NE, Py_GT, Py_LT, Py_GE, Py_LE,
 */
  __pyx_t_1 = PyList_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_le);
  __Pyx_GIVEREF(__pyx_n_s_le);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_le);
  __Pyx_INCREF(__pyx_n_s_lt);
  __Pyx_GIVEREF(__pyx_n_s_lt);
  PyList_SET_ITEM(__pyx_t_1, 1, __pyx_n_s_lt);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_operator, __pyx_t_1, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_le); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_le, __pyx_t_1) < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_lt); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_lt, __pyx_t_1) < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pandas/_libs/interval.pyx":10
 * from cython import Py_ssize_t
 * 
 * import numpy as np             # <<<<<<<<<<<<<<
 * cimport numpy as cnp
 * from numpy cimport (
 */
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_numpy, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 10, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_np, __pyx_t_2) < 0) __PYX_ERR(0, 10, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pandas/_libs/interval.pyx":16
 *     ndarray,
 *     PyArray_ArgSort, NPY_QUICKSORT, PyArray_Take)
 * cnp.import_array()             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_3 = __pyx_f_5numpy_import_array(); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 16, __pyx_L1_error)

  /* "pandas/_libs/interval.pyx":24
 * from pandas._libs.tslibs.util cimport is_integer_object, is_float_object
 * 
 * from pandas._libs.tslibs import Timestamp             # <<<<<<<<<<<<<<
 * from pandas._libs.tslibs.timedeltas import Timedelta
 * from pandas._libs.tslibs.timezones cimport tz_compare
 */
  __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_Timestamp);
  __Pyx_GIVEREF(__pyx_n_s_Timestamp);
  PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_Timestamp);
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_pandas__libs_tslibs, __pyx_t_2, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_Timestamp); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_Timestamp, __pyx_t_2) < 0) __PYX_ERR(0, 24, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pandas/_libs/interval.pyx":25
 * 
 * from pandas._libs.tslibs import Timestamp
 * from pandas._libs.tslibs.timedeltas import Timedelta             # <<<<<<<<<<<<<<
 * from pandas._libs.tslibs.timezones cimport tz_compare
 * 
 */
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_Timedelta);
  __Pyx_GIVEREF(__pyx_n_s_Timedelta);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_Timedelta);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_pandas__libs_tslibs_timedeltas, __pyx_t_1, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_Timedelta); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_Timedelta, __pyx_t_1) < 0) __PYX_ERR(0, 25, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pandas/_libs/interval.pyx":29
 * 
 * 
 * _VALID_CLOSED = frozenset(['left', 'right', 'both', 'neither'])             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = __Pyx_PyFrozenSet_New(__pyx_tuple__45); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_VALID_CLOSED, __pyx_t_2) < 0) __PYX_ERR(0, 29, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pandas/_libs/interval.pyx":285
 *     True
 *     """
 *     _typ = "interval"             # <<<<<<<<<<<<<<
 * 
 *     cdef readonly object left
 */
  if (PyDict_SetItem((PyObject *)__pyx_ptype_6pandas_5_libs_8interval_Interval->tp_dict, __pyx_n_s_typ_2, __pyx_n_u_interval) < 0) __PYX_ERR(0, 285, __pyx_L1_error)
  PyType_Modified(__pyx_ptype_6pandas_5_libs_8interval_Interval);

  /* "pandas/_libs/interval.pyx":491
 * @cython.wraparound(False)
 * @cython.boundscheck(False)
 * def intervals_to_interval_bounds(ndarray intervals,             # <<<<<<<<<<<<<<
 *                                  bint validate_closed=True):
 *     """
 */
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_6pandas_5_libs_8interval_1intervals_to_interval_bounds, NULL, __pyx_n_s_pandas__libs_interval); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 491, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_intervals_to_interval_bounds, __pyx_t_2) < 0) __PYX_ERR(0, 491, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pandas/_libs/intervaltree.pxi":7
 * """
 * 
 * from pandas._libs.algos import is_monotonic             # <<<<<<<<<<<<<<
 * 
 * ctypedef fused scalar_t:
 */
  __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_is_monotonic);
  __Pyx_GIVEREF(__pyx_n_s_is_monotonic);
  PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_is_monotonic);
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_pandas__libs_algos, __pyx_t_2, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_is_monotonic); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_is_monotonic, __pyx_t_2) < 0) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pandas/_libs/intervaltree.pxi":117
 *         return is_monotonic(sort_order, False)[0]
 * 
 *     def get_loc(self, scalar_t key):             # <<<<<<<<<<<<<<
 *         """Return all positions corresponding to intervals that overlap with
 *         the given scalar key
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(5); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 117, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __pyx_FusedFunction_NewEx(&__pyx_fuse_0__pyx_mdef_6pandas_5_libs_8interval_12IntervalTree_17get_loc, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_IntervalTree_get_loc, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__49)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 117, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_2, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_float64_t, __pyx_t_2) < 0) __PYX_ERR(1, 117, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __pyx_FusedFunction_NewEx(&__pyx_fuse_1__pyx_mdef_6pandas_5_libs_8interval_12IntervalTree_19get_loc, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_IntervalTree_get_loc, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__49)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 117, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_2, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_float32_t, __pyx_t_2) < 0) __PYX_ERR(1, 117, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __pyx_FusedFunction_NewEx(&__pyx_fuse_2__pyx_mdef_6pandas_5_libs_8interval_12IntervalTree_21get_loc, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_IntervalTree_get_loc, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__49)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 117, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_2, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_int64_t, __pyx_t_2) < 0) __PYX_ERR(1, 117, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __pyx_FusedFunction_NewEx(&__pyx_fuse_3__pyx_mdef_6pandas_5_libs_8interval_12IntervalTree_23get_loc, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_IntervalTree_get_loc, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__49)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 117, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_2, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_int32_t, __pyx_t_2) < 0) __PYX_ERR(1, 117, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __pyx_FusedFunction_NewEx(&__pyx_fuse_4__pyx_mdef_6pandas_5_libs_8interval_12IntervalTree_25get_loc, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_IntervalTree_get_loc, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__49)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 117, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_2, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_uint64_t, __pyx_t_2) < 0) __PYX_ERR(1, 117, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __pyx_FusedFunction_NewEx(&__pyx_mdef_6pandas_5_libs_8interval_12IntervalTree_3get_loc, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_IntervalTree_get_loc, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__49)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 117, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_2, __pyx_empty_tuple);
  ((__pyx_FusedFunctionObject *) __pyx_t_2)->__signatures__ = __pyx_t_1;
  __Pyx_GIVEREF(__pyx_t_1);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_6pandas_5_libs_8interval_IntervalTree->tp_dict, __pyx_n_s_get_loc, __pyx_t_2) < 0) __PYX_ERR(1, 117, __pyx_L1_error)
  PyType_Modified(__pyx_ptype_6pandas_5_libs_8interval_IntervalTree);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pandas/_libs/intervaltree.pxi":154
 *         return uniques.astype('intp')
 * 
 *     def get_indexer(self, scalar_t[:] target):             # <<<<<<<<<<<<<<
 *         """Return the positions corresponding to unique intervals that overlap
 *         with the given array of scalar targets.
 */
  __pyx_t_4 = __Pyx_PyDict_NewPresized(5); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 154, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __pyx_FusedFunction_NewEx(&__pyx_fuse_0__pyx_mdef_6pandas_5_libs_8interval_12IntervalTree_29get_indexer, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_IntervalTree_get_indexer, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__51)); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 154, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_5, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_float64_t, __pyx_t_5) < 0) __PYX_ERR(1, 154, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __pyx_FusedFunction_NewEx(&__pyx_fuse_1__pyx_mdef_6pandas_5_libs_8interval_12IntervalTree_31get_indexer, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_IntervalTree_get_indexer, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__51)); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 154, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_5, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_float32_t, __pyx_t_5) < 0) __PYX_ERR(1, 154, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __pyx_FusedFunction_NewEx(&__pyx_fuse_2__pyx_mdef_6pandas_5_libs_8interval_12IntervalTree_33get_indexer, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_IntervalTree_get_indexer, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__51)); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 154, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_5, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_int64_t, __pyx_t_5) < 0) __PYX_ERR(1, 154, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __pyx_FusedFunction_NewEx(&__pyx_fuse_3__pyx_mdef_6pandas_5_libs_8interval_12IntervalTree_35get_indexer, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_IntervalTree_get_indexer, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__51)); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 154, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_5, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_int32_t, __pyx_t_5) < 0) __PYX_ERR(1, 154, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __pyx_FusedFunction_NewEx(&__pyx_fuse_4__pyx_mdef_6pandas_5_libs_8interval_12IntervalTree_37get_indexer, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_IntervalTree_get_indexer, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__51)); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 154, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_5, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_uint64_t, __pyx_t_5) < 0) __PYX_ERR(1, 154, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __pyx_FusedFunction_NewEx(&__pyx_mdef_6pandas_5_libs_8interval_12IntervalTree_9get_indexer, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_IntervalTree_get_indexer, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__51)); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 154, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_5, __pyx_empty_tuple);
  ((__pyx_FusedFunctionObject *) __pyx_t_5)->__signatures__ = __pyx_t_4;
  __Pyx_GIVEREF(__pyx_t_4);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_6pandas_5_libs_8interval_IntervalTree->tp_dict, __pyx_n_s_get_indexer, __pyx_t_5) < 0) __PYX_ERR(1, 154, __pyx_L1_error)
  PyType_Modified(__pyx_ptype_6pandas_5_libs_8interval_IntervalTree);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pandas/_libs/intervaltree.pxi":177
 *         return result.to_array().astype('intp')
 * 
 *     def get_indexer_non_unique(self, scalar_t[:] target):             # <<<<<<<<<<<<<<
 *         """Return the positions corresponding to intervals that overlap with
 *         the given array of scalar targets. Non-unique positions are repeated.
 */
  __pyx_t_6 = __Pyx_PyDict_NewPresized(5); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 177, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = __pyx_FusedFunction_NewEx(&__pyx_fuse_0__pyx_mdef_6pandas_5_libs_8interval_12IntervalTree_41get_indexer_non_unique, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_IntervalTree_get_indexer_non_uni, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__53)); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 177, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_7, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_6, __pyx_n_s_float64_t, __pyx_t_7) < 0) __PYX_ERR(1, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __pyx_FusedFunction_NewEx(&__pyx_fuse_1__pyx_mdef_6pandas_5_libs_8interval_12IntervalTree_43get_indexer_non_unique, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_IntervalTree_get_indexer_non_uni, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__53)); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 177, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_7, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_6, __pyx_n_s_float32_t, __pyx_t_7) < 0) __PYX_ERR(1, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __pyx_FusedFunction_NewEx(&__pyx_fuse_2__pyx_mdef_6pandas_5_libs_8interval_12IntervalTree_45get_indexer_non_unique, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_IntervalTree_get_indexer_non_uni, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__53)); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 177, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_7, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_6, __pyx_n_s_int64_t, __pyx_t_7) < 0) __PYX_ERR(1, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __pyx_FusedFunction_NewEx(&__pyx_fuse_3__pyx_mdef_6pandas_5_libs_8interval_12IntervalTree_47get_indexer_non_unique, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_IntervalTree_get_indexer_non_uni, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__53)); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 177, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_7, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_6, __pyx_n_s_int32_t, __pyx_t_7) < 0) __PYX_ERR(1, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __pyx_FusedFunction_NewEx(&__pyx_fuse_4__pyx_mdef_6pandas_5_libs_8interval_12IntervalTree_49get_indexer_non_unique, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_IntervalTree_get_indexer_non_uni, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__53)); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 177, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_7, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_6, __pyx_n_s_uint64_t, __pyx_t_7) < 0) __PYX_ERR(1, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __pyx_FusedFunction_NewEx(&__pyx_mdef_6pandas_5_libs_8interval_12IntervalTree_11get_indexer_non_unique, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_IntervalTree_get_indexer_non_uni, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__53)); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 177, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_7, __pyx_empty_tuple);
  ((__pyx_FusedFunctionObject *) __pyx_t_7)->__signatures__ = __pyx_t_6;
  __Pyx_GIVEREF(__pyx_t_6);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_6pandas_5_libs_8interval_IntervalTree->tp_dict, __pyx_n_s_get_indexer_non_unique, __pyx_t_7) < 0) __PYX_ERR(1, 177, __pyx_L1_error)
  PyType_Modified(__pyx_ptype_6pandas_5_libs_8interval_IntervalTree);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pandas/_libs/intervaltree.pxi":231
 * # closed values
 * 
 * NODE_CLASSES = {}             # <<<<<<<<<<<<<<
 * 
 * cdef class Float32ClosedLeftIntervalNode:
 */
  __pyx_t_8 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 231, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_NODE_CLASSES, __pyx_t_8) < 0) __PYX_ERR(1, 231, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "pandas/_libs/intervaltree.pxi":334
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */
  __pyx_t_8 = __Pyx_PyDict_NewPresized(5); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 334, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __pyx_FusedFunction_NewEx(&__pyx_fuse_0__pyx_mdef_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_9__pyx_fuse_0query, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Float32ClosedLeftIntervalNode_2, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__55)); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 334, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_9, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_float64_t, __pyx_t_9) < 0) __PYX_ERR(1, 334, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __pyx_FusedFunction_NewEx(&__pyx_fuse_1__pyx_mdef_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_11__pyx_fuse_1query, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Float32ClosedLeftIntervalNode_3, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__55)); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 334, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_9, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_float32_t, __pyx_t_9) < 0) __PYX_ERR(1, 334, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __pyx_FusedFunction_NewEx(&__pyx_fuse_2__pyx_mdef_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_13__pyx_fuse_2query, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Float32ClosedLeftIntervalNode_4, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__55)); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 334, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_9, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_int64_t, __pyx_t_9) < 0) __PYX_ERR(1, 334, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __pyx_FusedFunction_NewEx(&__pyx_fuse_3__pyx_mdef_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_15__pyx_fuse_3query, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Float32ClosedLeftIntervalNode_5, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__55)); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 334, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_9, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_int32_t, __pyx_t_9) < 0) __PYX_ERR(1, 334, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __pyx_FusedFunction_NewEx(&__pyx_fuse_4__pyx_mdef_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_17__pyx_fuse_4query, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Float32ClosedLeftIntervalNode_6, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__55)); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 334, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_9, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_uint64_t, __pyx_t_9) < 0) __PYX_ERR(1, 334, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __pyx_FusedFunction_NewEx(&__pyx_mdef_6pandas_5_libs_8interval_29Float32ClosedLeftIntervalNode_3query, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Float32ClosedLeftIntervalNode_qu, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__55)); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 334, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_9, __pyx_empty_tuple);
  ((__pyx_FusedFunctionObject *) __pyx_t_9)->__signatures__ = __pyx_t_8;
  __Pyx_GIVEREF(__pyx_t_8);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode->tp_dict, __pyx_n_s_query, __pyx_t_9) < 0) __PYX_ERR(1, 334, __pyx_L1_error)
  PyType_Modified(__pyx_ptype_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "pandas/_libs/intervaltree.pxi":400
 *             return (m, (l, r))
 * 
 * NODE_CLASSES['float32',             # <<<<<<<<<<<<<<
 *              'left'] = Float32ClosedLeftIntervalNode
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_n_s_NODE_CLASSES); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 400, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  if (unlikely(PyObject_SetItem(__pyx_t_10, __pyx_tuple__56, ((PyObject *)__pyx_ptype_6pandas_5_libs_8interval_Float32ClosedLeftIntervalNode)) < 0)) __PYX_ERR(1, 400, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

  /* "pandas/_libs/intervaltree.pxi":504
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */
  __pyx_t_10 = __Pyx_PyDict_NewPresized(5); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 504, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_11 = __pyx_FusedFunction_NewEx(&__pyx_fuse_0__pyx_mdef_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_9__pyx_fuse_0query, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Float32ClosedRightIntervalNode_4, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__58)); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 504, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_11, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_10, __pyx_n_s_float64_t, __pyx_t_11) < 0) __PYX_ERR(1, 504, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  __pyx_t_11 = __pyx_FusedFunction_NewEx(&__pyx_fuse_1__pyx_mdef_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_11__pyx_fuse_1query, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Float32ClosedRightIntervalNode_5, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__58)); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 504, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_11, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_10, __pyx_n_s_float32_t, __pyx_t_11) < 0) __PYX_ERR(1, 504, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  __pyx_t_11 = __pyx_FusedFunction_NewEx(&__pyx_fuse_2__pyx_mdef_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_13__pyx_fuse_2query, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Float32ClosedRightIntervalNode_6, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__58)); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 504, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_11, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_10, __pyx_n_s_int64_t, __pyx_t_11) < 0) __PYX_ERR(1, 504, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  __pyx_t_11 = __pyx_FusedFunction_NewEx(&__pyx_fuse_3__pyx_mdef_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_15__pyx_fuse_3query, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Float32ClosedRightIntervalNode_7, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__58)); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 504, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_11, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_10, __pyx_n_s_int32_t, __pyx_t_11) < 0) __PYX_ERR(1, 504, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  __pyx_t_11 = __pyx_FusedFunction_NewEx(&__pyx_fuse_4__pyx_mdef_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_17__pyx_fuse_4query, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Float32ClosedRightIntervalNode_8, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__58)); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 504, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_11, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_10, __pyx_n_s_uint64_t, __pyx_t_11) < 0) __PYX_ERR(1, 504, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  __pyx_t_11 = __pyx_FusedFunction_NewEx(&__pyx_mdef_6pandas_5_libs_8interval_30Float32ClosedRightIntervalNode_3query, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Float32ClosedRightIntervalNode_q, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__58)); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 504, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_11, __pyx_empty_tuple);
  ((__pyx_FusedFunctionObject *) __pyx_t_11)->__signatures__ = __pyx_t_10;
  __Pyx_GIVEREF(__pyx_t_10);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode->tp_dict, __pyx_n_s_query, __pyx_t_11) < 0) __PYX_ERR(1, 504, __pyx_L1_error)
  PyType_Modified(__pyx_ptype_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode);
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "pandas/_libs/intervaltree.pxi":570
 *             return (m, (l, r))
 * 
 * NODE_CLASSES['float32',             # <<<<<<<<<<<<<<
 *              'right'] = Float32ClosedRightIntervalNode
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_12, __pyx_n_s_NODE_CLASSES); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 570, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  if (unlikely(PyObject_SetItem(__pyx_t_12, __pyx_tuple__59, ((PyObject *)__pyx_ptype_6pandas_5_libs_8interval_Float32ClosedRightIntervalNode)) < 0)) __PYX_ERR(1, 570, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;

  /* "pandas/_libs/intervaltree.pxi":674
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */
  __pyx_t_12 = __Pyx_PyDict_NewPresized(5); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 674, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __pyx_t_13 = __pyx_FusedFunction_NewEx(&__pyx_fuse_0__pyx_mdef_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_9__pyx_fuse_0query, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Float32ClosedBothIntervalNode_2, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__61)); if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 674, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_13, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_12, __pyx_n_s_float64_t, __pyx_t_13) < 0) __PYX_ERR(1, 674, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
  __pyx_t_13 = __pyx_FusedFunction_NewEx(&__pyx_fuse_1__pyx_mdef_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_11__pyx_fuse_1query, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Float32ClosedBothIntervalNode_3, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__61)); if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 674, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_13, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_12, __pyx_n_s_float32_t, __pyx_t_13) < 0) __PYX_ERR(1, 674, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
  __pyx_t_13 = __pyx_FusedFunction_NewEx(&__pyx_fuse_2__pyx_mdef_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_13__pyx_fuse_2query, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Float32ClosedBothIntervalNode_4, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__61)); if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 674, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_13, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_12, __pyx_n_s_int64_t, __pyx_t_13) < 0) __PYX_ERR(1, 674, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
  __pyx_t_13 = __pyx_FusedFunction_NewEx(&__pyx_fuse_3__pyx_mdef_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_15__pyx_fuse_3query, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Float32ClosedBothIntervalNode_5, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__61)); if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 674, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_13, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_12, __pyx_n_s_int32_t, __pyx_t_13) < 0) __PYX_ERR(1, 674, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
  __pyx_t_13 = __pyx_FusedFunction_NewEx(&__pyx_fuse_4__pyx_mdef_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_17__pyx_fuse_4query, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Float32ClosedBothIntervalNode_6, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__61)); if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 674, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_13, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_12, __pyx_n_s_uint64_t, __pyx_t_13) < 0) __PYX_ERR(1, 674, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
  __pyx_t_13 = __pyx_FusedFunction_NewEx(&__pyx_mdef_6pandas_5_libs_8interval_29Float32ClosedBothIntervalNode_3query, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Float32ClosedBothIntervalNode_qu, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__61)); if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 674, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_13, __pyx_empty_tuple);
  ((__pyx_FusedFunctionObject *) __pyx_t_13)->__signatures__ = __pyx_t_12;
  __Pyx_GIVEREF(__pyx_t_12);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode->tp_dict, __pyx_n_s_query, __pyx_t_13) < 0) __PYX_ERR(1, 674, __pyx_L1_error)
  PyType_Modified(__pyx_ptype_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode);
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

  /* "pandas/_libs/intervaltree.pxi":740
 *             return (m, (l, r))
 * 
 * NODE_CLASSES['float32',             # <<<<<<<<<<<<<<
 *              'both'] = Float32ClosedBothIntervalNode
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_14, __pyx_n_s_NODE_CLASSES); if (unlikely(!__pyx_t_14)) __PYX_ERR(1, 740, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  if (unlikely(PyObject_SetItem(__pyx_t_14, __pyx_tuple__62, ((PyObject *)__pyx_ptype_6pandas_5_libs_8interval_Float32ClosedBothIntervalNode)) < 0)) __PYX_ERR(1, 740, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;

  /* "pandas/_libs/intervaltree.pxi":844
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */
  __pyx_t_14 = __Pyx_PyDict_NewPresized(5); if (unlikely(!__pyx_t_14)) __PYX_ERR(1, 844, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __pyx_t_15 = __pyx_FusedFunction_NewEx(&__pyx_fuse_0__pyx_mdef_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_9__pyx_fuse_0query, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Float32ClosedNeitherIntervalNode_2, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__64)); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 844, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_15, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_14, __pyx_n_s_float64_t, __pyx_t_15) < 0) __PYX_ERR(1, 844, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  __pyx_t_15 = __pyx_FusedFunction_NewEx(&__pyx_fuse_1__pyx_mdef_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_11__pyx_fuse_1query, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Float32ClosedNeitherIntervalNode_3, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__64)); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 844, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_15, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_14, __pyx_n_s_float32_t, __pyx_t_15) < 0) __PYX_ERR(1, 844, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  __pyx_t_15 = __pyx_FusedFunction_NewEx(&__pyx_fuse_2__pyx_mdef_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_13__pyx_fuse_2query, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Float32ClosedNeitherIntervalNode_4, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__64)); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 844, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_15, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_14, __pyx_n_s_int64_t, __pyx_t_15) < 0) __PYX_ERR(1, 844, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  __pyx_t_15 = __pyx_FusedFunction_NewEx(&__pyx_fuse_3__pyx_mdef_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_15__pyx_fuse_3query, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Float32ClosedNeitherIntervalNode_5, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__64)); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 844, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_15, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_14, __pyx_n_s_int32_t, __pyx_t_15) < 0) __PYX_ERR(1, 844, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  __pyx_t_15 = __pyx_FusedFunction_NewEx(&__pyx_fuse_4__pyx_mdef_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_17__pyx_fuse_4query, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Float32ClosedNeitherIntervalNode_6, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__64)); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 844, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_15, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_14, __pyx_n_s_uint64_t, __pyx_t_15) < 0) __PYX_ERR(1, 844, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  __pyx_t_15 = __pyx_FusedFunction_NewEx(&__pyx_mdef_6pandas_5_libs_8interval_32Float32ClosedNeitherIntervalNode_3query, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Float32ClosedNeitherIntervalNode_7, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__64)); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 844, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_15, __pyx_empty_tuple);
  ((__pyx_FusedFunctionObject *) __pyx_t_15)->__signatures__ = __pyx_t_14;
  __Pyx_GIVEREF(__pyx_t_14);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode->tp_dict, __pyx_n_s_query, __pyx_t_15) < 0) __PYX_ERR(1, 844, __pyx_L1_error)
  PyType_Modified(__pyx_ptype_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode);
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

  /* "pandas/_libs/intervaltree.pxi":910
 *             return (m, (l, r))
 * 
 * NODE_CLASSES['float32',             # <<<<<<<<<<<<<<
 *              'neither'] = Float32ClosedNeitherIntervalNode
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_16, __pyx_n_s_NODE_CLASSES); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 910, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  if (unlikely(PyObject_SetItem(__pyx_t_16, __pyx_tuple__65, ((PyObject *)__pyx_ptype_6pandas_5_libs_8interval_Float32ClosedNeitherIntervalNode)) < 0)) __PYX_ERR(1, 910, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;

  /* "pandas/_libs/intervaltree.pxi":1014
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */
  __pyx_t_16 = __Pyx_PyDict_NewPresized(5); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 1014, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __pyx_t_17 = __pyx_FusedFunction_NewEx(&__pyx_fuse_0__pyx_mdef_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_9__pyx_fuse_0query, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Float64ClosedLeftIntervalNode_2, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__67)); if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 1014, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_17);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_17, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_16, __pyx_n_s_float64_t, __pyx_t_17) < 0) __PYX_ERR(1, 1014, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
  __pyx_t_17 = __pyx_FusedFunction_NewEx(&__pyx_fuse_1__pyx_mdef_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_11__pyx_fuse_1query, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Float64ClosedLeftIntervalNode_3, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__67)); if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 1014, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_17);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_17, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_16, __pyx_n_s_float32_t, __pyx_t_17) < 0) __PYX_ERR(1, 1014, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
  __pyx_t_17 = __pyx_FusedFunction_NewEx(&__pyx_fuse_2__pyx_mdef_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_13__pyx_fuse_2query, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Float64ClosedLeftIntervalNode_4, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__67)); if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 1014, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_17);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_17, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_16, __pyx_n_s_int64_t, __pyx_t_17) < 0) __PYX_ERR(1, 1014, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
  __pyx_t_17 = __pyx_FusedFunction_NewEx(&__pyx_fuse_3__pyx_mdef_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_15__pyx_fuse_3query, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Float64ClosedLeftIntervalNode_5, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__67)); if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 1014, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_17);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_17, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_16, __pyx_n_s_int32_t, __pyx_t_17) < 0) __PYX_ERR(1, 1014, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
  __pyx_t_17 = __pyx_FusedFunction_NewEx(&__pyx_fuse_4__pyx_mdef_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_17__pyx_fuse_4query, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Float64ClosedLeftIntervalNode_6, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__67)); if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 1014, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_17);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_17, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_16, __pyx_n_s_uint64_t, __pyx_t_17) < 0) __PYX_ERR(1, 1014, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
  __pyx_t_17 = __pyx_FusedFunction_NewEx(&__pyx_mdef_6pandas_5_libs_8interval_29Float64ClosedLeftIntervalNode_3query, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Float64ClosedLeftIntervalNode_qu, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__67)); if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 1014, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_17);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_17, __pyx_empty_tuple);
  ((__pyx_FusedFunctionObject *) __pyx_t_17)->__signatures__ = __pyx_t_16;
  __Pyx_GIVEREF(__pyx_t_16);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode->tp_dict, __pyx_n_s_query, __pyx_t_17) < 0) __PYX_ERR(1, 1014, __pyx_L1_error)
  PyType_Modified(__pyx_ptype_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode);
  __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;

  /* "pandas/_libs/intervaltree.pxi":1080
 *             return (m, (l, r))
 * 
 * NODE_CLASSES['float64',             # <<<<<<<<<<<<<<
 *              'left'] = Float64ClosedLeftIntervalNode
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_18, __pyx_n_s_NODE_CLASSES); if (unlikely(!__pyx_t_18)) __PYX_ERR(1, 1080, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  if (unlikely(PyObject_SetItem(__pyx_t_18, __pyx_tuple__68, ((PyObject *)__pyx_ptype_6pandas_5_libs_8interval_Float64ClosedLeftIntervalNode)) < 0)) __PYX_ERR(1, 1080, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;

  /* "pandas/_libs/intervaltree.pxi":1184
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */
  __pyx_t_18 = __Pyx_PyDict_NewPresized(5); if (unlikely(!__pyx_t_18)) __PYX_ERR(1, 1184, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  __pyx_t_19 = __pyx_FusedFunction_NewEx(&__pyx_fuse_0__pyx_mdef_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_9__pyx_fuse_0query, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Float64ClosedRightIntervalNode_4, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__70)); if (unlikely(!__pyx_t_19)) __PYX_ERR(1, 1184, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_19);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_19, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_18, __pyx_n_s_float64_t, __pyx_t_19) < 0) __PYX_ERR(1, 1184, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
  __pyx_t_19 = __pyx_FusedFunction_NewEx(&__pyx_fuse_1__pyx_mdef_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_11__pyx_fuse_1query, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Float64ClosedRightIntervalNode_5, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__70)); if (unlikely(!__pyx_t_19)) __PYX_ERR(1, 1184, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_19);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_19, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_18, __pyx_n_s_float32_t, __pyx_t_19) < 0) __PYX_ERR(1, 1184, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
  __pyx_t_19 = __pyx_FusedFunction_NewEx(&__pyx_fuse_2__pyx_mdef_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_13__pyx_fuse_2query, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Float64ClosedRightIntervalNode_6, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__70)); if (unlikely(!__pyx_t_19)) __PYX_ERR(1, 1184, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_19);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_19, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_18, __pyx_n_s_int64_t, __pyx_t_19) < 0) __PYX_ERR(1, 1184, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
  __pyx_t_19 = __pyx_FusedFunction_NewEx(&__pyx_fuse_3__pyx_mdef_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_15__pyx_fuse_3query, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Float64ClosedRightIntervalNode_7, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__70)); if (unlikely(!__pyx_t_19)) __PYX_ERR(1, 1184, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_19);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_19, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_18, __pyx_n_s_int32_t, __pyx_t_19) < 0) __PYX_ERR(1, 1184, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
  __pyx_t_19 = __pyx_FusedFunction_NewEx(&__pyx_fuse_4__pyx_mdef_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_17__pyx_fuse_4query, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Float64ClosedRightIntervalNode_8, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__70)); if (unlikely(!__pyx_t_19)) __PYX_ERR(1, 1184, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_19);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_19, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_18, __pyx_n_s_uint64_t, __pyx_t_19) < 0) __PYX_ERR(1, 1184, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
  __pyx_t_19 = __pyx_FusedFunction_NewEx(&__pyx_mdef_6pandas_5_libs_8interval_30Float64ClosedRightIntervalNode_3query, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Float64ClosedRightIntervalNode_q, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__70)); if (unlikely(!__pyx_t_19)) __PYX_ERR(1, 1184, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_19);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_19, __pyx_empty_tuple);
  ((__pyx_FusedFunctionObject *) __pyx_t_19)->__signatures__ = __pyx_t_18;
  __Pyx_GIVEREF(__pyx_t_18);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode->tp_dict, __pyx_n_s_query, __pyx_t_19) < 0) __PYX_ERR(1, 1184, __pyx_L1_error)
  PyType_Modified(__pyx_ptype_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode);
  __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;

  /* "pandas/_libs/intervaltree.pxi":1250
 *             return (m, (l, r))
 * 
 * NODE_CLASSES['float64',             # <<<<<<<<<<<<<<
 *              'right'] = Float64ClosedRightIntervalNode
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_20, __pyx_n_s_NODE_CLASSES); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 1250, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  if (unlikely(PyObject_SetItem(__pyx_t_20, __pyx_tuple__71, ((PyObject *)__pyx_ptype_6pandas_5_libs_8interval_Float64ClosedRightIntervalNode)) < 0)) __PYX_ERR(1, 1250, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

  /* "pandas/_libs/intervaltree.pxi":1354
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */
  __pyx_t_20 = __Pyx_PyDict_NewPresized(5); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 1354, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __pyx_t_21 = __pyx_FusedFunction_NewEx(&__pyx_fuse_0__pyx_mdef_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_9__pyx_fuse_0query, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Float64ClosedBothIntervalNode_2, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__73)); if (unlikely(!__pyx_t_21)) __PYX_ERR(1, 1354, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_21, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_20, __pyx_n_s_float64_t, __pyx_t_21) < 0) __PYX_ERR(1, 1354, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
  __pyx_t_21 = __pyx_FusedFunction_NewEx(&__pyx_fuse_1__pyx_mdef_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_11__pyx_fuse_1query, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Float64ClosedBothIntervalNode_3, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__73)); if (unlikely(!__pyx_t_21)) __PYX_ERR(1, 1354, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_21, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_20, __pyx_n_s_float32_t, __pyx_t_21) < 0) __PYX_ERR(1, 1354, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
  __pyx_t_21 = __pyx_FusedFunction_NewEx(&__pyx_fuse_2__pyx_mdef_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_13__pyx_fuse_2query, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Float64ClosedBothIntervalNode_4, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__73)); if (unlikely(!__pyx_t_21)) __PYX_ERR(1, 1354, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_21, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_20, __pyx_n_s_int64_t, __pyx_t_21) < 0) __PYX_ERR(1, 1354, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
  __pyx_t_21 = __pyx_FusedFunction_NewEx(&__pyx_fuse_3__pyx_mdef_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_15__pyx_fuse_3query, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Float64ClosedBothIntervalNode_5, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__73)); if (unlikely(!__pyx_t_21)) __PYX_ERR(1, 1354, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_21, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_20, __pyx_n_s_int32_t, __pyx_t_21) < 0) __PYX_ERR(1, 1354, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
  __pyx_t_21 = __pyx_FusedFunction_NewEx(&__pyx_fuse_4__pyx_mdef_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_17__pyx_fuse_4query, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Float64ClosedBothIntervalNode_6, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__73)); if (unlikely(!__pyx_t_21)) __PYX_ERR(1, 1354, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_21, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_20, __pyx_n_s_uint64_t, __pyx_t_21) < 0) __PYX_ERR(1, 1354, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
  __pyx_t_21 = __pyx_FusedFunction_NewEx(&__pyx_mdef_6pandas_5_libs_8interval_29Float64ClosedBothIntervalNode_3query, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Float64ClosedBothIntervalNode_qu, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__73)); if (unlikely(!__pyx_t_21)) __PYX_ERR(1, 1354, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_21, __pyx_empty_tuple);
  ((__pyx_FusedFunctionObject *) __pyx_t_21)->__signatures__ = __pyx_t_20;
  __Pyx_GIVEREF(__pyx_t_20);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode->tp_dict, __pyx_n_s_query, __pyx_t_21) < 0) __PYX_ERR(1, 1354, __pyx_L1_error)
  PyType_Modified(__pyx_ptype_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode);
  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;

  /* "pandas/_libs/intervaltree.pxi":1420
 *             return (m, (l, r))
 * 
 * NODE_CLASSES['float64',             # <<<<<<<<<<<<<<
 *              'both'] = Float64ClosedBothIntervalNode
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_22, __pyx_n_s_NODE_CLASSES); if (unlikely(!__pyx_t_22)) __PYX_ERR(1, 1420, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_22);
  if (unlikely(PyObject_SetItem(__pyx_t_22, __pyx_tuple__74, ((PyObject *)__pyx_ptype_6pandas_5_libs_8interval_Float64ClosedBothIntervalNode)) < 0)) __PYX_ERR(1, 1420, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;

  /* "pandas/_libs/intervaltree.pxi":1524
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */
  __pyx_t_22 = __Pyx_PyDict_NewPresized(5); if (unlikely(!__pyx_t_22)) __PYX_ERR(1, 1524, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_22);
  __pyx_t_23 = __pyx_FusedFunction_NewEx(&__pyx_fuse_0__pyx_mdef_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_9__pyx_fuse_0query, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Float64ClosedNeitherIntervalNode_2, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__76)); if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 1524, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_23);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_23, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_22, __pyx_n_s_float64_t, __pyx_t_23) < 0) __PYX_ERR(1, 1524, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
  __pyx_t_23 = __pyx_FusedFunction_NewEx(&__pyx_fuse_1__pyx_mdef_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_11__pyx_fuse_1query, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Float64ClosedNeitherIntervalNode_3, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__76)); if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 1524, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_23);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_23, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_22, __pyx_n_s_float32_t, __pyx_t_23) < 0) __PYX_ERR(1, 1524, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
  __pyx_t_23 = __pyx_FusedFunction_NewEx(&__pyx_fuse_2__pyx_mdef_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_13__pyx_fuse_2query, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Float64ClosedNeitherIntervalNode_4, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__76)); if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 1524, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_23);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_23, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_22, __pyx_n_s_int64_t, __pyx_t_23) < 0) __PYX_ERR(1, 1524, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
  __pyx_t_23 = __pyx_FusedFunction_NewEx(&__pyx_fuse_3__pyx_mdef_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_15__pyx_fuse_3query, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Float64ClosedNeitherIntervalNode_5, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__76)); if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 1524, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_23);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_23, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_22, __pyx_n_s_int32_t, __pyx_t_23) < 0) __PYX_ERR(1, 1524, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
  __pyx_t_23 = __pyx_FusedFunction_NewEx(&__pyx_fuse_4__pyx_mdef_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_17__pyx_fuse_4query, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Float64ClosedNeitherIntervalNode_6, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__76)); if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 1524, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_23);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_23, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_22, __pyx_n_s_uint64_t, __pyx_t_23) < 0) __PYX_ERR(1, 1524, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
  __pyx_t_23 = __pyx_FusedFunction_NewEx(&__pyx_mdef_6pandas_5_libs_8interval_32Float64ClosedNeitherIntervalNode_3query, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Float64ClosedNeitherIntervalNode_7, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__76)); if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 1524, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_23);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_23, __pyx_empty_tuple);
  ((__pyx_FusedFunctionObject *) __pyx_t_23)->__signatures__ = __pyx_t_22;
  __Pyx_GIVEREF(__pyx_t_22);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode->tp_dict, __pyx_n_s_query, __pyx_t_23) < 0) __PYX_ERR(1, 1524, __pyx_L1_error)
  PyType_Modified(__pyx_ptype_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode);
  __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;

  /* "pandas/_libs/intervaltree.pxi":1590
 *             return (m, (l, r))
 * 
 * NODE_CLASSES['float64',             # <<<<<<<<<<<<<<
 *              'neither'] = Float64ClosedNeitherIntervalNode
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_24, __pyx_n_s_NODE_CLASSES); if (unlikely(!__pyx_t_24)) __PYX_ERR(1, 1590, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_24);
  if (unlikely(PyObject_SetItem(__pyx_t_24, __pyx_tuple__77, ((PyObject *)__pyx_ptype_6pandas_5_libs_8interval_Float64ClosedNeitherIntervalNode)) < 0)) __PYX_ERR(1, 1590, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;

  /* "pandas/_libs/intervaltree.pxi":1694
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */
  __pyx_t_24 = __Pyx_PyDict_NewPresized(5); if (unlikely(!__pyx_t_24)) __PYX_ERR(1, 1694, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_24);
  __pyx_t_25 = __pyx_FusedFunction_NewEx(&__pyx_fuse_0__pyx_mdef_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_9__pyx_fuse_0query, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Int32ClosedLeftIntervalNode___py, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__79)); if (unlikely(!__pyx_t_25)) __PYX_ERR(1, 1694, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_25);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_25, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_24, __pyx_n_s_float64_t, __pyx_t_25) < 0) __PYX_ERR(1, 1694, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;
  __pyx_t_25 = __pyx_FusedFunction_NewEx(&__pyx_fuse_1__pyx_mdef_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_11__pyx_fuse_1query, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Int32ClosedLeftIntervalNode___py_2, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__79)); if (unlikely(!__pyx_t_25)) __PYX_ERR(1, 1694, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_25);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_25, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_24, __pyx_n_s_float32_t, __pyx_t_25) < 0) __PYX_ERR(1, 1694, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;
  __pyx_t_25 = __pyx_FusedFunction_NewEx(&__pyx_fuse_2__pyx_mdef_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_13__pyx_fuse_2query, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Int32ClosedLeftIntervalNode___py_3, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__79)); if (unlikely(!__pyx_t_25)) __PYX_ERR(1, 1694, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_25);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_25, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_24, __pyx_n_s_int64_t, __pyx_t_25) < 0) __PYX_ERR(1, 1694, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;
  __pyx_t_25 = __pyx_FusedFunction_NewEx(&__pyx_fuse_3__pyx_mdef_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_15__pyx_fuse_3query, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Int32ClosedLeftIntervalNode___py_4, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__79)); if (unlikely(!__pyx_t_25)) __PYX_ERR(1, 1694, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_25);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_25, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_24, __pyx_n_s_int32_t, __pyx_t_25) < 0) __PYX_ERR(1, 1694, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;
  __pyx_t_25 = __pyx_FusedFunction_NewEx(&__pyx_fuse_4__pyx_mdef_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_17__pyx_fuse_4query, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Int32ClosedLeftIntervalNode___py_5, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__79)); if (unlikely(!__pyx_t_25)) __PYX_ERR(1, 1694, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_25);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_25, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_24, __pyx_n_s_uint64_t, __pyx_t_25) < 0) __PYX_ERR(1, 1694, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;
  __pyx_t_25 = __pyx_FusedFunction_NewEx(&__pyx_mdef_6pandas_5_libs_8interval_27Int32ClosedLeftIntervalNode_3query, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Int32ClosedLeftIntervalNode_quer, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__79)); if (unlikely(!__pyx_t_25)) __PYX_ERR(1, 1694, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_25);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_25, __pyx_empty_tuple);
  ((__pyx_FusedFunctionObject *) __pyx_t_25)->__signatures__ = __pyx_t_24;
  __Pyx_GIVEREF(__pyx_t_24);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode->tp_dict, __pyx_n_s_query, __pyx_t_25) < 0) __PYX_ERR(1, 1694, __pyx_L1_error)
  PyType_Modified(__pyx_ptype_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode);
  __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;

  /* "pandas/_libs/intervaltree.pxi":1760
 *             return (m, (l, r))
 * 
 * NODE_CLASSES['int32',             # <<<<<<<<<<<<<<
 *              'left'] = Int32ClosedLeftIntervalNode
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_26, __pyx_n_s_NODE_CLASSES); if (unlikely(!__pyx_t_26)) __PYX_ERR(1, 1760, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_26);
  if (unlikely(PyObject_SetItem(__pyx_t_26, __pyx_tuple__80, ((PyObject *)__pyx_ptype_6pandas_5_libs_8interval_Int32ClosedLeftIntervalNode)) < 0)) __PYX_ERR(1, 1760, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;

  /* "pandas/_libs/intervaltree.pxi":1864
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */
  __pyx_t_26 = __Pyx_PyDict_NewPresized(5); if (unlikely(!__pyx_t_26)) __PYX_ERR(1, 1864, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_26);
  __pyx_t_27 = __pyx_FusedFunction_NewEx(&__pyx_fuse_0__pyx_mdef_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_9__pyx_fuse_0query, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Int32ClosedRightIntervalNode___p, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__82)); if (unlikely(!__pyx_t_27)) __PYX_ERR(1, 1864, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_27);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_27, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_26, __pyx_n_s_float64_t, __pyx_t_27) < 0) __PYX_ERR(1, 1864, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_27); __pyx_t_27 = 0;
  __pyx_t_27 = __pyx_FusedFunction_NewEx(&__pyx_fuse_1__pyx_mdef_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_11__pyx_fuse_1query, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Int32ClosedRightIntervalNode___p_2, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__82)); if (unlikely(!__pyx_t_27)) __PYX_ERR(1, 1864, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_27);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_27, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_26, __pyx_n_s_float32_t, __pyx_t_27) < 0) __PYX_ERR(1, 1864, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_27); __pyx_t_27 = 0;
  __pyx_t_27 = __pyx_FusedFunction_NewEx(&__pyx_fuse_2__pyx_mdef_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_13__pyx_fuse_2query, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Int32ClosedRightIntervalNode___p_3, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__82)); if (unlikely(!__pyx_t_27)) __PYX_ERR(1, 1864, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_27);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_27, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_26, __pyx_n_s_int64_t, __pyx_t_27) < 0) __PYX_ERR(1, 1864, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_27); __pyx_t_27 = 0;
  __pyx_t_27 = __pyx_FusedFunction_NewEx(&__pyx_fuse_3__pyx_mdef_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_15__pyx_fuse_3query, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Int32ClosedRightIntervalNode___p_4, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__82)); if (unlikely(!__pyx_t_27)) __PYX_ERR(1, 1864, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_27);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_27, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_26, __pyx_n_s_int32_t, __pyx_t_27) < 0) __PYX_ERR(1, 1864, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_27); __pyx_t_27 = 0;
  __pyx_t_27 = __pyx_FusedFunction_NewEx(&__pyx_fuse_4__pyx_mdef_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_17__pyx_fuse_4query, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Int32ClosedRightIntervalNode___p_5, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__82)); if (unlikely(!__pyx_t_27)) __PYX_ERR(1, 1864, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_27);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_27, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_26, __pyx_n_s_uint64_t, __pyx_t_27) < 0) __PYX_ERR(1, 1864, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_27); __pyx_t_27 = 0;
  __pyx_t_27 = __pyx_FusedFunction_NewEx(&__pyx_mdef_6pandas_5_libs_8interval_28Int32ClosedRightIntervalNode_3query, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Int32ClosedRightIntervalNode_que, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__82)); if (unlikely(!__pyx_t_27)) __PYX_ERR(1, 1864, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_27);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_27, __pyx_empty_tuple);
  ((__pyx_FusedFunctionObject *) __pyx_t_27)->__signatures__ = __pyx_t_26;
  __Pyx_GIVEREF(__pyx_t_26);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode->tp_dict, __pyx_n_s_query, __pyx_t_27) < 0) __PYX_ERR(1, 1864, __pyx_L1_error)
  PyType_Modified(__pyx_ptype_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode);
  __Pyx_DECREF(__pyx_t_27); __pyx_t_27 = 0;

  /* "pandas/_libs/intervaltree.pxi":1930
 *             return (m, (l, r))
 * 
 * NODE_CLASSES['int32',             # <<<<<<<<<<<<<<
 *              'right'] = Int32ClosedRightIntervalNode
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_28, __pyx_n_s_NODE_CLASSES); if (unlikely(!__pyx_t_28)) __PYX_ERR(1, 1930, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_28);
  if (unlikely(PyObject_SetItem(__pyx_t_28, __pyx_tuple__83, ((PyObject *)__pyx_ptype_6pandas_5_libs_8interval_Int32ClosedRightIntervalNode)) < 0)) __PYX_ERR(1, 1930, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;

  /* "pandas/_libs/intervaltree.pxi":2034
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */
  __pyx_t_28 = __Pyx_PyDict_NewPresized(5); if (unlikely(!__pyx_t_28)) __PYX_ERR(1, 2034, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_28);
  __pyx_t_29 = __pyx_FusedFunction_NewEx(&__pyx_fuse_0__pyx_mdef_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_9__pyx_fuse_0query, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Int32ClosedBothIntervalNode___py, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__85)); if (unlikely(!__pyx_t_29)) __PYX_ERR(1, 2034, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_29);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_29, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_28, __pyx_n_s_float64_t, __pyx_t_29) < 0) __PYX_ERR(1, 2034, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_29); __pyx_t_29 = 0;
  __pyx_t_29 = __pyx_FusedFunction_NewEx(&__pyx_fuse_1__pyx_mdef_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_11__pyx_fuse_1query, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Int32ClosedBothIntervalNode___py_2, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__85)); if (unlikely(!__pyx_t_29)) __PYX_ERR(1, 2034, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_29);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_29, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_28, __pyx_n_s_float32_t, __pyx_t_29) < 0) __PYX_ERR(1, 2034, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_29); __pyx_t_29 = 0;
  __pyx_t_29 = __pyx_FusedFunction_NewEx(&__pyx_fuse_2__pyx_mdef_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_13__pyx_fuse_2query, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Int32ClosedBothIntervalNode___py_3, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__85)); if (unlikely(!__pyx_t_29)) __PYX_ERR(1, 2034, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_29);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_29, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_28, __pyx_n_s_int64_t, __pyx_t_29) < 0) __PYX_ERR(1, 2034, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_29); __pyx_t_29 = 0;
  __pyx_t_29 = __pyx_FusedFunction_NewEx(&__pyx_fuse_3__pyx_mdef_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_15__pyx_fuse_3query, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Int32ClosedBothIntervalNode___py_4, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__85)); if (unlikely(!__pyx_t_29)) __PYX_ERR(1, 2034, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_29);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_29, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_28, __pyx_n_s_int32_t, __pyx_t_29) < 0) __PYX_ERR(1, 2034, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_29); __pyx_t_29 = 0;
  __pyx_t_29 = __pyx_FusedFunction_NewEx(&__pyx_fuse_4__pyx_mdef_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_17__pyx_fuse_4query, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Int32ClosedBothIntervalNode___py_5, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__85)); if (unlikely(!__pyx_t_29)) __PYX_ERR(1, 2034, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_29);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_29, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_28, __pyx_n_s_uint64_t, __pyx_t_29) < 0) __PYX_ERR(1, 2034, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_29); __pyx_t_29 = 0;
  __pyx_t_29 = __pyx_FusedFunction_NewEx(&__pyx_mdef_6pandas_5_libs_8interval_27Int32ClosedBothIntervalNode_3query, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Int32ClosedBothIntervalNode_quer, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__85)); if (unlikely(!__pyx_t_29)) __PYX_ERR(1, 2034, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_29);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_29, __pyx_empty_tuple);
  ((__pyx_FusedFunctionObject *) __pyx_t_29)->__signatures__ = __pyx_t_28;
  __Pyx_GIVEREF(__pyx_t_28);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode->tp_dict, __pyx_n_s_query, __pyx_t_29) < 0) __PYX_ERR(1, 2034, __pyx_L1_error)
  PyType_Modified(__pyx_ptype_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode);
  __Pyx_DECREF(__pyx_t_29); __pyx_t_29 = 0;

  /* "pandas/_libs/intervaltree.pxi":2100
 *             return (m, (l, r))
 * 
 * NODE_CLASSES['int32',             # <<<<<<<<<<<<<<
 *              'both'] = Int32ClosedBothIntervalNode
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_30, __pyx_n_s_NODE_CLASSES); if (unlikely(!__pyx_t_30)) __PYX_ERR(1, 2100, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_30);
  if (unlikely(PyObject_SetItem(__pyx_t_30, __pyx_tuple__86, ((PyObject *)__pyx_ptype_6pandas_5_libs_8interval_Int32ClosedBothIntervalNode)) < 0)) __PYX_ERR(1, 2100, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_30); __pyx_t_30 = 0;

  /* "pandas/_libs/intervaltree.pxi":2204
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */
  __pyx_t_30 = __Pyx_PyDict_NewPresized(5); if (unlikely(!__pyx_t_30)) __PYX_ERR(1, 2204, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_30);
  __pyx_t_31 = __pyx_FusedFunction_NewEx(&__pyx_fuse_0__pyx_mdef_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_9__pyx_fuse_0query, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Int32ClosedNeitherIntervalNode_4, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__88)); if (unlikely(!__pyx_t_31)) __PYX_ERR(1, 2204, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_31);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_31, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_30, __pyx_n_s_float64_t, __pyx_t_31) < 0) __PYX_ERR(1, 2204, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_31); __pyx_t_31 = 0;
  __pyx_t_31 = __pyx_FusedFunction_NewEx(&__pyx_fuse_1__pyx_mdef_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_11__pyx_fuse_1query, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Int32ClosedNeitherIntervalNode_5, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__88)); if (unlikely(!__pyx_t_31)) __PYX_ERR(1, 2204, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_31);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_31, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_30, __pyx_n_s_float32_t, __pyx_t_31) < 0) __PYX_ERR(1, 2204, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_31); __pyx_t_31 = 0;
  __pyx_t_31 = __pyx_FusedFunction_NewEx(&__pyx_fuse_2__pyx_mdef_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_13__pyx_fuse_2query, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Int32ClosedNeitherIntervalNode_6, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__88)); if (unlikely(!__pyx_t_31)) __PYX_ERR(1, 2204, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_31);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_31, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_30, __pyx_n_s_int64_t, __pyx_t_31) < 0) __PYX_ERR(1, 2204, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_31); __pyx_t_31 = 0;
  __pyx_t_31 = __pyx_FusedFunction_NewEx(&__pyx_fuse_3__pyx_mdef_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_15__pyx_fuse_3query, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Int32ClosedNeitherIntervalNode_7, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__88)); if (unlikely(!__pyx_t_31)) __PYX_ERR(1, 2204, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_31);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_31, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_30, __pyx_n_s_int32_t, __pyx_t_31) < 0) __PYX_ERR(1, 2204, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_31); __pyx_t_31 = 0;
  __pyx_t_31 = __pyx_FusedFunction_NewEx(&__pyx_fuse_4__pyx_mdef_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_17__pyx_fuse_4query, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Int32ClosedNeitherIntervalNode_8, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__88)); if (unlikely(!__pyx_t_31)) __PYX_ERR(1, 2204, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_31);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_31, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_30, __pyx_n_s_uint64_t, __pyx_t_31) < 0) __PYX_ERR(1, 2204, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_31); __pyx_t_31 = 0;
  __pyx_t_31 = __pyx_FusedFunction_NewEx(&__pyx_mdef_6pandas_5_libs_8interval_30Int32ClosedNeitherIntervalNode_3query, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Int32ClosedNeitherIntervalNode_q, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__88)); if (unlikely(!__pyx_t_31)) __PYX_ERR(1, 2204, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_31);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_31, __pyx_empty_tuple);
  ((__pyx_FusedFunctionObject *) __pyx_t_31)->__signatures__ = __pyx_t_30;
  __Pyx_GIVEREF(__pyx_t_30);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode->tp_dict, __pyx_n_s_query, __pyx_t_31) < 0) __PYX_ERR(1, 2204, __pyx_L1_error)
  PyType_Modified(__pyx_ptype_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode);
  __Pyx_DECREF(__pyx_t_31); __pyx_t_31 = 0;

  /* "pandas/_libs/intervaltree.pxi":2270
 *             return (m, (l, r))
 * 
 * NODE_CLASSES['int32',             # <<<<<<<<<<<<<<
 *              'neither'] = Int32ClosedNeitherIntervalNode
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_32, __pyx_n_s_NODE_CLASSES); if (unlikely(!__pyx_t_32)) __PYX_ERR(1, 2270, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  if (unlikely(PyObject_SetItem(__pyx_t_32, __pyx_tuple__89, ((PyObject *)__pyx_ptype_6pandas_5_libs_8interval_Int32ClosedNeitherIntervalNode)) < 0)) __PYX_ERR(1, 2270, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "pandas/_libs/intervaltree.pxi":2374
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */
  __pyx_t_32 = __Pyx_PyDict_NewPresized(5); if (unlikely(!__pyx_t_32)) __PYX_ERR(1, 2374, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  __pyx_t_33 = __pyx_FusedFunction_NewEx(&__pyx_fuse_0__pyx_mdef_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_9__pyx_fuse_0query, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Int64ClosedLeftIntervalNode___py, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__91)); if (unlikely(!__pyx_t_33)) __PYX_ERR(1, 2374, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_33);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_33, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_32, __pyx_n_s_float64_t, __pyx_t_33) < 0) __PYX_ERR(1, 2374, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_33); __pyx_t_33 = 0;
  __pyx_t_33 = __pyx_FusedFunction_NewEx(&__pyx_fuse_1__pyx_mdef_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_11__pyx_fuse_1query, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Int64ClosedLeftIntervalNode___py_2, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__91)); if (unlikely(!__pyx_t_33)) __PYX_ERR(1, 2374, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_33);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_33, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_32, __pyx_n_s_float32_t, __pyx_t_33) < 0) __PYX_ERR(1, 2374, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_33); __pyx_t_33 = 0;
  __pyx_t_33 = __pyx_FusedFunction_NewEx(&__pyx_fuse_2__pyx_mdef_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_13__pyx_fuse_2query, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Int64ClosedLeftIntervalNode___py_3, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__91)); if (unlikely(!__pyx_t_33)) __PYX_ERR(1, 2374, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_33);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_33, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_32, __pyx_n_s_int64_t, __pyx_t_33) < 0) __PYX_ERR(1, 2374, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_33); __pyx_t_33 = 0;
  __pyx_t_33 = __pyx_FusedFunction_NewEx(&__pyx_fuse_3__pyx_mdef_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_15__pyx_fuse_3query, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Int64ClosedLeftIntervalNode___py_4, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__91)); if (unlikely(!__pyx_t_33)) __PYX_ERR(1, 2374, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_33);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_33, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_32, __pyx_n_s_int32_t, __pyx_t_33) < 0) __PYX_ERR(1, 2374, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_33); __pyx_t_33 = 0;
  __pyx_t_33 = __pyx_FusedFunction_NewEx(&__pyx_fuse_4__pyx_mdef_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_17__pyx_fuse_4query, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Int64ClosedLeftIntervalNode___py_5, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__91)); if (unlikely(!__pyx_t_33)) __PYX_ERR(1, 2374, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_33);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_33, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_32, __pyx_n_s_uint64_t, __pyx_t_33) < 0) __PYX_ERR(1, 2374, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_33); __pyx_t_33 = 0;
  __pyx_t_33 = __pyx_FusedFunction_NewEx(&__pyx_mdef_6pandas_5_libs_8interval_27Int64ClosedLeftIntervalNode_3query, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Int64ClosedLeftIntervalNode_quer, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__91)); if (unlikely(!__pyx_t_33)) __PYX_ERR(1, 2374, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_33);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_33, __pyx_empty_tuple);
  ((__pyx_FusedFunctionObject *) __pyx_t_33)->__signatures__ = __pyx_t_32;
  __Pyx_GIVEREF(__pyx_t_32);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode->tp_dict, __pyx_n_s_query, __pyx_t_33) < 0) __PYX_ERR(1, 2374, __pyx_L1_error)
  PyType_Modified(__pyx_ptype_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode);
  __Pyx_DECREF(__pyx_t_33); __pyx_t_33 = 0;

  /* "pandas/_libs/intervaltree.pxi":2440
 *             return (m, (l, r))
 * 
 * NODE_CLASSES['int64',             # <<<<<<<<<<<<<<
 *              'left'] = Int64ClosedLeftIntervalNode
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_34, __pyx_n_s_NODE_CLASSES); if (unlikely(!__pyx_t_34)) __PYX_ERR(1, 2440, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_34);
  if (unlikely(PyObject_SetItem(__pyx_t_34, __pyx_tuple__92, ((PyObject *)__pyx_ptype_6pandas_5_libs_8interval_Int64ClosedLeftIntervalNode)) < 0)) __PYX_ERR(1, 2440, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_34); __pyx_t_34 = 0;

  /* "pandas/_libs/intervaltree.pxi":2544
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */
  __pyx_t_34 = __Pyx_PyDict_NewPresized(5); if (unlikely(!__pyx_t_34)) __PYX_ERR(1, 2544, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_34);
  __pyx_t_35 = __pyx_FusedFunction_NewEx(&__pyx_fuse_0__pyx_mdef_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_9__pyx_fuse_0query, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Int64ClosedRightIntervalNode___p, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__94)); if (unlikely(!__pyx_t_35)) __PYX_ERR(1, 2544, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_35);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_35, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_34, __pyx_n_s_float64_t, __pyx_t_35) < 0) __PYX_ERR(1, 2544, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_35); __pyx_t_35 = 0;
  __pyx_t_35 = __pyx_FusedFunction_NewEx(&__pyx_fuse_1__pyx_mdef_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_11__pyx_fuse_1query, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Int64ClosedRightIntervalNode___p_2, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__94)); if (unlikely(!__pyx_t_35)) __PYX_ERR(1, 2544, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_35);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_35, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_34, __pyx_n_s_float32_t, __pyx_t_35) < 0) __PYX_ERR(1, 2544, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_35); __pyx_t_35 = 0;
  __pyx_t_35 = __pyx_FusedFunction_NewEx(&__pyx_fuse_2__pyx_mdef_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_13__pyx_fuse_2query, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Int64ClosedRightIntervalNode___p_3, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__94)); if (unlikely(!__pyx_t_35)) __PYX_ERR(1, 2544, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_35);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_35, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_34, __pyx_n_s_int64_t, __pyx_t_35) < 0) __PYX_ERR(1, 2544, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_35); __pyx_t_35 = 0;
  __pyx_t_35 = __pyx_FusedFunction_NewEx(&__pyx_fuse_3__pyx_mdef_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_15__pyx_fuse_3query, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Int64ClosedRightIntervalNode___p_4, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__94)); if (unlikely(!__pyx_t_35)) __PYX_ERR(1, 2544, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_35);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_35, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_34, __pyx_n_s_int32_t, __pyx_t_35) < 0) __PYX_ERR(1, 2544, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_35); __pyx_t_35 = 0;
  __pyx_t_35 = __pyx_FusedFunction_NewEx(&__pyx_fuse_4__pyx_mdef_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_17__pyx_fuse_4query, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Int64ClosedRightIntervalNode___p_5, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__94)); if (unlikely(!__pyx_t_35)) __PYX_ERR(1, 2544, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_35);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_35, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_34, __pyx_n_s_uint64_t, __pyx_t_35) < 0) __PYX_ERR(1, 2544, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_35); __pyx_t_35 = 0;
  __pyx_t_35 = __pyx_FusedFunction_NewEx(&__pyx_mdef_6pandas_5_libs_8interval_28Int64ClosedRightIntervalNode_3query, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Int64ClosedRightIntervalNode_que, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__94)); if (unlikely(!__pyx_t_35)) __PYX_ERR(1, 2544, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_35);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_35, __pyx_empty_tuple);
  ((__pyx_FusedFunctionObject *) __pyx_t_35)->__signatures__ = __pyx_t_34;
  __Pyx_GIVEREF(__pyx_t_34);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode->tp_dict, __pyx_n_s_query, __pyx_t_35) < 0) __PYX_ERR(1, 2544, __pyx_L1_error)
  PyType_Modified(__pyx_ptype_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode);
  __Pyx_DECREF(__pyx_t_35); __pyx_t_35 = 0;

  /* "pandas/_libs/intervaltree.pxi":2610
 *             return (m, (l, r))
 * 
 * NODE_CLASSES['int64',             # <<<<<<<<<<<<<<
 *              'right'] = Int64ClosedRightIntervalNode
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_36, __pyx_n_s_NODE_CLASSES); if (unlikely(!__pyx_t_36)) __PYX_ERR(1, 2610, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_36);
  if (unlikely(PyObject_SetItem(__pyx_t_36, __pyx_tuple__95, ((PyObject *)__pyx_ptype_6pandas_5_libs_8interval_Int64ClosedRightIntervalNode)) < 0)) __PYX_ERR(1, 2610, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_36); __pyx_t_36 = 0;

  /* "pandas/_libs/intervaltree.pxi":2714
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */
  __pyx_t_36 = __Pyx_PyDict_NewPresized(5); if (unlikely(!__pyx_t_36)) __PYX_ERR(1, 2714, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_36);
  __pyx_t_37 = __pyx_FusedFunction_NewEx(&__pyx_fuse_0__pyx_mdef_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_9__pyx_fuse_0query, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Int64ClosedBothIntervalNode___py, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__97)); if (unlikely(!__pyx_t_37)) __PYX_ERR(1, 2714, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_37);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_37, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_36, __pyx_n_s_float64_t, __pyx_t_37) < 0) __PYX_ERR(1, 2714, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_37); __pyx_t_37 = 0;
  __pyx_t_37 = __pyx_FusedFunction_NewEx(&__pyx_fuse_1__pyx_mdef_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_11__pyx_fuse_1query, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Int64ClosedBothIntervalNode___py_2, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__97)); if (unlikely(!__pyx_t_37)) __PYX_ERR(1, 2714, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_37);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_37, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_36, __pyx_n_s_float32_t, __pyx_t_37) < 0) __PYX_ERR(1, 2714, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_37); __pyx_t_37 = 0;
  __pyx_t_37 = __pyx_FusedFunction_NewEx(&__pyx_fuse_2__pyx_mdef_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_13__pyx_fuse_2query, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Int64ClosedBothIntervalNode___py_3, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__97)); if (unlikely(!__pyx_t_37)) __PYX_ERR(1, 2714, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_37);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_37, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_36, __pyx_n_s_int64_t, __pyx_t_37) < 0) __PYX_ERR(1, 2714, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_37); __pyx_t_37 = 0;
  __pyx_t_37 = __pyx_FusedFunction_NewEx(&__pyx_fuse_3__pyx_mdef_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_15__pyx_fuse_3query, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Int64ClosedBothIntervalNode___py_4, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__97)); if (unlikely(!__pyx_t_37)) __PYX_ERR(1, 2714, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_37);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_37, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_36, __pyx_n_s_int32_t, __pyx_t_37) < 0) __PYX_ERR(1, 2714, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_37); __pyx_t_37 = 0;
  __pyx_t_37 = __pyx_FusedFunction_NewEx(&__pyx_fuse_4__pyx_mdef_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_17__pyx_fuse_4query, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Int64ClosedBothIntervalNode___py_5, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__97)); if (unlikely(!__pyx_t_37)) __PYX_ERR(1, 2714, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_37);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_37, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_36, __pyx_n_s_uint64_t, __pyx_t_37) < 0) __PYX_ERR(1, 2714, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_37); __pyx_t_37 = 0;
  __pyx_t_37 = __pyx_FusedFunction_NewEx(&__pyx_mdef_6pandas_5_libs_8interval_27Int64ClosedBothIntervalNode_3query, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Int64ClosedBothIntervalNode_quer, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__97)); if (unlikely(!__pyx_t_37)) __PYX_ERR(1, 2714, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_37);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_37, __pyx_empty_tuple);
  ((__pyx_FusedFunctionObject *) __pyx_t_37)->__signatures__ = __pyx_t_36;
  __Pyx_GIVEREF(__pyx_t_36);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode->tp_dict, __pyx_n_s_query, __pyx_t_37) < 0) __PYX_ERR(1, 2714, __pyx_L1_error)
  PyType_Modified(__pyx_ptype_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode);
  __Pyx_DECREF(__pyx_t_37); __pyx_t_37 = 0;

  /* "pandas/_libs/intervaltree.pxi":2780
 *             return (m, (l, r))
 * 
 * NODE_CLASSES['int64',             # <<<<<<<<<<<<<<
 *              'both'] = Int64ClosedBothIntervalNode
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_38, __pyx_n_s_NODE_CLASSES); if (unlikely(!__pyx_t_38)) __PYX_ERR(1, 2780, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_38);
  if (unlikely(PyObject_SetItem(__pyx_t_38, __pyx_tuple__98, ((PyObject *)__pyx_ptype_6pandas_5_libs_8interval_Int64ClosedBothIntervalNode)) < 0)) __PYX_ERR(1, 2780, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_38); __pyx_t_38 = 0;

  /* "pandas/_libs/intervaltree.pxi":2884
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */
  __pyx_t_38 = __Pyx_PyDict_NewPresized(5); if (unlikely(!__pyx_t_38)) __PYX_ERR(1, 2884, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_38);
  __pyx_t_39 = __pyx_FusedFunction_NewEx(&__pyx_fuse_0__pyx_mdef_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_9__pyx_fuse_0query, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Int64ClosedNeitherIntervalNode_4, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__100)); if (unlikely(!__pyx_t_39)) __PYX_ERR(1, 2884, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_39);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_39, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_38, __pyx_n_s_float64_t, __pyx_t_39) < 0) __PYX_ERR(1, 2884, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_39); __pyx_t_39 = 0;
  __pyx_t_39 = __pyx_FusedFunction_NewEx(&__pyx_fuse_1__pyx_mdef_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_11__pyx_fuse_1query, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Int64ClosedNeitherIntervalNode_5, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__100)); if (unlikely(!__pyx_t_39)) __PYX_ERR(1, 2884, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_39);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_39, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_38, __pyx_n_s_float32_t, __pyx_t_39) < 0) __PYX_ERR(1, 2884, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_39); __pyx_t_39 = 0;
  __pyx_t_39 = __pyx_FusedFunction_NewEx(&__pyx_fuse_2__pyx_mdef_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_13__pyx_fuse_2query, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Int64ClosedNeitherIntervalNode_6, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__100)); if (unlikely(!__pyx_t_39)) __PYX_ERR(1, 2884, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_39);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_39, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_38, __pyx_n_s_int64_t, __pyx_t_39) < 0) __PYX_ERR(1, 2884, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_39); __pyx_t_39 = 0;
  __pyx_t_39 = __pyx_FusedFunction_NewEx(&__pyx_fuse_3__pyx_mdef_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_15__pyx_fuse_3query, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Int64ClosedNeitherIntervalNode_7, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__100)); if (unlikely(!__pyx_t_39)) __PYX_ERR(1, 2884, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_39);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_39, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_38, __pyx_n_s_int32_t, __pyx_t_39) < 0) __PYX_ERR(1, 2884, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_39); __pyx_t_39 = 0;
  __pyx_t_39 = __pyx_FusedFunction_NewEx(&__pyx_fuse_4__pyx_mdef_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_17__pyx_fuse_4query, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Int64ClosedNeitherIntervalNode_8, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__100)); if (unlikely(!__pyx_t_39)) __PYX_ERR(1, 2884, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_39);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_39, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_38, __pyx_n_s_uint64_t, __pyx_t_39) < 0) __PYX_ERR(1, 2884, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_39); __pyx_t_39 = 0;
  __pyx_t_39 = __pyx_FusedFunction_NewEx(&__pyx_mdef_6pandas_5_libs_8interval_30Int64ClosedNeitherIntervalNode_3query, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Int64ClosedNeitherIntervalNode_q, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__100)); if (unlikely(!__pyx_t_39)) __PYX_ERR(1, 2884, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_39);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_39, __pyx_empty_tuple);
  ((__pyx_FusedFunctionObject *) __pyx_t_39)->__signatures__ = __pyx_t_38;
  __Pyx_GIVEREF(__pyx_t_38);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode->tp_dict, __pyx_n_s_query, __pyx_t_39) < 0) __PYX_ERR(1, 2884, __pyx_L1_error)
  PyType_Modified(__pyx_ptype_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode);
  __Pyx_DECREF(__pyx_t_39); __pyx_t_39 = 0;

  /* "pandas/_libs/intervaltree.pxi":2950
 *             return (m, (l, r))
 * 
 * NODE_CLASSES['int64',             # <<<<<<<<<<<<<<
 *              'neither'] = Int64ClosedNeitherIntervalNode
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_40, __pyx_n_s_NODE_CLASSES); if (unlikely(!__pyx_t_40)) __PYX_ERR(1, 2950, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_40);
  if (unlikely(PyObject_SetItem(__pyx_t_40, __pyx_tuple__101, ((PyObject *)__pyx_ptype_6pandas_5_libs_8interval_Int64ClosedNeitherIntervalNode)) < 0)) __PYX_ERR(1, 2950, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_40); __pyx_t_40 = 0;

  /* "pandas/_libs/intervaltree.pxi":3054
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */
  __pyx_t_40 = __Pyx_PyDict_NewPresized(5); if (unlikely(!__pyx_t_40)) __PYX_ERR(1, 3054, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_40);
  __pyx_t_41 = __pyx_FusedFunction_NewEx(&__pyx_fuse_0__pyx_mdef_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_9__pyx_fuse_0query, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Uint64ClosedLeftIntervalNode___p, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__103)); if (unlikely(!__pyx_t_41)) __PYX_ERR(1, 3054, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_41);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_41, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_40, __pyx_n_s_float64_t, __pyx_t_41) < 0) __PYX_ERR(1, 3054, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_41); __pyx_t_41 = 0;
  __pyx_t_41 = __pyx_FusedFunction_NewEx(&__pyx_fuse_1__pyx_mdef_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_11__pyx_fuse_1query, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Uint64ClosedLeftIntervalNode___p_2, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__103)); if (unlikely(!__pyx_t_41)) __PYX_ERR(1, 3054, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_41);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_41, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_40, __pyx_n_s_float32_t, __pyx_t_41) < 0) __PYX_ERR(1, 3054, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_41); __pyx_t_41 = 0;
  __pyx_t_41 = __pyx_FusedFunction_NewEx(&__pyx_fuse_2__pyx_mdef_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_13__pyx_fuse_2query, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Uint64ClosedLeftIntervalNode___p_3, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__103)); if (unlikely(!__pyx_t_41)) __PYX_ERR(1, 3054, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_41);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_41, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_40, __pyx_n_s_int64_t, __pyx_t_41) < 0) __PYX_ERR(1, 3054, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_41); __pyx_t_41 = 0;
  __pyx_t_41 = __pyx_FusedFunction_NewEx(&__pyx_fuse_3__pyx_mdef_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_15__pyx_fuse_3query, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Uint64ClosedLeftIntervalNode___p_4, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__103)); if (unlikely(!__pyx_t_41)) __PYX_ERR(1, 3054, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_41);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_41, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_40, __pyx_n_s_int32_t, __pyx_t_41) < 0) __PYX_ERR(1, 3054, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_41); __pyx_t_41 = 0;
  __pyx_t_41 = __pyx_FusedFunction_NewEx(&__pyx_fuse_4__pyx_mdef_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_17__pyx_fuse_4query, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Uint64ClosedLeftIntervalNode___p_5, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__103)); if (unlikely(!__pyx_t_41)) __PYX_ERR(1, 3054, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_41);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_41, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_40, __pyx_n_s_uint64_t, __pyx_t_41) < 0) __PYX_ERR(1, 3054, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_41); __pyx_t_41 = 0;
  __pyx_t_41 = __pyx_FusedFunction_NewEx(&__pyx_mdef_6pandas_5_libs_8interval_28Uint64ClosedLeftIntervalNode_3query, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Uint64ClosedLeftIntervalNode_que, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__103)); if (unlikely(!__pyx_t_41)) __PYX_ERR(1, 3054, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_41);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_41, __pyx_empty_tuple);
  ((__pyx_FusedFunctionObject *) __pyx_t_41)->__signatures__ = __pyx_t_40;
  __Pyx_GIVEREF(__pyx_t_40);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode->tp_dict, __pyx_n_s_query, __pyx_t_41) < 0) __PYX_ERR(1, 3054, __pyx_L1_error)
  PyType_Modified(__pyx_ptype_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode);
  __Pyx_DECREF(__pyx_t_41); __pyx_t_41 = 0;

  /* "pandas/_libs/intervaltree.pxi":3120
 *             return (m, (l, r))
 * 
 * NODE_CLASSES['uint64',             # <<<<<<<<<<<<<<
 *              'left'] = Uint64ClosedLeftIntervalNode
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_42, __pyx_n_s_NODE_CLASSES); if (unlikely(!__pyx_t_42)) __PYX_ERR(1, 3120, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_42);
  if (unlikely(PyObject_SetItem(__pyx_t_42, __pyx_tuple__104, ((PyObject *)__pyx_ptype_6pandas_5_libs_8interval_Uint64ClosedLeftIntervalNode)) < 0)) __PYX_ERR(1, 3120, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_42); __pyx_t_42 = 0;

  /* "pandas/_libs/intervaltree.pxi":3224
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */
  __pyx_t_42 = __Pyx_PyDict_NewPresized(5); if (unlikely(!__pyx_t_42)) __PYX_ERR(1, 3224, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_42);
  __pyx_t_43 = __pyx_FusedFunction_NewEx(&__pyx_fuse_0__pyx_mdef_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_9__pyx_fuse_0query, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Uint64ClosedRightIntervalNode_2, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__106)); if (unlikely(!__pyx_t_43)) __PYX_ERR(1, 3224, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_43);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_43, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_42, __pyx_n_s_float64_t, __pyx_t_43) < 0) __PYX_ERR(1, 3224, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_43); __pyx_t_43 = 0;
  __pyx_t_43 = __pyx_FusedFunction_NewEx(&__pyx_fuse_1__pyx_mdef_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_11__pyx_fuse_1query, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Uint64ClosedRightIntervalNode_3, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__106)); if (unlikely(!__pyx_t_43)) __PYX_ERR(1, 3224, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_43);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_43, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_42, __pyx_n_s_float32_t, __pyx_t_43) < 0) __PYX_ERR(1, 3224, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_43); __pyx_t_43 = 0;
  __pyx_t_43 = __pyx_FusedFunction_NewEx(&__pyx_fuse_2__pyx_mdef_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_13__pyx_fuse_2query, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Uint64ClosedRightIntervalNode_4, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__106)); if (unlikely(!__pyx_t_43)) __PYX_ERR(1, 3224, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_43);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_43, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_42, __pyx_n_s_int64_t, __pyx_t_43) < 0) __PYX_ERR(1, 3224, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_43); __pyx_t_43 = 0;
  __pyx_t_43 = __pyx_FusedFunction_NewEx(&__pyx_fuse_3__pyx_mdef_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_15__pyx_fuse_3query, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Uint64ClosedRightIntervalNode_5, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__106)); if (unlikely(!__pyx_t_43)) __PYX_ERR(1, 3224, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_43);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_43, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_42, __pyx_n_s_int32_t, __pyx_t_43) < 0) __PYX_ERR(1, 3224, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_43); __pyx_t_43 = 0;
  __pyx_t_43 = __pyx_FusedFunction_NewEx(&__pyx_fuse_4__pyx_mdef_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_17__pyx_fuse_4query, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Uint64ClosedRightIntervalNode_6, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__106)); if (unlikely(!__pyx_t_43)) __PYX_ERR(1, 3224, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_43);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_43, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_42, __pyx_n_s_uint64_t, __pyx_t_43) < 0) __PYX_ERR(1, 3224, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_43); __pyx_t_43 = 0;
  __pyx_t_43 = __pyx_FusedFunction_NewEx(&__pyx_mdef_6pandas_5_libs_8interval_29Uint64ClosedRightIntervalNode_3query, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Uint64ClosedRightIntervalNode_qu, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__106)); if (unlikely(!__pyx_t_43)) __PYX_ERR(1, 3224, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_43);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_43, __pyx_empty_tuple);
  ((__pyx_FusedFunctionObject *) __pyx_t_43)->__signatures__ = __pyx_t_42;
  __Pyx_GIVEREF(__pyx_t_42);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode->tp_dict, __pyx_n_s_query, __pyx_t_43) < 0) __PYX_ERR(1, 3224, __pyx_L1_error)
  PyType_Modified(__pyx_ptype_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode);
  __Pyx_DECREF(__pyx_t_43); __pyx_t_43 = 0;

  /* "pandas/_libs/intervaltree.pxi":3290
 *             return (m, (l, r))
 * 
 * NODE_CLASSES['uint64',             # <<<<<<<<<<<<<<
 *              'right'] = Uint64ClosedRightIntervalNode
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_44, __pyx_n_s_NODE_CLASSES); if (unlikely(!__pyx_t_44)) __PYX_ERR(1, 3290, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_44);
  if (unlikely(PyObject_SetItem(__pyx_t_44, __pyx_tuple__107, ((PyObject *)__pyx_ptype_6pandas_5_libs_8interval_Uint64ClosedRightIntervalNode)) < 0)) __PYX_ERR(1, 3290, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_44); __pyx_t_44 = 0;

  /* "pandas/_libs/intervaltree.pxi":3394
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */
  __pyx_t_44 = __Pyx_PyDict_NewPresized(5); if (unlikely(!__pyx_t_44)) __PYX_ERR(1, 3394, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_44);
  __pyx_t_45 = __pyx_FusedFunction_NewEx(&__pyx_fuse_0__pyx_mdef_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_9__pyx_fuse_0query, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Uint64ClosedBothIntervalNode___p, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__109)); if (unlikely(!__pyx_t_45)) __PYX_ERR(1, 3394, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_45);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_45, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_44, __pyx_n_s_float64_t, __pyx_t_45) < 0) __PYX_ERR(1, 3394, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_45); __pyx_t_45 = 0;
  __pyx_t_45 = __pyx_FusedFunction_NewEx(&__pyx_fuse_1__pyx_mdef_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_11__pyx_fuse_1query, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Uint64ClosedBothIntervalNode___p_2, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__109)); if (unlikely(!__pyx_t_45)) __PYX_ERR(1, 3394, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_45);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_45, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_44, __pyx_n_s_float32_t, __pyx_t_45) < 0) __PYX_ERR(1, 3394, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_45); __pyx_t_45 = 0;
  __pyx_t_45 = __pyx_FusedFunction_NewEx(&__pyx_fuse_2__pyx_mdef_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_13__pyx_fuse_2query, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Uint64ClosedBothIntervalNode___p_3, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__109)); if (unlikely(!__pyx_t_45)) __PYX_ERR(1, 3394, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_45);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_45, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_44, __pyx_n_s_int64_t, __pyx_t_45) < 0) __PYX_ERR(1, 3394, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_45); __pyx_t_45 = 0;
  __pyx_t_45 = __pyx_FusedFunction_NewEx(&__pyx_fuse_3__pyx_mdef_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_15__pyx_fuse_3query, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Uint64ClosedBothIntervalNode___p_4, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__109)); if (unlikely(!__pyx_t_45)) __PYX_ERR(1, 3394, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_45);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_45, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_44, __pyx_n_s_int32_t, __pyx_t_45) < 0) __PYX_ERR(1, 3394, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_45); __pyx_t_45 = 0;
  __pyx_t_45 = __pyx_FusedFunction_NewEx(&__pyx_fuse_4__pyx_mdef_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_17__pyx_fuse_4query, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Uint64ClosedBothIntervalNode___p_5, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__109)); if (unlikely(!__pyx_t_45)) __PYX_ERR(1, 3394, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_45);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_45, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_44, __pyx_n_s_uint64_t, __pyx_t_45) < 0) __PYX_ERR(1, 3394, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_45); __pyx_t_45 = 0;
  __pyx_t_45 = __pyx_FusedFunction_NewEx(&__pyx_mdef_6pandas_5_libs_8interval_28Uint64ClosedBothIntervalNode_3query, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Uint64ClosedBothIntervalNode_que, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__109)); if (unlikely(!__pyx_t_45)) __PYX_ERR(1, 3394, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_45);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_45, __pyx_empty_tuple);
  ((__pyx_FusedFunctionObject *) __pyx_t_45)->__signatures__ = __pyx_t_44;
  __Pyx_GIVEREF(__pyx_t_44);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode->tp_dict, __pyx_n_s_query, __pyx_t_45) < 0) __PYX_ERR(1, 3394, __pyx_L1_error)
  PyType_Modified(__pyx_ptype_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode);
  __Pyx_DECREF(__pyx_t_45); __pyx_t_45 = 0;

  /* "pandas/_libs/intervaltree.pxi":3460
 *             return (m, (l, r))
 * 
 * NODE_CLASSES['uint64',             # <<<<<<<<<<<<<<
 *              'both'] = Uint64ClosedBothIntervalNode
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_46, __pyx_n_s_NODE_CLASSES); if (unlikely(!__pyx_t_46)) __PYX_ERR(1, 3460, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_46);
  if (unlikely(PyObject_SetItem(__pyx_t_46, __pyx_tuple__110, ((PyObject *)__pyx_ptype_6pandas_5_libs_8interval_Uint64ClosedBothIntervalNode)) < 0)) __PYX_ERR(1, 3460, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_46); __pyx_t_46 = 0;

  /* "pandas/_libs/intervaltree.pxi":3564
 *     @cython.boundscheck(False)
 *     @cython.initializedcheck(False)
 *     cpdef query(self, Int64Vector result, scalar_t point):             # <<<<<<<<<<<<<<
 *         """Recursively query this node and its sub-nodes for intervals that
 *         overlap with the query point.
 */
  __pyx_t_46 = __Pyx_PyDict_NewPresized(5); if (unlikely(!__pyx_t_46)) __PYX_ERR(1, 3564, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_46);
  __pyx_t_47 = __pyx_FusedFunction_NewEx(&__pyx_fuse_0__pyx_mdef_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_9__pyx_fuse_0query, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Uint64ClosedNeitherIntervalNode_4, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__112)); if (unlikely(!__pyx_t_47)) __PYX_ERR(1, 3564, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_47);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_47, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_46, __pyx_n_s_float64_t, __pyx_t_47) < 0) __PYX_ERR(1, 3564, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_47); __pyx_t_47 = 0;
  __pyx_t_47 = __pyx_FusedFunction_NewEx(&__pyx_fuse_1__pyx_mdef_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_11__pyx_fuse_1query, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Uint64ClosedNeitherIntervalNode_5, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__112)); if (unlikely(!__pyx_t_47)) __PYX_ERR(1, 3564, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_47);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_47, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_46, __pyx_n_s_float32_t, __pyx_t_47) < 0) __PYX_ERR(1, 3564, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_47); __pyx_t_47 = 0;
  __pyx_t_47 = __pyx_FusedFunction_NewEx(&__pyx_fuse_2__pyx_mdef_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_13__pyx_fuse_2query, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Uint64ClosedNeitherIntervalNode_6, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__112)); if (unlikely(!__pyx_t_47)) __PYX_ERR(1, 3564, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_47);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_47, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_46, __pyx_n_s_int64_t, __pyx_t_47) < 0) __PYX_ERR(1, 3564, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_47); __pyx_t_47 = 0;
  __pyx_t_47 = __pyx_FusedFunction_NewEx(&__pyx_fuse_3__pyx_mdef_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_15__pyx_fuse_3query, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Uint64ClosedNeitherIntervalNode_7, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__112)); if (unlikely(!__pyx_t_47)) __PYX_ERR(1, 3564, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_47);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_47, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_46, __pyx_n_s_int32_t, __pyx_t_47) < 0) __PYX_ERR(1, 3564, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_47); __pyx_t_47 = 0;
  __pyx_t_47 = __pyx_FusedFunction_NewEx(&__pyx_fuse_4__pyx_mdef_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_17__pyx_fuse_4query, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Uint64ClosedNeitherIntervalNode_8, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__112)); if (unlikely(!__pyx_t_47)) __PYX_ERR(1, 3564, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_47);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_47, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_46, __pyx_n_s_uint64_t, __pyx_t_47) < 0) __PYX_ERR(1, 3564, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_47); __pyx_t_47 = 0;
  __pyx_t_47 = __pyx_FusedFunction_NewEx(&__pyx_mdef_6pandas_5_libs_8interval_31Uint64ClosedNeitherIntervalNode_3query, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Uint64ClosedNeitherIntervalNode_9, NULL, __pyx_n_s_pandas__libs_interval, __pyx_d, ((PyObject *)__pyx_codeobj__112)); if (unlikely(!__pyx_t_47)) __PYX_ERR(1, 3564, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_47);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_47, __pyx_empty_tuple);
  ((__pyx_FusedFunctionObject *) __pyx_t_47)->__signatures__ = __pyx_t_46;
  __Pyx_GIVEREF(__pyx_t_46);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode->tp_dict, __pyx_n_s_query, __pyx_t_47) < 0) __PYX_ERR(1, 3564, __pyx_L1_error)
  PyType_Modified(__pyx_ptype_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode);
  __Pyx_DECREF(__pyx_t_47); __pyx_t_47 = 0;

  /* "pandas/_libs/intervaltree.pxi":3630
 *             return (m, (l, r))
 * 
 * NODE_CLASSES['uint64',             # <<<<<<<<<<<<<<
 *              'neither'] = Uint64ClosedNeitherIntervalNode
 */
  __Pyx_GetModuleGlobalName(__pyx_t_48, __pyx_n_s_NODE_CLASSES); if (unlikely(!__pyx_t_48)) __PYX_ERR(1, 3630, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_48);
  if (unlikely(PyObject_SetItem(__pyx_t_48, __pyx_tuple__113, ((PyObject *)__pyx_ptype_6pandas_5_libs_8interval_Uint64ClosedNeitherIntervalNode)) < 0)) __PYX_ERR(1, 3630, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_48); __pyx_t_48 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_IntervalMixin(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_48 = PyCFunction_NewEx(&__pyx_mdef_6pandas_5_libs_8interval_3__pyx_unpickle_IntervalMixin, NULL, __pyx_n_s_pandas__libs_interval); if (unlikely(!__pyx_t_48)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_48);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_IntervalMixin, __pyx_t_48) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_48); __pyx_t_48 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle_IntervalMixin__set_state(<IntervalMixin> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_IntervalMixin__set_state(IntervalMixin __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */
  __pyx_t_48 = PyCFunction_NewEx(&__pyx_mdef_6pandas_5_libs_8interval_5__pyx_unpickle_IntervalTree, NULL, __pyx_n_s_pandas__libs_interval); if (unlikely(!__pyx_t_48)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_48);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_IntervalTree, __pyx_t_48) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_48); __pyx_t_48 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_Float32ClosedLeftIntervalNode(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_48 = PyCFunction_NewEx(&__pyx_mdef_6pandas_5_libs_8interval_7__pyx_unpickle_Float32ClosedLeftIntervalNode, NULL, __pyx_n_s_pandas__libs_interval); if (unlikely(!__pyx_t_48)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_48);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_Float32ClosedLeft, __pyx_t_48) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_48); __pyx_t_48 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle_Float32ClosedLeftIntervalNode__set_state(<Float32ClosedLeftIntervalNode> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Float32ClosedLeftIntervalNode__set_state(Float32ClosedLeftIntervalNode __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):
 */
  __pyx_t_48 = PyCFunction_NewEx(&__pyx_mdef_6pandas_5_libs_8interval_9__pyx_unpickle_Float32ClosedRightIntervalNode, NULL, __pyx_n_s_pandas__libs_interval); if (unlikely(!__pyx_t_48)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_48);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_Float32ClosedRigh, __pyx_t_48) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_48); __pyx_t_48 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_Float32ClosedBothIntervalNode(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_48 = PyCFunction_NewEx(&__pyx_mdef_6pandas_5_libs_8interval_11__pyx_unpickle_Float32ClosedBothIntervalNode, NULL, __pyx_n_s_pandas__libs_interval); if (unlikely(!__pyx_t_48)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_48);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_Float32ClosedBoth, __pyx_t_48) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_48); __pyx_t_48 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle_Float32ClosedBothIntervalNode__set_state(<Float32ClosedBothIntervalNode> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Float32ClosedBothIntervalNode__set_state(Float32ClosedBothIntervalNode __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):
 */
  __pyx_t_48 = PyCFunction_NewEx(&__pyx_mdef_6pandas_5_libs_8interval_13__pyx_unpickle_Float32ClosedNeitherIntervalNode, NULL, __pyx_n_s_pandas__libs_interval); if (unlikely(!__pyx_t_48)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_48);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_Float32ClosedNeit, __pyx_t_48) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_48); __pyx_t_48 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_Float64ClosedLeftIntervalNode(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_48 = PyCFunction_NewEx(&__pyx_mdef_6pandas_5_libs_8interval_15__pyx_unpickle_Float64ClosedLeftIntervalNode, NULL, __pyx_n_s_pandas__libs_interval); if (unlikely(!__pyx_t_48)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_48);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_Float64ClosedLeft, __pyx_t_48) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_48); __pyx_t_48 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle_Float64ClosedLeftIntervalNode__set_state(<Float64ClosedLeftIntervalNode> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Float64ClosedLeftIntervalNode__set_state(Float64ClosedLeftIntervalNode __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):
 */
  __pyx_t_48 = PyCFunction_NewEx(&__pyx_mdef_6pandas_5_libs_8interval_17__pyx_unpickle_Float64ClosedRightIntervalNode, NULL, __pyx_n_s_pandas__libs_interval); if (unlikely(!__pyx_t_48)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_48);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_Float64ClosedRigh, __pyx_t_48) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_48); __pyx_t_48 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_Float64ClosedBothIntervalNode(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_48 = PyCFunction_NewEx(&__pyx_mdef_6pandas_5_libs_8interval_19__pyx_unpickle_Float64ClosedBothIntervalNode, NULL, __pyx_n_s_pandas__libs_interval); if (unlikely(!__pyx_t_48)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_48);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_Float64ClosedBoth, __pyx_t_48) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_48); __pyx_t_48 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle_Float64ClosedBothIntervalNode__set_state(<Float64ClosedBothIntervalNode> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Float64ClosedBothIntervalNode__set_state(Float64ClosedBothIntervalNode __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):
 */
  __pyx_t_48 = PyCFunction_NewEx(&__pyx_mdef_6pandas_5_libs_8interval_21__pyx_unpickle_Float64ClosedNeitherIntervalNode, NULL, __pyx_n_s_pandas__libs_interval); if (unlikely(!__pyx_t_48)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_48);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_Float64ClosedNeit, __pyx_t_48) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_48); __pyx_t_48 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_Int32ClosedLeftIntervalNode(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_48 = PyCFunction_NewEx(&__pyx_mdef_6pandas_5_libs_8interval_23__pyx_unpickle_Int32ClosedLeftIntervalNode, NULL, __pyx_n_s_pandas__libs_interval); if (unlikely(!__pyx_t_48)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_48);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_Int32ClosedLeftIn, __pyx_t_48) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_48); __pyx_t_48 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle_Int32ClosedLeftIntervalNode__set_state(<Int32ClosedLeftIntervalNode> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Int32ClosedLeftIntervalNode__set_state(Int32ClosedLeftIntervalNode __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):
 */
  __pyx_t_48 = PyCFunction_NewEx(&__pyx_mdef_6pandas_5_libs_8interval_25__pyx_unpickle_Int32ClosedRightIntervalNode, NULL, __pyx_n_s_pandas__libs_interval); if (unlikely(!__pyx_t_48)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_48);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_Int32ClosedRightI, __pyx_t_48) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_48); __pyx_t_48 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_Int32ClosedBothIntervalNode(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_48 = PyCFunction_NewEx(&__pyx_mdef_6pandas_5_libs_8interval_27__pyx_unpickle_Int32ClosedBothIntervalNode, NULL, __pyx_n_s_pandas__libs_interval); if (unlikely(!__pyx_t_48)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_48);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_Int32ClosedBothIn, __pyx_t_48) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_48); __pyx_t_48 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle_Int32ClosedBothIntervalNode__set_state(<Int32ClosedBothIntervalNode> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Int32ClosedBothIntervalNode__set_state(Int32ClosedBothIntervalNode __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):
 */
  __pyx_t_48 = PyCFunction_NewEx(&__pyx_mdef_6pandas_5_libs_8interval_29__pyx_unpickle_Int32ClosedNeitherIntervalNode, NULL, __pyx_n_s_pandas__libs_interval); if (unlikely(!__pyx_t_48)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_48);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_Int32ClosedNeithe, __pyx_t_48) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_48); __pyx_t_48 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_Int64ClosedLeftIntervalNode(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_48 = PyCFunction_NewEx(&__pyx_mdef_6pandas_5_libs_8interval_31__pyx_unpickle_Int64ClosedLeftIntervalNode, NULL, __pyx_n_s_pandas__libs_interval); if (unlikely(!__pyx_t_48)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_48);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_Int64ClosedLeftIn, __pyx_t_48) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_48); __pyx_t_48 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle_Int64ClosedLeftIntervalNode__set_state(<Int64ClosedLeftIntervalNode> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Int64ClosedLeftIntervalNode__set_state(Int64ClosedLeftIntervalNode __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):
 */
  __pyx_t_48 = PyCFunction_NewEx(&__pyx_mdef_6pandas_5_libs_8interval_33__pyx_unpickle_Int64ClosedRightIntervalNode, NULL, __pyx_n_s_pandas__libs_interval); if (unlikely(!__pyx_t_48)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_48);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_Int64ClosedRightI, __pyx_t_48) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_48); __pyx_t_48 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_Int64ClosedBothIntervalNode(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_48 = PyCFunction_NewEx(&__pyx_mdef_6pandas_5_libs_8interval_35__pyx_unpickle_Int64ClosedBothIntervalNode, NULL, __pyx_n_s_pandas__libs_interval); if (unlikely(!__pyx_t_48)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_48);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_Int64ClosedBothIn, __pyx_t_48) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_48); __pyx_t_48 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle_Int64ClosedBothIntervalNode__set_state(<Int64ClosedBothIntervalNode> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Int64ClosedBothIntervalNode__set_state(Int64ClosedBothIntervalNode __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):
 */
  __pyx_t_48 = PyCFunction_NewEx(&__pyx_mdef_6pandas_5_libs_8interval_37__pyx_unpickle_Int64ClosedNeitherIntervalNode, NULL, __pyx_n_s_pandas__libs_interval); if (unlikely(!__pyx_t_48)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_48);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_Int64ClosedNeithe, __pyx_t_48) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_48); __pyx_t_48 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_Uint64ClosedLeftIntervalNode(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_48 = PyCFunction_NewEx(&__pyx_mdef_6pandas_5_libs_8interval_39__pyx_unpickle_Uint64ClosedLeftIntervalNode, NULL, __pyx_n_s_pandas__libs_interval); if (unlikely(!__pyx_t_48)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_48);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_Uint64ClosedLeftI, __pyx_t_48) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_48); __pyx_t_48 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle_Uint64ClosedLeftIntervalNode__set_state(<Uint64ClosedLeftIntervalNode> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Uint64ClosedLeftIntervalNode__set_state(Uint64ClosedLeftIntervalNode __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):
 */
  __pyx_t_48 = PyCFunction_NewEx(&__pyx_mdef_6pandas_5_libs_8interval_41__pyx_unpickle_Uint64ClosedRightIntervalNode, NULL, __pyx_n_s_pandas__libs_interval); if (unlikely(!__pyx_t_48)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_48);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_Uint64ClosedRight, __pyx_t_48) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_48); __pyx_t_48 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_Uint64ClosedBothIntervalNode(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_48 = PyCFunction_NewEx(&__pyx_mdef_6pandas_5_libs_8interval_43__pyx_unpickle_Uint64ClosedBothIntervalNode, NULL, __pyx_n_s_pandas__libs_interval); if (unlikely(!__pyx_t_48)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_48);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_Uint64ClosedBothI, __pyx_t_48) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_48); __pyx_t_48 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle_Uint64ClosedBothIntervalNode__set_state(<Uint64ClosedBothIntervalNode> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Uint64ClosedBothIntervalNode__set_state(Uint64ClosedBothIntervalNode __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.center_left_indices = __pyx_state[0]; __pyx_result.center_left_values = __pyx_state[1]; __pyx_result.center_right_indices = __pyx_state[2]; __pyx_result.center_right_values = __pyx_state[3]; __pyx_result.indices = __pyx_state[4]; __pyx_result.is_leaf_node = __pyx_state[5]; __pyx_result.leaf_size = __pyx_state[6]; __pyx_result.left = __pyx_state[7]; __pyx_result.left_node = __pyx_state[8]; __pyx_result.max_right = __pyx_state[9]; __pyx_result.min_left = __pyx_state[10]; __pyx_result.n_center = __pyx_state[11]; __pyx_result.n_elements = __pyx_state[12]; __pyx_result.pivot = __pyx_state[13]; __pyx_result.right = __pyx_state[14]; __pyx_result.right_node = __pyx_state[15]
 *     if len(__pyx_state) > 16 and hasattr(__pyx_result, '__dict__'):
 */
  __pyx_t_48 = PyCFunction_NewEx(&__pyx_mdef_6pandas_5_libs_8interval_45__pyx_unpickle_Uint64ClosedNeitherIntervalNode, NULL, __pyx_n_s_pandas__libs_interval); if (unlikely(!__pyx_t_48)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_48);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_Uint64ClosedNeith, __pyx_t_48) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_48); __pyx_t_48 = 0;

  /* "pandas/_libs/interval.pyx":1
 * import numbers             # <<<<<<<<<<<<<<
 * from operator import le, lt
 * 
 */
  __pyx_t_48 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_48)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_48);
  if (PyDict_SetItem(__pyx_t_48, __pyx_kp_u_IntervalMixin_is_empty___get___l, __pyx_kp_u_Indicates_if_an_interval_is_emp) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_48, __pyx_kp_u_Interval_overlaps_line_427, __pyx_kp_u_Check_whether_two_Interval_obje) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_48) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_48); __pyx_t_48 = 0;

  /* "View.MemoryView":209
 *         info.obj = self
 * 
 *     __pyx_getbuffer = capsule(<void *> &__pyx_array_getbuffer, "getbuffer(obj, view, flags)")             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(array self):
 */
  __pyx_t_48 = __pyx_capsule_create(((void *)(&__pyx_array_getbuffer)), ((char *)"getbuffer(obj, view, flags)")); if (unlikely(!__pyx_t_48)) __PYX_ERR(2, 209, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_48);
  if (PyDict_SetItem((PyObject *)__pyx_array_type->tp_dict, __pyx_n_s_pyx_getbuffer, __pyx_t_48) < 0) __PYX_ERR(2, 209, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_48); __pyx_t_48 = 0;
  PyType_Modified(__pyx_array_type);

  /* "View.MemoryView":286
 *         return self.name
 * 
 * cdef generic = Enum("<strided and direct or indirect>")             # <<<<<<<<<<<<<<
 * cdef strided = Enum("<strided and direct>") # default
 * cdef indirect = Enum("<strided and indirect>")
 */
  __pyx_t_48 = __Pyx_PyObject_Call(((PyObject *)__pyx_MemviewEnum_type), __pyx_tuple__158, NULL); if (unlikely(!__pyx_t_48)) __PYX_ERR(2, 286, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_48);
  __Pyx_XGOTREF(generic);
  __Pyx_DECREF_SET(generic, __pyx_t_48);
  __Pyx_GIVEREF(__pyx_t_48);
  __pyx_t_48 = 0;

  /* "View.MemoryView":287
 * 
 * cdef generic = Enum("<strided and direct or indirect>")
 * cdef strided = Enum("<strided and direct>") # default             # <<<<<<<<<<<<<<
 * cdef indirect = Enum("<strided and indirect>")
 * 
 */
  __pyx_t_48 = __Pyx_PyObject_Call(((PyObject *)__pyx_MemviewEnum_type), __pyx_tuple__159, NULL); if (unlikely(!__pyx_t_48)) __PYX_ERR(2, 287, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_48);
  __Pyx_XGOTREF(strided);
  __Pyx_DECREF_SET(strided, __pyx_t_48);
  __Pyx_GIVEREF(__pyx_t_48);
  __pyx_t_48 = 0;

  /* "View.MemoryView":288
 * cdef generic = Enum("<strided and direct or indirect>")
 * cdef strided = Enum("<strided and direct>") # default
 * cdef indirect = Enum("<strided and indirect>")             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_48 = __Pyx_PyObject_Call(((PyObject *)__pyx_MemviewEnum_type), __pyx_tuple__160, NULL); if (unlikely(!__pyx_t_48)) __PYX_ERR(2, 288, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_48);
  __Pyx_XGOTREF(indirect);
  __Pyx_DECREF_SET(indirect, __pyx_t_48);
  __Pyx_GIVEREF(__pyx_t_48);
  __pyx_t_48 = 0;

  /* "View.MemoryView":291
 * 
 * 
 * cdef contiguous = Enum("<contiguous and direct>")             # <<<<<<<<<<<<<<
 * cdef indirect_contiguous = Enum("<contiguous and indirect>")
 * 
 */
  __pyx_t_48 = __Pyx_PyObject_Call(((PyObject *)__pyx_MemviewEnum_type), __pyx_tuple__161, NULL); if (unlikely(!__pyx_t_48)) __PYX_ERR(2, 291, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_48);
  __Pyx_XGOTREF(contiguous);
  __Pyx_DECREF_SET(contiguous, __pyx_t_48);
  __Pyx_GIVEREF(__pyx_t_48);
  __pyx_t_48 = 0;

  /* "View.MemoryView":292
 * 
 * cdef contiguous = Enum("<contiguous and direct>")
 * cdef indirect_contiguous = Enum("<contiguous and indirect>")             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_48 = __Pyx_PyObject_Call(((PyObject *)__pyx_MemviewEnum_type), __pyx_tuple__162, NULL); if (unlikely(!__pyx_t_48)) __PYX_ERR(2, 292, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_48);
  __Pyx_XGOTREF(indirect_contiguous);
  __Pyx_DECREF_SET(indirect_contiguous, __pyx_t_48);
  __Pyx_GIVEREF(__pyx_t_48);
  __pyx_t_48 = 0;

  /* "View.MemoryView":316
 * 
 * DEF THREAD_LOCKS_PREALLOCATED = 8
 * cdef int __pyx_memoryview_thread_locks_used = 0             # <<<<<<<<<<<<<<
 * cdef PyThread_type_lock[THREAD_LOCKS_PREALLOCATED] __pyx_memoryview_thread_locks = [
 *     PyThread_allocate_lock(),
 */
  __pyx_memoryview_thread_locks_used = 0;

  /* "View.MemoryView":317
 * DEF THREAD_LOCKS_PREALLOCATED = 8
 * cdef int __pyx_memoryview_thread_locks_used = 0
 * cdef PyThread_type_lock[THREAD_LOCKS_PREALLOCATED] __pyx_memoryview_thread_locks = [             # <<<<<<<<<<<<<<
 *     PyThread_allocate_lock(),
 *     PyThread_allocate_lock(),
 */
  __pyx_t_49[0] = PyThread_allocate_lock();
  __pyx_t_49[1] = PyThread_allocate_lock();
  __pyx_t_49[2] = PyThread_allocate_lock();
  __pyx_t_49[3] = PyThread_allocate_lock();
  __pyx_t_49[4] = PyThread_allocate_lock();
  __pyx_t_49[5] = PyThread_allocate_lock();
  __pyx_t_49[6] = PyThread_allocate_lock();
  __pyx_t_49[7] = PyThread_allocate_lock();
  memcpy(&(__pyx_memoryview_thread_locks[0]), __pyx_t_49, sizeof(__pyx_memoryview_thread_locks[0]) * (8));

  /* "View.MemoryView":549
 *         info.obj = self
 * 
 *     __pyx_getbuffer = capsule(<void *> &__pyx_memoryview_getbuffer, "getbuffer(obj, view, flags)")             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_48 = __pyx_capsule_create(((void *)(&__pyx_memoryview_getbuffer)), ((char *)"getbuffer(obj, view, flags)")); if (unlikely(!__pyx_t_48)) __PYX_ERR(2, 549, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_48);
  if (PyDict_SetItem((PyObject *)__pyx_memoryview_type->tp_dict, __pyx_n_s_pyx_getbuffer, __pyx_t_48) < 0) __PYX_ERR(2, 549, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_48); __pyx_t_48 = 0;
  PyType_Modified(__pyx_memoryview_type);

  /* "View.MemoryView":995
 *         return self.from_object
 * 
 *     __pyx_getbuffer = capsule(<void *> &__pyx_memoryview_getbuffer, "getbuffer(obj, view, flags)")             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_48 = __pyx_capsule_create(((void *)(&__pyx_memoryview_getbuffer)), ((char *)"getbuffer(obj, view, flags)")); if (unlikely(!__pyx_t_48)) __PYX_ERR(2, 995, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_48);
  if (PyDict_SetItem((PyObject *)__pyx_memoryviewslice_type->tp_dict, __pyx_n_s_pyx_getbuffer, __pyx_t_48) < 0) __PYX_ERR(2, 995, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_48); __pyx_t_48 = 0;
  PyType_Modified(__pyx_memoryviewslice_type);

  /* "(tree fragment)":1
 * def __pyx_unpickle_Enum(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_48 = PyCFunction_NewEx(&__pyx_mdef_15View_dot_MemoryView_1__pyx_unpickle_Enum, NULL, __pyx_n_s_View_MemoryView); if (unlikely(!__pyx_t_48)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_48);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_Enum, __pyx_t_48) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_48); __pyx_t_48 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle_Enum__set_state(<Enum> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Enum__set_state(Enum __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.name = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 */

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_XDECREF(__pyx_t_16);
  __Pyx_XDECREF(__pyx_t_17);
  __Pyx_XDECREF(__pyx_t_18);
  __Pyx_XDECREF(__pyx_t_19);
  __Pyx_XDECREF(__pyx_t_20);
  __Pyx_XDECREF(__pyx_t_21);
  __Pyx_XDECREF(__pyx_t_22);
  __Pyx_XDECREF(__pyx_t_23);
  __Pyx_XDECREF(__pyx_t_24);
  __Pyx_XDECREF(__pyx_t_25);
  __Pyx_XDECREF(__pyx_t_26);
  __Pyx_XDECREF(__pyx_t_27);
  __Pyx_XDECREF(__pyx_t_28);
  __Pyx_XDECREF(__pyx_t_29);
  __Pyx_XDECREF(__pyx_t_30);
  __Pyx_XDECREF(__pyx_t_31);
  __Pyx_XDECREF(__pyx_t_32);
  __Pyx_XDECREF(__pyx_t_33);
  __Pyx_XDECREF(__pyx_t_34);
  __Pyx_XDECREF(__pyx_t_35);
  __Pyx_XDECREF(__pyx_t_36);
  __Pyx_XDECREF(__pyx_t_37);
  __Pyx_XDECREF(__pyx_t_38);
  __Pyx_XDECREF(__pyx_t_39);
  __Pyx_XDECREF(__pyx_t_40);
  __Pyx_XDECREF(__pyx_t_41);
  __Pyx_XDECREF(__pyx_t_42);
  __Pyx_XDECREF(__pyx_t_43);
  __Pyx_XDECREF(__pyx_t_44);
  __Pyx_XDECREF(__pyx_t_45);
  __Pyx_XDECREF(__pyx_t_46);
  __Pyx_XDECREF(__pyx_t_47);
  __Pyx_XDECREF(__pyx_t_48);
  if (__pyx_m) {
    if (__pyx_d) {
      __Pyx_AddTraceback("init pandas._libs.interval", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    Py_CLEAR(__pyx_m);
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init pandas._libs.interval");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  return (__pyx_m != NULL) ? 0 : -1;
  #elif PY_MAJOR_VERSION >= 3
  return __pyx_m;
  #else
  return;
  #endif
}

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule(modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, "RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* PyObjectGetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStr(__pyx_b, name);
    if (unlikely(!result)) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

/* BytesEquals */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY
    return PyObject_RichCompareBool(s1, s2, equals);
#else
    if (s1 == s2) {
        return (equals == Py_EQ);
    } else if (PyBytes_CheckExact(s1) & PyBytes_CheckExact(s2)) {
        const char *ps1, *ps2;
        Py_ssize_t length = PyBytes_GET_SIZE(s1);
        if (length != PyBytes_GET_SIZE(s2))
            return (equals == Py_NE);
        ps1 = PyBytes_AS_STRING(s1);
        ps2 = PyBytes_AS_STRING(s2);
        if (ps1[0] != ps2[0]) {
            return (equals == Py_NE);
        } else if (length == 1) {
            return (equals == Py_EQ);
        } else {
            int result;
#if CYTHON_USE_UNICODE_INTERNALS
            Py_hash_t hash1, hash2;
            hash1 = ((PyBytesObject*)s1)->ob_shash;
            hash2 = ((PyBytesObject*)s2)->ob_shash;
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                return (equals == Py_NE);
            }
#endif
            result = memcmp(ps1, ps2, (size_t)length);
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & PyBytes_CheckExact(s2)) {
        return (equals == Py_NE);
    } else if ((s2 == Py_None) & PyBytes_CheckExact(s1)) {
        return (equals == Py_NE);
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
#endif
}

/* UnicodeEquals */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY
    return PyObject_RichCompareBool(s1, s2, equals);
#else
#if PY_MAJOR_VERSION < 3
    PyObject* owned_ref = NULL;
#endif
    int s1_is_unicode, s2_is_unicode;
    if (s1 == s2) {
        goto return_eq;
    }
    s1_is_unicode = PyUnicode_CheckExact(s1);
    s2_is_unicode = PyUnicode_CheckExact(s2);
#if PY_MAJOR_VERSION < 3
    if ((s1_is_unicode & (!s2_is_unicode)) && PyString_CheckExact(s2)) {
        owned_ref = PyUnicode_FromObject(s2);
        if (unlikely(!owned_ref))
            return -1;
        s2 = owned_ref;
        s2_is_unicode = 1;
    } else if ((s2_is_unicode & (!s1_is_unicode)) && PyString_CheckExact(s1)) {
        owned_ref = PyUnicode_FromObject(s1);
        if (unlikely(!owned_ref))
            return -1;
        s1 = owned_ref;
        s1_is_unicode = 1;
    } else if (((!s2_is_unicode) & (!s1_is_unicode))) {
        return __Pyx_PyBytes_Equals(s1, s2, equals);
    }
#endif
    if (s1_is_unicode & s2_is_unicode) {
        Py_ssize_t length;
        int kind;
        void *data1, *data2;
        if (unlikely(__Pyx_PyUnicode_READY(s1) < 0) || unlikely(__Pyx_PyUnicode_READY(s2) < 0))
            return -1;
        length = __Pyx_PyUnicode_GET_LENGTH(s1);
        if (length != __Pyx_PyUnicode_GET_LENGTH(s2)) {
            goto return_ne;
        }
#if CYTHON_USE_UNICODE_INTERNALS
        {
            Py_hash_t hash1, hash2;
        #if CYTHON_PEP393_ENABLED
            hash1 = ((PyASCIIObject*)s1)->hash;
            hash2 = ((PyASCIIObject*)s2)->hash;
        #else
            hash1 = ((PyUnicodeObject*)s1)->hash;
            hash2 = ((PyUnicodeObject*)s2)->hash;
        #endif
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                goto return_ne;
            }
        }
#endif
        kind = __Pyx_PyUnicode_KIND(s1);
        if (kind != __Pyx_PyUnicode_KIND(s2)) {
            goto return_ne;
        }
        data1 = __Pyx_PyUnicode_DATA(s1);
        data2 = __Pyx_PyUnicode_DATA(s2);
        if (__Pyx_PyUnicode_READ(kind, data1, 0) != __Pyx_PyUnicode_READ(kind, data2, 0)) {
            goto return_ne;
        } else if (length == 1) {
            goto return_eq;
        } else {
            int result = memcmp(data1, data2, (size_t)(length * kind));
            #if PY_MAJOR_VERSION < 3
            Py_XDECREF(owned_ref);
            #endif
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & s2_is_unicode) {
        goto return_ne;
    } else if ((s2 == Py_None) & s1_is_unicode) {
        goto return_ne;
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        #if PY_MAJOR_VERSION < 3
        Py_XDECREF(owned_ref);
        #endif
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
return_eq:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_EQ);
return_ne:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_NE);
#endif
}

/* GetTopmostException */
#if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem *
__Pyx_PyErr_GetTopmostException(PyThreadState *tstate)
{
    _PyErr_StackItem *exc_info = tstate->exc_info;
    while ((exc_info->exc_type == NULL || exc_info->exc_type == Py_None) &&
           exc_info->previous_item != NULL)
    {
        exc_info = exc_info->previous_item;
    }
    return exc_info;
}
#endif

/* SaveResetException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    *type = exc_info->exc_type;
    *value = exc_info->exc_value;
    *tb = exc_info->exc_traceback;
    #else
    *type = tstate->exc_type;
    *value = tstate->exc_value;
    *tb = tstate->exc_traceback;
    #endif
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
}
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = type;
    exc_info->exc_value = value;
    exc_info->exc_traceback = tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = type;
    tstate->exc_value = value;
    tstate->exc_traceback = tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
#endif

/* PyErrExceptionMatches */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx_PyErr_ExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        if (__Pyx_PyErr_GivenExceptionMatches(exc_type, PyTuple_GET_ITEM(tuple, i))) return 1;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err) {
    PyObject *exc_type = tstate->curexc_type;
    if (exc_type == err) return 1;
    if (unlikely(!exc_type)) return 0;
    if (unlikely(PyTuple_Check(err)))
        return __Pyx_PyErr_ExceptionMatchesTuple(exc_type, err);
    return __Pyx_PyErr_GivenExceptionMatches(exc_type, err);
}
#endif

/* GetException */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb)
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb)
#endif
{
    PyObject *local_type, *local_value, *local_tb;
#if CYTHON_FAST_THREAD_STATE
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    local_type = tstate->curexc_type;
    local_value = tstate->curexc_value;
    local_tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#else
    PyErr_Fetch(&local_type, &local_value, &local_tb);
#endif
    PyErr_NormalizeException(&local_type, &local_value, &local_tb);
#if CYTHON_FAST_THREAD_STATE
    if (unlikely(tstate->curexc_type))
#else
    if (unlikely(PyErr_Occurred()))
#endif
        goto bad;
    #if PY_MAJOR_VERSION >= 3
    if (local_tb) {
        if (unlikely(PyException_SetTraceback(local_value, local_tb) < 0))
            goto bad;
    }
    #endif
    Py_XINCREF(local_tb);
    Py_XINCREF(local_type);
    Py_XINCREF(local_value);
    *type = local_type;
    *value = local_value;
    *tb = local_tb;
#if CYTHON_FAST_THREAD_STATE
    #if CYTHON_USE_EXC_INFO_STACK
    {
        _PyErr_StackItem *exc_info = tstate->exc_info;
        tmp_type = exc_info->exc_type;
        tmp_value = exc_info->exc_value;
        tmp_tb = exc_info->exc_traceback;
        exc_info->exc_type = local_type;
        exc_info->exc_value = local_value;
        exc_info->exc_traceback = local_tb;
    }
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = local_type;
    tstate->exc_value = local_value;
    tstate->exc_traceback = local_tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#else
    PyErr_SetExcInfo(local_type, local_value, local_tb);
#endif
    return 0;
bad:
    *type = 0;
    *value = 0;
    *tb = 0;
    Py_XDECREF(local_type);
    Py_XDECREF(local_value);
    Py_XDECREF(local_tb);
    return -1;
}

/* RaiseDoubleKeywords */
static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

/* ParseKeywords */
static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    while (PyDict_Next(kwds, &pos, &key, &value)) {
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
            continue;
        }
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_CheckExact(key)) || likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (**name == key) ? 0 :
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (PyUnicode_GET_SIZE(**name) != PyUnicode_GET_SIZE(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key);
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (PyUnicode_GET_SIZE(**argname) != PyUnicode_GET_SIZE(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    return -1;
}

/* RaiseArgTupleInvalid */
static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* PyFunctionFastCall */
#if CYTHON_FAST_PYCALL
static PyObject* __Pyx_PyFunction_FastCallNoKw(PyCodeObject *co, PyObject **args, Py_ssize_t na,
                                               PyObject *globals) {
    PyFrameObject *f;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject **fastlocals;
    Py_ssize_t i;
    PyObject *result;
    assert(globals != NULL);
    /* XXX Perhaps we should create a specialized
       PyFrame_New() that doesn't take locals, but does
       take builtins without sanity checking them.
       */
    assert(tstate != NULL);
    f = PyFrame_New(tstate, co, globals, NULL);
    if (f == NULL) {
        return NULL;
    }
    fastlocals = __Pyx_PyFrame_GetLocalsplus(f);
    for (i = 0; i < na; i++) {
        Py_INCREF(*args);
        fastlocals[i] = *args++;
    }
    result = PyEval_EvalFrameEx(f,0);
    ++tstate->recursion_depth;
    Py_DECREF(f);
    --tstate->recursion_depth;
    return result;
}
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs) {
    PyCodeObject *co = (PyCodeObject *)PyFunction_GET_CODE(func);
    PyObject *globals = PyFunction_GET_GLOBALS(func);
    PyObject *argdefs = PyFunction_GET_DEFAULTS(func);
    PyObject *closure;
#if PY_MAJOR_VERSION >= 3
    PyObject *kwdefs;
#endif
    PyObject *kwtuple, **k;
    PyObject **d;
    Py_ssize_t nd;
    Py_ssize_t nk;
    PyObject *result;
    assert(kwargs == NULL || PyDict_Check(kwargs));
    nk = kwargs ? PyDict_Size(kwargs) : 0;
    if (Py_EnterRecursiveCall((char*)" while calling a Python object")) {
        return NULL;
    }
    if (
#if PY_MAJOR_VERSION >= 3
            co->co_kwonlyargcount == 0 &&
#endif
            likely(kwargs == NULL || nk == 0) &&
            co->co_flags == (CO_OPTIMIZED | CO_NEWLOCALS | CO_NOFREE)) {
        if (argdefs == NULL && co->co_argcount == nargs) {
            result = __Pyx_PyFunction_FastCallNoKw(co, args, nargs, globals);
            goto done;
        }
        else if (nargs == 0 && argdefs != NULL
                 && co->co_argcount == Py_SIZE(argdefs)) {
            /* function called with no arguments, but all parameters have
               a default value: use default values as arguments .*/
            args = &PyTuple_GET_ITEM(argdefs, 0);
            result =__Pyx_PyFunction_FastCallNoKw(co, args, Py_SIZE(argdefs), globals);
            goto done;
        }
    }
    if (kwargs != NULL) {
        Py_ssize_t pos, i;
        kwtuple = PyTuple_New(2 * nk);
        if (kwtuple == NULL) {
            result = NULL;
            goto done;
        }
        k = &PyTuple_GET_ITEM(kwtuple, 0);
        pos = i = 0;
        while (PyDict_Next(kwargs, &pos, &k[i], &k[i+1])) {
            Py_INCREF(k[i]);
            Py_INCREF(k[i+1]);
            i += 2;
        }
        nk = i / 2;
    }
    else {
        kwtuple = NULL;
        k = NULL;
    }
    closure = PyFunction_GET_CLOSURE(func);
#if PY_MAJOR_VERSION >= 3
    kwdefs = PyFunction_GET_KW_DEFAULTS(func);
#endif
    if (argdefs != NULL) {
        d = &PyTuple_GET_ITEM(argdefs, 0);
        nd = Py_SIZE(argdefs);
    }
    else {
        d = NULL;
        nd = 0;
    }
#if PY_MAJOR_VERSION >= 3
    result = PyEval_EvalCodeEx((PyObject*)co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, kwdefs, closure);
#else
    result = PyEval_EvalCodeEx(co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, closure);
#endif
    Py_XDECREF(kwtuple);
done:
    Py_LeaveRecursiveCall();
    return result;
}
#endif
#endif

/* PyCFunctionFastCall */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject * __Pyx_PyCFunction_FastCall(PyObject *func_obj, PyObject **args, Py_ssize_t nargs) {
    PyCFunctionObject *func = (PyCFunctionObject*)func_obj;
    PyCFunction meth = PyCFunction_GET_FUNCTION(func);
    PyObject *self = PyCFunction_GET_SELF(func);
    int flags = PyCFunction_GET_FLAGS(func);
    assert(PyCFunction_Check(func));
    assert(METH_FASTCALL == (flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)));
    assert(nargs >= 0);
    assert(nargs == 0 || args != NULL);
    /* _PyCFunction_FastCallDict() must not be called with an exception set,
       because it may clear it (directly or indirectly) and so the
       caller loses its exception */
    assert(!PyErr_Occurred());
    if ((PY_VERSION_HEX < 0x030700A0) || unlikely(flags & METH_KEYWORDS)) {
        return (*((__Pyx_PyCFunctionFastWithKeywords)(void*)meth)) (self, args, nargs, NULL);
    } else {
        return (*((__Pyx_PyCFunctionFast)(void*)meth)) (self, args, nargs);
    }
}
#endif

/* PyObjectCall */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = func->ob_type->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallMethO */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = PyCFunction_GET_FUNCTION(func);
    self = PyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallOneArg */
#if CYTHON_COMPILING_IN_CPYTHON
static PyObject* __Pyx__PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_New(1);
    if (unlikely(!args)) return NULL;
    Py_INCREF(arg);
    PyTuple_SET_ITEM(args, 0, arg);
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, &arg, 1);
    }
#endif
    if (likely(PyCFunction_Check(func))) {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_O)) {
            return __Pyx_PyObject_CallMethO(func, arg);
#if CYTHON_FAST_PYCCALL
        } else if (PyCFunction_GET_FLAGS(func) & METH_FASTCALL) {
            return __Pyx_PyCFunction_FastCall(func, &arg, 1);
#endif
        }
    }
    return __Pyx__PyObject_CallOneArg(func, arg);
}
#else
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_Pack(1, arg);
    if (unlikely(!args)) return NULL;
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
#endif

/* PyErrFetchRestore */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
}
#endif

/* RaiseException */
#if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb,
                        CYTHON_UNUSED PyObject *cause) {
    __Pyx_PyThreadState_declare
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    if (PyType_Check(type)) {
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
    }
    __Pyx_PyThreadState_assign
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
    if (cause) {
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
#if CYTHON_COMPILING_IN_PYPY
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#else
        PyThreadState *tstate = __Pyx_PyThreadState_Current;
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

/* GetAttr */
static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *o, PyObject *n) {
#if CYTHON_USE_TYPE_SLOTS
#if PY_MAJOR_VERSION >= 3
    if (likely(PyUnicode_Check(n)))
#else
    if (likely(PyString_Check(n)))
#endif
        return __Pyx_PyObject_GetAttrStr(o, n);
#endif
    return PyObject_GetAttr(o, n);
}

/* GetAttr3 */
static PyObject *__Pyx_GetAttr3Default(PyObject *d) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (unlikely(!__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        return NULL;
    __Pyx_PyErr_Clear();
    Py_INCREF(d);
    return d;
}
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *o, PyObject *n, PyObject *d) {
    PyObject *r = __Pyx_GetAttr(o, n);
    return (likely(r)) ? r : __Pyx_GetAttr3Default(d);
}

/* PyDictVersioning */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    return likely(dict) ? __PYX_GET_DICT_VERSION(dict) : 0;
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj) {
    PyObject **dictptr = NULL;
    Py_ssize_t offset = Py_TYPE(obj)->tp_dictoffset;
    if (offset) {
#if CYTHON_COMPILING_IN_CPYTHON
        dictptr = (likely(offset > 0)) ? (PyObject **) ((char *)obj + offset) : _PyObject_GetDictPtr(obj);
#else
        dictptr = _PyObject_GetDictPtr(obj);
#endif
    }
    return (dictptr && *dictptr) ? __PYX_GET_DICT_VERSION(*dictptr) : 0;
}
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    if (unlikely(!dict) || unlikely(tp_dict_version != __PYX_GET_DICT_VERSION(dict)))
        return 0;
    return obj_dict_version == __Pyx_get_object_dict_version(obj);
}
#endif

/* GetModuleGlobalName */
#if CYTHON_USE_DICT_VERSIONS
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value)
#else
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name)
#endif
{
    PyObject *result;
#if !CYTHON_AVOID_BORROWED_REFS
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1
    result = _PyDict_GetItem_KnownHash(__pyx_d, name, ((PyASCIIObject *) name)->hash);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    } else if (unlikely(PyErr_Occurred())) {
        return NULL;
    }
#else
    result = PyDict_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
#endif
#else
    result = PyObject_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
    PyErr_Clear();
#endif
    return __Pyx_GetBuiltinName(name);
}

/* HasAttr */
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *o, PyObject *n) {
    PyObject *r;
    if (unlikely(!__Pyx_PyBaseString_Check(n))) {
        PyErr_SetString(PyExc_TypeError,
                        "hasattr(): attribute name must be string");
        return -1;
    }
    r = __Pyx_GetAttr(o, n);
    if (unlikely(!r)) {
        PyErr_Clear();
        return 0;
    } else {
        Py_DECREF(r);
        return 1;
    }
}

/* ArgTypeTest */
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact)
{
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    else if (exact) {
        #if PY_MAJOR_VERSION == 2
        if ((type == &PyBaseString_Type) && likely(__Pyx_PyBaseString_CheckExact(obj))) return 1;
        #endif
    }
    else {
        if (likely(__Pyx_TypeCheck(obj, type))) return 1;
    }
    PyErr_Format(PyExc_TypeError,
        "Argument '%.200s' has incorrect type (expected %.200s, got %.200s)",
        name, type->tp_name, Py_TYPE(obj)->tp_name);
    return 0;
}

/* PyObjectCall2Args */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2) {
    PyObject *args, *result = NULL;
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyFunction_FastCall(function, args, 2);
    }
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyCFunction_FastCall(function, args, 2);
    }
    #endif
    args = PyTuple_New(2);
    if (unlikely(!args)) goto done;
    Py_INCREF(arg1);
    PyTuple_SET_ITEM(args, 0, arg1);
    Py_INCREF(arg2);
    PyTuple_SET_ITEM(args, 1, arg2);
    Py_INCREF(function);
    result = __Pyx_PyObject_Call(function, args, NULL);
    Py_DECREF(args);
    Py_DECREF(function);
done:
    return result;
}

/* DictGetItem */
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key) {
    PyObject *value;
    value = PyDict_GetItemWithError(d, key);
    if (unlikely(!value)) {
        if (!PyErr_Occurred()) {
            if (unlikely(PyTuple_Check(key))) {
                PyObject* args = PyTuple_Pack(1, key);
                if (likely(args)) {
                    PyErr_SetObject(PyExc_KeyError, args);
                    Py_DECREF(args);
                }
            } else {
                PyErr_SetObject(PyExc_KeyError, key);
            }
        }
        return NULL;
    }
    Py_INCREF(value);
    return value;
}
#endif

/* PyObjectCallNoArg */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, NULL, 0);
    }
#endif
#ifdef __Pyx_CyFunction_USED
    if (likely(PyCFunction_Check(func) || __Pyx_CyFunction_Check(func)))
#else
    if (likely(PyCFunction_Check(func)))
#endif
    {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_NOARGS)) {
            return __Pyx_PyObject_CallMethO(func, NULL);
        }
    }
    return __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL);
}
#endif

/* RaiseTooManyValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected) {
    PyErr_Format(PyExc_ValueError,
                 "too many values to unpack (expected %" CYTHON_FORMAT_SSIZE_T "d)", expected);
}

/* RaiseNeedMoreValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index) {
    PyErr_Format(PyExc_ValueError,
                 "need more than %" CYTHON_FORMAT_SSIZE_T "d value%.1s to unpack",
                 index, (index == 1) ? "" : "s");
}

/* IterFinish */
static CYTHON_INLINE int __Pyx_IterFinish(void) {
#if CYTHON_FAST_THREAD_STATE
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject* exc_type = tstate->curexc_type;
    if (unlikely(exc_type)) {
        if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) {
            PyObject *exc_value, *exc_tb;
            exc_value = tstate->curexc_value;
            exc_tb = tstate->curexc_traceback;
            tstate->curexc_type = 0;
            tstate->curexc_value = 0;
            tstate->curexc_traceback = 0;
            Py_DECREF(exc_type);
            Py_XDECREF(exc_value);
            Py_XDECREF(exc_tb);
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#else
    if (unlikely(PyErr_Occurred())) {
        if (likely(PyErr_ExceptionMatches(PyExc_StopIteration))) {
            PyErr_Clear();
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#endif
}

/* UnpackItemEndCheck */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected) {
    if (unlikely(retval)) {
        Py_DECREF(retval);
        __Pyx_RaiseTooManyValuesError(expected);
        return -1;
    } else {
        return __Pyx_IterFinish();
    }
    return 0;
}

/* ExtTypeTest */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type) {
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    if (likely(__Pyx_TypeCheck(obj, type)))
        return 1;
    PyErr_Format(PyExc_TypeError, "Cannot convert %.200s to %.200s",
                 Py_TYPE(obj)->tp_name, type->tp_name);
    return 0;
}

/* GetItemInt */
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j) {
    PyObject *r;
    if (!j) return NULL;
    r = PyObject_GetItem(o, j);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyList_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyList_GET_SIZE(o)))) {
        PyObject *r = PyList_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyTuple_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyTuple_GET_SIZE(o)))) {
        PyObject *r = PyTuple_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i, int is_list,
                                                     CYTHON_NCP_UNUSED int wraparound,
                                                     CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyList_GET_SIZE(o);
        if ((!boundscheck) || (likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o))))) {
            PyObject *r = PyList_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    }
    else if (PyTuple_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyTuple_GET_SIZE(o);
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyTuple_GET_SIZE(o)))) {
            PyObject *r = PyTuple_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    } else {
        PySequenceMethods *m = Py_TYPE(o)->tp_as_sequence;
        if (likely(m && m->sq_item)) {
            if (wraparound && unlikely(i < 0) && likely(m->sq_length)) {
                Py_ssize_t l = m->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return NULL;
                    PyErr_Clear();
                }
            }
            return m->sq_item(o, i);
        }
    }
#else
    if (is_list || PySequence_Check(o)) {
        return PySequence_GetItem(o, i);
    }
#endif
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
}

/* SetItemInt */
static int __Pyx_SetItemInt_Generic(PyObject *o, PyObject *j, PyObject *v) {
    int r;
    if (!j) return -1;
    r = PyObject_SetItem(o, j, v);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE int __Pyx_SetItemInt_Fast(PyObject *o, Py_ssize_t i, PyObject *v, int is_list,
                                               CYTHON_NCP_UNUSED int wraparound, CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = (!wraparound) ? i : ((likely(i >= 0)) ? i : i + PyList_GET_SIZE(o));
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o)))) {
            PyObject* old = PyList_GET_ITEM(o, n);
            Py_INCREF(v);
            PyList_SET_ITEM(o, n, v);
            Py_DECREF(old);
            return 1;
        }
    } else {
        PySequenceMethods *m = Py_TYPE(o)->tp_as_sequence;
        if (likely(m && m->sq_ass_item)) {
            if (wraparound && unlikely(i < 0) && likely(m->sq_length)) {
                Py_ssize_t l = m->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return -1;
                    PyErr_Clear();
                }
            }
            return m->sq_ass_item(o, i, v);
        }
    }
#else
#if CYTHON_COMPILING_IN_PYPY
    if (is_list || (PySequence_Check(o) && !PyDict_Check(o)))
#else
    if (is_list || PySequence_Check(o))
#endif
    {
        return PySequence_SetItem(o, i, v);
    }
#endif
    return __Pyx_SetItemInt_Generic(o, PyInt_FromSsize_t(i), v);
}

/* ObjectGetItem */
#if CYTHON_USE_TYPE_SLOTS
static PyObject *__Pyx_PyObject_GetIndex(PyObject *obj, PyObject* index) {
    PyObject *runerr;
    Py_ssize_t key_value;
    PySequenceMethods *m = Py_TYPE(obj)->tp_as_sequence;
    if (unlikely(!(m && m->sq_item))) {
        PyErr_Format(PyExc_TypeError, "'%.200s' object is not subscriptable", Py_TYPE(obj)->tp_name);
        return NULL;
    }
    key_value = __Pyx_PyIndex_AsSsize_t(index);
    if (likely(key_value != -1 || !(runerr = PyErr_Occurred()))) {
        return __Pyx_GetItemInt_Fast(obj, key_value, 0, 1, 1);
    }
    if (PyErr_GivenExceptionMatches(runerr, PyExc_OverflowError)) {
        PyErr_Clear();
        PyErr_Format(PyExc_IndexError, "cannot fit '%.200s' into an index-sized integer", Py_TYPE(index)->tp_name);
    }
    return NULL;
}
static PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject* key) {
    PyMappingMethods *m = Py_TYPE(obj)->tp_as_mapping;
    if (likely(m && m->mp_subscript)) {
        return m->mp_subscript(obj, key);
    }
    return __Pyx_PyObject_GetIndex(obj, key);
}
#endif

/* SliceObject */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetSlice(PyObject* obj,
        Py_ssize_t cstart, Py_ssize_t cstop,
        PyObject** _py_start, PyObject** _py_stop, PyObject** _py_slice,
        int has_cstart, int has_cstop, CYTHON_UNUSED int wraparound) {
#if CYTHON_USE_TYPE_SLOTS
    PyMappingMethods* mp;
#if PY_MAJOR_VERSION < 3
    PySequenceMethods* ms = Py_TYPE(obj)->tp_as_sequence;
    if (likely(ms && ms->sq_slice)) {
        if (!has_cstart) {
            if (_py_start && (*_py_start != Py_None)) {
                cstart = __Pyx_PyIndex_AsSsize_t(*_py_start);
                if ((cstart == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
            } else
                cstart = 0;
        }
        if (!has_cstop) {
            if (_py_stop && (*_py_stop != Py_None)) {
                cstop = __Pyx_PyIndex_AsSsize_t(*_py_stop);
                if ((cstop == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
            } else
                cstop = PY_SSIZE_T_MAX;
        }
        if (wraparound && unlikely((cstart < 0) | (cstop < 0)) && likely(ms->sq_length)) {
            Py_ssize_t l = ms->sq_length(obj);
            if (likely(l >= 0)) {
                if (cstop < 0) {
                    cstop += l;
                    if (cstop < 0) cstop = 0;
                }
                if (cstart < 0) {
                    cstart += l;
                    if (cstart < 0) cstart = 0;
                }
            } else {
                if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                    goto bad;
                PyErr_Clear();
            }
        }
        return ms->sq_slice(obj, cstart, cstop);
    }
#endif
    mp = Py_TYPE(obj)->tp_as_mapping;
    if (likely(mp && mp->mp_subscript))
#endif
    {
        PyObject* result;
        PyObject *py_slice, *py_start, *py_stop;
        if (_py_slice) {
            py_slice = *_py_slice;
        } else {
            PyObject* owned_start = NULL;
            PyObject* owned_stop = NULL;
            if (_py_start) {
                py_start = *_py_start;
            } else {
                if (has_cstart) {
                    owned_start = py_start = PyInt_FromSsize_t(cstart);
                    if (unlikely(!py_start)) goto bad;
                } else
                    py_start = Py_None;
            }
            if (_py_stop) {
                py_stop = *_py_stop;
            } else {
                if (has_cstop) {
                    owned_stop = py_stop = PyInt_FromSsize_t(cstop);
                    if (unlikely(!py_stop)) {
                        Py_XDECREF(owned_start);
                        goto bad;
                    }
                } else
                    py_stop = Py_None;
            }
            py_slice = PySlice_New(py_start, py_stop, Py_None);
            Py_XDECREF(owned_start);
            Py_XDECREF(owned_stop);
            if (unlikely(!py_slice)) goto bad;
        }
#if CYTHON_USE_TYPE_SLOTS
        result = mp->mp_subscript(obj, py_slice);
#else
        result = PyObject_GetItem(obj, py_slice);
#endif
        if (!_py_slice) {
            Py_DECREF(py_slice);
        }
        return result;
    }
    PyErr_Format(PyExc_TypeError,
        "'%.200s' object is unsliceable", Py_TYPE(obj)->tp_name);
bad:
    return NULL;
}

/* PyObjectGetMethod */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method) {
    PyObject *attr;
#if CYTHON_UNPACK_METHODS && CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_PYTYPE_LOOKUP
    PyTypeObject *tp = Py_TYPE(obj);
    PyObject *descr;
    descrgetfunc f = NULL;
    PyObject **dictptr, *dict;
    int meth_found = 0;
    assert (*method == NULL);
    if (unlikely(tp->tp_getattro != PyObject_GenericGetAttr)) {
        attr = __Pyx_PyObject_GetAttrStr(obj, name);
        goto try_unpack;
    }
    if (unlikely(tp->tp_dict == NULL) && unlikely(PyType_Ready(tp) < 0)) {
        return 0;
    }
    descr = _PyType_Lookup(tp, name);
    if (likely(descr != NULL)) {
        Py_INCREF(descr);
#if PY_MAJOR_VERSION >= 3
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || (Py_TYPE(descr) == &PyMethodDescr_Type) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr) || (Py_TYPE(descr) == &PyMethodDescr_Type)))
        #endif
#else
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr)))
        #endif
#endif
        {
            meth_found = 1;
        } else {
            f = Py_TYPE(descr)->tp_descr_get;
            if (f != NULL && PyDescr_IsData(descr)) {
                attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
                Py_DECREF(descr);
                goto try_unpack;
            }
        }
    }
    dictptr = _PyObject_GetDictPtr(obj);
    if (dictptr != NULL && (dict = *dictptr) != NULL) {
        Py_INCREF(dict);
        attr = __Pyx_PyDict_GetItemStr(dict, name);
        if (attr != NULL) {
            Py_INCREF(attr);
            Py_DECREF(dict);
            Py_XDECREF(descr);
            goto try_unpack;
        }
        Py_DECREF(dict);
    }
    if (meth_found) {
        *method = descr;
        return 1;
    }
    if (f != NULL) {
        attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
        Py_DECREF(descr);
        goto try_unpack;
    }
    if (descr != NULL) {
        *method = descr;
        return 0;
    }
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'%.50s' object has no attribute '%U'",
                 tp->tp_name, name);
#else
                 "'%.50s' object has no attribute '%.400s'",
                 tp->tp_name, PyString_AS_STRING(name));
#endif
    return 0;
#else
    attr = __Pyx_PyObject_GetAttrStr(obj, name);
    goto try_unpack;
#endif
try_unpack:
#if CYTHON_UNPACK_METHODS
    if (likely(attr) && PyMethod_Check(attr) && likely(PyMethod_GET_SELF(attr) == obj)) {
        PyObject *function = PyMethod_GET_FUNCTION(attr);
        Py_INCREF(function);
        Py_DECREF(attr);
        *method = function;
        return 1;
    }
#endif
    *method = attr;
    return 0;
}

/* PyObjectCallMethod0 */
static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name) {
    PyObject *method = NULL, *result = NULL;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_CallOneArg(method, obj);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) goto bad;
    result = __Pyx_PyObject_CallNoArg(method);
    Py_DECREF(method);
bad:
    return result;
}

/* RaiseNoneIterError */
static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
}

/* UnpackTupleError */
static void __Pyx_UnpackTupleError(PyObject *t, Py_ssize_t index) {
    if (t == Py_None) {
      __Pyx_RaiseNoneNotIterableError();
    } else if (PyTuple_GET_SIZE(t) < index) {
      __Pyx_RaiseNeedMoreValuesError(PyTuple_GET_SIZE(t));
    } else {
      __Pyx_RaiseTooManyValuesError(index);
    }
}

/* UnpackTuple2 */
static CYTHON_INLINE int __Pyx_unpack_tuple2_exact(
        PyObject* tuple, PyObject** pvalue1, PyObject** pvalue2, int decref_tuple) {
    PyObject *value1 = NULL, *value2 = NULL;
#if CYTHON_COMPILING_IN_PYPY
    value1 = PySequence_ITEM(tuple, 0);  if (unlikely(!value1)) goto bad;
    value2 = PySequence_ITEM(tuple, 1);  if (unlikely(!value2)) goto bad;
#else
    value1 = PyTuple_GET_ITEM(tuple, 0);  Py_INCREF(value1);
    value2 = PyTuple_GET_ITEM(tuple, 1);  Py_INCREF(value2);
#endif
    if (decref_tuple) {
        Py_DECREF(tuple);
    }
    *pvalue1 = value1;
    *pvalue2 = value2;
    return 0;
#if CYTHON_COMPILING_IN_PYPY
bad:
    Py_XDECREF(value1);
    Py_XDECREF(value2);
    if (decref_tuple) { Py_XDECREF(tuple); }
    return -1;
#endif
}
static int __Pyx_unpack_tuple2_generic(PyObject* tuple, PyObject** pvalue1, PyObject** pvalue2,
                                       int has_known_size, int decref_tuple) {
    Py_ssize_t index;
    PyObject *value1 = NULL, *value2 = NULL, *iter = NULL;
    iternextfunc iternext;
    iter = PyObject_GetIter(tuple);
    if (unlikely(!iter)) goto bad;
    if (decref_tuple) { Py_DECREF(tuple); tuple = NULL; }
    iternext = Py_TYPE(iter)->tp_iternext;
    value1 = iternext(iter); if (unlikely(!value1)) { index = 0; goto unpacking_failed; }
    value2 = iternext(iter); if (unlikely(!value2)) { index = 1; goto unpacking_failed; }
    if (!has_known_size && unlikely(__Pyx_IternextUnpackEndCheck(iternext(iter), 2))) goto bad;
    Py_DECREF(iter);
    *pvalue1 = value1;
    *pvalue2 = value2;
    return 0;
unpacking_failed:
    if (!has_known_size && __Pyx_IterFinish() == 0)
        __Pyx_RaiseNeedMoreValuesError(index);
bad:
    Py_XDECREF(iter);
    Py_XDECREF(value1);
    Py_XDECREF(value2);
    if (decref_tuple) { Py_XDECREF(tuple); }
    return -1;
}

/* dict_iter */
static CYTHON_INLINE PyObject* __Pyx_dict_iterator(PyObject* iterable, int is_dict, PyObject* method_name,
                                                   Py_ssize_t* p_orig_length, int* p_source_is_dict) {
    is_dict = is_dict || likely(PyDict_CheckExact(iterable));
    *p_source_is_dict = is_dict;
    if (is_dict) {
#if !CYTHON_COMPILING_IN_PYPY
        *p_orig_length = PyDict_Size(iterable);
        Py_INCREF(iterable);
        return iterable;
#elif PY_MAJOR_VERSION >= 3
        static PyObject *py_items = NULL, *py_keys = NULL, *py_values = NULL;
        PyObject **pp = NULL;
        if (method_name) {
            const char *name = PyUnicode_AsUTF8(method_name);
            if (strcmp(name, "iteritems") == 0) pp = &py_items;
            else if (strcmp(name, "iterkeys") == 0) pp = &py_keys;
            else if (strcmp(name, "itervalues") == 0) pp = &py_values;
            if (pp) {
                if (!*pp) {
                    *pp = PyUnicode_FromString(name + 4);
                    if (!*pp)
                        return NULL;
                }
                method_name = *pp;
            }
        }
#endif
    }
    *p_orig_length = 0;
    if (method_name) {
        PyObject* iter;
        iterable = __Pyx_PyObject_CallMethod0(iterable, method_name);
        if (!iterable)
            return NULL;
#if !CYTHON_COMPILING_IN_PYPY
        if (PyTuple_CheckExact(iterable) || PyList_CheckExact(iterable))
            return iterable;
#endif
        iter = PyObject_GetIter(iterable);
        Py_DECREF(iterable);
        return iter;
    }
    return PyObject_GetIter(iterable);
}
static CYTHON_INLINE int __Pyx_dict_iter_next(
        PyObject* iter_obj, CYTHON_NCP_UNUSED Py_ssize_t orig_length, CYTHON_NCP_UNUSED Py_ssize_t* ppos,
        PyObject** pkey, PyObject** pvalue, PyObject** pitem, int source_is_dict) {
    PyObject* next_item;
#if !CYTHON_COMPILING_IN_PYPY
    if (source_is_dict) {
        PyObject *key, *value;
        if (unlikely(orig_length != PyDict_Size(iter_obj))) {
            PyErr_SetString(PyExc_RuntimeError, "dictionary changed size during iteration");
            return -1;
        }
        if (unlikely(!PyDict_Next(iter_obj, ppos, &key, &value))) {
            return 0;
        }
        if (pitem) {
            PyObject* tuple = PyTuple_New(2);
            if (unlikely(!tuple)) {
                return -1;
            }
            Py_INCREF(key);
            Py_INCREF(value);
            PyTuple_SET_ITEM(tuple, 0, key);
            PyTuple_SET_ITEM(tuple, 1, value);
            *pitem = tuple;
        } else {
            if (pkey) {
                Py_INCREF(key);
                *pkey = key;
            }
            if (pvalue) {
                Py_INCREF(value);
                *pvalue = value;
            }
        }
        return 1;
    } else if (PyTuple_CheckExact(iter_obj)) {
        Py_ssize_t pos = *ppos;
        if (unlikely(pos >= PyTuple_GET_SIZE(iter_obj))) return 0;
        *ppos = pos + 1;
        next_item = PyTuple_GET_ITEM(iter_obj, pos);
        Py_INCREF(next_item);
    } else if (PyList_CheckExact(iter_obj)) {
        Py_ssize_t pos = *ppos;
        if (unlikely(pos >= PyList_GET_SIZE(iter_obj))) return 0;
        *ppos = pos + 1;
        next_item = PyList_GET_ITEM(iter_obj, pos);
        Py_INCREF(next_item);
    } else
#endif
    {
        next_item = PyIter_Next(iter_obj);
        if (unlikely(!next_item)) {
            return __Pyx_IterFinish();
        }
    }
    if (pitem) {
        *pitem = next_item;
    } else if (pkey && pvalue) {
        if (__Pyx_unpack_tuple2(next_item, pkey, pvalue, source_is_dict, source_is_dict, 1))
            return -1;
    } else if (pkey) {
        *pkey = next_item;
    } else {
        *pvalue = next_item;
    }
    return 1;
}

/* Import */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level) {
    PyObject *empty_list = 0;
    PyObject *module = 0;
    PyObject *global_dict = 0;
    PyObject *empty_dict = 0;
    PyObject *list;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_import;
    py_import = __Pyx_PyObject_GetAttrStr(__pyx_b, __pyx_n_s_import);
    if (!py_import)
        goto bad;
    #endif
    if (from_list)
        list = from_list;
    else {
        empty_list = PyList_New(0);
        if (!empty_list)
            goto bad;
        list = empty_list;
    }
    global_dict = PyModule_GetDict(__pyx_m);
    if (!global_dict)
        goto bad;
    empty_dict = PyDict_New();
    if (!empty_dict)
        goto bad;
    {
        #if PY_MAJOR_VERSION >= 3
        if (level == -1) {
            if (strchr(__Pyx_MODULE_NAME, '.')) {
                module = PyImport_ImportModuleLevelObject(
                    name, global_dict, empty_dict, list, 1);
                if (!module) {
                    if (!PyErr_ExceptionMatches(PyExc_ImportError))
                        goto bad;
                    PyErr_Clear();
                }
            }
            level = 0;
        }
        #endif
        if (!module) {
            #if PY_MAJOR_VERSION < 3
            PyObject *py_level = PyInt_FromLong(level);
            if (!py_level)
                goto bad;
            module = PyObject_CallFunctionObjArgs(py_import,
                name, global_dict, empty_dict, list, py_level, (PyObject *)NULL);
            Py_DECREF(py_level);
            #else
            module = PyImport_ImportModuleLevelObject(
                name, global_dict, empty_dict, list, level);
            #endif
        }
    }
bad:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_import);
    #endif
    Py_XDECREF(empty_list);
    Py_XDECREF(empty_dict);
    return module;
}

/* UnicodeAsUCS4 */
static CYTHON_INLINE Py_UCS4 __Pyx_PyUnicode_AsPy_UCS4(PyObject* x) {
   Py_ssize_t length;
   #if CYTHON_PEP393_ENABLED
   length = PyUnicode_GET_LENGTH(x);
   if (likely(length == 1)) {
       return PyUnicode_READ_CHAR(x, 0);
   }
   #else
   length = PyUnicode_GET_SIZE(x);
   if (likely(length == 1)) {
       return PyUnicode_AS_UNICODE(x)[0];
   }
   #if Py_UNICODE_SIZE == 2
   else if (PyUnicode_GET_SIZE(x) == 2) {
       Py_UCS4 high_val = PyUnicode_AS_UNICODE(x)[0];
       if (high_val >= 0xD800 && high_val <= 0xDBFF) {
           Py_UCS4 low_val = PyUnicode_AS_UNICODE(x)[1];
           if (low_val >= 0xDC00 && low_val <= 0xDFFF) {
               return 0x10000 + (((high_val & ((1<<10)-1)) << 10) | (low_val & ((1<<10)-1)));
           }
       }
   }
   #endif
   #endif
   PyErr_Format(PyExc_ValueError,
                "only single character unicode strings can be converted to Py_UCS4, "
                "got length %" CYTHON_FORMAT_SSIZE_T "d", length);
   return (Py_UCS4)-1;
}

/* object_ord */
static long __Pyx__PyObject_Ord(PyObject* c) {
    Py_ssize_t size;
    if (PyBytes_Check(c)) {
        size = PyBytes_GET_SIZE(c);
        if (likely(size == 1)) {
            return (unsigned char) PyBytes_AS_STRING(c)[0];
        }
#if PY_MAJOR_VERSION < 3
    } else if (PyUnicode_Check(c)) {
        return (long)__Pyx_PyUnicode_AsPy_UCS4(c);
#endif
#if (!CYTHON_COMPILING_IN_PYPY) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    } else if (PyByteArray_Check(c)) {
        size = PyByteArray_GET_SIZE(c);
        if (likely(size == 1)) {
            return (unsigned char) PyByteArray_AS_STRING(c)[0];
        }
#endif
    } else {
        PyErr_Format(PyExc_TypeError,
            "ord() expected string of length 1, but %.200s found", c->ob_type->tp_name);
        return (long)(Py_UCS4)-1;
    }
    PyErr_Format(PyExc_TypeError,
        "ord() expected a character, but string of length %zd found", size);
    return (long)(Py_UCS4)-1;
}

/* BufferIndexError */
static void __Pyx_RaiseBufferIndexError(int axis) {
  PyErr_Format(PyExc_IndexError,
     "Out of bounds on buffer access (axis %d)", axis);
}

/* MemviewSliceInit */
static int
__Pyx_init_memviewslice(struct __pyx_memoryview_obj *memview,
                        int ndim,
                        __Pyx_memviewslice *memviewslice,
                        int memview_is_new_reference)
{
    __Pyx_RefNannyDeclarations
    int i, retval=-1;
    Py_buffer *buf = &memview->view;
    __Pyx_RefNannySetupContext("init_memviewslice", 0);
    if (memviewslice->memview || memviewslice->data) {
        PyErr_SetString(PyExc_ValueError,
            "memviewslice is already initialized!");
        goto fail;
    }
    if (buf->strides) {
        for (i = 0; i < ndim; i++) {
            memviewslice->strides[i] = buf->strides[i];
        }
    } else {
        Py_ssize_t stride = buf->itemsize;
        for (i = ndim - 1; i >= 0; i--) {
            memviewslice->strides[i] = stride;
            stride *= buf->shape[i];
        }
    }
    for (i = 0; i < ndim; i++) {
        memviewslice->shape[i]   = buf->shape[i];
        if (buf->suboffsets) {
            memviewslice->suboffsets[i] = buf->suboffsets[i];
        } else {
            memviewslice->suboffsets[i] = -1;
        }
    }
    memviewslice->memview = memview;
    memviewslice->data = (char *)buf->buf;
    if (__pyx_add_acquisition_count(memview) == 0 && !memview_is_new_reference) {
        Py_INCREF(memview);
    }
    retval = 0;
    goto no_fail;
fail:
    memviewslice->memview = 0;
    memviewslice->data = 0;
    retval = -1;
no_fail:
    __Pyx_RefNannyFinishContext();
    return retval;
}
#ifndef Py_NO_RETURN
#define Py_NO_RETURN
#endif
static void __pyx_fatalerror(const char *fmt, ...) Py_NO_RETURN {
    va_list vargs;
    char msg[200];
#ifdef HAVE_STDARG_PROTOTYPES
    va_start(vargs, fmt);
#else
    va_start(vargs);
#endif
    vsnprintf(msg, 200, fmt, vargs);
    va_end(vargs);
    Py_FatalError(msg);
}
static CYTHON_INLINE int
__pyx_add_acquisition_count_locked(__pyx_atomic_int *acquisition_count,
                                   PyThread_type_lock lock)
{
    int result;
    PyThread_acquire_lock(lock, 1);
    result = (*acquisition_count)++;
    PyThread_release_lock(lock);
    return result;
}
static CYTHON_INLINE int
__pyx_sub_acquisition_count_locked(__pyx_atomic_int *acquisition_count,
                                   PyThread_type_lock lock)
{
    int result;
    PyThread_acquire_lock(lock, 1);
    result = (*acquisition_count)--;
    PyThread_release_lock(lock);
    return result;
}
static CYTHON_INLINE void
__Pyx_INC_MEMVIEW(__Pyx_memviewslice *memslice, int have_gil, int lineno)
{
    int first_time;
    struct __pyx_memoryview_obj *memview = memslice->memview;
    if (!memview || (PyObject *) memview == Py_None)
        return;
    if (__pyx_get_slice_count(memview) < 0)
        __pyx_fatalerror("Acquisition count is %d (line %d)",
                         __pyx_get_slice_count(memview), lineno);
    first_time = __pyx_add_acquisition_count(memview) == 0;
    if (first_time) {
        if (have_gil) {
            Py_INCREF((PyObject *) memview);
        } else {
            PyGILState_STATE _gilstate = PyGILState_Ensure();
            Py_INCREF((PyObject *) memview);
            PyGILState_Release(_gilstate);
        }
    }
}
static CYTHON_INLINE void __Pyx_XDEC_MEMVIEW(__Pyx_memviewslice *memslice,
                                             int have_gil, int lineno) {
    int last_time;
    struct __pyx_memoryview_obj *memview = memslice->memview;
    if (!memview ) {
        return;
    } else if ((PyObject *) memview == Py_None) {
        memslice->memview = NULL;
        return;
    }
    if (__pyx_get_slice_count(memview) <= 0)
        __pyx_fatalerror("Acquisition count is %d (line %d)",
                         __pyx_get_slice_count(memview), lineno);
    last_time = __pyx_sub_acquisition_count(memview) == 1;
    memslice->data = NULL;
    if (last_time) {
        if (have_gil) {
            Py_CLEAR(memslice->memview);
        } else {
            PyGILState_STATE _gilstate = PyGILState_Ensure();
            Py_CLEAR(memslice->memview);
            PyGILState_Release(_gilstate);
        }
    } else {
        memslice->memview = NULL;
    }
}

/* IsLittleEndian */
static CYTHON_INLINE int __Pyx_Is_Little_Endian(void)
{
  union {
    uint32_t u32;
    uint8_t u8[4];
  } S;
  S.u32 = 0x01020304;
  return S.u8[0] == 4;
}

/* BufferFormatCheck */
static void __Pyx_BufFmt_Init(__Pyx_BufFmt_Context* ctx,
                              __Pyx_BufFmt_StackElem* stack,
                              __Pyx_TypeInfo* type) {
  stack[0].field = &ctx->root;
  stack[0].parent_offset = 0;
  ctx->root.type = type;
  ctx->root.name = "buffer dtype";
  ctx->root.offset = 0;
  ctx->head = stack;
  ctx->head->field = &ctx->root;
  ctx->fmt_offset = 0;
  ctx->head->parent_offset = 0;
  ctx->new_packmode = '@';
  ctx->enc_packmode = '@';
  ctx->new_count = 1;
  ctx->enc_count = 0;
  ctx->enc_type = 0;
  ctx->is_complex = 0;
  ctx->is_valid_array = 0;
  ctx->struct_alignment = 0;
  while (type->typegroup == 'S') {
    ++ctx->head;
    ctx->head->field = type->fields;
    ctx->head->parent_offset = 0;
    type = type->fields->type;
  }
}
static int __Pyx_BufFmt_ParseNumber(const char** ts) {
    int count;
    const char* t = *ts;
    if (*t < '0' || *t > '9') {
      return -1;
    } else {
        count = *t++ - '0';
        while (*t >= '0' && *t <= '9') {
            count *= 10;
            count += *t++ - '0';
        }
    }
    *ts = t;
    return count;
}
static int __Pyx_BufFmt_ExpectNumber(const char **ts) {
    int number = __Pyx_BufFmt_ParseNumber(ts);
    if (number == -1)
        PyErr_Format(PyExc_ValueError,\
                     "Does not understand character buffer dtype format string ('%c')", **ts);
    return number;
}
static void __Pyx_BufFmt_RaiseUnexpectedChar(char ch) {
  PyErr_Format(PyExc_ValueError,
               "Unexpected format string character: '%c'", ch);
}
static const char* __Pyx_BufFmt_DescribeTypeChar(char ch, int is_complex) {
  switch (ch) {
    case 'c': return "'char'";
    case 'b': return "'signed char'";
    case 'B': return "'unsigned char'";
    case 'h': return "'short'";
    case 'H': return "'unsigned short'";
    case 'i': return "'int'";
    case 'I': return "'unsigned int'";
    case 'l': return "'long'";
    case 'L': return "'unsigned long'";
    case 'q': return "'long long'";
    case 'Q': return "'unsigned long long'";
    case 'f': return (is_complex ? "'complex float'" : "'float'");
    case 'd': return (is_complex ? "'complex double'" : "'double'");
    case 'g': return (is_complex ? "'complex long double'" : "'long double'");
    case 'T': return "a struct";
    case 'O': return "Python object";
    case 'P': return "a pointer";
    case 's': case 'p': return "a string";
    case 0: return "end";
    default: return "unparseable format string";
  }
}
static size_t __Pyx_BufFmt_TypeCharToStandardSize(char ch, int is_complex) {
  switch (ch) {
    case '?': case 'c': case 'b': case 'B': case 's': case 'p': return 1;
    case 'h': case 'H': return 2;
    case 'i': case 'I': case 'l': case 'L': return 4;
    case 'q': case 'Q': return 8;
    case 'f': return (is_complex ? 8 : 4);
    case 'd': return (is_complex ? 16 : 8);
    case 'g': {
      PyErr_SetString(PyExc_ValueError, "Python does not define a standard format string size for long double ('g')..");
      return 0;
    }
    case 'O': case 'P': return sizeof(void*);
    default:
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
}
static size_t __Pyx_BufFmt_TypeCharToNativeSize(char ch, int is_complex) {
  switch (ch) {
    case 'c': case 'b': case 'B': case 's': case 'p': return 1;
    case 'h': case 'H': return sizeof(short);
    case 'i': case 'I': return sizeof(int);
    case 'l': case 'L': return sizeof(long);
    #ifdef HAVE_LONG_LONG
    case 'q': case 'Q': return sizeof(PY_LONG_LONG);
    #endif
    case 'f': return sizeof(float) * (is_complex ? 2 : 1);
    case 'd': return sizeof(double) * (is_complex ? 2 : 1);
    case 'g': return sizeof(long double) * (is_complex ? 2 : 1);
    case 'O': case 'P': return sizeof(void*);
    default: {
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
  }
}
typedef struct { char c; short x; } __Pyx_st_short;
typedef struct { char c; int x; } __Pyx_st_int;
typedef struct { char c; long x; } __Pyx_st_long;
typedef struct { char c; float x; } __Pyx_st_float;
typedef struct { char c; double x; } __Pyx_st_double;
typedef struct { char c; long double x; } __Pyx_st_longdouble;
typedef struct { char c; void *x; } __Pyx_st_void_p;
#ifdef HAVE_LONG_LONG
typedef struct { char c; PY_LONG_LONG x; } __Pyx_st_longlong;
#endif
static size_t __Pyx_BufFmt_TypeCharToAlignment(char ch, CYTHON_UNUSED int is_complex) {
  switch (ch) {
    case '?': case 'c': case 'b': case 'B': case 's': case 'p': return 1;
    case 'h': case 'H': return sizeof(__Pyx_st_short) - sizeof(short);
    case 'i': case 'I': return sizeof(__Pyx_st_int) - sizeof(int);
    case 'l': case 'L': return sizeof(__Pyx_st_long) - sizeof(long);
#ifdef HAVE_LONG_LONG
    case 'q': case 'Q': return sizeof(__Pyx_st_longlong) - sizeof(PY_LONG_LONG);
#endif
    case 'f': return sizeof(__Pyx_st_float) - sizeof(float);
    case 'd': return sizeof(__Pyx_st_double) - sizeof(double);
    case 'g': return sizeof(__Pyx_st_longdouble) - sizeof(long double);
    case 'P': case 'O': return sizeof(__Pyx_st_void_p) - sizeof(void*);
    default:
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
}
/* These are for computing the padding at the end of the struct to align
   on the first member of the struct. This will probably the same as above,
   but we don't have any guarantees.
 */
typedef struct { short x; char c; } __Pyx_pad_short;
typedef struct { int x; char c; } __Pyx_pad_int;
typedef struct { long x; char c; } __Pyx_pad_long;
typedef struct { float x; char c; } __Pyx_pad_float;
typedef struct { double x; char c; } __Pyx_pad_double;
typedef struct { long double x; char c; } __Pyx_pad_longdouble;
typedef struct { void *x; char c; } __Pyx_pad_void_p;
#ifdef HAVE_LONG_LONG
typedef struct { PY_LONG_LONG x; char c; } __Pyx_pad_longlong;
#endif
static size_t __Pyx_BufFmt_TypeCharToPadding(char ch, CYTHON_UNUSED int is_complex) {
  switch (ch) {
    case '?': case 'c': case 'b': case 'B': case 's': case 'p': return 1;
    case 'h': case 'H': return sizeof(__Pyx_pad_short) - sizeof(short);
    case 'i': case 'I': return sizeof(__Pyx_pad_int) - sizeof(int);
    case 'l': case 'L': return sizeof(__Pyx_pad_long) - sizeof(long);
#ifdef HAVE_LONG_LONG
    case 'q': case 'Q': return sizeof(__Pyx_pad_longlong) - sizeof(PY_LONG_LONG);
#endif
    case 'f': return sizeof(__Pyx_pad_float) - sizeof(float);
    case 'd': return sizeof(__Pyx_pad_double) - sizeof(double);
    case 'g': return sizeof(__Pyx_pad_longdouble) - sizeof(long double);
    case 'P': case 'O': return sizeof(__Pyx_pad_void_p) - sizeof(void*);
    default:
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
}
static char __Pyx_BufFmt_TypeCharToGroup(char ch, int is_complex) {
  switch (ch) {
    case 'c':
        return 'H';
    case 'b': case 'h': case 'i':
    case 'l': case 'q': case 's': case 'p':
        return 'I';
    case 'B': case 'H': case 'I': case 'L': case 'Q':
        return 'U';
    case 'f': case 'd': case 'g':
        return (is_complex ? 'C' : 'R');
    case 'O':
        return 'O';
    case 'P':
        return 'P';
    default: {
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
  }
}
static void __Pyx_BufFmt_RaiseExpected(__Pyx_BufFmt_Context* ctx) {
  if (ctx->head == NULL || ctx->head->field == &ctx->root) {
    const char* expected;
    const char* quote;
    if (ctx->head == NULL) {
      expected = "end";
      quote = "";
    } else {
      expected = ctx->head->field->type->name;
      quote = "'";
    }
    PyErr_Format(PyExc_ValueError,
                 "Buffer dtype mismatch, expected %s%s%s but got %s",
                 quote, expected, quote,
                 __Pyx_BufFmt_DescribeTypeChar(ctx->enc_type, ctx->is_complex));
  } else {
    __Pyx_StructField* field = ctx->head->field;
    __Pyx_StructField* parent = (ctx->head - 1)->field;
    PyErr_Format(PyExc_ValueError,
                 "Buffer dtype mismatch, expected '%s' but got %s in '%s.%s'",
                 field->type->name, __Pyx_BufFmt_DescribeTypeChar(ctx->enc_type, ctx->is_complex),
                 parent->type->name, field->name);
  }
}
static int __Pyx_BufFmt_ProcessTypeChunk(__Pyx_BufFmt_Context* ctx) {
  char group;
  size_t size, offset, arraysize = 1;
  if (ctx->enc_type == 0) return 0;
  if (ctx->head->field->type->arraysize[0]) {
    int i, ndim = 0;
    if (ctx->enc_type == 's' || ctx->enc_type == 'p') {
        ctx->is_valid_array = ctx->head->field->type->ndim == 1;
        ndim = 1;
        if (ctx->enc_count != ctx->head->field->type->arraysize[0]) {
            PyErr_Format(PyExc_ValueError,
                         "Expected a dimension of size %zu, got %zu",
                         ctx->head->field->type->arraysize[0], ctx->enc_count);
            return -1;
        }
    }
    if (!ctx->is_valid_array) {
      PyErr_Format(PyExc_ValueError, "Expected %d dimensions, got %d",
                   ctx->head->field->type->ndim, ndim);
      return -1;
    }
    for (i = 0; i < ctx->head->field->type->ndim; i++) {
      arraysize *= ctx->head->field->type->arraysize[i];
    }
    ctx->is_valid_array = 0;
    ctx->enc_count = 1;
  }
  group = __Pyx_BufFmt_TypeCharToGroup(ctx->enc_type, ctx->is_complex);
  do {
    __Pyx_StructField* field = ctx->head->field;
    __Pyx_TypeInfo* type = field->type;
    if (ctx->enc_packmode == '@' || ctx->enc_packmode == '^') {
      size = __Pyx_BufFmt_TypeCharToNativeSize(ctx->enc_type, ctx->is_complex);
    } else {
      size = __Pyx_BufFmt_TypeCharToStandardSize(ctx->enc_type, ctx->is_complex);
    }
    if (ctx->enc_packmode == '@') {
      size_t align_at = __Pyx_BufFmt_TypeCharToAlignment(ctx->enc_type, ctx->is_complex);
      size_t align_mod_offset;
      if (align_at == 0) return -1;
      align_mod_offset = ctx->fmt_offset % align_at;
      if (align_mod_offset > 0) ctx->fmt_offset += align_at - align_mod_offset;
      if (ctx->struct_alignment == 0)
          ctx->struct_alignment = __Pyx_BufFmt_TypeCharToPadding(ctx->enc_type,
                                                                 ctx->is_complex);
    }
    if (type->size != size || type->typegroup != group) {
      if (type->typegroup == 'C' && type->fields != NULL) {
        size_t parent_offset = ctx->head->parent_offset + field->offset;
        ++ctx->head;
        ctx->head->field = type->fields;
        ctx->head->parent_offset = parent_offset;
        continue;
      }
      if ((type->typegroup == 'H' || group == 'H') && type->size == size) {
      } else {
          __Pyx_BufFmt_RaiseExpected(ctx);
          return -1;
      }
    }
    offset = ctx->head->parent_offset + field->offset;
    if (ctx->fmt_offset != offset) {
      PyErr_Format(PyExc_ValueError,
                   "Buffer dtype mismatch; next field is at offset %" CYTHON_FORMAT_SSIZE_T "d but %" CYTHON_FORMAT_SSIZE_T "d expected",
                   (Py_ssize_t)ctx->fmt_offset, (Py_ssize_t)offset);
      return -1;
    }
    ctx->fmt_offset += size;
    if (arraysize)
      ctx->fmt_offset += (arraysize - 1) * size;
    --ctx->enc_count;
    while (1) {
      if (field == &ctx->root) {
        ctx->head = NULL;
        if (ctx->enc_count != 0) {
          __Pyx_BufFmt_RaiseExpected(ctx);
          return -1;
        }
        break;
      }
      ctx->head->field = ++field;
      if (field->type == NULL) {
        --ctx->head;
        field = ctx->head->field;
        continue;
      } else if (field->type->typegroup == 'S') {
        size_t parent_offset = ctx->head->parent_offset + field->offset;
        if (field->type->fields->type == NULL) continue;
        field = field->type->fields;
        ++ctx->head;
        ctx->head->field = field;
        ctx->head->parent_offset = parent_offset;
        break;
      } else {
        break;
      }
    }
  } while (ctx->enc_count);
  ctx->enc_type = 0;
  ctx->is_complex = 0;
  return 0;
}
static PyObject *
__pyx_buffmt_parse_array(__Pyx_BufFmt_Context* ctx, const char** tsp)
{
    const char *ts = *tsp;
    int i = 0, number;
    int ndim = ctx->head->field->type->ndim;
;
    ++ts;
    if (ctx->new_count != 1) {
        PyErr_SetString(PyExc_ValueError,
                        "Cannot handle repeated arrays in format string");
        return NULL;
    }
    if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
    while (*ts && *ts != ')') {
        switch (*ts) {
            case ' ': case '\f': case '\r': case '\n': case '\t': case '\v':  continue;
            default:  break;
        }
        number = __Pyx_BufFmt_ExpectNumber(&ts);
        if (number == -1) return NULL;
        if (i < ndim && (size_t) number != ctx->head->field->type->arraysize[i])
            return PyErr_Format(PyExc_ValueError,
                        "Expected a dimension of size %zu, got %d",
                        ctx->head->field->type->arraysize[i], number);
        if (*ts != ',' && *ts != ')')
            return PyErr_Format(PyExc_ValueError,
                                "Expected a comma in format string, got '%c'", *ts);
        if (*ts == ',') ts++;
        i++;
    }
    if (i != ndim)
        return PyErr_Format(PyExc_ValueError, "Expected %d dimension(s), got %d",
                            ctx->head->field->type->ndim, i);
    if (!*ts) {
        PyErr_SetString(PyExc_ValueError,
                        "Unexpected end of format string, expected ')'");
        return NULL;
    }
    ctx->is_valid_array = 1;
    ctx->new_count = 1;
    *tsp = ++ts;
    return Py_None;
}
static const char* __Pyx_BufFmt_CheckString(__Pyx_BufFmt_Context* ctx, const char* ts) {
  int got_Z = 0;
  while (1) {
    switch(*ts) {
      case 0:
        if (ctx->enc_type != 0 && ctx->head == NULL) {
          __Pyx_BufFmt_RaiseExpected(ctx);
          return NULL;
        }
        if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
        if (ctx->head != NULL) {
          __Pyx_BufFmt_RaiseExpected(ctx);
          return NULL;
        }
        return ts;
      case ' ':
      case '\r':
      case '\n':
        ++ts;
        break;
      case '<':
        if (!__Pyx_Is_Little_Endian()) {
          PyErr_SetString(PyExc_ValueError, "Little-endian buffer not supported on big-endian compiler");
          return NULL;
        }
        ctx->new_packmode = '=';
        ++ts;
        break;
      case '>':
      case '!':
        if (__Pyx_Is_Little_Endian()) {
          PyErr_SetString(PyExc_ValueError, "Big-endian buffer not supported on little-endian compiler");
          return NULL;
        }
        ctx->new_packmode = '=';
        ++ts;
        break;
      case '=':
      case '@':
      case '^':
        ctx->new_packmode = *ts++;
        break;
      case 'T':
        {
          const char* ts_after_sub;
          size_t i, struct_count = ctx->new_count;
          size_t struct_alignment = ctx->struct_alignment;
          ctx->new_count = 1;
          ++ts;
          if (*ts != '{') {
            PyErr_SetString(PyExc_ValueError, "Buffer acquisition: Expected '{' after 'T'");
            return NULL;
          }
          if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
          ctx->enc_type = 0;
          ctx->enc_count = 0;
          ctx->struct_alignment = 0;
          ++ts;
          ts_after_sub = ts;
          for (i = 0; i != struct_count; ++i) {
            ts_after_sub = __Pyx_BufFmt_CheckString(ctx, ts);
            if (!ts_after_sub) return NULL;
          }
          ts = ts_after_sub;
          if (struct_alignment) ctx->struct_alignment = struct_alignment;
        }
        break;
      case '}':
        {
          size_t alignment = ctx->struct_alignment;
          ++ts;
          if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
          ctx->enc_type = 0;
          if (alignment && ctx->fmt_offset % alignment) {
            ctx->fmt_offset += alignment - (ctx->fmt_offset % alignment);
          }
        }
        return ts;
      case 'x':
        if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
        ctx->fmt_offset += ctx->new_count;
        ctx->new_count = 1;
        ctx->enc_count = 0;
        ctx->enc_type = 0;
        ctx->enc_packmode = ctx->new_packmode;
        ++ts;
        break;
      case 'Z':
        got_Z = 1;
        ++ts;
        if (*ts != 'f' && *ts != 'd' && *ts != 'g') {
          __Pyx_BufFmt_RaiseUnexpectedChar('Z');
          return NULL;
        }
        CYTHON_FALLTHROUGH;
      case 'c': case 'b': case 'B': case 'h': case 'H': case 'i': case 'I':
      case 'l': case 'L': case 'q': case 'Q':
      case 'f': case 'd': case 'g':
      case 'O': case 'p':
        if (ctx->enc_type == *ts && got_Z == ctx->is_complex &&
            ctx->enc_packmode == ctx->new_packmode) {
          ctx->enc_count += ctx->new_count;
          ctx->new_count = 1;
          got_Z = 0;
          ++ts;
          break;
        }
        CYTHON_FALLTHROUGH;
      case 's':
        if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
        ctx->enc_count = ctx->new_count;
        ctx->enc_packmode = ctx->new_packmode;
        ctx->enc_type = *ts;
        ctx->is_complex = got_Z;
        ++ts;
        ctx->new_count = 1;
        got_Z = 0;
        break;
      case ':':
        ++ts;
        while(*ts != ':') ++ts;
        ++ts;
        break;
      case '(':
        if (!__pyx_buffmt_parse_array(ctx, &ts)) return NULL;
        break;
      default:
        {
          int number = __Pyx_BufFmt_ExpectNumber(&ts);
          if (number == -1) return NULL;
          ctx->new_count = (size_t)number;
        }
    }
  }
}

/* BufferGetAndValidate */
  static CYTHON_INLINE void __Pyx_SafeReleaseBuffer(Py_buffer* info) {
  if (unlikely(info->buf == NULL)) return;
  if (info->suboffsets == __Pyx_minusones) info->suboffsets = NULL;
  __Pyx_ReleaseBuffer(info);
}
static void __Pyx_ZeroBuffer(Py_buffer* buf) {
  buf->buf = NULL;
  buf->obj = NULL;
  buf->strides = __Pyx_zeros;
  buf->shape = __Pyx_zeros;
  buf->suboffsets = __Pyx_minusones;
}
static int __Pyx__GetBufferAndValidate(
        Py_buffer* buf, PyObject* obj,  __Pyx_TypeInfo* dtype, int flags,
        int nd, int cast, __Pyx_BufFmt_StackElem* stack)
{
  buf->buf = NULL;
  if (unlikely(__Pyx_GetBuffer(obj, buf, flags) == -1)) {
    __Pyx_ZeroBuffer(buf);
    return -1;
  }
  if (unlikely(buf->ndim != nd)) {
    PyErr_Format(PyExc_ValueError,
                 "Buffer has wrong number of dimensions (expected %d, got %d)",
                 nd, buf->ndim);
    goto fail;
  }
  if (!cast) {
    __Pyx_BufFmt_Context ctx;
    __Pyx_BufFmt_Init(&ctx, stack, dtype);
    if (!__Pyx_BufFmt_CheckString(&ctx, buf->format)) goto fail;
  }
  if (unlikely((size_t)buf->itemsize != dtype->size)) {
    PyErr_Format(PyExc_ValueError,
      "Item size of buffer (%" CYTHON_FORMAT_SSIZE_T "d byte%s) does not match size of '%s' (%" CYTHON_FORMAT_SSIZE_T "d byte%s)",
      buf->itemsize, (buf->itemsize > 1) ? "s" : "",
      dtype->name, (Py_ssize_t)dtype->size, (dtype->size > 1) ? "s" : "");
    goto fail;
  }
  if (buf->suboffsets == NULL) buf->suboffsets = __Pyx_minusones;
  return 0;
fail:;
  __Pyx_SafeReleaseBuffer(buf);
  return -1;
}

/* BufferFallbackError */
  static void __Pyx_RaiseBufferFallbackError(void) {
  PyErr_SetString(PyExc_ValueError,
     "Buffer acquisition failed on assignment; and then reacquiring the old buffer failed too!");
}

/* PyObjectFormatAndDecref */
  static CYTHON_INLINE PyObject* __Pyx_PyObject_FormatSimpleAndDecref(PyObject* s, PyObject* f) {
    if (unlikely(!s)) return NULL;
    if (likely(PyUnicode_CheckExact(s))) return s;
    #if PY_MAJOR_VERSION < 3
    if (likely(PyString_CheckExact(s))) {
        PyObject *result = PyUnicode_FromEncodedObject(s, NULL, "strict");
        Py_DECREF(s);
        return result;
    }
    #endif
    return __Pyx_PyObject_FormatAndDecref(s, f);
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_FormatAndDecref(PyObject* s, PyObject* f) {
    PyObject *result = PyObject_Format(s, f);
    Py_DECREF(s);
    return result;
}

/* JoinPyUnicode */
  static PyObject* __Pyx_PyUnicode_Join(PyObject* value_tuple, Py_ssize_t value_count, Py_ssize_t result_ulength,
                                      CYTHON_UNUSED Py_UCS4 max_char) {
#if CYTHON_USE_UNICODE_INTERNALS && CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    PyObject *result_uval;
    int result_ukind;
    Py_ssize_t i, char_pos;
    void *result_udata;
#if CYTHON_PEP393_ENABLED
    result_uval = PyUnicode_New(result_ulength, max_char);
    if (unlikely(!result_uval)) return NULL;
    result_ukind = (max_char <= 255) ? PyUnicode_1BYTE_KIND : (max_char <= 65535) ? PyUnicode_2BYTE_KIND : PyUnicode_4BYTE_KIND;
    result_udata = PyUnicode_DATA(result_uval);
#else
    result_uval = PyUnicode_FromUnicode(NULL, result_ulength);
    if (unlikely(!result_uval)) return NULL;
    result_ukind = sizeof(Py_UNICODE);
    result_udata = PyUnicode_AS_UNICODE(result_uval);
#endif
    char_pos = 0;
    for (i=0; i < value_count; i++) {
        int ukind;
        Py_ssize_t ulength;
        void *udata;
        PyObject *uval = PyTuple_GET_ITEM(value_tuple, i);
        if (unlikely(__Pyx_PyUnicode_READY(uval)))
            goto bad;
        ulength = __Pyx_PyUnicode_GET_LENGTH(uval);
        if (unlikely(!ulength))
            continue;
        if (unlikely(char_pos + ulength < 0))
            goto overflow;
        ukind = __Pyx_PyUnicode_KIND(uval);
        udata = __Pyx_PyUnicode_DATA(uval);
        if (!CYTHON_PEP393_ENABLED || ukind == result_ukind) {
            memcpy((char *)result_udata + char_pos * result_ukind, udata, (size_t) (ulength * result_ukind));
        } else {
            #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030300F0 || defined(_PyUnicode_FastCopyCharacters)
            _PyUnicode_FastCopyCharacters(result_uval, char_pos, uval, 0, ulength);
            #else
            Py_ssize_t j;
            for (j=0; j < ulength; j++) {
                Py_UCS4 uchar = __Pyx_PyUnicode_READ(ukind, udata, j);
                __Pyx_PyUnicode_WRITE(result_ukind, result_udata, char_pos+j, uchar);
            }
            #endif
        }
        char_pos += ulength;
    }
    return result_uval;
overflow:
    PyErr_SetString(PyExc_OverflowError, "join() result is too long for a Python string");
bad:
    Py_DECREF(result_uval);
    return NULL;
#else
    result_ulength++;
    value_count++;
    return PyUnicode_Join(__pyx_empty_unicode, value_tuple);
#endif
}

/* ImportFrom */
  static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name) {
    PyObject* value = __Pyx_PyObject_GetAttrStr(module, name);
    if (unlikely(!value) && PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Format(PyExc_ImportError,
        #if PY_MAJOR_VERSION < 3
            "cannot import name %.230s", PyString_AS_STRING(name));
        #else
            "cannot import name %S", name);
        #endif
    }
    return value;
}

/* WriteUnraisableException */
  static void __Pyx_WriteUnraisable(const char *name, CYTHON_UNUSED int clineno,
                                  CYTHON_UNUSED int lineno, CYTHON_UNUSED const char *filename,
                                  int full_traceback, CYTHON_UNUSED int nogil) {
    PyObject *old_exc, *old_val, *old_tb;
    PyObject *ctx;
    __Pyx_PyThreadState_declare
#ifdef WITH_THREAD
    PyGILState_STATE state;
    if (nogil)
        state = PyGILState_Ensure();
#ifdef _MSC_VER
    else state = (PyGILState_STATE)-1;
#endif
#endif
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&old_exc, &old_val, &old_tb);
    if (full_traceback) {
        Py_XINCREF(old_exc);
        Py_XINCREF(old_val);
        Py_XINCREF(old_tb);
        __Pyx_ErrRestore(old_exc, old_val, old_tb);
        PyErr_PrintEx(1);
    }
    #if PY_MAJOR_VERSION < 3
    ctx = PyString_FromString(name);
    #else
    ctx = PyUnicode_FromString(name);
    #endif
    __Pyx_ErrRestore(old_exc, old_val, old_tb);
    if (!ctx) {
        PyErr_WriteUnraisable(Py_None);
    } else {
        PyErr_WriteUnraisable(ctx);
        Py_DECREF(ctx);
    }
#ifdef WITH_THREAD
    if (nogil)
        PyGILState_Release(state);
#endif
}

/* None */
  static CYTHON_INLINE Py_ssize_t __Pyx_div_Py_ssize_t(Py_ssize_t a, Py_ssize_t b) {
    Py_ssize_t q = a / b;
    Py_ssize_t r = a - q*b;
    q -= ((r != 0) & ((r ^ b) < 0));
    return q;
}

/* decode_c_string */
  static CYTHON_INLINE PyObject* __Pyx_decode_c_string(
         const char* cstring, Py_ssize_t start, Py_ssize_t stop,
         const char* encoding, const char* errors,
         PyObject* (*decode_func)(const char *s, Py_ssize_t size, const char *errors)) {
    Py_ssize_t length;
    if (unlikely((start < 0) | (stop < 0))) {
        size_t slen = strlen(cstring);
        if (unlikely(slen > (size_t) PY_SSIZE_T_MAX)) {
            PyErr_SetString(PyExc_OverflowError,
                            "c-string too long to convert to Python");
            return NULL;
        }
        length = (Py_ssize_t) slen;
        if (start < 0) {
            start += length;
            if (start < 0)
                start = 0;
        }
        if (stop < 0)
            stop += length;
    }
    length = stop - start;
    if (unlikely(length <= 0))
        return PyUnicode_FromUnicode(NULL, 0);
    cstring += start;
    if (decode_func) {
        return decode_func(cstring, length, errors);
    } else {
        return PyUnicode_Decode(cstring, length, encoding, errors);
    }
}

/* SwapException */
  #if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = *type;
    exc_info->exc_value = *value;
    exc_info->exc_traceback = *tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = *type;
    tstate->exc_value = *value;
    tstate->exc_traceback = *tb;
    #endif
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    PyErr_GetExcInfo(&tmp_type, &tmp_value, &tmp_tb);
    PyErr_SetExcInfo(*type, *value, *tb);
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#endif

/* FastTypeChecks */
  #if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = a->tp_base;
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
#if PY_MAJOR_VERSION == 2
static int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject* exc_type2) {
    PyObject *exception, *value, *tb;
    int res;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&exception, &value, &tb);
    res = exc_type1 ? PyObject_IsSubclass(err, exc_type1) : 0;
    if (unlikely(res == -1)) {
        PyErr_WriteUnraisable(err);
        res = 0;
    }
    if (!res) {
        res = PyObject_IsSubclass(err, exc_type2);
        if (unlikely(res == -1)) {
            PyErr_WriteUnraisable(err);
            res = 0;
        }
    }
    __Pyx_ErrRestore(exception, value, tb);
    return res;
}
#else
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    int res = exc_type1 ? __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type1) : 0;
    if (!res) {
        res = __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
    return res;
}
#endif
static int __Pyx_PyErr_GivenExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    assert(PyExceptionClass_Check(exc_type));
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        PyObject *t = PyTuple_GET_ITEM(tuple, i);
        #if PY_MAJOR_VERSION < 3
        if (likely(exc_type == t)) return 1;
        #endif
        if (likely(PyExceptionClass_Check(t))) {
            if (__Pyx_inner_PyErr_GivenExceptionMatches2(exc_type, NULL, t)) return 1;
        } else {
        }
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        if (likely(PyExceptionClass_Check(exc_type))) {
            return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
        } else if (likely(PyTuple_Check(exc_type))) {
            return __Pyx_PyErr_GivenExceptionMatchesTuple(err, exc_type);
        } else {
        }
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    assert(PyExceptionClass_Check(exc_type1));
    assert(PyExceptionClass_Check(exc_type2));
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* PyIntBinop */
  #if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_AddObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, int inplace, int zerodivision_check) {
    (void)inplace;
    (void)zerodivision_check;
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long x;
        long a = PyInt_AS_LONG(op1);
            x = (long)((unsigned long)a + b);
            if (likely((x^a) >= 0 || (x^b) >= 0))
                return PyInt_FromLong(x);
            return PyLong_Type.tp_as_number->nb_add(op1, op2);
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        const long b = intval;
        long a, x;
#ifdef HAVE_LONG_LONG
        const PY_LONG_LONG llb = intval;
        PY_LONG_LONG lla, llx;
#endif
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        const Py_ssize_t size = Py_SIZE(op1);
        if (likely(__Pyx_sst_abs(size) <= 1)) {
            a = likely(size) ? digits[0] : 0;
            if (size == -1) a = -a;
        } else {
            switch (size) {
                case -2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                default: return PyLong_Type.tp_as_number->nb_add(op1, op2);
            }
        }
                x = a + b;
            return PyLong_FromLong(x);
#ifdef HAVE_LONG_LONG
        long_long:
                llx = lla + llb;
            return PyLong_FromLongLong(llx);
#endif
        
        
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
        double a = PyFloat_AS_DOUBLE(op1);
            double result;
            PyFPE_START_PROTECT("add", return NULL)
            result = ((double)a) + (double)b;
            PyFPE_END_PROTECT(result)
            return PyFloat_FromDouble(result);
    }
    return (inplace ? PyNumber_InPlaceAdd : PyNumber_Add)(op1, op2);
}
#endif

/* None */
  static CYTHON_INLINE void __Pyx_RaiseUnboundLocalError(const char *varname) {
    PyErr_Format(PyExc_UnboundLocalError, "local variable '%s' referenced before assignment", varname);
}

/* None */
  static CYTHON_INLINE long __Pyx_div_long(long a, long b) {
    long q = a / b;
    long r = a - q*b;
    q -= ((r != 0) & ((r ^ b) < 0));
    return q;
}

/* CallNextTpTraverse */
  static int __Pyx_call_next_tp_traverse(PyObject* obj, visitproc v, void *a, traverseproc current_tp_traverse) {
    PyTypeObject* type = Py_TYPE(obj);
    while (type && type->tp_traverse != current_tp_traverse)
        type = type->tp_base;
    while (type && type->tp_traverse == current_tp_traverse)
        type = type->tp_base;
    if (type && type->tp_traverse)
        return type->tp_traverse(obj, v, a);
    return 0;
}

/* CallNextTpClear */
  static void __Pyx_call_next_tp_clear(PyObject* obj, inquiry current_tp_clear) {
    PyTypeObject* type = Py_TYPE(obj);
    while (type && type->tp_clear != current_tp_clear)
        type = type->tp_base;
    while (type && type->tp_clear == current_tp_clear)
        type = type->tp_base;
    if (type && type->tp_clear)
        type->tp_clear(obj);
}

/* PyObject_GenericGetAttrNoDict */
  #if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject *__Pyx_RaiseGenericGetAttributeError(PyTypeObject *tp, PyObject *attr_name) {
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'%.50s' object has no attribute '%U'",
                 tp->tp_name, attr_name);
#else
                 "'%.50s' object has no attribute '%.400s'",
                 tp->tp_name, PyString_AS_STRING(attr_name));
#endif
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name) {
    PyObject *descr;
    PyTypeObject *tp = Py_TYPE(obj);
    if (unlikely(!PyString_Check(attr_name))) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    assert(!tp->tp_dictoffset);
    descr = _PyType_Lookup(tp, attr_name);
    if (unlikely(!descr)) {
        return __Pyx_RaiseGenericGetAttributeError(tp, attr_name);
    }
    Py_INCREF(descr);
    #if PY_MAJOR_VERSION < 3
    if (likely(PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_HAVE_CLASS)))
    #endif
    {
        descrgetfunc f = Py_TYPE(descr)->tp_descr_get;
        if (unlikely(f)) {
            PyObject *res = f(descr, obj, (PyObject *)tp);
            Py_DECREF(descr);
            return res;
        }
    }
    return descr;
}
#endif

/* PyObject_GenericGetAttr */
  #if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name) {
    if (unlikely(Py_TYPE(obj)->tp_dictoffset)) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    return __Pyx_PyObject_GenericGetAttrNoDict(obj, attr_name);
}
#endif

/* SetupReduce */
  static int __Pyx_setup_reduce_is_named(PyObject* meth, PyObject* name) {
  int ret;
  PyObject *name_attr;
  name_attr = __Pyx_PyObject_GetAttrStr(meth, __pyx_n_s_name_2);
  if (likely(name_attr)) {
      ret = PyObject_RichCompareBool(name_attr, name, Py_EQ);
  } else {
      ret = -1;
  }
  if (unlikely(ret < 0)) {
      PyErr_Clear();
      ret = 0;
  }
  Py_XDECREF(name_attr);
  return ret;
}
static int __Pyx_setup_reduce(PyObject* type_obj) {
    int ret = 0;
    PyObject *object_reduce = NULL;
    PyObject *object_reduce_ex = NULL;
    PyObject *reduce = NULL;
    PyObject *reduce_ex = NULL;
    PyObject *reduce_cython = NULL;
    PyObject *setstate = NULL;
    PyObject *setstate_cython = NULL;
#if CYTHON_USE_PYTYPE_LOOKUP
    if (_PyType_Lookup((PyTypeObject*)type_obj, __pyx_n_s_getstate)) goto GOOD;
#else
    if (PyObject_HasAttr(type_obj, __pyx_n_s_getstate)) goto GOOD;
#endif
#if CYTHON_USE_PYTYPE_LOOKUP
    object_reduce_ex = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto BAD;
#else
    object_reduce_ex = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto BAD;
#endif
    reduce_ex = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce_ex); if (unlikely(!reduce_ex)) goto BAD;
    if (reduce_ex == object_reduce_ex) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_reduce = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto BAD;
#else
        object_reduce = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto BAD;
#endif
        reduce = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce); if (unlikely(!reduce)) goto BAD;
        if (reduce == object_reduce || __Pyx_setup_reduce_is_named(reduce, __pyx_n_s_reduce_cython)) {
            reduce_cython = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce_cython); if (unlikely(!reduce_cython)) goto BAD;
            ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce, reduce_cython); if (unlikely(ret < 0)) goto BAD;
            ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce_cython); if (unlikely(ret < 0)) goto BAD;
            setstate = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_setstate);
            if (!setstate) PyErr_Clear();
            if (!setstate || __Pyx_setup_reduce_is_named(setstate, __pyx_n_s_setstate_cython)) {
                setstate_cython = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_setstate_cython); if (unlikely(!setstate_cython)) goto BAD;
                ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate, setstate_cython); if (unlikely(ret < 0)) goto BAD;
                ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate_cython); if (unlikely(ret < 0)) goto BAD;
            }
            PyType_Modified((PyTypeObject*)type_obj);
        }
    }
    goto GOOD;
BAD:
    if (!PyErr_Occurred())
        PyErr_Format(PyExc_RuntimeError, "Unable to initialize pickling for %s", ((PyTypeObject*)type_obj)->tp_name);
    ret = -1;
GOOD:
#if !CYTHON_USE_PYTYPE_LOOKUP
    Py_XDECREF(object_reduce);
    Py_XDECREF(object_reduce_ex);
#endif
    Py_XDECREF(reduce);
    Py_XDECREF(reduce_ex);
    Py_XDECREF(reduce_cython);
    Py_XDECREF(setstate);
    Py_XDECREF(setstate_cython);
    return ret;
}

/* SetVTable */
  static int __Pyx_SetVtable(PyObject *dict, void *vtable) {
#if PY_VERSION_HEX >= 0x02070000
    PyObject *ob = PyCapsule_New(vtable, 0, 0);
#else
    PyObject *ob = PyCObject_FromVoidPtr(vtable, 0);
#endif
    if (!ob)
        goto bad;
    if (PyDict_SetItem(dict, __pyx_n_s_pyx_vtable, ob) < 0)
        goto bad;
    Py_DECREF(ob);
    return 0;
bad:
    Py_XDECREF(ob);
    return -1;
}

/* TypeImport */
  #ifndef __PYX_HAVE_RT_ImportType
#define __PYX_HAVE_RT_ImportType
static PyTypeObject *__Pyx_ImportType(PyObject *module, const char *module_name, const char *class_name,
    size_t size, enum __Pyx_ImportType_CheckSize check_size)
{
    PyObject *result = 0;
    char warning[200];
    Py_ssize_t basicsize;
#ifdef Py_LIMITED_API
    PyObject *py_basicsize;
#endif
    result = PyObject_GetAttrString(module, class_name);
    if (!result)
        goto bad;
    if (!PyType_Check(result)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s.%.200s is not a type object",
            module_name, class_name);
        goto bad;
    }
#ifndef Py_LIMITED_API
    basicsize = ((PyTypeObject *)result)->tp_basicsize;
#else
    py_basicsize = PyObject_GetAttrString(result, "__basicsize__");
    if (!py_basicsize)
        goto bad;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = 0;
    if (basicsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
#endif
    if ((size_t)basicsize < size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    if (check_size == __Pyx_ImportType_CheckSize_Error && (size_t)basicsize != size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    else if (check_size == __Pyx_ImportType_CheckSize_Warn && (size_t)basicsize > size) {
        PyOS_snprintf(warning, sizeof(warning),
            "%s.%s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        if (PyErr_WarnEx(NULL, warning, 0) < 0) goto bad;
    }
    return (PyTypeObject *)result;
bad:
    Py_XDECREF(result);
    return NULL;
}
#endif

/* GetVTable */
  static void* __Pyx_GetVtable(PyObject *dict) {
    void* ptr;
    PyObject *ob = PyObject_GetItem(dict, __pyx_n_s_pyx_vtable);
    if (!ob)
        goto bad;
#if PY_VERSION_HEX >= 0x02070000
    ptr = PyCapsule_GetPointer(ob, 0);
#else
    ptr = PyCObject_AsVoidPtr(ob);
#endif
    if (!ptr && !PyErr_Occurred())
        PyErr_SetString(PyExc_RuntimeError, "invalid vtable found for imported type");
    Py_DECREF(ob);
    return ptr;
bad:
    Py_XDECREF(ob);
    return NULL;
}

/* pyfrozenset_new */
  static CYTHON_INLINE PyObject* __Pyx_PyFrozenSet_New(PyObject* it) {
    if (it) {
        PyObject* result;
#if CYTHON_COMPILING_IN_PYPY
        PyObject* args;
        args = PyTuple_Pack(1, it);
        if (unlikely(!args))
            return NULL;
        result = PyObject_Call((PyObject*)&PyFrozenSet_Type, args, NULL);
        Py_DECREF(args);
        return result;
#else
        if (PyFrozenSet_CheckExact(it)) {
            Py_INCREF(it);
            return it;
        }
        result = PyFrozenSet_New(it);
        if (unlikely(!result))
            return NULL;
        if (likely(PySet_GET_SIZE(result)))
            return result;
        Py_DECREF(result);
#endif
    }
#if CYTHON_USE_TYPE_SLOTS
    return PyFrozenSet_Type.tp_new(&PyFrozenSet_Type, __pyx_empty_tuple, NULL);
#else
    return PyObject_Call((PyObject*)&PyFrozenSet_Type, __pyx_empty_tuple, NULL);
#endif
}

/* FetchCommonType */
  static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type) {
    PyObject* fake_module;
    PyTypeObject* cached_type = NULL;
    fake_module = PyImport_AddModule((char*) "_cython_" CYTHON_ABI);
    if (!fake_module) return NULL;
    Py_INCREF(fake_module);
    cached_type = (PyTypeObject*) PyObject_GetAttrString(fake_module, type->tp_name);
    if (cached_type) {
        if (!PyType_Check((PyObject*)cached_type)) {
            PyErr_Format(PyExc_TypeError,
                "Shared Cython type %.200s is not a type object",
                type->tp_name);
            goto bad;
        }
        if (cached_type->tp_basicsize != type->tp_basicsize) {
            PyErr_Format(PyExc_TypeError,
                "Shared Cython type %.200s has the wrong size, try recompiling",
                type->tp_name);
            goto bad;
        }
    } else {
        if (!PyErr_ExceptionMatches(PyExc_AttributeError)) goto bad;
        PyErr_Clear();
        if (PyType_Ready(type) < 0) goto bad;
        if (PyObject_SetAttrString(fake_module, type->tp_name, (PyObject*) type) < 0)
            goto bad;
        Py_INCREF(type);
        cached_type = type;
    }
done:
    Py_DECREF(fake_module);
    return cached_type;
bad:
    Py_XDECREF(cached_type);
    cached_type = NULL;
    goto done;
}

/* CythonFunction */
  #include <structmember.h>
static PyObject *
__Pyx_CyFunction_get_doc(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *closure)
{
    if (unlikely(op->func_doc == NULL)) {
        if (op->func.m_ml->ml_doc) {
#if PY_MAJOR_VERSION >= 3
            op->func_doc = PyUnicode_FromString(op->func.m_ml->ml_doc);
#else
            op->func_doc = PyString_FromString(op->func.m_ml->ml_doc);
#endif
            if (unlikely(op->func_doc == NULL))
                return NULL;
        } else {
            Py_INCREF(Py_None);
            return Py_None;
        }
    }
    Py_INCREF(op->func_doc);
    return op->func_doc;
}
static int
__Pyx_CyFunction_set_doc(__pyx_CyFunctionObject *op, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp = op->func_doc;
    if (value == NULL) {
        value = Py_None;
    }
    Py_INCREF(value);
    op->func_doc = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_name(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    if (unlikely(op->func_name == NULL)) {
#if PY_MAJOR_VERSION >= 3
        op->func_name = PyUnicode_InternFromString(op->func.m_ml->ml_name);
#else
        op->func_name = PyString_InternFromString(op->func.m_ml->ml_name);
#endif
        if (unlikely(op->func_name == NULL))
            return NULL;
    }
    Py_INCREF(op->func_name);
    return op->func_name;
}
static int
__Pyx_CyFunction_set_name(__pyx_CyFunctionObject *op, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    tmp = op->func_name;
    Py_INCREF(value);
    op->func_name = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_qualname(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    Py_INCREF(op->func_qualname);
    return op->func_qualname;
}
static int
__Pyx_CyFunction_set_qualname(__pyx_CyFunctionObject *op, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    tmp = op->func_qualname;
    Py_INCREF(value);
    op->func_qualname = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_self(__pyx_CyFunctionObject *m, CYTHON_UNUSED void *closure)
{
    PyObject *self;
    self = m->func_closure;
    if (self == NULL)
        self = Py_None;
    Py_INCREF(self);
    return self;
}
static PyObject *
__Pyx_CyFunction_get_dict(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    if (unlikely(op->func_dict == NULL)) {
        op->func_dict = PyDict_New();
        if (unlikely(op->func_dict == NULL))
            return NULL;
    }
    Py_INCREF(op->func_dict);
    return op->func_dict;
}
static int
__Pyx_CyFunction_set_dict(__pyx_CyFunctionObject *op, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
    if (unlikely(value == NULL)) {
        PyErr_SetString(PyExc_TypeError,
               "function's dictionary may not be deleted");
        return -1;
    }
    if (unlikely(!PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
               "setting function's dictionary to a non-dict");
        return -1;
    }
    tmp = op->func_dict;
    Py_INCREF(value);
    op->func_dict = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_globals(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    Py_INCREF(op->func_globals);
    return op->func_globals;
}
static PyObject *
__Pyx_CyFunction_get_closure(CYTHON_UNUSED __pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    Py_INCREF(Py_None);
    return Py_None;
}
static PyObject *
__Pyx_CyFunction_get_code(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    PyObject* result = (op->func_code) ? op->func_code : Py_None;
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_init_defaults(__pyx_CyFunctionObject *op) {
    int result = 0;
    PyObject *res = op->defaults_getter((PyObject *) op);
    if (unlikely(!res))
        return -1;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    op->defaults_tuple = PyTuple_GET_ITEM(res, 0);
    Py_INCREF(op->defaults_tuple);
    op->defaults_kwdict = PyTuple_GET_ITEM(res, 1);
    Py_INCREF(op->defaults_kwdict);
    #else
    op->defaults_tuple = PySequence_ITEM(res, 0);
    if (unlikely(!op->defaults_tuple)) result = -1;
    else {
        op->defaults_kwdict = PySequence_ITEM(res, 1);
        if (unlikely(!op->defaults_kwdict)) result = -1;
    }
    #endif
    Py_DECREF(res);
    return result;
}
static int
__Pyx_CyFunction_set_defaults(__pyx_CyFunctionObject *op, PyObject* value, CYTHON_UNUSED void *context) {
    PyObject* tmp;
    if (!value) {
        value = Py_None;
    } else if (value != Py_None && !PyTuple_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "__defaults__ must be set to a tuple object");
        return -1;
    }
    Py_INCREF(value);
    tmp = op->defaults_tuple;
    op->defaults_tuple = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_defaults(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context) {
    PyObject* result = op->defaults_tuple;
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (__Pyx_CyFunction_init_defaults(op) < 0) return NULL;
            result = op->defaults_tuple;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_set_kwdefaults(__pyx_CyFunctionObject *op, PyObject* value, CYTHON_UNUSED void *context) {
    PyObject* tmp;
    if (!value) {
        value = Py_None;
    } else if (value != Py_None && !PyDict_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "__kwdefaults__ must be set to a dict object");
        return -1;
    }
    Py_INCREF(value);
    tmp = op->defaults_kwdict;
    op->defaults_kwdict = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_kwdefaults(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context) {
    PyObject* result = op->defaults_kwdict;
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (__Pyx_CyFunction_init_defaults(op) < 0) return NULL;
            result = op->defaults_kwdict;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_set_annotations(__pyx_CyFunctionObject *op, PyObject* value, CYTHON_UNUSED void *context) {
    PyObject* tmp;
    if (!value || value == Py_None) {
        value = NULL;
    } else if (!PyDict_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "__annotations__ must be set to a dict object");
        return -1;
    }
    Py_XINCREF(value);
    tmp = op->func_annotations;
    op->func_annotations = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_annotations(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context) {
    PyObject* result = op->func_annotations;
    if (unlikely(!result)) {
        result = PyDict_New();
        if (unlikely(!result)) return NULL;
        op->func_annotations = result;
    }
    Py_INCREF(result);
    return result;
}
static PyGetSetDef __pyx_CyFunction_getsets[] = {
    {(char *) "func_doc", (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "__doc__",  (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "func_name", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__name__", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__qualname__", (getter)__Pyx_CyFunction_get_qualname, (setter)__Pyx_CyFunction_set_qualname, 0, 0},
    {(char *) "__self__", (getter)__Pyx_CyFunction_get_self, 0, 0, 0},
    {(char *) "func_dict", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "__dict__", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "func_globals", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "__globals__", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "func_closure", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "__closure__", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "func_code", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "__code__", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "func_defaults", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {(char *) "__defaults__", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {(char *) "__kwdefaults__", (getter)__Pyx_CyFunction_get_kwdefaults, (setter)__Pyx_CyFunction_set_kwdefaults, 0, 0},
    {(char *) "__annotations__", (getter)__Pyx_CyFunction_get_annotations, (setter)__Pyx_CyFunction_set_annotations, 0, 0},
    {0, 0, 0, 0, 0}
};
static PyMemberDef __pyx_CyFunction_members[] = {
    {(char *) "__module__", T_OBJECT, offsetof(PyCFunctionObject, m_module), PY_WRITE_RESTRICTED, 0},
    {0, 0, 0,  0, 0}
};
static PyObject *
__Pyx_CyFunction_reduce(__pyx_CyFunctionObject *m, CYTHON_UNUSED PyObject *args)
{
#if PY_MAJOR_VERSION >= 3
    return PyUnicode_FromString(m->func.m_ml->ml_name);
#else
    return PyString_FromString(m->func.m_ml->ml_name);
#endif
}
static PyMethodDef __pyx_CyFunction_methods[] = {
    {"__reduce__", (PyCFunction)__Pyx_CyFunction_reduce, METH_VARARGS, 0},
    {0, 0, 0, 0}
};
#if PY_VERSION_HEX < 0x030500A0
#define __Pyx_CyFunction_weakreflist(cyfunc) ((cyfunc)->func_weakreflist)
#else
#define __Pyx_CyFunction_weakreflist(cyfunc) ((cyfunc)->func.m_weakreflist)
#endif
static PyObject *__Pyx_CyFunction_New(PyTypeObject *type, PyMethodDef *ml, int flags, PyObject* qualname,
                                      PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
    __pyx_CyFunctionObject *op = PyObject_GC_New(__pyx_CyFunctionObject, type);
    if (op == NULL)
        return NULL;
    op->flags = flags;
    __Pyx_CyFunction_weakreflist(op) = NULL;
    op->func.m_ml = ml;
    op->func.m_self = (PyObject *) op;
    Py_XINCREF(closure);
    op->func_closure = closure;
    Py_XINCREF(module);
    op->func.m_module = module;
    op->func_dict = NULL;
    op->func_name = NULL;
    Py_INCREF(qualname);
    op->func_qualname = qualname;
    op->func_doc = NULL;
    op->func_classobj = NULL;
    op->func_globals = globals;
    Py_INCREF(op->func_globals);
    Py_XINCREF(code);
    op->func_code = code;
    op->defaults_pyobjects = 0;
    op->defaults = NULL;
    op->defaults_tuple = NULL;
    op->defaults_kwdict = NULL;
    op->defaults_getter = NULL;
    op->func_annotations = NULL;
    PyObject_GC_Track(op);
    return (PyObject *) op;
}
static int
__Pyx_CyFunction_clear(__pyx_CyFunctionObject *m)
{
    Py_CLEAR(m->func_closure);
    Py_CLEAR(m->func.m_module);
    Py_CLEAR(m->func_dict);
    Py_CLEAR(m->func_name);
    Py_CLEAR(m->func_qualname);
    Py_CLEAR(m->func_doc);
    Py_CLEAR(m->func_globals);
    Py_CLEAR(m->func_code);
    Py_CLEAR(m->func_classobj);
    Py_CLEAR(m->defaults_tuple);
    Py_CLEAR(m->defaults_kwdict);
    Py_CLEAR(m->func_annotations);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_XDECREF(pydefaults[i]);
        PyObject_Free(m->defaults);
        m->defaults = NULL;
    }
    return 0;
}
static void __Pyx__CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    if (__Pyx_CyFunction_weakreflist(m) != NULL)
        PyObject_ClearWeakRefs((PyObject *) m);
    __Pyx_CyFunction_clear(m);
    PyObject_GC_Del(m);
}
static void __Pyx_CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    PyObject_GC_UnTrack(m);
    __Pyx__CyFunction_dealloc(m);
}
static int __Pyx_CyFunction_traverse(__pyx_CyFunctionObject *m, visitproc visit, void *arg)
{
    Py_VISIT(m->func_closure);
    Py_VISIT(m->func.m_module);
    Py_VISIT(m->func_dict);
    Py_VISIT(m->func_name);
    Py_VISIT(m->func_qualname);
    Py_VISIT(m->func_doc);
    Py_VISIT(m->func_globals);
    Py_VISIT(m->func_code);
    Py_VISIT(m->func_classobj);
    Py_VISIT(m->defaults_tuple);
    Py_VISIT(m->defaults_kwdict);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_VISIT(pydefaults[i]);
    }
    return 0;
}
static PyObject *__Pyx_CyFunction_descr_get(PyObject *func, PyObject *obj, PyObject *type)
{
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    if (m->flags & __Pyx_CYFUNCTION_STATICMETHOD) {
        Py_INCREF(func);
        return func;
    }
    if (m->flags & __Pyx_CYFUNCTION_CLASSMETHOD) {
        if (type == NULL)
            type = (PyObject *)(Py_TYPE(obj));
        return __Pyx_PyMethod_New(func, type, (PyObject *)(Py_TYPE(type)));
    }
    if (obj == Py_None)
        obj = NULL;
    return __Pyx_PyMethod_New(func, obj, type);
}
static PyObject*
__Pyx_CyFunction_repr(__pyx_CyFunctionObject *op)
{
#if PY_MAJOR_VERSION >= 3
    return PyUnicode_FromFormat("<cyfunction %U at %p>",
                                op->func_qualname, (void *)op);
#else
    return PyString_FromFormat("<cyfunction %s at %p>",
                               PyString_AsString(op->func_qualname), (void *)op);
#endif
}
static PyObject * __Pyx_CyFunction_CallMethod(PyObject *func, PyObject *self, PyObject *arg, PyObject *kw) {
    PyCFunctionObject* f = (PyCFunctionObject*)func;
    PyCFunction meth = f->m_ml->ml_meth;
    Py_ssize_t size;
    switch (f->m_ml->ml_flags & (METH_VARARGS | METH_KEYWORDS | METH_NOARGS | METH_O)) {
    case METH_VARARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0))
            return (*meth)(self, arg);
        break;
    case METH_VARARGS | METH_KEYWORDS:
        return (*(PyCFunctionWithKeywords)(void*)meth)(self, arg, kw);
    case METH_NOARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
            size = PyTuple_GET_SIZE(arg);
            if (likely(size == 0))
                return (*meth)(self, NULL);
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes no arguments (%" CYTHON_FORMAT_SSIZE_T "d given)",
                f->m_ml->ml_name, size);
            return NULL;
        }
        break;
    case METH_O:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
            size = PyTuple_GET_SIZE(arg);
            if (likely(size == 1)) {
                PyObject *result, *arg0;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                arg0 = PyTuple_GET_ITEM(arg, 0);
                #else
                arg0 = PySequence_ITEM(arg, 0); if (unlikely(!arg0)) return NULL;
                #endif
                result = (*meth)(self, arg0);
                #if !(CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
                Py_DECREF(arg0);
                #endif
                return result;
            }
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes exactly one argument (%" CYTHON_FORMAT_SSIZE_T "d given)",
                f->m_ml->ml_name, size);
            return NULL;
        }
        break;
    default:
        PyErr_SetString(PyExc_SystemError, "Bad call flags in "
                        "__Pyx_CyFunction_Call. METH_OLDARGS is no "
                        "longer supported!");
        return NULL;
    }
    PyErr_Format(PyExc_TypeError, "%.200s() takes no keyword arguments",
                 f->m_ml->ml_name);
    return NULL;
}
static CYTHON_INLINE PyObject *__Pyx_CyFunction_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    return __Pyx_CyFunction_CallMethod(func, ((PyCFunctionObject*)func)->m_self, arg, kw);
}
static PyObject *__Pyx_CyFunction_CallAsMethod(PyObject *func, PyObject *args, PyObject *kw) {
    PyObject *result;
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *) func;
    if ((cyfunc->flags & __Pyx_CYFUNCTION_CCLASS) && !(cyfunc->flags & __Pyx_CYFUNCTION_STATICMETHOD)) {
        Py_ssize_t argc;
        PyObject *new_args;
        PyObject *self;
        argc = PyTuple_GET_SIZE(args);
        new_args = PyTuple_GetSlice(args, 1, argc);
        if (unlikely(!new_args))
            return NULL;
        self = PyTuple_GetItem(args, 0);
        if (unlikely(!self)) {
            Py_DECREF(new_args);
            return NULL;
        }
        result = __Pyx_CyFunction_CallMethod(func, self, new_args, kw);
        Py_DECREF(new_args);
    } else {
        result = __Pyx_CyFunction_Call(func, args, kw);
    }
    return result;
}
static PyTypeObject __pyx_CyFunctionType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    "cython_function_or_method",
    sizeof(__pyx_CyFunctionObject),
    0,
    (destructor) __Pyx_CyFunction_dealloc,
    0,
    0,
    0,
#if PY_MAJOR_VERSION < 3
    0,
#else
    0,
#endif
    (reprfunc) __Pyx_CyFunction_repr,
    0,
    0,
    0,
    0,
    __Pyx_CyFunction_CallAsMethod,
    0,
    0,
    0,
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC,
    0,
    (traverseproc) __Pyx_CyFunction_traverse,
    (inquiry) __Pyx_CyFunction_clear,
    0,
#if PY_VERSION_HEX < 0x030500A0
    offsetof(__pyx_CyFunctionObject, func_weakreflist),
#else
    offsetof(PyCFunctionObject, m_weakreflist),
#endif
    0,
    0,
    __pyx_CyFunction_methods,
    __pyx_CyFunction_members,
    __pyx_CyFunction_getsets,
    0,
    0,
    __Pyx_CyFunction_descr_get,
    0,
    offsetof(__pyx_CyFunctionObject, func_dict),
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if PY_VERSION_HEX >= 0x030400a1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b1
    0,
#endif
};
static int __pyx_CyFunction_init(void) {
    __pyx_CyFunctionType = __Pyx_FetchCommonType(&__pyx_CyFunctionType_type);
    if (unlikely(__pyx_CyFunctionType == NULL)) {
        return -1;
    }
    return 0;
}
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *func, size_t size, int pyobjects) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults = PyObject_Malloc(size);
    if (unlikely(!m->defaults))
        return PyErr_NoMemory();
    memset(m->defaults, 0, size);
    m->defaults_pyobjects = pyobjects;
    return m->defaults;
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *func, PyObject *tuple) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_tuple = tuple;
    Py_INCREF(tuple);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_kwdict = dict;
    Py_INCREF(dict);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->func_annotations = dict;
    Py_INCREF(dict);
}

/* FusedFunction */
  static PyObject *
__pyx_FusedFunction_New(PyTypeObject *type, PyMethodDef *ml, int flags,
                        PyObject *qualname, PyObject *self,
                        PyObject *module, PyObject *globals,
                        PyObject *code)
{
    __pyx_FusedFunctionObject *fusedfunc =
        (__pyx_FusedFunctionObject *) __Pyx_CyFunction_New(type, ml, flags, qualname,
                                                           self, module, globals, code);
    if (!fusedfunc)
        return NULL;
    fusedfunc->__signatures__ = NULL;
    fusedfunc->type = NULL;
    fusedfunc->self = NULL;
    return (PyObject *) fusedfunc;
}
static void
__pyx_FusedFunction_dealloc(__pyx_FusedFunctionObject *self)
{
    PyObject_GC_UnTrack(self);
    Py_CLEAR(self->self);
    Py_CLEAR(self->type);
    Py_CLEAR(self->__signatures__);
    __Pyx__CyFunction_dealloc((__pyx_CyFunctionObject *) self);
}
static int
__pyx_FusedFunction_traverse(__pyx_FusedFunctionObject *self,
                             visitproc visit,
                             void *arg)
{
    Py_VISIT(self->self);
    Py_VISIT(self->type);
    Py_VISIT(self->__signatures__);
    return __Pyx_CyFunction_traverse((__pyx_CyFunctionObject *) self, visit, arg);
}
static int
__pyx_FusedFunction_clear(__pyx_FusedFunctionObject *self)
{
    Py_CLEAR(self->self);
    Py_CLEAR(self->type);
    Py_CLEAR(self->__signatures__);
    return __Pyx_CyFunction_clear((__pyx_CyFunctionObject *) self);
}
static PyObject *
__pyx_FusedFunction_descr_get(PyObject *self, PyObject *obj, PyObject *type)
{
    __pyx_FusedFunctionObject *func, *meth;
    func = (__pyx_FusedFunctionObject *) self;
    if (func->self || func->func.flags & __Pyx_CYFUNCTION_STATICMETHOD) {
        Py_INCREF(self);
        return self;
    }
    if (obj == Py_None)
        obj = NULL;
    meth = (__pyx_FusedFunctionObject *) __pyx_FusedFunction_NewEx(
                    ((PyCFunctionObject *) func)->m_ml,
                    ((__pyx_CyFunctionObject *) func)->flags,
                    ((__pyx_CyFunctionObject *) func)->func_qualname,
                    ((__pyx_CyFunctionObject *) func)->func_closure,
                    ((PyCFunctionObject *) func)->m_module,
                    ((__pyx_CyFunctionObject *) func)->func_globals,
                    ((__pyx_CyFunctionObject *) func)->func_code);
    if (!meth)
        return NULL;
    Py_XINCREF(func->func.func_classobj);
    meth->func.func_classobj = func->func.func_classobj;
    Py_XINCREF(func->__signatures__);
    meth->__signatures__ = func->__signatures__;
    Py_XINCREF(type);
    meth->type = type;
    Py_XINCREF(func->func.defaults_tuple);
    meth->func.defaults_tuple = func->func.defaults_tuple;
    if (func->func.flags & __Pyx_CYFUNCTION_CLASSMETHOD)
        obj = type;
    Py_XINCREF(obj);
    meth->self = obj;
    return (PyObject *) meth;
}
static PyObject *
_obj_to_str(PyObject *obj)
{
    if (PyType_Check(obj))
        return PyObject_GetAttr(obj, __pyx_n_s_name_2);
    else
        return PyObject_Str(obj);
}
static PyObject *
__pyx_FusedFunction_getitem(__pyx_FusedFunctionObject *self, PyObject *idx)
{
    PyObject *signature = NULL;
    PyObject *unbound_result_func;
    PyObject *result_func = NULL;
    if (self->__signatures__ == NULL) {
        PyErr_SetString(PyExc_TypeError, "Function is not fused");
        return NULL;
    }
    if (PyTuple_Check(idx)) {
        PyObject *list = PyList_New(0);
        Py_ssize_t n = PyTuple_GET_SIZE(idx);
        PyObject *string = NULL;
        PyObject *sep = NULL;
        int i;
        if (!list)
            return NULL;
        for (i = 0; i < n; i++) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            PyObject *item = PyTuple_GET_ITEM(idx, i);
#else
            PyObject *item = PySequence_ITEM(idx, i);
#endif
            string = _obj_to_str(item);
#if !(CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
            Py_DECREF(item);
#endif
            if (!string || PyList_Append(list, string) < 0)
                goto __pyx_err;
            Py_DECREF(string);
        }
        sep = PyUnicode_FromString("|");
        if (sep)
            signature = PyUnicode_Join(sep, list);
__pyx_err:
;
        Py_DECREF(list);
        Py_XDECREF(sep);
    } else {
        signature = _obj_to_str(idx);
    }
    if (!signature)
        return NULL;
    unbound_result_func = PyObject_GetItem(self->__signatures__, signature);
    if (unbound_result_func) {
        if (self->self || self->type) {
            __pyx_FusedFunctionObject *unbound = (__pyx_FusedFunctionObject *) unbound_result_func;
            Py_CLEAR(unbound->func.func_classobj);
            Py_XINCREF(self->func.func_classobj);
            unbound->func.func_classobj = self->func.func_classobj;
            result_func = __pyx_FusedFunction_descr_get(unbound_result_func,
                                                        self->self, self->type);
        } else {
            result_func = unbound_result_func;
            Py_INCREF(result_func);
        }
    }
    Py_DECREF(signature);
    Py_XDECREF(unbound_result_func);
    return result_func;
}
static PyObject *
__pyx_FusedFunction_callfunction(PyObject *func, PyObject *args, PyObject *kw)
{
     __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *) func;
    int static_specialized = (cyfunc->flags & __Pyx_CYFUNCTION_STATICMETHOD &&
                              !((__pyx_FusedFunctionObject *) func)->__signatures__);
    if (cyfunc->flags & __Pyx_CYFUNCTION_CCLASS && !static_specialized) {
        return __Pyx_CyFunction_CallAsMethod(func, args, kw);
    } else {
        return __Pyx_CyFunction_Call(func, args, kw);
    }
}
static PyObject *
__pyx_FusedFunction_call(PyObject *func, PyObject *args, PyObject *kw)
{
    __pyx_FusedFunctionObject *binding_func = (__pyx_FusedFunctionObject *) func;
    Py_ssize_t argc = PyTuple_GET_SIZE(args);
    PyObject *new_args = NULL;
    __pyx_FusedFunctionObject *new_func = NULL;
    PyObject *result = NULL;
    PyObject *self = NULL;
    int is_staticmethod = binding_func->func.flags & __Pyx_CYFUNCTION_STATICMETHOD;
    int is_classmethod = binding_func->func.flags & __Pyx_CYFUNCTION_CLASSMETHOD;
    if (binding_func->self) {
        Py_ssize_t i;
        new_args = PyTuple_New(argc + 1);
        if (!new_args)
            return NULL;
        self = binding_func->self;
#if !(CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
        Py_INCREF(self);
#endif
        Py_INCREF(self);
        PyTuple_SET_ITEM(new_args, 0, self);
        for (i = 0; i < argc; i++) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            PyObject *item = PyTuple_GET_ITEM(args, i);
            Py_INCREF(item);
#else
            PyObject *item = PySequence_ITEM(args, i);  if (unlikely(!item)) goto bad;
#endif
            PyTuple_SET_ITEM(new_args, i + 1, item);
        }
        args = new_args;
    } else if (binding_func->type) {
        if (argc < 1) {
            PyErr_SetString(PyExc_TypeError, "Need at least one argument, 0 given.");
            return NULL;
        }
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        self = PyTuple_GET_ITEM(args, 0);
#else
        self = PySequence_ITEM(args, 0);  if (unlikely(!self)) return NULL;
#endif
    }
    if (self && !is_classmethod && !is_staticmethod) {
        int is_instance = PyObject_IsInstance(self, binding_func->type);
        if (unlikely(!is_instance)) {
            PyErr_Format(PyExc_TypeError,
                         "First argument should be of type %.200s, got %.200s.",
                         ((PyTypeObject *) binding_func->type)->tp_name,
                         self->ob_type->tp_name);
            goto bad;
        } else if (unlikely(is_instance == -1)) {
            goto bad;
        }
    }
#if !(CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
    Py_XDECREF(self);
    self = NULL;
#endif
    if (binding_func->__signatures__) {
        PyObject *tup;
        if (is_staticmethod && binding_func->func.flags & __Pyx_CYFUNCTION_CCLASS) {
            tup = PyTuple_Pack(3, args,
                               kw == NULL ? Py_None : kw,
                               binding_func->func.defaults_tuple);
            if (unlikely(!tup)) goto bad;
            new_func = (__pyx_FusedFunctionObject *) __Pyx_CyFunction_CallMethod(
                func, binding_func->__signatures__, tup, NULL);
        } else {
            tup = PyTuple_Pack(4, binding_func->__signatures__, args,
                               kw == NULL ? Py_None : kw,
                               binding_func->func.defaults_tuple);
            if (unlikely(!tup)) goto bad;
            new_func = (__pyx_FusedFunctionObject *) __pyx_FusedFunction_callfunction(func, tup, NULL);
        }
        Py_DECREF(tup);
        if (unlikely(!new_func))
            goto bad;
        Py_XINCREF(binding_func->func.func_classobj);
        Py_CLEAR(new_func->func.func_classobj);
        new_func->func.func_classobj = binding_func->func.func_classobj;
        func = (PyObject *) new_func;
    }
    result = __pyx_FusedFunction_callfunction(func, args, kw);
bad:
#if !(CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
    Py_XDECREF(self);
#endif
    Py_XDECREF(new_args);
    Py_XDECREF((PyObject *) new_func);
    return result;
}
static PyMemberDef __pyx_FusedFunction_members[] = {
    {(char *) "__signatures__",
     T_OBJECT,
     offsetof(__pyx_FusedFunctionObject, __signatures__),
     READONLY,
     0},
    {0, 0, 0, 0, 0},
};
static PyMappingMethods __pyx_FusedFunction_mapping_methods = {
    0,
    (binaryfunc) __pyx_FusedFunction_getitem,
    0,
};
static PyTypeObject __pyx_FusedFunctionType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    "fused_cython_function",
    sizeof(__pyx_FusedFunctionObject),
    0,
    (destructor) __pyx_FusedFunction_dealloc,
    0,
    0,
    0,
#if PY_MAJOR_VERSION < 3
    0,
#else
    0,
#endif
    0,
    0,
    0,
    &__pyx_FusedFunction_mapping_methods,
    0,
    (ternaryfunc) __pyx_FusedFunction_call,
    0,
    0,
    0,
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_BASETYPE,
    0,
    (traverseproc) __pyx_FusedFunction_traverse,
    (inquiry) __pyx_FusedFunction_clear,
    0,
    0,
    0,
    0,
    0,
    __pyx_FusedFunction_members,
    __pyx_CyFunction_getsets,
    &__pyx_CyFunctionType_type,
    0,
    __pyx_FusedFunction_descr_get,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if PY_VERSION_HEX >= 0x030400a1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b1
    0,
#endif
};
static int __pyx_FusedFunction_init(void) {
    __pyx_FusedFunctionType = __Pyx_FetchCommonType(&__pyx_FusedFunctionType_type);
    if (__pyx_FusedFunctionType == NULL) {
        return -1;
    }
    return 0;
}

/* CLineInTraceback */
  #ifndef CYTHON_CLINE_IN_TRACEBACK
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line) {
    PyObject *use_cline;
    PyObject *ptype, *pvalue, *ptraceback;
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject **cython_runtime_dict;
#endif
    if (unlikely(!__pyx_cython_runtime)) {
        return c_line;
    }
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
#if CYTHON_COMPILING_IN_CPYTHON
    cython_runtime_dict = _PyObject_GetDictPtr(__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
        __PYX_PY_DICT_LOOKUP_IF_MODIFIED(
            use_cline, *cython_runtime_dict,
            __Pyx_PyDict_GetItemStr(*cython_runtime_dict, __pyx_n_s_cline_in_traceback))
    } else
#endif
    {
      PyObject *use_cline_obj = __Pyx_PyObject_GetAttrStr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback);
      if (use_cline_obj) {
        use_cline = PyObject_Not(use_cline_obj) ? Py_False : Py_True;
        Py_DECREF(use_cline_obj);
      } else {
        PyErr_Clear();
        use_cline = NULL;
      }
    }
    if (!use_cline) {
        c_line = 0;
        PyObject_SetAttr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback, Py_False);
    }
    else if (use_cline == Py_False || (use_cline != Py_True && PyObject_Not(use_cline) != 0)) {
        c_line = 0;
    }
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache */
  static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, (size_t)new_max*sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}

/* AddTraceback */
  #include "compile.h"
#include "frameobject.h"
#include "traceback.h"
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyObject *py_srcfile = 0;
    PyObject *py_funcname = 0;
    #if PY_MAJOR_VERSION < 3
    py_srcfile = PyString_FromString(filename);
    #else
    py_srcfile = PyUnicode_FromString(filename);
    #endif
    if (!py_srcfile) goto bad;
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        #else
        py_funcname = PyUnicode_FromString(funcname);
        #endif
    }
    if (!py_funcname) goto bad;
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    Py_DECREF(py_funcname);
    return py_code;
bad:
    Py_XDECREF(py_srcfile);
    Py_XDECREF(py_funcname);
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) goto bad;
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}

#if PY_MAJOR_VERSION < 3
static int __Pyx_GetBuffer(PyObject *obj, Py_buffer *view, int flags) {
    if (PyObject_CheckBuffer(obj)) return PyObject_GetBuffer(obj, view, flags);
        if (__Pyx_TypeCheck(obj, __pyx_ptype_5numpy_ndarray)) return __pyx_pw_5numpy_7ndarray_1__getbuffer__(obj, view, flags);
        if (__Pyx_TypeCheck(obj, __pyx_array_type)) return __pyx_array_getbuffer(obj, view, flags);
        if (__Pyx_TypeCheck(obj, __pyx_memoryview_type)) return __pyx_memoryview_getbuffer(obj, view, flags);
    PyErr_Format(PyExc_TypeError, "'%.200s' does not have the buffer interface", Py_TYPE(obj)->tp_name);
    return -1;
}
static void __Pyx_ReleaseBuffer(Py_buffer *view) {
    PyObject *obj = view->obj;
    if (!obj) return;
    if (PyObject_CheckBuffer(obj)) {
        PyBuffer_Release(view);
        return;
    }
    if ((0)) {}
        else if (__Pyx_TypeCheck(obj, __pyx_ptype_5numpy_ndarray)) __pyx_pw_5numpy_7ndarray_3__releasebuffer__(obj, view);
    view->obj = NULL;
    Py_DECREF(obj);
}
#endif


  /* MemviewSliceIsContig */
  static int
__pyx_memviewslice_is_contig(const __Pyx_memviewslice mvs, char order, int ndim)
{
    int i, index, step, start;
    Py_ssize_t itemsize = mvs.memview->view.itemsize;
    if (order == 'F') {
        step = 1;
        start = 0;
    } else {
        step = -1;
        start = ndim - 1;
    }
    for (i = 0; i < ndim; i++) {
        index = start + step * i;
        if (mvs.suboffsets[index] >= 0 || mvs.strides[index] != itemsize)
            return 0;
        itemsize *= mvs.shape[index];
    }
    return 1;
}

/* OverlappingSlices */
  static void
__pyx_get_array_memory_extents(__Pyx_memviewslice *slice,
                               void **out_start, void **out_end,
                               int ndim, size_t itemsize)
{
    char *start, *end;
    int i;
    start = end = slice->data;
    for (i = 0; i < ndim; i++) {
        Py_ssize_t stride = slice->strides[i];
        Py_ssize_t extent = slice->shape[i];
        if (extent == 0) {
            *out_start = *out_end = start;
            return;
        } else {
            if (stride > 0)
                end += stride * (extent - 1);
            else
                start += stride * (extent - 1);
        }
    }
    *out_start = start;
    *out_end = end + itemsize;
}
static int
__pyx_slices_overlap(__Pyx_memviewslice *slice1,
                     __Pyx_memviewslice *slice2,
                     int ndim, size_t itemsize)
{
    void *start1, *end1, *start2, *end2;
    __pyx_get_array_memory_extents(slice1, &start1, &end1, ndim, itemsize);
    __pyx_get_array_memory_extents(slice2, &start2, &end2, ndim, itemsize);
    return (start1 < end2) && (start2 < end1);
}

/* Capsule */
  static CYTHON_INLINE PyObject *
__pyx_capsule_create(void *p, CYTHON_UNUSED const char *sig)
{
    PyObject *cobj;
#if PY_VERSION_HEX >= 0x02070000
    cobj = PyCapsule_New(p, sig, NULL);
#else
    cobj = PyCObject_FromVoidPtr(p, NULL);
#endif
    return cobj;
}

/* TypeInfoCompare */
  static int
__pyx_typeinfo_cmp(__Pyx_TypeInfo *a, __Pyx_TypeInfo *b)
{
    int i;
    if (!a || !b)
        return 0;
    if (a == b)
        return 1;
    if (a->size != b->size || a->typegroup != b->typegroup ||
            a->is_unsigned != b->is_unsigned || a->ndim != b->ndim) {
        if (a->typegroup == 'H' || b->typegroup == 'H') {
            return a->size == b->size;
        } else {
            return 0;
        }
    }
    if (a->ndim) {
        for (i = 0; i < a->ndim; i++)
            if (a->arraysize[i] != b->arraysize[i])
                return 0;
    }
    if (a->typegroup == 'S') {
        if (a->flags != b->flags)
            return 0;
        if (a->fields || b->fields) {
            if (!(a->fields && b->fields))
                return 0;
            for (i = 0; a->fields[i].type && b->fields[i].type; i++) {
                __Pyx_StructField *field_a = a->fields + i;
                __Pyx_StructField *field_b = b->fields + i;
                if (field_a->offset != field_b->offset ||
                    !__pyx_typeinfo_cmp(field_a->type, field_b->type))
                    return 0;
            }
            return !a->fields[i].type && !b->fields[i].type;
        }
    }
    return 1;
}

/* MemviewSliceValidateAndInit */
  static int
__pyx_check_strides(Py_buffer *buf, int dim, int ndim, int spec)
{
    if (buf->shape[dim] <= 1)
        return 1;
    if (buf->strides) {
        if (spec & __Pyx_MEMVIEW_CONTIG) {
            if (spec & (__Pyx_MEMVIEW_PTR|__Pyx_MEMVIEW_FULL)) {
                if (buf->strides[dim] != sizeof(void *)) {
                    PyErr_Format(PyExc_ValueError,
                                 "Buffer is not indirectly contiguous "
                                 "in dimension %d.", dim);
                    goto fail;
                }
            } else if (buf->strides[dim] != buf->itemsize) {
                PyErr_SetString(PyExc_ValueError,
                                "Buffer and memoryview are not contiguous "
                                "in the same dimension.");
                goto fail;
            }
        }
        if (spec & __Pyx_MEMVIEW_FOLLOW) {
            Py_ssize_t stride = buf->strides[dim];
            if (stride < 0)
                stride = -stride;
            if (stride < buf->itemsize) {
                PyErr_SetString(PyExc_ValueError,
                                "Buffer and memoryview are not contiguous "
                                "in the same dimension.");
                goto fail;
            }
        }
    } else {
        if (spec & __Pyx_MEMVIEW_CONTIG && dim != ndim - 1) {
            PyErr_Format(PyExc_ValueError,
                         "C-contiguous buffer is not contiguous in "
                         "dimension %d", dim);
            goto fail;
        } else if (spec & (__Pyx_MEMVIEW_PTR)) {
            PyErr_Format(PyExc_ValueError,
                         "C-contiguous buffer is not indirect in "
                         "dimension %d", dim);
            goto fail;
        } else if (buf->suboffsets) {
            PyErr_SetString(PyExc_ValueError,
                            "Buffer exposes suboffsets but no strides");
            goto fail;
        }
    }
    return 1;
fail:
    return 0;
}
static int
__pyx_check_suboffsets(Py_buffer *buf, int dim, CYTHON_UNUSED int ndim, int spec)
{
    if (spec & __Pyx_MEMVIEW_DIRECT) {
        if (buf->suboffsets && buf->suboffsets[dim] >= 0) {
            PyErr_Format(PyExc_ValueError,
                         "Buffer not compatible with direct access "
                         "in dimension %d.", dim);
            goto fail;
        }
    }
    if (spec & __Pyx_MEMVIEW_PTR) {
        if (!buf->suboffsets || (buf->suboffsets[dim] < 0)) {
            PyErr_Format(PyExc_ValueError,
                         "Buffer is not indirectly accessible "
                         "in dimension %d.", dim);
            goto fail;
        }
    }
    return 1;
fail:
    return 0;
}
static int
__pyx_verify_contig(Py_buffer *buf, int ndim, int c_or_f_flag)
{
    int i;
    if (c_or_f_flag & __Pyx_IS_F_CONTIG) {
        Py_ssize_t stride = 1;
        for (i = 0; i < ndim; i++) {
            if (stride * buf->itemsize != buf->strides[i] &&
                    buf->shape[i] > 1)
            {
                PyErr_SetString(PyExc_ValueError,
                    "Buffer not fortran contiguous.");
                goto fail;
            }
            stride = stride * buf->shape[i];
        }
    } else if (c_or_f_flag & __Pyx_IS_C_CONTIG) {
        Py_ssize_t stride = 1;
        for (i = ndim - 1; i >- 1; i--) {
            if (stride * buf->itemsize != buf->strides[i] &&
                    buf->shape[i] > 1) {
                PyErr_SetString(PyExc_ValueError,
                    "Buffer not C contiguous.");
                goto fail;
            }
            stride = stride * buf->shape[i];
        }
    }
    return 1;
fail:
    return 0;
}
static int __Pyx_ValidateAndInit_memviewslice(
                int *axes_specs,
                int c_or_f_flag,
                int buf_flags,
                int ndim,
                __Pyx_TypeInfo *dtype,
                __Pyx_BufFmt_StackElem stack[],
                __Pyx_memviewslice *memviewslice,
                PyObject *original_obj)
{
    struct __pyx_memoryview_obj *memview, *new_memview;
    __Pyx_RefNannyDeclarations
    Py_buffer *buf;
    int i, spec = 0, retval = -1;
    __Pyx_BufFmt_Context ctx;
    int from_memoryview = __pyx_memoryview_check(original_obj);
    __Pyx_RefNannySetupContext("ValidateAndInit_memviewslice", 0);
    if (from_memoryview && __pyx_typeinfo_cmp(dtype, ((struct __pyx_memoryview_obj *)
                                                            original_obj)->typeinfo)) {
        memview = (struct __pyx_memoryview_obj *) original_obj;
        new_memview = NULL;
    } else {
        memview = (struct __pyx_memoryview_obj *) __pyx_memoryview_new(
                                            original_obj, buf_flags, 0, dtype);
        new_memview = memview;
        if (unlikely(!memview))
            goto fail;
    }
    buf = &memview->view;
    if (buf->ndim != ndim) {
        PyErr_Format(PyExc_ValueError,
                "Buffer has wrong number of dimensions (expected %d, got %d)",
                ndim, buf->ndim);
        goto fail;
    }
    if (new_memview) {
        __Pyx_BufFmt_Init(&ctx, stack, dtype);
        if (!__Pyx_BufFmt_CheckString(&ctx, buf->format)) goto fail;
    }
    if ((unsigned) buf->itemsize != dtype->size) {
        PyErr_Format(PyExc_ValueError,
                     "Item size of buffer (%" CYTHON_FORMAT_SSIZE_T "u byte%s) "
                     "does not match size of '%s' (%" CYTHON_FORMAT_SSIZE_T "u byte%s)",
                     buf->itemsize,
                     (buf->itemsize > 1) ? "s" : "",
                     dtype->name,
                     dtype->size,
                     (dtype->size > 1) ? "s" : "");
        goto fail;
    }
    for (i = 0; i < ndim; i++) {
        spec = axes_specs[i];
        if (!__pyx_check_strides(buf, i, ndim, spec))
            goto fail;
        if (!__pyx_check_suboffsets(buf, i, ndim, spec))
            goto fail;
    }
    if (buf->strides && !__pyx_verify_contig(buf, ndim, c_or_f_flag))
        goto fail;
    if (unlikely(__Pyx_init_memviewslice(memview, ndim, memviewslice,
                                         new_memview != NULL) == -1)) {
        goto fail;
    }
    retval = 0;
    goto no_fail;
fail:
    Py_XDECREF(new_memview);
    retval = -1;
no_fail:
    __Pyx_RefNannyFinishContext();
    return retval;
}

/* ObjectToMemviewSlice */
  static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_float64_t(PyObject *obj, int writable_flag) {
    __Pyx_memviewslice result = { 0, 0, { 0 }, { 0 }, { 0 } };
    __Pyx_BufFmt_StackElem stack[1];
    int axes_specs[] = { (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_STRIDED) };
    int retcode;
    if (obj == Py_None) {
        result.memview = (struct __pyx_memoryview_obj *) Py_None;
        return result;
    }
    retcode = __Pyx_ValidateAndInit_memviewslice(axes_specs, 0,
                                                 PyBUF_RECORDS_RO | writable_flag, 1,
                                                 &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, stack,
                                                 &result, obj);
    if (unlikely(retcode == -1))
        goto __pyx_fail;
    return result;
__pyx_fail:
    result.memview = NULL;
    result.data = NULL;
    return result;
}

/* ObjectToMemviewSlice */
  static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_float32_t(PyObject *obj, int writable_flag) {
    __Pyx_memviewslice result = { 0, 0, { 0 }, { 0 }, { 0 } };
    __Pyx_BufFmt_StackElem stack[1];
    int axes_specs[] = { (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_STRIDED) };
    int retcode;
    if (obj == Py_None) {
        result.memview = (struct __pyx_memoryview_obj *) Py_None;
        return result;
    }
    retcode = __Pyx_ValidateAndInit_memviewslice(axes_specs, 0,
                                                 PyBUF_RECORDS_RO | writable_flag, 1,
                                                 &__Pyx_TypeInfo_nn___pyx_t_5numpy_float32_t, stack,
                                                 &result, obj);
    if (unlikely(retcode == -1))
        goto __pyx_fail;
    return result;
__pyx_fail:
    result.memview = NULL;
    result.data = NULL;
    return result;
}

/* ObjectToMemviewSlice */
  static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(PyObject *obj, int writable_flag) {
    __Pyx_memviewslice result = { 0, 0, { 0 }, { 0 }, { 0 } };
    __Pyx_BufFmt_StackElem stack[1];
    int axes_specs[] = { (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_STRIDED) };
    int retcode;
    if (obj == Py_None) {
        result.memview = (struct __pyx_memoryview_obj *) Py_None;
        return result;
    }
    retcode = __Pyx_ValidateAndInit_memviewslice(axes_specs, 0,
                                                 PyBUF_RECORDS_RO | writable_flag, 1,
                                                 &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, stack,
                                                 &result, obj);
    if (unlikely(retcode == -1))
        goto __pyx_fail;
    return result;
__pyx_fail:
    result.memview = NULL;
    result.data = NULL;
    return result;
}

/* ObjectToMemviewSlice */
  static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int32_t(PyObject *obj, int writable_flag) {
    __Pyx_memviewslice result = { 0, 0, { 0 }, { 0 }, { 0 } };
    __Pyx_BufFmt_StackElem stack[1];
    int axes_specs[] = { (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_STRIDED) };
    int retcode;
    if (obj == Py_None) {
        result.memview = (struct __pyx_memoryview_obj *) Py_None;
        return result;
    }
    retcode = __Pyx_ValidateAndInit_memviewslice(axes_specs, 0,
                                                 PyBUF_RECORDS_RO | writable_flag, 1,
                                                 &__Pyx_TypeInfo_nn___pyx_t_5numpy_int32_t, stack,
                                                 &result, obj);
    if (unlikely(retcode == -1))
        goto __pyx_fail;
    return result;
__pyx_fail:
    result.memview = NULL;
    result.data = NULL;
    return result;
}

/* ObjectToMemviewSlice */
  static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_uint64_t(PyObject *obj, int writable_flag) {
    __Pyx_memviewslice result = { 0, 0, { 0 }, { 0 }, { 0 } };
    __Pyx_BufFmt_StackElem stack[1];
    int axes_specs[] = { (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_STRIDED) };
    int retcode;
    if (obj == Py_None) {
        result.memview = (struct __pyx_memoryview_obj *) Py_None;
        return result;
    }
    retcode = __Pyx_ValidateAndInit_memviewslice(axes_specs, 0,
                                                 PyBUF_RECORDS_RO | writable_flag, 1,
                                                 &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint64_t, stack,
                                                 &result, obj);
    if (unlikely(retcode == -1))
        goto __pyx_fail;
    return result;
__pyx_fail:
    result.memview = NULL;
    result.data = NULL;
    return result;
}

/* CIntToPy */
  static CYTHON_INLINE PyObject* __Pyx_PyInt_From_npy_int64(npy_int64 value) {
    const npy_int64 neg_one = (npy_int64) ((npy_int64) 0 - (npy_int64) 1), const_zero = (npy_int64) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(npy_int64) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(npy_int64) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(npy_int64) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(npy_int64) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(npy_int64) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(npy_int64),
                                     little, !is_unsigned);
    }
}

/* CIntFromPyVerify */
  #define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* MemviewDtypeToObject */
  static CYTHON_INLINE PyObject *__pyx_memview_get_nn___pyx_t_5numpy_int64_t(const char *itemp) {
    return (PyObject *) __Pyx_PyInt_From_npy_int64(*(__pyx_t_5numpy_int64_t *) itemp);
}
static CYTHON_INLINE int __pyx_memview_set_nn___pyx_t_5numpy_int64_t(const char *itemp, PyObject *obj) {
    __pyx_t_5numpy_int64_t value = __Pyx_PyInt_As_npy_int64(obj);
    if ((value == ((npy_int64)-1)) && PyErr_Occurred())
        return 0;
    *(__pyx_t_5numpy_int64_t *) itemp = value;
    return 1;
}

/* MemviewDtypeToObject */
  static CYTHON_INLINE PyObject *__pyx_memview_get_nn___pyx_t_5numpy_float32_t(const char *itemp) {
    return (PyObject *) PyFloat_FromDouble(*(__pyx_t_5numpy_float32_t *) itemp);
}
static CYTHON_INLINE int __pyx_memview_set_nn___pyx_t_5numpy_float32_t(const char *itemp, PyObject *obj) {
    __pyx_t_5numpy_float32_t value = __pyx_PyFloat_AsFloat(obj);
    if ((value == ((npy_float32)-1)) && PyErr_Occurred())
        return 0;
    *(__pyx_t_5numpy_float32_t *) itemp = value;
    return 1;
}

/* MemviewDtypeToObject */
  static CYTHON_INLINE PyObject *__pyx_memview_get_nn___pyx_t_5numpy_float64_t(const char *itemp) {
    return (PyObject *) PyFloat_FromDouble(*(__pyx_t_5numpy_float64_t *) itemp);
}
static CYTHON_INLINE int __pyx_memview_set_nn___pyx_t_5numpy_float64_t(const char *itemp, PyObject *obj) {
    __pyx_t_5numpy_float64_t value = __pyx_PyFloat_AsDouble(obj);
    if ((value == ((npy_float64)-1)) && PyErr_Occurred())
        return 0;
    *(__pyx_t_5numpy_float64_t *) itemp = value;
    return 1;
}

/* CIntToPy */
  static CYTHON_INLINE PyObject* __Pyx_PyInt_From_npy_int32(npy_int32 value) {
    const npy_int32 neg_one = (npy_int32) ((npy_int32) 0 - (npy_int32) 1), const_zero = (npy_int32) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(npy_int32) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(npy_int32) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(npy_int32) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(npy_int32) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(npy_int32) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(npy_int32),
                                     little, !is_unsigned);
    }
}

/* MemviewDtypeToObject */
  static CYTHON_INLINE PyObject *__pyx_memview_get_nn___pyx_t_5numpy_int32_t(const char *itemp) {
    return (PyObject *) __Pyx_PyInt_From_npy_int32(*(__pyx_t_5numpy_int32_t *) itemp);
}
static CYTHON_INLINE int __pyx_memview_set_nn___pyx_t_5numpy_int32_t(const char *itemp, PyObject *obj) {
    __pyx_t_5numpy_int32_t value = __Pyx_PyInt_As_npy_int32(obj);
    if ((value == ((npy_int32)-1)) && PyErr_Occurred())
        return 0;
    *(__pyx_t_5numpy_int32_t *) itemp = value;
    return 1;
}

/* CIntToPy */
  static CYTHON_INLINE PyObject* __Pyx_PyInt_From_npy_uint64(npy_uint64 value) {
    const npy_uint64 neg_one = (npy_uint64) ((npy_uint64) 0 - (npy_uint64) 1), const_zero = (npy_uint64) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(npy_uint64) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(npy_uint64) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(npy_uint64) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(npy_uint64) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(npy_uint64) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(npy_uint64),
                                     little, !is_unsigned);
    }
}

/* MemviewDtypeToObject */
  static CYTHON_INLINE PyObject *__pyx_memview_get_nn___pyx_t_5numpy_uint64_t(const char *itemp) {
    return (PyObject *) __Pyx_PyInt_From_npy_uint64(*(__pyx_t_5numpy_uint64_t *) itemp);
}
static CYTHON_INLINE int __pyx_memview_set_nn___pyx_t_5numpy_uint64_t(const char *itemp, PyObject *obj) {
    __pyx_t_5numpy_uint64_t value = __Pyx_PyInt_As_npy_uint64(obj);
    if ((value == ((npy_uint64)-1)) && PyErr_Occurred())
        return 0;
    *(__pyx_t_5numpy_uint64_t *) itemp = value;
    return 1;
}

/* CIntToPy */
  static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value) {
    const int neg_one = (int) ((int) 0 - (int) 1), const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(int),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
  static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
    const long neg_one = (long) ((long) 0 - (long) 1), const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
    }
}

/* Declarations */
  #if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      return ::std::complex< float >(x, y);
    }
  #else
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      return x + y*(__pyx_t_float_complex)_Complex_I;
    }
  #endif
#else
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      __pyx_t_float_complex z;
      z.real = x;
      z.imag = y;
      return z;
    }
#endif

/* Arithmetic */
  #if CYTHON_CCOMPLEX
#else
    static CYTHON_INLINE int __Pyx_c_eq_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
       return (a.real == b.real) && (a.imag == b.imag);
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_sum_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real + b.real;
        z.imag = a.imag + b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_diff_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real - b.real;
        z.imag = a.imag - b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_prod_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real * b.real - a.imag * b.imag;
        z.imag = a.real * b.imag + a.imag * b.real;
        return z;
    }
    #if 1
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quot_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        if (b.imag == 0) {
            return __pyx_t_float_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else if (fabsf(b.real) >= fabsf(b.imag)) {
            if (b.real == 0 && b.imag == 0) {
                return __pyx_t_float_complex_from_parts(a.real / b.real, a.imag / b.imag);
            } else {
                float r = b.imag / b.real;
                float s = (float)(1.0) / (b.real + b.imag * r);
                return __pyx_t_float_complex_from_parts(
                    (a.real + a.imag * r) * s, (a.imag - a.real * r) * s);
            }
        } else {
            float r = b.real / b.imag;
            float s = (float)(1.0) / (b.imag + b.real * r);
            return __pyx_t_float_complex_from_parts(
                (a.real * r + a.imag) * s, (a.imag * r - a.real) * s);
        }
    }
    #else
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quot_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        if (b.imag == 0) {
            return __pyx_t_float_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else {
            float denom = b.real * b.real + b.imag * b.imag;
            return __pyx_t_float_complex_from_parts(
                (a.real * b.real + a.imag * b.imag) / denom,
                (a.imag * b.real - a.real * b.imag) / denom);
        }
    }
    #endif
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_neg_float(__pyx_t_float_complex a) {
        __pyx_t_float_complex z;
        z.real = -a.real;
        z.imag = -a.imag;
        return z;
    }
    static CYTHON_INLINE int __Pyx_c_is_zero_float(__pyx_t_float_complex a) {
       return (a.real == 0) && (a.imag == 0);
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_conj_float(__pyx_t_float_complex a) {
        __pyx_t_float_complex z;
        z.real =  a.real;
        z.imag = -a.imag;
        return z;
    }
    #if 1
        static CYTHON_INLINE float __Pyx_c_abs_float(__pyx_t_float_complex z) {
          #if !defined(HAVE_HYPOT) || defined(_MSC_VER)
            return sqrtf(z.real*z.real + z.imag*z.imag);
          #else
            return hypotf(z.real, z.imag);
          #endif
        }
        static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_pow_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
            __pyx_t_float_complex z;
            float r, lnr, theta, z_r, z_theta;
            if (b.imag == 0 && b.real == (int)b.real) {
                if (b.real < 0) {
                    float denom = a.real * a.real + a.imag * a.imag;
                    a.real = a.real / denom;
                    a.imag = -a.imag / denom;
                    b.real = -b.real;
                }
                switch ((int)b.real) {
                    case 0:
                        z.real = 1;
                        z.imag = 0;
                        return z;
                    case 1:
                        return a;
                    case 2:
                        z = __Pyx_c_prod_float(a, a);
                        return __Pyx_c_prod_float(a, a);
                    case 3:
                        z = __Pyx_c_prod_float(a, a);
                        return __Pyx_c_prod_float(z, a);
                    case 4:
                        z = __Pyx_c_prod_float(a, a);
                        return __Pyx_c_prod_float(z, z);
                }
            }
            if (a.imag == 0) {
                if (a.real == 0) {
                    return a;
                } else if (b.imag == 0) {
                    z.real = powf(a.real, b.real);
                    z.imag = 0;
                    return z;
                } else if (a.real > 0) {
                    r = a.real;
                    theta = 0;
                } else {
                    r = -a.real;
                    theta = atan2f(0.0, -1.0);
                }
            } else {
                r = __Pyx_c_abs_float(a);
                theta = atan2f(a.imag, a.real);
            }
            lnr = logf(r);
            z_r = expf(lnr * b.real - theta * b.imag);
            z_theta = theta * b.real + lnr * b.imag;
            z.real = z_r * cosf(z_theta);
            z.imag = z_r * sinf(z_theta);
            return z;
        }
    #endif
#endif

/* Declarations */
  #if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      return ::std::complex< double >(x, y);
    }
  #else
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      return x + y*(__pyx_t_double_complex)_Complex_I;
    }
  #endif
#else
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      __pyx_t_double_complex z;
      z.real = x;
      z.imag = y;
      return z;
    }
#endif

/* Arithmetic */
  #if CYTHON_CCOMPLEX
#else
    static CYTHON_INLINE int __Pyx_c_eq_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
       return (a.real == b.real) && (a.imag == b.imag);
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_sum_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real + b.real;
        z.imag = a.imag + b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_diff_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real - b.real;
        z.imag = a.imag - b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_prod_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real * b.real - a.imag * b.imag;
        z.imag = a.real * b.imag + a.imag * b.real;
        return z;
    }
    #if 1
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        if (b.imag == 0) {
            return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else if (fabs(b.real) >= fabs(b.imag)) {
            if (b.real == 0 && b.imag == 0) {
                return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.imag);
            } else {
                double r = b.imag / b.real;
                double s = (double)(1.0) / (b.real + b.imag * r);
                return __pyx_t_double_complex_from_parts(
                    (a.real + a.imag * r) * s, (a.imag - a.real * r) * s);
            }
        } else {
            double r = b.real / b.imag;
            double s = (double)(1.0) / (b.imag + b.real * r);
            return __pyx_t_double_complex_from_parts(
                (a.real * r + a.imag) * s, (a.imag * r - a.real) * s);
        }
    }
    #else
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        if (b.imag == 0) {
            return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else {
            double denom = b.real * b.real + b.imag * b.imag;
            return __pyx_t_double_complex_from_parts(
                (a.real * b.real + a.imag * b.imag) / denom,
                (a.imag * b.real - a.real * b.imag) / denom);
        }
    }
    #endif
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_neg_double(__pyx_t_double_complex a) {
        __pyx_t_double_complex z;
        z.real = -a.real;
        z.imag = -a.imag;
        return z;
    }
    static CYTHON_INLINE int __Pyx_c_is_zero_double(__pyx_t_double_complex a) {
       return (a.real == 0) && (a.imag == 0);
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_conj_double(__pyx_t_double_complex a) {
        __pyx_t_double_complex z;
        z.real =  a.real;
        z.imag = -a.imag;
        return z;
    }
    #if 1
        static CYTHON_INLINE double __Pyx_c_abs_double(__pyx_t_double_complex z) {
          #if !defined(HAVE_HYPOT) || defined(_MSC_VER)
            return sqrt(z.real*z.real + z.imag*z.imag);
          #else
            return hypot(z.real, z.imag);
          #endif
        }
        static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_pow_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
            __pyx_t_double_complex z;
            double r, lnr, theta, z_r, z_theta;
            if (b.imag == 0 && b.real == (int)b.real) {
                if (b.real < 0) {
                    double denom = a.real * a.real + a.imag * a.imag;
                    a.real = a.real / denom;
                    a.imag = -a.imag / denom;
                    b.real = -b.real;
                }
                switch ((int)b.real) {
                    case 0:
                        z.real = 1;
                        z.imag = 0;
                        return z;
                    case 1:
                        return a;
                    case 2:
                        z = __Pyx_c_prod_double(a, a);
                        return __Pyx_c_prod_double(a, a);
                    case 3:
                        z = __Pyx_c_prod_double(a, a);
                        return __Pyx_c_prod_double(z, a);
                    case 4:
                        z = __Pyx_c_prod_double(a, a);
                        return __Pyx_c_prod_double(z, z);
                }
            }
            if (a.imag == 0) {
                if (a.real == 0) {
                    return a;
                } else if (b.imag == 0) {
                    z.real = pow(a.real, b.real);
                    z.imag = 0;
                    return z;
                } else if (a.real > 0) {
                    r = a.real;
                    theta = 0;
                } else {
                    r = -a.real;
                    theta = atan2(0.0, -1.0);
                }
            } else {
                r = __Pyx_c_abs_double(a);
                theta = atan2(a.imag, a.real);
            }
            lnr = log(r);
            z_r = exp(lnr * b.real - theta * b.imag);
            z_theta = theta * b.real + lnr * b.imag;
            z.real = z_r * cos(z_theta);
            z.imag = z_r * sin(z_theta);
            return z;
        }
    #endif
#endif

/* CIntToPy */
  static CYTHON_INLINE PyObject* __Pyx_PyInt_From_enum__NPY_TYPES(enum NPY_TYPES value) {
    const enum NPY_TYPES neg_one = (enum NPY_TYPES) ((enum NPY_TYPES) 0 - (enum NPY_TYPES) 1), const_zero = (enum NPY_TYPES) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum NPY_TYPES) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(enum NPY_TYPES) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(enum NPY_TYPES) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum NPY_TYPES) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(enum NPY_TYPES) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(enum NPY_TYPES),
                                     little, !is_unsigned);
    }
}

/* MemviewSliceCopyTemplate */
  static __Pyx_memviewslice
__pyx_memoryview_copy_new_contig(const __Pyx_memviewslice *from_mvs,
                                 const char *mode, int ndim,
                                 size_t sizeof_dtype, int contig_flag,
                                 int dtype_is_object)
{
    __Pyx_RefNannyDeclarations
    int i;
    __Pyx_memviewslice new_mvs = { 0, 0, { 0 }, { 0 }, { 0 } };
    struct __pyx_memoryview_obj *from_memview = from_mvs->memview;
    Py_buffer *buf = &from_memview->view;
    PyObject *shape_tuple = NULL;
    PyObject *temp_int = NULL;
    struct __pyx_array_obj *array_obj = NULL;
    struct __pyx_memoryview_obj *memview_obj = NULL;
    __Pyx_RefNannySetupContext("__pyx_memoryview_copy_new_contig", 0);
    for (i = 0; i < ndim; i++) {
        if (from_mvs->suboffsets[i] >= 0) {
            PyErr_Format(PyExc_ValueError, "Cannot copy memoryview slice with "
                                           "indirect dimensions (axis %d)", i);
            goto fail;
        }
    }
    shape_tuple = PyTuple_New(ndim);
    if (unlikely(!shape_tuple)) {
        goto fail;
    }
    __Pyx_GOTREF(shape_tuple);
    for(i = 0; i < ndim; i++) {
        temp_int = PyInt_FromSsize_t(from_mvs->shape[i]);
        if(unlikely(!temp_int)) {
            goto fail;
        } else {
            PyTuple_SET_ITEM(shape_tuple, i, temp_int);
            temp_int = NULL;
        }
    }
    array_obj = __pyx_array_new(shape_tuple, sizeof_dtype, buf->format, (char *) mode, NULL);
    if (unlikely(!array_obj)) {
        goto fail;
    }
    __Pyx_GOTREF(array_obj);
    memview_obj = (struct __pyx_memoryview_obj *) __pyx_memoryview_new(
                                    (PyObject *) array_obj, contig_flag,
                                    dtype_is_object,
                                    from_mvs->memview->typeinfo);
    if (unlikely(!memview_obj))
        goto fail;
    if (unlikely(__Pyx_init_memviewslice(memview_obj, ndim, &new_mvs, 1) < 0))
        goto fail;
    if (unlikely(__pyx_memoryview_copy_contents(*from_mvs, new_mvs, ndim, ndim,
                                                dtype_is_object) < 0))
        goto fail;
    goto no_fail;
fail:
    __Pyx_XDECREF(new_mvs.memview);
    new_mvs.memview = NULL;
    new_mvs.data = NULL;
no_fail:
    __Pyx_XDECREF(shape_tuple);
    __Pyx_XDECREF(temp_int);
    __Pyx_XDECREF(array_obj);
    __Pyx_RefNannyFinishContext();
    return new_mvs;
}

/* CIntFromPy */
  static CYTHON_INLINE npy_int64 __Pyx_PyInt_As_npy_int64(PyObject *x) {
    const npy_int64 neg_one = (npy_int64) ((npy_int64) 0 - (npy_int64) 1), const_zero = (npy_int64) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(npy_int64) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(npy_int64, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (npy_int64) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (npy_int64) 0;
                case  1: __PYX_VERIFY_RETURN_INT(npy_int64, digit, digits[0])
                case 2:
                    if (8 * sizeof(npy_int64) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_int64, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_int64) >= 2 * PyLong_SHIFT) {
                            return (npy_int64) (((((npy_int64)digits[1]) << PyLong_SHIFT) | (npy_int64)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(npy_int64) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_int64, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_int64) >= 3 * PyLong_SHIFT) {
                            return (npy_int64) (((((((npy_int64)digits[2]) << PyLong_SHIFT) | (npy_int64)digits[1]) << PyLong_SHIFT) | (npy_int64)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(npy_int64) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_int64, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_int64) >= 4 * PyLong_SHIFT) {
                            return (npy_int64) (((((((((npy_int64)digits[3]) << PyLong_SHIFT) | (npy_int64)digits[2]) << PyLong_SHIFT) | (npy_int64)digits[1]) << PyLong_SHIFT) | (npy_int64)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (npy_int64) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(npy_int64) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(npy_int64, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(npy_int64) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(npy_int64, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (npy_int64) 0;
                case -1: __PYX_VERIFY_RETURN_INT(npy_int64, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(npy_int64,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(npy_int64) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_int64, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_int64) - 1 > 2 * PyLong_SHIFT) {
                            return (npy_int64) (((npy_int64)-1)*(((((npy_int64)digits[1]) << PyLong_SHIFT) | (npy_int64)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(npy_int64) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_int64, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_int64) - 1 > 2 * PyLong_SHIFT) {
                            return (npy_int64) ((((((npy_int64)digits[1]) << PyLong_SHIFT) | (npy_int64)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(npy_int64) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_int64, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_int64) - 1 > 3 * PyLong_SHIFT) {
                            return (npy_int64) (((npy_int64)-1)*(((((((npy_int64)digits[2]) << PyLong_SHIFT) | (npy_int64)digits[1]) << PyLong_SHIFT) | (npy_int64)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(npy_int64) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_int64, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_int64) - 1 > 3 * PyLong_SHIFT) {
                            return (npy_int64) ((((((((npy_int64)digits[2]) << PyLong_SHIFT) | (npy_int64)digits[1]) << PyLong_SHIFT) | (npy_int64)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(npy_int64) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_int64, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_int64) - 1 > 4 * PyLong_SHIFT) {
                            return (npy_int64) (((npy_int64)-1)*(((((((((npy_int64)digits[3]) << PyLong_SHIFT) | (npy_int64)digits[2]) << PyLong_SHIFT) | (npy_int64)digits[1]) << PyLong_SHIFT) | (npy_int64)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(npy_int64) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_int64, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_int64) - 1 > 4 * PyLong_SHIFT) {
                            return (npy_int64) ((((((((((npy_int64)digits[3]) << PyLong_SHIFT) | (npy_int64)digits[2]) << PyLong_SHIFT) | (npy_int64)digits[1]) << PyLong_SHIFT) | (npy_int64)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(npy_int64) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(npy_int64, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(npy_int64) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(npy_int64, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            npy_int64 val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (npy_int64) -1;
        }
    } else {
        npy_int64 val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (npy_int64) -1;
        val = __Pyx_PyInt_As_npy_int64(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to npy_int64");
    return (npy_int64) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to npy_int64");
    return (npy_int64) -1;
}

/* CIntFromPy */
  static CYTHON_INLINE npy_int32 __Pyx_PyInt_As_npy_int32(PyObject *x) {
    const npy_int32 neg_one = (npy_int32) ((npy_int32) 0 - (npy_int32) 1), const_zero = (npy_int32) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(npy_int32) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(npy_int32, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (npy_int32) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (npy_int32) 0;
                case  1: __PYX_VERIFY_RETURN_INT(npy_int32, digit, digits[0])
                case 2:
                    if (8 * sizeof(npy_int32) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_int32, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_int32) >= 2 * PyLong_SHIFT) {
                            return (npy_int32) (((((npy_int32)digits[1]) << PyLong_SHIFT) | (npy_int32)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(npy_int32) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_int32, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_int32) >= 3 * PyLong_SHIFT) {
                            return (npy_int32) (((((((npy_int32)digits[2]) << PyLong_SHIFT) | (npy_int32)digits[1]) << PyLong_SHIFT) | (npy_int32)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(npy_int32) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_int32, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_int32) >= 4 * PyLong_SHIFT) {
                            return (npy_int32) (((((((((npy_int32)digits[3]) << PyLong_SHIFT) | (npy_int32)digits[2]) << PyLong_SHIFT) | (npy_int32)digits[1]) << PyLong_SHIFT) | (npy_int32)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (npy_int32) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(npy_int32) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(npy_int32, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(npy_int32) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(npy_int32, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (npy_int32) 0;
                case -1: __PYX_VERIFY_RETURN_INT(npy_int32, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(npy_int32,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(npy_int32) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_int32, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_int32) - 1 > 2 * PyLong_SHIFT) {
                            return (npy_int32) (((npy_int32)-1)*(((((npy_int32)digits[1]) << PyLong_SHIFT) | (npy_int32)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(npy_int32) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_int32, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_int32) - 1 > 2 * PyLong_SHIFT) {
                            return (npy_int32) ((((((npy_int32)digits[1]) << PyLong_SHIFT) | (npy_int32)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(npy_int32) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_int32, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_int32) - 1 > 3 * PyLong_SHIFT) {
                            return (npy_int32) (((npy_int32)-1)*(((((((npy_int32)digits[2]) << PyLong_SHIFT) | (npy_int32)digits[1]) << PyLong_SHIFT) | (npy_int32)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(npy_int32) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_int32, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_int32) - 1 > 3 * PyLong_SHIFT) {
                            return (npy_int32) ((((((((npy_int32)digits[2]) << PyLong_SHIFT) | (npy_int32)digits[1]) << PyLong_SHIFT) | (npy_int32)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(npy_int32) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_int32, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_int32) - 1 > 4 * PyLong_SHIFT) {
                            return (npy_int32) (((npy_int32)-1)*(((((((((npy_int32)digits[3]) << PyLong_SHIFT) | (npy_int32)digits[2]) << PyLong_SHIFT) | (npy_int32)digits[1]) << PyLong_SHIFT) | (npy_int32)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(npy_int32) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_int32, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_int32) - 1 > 4 * PyLong_SHIFT) {
                            return (npy_int32) ((((((((((npy_int32)digits[3]) << PyLong_SHIFT) | (npy_int32)digits[2]) << PyLong_SHIFT) | (npy_int32)digits[1]) << PyLong_SHIFT) | (npy_int32)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(npy_int32) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(npy_int32, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(npy_int32) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(npy_int32, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            npy_int32 val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (npy_int32) -1;
        }
    } else {
        npy_int32 val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (npy_int32) -1;
        val = __Pyx_PyInt_As_npy_int32(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to npy_int32");
    return (npy_int32) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to npy_int32");
    return (npy_int32) -1;
}

/* CIntFromPy */
  static CYTHON_INLINE npy_uint64 __Pyx_PyInt_As_npy_uint64(PyObject *x) {
    const npy_uint64 neg_one = (npy_uint64) ((npy_uint64) 0 - (npy_uint64) 1), const_zero = (npy_uint64) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(npy_uint64) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(npy_uint64, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (npy_uint64) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (npy_uint64) 0;
                case  1: __PYX_VERIFY_RETURN_INT(npy_uint64, digit, digits[0])
                case 2:
                    if (8 * sizeof(npy_uint64) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_uint64, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_uint64) >= 2 * PyLong_SHIFT) {
                            return (npy_uint64) (((((npy_uint64)digits[1]) << PyLong_SHIFT) | (npy_uint64)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(npy_uint64) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_uint64, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_uint64) >= 3 * PyLong_SHIFT) {
                            return (npy_uint64) (((((((npy_uint64)digits[2]) << PyLong_SHIFT) | (npy_uint64)digits[1]) << PyLong_SHIFT) | (npy_uint64)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(npy_uint64) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_uint64, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_uint64) >= 4 * PyLong_SHIFT) {
                            return (npy_uint64) (((((((((npy_uint64)digits[3]) << PyLong_SHIFT) | (npy_uint64)digits[2]) << PyLong_SHIFT) | (npy_uint64)digits[1]) << PyLong_SHIFT) | (npy_uint64)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (npy_uint64) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(npy_uint64) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(npy_uint64, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(npy_uint64) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(npy_uint64, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (npy_uint64) 0;
                case -1: __PYX_VERIFY_RETURN_INT(npy_uint64, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(npy_uint64,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(npy_uint64) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_uint64, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_uint64) - 1 > 2 * PyLong_SHIFT) {
                            return (npy_uint64) (((npy_uint64)-1)*(((((npy_uint64)digits[1]) << PyLong_SHIFT) | (npy_uint64)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(npy_uint64) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_uint64, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_uint64) - 1 > 2 * PyLong_SHIFT) {
                            return (npy_uint64) ((((((npy_uint64)digits[1]) << PyLong_SHIFT) | (npy_uint64)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(npy_uint64) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_uint64, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_uint64) - 1 > 3 * PyLong_SHIFT) {
                            return (npy_uint64) (((npy_uint64)-1)*(((((((npy_uint64)digits[2]) << PyLong_SHIFT) | (npy_uint64)digits[1]) << PyLong_SHIFT) | (npy_uint64)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(npy_uint64) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_uint64, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_uint64) - 1 > 3 * PyLong_SHIFT) {
                            return (npy_uint64) ((((((((npy_uint64)digits[2]) << PyLong_SHIFT) | (npy_uint64)digits[1]) << PyLong_SHIFT) | (npy_uint64)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(npy_uint64) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_uint64, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_uint64) - 1 > 4 * PyLong_SHIFT) {
                            return (npy_uint64) (((npy_uint64)-1)*(((((((((npy_uint64)digits[3]) << PyLong_SHIFT) | (npy_uint64)digits[2]) << PyLong_SHIFT) | (npy_uint64)digits[1]) << PyLong_SHIFT) | (npy_uint64)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(npy_uint64) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_uint64, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_uint64) - 1 > 4 * PyLong_SHIFT) {
                            return (npy_uint64) ((((((((((npy_uint64)digits[3]) << PyLong_SHIFT) | (npy_uint64)digits[2]) << PyLong_SHIFT) | (npy_uint64)digits[1]) << PyLong_SHIFT) | (npy_uint64)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(npy_uint64) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(npy_uint64, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(npy_uint64) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(npy_uint64, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            npy_uint64 val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (npy_uint64) -1;
        }
    } else {
        npy_uint64 val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (npy_uint64) -1;
        val = __Pyx_PyInt_As_npy_uint64(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to npy_uint64");
    return (npy_uint64) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to npy_uint64");
    return (npy_uint64) -1;
}

/* CIntFromPy */
  static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
    const long neg_one = (long) ((long) 0 - (long) 1), const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(long) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case  1: __PYX_VERIFY_RETURN_INT(long, digit, digits[0])
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 2 * PyLong_SHIFT) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 3 * PyLong_SHIFT) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 4 * PyLong_SHIFT) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (long) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(long) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case -1: __PYX_VERIFY_RETURN_INT(long, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(long,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(long) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(long) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            long val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (long) -1;
        }
    } else {
        long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* CIntFromPy */
  static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
    const int neg_one = (int) ((int) 0 - (int) 1), const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(int) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int, digit, digits[0])
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 2 * PyLong_SHIFT) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 3 * PyLong_SHIFT) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 4 * PyLong_SHIFT) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(int) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(int,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(int) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(int) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            int val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int) -1;
        }
    } else {
        int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* BytesContains */
  static CYTHON_INLINE int __Pyx_BytesContains(PyObject* bytes, char character) {
    const Py_ssize_t length = PyBytes_GET_SIZE(bytes);
    char* char_start = PyBytes_AS_STRING(bytes);
    return memchr(char_start, (unsigned char)character, (size_t)length) != NULL;
}

/* ImportNumPyArray */
  static PyObject* __Pyx__ImportNumPyArray(void) {
    PyObject *numpy_module, *ndarray_object = NULL;
    numpy_module = __Pyx_Import(__pyx_n_s_numpy, NULL, 0);
    if (likely(numpy_module)) {
        ndarray_object = PyObject_GetAttrString(numpy_module, "ndarray");
        Py_DECREF(numpy_module);
    }
    if (unlikely(!ndarray_object)) {
        PyErr_Clear();
    }
    if (unlikely(!ndarray_object || !PyObject_TypeCheck(ndarray_object, &PyType_Type))) {
        Py_XDECREF(ndarray_object);
        Py_INCREF(Py_None);
        ndarray_object = Py_None;
    }
    return ndarray_object;
}
static CYTHON_INLINE PyObject* __Pyx_ImportNumPyArrayTypeIfAvailable(void) {
    if (unlikely(!__pyx_numpy_ndarray)) {
        __pyx_numpy_ndarray = __Pyx__ImportNumPyArray();
    }
    Py_INCREF(__pyx_numpy_ndarray);
    return __pyx_numpy_ndarray;
}

/* CIntFromPy */
  static CYTHON_INLINE char __Pyx_PyInt_As_char(PyObject *x) {
    const char neg_one = (char) ((char) 0 - (char) 1), const_zero = (char) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(char) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(char, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (char) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (char) 0;
                case  1: __PYX_VERIFY_RETURN_INT(char, digit, digits[0])
                case 2:
                    if (8 * sizeof(char) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) >= 2 * PyLong_SHIFT) {
                            return (char) (((((char)digits[1]) << PyLong_SHIFT) | (char)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(char) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) >= 3 * PyLong_SHIFT) {
                            return (char) (((((((char)digits[2]) << PyLong_SHIFT) | (char)digits[1]) << PyLong_SHIFT) | (char)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(char) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) >= 4 * PyLong_SHIFT) {
                            return (char) (((((((((char)digits[3]) << PyLong_SHIFT) | (char)digits[2]) << PyLong_SHIFT) | (char)digits[1]) << PyLong_SHIFT) | (char)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (char) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(char) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(char, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(char) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(char, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (char) 0;
                case -1: __PYX_VERIFY_RETURN_INT(char, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(char,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(char) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) - 1 > 2 * PyLong_SHIFT) {
                            return (char) (((char)-1)*(((((char)digits[1]) << PyLong_SHIFT) | (char)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(char) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) - 1 > 2 * PyLong_SHIFT) {
                            return (char) ((((((char)digits[1]) << PyLong_SHIFT) | (char)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(char) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) - 1 > 3 * PyLong_SHIFT) {
                            return (char) (((char)-1)*(((((((char)digits[2]) << PyLong_SHIFT) | (char)digits[1]) << PyLong_SHIFT) | (char)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(char) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) - 1 > 3 * PyLong_SHIFT) {
                            return (char) ((((((((char)digits[2]) << PyLong_SHIFT) | (char)digits[1]) << PyLong_SHIFT) | (char)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(char) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) - 1 > 4 * PyLong_SHIFT) {
                            return (char) (((char)-1)*(((((((((char)digits[3]) << PyLong_SHIFT) | (char)digits[2]) << PyLong_SHIFT) | (char)digits[1]) << PyLong_SHIFT) | (char)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(char) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) - 1 > 4 * PyLong_SHIFT) {
                            return (char) ((((((((((char)digits[3]) << PyLong_SHIFT) | (char)digits[2]) << PyLong_SHIFT) | (char)digits[1]) << PyLong_SHIFT) | (char)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(char) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(char, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(char) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(char, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            char val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (char) -1;
        }
    } else {
        char val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (char) -1;
        val = __Pyx_PyInt_As_char(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to char");
    return (char) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to char");
    return (char) -1;
}

/* CheckBinaryVersion */
  static int __Pyx_check_binary_version(void) {
    char ctversion[4], rtversion[4];
    PyOS_snprintf(ctversion, 4, "%d.%d", PY_MAJOR_VERSION, PY_MINOR_VERSION);
    PyOS_snprintf(rtversion, 4, "%s", Py_GetVersion());
    if (ctversion[0] != rtversion[0] || ctversion[2] != rtversion[2]) {
        char message[200];
        PyOS_snprintf(message, sizeof(message),
                      "compiletime version %s of module '%.100s' "
                      "does not match runtime version %s",
                      ctversion, __Pyx_MODULE_NAME, rtversion);
        return PyErr_WarnEx(NULL, message, 1);
    }
    return 0;
}

/* FunctionImport */
  #ifndef __PYX_HAVE_RT_ImportFunction
#define __PYX_HAVE_RT_ImportFunction
static int __Pyx_ImportFunction(PyObject *module, const char *funcname, void (**f)(void), const char *sig) {
    PyObject *d = 0;
    PyObject *cobj = 0;
    union {
        void (*fp)(void);
        void *p;
    } tmp;
    d = PyObject_GetAttrString(module, (char *)"__pyx_capi__");
    if (!d)
        goto bad;
    cobj = PyDict_GetItemString(d, funcname);
    if (!cobj) {
        PyErr_Format(PyExc_ImportError,
            "%.200s does not export expected C function %.200s",
                PyModule_GetName(module), funcname);
        goto bad;
    }
#if PY_VERSION_HEX >= 0x02070000
    if (!PyCapsule_IsValid(cobj, sig)) {
        PyErr_Format(PyExc_TypeError,
            "C function %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), funcname, sig, PyCapsule_GetName(cobj));
        goto bad;
    }
    tmp.p = PyCapsule_GetPointer(cobj, sig);
#else
    {const char *desc, *s1, *s2;
    desc = (const char *)PyCObject_GetDesc(cobj);
    if (!desc)
        goto bad;
    s1 = desc; s2 = sig;
    while (*s1 != '\0' && *s1 == *s2) { s1++; s2++; }
    if (*s1 != *s2) {
        PyErr_Format(PyExc_TypeError,
            "C function %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), funcname, sig, desc);
        goto bad;
    }
    tmp.p = PyCObject_AsVoidPtr(cobj);}
#endif
    *f = tmp.fp;
    if (!(*f))
        goto bad;
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(d);
    return -1;
}
#endif

/* InitStrings */
  static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION < 3
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        #else
        if (t->is_unicode | t->is_str) {
            if (t->intern) {
                *t->p = PyUnicode_InternFromString(t->s);
            } else if (t->encoding) {
                *t->p = PyUnicode_Decode(t->s, t->n - 1, t->encoding, NULL);
            } else {
                *t->p = PyUnicode_FromStringAndSize(t->s, t->n - 1);
            }
        } else {
            *t->p = PyBytes_FromStringAndSize(t->s, t->n - 1);
        }
        #endif
        if (!*t->p)
            return -1;
        if (PyObject_Hash(*t->p) == -1)
            return -1;
        ++t;
    }
    return 0;
}

static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    return __Pyx_PyUnicode_FromStringAndSize(c_str, (Py_ssize_t)strlen(c_str));
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
#if !CYTHON_PEP393_ENABLED
static const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    char* defenc_c;
    PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
    if (!defenc) return NULL;
    defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    {
        char* end = defenc_c + PyBytes_GET_SIZE(defenc);
        char* c;
        for (c = defenc_c; c < end; c++) {
            if ((unsigned char) (*c) >= 128) {
                PyUnicode_AsASCIIString(o);
                return NULL;
            }
        }
    }
#endif
    *length = PyBytes_GET_SIZE(defenc);
    return defenc_c;
}
#else
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
}
#endif
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject* x) {
    int retval;
    if (unlikely(!x)) return -1;
    retval = __Pyx_PyObject_IsTrue(x);
    Py_DECREF(x);
    return retval;
}
static PyObject* __Pyx_PyNumber_IntOrLongWrongResultType(PyObject* result, const char* type_name) {
#if PY_MAJOR_VERSION >= 3
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type %.200s).  "
                "The ability to return an instance of a strict subclass of int "
                "is deprecated, and may be removed in a future version of Python.",
                Py_TYPE(result)->tp_name)) {
            Py_DECREF(result);
            return NULL;
        }
        return result;
    }
#endif
    PyErr_Format(PyExc_TypeError,
                 "__%.4s__ returned non-%.4s (type %.200s)",
                 type_name, type_name, Py_TYPE(result)->tp_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_Check(x) || PyLong_Check(x)))
#else
  if (likely(PyLong_Check(x)))
#endif
    return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  #if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = m->nb_int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = m->nb_long(x);
  }
  #else
  if (likely(m && m->nb_int)) {
    name = "int";
    res = m->nb_int(x);
  }
  #endif
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
    res = PyNumber_Int(x);
  }
#endif
  if (likely(res)) {
#if PY_MAJOR_VERSION < 3
    if (unlikely(!PyInt_Check(res) && !PyLong_Check(res))) {
#else
    if (unlikely(!PyLong_CheckExact(res))) {
#endif
        return __Pyx_PyNumber_IntOrLongWrongResultType(res, name);
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(b);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    const digit* digits = ((PyLongObject*)b)->ob_digit;
    const Py_ssize_t size = Py_SIZE(b);
    if (likely(__Pyx_sst_abs(size) <= 1)) {
        ival = likely(size) ? digits[0] : 0;
        if (size == -1) ival = -ival;
        return ival;
    } else {
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return b ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


#endif /* Py_PYTHON_H */
